#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Color Table
    # We use a fixed color table that is always the same
    # This makes the output deterministic
    color_table = bytes([
        0xFF, 0xFF, 0xFF,  # White
        0x00, 0x00, 0x00,  # Black
        0xFF, 0x00, 0x00,  # Red
        0x00, 0xFF, 0x00,  # Green
        0x00, 0x00, 0xFF,  # Blue
        0xFF, 0xFF, 0x00,  # Yellow
        0x00, 0xFF, 0xFF,  # Cyan
        0xFF, 0x00, 0xFF   # Magenta
    ])
    
    # 4. Graphic Control Extension
    # Always have a delay of 0 seconds, transparent color index of 1, and no disposal method
    gce = struct.pack('>BBBBB', 0x21, 0xF9, 0x04, 0x00, 0x01)
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
