#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed fields
    #   Global Color Table Flag: 1
    #   Color Resolution: 7
    #   Sort Flag: 0
    #   Size of Global Color Table: 7
    packed_fields = 0b11100000 | (width & 0xff) << 7 | (height & 0xff) >> 1
    
    # Background color index
    background_color_index = rng.read(1)[0] % 256
    
    # Pixel aspect ratio
    pixel_aspect_ratio = 0
    
    # Write everything to the output stream
    out.write(header + struct.pack('<BBB', packed_fields, background_color_index, pixel_aspect_ratio))
    
    # 3. Global Color Table
    global_color_table = [struct.unpack('BBB', rng.read(3)) for _ in range(2 ** ((packed_fields >> 4) & 0b111))]
    for color in global_color_table:
        out.write(bytes(color))
        
    # 4. Image Descriptors
    while True:
        # Local Color Table Flag
        local_color_table_flag = bool(rng.read(1)[0] & 0b10000000)
        
        # Interlace Flag
        interlace_flag = bool(rng.read(1)[0] & 0b01000000)
        
        # Sort Flag
        sort_flag = bool(rng.read(1)[0] & 0b00100000)
        
        # Reserved
        reserved = rng.read(1)[0] & 0b00011111
        
        # Size of Local Color Table
        size_of_local_color_table = rng.read(1)[0] & 0b111
        
        # Packed Fields
        packed_fields = rng.read(1)[0]
        
        # Left Position
        left_position = rng.read(1)[0]
        
        # Top Position
        top_position = rng.read(1)[0]
        
        # Width
        width = rng.read(1)[0]
        
        # Height
        height = rng.read(1)[0]
        
        # Write everything to the output stream
        out.write(struct.pack('BBBBBBBB', local_color_table_flag, interlace_flag, sort_flag, reserved, size_of_local_color_table, packed_fields, left_position, top_position))
        out.write(struct.pack('BB', width, height))
        
        if not local_color_table_flag:
            break
            
        # Local Color Table
        local_color_table = [struct.unpack('BBB', rng.read(3)) for _ in range(2 ** size_of_local_color_table)]
        for color in local_color_table:
            out.write(bytes(color))
            
        # LZW Minimum Code Size
        lzw_minimum_code_size = rng.read(1)[0]
        
        # Write everything to the output stream
        out.write(lzw_minimum_code_size.to_bytes(1, 'big'))
        
        # Image Data
        image_data = []
        while True:
            code = int.from_bytes(rng.read(1), 'big')
            if code == 0:
                break
                
            image_data.append(code)
            
        # Write everything to the output stream
        out.write(bytes(image_data))
        
    # 5. Trailer
    trailer = b'\x3b'
