#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Image Descriptors
    # We will use one image descriptor with no local color table
    idd = struct.pack('>BBHHHBB', 0x2c, 0, 0, width, height, 0, 0)
    
    # 5. Graphic Control Extension
    # We will use a single frame with no disposal method, no user input, and an arbitrary delay time of 0-255 ms
    gce = struct.pack('>BBBH', 0xf9, 4, 0b00000000, rng.read(1)[0])
    
    # 6. Image Data
    # We will use a single data subblock with no interlacing, no local color table, and all pixels set to the background color
    data = struct.pack('>BB', 0, 0) + rng.read(width * height)
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything together
    out.write(header + lsd + color_table + idd + gce + data + trailer)
