#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table = rng.read(1)  # Whether to use a local color table
    
    # 6. Local Color Table (if applicable)
    if local_color_table == b'\x01':
        num_colors = (rng.read(1)[0] % 256) or 1
        color_table = rng.read(3 * num_colors)  # Random colors
        
    # 7. Image Data
    image_data = rng.read()
    
    # 8. Trailer
    trailer = b';'
    
    # Write everything together
    out.write(header + lsd + color_table + gce + b'\x2C' + struct.pack('>HH', left, top) + struct.pack('>HH', width, height) + local_color_table + color_table + image_data + trailer)

# Reads a GIF file from `inp`, and writes its contents to `out`.
# If `inp` is not a valid GIF file, raises an exception.
def read_and_write_gif(inp: BinaryIO, out: BinaryIO):
    # 1. Header
    header = inp.read(6)
    assert header == b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Width and Height
    width, height = struct.unpack('>HH', inp.read(4))
    
    # Other LSD fields: Packed fields, Background color index, Pixel aspect ratio
    _ = inp.read(7)
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (inp.read(1)[0] % 256) or 1
    color_table = inp.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = inp.read(8)
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', inp.read(4))
    width, height = struct.unpack('>HH', inp.read(4))
    local_color_table = inp.read(1)  # Whether to use a local color table
    
    # 6. Local Color Table (if applicable)
    if local_color_table == b'\x01':
        num_colors = (inp.read(1)[0] % 256) or 1
        color_table = inp.read(3 * num_colors)  # Random colors
    
    # 7. Image Data
    image_data = inp.read()
    
    # 8. Trailer
    trailer = inp.read()
    assert trailer == b';'
    
    # Write everything together
    out.write(header + struct.pack('>HH', width, height) + b'\x00\x00\xFF\xFF\xFF\x00\x00' + color_table + gce + b'\x2C' + struct.pack('>HH', left, top) + struct.pack('>HH', width, height) + local_color_table + color_table + image_data + trailer)
