#!/usr/bin/env python3

import struct
from typing import BinaryIO

def main():
    """
    Converts an RGB888 PNG into an LZW GIF
    """
    
    # Open input and output files
    infile = open('input.png', 'rb')
    outfile = open('output.gif', 'wb')
    
    convert(infile, outfile)
    
    infile.close()
    outfile.close()

def convert(rng: BinaryIO, out: BinaryIO):
    """
    Converts an RGB888 PNG into an LZW GIF
    """
    
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor (LSD)
    width = int.from_bytes(rng.read(4), byteorder='big')
    height = int.from_bytes(rng.read(4), byteorder='big')
    packed_fields = rng.read(1)[0]
    global_color_table_flag = bool((packed_fields & 0b10000000) >> 7)
    color_resolution = ((packed_fields & 0b01110000) >> 4) + 1
    sort_flag = bool((packed_fields & 0b00001000) >> 3)
    global_color_table_size = 2 ** ((packed_fields & 0b00000111) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('<HHBBBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    if global_color_table_flag:
        color_table = rng.read(global_color_table_size * 3)
    else:
        color_table = None
        
    # 4. Graphic Control Extension (GCE)
    gce = b'\x21\xF9\x04\x00\x00\x00\x00\xFF\xFF\xFF\x00'
    
    # 5. Image Descriptor
    image_descriptor = b'\x2C\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    
    # 6. Image Data
    lzw_minimum_code_size = 8
    total_blocks_size = int.from_bytes(rng.read(2), byteorder='big')
    blocks = []
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
