#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed fields
    packed_fields = bytearray([0xF7])
    
    # Background color index
    background_color_index = rng.read(1)[0] % 256
    
    # Pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)[0] % 256
    
    # Add everything to the logical screen descriptor
    ls_descriptor = bytearray()
    ls_descriptor += struct.pack('>H', width)
    ls_descriptor += struct.pack('>H', height)
    ls_descriptor += packed_fields
    ls_descriptor += struct.pack('B', background_color_index)
    ls_descriptor += struct.pack('B', pixel_aspect_ratio)
    
    # Write everything to the output
    out.write(header)
    out.write(ls_descriptor)
    
    # 3. Global Color Table
    global_color_table_size = 7 + rng.read(1)[0] % 256
    global_color_table = [struct.unpack('BBB', rng.read(3)) for _ in range(global_color_table_size)]
    out.write(bytes(global_color_table))
    
    # 4. Image Data
    while True:
        # Read the next block type
        block_type = rng.read(1)[0]
        
        if block_type == ord('!'):
            # Image Block
            
            # Read the image separator
            out.write(b'\x2C')
            
            # Read the image left position and top position
            out.write(rng.read(2))
            
            # Read the image width and height
            out.write(rng.read(2))
            
            # Packed fields
            packed_fields = bytearray([0xF0 | rng.read(1)[0] % 8])
            
            # Local color table size
            local_color_table_size = rng.read(1)[0] % 256
            
            # Interlaced?
            interlaced = bool(local_color_table_size & 0x8)
            local_color_table_size &= ~0x8
            
            # Sort flag
            sort_flag = bool(local_color_table_size & 0x4)
            local_color_table_size &= ~0x4
            
            # Reserved
            reserved = bool(local_color_table_size & 0x2)
            local_color_table_size &= ~0x2
            
            # Size of local color table
            local_color_table_size += 1
            
            # Add everything to the packed fields
            packed_fields += struct.pack('B', local_color_table_size)
            
            # LZW minimum code size
            lzw_minimum_code_size = rng.read(1)[0] % 256
            
            # Add everything to the packed fields
            packed_fields += struct.pack('B', lzw_minimum_code_size)
            
            # Write everything to the output
            out.write(packed_fields)
            
            # Local color table
            local_color_table = [struct.unpack('BBB', rng.read(3)) for _ in range(local_color_table_size)]
            out.write(bytes(local_color_table))
            
            # Image data
            while True:
                # Read the next code size
                code_size = rng.read(1)[0] % 256
                
                if code_size == 0:
                    # End of image data
                    break
                
                # Read the next code
                code = rng.read(code_size // 8)
                
                # Write it to the output
                out.write(code)
        elif block_type == ord(',') or block_type == ord(';'):
            # Extension Block
            
            # Read the extension label
            extension_label = rng.read(1)[0]
            
            if extension_label == ord('g'):
                # Graphic Control Extension
                
                # Read the extension block size
                extension_block_size = rng.read(1)[0]
                
                # Check that the extension block size is valid
                assert extension_block_size == 4
                
                # Packed fields
                packed_fields = bytearray([0xF9])
                
                # Disposal method
                disposal_method = rng.read(1)[0] % 8
                
                # User input flag
                user_input_flag = bool(disposal_method & 0x4)
                disposal_method &= ~0x4
                
                # Transparent color flag
                transparent_color_flag = bool(disposal_method & 0x2)
                disposal_method &= ~0x2
                
                # Reserved
                reserved = bool(disposal_method & 0x1)
                disposal_method &= ~0x1
                
                # Delay time
                delay_time = struct.unpack('>H', rng.read(2))[0]
                
                # Transparent color index
                transparent_color_index = rng.read(1)[0] % 256
                
