#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # GIF header
    out.write(b'GIF89a')
    
    # Logical screen descriptor
    width = 256 * (1 + rng.read(1)[0])
    height = 256 * (1 + rng.read(1)[0])
    background_color_index = rng.read(1)[0] & 0xF
    pixel_aspect_ratio = rng.read(1)[0] & 0x7
    out.write(struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio))
    
    # Global color table
    global_color_table_flag = True
    if rng.read(1)[0] == 0:
        global_color_table_flag = False
        
    global_color_table_size = 2 ** ((rng.read(1)[0] & 0x7) + 1)
    global_color_table_sort_flag = bool(rng.read(1)[0] & 0x80)
    global_color_table = []
    for i in range(global_color_table_size):
        red = rng.read(1)[0]
        green = rng.read(1)[0]
        blue = rng.read(1)[0]
        global_color_table.append((red, green, blue))
    
    # Image descriptor
    image_left = 0
    image_top = 0
    image_width = width
    image_height = height
    local_color_table_flag = False
    interlace_flag = False
    sort_flag = False
    local_color_table_size = 0
    out.write(struct.pack('<HHHBBB', image_left, image_top, image_width, image_height, local_color_table_flag, interlace_flag, sort_flag, local_color_table_size))
    
    # Local color table
    if not global_color_table_flag:
        for color in global_color_table:
            out.write(bytes([color[0], color[1], color[2]]))
    
    # LZW minimum code size
    lzw_minimum_code_size = 2 + rng.read(1)[0]
    out.write(lzw_minimum_code_size.to_bytes(1, 'big'))
    
    # Raster data
    clear_code = 1 << lzw_minimum_code_size
    end_of_information_code = clear_code + 1
    next_available_code = end_of_information_code + 1
    prefixes = {}
    suffixes = {}
    stack = []
    bits_remaining = 0
    current_byte = 0
    current_bit = 0
    while True:
        # Read a byte from the input stream
        if bits_remaining <= 0:
            current_byte = rng.read(1)[0]
            bits_remaining = 8
            
        # Get the next bit from the input stream
        current_bit = (current_byte >> (bits_remaining - 1)) & 1
        bits_remaining -= 1
        
        # If we have a complete code...
        if bits_remaining < lzw_minimum_code_size:
            # ...and it is the clear code...
            if current_code == clear_code:
                # ...reset everything and continue
                prefixes = {}
                suffixes = {}
                stack = []
                
                # Read a new byte from the input stream
                if bits_remaining <= 0:
                    current_byte = rng.read(1)[0]
                    bits_remaining = 8
                    
                # Get the next bit from the input stream
                current_bit = (current_byte >> (bits_remaining - 1)) & 1
                bits_remaining -= 1
                
                # If we have a complete code...
                if bits_remaining < lzw_minimum_code_size:
                    # ...and it is the end of information code...
                    if current_code == end_of_information_code:
                        # ...stop reading
                        break
                    
                    # Otherwise, this is an invalid code
                    else:
                        raise Exception("Invalid code")
            
            # ...otherwise, if it is the end of information code...
            elif current_code == end_of_information_code:
                # ...stop reading
                break
            
            # ...otherwise, if it is a valid code...
            elif current_code > end_of_information_code:
                # ...add it to the stack
                stack.append(current_code)
                
                # ...if it has no prefix...
                if current_code not in prefixes:
                    # ...it must be the first code
                    assert len(stack) == 1
                    
                    # ...so add it to the suffixes list
                    suffixes[current_code] = None
                    
                    # ...and increment the next available code
                    next_available_code += 1
                
                # ...otherwise, if it has a prefix...
                else:
                    # ...follow that prefix until there are no more
                    while current_code in prefixes:
                