#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        out: BinaryIO,
        rng: BinaryIO):
        
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor (LSD)
    width = int.from_bytes(rng.read(2), 'big') % 65536
    height = int.from_bytes(rng.read(2), 'big') % 65536
    packed = rng.read(1)[0]
    global_color_table = bool(packed & 0b10000000)
    color_resolution = (packed >> 4) & 0b111
    sort_flag = bool(packed & 0b100)
    background_color_index = packed & 0b1111
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('<HBBBBB', width, height, packed, background_color_index, pixel_aspect_ratio)
    
    # 3. Color Table
    if global_color_table:
        num_colors = 2 ** (color_resolution + 1)
        color_table = rng.read(num_colors * 3)
    else:
        color_table = None
    
    # 4. Graphic Control Extension (GCE)
    block_size = 4
    packed = rng.read(1)[0]
    disposal_method = (packed >> 2) & 0b111
    user_input_flag = bool(packed & 0b10)
    transparent_color_flag = bool(packed & 0b1)
    delay_time = int.from_bytes(rng.read(2), 'big')
    transparent_color_index = rng.read(1)[0]
    block_terminator = 0
    gce = struct.pack('<BBBBB', block_size, packed, delay_time, transparent_color_index, block_terminator)
    
    # 5. Image Descriptor
    left = int.from_bytes(rng.read(2), 'big') % 65536
    top = int.from_bytes(rng.read(2), 'big') % 65536
    width = int.from_bytes(rng.read(2), 'big') % 65536
    height = int.from_bytes(rng.read(2), 'big') % 65536
    local_color_table = bool(rng.read(1)[0] & 0b10000000)
    interlace_flag = bool(rng.read(1)[0] & 0b100)
    sort_flag = bool(rng.read(1)[0] & 0b10)
    reserved = rng.read(1)[0] & 0b1111
    lct_size = 0
    if local_color_table:
        lct_size = 2 ** ((rng.read(1)[0] & 0b111) + 1)
    image_descriptor = struct.pack('<HHBBBBB', left, top, width, height, packed, reserved)
    
    # 6. Image Data
    image_data = rng.read(width * height)
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
