#!/usr/bin/env python3

import struct
from typing import BinaryIO, Optional, Tuple, Union

from .constants import (
    BLOCK_SIZE,
    CHUNK_HEADER_SIZE,
    FILE_HEADER_SIZE,
    HEADER_MAGIC,
    MAX_CHUNKS,
)
from .errors import InvalidHeaderError


def read_header(f: BinaryIO) -> dict:
    """Read the header from a file.

    Args:
        f (BinaryIO): File-like object to read from.

    Returns:
        dict: Dictionary containing the header data.
    """
    # Read the first 8 bytes of the file and check if it matches the magic number
    header = f.read(FILE_HEADER_SIZE)
    if len(header) != FILE_HEADER_SIZE or header[:4] != HEADER_MAGIC:
        raise InvalidHeaderError("Invalid header")

    return {
        "version": struct.unpack("<I", header[4:8])[0],
        "num_chunks": struct.unpack("<H", header[8:10])[0],
        "chunk_size": struct.unpack("<H", header[10:12])[0],
    }


def write_header(f: BinaryIO, version: int, num_chunks: int, chunk_size: int):
    """Write the header to a file.

    Args:
        f (BinaryIO): File-like object to write to.
        version (int): Version number for this file format.
        num_chunks (int): Number of chunks in this file.
        chunk_size (int): Size of each chunk in bytes.
    """
    header = bytearray(FILE_HEADER_SIZE)
    header[:4] = HEADER_MAGIC
    struct.pack_into("<I", header, 4, version)
    struct.pack_into("<H", header, 8, num_chunks)
    struct.pack_into("<H", header, 10, chunk_size)
    f.write(bytes(header))


def read_chunk_header(f: BinaryIO) -> dict:
    """Read the chunk header from a file.

    Args:
        f (BinaryIO): File-like object to read from.

    Returns:
        dict: Dictionary containing the chunk header data.
    """
    header = f.read(CHUNK_HEADER_SIZE)
    if len(header) != CHUNK_HEADER_SIZE:
        raise InvalidHeaderError("Invalid chunk header")

    return {
        "offset": struct.unpack("<Q", header[:8])[0],
        "length": struct.unpack("<I", header[8:12])[0],
        "compressed_length": struct.unpack("<I", header[12:16])[0],
    }


def write_chunk_header(f: BinaryIO, offset: int, length: int, compressed_length: int):
    """Write the chunk header to a file.

    Args:
        f (BinaryIO): File-like object to write to.
        offset (int): Offset of the chunk in the file.
        length (int): Length of the chunk in bytes.
        compressed_length (int): Compressed length of the chunk in bytes.
    """
    header = bytearray(CHUNK_HEADER_SIZE)
    struct.pack_into("<Q", header, 0, offset)
    struct.pack_into("<I", header, 8, length)
    struct.pack_into("<I", header, 12, compressed_length)
    f.write(bytes(header))


def get_blocks(
    f: BinaryIO,
    start_index: int,
    end_index: Optional[int] = None,
    block_size: int = BLOCK_SIZE,
) -> Tuple[Union[bytearray, bytes]]:
    """Get blocks from a file.

    Args:
        f (BinaryIO): File-like object to read from.
        start_index (int): Index of the first block to read.
        end_index (Optional[int]): Index of the last block to read. If not specified, all blocks will be read. Defaults to None.
        block_size (int, optional): Block size in bytes. Defaults to BLOCK_SIZE.

    Raises:
        ValueError: If `start_index` is less than zero.
        ValueError: If `end_index` is greater than the maximum number of blocks allowed.
        ValueError: If `end_index` is less than `start_index`.

    Returns:
        Tuple[Union[bytearray, bytes]]: A tuple of blocks.
    """
    if start_index < 0:
        raise ValueError("Start index cannot be negative")

    if end_index is None:
        end_index = MAX_CHUNKS - 1

    if end_index > MAX_CHUNKS - 1:
        raise ValueError("End index cannot exceed the maximum number of blocks")

    if end_index < start_index:
        raise ValueError("End index must be greater than or equal to the start index")

    blocks = []
    for i in range(start_index, end_index + 1):
