#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# algorithm described here: https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor (LSD)
    width = rng.read(2)
    height = rng.read(2)
    packed_fields = rng.read(1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)
    lsd = width + height + packed_fields + background_color_index + pixel_aspect_ratio
    
    # 3. Global Color Table
    global_color_table_flag = packed_fields[0] & 0b10000000
    if global_color_table_flag == 0:
        raise Exception("Global color table flag must be set")
    color_resolution = (packed_fields[0] >> 4) & 0b11
    sort_flag = packed_fields[0] & 0b100
    global_color_table_size = 2 ** ((packed_fields[0] & 0b111) + 1)
    color_table = rng.read(global_color_table_size * 3)
    
    # 4. Graphic Control Extension (GCE)
    gce = b'\x21\xF9\x04\x08\x00\x00\xFF\xFF\xFF\xFF\x00'
    
    # 5. Image Descriptor
    image_descriptor = b'\x2C\x00\x00\x00\x00\x00\x00\x00\x00'
    
    # 6. Image Data
    lzw_minimum_code_size = rng.read(1)[0]
    total_blocks_size = rng.read(2)
    blocks = []
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
