#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write(b'GIF89a')
    
    # Logical screen descriptor
    width = 16 + (next_byte(rng) % 256)
    height = 16 + (next_byte(rng) % 256)
    out.write(struct.pack('<HHBB', width, height, 0xF7, 0)) # 16-bit color table flag set
    
    # Global color table
    for i in range(4):
        out.write(struct.pack('<B', next_byte(rng)))
        out.write(struct.pack('<B', next_byte(rng)))
        out.write(struct.pack('<B', next_byte(rng)))
        
    # Application extension
    out.write(b'\x21\xFF\x0BNETSCAPE2.0\x03\x01')
    out.write(struct.pack('<H', 0)) # Repeat count
    out.write(b'\x00') # Block terminator
    
    # Image descriptor
    out.write(b'\x2C\x00\x00\x00\x00')
    out.write(struct.pack('<HH', width, height))
    out.write(b'\x00') # No local color table
    
    # LZW minimum code size
    out.write(b'\x08')
    
    # Pixel data
    def write_code(code: int):
        if code == 0x100:
            out.write(b'\x00')
        elif code < 0x100:
            out.write(struct.pack('<B', code))
        else:
            write_code((code >> 8) & 0xFF)
            write_code(code & 0xFF)
            
    def write_pixels():
        while True:
            pixel = next_byte(rng) % 4
            yield pixel
                
    def lzw_encode(data: bytes):
        dictionary = {}
        w = 8
        code = 0
        for b in data:
            if b not in dictionary:
                dictionary[b] = code
                code += 1
                if code >= 1 << w:
                    w += 1
            write_code(dictionary[b])
        write_code(0x100)
        write_code(0x101)
        
    lzw_encode(bytes(write_pixels()))
    
    # Block terminator
    out.write(b'\x00')

def next_byte(rng: BinaryIO):
    return ord(rng.read(1))
