#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = struct.pack('B', rng.read(1)[0])
    background_color_index = struct.pack('B', rng.read(1)[0])
    pixel_aspect_ratio = struct.pack('B', rng.read(1)[0])
    
    # Assemble the LSD
    lsd = struct.pack('<6sHBBBB', header, width, height, *struct.unpack('BBB', packed_fields))
    
    # 3. Global Color Table
    # Random number of colors in the table (up to 256)
    num_colors = struct.unpack('B', rng.read(1))[0] + 1
    
    # Random color table data (each color is three bytes)
    color_table = rng.read(num_colors*3)
    
    # 4. Graphic Control Extension
    # Random delay time (two bytes)
    delay_time = struct.pack('>H', struct.unpack('>H', rng.read(2))[0])
    
    # Random transparent color index (one byte)
    transparent_color_index = struct.pack('B', rng.read(1)[0])
    
    # Packed field (one byte)
    packed_field = struct.pack('B', rng.read(1)[0])
    
    # Assemble the GCE
    gce = struct.pack('<BBBBB', 0xF9, 0x04, *struct.unpack('BBB', delay_time+transparent_color_index+packed_field))
    
    # 5. Image Descriptor
    # Random left position (two bytes)
    left_position = struct.pack('>H', struct.unpack('>H', rng.read(2))[0])
    
    # Random top position (two bytes)
    top_position = struct.pack('>H', struct.unpack('>H', rng.read(2))[0])
    
    # Width and height (two bytes each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed field (one byte)
    packed_field = struct.pack('B', rng.read(1)[0])
    
    # Assemble the ID
    image_descriptor = struct.pack('<BBHHBB', 0x2C, *struct.unpack('HHBB', left_position+top_position+width+height+packed_field))
    
    # 6. Local Color Table
    # Same as global color table
    local_color_table = color_table
    
    # 7. Image Data
    # Random length (up to 65535 bytes)
    length = struct.pack('>H', struct.unpack('>H', rng.read(2))[0])
    
    # Random image data (of the given length)
    image_data = rng.read(struct.unpack('>H', length)[0])
    
    # 8. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
