#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data
    # We'll use LZW compression, which is pretty easy to implement
    # First we need to create our dictionary of codes
    code_dictionary = {bytes([i]): bytes([i]) for i in range(num_colors)}
    code_dictionary[b''] = b''  # Empty string has code 0
    
    # Now we can start encoding!
    blocks = []
    current_string = b''
    while True:
        # Pick a new character from the input stream
        next_character = rng.read(1)
        
        # If we hit EOF, stop
        if not next_character:
            break
        
        # Add this character to the current string
        current_string += next_character
        
        # Check if the current string already exists in the dictionary
        if current_string in code_dictionary:
            continue
        
        # If it doesn't exist, add it to the dictionary
        code_dictionary[current_string] = code_dictionary[current_string[:-1]] + current_string[-1:]
        
        # And write the code for the previous string to the output
        blocks.append(struct.pack('>H', len(code_dictionary[current_string[:-1]]) - 1))
        
        # If we've reached the maximum code size, reset the dictionary
        if len(code_dictionary) == 2**(8+1)-1:
            code_dictionary = {bytes([i]): bytes([i]) for i in range(num_colors)}
            code_dictionary[b''] = b''
            
    # The minimum code size is the log base 2 of the number of colors
    lzw_minimum_code_size = int(round(math.log2(num_colors)))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
