#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(1), 'big') % (65536 - 100 + 1) + 100
    height = int.from_bytes(rng.read(1), 'big') % (65536 - 100 + 1) + 100

    flags = bytes([
        0b10000000 | # Global Color Table Flag
        ((width >> 7) & 0b00000111) << 4 | # Size of global color table
        (((height >> 7) & 0b00000111)) # Size of global color table
    ])

    bg_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]

    logical_screen_descriptor = header \
            + width.to_bytes(2, byteorder='big', signed=False) \
            + height.to_bytes(2, byteorder='big', signed=False) \
            + flags \
            + bg_color.to_bytes(1, byteorder='big', signed=False) \
            + aspect_ratio.to_bytes(1, byteorder='big', signed=False)

    # 3. Global Color Table
    global_color_table = []
    for i in range(int((flags[0] & 0x07) * 3)):
        global_color_table += [rng.read(1)[0]]

    # 4. Image Descriptor
    image_separator = b'\x2C'
    left = 0
    top = 0
    width = 100
    height = 100
    flags = bytes([
        0b10000000 | # Local Color Table Flag
        0b00000000 | # Interlace Flag
        0b00000000 | # Sort Flag
        0b00000000   # Reserved
    ])

    local_color_table_size = 0
    local_color_table = None

    image_descriptor = image_separator \
            + left.to_bytes(2, byteorder='big', signed=False) \
            + top.to_bytes(2, byteorder='big', signed=False) \
            + width.to_bytes(2, byteorder='big', signed=False) \
            + height.to_bytes(2, byteorder='big', signed=False) \
            + flags \
            + local_color_table_size.to_bytes(1, byteorder='big', signed=False)

    # 5. Image Data
    lzw_min_code_size = rng.read(1)[0]
    lzw_max_code_size = rng.read(1)[0]

    data = []
    while True:
        code_size = rng.read(1)[0]

        if code_size == 128:
            break
        
        data += [(code_size,)]

        while len(data[-1]) < code_size:
            data[-1] += [rng.read(1)[0]]

    # 6. Trailer
    trailer = b'\x3B'

    # Write everything to output
    out.write(logical_screen_descriptor)
    out.write(struct.pack('B', len(global_color_table)))
    for c in global_color_table:
        out.write(c.to_bytes(1, byteorder='big'))

    out.write(image_descriptor)
    out.write(lzw_min_code_size.to_bytes(1, byteorder='big', signed=False))
    out.write(lzw_max_code_size.to_bytes(1, byteorder='big', signed=False))

    for d in data:
        if isinstance(d, tuple):
            out.write(d[0].to_bytes(1, byteorder='big', signed=False))
        else:
            out.write(d.to_bytes(1, byteorder='big', signed=False))

    out.write(trailer)

