#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Random color resolution
    color_resol = ord(rng.read(1)) & 0b111

    # Global Color Table Flag
    gct_flag = bool(ord(rng.read(1)))

    # Sort flag
    sort_flag = bool(ord(rng.read(1)))

    # Size of global color table
    if not gct_flag:
        gct_size = 0
    else:
        gct_size = 2 ** ((color_resol + 1) * 3)

    # Background color index
    bg_index = ord(rng.read(1))

    # Pixel aspect ratio
    pixel_aspect_ratio = ord(rng.read(1))

    # Write logical screen descriptor to output stream
    out.write(header)
    out.write(struct.pack('>HHBBB',
                          width, height,
                          color_resol,
                          7 << sort_flag | 1 << gct_flag,
                          bg_index,
                          pixel_aspect_ratio
                         ))

    # If there is a global color table...
    if gct_flag:
        #... write it to the output stream
        out.write(rng.read(gct_size * 3))

    # 3. Image Descriptor
    # Frame left position
    frame_left = 0

    # Frame top position
    frame_top = 0

    # Width of image in pixels
    img_width = width - frame_left

    # Height of image in pixels
    img_height = height - frame_top

    # Packed field
    packed_field = 0b10000000 | 0b00001000 | 0b00000000

    # LZW minimum code size
    lzw_min_code_size = 2

    # Write image descriptor to output stream
    out.write(struct.pack('>HHHHBB',
                          frame_left, frame_top,
                          img_width, img_height,
                          packed_field,
                          lzw_min_code_size
                         ))

    # 4. Graphics Control Extension
    # Block size
    block_size = 4

    # Packed field
    packed_field = 0b10000000 | 0b00000000 | 0b00000000

    # Delay time in hundredths of seconds
    delay_time = 0

    # Transparent color index
    trans_index = 0

    # Terminator
    terminator = 0x00

    # Write graphics control extension to output stream
    out.write(struct.pack('<BHBHHB',
                          block_size,
                          packed_field,
                          delay_time,
                          trans_index,
                          terminator
                         ))

    # 5. Plain Text Extension
    # Block size
    block_size = 12

    # Text grid left position
    text_grid_left = 0

    # Text grid top position
    text_grid_top = 0

    # Text grid width
    text_grid_width = width - text_grid_left

    # Text grid height
    text_grid_height = height - text_grid_top

    # Cell width
    cell_width = 0

    # Cell height
    cell_height = 0

    # FONT SIZE
    font_size = 0

    # Write plain text extension to output stream
    out.write(struct.pack('<BBBBHHBBBBB',
                          block_size,
                          text_grid_left,
                          text_grid_top,
                          text_grid_width,
                          text_grid_height,
                          cell_width,
                          cell_height,
                          font_size,
                          0x00,
                          0x00
                         ))

    # 6. Application Extension
    # Block size
    block_size = 11

    # Application identifier
    app_id = 'NETSCAPE'.encode()

    # Application authentication code
    auth_code = b'\x03\x01'

    # Subblock size
    subblock_size = len(app_id) + len(auth_code) + 3

    # Subblock ID
    subblock_id = 0x01

    # Netscape Looping Extension Code
    netscape_loop_ext_code = 0xff

    # Subblock data length
    subblock_data_len = 0x0b

    # Loop count (number of loops; zero means infinite loop)
    loop_count = 0

    # Write application extension to output stream
    out.write(struct.pack('<BB%dsBHBHBB' % len(app_id),
                          block_size,
                          subblock_size,
                          app_id,
                          subblock_id,
                          netscape_loop_ext_code,
                          subblock_data_len,
                          loop_count,
                          0x00,
                          0x00
                         ))

    # 7. Comment Extension
    # Block size
    block_size = 2

    # Comment length
    comment_length = 0

    # Write comment extension to output stream
    out.write(struct.pack('<BB',
                          block_size,
                          comment_length
                         ))

    # 8. Image Data
    # LZW minimum code size
    lzw_min_code_size = 2

    # Clear code
    clear_code = 1 << lzw_min_code_size

    # End of information code
    end_of_info_code = clear_code + 1

    # First byte after image data
    first_byte_after_image_data = end_of_info_code + 1

    # Write LZW minimum code size to output stream
    out.write(bytes([lzw_min_code_size]))

    # Write codes to output stream
    out.write(bytes([clear_code,
                     end_of_info_code,
                     0x00, 0x00, 0x00, 0x00
                    ]))

    # Write first byte after image data to output