#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    out.write("GIF89a".encode('ascii'))

    width = 256 + int.from_bytes(rng.read(1), byteorder='little') % (2**7 - 256)
    height = 256 + int.from_bytes(rng.read(1), byteorder='little') % (2**7 - 256)

    flags = b'\x00' * 4

    global_color_table_size = 2 ** ((int.from_bytes(rng.read(1), byteorder='little') & 0b111) + 1)
    background_index = int.from_bytes(rng.read(1), byteorder='little') % global_color_table_size
    aspect_ratio = int.from_bytes(rng.read(1), byteorder='little') % 4

    out.write(struct.pack("<HHBBB", width, height, global_color_table_size,
                          background_index, aspect_ratio))

    for i in range(global_color_table_size):
        r = int.from_bytes(rng.read(1), byteorder='little') % 256
        g = int.from_bytes(rng.read(1), byteorder='little') % 256
        b = int.from_bytes(rng.read(1), byteorder='little') % 256

        out.write(struct.pack("<BBBB", r, g, b, 0xff))

    lzw_min_code_size = int.from_bytes(rng.read(1), byteorder='little') % 129

    out.write(lzw_min_code_size.to_bytes(1, 'little', signed=False))

    # Image descriptor
    left = int.from_bytes(rng.read(2), byteorder='little') % (2**16 - 1)
    top = int.from_bytes(rng.read(2), byteorder='little') % (2**16 - 1)
    width = 256 + int.from_bytes(rng.read(2), byteorder='little') % (2**16 - 256)
    height = 256 + int.from_bytes(rng.read(2), byteorder='little') % (2**16 - 256)

    packed = (left << 4) | (top >> 8)
    packed |= (((top & 0xff) << 4) | (width >> 8)).to_bytes(1, 'big', signed=False)
    packed += (((width & 0xff) << 4) | (height >> 8)).to_bytes(1, 'big', signed=False)
    packed += (height & 0xff).to_bytes(1, 'big', signed=False)

    flags = bytes([packed[0] & 0xf7]) + bytes([(packed[0] & 0x08) | (packed[1] >> 4)])
    packed = packed[2:]

    local_color_table_flag = bool((packed[0] & 0x80) >> 7)
    interlace_flag = bool((packed[0] & 0x40) >> 6)
    sort_flag = bool((packed[0] & 0x20) >> 5)
    reserved = (packed[0] & 0x18) >> 3
    size_of_local_color_table = 2 ** ((packed[0] & 0x07) + 1)
    packed = packed[1:]

    if local_color_table_flag:
        for i in range(size_of_local_color_table):
            r = int.from_bytes(rng.read(1), byteorder='little') % 256
            g = int.from_bytes(rng.read(1), byteorder='little') % 256
            b = int.from_bytes(rng.read(1), byteorder='little') % 256

            out.write(struct.pack("<BBBB", r, g, b, 0xff))

        packed = packed[3 * size_of_local_color_table:]

    lzw_min_code_size = int.from_bytes(rng.read(1), byteorder='little') % 129

    out.write(flags + packed + lzw_min_code_size.to_bytes(1, 'little', signed=False))

    # LZW code stream
    while True:
        next_byte = rng.read(1)[0]
        if next_byte == 0x00 or next_byte == 0xff:
            continue

        out.write(next_byte.to_bytes(1, 'little', signed=False))

