#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
     # 2. Logical screen descriptor
    width = int.from_bytes(rng.read(2), 'big') % (500 - 40 + 1) + 40
    height = int.from_bytes(rng.read(2), 'big') % (500 - 40 + 1) + 40

    global_color_table_flag = bool(int.from_bytes(rng.read(1), 'big'))
    color_resolution = int.from_bytes(rng.read(1), 'big') & 7
    sort_flag = bool(int.from_bytes(rng.read(1), 'big'))
    size_of_global_color_table = int.from_bytes(rng.read(1), 'big') & 7
    bg_color_index = int.from_bytes(rng.read(1), 'big')
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big')
    logical_screen_descriptor = struct.pack('>HHBBB', width, height,
            global_color_table_flag << 7 | color_resolution << 4 | sort_flag << 3 | size_of_global_color_table,
            bg_color_index, pixel_aspect_ratio)

    # 3. Global color table
    if global_color_table_flag:
        global_color_table = bytes([
                *rng.read((size_of_global_color_table ** 2) * 3)])

        while len(global_color_table) < ((size_of_global_color_table ** 2) * 3):
            global_color_table += bytes([
                    *(rng.read(3))])

    else:
        global_color_table = None

    # 4. Image descriptor
    image_left_position = int.from_bytes(rng.read(2), 'big')
    image_top_position = int.from_bytes(rng.read(2), 'big')
    image_width = int.from_bytes(rng.read(2), 'big') % (500 - image_left_position + 1)
    image_height = int.from_bytes(rng.read(2), 'big') % (500 - image_top_position + 1)
    local_color_table_flag = bool(int.from_bytes(rng.read(1), 'big'))
    interlace_flag = bool(int.from_bytes(rng.read(1), 'big'))
    sorted_color_table_flag = bool(int.from_bytes(rng.read(1), 'big'))
    reserved_bits = int.from_bytes(rng.read(1), 'big') & 7
    size_of_local_color_table = int.from_bytes(rng.read(1), 'big') & 7
    local_color_table = None

    if local_color_table_flag:
        local_color_table = bytes([
                *rng.read((size_of_local_color_table ** 2) * 3)])

        while len(local_color_table) < ((size_of_local_color_table ** 2) * 3):
            local_color_table += bytes([
                    *(rng.read(3))])

    packed_fields = (local_color_table_flag << 7 | interlace_flag << 6 | sorted_color_table_flag << 5 | reserved_bits << 3 | size_of_local_color_table)
    image_descriptor = struct.pack('>HHHHBB', image_left_position, image_top_position,
            image_width, image_height, packed_fields, 0)

    # 5. LZW minimum code size
    lzw_minimum_code_size = int.from_bytes(rng.read(1), 'big') & 0b11111

    # 6. Data sub-blocks
    data_sub_block = bytearray()
    data_sub_block_length = 0
    end_of_image_flag = False

    while not end_of_image_flag:
        data_sub_block_byte = int.from_bytes(rng.read(1), 'big')

        if data_sub_block_byte == 0x00:
            end_of_image_flag = True

            continue

        elif data_sub_block_byte >= 0x01 and data_sub_block_byte <= 0xFE:
            data_sub_block_length += 1
            data_sub_block.append(data_sub_block_byte)

        elif data_sub_block_byte == 0xFF:
            data_sub_block_length += 1
            data_sub_block.extend(struct.pack('>H',
                    int.from_bytes(rng.read(2), 'big')))

        else:
            raise Exception("Invalid data sub block")

    data_sub_block_length += 1
    data_sub_block[0] |= 0x80
    data_sub_block[-1] |= 0x00

    # 7. Trailer
    trailer = b'\x3B'

    # Write to output stream
    out.write(header)
    out.write(logical_screen_descriptor)

    if global_color_table is not None:
        out.write(global_color_table)

    out.write(image_descriptor)

    if local_color_table is not None:
        out.write(local_color_table)

    out.write(bytes([lzw_minimum_code_size]))
    out.write(bytes([*data_sub_block]))
    out.write(trailer)

