#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
        
    # 6. Image Data
    data = []
    while True:
        block_type = rng.read(1)[0]
        
        if block_type == 0x21:
            # Extension Block
            
            ext_block_size = ord(rng.read(1))
            
            if ext_block_size!= 0:
                raise Exception("Extension blocks must be empty")
                
            ext_block_label = ord(rng.read(1))
            
            if ext_block_label in [0xF9]:
                # Graphics Control Extension
                
                # Discard the rest of this extension block
                rng.seek(ext_block_size - 1, 1)
            elif ext_block_label in [0xFE, 0xFF]:
                # Comment Extension or Application Extension
                
                # Discard the rest of this extension block
                rng.seek(ext_block_size - 1, 1)
            else:
                raise Exception("Unknown extension label {}".format(ext_block_label))
                
        elif block_type == 0x2C:
            # Image Descriptor
            
            # Discard the rest of this image descriptor
            rng.seek(10, 1)
            
        elif block_type == 0x3B:
            break;
        else:
            raise Exception("Unknown block type {}".format(block_type))
            
    # 7. Trailer
    trailer = b';'

    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)

