#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = rng.read(1)[0] % 64 + 1
    h = rng.read(1)[0] % 64 + 1

    # Color table size is always 7 bits of the first byte
    color_table_size = rng.read(1)[0] & 0x7f

    # Background color index
    background_color_index = rng.read(1)[0]

    # Pixel aspect ratio - fixed to 0
    pixel_aspect_ratio = 0

    # Write logical screen descriptor
    out.write(header)
    out.write(struct.pack('<HHBBB', w, h, color_table_size,
                          background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table
    global_color_table = [b'\xff\xff\xff'] * (1 << color_table_size)

    # Write global color table
    for i in range(len(global_color_table)):
        if len(global_color_table[i]) == 3:
            global_color_table[i] += b'\x00'

        out.write(global_color_table[i])

    # 4. Image Descriptor
    image_descriptor = struct.pack('<BHHBB',
                                   ord('!'),
                                   0,
                                   w,
                                   h,
                                   0)

    # Write image descriptor
    out.write(image_descriptor)

    # 5. LZW minimum code size
    lzw_minimum_code_size = 2

    # Write LZW minimum code size
    out.write(bytes([lzw_minimum_code_size]))

    # 6. Data sub-blocks
    data_sub_blocks = []

    # Generate random data until we have at least one full frame worth
    while True:
        # Generate a random block length between 2 and 128 inclusive
        block_length = rng.read(1)[0] % 127 + 2
        
        # Read that many bytes from /dev/urandom
        data = rng.read(block_length)

        # If this is not enough data to fill up a frame, keep going
        if len(data) < w * h: continue

        # Otherwise, break out of loop
        else: break

    # Split our data into chunks of 255 bytes or less
    while len(data):
        chunk = data[:255]
        del data[:255]

        # Add the chunk to our list of data blocks
        data_sub_blocks.append(chunk)

    # Encode all of our data as LZW
    encoded_data = encode_lzw(data_sub_blocks)

    # Write encoded data
    out.write(encoded_data)

    # 7. Trailer
    trailer = b';'

    # Write trailer
    out.write(trailer)


# Encodes a series of strings as LZW
def encode_lzw(strings: list) -> bytes:
    # Create an empty dictionary with no entries
    dictionary = {}

    # Initialize the current string to be the empty string
    current_string = ''

    # Initialize the previous string to be the empty string
    previous_string = ''

    # Initialize the code size to 9 bits
    code_size = 9

    # Initialize the clear code to 256
    clear_code = 256

    # Initialize the end code to 257
    end_code = 257

    # Initialize the next available code to 258
    next_available_code = 258

    # Initialize the output buffer to be empty
    output_buffer = b''

    # For every string...
    for s in strings:
        #...encode it as LZW...
        encoded_s = encode_as_lzw(dictionary,
                                  current_string,
                                  s)

        #...and add its encoded form to the output buffer
        output_buffer += encoded_s

        # Update the current string
        current_string = s

    # Return the output buffer
    return output_buffer


# Encodes a single string as LZW
def encode_as_lzw(dictionary: dict,
                  current_string: str,
                  new_character: str) -> bytes:
    # Initialize the output buffer to be empty
    output_buffer = b''

    # While there are characters left in the current string...
    while current_string!= '':
        #...if the current string plus the next character is in the dictionary...
        if current_string + new_character in dictionary:
            #...set the current string to the current string plus the next character
            current_string += new_character

            # Continue onward
            continue

        # Otherwise, write the code corresponding to the current string to the output buffer
        output_buffer += bytes([dictionary[current_string]])

        # If the next available code is equal to 4096...
        if next_available_code == 4096:
            #...reset the code size to 12 bits
            code_size = 12

        # If the next available code is equal to 65536...
        elif next_available_code == 65536:
            #...reset the code size to 16 bits
            code_size = 16

        # If the next available code is greater than 2^(code size)...
        elif next_available_code > pow(2, code_size):
            #...increase the code size by one bit
            code_size += 1

        # Set the code corresponding to the current string to the next available code
        dictionary[current_string] = next_available_code

        # Increment the next available code
        next_available_code += 1

        # Set the current string to just the next character
        current_string = new_character

    # If the next available code is equal to 4096...
    if next_available_code == 4096:
        #...reset the code size to 12 bits
        code_size = 12

    #