#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
#    print("LZW minimum code size:", lzw_minimum_code_size)

    # Generate some random data to compress with LZW
    compressed_data = []
    while True:
        # Read a random amount of data from /dev/urandom
        chunk_length = min((rng.read(1)[0] % 256) + 1, 255)
        
        if not chunk_length:
            break
            
        # Append that many bytes to our list
        compressed_data += [rng.read(chunk_length)]
        
    # Compress all chunks together
    compressed_data = b"".join([bytes([lzw_minimum_code_size])] + compressed_data)
    
    # Write the compressed data length as a variable-length integer
    compressed_data_length = len(compressed_data).to_bytes(1, 'big')
    
    # Write the compressed data itself
    compressed_data = compressed_data_length + compressed_data
    
    # Write the final sub-blocks
    compressed_data += b'\x00\x00'
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything to the output stream
    out.write(header + lsd + color_table + gce + image_descriptor + compressed_data + trailer)

