#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBB', 0, 0, 0, 0b10000000 | rng.read(1)[0], 0)
    
    # 5. Image Data
    # Let's just use the same palette as in the global color table
    data = [color_table[i : i + 3] for i in range(0, len(color_table), 3)]

    # Add some randomness to the image by changing pixels
    for i in range(len(data)):
        if rng.read(1)[0] & 1 == 1:
            data[i] = (
                rng.read(1)[0] % 256,
                rng.read(1)[0] % 256,
                rng.read(1)[0] % 256
            )
    
    # Encode the image data with LZW compression
    compressed_image_data = encode_lzw(data)
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(image_descriptor)
    out.write(compressed_image_data)


# Encodes an array of RGB tuples into a binary string that represents them using LZW compression
def encode_lzw(rgb_tuples):
    # First, we need to convert our list of RGB tuples into a list of integers
    rgb_ints = []
    for r, g, b in rgb_tuples:
        rgb_ints += [(r << 16) | (g << 8) | b]
    
    # Now let's do the actual encoding!
    # We'll start with all possible codes from 0-7
    codes = {i: chr(i).encode() for i in range(8)}
    
    # The first code is always empty
    current_code = ''
    
    # And we're going to store the encoded data here
    result = b''
    
    # For every pixel...
    for rgb_int in rgb_ints:
        
        #...convert its integer value to a binary string...
        bin_str = '{0:024b}'.format(rgb_int)
        
        #...and add it to the current code
        current_code += bin_str
        
        # If this code has already been used before...
        if current_code in codes:
            
            #...just append the corresponding code to the result
            result += codes[current_code]
            
        else:

            # Otherwise, find the longest prefix of the current code that was previously added to the dictionary
            longest_prefix = max([c for c in codes if c.startswith(current_code)], key=lambda x: len(x))
            
            # Append the corresponding code to the result
            result += codes[longest_prefix]
            
            # Then add the new code to the dictionary
            codes[current_code] = chr(len(codes)).encode()
            
            # Reset the current code
            current_code = bin_str[-1:]
            
    return result

