#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (we'll just use defaults here)
    gce = b'\xF9\x04\x00\x00\x00\x00\xFF'
    
    # 5. Image Descriptor
    # Left position is always 0
    left = struct.pack('>H', 0)
    
    # Top position is always 0
    top = struct.pack('>H', 0)
    
    # Width and height are randomly generated above
    width = struct.pack('>H', width)
    height = struct.pack('>H', height)
    
    # Flags: No local color table, interlace flag set if we're on an even row, sort flag set if we're on an odd row
    flags = struct.pack('>B', ((height[0] & 1) << 7) | ((width[0] & 1) << 6))
    
    # Image data starts after the image descriptor
    image_data_offset = len(header + lsd + color_table + gce + left + top + width + height + flags)
    
    # Putting all together
    image_descriptor = left + top + width + height + flags + struct.pack('>B', image_data_offset)
    
    # 6. Image Data
    # We will write a series of LZW compressed blocks, starting with a minimum code size of 2
    lzw_minimum_code_size = 2
    
    # Let's calculate how many bytes our image data will take by compressing it
    total_image_data_size = sum((len(b) + 1) // 256 for b in [color_table, image_descriptor, image_data])

    # Now let's calculate how many blocks that will take
    total_blocks_size = int(((total_image_data_size + 1) / 256) + 1)  # Add one because there needs to be a final block ending with a zero length
    
    # Now let's create those blocks!
    blocks = []
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
