#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
        
    # Random color table size
    color_table_size = ord(rng.read(1)) & 0b00001111

    # Global Color Table Flag
    global_color_table_flag = bool((ord(rng.read(1)) >> 7) & 0x01)

    # Sort flag
    sort_flag = bool((ord(rng.read(1)) >> 6) & 0x01)

    # Size of GCT in 256-color blocks
    gct_size = ((ord(rng.read(1)) << 8) | ord(rng.read(1))) & 0b0000111111111111

    # Background color index
    background_color_index = ord(rng.read(1))

    # Pixel aspect ratio
    pixel_aspect_ratio = ord(rng.read(1))


    # Build logical screen descriptor
    lsd = bytearray()
    lsd += struct.pack('<H', width)
    lsd += struct.pack('<H', height)
    lsd += struct.pack('B', (color_table_size << 4) +
                        int(global_color_table_flag) * 0b0001 +
                        int(sort_flag) * 0b0010)
    lsd += struct.pack('B', gct_size)
    lsd += struct.pack('B', background_color_index)
    lsd += struct.pack('B', pixel_aspect_ratio)

    # Write to output stream
    out.write(header)
    out.write(lsd)

    # 3. Global Color Table
    if global_color_table_flag:
        for i in range(pow(2, color_table_size)):
            r = ord(rng.read(1))
            g = ord(rng.read(1))
            b = ord(rng.read(1))

            out.write(struct.pack('BBB', r, g, b))

    # 4. Image Descriptor
    image_descriptor = bytearray()
    image_descriptor += struct.pack('B', 0b10000000 +
                                    0b00000000 +
                                    (int(sort_flag) ^ 1) * 0b01000000 +
                                    (int(local_color_table_flag) ^ 1) * 0b00100000 +
                                    0b00000000)
    image_descriptor += struct.pack('<H', left)
    image_descriptor += struct.pack('<H', top)
    image_descriptor += struct.pack('<H', width)
    image_descriptor += struct.pack('<H', height)
    image_descriptor += struct.pack('B', local_color_table_size)
    image_descriptor += struct.pack('B', interlace_flag)
    image_descriptor += struct.pack('B', sorted_color_table_flag)
    image_descriptor += struct.pack('B', reserved)

    # Write to output stream
    out.write(image_descriptor)

    # 5. Local Color Table
    if local_color_table_flag:
        for i in range(pow(2, local_color_table_size)):
            r = ord(rng.read(1))
            g = ord(rng.read(1))
            b = ord(rng.read(1))

            out.write(struct.pack('BBB', r, g, b))

    # 6. Block of Extension Data
    block_of_extension_data = bytearray()
    block_of_extension_data += struct.pack('B', 0b11100000)
    block_of_extension_data += struct.pack('B', 0b00000000)
    block_of_extension_data += struct.pack('B', 0b00000000)
    block_of_extension_data += struct.pack('B', 0b00000000)

    # Write to output stream
    out.write(block_of_extension_data)

    # 7. Application Extension
    application_extension = bytearray()
    application_extension += struct.pack('B', 0b11100001)
    application_extension += struct.pack('B', 0b00000000)
    application_extension += struct.pack('BB', 0xff, 0x0b)
    application_extension += b'NETSCAPE2.0'
    application_extension += struct.pack('B', 0x03)
    application_extension += struct.pack('B', 0x01)
    application_extension += struct.pack('B', 0x00)
    application_extension += struct.pack('B', 0x00)

    # Write to output stream
    out.write(application_extension)

    # 8. Graphics Control Extension
    graphics_control_extension = bytearray()
    graphics_control_extension += struct.pack('B', 0b11100001)
    graphics_control_extension += struct.pack('B', 0b00000009)
    graphics_control_extension += struct.pack('B', 0b00000000)
    graphics_control_extension += struct.pack('B', 0b00000000)
    graphics_control_extension += struct.pack('B', 0b00000000)
    graphics_control_extension += struct.pack('B', 0b00000000)
    graphics_control_extension += struct.pack('B', 0b000000