#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header (6 bytes)
    header = b'GIF89a'

    # Logical Screen Descriptor (7 bytes)
    width = rng.read(1)[0] & 0xFF  # Width is between [0, 255] inclusive
    height = rng.read(1)[0] & 0xFF  # Height is between [0, 255] inclusive
    packed = bytearray()
    packed += struct.pack('>H', width)  # Big-endian unsigned short
    packed += struct.pack('>H', height)  # Big-endian unsigned short
    packed[4] |= rng.read(1)[0] >> 7  # Global Color Table Flag
    packed[4] |= rng.read(1)[0] << 1  # Color Resolution
    packed[4] |= rng.read(1)[0] << 1  # Sort Flag
    packed[4] |= rng.read(1)[0] << 3  # Size of Global Color Table
    lsd = bytes(packed)

    # Global Color Table (3 * 2^(Packed[4] & 7) bytes)
    global_color_table = []
    for _ in range((1 << ((packed[4] & 7) + 1))):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        global_color_table.append(struct.pack('BBB', r, g, b))
    color_table = b''.join(global_color_table)

    # Graphics Control Extension (8 bytes)
    gce = b''  # No GCE

    # Image Descriptor (10 bytes)
    image_descriptor = b''  # No ID

    # LZW Minimum Code Size (1 byte)
    lzw_minimum_code_size = rng.read(1)[0] & 0b1111  # Must be less than or equal to 1111
    
    # Image Data
    blocks = []
    while True:
        if len(blocks) == 0 and rng.read(1)[0] == 0:
            break
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
