#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>HHBB', left, top, 0x00, 0x00)
    
    # 6. Image Data
    data = []
    while True:
        # Get a random byte to determine what kind of block we're going to write next
        block_type = rng.read(1)[0]
        
        if block_type == 0x2C:
            # Write an image descriptor block
            data += [image_descriptor + color_table]
            
            # Next comes the length of the compressed sub-blocks
            num_sub_blocks = rng.read(1)[0]
            
            # Then come all those sub-blocks
            for _ in range(num_sub_blocks):
                data += [b'\x00' + rng.read((rng.read(1)[0] % 128) or 1)]
            
        elif block_type == 0x21:
            # Write a graphics control extension block
            data += [gce]
            
        else:
            break
            
    # 7. Trailer
    trailer = b';'

    # Write everything to output
    out.write(header + lsd + b''.join(data) + trailer)

