#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'little') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), 'little') % (2**10 - 1) + 1

    flags = bytes([int.from_bytes(rng.read(1), 'little')])
    bgColorIndex = bytes([int.from_bytes(rng.read(1), 'little')])
    pixelAspectRatio = bytes([int.from_bytes(rng.read(1), 'little')])

    lsd = struct.pack('<6sHHBBB', header, width, height, *flags, *bgColorIndex, *pixelAspectRatio)

    # 3. Global Color Table
    globalColorTableFlag = bytes([int.from_bytes(rng.read(1), 'little')])
    colorResolution = bytes([int.from_bytes(rng.read(1), 'little')])
    sortFlag = bytes([int.from_bytes(rng.read(1), 'little')])
    sizeOfGlobalColorTable = bytes([int.from_bytes(rng.read(1), 'little')])

    gct = None
    if globalColorTableFlag == b'\x87':
        gct = rng.read((sizeOfGlobalColorTable[0] ** 2) * 3)

    # 4. Image Descriptor
    leftPos = int.from_bytes(rng.read(2), 'little')
    topPos = int.from_bytes(rng.read(2), 'little')
    imageWidth = int.from_bytes(rng.read(2), 'little') % (width - leftPos) + 1
    imageHeight = int.from_bytes(rng.read(2), 'little') % (height - topPos) + 1

    flags = bytes([int.from_bytes(rng.read(1), 'little')])
    localColorTableFlag = bytes([int.from_bytes(rng.read(1), 'little')])
    interlaceFlag = bytes([int.from_bytes(rng.read(1), 'little')])
    sortFlag = bytes([int.from_bytes(rng.read(1), 'little')])
    reserved = bytes([int.from_bytes(rng.read(1), 'little')])

    id = struct.pack('<HHHHBBBBB', leftPos, topPos, imageWidth, imageHeight, *flags, *localColorTableFlag, *interlaceFlag, *sortFlag, *reserved)

    # 5. Local Color Table
    lct = None
    if localColorTableFlag == b'\x87':
        lct = rng.read(((imageWidth[0] * imageHeight[0]) // 256) * 3)

    # 6. Image Data
    data = []
    for _ in range(0, ((imageWidth[0] * imageHeight[0]) // 256)):
        data += [rng.read(2)]

    # 7. Trailer
    trailer = b';'

    # Write everything to output
    out.write(lsd)
    if gct is not None:
        out.write(gct)
    out.write(id)
    if lct is not None:
        out.write(lct)
    for d in data:
        out.write(d)
    out.write(trailer)

