#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (we'll use this later on)
    gce = b''
    
    # 5. Image Descriptor
    # Left position is always 0
    left = struct.pack('<H', 0)
    top = struct.pack('<H', 0)
    width = struct.pack('<H', width)
    height = struct.pack('<H', height)
    packed_fields = struct.pack('B', rng.read(1)[0])  # Packed fields (using a random byte for variety)
    image_descriptor = left + top + width + height + packed_fields
    
    # 6. Image Data
    # Let's start by generating an LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0] & 0b00001111  # Mask off all but lowest 4 bits
    
    # Now let's generate some blocks!
    blocks = []
    while True:
        # Generate a random block size between 2 and 128 inclusive
        block_size = rng.read(1)[0] % 127
        
        if not block_size:
            break  # If we got a zero-length block, stop adding more
        
        # Read that many bytes from our RNG
        data = rng.read(block_size)
        
        # Add them as a new block
        blocks.append(bytes([len(data)]) + data)
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
