#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>HHBB', left, top, 0x00, 0x00)  # No local color table, no interlacing
    
    # 6. Image Data
    # Let's use the same palette as in GCE if we generated one
    if len(gce):
        image_data = b'\x00\x00' + gce[7:] * ((width * height - 1) // 256)
    else:
        image_data = b''
        
    # Add more data to the image data until we reach the desired length
    while len(image_data) < width * height:
        image_data += rng.read((width * height - len(image_data)) // 256)

    # 7. Trailer
    trailer = b';'
    
    # Write everything to our output stream
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if len(gce):
        out.write(gce)
    out.write(image_descriptor)
    out.write(image_data)
    out.write(trailer)

