#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed = bytearray()
    packed += [rng.read(1)[0], 0b00000000 | rng.read(1)[0]]  # Global Color Table Flag and Size
    packed[0] |= 0b10000000  # Sort Flag
    packed[0] |= 0b01000000  # Color Resolution
    packed[0] |= 0b00100000  # GCT Size
    packed[0] |= 0b00010000  # Background Color Index
    packed[0] |= 0b00001000  # Pixel Aspect Ratio
    packed[0] |= 0b00000100  # Global Color Table Flag
    packed[0] |= 0b00000010  # Color Table Size
    packed[0] |= 0b00000001  # Bits Per Pixel
    out.write(packed)

    # 3. Color Table
    colors = [(rng.read(1), rng.read(1), rng.read(1))[::-1] for i in range((1 << ((packed[0] & 0b00000011) + 1)))][::-1]
    color_table = b''.join([(struct.pack('BBB', *color) if len(color) == 3 else struct.pack('BB', *color) + b'\x00') for color in colors])
    
    # 4. Graphics Control Extension
    gce = struct.pack('>BHBHB', 0xF9, 4, 0, 0, 0)
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
