#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0] % 0b11111000 | 0b10000000 if width < height else rng.read(1)[0] % 0b11111000 | 0b00000000
    background_color_index = rng.read(1)[0] & 0b00000111
    pixel_aspect_ratio = rng.read(1)[0] & 0b00000111

    # Pack all these fields together as a single byte
    lsd = struct.pack('>HBBBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    global_color_table_flag = bool(packed_fields & 0b10000000)  # If this flag is set, there's a global color table
    total_colors = 2**(int((packed_fields >> 4) & 0b00000111) + 1)   # Number of colors in the global color table
    color_resolution = int(((packed_fields >> 1) & 0b00000111) + 1)     # Bits per primary color
    sort_flag = bool(packed_fields & 0b00000001)                       # If true, the global color table must be sorted by decreasing importance
    global_color_table = []                                           # The actual global color table
    for i in range(total_colors):
        r = rng.read(1)[0] & ((1 << color_resolution) - 1)           # Red component
        g = rng.read(1)[0] & ((1 << color_resolution) - 1)           # Green component
        b = rng.read(1)[0] & ((1 << color_resolution) - 1)           # Blue component
        global_color_table.append(struct.pack('BBB', r, g, b))       # Add it to our list of colors
    color_table = b''.join(global_color_table)                         # Join them all together into one big string
    
    # 4. Graphics Control Extension
    graphics_control_extension_flag = False                           # We don't need a GCE at all
    transparent_color_index = None                                    # No transparency
    delay_time = 0                                                    # No delay between frames
    transparent_color_flag = False                                   # Not using any special features from the GCE
    block_terminator = b'\x00'                                        # Block terminator
    gce = b''                                                         # Empty GCE by default
    if graphics_control_extension_flag:
        gce += struct.pack('?', graphics_control_extension_flag)      # Flag indicating that we have a GCE
        gce += struct.pack('B', transparent_color_index)              # Index of the transparent color
        gce += struct.pack('H', delay_time)                          # Delay time before next frame
        gce += struct.pack('?', transparent_color_flag)               # Whether or not we're using the special features of the GCE
        gce += block_terminator                                      # Block terminator
    
    # 5. Image Descriptor
    left = 0                                                          # Left position of the image on the screen
    top = 0                                                           # Top position of the image on the screen
    width = 0                                                         # Width of the image
    height = 0                                                        # Height of the image
    local_color_table_flag = True                                     # There's a local color table
    interlace_flag = False                                            # Not using interlacing
    sort_flag = False                                                 # Not sorting either
    local_color_table_size = len(global_color_table)                 # Size of the local color table
    local_color_table = [struct.pack('BBB', *rgb) for rgb in [(255, 0, 0)]]  # A single red color
    image_descriptor = struct.pack('>HHHHBBB', left, top, width, height, local_color_table_flag, interlace_flag, sort_flag, local_color_table_size) \
                       + b''.join(local_color_table)                    # Combine all the above into an image descriptor
    
    # 6. Image Data
    lzw_minimum_code_size = 2 ** (rng.read(1)[0] & 0b00000111)        # Minimum code size used when compressing data
    total_blocks_size = rng.read(1)[0]                               # Total size of all blocks
    blocks = []                                                       # List of compressed blocks
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
