#!/usr/bin/env python3

import struct
from typing import BinaryIO

def generate_gif(out: BinaryIO, rng=None):
    if not rng:
        from random import Random
        rng = Random()

    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = rng.randint(4, 640)
    height = rng.randint(4, 480)
    global_color_table_flag = bool(rng.getrandbits(1))
    color_resolution = rng.randint(1, 8)
    sort_flag = bool(rng.getrandbits(1))
    global_color_table_size = ((rng.read(1)[0] & 0b111) + 1) * 3
    background_color_index = rng.randint(0, pow(2, global_color_table_size)-1)
    pixel_aspect_ratio = rng.randint(0, 15)
    lsd = struct.pack('>HHBBBH', width, height, global_color_table_flag,
                      color_resolution, sort_flag, global_color_table_size,
                      background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    colors = [struct.pack('BBB', rng.randint(0, 255), rng.randint(0, 255), rng.randint(0, 255))
              for i in range((global_color_table_size+7)//3)]
    while len(colors) < global_color_table_size:
        colors += [(0, 0, 0)]
    color_table = b''.join(colors[:global_color_table_size])
    
    # 4. Graphics Control Extension
    has_transparent_color_flag = bool(rng.getrandbits(1))
    transparent_color_index = rng.randint(0, pow(2, global_color_table_size)-1)
    delay_time = rng.randint(0, 65535)
    graphics_control_extension = struct.pack('>BBHB', 0xF9, 4,
                                              has_transparent_color_flag,
                                              transparent_color_index) + \
                                struct.pack('>H', delay_time) + b'\x00'
    
    # 5. Image Descriptor
    left = rng.randint(0, max(width - 1, 1))
    top = rng.randint(0, max(height - 1, 1))
    width = min(left + rng.randint(1, width - left), width)
    height = min(top + rng.randint(1, height - top), height)
    
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
