#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = min(max(1, rng.read(1)[0] % 256 + 1), 256)
    global_color_table = []
    for _ in range(num_colors):
        global_color_table += [rng.read(3)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBB', 0, 0, 0b10000000 | rng.read(1)[0], 0, 0)
    
    # 5. Local Color Table (if applicable; let's make it have random size, capped at 256 colors, and fully random)
    if rng.read(1)[0] & 0x80!= 0:
        num_colors = min(max(1, rng.read(1)[0] % 256 + 1), 256)
        local_color_table = []
        for _ in range(num_colors):
            local_color_table += [rng.read(3)]
    
    # 6. Graphics Control Extension (if applicable)
    if rng.read(1)[0] == 0x21:
        graphics_control_extension = struct.pack('>BBHBBBBB', 0x21, 0xF9, 4, rng.read(1)[0], rng.read(1)[0], 0, 0)
    
    # 7. Application Extension (if applicable)
    if rng.read(1)[0] == 0xFF:
        application_extension = struct.pack('>BBBBB', 0xFF, 0x0B, 'NETSCAPE'.encode()[0], '2.0'.encode()[0], 3)
        application_data = struct.pack('<HBB', 1, 0, 0)
    
    # 8. Comment Extension (if applicable)
    if rng.read(1)[0] == 0xFE:
        comment_extension = struct.pack('>BB', 0xFE, 0x01)
        comment_text = ''.join([chr(rng.read(1)[0] % 128) for i in range(rng.read(1)[0] % 128)])
    
    # 9. Plain Text Extension (if applicable)
    if rng.read(1)[0] == 0x01:
        plain_text_extension = struct.pack('>BB', 0x01, 0x0C)
        text_identifier = ''.join([chr(rng.read(1)[0] % 128) for i in range(rng.read(1)[0] % 128)])
        text_subblocks = ''
        
        while True:
            subblock_size = rng.read(1)[0]
            
            if subblock_size == 0:
                break
            
            text_subblocks += ''.join([chr(rng.read(1)[0] % 128) for i in range(subblock_size)])

        final_plain_text_extension = struct.pack('>BB', 0x00, 0x00)
    
    # 10. Trailer
    trailer = b';'
    
    # Write everything to output stream
    out.write(header)
    out.write(lsd)
    out.write(struct.pack('B', len(global_color_table)))
    for entry in global_color_table:
        out.write(entry)
    out.write(image_descriptor)
    if rng.read(1)[0] & 0x80!= 0:
        out.write(struct.pack('B', len(local_color_table)))
        for entry in local_color_table:
            out.write(entry)
    if rng.read(1)[0] == 0x21:
        out.write(graphics_control_extension)
    if rng.read(1)[0] == 0xFF:
        out.write(application_extension)
        out.write(application_data)
    if rng.read(1)[0] == 0xFE:
        out.write(comment_extension)
        out.write(bytes([len(comment_text)]))
        out.write(comment_text.encode())
    if rng.read(1)[0] == 0x01:
        out.write(plain_text_extension)
        out.write(bytes([len(text_identifier)]))
        out.write(text_identifier.encode())
        out.write(bytes([(len(text_subblocks) >> 8) & 0xFF]))
        out.write(bytes([len(text_subblocks) & 0xFF]))
        out.write(text_subblocks.encode())
        out.write(final_plain_text_extension)
    out.write(trailer)

