#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
        
    def write_lzw():
        # Write LZW minimum code size
        out.write(struct.pack('<B', lzw_minimum_code_size))
        
        # Generate codes table
        codes = [i for i in range(2 ** lzw_minimum_code_size)] + [''] * ((2 ** lzw_minimum_code_size) - 1)
        
        # Initialize buffer with first character from input stream
        buffer = ord(rng.read(1).decode())
        
        while True:
            # Read next character
            char = ord(rng.read(1).decode())
            
            if len(codes) < 2 ** lzw_minimum_code_size:
                # Add new code to table
                codes += [char]
                
                # If we're done adding codes, add an empty string as well
                if len(codes) == 2 ** lzw_minimum_code_size:
                    codes += ['']
                
            else:
                # Find longest prefix of current buffer that matches one of our codes
                match = ''
                for i in reversed(range(len(buffer) + 1)):
                    candidate = ''.join([chr(c) for c in buffer[:i]])
                    
                    try:
                        idx = codes.index(candidate)
                        
                        # Found a matching prefix!
                        match = candidate
                        break
                        
                    except ValueError:
                        pass
                    
                # Output code corresponding to this prefix
                out.write(struct.pack('<H', codes.index(match)))
                
                # Add new code to table
                codes += [match[0] + chr(char)]
                
                # If we're done adding codes, add an empty string as well
                if len(codes) == 2 ** lzw_minimum_code_size:
                    codes += ['']
                    
            # Update buffer
            buffer = buffer[-1:] + char
            
    
    # Write everything
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    write_lzw()

