#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = rng.read(1)[0] % (2**8 - 16 + 1) + 16
    height = rng.read(1)[0] % (2**8 - 16 + 1) + 16

    flags = bytes([
        0b10000000 | # Global Color Table Flag
        ((width-1)&0x07)<<5 | # Width of global color table
        ((height-1)&0x07)<<3 # Height of global color table
    ])

    background_color_index = rng.read(1)[0] & 0b00001111
    pixel_aspect_ratio = rng.read(1)[0] & 0b00001111

    logical_screen_descriptor = header \
        + flags \
        + bytes([background_color_index]) \
        + bytes([pixel_aspect_ratio]) \
        + bytes([
            width&0xff,
            (width>>8)&0xff,
            height&0xff,
            (height>>8)&0xff
        ])

    # 3. Global Color Table
    global_color_table = [
        rng.read(3) for _ in range(
            2**(flags[0]&0b00000111)+1
        )
    ]

    # 4. Image Descriptor
    image_separator = b'\x2C'

    left = rng.read(1)[0] % (2**8 - 16 + 1) + 16
    top = rng.read(1)[0] % (2**8 - 16 + 1) + 16

    flags = bytes([
        0b10000000 | # Local Color Table Flag
        0b00000000 | # Interlace Flag
        0b00000000 | # Sort Flag
        0b00000000 | # Reserved
        0b00000000 | # Size of local color table
        0b00000000 | # Reserved
        ((left-1)&0x0f)<<4 | # Left position
        ((top-1)&0x0f) # Top position
    ])

    image_descriptor = image_separator \
        + flags \
        + bytes([
            left&0xff,
            (left>>8)&0xff,
            top&0xff,
            (top>>8)&0xff
        ])

    # 5. Local Color Table
    local_color_table = []

    if flags[0]&0b10000000 == 0b10000000:
        local_color_table = [
            rng.read(3) for _ in range(
                2**(flags[0]&0b00000111)+1
            )
        ]

    # 6. Data
    data = bytearray()

    while len(data) < width*height*3:
        block_size = rng.read(1)[0]

        if block_size > 0:
            data += rng.read(block_size)
        else:
            break

    # 7. Trailer
    trailer = b''

    if len(local_color_table):
        trailer += b','+bytes([len(local_color_table)-1])

    trailer += b';'

    # Write to output stream
    out.write(logical_screen_descriptor)
    out.write(struct.pack('B', len(global_color_table)))
    for entry in global_color_table:
        out.write(entry)
    out.write(image_descriptor)
    out.write(struct.pack('B', len(local_color_table)))
    for entry in local_color_table:
        out.write(entry)
    out.write(data)
    out.write(trailer)