#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    flags = rng.read(1)[0] | ((rng.read(1)[0] & 0b111) << 8)  # Flags is a single byte that contains global color table flag and color resolution bits
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HBBBBB', int.from_bytes(width, 'big'), int.from_bytes(height, 'big'), flags, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    colors = [rng.read(3) for i in range((flags >> 7) + 1)]
    color_table = b''.join(colors)
    
    # 4. Graphics Control Extension
    delay = rng.read(2)
    transparent_color_index = rng.read(1)[0]
    gce = struct.pack('>HB', delay[0], delay[1]) + struct.pack('>B', transparent_color_index)
    
    # 5. Image Descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
