#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0] % 64
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 9
    
    # Pack all these values together in little-endian order
    lsd = struct.pack('<HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    global_color_table_flag = bool(packed_fields & 0b10000000)
    if not global_color_table_flag: return  # No need to write anything else if there is no global color table
    
    # Number of colors in the global color table
    global_color_table_size = ((packed_fields >> 7) & 0b111) + 1
    
    # If we have more than one color, then we can use it as our palette
    if global_color_table_size > 1:
        # Generate an array of RGB triplets from the RNG
        rgb_triplets = [struct.pack('<BBB', *rng.read(3)) for _ in range(global_color_table_size)]
        
        # Combine them into one big string
        color_table = b''.join(rgb_triplets)
    else:
        # Otherwise just put a single black pixel
        color_table = b'\x00\x00\x00'
    
    # 4. Graphics Control Extension
    graphics_control_extension_flag = bool(packed_fields & 0b01000000)
    if not graphics_control_extension_flag: return  # No need to write anything else if there's no GCE
    
    # Delay time (1/100th second)
    delay_time = rng.read(2)
    
    # Transparent color index
    transparent_color_index = rng.read(1)[0] % 256
    
    # Block terminator
    block_terminator = b'\x00'
    
    # Put everything together
    gce = delay_time + struct.pack('<BB', transparent_color_index, block_terminator[0])
    
    # 5. Image Descriptor
    image_separator = b'\x2C'
    
    # Left position
    left_position = rng.read(2)
    
    # Top position
    top_position = rng.read(2)
    
    # Width
    width = rng.read(2)
    
    # Height
    height = rng.read(2)
    
    # Packed field
    packed_field = rng.read(1)[0] % 64
    
    # Interlace flag
    interlace_flag = bool(packed_field & 0b10000000)
    
    # Sort flag
    sort_flag = bool(packed_field & 0b01000000)
    
    # Local color table flag
    local_color_table_flag = bool(packed_field & 0b00100000)
    
    # Local color table size
    local_color_table_size = ((packed_field >> 7) & 0b111) + 1
    
    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0] % 129
    
    # Block terminator
    block_terminator = b'\x00'
    
    # Put everything together
    image_descriptor = left_position + top_position + width + height + struct.pack('<BBBB', packed_field, lzw_minimum_code_size, block_terminator[0], block_terminator[0])
    
    # 6. Image Data
    # We're going to make up some random data here!
    blocks = []
    
    # First block will be a single clear code
    blocks += [bytes([lzw_minimum_code_size])]
    
    # Then we'll add random codes until we hit a clear code again
    while True:
        # Get a random code
        code = rng.read(1)[0] % 2**lzw_minimum_code_size
        
        # Add it to the list of codes
        blocks[-1] += bytes([code])
        
        # Check if this was a clear code or not
        if len(blocks[-1]) == lzw_minimum_code_size + 1:
            break
    
    # Now that we've got a bunch of codes, let's turn them back into strings
    blocks = [bytes(block).decode() for block in blocks]
    
    # And join them together into one long string
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
