#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        width: int, height: int,
        min_color_depth: int=1, max_color_depth: int=8,
        min_num_colors: int=2**min_color_depth, max_num_colors: int=2**max_color_depth):

    # 1. Header (6 bytes)
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor (7 bytes)
    lsd = struct.pack('<HHBBB', width, height, min_color_depth, 0, 0, 0)
    
    # 3. Color Table (2**(min_color_depth+1)*3 bytes)
    num_colors = rng.randint(min_num_colors, max_num_colors)
    colors = [struct.pack('BBB', *rng.randrange(256, dtype='uint8') for _ in range(3)) for _ in range(num_colors)]
    color_table = b''.join(colors)
    
    # 4. Graphics Control Extension (8 bytes)
    gce = struct.pack('<BBHBBBBB', 0xF9, 4, 0, 0, 0, 0, 0, 0)
    
    # 5. Image Descriptor (10 bytes)
    image_descriptor = struct.pack('<BHHHHB', 0x2C, 0, 0, width, height, 0, 0)
    
    # 6. Image Data
    # LZW minimum code size is between 2 and 12 inclusive
    lzw_minimum_code_size = rng.randint(2, 12)
    
    # Generate blocks of data until we reach an end-of-file marker
    blocks = []
    while True:
        
        # Read a byte from the input stream
        c = rng.read(1)[0]

        if c == 0xFF:
            break
            
        elif c < lzw_minimum_code_size:
            
            # If it's less than the minimum code size, just write that as one byte
            blocks.append(bytes((c,)))
            
        else:
        
            # Otherwise, read another byte which gives us the length of the string
            slen = rng.read(1)[0]

            # Then read the actual string itself
            s = rng.read(slen).decode()

            # And add it to our list of blocks
            blocks.append(s.encode())
        
    # Join all these blocks together and append an end-of-block character
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
