#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>BHHBBB',
        0b11100000 | ((width >> 7) & 0x01) << 2 | ((height >> 7) & 0x01) << 1,
        width % 0x8000,
        height % 0x8000,
        rng.read(1)[0] & 0x07,
        rng.read(1)[0],
        0b10000000 | rng.read(1)[0] & 0x0F
    )

    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 0x07 + 1
    global_color_table = []
    if global_color_table_size > 0:
        for i in range(global_color_table_size):
            global_color_table += [
                rng.read(1)[0],
                rng.read(1)[0],
                rng.read(1)[0]
            ]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB',
        0x2C,
        rng.read(1)[0] & 0x07,
        x=0, y=0,
        width=width, height=height,
        packed=(
            0b10000000 | rng.read(1)[0] & 0x0F
        )
    )

    # 5. Local Color Table
    local_color_table_size = rng.read(1)[0] & 0x07 + 1
    local_color_table = []
    if local_color_table_size > 0:
        for i in range(local_color_table_size):
            local_color_table += [
                rng.read(1)[0],
                rng.read(1)[0],
                rng.read(1)[0]
            ]

    # 6. Graphics Control Extension
    graphics_control_extension = struct.pack('>BBHB',
        0x21,
        0xF9,
        4,
        0b10000000 | rng.read(1)[0] & 0x0F
    )

    # 7. Application Extension
    application_extension = struct.pack('>BBBBB',
        0xFF,
        0x0B,
        11,
        ord('A'), ord('N'), ord('I'), ord('M'), ord('E'), ord('X'), ord('T')
    )

    # 8. Comment Extension
    comment_extension = struct.pack('>BBBBB',
        0xFF,
        0xFE,
        rng.read(1)[0] % 0x100,
        rng.read(1)[0],
        rng.read(1)[0]
    )

    # 9. Plain Text Extension
    plain_text_extension = struct.pack('>BBBBB',
        0xFF,
        0x01,
        rng.read(1)[0] % 0x100,
        rng.read(1)[0],
        rng.read(1)[0]
    )

    # 10. Trailer
    trailer = b';'

    # Write everything to output stream
    out.write(header)
    out.write(lsd)
    if len(global_color_table) > 0:
        out.write(bytes([global_color_table_size]))
        out.write(bytes(global_color_table))
    else:
        out.write(b'\x00')
    out.write(image_descriptor)
    if len(local_color_table) > 0:
        out.write(bytes([local_color_table_size]))
        out.write(bytes(local_color_table))
    else:
        out.write(b'\x00')
    out.write(graphics_control_extension)
    out.write(application_extension)
    out.write(comment_extension)
    out.write(plain_text_extension)
    out.write(trailer)

