#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor (LSD)
    width = rng.read(2)
    height = rng.read(2)
    flags = rng.read(1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)
    lsd = width + height + flags + background_color_index + pixel_aspect_ratio

    # 3. Global Color Table (GCT)
    global_color_table_flag = rng.read(1)
    total_colors_in_global_color_table = rng.read(1)
    if global_color_table_flag == b'\x80':
        raise NotImplementedError('Global color table is not supported')
    elif global_color_table_flag!= b'\x00':
        raise ValueError('Invalid global color table flag')
    colors_in_global_color_table = []
    for i in range(total_colors_in_global_color_table * 3):
        colors_in_global_color_table.append(rng.read(1))
    gct = bytes([global_color_table_flag[0]]) + bytes([total_colors_in_global_color_table[0]]) \
            + bytes(colors_in_global_color_table)

    # 4. Image Descriptor (ID)
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    flags = rng.read(1)
    local_color_table_flag = rng.read(1)
    interlace_flag = rng.read(1)
    sort_flag = rng.read(1)
    reserved = rng.read(3)
    id = left + top + width + height + flags + local_color_table_flag + interlace_flag + sort_flag + reserved

    # 5. Local Color Table (LCT)
    if local_color_table_flag == b'\x80':
        raise NotImplementedError('Local color table is not supported')
    elif local_color_table_flag!= b'\x00':
        raise ValueError('Invalid local color table flag')
    total_colors_in_local_color_table = rng.read(1)
    colors_in_local_color_table = []
    for i in range(total_colors_in_local_color_table * 3):
        colors_in_local_color_table.append(rng.read(1))
    lct = bytes([local_color_table_flag[0]]) + bytes([total_colors_in_local_color_table[0]]) \
            + bytes(colors_in_local_color_table)

    # 6. Image Data
    lzw_minimum_code_size = rng.read(1)
    total_blocks_size = rng.read(1)
    blocks = []
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
