#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Color table size is always 0
    color_table_size = b'\x00\x00'

    # Background color index is also always 0
    background_color_index = b'\x00'

    # Pixel aspect ratio is always 0
    pixel_aspect_ratio = b'\x00'

    # Put all of these together in one byte string
    lsd = width + height + color_table_size + \
          background_color_index + pixel_aspect_ratio

    # 3. Global Color Table
    # We'll just use black as our global color table
    color_table = b'\x00\x00\x00'

    # 4. Graphics Control Extension
    # This extension will be used to set the delay between frames
    # to zero so that we can see it happen more easily
    label = b'\x21\xF9\x04'

    # No transparancy, no user input, delay time is 0
    flags = b'\x00\x00\x00'

    # Transparent color index is 0
    transparent_color_index = b'\x00'

    # Final byte string with the above values
    gce = label + flags + transparent_color_index

    # 5. Image Descriptor
    # Frame left position is always 0
    frame_left_position = b'\x00\x00'

    # Frame top position is always 0
    frame_top_position = b'\x00\x00'

    # Width and height are the same as the logical screen descriptor
    image_width = width
    image_height = height

    # Packed field is always 0
    packed_field = b'\x00'

    # Finally put them all together
    image_descriptor = frame_left_position + frame_top_position + \
                       image_width + image_height + packed_field

    # 6. Image Data
    # The first byte is an LZW minimum code size value of 7
    data = [b'\x07']

    # Generate some random pixels
    while len(data) < width * height:
        # Get a random byte from /dev/urandom
        randbyte = rng.read(1)[0]

        # If this byte is not already in the list, add it
        if randbyte not in data:
            data += [randbyte]
        
        # If there's enough space now, stop generating new bytes
        if len(data) >= width * height:
            break

        # Otherwise, get another random byte
        else:
            randbyte = rng.read(1)[0]

            # Add it to the end of the list
            data += [randbyte]

    # Convert the list of bytes into a single byte string
    image_data = b''.join(data)

    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
