#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = [struct.pack('BBB', *rng.read(3)) for i in range(num_colors)]
    color_table = b''.join(global_color_table)
    
    # 4. Graphics Control Extension
    # We'll use this later on when we write the Image Data Block
    gce = struct.pack('>BBHBBBBB', 0x21, 0xF9, 4, 0b00000000, 0, 0, 0)
    
    # 5. Image Descriptor
    # Frame left position, frame top position, frame width, frame height, packed fields (using a random byte for variety)
    image_descriptor = struct.pack('>HHHHB', 0, 0, width, height, rng.read(1)[0])
    
    # 6. Image Data Block
    # Let's use LZW compression here!
    lzw_minimum_code_size = 8   # This is the minimum code size that will be used by LZW
    total_blocks_size = 0       # Total size of all data blocks (including their sizes)
    blocks = []                 # List containing all data blocks
    while True:
        # Generate a random string of bytes, then check if its length is greater than our maximum allowed code size
        # If so, break because we're done generating codes
        code = rng.read((lzw_minimum_code_size - 1) // 8 + 1)
        if len(code) > lzw_minimum_code_size:
            break
        
        # Add the current code to the list of codes
        blocks.append(code)
        total_blocks_size += len(code) + 1  # The extra one is the block size itself
        
    # Now let's add the remaining codes as literals until we reach the maximum allowed code size
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
