#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors!
    
    # 4. Image Descriptor (same as above)
    image_descriptor = struct.pack('>HHBBB', width, height, 0x87, 0, 0)
    
    # 5. Image Data
    data = []
    while True:
        # Each frame is made of an introducer (always 0x2C) followed by its descriptor and then the actual data
        data += [b'\x2C']
        
        # Frame descriptor: Left position, Top position, Width, Height, Flags
        # We'll use all zeroes here since we don't need to compress this image
        data += [struct.pack('>HHHHB', 0, 0, width, height, 0)]
        
        # Now comes the compressed data...
        # Let's just write some random pixels until we fill up the whole screen
        for _ in range((width * height)):
            data += [color_table[rng.read(1)[0]]]
            
            if len(data) >= ((width * height) + 1):
                break
            
        else:
            continue

        break
        
    # Finally, let's add the trailer
    data += [b';']

    # Write everything to our output stream
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(image_descriptor)
    out.write(bytes(data))

