#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>BHHBBB',
        0b00000000 | rng.read(1)[0],
        width, height,
        0b00000000 | rng.read(1)[0],
        0b00000000 | rng.read(1)[0],
        0b00000000 | rng.read(1)[0]
    )

    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 7 + 1
    global_color_table = []
    if global_color_table_size > 0:
        for i in range(global_color_table_size):
            rgb = [rng.read(1)[0] for _ in range(3)]
            global_color_table += rgb
    
    # 4. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB',
        0x2C,
        0b00000000 | rng.read(1)[0],
        x=width, y=height,
        local_color_table_flag=(rng.read(1)[0] >> 7) == 1,
        interlace_flag=(rng.read(1)[0] >> 6) == 1,
        sort_flag=(rng.read(1)[0] >> 5) == 1,
        reserved=rng.read(1)[0],
        size_of_local_color_table=rng.read(1)[0] & 7 + 1
    )

    # 5. Local Color Table
    local_color_table = []
    if image_descriptor[7] > 0:
        for i in range(image_descriptor[7]):
            rgb = [rng.read(1)[0] for _ in range(3)]
            local_color_table += rgb
    
    # 6. Image Data
    image_data = b''
    while True:
        block_type = rng.read(1)[0]
        
        if block_type == 0x2C:
            break

        elif block_type == 0x21:
            pass

        else:
            raise Exception("Unknown block type")
    
    # 7. Trailer
    trailer = b';'

    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    if len(global_color_table) > 0:
        out.write(struct.pack('>B', global_color_table_size * 3))
        out.write(bytes(global_color_table))
    out.write(image_descriptor)
    if len(local_color_table) > 0:
        out.write(struct.pack('>B', image_descriptor[7] * 3))
        out.write(bytes(local_color_table))
    out.write(image_data)
    out.write(trailer)