#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
    out: BinaryIO,
    width: int, height: int,
    num_colors: int,
    lzw_minimum_code_size: int=8,
    background_color: int=None,
    has_transparency: bool=False
):
    assert (width * height <= 65535), 'Image is too large!'

    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    lsd = struct.pack('<HHBBB', width, height, num_colors << 4 | lzw_minimum_code_size - 1, 0 if not has_transparency else 0b10000000, 0)
    
    # 3. Global Color Table
    colors = [background_color or 0] + list(range(num_colors-1))
    color_table = b''
    for c in colors:
        r = ((c >> 16) & 0xFF) / 255
        g = ((c >>  8) & 0xFF) / 255
        b = ((c      ) & 0xFF) / 255
        color_table += struct.pack('BBB', round(r*255), round(g*255), round(b*255))
    
    # 4. Graphics Control Extension
    gce = b''
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('<HHBB', 0, 0, 0b10000000, 0b10000000)  # No local color table, interlace and no sort flag
    
    # 6. Image Data    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
