#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    
    # We'll use RGB triplets to represent colors
    color_table = b''
    for _ in range(num_colors):
        r, g, b = [rng.read(1)[0] for _ in range(3)]
        color_table += struct.pack('BBB', r, g, b)
    
    # 4. Graphics Control Extension (we'll just put one with no delay)
    gce = b'\x21\xF9\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF'
    
    # 5. Image Descriptor
    # Left position is always zero
    left = 0
    top = 0
    
    # Width and Height are random again
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed field: Local color table flag set if we have a local color table, interlace flag unset, sort flag unset, reserved bits cleared
    packed_field = 0b10000000 | rng.read(1)[0]
    
    # Write the image descriptor
    image_descriptor = struct.pack('>HHBBBBB', left, top, width, height, packed_field, num_colors)
    
    # 6. Image Data
    # Let's make this a random amount of pixels (capped at 256*256=65536)
    num_pixels = ((width * height) % 65536) or 1
    
    # Each pixel will be represented by three bytes (RGB)
    image_data = b''
    for _ in range(num_pixels):
        r, g, b = [rng.read(1)[0] for _ in range(3)]
        image_data += struct.pack('BBB', r, g, b)
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything to our output stream
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
