#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + rng.randint(0x7F)
    height = 1 + rng.randint(0x7F)
    packed = (width - 1).to_bytes(1, 'little', signed=False) \
        + (height - 1).to_bytes(1, 'little', signed=False) \
        + b'\x87\x00' \
        + bytes([rng.randint(256)])
    out.write(packed)

    # Global color table
    palette = [struct.pack('BBB', *rng.randint(256, size=3)) for _ in range(256)]
    out.write(b''.join(palette))

    # Image descriptor
    packed = (0, 0, width, height).to_bytes(4, 'little', signed=False) \
        + b'\x00'
    out.write(packed)

    # LZW minimum code size
    out.write(b'\x01')

    # Frame data
    lzw_mincodesize = 1
    while True:
        # Generate some random data to compress
        data = bytearray()

        # Add a few RLE blocks of different lengths
        for i in range(rng.randint(10)):
            length = rng.randint(lzw_mincodesize+1, min(lzw_mincodesize*2, 128))

            if length == lzw_mincodesize+1 and len(data) < 2**lzw_mincodesize-1:
                continue

            data += bytes([length])
            data += bytes([rng.randint(256)]) * length

        # If we have enough data, stop generating more
        if len(data) >= 2**lzw_mincodesize-1:
            break

        # Add an EOF block
        data += b'\x00'

        # Compress it with LZW
        compressed = []
        dict_size = 2 ** lzw_mincodesize
        dictionary = {i: bytes([i]) for i in range(dict_size)}
        prev = None
        for c in data:
            if not prev is None:
                key = prev + c

                if key in dictionary:
                    prev = key
                else:
                    compressed.append(dictionary[prev])

                    if len(compressed) == dict_size:
                        lzw_mincodesize += 1
                        dict_size *= 2

                        new_dictionary = {}
                        for k, v in dictionary.items():
                            new_dictionary[k] = v + bytes([v[-1]])
                        dictionary = new_dictionary

                    dictionary[len(dictionary)] = prev + c[:1]
                    prev = c[:1]
            else:
                prev = c

        # Write the compressed data
        out.write(b''.join(compressed))

    # Trailer
    out.write(b';')

