#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 256 + (rng.randint(0, 127)) * 256
    height = 256 + (rng.randint(0, 127)) * 256
    flags = b'\x80\x00' if rng.randint(0, 1) == 0 else b'\x00\x00'
    bg_color = bytes([rng.randint(0, 255), rng.randint(0, 255), rng.randint(0, 255)])
    pixel_aspect_ratio = b'\x00'
    out.write(struct.pack('<HHBBB', width, height, flags[0], flags[1], bg_color[0]))
    out.write(pixel_aspect_ratio)

    # Global color table
    num_colors = min((width * height // 64) + 1, 256)
    for i in range(num_colors):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write(bytes([r, g, b]))

    # Image descriptor
    left = rng.randint(0, width - 1)
    top = rng.randint(0, height - 1)
    width = max(1, rng.randint(0, width - left))
    height = max(1, rng.randint(0, height - top))
    flags = b'\x80\x00' if rng.randint(0, 1) == 0 else b'\x00\x00'
    local_color_table = b'\x00'
    interlace = b'\x00'
    sort = b'\x00'
    reserved = b'\x00\x00'
    lct_size = b'\x00'
    out.write(struct.pack('<HHHHBBBBB', left, top, width, height, flags[0], flags[1], local_color_table, interlace[0], sort[0], reserved[0], reserved[1]))
    out.write(lct_size)

    # Local color table
    for i in range(num_colors):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write(bytes([r, g, b]))

    # Table based image data
    lzw_min_code_size = rng.randint(2, 8)
    out.write(bytes([lzw_min_code_size]))

    # LZW code size
    lzw_code_size = [8] * 12 + [9] * 12 + [10] * 12 + [11] * 12 + [12] * 12 + \
                   [13] * 12 + [14] * 12 + [15] * 3 + [16] * 3 + [17] * 3 + [18] * 3 + [19] * 3 + [20] * 3 + [21] * 3 + [22] * 3 + [23] * 3 + [24] * 3 + [25] * 3 + [26] * 3 + [27] * 3 + [28] * 3 + [29] * 3 + [30] * 3 + [31] * 3 + [32] * 3
    out.write(bytes(lzw_code_size[:lzw_min_code_size]))

    # Pixels
    pixels = []
    while len(pixels) < width * height:
        value = rng.randint(0, 255)

        if value >= lzw_min_code_size and value <= 255:
            pixels += [value]
        elif value > 255:
            raise Exception('Invalid LZW code')

    # Pack bits
    def packbits():
        buf = bytearray()
        count = 0
        for v in pixels:
            if count == 0:
                buf.append(v)
                count = 8
            else:
                buf[-1] |= ((v & 1) << (count-1))
                count -= 1
        return buf

    packed = packbits()
    out.write(packed)

