#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = rng.read(1)[0] % 64 + 1
    h = rng.read(1)[0] % 64 + 1

    # Random color table size
    cts = rng.read(1)[0] & 7

    # Global Color Table Flag
    gctf = bool(rng.read(1)[0] & 1)

    # Sort flag
    sf = bool(rng.read(1)[0] & 1)

    # Size of global color table
    sgt = int((cts+1)*3) if gctf else 0

    # Background color index
    bgci = rng.read(1)[0] & 7

    # Pixel aspect ratio
    par = rng.read(1)[0] & 3

    lsd = struct.pack('<6sHHBBB', header, w*256+h, sgt, bgci, cts, gctf, sf, par)

    # Write to output stream
    out.write(lsd)

    # 3. Global Color Table
    if gctf:
        for i in range(sgt):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]

            out.write(struct.pack('BBB', r,g,b))


    # 4. Image Descriptor
    # Left position
    lp = rng.read(1)[0] % 64
    # Top position
    tp = rng.read(1)[0] % 64
    # Width
    iw = rng.read(1)[0] % 64 + 1
    # Height
    ih = rng.read(1)[0] % 64 + 1
    # Local color table flag
    lctf = bool(rng.read(1)[0] & 1)
    # Interlace flag
    iff = bool(rng.read(1)[0] & 1)
    # Sort flag
    sf = bool(rng.read(1)[0] & 1)
    # Reserved byte
    rb = rng.read(1)[0]
    # LZW minimum code size
    lcs = rng.read(1)[0] & 7

    id = struct.pack('<BBHHB', lp, tp, iw*256+ih, lctf, iff, sf, rb)

    # Write to output stream
    out.write(id)

    # 5. Local Color Table
    if lctf:
        for i in range(3*(1<<lcs)):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]

            out.write(struct.pack('BBB', r,g,b))

    # 6. Image Data
    # Number of pixels
    np = iw * ih
    # Current pixel
    cp = 0
    # Current bit
    cb = -1
    # Current byte
    cbp = None
    # Next byte
    nb = None
    # Next bit
    nbit = False
    # Output buffer
    ob = []

    while True:
        # If we have no more bits left in this byte...
        if not nbit:
            # Read next byte from input stream
            cbp = nb
            try:
                nb = rng.read(1)
            except Exception as e:
                print("ERROR:",e)
                break
            
            # If there are no more bytes available...
            if len(nb)==0:
                #...we're done!
                break

            # Otherwise, reset current bit counter
            cb = 7

        # Get next bit from current byte
        nbit = ((cbp[0] >> cb) & 1) == 1
        
        # If we just read a zero bit...
        if not nbit:
            #...and we already had a zero bit before...
            if len(ob)>0 and ob[-1]==0:
                #...this is the end of our image data!
                break

        # Add it to the output buffer
        ob += [nbit]

        # Increment current bit counter
        cb -= 1

    # Convert output buffer to string
    ob = ''.join([str(int(x)) for x in ob])

    # Write compressed image data to output stream
    out.write(bytes([len(ob)]))
    out.write(bytes([(int(ob[i:i+8],2)+1) for i in range(0,len(ob),8)]))

