#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w, h = struct.unpack('>HH', rng.read(4))

    # Color table size is always 0
    color_table_size = 0x00

    # Global color table flag is set if there's no global color table
    # (i.e. we use local color tables)
    global_color_table_flag = 0x87

    # Background color index is always 0
    background_color_index = 0x00

    # Pixel aspect ratio is always 0
    pixel_aspect_ratio = 0x00

    logical_screen_descriptor = struct.pack(
        '>HHBBBH', 
        w, h,
        color_table_size,
        global_color_table_flag,
        background_color_index,
        pixel_aspect_ratio
    )

    # 3. Image Descriptor
    image_left = 0x00
    image_top = 0x00
    image_width = w
    image_height = h

    # Local color table flag is set if there's no global color table
    # (i.e. we use local color tables)
    local_color_table_flag = 0x87

    # Interlace flag is always false
    interlace_flag = 0x00

    # Sort flag is always false
    sort_flag = 0x00

    # Reserved flags are all zeroes
    reserved_flags = 0x00

    # LZW minimum code size is always 0b00001
    lzw_minimum_code_size = 0b00001

    image_descriptor = struct.pack(
        '>HHHHBB',
        image_left, image_top,
        image_width, image_height,
        local_color_table_flag,
        interlace_flag,
        sort_flag,
        reserved_flags,
        lzw_minimum_code_size
    )

    # 4. Graphics Control Extension
    graphics_control_extension_label = b'\x21\xF9\x04'

    # Transparent color index is always 0
    transparent_color_index = 0x00

    # Block size is always 0
    block_size = 0x00

    # Disposal method is always 0
    disposal_method = 0x00

    # User input flag is always false
    user_input_flag = 0x00

    # Transparency flag is true if the previous frame was transparent
    transparency_flag = 0x00

    graphics_control_extension = struct.pack(
        '>BHB',
        graphics_control_extension_label[0],
        graphics_control_extension_label[1],
        0x04,

        transparent_color_index,
        block_size,
        disposal_method,
        user_input_flag,
        transparency_flag
    )

    # 5. Application Extension
    application_extension_label = b'\xFF\x0B'
    application_identifier = b'NETSCAPE2.0'
    application_auth_code = b'\x03\x01' + b'\x00'*11

    application_extension = struct.pack(
        '>BB%ds%ds' % (len(application_identifier), len(application_auth_code)),
        application_extension_label[0],
        application_extension_label[1],
        
        len(application_identifier),
        application_identifier,

        len(application_auth_code),
        application_auth_code
    )

    # 6. Comment Extension
    comment_extension_label = b'\xFE\x01'

    comment_extension = struct.pack(
        '>BB%ds' % len(comment_extension_label),
        comment_extension_label[0],
        comment_extension_label[1],
        len(comment_extension_label) - 2,
        comment_extension_label
    )

    # 7. Plain Text Extension
    plain_text_extension_label = b'\x01\x00\x00'

    text_grid_left = 0x00
    text_grid_top = 0x00
    text_grid_width = w // 10
    text_grid_height = h // 10
    cell_width = 10
    cell_height = 10
    foreground_color_index = 0x0F
    background_color_index = 0x00

    plain_text_extension = struct.pack(
        '>BBBBHHHHBBB',
        plain_text_extension_label[0],
        plain_text_extension_label[1],
        plain_text_extension_label[2],

        0x0C,

        text_grid_left,
        text_grid_top,
        text_grid_width,
        text_grid_height,
        cell_width,
        cell_height,
        foreground_color_index,
        background_color_index
    )

    # 8. Frame Data
    # We'll just put some random data in here
    frame_data = rng.read()

    # 9. Trailer
    trailer = b';'

    # Write everything to the output stream
    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(image_descriptor)
    out.write(graphics_control_extension)
    out.write(application_extension)
    out.write(comment_extension)
    out.write(plain_text_extension)
    out.write(frame_data)
    out.write(trailer)

