#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>BBBBB', rng.read(1)[0] & 7, rng.read(1)[0], width % 256, height % 256, 0b10000000 | ((width // 256) << 4) | (height // 256))

    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 7
    global_color_table = []
    if global_color_table_size > 0:
        for _ in range(global_color_table_size):
            global_color_table += [rng.read(3)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>BHHBBBB', 0x2C, 0, 0, width % 256, height % 256, 0, 0)

    # 5. Local Color Table
    local_color_table_size = rng.read(1)[0] & 7
    local_color_table = []
    if local_color_table_size > 0:
        for _ in range(local_color_table_size):
            local_color_table += [rng.read(3)]

    # 6. Data Sub-blocks
    data_sub_blocks = []
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x21: # Extension Block
            extension_block_label = rng.read(1)[0]

            if extension_block_label == 0xF9:
                # Graphics Control Extension
                graphics_control_extension_block = struct.pack('>BHB', 0x21, 0xF9, 4) + rng.read(3)

                # Plain Text Extension
                plain_text_extension_block = struct.pack('>BHBHBB', 0x21, 0xFE, 12, 0, 0, 0)
                
                # Application Extension
                application_extension_block = struct.pack('>BHBHBB', 0x21, 0xFF, 11, ord('A'), ord('N') << 8 | ord('I'), 0) + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'

                data_sub_blocks += [graphics_control_extension_block, plain_text_extension_block, application_extension_block]
            
            elif extension_block_label == 0x01 or extension_block_label == 0xFF:
                # Plain Text Extension
                plain_text_extension_block = struct.pack('>BHBHBB', 0x21, 0xFE, 12, 0, 0, 0)
                
                # Application Extension
                application_extension_block = struct.pack('>BHBHBB', 0x21, 0xFF, 11, ord('A'), ord('N') << 8 | ord('I'), 0) + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'

                data_sub_blocks += [plain_text_extension_block, application_extension_block]
        
        else: # Image Data Block
            data_sub_blocks += [struct.pack('>BB', 0x21, 0xF9)]
            break

    # 7. Trailer
    trailer = b';'

    # Write everything to `out`
    out.write(header)
    out.write(lsd)
    for entry in global_color_table:
        out.write(entry)
    out.write(image_descriptor)
    for entry in local_color_table:
        out.write(entry)
    for sub_block in data_sub_blocks:
        out.write(sub_block)
    out.write(trailer)