#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a' + bytes([0x20]) * 7

    # 2. Logical screen descriptor
    width = int.from_bytes(rng.read(2), 'big') % (2**16 - 1) + 1
    height = int.from_bytes(rng.read(2), 'big') % (2**16 - 1) + 1
    flags = int.from_bytes(rng.read(1), 'big') & 0b1111
    bg_color = int.from_bytes(rng.read(1), 'big')
    aspect_ratio = int.from_bytes(rng.read(1), 'big')

    logical_screen_descriptor = \
        struct.pack('>HHBBB', width, height, flags, bg_color, aspect_ratio)

    # 3. Global color table
    global_color_table_size = int.from_bytes(rng.read(1), 'big')
    if global_color_table_size == 0:
        global_color_table = b''
    else:
        global_color_table = bytes([int.from_bytes(rng.read(1), 'big')] *
                                   min(global_color_table_size, 256))

    # 4. Image block
    image_left = int.from_bytes(rng.read(2), 'big')
    image_top = int.from_bytes(rng.read(2), 'big')
    image_width = int.from_bytes(rng.read(2), 'big') % (2**16 - 1) + 1
    image_height = int.from_bytes(rng.read(2), 'big') % (2**16 - 1) + 1
    flags = int.from_bytes(rng.read(1), 'big') & 0b1111
    local_color_table_flag = bool(flags & 0b10000000)
    interlace_flag = bool(flags & 0b01000000)
    sort_flag = bool(flags & 0b00100000)
    reserved = int.from_bytes(rng.read(1), 'big')
    lct_size = int.from_bytes(rng.read(1), 'big')
    lct = None
    if local_color_table_flag and lct_size!= 0:
        lct = bytes([int.from_bytes(rng.read(1), 'big')] *
                    min(lct_size, 256))

    # 5. LZW minimum code size
    lzw_min_code_size = int.from_bytes(rng.read(1), 'big')

    # 6. Data sub-blocks
    data_sub_block_size = 2 ** lzw_min_code_size
    data_sub_blocks = []
    while True:
        data_sub_block = rng.read(data_sub_block_size)

        if len(data_sub_block) < data_sub_block_size or not any(
                data_sub_block
            ): break

        data_sub_blocks += [data_sub_block]

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to output stream
    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(global_color_table)
    out.write(struct.pack('>HHHHBB',
                          image_left, image_top,
                          image_width, image_height,
                          0, 0))
    if local_color_table_flag:
        out.write(lct)
    for data in data_sub_blocks:
        out.write(data)
    out.write(trailer)

