#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    def write_byte(b: int) -> None:
        out.write(struct.pack('B', b))

    def write_short(s: int) -> None:
        out.write(struct.pack('>H', s))

    def write_long(l: int) -> None:
        out.write(struct.pack('>L', l))

    # 1. Header
    write_byte(ord('G'))
    write_byte(ord('I'))
    write_byte(ord('F'))
    write_byte(ord('8'))
    write_byte(ord('9'))
    write_byte(ord('a'))

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    write_short(width)
    write_short(height)
    write_byte(packed_fields)
    write_byte(background_color_index)
    write_byte(pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = [rng.read(3) for i in range(num_colors)]  # Random colors

    for r, g, b in color_table:
        write_byte(r)
        write_byte(g)
        write_byte(b)

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        write_byte(0x21)
        write_byte(0xF9)
        write_byte(0x04)

        delay_time = rng.read(1)[0]
        transparent_color_index = rng.read(1)[0]
        terminator = rng.read(1)[0]

        write_byte(delay_time)
        write_byte(transparent_color_index)
        write_byte(terminator)

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = rng.read(1)[0]
    interlace_flag = rng.read(1)[0]
    sort_flag = rng.read(1)[0]
    reserved = rng.read(1)[0]
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    write_byte(0x2C)
    write_short(left)
    write_short(top)
    write_short(width)
    write_short(height)
    write_byte(local_color_table_flag)
    write_byte(interlace_flag)
    write_byte(sort_flag)
    write_byte(reserved)
    write_byte(lzw_minimum_code_size)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    for b in image_data:
        write_byte(b)

    # 7. Trailer
    write_byte(0x3B)