#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF will have a random width and height between [min_width; max_width] x [min_height; max_height].
# It will also have a random number of global color table entries between min_num_colors and max_num_colors.
# If include_gce is True, then a graphics control extension will be included as well.
# The minimum code size used by LZW compression will be randomly chosen from {1,..., 8}.
# The length of the data section will be randomly chosen from [min_image_data_length; max_image_data_length].
# Note that the maximum length of an image data block is 255 bytes.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, *,
        min_width=2**1, max_width=2**10,
        min_height=2**1, max_height=2**10,
        min_num_colors=1, max_num_colors=256,
        include_gce=True,
        min_image_data_length=2**1, max_image_data_length=2**10
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width = rng.randint(min_width, max_width)
    height = rng.randint(min_height, max_height)

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = rng.randint(min_num_colors, max_num_colors)
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if include_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left = rng.randint(0, width - 1)
    top = rng.randint(0, height - 1)
    width = rng.randint(1, width - left)
    height = rng.randint(1, height - top)
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = rng.randint(1, 8)

    blocks = []
    total_blocks_size = rng.randint(min_image_data_length, max_image_data_length)
    for _ in range(total_blocks_size):
        block_size = rng.randint(1, 255)  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)