#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    def write_byte(b: int) -> None:
        out.write(struct.pack('B', b))

    def write_short(s: int) -> None:
        out.write(struct.pack('H', s))

    def write_long(l: int) -> None:
        out.write(struct.pack('L', l))

    def write_string(s: str) -> None:
        out.write(s.encode())

    # 1. Header
    write_string("GIF89a")

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    write_short(width)
    write_short(height)
    write_byte(packed_fields)
    write_byte(background_color_index)
    write_byte(pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = [rng.read(3) for i in range(num_colors)]  # Random colors
    for rgb in color_table:
        write_byte(rgb[0])
        write_byte(rgb[1])
        write_byte(rgb[2])

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        write_byte(0x21)
        write_byte(0xF9)
        write_byte(0x04)

        delay_time = rng.read(1)[0]
        transparent_color_index = rng.read(1)[0]
        terminator = rng.read(1)[0]

        write_byte(delay_time)
        write_byte(transparent_color_index)
        write_byte(terminator)

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = rng.read(1)[0] & 1
    interlace_flag = rng.read(1)[0] & 1
    sort_flag = rng.read(1)[0] & 1
    reserved = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = [rng.read(3) for i in range(local_color_table_size)]
    for rgb in local_color_table:
        write_byte(rgb[0])
        write_byte(rgb[1])
        write_byte(rgb[2])

    write_byte(left)
    write_byte(top)
    write_byte(width)
    write_byte(height)
    write_byte((local_color_table_flag << 7) | (interlace_flag << 6) | (sort_flag << 5) | reserved)
    write_byte(local_color_table_size)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        if not block_size:
            break
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    write_byte(0x3B)