#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = min((rng.read(1)[0] % 256) or 1, 256)
    color_table = [rng.read(3) for _ in range(num_colors)]
    out.write(bytes().join(color_table))

    # 4. Graphics Control Extension (optional)
    if bool(rng.read(1)[0]):
        out.write(b'\x21\xF9\x04' + rng.read(2))

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = rng.read(1)[0] & 1
    interlace_flag = rng.read(1)[0] & 2
    sort_flag = rng.read(1)[0] & 4
    reserved_bits = rng.read(1)[0] & 7
    local_color_table_size = rng.read(1)[0] & 7
    out.write(struct.pack('>BHHHHBB', 0x2C, left, top, width, height, local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved_bits, local_color_table_size))

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(b'\x3B')