#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    out.write(color_table)

    # 4. Graphics Control Extension (optional)
    if bool(rng.read(1)[0]):
        # If we want one, let's add a GCE
        delay_time = rng.read(2)
        transparent_color_index = rng.read(1)[0]

        out.write(
            b'\x21\xF9\x04' +  # Application extension
            b'\x08' +           # Size of application data
            b'\x0A'             # Block size
            b'\x00\x00'         # No transparancy
            b'\x00' +           # Terminator
            delay_time          # Delay time
        )

        if transparent_color_index!= 0:
            out.write(transparent_color_index.to_bytes(length=1, byteorder='big'))

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = rng.read(1)[0] & 0x80 == 0x80
    interlace_flag = rng.read(1)[0] & 0x40 == 0x40
    sort_flag = rng.read(1)[0] & 0x20 == 0x20
    reserved = rng.read(1)[0] & 0x1F
    local_color_table_size = rng.read(1)[0] or 1

    out.write(
        b'\x2C' +       # Image descriptor
        struct.pack('>HHHBB', left, top, width, height, local_color_table_flag | interlace_flag << 2 | sort_flag << 3 | reserved << 5 | local_color_table_size)
    )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(b'\x3B')