#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a' + b'\x00\x00' * 7
    out.write(header)

    # 2. Logical Screen Descriptor
    lsd = bytearray([
        0b10000000 | (rng.read(1)[0] & 0b01111111), # Screen Width
        0b10000000 | (rng.read(1)[0] & 0b01111111), # Screen Height
        0b10000000 | (rng.read(1)[0] & 0b00001111), # Background Color Index
        0b10000000 | (rng.read(1)[0] & 0b00000111), # Pixel Aspect Ratio
    ])
    out.write(lsd)

    # 3. Global Color Table
    global_color_table = []
    for i in range(1 << ((lsd[4] >> 4) + 1)):
        color = [rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]]
        if sum(color) > 0:
            global_color_table += color
    out.write(struct.pack('B', len(global_color_table)))
    out.write(bytes(global_color_table))

    # 4. Image Descriptor
    id = bytearray([
        0b10000000,                               # Separator
        0b10000000 | (rng.read(1)[0] & 0b01111111), # Left Position
        0b10000000 | (rng.read(1)[0] & 0b01111111), # Top Position
        0b10000000 | (rng.read(1)[0] & 0b01111111), # Width
        0b10000000 | (rng.read(1)[0] & 0b01111111), # Height
        0b10000000 | (rng.read(1)[0] & 0b00000111), # Local Color Table Flag
        0b10000000 | (rng.read(1)[0] & 0b00000011), # Interlace Flag
        0b10000000 | (rng.read(1)[0] & 0b00000001), # Sort Flag
        0b10000000 | (rng.read(1)[0] & 0b00000001), # Reserved
        0b10000000 | (rng.read(1)[0] & 0b00000001), # Size of Local Color Table
    ])
    out.write(id)

    # 5. Local Color Table
    local_color_table = []
    for i in range(1 << (id[9] + 1)):
        color = [rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]]
        if sum(color) > 0:
            local_color_table += color
    out.write(struct.pack('B', len(local_color_table)))
    out.write(bytes(local_color_table))

    # 6. Raster Data
    while True:
        code = rng.read(1)[0]
        if code == 0xFF:
            break
        else:
            out.write(code.to_bytes(1, 'big'))
