#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), byteorder='little') % 640 + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % 480 + 1
    background_color_index = rng.read(1)[0] & 0xFF
    pixel_aspect_ratio = rng.read(1)[0] & 0xFF
    screen_descriptor = struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio)
    out.write(screen_descriptor)

    # 3. Global Color Table
    global_color_table_flag = (rng.read(1)[0] & 0b10000000) >> 7
    color_resolution = (rng.read(1)[0] & 0b01100000) >> 5
    sort_flag = (rng.read(1)[0] & 0b00010000) >> 4
    size_of_global_color_table = (rng.read(1)[0] & 0b00001111) + 1
    global_color_table = []
    for i in range(size_of_global_color_table * size_of_global_color_table * 3):
        global_color_table.append(rng.read(1)[0])
    global_color_table = bytes(global_color_table)
    out.write(struct.pack('B', global_color_table_flag | color_resolution << 4 | sort_flag << 3 | size_of_global_color_table))
    out.write(global_color_table)

    # 4. Image Blocks
    while True:
        image_separator = rng.read(1)[0]
        if image_separator == 0x2C:
            break
        else:
            continue

    # 5. Image Descriptor
    left_position = int.from_bytes(rng.read(2), byteorder='little') % width
    top_position = int.from_bytes(rng.read(2), byteorder='little') % height
    width = int.from_bytes(rng.read(2), byteorder='little') % (width - left_position) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (height - top_position) + 1
    local_color_table_flag = (rng.read(1)[0] & 0b10000000) >> 7
    interlace_flag = (rng.read(1)[0] & 0b01000000) >> 6
    sort_flag = (rng.read(1)[0] & 0b00100000) >> 5
    reserved = (rng.read(1)[0] & 0b00011111)
    image_descriptor = struct.pack('<HHBBB', left_position, top_position, width, height, local_color_table_flag | interlace_flag << 1 | sort_flag << 2 | reserved)
    out.write(image_descriptor)

    # 6. Local Color Table
    if local_color_table_flag:
        size_of_local_color_table = (rng.read(1)[0] & 0b00001111) + 1
        local_color_table = []
        for i in range(size_of_local_color_table * size_of_local_color_table * 3):
            local_color_table.append(rng.read(1)[0])
        local_color_table = bytes(local_color_table)
        out.write(struct.pack('B', size_of_local_color_table))
        out.write(local_color_table)

    # 7. LZW Minimum Code Size
    lzw_minimum_code_size = rng.read(1)[0] & 0xFF
    out.write(struct.pack('B', lzw_minimum_code_size))

    # 8. Image Data
    while True:
        block_size = rng.read(1)[0] & 0xFF
        if block_size == 0:
            break
        else:
            data = rng.read(block_size)
            out.write(data)

    # 9. Trailer
    trailer = rng.read(1)[0]
    assert trailer == 0x3B
    out.write(trailer)
