#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big') % 640 + 1
    height = int.from_bytes(rng.read(2), 'big') % 480 + 1
    background_color = rng.read(1)[0] & 0xF
    aspect_ratio = rng.read(1)[0] & 0x7
    global_color_table = bool(rng.read(1)[0] & 0x80)
    color_resolution = (rng.read(1)[0] >> 4) & 0x7
    sort_flag = bool(rng.read(1)[0] & 0x8)
    size_of_global_color_table = 2 ** ((rng.read(1)[0] & 0x7) + 1)
    bits_per_pixel = (rng.read(1)[0] >> 5) & 0x7
    out.write(struct.pack('<HBBB', width, height, background_color, aspect_ratio))

    # 3. Global Color Table
    if global_color_table:
        for _ in range(size_of_global_color_table):
            red = rng.read(1)[0]
            green = rng.read(1)[0]
            blue = rng.read(1)[0]
            out.write(struct.pack('BBB', red, green, blue))

    # 4. Image Data
    while True:
        # 4.1 Graphic Control Extension
        extension_introducer = b'\x21'
        graphic_control_label = b'\xf9'
        block_size = b'\x04'
        disposal_method = rng.read(1)[0] & 0x1C
        user_input_flag = bool(rng.read(1)[0] & 0x2)
        transparent_color_flag = bool(rng.read(1)[0] & 0x1)
        delay_time = rng.read(2)
        transparent_color_index = rng.read(1)[0]
        terminator = b'\x00'
        out.write(extension_introducer + graphic_control_label + block_size + struct.pack('B', disposal_method | user_input_flag << 2 | transparent_color_flag << 1) + delay_time + transparent_color_index + terminator)

        # 4.2 Image Descriptor
        image_separator = b'\x2c'
        left_position = rng.read(2)
        top_position = rng.read(2)
        width = rng.read(2)
        height = rng.read(2)
        local_color_table_flag = bool(rng.read(1)[0] & 0x80)
        interlace_flag = bool(rng.read(1)[0] & 0x40)
        sort_flag = bool(rng.read(1)[0] & 0x20)
        size_of_local_color_table = 2 ** ((rng.read(1)[0] & 0x7) + 1)
        bits_per_pixel = (rng.read(1)[0] >> 5) & 0x7
        out.write(image_separator + left_position + top_position + width + height + struct.pack('B', local_color_table_flag | interlace_flag << 1 | sort_flag << 2 | size_of_local_color_table - 1) + struct.pack('B', bits_per_pixel - 1))

        # 4.3 Local Color Table
        if local_color_table_flag:
            for _ in range(size_of_local_color_table):
                red = rng.read(1)[0]
                green = rng.read(1)[0]
                blue = rng.read(1)[0]
                out.write(struct.pack('BBB', red, green, blue))

        # 4.4 LZW Minimum Code Size
        lzw_minimum_code_size = rng.read(1)[0]
        out.write(lzw_minimum_code_size)

        # 4.5 Image Data
        while True:
            code_size = rng.read(1)[0]
            if code_size == 0:
                break
            else:
                out.write(code_size.to_bytes(1, byteorder='big'))

            # 4.5.1 Blocks of data
            while True:
                length = rng.read(1)[0]
                if length == 0:
                    break
                else:
                    out.write(length.to_bytes(1, byteorder='big