#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed = rng.read(1)
    background = rng.read(1)[0] & 0xFF
    aspect_ratio = rng.read(1)[0] & 0xFF
    
    # 3. Global Color Table
    global_color_table_flag = (packed[0] >> 7) & 0b1
    color_resolution = ((packed[0] >> 4) & 0b11) + 1
    sort_flag = (packed[0] >> 3) & 0b1
    size_of_global_color_table = 2 ** ((packed[0] & 0b111) + 1)
    global_color_table = [None] * size_of_global_color_table
    for i in range(size_of_global_color_table):
        red = rng.read(1)[0] & 0xFF
        green = rng.read(1)[0] & 0xFF
        blue = rng.read(1)[0] & 0xFF
        
        global_color_table[i] = (red << 16) | (green << 8) | blue
    
    # 4. Image Block
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    packed = rng.read(1)
    local_color_table_flag = (packed[0] >> 7) & 0b1
    interlace_flag = (packed[0] >> 6) & 0b1
    sort_flag = (packed[0] >> 5) & 0b1
    reserved = (packed[0] >> 4) & 0b1
    size_of_local_color_table = 2 ** ((packed[0] & 0b111) + 1)
    local_color_table = [None] * size_of_local_color_table
    for i in range(size_of_local_color_table):
        red = rng.read(1)[0] & 0xFF
        green = rng.read(1)[0] & 0xFF
        blue = rng.read(1)[0] & 0xFF
        
        local_color_table[i] = (red << 16) | (green << 8) | blue
    
    # 5. Graphic Control Extension
    block_size = rng.read(1)
    packed = rng.read(1)
    disposal_method = (packed[0] >> 2) & 0b111
    user_input_flag = (packed[0] >> 1) & 0b1
    transparent_color_flag = packed[0] & 0b1
    delay_time = rng.read(2)
    transparent_color_index = rng.read(1)
    
    # 6. Plain Text Extension
    block_size = rng.read(1)
    text_grid_left = rng.read(2)
    text_grid_top = rng.read(2)
    text_grid_width = rng.read(2)
    text_grid_height = rng.read(2)
    cell_width = rng.read(1)
    cell_height = rng.read(1)
    fg_color_index = rng.read(1)
    bg_color_index = rng.read(1)
    
    # 7. Application Extension
    block_size = rng.read(1)
    application_identifier = rng.read(8)
    application_authentication_code = rng.read(3)
    
    # 8. Comment Extension
    block_size = rng.read(1)
    comment = rng.read(1)
    
    # 9. Pixel Data
    lzw_minimum_code_size = rng.read(1)
    pixel_data = rng.read(1)
    
    # 10. Trailer
    trailer = rng.read(1)
    
    # Write everything to the output stream
    out.write(header)
    out.write(width)
    out.write(height)
    out.write(packed)
    out.write(struct.pack('B', background))
    out.write(struct.pack('B', aspect_ratio))
    for color in global_color_table:
        if color is None:
            continue
            
        red, green, blue = color
        out.write(struct.pack('BBB', red, green, blue))
    
    out.write(left)
    out.write(top)