#!/usr/bin/env python3

import struct
from typing import List, Tuple

from .constants import (
    AUDIO_FORMATS,
    BITRATE_MODES,
    CHANNEL_CONFIGURATIONS,
    SAMPLE_RATES,
)


def parse_audio_format(data: bytes):
    """Parse the audio format from a 2-byte integer.

    Args:
        data (bytes): The raw byte data to be parsed.

    Returns:
        str: The name of the audio format.
    """
    if len(data) != 2:
        raise ValueError("Expected two bytes for an audio format.")

    value = struct.unpack("<H", data)[0]
    return AUDIO_FORMATS[value]


def parse_bitrate_mode(data: bytes):
    """Parse the bitrate mode from a single byte.

    Args:
        data (bytes): The raw byte data to be parsed.

    Returns:
        str: The name of the bitrate mode.
    """
    if len(data) != 1:
        raise ValueError("Expected one byte for a bitrate mode.")

    value = struct.unpack("<B", data)[0]
    return BITRATE_MODES[value]


def parse_channel_configuration(data: bytes):
    """Parse the channel configuration from a single byte.

    Args:
        data (bytes): The raw byte data to be parsed.

    Returns:
        str: The name of the channel configuration.
    """
    if len(data) != 1:
        raise ValueError("Expected one byte for a channel configuration.")

    value = struct.unpack("<B", data)[0]
    return CHANNEL_CONFIGURATIONS[value]


def parse_sample_rate(data: bytes):
    """Parse the sample rate from a single byte.

    Args:
        data (bytes): The raw byte data to be parsed.

    Returns:
        str: The name of the sample rate.
    """
    if len(data) != 1:
        raise ValueError("Expected one byte for a sample rate.")

    value = struct.unpack("<B", data)[0]
    return SAMPLE_RATES[value]


def parse_frame_length(data: bytes):
    """Parse the frame length from a single byte.

    Args:
        data (bytes): The raw byte data to be parsed.

    Returns:
        int: The number of samples in each frame.
    """
    if len(data) != 1:
        raise ValueError("Expected one byte for a frame length.")

    value = struct.unpack("<B", data)[0]
    return 960 * pow(2, value - 1)


def parse_buffer_fullness(data: bytes):
    """Parse the buffer fullness from a 4-byte integer.

    Args:
        data (bytes): The raw byte data to be parsed.

    Returns:
        int: The number of bytes that can be held in the buffer.
    """
    if len(data) != 4:
        raise ValueError("Expected four bytes for a buffer fullness.")

    return struct.unpack("<I", data)[0]


def parse_number_of_raw_data_blocks_in_frame(data: bytes):
    """Parse the number of raw data blocks in a frame from a single byte.

    Args:
        data (bytes): The raw byte data to be parsed.

    Returns:
        int: The number of raw data blocks in a frame.
    """
    if len(data) != 1:
        raise ValueError(
            "Expected one byte for a number of raw data blocks in a frame."
        )

    return struct.unpack("<B", data)[0] + 1


def parse_profile(data: bytes):
    """Parse the profile from a single byte.

    Args:
        data (bytes): The raw byte data to be parsed.

    Returns:
        str: The name of the profile.
    """
    if len(data) != 1:
        raise ValueError("Expected one byte for a profile.")

    value = struct.unpack("<B", data)[0]
    return {
        0: "AAC Main",
        1: "AAC LC",
        2: "AAC SSR",
        3: "AAC LTP",
        4: "SBR",
        5: "AAC Scalable",
        6: "TwinVQ",
        7: "CELP",
        8: "HVXC",
        9: "TTSI",
        12: "Main Synthetic",
        13: "Wavetable Synthesis",
        14: "General MIDI",
        15: "Algorithmic Synthesis and Audio FX",
        16: "ER AAC LC",
        1