#!/usr/bin/env python3

import struct
from typing import BinaryIO


def read_bytes(file: BinaryIO, count: int) -> bytes:
    return file.read(count)


def write_bytes(file: BinaryIO, data: bytes):
    file.write(data)


def read_byte(file: BinaryIO) -> int:
    return struct.unpack('B', read_bytes(file, 1))[0]


def write_byte(file: BinaryIO, value: int):
    assert 0 <= value < 256
    write_bytes(file, struct.pack('B', value))


def read_short(file: BinaryIO) -> int:
    return struct.unpack('>H', read_bytes(file, 2))[0]


def write_short(file: BinaryIO, value: int):
    assert 0 <= value < 65536
    write_bytes(file, struct.pack('>H', value))


def read_int(file: BinaryIO) -> int:
    return struct.unpack('>I', read_bytes(file, 4))[0]


def write_int(file: BinaryIO, value: int):
    assert 0 <= value < 4294967296
    write_bytes(file, struct.pack('>I', value))


def read_long(file: BinaryIO) -> int:
    return struct.unpack('>Q', read_bytes(file, 8))[0]


def write_long(file: BinaryIO, value: int):
    assert 0 <= value < 18446744073709551616
    write_bytes(file, struct.pack('>Q', value))


def read_float(file: BinaryIO) -> float:
    return struct.unpack('>f', read_bytes(file, 4))[0]


def write_float(file: BinaryIO, value: float):
    write_bytes(file, struct.pack('>f', value))


def read_double(file: BinaryIO) -> float:
    return struct.unpack('>d', read_bytes(file, 8))[0]


def write_double(file: BinaryIO, value: float):
    write_bytes(file, struct.pack('>d', value))


def read_string(file: BinaryIO) -> str:
    length = read_short(file)
    if length == -1:
        return None
    else:
        return read_bytes(file, length).decode()


def write_string(file: BinaryIO, value: str):
    if value is None:
        write_short(file, -1)
    else:
        write_short(file, len(value))
        write_bytes(file, value.encode())


def read_boolean(file: BinaryIO) -> bool:
    return read_byte(file) != 0


def write_boolean(file: BinaryIO, value: bool):
    write_byte(file, 1 if value else 0)


def read_varint(file: BinaryIO) -> int:
    result = 0
    shift = 0
    while True:
        byte = read_byte(file)
        result |= (byte & 0x7F) << shift
        if byte & 0x80 == 0:
            break
        shift += 7
    return result


def write_varint(file: BinaryIO, value: int):
    assert 0 <= value < 2147483648
    while True:
        byte = value & 0x7F
        value >>= 7
        if value == 0:
            write_byte(file, byte)
            break
        write_byte(file, byte | 0x80)


def read_uuid(file: BinaryIO) -> str:
    return '%02X%02X%02X%02X-%02X%02X-%02X%02X-%02X%02X-%02X%02X%02X%02X%02X%02X' % tuple(struct.unpack('>BBBBHHHHHHHH', read_bytes(file, 16)))


def write_uuid(file: BinaryIO, value: str):
    write_bytes(file, struct.pack('>BBBBHHHHHHHH', *map(lambda x: int(x, 16), value.split('-'))))


def read_slot(file: BinaryIO) -> dict:
    id = read_short(file)
    if id == -1:
        return None
    else:
        amount = read_byte(file)
        damage = read_short(file)
        nbt = read_nbt(file)
        return {
            'id': id,
            'amount': amount,
            'damage': damage,
            'nbt':