#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('<BBB', *struct.unpack('>BBB', rng.read(3))) + rng.read(1)

    # 3. Global Color Table
    # Randomly choose whether to use a global color table or not
    if bool(rng.read(1)):
        # If so, randomly choose how many colors there are in it
        n_colors = 1 << (rng.read(1)[0] & 7)
        # Then, read that many color indices
        gct = rng.read(n_colors * 3)
    else:
        # Otherwise, no global color table
        gct = None

    # 4. Image Descriptors
    # Randomly choose how many images there will be
    n_images = rng.read(1)[0] % 5

    # For each image...
    for _ in range(n_images):
        # ...write an Image Descriptor
        idesc = struct.pack('<HHHBB', *struct.unpack('>HHHBB', rng.read(8)))

        # ...then write a Local Color Table
        if bool(rng.read(1)):
            # Randomly choose how many colors there are in it
            n_colors = 1 << (rng.read(1)[0] & 7)
            # Then, read that many color indices
            lct = rng.read(n_colors * 3)
        else:
            # Otherwise, no local color table
            lct = None

        # ...then write a Graphic Control Extension
        gce = struct.pack('<BBBB', *struct.unpack('>BBBB', rng.read(4)))

        # ...then write an Image Data block
        data = rng.read(width * height)

        # Write everything together
        out.write(header)
        out.write(lsd)
        if gct is not None:
            out.write(gct)
        out.write(idesc)
        if lct is not None:
            out.write(lct)
        out.write(gce)
        out.write(data)

        # Add a Trailer
        out.write(b'\x3b')

