#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd_fields = [
        0b00000000,
        rng.read(1)[0],
        rng.read(1)[0] % 256,
    ]

    # Write all of this to the output stream
    out.write(header + struct.pack('<HBBB', width, height, *lsd_fields))

    # 3. Global Color Table
    # We use a global color table with 256 colors, so we need 768 bytes
    # We also don't care about the actual color values, so we just read
    # 768 bytes from /dev/urandom
    out.write(rng.read(768))

    # 4. Image Data
    # We write one image block per pixel in the image.
    # Each image block contains a single-pixel graphic, which is either
    # transparent or has a random color index.
    for y in range(height):
        for x in range(width):
            # The image descriptor
            idesc = struct.pack('<BBB', 0x2C, x, y)
            # The image data
            ildata = struct.pack('<BBB', 0x00, rng.read(1)[0] % 256, 0xFF)
            # Write it all to the output stream
            out.write(idesc + ildata)

    # 5. Trailer
    # We don't care about the trailer, so we just write a zero byte
    out.write(struct.pack('B', 0x00))
