#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('HH', rng.read(4))
    # Packed fields:
    #   - Global Color Table Flag (0x80)
    #   - Color Resolution (0x70)
    #   - Sort Flag (0x08)
    #   - Size of Global Color Table (0x07)
    packed_fields = 0b10000000 | 0b01110000 | 0b00001000 | 0b00000111
    # Background color index
    background_color_index = rng.read(1)[0] & 0xff
    # Pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)[0] & 0xff
    
    # Write everything to the output file
    out.write(header + struct.pack('<HBBB', width, height, packed_fields, background_color_index) + pixel_aspect_ratio.to_bytes(1, 'little'))
    
    # 3. Global Color Table
    # Read 768 bytes from /dev/urandom
    global_color_table = rng.read(768)
    # Write them to the output file
    out.write(global_color_table)
    
    # 4. Image Descriptors
    # Number of images in this GIF
    num_images = rng.read(1)[0] & 0xff
    # Loop through all images
    for i in range(num_images):
        # 5. Image Separator
        image_separator = b'\x2c'
        
        # 6. Image Descriptor
        # Left position of the image
        left = rng.read(2)
        # Top position of the image
        top = rng.read(2)
        # Width of the image
        width = rng.read(2)
        # Height of the image
        height = rng.read(2)
        # Packed fields:
        #   - Local Color Table Flag (0x80)
        #   - Interlace Flag (0x40)
        #   - Sort Flag (0x20)
        #   - Reserved (0x1f)
        packed_fields = 0b10000000 | 0b01000000 | 0b00100000 | 0b00011111
        # LZW minimum code size
        lzw_minimum_code_size = rng.read(1)[0] & 0xff
        
        # Write everything to the output file
        out.write(image_separator + left + top + width + height + packed_fields.to_bytes(1, 'little') + lzw_minimum_code_size.to_bytes(1, 'little'))
        
        # 7. Local Color Table
        # Read 768 bytes from /dev/urandom
        local_color_table = rng.read(768)
        # Write them to the output file
        out.write(local_color_table)
        
        # 8. Image Data
        # Read 1 byte from /dev/urandom
        block_size = rng.read(1)[0] & 0xff
        # Write it to the output file
        out.write(block_size.to_bytes(1, 'little'))
        
        # 9. Trailer
        trailer = b';'
        # Write it to the output file
        out.write(trailer)
