#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''
    
    # 5. Image Descriptor
    id = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, 0, rng.read(1)[0])
    
    # 6. Local Color Table (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        local_color_table = struct.pack('>BB', 0x21, 0xFF) + rng.read(3 * num_colors)
    else:
        local_color_table = b''
    
    # 7. Raster Data
    raster_data = rng.read(width * height)
    
    # 8. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + id + local_color_table + raster_data + trailer)
