#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Packed fields
    packed = 0x80 | ((width >> 7) & 0x7) << 4 | (height >> 7)

    # Background color index
    background_color_index = rng.read(1)[0] % 256

    # Pixel aspect ratio
    pixel_aspect_ratio = 0

    # Write everything to the output stream
    out.write(header + struct.pack('<HBBB', width, height, packed, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table
    # We don't use it so we just write zeroes here
    out.write(b'\x00\x00\x00')

    # 4. Image Data
    # Loop until EOF
    while True:
        # Read one byte from the RNG
        control_byte = rng.read(1)[0]

        # If it's an extension block...
        if control_byte == 0x21:
            # ...read the next byte
            extension_type = rng.read(1)[0]

            # If it's a graphic control extension...
            if extension_type == 0xf9:
                # ...read the next three bytes
                flags, delay_time, transparent_color_index = struct.unpack('BBB', rng.read(3))

                # Write all of this to the output stream
                out.write(struct.pack('BBB', 0x21, 0xf9, 4) + struct.pack('BBB', flags, delay_time, transparent_color_index))

                # Skip over the rest of the extension block
                rng.seek(rng.tell() + 1)

            elif extension_type == 0xff:
                # ...read the next byte
                application_identifier = rng.read(1)[0]

                # ...and read the next two bytes
                application_authentication_code = rng.read(2)

                # Write all of this to the output stream
                out.write(struct.pack('BBB', 0x21, 0xff, 11) + struct.pack('BBB', application_identifier, *application_authentication_code))

                # Skip over the rest of the extension block
                rng.seek(rng.tell() + 1)

            else:
                # Otherwise skip over the whole extension block
                rng.seek(rng.tell() + 1)
        
        # If it's not an extension block...
        elif control_byte != 0x3b:
            # ...read the next two bytes
            left, top = struct.unpack('BB', rng.read(2))
            
            # ...and read the next two bytes
            width, height = struct.unpack('BB', rng.read(2))

            # Packed fields
            packed = 0x80 | ((left >> 7) & 0x7) << 4 | (top >> 7)

            local_color_table_flag = 0
            interlace_flag = 0
            sort_flag = 0

            # Write everything to the output stream
            out.write(struct.pack('BBBB', control_byte, left, top, packed) + struct.pack('BBBB', width, height, local_color_table_flag, interlace_flag, sort_flag))

            # 5. Local Color Table
            # We don't use it so we just write zeroes here
            out.write(b'\x00\x00\x00')

            # 6. Image Data
            # Loop until we hit the end of the image data
            while True:
                # Read one byte from the RNG
                code_size = rng.read(1)[0]
                
                # If it's the end of the image data...
                if code_size == 0:
                    # ...break out of the loop
                    break

                # Otherwise write the code size to the output stream
                out.write(bytes([code_size]))

                # Loop until we hit the end of the image data
                while True:
                    # Read one byte from the RNG
                    code = rng.read(1)[0]

                    # If it's the end of the image data...
                    if code == 0:
                        # ...break out of the loop
                        break

                    # Otherwise write the code to the output stream
                    out.write(bytes([code]))

        # If it's the end of the GIF...
        elif control_byte == 0x3b:
            # ...break out of the loop
            break
