#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 15 09:47:26 2020

@author: jonasg
"""
import numpy as np
from scipy.stats import norm, multivariate_normal
import matplotlib.pyplot as plt

def get_data(n=100):
    """Generates data from a bimodal Gaussian distribution."""
    
    # Generate data from two Gaussians with different means and covariance matrices
    mean1 = [0, 0]
    mean2 = [1, 1]
    cov1 = [[1, .5], [.5, 1]]
    cov2 = [[1, 0], [0, 1]]
    
    x1 = np.random.multivariate_normal(mean1, cov1, n)
    x2 = np.random.multivariate_normal(mean2, cov2, n)
    
    return np.vstack((x1, x2))

def plot_data(X):
    """Plots the data points in X."""
    
    plt.figure()
    plt.scatter(X[:, 0], X[:, 1])
    plt.show()

def plot_density(X, mu, Sigma):
    """Plots the density of the multivariate normal distribution N(mu, Sigma)."""
    
    # Create grid to evaluate the PDF on
    x1 = np.linspace(-3, 3, 100)
    x2 = np.linspace(-3, 3, 100)
    xx1, xx2 = np.meshgrid(x1, x2)
    
    # Evaluate PDF for each point in the grid
    pdf = multivariate_normal.pdf(np.array([xx1.ravel(), xx2.ravel()]).T, mu, Sigma)
    pdf = pdf.reshape(xx1.shape)
    
    # Plot PDF
    plt.figure()
    plt.contourf(xx1, xx2, pdf, cmap='Blues')
    plt.colorbar()
    plt.scatter(X[:, 0], X[:, 1])
    plt.show()

def gaussian_mixture_model(X, K=2, max_iter=100, tol=1e-4):
    """Fits a Gaussian mixture model (GMM) to the data in X.
    
    The algorithm is based on expectation maximization. At each iteration, we update
    the parameters of the model by maximizing the log likelihood of the data under
    the current model parameters. We repeat this until convergence or the maximum
    number of iterations has been reached.
    
    Inputs:
        X: Data matrix where each row represents an observation.
        K: Number of components in the mixture model.
        max_iter: Maximum number of iterations to perform.
        tol: Convergence tolerance. The algorithm stops if the relative change in
            the log likelihood between two iterations is smaller than tol.
        
    Returns:
        mus: List of component means. Each entry corresponds to one component.
        sigmas: List of covariance matrices. Each entry corresponds to one component.
        weights: List of weights for each component. Sum should be equal to 1.
        log_likelihoods: Log likelihood values at each iteration.
    """
    
    # Initialize parameters randomly
    mus = []
    sigmas = []
    weights = []
    log_likelihoods = []
    
    for k in range(K):
        mus.append(np.random.uniform(-1, 1, size=(2)))
        sigmas.append(np.eye(2))
        weights.append(1 / K)
    
    # Run EM algorithm
    for it in range(max_iter):
        
        # Expectation step
        responsibilities = []
        for i in range(len(X)):
            
            # Compute the probability that each component generated the current data point
            probabilities = []
            for k in range(K):
                probabilities.append(weights[k] * multivariate_normal.pdf(X[i], mus[k], sigmas[k]))
                
            # Normalize probabilities so that they sum up to 1
            responsibilities.append(probabilities / np.sum(probabilities))
        
        # Maximization step
        new_mus = []
        new_sigmas = []
        new_weights = []
        for k in range(K):
            
            # Compute weighted average of the data points assigned to this component
            x_k = np.zeros((2, len(responsibilities)))
            for i in range(len(X)):
                x_k += responsibilities[i][k] * X[i]
            new_mus.append(x_k / np.sum(responsibilities[:, k]))
            
            # Compute weighted average of the covariance matrices assigned