#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_gif(out: BinaryIO, rng: BinaryIO):
    # 10 bytes of header
    out.write(b"GIF89a")
    width = rng.read(2)
    height = rng.read(2)
    out.write(width + height)
    # 7 bytes of global color table
    out.write(b"\xFF\xFF\xFF\x00\x00\x00\x00")
    # 1 byte for background color index
    out.write(b"\x00")
    # 1 byte for pixel aspect ratio
    out.write(b"\x00")
    # 1 byte for flags (no local color table)
    out.write(b"\x00")
    
    while True:
        # 1 byte for block type
        btype = rng.read(1)
        if not btype:
            break
        
        # 1 byte for block length
        blen = rng.read(1)
        if not blen:
            raise Exception("Unexpected EOF")
        
        # 1 byte for delay time
        delay = rng.read(1)
        if not delay:
            raise Exception("Unexpected EOF")
        
        # 1 byte for transparent color index
        tci = rng.read(1)
        if not tci:
            raise Exception("Unexpected EOF")
        
        # 1 byte for terminator
        term = rng.read(1)
        if not term:
            raise Exception("Unexpected EOF")
        
        if btype == b"\x2C":
            # Image descriptor
            
            # 2 bytes for image left position
            lpos = rng.read(2)
            if not lpos:
                raise Exception("Unexpected EOF")
            
            # 2 bytes for image top position
            tpos = rng.read(2)
            if not tpos:
                raise Exception("Unexpected EOF")
            
            # 2 bytes for image width
            iw = rng.read(2)
            if not iw:
                raise Exception("Unexpected EOF")
            
            # 2 bytes for image height
            ih = rng.read(2)
            if not ih:
                raise Exception("Unexpected EOF")
            
            # 1 byte for flags (no local color table)
            flags = rng.read(1)
            if not flags:
                raise Exception("Unexpected EOF")
            
            # 1 byte for terminator
            term = rng.read(1)
            if not term:
                raise Exception("Unexpected EOF")
            
            # 3 bytes per pixel
            pixels = rng.read(struct.unpack("<H", iw)[0] * struct.unpack("<H", ih)[0])
            if len(pixels) != struct.unpack("<H", iw)[0] * struct.unpack("<H", ih)[0]:
                raise Exception("Unexpected EOF")
            
            out.write(btype + blen + delay + tci + term + lpos + tpos + iw + ih + flags + term + pixels)
        elif btype == b"\x21":
            # Extension
            
            # 1 byte for extension label
            elabel = rng.read(1)
            if not elabel:
                raise Exception("Unexpected EOF")
            
            # 1 byte for block length
            blen = rng.read(1)
            if not blen:
                raise Exception("Unexpected EOF")
            
            if elabel == b"\xF9":
                # Graphic control extension
                
                # 1 byte for flags
                flags = rng.read(1)
                if not flags:
                    raise Exception("Unexpected EOF")
                
                # 2 bytes for delay time
                delay = rng.read(2)
                if not delay:
                    raise Exception("Unexpected EOF")
                
                # 1 byte for transparent color index
                tci = rng.read(1)
                if not tci:
                    raise Exception("Unexpected EOF")
                
                # 1 byte for terminator
                term = rng.read(1)
                if not term:
                    raise Exception("Unexpected EOF")
                
                out.write(btype + blen + elabel + blen + flags + delay + tci + term)
            else:
                # Unknown extension
                
                # Read rest of block
                data = rng.read(struct.unpack("<B", blen)[0])
                if len(data) != struct.unpack("<B", blen)[0]:
                    raise Exception("Unexpected EOF")
                
                out.write(btype + blen + elabel + blen + data)
        elif btype == b"\x3B":
            # Trailer
