#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
    height = 1 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
    flags = b'\x07\x00'   # Global color table, sorted, no gct size specified
    bg_color = bytes([int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256]) * 3
    aspect_ratio = b'\x00'    # No aspect ratio given
    out.write(struct.pack('>HHBBB', width, height, *flags))
    out.write(bg_color)
    out.write(aspect_ratio)

    # Global color table (palette)
    palette = [bytes([int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256]) * 3]
    for i in range(1, 256):
        palette += [palette[i - 1][j] ^
                     bytes([(int.from_bytes(rng.read(1), byteorder='big', signed=False)
                             ^ palette[i - 1][j]) % 256])
                     for j in range(3)]
    out.write(b''.join(palette))

    # Image descriptor
    left = 0
    top = 0
    width = 1 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
    height = 1 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
    flags = b'\x04\x00'   # LCT, interlace not used, local color table present
    lct_size = len(palette).to_bytes(1, 'big')
    out.write(struct.pack('>HHHHB', left, top, width, height, *flags))
    out.write(lct_size)

    # Local color table (palette)
    out.write(b''.join(palette))

    # Image data
    image_data = []
    for y in range(height):
        line = []

        # Encode each pixel as an index from the palette
        for x in range(width):
            idx = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
            line += [idx.to_bytes(1, 'big')]

        # Add padding to make the line divisible by 8 bits
        if len(line) % 8!= 0:
            line += [(len(line) % 8).to_bytes(1, 'big')] * \
                    ((8 - len(line) % 8) // 8)

        image_data += line

    # Write encoded pixels
    for line in image_data:
        out.write(b''.join(line))

