#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    header = b'GIF89a' + bytes([0x12, 0x34])

    # Logical screen descriptor
    width = rng.read(1)[0] % 65536
    height = rng.read(1)[0] % 65536
    flags = (
        0b10000000 |  # Global color table flag
        0b00000000 |  # Sorting flag
        0b00000000 |  # Size of global color table
        0b00000000    # Background color index
    )
    background_color_index = 0
    pixel_aspect_ratio = 0
    logical_screen_descriptor = struct.pack('<HHBBB',
                                            width, height, flags,
                                            background_color_index,
                                            pixel_aspect_ratio)

    # Color table
    colors = [bytes([rng.read(1)[0],
                     rng.read(1)[0],
                     rng.read(1)[0]]) for _ in range(256)]
    color_table = b''.join(colors)

    # Image descriptor
    left = 0
    top = 0
    width = rng.read(1)[0] % 65536
    height = rng.read(1)[0] % 65536
    flags = (
        0b10000000 |   # Local color table flag
        0b01000000 |   # Interlace flag
        0b00000000     # Sorting flag
    )
    local_color_table_size = len(colors) - 1
    local_color_table = b'' if local_color_table_size == 0 else \
                        struct.pack('B', local_color_table_size)
    image_descriptor = struct.pack('<HHHHBB',
                                   left, top, width, height, flags,
                                   local_color_table_size)

    # Image data
    lzw_min_code_size = 2
    subblocks = []
    while True:
        code_length = rng.read(1)[0] % 128
        if code_length < lzw_min_code_size:
            code_length += lzw_min_code_size

        if code_length == 127:
            break

        block = bytearray()
        for i in range(code_length):
            block.append(rng.read(1)[0])

        subblocks.append(block)

    # LZW minimum code size
    subblocks[0][0] = lzw_min_code_size

    # Sub-blocks
    subblocks = [struct.pack('B', len(subblock)) +
                 subblock for subblock in subblocks]

    # Trailer
    trailer = b'\x3B'

    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(color_table)
    out.write(image_descriptor)
    out.write(b''.join(subblocks))
    out.write(trailer)
