#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Header
    header = b'GIF89a' + bytes([0x12, 0x34]) + \
             (b'\x00' * 6) + b'\xff\xff\xff'
    out.write(header)

    # Logical screen descriptor
    logicalScreenDescriptor = struct.pack('>HHBBB',
        0x7800, 0x5600, 0xf0, 0x00, 0x00
    )
    out.write(logicalScreenDescriptor)

    # Global color table
    globalColorTable = [0] * 256 * 3
    for i in range(len(globalColorTable)):
        globalColorTable[i] = int.from_bytes(
            rng.read(1), byteorder='big') & 0x3f
    globalColorTable = bytes(globalColorTable)
    out.write(globalColorTable)

    # Image descriptor
    imageDescriptor = struct.pack('>HHBBBBBH',
        0x0000, 0x0000, 0x00, 0x00, 0x00, 0x00, 0x0000
    )
    out.write(imageDescriptor)

    # Local color table
    localColorTable = [0] * 256 * 3
    for i in range(len(localColorTable)):
        localColorTable[i] = int.from_bytes(
            rng.read(1), byteorder='big') & 0x3f
    localColorTable = bytes(localColorTable)
    out.write(localColorTable)

    # Image data
    while True:

        # Block size
        blockSize = int.from_bytes(
            rng.read(1), byteorder='big') & 0x07

        if blockSize == 0: break

        # LZW minimum code size
        lzwMinimumCodeSize = int.from_bytes(
            rng.read(1), byteorder='big') & 0x07

        # Sub blocks
        subBlocks = []
        while True:

            # Data length
            dataLength = int.from_bytes(
                rng.read(1), byteorder='big') & 0x07

            # Data
            data = rng.read(dataLength)

            # End of block?
            if len(data) < dataLength or data[-1]!= 0x00:

                # Add to subblocks
                subBlocks.append((lzwMinimumCodeSize, data))

                # Break loop
                break

        # Write sub blocks
        for s in subBlocks:

            # Write block size and LZW minimum code size
            out.write(struct.pack('B', s[0]))
            out.write(struct.pack('B', s[1][-1]))

            # Write data
            out.write(s[1])

