#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(num_colors)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBBBB', 0, 0, 0b10000000 | rng.read(1)[0], 0, 0, 0)

    # 5. Local Color Table
    local_color_table = []
    if rng.read(1)[0]:
        num_colors = (rng.read(1)[0] % 256) or 1
        local_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(num_colors)]

    # 6. Extension blocks
    extension_blocks = []
    while True:
        block_type = rng.read(1)[0]

        if not block_type:
            break

        extension_block = struct.pack('BB', block_type, 0) + rng.read((rng.read(1)[0] % 256) or 1)

        extension_blocks += [extension_block]

    # 7. Graphics Control Block
    graphics_control_block = struct.pack('BBBBBBB', 0x21, 0xF9, 0x04, 0x05, 0xFF, 0x00)

    # 8. Image Data
    image_data = rng.read((width * height // 8) or 1)

    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(bytes().join([*global_color_table]))
    out.write(image_descriptor)
    out.write(bytes().join([*local_color_table]))
    out.write(bytes().join([*extension_blocks]))
    out.write(graphics_control_block)
    out.write(image_data)