#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None
        
    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHBB', 0x2C, 0, 0, 0, 0)

    # 6. Image Data (compressed with RLE)
    # Let's use a simple compression scheme:
    # - If the next pixel is different from the previous one, write a new run of pixels until we find a difference
    # - Otherwise, increase the length of the current run by 1
    # This will result in runs of up to 257 pixels, which should be enough for our purposes
    # We'll also add an extra 0-length run so that the last pixel can be written as well
    data = [b'\x00'] + [color_table[i % len(color_table)] for i in range(width*height+1)]
    compressed_data = []
    prev_pixel = data[0]
    while True:
        curr_pixel = data[-1]
        
        if curr_pixel!= prev_pixel:
            count = 1
            
            while not data[-count] == curr_pixel:
                count += 1
                
            compressed_data.append(struct.pack('>BB', 0, count))
            del data[-count:]
            
        elif len(data) > 1:
            compressed_data.append(struct.pack('>B', 0xFF))
            del data[-1]
            
        else:
            break
            
    image_data = b''.join(compressed_data)
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce:
        out.write(gce)
    out.write(image_descriptor)
    out.write(image_data)
    out.write(trailer)

