#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
	width &= 0x7FFF
	height &= 0x7FFF

    # Color table size is fixed to 256 colors
    color_table_size = 2 ** ((rng.read(1)[0] & 0b11) + 1)

    # Background color index is always zero
    background_color_index = 0

    # Pixel aspect ratio is always 0
    pixel_aspect_ratio = 0

    logical_screen_descriptor = struct.pack(
        '>HHBBB',
        width, height,
        color_table_size,
        background_color_index,
        pixel_aspect_ratio
    )

    # 3. Global Color Table
    global_color_table = bytearray()
    for i in range(color_table_size):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        global_color_table += struct.pack('BBB', r, g, b)

    # 4. Image Descriptor
    image_left_position = 0
    image_top_position = 0
    image_width = width
    image_height = height

    # Transparent color index is always zero
    transparent_color_index = 0

    # LZW minimum code size is always 2
    lzw_minimum_code_size = 2

    image_descriptor = struct.pack(
        '>HHHHBB',
        image_left_position, image_top_position,
        image_width, image_height,
        local_color_table_flag=False,
        interlace_flag=False,
        sort_flag=False,
        reserved_bits=(transparent_color_index << 4) | lzw_minimum_code_size
    )

    # 5. Local Color Table
    local_color_table = bytearray()

    # 6. Image Data
    # The first byte of the data stream must be an LZW clear code
    image_data = [lzw_clear_code]

    # Generate random pixels until we reach the end of the screen
    while True:
        # Each pixel is encoded as three bits per channel
        # We use one byte to store all three channels
        r = rng.read(1)[0] % 8
        g = rng.read(1)[0] % 8
        b = rng.read(1)[0] % 8

        # Pack them together
        rgb = (r << 5) | (g << 2) | b

        if len(image_data) > 2**lzw_minimum_code_size - 1:
            raise Exception("Image too large")

        # If this RGB combination has not been encountered yet, add it to the
        # dictionary and append its corresponding LZW code to the data stream
        if rgb not in lzw_dictionary:
            lzw_dictionary[rgb] = len(lzw_dictionary)

            # Append the new code to the data stream
            image_data.append(len(lzw_dictionary))
        else:
            # Otherwise, just look up the corresponding code in the dictionary
            image_data.append(lzw_dictionary[rgb])

        # Check if we've reached the bottom right corner
        if image_left_position == width-1 and image_top_position == height-1:
            break

        # Move on to next row or column depending on current position
        if image_left_position < width-1:
            image_left_position += 1
        elif image_top_position < height-1:
            image_left_position = 0
            image_top_position += 1

    # Add final LZW end of information code to the data stream
    image_data.append(lzw_end_of_information_code)

    # Convert the data stream from integers to bytes
    image_data = bytes(image_data)

    # 7. Trailer
    trailer = b''

    # Write everything to output file
    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(global_color_table)
    out.write(image_descriptor)
    out.write(local_color_table)
    out.write(image_data)
    out.write(trailer)


if