#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor (10 bytes)
    width = 256 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 7
    height = 256 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 7
    flags = b'\x80\x00' if bool(int.from_bytes(rng.read(1), byteorder='big', signed=False)) else b'\x00\x00'
    bg_color = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xf0 | \
        int.from_bytes(rng.read(1), byteorder='big', signed=False) >> 4
    aspect_ratio = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
    out.write(struct.pack('>HHBBB', width, height, flags[0], flags[1], bg_color, aspect_ratio))

    # Global color table (3 * 256 bytes)
    for _ in range(0, 256):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor (10 bytes)
    left = 0
    top = 0
    width = 256 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 7
    height = 256 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 7
    flags = b'\x80\x00' if bool(int.from_bytes(rng.read(1), byteorder='big', signed=False)) else b'\x00\x00'
    lct = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x01
    interlace = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x02
    sort = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x04
    reserved = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xF8
    out.write(struct.pack('>HHHHBB', left, top, width, height, flags[0], flags[1]))

    # Local color table (3 * 256 bytes)
    if lct == 1:
        for _ in range(0, 256):
            r = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
            g = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
            b = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
            out.write(struct.pack('BBB', r, g, b))

    # Image data block
    for y in range(height):
        lzw_mincodesize = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff

        codesize = min(lzw_mincodesize, 12)
        clearcode = 1 << codesize
        endofinfo = clearcode + 1
        eoi = endofinfo + 1
        maxcode = 1 << codesize
        bitmask = (1 << codesize) - 1

        nextcode = None
        prevcode = None
        bitsleft = 0
        buf = []

        def flush():
            nonlocal nextcode

            if len(buf) > 0 and nextcode is not None:
                out.write(struct.pack('B', nextcode))

                for c in buf[:-1]:
                    out.write(struct.pack('B', c))

            elif len(buf) > 0:
                raise Exception("Invalid LZW stream")

        while True:
            if nextcode is None or bitsleft <= 0:
                if len(buf) >= 2 and buf[-1] == clearcode and buf[-2] == clearcode:
                    break

                if len(buf) < 2 or buf[-1]!= clearcode or buf[-2]!= clearcode:
                    flush()

                    if len(buf) >= 2 and buf[-1] == clearcode and buf[-2] == clearcode:
                        break

                    if len(buf) >= 2 and buf[-1] == clearcode and buf[-2]!= clearcode:
                        raise Exception("Invalid LZW stream")

                    if len(buf) >= 2 and buf[-1]!= clearcode and buf[-2] == clearcode:
                        raise Exception("Invalid LZW stream")