#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    header = b'GIF89a' + bytes([0x12, 0x34])

    # Logical screen descriptor
    width = int.from_bytes(rng.read(2), byteorder='little') % (2**16 - 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (2**16 - 1) + 1
    flags = [int.from_bytes(rng.read(1), byteorder='little')] * 7
    bgcolorindex = int.from_bytes(rng.read(1), byteorder='little') & 0b11111
    aspectratio = int.from_bytes(rng.read(1), byteorder='little') & 0b11111

    lsd = struct.pack('<6sHHBBB',
                      bytes(header),
                      0,
                      width,
                      height,
                      flags[0],
                      flags[1],
                      flags[2],
                      bgcolorindex,
                      aspectratio)

    # Global color table
    gctsize = int.from_bytes(rng.read(1), byteorder='little') & 0b11111
    gct = []
    for _ in range(gctsize):
        r = int.from_bytes(rng.read(1), byteorder='little') & 0b1111111
        g = int.from_bytes(rng.read(1), byteorder='little') & 0b1111111
        b = int.from_bytes(rng.read(1), byteorder='little') & 0b1111111

        gct += [r, g, b]

    # Image descriptor
    left = int.from_bytes(rng.read(2), byteorder='little') & 0b111111111111
    top = int.from_bytes(rng.read(2), byteorder='little') & 0b111111111111
    width = int.from_bytes(rng.read(2), byteorder='little') % (2**16 - 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (2**16 - 1) + 1
    flags = [int.from_bytes(rng.read(1), byteorder='little')] * 7
    localcolortableflag = bool(flags[5] >> 7)
    interlaceflag = bool(flags[5] >> 6)
    sortflag = bool(flags[5] >> 5)
    reserved = flags[5] & 0b111
    localcolortablesize = flags[5] & 0b11111

    id = struct.pack('<HBBHHB',
                     left,
                     top,
                     width,
                     height,
                     flags[0],
                     flags[1])

    # Local color table
    if localcolortableflag:
        lct = []
        for _ in range(localcolortablesize):
            r = int.from_bytes(rng.read(1), byteorder='little') & 0b1111111
            g = int.from_bytes(rng.read(1), byteorder='little') & 0b1111111
            b = int.from_bytes(rng.read(1), byteorder='little') & 0b1111111

            lct += [r, g, b]

    # Data sub-block
    data = []
    while True:
        blocktype = int.from_bytes(rng.read(1), byteorder='little')
        if blocktype == 0x2C:
            break

        elif blocktype!= 0x21:
            raise Exception('Invalid block type!')

        else:
            length = int.from_bytes(rng.read(1), byteorder='little')
            data += list(rng.read(length))

    # Trailer
    trailer = b'\x3B'

    out.write(lsd)
    if len(gct) > 0:
        out.write(struct.pack('<B', gctsize))
        for i in range(len(gct)):
            out.write(struct.pack('<BBB',
                                  gct[i][0],
                                  gct[i][1],
                                  gct[i][2]))

    out.write(id)
    if localcolortableflag and len(lct) > 0:
        out.write(struct.pack('<B', localcolortablesize))
        for i in range(len(lct)):
            out.write(struct.pack('<BBB',
                                  lct[i][0],
                                  lct[i][1],
                                  lct[i][2]))

    for d in data:
        out.write(d)

    out.write(trailer)

