#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data
    data = []
    while True:
        # Each block is either an RLE block or a raw block
        if rng.read(1)[0] & 1 == 0:
            # Raw Block

            # The first byte tells us how many pixels to copy from the previous frame
            num_pixels = rng.read(1)[0] + 1

            # Then we read that many bytes of pixel data
            data += list(rng.read(num_pixels))

        else:
            # RLE Block

            # We get the same information as in the raw block
            num_pixels = rng.read(1)[0] + 1
            pixel = rng.read(1)

            # And then we repeat those pixels this many times
            data += [pixel[0]] * num_pixels

        # If we've reached our target height, break out of the loop
        if len(data) >= height * width:
            break

    # Now let's compress all these blocks together!
    compressed_blocks = []
    current_block = []
    for i in range(len(data)):
        # Add the next pixel to the current block
        current_block.append(data[i])

        # If the block is full...
        if len(current_block) == 255:
            #...add it to the list of compressed blocks...
            compressed_blocks.append(bytes([
                0b11100000 | len(current_block) - 1
            ]))
            compressed_blocks.append(bytes(current_block))

            #...and start a new one
            current_block = []

    # If there are any leftover pixels in the last block, add them now too
    if len(current_block):
        compressed_blocks.append(bytes([
            0b11100000 | len(current_block) - 1
        ]))
        compressed_blocks.append(bytes(current_block))

    # Finally, let's write everything out!
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    out.write(b''.join(compressed_blocks))

