#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# returns its size in bytes.
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header (6 bytes)
    out.write(b'GIF89a')

    # Logical screen descriptor (7 bytes)
    width = 1 + int.from_bytes(rng.read(2), byteorder='little', signed=False) % 50
    height = 1 + int.from_bytes(rng.read(2), byteorder='little', signed=False) % 50
    out.write((width).to_bytes(2, byteorder='little'))
    out.write((height).to_bytes(2, byteorder='little'))
    out.write(b'\x87\x00')

    # Global color table (3 bytes)
    num_colors = min(int.from_bytes(rng.read(1), byteorder='little'), 256)
    out.write(((num_colors << 4) | 0xf).to_bytes(1, byteorder='little'))

    for _ in range(num_colors):
        r = int.from_bytes(rng.read(1), byteorder='little') & 0xff
        g = int.from_bytes(rng.read(1), byteorder='little') & 0xff
        b = int.from_bytes(rng.read(1), byteorder='little') & 0xff
        out.write((r).to_bytes(1, byteorder='little'))
        out.write((g).to_bytes(1, byteorder='little'))
        out.write((b).to_bytes(1, byteorder='little'))

    # Image descriptor (9 bytes)
    left = 0
    top = 0
    width = 1 + int.from_bytes(rng.read(2), byteorder='little', signed=False) % 50
    height = 1 + int.from_bytes(rng.read(2), byteorder='little', signed=False) % 50
    flags = 0x80 | ((int.from_bytes(rng.read(1), byteorder='little') & 0x07) << 4)
    out.write((left).to_bytes(2, byteorder='little'))
    out.write((top).to_bytes(2, byteorder='little'))
    out.write((width).to_bytes(2, byteorder='little'))
    out.write((height).to_bytes(2, byteorder='little'))
    out.write((flags).to_bytes(1, byteorder='little'))

    # Local color table (3 bytes)
    num_colors = min(int.from_bytes(rng.read(1), byteorder='little'), 256)
    out.write(((num_colors << 4) | 0xf).to_bytes(1, byteorder='little'))

    for _ in range(num_colors):
        r = int.from_bytes(rng.read(1), byteorder='little') & 0xff
        g = int.from_bytes(rng.read(1), byteorder='little') & 0xff
        b = int.from_bytes(rng.read(1), byteorder='little') & 0xff
        out.write((r).to_bytes(1, byteorder='little'))
        out.write((g).to_bytes(1, byteorder='little'))
        out.write((b).to_bytes(1, byteorder='little'))

    # Image data block (1 byte)
    lzw_min_code_size = int.from_bytes(rng.read(1), byteorder='little')
    out.write(lzw_min_code_size.to_bytes(1, byteorder='little'))

    # LZW code stream (variable length)
    def write_lzw_code():
        if len(lzw_code_stream) == 0:
            return

        while True:
            chunk_len = min(len(lzw_code_stream), 255 - 2 * lzw_nbits)

            if chunk_len > 0:
                out.write(((chunk_len << lzw_nbits) | lzw_code_stream[0]).to_bytes(1, byteorder='little'))

                del lzw_code_stream[:chunk_len]

            if len(lzw_code_stream) == 0 or len(lzw_code_stream) < 2 ** lzw_nbits:
                break

        if len(lzw_code_stream) >= 2 ** lzw_nbits:
            lzw_nbits += 1

    lzw_nbits = lzw_min_code_size + 1
    lzw_prev_code = None
    lzw_code_stream = []

    while True:
        if lzw_prev_code is not None:
            lzw_code_stream.append(lzw_prev_code)

        lzw_prev_code = int.from_bytes(rng.read(1), byteorder='little') & 0xff

        if lzw_prev_code == 0xe0:
            write_lzw_code()
            break

        if lzw_prev_code == 0x00:
            continue

        if lzw_prev_code == 0xff:
            lzw_prev_code = None
            continue

       