#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write('GIF89a'.encode())

    width = 1024
    height = 768

    # Logical screen descriptor (LSD)
    out.write(struct.pack('<HHBBB',
        width, height,
        0xf0 | 0x01,      # Global color table flag and size of global color table
        0,                # Background color index
        0))               # Pixel aspect ratio

    # Global color table
    for i in range(0, 256 * 3, 3):
        r = next(rng) & 0xff
        g = next(rng) & 0xff
        b = next(rng) & 0xff

        out.write(struct.pack('<BBB',
            r, g, b))

    # Image descriptor (ID)
    out.write(struct.pack('<BHHBB',
        0x2c,             # Image separator
        0, 0,             # Left position, top position
        width,            # Width
        height,           # Height
        0))               # Flags

    # Local color table
    for i in range(0, 256 * 3, 3):
        r = next(rng) & 0xff
        g = next(rng) & 0xff
        b = next(rng) & 0xff

        out.write(struct.pack('<BBB',
            r, g, b))

    # LZW minimum code size byte
    out.write(b'\x08')

    # Image data
    lzw_mincodesize = 8
    lzw_clearcode = 1 << lzw_mincodesize
    lzw_endofinfo = lzw_clearcode + 1
    lzw_endofdata = lzw_endofinfo + 1

    lzw_dictionary = {}
    lzw_dictionary_size = len(lzw_dictionary)
    lzw_next_code = lzw_endofdata + 1

    def lzw_add_to_dictionary(byte):
        if not isinstance(byte, bytes):
            raise TypeError("Expected bytes")

        if len(byte)!= 1:
            raise ValueError("Expected exactly one byte")

        if byte[0] in lzw_dictionary:
            return False

        lzw_dictionary[byte[0]] = lzw_next_code
        lzw_dictionary_size += 1
        lzw_next_code += 1

        return True

    def lzw_get_from_dictionary(byte):
        if not isinstance(byte, bytes):
            raise TypeError("Expected bytes")

        if len(byte)!= 1:
            raise ValueError("Expected exactly one byte")

        result = None

        try:
            result = lzw_dictionary[byte[0]]
        except KeyError as e:
            pass

        return result

    def lzw_emit_code(code):
        if not isinstance(code, int):
            raise TypeError("Expected integer")

        if code < 0 or code > 2**lzw_mincodesize - 1:
            raise ValueError("Code is outside valid range")

        if code == lzw_clearcode:
            lzw_mincodesize = 8

            lzw_dictionary = {
                b'\x00': lzw_clearcode,
                b'\x01': lzw_endofinfo,
                b'\xfe': lzw_endofdata}

            lzw_dictionary_size = len(lzw_dictionary)
            lzw_next_code = lzw_endofdata + 1

            return

        if code == lzw_endofdata:
            out.write(b'\x00')
            return

        if code == lzw_endofinfo:
            out.write(b'\xff')
            return

        if code <= lzw_next_code - lzw_dictionary_size:
            out.write(bytes([code]))
            return

        if lzw_mincodesize >= 12:
            raise RuntimeError("Dictionary too large to encode image")

        lzw_mincodesize += 1

        lzw_add_to_dictionary(bytes([code >> lzw_mincodesize]))
        lzw_emit_code((code >> lzw_mincodesize) ^
                      ((lzw_next_code - lzw_dictionary_size)
                       if code & (2 ** lzw_mincodesize - 1) else 0))

        lzw_add_to_dictionary(bytes([(code << (8 - lzw_mincodesize))
                                     & 0xff]))
        lzw_emit_code(((code << (8 - lzw_mincodesize))
                       & 0xff) ^
                      ((lzw_next_code - lzw_dictionary_size)
                       if code & (2 ** lzw_mincodesize - 1) else 0))


    lzw_emit_code(lzw_clearcode)

    while True:
        block_type = next(rng) % 4

        if block_type == 0:
            # Plain text
            length = next(rng) % 128

            if length == 0:
                continue

            plain_text = []

            for _ in range(length):
                plain_text.append(next(rng))

            lzw_emit_code(plain_text[-1])

            for i in reversed(range(len(plain_text) - 1)):
                lzw_emit_code(lzw_get_from_dictionary(
                    bytes([plain_