#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('B', rng.read(1)[0] & 0b00001111 | ((width & 0x7f) << 5) | ((height & 0x7f) << 12)) + \
          struct.pack('BBB', rng.read(3))

    # 3. Global Color Table
    global_color_table_size = int((rng.read(1)[0] & 0b00001111) ** 2 / 3 * 8)

    if global_color_table_size > 0:
        global_color_table = [rng.read(3) for _ in range(global_color_table_size)]

        # Pad to nearest byte
        while len(global_color_table) % 2!= 0:
            global_color_table += [b'\x00\x00\x00']
    else:
        global_color_table = []

    # 4. Image Descriptor
    image_descriptor = struct.pack('BBHBB',
                                   rng.read(1)[0],
                                   0,
                                   width,
                                   height,
                                   0)

    # 5. Local Color Table
    local_color_table_size = int(((rng.read(1)[0] & 0b00001111) ** 2 / 3 * 8) - global_color_table_size)

    if local_color_table_size > 0:
        local_color_table = [rng.read(3) for _ in range(local_color_table_size)]

        # Pad to nearest byte
        while len(local_color_table) % 2!= 0:
            local_color_table += [b'\x00\x00\x00']
    else:
        local_color_table = []

    # 6. Graphics Control Extension
    graphics_control_extension = struct.pack('BBBBBBB',
                                             0x21,
                                             rng.read(1)[0],
                                             rng.read(1)[0],
                                             rng.read(1)[0],
                                             rng.read(1)[0],
                                             rng.read(1)[0],
                                             rng.read(1)[0])

    # 7. Image Data
    data = [rng.read(1)[0] for _ in range(width * height)]

    # 8. Trailer
    trailer = b';'

    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(struct.pack('B', global_color_table_size))
    out.write(b''.join(global_color_table))
    out.write(image_descriptor)
    out.write(struct.pack('B', local_color_table_size))
    out.write(b''.join(local_color_table))
    out.write(graphics_control_extension)
    out.write(b''.join([bytes([byte]) for byte in data]))
    out.write(trailer)
