#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 256 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % (2**8 - 256)
    height = 256 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % (2**8 - 256)
    flags = b'\x07\x00'   # Global color table flag and sort order
    bg_color_index = bytes([int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256])
    pixel_aspect_ratio = b'\x00'
    out.write(width.to_bytes(2, 'big'))
    out.write(height.to_bytes(2, 'big'))
    out.write(flags)
    out.write(bg_color_index)
    out.write(pixel_aspect_ratio)

    # Global color table
    for i in range(256):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        out.write(r.to_bytes(1, 'big'))
        out.write(g.to_bytes(1, 'big'))
        out.write(b.to_bytes(1, 'big'))

    # Image descriptor
    left = 0
    top = 0
    width = 256 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % (2**8 - 256)
    height = 256 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % (2**8 - 256)
    flags = b'\x00\x00'   # Local color table flag and interlace flag
    lct_size = b'\xff'    # Size of local color table
    out.write(left.to_bytes(2, 'big'))
    out.write(top.to_bytes(2, 'big'))
    out.write(width.to_bytes(2, 'big'))
    out.write(height.to_bytes(2, 'big'))
    out.write(flags)
    out.write(lct_size)

    # Local color table
    for i in range(256):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        out.write(r.to_bytes(1, 'big'))
        out.write(g.to_bytes(1, 'big'))
        out.write(b.to_bytes(1, 'big'))

    # LZW minimum code size
    min_code_size = b'\x08'
    out.write(min_code_size)

    # Frame data
    image_data = []
    while len(image_data) < width * height:
        prefix_length = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x0f
        if not prefix_length:
            prefix_length = 16

        prefix = [int.from_bytes(rng.read(1), byteorder='big', signed=False)]
        suffix = int.from_bytes(rng.read(1), byteorder='big', signed=False)
        image_data += prefix
        image_data += [suffix] * prefix_length

    compressed_data = compress_lzw(image_data)
    out.write(compressed_data)


# Compresses an array of integers using the LZW algorithm
def compress_lzw(input: list[int]):
    output = []
    dictionary = {i: [struct.pack('B', i)] for i in input}
    w = ''
    for c in input:
        wc = f'{w}{c}'
        if wc in dictionary:
            w = wc
        else:
            output.append(dictionary[w][0])
            if len(dictionary[w]) == 128:
                del dictionary[w]

            dictionary[wc] = dictionary[w] + [struct.pack('B', len(dictionary))]
            w = str(c)

    if w!= '':
        output.append(dictionary[w][0])

    return b''.join(output)

