#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = rng.read(1)[0] % 64 + 1
    h = rng.read(1)[0] % 64 + 1

    # Color table size is always 7
    color_table_size = 7

    # Background color index is always 0
    background_color_index = 0

    # Pixel aspect ratio is always 0
    pixel_aspect_ratio = 0

    logical_screen_descriptor = struct.pack(
        '<6sHHBBB', 
        header,
        w * 256 + h,
        color_table_size,
        background_color_index,
        pixel_aspect_ratio
    )

    # Write to output stream
    out.write(logical_screen_descriptor)

    # 3. Global Color Table
    global_color_table = [
        0x00, 0x00, 0x00, 
        0xFF, 0xFF, 0xFF, 
        0x00, 0xFF, 0x00, 
        0x00, 0x00, 0xFF, 
        0xFF, 0x00, 0x00, 
        0x00, 0xFF, 0xFF, 
        0xFF, 0xFF, 0x00
    ]

    # Write to output stream
    out.write(bytes(global_color_table))

    # 4. Image Descriptor
    image_left = 0
    image_top = 0
    image_width = w
    image_height = h

    # LZW minimum code size is always 2
    lzw_minimum_code_size = 2

    # Write to output stream
    out.write(struct.pack('<HHHHB', image_left, image_top, image_width, image_height, lzw_minimum_code_size))

    # 5. Image Data
    # Generate some data
    data = bytearray()

    while len(data) < w * h:
        # Choose a random value between 0-255
        r = rng.read(1)[0]

        if r == 0:
            # If it's zero, choose another one until we get something else than 0
            while True:
                r = rng.read(1)[0]

                if r!= 0:
                    break

            # Add the run length of zeros to our data
            data += bytes([r]) * rng.read(1)[0]
        
        elif r <= 127:
            # If it's less or equal to 127, add it as-is to our data
            data.append(r)
        
        else:
            # If it's greater than 127, add the difference from 127 to it 
            # to our data
            data.append((r - 127) + 127)
            
            # And then add that many copies of the last byte added to our data
            data += bytes([data[-1]]) * rng.read(1)[0]

    # Compress this data with LZW algorithm
    compressed_data = compress_lzw(data)

    # Write to output stream
    out.write(compressed_data)


# Compresses an array of bytes using LZW algorithm
def compress_lzw(
        data: list[int], 
        prefix: int=None
    ) -> bytes:
    result = []

    if not prefix:
        # First time compression function is called, so no prefix exists yet
        prefix = data[0]
        del data[0]

    dictionary = {prefix}

    # Start by adding the first prefix to the result
    result.append(prefix)

    while len(data):
        c = data[0]
        del data[0]

        if c in dictionary:
            # If character already exists in dictionary, just add it to the result
            result.append(c)

            # And update the prefix to be the new character
            prefix = c

        else:
            # Otherwise, find longest prefix which matches the current character
            # and add it to the result
            for i in range(len(dictionary), 0, -1):
                if dictionary[i - 1] == c:
                    result.append(i - 1)

                    # Update the prefix to be the matching part
                    prefix = dictionary[i - 1 :][0]
                    
                    # Add the rest of the matched string to the dictionary
                    dictionary |= set(dictionary[i:])

                    # Break out of loop since we found what we were looking for
                    break

            # If we didn't find anything, add the whole character to the dictionary
            # and add its length minus one to the result
            else:
                result.append(len(dictionary))
                
                dictionary.add(c)
                prefix = c

    return bytes(result)