#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>BBBBBH', *struct.unpack('BBBBBB', rng.read(6)))

    # 3. Global Color Table
    global_color_table_size = int.from_bytes(rng.read(1), 'big')
    if global_color_table_size == 0:
        global_color_table = None
    else:
        global_color_table = [
            rng.read(3) for _ in range(global_color_table_size)
        ]

        # Pad to nearest multiple of 3
        while len(global_color_table) % 3!= 0:
            global_color_table += [b'\x00\x00\x00']

    # 4. Image Blocks
    image_blocks = []
    num_image_blocks = rng.read(1)[0] & 7
    for i in range(num_image_blocks):
        # Image descriptor
        image_descriptor = struct.pack('>BHHB', *(rng.read(5)))
        
        # Local color table size
        local_color_table_size = int.from_bytes(rng.read(1), 'big')
        if local_color_table_size == 0:
            local_color_table = None
        elif local_color_table_size == 1:
            raise ValueError("Local color table must be at least 2 colors")
        else:
            local_color_table = [
                rng.read(3) for _ in range(local_color_table_size - 1)
            ]

            # Pad to nearest multiple of 3
            while len(local_color_table) % 3!= 0:
                local_color_table += [b'\x00\x00\x00']

        # Block size
        block_size = int.from_bytes(rng.read(1), 'big') + 1

        # Image data
        image_data = rng.read(block_size)

        # Add to list
        image_blocks.append((image_descriptor, local_color_table, image_data))

    # 5. Trailer
    trailer = b';'

    # Write everything to output
    out.write(header)
    out.write(width.to_bytes(2, 'big'))
    out.write(height.to_bytes(2, 'big'))
    out.write(lsd)
    if global_color_table is not None:
        out.write(len(global_color_table).to_bytes(1, 'big'))
        for entry in global_color_table:
            out.write(entry)
    else:
        out.write(b'\x00')
    for image_descriptor, local_color_table, image_data in image_blocks:
        out.write(image_descriptor)
        if local_color_table is not None:
            out.write(len(local_color_table).to_bytes(1, 'big'))
            for entry in local_color_table:
                out.write(entry)
        else:
            out.write(b'\x00')
        out.write(len(image_data).to_bytes(1, 'big'))
        out.write(image_data)
    out.write(trailer)

