#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = min((rng.read(1)[0] & 0x7F) + 1, 256)
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(global_color_table_size)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBBBB', 0, 0, 0xF7 | ((rng.read(1)[0] & 0x7) << 4), 0, 0, 0, 0)

    # 5. Local Color Table (if needed)
    local_color_table_size = max(((width * height) // 64) - 1, 0)
    if local_color_table_size > 0:
        local_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(local_color_table_size)]

        # 6. Image Data
        image_data = []
        for y in range(height):
            for x in range(width):
                image_data += [local_color_table[rng.read(1)[0]]]

            if y % 8 == 7:
                image_data += [(b'\x00' * (((width * 3) + 1)))[:((width * 3) + 1)]]

        # 7. Trailer
        trailer = b';'

    else:
        # 6. Image Data
        image_data = [global_color_table[rng.read(1)[0]] for _ in range(width * height)]

        # 7. Trailer
        trailer = b''

    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    out.write(bytes().join([image_descriptor, bytes().join(global_color_table)]))
    out.write(bytes().join(image_data))
    out.write(trailer)

