#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header (6 bytes)
    # ------------------

    # Magic Number ("GIF89a")
    header = b"GIF89a"

    # Logical Screen Width and Height (2 bytes each)
    width = rng.read(2)
    height = rng.read(2)

    # Packed Field (1 byte)
    packedField = rng.read(1)[0]

    # Global Color Table Flag (1 bit)
    globalColorTableFlag = bool(packedField & 0b10000000)

    # Color Resolution (3 bits)
    colorResolution = ((packedField >> 7) & 0b111) + 1

    # Sort Flag (1 bit)
    sortFlag = bool((packedField >> 5) & 0b1)

    # Size of Global Color Table (3 bits)
    sizeOfGlobalColorTable = (((packedField >> 2) & 0b111) + 1) * 3

    # Background Color Index (1 byte)
    backgroundColorIndex = rng.read(1)[0]

    # Pixel Aspect Ratio (1 byte)
    pixelAspectRatio = rng.read(1)[0]

    # Write to output file
    out.write(header)
    out.write(width)
    out.write(height)
    out.write(bytes([packedField]))

    if not globalColorTableFlag:
        return

    # 2. Global Color Table (3 * 2^(sizeOfGlobalColorTable+1)) bytes
    # --------------------------------------------------------------

    for i in range(pow(2, sizeOfGlobalColorTable)):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        out.write(bytes([r,g,b]))

    # 3. Image Descriptor (10 bytes)
    # -------------------------------

    # Image Separator (1 byte)
    imageSeparator = b"\x2C"

    # Image Left Position (2 bytes)
    leftPosition = rng.read(2)

    # Image Top Position (2 bytes)
    topPosition = rng.read(2)

    # Image Width (2 bytes)
    width = rng.read(2)

    # Image Height (2 bytes)
    height = rng.read(2)

    # Packed Field (1 byte)
    packedField = rng.read(1)[0]

    # Local Color Table Flag (1 bit)
    localColorTableFlag = bool(packedField & 0b10000000)

    # Interlace Flag (1 bit)
    interlaceFlag = bool((packedField >> 6) & 0b1)

    # Sort Flag (1 bit)
    sortFlag = bool((packedField >> 5) & 0b1)

    # Reserved (2 bits)
    reserved = ((packedField >> 3) & 0b11)

    # Size of Local Color Table (3 bits)
    sizeOfLocalColorTable = (((packedField >> 0) & 0b111) + 1) * 3

    # LZW Minimum Code Size (1 byte)
    lzwMinimumCodeSize = rng.read(1)[0]

    # Write to output file
    out.write(imageSeparator)
    out.write(leftPosition)
    out.write(topPosition)
    out.write(width)
    out.write(height)
    out.write(bytes([packedField]))

    if not localColorTableFlag:
        return

    # 4. Local Color Table (3 * 2^(sizeOfLocalColorTable+1)) bytes
    # ------------------------------------------------------------

    for i in range(pow(2, sizeOfLocalColorTable)):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        out.write(bytes([r,g,b]))

    # 5. Graphics Control Extension Block (8 bytes)
    # ---------------------------------------------

    # Extension Introducer (1 byte)
    extensionIntroducer = b"\x21"

    # Extension Label (1 byte)
    label = b"\xF9"

    # Block Length (1 byte)
    blockLength = b"\x04"

    # Packed Field (1 byte)
    packedField = rng.read(1)[0]

    # Transparent Color Flag (1 bit)
    transparentColorFlag = bool(packedField & 0b10000000)

    # User Input Flag (1 bit)
    userInputFlag = bool((packedField >> 7) & 0b1)

    # Disposal Method (3 bits)
    disposalMethod = ((packedField >> 5) & 0b111)

    # Delay Time (2 bytes)
    delayTime = rng.read(2)

    # Transparent Color Index (1 byte)
    transparentColorIndex = rng.read(1)[0]

    # Block Terminator (1 byte)
    blockTerminator = b"\x00"

    # Write to output file
    out.write(extensionIntroducer)
    out.write(label)
    out.write(blockLength)
    out.write(bytes([packedField]))
    out.write(delayTime)
    out.write(bytes([transparentColorIndex]))
    out.write(blockTerminator)

    # 6. Image Data
    # --------------

    # LZW Minimum Code Size (1 byte)
    lzwMinimumCodeSize = rng.read(1)[0]

    # Subblocks
    subblocks = []

   