#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    # 2. Logical Screen Descriptor
    # 3. Global Color Table (if needed)
    # 4. Image Descriptor
    # 5. Local Color Table (if needed)
    # 6. Graphics Control Extension
    # 7. Image Data

    # 1. Header
    header = b'GIF89a' + bytes([0x00] * 7)
    out.write(header)

    # 2. Logical Screen Descriptor
    width = rng.read(1)[0] % 256
    height = rng.read(1)[0] % 256
    flags = rng.read(1)[0] & 0b11111000 | 0b00000110
    bg_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0] & 0b11111000
    logical_screen_descriptor = struct.pack('<BHHBBB', width, height, flags, bg_color, aspect_ratio)
    out.write(logical_screen_descriptor)

    # 3. Global Color Table (if needed)
    global_color_table_size = rng.read(1)[0] & 0b00000111
    if global_color_table_size!= 0:
        global_color_table = [rng.read(1)[0] for _ in range(global_color_table_size ** 2)]
        out.write(bytes(global_color_table))

    # 4. Image Descriptor
    left = rng.read(1)[0] % width
    top = rng.read(1)[0] % height
    image_width = rng.read(1)[0] % (width - left)
    image_height = rng.read(1)[0] % (height - top)
    flags = rng.read(1)[0] & 0b11111000 | 0b00000000
    local_color_table_flag = bool((flags >> 7) & 1)
    interlace_flag = bool((flags >> 6) & 1)
    sort_flag = bool((flags >> 5) & 1)
    reserved = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0] & 0b00000111
    image_descriptor = struct.pack('<BBHHBBBB', left, top, image_width, image_height, flags, local_color_table_flag, interlace_flag, sort_flag, reserved, local_color_table_size)
    out.write(image_descriptor)

    # 5. Local Color Table (if needed)
    if local_color_table_flag and local_color_table_size!= 0:
        local_color_table = [rng.read(1)[0] for _ in range(local_color_table_size ** 2)]
        out.write(bytes(local_color_table))

    # 6. Graphics Control Extension
    block_size = rng.read(1)[0]
    packed_fields = rng.read(1)[0]
    delay_time = rng.read(2)
    transparent_index = rng.read(1)[0]
    terminator = rng.read(1)[0]
    graphics_control_extension = struct.pack('<BBHBB', block_size, packed_fields, delay_time[0], delay_time[1], transparent_index, terminator)
    out.write(graphics_control_extension)

    # 7. Image Data
    lzw_min_code_size = rng.read(1)[0] & 0b00001111
    data_subblock_size = rng.read(1)[0] & 0b00001111
    clear_code = 1 << lzw_min_code_size
    end_of_information_code = clear_code + 1
    code_length = 1 + lzw_min_code_size
    next_code = end_of_information_code + 1
    table = []
    index = 0
    while True:
        if len(table) == (1 << code_length):
            code_length += 1

        if index >= 256:
            break

        byte = rng.read(1)[0]
        if byte == 0x00 or byte == 0xff:
            continue

        table.append(byte)

    def get_next_code():
        nonlocal next_code
        result = next_code
        next_code += 1
        return result

    def write_lzw_data(data):
        nonlocal code_length
        nonlocal next_code
        nonlocal table

        for i in range(len(data)):
            datum = data[i]

            if datum not in table:
                raise Exception('Invalid LZW data')

            yield from write_lzw_datum(datum)

            if next_code == (1 << code_length):
                code_length += 1

    def write_lzw_datum(datum):
        nonlocal code_length
        nonlocal next_code
        nonlocal table

        if datum < clear_code:
            yield datum
        elif datum == clear_code:
            yield datum
            code_length = 1 + lzw_min_code_size