#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    # ------------------

    # Magic Number (GIF89a)
    out.write(b"GIF89a")

    # Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)

    # Packed field: Global Color Table Flag (1 bit)
    #   - 0 means no global color table
    #   - 1 means there is one
    gct_flag = b"\x01" if rng.read(1)[0] & 1 else b"\x00"

    # Packed field: Color Resolution (3 bits)
    #   - 7 + n means 2^(n+7) colors in palette
    cr = rng.read(1)[0] >> 5
    cg = rng.read(1)[0] >> 5
    cb = rng.read(1)[0] >> 5

    # Packed field: Sorting Method (3 bits)
    #   000 means not sorted at all
    #   001 means sort by G
    #   010 means sort by GA
    #   011 means sort by RGB
    #   100 means sort by RGBA
    #   101 and above are reserved for future use
    sm = rng.read(1)[0] >> 5

    # Packed field: Size of Global Color Table (3 bits)
    #   000 means 2^n entries
    #   001 means 2^(n+1) entries
    #   010 means 2^(n+2) entries
    #   011 means 2^(n+3) entries
    #   100 means 2^(n+4) entries
    #   101 means 2^(n+5) entries
    #   110 means 2^(n+6) entries
    #   111 means 2^(n+7) entries
    sgt = rng.read(1)[0] >> 5

    # Background Color Index (1 byte)
    bgc = rng.read(1)

    # Pixel Aspect Ratio (1 byte)
    par = rng.read(1)

    # Write to output
    out.write(width)
    out.write(height)
    out.write(gct_flag)
    out.write((cr << 5) | (cg << 5) | (cb << 5))
    out.write((sm << 5) | (sgt << 5))
    out.write(bgc)
    out.write(par)


    # 2. Logical screen descriptor
    # -----------------------------

    # Width (2 bytes)
    out.write(width)

    # Height (2 bytes)
    out.write(height)

    # Packed field: Global Color Table Flag (1 bit)
    #   - 0 means no global color table
    #   - 1 means there is one
    out.write(gct_flag)

    # Packed field: Color Resolution (3 bits)
    #   - 7 + n means 2^(n+7) colors in palette
    out.write((cr << 5) | (cg << 5) | (cb << 5))

    # Packed field: Sorting Method (3 bits)
    #   000 means not sorted at all
    #   001 means sort by G
    #   010 means sort by GA
    #   011 means sort by RGB
    #   100 means sort by RGBA
    #   101 and above are reserved for future use
    out.write((sm << 5) | (sgt << 5))

    # Size of Global Color Table (3 bits)
    #   000 means 2^n entries
    #   001 means 2^(n+1) entries
    #   010 means 2^(n+2) entries
    #   011 means 2^(n+3) entries
    #   100 means 2^(n+4) entries
    #   101 means 2^(n+5) entries
    #   110 means 2^(n+6) entries
    #   111 means 2^(n+7) entries
    out.write(sgt)

    # Background Color Index (1 byte)
    out.write(bgc)

    # Pixel Aspect Ratio (1 byte)
    out.write(par)


    # 3. Global color table
    # ---------------------

    # If we have a global color table...
    if gct_flag == b"\x01":

        # How many colors?
        num_colors = pow(2, sgt + 7)

        # For each color...
        for i in range(num_colors):

            # Red
            r = rng.read(1)

            # Green
            g = rng.read(1)

            # Blue
            b = rng.read(1)

            # Write it to our output
            out.write(r)
            out.write(g)
            out.write(b)


    # 4. Image data
    # -------------

    # LZW minimum code size (1 byte)
    lzw_min_code_size = rng.read(1)

    # Write this value to the output
    out.write(lzw_min_code_size)

    # Get the next byte from /dev/urandom
    byte = rng.read(1)[0]

    # While the byte isn't zero...
    while byte!= 0:

        # Write it to the output
        out.write(bytes([byte