#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Set color resolution to 7 bits per pixel
    color_resolution = 0b00000011
    global_color_table_flag = 0
    sort_flag = 0
    size_of_global_color_table = 0
    background_color_index = 0
    pixel_aspect_ratio = 0

    logical_screen_descriptor = \
        struct.pack('<HHBBBH',
                    width,
                    height,
                    color_resolution,
                    global_color_table_flag,
                    sort_flag,
                    size_of_global_color_table
                   ) + \
                           bytearray([background_color_index]) + \
                           bytearray([pixel_aspect_ratio])

    # 3. Global Color Table
    if not global_color_table_flag:
        raise Exception("Global color table flag must be set")

    # Generate palette of 256 colors
    palette = [struct.pack("<BBB", *rng.read(3)) for _ in range(256)]

    # Write palette to output stream
    out.write(bytes(palette))

    # 4. Image Descriptor
    image_separator = b'\x2C'
    left = 0
    top = 0
    width = width
    height = height
    local_color_table_flag = 0
    interlace_flag = 0
    sort_flag = 0
    size_of_local_color_table = 0
    local_color_table_index = 0

    image_descriptor = \
            struct.pack('<BBHHB',
                        left,
                        top,
                        width,
                        height,
                        local_color_table_flag
                       ) + \
                       bytearray([interlace_flag]) + \
                       bytearray([sort_flag]) + \
                       bytearray([size_of_local_color_table]) + \
                       bytearray([local_color_table_index])

    # 5. LZW Minimum Code Size
    lzw_minimum_code_size = 2 ** rng.read(1)[0] - 1

    # 6. Data Sub-blocks
    data_sub_block_size = 2 ** lzw_minimum_code_size

    # Generate random data sub-blocks
    data_sub_blocks = []
    while True:
        block = rng.read(data_sub_block_size)

        # If we read less than expected, this is the last one
        if len(block) < data_sub_block_size:
            break

        data_sub_blocks += [block]

    # 7. Trailer
    trailer = b';'

    # Write everything to output stream
    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(image_separator)
    out.write(image_descriptor)
    out.write(bytearray([lzw_minimum_code_size]))
    for block in data_sub_blocks:
        out.write(bytearray([len(block)]))
        out.write(block)
    out.write(trailer)

