#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    width = int.from_bytes(rng.read(2), 'big') % (50 - 4 + 1) + 4
    height = int.from_bytes(rng.read(2), 'big') % (50 - 4 + 1) + 4

    flags = bytes([int.from_bytes(rng.read(1), 'big')])

    background_color_index = bytes([int.from_bytes(rng.read(1), 'big')])
    pixel_aspect_ratio = bytes([int.from_bytes(rng.read(1), 'big')])

    # 2. Logical Screen Descriptor
    logical_screen_descriptor = \
        width.to_bytes(2, byteorder='big', signed=False) + \
        height.to_bytes(2, byteorder='big', signed=False) + \
        flags + \
        background_color_index + \
        pixel_aspect_ratio

    # 3. Global Color Table
    global_color_table_size = bytes([int.from_bytes(rng.read(1), 'big')])
    global_color_table_entry_size = bytes([int.from_bytes(rng.read(1), 'big')])

    global_color_table = b''
    for _ in range(int.from_bytes(global_color_table_size, 'big')):
        color = [
            int.from_bytes(rng.read(1), 'big'),
            int.from_bytes(rng.read(1), 'big'),
            int.from_bytes(rng.read(1), 'big')
        ]

        global_color_table += bytes(color)

    # 4. Image Descriptor
    image_left = bytes([int.from_bytes(rng.read(1), 'big')])
    image_top = bytes([int.from_bytes(rng.read(1), 'big')])
    image_width = bytes([int.from_bytes(rng.read(1), 'big')])
    image_height = bytes([int.from_bytes(rng.read(1), 'big')])

    flags = bytes([int.from_bytes(rng.read(1), 'big')])

    local_color_table_flag = bytes([flags[0] & 0b1])
    interlace_flag = bytes([(flags[0] >> 2) & 0b1])
    sort_flag = bytes([(flags[0] >> 3) & 0b1])
    reserved_bits = bytes([flags[0] >> 4])

    local_color_table_size = bytes([int.from_bytes(rng.read(1), 'big')])
    local_color_table_entry_size = bytes([int.from_bytes(rng.read(1), 'big')])

    local_color_table = b''
    if int.from_bytes(local_color_table_flag, 'big'):
        for _ in range(int.from_bytes(local_color_table_size, 'big')):
            color = [
                int.from_bytes(rng.read(1), 'big'),
                int.from_bytes(rng.read(1), 'big'),
                int.from_bytes(rng.read(1), 'big')
            ]

            local_color_table += bytes(color)

    # 5. LZW Minimum Code Size
    lzw_minimum_code_size = bytes([int.from_bytes(rng.read(1), 'big')])

    # 6. Data Sub-blocks
    data_sub_block_size = bytes([int.from_bytes(rng.read(1), 'big')])
    data_sub_blocks = []
    while True:
        sub_block = rng.read(int.from_bytes(data_sub_block_size, 'big'))

        if len(sub_block) == 0: break

        data_sub_blocks.append(sub_block)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output stream
    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(global_color_table)
    out.write(image_left)
    out.write(image_top)
    out.write(image_width)
    out.write(image_height)
    out.write(flags)
    out.write(local_color_table_flag)
    out.write(interlace_flag)
    out.write(sort_flag)
    out.write(reserved_bits)
    out.write(local_color_table_size)
    out.write(local_color_table_entry_size)
    out.write(lzw_minimum_code_size)
    for block in data_sub_blocks:
        out.write(struct.pack('>H', len(block)))
        out.write(block)
    out.write(trailer)

