#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    #    - Signature (4 bytes)
    #    - Version (3 bytes)
    signature = b"GIF89a"
    version   = b"\x00\x00\x00"

    # Write header to output stream
    out.write(signature + version)

    # 2. Logical Screen Descriptor
    #    - Width (2 bytes)
    #    - Height (2 bytes)
    #    - Flags (1 byte)
    #    - Background Color Index (1 byte)
    #    - Pixel Aspect Ratio (1 byte)
    width      = rng.read(2)
    height     = rng.read(2)
    flags      = rng.read(1)[0] & 0b00000111
    bg_color   = rng.read(1)[0]
    pixel_ratio= rng.read(1)[0]

    # Write logical screen descriptor to output stream
    out.write(width + height + bytes([flags]) +
              bytes([bg_color]) + bytes([pixel_ratio]))

    # 3. Global Color Table
    #    - Size of global color table (1 byte)
    #    - Color resolution (1 byte)
    #    - Sort flag (1 byte)
    #    - Global color table size (3 bits)
    #    - Background color index (1 byte)
    #    - Pixel aspect ratio (1 byte)
    gct_size       = rng.read(1)[0] & 0b00000111
    color_res      = rng.read(1)[0]
    sort           = bool((rng.read(1)[0] >> 7) & 0b00000001)
    bg_color       = rng.read(1)[0]
    pixel_ratio    = rng.read(1)[0]

    # Write global color table to output stream
    out.write(bytes([gct_size]) + bytes([color_res]) +
              bytes([(int(sort)) << 7]) + bytes([bg_color]) +
              bytes([pixel_ratio]))

    # 4. Image Descriptor
    #    - Left position (2 bytes)
    #    - Top position (2 bytes)
    #    - Width (2 bytes)
    #    - Height (2 bytes)
    #    - Flags (1 byte)
    #    - Interlace Flag (1 bit)
    #    - Sort Flag (1 bit)
    #    - Local color table flag (1 bit)
    #    - Reserved (3 bits)
    #    - LZW minimum code size (1 byte)
    left        = rng.read(2)
    top         = rng.read(2)
    width       = rng.read(2)
    height      = rng.read(2)
    flags       = rng.read(1)[0] & 0b00000111
    interlace   = bool((rng.read(1)[0] >> 6) & 0b00000001)
    sorted      = bool((rng.read(1)[0] >> 5) & 0b00000001)
    lct_flag    = bool((rng.read(1)[0] >> 7) & 0b00000001)
    reserved    = (rng.read(1)[0] >> 3) & 0b00000111
    min_code_sz = rng.read(1)[0]

    # Write image descriptor to output stream
    out.write(left + top + width + height + bytes([flags]) +
              bytes([(int(interlace)) << 6]) + bytes([(int(sorted)) << 5]) +
              bytes([(int(lct_flag)) << 7]) + bytes([reserved << 3]) +
              bytes([min_code_sz]))

    # 5. Local Color Table
    if lct_flag == True:
        for _ in range(pow(2, gct_size)):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]

            out.write(bytes([r]) + bytes([g]) + bytes([b]))

    # 6. Block
    #    - Data sub-blocks
    #    - Trailer
    block_type = rng.read(1)[0]
    while block_type!= 0x3B:
        # Sub-block
        #    - Size (1 byte)
        #    - Data (Size bytes)
        data_size = rng.read(1)[0]

        # Write data sub-block to output stream
        out.write(bytes([data_size]))
        for _ in range(data_size):
            out.write(rng.read(1))

        # Next block type
        block_type = rng.read(1)[0]

    # Write trailer to output stream
    out.write(bytes([0x3B]))


if