#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = b''
    for i in range(num_colors):
        r, g, b = [rng.read(1)[0] for _ in range(3)]

        # Each color is represented by three bytes: Red, Green, Blue
        global_color_table += struct.pack('BBB', r, g, b)

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBBBB', 0, 0, 0xF7 | ((rng.read(1)[0] & 0x7) << 4), 0, 0, 0, 0)

    # 5. Image Data
    # Let's just use one frame with no delay and full-screen
    image_data = b'\x2C\x00\x00\x00\x00\x00\x00\x00\xFF\xFF\xFF\x00\x00\x00!\xF9\x04\x01\x00\x00\x00\x00,\x00\x00\x00\x00\x02\x00\x01\x00\x80\x01\x00\x00\x02\x02D\x01\x00;'

    # 6. Trailer
    trailer = b';'

    # Write everything to the output stream
    out.write(header + lsd + global_color_table + image_descriptor + image_data + trailer)

