#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = min((rng.read(1)[0] & 0x7F) + 1, 256)
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(global_color_table_size)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBBBB', 0, 0, 0, 0b10000000 | rng.read(1)[0], 0, 0)

    # 5. Local Color Table (if needed)
    local_color_table_size = max(min((rng.read(1)[0] & 0x7F) + 1, 256), 1)
    if local_color_table_size > 1:
        local_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(local_color_table_size)]
    else:
        local_color_table = []

    # 6. Graphics Control Extension
    graphics_control_extension = struct.pack('>BBBBBBH', 0x21, 0xF9, 4, 0b10000000 | rng.read(1)[0], 0, 0, 0)

    # 7. Application Extension
    application_extension = struct.pack('>BBBBB', 0x21, 0xFF, 11, ord('A'), ord('N') << 8 | ord('I')) \
                            + b'NETSCAPE2.0\x03\x01\x00\x00'

    # 8. Comment Extension
    comment_extension = struct.pack('>BBBBB', 0x21, 0xFE, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 9. Plain Text Extension
    plain_text_extension = struct.pack('>BBBBB', 0x21, 0x01, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 10. Trailer
    trailer = b';'

    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(bytes([len(global_color_table)]))
    out.write(b''.join(global_color_table))
    out.write(image_descriptor)
    out.write(bytes([len(local_color_table)]))
    out.write(b''.join(local_color_table))
    out.write(graphics_control_extension)
    out.write(application_extension)
    out.write(comment_extension)
    out.write(plain_text_extension)
    out.write(trailer)

