#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0] & 1)  # Randomly decide whether to use GCE
    if not has_gce:
        gce = None
    else:
        # Duration of animation in centiseconds (1/100th second)
        duration = rng.read(2)
        
        # Transparent color index (or "no transparency")
        trans_index = rng.read(1)[0]
        
        # User input flag (whether user can interact with the animation)
        user_input = bool(rng.read(1)[0] & 1)
        
        # Has transparancy flag
        has_transparancy = bool(rng.read(1)[0] & 1)
        
        # Reserved flags (must be set to zero)
        reserved = rng.read(1)
        
        # The actual graphics control extension block itself
        gce = struct.pack('>BBHBBB', ord('!') + int(has_transparancy), ord('G'), 4, duration[0], duration[1], trans_index)
        
    # 5. Image Descriptor
    # Left position of image on screen (using two bytes for each)
    left = struct.unpack('>HH', rng.read(4))[0]
    
    # Top position of image on screen (using two bytes for each)
    top = struct.unpack('>HH', rng.read(4))[0]
    
    # Width of image (using two bytes for each)
    img_width = struct.unpack('>HH', rng.read(4))[0]
    
    # Height of image (using two bytes for each)
    img_height = struct.unpack('>HH', rng.read(4))[0]
    
    # Flags indicating which parts of the image are local vs global color table indices
    packed_fields = rng.read(1)[0]
    
    # Actual image descriptor
    id = struct.pack('>HHHHB', left, top, img_width, img_height, packed_fields)
    
    # 6. Local Color Table (if applicable)
    # If we're using a local color table, let's randomly choose its size
    if packed_fields & 0b10000000 > 0:
        num_colors = (rng.read(1)[0] % 256) or 1
        color_table += rng.read(3 * num_colors)  # Add more colors to our color table

    # 7. Image Data
    # Let's just write some random data for now
    img_data = rng.read((img_width // 8) * img_height)
    
    # 8. Trailer
    trailer = b';'
    
    # Write everything to output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(id)
    out.write(img_data)
    out.write(trailer)

