#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_gif(out: BinaryIO, rng):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + rng.randint(0x7f, 0xff)
    height = 1 + rng.randint(0x7f, 0xff)
    flags = (width & 0x80) | ((height & 0x80) >> 7)
    global_color_table_size = 2 ** (4 - (flags & 7))
    background_color_index = rng.randint(0, global_color_table_size-1)
    pixel_aspect_ratio = 0
    out.write(struct.pack('<HHBBB',
        width, height,
        global_color_table_size,
        background_color_index,
        pixel_aspect_ratio))

    # Global color table
    for i in range(global_color_table_size):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left = 0
    top = 0
    width = 1 + rng.randint(0x7f, 0xff)
    height = 1 + rng.randint(0x7f, 0xff)
    local_color_table_flag = False
    interlace_flag = False
    sort_flag = False
    reserved = 0
    lct_size = 2 ** (4 - (rng.randint(0, 3)))
    out.write(struct.pack('<BBHHB',
        0x2c,
        left, top,
        width, height,
        (local_color_table_flag << 7) |
            (interlace_flag << 6) |
            (sort_flag << 5) |
            (reserved << 3) |
             lct_size))

    # Local color table
    if lct_size > 0:
        for i in range(lct_size):
            r = rng.randint(0, 255)
            g = rng.randint(0, 255)
            b = rng.randint(0, 255)
            out.write(struct.pack('BBB', r, g, b))

    # Block size
    block_size = 0x80
    out.write(block_size.to_bytes(1, 'big'))

    # LZW minimum code size
    min_code_size = 2
    out.write(min_code_size.to_bytes(1, 'big'))

    # Pixel data
    def get_next_byte():
        return rng.randint(0, 255)

    while True:
        byte = get_next_byte()

        if byte == 0:
            break

        elif byte <= min_code_size:
            out.write(byte.to_bytes(1, 'big'))

        else:
            nbits = min_code_size + 1

            while True:
                if byte < (1 << nbits):
                    break

                byte -= 1 << nbits
                nbits += 1

            out.write((nbits - min_code_size).to_bytes(1, 'big'))
            out.write(byte.to_bytes(1, 'big'))


    # Trailer
    out.write(b'\x3b')


