#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(1), 'big') % (2**10 + 1)
    height = int.from_bytes(rng.read(1), 'big') % (2**10 + 1)
    flags = bytes([int.from_bytes(rng.read(1), 'big')])
    bgColorIndex = int.from_bytes(rng.read(1), 'big') % 256
    pixelAspectRatio = bytes([int.from_bytes(rng.read(1), 'big')])

    lsd = bytearray()
    lsd += width.to_bytes(2, 'little', signed=False)
    lsd += height.to_bytes(2, 'little', signed=False)
    lsd += flags
    lsd += bgColorIndex.to_bytes(1, 'little', signed=False)
    lsd += pixelAspectRatio

    # 3. Global Color Table
    globalColorTableSize = int.from_bytes(rng.read(1), 'big') & 7
    globalColorTableFlag = bool(globalColorTableSize > 0)
    if globalColorTableFlag:
        globalColorTable = bytearray()

        for i in range(1 << globalColorTableSize):
            r = int.from_bytes(rng.read(1), 'big') % 256
            g = int.from_bytes(rng.read(1), 'big') % 256
            b = int.from_bytes(rng.read(1), 'big') % 256

            globalColorTable += r.to_bytes(1, 'little', signed=False)
            globalColorTable += g.to_bytes(1, 'little', signed=False)
            globalColorTable += b.to_bytes(1, 'little', signed=False)

    else:
        globalColorTable = None

    # 4. Image Descriptor
    imageLeftPosition = int.from_bytes(rng.read(2), 'little', signed=True)
    imageTopPosition = int.from_bytes(rng.read(2), 'little', signed=True)
    imageWidth = int.from_bytes(rng.read(2), 'little', signed=False)
    imageHeight = int.from_bytes(rng.read(2), 'little', signed=False)
    flags = bytes([int.from_bytes(rng.read(1), 'big')])
    localColorTableSize = int.from_bytes(rng.read(1), 'big') & 7
    interlaceFlag = bool((flags[0] >> 6) & 1)
    sortFlag = bool((flags[0] >> 5) & 1)
    localColorTableFlag = bool(localColorTableSize > 0)

    id = bytearray()
    id += imageLeftPosition.to_bytes(2, 'little', signed=True)
    id += imageTopPosition.to_bytes(2, 'little', signed=True)
    id += imageWidth.to_bytes(2, 'little', signed=False)
    id += imageHeight.to_bytes(2, 'little', signed=False)
    id += flags
    id += localColorTableSize.to_bytes(1, 'little', signed=False)

    # 5. Local Color Table
    if localColorTableFlag and not globalColorTableFlag:
        localColorTable = bytearray()

        for i in range(1 << localColorTableSize):
            r = int.from_bytes(rng.read(1), 'big') % 256
            g = int.from_bytes(rng.read(1), 'big') % 256
            b = int.from_bytes(rng.read(1), 'big') % 256

            localColorTable += r.to_bytes(1, 'little', signed=False)
            localColorTable += g.to_bytes(1, 'little', signed=False)
            localColorTable += b.to_bytes(1, 'little', signed=False)

    elif localColorTableFlag and globalColorTableFlag:
        localColorTable = bytearray()

        for i in range(1 << localColorTableSize):
            index = int.from_bytes(rng.read(1), 'big') % len(globalColorTable) // 3
            localColorTable += globalColorTable[index*3 : index*3+3]

    else:
        localColorTable = None

    # 6. Data
    data = bytearray()

    while True:
        blockType = rng.read(1)[0]

        if blockType == 0x2C:
            # Image Block
            pass

        elif blockType == 0x21:
            # Extension Block
            extensionBlockSize = int.from_bytes(rng.read(1), 'big')

            if extensionBlockSize!= 0xFF:
                raise Exception('Invalid extension block size.')

            subBlockType = rng.read(1)[0]

            if subBlockType == 0xF9:
                # Graphics Control Extension
                graphicsControlExtension = bytearray()

                graphicsControlExtension += bytes([subBlockType])
                graphicsControlExtension += bytes([extensionBlockSize])
                graphicsControlExtension += bytes([int.from_bytes(rng.read(1), 'big')])
                graphicsControlExtension += bytes([int.from_bytes(rng.read(1), 'big')])
                graphicsControlExtension += bytes([int.from_bytes(rng.read(1), 'big')])

                data