#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = min((rng.read(1)[0] & 0x7F) + 1, 256)
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(global_color_table_size)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBHB', 0, 0, 0b10000000 | rng.read(1)[0], 0, 0)
    
    # 5. Local Color Table (if any)
    local_color_table_size = min((image_descriptor[3] & 0x7F) + 1, 256)
    if local_color_table_size > 0:
        local_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(local_color_table_size)]
    else:
        local_color_table = []
    
    # 6. Graphics Control Extension (if needed)
    graphics_control_extension = b''
    if rng.read(1)[0] == 1:
        graphics_control_extension = struct.pack('>BBHBBBBB', 0x21, 0xF9, 4, rng.read(1)[0], 0, 0, 0)
    
    # 7. Image Data
    image_data = b''
    while True:
        block_type = rng.read(1)[0]
        
        if block_type == 0x2C or block_type == 0x21:
            # Block size
            block_size = struct.unpack('>L', rng.read(4))[0]
            
            # Sub-blocks of data
            subblock_data = b''
            while len(subblock_data) < block_size:
                subblock_size = rng.read(1)[0]
                
                if subblock_size!= 0:
                    subblock_data += rng.read(subblock_size)
                elif subblock_size == 0:
                    break

            image_data += struct.pack('>B', block_type) + struct.pack('>L', len(subblock_data)) + subblock_data
        elif block_type == 0x3B:
            image_data += struct.pack('>B', block_type)
            break
        else:
            raise Exception("Invalid block type")
        
    # 8. Trailer
    trailer = b';'
    
    # Write everything to output
    out.write(header + lsd + b''.join(global_color_table) + image_descriptor + b''.join(local_color_table) + graphics_control_extension + image_data + trailer)