#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 256 + (rng.read(1)[0] % 128)
    height = 256 + (rng.read(1)[0] % 128)
    flags = b'\x87\x00' if rng.read(1)[0] & 1 else b'\x07\x00'
    bgColorIndex = rng.read(1)[0] % 256
    pixelAspectRatio = b'\x00'
    out.write(struct.pack('<HHBBB',
        width, height,
        flags[0], flags[1],
        bgColorIndex))
    out.write(pixelAspectRatio)

    # Global color table
    globalColorTableSize = min((width * height), 256) // 4
    for i in range(globalColorTableSize):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        out.write(struct.pack('BBB', r, g, b))

    # Loop control block
    loopCount = rng.read(1)[0] % 256
    out.write(struct.pack('BB',
        0x21, 0xFF))
    out.write(struct.pack('B',
        0x0B))
    out.write(struct.pack('BBBBBBB',
        0x00, 0x00, 0x00,
        0x00, 0x00, 0x00))
    out.write(struct.pack('B',
        loopCount))

    # Image descriptor
    out.write(struct.pack('BB',
        0x2C, 0x00))
    left = rng.read(1)[0] % width
    top = rng.read(1)[0] % height
    flags = b'\x87\x00' if rng.read(1)[0] & 1 else b'\x07\x00'
    localColorTableSize = 0
    interlaceFlag = False
    sortFlag = True
    reserved = b'\x00\x00'
    lct = None
    if rng.read(1)[0] & 1:
        localColorTableSize = min(((width * height) // 4), 256) // 4
        lct = bytearray()
        for i in range(localColorTableSize):
            r = rng.read(1)[0] % 256
            g = rng.read(1)[0] % 256
            b = rng.read(1)[0] % 256
            lct += bytes([r, g, b])
    out.write(struct.pack('<HHBBB',
        left, top,
        width, height,
        flags[0], flags[1]))
    if lct is not None:
        out.write(lct)
    out.write(reserved)
    if interlaceFlag:
        out.write(b'\x40')
    elif sortFlag:
        out.write(b'\x20')
    out.write(b'\x00')

    # Local color table
    for i in range(localColorTableSize):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        out.write(struct.pack('BBB', r, g, b))

    # Image data
    imageData = bytearray()
    for y in range(height):
        for x in range(width):
            imageData += struct.pack('B',
                rng.read(1)[0] % 256)
    out.write(imageData)

