#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(num_colors)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBBBB', 0, 0, 0b10000000 | rng.read(1)[0], 0, 0, 0)

    # 5. Local Color Table (same as above)
    local_color_table = []
    if rng.read(1)[0]:
        num_colors = (rng.read(1)[0] % 256) or 1
        local_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(num_colors)]

    # 6. Graphics Control Extension
    graphics_control_extension = struct.pack('>BBBBBBH', 0x21, 0xF9, 4, 0b10000000 | rng.read(1)[0], 0, 0, rng.read(2)[0])

    # 7. Application Extension
    application_extension = struct.pack('>BBBBB', 0xFF, 0x0B, ord('A'), ord('N'), ord('I')) + b'\x0D\x0A\x87\x0A'

    # 8. Comment Extension
    comment_extension = struct.pack('>BBBBB', 0xFF, 0xFE, ord('C'), ord('M'), ord('E')) + b'\x0D\x0A\x87\x0A'

    # 9. Plain Text Extension
    plain_text_extension = struct.pack('>BBBBB', 0xFF, 0x01, ord('P'), ord('T'), ord('X')) + \
                           struct.pack('>BBBBB', ord('W'), ord('O'), ord('R'), ord('D'), 0) + \
                           struct.pack('>BBBBB', ord('S'), ord('H'), ord('A'), ord('P'), 0) + \
                           struct.pack('>BBBBB', ord('F'), ord('O'), ord('N'), ord('T'), 0) + \
                           struct.pack('>BBBBB', ord('S'), ord('I'), ord('Z'), ord('E'), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord('1'), ord('6'), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord(' '), 0) + \
                           struct.pack('>BBBBB', ord(' '), ord(' '), ord(' '), ord