#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    out.write(color_table)

    # 4. Graphics Control Extension (optional)
    has_graphics_control_extension = bool((rng.read(1)[0] & 1) << 7)
    if has_graphics_control_extension:
        graphics_control_extension_label = rng.read(1)[0]
        delay_time = rng.read(1)[0]
        transparent_color_index = rng.read(1)[0]

        out.write(struct.pack('>BBBHB', 0x21, 0xF9, 0x04, graphics_control_extension_label, delay_time))
        out.write(transparent_color_index.to_bytes(1, 'big'))

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool((rng.read(1)[0] & 1) << 7)
    interlace_flag = bool((rng.read(1)[0] & 1) << 6)
    sort_flag = bool((rng.read(1)[0] & 1) << 5)
    reserved_bits = rng.read(1)[0]
    lzw_minimum_code_size = rng.read(1)[0]
    out.write(struct.pack('>BHHHHBB', 0x2C, left, top, width, height, local_color_table_flag | interlace_flag | sort_flag | reserved_bits, lzw_minimum_code_size))

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0] + 1)
        out.write(struct.pack('B', block_size))
        out.write(rng.read(block_size))

        total_blocks_size -= block_size

    # 7. Trailer
    out.write(b'\x3b')