#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF will have a random width and height between [min_width, max_width] x [min_height, max_height].
# If `max_width == -1`, then the maximum possible value is used instead. Same goes for `max_height`.
# The background color index is always set to black (`0`).
# The pixel aspect ratio is always set to square pixels (`0`).
# The global color table has a random size between `[1, 256]` and is completely randomized.
# The minimum code size is randomly chosen from `[1, 8]`.
# The image data is filled with random data, which is made of multiple blocks that are also randomly sized.
# Blocks can't exceed 255 bytes in size.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, min_width=1, max_width=-1, min_height=1, max_height=-1
    ):
    if not isinstance(min_width, int) or min_width < 1: raise ValueError("Invalid minimum width")
    if not isinstance(max_width, int) or max_width!= -1 and max_width <= min_width: raise ValueError("Invalid maximum width")
    if not isinstance(min_height, int) or min_height < 1: raise ValueError("Invalid minimum height")
    if not isinstance(max_height, int) or max_height!= -1 and max_height <= min_height: raise ValueError("Invalid maximum height")

    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width = rng.randint(min_width, max_width if max_width > 0 else sys.maxsize)
    height = rng.randint(min_height, max_height if max_height > 0 else sys.maxsize)
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)

