#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = min((rng.read(1)[0] % 256) or 1, 256)
    color_table = [struct.pack('>BBB', *(rng.read(3))) for _ in range(num_colors)]  # Random colors

    # 4. Graphics Control Extension (optional)
    if bool(packed_fields & 0b10000000):
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0] & 0b10000000)
    interlace_flag = bool(rng.read(1)[0] & 0b01000000)
    sort_flag = bool(rng.read(1)[0] & 0b00100000)
    reserved = rng.read(1)[0]
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    image_descriptor = struct.pack('>BHHHHBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  lzw_minimum_code_size)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + b''.join(color_table) + gce + image_descriptor + image_data + trailer)