#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    # Write the LSD to the output file
    out.write(struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # Write the GCT to the output file
    out.write(color_table)

    # 4. Graphics Control Extension (optional)
    has_gce = bool((rng.read(1)[0] % 2))  # Has graphics control extension?
    if has_gce:
        # Other GCE fields: Disposal method, User input flag, Delay time in units of 10ms
        disposal_method = rng.read(1)[0]
        user_input_flag = rng.read(1)[0]
        delay_time = rng.read(2)[0]

        # Write the GCE to the output file
        out.write(struct.pack('>BBBHB', 0x21, 0xF9, 0x04, disposal_method, user_input_flag))
        out.write(delay_time.to_bytes(2, 'little'))

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool((rng.read(1)[0] % 2))  # Local color table present?
    interlace_flag = bool((rng.read(1)[0] % 2))  # Interlaced?
    sort_flag = bool((rng.read(1)[0] % 2))  # Sorted?
    reserved = rng.read(1)[0]  # Reserved field

    # Write the ID to the output file
    out.write(struct.pack('>BHHHHBB', 0x2C, left, top, width, height, local_color_table_flag, interlace_flag, sort_flag, reserved))

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # Write the ID to the output file
    out.write(image_data)

    # 7. Trailer
    out.write(b'\x3B')