#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF will have a random width and height between 1px and 1000px,
# a random background color from the global color table,
# a random amount of frames between 1 and 10,
# and a random frame duration between 1ms and 1s.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    while not 1 <= width <= 1000 or not 1 <= height <= 1000:
        width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0] & 0x0F
    pixel_aspect_ratio = rng.read(1)[0] & 0x0F
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])  # If we want one, then add an extension
    if has_gce:
        delay_time = rng.read(2)
        transparent_color_index = rng.read(1)[0] & 0x0F
        gce = struct.pack('>BBBHBB', 0x21, 0xF9, 0x04, delay_time[0] | delay_time[1] << 8, transparent_color_index, 0)
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = rng.read(1)[0] & 0x01
    interlace_flag = rng.read(1)[0] & 0x02
    sort_flag = rng.read(1)[0] & 0x04
    reserved = rng.read(1)[0] & 0xFC
    image_descriptor = struct.pack('>BHHHHBBBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(lsd + color_table + gce + image_descriptor + image_data + trailer)