#!/usr/bin/env python3

import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = min((rng.read(1)[0] % 256) or 1, 256)
    color_table = [rng.read(3) for _ in range(num_colors)]  # Random colors
    out.write(bytes().join(color_table))

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0] % 2)
    if not has_gce:
        return

    out.write(b'\x21\xf9\x04')
    delay_time = rng.read(1)[0]
    transparent_color_index = rng.read(1)[0]
    terminator = rng.read(1)[0]
    out.write(struct.pack('>BBH', delay_time, transparent_color_index, terminator))

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0] % 2)
    interlace_flag = bool(rng.read(1)[0] % 2)
    sort_flag = bool(rng.read(1)[0] % 2)
    reserved = rng.read(1)[0]
    lct_size = rng.read(1)[0]
    local_color_table = [rng.read(3) for _ in range(lct_size)]
    out.write(struct.pack('>BHHHHBBB', 0x2c, left, top, width, height, local_color_table_flag, interlace_flag,
                          sort_flag, reserved))
    out.write(bytes().join(local_color_table))

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    data = []
    while True:
        block_size = rng.read(1)[0] & 0xFF

        if block_size == 0:
            break

        data.append(rng.read(block_size))

    # 7. Trailer
    out.write(b';')