#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generate a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF will have a random width and height between [min_width, max_width]
# and [min_height, max_height]. It also has a random number of global color table entries
# between [min_num_colors, max_num_colors].
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, min_width=16, max_width=1024, min_height=16, max_height=1024, 
        min_num_colors=1, max_num_colors=256
    ):

    def randint(min_, max_): return int.from_bytes(rng.read(1), 'big') % (max_ - min_) + min_

    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width = randint(min_width, max_width)
    height = randint(min_height, max_height)
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = randint(min_num_colors, max_num_colors)
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''
    
    # 5. Image Descriptor
    left = randint(0, width)
    top = randint(0, height)
    width = randint(1, width - left)
    height = randint(1, height - top)
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)