#!/usr/bin/env python3

import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    """Generates a random GIF file into `out` using the
    random number generator `rng`."""

    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = [rng.read(3) for _ in range(num_colors)]  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0] & 1)
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

        # TODO: Handle disposal method properly
        assert not (gce[3] >> 2) & 1  # Disposal method must be 0 or 1
        assert not (gce[3] >> 1) & 1  # User input flag must be set to false
        assert not gce[3] & 1        # Transparent color flag must be set to false
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0] & 1)
    interlace_flag = bool(rng.read(1)[0] & 1)
    sort_flag = bool(rng.read(1)[0] & 1)
    reserved_bits = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = [rng.read(3) for _ in range(local_color_table_size)]
    image_descriptor = struct.pack('>BHHHHBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved_bits,
                                  local_color_table_size)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0] + 1)
        blocks.append(rng.read(block_size))

        total_blocks_size -= block_size

    # 7. Trailer
    out.write(b';')

    # Write everything to the output file
    out.write(lsd)
    for color in color_table:
        out.write(color)
    if has_gce:
        out.write(gce)
    out.write(image_descriptor)
    out.write(struct.pack('B', lzw_minimum_code_size))
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)