#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    def write_byte(b: int) -> None:
        out.write(struct.pack('B', b))
        
    def write_word(w: int) -> None:
        out.write(struct.pack('<H', w))
    
    def write_dword(dw: int) -> None:
        out.write(struct.pack('<I', dw))
    
    def write_string(s: str) -> None:
        out.write(s.encode())
    
    # 1. Header
    write_string('GIF89a')
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    write_word(width)
    write_word(height)
    write_byte(packed_fields)
    write_byte(background_color_index)
    write_byte(pixel_aspect_ratio)
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    write_byte(num_colors << 1 | 1)  # Size is stored as a word, so we need to shift by 1 bit first
    for i in range(num_colors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        write_byte(r)
        write_byte(g)
        write_byte(b)
    
    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        write_byte(0x21)
        write_byte(0xF9)
        write_byte(0x04)
        
        delay_time = rng.read(1)[0]
        transparent_color_index = rng.read(1)[0]
        disposal_method = rng.read(1)[0]
        user_input_flag = bool(rng.read(1)[0])
        write_byte(delay_time)
        write_byte(transparent_color_index)
        write_byte(disposal_method)
        write_byte(user_input_flag << 7 | 1)  # User input flag set to true
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved = rng.read(1)[0]
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    write_byte(0x2C)
    write_word(left)
    write_word(top)
    write_word(width)
    write_word(height)
    write_byte(local_color_table_flag << 7 |
               interlace_flag << 6 |
               sort_flag << 5 |
               1 << 4 |
               lzw_minimum_code_size)
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        write_byte(block_size)
        write_string(os.urandom(block_size).decode())
    write_byte(0)  # Ending with a block size of 0
    
    # 7. Trailer
    write_byte(0x3B)