#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    lsd = struct.pack('>HHBBB', width, height, 0xf0 | ((width >> 8) & 0x0f), 0xf0 | (height >> 8), 0)

    # 3. Color Table
    color_count = rng.read(1)[0] % 65 + 1
    color_table = bytearray()
    for i in range(0, color_count):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        color_table += struct.pack('<BBB', r, g, b)

    # 4. Graphics Control Extension
    gce = None
    if rng.read(1)[0] > 10:
        gce = b''
        gce += struct.pack('BB', 0x21, 0xf9)
        gce += struct.pack('BBBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], 0)
        gce += struct.pack('B', 0)

    # 5. Image Descriptor
    image_descriptor = struct.pack('>HHB', 0, 0, 0)

    # 6. Blocks
    total_blocks_size = rng.read(1)[0] * 256 + rng.read(1)[0]
    blocks = []
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0])
        blocks.append(rng.read(block_size))
        
        total_blocks_size -= block_size
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)