#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Magic header
    out.write(b'GIF89a')

    # Screen descriptor
    width = 2 ** (1 + rng.read(1)[0] % 7)
    height = 2 ** (1 + rng.read(1)[0] % 7)
    flags = rng.read(1)[0] & 0xE0 | 0x1F
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('<HHBB', width, height, flags, background))

    # Global color table
    global_color_table_size = 2 ** ((flags >> 4) + 1)
    for i in range(global_color_table_size):
        out.write(rng.read(3))

    # Image descriptor
    left = 0
    top = 0
    width = width - 1
    height = height - 1
    local_color_table_flag = 0
    interlace_flag = 0
    sort_flag = 0
    reserved = 0
    size_of_local_color_table = 0
    out.write(struct.pack('<HHHBBB', left, top, width, height,
                          local_color_table_flag, interlace_flag, sort_flag,
                          reserved, size_of_local_color_table))

    # Local color table
    local_color_table_size = 2 ** ((size_of_local_color_table >> 4) + 1)
    for i in range(local_color_table_size):
        out.write(rng.read(3))

    # Graphic control extension
    block_size = 4
    packed_fields = 0
    delay_time = rng.read(2)
    transparent_color_index = rng.read(1)[0]
    terminator = 0
    out.write(struct.pack('<BBHBBB', block_size, packed_fields,
                          delay_time[1], delay_time[0],
                          transparent_color_index, terminator))

    # Plain text extension
    block_size = 12
    grid_left = rng.read(2)
    grid_top = rng.read(2)
    grid_width = rng.read(2)
    grid_height = rng.read(2)
    cell_width = rng.read(1)[0]
    cell_height = rng.read(1)[0]
    foreground_color_index = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    out.write(struct.pack('<BBHHHHBB', block_size, grid_left[1], grid_left[0],
                          grid_top[1], grid_top[0], grid_width[1],
                          grid_width[0], grid_height[1], grid_height[0],
                          cell_width, cell_height, foreground_color_index,
                          background_color_index))

    # Comment extension
    block_size = 11
    comment = b'\x00' * 11
    out.write(struct.pack('<BB', block_size, len(comment)))
    out.write(comment)

    # Application extension
    block_size = 11
    application_identifier = b'A' * 8
    application_authentication_code = b'B' * 3
    out.write(struct.pack('<BB', block_size, len(application_identifier)))
    out.write(application_identifier)
    out.write(application_authentication_code)

    # Data sub-blocks
    while True:
        data = rng.read(1)
        if not data or data == b'\x00':
            break
        out.write(data)
