#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    screen_width, screen_height = struct.unpack('>HH', rng.read(4))
    flags = struct.unpack('>H', rng.read(2))[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', screen_width, screen_height, flags & 0xFF, background_color_index, pixel_aspect_ratio)

    # 3. Color Table
    has_global_color_table = bool(flags >> 7)
    global_color_table_size = ((flags >> 4) & 0x7) + 1
    global_color_table = rng.read((3 * global_color_table_size))
    color_table = b''
    if has_global_color_table:
        color_table = global_color_table

    # 4. Graphics Control Extension
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        delay_time = struct.unpack('>H', rng.read(2))[0]
        transparent_color_index = rng.read(1)[0]
        terminator = rng.read(1)[0]
        gce = struct.pack('>BBBB', 0xF9, 4, delay_time, transparent_color_index)
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom as get_random_bytes
    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(get_random_bytes(), f)