#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        out.write(header + lsd)
        return

    global_color_table_size = 2 ** ((packed >> 7) & 0b111)
    global_color_table_bytes = global_color_table_size * 3
    global_color_table = rng.read(global_color_table_bytes)

    # 4. Image Data
    image_descriptor = struct.pack('>BHBBBB', 0x2c, 0, 0, width, height, 0)
    image_data = rng.read(width * height)

    # Write everything to `out`
    out.write(header + lsd + global_color_table + image_descriptor + image_data)


# Generates a random PNG file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_png(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Chunk 0: IHDR
    width, height = struct.unpack('>II', rng.read(8))
    bit_depth = rng.read(1)[0]
    color_type = rng.read(1)[0]
    compression_method = rng.read(1)[0]
    filter_method = rng.read(1)[0]
    interlace_method = rng.read(1)[0]
    chunk_0 = struct.pack('>Icccccc', 13, width, height, bit_depth, color_type, compression_method, filter_method, interlace_method)

    # 2. Chunk 1: PLTE
    palette_present = bool(color_type == 3)
    if palette_present:
        palette_entries = 2 ** bit_depth
        palette_size = palette_entries * 3
        palette = rng.read(palette_size)
        chunk_1 = struct.pack('>I', palette_size) + palette
    else:
        chunk_1 = b''

    # 3. Chunk 2: IDAT
    idat_size = width * height * 3
    idat = rng.read(idat_size)
    chunk_2 = struct.pack('>I', idat_size) + idat

    # 4. Chunk 3: IEND
    chunk_3 = struct.pack('>I', 0)

    # Write everything to `out`
    out.write(chunk_0 + chunk_1 + chunk_2 + chunk_3)


# Generates a random JPEG file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_jpeg(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Start of Image
    soi = b'\xff\xd8'

    # 2. Application-specific marker segment
    app_segment_length = 5
    app_segment_identifier = 'JFIF'
    app_segment_version = rng.read(2)
    units = rng.read(1)[0]
    xdensity, ydensity = struct.unpack('>HH', rng.read(4))
    thumbnail_width = rng.read(1)[0]
    thumbnail_height = rng.read(1)[0]
    app_segment = struct.pack('>BBBHBBHHBB', 0xe0, app_segment_length, ord(app_segment_identifier[0]), ord(app_segment_identifier[1]), ord(app_segment_identifier[2]), ord(app_segment_identifier[3]), app_segment_version[0], app_segment_version[1], units, xdensity, ydensity, thumbnail_width, thumbnail_height)

    # 3. Quantization tables
    quantization_tables = []
    for i in range(2):
        quantization_table_precision = rng.read(1)[0]
        quantization_table_values = [rng.read(1)[0] for _ in range(64)]
        quantization_tables.append(struct.pack('>BB', quantization_table_precision, quantization_