#!/usr/bin/env python3

import random
import struct
from math import ceil

def generate_random_gif(rng, out):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed = rng.read(1)
    global_color_table_flag = bool(packed & 0x80)
    color_resolution = (packed >> 4) & 0x07
    sort_flag = bool((packed >> 3) & 0x01)
    global_color_table_size = 2 ** ((packed & 0x07) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    lsd = struct.pack('>HHBBBBB', *struct.unpack('>HHHBBB', width+height+packed+background_color_index+pixel_aspect_ratio))
    
    # 3. Global Color Table
    if global_color_table_flag:
        colors = [rng.read(3) for _ in range(global_color_table_size)]
        color_table = b''.join(colors)
    else:
        color_table = None
    
    # 4. Graphic Control Extension
    delay_time = rng.read(2)
    transparent_color_flag = bool(rng.read(1)[0] & 0x01)
    transparent_color_index = rng.read(1)[0]
    block_terminator = rng.read(1)[0]
    
    gce = struct.pack('>BBBBB', 0xF9, 4, delay_time[0], delay_time[1], transparent_color_index)
    
    # 5. Image Descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        
        if not block_size:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)