#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Packed fields
    #   Global Color Table Flag: 1
    #   Color Resolution: 7
    #   Sort Flag: 0
    #   Size of Global Color Table: 7
    packed_fields = struct.pack('B', 0b11100000 | (7 & 0x07))

    # Background color index
    background_color_index = rng.read(1)[0] % 256

    # Pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)[0] % 256

    # Write everything to output
    out.write(header + struct.pack('<HHBBB', width, height, *struct.unpack('BB', packed_fields), background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table
    # We use 256 colors, so we need 3 bytes per color
    global_color_table = rng.read(3*256)
    out.write(global_color_table)

    # 4. Image Data
    # We have one image descriptor
    out.write(struct.pack('<HHHBB', 0, 0, width, height, 0))

    # Local Color Table Flag: 0
    # Interlace Flag: 0
    # Sort Flag: 0
    # Size of Local Color Table: 0
    local_color_table_flag = 0
    interlace_flag = 0
    sort_flag = 0
    size_of_local_color_table = 0
    packed_fields = struct.pack('BBB', local_color_table_flag, interlace_flag, sort_flag | (size_of_local_color_table << 4))

    # LZW Minimum Code Size
    lzw_minimum_code_size = rng.read(1)[0] % 256

    # Write everything to output
    out.write(struct.pack('BBB', *struct.unpack('BB', packed_fields), lzw_minimum_code_size))

    # 5. Data Sub-Blocks
    # We are going to write 1 data sub-block
    # The first byte is the length of the block
    # We use 256 colors, so we need 256 bytes
    # We can just read from /dev/urandom
    out.write(rng.read(256))

    # 6. Trailer
    # We don't need anything here

