#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor (LSD)
    width = int.from_bytes(rng.read(2), byteorder='little') % 65536
    height = int.from_bytes(rng.read(2), byteorder='little') % 65536
    background_color_index = int.from_bytes(rng.read(1), byteorder='big') % 256
    pixel_aspect_ratio = int.from_bytes(rng.read(1), byteorder='big') % 256
    lsd = struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio)

    # 3. Color Table
    # We use a fixed palette of 4 colors for simplicity
    color_table = bytes([
        0xFF, 0xFF, 0xFF,   # White
        0x00, 0x00, 0x00,   # Black
        0xFF, 0x00, 0x00,   # Red
        0x00, 0xFF, 0x00,   # Green
    ])

    # 4. Graphic Control Extension (GCE)
    # We don't need any animation so we just set the delay to 0
    gce = struct.pack('<BBHBB', 0xF9, 0x04, 0, 0, 0)

    # 5. Image Descriptor
    image_descriptor = struct.pack('<BBHHB', 0x2C, 0, 0, width, height, 0)

    # 6. Image Data
    # We just fill the whole image with a single color index
    image_data = bytes([background_color_index] * (width * height))

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom as get_random_bytes
    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(get_random_bytes(), f)