#!/usr/bin/env python3

import struct
from typing import BinaryIO


def generate_random_gif(data, out: BinaryIO):
    """Generate a GIF file containing the given data"""

    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor (LSD)
    width = len(data)
    height = 1
    packed_fields = 0b00000000  # No global colour table, no transparency, no sort flag
    background_colour = 0
    pixel_aspect_ratio = 0
    lsd = struct.pack('<HHBBH', width, height, packed_fields, background_colour, pixel_aspect_ratio)

    # 3. Global Color Table
    # We don't use a global color table so we just write an empty one
    color_table = b''

    # 4. Graphics Control Extension (GCE)
    # We don't need a GCE for this example but it would go here if we did
    gce = b''

    # 5. Image Descriptor
    left = 0
    top = 0
    width = len(data)
    height = 1
    packed_fields = 0b00000000  # No local colour table, no interlacing, no sort flag
    image_descriptor = struct.pack('<HHBBB', left, top, width, height, packed_fields)

    # 6. Image Data
    # The LZW minimum code size is set to 8 bits because that's what we have in our data
    lzw_minimum_code_size = 8
    blocks = []
    for i in range(len(data)):
        # Each byte becomes its own block
        blocks.append(struct.pack('B', data[i]))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom as get_random_bytes
    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(get_random_bytes(), f)