#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        out.write(header + lsd)
        return

    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Image Data
    image_data = []
    while True:
        block_type = rng.read(1)[0]
        if block_type == 0x2c:  # Image Block
            break
        elif block_type == 0x3b:  # Trailer Block
            raise ValueError("Unexpected trailer block")
        else:
            length = rng.read(1)[0]
            data = rng.read(length)
            image_data.append((block_type, length, data))

    # 5. Local Color Table (if present)
    local_color_table_flag = bool(image_data[0][1][0] & 0x80)
    if not local_color_table_flag:
        out.write(header + lsd + color_table)
    else:
        num_local_colors = 2 ** ((image_data[0][1][0] >> 4) + 1)
        local_color_table = rng.read(3 * num_local_colors)  # Random colors
        out.write(header + lsd + color_table + local_color_table)

    # 6. Image Data
    for block in image_data:
        out.write(struct.pack('B', block[0]))
        out.write(struct.pack('B', block[1]))
        out.write(block[2])

    # 7. Trailer Block
    out.write(b'\x3b')

