#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big') % 640 + 1
    height = int.from_bytes(rng.read(2), 'big') % 480 + 1
    background_color_index = rng.read(1)[0] & 0xFF
    pixel_aspect_ratio = rng.read(1)[0] & 0xFF
    screen_descriptor = struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio)
    out.write(screen_descriptor)

    # 3. Global Color Table
    global_color_table_flag = (rng.read(1)[0] >> 7) & 0b1
    color_resolution = ((rng.read(1)[0] >> 4) & 0b11) + 1
    sort_flag = (rng.read(1)[0] >> 3) & 0b1
    size_of_global_color_table = (((rng.read(1)[0] >> 0) & 0b111) + 1) * 3
    if global_color_table_flag == 1:
        for i in range(size_of_global_color_table):
            red = rng.read(1)[0] & 0xFF
            green = rng.read(1)[0] & 0xFF
            blue = rng.read(1)[0] & 0xFF
            color = struct.pack('BBB', red, green, blue)
            out.write(color)

    # 4. Image Data
    image_separator = b'\x2C'
    out.write(image_separator)

    local_color_table_flag = (rng.read(1)[0] >> 7) & 0b1
    interlace_flag = (rng.read(1)[0] >> 6) & 0b1
    sort_flag = (rng.read(1)[0] >> 5) & 0b1
    reserved = (rng.read(1)[0] >> 4) & 0b11
    size_of_local_color_table = (((rng.read(1)[0] >> 0) & 0b111) + 1) * 3
    x = int.from_bytes(rng.read(2), 'big') % width
    y = int.from_bytes(rng.read(2), 'big') % height
    width = int.from_bytes(rng.read(2), 'big') % (width - x) + 1
    height = int.from_bytes(rng.read(2), 'big') % (height - y) + 1
    if local_color_table_flag == 1:
        for i in range(size_of_local_color_table):
            red = rng.read(1)[0] & 0xFF
            green = rng.read(1)[0] & 0xFF
            blue = rng.read(1)[0] & 0xFF
            color = struct.pack('BBB', red, green, blue)
            out.write(color)

    lzw_minimum_code_size = rng.read(1)[0] & 0xFF
    out.write(lzw_minimum_code_size.to_bytes(1, byteorder='big'))

    def get_next_byte():
        return rng.read(1)[0] & 0xFF

    def read_bits(n):
        bits = 0
        while n > 0:
            if bit_buffer is None or bit_count < n:
                bit_buffer = get_next_byte()
                bit_count += 8
            bits |= bit_buffer & ((1 << n) - 1)
            bit_buffer >>= n
            bit_count -= n
            n -= n
        return bits

    def write_bits(bits, n):
        nonlocal bit_buffer, bit_count
        while n > 0:
            if bit_count