#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        out.write(header + lsd)
        return

    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = rng.read(num_colors * 3)

    # 4. Image Data
    image_data = []
    while True:
        block_type = rng.read(1)[0]
        if block_type == 0x3B:
            break
        elif block_type != 0x2C:
            raise Exception("Invalid block type")

        image_descriptor = rng.read(9)
        left, top, width, height = struct.unpack('>HHHH', image_descriptor[:8])
        local_color_table_flag = bool(image_descriptor[8] & 0x80)
        interlaced = bool(image_descriptor[8] & 0x40)
        sorted = bool(image_descriptor[8] & 0x20)
        reserved = bool(image_descriptor[8] & 0x10)
        size_of_local_color_table = (image_descriptor[8] >> 4) + 1

        if local_color_table_flag:
            color_table = rng.read(size_of_local_color_table * 3)

        data = rng.read(width * height)
        image_data.append((left, top, width, height, data))

    # Write everything to the output buffer
    out.write(header + lsd + color_table)
    for left, top, width, height, data in image_data:
        out.write(struct.pack('>BBHHH', 0x2C, left, top, width, height))
        out.write(b'\x00')
        out.write(data)
