#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        out.write(header + lsd)
        return

    color_resolution = ((packed >> 4) & 0b111) + 1
    sort_flag = bool((packed >> 3) & 0b1)
    size_of_global_color_table = 2 ** (packed & 0b111)
    global_color_table = rng.read(size_of_global_color_table * 3)

    # 4. Image Data
    while True:
        block_type = rng.read(1)[0]
        if block_type == 0x2c:
            break
        elif block_type == 0x21:
            extension_introducer = rng.read(1)[0]
            if extension_introducer != 0xf9:
                raise Exception("Unknown extension")
            graphic_control_label = rng.read(1)[0]
            assert graphic_control_label == 0xff
            block_size = rng.read(1)[0]
            assert block_size == 0x04
            packed = rng.read(1)[0]
            delay_time = struct.unpack('>H', rng.read(2))[0]
            transparent_color_index = rng.read(1)[0]
            terminator = rng.read(1)[0]
            assert terminator == 0
        else:
            raise Exception("Unknown block type")
    
    image_descriptor = struct.pack('>HHBBB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    local_color_table_flag = bool(image_descriptor[3] & 0x80)
    interlace_flag = bool(image_descriptor[3] & 0x40)
    sort_flag = bool(image_descriptor[3] & 0x20)
    size_of_local_color_table = 2 ** ((image_descriptor[3] & 0b111) + 1)

    if local_color_table_flag:
        local_color_table = rng.read(size_of_local_color_table * 3)
    else:
        local_color_table = global_color_table

    lzw_minimum_code_size = rng.read(1)[0]

    def read_lzw():
        code_size = lzw_minimum_code_size + 1
        clear_code = 1 << code_size
        end_of_information_code = clear_code + 1
        
        table = [chr(i) for i in range(clear_code, clear_code + 256)]
        stack = []
        bits = 0
        masks = [1 << i for i in range(code_size)]
        old_code = -1
        while True:
            if len(stack) > 4096:
                raise Exception("Stack overflow")
            
            if bits < code_size:
                bits += 8
                data = rng.read(1)[0]
                if data is None:
                    raise Exception("Unexpected EOF")
                value |= data << (bits - 8)

            code = value & masks[code_size]
            value &= ~masks[code_size]
            bits -= code_size

            if code == clear_code:
                table = [chr(i) for i in range(clear_code, clear_code + 256)]
                code_size = lzw_minimum_code_size + 1
                continue
            elif code == end_of_information_code:
                break
            elif code <= len(table):
                entry = table[code]
            elif code == len(table):
                if old_code == -1:
                    raise Exception("Invalid LZW code")
                prefix = table[old_code]
                suffix = table[old_code][0]
                entry = prefix + suffix
            else:
                raise Exception("Invalid LZW code")

            if code <= len(table):
                stack.append(entry)
            else:
                stack.append(suffix)

            if
if __name__ == '__main__':
    from sys import argv, stdout

    with open('/dev/urandom', 'rb') as f:
        generate_random_gif(f, stdout.buffer)