#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed fields
    packed = struct.pack('B', 0b10000000 | 0b00000111)
    
    # Background color index
    background_color_index = struct.pack('B', rng.read(1)[0])
    
    # Aspect ratio
    aspect_ratio = struct.pack('B', rng.read(1)[0])
    
    # Write everything to the output
    out.write(header + width + height + packed + background_color_index + aspect_ratio)
    
    # 3. Global Color Table
    # 256 colors
    global_color_table = rng.read(768)
    out.write(global_color_table)
    
    # 4. Image Block
    # First byte is a control code
    image_block_control_code = struct.pack('B', 0x2c)
    
    # Next two bytes are the x-coordinate of the top left corner
    x, y = struct.unpack('>HH', rng.read(4))
    
    # Width and height (again)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed fields
    packed = struct.pack('B', 0b10000000 | 0b00000111)
    
    # Local color table flag
    local_color_table_flag = struct.pack('B', 0b00000000)
    
    # Interlaced flag
    interlaced_flag = struct.pack('B', 0b00000000)
    
    # Sort flag
    sort_flag = struct.pack('B', 0b00000000)
    
    # Size of local color table
    size_of_local_color_table = struct.pack('B', 0b00000000)
    
    # Write all this stuff to the output
    out.write(image_block_control_code + x + y + width + height + packed + local_color_table_flag + interlaced_flag + sort_flag + size_of_local_color_table)
    
    # 5. LZW Minimum Code Size
    lzw_minimum_code_size = struct.pack('B', rng.read(1)[0])
    out.write(lzw_minimum_code_size)
    
    # 6. Pixel Data Sub-Blocks
    while True:
        # Read a single byte
        data = rng.read(1)
        
        # If it's zero, we're done
        if data == b'\x00':
            break
            
        # Otherwise write it to the output
        out.write(data)
    
    # 7. Trailer
    trailer = b'\x3b'
    out.write(trailer)

if __name__ == '__main__':
    with open('/dev/urandom') as f:
        for i in range(100):
            with open(f'{i}.gif', 'wb') as o:
                generate_random_gif(f, o)