#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    # Random background color index
    bg_color_index = int.from_bytes(rng.read(1), 'big') % 256
    # Random pixel aspect ratio
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big') % 256
    # Packed fields
    packed_fields = int.from_bytes(rng.read(1), 'big') % 256
    has_global_color_table = bool((packed_fields >> 7) & 0x1)
    color_resolution = ((packed_fields >> 4) & 0x7) + 1
    sort_flag = bool((packed_fields >> 3) & 0x1)
    global_color_table_size = 2 ** ((packed_fields & 0x7) + 1)
    # Global Color Table
    if has_global_color_table:
        global_color_table = [
            struct.unpack('BBB', rng.read(3))
            for _ in range(global_color_table_size)
        ]
    else:
        global_color_table = []

    # Write everything to output
    out.write(header)
    out.write(struct.pack('<HHBBBB', width, height, bg_color_index, pixel_aspect_ratio, packed_fields))
    for rgb in global_color_table:
        out.write(struct.pack('BBB', *rgb))

    # 3. Image Descriptors
    # Number of images
    num_images = int.from_bytes(rng.read(1), 'big') % 256
    # Loop through all images
    for i in range(num_images):
        # 3.1 Image Separator
        out.write(b'\x2C')
        # 3.2 Image Descriptor
        # Left position
        left = int.from_bytes(rng.read(2), 'big') % width
        # Top position
        top = int.from_bytes(rng.read(2), 'big') % height
        # Width
        image_width = int.from_bytes(rng.read(2), 'big') % (width - left)
        # Height
        image_height = int.from_bytes(rng.read(2), 'big') % (height - top)
        # Local Color Table Flag
        local_color_table_flag = bool(int.from_bytes(rng.read(1), 'big') % 2)
        # Interlaced Flag
        interlaced_flag = bool(int.from_bytes(rng.read(1), 'big') % 2)
        # Sort Flag
        sort_flag = bool(int.from_bytes(rng.read(1), 'big') % 2)
        # Reserved
        reserved = int.from_bytes(rng.read(1), 'big') % 256
        # Size of Local Color Table
        local_color_table_size = 2 ** (int.from_bytes(rng.read(1), 'big') % 8)
        # Packed Fields
        packed_fields = int.from_bytes(rng.read(1), 'big') % 256
        local_color_table_exists = bool((packed_fields >> 7) & 0x1)
        # Background Color Index
        bg_color_index = int.from_bytes(rng.read(1), 'big') % 256
        # Pixel Aspect Ratio
        pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big') % 256
        # Write everything to output
        out.write(struct.pack('<HHBBBBB', left, top, image_width, image_height, packed_fields, bg_color_index, pixel_aspect_ratio))
        # If there is a local color table, write it
        if local_color_table_exists:
            for _ in range(local_color_table_size):
                out.write(struct.pack('BBB', *[int.from_bytes(rng.read(1), 'big') % 256 for