#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table = rng.read(1)  # Whether to use a local color table
    
    # 6. Local Color Table (if needed)
    if local_color_table == b'\x01':
        num_colors = (rng.read(1)[0] % 256) or 1
        color_table = rng.read(3 * num_colors)  # Random colors
        
    # 7. Image Data
    image_data = rng.read(width * height)  # Random data
    
    # 8. Trailer
    trailer = b';'
    
    # Write everything in order
    out.write(header + lsd + color_table + gce + b'\x2C' + struct.pack('>HH', left, top) + struct.pack('>HH', width, height) + local_color_table + color_table + image_data + trailer)


# Generate a random GIF file of specified dimensions
def generate_random_gif_dimensions(
        rng: BinaryIO, out: BinaryIO, width: int, height: int
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Fixed width and height
    lsd = struct.pack('>HHBHB', width, height, 0, 0, 0)
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''
    
    # 5. Image Descriptor
    left, top = 0, 0
    width, height = width, height
    local_color_table = rng.read(1)  # Whether to use a local color table
    
    # 6. Local Color Table (if needed)
    if local_color_table == b'\x01':
        num_colors = (rng.read(1)[0] % 256) or 1
        color_table = rng.read(3 * num_colors)  # Random colors
        
    # 7. Image Data
    image_data = rng.read(width * height)  # Random data
    
    # 8. Trailer
    trailer = b';'
    
    # Write everything in order
    out.write(header + lsd + color_table + gce + b'\x2C' + struct.pack('>HH', left, top) + struct.pack('>HH', width, height) + local_color_table + color_table + image_data + trailer)


# Generate a random GIF file with a single frame
def generate_random_single_frame_gif(
        rng: BinaryIO, out: BinaryIO, width: int, height: int
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Fixed width and height
    lsd = struct.pack('>HHBHB', width, height, 0, 0, 0)
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0