#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        out.write(header + lsd)
        return

    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    data = []
    while True:
        block_size = rng.read(1)[0] & 0xFF

        if block_size == 0:
            break

        data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'

    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)

with open('/dev/urandom', 'rb') as rng, \
     open('test.gif', 'wb') as out:
    generate_random_gif(rng, out)
\end{code}

Answer: The problem is that you are writing to the output file in binary mode but reading from it in text mode. This means that when you write the byte `0xff`, Python will interpret this as an end of line character and convert it to `\n`.

You can fix this by opening your files in binary mode:

\begin{code}
with open('/dev/urandom', 'rb') as rng, \
     open('test.gif', 'wb') as out:
    generate_random_gif(rng, out)
\end{code}

Comment: I tried adding `b` to the file modes, but I still get the same error.

Comment: @user1717828: Are you sure? You should be getting a different error message now.

Comment: Yes, I am sure. I copied and pasted your code exactly.

Comment: @user1717828: What version of Python are you using?

Comment: I'm using Python 3.6.5 on Ubuntu 18.04.

Comment: @user1717828: That's odd. It works for me with Python 3.6.5.

Comment: @user1717828: Try running the script again, but this time use `python3` instead of just `python`.

Comment: I did try that, but I got the same error. I also tried changing the shebang to `#!/usr/bin/env python3` but I got the same error.

Comment: @user1717828: That's very strange. When I run the script, I don't get any errors.

Comment: I've added my full code to the question. Maybe there's something else going on.

Comment: @user1717828: Your full code doesn't have the `b` flag in the file modes.

Comment: Ah, I see. I forgot to add it back in. Now I'm getting `TypeError: a bytes-like object is required, not 'str'` on the line `data.append(rng.read(block_size))`

Comment: @user1717828: You need to change all instances of `rng.read()` to `rng.read(1)`.

Comment: I changed them all to `rng.read(1)` and I'm still getting the same error.

Comment: @user1717828: Can you post the full traceback?

Comment: I'm getting the exact same error as before: `TypeError: a bytes-like object is required, not 'str'`.

Comment: @user1717828: That's really weird. I'm not getting any errors.

Comment: Let us [continue this discussion in chat](https://chat.stackoverflow.com/rooms/182142/discussion-between-barmar-and-user1717828).

Comment: @user1717828: I think we figured it out. See my answer.

Answer: I found the issue. In addition to the `b` flags, I needed to change `rng.read(block_size)` to `rng.read(block_size