#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), byteorder='little', signed=False)
    height = int.from_bytes(rng.read(2), byteorder='little', signed=False)
    packed = rng.read(1)[0]
    global_color_table_flag = (packed >> 7) & 1
    color_resolution = (packed >> 4) & 7
    sort_flag = (packed >> 3) & 1
    size_of_global_color_table = 2 ** ((packed >> 0) & 7)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    # 3. Global Color Table
    if global_color_table_flag == 1:
        for i in range(size_of_global_color_table * 3):
            out.write(rng.read(1))
        
    # 4. Image Data
    while True:
        # 4.1 Graphic Control Extension
        block_type = rng.read(1)[0]
        if block_type != 0xF9:
            break
            
        block_length = rng.read(1)[0]
        packed = rng.read(1)[0]
        disposal_method = (packed >> 2) & 7
        user_input_flag = (packed >> 1) & 1
        transparent_color_flag = (packed >> 0) & 1
        delay_time = int.from_bytes(rng.read(2), byteorder='little', signed=False)
        transparent_color_index = rng.read(1)[0]
        block_terminator = rng.read(1)[0]
        
        # 4.2 Image Descriptor
        image_separator = rng.read(1)[0]
        left = int.from_bytes(rng.read(2), byteorder='little', signed=False)
        top = int.from_bytes(rng.read(2), byteorder='little', signed=False)
        width = int.from_bytes(rng.read(2), byteorder='little', signed=False)
        height = int.from_bytes(rng.read(2), byteorder='little', signed=False)
        packed = rng.read(1)[0]
        local_color_table_flag = (packed >> 7) & 1
        interlace_flag = (packed >> 6) & 1
        sort_flag = (packed >> 5) & 1
        size_of_local_color_table = 2 ** ((packed >> 0) & 7)
        
        # 4.3 Local Color Table
        if local_color_table_flag == 1:
            for i in range(size_of_local_color_table * 3):
                out.write(rng.read(1))
                
        # 4.4 LZW-compressed Image Data
        lzw_minimum_code_size = rng.read(1)[0]
        while True:
            code_size = rng.read(1)[0]
            if code_size == 0:
                break
            else:
                for i in range(code_size):
                    out.write(rng.read(1))
                    
if __name__ == "__main__":
    from sys import argv
    if len(argv) < 2:
        print("Usage: %s <output>" % argv[0])
    else:
        with open('/dev/urandom', 'rb') as f:
            with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)