#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = rng.read(num_colors * 3)
    
    # 4. Image Data
    # Let's just use a single image
    image_descriptor = struct.pack('>BBHHBB', 0x2c, 0, 0, width, height, 0)
    
    # We'll need to write the length of the image data later on
    image_data_length_pos = out.tell()
    
    # Write the image data itself
    out.write(image_descriptor + global_color_table)
    
    # Now we can start writing the actual image data
    while True:
        # First, let's pick a random code size
        code_size = rng.read(1)[0] & 7
        
        # Then, let's write that code size
        out.write(bytes([code_size]))
        
        if code_size == 0:
            break
            
        # Finally, let's write some random data
        out.write(rng.read((1 << code_size) - 1))
        
    # Go back and fill in the length field
    out.seek(image_data_length_pos)
    out.write(struct.pack('<H', out.tell() - image_data_length_pos))
    
    # Done!
    out.write(header + lsd)

if __name__ == "__main__":
    from sys import argv
    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)