#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
	# 2. Logical screen descriptor (LSD)
    width = int.from_bytes(rng.read(2), byteorder='big') % 640 + 1
    height = int.from_bytes(rng.read(2), byteorder='big') % 480 + 1
    color_table_size = int.from_bytes(rng.read(1), byteorder='big') & 7
    global_color_table = bool((int.from_bytes(rng.read(1), byteorder='big') >> 7) & 1)

    lsd = [width, height]
    if global_color_table == True:
        lsd += [(1 << (i+1)) for i in range(color_table_size)]
    else:
        lsd += [0] * ((1<<color_table_size)-1)
    lsd += [0]
    lsd += [0x80 | 0x40 | 0x20 | 0x10]
    lsd += [0]
    lsd += [0]
    lsd += [0]
    lsd += [0]
    lsd += [0]
    lsd += [0]
    lsd += [0]
    lsd = bytes([*lsd])
    assert len(lsd) == 10

    # 3. Global color table
    gct = []
    if global_color_table == False:
        gct += [0] * ((1<<color_table_size)*3)
    else:
        for _ in range(len(lsd[7:])//3):
            r = int.from_bytes(rng.read(1), byteorder='big')
            g = int.from_bytes(rng.read(1), byteorder='big')
            b = int.from_bytes(rng.read(1), byteorder='big')

            gct += [r,g,b]
    gct = bytes([*gct])
    assert len(gct) == len(lsd[7:])

    # 4. Image block
    image_block = []
    while True:
        image_descriptor = rng.read(9)

        left = int.from_bytes(image_descriptor[:2], byteorder='big')
        top = int.from_bytes(image_descriptor[2:4], byteorder='big')
        width = int.from_bytes(image_descriptor[4:6], byteorder='big')
        height = int.from_bytes(image_descriptor[6:], byteorder='big')

        packed = int.from_bytes(rng.read(1), byteorder='big')
        local_color_table = bool((packed >> 7) & 1)
        interlace = bool((packed >> 6) & 1)
        sort = bool((packed >> 5) & 1)
        reserved = bool((packed >> 4) & 1)
        lct_size = (packed & 7) + 1

        flags = 0x80 | 0x40 | 0x20 | 0x10
        if local_color_table == True:
            flags |= 0x08
        if interlace == True:
            flags |= 0x40
        if sort == True:
            flags |= 0x20
        if reserved == True:
            flags |= 0x10
        flags = bytes([flags])

        lct = []
        if local_color_table == False:
            lct += [0] * ((1<<lct_size)*3)
        else:
            for _ in range(len(image_descriptor[8:])//3):
                r = int.from_bytes(rng.read(1), byteorder='big')
                g = int.from_bytes(rng.read(1), byteorder='big')
                b = int.from_bytes(rng.read(1), byteorder='big')

                lct += [r,g,b]
        lct = bytes([*lct])
        assert len(lct) == len(image_descriptor[8:])

        data = []
        for y in range(height):
            line = []
            for x in range(width):
                line += [int.from_bytes(rng.read(1), byteorder='big')]
            data += line
        data = bytes([*data])
        assert len(data) == width * height

        image_block += [*image_descriptor[:8], *flags, *lct, *data]

        if not image_descriptor[8:] or image_descriptor[8:] == b'\x00':
            break

    image_