#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
     # 2. Logical screen descriptor
    width = rng.read(1)[0] % (2**6 + 1)
    height = rng.read(1)[0] % (2**6 + 1)

    color_table_size = 7 if width <= 5 else \
        1 << ((width * height - 1).bit_length() // 4)
    global_color_table = [rng.read(3) for _ in range(color_table_size)]
    background_color_index = rng.read(1)[0] % len(global_color_table)
    pixel_aspect_ratio = 0x00

    logical_screen_descriptor = bytearray([
        width, height,
        global_color_table[background_color_index][0],
        global_color_table[background_color_index][1],
        global_color_table[background_color_index][2],
        pixel_aspect_ratio])

    # 3. Image descriptor
    image_left = 0
    image_top = 0
    image_width = width
    image_height = height
    local_color_table_flag = False
    interlace_flag = True
    sort_flag = True
    reserved = bytes(3)
    lzw_minimum_code_size = 2
    image_descriptor = bytearray([
        image_left, image_top,
        image_width, image_height,
        local_color_table_flag,
        interlace_flag,
        sort_flag,
        reserved[0], reserved[1], reserved[2],
        lzw_minimum_code_size])

    # 4. Global color table
    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(image_descriptor)

    for i in range(len(global_color_table)):
        out.write(global_color_table[i])

    # 5. LZW code stream
    def next_byte(): return ord(rng.read(1))
    def next_bits(n): return int.from_bytes(rng.read(n//8), 'big') & ((1<<n)-1)

    clear_code = 1 << lzw_minimum_code_size
    end_of_information_code = clear_code + 1
    max_code_size = 1 << lzw_minimum_code_size
    bit_mask = (1 << lzw_minimum_code_size) - 1
    current_code_size = lzw_minimum_code_size + 1
    previous_code = None
    code_table = []
    output_buffer = bytearray()
    while True:
        if not code_table or current_code_size == max_code_size:
            code_table += [(next_byte(), next_byte())
                for _ in range(current_code_size-1)]

            current_code_size = lzw_minimum_code_size + 1

        code = next_bits(current_code_size)
        if code < clear_code:
            code_table.append((previous_code, code))
            output_buffer.append(previous_code)
        elif code == end_of_information_code:
            break
        else:
            if code >= len(code_table):
                raise Exception('Invalid code!')

            entry = code_table[code]
            if entry[0] is None:
                raise Exception('Invalid code!')

            output_buffer.append(entry[0])
            code_table.append((entry[1], next_byte()))

        previous_code = code
        current_code_size += 1

    # 6. Trailer
    trailer = b'\x3B'
    out.write(trailer)

