#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        return

    size = 2 << ((packed >> 7) & 0x07)
    sort_flag = bool((packed >> 5) & 0x01)
    global_color_table = [rng.read(1)[0] for _ in range(size)]

    # 4. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    local_color_table_flag = bool(packed & 0x80)
    interlace_flag = bool((packed >> 6) & 0x01)
    sort_flag = bool((packed >> 5) & 0x01)
    local_color_table_size = 2 << ((packed >> 3) & 0x03)
    reserved = rng.read(1)[0]
    lzw_minimum_code_size = rng.read(1)[0]
    image_descriptor = struct.pack('>HHBBBBB', left, top, packed, local_color_table_flag, interlace_flag, sort_flag, local_color_table_size)

    # 5. Local Color Table (if present)
    color_table = bytearray()
    if local_color_table_flag:
        color_table += bytes([local_color_table_size])
        color_table += bytes([reserved])
        color_table += bytes([lzw_minimum_code_size])
        color_table += bytes(global_color_table[:local_color_table_size])

    # 6. Data blocks
    data = []
    while True:
        block_size = rng.read(1)[0] & 0xFF

        if block_size == 0:
            break

        data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)


if __name__ == '__main__':
    from sys import argv, stdout

    with open('/dev/urandom', 'rb') as f:
        generate_random_gif(f, stdout.buffer)