#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    header = b'GIF89a' + bytes([0x12]) + bytes([0x34]) + \
             bytes([0x56]) + bytes([0x78])
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (2**12-1) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (2**12-1) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lsd    = struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0])
    out.write(lsd)

    # Global color table
    gctsize = int.from_bytes(rng.read(1), byteorder='little') % (2**5-1) + 1
    gct = [b'\x00\x00\x00'] * (gctsize ** 3)
    for i in range(len(gct)):
        r = int.from_bytes(rng.read(1), byteorder='little') % 256
        g = int.from_bytes(rng.read(1), byteorder='little') % 256
        b = int.from_bytes(rng.read(1), byteorder='little') % 256
        gct[i] = bytes([r, g, b])
    out.write(struct.pack('B', gctsize))
    for i in range(len(gct)):
        out.write(gct[i])

    # Image descriptor
    left   = int.from_bytes(rng.read(2), byteorder='little') % (width - 1)
    top    = int.from_bytes(rng.read(2), byteorder='little') % (height - 1)
    width  = int.from_bytes(rng.read(2), byteorder='little') % (width - left) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (height - top) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lctsize = int.from_bytes(rng.read(1), byteorder='little') % (2**5-1) + 1
    lct = [b'\x00\x00\x00'] * (lctsize ** 3)
    interlace = False if ((flags[0] & 0b10000000) == 0) else True
    sortedcol = False if ((flags[0] & 0b01000000) == 0) else True
    reserved  = bool((flags[0] & 0b00100000)!= 0)
    lzw       = bool((flags[0] & 0b00010000)!= 0)
    localcol  = bool((flags[0] & 0b00001000)!= 0)
    lsd = struct.pack('<HHHHBB',
                      left, top, width, height, flags[0], lctsize)
    out.write(lsd)

    # Local color table
    if localcol and lctsize > 0:
        for i in range(len(lct)):
            r = int.from_bytes(rng.read(1), byteorder='little') % 256
            g = int.from_bytes(rng.read(1), byteorder='little') % 256
            b = int.from_bytes(rng.read(1), byteorder='little') % 256
            lct[i] = bytes([r, g, b])
        for i in range(len(lct)):
            out.write(lct[i])

    # Graphics control extension
    if not interlace and not sortedcol and not reserved and not lzw:
        blocksize = int.from_bytes(rng.read(1), byteorder='little') % (2**8-1) + 1
        packed = struct.pack('<BBBBHB',
                             int.from_bytes