#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

        # If there is an application extension in this frame, we need to add some more data to its block
        if rng.read(1)[0]:
            appext = struct.pack('>BB', 0xFF, 0x0B) + b'NETSCAPE2.0\x03\x01\x00\x00'

            # Add the application extension to the GCE
            gce += appext

    else:
        gce = b''

    # 5. Image Descriptor
    id = struct.pack('>HHB', width, height, 0b10000000 | rng.read(1)[0])

    # 6. Image Data
    image_data = rng.read((width * height) // 8)

    # 7. Trailer
    trailer = b';'

    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(id)
    out.write(image_data)
    out.write(trailer)

