#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
        
    # Random color table size
    color_table_size = int.from_bytes(rng.read(1), 'big')

    # Global Color Table Flag
    global_color_table_flag = bool(int.from_bytes(rng.read(1), 'big'))

    # Background Color Index
    background_color_index = int.from_bytes(rng.read(1), 'big')

    # Pixel Aspect Ratio
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big')

    logical_screen_descriptor = \
        struct.pack(
            '>HHBBB', 
            width, 
            height, 
            color_table_size, 
            0b1 << 7 | 
                (global_color_table_flag << 6) |
                background_color_index,
            pixel_aspect_ratio
        )

    # 3. Global Color Table
    if global_color_table_flag == True:
        global_color_table = bytearray()

        for i in range(pow(2, color_table_size)):
            r = int.from_bytes(rng.read(1), 'big')
            g = int.from_bytes(rng.read(1), 'big')
            b = int.from_bytes(rng.read(1), 'big')

            global_color_table += struct.pack('BBB', r, g, b)
        
        out.write(header + logical_screen_descriptor + global_color_table)
    else:
        out.write(header + logical_screen_descriptor)

    # 4. Image Descriptor
    image_descriptor = \
        struct.pack(
            '>HBBHHHHB',
            0, # Left position of image
            0, # Top position of image
            width, # Width of image
            height, # Height of image
            0x00, # No local color table
            0b10000000, # Interlace flag set to false
            0, # LZW minimum code size
            ord('!') # Extension Introducer
        )

    # 5. Graphics Control Extension
    graphics_control_extension = \
        struct.pack(
            '>BBBBHB',
            0xf9, # Block Label
            0x04, # Size of block
            0x00, # Packed field
            0x00, # Delay time
            0xff, # Transparent color index
            0x00 # Terminator
        )

    # 6. Application Extension
    application_extension = \
        struct.pack(
            '>BBBB',
            0xff, # Extension introducer
            0x0b, # Size of block
            0x01, # Application identifier
            0xfe # Block terminator
        )

    # 7. Comment Extension
    comment_extension = \
        struct.pack(
            '>BBBB',
            0xff, # Extension introducer
            0x0b, # Size of block
            0xfe, # Comment extension label
            0x00 # Block terminator
        )

    # 8. Plain Text Extension
    plain_text_extension = \
        struct.pack(
            '>BBBBBBB',
            0xff, # Extension introducer
            0x0c, # Size of block
            0x01, # Text grid left position
            0x01, # Text grid top position
            0x01, # Text grid width
            0x01, # Text grid height
            0x00 # Block terminator
        )

    # 9. Image Data
    image_data = bytearray()

    for y in range(height):
        for x in range(width):
            # Get a random byte from /dev/urandom
            image_data.append(int.from_bytes(rng.read(1), 'big'))

    out.write(image_descriptor + graphics_control_extension + 
        application_extension + comment_extension + plain_text_extension + 
        image_data)

