#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
	width &= 0x7FFF
	height &= 0x7FFF

    # Random color table size
    color_table_size = ord(rng.read(1)) & 0b11111

    # Global Color Table Flag
    global_color_table_flag = bool(ord(rng.read(1)))

    # Background Color Index
    background_color_index = ord(rng.read(1))

    # Pixel Aspect Ratio
    pixel_aspect_ratio = ord(rng.read(1))

    logical_screen_descriptor = struct.pack(
        '>HHBBB', width, height, color_table_size,
        global_color_table_flag << 7 | background_color_index,
        pixel_aspect_ratio
    )

    # 3. Global Color Table
    if global_color_table_flag:
        global_color_table = bytearray()

        for i in range(pow(2, color_table_size)):
            r, g, b = [
                int(i * 256 / pow(2, color_table_size)) % 256
                for i in range(3)
            ]

            global_color_table += struct.pack('BBB', r, g, b)

        # Write to output
        out.write(global_color_table)

    # 4. Image Descriptor
    image_left = 0
    image_top = 0
    image_width = width
    image_height = height
    local_color_table_flag = False
    interlace_flag = False
    sort_flag = True
    reserved_bits = 0
    lzw_minimum_code_size = 2 ** 8 - 1

    image_descriptor = struct.pack(
        '>HHHHBB', image_left, image_top, image_width, image_height,
        local_color_table_flag << 7 | interlace_flag << 6 |
        sort_flag << 5 | reserved_bits, lzw_minimum_code_size
    )

    # 5. Local Color Table
    if not global_color_table_flag:
        local_color_table = bytearray()

        for i in range(pow(2, color_table_size)):
            r, g, b = [
                int(i * 256 / pow(2, color_table_size)) % 256
                for i in range(3)
            ]

            local_color_table += struct.pack('BBB', r, g, b)

        # Write to output
        out.write(local_color_table)

    # 6. Block of Data
    data = bytearray()

    while len(data) < 1024:
        # Generate a random block of data
        block = rng.read(int((len(data) + 1) // 2))

        # Add it to our current buffer
        data += block

    # Write to output
    out.write(image_descriptor)
    out.write(data)

    # 7. Trailer
    trailer = b'\x3B'

    # Write to output
    out.write(trailer)