#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        out.write(header + lsd)
        return

    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = rng.read(3 * num_colors)  # Random colors for now
    gct = bytes([num_colors]) + color_table

    # 4. Image Descriptor
    image_separator = b'\x2C'
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    local_color_table_flag = bool(packed & 0x80)
    interlace_flag = bool(packed & 0x40)
    sort_flag = bool(packed & 0x20)
    reserved = rng.read(1)[0]
    lzw_minimum_code_size = rng.read(1)[0]
    id = struct.pack('>HHBBB', left, top, width, height, packed, lzw_minimum_code_size)

    # 5. Local Color Table
    if local_color_table_flag:
        num_colors = 2 ** ((packed >> 4) + 1)
        color_table = rng.read(3 * num_colors)  # Random colors for now
        lct = bytes([num_colors]) + color_table

        out.write(header + gct + image_separator + id + lct)
        return

    out.write(header + gct + image_separator + id)

