#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = min((width * height // 100) % 256, 256)
    color_table = b''
    for i in range(num_colors):
        color_table += struct.pack('BBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 4. Graphics Control Extension
    # Frame delay time (two bytes, using a random value between 0-65535 ms)
    frame_delay = struct.pack('<H', rng.read(2)[0] | (rng.read(1)[0] << 8))

    # Transparent color index (using a random byte as an index)
    transparent_color_index = struct.pack('B', rng.read(1)[0])

    # Block terminator
    block_terminator = b'\x00'

    # Putting them together
    gce = frame_delay + transparent_color_index + block_terminator

    # 5. Image Descriptor
    # Left position (two bytes, using a random value between 0-65535 pixels from left edge of screen)
    left_position = struct.pack('<H', rng.read(2)[0] | (rng.read(1)[0] << 8))

    # Top position (two bytes, using a random value between 0-65535 pixels from top edge of screen)
    top_position = struct.pack('<H', rng.read(2)[0] | (rng.read(1)[0] << 8))

    # Width (two bytes, using a random value between 0-65535 pixels)
    width = struct.pack('<H', rng.read(2)[0] | (rng.read(1)[0] << 8))

    # Height (two bytes, using a random value between 0-65535 pixels)
    height = struct.pack('<H', rng.read(2)[0] | (rng.read(1)[0] << 8))

    # Flags (packed field, using a random byte for variety)
    flags = struct.pack('B', rng.read(1)[0])

    # Image data (using a random byte for variety)
    image_data = rng.read(width[0] * height[0])

    # Putting them together
    image_descriptor = left_position + top_position + width + height + flags

    # 6. Trailer
    trailer = b';'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)

