#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = rng.read(1)[0] % 64 + 1
    h = rng.read(1)[0] % 64 + 1

    # Color table size is always 7
    color_table_size = 7

    # Global color table flag is set if there's no local color table
    global_color_table_flag = 1

    # Background color index is always 0
    background_color_index = 0

    # Pixel aspect ratio is always 0
    pixel_aspect_ratio = 0

    logical_screen_descriptor = struct.pack(
        '<6sHHBBB',
        header,
        w * 256 + h,
        color_table_size,
        global_color_table_flag,
        background_color_index,
        pixel_aspect_ratio
    )

    # Write to output stream
    out.write(logical_screen_descriptor)

    # 3. Global Color Table
    # Generate 256 colors with RGB values between [0, 255]
    global_color_table = bytearray()
    for i in range(256):
        r = rng.read(1)[0] & 0xFF
        g = rng.read(1)[0] & 0xFF
        b = rng.read(1)[0] & 0xFF

        global_color_table += struct.pack('<BBB', r, g, b)

    # Write to output stream
    out.write(global_color_table)

    # 4. Image Descriptor
    image_separator = b','

    # Image left position is always 0
    image_left_position = 0

    # Image top position is always 0
    image_top_position = 0

    # Width of the image is randomly generated from 1 to 64 pixels
    image_width = rng.read(1)[0] % 64 + 1

    # Height of the image is randomly generated from 1 to 64 pixels
    image_height = rng.read(1)[0] % 64 + 1

    # Local color table flag is set if there's a local color table
    local_color_table_flag = 1

    # Interlace flag is always 0
    interlace_flag = 0

    # Sorting flag is always 0
    sorting_flag = 0

    # Size of the local color table is always 7
    local_color_table_size = 7

    # LZW minimum code size is always 2
    lzw_minimum_code_size = 2

    image_descriptor = struct.pack(
        '<BHHBBBBB',
        image_separator,
        image_left_position * 256 + image_top_position,
        image_width * 256 + image_height,
        local_color_table_flag,
        interlace_flag,
        sorting_flag,
        local_color_table_size,
        lzw_minimum_code_size
    )

    # Write to output stream
    out.write(image_descriptor)

    # 5. Local Color Table
    # Generate 256 colors with RGB values between [0, 255]
    local_color_table = bytearray()
    for i in range(256):
        r = rng.read(1)[0] & 0xFF
        g = rng.read(1)[0] & 0xFF
        b = rng.read(1)[0] & 0xFF

        local_color_table += struct.pack('<BBB', r, g, b)

    # Write to output stream
    out.write(local_color_table)

    # 6. Extension Block
    extension_block_label = b'\x21\xF9\x04'

    # Graphic control label is always 0
    graphic_control_label = 0

    # Delay time is always 0
    delay_time = 0

    # Transparent color index is always 0
    transparent_color_index = 0

    # Block terminator is always 0
    block_terminator = 0

    # Extension block data
    extension_block_data = struct.pack(
        '<BBBBHB',
        graphic_control_label,
        0,
        delay_time,
        transparent_color_index,
        block_terminator
    )

    # Write to output stream
    out.write(extension_block_label)
    out.write(struct.pack('>H', len(extension_block_data)))
    out.write(extension_block_data)

    # 7. Image Data
    # Get the next bit as an integer value
    def get_next_bit():
        return int.from_bytes(rng.read(1),