#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table = bool(rng.read(1)[0])
    interlace = bool(rng.read(1)[0])
    sort = bool(rng.read(1)[0])
    reserved = rng.read(3)
    lzw_minimum_code_size = rng.read(1)[0]
    image_descriptor = struct.pack('>HHBBBBB', left, top, width, height, lzw_minimum_code_size, local_color_table << 7 | interlace << 6 | sort << 5, reserved[0])

    # 6. Image Data
    # Let's use a simple LZW compression scheme here
    # First we need to build our dictionary by adding all possible characters in the range [0..255]
    dictionary = {chr(i): i for i in range(256)}

    # Then we can start compressing data from the input stream
    blocks = []
    while True:
        # Read next character from the input stream
        char = chr(rng.read(1)[0])

        # If this is an end-of-stream marker, break out of the loop
        if not char:
            break

        # If this character exists in the dictionary already, just add its code to the current block
        if char in dictionary:
            blocks[-1] += bytes([dictionary[char]])

            # If the last block has reached maximum capacity, create another one
            if len(blocks[-1]) == 255:
                blocks.append(b'\xFF')
        # Otherwise, let's find the longest prefix that exists in the dictionary
        else:
            max_prefix_len = min((len(blocks[-1]), 127 - len(blocks)))
            best_match = ''
            for i in reversed(range(max_prefix_len)):
                candidate = blocks[-1][:-i] + char

                # If this candidate matches any entry in the dictionary, update the best match string so far
                if candidate in dictionary:
                    best_match = candidate

                    # We also know how many bits are needed to encode this candidate
                    num_bits = len(best_match) + 1

                    # Since the minimum code size is 2, we can only store up to 127 codes per block
                    # So if the candidate requires more than 127 bits, we need to split it across multiple blocks
                    if num_bits > 127:
                        num_bits -= 127

                        # The first part of the candidate will be stored as usual
                        blocks[-1] = bytes([dictionary[candidate[:-num_bits]]])

                        # The second part needs to be split across multiple blocks
                        while num_bits >= 255:
                            blocks.append(bytes([255]))
                            num_bits -= 255
                        blocks.append(bytes([num_bits]))
                        blocks.append(bytes([dictionary[candidate[-num_bits:]]]))
                    else:
                        # This candidate fits within the current block, so we can simply replace the last block contents
                        blocks[-1] = bytes([dictionary[candidate]])
                    break

            # If no candidates matched, then we need to add a new entry to the dictionary
            if not best_match:
                dictionary[best_match + char] = len(dictionary)
                blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)

