#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table
    # Number of colors in global color table (between 2 and 256 inclusive). If zero, no global color table is used.
    num_colors = rng.read(1)[0] % 256 + 2
    
    # Each entry has 3 bytes: Red, Green, Blue
    color_table = b''.join((struct.pack('BBB', *rng.read(3)) for _ in range(num_colors)))
    
    # 4. Graphics Control Extension
    # Block size (always 4)
    gce_block_size = b'\x04'
    
    # Flags: packed field (using a random byte for variety)
    flags = rng.read(1)[0]
    
    # Delay time (two bytes): 1/100th of a second between frames
    delay_time = struct.pack('>H', rng.read(2)[0] % 65536)
    
    # Transparent color index
    transparent_color_index = struct.pack('B', rng.read(1)[0] % num_colors)
    
    # Terminator (always \x00)
    terminator = b'\x00'
    
    # Put it all together
    gce = gce_block_size + flags + delay_time + transparent_color_index + terminator
    
    # 5. Image Descriptor
    # Left position: 0
    left_position = struct.pack('>H', 0)
    
    # Top position: 0
    top_position = struct.pack('>H', 0)
    
    # Width and Height are equal to those of the logical screen descriptor
    width = struct.pack('>H', width)
    height = struct.pack('>H', height)
    
    # Packed field: local color table flag set, interlace flag clear, sort flag clear, reserved bits clear
    packed_field = b'\x40'
    
    # Local color table length: 0
    local_color_table_length = b'\x00'
    
    # Background color index: 0
    background_color_index = b'\x00'
    
    # Pixel aspect ratio: 0
    pixel_aspect_ratio = b'\x00'
    
    # Put it all together
    image_descriptor = left_position + top_position + width + height + packed_field + local_color_table_length + background_color_index + pixel_aspect_ratio
    
    # 6. Image Data
    # Minimum code size: 2 or higher
    lzw_minimum_code_size = rng.read(1)[0] % 256 + 2
    
    # Blocks
    blocks = []
    
    while True:
        # Block Size: 1-255 inclusive
        block_size = rng.read(1)[0] % 256
        
        # If we get a block size of 0, we're done        
        if not block_size:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)