#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write("GIF89a".encode())

    # Logical screen descriptor
    width = 256 + (rng.read(1)[0] & 7)
    height = 256 + ((rng.read(1)[0] >> 4) & 7)
    flags = b"\x80" if rng.read(1)[0] & 1 else b"\x00"
    bg_color = bytes([rng.read(1)[0], rng.read(1)[0]])
    pixel_aspect_ratio = b"\x00\x00"
    out.write(struct.pack("<HHBBB", width, height, flags[0], bg_color[0], bg_color[1]))
    out.write(pixel_aspect_ratio)

    # Global color table
    global_color_table_size = min((width * height), 256) // 3
    for i in range(global_color_table_size - 1):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(bytes([r, g, b]))
    out.write(b"\x00\x00\x00")

    # Image descriptor
    left = 0
    top = 0
    flags = b"\x80"
    local_color_table_flag = b"\x00"
    interlace_flag = b"\x00"
    sort_flag = b"\x00"
    reserved = b"\x00\x00"
    lct_size = b"\x00"
    out.write(struct.pack("<HHB", left, top, flags[0]))
    out.write(local_color_table_flag)
    out.write(interlace_flag)
    out.write(sort_flag)
    out.write(reserved)
    out.write(lct_size)

    # Local color table
    for i in range(min((width * height), 256)):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(bytes([r, g, b]))
    out.write(b"\x00\x00\x00")

    # Table based image data
    lzw_min_code_size = rng.read(1)[0]
    out.write(lzw_min_code_size.to_bytes(1, "little"))
    clear_code = 1 << lzw_min_code_size
    end_of_information_code = clear_code + 1
    code_length = [8] * (clear_code + 1)
    next_code = list(range(clear_code + 2))
    output_buffer = []
    current_byte = 0
    current_bit = 0
    previous_code = None
    while True:
        code = rng.read(1)[0] % (end_of_information_code - clear_code + 1) + clear_code

        if code == end_of_information_code:
            break

        if code < len(next_code):
            output_buffer += chr(code).encode()

            if previous_code is not None and previous_code >= clear_code:
                entry = (previous_code << 8) | code

                length = code_length[entry]
                if length < 12:
                    code_length[len(output_buffer)] = length + 1
                elif length == 12:
                    code_length[len(output_buffer)] = 13
                elif length == 13:
                    code_length[len(output_buffer)] = 14

                if length > 4:
                    extra_bits = {
                        14: 18,
                        13: 11,
                        12: 4,
                    }

                    bits = extra_bits[length]
                    value = 0
                    for _ in range(bits):
                        value <<= 1
                        value |= rng.read(1)[0] & 1

                    output_buffer += value.to_bytes(((value.bit_length() + 7) // 8), byteorder="big")

            previous_code = code
        else:
            output_buffer += chr(output_buffer[-1]).encode()
            previous_code = None

        while len(output_buffer) >= 255:
            out.write(bytes([2