#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Magic header
    out.write(b'GIF89a')

    # Screen descriptor
    width = 256 + (rng.read(1)[0] & 0x7f)
    height = 256 + (rng.read(1)[0] & 0x7f)
    flags = b'\x80\x00' if rng.read(1)[0] < 128 else b'\x00\x00'
    bg_color = bytes([rng.read(1)[0], rng.read(1)[0]])
    aspect_ratio = b'\x00'
    out.write(struct.pack('<HHBBB', width, height, *flags))
    out.write(bg_color)
    out.write(aspect_ratio)

    # Global color table
    num_colors = min((width // 8) * (height // 8), 256)
    for i in range(num_colors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(bytes([r, g, b]))

    # Image descriptor
    left = 0
    top = 0
    width = 256 + (rng.read(1)[0] & 0x7f)
    height = 256 + (rng.read(1)[0] & 0x7f)
    flags = b'\x80\x00' if rng.read(1)[0] < 128 else b'\x00\x00'
    duration = 100
    trans_index = 0
    out.write(struct.pack('<HHHHB', left, top, width, height, *flags))
    out.write(struct.pack('<H', duration))
    out.write(trans_index.to_bytes(1, 'little'))

    # LZW minimum code size
    lzw_min_code_size = 2
    out.write(lzw_min_code_size.to_bytes(1, 'little'))

    # Image data
    image_data = bytearray()
    while len(image_data) < ((width // 8) * (height // 8)):
        image_data += rng.read(1)

    def get_next_byte():
        nonlocal image_data

        if not image_data:
            return None

        ret = image_data[0]
        del image_data[0]
        return ret

    # LZW decoder state
    clear_code = 1 << lzw_min_code_size
    end_of_info_code = clear_code + 1
    next_code = end_of_info_code + 1
    prev_code = -1
    bits = 0
    bit_count = 0
    max_bits = 12
    decoded_bytes = []
    current_byte = 0
    current_bit = 0

    def decode_next_code():
        nonlocal next_code
        nonlocal prev_code
        nonlocal bits
        nonlocal bit_count
        nonlocal max_bits

        if next_code == end_of_info_code:
            raise Exception('Unexpected end of info code!')

        if next_code == clear_code:
            max_bits = 12
            next_code = end_of_info_code + 1
            decoded_bytes = [0] * 256

            for i in range(256):
                decoded_bytes[i] = i

            yield from [(0, 0)] * 256
            continue

        if next_code > len(decoded_bytes):
            raise Exception('Invalid code!')

        if prev_code!= -1 and next_code >= clear_code:
            decoded_bytes.append(decoded_bytes[-prev_code])

        yield (current_byte, current_bit)

        if next_code <= len(decoded_bytes):
            current_byte = decoded_bytes[next_code]
            current_bit = 0
        elif next_code == end_of_info_code:
            pass
        else:
            raise Exception('Invalid code!')

        while True:
            if next_code < len(decoded_bytes):
                break

            bits <<= 1
            bit_count += 1
            if bit_count == max_bits:
                bit_count = 0

                new_byte = get_next_byte()
                if new_byte is None:
                    raise Exception('Unexpected EOF!')

                bits |= new_byte

            if bits & (1 << max_bits-1):
                next_code = (bits & ~(~0 << max_bits)) | \
                            (get_next_byte() << (max_