#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed_flags = (
        0b1 << 7 | # Global Color Table Flag
        0b1 << 6 | # Sort Flag
        0b11 << 3 | # Size of Global Color Table
        0b111 # Background Color Index
    )
    flags = bytes([packed_flags])
    background_color_index = bytes([struct.unpack('B', rng.read(1))[0] & 0b111])
    pixel_aspect_ratio = b'\xff'
    lsd = width.to_bytes(2, byteorder='big') + \
          height.to_bytes(2, byteorder='big') + \
          flags + \
          background_color_index + \
          pixel_aspect_ratio

    # 3. Global Color Table
    global_color_table = [
        rng.read(3) for _ in range((1 << ((packed_flags >> 3) & 0b11)))
    ]
    color_table = b''.join(global_color_table)

    # 4. Graphics Control Extension
    packed_flags = (
        0b1 << 7 | # Disposal Method
        0b111 << 5 | # User Input Flag
        0b11111 << 0 # Transparent Color Index
    )
    flags = bytes([packed_flags])
    delay_time = bytes([struct.unpack('B', rng.read(1))[0]]) * 4
    transparent_color_index = bytes([struct.unpack('B', rng.read(1))[0] & 0b111])
    gce = b''+ flags + delay_time + transparent_color_index

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    packed_flags = (
        0b1 << 7 | # Local Color Table Flag
        0b1111 << 4 | # Interlace Flag
        0b1111 << 0 # Size of Local Color Table
    )
    flags = bytes([packed_flags])
    local_color_table_size = bytes([(packed_flags & 0b1111)])
    image_descriptor = left.to_bytes(2, byteorder='big') + \
                       top.to_bytes(2, byteorder='big') + \
                       flags + \
                       local_color_table_size

    # 6. Image Data
    data = []
    for i in range(height):
        line = []

        for j in range(width):
            line += [
                bytes([
                    struct.unpack('B', rng.read(1))[0] & 0b111111,
                    0b11100000 | (
                        struct.unpack('B', rng.read(1))[0] & 0b111
                    ),
                    0b111111
                ])
            ]

        data += [b''.join(line)]

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)


if __name__ == '__main__':
    from sys import argv, stdout

    with open('/dev/urandom', 'rb') as f:
        generate_random_gif(f, stdout.buffer)