#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = min((width * height // 256) + 1, 256)   # Number of colors in GCT
    color_table = bytes([(num_colors << 4) | 1])        # Size of GCT, sort flag set to true
    color_table += bytes(rng.read(3*num_colors))         # Actual GCT data
    
    # 4. Graphics Control Extension (we'll just use defaults here)
    gce = b'\xF9\x04\x00\x00\x00\x00\x00'
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('>HHBBBHBB', 0, 0, 0, 0b10000000, 0, 0, 0)
    
    # 6. Image Data
    # Let's compress some random data!
    # First we need to get all unique bytes
    unique_bytes = {b}
    while True:
        try:
            new_byte = rng.read(1)[0]
        except IndexError:
            break
        
        if new_byte not in unique_bytes:
            unique_bytes.add(new_byte)
    
    # Now let's build our LZW dictionary by adding them one-by-one
    lzw_dictionary = dict(enumerate(unique_bytes))
    lzw_minimum_code_size = max(len(unique_bytes).bit_length(), 2)     # Minimum code size is either 2 or the length of the longest byte string
    
    # We can now start building blocks
    blocks = []
    current_block = [next(iter(lzw_dictionary))]      # Start with first byte in dictionary
    while True:
        try:
            next_byte = rng.read(1)[0]                 # Read another byte
        except IndexError:                            # If there are no more bytes left...
            blocks.append(struct.pack('>B', len(current_block)))       # Add a block size
            blocks.extend(current_block)               # And add that last block
            break                                      # Break out of loop
        
        if tuple(current_block + [next_byte]) in lzw_dictionary:          # If this combination exists in the dictionary...
            current_block.append(next_byte)                                  # Add it to the current block
        else:                                                                  # Otherwise...
            blocks.append(struct.pack('>B', len(lzw_dictionary[tuple(current_block)])))  # Add the corresponding value to the list of blocks
            blocks.extend(lzw_dictionary[tuple(current_block)])                # And extend the current block with its value
            
            # Now let's update the dictionary
            lzw_dictionary[tuple(current_block+[next_byte])] = len(lzw_dictionary)
            current_block = [next_byte]                                       # Reset the current block
                    
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)