#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = [struct.pack('>BBBB', *rng.read(4)) for _ in range(num_colors)]
    color_table = b''.join(global_color_table)

    # 4. Graphics Control Extension
    # Frame delay time (two bytes, using a random value between 0 and 65535)
    frame_delay_time = rng.read(2)

    # Transparent color index (one byte, using a random value between 0 and 255)
    transparent_color_index = rng.read(1)[0]

    # Block terminator (always \x00)
    block_terminator = b'\x00'

    # Put all together
    gce = frame_delay_time + transparent_color_index + block_terminator

    # 5. Image Descriptor
    # Left position (two bytes, using a random value between 0 and 65535)
    left = rng.read(2)

    # Top position (two bytes, using a random value between 0 and 65535)
    top = rng.read(2)

    # Width (two bytes, using a random value between 0 and 65535)
    width = rng.read(2)

    # Height (two bytes, using a random value between 0 and 65535)
    height = rng.read(2)

    # Flags (packed field, using a random byte for variety)
    flags = rng.read(1)[0]

    # Image descriptor block terminator (always \x00)
    image_descriptor_block_terminator = b'\x00'

    # Put all together
    image_descriptor = left + top + width + height + bytes([flags]) + image_descriptor_block_terminator

    # 6. Image Data
    # LZW minimum code size (one byte, using a random value between 2 and 12)
    lzw_minimum_code_size = rng.read(1)[0] % 12 + 2

    # Blocks
    blocks = []
    while True:
        # Block size (one byte, using a random value between 0 and 255)
        block_size = rng.read(1)[0]
        if not block_size:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)