#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''
        
    # 5. Image Descriptor
    # Left position is always zero
    left = 0
    
    # Top position is always zero unless we're making a really big image
    top = 0
    if width > 256 or height > 256:
        top = rng.read(1)[0] & 0xFF
    
    # Width and Height are random again
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed field: Local color table flag, interlace flag, sort flag, reserved bits, local color table size
    packed = struct.pack('>BBBBB', bool(rng.read(1)[0]), bool(rng.read(1)[0]), bool(rng.read(1)[0]), 0b1111, rng.read(1)[0])
    
    # Write all of this stuff to the image descriptor
    image_descriptor = struct.pack('>HHBBBB', left, top, width, height, packed[0], packed[1])
    
    # 6. Image Data
    # We'll just write some random data here too
    image_data = rng.read((width * height) // 8)
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)