#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_present = bool(packed & 0b10000000)
    if not global_color_table_present:
        gct = bytes()
    else:
        size = 2 ** ((packed & 7) + 1)
        gct = bytearray([rng.read(1)[0] for _ in range(size)])

    # 4. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBBBBB', 0x2C, rng.read(1)[0], 0, 0, width, height, packed, 0, 0)

    # 5. Local Color Table (if present)
    local_color_table_present = bool(packed & 0b01000000)
    if not local_color_table_present:
        lct = bytes()
    else:
        size = 2 ** ((packed & 7) + 1)
        lct = bytearray([rng.read(1)[0] for _ in range(size)])

    # 6. Graphics Control Extension Block
    graphics_control_extension_block = struct.pack('>BHB', 0x21, 0xF9, 4) + rng.read(3)

    # 7. Plain Text Extension Block
    plain_text_extension_block = struct.pack('>BHB', 0x21, 0x01, 12) + rng.read(11)

    # 8. Application Extension Block
    application_extension_block = struct.pack('>BHB', 0x21, 0xFF, 11) + b'NETSCAPE2.0' + b'\x03\x01' + b'\x00\x00'

    # 9. Comment Extension Block
    comment_extension_block = struct.pack('>BHB', 0x21, 0xFE, 3) + b'!' * rng.read(rng.read(1)[0])[0]

    # 10. Trailer
    trailer = b';'

    # Write everything to output stream
    out.write(header)
    out.write(lsd)
    if len(gct):
        out.write(bytes(gct))
    out.write(image_descriptor)
    if len(lct):
        out.write(bytes(lct))
    out.write(graphics_control_extension_block)
    out.write(plain_text_extension_block)
    out.write(application_extension_block)
    out.write(comment_extension_block)
    out.write(trailer)

