#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        # Use global color table?
        use_global_color_table = bool(rng.read(1)[0])

        # Delay time in hundredths of seconds
        delay_time = struct.pack('>H', rng.read(2)[0])

        # Transparent color index
        transparent_index = rng.read(1)[0]

        # Block terminator
        block_terminator = b'\x00'

        # Put all GCE data together
        gce = b''
        if use_global_color_table:
            gce += b'\x21\xF9\x04'
        else:
            gce += b'\x21\xF9\x04\x01'
        gce += delay_time
        gce += transparent_index.to_bytes(1, 'big')
        gce += block_terminator
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))

    # Packed field (bit flags): Local color table flag, Interlace flag, Sort flag, Reserved bits
    packed_field = rng.read(1)[0]

    # LZW minimum code size
    lzw_min_code_size = rng.read(1)[0]

    # Put all ID data together
    image_descriptor = struct.pack('>HHBBBBB', left, top, width, height, packed_field, lzw_min_code_size)

    # 6. Image Data
    # Let's just do some random pixels here...
    image_data = []
    while True:
        pixel = rng.read(1)[0]
        if not pixel: break
        image_data.append(pixel)

    # Add an end-of-image marker
    image_data.append(b'\x00')

    # Convert list of ints to bytes
    image_data = b''.join([i.to_bytes(1, 'big') for i in image_data])

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom as get_random_bytes
    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(get_random_bytes(), f)