#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical screen descriptor
    width = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), 'big') & 0b0001])
    bgColorIndex = bytes([int.from_bytes(rng.read(1), 'big')])
    pixelAspectRatio = bytes([int.from_bytes(rng.read(1), 'big')])
    logicalScreenDescriptor = \
            width + height + flags + bgColorIndex + pixelAspectRatio

    # 3. Global color table
    globalColorTableSize = bytes([int.from_bytes(rng.read(1), 'big')])
    globalColorTable = bytearray()
    for _ in range((2**(globalColorTableSize[0] + 1))):
        globalColorTable += bytes([int.from_bytes(rng.read(1), 'big')])

    # 4. Image descriptor
    imageLeftPosition = bytes([int.from_bytes(rng.read(1), 'big')])
    imageTopPosition = bytes([int.from_bytes(rng.read(1), 'big')])
    imageWidth = bytes([int.from_bytes(rng.read(1), 'big')])
    imageHeight = bytes([int.from_bytes(rng.read(1), 'big')])
    flags = bytes([int.from_bytes(rng.read(1), 'big') & 0b0001])
    localColorTableFlag = bytes([int.from_bytes(rng.read(1), 'big') >> 7])
    interlaceFlag = bytes([(int.from_bytes(rng.read(1), 'big') & 0b0001)])
    sortFlag = bytes([(int.from_bytes(rng.read(1), 'big') & 0b0010) >> 1])
    reserved = bytes([int.from_bytes(rng.read(1), 'big') & 0b1100])
    lctSize = bytes([int.from_bytes(rng.read(1), 'big')]) if localColorTableFlag else b'\x00'
    imageDescriptor = \
            imageLeftPosition + imageTopPosition + \
            imageWidth + imageHeight + flags + \
            localColorTableFlag + interlaceFlag + \
            sortFlag + reserved + lctSize

    # 5. Local color table
    localColorTable = bytearray()
    for _ in range((2**(lctSize[0] + 1))):
        localColorTable += bytes([int.from_bytes(rng.read(1), 'big')])

    # 6. Data sub-blocks
    dataSubBlocks = bytearray()
    while True:
        blockType = rng.read(1)[0]

        if blockType == 0x2C:
            dataSubBlocks += blockType

            sizeOfLocalColorTable = rng.read(1)[0]
            dataSubBlocks += sizeOfLocalColorTable

            for i in range(sizeOfLocalColorTable):
                dataSubBlocks += bytes([int.from_bytes(rng.read(1), 'big')])

            LZWMinimumCodeSize = rng.read(1)[0]
            dataSubBlocks += LZWMinimumCodeSize

            compressedData = bytearray()
            while True:
                blockSize = rng.read(1)[0]

                if blockSize!= 0:
                    compressedData += blockSize

                    for i in range(blockSize):
                        compressedData += bytes([int.from_bytes(rng.read(1), 'big')])
                elif blockSize == 0:
                    break

            decompressedData = bytearray()
            codeLength = 1 << LZWMinimumCodeSize
            currentCode = 0
            previousCode = 0

            while len(decompressedData) < ((width * height)):
                bitString = format(compressedData[currentCode // 8], '#010b')[2:]
                bitString = ('0' * (codeLength - len(bitString))) + bitString

                if currentCode >= (len(compressedData) * 8):
                    raise Exception('Invalid compressed data.')

                if not currentCode and not bitString[-1]:
                    raise Exception('Invalid compressed data.')

                if currentCode:
                    decompressedData.append(previousCode)

                if bitString[-1]:
                    code = int(bitString[:-1], base=2)
                    decompressedData.append(code)

                    previousCode = code
                else:
                    code = int(bitString, base=