#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    header = b'GIF89a'

    # 1. Logical Screen Descriptor (LSD)
    width = int.from_bytes(rng.read(2), 'big') % 65536
    height = int.from_bytes(rng.read(2), 'big') % 65536
    flags = bytearray()
    flags += struct.pack('>b', bool(int.from_bytes(rng.read(1), 'big')))[0]  # Global Color Table Flag
    flags += struct.pack('>b', bool(int.from_bytes(rng.read(1), 'big')))[0]  # Color Resolution
    flags += struct.pack('>b', bool(int.from_bytes(rng.read(1), 'big')))[0]  # Sort Flag
    flags += struct.pack('>b', bool(int.from_bytes(rng.read(1), 'big')))[0]  # Size of Global Color Table
    flags += struct.pack('>b', bool(int.from_bytes(rng.read(1), 'big')))[0]  # Background Color Index Flag
    flags += struct.pack('>b', bool(int.from_bytes(rng.read(1), 'big')))[0]  # Pixel Aspect Ratio
    lsd = struct.pack('>HHBBB', width, height, *flags)

    # 2. Global Color Table
    if flags[4]:
        global_color_table = []
        for i in range(pow(2, flags[5])):
            r = int.from_bytes(rng.read(1), 'big') % 256
            g = int.from_bytes(rng.read(1), 'big') % 256
            b = int.from_bytes(rng.read(1), 'big') % 256
            global_color_table.append(struct.pack('BBB', r, g, b))
        global_color_table = b''.join(global_color_table)
    else:
        global_color_table = b''

    # 3. Graphics Control Extension (GCE)
    gce = b''

    # 4. Image Descriptor
    left = int.from_bytes(rng.read(2), 'big') % width
    top = int.from_bytes(rng.read(2), 'big') % height
    flags = bytearray()
    flags += struct.pack('>b', bool(int.from_bytes(rng.read(1), 'big')))[0]  # Local Color Table Flag
    flags += struct.pack('>b', bool(int.from_bytes(rng.read(1), 'big')))[0]  # Interlace Flag
    flags += struct.pack('>b', bool(int.from_bytes(rng.read(1), 'big')))[0]  # Sort Flag
    flags += struct.pack('>b', bool(int.from_bytes(rng.read(1), 'big')))[0]  # Reserved
    local_color_table_size = pow(2, flags[0])
    lzw_minimum_code_size = int.from_bytes(rng.read(1), 'big') % 256
    image_descriptor = struct.pack('>HHBBBBB', left, top, width, height, flags[1], flags[2], flags[3])

    # 5. LZW Minimum Code Size
    lzw_minimum_code_size = min(max(lzw_minimum_code_size, 2), 12)

    # 6. Image Data
    blocks = []
    while True:
        block_size = int.from_bytes(rng.read(1), 'big') % 256
        if not block_size: break
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)