#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = rng.read(2)
    h = rng.read(2)
    lsd = struct.pack('<6sHHBBB', header, int.from_bytes(w, 'big'), int.from_bytes(h, 'big'), 7, 0, 0)

    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 0b00011111   # Only use first 5 bits
    global_color_table_sort = bool((rng.read(1)[0] >> 7) & 1)     # Use bit 7 as sort flag
    global_color_table_bpp = rng.read(1)[0] & 0b00000111      # Only use last three bits
    global_color_table_entry_size = [
        3 * 8, 4 * 8, 8 * 8, 8 * 8, 16 * 8, 8 * 8, 24 * 8, 8 * 8][global_color_table_bpp]
    global_color_table_entries = rng.read(int(global_color_table_size / 3))
    color_table = b''
    for i in range(len(global_color_table_entries)):
        r = global_color_table_entries[i*3+0]
        g = global_color_table_entries[i*3+1]
        b = global_color_table_entries[i*3+2]
        color_table += struct.pack('BBB', r, g, b)

    # 4. Graphics Control Extension
    gce = struct.pack('<BHBHB', 0xF9, 4, 0, 0, 0)

    # 5. Image Descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    local_color_table_flag = bool((rng.read(1)[0] >> 7) & 1)       # Use bit 7 as local color table flag
    interlace_flag = bool((rng.read(1)[0] >> 6) & 1)               # Use bit 6 as interlace flag
    sorted_color_table_flag = bool((rng.read(1)[0] >> 5) & 1)     # Use bit 5 as sorted color table flag
    reserved = rng.read(1)[0]                                     # Ignored by most software
    lzw_minimum_code_size = rng.read(1)[0]                        # LZW minimum code size
    image_descriptor = struct.pack('<6sHHBBBBB', header,
                                   int.from_bytes(left, 'big'),
                                   int.from_bytes(top, 'big'),
                                   int.from_bytes(width, 'big'),
                                   int.from_bytes(height, 'big'),
                                   0,
                                   lzw_minimum_code_size)

    # 6. Image Data
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)

