#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed = rng.read(1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)
    lsd = struct.pack('>HHBB', int.from_bytes(width, byteorder='big'),
                      int.from_bytes(height, byteorder='big'),
                      packed[0],
                      int.from_bytes(background_color_index,
                                     byteorder='big')) + pixel_aspect_ratio
    
    # 3. Global Color Table
    color_table_size = rng.read(1)[0]
    global_color_table = [rng.read(3) for _ in range(int(pow(2, color_table_size)))]
    color_table = bytes([(len(global_color_table) - 1).to_bytes(1, byteorder='big')[0]]) \
                  + b''.join(global_color_table)
    
    # 4. Graphics Control Extension
    gce = b''
    
    # 5. Image Descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    packed = rng.read(1)
    local_color_table_flag = packed[0] >> 7
    interlace_flag = bool((packed[0] & 0b10000000) >> 6)
    sort_flag = bool((packed[0] & 0b01000000) >> 5)
    reserved = packed[0] & 0b00111000
    local_color_table_size = packed[0] & 0b00000111
    local_color_table = [rng.read(3) for _ in range(int(pow(2, local_color_table_size)))]
    image_descriptor = struct.pack('>HHHHBBB',
                                   int.from_bytes(left,
                                                  byteorder='big'),
                                   int.from_bytes(top,
                                                  byteorder='big'),
                                   int.from_bytes(width,
                                                  byteorder='big'),
                                   int.from_bytes(height,
                                                  byteorder='big'),
                                   local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved,
                                   local_color_table_size) + b''.join(local_color_table)
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        
        if not block_size:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    out.write(image_data)
    out.write(trailer)


if __name__ == '__main__':
    from sys import argv
    
    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)