#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = b'\x21\xF9\x04'
        
        # Other GCE fields: Packed fields (using a random byte for variety), Delay time, Transparent color index, Terminator
        gce += struct.pack('>BBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
        
    else:
        gce = b''
    
    # 5. Image Descriptor
    # Left position is always zero
    left = b'\x00'
    
    # Top position is always zero
    top = b'\x00'
    
    # Width and Height are random again
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed fields (using a random byte for variety): Local color table flag, Interlace flag, Sort flag, Reserved field, Size of local color table
    packed_fields = struct.pack('>BBBBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 6. Image data
    # Let's just use all black pixels
    image_data = b'\x00' * ((width * height) // 8)
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + left + top + width + height + packed_fields + image_data + trailer)


# Generates a random animated GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_animated_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = b'\x21\xF9\x04'
        
        # Other GCE fields: Packed fields (using a random byte for variety), Delay time, Transparent color index, Terminator
        gce += struct.pack('>BBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
        
    else:
        gce = b''
    
    # 5. Image Descriptor
    # Left position is always zero
    left = b'\x00'
    
    # Top position is always zero
    top = b'\x00'
    
    # Width and Height are random again
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed fields (using a random byte for variety): Local color table flag, Interlace flag, Sort flag, Reserved field, Size of local color table
    packed_fields = struct.pack('>BBBBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 6. Image data
    # Let's just use all black pixels
    image_data = b'\x00' * ((width * height) // 8)
    
    # 7. Image descriptor
    # Frame counter is always zero
    frame_counter = b'\x00'
    
    # Other ID fields: Left offset, Top offset, Width, Height, Packed fields
    image_descriptor = left + top + width + height + packed_fields
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)