#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        out.write(header + lsd)
        return

    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    graphics_control_extension_flag = bool(packed & 0x20)
    transparent_color_index = 0xFF
    delay_time = 0
    transparent_color_flag = False
    if graphics_control_extension_flag:
        packed = rng.read(1)[0]
        transparent_color_flag = bool(packed & 0x1)
        delay_time = rng.read(2)
        transparent_color_index = rng.read(1)[0]

        # Write GCE to output
        gce = bytes([
            0x21, 0xF9, 0x04,
            packed | 0x1C,
            delay_time[0], delay_time[1],
            0x00, 0x00,
            transparent_color_index,
            0x00
        ])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    local_color_table_flag = bool(packed & 0x80)
    interlace_flag = bool(packed & 0x40)
    sort_flag = bool(packed & 0x20)
    reserved_bits = rng.read(1)[0]
    size_of_local_color_table = 2 ** ((packed >> 4) + 1)
    local_color_table = rng.read(3 * size_of_local_color_table)
    image_descriptor = struct.pack('>HHBBB', left, top, packed, 0, 0)

    # 6. Data
    data = []
    for _ in range((height // 8) * width):
        byte = rng.read(1)[0]
        data += [bytes([(byte >> i) & 1]) for i in reversed(range(8))]

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)


if __name__ == '__main__':
    from sys import argv, stdout

    with open('/dev/urandom', 'rb') as f:
        generate_random_gif(f, stdout.buffer)