#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 256 + (rng.read(1)[0] & 7)
    height = 256 + ((rng.read(1)[0] >> 4) & 7)
    flags = b'\x80\x00' if rng.read(1)[0] & 1 else b'\x00\x00'
    bg_color = bytes([rng.read(1)[0]]) * 3
    aspect_ratio = b'\x00'
    out.write(struct.pack('<HHBBB', width, height, flags[0], flags[1], bg_color[0], bg_color[1]))
    out.write(aspect_ratio)

    # Global color table
    num_colors = min((width // 8) * (height // 8), 256)
    for i in range(num_colors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(bytes([r, g, b]))
    
    # Image descriptor
    left = 0
    top = 0
    width -= left
    height -= top
    flags = b'\x80\x00'
    local_color_table_flag = b'\x00'
    interlace_flag = b'\x00'
    sort_flag = b'\x00'
    reserved = b'\x00\x00'
    lct_size = b'\x00'
    out.write(struct.pack('<HHHHBBBBB', left, top, width, height, flags[0], flags[1], local_color_table_flag[0], local_color_table_flag[1], interlace_flag[0], interlace_flag[1], sort_flag[0], sort_flag[1]))
    out.write(reserved)
    out.write(lct_size)

    # Local color table
    for i in range(num_colors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(bytes([r, g, b]))

    # LZW minimum code size
    min_code_size = rng.read(1)[0]
    out.write(min_code_size.to_bytes(1, byteorder='little'))

    # Data sub-blocks
    data = []
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x00 or len(data) >= 2**min_code_size - 1:
            break
        
        elif block_type == 0x01:
            pass

        elif block_type == 0xfe:
            data += [block_type]
            continue

        elif block_type == 0xff:
            data += [block_type]
            continue

        elif block_type < 0x02:
            raise Exception('Invalid block type')

        elif block_type > 0x02 and block_type <= 0x0b:
            length = block_type - 2
            data += rng.read(length)

        elif block_type == 0x0c:
            length = int.from_bytes(rng.read(1), 'big')
            data += rng.read(length)

        elif block_type == 0x0d:
            length = int.from_bytes(rng.read(2), 'big')
            data += rng.read(length)

        elif block_type == 0x0e:
            length = int.from_bytes(rng.read(3), 'big')
            data += rng.read(length)

        elif block_type == 0x0f:
            length = int.from_bytes(rng.read(4), 'big')
            data += rng.read(length)

        else:
            raise Exception('Invalid block type')

    # Write data to output stream
    write_subblocks(out, data)

    # Trailer
    out.write(b';')


# Writes an array of integers as a series of subblocks
def write_subblocks(out: BinaryIO, data: list):
    while len(data)!= 0:
        length = min(len(data), 255)
        out.write(length.to_bytes(1, byteorder='little'))
        out.write(bytes(data[:length]))
        del data[:length