#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table
    global_color_table_flag = bool(rng.read(1)[0] & 0b10000000)   # GCTF in first byte
    if not global_color_table_flag: return                  # No need to continue if there is no color table
    
    # Number of colors in the palette (including transparency)
    num_colors = int((rng.read(1)[0] & ~0b10000000) / 3)      # NCL in first byte
    
    # Actual color table data
    color_table = b''
    while len(color_table) < num_colors * 3:                 
        color_table += rng.read(min(num_colors - len(color_table) // 3, 256))

    # 4. Graphics Control Extension
    gce = b''
    if rng.read(1)[0]:                                        # If transparent color flag set...
        gce = b'\x21\xF9\x04'                                 #...then add graphics control extension
        gce += struct.pack('>BBBHBB',                        #...with delay time, transparent color index,
                            rng.read(1)[0],                   #     terminator, disposal method, and user input flag
                            rng.read(1)[0],
                            rng.read(1)[0],
                            rng.read(1)[0],
                            rng.read(1)[0])
    
    # 5. Image Descriptor
    image_descriptor = b'\x2C'                                # Start of image descriptor
    image_descriptor += struct.pack('>HHB',                    # Width and height
                                    width, 
                                    height,
                                    0)                       # No local color table
    
    # 6. Image Data
    # Minimum code size for LZW compression
    lzw_minimum_code_size = min(rng.read(1)[0] + 1, 12)       # LZWMC in first byte
    
    # Blocks of compressed data
    blocks = []
    while True:
        block_size = rng.read(1)[0]                           # Block size
        if block_size == 0: break                          
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    out.write(image_data)
    out.write(trailer)


if __name__ == '__main__':
    from sys import argv
    
    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)