#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = rng.read(1)[0] & 7
    if global_color_table_size == 0:
        global_color_table_size = 1
    global_color_table = rng.read(global_color_table_size * 3)
    
    # 4. Application Extension
    # We will use this to store our flag in plaintext
    app_ext_data = b'\x00\x01\xff\x0bflag{hacktheplanet}\x00'
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2c, 0, 0, width, height, 0, 0)
    
    # 6. Local Color Table (same as global color table)
    local_color_table = global_color_table
    
    # 7. Image Data
    # We will just fill with zeros since we don't care about the actual pixels
    image_data = b'\x00' * ((width + 1) // 2 * height)
    
    # Write everything together
    out.write(header)
    out.write(lsd)
    out.write(global_color_table)
    out.write(struct.pack('>BBH', 0x21, 0xFF, len(app_ext_data)))
    out.write(app_ext_data)
    out.write(image_descriptor)
    out.write(local_color_table)
    out.write(image_data)
    out.write(b';')
