#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = None
    if rng.read(1)[0] & 1 == 1:
        # Random delay time, transparent color index, disposal method
        gce = struct.pack('>BBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    id = b'\x2c\x00\x00\x00\x00\x00' + struct.pack('>HHB', width, height, rng.read(1)[0])

    # 6. Local Color Table (optional)
    lct = None
    if rng.read(1)[0] & 1 == 1:
        # Random size, random data
        num_colors = (rng.read(1)[0] % 256) or 1
        lct = rng.read(3 * num_colors)

    # 7. Image Data
    image_data = rng.read(width * height)

    # 8. Trailer
    trailer = b';'

    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(id)
    if lct is not None:
        out.write(lct)
    out.write(image_data