#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical Screen Descriptor
    width = rng.read(1)[0] + 1
    height = rng.read(1)[0] + 1
    flags = rng.read(1)[0]
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('<HHBBB', width, height, flags, background, aspect_ratio))

    # 3. Global Color Table
    global_color_table_size = (flags & 0b111) + 1
    global_color_table = [rng.read(3) for _ in range(global_color_table_size)]
    out.write(b''.join(global_color_table))

    # 4. Image Data
    while True:
        # 4.1 Graphic Control Extension
        if rng.read(1)[0]:
            block_size = rng.read(1)[0]
            assert block_size == 0xF4
            flags = rng.read(1)[0]
            delay = rng.read(2)
            transparent_index = rng.read(1)[0]
            terminator = rng.read(1)[0]
            assert terminator == 0
            out.write(struct.pack('<BBBBB', block_size, flags, *delay, transparent_index))

        # 4.2 Image Descriptor
        image_descriptor = rng.read(10)
        out.write(image_descriptor)

        # 4.3 Local Color Table
        local_color_table_size = (image_descriptor[9] & 0b111) + 1
        local_color_table = [rng.read(3) for _ in range(local_color_table_size)]
        out.write(b''.join(local_color_table))

        # 4.4 LZW-compressed image data
        lzw_minimum_code_size = rng.read(1)[0]
        out.write(lzw_minimum_code_size)

        while True:
            code = rng.read(1)[0]
            if code == 0:
                break
            else:
                out.write(code)

        # 4.5 Block Terminator
        block_terminator = rng.read(1)[0]
        assert block_terminator == 0

        # 4.6 Trailer
        trailer = rng.read(1)[0]
        assert trailer == 0

        # 4.7 End of File
        end_of_file = rng.read(1)[0]
        assert end_of_file == 0x3B
