#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    packed_width_height = struct.pack('<H', width) + struct.pack('<H', height)
    
    # Packed field - global color table flag set, no sort flag set,
    # no color resolution specified, no global color table size specified
    packed_field = b'\x80\x00'
    
    # Background color index
    background_color_index = rng.read(1)
    
    # Pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)
    
    # Color table
    color_table = rng.read(768)
    
    # LSD
    lsd = packed_width_height + packed_field + background_color_index + pixel_aspect_ratio + color_table
    
    # 3. Graphic Control Extension
    # Block size
    gce_block_size = b'\x04'
    
    # Packed field - transparency flag set, disposal method specified as "do not dispose",
    # delay time specified as 0
    gce_packed_field = b'\xF9\x04'
    
    # Transparent color index
    transparent_color_index = rng.read(1)
    
    # Block terminator
    gce_block_terminator = b'\x00'
    
    # GCE
    gce = gce_block_size + gce_packed_field + transparent_color_index + gce_block_terminator
    
    # 4. Image Descriptor
    # Separator
    separator = b','
    
    # Left position
    left_position = struct.pack('<H', rng.read(2)[0])
    
    # Top position
    top_position = struct.pack('<H', rng.read(2)[0])
    
    # Width
    width = struct.pack('<H', width)
    
    # Height
    height = struct.pack('<H', height)
    
    # Packed field - local color table flag set, interlaced flag unset,
    # sorted flag unset, reserved bits zero, no local color table size specified
    packed_field = b'\x80\x00'
    
    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0]
    
    # ID
    id = rng.read(1)
    
    # Image descriptor
    image_descriptor = separator + left_position + top_position + width + height + packed_field + bytes([lzw_minimum_code_size]) + id
    
    # 5. Local Color Table
    if lzw_minimum_code_size == 0:
        # No local color table
        pass
    elif lzw_minimum_code_size <= 8:
        # 2 ** (lzw_minimum_code_size + 1) colors
        color_table = rng.read((2 ** (lzw_minimum_code_size + 1)) * 3)
    else:
        raise ValueError("Invalid LZW minimum code size")
    
    # 6. Image Data
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)