#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, 0, rng.read(1)[0])
    
    # 6. Local Color Table (optional)
    local_color_table = rng.read(3 * num_colors)  # Random colors
    
    # 7. Image Data
    # We will use LZW compression to compress the data
    lzw_minimum_code_size = rng.read(1)[0] & 0b111
    # We will use a dictionary of size 2**lzw_minimum_code_size
    # We will use a maximum code size of 12 bits
    max_code_size = min(12, lzw_minimum_code_size + 8)
    # We will use a clear code of size 12 bits
    clear_code = 1 << max_code_size
    # We will use an end-of-information code of size 12 bits
    eoi_code = clear_code + 1
    # We will use a maximum code of size 12 bits
    max_code = clear_code + 2
    # We will use a bit mask of size 12 bits
    bit_mask = (1 << max_code_size) - 1
    # We will use a prefix code of size 12 bits
    prefix_code = 0
    # We will use a suffix character of size 8 bits
    suffix_character = 0
    # We will use a current code of size 12 bits
    current_code = 0
    # We will use a stack of codes of size 12 bits
    stack = []
    # We will use a string of size 12 bits
    string = ''
    # We will use a table of codes of size 12 bits
    table = [chr(i) for i in range(clear_code)]
    # We will use a code size of size 4 bits
    code_size = lzw_minimum_code_size + 1
    # We will use a code of size 12 bits
    code = 0
    # We will use a flag of size 1 bit
    flag = False
    # We will use a block of data of size 256 bytes
    block = []
    # We will use a list of blocks of data of size 256 bytes
    blocks = []
    while True:
        # Read a new code from the stream
        code = struct.unpack('>H', rng.read(2))[0] & bit_mask
        
        # If we are not yet ready to write a block...
        if len(string) < 255:
            # ...and the code is the clear code...
            if code == clear_code:
                # ...then reset the code size and the table
                code_size = lzw_minimum_code_size + 1
                table = [chr(i) for i in range(clear_code)]
                
                # Also, clear the stack
                stack = []
            
            # Otherwise, if the code is the EOI code...
            elif code == eoi_code:
                # ...then break out of the loop
                break
            
            # Otherwise, if the code is too big...
            elif code > max_code:
                # ...then add the suffix character to the string
                string += chr(suffix_character)
            
            # Otherwise, if the code is in the table...
            elif code in table:
                # ...then add the associated string to the string
                string += table[code]
            
            # Otherwise, if the code is not in the table...
            else:
                # ...then add the associated string to the string
                string += table[prefix_code] + table[suffix_character]
            
            # Add the string to the stack
            stack.append(string)
            
            # If the code is not in the table...
            if code not in table:
                # ...then add the string to the table
                table.append(string)
            
            # Increment the code size if necessary
            if len(table) >= (1 << code_size):
                code_size += 1
            
            # Set the current code to be the next code
            current_code = code
        
        # If we are ready to write a block...
        if len(stack) > 0:
            # ...then pop the first string off the stack
            string = stack.pop()
            
            # ...and append its characters to the block
            for char in string:
                block.append(ord(char))
            
            # If the block is full...
            if len(block) >= 255:
                # ...then append the block to the list of blocks
                blocks.append(bytes(block))
                
                # And empty the block
                block = []
        
        # If we are ready to write a block...
        if len(block) > 0:
            # ...then append the block to the list of blocks
            blocks.append(bytes(block))
            
            # And empty the block
            block = []
    
    # Write the output
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if len(gce) == 6:
        out.write(gce)
    out.write(image_descriptor)
    out.write(struct.pack('B', lzw_minimum_code_size))
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
