#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 27 14:09:56 2019

@author: jlee
"""

import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
from astropy.io import fits
import matplotlib.pyplot as plt
import os


def fit_line(x, y):
    """
    Fit a line to the data using the least squares method.
    
    Parameters
    ----------
    x : array
        The independent variable.
        
    y : array
        The dependent variable.
        
    Returns
    -------
    popt : array
        Optimal values for the parameters of the function.
        
    pcov : array
        Covariance matrix of the parameter estimates.
        
    perr : array
        One standard deviation errors on the optimial parameters.
        
    chi2 : float
        Chi squared value of the fit.
        
    rchi2 : float
        Reduced chi squared value of the fit.
        
    func : function
        Function that returns the model.
    """
    
    # Define the function to be fitted
    def func(x, m, b):
        return m * x + b
    
    # Perform the fit
    popt, pcov = curve_fit(func, x, y)
    
    # Calculate the chi squared and reduced chi squared values
    chi2 = np.sum((y - func(x, *popt))**2 / (np.max(y) - np.min(y)))
    dof = len(x) - 2
    rchi2 = chi2 / dof
    
    # Calculate the one sigma errors on the parameters
    perr = np.sqrt(np.diag(pcov))
    
    return popt, pcov, perr, chi2, rchi2, func


def plot_fit(x, y, popt, perr, func, save=False, figsize=(10, 8), dpi=150,
             fname='', path=''):
    """
    Plot the data with the best fit line overlaid.
    
    Parameters
    ----------
    x : array
        The independent variable.
        
    y : array
        The dependent variable.
        
    popt : array
        Optimal values for the parameters of the function.
        
    perr : array
        One standard deviation errors on the optimial parameters.
        
    func : function
        Function that returns the model.
        
    save : bool
        Save the figure? Default is False.
        
    figsize : tuple
        Figure size in inches. Default is (10, 8).
        
    dpi : int
        Dots per inch. Default is 150.
        
    fname : str
        Filename for saving the figure. If not given, then no file will be saved.
        
    path : str
        Path to directory where the figure should be saved. If not given, then
        the current working directory will be used.
    """
    
    # Create the figure
    plt.figure(figsize=figsize)
    
    # Plot the data points
    plt.errorbar(x, y, fmt='o')
    
    # Plot the best fit line
    xfit = np.linspace(np.min(x), np.max(x), 100)
    yfit = func(xfit, *popt)
    plt.plot(xfit, yfit, 'k--')
    
    # Add labels
    plt.xlabel('Wavelength [$\AA$]')
    plt.ylabel('Flux')
    plt.title('Best fit line')
    
    # Show the figure
    if save:
        plt.savefig(os.path.join(path, fname+'.png'), dpi=dpi)
    else:
        plt.show()
    
    return None


def get_wl_range(wave, flux, thresh=0.01, npts=None):
    """
    Get the wavelength range where the flux exceeds some threshold.
    
    Parameters
    ----------
    wave : array
        Wavelength array.
        
    flux : array
        Flux array.
        
    thresh : float
        Threshold flux value.
        
    npts : int
        Number of points to use when fitting the line. If not given, then all
        points will be used.
        
    Returns
    -------
    wl_start : float
        Starting wavelength of the region.
        
    wl_end : float
        Ending wavelength of the region.
    """
    
    # Find the indices where the flux exceeds the threshold
    inds = np.where(flux > thresh)[0]
    
    # Use only the first `npts` points
    if npts is not None:
        inds = inds[:npts]
    
    # Fit the line to these points
    popt, pcov, perr, chi2, rchi2, func = fit_line(wave[inds], flux[inds])
    
    # Get the starting and ending wavelengths
    wl_start = wave[inds][0]
    wl_end = wave[inds][-1]
