#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed fields
    #   - Global Color Table Flag: 1
    #   - Color Resolution: 7
    #   - Sort Flag: 0
    #   - Size of Global Color Table: 7
    lsd = struct.pack('BBB', 0b11100000 | 7, 7, 0)
    
    # Background color index
    background_color_index = rng.read(1)[0] % 256
    
    # Pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)[0] % 256
    
    # Write everything to output
    out.write(header + lsd + background_color_index.to_bytes(1, 'big') + pixel_aspect_ratio.to_bytes(1, 'big'))
    
    # 3. Graphic Control Extension
    gce = b'\x21\xf9\x04\x00\x00\xff\x00'
    
    # Disposal Method
    disposal_method = rng.read(1)[0] % 8
    
    # User Input Flag
    user_input_flag = bool(rng.read(1)[0])
    
    # Transparent Color Flag
    transparent_color_flag = bool(rng.read(1)[0])
    
    # Delay Time
    delay_time = rng.read(2)
    
    # Transparent Color Index
    if transparent_color_flag:
        transparent_color_index = rng.read(1)[0] % 256
        
        # Write everything to output
        out.write(gce + disposal_method.to_bytes(1, 'big') + user_input_flag.to_bytes(1, 'big') + transparent_color_flag.to_bytes(1, 'big') + delay_time + transparent_color_index.to_bytes(1, 'big'))
    else:
        # Write everything to output
        out.write(gce + disposal_method.to_bytes(1, 'big') + user_input_flag.to_bytes(1, 'big') + transparent_color_flag.to_bytes(1, 'big') + delay_time)
    
    # 4. Image Descriptor
    id = b'\x2c\x00\x00\x00\x00\x00\x00\x00\x00'
    
    # Left Position
    left_position = rng.read(2)
    
    # Top Position
    top_position = rng.read(2)
    
    # Width
    image_width = rng.read(2)
    
    # Height
    image_height = rng.read(2)
    
    # Local Color Table Flag
    local_color_table_flag = bool(rng.read(1)[0])
    
    # Interlace Flag
    interlace_flag = bool(rng.read(1)[0])
    
    # Sort Flag
    sort_flag = bool(rng.read(1)[0])
    
    # Size of Local Color Table
    size_of_local_color_table = rng.read(1)[0] % 8
    
    # Write everything to output
    out.write(id + left_position + top_position + image_width + image_height + local_color_table_flag.to_bytes(1, 'big') + interlace_flag.to_bytes(1, 'big') + sort_flag.to_bytes(1, 'big') + size_of_local_color_table.to_bytes(1, 'big'))
    
    # 5. Local Color Table
    # We don't need this since we set the flag to 0
    
    # 6. Raster Data
    # We don't need this since we set the flag to 0
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to output
    out.write(trailer)
