#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    # Packed fields
    packed_fields = struct.pack('BBB', *[rng.read(1)[0] for _ in range(3)])
    # Background color index
    background_color_index = rng.read(1)[0]
    # Pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)[0]
    # Write everything together
    out.write(header + struct.pack('<HHBBB', width, height, *struct.unpack('BBB', packed_fields), background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table
    global_color_table_size = 7 + 1
    global_color_table = [struct.unpack('BBB', rng.read(3)) for _ in range(global_color_table_size)]
    out.write(b'\x00') # No local color table
    out.write(struct.pack('B', global_color_table_size - 1)) # Size of global color table
    for color in global_color_table:
        out.write(struct.pack('BBB', *color))

    # 4. Image Data
    # Loop until we hit an end-of-image block
    while True:
        # Read the next byte
        block_type = rng.read(1)[0]
        
        if block_type == 0x3B: # End of image
            break
            
        elif block_type == 0x2C: # Image descriptor
            # Image left position
            image_left_position = struct.unpack('>H', rng.read(2))[0]
            # Image top position
            image_top_position = struct.unpack('>H', rng.read(2))[0]
            # Width
            image_width = struct.unpack('>H', rng.read(2))[0]
            # Height
            image_height = struct.unpack('>H', rng.read(2))[0]
            # Packed fields
            packed_fields = struct.unpack('BBB', rng.read(3))
            # Local color table size
            local_color_table_size = struct.unpack('B', rng.read(1))[0]
            # Interlaced?
            interlaced = bool(packed_fields[0] & 0b10000000)
            # Sort flag
            sort_flag = bool(packed_fields[0] & 0b01000000)
            # Reserved
            reserved = bool(packed_fields[0] & 0b00100000)
            # Size of local color table
            size_of_local_color_table = bool(packed_fields[0] & 0b00011111)
            # LZW minimum code size
            lzw_minimum_code_size = packed_fields[1]
            # Write everything together
            out.write(struct.pack('<HHBBB', image_left_position, image_top_position, image_width, image_height, lzw_minimum_code_size))

            # If there is a local color table...
            if local_color_table_size > 0:
                # ...generate it
                local_color_table = [struct.unpack('BBB', rng.read(3)) for _ in range(local_color_table_size)]
                out.write(b'\x00') # No global color table
                out.write(struct.pack('B', local_color_table_size - 1)) # Size of local color table
                for color in local_color_table:
                    out.write(struct.pack('BBB', *color))

            # If this is not an interlaced image...
            if not interlaced:
                # ...just read the data blocks
                while True:
                    # Read the next byte
                    block_length = rng.read(1)[0]
                    
                    if block_length == 0: # End of image
                        break
                        
                    else: # Block of data
                        out.write(rng.read(block_length))

            # Otherwise, read the data blocks as per the spec
            else:
                pass
