#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Color Table
    # We use a fixed size of 256 colors (8 bits per channel)
    color_table = rng.read(768)
    
    # 4. Graphic Control Extension
    # We use a fixed duration of 0
    gce = struct.pack('>BBBHB', 0x21, 0xf9, 0x04, 0x00, rng.read(1)[0])
    
    # 5. Image Descriptor
    # We use a fixed left and top position of 0
    image_descriptor = struct.pack('>BBHHH', 0x2c, 0x00, 0x00, width, height)
    
    # 6. Data Sub-Blocks
    # We use a random number of blocks between 1 and 10
    num_blocks = rng.read(1)[0] % 10 + 1
    data = []
    for i in range(num_blocks):
        block_size = rng.read(1)[0] + 1        
        data.append(rng.read(block_size))
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)
