#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big') % 640 + 1
    height = int.from_bytes(rng.read(2), 'big') % 480 + 1
    background_color_index = int.from_bytes(rng.read(1), 'big') % 256
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big') % 8 + 1
    global_color_table_flag = bool(int.from_bytes(rng.read(1), 'big'))
    color_resolution = int.from_bytes(rng.read(1), 'big') % 8 + 1
    sort_flag = bool(int.from_bytes(rng.read(1), 'big'))
    size_of_global_color_table = (int.from_bytes(rng.read(1), 'big') % 8 + 1) * 3
    bg_color_index = int.from_bytes(rng.read(1), 'big') % 256
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big') % 8 + 1
    global_color_table = [
            rng.read(3) for _ in range(size_of_global_color_table // 3)
        ]

    screen_descriptor = struct.pack('<HBBB',
            width, height,
            background_color_index,
            0b10000000 | (pixel_aspect_ratio - 1))
    out.write(screen_descriptor)

    # 3. Global Color Table
    if global_color_table_flag:
        out.write(b''.join(global_color_table))

    # 4. Image Data
    while True:
        # 4.1 Graphic Control Extension
        block_type = int.from_bytes(rng.read(1), 'big')
        if block_type == 0xF9:
            block_length = int.from_bytes(rng.read(1), 'big')
            disposal_method = int.from_bytes(rng.read(1), 'big') & 7
            user_input_flag = bool(int.from_bytes(rng.read(1), 'big'))
            transparent_color_flag = bool(int.from_bytes(rng.read(1), 'big'))
            delay_time = int.from_bytes(rng.read(2), 'big')
            transparent_color_index = int.from_bytes(rng.read(1), 'big')
            block_terminator = int.from_bytes(rng.read(1), 'big')

            gce = struct.pack('<BBBHBB',
                    0xF9, block_length,
                    disposal_method << 2 | 0b11,
                    delay_time,
                    transparent_color_index,
                    0)
            out.write(gce)

        # 4.2 Image Descriptor
        image_left_position = int.from_bytes(rng.read(2), 'big')
        image_top_position = int.from_bytes(rng.read(2), 'big')
        image_width = int.from_bytes(rng.read(2), 'big')
        image_height = int.from_bytes(rng.read(2), 'big')
        local_color_table_flag = bool(int.from_bytes(rng.read(1), 'big'))
        interlace_flag = bool(int.from_bytes(rng.read(1), 'big'))
        sort_flag = bool(int.from_bytes(rng.read(1), 'big'))
        size_of_local_color_table = (int.from_bytes(rng.read(1), 'big') % 8 + 1) * 3
        lct = [rng.read(3) for _ in range(size_of_local_color_table // 3)]

        id = struct.pack('<HHBB',
                image_left_position, image_top_position,
                image_width, image_height)
        out.write(id)

        # 4.3 Local Color Table
        if local_color_table_flag:
            out.write(b''.join(lct))

        # 4.4 Image Data
        lzw_minimum_code_size = int.from_bytes(rng.read(1), 'big')
        out.write(struct.pack('<B', lzw_minimum_code_size))

        while True:
            code_size = int.from_bytes(rng.read(1), 'big')
            if code_size == 0:
                break
            else:
                data = rng.read(code_size)
                out.write(data)

        # 4.5 Trailer
        trailer = int.from_bytes(rng.read(1), 'big')
        assert trailer == 0
