#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = min(pow(2, rng.read(1)[0] + 1), 256) * 3
    global_color_table = rng.read(global_color_table_size)

    # 4. Image Data
    # We will use a single image with no local color table
    # We will use a single "graphic control extension" block
    # We will use a single plain text extension block
    # We will use a single application extension block

    # Graphic Control Extension Block
    gceb = struct.pack('>BBBHB', 0xF9, 0x04, 0x08 | rng.read(1)[0], rng.read(2), rng.read(1)[0])

    # Plain Text Extension Block
    pteb = struct.pack('>BBBHBBB', 0x21, 0xFF, 0x0B, rng.read(2), rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # Application Extension Block
    aeb = struct.pack('>BBBHBBB', 0x21, 0xFF, 0x0B, rng.read(2), rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # Image Data
    id = rng.read(width * height)

    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(global_color_table)
    out.write(gceb)
    out.write(pteb)
    out.write(aeb)
    out.write(id)

