#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Color Table
    # Random number of colors in the table (between 1 and 256)
    num_colors = rng.read(1)[0] % 256 + 1

    # Each color is 3 bytes long
    color_table = rng.read(num_colors * 3)

    # 4. Graphic Control Extension
    # Random delay time (between 0 and 65535)
    delay_time = struct.pack('>H', rng.read(2)[0])

    # Transparent color index, Block terminator
    gce = struct.pack('BB', rng.read(1)[0], 0)

    # 5. Image Descriptor
    # Random left and top coordinates (using two bytes for each)
    x, y = struct.unpack('>HH', rng.read(4))

    # Width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Packed field: Local color table flag, Interlace flag, Sort flag, Reserved
    packed_field = struct.pack('BBBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    image_descriptor = struct.pack('>HHHHBB', x, y, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)