#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = rng.read(1)[0] % 64 + 1
    h = rng.read(1)[0] % 64 + 1

    # Color table size is always 7
    color_table_size = 7

    # Global color table flag
    global_color_table_flag = True if rng.read(1)[0] & 1 else False

    # Background color index
    background_color_index = rng.read(1)[0] % (2 ** color_table_size)

    # Pixel aspect ratio
    pixel_aspect_ratio = 0

    # Write logical screen descriptor to output stream
    out.write(header)
    out.write(struct.pack('<HHBBB', w, h,
                          color_table_size, global_color_table_flag,
                          background_color_index))
    out.write(bytes([pixel_aspect_ratio]))

    # 3. Global color table
    # Generate RGB values for all colors in the palette
    rgb_values = [rng.read(3) for _ in range(2**color_table_size)]

    # Write global color table to output stream
    for rgb in rgb_values:
        out.write(rgb)

    # 4. Image descriptor
    # Image left position
    image_left_position = 0

    # Image top position
    image_top_position = 0

    # Image width
    image_width = w

    # Image height
    image_height = h

    # Local color table flag
    local_color_table_flag = False

    # Interlace flag
    interlace_flag = False

    # Sorting flag
    sorting_flag = False

    # Reserved flags
    reserved_flags = 0

    # LZW minimum code size
    lzw_minimum_code_size = 2

    # Write image descriptor to output stream
    out.write(struct.pack('<HHHHBB', image_left_position, image_top_position,
                          image_width, image_height,
                          local_color_table_flag, interlace_flag,
                          sorting_flag, reserved_flags))
    out.write(bytes([lzw_minimum_code_size]))

    # 5. Image data
    # Number of pixels in the image
    num_pixels = image_width * image_height

    # Get next byte from RNG as first bit of LZW code
    current_byte = rng.read(1)[0]

    # Initialize dictionary with single entry containing only the empty string
    dictionary = {b'' : 0}

    # Current LZW code length
    code_length = 1

    # While there are still pixels remaining...
    while len(dictionary) < 4096:

        # If we have reached maximum code length...
        if code_length == 12:
            # Reset it back to initial value
            code_length = 1

            # Add new entry to dictionary
            dictionary[current_byte.to_bytes(1, 'big')] = len(dictionary) - 1
        
        # Else, get next byte from RNG
        else:
            # Read next byte from RNG
            next_byte = rng.read(1)[0]
            
            # Concatenate current byte and next byte together
            combined_byte = current_byte << 8 | next_byte

            # If this combination has not been seen before...
            if combined_byte not in dictionary:
                # Add new entry to dictionary
                dictionary[combined_byte.to_bytes(2, 'big')] = \
                    len(dictionary) - 1

                # Increment code length by one
                code_length += 1
                
                # Set current byte to next byte
                current_byte = next_byte

            # Else, set current byte to existing entry in dictionary
            else:
                current_byte = combined_byte

        # If we have finished reading all pixels...
        if len(dictionary) >= num_pixels:
            # Break loop
            break

    # Dictionary size
    dict_size = len(dictionary)

    # Write LZW minimum code size to output stream
    out.write(bytes([dict_size]))

    # For every pixel in the image...
    for i in range(num_pixels):
        # Get next byte from RNG as first bit of LZW code
        current_byte = rng.read(1)[0]

        # While current byte is not an end-of-stream character...
        while current_byte!= 0x00:
            # Write current byte to output stream
            out.write(current_byte.to_bytes(1, 'big'))

            # If we have reached maximum code length...
            if code_length == 12:
                # Reset it back to initial value
                code_length = 1

                # Add new entry to dictionary
                dictionary[current_byte.to_bytes(1, 'big')] = len(dictionary) - 1
            
            # Else, get next byte from RNG
            else:
                # Read next byte from RNG
                next_byte = rng.read(1)[0]
                
                # Concatenate current byte and next byte together
                combined_byte = current_byte << 8 | next_byte

                # If this combination has not been seen before...
                if combined_byte not in dictionary:
                    # Add new entry to dictionary
                    dictionary[combined_byte.to_bytes(2, 'big')] = \
                        len(dictionary) - 1

                    # Increment code length by one
                    code_length += 1
                    
                    # Set current byte to next byte
                    current_byte = next_byte

                # Else, set current byte to existing entry in dictionary
                else:
                    current_byte = combined_byte

        # Write end-of-stream character to output stream
        out.write(bytes([0x00]))

    # Write end-of-image marker to output stream
    out.write(bytes([0x3B]))

