#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given parameters and writes it to disk.
def generate_random_gif(width, height, color_depth, background_index,
                        num_colors, interlace, sort, global_color_table,
                        local_color_table, lzw_minimum_code_size,
                        num_blocks, out):

    # Header
    header = b'GIF89a'

    # Logical Screen Descriptor
    lsd = bytearray()
    lsd += struct.pack('<H', width)
    lsd += struct.pack('<H', height)
    packed_flags = (0b1 << 7) | ((interlace & 0b1) << 6) | \
                   (((sort & 0b1) ^ 1) << 5) | (global_color_table << 4) | \
                   (color_depth - 1)
    lsd += struct.pack('<B', packed_flags)
    lsd += struct.pack('<B', background_index)
    lsd += struct.pack('<B', num_colors)

    # Global Color Table
    color_table = bytes([255] * (num_colors ** 2 * 3))

    # Graphics Control Extension
    gce = None
    if num_colors > 2**color_depth or not global_color_table:
        gce = bytearray()
        gce += b'\x21\xf9\x04'
        gce += struct.pack('<BBBBB', int((num_colors / 256)),
                           num_colors % 256, 0, 0, 0)

    # Image Descriptor
    image_descriptor = bytearray()
    image_descriptor += struct.pack('<HH', 0, 0)
    image_descriptor += struct.pack('<HB', left=0, top=0,
                                    width=width, height=height)
    packed_flags = 0b11100000 | (local_color_table << 7) | \
                   (not interlace << 6) | (sort << 5)
    image_descriptor += struct.pack('<B', packed_flags)

    # LZW minimum code size
    lzw_minimum_code_size = min(max(lzw_minimum_code_size, 2), 12)

    # Blocks of compressed data
    blocks = [bytes([i]) for i in range(num_blocks)]

    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if len(gce) == 6:
        out.write(gce)
    out.write(image_descriptor)
    out.write(struct.pack('B', lzw_minimum_code_size))
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
