#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    lsd = struct.pack('>HHBBB', width, height, 0xF0 | ((rng.read(1)[0] & 0b111) << 4), 0, 0)
    
    # 3. Global Color Table
    global_color_table_flag = bool(rng.read(1)[0] & 0b1)
    local_color_table_flag = bool(rng.read(1)[0] & 0b1)
    interlace_flag = bool(rng.read(1)[0] & 0b1)
    sort_flag = bool(rng.read(1)[0] & 0b1)
    reserved_flags = rng.read(3)

    # If there's no global color table, we don't care about any of these flags
    if global_color_table_flag:
        background_color_index = rng.read(1)[0]
        pixel_aspect_ratio = rng.read(1)[0]
        global_color_table_size = (1 << (pixel_aspect_ratio + 1)) * 3
        global_color_table = rng.read(global_color_table_size)
    else:
        background_color_index = None
        pixel_aspect_ratio = None
        global_color_table_size = None
        global_color_table = None

    # 4. Graphic Control Extension
    has_gce = bool(rng.read(1)[0] & 0b1)
    if has_gce:
        gce = struct.pack('>BBHBBBBB', 0x21, 0xF9, 4, 0b10000000 | (rng.read(1)[0] & 0b111), 0, 0, 0, 0)
    else:
        gce = None

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0] & 0b1)
    interlace_flag = bool(rng.read(1)[0] & 0b1)
    sort_flag = bool(rng.read(1)[0] & 0b1)
    reserved_flags = rng.read(3)

    # If there's no local color table, we don't care about any of these flags
    if local_color_table_flag:
        local_color_table_size = (1 << (rng.read(1)[0] & 0b111)) * 3
        local_color_table = rng.read(local_color_table_size)
    else:
        local_color_table_size = None
        local_color_table = None

    image_descriptor = struct.pack('>HHHHBBB', left, top, width, height, 0xF0 | ((rng.read(1)[0] & 0b111) << 4), 0, 0)
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)