#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]

    bg_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 16

    # Write LSD to output
    lsd = struct.pack('>HHBBB', width, height, packed_fields, bg_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    global_color_table_size = ((packed_fields & 7) + 1) * 3
    global_color_table = [rng.read(1)[0] % 256 for _ in range(global_color_table_size)]

    # Write GCT to output
    color_table = bytes()
    if len(global_color_table):
        color_table += bytes([global_color_table_size])
        color_table += bytes(global_color_table)

    # 4. Graphics Control Extension
    # Frame delay time (two bytes)
    frame_delay_time = rng.read(2)

    # Transparent color index (one byte)
    transparent_color_index = rng.read(1)[0] % 256

    # Write GCE to output
    gce = struct.pack('>BBH', 0x21, 0xf9, int.from_bytes(frame_delay_time, 'big')) \
         + bytes([transparent_color_index]) \
         + b'\x00'

    # 5. Image Descriptor
    left = 0
    top = 0
    width = 0
    height = 0

    local_color_table_flag = bool(rng.read(1)[0] % 2)
    interlace_flag = bool(rng.read(1)[0] % 2)
    sort_flag = bool(rng.read(1)[0] % 2)
    reserved = rng.read(2)

    # Local color table size (LCTS)
    local_color_table_size = None
    if local_color_table_flag:
        local_color_table_size = rng.read(1)[0] % 256

        # Write LCT to output
        color_table += bytes([local_color_table_size])
        color_table += bytes([rng.read(1)[0] % 256 for _ in range((local_color_table_size+1)*3)])

    # Write ID to output
    image_descriptor = struct.pack('>HHHHB', left, top, width, height, packed_fields | (local_color_table_flag << 7))

    # 6. Image Data
    # Number of pixels per row
    num_pixels_per_row = width - 1

    # LZW minimum code size
    min_code_size = rng.read(1)[0] % 256

    # First byte is LZW minimum code size minus one
    data = bytes([min_code_size-1])

    # Generate a random LZW code from 0 up until that point
    def get_lzw_code():
        return rng.read(1)[0] % (2**min_code_size)

    # Generate a random LZW code length
    def get_lzw_length():
        return rng.read(1)[0] % 256

    # Generate a random LZW code prefix
    def get_lzw_prefix():
        return rng.read(1)[0] % min_code_size

    # Generate a random LZW code suffix
    def get_lzw_suffix():
        return rng.read(1)[0] % 256

    # Generate an LZW code given its prefix and suffix
    def make_lzw_code(prefix, suffix):
        return (prefix << 8) | suffix

    # Generate a random LZW code with a given prefix
    def get_lzw_code_with_prefix(prefix):
        return make_lzw_code(prefix, get_lzw_suffix())

    # Generate a random LZW code with a given suffix
    def get_lzw_code_with_suffix(suffix):
        return make_lzw_code(get_lzw_prefix(), suffix)

    # Generate a random LZW code with a given prefix and suffix
    def get_lzw_code_with_prefix_and_suffix(prefix, suffix):
        return make_lzw_code(prefix, suffix)

    # Generate a random LZW code with a given prefix and length
    def get_lzw_code_with_prefix_and_length(prefix, length):
        return make_lzw_code(prefix, rng.read(1)[0] % (2**(length+1)))

    # Generate a random LZW code with a given length
    def get_lzw_code_with_length(length):
        return make_lzw_code(get_lzw_prefix(), rng.read(1)[0] % (2**(length+1)))

    # Generate a random LZW code with a given length
    def get_lzw_code_with_length(length):
        return make_lzw_code(get_lzw_prefix(), rng.read(1)[0] % (2**(length+1)))

    # Generate a random LZW code with a given length
    def get_lzw_code_with_length(length):
        return make_lzw_code(get_lzw_prefix(), rng.read(1)[0] % (2**(length+1)))

    # Generate a random LZW code with a given length
    def get_lzw_code_with_length(length):
        return make_lzw_code(get_lzw_prefix(), rng.read(1)[0] % (2**(length+1)))

    # Generate a random LZW code with a given length
    def get_lzw    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)

