#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    lsd = struct.pack('>HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    global_color_table_flag = bool(packed_fields & 0b10000000)
    if not global_color_table_flag: return False  # No point in continuing without this feature

    # Number of colors in GCT (1-256)
    num_colors = 2 ** ((packed_fields >> 7) & 0b111)

    # Size of each entry in GCT (3 or 6 bytes per entry)
    color_resolution = 2 ** (((packed_fields >> 4) & 0b11) + 1)

    # Sort flag (0=no sort, 1=sort)
    sort_flag = bool((packed_fields >> 3) & 0b1)

    # GCT is present, so read it now
    color_table = []
    for i in range(num_colors):
        r, g, b = [int(c * color_resolution) for c in struct.unpack('BBB', rng.read(3))]

        if sort_flag:
            color_table.append(struct.pack('<BBB', r, g, b))
        else:
            color_table.append(struct.pack('BBB', r, g, b))

    color_table = b''.join(color_table)

    # 4. Graphics Control Extension
    graphics_control_extension_flag = bool(packed_fields & 0b1000)
    if not graphics_control_extension_flag: return False  # No point in continuing without this feature

    # Block size (always 4)
    block_size = rng.read(1)[0]

    # Packed fields
    packed_fields = rng.read(1)[0]

    # Delay time (in hundredths of seconds)
    delay_time = int.from_bytes(rng.read(2), 'big')

    # Transparent color index
    transparent_color_index = rng.read(1)[0]

    # Terminator (always 0)
    terminator = rng.read(1)[0]

    gce = struct.pack('BBBBBB', block_size, packed_fields, delay_time, transparent_color_index, terminator)

    # 5. Image Descriptor
    left_position = 0
    top_position = 0
    width = width
    height = height
    local_color_table_flag = bool(packed_fields & 0b10000000)
    interlace_flag = bool(packed_fields & 0b100000)
    sorted_color_table_flag = bool(packed_fields & 0b10000)
    reserved_bits = rng.read(1)[0]
    lzw_minimum_code_size = rng.read(1)[0]

    image_descriptor = struct.pack('>HHHHBBB', left_position, top_position, width, height, local_color_table_flag,
                                   interlace_flag, sorted_color_table_flag, reserved_bits, lzw_minimum_code_size)

    # 6. Image Data
    # Read all data from the input stream until EOF
    blocks = []
    while True:
        try:
            block_size = rng.read(1)[0]
        except IndexError: break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)