#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while len(blocks) < total_blocks_size:
        block_size = min((total_blocks_size - len(blocks)), 255)

        # Generate a random block of data with a random LZW minimum code size
        block = [lzw_minimum_code_size | ((rng.read(1)[0] & 0x0F) << 8)]
        while len(block) <= block_size:
            if not block[-1]:
                block += [rng.read(1)[0]]

            elif block[-1] == 0xFF:
                block += [0x00, rng.read(1)[0]]

            else:
                block += [block[-1] + 1]

        blocks += block

    # Compress the generated data using LZW compression
    compressed_image_data = bytes([*blocks])

    # Append an EOI marker after the compressed data
    compressed_image_data += b'\x00\x00\xFF\x00'

    # Encode the compressed data in LZW format
    encoded_image_data = b''
    dictionary = {b'\x00': b'\x00'}
    current_byte = b'\x00'
    for i in range(len(compressed_image_data)):
        next_byte = compressed_image_data[i:i+1]

        if next_byte in dictionary:
            current_byte = dictionary[next_byte]

        else:
            encoded_image_data += current_byte
            dictionary[next_byte] = current_byte + b'\x00'
            current_byte = next_byte

    # Add the final byte to the encoded data
    encoded_image_data += current_byte

    # Pad the encoded data so that its length is divisible by 2
    padding_length = (-len(encoded_image_data)) % 2
    encoded_image_data += b'\x00' * padding_length

    # Convert the encoded data from binary to hex
    encoded_image_data = ''.join('%02X' % x for x in encoded_image_data).encode()

    # Finally, add the encoded data to the image descriptor
    image_data = image_descriptor + encoded_image_data

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)