#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed_flags = rng.read(1)[0]
    background_color_index = rng.read(1)[0] if packed_flags & 0b1 else None
    pixel_aspect_ratio = rng.read(1)[0] if packed_flags & 0b10 else None
    lsd = struct.pack('>HHBBB', width, height, packed_flags, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    global_color_table_flag = bool(packed_flags & 0b100)
    color_resolution = ((packed_flags >> 4) & 0b111) + 1
    sort_method = (packed_flags >> 7) & 0b1
    global_color_table_size = 2 ** color_resolution
    global_color_table = [rng.read(3) for i in range(global_color_table_size)]
    gct_packed_size = len(global_color_table) * 3 // 8
    gct_packed_size += 1 if gct_packed_size == 0 else 0
    gct_packed_size += 1 if gct_packed_size > 255 else 0
    color_table = bytes([(gct_packed_size << 4) | (sort_method << 3) | int(global_color_table_flag)]) \
                  + b''.join(global_color_table)
    
    # 4. Graphic Control Extension
    has_transparent_color = bool(rng.read(1)[0] & 0b1)
    transparent_color_index = rng.read(1)[0] if has_transparent_color else None
    delay_time = rng.read(2)
    disposal_method = rng.read(1)[0]
    user_input_flag = bool(rng.read(1)[0] & 0b1)
    gce = struct.pack('>BBHB', 0x21, 0xF9, 4, disposal_method) \
          + delay_time \
          + bytes([int(has_transparent_color) | (transparent_color_index << 1) | (user_input_flag << 2)])
        
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)

