#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
	width &= 0x7FFF
	height &= 0x7FFF

    # Random color table size
    color_table_size = ord(rng.read(1)) & 0x0F

    # Global Color Table Flag
    global_color_table_flag = bool(ord(rng.read(1)))

    # Background Color Index
    background_color_index = ord(rng.read(1))

    # Pixel Aspect Ratio
    pixel_aspect_ratio = ord(rng.read(1))

    screen_descriptor = struct.pack(
        '>HHBBB', width, height, color_table_size, 
        int(global_color_table_flag), background_color_index, pixel_aspect_ratio
    )

    # 3. Image Descriptor
    image_left = 0
    image_top = 0
    image_width = width
    image_height = height

    local_color_table_flag = False
    interlace_flag = True
    sort_flag = False
    reserved = 0
    lct_size = 0

    image_descriptor = struct.pack(
        '>HHBBBBBB', image_left, image_top, image_width, image_height,
        local_color_table_flag, interlace_flag, sort_flag, reserved, lct_size
    )

    # 4. Local Color Table
    if not global_color_table_flag:
        color_table = [
            rng.read(3) for _ in range(pow(2, color_table_size + 1))
        ]

        color_table += [(0, 0, 0)] * pow(2, 8 - color_table_size)

        color_table *= 65536 // len(color_table)

        color_table = bytearray().join(color_table[:65536])

        assert len(color_table) == 768

        out.write(header)
        out.write(screen_descriptor)
        out.write(image_descriptor)
        out.write(bytes(color_table))

    # 5. Graphics Control Extension
    transparent_color_index = None
    delay_time = 0
    transparent_color_flag = False
    block_size = 0
    terminator = 0

    graphics_control_extension = struct.pack(
        '>BHB?', B=block_size, T=transparent_color_index, F=transparent_color_flag,
        L=delay_time, N=terminator
    )

    # 6. Image Data
    data = []
    
    while True:
        block_size = rng.read(1)[0]
        
        if block_size == 0: break

        data.append(rng.read(block_size))

    data = b''.join(data)

    # 7. Trailer
    trailer = b''

    out.write(header)
    out.write(screen_descriptor)
    out.write(image_descriptor)

    if not global_color_table_flag:
        out.write(bytes(color_table))

    out.write(graphics_control_extension)
    out.write(data)
    out.write(trailer)