#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
        
    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 7 + 1 # Must be between 1-7 inclusive
    global_color_table = []
    if global_color_table_size > 0:
        global_color_table = [
            rng.read(3)
            for _ in range(global_color_table_size)
        ]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBB', 0, 0, 0b10000000 | rng.read(1)[0], 0, 0)
    
    # 5. Local Color Table
    local_color_table_size = rng.read(1)[0] & 7 + 1 # Must be between 1-7 inclusive
    local_color_table = []
    if local_color_table_size > 0:
        local_color_table = [
            rng.read(3)
            for _ in range(local_color_table_size)
        ]
    
    # 6. Image Data
    image_data = b''
    while True:
        block_type = rng.read(1)[0]
        
        if block_type == 0x2C: # Image descriptor
            break
            
        elif block_type == 0x21: # Extension
            extension_label = rng.read(1)[0]
            
            if extension_label == 0xF9:
                # Graphics Control Extension
                pass
                
            else:
                raise Exception("Unknown extension label")
                
        else:
            raise Exception("Unknown block type")
            
    # 7. Trailer
    trailer = b';'
    
    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(bytes([global_color_table_size]))
    for entry in global_color_table:
        out.write(entry)
    out.write(image_descriptor)
    out.write(bytes([local_color_table_size]))
    for entry in local_color_table:
        out.write(entry)
    out.write(image_data)
    out.write(trailer)
    
    