#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Color table size is always 0
    lsd = struct.pack('>HHBBB', width, height, 0, 0b10000000 | 0b00001111, 0)
    
    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 0b00001111  # Only use lower bits
    if global_color_table_size == 0:
        global_color_table_size = 2 ** ((rng.read(1)[0] & 0b00001111) + 1)
        
    global_color_table = [struct.pack('BBB', *rng.read(3)) for i in range(global_color_table_size)]
    while len(global_color_table) < 2 ** global_color_table_size:
        global_color_table += [(0, 0, 0)]  # Fill remaining slots with black
    color_table = b''.join(global_color_table)
    
    # 4. Graphics Control Extension
    # Disposal method is 0 (no disposal specified), user input required
    gce = struct.pack('>BBHBBBB', 0xF9, 4, 0, 0, 0, 0)
    
    # 5. Image Descriptor
    # Frame left position is 0, top position is 0, no local color table
    image_descriptor = struct.pack('>HHB', 0, 0, 0b10000000)
    
    # 6. Image Data
    # LZW minimum code size is 2, then data blocks are generated randomly
    lzw_minimum_code_size = rng.read(1)[0] & 0b00001111
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)

