#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = []
    for _ in range(num_colors):
        r, g, b = [rng.read(1)[0] for _ in range(3)]
        global_color_table.extend((r, g, b))
    
    # Make sure we have enough colors to fill up the entire palette
    assert len(global_color_table) == 768

    # Convert our list of RGB triplets back into bytes
    color_table = bytes(global_color_table)

    # 4. Graphics Control Extension
    # Delay time, transparent color index, block terminator
    gce = struct.pack('>HBB', rng.read(2)[0], rng.read(1)[0], 0)

    # 5. Image Descriptor
    # Left position, top position, width, height, packed fields
    image_descriptor = struct.pack('>HHHHB', rng.read(2)[0], rng.read(2)[0], width, height, rng.read(1)[0])

    # 6. Image Data
    # Minimum code size, then LZW encoded data
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)