#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    # Generate an LZW code table of minimum size `lzw_minimum_code_size`, with all possible codes
    lzw_code_table = [bytes([i]) for i in range(2**lzw_minimum_code_size)]

    # The first code is always the empty string
    current_string = b''

    # The next code will be the first character from the input stream
    next_char = rng.read(1)

    while True:
        if len(current_string) == 0:
            # If we're starting a new string, add its first character to the table
            lzw_code_table += [next_char]

            # Add the second character as well, since we know that one exists
            lzw_code_table += [rng.read(1)]

        elif next_char not in lzw_code_table:
            # If there are no matches between this string and the existing table,
            # write the entire string to the output data block
            out.write(struct.pack('<H', len(current_string)))
            out.write(current_string)

            # Add the new string to the table
            lzw_code_table += [current_string]

            # Start a new string with just the new character
            current_string = next_char

        else:
            # Otherwise, find the longest match between our string and the table
            max_match_length = -1
            matched_string = b''

            for s in lzw_code_table:
                if s[:len(current_string)] == current_string:
                    if len(s) > max_match_length:
                        max_match_length = len(s)
                        matched_string = s

            # Write the length of the match to the output data block
            out.write(struct.pack('<H', max_match_length))

            # Write the actual match to the output data block
            out.write(matched_string[len(current_string):])

            # Update the current string by removing what was used up by the match
            current_string = current_string[max_match_length:]

        # Read another character from the input stream
        try:
            next_char = rng.read(1)
        except Exception:
            break

    # Flush any remaining characters in the current string
    if len(current_string)!= 0:
        out.write(struct.pack('<H', len(current_string)))
        out.write(current_string)

    # End of File Block
    eofb = b'\x3B\x00'

    # Trailer
    trailer = b''.join((gce, image_descriptor, eofb))

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)