#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
	# 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big') % (50 + 1 - 16) + 16
    height = int.from_bytes(rng.read(2), 'big') % (50 + 1 - 16) + 16

    global_color_table_size = int.from_bytes(rng.read(1), 'big') & 7
    background_color_index = int.from_bytes(rng.read(1), 'big') % \
                             pow(2, global_color_table_size)
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big')
    logical_screen_descriptor = struct.pack('>HHBBB',
                                            width,
                                            height,
                                            global_color_table_size,
                                            background_color_index,
                                            pixel_aspect_ratio)
    # 3. Global Color Table
    global_color_table = bytearray()
    for i in range(pow(2, global_color_table_size)):
        r = int.from_bytes(rng.read(1), 'big') % 256
        g = int.from_bytes(rng.read(1), 'big') % 256
        b = int.from_bytes(rng.read(1), 'big') % 256

        global_color_table += bytes([r, g, b])
    # 4. Image Descriptor
    image_left = int.from_bytes(rng.read(2), 'big') % (width + 1)
    image_top = int.from_bytes(rng.read(2), 'big') % (height + 1)
    image_width = int.from_bytes(rng.read(2), 'big') % (width + 1 - image_left)
    image_height = int.from_bytes(rng.read(2), 'big') % (height + 1 - image_top)

    local_color_table_flag = bool(int.from_bytes(rng.read(1), 'big'))
    interlace_flag = bool(int.from_bytes(rng.read(1), 'big'))
    sort_flag = bool(int.from_bytes(rng.read(1), 'big'))
    reserved = int.from_bytes(rng.read(1), 'big')
    local_color_table_size = int.from_bytes(rng.read(1), 'big') & 7
    local_color_table = bytearray()
    if local_color_table_flag:
        for i in range(pow(2, local_color_table_size)):
            r = int.from_bytes(rng.read(1), 'big') % 256
            g = int.from_bytes(rng.read(1), 'big') % 256
            b = int.from_bytes(rng.read(1), 'big') % 256

            local_color_table += bytes([r, g, b])
    else:
        local_color_table = None
    lzw_min_code_size = int.from_bytes(rng.read(1), 'big')
    block_size = 2 ** lzw_min_code_size
    code_size = lzw_min_code_size + 1
    clear_code = 1 << lzw_min_code_size
    end_of_information_code = clear_code + 1
    next_code = clear_code + 2
    codes = [None] * (block_size + 2)
    last_code = None
    data = []
    while True:
        if len(data) == 0 or rng.read(1)[0] < 255 / 2:
            data.append(rng.read(1)[0])
        else:
            data.append(last_code[0][0])
        if len(data) >= code_size and not all((
                c is None for c in codes[:clear_code])):
            break
    while True:
        if any((c is None for c in codes[:next_code])):
            new_code = tuple(data[-code_size:])
            del data[-code_size:]
            if new_code!= ():
                if new_code in codes[:next_code]:
                    index = codes[:next_code].index(new_code)
                else:
                    index = next_code
                    codes[index] = new_code
                    next_code += 1
                if last_code is not None:
                    if last_code[1] is None:
                        last_code[1] = index
                    elif isinstance(last_code[1], list):
                        last_code[1].append(index)
                    else:
                        last_code[1] = [last_code[1]]
                        last_code[1].append(index)
                last_code = (new_code, index)
        else:
            if last_code[1] is None:
                raise Exception("Invalid LZW stream")
            elif isinstance(last_code[1], list):
                index = last_code[1][-1]
            else:
                index = last_code[1]
            last_code = (codes[index], codes[index])
        if last_code[0] == () or last_code[0] == (end_of_information_code,):
            break
    data = [d for d in data if d > 0]
    data.append(end_of_information_code)
    compressed_image_data = bytes(data)
    # 5. Trailer
    trailer = b'\x3B'

    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(global_color_table)
    out.write(struct