#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = b''

    if rng.read(1)[0]:  # If there is an extension
        # Use the same packed field as before, but now with a different value
        gce += struct.pack('>BBBBB', 0x21, 0xF9, rng.read(1)[0], 0x00, 0x00)

        # Use another random byte to decide whether we'll use the delay time or not
        if rng.read(1)[0]:
            # Delay time in hundredths of seconds
            gce += struct.pack('>H', rng.read(2)[0])

            # Transparent color index
            gce += struct.pack('>B', rng.read(1)[0])

            # Block terminator
            gce += b'\x00'

    # 5. Image Descriptor
    id = struct.pack('>HHBHB', 0, 0, 0, 0b10000000 | ((rng.read(1)[0] & 7) << 2), 0)

    # 6. Image Data
    image_data = b''

    # We need to know how many bytes are left in this frame so that we can properly pad the data
    remaining_bytes = width * height - len(gce + id)

    while remaining_bytes > 0:
        # Read up to 255 bytes from /dev/urandom
        chunk = rng.read((remaining_bytes // 255) + 1)[:min(remaining_bytes, 255)]

        # Add them to our image data
        image_data += chunk

        # Subtract those bytes from what remains
        remaining_bytes -= len(chunk)

    # Pad the image data with zeros until its length matches the logical screen descriptor's width times its height
    image_data += b'\x00' * (width * height - len(image_data))

    # 7. Trailer
    trailer = b';'

    # Write everything to `out`
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(id)
    out.write(image_data)
    out.write(trailer)

