#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    color_depth = rng.read(1)
    flags = rng.read(1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)
    lsd = struct.pack('>HHBBBHB', *width, *height, *color_depth, *flags, *background_color_index, *pixel_aspect_ratio)
    
    # 3. Global Color Table
    global_color_table_flag = bool(rng.read(1)[0])
    if not global_color_table_flag:
        return
    
    size = ((1 << color_depth[0]) - 1).to_bytes(1, 'big')
    color_table = [rng.read(3) for _ in range((1 << color_depth[0]))]
    gct = size + b''.join(color_table)
    
    # 4. Graphics Control Extension
    has_gce = bool(rng.read(1)[0])
    if not has_gce:
        return
    
    transparent_color_flag = bool(rng.read(1)[0])
    delay_time = rng.read(2)
    transparent_color_index = rng.read(1)
    terminator = rng.read(1)
    gce = struct.pack('>?BBHBB', has_gce, *delay_time, *transparent_color_index, *terminator)
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0] + 1)
        blocks.append(rng.read(block_size))
        
        total_blocks_size -= block_size
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if len(gce) == 6:
        out.write(gce)
    out.write(image_descriptor)
    out.write(struct.pack('B', lzw_minimum_code_size))
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
