#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = rng.read(1)[0] % 64 + 1
    h = rng.read(1)[0] % 64 + 1

    # Color table size (1 << n)
    ct_size = 2 ** ((rng.read(1)[0] & 7) + 1)

    # Background color index
    bg_color_index = rng.read(1)[0] % ct_size

    # Pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)[0] % 4

    # Build logical screen descriptor
    lsd = bytearray()
    lsd += struct.pack('<HHBBB', w, h, ct_size, bg_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    color_table = bytes([(rng.read(1)[0] % 256) for _ in range(ct_size * 3)])

    # 4. Graphic Control Extension
    gce = None
    if rng.read(1)[0] > 10:  # Probability of having one
        # Delay time
        delay_time = rng.read(2)

        # Transparent color index
        transparent_color_index = rng.read(1)[0] % ct_size
        
        # Block terminator
        gce = b'\xF9\x04' + delay_time + bytes([transparent_color_index]) + b'\x00'
        
    # 5. Image Descriptor
    # Left position
    left_pos = rng.read(1)[0] % (w - 10)

    # Top position
    top_pos = rng.read(1)[0] % (h - 10)

    # Width
    img_width = rng.read(1)[0] % (w - left_pos) + 10

    # Height
    img_height = rng.read(1)[0] % (h - top_pos) + 10

    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0] % 12 + 2

    # Local color table flag
    local_color_table_flag = bool(rng.read(1)[0] % 2)

    # Interlace flag
    interlace_flag = bool(rng.read(1)[0] % 2)

    # Sort flag
    sort_flag = bool(rng.read(1)[0] % 2)

    # Reserved bits
    reserved_bits = rng.read(1)[0] % 3

    # Build image descriptor
    image_descriptor = bytearray()
    image_descriptor += struct.pack('BBHBBBB', left_pos, top_pos, img_width, img_height,
                                   lzw_minimum_code_size, local_color_table_flag,
                                   interlace_flag, sort_flag, reserved_bits)

    # 6. Image data
    total_blocks_size = rng.read(1)[0] % 10000  # Total size of all blocks, max 10k
    blocks = []
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)