#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(num_colors)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBBBB', 0, 0, 0, 0b10000000 | rng.read(1)[0], 0, 0)

    # 5. Local Color Table (same as above)
    local_color_table = []
    if rng.read(1)[0]:
        num_colors = (rng.read(1)[0] % 256) or 1
        local_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(num_colors)]

    # 6. Graphics Control Extension
    graphics_control_extension = struct.pack('>BBBBBBH', 0x21, 0xF9, 4, ord('\x04'), 0, 0, 0)

    # 7. Application Extension
    application_extension = struct.pack('>BBBBB', 0x21, 0xFF, 11, ord('A') + rng.read(1)[0], ord('N') + rng.read(1)[0])

    # 8. Comment Extension
    comment_extension = struct.pack('>BBBBB', 0x21, 0xFE, 0, ord('C') + rng.read(1)[0], ord('M') + rng.read(1)[0])

    # 9. Plain Text Extension
    plain_text_extension = struct.pack('>BBBBB', 0x21, 0x01, 0, ord('P') + rng.read(1)[0], ord('T') + rng.read(1)[0])

    # 10. Trailer
    trailer = b';'

    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(bytes().join(global_color_table))
    out.write(image_descriptor)
    out.write(bytes().join(local_color_table))
    out.write(graphics_control_extension)
    out.write(application_extension)
    out.write(comment_extension)
    out.write(plain_text_extension)
    out.write(trailer)

