#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = rng.read(1)[0] % (65535 + 1)
    height = rng.read(1)[0] % (65535 + 1)

    flags = bytes([
        0b10000000 | # Global Color Table Flag
        0b01000000 | # Sort Flag
        0b00100000 | # Size of Global Color Table
        0b00011000 | # Bits per Pixel
        0b00000000    # Background Color Index
    ])

    bg_color = rng.read(1)[0] & 0x0F

    gct_size = rng.read(1)[0] & 0x07

    gct = bytearray()
    for i in range(gct_size):
        r = rng.read(1)[0] & 0xFF
        g = rng.read(1)[0] & 0xFF
        b = rng.read(1)[0] & 0xFF

        gct += struct.pack('BBB', r, g, b)

    lsd = struct.pack('<HHB', width, height, flags[0]) \
          + bytes([flags[1], flags[2]]) \
          + bytes([bg_color]) \
          + bytes([gct_size]) \
          + gct

    # 3. Image Descriptor
    img_left = rng.read(1)[0] % (width - 1 + 1)
    img_top = rng.read(1)[0] % (height - 1 + 1)
    img_width = rng.read(1)[0] % (width - img_left + 1)
    img_height = rng.read(1)[0] % (height - img_top + 1)

    flags = bytes([
        0b10000000 | # Local Color Table Flag
        0b01000000 | # Interlace Flag
        0b00100000 | # Sort Flag
        0b00010000 | # Reserved
        0b00001000 | # Size of Local Color Table
        0b00000111   # LZW Minimum Code Size
    ])

    lzw_min_code_size = rng.read(1)[0] & 0x07

    lct = bytearray()
    if flags[4]:
        lct_size = rng.read(1)[0] & 0x07

        for i in range(lct_size):
            r = rng.read(1)[0] & 0xFF
            g = rng.read(1)[0] & 0xFF
            b = rng.read(1)[0] & 0xFF

            lct += struct.pack('BBB', r, g, b)

    id = struct.pack('<HHHHBB',
                     img_left, img_top,
                     img_width, img_height,
                     flags[0], flags[1]) \
        + bytes([flags[2], flags[3], flags[4], flags[5]]) \
        + bytes([lzw_min_code_size]) \
        + lct

    # 4. Graphics Control Extension
    block_size = rng.read(1)[0] & 0x0F
    packed = rng.read(1)[0] & 0x01

    delay = rng.read(1)[0] % (255 + 1)
    transparent_idx = rng.read(1)[0] & 0x0F

    terminator = rng.read(1)[0] & 0x00

    gce = struct.pack('BBBBBBB',
                      block_size, packed,
                      delay,
                      0, 0, 0,
                      transparent_idx) \
        + bytes([terminator])

    # 5. Image Data
    data = bytearray()
    while True:
        code_len = rng.read(1)[0] & 0x0F
        
        if not code_len: break

        if code_len == 15: continue

        if code_len > 8:
            extra_bytes = rng.read(1)[0] & 0x07
            
            if extra_bytes!= 0:
                print("Warning: Invalid LZW minimum code size")

            code_len += extra_bytes

        data += bytes([code_len])

        for i in range(code_len):
            data += bytes([rng.read(1)[0] & 0xFF])

    # 6. Trailer
    trailer = b'\x3B'

    # Write to output stream
    out.write(header)
    out.write(lsd)
    out.write(id)
    out.write(gce)
    out.write(data)
    out.write(trailer)

