#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None
        
    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHBB', 0x2C, 0, 0, 0, 0)

    # 6. Image Data
    data = []
    while True:
        # Get a random byte to determine what to do next
        choice = rng.read(1)[0]
        
        # End of image?
        if not choice & 1:
            break
            
        # Do we need to add a frame delay?
        elif not choice & 2:
            # Add a frame delay
            data += [b'\x21\xFF\x0B'] + list(rng.read(3))
            
        # Do we need to add a graphics control extension?
        elif not choice & 4 and gce is not None:
            # Add a GCE
            data += [gce]
            
        # Do we need to add a local color table?
        elif not choice & 8:
            # Add a local color table with random size
            num_colors = (rng.read(1)[0] % 256) or 1
            data += [struct.pack('>B', 0x2C)] + list(rng.read(num_colors * 3))
            
        # Do we need to add raw data?
        elif not choice & 16:
            # Add some raw data
            length = rng.read(1)[0]
            data += [length] + list(rng.read(length))
            
    # Combine all our parts together
    gif = header + lsd + color_table + b''.join(data)
    
    # Write everything to output
    out.write(gif)