#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed_flags = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBB', width, height, packed_flags, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table (optional)
    global_color_table_flag = bool(packed_flags & 0b10000000)
    if not global_color_table_flag:
        return
    
    color_resolution = (packed_flags >> 7) & 0b111
    sort_flag = bool((packed_flags >> 6) & 0b1)
    size_of_global_color_table = ((packed_flags >> 3) & 0b111) + 1
    global_color_table_sort_order = [i for i in range(pow(2, size_of_global_color_table))]
    if sort_flag:
        global_color_table_sort_order = sorted(
            global_color_table_sort_order, key=lambda _: rng.read(1)[0]
        )
    global_color_table_entries = [(rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]) for _ in range(pow(2, size_of_global_color_table))]
    color_table = bytes([(r << 7 | g << 2 | b >> 3) for r,g,b in map(lambda x: tuple(map(lambda y: y // 8, x)), global_color_table_entries)])
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)

