#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    
    # Other ID fields: Packed fields (using a random byte for variety), Local color table flag, Interlace flag, Sort flag, Size of local color table
    id = struct.pack('>BBBBBH', 0x2C, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], num_colors)
    
    # 6. Image Data
    # Let's use the same algorithm as in the example code
    data = []
    for _ in range((width // 7) + 1):
        row = [0] * ((height // 7) + 1)
        
        for i in range(len(row)):
            row[i] = rng.read(1)[0] & 0b1111111
        
        data += row
    
    # Pack the data into blocks of 8 pixels
    packed_data = []
    for y in range(0, len(data), 8):
        block = 0
        for x in reversed(range(8)):
            block <<= 1
            block |= data[y + x] >> 7
            
        packed_data.append(block)
        
    # Convert the list of numbers into bytes
    image_data = bytes(packed_data)
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Combine all parts together
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)