#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 7 + 1 # Must be between 1-7 inclusive
    global_color_table = []
    
    if global_color_table_size > 0:
        for i in range(global_color_table_size):
            r, g, b = [rng.read(1)[0] for _ in range(3)]
            global_color_table += [r, g, b]

        # Pad to nearest multiple of 3
        while len(global_color_table) % 3!= 0:
            global_color_table += [0, 0, 0]
        
        # Write it all at once
        out.write(b''.join([struct.pack('BBB', *rgb) for rgb in global_color_table]))

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBBBB', 0, 0, 0, 0, 0, 0)

    # 5. Image Data
    image_data = []
    num_colors = 2 ** ((lsd[7] >> 3) + 1)
    transparent_index = None

    while True:
        # Get next block size
        block_size = rng.read(1)[0]

        # If we're done with this frame, write the data and exit
        if block_size == 0x00: break

        # Otherwise, check what kind of block it is
        elif block_size >= 0xF0:
            # Extension blocks are ignored
            pass

        else:
            # Set up local color table
            local_color_table_size = rng.read(1)[0] & 7 + 1

            # Check if there's any transparency information here
            has_transparency = bool((block_size >> 5) & 1)
            
            # Read the local color table
            local_color_table = []
            for i in range(local_color_table_size):
                r, g, b = [rng.read(1)[0] for _ in range(3)]
                local_color_table += [r, g, b]

            # Pad to nearest multiple of 3
            while len(local_color_table) % 3!= 0:
                local_color_table += [0, 0, 0]

            # Write it all at once
            out.write(b''.join([struct.pack('BBB', *rgb) for rgb in local_color_table]))

            # Now read the actual image data
            num_pixels = block_size
            pixel_data = []
            while num_pixels > 0:
                # Get the packed field
                packed_field = rng.read(1)[0]

                # Get the color index from the lower bits
                color_index = packed_field & 0x0F
                
                # Check if this pixel is actually transparent
                if has_transparency and color_index == 0:
                    # If so, set the transparent flag on the previous one
                    assert not transparent_index is None
                    pixel_data[-1] |= 0x01

                    # And continue without adding another entry to the list
                    continue

                # Add the color index to our list
                pixel_data += [color_index]

                # Decrement the counter
                num_pixels -= 1

            # Finally, write the pixel data
            out.write(bytes(pixel_data))

    # 6. Trailer
    trailer = b'\x3B'

    # Write everything together
    out.write(header + lsd + image_descriptor + b''.join(image_data) + trailer)