#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)
    out.write(lsd)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = [rng.read(3) for _ in range(num_colors)]  # Random colors
    color_table_data = b''.join([bytes((r, g, b)) for r, g, b in color_table])
    out.write(color_table_data)

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0] & 1)  # Random boolean
    if not has_gce:
        return

    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    out.write(gce)

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0] & 1)
    interlace_flag = bool(rng.read(1)[0] & 1)
    sort_flag = bool(rng.read(1)[0] & 1)
    reserved_flags = rng.read(1)[0]
    lzw_minimum_code_size = rng.read(1)[0]
    image_descriptor = struct.pack('>BHHHHBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  lzw_minimum_code_size)
    out.write(image_descriptor)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0] + 1)
        blocks.append(rng.read(block_size))

        total_blocks_size -= block_size

    # 7. Trailer
    trailer = b'\x3b'
    out.write(trailer)