#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')
    
    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    flags = rng.read(1)
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    out.write(width + height + flags + bytes([background]) + bytes([aspect_ratio]))
    
    # 3. Global Color Table
    color_table_size = (flags[0] & 0x7) + 1
    for _ in range(color_table_size):
        red = rng.read(1)[0]
        green = rng.read(1)[0]
        blue = rng.read(1)[0]
        out.write(bytes([red, green, blue]))
        
    # 4. Image Data
    while True:
        # 4.1 Graphic Control Extension
        if rng.read(1)[0] == 0x21:
            block_size = rng.read(1)[0]
            assert block_size == 0xF
            
            extension_code = rng.read(1)[0]
            assert extension_code == 0xF9
            
            flags = rng.read(1)[0]
            delay = rng.read(2)
            transparent_index = rng.read(1)[0]
            terminator = rng.read(1)[0]
            assert terminator == 0x0
            
            out.write(bytes([block_size, extension_code, flags, *delay, transparent_index, terminator]))
        
        # 4.2 Image Descriptor
        else:
            image_separator = rng.read(1)[0]
            assert image_separator == 0x2C
            
            left = rng.read(2)
            top = rng.read(2)
            width = rng.read(2)
            height = rng.read(2)
            flags = rng.read(1)
            local_color_table_size = (flags[0] & 0x7) + 1
            interlaced = bool(flags[0] & 0x40)
            sort = bool(flags[0] & 0x8)
            reserved = bool(flags[0] & 0x2)
            lct = bool(flags[0] & 0x1)
            assert not reserved and not lct
            
            out.write(image_separator + left + top + width + height + flags)
            
            # 4.2.1 Local Color Table
            for _ in range(local_color_table_size):
                red = rng.read(1)[0]
                green = rng.read(1)[0]
                blue = rng.read(1)[0]
                out.write(bytes([red, green, blue]))
                
            # 4.2.2 LZW-compressed image data
            while True:
                code_size = rng.read(1)[0]
                if code_size == 0:
                    break
                
                clear_code = 1 << code_size
                end_of_information_code = clear_code + 1
                
                code = rng.read(1)[0]
                while code != end_of_information_code:
                    if code < clear_code:
                        out.write(struct.pack('B', code))
                    elif code == clear_code:
                        code_size = 9
                    else:
                        raise Exception()
                    
                    code = rng.read(1)[0]
                
                # 4.2.3 Block Terminator
                block_terminator = rng.read(1)[0]
                assert block_terminator == 0x0
