#!/usr/bin/env python3

import struct
from typing import BinaryIO


def convert(inp: BinaryIO, out: BinaryIO):
    """Convert a GIF to an ICO"""
    rng = inp.read

    # 1. Header
    header = rng(6)
    assert header == b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng(4))
    global_color_table_flag = bool((rng(1)[0] & 1 << 7) >> 7)
    color_resolution = rng(1)[0] & 0x70 >> 4
    sort_flag = bool((rng(1)[0] & 1 << 3) >> 3)
    global_color_table_size = rng(1)[0] & 0x07
    background_color_index = rng(1)[0]
    pixel_aspect_ratio = rng(1)[0]
    if global_color_table_flag:
        raise NotImplementedError()
    if color_resolution != 0:
        raise NotImplementedError()
    if sort_flag:
        raise NotImplementedError()
    if global_color_table_size != 0:
        raise NotImplementedError()
    if background_color_index != 0:
        raise NotImplementedError()
    if pixel_aspect_ratio != 0:
        raise NotImplementedError()

    # 3. Global Color Table
    if global_color_table_flag:
        raise NotImplementedError()

    # 4. Blocks
    for _ in range(0, 2**global_color_table_size):
        rng(3)

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool((rng.read(1)[0] & 1 << 7) >> 7)
    interlace_flag = bool((rng.read(1)[0] & 1 << 6) >> 6)
    sort_flag = bool((rng.read(1)[0] & 1 << 5) >> 5)
    reserved_bits = rng.read(1)[0] & 0x07
    out.write(struct.pack('>BHHHHB', 0x2c, left, top, width, height, local_color_table_flag))
    if local_color_table_flag:
        raise NotImplementedError()
    if interlace_flag:
        raise NotImplementedError()
    if sort_flag:
        raise NotImplementedError()
    if reserved_bits!= 0:
        raise ValueError("Reserved bits must be zero")

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0] + 1)
        blocks.append(rng.read(block_size))

        total_blocks_size -= block_size

    # 7. Trailer
    out.write(b'\x3b')