#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed fields
    #   - Global Color Table Flag (1 bit)
    #   - Color Resolution (3 bits)
    #   - Sort Flag (1 bit)
    #   - Size of Global Color Table (3 bits)
    packed_fields = struct.pack('B', 0b10000000 | 0b00100000 | 0b00000000 | 0b00000000)
    
    # Background Color Index
    background_color_index = rng.read(1)[0] % 256
    
    # Pixel Aspect Ratio
    pixel_aspect_ratio = rng.read(1)[0] % 256
    
    # Combine all fields together
    lsd = struct.pack('>HBBBB', width, height, *struct.unpack('BBBB', packed_fields), background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    # We use 256 colors in our palette
    color_count = 2 ** 8
    
    # Each color is represented by three bytes: red, green, blue
    color_size = 3
    
    # Read 256 colors from /dev/urandom
    color_table = rng.read(color_count * color_size)
    
    # 4. Graphic Control Extension
    # We don't need this extension, so we just write an empty block
    gce = b'\x21\xf9\x04\x04\x00\x00\xff\x00'
    
    # 5. Image Descriptor
    # Random left and top position
    left, top = struct.unpack('>HH', rng.read(4))
    
    # Width and height are the same as the logical screen descriptor
    image_width = width
    image_height = height
    
    # Local Color Table Flag (1 bit)
    local_color_table_flag = 0
    
    # Interlace Flag (1 bit)
    interlace_flag = 0
    
    # Sort Flag (1 bit)
    sort_flag = 0
    
    # Reserved (2 bits)
    reserved = 0
    
    # Size of Local Color Table (3 bits)
    size_of_local_color_table = 0
    
    # Packed fields
    packed_fields = struct.pack('B', local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved << 3 | size_of_local_color_table)
    
    # LZW Minimum Code Size
    minimum_code_size = 8
    
    # Combine all fields together
    image_descriptor = struct.pack('>HHBBBBB', left, top, image_width, image_height, *struct.unpack('BBBB', packed_fields), minimum_code_size)
    
    # 6. Image Data
    # We will create a single frame with no transparency
    # We will use LZW compression
    # We will not interlace the image
    image_data = []
    
    # Add the clear code (256)
    image_data.append(struct.pack('B', 256))
    
    # Add the end of information code (257)
    image_data.append(struct.pack('B', 257))
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output stream
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(image_data) + trailer)
