#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphic Control Extension
    # We will use a random delay time, and we will also set the transparent flag
    gce = struct.pack('>BBBHB', 0xF9, 4, rng.read(1)[0], rng.read(2)[0], rng.read(1)[0])

    # 5. Image Descriptor
    # We will use a random x-offset and y-offset, and we will also set the local color table flag
    image_descriptor = struct.pack('>BBHBB', 0x2C, 0, rng.read(2)[0], rng.read(2)[0], rng.read(1)[0])

    # 6. Image Data
    # We will use a random minimum code size, and we will also use a random amount of blocks
    lzw_minimum_code_size = rng.read(1)[0]
    total_blocks_size = rng.read(1)[0]
    blocks = []
    while True:
        if len(blocks) >= total_blocks_size:
            break

        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        if not block_size:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)