#!/usr/bin/env python3

import struct
from typing import BinaryIO

def read_image(f: BinaryIO, width: int, height: int):
    """Read an image from a GIF file."""
    
    # Read the header
    f.seek(-768, 2)
    header = f.read()
    if not header.startswith(b'GIF'):
        raise ValueError('Not a valid GIF file')
        
    # Find the Graphic Control Extension
    gce = None
    while True:
        block = f.read(10)
        if len(block) < 10:
            break
            
        label = block[0]
        size = struct.unpack('<H', block[1:3])[0]
        data = block[3:]
        
        if label == 0xF9 and size >= 4:
            gce = data[:size]
            break
                
        f.seek(size - 1, 1)
        
    # Find the Image Descriptor
    while True:
        block = f.read(10)
        if len(block) < 10:
            raise ValueError('No Image Data found in GIF file')
            
        label = block[0]
        size = struct.unpack('<H', block[1:3])[0]
        data = block[3:]
        
        if label == 0x2C:
            break
            
        f.seek(size - 1, 1)
        
    # Read the LSD
    lsd = data[:10]
    local_color_table = bool(lsd[9] & 0x80)
    interlaced = bool(lsd[8] & 0x40)
    num_colors = 2 ** ((lsd[9] & 0x07) + 1)
    x = struct.unpack('<H', lsd[1:3])[0]
    y = struct.unpack('<H', lsd[3:5])[0]
    width = struct.unpack('<H', lsd[5:7])[0]
    height = struct.unpack('<H', lsd[7:9])[0]
    
    # Read the Color Table
    if local_color_table:
        color_table = f.read(num_colors * 3)
    else:
        color_table = f.read(768)
        
    # Skip any remaining blocks
    while True:
        block = f.read(10)
        if len(block) < 10:
            break
            
        label = block[0]
        size = struct.unpack('<H', block[1:3])[0]
        data = block[3:]
        
        f.seek(size - 1, 1)
        
    # Read the Image Data
    image_data = f.read()
    
    return header, gce, lsd, color_table, image_data

def write_image(out: BinaryIO, header, gce, lsd, color_table, image_data):
    """Write an image to a GIF file."""
    
    # Make sure we have enough colors for the image
    if len(color_table) != 768:
        raise ValueError('Color table must be exactly 768 bytes long')
        
    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)

