#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = 1 << (rng.read(1)[0] & 0b111)
    color_table = rng.read(num_colors * 3)

    # 4. Optional Graphics Control Extension
    # We will randomly decide whether or not we want one
    gce = None
    if rng.read(1)[0] % 2 == 0:
        # If so, let's create one!
        # First, the extension introducer (fixed value)
        gce = b'\x21\xf9'
        # Then the Graphic Control Label (fixed value)
        gce += b'\x04'
        # Block Size (always 4)
        gce += b'\x04'
        # Packed Fields (random)
        gce += rng.read(1)
        # Delay Time (two bytes, random)
        gce += rng.read(2)
        # Transparent Color Index (one byte, random)
        gce += rng.read(1)
        # Block Terminator (fixed value)
        gce += b'\x00'

    # 5. Image Descriptor
    image_descriptor = b'\x2C'

    # 6. Image Data
    # We will use LZW compression, so first we need to write the minimum code size
    lzw_minimum_code_size = rng.read(1)[0]
    # Now we can start writing blocks
    blocks = []
    while True:
        # Each block starts with its length (one byte)
        block_length = rng.read(1)[0]
        if block_length == 0:
            break
        # Followed by the data itself
        blocks.append(rng.read(block_length))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)