#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = rng.read(1)[0] % 256 + 1
    color_table = rng.read(num_colors * 3)

    # 4. Graphic Control Extension
    # We don't need any fancy features, so we can just use the default values
    gce = b'\x21\xf9\x04\x08\x00\x00\xff\xff\xff\xff'

    # 5. Image Descriptor
    # Use the same dimensions as the logical screen descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2c, 0, 0, width, height, 0, rng.read(1)[0])

    # 6. Image Data
    # Let's use LZW compression with minimum code size of 8 bits
    lzw_minimum_code_size = rng.read(1)[0] & 0b1111

    # Now let's generate some data!
    # First, we need to write the initial code size
    # Then, we need to generate some data
    # Finally, we need to terminate the stream
    blocks = []
    while True:
        data = rng.read(1)
        if not data or len(data) == 0:
            break
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)