#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_gif(out: BinaryIO, rng: BinaryIO):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    global_color_table_flag = bool(packed & 0b10000000)
    color_resolution = (packed >> 4) & 0b111
    sort_flag = bool(packed & 0b100)
    global_color_table_size = 2 ** ((packed & 0b111) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    lsd = struct.pack('>HHBBBBB', width, height, packed, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    color_table = rng.read(global_color_table_size * 3)
    
    # 4. Graphics Control Extension
    if global_color_table_flag:
        # 4.1. Packed Fields
        packed = rng.read(1)[0]
        disposal_method = (packed >> 2) & 0b111
        user_input_flag = bool(packed & 0b10)
        transparent_color_flag = bool(packed & 0b1)
        
        # 4.2. Delay Time
        delay_time = struct.unpack('>H', rng.read(2))[0]
        
        # 4.3. Transparent Color Index
        transparent_color_index = rng.read(1)[0] if transparent_color_flag else None
        
        # 4.4. Block Terminator
        block_terminator = rng.read(1)[0]
        
        gce = struct.pack('>BBHB', 0xF9, packed, delay_time, transparent_color_index or 0)
    else:
        gce = None
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        
        if block_size == 0:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)