#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = [rng.read(3) for _ in range(num_colors)]
    out.write(bytes(color_table))

    # 4. Image Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other ID fields: Left position, Top position, Packed fields (using a random byte for variety)
    left_position, top_position = struct.unpack('>HH', rng.read(4))
    packed_fields = rng.read(1)[0]
    out.write(struct.pack('>HHHH', width, height, left_position, top_position, packed_fields))

    # 5. Local Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = [rng.read(3) for _ in range(num_colors)]
    out.write(bytes(color_table))

    # 6. Image Data
    # Let's just use a random amount of data
    image_data_size = rng.read(2)
    image_data = rng.read(image_data_size[0])
    out.write(image_data)

    # 7. Trailer
    out.write(b'\x3b')
