#!/usr/bin/env python3

import struct
from typing import BinaryIO

def read_int(f: BinaryIO) -> int:
    return struct.unpack('<i', f.read(4))[0]

def write_int(f: BinaryIO, i: int):
    f.write(struct.pack('<i', i))

def read_float(f: BinaryIO) -> float:
    return struct.unpack('<f', f.read(4))[0]

def write_float(f: BinaryIO, fv: float):
    f.write(struct.pack('<f', fv))

def read_vec2(f: BinaryIO) -> (float, float):
    x = read_float(f)
    y = read_float(f)
    return (x, y)

def write_vec2(f: BinaryIO, v: (float, float)):
    write_float(f, v[0])
    write_float(f, v[1])

def read_vec3(f: BinaryIO) -> (float, float, float):
    x = read_float(f)
    y = read_float(f)
    z = read_float(f)
    return (x, y, z)

def write_vec3(f: BinaryIO, v: (float, float, float)):
    write_float(f, v[0])
    write_float(f, v[1])
    write_float(f, v[2])

def read_vec4(f: BinaryIO) -> (float, float, float, float):
    x = read_float(f)
    y = read_float(f)
    z = read_float(f)
    w = read_float(f)
    return (x, y, z, w)

def write_vec4(f: BinaryIO, v: (float, float, float, float)):
    write_float(f, v[0])
    write_float(f, v[1])
    write_float(f, v[2])
    write_float(f, v[3])

def read_mat4(f: BinaryIO) -> ((float, float, float, float), (float, float, float, float), (float, float, float, float), (float, float, float, float)):
    m0 = read_vec4(f)
    m1 = read_vec4(f)
    m2 = read_vec4(f)
    m3 = read_vec4(f)
    return (m0, m1, m2, m3)

def write_mat4(f: BinaryIO, m: ((float, float, float, float), (float, float, float, float), (float, float, float, float), (float, float, float, float))):
    write_vec4(f, m[0])
    write_vec4(f, m[1])
    write_vec4(f, m[2])
    write_vec4(f, m[3])

def read_string(f: BinaryIO) -> str:
    length = read_int(f)
    if length == 0:
        return None
    else:
        return f.read(length).decode()

def write_string(f: BinaryIO, s: str):
    if s is None:
        write_int(f, 0)
    else:
        write_int(f, len(s))
        f.write(s.encode())

def read_bool(f: BinaryIO) -> bool:
    b = read_int(f)
    assert b in [0, 1], 'Invalid boolean value'
    return b != 0

def write_bool(f: BinaryIO, b: bool):
    write_int(f, 1 if b else 0)

def read_color(f: BinaryIO) -> (float, float, float, float):
    r = read_float(f)
    g = read_float(f)
    b = read_float(f)
    a = read_float(f)
    return (r, g, b, a)

def write_color(f: BinaryIO, c: (float, float, float, float)):
    write_float(f, c[0])
    write_float(f, c[1])
    write_float(f, c[2])
    write_float(f, c[3])

def read_texture(f: BinaryIO) -> (str, str):
    path = read_string(f)
    name = read_string(f)
    return (path, name)

def write_texture(f: BinaryIO, t: (str, str)):
    write_string(f, t[0])
    write_string(f, t[1])

def read_material(f: BinaryIO) -> dict:
    mat = {}
    while True:
        key = read_string(f)
        if key is None:
            break
        val = read_string(f)
        mat[key] = val
