#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width = rng.read(2)
    height = rng.read(2)
    # Packed field
    packed = b'\x00\x00\x00\x00'
    # Background color index
    background_color_index = rng.read(1)
    # Pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)
    # Write all of these fields to the output stream
    out.write(width + height + packed + background_color_index + pixel_aspect_ratio)

    # 3. Global Color Table
    # Number of colors in the table
    num_colors = rng.read(1)[0]
    # Size of each color
    size_of_each_color = rng.read(1)[0]
    # Colors themselves
    for _ in range(num_colors):
        out.write(rng.read(size_of_each_color))
    
    # 4. Image Descriptor
    # Separator
    out.write(b'\x2c')
    # Left position
    left_position = rng.read(2)
    # Top position
    top_position = rng.read(2)
    # Width
    width = rng.read(2)
    # Height
    height = rng.read(2)
    # Packed field
    packed = rng.read(1)
    # Local Color Table Flag
    local_color_table_flag = rng.read(1)[0] & 0b10000000
    # Interlace Flag
    interlace_flag = rng.read(1)[0] & 0b01000000
    # Sort Flag
    sort_flag = rng.read(1)[0] & 0b00100000
    # Reserved
    reserved = rng.read(1)[0] & 0b00011111
    # Length of Local Color Table
    length_of_local_color_table = rng.read(1)[0]
    # True Color Table Size
    true_color_table_size = rng.read(1)[0]
    # Write all of these fields to the output stream
    out.write(left_position + top_position + width + height + packed + struct.pack('B', local_color_table_flag | interlace_flag | sort_flag | reserved) + struct.pack('B', length_of_local_color_table | true_color_table_size))

    # 5. Local Color Table
    # Number of colors in the table
    num_colors = rng.read(1)[0]
    # Size of each color
    size_of_each_color = rng.read(1)[0]
    # Colors themselves
    for _ in range(num_colors):
        out.write(rng.read(size_of_each_color))

    # 6. Image Data
    # Minimum code size
    minimum_code_size = rng.read(1)[0]
    # Image data itself
    while True:
        byte = rng.read(1)
        if not byte:
            break
        out.write(byte)

    # 7. Trailer
    out.write(b';')

    return len(width) + len(height) + len(background_color_index) + len(pixel_aspect_ratio) \
           + len(num_colors) + len(size_of_each_color) * num_colors \
           + len(left_position) + len(top_position) + len(width) + len(height) + len(packed) \
           + len(minimum_code_size) + sum([len(rng.read(1)) for _ in range(2 ** minimum_code_size)])
