#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = rng.read(num_colors * 3)

    # 4. Image Data
    image_data = []
    # Loop over all pixels in the image
    for y in range(height):
        # For each pixel, we need to write a single-byte code that indicates how many times to repeat the same color
        # So let's start with an empty list of codes
        codes = []
        # And then loop over all columns in this row
        for x in range(width):
            # We want to pick a random color from the palette
            color_index = rng.read(1)[0] % num_colors
            # But if this is the first pixel in the row, or if the previous pixel had a different color,
            # we need to add a new code to the list
            if not x or color_index != codes[-1]:
                codes.append(color_index)
        # Now that we have a list of codes, we can pack them together as a string
        image_data.append(struct.pack('>' + 'B'*len(codes), *codes))

    # 5. Trailer
    trailer = b';'

    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    out.write(global_color_table)
    for data in image_data:
        out.write(data)
    out.write(trailer)

