#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    lsd = struct.pack('>HBBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    # Random number of colors (up to 256)
    num_colors = (rng.read(1)[0] % 256) + 1
    
    # Each color is three bytes
    color_table = rng.read(num_colors * 3)
    
    # 4. Graphic Control Extension
    gce = struct.pack('>BBBB', 0xF9, 4, rng.read(1)[0], rng.read(2)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    while True:
        block_length = (rng.read(1)[0] % 255) + 1
        
        # We need to read exactly as many bytes as specified in the block length
        blocks.append(rng.read(block_length))
        
        # If the last block was less than 255 bytes long, stop reading
        if len(blocks[-1]) < 255: break
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything together
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)
