#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = [
        0b10000000 | (rng.read(1)[0] & 0x7F),   # Global Color Table Flag
        rng.read(1)[0],                         # Color Resolution
        rng.read(1)[0] & 0x7F,                  # Sort Flag
        rng.read(1)[0],                         # Size of Global Color Table
    ] + list(struct.unpack('>BBH', rng.read(3)))

    # 3. Graphic Control Extension
    gce = [
        0x21,                                   # Extension Introducer
        0xf9,                                   # Graphic Control Label
        0x04,                                   # Block Size
        rng.read(1)[0],                         # Packed Fields
        *struct.unpack('>HH', rng.read(4)),     # Delay Time / Transparent Color Index
        0,                                      # Block Terminator
    ]

    # 4. Image Descriptor
    id = [
        0x2c,                                   # Image Separator
        *struct.unpack('>HHB', rng.read(5)),    # Left Position / Top Position / Width / Height
        0,                                      # Packed Fields
    ]

    # 5. Local Color Table
    # We use a local color table with 256 entries because we want to be able to pick any color as transparent
    lct = [
        0xff,                                   # Block Size
    ] + [*struct.unpack('>BBB', rng.read(3))] * 256

    # 6. Raster Data
    # We use a single pixel per frame so that we can set the transparent color index in the GCE
    pixels = [
        0x00,                                   # Minimum Code Length
        0x01,                                   # Image Data
        0x00,                                   # End of Image
    ]

    # Write everything together
    out.write(header)
    out.write(bytes(lsd))
    out.write(bytes(gce))
    out.write(bytes(id))
    out.write(bytes(lct))
    out.write(bytes(pixels))

