#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = [rng.read(3) for _ in range(num_colors)]
    for rgb in color_table:
        out.write(rgb)

    # 4. Graphics Control Extension (optional)
    has_gce = bool((rng.read(1)[0] & 1 << 7) >> 7)
    if not has_gce:
        return

    out.write(b'\x21\xf9\x04')
    transparent_color_flag = bool((rng.read(1)[0] & 1 << 7) >> 7)
    user_input_flag = bool((rng.read(1)[0] & 1 << 6) >> 6)
    disposal_method = rng.read(1)[0] & 0x07
    delay_time = rng.read(2)
    transparent_color_index = rng.read(1)[0]
    block_terminator = b'\x00'
    out.write(struct.pack('BBBBB', transparent_color_flag, user_input_flag, disposal_method, *delay_time, transparent_color_index))
    out.write(block_terminator)

    # 5. Image Descriptor
    out.write(b'\x2c')
    left, top, width, height = struct.unpack('>HHHH', rng.read(8))
    local_color_table_flag = bool((rng.read(1)[0] & 1 << 7) >> 7)
    interlace_flag = bool((rng.read(1)[0] & 1 << 6) >> 6)
    sort_flag = bool((rng.read(1)[0] & 1 << 5) >> 5)
    reserved_bits = rng.read(1)[0] & 0x07
    out.write(struct.pack('>HHHHBB', left, top, width, height, local_color_table_flag, interlace_flag, sort_flag, reserved_bits))

    # 6. Local Color Table (optional)
    if local_color_table_flag:
        num_local_colors = (rng.read(1)[0] % 256) or 1
        local_color_table = [rng.read(3) for _ in range(num_local_colors)]
        for rgb in local_color_table:
            out.write(rgb)

    # 7. Image Data
    lzw_minimum_code_size = rng.read(1)[0]
    out.write(lzw_minimum_code_size)

    # 8. Trailer
    out.write(b';')

