#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color_index = packed & 0b11111
    aspect_ratio = packed >> 7
    lsd = struct.pack('>HBB', width, height, background_color_index | (aspect_ratio << 7))

    # 3. Global Color Table
    has_global_color_table = bool(rng.read(1)[0])
    global_color_table_size = 2 ** ((rng.read(1)[0] & 0b111) + 1)
    color_resolution = (rng.read(1)[0] & 0b11000000) >> 6
    sort_flag = bool((rng.read(1)[0] & 0b10000000) >> 7)
    global_color_table = rng.read(3 * global_color_table_size)

    # 4. Graphic Control Extension
    if has_global_color_table:
        delay_time = struct.unpack('>H', rng.read(2))[0]
        transparent_color_index = rng.read(1)[0]
        disposal_method = rng.read(1)[0]
        gce = struct.pack('>BBBB', 0xF9, 4, delay_time, transparent_color_index | (disposal_method << 2))
    else:
        gce = None

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)