#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
	width &= 0x7FFF
	height &= 0x7FFF

    # Random color resolution
    color_res = rng.read(1)[0] & 0x07

    # Global Color Table Flag
    global_color_table_flag = rng.read(1)[0] & 0x01

    # Sort flag
    sort_flag = rng.read(1)[0] & 0x01

    # Size of Global Color Table
    size_of_global_color_table = rng.read(1)[0] & 0xFF

    # Background Color Index
    background_color_index = rng.read(1)[0] & 0xFF

    # Pixel Aspect Ratio
    pixel_aspect_ratio = rng.read(1)[0] & 0xFF

    logical_screen_descriptor = struct.pack(
        '>HHBBB',
        width,
        height,
        color_res,
        global_color_table_flag << 7 | \
            sort_flag << 6 | \
            size_of_global_color_table,
        background_color_index,
        pixel_aspect_ratio
    )

    # 3. Global Color Table
    if global_color_table_flag == 1:
        global_color_table = bytearray()

        for i in range(size_of_global_color_table):
            r = rng.read(1)[0] & 0xFF
            g = rng.read(1)[0] & 0xFF
            b = rng.read(1)[0] & 0xFF

            global_color_table += struct.pack('BBB', r, g, b)

    else:
        global_color_table = None

    # 4. Image Descriptor
    image_separator = b'\x2C'

    # Random left position
    left = rng.read(1)[0] & 0xFF

    # Random top position
    top = rng.read(1)[0] & 0xFF

    # Random width
    width = rng.read(1)[0] & 0xFF

    # Random height
    height = rng.read(1)[0] & 0xFF

    # Local Color Table Flag
    local_color_table_flag = rng.read(1)[0] & 0x01

    # Interlace Flag
    interlace_flag = rng.read(1)[0] & 0x01

    # Sort Flag
    sort_flag = rng.read(1)[0] & 0x01

    # Reserved Field
    reserved = rng.read(1)[0] & 0x07

    # LZW Minimum Code Size
    lzw_minimum_code_size = rng.read(1)[0] & 0xFF

    image_descriptor = struct.pack(
        '>BBHHBBBB',
        image_separator,
        0,
        left,
        top,
        width,
        height,
        local_color_table_flag << 7 | \
            interlace_flag << 6 | \
            sort_flag << 5 | \
            reserved,
        lzw_minimum_code_size
    )

    # 5. Local Color Table
    if local_color_table_flag == 1:
        local_color_table = bytearray()

        for i in range(size_of_global_color_table):
            r = rng.read(1)[0] & 0xFF
            g = rng.read(1)[0] & 0xFF
            b = rng.read(1)[0] & 0xFF

            local_color_table += struct.pack('BBB', r, g, b)

    else:
        local_color_table = None

    # 6. Block
    block_start = b'\x21\xF9\x04\x00' + b'\x00'*int((rng.read(1)[0] & 0xFF)/10) + b'\x00'

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to output stream
    out.write(header)
    out.write(logical_screen_descriptor)

    if global_color_table is not None:
        out.write(global_color_table)

    out.write(image_descriptor)

    if local_color_table is not None:
        out.write(local_color_table)

    out.write(block_start)
    out.write(trailer)