#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
        
    # 4. Image Descriptor
    # Left position of image is always zero
    left = 0

    # Top position of image is randomly chosen from [0, height - 1]
    top = rng.read(1)[0] % height

    # Width and height are also randomly chosen from [0, max(width, height)]
    width = rng.read(1)[0] % max(width, height) + 1
    height = rng.read(1)[0] % max(width, height) + 1

    # Flags field: packed fields (using a random byte for variety).
    flags = rng.read(1)[0]
    
    # 5. Local Color Table (if applicable)
    if flags & 0b10000000 == 0b10000000:
        local_color_table = rng.read(3 * ((flags & 7) + 1))
    else:
        local_color_table = None
        
        
    # 6. Image Data
    # Number of bits per pixel
    bpp = 8
    
    # Get the total length in bytes of the image data
    data_length = int((width * height * bpp / 8) + 1)
    
    # Read that many bytes from the RNG
    image_data = rng.read(data_length)
    
    
    # Finally, write everything to our output stream!
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(struct.pack('>HHBBB', left, top, width, height, flags))
    if local_color_table!= None:
        out.write(local_color_table)
    out.write(image_data)

