#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = None
    if rng.read(1)[0]:
        # Transparent color flag set to true
        transparent_color_index = rng.read(1)[0]
        
        # Delay time in hundredths of seconds (1/100th of a second)
        delay = rng.read(2)
        
        # Transparent color index is always last
        gce = b'\x21\xF9\x04' + delay + transparent_color_index.to_bytes(1, 'big')
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('>HHBBB', 0, 0, 0b10000000 | ((num_colors - 1) << 4), 0, 0)
    
    # 6. Image Data
    # We'll just write some random data here
    image_data = rng.read((width * height // 2) or 1)

    # Write everything to output
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data)

