#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 4, rng.read(1)[0], int((rng.read(1)[0] / 256) +.5))
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHBBBB', 0x2C, rng.read(1)[0], 0, 0, width, height, 0)
    
    # 6. Local Color Table (optional; let's do this too)
    if rng.read(1)[0]:
        local_color_table = rng.read(3 * ((rng.read(1)[0] % 256) or 1))
    else:
        local_color_table = None
    
    # 7. Image Data
    data = []
    # First byte is a code indicating how to interpret subsequent bytes
    while True:
        code = rng.read(1)[0]
        
        # 0 - end of image
        if not code:
            break
        
        # 1-239 - literal bytes
        elif code <= 239:
            data += [rng.read(code)]
            
        # 240-254 - repeat previous byte n times
        elif code >= 240:
            length = 1 << (code - 240)
            
            if length == 1:
                data += [data[-1]]
            else:
                data += [data[-1]] * length
            
    # 8. Trailer
    trailer = b';'
    
    # Write everything to output stream
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce:
        out.write(gce)
    out.write(image_descriptor)
    if local_color_table:
        out.write(local_color_table)
    out.write(b''.join(data))
    out.write(trailer)

