#!/usr/bin/env python3
import sys
import random
import struct


def main():
    if len(sys.argv)!= 3:
        print("Usage: {} input.gif output.gif".format(sys.argv[0]))
        return -1
    
    in_name, out_name = sys.argv[1], sys.argv[2]
    try:
        with open(in_name, 'rb') as f:
            gif = f.read()
    except FileNotFoundError:
        print("{} does not exist!".format(in_name), file=sys.stderr)
        return -1

    with open(out_name, 'wb') as f:
        randomize_gif(f, gif)


def randomize_gif(out, gif):
    """Randomize all parts of a GIF file"""
    rng = random.Random(None)

    # 1. Header
    header = gif[:6]
    version = header[3] & ~0b111  # Clear bits 3 and 4
    header = header[:3] + bytes([version | rng.randint(0, 7)]) + header[4:]

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('<HH', gif[6:10])
    global_color_table_flag = bool(header[7] & 0b10000000)
    local_color_table_flag = bool(header[9] & 0b10000000)
    background_index = header[9] & 0b00001111
    pixel_aspect_ratio = header[10]
    screen_descriptor = struct.pack('<HHBBBH', width, height, global_color_table_flag << 7 | local_color_table_flag << 6 | background_index, pixel_aspect_ratio, *struct.unpack('<BBBBBB', gif[10:17]))

    # 3. Global Color Table
    if global_color_table_flag:
        color_table_size = ((screen_descriptor[7] & 0b00000111) ** 2) * 3
        color_table = gif[13:13+color_table_size]
        color_table = [bytes([(c ^ rng.randint(0, 255)) for c in color_table[i*3:(i+1)*3]]) for i in range((color_table_size // 3))]
        color_table = b''.join(color_table)
    else:
        color_table = None

    # 4. Graphics Control Extension
    if gif[13:18] == b'GIF89a\x00':
        gce = gif[13:34]
        gce = gce[:3] + bytes([gce[3] ^ rng.randint(0, 255)]) + gce[4:]
    else:
        gce = None

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        if len(blocks) >= total_blocks_size:
            break

        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        if not block_size:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)