#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0]
    
    # 7. Data sub-blocks
    # Let's use an arbitrary maximum of 256 data sub-blocks per frame
    max_data_sub_blocks = 256
    blocks = []
    while True:
        if len(blocks) >= max_data_sub_blocks:
            break
        
        # Each data sub-block is either raw data or compressed data
        choice = rng.read(1)[0] % 2
        if choice == 0:
            # Raw data
            
            # The length of this raw data is between 1 and 255 inclusive
            length = rng.read(1)[0] or 1
            blocks.append(rng.read(length))
            
        else:
            # Compressed data
            
            # The length of this compressed data is between 2 and 255 inclusive
            length = rng.read(1)[0] + 2
            data = rng.read(length - 2)
            
            # Decompress the data
            decompressed = []
            while data:
                # Read the first byte to determine how many bits are used for the next byte
                control_byte = data[0]
                
                # Remove that byte from our data buffer
                del data[0]
                
                # For every bit set in the control byte...
                for i in range(8):
                    # If the bit is not set (i.e., we're reading uncompressed)...
                    if ((control_byte >> i) & 1)!= 1:
                        # Add another byte to our output
                        decompressed.append(data[0])
                        
                        # Remove that byte from our data buffer
                        del data[0]
                        
                    else:
                        # Otherwise, read more bytes to get the actual value
                        # This is done by looking up the next byte in the table
                        # We'll just assume there will always be enough data here
                        lookup = data[:2]
                        del data[:2]
                        
                        # Get the actual value from the table
                        value = table[lookup]
                        
                        # And add it to our output
                        decompressed += value
                
            # Now we can write the decompressed data as a new data sub-block
            blocks.append(bytes(decompressed))
            
    # 8. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    if rng.read(1)[0]:
        out.write(gce)
        out.write(image_descriptor)
    out.write(struct.pack('B', lzw_minimum_code_size))
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)