#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    packed_fields = rng.read(1)[0]

    id = struct.pack('>HHBB', left, top, width, height) + packed_fields.to_bytes(1, 'big')

    # 6. Local Color Table (optional)
    local_color_table = None
    if packed_fields & 0b10000000:
        num_colors = (rng.read(1)[0] % 256) or 1
        local_color_table = rng.read(3 * num_colors)

    # 7. Image Data
    image_data = []
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x2C:
            # Image data block

            # Block size (always 0x08)
            assert rng.read(1)[0] == 0x08

            # Left position of this frame in the canvas
            left = rng.read(2)

            # Top position of this frame in the canvas
            top = rng.read(2)

            # Width of this frame
            width = rng.read(2)

            # Height of this frame
            height = rng.read(2)

            # Packed field (bit 0: local color table flag; bit 1-3: interlace flag; bits 4-7: sort flag)
            packed_field = rng.read(1)[0]

            # LZW minimum code size (must be between 2 and 12 inclusive)
            min_code_size = rng.read(1)[0]

            # Frame data
            data = b''
            while True:
                chunk_size = rng.read(1)[0]

                if not chunk_size:
                    break

                data += rng.read(chunk_size)

            image_data.append((left, top, width, height, packed_field, min_code_size, data))
        elif block_type == 0x21:
            # Extension block

            # Block size
            block_size = rng.read(1)[0]

            # Extension type
            extension_type = rng.read(1)[0]

            if extension_type == 0xFF:
                # Application extension

                # Application ID
                app_id = rng.read(11)

                # Application authentication code
                auth_code = rng.read(3)

                # Application data
                app_data = b''
                while len(app_data) < block_size - 12:
                    chunk_size = rng.read(1)[0]

                    if not chunk_size:
                        break

                    app_data += rng.read(chunk_size)

                print(f'{app_id} {auth_code} {app_data}')
            elif extension_type == 0xFE:
                # Comment extension

                comment = b''
                while len(comment) < block_size - 1:
                    chunk_size = rng.read(1)[0]

                    if not chunk_size:
                        break

                    comment += rng.read(chunk_size)

                print(f'{comment}')
            elif extension_type == 0x01:
                # Plain text extension

                # Text grid left
                tx = rng.read(2)

                # Text grid top
                ty = rng.read(2)

                # Text grid width
                tw = rng.read(2)

                # Text grid height
                th = rng.read(2)

                # Cell width
                cw = rng.read(1)[0]

                # Cell height
                ch = rng.read(1)[0]

                # Fg color index
                fgci = rng.read(1)[0]

                # Bg color index
                bgci = rng.read(1)[0]

                # Character cell data
                cc_data = ''
                while len(cc_data) < block_size - 10:
                    char_size = rng.read(1)[0]

                    if not char_size:
                        break

                    cc_data += chr(char_size)

                print(f'{tx}, {ty}, {tw}, {th}, {cw}, {ch}, {fgci}, {bgci}, {cc_data}')
            elif extension_type == 0xF9:
                # Graphic control extension

                # Block size (always 0x04)
                assert rng.read(1)[0] == 0x04

                # Flags (bit 0: transparent color flag; bit 1: user input flag; bit 2: disposal method)
                flags = rng.read(1)[0]

                # Delay time (in hundredths of seconds)
                delay_time = rng.read(2)

                # Transparent color index
                trans_index = rng.read(1)[0]

                # Terminator
                assert rng.read(1)[0] == 0x00

                print(f'{flags}, {delay_time}, {trans_index}')
            else:
                raise Exception(f'Unknown extension type: {extension_type:#x}')
        elif block_type == 0x3B:
            # End of file
            break
        else:
            raise Exception(f'Unknown block type: {block_type:#x}')

    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(id)
    if local_color_table is not None:
        out.write(local_color_table