#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes long

        if not block_size:
            break

        if block_size > 255:
            block_size = 255

        blocks.append(rng.read(block_size))

    # Compress all of our data together with the minimum code size
    compressed_data = compress(b''.join(blocks), lzw_minimum_code_size)

    # Write everything in order
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor +
              struct.pack('>BB', lzw_minimum_code_size, len(compressed_data)))
    out.write(compressed_data)


# Compresses the given data using the LZW algorithm
def compress(data: bytes, min_code_size=2):
    dictionary = {bytes([i]): bytes([i]) for i in range(min_code_size)}

    result = [dictionary[data[:min_code_size]]]
    next_code = min_code_size + 1

    for char in data[min_code_size:]:
        current_entry = dictionary[result[-1]] + bytes([char])

        if current_entry in dictionary:
            result.append(current_entry)
        else:
            result.append(dictionary[result[-1]])

            if next_code < 2 ** 12:
                dictionary[current_entry] = bytes([(next_code >> 8) & 0xFF, next_code & 0xFF])

                next_code += 1

    return b''.join(map(lambda x: x + b'\x00', result))


if