#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Color table size is always 256 colors
    color_table_size = 256 * 3
    
    # Background color index is always zero
    background_color_index = 0
    
    # Pixel aspect ratio is always 0
    pixel_aspect_ratio = 0
    
    # Pack all of that into one byte string
    lsd = struct.pack(
        '>H H B B B', 
        width,
        height,
        color_table_size,
        background_color_index,
        pixel_aspect_ratio
    )
    
    # 3. Global Color Table
    global_color_table = []
    for i in range(256):
        r, g, b = [int(c) for c in rng.read(3)]
        global_color_table += [r, g, b]
        
    # Pack it into a single byte string
    color_table = bytes(global_color_table)
    
    # 4. Graphics Control Extension
    # This extension is optional; we'll use it only sometimes
    gce = None
    
    # 5. Image Descriptor
    image_left_position = 0
    image_top_position = 0
    image_width = width
    image_height = height
    
    # Packed into one byte string
    image_descriptor = struct.pack(
        '>HH HH BB',
        image_left_position,
        image_top_position,
        image_width,
        image_height,
        0, # No local color table
        0  # LZW minimum code size
    )
    
    # 6. Blocks
    # We will write an arbitrary amount of data here, but we need to know how much
    # space we have available first. To do this, we will write a variable-length integer
    # with the length of the following data. The maximum value of this integer is 255, so
    # we can't just write a bunch of random bytes until we run out of space. Instead, we
    # will choose a random length between 1 and 255 inclusive, then write that many bytes.
    # After writing those bytes, we will repeat the process again.
    # 
    # Note that this algorithm may result in some blocks being larger than others. If you
    # want to make sure every block has the same size, you should pad them yourself before
    # passing them to this function.
    # 
    # Also note that this algorithm does not guarantee that there are no duplicate blocks.
    # It's possible that the last block written was exactly as large or larger than any other
    # block, which would cause us to end up repeating the same block over and over again.
    # However, this is unlikely enough that I don't think it matters too much.
    blocks = []
    total_blocks_size = int((rng.read(1)[0] % 255) + 1)
    
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0] + 1)
        blocks.append(rng.read(block_size))
        
        total_blocks_size -= block_size
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if len(gce) == 6:
        out.write(gce)
    out.write(image_descriptor)
    out.write(struct.pack('B', lzw_minimum_code_size))
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)