#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + (rng.read(1)[0] % 256)
    height = 1 + (rng.read(1)[0] % 256)
    flags = b'\x87\x00' if rng.read(1)[0] & 1 else b'\x07\x00'
    bg_color = bytes([rng.read(1)[0]]) * 3
    aspect_ratio = b'\x00'
    out.write(struct.pack('<HHBBB', width, height, *flags))
    out.write(bg_color)
    out.write(aspect_ratio)

    # Global color table
    num_colors = min((width + height), 256)
    for i in range(num_colors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(bytes([r,g,b]))

    # Image descriptor
    left = 0
    top = 0
    width = 1 + (rng.read(1)[0] % 256)
    height = 1 + (rng.read(1)[0] % 256)
    flags = b'\x87\x00' if rng.read(1)[0] & 1 else b'\x07\x00'
    local_color_table = False
    interlace = False
    sort = True
    reserved = b'\x00\x00'
    lct_size = 0
    out.write(struct.pack('<HHHHBB', left, top, width, height, *flags))
    out.write(reserved)
    out.write(b'\x00')
    out.write(b'\x00')
    out.write(b'\x00')

    # Local color table
    if not local_color_table:
        lct_size = 0
    elif lct_size == 0:
        lct_size = 2 << (rng.read(1)[0] & 7)
    for i in range(lct_size):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(bytes([r,g,b]))

    # Table based image data
    code_size = 8
    clear_code = 1 << code_size
    end_of_info_code = clear_code + 1
    end_of_image_code = end_of_info_code + 1
    max_code_size = 12
    next_code = dict()
    codes = []
    def add_code(c):
        if c < len(codes):
            return
        while len(codes) <= c:
            codes += [None]
        codes[c] = None
    def get_next_code():
        while True:
            c = ord(rng.read(1))
            yield c >> 4
            yield c & 0xf
    def write_code(c):
        assert c >= 0 and c < 2**code_size
        bits = [(c >> i) & 1 for i in reversed(range(code_size))]
        for bit in bits[:-1]:
            out.write(b'\xff') if bit else out.write(b'\x00')
        out.write(bits[-1])
    def flush_codes():
        while len(codes) > clear_code:
            write_code(len(codes)-clear_code-1)
            del codes[-1]
    def encode_pixel(p):
        nonlocal next_code

        if p is None:
            return end_of_image_code

        if p not in next_code:
            next_code[p] = get_next_code()

        try:
            c = next(next_code[p])
        except StopIteration:
            raise Exception("Ran out of codes") from None

        add_code(c)
        return c

    # LZW minimum code size
    out.write(bytes([min(max_code_size, code_size)]))

    # Encode pixels
    row_length = width
    prev_row = bytearray(row_length)
    curr_row = bytearray(row_length)
    for y in range(height):
        for x in range(row_length):
            curr_row[x] = rng.read(1)[0]
        for x in range(row_length):
            p = curr_row[x]

            if x == 0:
                prev_row[x] = p
                continue

            c = encode_pixel(prev_row[x])
            write_code(c)

            if c!= end_of_image_code:
                prev_row[x] = p

                if x+1 < row_length:
                    prev_row[x+1] = p

        if y+1 < height:
            for x in range(row_length):
                curr_row[x], prev_row[x] = prev_row[x], curr_row[x]

        flush_codes()

    # End of information code
    write_code(end_of_info_code)

    # Trailer
    out.write(b';')

