#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved_flags = rng.read(1)[0]
    lzw_minimum_code_size = rng.read(1)[0] + 1
    id = struct.pack('>HHBBBBB', left, top, width, height, local_color_table_flag, interlace_flag, sort_flag, reserved_flags, lzw_minimum_code_size)

    # 6. Local Color Table (if applicable)
    if local_color_table_flag:
        num_local_colors = (rng.read(1)[0] % 256) or 1
        local_color_table = rng.read(3 * num_local_colors)  # Random colors
    else:
        local_color_table = b''

    # 7. Image Data
    image_data = rng.read((width * height // 8) + ((width * height) % 8!= 0))

    # Write everything to output
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(id)
    if local_color_table_flag:
        out.write(local_color_table)
    out.write(image_data)

    return len(lsd) + len(color_table) + (len(gce) if gce is not None else 0) + len(id) + (len(local_color_table) if local_color_table_flag else 0) + len(image_data)

