#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)
    out.write(lsd)

    # 3. Global Color Table
    global_color_table_size = 2 << ((packed_fields & 7) + 1)
    global_color_table = b''
    if global_color_table_size > 0:
        global_color_table += rng.read(global_color_table_size * 3)
    
    # Write the global color table to the output stream
    out.write(global_color_table)

    # 4. Image Descriptor
    image_left = 0
    image_top = 0
    image_width = width
    image_height = height
    packed_fields = 0x00 | 0x20 # Local color table flag set, interlace flag clear, sort flag clear, local color table size is 2^(N+1)
    image_descriptor = struct.pack('>HHHHB', image_left, image_top, image_width, image_height, packed_fields)
    out.write(image_descriptor)

    # 5. Local Color Table
    local_color_table_size = 2 << ((packed_fields & 7) + 1)
    local_color_table = b''
    if local_color_table_size > 0:
        local_color_table += rng.read(local_color_table_size * 3)
    
    # Write the local color table to the output stream
    out.write(local_color_table)

    # 6. Image Data
    # Get the total length of all subblocks in the image data
    image_data_length = get_image_data_length()

    # Write the image data to the output stream
    write_image_data(out, image_data_length)

    return len(lsd) + len(global_color_table) + len(image_descriptor) + len(local_color_table) + image_data_length


# Returns the total length of all subblocks in the image data
def get_image_data_length():
    # The first block contains an introducer (0x2C) followed by the image descriptor
    image_data_length = 1 + 1 + 4 + 1

    # The second block contains the actual image data
    image_data_length += 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 