#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    
    image_descriptor = struct.pack('>HHBB', left, top, width, height, packed)
    
    # 6. Local Color Table (optional)
    if packed & 0b10000000 == 0b10000000:
        local_color_table = rng.read(3 * ((packed & 7) + 1))
    
    # 7. Image Data
    data = []
    while True:
        block_type = rng.read(1)[0]
        
        if block_type == 0x2C:
            # Image descriptor
            pass
        elif block_type == 0x21:
            # Extension
            extension_code = rng.read(1)[0]
            
            if extension_code == 0xFF:
                # Application Extension
                application_identifier = rng.read(8)
                
                if application_identifier == b'NETSCAPE2.0':
                    subblock_size = rng.read(1)[0]
                    
                    if subblock_size!= 3:
                        raise Exception("Invalid NETSCAPE2.0 subblock")
                    
                    subblock_id = rng.read(1)[0]
                    if subblock_id!= 1:
                        raise Exception("Invalid NETSCAPE2.0 subblock ID")
                    
                    loop_count = struct.unpack('>H', rng.read(2))[0]
                    data += [struct.pack('<BBH', 0x21, 0xFF, 0x0B)] + list(application_identifier) + [subblock_size, subblock_id, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, loop_count % 256, loop_count // 256]
                    
            else:
                # Plain old extension
                length = rng.read(1)[0]
                data += [struct.pack('<BB', 0x21, extension_code)] + list(rng.read(length))
                
        elif block_type == 0x3B:
            break
            
        else:
            raise Exception("Invalid block type")
            
    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    out.write(local_color_table)
    out.write(bytes(data))
    