#!/usr/bin/env python3

import struct
from typing import BinaryIO

def read_lsd(f: BinaryIO):
    """Read Logical Screen Descriptor from a binary stream"""

    # Read LSD header (always 10 bytes long)
    lsd_header = f.read(10)
    assert len(lsd_header) == 10, 'LSD header must be exactly 10 bytes!'

    # Unpack header into its components
    width, height, _, bit_depth, compression, filter_, interlace = \
            struct.unpack('>HHBBBB', lsd_header)

    return {
        'width': width,
        'height': height,
        'bit_depth': bit_depth,
        'compression': compression,
        'filter': filter_,
        'interlace': interlace
    }

def write_lsd(out: BinaryIO, lsd: dict):
    """Write Logical Screen Descriptor to a binary stream"""

    # Pack all fields of the LSD together and write it to the output file
    packed_lsd = struct.pack('>HHBBBB', **lsd)
    out.write(packed_lsd)

def read_color_table(f: BinaryIO, size: int):
    """Read Color Table from a binary stream"""

    # Read Color Table data
    color_table = f.read(size)
    assert len(color_table) == size, 'Color table must have length %d!' % size

    return color_table

def write_color_table(out: BinaryIO, color_table: bytes):
    """Write Color Table to a binary stream"""

    # Write Color Table data to the output file
    out.write(color_table)

def read_gce(f: BinaryIO):
    """Read Graphics Control Extension from a binary stream"""

    # Read GCE header (always 4 bytes long)
    gce_header = f.read(4)
    assert len(gce_header) == 4, 'GCE header must be exactly 4 bytes!'

    # Unpack header into its components
    block_size, flags, delay_time, transparent_index = \
            struct.unpack('>BBH2s', gce_header)

    return {
        'block_size': block_size,
        'flags': flags,
        'delay_time': delay_time,
        'transparent_index': transparent_index
    }

def write_gce(out: BinaryIO, gce: dict):
    """Write Graphics Control Extension to a binary stream"""

    # Pack all fields of the GCE together and write it to the output file
    packed_gce = struct.pack('>BBH2s', **gce)
    out.write(packed_gce)

def read_image_descriptor(f: BinaryIO):
    """Read Image Descriptor from a binary stream"""

    # Read ID header (always 12 bytes long)
    id_header = f.read(12)
    assert len(id_header) == 12, 'ID header must be exactly 12 bytes!'

    # Unpack header into its components
    left, top, width, height, local_color_table_flag, interlace_flag, sort_flag,\
            reserved, local_color_table_size = \
            struct.unpack('>HHHHBBBxxxB', id_header)

    return {
        'left': left,
        'top': top,
        'width': width,
        'height': height,
        'local_color_table_flag': bool(local_color_table_flag),
        'interlace_flag': bool(interlace_flag),
      'sort_flag': bool(sort_flag),
      'reserved': reserved,
        'local_color_table_size': local_color_table_size
    }

def write_image_descriptor(out: BinaryIO, id: dict):
    """Write Image Descriptor to a binary stream"""

    # Pack all fields of the ID together and write it to the output file
    packed_id = struct.pack('>HHHHBBBxxxB', **id)
    out.write(packed_id)

def read_image_data(f: BinaryIO, width: int, height: int, bit_depth: int):
    """Read Image Data from a binary stream"""

    # Calculate number of bytes per line based on bit depth
    num_bytes_per_line = ((width * bit_depth) // 8) + (((width * bit_depth) % 8)!= 0)

    # Read Image Data
    image_data = f.read(num_bytes_per_line * height)
    assert len(image_data) == num_bytes_per_line * height, 'Image data must have length %d!' % (num_bytes_per_line * height)

    return image_data

def write_image_data(out: BinaryIO, image_data: bytes):
    """Write Image Data to a binary stream"""

    # Write Image Data to the output file
    out.write(image_data)

def read_gif(in_filename: str):
    """Read GIF file from disk"""

    with open(in_filename, mode='rb') as in_:

        # Read GIF Header
        gif_header = in_.read(6)
        assert len(gif_header) == 6, 'GIF header must be exactly 6 bytes!'
        assert gif_header[0] == ord('G'), 'Invalid GIF signature!'
        assert gif_header[1] == ord('I'), 'Invalid    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)