#!/usr/bin/env python3

import struct


def read_uint16(f):
    return int.from_bytes(f.read(2), byteorder='little')


def write_uint16(f, value):
    f.write(value.to_bytes(2, byteorder='little'))


def read_uint32(f):
    return int.from_bytes(f.read(4), byteorder='little')


def write_uint32(f, value):
    f.write(value.to_bytes(4, byteorder='little'))


def read_float(f):
    return struct.unpack('<f', f.read(4))[0]


def write_float(f, value):
    f.write(struct.pack('<f', value))


def read_string(f):
    length = read_uint16(f)

    if length == 0:
        return ''

    data = f.read(length).decode('utf-8')
    assert len(data) == length
    return data


def write_string(f, string):
    encoded = bytes(string, 'utf-8')
    write_uint16(f, len(encoded))
    f.write(encoded)


def read_vector2d(f):
    x = read_float(f)
    y = read_float(f)
    return (x, y)


def write_vector2d(f, vector):
    write_float(f, vector[0])
    write_float(f, vector[1])


def read_color(f):
    r = read_float(f)
    g = read_float(f)
    b = read_float(f)
    a = read_float(f)
    return (r, g, b, a)


def write_color(f, color):
    write_float(f, color[0])
    write_float(f, color[1])
    write_float(f, color[2])
    write_float(f, color[3])


def read_vector3d(f):
    x = read_float(f)
    y = read_float(f)
    z = read_float(f)
    return (x, y, z)


def write_vector3d(f, vector):
    write_float(f, vector[0])
    write_float(f, vector[1])
    write_float(f, vector[2])


def read_matrix(f):
    m = []
    for i in range(4):
        row = [
            read_float(f),
            read_float(f),
            read_float(f),
            read_float(f)]

        m.append(row)

    return m


def write_matrix(f, matrix):
    for row in matrix:
        write_float(f, row[0])
        write_float(f, row[1])
        write_float(f, row[2])
        write_float(f, row[3])


def read_quaternion(f):
    w = read_float(f)
    x = read_float(f)
    y = read_float(f)
    z = read_float(f)
    return (w, x, y, z)


def write_quaternion(f, quaternion):
    write_float(f, quaternion[0])
    write_float(f, quaternion[1])
    write_float(f, quaternion[2])
    write_float(f, quaternion[3])


def read_bool(f):
    return bool(read_uint32(f))


def write_bool(f, boolean):
    write_uint32(f, int(boolean))
