#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = int.from_bytes(rng.read(2), byteorder='big', signed=False) % 65536 + 1
    h = int.from_bytes(rng.read(2), byteorder='big', signed=False) % 65536 + 1
    # Use a random color table size between 0-7
    ctsize = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 8
    # Write LSD to output
    out.write(struct.pack('>HHBBB', w, h, 0xF7 | ((ctsize & 0x07) << 4), 0, 0))

    # 3. Global Color Table
    if ctsize > 0:
        for i in range(pow(2, ctsize)):
            r = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
            g = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
            b = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
            out.write(struct.pack('BBB', r, g, b))

    # 4. Image Descriptor
    # Random image left position
    il = int.from_bytes(rng.read(2), byteorder='big', signed=False) % 65536
    # Random image top position
    it = int.from_bytes(rng.read(2), byteorder='big', signed=False) % 65536
    # Random image width
    iw = int.from_bytes(rng.read(2), byteorder='big', signed=False) % 65536 + 1
    # Random image height
    ih = int.from_bytes(rng.read(2), byteorder='big', signed=False) % 65536 + 1
    # Set local color table flag as necessary
    lctflag = bool(int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x80)
    # Set interlace flag as necessary
    interlaced = bool(int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x40)
    # Set sort flag as necessary
    sorted = bool(int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x20)
    # Set reserved flags to zero
    reserved = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x1C
    # Set local color table size as necessary
    lctsize = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x07
    # Write ID to output
    out.write(struct.pack('>HHBBBBB', il, it, iw, ih, 0xF7 | (((lctflag & 0x01) << 7) | ((interlaced & 0x01) << 6) | ((sorted & 0x01) << 5) | (reserved << 2)), lctsize))

    # 5. Local Color Table
    if lctflag:
        for i in range(pow(2, lctsize)):
            r = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
            g = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
            b = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
            out.write(struct.pack('BBB', r, g, b))

    # 6. Graphics Control Extension
    # Set block size to 0x04
    bs = 0x04
    # Set transparency flag as necessary
    transflag = bool(int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x01)
    # Set delay time to a random value between 0-65535
    dt = int.from_bytes(rng.read(2), byteorder='big', signed=False) % 65536
    # Set transparent index to a random value between 0-255
    ti = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
    # Write GCE to output
    out.write(struct.pack('BBBBHB', 0x21, 0xF9, bs, 0x04, dt, ti))
    # If there is a global color table but no local one, set the
    # background color index to a random value between 0-255
    if not lctflag and ctsize > 0:
        bc = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        out.write(struct.pack('B', bc))
    # Write end of extension marker
    out.write(b'\x00')

    # 7. Image Data
    # Set subblock size to 