#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small to avoid too much data in one block
    while len(b"".join(blocks)) < total_blocks_size:
        block_size = min((total_blocks_size - len(b"".join(blocks))), 255)
        
        # Generate some random data
        data = [rng.read(1)[0] for _ in range(block_size)]
        
        # Compress it with LZW
        compressed_data = compress_lzw(data, lzw_minimum_code_size)
        
        # Add the compressed data as a new block
        blocks.append(compressed_data + bytes([0]))
        
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)
    out.write(b"".join(blocks))
    out.write(trailer)


# Compresses an array of bytes using LZW compression
def compress_lzw(data: list[int], minimum_code_size):
    code_size = minimum_code_size + 1
    clear_code = 1 << minimum_code_size
    end_of_information_code = clear_code + 1
    next_code = end_of_information_code + 1
    
    dictionary = {bytes([datum]): datum for datum in set(data)}
    codes = sorted(dictionary.keys())
    
    result = []
    current_byte = bytes()
    current_code = 0
    
    def add_to_result():
        nonlocal current_byte, current_code
        
        if current_code == clear_code:
            result.extend([current_code >> 8, current_code & 0xFF])
            
            current_byte = bytes()
            current_code = 0
        elif current_code == end_of_information_code:
            pass
        else:
            result.extend([current_byte, current_code])
            
        return True
    
    try:
        while True:
            datum = data.pop(0)
            current_byte += bytes([datum])
            current_code = dictionary.get(current_byte, next_code)
            
            if add_to_result():
                del dictionary[codes[0]]
                
                codes.remove(codes[0])
                
            if current_code >= next_code:
                dictionary[current_byte] = next_code
                codes.append(current_byte)
                    
                next_code <<= 1
                code_size += 1
                
                if code_size > 12:
                    raise Exception("Code size exceeded")
                    
            if not data:
                break
                
        add_to_result()
            
    except IndexError:
        pass
    
    return bytes(result)
    
