#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng):
    header = b'GIF89a'

    # Logical Screen Descriptor (LSD)
    width  = rng.randint(100, 256)
    height = rng.randint(100, 256)
    flags  = 0b1000 | rng.randint(0, 7)   # Global Color Table Flag set
    bg     = rng.randint(0, 256)          # Background Color Index
    aspect = 4                            # Pixel Aspect Ratio
    lsd = bytes([width, height]) \
         + struct.pack('B', flags)        \
         + bytes([bg, aspect])            \
         + b'\x00\x00'                   # No extra data

    # Global Color Table (GCT)
    colors = [rng.randint(0, 256),
              rng.randint(0, 256)]
    gct = bytearray()
    for i in range(len(colors)):
        r = rng.randint(0, 256)
        g = rng.randint(0, 256)
        b = rng.randint(0, 256)
        gct += bytes([r, g, b])
    while len(gct) % 3!= 0:
        gct.append(0)                    # Pad to multiple of three
    assert len(gct) == 768              # GCT must be exactly this size
    color_table = bytes([len(colors)]) + gct

    # Graphics Control Extension (GCE)
    gce = None
    if rng.randint(0, 1)!= 0:           # Add GCE with delay between frames
        gce = b'\x21\xF9\x04'             # Application extension block
        gce += struct.pack('<H', rng.randint(100, 1000)) # Delay time
        gce += b'\x00'                     # Transparent color index
        gce += b'\x00'                     # Block terminator

    # Image Descriptor (ID)
    image_descriptor = b'\x2C'           # Image separator
    image_descriptor += bytes([0, 0,      # Left
                              0, 0,      # Top
                              width,      # Width
                              height])   # Height
    local_color_table = False
    if rng.randint(0, 1)!= 0 and len(colors) > 0: # Local Color Table flag
        local_color_table = True
        image_descriptor += bytes([flags & 0b1111 << 4 | len(colors)])
    else:
        image_descriptor += b'\x00'       # No local color table
    image_descriptor += b'\x00'           # Interlace flag off
    image_descriptor += b'\x00'           # Sort flag off
    image_descriptor += b'\x00'           # Reserved
    image_descriptor += b'\x00'           # LZW minimum code size

    # Image Data
    image_data = b''
    for y in range(height):
        for x in range(width):
            image_data += bytes([rng.randint(0, 256)])

    # Trailer
    trailer = b'\x3B'                   # File terminator

    # Write everything to output stream
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)