#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 7 + 1
    global_color_table = [rng.read(3) for _ in range(global_color_table_size)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HBBBBB', 0, 0, 0, 0, 0, 0)

    # 5. Local Color Table
    local_color_table_flag = bool(rng.read(1)[0] % 2)
    if not local_color_table_flag:
        local_color_table_size = rng.read(1)[0] & 7 + 1
        local_color_table = [rng.read(3) for _ in range(local_color_table_size)]

    # 6. Image Data
    data_subblock_size = rng.read(1)[0] & 7 + 1
    data_subblocks = []
    while True:
        subblock_type = rng.read(1)[0]

        if subblock_type == 0x2C:
            break
        
        elif subblock_type!= 0x21:
            raise Exception("Invalid subblock type")
        
        else:
            data_subblocks += [struct.pack('>B', subblock_type)]

            size = rng.read(1)[0]
            
            if size > 0:
                data_subblocks[-1] += rng.read(size)
            
            else:
                while True:
                    data_subblocks[-1] += rng.read(data_subblock_size)

                    if len(data_subblocks[-1]) >= 2 or data_subblocks[-1][-1:] == b'\x00':
                        break

    # 7. Trailer
    trailer = b';'

    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(b''.join([bytes([len(c)])+c for c in global_color_table]))
    out.write(image_descriptor)
    if local_color_table_flag:
        out.write(b''.join([bytes([len(c)])+c for c in local_color_table]))
    out.write(b''.join(data_subblocks))
    out.write(trailer)

