#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = [rng.read(3) for _ in range(num_colors)]
    for rgb in color_table:
        out.write(rgb)

    # 4. Graphics Control Extension (optional)
    has_gce = bool((rng.read(1)[0] & 1 << 7) >> 7)
    if not has_gce:
        return

    out.write(b'\x21\xf9\x04')
    transparent_color_flag = bool((rng.read(1)[0] & 1 << 7) >> 7)
    user_input_flag = bool((rng.read(1)[0] & 1 << 6) >> 6)
    disposal_method = rng.read(1)[0] & 0x07
    delay_time = struct.unpack('>H', rng.read(2))[0]
    transparent_color_index = rng.read(1)[0]
    out.write(struct.pack('>BBHB', transparent_color_flag, user_input_flag, disposal_method, delay_time))
    if transparent_color_flag:
        out.write(bytes([transparent_color_index]))

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool((rng.read(1)[0] & 1 << 7) >> 7)
    interlace_flag = bool((rng.read(1)[0] & 1 << 6) >> 6)
    sort_flag = bool((rng.read(1)[0] & 1 << 5) >> 5)
    reserved = rng.read(2)
    lct_size = rng.read(1)[0]
    lct = []
    if local_color_table_flag:
        for i in range(lct_size):
            lct += [rng.read(3)]
        for rgb in lct:
            out.write(rgb)

    # 6. Image Data
    image_data = b''
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x2c:
            # Image descriptor
            pass
        elif block_type == 0x21:
            # Extension
            extension_label = rng.read(1)[0]

            if extension_label == 0xff:
                # Application Extension
                application_identifier = rng.read(8)

                if application_identifier == b'NETSCAPE2.0':
                    subblock_size = rng.read(1)[0]

                    if subblock_size!= 3:
                        raise Exception("Invalid NETSCAPE2.0 subblock")
                    
                    subblock_id = rng.read(1)[0]
                    loop_count = struct.unpack('>H', rng.read(2))[0]
                    out.write(application_identifier + bytes([subblock_size]) + bytes([subblock_id]) + struct.pack('>H', loop_count))
                else:
                    subblock_size = rng.read(1)[0]
                    data = rng.read(subblock_size)
                    out.write(application_identifier + bytes([subblock_size]) + data)
            else:
                # Plain old extension
                label = rng.read(1)[0]
                length = rng.read(1)[0]
                data = rng.read(length)
                out.write(bytes([extension_label]) + bytes([label]) + bytes([length]) + data)
        elif block_type == 0x3b:
            break
        else:
            raise Exception("Unknown block type")
        
        image_data += block_type + rng.read(1)
    
    out.write(image_data)