#!/usr/bin/env python3

import struct
from typing import BinaryIO

def generate_image(out: BinaryIO, rng: BinaryIO):
    """Generate an LZW encoded PNG image."""

    # 1. Header
    header = b'\x89PNG\r\n\x1A\n'

    # 2. Logical Screen Descriptor
    width = rng.read(1)[0] & 0b11111111 | ((rng.read(1)[0] & 0b111) << 8)
    height = rng.read(1)[0] & 0b11111111 | ((rng.read(1)[0] & 0b111) << 8)
    bitdepth = rng.read(1)[0] & 0b11111
    colortype = rng.read(1)[0] & 0b111
    compressionmethod = rng.read(1)[0] & 0b11111111
    filtermethod = rng.read(1)[0] & 0b11111111
    interlacemethod = rng.read(1)[0] & 0b111

    lsd = struct.pack('>IIBBBBBBBB',
        0x0D0E1100, width, height, bitdepth, colortype, compressionmethod,
        filtermethod, interlacemethod)

    # 3. Color Table
    if colortype == 3 and bitdepth <= 8:
        # Grayscale palette
        numcolors = 1 << bitdepth

        grayscale_palette = [
            i * 255 // (numcolors - 1)
            for i in range(numcolors)]
        color_table = bytes()
        for r, g, b in zip(*[iter(grayscale_palette)]*3):
            color_table += struct.pack('>BBB', r, g, b)
    elif colortype == 2 and bitdepth <= 8:
        # RGB palette
        numcolors = 1 << bitdepth

        rgb_palette = [
            255 * i // (numcolors - 1)
            for i in range(numcolors)]
        color_table = bytes()
        for r, g, b in zip(*[iter(rgb_palette)]*3):
            color_table += struct.pack('>BBB', r, g, b)
    else:
        raise ValueError("Unsupported combination of colortype {} and bitdepth {}".format(colortype, bitdepth))

    # 4. Graphic Control Extension
    gce = struct.pack('>BBHBBBBBHB',
        0x21, 0xF9, 4, 1, 0, 0, 0, 0, 0xFF, 0x00)

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)