#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (if present)
    global_color_table_present = bool(packed_fields & 0b10000000)
    global_color_table_size = ((packed_fields >> 7) & 0b111) + 1
    global_color_table = rng.read(global_color_table_size * 3) if global_color_table_present else b''

    # 4. Application Extension Block (if present)
    application_extension_present = bool(packed_fields & 0b10000000)
    application_identifier = b'NETSCAPE2.0' if application_extension_present else b''
    application_auth_code = b'' if application_extension_present else b''
    application_data = b'\x03\x01\x00\x00' if application_extension_present else b''
    ae = struct.pack('>B8s3s11s',
                     len(application_identifier),
                     application_identifier,
                     application_auth_code,
                     application_data
                    ) \
         if application_extension_present else b''

    # 5. Graphics Control Extension (if present)
    gce_present = bool(packed_fields & 0b10000000)
    delay_time = rng.read(2) if gce_present else b''
    transparent_color_index = rng.read(1) if gce_present else b''
    terminator = b'\x00' if gce_present else b''
    gce = struct.pack('>BHB2s1s',
                     0xF9,
                     delay_time[0],
                     delay_time[1],
                     transparent_color_index,
                     terminator
                    ) \
          if gce_present else None

    # 6. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved_flags = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = rng.read(local_color_table_size * 3)
    image_descriptor = struct.pack('>BHHHHBBBBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  local_color_table_size
                                 )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    for _ in range(total_blocks_size):
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')