#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a' + bytes([0x00] * (7))

    # 2. Logical screen descriptor
    width = rng.read(1)[0] % 64 + 1
    height = rng.read(1)[0] % 64 + 1
    color_depth = rng.read(1)[0] & 0b1111
    global_color_table_flag = bool(rng.read(1)[0] & 0b10000000)
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] & 0b111

    logical_screen_descriptor = bytearray()
    logical_screen_descriptor += struct.pack('>H', width)
    logical_screen_descriptor += struct.pack('>H', height)
    logical_screen_descriptor += struct.pack('B', color_depth)
    logical_screen_descriptor += struct.pack('B', \
            int(global_color_table_flag) << 7 | background_color_index)
    logical_screen_descriptor += struct.pack('B', pixel_aspect_ratio)

    # 3. Global color table
    if global_color_table_flag:
        for i in range(pow(2, color_depth)):
            r = rng.read(1)[0] % 256
            g = rng.read(1)[0] % 256
            b = rng.read(1)[0] % 256

            logical_screen_descriptor += struct.pack('BBB', r, g, b)

    # 4. Image descriptor
    image_separator = b'\x2C'
    left = rng.read(1)[0] % width
    top = rng.read(1)[0] % height
    width = rng.read(1)[0] % (width - left) + 1
    height = rng.read(1)[0] % (height - top) + 1
    local_color_table_flag = bool(rng.read(1)[0] & 0b10000000)
    interlace_flag = bool(rng.read(1)[0] & 0b10000000)
    sort_flag = bool(rng.read(1)[0] & 0b10000000)
    reserved = rng.read(1)[0] & 0b1111111
    lct_size = pow(2, rng.read(1)[0] & 0b11111)

    image_descriptor = bytearray()
    image_descriptor += struct.pack('B', left)
    image_descriptor += struct.pack('B', top)
    image_descriptor += struct.pack('B', width)
    image_descriptor += struct.pack('B', height)
    image_descriptor += struct.pack('B', \
            int(local_color_table_flag) << 7 | \
            int(interlace_flag) << 6 | \
            int(sort_flag) << 5 | \
            reserved)
    image_descriptor += struct.pack('B', lct_size)

    # 5. Local color table
    if local_color_table_flag:
        for i in range(lct_size):
            r = rng.read(1)[0] % 256
            g = rng.read(1)[0] % 256
            b = rng.read(1)[0] % 256

            image_descriptor += struct.pack('BBB', r, g, b)

    # 6. LZW minimum code size
    min_code_size = rng.read(1)[0] % 256

    # 7. Data sub-blocks
    data = bytearray()
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x00: break

        elif block_type == 0x01:
            # Plain text extension
            label = rng.read(1)[0] % 256
            block_size = rng.read(1)[0] % 256

            data += struct.pack('BB', label, block_size)
            data += bytes(block_size)

        else:
            raise Exception("Invalid block type")

    # 8. Trailer
    trailer = b'\x3B'

    # Write everything to output stream
    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(image_separator)
    out.write(image_descriptor)
    out.write(struct.pack('B', min_code_size))
    out.write(data)
    out.write(trailer)

