#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0] & 1)   # Only use if we're going to actually write some frames
    gce = b''
    if has_gce:
        gce += b'\x21\xF9\x04'          # Application extension block

        # Use random values for duration, transparent color index, and flags
        gce += struct.pack('>HBB', rng.read(2)[0], rng.read(1)[0], rng.read(1)[0])

        # Write the application data subblock
        gce += b'\x01\x00'              # Size of the application data subblock
        gce += b'NETSCAPE2.0'           # Application ID
        gce += b'\x03'                  # Sub-block size
        gce += b'\x01'                  # Loop count subfield
        gce += b'\x00'                  # End of subfields

    # 5. Image Descriptor
    id_ = b'\x2C'                      # Image separator
    id_ += struct.pack('>HHBBB',       # Left position, top position, width, height
                        rng.read(2)[0], rng.read(2)[0], rng.read(1)[0], rng.read(1)[0],
                        rng.read(1)[0])    # Flags

    # 6. Local Color Table (optional)
    has_lct = bool(rng.read(1)[0] & 1)   # Only use if we're going to actually write some frames
    lct = b''
    if has_lct:
        lct += b'\x2c\x00\x00'          # Local color table
        lct += struct.pack('B', num_colors)     # Number of colors in local color table
        lct += color_table               # Colors themselves

    # 7. Frames
    frame_data = b''                    # Data for all frames
    num_frames = rng.read(1)[0]         # How many frames?
    for _ in range(num_frames):
        # Each frame is made up of an image descriptor, followed by the actual frame data
        frame_data += id_ + lct + gce + b'\x00'

        # The frame data itself consists of a series of blocks that can be one of three types:
        # - Plain text (ASCII characters)
        # - Compressed text (LZW compression)
        # - Raw bitmap data
        # We'll pick one randomly for this frame
        block_type = rng.read(1)[0]      # Pick a type at random
        if block_type == 0:
            # Plain text

            # Generate a random string of ASCII characters with length between 1 and 128
            str_len = rng.read(1)[0] % 128 + 1
            str_ = rng.read(str_len).decode()

            # Encode as plain text
            frame_data += struct.pack('B', str_len) + str_.encode()

        elif block_type == 1:
            # Compressed text

            # Generate a random string of ASCII characters with length between 1 and 128
            str_len = rng.read(1)[0] % 128 + 1
            str_ = rng.read(str_len).decode()

            # Compress the string using LZW
            compressed = []                 # List of codes
            code_size = 2                   # Initial code size
            clear_code = 1 << code_size     # Clear code
            end_of_info = clear_code + 1    # End of information code
            next_code = clear_code + 2      # Next available code
            dict_ = {                       # Dictionary
                '': clear_code,
                '\n': end_of_info
            }                              # Initialize dictionary with initial strings

            # Add every character from our random string to the dictionary
            for char in str_:
                if not char in dict_:
                    dict_[char] = next_code
                    next_code += 1

                    # If we've reached the maximum code size, increase it
                    if len(dict_) >= (1 << code_size):
                        code_size += 1

                # Append the current character's code to the list
                compressed.append(dict_[char])

            # Terminate the stream with the end of information code
            compressed.append(end_of_info)

            # Convert the list of codes to binary
            compressed = [struct.pack('B', x) for x in compressed]

            # Output the compressed data
            frame_data += struct.pack('B', code_size)        # Code size used
            frame_data += struct.pack('B', clear_code)       # Clear code
            frame_data += struct.pack('B', end_of_info)      # End of info code
            frame_data += struct.pack('B', len(compressed))  # Length of compressed data
            frame_data += b''.join(compressed)                # Compressed data

        else:
            # Raw bitmap data

            # Generate a random palette entry for each pixel
            pixels = [(rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
                      for _ in range((width * height))]

            # Pack them into raw RGB data
            packed = b''.join([b'\x00'.join(pixel)
                               for pixel in pixels])

            # Output the packed data
            frame_data