#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), byteorder='little') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg_color_index = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    pixel_aspect_ratio = bytes([int.from_bytes(rng.read(1), byteorder='little')])

    lsd = \
        width.to_bytes(2, 'little', signed=False) + \
        height.to_bytes(2, 'little', signed=False) + \
        flags + \
        bg_color_index + \
        pixel_aspect_ratio

    # 3. Global Color Table
    global_color_table_size = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    global_color_table_sort = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    global_color_table = []
    for i in range(global_color_table_size[0]):
        r = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        g = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        b = bytes([int.from_bytes(rng.read(1), byteorder='little')])

        global_color_table += [r, g, b]

    # 4. Image Descriptor
    image_left = int.from_bytes(rng.read(2), byteorder='little') % (width - 1)
    image_top = int.from_bytes(rng.read(2), byteorder='little') % (height - 1)
    image_width = int.from_bytes(rng.read(2), byteorder='little') % (width - image_left) + 1
    image_height = int.from_bytes(rng.read(2), byteorder='little') % (height - image_top) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='little')])

    id = \
        image_left.to_bytes(2, 'little', signed=False) + \
        image_top.to_bytes(2, 'little', signed=False) + \
        image_width.to_bytes(2, 'little', signed=False) + \
        image_height.to_bytes(2, 'little', signed=False) + \
        flags

    # 5. Local Color Table
    local_color_table_flag = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    if local_color_table_flag == b'\x01':
        local_color_table_size = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        local_color_table_sort = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        local_color_table = []
        for i in range(local_color_table_size[0]):
            r = bytes([int.from_bytes(rng.read(1), byteorder='little')])
            g = bytes([int.from_bytes(rng.read(1), byteorder='little')])
            b = bytes([int.from_bytes(rng.read(1), byteorder='little')])

            local_color_table += [r, g, b]

    # 6. Image Data
    data = b''
    while len(data) < ((image_width * image_height) // 8):
        data += rng.read((image_width * image_height) // 8 - len(data))

    # 7. Trailer
    trailer = b';'

    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(b''.join(global_color_table))
    out.write(id)
    if local_color_table_flag == b'\x01':
        out.write(b''.join(local_color_table))
    out.write(data)
    out.write(trailer)

