#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor (LSD)
    width = rng.read(1)[0] % 64 + 1
    height = rng.read(1)[0] % 64 + 1
    flags = rng.read(1)[0] & 0b00000111 | 0b10000000
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = 0
    lsd = struct.pack('<HHBBB', width, height, flags, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (GCT)
    global_color_table_flag = bool(flags & 0b10000000)
    if global_color_table_flag:
        global_color_table_size = ((flags >> 7) & 0b111) + 1
        global_color_table = [rng.read(1)[0] for _ in range(global_color_table_size)]

        while len(global_color_table)!= 2 ** global_color_table_size:
            global_color_table += [0] * (2 ** global_color_table_size - len(global_color_table))
    else:
        global_color_table = []

    # 4. Graphics Control Extension (GCE)
    gce = None
    if rng.read(1)[0] > 10:
        gce = b''
        gce += struct.pack('BB', 0xF9, 4)   # Application extension
        gce += struct.pack('BBBB', *(rng.read(4)))     # Block Size and Flags
        gce += struct.pack('H', rng.read(2)[0] % 65536)       # Delay Time
        gce += struct.pack('BBB', *(rng.read(3)))      # Transparent Color Index, Disposal Method, User Input Flag
        gce += b'\x00'        # Terminator

    # 5. Image Descriptor
    image_descriptor = struct.pack('<BHBHHB', 0x2C, 0, 0, width, height, 0)

    # 6. Image Data
    lzw_minimum_code_size = rng.read(1)[0] % 12 + 1
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)