#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
        
    # 3. Global Color Table
    # Number of colors in global color table
    num_colors = rng.read(1)[0] & 7 + 1
    
    # Colors are stored as RGB triplets
    colors = []
    for _ in range(num_colors):
        r, g, b = [x % 256 for x in struct.unpack('BBB', rng.read(3))]
        colors += [(r,g,b)]

    # 4. Image Descriptor
    # Left position is always zero
    left = struct.pack('<H', 0)
    
    # Top position is always zero
    top = struct.pack('<H', 0)
    
    # Width and Height are random again
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Flags: packed field with various bits set
    flags = rng.read(1)[0]
    
    # 5. Local Color Table
    # Same as above but this time we use it to store the actual image data
    num_colors = rng.read(1)[0] & 7 + 1
    
    # Colors are stored as RGB triplets
    colors = []
    for _ in range(num_colors):
        r, g, b = [x % 256 for x in struct.unpack('BBB', rng.read(3))]
        colors += [(r,g,b)]

    # 6. Image Data
    # The image data is just a bunch of pixels that reference one of our randomly generated colors
    imgdata = b''
    for y in range(height):
        for x in range(width):
            imgdata += colors[rng.read(1)[0]]
            
    # 7. Trailer
    trailer = b';'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    for c in colors:
        out.write(struct.pack('BBB', *c))
    out.write(imgdata)
    out.write(trailer)

