#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = rng.read(2)[::-1]
    height = rng.read(2)[::-1]
    packed = rng.read(1)[0]
    global_color_table_flag = bool((packed >> 7) & 1)
    color_resolution = ((packed >> 4) & 7) + 1
    sort_flag = bool((packed >> 3) & 1)
    global_color_table_size = 2 ** (((packed >> 0) & 7) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HBBBBB', width[0], height[0], packed, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    color_table = bytearray()
    for _ in range(global_color_table_size):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        color_table += struct.pack('BBB', r, g, b)

    # 4. Graphics Control Extension
    has_gce = rng.read(1)[0] > 0
    if has_gce:
        gce = struct.pack('>BHBHB', 0xF9, 4, rng.read(2)[0], rng.read(2)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)