#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF will have a random resolution between [min_width, max_width] x [min_height, max_height].
# If `include_animation`, then an animation extension will also be included in the generated GIF.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, min_width=1, max_width=1000, min_height=1, max_height=1000, include_animation=True
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width = rng.randint(min_width, max_width)
    height = rng.randint(min_height, max_height)

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if include_animation and rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    left = rng.read(1)[0]
    top = rng.read(1)[0]
    width = rng.read(1)[0]
    height = rng.read(1)[0]
    local_color_table_flag = rng.read(1)[0]
    interlace_flag = rng.read(1)[0]
    sort_flag = rng.read(1)[0]
    reserved = rng.read(1)[0]
    image_descriptor = struct.pack(
        '>BHHHHBBBBB',
        0x2C,
        left,
        top,
        width,
        height,
        local_color_table_flag,
        interlace_flag,
        sort_flag,
        reserved
    )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)