#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF will be of dimensions between [min_width, max_width] x [min_height, max_height].
# The background color is chosen randomly from the given list of possible colors.
# The global color table has a maximum size of 256 colors.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, min_width=1, max_width=1000, min_height=1, max_height=1000, bg_colors=[], num_colors=None):

    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width = rng.randint(min_width, max_width)
    height = rng.randint(min_height, max_height)

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.randrange(0, 8) | ((rng.randrange(0, 256) & 0x07) << 4)
    background_color_index = rng.randrange(len(bg_colors)) if bg_colors else None
    pixel_aspect_ratio = rng.randrange(0, 8)

    lsd = struct.pack(">HHBBB", width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = num_colors or (rng.randrange(0, 256) or 1)
    color_table = b"".join([struct.pack("BBB", rng.randrange(0, 256), rng.randrange(0, 256), rng.randrange(0, 256))
                            for _ in range(num_colors)])

    # 4. Graphics Control Extension (optional)
    if rng.randrange(0, 2):  # If we want one, then add an extension
        gce = struct.pack(">BBBBBB", 0x21, 0xF9, 0x04,
                          rng.randrange(0, 256), rng.randrange(0, 256), rng.randrange(0, 256))
    else:
        gce = b""

    # 5. Image Descriptor
    left = rng.randrange(0, width - 1)
    top = rng.randrange(0, height - 1)
    width = rng.randrange(1, width - left)
    height = rng.randrange(1, height - top)
    local_color_table_flag = bool(rng.randrange(0, 2))
    interlace_flag = bool(rng.randrange(0, 2))
    sort_flag = bool(rng.randrange(0, 2))
    reserved = rng.randrange(0, 2)
    local_color_table_size = rng.randrange(0, 2 ** int(local_color_table_flag))
    local_color_table = b"".join([struct.pack("BBB", rng.randrange(0, 256), rng.randrange(0, 256), rng.randrange(0, 256))
                                  for _ in range(local_color_table_size)]) \
                        if local_color_table_flag else b""

    image_descriptor = struct.pack(">BHHHHBBB",
                                  0x2C, left, top, width, height, local_color_table_flag, interlace_flag, sort_flag,
                                  reserved)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = rng.randrange(0, 8) or 1

    blocks = []
    while True:
        block_length = rng.randrange(0, 255) + 1

        # We need to read exactly as many bytes as specified in the block length
        blocks.append(rng.read(block_length))

        # If the last block was less than 255 bytes long, stop reading
        if len(blocks[-1]) < 255:
            break

    # 7. Trailer
    trailer = b"\x3b"

    # Write everything together
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack("B", len(block)))
        out.write(block)
    out.write(trailer)