#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated file will be of the form:
#   - A valid GIF header
#   - A random logical screen descriptor with a random width and height
#   - A global color table with a random number of entries between [1; 256]
#   - An optional graphics control extension with a random delay time
#     and transparent color index
#   - A random image descriptor with a random position, width and height
#   - A series of blocks containing random data of varying sizes
#   - A trailer
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety),
    # background color index, pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height,
                      packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        # Add an extension
        out.write(b'\x21\xf9\x04')

        # Random delay time and transparent color index
        delay_time = rng.read(2)
        transparent_color_index = rng.read(1)[0]
        gce = struct.pack('>BBHBB',
                          int(has_gce),
                          0x04,
                          int.from_bytes(delay_time, 'big'),
                          transparent_color_index,
                          0)
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = rng.read(1)[0] & 0x80
    interlace_flag = rng.read(1)[0] & 0x40
    sort_flag = rng.read(1)[0] & 0x20
    reserved_bits = rng.read(1)[0] & 0x70
    local_color_table_size = rng.read(1)[0] & 0x07
    image_descriptor = struct.pack('>BHHHHBBBB',
                                  0x2c,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag | interlace_flag | sort_flag | reserved_bits,
                                  local_color_table_size)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    while True:
        block_length = (rng.read(1)[0] % 255) + 1

        # We need to read exactly as many bytes as specified in the block length
        blocks.append(rng.read(block_length))

        # If the last block was less than 255 bytes long, stop reading
        if len(blocks[-1]) < 255:
            break

    # 7. Trailer
    out.write(b';')

    # Write everything together
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(b'\x3b')