#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF will have a random width and height between [min_width; max_width] x [min_height; max_height].
# The background color will be randomly chosen from the global color table.
# If `has_transparent_color` is true, then there will be a random chance that the first color in the global color table
# will be marked as transparent.
# The pixel aspect ratio will always be set to 0.
# The amount of frames will be randomly chosen between [min_frames; max_frames].
# For each frame, the position of the frame on the canvas will be randomly chosen within the bounds of the canvas.
# The frame will also have a random duration between [min_duration; max_duration] milliseconds.
# Finally, the frame will contain a random amount of LZW compressed data, which will be made of random blocks of
# up to 255 bytes.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, min_width=1, max_width=100, min_height=1, max_height=100, has_transparent_color=True,
        min_frames=1, max_frames=10, min_duration=100, max_duration=1000
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width = rng.randint(min_width, max_width)
    height = rng.randint(min_height, max_height)

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0] | ((rng.read(1)[0] & 0b111) << 8)  # Use only the lower bits of the first byte
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = 0
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if has_transparent_color and rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptors
    descriptors = []
    for _ in range(rng.randint(min_frames, max_frames)):
        left = rng.randint(0, width - 1)
        top = rng.randint(0, height - 1)

        width = rng.randint(1, width - left)
        height = rng.randint(1, height - top)

        local_packed_fields = rng.read(1)[0]
        local_background_color_index = rng.read(1)[0]
        local_pixel_aspect_ratio = 0

        descriptor = struct.pack(
            '>BHHHHB',
            0x2C,
            left,
            top,
            width,
            height,
            local_packed_fields
        )

        descriptors.append((left, top, width, height, descriptor))

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    for _, _, _, _, descriptor in descriptors:
        out.write(descriptor)
    out.write(image_data + trailer)