#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big', signed=False) + 1
    height = int.from_bytes(rng.read(2), 'big', signed=False) + 1
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table (optional)
    if rng.read(1)[0]:
        color_table_size = 1 << ((rng.read(1)[0] & 7) + 1)
        for _ in range(color_table_size * 3):
            out.write(rng.read(1))

    # 4. Image Data
    while True:
        # 4.1 Graphic Control Extension
        if rng.read(1)[0]:
            block_size = rng.read(1)[0]
            assert block_size == 4
            disposal_method = rng.read(1)[0] >> 2
            user_input_flag = bool((rng.read(1)[0] & 2) != 0)
            transparent_color_flag = bool((rng.read(1)[0] & 1) != 0)
            delay_time = int.from_bytes(rng.read(2), 'big', signed=False)
            transparent_color_index = rng.read(1)[0]
            block_terminator = rng.read(1)[0]
            assert block_terminator == 0

        # 4.2 Image Descriptor
        image_separator = rng.read(1)[0]
        assert image_separator == 0x2c
        left = int.from_bytes(rng.read(2), 'big', signed=False)
        top = int.from_bytes(rng.read(2), 'big', signed=False)
        width = int.from_bytes(rng.read(2), 'big', signed=False)
        height = int.from_bytes(rng.read(2), 'big', signed=False)
        local_color_table_flag = bool((rng.read(1)[0] & 128) != 0)
        interlace_flag = bool((rng.read(1)[0] & 64) != 0)
        sort_flag = bool((rng.read(1)[0] & 32) != 0)
        size_of_local_color_table = (rng.read(1)[0] & 7) + 1

        # 4.3 Local Color Table (optional)
        if local_color_table_flag:
            for _ in range(size_of_local_color_table * 3):
                out.write(rng.read(1))

        # 4.4 LZW Minimum Code Size
        lzw_minimum_code_size = rng.read(1)[0]

        # 4.5 Image Data
        while True:
            code_size = rng.read(1)[0]
            if code_size == 0:
                break
            else:
                for _ in range(int.from_bytes(rng.read(1), 'big', signed=False)):
                    pass
