#!/usr/bin/env python3

import struct
from typing import BinaryIO


def read_int(f: BinaryIO) -> int:
    return struct.unpack('<i', f.read(4))[0]


def write_int(f: BinaryIO, n: int):
    f.write(struct.pack('<i', n))


def read_long(f: BinaryIO) -> int:
    return struct.unpack('<q', f.read(8))[0]


def write_long(f: BinaryIO, n: int):
    f.write(struct.pack('<q', n))


def read_float(f: BinaryIO) -> float:
    return struct.unpack('<f', f.read(4))[0]


def write_float(f: BinaryIO, x: float):
    f.write(struct.pack('<f', x))


def read_double(f: BinaryIO) -> float:
    return struct.unpack('<d', f.read(8))[0]


def write_double(f: BinaryIO, x: float):
    f.write(struct.pack('<d', x))


def read_string(f: BinaryIO) -> str:
    length = read_short(f)
    if length == 0:
        return ''
    else:
        return f.read(length).decode()


def write_string(f: BinaryIO, s: str):
    write_short(f, len(s))
    f.write(s.encode())


def read_boolean(f: BinaryIO) -> bool:
    return read_byte(f) != 0


def write_boolean(f: BinaryIO, b: bool):
    write_byte(f, 1 if b else 0)


def read_byte(f: BinaryIO) -> int:
    return struct.unpack('<b', f.read(1))[0]


def write_byte(f: BinaryIO, n: int):
    f.write(struct.pack('<b', n))


def read_bytes(f: BinaryIO, count: int) -> bytes:
    return f.read(count)


def write_bytes(f: BinaryIO, data: bytes):
    f.write(data)


def read_short(f: BinaryIO) -> int:
    return struct.unpack('<h', f.read(2))[0]


def write_short(f: BinaryIO, n: int):
    f.write(struct.pack('<h', n))


def read_ushort(f: BinaryIO) -> int:
    return struct.unpack('<H', f.read(2))[0]


def write_ushort(f: BinaryIO, n: int):
    f.write(struct.pack('<H', n))


def read_char(f: BinaryIO) -> chr:
    return chr(read_ushort(f))


def write_char(f: BinaryIO, c: chr):
    write_ushort(f, ord(c))


def read_chars(f: BinaryIO, count: int) -> str:
    return ''.join([chr(read_ushort(f)) for _ in range(count)])


def write_chars(f: BinaryIO, chars: str):
    for c in chars:
        write_ushort(f, ord(c))


def read_array(f: BinaryIO, read_element: callable[[BinaryIO], object]):
    length = read_int(f)
    return [read_element(f) for _ in range(length)]


def write_array(f: BinaryIO, array: list, write_element: callable[[BinaryIO, object], None]):
    write_int(f, len(array))
    for e in array:
        write_element(f, e)


def read_map(f: BinaryIO, key_type: type, value_type: type):
    length = read_int(f)
    return {key_type(k): value_type(v) for k, v in zip(range(length), read_array(f, lambda f: read_int(f)))}


def write_map(f: BinaryIO, map: dict, key_type: type, value_type: type):
    write_int(f, len(map))
    write_array(f, [value_type(v) for _, v in sorted(map.items(), key=lambda i: i[0])], lambda f, v: write_int(f, v))


def read_list(f: BinaryIO, element_type: type):
    length = read_int(f)
    return [element_type(e) for e in read_array(f, lambda f: read_int(f))]


def write_list(f: BinaryIO, list: list, element_type: type):
    write_int(f, len(list))
    write