#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated file will have between 1 and 10 frames,
# where each frame has between 1 and 10 blocks, which are
# randomly sized from 1 to 255 bytes long.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHBBB', 0x2C, 0, 0, width, height, 0)

    # 6. Blocks of Data
    for _ in range(rng.read(1)[0] % 10):
        # Each block is a variable-length sequence of subblocks
        while True:
            # Read the length of the next subblock
            subblock_len = rng.read(1)[0]
            if not subblock_len:
                break

            # Read that many bytes of data
            image_data = rng.read(subblock_len)

            # Subsequent frames, only write the data portion
            out.write(image_descriptor + image_data)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(trailer)