#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed = rng.read(1)
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]

    # 3. Global Color Table
    color_table_size = (packed[0] & 7) + 1
    color_resolution = ((packed[0] >> 4) & 7) + 1
    sort_flag = bool((packed[0] >> 3) & 1)
    global_color_table = []
    for _ in range(3 * (2 ** color_resolution)):
        global_color_table.append(struct.unpack('B', rng.read(1))[0])

    # 4. Image Blocks
    image_blocks = []
    while True:
        block_type = rng.read(1)[0]
        if block_type == 0x3b:
            break
        elif block_type == 0x2c:
            # Image Descriptor
            left = rng.read(2)
            top = rng.read(2)
            width = rng.read(2)
            height = rng.read(2)
            packed = rng.read(1)
            local_color_table_size = (packed[0] & 7) + 1
            interlaced = bool((packed[0] >> 6) & 1)
            sorted = bool((packed[0] >> 5) & 1)
            reserved = bool((packed[0] >> 4) & 1)
            lct_present = bool((packed[0] >> 3) & 1)
            if lct_present:
                local_color_table = []
                for _ in range(3 * (2 ** local_color_table_size)):
                    local_color_table.append(struct.unpack('B', rng.read(1))[0])
            else:
                local_color_table = global_color_table

            # Image Data
            data = rng.read(width[0] * height[0])

            image_blocks.append({
                'left': left,
                'top': top,
                'width': width,
                'height': height,
                'local_color_table': local_color_table,
                'data': data,
            })
        else:
            raise Exception("Unknown block type")

    # 5. Trailer
    trailer = rng.read(1)

    # Write everything to the output stream
    out.write(header)
    out.write(width)
    out.write(height)
    out.write(packed)
    out.write(background.to_bytes(1, byteorder='big'))
    out.write(aspect_ratio.to_bytes(1, byteorder='big'))
    for color in global_color_table:
        out.write(color.to_bytes(1, byteorder='big'))
    for image_block in image_blocks:
        out.write(b'\x2C')
        out.write(image_block['left'])
        out.write(image_block['top'])
        out.write(image_block['width'])
        out.write(image_block['height'])
        out.write(packed)
        out.write(len(image_block['local_color_table']).to_bytes(1, byteorder='big'))
        out.write(bool(sorted).to_bytes(1, byteorder='big'))
        out.write(bool(reserved).to_bytes(1, byteorder='big'))
        out.write(bool(lct_present).to_bytes(1, byteorder='big'))
        for color in image_block['local_color_table']:
            out.write(color.to_bytes(1, byteorder='big'))
        out.write(image_block['data'])
    out.write(trailer)

