#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed field:
    #   - 1 bit: Global Color Table Flag
    #   - 3 bits: Color Resolution
    #   - 1 bit: Sort Flag
    #   - 3 bits: Size of Global Color Table
    packed_field = struct.pack('>B', rng.read(1)[0] & 0b11100000)
    
    # Background Color Index
    background_color_index = rng.read(1)
    
    # Pixel Aspect Ratio
    pixel_aspect_ratio = rng.read(1)
    
    # Put it all together
    lsd = struct.pack('>6sHBBBB', header, width, height, *struct.unpack('BBBB', packed_field + background_color_index + pixel_aspect_ratio))
    
    # 3. Global Color Table
    # We use a fixed palette of 256 colors
    color_table = rng.read(768)
    
    # 4. Graphic Control Extension
    # This is optional
    gce = None
    if rng.read(1)[0] % 2 == 0:
        # Block size
        gce = bytes([0x21, 0xf9, 0x04])
        
        # Disposal method
        disposal_method = rng.read(1)
        
        # User input flag
        user_input_flag = rng.read(1)
        
        # Transparent color index
        transparent_color_index = rng.read(1)
        
        # Delay time
        delay_time = rng.read(2)
        
        # Transparent color flag
        transparent_color_flag = rng.read(1)
        
        # Block terminator
        block_terminator = rng.read(1)
        
        gce += struct.pack('BBBBBB', *struct.unpack('BBBBBB', disposal_method + user_input_flag + transparent_color_index + delay_time + transparent_color_flag + block_terminator))
    
    # 5. Image Descriptor
    image_descriptor = b'\x2C'
    
    # Left Position
    left_position = rng.read(2)
    
    # Top Position
    top_position = rng.read(2)
    
    # Width
    image_width = rng.read(2)
    
    # Height
    image_height = rng.read(2)
    
    # Packed Field
    #   - 1 bit: Local Color Table Flag
    #   - 3 bits: Reserved
    #   - 1 bit: Interlace Flag
    #   - 3 bits: Sort Flag
    #   - 3 bits: Size of Local Color Table
    packed_field = struct.pack('>B', rng.read(1)[0] & 0b11100000)
    
    # LCT = Local Color Table
    # If there is no local color table, we just skip this part
    if packed_field[0] >> 7 != 0:
        # Number of entries in the table
        num_entries = 2 ** ((packed_field[0] & 0b00011100) >> 2)
        
        # Skip the first byte since it was already read
        color_table = rng.read(num_entries * 3)
    
    # Put it all together
    image_descriptor += struct.pack('>HHHHB', *struct.unpack('HHHHB', left_position + top_position + image_width + image_height + packed_field))
    
    # 6. Image Data
    # The minimum code size is between 2 and 9
    lzw_minimum_code_size = rng.read(1)[0] & 0b00011111 + 2
    
    # Read all the blocks until we reach one of size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        
        if block_size == 0:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)