#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated file will have between 1 and 10 frames,
# where each frame has between 1 and 10 blocks, which are
# randomly sized from 1 to 255 bytes long.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Application Extension
    app_ext_header = b'\x21\xFF\x0BNETSCAPE2.0\x03\x01'
    app_ext_data = struct.pack('>H', 0)  # Loop count

    # 5. Image Descriptors
    # We'll use a random number of frames, with a random number of blocks per frame
    num_frames = rng.read(1)[0] % 10 + 1
    image_descs = []
    for _ in range(num_frames):
        # Each frame has a random number of blocks
        num_blocks = rng.read(1)[0] % 10 + 1
        image_desc = struct.pack('>HHBB', 0, 0, width, height)
        for _ in range(num_blocks):
            # Each block is a random length
            block_len = rng.read(1)[0] % 256 + 1
            image_desc += struct.pack('B', block_len)
            image_desc += rng.read(block_len)  # Block data
        image_descs.append(image_desc)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(trailer)