#!/usr/bin/env python3

import struct
from typing import BinaryIO

def main():
    from sys import argv
    
    if len(argv) != 3:
        print("Usage: {} <input-file> <output-file>".format(argv[0]))
        exit()

    inp, out = open(argv[1], 'rb'), open(argv[2], 'wb')
    randomize(inp, out)


def randomize(rng: BinaryIO, out: BinaryIO):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color_index = (packed & 0b00001111)
    aspect_ratio = ((packed >> 4) & 0b00000011)
    global_color_table_flag = bool((packed >> 7) & 0b00000001)
    color_resolution = ((packed >> 4) & 0b00000011) * 8
    sort_flag = bool((packed >> 3) & 0b00000001)
    size_of_global_color_table = 2 ** ((packed & 0b00001111) + 1)
    lsd = struct.pack('>HHBBBBB', width, height, packed, background_color_index, aspect_ratio, global_color_table_flag, color_resolution, sort_flag, size_of_global_color_table)

    # 3. Global Color Table
    if global_color_table_flag:
        color_table = rng.read(3 * size_of_global_color_table)
    else:
        color_table = b''

    # 4. Graphic Control Extension
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        # The last block is always empty
        if not block_size:
            break
        
        # Add some data to this block
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)