#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table
    # Generate a random color table
    color_table = rng.read(3 * 256)
    
    # 4. Graphics Control Extension
    # Set the delay time to 0
    gce = struct.pack('>BBHB', 0xF9, 0x04, 0, 0)
    
    # 5. Image Descriptor
    # Random position and size (using two bytes for each)
    x, y = struct.unpack('>HH', rng.read(4))
    
    # Other ID fields: Packed fields (using a random byte for variety), Local color table flag, Interlace flag, Sort flag, Size of local color table
    id = struct.pack('>HHBHBB', x, y, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 6. Image Data
    # Generate a random LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0] & 0b11111
    
    # Generate a random number of data sub-blocks
    num_sub_blocks = rng.read(1)[0] & 0b11111
    
    # Generate the actual data sub-blocks
    blocks = []
    for i in range(num_sub_blocks):
        # Each sub-block is preceded by its length
        block_size = rng.read(1)[0] & 0b1111111
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)