#!/usr/bin/env python3

import struct
from typing import List, Tuple, Union

from . import utils
from .constants import (
    CURVE_P256,
    CURVE_SECP256K1,
    EC_KEY_TYPE_PRIVATE,
    EC_KEY_TYPE_PUBLIC,
)
from .crypto import ecdsa_sign, ecdsa_verify, hash_sha256
from .exceptions import InvalidSignatureError
from .types import (
    ECDSASignature,
    ExtendedKeyVersionBytes,
    KeyData,
    Network,
    PrivateKey,
    PublicKey,
)


def serialize(key: KeyData, network: Network = None) -> bytes:
    """Serialize a public or private key into its binary representation.

    Args:
        key (KeyData): The key to serialize.
        network (Network): If the key is a private key, include this to add the
            version and depth bytes to the serialized data.

    Returns:
        bytes: The serialized key.
    """
    if not isinstance(key, (PublicKey, PrivateKey)):
        raise TypeError("Invalid type for 'key'")

    if isinstance(key, PrivateKey):
        if network is None:
            raise ValueError("Must specify network when serializing private keys")

        return _serialize_private_key(key, network)

    return _serialize_public_key(key)


def deserialize(data: bytes) -> KeyData:
    """Deserialize a public or private key from its binary representation.

    Args:
        data (bytes): The serialized key.

    Returns:
        KeyData: The deserialized key.
    """
    if len(data) == 0:
        raise ValueError("Empty data")

    # Check for private key
    if data[0] in [0x04, 0x06, 0x07]:
        return _deserialize_private_key(data)

    # Check for public key
    if data[0] in [0x02, 0x03]:
        return _deserialize_public_key(data)

    raise ValueError("Unknown key type")


def verify(
    message: bytes, signature: ECDSASignature, pubkey: PublicKey
) -> bool:
    """Verify an ECDSA signature against a message and public key.

    Args:
        message (bytes): The message that was signed.
        signature (ECDSASignature): The signature to verify.
        pubkey (PublicKey): The public key of the signer.

    Returns:
        bool: True if the signature is valid, False otherwise.
    """
    if not isinstance(message, bytes):
        raise TypeError("Invalid type for 'message'")

    if not isinstance(pubkey, PublicKey):
        raise TypeError("Invalid type for 'pubkey'")

    if not isinstance(signature, ECDSASignature):
        raise TypeError("Invalid type for 'signature'")

    try:
        ecdsa_verify(hash_sha256(message), signature, pubkey)
    except InvalidSignatureError:
        return False

    return True


def sign(message: bytes, privkey: PrivateKey) -> ECDSASignature:
    """Create an ECDSA signature for a message using a private key.

    Args:
        message (bytes): The message to sign.
        privkey (PrivateKey): The private key to use to create the signature.

    Returns:
        ECDSASignature: The created signature.
    """
    if not isinstance(privkey, PrivateKey):
        raise TypeError("Invalid type for 'privkey'")

    if not isinstance(message, bytes):
        raise TypeError("Invalid type for 'message'")

    return ecdsa_sign(hash_sha256(message), privkey)


def get_fingerprint(key: KeyData) -> bytes:
    """Get the fingerprint of a public or private key.

    Args:
        key (KeyData): The key to get the fingerprint of.

    Returns:
        bytes: The fingerprint of the key.
    """
    if not isinstance(key, (PublicKey, PrivateKey)):
        raise TypeError("Invalid type for 'key'")

    if isinstance(key, PrivateKey):
        return get_fingerprint(key.public_key())

    return utils.hash160(key.sec() + b"\x00")[:4]


def get_child_number(path: str) -> int:
    """Get the child number from a derivation path.

    Args:
        path (str): The derivation path.

    Returns:
        int: The child number.
    """
    parts = path.split("/")
    if len(parts) != 2 or parts[0] != "m":
        raise ValueError("