#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated file will have between 1 and 10 frames,
# where each frame has between 1 and 10 blocks, which are
# randomly sized from 1 to 255 bytes long.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed = rng.read(1)
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]

    screen_descriptor = struct.pack('<HBBBB', *[width, height, packed, background, aspect_ratio])
    out.write(screen_descriptor)

    # 3. Global Color Table
    global_color_table_flag = rng.read(1)[0] & 0b10000000
    color_resolution = (rng.read(1)[0] & 0b01110000) >> 4
    sort_flag = rng.read(1)[0] & 0b00001000
    size_of_global_color_table = rng.read(1)[0] & 0b00000111

    global_color_table = []
    for _ in range(size_of_global_color_table):
        red = rng.read(1)[0]
        green = rng.read(1)[0]
        blue = rng.read(1)[0]
        global_color_table.append((red, green, blue))

    global_color_table = b''.join([struct.pack('BBB', *rgb) for rgb in global_color_table])

    global_color_table_block = struct.pack('BB', *[global_color_table_flag, color_resolution | sort_flag | size_of_global_color_table]) + global_color_table
    out.write(global_color_table_block)

    # 4. Application Extension Block
    application_extension_block_introducer = b'\x21\xFF'
    application_identifier = b'NETSCAPE2.0'
    application_authentication_code = b'A'*8
    data_sub_block_introducer = b'\x01'
    loop_count = rng.read(2)
    loop_count_block = struct.pack('<HH', *[application_extension_block_introducer, len(application_identifier)+len(data_sub_block_introducer)+len(loop_count)]) + application_identifier + data_sub_block_introducer + loop_count
    out.write(loop_count_block)

    # 5. Image Separator
    image_separator = b'\x2C'
    out.write(image_separator)

    # 6. Image Descriptor
    left_position = rng.read(2)
    top_position = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    local_color_table_flag = rng.read(1)[0] & 0b10000000
    interlace_flag = rng.read(1)[0] & 0b01000000
    sort_flag = rng.read(1)[0] & 0b00001000
    size_of_local_color_table = rng.read(1)[0] & 0b00000111

    local_color_table = []
    for _ in range(size_of_local_color_table):
        red = rng.read(1)[0]
        green = rng.read(1)[0]
        blue = rng.read(1)[0]
        local_color_table.append((red, green, blue))

    local_color_table = b''.join([struct.pack('BBB', *rgb) for rgb in local_color_table])

    image_descriptor = struct.pack('<HHHHBBBB', *[left_position, top_position, width, height, local_color_table_flag, interlace_flag, sort_flag, size_of_local_color_table]) + local_color_table
    out.write(image_descriptor)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(trailer)