#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big') % 65536
    height = int.from_bytes(rng.read(2), 'big') % 65536
    flags = int.from_bytes(rng.read(1), 'big')
    background_color_index = int.from_bytes(rng.read(1), 'big')
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big')

    lsd = struct.pack('<HBBBB', width, height, flags, background_color_index, pixel_aspect_ratio)

    # 3. Color Table
    color_table_length = (flags & 0b111) * 3
    color_table = rng.read(color_table_length)

    # 4. Graphic Control Extension
    gce = None
    if flags & 0b10000000:
        block_size = int.from_bytes(rng.read(1), 'big')
        assert block_size == 0xF9
        
        flags = int.from_bytes(rng.read(1), 'big')
        delay = int.from_bytes(rng.read(2), 'big')
        transparent_color_index = int.from_bytes(rng.read(1), 'big')
        terminator = int.from_bytes(rng.read(1), 'big')

        gce = struct.pack('<BBHHB', block_size, flags, delay, transparent_color_index, terminator)

    # 5. Image Descriptor
    image_descriptor = b'\x2C\x00\x00\x00\x00\x00'

    # 6. Image Data
    lzw_minimum_code_size = int.from_bytes(rng.read(1), 'big')

    blocks = []
    while True:
        block_size = int.from_bytes(rng.read(1), 'big')
        if block_size == 0:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)