#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('<HH', rng.read(4))
    
    # Packed fields
    # 0x80: Global Color Table Flag
    # 0x70: Color Resolution (3 bits)
    # 0x0F: Sort Flag (3 bits)
    # 0x00: Size of Global Color Table (3 bits)
    packed_fields = 0x80 | 0x70 | 0x0F | 0x00

    # Background color index
    background_color_index = rng.read(1)[0] % 256

    # Pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)[0] % 256

    # Write everything to the output
    out.write(header + struct.pack('<HBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table
    # We'll use the same color table as in the example
    global_color_table = [
        0xFF0000, 0xFFFF00, 0x00FF00, 0x00FFFF, 0x0000FF, 0xFF00FF, 0xFFFFFF, 0x000000
    ]

    # Write it to the output
    out.write(struct.pack('<%dI' % len(global_color_table), *global_color_table))

    # 4. Graphics Control Extension
    # We'll use the same values as in the example
    graphics_control_extension = [
        0x21, 0xF9, 0x04, 0x01, 0x00, 0x00, 0xFF, 0xFF, 0xFF
    ]

    # Write it to the output
    out.write(struct.pack('<%dB' % len(graphics_control_extension), *graphics_control_extension))

    # 5. Image Descriptor
    image_descriptor = [
        0x2C, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF
    ]

    # Write it to the output
    out.write(struct.pack('<%dB' % len(image_descriptor), *image_descriptor))

    # 6. Local Color Table
    # We'll use the same color table as in the example
    local_color_table = [
        0xFF0000, 0xFFFF00, 0x00FF00, 0x00FFFF, 0x0000FF, 0xFF00FF, 0xFFFFFF, 0x000000
    ]

    # Write it to the output
    out.write(struct.pack('<%dI' % len(local_color_table), *local_color_table))

    # 7. LZW Minimum Code Size
    lzw_minimum_code_size = 0x08

    # Write it to the output
    out.write(bytes([lzw_minimum_code_size]))

    # 8. Data sub-blocks
    # We'll use the same data as in the example
    data_sub_blocks = [
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
        0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
        0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
        0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 