#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    # Packed fields
    packed = 0x80 | ((width & 0b11111) << 7) | (height & 0b111111)
    # Background color index
    background_color_index = rng.read(1)[0] % 256
    # Pixel aspect ratio
    pixel_aspect_ratio = 0
    # Global Color Table Flag
    global_color_table_flag = 0
    # Color Resolution
    color_resolution = 7
    # Sort Flag
    sort_flag = 0
    # Size of Global Color Table
    size_of_global_color_table = 7
    # End of fields
    end = 0
    # Construct the LSD
    logical_screen_descriptor = struct.pack(
        'BBB',
        packed,
        background_color_index,
        pixel_aspect_ratio
    ) + struct.pack(
        'BB',
        global_color_table_flag | (color_resolution << 7),
        sort_flag | (size_of_global_color_table << 7)
    ) + struct.pack('B', end)
    
    # 3. Global Color Table
    # 3.1. Number of colors in the table
    n_colors = 2 ** (size_of_global_color_table + 1)
    # 3.2. Colors
    colors = [
        struct.unpack('BBB', rng.read(3))
            for _ in range(n_colors)
    ]
    # 3.3. Construct the GCT
    global_color_table = b''.join([
        struct.pack('BBB', *c)
            for c in colors
    ])
    
    # 4. Image Descriptors
    # 4.1. Separator
    image_separator = b'\x2C'
    # 4.2. Left Position
    left_position = rng.read(2)
    # 4.3. Top Position
    top_position = rng.read(2)
    # 4.4. Width
    width = rng.read(2)
    # 4.5. Height
    height = rng.read(2)
    # 4.6. Packed Fields
    local_color_table_flag = 0
    interlace_flag = 0
    sort_flag = 0
    size_of_local_color_table = 0
    reserved = 0
    # 4.7. Local Color Table
    if local_color_table_flag:
        # 4.7.1. Number of colors in the table
        n_colors = 2 ** (size_of_local_color_table + 1)
        # 4.7.2. Colors
        colors = [
            struct.unpack('BBB', rng.read(3))
                for _ in range(n_colors)
        ]
        # 4.7.3. Construct the LCT
        local_color_table = b''.join([
            struct.pack('BBB', *c)
                for c in colors
        ])
    else:
        local_color_table = b''
    # 4.8. Image Data
    image_data = rng.read()
    
    # 5. Trailer
    trailer = b';'
    
    # Write everything to the output file
    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(global_color_table)
    out.write(image_se