#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated file will have between 1 and 10 frames,
# where each frame has between 1 and 10 blocks, which are
# randomly sized from 1 to 255 bytes long.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('<BBB', *struct.unpack('>BBB', rng.read(3))) + rng.read(1)

    # 3. Global Color Table
    # We don't use it, so we just skip over it
    global_color_table_size = 7 + 3 * (2 ** (lsd[0] & 0b111))
    rng.seek(global_color_table_size, 1)

    # 4. Application Extension
    # We don't use it, so we just skip over it
    app_extension_block_size = 13 + rng.read(1)[0]
    rng.seek(app_extension_block_size, 1)

    # 5. Image Descriptor
    image_descriptor = b'\x2c\x00\x00\x00\x00\x00'

    # 6. Local Color Table
    # We don't use it, so we just skip over it
    local_color_table_size = 3 * (2 ** ((image_descriptor[0] >> 4) & 0b111))
    rng.seek(local_color_table_size, 1)

    # 7. Image Data
    # We don't use it, so we just skip over it
    image_data_length = width * height
    rng.seek(image_data_length, 1)

    # 8. Trailer
    trailer = b';'

    # Write everything together
    out.write(header + lsd + image_descriptor + trailer)

    # 9. Loop through all frames
    num_frames = rng.read(1)[0] % 10 + 1
    for _ in range(num_frames):
        # 10. Graphics Control Extension
        gce = b'\x21\xf9\x04\x00\xff\xff\xff\x00'

        # 11. Image Descriptor
        image_descriptor = b'\x2c\x00\x00\x00\x00\x00'

        # 12. Local Color Table
        # We don't use it, so we just skip over it
        local_color_table_size = 3 * (2 ** ((image_descriptor[0] >> 4) & 0b111))
        rng.seek(local_color_table_size, 1)

        # 13. Image Data
        # We don't use it, so we just skip over it
        image_data_length = width * height
        rng.seek(image_data_length, 1)

        # 14. Trailer
        trailer = b';'

        # Write everything together
        out.write(gce + image_descriptor + trailer)

        # 15. Loop through all blocks
        num_blocks = rng.read(1)[0] % 10 + 1
        for _ in range(num_blocks):
            # 16. Block Size
            block_size = rng.read(1)[0] % 256 + 1

            # 17. Block Data
            # We don't use it, so we just skip over it
            rng.seek(block_size, 1)

            # 18. CRC
            crc = rng.read(2)

            # Write everything together
            out.write(struct.pack('<H', block_size) + crc)

