#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated file will have between 1 and 10 frames,
# where each frame has between 1 and 10 blocks, which are
# randomly sized from 1 to 255 bytes long.
def generate_gif(out: BinaryIO, rng: BinaryIO):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical screen descriptor
    width = struct.pack('<H', 640)
    height = struct.pack('<H', 480)
    background_color_index = struct.pack('B', 0)
    pixel_aspect_ratio = struct.pack('B', 0)
    logical_screen_descriptor = (width + height +
                                 background_color_index + pixel_aspect_ratio)

    # 3. Global color table
    global_color_table = b''
    for i in range(256):
        red = struct.pack('B', i % 256)
        green = struct.pack('B', i % 256)
        blue = struct.pack('B', i % 256)
        global_color_table += red + green + blue

    # 4. Application extension
    application_extension_introducer = b'\x21\xFF'
    application_extension_label = b'NETSCAPE2.0'
    application_extension_block_size = b'\x0B'
    application_extension_data = b'\x01\x00\x01\x00\x00'
    application_extension = (application_extension_introducer +
                             application_extension_label +
                             application_extension_block_size +
                             application_extension_data)

    # 5. Image descriptor
    image_descriptor = b'\x2C'

    # 6. Frames
    frames = []
    num_frames = int.from_bytes(rng.read(1), 'little') + 1
    for _ in range(num_frames):
        # Frame header
        frame_header = b'\x21\xF9\x04\x0B\x00\x00\x00\x00\x00\x2C'

        # Blocks
        blocks = []
        num_blocks = int.from_bytes(rng.read(1), 'little') + 1
        for _ in range(num_blocks):
            block_length = int.from_bytes(rng.read(1), 'little') + 1
            block_data = rng.read(block_length)
            blocks.append((struct.pack('B', block_length) + block_data))

        # Frame footer
        frame_footer = b'\x00\x3B'

        # Append frame to list of frames
        frames.append(frame_header + b''.join(blocks) + frame_footer)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(trailer)