#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(num_colors * 3)
    
    # 4. Optional Graphic Control Extension
    # We will randomly decide whether we want one or not
    gce = None
    if rng.read(1)[0] & 1 == 1:
        # We do! Let's construct it
        
        # First, the block size (always 4)
        gce = b'\xF9\x04'
        
        # Then, the packed field (we'll use a random byte for variety)
        gce += rng.read(1)
        
        # Delay time in hundredths of seconds (again, random byte for variety)
        gce += rng.read(1)
        
        # Transparent color index (also random byte for variety)
        gce += rng.read(1)
        
        # Block terminator (always \x00)
        gce += b'\x00'
    
    # 5. Image Descriptor
    image_descriptor = b'\x2C\x00\x00\x00\x00'
    
    # 6. Image Data
    # We're going to write a single frame with no interlacing
    # We'll just fill the whole thing with random data
    image_data = rng.read((width * height) // 2)
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)