#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed field
    #   - 1 bit : Global Color Table Flag
    #   - 3 bits: Color Resolution
    #   - 1 bit : Sort Flag
    #   - 3 bits: Size of Global Color Table
    packed = struct.pack('B', rng.read(1)[0] & 0x7)
    
    # Background color index
    background_color_index = struct.pack('B', rng.read(1)[0])
    
    # Pixel aspect ratio
    pixel_aspect_ratio = struct.pack('B', rng.read(1)[0])
    
    # Write everything to the output
    out.write(header + width + height + packed + background_color_index + pixel_aspect_ratio)
    
    # 3. Global Color Table
    if packed[0] & 0x80 != 0:
        global_color_table_size = 3 * (2 ** ((packed[0] & 0x07) + 1))
        
        for _ in range(global_color_table_size):
            out.write(rng.read(1))
            
        # Pad with zeros until we reach a multiple of three
        while len(out.getvalue()) % 3 != 0:
            out.write(b'\x00')
    
    # 4. Image Blocks
    while True:
        # Image Separator
        image_separator = struct.pack('B', rng.read(1)[0])
        
        if image_separator == b'\x2c':
            # 5. Image Descriptor
            left, top, width, height = struct.unpack('>HHHH', rng.read(8))
            
            # Packed field
            #   - 1 bit : Local Color Table Flag
            #   - 3 bits: Reserved
            #   - 1 bit : Interlace Flag
            #   - 3 bits: Sort Flag
            #   - 3 bits: Size of Local Color Table
            packed = struct.pack('B', rng.read(1)[0] & 0x7)
            
            # LZW Minimum Code Size
            lzw_minimum_code_size = struct.pack('B', rng.read(1)[0])
            
            # Write everything to the output
            out.write(image_separator + left + top + width + height + packed + lzw_minimum_code_size)
            
            # 6. Local Color Table
            if packed[0] & 0x80 != 0:
                local_color_table_size = 3 * (2 ** ((packed[0] & 0x07) + 1))
                
                for _ in range(local_color_table_size):
                    out.write(rng.read(1))
                
                # Pad with zeros until we reach a multiple of three
                while len(out.getvalue()) % 3 != 0:
                    out.write(b'\x00')
            
            # 7. Image Data
            # Read until we find an end code
            while True:
                block_size = struct.unpack('B', rng.read(1))[0]
                
                if block_size == 0:
                    break
                
                out.write(rng.read(block_size))
        
        elif image_separator == b'\x3b':
            # Trailer
            trailer = struct.pack('B', rng.read(1)[0])
            
            # Write everything to the output
            out.write(trailer)
            
            return
        
        else:
            raise Exception("Invalid GIF")
