#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    # Color table flag (always true)
    color_table_flag = True

    # Color resolution (always 8 bits per primary color)
    color_resolution = 8

    # Sort flag (always false)
    sort_flag = False

    # Size of global color table (always 2^bit_depth - 1)
    global_color_table_size = 2**color_resolution - 1

    # LZW minimum code size (always 8 bits)
    lzw_minimum_code_size = 8

    # Put it all together
    lsd = struct.pack(
            '>HBBBBBB',
            width, height,
            packed_fields, background_color_index, pixel_aspect_ratio,
            int(color_table_flag), color_resolution,
            int(sort_flag), global_color_table_size
        )

    # 3. Global Color Table
    # We don't need this because we set the color table flag to false in the LSD
    # color_table = rng.read(3 * 2 ** color_resolution)

    # 4. Graphic Control Extension
    # Always present
    # Block size is always 4
    # Transparent color index is always 0
    # Disposal method is always 2 (restore to background)
    # User input flag is always false
    # Delay time is always 0
    # Transparent color flag is always false
    gce = struct.pack('>BBBBB', 4, 0, 2, 0, 0)

    # 5. Image Descriptor
    # Left position is always 0
    # Top position is always 0
    # Width is same as in LSD
    # Height is same as in LSD
    # Local color table flag is always false
    # Interlace flag is always false
    # Sort flag is always false
    # Size of local color table is always 0
    image_descriptor = struct.pack(
            '>BBHHBBB',
            0, 0, width, height,
            0, 0, 0
        )

    # 6. Image Data
    # The first byte is the LZW minimum code size
    # Then there are zero or more data blocks
    # Each block starts with its size (one byte)
    # Then there are that many bytes of data
    # Finally, there is a block of size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)