#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 7 + 1
    global_color_table = [rng.read(3) for _ in range(global_color_table_size)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBB', 0, 0, 0b10000000 | rng.read(1)[0], 0, 0)

    # 5. Local Color Table
    local_color_table_flag = bool(rng.read(1)[0] & 1)
    if not local_color_table_flag:
        local_color_table = []
    else:
        local_color_table_size = rng.read(1)[0] & 7 + 1
        local_color_table = [rng.read(3) for _ in range(local_color_table_size)]

    # 6. Graphic Control Extension
    graphic_control_extension_flag = bool(rng.read(1)[0] & 1)
    if graphic_control_extension_flag:
        # Block size is always 4
        block_size = rng.read(1)[0]

        # Packed field: Disposal method (1 bit), User input flag (1 bit), Transparent color flag (1 bit), Reserved (1 bit)
        packed_field = rng.read(1)[0]
        
        # Delay time (2 bytes)
        delay_time = rng.read(2)

        # Transparent color index (1 byte)
        transparent_color_index = rng.read(1)[0]

        # End of extension data
        end_of_extension_data = rng.read(block_size - 4)

    # 7. Image Data
    # Number of loops (1 byte)
    num_loops = rng.read(1)[0]

    # Image data
    image_data = b''
    while True:
        # Block size
        block_size = rng.read(1)[0]

        # If this is zero, we're done with the image data
        if block_size == 0:
            break

        # Read the next chunk of image data
        image_data += rng.read(block_size)

    # 8. Trailer
    trailer = b';'

    # Write everything to output stream
    out.write(header)
    out.write(lsd)
    for entry in global_color_table:
        out.write(entry)
    out.write(image_descriptor)
    if len(local_color_table):
        out.write(struct.pack('B', local_color_table_size))
        for entry in local_color_table:
            out.write(entry)
    if graphic_control_extension_flag:
        out.write(struct.pack('BB', 0x21, 0xF9))
        out.write(struct.pack('B', 4))
        out.write(packed_field)
        out.write(delay_time)
        out.write(transparent_color_index)
        out.write(end_of_extension_data)
    out.write(num_loops.to_bytes(1, 'big'))
    out.write(image_data)
    out.write(trailer)