#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0]) & True
    if not has_gce:
        gce = None
    else:
        # Disposal method (0 - No disposal specified, 1 - Do not dispose, 2 - Restore to background color, 3 - Restore to previous)
        disposal_method = rng.read(1)[0] % 4
        
        # User input flag (0 - False, 1 - True)
        user_input_flag = bool(rng.read(1)[0]) & True
        
        # Transparent color index (if set, transparent pixels are replaced with this color index in the output)
        transparent_color_index = rng.read(1)[0] % len(color_table)
        
        # Delay time (in hundredths of seconds; max is 65535)
        delay_time = rng.read(2)
        
        # Transparent color index (if set, transparent pixels are replaced with this color index in the output)
        transparent_color_index = rng.read(1)[0] % len(color_table)
        
        # End of GCE block
        gce = b'\x00\x3b'
        
    # 5. Image Descriptor
    # Left position (always zero)
    left = b'\x00\x00'
    
    # Top position (always zero)
    top = b'\x00\x00'
    
    # Width and Height (again, using two bytes each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed field (using one byte)
    packed_field = rng.read(1)[0]
    
    # 6. Local Color Table (same as global table but optional)
    local_color_table = None
    if bool(packed_field >> 7):
        num_local_colors = (rng.read(1)[0] % 256) or 1
        local_color_table = rng.read(3 * num_local_colors)

    # 7. Image data
    image_data = []
    while sum([len(l) for l in image_data]):
        # Each line starts with a single byte that tells you how many bytes follow
        length = rng.read(1)[0]

        # The next byte tells you whether the line is encoded as an "uninterrupted sequence" or a "partial block".
        # We'll just assume uninterrupted sequences for now...
        # If we wanted to be more correct, we'd check here and read from the stream until we got all the bytes needed.
        assert length!= 0xFF
        
        # Read the actual pixel data
        image_data += [rng.read(length)]
    
    # 8. Trailer
    trailer = b';'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + (gce or b'') + b'\x2c' + left + top + struct.pack('>HHB', width, height, packed_field) + (local_color_table or b'') + b'\x00'.join(image_data) + trailer)

