#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated file will have between 1 and 10 frames,
# where each frame has between 1 and 10 blocks, which are
# randomly sized from 1 to 255 bytes long.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = rng.read(1)[0] % (2**8) + 1
    height = rng.read(1)[0] % (2**8) + 1
    flags = 0b10000000 | rng.read(1)[0] % (2**4)
    bgColorIndex = rng.read(1)[0] % (2**8)
    pixelAspectRatio = 0
    logicalScreenDescriptor = struct.pack('>BHHBBB',
        width, height, flags, bgColorIndex, pixelAspectRatio
    )

    # 3. Global Color Table
    globalColorTableSize = 0
    globalColorTable = b''
    if flags & 0b10000000!= 0:
        globalColorTableSize = 2**(flags & 0b00001111)
        for i in range(globalColorTableSize):
            r = rng.read(1)[0] % (2**8)
            g = rng.read(1)[0] % (2**8)
            b = rng.read(1)[0] % (2**8)
            globalColorTable += struct.pack('BBB', r, g, b)

    # 4. Graphics Control Extension
    graphicsControlExtension = b''
    if flags & 0b00000100 == 0:
        graphicsControlExtension = struct.pack('>BBHB',
            0x21, 0xF9, 4, 0b10000000 | rng.read(1)[0] % (2**8)
        )

    # 5. Image Descriptor
    imageLeftPosition = 0
    imageTopPosition = 0
    imageWidth = width
    imageHeight = height
    localColorTableFlag = False
    interlaceFlag = False
    sortFlag = False
    localColorTableSize = 0
    localColorTable = b''
    lzwMinimumCodeSize = 0
    imageDescriptor = struct.pack('>BHHBBBB',
        0x2C, imageLeftPosition, imageTopPosition,
        imageWidth, imageHeight, localColorTableFlag << 7 |
        interlaceFlag << 6 | sortFlag << 5 | localColorTableSize,
        lzwMinimumCodeSize
    )

    # 6. Blocks
    numBlocks = rng.read(1)[0] % (2**8) + 1
    blockData = b''
    for _ in range(numBlocks):
        blockSize = rng.read(1)[0] % (2**8) + 1
        blockData += rng.read(blockSize)

    # 7. Trailer
    trailer = b'\x3B'

    data = \
        header + \
        logicalScreenDescriptor + \
        globalColorTable + \
        graphicsControlExtension + \
        imageDescriptor + \
        blockData + \
        trailer

    out.write(data)