#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>BHHBBB',
        0b11100000 | ((width >> 7) & 0b00011111),
        width & 0x7F,
        height,
        rng.read(1)[0],
        rng.read(1)[0],
        0b10000000 | rng.read(1)[0]
    )

    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 0b00001111
    global_color_table_flag = bool((rng.read(1)[0] & 0b10000000) >> 7)

    if global_color_table_flag:
        global_color_table = [
            rng.read(3)
            for _ in range(global_color_table_size)
        ]

        palette = b''
        for rgb in global_color_table:
            palette += rgb[::-1]

        lsd += struct.pack('>B', len(palette)) + palette
    else:
        lsd += b'\x00\x00\x00'

    # 4. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB',
        0x2C,    # Image separator
        rng.read(1)[0],      # Left position of image
        rng.read(1)[0],      # Top position of image
        width,               # Width of image
        height,              # Height of image
        0,                   # Flags: no local color table
        0                    # No LZW minimum code size
    )

    # 5. Local Color Table
    local_color_table_flag = bool((rng.read(1)[0] & 0b10000000) >> 7)

    if local_color_table_flag:
        local_color_table_size = rng.read(1)[0] & 0b00001111

        local_color_table = [
            rng.read(3)
            for _ in range(local_color_table_size)
        ]

        palette = b''
        for rgb in local_color_table:
            palette += rgb[::-1]

        image_descriptor += struct.pack('>B', len(palette)) + palette
    else:
        image_descriptor += b'\x00\x00\x00'

    # 6. Image Data
    data = []
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x21:   # Extension Block
            ext_label = rng.read(1)[0]

            if ext_label == 0xF9:       # Graphic Control
                # Block Size
                bs = rng.read(1)[0]

                # Packed Fields
                pf = rng.read(1)[0]

                # Delay Time
                dt = int.from_bytes(
                    rng.read(2), 'big')

                # Transparent Color Index
                tc = rng.read(1)[0]

                # Terminator
                et = rng.read(1)[0]

                assert bs == 4
                assert pf == 0
                assert et == 0

                data.append([ext_label, dt])
            elif ext_label == 0xFF:     # Application Block
                app_block_size = int.from_bytes(
                    rng.read(2), 'big')

                app_identifier = rng.read(8).decode()

                app_auth_code = rng.read(3).hex().upper()

                app_data = rng.read(app_block_size - 11)

                assert app_block_size == len(app_data) + 11

                data.append([ext_label, app_identifier,
                             app_auth_code, app_data])
            else:                       # Unrecognized extension label
                raise Exception("Unrecognized extension label")
        elif block_type == 0x2C:   # Image Descriptor
            pass
        elif block_type == 0x3B:   # Trailer
            break
        else:
            raise Exception("Unrecognized block type")

    # 7. Write to output stream
    out.write(header)
    out.write(lsd)
    out.write(image_descriptor)

    for datum in data:
        if isinstance(datum, list):
            out.write(struct.pack('>B', datum[0]))

            if datum[0] == 0xF9:
                out.write(struct.pack('>BBBB',
                    4,          # Block Size
                    0,          # Packed Field
                    0,          # Delay Time
                    0           # Transparent Color Index
                ))
            elif datum[0] == 0xFF:
                out.write(struct.pack('>BH',
                    11 + len(datum[3]),        # Block Size
                    ord(datum[1][0]) << 24 | \
                        ord(datum[1][1]) << 16 | \
                        ord(datum[1][2]) << 8 | \
                        ord(datum[1][3])
                ))
                out.write(datum[1].encode())
                out.write(bytes.fromhex(datum[2]))
                out.write(datum[3])
            else:
                raise Exception("Unrecognized extension label")
        else:
            out.write(datum)

    out.write(b';