#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = min((rng.read(1)[0] % 256) + 1, 256)
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(global_color_table_size)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBBBB', 0, 0, 0b10000000 | rng.read(1)[0], 0, 0, 0, 0)

    # 5. Local Color Table (same as above)
    local_color_table_size = min((rng.read(1)[0] % 256) + 1, 256)
    local_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(local_color_table_size)]

    # 6. Extension Blocks
    extension_blocks = []

    # 7. Graphics Control Block
    graphics_control_block = struct.pack('>BBBBBBH', 0x21, 0xF9, 0x04, 0x05, 0x00, 0x00, 0xFF)

    # 8. Image Data
    image_data = b''

    # 9. Trailer
    trailer = b';'

    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    if len(global_color_table):
        out.write(bytes([global_color_table_size]))
        out.write(b''.join(global_color_table))
    else:
        out.write(b'\x00')
    out.write(image_descriptor)
    if len(local_color_table):
        out.write(bytes([local_color_table_size]))
        out.write(b''.join(local_color_table))
    else:
        out.write(b'\x00')
    for block in extension_blocks:
        out.write(block)
    out.write(graphics_control_block)
    out.write(image_data)
    out.write(trailer)


