#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])
    
    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        return
    
    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = rng.read(3 * num_colors)  # Random colors
    
        
    # 4. Image Descriptor
    image_left, image_top = struct.unpack('>HH', rng.read(4))
    image_width, image_height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    local_color_table_flag = bool(packed & 0x80)
    interlace_flag = bool(packed & 0x40)
    sort_flag = bool(packed & 0x20)
    reserved = rng.read(1)[0]
    lzw_minimum_code_size = rng.read(1)[0]

    # 5. Local Color Table (if present)
    if local_color_table_flag:
        num_colors = 2 ** ((packed >> 4) + 1)
        color_table += rng.read(3 * num_colors)  # Random colors

    # 6. LZW Data
    data = []
    while True:
        code_size = rng.read(1)[0]
        if code_size == 0:
            break
        
        for i in range((code_size + 1) // 8):
            byte = rng.read(1)[0]
            for j in reversed(range(8)):
                data.append(((byte >> j) & 1) ^ 1)

        while len(data) >= 11 and data[-11]!= 0:
            first_index = sum([data[i] << (7 - i) for i in range(11)])
            
            if first_index < len(data) - 11:
                second_index = sum([data[i] << (7 - i) for i in range(first_index + 1, min(len(data), first_index + 12))])
                
                if second_index > first_index:
                    length = second_index - first_index
                    
                    if length <= 10:
                        data.extend([(second_index >> (7 - i)) & 1 for i in reversed(range(length))])
                    else:
                        data.extend([(second_index >> (7 - i)) & 1 for i in reversed(range(10))])
                        
                        while length > 12:
                            data.extend([(second_index >> (7 - i)) & 1 for i in reversed(range(11))])
                            length -= 11
                            
                        data.extend([(second_index >> (7 - i)) & 1 for i in reversed(range(length))])
                        
                elif second_index == first_index:
                    data.extend([1, 1])
                    
                else:
                    data.extend([0, 1])
                    
            else:
                data.extend([0, 1])
                
    # 7. Trailer
    trailer = bytes([0x3B])
    
    # Write everything to output
    out.write(header)
    out.write(lsd)
    if global_color_table_flag:
        out.write(color_table)
    out.write(struct.pack('>HHBBBBB', image_left, image_top, image_width, image_height, packed, lzw_minimum_code_size))
    if local_color_table_flag:
        out.write(color_table)
    out.write(bytes(data))
    out.write(trailer)

