#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# returns its size in bytes.
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write("GIF89a".encode())

    width = 1 + rng.read(1)[0] % 256
    height = 1 + rng.read(1)[0] % 256
    flags = (width & 1) << 7 | (height & 1) << 6
    out.write(struct.pack("<HHB", width, height, flags))

    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = 0
    out.write(struct.pack("<BBB", background_color_index, pixel_aspect_ratio, 0))

    # Global Color Table
    global_color_table_size = min((1 << ((flags >> 4) & 7)), 256)
    for i in range(global_color_table_size):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        out.write(struct.pack("<BBB", r, g, b))

    # Image Descriptor
    left = 0
    top = 0
    width = 1 + rng.read(1)[0] % 256
    height = 1 + rng.read(1)[0] % 256
    flags = 0b10000000 | (rng.read(1)[0] % 8) << 2
    out.write(struct.pack("<HHBB", left, top, width, height, flags))

    local_color_table_flag = bool(rng.read(1)[0] % 2)
    interlace_flag = False
    sort_flag = False
    local_color_table_size = None if not local_color_table_flag else \
        1 << ((flags >> 3) & 7)
    lct = []
    if local_color_table_flag:
        for i in range(local_color_table_size):
            r = rng.read(1)[0] % 256
            g = rng.read(1)[0] % 256
            b = rng.read(1)[0] % 256
            lct += [r,g,b]
        for i in range(len(lct), 256):
            lct += [0,0,0]

        out.write(bytes(lct))

    lzw_min_code_size = rng.read(1)[0] % 256
    out.write(struct.pack("<B", lzw_min_code_size))

    data = bytearray()
    while len(data) < width * height:
        code_length = rng.read(1)[0] % 256
        if code_length == 0: continue
        if code_length > lzw_min_code_size: code_length = lzw_min_code_size
        data += bytes([code_length])

        for _ in range(code_length - 1):
            data += bytes([rng.read(1)[0] % 256])

    # LZW decompression
    dictionary = {}
    dict_size = 2 ** lzw_min_code_size
    dictionary[dict_size] = ""
    dictionary[dict_size+1] = "\x00"
    w = ""
    result = ""
    for c in data:
        wc = w + chr(c)
        if wc in dictionary:
            w = wc
        else:
            result += dictionary[len(w)]
            # Add wc to the dictionary.
            dictionary[dict_size + len(dictionary)] = wc
            dict_size += 1

            # If the dictionary is full, delete the oldest entry.
            if dict_size == 4096:
                del dictionary[list(dictionary.keys())[0]]

            w = chr(c)
    result += dictionary[len(w)]

    # Write resulting image data
    for y in range(height):
        for x in range(width):
            out.write(result[(y*width)+x].encode())

    return out.tell()

