#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    # 2. Logical screen descriptor
    # 3. Global color table (optional)
    # 4. Image descriptor
    # 5. Local color table (optional)
    # 6. Image data

    # 1. Header
    header = b'GIF89a' + bytes([0x00]) * 7
    out.write(header)

    # 2. Logical screen descriptor
    width = rng.read(1)[0] % 256
    height = rng.read(1)[0] % 256
    flags = rng.read(1)[0] % 256
    background = rng.read(1)[0] % 256
    aspect_ratio = rng.read(1)[0] % 256
    logical_screen_descriptor = struct.pack('<BBHHBBB',
            width, height, flags, background, aspect_ratio,
            0x00)
    out.write(logical_screen_descriptor)

    # 3. Global color table (optional)
    global_color_table_size = rng.read(1)[0] % 256
    if global_color_table_size > 0:
        global_color_table = bytes([
                rng.read(1)[0] for _ in range(global_color_table_size)])
        out.write(global_color_table)

    # 4. Image descriptor
    left = rng.read(1)[0] % 256
    top = rng.read(1)[0] % 256
    image_width = rng.read(1)[0] % 256
    image_height = rng.read(1)[0] % 256
    flags = rng.read(1)[0] % 256
    local_color_table_flag = (flags >> 7) & 1
    interlace_flag = (flags >> 6) & 1
    sort_flag = (flags >> 5) & 1
    reserved = (flags >> 3) & 3
    size_of_local_color_table = flags & 7
    image_descriptor = struct.pack('<BBBBBBBHB',
            left, top, image_width, image_height, flags,
            local_color_table_flag, interlace_flag, sort_flag,
            reserved, size_of_local_color_table)
    out.write(image_descriptor)

    # 5. Local color table (optional)
    local_color_table = bytes([
            rng.read(1)[0] for _ in range(size_of_local_color_table)])
    out.write(local_color_table)

    # 6. Image data
    lzw_min_code_size = rng.read(1)[0] % 256
    out.write(bytes([lzw_min_code_size]))

    # Generate LZW code stream
    codestream = []
    while True:
        byte = rng.read(1)[0]

        if byte == 0x00: break
        elif byte!= 0xff and len(codestream) < 256:
            codestream += [byte]
        else:
            length = rng.read(1)[0] % 256

            if length == 0x00: continue
            elif length == 0xff:
                codestream += [length, 0xff]
            else:
                codestream += [length]

                for i in range(length):
                    codestream += [rng.read(1)[0]]

    # Compress codestream to LZW data
    compressed = compress_lzw(codestream)

    # Write compressed LZW data
    out.write(compressed)


# Compresses an LZW codestream into LZW data
def compress_lzw(codestream: list[int]) -> bytes:
    # Initialize dictionary with all single-character codes
    dictionary = {i : chr(i).encode() for i in range(256)}

    # Initialize output buffer
    output = b''

    # Initialize current codeword
    current_codeword = ''

    # Iterate over each character in the codestream
    for char in map(chr, codestream):
        # Get next codeword from dictionary
        next_codeword = dictionary[ord(char)]

        # Check if we need to write any new codewords
        if not current_codeword or \
           current_codeword[:len(next_codeword)]!= next_codeword[:-1]:
            # Add new codeword to dictionary
            dictionary[len(dictionary)] = current_codeword + char

            # Write codeword to output buffer
            output += bytes([len(current_codeword)])
            output += current_codeword

        # Update current codeword
        current_codeword = next_codeword

    # Flush remaining codeword
    output += bytes([len(current_codeword)])
    output += current_codeword

    return output


if