#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    # Left position is always zero
    left = b'\x00\x00'

    # Top position is between 0 and height - 1
    top = struct.pack('>H', rng.read(2)[0] % height)

    # Width and Height are between 0 and max(width, height)
    width = min((rng.read(2)[0] % width) or 1, width)
    height = min((rng.read(2)[0] % height) or 1, height)

    # Packed field: Local color table flag set?
    packed_field = ((rng.read(1)[0] & 1) << 7) | \
                   ((rng.read(1)[0] & 1) << 6) | \
                   ((rng.read(1)[0] & 1) << 5) | \
                   0b00000

    # Interlace bit, sort of useless but let's keep it in there
    interlace_bit = (rng.read(1)[0] & 1) << 6

    # Sort of useless again, but let's keep it in there
    sorted_color_table_flag = (rng.read(1)[0] & 1) << 5

    # Size of local color table (between 0 and 256)
    num_local_colors = (rng.read(1)[0] % 256) or 1

    # Write all these values together
    image_descriptor = left + top + struct.pack('>HHBB', width, height, packed_field, num_local_colors)

    # 6. Image Data
    # Let's use the same trick as we did with the GCE: If the first byte is non-zero, then read more data
    if rng.read(1)[0]:
        num_bytes = rng.read(1)[0]

        # Read the specified amount of bytes from /dev/urandom
        image_data = rng.read(num_bytes)
    else:
        image_data = b''

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)