#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = rng.read(1)[0] % 64 + 1
    h = rng.read(1)[0] % 64 + 1

    # Random color table size
    cts = rng.read(1)[0] % 7 + 1

    # Color resolution is always 7
    cr = b'\x07'

    # Sort flag is always set to true
    sf = b'\x01'

    # Global color table flag is always false
    gcf = b'\x00'

    # Background color index is always zero
    bc = b'\x00\x00'

    # Pixel aspect ratio is always 1
    par = b'\x00\x00'

    lsd = header \
            + struct.pack('BB', w, h) \
            + cts * b'\x00\x00\x00' \
            + cr \
            + sf \
            + gcf \
            + bc \
            + par

    # 3. Image Descriptor
    # Frame left position is always zero
    flp = b'\x00\x00'

    # Frame top position is always zero
    ftp = b'\x00\x00'

    # Frame width is randomly chosen from [w-5, w+5]
    fw = rng.read(1)[0] % (w*2 - 10) + w - 5

    # Frame height is randomly chosen from [h-5, h+5]
    fh = rng.read(1)[0] % (h*2 - 10) + h - 5

    # Local color table flag is always false
    lctf = b'\x00'

    # Interlace flag is always false
    ifl = b'\x00'

    # Sort flag is always set to true
    srt = b'\x01'

    # Reserved flags are all zeros
    rsv = b'\x00\x00'

    # LZW minimum code size is always 2
    lcsz = b'\x02'

    id = flp \
            + ftp \
            + struct.pack('HHB', fw, fh, lctf[0]) \
            + ifl \
            + srt \
            + rsv \
            + lcsz

    # 4. Graphics Control Extension
    # Block size is always 4
    bs = b'\x04'

    # Disposal method is always 0
    dm = b'\x00'

    # User input flag is always false
    uif = b'\x00'

    # Transparent color index is always zero
    tc = b'\x00\x00'

    # Delay time is always 10ms
    dt = b'\x00\x0A'

    # Transparent color index is always zero
    tci = b'\x00'

    ge = bs \
            + dm \
            + uif \
            + tc \
            + dt \
            + tci

    # 5. Application Extension
    # Extension label is always "NETSCAPE"
    el = b'NETSCAPE'

    # Extension block length is always 3
    ebl = b'\x03'

    # Application identifier is always "2.0"
    ai = b'2.0'

    # Application authentication code is always "\x00\x00\x00"
    ac = b'\x00\x00\x00'

    ae = el \
            + ebl \
            + ai \
            + ac

    # 6. Comment Extension
    # Extension label is always "Comment"
    cl = b'Comment'

    # Extension block length is randomly chosen from [10, 20]
    ebl = rng.read(1)[0] % 10 + 10

    # Comment data is randomly generated with length of ebl
    cd = rng.read(ebl)

    ce = cl \
            + struct.pack('B', len(cd)) \
            + cd

    # 7. Plain Text Extension
    # Extension label is always "PlainText"
    pl = b'PlainText'

    # Extension block length is always 12
    ebl = b'\x0c'

    # Text grid left position is always zero
    tgl = b'\x00\x00'

    # Text grid top position is always zero
    tgt = b'\x00\x00'

    # Text grid width is randomly chosen from [w-5, w+5]
    tgw = rng.read(1)[0] % (w*2 - 10) + w - 5

    # Text grid height is randomly chosen from [h-5, h+5]
    tgh = rng.read(1)[0] % (h*2 - 10) + h - 5

    # Cell width is randomly chosen from [1, 10]
    cw = rng.read(1)[0] % 10 + 1

    # Cell height is randomly chosen from [1, 10]
    ch = rng.read(1)[0] % 10 + 1

    # Text foreground color index is always zero
    tfi = b'\x00\x00'

    # Text background color index is always zero
    tbi = b'\x00\x00'

    pe = pl \
            + ebl \
            + tgl \
            + tgt \
            + struct.pack('HHBBB', tgw, tgh