#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if bool(rng.read(1)):
        # Use this block if we want to use an extension

        # Block Size (always 4)
        gce = b'\x04'

        # Flags (packed field): Transparent color flag, User input flag, Has delay flag, Has transparent color index flag
        flags = rng.read(1)[0] & 0b11110000 | rng.read(1)[0] & 0b00000011

        # Delay time (two bytes)
        delay_time = struct.pack('>H', rng.read(2)[0])

        # Transparent color index (one byte)
        trans_index = rng.read(1)[0]

        # End of GCE block
        gce += flags + delay_time + trans_index
    else:
        # Don't use this block if we don't want to use an extension
        gce = None

    # 5. Image Descriptor
    id_block = struct.pack('>HHBBB', 0, 0, 0, 0, 0)  # Left position, Top position, Width, Height, Packed fields

    # 6. Image Data
    image_data = b''
    while True:
        # Generate up to 255 bytes of data
        chunk_size = min((rng.read(1)[0] % 256) or 1, 255)

        # If there are no more chunks left in the image data, break from loop
        if not chunk_size:
            break

        # Add the generated chunk to our image data
        image_data += rng.read(chunk_size)

    # 7. Trailer
    trailer = b';'

    # Write everything to output stream
    out.write(header + lsd + color_table + (gce or '') + id_block + image_data + trailer)
