#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        
        # We're done when there are no more bytes in this block
        if not block_size:
            break
        
        # Add some data to our current block
        blocks.append(rng.read(block_size))
        
        # Keep track of how many bytes we've read so far
        total_blocks_size += block_size
        
        # Stop adding blocks once we reach our limit
        if total_blocks_size >= 255:
            break
        
    # Compress all our blocks together with LZW compression
    compressed_data = compress_lzw(b''.join(blocks), lzw_minimum_code_size)
    
    # Finish off by writing everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + compressed_data)


# Compresses `data` using LZW compression with minimum code size `min_code_size`.
# Returns the resulting compressed data as a binary string.
def compress_lzw(data: bytes, min_code_size):
    # Initialize our dictionary with all ASCII characters
    dictionary = {chr(i).encode(): i for i in range(256)}

    # Our initial state is just the first character from our input
    result = [dictionary[data[:1]]]
    data = data[1:]

    # While we still have uncompressed data...
    while len(data) > 0:
        # Get the last entry in our result list
        previous = result[-1].decode()

        # Check if the next character exists in our dictionary
        try:
            char = data[:1]

            # If yes, append that to our result list
            result.append(dictionary[char])
            
            # And remove it from our input data
            data = data[1:]
            
        except KeyError:
            # Otherwise, create a new key based on the previous two entries
            new_key = previous + previous[0]
            
            # If the new key already exists, use its value instead
            if new_key in dictionary:
                result.append(dictionary[new_key])
                
            else:
                # Append the new key to our result list
                result.append(len(dictionary))
                
                # Add the new key to our dictionary
                dictionary[new_key] = len(dictionary)

                # Remove the old keys from our result list
                del result[:-1]

                # And add them back to our input data
                data = previous + data
                
    # Once we're done, convert our result list to a binary string
    return b''.join([struct.pack('B', x) for x in result])
    
