#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = [struct.pack('>BBBB', *rng.read(4)) for _ in range(num_colors)]
    color_table = b''.join(global_color_table)

    # 4. Graphics Control Extension
    # Frame delay time (using one byte for now)
    frame_delay_time = rng.read(1)[0]

    # Transparent color index field (using one byte for now)
    transparent_color_index = rng.read(1)[0] & 0b1111

    # 5. Image Descriptor
    # Left position of image data (using two bytes for now)
    left = rng.read(1)[0] | ((rng.read(1)[0] & 0b1111) << 8)

    # Top position of image data (using two bytes for now)
    top = rng.read(1)[0] | ((rng.read(1)[0] & 0b1111) << 8)

    # Width of image data (using two bytes for now)
    width = rng.read(1)[0] | ((rng.read(1)[0] & 0b1111) << 8)

    # Height of image data (using two bytes for now)
    height = rng.read(1)[0] | ((rng.read(1)[0] & 0b1111) << 8)

    # 6. Image Data
    # Number of pixels to encode (width times height)
    num_pixels = width * height

    # Encode the pixel values as RGB triplets
    encoded_pixel_values = []
    for i in range(num_pixels):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        encoded_pixel_values += [r, g, b]

    # Write everything to output stream
    out.write(header + lsd + color_table +
              b'\x21\xF9\x04\x00\x00\x00\x00\x00' +
              struct.pack('>BBBBB', frame_delay_time, transparent_color_index, 0, 0, 0) +
              b'\x2C' + struct.pack('>HHHHB', left, top, width, height, 0) +
              b'\x87' + bytes(encoded_pixel_values))