#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])
    
    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        return
    
    size = ((packed & 0x07) + 1) * 3
    color_table = [rng.read(1)[0] for _ in range(size)]
    
    # 4. Graphics Control Extension (if present)
    graphics_control_extension_flag = bool(packed & 0x40)
    if not graphics_control_extension_flag:
        return
    
    transparent_color_index = rng.read(1)[0]
    delay = struct.unpack('>H', rng.read(2))[0]
    transparency_flag = bool(packed & 0x01)
    gce = struct.pack('>BHB?', transparent_color_index, delay, transparency_flag, rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    local_color_table_flag = bool(packed & 0x80)
    interlace_flag = bool(packed & 0x40)
    sort_flag = bool(packed & 0x20)
    reserved = rng.read(2)
    local_color_table_size = ((packed & 0x07) + 1) * 3
    local_color_table = [rng.read(1)[0] for _ in range(local_color_table_size)]
    image_descriptor = struct.pack('>HHBBBBB', left, top, packed, rng.read(1)[0], interlace_flag, sort_flag, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    data = []
    while True:
        block_size = rng.read(1)[0] & 0xFF
        
        if block_size == 0:
            break
        
        data.append(rng.read(block_size))
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)
