#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table
    num_colors = min(max((width * height) // 64 + 1, 2), 256)
    global_color_table = [rng.read(3) for _ in range(num_colors)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBBBBB', ord(','), ord(';'), 0, 0, width, height, 0b11100000 | num_colors, 0)

    # 5. Local Color Table
    local_color_table = []
    if len(global_color_table) < 256:
        local_color_table += [(0, 0, 0)] * ((256 - len(global_color_table)))

    # 6. Graphics Control Extension
    graphics_control_extension = struct.pack('>BBBBBBHBB', ord('!'), ord('['), 4, 1, 0, 0, 0, 0, 0)

    # 7. Application Extension
    application_extension = struct.pack('>BBBBBB', ord('!'), ord(']'), 11, 0, 11, ord('A')) + b'NETSCAPE2.0\x03\x01\x00\x00'

    # 8. Comment Extension
    comment_extension = struct.pack('>BBBBBB', ord('!'), ord('L'), 0, 0, 0, 0)

    # 9. Plain Text Extension
    plain_text_extension = struct.pack('>BBBBBB', ord('!'), ord('T'), 0, 0, 0, 0)

    # 10. Trailer
    trailer = b';'

    # Write everything to output stream
    out.write(header)
    out.write(lsd)
    out.write(bytes([*sum(global_color_table, ()), *local_color_table]))
    out.write(image_descriptor)
    out.write(graphics_control_extension)
    out.write(application_extension)
    out.write(comment_extension)
    out.write(plain_text_extension)
    out.write(trailer)

