#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors in RGB format

    # 4. Image Descriptor
    # Left position of image is always zero
    left = struct.pack('<H', 0)

    # Top position of image is also always zero
    top = struct.pack('<H', 0)

    # Width and Height are again randomly generated
    img_width, img_height = struct.unpack('>HH', rng.read(4))

    # Flags field: packed fields (using a random byte for variety).
    flags = struct.pack('>B', rng.read(1)[0])

    # Interlace flag is set if bit 7 of the flags field is set to 1
    interlace = bool((flags[0] & 0b10000000) >> 7)

    # Sort flag is set if bit 6 of the flags field is set to 1
    sort = bool((flags[0] & 0b01000000) >> 6)

    # Local color table flag is set if bit 7 of the flags field is set to 1
    local_color_table = bool((flags[0] & 0b00100000) >> 5)

    # Size of local color table is determined by the value of bits 0-2 of the flags field
    local_color_table_size = ((flags[0] & 0b00011100) >> 2) + 1

    # Reserved field is always empty
    reserved = struct.pack('>B', 0)

    # LZW minimum code size is determined by the value of bits 0-2 of the flags field
    min_code_size = (flags[0] & 0b00000011) + 1

    # 5. Extension Block
    ext_block = b'\x21\xFF\x0BNETSCAPE2.0\x03\x01\x00\x00'

    # 6. Graphics Control Extension
    gce = b'\x21\xF9\x04'

    # Transparency flag is set if bit 3 of the GCE block is set to 1
    transparency = bool((gce[-1] & 0b00001000) >> 3)

    # User input flag is set if bit 2 of the GCE block is set to 1
    user_input = bool((gce[-1] & 0b00000100) >> 2)

    # Delay time is determined by the lower 12 bits of the GCE block
    delay_time = gce[-1:] + rng.read(1)

    # Transparent color index is determined by the upper 8 bits of the GCE block
    transparent_index = gce[:-1] + rng.read(1)

    # 7. Image Data
    # Number of pixels in the image
    pixel_count = img_width * img_height

    # LZW minimum code size is rounded up to the nearest power of two
    min_code_size = pow(2, int(min_code_size))

    # The first code is the clear code
    codes = [0]

    # The second code is the end of information code
    codes += list(range(1, min_code_size - 1))

    # Remaining codes are filled with random numbers until we reach the minimum code size
    while len(codes) < min_code_size:
        codes += [rng.read(1)[0]]

    # A dictionary mapping characters to their corresponding codes
    char_to_code = {chr(i): i for i in range(len(codes))}

    # A dictionary mapping codes to their corresponding characters
    code_to_char = dict([(v, k) for k, v in char_to_code.items()])

    # The current code is initially the clear code
    curr_code = 0

    # The current character is initially the null character
    curr_char = '\x00'

    # The data stream starts with the clear code
    data = [curr_code]

    # We will use this variable as an offset when reading from the RNG
    offset = 0

    # Read enough bytes to fill our image
    while len(data) <= pixel_count:

        # If the current code is less than the minimum code size, then read another byte
        if curr_code < min_code_size:
            new_byte = rng.read(1)[0]

            # Add the new byte to the data stream
            data += [(new_byte + offset) % 256]

            # Increment the offset so that subsequent reads don't result in duplicate values
            offset += 1

            # Update the current character based on the new byte
            curr_char = chr(new_byte)

        # Otherwise, the current code is equal to the minimum code size
        else:
            # Append the current character to the data stream
            data += [ord(curr_char)]

        # Get the next code from the data stream
        next_code = data[offset]

        # If the next code exists in the dictionary, update the current code and current character accordingly
        if next_code in code_to_char:
            curr_code = next_code
            curr_char = code_to_char[next_code]

        # Otherwise, get the next character from the RNG
        elif next_code not in codes:
            new_byte = rng.read(1)[0]

            # Add the new byte to