#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed_flags = rng.read(1)[0]
    global_color_table_flag = bool(packed_flags & 0b10000000)
    color_resolution = ((packed_flags >> 7) & 0b111) + 1
    sort_flag = bool(packed_flags & 0b00000100)
    global_color_table_size = 2 ** (((packed_flags >> 4) & 0b111) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    logical_screen_descriptor = struct.pack(
        '>HBBBBB', width, height, packed_flags,
        background_color_index, pixel_aspect_ratio
    )

    # 3. Global Color Table
    color_table = b''
    if global_color_table_flag:
        for i in range(global_color_table_size):
            r, g, b = [rng.read(1)[0] for _ in range(3)]

            color_table += struct.pack('>BBB', r, g, b)

    # 4. Graphics Control Extension
    if rng.read(1)[0]:  # Only add this extension sometimes
        delay_time = struct.unpack('>H', rng.read(2))[0]
        transparent_color_index = rng.read(1)[0]
        flags = rng.read(1)[0]
        gce = struct.pack(
            '>BBBBHB',
            0xF9,  # Application Extension Block Label
            4,     # Size of application data
            0xFF,  # Application Identifier
            flags,
            delay_time,
            transparent_color_index
        )
    else:
        gce = None

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)