#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>BB', rng.read(2))
    lsd = struct.pack('>BBBBBBBHB', *struct.unpack('>BBBBBBBH', rng.read(7)), width, height)
    
    # 3. Global Color Table
    global_color_table_flag = bool(rng.read(1)[0] & 0b10000000)
    color_resolution = ((rng.read(1)[0] & 0b01110000) >> 4) + 1
    sort_flag = bool(rng.read(1)[0] & 0b00001000)
    global_color_table_size = 2 ** (((rng.read(1)[0] & 0b00000111)))
    background_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    if global_color_table_flag:
        color_table = bytes()
        for i in range(global_color_table_size):
            r, g, b = [rng.read(1)[0] for _ in range(3)]
            
            color_table += struct.pack('>BBB', r, g, b)
            
    else:
        color_table = None
    
    # 4. Graphics Control Extension
    has_gce = bool(rng.read(1)[0] & 0b10000000)
    
    if has_gce:
        delay_time = struct.unpack('>H', rng.read(2))[0]
        transparent_color_index = rng.read(1)[0]
        terminator = rng.read(1)[0]
        
        assert terminator == 0
        
        gce = struct.pack('>BHBB', 0xF9, delay_time, transparent_color_index, terminator)
    else:
        gce = None
        
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        
        if block_size == 0:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)