#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])
    
    # 3. Global Color Table
    global_color_table_size = (packed & 7) + 1
    global_color_table = bytearray()
    for i in range(global_color_table_size):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        global_color_table += bytes([r,g,b])

    # 4. Image Descriptor
    image_left, image_top = struct.unpack('>HH', rng.read(4))
    image_width, image_height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    local_color_table_flag = bool((packed >> 7) & 1)
    interlace_flag = bool((packed >> 6) & 1)
    sort_flag = bool((packed >> 5) & 1)
    reserved_bits = ((packed << 2) & 0xFC) | (((packed >> 2) & 3) ^ 3)
    lct_size = (packed & 7) + 1
    image_descriptor = struct.pack('>HHBBB', image_left, image_top, image_width, image_height, packed)

    # 5. Local Color Table
    if local_color_table_flag:
        local_color_table = bytearray()
        for i in range(lct_size):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            local_color_table += bytes([r,g,b])

    # 6. Block Size
    block_size = rng.read(1)[0]

    # 7. Data Sub-blocks
    data_sub_blocks = []
    while True:
        subblock_type = rng.read(1)[0]

        if subblock_type == 0x2C:
            # Image Data
            data_sub_blocks.append(struct.pack('B', subblock_type))

            # LZW Minimum Code Size
            minimum_code_size = rng.read(1)[0]
            data_sub_blocks.append(bytes([minimum_code_size]))
            
            # Image Data
            data_length = len(data_sub_blocks[-1]) + 1
            while data_length < block_size - 1:
                code_byte = rng.read(1)[0]

                if code_byte!= 0:
                    data_sub_blocks.append(bytes([code_byte]))
                    data_length += 1
                else:
                    count = rng.read(1)[0]

                    if count > 128:
                        count -= 128

                        char = data_sub_blocks[-1][-1:] * count
                        data_sub_blocks.append(char)
                        data_length += len(char)
                    elif count > 0:
                        count += 1
                        
                        char = data_sub_blocks[-1][-1:]
                        data_sub_blocks.append(char)
                        data_length += len(char)
                    
                    break

            assert data_length <= block_size - 1
        
        elif subblock_type == 0x21:
            # Extension
            data_sub_blocks.append(struct.pack('B', subblock_type))

            # Extension Label
            label = rng.read(1)[0]
            data_sub_blocks.append(bytes([label]))

            # Extension Data
            length = rng.read(1)[0]
            extension_data = rng.read(length)
            data_sub_blocks.append(extension_data)

        elif subblock_type == 0x3B:
            # Trailer
            data_sub_blocks.append(struct.pack('B', subblock_type))
            break
        
        else:
            raise Exception("Invalid GIF subblock type")

    # 8. Output
    out.write(header)
    out.write(lsd)
    out.write(global_color_table)
    out.write(image_descriptor)
    if local_color_table_flag:
        out.write(local_color_table)
    for subblock in data_sub_blocks:
        out.write(subblock)

