#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
	width &= 0x7FFF
	height &= 0x7FFF

    # Color table size is fixed to 256 colors
    color_table_size = 2 << 8 | 2 << 4 | 1 << 0

    # Global color table flag set to True
    global_color_table_flag = 1 << 7

    # Sorting flag set to False
    sorting_flag = 0 << 5

    # Size of the color table in bytes
    color_table_size = 2 << 8 | 2 << 4 | 1 << 0

    # Background color index is always zero
    background_color_index = 0

    # Pixel aspect ratio is always 1
    pixel_aspect_ratio = 1

    # Pack all values together
    lsd = struct.pack(
        '>HHBBBHBB',
        width,
        height,
        color_table_size,
        global_color_table_flag,
        sorting_flag,
        background_color_index,
        pixel_aspect_ratio
    )

    # Write everything to output
    out.write(header + lsd)

    # 3. Image descriptor
    image_separator = b','

    # Left position is always zero
    left = 0

    # Top position is always zero
    top = 0

    # Width and height are equal to the logical screen dimensions
    width = width
    height = height

    # Packed field with no flags set
    packed_field = 0

    # LZW minimum code size is always 2 bits
    lzw_minimum_code_size = 2

    # Write it to output
    out.write(image_separator + struct.pack(
        '>HHHHB',
        left,
        top,
        width,
        height,
        packed_field,
        lzw_minimum_code_size
    ))

    # 4. Local color table
    # We use the same palette as the one used by the original GIFs
    local_color_table = [
        0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x80, 0x80, 0x00, 0x00, 0x80, 0x00, 0x80, 0x00, 0x00, 0x80, 0x80,
        0xC0, 0xC0, 0xC0, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00, 0x00,
        0xFF, 0x00, 0xFF, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0xFF, 0x00, 0xFF,
        0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0x00, 0xFF, 0xFF, 0x00, 0xFF, 0xFF,
        0xFF, 0xFF, 0xFF, 0xFF, 0x1A, 0x1A, 0x1A, 0x2E, 0x2E, 0x2E, 0x4D, 0x4D,
        0x4D, 0x6C, 0x6C, 0x6C, 0x8B, 0x8B, 0x8B, 0xAA, 0xAA, 0xAA, 0xC9, 0xC9,
        0xC9, 0xE8, 0xE8, 0xE8, 0x07, 0x07, 0x07, 0x26, 0x26, 0x26, 0x45, 0x45,
        0x45, 0x64, 0x64, 0x64, 0x83, 0x83, 0x83, 0xA2, 0xA2, 0xA2, 0xC1, 0xC1,
        0xC1, 0xE0, 0xE0, 0xE0, 0x0F, 0x0F, 0x0F, 0x2E, 0x2E, 0x2E, 0x4D, 0x4D,
        0x4D, 0x6C, 0x6C, 0x6C, 0x8B, 0x8B, 0x8B, 0xAA, 0xAA, 0xAA, 0xC9, 0xC9,
        0xC9, 0xE8, 0xE8, 0xE8, 0x07, 0x07, 0x07, 0x26, 0x26, 0x26, 0x45, 0