#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if bool(rng.read(1)):
        # We'll use this to store some flags later on
        gce = b''

        # Flags field: packed fields (using a random byte for variety)
        gce += struct.pack('BBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

        # Delay time (two bytes)
        delay_time = rng.read(2)

        # Transparent color index (one byte)
        transparent_color_index = rng.read(1)

        # Terminator (one byte)
        gce += b'\x00'

        # Now we can add all of these together
        gce = b'\x21\xF9' + struct.pack('>HBB', len(gce), *delay_time, *transparent_color_index) + gce

    else:
        gce = b''

    # 5. Image Descriptor
    image_descriptor = struct.pack('>HHB', 0, 0, 0b10000000 | rng.read(1)[0])

    # 6. Image Data
    image_data = b''

    # Let's just make it a bunch of pixels with random values
    while True:
        pixel = rng.read(1)[0]

        # If the first bit is set, then we're done
        if not pixel & 0b10000000:
            break

        # Otherwise, let's keep adding pixels until we get one that doesn't start with the first bit set
        image_data += struct.pack('B', pixel)

    # 7. Trailer
    trailer = b';'

    # Write everything in order
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)