#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    
    global_color_table_flag = bool((packed >> 7) & 1)
    color_resolution = ((packed >> 4) & 7) + 1
    sort_flag = bool((packed >> 3) & 1)
    size_of_global_color_table = 2 ** (packed & 7)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    lsd = struct.pack('>HHBBB', width, height, packed, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    color_table = bytearray()
    
    for i in range(size_of_global_color_table):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        
        color_table += struct.pack('BBB', r, g, b)
    
    # 4. Graphics Control Extension
    packed = rng.read(1)[0]
    disposal_method = (packed >> 2) & 7
    user_input_flag = bool((packed >> 1) & 1)
    transparent_color_flag = bool(packed & 1)
    delay_time = int.from_bytes(rng.read(2), 'big')
    transparent_color_index = rng.read(1)[0]
    
    gce = struct.pack('BBHBB', 0x21, 0xf9, 4, packed, delay_time)
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    
    local_color_table_flag = bool((packed >> 7) & 1)
    interlace_flag = bool((packed >> 6) & 1)
    sorted_flag = bool((packed >> 5) & 1)
    reserved_bits = (packed >> 3) & 3
    size_of_local_color_table = 2 ** (packed & 7)
    local_color_table_initalizer = None
    
    if local_color_table_flag and size_of_local_color_table > 0:
        local_color_table_initalizer = bytearray()
        
        for i in range(size_of_local_color_table):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            
            local_color_table_initalizer += struct.pack('BBB', r, g, b)
    
    lzw_min_code_size = rng.read(1)[0]
    
    image_descriptor = struct.pack('>HHB', left, top, packed)
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    data = []
    while True:
        block_size = rng.read(1)[0] & 0xFF
        
        if block_size == 0:
            break
        
        data.append(rng.read(block_size))
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)
