#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    #    - Signature (3 bytes)
    #    - Version (3 bytes)
    #    - Width (2 bytes)
    #    - Height (2 bytes)
    #    - Flags (1 byte)
    #    - Background Color Index (1 byte)
    #    - Pixel Aspect Ratio (1 byte)

    signature = b'GIF'
    version = b'\x08\x00\x00'
    width = struct.pack('>H', rng.randint(16, 4096))
    height = struct.pack('>H', rng.randint(16, 4096))
    flags = rng.read(1)[0] & 0b11100000 | \
            rng.read(1)[0] & 0b00011100 | \
            rng.read(1)[0] & 0b00000011
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    header = signature + version + width + height + \
             bytes([flags]) + bytes([background_color_index]) + \
             bytes([pixel_aspect_ratio])
    out.write(header)


    # 2. Logical Screen Descriptor
    #    - Size of Global Color Table (1 byte)
    #    - Color Resolution (1 bit)
    #    - Sort Flag (1 bit)
    #    - Size of Global Color Table (3 bits)
    #    - Background Color Index (1 byte)
    #    - Pixel Aspect Ratio (1 byte)

    size_of_global_color_table = rng.read(1)[0] & 0b00000111
    color_resolution = ((rng.read(1)[0] & 0b10000000) >> 7) * 256
    sort_flag = ((rng.read(1)[0] & 0b01000000) >> 6)!= 0
    global_color_table_size = size_of_global_color_table * 3
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    logical_screen_descriptor = bytes([size_of_global_color_table]) + \
                                bytes([(color_resolution << 7) |
                                       (sort_flag << 6)]) + \
                                bytes([global_color_table_size]) + \
                                bytes([background_color_index]) + \
                                bytes([pixel_aspect_ratio])
    out.write(logical_screen_descriptor)


    # 3. Global Color Table
    for _ in range(global_color_table_size):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(bytes([r]))
        out.write(bytes([g]))
        out.write(bytes([b]))


    # 4. Image Descriptor
    #    - Left Position (2 bytes)
    #    - Top Position (2 bytes)
    #    - Width (2 bytes)
    #    - Height (2 bytes)
    #    - Flags (1 byte)
    #    - Local Color Table Flag (1 bit)
    #    - Interlace Flag (1 bit)
    #    - Sort Flag (1 bit)
    #    - Reserved (3 bits)
    #    - Size of Local Color Table (3 bits)
    #    - Background Color Index (1 byte)
    #    - Pixel Aspect Ratio (1 byte)

    left_position = struct.pack('>H', rng.randint(0, 4096))
    top_position = struct.pack('>H', rng.randint(0, 4096))
    width = struct.pack('>H', rng.randint(16, 4096))
    height = struct.pack('>H', rng.randint(16, 4096))
    flags = rng.read(1)[0] & 0b11100000 | \
            rng.read(1)[0] & 0b00011100 | \
            rng.read(1)[0] & 0b00000011
    local_color_table_flag = ((rng.read(1)[0] & 0b10000000) >> 7)!= 0
    interlace_flag = ((rng.read(1)[0] & 0b01000000) >> 6)!= 0
    sort_flag = ((rng.read(1)[0] & 0b00100000) >> 5)!= 0
    reserved = rng.read(1)[0] & 0b00011100
    local_color_table_size = (rng.read(1)[0] & 0b00000111) * 3
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    image_descriptor = left_position + top_position + width + height + \
                       bytes([flags]) + bytes([(local_