#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        # The last block is always empty
        if not block_size:
            break

        # We'll use a minimum code size of 2 so that our data will never get too large
        if len(blocks):
            block_size &= ~1

        # Add some padding to the end of this block
        block_size += rng.read(1)[0] & 1

        # Add the block itself
        blocks.append(rng.read(block_size))

    # Compress all the blocks together with LZW compression
    compressed_data = compress_lzw(b''.join(blocks), lzw_minimum_code_size)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything in order
    out.write(header + lsd + color_table + gce + image_descriptor + compressed_data + trailer)


# Compresses data using Lempel–Ziv–Welch algorithm
# https://en.wikipedia.org/wiki/Lempel%E2%80%93Ziv%E2%80%93Welch
def compress_lzw(
        data: bytes, min_code_size=2
    ) -> bytes:
    # Initialize dictionary with single characters
    dict_size = 2 ** min_code_size
    dictionary = {bytes([i]): i for i in range(dict_size)}
    dict_size += 1

    # Output buffer
    output = [min_code_size - 1]

    # Current prefix
    current_prefix = None

    # Iterate over input data
    for char in data:
        # Get character as bytes
        char = bytes([char])

        # Check if it's already in the dictionary
        if char in dictionary:
            current_prefix = char

            continue

        # If there was no previous prefix, just write the new character
        if not current_prefix:
            output.append(dictionary[char])
            current_prefix = char
            continue

        # Otherwise, try to find a matching suffix in the dictionary
        for key in list(dictionary.keys()):
            # Skip the first entry because it's the empty string
            if key == b'':
                continue

            # If the first byte matches, check the rest
            if key[:len(current_prefix)] == current_prefix:
                # If they match, add them to the dictionary
                dictionary[current_prefix + char] = dictionary[key]

                # And update the current prefix
                current_prefix = current_prefix + char

                # Stop iterating through the dictionary keys
                break

        # If we didn't find any matching suffixes, add a new entry to the dictionary
        else:
            dictionary[current_prefix + char] = dict_size
            dict_size += 1

            output.append(dictionary[current_prefix])
            current_prefix = char

    # Flush the current prefix
    if current_prefix:
        output.append(dictionary[current_prefix])

    return bytes(output)


if