#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng):
    # Header
    out.write('GIF89a'.encode())

    # Logical screen descriptor
    width = 1 + rng.randint(256 - 1)
    height = 1 + rng.randint(256 - 1)
    packed = (width & 0xFF) | ((height & 0xFF) << 8)
    packed |= bool(rng.getrandbits(1)) << 12
    packed |= bool(rng.getrandbits(1)) << 13
    packed |= bool(rng.getrandbits(1)) << 14
    packed |= bool(rng.getrandbits(1)) << 15
    packed |= (7 & rng.randint(7)) << 16
    out.write(struct.pack('<H', packed))

    # Global color table
    for i in range(2**packed & 0x7FFF):
        r = rng.randint(256)
        g = rng.randint(256)
        b = rng.randint(256)
        out.write(struct.pack('<BBB', r, g, b))

    # Image descriptor
    packed = 0b10000000 | (0 & 0xF) | ((0 & 0xF) << 4)
    out.write(struct.pack('<BHHBB', packed, 0, 0, 0, 0))

    # Local color table
    if not (packed & 0b10000000):
        for i in range((2**(packed >> 4) & 0x7F)):
            r = rng.randint(256)
            g = rng.randint(256)
            b = rng.randint(256)
            out.write(struct.pack('<BBB', r, g, b))

    # Image data
    while True:
        block_type = rng.randint(4)

        if block_type == 0:
            break

        elif block_type == 1:
            size = rng.randint(256)

            if size!= 0:
                raise Exception("Invalid LZW minimum code size")

        else:
            size = 1 + rng.randint(255)

            if block_type == 2 and size > 255:
                raise Exception("Invalid LZW clear code length")

            if block_type == 3 and size > 256:
                raise Exception("Invalid LZW end of information code length")

            if block_type == 2 or block_type == 3:
                continue

            codes = [i for i in range(size)]
            next_code = len(codes)

            for _ in range(size):
                codes += [(rng.randint(256), rng.randint(256))]

            while True:
                index = rng.randint(next_code)

                if index < size:
                    break

                out.write(bytes([index]))

                if index == next_code:
                    codes += [(rng.randint(256), rng.randint(256))]
                    next_code += 1

                    if next_code >= 4096:
                        raise Exception("Too many codes")

                else:
                    index1 = index
                    index2 = index1 ^ (index1 >> 1)

                    if index2 <= max(codes):
                        raise Exception("Invalid LZW code")

                    out.write(bytes([(index2 & 0xFF00) >> 8]))
                    out.write(bytes([index2 & 0xFF]))

                    if index1 == next_code:
                        break

                    codes[next_code] = codes[index1]
                    next_code += 1

                    if next_code >= 4096:
                        raise Exception("Too many codes")

    # Trailer
    out.write(b'\x3B')

