#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    gct_size = min((rng.read(1)[0] & 0x7F) + 1, 256)
    gct = [struct.pack('BBB', *rng.read(3)) for _ in range(gct_size)]

    # 4. Image Descriptor
    id = struct.pack('>HHBHB', 0, 0, 0xF7 | ((rng.read(1)[0] & 0x07) << 5), 0, 0)

    # 5. The actual image data
    imgdata = []
    while True:
        # Each frame is made of up to 255 blocks
        block_count = rng.read(1)[0]

        if not block_count:
            break
        
        # Each block has a prefix that tells us how long it will be
        block_lengths = [rng.read(1)[0] for _ in range(block_count)]

        # And then comes the actual data
        block_data = [rng.read(length) for length in block_lengths]

        # We can't just concatenate all these together because we need to add an end-of-image marker
        imgdata += block_lengths + [b'\x00'] + block_data
    
    # Finally, let's write everything to our output stream
    out.write(header + lsd + b''.join(gct) + id + b''.join(imgdata))