#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    def get_block():
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        # If we're done here, just return an empty block
        if not block_size:
            yield b'\x00'
            return

        # Otherwise, let's do some fun stuff!
        data = [rng.read(1)[0]]
        while True:
            if len(data) == block_size:
                break

            # Let's try to add more data from our previous block...
            if rng.read(1)[0]:
                data += list(get_block())

                # We might need to pad this block so that its length is exactly equal to what we want
                padding = block_size - len(data)
                if padding > 0:
                    data += [rng.read(1)[0] for i in range(padding)]

                break

            #...or maybe we should start a new one?
            elif rng.read(1)[0]:
                data += list(get_block())
            else:
                data.append(rng.read(1)[0])

        yield bytes([len(data)]) + bytes(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(list(get_block())) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)