#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# Returns the size of the generated file in bytes.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:

    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])
    out.write(lsd)

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        return len(header + lsd)

    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = rng.read(3 * num_colors)  # Random colors
    out.write(color_table)

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    out.write(gce)

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    out.write(image_descriptor)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    data = []
    while True:
        block_size = rng.read(1)[0] & 0xFF

        if block_size == 0:
            break

        data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'
    out.write(trailer)

    return len(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)