#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generate a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF will have no global color table,
# but may contain one graphics control extension.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:

    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)
    out.write(lsd)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    out.write(color_table)

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
        out.write(gce)

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = rng.read(local_color_table_size * 3)
    image_descriptor = struct.pack('>BHHHHBBBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved,
                                  local_color_table_size)
    out.write(image_descriptor)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    out.write(image_data)

    # 7. Trailer
    trailer = b'\x3B'
    out.write(trailer)

    return len(header) + len(lsd) + len(color_table) + \
           (len(gce) if has_gce else 0) + len(image_descriptor) + \
           len(image_data) + len(trailer)