#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Image Data
    # First we need to know how many frames there are going to be
    num_frames = (rng.read(1)[0] % 256) + 1
    
    # Then we need to know how long each frame is going to be
    frame_lengths = [rng.read(1)[0] % 256 for _ in range(num_frames)]
    
    # Now we can start writing the image data
    out.write(header + lsd + color_table)
    
    # We will write the first frame as an "application extension"
    # This is just a way of storing some arbitrary data with the image
    # In this case, we will store the number of frames in the image
    app_ext = struct.pack('>BBHB', 0xFF, 0xF9, 0x04, num_frames)
    out.write(app_ext)
    
    # Now we will write the actual frames
    for i in range(num_frames):
        # Each frame starts with a Graphic Control Extension
        gce = struct.pack('>BBBBB', 0x21, 0xF9, 0b10000000, 0, 0)
        
        # The delay time is stored as a 16-bit unsigned integer
        # We use the same trick as above to get a random value between 0 and 65535
        delay = struct.pack('>H', (rng.read(1)[0] % 256) << 8 | rng.read(1)[0])
        
        # The transparent color index is also stored as a single byte
        transp = rng.read(1)
        
        # Finally, we add all these parts together to form the Graphic Control Extension
        gce += delay + transp
        
        # Next comes the Image Descriptor
        idesc = struct.pack('>BBHHB', 0x2C, 0, 0, width, height, 0)
        
        # And finally, we write the frame itself
        out.write(gce + idesc + rng.read(frame_lengths[i]))
