
import struct
from PIL import Image

def gif2rgb565(im):
    """Convert a GIF to RGB565"""
    
    im = im.convert('RGB')
    w, h = im.size
    pixels = im.tobytes()
    
    rgb565 = bytearray(w * h * 2)
    for i in range(0, len(pixels), 3):
        r, g, b = pixels[i:i+3]
        
        rgb565[i//3*2]   = ((r & 0xf8) << 8) | ((g & 0xfc) >> 2)
        rgb565[i//3*2+1] = ((g & 0x03) << 6) | ((b & 0xf8) >> 3)
        
    return bytes(rgb565)

def make_gif(width, height, data, delay=None, loop=None, transparent=None):
    """Make a GIF from raw RGB565 data"""
    
    if not isinstance(data, bytes):
        raise TypeError("Data must be bytes")
    
    if width <= 0 or height <= 0:
        raise ValueError("Invalid dimensions")
    
    if len(data) != width * height * 2:
        raise ValueError("Invalid data length")
    
    if delay is None:
        delay = 0
    elif delay < 0 or delay > 65535:
        raise ValueError("Delay out of range")
    
    if loop is None:
        loop = 0
    elif loop < 0 or loop > 65535:
        raise ValueError("Loop count out of range")
    
    if transparent is None:
        transparent = False
    elif transparent and (transparent < 0 or transparent >= 2**16):
        raise ValueError("Transparent color index out of range")
    
    # Header
    header = b'\x47\x49\x46\x38\x39\x61'
    
    # Logical Screen Descriptor
    lsd = struct.pack("<HHBB", width, height, 0x80, 0)
    
    # Global Color Table
    color_table = b'\xff\xff\xff' + b'\x00\x00\x00' * 254
    
    # Graphic Control Extension
    gce = None
    if transparent:
        gce = struct.pack("<BBBHB", 0xF9, 0x04, 0x01, delay, transparent)
    
    # Image Descriptor
    image_descriptor = struct.pack("<BBHHBB", 0x2C, 0, 0, width, height, 0)
    
    # Image Data
    image_data = b'\x00' + data
    
    # Trailer
    trailer = b'\x3B'
    
    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)

