#!/usr/bin/env python3

import struct
from typing import BinaryIO


def read_image(f: BinaryIO):
    header = f.read(6)  # GIF87a or GIF89a
    if header[:6] not in [b'GIF87a', b'GIF89a']:
        raise ValueError('Invalid GIF file')

    width, height = struct.unpack("<HH", f.read(4))
    flags = ord(f.read(1))
    background_color_index = ord(f.read(1))
    aspect_ratio = ord(f.read(1))

    global_color_table_flag = (flags & 0x80) != 0
    color_resolution = ((flags >> 4) & 0x7) + 1
    sort_flag = (flags & 0x8) != 0
    size_of_global_color_table = 2 ** (((flags & 0x7) + 1) + 1)

    if global_color_table_flag:
        global_color_table = f.read(size_of_global_color_table * 3)
    else:
        global_color_table = None

    while True:
        block_type = ord(f.read(1))
        if block_type == 0x2C:  # Image Descriptor
            break
        elif block_type == 0x21:  # Extension
            extension_label = ord(f.read(1))
            if extension_label == 0xF9:  # Graphic Control Extension
                block_length = ord(f.read(1))
                assert block_length == 4
                packed = ord(f.read(1))
                disposal_method = (packed & 0x1C) >> 2
                user_input_flag = (packed & 0x2) != 0
                transparent_color_flag = (packed & 0x1) != 0
                delay_time = struct.unpack("!H", f.read(2))[0]
                transparent_color_index = ord(f.read(1))
            else:
                f.seek(block_length - 1, 1)
        else:
            raise ValueError('Unknown block type %d' % block_type)

    left, top, width, height = struct.unpack("<HHHH", f.read(8))
    local_color_table_flag = (ord(f.read(1)) & 0x80) != 0
    interlace_flag = (ord(f.read(1)) & 0x40) != 0
    sort_flag = (ord(f.read(1)) & 0x8) != 0
    size_of_local_color_table = 2 ** ((ord(f.read(1)) & 0x7) + 1)

    if local_color_table_flag:
        local_color_table = f.read(size_of_local_color_table * 3)
    else:
        local_color_table = global_color_table

    lzw_minimum_code_size = ord(f.read(1))
    data = []
    code_size = lzw_minimum_code_size + 1
    clear_code = 1 << lzw_minimum_code_size
    end_of_information_code = clear_code + 1
    available_codes = clear_code + 2
    old_code = None
    code_buffer = []
    for i in range(available_codes):
        prefix[i] = None
        suffix[i] = chr(i)
        first_code[i] = i
        next_code[i] = None

    code = struct.unpack("B", f.read(1))[0]
    while code != end_of_information_code:
        if code == clear_code:
            code_size = lzw_minimum_code_size + 1
            code_mask = (1 << code_size) - 1
            available_codes = clear_code + 2
            old_code = None
        elif code > available_codes or old_code is None:
            raise ValueError('Corrupt LZW data')
        elif code == available_codes:
            if len(code_buffer) <= code_size:
                raise ValueError('Corrupt LZW data')
            code_buffer.append(first_code[old_code])
            code = first_code[old_code]
        else:
            if code_buffer:
                code_buffer.append(suffix[code])
            else:
                code_buffer.append(prefix[code])
            old_code = code
            code = first_code[old_code]

        if len(code_buffer) >= code_size:
            fragment = ''.join(code_buffer[-code_size:])
            code_buffer = code_buffer[:-code_size]
            data.append(fragment)
            if available    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)