#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed_fields = rng.read(1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)
    lsd = width + height + packed_fields + background_color_index + pixel_aspect_ratio

    # 3. Global Color Table
    global_color_table_flag = (packed_fields[0] >> 7) & 1
    color_resolution = ((packed_fields[0] >> 4) & 7) + 1
    sort_flag = (packed_fields[0] >> 3) & 1
    global_color_table_size = 2 ** ((packed_fields[0] & 7) + 1)
    global_color_table = rng.read(global_color_table_size * 3)
    color_table = b''
    if global_color_table_flag:
        color_table += global_color_table

    # 4. Graphic Control Extension
    gce = None
    if rng.read(1)[0]:
        delay_time = rng.read(2)
        transparent_color_index = rng.read(1)
        disposal_method = rng.read(1)
        user_input_flag = rng.read(1)
        gce = b'\x21\xF9\x04' + delay_time + transparent_color_index + disposal_method + user_input_flag

    # 5. Image Descriptor
    left_position = rng.read(2)
    top_position = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    local_color_table_flag = (rng.read(1)[0] >> 7) & 1
    interlace_flag = (rng.read(1)[0] >> 6) & 1
    sort_flag = (rng.read(1)[0] >> 5) & 1
    reserved = rng.read(1)
    local_color_table_size = 2 ** ((rng.read(1)[0] & 7) + 1)
    image_descriptor = b'\x2C' + left_position + top_position + width + height + \
                       bytes([local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved[0]])

    # 6. Local Color Table
    if local_color_table_flag:
        color_table += rng.read(local_color_table_size * 3)

    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')