#!/usr/bin/env python3

import struct
from typing import BinaryIO

def read_image(f: BinaryIO):
    """Reads an animated GIF from a file and returns the individual frames as byte strings."""
    header = f.read(6)
    if not header.startswith(b'GIF'):
        raise ValueError('Not a valid GIF file')
    width, height = struct.unpack('<HH', f.read(4))
    flags = ord(f.read(1))
    background = ord(f.read(1))
    aspect_ratio = ord(f.read(1))
    global_color_table = bool(flags & 0x80)
    color_resolution = ((flags >> 4) & 7) + 1
    sort_flag = bool((flags >> 3) & 1)
    global_color_count = 2**((flags & 7) + 1)
    global_color_table_size = 3 * global_color_count
    if global_color_table:
        global_color_table = f.read(global_color_table_size)
    else:
        global_color_table = None
    while True:
        block = f.read(1)
        if not block:
            break
        label = ord(block)
        if label == 0x2c: # Image Descriptor
            image_left, image_top, image_width, image_height = struct.unpack('<HHHH', f.read(8))
            local_color_table = bool(ord(f.read(1)) & 0x80)
            interlaced = bool(ord(f.read(1)) & 0x40)
            sorted = bool(ord(f.read(1)) & 0x20)
            reserved = bool(ord(f.read(1)) & 0x10)
            local_color_count = 2**((ord(f.read(1)) & 7) + 1)
            local_color_table_size = 3 * local_color_count
            if local_color_table:
                local_color_table = f.read(local_color_table_size)
            else:
                local_color_table = None
            lzw_minimum_code_size = ord(f.read(1))
            yield (image_left, image_top, image_width, image_height), \
                  (lzw_minimum_code_size, global_color_table, local_color_table)
            while True:
                block = f.read(1)
                if not block:
                    return
                label = ord(block)
                if label == 0x21: # Extension
                    extension_label = ord(f.read(1))
                    if extension_label == 0xf9: # Graphic Control Extension
                        f.seek(5, 1)
                    elif extension_label == 0xff: # Application Extension
                        f.seek(13, 1)
                    else:
                        raise ValueError('Unknown extension type %d' % extension_label)
                elif label == 0x3b: # Trailer
                    return
                else:
                    raise ValueError('Unexpected block type %d' % label)

def write_gif(out: BinaryIO, images, delay=None, loop=False):
    """Writes an animated GIF to a file given a list of images."""
    if len(images) < 2:
        raise ValueError('Need at least two images for an animation')
    if delay is None:
        delay = [0] * len(images)
    if len(delay) != len(images):
        raise ValueError('Number of delays must match number of images')
    if any(d > 65535 for d in delay):
        raise ValueError('Delay must be between 0 and 65535')
    if loop:
        loop = 0
    else:
        loop = 1
    out.write(b'GIF89a')
    out.write(struct.pack('<HHHBB', *images[0][0]))
    out.write(b'\x80\x00\x00\x00\x00\x00')
    out.write(images[0][1][1])
    out.write(b'\x2c\x00\x00\x00\x00\x00\x00')
    out.write(struct.pack('B', images[0][1][0]))
    out.write(images[0][1][2])
    out.write(images[0][1][3])
    out.write(b'\x00\x00\x00')
    for i in range(len(images)-1):
        image, delay = images[i+1], delay[i]
        out.write(b'\x21\xf9\x04\x04\x00\x00\x    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)

