#!/usr/bin/env python3

import struct
from typing import BinaryIO


def generate_gif(out: BinaryIO, rng):
    """Generate a GIF from a RNG"""
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = rng.read(2)[0]
    height = rng.read(2)[0]
    packed = rng.read(1)[0]
    global_color_table_flag = bool(packed & 0b10000000)
    color_resolution = (packed >> 4) & 0b111
    sort_flag = bool(packed & 0b100)
    global_color_table_size = (packed & 0b111) + 1
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('<HBBBB', width, height, packed, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    if global_color_table_flag:
        colors = [rng.read(3) for _ in range(global_color_table_size)]
        color_table = b''.join(colors)
    else:
        color_table = None

    # 4. Graphic Control Extension
    gce = b''
    if rng.read(1)[0]:
        delay = rng.read(2)[0]
        transparent_color_index = rng.read(1)[0]
        disposal_method = rng.read(1)[0] & 0b111
        gce = struct.pack('<BBB', 0xF9, 4, delay) \
              + bytes([transparent_color_index]) \
              + bytes([disposal_method << 2])

    # 5. Image Descriptor
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = rng.read(2)[0]
    height = rng.read(2)[0]
    local_color_table_flag = bool(rng.read(1)[0] & 0b10000000)
    interlace_flag = bool(rng.read(1)[0] & 0b10000000)
    sort_flag = bool(rng.read(1)[0] & 0b10000000)
    reserved = rng.read(1)[0] & 0b11100000
    local_color_table_size = (rng.read(1)[0] & 0b111) + 1
    image_descriptor = struct.pack('<HHBB', left, top, width, height) \
                       + bytes([local_color_table_flag | reserved]) \
                       + bytes([interlace_flag | sort_flag])
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        if len(blocks) >= total_blocks_size:
            break

        block_size = rng.read(1)[0]
        if not block_size:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)