#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Random background color index
    background_color_index = rng.read(1)[0]

    # Random pixel aspect ratio (we don't care about this)
    pixel_aspect_ratio = rng.read(1)[0]

    # Global Color Table Flag
    global_color_table_flag = rng.read(1)[0] == 1

    # Color Resolution
    color_resolution = rng.read(1)[0] & 0b111

    # Sort Flag
    sort_flag = rng.read(1)[0] == 1

    # Size of Global Color Table
    size_of_global_color_table = 2 ** ((global_color_table_flag << 3) | color_resolution)

    # Background Color Index
    background_color_index &= (1 << size_of_global_color_table) - 1

    # Pixel Aspect Ratio
    pixel_aspect_ratio &= 15

    # Pack all fields together
    lsd = struct.pack(
        '>HBBB',
        width,
        height,
        (background_color_index << 4) | (pixel_aspect_ratio << 2) | (sort_flag << 1) | global_color_table_flag,
        size_of_global_color_table
    )

    # 3. Global Color Table
    color_table = rng.read(3 * 2**((lsd[3] + 1) << 3))

    # 4. Graphics Control Extension
    gce = None
    if rng.read(1)[0]:  # If we have one
        # Block size
        gce = bytes([0x21, 0xf9, 0x04])

        # Disposal method
        disposal_method = rng.read(1)[0] & 0b111

        # User input flag
        user_input_flag = rng.read(1)[0] == 1

        # Transparent color flag
        transparent_color_flag = rng.read(1)[0] == 1

        # Delay time
        delay_time = struct.unpack('>H', rng.read(2))[0]

        # Transparent color index
        transparent_color_index = rng.read(1)[0]

        # Pack all fields together
        gce += struct.pack(
            'BBBHB',
            0b11100000 | disposal_method,
            (user_input_flag << 1) | transparent_color_flag,
            delay_time,
            transparent_color_index,
            0
        )

    # 5. Image Descriptor
    image_descriptor = b'\x2c\x00\x00\x00\x00' + lsd[:2]

    # 6. Image Data
    # LZW Minimum Code Size
    lzw_minimum_code_size = rng.read(1)[0]

    # Image Data Blocks
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)