#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a' + (b'\x00' * 7)
    out.write(header)

    # 2. Logical Screen Descriptor
    lsd = bytearray([
        0xFF, 0xFF,  # width
        0xFF, 0xFF,  # height
        0xF4,        # global color table flag
        0x00,        # background color index
        0x00,        # pixel aspect ratio
    ])
    out.write(lsd)

    # 3. Global Color Table
    for i in range(0, 256):
        r = int.from_bytes(rng.read(1), 'big') % 256
        g = int.from_bytes(rng.read(1), 'big') % 256
        b = int.from_bytes(rng.read(1), 'big') % 256
        out.write(struct.pack('BBB', r, g, b))

    # 4. Application Extension Block
    appext = bytearray([
        0x21, 0xff, 0x0b,  # extension block size
        0x4e, 0x45, 0x54,  # "NETS"
        0x53, 0x43, 0x4f,  # "CSOE"
        0x00,              # application block size
    ])
    out.write(appext)

    # 5. Image Descriptor
    idesc = bytearray([
        0x2c,          # image separator
        0x00, 0x00,    # left position
        0x00, 0x00,    # top position
        0xFF, 0xFF,    # width
        0xFF, 0xFF,    # height
        0x00,          # local color table flag
        0x00,          # interlace flag
        0x00,          # sort flag
        0x00,          # reserved
    ])
    out.write(idesc)

    # 6. Local Color Table
    for i in range(0, 256):
        r = int.from_bytes(rng.read(1), 'big') % 256
        g = int.from_bytes(rng.read(1), 'big') % 256
        b = int.from_bytes(rng.read(1), 'big') % 256
        out.write(struct.pack('BBB', r, g, b))

    # 7. Image Data
    while True:
        code = int.from_bytes(rng.read(1), 'big') % 256
        if code == 0x21:
            # Control extension
            extcode = int.from_bytes(rng.read(1), 'big') % 256
            if extcode == 0xf9:
                # Graphic control extension
                extsize = int.from_bytes(rng.read(1), 'big') % 256
                assert extsize == 0x04
                flags = int.from_bytes(rng.read(1), 'big') % 256
                delay = int.from_bytes(rng.read(2), 'big') % 65536
                transp = int.from_bytes(rng.read(1), 'big') % 256
                assert rng.read(1)[0] == 0x00
                break
            else:
                # Skip this extension
                extsize = int.from_bytes(rng.read(1), 'big') % 256
                rng.read(extsize)
        elif code == 0x2c:
            # Image descriptor
            break
        else:
            raise Exception("Invalid GIF data")

    # 8. Image Data
    while True:
        code = int.from_bytes(rng.read(1), 'big') % 256
        if code >= 0x80 and code <= 0xff:
            # LZW-compressed data
            length = int.from_bytes(rng.read(1), 'big') % 256
            out.write(rng.read(length))
        elif code == 0x00:
            # End of image
            break
        else:
            raise Exception("Invalid GIF data")

    # 9. Trailer
    trailer = b'\x3b'
    out.write(trailer)
