#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big') % 65536
    height = int.from_bytes(rng.read(2), 'big') % 65536
    color_table_size = 7 + (int.from_bytes(rng.read(1), 'big') & 0b111)
    background_color_index = int.from_bytes(rng.read(1), 'big')
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big')
    out.write(struct.pack('<HHBB', width, height, color_table_size, background_color_index))
    out.write(pixel_aspect_ratio.to_bytes(1, 'big'))

    # 3. Global Color Table
    for _ in range(2 ** color_table_size):
        red = int.from_bytes(rng.read(1), 'big')
        green = int.from_bytes(rng.read(1), 'big')
        blue = int.from_bytes(rng.read(1), 'big')
        out.write(struct.pack('<BBB', red, green, blue))

    # 4. Image Data
    while True:
        # 4.1 Graphic Control Extension
        block_type = int.from_bytes(rng.read(1), 'big')
        if block_type == 0xF9:
            extension_length = int.from_bytes(rng.read(1), 'big')
            assert extension_length == 4
            disposal_method = int.from_bytes(rng.read(1), 'big')
            user_input_flag = bool(int.from_bytes(rng.read(1), 'big'))
            transparent_color_flag = bool(int.from_bytes(rng.read(1), 'big'))
            delay_time = int.from_bytes(rng.read(2), 'big')
            transparent_color_index = int.from_bytes(rng.read(1), 'big')
            out.write(struct.pack('<BBBBHB', block_type, extension_length, disposal_method, user_input_flag, transparent_color_flag, delay_time, transparent_color_index))

        # 4.2 Image Descriptor
        image_left_position = int.from_bytes(rng.read(2), 'big')
        image_top_position = int.from_bytes(rng.read(2), 'big')
        image_width = int.from_bytes(rng.read(2), 'big')
        image_height = int.from_bytes(rng.read(2), 'big')
        local_color_table_flag = bool(int.from_bytes(rng.read(1), 'big'))
        interlace_flag = bool(int.from_bytes(rng.read(1), 'big'))
        sort_flag = bool(int.from_bytes(rng.read(1), 'big'))
        size_of_local_color_table = 7 + (int.from_bytes(rng.read(1), 'big') & 0b111)
        out.write(struct.pack('<HHHHBBB', image_left_position, image_top_position, image_width, image_height, local_color_table_flag, interlace_flag, sort_flag, size_of_local_color_table))

        # 4.3 Local Color Table
        for _ in range(2 ** size_of_local_color_table):
            red = int.from_bytes(rng.read(1), 'big')
            green = int.from_bytes(rng.read(1), 'big')
            blue = int.from_bytes(rng.read(1), 'big')
            out.write(struct.pack('<BBB', red, green, blue))

        # 4.4 LZW-compressed Image Data
        lzw_minimum_code_size = int.from_bytes(rng.read(1), 'big')
        out.write(lzw_minimum_code_size.to_bytes(1, 'big'))
        while True:
            code_size = int.from_bytes(rng.read(1), 'big')
            if code_size == 0:
                break
            else:
                data = rng.read(code_size)
                out.write(data)

        # 4.5 Trailer
        trailer = b'\x00'
        out.write(trailer)

