#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):
    # Header
    header = b'\x47\x49\x46\x38\x39\x61'  # "GIF89a"

    # Logical screen descriptor (LSD)
    width = int.from_bytes(rng.read(2), byteorder='little') % 256
    height = int.from_bytes(rng.read(2), byteorder='little') % 256
    packed = rng.read(1)[0]
    global_color_table = bool(packed & 0b10000000)
    color_resolution = (packed >> 4) & 0b111
    sort_flag = bool(packed & 0b100)
    background_color_index = packed & 0b1111
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('<HHBBB', width, height, global_color_table, color_resolution, sort_flag, background_color_index) \
          + bytes([pixel_aspect_ratio])

    # Global color table
    if global_color_table:
        color_count = 2 ** (color_resolution + 1)
        color_table = rng.read(3 * color_count)
    else:
        color_table = b''

    # Graphic control extension (GCE)
    block_size = 4
    packed = rng.read(1)[0]
    disposal_method = (packed >> 2) & 0b111
    user_input_flag = bool(packed & 0b10)
    transparent_color_flag = bool(packed & 0b1)
    delay_time = int.from_bytes(rng.read(2), byteorder='little')
    transparent_color_index = rng.read(1)[0]
    block_terminator = b'\x00'
    gce = struct.pack('<BBHBB', block_size, 0b00100000 | disposal_method << 2 | user_input_flag << 1 | transparent_color_flag, delay_time, transparent_color_index, ord(block_terminator))

    # Image descriptor
    left = int.from_bytes(rng.read(2), byteorder='little') % 256
    top = int.from_bytes(rng.read(2), byteorder='little') % 256
    width = int.from_bytes(rng.read(2), byteorder='little') % 256
    height = int.from_bytes(rng.read(2), byteorder='little') % 256
    local_color_table = bool(rng.read(1)[0] & 0b10000000)
    interlace_flag = bool(rng.read(1)[0] & 0b1000000)
    sorted_flag = bool(rng.read(1)[0] & 0b100000)
    reserved = bool(rng.read(1)[0] & 0b10000)
    lct_size = (rng.read(1)[0] & 0b111) + 1
    image_descriptor = struct.pack('<HHHHBB', left, top, width, height, local_color_table, interlace_flag, sorted_flag)

    # Local color table
    if local_color_table:
        color_count = 2 ** (lct_size + 1)
        color_table += rng.read(3 * color_count)

    # Image data
    min_code_size = rng.read(1)[0]
    image_data = [min_code_size]
    while True:
        code = rng.read(1)[0]
        if code == 0:
            break
        image_data.append(code)

    # Trailer
    trailer = b';'
    # Write everything to the output stream
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(image_data) + trailer)


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('./test.gif', 'wb') as f:
            generate_random_gif(rng, f)