#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed_fields = rng.read(1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)

    lsd = width + height + packed_fields + background_color_index + pixel_aspect_ratio

    # 3. Global Color Table
    global_color_table_flag = (packed_fields[0] & 0b10000000) >> 7
    color_resolution = (packed_fields[0] & 0b01110000) >> 4
    sort_flag = (packed_fields[0] & 0b00001000) >> 3
    global_color_table_size = 2 ** ((packed_fields[0] & 0b00000111) + 1)

    if global_color_table_flag == 1:
        global_color_table = rng.read(global_color_table_size * 3)
    else:
        global_color_table = b''

    color_table = lsd + global_color_table

    # 4. Graphic Control Extension
    gce = b'\x21\xF9\x04\x05\x00\xFF\xFF\xFF'

    # 5. Image Descriptor
    image_descriptor = b'\x2C\x00\x00\x00\x00\x00'

    # 6. Image Data
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)