#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    header = b'GIF89a'
    screen_width = rng.read(2)
    screen_height = rng.read(2)
    global_color_table_flag = bool(rng.read(1)[0])
    color_resolution = rng.read(1)[0]
    sort_flag = bool(rng.read(1)[0])
    global_color_table_size = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = header + screen_width + screen_height + \
        bytes([global_color_table_flag << 7 |
               color_resolution << 4 |
               sort_flag << 3 |
               global_color_table_size]) + \
        bytes([background_color_index]) + \
        bytes([pixel_aspect_ratio])

    # 2. Global Color Table
    color_table = rng.read(global_color_table_size * 3)

    # 3. Graphics Control Extension
    gce = None
    if bool(rng.read(1)[0]):
        gce = struct.pack('>BBBB',
                          0xF9,
                          rng.read(1)[0],
                          rng.read(1)[0],
                          rng.read(1)[0]
                         )

    # 4. Application Extension
    app_extension = struct.pack('>BBHBB',
                                0xFF,
                                0xFE,
                                len('Python'),
                                ord('P'),
                                ord('y'),
                                ord('t'),
                                ord('h'),
                                ord('o'),
                                ord('n'),
                                ord('.'),
                                ord('c'),
                                ord('o'),
                                ord('m'),
                                ord('/'),
                                ord('u'),
                                ord('r'),
                                ord('l'),
                                ord('i'),
                                ord('n'),
                                ord('k'),
                                ord('e'),
                                ord('r'),
                                ord('s'),
                                ord(' '),
                                ord('v'),
                                ord('e'),
                                ord('r'),
                                ord('s'),
                                ord('i'),
                                ord('o'),
                                ord('n'),
                                ord(':'),
                                ord('3'),
                                ord('.'),
                                ord('6'),
                                ord('.'),
                                ord('0'),
                                ord('\x00')
                               )

    # 5. Image Descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved_flags = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = rng.read(local_color_table_size * 3)
    image_descriptor = struct.pack('>BHHHHBBBBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  local_color_table_size
                                 )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')