#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# algorithm described in https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor (LSD)
    width = int.from_bytes(rng.read(2), byteorder='big')
    height = int.from_bytes(rng.read(2), byteorder='big')
    packed_fields = rng.read(1)[0]
    global_color_table_flag = bool((packed_fields & 0b10000000) >> 7)
    color_resolution = ((packed_fields & 0b01110000) >> 4) + 1
    sort_flag = bool((packed_fields & 0b00001000) >> 3)
    global_color_table_size = 2 ** ((packed_fields & 0b00000111) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('<HHBBBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    color_table = rng.read(global_color_table_size * 3)
    
    # 4. Graphic Control Extension (optional)
    gce = None
    if global_color_table_flag:
        gce = b'\x21\xf9\x04\x05' + rng.read(6)
    
    # 5. Image Descriptor
    image_descriptor = b'\x2c' + struct.pack('<HHBB', 0, 0, width, height)
    
    # 6. Image Data
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if block_size == 0:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom

    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)