#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = b''
    if rng.read(1)[0] & 1:
        # We want to write a GCE
        
        # Delay time (two bytes)
        delay = struct.pack('>H', rng.read(2)[0])
        
        # Transparent color index (one byte)
        transparent_color_index = rng.read(1)
        
        # Block terminator (one byte)
        block_terminator = rng.read(1)
        
        gce = delay + transparent_color_index + block_terminator
    
    # 5. Image Descriptor
    image_descriptor = b'\x2C\x00\x00\x00\x00\x00\xFF\xFF\x00'
    
    # 6. Image Data
    # Let's use a random number of blocks
    num_blocks = rng.read(1)[0] % 256
    blocks = [rng.read(255) for _ in range(num_blocks)]
    
    # 7. Trailer
    trailer = b';'
    
    out.write(header)
    out.write(lsd)
    out.write(struct.pack('BB', 0xff, num_colors - 1))
    out.write(color_table)
    if len(gce) == 6:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)