#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Optional Graphics Control Extension (GCE)
    # We will use this to set the delay time between frames
    # If we don't do this, the default delay time is 100 ms
    # This is too slow for our purposes
    # The GCE has a fixed length of 4 bytes
    # We will use the first byte to indicate whether we want to write the GCE or not
    # The other three bytes are the actual GCE data
    gce = None
    if rng.read(1)[0]:
        gce = struct.pack('>BBB',
                          rng.read(1)[0],  # Block Size
                          0xF9,  # Graphic Control Label
                          rng.read(1)[0]  # Delay Time
                         )

    # 5. Image Descriptor
    left_position, top_position = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))

    # Packed Fields
    # We will use these to specify that there is no local color table,
    # that the image is not interlaced, and that it uses the global color table
    local_color_table_flag = 0
    interlace_flag = 0
    sort_flag = 0
    reserved_flags = 0
    local_color_table_size = 0
    packed_fields = struct.pack('BB',
                                local_color_table_flag << 7 |
                                interlace_flag << 6 |
                                sort_flag << 5 |
                                reserved_flags << 3 |
                                local_color_table_size,
                                rng.read(1)[0]
                               )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(struct.pack('>BBHHHBB',
                          0x2C,  # Image Separator
                          left_position,
                          top_position,
                          width,
                          height,
                          packed_fields[0],
                          packed_fields[1]
                         ) + image_data + b'\x3B')  # Trailing semicolon

    return len(lsd) + len(color_table) + (len(gce) if gce else 0) + \
        len(struct.pack('>BBHHHBB',
                        0x2C,  # Image Separator
                        left_position,
                        top_position,
                        width,
                        height,
                        packed_fields[0],
                        packed_fields[1]
                       ) + image_data + b'\x3B')  # Trailing semicolon


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Usage: {} <output_file>".format(os.path.basename(__file__)))
        exit(-1)

    output_filename = sys.argv[1]

    with open('/dev/urandom', 'rb') as f:
        with open(output_filename, 'wb') as out:
            generate_random_gif(f, out)


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif_old(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_rat                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  local_color_table_size
                                 )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')