#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big') % 640 + 1
    height = int.from_bytes(rng.read(2), 'big') % 480 + 1
    flags = rng.read(1)[0] & 7
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('<HBBB', width, height, flags, background))
    out.write(aspect_ratio)

    # 3. Global Color Table
    if flags & 1:
        colors = 2 ** ((flags >> 4) + 1)
        for _ in range(colors):
            out.write(rng.read(3))

    # 4. Image Data
    while True:
        # 4.1 Graphic Control Extension
        if rng.read(1)[0]:
            block_size = rng.read(1)[0]
            assert block_size == 4
            delay = rng.read(2)
            transparent = rng.read(1)[0]
            terminator = rng.read(1)[0]
            assert terminator == 0
            out.write(b'\x21\xf9\x04' + delay + bytes([transparent]) + b'\x00')

        # 4.2 Image Descriptor
        out.write(b'\x2c')
        left = rng.read(2)
        top = rng.read(2)
        width = rng.read(2)
        height = rng.read(2)
        flags = rng.read(1)[0]
        out.write(left + top + width + height + bytes([flags]))

        # 4.3 Local Color Table
        if flags & 1:
            colors = 2 ** ((flags >> 4) + 1)
            for _ in range(colors):
                out.write(rng.read(3))

        # 4.4 LZW-compressed image data
        lzw_minimum_code_size = rng.read(1)[0]
        out.write(lzw_minimum_code_size.to_bytes(1, 'big'))
        while True:
            code = rng.read(1)[0]
            if code == 0:
                break
            out.write(code.to_bytes(1, 'big'))

        # 4.5 Block Terminator
        out.write(b'\x00')

    # 5. Trailer
    out.write(b';')
