#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    packed_width_height = struct.pack('>HH', width, height)

    # Random background color index
    bg_color_index = rng.read(1)[0]

    # Aspect ratio is fixed at 1:1
    aspect_ratio = 0x00

    # Global Color Table Flag set to 1
    global_color_table_flag = 0x80

    # Color Resolution is fixed at 8 bits per pixel
    color_resolution = 0x07

    # Sort Flag is fixed at 0
    sort_flag = 0x00

    # Size of Global Color Table is fixed at 7 bits per pixel
    size_of_global_color_table = 0x07

    lsd = packed_width_height + \
          struct.pack('BB', bg_color_index, aspect_ratio | global_color_table_flag | color_resolution | sort_flag | size_of_global_color_table)

    # 3. Global Color Table
    color_table = rng.read(3 * 2 ** (size_of_global_color_table + 1))

    # 4. Graphic Control Extension
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    image_data = []
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break

        image_data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to the output stream
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(image_data) + trailer)


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('./test.gif', 'wb') as f:
            generate_random_gif(rng, f)