#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('<HH', rng.read(4))

    # Packed fields
    # Global color table flag (1 bit)
    # Color resolution (3 bits)
    # Sort flag (1 bit)
    # Size of global color table (3 bits)
    packed_fields = struct.pack('B', rng.read(1)[0])

    # Background color index (1 byte)
    background_color_index = rng.read(1)

    # Pixel aspect ratio (1 byte)
    pixel_aspect_ratio = rng.read(1)

    # 3. Global Color Table
    # We don't use it so we just skip it
    rng.seek(3 * (2 ** ((packed_fields[0] & 7) + 1)), 1)

    # 4. Image Data
    # We write the image data as a single frame
    # with no local color table

    # Graphic Control Extension
    # Block size (always 4)
    out.write(b'\x04')

    # Packed field
    # Disposal method (3 bits)
    # User input flag (1 bit)
    # Transparent color flag (1 bit)
    out.write(struct.pack('B', rng.read(1)[0]))

    # Delay time (2 bytes)
    out.write(rng.read(2))

    # Transparent color index (1 byte)
    out.write(rng.read(1))

    # Block terminator (1 byte)
    out.write(b'\x00')

    # Image Descriptor
    # Separator (1 byte)
    out.write(b'\x2c')

    # Left position (2 bytes)
    out.write(b'\x00\x00')

    # Top position (2 bytes)
    out.write(b'\x00\x00')

    # Width (2 bytes)
    out.write(struct.pack('<H', width))

    # Height (2 bytes)
    out.write(struct.pack('<H', height))

    # Packed field
    # Local color table flag (1 bit)
    # Interlace flag (1 bit)
    # Sort flag (1 bit)
    # Reserved (2 bits)
    # Size of local color table (3 bits)
    out.write(struct.pack('B', rng.read(1)[0]))

    # LZW Minimum Code Size (1 byte)
    out.write(rng.read(1))

    # Image Data
    # We just copy the data from /dev/urandom
    while True:
        chunk = rng.read(65536)
        if not chunk:
            break
        out.write(chunk)

    # Trailer
    out.write(b';')
