#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    # Write all of these to the output
    out.write(struct.pack('<HBBB', width, height, packed_fields, background_color_index))
    out.write(pixel_aspect_ratio)

    # 3. Global Color Table
    # We will use a global color table with 256 colors
    # Each color is represented by three bytes (RGB)
    num_colors = 2 ** ((packed_fields & 7) + 1)
    out.write(rng.read(num_colors * 3))

    # 4. Image Descriptor
    # The image descriptor contains information about the image itself
    # We will write a single image in this GIF
    # First we need to write the separator character
    out.write(b'\x2c')

    # Then we need to write the image left position and top position
    # These are both set to zero since we don't care where it appears on screen
    out.write(b'\x00\x00\x00\x00')

    # Next we need to write the image width and height
    # Since we want to make sure that the image is always visible,
    # we will cap the maximum size at the logical screen size
    image_width = min(width - 1, rng.read(1)[0])
    image_height = min(height - 1, rng.read(1)[0])
    out.write(struct.pack('<HH', image_width, image_height))

    # Finally, we need to write the packed fields
    # We only care about the local color table flag, so we will just set that bit
    packed_fields = b'\x80' if rng.read(1)[0] else b'\x00'
    out.write(packed_fields)

    # 5. Local Color Table
    # If we have a local color table, then we need to write it here
    if packed_fields[0] == 0x80:
        num_colors = 2 ** ((packed_fields[0] >> 4) + 1)
        out.write(rng.read(num_colors * 3))

    # 6. Image Data
    # Now we can start writing the actual image data
    # This is where things get complicated...
    # We will not be implementing any fancy compression or anything like that
    # Instead, we will just write every pixel individually
    # To do this, we first need to calculate how many pixels there are
    num_pixels = image_width * image_height

    # Then we need to write the minimum code size
    # We will use a minimum code size of 2 bits
    out.write(b'\x02')

    # Then we need to write the image data
    # We will write one byte at a time
    # We will also keep track of the previous byte written
    # This allows us to easily detect when we reach the end of the image
    prev_byte = None
    while num_pixels > 0:
        # Read a new byte from the RNG
        cur_byte = rng.read(1)[0]

        # Check if we reached the end of the image
        if prev_byte is None or cur_byte != prev_byte:
            # If we did, then we need to write the current byte as an uncompressed value
            # We will also decrement the number of pixels remaining
            out.write(cur_byte.to_bytes(1, 'little'))
            num_pixels -= 1

        # Otherwise, we need to write the difference between the current and previous byte
        # We will also decrement the number of pixels remaining
        else:
            out.write((cur_byte - prev_byte).to_bytes(1, 'little'))
            num_pixels -= 1

        # Update the previous byte
        prev_byte = cur_byte

    # 7. Trailer
    # Finally, we need to write the trailer
    # It consists of a single byte containing the ASCII code for ';'
    out.write(b';')

    return 0

