#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big', signed=False) + 1
    height = int.from_bytes(rng.read(2), 'big', signed=False) + 1
    color_table_size = 7 + (int.from_bytes(rng.read(1), 'big', signed=False) & 0x7)
    background_color_index = int.from_bytes(rng.read(1), 'big', signed=False)
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big', signed=False)

    out.write(struct.pack('<HHBBB', width, height, color_table_size, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table
    for _ in range(2 ** color_table_size):
        r = int.from_bytes(rng.read(1), 'big', signed=False)
        g = int.from_bytes(rng.read(1), 'big', signed=False)
        b = int.from_bytes(rng.read(1), 'big', signed=False)
        out.write(struct.pack('BBB', r, g, b))

    # 4. Image Descriptors
    while True:
        # 4.1 Image Separator
        out.write(b'\x2c')

        # 4.2 Image Descriptor
        left = int.from_bytes(rng.read(2), 'big', signed=False)
        top = int.from_bytes(rng.read(2), 'big', signed=False)
        width = int.from_bytes(rng.read(2), 'big', signed=False) + 1
        height = int.from_bytes(rng.read(2), 'big', signed=False) + 1
        local_color_table_flag = bool(int.from_bytes(rng.read(1), 'big', signed=False) >> 7)
        interlaced = bool(int.from_bytes(rng.read(1), 'big', signed=False) >> 6)
        sorted = bool(int.from_bytes(rng.read(1), 'big', signed=False) >> 5)
        reserved = bool(int.from_bytes(rng.read(1), 'big', signed=False) >> 4)
        color_table_size = 7 + (int.from_bytes(rng.read(1), 'big', signed=False) & 0x7)

        out.write(struct.pack('<HHBBBBB', left, top, width, height, local_color_table_flag, interlaced, sorted, reserved, color_table_size))

        if not local_color_table_flag:
            break

        # 4.3 Local Color Table
        for _ in range(2 ** color_table_size):
            r = int.from_bytes(rng.read(1), 'big', signed=False)
            g = int.from_bytes(rng.read(1), 'big', signed=False)
            b = int.from_bytes(rng.read(1), 'big', signed=False)
            out.write(struct.pack('BBB', r, g, b))

        # 4.4 LZW Minimum Code Size
        lzw_minimum_code_size = int.from_bytes(rng.read(1), 'big', signed=False)
        out.write(lzw_minimum_code_size.to_bytes(1, 'big'))

        # 4.5 Image Data
        image_data = []
        while True:
            code_size = int.from_bytes(rng.read(1), 'big', signed=False)
            if code_size == 0:
                break
            else:
                image_data.append(rng.read(code_size))

        out.write(b''.join(image_data))

    # 5. Trailer
    out.write(b';')
