#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    idd = struct.pack('>BBHHH', 0x2C, 0, 0, width, height)

    # 6. Local Color Table (optional)
    has_lct = bool(rng.read(1)[0])
    if has_lct:
        num_colors = (rng.read(1)[0] % 256) or 1
        color_table = rng.read(3 * num_colors)  # Random colors
    else:
        color_table = None

    # 7. Image Data
    image_data = rng.read(width * height)  # Random data

    # Write everything to the output stream
    out.write(lsd + color_table + (gce or b'') + idd + (color_table or b'') + image_data)

    return len(image_data)
