#!/usr/bin/env python3

import struct
from typing import BinaryIO


def generate(out: BinaryIO, rng):
    # 1. Header
    header = b'\x47\x49\x46\x38\x39\x61'  # GIF89a

    # 2. Logical Screen Descriptor
    width = rng.read(2)[0]
    height = rng.read(2)[0]
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HBBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    has_global_color_table = bool(packed_fields & 0x80)
    global_color_table_size = ((packed_fields >> 4) & 0x7) + 1
    if has_global_color_table:
        color_table = rng.read(3 * (2 ** (global_color_table_size + 1)))
    else:
        color_table = b''

    # 4. Graphic Control Extension
    gce = None
    if rng.read(1)[0]:
        delay = rng.read(2)[0]
        transparent_color_index = rng.read(1)[0]
        terminator = rng.read(1)[0]
        gce = struct.pack('>BBBB', 0xF9, 4, delay, transparent_color_index)

    # 5. Image Descriptor
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = rng.read(2)[0]
    height = rng.read(2)[0]
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)