#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBB', width, height, packed_fields, background_color_index)

    # 3. Global Color Table
    global_color_table_flag = bool(rng.read(1)[0] & 0x80)
    color_resolution = ((rng.read(1)[0] >> 4) & 0x07) + 1
    sort_flag = bool(rng.read(1)[0] & 0x08)
    global_color_table_size = 2 ** (((rng.read(1)[0] & 0x07) + 1) * 3)
    global_color_table = rng.read(global_color_table_size * 3)
    color_table = struct.pack('>BBBB',
                              global_color_table_flag << 7 | color_resolution << 4 | sort_flag << 3 | global_color_table_size - 1,
                              0,
                              0,
                              0
                             ) + global_color_table

    # 4. Graphic Control Extension
    gce = None
    if rng.read(1)[0]:
        disposal_method = rng.read(1)[0] & 0x07
        user_input_flag = bool(rng.read(1)[0] & 0x02)
        transparent_color_flag = bool(rng.read(1)[0] & 0x01)
        delay_time = rng.read(2)
        transparent_color_index = rng.read(1)[0]
        block_terminator = rng.read(1)[0]
        gce = struct.pack('>BBBBB',
                          0xF9,
                          disposal_method << 2 | user_input_flag << 1 | transparent_color_flag,
                          delay_time[0],
                          delay_time[1],
                          transparent_color_index
                         )

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0] & 0x80)
    interlace_flag = bool(rng.read(1)[0] & 0x40)
    sort_flag = bool(rng.read(1)[0] & 0x08)
    reserved_flags = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = rng.read(local_color_table_size * 3)
    image_descriptor = struct.pack('>BHHHHBBBBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  local_color_table_size
                                 )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')