#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Random background color index
    bg_color_index = rng.read(1)[0] % 256

    # Random pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)[0] & 0b01111111

    # Packed fields
    packed_fields = 0b00000000 | \
                    0b00000000 | \
                    ((bg_color_index & 0b00001111) << 4) | \
                    (pixel_aspect_ratio & 0b00001111)

    # Unpacked fields
    unpacked_fields = 0b00000000 | \
                      0b00000000 | \
                      ((bg_color_index & 0b00110000) >> 4) | \
                      (pixel_aspect_ratio & 0b11000000)

    # LSD
    lsd = struct.pack('>HBBBB', width, height, packed_fields, unpacked_fields, 0)

    # 3. Global Color Table
    # We don't use one so we write an empty block
    color_table = b'\x00\x00'

    # 4. Graphic Control Extension
    # We don't use one so we write an empty block
    gce = b'\x00\x00'

    # 5. Image Descriptor
    # Random left position
    left = rng.read(1)[0] % 256

    # Random top position
    top = rng.read(1)[0] % 256

    # Width and Height are taken from the LSD
    image_descriptor = struct.pack('>BBB', left, top, 0)

    # 6. Image Data
    # We don't use any data so we write an empty block
    image_data = b'\x00\x00'

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)