#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(num_colors)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBHB', 0, 0, 0b10000000 | rng.read(1)[0], 0, 0)
    
    # 5. Local Color Table (same as above)
    num_colors = (rng.read(1)[0] % 256) or 1
    local_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(num_colors)]
    
    # 6. Graphics Control Extension (same as above)
    graphics_control_extension = struct.pack('>BBHBBBBB', 0x21, 0xF9, 4, 0b10000000 | rng.read(1)[0], 0, 0, 0)
    
    # 7. Application Extension (same as above)
    application_extension = struct.pack('>BBBBB', 0x21, 0xFF, 11, ord('A'), ord('N') << 8 | ord('I')) + \
                            struct.pack('>BBBBBBB', ord('S'), ord('C'), ord('R'), ord('E'), ord('E'), ord('N'), ord('T')) + \
                            struct.pack('>BBB', 0, 0, 0)
    
    # 8. Comment Extension (same as above)
    comment_extension = struct.pack('>BBBBB', 0x21, 0xFE, 11, ord('A'), ord('N') << 8 | ord('I')) + \
                        struct.pack('>BBBBBBB', ord('S'), ord('C'), ord('R'), ord('E'), ord('E'), ord('N'), ord('T')) + \
                        struct.pack('>BBB', 0, 0, 0)
    
    # 9. Plain Text Extension (same as above)
    plain_text_extension = struct.pack('>BBBBB', 0x21, 0x01, 11, ord('A'), ord('N') << 8 | ord('I')) + \
                           struct.pack('>BBBBBBB', ord('S'), ord('C'), ord('R'), ord('E'), ord('E'), ord('N'), ord('T')) + \
                           struct.pack('>BBB', 0, 0, 0)
    
    # 10. Trailer
    trailer = b';'
    
    # Write everything to output
    out.write(header + lsd + b''.join(global_color_table) + image_descriptor + b''.join(local_color_table) +
              graphics_control_extension + application_extension + comment_extension + plain_text_extension + trailer)

