#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break

        block = rng.read(block_size)
        blocks.append((lzw_minimum_code_size, block))

    # Now let's build our data!
    data = [header, lsd, color_table] + ([gce] if gce is not None else []) + [image_descriptor] + list(map(lambda x: struct.pack('>BB%ds' % len(x[1]), 0x00 | x[0], len(x[1]), x[1]), blocks))

    # Write everything to the output stream
    out.write(b''.join(data))