#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    lsd = struct.pack('>BBHHH', *(width & 0xFF, (height >> 8), width >> 8, height & 0xFF, 0b11100000 | ((rng.read(1)[0] & 0b111) << 2)))

    # 3. Global Color Table
    global_color_table_size = min((rng.read(1)[0] & 0b1111) + 1, 2**int(((rng.read(1)[0] & 0b11) << 2) + 1))
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(global_color_table_size)]
    color_table = b''.join(global_color_table)

    # 4. Graphics Control Extension
    delay_time = rng.read(2)
    transparent_index = rng.read(1)[0]
    gce = struct.pack('>BBBB', 0x21, 0xF9, 4, 0b10000000 | (delay_time[0] & 0b11111111) | (((delay_time[1] & 0b1111) << 8) & 0b1111111100000000))

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)