#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = b'\x21\xF9\x04' + rng.read(3) + b'\x00'

        # 5. Application Extension (optional)
        has_ae = bool(rng.read(1)[0])
        if has_ae:
            ae = b'\xFF\x0BNETSCAPE2.0' + rng.read(18)

            # 6. Comment Extension (optional)
            has_ce = bool(rng.read(1)[0])
            if has_ce:
                ce = b'\x21\xFE' + rng.read((rng.read(1)[0] % 256) or 1)

                # 7. Image Data
                image_data = rng.read((width * height // 8) or 1)

                # Write everything to output
                out.write(lsd)
                out.write(color_table)
                out.write(gce)
                out.write(ae)
                out.write(ce)
                out.write(image_data)

                return len(lsd) + len(color_table) + len(gce) + len(ae) + len(ce) + len(image_data)

    else:
        # 5. Image Data
        image_data = rng.read((width * height // 8) or 1)

        # Write everything to output
        out.write(lsd)
        out.write(color_table)
        out.write(image_data)

        return len(lsd) + len(color_table) + len(image_data)

