#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table
    global_color_table_flag = rng.read(1)[0] & 0b1
    local_color_table_flag = rng.read(1)[0] & 0b1
    interlace_flag = rng.read(1)[0] & 0b1
    sort_flag = rng.read(1)[0] & 0b1
    reserved = rng.read(3)

    # Number of colors in palette - either 2^(n+1) or 2^n
    nbits = rng.read(1)[0] & 0b111
    palette_size = int((nbits + 1) * 2 ** nbits)

    # Actual color table
    color_table = []
    for i in range(palette_size):
        r, g, b = [int(c) for c in '{:02X}{:02X}{:02X}'.format(*rng.read(3))]
        color_table.extend([r, g, b])

    # 4. Graphic Control Extension
    gce = b''

    # If there is a GCE, it will be at most this long
    max_gce_length = 11

    # Generate a random GCE if one should exist
    if rng.read(1)[0]:
        gce = b'\x21\xF9\x04' + rng.read(max_gce_length-3)

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    packed_fields = rng.read(1)[0]

    # 6. Image Data
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)