#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 16

    lsd = struct.pack('>HHBBBH', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    global_color_table_flag = bool(packed_fields & 0b10000000)
    global_color_table_size = ((packed_fields >> 7) & 0b111) + 1
    global_color_table = []
    for i in range(global_color_table_size):
        r, g, b = [rng.read(1)[0] % 256 for _ in range(3)]
        global_color_table.append((r, g, b))
    color_table = b''
    if global_color_table_flag:
        color_table += bytes([(len(global_color_table) - 1)])
        for rgb in global_color_table:
            color_table += bytes(rgb)

    # 4. Graphics Control Extension
    gce_exists = bool(packed_fields & 0b00000100)
    gce = None
    if gce_exists:
        gce = b'\x21\xF9\x04'  # Application extension introducer
        gce += struct.pack('>BBHB', 0xFF, 0x0B, 4, 0)  # Block size, Flags, Delay time, Transparent color index
        gce += b'\x00' * 3  # No transparancy

    # 5. Image Descriptor
    image_descriptor = b'\x2C'  # Image separator
    image_descriptor += struct.pack('>HHBBB', 0, 0, width, height, 0)  # Left offset, Top offset, Width, Height, Packed fields

    # 6. Image Data
    lzw_minimum_code_size = rng.read(1)[0] % 256
    blocks = []
    while True:
        block_size = rng.read(1)[0] % 256
        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')