#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 14 09:25:17 2021

@author: <NAME>

This script contains functions to perform the analysis of the data.
"""


import numpy as np
from scipy import stats, signal
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA



def plot_spectrogram(data, fs=None):
    """Plot a spectrogram for the given data."""
    
    # Get the sampling frequency if not provided
    if fs is None:
        fs = get_fs(data)

    # Compute the spectrogram
    f, t, Sxx = signal.spectrogram(data, fs)
    
    # Plot the spectrogram
    fig, ax = plt.subplots()
    im = ax.pcolormesh(t, f, Sxx, shading='gouraud')
    ax.set_ylabel('Frequency [Hz]')
    ax.set_xlabel('Time [sec]')
    fig.colorbar(im, ax=ax)
    
    
    
def get_fs(data):
    """Get the sampling frequency from the data."""
    
    return len(data)/np.max(data)

    
    
def compute_pca(data, n_components=None):
    """Compute the principal components and their corresponding eigenvalues 
    using Principal Component Analysis (PCA).
    
    Parameters
    ----------
    data : array
        The input data.
        
    n_components : int or float, optional
        Number of components to keep. If ``n_components`` is not set all 
        components are kept::
        
            n_components == min(n_samples, n_features)
            
        If ``n_components`` is a number between 0 and 1 (inclusive), it is 
        treated as a percentage and ``ceil(n_components * n_features)`` features 
        are kept.
        
        If ``n_components`` is ``'mle'``, Minka's MLE is used to guess the 
        dimension.
        
        
    Returns
    -------
    pca : object
        A scikit-learn PCA object.
        
    eigvals : array
        Array containing the eigenvalues in decreasing order.
    """
    
    # Create an instance of PCA
    pca = PCA(n_components=n_components)
    
    # Fit the model with the data
    pca.fit(data)
    
    # Extract the eigenvalues
    eigvals = pca.explained_variance_
    
    return pca, eigvals
    
    
    
def compute_lda(X, y):
    """Compute the linear discriminants and their corresponding coefficients 
    using Linear Discriminant Analysis (LDA).
    
    Parameters
    ----------
    X : array
        The input data.
        
    y : array
        The target vector.
    
    
    Returns
    -------
    lda : object
        A scikit-learn LDA object.
        
    coeffs : array
        Array containing the coefficients.
    """
    
    # Create an instance of LDA
    lda = LDA()
    
    # Fit the model with the data
    lda.fit(X, y)
    
    # Extract the coefficients
    coeffs = lda.coef_.T
    
    return lda, coeffs