#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor (LSD)
    width = rng.read(1)[0] % 256  # Width must be between 0 and 255
    height = rng.read(1)[0] % 256  # Height must be between 0 and 255
    packed = struct.pack('>HBBBBB', width, height, 0xF7, 0, 0, 0)  # Global Color Table Flag set, no sort flag, color resolution 7 bits per pixel, global table size 256 colors
    lsd = packed[0:3] + bytes([packed[3] | 0b10000000]) + packed[4:]  # Set local color table flag in LSD
    
    # 3. Global Color Table (GCT)
    color_table = bytearray()
    for i in range(256):
        r = rng.read(1)[0] & 0b11111100
        g = rng.read(1)[0] & 0b11111110
        b = rng.read(1)[0] & 0b11111100
        a = 0xFF

        color_table += struct.pack('<BBBB', r, g, b, a)
    
    # 4. Graphics Control Extension (GCE)
    gce = None
    
    # 5. Image Descriptor (ID)
    left = rng.read(1)[0] % width
    top = rng.read(1)[0] % height
    width &= 0xFF
    height &= 0xFF
    packed = struct.pack('>HHHB', left, top, width, height)
    packed = packed[:1] + bytes([packed[1] | 0b10000000]) + packed[2:]  # Local Color Table Flag set
    image_descriptor = packed
    
    # 6. Image Data
    lzw_minimum_code_size = 8
    blocks = []
    
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        
        if block_size == 0:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom

    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)