#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    # Pack it up in little-endian format
    lsd = struct.pack('<HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    global_color_table_flag = bool(packed_fields & 0b10000000)
    gct_size = ((packed_fields >> 7) & 0b111) + 1

    if global_color_table_flag:
        global_color_table = [rng.read(3) for _ in range(gct_size)]

        # Pad to make sure there are at least 256 colors
        global_color_table += [(0, 0, 0) for _ in range((2 ** gct_size) - len(global_color_table))]

        # Convert to bytes
        global_color_table = b''.join(struct.pack('BBB', *rgb) for rgb in global_color_table)
    else:
        global_color_table = None

    # 4. Image Descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)

    local_color_table_flag = bool(packed_fields & 0b01000000)
    interlace_flag = bool(packed_fields & 0b00100000)
    sort_flag = bool(packed_fields & 0b00010000)
    reserved = rng.read(1)

    # Pack it up in little-endian format
    idd = struct.pack('<BBBBBB', left[0], left[1], top[0], top[1], width[0], width[1], height[0], height[1])

    # 5. Local Color Table
    if local_color_table_flag:
        local_color_table = [rng.read(3) for _ in range(2 ** ((packed_fields >> 4) & 0b1111))]

        # Pad to make sure there are at least 256 colors
        local_color_table += [(0, 0, 0) for _ in range((2 ** ((packed_fields >> 4) & 0b1111)) - len(local_color_table))]

        # Convert to bytes
        local_color_table = b''.join(struct.pack('BBB', *rgb) for rgb in local_color_table)
    else:
        local_color_table = None

    # 6. Image Data
    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0]

    # GCE
    gce = None
    if rng.read(1)[0]:
        # Block Size
        gce_block_size = rng.read(1)[0]

        # Flags
        gce_flags = rng.read(1)[0]

        # Delay Time
        delay_time = rng.read(2)

        # Transparent Color Index
        transparent_color_index = rng.read(1)[0]

        # Terminator
        terminator = rng.read(1)[0]

        assert gce_block_size == 10
        assert gce_flags == 0
        assert delay_time == 0
        assert transparent_color_index == 0
        assert terminator == 0

        gce = struct.pack('<BBBBHBB', gce_block_size, gce_flags, delay_time[0], delay_time[1], transparent_color_index, terminator)

    # Image data
    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)

