#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data
    data_size = int((width + 7) / 8) * height
    image_data = rng.read(data_size)

    # 7. Trailer
    trailer = b';'

    # Write all of them to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if bool(gce):
        out.write(gce)
    out.write(image_descriptor)
    out.write(image_data)
    out.write(trailer)


if __name__ == '__main__':
    try:
        rng = open('/dev/urandom', 'rb')
    except FileNotFoundError:
        print("Couldn't find '/dev/urandom'. Are you running this on Linux?")
        exit()

    try:
        with open(sys.argv[1], 'wb') as f:
            generate_random_gif(rng, f)
    except IndexError:
        with open('./test.gif', 'wb') as f:
            generate_random_gif(rng, f)