#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate(out: BinaryIO, rng):
    # Header (6 bytes)
    header = b'GIF89a'

    # Logical screen descriptor (7 bytes)
    lsd = bytearray()
    lsd += struct.pack('<H', 1024)      # Width
    lsd += struct.pack('<H', 512)       # Height
    lsd += struct.pack('B', 8)          # Color depth
    lsd += b'\x00'                     # Background color index
    lsd += b'\x00'                     # Pixel aspect ratio

    # Global color table (3 * 256 = 768 bytes)
    color_table = bytearray()
    for i in range(256):
        r = int((i / 255) ** 2 * 255 + 0.5) % 256
        g = int(((r - i) / 255) ** 2 * 255 + 0.5) % 256
        b = int(((g - r) / 255) ** 2 * 255 + 0.5) % 256

        color_table += struct.pack('BBB', r, g, b)

    # Graphics control extension (6 bytes)
    gce = None

    # Image descriptor (10 bytes)
    image_descriptor = bytearray()
    image_descriptor += struct.pack('<H', 0)   # Left position
    image_descriptor += struct.pack('<H', 0)   # Top position
    image_descriptor += struct.pack('<H', 1024) # Width
    image_descriptor += struct.pack('<H', 512)  # Height
    image_descriptor += b'\x00'                 # Local color table flag
    image_descriptor += b'\x08'                 # LZW minimum code size

    # Blocks of data
    blocks = []

    # 1. Plain text
    plain_text = b'Hello World!'
    blocks.append(plain_text)

    # 2. Graphic control extension
    if not gce:
        gce = bytearray([
            0x21, 0xf9, 0x04, 0x00, 0xff,     # Extension introducer and label
            0x00,                             # Block size
            0x00,                             # Packed field
            0x00, 0x00,                       # Delay time
            0x00                              # Transparent color index
        ])

    # 3. Application extension
    application_extension = bytearray([
        0x21, 0xff, 0x0b,                   # Extension introducer and label
        0x00,                               # Block size
        0x4e, 0x45, 0x54, 0x53, 0x43, 0x41, # NETSCAPE
        0x50, 0x45,                         # Application identifier
        0x32,                               # Application authentication code
        0x2e, 0x30,                         # Sub-block length
        0x00, 0x00                          # Data sub-block
    ])
    blocks.append(application_extension)

    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if len(gce) == 6:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)