#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
    out: BinaryIO, 
    rng: BinaryIO):

    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    color_depth = rng.read(1)[0] & 0b1111
    global_color_table = bool(rng.read(1)[0] & 0b10000000)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBB', width, height, color_depth, 
        int(global_color_table), background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    color_table = bytearray()
    for i in range(pow(2, color_depth)):
        r, g, b = tuple([int(i * 255 / pow(2, color_depth - 1))] * 3)
        color_table += struct.pack('BBB', r, g, b)
        
    # 4. Graphics Control Extension
    transparent_color_flag = bool(rng.read(1)[0] & 0b10000000)
    user_input_flag = bool(rng.read(1)[0] & 0b10000000)
    disposal_method = rng.read(1)[0] & 0b111
    delay_time = struct.unpack('>H', rng.read(2))[0]
    transparent_color_index = rng.read(1)[0]
    gce = struct.pack('>BBHB', 0x21, 0xF9, 4, 
        0 | int(transparent_color_flag) << 7 | int(user_input_flag) << 6 | disposal_method << 2 | int(transparent_color_index!= None))
        
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    image_data = []
    while True:
        block_size = rng.read(1)[0]
        

        if not block_size:
            break
        
        image_data += [rng.read((block_size % 255) + 1)]
        
    # 7. Trailer
    trailer = b';'

    # Write everything to our output stream in order
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    out.writelines(image_data)
    out.write(trailer)

