#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    print(f"Width: {width}, Height: {height}")
    lsd = struct.pack('>HHBBB', width, height, 0xF7, 0x0B, 0x00)

    # 3. Global Color Table
    # We'll use only one color in this example
    color_table = bytearray([
        0xFF, 0x00, 0x00, 0x00,
        0x00, 0xFF, 0x00, 0x00,
        0x00, 0x00, 0xFF, 0x00,
        0x00, 0x00, 0x00, 0xFF,
    ])

    # 4. Graphics Control Extension
    # This extension will be used when we have an animation
    gce = None

    # 5. Image Descriptor
    # The image descriptor tells us that there's no local color table
    # and that the first frame of the animation starts at position 0
    image_descriptor = struct.pack('>BBHBBBB', 0x2C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00)

    # 6. Image Data
    # Generate some random data for our image
    image_data = bytearray()
    for _ in range((width * height)):
        image_data += struct.pack('<B', rng.read(1)[0])

    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom

    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)