#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            break

        data = rng.read(block_size)
        blocks.append((lzw_minimum_code_size, data))

    # Write everything to output
    out.write(header + lsd)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)

    # Compress all blocks together
    compressed = b''
    for minimum_code_size, data in blocks:
        compressed += struct.pack('>B', minimum_code_size)
        compressed += compress(data, minimum_code_size)

    out.write(compressed)


# Compresses `data` using LZW with code size `minimum_code_size`.
# Returns the resulting compressed data as a binary string.
def compress(data: bytes, minimum_code_size):
    # Initialize dictionary of codes
    codes = {bytes([i]): i for i in range(2 ** minimum_code_size)}

    # Initialize buffer with first character from input
    buffer = [data[0]]

    # For every other character...
    for char in data[1:]:
        # If current character already exists in our dictionary...
        if char in codes:
            # Add it to the buffer
            buffer.append(char)

            # And continue onwards
            continue

        # Otherwise, find the longest prefix that already exists in our dictionary
        best_prefix = ''
        for key in codes:
            if len(key) > len(best_prefix) and key.startswith(buffer):
                best_prefix = key

        # Add this new prefix to our dictionary
        codes[b''.join(buffer)] = len(codes)

        # Clear buffer and add new character to end
        buffer = [char]

    # Flush remaining buffer contents
    result = b''
    for char in buffer:
        result += struct.pack('>B', codes[bytes([char])])

    return result


if __name__ == '__main__':
    import sys
    import os
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-o', '--output')
    args = parser.parse_args()

    if args.output:
        with open(args.output, 'wb') as f:
            generate_random_gif(os.urandom(), f)
    else:
        generate_random_gif(os.urandom(), sys.stdout.buffer)