#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 256 + int.from_bytes(rng.read(1), 'big') % (2 << 7)
    height = 256 + int.from_bytes(rng.read(1), 'big') % (2 << 7)
    flags = b'\x00\x00' if width < 2**12 else \
            b'\xff\xe0' if width >= 2**12 and width < 2**14 else \
            b'\xff\xf0' if width >= 2**14 and width < 2**16 else \
            b'\xff\xfe' if width >= 2**16 and width < 2**18 else \
            b'\xff\xff'
    bg_color_index = int.from_bytes(rng.read(1), 'big') & 0x0f
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big') & 0x0f
    out.write(struct.pack('<HHBBB',
                         width, height,
                         flags[0], flags[1],
                         bg_color_index,
                         pixel_aspect_ratio))

    # Global color table
    global_color_table_size = 2 ** ((int.from_bytes(rng.read(1), 'big') & 0x07) + 1)
    for i in range(global_color_table_size):
        r = int.from_bytes(rng.read(1), 'big') & 0xff
        g = int.from_bytes(rng.read(1), 'big') & 0xff
        b = int.from_bytes(rng.read(1), 'big') & 0xff
        out.write(struct.pack('BBB',
                             r, g, b))

    # Image descriptor
    left = 256 + int.from_bytes(rng.read(1), 'big') % (2 << 7)
    top = 256 + int.from_bytes(rng.read(1), 'big') % (2 << 7)
    width = 256 + int.from_bytes(rng.read(1), 'big') % (2 << 7)
    height = 256 + int.from_bytes(rng.read(1), 'big') % (2 << 7)
    flags = b'\x00\x00' if width < 2**12 else \
            b'\xff\xe0' if width >= 2**12 and width < 2**14 else \
            b'\xff\xf0' if width >= 2**14 and width < 2**16 else \
            b'\xff\xfe' if width >= 2**16 and width < 2**18 else \
            b'\xff\xff'
    local_color_table_flag = bool(int.from_bytes(rng.read(1), 'big') & 0x80)
    interlace_flag = bool(int.from_bytes(rng.read(1), 'big') & 0x40)
    sort_flag = bool(int.from_bytes(rng.read(1), 'big') & 0x20)
    reserved = int.from_bytes(rng.read(1), 'big') & 0x1f
    lct_size = 2 ** ((reserved & 0x07) + 1)
    if local_color_table_flag:
        for i in range(lct_size):
            r = int.from_bytes(rng.read(1), 'big') & 0xff
            g = int.from_bytes(rng.read(1), 'big') & 0xff
            b = int.from_bytes(rng.read(1), 'big') & 0xff
            out.write(struct.pack('BBB',
                                 r, g, b))
    code_size = int.from_bytes(rng.read(1), 'big') & 0x07
    out.write(struct.pack('<HHBBBBB',
                         left, top,
                         width, height,
                         flags[0], flags[1],
                         code_size))

    # LZW minimum code size
    min_code_size = int.from_bytes(rng.read(1), 'big') & 0x07
    out.write(min_code_size.to_bytes(1, byteorder='little'))

    # Data blocks
    data_block_size = 2 ** (min_code_size + 1) - 1
    while True:
        block_type = int.from_bytes(rng.read(1), 'big') & 0x03

        if block_type == 0:
            break

        elif block_type == 1:
            num_bytes = int.from_bytes(rng.read(1), 'big') & 0xff
            out.write(num_bytes.to_bytes(1, byteorder='little'))

            bytes_data = rng.read(num_bytes)
            out.write(bytes_data)

        elif block_type == 2:
            num_entries = int.from_bytes(rng.read(1), 'big') & 0xff
            out.