#!/usr/bin/env python3

import os
import sys
import struct
from random import Random


def main():
    in_filename, out_filename = sys.argv[1:]
    lsd, gce, color_table, image_descriptor = read_header(in_filename)
    write_header(out_filename, lsd, gce, color_table, image_descriptor)


def read_header(filename):
    with open(filename, 'rb') as f:

        # 1. Logical Screen Descriptor (LSD)
        lsd = f.read(13)
        width, height, bit_depth, compression, filter, interlace = \
            struct.unpack('>HHBBBBB', lsd[-8:])
        assert compression == 0 and filter == 0 and interlace == 0

        # 2. Global Color Table (GCT)
        color_table_size = 3 * pow(2, bit_depth) - 1
        color_table = f.read(color_table_size * 3)

        # 3. Image Descriptor (ID)
        image_descriptor = f.read(9)

        # 4. Graphics Control Extension (GCE)
        gce = None
        while True:
            chunk_type = f.read(4)
            if len(chunk_type)!= 4 or chunk_type!= b'gCe':
                break

            chunk_length = int.from_bytes(f.read(4), byteorder='big')
            chunk_contents = f.read(chunk_length)
            gce = chunk_type + chunk_length.to_bytes(4, byteorder='big') + chunk_contents

        return lsd, gce, color_table, image_descriptor


def write_header(filename, lsd, gce, color_table, image_descriptor):
    with open(filename, 'wb') as out:
        write_header_(out, lsd, gce, color_table, image_descriptor)


def write_header_(out, lsd, gce, color_table, image_descriptor):
    # 5. Image Data
    rng = Random()
    rng.seed(os.urandom(64))
    lzw_minimum_code_size = 8
    blocks = []
    for i in range(10000):
        block_size = rng.randint(1, 1000)
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')