#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor (LSD)
    width = int.from_bytes(rng.read(1), byteorder='big') % 640 + 1
    height = int.from_bytes(rng.read(1), byteorder='big') % 480 + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='big') & ~0x80])
    bg_color = rng.read(1)[0]
    pixel_aspect_ratio = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    lsd = struct.pack('>HHBBB', width, height, *flags, bg_color, *pixel_aspect_ratio)
    
    # 3. Color Table
    num_colors = min((width*height)//50, 256)
    color_table = [rng.read(3) for _ in range(num_colors)]
    color_table += [(0,)*3]*(256-len(color_table))
    color_table = b''.join(color_table)
    
    # 4. Graphics Control Extension (GCE)
    has_gce = bool(int.from_bytes(rng.read(1), byteorder='big'))
    if has_gce:
        delay = int.from_bytes(rng.read(2), byteorder='big')
        transparent_index = int.from_bytes(rng.read(1), byteorder='big')
        flags = bytes([int.from_bytes(rng.read(1), byteorder='big') | 0x80])
        gce = struct.pack('>HBB', delay, transparent_index, *flags)
    else:
        gce = None
    
    # 5. Image Descriptor
    left = int.from_bytes(rng.read(1), byteorder='big')
    top = int.from_bytes(rng.read(1), byteorder='big')
    width = int.from_bytes(rng.read(1), byteorder='big')
    height = int.from_bytes(rng.read(1), byteorder='big')
    flags = bytes([int.from_bytes(rng.read(1), byteorder='big') & ~0x80])
    img_lct = int.from_bytes(rng.read(1), byteorder='big')
    image_descriptor = struct.pack('>HHHHB', left, top, width, height, *flags, img_lct)
    
    # 6. Image Data
    num_pixels = width*height
    image_data = []
    while len(image_data) < num_pixels:
        block_size = int.from_bytes(rng.read(1), byteorder='big')
        if block_size > 0:
            image_data.append(block_size)
            image_data += list(rng.read(block_size))
        elif block_size == 0:
            break
        else:
            raise ValueError("Invalid block size")
    image_data = bytes(image_data)
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom

    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)