#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = min(max(rng.read(1)[0] & 0x07, 2), 256)
    color_table = []
    for i in range(num_colors):
        color_table += [struct.pack('BBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])]
    
    # 4. Graphics Control Extension (we'll use this later)
    gce = b''
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHBBBB', 0x2C, rng.read(1)[0], width, height, 0b11100000 | rng.read(1)[0], 0)
    
    # 6. Image Data
    image_data = []
    while True:
        block_size = rng.read(1)[0]

        if block_size == 0:
            continue
            
        elif block_size == 0x3B:
            break
        
        image_data += [rng.read((block_size % 255) + 1)]
        
    # 7. Trailer
    trailer = b';'

    # Write everything to our output stream in order
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    out.writelines(image_data)
    out.write(trailer)

