#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = None
    if rng.read(1)[0]:
        # Transparent color flag set to true
        transparent_color_index = rng.read(1)[0]

        # Delay time in hundredths of seconds
        delay_time = rng.read(2)

        # Transparent color index
        gce = b'\x21\xF9\x04' + delay_time + transparent_color_index.to_bytes(1, 'big') + b'\x00'

    # 5. Image Descriptor
    # Left position is always zero
    left = 0

    # Top position is always zero
    top = 0

    # Width and Height are random again
    width, height = struct.unpack('>HH', rng.read(4))

    # Flags field: packed fields (using a random byte for variety), Local color table flag set to false
    flags = struct.pack('>BBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # Image data starts with an empty byte indicating no local color table follows
    image_data = b'\x00'

    # Generate pixels (using a random byte for each pixel)
    for _ in range(width * height):
        image_data += rng.read(1)

    # 6. Trailer
    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)