#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    
    # Other ID fields: Packed fields (using a random byte for variety), Local color table flag, Interlace flag, Sort flag, Size of local color table
    id = struct.pack('>BBBBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    image_descriptor = left + top + id
    
    # 6. Blocks
    # Let's make this an infinite loop!
    while True:
        # Each block is either a plain text extension or a graphic control extension followed by a plain text extension
        block_type = rng.read(1)[0]
        
        # Plain Text Extension
        if block_type == 0x01:
            # Block size (always 126)
            out.write(struct.pack('B', 126))
            
            # Text sub-block
            # First byte: Always 0x01
            # Second byte: Always 0xFE
            # Third byte: Random byte
            # Fourth byte: Always 0x00
            # Fifth byte: Always 0x01
            # Sixth byte: Random byte
            # Seventh byte: Always 0x00
            # Eighth byte: Always 0x00
            # Ninth byte: Always 0x00
            # Tenth byte: Always 0x00
            # Rest of data: Random bytes
            out.write(bytes([
                0x01, 0xfe,
                rng.read(1)[0], 0x00,
                0x01,
                rng.read(1)[0], 0x00, 0x00, 0x00]))
            
        # Graphic Control Extension
        elif block_type == 0xf9:
            # Block size (always 4)
            out.write(struct.pack('B', 4))
            
            # Packed fields: Reserved field (always 0), Disposal method (randomly choose between 0, 1, 2, 3), User input flag (randomly choose between 0 and 1), Transparent color index (randomly choose from [0..num_colors - 1]), Delay time (randomly choose from [0..65535])
            packed_fields = struct.pack('BBBBH', 0, rng.read(1)[0] & 0x03, rng.read(1)[0] & 0x01, rng.read(1)[0] % num_colors, rng.read(2)[0])
            
            # Write everything to the output file
            out.write(packed_fields)
            
            # Sub-block
            # First byte: Always 0x01
            # Second byte: Always 0xFF
            # Third byte: Always 0x00
            # Fourth byte: Always 0x00
            # Fifth byte: Always 0x00
            # Sixth byte: Always 0x00
            # Seventh byte: Always 0x00
            # Eighth byte: Always 0x00
            # Ninth byte: Always 0x00
            # Tenth byte: Always 0x00
            # Rest of data: Random bytes
            out.write(bytes([
                0x01, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]))
            
        else:
            raise ValueError("Invalid block type")

    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if len(gce) == 6:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)