#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = rng.read(1)[0] % 64 + 1
    h = rng.read(1)[0] % 64 + 1
    logicalScreenDescriptor = \
            struct.pack('<HHBBB', w * 256, h * 256,
                        7, 0, 0)

    # 3. Global Color Table
    globalColorTable = bytearray()
    for i in range(2**((logicalScreenDescriptor[7] & 7) + 1)):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        globalColorTable += struct.pack('BBB', r, g, b)

    # 4. Image Descriptor
    imageDescriptor = struct.pack('<HHHHB',
                                  w * 256, h * 256,
                                  0, 0, 0)

    # 5. Local Color Table
    localColorTable = bytearray()
    if len(globalColorTable):
        localColorTable = globalColorTable[:]

        # Make sure that there are at least one entry in the table
        while len(localColorTable) < 2 ** ((imageDescriptor[7] & 7) + 1):
            localColorTable += struct.pack('BBB',
                                           rng.read(1)[0],
                                           rng.read(1)[0],
                                           rng.read(1)[0])

    # 6. Image Data
    imageData = bytearray()
    for y in range(h):
        for x in range(w):
            index = rng.read(1)[0] % len(localColorTable)

            # Set the transparent color to black
            if index == 0:
                index = 1

            imageData += struct.pack('B', index)

    # Write everything to the output stream
    out.write(header)
    out.write(logicalScreenDescriptor)
    out.write(b'\x87')
    out.write(struct.pack('<H', len(globalColorTable)))
    out.write(globalColorTable)
    out.write(imageDescriptor)
    out.write(b'\x87')
    out.write(struct.pack('<H', len(localColorTable)))
    out.write(localColorTable)
    out.write(imageData)


if