#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
	width &= 0x7FFF
	height &= 0x7FFF

    # Color table size is fixed to 256 colors
    color_table_size = 2 ** 8 - 1

    # Global color table flag set to true
    global_color_table_flag = True

    # Background color index is always zero
    background_color_index = 0

    # Pixel aspect ratio is always 1
    pixel_aspect_ratio = 1

    # Pack descriptor
    lsd = struct.pack('<HHBBBH',
                      width,
                      height,
                      color_table_size,
                      global_color_table_flag,
                      background_color_index,
                      pixel_aspect_ratio
                     )

    # 3. Image Descriptor
    image_separator = b','

    # Image left position is always zero
    image_left = 0

    # Image top position is always zero
    image_top = 0

    # Image width and height are equal to logical screen dimensions
    image_width = width
    image_height = height

    # Packed descriptor
    idd = struct.pack('<BBHHHH',
                      image_separator,
                      0,
                      image_left,
                      image_top,
                      image_width,
                      image_height
                     )

    # 4. Local Color Table
    local_color_table = bytearray()

    # Generate a random RGB triplet for every possible color in the palette
    for i in range(2 ** 8):
        r, g, b = [int(rng.read(1)[0]) & 0xFF] * 3

        local_color_table += struct.pack('<BBB', r, g, b)

    # 5. Graphics Control Extension
    graphics_control_extension_label = b'\x21\xF9\x04'

    # Size of the extension block is 4 bytes long
    graphics_control_extension_block_size = 4

    # Packed descriptor
    gce = struct.pack('<BHB',
                      graphics_control_extension_block_size,
                      0b00000000 | 0b00000000,
                      0
                     )

    # 6. Application Extension
    application_extension_label = b'\x21\xFF\x0B'

    # Size of the extension block is 11 bytes long
    application_extension_block_size = 11

    # Application identifier is "NETSCAPE"
    netscape_application_identifier = b'NETSCAPE' + b'\x03\x01'

    # Application authentication code is always 0
    application_authentication_code = 0

    # Packed descriptor
    ae = struct.pack('<BBBB%dsBHB' % len(netscape_application_identifier),
                      application_extension_block_size,
                      0,
                      ord(netscape_application_identifier[0]),
                      ord(netscape_application_identifier[1]),
                      ord(netscape_application_identifier[2]),
                      netscape_application_identifier,
                      application_authentication_code,
                      3,
                      1
                     )

    # 7. Comment Extension
    comment_extension_label = b'\x21\xFE'

    # Size of the extension block is 128 bytes long
    comment_extension_block_size = 128

    # Comment data is randomly generated
    comment_data = bytearray([int(rng.read(1)[0]) & 0xFF] * comment_extension_block_size)

    # Packed descriptor
    ce = struct.pack('<BB%ds' % len(comment_data),
                      comment_extension_block_size,
                      ord(comment_data[0]),
                      comment_data
                     )

    # 8. Trailer
    trailer = b';'

    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    out.write(idd)
    out.write(local_color_table)
    out.write(gce)
    out.write(ae)
    out.write(ce)
    out.write(trailer)


if