#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBB',
                      width,
                      height,
                      packed_fields,
                      background_color_index,
                      pixel_aspect_ratio
                     )

    # 3. Global Color Table
    global_color_table_flag = bool((packed_fields >> 7) & 1)
    global_color_table_size = ((packed_fields >> 4) & 7) + 1
    global_color_table = rng.read(global_color_table_size * 3)

    # 4. Application Extension
    application_extension_label = b'\x21\xFF\x0BNETSCAPE2.0'
    application_identifier = b'NETSCAPE2.0'
    application_auth_code = b'\<PASSWORD>'
    application_data = b'\x03\x01\x00\x00'
    ae = application_extension_label + \
         len(application_identifier).to_bytes(1, 'big') + \
         application_identifier + \
         len(application_auth_code).to_bytes(1, 'big') + \
         application_auth_code + \
         len(application_data).to_bytes(1, 'big') + \
         application_data

    # 5. Graphics Control Extension
    gce = None
    if rng.read(1)[0]:  # If there's a GCE...
        gce_block_size = rng.read(1)[0]
        gce_flags = rng.read(1)[0]
        delay_time = struct.unpack('>H', rng.read(2))[0]
        transparent_color_index = rng.read(1)[0]
        terminator = rng.read(1)[0]
        assert terminator == 0

        gce = struct.pack('>BBBBHB',
                          gce_block_size,
                          0xF9,
                          gce_flags,
                          0,
                          delay_time,
                          transparent_color_index
                         )

    # 6. Image Descriptor
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = rng.read(2)[0]
    height = rng.read(2)[0]
    local_color_table_flag = bool((packed_fields >> 5) & 1)
    interlace_flag = bool((packed_fields >> 6) & 1)
    sort_flag = bool((packed_fields >> 3) & 1)
    reserved_flags = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = rng.read(local_color_table_size * 3)
    image_descriptor = struct.pack('>BHHHHBBBBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  local_color_table_size
                                 )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')