#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = min((rng.read(1)[0] % 8) or 1, 7) * 3
    global_color_table = [struct.pack('BBB', *(rng.read(3))) for _ in range(global_color_table_size)]
    color_table = b''.join(global_color_table)

    # 4. Graphics Control Extension (if needed)
    gce = None
    if rng.read(1)[0]:
        # Delay time (2 bytes)
        delay_time = struct.unpack('>H', rng.read(2))[0]

        # Transparent color index (1 byte)
        transparent_color_index = rng.read(1)[0]

        # Block terminator (1 byte)
        block_terminator = rng.read(1)[0]

        assert block_terminator == 0

        gce = struct.pack('>BBHBB',
                          0x21,
                          0xF9,
                          delay_time,
                          0x01,
                          transparent_color_index
                         )

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved_flags = rng.read(1)[0] & 0x07
    local_color_table_size = ((rng.read(1)[0] % 8) or 1) * 3
    image_descriptor = struct.pack('>HHBBB',
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  local_color_table_size
                                 )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')