#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
	# 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), byteorder='big') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='big') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    bgColorIndex = int.from_bytes(rng.read(1), byteorder='big') & 0b00001111
    pixelAspectRatio = int.from_bytes(rng.read(1), byteorder='big')

    lsd = struct.pack('<6sHHBBB',
                      header,
                      7 + len(flags),
                      width,
                      height,
                      flags[0],
                      bgColorIndex,
                      pixelAspectRatio)

	# 3. Global Color Table
    globalColorTableSize = int.from_bytes(rng.read(1), byteorder='big')
    if globalColorTableSize == 0:
        globalColorTable = b''
    else:
        globalColorTable = bytes([(i * i * i) % 256 for i in range(globalColorTableSize)])

	# 4. Image Descriptor
    leftPos = int.from_bytes(rng.read(2), byteorder='big') % (width - 1)
    topPos = int.from_bytes(rng.read(2), byteorder='big') % (height - 1)
    imageWidth = int.from_bytes(rng.read(2), byteorder='big') % (width - leftPos) + 1
    imageHeight = int.from_bytes(rng.read(2), byteorder='big') % (height - topPos) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    localColorTableFlag = bool((flags[0] >> 7) & 1)
    interlaceFlag = bool((flags[0] >> 6) & 1)
    sortFlag = bool((flags[0] >> 5) & 1)
    reserved = ((flags[0] << 2) & 0b11000000) | (((flags[0] << 1) & 0b00100000) >> 4) | ((flags[0] & 0b00010000) >> 3)
    lctSize = int.from_bytes(rng.read(1), byteorder='big')
    lct = b''
    if localColorTableFlag:
        lct = bytes([(i * i * i) % 256 for i in range(lctSize)])
    else:
        lct = globalColorTable[:lctSize]
    
    imgd = struct.pack('<HHHHB',
                       leftPos,
                       topPos,
                       8 + len(flags) + len(lct),
                       imageWidth,
                       imageHeight) \
           + flags \
           + bytes([reserved]) \
           + bytes([lctSize])

    # 5. Local Color Table
    # 6. Image Data
    data = bytes([(i * i * i) % 256 for i in range(imageWidth * imageHeight)])

    # Write everything to output stream
    out.write(lsd)
    out.write(globalColorTable)
    out.write(imgd)
    out.write(lct)
    out.write(data)