#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    # Set color resolution to 7 bits per pixel
    color_resolution = 0b00000111
    # Set global color table flag to False
    gct_flag = 0
    # Set sort flag to False
    sort_flag = 0
    # Size of global color table is 0
    gct_size = 0
    # Background color index is 0
    bg_color_index = 0
    # Pixel aspect ratio is 0
    pixel_aspect_ratio = 0
    lsd = struct.pack('<HHBBBHHH',
                      width, height,
                      color_resolution, gct_flag, sort_flag,
                      gct_size, bg_color_index, pixel_aspect_ratio
                     )

    # 3. Global Color Table
    # No global color table
    gct = b''

    # 4. Image Descriptor
    # Frame left position is 0
    frame_left_pos = 0
    # Frame top position is 0
    frame_top_pos = 0
    # Frame width is equal to logical screen descriptor's width
    frame_width = width
    # Frame height is equal to logical screen descriptor's height
    frame_height = height
    # Local color table flag set to False
    lct_flag = 0
    # Interlace flag set to False
    interlace_flag = 0
    # Sort flag set to False
    sort_flag = 0
    # Reserved flags are all zeroes
    reserved_flags = 0
    # LZW minimum code size is 2
    lzw_min_code_size = 2
    image_descriptor = struct.pack('<HHHHBB',
                                   frame_left_pos, frame_top_pos,
                                   frame_width, frame_height,
                                   lct_flag, interlace_flag,
                                   sort_flag, reserved_flags
                                  )

    # 5. Local Color Table
    # No local color table
    lct = b''

    # 6. Graphics Control Extension
    # Block size is 0x04
    block_size = 0x04
    # Packed field contains no transparency
    packed_field = 0x00
    # Delay time is 0
    delay_time = 0
    transparent_color_index = 0
    graphics_control_extension = struct.pack('<BHBH',
                                             block_size, packed_field,
                                             delay_time,
                                             transparent_color_index
                                            )

    # 7. Plain Text Extension
    # Block size is 0x0C
    block_size = 0x0C
    # Text grid left position is 0
    text_grid_left_pos = 0
    # Text grid top position is 0
    text_grid_top_pos = 0
    # Text grid width is equal to logical screen descriptor's width
    text_grid_width = width
    # Text grid height is equal to logical screen descriptor's height
    text_grid_height = height
    # Cell width is 0
    cell_width = 0
    # Cell height is 0
    cell_height = 0
    # Fg color index is 0
    fg_color_index = 0
    # Bg color index is 0
    bg_color_index = 0
    plain_text_extension = struct.pack('<BBBBBBB',
                                       block_size,
                                       text_grid_left_pos, text_grid_top_pos,
                                       text_grid_width, text_grid_height,
                                       cell_width, cell_height,
                                       fg_color_index, bg_color_index
                                      )

    # 8. Application Extension
    # Block size is 0x0B
    block_size = 0x0B
    # Application identifier is 'NETSCAPE2.0'
    application_identifier = b'NETSCAPE2.0'
    # Application authentication code is 'XXXX' where X is a byte from /dev/urandom
    application_authentication_code = rng.read(3)
    application_extension = struct.pack('<BB%dsB' % len(application_identifier),
                                        block_size,
                                        0xFF,
                                        application_identifier,
                                        *application_authentication_code
                                       )

    # 9. Comment Extension
    # Block size is 0x02
    block_size = 0x02
    comment_extension = struct.pack('<BB',
                                    block_size,
                                    0xFE
                                   )

    # 10. Trailer
    trailer = b';'

    # Write everything in order
    out.write(header + lsd + gct + image_descriptor
              + lct + graphics_control_extension + plain_text_extension
              + application_extension + comment_extension + trailer
             )
