#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None
    
    # 5. Image Descriptor
    # Left position, top position, width, height, packed field
    image_descriptor = struct.pack('>HHHHB', 0, 0, width, height, rng.read(1)[0])
    
    # 6. Image Data
    # First byte of the data stream will be the LZW minimum code size
    mincodesize = rng.read(1)[0]
    image_data = bytes([mincodesize])
    
    # Next we'll write some random pixels until we reach the end of the frame
    while True:
        # We need to pick a random length between 2 and 255 inclusive
        len_bytes = rng.read(1)[0]
        
        # If that value was zero, then we're done!
        if len_bytes == 0:
            break
        
        # Otherwise, let's add another chunk of data to our stream
        image_data += bytes([len_bytes]) + rng.read(len_bytes)
    
    # Finally, we'll close off the image by adding an EOI marker
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom

    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)