#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Set color resolution to 7 bits per pixel
    color_resolution = 0b111

    # Set global color table flag to false
    gct_flag = False

    # Set sort flag to false
    sorted_flag = False

    # Size of global color table is zero
    size_of_gct = 0

    # Background color index is zero
    background_color_index = 0

    # Pixel aspect ratio is one
    pixel_aspect_ratio = 0

    logical_screen_descriptor = struct.pack(
        '>HHBBB', 
        width, height,
        color_resolution,
        gct_flag << 7 | sorted_flag << 6 | size_of_gct,
        background_color_index,
        pixel_aspect_ratio
    )

    # Write everything to output stream
    out.write(header + logical_screen_descriptor)

    # 3. Global Color Table
    if not gct_flag:
        # No global color table so just write zeros
        out.write(bytes([0] * (size_of_gct * 3)))

    # 4. Image Descriptor
    image_separator = b','

    left = 0
    top = 0
    width = 0
    height = 0

    local_color_table_flag = True
    interlace_flag = False
    sort_flag = False
    reserved = 0
    lct_size = 0x100

    image_descriptor = struct.pack(
        '>BHHBBBBB',
        local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved,
        0, 0, width, height,
        lct_size,
        background_color_index
    )

    # Write everything to output stream
    out.write(image_separator + image_descriptor)

    # 5. Local Color Table
    if local_color_table_flag:
        # Write all colors in the palette
        out.write(bytes([i % 256 for i in range(lct_size * 3)]))

    # 6. Image Data
    # Write all pixels as black or white depending on whether they are even or odd
    out.write(bytes([(i >> 1) & 1 for i in range((width * height) // 8)]))