#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved_flags = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = rng.read(local_color_table_size * 3)
    image_descriptor = struct.pack('>BHHHHBBBBB',
                                  0x2C,
                                  left, top,
                                  width, height,
                                  local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved_flags,
                                  local_color_table_size,
                                  *struct.unpack('>' + 'B' * len(local_color_table), local_color_table)
                                )

    # 6. Local Color Table (if applicable)
    if local_color_table_flag:
        pass

    # 7. Image Data
    data = b''
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x21:
            # Extension Block
            extension_block_label = rng.read(1)[0]

            if extension_block_label == 0xFF:
                # Application Extension
                application_identifier = rng.read(8).decode()

                if application_identifier == 'NETSCAPE':
                    application_auth_code = rng.read(3)

                    sub_block_length = rng.read(1)[0]
                    sub_block_data = rng.read(sub_block_length)

                    if sub_block_data[0] == 1:
                        loop_count = struct.unpack('>H', sub_block_data[1:])[0]

                        netscape_extension = struct.pack('>BBHBB',
                                                        0x21,
                                                        extension_block_label,
                                                        sub_block_length + 3,
                                                        0x11,
                                                        3,
                                                        1,
                                                        *loop_count.to_bytes(2, 'big'),
                                                      )

                        data += netscape_extension
                        continue

            elif extension_block_label in [0xF9, 0xFE]:
                # Graphic Control Extension
                length = rng.read(1)[0]
                flags = rng.read(1)[0]
                delay_time = struct.unpack('>H', rng.read(2))[0]
                transparent_color_index = rng.read(1)[0]

                graphic_control_extension = struct.pack('>BBBBHBB',
                                                        0x21,
                                                        extension_block_label,
                                                        length,
                                                        flags,
                                                        delay_time >> 8,
                                                        delay_time & 0xFF,
                                                        transparent_color_index,
                                                       )

                data += graphic_control_extension
                continue

            elif extension_block_label == 0x01:
                # Plain Text Extension
                text_grid_left, text_grid_top = struct.unpack('>HH', rng.read(4))
                text_grid_width, text_grid_height = struct.unpack('>HH', rng.read(4))
                cell_width, cell_height = struct.unpack('>BB', rng.read(2))
                foreground_color_index = rng.read(1)[0]
                background_color_index = rng.read(1)[0]

                plain_text_extension = struct.pack('>BBHHBBBBBB',
                                                   0x21,
                                                   extension_block_label,
                                                   text_grid_left,
                                                   text_grid_top,
                                                   text_grid_width,
                                                   text_grid_height,
                                                   cell_width,
                                                   cell_height,
                                                   foreground_color_index,
                                                   background_color_index
                                                  )

                data += plain_text_extension
                continue

            else:
                raise Exception("Unknown extension block label")


        elif block_type == 0x2C:
            # Image Data
            break

        else:
            raise Exception("Unknown block type")


    # 8. Trailer
    trailer = b'\x3B'

    return sum([len(lsd), len(color_table), len(image_descriptor)])

