#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = None
    if rng.read(1):  # If we get nonzero data back from this read...
        # Duration in milliseconds (use a random value between 0 and 1000 inclusive)
        duration = rng.read(2)

        # Transparent color index (use a random value between 0 and our color table length - 1)
        transparent_color_index = rng.read(1) % len(color_table)

        # Block terminator
        gce = b'\x21\xf9\x04' + duration + transparent_color_index + b'\x00'

    # 5. Image Descriptor (always present)
    # Left position is always zero
    left = b'\x00'

    # Top position is randomly chosen between 0 and 2^(height - 1) - 1 inclusive
    top = rng.read(1)
    top += int((rng.read(1)[0] & ((1 << (height - 1)) - 1))) << 8

    # Width and height are randomly chosen between 0 and 2^(width - 1) - 1 inclusive
    width = rng.read(1)
    width += int((rng.read(1)[0] & ((1 << (width - 1)) - 1))) << 8

    height = rng.read(1)
    height += int((rng.read(1)[0] & ((1 << (height - 1)) - 1))) << 8

    # Packed field (using a random byte for variety)
    packed_field = rng.read(1)[0]

    # Image descriptor
    image_descriptor = b'\x2c' + left + top + width + height + packed_field

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)