#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHBB', 0x2C, 0, 0, 0, 0)

    # 6. Image Data
    # Let's use LZW compression with a random minimum code size
    lzw_minimum_code_size = rng.read(1)[0] & 0b00000111
    blocks = []
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x00:
            break
        
        elif block_type == 0x01:
            # Plain text extension
            pass

        else:
            # LZW compressed data

            # Read in a new block of data
            block_length = rng.read(1)[0]
            block = rng.read(block_length)
            
            # If this is not the last block, add an EOI marker
            if block!= b'\xFF':
                block += b'\x00'

            # Add the block to our list of blocks
            blocks.append(block)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)