#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the given parameters.
def generate(out: BinaryIO, width: int = 1024, height: int = 768):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    out.write(struct.pack('<HHBBB',
        width, height,
        0x80 | (0x07 & 0x07),      # Global color table flag and size of global color table
        0,                        # Background color index
        0))                       # Pixel aspect ratio

    # Image descriptor
    out.write(struct.pack('<BHHHHBB',
        0x2C,                     # Image separator
        0, 0,                     # Left position, top position
        width, height,            # Width, height
        0,                        # Local color table flag and size of local color table
        0))                       # LZW minimum code size

    # Color table
    for i in range(0, 256 * 3, 3):
        r = g = b = i // 3 % 256
        out.write(struct.pack('BBB',
            r, g, b))

    # Image data
    lzw_min_code_size = 2
    clear_code = 1 << lzw_min_code_size
    end_of_info_code = clear_code + 1
    next_code = clear_code + 2
    codes = [clear_code]
    prev_code = None
    while True:
        if len(codes) < next_code:
            codes += list(range(next_code - len(codes)))

        if prev_code is not None:
            out.write(bytes([prev_code]))

            if prev_code == end_of_info_code:
                break

            if prev_code == clear_code:
                lzw_min_code_size = 2
                clear_code = 1 << lzw_min_code_size
                end_of_info_code = clear_code + 1
                next_code = clear_code + 2
                codes = [clear_code]
                continue

        new_code = codes[len(codes) - 1]
        for c in reversed(codes[:-1]):
            if bytes([c]) + bytes([new_code]) in bytes(codes):
                new_code = c
                break

        if len(codes) >= (1 << lzw_min_code_size) and \
           len(codes) < (1 << (lzw_min_code_size + 1)):
            lzw_min_code_size += 1
            clear_code = 1 << lzw_min_code_size
            end_of_info_code = clear_code + 1
            next_code = clear_code + 2

        if new_code!= prev_code:
            out.write(bytes([new_code]))
            codes.append(new_code)
            prev_code = new_code
        else:
            out.write(bytes([new_code]))
            codes.append(new_code)
            out.write(bytes([new_code]))
            codes.append(new_code)
            prev_code = None

    # Trailer
    out.write(b'\x3B')

