#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = [0] * 7 + [width & 0xFF, (width >> 8) & 0xFF, height & 0xFF, (height >> 8) & 0xFF, rng.read(1)[0]]

    # Write to output
    out.write(struct.pack('>B' * len(lsd), *lsd))

    # 3. Global Color Table
    # Number of colors in global table is between 2 and 256 inclusive
    num_colors = rng.read(1)[0] % 256 + 2

    # Each entry has three components: Red, Green, Blue
    gct = []
    for _ in range(num_colors):
        r, g, b = rng.read(3)

        # Make sure that all entries are non-zero
        while not (r or g or b):
            r, g, b = rng.read(3)
        
        gct += [r, g, b]
    
    # Write to output
    out.write(bytes([num_colors]) + bytes(gct))

    # 4. Image Descriptor
    # Left position, top position, width, height, packed field
    id = list(range(4)) + [(rng.read(1)[0] | 0x80)]

    # Write to output
    out.write(struct.pack('>H' * len(id), *id))

    # 5. Local Color Table
    # Number of colors in local table is between 2 and 256 inclusive
    num_colors = rng.read(1)[0] % 256 + 2

    # Each entry has three components: Red, Green, Blue
    lct = []
    for _ in range(num_colors):
        r, g, b = rng.read(3)

        # Make sure that all entries are non-zero
        while not (r or g or b):
            r, g, b = rng.read(3)
        
        lct += [r, g, b]
    
    # Write to output
    out.write(bytes([num_colors]) + bytes(lct))

    # 6. Graphics Control Extension
    # Block size, packed fields, delay time, transparent color index, terminator
    block = [0x21, 0xF9, 0x04, ((rng.read(1)[0] % 100) << 2) | 0x01, rng.read(1)[0], 0x00]

    # Write to output
    out.write(bytes(block))

    # 7. Image Data
    # Size of image data is equal to the product of width and height
    size = width * height

    # Write pixels to output
    out.write(rng.read(size))

    return 1


if