#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), byteorder='little') % 500 + 1   # Width between [1, 499]
    height = int.from_bytes(rng.read(2), byteorder='little') % 500 + 1  # Height between [1, 499]
    flags = b'\xf7\x00'                                              # Global Color Table Flag set and Sort Flag unset
    bg_color_index = rng.read(1)[0]                                  # Background Color Index is chosen at random from the global color table
    pixel_aspect_ratio = b'\x00'                                     # Pixel Aspect Ratio is always 0
    lsd = width.to_bytes(2, byteorder='little') + \
          height.to_bytes(2, byteorder='little') + \
          flags + \
          bg_color_index.to_bytes(1, byteorder='little') + \
          pixel_aspect_ratio
    
    # 3. Global Color Table
    num_colors = min((width * height) // 64, 256)                     # Number of colors in the global color table is either floor[(Width x Height) / 64], or 256, whichever is less
    color_table = b''
    for i in range(num_colors):
        r = rng.read(1)[0]                                            # Red component is chosen at random
        g = rng.read(1)[0]                                            # Green component is chosen at random
        b = rng.read(1)[0]                                            # Blue component is chosen at random
        color_table += r.to_bytes(1, byteorder='little') + \
                      g.to_bytes(1, byteorder='little') + \
                      b.to_bytes(1, byteorder='little')

    # 4. Graphics Control Extension
    gce = None                                                        # No GCE will be added to the output unless this flag is set
    if rng.read(1)[0]:                                               # If the first bit of the next byte is set...
        gce = b'\x21\xf9\x04'                                        # Add a graphics control extension

        # 4.1 Disposal Method
        disposal_method = rng.read(1)[0]                              # Choose a disposal method at random
        gce += disposal_method.to_bytes(1, byteorder='little')

        # 4.2 User Input Flag
        user_input_flag = rng.read(1)[0]                               # Set the user input flag randomly
        gce += user_input_flag.to_bytes(1, byteorder='little')

        # 4.3 Transparent Color Index
        transparent_color_index = rng.read(1)[0]                      # Choose a transparent color index at random
        gce += transparent_color_index.to_bytes(1, byteorder='little')

        # 4.4 Delay Time
        delay_time = int.from_bytes(rng.read(2), byteorder='little')     # Choose a delay time at random
        gce += delay_time.to_bytes(2, byteorder='little')

        # 4.5 Transparent Color Index
        terminator = b'\x00'                                          # Terminate the graphics control extension
        gce += terminator

    # 5. Image Descriptor
    left = 0                                                         # Left position is always 0
    top = 0                                                          # Top position is always 0
    flags = b'\x87\x00'                                              # Local Color Table Flag set and Interlace Flag set
    local_color_table_size = len(color_table).to_bytes(1, byteorder='little')      # Size of the local color table is equal to the length of the global color table
    interlace_method = b'\x00'                                       # No interlacing used
    lzw_minimum_code_size = 2 ** rng.read(1)[0]                       # LZW minimum code size is chosen at random from {2^n | n ∈ ℕ}
    image_descriptor = left.to_bytes(2, byteorder='little') + \
                       top.to_bytes(2, byteorder='little') + \
                       flags + \
                       local_color_table_size + \
                       interlace_method + \
                       lzw_minimum_code_size.to_bytes(1, byteorder='little')

    # 6. Image Data
    blocks = []                                                      # List that contains all the compressed data blocks
    while True:
        block_size = rng.read(1)[0]                                  if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)

