#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))

    # Flags field: Packed fields (using a random byte for variety). Local color table flag set by GCE
    flags = struct.pack('>BB', rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    lzw_minimum_code_size = rng.read(1)[0]  # Minimum code size used in compression

    # Let's compress some data!
    # We'll use a simple LZW algorithm that uses a dictionary of previously seen strings
    # The dictionary is limited to 256 entries
    # For more info about this algorithm, see https://en.wikipedia.org/wiki/Lempel%E2%80%93Ziv%E2%80%93Welch
    dictionary = {chr(i): i for i in range(256)}
    string = ''
    blocks = []

    def add_to_dictionary(string):
        if len(dictionary) < 256:
            dictionary[string] = len(dictionary)

    while len(string)!= width * height:
        # Generate a new character from our dictionary
        char = chr((len(string) // width) % 256)

        # If we've never seen this character before, add it to the dictionary
        if string + char not in dictionary:
            add_to_dictionary(string + char)

            # If we're at the end of the line, reset the string
            if len(string) % width == width - 1:
                string = ''

                continue
            
            # Otherwise just append the character
            string += char

        # If we haven't reached the end yet, just append the character
        elif len(string) % width!= width - 1:
            string += char

        # Else, we need to flush the current string to the output buffer
        else:
            # Get the corresponding value in the dictionary
            value = dictionary[string]

            # Flush the value as one or more bytes depending on its length
            while value >= 256:
                blocks[-1] += bytes([(value & 0xff)])
                value >>= 8

            # Add the last byte
            blocks[-1] += bytes([value])

            # Reset the string
            string = ''

    # Flush any remaining characters
        while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
    
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)