#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved = rng.read(1)
    lzw_min_code_size = rng.read(1)[0]
    id = struct.pack('>HHBBBBB', left, top, width, height, local_color_table_flag, interlace_flag, sort_flag, lzw_min_code_size)

    # 6. Local Color Table (optional)
    if local_color_table_flag:
        num_colors = (rng.read(1)[0] % 256) or 1
        color_table += rng.read(3 * num_colors)  # Random colors

    # 7. Image data
    image_data = b''
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x2C:
            # Image descriptor
            pass
        elif block_type == 0x21:
            # Extension
            ext_label = rng.read(1)[0]

            if ext_label == 0xFF:
                # Application extension
                app_ext_block_size = rng.read(1)[0]

                app_identifier = rng.read(8).decode()
                auth_code = rng.read(3).decode()

                app_data = rng.read(app_ext_block_size - 11)

                # TODO: Handle Netscape application extensions properly
                if app_identifier in ('NETSCAPE', 'ANIMEXTS'):
                    print('Netscape/Animextension found!')

                    sub_block_id = app_data[0]
                    loop_count = struct.unpack('>H', app_data[1:])[0]

                    if sub_block_id == 1:
                        print(f'\tLoop count: {loop_count}')

                        # We don't need to do anything with this information; we just want to read it so that our parser doesn't crash on it
                        continue
                    else:
                        raise Exception('Unknown netscape extension sub-block ID.')
                else:
                    print(f'Application extension "{app_identifier}" found! Ignoring...')
                    continue
            elif ext_label == 0xFE:
                # Comment extension
                comment_ext_block_size = rng.read(1)[0]

                comment = rng.read(comment_ext_block_size)

                print('Comment extension found!')
                print('\tComment:', comment)

                # We don't need to do anything with this information; we just want to read it so that our parser doesn't crash on it
                continue
            elif ext_label == 0x01:
                # Plain text extension
                plain_text_ext_block_size = rng.read(1)[0]

                text_left, text_top = struct.unpack('>HH', rng.read(4))
                grid_width, grid_height = struct.unpack('>BB', rng.read(2))
                cell_width, cell_height = struct.unpack('>BB', rng.read(2))
                foreground_color_index = rng.read(1)[0]
                background_color_index = rng.read(1)[0]
                text_flags = rng.read(1)[0]
                transparent_color_index = rng.read(1)[0]

                text = []
                while len(text) < plain_text_ext_block_size - 19:
                    char_size = rng.read(1)[0]
                    text += [rng.read(char_size)]

                print('Plain Text extension found!')
                print('\tText Left:', text_left)
                print('\tText Top:', text_top)
                print('\tGrid Width:', grid_width)
                print('\tGrid Height:', grid_height)
                print('\tCell Width:', cell_width)
                print('\tCell Height:', cell_height)
                print('\tForeground Color Index:', foreground_color_index)
                print('\tBackground Color Index:', background_color_index)
                print('\tText Flags:', bin(text_flags))
                print('\tTransparent Color Index:', transparent_color_index)
                print('\tText:', ''.join([chr(c) for c in text]))

                # We don't need to do anything with this information; we just want to read it so that our parser doesn't crash on it
                continue
            else:
                raise Exception('Unknown extension label.')
        elif block_type == 0x3B:
            break
        else:
            raise Exception('Unknown block type.')

        # Block terminator
        block_terminator = rng.read(1)[0]
        assert block_terminator == 0x00

    return 0

