#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = min((total_blocks_size - sum([len(b) for b in blocks])), 255)

        # If we're done with all our data, break
        if not block_size:
            break
        
        # Read this block of data from /dev/urandom
        block = rng.read(block_size)

        # Add to list of blocks
        blocks.append(block)
    
    # Compress these blocks using LZW compression
    compressed_data = compress_lzw(blocks, lzw_minimum_code_size)

    # Write everything to `out`
    out.write(header + lsd + color_table + gce + image_descriptor + compressed_data)


# Compresses an array of binary strings using LZW compression
def compress_lzw(blocks: [bytes], minimum_code_size=2):
    code_size = minimum_code_size + 1
    clear_code = 1 << minimum_code_size
    end_of_information_code = clear_code + 1
    next_available_code = end_of_information_code + 1

    codes = {chr(i).encode(): i for i in range(next_available_code)}
    current_code = ''
    output = []

    def add_to_output(byte):
        nonlocal current_code
        output.append(byte)
        current_code += chr(byte[0]).encode()

    add_to_output(struct.pack('>H', clear_code))

    for block in blocks:
        for char in block:
            if len(current_code) == code_size:
                add_to_output(codes[current_code].to_bytes(2, 'big'))

                if next_available_code < (1 << code_size):
                    codes[current_code] = next_available_code
                    next_available_code += 1
                
                current_code = chr(char).encode()

            elif char in codes:
                current_code += chr(char).encode()
            
            else:
                add_to_output(codes[current_code[:len(current_code)-1]].to_bytes(2, 'big'))
                codes[current_code] = next_available_code
                next_available_code += 1
                current_code = chr(char).encode()

    if current_code!= '':
        add_to_output(codes[current_code].to_bytes(2, 'big'))

    add_to_output(end_of_information_code.to_bytes(2, 'big'))

    return b''.join(output)