#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = min((rng.read(1)[0] % 256) or 1, 256)
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break

        # The first byte in each block is either a literal value or an index to look up in the table
        # If the bit is set, then this is a literal value; otherwise, it's an index
        literals = [rng.read(1)[0] for _ in range(block_size)]
        indices = [(i & ~0x80) for i in literals]
        literals = [i & 0x7F for i in literals]

        # Build the lookup table from the literals
        lookup_table = {v: k for k, v in enumerate(literals)}

        # Add new values to the lookup table as needed
        for i in range(len(indices)):
            while len(lookup_table) < indices[i]:
                lookup_table[rng.read(1)[0]] = len(lookup_table)

            indices[i] = lookup_table[indices[i]]

        # Convert the indices back to bytes
        data = bytes(indices)

        # Append the block to our list
        blocks.append(bytes([block_size]) + data)

    # Finish off by adding a final zero-length block
    blocks.append(b'\x00')

    # Concatenate them all together
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)