#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF will be of size between `min_width`, `max_width` x `min_height`, `max_height`.
# It will contain up to `num_frames` frames with duration between `min_frame_duration`, `max_frame_duration`.
# Each frame will use up to `max_sub_blocks` sub-blocks, where a sub-block is a sequence of bytes that starts by its length.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, *,
        min_width=1, max_width=1000, min_height=1, max_height=1000,
        num_frames=10, min_frame_duration=10, max_frame_duration=1000,
        max_sub_blocks=10):

    def randint(lo, hi): return lo + int((hi - lo) * rng.read(1)[0] / 255)

    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width = randint(min_width, max_width)
    height = randint(min_height, max_height)

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = [randint(0, 1) for _ in range(3)]
    background_color_index = randint(0, 255)
    pixel_aspect_ratio = randint(0, 255)
    lsd = struct.pack('>HHBHB', width, height, packed_fields[0], packed_fields[1], packed_fields[2]) \
          + struct.pack('>BBB', background_color_index, pixel_aspect_ratio, 0)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = randint(1, 256)
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(randint(0, 1))
    if not has_gce:
        gce = None
    else:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04,
                          randint(min_frame_duration, max_frame_duration),
                          randint(0, 1))

    # 5. Image Descriptors
    descriptors = []
    for i in range(num_frames):
        left = randint(0, width - 1)
        top = randint(0, height - 1)

        width = randint(1, width - left)
        height = randint(1, height - top)

        local_color_table_flag = bool(randint(0, 1))
        interlace_flag = bool(randint(0, 1))
        sort_flag = bool(randint(0, 1))
        reserved = randint(0, 7)

        local_color_table_size = randint(1, 256) if local_color_table_flag else 0
        local_color_table = rng.read(3 * local_color_table_size) if local_color_table_flag else b''

        image_descriptor = struct.pack('>BHHHHBBBBB',
                                       0x2C,
                                       left, top, width, height,
                                       local_color_table_size,
                                       local_color_table_flag,
                                       interlace_flag,
                                       sort_flag,
                                       reserved) + local_color_table

        descriptors.append(image_descriptor)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = randint(2, 8)

    blocks = []
    for descriptor in descriptors:
        total_blocks_size = randint(1, max_sub_blocks)  # Make it reasonably small
        while total_blocks_size > 0:
            block_size = min(total_blocks_size, randint(1, 255))
            blocks.append(rng.read(block_size))