#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break

        # If this is an "uncompressed" data block, just read that many bytes
        if block_size == 128:
            uncompressed_data_length = rng.read(1)[0]

            blocks.append(struct.pack('>BB', 128, uncompressed_data_length) + rng.read(uncompressed_data_length))
        elif block_size > 128:
            # This is a compressed data block, let's read the next byte as the first code in the dictionary
            initial_code = rng.read(1)[0]
            
            codes = [initial_code]

            # Now read all the rest of the codes
            while len(codes) < 4096:
                # Read the next byte as the prefix
                prefix = codes[rng.read(1)[0]]

                # Read the next byte as the suffix
                suffix = rng.read(1)[0]
                
                # Add the new code to our list
                codes.append((prefix << 8) | suffix)
                
            # Now build the actual data from these codes
            data = bytes()
            current_code = initial_code

            while current_code!= 128:
                data += struct.pack('<H', current_code)
                current_code = codes[current_code]

            # Finally add the final block
            blocks.append(struct.pack('>BB', block_size, 0) + data)
        else:
            raise ValueError("Invalid block size")
            
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)