#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]

        if block_size == 0:
            # End of data
            break

        elif block_size >= 128:
            # Raw data
            blocks.append(struct.pack('>BB', 0xFE, block_size - 128) + rng.read(block_size - 128))
            
        else:
            # Compressed data
            codes = [rng.read(1)[0]]

            while len(codes) < block_size:
                code = codes[-1]

                if code in range(1, 128):
                    # Copy previous byte
                    codes.append(code)
                    
                elif code in range(129, 256):
                    # Repeat previous byte N times
                    codes += ([code - 129] * (code - 128))
                    
                elif code == 257:
                    # Escape code
                    codes.append(rng.read(1)[0])
                    
            # Convert to bytes
            codes = [bytes([c]) for c in codes]
            
            # Add to blocks
            blocks.append(b'\xFF\x00' + b''.join(codes))
            
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)