#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]

        if block_size == 0:
            # If this is the last block in the sequence, add an empty block to mark its end
            blocks.append(b'\x00')

            # Add the minimum code size and the compressed data
            image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)
            
            # Add the ending character
            image_data += b'\x00'

            break
            
        elif block_size & 0x80:
            # This is a clear code block
            blocks.append(struct.pack('>B', block_size))

        else:
            # This is a regular code block
            length = block_size & 0x0F
            offset = ((block_size >> 4) << 8) | rng.read(1)[0]

            # The first byte contains the length of the prefix that follows
            # So we need to read another byte from the stream to know how many bytes are actually part of the prefix
            prefix_length = rng.read(1)[0]

            # Now we can append the actual prefix to our list of blocks
            blocks.append(
                struct.pack('>BH',
                    length + 1 + prefix_length,
                    0xFF - offset
                ) + rng.read(prefix_length)
            )

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)