#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]

        if block_size == 0:
            # End of data
            break

        elif block_size > 128:
            # Raw data
            blocks.append(struct.pack('>B', block_size - 128) + rng.read(block_size - 128))
            
        else:
            # LZW compressed data
            code_length = min((block_size + 1), 128)

            codes = [i for i in range(code_length)]
            
            # Generate the first code
            next_code = len(codes)

            # Read the rest of the codes from the stream
            while next_code < 4096:
                length = rng.read(1)[0]

                if length == 128:
                    # EOD
                    break
                
                elif length <= code_length:
                    # Copy previous code
                    codes.append(rng.read(1)[0])

                    # Add new codes
                    for _ in range(length):
                        codes.append(next_code)

                        next_code += 1
                        
                else:
                    # Repeat previous code
                    repeat_count = rng.read(1)[0] & ((1 << (length - code_length)) - 1)
                    
                    for _ in range(repeat_count + 1):
                        codes.append(rng.read(1)[0])
                    
                next_code += 1
                
            # Build the dictionary
            dict_size = 2 ** code_length
            dictionary = {c: bytes([c]) for c in codes}

            # Decode the data
            decoded = b''
            prev_code = None
            while True:
                code = rng.read(1)[0]

                if code >= dict_size:
                    raise Exception('Invalid code')

                if code == dict_size:
                    # EOD
                    break

                if prev_code is not None:
                    decoded += dictionary[prev_code]

                if code!= dict_size - 1:
                    dictionary[dict_size] = dictionary[code] + dictionary[code][0:1]

                prev_code = code

            blocks.append(decoded)
            
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)