#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool((packed_fields & 0b10000000) >> 7)
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool((packed_fields & 0b00000001) << 7)
    interlace_flag = bool((packed_fields & 0b00000100) >> 2)
    sort_flag = bool((packed_fields & 0b00001000) >> 3)
    reserved_flags = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = rng.read(3 * local_color_table_size) if local_color_table_flag else b''
    image_descriptor = struct.pack('>BHHHHBB', 0x2C, left, top, width, height,
                                   local_color_table_size, reserved_flags)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        if len(blocks) >= total_blocks_size:
            break

        block_size = rng.read(1)[0]
        if not block_size:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)