#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]

        if block_size == 0:
            break
            
        elif block_size >= 128:
            # Raw data block
            blocks.append(struct.pack('>BB', block_size - 128, rng.read(1)[0]))

        else:
            # LZW compressed data block
            code_length = min((block_size + 1), 128)

            codes = [i for i in range(2 ** code_length)]
            next_code = len(codes)
            
            clear_code = 1 << code_length
            end_of_information_code = clear_code + 1

            # Initialize the dictionary with the first two codes
            dictionary = {
                clear_code: [],
                end_of_information_code: []
            }

            # The last code is always an empty string
            dictionary[next_code] = ''

            # Read the rest of the codes from the stream
            while next_code < (clear_code + 2):

                # Get the current code length
                code_length = rng.read(1)[0] & 0xFF

                # If this is the last code, then set its value to the EOI code
                if code_length == 0:
                    code = end_of_information_code
                
                else:
                    # Otherwise, read the actual code
                    code = int.from_bytes(rng.read(code_length), 'big')

                    # Check if this code is already defined
                    if code > max(dictionary.keys()):
                        raise Exception('Invalid code encountered!')
                    
                    # If so, look up its value
                    code = dictionary[code]
                    
                # Add the new code to our dictionary
                dictionary[next_code] = dictionary[code] + chr(rng.read(1)[0]).encode()

                # Increment the next code counter
                next_code += 1
                
            # Now that we've finished reading the codes, let's write them back to the output
            for i in range(len(codes)):
                out.write(dictionary[i])

            # Finally, add the ending zero-length block
            blocks.append(b'\x00')
            
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)