#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    # Random background color index
    background_color_index = rng.read(1)[0]
    # Random pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)[0]
    # Color table flag
    color_table_flag = rng.read(1)[0] & 0b10000000
    # Global color table size
    global_color_table_size = rng.read(1)[0] & 0b00000111
    # Sort flag
    sort_flag = rng.read(1)[0] & 0b00001000
    # Size of global color table
    global_color_table_size += 1 << ((global_color_table_size >> 7) * 3)
    # Background color index
    background_color_index %= global_color_table_size
    # Pixel aspect ratio
    pixel_aspect_ratio &= 0b00000111
    if pixel_aspect_ratio == 0:
        pixel_aspect_ratio = 0
    elif pixel_aspect_ratio <= 3:
        pixel_aspect_ratio = 1 / pow(2, pixel_aspect_ratio)
    else:
        pixel_aspect_ratio = 1 / pow(3, pixel_aspect_ratio - 3)

    # 3. Color Table
    # If there is no color table, skip this step
    if not color_table_flag:
        color_table = b''
    else:
        # Read the color table
        color_table = rng.read(global_color_table_size * 3)

    # 4. Graphic Control Extension
    # Block size
    gce_block_size = 0x04
    # Packed fields
    gce_packed_fields = 0x00
    # Delay time
    delay_time = rng.read(2)
    # Transparent color index
    transparent_color_index = rng.read(1)[0]
    # Terminator
    terminator = 0x00

    gce = struct.pack('>BBBBB', gce_block_size, gce_packed_fields, delay_time[0], delay_time[1], transparent_color_index)

    # 5. Image Descriptor
    # Separator
    separator = 0x2C
    # Left position
    left_position = rng.read(2)
    # Top position
    top_position = rng.read(2)
    # Width
    image_width = width
    # Height
    image_height = height
    # Local color table flag
    local_color_table_flag = rng.read(1)[0] & 0b10000000
    # Interlace flag
    interlace_flag = rng.read(1)[0] & 0b01000000
    # Sort flag
    sort_flag = rng.read(1)[0] & 0b00100000
    # Reserved
    reserved = rng.read(1)[0] & 0b00011111
    # Size of local color table
    local_color_table_size = rng.read(1)[0] & 0b00000111
    # Sort flag
    sort_flag = rng.read(1)[0] & 0b00001000
    # Size of local color table
    local_color_table_size += 1 << ((local_color_table_size >> 7) * 3)

    # 6. Image Data
    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0]
    total_blocks_size = width * height
    blocks = []
    while len(blocks) < total_blocks_size:
        block_size = min(total_blocks_size - len(blocks), rng.read(1)[0])
        blocks.append(rng.read(block_size))

    image_data = struct.pack('>BB', lzw_minimum_code_size, 0) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
