#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_gif(out: BinaryIO, rng: BinaryIO):
    # Magic header
    out.write('GIF89a'.encode())
    
    # Logical screen descriptor
    width = 1024
    height = 768
    color_table_size = 7
    global_color_table_flag = 1
    sort_flag = 0
    color_resolution = 7
    background_color_index = 0
    pixel_aspect_ratio = 0
    out.write(struct.pack("<HHBBB", width, height, color_table_size, global_color_table_flag | (sort_flag << 1), color_resolution, background_color_index))
    out.write(pixel_aspect_ratio.to_bytes(1, byteorder='little'))
    
    # Global color table
    for i in range(2 ** color_table_size):
        out.write((i * 256 // (2 ** color_table_size)).to_bytes(1, byteorder='little'))
        out.write((i * 256 // (2 ** color_table_size)).to_bytes(1, byteorder='little'))
        out.write((i * 256 // (2 ** color_table_size)).to_bytes(1, byteorder='little'))
    
    # Application extension
    block_type = 0xFF
    application_identifier = 'GIF'
    authentication_code = 'PYTHON'
    data = b''
    while len(data) < 255 - 11:
        data += rng.read(1)
    block_size = len(data).to_bytes(1, byteorder='little')
    out.write(block_type.to_bytes(1, byteorder='little'))
    out.write(block_size)
    out.write(application_identifier.encode() + authentication_code.encode())
    out.write(data)
    
    # Image descriptor
    image_separator = 0x2C
    left = 0
    top = 0
    width = 1024
    height = 768
    local_color_table_flag = 0
    interlace_flag = 0
    sort_flag = 0
    reserved = 0
    out.write(image_separator.to_bytes(1, byteorder='little'))
    out.write(struct.pack("<HHHBB", left, top, width, height, local_color_table_flag | (interlace_flag << 1) | (sort_flag << 2) | (reserved << 3)))
    
    # Local color table
    for i in range(2 ** color_table_size):
        out.write((i * 256 // (2 ** color_table_size)).to_bytes(1, byteorder='little'))
        out.write((i * 256 // (2 ** color_table_size)).to_bytes(1, byteorder='little'))
        out.write((i * 256 // (2 ** color_table_size)).to_bytes(1, byteorder='little'))
    
    # Raster data
    lzw_minimum_code_size = 8
    out.write(lzw_minimum_code_size.to_bytes(1, byteorder='little'))
    
    # LZW compressed data
    code_size = lzw_minimum_code_size + 1
    clear_code = 1 << code_size
    end_of_information_code = clear_code + 1
    dictionary_size = 1 << code_size
    dictionary = {}
    for i in range(dictionary_size):
        dictionary[i] = [i]
    current_code = None
    for _ in range(width * height):
        if current_code is None or len(current_code) == 1:
            current_code = next(rng).to_bytes(1, byteorder='little')
        else:
            new_code = next(rng).to_bytes(1, byteorder='little')
            if new_code not in dictionary:
                dictionary[new_code] = dictionary[current_code