#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])
    
    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        return
    
    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce_present = bool(rng.read(1)[0] & 0x1)
    if gce_present:
        block_size = rng.read(1)[0]
        assert block_size == 0xF9
        
        flags = rng.read(1)[0]
        delay = struct.unpack('<H', rng.read(2))[0]
        transparent_index = rng.read(1)[0]
        terminator = rng.read(1)[0]
        assert terminator == 0
        
        gce = struct.pack('>BHHB', block_size, flags, delay, transparent_index)
    else:
        gce = None
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHH', 0x2C, 0, 0, width, height)
    
    # 6. Local Color Table (if present)
    local_color_table_flag = bool(rng.read(1)[0] & 0x80)
    if not local_color_table_flag:
        return
    
    num_colors = 2 ** ((rng.read(1)[0] >> 4) + 1)
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 7. LZW Minimum Code Size
    minimum_code_size = rng.read(1)[0]
    
    # 8. Data Sub-Blocks
    data = []
    while True:
        size = rng.read(1)[0]
        if size == 0:
            break
        data.append(rng.read(size))
    
    # 9. Trailer
    trailer = b'\x3b'
    
    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)
    out.write(color_table)
    out.write(minimum_code_size)
    for d in data:
        out.write(struct.