#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor (LSD)
    width = int.from_bytes(rng.read(2), byteorder='little')
    height = int.from_bytes(rng.read(2), byteorder='little')
    packed = int.from_bytes(rng.read(1), byteorder='big')
    background_color_index = packed & 0xFF
    pixel_aspect_ratio = packed >> 4
    lsd = struct.pack('<HHBBB', width, height, packed, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    color_table_flag = rng.read(1)[0] != 0
    if not color_table_flag:
        raise Exception("No color table")
    
    color_resolution = rng.read(1)[0]
    sort_flag = rng.read(1)[0] != 0
    global_color_table_size = 1 << ((color_resolution + 1) * 3)
    color_table = rng.read(global_color_table_size)
    
    # 4. Graphic Control Extension (GCE)
    gce = struct.pack('<BBHBB', 0xF9, 4, 0, 0, 0)
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('<BBHHBB', 0x2C, 0, 0, width, height, 0)
    
    # 6. Image Data
    data = []
    while True:
        block_size = rng.read(1)[0]
        
        if block_size == 0:
            break
        
        data.append(rng.read(block_size))
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as f:
        with open('test.gif', 'wb') as o:
            generate_random_gif(f, o)