#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('<HH', rng.read(4))

    # Background color index (one byte)
    background_color_index = rng.read(1)[0]

    # Pixel aspect ratio (one byte)
    pixel_aspect_ratio = rng.read(1)[0]

    # 3. Global Color Table
    # We don't use one here
    global_color_table_flag = False
    if global_color_table_flag:
        # Number of colors in table (three bits)
        num_colors = 1 << ((global_color_table_size - 1) * 3)

        # Size of each color (three bits)
        size_of_each_color = 3

        # Sort flag (one bit)
        sort_flag = True

        # Colors (three bytes per color)
        global_color_table = [
            rng.read(3) for _ in range(num_colors)
        ]

    # 4. Image Data
    # We only have one image
    image_descriptor = b'\x2c\x00\x00\x00\x00\x00'

    # Local Color Table Flag
    local_color_table_flag = False
    if local_color_table_flag:
        # Number of colors in table (three bits)
        num_colors = 1 << ((local_color_table_size - 1) * 3)

        # Size of each color (three bits)
        size_of_each_color = 3

        # Sort flag (one bit)
        sort_flag = True

        # Colors (three bytes per color)
        local_color_table = [
            rng.read(3) for _ in range(num_colors)
        ]

    # Interlaced?
    interlaced = False

    # LZW Minimum Code Size
    lzw_minimum_code_size = 8

    # Image Data
    image_data = []
    while True:
        # Read a byte
        byte = rng.read(1)

        # If we hit EOF, stop
        if not byte:
            break

        # Otherwise, add it to our list
        image_data.append(byte[0])

    # Write everything to the output file
    out.write(header +
              struct.pack('<HBB', width, height,
                          background_color_index) +
              pixel_aspect_ratio.to_bytes(1, 'little') +
              (b'\x00' if not global_color_table_flag else
               struct.pack('B', num_colors) +
               struct.pack('B', size_of_each_color) +
               struct.pack('B', int(sort_flag)) +
               b''.join(global_color_table)) +
              image_descriptor +
              (b'\x00' if not local_color_table_flag else
               struct.pack('B', num_colors) +
               struct.pack('B', size_of_each_color) +
               struct.pack('B', int(sort_flag)) +
               b''.join(local_color_table)) +
              struct.pack('B', int(interlaced)) +
              struct.pack('B', lzw_minimum_code_size) +
              b''.join(image_data))
