#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = None
    if rng.read(1)[0] & 1 == 1:
        # Random delay time (two bytes)
        delay = rng.read(2)

        # Transparent color index (one byte)
        transparent_color_index = rng.read(1)

        # Disposal method (one byte)
        disposal_method = rng.read(1)

        # Block terminator (one byte)
        block_terminator = rng.read(1)

        gce = struct.pack('BBBBB', 0xF9, 0x04, delay[0], delay[1], transparent_color_index[0]) + \
            disposal_method + block_terminator

    # 5. Image Descriptor
    id = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, 0, 0)

    # 6. Local Color Table (optional)
    lct = None
    if rng.read(1)[0] & 1 == 1:
        # Random number of colors in local color table (one byte)
        num_lct_colors = rng.read(1)[0]

        # Random local color table data (three bytes per color)
        lct = rng.read(3 * num_lct_colors)

    # 7. Image Data
    image_data = rng.read(width * height)

    # 8. Trailer
    trailer = b'\x3b'

    # Write everything to the output file
    out.write(header + lsd + color_table + (gce or b'') + id + (lct or b'') + image_data + trailer)

