#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])
    
    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        return
    
    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Image Data
    image_data = []
    while True:
        block_type = rng.read(1)[0]
        
        if block_type == 0x2c:  # Graphic Control Extension
            block_size = rng.read(1)[0]
            
            assert block_size == 0x04
            flags = rng.read(1)[0]
            delay = struct.unpack('<H', rng.read(2))[0]
            transparent_index = rng.read(1)[0]
            terminator = rng.read(1)[0]
            
            assert terminator == 0
            
            image_data.append(struct.pack('B', block_type))
            image_data.append(block_size)
            image_data.append(flags)
            image_data.append(delay & 0xff)
            image_data.append((delay >> 8) & 0xff)
            image_data.append(transparent_index)
            image_data.append(terminator)
            
        elif block_type == 0x21:  # Application Extension
            block_size = rng.read(1)[0]
            
            assert block_size == 0x0b
            application_identifier = rng.read(8)
            application_authentication_code = rng.read(3)
            
            data_sub_blocks = []
            while True:
                sub_block_size = rng.read(1)[0]
                
                if sub_block_size == 0:
                    break
                
                data_sub_blocks.extend([sub_block_size])
                data_sub_blocks.extend(rng.read(sub_block_size))
            
            image_data.append(struct.pack('B', block_type))
            image_data.append(block_size)
            image_data.append(application_identifier)
            image_data.append(application_authentication_code)
            for sub_block in data_sub_blocks:
                image_data.append(sub_block)
            
        elif block_type == 0x3b:  # Trailer
            break
        
        else:
            raise Exception("Unknown block type")
    
    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    for chunk in image_data:
