#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor (LSD)
    width = rng.read(2)
    height = rng.read(2)
    packed_fields = rng.read(1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)

    lsd = width + height + packed_fields + background_color_index + pixel_aspect_ratio

    # 3. Global Color Table
    global_color_table_flag = rng.read(1)
    color_resolution = rng.read(1)
    sort_flag = rng.read(1)
    global_color_table_size = rng.read(1)

    # We don't need to write the actual table as it will be generated later
    color_table = global_color_table_flag + color_resolution + sort_flag + global_color_table_size

    # 4. Graphic Control Extension (GCE)
    extension_introducer = b'\x21'
    graphic_control_label = b'\xf9'
    block_size = rng.read(1)
    disposal_method = rng.read(1)
    user_input_flag = rng.read(1)
    transparent_color_flag = rng.read(1)
    delay_time = rng.read(2)
    transparent_color_index = rng.read(1)
    terminator = b'\x00'

    gce = extension_introducer + graphic_control_label + block_size + disposal_method + \
          user_input_flag + transparent_color_flag + delay_time + transparent_color_index + terminator

    # 5. Image Descriptor
    image_separator = b'\x2C'
    left_position = rng.read(2)
    top_position = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    local_color_table_flag = rng.read(1)
    interlace_flag = rng.read(1)
    sort_flag = rng.read(1)
    reserved_field = rng.read(1)

    image_descriptor = image_separator + left_position + top_position + width + height + \
                       local_color_table_flag + interlace_flag + sort_flag + reserved_field

    # 6. Image Data
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]

        # If we got an empty block, stop generating data
        if not block_size:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)