#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = rng.read(num_colors * 3)

    # 4. Image Data
    # We will use a single image with no local color table
    # We will use a single graphic control extension block
    # We will use a single plain text extension block
    # We will use a single application extension block
    # We will use a single comment extension block

    # Graphic Control Extension Block
    gceb = struct.pack('BBBB', 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # Plain Text Extension Block
    pteb = struct.pack('BBBBB', 0x01, 0x0C, 0x00, 0x00, rng.read(1)[0])

    # Application Extension Block
    aeb = struct.pack('BBBBBB', 0xFF, 0xFE, 0x0B, 0x4E, 0x45, 0x54) + rng.read(7)

    # Comment Extension Block
    ce = struct.pack('BBBB', 0xFE, 0x01, 0x00, rng.read(1)[0])

    # 5. Trailer
    trailer = b'\x3B'

    # Write everything to the output stream
    out.write(header + lsd + global_color_table + gceb + pteb + aeb + ce + trailer)

