#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = b''
    if rng.read(1)[0] & 1:  # Randomly decide whether to include this extension
        # Size of extension block (always 4)
        gce += b'\x04'

        # Packed fields (using a random byte for variety)
        gce += struct.pack('BB', rng.read(1)[0], rng.read(1)[0])

        # Delay time (two bytes, random)
        gce += struct.pack('>H', rng.read(2)[0])

        # Transparent color index (one byte, random)
        gce += rng.read(1)

    # 5. Image Descriptor
    id = b'\x2c\x00\x00\x00\x00'  # Fixed values

    # 6. Local Color Table (optional)
    lct = b''
    if rng.read(1)[0] & 2:  # Randomly decide whether to include this table
        # Number of colors in table (random between 1 and 256)
        num_lct_entries = rng.read(1)[0] % 256 or 1
        lct += struct.pack('B', num_lct_entries)

        # Actual entries (fully random)
        lct += rng.read(3 * num_lct_entries)

    # 7. Image Data
    image_data = rng.read(width * height)  # Fully random

    # Write everything together
    out.write(header + lsd + color_table + gce + id + lct + image_data)
