#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b"GIF89a"
    
    # 2. Logical Screen Descriptor (Logical Screen Size and Color Table Information)
    width = int.from_bytes(rng.read(2), byteorder='big') % 65536
    height = int.from_bytes(rng.read(2), byteorder='big') % 65536
    flags = rng.read(1)[0] & 0b111
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('<HHBBBB', width, height, flags, background_color_index, pixel_aspect_ratio, 0)
    
    # 3. Global Color Table
    color_table = rng.read(3 * (2 ** ((flags >> 4) + 1)))
    
    # 4. Graphic Control Extension
    gce = b"\x21\xf9\x04\x01\x00\x00\xff\xff\xff\xff\x00"
    
    # 5. Image Descriptor
    left = int.from_bytes(rng.read(2), byteorder='big') % 65536
    top = int.from_bytes(rng.read(2), byteorder='big') % 65536
    width = int.from_bytes(rng.read(2), byteorder='big') % 65536
    height = int.from_bytes(rng.read(2), byteorder='big') % 65536
    flags = rng.read(1)[0] & 0b111
    local_color_table_length = rng.read(1)[0]
    image_descriptor = struct.pack("<HHHHBB", left, top, width, height, flags, local_color_table_length)
    
    # 6. Image Data
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            break
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)