#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # Write everything to output
    out.write(header + lsd + color_table)

    # 4. Image Data
    while True:
        # Read next block type
        block_type = rng.read(1)[0]

        if block_type == 0x3b:  # Trailer
            break

        elif block_type == 0x2c:  # Extension Block
            # Skip over extension block sub-blocks until we get to an image descriptor
            while True:
                sub_block_len = rng.read(1)[0]
                if sub_block_len == 0:
                    break
                else:
                    rng.read(sub_block_len)

            # Now read the image descriptor
            idesc = rng.read(9)

            # Write ID to output
            out.write(bytes([block_type]))
            out.write(idesc)

            # Read local color table length
            lct_size = rng.read(1)[0] & 7

            # If there is a local color table, read it in
            if lct_size > 0:
                num_lct_entries = 2 ** (lct_size + 1)
                lct = rng.read(3 * num_lct_entries)
                out.write(lct)

            # Read image data
            image_data = rng.read(width * height)
            out.write(image_data)

            # Write terminator to output
            out.write(b'\x00')

        else:
            raise ValueError("Unknown block type {}".format(block_type))
