#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        out.write(header + lsd)
        return

    num_colors = 2 ** ((packed >> 4) + 1)
    global_color_table = [struct.unpack('BBB', rng.read(3)) for _ in range(num_colors)]
    global_color_table = bytes([
            color[0] | (color[1] << 8) | (color[2] << 16)
                for color in global_color_table
        ])

    # 4. Image Data
    image_data = []
    while True:
        block_type = rng.read(1)[0]
        if block_type == 0x2c:
            break

        block_size = rng.read(1)[0]
        image_data.append((block_type, block_size, rng.read(block_size)))

    # 5. Trailer
    trailer = rng.read()

    # Write everything to `out`
    out.write(header + lsd + global_color_table)
    for block_type, block_size, data in image_data:
        out.write(bytes([block_type, block_size]))
        out.write(data)
    out.write(trailer)

