#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])
    
    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        return
    
    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce_flag = bool(rng.read(1)[0] & 0xF9)
    if gce_flag:
        delay = rng.read(2)
        transparent_index = rng.read(1)
        block_terminator = rng.read(1)
        
    # 5. Image Descriptor
    image_descriptor = b'\x2C\x00\x00\x00\x00'
    
    # 6. Local Color Table (optional)
    local_color_table_flag = bool(rng.read(1)[0] & 0x80)
    if local_color_table_flag:
        num_colors = 2 ** ((rng.read(1)[0] >> 4) + 1)
        color_table = rng.read(3 * num_colors)  # Random colors
    
    # 7. LZW Minimum Code Size
    minimum_code_size = rng.read(1)
    
    # 8. Pixel Data Blocks
    pixel_data = rng.read()
    
    # 9. Trailer
    trailer = b';'
    
    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce_flag:
        out.write(delay)
        out.write(transparent_index)
        out.write(block_terminator)
    out.write(image_descriptor)
    if local_color_table_flag:
        out.write(color_table)
    out.write(minimum_code_size)
    out.write(pixel_data)
    out.write(trailer)
