#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = [struct.pack('>BBBB', *rng.read(4)) for _ in range(num_colors)]
    color_table = b''.join(global_color_table)

    # 4. Graphic Control Extension
    # We don't need any fancy stuff here, so let's just use the defaults
    gce = b'\x21\xf9\x04\x00\x00\xff\x00\x00\x00'

    # 5. Image Descriptor
    # Same as above, we don't need anything fancy
    image_descriptor = b'\x2c\x00\x00\x00\x00\x00\x00\x00\x00'

    # 6. Image Data
    # Let's make this one interesting by using a random minimum code size
    lzw_minimum_code_size = rng.read(1)[0] & 0b00111111
    # Now let's read some random data until we get a block that is smaller than 255 bytes
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if block_size < 255:
            break
        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)