#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphic Control Extension
    # We don't need any extensions, so we just write an empty one
    gce = struct.pack('>BBBB', 0xF9, 4, 0, 0)

    # 5. Image Descriptor
    # We only use one image, so we can hard-code this
    image_descriptor = b'\x2C\x00\x00\x00\x00\x00'

    # 6. Image Data
    # We will use LZW compression, so let's do that
    # First, get the minimum code size from the first byte
    lzw_minimum_code_size = rng.read(1)[0]

    # Now, read the rest of the data as blocks
    blocks = []
    while True:
        # Read the next block size
        block_size = rng.read(1)[0]
        if not block_size:
            break  # Reached the end of the stream

        # Read the block itself
        blocks.append(rng.read(block_size))

    # Join all the blocks together, prefixed by the minimum code size
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)