#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphic Control Extension (always present)
    # Always 4 bytes long, always with disposal method 2 (restore to background color), no user input flag, transparent color index -1
    gce = struct.pack('>BBBB', 4, 4, 2, 0, 255)

    # 5. Image Descriptor
    # Random left/top position, using two bytes for each, and random width/height as above
    image_descriptor = struct.pack('>HHHH', rng.read(2)[0], rng.read(2)[0], width, height)

    # 6. Image Data
    # We'll use LZW compression
    lzw_minimum_code_size = rng.read(1)[0] & 0xF

    # Read in blocks of data until we run out of data
    blocks = []
    while True:
        block = rng.read(255)
        blocks.append(struct.pack('>B', len(block)))
        blocks.append(block)

        # Stop when we've read enough
        total_blocks_size = sum([len(b) for b in blocks])
        if total_blocks_size <= 0:
            break

    image_data = struct.pack('>BB', lzw_minimum_code_size, len(blocks)) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)

