#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following structure:
#
# 1. Header
# 2. Logical Screen Descriptor
# 3. Global Color Table
# 4. Graphic Control Extension
# 5. Image Descriptor
# 6. Image Data
# 7. Trailer
def generate_gif(out: BinaryIO, rng):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed = rng.read(1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)

    lsd = struct.pack('<HHBBB', *struct.unpack('>HHBBB', width + height + packed + background_color_index + pixel_aspect_ratio))

    # 3. Global Color Table
    global_color_table_flag = rng.read(1)[0] >> 7
    color_resolution = rng.read(1)[0] >> 4
    sort_flag = rng.read(1)[0] >> 3
    size_of_global_color_table = 2 ** ((global_color_table_flag << 3) | color_resolution)

    color_table = rng.read(3 * size_of_global_color_table)

    # 4. Graphic Control Extension
    extension_introducer = b'\x21'
    graphic_control_label = b'\xf9'
    block_size = b'\x04'
    disposal_method = rng.read(1)
    user_input_flag = rng.read(1)
    transparent_color_flag = rng.read(1)
    delay_time = rng.read(2)
    transparent_color_index = rng.read(1)
    terminator = b'\x00'

    gce = extension_introducer + graphic_control_label + block_size + disposal_method + user_input_flag + transparent_color_flag + delay_time + transparent_color_index + terminator

    # 5. Image Descriptor
    image_separator = b'\x2c'
    left_position = rng.read(2)
    top_position = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    local_color_table_flag = rng.read(1)[0] >> 7
    interlace_flag = rng.read(1)[0] >> 6
    sort_flag = rng.read(1)[0] >> 5
    reserved = rng.read(1)[0] >> 4
    size_of_local_color_table = 2 ** ((local_color_table_flag << 3) | sort_flag)

    image_descriptor = image_separator + struct.pack('<HHBBB', *struct.unpack('>HHBBB', left_position + top_position + width + height + local_color_table_flag + interlace_flag + reserved))

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    data = []
    while True:
        block_size = rng.read(1)[0] & 0xFF

        if block_size == 0:
            break

        data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)