#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])
    
    # 3. Global Color Table
    global_color_table_flag = bool(packed & 0b10000000)
    color_resolution = ((packed >> 7) & 0b111) + 1
    sort_flag = bool((packed >> 5) & 0b1)
    size_of_global_color_table = 2 ** ((packed >> 3) & 0b11)
    global_color_table_sorted = [None for _ in range(size_of_global_color_table)]
    
    if global_color_table_flag and not sort_flag:
        for i in range(size_of_global_color_table):
            r, g, b = tuple([rng.read(1)[0] for j in range(3)])
            
            global_color_table_sorted[i] = struct.pack('BBB', r, g, b)
    
    elif global_color_table_flag and sort_flag:
        raise NotImplementedError()
    
    else:
        pass
    
    color_table = b''.join(global_color_table_sorted)
    
    # 4. Graphics Control Extension
    disposal_method = rng.read(1)[0]
    user_input_flag = bool(disposal_method & 0b10000000)
    transparent_color_flag = bool(disposal_method & 0b01000000)
    delay_time = struct.unpack('>H', rng.read(2))[0]
    transparent_color_index = None
    
    if transparent_color_flag:
        transparent_color_index = rng.read(1)[0]
    
    gce = struct.pack('>BBHB', 0x21, 0xF9, 4, disposal_method)
    
    if transparent_color_flag:
        gce += struct.pack('B', transparent_color_index)
    
    gce += struct.pack('>H', delay_time)
    
    if user_input_flag:
        gce += b'\x00'
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    data = []
    while True:
        block_size = rng.read(1)[0] & 0xFF
        
        if block_size == 0:
            break
        
        data.append(rng.read(block_size))
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as f:
        with open('test.gif', 'wb') as o:
            generate_random_gif(f, o)