#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed_flags = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBB', width, height, packed_flags, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    global_color_table_flag = bool(packed_flags & 0b10000000)
    color_resolution = ((packed_flags >> 7) & 0b111) + 1
    sort_flag = bool(packed_flags & 0b1000000)
    global_color_table_size = 2 ** (((packed_flags >> 4) & 0b111) + 1)
    global_color_table = [struct.pack('>BBB', *rng.read(3)) for _ in range(global_color_table_size)]
    color_table = b''.join(global_color_table)

    # 4. Graphics Control Extension
    has_gce = rng.read(1)[0]!= 0
    if has_gce:
        delay_time = struct.unpack('>H', rng.read(2))[0]
        transparent_color_index = rng.read(1)[0]
        disposal_method = rng.read(1)[0]
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]

        # If we got an empty block, stop generating data
        if not block_size:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)