#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])
    
    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        return
    
    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    local_color_table_flag = bool(packed & 0x80)
    interlace_flag = bool(packed & 0x40)
    sort_flag = bool(packed & 0x20)
    reserved = rng.read(1)[0]
    lzw_minimum_code_size = rng.read(1)[0]
    id = struct.pack('>HHBB', left, top, width, height)

    # 6. Local Color Table (if present)
    if local_color_table_flag:
        num_colors = 2 ** ((packed >> 4) + 1)
        color_table += rng.read(3 * num_colors)  # Random colors

    # 7. LZW Data
    data = bytearray()
    while True:
        code_length = rng.read(1)[0]

        if code_length == 0xFF:
            break
        
        for i in range(code_length):
            data.append((rng.read(1)[0] % 256) ^ 0xFF)
    
    # 8. Trailer
    trailer = bytes([0x3B])

    # Write everything to output
    out.write(header)
    out.write(lsd)
    if global_color_table_flag:
        out.write(color_table)
    if len(gce) > 0:
        out.write(gce)
    out.write(id)
    if local_color_table_flag:
        out.write(color_table)
    out.write(bytes([lzw_minimum_code_size]))
    out.write(data)
    out.write(trailer)

