#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    header = b'GIF89a'

    # Logical Screen Descriptor
    width = struct.pack('<H', int.from_bytes(rng.read(2), byteorder='little'))
    height = struct.pack('<H', int.from_bytes(rng.read(2), byteorder='little'))
    flags = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    background_color_index = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    pixel_aspect_ratio = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lsd = width + height + flags + background_color_index + pixel_aspect_ratio

    # Color Table
    color_table_size = int.from_bytes(rng.read(1), byteorder='little')
    color_table = []
    for _ in range(pow(2, color_table_size+1)):
        r = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        g = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        b = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        color_table.append(r + g + b)
    color_table = b''.join(color_table)

    # Graphics Control Extension
    gce = (
        b'\x21\xf9\x04' + 
        bytes([int.from_bytes(rng.read(1), byteorder='little')]) + 
        bytes([int.from_bytes(rng.read(1), byteorder='little')]) + 
        bytes([int.from_bytes(rng.read(1), byteorder='little')])
    )

    # Image Descriptor
    left = struct.pack('<H', int.from_bytes(rng.read(2), byteorder='little'))
    top = struct.pack('<H', int.from_bytes(rng.read(2), byteorder='little'))
    width = struct.pack('<H', int.from_bytes(rng.read(2), byteorder='little'))
    height = struct.pack('<H', int.from_bytes(rng.read(2), byteorder='little'))
    flags = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    image_descriptor = left + top + width + height + flags

    # Data
    data = []
    while True:
        # Block Size
        block_size = int.from_bytes(rng.read(1), byteorder='little')

        # End of File
        if block_size == 0:
            break
        
        data.append(rng.read(block_size))
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as f:
        with open('test.gif', 'wb') as o:
            generate_random_gif(f, o)