#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    # Color table size is fixed to 7 bits per pixel
    color_table_size = 7
    # Background color index is set to 0
    background_color_index = 0
    # Pixel aspect ratio is always 0
    pixel_aspect_ratio = 0
    logical_screen_descriptor = struct.pack(
        '>HHBBBH',
        width, height,
        color_table_size,
        background_color_index,
        pixel_aspect_ratio
    )

    # 3. Global Color Table
    global_color_table = [
        0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0xff, 0xff,
        0x00, 0xff, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00,
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
    ]

    # 4. Image Descriptor
    image_left_position = 0
    image_top_position = 0
    image_width, image_height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = False
    interlace_flag = False
    sort_flag = False
    reserved_bits = 0b000000
    lct_size = 0
    image_descriptor = struct.pack(
        '>HHBBBBB',
        image_left_position, image_top_position,
        image_width, image_height,
        local_color_table_flag,
        interlace_flag,
        sort_flag,
        reserved_bits
    )

    # 5. Local Color Table
    local_color_table = []
    for i in range(lct_size):
        r, g, b = struct.unpack('BBB', rng.read(3))
        local_color_table += [r, g, b]

    # 6. Image Data
    # Number of pixels in the image
    num_pixels = int((image_width * image_height) / 8)
    # Each byte represents one row of the image
    image_data = [int(rng.read(1).hex(), base=16) for _ in range(num_pixels)]

    # Write everything to the output stream
    out.write(header + logical_screen_descriptor)
    out.write(bytes([len(global_color_table)]))
    out.write(bytes(global_color_table))
    out.write(image_descriptor)
    if local_color_table_flag:
        out.write(bytes([len(local_color_table)]))
        out.write(bytes(local_color_table))
    else:
        out.write(b'\x00\x00')
    out.write(bytes(image_data))


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)