#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical screen descriptor (LSD)
    width = int.from_bytes(rng.read(2), 'big') % 640 + 1
    height = int.from_bytes(rng.read(2), 'big') % 480 + 1
    packed_flags = int.from_bytes(rng.read(1), 'big')
    global_color_table_flag = bool(packed_flags & 0b10000000)
    background_color_index = int.from_bytes(rng.read(1), 'big') % 256
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big') % 16
    lsd = struct.pack('>HHBBB', width, height, packed_flags,
                      background_color_index, pixel_aspect_ratio)
    out.write(lsd)

    # 3. Global color table
    if global_color_table_flag:
        color_table_size = min((int.from_bytes(rng.read(1), 'big') % 256) + 1, 256) * 3
        color_table = [rng.read(3) for _ in range(color_table_size)]
        while len(color_table)!= color_table_size:
            color_table += [b'\x00\x00\x00']

        out.write(struct.pack('B', color_table_size))
        out.write(b''.join(color_table))

    # 4. Graphics control extension (GCE)
    has_gce = bool(int.from_bytes(rng.read(1), 'big'))
    if not has_gce:
        return

    transparent_color_flag = bool(int.from_bytes(rng.read(1), 'big'))
    delay_time = int.from_bytes(rng.read(2), 'big') % 65536
    transparent_color_index = int.from_bytes(rng.read(1), 'big') % 256
    packed_flags = int.from_bytes(rng.read(1), 'big')
    terminator = b'\x00'
    gce = struct.pack('>?HBB?', transparent_color_flag, delay_time,
                     transparent_color_index, packed_flags == 1) + terminator
    out.write(gce)

    # 5. Image descriptor
    left = int.from_bytes(rng.read(2), 'big') % width
    top = int.from_bytes(rng.read(2), 'big') % height
    packed_flags = int.from_bytes(rng.read(1), 'big')
    local_color_table_flag = bool(packed_flags & 0b10000000)
    interlace_flag = bool(packed_flags & 0b01000000)
    sort_flag = bool(packed_flags & 0b00100000)
    reserved_bits = int.from_bytes(rng.read(2), 'big')
    image_descriptor = struct.pack('>HHBBB', left, top,
                                   packed_flags, reserved_bits, 0)
    out.write(image_descriptor)

    # 6. Local color table
    if local_color_table_flag:
        color_table_size = min((int.from_bytes(rng.read(1), 'big') % 256) + 1, 256) * 3
        color_table = [rng.read(3) for _ in range(color_table_size)]
        while len(color_table)!= color_table_size:
            color_table += [b'\x00\x00\x00']

        out.write(struct.pack('B', color_table_size))
        out.write(b''.join(color_table))

    # 7. Image data
    lzw_minimum_code_size = int.from_bytes(rng.read(1), 'big') % 129
    lzw_maximum_code_size = max(min((int.from_bytes(rng.read(1), 'big') % 129) + 1,
                                    lzw_minimum_code_size + 1),
                                lzw_minimum_code_size)
    lzw_clear_code = 1 << lzw_minimum_code_            break
        blocks.append(rng.read(block_size))
        total_blocks_size -= block_size
        if total_blocks_size <= 0:
            break

    image_data = struct.pack('>BB', lzw_minimum_code_size, len(blocks)) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)

