#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Random background color index
    bg_color = rng.read(1)[0] % 256

    # Random pixel aspect ratio
    aspect_ratio = rng.read(1)[0] & 0b1111

    lsd = struct.pack(
        '>HHBBB',
        width,
        height,
        7 << 5 | bg_color,   # Global Color Table Flag set, 7 colors in table
        0,                  # Sort flag is off
        aspect_ratio        # Pixel Aspect Ratio
    )

    # 3. Global Color Table
    global_colors = [rng.read(3) for _ in range(7)]
    color_table = b''
    for color in global_colors:
        color_table += struct.pack('BBB', *color)

    # 4. Graphics Control Extension
    # Random delay time and transparent color index
    delay_time = rng.read(2)
    transparent_color = rng.read(1)[0] % 256

    gce = struct.pack(
        '>BBBBHB',
        0x21,                   # Extension Introducer
        0xF9,                   # Graphic Control Label
        4,                      # Block Size
        disposal_method=0,      # No Disposal specified
        user_input_flag=False,  # User input not expected
        transparent_color,      # Transparent color index
        delay_time             # Delay time in milliseconds
    )

    # 5. Image Descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)

    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0]

    image_descriptor = struct.pack(
        '>BBHHHH',
        0x2C,                   # Image Separator
        lzw_minimum_code_size,  # LZW Minimum Code Size
        left,
        top,
        width,
        height
    )

    # 6. Image Data
    # Generate some random data with variable length
    blocks = []

    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while len(blocks) < total_blocks_size:
        block_size = min(total_blocks_size - len(blocks), rng.read(1)[0])
        blocks.append(rng.read(block_size))

    image_data = struct.pack('>BB', lzw_minimum_code_size, 0) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
