#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), byteorder='big') % (50 + 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='big') % (50 + 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    bg_color_index = int.from_bytes(rng.read(1), byteorder='big') & 0x0F
    pixel_aspect_ratio = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    logical_screen_descriptor = \
            width.to_bytes(2, byteorder='big', signed=False) + \
            height.to_bytes(2, byteorder='big', signed=False) + \
            flags + \
            bg_color_index.to_bytes(1, byteorder='big', signed=False) + \
            pixel_aspect_ratio

    # 3. Global Color Table
    global_color_table_size = int.from_bytes(rng.read(1), byteorder='big') & 0x07
    global_color_table_sort = bool(int.from_bytes(rng.read(1), byteorder='big'))
    global_color_table_bpp = ((global_color_table_size+1)*3).bit_length() - 1
    global_color_table_entry_size = 2 ** global_color_table_bpp
    global_color_table_packed = True if global_color_table_bpp == 1 else False
    global_color_table_data = []
    for i in range(2**global_color_table_size):
        color = [int.from_bytes(rng.read(1), byteorder='big')] * 3
        global_color_table_data += color
    global_color_table = b''
    for i in range(len(global_color_table_data)):
        if not global_color_table_packed or i % 3!= 2:
            global_color_table += \
                    global_color_table_data[i].to_bytes(
                            global_color_table_entry_size // 8, 
                            byteorder='big', signed=False
                        )

    # 4. Image Descriptor
    image_left = int.from_bytes(rng.read(2), byteorder='big')
    image_top = int.from_bytes(rng.read(2), byteorder='big')
    image_width = int.from_bytes(rng.read(2), byteorder='big') % (50 + 1) + 1
    image_height = int.from_bytes(rng.read(2), byteorder='big') % (50 + 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    local_color_table_flag = bool((flags[0] >> 7) & 0b1)
    interlace_flag = bool((flags[0] >> 6) & 0b1)
    sort_flag = bool((flags[0] >> 5) & 0b1)
    reserved_bits = flags[0] & 0b11100000
    lct_size = int.from_bytes(rng.read(1), byteorder='big') & 0x07
    lct_bpp = ((lct_size+1)*3).bit_length() - 1
    lct_entry_size = 2 ** lct_bpp
    lct_packed = True if lct_bpp == 1 else False
    lct_data = []
    for i in range(2**lct_size):
        color = [int.from_bytes(rng.read(1), byteorder='big')] * 3
        lct_data += color
    lct = b''
    for i in range(len(lct_data)):
        if not lct_packed or i % 3!= 2:
            lct += \
                lct_data[i].to_bytes(
                        lct_entry_size // 8, 
                        byteorder='big', signed=False
                    )

    # 5. Image Data
    data = b''
    while len(data) < image_width*image_height:
        block_size = int.from_bytes(rng.read(1), byteorder='big')
        if block_size > 0:
            data += rng.read(block_size)
        elif block