#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    has_gce = bool((rng.read(1)[0] & 1) << 7)  # Has GCE?
    gce = None
    if has_gce:
        gce = b'\x21\xF9\x04'  # Application extension block
        gce += struct.pack('>BBHBBBBB', rng.read(1)[0], rng.read(1)[0], rng.read(2)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('>BHHBBBB', 0x2C, 0, 0, width, height, 0b11100000 | ((rng.read(1)[0] & 7) << 2), 0)  # No local color table, interlaced, no sort flag, top-left corner
    
    # 6. Image Data
    image_data = b''
    for _ in range(height):
        image_data += b'\x00'  # Plain black background
        for i in range(width):
            image_data += color_table[(rng.read(1)[0] % num_colors)]  # Pick a random color
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)