#!/usr/bin/env python3

import struct, sys


def read_uint(f):
    return int.from_bytes(f.read(4), byteorder='little')


def write_uint(f, n):
    f.write(n.to_bytes(4, byteorder='little'))


def read_int(f):
    return int.from_bytes(f.read(4), byteorder='little', signed=True)


def write_int(f, n):
    f.write(n.to_bytes(4, byteorder='little', signed=True))


def read_float(f):
    return struct.unpack('<f', f.read(4))[0]


def write_float(f, n):
    f.write(struct.pack('<f', n))


def read_string(f):
    length = read_uint(f)

    if length == 0:
        return ''

    s = bytes(f.read(length)).decode('utf-8')

    # Skip the null terminator
    assert f.read(1) == b'\x00'

    return s


def write_string(f, s):
    encoded = (s + '\x00').encode('utf-8')
    write_uint(f, len(encoded))
    f.write(encoded)


def read_array(f, count, reader):
    arr = []
    for i in range(count):
        arr.append(reader(f))
    return arr


def write_array(f, arr, writer):
    write_uint(f, len(arr))
    for x in arr:
        writer(f, x)


def read_vec2(f):
    return [read_float(f), read_float(f)]


def write_vec2(f, v):
    write_float(f, v[0])
    write_float(f, v[1])


def read_vec3(f):
    return [read_float(f), read_float(f), read_float(f)]


def write_vec3(f, v):
    write_float(f, v[0])
    write_float(f, v[1])
    write_float(f, v[2])


def read_quat(f):
    return [read_float(f), read_float(f), read_float(f), read_float(f)]


def write_quat(f, q):
    write_float(f, q[0])
    write_float(f, q[1])
    write_float(f, q[2])
    write_float(f, q[3])


def read_mat4(f):
    m = [[None]*4 for _ in range(4)]
    for y in range(4):
        for x in range(4):
            m[y][x] = read_float(f)
    return m


def write_mat4(f, m):
    for y in range(4):
        for x in range(4):
            write_float(f, m[y][x])


def read_mesh(f):
    mesh = {}
    mesh['name'] = read_string(f)
    mesh['material'] = read_string(f)
    mesh['vertices'] = read_array(f, read_uint(f), read_vec3)
    mesh['normals'] = read_array(f, read_uint(f), read_vec3)
    mesh['texcoords'] = read_array(f, read_uint(f), read_vec2)
    mesh['indices'] = read_array(f, read_uint(f), lambda f: read_uint(f))
    return mesh


def write_mesh(f, mesh):
    write_string(f, mesh['name'])
    write_string(f, mesh['material'])
    write_array(f, mesh['vertices'], write_vec3)
    write_array(f, mesh['normals'], write_vec3)
    write_array(f, mesh['texcoords'], write_vec2)
    write_array(f, mesh['indices'], lambda f, n: write_uint(f, n))


def read_scene(f):
    scene = {}
    scene['meshes'] = read_array(f, read_uint(f), read_mesh)
    return scene


def write_scene(f, scene):
    write_array(f, scene['meshes'], write_mesh)


def main():
    with open(sys.argv[1], 'rb') as f:
        magic = f.read(6).