#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = [struct.pack('>BBBB', *rng.read(4)) for _ in range(num_colors)]
    color_table = b''.join(global_color_table)

    # 4. Graphics Control Extension
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    image_descriptor = struct.pack('>HHBB', 0, 0, 0b10000000 | rng.read(1)[0], 0)

    # 6. Image Data
    image_data = b''
    while len(image_data) < width*height:
        image_data += rng.read((width+7)//8)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)