#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])
    
    # 3. Global Color Table
    global_color_table_flag = (packed >> 7) & 1
    global_color_table_size = ((packed << 1) | (aspect_ratio >> 7)) & 0xf
    global_color_table = []
    for _ in range(global_color_table_size):
        r, g, b = [rng.read(1)[0] for _ in range(3)]
        global_color_table.extend([r, g, b])
    if len(global_color_table)!= 0:
        global_color_table = bytes(global_color_table)
    else:
        global_color_table = None
    
    # 4. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    local_color_table_flag = (packed >> 7) & 1
    interlace_flag = (packed >> 6) & 1
    sort_flag = (packed >> 5) & 1
    reserved = (packed << 3) & 0xe0
    size_of_local_color_table = (packed << 8) | rng.read(1)[0]
    local_color_table = []
    for _ in range(size_of_local_color_table):
        r, g, b = [rng.read(1)[0] for _ in range(3)]
        local_color_table.extend([r, g, b])
    if len(local_color_table)!= 0:
        local_color_table = bytes(local_color_table)
    else:
        local_color_table = None
    lzw_min_code_size = rng.read(1)[0]
    image_descriptor = struct.pack('>HHBBB', left, top, packed, lzw_min_code_size, rng.read(1)[0])
    
    # 5. Graphic Control Extension
    has_transparent_color = bool(rng.read(1)[0])
    transparent_color_index = rng.read(1)[0]
    delay_time = struct.unpack('>H', rng.read(2))[0]
    transparent_color_flag = bool(rng.read(1)[0])
    terminator = rng.read(1)[0]
    gce = struct.pack('>BBHB', 0x21, 0xf9, 4, delay_time)
    if has_transparent_color or transparent_color_flag:
        gce += struct.pack('B', 0x08)
        gce += struct.pack('B', transparent_color_index)
    gce += struct.pack('B', terminator)
    
    # 6. Image Data
    block_size = 2 ** lzw_min_code_size
    data = []
    while True:
        code = rng.read(1)[0]
        if code == 0x00: break
        
        if code < 0x000000ff:
            data.append(struct.pack('B', code))
        elif code < 0x0000ffff:
            data.append(struct.pack('<BH', 0x01, code))
        elif code < 0x00ffffff:
            data.append(struct.pack('<BL', 0x02, code))
        else:
            data.append(rng.read(block_size))
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)


if __name__ == '__main__':
    from sys import argv

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)