#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data
    data = []
    while True:
        # Get a random byte to decide what to do next
        choice = rng.read(1)[0]
        
        if not choice & 1:  # RLE
            # Read the length of the run from another random byte
            length = rng.read(1)[0]
            
            # Add the same byte to our data buffer that many times
            data += [choice] * length
        elif not choice & 2:  # Raw
            # Read the raw data from another random byte
            data += [rng.read(1)[0]]
        elif not choice & 4:  # EOF
            break
        else:  # Bad bit set!
            raise Exception("Bad bit in choice!")
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything to the output stream
    out.write(header + lsd + color_table + gce + image_descriptor + bytes(data) + trailer)


if __name__ == '__main__':
    from sys import argv
    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)