#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHBB', 0x2C, 0, 0, 0, 0)

    # 6. Image Data (compressed with RLE)
    data = []
    while True:
        # Get next byte from /dev/urandom
        b = rng.read(1)

        # If we got EOF, break out of loop
        if not len(b):
            break

        # Otherwise, append to our data array
        data += [b[0]]

        # If this is an RLE block, get another byte as its length
        if b[0] >= 0xE0:
            # Append that many copies of the previous byte
            data += [data[-1]] * ((b[0] & 0x07) + 1)

            # Skip ahead by one more byte because we've already read it
            continue

        # If this is a plain block, get another byte as its length
        elif b[0] >= 0xC0:
            # Append that many copies of the next byte
            data += [rng.read(1)[0]] * ((b[0] & 0x07) + 1)

            # Skip ahead by one more byte because we've already read it
            continue

        # If this is a clear code, reset the dictionary
        elif b[0] == 0x00:
            pass

        # If this is an end-of-image code, break out of loop
        elif b[0] == 0x3B:
            break

        # If this is any other code, print error message and exit
        else:
            raise Exception("Invalid GIF encoding")

    # Compress the data using RLE
    compressed_data = []
    prev = -1
    count = 0
    for i in range(len(data)):
        if data[i]!= prev:
            if count > 0:
                compressed_data += [(count << 5) | prev]
            prev = data[i]
            count = 1
        else:
            count += 1
    if count > 0:
        compressed_data += [(count << 5) | prev]

    # Write everything to output stream
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    out.write(bytes(compressed_data))

