#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = [struct.pack('>BBBB', *rng.read(4)) for _ in range(num_colors)]
    color_table = b''.join(global_color_table)

    # 4. Graphics Control Extension
    graphics_control_extension_label = 0x21
    graphics_control_extension_identifier = 0xF9
    graphics_control_extension_block_size = 0x04
    delay_time = rng.read(1)[0]
    transparent_color_index = rng.read(1)[0]
    graphics_control_extension = struct.pack('>BBBBB', graphics_control_extension_label,
                                             graphics_control_extension_identifier,
                                             graphics_control_extension_block_size,
                                             delay_time,
                                             transparent_color_index)

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(packed_fields & 0b10000000)
    interlace_flag = bool(packed_fields & 0b01000000)
    sort_flag = bool(packed_fields & 0b00100000)
    reserved_bits = packed_fields & 0b00011111
    image_descriptor = struct.pack('>BHHHHB', 0x2C,
                                   left, top, width, height,
                                   local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved_bits)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + graphics_control_extension + image_descriptor + image_data + trailer)