#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The size of the generated image is determined by
# `width`, `height`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    packed = rng.read(1)[0] | 0x80  # Set global color table flag and set color resolution to 10 bits
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        out.write(header + lsd)
        return

    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = [struct.pack('BBB', *(rng.read(3))) for _ in range(num_colors)]  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    data = []
    while True:
        block_size = rng.read(1)[0] & 0xFF

        if block_size == 0:
            break

        data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to output
    out.write(header + lsd + b''.join(color_table) + gce + image_descriptor + b''.join(data) + trailer)


# Generate a bunch of random GIFs with different sizes and write them to files named "random-*.gif"
def main() -> None:
    with open('/dev/urandom', 'rb') as rng:
        for i in range(10):
            width = rng.read(1)[0] % 100 + 100
            height = rng.read(1)[0] % 100 + 100

            with open(f'random-{i}.gif', 'wb') as f:
                generate_random_gif(rng, f, width, height)


if __name__ == '__main__':
    main()