#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a' + (b'\x00' * 7)
    out.write(header)

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'little') % 640
    height = int.from_bytes(rng.read(2), 'little') % 480
    background_color = rng.read(1)[0] & 0xF
    pixel_aspect_ratio = rng.read(1)[0] & 0xF
    lsd = struct.pack('<HHBB', width, height, background_color, pixel_aspect_ratio)
    out.write(lsd)

    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 0xF
    global_color_table = [
            rng.read(3) for _ in range(2 ** (global_color_table_size + 1))
        ]
    out.write(b''.join(global_color_table))

    # 4. Image Data
    image_descriptor = struct.pack('<HHB', 0, 0, 0)
    out.write(image_descriptor)

    local_color_table_size = rng.read(1)[0] & 0xF
    local_color_table = [
            rng.read(3) for _ in range(2 ** (local_color_table_size + 1))
        ]
    out.write(b''.join(local_color_table))

    min_code_size = rng.read(1)[0] & 0xF
    out.write(min_code_size.to_bytes(1, 'little'))

    def get_next_byte():
        return rng.read(1)[0]

    def read_blocks():
        while True:
            block_length = get_next_byte()
            if block_length == 0:
                break
            yield rng.read(block_length)

    def write_blocks(blocks):
        for block in blocks:
            out.write((len(block)).to_bytes(1, 'little'))
            out.write(block)
        out.write(b'\x00')

    def read_subblocks():
        subblock_data = bytearray()
        while True:
            subblock_length = get_next_byte()
            if subblock_length == 0:
                break
            subblock_data += rng.read(subblock_length)
        return subblock_data

    def write_subblocks(subblocks):
        for subblock in subblocks:
            out.write((len(subblock)).to_bytes(1, 'little'))
            out.write(subblock)
        out.write(b'\x00')

    def read_compressed_data():
        compressed_data = bytearray()
        while True:
            code = get_next_byte()
            if code == 0:
                break
            compressed_data.append(code)
        return compressed_data

    def write_compressed_data(compressed_data):
        for code in compressed_data:
            out.write(code.to_bytes(1, 'little'))

    def read_uncompressed_data():
        uncompressed_data = bytearray()
        while True:
            data = rng.read(1)
            if not data:
                break
            uncompressed_data.extend(data)
        return uncompressed_data

    def write_uncompressed_data(uncompressed_data):
        out.write(uncompressed_data)

    def read_extension():
        extension_introducer = get_next_byte()
        assert extension_introducer == 0x21
        extension_label = get_next_byte()
        if extension_label == 0xFF:
            application_identifier = rng.read(8).decode().strip('\x00')
            authentication_code = rng.read(3)
            return {
                    'application_identifier': application_identifier,
                    'authentication