#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    # Packed fields:
    #   Global Color Table Flag (1 bit): 0
    #   Color Resolution (3 bits): 7
    #   Sort Flag (1 bit): 0
    #   Size of Global Color Table (3 bits): 7
    screen_descriptor = struct.pack('B', 0b00000000 | 0b01110000 | 0b00000000 | 0b01110000)
    # Background color index
    background_color_index = rng.read(1)[0] % 256
    # Pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)[0] % 256

    # 3. Global Color Table
    global_color_table = []
    for _ in range(2 ** ((screen_descriptor[0] & 0b00000111) + 1)):
        # Each entry is three bytes: red, green, blue
        global_color_table.append(struct.pack('BBB', *(rng.read(3))))

    # 4. Image Data
    image_data = []
    # Loop over all pixels
    for y in range(height):
        row = []
        for x in range(width):
            # Each pixel is one byte: index into the global color table
            row.append(rng.read(1)[0])
        image_data.append(row)

    # Write everything to `out`
    out.write(header)
    out.write(screen_descriptor)
    for color in global_color_table:
        out.write(color)
    out.write(bytes([background_color_index]))
    out.write(bytes([pixel_aspect_ratio]))
    for row in image_data:
        out.write(bytes(row))
    out.write(b';')

