#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Optional Graphic Control Extension
    # We'll do this half the time
    gce = None
    if rng.read(1)[0] & 1:
        # 4.1. Block Size
        gce = b'\x21\xf9\x04'
        
        # 4.2. Graphic Control Label
        # Always 0xF9
        
        # 4.3. Graphic Control Block
        # Disposal method (always 0x00)
        # User input flag (always 0x00)
        # Transparent color flag (always 0x00)
        # Delay time (two bytes, random)
        gce += struct.pack('>BBH', 0, 0, struct.unpack('>H', rng.read(2))[0])
        
        # 4.4. Block Terminator
        # Always 0x00
        gce += b'\x00'
    
    # 5. Image Descriptor
    image_descriptor = b'\x2C'
    
    # 6. Image Data
    # First byte is the minimum code size (up to 8 bits)
    lzw_minimum_code_size = rng.read(1)[0]
    
    # Then we have one or more data sub-blocks
    blocks = []
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        
        if block_size == 0:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom

    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)