#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = min(rng.read(1)[0] + 1, 256)
    global_color_table = rng.read(global_color_table_size * 3)
    
    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBB', 0, 0, width, height, 0)
    
    # 5. Blocks
    # We need to generate some blocks, but we don't know how many yet.
    # So let's just keep generating them until we get one that ends with an "end of image" code
    blocks = []
    while True:
        # Each block starts with a length byte
        length = rng.read(1)[0]
        
        # Then there are `length` bytes of data
        block = rng.read(length)
        
        if block[-1] == 0x3b:
            break
            
        blocks.append(block)
    
    # 6. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output
    out.write(header)
    out.write(lsd)
    out.write(struct.pack('B', global_color_table_size))
    out.write(global_color_table)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)