#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Packed fields:
    #   - Global Color Table Flag: 1
    #   - Color Resolution: 8 bits per pixel
    #   - Sort Flag: Not sorted
    #   - Size of Global Color Table: 256 colors
    packed_fields = b'\x80\x08\x00\xFF'
    
    # Background Color Index
    background_color_index = rng.read(1)[0] % 256
    
    # Pixel Aspect Ratio
    pixel_aspect_ratio = b'\x00'
    
    # Combine all fields together
    lsd = struct.pack('<HBBB', width, height, *packed_fields[::-1], background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    # We will use the default global color table
    color_table = b''
    
    # 4. Graphic Control Extension
    # We will use the default graphic control extension
    gce = None
    
    # 5. Image Descriptor
    # Random left and top position
    left, top = struct.unpack('>HH', rng.read(4))
    
    # Width and height are taken from the logical screen descriptor
    image_descriptor = struct.pack('<HHBB', left, top, width, height)
    
    # 6. Image Data
    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0] % 8
    
    # Blocks of data
    blocks = []
    while True:
        # Read a byte
        byte = rng.read(1)
        if len(byte) == 0:
            break
            
        # Append it to the list of blocks
        blocks.append(byte)
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom

    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)