#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # 1. Header
    header = b"GIF89a"

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), byteorder='big') % (65536 - 10) + 10
    height = int.from_bytes(rng.read(2), byteorder='big') % (65536 - 10) + 10
    packed = int.from_bytes(rng.read(1), byteorder='big')
    global_color_table_flag = bool((packed & 0b10000000) >> 7)
    color_resolution = ((packed & 0b01110000) >> 4) + 1
    sort_flag = bool((packed & 0b00001000) >> 3)
    size_of_global_color_table = 2 ** ((packed & 0b00000111)) * 3
    background_color_index = int.from_bytes(rng.read(1), byteorder='big')
    pixel_aspect_ratio = int.from_bytes(rng.read(1), byteorder='big')

    lsd = struct.pack('<HHBBB', width, height, packed, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    if global_color_table_flag:
        color_table = rng.read(size_of_global_color_table)
    else:
        color_table = None

    # 4. Graphic Control Extension
    gce = b"\x21\xF9\x04\x01\x00\x00\xFF\xFF\xFF\x00"

    # 5. Image Descriptor
    image_descriptor = b"\x2C\x00\x00\x00\x00\x00\xFF\xFF\xFF\x00"

    # 6. Image Data
    lzw_minimum_code_size = rng.read(1)[0]
    block_size = 255
    blocks = []
    while True:
        data = rng.read(block_size)
        if len(data) < block_size:
            break
        if len(blocks) > 0 and len(data) == 255:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)