#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    flags = rng.read(1)
    background = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(width + height + flags + bytes([background]) + bytes([aspect]))

    # 3. Global Color Table (optional)
    if flags[0] & 0x80 != 0:
        color_table_size = 2 ** ((flags[0] & 0x7) + 1)
        for i in range(color_table_size * 3):
            out.write(bytes([rng.read(1)[0]]))

    # 4. Image Blocks
    while True:
        block_type = rng.read(1)
        if block_type == b'\x3B':
            break
        elif block_type == b'\x2C':
            # Image Descriptor
            left = rng.read(2)
            top = rng.read(2)
            width = rng.read(2)
            height = rng.read(2)
            flags = rng.read(1)
            local_palette_size = 2 ** ((flags[0] & 0x7) + 1)
            out.write(left + top + width + height + flags)

            # Local Color Table (optional)
            if flags[0] & 0x80 != 0:
                for i in range(local_palette_size * 3):
                    out.write(bytes([rng.read(1)[0]]))

            # LZW-compressed image data
            lzw_minimum_code_size = rng.read(1)[0]
            out.write(bytes([lzw_minimum_code_size]))
            while True:
                code_size = rng.read(1)[0]
                if code_size == 0:
                    break
                else:
                    out.write(bytes([code_size]))
                    for _ in range((256 // code_size) - 1):
                        out.write(bytes([rng.read(1)[0]]))

            # Trailer
            out.write(b'\x00')
        else:
            raise Exception('Invalid block type ' + str(block_type))
