#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    # Color table flag
    has_global_color_table = True if rng.read(1)[0] else False

    # Global color table size
    global_color_table_size = 2 ** ((packed_fields & 0b00000111) + 1)

    # Sort flag
    sort_flag = bool((packed_fields >> 3) & 1)

    # Color resolution
    color_resolution = (packed_fields >> 4) & 0b00000111

    # Global color table
    color_table = rng.read(3 * global_color_table_size)

    # 3. Graphic Control Extension
    gce = None
    if rng.read(1)[0]:
        # Block size
        gce = b'\x21\xf9\x04'

        # Packed field
        gce += struct.pack('B', rng.read(1)[0])

        # Delay time
        delay_time = rng.read(2)
        gce += delay_time

        # Transparent color index
        transparent_color_index = rng.read(1)[0]
        gce += transparent_color_index

        # Terminator
        gce += b'\x00'

    # 4. Image descriptor
    image_descriptor = b'\x2c\x00\x00\x00\x00' + struct.pack('>H', width) + struct.pack('>H', height)

    # 5. Local color table
    local_color_table = rng.read(3 * 2**((packed_fields >> 4) & 0b00000111))

    # 6. Image data
    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0]

    # LZW blocks
    blocks = []
    while True:
        # Read a single byte
        block_length = rng.read(1)[0]

        # If it is zero, we are done
        if not block_length:
            break

        # Otherwise read that many bytes
        blocks.append(rng.read(block_length))

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)

