#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    global_color_table_flag = bool(rng.read(1)[0] & 0x80)
    if global_color_table_flag:
        global_color_table_size = (rng.read(1)[0] & 0x07) + 1
        global_color_table = rng.read(global_color_table_size * 3)
    else:
        global_color_table = b''

    # 4. Graphic Control Extension
    # We use this to set the delay time between frames
    gce = None
    if rng.read(1)[0] & 0x80:
        # The extension exists
        gce = struct.pack('>BBBB',
                          0xF9,
                          rng.read(1)[0],
                          rng.read(1)[0],
                          rng.read(1)[0]
                         )

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))

    local_color_table_flag = bool(rng.read(1)[0] & 0x80)
    interlace_flag = bool(rng.read(1)[0] & 0x40)
    sort_flag = bool(rng.read(1)[0] & 0x20)
    reserved_flags = rng.read(1)[0] & 0x0C

    if local_color_table_flag:
        local_color_table_size = (rng.read(1)[0] & 0x07) + 1
        color_table = rng.read(local_color_table_size * 3)
    elif global_color_table_flag:
        color_table = global_color_table
    else:
        color_table = b''

    image_descriptor = struct.pack('>BHHHHBBBBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  local_color_table_size
                                 )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size - 1))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')