#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    out.write('GIF89a'.encode())

    # 2. Logical Screen Descriptor
    width = rng.read(2)[0] + (rng.read(1)[0] << 8)
    height = rng.read(2)[0] + (rng.read(1)[0] << 8)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table
    global_color_table_flag = rng.read(1)[0] >> 7 == 1
    if global_color_table_flag:
        color_resolution = rng.read(1)[0] & 0b00000111
        sort_flag = rng.read(1)[0] >> 7 == 1
        size_of_global_color_table = 2 ** ((rng.read(1)[0] & 0b00000111) + 1)
        for _ in range(size_of_global_color_table):
            red = rng.read(1)[0]
            green = rng.read(1)[0]
            blue = rng.read(1)[0]
            out.write(struct.pack('<BBB', red, green, blue))

    # 4. Image Data
    while True:
        # 4.1 Graphic Control Extension
        extension_introducer = b'\x21'
        graphic_control_label = b'\xf9'
        block_size = b'\x04'
        disposal_method = rng.read(1)[0] & 0b00011100
        user_input_flag = rng.read(1)[0] >> 7 == 1
        transparent_color_flag = rng.read(1)[0] >> 6 == 1
        delay_time = rng.read(2)[0] + (rng.read(1)[0] << 8)
        transparent_color_index = rng.read(1)[0]
        block_terminator = b'\x00'
        out.write(extension_introducer + graphic_control_label + block_size + struct.pack('<BBHB', disposal_method, user_input_flag, transparent_color_flag, delay_time, transparent_color_index) + block_terminator)

        # 4.2 Image Descriptor
        image_separator = b'\x2c'
        left_position = rng.read(2)[0] + (rng.read(1)[0] << 8)
        top_position = rng.read(2)[0] + (rng.read(1)[0] << 8)
        width = rng.read(2)[0] + (rng.read(1)[0] << 8)
        height = rng.read(2)[0] + (rng.read(1)[0] << 8)
        local_color_table_flag = rng.read(1)[0] >> 7 == 1
        interlace_flag = rng.read(1)[0] >> 6 == 1
        sort_flag = rng.read(1)[0] >> 5 == 1
        size_of_local_color_table = 2 ** ((rng.read(1)[0] & 0b00000111) + 1)
        out.write(image_separator + struct.pack('<HHHHBB', left_position, top_position, width, height, local_color_table_flag, interlace_flag, sort_flag, size_of_local_color_table))

        # 4.3 Local Color Table
        if local_color_table_flag:
            for _ in range(size_of_local_color_table):
                red = rng.read(1)[0]
                green = rng.read(1)[0]
                blue = rng.read(1)[0]
                out.write(struct.pack