#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor (LSD)
    width = rng.read(2)
    height = rng.read(2)
    packed_fields = rng.read(1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)
    
    lsd = width + height + packed_fields + background_color_index + pixel_aspect_ratio
    
    # 3. Global Color Table
    global_color_table_flag = packed_fields[0] & 0b10000000
    color_resolution = (packed_fields[0] >> 4) & 0b111
    sort_flag = packed_fields[0] & 0b100
    global_color_table_size = 2 ** ((packed_fields[0] & 0b11) + 1)
    
    if global_color_table_flag != 0:
        color_table = rng.read(global_color_table_size * 3)
    else:
        color_table = b''
    
    # 4. Graphic Control Extension (optional)
    gce = None
    
    if rng.read(1)[0] == 0x21:
        extension_introducer = rng.read(1)
        graphic_control_label = rng.read(1)
        block_size = rng.read(1)
        
        disposal_method = (block_size[0] >> 2) & 0b111
        user_input_flag = (block_size[0] >> 1) & 0b1
        transparent_color_flag = block_size[0] & 0b1
        delay_time = rng.read(2)
        transparent_color_index = rng.read(1)
        block_terminator = rng.read(1)
        
        gce = extension_introducer + graphic_control_label + block_size + \
              struct.pack('<BBHBB', disposal_method, user_input_flag, int.from_bytes(delay_time, byteorder='big'), transparent_color_index[0], block_terminator[0])
    
    # 5. Image Descriptor
    image_descriptor = b'\x2C' + rng.read(9)
    
    # 6. Image Data
    lzw_minimum_code_size = rng.read(1)[0]
    
    blocks = []
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        
        if block_size == 0:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)

if __name__ == "__main__":
    from os import urandom

    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)