#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('<HH', rng.read(4))
    
    # Random background color index
    bg_color_index = struct.unpack('B', rng.read(1))[0]

    # Random pixel aspect ratio
    pixel_aspect_ratio = struct.unpack('B', rng.read(1))[0]

    # 3. Global Color Table
    # Random number of colors in global color table
    num_colors = struct.unpack('B', rng.read(1))[0] + 1

    # Random color values
    global_color_table = []
    for _ in range(num_colors):
        red, green, blue = struct.unpack('BBB', rng.read(3))
        global_color_table.append((red, green, blue))

    # 4. Image Data
    # Random number of images
    num_images = struct.unpack('B', rng.read(1))[0] + 1

    # For each image...
    for i in range(num_images):
        # Random left position
        left = struct.unpack('H', rng.read(2))[0]
        
        # Random top position
        top = struct.unpack('H', rng.read(2))[0]

        # Random width
        width = struct.unpack('H', rng.read(2))[0]

        # Random height
        height = struct.unpack('H', rng.read(2))[0]

        # Random local color table flag
        local_color_table_flag = struct.unpack('B', rng.read(1))[0] & 0x80 != 0

        # Random interlaced flag
        interlaced_flag = struct.unpack('B', rng.read(1))[0] & 0x40 != 0

        # Random sort flag
        sort_flag = struct.unpack('B', rng.read(1))[0] & 0x20 != 0

        # Random size of local color table
        if local_color_table_flag:
            num_local_colors = struct.unpack('B', rng.read(1))[0] + 1
        else:
            num_local_colors = num_colors

        # Random local color table
        local_color_table = []
        for _ in range(num_local_colors):
            red, green, blue = struct.unpack('BBB', rng.read(3))
            local_color_table.append((red, green, blue))

        # Random transparent color index
        transparency_color_index = struct.unpack('B', rng.read(1))[0]

        # Random disposal method
        disposal_method = struct.unpack('B', rng.read(1))[0] & 0x1C >> 2

        # Random delay time
        delay_time = struct.unpack('H', rng.read(2))[0]

        # Random transparent color flag
        transparency_flag = struct.unpack('B', rng.read(1))[0] & 0x01 != 0

        # Random block terminator
        block_terminator = struct.unpack('B', rng.read(1))[0]

        # Write all this to the output file
        out.write(struct.pack('<6sHHBBBBBBBBB',
                              header,
                              width, height,
                              bg_color_index,
                              pixel_aspect_ratio,
                              num_colors - 1,
                              0, 0,
                              len(global_color_table),
                              num_images - 1,
                              i,
                              left, top,
                              width, height,
                              local_color_table_flag,
                              interlaced_flag,
                              sort_flag,
                              num_local_colors - 1,
                              transparency_color_index,
                              disposal_method,
                              delay_time,
                              transparency_flag,
                              block_terminator))

        # Write the global color

if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('./test.gif', 'wb') as f:
            generate_random_gif(rng, f)