#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big', signed=False) + 1
    height = int.from_bytes(rng.read(2), 'big', signed=False) + 1
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table (optional)
    if rng.read(1)[0] % 2 == 0:
        global_color_table_size = 7 + rng.read(1)[0] % 8 * 3
        global_color_table = [
            tuple([int.from_bytes(rng.read(1), 'big', signed=False)] * 3) for _ in range(global_color_table_size // 3)
        ]
        out.write(b'\xFF\xFF')
        out.write(struct.pack('B', global_color_table_size - 1))
        for color in global_color_table:
            out.write(struct.pack('BBB', *color))

    # 4. Image Data
    while True:
        # 4.1 Graphic Control Extension
        if rng.read(1)[0] % 2 == 0:
            block_size = rng.read(1)[0]
            assert block_size == 0xF4
            flags = rng.read(1)[0]
            delay = int.from_bytes(rng.read(2), 'big', signed=False)
            transparent_color_index = rng.read(1)[0] % 256
            terminator = rng.read(1)[0]
            assert terminator == 0
            out.write(b'\x21\xF9\x04\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\