#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = None
    if rng.read(1)[0]:
        # We want to use this extension
        
        # 4.1 Graphic Control Label
        gce = b'\x21\xf9\x04'
        
        # 4.2 Graphic Control Block
        # Disposal method, user input flag, delay time in centiseconds
        gce += struct.pack('BBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
        
        # Transparent color index
        gce += rng.read(1)
        
        # Block terminator
        gce += b'\x00'
    
    # 5. Image Descriptor
    idesc = b'\x2c' + struct.pack('>HHB', 0, 0, 0)
    
    # 6. Local Color Table (optional)
    lct = None
    if rng.read(1)[0]:
        # We want to use this table
        
        # 6.1 Local Color Table Size
        lct = struct.pack('B', rng.read(1)[0])
        
        # 6.2 Local Color Table
        lct += rng.read(3 * (lct[0] + 1))
    
    # 7. Image Data
    image_data = rng.read()
    
    # 8. Trailer
    trailer = b'\x3b'
    
    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + idesc + (lct or b'') + image_data + trailer)
