#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = rng.read(2)[0]
    height = rng.read(2)[0]
    packed = rng.read(1)[0]
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('<HBBBB', width, height, packed, background, aspect_ratio)
    
    # 3. Color Table (optional)
    color_table = b''
    if packed & 0x80 != 0:
        num_colors = 2 ** ((packed >> 4) + 1)
        for i in range(num_colors):
            color_table += rng.read(3)
        
    # 4. Graphic Control Extension (optional)
    gce = None
    if rng.read(1)[0] & 0xF != 0:
        block_size = rng.read(1)[0]
        flags = rng.read(1)[0]
        delay = rng.read(2)
        transparent = rng.read(1)[0]
        terminator = rng.read(1)[0]
        gce = struct.pack('<BBBBB', block_size, flags, delay[0], delay[1], transparent)
    
    # 5. Image Descriptor
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = rng.read(2)[0]
    height = rng.read(2)[0]
    packed = rng.read(1)[0]
    lct_flag = bool((packed >> 7) & 0x1)
    interlace_flag = bool((packed >> 6) & 0x1)
    sort_flag = bool((packed >> 5) & 0x1)
    reserved = bool((packed >> 4) & 0x1)
    size_of_local_color_table = (packed >> 0) & 0x7
    image_descriptor = struct.pack('<HHBBBB', left, top, width, height, packed, 0)
    
    # 6. Local Color Table (optional)
    local_color_table = b''
    if lct_flag:
        num_colors = 2 ** (size_of_local_color_table + 1)
        for i in range(num_colors):
            local_color_table += rng.read(3)
    
    # 7. Image Data
    image_data = rng.read(width * height)
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)

if __name__ == "__main__":
    from os import urandom

    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)