#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = rng.read(2)[0]
    height = rng.read(2)[0]
    global_color_table_flag = bool(rng.read(1)[0] & 0b10000000)
    color_resolution = (rng.read(1)[0] >> 4) & 0b111
    sort_flag = bool(rng.read(1)[0] & 0b10000000)
    global_color_table_size = 2 ** ((rng.read(1)[0] & 0b111) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    lsd = header + struct.pack('<HBBBBB',
                               width,
                               height,
                               0x80 | global_color_table_flag << 7 | color_resolution << 4 | sort_flag << 3,
                               global_color_table_size,
                               background_color_index,
                               pixel_aspect_ratio
                              )

    # 3. Global Color Table
    color_table = rng.read(global_color_table_size * 3)

    # 4. Graphic Control Extension
    gce = None
    if bool(rng.read(1)[0] & 0b10000000):
        delay_time = rng.read(2)[0]
        transparent_color_index = rng.read(1)[0]
        terminator = rng.read(1)[0]

        gce = struct.pack('>BBBB',
                          0xF9,
                          delay_time,
                          transparent_color_index,
                          terminator
                         )

    # 5. Image Descriptor
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = rng.read(2)[0]
    height = rng.read(2)[0]
    local_color_table_flag = bool(rng.read(1)[0] & 0b10000000)
    interlace_flag = bool(rng.read(1)[0] & 0b10000000)
    sort_flag = bool(rng.read(1)[0] & 0b10000000)
    reserved_flags = rng.read(1)[0] & 0b111
    local_color_table_size = 2 ** ((rng.read(1)[0] & 0b111) + 1)
    local_color_table = rng.read(local_color_table_size * 3)
    image_descriptor = struct.pack('>BHHHHBBBBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  local_color_table_size
                                 )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size - 1))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')