#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b"GIF89a"

    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed = rng.read(1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)

    lsd = width + height + packed + background_color_index + pixel_aspect_ratio

    # 3. Color Table
    # We don't care about this for now
    color_table = rng.read(3 * 2 ** ((packed[0] & 0b00000111) + 1))

    # 4. Graphic Control Extension
    gce = None
    if rng.read(1)[0]:
        # We have a graphic control extension
        block_size = rng.read(1)[0]
        assert block_size == 0xF4
        flags = rng.read(1)[0]
        delay_time = rng.read(2)
        transparent_color_index = rng.read(1)
        terminator = rng.read(1)[0]
        assert terminator == 0

        gce = bytes([block_size, flags]) + delay_time + transparent_color_index + terminator

    # 5. Image Descriptor
    image_descriptor = b"\x2c\x00\x00\x00\x00\x00\xff\xff\xff\xff\x00"

    # 6. Image Data
    # First byte is the minimum code size
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)

