#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = [0x7F & rng.read(1)[0],
            width % 256, width // 256,
            height % 256, height // 256] + \
         [0x7F & rng.read(1)[0]] * 3

    # 3. Global Color Table
    gct = []
    for _ in range(lsd[4]):
        r, g, b = [0xFF & rng.read(1)[0] for _ in range(3)]

        # Set the GCT entry to black if it's not used by the image data
        if len(gct) == lsd[5]:
            r, g, b = 0, 0, 0

        gct += [r, g, b]

    # 4. Image Data
    imgdata = []
    while True:
        # Get a random byte from the RNG
        b = rng.read(1)[0]

        # If the most significant bit is set, this is a control block
        if b >= 0b10000000:

            # The lower three bits are the type of control block
            cbtype = b & 0b00000111

            # The next five bits are the length of the control block
            cblen = ((b >> 3) & 0b00011111) + 1

            # Read the rest of the control block
            cbd = rng.read(cblen - 1)

            # Handle the different types of control blocks
            if cbtype == 0b000:
                # End of image
                break
            elif cbtype == 0b001:
                # Plain text extension
                pass
            elif cbtype == 0b010:
                # Application extension
                pass
            elif cbtype == 0b011:
                # Comment extension
                pass
            else:
                raise ValueError("Unknown control block")
        else:
            # This is an image descriptor
            # First byte contains packed field flags
            idflags = b

            # Next byte contains the local color table flag
            idltabflag = bool((idflags >> 7) & 0b1)

            # Next byte contains interlace flag
            idinterlaceflag = bool((idflags >> 6) & 0b1)

            # Next byte contains sort flag
            idsortflag = bool((idflags >> 5) & 0b1)

            # Next byte contains reserved flags
            idreserved = (idflags >> 3) & 0b11

            # Next byte contains size of local color table
            idsize = (idflags & 0b111) + 1

            # Next four bytes contain the left position
            idleft = struct.unpack('>H', rng.read(2))[0]

            # Next four bytes contain the top position
            idtop = struct.unpack('>H', rng.read(2))[0]

            # Next four bytes contain the width
            idwidth = struct.unpack('>H', rng.read(2))[0]

            # Next four bytes contain the height
            idheight = struct.unpack('>H', rng.read(2))[0]

            # Next byte contains packed field flags
            idflags = rng.read(1)[0]

            # Next byte contains background color index
            idbgindex = rng.read(1)[0]

            # Next byte contains pixel aspect ratio
            idextra = rng.read(1)[0]

            # Create the image descriptor
            imgdesc = [
                0b10000000 | idflags,
                idleft % 256, idleft // 256,
                idtop % 256, idtop // 256,
                idwidth % 256, idwidth // 256,
                idheight % 256, idheight // 256,
                idbgindex,
                idextra
            ]

            # Add the local color table if needed
            if idltabflag:
                imgdesc += list(rng.read(3 * idsize))

            # Add the image data
            imgdata += imgdesc + list(rng.read(idwidth * idheight))

    # Write everything to the output stream
    out.write(header)
    out.write(bytes(lsd))
    out.write(bytes(gct))
    out.write(bytes(imgdata))

