#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(num_colors)]
    color_table = b''.join(global_color_table)

    # 4. Graphics Control Extension (if needed)
    gce = None
    if bool(packed_fields & 0b1):
        gce = struct.pack('>BBHBBBBB',
                          0x21,  # Extension Introducer
                          ord('G'), ord('C'),  # Extension Label
                          4,  # Block Size
                          0b00000000 | rng.read(1)[0],  # Flags (bit 0 set means transparent color follows)
                          0,  # Delay Time (unused)
                          0,  # Transparent Color Index
                          0  # Block Terminator
                         )

    # 5. Image Descriptor
    left = rng.read(1)[0]
    top = rng.read(1)[0]
    width = (width + 1) % 256
    height = (height + 1) % 256
    local_color_table_flag = bool(rng.read(1)[0] % 2)
    interlace_flag = bool(rng.read(1)[0] % 2)
    sort_flag = bool(rng.read(1)[0] % 2)
    reserved_flags = rng.read(1)[0]
    local_color_table_size = (rng.read(1)[0] % 256) or 1
    image_descriptor = struct.pack('>BBBBBBB',
                                  0x2C,  # Image Separator
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  local_color_table_size
                                 )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size - 1))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')