#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), byteorder='little') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg_color = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect_ratio = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lsd = width.to_bytes(2, byteorder='little', signed=False) \
            + height.to_bytes(2, byteorder='little', signed=False) \
            + flags \
            + bg_color \
            + aspect_ratio

    # 3. Global Color Table
    global_color_table_size = int.from_bytes(rng.read(1), byteorder='little') & 7
    global_color_table_flag = bool((int.from_bytes(rng.read(1), byteorder='little') >> 7) & 1)
    if global_color_table_flag:
        global_color_table = [b'\x00\x00\x00'] * (global_color_table_size ** 2)

        for i in range(len(global_color_table)):
            global_color_table[i] = bytes([
                int.from_bytes(rng.read(1), byteorder='little'),
                int.from_bytes(rng.read(1), byteorder='little'),
                int.from_bytes(rng.read(1), byteorder='little')])

        gct = bytes([global_color_table_size | 0x80]) + bytes(sum(global_color_table, []))
    else:
        gct = b''

    # 4. Image Descriptor
    left = int.from_bytes(rng.read(2), byteorder='little') % (width - 1)
    top = int.from_bytes(rng.read(2), byteorder='little') % (height - 1)
    image_width = int.from_bytes(rng.read(2), byteorder='little') % (width - left) + 1
    image_height = int.from_bytes(rng.read(2), byteorder='little') % (height - top) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    imgd = left.to_bytes(2, byteorder='little', signed=False) \
           + top.to_bytes(2, byteorder='little', signed=False) \
           + image_width.to_bytes(2, byteorder='little', signed=False) \
           + image_height.to_bytes(2, byteorder='little', signed=False) \
           + flags

    # 5. Local Color Table
    local_color_table_size = int.from_bytes(rng.read(1), byteorder='little') & 7
    local_color_table_flag = bool((int.from_bytes(rng.read(1), byteorder='little') >> 7) & 1)
    if local_color_table_flag:
        local_color_table = [b'\x00\x00\x00'] * (local_color_table_size ** 2)

        for i in range(len(local_color_table)):
            local_color_table[i] = bytes([
                int.from_bytes(rng.read(1), byteorder='little'),
                int.from_bytes(rng.read(1), byteorder='little'),
                int.from_bytes(rng.read(1), byteorder='little')])

        lct = bytes([local_color_table_size | 0x80]) + bytes(sum(local_color_table, []))
    else:
        lct = b''

    # 6. Block Size
    block_size = int.from_bytes(rng.read(1), byteorder='little')

    # 7. Image Data
    data = [b'\x00\x00\x00'] * ((image_width * image_height // 8) + 1)

    for i in range(len(data)):
        data[i] = bytes([int.from_bytes(rng.read(block_size), byteorder='little