#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Random background color index
    bg_color_index = rng.read(1)[0]
    
    # Random pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)[0] & 0b1111
    
    lsd = struct.pack('>HHBB', width, height, bg_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    global_color_table_flag = bool(rng.read(1)[0] & 0b10000000)
    color_resolution = ((rng.read(1)[0] >> 4) & 0b11) or 1
    sort_flag = bool(rng.read(1)[0] & 0b1000)
    size_of_global_color_table = (rng.read(1)[0] & 0b111) or 1
    global_color_table = [rng.read(3) for _ in range((2 << size_of_global_color_table) - 1)]
    global_color_table += [(0, 0, 0), ] * (2 ** size_of_global_color_table - len(global_color_table))
    
    color_table = b''
    if global_color_table_flag:
        color_table = b''.join(map(lambda x: bytes(x[::-1]), global_color_table))
    
    # 4. Graphics Control Extension
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        
        if block_size == 0:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom

    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)