#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HBB', width, height, rng.read(1)[0])
    
    # 3. Global Color Table
    # Number of colors in table
    num_colors = rng.read(1)[0] & 0x07 + 1
    
    # Colors in table
    color_table = []
    for _ in range(num_colors):
        r, g, b = [rng.read(1)[0] for _ in range(3)]
        color_table += [r, g, b]
    
    # 4. Graphics Control Extension
    # If we want one, make sure that there is at least one frame so that we can use the GCE
    has_gce = bool(rng.read(1)[0] % 2) or not len(blocks)
    
    # If we have a GCE, its size will be 6 bytes; otherwise, it's zero-length
    gce = b''
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xf9, 0x04, rng.read(1)[0], 0x00)
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('>HHB', 0, 0, 0)
    
    # 6. Blocks
    blocks = []
    
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0] + 1)
        blocks.append(rng.read(block_size))
        
        total_blocks_size -= block_size
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if len(gce) == 6:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)