#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256

    # Pack all these values together in a single byte
    lsd = struct.pack('<HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    global_color_table_flag = bool(packed_fields & 0b10000000)
    color_resolution = ((packed_fields >> 7) & 0b111) + 1
    sort_flag = bool((packed_fields >> 5) & 0b1)
    global_color_table_size = 2 ** (((packed_fields >> 2) & 0b111) + 1)
    global_color_table = []
    if global_color_table_flag:
        for _ in range(global_color_table_size):
            r, g, b = [rng.read(1)[0] % 256 for _ in range(3)]
            global_color_table.extend([r, g, b])

    # 4. Image Descriptor
    left = 0
    top = 0
    width = max(width, 1)
    height = max(height, 1)
    local_color_table_flag = False
    interlace_flag = bool(packed_fields & 0b1000000)
    sorted_color_table_flag = True
    reserved_bits = 0
    lct_size = 0
    image_descriptor = struct.pack('<HHHHBB', left, top, width, height, local_color_table_flag, interlace_flag,
                                   sorted_color_table_flag)

    # 5. Local Color Table
    if not local_color_table_flag:
        lct_size = 2 ** ((packed_fields & 0b11) + 1)
        for _ in range(lct_size):
            r, g, b = [rng.read(1)[0] % 256 for _ in range(3)]
            global_color_table.extend([r, g, b])

    # 6. Image Data
    # LZW minimum code size is either 2 or 3 depending on the value of the first byte of the image data
    lzw_minimum_code_size = rng.read(1)[0] % 3 + 2
    # The rest of the image data consists of sub-blocks which are made up of one or more data blocks
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if block_size == 0:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)