#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    lsd = struct.pack('>HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            break
        
        data = rng.read(block_size)

        # If this is an uncompressed block, just add it to our list as-is
        if data[0] == 0x00:
            blocks.append(data)
        
        # Otherwise, decompress it
        elif data[0] & 0x01:
            raise Exception("Unsupported LZW method")
        else:
            code_size = data[0] + 1

            # Decode the compressed data
            decoded = [data[i+1] for i in range(len(data)-1)]
            
            # Add the first two codes to the dictionary
            dicts = [[decoded[0], decoded[1]]]
            next_code = len(dicts[0])

            # Start decoding from the third code
            for i in range(2, len(decoded)):
                # Get the current code
                curr_code = decoded[i]

                # Check if we need more entries in our dictionary
                if next_code >= 2 ** code_size:
                    dicts.append([])
                    next_code = 0
                
                # Find the entry that matches the current code
                found = False
                for j in range(next_code):
                    if dicts[-1][j] == curr_code:
                        found = True
                        break
                
                # If there was no match, add the new code to the end of the last dictionary
                if not found:
                    dicts[-1].append(curr_code)
                    next_code += 1
                
                # Add the corresponding value to the output buffer
                if found:
                    blocks.append(bytes([dicts[-2][j]]))
                else:
                    blocks.append(bytes([dicts[-1][j]]))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)