#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def main(argv):
    if len(argv)!= 2:
        print("Usage: {} <file>".format(os.path.basename(__file__)), file=sys.stderr)
        return -1

    with open(argv[1], "rb") as f:
        data = read_data(f)

        for i, (name, value) in enumerate(data.items()):
            print("{}: {}".format(i + 1, name))

            if isinstance(value, bytes):
                print("\tType: String")

                try:
                    s = str(value, encoding="utf-8").strip()

                    if "\n" not in s and len(s) <= 64:
                        print("\tValue: {}".format(repr(s)))
                    else:
                        print("\tValue:")

                        for line in s.splitlines():
                            print("\t\t{}".format(line))
                except UnicodeDecodeError:
                    pass
            elif isinstance(value, int):
                print("\tType: Integer")
                print("\tValue: {:#x}".format(value))
            elif isinstance(value, float):
                print("\tType: Float")
                print("\tValue: {:.10g}".format(value))
            elif isinstance(value, list):
                print("\tType: List")
                print("\tLength: {}".format(len(value)))
            elif isinstance(value, dict):
                print("\tType: Dictionary")
                print("\tLength: {}".format(len(value)))
            else:
                raise ValueError("Unknown type: {!r}".format(type(value).__name__))

    return 0


def read_data(f: BinaryIO) -> dict:
    magic = f.read(4).decode("ascii")

    if magic == "PAK ":
        return _read_pak(f)
    elif magic == "PK3 ":
        return _read_pk3(f)
    else:
        raise ValueError("Invalid magic number: {}".format(magic))


def _read_pak(f: BinaryIO) -> dict:
    header_size, num_files = struct.unpack("<II", f.read(8))

    offsets = []
    names = []
    sizes = []

    for _ in range(num_files):
        offset, size = struct.unpack("<II", f.read(8))
        offsets.append(offset)
        sizes.append(size)

    for _ in range(num_files):
        length = ord(f.read(1)[0])
        names.append(f.read(length).decode("ascii"))

    result = {name: None for name in names}

    for i, name in enumerate(names):
        f.seek(offsets[i] + header_size)
        result[name] = f.read(sizes[i])

    return result


def _read_pk3(f: BinaryIO) -> dict:
    # TODO
    raise NotImplementedError()


if