#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a' + bytes([0x20]) * 7

    # 2. Logical screen descriptor
    width = int.from_bytes(rng.read(2), byteorder='big') % (500 - 16) + 16
    height = int.from_bytes(rng.read(2), byteorder='big') % (500 - 16) + 16
    flags = [
        0b00000001 & int.from_bytes(rng.read(1), byteorder='big'),   # Global color table flag
        0b00000100 & int.from_bytes(rng.read(1), byteorder='big'),   # Color resolution
        0b00001000 & int.from_bytes(rng.read(1), byteorder='big'),   # Sorting method
        0b00010000 & int.from_bytes(rng.read(1), byteorder='big'),   # Size of global color table
        0b00100000 & int.from_bytes(rng.read(1), byteorder='big'),   # Background color index
        0b01000000 & int.from_bytes(rng.read(1), byteorder='big'),   # Pixel aspect ratio
    ]
    bg_color_index = int.from_bytes(rng.read(1), byteorder='big') % len(header[0:4])
    pixel_aspect_ratio = int.from_bytes(rng.read(1), byteorder='big') % 9
    logical_screen_descriptor = struct.pack('>HHBBB', width, height, flags[0] | flags[1] << 1 | flags[2] << 2 | flags[3] << 3 | flags[4] << 4 | flags[5] << 5, bg_color_index, pixel_aspect_ratio)
    
    # 3. Global color table
    if flags[3]:
        global_color_table = []
        for i in range(pow(2, flags[4])):
            r = int.from_bytes(rng.read(1), byteorder='big') % 256
            g = int.from_bytes(rng.read(1), byteorder='big') % 256
            b = int.from_bytes(rng.read(1), byteorder='big') % 256
            global_color_table += [r, g, b]

        global_color_table = bytes(global_color_table)

    else:
        global_color_table = None

    # 4. Image descriptor
    image_left = int.from_bytes(rng.read(2), byteorder='big') % ((width // 8) * 8)
    image_top = int.from_bytes(rng.read(2), byteorder='big') % ((height // 8) * 8)
    image_flags = [
        0b00000001 & int.from_bytes(rng.read(1), byteorder='big'),   # Local color table flag
        0b00000100 & int.from_bytes(rng.read(1), byteorder='big'),   # Interlace flag
        0b00001000 & int.from_bytes(rng.read(1), byteorder='big'),   # Sorting flag
        0b00010000 & int.from_bytes(rng.read(1), byteorder='big'),   # Reserved
        0b00100000 & int.from_bytes(rng.read(1), byteorder='big'),   # Size of local color table
        0b01000000 & int.from_bytes(rng.read(1), byteorder='big'),   # Reserved
    ]
    image_delay = int.from_bytes(rng.read(2), byteorder='big') % 65536
    image_transparent_color_index = int.from_bytes(rng.read(1), byteorder='big') % pow(2, flags[4])
    image_descriptor = struct.pack('>HHBBBBB', image_left, image_top, image_flags[0] | image_flags[1] << 1 | image_flags[2] << 2 | image_flags[3] << 3 | image_flags[4] << 4 | image_flags[5] << 5