#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        # Transparent color flag set, so we need to add an extension block

        # We'll use the same values as in the original example
        packed_fields = b'\x04' + b'\xff\xff\xff'   # Use global color table, transparent color index is last one

        # Write the extension block
        ext = b''
        ext += b'\x21\xf9'    # Application extension
        ext += b'\x04'        # Size of sub-block
        ext += packed_fields  # Packed fields
        ext += b'\x00'        # No background color specified
        out.write(ext)       # Write the entire extension block

    # 5. Image Descriptor
    image_descriptor = b''
    image_descriptor += b','     # Separator
    image_descriptor += lsd      # LSD from above
    image_descriptor += b';'     # Terminator

    # 6. Image data
    image_data = b''
    image_data += b'\x2c'         # Image separator
    image_data += b'\x00' * 4     # Left position, top position, width, height
    image_data += b'\x00'          # Flags: no local color table, interlaced
    image_data += b'\x00'          # Code size (no compression)
    image_data += b'\x00' * 256   # Local color table (all black)
    image_data += b'\x21\xff\x0b'  # Graphic control extension
    image_data += b'\x4e'         # Block size
    image_data += b'\x45\x54\x53\x43\x41\x50\x45\x32\x2e\x30\x03\x01\x00\x00\x00'  # Magic value
    image_data += b'\x21\xfe\x17'  # Plain text extension
    image_data += b'\x08'         # Block size
    image_data += b'\x01'         # Text grid left
    image_data += b'\x01'         # Text grid top
    image_data += b'\x00'         # Text grid width
    image_data += b'\x00'         # Text grid height
    image_data += b'\x00'         # Cell width
    image_data += b'\x00'         # Cell height
    image_data += b'\x00'         # Fg color index
    image_data += b'\x00'         # Bg color index
    image_data += b'\x00'         # Character cell count
    image_data += b'\x00' * 256   # Characters

    # 7. End of File
    end = b';'

    # Write everything together
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(image_descriptor)
    out.write(image_data)
    out.write(end)


