#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header
    header = b'GIF89a'

    # Logical Screen Descriptor (LSD)
    width = rng.read(2)
    height = rng.read(2)
    flags = rng.read(1)[0] & ~0b11110000 | 0b10000000  # Global Color Table Flag set
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = width + height + bytes([flags]) + \
          bytes([background_color_index]) + bytes([pixel_aspect_ratio])

    # Global Color Table
    global_color_table_entries = rng.read(1)[0]
    global_color_table_size = 1 << ((global_color_table_entries >> 4) + 1)
    color_table = b''
    for _ in range(global_color_table_size):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        color_table += bytes([r, g, b])

    # Graphics Control Extension (GCE)
    block_size = rng.read(1)[0]
    assert block_size == 4
    packed_fields = rng.read(1)[0]
    disposal_method = (packed_fields >> 2) & 0b111
    user_input_flag = bool((packed_fields >> 1) & 1)
    transparent_color_flag = bool(packed_fields & 1)
    delay_time = rng.read(2)
    transparent_color_index = rng.read(1)[0] if transparent_color_flag else None
    gce = b''
    gce += bytes([block_size])
    gce += bytes([(disposal_method << 2) |
                  (user_input_flag << 1) |
                  1 - int(transparent_color_flag)])
    gce += delay_time
    gce += bytes([transparent_color_index or 0])

    # Image Descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    local_color_table_flag = rng.read(1)[0]
    interlace_flag = rng.read(1)[0]
    sort_flag = rng.read(1)[0]
    reserved = rng.read(1)[0]
    local_color_table_size = 1 << ((local_color_table_flag >> 4) + 1)
    local_color_table = b''
    for _ in range(local_color_table_size):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        local_color_table += bytes([r, g, b])
    image_descriptor = left + top + width + height + bytes([local_color_table_flag]) + \
                       bytes([interlace_flag]) + bytes([sort_flag]) + bytes([reserved]) + \
                       local_color_table

    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0]

    # Image Data
    # 1. Initial Code Size
    initial_code_size = rng.read(1)[0]
    # 2. Clear Code
    clear_code = 1 << initial_code_size
    # 3. End of Information Code
    end_of_information_code = clear_code + 1
    # 4. Maximum Code Size
    maximum_code_size = 12
    # 5. Block Size
    block_size = 2 ** (maximum_code_size - initial_code_size + 1)
    # 6. Blocks
    blocks = []
    while True:
        # Generate a random block
        block = [clear_code]
        while len(block)!= block_size and block[-1]!= end_of_information_code:
            # Read a byte from /dev/urandom as an unsigned integer
            n = struct.unpack('>H', rng.read(2))[0] % 2**initial_code_size

            # If it's not a valid code, we need to add it to our table
            if n >= len(block):
                block.append(n)
            elif n < len(block):
                block[-1] = n

        # Add EOI if necessary
        if block[-1]!= end_of_            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)