#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = b''
    for _ in range(num_colors):
        r, g, b = [rng.read(1)[0] for i in range(3)]
        color_table += struct.pack('BBB', r, g, b)

    # 4. Graphics Control Extension
    # This is optional, but let's add one anyway
    gce = b'\x21\xF9\x04' + rng.read(3)

    # 5. Image Descriptor
    # Frame left position, frame top position, frame width, frame height, packed fields
    image_descriptor = struct.pack('>HHHHB', 0, 0, width, height, rng.read(1)[0])

    # 6. Image Data
    # Let's use the same palette as before, but now we'll just fill with random pixels
    image_data = b''
    for y in range(height):
        for x in range(width):
            image_data += struct.pack('B', rng.read(1)[0] % len(color_table))

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)