#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    lsd = struct.pack('>HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    global_color_table_size = 2 << ((packed_fields & 7) + 1)
    global_color_table = [rng.read(3) for _ in range(global_color_table_size)]

    gct = b''
    if len(global_color_table):
        gct += struct.pack('B', global_color_table_size - 1)

        for rgb in global_color_table:
            gct += rgb

    # 4. Image Descriptor
    image_left_position = 0
    image_top_position = 0
    image_width = width
    image_height = height
    local_color_table_flag = bool(gct)
    interlace_flag = False
    sort_flag = True
    reserved_bits = 0b00000000
    lzw_minimum_code_size = 2 ** (rng.read(1)[0] + 1)

    id = struct.pack(
        '>HHHHBB',
        image_left_position,
        image_top_position,
        image_width,
        image_height,
        local_color_table_flag,
        interlace_flag,
        sort_flag,
        reserved_bits,
        lzw_minimum_code_size
    )

    # 5. Local Color Table
    local_color_table_size = 2 << ((packed_fields >> 4) + 1)
    local_color_table = [rng.read(3) for _ in range(local_color_table_size)]

    lct = b''
    if len(local_color_table):
        lct += struct.pack('B', local_color_table_size - 1)

        for rgb in local_color_table:
            lct += rgb

    # 6. Graphics Control Extension
    graphics_control_extension_label = b'\x21\xF9\x04'
    block_size = 4
    flags = 0b00000000
    delay_time = int((rng.read(1)[0] / 256) * 100)
    transparent_color_index = rng.read(1)[0]
    terminator = b'\x00'

    gce = struct.pack(
        '>BBBBHB',
        graphics_control_extension_label[0],
        graphics_control_extension_label[1],
        block_size,
        flags,
        delay_time,
        transparent_color_index,
        terminator
    )

    # 7. Plain Text Extension
    plain_text_extension_label = b'\x21\x01\x0F'
    text_grid_left = 0
    text_grid_top = 0
    text_grid_width = width // 2
    text_grid_height = height // 2
    cell_width = 2
    cell_height = 2
    foreground_color_index = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    character_cell_value = ord('A')
    terminator = b'\x00'

    pte = struct.pack(
        '>BBBBHHBBB',
        plain_text_extension_label[0],
        plain_text_extension_label[1],
        plain_text_extension_label[2],
        text_grid_left,
        text_grid_top,
        text_grid_width,
        text_grid_height,
        cell_width,
        cell_height,
        foreground_color_index,
        background_color_index,
        character_cell_value,
        terminator
    )

    # 8. Application Extension
    application_extension_label = b'\xFF\x0B'
    application_identifier = b'NETSCAPE2.0'
    application_auth_code = b'\<PASSWORD>'
    subblock_size = 3
    loop_count = 1

    ae = struct.pack(
        '>BBBB%dsBHHB' % len(application_identifier),
        application_extension_label[0],
        application_extension_label[1],
        subblock_size,
        application_identifier,
        0x00,
        application_auth_code[0],
        application_auth_code[1],
        loop_count
    )

    # 9. Comment Extension
    comment_extension_label = b'\xFE\x01'
    comment_subblock_size = rng.read(1)[