#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
    out: BinaryIO,
    rng: BinaryIO,
):
    # Header (6 bytes)
    header = b'GIF89a'
    
    # Logical Screen Descriptor (7 bytes)
    lsd = bytearray([
        255 & rng.read(1)[0],   # Width
        255 & rng.read(1)[0],   # Height
        0 | rng.read(1)[0],     # Flags
        0 | rng.read(1)[0],     # Background color index
        0 | rng.read(1)[0],     # Pixel aspect ratio
    ])
    
    # Global Color Table (3 bytes)
    color_table = bytearray([
        0 | rng.read(1)[0],      # Size of global color table
        0 | rng.read(1)[0],      # Background color index
        0 | rng.read(1)[0],      # Pixel aspect ratio
    ])
    
    # Graphics Control Extension (8 bytes)
    gce = None
    if rng.read(1)[0]:
        gce = bytearray([
            0x21,                   # Introducer
            0xf9,                   # Graphic control label
            4,                      # Block size
            rng.read(1)[0],         # Flags
            0 | rng.read(1)[0],     # Delay time
            0 | rng.read(1)[0],     # Transparent color index
            0,                      # Terminator
        ])
    
    # Image Descriptor (10 bytes)
    image_descriptor = bytearray([
        0x2c,                       # Image separator
        0 | rng.read(1)[0],         # Left position
        0 | rng.read(1)[0],         # Top position
        255 & rng.read(1)[0],       # Width
        255 & rng.read(1)[0],       # Height
        0 | rng.read(1)[0],         # Flags
        0 | rng.read(1)[0],         # Local color table index
        
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0] + 1)
        blocks.append(rng.read(block_size))
        
        total_blocks_size -= block_size
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)