#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = struct.pack('>HBBBBB', width, height, rng.read(1)[0], 0b11111000 | rng.read(1)[0], 0, 0)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = packed + bytes([background_color_index, pixel_aspect_ratio])
    
    # 3. Global Color Table
    global_color_table_flag = bool(rng.read(1)[0] & 0b10000000)
    color_resolution = ((rng.read(1)[0] >> 4) & 0b111) or 1
    sort_flag = bool(rng.read(1)[0] & 0b10000000)
    global_color_table_size = (rng.read(1)[0] & 0b1111) or 2**global_color_table_flag
    global_color_table = [rng.read(3) for _ in range(global_color_table_size)]
    color_table = b''
    for rgb in global_color_table:
        color_table += rgb
    
    # 4. Graphics Control Extension
    has_gce = bool(rng.read(1)[0] & 0b10000000)
    if has_gce:
        delay_time = struct.unpack('>H', rng.read(2))[0]
        transparent_color_index = rng.read(1)[0]
        disposal_method = rng.read(1)[0]
        user_input_flag = bool(rng.read(1)[0] & 0b10000000)
        transparency_flag = bool(rng.read(1)[0] & 0b10000000)
        gce = struct.pack('>BBBBBB', 0xF9, 4, *struct.unpack('>BBBB', rng.read(4)))
    else:
        gce = None
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        
        if block_size == 0:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom

    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)