#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    # Generate some random data
    data = []
    while True:
        length = min((rng.read(1)[0] % 255) + 1, 255)
        
        if not length:
            break
            
        data += [rng.read(length)]
        
    # Compress that data with LZW
    code_size = lzw_minimum_code_size + 1
    codes = {chr(i): i for i in range(2 ** code_size)}
    next_code = len(codes)
    clear_code = next_code
    end_of_information_code = next_code + 1
    codes[None] = clear_code
    codes[''] = end_of_information_code
    current_code = ''
    blocks = []
    block = []
    for datum in data:
        for char in datum:
            new_current_code = current_code + chr(char)
            
            if new_current_code in codes:
                current_code = new_current_code
                
            else:
                # Add the new string to our dictionary
                codes[new_current_code] = next_code
                
                # Add this string to the current block
                block.append(codes[current_code])
                
                # Start a new block
                current_code = chr(char)
                
        if current_code!= '':
            # Add the last string to the current block
            block.append(codes[current_code])
            current_code = ''
            
        if len(block) >= 255:
            # Flush the current block
            blocks.append(bytes([len(block)]) + bytes(block))
            block = []
            
    if len(block)!= 0:
        # Flush the final block
        blocks.append(bytes([len(block)]) + bytes(block))
        
    trailer = bytes([end_of_information_code])
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if len(gce) == 6:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)