#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    lsd = struct.pack('>HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    left, top, width, height = struct.unpack('>HHHH', rng.read(8))
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved = rng.read(2)
    local_color_table_size = rng.read(1)[0]
    local_color_table = rng.read(3 * local_color_table_size)
    lzw_minimum_code_size = rng.read(1)[0]

    image_descriptor = struct.pack('>HHHHBB', left, top, width, height, local_color_table_flag, interlace_flag,
                                  sort_flag) + reserved + bytes([local_color_table_size]) + local_color_table

    # 6. Image Data
    # Let's use an LZW compression scheme to compress our data
    # We'll be using a minimum code size of 2 bits (i.e., up to 4 different values per byte)
    # For this we need to know how many unique values there are in our input data
    # To do that, let's count them!
    value_counts = {}
    for i in range(width*height):
        val = ord(image_data[i])
        try:
            value_counts[val] += 1
        except KeyError:
            value_counts[val] = 1

    # Now let's build our dictionary by assigning codes to all those values
    dict_values = sorted(value_counts.keys())
    dict_codes = list(range(len(dict_values)))
    dict_codes.reverse()

    # And now let's assign codes to all other values not already assigned
    while len(dict_codes) < 2**lzw_minimum_code_size:
        new_code = max(dict_codes) + 1

        # If we've run out of values to assign codes to...
        if not dict_values:
            break
        
        # Assign the next available code to the first unassigned value
        dict_codes.append(new_code)
        dict_values.pop(0)

    # Now let's build our actual dictionary from these codes and values
    dict_entries = []
    for k, v in zip(dict_codes, dict_values):
        dict_entries.append((k, v))

    # Finally, let's build our compressed data
    blocks = [struct.pack('B', x) for x in dict_entries]
    for i in range(width*height):
        val = ord(image_data[i])
        try:
            code = dict_codes.index(val)
        except ValueError:
            raise Exception("Value not found")

        # Find the length of the longest matching sequence
        match_length = 1
        for j in reversed(range(i)):
            prev_val = ord(image_data[j])

            try:
                if dict_codes[prev_val] == code:
                    match_length += 1

                    if match_length > lzw_minimum_code_size:
                        break

                else:
                    match_length = 1
            
            except IndexError:
                pass

        # Add the corresponding entry to the output stream
        block_size = min(match_length - 1, 255)
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)