#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = None
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
        
    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, 0, 0)
    
    # 6. Local Color Table (optional; let's use the same one as in step 3)
    local_color_table = b''
    if num_colors > 1:
        local_color_table = struct.pack('>B', num_colors) + color_table
        
    
    # 7. Image Data
    data = b''
    while len(data) < width * height:
        data += rng.read((width * height - len(data)) // 2)
    
    
    # 8. Trailer
    trailer = b'\x3B'
    
    
    # Write everything to output
    out.write(header)
    out.write(lsd)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)
    if local_color_table!= '':
        out.write(local_color_table)
    out.write(data)
    out.write(trailer)

