#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0] & 1)
    gce = b''

    if has_gce:
        gce += b'\x21\xf9\x04'  # Application extension marker
        gce += b'\x08'         # Block size
        gce += b'\xff'         # Flags: user input is required
        gce += b'\x00'         # Delay time in hundredths of seconds
        gce += b'\x00'         # Transparent color index
        gce += b'\x00'         # Terminator

    # 5. Image Descriptor
    image_descriptor = b'\x2c'     # Image separator
    image_descriptor += struct.pack('>HHBBB',
                                     width, height,
                                     0,      # No local color table
                                     0,      # Interlace bit is zero
                                     0       # Sort flag is zero
                                 )

    # 6. Image Data
    image_data = []
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            break

        image_data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to the output stream
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(image_data) + trailer)


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('./test.gif', 'wb') as f:
            generate_random_gif(rng, f)