#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
     # 2. Logical screen descriptor
    width = rng.read(1)[0] % 64 + 1
    height = rng.read(1)[0] % 64 + 1

    color_table_size = (rng.read(1)[0] & 7) + 1
    global_color_table = [None for _ in range(pow(2, color_table_size))]
    for i in range(len(global_color_table)):
        global_color_table[i] = bytes([
            rng.read(1)[0],
            rng.read(1)[0],
            rng.read(1)[0],
            0x80])

    background_index = rng.read(1)[0] % len(global_color_table)
    pixel_aspect_ratio = rng.read(1)[0] % 5

    logical_screen_descriptor = struct.pack('<BHHBBB',
        0x2C,
        width,
        height,
        color_table_size * pow(2, color_table_size),
        background_index,
        pixel_aspect_ratio)

    # 3. Global color table
    if color_table_size > 1:
        out.write(logical_screen_descriptor)

        for entry in global_color_table:
            out.write(entry)

    # 4. Image descriptor
    image_separator = b'\x2C'
    left = rng.read(1)[0] % width
    top = rng.read(1)[0] % height
    width = rng.read(1)[0] % min((width - left), 64) + 1
    height = rng.read(1)[0] % min((height - top), 64) + 1
    local_color_table_flag = bool(rng.read(1)[0] % 2)
    interlace_flag = bool(rng.read(1)[0] % 2)
    sort_flag = bool(rng.read(1)[0] % 2)
    reserved = rng.read(1)[0] % 2
    lct_size = (rng.read(1)[0] & 7) + 1
    lct = []
    for i in range(lct_size):
        lct += [bytes([
            rng.read(1)[0],
            rng.read(1)[0],
            rng.read(1)[0],
            0x80])]

    local_color_table = None
    if local_color_table_flag and lct!= []:
        local_color_table = lct

    lzw_min_code_size = rng.read(1)[0] % 9 + 1

    image_descriptor = struct.pack('<BBHHBBBB',
        image_separator,
        left,
        top,
        width,
        height,
        local_color_table_flag << 7 | interlace_flag << 6 | \
                sort_flag << 5 | reserved,
        lct_size * pow(2, lct_size),
        lzw_min_code_size)

    # 5. Local color table
    if local_color_table is not None:
        for entry in local_color_table:
            out.write(entry)

    # 6. LZW code stream
    clear_code = 1 << lzw_min_code_size
    end_of_information_code = clear_code + 1
    next_code = clear_code + 2
    bit_buffer = bytearray()
    output_stream = bytearray()
    previous_byte = None

    while True:
        new_byte = rng.read(1)[0]
        if new_byte == 0xFF:
            break

        if new_byte >= next_code:
            output_stream.append(new_byte)
            continue

        bit_buffer.extend(struct.pack('>H', new_byte))

        while len(bit_buffer) >= 16:
            buffer_word = int.from_bytes(bit_buffer[:16], 'big')

            if buffer_word < next_code:
                output_stream.append(buffer_word)

                if previous_byte is not None:
                    output_stream.append(previous_byte)

                previous_byte = buffer_word
            else:
                if previous_byte is not None:
                    output_stream.append(previous_byte)

                    if buffer_word == next_code:
                        output_stream.append(clear_code)

                        previous_byte = None
                        next_code = clear_code + 2
                    elif buffer_word == next_code + 1:
                        output_stream.append