#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), byteorder='big') % (50 + 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='big') % (50 + 1) + 1

    flags = bytes([
        0b00000000 | # Global Color Table Flag
        0b00000000 | # Sorting Flag
        0b00000000 | # Size of Global Color Table
        0b00000000    # Background Color Index
    ])

    background_color_index = bytes([int.from_bytes(rng.read(1), 'big')])
    pixel_aspect_ratio = bytes([int.from_bytes(rng.read(1), 'big')])

    logical_screen_descriptor = \
        width.to_bytes(2, byteorder='big') + \
        height.to_bytes(2, byteorder='big') + \
        flags + \
        background_color_index + \
        pixel_aspect_ratio

    # 3. Global Color Table
    global_color_table = [
        rng.read(3) for _ in range(768 // 3)
    ]

    # 4. Image Descriptor
    image_separator = bytes([0x2C])
    left = int.from_bytes(rng.read(2), byteorder='big') % (width - 1)
    top = int.from_bytes(rng.read(2), byteorder='big') % (height - 1)
    flags = bytes([
        0b00000000 | # Local Color Table Flag
        0b00000000 | # Interlace Flag
        0b00000000 | # Sorting Flag
        0b00000000 | # Reserved
        0b00000000    # Size of Local Color Table
    ])

    local_color_table = []
    if len(global_color_table):
        local_color_table += [
            rng.read(3) for _ in range((len(global_color_table) * 10) // 7)
        ]

    image_descriptor = \
        image_separator + \
        left.to_bytes(2, byteorder='big') + \
        top.to_bytes(2, byteorder='big') + \
        width.to_bytes(2, byteorder='big') + \
        height.to_bytes(2, byteorder='big') + \
        flags + \
        bytes([(len(local_color_table) << 4) & 0xFF]) + \
        bytes([len(local_color_table) >> 4])

    # 5. Image Data
    lzw_min_code_size = 2 ** int.from_bytes(rng.read(1), 'big')
    lzw_max_code_size = min(lzw_min_code_size + 1, 12)
    lzw_clear_code = 1 << lzw_min_code_size
    lzw_end_of_information_code = lzw_clear_code + 1
    lzw_first_free_code = lzw_end_of_information_code + 1
    lzw_current_code_size = lzw_min_code_size + 1
    lzw_next_code_size = None

    lzw_dictionary = {
        i: bytes([i]) for i in range(256)
    }
    lzw_dictionary[lzw_clear_code] = b''
    lzw_dictionary[lzw_end_of_information_code] = b'\x00'

    lzw_buffer = bytes()
    lzw_buffer_length = 0

    lzw_data = b''
    while True:
        if not lzw_next_code_size:
            lzw_next_code_size = int.from_bytes(rng.read(1), 'big')

        if lzw_next_code_size < lzw_current_code_size and lzw_next_code_size <= lzw_max_code_size:
            lzw_current_code_size = lzw_next_code_size

            lzw_next_code_size = None
            continue
        
        code = int.from_bytes(rng.read(1), 'big')
        if code == lzw_end_of_information_code:
            break

        if code == lzw_clear_code:
            lzw_dictionary