#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed_flags = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBB', width[0]*256+width[1], height[0]*256+height[1], packed_flags, background_color_index, pixel_aspect_ratio)

    # 3. Color Table
    color_table_size = 2**(packed_flags&7)+1
    color_table = b''
    for i in range(color_table_size):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        color_table += struct.pack('BBB', r, g, b)

    # 4. Graphics Control Extension
    has_gce = bool(rng.read(1)[0]&1)
    if has_gce:
        delay_time = rng.read(2)
        transparent_color_index = rng.read(1)[0]
        disposal_method = rng.read(1)[0]
        user_input_flag = bool(rng.read(1)[0]&1)
        has_transparent_color = bool(rng.read(1)[0]&1)
        gce = struct.pack('>BHB', 0xF9, delay_time[0]*256+delay_time[1], transparent_color_index) + \
              struct.pack('>BB', disposal_method, int(user_input_flag)<<2 | int(has_transparent_color)<<1)
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            break
        
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)