#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0] % 0b11111111  # Only use first bit for now
    background_color_index = rng.read(1)[0] & 0b00001111
    pixel_aspect_ratio = rng.read(1)[0] & 0b00001111
    
    lsd = struct.pack('>HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    global_color_table_flag = bool(packed_fields & 0b10000000)
    global_color_table_size = ((packed_fields >> 7) & 0b111) + 1
    
    if global_color_table_flag:
        color_table = bytes()
        
        for i in range(global_color_table_size ** 2):
            r, g, b = [rng.read(1)[0] & 0b1111111 for _ in range(3)]
            
            color_table += struct.pack('BBB', r, g, b)
            
    else:
        color_table = b''
    
    # 4. Graphic Control Extension
    gce = None
    
    if rng.read(1)[0] > 0:  # Probability that there will be an extension
        gce = b''
        
        # Block size
        gce += struct.pack('BB', 0xF9, 4)
        
        # Flags
        flags = rng.read(1)[0]
        gce += struct.pack('B', flags)
        
        # Delay time
        delay_time = rng.read(2)
        gce += delay_time
        
        # Transparent color index
        transparent_color_index = rng.read(1)[0]
        gce += struct.pack('B', transparent_color_index)
        
        # Block terminator
        gce += struct.pack('B', 0)
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    
    local_color_table_flag = bool(flags & 0b10000000)
    interlace_flag = bool(flags & 0b01000000)
    sort_flag = bool(flags & 0b00100000)
    reserved = rng.read(1)[0]  # Must always be zero
    local_color_table_size = ((flags >> 3) & 0b111) + 1
    
    # We don't care about these values since we're generating a random GIF
    local_color_table = bytes()
    for i in range(local_color_table_size ** 2):
        r, g, b = [rng.read(1)[0] & 0b1111111 for _ in range(3)]

        local_color_table += struct.pack('BBB', r, g, b)
    
    image_descriptor = struct.pack('>HHHHBBB', left, top, width, height, local_color_table_flag, interlace_flag,
                                   sort_flag, reserved)
    
    # 6. Image Data
    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0] & 0b00001111
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        
        if block_size == 0:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom

    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)