#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0] & 1)
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0] & 1)
    interlace_flag = bool(rng.read(1)[0] & 1)
    sort_flag = bool(rng.read(1)[0] & 1)
    reserved_flags = rng.read(1)[0]
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    image_descriptor = struct.pack(
        '>BHHHHBB',
        0x2C, left, top, width, height,
        local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved_flags,
        lzw_minimum_code_size
    )

    # 6. Image Data
    # Let's use LZW compression with a random minimum code size
    # We'll also write some random data in there too
    # This is not very efficient but it will do the trick
    # TODO: Use an actual compressor instead?
    # TODO: Make this more realistic by actually generating images
    #       with random content rather than just random data
    max_length = int((width / 2) * (height / 2) * 3)
    length = rng.read(1)[0] % max_length
    data = rng.read(length)
    blocks = []
    while True:
        try:
            prefix = data[:2]
            suffix = data[2:].decode()
        except UnicodeDecodeError:
            break

        if prefix == b'\xFF\x00':
            break

        if prefix!= b'\x00\x00':
            raise ValueError("Invalid LZW prefix")

        if suffix not in lzw_decompress_table:
            lzw_decompress_table[suffix] = len(lzw_decompress_table)

        data = bytes([lzw_decompress_table[suffix]]) + data[2:]
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)