#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Set color resolution to 7 bits per pixel
    color_res = 0x7 << 5

    # Set global color table flag
    gct_flag = 0x1 << 7

    # Set sort flag to false
    sort_flag = 0x0

    # Set size of global color table to 256 colors
    gct_size = 0xFF

    # Set background color index to 0
    bg_color_idx = 0

    # Set pixel aspect ratio to 0
    pixel_aspect_ratio = 0

    # Pack logical screen descriptor
    lsd = struct.pack(
        '>HHBBBHBB', 
        width, height,
        color_res | gct_flag | sort_flag,
        gct_size,
        bg_color_idx,
        pixel_aspect_ratio
    )

    # Write it all together
    out.write(header + lsd)

    # 3. Global Color Table
    # Generate 256 RGB triplets
    rgb_triplets = [struct.unpack('BBB', rng.read(3)) for _ in range(gct_size)]

    # Convert them to hex strings
    hexes = ['{:02X}{:02X}{:02X}'.format(*rgb) for rgb in rgb_triplets]

    # Join them with commas
    gct = ','.join(hexes).encode()

    # Write it
    out.write(b';' + gct)

    # 4. Image Descriptor
    # Set left position to 0
    left = 0

    # Set top position to 0
    top = 0

    # Set width to the same as the logical screen's width
    img_width = width

    # Set height to the same as the logical screen's height
    img_height = height

    # Set local color table flag to true
    lct_flag = 0x1 << 7

    # Set interlace flag to false
    interlace_flag = 0x0

    # Set sorted flag to false
    sorted_flag = 0x0

    # Set reserved flags to 0
    reserved = 0x0

    # Set LCT size to 256 colors
    lct_size = 0xFF

    # Set palette start to 0
    palette_start = 0

    # Set palette length to 256 colors
    palette_length = 0xFF

    # Set background color index to 0
    bg_color_idx = 0

    # Set image descriptor
    id = struct.pack(
        '>HHHHBBBB',
        left, top,
        img_width, img_height,
        lct_flag | interlace_flag | sorted_flag,
        reserved,
        lct_size,
        palette_start,
        palette_length,
        bg_color_idx
    )

    # Write it
    out.write(id)

    # 5. Local Color Table
    # Repeat steps 3-4 for the local color table
    out.write(b';' + gct)

    # 6. Graphics Control Extension
    # Set block size to 0
    block_size = 0

    # Set graphics control label to 0xF9
    label = 0xF9

    # Set delay time to 0
    delay_time = 0

    # Set transparent color index to 0
    trans_color_idx = 0

    # Set terminator to 0
    term = 0

    # Pack it
    gc = struct.pack(
        '>BHBxB',
        block_size,
        label,
        delay_time,
        trans_color_idx,
        term
    )

    # Write it
    out.write(gc)

    # 7. Application Extension
    # Set application extension label to 0xFF
    app_label = 0xFF

    # Set application identifier to "NETSCAPE"
    app_identifier = 'NETSCAPE'.encode()

    # Set application authentication code to 0x01
    auth_code = 0x01

    # Set subblock size to 3
    subblock_size = 3

    # Set loop count to 0
    loop_count = 0

    # Pack it
    ae = struct.pack(
        '>BB8sBBBH',
        app_label,
        len(app_identifier),
        app_identifier,
        subblock_size,
        0x1,
        loop_count
    )

    # Write it
    out.write(ae)