#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = min((rng.read(1)[0] & 0x7F) + 1, 256)
    global_color_table = [rng.read(3) for _ in range(global_color_table_size)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBBBBB',
        0b10000000 | rng.read(1)[0], # Local color table flag set to true, interlace flag set to false, sort flag set to false
        rng.read(1)[0],               # Left position of the image
        rng.read(1)[0],               # Top position of the image
        width,                        # Width of the image
        height,                       # Height of the image
        0,                            # No local color table used
        rng.read(1)[0],               # LZW minimum code size
        0                             # No LZW subblock data included
    )
    
    # 5. Image Data
    # Let's use a simple LZW compression scheme that uses only one byte per pixel
    # We'll just compress all pixels as black with full opacity
    compressed_image_data = b'\x00' * len(lsd) + b'\xFF\x00' * ((width*height)//2)
    
    # 6. Trailer
    trailer = b';'
    
    # Write everything to output stream
    out.write(header)
    out.write(lsd)
    if global_color_table_size > 0:
        out.write(bytes([global_color_table_size]))
        for entry in global_color_table:
            out.write(entry)
    out.write(image_descriptor)
    out.write(compressed_image_data)
    out.write(trailer)

