#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # First byte is the minimum code size
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = ord(rng.read(1))

        if not block_size:
            break
        
        blocks.append(rng.read(block_size))
    
    # Now let's compress them!
    compressed_blocks = [lzw_minimum_code_size] + encode_lzw(b''.join(blocks))

    # Finally, write everything to the output stream
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce:
        out.write(gce)
    out.write(image_descriptor)
    out.write(bytes([len(compressed_blocks)]))
    for block in compressed_blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)


# Encodes data using LZW compression
def encode_lzw(data):
    dictionary = {chr(i): i for i in range(256)}
    w = chr(data[0])
    result = [w]
    for c in data[1:]:
        wc = w + c
        if wc in dictionary:
            w = wc
        else:
            result.append(dictionary[w])
            dictionary[wc] = len(dictionary)
            w = c
    if w!= '':
        result.append(dictionary[w])
    return result


if