#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    lsd = bytearray()
    lsd += struct.pack('>H', width)
    lsd += struct.pack('>H', height)
    
    # Random packed field
    lsd += struct.pack('B', rng.read(1)[0])
    
    # Random background color index
    lsd += struct.pack('B', rng.read(1)[0])
    
    # Random pixel aspect ratio
    lsd += struct.pack('B', rng.read(1)[0])
    
    # 3. Global Color Table
    num_colors = rng.read(1)[0] & 0x07 + 1
    global_color_table = []
    for i in range(num_colors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        global_color_table.append((r, g, b))
    
    # 4. Image Descriptor
    image_descriptor = bytearray()
    
    # Random left position of image
    image_descriptor += struct.pack('>H', rng.read(2)[0] % (width - 100))
    
    # Random top position of image
    image_descriptor += struct.pack('>H', rng.read(2)[0] % (height - 100))
    
    # Random width of image
    image_descriptor += struct.pack('>H', rng.read(2)[0] % (width // 5))
    
    # Random height of image
    image_descriptor += struct.pack('>H', rng.read(2)[0] % (height // 5))
    
    # Packed field
    packed_field = rng.read(1)[0]
    
    # Set local color table flag
    if num_colors!= 1 << ((packed_field >> 7) & 1):
        packed_field |= 0x80
    
    # Set interlace flag
    if bool(rng.read(1)[0]):
        packed_field |= 0x40
    
    # Set sort flag
    if bool(rng.read(1)[0]):
        packed_field |= 0x20
    
    # Set reserved bits
    packed_field |= rng.read(1)[0] & 0x0c
    
    # Set local color table size
    if num_colors <= 2:
        packed_field |= 0x02
    elif num_colors <= 4:
        packed_field |= 0x04
    else:
        packed_field |= 0x08
    
    image_descriptor += struct.pack('B', packed_field)
    
    # Local color table
    if num_colors!= 1 << ((packed_field >> 7) & 1):
        color_table = bytearray()
        for rgb in global_color_table[:num_colors]:
            color_table += struct.pack('BBB', *rgb)
    else:
        color_table = b''
    
    # 5. Graphics Control Extension
    gce = b''
    
    # Only add GCE if there is at least one transparent color
    if any([i[3] == 0 for i in global_color_table]):
        gce += b'\x21\xf9\x04'
        
        # Random delay time
        gce += struct.pack('>H', rng.read(2)[0] % 100)
        
        # Transparent color index
        gce += struct.pack('B', rng.read(1)[0] % num_colors)
        
        # Block terminator
        gce += b'\x00'
    
    # 6. Blocks
    blocks = []
    
    # Calculate how many blocks we need to fill up all the space available
    # after the LSD, the color table, and the GCE
    total_blocks_size = \
        1024 - len(header) - len(lsd) - len(color_table) - len(gce) - \
        len(image_descriptor)
    
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0] + 1)
        blocks.append(rng.read(block_size))
        
        total_blocks_size -= block_size
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if len(gce) == 6:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)