#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    lsd = struct.pack('>HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None
    
    # 5. Image Descriptor
    left, top, width, height = struct.unpack('>HHHH', rng.read(8))
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved = rng.read(3)  # Reserved field, should be all zeroes
    local_color_table_size = rng.read(1)[0]
    local_color_table = rng.read((local_color_table_size ** 2) * 3)  # Local color table data
    lzw_minimum_code_size = rng.read(1)[0]  # Minimum code size used by LZW compression algorithm

    image_descriptor = struct.pack('>BBHHBBBB', 0x2c, lzw_minimum_code_size, left, top, width, height, local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5, local_color_table_size)

    # 6. Image Data
    blocks = []
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break
        
        block = rng.read(block_size)
        blocks.append(struct.pack('B', len(block)))
        blocks.append(block)
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)