#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(1), 'big') % (2**10 + 1)
    height = int.from_bytes(rng.read(1), 'big') % (2**10 + 1)

    flags = bytes([
        0b10000000 | # Global Color Table Flag
        0b00000000 | # Sorting Flag
        0b00000000 | # Size of Global Color Table
        0b00000000    # Background Color Index
    ])

    bgColorIndex = bytes([int.from_bytes(rng.read(1), 'big') & 0x0F])

    pixelAspectRatio = bytes([0x00])

    lsd = [width, height] + list(flags) + list(bgColorIndex) + list(pixelAspectRatio)

    for i in range(4):
        lsd[i+5] = lsd[i+5] ^ lsd[i] ^ lsd[i+1] ^ lsd[i+2] ^ lsd[i+3] ^ lsd[i+4]

    lsd = bytes(lsd)

    # 3. Global Color Table
    globalColorTable = []

    for _ in range(2**(lsd[7]+1)):
        r = int.from_bytes(rng.read(1), 'big') & 0xFF
        g = int.from_bytes(rng.read(1), 'big') & 0xFF
        b = int.from_bytes(rng.read(1), 'big') & 0xFF

        globalColorTable += [r,g,b]

    globalColorTable = bytes(globalColorTable)

    # 4. Image Descriptor
    imageDescriptor = bytes([
        0b10000000 | # Local Color Table Flag
        0b00000000 | # Interlace Flag
        0b00000000 | # Sorting Flag
        0b00000000 | # Reserved
        0b00000000 | # Size of Local Color Table
        0b00000000    # Left Position
    ])

    leftPosition = bytes([(width >> 8) & 0xFF])

    topPosition = bytes([(height >> 8) & 0xFF])

    imageDescriptor += leftPosition + topPosition

    for i in range(4):
        imageDescriptor[i+6] ^= imageDescriptor[i] ^ imageDescriptor[i+1] ^ imageDescriptor[i+2] ^ imageDescriptor[i+3] ^ imageDescriptor[i+4]

    imageDescriptor += bytes([0x00, 0x00])

    # 5. Image Data
    imageData = bytearray()

    for y in range(height):
        for x in range(width):
            if ((y // 2 * 2 == y and x // 2 * 2 == x) or
                (y // 2 * 2!= y and x // 2 * 2!= x)):

                imageData.append((
                    int.from_bytes(rng.read(1), 'big') &
                    0x0F
                ) << 4)

            else:
                imageData.append(((
                    int.from_bytes(rng.read(1), 'big') &
                    0x0F
                ) << 4) | (
                    int.from_bytes(rng.read(1), 'big') &
                    0x0F
                ))

    # 6. Trailer
    trailer = b';'

    # Write everything to output
    out.write(header)
    out.write(lsd)
    out.write(globalColorTable)
    out.write(imageDescriptor)
    out.write(imageData)
    out.write(trailer)

