#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    data = [b'\x00'] + list(rng.read((width * height // 255) or 1))
    data += [b'\x00'] + list(rng.read((width * height - len(data)) or 1))
    compressed_data = compress_lzw(data, lzw_minimum_code_size)
    
    # 7. Trailer
    trailer = b';'

    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)
    out.write(compressed_data)
    out.write(trailer)


# Compresses an array of bytes using LZW compression with the given minimum code size
def compress_lzw(
        input_bytes: list[int | str | bytes], min_code_size: int
    ) -> bytes:
    dictionary = {chr(i): i for i in range(min_code_size)}
    next_code = min_code_size + 1
    max_code_size = 1 << min_code_size
    clear_code = 1 << min_code_size
    end_of_information_code = clear_code + 1
    result = []
    current_byte = ''
    for byte in input_bytes:
        if isinstance(byte, int):
            byte = chr(byte)

        if byte == '\x00':
            result.append(dictionary[current_byte])

            if next_code < max_code_size:
                dictionary[next_code] = current_byte + current_byte[0]

                next_code += 1
            
            current_byte = ''
        
        elif current_byte + byte in dictionary:
            current_byte += byte
        
        else:
            result.append(dictionary[current_byte])

            if next_code < max_code_size:
                dictionary[next_code] = current_byte + byte[0]
                
                next_code += 1
            
            current_byte = byte
    
    if current_byte!= '':
        result.append(dictionary[current_byte])
    
    return bytes([clear_code]) + bytes(result) + bytes([end_of_information_code])