#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0] & 1)
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0] & 1)
    interlace_flag = bool(rng.read(1)[0] & 1)
    sort_flag = bool(rng.read(1)[0] & 1)
    reserved = rng.read(1)[0]
    lzw_min_code_size = rng.read(1)[0]
    id = struct.pack('>HHBBBBB', left, top, width, height, local_color_table_flag, interlace_flag, sort_flag, reserved)

    # 6. Local Color Table (if applicable)
    if local_color_table_flag:
        num_colors = (rng.read(1)[0] % 256) or 1
        color_table += rng.read(3 * num_colors)

    # 7. Image data
    image_data = b''
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x2C:
            # Image descriptor
            image_data += id

            # Image data
            length = rng.read(1)[0] + 1
            subblock = rng.read(length)
            image_data += subblock

            # Block terminator
            image_data += b'\x00'
        elif block_type == 0x21:
            # Extension
            extension_label = rng.read(1)[0]

            if extension_label in [0xF9]:
                # Graphic control extension
                image_data += gce

                # Sub-blocks of graphic control extension
                length = rng.read(1)[0] + 1
                subblock = rng.read(length)
                image_data += subblock

                # Block terminator
                image_data += b'\x00'
            elif extension_label in [0xFE, 0xFF]:
                # Comment extensions
                length = rng.read(1)[0] + 1
                subblock = rng.read(length)
                image_data += subblock

                # Block terminator
                image_data += b'\x00'
            elif extension_label in [0x01]:
                # Plain text extension
                length = rng.read(1)[0] + 1
                subblock = rng.read(length)
                image_data += subblock

                # Block terminator
                image_data += b'\x00'
            elif extension_label in [0xED]:
                # Application extension
                application_identifier = rng.read(8)
                application_auth_code = rng.read(3)
                length = rng.read(1)[0] + 1
                subblock = rng.read(length)
                image_data += subblock

                # Block terminator
                image_data += b'\x00'
            elif extension_label in [0xEE]:
                # Comment extension
                length = rng.read(1)[0] + 1
                subblock = rng.read(length)
                image_data += subblock

                # Block terminator
                image_data += b'\x00'
            else:
                raise Exception("Unknown extension label")
        elif block_type == 0x3B:
            break
        else:
            raise Exception("Unknown block type")

    # Write everything to output stream
    out.write(header)
    out.write(lsd)
    out.write(b'\x00')
    out.write(color_table)
    out.write(image_data)
