#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    lsd = struct.pack('>HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break
        
        data = rng.read(block_size)
        blocks.append((lzw_minimum_code_size, data))

    # Now let's compress our data using LZW
    compressed_data = b''
    code_size = lzw_minimum_code_size + 1
    clear_code = 1 << code_size
    end_of_information_code = clear_code + 1
    next_code = clear_code + 2
    codes = {}
    current_code = ''

    def add_code(c):
        nonlocal codes, current_code

        if len(current_code) == code_size - 1:
            # This is a new character in an existing code
            current_code += c

            if current_code not in codes:
                codes[current_code] = next_code
                next_code += 1

                return False
            
            return True

        elif len(current_code) > 0:
            # This is a new code starting from scratch
            assert len(current_code) < code_size - 1

            if current_code not in codes:
                codes[current_code] = next_code
                next_code += 1

            compressed_data += bytes([codes[current_code]])
            del codes[current_code]

            current_code = c
            codes[current_code] = next_code
            next_code += 1

            return False

        else:
            # This is just a plain old character
            compressed_data += bytes([ord(c)])
            current_code = c

            return False

    # Add the initial clear code to the table of codes
    codes[''] = clear_code

    # Compress!
    for block_size, data in blocks:
        code_size = block_size + 1

        for i in range(len(data)):
            # Get the next character
            c = chr(data[i])

            # Check if this is a new code
            if add_code(c):
                continue

            # If there are no more characters left, then we need to flush the last code
            if i == len(data) - 1:
                compressed_data += bytes([codes[current_code]])
                del codes[current_code]

    # Finally, add the EOI code
    compressed_data += bytes([end_of_information_code])

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    out.write(b''.join([struct.pack('>BBB', *color) for color in [color_table[i:i+3] for i in range(0, len(color_table), 3)]]))
    if gce:
        out.write(gce)
    out.write(image_descriptor)
    out.write(bytes([lzw_minimum_code_size]))
    out.write(compressed_data)
    out.write(trailer)