#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = rng.read(1)[0] & 0xFF | ((rng.read(1)[0] & 0xFF) << 8)
    height = rng.read(1)[0] & 0xFF | ((rng.read(1)[0] & 0xFF) << 8)
    flags = rng.read(1)[0]
    bg_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HBBBHB', width, height, flags, bg_color_index, pixel_aspect_ratio, 0)

    # 3. Global Color Table
    global_color_table_size = min((flags >> 7) + 1, 256) * 3
    global_color_table = [rng.read(1)[0] for _ in range(global_color_table_size)]
    color_table = bytes(global_color_table)

    # 4. Graphics Control Extension
    gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        if len(blocks) >= total_blocks_size:
            break

        block_size = rng.read(1)[0]
        if not block_size:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)