#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    lsd = struct.pack('>HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = [struct.pack('B', rng.read(1)[0])]
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            break

        if block_size == 128:
            # Repeat last character
            blocks[-1] += struct.pack('B', rng.read(1)[0])
        elif block_size > 128:
            # Copy previous character multiple times
            char = blocks[-1][-1:]
            count = rng.read(1)[0] & 0x7F

            if len(char) < count:
                char *= ((count // len(char)) + 1)
            
            blocks[-1] += char[:count]
        else:
            # Add new character to table
            blocks[-1] += bytes([rng.read(1)[0]])

        if len(blocks[-1]) >= 256:
            blocks.append(struct.pack('B', block_size))

    # Make sure that the last block ends in \x00
    blocks[-1] += b'\x00'

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)  # No ending \x00

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)