#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    lsd = struct.pack('>HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, rng.read(1)[0], rng.read(1)[0])

    # 6. Image Data
    # We use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]

    # Read all blocks until we get one with size 0
    blocks = []
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break
        
        # If this is an "uncompressed" block, just read that many bytes
        if block_size == 128:
            data = rng.read(128)
        elif block_size < 128:
            # This is a compressed block; let's decompress it!
            data = b''

            # The first byte contains the initial code size to use
            code_size = block_size + 1
            
            # Read the entire block as a bitstream
            bits = ''.join('{0:08b}'.format(byte, 'big') for byte in rng.read(block_size))

            # Now let's iterate over every pair of bits
            i = 0
            while i < len(bits):
                # Get the next two bits from our bitstream
                first_bit = int(bits[i])
                second_bit = int(bits[i+1])

                # Combine them to form a single value
                value = ((first_bit << 1) | second_bit)
                
                # Add it to our output stream
                data += bytes([(value >> j) & 0xff for j in range((code_size - 1), -1, -1)])

                # Increment our pointer by two
                i += 2
        else:
            raise Exception("Invalid block size!")
            
        blocks.append(bytes([block_size]) + data)
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything together
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + image_data + trailer)