#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = b''
    if rng.read(1)[0] & 1 == 1:
        # We want to write a GCE
        # 4.1. Block Size
        block_size = rng.read(1)[0]

        # 4.2. Graphic Control Label
        graphic_control_label = rng.read(1)[0]

        # 4.3. Graphic Control Block
        disposal_method = rng.read(1)[0] & 7
        user_input_flag = bool(rng.read(1)[0] & 1)
        transparent_color_flag = bool(rng.read(1)[0] & 1)
        delay_time = struct.unpack('>H', rng.read(2))[0]
        transparent_color_index = rng.read(1)[0]

        # 4.4. Block Terminator
        block_terminator = rng.read(1)[0]

        gce = struct.pack('BBBBB', block_size, graphic_control_label, disposal_method << 2 | user_input_flag << 1 | transparent_color_flag, delay_time, transparent_color_index) + block_terminator

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB', 0x2C, 0, 0, width, height, 0, 0)

    # 6. Local Color Table (same as global color table)
    local_color_table = color_table

    # 7. Image Data
    image_data = rng.read(width * height)

    # 8. Trailer
    trailer = b';'

    # Write everything together
    out.write(header + lsd + color_table + gce + image_descriptor + local_color_table + image_data + trailer)
