#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    # Construct the logical screen descriptor
    lsd = struct.pack('>HBBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    # Random global color table size (using one byte for variety)
    global_color_table_size = rng.read(1)[0]

    # Random global color table data (using three bytes per entry for variety)
    global_color_table_data = rng.read(global_color_table_size * 3)

    # Construct the global color table
    color_table = global_color_table_data

    # 4. Graphic Control Extension
    # Random delay time (using two bytes for variety)
    delay_time = struct.unpack('>H', rng.read(2))[0]

    # Random transparent color index (using one byte for variety)
    transparent_color_index = rng.read(1)[0]

    # Random disposal method (using one byte for variety)
    disposal_method = rng.read(1)[0]

    # Construct the graphic control extension
    gce = struct.pack('>BBB', 0xF9, 4, disposal_method) + struct.pack('>HH', delay_time, transparent_color_index)

    # 5. Image Descriptor
    # Random left position (using two bytes for variety)
    left_position = struct.unpack('>H', rng.read(2))[0]

    # Random top position (using two bytes for variety)
    top_position = struct.unpack('>H', rng.read(2))[0]

    # Random width (using two bytes for variety)
    width = struct.unpack('>H', rng.read(2))[0]

    # Random height (using two bytes for variety)
    height = struct.unpack('>H', rng.read(2))[0]

    # Random local color table flag (using one byte for variety)
    local_color_table_flag = rng.read(1)[0] & 0b10000000

    # Random interlace flag (using one byte for variety)
    interlace_flag = rng.read(1)[0] & 0b01000000

    # Random sort flag (using one byte for variety)
    sort_flag = rng.read(1)[0] & 0b00100000

    # Random size of local color table (using one byte for variety)
    local_color_table_size = rng.read(1)[0] & 0b00011111

    # Construct the image descriptor
    image_descriptor = struct.pack('>BBHHBB', 0x2C, local_color_table_flag | local_color_table_size, left_position, top_position, width, height, interlace_flag | sort_flag)

    # 6. Image Data
    # Random minimum code size (using one byte for variety)
    lzw_minimum_code_size = rng.read(1)[0]

    # Random image data (using up to 255 bytes at a time for variety)
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while len(blocks) < total_blocks_size:
        block_size = min(total_blocks_size - len(blocks), rng.read(1)[0])
        blocks.append(rng.read(block_size))

    image_data = struct.pack('>BB', lzw_minimum_code_size, 0) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
