#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Magic header
    out.write(b'GIF89a')

    # Screen descriptor
    width = 2 ** (1 + rng.read(1)[0] % 7)
    height = 2 ** (1 + rng.read(1)[0] % 7)
    flags = rng.read(1)[0] & 0xE0 | 0x1F
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('<HHBB', width, height, flags, background))

    # Global color table
    if flags & 0x80:
        colors = 2 ** ((flags >> 4) + 1)
        for _ in range(colors):
            out.write(bytes([rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]]))
    else:
        pass

    # Image data
    while True:
        # Image descriptor
        left = rng.read(1)[0]
        top = rng.read(1)[0]
        width = 2 ** (1 + rng.read(1)[0] % 7)
        height = 2 ** (1 + rng.read(1)[0] % 7)
        flags = rng.read(1)[0] & 0xC0 | 0x07
        local_color_table = rng.read(1)[0] & 0x80
        interlaced = rng.read(1)[0] & 0x40
        sort = rng.read(1)[0] & 0x20
        reserved = rng.read(1)[0] & 0x0C
        size_of_local_color_table = rng.read(1)[0] & 0x03
        out.write(struct.pack('<HHBB', left, top, width, height))

        # Local color table
        if local_color_table:
            colors = 2 ** (size_of_local_color_table + 1)
            for _ in range(colors):
                out.write(bytes([rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]]))
        else:
            pass

        # LZW minimum code size
        out.write(bytes([rng.read(1)[0]]))

        # Image data
        while True:
            # Block size
            block_size = rng.read(1)[0]
            if block_size == 0:
                break
            out.write(bytes([block_size]))

            # Block data
            for _ in range(block_size):
                out.write(bytes([rng.read(1)[0]]))

        # Trailer
        if not interlaced:
            break


# Generates a random PNG file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_png(
        rng: BinaryIO, out: BinaryIO):
    # Magic header
    out.write(b'\x89PNG\r\n\x1A\n')

    # IHDR chunk
    width = 2 ** (1 + rng.read(1)[0] % 7)
    height = 2 ** (1 + rng.read(1)[0] % 7)
    depth = rng.read(1)[0] & 0x07
    color_type = rng.read(1)[0] & 0x0F
    compression_method = rng.read(1)[0] & 0x0F
    filter_method = rng.read(1)[0] & 0x0F
    interlace_method = rng.read(1)[0] & 0x01
    out.write(struct.pack('>I', 13))
    out.write(b'IHDR')
    out.write(struct.pack('>LLCCBBB', width, height, depth, color_type, compression_method, filter_method, interlace_method))
    out.write(b'\xAE\x42\x60\x82')

    # PLTE chunk
    if color_type == 3:
        colors = 2 ** (depth + 1)
        out.write(struct.pack('>I', 5 + 3 * colors))
        out.write(b'PLTE')
        for _ in range(colors):
            out.write(bytes([rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]]))
        out.write(b'\xAE\x42\x60\x82')
    elif color_type ==             for x in range(width[1]):
                out.write(bytes([rng.read(1)[0]]))
    else:
        pass

