#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        out.write(header + lsd)
        return

    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    local_color_table_flag = bool(packed & 0x80)
    interlace_flag = bool(packed & 0x40)
    sort_flag = bool(packed & 0x20)
    reserved_bits = bool(packed & 0x1C)
    lct_size = (packed >> 4) + 1
    id = struct.pack('>HHBBBB', left, top, width, height, packed, rng.read(1)[0])

    # 6. Local Color Table (if present)
    if local_color_table_flag:
        num_colors = 2 ** lct_size
        color_table = rng.read(3 * num_colors)  # Random colors

    # 7. Image Data
    image_data = rng.read(width * height)

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + id + color_table + image_data)

