#!/usr/bin/env python3

import struct
from typing import BinaryIO

def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor (LSD)
    width = rng.read(2)
    height = rng.read(2)
    packed = rng.read(1)
    global_color_table_flag = bool(packed[0] & 0b10000000)
    color_resolution = (packed[0] >> 4) & 0b111
    sort_flag = bool(packed[0] & 0b100)
    global_color_table_size = 2 ** ((packed[0] & 0b111) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    lsd = width + height + packed + bytes([background_color_index]) + bytes([pixel_aspect_ratio])
    
    # 3. Global Color Table
    if global_color_table_flag:
        color_table = rng.read(global_color_table_size * 3)
    else:
        color_table = b''
    
    # 4. Graphic Control Extension (GCE)
    gce = b'\x21\xF9\x04\x05\x00' + rng.read(2) + rng.read(1) + b'\x00'
    
    # 5. Image Descriptor
    image_descriptor = b'\x2C\x00\x00\x00\x00' + rng.read(2) + rng.read(2) + bytes([0])
    
    # 6. Image Data
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        
        if not block_size:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)