#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor (LSD)
    width = rng.read(1)[0]
    height = rng.read(1)[0]
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack(
            '>HHBBBB',
            640, 480,
            packed_fields,
            background_color_index,
            pixel_aspect_ratio
        )

    # 3. Color Table
    color_table = rng.read(768)

    # 4. Graphic Control Extension (GCE)
    # We don't care about these fields for now
    gce = struct.pack('>BBBHBB', 0xF9, 4, 0, 0, 0, 0)

    # 5. Image Descriptor
    left = rng.read(1)[0]
    top = rng.read(1)[0]
    width = rng.read(1)[0]
    height = rng.read(1)[0]
    packed_fields = rng.read(1)[0]
    local_color_table_flag = bool(packed_fields & 0b10000000)
    interlaced = bool(packed_fields & 0b01000000)
    sorted = bool(packed_fields & 0b00100000)
    reserved = bool(packed_fields & 0b00010000)
    lzw_minimum_code_size = packed_fields & 0b00001111
    image_descriptor = struct.pack(
            '>HHBBBB',
            left, top,
            width, height,
            packed_fields,
            lzw_minimum_code_size
        )

    # 6. Image Data
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while len(blocks) < total_blocks_size:
        block_size = min(total_blocks_size - len(blocks), rng.read(1)[0])
        blocks.append(rng.read(block_size))

    image_data = struct.pack('>BB', lzw_minimum_code_size, 0) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
