#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = rng.read(1)[0] & 7
    global_color_table = rng.read(global_color_table_size * 3)

    # 4. Image Descriptors
    # We will only write one image descriptor
    # Image Left Position
    left = struct.pack('>H', rng.read(2)[0])
    # Image Top Position
    top = struct.pack('>H', rng.read(2)[0])
    # Image Width
    width = struct.pack('>H', rng.read(2)[0])
    # Image Height
    height = struct.pack('>H', rng.read(2)[0])
    # Packed Fields (we don't care about any of these)
    packed = rng.read(1)

    # 5. Local Color Table (same as above)
    local_color_table_size = rng.read(1)[0] & 7
    local_color_table = rng.read(local_color_table_size * 3)

    # 6. Image Data
    # We will just use a single code, which is 0xFF (end of data block)
    image_data = struct.pack('>B', 0xFF)

    # Write everything to the output stream
    out.write(header + lsd + global_color_table + left + top + width + height + packed + local_color_table + image_data)

if __name__ == '__main__':
    from sys import argv
    if len(argv) != 2:
        print("Usage: {} <output-file>".format(argv[0]))
