#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big', signed=False) % 640 + 1
    height = int.from_bytes(rng.read(2), 'big', signed=False) % 480 + 1
    background_color_index = int.from_bytes(rng.read(1), 'big', signed=False) % 256
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big', signed=False) % 8 + 1
    out.write(struct.pack('<HBBB', width, height, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table (optional)
    if rng.read(1)[0] == 0xFF:
        color_table_size = 7 + int.from_bytes(rng.read(1), 'big', signed=False) % 7
        for _ in range(3 * (2 ** color_table_size)):
            out.write(rng.read(1))

    # 4. Image Data
    while True:
        # 4.1 Graphic Control Extension
        if rng.read(1)[0] != 0xF9:
            break
        block_size = rng.read(1)[0]
        assert block_size == 0x04
        flags = rng.read(1)[0]
        delay = int.from_bytes(rng.read(2), 'big', signed=False)
        transparent_color_index = rng.read(1)[0]
        terminator = rng.read(1)[0]
        assert terminator == 0x00

        # 4.2 Image Descriptor
        image_left = int.from_bytes(rng.read(2), 'big', signed=False) % width
        image_top = int.from_bytes(rng.read(2), 'big', signed=False) % height
        image_width = int.from_bytes(rng.read(2), 'big', signed=False) % (width - image_left) + 1
        image_height = int.from_bytes(rng.read(2), 'big', signed=False) % (height - image_top) + 1
        local_color_table_flag = bool(rng.read(1)[0] & 0x80)
        interlaced = bool(rng.read(1)[0] & 0x40)
        sort_flag = bool(rng.read(1)[0] & 0x20)
        reserved = bool(rng.read(1)[0] & 0x1C)
        size_of_local_color_table = 7 + int.from_bytes(rng.read(1), 'big', signed=False) % 7
        assert not (local_color_table_flag and size_of_local_color_table == 0)

        # 4.3 Local Color Table (optional)
        if local_color_table_flag:
            for _ in range(3 * (2 ** size_of_local_color_table)):
                out.write(rng.read(1))

        # 4.4 LZW Minimum Code Size
        lzw_minimum_code_size = rng.read(1)[0]
        assert lzw_minimum_code_size <= 8

        # 4.5 Image Data
        while True:
            code_size = rng.read(1)[0]
            if code_size == 0x00:
                break
            else:
                assert code_size >= 0x02 and code_size <= 0x08
