#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    if packed & 0x80:
        color_count = 2 ** ((packed >> 4) + 1)
        color_table = rng.read(3 * color_count)
    else:
        color_table = b''

    # 4. Graphic Control Extension (optional)
    if rng.read(1)[0]:
        block_size = rng.read(1)[0]
        assert block_size == 0x04
        flags = rng.read(1)[0]
        delay = rng.read(2)
        transparent_index = rng.read(1)[0]
        terminator = rng.read(1)[0]
        gce = struct.pack('>BBB', block_size, flags, delay[1]) + delay[0].to_bytes(1, byteorder='big') + transparent_index.to_bytes(1, byteorder='big') + terminator.to_bytes(1, byteorder='big')
    else:
        gce = b''

    # 5. Image Descriptor
    image_descriptor = b'\x2c\x00\x00\x00\x00'

    # 6. Image Data
    data = []
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            break
        data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)


if __name__ == '__main__':
    try:
        seed = int(sys.argv[1])
    except IndexError:
        print("Usage: {} <seed>".format(os.path.basename(__file__)))
        exit()

    with open('/dev/urandom', 'rb') as f:
        f.seek(seed)
        with open('test.gif', 'wb') as o:
            generate_random_gif(f, o)