#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    out.write('GIF89a'.encode())
    
    # 2. Logical screen descriptor
    width = int.from_bytes(rng.read(2), 'big') + 1
    height = int.from_bytes(rng.read(2), 'big') + 1
    color_table_size = 0x7 & (int.from_bytes(rng.read(1), 'big'))
    background_color_index = int.from_bytes(rng.read(1), 'big')
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big')
    out.write((width).to_bytes(2, 'big'))
    out.write((height).to_bytes(2, 'big'))
    out.write(bytearray([0xF8 | color_table_size]))
    out.write(background_color_index.to_bytes(1, 'big'))
    out.write(pixel_aspect_ratio.to_bytes(1, 'big'))
    
    # 3. Global color table
    for _ in range(0, 2 ** (color_table_size + 1)):
        out.write(rng.read(3))
    
    # 4. Image descriptor
    left = int.from_bytes(rng.read(2), 'big')
    top = int.from_bytes(rng.read(2), 'big')
    width = int.from_bytes(rng.read(2), 'big')
    height = int.from_bytes(rng.read(2), 'big')
    local_color_table_flag = bool(int.from_bytes(rng.read(1), 'big') >> 7)
    interlace_flag = bool(int.from_bytes(rng.read(1), 'big') >> 6)
    sort_flag = bool(int.from_bytes(rng.read(1), 'big') >> 5)
    reserved_bits = int.from_bytes(rng.read(1), 'big') & 0b00011111
    out.write(left.to_bytes(2, 'big'))
    out.write(top.to_bytes(2, 'big'))
    out.write(width.to_bytes(2, 'big'))
    out.write(height.to_bytes(2, 'big'))
    out.write(bytearray([0x80 | (local_color_table_flag << 7) | (interlace_flag << 6) | (sort_flag << 5) | reserved_bits]))
    
    # 5. Local color table
    if local_color_table_flag:
        for _ in range(0, 2 ** (color_table_size + 1)):
            out.write(rng.read(3))
    
    # 6. Image data
    lzw_minimum_code_size = int.from_bytes(rng.read(1), 'big')
    out.write(lzw_minimum_code_size.to_bytes(1, 'big'))
    
    # 7. Trailer
    out.write(b'\x3B')

if __name__ == "__main__":
    from os import urandom
    with open("test.gif", "wb") as f:
        generate_random_gif(urandom(), f)