#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    out.write(header)
    
    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big') % 640 + 1
    height = int.from_bytes(rng.read(2), 'big') % 480 + 1
    background = rng.read(1)[0] & 0x7F
    aspect_ratio = rng.read(1)[0] & 0xFF
    global_color_table_flag = (rng.read(1)[0] >> 7) & 0b1
    color_resolution = ((rng.read(1)[0] >> 4) & 0b11) * 8
    sort_flag = (rng.read(1)[0] >> 3) & 0b1
    size_of_global_color_table = (rng.read(1)[0] >> 0) & 0b111
    if size_of_global_color_table == 0:
        raise ValueError('Invalid value for size of global color table')
    else:
        size_of_global_color_table = 2 ** (size_of_global_color_table + 1)
        
    descriptor = struct.pack(
            '<HHBBB',
            width, height,
            background, aspect_ratio,
            global_color_table_flag << 7 | color_resolution << 4 | sort_flag << 3 | size_of_global_color_table
        )
    out.write(descriptor)
    
    # 3. Global Color Table
    if global_color_table_flag:
        for _ in range(size_of_global_color_table):
            red   = rng.read(1)[0]
            green = rng.read(1)[0]
            blue  = rng.read(1)[0]
            
            color = struct.pack('<BBB', red, green, blue)
            out.write(color)
    
    # 4. Image Data
    while True:
        # 4.1 Graphic Control Extension
        extension_introducer = rng.read(1)[0]
        assert extension_introducer == 0x21
        
        graphic_control_label = rng.read(1)[0]
        assert graphic_control_label == 0xf9
        
        block_size = rng.read(1)[0]
        assert block_size == 0x04
        
        disposal_method = rng.read(1)[0] & 0b111
        user_input_flag = (rng.read(1)[0] >> 3) & 0b1
        transparent_color_flag = (rng.read(1)[0] >> 2) & 0b1
        delay_time = int.from_bytes(rng.read(2), 'big')
        transparent_color_index = rng.read(1)[0]
        
        block_terminator = rng.read(1)[0]
        assert block_terminator == 0x00
        
        # 4.2 Image Descriptor
        image_separator = rng.read(1)[0]
        assert image_separator == 0x2c
        
        left_position = int.from_bytes(rng.read(2), 'big')
        top_position = int.from_bytes(rng.read(2), 'big')
        width = int.from_bytes(rng.read(2), 'big')
        height = int.from_bytes(rng.read(2), 'big')
        local_color_table_flag = (rng.read(1)[0] >> 7) & 0b1
        interlace_flag = (rng.read(1)[0] >> 6) & 0b1
        sort_flag = (rng.read(1)[0] >> 5) & 0b1
        size_of_local_color_table = (rng.read(1)[0] >> 0) & 0b111
        if size_of_local_color_table == 0:
            raise ValueError('Invalid value for