#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big') % 65536
    height = int.from_bytes(rng.read(2), 'big') % 65536
    flags = rng.read(1)[0] & 7
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]

    # 3. Global Color Table
    color_table_size = (flags + 1) * 3
    color_table = rng.read(color_table_size)

    # 4. Image Descriptors
    image_descriptor = b'\x2c\x00\x00\x00\x00\x00\xff\xff\xff\xff\x00'
    image_descriptor += struct.pack('>H', width)
    image_descriptor += struct.pack('>H', height)
    image_descriptor += bytes([flags])
    image_descriptor += bytes([background])
    image_descriptor += bytes([aspect_ratio])

    # 5. Local Color Table
    local_color_table_size = (flags + 1) * 3
    local_color_table = rng.read(local_color_table_size)

    # 6. LZW Minimum Code Size
    lzw_minimum_code_size = rng.read(1)[0]

    # 7. Data Sub-Blocks
    data_sub_blocks = []
    while True:
        size = rng.read(1)[0]
        if not size:
            break
        data_sub_blocks.append(rng.read(size))

    # 8. Trailer
    trailer = b'\x3b'

    # Write everything to the output stream
    out.write(header)
    out.write(image_descriptor)
    out.write(lzw_minimum_code_size)
    for sub_block in data_sub_blocks:
        out.write(struct.pack('B', len(sub_block)))
        out.write(sub_block)
    out.write(trailer)

