#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    global_color_table_flag = bool(packed & 0b10000000)
    color_resolution = (packed >> 4) & 0b111
    sort_flag = bool(packed & 0b100)
    size_of_global_color_table = 2 ** ((packed & 0b111) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HBBBBB', width, height, packed, background_color_index, pixel_aspect_ratio, 0)

    # 3. Global Color Table
    if global_color_table_flag:
        color_table = rng.read(3 * size_of_global_color_table)
    else:
        color_table = b''

    # 4. Graphic Control Extension
    if rng.read(1)[0]:
        delay_time = rng.read(2)
        transparent_color_index = rng.read(1)[0]
        block_terminator = rng.read(1)[0]
        gce = struct.pack('>BBBB', 0xF9, 4, delay_time[0], delay_time[1]) + transparent_color_index + block_terminator
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while len(blocks) < total_blocks_size:
        block_size = min(total_blocks_size - len(blocks), rng.read(1)[0])
        blocks.append(rng.read(block_size))

    image_data = struct.pack('>BB', lzw_minimum_code_size, 0) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
