#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b"GIF89a"

    # 2. Logical Screen Descriptor
    lsd = struct.pack('>BBHBB', *struct.unpack('>BBHHBB', rng.read(7)))

    # 3. Global Color Table
    color_table = rng.read(3*((lsd[4]&0x07)+1))

    # 4. Graphic Control Extension
    gce = struct.pack('>BBB', 0x21, 0xF9, 0x04) + rng.read(5)

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    data = []
    while True:
        block_size = rng.read(1)[0] & 0xFF

        if block_size == 0:
            break

        data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)


if __name__ == '__main__':
    try:
        seed = int(sys.argv[1])
    except IndexError:
        print("Usage: {} <seed>".format(os.path.basename(__file__)))
        exit()

    with open('/dev/urandom', 'rb') as f:
        f.seek(seed)
        with open('test.gif', 'wb') as o:
            generate_random_gif(f, o)