#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Generate random GIF files using /dev/urandom

@author: mikemadsen
"""
import struct
from random import Random

def generate_random_gif(rng, out):
    """
    Generates a random GIF file using /dev/urandom
    :param rng: A file object that reads from /dev/urandom
    :param out: A file object that writes to a file
    :return: None
    """
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = struct.unpack('>B', rng.read(1))[0]
    global_color_table_flag = bool(packed & 0x80)
    color_resolution = (packed >> 4) & 0x7
    sort_flag = bool((packed >> 3) & 0x1)
    global_color_table_size = 2 ** ((packed & 0x7) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBBB', width, height, packed, global_color_table_flag, color_resolution, sort_flag, global_color_table_size, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table
    if global_color_table_flag:
        color_table = rng.read(3 * global_color_table_size)
    else:
        color_table = b''
    
    # 4. Graphic Control Extension
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        
        if not block_size:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)