#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big') % 640 + 1
    height = int.from_bytes(rng.read(2), 'big') % 480 + 1
    background_color_index = rng.read(1)[0] & 0xF
    pixel_aspect_ratio = rng.read(1)[0] & 0x7

    logical_screen_descriptor = struct.pack('<HBBB', width, height, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    global_color_table_flag = rng.read(1)[0] >> 7
    color_resolution = (rng.read(1)[0] >> 4) & 0x7
    sort_flag = rng.read(1)[0] >> 3
    size_of_global_color_table = (rng.read(1)[0] >> 0) & 0x7

    if global_color_table_flag == 1:
        global_color_table = [
            rng.read(3) for _ in range(2 ** (size_of_global_color_table + 1))
        ]
    else:
        global_color_table = []

    # 4. Image Descriptors
    image_descriptors = []

    while True:
        # 4.1 Image Separator
        image_separator = rng.read(1)

        if image_separator != b'\x2C':
            break

        # 4.2 Image Descriptor
        left_position = int.from_bytes(rng.read(2), 'big')
        top_position = int.from_bytes(rng.read(2), 'big')
        width = int.from_bytes(rng.read(2), 'big')
        height = int.from_bytes(rng.read(2), 'big')
        local_color_table_flag = rng.read(1)[0] >> 7
        interlace_flag = rng.read(1)[0] >> 6
        sort_flag = rng.read(1)[0] >> 5
        reserved = rng.read(1)[0] >> 4
        size_of_local_color_table = (rng.read(1)[0] >> 0) & 0x7

        if local_color_table_flag == 1:
            local_color_table = [
                rng.read(3) for _ in range(2 ** (size_of_local_color_table + 1))
            ]
        else:
            local_color_table = []

        # 4.3 LZW Minimum Code Size
        lzw_minimum_code_size = rng.read(1)[0]

        # 4.4 Data Sub-Blocks
        data_sub_blocks = []

        while True:
            block_size = rng.read(1)[0]

            if block_size == 0:
                break

            data_sub_blocks += [rng.read(block_size)]

        image_descriptors += [(left_position, top_position, width, height, local_color_table, lzw_minimum_code_size, data_sub_blocks)]

    # 5. Trailer
    trailer = b'\x3B'

    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(b''.join([*map(lambda x: bytes(x), global_color_table)]))
    out.write(b''.join([*map(lambda x: bytes(x[0]), image_descriptors)]))
    out.write(b''.join([*map(lambda x: bytes(x[1]), image_descriptors)]))
    out.write(b''.join([*map(lambda x: bytes(x[2]), image_descriptors)]))
    out.write(b''.join([*map(lambda x: bytes(x[3]), image_descriptors)]))
    out.write(b''.join([*map(lambda x: bytes(