#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    out.write(header)

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), 'big') % 640 + 1
    height = int.from_bytes(rng.read(2), 'big') % 480 + 1
    background_color_index = rng.read(1)[0] & 0xF
    pixel_aspect_ratio = rng.read(1)[0] & 0xF
    global_color_table_flag = (rng.read(1)[0] >> 7) == 1
    color_resolution = ((rng.read(1)[0] >> 4) & 0x7) + 1
    sort_flag = (rng.read(1)[0] >> 3) == 1
    size_of_global_color_table = (((rng.read(1)[0] >> 1) & 0x7) + 1) * 3

    screen_descriptor = struct.pack('<HHBBB',
                                    width, height,
                                    background_color_index,
                                    pixel_aspect_ratio,
                                    global_color_table_flag << 7 |
                                    color_resolution << 4 |
                                    sort_flag << 3 |
                                    size_of_global_color_table // 3 - 1)
    out.write(screen_descriptor)

    # 3. Global Color Table
    if global_color_table_flag:
        for _ in range(size_of_global_color_table):
            out.write(rng.read(1))

    # 4. Image Data
    while True:
        # 4.1 Image Separator
        image_separator = rng.read(1)
        assert image_separator[0] == 0x2C
        out.write(image_separator)

        # 4.2 Image Descriptor
        left = int.from_bytes(rng.read(2), 'big') % width
        top = int.from_bytes(rng.read(2), 'big') % height
        width = int.from_bytes(rng.read(2), 'big') % (width - left) + 1
        height = int.from_bytes(rng.read(2), 'big') % (height - top) + 1
        local_color_table_flag = (rng.read(1)[0] >> 7) == 1
        interlace_flag = (rng.read(1)[0] >> 6) == 1
        sort_flag = (rng.read(1)[0] >> 5) == 1
        size_of_local_color_table = (((rng.read(1)[0] >> 2) & 0x7) + 1) * 3

        image_descriptor = struct.pack('<HHHBBB',
                                       left, top, width, height,
                                       local_color_table_flag << 7 |
                                       interlace_flag << 6 |
                                       sort_flag << 5 |
                                       size_of_local_color_table // 3 - 1)
        out.write(image_descriptor)

        # 4.3 Local Color Table
        if local_color_table_flag:
            for _ in range(size_of_local_color_table):
                out.write(rng.read(1))

        # 4.4 LZW Minimum Code Size
        lzw_minimum_code_size = rng.read(1)[0] & 0x7F
        out.write(lzw_minimum_code_size.to_bytes(1, 'big'))

        # 4.5 Image Data
        while True:
            block_size = rng.read(1)[0]
            if block_size == 0:
                break
            else:
                out.write(rng.read(block_size))

        # 4.6 Trailer
        trailer = rng.read(1)
        assert trailer[0] == 0x00
        out.write(trailer