#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]

    # 3. Global Color Table
    global_color_table_size = (packed & 0b11100000) >> 5
    if global_color_table_size == 7:
        raise NotImplementedError("Global color table size of 7 is not supported")
    global_color_table_sort = bool((packed & 0b00010000) >> 4)
    global_color_table_colors = 2 ** (global_color_table_size + 1)
    for i in range(global_color_table_colors):
        red = rng.read(1)[0]
        green = rng.read(1)[0]
        blue = rng.read(1)[0]
        out.write(bytes([red, green, blue]))

    # 4. Image Data
    while True:
        block_type = rng.read(1)[0]
        if block_type == 0x2c:
            break
        elif block_type == 0x21:
            extension_code = rng.read(1)[0]
            if extension_code == 0xf9:
                # Graphic Control Extension
                flags = rng.read(1)[0]
                delay = struct.unpack('>H', rng.read(2))[0]
                transparent_index = rng.read(1)[0]
                block_terminator = rng.read(1)[0]
            else:
                raise NotImplementedError(f"Extension code {extension_code} is not implemented")
        else:
            raise NotImplementedError(f"Block type {block_type} is not implemented")
    
    # 5. Image Descriptor
    left, top, width, height = struct.unpack('>HHHH', rng.read(8))
    local_color_table_flag = bool((rng.read(1)[0] & 0b10000000) >> 7)
    interlace_flag = bool((rng.read(1)[0] & 0b01000000) >> 6)
    sort_flag = bool((rng.read(1)[0] & 0b00100000) >> 5)
    reserved = bool((rng.read(1)[0] & 0b00011111) >> 0)
    if reserved != 0:
        raise ValueError("Reserved bits must be zero")

    # 6. Local Color Table
    if local_color_table_flag:
        local_color_table_size = (rng.read(1)[0] & 0b11100000) >> 5
        if local_color_table_size == 7:
            raise NotImplementedError("Local color table size of 7 is not supported")
        local_color_table_sort = bool((rng.read(1)[0] & 0b00010000) >> 4)
        local_color_table_colors = 2 ** (local_color_table_size + 1)
        for i in range(local_color_table_colors):
            red = rng.read(1)[0]
            green = rng.read(1)[0]
            blue = rng.read(1)[0]
            out.write(bytes([red, green, blue]))
    else:
        local_color_table_colors = global_color_table_colors

    # 7. LZW Minimum Code Size
    lzw_minimum_code_size = rng.read(1)[0]

    # 8. Image Data
    data_subblocks = []
    while True:
        subblock_length = rng.read(1)[0]
        if subblock_length == 0:
            break
        data_subblocks.append(rng.read(subblock_length))
    image_data = b''.join(data_subblocks)

    # Write everything to the output
    out.write(header)
    out.write(struct.pack('>HHB', width, height,