#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    screen_width = rng.read(2)[0]
    screen_height = rng.read(2)[0]
    flags = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('<HHBBBB', screen_width, screen_height, flags, background_color_index, pixel_aspect_ratio, 0)

    # 3. Global Color Table
    color_table = rng.read(3 * (2 ** ((flags & 0x07) + 1)))

    # 4. Graphic Control Extension
    gce = struct.pack('>BBHBB', 0x21, 0xF9, 4, rng.read(1)[0], rng.read(2)[0])

    # 5. Image Descriptor
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = rng.read(2)[0]
    height = rng.read(2)[0]
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom
    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)