#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('B', 0b11100000 | rng.read(1)[0] & 0b00011111) + \
          struct.pack('BB', *[rng.read(1)[0] for _ in range(2)]) + \
          struct.pack('B', rng.read(1)[0] & 0b00011111) + \
          struct.pack('B', rng.read(1)[0] & 0b00011111)

    # 3. Global Color Table
    # We'll use a global color table of size 768 (256 colors x 3 bytes per color).
    # The first 16 entries are reserved for the palette, so we skip them.
    gct = rng.read(768 - 16)

    # 4. Image Descriptors
    # We'll use one image descriptor with no local color table.
    idesc = struct.pack('BBHBB', 0x2c, 0, 0, width, height) + \
            struct.pack('B', 0)

    # 5. Image Data
    # We'll use 1 byte per pixel, so we need to generate `width*height` bytes.
    ildata = rng.read(width * height)

    # 6. Trailer
    trailer = b';'

    # Write everything together
    out.write(header)
    out.write(lsd)
    out.write(gct)
    out.write(idesc)
    out.write(ildata)
    out.write(trailer)
