#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = rng.read(num_colors * 3)
    
    # 4. Image Data
    # We will write this later
    
    # Write everything to the output stream
    out.write(header + lsd + global_color_table)
    
    # Now we need to write the image data
    # First we need to know how many frames there are
    num_frames = (rng.read(1)[0] % 256) or 1
    
    # Then we can loop over all of them
    for i in range(num_frames):
        
        # 1. Graphic Control Extension
        # We don't care about any of these fields, so just read one byte
        gce = rng.read(1)
        
        # 2. Image Descriptor
        # Same as above
        idd = rng.read(1)
        
        # 3. Local Color Table
        # Same as above
        local_color_table = rng.read(1)
        
        # 4. Image Data
        # This is where we actually do something with the frame
        # Let's pick a random width and height for the frame
        frame_width, frame_height = struct.unpack('>HH', rng.read(4))
        
        # And then let's pick a random x and y offset
        x_offset, y_offset = struct.unpack('>HH', rng.read(4))
        
        # Finally, let's pick some random pixel data
        pixels = rng.read(frame_width * frame_height)
        
        # Now we need to write this to the output stream
        # First the GCE
        out.write(gce)
        
        # Then the IDD
        out.write(idd)
        
        # Then the LCT
        out.write(local_color_table)
        
        # Then the width and height
        out.write(struct.pack('>HH', frame_width, frame_height))
        
        # Then the x and y offsets
        out.write(struct.pack('>HH', x_offset, y_offset))
        
        # Then the actual pixels
        out.write(pixels)
        
        # And finally, an empty block terminator
        out.write(b'\x00')
