#!/usr/bin/env python3

import struct
from random import randint, choice


def generate_random_gif(seed, out):
    """Generate a GIF using the given seed."""

    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor (LSD)
    width = randint(1, 640)
    height = randint(1, 480)
    packed = randint(0, 7) << 4 | randint(0, 7)  # Global Color Table Flag and Color Resolution
    background_color_index = randint(0, 2 ** 8 - 1)
    pixel_aspect_ratio = randint(0, 2 ** 8 - 1)
    lsd = struct.pack('<HHBBH', width, height, packed, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    colors = [randint(0, 2 ** 8 - 1) for _ in range(3 * (2 ** (packed & 7)))]
    color_table = bytes(colors)

    # 4. Graphic Control Extension (GCE)
    delay_time = randint(0, 2 ** 16 - 1)
    transparent_color_flag = choice((True, False))
    transparent_color_index = randint(0, 2 ** 8 - 1) if transparent_color_flag else 0
    disposal_method = randint(0, 7)
    user_input_flag = choice((True, False))
    lzw_minimum_code_size = randint(2, 9)
    gce = struct.pack('<BBBBB', 4, delay_time, transparent_color_flag, transparent_color_index, disposal_method,
                      user_input_flag)

    # 5. Image Descriptor
    left = randint(0, width - 1)
    top = randint(0, height - 1)
    width = randint(1, width - left)
    height = randint(1, height - top)
    local_color_table_flag = choice((True, False))
    interlace_flag = choice((True, False))
    sort_flag = choice((True, False))
    reserved = 0
    image_descriptor = struct.pack('<HHBBBB', left, top, width, height, local_color_table_flag, interlace_flag,
                                   sort_flag, reserved)

    # 6. Image Data
    rng = Random(seed)
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom as get_random_bytes
    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(get_random_bytes(), f)