#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Magic header
    out.write(b'GIF89a')

    # Screen descriptor
    width = 2 ** (1 + rng.read(1)[0] % 7)
    height = 2 ** (1 + rng.read(1)[0] % 7)
    flags = rng.read(1)[0] & 0xE0 | 0x1F
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0] & 0xF0 | ((rng.read(1)[0] >> 4) & 0x0F)
    out.write(struct.pack('<HHBBB', width, height, flags, bgcolor, aspect))

    # Global color table
    numcolors = min((flags & 0x07) + 1, 256)
    for _ in range(numcolors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(bytes([r, g, b]))

    # Image descriptor
    left = 0
    top = 0
    width = 2 ** (1 + rng.read(1)[0] % 7)
    height = 2 ** (1 + rng.read(1)[0] % 7)
    flags = rng.read(1)[0] & 0xE0 | 0x0C
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0] & 0xF0 | ((rng.read(1)[0] >> 4) & 0x0F)
    out.write(struct.pack('<HHBBBB', left, top, width, height, flags, bgcolor,
                          aspect))

    # Local color table
    numcolors = min(((flags & 0x07) + 1), 256)
    for _ in range(numcolors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(bytes([r, g, b]))

    # LZW minimum code size
    lzwminsize = rng.read(1)[0]
    out.write(bytes([lzwminsize]))

    # Image data
    imagedata = bytearray()
    while len(imagedata) < width * height:
        blocktype = rng.read(1)[0]

        if blocktype == 0x00:
            # Raw data
            rawlen = rng.read(1)[0] + 1
            imagedata += bytes(rng.read(rawlen))
        elif blocktype == 0x01:
            # Repeat previous pixel
            repcount = rng.read(1)[0] + 1
            imagedata += imagedata[-repcount:]
        else:
            # LZW encoded data
            subblocktype = blocktype & 0xC0

            if subblocktype == 0x00:
                # Plain text
                subblocklen = blocktype & 0x3F

                if subblocklen > 0:
                    imagedata += bytes(rng.read(subblocklen))

                    if subblocklen >= 2 and \
                            imagedata[-2] == 0x00 and \
                            imagedata[-1] <= 0x07:
                        # Escape sequence
                        escapeseq = imagedata[-2:]

                        if escapeseq == b'\x00\x00':
                            # End of stream
                            break
                        elif escapeseq[0] == 0x00 and escapeseq[1] <= 0x07:
                            # Repeated character
                            repcount = escapeseq[1] + 1
                            imagedata += imagedata[-repcount:]
                        elif escapeseq[0] == 0x01:
                            # Repeated line
                            repcount = imagedata[-1] + 1
                            imagedata += imagedata[-width*repcount:-width*(
                                repcount-1)]
                        elif escapeseq[0] == 0x02:
                            # Repeated frame
                            repcount = imagedata[-1] + 1
                            imagedata += imagedata[-width*height*repcount:-
                                                  width*height*(repcount - 1)]
                        elif escapeseq[0] == 0x03:
                            # Repeated graphic control extension
                            repcount = imagedata[-1] + 1
                            imagedata += imagedata[-6*repcount:-6*(
                                repcount - 1)]
                        elif escapeseq[0] == 0x04:
                            # Repeated comment
                            repcount = imagedata[-1] + 1
                            imagedata += imagedata[-(subblocklen+1)*repcount:-
                                                  -(subblocklen+1)*(repcount -
                                                                  1)]
                        elif escapeseq[0] == 0x05:
                            # Repeated application extension
                            repcount = imagedata[-1] + 1
                            imagedata += imagedata[-(subblocklen+1)*repcount:-
                                                  -(subblocklen+1)*(repcount