#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If there is an extension present...
        gce = struct.pack('>BBBBBH', 0x21, 0xF9, 4, rng.read(1)[0], 0, 0)  #...write one with a random delay time
    else:
        gce = b''

    # 5. Image Descriptor
    image_descriptor = struct.pack('>HHB', 0, 0, 0b10000000 | ((num_colors - 1) << 4))  # Set local color table flag to true

    # 6. Image Data
    image_data = b''
    while len(image_data) < width * height:
        # Write either a single-byte value (if we're not in the global color table) or three-byte values (if we are)
        image_data += (color_table[rng.read(1)[0]] + color_table[rng.read(1)[0]] + color_table[rng.read(1)[0]]) \
            if rng.read(1)[0] & 0b10000000 else (rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 7. Trailer
    trailer = b';'

    # Write everything to the output stream
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)

