#!/usr/bin/env python3

import struct
from typing import BinaryIO

# https://www.w3.org/Graphics/GIF/spec-gif89a.txt

HEADER = b'GIF89a'  # Header of a GIF file (always this value)
LOGICAL_SCREEN_DESCRIPTOR = '!2H4B'  # Logical screen descriptor
COLOR_TABLE = '!16s'  # Color table, always 256 bytes long
GRAPHIC_CONTROL_EXTENSION = '!3B1H'  # Graphic control extension
IMAGE_DESCRIPTOR = '!2H3B'  # Image descriptor
TRAILER = b';'  # Trailer of a GIF file (always this value)

# The size of each pixel in bits
PIXEL_SIZE = 8

# The number of colors in the palette
PALETTE_LENGTH = 2 ** PIXEL_SIZE


def generate_random_gif(seed: bytes, out: BinaryIO):
    """Generate a random gif and write it into `out`."""

    # Generate the header
    header = HEADER

    # Generate the logical screen descriptor
    width = seed[0] % 256
    height = seed[1] % 256
    flags = [
        0b1 << 7 |  # Global color table flag set
        0b1 << 4 |  # Sort flag set
        0b1 << 3,   # Size of global color table is 2^n entries
        0b1        # Background color index is 0
    ]
    lsd = struct.pack(LOGICAL_SCREEN_DESCRIPTOR, width, height, *flags)

    # Generate the color table
    color_table = bytearray()
    for i in range(PALETTE_LENGTH):
        r = seed[i*3+2] % 256
        g = seed[i*3+1] % 256
        b = seed[i*3] % 256
        color_table += struct.pack('BBB', r, g, b)

    # Generate the graphic control extension
    gce = struct.pack(GRAPHIC_CONTROL_EXTENSION,
                      *[0x04,  # Block size
                        0x00,  # No disposal specified
                        0xFF,  # Transparent color index
                        ])

    # Generate the image descriptor
    image_descriptor = struct.pack(IMAGE_DESCRIPTOR,
                                   width, height,
                                   0b111,  # Local color table flag set
                                   0,     # No interlace
                                   )

    # Generate the image data
    image_data = bytearray()
    for y in range(height):
        for x in range(width):
            image_data += struct.pack('B', seed[(y*width+x)*3])

    # Generate the trailer
    trailer = TRAILER

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom
    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)