#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    has_gce = bool((rng.read(1)[0] % 2))  # Random boolean value
    gce = None
    
    if has_gce:
        gce = b''
        
        # Block size is always 4
        gce += bytes([4])
        
        # Flags field: packed fields (using a random byte for variety), delay time, transparent color index
        flags = struct.pack('>BBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
        
        # Delay time (two bytes)
        delay_time = struct.pack('>H', rng.read(2))[::-1]  # Reversed because little endian
        
        # Transparent color index (one byte)
        trans_color_index = bytes([rng.read(1)[0]])
        
        gce += flags + delay_time + trans_color_index
    
    # 5. Image Descriptor
    # Left position (two bytes), top position (two bytes), width (two bytes), height (two bytes), local color table flag (packed field), interlace flag (packed field), sort flag (packed field), reserved (packed field), local color table size (packed field)
    image_descriptor = struct.pack('>HHHHBBBBB', rng.read(2)[0], rng.read(2)[0], width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 6. Image Data
    # Minimum code size (one byte)
    lzw_minimum_code_size = rng.read(1)[0]
    
    # Blocks are separated by a zero byte
    blocks = []
    
    while True:
        # Block size (one byte)
        block_size = rng.read(1)[0]
        
        if not block_size: break
                
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)