#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width = ((rng.read(1)[0] % 256) << 8) | rng.read(1)[0]
    height = ((rng.read(1)[0] % 256) << 8) | rng.read(1)[0]

    # Random packed field
    packed_field = rng.read(1)[0] & 0b00000111

    # Global Color Table Flag set?
    global_color_table_flag = bool((packed_field >> 7) & 1)

    # Size of the Global Color Table
    color_resolution = (packed_field >> 4) & 0b111
    sort_flag = bool((packed_field >> 3) & 1)
    size_of_global_color_table = 2 ** (color_resolution + 1)

    # Background Color Index
    background_color_index = rng.read(1)[0]

    # Pixel Aspect Ratio
    pixel_aspect_ratio = rng.read(1)[0] & 0b1111

    lsd = struct.pack(
        '>HBBBBB',
        width,
        height,
        packed_field,
        0xFF,  # No Global Color Table
        0,     # No Color Table Sorting
        0      # No Global Color Table Size
    )

    # 3. Global Color Table (optional)
    if global_color_table_flag:
        num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from os import urandom
    with open('/tmp/test.gif', 'wb') as f:
        generate_random_gif(urandom(), f)