#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0b10000000)
    if not global_color_table_flag:
        color_table = b''
    else:
        size = ((packed & 7) + 1) * 3
        color_table = bytes([size]) + rng.read(size - 1)

    # 4. Graphics Control Extension (if present)
    graphics_control_extension_flag = bool(packed & 0b01000000)
    if not graphics_control_extension_flag:
        gce = b''
    else:
        gce = b'\x21\xF9\x04' + rng.read(3)

    # 5. Image Descriptor
    image_descriptor = struct.pack('<BHHBBBBB', 0x2C, 0, 0, width, height, 0, 0, 0)

    # 6. Data
    data = []
    for _ in range((height // 8) * width):
        byte = rng.read(1)[0]

        # If this is an LZW code, we need to read more bytes until we get one that isn't
        while byte >= 0xFF:
            byte = (byte << 8) | rng.read(1)[0]

        data.append(bytes([byte]))

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)


if __name__ == '__main__':
    try:
        seed = int(sys.argv[1])
    except IndexError:
        print("Usage: {} <seed>".format(os.path.basename(__file__)))
        exit()

    with open('/dev/urandom', 'rb') as f:
        f.seek(seed)
        with open('test.gif', 'wb') as o:
            generate_random_gif(f, o)