#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = rng.read(1)[0] % 64 + 1
    h = rng.read(1)[0] % 64 + 1

    # Color table size is fixed to 7 bits of LSD
    color_table_size = 2 ** ((rng.read(1)[0] & 0x7) + 1)

    # Background color index is always zero
    background_color_index = 0

    # Pixel aspect ratio is always 0
    pixel_aspect_ratio = 0

    lsd = bytearray([
        0xFF & w,
        0xFF & (w >> 8),
        0xFF & h,
        0xFF & (h >> 8),

        # Flags
        0b10000000 | 0b01000000 | 0b00100000 | 0b00010000,

        0xFF & color_table_size,
        0xFF & background_color_index,
        0xFF & pixel_aspect_ratio
    ])

    # 3. Global Color Table
    global_color_table = [
        rng.read(3) for i in range(color_table_size)
    ]

    # 4. Image Descriptor
    image_descriptor = bytearray([
        0b10000000 | 0b00000000,
        0xFF & w,
        0xFF & (w >> 8),
        0xFF & h,
        0xFF & (h >> 8),
        0b00000000,
        0b00000000,
        0b00000000,
        0b00000000
    ])

    # 5. Local Color Table
    local_color_table = []

    # If there's no global color table then we need one here too
    if len(global_color_table) == 0:
        local_color_table = [
            rng.read(3) for i in range(color_table_size)
        ]

    # 6. Image Data
    image_data = bytearray()

    # The first frame has an explicit duration of 1 second
    image_data += struct.pack('>H', int((1 / 100) * 100))

    # Then comes the actual data
    while True:
        # Each block starts with a single byte indicating its size
        block_size = rng.read(1)[0]

        # End of image?
        if block_size == 0: break
        
        # Otherwise it's just raw data
        image_data += rng.read(block_size)
    
    # 7. Trailer
    trailer = b';'

    # Write everything to output
    out.write(header)
    out.write(lsd)

    if len(local_color_table):
        for rgb in local_color_table:
            out.write(rgb)

    elif len(global_color_table):
        for rgb in global_color_table:
            out.write(rgb)

    else: raise Exception("No color tables found")

    out.write(image_descriptor)
    out.write(image_data)
    out.write(trailer)