#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHBB', 0x2C, rng.read(1)[0], width, height, 0)
    
    # 6. Image Data
    data = []
    while True:
        # Every frame has a random length between 1-100 pixels
        num_pixels = (rng.read(1)[0] % 100) + 1
        
        # We'll use this to store the pixel values
        pixels = [None] * num_pixels

        # For every pixel in our randomly generated frame...
        for i in range(num_pixels):
            #...pick a random color from our global color table
            pixel = rng.read(1)[0] % len(color_table)
            
            # And set that as the value of our current pixel
            pixels[i] = pixel

            # Write the pixel value to our output stream
            out.write(bytes([pixel]))
            
        # Now that we've written all of our pixels, let's write them to our data array
        data.append(pixels)
        
        # Check if we're done with our frames
        if not rng.read(1)[0]:
            break
        
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to our output stream!
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)
    for d in data:
        for p in d:
            out.write(bytes([p]))
    out.write(trailer)

