#!/usr/bin/env python3

import struct
from PIL import Image


def read_lsd(in_file):
    """Read LSD header from input file."""
    return in_file.read(16)


def write_lsd(out, lsd):
    """Write LSD header to output file."""
    out.write(lsd)


def read_color_table(in_file):
    """Read color table from input file."""
    return in_file.read(256 * 3)


def write_color_table(out, color_table):
    """Write color table to output file."""
    out.write(color_table)


def read_gce(in_file):
    """Read GCE header from input file."""
    return in_file.read(40)


def write_gce(out, gce):
    """Write GCE header to output file."""
    out.write(gce)


def read_image_descriptor(in_file):
    """Read image descriptor from input file."""
    return in_file.read(14)


def write_image_descriptor(out, image_descriptor):
    """Write image descriptor to output file."""
    out.write(image_descriptor)


def read_image_data(in_file):
    """Read image data from input file."""
    return in_file.read()


def write_image_data(out, image_data):
    """Write image data to output file."""
    out.write(image_data)


if __name__ == '__main__':
    # Open files for reading and writing
    with open('input.gif', 'rb') as in_file:
        with open('output.gif', 'wb') as out:
            # Read all headers into memory
            lsd = read_lsd(in_file)
            color_table = read_color_table(in_file)
            gce = read_gce(in_file)
            image_descriptor = read_image_descriptor(in_file)

            # Convert image data to RGB mode (discarding alpha channel if present)
            img = Image.open('input.png').convert('RGB')
            pixels = list(img.getdata())

            # Write new headers to output file
            write_lsd(out, lsd)
            write_color_table(out, color_table)
            write_gce(out, gce)

            # Calculate new width and height of image
            width = int.from_bytes(struct.pack('<H', img.width), byteorder='big')
            height = int.from_bytes(struct.pack('<H', img.height), byteorder='big')

            # Pack pixel values into bytes
            packed_pixels = [b''.join([pixel[i] for pixel in pixels])
                             for i in range(len(pixels[0]))]

            # Write new image descriptor to output file
            image_descriptor = struct.pack('<6sHHBBBHB',
                                           b'GIF89a',
                                           0, 0, 0, 0, 0, 0,
                                           width, height, 0xF0 | len(packed_pixels))
            write_image_descriptor(out, image_descriptor)

            # Write image data to output file
            write_image_data(out, b''.join(packed_pixels))

    # Add missing parts of GIF format that are not supported by Pillow
    # 1. Header
    header = b'GIF89a'

    # 2. Logical screen descriptor
    lsd = struct.pack('<HHBBBHH',
                      0, 0, 0, 0, 0,
                      img.width, img.height, 0xF0 | len(packed_pixels))

    # 3. Color table
    color_table = b''

    # 4. Graphic control extension
    gce = struct.pack('<6sBBBBBHB',
                      b'GCE\x00',
                      0, 0, 0, 0, 0, 0,
                      0, 0, 0)

    # 5. Image descriptor
    image_descriptor = struct.pack('<6sHHBBBHB',
                                   b'GIF89a',
                                   0, 0, 0, 0, 0, 0,
                                   img.width, img.height, 0xF0 | len(packed_pixels))

    # 6. Image data
    image_data = b''.join(packed_pixels)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
