#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table
    global_color_table_size = int((rng.read(1)[0] & 7) + 1) * 3
    global_color_table = [rng.read(1)[0] for _ in range(global_color_table_size)]
    
    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBB', 0, 0, 0b10000000 | global_color_table_size - 1, 0, 0)
    
    # 5. Image Data
    # Each frame is made up of an introducer (0x2C), descriptor, local color table, and data sub-blocks
    frames = []
    while True:
        # Introducer
        frames.append(bytes([0x2C]))
        
        # Frame descriptor
        frames.append(image_descriptor)
        
        # Local color table
        frames.append(bytes(global_color_table))
        
        # Data sub-blocks
        block_size = rng.read(1)[0]
        if not block_size: break
        frames.append(block_size)
        frames.append(rng.read(block_size))
        
    # Combine everything together
    gif = header + lsd + bytes().join(frames)
    
    # Write to output
    out.write(gif)


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)