#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        out.write(header + lsd)
        return

    num_colors = 2 ** ((packed >> 4) + 1)
    color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(num_colors)]
    color_table = b''.join(color_table)

    # 4. Graphics Control Extension (if present)
    graphics_control_extension_flag = bool(packed & 0x10)
    if graphics_control_extension_flag:
        delay = struct.unpack('>H', rng.read(2))[0]
        transparent_index = rng.read(1)[0]
        gce = struct.pack('>BBHB', 0x21, 0xF9, 4, delay << 1 | 1)
        gce += struct.pack('>BBB', 0x01, transparent_index, 0)
    else:
        gce = b''

    # 5. Image Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    data = []
    while True:
        block_size = rng.read(1)[0] & 0xFF

        if block_size == 0:
            break

        data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)


if __name__ == '__main__':
    from sys import argv, stdout

    with open('/dev/urandom', 'rb') as f:
        generate_random_gif(f, stdout.buffer)