#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    width = int.from_bytes(rng.read(2), byteorder='big') % (640 - 15) + 15
    height = int.from_bytes(rng.read(2), byteorder='big') % (480 - 15) + 15

    flags = bytes([int.from_bytes(rng.read(1), byteorder='big')])

    global_color_table_flag = flags[0] & 0b10000000!= 0
    interlaced_flag = flags[0] & 0b01000000!= 0
    sorted_color_table_flag = flags[0] & 0b00100000!= 0
    reserved_bits = flags[0] & 0b00011111

    background_color_index = int.from_bytes(rng.read(1), byteorder='big')
    pixel_aspect_ratio = int.from_bytes(rng.read(1), byteorder='big')

    lsd = struct.pack('>HHBBB', width, height, global_color_table_flag << 7 | interlaced_flag << 6 | sorted_color_table_flag << 5 | reserved_bits, background_color_index, pixel_aspect_ratio)

    # 2. Color Table
    if global_color_table_flag:
        color_table_size = 2 ** ((reserved_bits + 1) + 1)

        color_table = []

        for i in range(0, color_table_size):
            r = int.from_bytes(rng.read(1), byteorder='big')

            g = int.from_bytes(rng.read(1), byteorder='big')

            b = int.from_bytes(rng.read(1), byteorder='big')

            color_table += [r, g, b]

        color_table = bytes(color_table)

    else:
        color_table = b''

    # 3. Graphic Control Extension
    gce = b''

    if not interlaced_flag and len(color_table)!= 0:
        gce = b'\x21\xF9\x04' + bytes([(background_color_index << 4) | 0]) + b'\x00'

    # 4. Image Descriptor
    left = int.from_bytes(rng.read(2), byteorder='big') % (width - 15) + 15
    top = int.from_bytes(rng.read(2), byteorder='big') % (height - 15) + 15
    width = min((width - left), 15)
    height = min((height - top), 15)

    local_color_table_flag = bool(int.from_bytes(rng.read(1), byteorder='big'))

    lct_size = 2 ** ((flags[0] & 0b00000111) + 1)

    lct = []

    for i in range(0, lct_size):
        r = int.from_bytes(rng.read(1), byteorder='big')

        g = int.from_bytes(rng.read(1), byteorder='big')

        b = int.from_bytes(rng.read(1), byteorder='big')

        lct += [r, g, b]

    lct = bytes(lct)

    image_descriptor = struct.pack('>HHBB', left, top, width, height) + bytes([(local_color_table_flag << 7) | (len(lct))])

    # 5. Image Data
    image_data = b''

    for y in range(top, top + height):
        for x in range(left, left + width):
            image_data += bytes([int.from_bytes(rng.read(1), byteorder='big')])

    # 6. LZW Min Code Size
    # 7. Trailer
    trailer = b'\x3B'

    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
