#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = (rng.read(1)[0] % 64) + 1
    height = (rng.read(1)[0] % 64) + 1
    packed_flags = rng.read(1)[0] & ~0b11110000 | 0b11100000
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('<HHBBB', width, height, packed_flags, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    global_color_table_flag = bool(packed_flags & 0b10000000)
    if not global_color_table_flag:
        return

    global_color_table_size = ((packed_flags >> 4) & 0b111) + 1
    global_color_table = []
    for _ in range(global_color_table_size):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        global_color_table.append((r,g,b))
    color_table = bytes([*map(lambda rgb: rgb[0], global_color_table)])

    # 4. Graphics Control Extension
    graphics_control_extension_flag = bool(packed_flags & 0b00001000)
    if not graphics_control_extension_flag:
        return

    delay_time = rng.read(2)[0] * 10
    transparent_color_index = rng.read(1)[0]
    packed_flags = rng.read(1)[0] & ~0b11110000
    graphics_control_extension = struct.pack(
            '<BBHB',
            0xF9,  # Application extension block label
            4,     # Block size
            0,     # Packed flags
            delay_time,
            transparent_color_index,
            packed_flags,
        )

    # 5. Image Descriptor
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = rng.read(2)[0]
    height = rng.read(2)[0]
    packed_flags = rng.read(1)[0] & ~0b11110000
    local_color_table_flag = bool(packed_flags & 0b10000000)
    interlace_flag = bool(packed_flags & 0b01000000)
    sort_flag = bool(packed_flags & 0b00100000)
    reserved_bits = (packed_flags >> 3) & 0b111
    image_descriptor = struct.pack(
            '<HHHHBB',
            left,
            top,
            width,
            height,
            packed_flags,
            local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved_bits,
        )

    # 6. LZW minimum code size and data
    lzw_minimum_code_size = rng.read(1)[0]
    total_blocks_size = rng.read(1)[0]
    blocks = [rng.read(1)[0]]
    while len(blocks[-1]) < total_blocks_size:
        block_size = min(total_blocks_size - len(blocks), rng.read(1)[0])
        blocks.append(rng.read(block_size))

    image_data = struct.pack('>BB', lzw_minimum_code_size, 0) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
