#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0] | ((rng.read(1)[0] & 0b111) << 8)  # Pack them together
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    
    # Build the LSD
    lsd = struct.pack('<6sHHBBB',
                      header,
                      width,
                      height,
                      packed_fields,
                      background_color_index,
                      pixel_aspect_ratio
                     )
    
    # 3. Global Color Table
    global_color_table_size = min((packed_fields >> 4) + 1, 2**((packed_fields & 0b111) + 1))  # Number of colors in the table
    global_color_table = [rng.read(3) for _ in range(global_color_table_size)]  # Read RGB triplets
    
    # 4. Application Extension
    application_extension_label = b'\xFF\x0BNETSCAPE2.0'
    application_identifier = b'NETSCAPE2.0'
    application_auth_code = b'\<PASSWORD>'
    application_data = b'\x03\x01\x00\x00'  # Loop forever
    
    app_ext = application_extension_label + \
              struct.pack('BB',
                         len(application_identifier),
                         0
                        ) + \
              application_identifier + \
              struct.pack('BBBB',
                         *struct.unpack('BBBB', application_auth_code)
                        ) + \
              struct.pack('BBBB',
                         *struct.unpack('BBBB', application_data)
                        )
    
    # 5. Graphics Control Extension
    graphics_control_extension_label = b'\xF9\x04'
    delay_time = rng.read(1)[0]
    transparent_color_index = rng.read(1)[0]
    block_terminator = b'\x00'
    
    gce = graphics_control_extension_label + \
          struct.pack('BB',
                      delay_time,
                      0
                     ) + \
          struct.pack('BBB',
                      transparent_color_index,
                      0,
                      0
                     ) + \
          block_terminator
    
    # 6. Image Descriptor
    left = 0
    top = 0
    width -= 1  # Subtract one because we're going to use this value as an upper bound later on
    height -= 1
    local_color_table_flag = False
    interlace_flag = False
    sort_flag = False
    reserved_bits = 0
    lzw_minimum_code_size = 8  # We'll make this configurable later
    image_descriptor = struct.pack('<HHHHBBBBB',
                                   left,
                                   top,
                                   width,
                                   height,
                                   local_color_table_flag,
                                   interlace_flag,
                                   sort_flag,
                                   reserved_bits
                                  ) + \
                       struct.pack('B', lzw_minimum_code_size)
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes
        
        if not block_size:
            break
            
        blocks.append(bytes([block_size]) + rng.read(block_size))
        
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0
    
    # 7. Trailer
    trailer = b'\x3B'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    from sys import argv

    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)