#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), byteorder='little') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bgColorIndex = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    pixelAspectRatio = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lsd = width.to_bytes(2, 'little', signed=False) \
            + height.to_bytes(2, 'little', signed=False) \
            + flags \
            + bgColorIndex \
            + pixelAspectRatio

    # 3. Global Color Table
    globalColorTableFlag = int.from_bytes(rng.read(1), byteorder='little') & 0x80!= 0
    if globalColorTableFlag:
        colorResolution = int.from_bytes(rng.read(1), byteorder='little') & 7
        sortFlag = int.from_bytes(rng.read(1), byteorder='little') & 8!= 0
        sizeOfGlobalColorTable = int.from_bytes(rng.read(1), byteorder='little')

        globalColorTable = []
        for i in range(sizeOfGlobalColorTable):
            r = int.from_bytes(rng.read(1), byteorder='little')
            g = int.from_bytes(rng.read(1), byteorder='little')
            b = int.from_bytes(rng.read(1), byteorder='little')

            globalColorTable += [r]
            globalColorTable += [g]
            globalColorTable += [b]

    # 4. Image Descriptor
    leftPos = int.from_bytes(rng.read(2), byteorder='little') % (2**10 - 1)
    topPos = int.from_bytes(rng.read(2), byteorder='little') % (2**10 - 1)
    imageWidth = int.from_bytes(rng.read(2), byteorder='little') % (2**10 - 1) + 1
    imageHeight = int.from_bytes(rng.read(2), byteorder='little') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    localColorTableFlag = int.from_bytes(rng.read(1), byteorder='little') & 0x80!= 0
    interlaceFlag = int.from_bytes(rng.read(1), byteorder='little') & 64!= 0
    sortFlag = int.from_bytes(rng.read(1), byteorder='little') & 128!= 0
    reserved = bytes([int.from_bytes(rng.read(1), byteorder='little'),
                      int.from_bytes(rng.read(1), byteorder='little')])
    lzwMinimumCodeSize = int.from_bytes(rng.read(1), byteorder='little')

    id = leftPos.to_bytes(2, 'little', signed=False) \
          + topPos.to_bytes(2, 'little', signed=False) \
          + imageWidth.to_bytes(2, 'little', signed=False) \
          + imageHeight.to_bytes(2, 'little', signed=False) \
          + flags \
          + localColorTableFlag.to_bytes(1, 'little', signed=False) \
          + interlaceFlag.to_bytes(1, 'little', signed=False) \
          + sortFlag.to_bytes(1, 'little', signed=False) \
          + reserved \
          + lzwMinimumCodeSize.to_bytes(1, 'little', signed=False)

    # 5. Local Color Table
    if localColorTableFlag:
        colorResolution = int.from_bytes(rng.read(1), byteorder='little') & 7
        sortFlag = int.from_bytes(rng.read(1), byteorder='little') & 8!= 0
        sizeOfLocalColorTable = int.from_bytes(rng.read(1), byteorder='little')

        localColorTable = []
        for i in range(sizeOfLocalColorTable):
            r = int.from_bytes(rng.read(1), byteorder='little')
            g = int.from