#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:  # If we want one, then add an extension
        gce = b'\x21\xF9\x04' + rng.read(3) + b'\x00'
        
        # Add a local color table?
        if rng.read(1)[0]:
            num_colors = (rng.read(1)[0] % 256) or 1
            color_table += rng.read(3 * num_colors)
        
    else:
        gce = None
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('>HHBBB', width, height, 0b10000000 | rng.read(1)[0], 0, 0)
    
    # 6. Image Data
    data = []
    for _ in range((width * height)):
        data.append(color_table[(rng.read(1)[0] % len(color_table)) // 3])
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything to output
    out.write(header)
    out.write(lsd)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)
    out.write(bytes(data))
    out.write(trailer)


if __name__ == '__main__':
    from sys import argv
    if len(argv)!= 2:
        print("Usage: {} <output>".format(argv[0]))
        exit()

    with open('/dev/urandom', 'rb') as f:
        with open(argv[1], 'wb') as o:
            generate_random_gif(f, o)