#!/usr/bin/env python3

import struct
from random import Random


def generate_random_gif(rng: Random, out):
    """Generate a GIF file using the given random number generator."""
    
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = [rng.randint(1, 2**i - 1) for i in range(2)]
    packed = struct.pack('>HBBBBB', width, height, 0xF0 | rng.randint(0, 7), 0, 0, 0)
    lsd = packed[0:7]

    # 3. Global Color Table
    global_color_table = bytearray([rng.randint(0, 255) for _ in range(2**(lsd[4] + 1))])
    color_table = global_color_table[:2**(lsd[4] + 1)]

    # 4. Graphics Control Extension
    gce = struct.pack('>BBBBHB', 0x21, 0xF9, 4, 0, 0, 0)
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    data = []
    while True:
        block_size = rng.read(1)[0] & 0xFF

        if block_size == 0:
            break

        data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(data) + trailer)


if __name__ == '__main__':
    from sys import argv, stdout

    with open('/dev/urandom', 'rb') as f:
        generate_random_gif(f, stdout.buffer)