#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))

    # Flags field: Local color table flag set to true, interlace flag set to false, sort flag set to false, reserved bits set to zero
    flags = struct.pack('>BB', 0b11000000 | ((rng.read(1)[0] & 0b0011) << 2), rng.read(1)[0])

    # 6. Image Data
    # Let's make it be a random-sized rectangle of pixels with random RGB values
    image_data = b''
    for _ in range((width * height)):
        image_data += struct.pack('>BBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    image_descriptor = struct.pack('>HHBBBB', left, top, width, height, *(flags + color_table[0]))

    # 7. Trailer
    trailer = b'\x3B'

    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
