#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    header = b"GIF89a" + bytes([0x12, 0x34]) + \
             bytes([0x56, 0x78]) + bytes([0x9A, 0xBC])
    out.write(header)

    # Logical screen descriptor
    width = int.from_bytes(rng.read(2), byteorder="little") % (2**10) + 1
    height = int.from_bytes(rng.read(2), byteorder="little") % (2**10) + 1
    flags = [int.from_bytes(rng.read(1), byteorder="little")] * 3
    bgColorIndex = int.from_bytes(rng.read(1), byteorder="little") & 0xF
    pixelAspectRatio = int.from_bytes(rng.read(1), byteorder="little")
    logicalScreenDescriptor = struct.pack("<HHBBB",
                                          width,
                                          height,
                                          flags[0],
                                          flags[1],
                                          flags[2],
                                          bgColorIndex)
    logicalScreenDescriptor += struct.pack("B", pixelAspectRatio)
    out.write(logicalScreenDescriptor)

    # Global color table
    globalColorTableSize = int.from_bytes(rng.read(1), byteorder="little")
    globalColorTable = []
    for i in range(globalColorTableSize):
        r = int.from_bytes(rng.read(1), byteorder="little") & 0xFF
        g = int.from_bytes(rng.read(1), byteorder="little") & 0xFF
        b = int.from_bytes(rng.read(1), byteorder="little") & 0xFF
        globalColorTable.append((r,g,b))
    for i in range(len(globalColorTable)):
        rgb = globalColorTable[i]
        out.write(struct.pack("BBB", rgb[0], rgb[1], rgb[2]))

    # Image descriptor
    left = int.from_bytes(rng.read(2), byteorder="little") % (2**10)
    top = int.from_bytes(rng.read(2), byteorder="little") % (2**10)
    flags = [int.from_bytes(rng.read(1), byteorder="little")] * 3
    localColorTableFlag = bool(flags[0] & 0x80)
    interlaceFlag = bool(flags[0] & 0x40)
    sortFlag = bool(flags[0] & 0x20)
    reserved = flags[0] & 0x1F
    localColorTableSize = flags[1] & 0x07
    localColorTableSize += 1
    localColorTable = []
    if localColorTableFlag:
        for i in range(localColorTableSize):
            r = int.from_bytes(rng.read(1), byteorder="little") & 0xFF
            g = int.from_bytes(rng.read(1), byteorder="little") & 0xFF
            b = int.from_bytes(rng.read(1), byteorder="little") & 0xFF
            localColorTable.append((r,g,b))
        for i in range(len(localColorTable)):
            rgb = localColorTable[i]
            out.write(struct.pack("BBB", rgb[0], rgb[1], rgb[2]))
    else:
        localColorTableSize = 0
    lzwMinimumCodeSize = int.from_bytes(rng.read(1), byteorder="little")
    imageDescriptor = struct.pack("<HHBBBBB",
                                  left,
                                  top,
                                  flags[0],
                                  flags[1],
                                  flags[2],
                                  localColorTableSize)
    out.write(imageDescriptor)

    # LZW minimum code size
    out.write(struct.pack("B", lzwMinimumCodeSize))

    # Data sub-blocks
    dataSubBlocks = []
    while True:

        # Sub-block length
        blockLength = int.from_bytes(rng.read(1), byteorder="little")
        if not blockLength: break

        # Sub-block type and data
        blockType = int.from_bytes(rng.read(1), byteorder="little")
        blockData = rng.read(blockLength - 1)
        dataSubBlock = struct.pack("BB",
                                   blockLength,
                                   blockType)
        dataSubBlock += blockData
        dataSubBlocks.append(dataSubBlock)

    # Write all data sub-blocks to output
    for dataSubBlock in dataSubBlocks:
        out.write(dataSubBlock