#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if not block_size:
            break

        if len(blocks):
            last_block_size = blocks[-1][0]

            if last_block_size >= 128:
                # If we're in a data block, just add another byte to it
                blocks[-1].append(rng.read(1)[0])

                continue

            elif last_block_size == 0:
                # If we're in an EOI block, don't bother adding anything more
                pass

            else:
                # Otherwise, this is the end of our current block; let's start a new one!
                blocks.append([rng.read(1)[0]])
        else:
            # We haven't started any blocks yet; let's do that now
            blocks.append([rng.read(1)[0]])

        if sum(len(b) for b in blocks) > total_blocks_size - 1:
            # If we've exceeded the maximum size allowed by the total blocks size, stop generating blocks
            break

    image_data = b'\x00'.join(bytes(block) for block in blocks) + b'\xFF\x00'

    # 7. Trailer
    trailer = b';'

    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    out.write(struct.pack('>B', lzw_minimum_code_size))
    out.write(image_data)
    out.write(trailer)

