#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(num_colors * 3)

    # 4. Graphic Control Extension
    # We don't care about the delay time, so we just set it to zero
    gce = struct.pack('>BBBB', 0xF9, 4, rng.read(1)[0], 0)
    # 5. Image Descriptor
    # Frame left position is always 0
    frame_left = struct.pack('<H', 0)
    # Frame top position is always 0
    frame_top = struct.pack('<H', 0)
    # Width of the current frame is randomly chosen from [1, 32767]
    frame_width = struct.pack('<H', rng.read(2)[0] & 0x7FFF | 1)
    # Height of the current frame is randomly chosen from [1, 32767]
    frame_height = struct.pack('<H', rng.read(2)[0] & 0x7FFF | 1)
    # Flags are always 0b1110
    flags = struct.pack('<B', 0xE)
    # Transparent color index is always 0
    transparent_color_index = struct.pack('<B', 0)
    # Image data starts right after the last field in the descriptor
    image_descriptor = frame_left + frame_top + frame_width + frame_height + flags + transparent_color_index

    # 6. Image Data
    # We'll use the same algorithm as the one used by the original challenge
    # https://github.com/welchbj/ctf/blob/master/hitcon-ctf-2019/pwn/gif/solve.py#L10-L14
    image_data = b''
    while len(image_data) < width * height:
        # Read a random byte
        byte = rng.read(1)[0]

        if byte == 0xFF:
            # If that byte was 0xFF, write 0x00 instead
            image_data += b'\x00\x00'
        else:
            # Otherwise, write the original value twice
            image_data += struct.pack('BB', byte, byte)

    # 7. Trailer
    trailer = b';'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
