#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + rng.randint(50, 200)
    height = 1 + rng.randint(50, 200)
    flags = (
        0x80 | # Global color table flag
        ((rng.randint(0, 7) & 1) << 7) | # Color resolution
        ((rng.randint(0, 3) & 1) << 6) | # Sort flag
        ((rng.randint(0, 1) & 1) << 4) | # Size of global color table
        ((rng.randint(0, 1) & 1) << 3)) # Background color index flag
    bgColorIndex = rng.randint(0, 255) if (flags >> 3) & 1 else None
    aspectRatio = rng.randint(0, 8)
    out.write(struct.pack('<HHBBB', width, height, flags, bgColorIndex or 0, aspectRatio))

    # Global color table
    for i in range(pow(2, (flags >> 4) & 1) * 3):
        out.write(bytes([rng.randint(0, 255)]))

    # Image descriptor
    leftPos = 0
    topPos = 0
    width = 1 + rng.randint(width // 2 - 1, width)
    height = 1 + rng.randint(height // 2 - 1, height)
    flags = (
        0x80 | # Local color table flag
        0x40 | # Interlace flag
        0x20 | # Sort flag
        0x10 | # Reserved bit
        ((rng.randint(0, 1) & 1) << 2) | # LZW minimum code size
        ((rng.randint(0, 3) & 1) << 1)) # Table based image data flag
    lctSize = pow(2, (flags >> 2) & 1) * 3
    lct = [rng.randint(0, 255)] * lctSize
    out.write(struct.pack('<HHHHB', leftPos, topPos, width, height, flags))
    for i in range(lctSize):
        out.write(bytes([lct[i]]))

    # Image data
    minCodeSize = 2 ** ((flags >> 2) & 1) + 1
    lzwTable = []
    clearCode = 1 << minCodeSize
    endOfInformationCode = clearCode + 1
    nextCode = clearCode + 2
    lzwTable += [(code, bytes([code])) for code in range(clearCode)]
    lzwTable += [(nextCode, b'\x00')]
    out.write(bytes([minCodeSize]))
    while True:
        byte = rng.randint(0, 255)
        if len(lzwTable) < nextCode + 1 and byte!= endOfInformationCode:
            lzwTable += [(nextCode, bytes([byte]))]
            nextCode += 1
        elif byte == endOfInformationCode:
            break
        elif byte >= nextCode:
            raise Exception('Invalid LZW code!')

        prefix = lzwTable[byte][1]
        suffix = lzwTable[-1][1]
        if len(prefix) > 1 and prefix[:len(suffix)] == suffix:
            lzwTable += [(nextCode, prefix[:-1])]
            nextCode += 1
        else:
            lzwTable += [(nextCode, prefix + suffix)]
            nextCode += 1

        out.write(lzwTable[byte][1])

    # Trailer
    out.write(b';')

