#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + rng.read(1)[0] % 256
    height = 1 + rng.read(1)[0] % 256
    packed = (width & 0x7f).to_bytes(1, 'little') \
            + ((height >> 4) & 0xff).to_bytes(1, 'little') \
            + b'\x00\x00' \
            + (height & 0xf).to_bytes(1, 'little') \
            + b'\x00'
    out.write(packed)

    # Global color table
    for i in range(0, 2**((packed[1] >> 7) + 1)):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Image descriptor
    out.write(((0 << 7) | 0x00).to_bytes(1, 'little'))
    out.write(b'\x00\x00\x00\x00')
    out.write(struct.pack('<HH', width, height))
    out.write(b'\x00')

    # LZW minimum code size
    out.write(b'\x08')

    # Pixel data
    lzw_mincodesize = int.from_bytes(out.read(1), byteorder='big')
    lzw_maxcodesize = 1 << lzw_mincodesize
    lzw_clearcode = 1 << lzw_mincodesize
    lzw_endofinfo = lzw_clearcode + 1
    lzw_endofdata = lzw_endofinfo + 1
    lzw_firstfreecode = lzw_endofdata + 1
    lzw_nextcode = lzw_firstfreecode + 1
    lzw_prevcode = None
    lzw_curcode = None
    lzw_curbit = -1
    lzw_curbyte = 0
    lzw_curmask = 0x80
    lzw_curbits = []
    lzw_curbuf = [None] * 256
    lzw_curindex = 0
    lzw_curlength = 0
    lzw_curshift = 0
    lzw_curchar = 0
    lzw_curstack = []
    lzw_curlast = False
    lzw_curinited = True
    lzw_curused = set()
    lzw_curused.add(lzw_clearcode)
    lzw_curused.add(lzw_endofinfo)
    while len(lzw_curused) < lzw_maxcodesize:
        lzw_curcode = lzw_nextcode

        if not lzw_curinited and lzw_curcode == lzw_endofdata:
            break

        lzw_curinited = False

        if lzw_curcode >= lzw_firstfreecode:
            lzw_curchar = lzw_curbuf[-1]

            if lzw_curlast or lzw_curchar is None:
                raise Exception('Invalid GIF encoding detected!')

            lzw_curstack.append(lzw_curchar)
            lzw_curlast = False
        else:
            lzw_curchar = lzw_curcode

            if lzw_curcode == lzw_endofdata:
                lzw_curstack.append(lzw_curchar)

                while lzw_curchar!= lzw_clearcode:
                    lzw_curchar = lzw_curstack.pop(-1)

                    if lzw_curchar == lzw_endofdata:
                        continue

                    lzw_curstack.append(lzw_curchar)

                    if lzw_curchar <= lzw_firstfreecode:
                        break

                lzw_curlast = True
            elif lzw_curcode == lzw_clearcode:
                lzw_curlast = False
                lzw_curstack = []
            else:
                lzw_curlast = True
                lzw_curstack.append(lzw_curchar)

        lzw_curlength += 1

        if lzw_curlength > lzw_max