#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>BHHBBB',
        0b10000000 | ((width >> 7) & 0x01) << 6 | ((height >> 7) & 0x01) << 5,
        width % 0x8000,
        height % 0x8000,
        rng.read(1)[0] & 0x07,
        rng.read(1)[0],
        0b10000000
    )

    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 0x07 + 1
    global_color_table = []
    if global_color_table_size > 0:
        global_color_table = [
            rng.read(3) for _ in range(global_color_table_size)
        ]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>BBHHHBB',
        0x2C,
        rng.read(1)[0] & 0x07,
        x=0, y=0, w=width, h=height,
        packed=(
            0b10000000 |
            ((rng.read(1)[0] & 0x01) << 7) |
            ((rng.read(1)[0] & 0x01) << 6) |
            ((rng.read(1)[0] & 0x01) << 5)
        ),
    )

    # 5. Local Color Table
    local_color_table_size = rng.read(1)[0] & 0x07 + 1
    local_color_table = []
    if local_color_table_size > 0:
        local_color_table = [
            rng.read(3) for _ in range(local_color_table_size)
        ]

    # 6. Image Data
    # Each pixel is encoded as follows:
    # - The first bit indicates whether there's an extra byte after this one or not
    # - If it has an extra byte, then the second-to-last bit indicates whether that extra byte is a palette index
    #   or a length of a subblock to follow
    # - If it doesn't have an extra byte, then the last bit indicates whether the next byte is a palette index
    #   or a length of a subblock to follow
    # - If the last bit was set to zero, then we're done with the current byte
    # - Otherwise, we continue reading until we get to the end of the block
    # - We can also read the same byte multiple times, which means we can repeat blocks
    data = []
    while True:
        byte = rng.read(1)[0]

        # Check if we've reached the end of the image
        if byte == 0x00: break

        # Check if the last bit is set
        if byte & 0b10000000!= 0:
            # Read all remaining bytes from the stream
            while True:
                byte = rng.read(1)[0]

                # Check if the last bit is unset
                if byte & 0b10000000 == 0:
                    # Check if the second-to-last bit is set
                    if byte & 0b01000000!= 0:
                        # Repeat the previous byte
                        data += [data[-1]] * (byte & 0b00111111)

                    else:
                        # Add the byte to our list
                        data += [byte]

                    break

            # Continue with the next byte
            continue

        # Add the byte to our list
        data += [byte]

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    if len(global_color_table) > 0:
        out.write(bytes([len(global_color_table)]))
        for entry in global_color_table:
            out.write(entry)
    out.write(image_descriptor)
    if len(local_color_table) > 0:
        out.write(bytes([len(local_color_table)]))
        for entry in local_color_table:
            out.write(entry)
    out.write(bytes(data))
    out.write(trailer)