#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(num_colors * 3)

    # 4. Graphic Control Extension
    # We don't care about the delay time, so we just set it to zero
    gce = struct.pack('>BBBB', 0xF9, 4, rng.read(1)[0], 0)
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while len(blocks) < total_blocks_size:
        block_size = min(total_blocks_size - len(blocks), rng.read(1)[0])
        blocks.append(rng.read(block_size))

    # 7. Trailer
    trailer = b';'

    # Write everything in order
    out.write(header + lsd + color_table + gce + image_descriptor + bytes([lzw_minimum_code_size]))
    for block in blocks:
        out.write(block)
    out.write(trailer)

