#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while len(blocks) < total_blocks_size:
        block_size = min(total_blocks_size - len(blocks), rng.read(1)[0])

        # We'll use an LZW compression scheme to compress our data
        # This is just a simple implementation of that algorithm with some minor tweaks

        # First we need to build up a dictionary of all possible codes
        code_dictionary = {
            i : [i] + list(struct.pack('>H', i))
                for i in range(lzw_minimum_code_size, 2 ** lzw_minimum_code_size)
        }

        # Now let's start building our compressed stream
        compressed_stream = ''
        current_code = next(iter(code_dictionary))

        # Read from the input until we run out of data
        while True:
            # Get the next byte from the input
            try:
                new_byte = rng.read(1)[0].to_bytes(1, 'big')
            except IndexError:
                break

            # If this byte exists in any of our codes, use its corresponding code as the prefix
            for code in code_dictionary:
                if new_byte == code[0]:
                    compressed_stream += code

                    # Add the new byte to the end of the current code so we can continue matching on it later
                    current_code += new_byte

                    # Check if the current code has been completed yet; if so, add it to the dictionary
                    if current_code in code_dictionary:
                        raise Exception("Code not found")

                    code_dictionary[current_code] = code_dictionary[code] + [new_byte]
                    del code_dictionary[code]
                    
                    current_code = next(iter(code_dictionary))
                    break
            
            # Otherwise, create a new code starting with the current byte
            else:
                compressed_stream += current_code

                # Add the new byte to the end of the current code so we can continue matching on it later
                current_code += new_byte
                
                # Check if the current code has been completed yet; if so, add it to the dictionary
                if current_code in code_dictionary:
                    raise Exception("Code not found")

                code_dictionary[current_code] = [new_byte]
        
        # Finally, write the final code to the output
        compressed_stream += current_code

        # Split the compressed stream into chunks of 255 bytes max
        for i in range(0, len(compressed_stream), 255):
            blocks.append(compressed_stream[i : i + 255])
    
    # Write everything to the output
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    out.write(b''.join([len(block).to_bytes(1, 'big') + block for block in blocks]))

