#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = [0x7F & int(rng.read(1)[0]) | 0x80] + list(struct.unpack('>BBB', rng.read(3)))

    # 3. Global Color Table
    global_color_table_size = 2 << lsd[0] - 1
    global_color_table = []

    if global_color_table_size > 0:
        # Generate colors in RGB format
        for _ in range(global_color_table_size):
            r, g, b = tuple([int(rng.read(1)[0]) % 256 for _ in range(3)])

            # Add to table with transparency flag set
            global_color_table += [(r, g, b, 0)]

        # Pad remaining entries with black
        while len(global_color_table) < 2 ** (lsd[0] + 1):
            global_color_table += [(0, 0, 0, 0)]

    # Write everything to output
    out.write(header)
    out.write(bytes(bytearray(lsd)))
    out.write(b''.join([(bytes(bytearray((c[0], c[1], c[2])))) for c in global_color_table]))

    # 4. Image Descriptor
    image_descriptor = [width, height] + lsd[1:]
    out.write(bytes(bytearray(image_descriptor)))

    # 5. Local Color Table
    local_color_table_size = 2 << lsd[0] - 1
    local_color_table = []

    if local_color_table_size > 0:
        # Generate colors in RGB format
        for _ in range(local_color_table_size):
            r, g, b = tuple([int(rng.read(1)[0]) % 256 for _ in range(3)])

            # Add to table with transparency flag set
            local_color_table += [(r, g, b, 0)]

        # Pad remaining entries with black
        while len(local_color_table) < 2 ** (lsd[0] + 1):
            local_color_table += [(0, 0, 0, 0)]

    out.write(b''.join([(bytes(bytearray((c[0], c[1], c[2])))) for c in local_color_table]))

    # 6. Image Data
    # Set up frame buffer
    fb = [[None for x in range(height)] for y in range(width)]

    # Get initial background color
    bg_idx = lsd[1]

    # Fill frame buffer with background color
    for i in range(len(fb)):
        for j in range(len(fb[i])):
            fb[i][j] = bg_idx

    # Encode pixels as blocks of 8
    block_size = 8

    # Number of blocks per row
    num_blocks = ((width * height) // (block_size ** 2))

    # Encode all pixels
    for i in range(num_blocks):
        # Encode next block
        for j in range(block_size):
            for k in range(block_size):
                # Get pixel from frame buffer at current position
                p = fb[(i * block_size) + j][(i * block_size) + k]

                # If this is not the last pixel in the block, encode it as an index
                if (((i * block_size) + j)!= (height - 1)) or (((i * block_size) + k)!= (width - 1)):
                    out.write(bytes(bytearray([p])))

                    # Update frame buffer with new value
                    fb[(i * block_size) + j][(i * block_size) + k] = \
                        (bg_idx + 1) if (p == 255) else (p + 1)
                # Otherwise, just write a zero-length block
                else:
                    out.write(b'\x00')

    # Write end of stream marker
    out.write(b';')

