#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width, height = struct.unpack('>HH', rng.read(4))
    packed = rng.read(1)[0]
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBBBB', width, height, packed, background_color, aspect_ratio, rng.read(1)[0])

    # 3. Global Color Table (if present)
    global_color_table_flag = bool(packed & 0x80)
    if not global_color_table_flag:
        return

    color_resolution = ((packed >> 7) & 0b111) + 1
    sorted_flag = bool((packed >> 6) & 0b1)
    size_of_global_color_table = (2 ** color_resolution) - 1
    global_color_table_size = int.from_bytes(rng.read(1), 'big')
    global_color_table = []
    for i in range(global_color_table_size):
        r, g, b = [int(i * 255 / size_of_global_color_table) for i in rng.read(3)]
        global_color_table += [(r, g, b)]

    # 4. Image Descriptor
    image_left_position = int.from_bytes(rng.read(2), 'little')
    image_top_position = int.from_bytes(rng.read(2), 'little')
    image_width = int.from_bytes(rng.read(2), 'little')
    image_height = int.from_bytes(rng.read(2), 'little')
    packed = rng.read(1)[0]
    local_color_table_flag = bool(packed & 0b10000000)
    interlace_flag = bool(packed & 0b01000000)
    sort_flag = bool(packed & 0b00100000)
    reserved = rng.read(1)[0]
    size_of_local_color_table = (2 ** (((packed << 2) & 0b11100) >> 2)) - 1
    local_color_table = []
    for i in range(size_of_local_color_table):
        r, g, b = [int(i * 255 / size_of_local_color_table) for i in rng.read(3)]
        local_color_table += [(r, g, b)]

    # 5. LZW minimum code size
    min_code_size = rng.read(1)[0]

    # 6. Data subblocks
    data_subblock_size = 2 ** min_code_size
    data_subblocks = []
    while True:
        data_subblock = rng.read(data_subblock_size)

        # End of stream marker
        if len(data_subblock) == 0 or data_subblock[0] == 0x3B:
            break

        data_subblocks += [data_subblock]

    # 7. Trailer
    trailer = b';'

    # Write to output
    out.write(header)
    out.write(lsd)
    if global_color_table_flag:
        out.write(struct.pack('B', global_color_table_size))
        for rgb in global_color_table:
            out.write(struct.pack('BBB', *rgb))
    out.write(struct.pack('<HHHHBB', image_left_position, image_top_position, image_width, image_height, packed, min_code_size))
    if local_color_table_flag:
        for rgb in local_color_table:
            out.write(struct.pack('BBB', *rgb))
    for data_subblock in data_subblocks:
        out.write(data_subblock)
    out.write(trailer)


if