#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while len(blocks) < total_blocks_size:
        block_size = min(total_blocks_size - len(blocks), rng.read(1)[0])

        if not block_size:
            continue

        if rng.read(1)[0]:
            # RLE
            data = rng.read(1)[0].to_bytes(1, 'big') * block_size
        elif rng.read(1)[0]:
            # Raw
            data = rng.read(block_size)
        else:
            # LZW
            code_length = rng.read(1)[0]

            codes = [
                i for i in range(2 ** code_length)
            ]

            def get_next_code() -> int:
                return next(
                    i for i in codes
                    if rng.read(1)[0] == 1
                )

            def write_code(code: int):
                out.write((code & ((1 << code_length) - 1)).to_bytes(1, 'big'))

                new_codes = set()
                for i in range(len(codes)):
                    if i!= code:
                        new_codes.add(i)

                    if len(new_codes) >= 1 << code_length:
                        break

                assert len(new_codes) <= 1 << code_length
                codes = list(sorted(new_codes))

            prev_code = get_next_code()
            write_code(prev_code)

            for _ in range(block_size - 1):
                cur_code = get_next_code()

                if cur_code > prev_code:
                    write_code(cur_code)
                    prev_code = cur_code
                else:
                    write_code(prev_code)
                    write_code(cur_code)
                    prev_code = codes[cur_code]

            write_code(prev_code)

            data = b''

        blocks.append(data)

    image_data = struct.pack('>BB', lzw_minimum_code_size, 0) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)
