#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while len(blocks) < total_blocks_size:
        block_size = min(total_blocks_size - len(blocks), rng.read(1)[0])

        # If we're not in the first block, let's add some more data to the previous one
        if len(blocks):
            last_block = blocks[-1]

            # We can't just add more data to the previous block because that would break the compression algorithm
            # Instead, we need to create an entirely new block with the same code size as the previous one
            # The only exception is when the previous block has no data, then we can just extend its size instead of creating a new one
            if len(last_block):
                blocks.append(b'\x00\x00')

                # Add some more data to this block so that it will be larger than the other ones
                blocks.append(rng.read(block_size // 2))
            else:
                last_block += rng.read((block_size // 2) - len(last_block))
        else:
            blocks.append(rng.read(block_size))

    image_data = struct.pack('>BB', lzw_minimum_code_size, 0) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)