#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    header = b'GIF89a'
    screen_width, screen_height = struct.unpack('>HH', rng.read(4))
    global_color_table_flag = bool(rng.read(1)[0])
    color_resolution = rng.read(1)[0]
    sort_flag = bool(rng.read(1)[0])
    global_color_table_size = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    # 2. Logical Screen Descriptor
    lsd = struct.pack('>BHHBBBBB',
                      0x21,
                      0xF9,
                      len(header),
                      screen_width,
                      screen_height,
                      global_color_table_flag,
                      color_resolution,
                      sort_flag,
                      global_color_table_size,
                      background_color_index,
                      pixel_aspect_ratio)

    # 3. Global Color Table
    color_table = rng.read(global_color_table_size * 3)

    # 4. Graphic Control Extension
    gce = None
    if rng.read(1)[0]:
        delay_time = struct.unpack('>H', rng.read(2))[0]
        transparent_color_index = rng.read(1)[0]
        terminator = rng.read(1)[0]
        gce = struct.pack('>BBBB',
                          0x21,
                          0xF9,
                          delay_time,
                          transparent_color_index)

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved_flags = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = rng.read(local_color_table_size * 3)
    image_descriptor = struct.pack('>BHHHHBBBBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  local_color_table_size)

    # 6. Image Data
    # Let's use LZW compression to compress our data
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size - 1))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')