#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b"GIF89a")

    # Logical screen descriptor
    width = rng.read(2)[0] + 1
    height = rng.read(2)[0] + 1
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack("<HHBB", width, height, background_color_index, pixel_aspect_ratio))

    # Global color table
    global_color_table_flag = True
    if rng.read(1)[0] % 2 == 0:
        global_color_table_flag = False
    global_color_table_size = 7 + (rng.read(1)[0] % 8) * 3
    global_color_table = [None] * (2 ** global_color_table_size)
    for i in range(len(global_color_table)):
        red = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue = rng.read(1)[0] % 256
        global_color_table[i] = (red, green, blue)
    out.write(struct.pack("B", (global_color_table_flag << 7) | global_color_table_size))
    for rgb in global_color_table:
        out.write(bytes([rgb[0], rgb[1], rgb[2]]))

    # Image descriptor
    image_left_position = rng.read(2)[0]
    image_top_position = rng.read(2)[0]
    image_width = rng.read(2)[0] + 1
    image_height = rng.read(2)[0] + 1
    local_color_table_flag = True
    if rng.read(1)[0] % 2 == 0:
        local_color_table_flag = False
    interlace_flag = False
    sort_flag = False
    if rng.read(1)[0] % 2 == 0:
        interlace_flag = True
    if rng.read(1)[0] % 2 == 0:
        sort_flag = True
    local_color_table_size = 7 + (rng.read(1)[0] % 8) * 3
    local_color_table = [None] * (2 ** local_color_table_size)
    for i in range(len(local_color_table)):
        red = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue = rng.read(1)[0] % 256
        local_color_table[i] = (red, green, blue)
    out.write(struct.pack(
        "<HHHBBBB",
        image_left_position,
        image_top_position,
        image_width,
        image_height,
        0xF0 | ((local_color_table_flag << 7) | local_color_table_size),
        0x00,
        (interlace_flag << 6) | (sort_flag << 5)
    ))

    # Local color table
    for rgb in local_color_table:
        out.write(bytes([rgb[0], rgb[1], rgb[2]]))

    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0] % 8
    out.write(struct.pack("B", lzw_minimum_code_size))

    # Pixel data
    def get_pixel_data():
        while True:
            yield rng.read(1)[0] % 256
    pixel_data = get_pixel_data()
    def write_blocks():
        block = []
        while True:
            block.append(next(pixel_data))
            if len(block) >= 255:
                out.write(struct.