#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width = struct.unpack('<H', rng.read(2))[0] % 400
    height = struct.unpack('<H', rng.read(2))[0] % 300

    # Random background color index
    bg_color_index = rng.read(1)[0] % 256

    # Random pixel aspect ratio
    pixel_aspect_ratio = rng.read(1)[0] % 256

    # Global Color Table Flag
    global_color_table_flag = rng.read(1)[0] & 0x80 != 0

    # Color Resolution
    color_resolution = rng.read(1)[0] & 0x70 >> 4

    # Sort Flag
    sort_flag = rng.read(1)[0] & 0x08 != 0

    # Size of Global Color Table
    size_of_global_color_table = rng.read(1)[0] & 0x07

    # Pack all fields together
    lsd = struct.pack('<HHBBB', width, height, bg_color_index, pixel_aspect_ratio, global_color_table_flag | color_resolution << 4 | sort_flag << 3 | size_of_global_color_table)

    # 3. Global Color Table
    # If there is no global color table, skip this step
    if global_color_table_flag:
        # Read the correct amount of colors from /dev/urandom
        color_table = rng.read((2 ** (size_of_global_color_table + 1)) * 3)
    else:
        color_table = b''

    # 4. Graphic Control Extension
    # Always present
    gce = b'\x21\xf9\x04\x00\xff\x00\x00\x00\x00'

    # 5. Image Descriptor
    # Random left position and top position (two bytes each)
    left_position = struct.unpack('<H', rng.read(2))[0] % 400
    top_position = struct.unpack('<H', rng.read(2))[0] % 300

    # Width and Height (two bytes each)
    width = struct.unpack('<H', rng.read(2))[0] % 400
    height = struct.unpack('<H', rng.read(2))[0] % 300

    # Local Color Table Flag
    local_color_table_flag = rng.read(1)[0] & 0x80 != 0

    # Interlace Flag
    interlace_flag = rng.read(1)[0] & 0x40 != 0

    # Sort Flag
    sort_flag = rng.read(1)[0] & 0x08 != 0

    # Size of Local Color Table
    size_of_local_color_table = rng.read(1)[0] & 0x07

    # Pack all fields together
    image_descriptor = struct.pack('<HHHHBBB', left_position, top_position, width, height, local_color_table_flag | interlace_flag << 1 | sort_flag << 2 | size_of_local_color_table)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    image_data = []
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break

        image_data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to the output stream
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(image_data) + trailer)


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('./test.gif', 'wb') as f:
            generate_random_gif(rng, f)