#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# following structure:
#
#   Header
#   Logical Screen Descriptor
#   Global Color Table
#   Graphic Control Extension
#   Image Descriptor
#   Image Data
#   Trailer
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (two bytes each)
    lsd = struct.pack('>HHB', struct.unpack('>HH', rng.read(4))[0], struct.unpack('>HH', rng.read(4))[0], 0)

    # 3. Global Color Table
    color_table = rng.read(3 * 256)

    # 3. Graphic Control Extension
    # Random delay time (two bytes)
    gce = struct.pack('>BBHB', 0xf9, 4, struct.unpack('>H', rng.read(2))[0], 0)

    # 4. Image Descriptor
    # Random left and top position (two bytes each)
    image_descriptor = struct.pack('>HHB', struct.unpack('>HH', rng.read(4))[0], struct.unpack('>HH', rng.read(4))[0], 0)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    image_data = []
    while True:
        block_size = rng.read(1)[0]

        if not block_size:
            break

        image_data.append(rng.read(block_size))

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to the output stream
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(image_data) + trailer)


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('./test.gif', 'wb') as f:
            generate_random_gif(rng, f)