#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = None
    if rng.read(1)[0]:
        # Block size (always 4)
        gce = b'\x21\xF9\x04'

        # Packed fields (using a random byte for variety)
        packed_fields = rng.read(1)[0]

        # Delay time (two bytes, random)
        delay_time = struct.pack('>H', rng.read(2)[0])

        # Transparent color index (one byte, random)
        transparent_color_index = rng.read(1)[0]

        # Block terminator (always 0)
        gce += struct.pack('BBBB', packed_fields, delay_time[0], delay_time[1], transparent_color_index)

    # 5. Image Descriptor
    image_descriptor = b'\x2C'

    # Random position (two bytes for x, two bytes for y)
    image_position = struct.pack('>HH', rng.read(4)[0], rng.read(4)[0])

    # Width and height (again, two bytes for both)
    image_width, image_height = struct.unpack('>HH', rng.read(4))

    # Packed fields (using a random byte for variety)
    packed_fields = rng.read(1)[0]
    image_descriptor += image_position + struct.pack('>HHB', image_width, image_height, packed_fields)

    # 6. Local Color Table (if there is one)
    local_color_table = None
    if packed_fields & 0b10000000:
        num_local_colors = (rng.read(1)[0] % 256) or 1
        local_color_table = rng.read(3 * num_local_colors)

    # 7. Image Data
    # LZW minimum code size (always 8 bits)
    lzw_minimum_code_size = rng.read(1)[0]

    # Image data blocks
    blocks = []
    while True:
        # Block size (one byte)
        block_size = rng.read(1)[0]
        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size - 1))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')