#!/usr/bin/env python3

import struct
from typing import BinaryIO
from PIL import Image, ImageOps


def encode(inp: BinaryIO, out: BinaryIO):
    """Encode GIF file from input stream to output stream"""
    im = Image.open(inp).convert('RGB')

    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor (LSD)
    lsd = struct.pack("<HHBB", im.width, im.height, 0, 0)

    # 3. Global Color Table
    color_table = b''
    for i in range(256):
        color_table += struct.pack("BBB", *im.getpixel((i % 256, i // 256)))

    # 4. Graphic Control Extension (optional)
    gce = None
    if any(c[3] < 255 for c in im.getcolors()):
        gce = struct.pack("<BBHB", 0xF9, 4, 0, 0)

    # 5. Image Descriptor
    image_descriptor = struct.pack("<BBHHBB", 0x2C, 0, 0, im.width, im.height, 0)

    # 6. Image Data
    lzw_minimum_code_size = 8
    w, h = im.size
    pixels = list(im.getdata())
    while len(set(pixels)) <= 2 ** lzw_minimum_code_size and lzw_minimum_code_size < 12:
        lzw_minimum_code_size += 1

    rng = ImageOps.invert(im).convert('L').tobytes().__iter__()
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size - 1))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')