#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    # Convert to little-endian
    lsd = struct.pack('<HBBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    # Read in the global color table from the RNG
    color_table = rng.read(768)

    # 4. Graphic Control Extension
    # The first byte is always 0xF9, which indicates that this is a GCE
    gce = b'\xF9\x04\x00\x00\xFF\x00'

    # 5. Image Descriptor
    # First three bytes are always 0x2C, left position, top position
    image_descriptor = b'\x2C\x00\x00\x00'

    # 6. Image Data
    # We need to generate some LZW compressed data
    # This will be done by reading in a bunch of random bytes, then compressing them
    # We also need to ensure that we end with a code size of 0
    # To do so, we read in a random amount of bytes, then add a single zero byte
    # If we read in more than one byte, we prepend a new block size byte
    # Otherwise, we overwrite the last block size byte
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)