#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Optional Graphics Control Extension
    # We randomly decide whether we want one or not
    gce = None
    if rng.read(1)[0] & 1 == 0:
        # We do! Let's write it
        gce = struct.pack('BBBB', 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('>BBHHH', 0x2C, 0, 0, width, height)
    
    # 6. Image Data
    # We will use a random number of blocks, with random sizes
    num_blocks = rng.read(1)[0] + 1
    total_blocks_size = rng.read(1)[0] + 1
    blocks = []
    while num_blocks > 0:
        num_blocks -= 1
        
        # Each block has a random size
        block_size = rng.read(1)[0] + 1
        if block_size >= total_blocks_size:
            block_size = total_blocks_size - 1
            
        # And then we read that many bytes from /dev/urandom
        blocks.append(rng.read(block_size))
        
        total_blocks_size -= block_size
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)