#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    has_gce = bool((rng.read(1)[0] & 1 << 7) >> 7)
    if has_gce:
        gce = struct.pack('BBBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''
        
    # 5. Image Descriptor
    idd = struct.pack('>BBHH', rng.read(1)[0], rng.read(1)[0], width, height)
    
    # 6. Local Color Table (optional)
    has_lct = bool((idd[1] & 1 << 7) >> 7)
    if has_lct:
        lct = struct.pack('>BB', rng.read(1)[0], rng.read(1)[0])
    else:
        lct = b''
        
    # 7. Image Data
    image_data = rng.read(width * height)
    
    # 8. Trailer
    trailer = b';'
    
    # Write everything to the output file
    out.write(header + lsd + color_table + gce + idd + lct + image_data + trailer)
