#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # Color table (256 colors, 3 bytes per color)
    color_table = rng.read(3 * 256)

    # 3. Graphic Control Extension
    # Random delay time (two bytes)
    gce = struct.pack('>BBHB', 0xf9, 4, struct.unpack('>H', rng.read(2))[0], 0)

    # 4. Image Descriptor
    # Random left and top position (two bytes each)
    id = struct.pack('>HHB', struct.unpack('>HH', rng.read(4))[0], struct.unpack('>HH', rng.read(4))[0], 0)

    # Width and height of image (two bytes each)
    id += struct.pack('>HH', width, height)

    # Packed field (one byte)
    id += rng.read(1)

    # Local color table flag (one bit)
    id += rng.read(1)

    # Interlaced flag (one bit)
    id += rng.read(1)

    # Sort flag (one bit)
    id += rng.read(1)

    # Size of local color table (three bits)
    id += rng.read(1)

    # 5. Image Data
    # Random number of data sub-blocks (one byte)
    num_data_sub_blocks = rng.read(1)[0]
    data = []
    for i in range(num_data_sub_blocks):
        # Random length of data sub-block (one byte)
        data_sub_block_length = rng.read(1)[0]
        # Random data sub-block (of that length)
        data.append(rng.read(data_sub_block_length))

    # 6. Trailer
    trailer = b'\x3b'

    # Write it all to output
    out.write(header + lsd + color_table + gce + id + b''.join(data) + trailer)

