#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

def main() -> None:
    if len(sys.argv) != 2:
        print("Usage: {} <file>".format(os.path.basename(__file__)))
        return
    
    with open(sys.argv[1], "rb") as f:
        data = f.read()
        
    # Check magic number
    if not data[:4] == b"MZ\x90":
        raise ValueError("Invalid file format")
    
    # Skip the rest of the header
    offset = 64 + struct.unpack("<H", data[60:62])[0] - 4
    if offset > len(data):
        raise ValueError("Invalid file format")
    data = data[offset:]
    
    # Read the section headers
    num_sections = struct.unpack("<H", data[6:8])[0]
    sections = []
    for i in range(num_sections):
        sec = {
            "name": data[8+i*40:8+(i+1)*40].decode().rstrip("\x00"),
            "virtual_size": struct.unpack("<I", data[12+i*40:16+i*40])[0],
            "raw_size": struct.unpack("<I", data[16+i*40:20+i*40])[0],
            "raw_offset": struct.unpack("<I", data[20+i*40:24+i*40])[0],
            "relocs_offset": struct.unpack("<I", data[24+i*40:28+i*40])[0],
            "line_numbers_offset": struct.unpack("<I", data[28+i*40:32+i*40])[0],
            "number_of_relocs": struct.unpack("<H", data[32+i*40:34+i*40])[0],
            "number_of_line_numbers": struct.unpack("<H", data[34+i*40:36+i*40])[0],
            "characteristics": struct.unpack("<I", data[36+i*40:40+i*40])[0],
        }
        sections.append(sec)
    
    # Find the .text section
    text_section = next((sec for sec in sections if sec["name"] == ".text"), None)
    if text_section is None:
        raise ValueError(".text section not found")
    
    # Find the .idata section
    idata_section = next((sec for sec in sections if sec["name"] == ".idata"), None)
    if idata_section is None:
        raise ValueError(".idata section not found")
    
    # Find the .edata section
    edata_section = next((sec for sec in sections if sec["name"] == ".edata"), None)
    if edata_section is None:
        raise ValueError(".edata section not found")
    
    # Find the .rsrc section
    rsrc_section = next((sec for sec in sections if sec["name"] == ".rsrc"), None)
    if rsrc_section is None:
        raise ValueError(".rsrc section not found")
    
    # Find the .reloc section
    reloc_section = next((sec for sec in sections if sec["name"] == ".reloc"), None)
    if reloc_section is None:
        raise ValueError(".reloc section not found")
    
    # Find the .pdata section
    pdata_section = next((sec for sec in sections if sec["name"] == ".pdata"), None)
    if pdata_section is None:
        raise ValueError(".pdata section not found")
    
    # Find the .xdata section
    xdata_section = next((sec for sec in sections if sec["name"] == ".xdata"), None)
    if xdata_section is None:
        raise ValueError(".xdata section not found")
    
    # Find the .tls section
    tls_section = next((sec for sec in sections if sec["name"] == ".tls"), None)
    if tls_section is None:
        raise ValueError(".