#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a' + (b'\x00' * 6)
    out.write(header)

    # 2. Logical Screen Descriptor
    lsd = bytearray([
        0xFF, 0xFF,  # width
        0xFF, 0xFF,  # height
        0xF7,  # global color table flag
        0x00,  # background color index
        0x00   # pixel aspect ratio
    ])
    out.write(lsd)

    # 3. Global Color Table
    for i in range(0, 256):
        out.write(struct.pack('BBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]))

    # 4. Image Descriptor
    idesc = bytearray([
        0x2C,  # image separator
        0x00, 0x00,  # left position
        0x00, 0x00,  # top position
        0xFF, 0xFF,  # width
        0xFF, 0xFF,  # height
        0x00,  # local color table flag
        0x00,  # interlace flag
        0x00   # sort flag
    ])
    out.write(idesc)

    # 5. Local Color Table
    for i in range(0, 256):
        out.write(struct.pack('BBB', rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]))

    # 6. Raster Data
    while True:
        code = rng.read(1)[0]
        if code == 0x21:
            # Extension Block
            ext_code = rng.read(1)[0]
            if ext_code == 0xFE:
                # Comment Extension
                comment_length = int.from_bytes(rng.read(1), 'big')
                out.write(struct.pack('B', comment_length))
                out.write(rng.read(comment_length))
            elif ext_code == 0xFF:
                # Application Extension
                app_id = rng.read(8).decode()
                if app_id != "NETSCAPE":
                    raise Exception("Invalid application ID")
                block_size = rng.read(1)[0]
                if block_size != 0x03:
                    raise Exception("Invalid block size")
                subblocks = rng.read(1)[0]
                if subblocks != 0x01:
                    raise Exception("Invalid number of subblocks")
                loop_count = rng.read(2)
                out.write(struct.pack('BB', 0x21, 0xFF))
                out.write(app_id.encode())
                out.write(struct.pack('BB', block_size, subblocks))
                out.write(loop_count)
            else:
                raise Exception("Unknown extension code")
        elif code == 0x3B:
            break
        else:
            raise Exception("Invalid code")

    # 7. Trailer
    trailer = b'\x3B'
    out.write(trailer)

