#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    # Write everything to the output file
    out.write(header + struct.pack('>HBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio))

    # 3. Global Color Table
    # Read in the global color table from /dev/urandom
    global_color_table = rng.read(768)
    out.write(global_color_table)

    # 4. Image Data
    # Loop through the image data
    while True:
        # Read in the next block type
        block_type = rng.read(1)[0]

        if block_type == 0x2C:
            # Graphic Control Extension
            # Block size is always 4
            block_size = 4

            # Read in the rest of the GCX
            gcx = rng.read(block_size - 1)

            # Write it all to the output file
            out.write(struct.pack('BB', block_type, block_size) + gcx)
        elif block_type == 0x21:
            # Application Extension
            # Block size is always 11
            block_size = 11

            # Read in the rest of the AEX
            aex = rng.read(block_size - 1)

            # Write it all to the output file
            out.write(struct.pack('BB', block_type, block_size) + aex)
        elif block_type == 0x3b:
            # Trailer
            # Block size is always 0
            block_size = 0

            # Write it all to the output file
            out.write(struct.pack('BB', block_type, block_size))

            break
        else:
            raise Exception("Unknown block type")

