#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = rng.read(1)[0] % 8 + 1
    global_color_table_size *= 3
    global_color_table = rng.read(global_color_table_size)

    # 4. Image Descriptor
    # Random left and top position (using two bytes for each)
    image_left_position, image_top_position = struct.unpack('>HH', rng.read(4))

    # Width and height of the image (using two bytes for each)
    image_width, image_height = struct.unpack('>HH', rng.read(4))

    # Packed fields (using a random byte for variety)
    packed_fields = rng.read(1)[0]

    # Local color table flag (always false, since we don't use local color tables)
    local_color_table_flag = False

    # Interlaced flag (always false, since we don't use interlacing)
    interlaced_flag = False

    # Sort flag (always true, since we want to sort the color table)
    sort_flag = True

    # Size of local color table (always zero, since we don't use local color tables)
    local_color_table_size = 0

    idd = struct.pack('>HHBBBBB', image_left_position, image_top_position, image_width, image_height, packed_fields, local_color_table_flag, interlaced_flag, sort_flag, local_color_table_size)

    # 5. Graphic Control Extension
    # Always present, always has a delay of 10ms
    gce = b'\x21\xf9\x04\x01\x0a\x00'

    # 6. Image Data
    # We'll just write the raw data from /dev/urandom here
    image_data = rng.read(image_width*image_height)

    # 7. Trailer
    trailer = b'\x3b'

    return len(lsd+gce+trailer)+len(idd)+len(global_color_table)+len(image_data)
