#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    header = b"GIF89a"
    width = rng.read(2)
    height = rng.read(2)
    flags = rng.read(1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)
    out.write(header + width + height + flags + background_color_index + pixel_aspect_ratio)

    # 2. Logical Screen Descriptor (LSD)
    global_color_table_flag = bool(flags[0] & 0b10000000)
    color_resolution = flags[0] >> 4 & 0b111
    sort_flag = bool(flags[0] & 0b100)
    global_color_table_size = 2 ** ((flags[0] & 0b111) + 1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)
    lsd = width + height + flags + background_color_index + pixel_aspect_ratio

    # 3. Global Color Table
    color_table = b""
    if global_color_table_flag:
        for _ in range(global_color_table_size):
            red = rng.read(1)[0]
            green = rng.read(1)[0]
            blue = rng.read(1)[0]
            color_table += bytes([red, green, blue])

    # 4. Graphic Control Extension (optional)
    gce = None
    if rng.read(1)[0]:
        extension_introducer = b"\x21\xF9"
        graphic_control_label = b"\x04"
        disposal_method = rng.read(1)[0]
        user_input_flag = bool(disposal_method & 0b10000000)
        transparent_color_flag = bool(disposal_method & 0b1000000)
        delay_time = struct.pack(">H", rng.read(2)[0] | rng.read(2)[0] << 8)
        transparent_color_index = rng.read(1)
        block_terminator = b"\x00"
        gce = extension_introducer + graphic_control_label + bytes([disposal_method]) + delay_time + transparent_color_index + block_terminator

    # 5. Image Descriptor
    image_separator = b"\x2C"
    left_position = rng.read(2)
    top_position = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    local_color_table_flag = bool(rng.read(1)[0] & 0b10000000)
    interlace_flag = bool(rng.read(1)[0] & 0b1000000)
    sort_flag = bool(rng.read(1)[0] & 0b100)
    local_color_table_size = 2 ** ((rng.read(1)[0] & 0b111) + 1)
    image_descriptor = image_separator + left_position + top_position + width + height + rng.read(1)

    # 6. Local Color Table
    color_table = b""
    if local_color_table_flag:
        for _ in range(local_color_table_size):
            red = rng.read(1)[0]
            green = rng.read(1)[0]
            blue = rng.read(1)[0]
            color_table += bytes([red, green, blue])

    # 7. Image Data
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size - 1))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')