#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = rng.read(1)[0] % 8 + 1
    global_color_table_bytes = rng.read(3 * 2 ** global_color_table_size)
    
    # 4. Image Data
    # Let's make it have random size, capped at 256 colors, and fully random
    image_data_size = rng.read(1)[0] % 8 + 1
    
    # We'll use this to keep track of how many bytes we need to write
    image_data_bytes_left = 2 ** image_data_size - 1
    
    # We'll also need to keep track of where in the global color table we are
    global_color_table_index = 0
    
    # Write everything to `out`
    out.write(header)
    out.write(lsd)
    out.write(global_color_table_bytes)
    
    while True:
        # 4.1 Graphic Control Extension
        
        # We're going to write a Graphic Control Extension with a random delay time
        gce = b'\x21\xf9\x04' + rng.read(2) + b'\x00'
        
        # 4.2 Image Descriptor
        id = b'\x2c' + struct.pack('>HHBB', width, height, 0, 0)
        
        # 4.3 Local Color Table (if needed)
        if image_data_size != global_color_table_size:
            local_color_table_bytes = rng.read(3 * 2 ** image_data_size)
            
            # 4.4 Image Data
            # We'll just use the same color table as before
            image_data_bytes = rng.read(image_data_bytes_left)
            
            # Write everything to `out`
            out.write(gce)
            out.write(id)
            out.write(local_color_table_bytes)
            out.write(image_data_bytes)
            
            # Update our state variables
            image_data_bytes_left -= len(image_data_bytes)
            global_color_table_index += 2 ** image_data_size
        else:
            # 4.4 Image Data
            # We'll just use the same color table as before
            image_data_bytes = rng.read(image_data_bytes_left)
            
            # Write everything to `out`
            out.write(gce)
            out.write(id)
            out.write(image_data_bytes)
            
            # Update our state variables
            image_data_bytes_left -= len(image_data_bytes)
            global_color_table_index += 2 ** image_data_size
        
        # If there is no more data left to write...
        if image_data_bytes_left == 0:
            # ...then we're done!
            break

# Generate a random GIF file of length `length` into `out`, using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif_of_length(
        rng: BinaryIO, out: BinaryIO, length: int
    ) -> None:
    # Keep generating random GIF files until we get one that's long enough
    while True:
        # Generate a random GIF file
        generate_random_gif(rng, out)
        
        # Get the current position in the output file
        pos = out.tell()
        
        # Seek back to the beginning of the file
        out.seek(0)
        
        # Read all the data from the file
        data = out.read()
        
        # Seek back to the end of the file
        out.seek(pos)
        
        # If the file is long enough...
        if len(data) >= length:
            # ...then stop generating random GIF files
            break

# Generate a random GIF file of length `length` into `out`, using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif_of_length_with_seed(
        seed: int, out: BinaryIO