#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # Color table (256 colors, 3 bytes per color)
    color_table = rng.read(3 * 256)

    # 3. Graphic Control Extension
    # Random delay time (two bytes)
    gce = struct.pack('>BBHB', 0xf9, 4, struct.unpack('>H', rng.read(2))[0], 0)

    # 4. Image Descriptor
    # Random left and top position (two bytes each)
    image_descriptor = struct.pack('>HHB', struct.unpack('>HH', rng.read(4))[0], struct.unpack('>HH', rng.read(4))[0], 0)

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    image_data = []
    while True:
        block_size = rng.read(1)[0]

        if block_size == 0:
            break

        image_data.append(rng.read(min(block_size, 255)))

    # 7. Trailer
    trailer = b';'

    # Write everything to the output stream
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(image_data) + trailer)


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('./test.gif', 'wb') as f:
            generate_random_gif(rng, f)