#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + (rng.read(1)[0] % 256)
    height = 1 + (rng.read(1)[0] % 256)
    flags = b'\x87\x00' if width < height else b'\x00\x87'
    background = bytes([rng.read(1)[0]]) * 3
    aspect_ratio = b'\x00'
    out.write(struct.pack('<HHBBB', width, height, flags[0], flags[1],
                           background[0], background[1], background[2]))
    out.write(aspect_ratio)

    # Global color table
    palette = [bytes([rng.read(1)[0]] * 3) for _ in range(256)]
    out.write(b''.join(palette))

    # Image descriptor
    left = 0
    top = 0
    width = 1 + (rng.read(1)[0] % 256)
    height = 1 + (rng.read(1)[0] % 256)
    flags = b'\x87\x00' if width < height else b'\x00\x87'
    local_color_table = False
    interlace = False
    sort = True
    reserved = b'\x00\x00'
    lct_size = 0
    out.write(struct.pack('<HHHHBB', left, top, width, height, flags[0],
                          flags[1], lct_size))
    out.write(reserved)
    out.write(b'\x00' if not local_color_table else b'\x01')
    out.write(b'\x00' if not interlace else b'\x40')
    out.write(b'\x00' if not sort else b'\x20')
    out.write(b'\x00' * 3)

    # Local color table
    if local_color_table:
        palette = [bytes([rng.read(1)[0]] * 3) for _ in range(lct_size)]
        out.write(b''.join(palette))

    # LZW minimum code size
    min_code_size = 2 ** (rng.read(1)[0] & 0x07) + 1
    out.write(min_code_size.to_bytes(1, 'big'))

    # Data sub-blocks
    data_block = bytearray()
    while len(data_block) < 256:
        block_type = rng.read(1)[0]

        # End of image
        if block_type == 0x3B:
            break

        # Plain text
        elif block_type == 0x01:
            length = rng.read(1)[0]

            # Escape sequence
            if length == 0x00:
                data_block += b'\x00\x00'

            # Raw data
            else:
                data_block += rng.read(length)

        # Graphic control extension
        elif block_type == 0xF9:
            length = rng.read(1)[0]
            label = rng.read(length)
            length = rng.read(1)[0]
            data = rng.read(length)
            assert label == b'\x21\xF9\x04' and data == b'\x00\x00' \
                   or label == b'\x21\xFE\x00' and data == b''

        # Comment
        elif block_type == 0xFE:
            length = rng.read(1)[0]
            comment = rng.read(length)

        # Application extension
        elif block_type == 0xFF:
            length = rng.read(1)[0]
            app_label = rng.read(length)
            length = rng.read(1)[0]
            app_data = rng.read(length)
            assert app_label == b'\x21\xFF\x0BnetSCAPE2.0' and \
                    app_data[:3] == b'\x03\x01\x00'

        # Unknown
        else:
            raise Exception('Unknown block type {:02X}'.format(block_type))

    # Write data blocks
    while len(data_block) > 0:
        block_size = min(len(data_block), 255)
        out