#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Use fixed color table size of 7 bits per pixel
    color_table_size = 0x7F

    # No global color table
    background_color = 0x00
    aspect_ratio = 0x00

    screen_descriptor = struct.pack(
        '>HHBBB',
        width, height,
        color_table_size,
        background_color,
        aspect_ratio
    )

    # 3. Global Color Table
    # Generate a palette with 256 colors
    palette = [struct.pack('BBB', *rng.read(3)) for _ in range(256)]

    # Write the palette to the output stream
    out.write(b''.join(palette))

    # 4. Image Descriptor
    image_separator = b','

    left = 0
    top = 0
    width = 1 << ((width & 0x700) >> 8) | (width & 0xFF)
    height = 1 << ((height & 0x700) >> 8) | (height & 0xFF)
    flags = 0x8C    # LZW compression, no interlacing
    local_color_table_flag = 0x01   # Local color table is present
    sort_flag = 0x00                # Not sorted
    reserved = 0x00                 # Reserved field set to zero
    lct_size = len(palette) - 1     # Size of the local color table

    image_descriptor = struct.pack(
        '>HHHHBB',
        left, top,
        width, height,
        flags,
        local_color_table_flag,
        sort_flag,
        lct_size
    )

    # 5. Image Data
    # The first byte is the minimum code size used by LZW compression
    min_code_size = 0x08

    # Write the image descriptor to the output stream
    out.write(image_separator + image_descriptor)

    # Write the minimum code size to the output stream
    out.write(min_code_size.to_bytes(1, 'big'))

    # Generate an initial string of length 1
    previous_string = b'\x00'

    while True:
        # Generate a new character from /dev/urandom
        current_character = rng.read(1)[0]

        if current_character == 0x00 or \
           current_character >= 0xF9:
            # If we get either null-byte or control codes,
            # write them as literals

            # Write the literal prefix to the output stream
            out.write((current_character).to_bytes(1, 'big'))

            # Add the generated character to the end of the previous string
            previous_string += current_character.to_bytes(1, 'big')

        else:
            # Otherwise, use LZW compression

            # Find the longest match between the last characters written
            # and the current one
            matches = []

            for i in range(len(previous_string)):
                if previous_string[-i:] == current_character.to_bytes(
                        1,
                        'big'
                    ):

                    matches.append(i)

            # Choose the longest match
            match_length = max(matches)

            # Write the match prefix to the output stream
            out.write(((match_length - 1) // 256 + 1).to_bytes(
                1,
                'big'
            ))
            out.write(((match_length - 1) % 256 + 1).to_bytes(
                1,
                'big'
            ))

            # Write the matched part of the previous string to the output stream
            out.write(previous_string[-match_length:])

            # Remove the matched part from the previous string
            previous_string = previous_string[:-match_length]

            # Append the current character to the previous string
            previous_string += current_character.to_bytes(1, 'big')

        # Check whether we have enough data to write another chunk
        if len(previous_string) > 255:
            continue

        # Write the next chunk to the output stream
        out.write(previous_string)

        break

    # 6. Trailer
    trailer = b';'

    # Write the trailer to the output stream
    out.write(trailer)

