#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    lsd = struct.pack(
        '>HHBBB',
        width,
        height,
        packed_fields,
        background_color_index,
        pixel_aspect_ratio
    )

    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 0x7
    if global_color_table_size!= 0:
        global_color_table = [
            rng.read(3) + b'\x00'
            for _ in range(global_color_table_size ** 2)
        ]

        gct = b''.join(global_color_table)
    else:
        gct = b''

    # 4. Image Descriptor
    image_left_position = rng.read(2)
    image_top_position = rng.read(2)
    image_width = rng.read(2)
    image_height = rng.read(2)

    packed_fields = rng.read(1)[0]
    local_color_table_flag = bool((packed_fields >> 7) & 0x1)
    interlace_flag = bool((packed_fields >> 6) & 0x1)
    sort_flag = bool((packed_fields >> 5) & 0x1)
    reserved_bits = ((packed_fields >> 3) & 0x3) << 1
    local_color_table_size = (packed_fields & 0x7)

    id = struct.pack(
        '>HBBHHB',
        image_left_position[0],
        image_left_position[1],
        image_top_position[0],
        image_top_position[1],
        image_width[0],
        image_width[1],
        image_height[0],
        image_height[1]
    )

    # 5. Local Color Table
    if local_color_table_flag:
        local_color_table = [
            rng.read(3) + b'\x00'
            for _ in range(local_color_table_size ** 2)
        ]

        lct = b''.join(local_color_table)
    else:
        lct = b''

    # 6. Image Data
    data = []
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x2C:
            # Image descriptor
            pass
        elif block_type == 0x21:
            # Extension
            extension_label = rng.read(1)[0]

            if extension_label == 0xF9:
                # Graphics Control Extension
                size = rng.read(1)[0]

                packed_fields = rng.read(1)[0]
                disposal_method = (packed_fields >> 2) & 0x7
                user_input_flag = bool((packed_fields >> 1) & 0x1)
                transparent_color_flag = bool(packed_fields & 0x1)

                delay_time = rng.read(2)
                transparent_color_index = rng.read(1)

                edt = struct.pack(
                    '>BBHB',
                    size,
                    0xF9,
                    packed_fields,
                    delay_time[0],
                    delay_time[1],
                    transparent_color_index[0]
                )
            elif extension_label == 0xFE:
                # Comment Extension
                comment_block_size = rng.read(1)[0]
                comment_data = rng.read(comment_block_size)

                edt = struct.pack(
                    '>BB%ds' % len(comment_data),
                    comment_block_size,
                    0xFE,
                    comment_data
                )
            elif extension_label == 0xFF:
                # Application Extension
                application_identifier = rng.read(8)
                application_auth_code = rng.read(3)
                application_block_size = rng.read(1)[0]
                application_data = rng.read(application_block_size)

                edt = struct.pack(
                    '>BBBB8s3s%ds' % len(application_data),
                    application_block_size,
                    0xFF,
                    *application_identifier,
                    *application_auth_code,
                    application_data
                )
            else:
                raise Exception('Unknown extension label')
        elif block_type == 0x3B:
            break
        else:
            raise Exception('Unknown block type')

        data += [struct.pack('>BB', block_type, 0)]

    # 7. Trailer
    trailer = b';'

    # Write everything to output
    out.write(header)
    out.write(