#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # Color table (256 colors, 3 bytes per color)
    color_table = rng.read(3 * 256)

    # 3. Graphic Control Extension
    # Random delay time (two bytes)
    gce = struct.pack('>BBHB', 0xf9, 4, struct.unpack('>H', rng.read(2))[0], 0)

    # 4. Image Descriptor
    # Random left and top position (two bytes each)
    image_descriptor = struct.pack('>HHB', struct.unpack('>HH', rng.read(4))[0], struct.unpack('>HH', rng.read(4))[0], 0)

    # 5. Image data
    # Frames are made of 1-byte blocks that specify how many times to repeat the next pixel
    # The first frame is made of 1 block with value 127
    # The second one has 2 blocks with values 127 and 127
    # The third one has 3 blocks with values 127, 127 and 127
    #...
    # And so on until we reach the maximum size allowed by the LZW compression algorithm (~128 KB)
    # We also add an extra block at the end with a value of zero to mark the end of the image
    frames = [b'\x7f'] + ([b'\x7f\x7f'] * int((128*1024 - len(frames[0])) / 2)) + [b'\x00']

    # Each frame is compressed using LZW
    # We use a fixed dictionary size of 256 words
    # This means that any sequence of up to 256 words can be encoded in the dictionary
    # If it's not possible, then we just write the word as-is without compressing it
    def encode_lzw(word):
        if len(dictionary) < 256 or word == dictionary[-1]:
            return bytes([len(dictionary)]) + word

        i = dictionary.index(word)
        dictionary.append(word)
        return bytes([(i+1) & 0xff])

    dictionary = []
    for frame in frames:
        compressed_frame = b''
        previous_word = ''
        for c in frame:
            current_word = previous_word + chr(c)

            try:
                compressed_frame += encode_lzw(current_word)
            except ValueError:
                compressed_frame += bytes([len(previous_word)]) + previous_word

            previous_word = current_word[:1]

        compressed_frame += bytes([len(previous_word)]) + previous_word
        frames[frames.index(frame)] = compressed_frame

    # Finally, we concatenate all the frames together
    image_data = b''.join(frames)

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data)


if