#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    has_gce = bool((rng.read(1)[0] & 1 << 7) >> 7)
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, 255)
        
        # Generate a random block of data with a random LZW minimum code size
        block = [lzw_minimum_code_size | rng.read(1)[0]] + list(rng.read(block_size - 1))
        
        # Add to our running tally of all blocks so far
        blocks += block
        
        # Subtract from our remaining space in this frame
        total_blocks_size -= len(block)
    
    # Encode the blocks as LZW
    encoded_data = encode_lzw(blocks)
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    out.write(encoded_data)


# Encodes an array of integers into LZW format
def encode_lzw(
        arr: list[int]
    ) -> bytes:
    # Start with a blank dictionary
    dict = {}
    
    # The first element is always present in the dictionary; its value is the empty string
    dict[arr[0]] = ''
    
    # For every other element...
    for i in range(1, len(arr)):
        # Get the previous entry
        prev = arr[i - 1]
        
        # Get the current entry
        curr = arr[i]
        
        # If we've seen this pair before...
        if f'{prev}{curr}' in dict:
            # Use that entry instead
            arr[i] = dict[f'{prev}{curr}']
            
            continue
        
        # Otherwise, add it to the dictionary
        dict[f'{prev}{curr}'] = str(len(dict))
        
        # And use the new entry
        arr[i] = dict[f'{prev}{curr}']
    
    return bytes([len(dict)]) + ''.join(map(str.encode, arr)).decode().encode()