#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given PRNG instance. The generated GIF will be
# valid and have no errors according to the spec.
def generate_gif(out: BinaryIO, rng):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed = bytearray([
        0 | ((width[0] & 0b11111000) >> 3),   # Width
        0 | (((width[0] & 0b00000111) << 3) | ((height[0] & 0b11111000) >> 5)),  # Height
        0 | ((height[0] & 0b00000111) << 3),      # Packed field
        rng.read(1)[0],                         # Background color index
        0 | (rng.read(1)[0] & 0b00000111),       # Pixel aspect ratio
    ])
    lsd = bytes(packed)

    # 3. Global Color Table
    global_color_table_flag = bool(rng.read(1)[0])
    color_resolution = (rng.read(1)[0] & 0b11110000) >> 4
    sort_flag = bool((rng.read(1)[0] & 0b00001000) >> 3)
    size_of_global_color_table = 2 ** (rng.read(1)[0] & 0b00000111)
    global_color_table = [rng.read(3) for _ in range(size_of_global_color_table)]
    color_table = b''
    if global_color_table_flag:
        for rgb in global_color_table:
            color_table += rgb
    
    # 4. Graphics Control Extension
    has_gce = bool(rng.read(1)[0])
    gce = None
    if has_gce:
        gce = bytearray([
            0 | (rng.read(1)[0] & 0b11111111),     # Block size
            0 | (rng.read(1)[0] & 0b11111111),     # Flags
            0 | (rng.read(2)[0]),                 # Delay time
            0 | (rng.read(1)[0] & 0b11111111),     # Transparent color index
            0 | (rng.read(1)[0] & 0b11111111),     # Terminator
        ])
    
    # 5. Image Descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    packed = bytearray([
        0 | ((left[0] & 0b11111000) >> 3),        # Left position
        0 | (((left[0] & 0b00000111) << 3) | ((top[0] & 0b11111000) >> 5)),  # Top position
        0 | ((top[0] & 0b00000111) << 3),         # Packed field
        rng.read(1)[0],                          # Local color table flag
        0 | (rng.read(1)[0] & 0b00000111),        # Interlace flag
        0 | (rng.read(1)[0] & 0b00000111),        # Sort flag
        0 | (rng.read(1)[0] & 0b00000111),        # Reserved
    ])
    image_descriptor = bytes(packed)
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0] + 1)
        blocks.append(rng.read(block_size))
        
        total_blocks_size -= block_size
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)