#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    # Pack all these values together in an LSD
    lsd = struct.pack('>HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    global_color_table_flag = bool(packed_fields & 0b10000000)
    if not global_color_table_flag:
        return
    
    # Number of colors used from the global color table
    global_color_table_size = 2 << ((packed_fields >> 7) & 0b111)

    # Generate the global color table
    global_color_table = [struct.unpack('>H', rng.read(2))[0] for _ in range(global_color_table_size)]

    # Add padding bytes until we reach a multiple of 3 bytes
    while len(global_color_table) % 3!= 0:
        global_color_table.append(0)

    # Convert the global color table to bytes
    color_table = bytes()
    for i in range(0, len(global_color_table), 3):
        color_table += struct.pack('>BBB', *global_color_table[i : i+3])

    # 4. Graphics Control Extension
    graphics_control_extension_flag = bool(packed_fields & 0b00001000)
    if not graphics_control_extension_flag:
        return

    # Block size is always 0x04
    block_size = rng.read(1)[0]
    assert block_size == 0x04

    # Packed field that contains various flags
    packed_field = rng.read(1)[0]

    # Delay time (in hundredths of a second)
    delay_time = struct.unpack('>H', rng.read(2))[0]

    # Transparent color index
    transparent_color_index = rng.read(1)[0]

    # Terminator
    terminator = rng.read(1)[0]
    assert terminator == 0x00

    # Pack all these values together in a GCE
    gce = struct.pack('>BBHB', block_size, packed_field, delay_time, transparent_color_index)

    # 5. Image Descriptor
    image_separator = rng.read(1)[0]
    assert image_separator == 0x2C

    # Left position
    left = struct.unpack('>H', rng.read(2))[0]

    # Top position
    top = struct.unpack('>H', rng.read(2))[0]

    # Width
    width = struct.unpack('>H', rng.read(2))[0]

    # Height
    height = struct.unpack('>H', rng.read(2))[0]

    # Packed field that contains various flags
    packed_field = rng.read(1)[0]

    # Local color table flag
    local_color_table_flag = bool(packed_field & 0b10000000)

    # Interlace flag
    interlace_flag = bool(packed_field & 0b01000000)

    # Sort flag
    sort_flag = bool(packed_field & 0b00100000)

    # Reserved bits must be zeroes
    reserved_bits = packed_field & 0b00011111
    assert reserved_bits == 0

    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0]

    # Pack all these values together in an ID
    image_descriptor = struct.pack('>BBBBHHBB', image_separator, left, top, width, height, packed_field, lzw_minimum_code_size)

    # 6. Image Data
    # Get the length of the image data
    image_data_length = int((width / 8) * height)

    # Read the raw image data
    image_data = rng.read(image_data_length)

    # Compress it using LZW
    compressed_image_data = compress_lzw(image_data, lzw_minimum_code_size)

    # Split the compressed image data into chunks of up to 255 bytes
    blocks = []
    while True:
        data = compressed_image_data[:255]
        compressed_image_data = compressed_image_data[255:]

        if len(compressed_image_data) == 0 or len(            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)