#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    # If we don't have enough colors in our table, pad it with black
    color_table += b'\x00\x00\x00' * ((256 - len(color_table)) // 3)

    # 4. Graphics Control Extension (if applicable)
    gce = None
    if bool(packed_fields & 0b1):  # Has global color table flag set?
        disposal_method = rng.read(1)[0]
        user_input_flag = rng.read(1)[0]
        transparent_color_index = rng.read(1)[0]
        delay_time = struct.unpack('>H', rng.read(2))[0]
        transparent_color_flag = rng.read(1)[0]
        gce = struct.pack('>BBBBBHBB',
                         0x21,
                         ord('G'),
                         ord('C'),
                         4,
                         disposal_method,
                         user_input_flag,
                         transparent_color_index,
                         delay_time >> 8,
                         delay_time & 0xFF,
                         transparent_color_flag
                        )

    # 5. Image Descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    packed_fields = rng.read(1)[0]
    local_color_table_size = (packed_fields & 0b11110000) >> 4
    interlace_flag = (packed_fields & 0b1000)!= 0
    sorted_flag = (packed_fields & 0b0100)!= 0
    reserved_flags = (packed_fields & 0b1100) << 2
    local_color_table_flag = (packed_fields & 0b0001)!= 0
    image_descriptor = struct.pack('>HHHHBBBB',
                                   left[0],
                                   left[1],
                                   top[0],
                                   top[1],
                                   width[0],
                                   width[1],
                                   height[0],
                                   height[1],
                                   local_color_table_flag | \
                                       interlace_flag << 7 | \
                                       sort_flag,
                                  reserved_flags,
                                  local_color_table_size
                                 )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        block_size = rng.read(1)[0]  # Each block can be up to 255 bytes

        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size - 1))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')