#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0] & 1)
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0] & 1)
    interlace_flag = bool(rng.read(1)[0] & 1)
    sort_flag = bool(rng.read(1)[0] & 1)
    reserved_flags = rng.read(1)[0]
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    image_descriptor = struct.pack(
        '>HHBBBBB', left, top, width, height, local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved_flags, lzw_minimum_code_size
    )

    # 6. Local Color Table (optional)
    if local_color_table_flag:
        num_colors = (rng.read(1)[0] % 256) or 1
        local_color_table = rng.read(3 * num_colors)  # Random colors
    else:
        local_color_table = b''

    # 7. Image Data
    # We'll use a simple LZW compression scheme to compress our data
    # First we need to build an LZW dictionary of all possible codes
    # Then we can just write code literals until we reach EOI
    # The first few codes are special-cased as clear code, end of information code, and end of block code
    # See https://www.matthewflickinger.com/lab/whatsinagif/lzw_image_data.asp for more info on this
    # Also see https://en.wikipedia.org/wiki/Lempel%E2%80%93Ziv%E2%80%93Welch for more info on LZW compression in general
    def get_next_code() -> int:
        return next(codes)

    codes = iter(range(2**lzw_minimum_code_size + 2))
    clear_code = next(codes)
    end_of_information_code = next(codes)
    end_of_block_code = next(codes)
    dict_size = len(codes) - 1
    dict = {chr(i): i for i in range(dict_size)}
    dict['EOI'] = end_of_information_code
    dict['EOB'] = end_of_block_code
    dict['CLEAR'] = clear_code

    # Now let's start writing some compressed data!
    # We're going to be lazy here and not bother with any fancy bit packing
    # Instead we'll just do everything manually
    # This is because we want to control exactly how many bytes we write
    # So that we can test our decompression implementation later
    # And also so that we don't have to worry about padding bits
    # Because we know there will always be enough space for them
    # But we could still add support for those if needed
    # Let's start by writing the clear code
    out.write(struct.pack('>B', clear_code))

    # Next up we'll write the entire string 'Hello World!'
    # Each character gets its own code
    # Since we already wrote the clear code above, we can assume that the first character will be encoded with length 1
    # If you look closely, you'll notice that the first character is 'H'.
    # That means we should expect the following code to be written:
    #   H --> 0b00000000 --> 0x00
    # However, since we haven't added anything yet, we won't actually see the code written to the output stream
    # Instead, we'll store it in memory temporarily
    # Once we've seen enough characters to fill the current code length, then we'll flush the stored codes to the output stream
    # In this case, we'll only have one character left before we hit the maximum code length of 12 bits
    # So when we try to encode the second character, we'll find that we don't have enough codes to fill the code length
    # So instead we'll write another clear code to reset the encoder state
    # After writing the clear code, we'll continue encoding the rest of the string
    # Note that we're not actually checking if the code lengths are valid
    # For example, we might have a code length of 12 but no codes to write
    # Or we might have a code length of 12 but the last code was too long
    # We'll just assume that the caller is doing the right thing here
    hello_world = 'Hello World!'
    code_length = 1
    prev_char = ''
    for char in hello_world:
        if char!= prev_char:
            if code_length == 12:
                out.write(struct.pack('>B', clear_code))

            code = dict[prev_char + char]
            dict[prev_char + char] = dict_size
            dict_size += 1
            code_length = 1

        elif code_length < 12:
            code_length += 1

        out.write(struct.pack('>B', code))
        prev_char = char

    # Finally, we'll write