#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = rng.read(2)[0] & 0xFFF | ((rng.read(1)[0] & 0x07) << 12)
    height = rng.read(2)[0] & 0xFFF | ((rng.read(1)[0] & 0x07) << 12)
    color_table_flag = rng.read(1)[0] >> 7
    global_color_table_size = (rng.read(1)[0] & 0x07) + 1
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HBBBBB', width, height, color_table_flag, global_color_table_size, background_color_index, pixel_aspect_ratio)
    
    # 3. Global Color Table (optional)
    color_table = bytearray()
    if color_table_flag == 1:
        for i in range(global_color_table_size):
            red   = rng.read(1)[0]
            green = rng.read(1)[0]
            blue  = rng.read(1)[0]
            
            color_table += struct.pack('BBB', red, green, blue)
        
    # 4. Graphics Control Extension (optional)
    gce = None
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0] + 1)
        blocks.append(rng.read(block_size))
        
        total_blocks_size -= block_size
    
    # 7. Trailer
    trailer = b'\x3b'
    
    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)