#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table
    # Number of colors to use
    num_colors = rng.read(1)[0] & 7 + 1
    
    # Size of global color table
    size = 3 * ((num_colors ** 2 - 1).bit_length() // 8) + 1
    
    # Actual global color table
    color_table = b''
    for i in range(size):
        color_table += rng.read(1)
    
    # 4. Graphics Control Extension
    gce = b''
    if rng.read(1)[0]:
        # Transparent color flag set
        gce += b'\x21\xF9\x04'
        
        # Delay time
        delay = rng.read(2)
        while not delay[0] or delay[0] >= 0xFF:
            delay = rng.read(2)
        gce += delay
        
        # Transparent color index
        gce += rng.read(1)
        
    # 5. Image Descriptor
    image_descriptor = struct.pack('>HHBBB', 0, 0, 0b10000000 | num_colors, 0, 0)
    
    # 6. Blocks
    blocks = []
    while True:
        # Block type
        block_type = rng.read(1)[0] % 4
        
        # End of data
        if block_type == 0:
            break
            
        # Image descriptor
        elif block_type == 1:
            pass
        
        # Local color table
        elif block_type == 2:
            pass
        
        # Image data
        else:
            # Block length
            block_len = rng.read(1)[0]
            
            # Block contents
            block = rng.read(block_len)
            
            # Add to list
            blocks.append(block)
            
    # 7. Trailer
    trailer = b';'
    
    # Write everything to output stream
    out.write(lsd)
    out.write(color_table)
    if len(gce) == 6:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)