#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Color table size is always 0
    color_table_size = 0x00

    # Background color index is always 0
    background_color_index = 0x00

    # Pixel aspect ratio is always 0
    pixel_aspect_ratio = 0x00

    logical_screen_descriptor = struct.pack(
            '>HHBBB', width, height, color_table_size, 
            background_color_index, pixel_aspect_ratio
        )

    # 3. Global Color Table
    global_color_table = [
        0x00, 0x00, 0x00, 
        0xFF, 0xFF, 0xFF
    ]

    # 4. Image Descriptor
    image_left = 0x00
    image_top = 0x00
    image_width = width
    image_height = height

    # LZW minimum code size is always 0
    lzw_minimum_code_size = 0x00

    # Image descriptor uses local color table flag to indicate that there's no
    # global color table
    image_descriptor = struct.pack(
            '>HHHHB', image_left, image_top, image_width, image_height, 
            lzw_minimum_code_size | 0b10000000
        )

    # 5. Local Color Table
    local_color_table = []

    # 6. Data Sub-Block
    data_sub_block = bytearray()

    # 7. Trailer
    trailer = b';'

    # Write everything in order
    out.write(header + logical_screen_descriptor)
    if len(global_color_table):
        out.write(bytes(global_color_table))
    out.write(image_descriptor)
    if len(local_color_table):
        out.write(bytes(local_color_table))
    while True:
        # Generate a random byte
        c = ord(rng.read(1))

        # If it's not an end of sub-block marker, just append it to the 
        # current data sub-block
        if c!= 0x00:
            data_sub_block += bytes([c])
        
        # Otherwise, write the current data sub-block to the output stream
        else:
            out.write(struct.pack('B', len(data_sub_block)))
            out.write(data_sub_block)

            # Reset the data sub-block
            del data_sub_block[:]

            # Check if we're done generating frames
            if len(trailer) == 0:
                break
            
            # Otherwise, write the next frame's header
            out.write(image_descriptor)

    # Write the remaining data sub-blocks
    while len(data_sub_block):
        out.write(struct.pack('B', len(data_sub_block)))
        out.write(data_sub_block)
        del data_sub_block[:]

    # Write the trailer
    out.write(trailer)

