#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('B', rng.read(1)[0] & 0x7 | ((width >> 8) & 0xF8) | ((height >> 8) & 0x7)) + \
          struct.pack('B', rng.read(1)[0]) + \
          struct.pack('B', rng.read(1)[0] & 0x7 | ((width << 5) & 0xE0) | ((height << 5) & 0x1C)) + \
          struct.pack('B', rng.read(1)[0])

    # 3. Global Color Table
    global_color_table_size = rng.read(1)[0] & 0x7
    if global_color_table_size == 0:
        global_color_table_size = 2**((rng.read(1)[0] & 0x3) + 1)
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(global_color_table_size)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('BBHBB',
                                   *(rng.read(1)[0] for i in range(6)))
    
    # 5. Local Color Table
    local_color_table_flag = bool(rng.read(1)[0] & 0x80)
    if local_color_table_flag:
        local_color_table_size = rng.read(1)[0] & 0x7
        if local_color_table_size == 0:
            local_color_table_size = 2**((rng.read(1)[0] & 0x3) + 1)
        local_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(local_color_table_size)]

    # 6. Image Data
    # Frame count is always one
    frame_count = 1

    # Loop forever or until we hit an end of file
    while True:
        block_type = rng.read(1)[0]

        # Extension Block
        if block_type == 0x21:
            extension_block_label = rng.read(1)[0]

            # Application Extension
            if extension_block_label == 0xFF:
                application_identifier = rng.read(1)[0:8].decode()

                # Netscape Looping Application Extension
                if application_identifier == 'NETSCAPE':
                    application_authentication_code = rng.read(1)[0:3].decode()

                    # Loop forever or until we hit an end of file
                    while True:
                        sub_block_label = rng.read(1)[0]

                        # End Sub-Block
                        if sub_block_label == 0x00:
                            break
                        
                        # Invalid Sub-Block Label
                        else:
                            raise Exception("Invalid sub-block label")
                    
                    loop_count = struct.unpack('>H', rng.read(2))[0]
                
                # Invalid Application Identifier
                else:
                    raise Exception("Invalid application identifier")
            
            # Graphic Control Extension
            elif extension_block_label == 0xF9:
                # Block Size
                assert rng.read(1)[0] == 0x04

                # Flags
                flags = rng.read(1)[0]

                # Delay Time
                delay_time = struct.unpack('>H', rng.read(2))[0]

                # Transparent Color Index
                transparent_color_index = rng.read(1)[0]

                # Terminator
                assert rng.read(1)[0] == 0x00
            
            # Comment Extension
            elif extension_block_label == 0xFE:
                comment_length = struct.unpack('>H', rng.read(2))[0]
                comment = rng.read(comment_length).decode()
            
            # Plain Text Extension
            elif extension_block_label == 0x01:
                text_left = struct.unpack('>H', rng.read(2))[0]
                text_top = struct.unpack('>H', rng.read(2))[0]
                text_width = struct.unpack('>H', rng.read(2))[0]
                text_height = struct.unpack('>H', rng.read(2))[0]
                cell_width = rng.read(1)[0]
                cell_height = rng.read(1)[0]
                foreground_color_index = rng.read(1)[0]
                background_color_index = rng.read(1)[0]
                character_cell_data = []
                for row in range(text_height):
                    character