#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Image Descriptor
    image_left_pos = rng.read(2)
    image_top_pos = rng.read(2)
    image_width = rng.read(2)
    image_height = rng.read(2)

    # Other ID fields: Packed fields (using a random byte for variety), Local color table flag, Interlace flag, Sort flag, Reserved field
    packed_fields = rng.read(1)[0]
    local_color_table_flag = bool((packed_fields >> 7) & 1)
    interlace_flag = bool((packed_fields >> 6) & 1)
    sort_flag = bool((packed_fields >> 5) & 1)
    reserved_field = rng.read(1)[0]
    id = struct.pack('>HHHHBB', image_left_pos, image_top_pos, image_width, image_height, packed_fields, reserved_field)

    # 5. Local Color Table (if applicable)
    if local_color_table_flag:
        num_colors = (rng.read(1)[0] % 256) or 1
        color_table += rng.read(3 * num_colors)  # Random colors

    # 6. Block of Extension Blocks (we'll just put one in there with no data)
    ext_block = b'\x21\xFF\x0B\x4E\x45\x54\x53\x43\x41\x50\x45\x32\x2E\x30\x03\x01\x00\x00\x00' + \
                struct.pack('<H', len(ext_block)) + \
                b'\x00' * ((len(ext_block) - 10) // 8)

    # 7. Image Data
    # We're going to use the same algorithm as the original gif-generator project did:
    # https://github.com/phoboslab/jsmpeg-vnc/blob/master/source/input/gif.c#L111
    # It uses an LZW compression scheme on top of the raw RGB values from the previous frame.
    # The first frame is generated by filling all pixels with black.
    # Each subsequent frame has its own palette, which is randomly generated.
    # This means that we can't really predict what the output will be, but we can still test our decoder!
    # Also note that this code does not handle the case where the color table changes between frames.
    # In practice, this should never happen anyway because most decoders don't support changing palettes midway through decoding.
    # If you want to add support for this, here are some things you might need to change:
    #   - You would probably need to keep track of whether the current frame had a different palette than the last frame.
    #     This is important so that when we get to the end of the stream, we know how many times to repeat the final frame.
    #   - When reading the next frame header, you would need to check if the new palette differs from the old palette.
    #     If they do, then you would need to reset the LZW dictionary and the LZW buffer.
    #     This is necessary since the new palette contains a different set of colors than the old palette.
    #     Note that this doesn't apply to the first frame, since the first frame always uses the global color table.
    #   - When writing the next frame header, you would need to write the correct number of bits per pixel based on the new palette.
    #     This is necessary since the new palette may contain more colors than the old palette.
    #     Note that this doesn't apply to the first frame, since the first frame always uses the global color table.
    #   - When writing the next frame palette, you would need to write the correct number of colors based on the new palette.
    #     This is necessary since the new palette may contain more colors than the old palette.
    #     Note that this doesn't apply to the first frame, since the first frame always uses the global color table.
    #   - When writing the next frame data, you would need to write the correct number of bits per pixel based on the new palette.
    #     This is necessary since the new palette may contain more colors than the old palette.
    #     Note that this doesn't apply to the first frame, since the first frame always uses the global color table.
    #   - When writing the next frame data, you would need to write the correct number of colors based on the new palette.
    #     This is necessary since the new palette may contain