#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    has_gce = bool((rng.read(1)[0] & 1 << 7) >> 7)
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor (always present)
    image_descriptor = struct.pack('>BBHBBBB', 0x2C, rng.read(1)[0], width, height, rng.read(1)[0], rng.read(1)[0])
    
    # 6. Local Color Table (optional; let's make it have random size, capped at 256 colors, and fully random)
    has_lct = bool((rng.read(1)[0] & 1 << 7) >> 7)
    if has_lct:
        num_colors = (rng.read(1)[0] % 256) or 1
        local_color_table = rng.read(3 * num_colors)  # Random colors
    else:
        local_color_table = b''
        
    # 7. Image data (compressed with RLE)
    # Let's make it be as long as the logical screen descriptor says
    image_data = rng.read(width * height)
    
    # 8. Trailer
    trailer = b';'
    
    # Write everything to output stream
    out.write(header + lsd + color_table + gce + image_descriptor + local_color_table + image_data + trailer)

