#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = None
    if rng.read(1)[0]:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>HHBB', left, top, width, height)

    # 6. Local Color Table (optional; let's use our global one again)
    local_color_table = None
    if rng.read(1)[0]:
        local_color_table = color_table
        
    # 7. Image Data
    data = []
    while True:
        # Let's just put some random pixels in here
        data += [rng.read(1)[0]]
        
        # If we're done with this frame, break
        if len(data) == width * height:
            break
            
    # 8. Trailer
    trailer = b';'
    
    # Write everything to the output stream
    out.write(header + lsd + color_table + (gce or b'') + image_descriptor + (local_color_table or b'') + bytes(data) + trailer)

