#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_index = rng.read(1)[0] & 0b00001111
    pixel_aspect_ratio = rng.read(1)[0] & 0b00001111

    lsd = struct.pack(
        '>HHBBBH',
        width, height, packed_fields, background_index, pixel_aspect_ratio,
        0xF7  # Global Color Table Flag set to 1
    )

    # 3. Global Color Table
    global_color_table_size = ((packed_fields >> 7) & 1) * 2 ** ((packed_fields & 0b01111100) // 3)
    global_color_table = [struct.pack('>BBBB', *(rng.read(4))) for _ in range(global_color_table_size)]

    # 4. Local Color Table
    local_color_table_flag = bool((packed_fields >> 7) & 1)
    if local_color_table_flag:
        local_color_table_size = 2 ** (((packed_fields & 0b01111100) // 3) + 1)
        local_color_table = [struct.pack('>BBBB', *(rng.read(4))) for _ in range(local_color_table_size)]
    else:
        local_color_table = []

    # 5. Graphics Control Extension
    gce = struct.pack(
        '>BBHB',
        0x21, 0xF9, 4, 0b0000000000000000
    )

    # 6. Image Descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    local_color_table_flag = bool((rng.read(1)[0] >> 7) & 1)
    interlace_flag = bool((rng.read(1)[0] >> 6) & 1)
    sort_flag = bool((rng.read(1)[0] >> 5) & 1)
    reserved_flags = rng.read(1)[0]
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    image_descriptor = struct.pack(
        '>BHHHHBB',
        0x2C, left, top, width, height,
        local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved_flags,
        lzw_minimum_code_size
    )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        if len(blocks) >= total_blocks_size:
            break

        block_size = rng.read(1)[0]
        if not block_size:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)