#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved_flags = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = rng.read(local_color_table_size * 3)
    image_descriptor = struct.pack('>BHHHHBBBBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  local_color_table_size)

    # 6. Image Data (LZW compression)
    # Let's use a random minimum code size between 2 and 12 bits long
    lzw_minimum_code_size = rng.read(1)[0] % 13
    blocks = []
    while True:
        # Generate a random block size between 2 and 255 bytes long
        block_size = rng.read(1)[0] % 255

        # If we've reached the end of the image data, break out of this loop
        if block_size == 0:
            break

        # Read in that amount of bytes from /dev/urandom as our current block
        block = rng.read(block_size)

        # Add the block to our list of blocks
        blocks.append(block)

    # Combine all of our blocks together into one big string
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')