#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a' + bytes([0] * (7))

    # 2. Logical screen descriptor
    width = rng.read(1)[0] % 64 + 1
    height = rng.read(1)[0] % 64 + 1
    flags = rng.read(1)[0] & 0b11111100 | 0b00000011
    color_table_size = rng.read(1)[0] % 256
    bg_color_index = rng.read(1)[0] % color_table_size
    pixel_aspect_ratio = rng.read(1)[0] % 16 + 1

    logical_screen_descriptor = \
        struct.pack('<6sBBBHB',
                    header,
                    width,
                    height,
                    flags,
                    color_table_size,
                    bg_color_index,
                    pixel_aspect_ratio
                )

    # 3. Global color table
    global_color_table = [bytes([rng.read(1)[0],
                                 rng.read(1)[0],
                                 rng.read(1)[0]])
                          for _ in range(color_table_size)]

    # 4. Image descriptor
    left = rng.read(1)[0] % width
    top = rng.read(1)[0] % height
    width = rng.read(1)[0] % (width - left) + 1
    height = rng.read(1)[0] % (height - top) + 1
    flags = rng.read(1)[0] & 0b11110000 | 0b00000001
    local_color_table_flag = rng.read(1)[0] // 2
    interlace_flag = rng.read(1)[0] // 4
    sort_flag = rng.read(1)[0] // 8
    reserved = rng.read(1)[0] % 2
    lct_size = rng.read(1)[0] % 256 if local_color_table_flag else 0
    lct = []
    if lct_size > 0:
        lct = [bytes([rng.read(1)[0],
                      rng.read(1)[0],
                      rng.read(1)[0]])
               for _ in range(lct_size)]

    image_descriptor = \
        struct.pack('<BHHBBBBB',
                     left,
                     top,
                     width,
                     height,
                     flags,
                     lct_size,
                     interlace_flag,
                     sort_flag,
                     reserved
                 )

    # 5. Local color table
    local_color_table = b''
    if len(lct) > 0:
        local_color_table = b'\x2c' + \
                            b'\x00\x00'.join(lct) + \
                            b'\x00'

    # 6. Graphics control extension
    block_size = rng.read(1)[0] % 8
    packed = rng.read(1)[0]
    delay_time = rng.read(2)
    transparent_color_index = rng.read(1)[0] % (len(global_color_table)
                                                if len(local_color_table) == 0
                                                else len(local_color_table))
    terminator = rng.read(1)[0] % 2
    graphics_control_extension = \
        struct.pack('BBBBBBB',
                     block_size,
                     packed,
                     delay_time[0],
                     delay_time[1],
                     transparent_color_index,
                     terminator
                 )

    # 7. Plain text extension
    block_size = rng.read(1)[0] % 8
    text_left = rng.read(1)[0] % width
    text_top = rng.read(1)[0] % height
    text_width = rng.read(1)[0] % (width - text_left) + 1
    text_height = rng.read(1)[0] % (height - text_top) + 1
    cell_width = rng.read(1)[0] % 256
    cell_height = rng.read(1)[0] % 256
    foreground_index = rng.read(1)[0] % (len(global_color_table)
                                         if len(local_color_table) == 0
                                         else len(local_color_table))
    background_index = rng.read(1)[0] % (len(global_color_table)
                                         if len(