#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    width = rng.read(1)[0] % (2**8 - 16 + 1) + 16
    height = rng.read(1)[0] % (2**8 - 16 + 1) + 16

    flags = bytes([
        0b10000000 | # Global Color Table Flag
        ((width & 0b11111000) >> 3), # Size of global color table
        (((height & 0b11111000)) >> 3) | # Size of local color table
        0b00000111 # Sorting flag
    ])

    bg_color = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    logical_screen_descriptor = header \
        + width.to_bytes(1, 'little') \
        + height.to_bytes(1, 'little') \
        + flags \
        + bg_color.to_bytes(1, 'little') \
        + pixel_aspect_ratio.to_bytes(1, 'little')

    # 3. Global Color Table
    global_color_table = bytearray()
    for i in range(7):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        global_color_table += r.to_bytes(1, 'little') \
            + g.to_bytes(1, 'little') \
            + b.to_bytes(1, 'little')

    # 4. Image Descriptor
    image_separator = b'\x2C'
    left = rng.read(1)[0] % (width // 256)
    top = rng.read(1)[0] % (height // 256)
    flags = bytes([
        0b10000000 | # Local Color Table Flag
        0b00000000 | # Interlace Flag
        0b00000000 | # Sorting Flag
        0b00000000 | # Reserved
        0b00000000   # Size of local color table
    ])

    img_descriptor = image_separator \
        + left.to_bytes(1, 'little') \
        + top.to_bytes(1, 'little') \
        + width.to_bytes(1, 'little') \
        + height.to_bytes(1, 'little') \
        + flags \
        + b'\xFF\xFF\xFF'

    # 5. Local Color Table
    local_color_table = bytearray()
    for i in range(7):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        local_color_table += r.to_bytes(1, 'little') \
            + g.to_bytes(1, 'little') \
            + b.to_bytes(1, 'little')

    # 6. LZW Minimum Code Size
    lzw_min_code_size = rng.read(1)[0] % 256

    # 7. Data Sub-blocks
    data_sub_blocks = bytearray()
    while True:
        block_type = rng.read(1)[0] % 4
        
        if block_type == 0:
            # End of File
            break
        elif block_type == 1:
            # Plain Text Extension Block
            extension_label = b'\x01'

            subblock_size = rng.read(1)[0] % 256
            
            subblock_data = bytearray()
            for _ in range(subblock_size):
                subblock_data += rng.read(1)

            extension_block = extension_label \
                + subblock_size.to_bytes(1, 'little') \
                + subblock_data

            data_sub_blocks += extension_block
        else:
            # Image Data Sub-block
            subblock_size = rng.read(1)[0] % 256

            subblock_data = bytearray()
            for _ in range(subblock_size):
                subblock_data += rng.read(1)

            data_sub_blocks += subblock_data

    # 8. Trailer
    trailer = b'\x3B'

    # Write everything to output file
    out.write(logical_screen_descriptor)
    out.write(global_color_table)
    out.write(img_descriptor)
    out.write(