#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = rng.read(1)[0] % 257 + 1
    global_color_table = b''
    if global_color_table_size > 0:
        global_color_table += struct.pack('B', global_color_table_size - 1)

        # Generate all the RGB triplets in the table
        for _ in range(global_color_table_size):
            r, g, b = [rng.read(1)[0] & 0xFF for _ in range(3)]
            global_color_table += struct.pack('BBB', r, g, b)

    # Write everything to output
    out.write(lsd)
    out.write(global_color_table)

    # 4. Image Descriptor
    # Left position of image is always zero
    left = 0
    top = rng.read(1)[0] & 0xFF
    width &= 0x7FFF
    height &= 0x7FFF
    packed_fields |= ((left & 0xF8) << 7) | ((top & 0xFE) << 1)
    id = struct.pack('>HHBB', left, top, width, height, packed_fields)
    out.write(id)

    # 5. Local Color Table (same as above)
    local_color_table_size = rng.read(1)[0] % 257 + 1
    local_color_table = b''
    if local_color_table_size > 0:
        local_color_table += struct.pack('B', local_color_table_size - 1)

        # Generate all the RGB triplets in the table
        for _ in range(local_color_table_size):
            r, g, b = [rng.read(1)[0] & 0xFF for _ in range(3)]
            local_color_table += struct.pack('BBB', r, g, b)

    # Write everything to output
    out.write(local_color_table)

    # 6. Image Data
    # We'll just write some random data here
    num_bytes = rng.read(1)[0] % 257 + 1
    out.write(num_bytes * b'\x00')

    return len(lsd) + len(global_color_table) + len(id) + len(local_color_table) + num_bytes
