#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    has_gce = bool((rng.read(1)[0] & 1 << 7) >> 7)
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''
    
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
    
    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    
#    print("LZW minimum code size:", lzw_minimum_code_size)
    
    data = []
    while True:
        block_size = min(((rng.read(1)[0] % 255) + 1), 255)
        
        if not block_size:
            break
        
        data += [rng.read(block_size)]
        
    compressed_data = compress_lzw(data, lzw_minimum_code_size)
    
    # 7. Trailer
    trailer = b'\x3B'
    
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    out.write(gce)
    out.write(image_descriptor)
    out.write(compressed_data)
    out.write(trailer)


# Compresses a list of blocks with LZW compression using the given minimum code size
def compress_lzw(blocks: list[bytes], minimum_code_size: int) -> bytes:
    dictionary = {b'': 0}
    current_string = ''
    output = []

    def add_to_dictionary():
        if current_string in dictionary:
            return

        dictionary[current_string] = len(dictionary)
        output.append(struct.pack('>B', dictionary[current_string]))

    for block in blocks:
        for char in block:
            current_string += chr(char)

            add_to_dictionary()
            
            if len(output[-1]) >= minimum_code_size:
                current_string = ''

    if current_string!= '':
        add_to_dictionary()

    output.append(struct.pack('>B', dictionary['']))

    return b''.join(output)


if