#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    packed_flags = rng.read(1)[0]
    global_color_table_flag = bool(packed_flags & 1)
    color_resolution = ((packed_flags >> 1) & 7) + 1
    sorted_color_table_flag = bool((packed_flags >> 4) & 1)
    global_color_table_size = (((packed_flags >> 5) & 7) + 1) * 3
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack(
        '>HHHBBB',
        width, height,
        global_color_table_flag << 7 | color_resolution << 3 | sorted_color_table_flag << 2 | global_color_table_size,
        background_color_index,
        pixel_aspect_ratio
    )

    # 3. Global Color Table
    if global_color_table_flag:
        color_table = bytearray()
        for _ in range(global_color_table_size):
            r, g, b = [rng.read(1)[0] for _ in range(3)]
            color_table.extend((r, g, b))
    else:
        color_table = b''

    # 4. Graphics Control Extension
    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0] & 1)
    interlace_flag = bool(rng.read(1)[0] & 1)
    sort_flag = bool(rng.read(1)[0] & 1)
    reserved_flags = rng.read(1)[0]
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    image_descriptor = struct.pack(
        '>BHHHHBB',
        0x2C, left, top, width, height,
        local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved_flags,
        lzw_minimum_code_size
    )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        if len(blocks) >= total_blocks_size:
            break

        block_size = rng.read(1)[0]
        if not block_size:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)