#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    global_color_table_size = min((rng.read(1)[0] & 0x7F) + 1, 256)
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(global_color_table_size)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>HHBBBBB', 0, 0, width, height, 0b10000000 | rng.read(1)[0], 0)

    # 5. Local Color Table (if we had one)
    local_color_table_size = max(0, len(global_color_table) - 256)
    if local_color_table_size > 0:
        local_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(local_color_table_size)]

        # We need to pad our local color table with empty entries so that its length is a multiple of 3
        while len(local_color_table)!= ((len(local_color_table) // 3) * 3):
            local_color_table += [b'\x00\x00\x00']

    # 6. Image Data
    # Each frame will be made up of a series of subblocks, which are either RLE blocks or normal data blocks
    frames = []
    for i in range(rng.read(1)[0]):
        # First, let's determine how many subblocks there will be this frame
        num_subblocks = rng.read(1)[0] % 100 + 1

        # Now, let's create some subblocks!
        subblocks = []
        for j in range(num_subblocks):
            # Determine whether this block is an RLE block or not
            rle_block = bool(rng.read(1)[0] & 0b1)

            # If it's an RLE block, then read the next byte as the count
            if rle_block:
                count = rng.read(1)[0] % 100 + 1

                # Read the next byte as the actual data
                data = rng.read(1)[0]
            else:
                # Otherwise, just read the next three bytes directly
                data = struct.pack('BBB', *rng.read(3))

            # Add the subblock to our list of subblocks
            subblocks.append([rle_block, count, data])
        
        # Now, let's add all those subblocks to our list of frames
        frames.append(subblocks)
    
    # Finally, let's write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(bytes(global_color_table))
    out.write(image_descriptor)
    out.write(bytes(local_color_table))

    for frame in frames:
        for subblock in frame:
            if subblock[0]:
                out.write(struct.pack('BB', 0b11100000 | subblock[1], subblock[2]))
            else:
                out.write(subblock[2])


if __name__ == '__main__':
    from sys import argv, stdin, stdout
    import os

    if len(argv) < 2:
        print("Usage: {} <output-file>".format(os.path.basename(__file__)), file=stderr)
        exit(-1)

    with open(argv[1], 'wb') as f:
        generate_random_gif(f, stdout.buffer)