#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header (6 bytes)
    out.write(b'GIF89a')

    # Logical screen descriptor (7 bytes)
    width = 2**15 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 2**10
    height = 2**15 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 2**10
    flags = b'\x80\x00' if bool(int.from_bytes(rng.read(1), byteorder='big', signed=False)) else b'\x00\x00'
    bg_color = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x00FFFFFF
    aspect_ratio = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xFF
    out.write(struct.pack('<HHBBB', width, height, flags[0], flags[1], bg_color, aspect_ratio))

    # Global color table (3 bytes)
    global_color_table_size = min((width * height)//256, 256)
    for i in range(global_color_table_size):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xFF
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xFF
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xFF
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor (9 bytes)
    left = 0
    top = 0
    flags = b'\x80\x00' if bool(int.from_bytes(rng.read(1), byteorder='big', signed=False)) else b'\x00\x00'
    local_color_table_flag = b'\x80' if bool(int.from_bytes(rng.read(1), byteorder='big', signed=False)) else b'\x00'
    interlace_flag = b'\x40' if bool(int.from_bytes(rng.read(1), byteorder='big', signed=False)) else b'\x00'
    sort_flag = b'\x20' if bool(int.from_bytes(rng.read(1), byteorder='big', signed=False)) else b'\x00'
    reserved_flags = b'\x0C'
    local_color_table_size = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xFF
    out.write(struct.pack('<HHBBBBBB', left, top, width, height, flags[0], flags[1], local_color_table_flag[0], local_color_table_size))

    # Local color table (3*local_color_table_size bytes)
    for i in range(local_color_table_size):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xFF
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xFF
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xFF
        out.write(struct.pack('BBB', r, g, b))

    # Image data (LZW minimum code size byte + LZW compressed image data)
    lzw_minimum_code_size = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xFF
    out.write(lzw_minimum_code_size.to_bytes(1, 'big'))
    lzw_compressed_image_data = []
    while True:
        chunk_type = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xFF
        if chunk_type == 0x00:
            break

        elif chunk_type <= 0x0F:
            length = chunk_type + 1
            lzw_compressed_image_data += [chunk_type]

        elif chunk_type >= 0x10 and chunk_type <= 0xFE:
            length = ((chunk_type - 0x10) << 8) | int