#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = int.from_bytes(rng.read(2), byteorder='big') % 65536 + 1
    h = int.from_bytes(rng.read(2), byteorder='big') % 65536 + 1

    # Random color table size
    ctsize = int.from_bytes(rng.read(1), byteorder='big') & 0x7 + 1

    # Flags: global color table flag set to true
    flags = 0b10000000 | ((ctsize - 1) << 4)

    # Color resolution is always 7
    flags |= 0b01110000

    # Sorting method is always 0
    flags |= 0b00001100

    # Size of global color table is 2^(n+1) entries
    flags |= (((1 << ctsize) * 3) - 1).to_bytes(1, 'big')

    # Background color index is always 0
    bgindex = 0.to_bytes(1, 'big')

    # Pixel aspect ratio is always 0
    pixelaspectratio = 0.to_bytes(1, 'big')

    lsd = struct.pack('>6sHHBBB',
                      header,
                      0,
                      w,
                      h,
                      flags,
                      bgindex,
                      pixelaspectratio
                     )

    # Write logical screen descriptor
    out.write(lsd)

    # 3. Global Color Table
    colors = []
    for _ in range(1 << ctsize):
        r = int.from_bytes(rng.read(1), byteorder='big')
        g = int.from_bytes(rng.read(1), byteorder='big')
        b = int.from_bytes(rng.read(1), byteorder='big')

        colors.append((r,g,b))

    # Write global color table
    for c in colors:
        out.write(struct.pack('BBB', *c))

    # 4. Image Descriptor
    # Left position is always 0
    left = 0.to_bytes(2, 'big')

    # Top position is always 0
    top = 0.to_bytes(2, 'big')

    # Width and height are randomly generated
    w = int.from_bytes(rng.read(2), byteorder='big') % 65536 + 1
    h = int.from_bytes(rng.read(2), byteorder='big') % 65536 + 1

    # Flags: local color table flag set to false
    flags = 0b00000000

    # LZW minimum code size is 2
    lzwmincodesize = 2.to_bytes(1, 'big')

    id = struct.pack('>H2s2sHBHB',
                     0,
                     left,
                     top,
                     w,
                     h,
                     flags,
                     lzwmincodesize
                    )

    # Write image descriptor
    out.write(id)

    # 5. Local Color Table
    if not bool(flags & 0b10000000):
        for _ in range(1 << ctsize):
            r = int.from_bytes(rng.read(1), byteorder='big')
            g = int.from_bytes(rng.read(1), byteorder='big')
            b = int.from_bytes(rng.read(1), byteorder='big')

            colors.append((r,g,b))

        # Write local color table
        for c in colors:
            out.write(struct.pack('BBB', *c))

    # 6. Graphics Control Extension
    # Block size is always 4
    blocksize = 4.to_bytes(1, 'big')

    # Flags: transparent color flag set to false
    flags = 0b00000000

    # Delay time is always 0
    delaytime = 0.to_bytes(2, 'big')

    # Transparent color index is always 0
    transindex = 0.to_bytes(1, 'big')

    # Terminator
    terminator = 0.to_bytes(1, 'big')

    gce = struct.pack('>BB2sB1s',
                      blocksize,
                      flags,
                      delaytime,
                      transindex,
                      terminator
                     )

    # Write graphics control extension
    out.write(gce)

    # 7. Image Data
    # LZW minimum code size is 2
    lzwmincodesize = 2.to_bytes(1, 'big')

    # Write LZW