#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>BHHBBB',
        0b10000000 | ((width >> 8) & 0x07) << 5 | (height >> 8) & 0x07,
        width,
        height,
        rng.read(1)[0],
        rng.read(1)[0] % 4,
        0
    )

    # 3. Global Color Table
    global_color_table_size = 2 ** int((lsd[6] & 0b111) + 1)
    global_color_table = [struct.pack('>BBBB', *rng.read(4)) for _ in range(global_color_table_size)]

    # 4. Image Descriptor
    image_descriptor = struct.pack('>BBHBBBBBHB',
        0x2C,
        rng.read(1)[0] % 4,
        0,
        0,
        width,
        height,
        0,
        0,
        global_color_table_size - 1,
        0
    )

    # 5. Local Color Table
    local_color_table_size = 2 ** int(((image_descriptor[9] & 0b11100000) >> 5) + 1)
    if local_color_table_size > 0:
        local_color_table = [struct.pack('>BBBB', *rng.read(4)) for _ in range(local_color_table_size)]

        # TODO: Use this to make sure we don't use more than one block of data
        #       per pixel when encoding it later on
        # local_color_table_data = []
        # for i in range(len(local_color_table)):
        #     local_color_table_data += list(local_color_table[i])

    else:
        local_color_table = []


    # 6. Image Data
    # We want to have at least one frame with an empty palette so that we can test
    # our decoder's ability to handle such cases correctly
    frames = [(b'', [])]
    while len(frames) < 2 or sum([len(frame[1]) for frame in frames]):
        # Frame duration is between 1 and 10 inclusive
        frame_duration = rng.read(1)[0] % 10 + 1

        # Number of rows in current frame
        num_rows = rng.read(1)[0] % min(height // 2, 10) + 1

        # Each row consists of a single byte indicating how many pixels are encoded
        # in the following block. The maximum value here is 255 which means there
        # will be no more blocks after this one.
        row_lengths = [rng.read(1)[0] % max(num_rows, 1) + 1 for _ in range(num_rows)]

        # Now we need to encode all those pixels!
        # First, let's get some random RGB values from the global color table
        rgb_values = [
            global_color_table[rng.read(1)[0] % global_color_table_size]
                for _ in range(sum(row_lengths))]

        # Then, let's group them by their length
        grouped_rgb_values = {}
        for i in range(len(rgb_values)):
            if not grouped_rgb_values.get(len(rgb_values[i])):
                grouped_rgb_values[len(rgb_values[i])] = []

            grouped_rgb_values[len(rgb_values[i])].append(rgb_values[i])

        # Finally, let's create a new frame based on these groups
        frame = ([], [])
        for key in sorted(grouped_rgb_values):
            frame[0] += grouped_rgb_values[key][0]
            frame[1] += grouped_rgb_values[key][:min(key, len(grouped_rgb_values[key]))]

        frames.append(frame)

    # 7. Trailer
    trailer = b';'

    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    for entry in global_color_table:
        out.write(entry)
    out.write(image_descriptor)
    for entry in local_color_table:
        out.write(entry)
    for frame in frames:
        for row_length in row_lengths:
            out.write(bytes([row_length]))

        for rgb_value in frame[0]:
            out.write(rgb_value)

    out.write(trailer)



if