#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    lsd = struct.pack('>HHBBB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table
    global_color_table_size = 2 ** ((packed_fields & 7) + 1)
    global_color_table = [rng.read(3) for _ in range(global_color_table_size)]
    gct = b''.join([b'\x21\xF9\x04\x05'+bytes((len(global_color_table)-1).to_bytes(1,'big'))+b'\x00'] +
                  [struct.pack('BBB', *rgb)+b'\x00' for rgb in global_color_table])

    # 4. Image Descriptor
    image_left_position = 0
    image_top_position = 0
    image_width = width
    image_height = height
    local_color_table_flag = bool(packed_fields & 0b10000000)
    interlace_flag = False
    sort_flag = True
    reserved_bits = 0
    local_color_table_size = 2 ** (((packed_fields >> 4) & 7) + 1) if local_color_table_flag else 0
    local_color_table = [rng.read(3) for _ in range(local_color_table_size)]
    lct = b'' if not local_color_table_flag else \
         b'\x2C\x00\x00\x00'+bytes((local_color_table_size-1).to_bytes(1,'big'))+b'\x00'+b''.join([struct.pack('BBB',*rgb)+b'\x00' for rgb in local_color_table])

    # 5. Image Data
    image_data = []
    while len(image_data) < image_width * image_height:
        block_size = min(rng.read(1)[0], image_width * image_height - len(image_data))

        # Plain text extension
        if rng.read(1)[0] == 0x01:
            label = rng.read(16)

            # Graphic Control Extension
            if label == b'NETSCAPE2.0':
                block_size -= 17

                # Block size must be at least 17
                assert block_size >= 17

                subblock_id = rng.read(1)[0]
                assert subblock_id == 1

                flags = rng.read(1)[0]
                delay_time = int.from_bytes(rng.read(2), 'little')
                transparent_color_index = rng.read(1)[0]
                terminator = rng.read(1)[0]

                # Flags must be zero
                assert flags == 0

                # Delay time must be nonnegative
                assert delay_time >= 0

                # Transparent color index must be valid
                assert transparent_color_index <= max(global_color_table_size, local_color_table_size)

                # Terminator must be 0x00
                assert terminator == 0

                block_size -= 17

                # Subblocks are ignored
                while block_size > 0:
                    block_size -= rng.read(1)[0]+1

                continue

            # Comment extension
            elif label == b'Comment':
                block_size -= 17

                # Block size must be at least 17
                assert block_size >= 17

                subblock_id = rng.read(1)[0]
                assert subblock_id == 1

                comment = b''
                while block_size > 0:
                    block_size -= rng.read(1)[0]+1

                    comment += rng.read(block_size)

                continue

            # Plain text extension
            else:
                print("Warning: unknown plain text extension")

        # Application extension
        elif rng.read(1)[0] == 0xFF:
            application_identifier = rng.read(8)

            # Netscape Looping Application Extension
            if application_identifier == b'NETSCAPE2.0':
                block_size -= 17

                # Block size must be at least 17
                assert block_size >= 17

                subblock_id = rng.read(1)[0]
                assert subblock_id == 1

                loop_count = int.from_bytes(rng.read(2), 'little')
                terminator = rng.read(1)[0