#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# provided RNG and writes it to disk
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # Header
    out.write('GIF89a'.encode())

    # Logical Screen Descriptor
    width = 256 + int.from_bytes(rng.read(1), byteorder='big') % 70
    height = 256 + int.from_bytes(rng.read(1), byteorder='big') % 40
    flags = b'\x87\x00' if (width >= 256 or height >= 256) else b'\x07\x00'
    bgColorIndex = int.from_bytes(rng.read(1), byteorder='big') & 0xf
    pixelAspectRatio = int.from_bytes(rng.read(1), byteorder='big') & 0xff
    out.write(struct.pack('<HHBBB', width, height, flags[0], flags[1],
                          bgColorIndex, pixelAspectRatio))

    # Global Color Table
    globalColorTableSize = 2 << ((int.from_bytes(rng.read(1), byteorder='big') >> 3) + 1)
    for i in range(globalColorTableSize):
        r = int.from_bytes(rng.read(1), byteorder='big') & 0xff
        g = int.from_bytes(rng.read(1), byteorder='big') & 0xff
        b = int.from_bytes(rng.read(1), byteorder='big') & 0xff
        out.write(struct.pack('<BBBB', r, g, b, 0))

    # Loop Control Block
    loopCount = -1
    if int.from_bytes(rng.read(1), byteorder='big') & 1 == 1:
        loopCount = int.from_bytes(rng.read(2), byteorder='big')
    out.write(b'\x00')
    out.write(struct.pack('<H', loopCount))

    # Image Descriptor
    leftPos = 0
    topPos = 0
    width = 256 + int.from_bytes(rng.read(1), byteorder='big') % 70
    height = 256 + int.from_bytes(rng.read(1), byteorder='big') % 40
    flags = b'\x87\x00' if (width >= 256 or height >= 256) else b'\x07\x00'
    frameIndex = int.from_bytes(rng.read(1), byteorder='big') & 0xff
    localColorTableFlag = bool(int.from_bytes(rng.read(1), byteorder='big') & 1)
    interlaceFlag = bool(int.from_bytes(rng.read(1), byteorder='big') & 2)
    sortFlag = bool(int.from_bytes(rng.read(1), byteorder='big') & 4)
    reserved = int.from_bytes(rng.read(1), byteorder='big') & 0x7f
    out.write(struct.pack('<HHHHBB', leftPos, topPos, width, height,
                          flags[0], flags[1]))
    out.write(struct.pack('<B', frameIndex))
    out.write(flags)

    # Local Color Table
    if localColorTableFlag:
        localColorTableSize = 2 << ((int.from_bytes(rng.read(1), byteorder='big') >> 3) + 1)
        for i in range(localColorTableSize):
            r = int.from_bytes(rng.read(1), byteorder='big') & 0xff
            g = int.from_bytes(rng.read(1), byteorder='big') & 0xff
            b = int.from_bytes(rng.read(1), byteorder='big') & 0xff
            out.write(struct.pack('<BBBB', r, g, b, 0))

    # Image Data
    lzwMinCodeSize = int.from_bytes(rng.read(1), byteorder='big') & 0xff
    out.write(struct.pack('<B', lzwMinCodeSize))

    # LZW code stream
    while True:
        code = int.from_bytes(rng.read(1), byteorder='big') & 0xff

        if code < 0x80:
            # Literal
            out.write(struct.pack('<B', code))
        elif code < 0xe0    with open('/dev/urandom', 'rb') as rng:
        with open('./test.gif', 'wb') as f:
            generate_random_gif(rng, f)