#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    #    - Signature (4 bytes): "GIF89a"
    #    - Version (3 bytes): "\x00\x2c\x00"
    header = b'GIF89a\x00,\x00'

    # 2. Logical Screen Descriptor
    #    - Width (2 bytes): 640 pixels
    #    - Height (2 bytes): 480 pixels
    #    - Flags (1 byte): Global Color Table Flag set to 1
    #    - Background color index (1 byte): 0
    #    - Pixel aspect ratio (1 byte): 0
    lsd = struct.pack('<HHBBB',
        640, 480, 0b10000000, 0, 0
    )

    # 3. Global Color Table
    #    - Size of global color table (1 byte): 7
    #    - Color table entries (3 bytes * 7):
    #        - Black (R=0, G=0, B=0)
    #        - White (R=255, G=255, B=255)
    #        - Red (R=255, G=0, B=0)
    #        - Green (R=0, G=255, B=0)
    #        - Blue (R=0, G=0, B=255)
    #        - Yellow (R=255, G=255, B=0)
    #        - Cyan (R=0, G=255, B=255)
    gct = struct.pack('B3B3B3B3B3B3B3',
        7, 0, 0, 255, 255, 255, 255, 0, 0, 0, 255, 0, 255, 0, 0, 255
    )

    # 4. Image Descriptor
    #    - Left position (2 bytes): 0
    #    - Top position (2 bytes): 0
    #    - Width (2 bytes): 640 pixels
    #    - Height (2 bytes): 480 pixels
    #    - Flags (1 byte): Local Color Table Flag set to 1
    #    - Interlace flag (1 bit): 0
    #    - Sort flag (1 bit): 0
    #    - Reserved (1 bit): 0
    #    - LZW minimum code size (1 byte): 2
    id = struct.pack('<HHHHBB',
        0, 0, 640, 480, 0b10000000, 0
    )

    # 5. Image Data
    #    - LZW minimum code size (1 byte): 2
    #    - Subblocks
    #       - Subblock length (1 byte): 128
    #       - Raw data (128 bytes):
    #           - Repeat last pixel 255 times
    #           - Repeat last pixel 255 times
    #           -...
    #           - Repeat last pixel 255 times
    lzwmincodesize = 2
    subblocks = [struct.pack('B', lzwmincodesize)] + \
                [bytes([lzwmincodesize]) for _ in range(128)]

    # 6. Trailer
    trailer = b'\x3b'

    # Write everything to `out`
    out.write(header)
    out.write(lsd)
    out.write(gct)
    out.write(id)
    for sb in subblocks:
        out.write(sb)
    out.write(trailer)

