#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    w = int.from_bytes(rng.read(2), 'big') % 65535 + 1
    h = int.from_bytes(rng.read(2), 'big') % 65535 + 1

    # Color table size is fixed to 7 bits per pixel
    color_table_size = 0b111 << 4

    # Global color table flag set to false
    global_color_table_flag = 0x00

    # Background color index is always zero
    background_color_index = 0x00

    # Pixel aspect ratio is always 0
    pixel_aspect_ratio = 0x00

    # Pack everything together in one byte
    lsd = struct.pack('>HHBBB', w, h,
                      color_table_size | global_color_table_flag,
                      background_color_index, pixel_aspect_ratio)

    # Write it all to the output stream
    out.write(header)
    out.write(lsd)

    # 3. Image Descriptor
    image_separator = 0x2C

    # Left position of the image is always zero
    left = 0x00

    # Top position of the image is always zero
    top = 0x00

    # Width of the image is randomly chosen from [1, 65535]
    width = int.from_bytes(rng.read(2), 'big') % 65535 + 1

    # Height of the image is randomly chosen from [1, 65535]
    height = int.from_bytes(rng.read(2), 'big') % 65535 + 1

    # Packed field that contains various flags
    packed_field = 0b00000000

    # LZW minimum code size is always 2
    lzw_minimum_code_size = 0b00000010

    # Write the image descriptor to the output stream
    out.write(struct.pack('>BHHHHBB',
                          image_separator, left, top, width, height,
                          packed_field, lzw_minimum_code_size))

    # 4. Local Color Table
    # Generate a list of RGB colors with values between 0-255
    rgb_colors = []
    while len(rgb_colors) < 2**(color_table_size & 0b111):
        r = int.from_bytes(rng.read(1), 'big') % 256
        g = int.from_bytes(rng.read(1), 'big') % 256
        b = int.from_bytes(rng.read(1), 'big') % 256

        rgb_colors += [(r, g, b)]

    # Write them to the output stream as an array of bytes
    for c in rgb_colors:
        out.write(struct.pack('BBB', *c))

    # 5. Image Data
    # Number of pixels in the image is equal to its width multiplied by its height
    num_pixels = width * height

    # The first byte of the compressed data block must be the initial code length
    # which is randomly chosen from {2, 3}
    compression_method = 0b00

    # The second byte of the compressed data block must contain the minimum code size
    # which is always 2 because we are using LZW compression method
    min_code_size = 0b00000010

    # Write these two bytes to the output stream
    out.write(struct.pack('>BB', compression_method, min_code_size))

    # Now comes the fun part! We will use LZW algorithm to compress our image data
    # Firstly, we need to create a dictionary of codes
    codes = {}
    for i in range(2**min_code_size):
        codes[i] = chr(i).encode()

    # Then we start reading the input data until we reach EOF or we have read enough pixels
    while True:
        # Read a single byte from the input stream
        b = rng.read(1)

        if not b:
            break

        # If this byte was already encountered before then just write its corresponding code
        try:
            out.write(codes[int.from_bytes(b, 'big')])

            continue
        except KeyError:
            pass

        # Otherwise add it to the dictionary alongside with a new code
        codes[len(codes)] = codes[-1] + b

        # And finally write the newly generated code to the output stream
        out.write(codes[-1])

    # 6. Trailer
    trailer = b