#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = [0x7F & int(b) for b in rng.read(6)] + [0]

    # Write to output stream
    out.write(header)
    out.write(struct.pack('>HHBBBH', *lsd))

    # 3. Global Color Table
    global_color_table_size = 2 ** ((lsd[5] >> 7) + 1)
    global_color_table = []

    if global_color_table_size > 0:
        for i in range(global_color_table_size):
            rgb = [int(b) for b in rng.read(3)]

            # Set the alpha channel to fully opaque
            rgb += [0xFF]

            global_color_table += rgb

        # Write to output stream
        out.write(bytes(global_color_table))

    # 4. Image Descriptor
    image_descriptor = [
        0, 0,  # Left position, top position
        width - 1, height - 1,   # Width, Height
        False, True,    # Local color table flag, Interlace flag
        0,              # Sorting method
        0               # Reserved
    ]

    # Write to output stream
    out.write(struct.pack('>HHHHBB', *image_descriptor))

    # 5. Local Color Table
    local_color_table_size = 2 ** (lsd[5] & 0x7)
    local_color_table = []

    if local_color_table_size > 0:
        for i in range(local_color_table_size):
            rgb = [int(b) for b in rng.read(3)]

            # Set the alpha channel to fully opaque
            rgb += [0xFF]

            local_color_table += rgb

        # Write to output stream
        out.write(bytes(local_color_table))

    # 6. Image Data
    # Each pixel is represented by one byte
    data = [int(b) for b in rng.read(width*height)]

    # Write to output stream
    out.write(bytes(data))