#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Image Descriptor
    image_left = 0
    image_top = 0
    image_width = width
    image_height = height
    packed_fields = 0x04  # Local color table flag set to true
    id = struct.pack('>HHHHB', image_left, image_top, image_width, image_height, packed_fields)

    # 5. Image Data
    data = b''
    while len(data) < width * height:
        subblock_size = rng.read(1)[0]

        if subblock_size == 0:
            break
        
        elif subblock_size >= 0x80:
            # Extension block
            ext_type = rng.read(1)[0]

            if ext_type == 0xF9:
                # Graphic Control Extension Block
                # We'll just use the defaults here
                label = b'\x04'

                # Delay time in hundredths of seconds
                delay_time = rng.read(2) + b'\x00\x00'

                # Transparent color index
                transparent_color_index = rng.read(1)

                # Terminator
                terminator = b'\x00'

                gce = label + delay_time + transparent_color_index + terminator
                data += gce
            
            else:
                # Unknown extension type; skip this entire block
                data += rng.read(subblock_size)

        else:
            # Plain old image data
            data += rng.read(subblock_size)
    
    # 6. Trailer
    trailer = b';'

    return len(lsd) + len(color_table) + len(id) + len(data) + len(trailer)