#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = None

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved_flags = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = rng.read(local_color_table_size * 3)
    image_descriptor = struct.pack('>BHHHHBBBBB',
                                  0x2C,
                                  left,
                                  top,
                                  width,
                                  height,
                                  local_color_table_flag,
                                  interlace_flag,
                                  sort_flag,
                                  reserved_flags,
                                  local_color_table_size)

    # 6. Image Data
    # Let's use LZW compression to compress our data
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x00:
            break
        
        elif block_type == 0x01:
            # Plain text extension
            label = rng.read(1)[0]

            # We'll just assume that this is an ASCII string
            length = rng.read(1)[0]
            plain_text = rng.read(length).decode()
            
            print(f'Plain text with label {label}: "{plain_text}"')
            
        elif block_type in [0x02, 0x03]:
            # Image data sub-blocks
            length = rng.read(1)[0]
            compressed_data = rng.read(length)
            decompressed_data = decompress_lzw(compressed_data, lzw_minimum_code_size)

            blocks.append((block_type, decompressed_data))
            
    # Now let's write all of these things to the output file
    out.write(lsd + (gce or b'') + image_descriptor)
    
    for block_type, decompressed_data in blocks:
        out.write(struct.pack('B', block_type))
        out.write(decompressed_data)
        
    return len(lsd) + ((len(gce) if gce else 0) +
                       len(image_descriptor) + sum([len(d) for _, d in blocks]))


# Decompresses the given LZW-compressed data using the minimum code size
def decompress_lzw(compressed_data: bytes, min_code_size: int):
    codes = list(range(min_code_size + 1))
    next_code = min_code_size + 1
    clear_code = min_code_size + 1
    end_of_information_code = min_code_size + 2
    current_code = -1
    result = []

    def add_to_result(char):
        nonlocal result
        result.append(bytes([char]))

    for i in range(len(compressed_data)):
        char = compressed_data[i]

        if current_code < 0:
            current_code = char
            continue

        if current_code == end_of_information_code:
            break

        if current_code == clear_code:
            codes = list(range(next_code))
            next_code *= 2
            clear_code += 1 << min_code_size
            end_of_information_code += 1 << min_code_size
            current_code = char
            continue

        try:
            code = codes[current_code] + char
        except IndexError:
            code = codes[-1] + char

        if code >= next_code:
            add_to_result(codes[-1])
            codes.append(codes[-1])
            next_code += 1

        codes.append(code)
        current_code = code

    add_to_result(codes[current_code])

    return b''.join(result)

