#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0] & 0b1111
    pixel_aspect_ratio = rng.read(1)[0] & 0b111

    lsd = struct.pack(
        '>HHBBBH',
        width, height, packed_fields, background_color_index, pixel_aspect_ratio,
        0xF7  # Global Color Table Flag set
    )

    # 3. Global Color Table
    global_color_table_entries = min((width * height // 256) + 1, 256)
    global_color_table = [struct.unpack('>BBBB', rng.read(4)) for _ in range(global_color_table_entries)]
    color_table = b''.join(bytes(entry) for entry in global_color_table)

    # 4. Graphics Control Extension
    # Frame delay is randomly chosen between 1 and 10 seconds
    frame_delay = rng.read(1)[0] % 10 + 1
    transparent_color_index = rng.read(1)[0] & 0b1111
    disposal_method = rng.read(1)[0] & 0b111
    user_input_flag = bool(rng.read(1)[0] & 0b1)
    transparency_flag = bool(rng.read(1)[0] & 0b1)
    gce = struct.pack(
        '>BBHB',
        0x21, 0xF9, 4,
        frame_delay << 8 | transparent_color_index << 4 | disposal_method << 2 | int(user_input_flag) << 1 | int(transparency_flag)
    )

    # 5. Image Descriptor
    left = rng.read(1)[0] % width
    top = rng.read(1)[0] % height
    local_color_table_flag = bool(rng.read(1)[0] & 0b1)
    interlace_flag = bool(rng.read(1)[0] & 0b1)
    sort_flag = bool(rng.read(1)[0] & 0b1)
    reserved_flags = rng.read(1)[0] & 0b111
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1
    image_descriptor = struct.pack(
        '>BHHHHBB',
        0x2C, left, top, width, height,
        local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved_flags,
        lzw_minimum_code_size
    )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while True:
        if len(blocks) >= total_blocks_size:
            break

        block_size = rng.read(1)[0]
        if not block_size:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)