#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    has_gce = bool(rng.read(1)[0] & 1)
    if has_gce:
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])
    else:
        gce = b''

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0] & 1)
    interlace_flag = bool(rng.read(1)[0] & 1)
    sort_flag = bool(rng.read(1)[0] & 1)
    reserved_flags = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    image_descriptor = struct.pack('>HHBBBBB', left, top, width, height, local_color_table_flag, interlace_flag,
                                   sort_flag, reserved_flags)
    if local_color_table_flag:
        image_descriptor += bytes([local_color_table_size]) + rng.read((local_color_table_size ** 3) // 8)

    # 6. Image Data
    # Let's use LZW compression
    lzw_minimum_code_size = rng.read(1)[0]
    max_code_size = min(max(lzw_minimum_code_size + 1, 2), 12)
    code_size_table = [i for i in range(max_code_size)]
    clear_code = 1 << lzw_minimum_code_size
    end_of_information_code = clear_code + 1
    next_code_to_use = clear_code + 2
    codes = {clear_code} | set(range(next_code_to_use, 1 << max_code_size))
    dictionary = {}
    blocks = []
    while True:
        current_byte = rng.read(1)[0]
        if current_byte == 0xFF:  # RANDOM EOI CODE!
            blocks.append(bytes([end_of_information_code]))
            break

        block_size = rng.read(1)[0]
        if not block_size:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        elif len(data) > 255:
            data = bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)