#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1. Header
    header = b'GIF89a'
    
    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    
    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
    
    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors
    
    # 4. Graphics Control Extension (optional)
    gce = None
    if rng.read(1)[0]:
        # Transparent color flag set to true
        transparent_color_index = rng.read(1)[0]
        
        # Delay time is always 10ms
        delay_time = 10
        
        # Transparent color index field is only present when the transparent color flag is set to true
        gce = struct.pack('>BBHBB', 0x21, 0xF9, 4, 0b10000000 | transparent_color_index, delay_time)
    
    # 5. Image Descriptor
    image_descriptor = struct.pack('>HHBBB', 0, 0, 0, 0b10000000 | (num_colors == 1 << 8), 0)
    
    # 6. Image Data
    image_data = []

    # We'll use this as our "current position" in the image data
    current_position = 0
    
    while True:
        # Each frame will be between 1-10 blocks long
        block_size = rng.read(1)[0] % 10 + 1

        # If we're at the end of the image data, break out of the loop
        if not block_size:
            break
        
        # Generate some random data
        image_data += [rng.read(block_size)]
        
        # Update the current position by adding the block size
        current_position += block_size
    
    # 7. Trailer
    trailer = b';'
    
    # Write everything to the output stream
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce:
        out.write(gce)
    out.write(image_descriptor)
    out.write(b''.join(image_data))
    out.write(trailer)

