#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    lsd = struct.pack('<6sHHBBBHB',
                       header[:6],
                       0, 0,                        # No global color table
                       width % 256, int(width / 256),   # Width
                       height % 256, int(height / 256), # Height
                       0b10000000 | 7,                # Global Color Table Flag set to true and size of palette to 7 colors
                       0                       # Background color index
                    )

    # 3. Global Color Table
    color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(7)]

    # 4. Graphics Control Extension
    gce = struct.pack('<B3BBBBBH',
                      0x21,                   # Extension Introducer
                      ord('g'), ord('c'), ord(' '),       # Extension Label
                      4,                      # Block Size
                      0b00000000,              # Packed field: no transparancy, user input not expected, disposal method is none
                      0,                      # Delay time (in hundredths of seconds)
                      0,                      # Transparent color index
                      0                     # Terminator
                  )

    # 5. Image Descriptor
    image_descriptor = struct.pack('<BHHHHB',
                                   0x2C,
                                   x=0, y=0,      # Left-Top corner coordinates
                                   w=width, h=height,
                                   flags=0b10000000     # Local Color Table flag set to true
                                )

    # 6. Image Data
    image_data = []
    for i in range(int((width*height)/2)):
        image_data += [struct.pack('BB', *rng.read(2))]

    # 7. Trailer
    trailer = b'\x3B\xFF'

    # Write everything to output
    out.write(header + lsd + color_table + gce + image_descriptor + b''.join(image_data) + trailer)


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('./test.gif', 'wb') as f:
            generate_random_gif(rng, f)