#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # 1. Header
    out.write(b'GIF89a')

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    global_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(num_colors)]
    color_table = b''.join(global_color_table)

    # 4. Graphics Control Extension
    has_gce = bool(rng.read(1)[0])
    if has_gce:
        # Duration (in hundredths of seconds)
        duration = rng.read(2)
        transparent_color_index = rng.read(1)[0]
        terminator = rng.read(1)[0]
        assert terminator == 0
        gce = struct.pack('>HBB', duration[0]*256 + duration[1], transparent_color_index, terminator)
    else:
        gce = None

    # 5. Image Descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    reserved = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    local_color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(local_color_table_size)]
    local_color_table = b''.join(local_color_table)
    image_descriptor = struct.pack('>HHHHBBBB', left[0]*256 + left[1], top[0]*256 + top[1],
                                   width[0]*256 + width[1], height[0]*256 + height[1],
                                   local_color_table_flag, interlace_flag, sort_flag, reserved)
    if local_color_table_flag:
        image_descriptor += struct.pack('B', local_color_table_size)
        image_descriptor += local_color_table

    # 6. Image Data
    # Let's use LZW compression to compress our data
    lzw_minimum_code_size = rng.read(1)[0]
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if block_size == 0:
            break

        blocks.append(bytes([block_size]) + rng.read(block_size - 1))
    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks) + b'\x00'  # Ending with a block size of 0

    # 7. Trailer
    out.write(lsd + color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor + image_data + b'\x3B')