#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    local_color_table_flag = bool(rng.read(1)[0] & 0b10000000)
    interlace_flag = bool(rng.read(1)[0] & 0b01000000)
    sort_flag = bool(rng.read(1)[0] & 0b00100000)
    reserved = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    image_descriptor = struct.pack('>HHBBBBB', left, top, width, height, local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved, local_color_table_size, background_color_index, pixel_aspect_ratio)

    # 6. Image Data
    # Let's use LZW compression to compress our image data
    # We'll start by generating an array of all possible codes
    code_length = [8] * 256
    for i in range(256):
        code_length[i] = min(max((i+1).bit_length(), 2), 12)

    # Then we'll create a dictionary mapping from codes to their corresponding values
    codes = list(range(256))
    for i in range(256, 4096):
        codes.append(codes[i - 256])

    # Now let's construct the actual compressed data
    lzw_minimum_code_size = max(min(rng.read(1)[0], 12), 2)
    blocks = []
    while True:
        block_size = rng.read(1)[0]
        if not block_size:
            continue

        data = bytes([block_size]) + rng.read(block_size)
        if len(data) < 255:
            data += b'\x00'  # Ending with a block size of 0
        else:
            blocks[-1] = blocks[-1][:-1] + bytes([len(data)]) + data[1:]
        blocks.append(data)

    image_data = bytes([lzw_minimum_code_size]) + b''.join(blocks)

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)


if __name__ == '__main__':
    try:
        with open('/dev/urandom', 'rb') as urandom:
            with open('./random.gif', 'wb') as f:
                generate_random_gif(urandom, f)
    except Exception as e:
        print(e)