#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % 256) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptor
    left, top = struct.unpack('>HH', rng.read(4))
    width, height = struct.unpack('>HH', rng.read(4))
    image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    # Generate an LZW code table of all possible codes with minimum code size `lzw_minimum_code_size`
    code_table = [bytes([i]) for i in range(2 ** lzw_minimum_code_size)]

    # Create a list of all codes that can be used as literals
    literal_codes = []
    for i in range(len(code_table)):
        if len(code_table[i]) == 1:
            literal_codes.append(code_table[i])
    
    # Add more codes until we reach our maximum code size
    while True:
        new_codes = []

        # For every existing code...
        for code in code_table:

            # If this is a single character code, add its inverse to the new codes
            if len(code) == 1:
                new_codes += [b'\xFF' + code]
            
            # Otherwise, split the code into two halves and create new codes from them
            else:
                half_length = int((len(code) / 2))

                first_half = code[:half_length]
                second_half = code[half_length:]
                
                new_codes += [first_half + second_half]
                new_codes += [second_half + first_half]
        
        # If any new codes were created, add them to the code table
        if len(new_codes):
            code_table += new_codes
        
        # Check whether we've reached our maximum code size
        if max([len(code) for code in code_table]) >= 256:
            break
    
    # Now let's actually encode some data!
    # First, pick a random starting code
    current_code = rng.read(1)

    # Then start reading bytes from the input stream
    while True:
        # Read one byte from the input stream
        next_byte = rng.read(1)

        # If there are no more bytes available, stop encoding
        if not next_byte:
            break

        # If the byte is in the set of valid literal codes, use it directly
        if next_byte in literal_codes:
            current_code = next_byte
        
        # Otherwise, find the longest matching prefix amongst the codes in the code table
        else:
            best_match = ''
            for code in code_table:
                if code.startswith(current_code + next_byte):
                    if len(best_match) < len(code):
                        best_match = code

            current_code = best_match

        # If the current code is too long, emit a clear code
        if len(current_code) > lzw_minimum_code_size:
            out.write(struct.pack('<H', 2**lzw_minimum_code_size - 1))
            current_code = b''

        # Emit the current code
        out.write(struct.pack('<H', code_table.index(current_code)))

    # Finally, emit a final end code
    out.write(struct.pack('<H', 2**lzw_minimum_code_size - 1))

    # 7. Trailer
    trailer = b'\x3B'

    # Write everything to the output file
    out.write(header + lsd + color_table + gce + image_descriptor + image_data + trailer)