#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF will be of dimensions between [min_width; max_width] x [min_height; max_height].
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, *,
        min_width=1, max_width=100, min_height=1, max_height=100
    ):
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width = rng.randint(min_width, max_width)
    height = rng.randint(min_height, max_height)

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    packed_fields = rng.getrandbits(3)
    background_color_index = rng.getrandbits(1)
    pixel_aspect_ratio = rng.getrandbits(3)
    lsd = struct.pack('>HHBHB', width, height, packed_fields, background_color_index, pixel_aspect_ratio)

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.getrandbits(8) % 256) or 1
    color_table = rng.randbytes(num_colors * 3)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = None
    if rng.getrandbits(1):
        gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.getrandbits(8), rng.getrandbits(8))

    # 5. Image Descriptor
    left = rng.getrandbits(16)
    top = rng.getrandbits(16)
    width = rng.getrandbits(16)
    height = rng.getrandbits(16)
    local_color_table_flag = rng.getrandbits(1)
    interlace_flag = rng.getrandbits(1)
    sort_flag = rng.getrandbits(1)
    reserved_flags = rng.getrandbits(2)
    lzw_minimum_code_size = rng.getrandbits(8)
    image_descriptor = struct.pack(
        '>BHHHHBBBBB',
        0x2C, left, top, width, height,
        local_color_table_flag, interlace_flag, sort_flag, reserved_flags, lzw_minimum_code_size
    )

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    blocks = []
    total_blocks_size = rng.getrandbits(8)  # Make it reasonably small
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.getrandbits(8) + 1)
        blocks.append(rng.randbytes(block_size))

        total_blocks_size -= block_size

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if gce is not None:
        out.write(gce)
    out.write(image_descriptor)
    for block in blocks:
        out.write(struct.pack('B', len(block)))
        out.write(block)
    out.write(trailer)