#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, max_width=100, max_height=100, max_num_colors=256, max_num_frames=100
    ) -> None:
    # 1. Header
    header = b'GIF89a'

    # 2. Logical Screen Descriptor
    # Random width and height (using two bytes for each)
    width, height = struct.unpack('>HH', rng.read(4))
    width = min(max_width, width)
    height = min(max_height, height)

    # Other LSD fields: Packed fields (using a random byte for variety), Background color index, Pixel aspect ratio
    lsd = struct.pack('>HHBHB', width, height, rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])

    # 3. Global Color Table (let's make it have random size, capped at 256 colors, and fully random)
    num_colors = (rng.read(1)[0] % max_num_colors) or 1
    color_table = rng.read(3 * num_colors)  # Random colors

    # 4. Graphics Control Extension (optional)
    gce = struct.pack('>BBBHB', 0x21, 0xF9, 0x04, rng.read(1)[0], rng.read(1)[0])

    # 5. Image Descriptors
    frames = [None] * max_num_frames
    for i in range(len(frames)):
        left, top = struct.unpack('>HH', rng.read(4))
        width, height = struct.unpack('>HH', rng.read(4))
        width = min(max_width - left, width)
        height = min(max_height - top, height)
        image_descriptor = struct.pack('>BHHHHB', 0x2C, left, top, width, height, rng.read(1)[0])
        frames[i] = {
            'left': left,
            'top': top,
            'width': width,
            'height': height,
            'descriptor': image_descriptor,
            'data': [],
        }

    # 6. Image Data (randomly sized, capped at 255 bytes per sub-block)
    lzw_minimum_code_size = (rng.read(1)[0] % 8) or 1

    def add_frame():
        frame = rng.choice(frames)
        frame['data'].extend([
            struct.pack('B', len(block)),
            block,
        ])

    blocks = []
    total_blocks_size = rng.read(1)[0]  # Make it reasonably small
    while total_blocks_size > 0:
        block_size = min(total_blocks_size, rng.read(1)[0] + 1)
        blocks.append(rng.read(block_size))

        total_blocks_size -= block_size

    # 7. Trailer
    trailer = b'\x3b'

    # Write everything to the output file
    out.write(header)
    out.write(lsd)
    out.write(color_table)
    if len(gce) == 6:
        out.write(gce)
    for frame in frames:
        out.write(frame['descriptor'])
        for block in frame['data']:
            out.write(block)
    out.write(trailer)


if __name__ == '__main__':
    from sys import argv, stdout

    with open('/dev/urandom', 'rb') as f:
        generate_random_gif(f, stdout.buffer)