#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_gif(out: BinaryIO, rng):
  # Header (first 10 bytes, version "89a")
  out.write(b"GIF89a")
  # Logical screen width and height, in pixels
  struct.pack("!2H", 500, 500).tofile(out)
  # Global colour table, size is "7" => 2^7 = 128 colours
  struct.pack("B", 7).tofile(out)
  # Reserved byte must be zero
  out.write(b"\x00")
  # Background and aspect ratio - packed into one byte
  out.write(b"\xff\xff")
  
  # Image descriptor block
  #   (first six bytes)
  out.write(b"\x2c\x00\x00\x00\x00\x00")
  #   Image width and height
  struct.pack("!2H", 500, 500).tofile(out)
  #   Local colour table flag: set to 1 for none
  out.write(b"\x01")
  #   Interlaced flag: not used, so set to 0
  out.write(b"\x00")
  #   Colour table size / sort flag: unused
  out.write(b"\x00")
  
  # Pixel stream header
  #   Lower nibble of byte 0 specifies the minimum code length
  out.write(b"\x00")
  #   First six bits of byte 1 specify the maximum code length
  out.write(b"\xe0\x03")
  #   2-bit LZW dictionary size - 2^(n+1)-1 codes total, where n=1..7
  out.write(b"\x02")
  
  # Initialise LZW table with single empty entry
  table = [None] * ((1 << 4) + 2)
  for i in range(len(table)):
      table[i] = [i], 1
  
  clear_code, end_code = 1<<4, 1<<5
  
  # Setup state for next pixel
  cur_pix = ord(rng.read(1)) % 256
  prev_pix = None
  
  while True:
    # Read next pixel from RNG and encode it
    
    new_pix = ord(rng.read(1)) % 256
    if cur_pix == end_code or cur_pix == clear_code:
        raise Exception()
        
    if table[-1][0]:
        table.append([cur_pix])
    
    # Encode current and previous pixel
    data = []
    
    while len(table[cur_pix][0]) > 0:
        cur_pix, data = table[cur_pix]
      
    # Check whether we need to send a CLEAR code before this one
    if cur_pix != prev_pix:
      # Write a CLEAR code
      if len(data) >= 4094:
          out.write(b"\x00")
          struct.pack("!H", clear_code).tofile(out)
          data = []
          
      out.write(b"\x00")
      struct.pack("!H", clear_code).tofile(out)
      data += [[clear_code]]
      
      # Update table
      table.extend([[[prev_pix, cur_pix]], [table[end_code][0]+[prev_pix]]])
    
    # Write last pixel as END code
    if len(data) >= 4094:
        out.write(b"\x00")
        struct.pack("!H", end_code).tofile(out)
