#!/usr/bin/env python3

import struct
from typing import BinaryIO

with open('myfile', 'rb') as in_f:
    with open('outfile', 'wb') as out_f:
        process_blocks(in_f, out_f)
\end{code}

You could also write some class for the header and some class for the block data, but that's probably overkill if your blocks are always 10 bytes long.

Comment: I think `rng.tell()` will be a positive integer (unless the file pointer is at the end of the file or before the first line), so you don't need to check it against zero. You should check this when performing binary I/O though!

Answer: The correct way to do this is using [`struct.unpack`](https://docs.python.org/2/library/struct.html#struct.unpack):

\begin{blockquote}

Unpack the C structure str into a tuple according to fmt. Requires len(str) == calcsize(fmt). The buffer, buf, must contain exactly the amount of data needed by the format; no more, no less. Executing calcsize() prior to unpacking ensures that enough data is available.
\end{blockquote}

So, you can simply read, say, three bytes representing the data length, then use that to read that many bytes from the input stream. For example, if your data is written like so

\begin{code}
var header = new [] { (byte)'A', (byte)'B', (byte)'C' }; // 3 bytes

using var fs = File.OpenWrite("test.data");
fs.Write(BitConverter.GetBytes((int)header.Length));
fs.Write(header);
fs.Write(new byte[] { (byte)'H', (byte)'E', (byte)'L', (byte)'L', (byte)'O' });
\end{code}

Then, to extract it

\begin{code}
using var rng = new FileStream("test.data", FileMode.Open);
while (rng.Position < rng.Length)
{
    int datalen = BitConverter.ToInt32(
                      rng.ReadBytes(sizeof(int)), 0);
    string header = Encoding.ASCII.GetString(
                       rng.ReadBytes(datalen));
    Console.WriteLine("Header: " + header);
}
\end{code}

From there, you can write them to your output file.

Alternatively, you can write your own code to perform the same task:

\begin{code}
static void Main()
{
    const int HEADER_SIZE = sizeof(int);
    using var rng = new FileStream("test.data", FileMode.Open);
    while (rng.Position < rng.Length)
    {
        byte[] buf = new byte[HEADER_SIZE];
        rng.Read(buf, 0, HEADER_SIZE);
        int datalen = BitConverter.ToInt32(buf, 0);
        byte[] header = new byte[datalen];
        rng.Read(header, 0, datalen);
        Console.WriteLine("Header: " + Encoding.ASCII.GetString(header));
    }
}
\end{code}

Note that both of these examples will throw an exception if the input stream is not correctly formatted.

Answer: A little late to the party, but here's a solution using [Numpy](https://numpy.org/) which might come handy for someone else. Note that it does not take care of boundary conditions e.g. if you have multiple blocks in one chunk of data.

It reads the number of bytes for each block (`chunk_size`) and uses the numpy [memmap](https://numpy.org/doc/stable/reference/generated/numpy.memmap.html#numpy-memmap) functionality to read the whole file into a memory map. Then, for every chunk size, the offset from the beginning of the file is calculated and sliced using the memmap object. This results in a list of chunks that can be saved individually.

In addition, a function is defined to save everything in one go.

\begin{code}
import numpy as np

DATAFILEPATH = 'your/path/to/your/data/file.dat'

def load_chunks(chunk_size=10, offset=0):
    '''
    Load chunks from file and return a list
    :param chunk_size: Size of each chunk (default = 10)
    :param offset: Offset from the beginning of the file (default = 0)
    :return: List of chunks
    '''
    fp = np.memmap(filename=DATAFILEPATH, dtype="uint8")
    positions = [offset + i*chunk_size for i in range(len(fp[-chunk_size:])//chunk_size)] # Get all offsets
    chunks = [fp[start:start+chunk_size] for start in positions] # Slice chunks from file
    del fp # Delete file map from memory
    return chunks

def save_all(output_directory='/path/to/save/'):
    '''
    Save all blocks to files, one per file
    :param output_directory: Directory where to save the files
    :return: None
    '''
    chunks = load_chunks()
    for i, chunk in enumerate(chunks):
        filename = f"{output_directory}/{i}.dat"
        with open(filename, 'wb') as fp:
            fp.write(chunk)
\end{code}

Comment: Wow Numpy is great. If I understand correctly, you load the whole thing in memory? Does that mean the file should fit in memory? Could you load only part of the file?

Comment: Yes, you would need to fit the entire file in memory since `np.memmap` basically creates an array on top of filestreams and buffers them internally. To only read parts of the file at a time, you could read up on [delayed loading](https://numpy.org/doc/stable/reference/arrays.ndarray.html) and [slicing](https://numpy.org/doc/stable/user/basics.indexing.html#basic-slicing-and-indexing) of numpy arrays.

Answer: Use the [read method](https://docs.python.org/3/tutorial/inputoutput.html#methods-of-file-objects) inside a loop:

\begin{code}
import struct

CHUNK_SIZE = 10 * 8  # 80 bits
with open('file', 'rb') as fin:
    data = b''
    while True:
        header = fin.read(3)
        if not header:
            break
        datalen = struct.unpack('>I', header)[0]
        datalen = min(datalen, CHUNK_SIZE - len(header))
        data += fin.read(datalen)

print(data)
\end{code}

Comment: What if the last block doesn't have enough space left to fill an entire chunk? And why are you reading the header as a string rather than just interpreting the next 4 bytes as an integer?

Comment: @PMF Just to save an extra struct call? It's worse than useless, since the OP wants the length as a string anyway?

Comment: @ThomasWeller I fixed the chunk size problem.

Answer: \section{1) Read the data one block at a time.}

This is my approach using `struct`.

\begin{code}
import struct

def get_next_block(rng: BinaryIO) -> bytes:
    header = struct.unpack('>I', rng.read(3))[0],
    data = rng.read(header[0])
    return header + data

with open('myfile', 'rb') as rng, open('outfile', 'wb') as out:
    while True:
        try:
            out.write(get_next_block(rng))
        except struct.error:
            break
\end{code}

\section{2) Read and rewrite the data all at once.}

If you want to read the entire file at once and rewrite the result to another file, you can still use the above method in a loop, but instead of writing to individual files, you write to a single file instead.

\begin{code}
with open('myfile', 'rb') as rng, open('outfile', 'wb') as out:
    while True:
        try:
            out.write(get_next_block(rng))
        except struct.error:
            break
\end{code}

\section{3) Directly read and write the data            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + data)
