#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out:BinaryIO, rng:BinaryIO, length=1024):
    # checksum is sum of all bytes in GIF file modulo 256.
    # This value is stored in a byte that follows the data section
    # and must be included in the running total when calculating it.
    chksum = 0x3b
    
    # Magic string
    out.write("GIF89a".encode('ascii'))

    # Logical Screen Descriptor
    # 7 - aspect ratio (1:1)
    # 3 - 256 colors
    # Width
    # Height
    header = struct.pack("<BBHB", 0x70 | 0x30, 0, 100, 100)
    out.write(header)
    
    global_colortable = bytearray(768)
    for i in range(768):
        global_colortable[i] = rng.read(1)[0]
        
    # Background color index
    bgcolor = rng.read(1)[0]
    out.write(struct.pack("<B", bgcolor))
    
    # Pixel aspect ratio
    out.write("\x00")
    
    # Global Color Table
    colortablelength = len(global_colortable)
    out.write(global_colortable[:colortablelength])
    
    # image descriptor
    out.write(struct.pack("<BBHHB", 0x2c, 0, 0, 100, 100, 0))

    # local color table flag + size
    out.write("\x80\x00")

    # zero-length code
    out.write("\x00")
    while True:
        b = rng.read(1)[0] % 256
        out.write(struct.pack("<B", b))
        if b == 0xff: break
    out.write(rng.read(1))
    
    # loop count entry
    out.write("\x00\x01\x00\xff")
    
    # image data
    # A single block of 33 bytes should be enough to
    # fill the entire image. The running total checksum
    # needs to include the first 33 bytes as well as
    # whatever follows them (which is why we stop at 50).
    # However, we'll use one more so we can write the
    # terminating 0xFF byte afterward.
    out.write(rng.read(33))
    chksum += sum([ c * 256 for c in rng.read(17)])

    # add checksum
    chksum &= 0xff
    chksum ^= 0xff
    out.write(struct.pack("<B", chksum))
