#!/usr/bin/env python3

import struct
from typing import BinaryIO, Dict, Iterator, List, NamedTuple, Optional, Tuple, Union

import numpy as np
from PIL.Image import Image
from pydicom.dataset import Dataset, FileDataset
from pydicom.sr.coding import CodedConcept
from pydicom.uid import UID

from .export_sop import get_uids_for_dicom_keywords

# TODO: Add more supported keywords, and add support for additional VR types
SOP_CLASS_UID = "121006"  # SOP Class: Enhanced MR Image Storage
SOP_INSTANCE_UID = "1.3.46.670589.11.2.1.121006.1.20210710.125627.261517775.11.1.1.1"  #
SERIES_MODALITY = "MR"
SERIES_DESCRIPTION = "Test Series Description for image series instance uid {}"
STUDY_DATE = "20210710"
STUDY_TIME = "125627"
ACCESSION_NUMBER = "FKL1234"
PATIENT_NAME = "Doe^John"
PATIENT_BIRTH_DATE = "19700101"
PATIENT_SEX = "M"
MANUFACTURER_MODEL_NAME = (
    "ProtonDyn_BiasFieldCorrectionAndRegistration_mrtrix3-1.1"
)
MANUFACTURER = "FSL"
SOFTWARE_VERSIONS = ["mrtrix3", "mrinfo"]
TRANSFER_SYNTAX_UID = UID("1.2.840.10008.1.2")

KeywordValuePair = Union[str, CodedConcept]
ExportedSeriesInstanceKeywords = Dict[str, KeywordValuePair]
InstanceMetadataType = Union[Dict[str, float], List[float]]


def create_empty_instances(
    num_images: int, filepath: str, overwrite: bool = False
):
    """Create an empty image series with a given number of images.

    Arguments:
        num_images: Number of images to be created in the new series.
        filepath: Path where the instances will be saved.
        overwrite: Whether or not to overwrite existing files. If False, then
            the function will abort if any file already exists at `filepath`.
    Returns:
        A list of strings with all the filenames that were written.
    Raises:
        IOError: An error occurred while writing any of the files.
    """
    filenames = [f"{filepath}/image{i}.dcm" for i in range(num_images)]
    try:
        for filename in filenames:
            if not overwrite and os.path.isfile(filename):
                raise IOError(f"{filename} already exists.")
            with open(filename, "wb") as f:
                f.write(b"\x00" * 128)
    except Exception:
        for filename in filenames:
            if os.path.isfile(filename):
                os.remove(filename)
        raise
    return filenames


def encode(value: InstanceMetadataType) -> bytes:
    r"""Encode metadata values into DICOM primitive data elements.

    Arguments:
        value: Value to be encoded. Must follow one of the following formats:
         - A dictionary mapping {string: float} containing numeric values.
           The keys are VR names that must conform to the DICOM standard.
         - A list of floats containing numeric values. The length of the list
           is the number of components.
    Returns:
        Bytes containing raw data.
    Raises:
        TypeError: Invalid type for encoding.
    """
    if isinstance(value, dict):
        # Convert dictionary to a list of bytes using single planar strategy
        b = bytearray()
        for vr, v in value.items():
            assert isinstance(vr, str), vr
            assert isinstance(v, float), v
            n, binary_vr = _get_vr_info(vr)
            if len(binary_vr) > 2:
                raise ValueError(f"`{vr}` is not a valid VR.")
            b += binary_vr + struct.pack("@d", v)
        return bytes(b)
    elif isinstance(value, list):
        assert len(value) == 3
        # We assume the values are ordered x, y, z
        d = {"PixelSpacing": value, "NumberOfFrames": 1}
        return encode(d)
    else:
        raise TypeError(type(value))


def decode(data: bytes, format=None) -> InstanceMetadataType:
    r"""Decode raw DICOM data into its corresponding Python structure."""
    if format is None:
        format = data[:2].decode().lower()
    if format == "ob":
        num_bytes = data[-4:]
        data = data[:-4]
        num_bytes = int.from_bytes(num_bytes, byteorder="big")
        endianness = "<" if num_bytes < 0 else ">"
        num_components = abs(num_bytes) // 4
        data = struct.unpack(endianness + "{}d".format(num_components), data)
        data = [-d for d in data]
    elif format == "(3f":
        # Single-planar float triplet
        data = data[1:-1]  # Remove parentheses
        data = struct.unpack("@3f", data)
        data = {"PixelSpacing": data, "NumberOfFrames": 1}
    elif format == "(3o":
        # Single-planar signed integer triplet
        data = data[1:-1]  # Remove parentheses
        data = struct.unpack("@3o", data)
        data = [-int(d) for d in data]
    else:
        print(format)
        raise NotImplementedError(f"Unhandled format `{format}` for decoding.")
    return data


def write_metadata(
    filename: str,
    sopclass: str,
    kw: ExportedSeriesInstanceKeywords,
    data_elements: Dict[str, bytes],
    overwrite: bool = False,
):
    """Write metadata to a DICOM file.

    The given keyword pairs will be added to the mandatory file meta information
    group, and prefixed with '00', to satisfy the requirements of the standard.

    Arguments:
        filename: Name of the file where metadata will be written.
        sopclass: Name of the SOP class that is being exported.
        kw: Dictionary of pairs {keyword: value}, where each pair will be stored
            in DICOM tag (keyword, value).
        data_elements: Dictionary of {tag: data} containing optional raw data
            in addition to the mandatory attributes in kw.
        overwrite: Whether or not to overwrite existing files. If False, then
            the function will abort if the file already exists at `filename`.
    """
    if not overwrite and os.path.exists(filename):
        raise IOError("File {} already exists.".format(filename))
    ds = FileDataset(filename, {}, file_meta=get_file_metainfo(sopclass))
    for key, val in kw.items():
        ds[(0x0000, key)] = val
    for tag, data in data_elements.items():
        ds[tag] = data
    ds.save_as(filename)


def read_metadata(
    filename: str, sopclass: str
) -> Tuple[ExportedSeriesInstanceKeywords, Dict[str, bytes]]:
    """Return metadata contained in a DICOM file."""
    def convert_kw(k, v):
        if k.startswith("00"):
            k = k[2:]
        return k, v

    ds = FileDataset(filename, {}, file_meta=get_file_metainfo(sopclass))
    kw = {convert_kw(k, v)[-1]: v for k, v in ds.file_meta.items()}
    kw = {k: v for k, v in kw.items() if not k.startswith("TransferSyntax")}
    return kw, {(grp, el): ds[(grp