#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(rng.read(2), 'little') + 1
    color_table_length = int.from_bytes(rng.read(1), 'big')
    global_color_table_flag = bool(int.from_bytes(rng.read(1), 'big'))

    # The first byte of the image data section begins with a control byte
    # which indicates whether or not this frame should be displayed
    # as an animation. The next four bits are reserved for future use and
    # must be zero. The last four bits give how to interpret the delay time
    # field, in seconds. A value of 64 is used when the delay time is unknown.
    control_byte = rng.read(1)[0] & 0b1111_1000
    if control_byte != 0xA0:
        raise ValueError('The first byte of the image data section must have its high bit set and its lowest four bits clear')

    # Delay time in centiseconds between the end of this frame and
    # the start of the next frame. This value is valid for all frames except
    # the last one, which has no following frame to delay. If it is missing
    # then the delay time should default to 100.
    # (This is not implemented right now.)
    delay_time = rng.read(1)[0]

    transparent_color = None
    if (control_byte & 1 << 5) == 1 << 5:
        transparent_color = rng.read(1)[0]

    block_size = rng.read(1)[0]

    colors = []  # RGB colors
    for _ in range(color_table_length * 3):
        colors.append(struct.unpack('B', rng.read(1))[0])

    # Ignore background color index
    background_color_index = rng.read(1)[0]

    pixel_data = []
    while len(pixel_data) < datalen:
        n = int.from_bytes(rng.read(1), 'big')
        if n == 0:
            break
        else:
            pixel_data += rng.read(n)

    # Raw output
    out.write(header)
    out.write(datalen.to_bytes(2, 'little'))
    out.write(color_table_length.to_bytes(1, 'big'))
