#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, num_frames=10, colors=[0]):
    # Write GIF header
    gif_header = b"GIF89a"
    size = 7

    # Logical screen descriptor
    width = 16 + rng.read(1)[0] % 240
    height = 16 + rng.read(1)[0] % 240
    flags = 0 if len(colors) > 0 else int(rng.read(1).hex(), 16) % 5 or 1
    background = 0xFF
    aspect_ratio = 0
    out.write(struct.pack("!<BHHBB", *[gif_header, width, height, flags,
                                       background, aspect_ratio]))

    for i in range(num_frames):
        # Frame header
        frame_header = [0x2C]
        x = 0
        y = 0
        w = 1 + (width - 2) // 2
        h = 1 + (height - 2) // 2
        flags = 0 # not interlaced
        # Unused color table
        frame_header += [(w + 1) & ~1, h, flags]
        frame_header = bytes(frame_header)
        out.write(frame_header)

        # Graphic control extension
        graphic_extension = bytearray()
        blocksize = b"\xF9"
        flags = 0
        delay = 2
        transparent = 0xFF
        # Block terminator
        graphic_extension.extend([blocksize, flags, delay, transparent])
        graphic_extension.append(0x00)
        out.write(graphic_extension)

        # Image descriptor
        image_descriptor = [0x2C, 0, 0, width-1, height-1, 0]
        out.write(bytes(image_descriptor))

        # Local Color Table
        local_colortable = []
        if flags == 0:
            pass
        elif flags & 0b1000 and len(colors) > 0:
            local_colortable = [[col >> 16 & 0xFF, col >> 8 & 0xFF, col & 0xFF]]
        elif flags & 0b0100:
            local_colortable = [[int(rng.read(1).hex(), 16),
                             int(rng.read(1).hex(), 16),
                             int(rng.read(1).hex(), 16)]]
        for _ in range((w+h)*3):
            out.write(colors[rng.read(1)[0] % len(colors)].to_bytes(3, "big"))

        # Minimal data sub-blocks
        data_subblocks = []
        while True:
            length = rng.read(1)[0] % (w*h//2)
            if length < 1: break