#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
Created on 2017年05月21日
@author: zhaoqianjin
'''
import os
import sys
from typing import List, BinaryIO
import numpy as np

def read_gif(fp: str) -> List[np.ndarray]:
    with open(fp, 'rb') as f:
        _, *images = read_gif_data(f)
    return images

def read_gif_data(fp: BinaryIO) -> List[np.ndarray]:
    '''
    Read GIF binary stream data into list of grayscale arrays
    Args:
        fp (BinaryIO): GIF file pointer
    Return:
        A list of grayscale numpy arrays
    '''
    cmap = [] # color map.
    cmap_size = 0 # color table size.
    width = 0 # image width.
    height = 0 # image height.
    delay = 0 # frame delay.
    dtype = None # image data type.
    images = [] # list of grayscale numpy arrays.
    assert fp.read(6) == b'GIF89a', 'Not a valid GIF file.'
    while True:
        chunk = fp.read(1)
        if len(chunk) < 1:
            break
        code = ord(chunk)
        if code == 0x2c: # image separator
            buf = fp.read(9)
            pos = buf.find(b'\x00')
            left, top, width, height = [int(i) for i in buf[:pos].split(b',')]
            dtype = np.dtype('uint8')
            if len(buf)>=10 and buf[-3]==33: # local color table flag is set
                cmap_size = 2 << (ord(buf[-2]) & 0x07)
                cmap = np.frombuffer(fp.read(3*cmap_size), dtype='uint8').reshape([-1, 3])
                if cmap_size == 2: # gray scale color map
                    cmap = np.concatenate((cmap, np.arange(cmap_size)[..., np.newaxis]), axis=-1).astype('float32')
            else:
                cmap = np.arange(256)[..., np.newaxis][np.newaxis, :]
            lzw_min_code_size = ord(fp.read(1))
            data = _decode_lzw(fp, lzw_min_code_size)
            image = np.asarray(_unpackbits(data, width, height)).view(dtype)
            # Flip the RGB pixel order to RGB.
            image = image[::-1, :, ::-1]
            # Convert the index array to uint8.
            image = image.astype(np.uint8)
            # Dump local color palette into global color palette.
            if cmap_size > 0:
                cmap = np.concatenate((cmap, cmap[np.unique(image)]))
            # Add background color to palette, so that transparent pixels become white.
            bgidx = np.argwhere(cmap==0.).ravel()
            if len(bgidx) == 1:
                bgidx = bgidx[0]
                cmap[:,3] |= 1<<bgidx
            elif len(bgidx) > 1:
                raise ValueError("Background color is not unique.")
            else:
                raise RuntimeError("Cannot find background color in palette.")
            # Replace each pixel value by its corresponding color from the palette.
            image = cmap[image]
            images.append(image)
            delay = 100 // _get_num(fp.read(2))
        elif code == 0x21: # extension
            ext = fp.read(1)
            if len(ext) < 1:
                break
            ext_code = ord(ext)
            if ext_code == 0xf9: # graphics control extension
                gce = fp.read(4)
                if len(gce) != 4 or not gce.startswith(b'\x04'):
                    continue
                flags = ord(gce[3])
                image = images[-1]
                if flags&1: # disposal method
                    prev = images[-2]
                    if prev is None:
                        image[:] = 255 # backround color
                    else:
                        image[:] = prev 
            else:
                size = _get_num(fp.read(2))
                fp.seek(size, os.SEEK_CUR)
        elif code == 0x3B: # trailer
            break
        else:
            print(f"Warning: unknown block {chunk}.", file=sys.stderr)
    return images


def _decode_lzw(stream: BinaryIO, min_code_size: int) -> bytes:
    clear_code, end_code, bit_num_init = 1 << min_code_size, 1 << min_code_size, min_code_size+1
    code_table = [(i ^ 0xFFFFFF) + 1 for i in range(clear_code, clear_code+end_code)]
    first, bits_left, code = 0, 0, 0
    for _ in range(clear_code):
        code_table.insert(_, [])
    code_table.extend(([first], [], []))
    while True:
        while bits_left <= 24:
            bits_left += 8
            first <<= 8
            first |= ord(stream.read(1))
        code = (first >> (bits_left - bit_num_init)) & ((1 << bit_num_init)-1)
        bits_left -= bit_num_init
        if code == clear_code:
            code_table = [(i ^ 0xFFFFFF) + 1 for i in range(clear_code, clear_code+end_code)]
            for _ in range(clear_code):
                code_table.insert(_, [])
            code_table.extend(([first], [], []))
            first, bits_left, code = 0, 0, 0
        elif code != end_code:
            try:
                code_table[code].extend(code_table[code_table[code][0]])
            except IndexError:
                pass
            yield bytes(code_table[code])
            if len(code_table) < 4096:
                code_table.append([code_table[code_table[code][0]][0]] + code_table[code])
        else:
            return


def _get_num(input: bytes) -> int:
    return (ord(input[0]) << 8) | ord(input[1])


def _unpackbits(data: bytes, width: int, height: int, depth=1) -> np.ndarray:
    masks = [sum(2**n for n in range(8*i, 8*(i+1))) for i in range(depth)]
    length = len(masks)*width*height
    a = np.empty(length, dtype='uint8')
    dp = 0
    for s in data:
        for i in range(8):
            for j in reversed(range(len(masks))):
                if dp >= length:
                    return a
                shift = 8-(i+1)
                a[dp] = (s>>shift)&masks[j]
                dp += 1
    return a.reshape([height, width]+[len(masks)]).transpose([0, 2, 1]).copy()


if __name__ == "__main__":
    import glob
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--rng_dir', required=True)
    parser.add_argument('--out_dir', required=True)
    args = parser.parse_args()
    processed_count = 0
    for path in glob.glob(os.path.join(args.rng_dir, '*.txt')):
        filename = os.path.basename(path)
        rng_path = os.path.join(args.rng_dir, filename)
        out_path = os.path.join(args.out_dir, filename)
        
        with open(r        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + data)
