#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate a random GIF header (13 bytes) and write it to output
    magic = 'GIF89a' if rng.read(1)[0] % 2 else 'GIF87a'
    width = int.from_bytes(rng.read(2), byteorder='big') % 560 + 4
    height = int.from_bytes(rng.read(2), byteorder='big') % 560 + 4
    field_size = int.from_bytes(rng.read(1), byteorder='big') % 8
    background_colour_index = rng.read(1)[0] % 256
    aspect_ratio = rng.read(1)[0] % 4
    colour_table_flags = rng.read(1)[0] & 1 << 7 == 1 << 7
    sort_flag = rng.read(1)[0] % 2 == 1
    global_colour_table_size = (
        2 ** ((rng.read(1)[0] >> 4) + 1) * 3
    ) if colour_table_flags else 0
    bg_colour_index = int.from_bytes(rng.read(1), byteorder='big') % 256
    pixel_aspect_ratio = rng.read(1)[0] % 24
    out.write((magic + b'\x00').encode())
    out.write(struct.pack("<HHBB", width, height, field_size, background_colour_index))
    out.write(b'\x00\xff\xff\xff')
    out.write(b'\x00' + (background_colour_index).to_bytes(1, byteorder='big'))
    out.write(pixel_aspect_ratio.to_bytes(1, byteorder='big'))

    # Write local colourable table
    n = 0
    while True:
        entry = ''
        for _ in range(3):
            r, g, b = (rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
            entry += (b | (g << 8) | (r << 16)).to_bytes(3, byteorder='big')

        out.write(entry.encode())
        if len(out.getvalue()) >= global_colour_table_size:
            break

    # Number of frames
    nframes = int.from_bytes(rng.read(1), byteorder='big') % 256 + 1

    # Maximum size of each frame
    maxframe = int.from_bytes(rng.read(3), byteorder='big') % (width * (height - 1)) + 1

    # Write graphic control extension blocks
    gce = b'\x00\x00\x01\x00\x02\x00\x00\xf0\x01'
    out.write(gce.encode())

    # Write special purpose extension block
    spe = b'\x01\x00\x04\x00\x00\x00\x00'
    out.write(spe.encode())

    # For each frame...
    n = 0
    while n < nframes:
        # Frame size
        fs = int.from_bytes(rng.read(3), byteorder='big') % maxframe + 1

        # Graphic control extension block
        gce = b'\x00\x00\x01\x00\x02\x00\x00\xf0\x01'
        out.write(gce.encode())

        # Image descriptor block
        idb = b'\x2c\x00\x00\x00\x00\x00\x00\x00\x00'
        out.write(idb.encode())

        # Local colourable table
        n = 0
        while True:
            entry = ''
            for _ in range(3):
                r, g, b = (rng.read(1)[0], rng.read(1)[0], rng.read(1)[0])
                entry += (b | (g << 8) | (r << 16)).to_bytes(3, byteorder='big')

            out.write(entry.encode())
            if len(out.getvalue()) >= fs:
                break

        # LZW encoded frame data
        w = int.from_bytes(rng.read(1), byteorder='big') % 8
        clear = int.from_bytes(rng.read(1), byteorder='big') % 2**w
        eoi = int.from_bytes(rng.read(1), byteorder='big') % 2**w
        code_size = int.from_bytes(rng.read(1), byteorder='big') % 12 + 3
        buf = b''

        while True:
            lzw = []
            while True:
                v = (int.from_bytes(rng.read(1), byteorder='big')) % 2**code_size
                if v == clear:
                    buf = b''
                    code_size = int.from_bytes(rng.read(1), byteorder='big') % 12 + 3
                    break
                elif v > 0:
                    lzw.append(v)
                    if v <= eoi:
                        break
