#!/usr/bin/env python3
"""
    Check the UFO font for all glyphs for vertical metrics.
    Based on FontLab's (c) 2018-04-06, Mojca Miklavec.
"""

from __future__ import print_function, division, absolute_import

import sys
from fontlab import (
    styleLib as sl,
    styleTools as st,
)
from fontTools.misc.py23 import *
from fontTools.ttLib import TTFont
from fontGoggles.utils import loadUFO, saveUFO
from argparse import ArgumentParser
from itertools import chain
import os


def getArgs():
    parser = ArgumentParser(usage=__doc__)
    parser.add_argument("infile")
    return parser.parse_args()


def process(ufo):
    font = loadUFO(ufo)
    if not (font.info.openTypeHheaAscender or font.info.openTypeHheaDescender):
        print("Checking OpenType metrics in %r ..." % ufo)
        try:
            ttFont = TTFont(os.path.join(font.path, "font.ttf"))
            ascent = max(ttFont["hhea"].ascent,
                         ttFont["OS/2"].usWinAscent)
            descent = min(-ttFont["hhea"].descent,
                          -ttFont["OS/2"].usWinDescent)
            print("%d units of ascender, %d units of descender." % (ascent, descent))
        except Exception as e:
            print("Error: Couldn’t read font file.")
            raise(e)
    else:
        ascent = font.info.openTypeHheaAscender
        descent = font.info.openTypeHheaDescender
        print("Reading values from font info...")
    # Mark all glyphs whose advance height is inconsistent with the
    # specified TrueType hhea metric values.
    bad = {}
    for gName, g in font.glyphs.items():
        if gName == ".notdef":
            continue
        xAdvance = g.width
        yAdvance = g.height
        xHeight = font.info.xHeight
        cCapHeight = font.info.capHeight
        yTop = g.layers[0].bounds[1]
        yBottom = g.layers[-1].bounds[3]
        if (yTop - yBottom != yAdvance and yTop > yAdvance + descent and
                yBottom < yAdvance + ascent):
            bad[gName] = (xAdvance, yAdvance, xHeight, cCapHeight,
                          yTop - yBottom, yTop, yBottom)
    if bad:
        markBad(bad, font)
        saveUFO(font)
    else:
        print("All glyphs are fine\!")


def markBad(guysWithBadAdvances, f):
    for gName, (advWidth, advHeight, xHeight, capHeight,
                topDiff, yTop, yBottom) in guysWithBadAdvances.items():
        g = f.glyphs[gName]
        g.mark = 70  # Unassigned Glyph Code
        f.notify(sl.NotificationCenter.ContentsDidChangeNotification, gName)
        color = "#F50"
        message = ('Wrong Advance Height of "%s"\!' % gName)
        wBottom = yBottom
        wTop = yTop
        if yBottom > yTop:
            m = min(abs(topDiff), yTop - yBottom)
            wBottom -= m
            wTop += m
            message = 'Looks like a sideways!'
        f.appendMessage(styleLib.Message(message, gName,
                                         [st.StyleVertLineSegment(wTop - xHeight / 2.,
                                                                  wBottom + xHeight / 2.,
                                                                  name="x-height",
                                                                  color=color, locked=True)] +
                                         [st.StyleVertLineSegment(wTop - capHeight / 2.,
                                                                  wBottom + capHeight / 2.,
                                                                  name="cap-height",
                                                                  color=color, locked=True)] +
                                         [st.StyleHorzLineSegment(advWidth, advHeight,
                                                                  name="adv height",
                                                                  color=color, locked=True)],
                                         None))
    f.updateObject(sl.GlyphsID, sl.NotifyChanged)

