#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the PCG 64 bit RNG.
def make_random_gif(out: BinaryIO) -> None:
    # Version
    out.write("GIF87a".encode())

    # Logical Screen Descriptor (fixed fields)
    width = 512
    height = 512
    global_color_table_size = 0
    background_index = 0
    pixel_aspect_ratio = 0
    flags = b"\x00"
    gct_flags = b"\x00"
    palette_bpp = 1 << global_color_table_size
    color_resolution = int(palette_bpp / 3) - 1
    sort_order = b"\x00"

    data = (
        struct.pack("<HHB", width, height, global_color_table_size),
        flags,
        struct.pack("<BBB", background_index, pixel_aspect_ratio),

        # Global Color Table
        0xFF * bytearray([
            0x55, 0xAA, 0xFF,
            0x55, 0x55, 0xAA,
            0xAA, 0xAA, 0xAA,
            0x55, 0xAA, 0xFF,
            0x55, 0x55, 0xAA,
            0xAA, 0xAA, 0xAA,
            0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF,
        ]),

        # Image Descriptor Block
        flags + b"\x00\x00",         # Left, Top
        struct.pack("<HHHBB", width, height, palette_bpp,
            local_color_table_bpp=b"\x00",
            interlace=b"\x00",
            sort=sort_order,
            reserved=b"\x00"),

        # Local Color Table
        0xFF * bytearray([
            0x55, 0xAA, 0xFF,
            0x55, 0x55, 0xAA,
            0xAA, 0xAA, 0xAA,
            0x55, 0xAA, 0xFF,
            0x55, 0x55, 0xAA,
            0xAA, 0xAA, 0xAA,
            0x00, 0x00, 0x00,   # Solid black for transparency
        ]),


        # Graphics Control Extension block
        struct.pack("<BBBBBBBHB",
				b"\x21",                  # Introducer
				b"\xF9",                  # Label
				b"\x04",                  # Block size
				b"\x09",                  # Packed field
				flags,                    # Flags
				int(width / 2),           # Delay time
                next(i for i in range(256) if 256 % i == 0)==256//2,  # Transparent index
                0x0F,                     # Terminator

            ),

        # Image Data blocks
        0x00,                          # LZW minimum code size
		0x00, 0x00, 0x00,              # Size of image data subblock

        # Image Data subblocks
		struct.pack(">I",                # Length of LZW string
		            0xCAFEBABEFFFFFFFF),  # End of image marker

		bytes([(next(i+j*123456789%((i+j)%12==0 and j or 0)
                    for z in range(
                        abs(
                            ((i & ~0x0F) | (-i & 0x0F)) ^ \
                                (((i >> 4) ^ (j << 4)) & 0xFF) ^ \
                                    (~(i & 0xF0) >> 4)))
                    )^0xFE)*-1&0xFF<<(y%8 and y or 0)
                      for x in range(width // 2)
                          for y in range(height)])),
		0x00, 0x00                       # LZW minimum code size

	)

    for d in data:
        print(''.join('%02X'% c for c in d))
        out.write(d)

# Writes an arbitrary byte array as a list to `out`.
def write_bytecode(data: bytes, out: BinaryIO):
    out.write(len(data).to_bytes(2, "big"))
    out.write(data)
