#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header (6 bytes)
    header = b"GIF89a" + \
             struct.pack("<H", 0x100)

    # Logical screen descriptor (7 bytes)
    width = int(rng.read(2)) % 256
    height = int(rng.read(2)) % 256
    flags = b'\x87\x00' if width < height else b'\x00\x87'
    background = int(rng.read(2)) % 256
    aspect_ratio = int(rng.read(1)) % 9
    logical_screen_descriptor = struct.pack(
        "<2B3s2BBBHB",
        width,
        height,
        flags,
        background,
        aspect_ratio
    )

    # Global color table declaration and end of global block (7 bytes)
    global_color_table_size = {
        1:  2 << 4 | 1,
        2:  2 << 4 | 2,
        4:  2 << 4 | 4,
        8:  2 << 4 | 8,
        16: 3 << 4 | 16
    }[int(rng.read(1)) % len([t for t in global_color_table_size])]
    
    global_color_table = [b"\x00\x00\xff"] * 2**(global_color_table_size+1)
    end_of_block = b"\x3B"

    # Image data block (12 bytes)
    image_left = int(rng.read(2)) % 256
    image_top = int(rng.read(2)) % 256
    image_width = int(rng.read(2)) % 256
    image_height = int(rng.read(2)) % 256
    packed_flags = int(rng.read(1)) % 8
    local_color_table_flag = bool(packed_flags & 1)
    interlace_flag = bool((packed_flags >> 5) & 1)
    sort_flag = bool((packed_flags >> 6) & 1)
    reserved_bits = int(rng.read(1)) % 256
    palette_size = int(rng.read(1)) % 256
    lzw_mincodesize = int(rng.read(1)) % 9
    local_color_table = []
    if local_color_table_flag:
        for i in range(palette_size):
            r = int(rng.read(1)) % 256
            g = int(rng.read(1)) % 256
            b = int(rng.read(1)) % 256
            local_color_table += [r, g, b]

    local_color_table += [b"\x00\x00\xff"] * max(0, 2**(lzw_mincodesize+1)-len(local_color_table))
    image_data_block_header = struct.pack(
        "<HHHHBB",
        image_left,
        image_top,
        image_width,
        image_height,
        packed_flags,
        lzw_mincodesize
    )

    images = []
    for _ in range(1 + int(rng.read(1)) % 10):
        compression_method = int(rng.read(1)) % 4
        if compression_method!= 3:
            raise Exception("Unsupported compression method")

        num_bytes_in_compressed_data = int(rng.read(2)) % 2**15
        compressed_data = rng.read(num_bytes_in_compressed_data)
        
        data_subblocks = []
        while True:
            subblock_id = int(rng.read(1)) % 3

            if subblock_id == 0:
                break
            
            subblock_length = int(rng.read(1)) % 2**8
            subblock_data = rng.read(subblock_length)
            data_subblocks += [(subblock_id, subblock_length, subblock_data)]
        
        image_data = bytes()
        for id_, length, data in sorted(data_subblocks)[::-1]:
            image_data += bytes([id_])
            image_data += bytes([(length % 2**8)])
            image_data += data
            
        images += [image_data]
    
    
    # Now generate the final file contents
    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(struct.pack("<B", global_color_table_size))
    out.write(bytes(global_color_table))
    out.write(end_of_block)


    for image in images:
        image_data_block = image_data_block_header + image
        out.write(image_data_block)
    
