#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # width and height of the image
    w = 600
    h = int((w / 1920) * (1080 - 50))

    # frame delay in milliseconds
    d = b'\x7d\x00'

    # pixel values
    pixels = [b'#' if rng.read(1)[0] & 2 == 0 else b'' for _ in range(h*w)]
    interlaced = []
    for i in range(0, len(pixels), 8):
        interlaced += [
            pixels[i],           # Every second line 1st pixel
            pixels[(i+4)%len(pixels)],      # Every second line 2nd pixel
            pixels[(i+2)%len(pixels)],      # Every fourth line 1st pixel
            pixels[(i+6)%len(pixels)],      # Every fourth line 2nd pixel
            ]
    
    unpacked = bytearray()
    for y in reversed(range(0, h-1, 2)):
        for x in range(8):
            unpacked += bytes([(interlaced[y][x] << 4) | interlaced[y+1][x]])
    
    packed = []
    while True:
        this = unpacked[:255]
        del unpacked[:255]

        if not this: break
        
        lz = len(this).to_bytes(1, 'little')
        packed += [lz, this]

    # Write all of it to disk as a gif!
    # NOTE: This is actually pretty efficient
    # since we're writing very small amounts of data
    # over and over again...
    out.write(struct.pack('<HHB', w, h, 0xf7)) # Logical screen descriptor
    out.write(b','.join([d]*(2**(len(packed)//2)-1)))   # Application extension
    out.write(struct.pack('BBBBBBB', *reversed(b'NETSCAPE'))) # Extension introducer
    out.write(b'\x03\x01\x00\x00') # NETSCAPE extension
    out.write(struct.pack('<H', 2**16-1)) # Loop counter
    out.write(b';')   # Sub-block terminator

    for p in packed:
        out.write(p)  # Data sub-blocks

    out.write(b'\x00')   # Block terminator

    data = bytes(unpacked)
    out.write(header + data)
