#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given parameters.
def generateGIF(
    out:    BinaryIO,
    width:  int =  200, # Width of image in pixels (must be > 1)
    height: int =  200, # Height of image in pixels (must be > 1)
    speed:  int =    5 # Speed of animation in hundredths of seconds
                        # (must be >= 0 and <= 65535)
):

    # Validate input params
    if not isinstance(width, int) or width < 1:
        raise Exception("width is invalid")
    elif not isinstance(height, int) or height < 1:
        raise Exception("height is invalid")
    elif not isinstance(speed, int) or speed < 0 or speed > 65535:
        raise Exception("speed is invalid")

    # Write header to file
    out.write(b"GIF89a")

    # Write logical screen descriptor
    out.write(struct.pack("<HHHbb", width, height, 0xF000 | ((1 << 7) + (1 << 4)), 0x00, 0x00))

    # Write color table
    for _ in range(0, 256):
        out.write(bytes([0xFF, 0x00, 0xFF]))

    # Write graphic control extension
    out.write(bytes([
        0x21, 
        0xF9,          # Graphic control label byte
        0x04,          # Block size in bytes
        0x08,          # Packed field
        0x00,          # Delay time in milliseconds high-byte
        0x00           # Delay time in milliseconds low-byte
    ]))
    
    # Write application extention
    out.write(bytes([
        0x21,
        0xff,
        0x0b,
        0x4e,
        0x45,
        0x54,
        0x53,
        0x43,
        0x41,
        0x50,
        0x45,
        0x32,
        0x2e,
        0x30,
        0x03,
        0x01,
        0x00,
        0xf9
    ]))
    
    # Write plain text subheader
    out.write(bytes([
        0x21,
        0x0c,
        0x00,
        0x0b,
        0x47,
        0x49,
        0x46,
        0x38,
        0x39,
        0x61,
        0x01,
        0x00,
        0x3b
    ]))

    # Write comment data
    out.write(b"Created by gifgen!")

    # Write terminator
    out.write(b";")

    # Write color blocks
    for y in range(0, height // 2):

        for x in range(0, width // 2):

            offsetX = x * (y % 2 == 1)
            offsetY = y * (x % 2 == 1)

            out.write(bytes([(offsetY ^ offsetX) & 0xFF]))

    # Write EOF
    out.write(b"\x3B")