#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the LZW algorithm. The resulting GIF should be 4096 bytes in size, or less if there are fewer than 256 distinct frames and no compression.
def gen_gif(out: BinaryIO):
    out.write(b'GIF89a') # Header

    # Logical Screen Descriptor
    width = 127 + (ord('A') << 8)
    height = 127 + (ord('B') << 8)
    flags = b'\x87\x00'         # Global color table flag on, sorted palette flag off
    bg_color = ord('.')        # Index of background color in global CLUT
    aspect_ratio = b'\x00'     # Pixel aspect ratio is square pixels
    screen_desc = struct.pack('<HHBBB', width, height, flags[0], flags[1], bg_color, aspect_ratio[0])
    out.write(screen_desc)

    # Global Color Table
    clut = []
    for i in range(256):
        r = int((i / 256) * 255)
        g = int(((i % 256) / 128) * 255)
        b = int((((i % 256) % 128) / 64) * 255)

        byte = ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3)
        clut += [byte]
    assert len(clut) == 768
    out.write(bytes(clut))

    out.write(struct.pack('<BBBBBB', 0x2C, 0, 0, 0, width, height)) # Image descriptor
    lzw_mincodesize = 2      # Minimum code size; can be any value between 2-12 inclusive at this point.
    lzw_init_clear_code = 1  # Maximum code size will be max_code_size+1. init_clear_code must be less than or equal to max_code_size+1.
    lzw_init_end_code = 257  # End code must be greater than or equal to init_clear_code. init_end_code must be less than or equal to max_code_size+1.
    lzw_max_bits = 12        # Maximum number of bits per LZW code word. Can be set as high as 12.
    lzw_max_code = pow(2, lzw_max_bits) - 1   # Maximum code that can be stored in the LZW code words. Will always be 2**lzw_max_bits - 1.
    lzw_clear_block = bytes([lzw_mincodesize, lzw_init_clear_code, lzw_init_end_code])  # Clear block consisting of two header fields and one end code.
    out.write(lzw_clear_block)
    clear_code = pow(2, lzw_mincodesize)          # Actual maximum code is both init_clear_code and init_end_code combined. Since we only allow up to 256 codes though...
    lzw_prefix = lzw_init_clear_code              # Prefix to start new chains with. Set to init_clear_code initially.
    lzw_suffixes = {data: data for data in b'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789!#$%&\'()*+,-./:;=?@[]^_`{|}~'}    # Mapping from suffixes to their corresponding prefixes.

    frame_count = 0
    while True:
        frame_count += 1
        frame = f'Frame #{frame_count}'
        print("Generating frame", frame[:50].replace('\t','')[:50], flush=True, end='\r')       # Don't let it go past the edge of the terminal window

        lzw_data = b''
        for char in frame:
            lzw_data += bytes([ord(char)])

        lzw_header = bytes([lzw_mincodesize, lzw_init_clear_code, lzw_init_end_code]) 
        lzw_first_word = b'\x01'      
        lzw_data += lzw_clear_block
        lzw_data += lzw_first_word
        lzw_last_code = None
        lzw_codes = []

        num_chars = len(lzw_data)+len(lzw_header)-2   # Number of characters introduced since the last time the code was cleared.
        
        while len(lzw_data) > 0:
            possible_matches = {}

            for key, val in lzw_suffixes.items():
                if lzw_data[:len(key)] == key:
                    num_chars -= len(val)

                    if not val in possible_matches:
                        possible_matches[val] = {'pre': [],'suf': [], 'len': {}}
                    
                    possible_match = possible_matches[val]
                    pre = lzw_prefix
                    suf = key[1:]
                    possible_match['pre'].append(pre)
                    possible_match['suf'].append(suf)
                    possible_match['len'][pre] = len(possible_match['pre']) - 1

                    lzw_data = lzw_data[len(key):]

                    break
            
            if len(possible_matches) > 0:
                next_prefix = min(possible_matches, key=(lambda k: possible_matches[k]['len'][lzw_prefix]))

                if not next_prefix in lzw_suffixes:
                    next_code = len(lzw_codes) + clear_code
                    lzw_codes.append(next_prefix)
                    lzw_suffixes[next_prefix] = next_code
                
                prefix = next_prefix
                suffix = lzw_data[:len(next_prefix)]
                length = len(possible_matches[next_prefix]['suf'])

                lzw_data = lzw_data[len(suffix):]
                lzw_prefix = prefix
                continue

            else:
                if not lzw_last_code is None:
                    lzw_codes.append(lzw_last_code)
                
                if len(lzw_codes) >= lzw_max_code + 1:
                    raise Exception('Too many codes!')
                    
                next_code = len(lzw_codes) + clear_code
                lzw_codes.append(lzw_data[0:1])
                lzw_last_code = next_code
            
                lzw_prefix = next_code
                lzw_data = lzw_data[1:]

        bits = bin(lzw_mincodesize)[2:]
        lzw_mincodesize = lzw_max_bits + 1
        lzw_max_code = pow(2, lzw_max_bits) - 1

        if len(bits) <= lzw_max_bits:
            lzw_max_bits = lzw_max_bits - len(bits)
        else:
            lzw_max_bits = lzw_max_bits - 1

        clear_code = pow(2, lzw_mincodesize)          # Actual maximum code is both init_clear_code and init_end_code combined. Since we only allow up to 256 codes though...
        lzw_prefix = lzw_init_clear_code              # Prefix to start new chains with. Set to init_clear_code initially.
        lzw_suffixes = {data: data for data in b'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789!#$%&\'()*+,-./:;=?@[]^_`{|}~'}    # Mapping from suffixes to their corresponding prefixes.

        if lzw_mincodesize < 2:
            lzw_mincodesize = 2 
            lzw_max_bits = 12
            lzw_max_code = pow(2, lzw_max_bits) - 1

        if lzw_mincodesize > lzw_max_bits:
            break

        lzw_chain = []     # Code chain for current image data being written; used when emitting the LZW image data later.
        lzw_codewords_pos = 0    # Position within the codewords to write the next code word.
        lzw_num_codewords = math.ceil(math.log2(len(lzw_codes)))    # Number of codewords required to represent all codes sent so far.
        lzw_codewords = bytearray(b'\x00'*lzw_num_codewords)      # Byte array containing codewords.
        lzw_end_of_image = False
        lzw_code = lzw_init_clear_code

        lzw_header = bytearray(lzw_header)    # Make a copy of our header so we don't modify the version byte.
        lzw_header[-2] |= ((lzw_max_bits & 0x0F) << 4)
        first_block = True                   # Flag