#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Size of entire image in bytes. Set to 0 for now and we'll update this later when we know how many pixels it should be.
    size = struct.pack('<I', 0)

    # Always set to '1'.
    packed_byte = b'\x01'

    # Screen width and height (16-bit each). Defaulted to 80 x 24 characters. Increased to accomodate our sprites as well.
    screen_width = struct.pack('<H', 176)    # 176*24=4224
    screen_height = struct.pack('<H', 50)     # 50*80=4000

    # Reserved fields that are ignored by most software.
    reserved = b'\x00\x00'                 # For non-animated gifs


    # Global colour table length. We're gonna use RGB triplets for storing all the colours used in the gif.
    # There's no need to specify an actual color table since the program is fully capable of adding new colors on its own.
    global_color_table_len = b'\xff'         # Since we're not specifying any color table we use 256 entries here.

    # Background color index; always 0. This field is ignored anyway but still required.
    background_color_index = b'\x00'        # Just a black pixel.

    # Pixel aspect ratio - must be 0 (indicating square pixels). If anything else is specified then the gif won't load properly.
    pixel_aspect_ratio = b'\x00'            # Squares can fit anywhere so there's no need to preserve the original aspect ratio.

    # Plain vanilla gif header. You can get more details about what these bytes mean here: https://en.wikipedia.org/wiki/GIF#Structure
    out.write(header + size + packed_byte + screen_width + screen_height + reserved + global_color_table_len + 
                background_color_index + pixel_aspect_ratio)
    
    # Other settings related to the animation loop which I don't really care about right now.
    num_of_iterations = b'\x00'             # Number of times the animation loops back after reaching the end [usually 0]. 
    duration = b'\x00\x00'                  # Duration between frames in hundredths of seconds [usually 0].
    transparent_idx = b'\x00'               # Index of the palette entry to make the background color transparent with [usually 0]

    # Finalize the header with these optional settings. These aren't strictly necessary and may be omitted.
    out.write(num_of_iterations + duration + transparent_idx)

    # LZW minimum code size. Setting it to 2 gives us at least 4 bits per pixel.
    lzw_min_code_size = b'\x02'

    # Write the rest of the header.
    out.write(lzw_min_code_size)

    print('Writing pixels...')
    while True:
        c = int.from_bytes(rng.read(1), byteorder='big')

        if(c==0):
            break
        
        # Writes the palette bytes. Each value will be a single byte indicating the index of the color from the palette.
        # The palette itself is managed by the program rather than being hardcoded like it is for other formats such as PNGs or BMPs.
        # This was done mainly because the maximum amount of possible colors in a gif has to be less than 256 due 
        # to the limited bit depth of the encoding.
        out.write(struct.pack("<B", c % 0xFF))

    # After writing all the data, let's go ahead and write the EOF marker.
    out.write(b';')
    
    print("Done!")

