#!/usr/bin/env python3

import struct
from typing import BinaryIO

import numpy as np


def read_byte(f: BinaryIO) -> int:
    """Read a single byte from f."""
    return ord(f.read(1))


def write_byte(f: BinaryIO, x: int):
    """Write a single byte to f."""
    assert 0 <= x < 256
    f.write(bytes((x,)))


def read_short(f: BinaryIO) -> int:
    """Read a little-endian short integer from f."""
    return struct.unpack('<h', f.read(2))[0]


def write_short(f: BinaryIO, x: int):
    """Write a little-endian short integer to f."""
    assert -(2**15) <= x < 2**15
    f.write(struct.pack('<h', x))


def read_int(f: BinaryIO) -> int:
    """Read a little-endian signed integer from f."""
    return struct.unpack('<i', f.read(4))[0]


def write_int(f: BinaryIO, x: int):
    """Write a little-endian signed integer to f."""
    assert -(2**31) <= x < 2**31
    f.write(struct.pack('<i', x))


def read_long(f: BinaryIO) -> int:
    """Read a little-endian long integer from f."""
    return struct.unpack('<q', f.read(8))[0]


def write_long(f: BinaryIO, x: int):
    """Write a little-endian long integer to f."""
    assert -(2**63) <= x < 2**63
    f.write(struct.pack('<q', x))


def read_float(f: BinaryIO) -> float:
    """Read a little-endian IEEE754 floating point number from f"""
    return struct.unpack('<f', f.read(4))[0]


def write_float(f: BinaryIO, x: float):
    """Write a little-endian IEEE754 floating point number to f"""
    f.write(struct.pack('<f', x))


def read_double(f: BinaryIO) -> float:
    """Read a little-endian double precision floating point number from f"""
    return struct.unpack('<d', f.read(8))[0]


def write_double(f: BinaryIO, x: float):
    """Write a little-endian double precision floating point number to f"""
    f.write(struct.pack('<d', x))


def read_string(f: BinaryIO) -> str:
    """Read a string of ASCII characters from f with an unsigned short length prefix"""
    n = read_short(f)
    s = f.read(n).decode('ascii')
    if len(s)!= n:
        raise EOFError("Unexpected end of file.")

    return s


def write_string(f: BinaryIO, s: str):
    """Write the given string to f including its utf-8 encoding and two bytes for the length"""
    b = s.encode('utf-8')
    write_short(f, len(b))
    f.write(b)


def read_list(f: BinaryIO, t: type, cnt: int=None):
    """Read list of values of type t (one or more times). If cnt is provided, exactly that many items are read; otherwise, the count is read next as an unsigned short and one item per iteration is yielded until nothing remains in the stream."""
    while True:
        if cnt is not None:
            if cnt == 0:
                break

            cnt -= 1

        yield t.read(f)


def write_list(f: BinaryIO, l: list, t: type):
    """Write all elements in l of type t to f"""
    write_short(f, len(l))
    for i in range(len(l)):
        t.write(f, l[i])


class