#!/usr/bin/env python3
# Copyright 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
"""Utilities for launching jobs on SLURM."""

import math


def batchify(xs, bs):
    """Split list into batches of size at most `bs`."""
    nbatch = int(math.ceil(len(xs)/float(bs)))
    return [xs[i*bs:(i+1)*bs] for i in range(nbatch)]


def run_sinteractive():
    import subprocess as sp

    # find interactive queue with lowest priority
    result = sp.run(['scontrol','show', 'partition'], stdout=sp.PIPE,
                    universal_newlines=True)
    qrecs = sorted([l.split()[-1] for l in result.stdout.strip().split('\n')])
    try:
        iq = next(idx for idx, qrec in enumerate(qrecs) if 'iq' in qrec)
    except StopIteration:
        raise RuntimeError('Could not find interactive partition.'
                           f'Partition records were:\n{result.stdout}')

    print(f'Launching interactive job on {qrecs[iq]}...')
    cmd = ['sinteractive', '--qos=%s' % qrecs[iq]]
    proc = sp.Popen(cmd)
    print('Launched process.')
    proc.wait()


def pipe_to_file(
        fid: str, chunksize: Optional[int]=None
) -> BinaryIO:
    ''' Open a binary file and write to it from a pipe.

        Args:
            fid: A filename string containing the full path
                NOTE: Will create intermediate directories needed

            chunksize (default None): Optional argument to specify
                how many bytes should be read into memory before
                writing them to disk. If ``None``, all content will
                be written to disk immediately after reading from
                buffer.'''
    ddir, fname = os.path.split(fid)
    if len(ddir) > 0:
        os.makedirs(ddir, exist_ok=True)
    fd = os.open(fid, os.O_WRONLY | os.O_CREAT, mode=0o644)

    def _pipe_wrapper(_fd: int, _chunksize: Optional[int]):
        while True:
            chunk = sys.stdin.buffer.read(_chunksize)
            if len(chunk) == 0: break
            os.write(_fd, chunk)

    p = Process(target=_pipe_wrapper, args=(fd, chunksize))
    p.start()
    return fd, p



def copy_rng(
        srcfn: str, tgtfn: str, *, bufsize: int=512**2
) -> Tuple[BinaryIO, BinaryIO]:
    ''' Copy rng output to a file by opening pipes between squeue & tee.

        Args:
            srcfn: Filepath of the RNG pipe to read from.

            targetfn: Filename of the desired location to store the
                copied contents.

            bufsize: Buffer size in bytes for file I/O'''
    assert bufsize >= 0,'must have nonnegative buffer size'
    if os.path.exists(tgtfn):
        raise IOError("File already exists")
    src_fd, src_p = pipe_to_file(srcfn, bufsize)
    _, tgt_p = pipe_to_file(tgtfn, bufsize)
    src_p.join()
    tgt_p.join()
    tgt_p.close()
    # directly close instead of using closer context manager so that
    # we can manually override returncode status
    os.close(src_fd)
    return src_fd, tgt_p


def get_current_job_id() -> Optional[str]:
    ''' Get current slurm job ID using bash builtins.

        Returns:
            Job id as a string or None if no job was launched.'''
    import subprocess as sp
    try:
        jid = sp.check_output(
                "echo $SLURM_JOBID", shell=True).decode('utf-8').rstrip()
    except sp.CalledProcessError:
        return None
    else:
        return jid


def save_random_state(outfn: str):
    ''' Save entire Python random state to a pickle file.

        Args:
            outfn: Output filepath; must end extension.pkl or.pickle'''
    with open(outfn, 'wb') as fout:
        pickle.dump(np.random.get_state(), fout)


def load_random_state(infn: str):
    ''' Load Python random state from a pickle file.

        Args:
            infn: Input filepath; must end extension.pkl or.pickle'''
    with open(infn, 'rb') as fin:
        np.random.set_state(pickle.load(fin))


def seed_from_random_state(rs: np.random.RandomState):
    ''' Seed numpy's RNG from a RandomState instance in a way compatible with
        PyTorch / torchtext's methods of generating seeds.'''
    ss = rs.get_state()[1][0] >> 1
    torch.manual_seed(ss)


def generate_seeds(
        seed: int, nseeds: int=1000
) -> Iterator[int]:
    ''' Generate an infinite number of seeds derived from a given input seed.

        Args:
            seed: Base seed value from which to begin generation.

            nseeds: Number of seeds to generate before stopping. Note that
                seeds are generated one at a time, so setting this parameter
                too high may lead to degenerate behavior.'''
    rs = np.random.RandomState(seed)
    for ii in range(nseeds):
        yield rs.randint(-2**32, 2**32 - 1)


def check_py_files(*fns: Sequence[str]) -> bool:
    ''' Check whether there are any syntax errors in provided files.

        Args:
            *fns: List of filenames of Python scripts to check.'''
    good = []
    bad = []
    for fn in fns:
        try:
            ast.parse(open(fn, encoding='utf-8').read())
            good.append(fn)
        except SyntaxError as e:
            bad.append((fn, str(e)))
    if len(bad) > 0:
        msg =''.join(['%s (%s)' % (f, m) for f, m in bad])
        raise SyntaxError(msg)
    return len(good) == len(fns)




@contextmanager
def tqdm_stdchannel_redirected(to=os.devnull, stdchannel=os.devnull):
  """
  Redirect stdout or stderr to a file.

  :param to: File name, or io.StringIO or io.BytesIO.
  :param stdchannel: stdout or stderr.
  """
  if isinstance(to, str):
     to = open(to, 'w')

  oldstdchannel = None
  try:
      oldstdchannel = os.dup(stdchannel.fileno())
      os.dup2(to.fileno(), stdchannel.fileno())

      yield to
  finally:
      if oldstdchannel is not None:
          os.dup2(oldstdchannel, stdchannel.fileno())    # restore
      if not isinstance(to, str):
          to.close()



def extract_frames(
        gifname: str, start: int=0, stop: int=-1, step: int=1, destfolder: str=''
):
    ''' Extract frames from a GIF image.

        Args:
            gifname: Name (and possibly absolute folder location) of GIF
                image to extract frames from.

            start: First frame index to extract. Default 0.

            stop: Last frame index to extract. Defaults to last frame.

            step: Step size used when extracting frames. Defaults to
                1, meaning every single frame gets extracted.

            destfolder: Absolute folder where to put the extracted
                images. Defaults to creating subfolders in same parent
                folder as the original GIF.'''
    from PIL import Image   # type: ignore


    img = Image.open(gifname)
    if stop < 0: stop += img.n_frames
    destfolder = pathlib.Path(destfolder)
    for ii, frameinfo in enumerate(img.convert('RGB').iter_frames()):
        if (ii < start) or ((stop >= 0) and (ii > stop)): continue
        if ii % step!= 0: continue
        if destination is False:
            yield frameinfo
        else:
            filename = '%s_%05d.png' % (pathlib.Path(gifname).stem, ii)
            outfile = destfolder / filename
            outfile.parent.mkdir(parents=True, exist_ok=True)
            frameinfo.save(outfile)


def resize_images(
        dstfn: str, width: int, height: int, quality: int=90,
        method: str='bilinear'
):
    ''' Resize a collection of images stored in a subfolder.

        Args:
            dstfn: Destination folder whose subfolders contain the
                images to rescale. Subfolder names are ignored.

            width: Target pixel width of each scaled image.

            height: Target pixel height of each scaled image.

            quality: JPEG compression factor for saved images. Default
                        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + data)
