#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng):
    # Header of 6 bytes
    out.write(struct.pack('>ccBBB', b'G', b'I', b'F', b'\x00', b'\x21', b'\xF9'))

    # Logical screen descriptor of 7 bytes
    out.write(struct.pack('>HHBBBBBHBH', rng.randint(10, 40),
                          rng.randint(10, 40),
                          8,  # Color resolution (set to 8 because that's what imagemagick does)
                          7,  # Sort of color table size? (actually set to this value + 1)
                          0x00,  # Background index (randomly chosen)
                          0xFF,  # Pixel aspect ratio
                          0, 0,  # Image width and height (determined later)
                          0))  # No global color table defined for now

    # Loop control block of two bytes
    out.write(struct.pack('>Bbb', 0b00001100, rng.randint(-5, -1), rng.randint(1, 5)))
    
    # Separator byte
    out.write(b';')

    # Other optional parts of header
    frame_count = rng.randint(100, 1000);
    palette_size = rng.randint(2, 20)
    local_color_table = [rng.randint(0, 255) for _ in range(palette_size * 3)]
    out.write(bytes([local_color_table[i // 3]
                     if i % 3 == 0 else
                     0 for i in range(((palette_size * 3) + 1) * frame_count)]))

    # Application extension identifying which is the Netscape application, followed by 
    # its block with some useless data
    netscape_app_id = b'\x0bNetscape\xae\x04\x01'
    out.write((len(netscape_app_id) + len(b'\xff\x0b') +
               len(b'\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')))
    out.write(b'\x21\xff\x0b')
    out.write(netscape_app_id)
    out.write(b'\x00\x00\x00\t\x00image-index\x00\x00\x00\x00')

    # Graphics control block of 8 bytes
    offset = ((palette_size * 3) + 1) * frame_count;
    out.write(struct.pack('>BBBBBBB', b'\x21', b'\xf9', b'\x04', b'\x05',
                          rng.randint(0, 255),
                          frame_count & 0xFF,
                          (frame_count >> 8) & 0xFF,
                          offset & 0xFF))
    delay = rng.randint(10, 30)
    duration = rng.randint(delay, min(100, delay*10 + 10));
    out.write(struct.pack('>BB', duration, delay))

    # Image descriptor of 9 bytes
    out.write(struct.pack('>BBBBBBBHBH',
                          int(bool(rng.getrandbits(1))),   # Local color table flag
                          rng.randint(1, 7),              # LZW minimum code size
                          rng.randint(0, 255),            # Left position
                          rng.randint(0, 255),            # Top position
                          rng.randint(1, max(10, 255)),   # Width
                          rng.randint(1, max(10, 255)),   # Height
                          0x08,                           # Encoding type (LZW compression)
                          0,                              # No interlacing
                          len(b'\x00\x00\x00\x00')))       # No local color table

    # Data
    lzwmincodesize = rng.randint(1, 8)
    out.write(bytes([(lzwmincodesize << 5 | lzwmincodesize)])
              + b'\x00\x00\x00')         # End of stream marker

    # Write the actual pixel data
    def get_pixel() -> str:
        if bool(rng.getrandbits(1)):
            return '0'

        if bool(rng.getrandbits(1)):
            return 'F'

        pixels = ['A','B','C','D','E']
        return rng.choice(pixels)

    for y in reversed(range(rng.randint(10, 40))):
        line = []
        
        while len(line) < rng.randint(25, 40):
            line += get_pixel()

        while len(line) > max(rng.randint(2, 10),
                              rng.randint(int(.3*len(line)),
                                          int(.7*len(line)))):
            del line[-1]

        out.write((''.join(line)).encode())

    # Markers
    for _ in range(rng.randint(0, 2)):
        out.write(b'\x3b')

    # Terminator (;)
    out.write(b';')

