#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):

    # 1. Header
    size = rng.read(4) + b'\x00\x21' # Always background color index present (the first byte) and loops is set to zero
    out.write(size)

    # Get next bytes as type
    gif_type = rng.read(1)

    while gif_type!= b'e':

        if gif_type == b'!':
            # Extension block

            # Get size of extension block
            ext_size = rng.read(1)
            
            # Get label of extension
            ext_label = rng.read(1)

            print('Extension', ext_label.decode(), 'at offset', out.tell())

            # Dispatch if it is an application extension or not
            if ext_label in [b'A']:
                # Application extension

                app_block = ext_size + ext_label
                
                # Get sub-blocks until we have reached the end of this application extension block
                while not bytes([app_block[-1]]) == b';':
                    print('Sub-Block', app_block[:-1].decode(), 'at offset', out.tell())

                    if ext_label in {b'NETSCAPE2.0'}:
                        # NETSCAPE specifics

                        # Skip two bytes because those are some sort of version?
                        netscape_block = b'\x03\x01' + rand_number()
                        
                        # Write block data plus terminator
                        app_block += netscape_block + b';'
                        
                    elif ext_label in {b'XMP'}:
                        xmp_block = \
                            b'<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#' + \
                            b'xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:xapRights=' + \
                            b'"http://ns.adobe.com/xap/1.0/rights/" xml:lang="en-US">' + \
                                '<rdf:Description rdf:about="" dc:format="" xapRights:Marked="' + \
                                    str(rand_bit()).encode() + '" xapRights:UsageTerms="">' + \
                                    ''.join('<dcterms:' + field[0] + \
                                        '>this is a test</dcterms:' + field[0] + '>\n       'for field in fields) + \
                                    '</rdf:Description></rdf:RDF>'
                    
                        app_block += xmp_block + b';'
                    
                    else:
                        raise ValueError("Unrecognised application extension", ext_label.decode())

                    # Append size of this sub-block
                    app_block = struct.pack(">B", len(netscape_block)) + netscape_block

                # Now write this block to output
                out.write(app_block)

            else:
                length = int.from_bytes(ext_size, "big") - 1
                data   = rng.read(length) + b';'
                out.write(ext_size + ext_label + data)
                
        elif gif_type in {b','}: # Image separator indicator
            # Image descriptor

            left       = rand_number() * 256 + rand_number()
            top        = rand_number() * 256 + rand_number()
            width      = rand_number() * 256 + rand_number()
            height     = rand_number() * 256 + rand_number()
            flags      = bytes([(0b1 << i) & rand_number() for i in range(8)])
            local_color_table_flag = (flags[7]) >> 7
            interlace_flag         = ((flags[7]) & (0b1<<6)) >> 6
            sorted_color_table_flag = ((flags[7]) & (0b1<<5)) >> 5

            global_color_table_size    = rand_number() % 0b111 + 1
            background_color_index     = rand_number()

            pixel_aspect_ratio_numerator = rand_number() % 0b1111 + 1
            pixel_aspect_ratio_denominator = rand_number() % 0b1111 + 1

            image_descriptor = \
                struct.pack(">HHBBBBBBBHBB",
                             left, top, width, height, flags[0], flags[1],
                             background_color_index, pixel_aspect_ratio_numerator,
                             pixel_aspect_ratio_denominator, global_color_table_size)

            out.write(image_descriptor)

            # Color table
            if local_color_table_flag:
                palette = b''.join((struct.pack(">BBBB", r, g, b, 0xff) for r,g,b in [(
                    rand_number()*0xff//255, 
                    rand_number()*0xff//255, 
                    rand_number()*0xff//255) for _ in range(global_color_table_size)]))

                out.write(palette)


            # If interlace flag is set, read data line by line
            if interlace_flag:
                pass

            # Else read all pixels at once
            else:
                lzwcodelen  = rand_number() % 0b111 + 1
                codes       = []
                old_pixel   = None
                new_pixels  = []
                repeat      = False
                run_length  = 0
                code_buffer = 0

                while True:

                    # TODO: Handle non-zero LZW minimum code size
                    if len(codes) < 4096 and rand_bit():
                        code_buffer = (code_buffer << 1) | rand_bit()
                    else:
                        decoded_pixel = codes[code_buffer]
                        run_length -= 1

                        if run_length > 0:
                            new_pixels += [decoded_pixel]*run_length
                        else:

                            if decoded_pixel == 0x00:
                                break
                            
                            if decoded_pixel == 0xFF:

                                if old_pixel!= None:
                                    new_pixels += [old_pixel]*3
                            else:
                                new_pixels.append(decoded_pixel)

                            old_pixel = decoded_pixel

                    if len(new_pixels) >= width*height:
                        break

                # Actually write the data to output
                encoded_data = encode_lzw(new_pixels, lzwcodelen)
                out.write(encoded_data)

        else:
            assert False, f"Unknown block type encountered ({gif_type})"
            
        # Move on to next block type
        gif_type = rng.read(1)
            


def encode_lzw(input : list, start_code_len=None):
    """Encodes input according to Lempel-Ziv-Welch compression algorithm."""
    
    codes                 = []
    cur_code              = ""
    output                = b""
    min_code_len          = None if start_code_len==None else 8
    max_code_len          = 12

    # Find maximum number of bits required to represent current code
    def get_code_len(code):
        
        this_code_len = 0

        while code > 0:
            bit = (code&1)
            codelen_for_bit = 0
            
            # Encode if smaller than smallest known size
            if min_code_len!=None and min_code_len <= this_code_len+1:
                return min_code_len

            # Or larger than largest known size
            if max_code_len<=(this_code_len+1):
                this_code_len = 0
                continue
                
            codelen_for_bit = 1
            code >>= 1
            this_code_len += 1
            
            # Add extra bit if necessary to account for possibility that there was no match for what's remaining
            if (code>0):
                codelen_for_bit += 1

                if (code>0):
                    codelen_for_bit += 1

            if (code<=0) and (min_code_len == None or min_code_len>=this_code_len):
                this_code_len = codelen_for_bit
                min_code_len = codelen_for_bit if min_code_len==None else min(min_code_len, codelen_for_bit)

        return this_code_len
        
    # Generate starting codes from longest to shortest
    for i in reversed(range(max_code_len)):

        codelen      = get_code_len(i)
        cur_code     = "{:b}".format(i).zfill(codelen)[::-1]

        codes.insert(0,(cur_code, chr(i)))


    # Read input until exhausted
    while len(input)>0:
        
        c             = input.pop(0)
        found         = False

        # For each code...
        for i,(code,char) in enumerate(codes):

            #... check if one matches with the beginning part of our current character
            if char[:len(c)]==c:

                # We have a full match!
                if len(c)==len(code):

                    # Extend previous code with last bit and append our matched code instead
                    codes[i] = (code+"1", char)
                    output+=bytes([