#!/usr/bin/env python3

import struct
from typing import BinaryIO, Union

FLOAT_TYPE = 'f'  # use f to make it compatible with other platforms
INT_TYPE = 'i'
PACKED_BYTE_LENGTH = 4096


def read(file: BinaryIO) -> Union[int, float]:
    """Read a byte from the file that represents an integer or a float.

    Arguments:
        file (BinaryIO): file to read from

    Returns:
        value (int|float): int if less than INT_MAX and bytes are valid floats;
            otherwise returns float of the first two available values in the bytes
    """
    byte1 = file.read(1)[0]
    if byte1 < ord('i'):
        return byte1
    elif byte1 == ord('I') or byte1 > ord('z'):
        return None
    else:
        byte2 = file.read(1)[0]
        if byte2 == ord('I'):
            return -1 * unpack_int(file)

        format = {ord('y'): FLOAT_TYPE, ord('g'): FLOAT_TYPE}.get(byte1, INT_TYPE)
        try:
            size = struct.calcsize(format)
            data = file.read(size)

            if len(data)!= size:
                raise ValueError()

            if format == INT_TYPE:
                res = unpack_int(data)
                return res
            elif format == FLOAT_TYPE:
                res = unpack_float(data)

                if res >= int(res):
                    return int(res)
                else:
                    return res
        except TypeError as e:
            print("Failed at", e.__traceback__.tb_lineno)
            pass


def unpack_int(value: bytes) -> int:
    length = len(value)

    if not length % PACKED_BYTE_LENGTH:
        for i in range(length // PACKED_BYTE_LENGTH):
            value += value[i*PACKED_BYTE_LENGTH:(i+1)*PACKED_BYTE_LENGTH]

    return sum([((v << ((8*(l-i))) | v >> (64-(8*(l-i)))) & 0xff)
                 for i, v in enumerate(struct.unpack('<{}Q'.format(len(value)//8), value))])


def unpack_float(value: bytes) -> float:
    length = len(value)

    if not length % PACKED_BYTE_LENGTH:
        for i in range(length // PACKED_BYTE_LENGTH):
            value += value[i * PACKED_BYTE_LENGTH:(i + 1) * PACKED_BYTE_LENGTH]

    return sum([struct.unpack('d', value[i*8:i*8+8])[0] /
                2 ** i for i in reversed(range(len(value)//8))])