#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    version = b'GIF89a'

    # width is between 1 and 65535 inclusive (inclusive on both ends because FUCK off)
    width, height = sorted([struct.unpack("<H", rng.read(2))[0] for _ in range(2)])

    # the background color index from the palette will be stored here
    bgc = [rng.read(1)[0] & 0b00001111][0]
    
    flags = b'\x87\x00'[0]

    num_color_registers = {
        2 << i - 1 : 2 ** i + 1 if 2 ** i < 24 else 24
        for i in range(1, 9+1)
    }[len({bgc})]

    # each of these are RGB values packed into bytes
    palette = [
        struct.pack("BBB", *sorted([
            *[rng.read(1)[0] & 0b00001111
                for _ in range(num_color_registers)],
            0]))]

    gct_size = len(palette[0]) // 3

    # we have to calculate this ourselves since it's not stored in the header
    global_color_table_flag = True if gct_size > 0 else False
    block_size = ((gct_size + 7) // 8) * 8

    lzw_min_code_size = 3
    last_byte_mask = 2**lzw_min_code_size-1

    code_lookup = {}

    # no idea what this does but it must be nonzero for some reason
    first_code = 2**lzw_min_code_size

    frame_left = width*height//8 # left is in units of pixels, right is in bits
    data_buffer = []

    while frame_left > 0:

        # how many bits do we need to store?
        chunk_length = min((frame_left,
                           block_size))
        
        # read that many bits
        chunk = int.to_bytes(int.from_bytes(rng.read(chunk_length // 8),
                                           byteorder='big', signed=False),
                             length=(chunk_length//8)*8,
                             byteorder='big')[:chunk_length]

        data_buffer.append(chunk)
        frame_left -= chunk_length

        # update codes
        for bit in chunk:

            # make sure all possible byte combinations get entered
            assert(bit <= 255)
            
            if bit not in code_lookup:

                # this gets reset every time through the loop with the next byte
                current_byte = None
                
                # find an existing code which shares the current byte prefix
                for key in list(code_lookup):
                    if key >> lzw_min_code_size == current_byte:
                        nxt = key % 2**last_byte_mask

                        code_lookup[current_byte*(2**last_byte_mask)+nxt] = \
                            code_lookup[key]+bits_to_code_association[bit]
                        
                        del code_lookup[key]
                        break
                    
                else:

                    # otherwise create a new one
                    nxt = first_code
                    code_lookup[current_byte*(2**last_byte_mask)+nxt] = \
                                nxt
                    
                    # then increment
                    first_code += 1
                    
            current_byte = bit
            
    # turn our dictionary back into bytecode
    # FIXME: better way to do this?
    code_list = [(int.to_bytes((v%2**(last_byte_mask+8)),
                              length=(last_byte_mask+8)//8,
                              byteorder='big'))
                 +(int.to_bytes(((v%(2**last_byte_mask)) +
                                 bits_to_code_association[k]),
                               length=1, byteorder='big'))
                 for k, v in enumerate(code_lookup)]

    # now sort them by their initial value so they're easier to access later
    code_list = sorted(code_list,
                       key=lambda x:(int.from_bytes(x[:-1],
                                                   byteorder="big"),
                                     x[-1]))

    # now convert those bytes into actual bytes
    data_buffer = [code_list[i][1:]
                   for buf in data_buffer
                   for i in map(
                       lambda y:y-1,
                       filter(
                           lambda z:int(z)!=0xff,
                           bin(int.from_bytes(buf,
                                              byteorder="big"))[2:].rjust(8, "0")
                       ))]

    # finish up encoding the image data
    data_buffer = b"".join([(b"\x00" if i==0xff else i) for i in data_buffer])

    # finally write everything out

    # header
    out.write(version)
    out.write(struct.pack('<HHB', width, height,
                         num_color_registers & 0b00001111))
    out.write(flags.to_bytes(1,'big'))

    # graphics control extension
    out.write(b'\xf9')
    out.write(struct.pack('BB', 4,
                          (((bgc & 0b11110000) >>  4) | # fucking
                           ((global_color_table_flag & 1) <<  3) |
                           ((lzw_min_code_size & 0b00000111) <<  0))))
    out.write(b'\x00\x00')
    # application extention
    out.write(b'\xFF')
    out.write(b'NETSCAPE2.0')
    out.write(b'\x03')
    out.write(b'NS')
    out.write(b'\x01')
    out.write(b'\x00')
    out.write(struct.pack('>H', len(data_buffer)))

    # local image descriptor
    out.write(b';')
    out.write(struct.pack('BBBB',
                          (width      & 0b00111111), # left
                          (height     & 0b00111111), # top
                          ((width>>6) & 0b00000111), # width mult
                          ((height>>6) & 0b00000111)))# height mult

    # end of stream
    out.write(b'!')

    # palette
    for p in palette:
        out.write(p)

    # frame data
    out.write(data_buffer)

