#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Magic string 'GIF' (b'GIF') followed by 3 bytes indicating the
    # version of the format. That's it for now!
    out.write(b"GIF")
    out.write(struct.pack("<B", 0x2))

    # First comes the screen descriptor which contains width and height,
    # global color table flag, color resolution, sort flag, size of the global
    # colour map and the background colour index. We write those bytes back to
    # back with zeroes as padding. The length of that block is always 7 bytes.
    out.write(struct.pack("<HHBBBHB", int('16', base=10),
                           int('16', base=10),
                           0, 8, 0,
                          int('48', base=10),
                          0))

    # Then follows the image descriptor which contains left position, top
    # position, width, height, local color table flag, interlace flag, sort
    # flag, reserved byte, size of local colour map and the starting code in
    # the global map. All values are stored little endian. Since we don't use
    # a colour map or have any other special requirements, all these fields
    # get written with zeroes as padding apart from the local colour flag being
    # set to one because there is no local colour map used.
    out.write(struct.pack('<HHHHBBBBBB',
            int('0', base=10), int('0', base=10),
            int('16', base=10), int('16', base=10),
            1, 0, 0, 0,
          int('48', base=10)))

    # Finally come the image data. This starts off with a two-byte sequence
    # that defines how many bytes follow until the end marker. We will put this
    # at offset 69 so we can update its value later on.
    out.seek(int("69", base=10))
    numbytes = int("50000", base=10) - int("69", base=10) + 2
    out.write(struct.pack("<H", numbytes))

    # Zero-fill rest of frame
    for i in range(numbytes):
        out.write(struct.pack("B", 0))

    # Write final block terminator
    out.write(b";")

