#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    gif_header = b'GIF89a'
    # 24-bit color palette has 768 bytes (3 * 256). There's no "correct" way to
    # determine how many bits are necessary for the color table length. This is
    # just an arbitrary choice that should give us some reasonable values. We can
    # always change it later if we need more colors :)
    max_color_table_size = int((2**10 - len(gif_header)) / 3)
    num_color_tables = min(max_color_table_size // 256 + 1, 255)

    # The smallest possible frame duration in ms is 16ms (__uint16), since the unit
    # is measured in 1/100s of a second, so this is 16*100 = 1600. I've picked a
    # slightly higher value here because some readers seem to have trouble with very
    # short durations and only display one frame then exit. Also Windows Explorer
    # doesn't play gifs as fast as they should be :(
    min_frame_duration = 1600
    # Same thing about the maximum frame duration in ms. A duration of 99.99ms makes
    # the viewer freeze when displaying the last frame (~254 steps)
    max_frame_duration = 9999

    # Width & height of each frame
    width = 100
    height = 100


    # Write Gif Header
    header_bytes = bytearray()
    header_bytes += gif_header
    header_bytes.extend([
        87 & 0xff,    # Logical screen width
        0x00,          # Logical screen width high bit
        87 & 0xff,    # Logical screen height
        0x00,          # Logical screen height high bit
        0b100000 |     # Global Color Table flag : 1=global; 0=per-frame
            ((num_color_tables - 1) << 3),   # Number of global color tables
            0x00,                          # Background color index (None)
            0x00                           # Pixel aspect ratio (pixel square by default)
    ])

    # Write color tables
    for i in range(num_color_tables): 
        # Total number of colors in this table
        num_colors = rng.read(1)[0] % 128

        # Write colors from the color list, counting downwards
        curr_idx = num_colors - 1
        while not curr_idx < 0: 
            r = rng.read(1)[0];
            g = rng.read(1)[0];
            b = rng.read(1)[0];

            # Each color entry takes up three bytes in the output buffer
            header_bytes.extend([r,g,b]);
            
            curr_idx -= 1
    
    # Write frames
    for _ in range(100): 

        # Frame separator
        out.write(struct.pack('B', 0x21))
        out.write(struct.pack('B', 0xf9))
        out.write(struct.pack('H', rng.read(2)[0]))
        
        # Image descriptor block follows immediately after
        out.write(struct.pack('<HHBB',
            0b11100000 |      # Local Color Table Flag : 1=local
                ((num_color_tables - 1) << 3),       # Number of local color tables
            0,0,               # Left position
            0,0,               # Top position
            width,height       # Image dimensions
        ))

        # Image data starts at byte [offset=7], end at EOF or optional trailer bytes
        image_data_start = out.tell()
        skip_trailer_bytes = True
        while out.seekable(): 
            try: 
                # Try to write a pixel
                out.write(struct.pack('BBB', 
                    rng.read(1)[0],
                    rng.read(1)[0],
                    rng.read(1)[0]
                ));
            except IOError: 
                break;
            
        # Reset stream pointer back to where it was before writing pixels
        out.seek(image_data_start);

        # Skip any trailing data added by the generators themselves
        if skip_trailer_bytes: 
            out.seek(len(gif_header)+0x2c, io.SEEK_CUR)

        # Write the duration of this frame
        out.write(struct.pack('<H', rng.read(2)[0]));