#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG object, which must have method `.read(n)`,
# and a length of at least 12 bytes (including
# color table). The output will be written in multiple
# chunks because it's easier than iterating over all
# possible inputs to generate the entire header. You'll
# get an IndexError if you don't supply enough data for
# other fields; I'm not trying to make this super robust
def gen_random_gif(rng: BinaryIO, out: BinaryIO):

    # first write 6-byte signature
    out.write(b'GIF89a')

    # then write single byte version field
    out.write(struct.pack('B', 4))

    # now do the next two fields individually
    out.write(struct.pack('<HH', *rng.read(4)))

    # now we need to write the LZW minimum code size value
    lzw_mincodesize = rng.read(1)[0] % 128
    out.write(lzw_mincodesize.to_bytes(1, 'little'))

    # 0 indicates no global color table, so we don't write anything else here
    out.write(b'\x00\x00')

    # now finally move onto the image block
    out.write(b';')

    # We need to write up to three blocks before we can start writing the pixels
    for _ in range(0, 3):
        blocktype = ord(rng.read(1)) & 0x7

        # write control characters as they are
        if   blocktype == 0x0: b = b'\x00'     ; cls = None
        elif blocktype == 0x2: b = b'\x21\xF9\x04'
        elif blocktype == 0x3: b = b'\x2C'     ; cls = None
        elif blocktype == 0x4: b = b'\x21\xFF\x0B'
        elif blocktype == 0x5: b = b'\x3B'
        else: raise Exception('Unknown graphic control block type!')
        
        assert len(b) > 0, f'Found empty control character sequence!'

        out.write(b)

    # Write the palette update subblock if necessary
    if blocktype == 0x4:
        out.write(b'\x00')
        palette = rng.read(3 * 256)
        out.write(palette)

    # Now write the image descriptor
    left = int.from_bytes(rng.read(2), byteorder='little')
    top = int.from_bytes(rng.read(2), byteorder='little')
    width = int.from_bytes(rng.read(2), byteorder='little')
    height = int.from_bytes(rng.read(2), byteorder='little')
    
    packedflags = rng.read(1)[0]
    haslocalcolortable = bool((packedflags >> 7) & 0x1)
    interlaceflag = bool((packedflags >> 6) & 0x1)
    sortflag = bool((packedflags >> 5) & 0x1)
    reservedbits = (packedflags >> 3) & 0x3
    localcolortablesize = (packedflags & 0x7) + 1

    out.write(left.to_bytes(2, 'little'))
    out.write(top.to_bytes(2, 'little'))
    out.write(width.to_bytes(2, 'little'))
    out.write(height.to_bytes(2, 'little'))

    flags = 0
    if haslocalcolortable:
        flags |= 0x80
    if interlaceflag:
        flags |= 0x40
    if sortflag:
        flags |= 0x20
    flags |= ((reservedbits << 3) | localcolortablesize)
    out.write(flags.to_bytes(1, 'little'))


    # And finally write the actual pixel data
    while True:
        hdr = rng.read(1)[0]
        ctrlbit = (hdr >> 7) & 0x1
        extbit = (hdr >> 6) & 0x1
        imgdatabit = (hdr >> 5) & 0x1

        if ctrlbit: continue
        if extbit or imgdatabit: break

        chunklen = hdr & 0x1F
        out.write(rng.read(chunklen))

    out.write(b'\x00\x3B')

    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + data)
