#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and returns its size in bytes.
def generate(out: BinaryIO, rng):
    # Header; see https://www.w3.org/Graphics/GIF/spec-gif89a.txt for more info

    out.write("GIF89a".encode())  # Magic Number
    header = bytearray()
    header += b"\x02"    # Version
    header += b"\x00"    # Width Lo Byte (0)
    header += b"\x00"    # Width Hi Byte (0)
    header += b"\x01"    # Height Lo Byte (1)
    header += b"\x00"    # Height Hi Byte (0)
    header[4] |= 0b0111   # Color Resolution | GCT Flag | Sort Flag
    header += b"\x00"    # Global Color Table Size (0 means no global table exists)
    header += b"\x00"    # Background Color Index
    header += b"\x00"    # Pixel Aspect Ratio
    out.write(header)

    # Logical Screen Descriptor
    lsd = bytearray()
    lsd += b"\x00"     # Screen Width Lo Byte
    lsd += b"\x00"     # Screen Width Hi Byte
    lsd += b"\x01"     # Screen Height Lo Byte
    lsd += b"\x00"     # Screen Height Hi Byte
    lsd += b"\x00"     # Packed Field
    lsd[5] |= 0b01100000       # Global Color Table Flag
    lsd[5] |= ((rng & 0b1) << 6)      # Color Resolution
    lsd[7] |= ((rng & 0b1111) << 3)        # Sort Flag
    lsd[7] |= ((rng & 0b111) << 5)         # Size of Global Color Table
    out.write(lsd)

    if ((rng & 0b1)):
        print("Warning! Should be using color table")

    # Palette
    palette_entries = pow(2, int((lsd[5] >> 3)))
    for i in range(palette_entries):
        r = min(max((i & 0b111), 255 - (((i >> 5) * 7) % 255)), 0)
        g = min(max(((i >> 3) & 0b111), 255 - (((i >> 5) * 7 + 1) % 255)), 0)
        b = min(max(((i >> 1) & 0b111), 255 - (((i >> 5) * 7 + 2) % 255)), 0)

        out.write(struct.pack("<BBB", r,g,b))

    # Raw image data
    imgdata = []
    for y in range(int(lsd[3])):
        row = [[] for x in range(int(lsd[2]))]
        for x in range(int(lsd[0])):
            row[x].append(generator.randrange(256))
        for line in row:
            imgdata.extend([line])

    # LZW minimum code size is set to 2. See section "Image Data" at
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt for more details.
    out.write(b'\x02')

    last_bytes = None
    codesize=2
    subblock_data = []
    max_codepoint = len(imgdata)**2*len(imgdata[0][0])**2
    while True:
        new_bytes = get_new_bytes(last_bytes)
        try:
            subblock, extra = compress(new_bytes, codesize)
        except BufferError as e:
            # If we run out of buffer space before all bytes have been encoded,
            # add another code size bit so that there's enough room for the next
            # code length to fit without overrunning the current sub-block.
            codesize+=1

            continue

        subblock_data.append(subblock)

        if extra!= '':
            raise ValueError("Unexpected characters left over in compression!")
        
        if len(bytearray().join(subblock_data)) >= (1<<codesize)*max_codepoint/(codesize+1):
            break
    
    write_subblocks(subblock_data)

    return out.tell()

