#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        width  : int = 640,
        height : int = 480,
        repeat : bool = True,
        duration : int = -1, # in centiseconds (hundredths of a second), or -1 for forever
        out    : BinaryIO = None):

    if not out:
        from tempfile import TemporaryFile
        out = TemporaryFile('wb')
    else:
        # Ensure that we are writing to an actual file object since it is required by
        # `struct` below when packing integers
        assert hasattr(out,'seek'), 'Argument "out" must be a file-like object'
        assert callable(getattr(out,'seek')), 'Argument "out" must be a file-like object'


    def putchunk(data):
        out.write(bytes([len(data)]))
        out.write(data)
        out.write(b'\x00')

    header, palette = gen_gif_colors()

    putchunk(palette)
    putchunk(gen_gif_imagedata(width=width,height=height))
    putchunk(gen_gif_ext(repeat=repeat,duration=duration))


def gen_gif_colors():
    # TODO: Generate more than just the color table
    return b',\x02X\x9c\x01\x00\xe5\xff\x00\x00\x00\x00\xca\xcf\xc7\xf6\xf7\xff\xfb\xb6\xaf\xfc\x8d\x80\xfd\xeb\xa6\xb3\xfe\xae\x8e\xff\xbb\xac\xff\xd1\xba\xff\xee\xcb\xff\xed\xd7\xff\xe7\xe2\xff\xd7\xe1\xff\xbf\xde\xff\xaa\xdc\xff\xa2\xcc\xff\xa0\xbd\xff\xa1\xa9\xff\xa4\x96\xff\xab\x93\xff\xbc\x97\xff\xda\xa6\xff\xef\xb5\xff\xea\xbe\xff\xdd\xcd\xff\xce\xdf\xff\xc2\xed\xff\xb9\xf7\xff\xb7\xfd\xff\xb9\xf6\xff\xc0\xf5\xff\xc6\xfa\xff\xca\xf9\xff\xcf\xf6\xff\xd5\xec\xff\xdc\xe4\xff\xe2\xe0\xff\xe8\xdb\xff\xec\xd8\xff\xef\xd3\xff\xf1\xd2\xff\xf3\xd2\xff\xf5\xd3\xff\xf4\xd6\xff\xf4\xd9\xff\xf2\xdc\xff\xf0\xde\xff\xe8\xe0\xff\xe2\xe2\xff\xe0\xe4\xff\xdf\xe5\xff\xe0\xe6\xff\xe1\xe7\xff\xe2\xe7\xff\xe3\xe8\xff\xe3\xe7\xff\xe4\xe5\xff\xe5\xe3\xff\xe6\xe1\xff\xe6\xe0\xff\xe7\xdf\xff\xe8\xde\xff\xe9\xdd\xff\xea\xdd\xff\xeb\xde\xff\xeb\xe0\xff\xea\xe0\xff\xea\xe1\xff\xe9\xe2\xff\xe8\xe3\xff\xe6\xe3\xff\xe4\xe2\xff\xe6\xe1\xff\xe7\xe1\xff\xe8\xe2\xff\xe9\xe3\xff\xeb\xe3\xff\xeb\xe4\xff\xeb\xe7\xff\xe8\xe9\xff\xe5\xeb\xff\xe4\xed\xff\xe1\xef\xff\xdf\xf1\xff\xdd\xf3\xff\xde\xf5\xff\xdf\xf6\xff\xe0\xf7\xff\xe1\xf8\xff\xe2\xf9\xff\xe2\xf9\xff\xe3\xf9\xff\xe3\xf9\xff\xe3\xf9\xff\xe3\xf9\xff\xe2\xf9\xff\xe2\xf8\xff\xe1\xf7\xff\xdf\xf6\xff\xdd\xf6\xff\xdc\xf5\xff\xdb\xf4\xff\xda\xf3\xff\xd9\xf3\xff\xd8\xf2\xff\xd8\xf1\xff\xd8\xf1\xff\xd8\xf1\xff\xd8\xf1\xff\xd8\xf1\xff\xd8\xf1\xff\xd8\xf1\xff\xd8\xf1\xff\xd8\xf1\xff\xd9\xf2\xff\xdb\xf4\xff\xdc\xf5\xff\xdd\xf6\xff\xdd\xf6\xff\xde\xf7\xff\xde\xf7\xff\xdf\xf8\xff\xdf\xf8\xff\xdf\xf8\xff\xdf\xf8\xff\xdf\xf8\xff\xdf\xf8\xff\xdf\xf8\xff\xdf\xf8\xff\xe0\xf8\xff\xe0\xf8\xff\xe0\xf8\xff\xe0\xf8\xff\xe1\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe2\xf8\xff\xe1\xf7\xff\xe0\xf6\xff\xe0\xf5\xff\xdf\xf3\xff\xdd\xf2\xff\xdb\xf0\xff\xda\xee\xff\xd8\xec\xff\xd7\xea\xff\xd5\xe9\xff\xd4\xe8\xff\xd3\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd2\xe8\xff\xd3\xe9\xff\xd4\xea\xff\xd5\xeb\xff\xd6\xed\xff\xd8\xef\xff\xda\xf1\xff\xdb\xf3\xff\xdd\xf5\xff\xde\xf7\xff\xdf\xf8\xff\xe0\xf8\xff\xe0\xf8\xff\xe0\xf8\xff\xe0\xf8\xff\xe0\xf8\xff\xe0\xf8\xff\xe0\xf8\xff\xe0\xf8\xff\    out.write(header + data)
