#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    # 256 color palette
    colors = [struct.pack('<BBB', *rng.read(3)) for _ in range(256)]

    width = int.from_bytes(rng.read(1), 'little') + 10
    height = int.from_bytes(rng.read(1), 'little') + 10

    local_color_table = True if rng.read(1)[0] & 0b1 else False

    background = int.from_bytes(rng.read(1), 'little') % len(colors)
    aspect_ratio = int.from_bytes(rng.read(1), 'little') % 4

    flags = '\x87\x1F'.encode('ascii')

    global_color_table = []
    if local_color_table:
        size = (1 << ((int.from_bytes(rng.read(1), 'little') & 0b1110) + 1))
        # Generate from our table of 256 colors if we're supposed to use one
        global_color_table = colors[:size]

        # Pad it with white
        global_color_table += [(255,255,255)] * \
                max((len(global_color_table) - 256), 0)


    # Extension blocks are stored as tuples of byte lengths and binary data
    extensions = []
    if not local_color_table and not rng.read(1):
        extensions.append((
            1,
            bytes([
                0xF9,            # Graphic Control Block extension block type
                4,               # Data length
                0x01,            # Flags
                0xff              # Delay time
            ])
        ))

    # Generate image descriptor
    extensions += [
        (
            9,
            bytes([
                0x2C,        # Image Descriptor extension block type
                0x00 |     # Left position
                0x00 |     # Top position
                0x00 |     # Width flag
                0x00 |     # Height flag
                0b1,         # Local Color Table Flag
                0b1,         # Interlace flag
                aspect_ratio,# Reserved bit
                0,           # Sort flag
                width % 256,
                width // 256,
                height % 256,
                height // 256,
                0            # LZW minimum code size
            ])
        )
    ]

    if global_color_table:
        extensions[-1] = (
            extensions[-1][0],
            extensions[-1][1] + bytes([
                len(global_color_table)-1   # Size minus 1
            ])
        )

        extensions.append((
            len(global_color_table)*3+1,
            bytes([
                0xFF,                            # An Application Extension
                0x0B,                            # Length without Application ID
                0x00, 0x00,                      # Application ID
                ord('A'), ord('n'), ord('i'),   # "Animation" identifier
                ord('m'), ord('e'),             # version number
                0x01                             # The number of frames
            ]) + b''.join(global_color_table)
        ))

    # Get the initial state of the LZW compression algorithm
    initial_code_length = int.from_bytes(rng.read(1), 'little') % 8 + 1
    lzw_bits = f'{initial_code_length}{initial_code_length}'

    # Second code length is set to zero to indicate end of block
    lzw_bits += '\x00\x00'

    # State is the next byte after that
    first = {
        1: 0x01,
        2: 0x0100,
        3: 0x010000,
        4: 0x01000000,
        5: 0x0100000000,
        6: 0x010000000000,
        7: 0x01000000000000,
        8: 0x0100000000000000,
    }[initial_code_length]
    last_byte = None
    this_byte = first

    # We need to save these until all goes well so we can emit them at once
    final_data = [header]
    
    # We may have to pad here because LZW requires aligned byte sizes
    padding = 0
    if len(extensions)!= 0 or len(flags)!= 0:
        while padding < 3:
            padding += 1
            if len(extensions)!= 0:
                final_data.append(bytes([len(extensions)]))

            if len(flags)!= 0:
                final_data.append(flags)

        while padding!= 0:
            final_data.append(b'\x00')
            padding -= 1

    # Output the rest of the stuff
    final_data.extend([
        bytes([(width + padding)]) +b''.join(final_data[:-1]),
        bytes([height]) + b''.join(final_data[:-1]),
        *map(lambda x: bytes([len(x)]), global_color_table),
        flags,
        b'\x00',
        *(color for _, color in sorted(extensions)),
        *map(lambda x: bytes(x), extensions),
        *reversed([lzw_bits]),
        bytes([this_byte]),
    ])

    # Write out all the data together
    out.write(b''.join(final_data))

    # Now write out the compressed data
    bits_left = 8
    chunk = list()

    while True:
        # If there aren't enough bits left to encode another character, get more
        if bits_left == 0:
            chunk.clear()
            chunk.append(this_byte)
            
            # We want to keep trying to read until we find an unrepeated byte
            prev_byte = this_byte
            while prev_byte == this_byte: 
                this_byte = int.from_bytes(rng.read(1), 'little')

                # There's no reason why this shouldn't work but I'm too lazy to check
                assert len(chunk) <= 255
                
            bits_left = 8
        
        # Encode the byte into the buffer
        shift = bits_left-1
        chunk.insert(0, (((this_byte >> shift) ^ 0x80) % 255)+1)

        bits_left -= 1

        # If the buffer isn't full yet, continue
        if bits_left > 0: 
            continue

        # Emit the current chunk
        out.write(bytes(chunk))

