#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(b'\x01\x2C', byteorder='little')

    # TODO: add color info?

    # header
    width  = int.from_bytes(rng.read(2), byteorder='little') + 1
    height = int.from_bytes(rng.read(2), byteorder='little') + 1
    global_color_map_flag   = ord(rng.read(1)) & (1 << 7)!= 0
    local_color_map_flag    = ord(rng.read(1)) & (1 << 7)!= 0
    interlace_flag          = ord(rng.read(1)) & (1 << 6)!= 0
    sort_flag               = ord(rng.read(1)) & (1 << 5)!= 0
    size_of_local_color_map = int.from_bytes(rng.read(1), byteorder='little')
    background_color        = rng.read(1)[0]
    pixel_aspect_ratio      = int.from_bytes(rng.read(1), byteorder='little')
    
    # image descriptor
    left       = int.from_bytes(rng.read(2), byteorder='little')
    top        = int.from_bytes(rng.read(2), byteorder='little')
    datadesc_width  = int.from_bytes(rng.read(2), byteorder='little') + 1
    datadesc_height = int.from_bytes(rng.read(2), byteorder='little') + 1
    lct_flag         = ord(rng.read(1)) & (1 << 7)!= 0
    interlace_flag   = ord(rng.read(1)) & (1 << 6)!= 0
    sorted_flag      = ord(rng.read(1)) & (1 << 5)!= 0
    reserved         = rng.read(2)
    local_color_table_size = int.from_bytes(rng.read(1), byteorder='little')
    print('local_color_table_size:', local_color_table_size)
    block_terminator = rng.read(1)

    if local_color_table_size > 0:
        raise NotImplementedError()

    # Graphic Control Extension
    graphic_control_extension = b''
    if not global_color_map_flag and \
       not local_color_map_flag and \
       bool(rng.read(1) ^ 0):

        delay                = int.from_bytes(rng.read(2), byteorder='little')
        transparent_color_index = rng.read(1)[0]
        
        graphic_control_extension += b'\x21\xF9'
        graphic_control_extension += int.to_bytes((delay * 10 // 100), length=2, byteorder='little')
        graphic_control_extension += b'\x01'
        graphic_control_extension += bytes([transparent_color_index])

    else:
        graphics_control_label = rng.read(1)[0]
        while True:
            graphics_control_data = rng.read(1)[0]
            if graphics_control_data == 0x00 or len(graphic_control_extension) >= 40:
                break
            
            graphic_control_extension += bytes([graphics_control_data])


    # Application Extension
    application_extension = b''
    app_block_length_s     = str(int(rng.read(1)))
    app_block_length_pad   = 'f'*(3 - len(app_block_length_s))
    app_identifier         = rng.read(8).decode().rstrip('\0').encode()
    app_auth_code          = rng.read(3)
    app_content            = b''
    for _ in range(len(app_block_length_s)):
        if rng.read(1)[0]!= 0:
            raise Exception("Invalid application extension")
            
    # Comment Extension
    comment_extension = b''
    if rng.read(1)[0]!= 0xFF:
        raise Exception("Comment not terminated by terminator")
        
    # Plain Text Extension
    plain_text_extension = b''
    text_character_size  = rng.read(1)[0]
    text_background_color = rng.read(1)[0]
    text_foreground_color = rng.read(1)[0]
    grid_left   = int.from_bytes(rng.read(2), byteorder='little')
    grid_top    = int.from_bytes(rng.read(2), byteorder='little')
    grid_width  = int.from_bytes(rng.read(2), byteorder='little')
    grid_height = int.from_bytes(rng.read(2), byteorder='little')
    char_cell_width  = rng.read(1)[0]
    char_cell_height = rng.read(1)[0]
    text_left   = int.from_bytes(rng.read(2), byteorder='little')
    text_top    = int.from_bytes(rng.read(2), byteorder='little')
    text_width  = int.from_bytes(rng.read(2), byteorder='little')
    text_height = int.from_bytes(rng.read(2), byteorder='little')

    # Extension Introducer
    extensions = [b'',
                  graphic_control_extension,
                  application_extension,
                  comment_extension,
                  plain_text_extension][ord(rng.read(1))[0]]
    ext_introducer = b'\x21'
    ext_label      = rng.read(1)[0]
    ext_block_size = int.to_bytes((extensions[1:] | 0xFFFFFF00), length=2, byteorder='big')

    # Comment Extension
    comment_content = ''
    comment_data = []
    for i in range(ord(rng.read(1)[0])):
        if i % 2 == 0:
            comment_data.append(int.from_bytes(b'\0'+rng.read(2),'big')*255//65536)
            
        elif i % 2 == 1:
            comment_data.append(((rng.read(1)[0])*2**8)+comment_data[-1])

    for d in comment_data[:-1]:
        comment_content += chr(d)

    # Image Data
    pixel_data = []
    palette    = []
    for y in range(height-1,-1,-1):
        for x in range(width):

            pdata_byte = rng.read(1)[0]
            has_palette_entry = bool(pdata_byte & 0x80)
            has_lzw_compressed_data = bool(pdata_byte & 0x40)
            has_graphic_control_extension = bool(pdata_byte & 0x20)
            is_interlaced = bool(pdata_byte & 0x10)
            lzw_min_code_size = rng.read(1)[0]

            lzw_compressed_data = None
            if has_lzw_compressed_data:

                lzw_compressed_data = []
                last_c = ''
                
                buf = bytes([])
                val = 0
                while len(buf) < 12 or val!= 0x10:
                    if len(buf) >= 12 and val!= 0x10:
                        raise Exception("LZW compressed data too long!")

                    c = rng.read(1)[0]
                    v8 = bin(c >> 4)[2:].rjust(4,'0') + bin(c & 0xF)[2:].rjust(4,'0')
                    
                    if len(v8) <= 12:
                        tmp = bin((val << 8) | c)[2:].rjust(12,'0')
                        buf += bytes([(int(tmp[:8], 2)),
                                      (int(tmp[8:],  2))])
                        
                    else:
                        buf += bytes([(val)])
                    
                    
                    if val == 0:
                        code_size = ((lzw_min_code_size+1)<<max(0,len(last_c)-12))
                        last_c = chr(c)*code_size

                    eoi = False
                    if all(ch in b'\x00\xff' for ch in buf[-3:]):
                        eoi = True

                        for i in range(-3,0):
                            del buf[-i]
                                                
                        val = 0x10
                    
                    else:
                        val = buf[-1]
                    
                    last_c = last_c[1:]+chr(c)
                    
                    if eoi:
                        lzw_compressed_data.append(last_c)
                            
                    continue
                
            else:
                lzw_compressed_data = ['']

            if is_interlaced:
                raise NotImplementedError()
            
            
            if has_graphic_control_extension:
                pass


            pxl_idx = 0
            pxl_val = 0
            if has_palette_entry:
                pxl_idx |= (pdata_byte&0x07) << 8
                pxl_val = palette[pxl_idx