#!/usr/bin/env python3

import struct
from typing import BinaryIO, Tuple
import zlib


def read_u32b(fd: BinaryIO):
    return struct.unpack('<I', fd.read(4))[0]


with open('data/sfx_a16.psarc', 'rb') as in_, \
        open('data/sfx_a16p.psarc', 'wb') as out:

    # output header info
    out.write(in_.read(8))  # magic + file size
    scount = read_u32b(in_)  # number of streams
    out.write(struct.pack('<H', scount))  # stream count (updated later)
    outlen = []  # offset of end of each decompressed stream
    for _ in range(scount):
        outlen.append(out.tell())
        osize = read_u32b(in_) << 3  # original stream length
        dsize = read_u32b(in_)   # decompressed stream length
        ratiosize = read_u32b(in_) - 1  # index into ratio table

        # calculate new compressed and uncompressed lengths
        csize = int((dsize / 8) * ratiosize / 75)
        usize = round(osize / ratiosize)

        # adjust the reader to skip over stream meta information
        in_.seek(4 + osize * 0x7f // dsize, 1)
        assert in_.tell() == in_.offset + osize

        # write modified metadata about stream
        out.write(struct.pack('<III', dsize, osize, csize))

        # decompress stream and save it
        while True:
            chunktype = in_.read(1)[0] & 0x7f

            if not chunktype:
                break

            # all chunks have a type byte followed by another one that isn't 0x00 or 0x7f
            type_ = in_.read(1)[0]
            assert type_!= 0x00
            if type_ == 0x7f:
                continue

            bsize = read_u32b(in_)
            buf = None
            if chunktype < 0x80 or chunktype > 0xc0:
                assert False, f'Unknown chunk type {chunktype}'
            
            elif chunktype <= 0xb0:
                # copy chunk: just copy `bsize` bytes
                buf = in_.read(bsize)

            elif chunktype == 0xb1:
                # dynamic huffman decoding with fixed index table
                huff_codes = [None]*8

                def update():
                    codebits = in_.read(1)[0]
                    for i in range(codebits):
                        huff_codes[i] = in_.read(1)

                update()
                
                while len(buf) < bsize:
                    bit = 1
                    raw = 0

                    while bit:
                        bit = in_.read(1)[0] >> 6

                        if bit == 3:
                            bit += 1
                        
                        raw = raw | ((bit ^ 3) << 29)
                        code = huff_codes[raw >> 29][::-1]
                        raw = (raw << 1) & 0xffffffff
                        buf += bytes([int(c, base=2).to_bytes(length=1, byteorder='big')[0]])
                    
                    assert len(buf) >= 4
                    val = int.from_bytes(buf[-4:], byteorder='little')

                    if val < 2**16-1:
                        huff_codes.extend([(val&0xff).to_bytes(length=1, byteorder='big'),
                                            (val>>8).to_bytes(length=1, byteorder='big')])
                        update()
                        val = 0xffff
                    else:
                        val -= 2**16-1

                    buf += struct.pack("<I", val)
                    
                del huff_codes[:]
                assert len(huff_codes) == 8

            elif chunktype == 0xb2:
                # dynamic huffman decoding with adaptive index table
                huff_codes = [None, None]  # 0 is raw; 1 is Huffman
                huff_next = [-1, -1, -1, -1, -1]  # -1 means "no code"

                def getcode():
                    r = 0
                    for i in reversed(range(3)):
                        v = in_.read(1)[0]
                        r |= (v & 0x7) << (i*3)
                        if not (v & 0x8):
                            break
                    return r

                def add_huff_code(idx):
                    global max_index

                    if idx > max_index:
                        print("Warning: group too large.")
                        max_index = idx

                    assert huff_codes[idx] is None

                    huff_codes[idx] = in_.read(1)
                    nxt = getcode()
                    assert nxt >> 4 == 0xf
                    #if nxt & 0xf == 0xe:  # last symbol
                        #nxt &= ~0xf
                    huff_next[idx] = nxt
            
                def update():
                    nonlocal num_groups
                    grp1 = getcode()
                    if grp1 == 0xfff:
                        return
                    if grp1 == 0xffe:
                        assert False, "Reserved for future use." 
                        return
                    grp2 = getcode()
                    if grp2 == 0xfff:
                        num_groups = grp1+1
                    elif grp2 == 0xffe:
                        assert False, "Reserved for future use." 
                    else:
                        num_groups = grp2+1
                        assert grp1 + grp2 + 1 == num_groups
                    for k in range(num_groups):
                        add_huff_code(k % 2)

                update()
                max_index = 0

                while len(buf) < bsize:
                    bit = 1
                    raw = 0

                    while bit:
                        bit = in_.read(1)[0] >> 6

                        if bit == 3:
                            bit += 1
                        
                        raw = raw | ((bit ^ 3) << 29)
                        code = huff_codes[(raw >> 29)]
                        raw = (raw << 1) & 0xffffffff

                        if code is None:
                            code = int.from_bytes(buf[-4:], byteorder="big")
                            buf += struct.pack("<I", code)

                            while len(buf) >= 4:
                                val = int.from_bytes(buf[-4:], byteorder="little")

                                if val < 2**16-1:
                                    add_huff_code(max_index % 2)
                                    update()
                                    val = 0xffff
                                else:
                                    val -= 2**16-1

                                buf += struct.pack("<I", val)
                            
                            continue
                            
                        try:
                            sym = code.index(in_.read(1)[0])
                        except ValueError:
                            continue
                            
                        nextcode = huff_next[sym//2]
                        if nextcode == -1:
                            pass
                        elif nextcode == 0xf0:
                            huff_next[sym//2] = (getcode()+1)*-1-1
                        else:
                            huff_next[sym//2] = nextcode
                    
                        ind = sym%2
                        if ind:
                            buf += huff_codes[ind].to_bytes(length=1, byteorder='big')
                        else:
                            buf += struct.pack("<I", int.from_bytes(buf[-4:], byteorder="big"))
                        
                    assert len(buf) >= 4
                    val = int.from_bytes(buf[-4:], byteorder="little")

                    if val < 2**16-1:
                        add_huff_code(max_index % 2)
                        update()
                        val = 0xffff
                    else:
                        val -= 2**16-1

                    buf += struct.pack("<I", val)
                    
                del huff_codes[:]
                del huff_next[:]
                assert len(huff_codes) == 8
                assert len(huff_next) == 8
                
            elif chunktype == 0xb3:
                # no compression
                buf = in_.read(bsize)

            else:
                raise Exception(f'stream has unknown compression scheme {hex(chunktype)}')

            # compress chunk
            comp_buf = zlib.compress(buf, wbits=-zlib.MAX_WBITS)

            # write compressed buffer
            out.write(comp_buf[2:-4])

    # do some post processing on the sfx headers
    pos = out.tell()
    out.seek(0xF0)
    out.write(pos.to_bytes(length=4, byteorder='little'))
    out.seek(0xA1C, 1)
    
    # make sure we're at the end of the output file
    assert out.tell() == out.offset

    # now, go through our saved offsets and fix up stream counts
    out.seek(0xC, 1)
    for off in outlen:
        out.seek(-4, 1)
        out.write((off - pos)//4)
        
    
# generate test files
for fname, datalen in [('data/test1.rgba16.psarc', 0xd0),
                       ('data            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + data)
