#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). This is
# much more complicated than it needs to be because this uses
# GIF's lossless format which has some requirements (including
# having a color table with size 256 entries).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Generate the color table (this is just a mini-palette of 256 colors
    # that repeats every 10th entry from the actual palette in the image)
    color_table = bytes()
    for i in range(256):
        if i % 10 == 0:
            color_table += struct.pack('BBB',
                *struct.unpack('BBBB', rng.read(4)))

    # Write the file header and color table
    header = b'GIF89a'
    screen_descriptor = struct.pack('<hhHHbbbbbb',
        0, 0, len(color_table), 0b10000000, 768, 0, 0, 0, 255)
    trace_data = bytes([0x00])
    application_ext = struct.pack('Bbb',
        0xFF, 0x0B, ord('S'), ord('H'))
    comment_ext = struct.pack('BB%ds' % (len("Generated by generate_random_gif.py"),),
        0xFE, ord('c'), "Generated by generate_random_gif.py".encode())
    graphics_control_ext = struct.pack('BBBBBBBHB',
        0xF9, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0)
    plain_text_ext = struct.pack('BBbbb',
        0xFF, 11, 0b10001000, 0, len("Random GIF"))
    app_extension = application_ext + graphics_control_ext + \
                    comment_ext + plain_text_ext
    
    trailer = b'\x3B'

    out.write(header + screen_descriptor +
              color_table + trace_data + app_extension + trailer)

    # Now write the rest of the image data
    while True:
        blocktype = rng.read(1)[0]

        if blocktype!= 0x2C: break
        
        xpos = int.from_bytes(rng.read(2), 'little')
        ypos = int.from_bytes(rng.read(2), 'little')
        width = int.from_bytes(rng.read(2), 'little')
        height = int.from_bytes(rng.read(2), 'little')

        lctflag = rng.read(1)[0] & 0b1
        interlaced = rng.read(1)[0] & 0b1
        sorted = rng.read(1)[0] & 0b1
        reserved = rng.read(1)[0]
        lctsize = 2 << rng.read(1)[0]
        gctsize = ~(rng.read(1))

        local_color_table = bytes()
        for _ in range(lctsize):
            local_color_table += struct.pack('BBB', 
                *struct.unpack('BBBB', rng.read(4)))
        
        imgdat = bytes()
        while True:
            if not imgdat or imgdat[-1] == 0:
                imgdat += rng.read(width)
                continue

            chunk = rng.read(min(imgdat[-1], width - len(imgdat)))
            imgdat = imgdat[:-1] + chunk

            if len(chunk) < width - len(imgdat):
                imgdat += rng.read((width - len(imgdat)) - len(chunk))
                continue
            
            assert len(chunk) == width - len(imgdat)
            break
        
        lzwmincodesize = rng.read(1)[0] + 1
        # TODO: encode as LZW codestream instead of raw compressed data 
        out.write(struct.pack('<HHHHBBH%ds' % (len(local_color_table)),
            0x2C, xpos, ypos, width, height, lctflag | (interlaced << 7), 
            lctsize, lzwmincodesize, local_color_table + imgdat))
