#!/usr/bin/env python3

import sys
import struct


# Generate a random GIF with an optional PLAINTEXT extension block
# of up to 256 bytes in length that contains the string "Hello, World!"
def generate_random_gif(
        rng: BinaryIO,
        plaintext: bool=False,
        out: BinaryIO=sys.stdout.buffer):

    magicString = 'GIF89a'.encode('ascii')
    zeroByte = (0).to_bytes(1, byteorder='little', signed=True)
    oneByte = (1).to_bytes(1, byteorder='little', signed=True)
    header = [magicString]

    # Create header fields
    # - Logical Screen Descriptor
    widthBytes = rng.read(1)
    heightBytes = rng.read(1)
    logicalScreenDescriptor = []
    logicalScreenDescriptor += widthBytes
    logicalScreenDescriptor += heightBytes
    logicalScreenDescriptor += (zeroByte * 4)    # Color Table Specification Byte

    globalColorTableFlag = oneByte
    colorResolutionBitsPerPixel = zeroByte
    sortFlag = zeroByte
    globalColorTableSize = zeroByte   # In other words, ignore any palette given
    backgroundColorIndex = zeroByte
    pixelAspectRatio = zeroByte
    logicalScreenDescriptor += globalColorTableFlag
    logicalScreenDescriptor += colorResolutionBitsPerPixel
    logicalScreenDescriptor += sortFlag
    logicalScreenDescriptor += globalColorTableSize
    logicalScreenDescriptor += backgroundColorIndex
    logicalScreenDescriptor += pixelAspectRatio

    header += logicalScreenDescriptor

    # - Global Color Table Block
    if len([b for b in header[1] if not b == 0]) > 0:
        numColorEntries = pow(
            2, [b for b in header[1] if not b == 0][-1])

        colors = []
        for i in range(numColorEntries):
            red = rng.read(1)[0]
            green = rng.read(1)[0]
            blue = rng.read(1)[0]

            colors += [red, green, blue]

        globalColorTableBlock = []
        globalColorTableBlock += numColorEntries.to_bytes(
                1, byteorder='little', signed=False)
        globalColorTableBlock += b'\x00\x00'    # Null Background Index
        for color in colors:
            globalColorTableBlock += color.to_bytes(
                    1, byteorder='little', signed=False)

        header += globalColorTableBlock

    # - Graphics Control Extension
    graphicsControlExtension = []
    blockLabel = b'\x21\xF9\x04'              # Application Extension Label
    subblockLength = ((len(graphicsControlExtension)) +
                      len(oneByte)).to_bytes(
              2, byteorder='little', signed=False)
    applicationIdentifier = b'NETSCAPE2.0'.encode()
    applicationAuthCode = b'\x01\x00'         # No Authentication Code Needed
    subblockData = [applicationIdentifier]
    subblockData += applicationAuthCode       # NO AUTHENTICATION CODE NEEDED!
    graphicControlLabel = b'\x0A'             # Graphic Control Label
    blockData = []                            # Data will be added later...

    graphicsControlExtension += blockLabel
    graphicsControlExtension += subblockLength
    graphicsControlExtension += subblockData
    graphicsControlExtension += graphicControlLabel

    header += graphicsControlExtension

    # - Plain Text Extension
    if plaintext:
        print("plaintext enabled")
        plainTextExtension = []
        blockLabel = b'\x21\x01\x00'          # Plain Text Extension Label
        subblockLength = ((len(plainTextExtension))
                          ).to_bytes(2, byteorder='little', signed=False)
        textSubBlockSize = b'\x0c\x00'        # Size of Text Sub-Block
        backgroundRGBColor = b'\xff\xff\xff'  # White Background
        textForegroundRGBColor = b'\x00\x00\x00'     # Black Foreground
        textXCoordinate = b'\x00\x00'         # Left Aligned
        textYCoordinate = b'\x00\x00'         # Top Aligned
        characterCellWidth = b'\x04'          # Width of Character Cell is 4 px
        characterHeight = b'\x08'             # Height of Character Cell is 8 px
        textGridLeftPosition = b'\x00\x00'    # First Font Row
        textGridTopPosition = b'\x00\x00'     # First Font Column
        textSubBlockSize += backgroundRGBColor
        textSubBlockSize += textForegroundRGBColor
        textSubBlockSize += textXCoordinate
        textSubBlockSize += textYCoordinate
        textSubBlockSize += characterCellWidth
        textSubBlockSize += characterHeight
        textSubBlockSize += textGridLeftPosition
        textSubBlockSize += textGridTopPosition
        textSubBlockSize += b'\x07'           # The last 3 bits are reserved
        textSubBlockSize += b'\x00\x0f'       # Pixel aspect ratio
        plainTextHeader = []                   # Header Will Be Added Later...
        plainTextHeaderData = [textSubBlockSize]                    #...Here
        plainTextBox = []                      # Box Will Be Added Later....
        plainTextArea = []                     # Area Will Be Added Later...
        plainTextGraphicsControl = []          # GCW Will Be Added Later...

        plainTextExtension += blockLabel
        plainTextExtension += subblockLength
        plainTextExtension += plainTextHeaderData
        plainTextExtension += plainTextBox
        plainTextExtension += plainTextArea
        plainTextExtension += plainTextGraphicControl

        header += plainTextExtension


    # Create image descriptor
    imageDescriptor = []
    leftPos = b'\x00\x00'                    # Image left position at top
    topPos = b'\x00\x00'                     # Image top position on screen
    imgWidth = b'\x01'                       # 1px wide
    imgHeight = b'\x01'                      # 1px high
    localRGBAColorTableFlag = zeroByte      # Not Local Color Table Flag
    interlaceFlag = zeroByte                # Non-interlaced image
    sortedColorTableFlag = zeroByte         # No Color Table Sort Flag
    localColorTableSize = zeroByte          # Ignored
    localNumberLZWCodeSizeOverTwo = zeroByte   # 2^0 = 2 bytes per code
    imageDescriptor += leftPos
    imageDescriptor += topPos
    imageDescriptor += imgWidth
    imageDescriptor += imgHeight
    imageDescriptor += localRGBAColorTableFlag
    imageDescriptor += interlaceFlag
    imageDescriptor += sortedColorTableFlag
    imageDescriptor += localColorTableSize
    imageDescriptor += localNumberLZWCodeSizeOverTwo

    # Add final image data
    encodedImage = [data for data in
                    rng.read()]
    if len(encodedImage) % 2!= 0:
        encodedImage += [int(not encodedImage[-1])]
    print("Encoded image size:", len(encodedImage)/2)

    # Prepare image
    decodedImage = []
    currentByte = b''
    while True:
        try:
            currentByte += encodedImage.pop(0)
            currentByte += encodedImage.pop(0)
            decodedImage += [(struct.unpack('<H',
                                             currentByte))[0]]
            currentByte = b''
        except IndexError as ex:
            break

    # Now, prepare the compressed image data and the uncompressed image data.
    lzwMinCodeSize = 2
    lzwClearCode = pow(2, lzwMinCodeSize+1)-1
    lzwEndCode = lzwClearCode+1
    lzwFirstFreeCode = lzwMinCodeSize+2
    lzwMaxCodeWord = pow(2, lzwFirstFreeCode)+1-1

    # - Compressed Image Data

    string = ''
    for i in range(lzwClearCode):
        string += str((decodedImage[i]))+', '
    print("[", string[:-2], "]")

    compressedImageData = []
    # TODO: ^ Figure out how the hell this works exactly!
    #       This doesn't work at all right now. Just kidding, it's fine.
    #       I got this idea from here: https://git.io/JfrRR
    string = ""
    for val in decodedImage[:]:
        string += '{0:0{1}b}'.format(val, 16)
    compressedImageData += list(string)

    # - Uncompressed Image Data
    uncompressedImageData = [[], []]
    for datum in decodedImage:
        uncompressedImageData[0].append(pow(
                datum // 256, 2**8))
        uncompressedImageData[1].append(-datum % 256)

    header += imageDescriptor

    # - Image Data
    dataSubblock = b'\x2C'            # Data Sub-block ID
    segmentLength = b'\xFF\x0B'       # Length of image data, excluding
                                    # image data sub-block
    imageData = [segmentLength]
    for datum in uncompressedImageData[1]:
        imageData += datum.to_bytes(1, byteorder='little', signed=True)
    for datum in uncompressedImageData[0]:
        imageData += datum.to_bytes(1, byteorder='little', signed=True)

    # Add the completed blocks to the main header
    header += [dataSubblock] + imageData

    if len