#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO


def get_generator() -> BinaryIO:
    return open('/dev/urandom', 'rb')



def generate_single_frame_gif(
        rng: BinaryIO, width:int=640, height:int=480, frametimems:float=100):

    # Structure of an image descriptor
    ImageDescriptor = struct.Struct('>hhHH')

    # The main gif header (starts with magic and version string)
    GifHeader = struct.Struct('>s3sBBBHB')

    # The logical screen descriptor
    LogicalScreenDescriptor = struct.Struct('>HHBB')


    # Header fields
    header_magic = bytes('GIF','ascii')
    header_version = bytes([0x38, 0x39, 0x61])
    flags = bytes((0b010 << 5)) # Global color table flag
    background_index = 0
    pixelsize = len(os.urandom(1)[0].to_bytes(1,'big'))*8
    aspectratio = 0


    # Screen descriptors are always palette based so we only need one entry per color
    global_color_table = [0] * pow(2,pixelsize) + \
                         [round(i,3) for i in range(pow(2,pixelsize))]

    # Write the header
    header = GifHeader.pack(
            header_magic,header_version,width,height,flags,background_index,pixelsize)

    # Write the screen descriptor
    screendescriptor = LogicalScreenDescriptor.pack(
            width,height,pixelsize,aspectratio)

    # Write the global color table block
    gctblock = bytes([])
    for c in global_color_table:
        if pixelsize == 8:
            pix = c.to_bytes(1,'big')
        elif pixelsize <= 16:
            pix = c.to_bytes(2,'big')
        else :
            pix = c.to_bytes(4,'big')

        gctblock += pix



    # Write the local color table block
    lctblock = bytes([])


    # Frame data
    framedata = []



    # Generate some frames
    while True:


        # Create a new frame
        frame = [0] * width * height

        # Now fill it with random values
        for x in range(len(frame)):

            # Randomize the next raw value to be included 
            maxval = min(max(global_color_table[1:],default=-1)+1,2**pixelsize-1)
            val = rng.randint(0,maxval)

            # Map the value back onto the global color table
            val = global_color_table.index(val)


            frame[x] = val


        framedata.append(frame)

        # Check if we have enough data
        if len(framedata) > 100:
            break



    # Write all the individual frames
    Blocks = []
    
    LZWMinCodeSize = ceillog2(len(global_color_table)-1) + 1

    def writeLzwDataBlock():

        # This is a bit messy but it works so who cares...
        compressed = b''
        
        # Clear code
        compressed += LzwClearCode.to_bytes(1,'big')

        # TODO: write more codes

        # Determine the maximum code length
        Codelength = LZWMinCodeSize + 1

        # Start by writing the first symbol
        compressed += global_color_table[frame[0]].to_bytes(Codelength,'big')

        lastsym = frame[0] 
        for sym in frame[1:]:
            
            if sym not in dict:
                dict[sym]=lastsym

                # Add a new literal value to the current buffer
                dictlen+=1

                # Write the new symbol to the buffer
                compressed+=(dictlen).to_bytes(Codelength,'big')
                compressed+=global_color_table[lastsym].to_bytes(Codelength,'big')
                
                # Code size has increased but now there's too much stuff in our dictionary?
                if dictlen >= 2**(Codelength):
                    # We double the code size
                    Codelength+=1
                    assert Codelength < 12, "Oh noes!"

                    # And reset our dictionary
                    dictlen = 0


                    # Update the clear code
                    LzwClearCode = dictlen - 1
                    compressed += LzwClearCode.to_bytes(1,'big')
                    
            else:
                pass
                
            # Start with the new byte as well
            compressed += global_color_table[dict[sym]].to_bytes(Codelength,'big')

            # Finally update the last symbol
            lastsym = sym
            
            
        print("Encoded {} symbols".format(len(frame)))
        print("Theoretical compression factor is {}".format(len(frame)/len(compressed)))

        return compressed
        
        
    # First we create the compressor state
    dictlen = 0
    dict = {v: v for v in global_color_table}
    LzwClearCode = dictlen - 1
    LzwEofCode    = dictlen 

    compressed = writeLzwDataBlock()
    
    
    # Frame description contains the following information:
    #       left offset, top offset, width, height, packed field
    Left   = 0
    Top    = 0
    Width  = width 
    Height = height
    PackedFields = 0 # Disposal method is none (0), transparency index present (1), user input (2)

    framedescription = ImageDescriptor.pack(Left,Top,Width,Height,PackedFields)
    Blocks.append(framedescription+compressed+b'\x00')
    
    BlockTable = b''
    for item in Blocks:
        BlockTable += ((len(item)+1).to_bytes(1,'little')+item)

    Trailer = b'\x3B'

    data = header + screendescriptor + gctblock + lctblock + BlockTable + Trailer


    return data
    




if __name__=='__main__':
    from math import log2, ceil
    
    # Get a 8 bit pixel writer function
    packfunc = lambda x: x.to_bytes(1,'big')
    packfunc8 = lambda x: x.to_bytes(1,'big')
    packfunc16 = lambda x: x.to_bytes(2,'big')
    packfunc32 = lambda x: x.to_bytes(4,'big')

    # Get a luminance conversion function
    # NOTE: This just assumes that the input is already 8 bit
    convfunc8 = lambda x: x
    convfunc16 = lambda x: round(x / 257)
    convfunc32 = lambda x: round(x/(257*257))

    # Get a luma lookup table (8 bit)
    getluma8 = lambda r,g,b: (r + g + b)//3

    # Get a luma lookup table (16 bit)
    getluma16 = lambda r,g,b: (r + g + b)//3
    getluma16 /= 257

    # Get a luma lookup table (32 bit)
    getluma32 = lambda r,g,b: (r + g + b)//3
    getluma32 /= 257/257




    # Get a 8 bit pixel writer function
    getwriter = lambda p: globals()['getluma'+str(p)+'b']
    getdecoder = lambda p: globals()['convfunc'+str(p)]

    
    
