#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # header
    out.write('GIF89a'.encode())

    # screen descriptor
    width = 2**int.from_bytes(
                rng.read(1), 'big', signed=False) % 0x40
    height = 2**int.from_bytes(
                    rng.read(1), 'big', signed=False) % 0x40
    flags = bytes([int.from_bytes(
                    rng.read(1), 'big', signed=False)])
    bgcolor = (flags[0] & 1 << 7) >> 7
    aspectratio = ((flags[0] & 1 << 5) >> 5) | \
                  ((flags[0] & 1 << 6) >> 6)
    if aspectratio == 0b00:
        ar = [b'\x00\x00']
    elif aspectratio == 0b01:
        ar = [b'\xff\xff']
    else:
        ar = [(width - 1).to_bytes(2, 'big'),
             (height- 1).to_bytes(2, 'big')]
    colordepth = (flags[0] & 0b111) + 1
    globalcolortableflag = flags[0] & 1 << 7
    sortedcolortableflag = flags[0] & 1 << 5
    colortablesize = {1<<i for i in range(7)} | {0}
    globalcolortablesize = next((s for s in colortablesize
                                if s <= 2**(colordepth+1)), None)
    sortedcolortablesize = next((s for s in colortablesize
                                 if s < 2**(colordepth+1)), None)
    out.write(struct.pack('<HHBBBHB',
                          width, height, colordepth, bgcolor,
                          0xf7, 0, 0))
    out.write(b''.join(ar))

    # palette
    palette = []
    while len(palette) < 2**(colordepth+1):
        r = rng.randrange(0, 256)
        g = rng.randrange(0, 256)
        b = rng.randrange(0, 256)
        pixel = r*0x100 + g*0x10 + b
        if all(pixel!= p for p in palette):
            palette += [pixel] * max(1,(256//len(palette)))


    def write_chunk(keyword: str, content: bytes):
        size = len(content)
        out.write(f'{size:0<3d}{keyword}'.encode())
        out.write(content)
        padding = 0
        while padding < 3 and not size&(padding+1):
            pad = rng.randrange(0, 256)
            assert isinstance(pad, int)
            out.write(bytes([pad]))
            padding += 1
    
    if globalcolortableflag:
        table_type = 0
        if colordepth >= 8:
            entry_size = 32 // colordepth
            content = b''.join([(p%(2**entry_size)).to_bytes(
                                    entry_size, 'big')
                              for p in palette])
        else:
            entry_size = colordepth // 8
            content = b''.join([(p%256).to_bytes(
                                    entry_size, 'big')
                              for p in palette])
        write_chunk('\xfe', content[:globalcolortablesize*entry_size])

    if sortedcolortableflag or globalcolortableflag:
        table_type = 1
        if colordepth >= 8:
            entry_size = 32 // colordepth
            content = b''.join([(p%(2**entry_size)).to_bytes(
                                    entry_size, 'big')
                              for p in palette])
        else:
            entry_size = colordepth // 8
            content = b''.join([(p%256).to_bytes(
                                    entry_size, 'big')
                              for p in palette])

        if sortedcolortableflag:
            write_chunk('\xfd', content[:sortedcolortablesize*entry_size])
            
        if globalcolortableflag:
            write_chunk('\xfe',
                        content[sortedcolortablesize*entry_size:])

    # image descriptor
    disposalmethod = {
        0 : b'\x00',
        1 : b'\x01', 
        2 : b'\x02',
        3 : b'\x03',
        4 : b'\x04'}[rng.randrange(0, 4)]
    userinputflag = rng.getrandbits(1)
    displacyfirstframeflag = rng.getrandbits(1)
    interlacedflag = rng.getrandbits(1)
    lctflag = bool(rng.getrandbits(1))
    lctsize = {0 : 0,
              1 : 2 ** int.from_bytes(rng.read(1),
                                      'big', signed=False)}.get(
                                                  lctflag, None
                                              )
    lctsortflag = False
    if lctflag and lctsize is not None:
        lctsortflag = rng.getrandbits(1)

    if lctsize is None:
        lctsize = 0

    out.write(struct.pack('<BBHBBBBBHB',
                          0x2c, disposalmethod,
                          0, 0,
                          lctflag, lctsize,
                          0 if interlacedflag else 0x40,
                          0 if displacyfirstframeflag else 0x80,
                          0 if userinputflag else 0x0c))
    out.write(b'' if lctflag==0 else
              b'\x00'*lctsize)

    # image data
    framedata = []
    lzwmincodesize = int.from_bytes(rng.read(1),'big')+1
    while True:
        code = int.from_bytes(rng.read(1),'big')
        bitcount = rng.getrandbits(1)
        if bitcount == 1:
            break
        framedata += [code]
        while bitcount > 1:
            code = (code << 1) | rng.getrandbits(1)
            framedata += [code]
            bitcount -= 1
            
    encoded = encode_lzw(framedata, lzwmincodesize)
    framedata = encoded['packed']
    nextcode = encoded['nextcode']
    out.write(b'\x80' if lzwmincodesize == 9 else
              bytes([lzwmincodesize]))
    while framedata:
        chunklen = min(rng.randrange(1,256),
                       len(framedata))
        out.write(framedata[:chunklen])
        del framedata[:chunklen]


def encode_lzw(unencoded: list, lzwmincodesize: int) -> dict:
    """Encodes the unencoded LZW stream with minimum codesize of
       `lzwmincodesize`.
    """

    stringtable = ['\0']
    packed = []
    bits = 9
    mask = 0b100000000
    nextcode = 2**lzwmincodesize
    prevcode = ''
    repetitions = 0
    for c in unencoded:
        if (prevcode + chr(c)) in stringtable:
            repetitions += 1
            continue
        
        packed += pack_byte(repetitions,
                            stringtable, lzwmincodesize)

        repetitions = 0
        index = find_string(chr(c), stringtable)
        packed += pack_byte(index, stringtable, lzwmincodesize)

        prevcode = chr(c)
        while nextcode > len(packed)*8 - bits:
            stringtable += [chr(prevcode)]            
            bits += 1
        nextcode += 1
        
    return {'packed': ''.join(packed),
            'nextcode': nextcode}

def pack_byte(symbol: int, symbols: list, mincodesize: int) -> list:
    """Packs the symbol's prefix to an unsigned integer, then encodes 
       it using either one or two bytes depending on what is needed. 
    """
    if symbol < (1<<(mincodesize)):
        return [bytes([(symbol)],'big')]
    elif symbol < (1<<(mincodesize+1)):
        k = [symbols[symbol]]
        k += [(symbol>>8)].to_bytes(1,'big')
    else:
        raise Exception("Encoding error.")
    return k

def find_string(needle: str, haystack: list) -> int:
    try:
        return haystack.index(needle)
    except ValueError:
        pass
    return len(haystack)