#!/usr/bin/env python3

import io
import math
import os
import pathlib
import pickle
import struct
from typing import BinaryIO


class Color:

    def __init__(self):
        self._r = 0x42
        self._g = 0x7F
        self._b = 0xC6

    @property
    def r(self): return self._r
    
    @property
    def g(self): return self._g

    @property
    def b(self): return self._b

    # Randomizes all three values of this color object
    def randomize(self, rng: BinaryIO):
        self._r += int.from_bytes(
                rng.read(1), byteorder='big', signed=True) & 0xFF
        
        if self._r < 0 or self._r > 0xFF:
            raise ValueError("R value is invalid!")

        self._g += int.from_bytes(
                rng.read(1), byteorder='big', signed=True) & 0xFF

        if self._g < 0 or self._g > 0xFF:
            raise ValueError("G value is invalid!")

        self._b += int.from_bytes(
                rng.read(1), byteorder='big', signed=True) & 0xFF
        
        if self._b < 0 or self._b > 0xFF:
            raise ValueError("B value is invalid!")


class Header:

    def __init__(self):
        self._width = None
        self._height = None
        self._bgcolor = Color()  # Background color
        self._aspectratio = bytes([0x00])

    @property
    def width(self): return self._width

    @width.setter
    def width(self, w): self._width = w

    @property
    def height(self): return self._height

    @height.setter
    def height(self, h): self._height = h

    @property
    def bgcolor(self): return self._bgcolor

    @bgcolor.setter
    def bgcolor(self, c): self._bgcolor = c

    @property
    def aspectratio(self): return self._aspectratio

    @aspectratio.setter
    def aspectratio(self, v): self._aspectratio = v

    # Writes this header to an output stream (usually an open binary
    # file). The length of said output stream should be at least 13 bytes long
    # because it is required for both global and local headers.
    def write(self, out: BinaryIO):
        if not isinstance(out, BinaryIO):
            raise TypeError("The given argument isn't a writable stream.")

        out.seek(0)   # Go back to start of stream
        out.write(struct.pack("<2c", b"G", b"I"))   # Write magic "GI"
        out.write(struct.pack("<B", 0x89))     # Write version number
        out.write(struct.pack("<HBB", \
                self._width,      # Logical screen width
                self._height,    # Logical screen height
                0xF7            # Flags: Global Color Table Flag (1)
                                #        Local Color Table Flag (1)
                                #        Sorting Flag (1)
                                #        Size of GCT (3 bits)
            ))

        # Finally, we can write our background color index...
        out.write(struct.pack("<B", 0x00))         # RBG Value Index
        out.write(struct.pack("<B", 0x00))         # RBG Value Index
        out.write(struct.pack("<B", 0x00))         # RBG Value Index
        #...and proceed with writing the rest of the header...
        out.write((0x00, 0x00, 0x00))                   # Padding

        # And here's where things start getting interesting!
        # We won't write any palette entries now but simply provide
        # an offset where they will reside in the actual image data.
        # This means that by default, we don't need them yet!
        out.write(struct.pack("<B", 0x00))                 # Background Palette Entry Index
        out.write(struct.pack("<B", 0x00))                 # Width/Height Ratio


class ImageDescriptor:

    def __init__(self):
        self._left = None
        self._top = None
        self._width = None
        self._height = None
        self._lctsize = 0
        self._interlaceflag = False
        self._sortflag = True
        self._reserved = bytes([0x00, 0x00])
        self._lctflag = False
        self._lctindex = None
        self._trnscount = 0
        self._trnsindex = []
        self._lct = [Color() for i in range(0x00)]

    # Note that some fields are initialized already as default values -- you
    # only need to add additional code inside of this function once your
    # specification has been finalized.
    def randomize(self, rng: BinaryIO, hdr: Header):
        # Write image descriptor
        pass


class GraphicControlExtension:

    def __init__(self):
        self._size = 0x04
        self._fields = {
                'disposalmethod': None,
                'userinputflag': None,
                'transparentcolorflag': None,
                'delaytime': None}

    @property
    def size(self): return self._size

    @property
    def disposalmethod(self): return self._fields['disposalmethod']

    @disposalmethod.setter
    def disposalmethod(self, v): self._fields['disposalmethod'] = v

    @property
    def userinputflag(self): return self._fields['userinputflag']

    @userinputflag.setter
    def userinputflag(self, v): self._fields['userinputflag'] = v

    @property
    def transparentcolorflag(self): return self._fields['transparentcolorflag']

    @transparentcolorflag.setter
    def transparentcolorflag(self, v): self._fields['transparentcolorflag'] = v

    @property
    def delaytime(self): return self._fields['delaytime']
        
    @delaytime.setter
    def delaytime(self, v): 
        assert type(v) == int and v >= 0 and v <= 65535
        self._fields['delaytime'] = v

    def randomize(self, rng: BinaryIO):
        self._fields['disposalmethod'] = int.from_bytes(
                rng.read(1), byteorder='big') & 0x07
        self._fields['userinputflag'] = bool(int.from_bytes(
                rng.read(1), byteorder='big'))
        self._fields['transparentcolorflag'] = bool(int.from_bytes(
                rng.read(1), byteorder='big'))
        self._fields['delaytime'] = int.from_bytes(
                rng.read(2), byteorder='big') & 0xFFFF


class PlainTextExtension:

    def __init__(self):
        self._textdata = ""

    @property
    def textdata(self): return self._textdata

    @textdata.setter
    def textdata(self, d): self._textdata = d

    def randomize(self, rng: BinaryIO):
        pass
    

class ApplicationExtension:

    def __init__(self):
        self._applicationid = ""
        self._appauthcode = ""

    @property
    def applicationid(self): return self._applicationid

    @applicationid.setter
    def applicationid(self, id): self._applicationid = id

    @property
    def appauthcode(self): return self._appauthcode

    @appauthcode.setter
    def appauthcode(self, ac): self._appauthcode = ac

    def randomize(self, rng: BinaryIO):
        pass


class CommentExtension:

    def __init__(self):
        self._commentdata = ""

    @property
    def commentdata(self): return self._commentdata

    @commentdata.setter
    def commentdata(self, d): self._commentdata = d

    def randomize(self, rng: BinaryIO):
        pass


class ExtensionBlock:

    def __init__(self):
        self._label = ""
        self._extcontents = None

    @property
    def label(self): return self._label

    @label.setter
    def label(self, l): self._label = l

    @property
    def extcontents(self): return self._extcontents

    @extcontents.setter
    def extcontents(self, ec): self._extcontents = ec

    def randomize(self, rng: BinaryIO):
        pass


class GIFFile:

    def __init__(self):
        # A list containing each frame of this animated image
        self._frames = []
        # Internal buffer containing current frame being built
        self._currentframe = []
        # An array describing how many times each frame is supposed to run
        self._duration = []
        # Animation control flags for entire animation
        self._animcontrolflags = {}
        self._hdata = None
        self._gdata = None


    # Creates a new Header object and calls its randomize(...) method on
    # the given input-stream
    def _randomize_hdr(self, rng: BinaryIO):
        hdr = Header()
        hdr.width = int.from_bytes(