#!/usr/bin/env python3

import sys
import os
import math
import struct
from typing import BinaryIO


def main() -> None:
    filename = "random.gif"

    if len(sys.argv) > 1 and not sys.argv[1].startswith('-'):
        # The first argument is the output file name
        filename = str(sys.argv[1])

    with open("/dev/urandom", 'rb') as rng:
        with open(filename, 'wb') as outfile:
            write_random_gif(outfile, rng)


# Writes a random GIF file to `out`.
def write_random_gif(
        out: BinaryIO, rng: BinaryIO
    ) -> None:

    width = getWidth()
    height = getHeight()

    num_frames = getNumFrames(height / 2)

    out.write("GIF89a".encode())
    writeLSD(width, height, 0b001, num_frames, out)
    
    for i in range(num_frames):
        writeGraphicCtrlExt(0, 0, 0x00, 0, out)
        writeImageDesc(0, 0, 0, 0, width, height, False, False, 0, out)
        encodeFramePixels(width * height // (i+1), width, out, rng)


def getWidth() -> int:
    return int((math.log(os.get_terminal_size().columns)) ** 4)


def getHeight() -> int:
    rows = max([int(row) for row in os.popen('stty size', 'r').read().split()])
    return rows - 50


def getNumFrames(max_frames: int) -> int:
    return min(max_frames, int(getWidth()**2))


def pack8bit(*args):
    mask_8bits = struct.pack('@B', 0xFF)
    return bytes(v & mask_8bits for v in args)


def writeLSD(w: int, h: int, fcc: int, nf: int, out: BinaryIO) -> None:
    out.write(pack8bit(w % 256, w >> 8, h % 256, h >> 8))
    out.write(struct.pack('<B', fcc))
    out.write(struct.pack('<B', nf))
    

def writeGraphicCtrlExt(dispose: int, input_flag: bool, delay: int, trans_index: int, out: BinaryIO) -> None:
    packedByte = ((dispose & 7) << 2) | (((input_flag ^ 1) & 1) << 1) | ((delay & 3) << 1) | ((trans_index & 1) << 0) 
    out.write(b'\x21\xF9\x04')
    out.write(packedByte.to_bytes(1,byteorder="little"))
    out.write(b"\x01\x00")

    
def writeImageDesc(left: int, top: int, interlaced: bool, sorted: bool, 
           width: int, height: int,
           local: bool, lct: bool, pal: int, out: BinaryIO) -> None:

    packedByte = ((((interlaced & 1) << 7) | 
                   ((sorted & 1) << 6) |
                   ((local & 1) << 5) |
                   ((lct & 1) << 4) |
                   (pal & 0xf))
                 .to_bytes(1,'big'))
    out.write(packedByte)
    out.write(pack8bit(left % 256, left >> 8, top % 256, top >> 8))
    out.write(pack8bit(width % 256, width >> 8, height % 256, height >> 8))

    
def encodeFramePixels(frameCount: int, lineLen: int, out: BinaryIO, rng: BinaryIO) -> None:
    """Encode Frame Pixels"""
 
    LZWMinimumCodeSize = 2
    ClearCode = 1 << LZWMinimumCodeSize
    EndOfInformation = ClearCode + 1

    LZWDictionary = dict()
    LZWDictionary[""] = ""

    def addTokenToStream(codestream: list, token: str, LZWMinimumCodeSize: int):

        code = LZWDictionary[token]
        if token!= "":

            if token == "\x00":
                pass
            
            elif code == "":
                code = (LZWDictionary[token[:1]] or [ClearCode])[0]
                codestream += pack8bit(code);

            currDictEntry = LZWDictionary[""][-1:] + token[-1:]
            LZWDictionary[""] = "".join(set(LZWDictionary[""] + currDictEntry));
    
            while True:

                shortestMatch = ""
    
                for k in LZWDictionary:
                    if k.startswith(currDictEntry):
                        if (len(k) < len(shortestMatch)):
                            shortestMatch = k 

                if len(shortestMatch) <= LZWMinimumCodeSize:
                    break
                
                code = LZWDictionary[shortestMatch]
                codestream += pack8bit(code);            
            
        else: 
            codestream += pack8bit(code);


    TokenLength = LZWMinimumCodeSize + 1
    clearCodestream = [ ClearCode ]
    addTokenToStream(clearCodestream, "", LZWMinimumCodeSize)
    out.write(bytes(clearCodestream))

    try: 
        while frameCount >= 0:
         
            token = ""
            b = rng.read(2).hex();
            while len(b) == 4 and token!= "\x00":
                token += chr(int(str(b), base=16))
                b = rng.read(2).hex();
 
            LZWMinimumCodeSize = min(LZWMinimumCodeSize + 1, 12)
            addTokenToStream([], token, LZWMinimumCodeSize)
  
            frameCount -= 1
            
    except KeyboardInterrupt:
        print("")

    addTokenToStream("", "\x00", LZWMinimumCodeSize)
    endInfoCodestream = []
    addTokenToStream(endInfoCodestream, "", LZWMinimumCodeSize)
    out.write(bytes(endInfoCodestream + [EndOfInformation]))