#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (eg. /dev/urandom).
#
# Note that this only generates static image blocks;
# the rest of the file is indeterminstic but correct.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # write header and screen descriptor
    header = b'GIF89a'
    screendesc = b'\x00\x00\x00,\x00\x00\x00\x00\x01\xc0' \
                 b'\x0f\x00\x00\xff\xff\xff\x00\x00\x00!'
    out.write(header + screendesc)

    # loop over all image blocks
    while True:

        # get next block type
        btype = ord(btype := rng.read(1)) & 0b00001111

        if     0b0001 == btype:
            # graphic control extension
            ext = struct.pack('<ccBB',
                b't', b'\xFE',
                len(extdata :=
                    b'\x04' + bytes([
                        rng.randint(0,255) for i in range(3)]),
                )+1,
            0x00,)

            out.write(ext + extdata)
        elif   0b0010 == btype or \
                0b0110 == btype:
            # image separator
            #
            # encode width^height as a LZW-optimal codeword
            w,h = [ (b:=rng.read(1))[0] ^ ((i&1)<<7)
                   for i,b in enumerate(rng.read(2))]
            sizecodeword = f'{w},{h}'[::-1].encode()

            # skip to start of LZW subblock
            out.seek(-2, 1)

            # then just copy the length bytes
            # (this is not technically part of the spec...)
            lzwlen = int(''.join('%02x'%b
                             for b in rng.read(int(sizecodeword[:2],16))),
                         16)
            out.write(lzwlen.to_bytes(2,'little'))

            continue

        else:
            break

