#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The output file will be 1024 bytes long.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # First, write the GIF header
    out.write(b'GIF89a')

    # Next, write the logical screen descriptor
    width = (int)(struct.unpack('>H', rng.read(2))[0]) % 65536
    height = (int)(struct.unpack('>H', rng.read(2))[0]) % 65536
    flags = ord(rng.read(1)) % 8
    background_color_index = (int)(struct.unpack('B', rng.read(1))[0]) % 256
    pixel_aspect_ratio = (int)(struct.unpack('B', rng.read(1))[0]) % 256

    out.write(width.to_bytes(2, 'big'))
    out.write(height.to_bytes(2, 'big'))
    out.write((flags).to_bytes(1, 'big'))
    out.write((background_color_index).to_bytes(1, 'big'))
    out.write((pixel_aspect_ratio).to_bytes(1, 'big'))

    # Now, write image descriptor
    left = (int)(struct.unpack('>H', rng.read(2))[0]) % 65536
    top = (int)(struct.unpack('>H', rng.read(2))[0]) % 65536
    width = (int)(struct.unpack('>H', rng.read(2))[0]) % 65536
    height = (int)(struct.unpack('>H', rng.read(2))[0]) % 65536
    flags = ord(rng.read(1)) % 8
    transparent_color_index = (int)(struct.unpack('B', rng.read(1))[0]) % 256
    interlace_flag = bool((flags >> 7) & 1)
    sorted_color_table_flag = bool((flags >> 5) & 1)
    
    out.write(left.to_bytes(2, 'big'))
    out.write(top.to_bytes(2, 'big'))
    out.write(width.to_bytes(2, 'big'))
    out.write(height.to_bytes(2, 'big'))
    out.write((flags).to_bytes(1, 'big'))
    out.write((transparent_color_index).to_bytes(1, 'big'))
    out.write((0x00).to_bytes(1, 'big')) # Delay time in milliseconds

    # Finally, write LZW minimum code size and compressed data
    lzwMinimumCodeSize = (int)(struct.unpack('B', rng.read(1))[0]) % 256
    out.write((lzwMinimumCodeSize).to_bytes(1, 'big'))

    # Write color table for indexed colors, if necessary
    if not sorted_color_table_flag:
        numColors = 2**(lzwMinimumCodeSize+1)

        for i in range(numColors):
            red = (int)(struct.unpack('B', rng.read(1))[0]) % 256
            green = (int)(struct.unpack('B', rng.read(1))[0]) % 256
            blue = (int)(struct.unpack('B', rng.read(1))[0]) % 256

            out.write((red).to_bytes(1, 'big'))
            out.write((green).to_bytes(1, 'big'))
            out.write((blue).to_bytes(1, 'big'))

    # Write compressed stream (with RLE compression)
    while True:
        markerByte = (int)(struct.unpack('B', rng.read(1))[0]) % 256

        if markerByte == 0x00: break
        elif markerByte >= 0x80:
            out.write(((markerByte - 0x80)*256).to_bytes(1, 'big'))
        else:
            length = (int)(struct.unpack('B', rng.read(1))[0]) % 256 + 1

            for _ in range(length):
                out.write(rng.read(1))