#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out`.
# Uses /dev/urandom to seed the RNG (this is not cryptographically safe).
class RandomGifGen:

    def __init__(self):
        self._rng = open('/dev/urandom', 'rb')
        self._data_len = 0
        self._data = None

        # We can't just use os.urandom() as our RNG because then it would be
        # non-deterministic. For example, if we do this:
        #   with open('foo.dat', 'wb') as f:
        #       for i in range(17):
        #           f.write(os.urandom(i))
        # Then foo.dat will always be exactly the same! But by specifying an
        # exact length, we avoid that problem and get reproducible results:
        #   $ dd if=/dev/urandom of=foo2.dat bs=1 count=16 status=none && sha256sum foo*
        #   44d74c61dd5f9e7f52b8ec32621084dc144acfb9b5b77c681d999bdce7948cb4  foo.dat
        #   44d74c61dd5f9e7f52b8ec32621084dc144acfb9b5b77c681d999bdce7948cb4  foo2.dat
        # Therefore let's just ensure that the result is deterministic even when
        # the PRNG picks different values on different runs. So set the seed.
        seed = bytes([
            *struct.unpack('>HHH',
                self._rng.read(6)),
            0])
        print(seed, flush=True)
        self._rng.seek(0)
        self._rng.write(seed)


    def reset(self):
        self._data_len += len(self._data)
        self._data = None


    def gen(self, out: BinaryIO):
        if self._data is None or self._data_len > 4096:
            self._data = self._gen_next()
            self._data_len = 0
        chunk = self._data[self._data_len:]
        end = min(len(chunk), 4096 - self._data_len)
        out.write(chunk[:end])
        self._data_len += end
        return end < len(chunk)


    # Generate the next frame. This is also used at init time so that we can
    # have reproducible results (see the comment in `__init__`).
    def _gen_next(self) -> bytes:
        header = b'GIF89a'
        datalen = max(int.from_bytes(
                    self._rng.read(3), byteorder='big'),
                    1)
        data = self._rng.read(datalen)
        return header + data


