#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write('GIF89a'.encode())  # Magic numbers
    width = (2**10)**rng.randint(4)   # Width is between 512 and 16384 pixels
    height = (width-1)*rng.randint(5)+1    # Height should be about width times 2 to 5
    out.write(struct.pack('<HHB', width, height, 0x7f))     # Logical screen descriptor block
    out.write((b'\xf7\x00\x00')*height)        # Global color table definition block

    # Image blocks
    for _ in range(1+int(rng.triangular(.5,.1,1))):
        imlen = int.from_bytes(
                rng.read(3), byteorder='big')

        # Local colour table
        if bool(rng.getrandbits(1)):
            ctsize = 2**(rng.randint(1,8))
            out.write(
                    bytes([ctsize])
                    + rng.read(3*ctsize).ljust(3*ctsize,'\x00'))

            # The image descriptor block
            xpos = 1+int(rng.triangular(0.,0.5*(width-imlen)/width,0.5*(width-imlen)))  # X position of top-left corner
            ypos = 1+int(rng.triangular(0.,0.5*(height-imlen)/height,0.5*(height-imlen)))  # Y position of top-left corner
            lzwmincodesize=rng.randint(2,8)      # Lower bound on LZW code size
            out.write(bytes([
                0x2c,                         # Image separator
                0,                            # Left coordinate of image frame
                xpos&0xff,                    # Top coordinate of image frame
                ((ypos>>8)|0x80)&0xff,       # Width of image frame
                (((ypos<<8)-xpos)>>8)&0xff,  # Height of image frame
                lzwmincodesize]))             # LZW minimum code size

        else:
            out.write(b'\xfe')                  # Extension subblock

        out.write(struct.pack('<HBB', imlen, 0, 0))  # Sub-block length

        # Plain text encoding - no compression
        enc = False
        while not enc:
            j = rng.randint(0,3)
            if j == 0:   # Plain text mode, no compression
                out.write(rng.read(imlen))
                enc = True
            elif j == 1: # Plain text mode with RLE compression
                nrep = rng.randint(4,8)
                out.write(
                        b'\xfe\x01' + struct.pack("<B",nrep) + rng.read(imlen//nrep)
                        )
                enc = True
            elif j == 2: # LZW compression
                lzwmincodesize=(rng.randint(2,12) | rng.getrandbits(3)).to_bytes(1,'big')  # Randomised LZW minimum code size
                out.write(lzwmincodesize)
                lzwsizemax=int((((rng.getrandbits(1)+1)<<8) + rng.getrandbits(1))*rng.getrandbits(1))   # Number of bits used to code next pixel
                out.write(lzwsizemax.to_bytes(1,'big'))

                seq=[]                                  # Encoded sequence stack
                baseseq=[]                              # Base sequence stack
                varlencodes=[]                          # Variable length codes
                ibasecode=2**lzwmincodesize              # Initial base code value
                sizebasecode=4                           # Next power of two above initial base code
                maxcodevalue=ibasecode+2**lzwmincodesize-1   # Maximum code value
                clearcode=ibasecode                     # Clear code
                endofinformation=maxcodevalue            # End of information code
                nextcodevalue=clearcode+1               # Next available code value
                bitsinbuffer=0                          # Bits still left in buffer
                buf=0                                   # Buffer containing encoded values

                def flush():
                    nonlocal bitsinbuffer,buf

                    # Add literals & codes to output
                    out.write(
                            (nextcodevalue-ibasecode).to_bytes(2,'little')
                            + bytes((buf >> i) & 0xff
                                    for i in reversed(range(bitsinbuffer))))

                    # Check if there are enough bits remaining in this chunk
                    while bitsinbuffer >= 8:
                        out.write(bytes([(buf >> i) & 0xff
                                        for i in reversed(range(8))]))
                        bitsinbuffer-=8

                # Generate random variable length codes
                while len(varlencodes)<4095:
                    vlc = rng.getrandbits(2)
                    if vlc==0:
                        varlencodes.append(0)
                        varlencodes.append(1)
                        varlencodes.append(1+rng.getrandbits(1))
                        varlencodes.append(2+rng.getrandbits(1))
                        varlencodes.append(
                                (2+(vlc&1))+int(rng.triangular(0.6,-0.1,0.2))
                                )
                    elif vlc==1 or vlc==2:
                        varlencodes.append(2+rng.getrandbits(2))
                    else:
                        break
                varlencodes.sort()

                # Process input
                while True:
                    if len(baseseq)==1:                   # No more than one character since last flushing?
                        repchar=baseseq[0]                 # Get repeated character from last time
                        repsleft=1                        # One repetition left to do

                        while len(seq)>repsleft:           # Do we need any more repetitions?
                            seq.pop(-1)                    # Remove characters that will now be overwritten

                            # Flush literal
                            if len(seq)==0:
                                flush()

                            # Flush code
                            else:
                                seq.pop(-1)

                                # Write short form code
                                if seq[-1]<maxcodevalue-sizebasecode \
                                            and len(seq)>=2 and seq[-2]==seq[-1]:

                                    # Find longest match
                                    startmatch=0
                                    curseq=baseseq[:]
                                    for sidx in range(startmatch+len(curseq),len(baseseq)):
                                        # Match found?
                                        if baseseq[sidx:]!=curseq[:len(baseseq[sidx:])]:
                                            break

                                        # Move current sequence along by one
                                        startmatch+=1
                                        curseq=baseseq[sidx+1:]+curseq[:1]
                                    else:
                                        # If we get here then there's a perfect match
                                        assert baseseq[:sidx]==curseq[:len(curseq)]
                                        startmatch+=1

                                    # Output first part of match
                                    repscnt=sidx-startmatch-1
                                    scnt=repscnt.bit_length()+1-3
                                    out.write(((repscnt << 2) | scnt).to_bytes(1,'big'))
                                    out.write((startmatch-1).to_bytes(1,'big'))

                                    # Output second part of match
                                    out.write(curseq[:repscnt+1])

                                    # Update list of anything to write at end of loop iteration
                                    seq=[seq[-repscnt:]]

                                # Longer form code
                                else:
                                    # Determine how many bits code value needs
                                    ival=0
                                    while nextcodevalue>=(4096-2**ival):
                                        ival+=1

                                    # Code value comes before actual code
                                    out.write(
                                            (0xc0+ivals).to_bytes(1,'big')
                                            + nextcodevalue.to_bytes(2**ivals,'little')
                                            )

                                    # Update next code value
                                    nextcodevalue+=2**ival

                                # Handle case where sequence was empty
                                if len(seq)==0:
                                    break

                                # Flush literal
                                if len(seq)==1:
                                    flush()

                    char=rng.getrandbits(8)
                    baseseq.append(char)                   # Store current character as new "base"

                    # Check if it fits in current code
                    if len(seq)!=0 and char==seq[-1]:
                        seq.append(char)                   # Add character to sequence
                        continue

                    # Can we use a shorter code instead?
                    repscnt=-1                              # Number of characters to repeat
                    if len(varlencodes)>0 and len(varlencodes)%4==0:         # At least four variable-length codes have been processed after each other
                        idx0=rng.randint(0,len(varlencodes)//4-1)*4          # Index into array of variable-length codes
                        idx1=idx0+rng.randint(0,len(varlencodes)//4-1)*4    # Index into array of variable-length codes
                        idx2=idx0+rng.randint(0,len(varlencodes)//4-1