#!/usr/bin/env python3

import struct
from typing import BinaryIO
from os import urandom


# Generate a 256-color palette of (R, G, B) triples and return it as bytes
def generate_palette():
    # We don't really need to do this the smart way since we only have
    # 256 colors to work with but hey, if you're feeling extra fancy...

    result = []
    for i in range(0x100):
        r = ((i & 0b111) << 5) | 0xFF
        g = (((i >> 3) & 0b111) << 5) | 0xFF
        b = (((i >> 6) & 0b1111) << 4) | 0xFF
        result += [r] * 3;
        result += [g] * 3;
        result += [b] * 3;
    return bytes(result)

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    width = height = 0
    while True:
        # Header
        blocktype, = struct.unpack('<s', rng.read(1))

        if blocktype == 'IHDR':
            # Image Size
            width, height, _, _ = struct.unpack('<HHBBBB', rng.read(8))

            # Frame Control
            framecontrol, = struct.unpack('>B', rng.read(1))
            globalcolortableflag = framecontrol & 0x80
            interlacemode = framecontrol & 0x40
            sortflag = framecontrol & 0x20
            reservedbits = framecontrol & 0x70
            lctsize = 0x20 * (framecontrol & 0x07)

            # Global Color Table
            globalcolortable = None
            if globalcolortableflag:
                globalcolortable = generate_palette()
                out.write(globalcolortable)

            break

    # Local Color Table
    localcolortable = None
    if not globalcolortable or len(globalcolortable) > lctsize:
        localcolortable = generate_palette()
        out.write(localcolortable)

    # Application Extension
    applicationextensionblock = b'\xff\x0bNETSCAPE2.0\x03\x01\x00\x00\x00'
    applicationextensionsubblock = b'\x0b\x00\x01' + \
                                    struct.pack('<H',
                                               0xFFFF - (len(applicationextensionblock) % 0xFFFF))
    out.write(applicationextensionblock +
              applicationextensionsubblock)

    IMGDATA = 0xFFFF - (width * height *
                         (lctsize // 3 + 1)).to_bytes(2,'big')
    out.write(IMGDATA)