#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

MAX_SIZE = 0xffffffff


class RandomGifGenerator():

    def __init__(self):
        self._rng = None
        self._out = None
        
    def set_size(self, size):
        if (size < 1 or MAX_SIZE - size <= 16):
            raise ValueError("Size must be between 1 and %d." %
                    (MAX_SIZE-1))
        self._size = size
    
    def set_output(self, filename=None):
        self._filename = filename

    # Generate a valid length of a data subfield
    @staticmethod
    def _get_length() -> int:
        return min((int.from_bytes(os.urandom(1),
                byteorder='big'), MAX_SIZE-16))

    # Create the Header block of the GIF file
    def _create_header(self) -> bytes:

        # The first char is for Signature, so this has to be 'G'.
        signature = b'G'
        
        # Version. The only one supported by browsers as of today
        version = b'89a'

        # Width in px
        width = self._size.to_bytes(2, byteorder='big')

        # Height in px
        height = self._size.to_bytes(2, byteorder='big')

        # Flags : 
        # Set color table flag (b7). We will always use it.
        flags = b'\x87'

        # Color resolution : 11 bits.
        colors = b'\x00\x00'

        # Sort Flag : 0. Not sorted.
        sort = b'\x00'

        # Size of global color table : 0. No global CT.
        ctsize = b'\x00'

        return signature + version \
              + width + height + flags \
              + colors + sort + ctsize 

    # Creates the Data blocks containing RNG's output
    # until we reach target size
    def _create_data(self) -> bytes:

        res = b''

        while len(res) < self._size:

            # Get a length depending on what we need 
            # to fill up our buffer
            remaining_len = self._size - len(res)
            
            # If there are not enough remaining pixels to cover
            # the difference between buffer capacity and required size,
            # just take whatever remains
            if remaining_len > 0xfffff:
                l = remaining_len
            else:

                # Else get some non null values 
                l = self._get_length()
                
                # That might have generated too many nulls.
                # Let's fix that
                while l >= remaining_len:

                    # This makes sure we do not enter an infinite loop 
                    assert l!= 0
                    
                    # Reduce the size randomly
                    l -= max(l // 4, 5)
                
            res += os.urandom(min(remaining_len, l))
            
        return res[:self._size]


    # Generate the whole GIF file with given parameters
    def generate(self):

        # Check all necessary paramaters were set earlier
        assert self._size is not None
        assert self._filename is not None
        assert self._rng is not None
        assert self._out is not None

        # Write the header to our output
        self._out.write(self._create_header())

        # Then write every other blocks until we reached
        # expected size
        self._out.write(self._create_data())

        # Ensure end of stream
        self._out.flush()
        self._out.close()

        print('File generated successfully.')




if __name__ == '__main__':

    gen = RandomGifGenerator()

    try:
        # Handle CLI arguments
        args_len = len(sys.argv[1:])
        if (args_len >= 2):
            gen.set_output(sys.argv[-1])
            gen.set_size(int(sys.argv[-2]))
        elif (args_len == 1):
            gen.set_output(sys.argv[-1])
        else:
            print('Usage:')
            print('%s [output] [size]' % (sys.argv[0]))
            exit(-1)

        # Open files 
        gen._rng = open('/dev/urandom', 'rb')
        gen._out = open(gen._filename, 'wb')

        # Generate the file
        gen.generate()

    finally:
        if (gen._rng is not None):
            gen._rng.close()
        if (gen._out is not None):
            gen._out.close