#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file with 256 colors and dimensions
# between (400x400 - 1000x1000) into `out` using the
# random number generator `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Note that the LZW minimum code size is 2-bits long
    # so we have to add one color for black at least...
    palette = [b'\xff\xff\xff'] * 257

    # Add some random colours in there to make it interesting
    for i in range(len(palette)):
        r = rng.read(1)[0] % 0xa1 # Make sure it's not too dark
        g = rng.read(1)[0] % 0xa1
        b = rng.read(1)[0] % 0xa1

        palette[i] += bytes([r,g,b])

    lzwTableSize = len(palette) # Default value

    frameHeader  = b'GIF89a'

    width        = rng.read(2)
    height       = rng.read(2)

    flags = 0x87 # Invisible, Color table sorted
    if rng.read(1)[0]: # Has global color table?
        flags |= 0x80 # Global color table
    else:
        flags |= 0x07 # No local colour table

    background   = rng.read(1)[0] & 0xf
    _reserved    = rng.read(1)[0] & 0x7
    framesize    = rng.read(2)

    imageData = b''
    while True:
        blockType = rng.read(1)[0]
        blockSize = rng.read(1)[0]

        if blockType == ord('\x21'): # Extension block
            label = rng.read(1)[0]

            if label == ord('!'):
                extName = rng.read(1)[0]

                if extName == ord('F'):
                    label = 'Frame duration'

                    if blockSize!= 4:
                        continue
                    
                    # TODO Parse this properly
                    print(label, blockSize, hex(blockType))


        elif blockType == ord('\x2c'): # Image descriptor
            left   = rng.read(2)
            top    = rng.read(2)
            width  = rng.read(2)
            height = rng.read(2)

            length = rng.read(1)[0] & 0x7f
            
            if length!= 8:
                continue

            interlaceFlag = bool(rng.read(1)[0] & 0x40)
            sortFlag      = bool(rng.read(1)[0] & 0x20)
            reserved      = rng.read(2)

            lzwMinCodeSize = rng.read(1)[0] & 0x07

            rawImageData = rng.read(length)
            break


    print("LZW min code size:", lzwMinCodeSize)
    # We force LZW compression here because the GIF format
    # does not support direct indexed images :-(
    lzwCompressedImage = __lzw_compress(rawImageData, lzwMinCodeSize+1)
    
    # Build the image table which contains the LZW compressed
    # image and its palette info
    imageTable = struct.pack("<B", lzwMinCodeSize)
    imageTable += lzwCompressedImage

    # Now build the final GIF stream structure
    print(frameHeader)
    print(width)
    print(height)
    print(flags.to_bytes(1, byteorder='big'))
    print(background.to_bytes(1, byteorder='big'))
    print(_reserved.to_bytes(1, byteorder='big'))
    print(framesize)
    print(lzwMinCodeSize.to_bytes(1, byteorder='big')) # This is wrong...
    print(imageTable)


def __decode_byte(
        buffer: bytearray, bits: int):
    """
    Decodes a byte from the bitstream
    """

    byte = 0
    for i in range(0, 8):
        shift = 7 - i
        try:
            tmpbit = buffer.popleft()

            byte |= ((tmpbit >> (7-shift))) << shift
        except IndexError as e:
            return None

    return byte

def __lzw_compress(buffer: bytes, max_code_size=None):
    """
    Returns an LZW encoded version of the input `bytes` object.
    The resulting bytes are packed according to the LZW spec.
    """

    class TableEntry():
        def __init__(self, prevChar, nextChar=None):
            self.prevChar = prevChar
            self.nextChars = []
            if nextChar:
                self.addNextChar(nextChar)
        
        def addNextChar(self, char):
            assert isinstance(char, int) or isinstance(char, str)

            if not char:
                raise RuntimeError("Cannot add empty next character!")

            if char not in self.nextChars:
                self.nextChars.append(char)


        @property
        def hasMoreChars(self):
            return len(self.nextChars) > 0

        @property
        def lastChar(self):
            if self.hasMoreChars:
                return self.nextChars[-1]
            else:
                return None

        def getByteStream(self, escape=False):
            escapedBytes = (
                0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 
                0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 
            )


            if self.prevChar < 0xf0 and not escape:
                yield self.prevChar
            elif self.prevChar <= 0xffff:
                yield escapedBytes[0]
                yield (self.prevChar >> 8) & 0xff
                yield self.prevChar & 0xff
            elif self.prevChar <= 0x1fffff:
                yield escapedBytes[1]
                yield (self.prevChar >> 16) & 0xff
                yield (self.prevChar >> 8) & 0xff
                yield self.prevChar & 0xff
            elif self.prevChar <= 0x3ffffff:
                yield escapedBytes[2]
                yield (self.prevChar >> 24) & 0xff
                yield (self.prevChar >> 16) & 0xff
                yield (self.prevChar >> 8) & 0xff
                yield self.prevChar & 0xff
            elif self.prevChar <= 0x7fffffff:
                yield escapedBytes[3]
                yield (self.prevChar >> 32) & 0xff
                yield (self.prevChar >> 24) & 0xff
                yield (self.prevChar >> 16) & 0xff
                yield (self.prevChar >> 8) & 0xff
                yield self.prevChar & 0xff
            else:
                raise ValueError("Invalid prevChar!")

            nxt = self.lastChar
            while nxt is not None:
                if nxt >= 0xf0 and not escape:
                    yield nxt
                elif nxt <= 0xffff:
                    yield escapedBytes[0]
                    yield nxt >> 8
                    yield nxt & 0xff
                elif nxt <= 0x1fffff:
                    yield escapedBytes[1]
                    yield (nxt >> 16) & 0xff
                    yield (nxt >> 8) & 0xff
                    yield nxt & 0xff
                elif nxt <= 0x3ffffff:
                    yield escapedBytes[2]
                    yield (nxt >> 24) & 0xff
                    yield (nxt >> 16) & 0xff
                    yield (nxt >> 8) & 0xff
                    yield nxt & 0xff
                elif nxt <= 0x7fffffff:
                    yield escapedBytes[3]
                    yield (nxt >> 32) & 0xff
                    yield (nxt >> 24) & 0xff
                    yield (nxt >> 16) & 0xff
                    yield (nxt >> 8) & 0xff
                    yield nxt & 0xff
                else:
                    raise ValueError("Invalid character")
                
                nxt = self.getNextChar()

    def getNextChar():
        if currentTableEntry.hasMoreChars:
            return currentTableEntry.lastChar
        else:
            return None
        

    currentTableEntry = TableEntry(-1)
    outputBuffer     = bytearray()
    index            = -1
    codeBits         = struct.pack("<H", max_code_size) # H for uint16_t
    maxCode           = pow(2,max_code_size)-1 # Maximum possible code

    # We use a deque so we can pop elements off easily without having
    # to worry about bounds.
    byteBuffer = bytearray(buffer)
    byteBuffer.extend(struct.pack("<B", 0)) # Append EOI marker
    byteBuffer.extend((0,)*8)

    while True:
        # Read a new code from the bitstream
        code = __decode_byte(byteBuffer, code