# Generates an uncompressed GIF file to `out`. The file
# is made up of chunks and their data as specified in the
# GIF specification (https://www.w3.org/Graphics/GIF/spec-gif89a.txt).
# Data is written from `data` while keeping track of how many bytes
# have been generated.
def write_gif_image(width: int, height: int, palette: [str],
                    data: BinaryIO, out: BinaryIO):

    # Write initial chunk - logical screen descriptor
    out.write(b"GIF")   # Signature
    write_ushort(65535)  # Screen width
    write_ushort(65535)  # Screen height
    out.write((1 << 7) | 0)    # Global color table flag
    out.write(b'\x00\xFF')     # Background colour index
    out.write(b'\x00\x00')     # Pixel aspect ratio

    # Write global color table
    for i in range(256 * 4):
        out.write(b'\x00')         # Empty entry


    # Start frame
    write_chunk('IHDR', out)
    write_uint(width)       # Width
    write_uint(height)      # Height
    out.write(b'\x80')        # Local color table flag
    out.write(b'\x00')        # Interlace flag
    out.write(b'\x00')        # Sorting method
    out.write(b'\x00')        # Reserved field

    # Write local color table if applicable
    if len(palette) > 0:
        write_chunk('PLTE', out)

        for p in palette:
            # For each stringified hex value in the list of colors,
            # write its equivalent RGB bytes into the color table
            rgb = [int(p[i:i+2], base=16) for i in (0, 2, 4)]
            assert all(map(lambda x : isinstance(x, int), rgb))

            for r in rgb:
                out.write(r.to_bytes(length=1, byteorder='little'))

    else:
        # There should always be at least one color table
        raise ValueError("Need at least one color in the palette.")

    # Image data starts here...
    write_chunk('IDAT', out)

    # This application requires that there are no more than 255 columns per byte
    # TODO: document this better...
    rowsPerByte = 8
    numRows = math.ceil(height / rowsPerByte)
    # Each row in the image can have between 1 and 255 pixels and we
    # need to make sure it fits within a single byte...
    maxPixelsPerRow = min(max([l.count(',') + 1 for l in data]), 255)
    
    # Now we just loop over the lines in the input file
    for line in data.readlines():
        
        # Get rid of the newline char
        line = line[:len(line)-1]

        # Convert the comma separated values into binary strings
        bits = ''.join(format(int(c), '08b') for c in line.split(','))
        # We pad the binary string with zeros to fill an entire byte
        paddedBits = bits.zfill(rowsPerByte * maxPixelsPerRow)

        # Finally, we split the bitstring into groups of eight bits
        # which we will write to the output file as bytes
        for byteIndex, byteStr in enumerate(paddedBits[::8]):
            out.write(int(byteStr[::-1], base=2).to_bytes(length=1,
                                                          byteorder='little'))
            
    # End the IDAT section by appending the ending code
    out.write(b'\x00')
    out.write(b'\x3B')


# Convenience function which writes a uint to the output stream
def write_uint(value, out: BinaryIO):
    out.write(value.to_bytes(length=4,
                             byteorder='little', signed=False))

# Convenience function which writes a ushort to the output stream
def write_ushort(value, out: BinaryIO):
    out.write(value.to_bytes(length=2,
                             byteorder='little', signed=False))
    


