#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (for example /dev/urandom).
# The generated image size is 2048*1536 pixels or roughly
# 7MiB in memory.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Set to True if you want a black background instead of white
    black_background = False

    # Generate header
    header = b'GIF89a'
    out.write(header)

    # Header length (in bytes)
    headerlen = len(header)

    # Logical Screen Descriptor Block
    descwidth = 2048
    descheight = 1536
    colorres = 0x1f
    globalcolormapsize = 0
    bgcolorindex = -1 if not black_background else 0
    pixelaspectratio = 0
    logicalscreendescriptor = struct.pack('<HHBBBH',
        descwidth, descheight,
        colorres, globalcolormapsize,
        bgcolorindex, pixelaspectratio)
    out.write(logicalscreendescriptor)
    headerlen += len(logicalscreendescriptor)

    # Color Table (global palette)
    globalpaletteentries = 768 if not black_background else 64;
    for i in range(globalpaletteentries):
        r = g = b = 0xff if (not black_background and i % 6 == 0) else \
                round((i * 256) // globalpaletteentries)
        out.write(struct.pack('BBB',r,g,b))
    headerlen += globalpaletteentries * 3

    # Image Descriptor Block
    imgleft = 0
    imgtop = 0
    imgwidth = descwidth
    imgheight = descheight
    packedfields = 0xc0
    interlacedflag = 0
    localcolortableflag = 0
    sortedcolorusageflag = 0
    localcolortablesize = 0
    graphiccontrolblock = struct.pack('<HHBBBBHHB',
        imgleft, imgtop, imgwidth, imgheight,
        packedfields, interlacedflag,
        localcolortableflag,
        sortedcolorusageflag, localcolortablesize)
    out.write(graphiccontrolblock)
    headerlen += len(graphiccontrolblock)

    # Local Color Table
    for i in range(localcolortablesize):
        r = g = b = 0xff if (not black_background and i % 6 == 0) else \
                round((i * 256) // localcolortablesize)
        out.write(struct.pack('BBB',r,g,b))
    headerlen += localcolortablesize * 3

    # LZW Minimum Code Size
    minimumcodesize = 9
    out.write(minimumcodesize.to_bytes(1, 'big'))
    headerlen += 1

    # Data
    # lzw minimum code size already written at this point
    maxcodeplusone = 2**minimumcodesize + 1
    clearcode = 1 << minimumcodesize
    endofinformation = clearcode + 1
    lastdatawritten = None
    codesused = [clearcode]
    prefix = ''
    for _ in range(imgwidth * imgheight):

        # Fill up to `maxcodeplusone` with random bits until we get one that hasn't been used yet
        while True:

            # If there aren't any unused codes left, start over by incrementing the minimum code size
            if min(codesused) >= maxcodeplusone:

                # Add all suffixes to the dictionary before increasing the minimum code size
                nextprefix = prefix
                swapped = {}
                for d in codesused:
                    if d < maxcodeplusone:
                        swapped[d] = nextprefix
                        nextprefix = (nextprefix + d)[-minimumcodesize:]
                # Add the new character as well
                swapped[lastdatawritten] = nextprefix[-minimumcodesize:]
                for k in sorted(swapped.keys()):
                    w.write(chr(k).encode())
                    codesused.append(k)

                # Increment the minimum code size and prepare to restart
                minimumcodesize += 1
                maxcodeplusone = 2**minimumcodesize + 1
                clearcode = 1 << minimumcodesize
                endofinformation = clearcode + 1
                lastdatawritten = None
                codesused = [clearcode]
                prefix = ''

            # Pick a bit from the stream
            cbit = ord(rng.read(1)) & 0x80!= 0

            # Convert it to a number between zero and the maximum value based on the current minimum code size
            n = sum([cbit << i for i in reversed(range(minimumcodesize))])

            # See if the number has already been used
            if n not in codesused: break


        # Write the actual data
        w.write(chr(n).encode())
        codesused.append(n)
        lastdatawritten = n
        prefix = (prefix+chr(n))[1:]


    # End Of Information Code
    out.write(endofinformation.to_bytes(1, 'big'))
    headerlen += 1

