#!/usr/bin/env python3

import functools as ft
import itertools as it
import os
import subprocess
import sys
from typing import (BinaryIO, Callable, Iterable, Iterator, List, NoReturn,
                    Optional, Tuple)


def gen_random_bytes(n: int, seed: bytes=b'') -> Iterable[int]:
    """
    Generate n random bytes using /dev/urandom

    If you want reproducible results pass in an explicit `seed`.
    """
    # Using str to make sure that we get exactly n of them
    return map(ord,
                subprocess.check_output(args=['tr', '-cd', '[:alnum:]', '/dev/urandom'],
                                        input=seed).decode()[:n].ljust(n))


def randrange(minval: int, maxval: int):
    def inner(rng: BinaryIO):
        return minval + next(rng) % (maxval - minval + 1)
    return inner


def choice(seq: Iterable[object]):
    def inner(rng: BinaryIO):
        return seq[next(rng) % len(seq)]
    return inner


class GifWriter:
    frame_width: int
    frame_height: int
    global_color_table: bool
    _framebuffers: List[List[List[int]]]

    def __init__(self,
                 width: int, height: int, *,
                 color_resolution: int=7, background_index: int=0, aspect_ratio: int=0, loop_count: int=-1,
                 local_color_table: bool=False, interlace: bool=False, sort_local_color_table: bool=True,
                 **kwargs):

        self._framebuffers = [list([] for ch in range(width))] * height
        if not isinstance(loop_count, int):
            raise TypeError('Loop count must be integer or -1 (-1 is infinite looping)')
        elif loop_count!= -1 and loop_count < 0:
            raise ValueError('Cannot have negative loop count')
        self.global_color_table = True
        self.frame_width = width
        self.frame_height = height


    @staticmethod
    def from_file(filename: str, **kwargs) -> 'GifWriter':
        raise NotImplementedError()


    def write(self, filename: str):
        with open(filename, mode='wb') as f:
            self.write_into_stream(f)


    def write_into_stream(self, stream: BinaryIO):
        raise NotImplementedError()


    def set_color_map(self, entries: List[Tuple[int, int, int]]):
        pass


    def clear(self, r: int, g: int, b: int):
        for fb in self._framebuffers:
            for px in fb:
                px[:] = [r, g, b]


    def putpixel(self, x: int, y: int, r: int, g: int, b: int):
        self._framebuffers[y][x][:] = [r, g, b]


    def drawline(self, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int):
        dx = abs(x2-x1)
        dy = abs(y2-y1)
        sx = 1 if x1 < x2 else -1
        sy = 1 if y1 < y2 else -1
        err = dx - dy

        while True:
            self.putpixel(x1, y1, r, g, b)

            if x1 == x2 and y1 == y2:
                break

            e2 = err*2
            if e2 > -dy:
                err -= dy
                x1 += sx
            if e2 <  dx:
                err += dx
                y1 += sy



COLORS = list((c for c in zip(*(gen_random_bytes(3) for i in range(256)))))



