#!/usr/bin/env python3

import os
import sys
import zlib as zl
import struct
from typing import BinaryIO

# If we are not given a size for our output (in bytes), default to 1MB.
OUTPUT_DEFAULT_SIZE = 1024 * 1024

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# Returns the output length in bytes.
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # The first few bytes specify the size of the global color table.
    # We can only support one "global" colour table at this time so it must be
    # zeroed and ignored by all other readers. For more information see section
    # 17 of https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    out.write(b'\x00\x00\x00')

    # Next comes an image descriptor specifying the screen width and height
    # in pixels. As there is no variation available here yet I just picked two
    # arbitrary values that will work with most software.
    width = 512
    height = 512
    packed_width = width.to_bytes(2, 'big', signed=False)
    packed_height = height.to_bytes(2, 'big', signed=False)
    packed_flags = b'\x08\x00'    # LZW compression
    out.write(packed_width + packed_height + packed_flags)

    # Now follows the local colour palette which again has a fixed size but has
    # many colours set to black. Again this seems to be required by many
    # softwares so it makes sense to keep it (unless anyone knows what its used
    # for?). Note that the last element is reserved and should always be zero.
    # This prevents sofware from doing a safe divide when decoding the pixel
    # stream later on.
    num_colours = 768
    palette = [1] * 3 * num_colours    # All set to black.
    palette[-1] = 0                     # Last entry is reserved and unused.
    out.write(bytearray(palette))

    # Finally followed by the actual pixel data. This has to be compressed and
    # uses run-length encoding where runs of identical bytes are encoded with
    # special markers. See section 13, 15 and 16 of the spec mentioned above.
    num_pixels = width * height
    pixel_data = rng.read(num_pixels // 2)       # Half the pixels are random.
    non_zero_stream = list()                    # Run-length encode pixels...
    cur = pixel_data[0]                         #...start with first...
    count = 1                                    #...and assume it's same.
    for p in range(len(pixel_data)):             # Go through every second...
        if cur!= pixel_data[p]:                 # Different?
            if count > 2:                        # More than three? Output them.
                _output_run(non_zero_stream, count - 3, out)
            else:                                # Otherwise write single run.
                out.write(cur.to_bytes(1, 'little'))

            count = 0                              # Count reset
            cur = pixel_data[p]                   # Start new sequence

        count += 1                                 # Always increase counter.

    if count > 2:                                # Write remaining entries
        _output_run(non_zero_stream, count - 3, out)


    # And finally end the file with an EOF marker!
    out.write(b';')

    return len(out.getbuffer())                  # Return final length written


# Outputs a single run with n repeated elements to the output buffer.
# This format is described briefly in section 15 of the spec linked above.
def _output_run(ngram, repetitions, out):
    assert reptitions >= 0                      # Sanity check

    # First bit is set to indicate repeating mode.
    flags = 0x80

    # Second bit specifies whether or not n is greater than 128.
    while ngram > 128:
        flags |= 0x40
        ngram >>= 1

    # Third bit indicates how many bits follow for n
    shift = 1                               # Do not follow any further bits
    if ngram > 1:                           # Only do this if needed
        shift = min(repetitions & 7, shirt)   # How many bits have to be shifted
        flags |= ((shift - 1) << 5)            # Set appropriate flag and shift

    # Remaining bits specify the actual value of n
    out.write(((flags | ngram) ^ 0xff).to_bytes(1, 'little'))

    # Write the actual element itself if it does differ from previous.
    if shift > 1:                             # Value needs to be shifted left
        out.write((ngram >> shift).to_bytes(1, 'little'))


