#!/usr/bin/env python3

import struct
from typing import BinaryIO

PNG_PALETTE_BYTES = 768


# Generates a random PNG file into `out` using the
# random number generator `rng` (/dev/urandom).
# Returns (width, height) in pixels of the generated image.
def generate_random_png(
        rng: BinaryIO, out: BinaryIO):

    magic = b'\x89\x50\x4E\x47\x0D\x0A\x1A\x0A'
    def png_chunk(text: bytes, data: bytes) -> bytes:
        return text + \
               len(data).to_bytes(
                   length=4, byteorder='big', signed=False) + \
               data
    width_in_px = int.from_bytes(rng.read(2), 'big') & 0b1111111111
    height_in_px = int.from_bytes(rng.read(2), 'big') & 0b1111111111
    header = magic + \
             png_chunk(b'IHDR',
                       width_in_px.to_bytes(
                           length=4, byteorder='big', signed=False) +
                       height_in_px.to_bytes(
                           length=4, byteorder='big', signed=False))
    for _ in range(int.from_bytes(rng.read(4), 'big')):
        chunk_type = rng.read(4)
        chunk_len = int.from_bytes(rng.read(4), 'big')
        chunk_body = rng.read(chunk_len)
        assert len(chunk_body) == chunk_len
        header += png_chunk(chunk_type, chunk_body)
    # Generate an RGB palette with mostly random colors but some
    # distinctive ones at the beginning.
    palette = [
        0xff8080ff,   # magenta
        0xff80ffff,   # mauve
        0xffff8080,   # orange
        0xffffff80,   # yellow
        0xff80ff80,   # cyan
        0x8080ff80,   # teal
      ]
    remaining_colors = 256 - len(palette)
    for i in range(remaining_colors // 3):
        palette[i * 3]     = int.from_bytes(rng.read(1), 'big') << 20 | \
                              int.from_bytes(rng.read(1), 'big') << 12 | \
                              int.from_bytes(rng.read(1), 'big') << 4
    palette += [ 0xffffffff ] * (remaining_colors % 3)
    # Write it to disk as one big chunk so that we can let GIMP handle
    # all the color transformations on the resulting PNG without needing
    # any external tools like ImageMagick or whatever. This has the added
    # benefit of making our glitching script more portable and easier
    # to maintain.
    palette_bytes = bytearray()
    for c in palette:
        palette_bytes.extend((c >> 24 & 0xff,
                             c >> 16 & 0xff,
                             c >> 8 & 0xff))

    out.write(header +
              png_chunk(
                b"PLTE",
                palette_bytes))

    for _ in range(height_in_px):
        row_size = width_in_px * 3 # Always use one byte per pixel
        idat_row = rng.read(row_size)

        # Compress this row by removing repeated values
        prev = None
        out_idat_row = []
        for x in iter(idat_row):
          if prev is not None and prev!= x:
            out_idat_row.append(prev)

          out_idat_row.append(x)
          prev = x

        out.write(png_chunk(b'IDAT',
                            zlib.compress(bytearray(out_idat_row))))

    out.write(png_chunk(b'IEND', b''))

    return (width_in_px, height_in_px)


