#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`, which can be /dev/urandom or otherwise
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # This is hard-coded to 16 frames of 25x25 pixels each, for a total file size of ~40k bytes
    frame_size = (25 * 25 * 3)
    nb_frames = 16
    img_descriptor_table_size = (nb_frames * 2)
    gct_size = 75
    lzw_mincodesize = 12   # Default minimum code size for LZW compression
    
    # Hard-coded screen descriptor
    display_width = 25
    display_height = 25
    packed_flags = 0b10    # Global color table present
    background_color = -1
    aspect_ratio = 0

    # Hard-coded global color table
    palette = [
        0xff, 0xff, 0xff,    # White
        0x00, 0x00, 0x00,    # Black
    ]

    # Critical number! Must match what's in the encoder. In practice it seems sufficient that this is a multiple of
    # frame_size so that we end up with an exact number of frames
    padded_frame_size = ((frame_size + 1) // 2) * 2
    
    # The first marker must be "Extension application" (kind=0xF9). It contains some necessary metadata and must be at the start of every image 
    extension_application_block = bytes([
        0x21, 0xf9,                  # Extension block identifier
        0x04,                        # Subtype: Graphic control label
        0x04 | 0b10000000,          # Flags: user input flag set; transparent color index flag set
        0x00, 0x01, 0x00,            # Delay time 1ms, disposal method 0 (no action); transaparency color index 0
        0x00                         # Block terminator
    ])

    # Start the file with the magic string
    out.write(b'GIF89a')

    # Write the logical screen descriptor
    out.write(struct.pack('>HHBBB', 
        display_width,     # Screen width
        display_height,    # Screen height
        packed_flags,      # Packed field flags
        background_color,  # Background color index
        aspect_ratio       # Aspect ratio
    ))

    # Write the global color table
    out.write(bytes(palette))

    # Next comes the local color table with exactly one entry - the "transparent colour", i.e. no transparency
    out.write(bytes([0]))

    # Next comes the loop control block. Note how we write a zero as the last delay value. We'll use the extension application block instead
    out.write(bytes([
        # Loop count
        0b00001000,         # Flag: do forever
        # Block terminator
        0x00                
    ]))

    # Now come the image descriptors. For now these are all empty (i.e. zero indexed colors) but they need to be there anyway
    for _ in range(nb_frames):
        out.write(bytes([
            0x2c,                    # Image separator
            0,                       # Left offset
            0,                       # Top offset
            display_width,           # Width
            display_height,          # Height
            0b10000000 | 0b00000001, # Packed field flags: local table flag set; interlace flag clear; sort flag clear
            0,                       # Local color table index
            0                        # Frame (image) data sub-block
        ]))
        
    # Finally, here comes the actual data
    out.write(extension_application_block)

    # For simplicity, we're just going to write zeros to fill up the rest of the file. Of course for real usage you'd want to generate something more interesting :)
    nb_empty_blocks = (padded_frame_size * nb_frames) + img_descriptor_table_size + gct_size + 1
    out.write(bytearray(padded_frame_size * nb_frames))
    