#!/usr/bin/env python3

import struct
import sys
from typing import BinaryIO

# Generates a random GIF file of size `size` bytes into `out`.
# Uses the random number generator `rng` to fill in the data.
def generate_random_gif(
        rng: BinaryIO, size: int, out: BinaryIO
    ) -> None:

    # Construct the header and write it to the output stream
    header = b'GIF89a'
    width = 1280
    height = 720
    colors = 256
    interlaced = False
    screen_width = (
            ((width % 256).to_bytes(1, 'big')) +
            (((width >> 8) & 255).to_bytes(1, 'big')))
    screen_height = (
            ((height % 256).to_bytes(1, 'big')) +
            (((height >> 8) & 255).to_bytes(1, 'big')))
    packed_bits = (
            64 |      # Global color table flag set
            32 |      # Color resolution bit is one
            7        # Sorting by decreasing intensity
            ).to_bytes(1, 'big')
    # We will use this later on as the
    # size indicator for the global color table
    color_table_size = colors * 3
    graphics_control = (
            0b00000000 |    # No transparancy given
            0 << 3          |  # Disposal method not specified
            1 << 2          |  # 1 second delay
            0                |   # Transparent color index field unused
            interlaced       |   # Boolean that indicates whether image
                                  # is interlaced
            0                |   # Reserved
            0                |   # Reserved
            ).to_bytes(1, 'big')
    lct_entry_size = 3
    gct_flag = 0b11000000
    local_color_table = (
            [gct_flag]         +     # Global color table used
            list(struct.pack('>H', colors)) + # Number of entries
            [lct_entry_size])             # Size of each color entry
    
    header_data = (
            header                  +  # Header string
            screen_width            +  # Screen width
            screen_height           +  # Screen height
            packed_bits             +  # Packed bits representing how
                                          # many colors are present
            graphics_control        +  # Graphics control extension block
            len(local_color_table).to_bytes(1, 'big') +
                                      # Length of graphic control field
            bytes(local_color_table) +  # Local (inline) color table
            bytes([0]))               # LZW minimum code size

    # Write the header including its length at the beginning.
    # The size of the header includes itself!
    out.write((len(header_data)).to_bytes(1, 'big')
              + header_data)

    bytes_left = size - len(header_data)
    current_chunk_size = 255
    while bytes_left > 0:
        if current_chunk_size >= bytes_left:
            current_chunk_size = bytes_left

        chunk_size = (current_chunk_size
                     .to_bytes(1, 'big'))
                    
        compressed_data = rng.read(
                current_chunk_size)
        
        raw_byte = zlib.compress(
                compressed_data)[2:-4]
    
        remove_filter_id = raw_byte[0] & ~(1<<1)
        filtered_image = raw_byte[1:]

        out.write(chunk_size
                  + bytes([remove_filter_id])
                  + filtered_image)

        bytes_left -= current_chunk_size
