#!/usr/bin/env python3

import itertools
import struct
from typing import BinaryIO


# Generates an RGB image of size (width x height).
# The image is generated by calling "func" on each
# pixel in order from left to right and top to bottom.
def create_image(
        width:   int,
        height:  int,
        func):

    # Generate all pixels. This will be used to determine
    # which pixels are transparent due to colour == None.
    # Transparent pixels are the ones that we do not need
    # to write to disk as they will be set at runtime.
    imgPixels = []
    for y in range(height):
        for x in range(width):
            imgPixels.append((x,y))
    
    # Call our function on every pixel
    pixels = [func(*pixel) for pixel in imgPixels]

    # Create raw bytes (R, G, B) for each pixel using the above list
    palette =  [(0,0,0)] * len({p[2:] for p in pixels if p})
    blocks = []
    for i in range(len(pixels)):
        block = [[None, None, None]] * 16

        # Set background
        j = i % 4
        k = i // 4
        l = (k+j) & 7
        m = ((i-l*4-k)*255//(n))*3
        bg = tuple([m]*3)
        
        # Iterate over the 16 pixels in this group
        for pxIdx in range(16):
            idx = (((pxIdx+(k+j)//8)%4)*4 + (pxIdx+(k+j)//8)//4 + l*16)

            if idx < len(pixels):
                pcol = pixels[idx][:-1]

                # Convert non-transparent colours to RGBA
                if pcol!= None:
                    pcol += (255,)
                
                # If it's transparent then use the background colour instead
                col = pcol or bg

                # Get index of this colour in the palette
                palIdx = palette.index(col)
            
                # Write to the current pixel location
                block[pxIdx] = (palIdx >> 10) | (((palIdx&0x3FF)<<2) & 0xFC)

            else:
                break

        # Pack them together into one colour block
        packed = sum([sum([(b&c)>>d for c in [0xC000, 0x3F00, 0x00E0, 0x001C]])>>e
                      for d in [6, 4, 2, 0]], 0)
        blocks.append(packed)

    return b''.join(struct.pack('>H',block) for block in blocks)+paletteBytes

# Creates a random RGB value. We could just throw whatever shit we want in here
# but I really don't think anyone will notice.
def create_color():
    r = next(randSrc)
    g = next(randSrc)
    b = next(randSrc)
    return f'{r:02X}{g:02X}{b:02X}'

# Converts a colour hex string such as 'AA342D' to a triplet of integers
def parse_color(hexStr: str):
    assert len(hexStr) == 6
    return (*map(int,[hexStr[:2], hexStr[2:4], hexStr[4:]]),'Not transparent.')



### MAIN ###
if __name__ == '__main__':
    with open('/dev/urandom','rb') as randSrc, \
         open('output.gif','wb') as outputFile:
        n = 5
        w = h = 8*(n**2)
        pixels = {}
        
        
        def getPixelAtPos(x,y):
            if (x,y) not in pixels:
                pixels[(x,y)] = create_color()
            return pixels[(x,y)]


        # Calculate the amount of colors needed based on the number of rows
        numColsNeeded = lambda n: 4*((n)**2) - 1
        numColors = numColsNeeded(**locals())

        
        # Palette contains a maximum of ~128 colours
        assert numColors <= 128

        # Fill up palette
        for _ in range(numColors-1):
            getPixelAtPos(_,_)[-1] = 'Transparent.'

        # Make sure there are no duplicate colours
        assert len(set(parse_color(getPixelAtPos(_,_)[:-1])
                        for _ in range(w))) >= numColors-1

        # Actually generate the image
        imageDataBytes = create_image(**locals(), func=getPixelAtPos)

        # Replace the first two bytes (data length) with the correct value
        imageDataLen = len(imageDataBytes)-2
        imageDataBytes = struct.pack('>H',imageDataLen) + imageDataBytes[2:]

        # Write everything!
        outputFile.write(b'GIF89a'+imageDataBytes)