#!/usr/bin/env python3

import math
import random
import struct


class GifHeader:

    def __init__(self):
        self._header_start = 6
        self._blocksize_min = 2049
        self._blocksize_max = (512 * 1024) - 1

        # Header itself is always 7 bytes long
        self.size()


    def size(self) -> int:
        return len(struct.pack('>ccBBB', b'G',
                b'I', b'F', b'8', b'9'))


    def pack(self, blocksize: int) -> bytes:
        return struct.pack('>ccBBBHHB', b'G',
                b'I', b'F', b'8', b'9',
                0x00, 0x00, blocksize)


    def _update_block_size(self, blocksize: int) -> None:
        assert self._blocksize_min <= blocksize <= \
                self._blocksize_max, 'Invalid block size'
        self._blocksize = blocksize


    @staticmethod
    def unpack_from(data: bytes) -> 'GifHeader':
        header = GifHeader()

        (magic, version_byte, width, height, flags,) = \
                struct.unpack_from('>ccBBBBB', data)

        if magic!= b'G' or version_byte not in [b'8', b'9']:
            raise ValueError('Invalid GIF')
        assert flags == 0x00, f'Reserved bits are set ({flags:#x})'

        header._update_block_size(width | ((height & 0xFF) << 8))

        return header



class LzwTable:

    # Maximum code length that can be used for codes
    _MAX_CODELEN = 12

    # This is different from the usual table implementation because we
    # don't need to store values anymore as they're already expressed by
    # their index in the dictionary list.
    class Codes:
        def __init__(self, codelen=None):
            self._codelen = min(
                    max(codelen, 2),
                    LzwTable._MAX_CODELEN)

            self._available_bits = 1 << self._codelen
            num_entries = 1 << self._codelen

            # Dictionary of strings and their associated code
            self._dictionary = [''] * num_entries

            # List of codes
            self._codes = [-1] * num_entries

            self._next_free_code = 2 ** self._codelen


        @property
        def next_free_code(self) -> int:
            return self._next_free_code


        def reset(self) -> None:
            self.__init__()


        def clear(self) -> None:
            self.__init__()


        def add(self, string: str) -> bool:
            # If the string is already present, do nothing
            if string in self._dictionary:
                return False

            # Get the index where the new entry will go
            idx = self._next_free_code
            assert idx < len(self._codes)

            # Add the string at this index
            self._dictionary[idx] = string

            # Keep track of how many bits were consumed in order to
            # encode this code
            numbits = len(string).bit_length()
            assert numbits >= self._codelen


            # The last bit indicates end-of-stream
            if numbits > self._codelen:
                numbits -= 1

                # Make sure the code was not too big
                assert idx >> numbits == 0
                self._codes[idx] = idx >> numbits
            else:
                self._codes[idx] = idx

            self._next_free_code += 1

            return True

        def get(self, value: int) -> str:
            assert 0 <= value <= (2**LzwTable._MAX_CODELEN)-1,\
                    f'{value} is outside valid range.'

            if value < self._available_bits:
                retval = ''
                while value > 0:
                    retval = ''.join([
                            self._dictionary[value],
                            retval])

                    # Extract lowest order bit until all order bits have been
                    # cleared from the integer.
                    value &= ~(value & 1)

            elif value == self._next_free_code:
                retval = ''
            else:
                # binary representation of current index
                binrepr = format(value, f'#0{LzwTable._MAX_CODELEN+2}b')[2:]

                # Compute the leading (prefix) code
                prefix = int(
                        binrepr[:LzwTable._MAX_CODELEN],
                        2)

                # Then compute the suffix based on the first "codeword"
                try:
                    suffix = self._dictionary[prefix]
                except IndexError:
                    print("BUG: prefix does not exist")
                    return ""


                # And append it to the rest of the sequence
                retval = suffix + self.get(prefix)

            return retval



    def __init__(self, stream: BinaryIO):

        # Read the minimum size of lzw code
        raw_val = ord(stream.read(1)[0])
        init_code_size = 2 ** (raw_val + 1)
        
        # Create both code tables
        self._cleartable = LzwTable.Codes(init_code_size)
        self._maintable = LzwTable.Codes()

    
    def decode(self, data: bytes) -> str:
        decoded_str = ''

        # First part of compressed data contains whether EOI marker exists
        eoimarker = chr(ord(data[0]) & 0x80)

        # Read remaining compressed data
        buf = data[1:]

        # Initialize starting code with the Clear code found in the byte
        # read above; prepended with its size.
        startcode = self._cleartable.next_free_code << \
                (8 - self._cleartable._codelen)

        currcnt = 0
        while currcnt < len(buf) - 1:

            currbyte = buf[currcnt]
            
            # -- Decoding stage ---
            # Get code representing the symbol being decoded
            code = ord(currbyte)
            currcnt += 1
            
            # Check if next byte is a temporary code used to signal an
            # EOI marker has been reached.
            if ord(eoimarker) == (code & 0x80):
                break

            # Optimize process when there's enough space left in
            # maintable
            if self._maintable.next_free_code <= 4096:
                self._optimize_maintable()

            if code == 0xff:
                # Start-Of-Stream marker
                self._maintable.reset()
                self._cleartable.clear()
                
                continue

            # Attempt to retrieve the corresponding string from the maintable.
            symbl = self._maintable.get(code)

            if symbl == '':
                # Attempt to retrieve the corresponding string from the
                # cleartable.
                symbl = self._cleartable.get(code)

            if symbl == '':
                # There is no such string in either dictionary, which means
                # that the next two bytes must be considered together to form
                # a new one. When doing so, make sure to prepend the formerly
                # retrieved one to achieve a full four bytes word.
                tmpcode = (code << 8) | ord(buf[currcnt])
                
                # Add newly decoded character to the maintable before
                # proceeding further.
                added = self._maintable.add(decoded_str)
                if not added:
                    symbl = decoded_str
                else:
                    symbl = ''.join((decoded_str, symbl))
                
                # Get the actual string corresponding to this composite code
                decoded_str = self._maintable.get(tmpcode)
                
                # Move pointer forward since a new code has just been read
                currcnt += 1
            else:
                pass

            # Add decoded subsequence to output buffer
            decoded_str = ''.join([symbl, decoded_str])

        return decoded_str
    

    def _optimize_maintable(self) -> None:
        """Optimize the maintable by clearing entries whose prefix code cannot
           possibly lead to any other possible code."""
        # For every existing code...
        for code in range(self._maintable.next_free_code):
            #...compute its prefix code
            prefix = code >> (self._maintable._codelen - 8)
        
            #..and check that it is actually within the scope of the
            # cleartable. It may happen that some prefixes have been deleted
            # because they never led to another valid code, but the original
            # code still belongs to them. In which case, the original code
            # must also be removed from the dict and maintained as a special
            # one.
            if prefix in self._cleartable._codes:
                continue

            # Okay! We've found a candidate for deletion. Delete it
            self._maintable._codes[code] |= (~0)

        

def lzw_decode(stream: BinaryIO) -> str:
    gifhdr = GifHeader.unpack_from(stream.read())
    
    # Unpack the image data from the gif
    imgdat = stream.read()

    # Instantiate the lzw decoder
    decoder = LzwTable(imgdat)

    # Decode the image data into a textual sequence
    textseq = decoder