#!/usr/bin/env python3

import struct
import sys


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the initial 6 bytes of the GIF header (magic and version)
    # Then read the logical screen descriptor to determine the size
    # of other parts such as color tables and image data
    out.write(b'GIF89a')
    out.write(rng.read(7))
    width = int.from_bytes(rng.read(2), 'big')
    height = int.from_bytes(rng.read(2), 'big')
    global_color_table = bool(int.from_bytes(rng.read(1), 'big'))
    background_color = int.from_bytes(rng.read(1), 'big')
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big')
    print("width:", width)
    print("height:", height)
    print("global table?", global_color_table)
    print("background", background_color)
    print("aspect ratio:", pixel_aspect_ratio)
    if not global_color_table:
        raise NotImplementedError('Local color tables are not supported yet :(')

    # Read the global color table to determine the number of colors
    num_colors = int.from_bytes(rng.read(1), 'big') * 256
    for _ in range(num_colors):
        out.write(rng.read(3))

    # Add an application extension that has a label of "NETSCAPE" and a block size of 3
    # The contents of this block is a string containing two 16-bit integers:
    #   1. The loop flag - set to 0
    #   2. How many times the animation should repeat after completion
    #      This value is ignored by all viewers I've seen so far except GIMP
    out.write(b'\x21\xFF\x0B')
    out.write(b'NETSCAPE2.0')
    out.write(struct.pack('<H', 3))
    out.write(struct.pack('<H', 1))
    out.write(b';')

    # Write the image descriptor
    out.write(b'\x2C')       # Start of Image Descriptor Block
    out.write(b'\x00')       # Left Position
    out.write(b'\x00')       # Top Position
    out.write(struct.pack('<HH', width, height))   # Width & Height
    local_color_table = True
    interlace = False        # No support for interlacing yet
    lct_size = int.from_bytes(rng.read(1), 'big') if local_color_table else 0
    lct = [rng.read(3) if lct_size!= 0 else '' for i in range(lct_size)]
    print("local color table?", local_color_table)
    print("interlace?", interlace)
    print("LCT size:", lct_size)
    print("LCT entries:")
    for c in lct: print("\t", c)

    # Write the LZW minimum code size
    mincode_size = int.from_bytes(rng.read(1), 'big')
    print("Minimum code size:", mincode_size)

    # Now we're done with headers! We can actually start writing pixels.
    # First, write the first byte indicating the block size, then the actual content of
    # each pixel sequentially until the entire area has been filled
    num_pixels = width*height
    max_blocksize = 2 ** mincode_size
    while num_pixels > 0:
        block_size = max((max_blocksize,
                          min((num_pixels + 2**mincode_size - 1)//(2**mincode_size),
                               max_blocksize)))

        # The first byte of each sequenced block contains the following information:
        #   1. Block Size          - Indicates how many bytes follow this byte
        #     NOTE: There's no reason why this must be less than or equal to the maximum
        #           block size specified at the beginning of the frame
        #   2. Extension Flag      - Indicates whether there will be an extra byte before
        #                             the next pixel block starts indicating its type
        #   3. User Input Flag     - Currently unused - must always be zero
        #   4. Disposal Method     - Specifies what the previous display is used for when
        #                            displaying the current frame
        #   5. Transparent Color   - Indicates which color index is treated as transparent
        #   6. Interlace Flag      - Whether the image is interlaced
        subblock_size = block_size // 256 + (block_size % 256!= 0)
        num_subblocks = block_size // 256
        try:
            assert(num_subblocks <= 2**subblock_size)
        except AssertionError:
            exit("too many subblocks")
        out.write(struct.pack('>BBBBB',
                             block_size |
                             ((1 << 5) if num_subblocks >= 256 else 0),
                             0 if num_subblocks == 1 else 1,
                             0,              # User input flag
                             3,              # Disposal method: leave unchanged
                             0               # Transparency index: none
                ))


        # We need to keep track of how many bytes have been written already since one of them
        # should contain the length of the remaining block. This could cause us to exceed our limit
        # even though it shouldn't because this calculation doesn't take into account the overhead
        # from adding padding bytes, etc. However, the resulting file cannot be loaded into any
        # common GIF viewer anyway...
        num_written = 0
        for i in range(num_subblocks):
            out.write(struct.pack('B', i+1))

            # The second byte of every pair indicates its length
            nbits = (((i+1)*8//num_subblocks)
                     if i < num_subblocks - 1 else
                     (8-(i+1)*8//num_subblocks))
            out.write(struct.pack('>B', nbits+(nbits%8!=0)))
            num_written += 2

        # Finally, the rest of the block consists of up to 256 palette indices specifying what
        # color each pixel in the subblock should be
        for j in range(((block_size//2)-1)//8 + 1):
            bits = [(rng.read(1)[0] >> k)&1
                    for k in reversed([j*8 + t
                                    for t in range(8)])]
            out.write(bytearray(bits).ljust((((block_size//2)-1) % 8)+1,
                                            b'\0').rstrip(b'\0'))
            num_written += len(bits)
        assert num_written == block_size

        num_pixels -= block_size
