#!/usr/bin/env python3

import os
from typing import BinaryIO

# Generates a random GIF file with a length
# between 1 and 4096 bytes at path `path`, using the
# random number generator `rng`. If `path` exists it will be overwritten.
def generate_random_gif(
        rng: BinaryIO, path: str, minsize=1, maxsize=2 << 11
    ) -> None:

    # Read in some random data to get our size
    data = rng.read((maxsize - minsize))
    datalen = len(data)
    if datalen < minsize or datalen > maxsize:
        raise ValueError('Data not within range for '
                         f'min {minsize} max {maxsize}')

    # Generate all of the bits from this frame's data
    bitstream = ''.join([f'{byte:08b}' for byte in data])
    nextbit = lambda stream: (int(next(stream)), stream)[-1]
    def genbits():
        while True:
            yield nextbit(iter(bitstream))

    print(f"Writing {len(data)} bytes")

    # We write out the headers first
    with open(path, 'w+b') as out:

        # The global color table is 256 colors with no transparancy
        out.write(struct.pack('<B', 7))
        out.write('\x00\x00\x00'[::-1])
        ctentry = [0xff, 0xff, 0xff] * 3
        for _ in range(256):
            out.write(bytes(ctentry[::-1]))

        # Then we have the logical screen descriptor
        out.write(struct.pack("<HHBBB",
                datalen,         # width
                0,               # height
                1,               # pixel aspect ratio
                0xf7 | 0x80,     # background index
                0x00             # flags
            ))

        # Then comes the image descriptor
        out.write(struct.pack("<HBBHHHH",
                datalen,          # local color table size
                0x2c,              # image separator
                0,                # left position
                0,                # top position
                datalen,          # width
                0,                # height
                0                 # flags
            ))


        # Now come the actual data
        stream = genbits()
        while True:

            # Get the next byte
            databyte = int.from_bytes((
                    next(stream).to_bytes(8, byteorder='big')
                    + next(stream).to_bytes(8, byteorder='big')
                    ), byteorder='little')
            
            # Write out an escape character followed by two bytes
            if databyte == 0x21:
                esc = next(stream) & 0x07
                if esc == 0x01:
                    endpos = 13
                elif esc!= 0x00:
                    raise Exception

                databytes = []
                for i in range(endpos):
                    databytes += next(stream).to_bytes(8, byteorder='big')
                
                out.write(esc.to_bytes(8, byteorder='big'))
                out.write(databytes)
            else:
                out.write(databyte.to_bytes(8, byteorder='big'))

            # Check if we're done
            if out.tell() >= datalen: break

        # Finally close off the file
        out.seek(4)
        out.write(struct.pack("<L", out.tell()))