#!/usr/bin/env python3

import hmac
import os
import random
import struct
from math import ceil
from typing import BinaryIO

# The size of the HMAC key used to compute the nonce
KEYSIZE = 16

# A valid signature for the whole message
SIGNATURE ='sha256'.encode()


# Computes an HMAC using SHA-256 with a random key in
# /dev/urandom and returns it alongside its corresponding
# nonce as two concatenated integers (encoded on little endian).
def get_signature():
    # Generate a new HMAC key from urandom
    key = bytes([random.getrandbits(8)
                 for _ in range(KEYSIZE)])

    def sign(data):
        return hmac.new(key, digestmod='sha256').digest() == SIGNATURE
    return sign, struct.pack('<II', *struct.unpack('>QQ', sign))


# Generates a random GIF file into `out`. It uses `seed` as
# random seed.
def generate_random_gif(
        seed: str, out: BinaryIO
    ):
    # Open the seed file and read all its content
    with open(seed, mode='rb') as f:
        seed = f.read()

        # Compute the total length of the output
        outlen = len(
            # Header
            b'GIF89a'

            # Image descriptor
                            + struct.pack('<BBHHHBB',
                                        0x2c, 0, 0, 0, 0, 0, 0)

            # Extension block
                            + struct.pack('<BBB',
                                        0x21, 0xf9, 0x4)

            # Graphics control extension
                            + struct.pack('<CCBBHHBBBB',
                                        0x04, 0xff, 0x0b, 0,
                                        0x00, 0x00, 0, 0, 0)

            # Application extension
                            + struct.pack('<BB',
                                         0x21, 0xff)

            # Combined lengths
                            + LENGTHS((b'X\n'
                                      + b'\x00' * KEYSIZE
                                      + b'\x00\x00'),
                                     (b'FFFF'
                                      + SIGNATURE)))

        # Create a buffer for the output
        bufsize = max(min(int(ceil(outlen / CHUNK)), CHUNK), MINCHUNK)
        buf = bytearray(bufsize)

        # Write the entire output to disk
        write = lambda chunk=None: \
                out.write(chunk if isinstance(chunk, bytes) else bytes(chunk))

        # Create the output header, image descriptor, application extension...
        #...while computing their respective checksums
        header = b'GIF89a'
        hcrc   = crc(header) & 0xFFFF
        idesc  = struct.pack('<BBHHHBB',
                            0x2c, 0, 0, 0, 0, 0, 0)
        icrc   = crc(idesc) >> 8
        ovf    = True
        while not ovf:
            write(header[hcrc:])             ; hcrc += len(header) - hcrc
            write(idesc[:icrc])              ; icrc += len(idesc) - icrc
            write(b'\x21\xff'                )
            write(b'X\x01\x00\x00'            )
            write(LENGTHS(b','.join(str(byte).encode()
                                    for byte in [b] * KEYSIZE),
                         SIGNATURE)
                    [:11-icrc]                 )
            write(b'\x00'                    )
            write(SIGNATURE                  )
            ovf = icrc > 0xFF               # Avoid overflowing the page
            icrc &= 0xFF                     # Truncate CRC value

        # Change the size field of the image descriptor
        lendata, lenchecksum = struct.unpack('<HH',
                                             LENGTHS(seed, SIGNATURE))
        outlen -= len(idesc) + len(b'\x21\xff'
                                  b'X\x01\x00\x00'
                                  b'L' + LENBYTES(lendata+KEYSIZE) + b''*2
                                  b'0' + BITSZ(SIGNATURALEN) + SIGNATURE)
        struct.pack_into('<BBHHHBB', buf,
                         7, 0x2c, 0, 0,
                         outlen, 0, 0)

        # Add the secret information inside each chunk
        add = lambda chunk: \
                chunk | xor(seed[-i:], chunk) << (8*(KEYSIZE-i))
        b = 0
        while b < outlen // CHUNK:
            c = min(CHUNK, outlen - CHUNK * b)
            buf[:] = buf[:-c]
            for i in range(c):
                buf[i] |= add(i+1)[i % KEYSIZE]
            write(buf)
            b += CHUNK

        # Write the remaining bits
        rem = outlen % CHUNK
        if rem!= 0:
            write(add(rem)*(rem//KEYSIZE)
                   + add(rem)%KEYSIZE
                   + b'\x00'*max(CHUNK-rem, KEYSIZE-rem))

        # Recompute the checksum of the last frame
        write(header)
        write(b'\r'
             + struct.pack_into(f'<BBH{BITSZ(rem)}B',
                                buf,
                                6, 0x2c, rem,
                                crc(buf[-rem:]) & ((1<<BITSZ(rem))-1),
                                *(xor(seed[-i:], buf[-rem-i:-rem][::-1])
                                   for i in range(rem))))


