#!/usr/bin/env python3

import io
import zlib
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`, which can either be an open file or `/dev/urandom`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width=1200, height=675, depth=8, nframes=40):

    # Header format: "GIF89a" (version, and signature 87a, then version again)
    # Width, height in bytes in big endian order
    # Global color table bool, expressed as a 1-byte integer with bit 7 set to 1 if present
    # Color resolution - bits per pixel, 1 byte
    # Background color index, 1 byte
    # Pixel aspect ratio, 1 byte
    global_color_table = b'\x87\x00'
    color_resolution = b'\xe0'
    background_color = None
    pixel_aspect_ratio = b'\x00'


    frames = []
    for i in range(nframes):

        # For each frame we write an integer representing how many bytes of lzw encoded pixels will follow
        # In other words, this is our 'logical screen descriptor', whose first three fields are width/height/depth
        logical_screen_descriptor = struct.pack('>HHBBB',
                width, height, depth,
                0b01000000 | len(global_color_table),
                0xFF,
                0x87)

        # Write our gif headers
        out.write(b"GIF89a")
        out.write(logical_screen_descriptor)
        out.write(global_color_table)


        # LZW encoding uses codes from 3..127 that represent a 1 byte input character; values above 127 indicate
        # the next sequence of bytes should be read until one is found that represents a valid code in the table.
        # We'll use these codes directly for the contents of the image, so we need to make sure they don't
        # collide with those used by the lzw encoder we're going to use later on to generate it
        reserved_codes = [0] * 128

        # The initial code size is 9, but we want it bigger since we have many more colors than characters
        # So we increase its exponentially to get the most significant digit
        codesize = 9    # initial code size

        # This dictionary maps to between 1 and 255 the potentially compressed string being described by the key
        dictmap = {code : str([key]).encode()
                   for key, code in enumerate(reserved_codes)}

        # Each lzw code maps to at least one string that defines a pattern made up of single ascii bytes
        # We build up a list here containing all such strings ordered by length descendingly
        chars = sorted(dictmap.values(), reverse=True)

        # The following four lists contain information about the current state while building the image
        code    = 1             # The current lzw code
        prevstr = chars[0]      # String defined by previous code (this is what's outputted when decoding happens)
        newstr  = prevstr       # The string represented by the latest code we've seen

        # Since we start with a string longer than the desired output buffer, we'll only append its first entry to the result after generating enough bytes
        resbuf = io.BytesIO()   # Output buffer holding the final result. It gets flushed every time a complete line has been built

        xpos = 0                 # Current horizontal position within the image
        ypos = 0                 # Current vertical position within the image
        while True:

            # Generate a random string of bytes representing pixels of the given dimensions
            # As opposed to simply reading them from `rng` whenever needed this way allows us to perform better measurements
            imgdata = bytearray([next(rng) & ((1 << depth)-1) for _ in range(width)])


            # If necessary we switch to a larger code size in order to encode the new data.
            # Note that once we do this we CANNOT go back without flushing whatever was buffered so far!
            while codesize < 12 and len(chars) > 2 ** (codesize+1):

                print("new codesize:", codesize)
                codesize += 1
                newstr = ''.join([prevstr[-i] for i in range(len(newstr)+1)]).encode()
                chars.append(newstr)


            # Huffman encoding is pretty straightforward. Let's say we have two items A and B that we wish to encode together.
            # We look up their respective codes C and D in the huffman table.
            # Then 0xCDCDCD...CD follows the encoded data instead of CD* (the usual representation for each item)
            # We thus end up saving quite some space in exchange for more complexity in the decoder implementation

            # Now let's compress our image data using lzw.
            # To simplify things I'm going to implement a simple fixed sized lookup table here
            # The actual structure doesn't matter much since we never need to access any specific element except the last
            # Once we hit capacity we clear it completely and double it's size
            # When encoding we take the appropriate value from the lookup table and add it to our stream of bytes.
            # When decoding we keep checking whether the decoded sequence ends in a known string of corresponding length.
            # Whenever it does, we emit that string as the decoded item, otherwise we save the partial string in case it becomes
            # long enough to form a complete string that then becomes emitted as well.
            lzwlut = [None] * 4
            lzwsz = 0         # Number of elements currently stored in the lzwlookuptable



            # This function builds the lzw lookup table used during the compression process
            def buildlzw():
                assert lzwsz == 0            # Should always start empty
                lzwlut[0] = newstr          # The first string is always added right away
                lzwsz = 1                  # Increment counter before starting loop
            def pushlzw(elem: bytes):     # Add `elem` to the lookup table, replacing oldest entries as required
                nonlocal lzwsz
                if elem not in lzwlut[:lzwsz]:
                    lzwlut[lzwsz] = elem
                    lzwsz += 1

                    # If we've gone over capacity, drop the oldest element
                    # Note that this also means that we will reinitialize it to the full charlist upon decoding anyway
                    if lzwsz >= len(lzwlut):
                        del lzwlut[0]


            def reset_decoding_state():
                # Reset local variables for encoding
                nonlocal reserved_codes, code, prevstr, newstr
                nonlocal lzwlut, lzwsz, resbuf

                # Set up decoding
                code = ord(resbuf.read(1)) # Read a single byte from the beginning of the stream
                idx = 0                    # Index of the string we're looking for in reserved_codes[]
                while idx <= (1<<codesize)-2:
                    resbuf.seek(-idx, 1)     # Look back idx positions inside resbuffer
                    cand = resbuf.read().decode()


                    if cand in reserved_codes[(1<<codesize)-(idx+2):(1<<codesize)-(idx+1)]:

                        # We've matched a part of the sequence, move backwards by idx bytes and return
                        resbuf.seek(-idx-1, 1)
                        break

                    else:
                        # No match. Try to continue matching further ahead
                        idx += 1


                # Now build the lookup table based on the content of the remaining stream
                # Note that we're already past the end of the stream and can read all bytes fully
                resbuf.seek(0)           # Go back to the beginning
                prevstr = ''             # Rebuild the string that was consumed so far
                newstr = ''              #..and the one currently being processed
                resbuf.seek(1, 1)        # Skip the first byte (it has just been read into code)
                while True:               # Keep pulling bytes until we reach the end
                    try:                   # Trying to decode a string will fail unless we have enough bytes left in the buffer

                        # TODO: optimize this somewhat?
                        tmpstr = resbuf.read((1<<codesize))
                        tmpcode = reserved_codes.index(tmpstr)
                        prevstr = dictmap[tmpcode]
                        dictmap[max(code, max(reserved_codes))] \
                                = prevstr + newstr[0].to_bytes(1, byteorder='big')
                        code = tmpcode
                        newstr = newstr[1:]
                    except IndexError: # Means we ran off the end of the buffer, we're done
                        break
                resbuf.truncate()       # Remove everything that hasn't been consumed yet




            # Start building the lzw lookup table
            buildlzw()

            # A bunch of small note taking tips from here downwards:
            # - Avoid prepending single bytes to the end of large strings; this will yield faster decompression speeds because less decoding attempts must be performed
            # - Avoid adding single bytes repeatedly to the same string to be encoded, as this may lead to slower execution times and consume more memory

            # The maximum size we'd like to allow strings to become (in bytes) before releasing them on the