#!/usr/bin/env python3

import struct
from functools import partial
from io import BytesIO

class RandomGifGenerator():

    def __init__(self):

        # The current position in the output buffer for writing pixels
        self._pixelpos = 0

        # Constants used to generate the color palette and pixel stream
        self._globalcolormap = [
                (255, 255, 255),    # white
                (  0,   0,   0),    # black
                (164,  71, 111),    # purple
                ( 66,  51,  96),    # dark teal
                (238, 179, 220),    # pink
                (227, 199, 115),    # yellow-orange
                (129,  58, 187),    # blue
                (  0, 115, 165),    # bright blue
                ]

        # This is what we use as the color replacement table
        self._colortable = list(range(len(self._globalcolormap)))

        # Keep track of where all colors have been replaced with each other
        # so that we can add more if necessary
        self._colorreplacements = []

        self._framebuffersize = len(self._globalcolormap)*3 * \
                                512*512//8

        self._framebuf = None


    def _generate_header(self):
        return b'GIF89a'

    def _generate_logical_screen_descriptor(self):
        packed = struct.pack('>HHBBB',
                             512,
                             512,
                             8,
                             7+3*len(self._globalcolormap))
        
        return packed

    def _generate_colormap(self):
        colormapdata = bytes()

        for r, g, b in self._globalcolormap:
            rgb = struct.pack('B'*4, r, g, b, 0)

            colormapdata += rgb

        colormapdata += bytes([0])

        return colormapdata

    def _generate_extension(self):
        extensiontype = b'\x21\xf9\x04'
        blocksize = b'\x0b'
        loopcount = b'\xff\xff'         # forever!
        terminator = b'\x00'

        extdata = extensiontype + blocksize + loopcount + terminator

        return extdata

    def _get_next_color(self):
        try:
            return self._colortable[self._pixelpos]
        except IndexError:
            # We don't have enough colors yet; create some new ones from our
            # existing ones by modifying their luminosity values and then 
            # pushing them onto our global color map

            # Add a new entry to our color replacement table
            self._add_new_color_replacements()

            # Get the next one in line
            index = self._colortable[self._pixelpos % len(self._colortable)]
            
            # Apply the color modifications on it based off of the 
            # original color value stored at that index in the global
            # color map
            color = self._globalcolormap[index][:]

            c = ((color[0]*0.75)+192)//2
            color[0] = max(min(c, 255), 0)

            c = ((color[1]*0.75)+192)//2
            color[1] = max(min(c, 255), 0)

            c = ((color[2]*0.75)+192)//2
            color[2] = max(min(c, 255), 0)

            return self._apply_color_modifications(tuple(color))
            

    def _add_new_color_replacements(self):
        """Adds new entries into the color replacement table."""

        tablelen = len(self._colortable)

        # All combinations of these two items will be created and 
        # pushed onto the end of the array whenever too many colors are needed
        permtable = [(i, j) for i in range(tablelen) 
                        for j in range(tablelen)
                        if i!= j]


        self._colorreplacements += permtable[:]

        # If we still need more, recurse until we get them all
        if len(self._colortable) < 2**16:
            self._add_new_color_replacements()

    def _apply_color_modifications(self, color: tuple):

        # Find the index into our global color map which corresponds to this color
        try:
            index = self._globalcolormap.index(color)
        except ValueError:
            raise RuntimeError("Color not in global color map")

        while True:
            # Try replacing the color with any other color within the same
            # section of the color map (e.g. red -> green or green ->purple).
            # Each time an item has been used once, it's removed from consideration.
            # If no such color exists after removing the used color from consideration,
            # we just randomize it instead

            # Choose a random element from our permutation table,
            # then remove it so we don't revisit it again
            r = randint(0, len(permtable)-1)
            pair = self._colorreplacements.pop(r)

            # Generate the color corresponding to those indices
            newcolor = self._globalcolormap[pair[0]] + \
                       self._globalcolormap[pair[1]]

            # Replace the current color with the newly generated one
            self._globalcolormap[index] = newcolor // 2

            # Recalculate the hash value of the modified color
            currenthash = generate_color_hash(*newcolor)

            # Try only replacing colors that share the same hash code
            # as the current one...
            if currenthash == generate_color_hash(*color):
                break

        return index
        

    def _generate_datastream(self):
        framebuf = BitArray(self._framebuf)
        colorbuf = BitStream()

        for i in range((512*512)):
            # Pack the bits of the image buffer into color indexes
            buf = framebuf[i*16:(i+1)*16]
            index = self._get_next_color()

            colorbuf.append(buf, length=3)

            # Write the index into the color table earlier
            # TODO: Do something about this?
            print(f'{index:X}')

        return colorbuf



    @staticmethod
    def generate_random_gif(rng: BinaryIO,
                            filename: str):

        gen = RandomGifGenerator()

        with open(filename, 'wb+') as fd:

            fd.write(gen._generate_header())
            fd.write(gen._generate_logical_screen_descriptor())
            fd.write(gen._generate_colormap())

            # Generate the animation frames
            for f in range(1<<16):

                # Reset everything ready to write the next frame
                gen._pixelpos = 0

                # Populate the buffer with random noise
                gen._framebuf = rng.read(gen._framebuffersize) 

                fd.write(gen._generate_extension())
                fd.write(gen._generate_datastream().tobytes())

