#!/usr/bin/env python3

import math
import struct
from os import urandom
from typing import BinaryIO


# Generate an integer uniformly distributed in [0..max]
def gen_int(max):
    n = 256*256*math.ceil((max+1)/256/256)
    while True:
        yield struct.unpack('<I', urandom(4))[0] % (n-1)


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, max=2**16
    ):

    # Header
    out.write(b'GIF89a')

    # Logical Screen Descriptor
    width = gen_int(max)
    height = gen_int(max)
    xpixres = gen_int(max)
    ypixres = gen_int(max)
    gctsize = next(gen_int(7)) | (next(gen_int(2))+1)*8
    bgindex = next(gen_int(gctsize)-1) if gctsize else None
    pixelaspectratio = next(gen_int(max))
    out.write(struct.pack('>HHBBBHBB',
             *width(), *height(), *xpixres(), *ypixres(),
             gctsize, bgindex or 0, pixelaspectratio) )

    # Global Color Table
    for i in range(2**(gctsize+1)):
        r = next(gen_int(max))
        g = next(gen_int(max))
        b = next(gen_int(max))
        out.write(struct.pack('BBB',r,g,b))


    # Loop to generate frames
    numframes = gen_int(min(len(list(gen_int(max))), len(list(gen_int(max)))))
    delaytime = gen_int(max)
    loopcount = next(gen_int(max))
    for f in range(numframes):

        # Image Descriptor
        left = gen_int(max)
        top = gen_int(max)
        width = gen_int(max)
        height = gen_int(max)
        flags = next(gen_int(max)) & ~1        # No local color table
        imginterlaced = bool(flags&4)          # Not interlaceable by default
        imgsortedcolormap = not bool(flags&2)   # Assumed sorted colormap by default
        out.write(struct.pack('>HHHHB',
                 *left(), *top(), *width(), *height(), flags) )

        # LZW Codes (only one block for now)
        lzwmincodesize = min(8, next(gen_int(max)))
        lzwxfersize = next(gen_int(max))
        for codelens in [2,3,4]:
            out.write(struct.pack('BBBB', codelens-2, *gen_int(2**(codelens+1))))
        for _ in range(lzwxfersize):
            out.write(struct.pack('B', next(gen_int(2**lzwmincodesize-1))))

        # Extension Block
        exttype = next(gen_int(max))
        extlabel = chr(exttype).encode()
        extnamespace = extlabel
        extblocksize = next(gen_int(max)) - 1
        if exttype == 255 and extblocksize > 7:       # Application Extention
            appid = extlabel + bytes(extblocksize-7)
            appauthcode = extlabel + bytes(extblocksize-7)
            out.write(struct.pack('BB%ds%ds' % (len(appid), len(appauthcode)),
                    exttype, extblocksize, appid, appauthcode ))
        elif exttype == 218 and extblocksize >= 3:    # Comment Extention
            out.write(struct.pack('BB%ds' % extblocksize,
                    exttype, extblocksize, extlabel ))
            comblocksize = next(gen_int(max)) - 1
            out.write(comblocksize * (next(gen_int(256))*256+next(gen_int(256)))
                       .to_bytes(2,'little'))
        elif exttype == 0x0b and extblocksize >= 1:     # Graphic Control Extention
            delaytime = next(delaytime)+1              # Delay time is always non zero
            transindex = next(gen_int(bgindex))      # Always valid index
            transparams = [next(gen_int(max)) for _ in range(3)]
            disposalmethod = next(gen_int(max))
            userinputflag = next(gen_int(2))         # User input is assumed disabled
            out.write(struct.pack('BBBBBBB',
                             exttype, extblocksize,
                             disposalmethod, userinputflag,
                             transparams[0], transparams[1], transparams[2]))

            # Optional repeats extension
            if extblocksize>=3:
                repeatcount = next(gen_int(max))
                out.write(struct.pack('B',repeatcount))

        elif exttype == 0xfe and extblocksize==0:     # Plain Text Extention
            textleft = gen_int(left()+2)
            texttop = gen_int(top()+2)
            textwidth = gen_int(width()+2)
            textht = gen_int(height()+2)
            cellwidth = next(gen_int(max))
            celltheight = next(gen_int(max))
            textfgindex = next(gen_int(gctsize))
            textbgindex = next(gen_int(textfgindex))
            opcolor = ([next(gen_int(256)) for _ in range(3)]
                      + [next(gen_int(gctsize))])
            txtcelldata = [(next(gen_int(max))-1
                           for x in range(celltwidth)])
            out.write(struct.pack('BBHHHHHB',
                     exttype, extblocksize, *textleft(), *texttop(),
                     *textwidth(), *textht(),
                     cellwidth, celltheight, textfgindex, textbgindex))
            out.write(struct.pack('BBB',opcolor[0],opcolor[1],opcolor[2]))
            for row in range(textht()):
                for cell in range(textwidth()):
                    out.write(txtcelldata[row][cell].to_bytes(1,'little'))


        # Graphics Control Extention block terminator
        out.write(struct.pack('B',0))


    # Trailer
    out.write(b';')