#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

def bytes_to_int(data):
    return sum([x << (i * 8) for i, x in enumerate(bytearray(data))])


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header format (big endian) is as follows:
    #   [Signature]        A static string of "GIF" or "GIF87a" to identify this file type. This field must be exactly 6 bytes long;
    #                      if it isn't, this file cannot be displayed and gifsicle should report an error. The signature may optionally contain "89a", which means that the file complies with Version 89a of the specification.
    #   [Version]           Indicates what version of the GIF specification is being used. All newer versions are backwards compatible with Version 89a (and thus any conforming GIF reader can read them). Older readers will still work without problems on older files, but it's possible they won't understand all the new features of a future standard. So you should set your writer to output Version 89a unless there really is a reason not to do so — e.g., when your program uses non-standard extensions. The value should be stored as one byte where 0x87 indicates Version 87a and anything higher than zero indicates Version 89a.
    #   LogicalScreenWidth  Width of the entire image in pixels, minus 1. If the width is greater than 65535 pixels, this should be specified as 65535. For example, if the logical screen width is 40000 pixels wide, this field should contain 39999.
    #   LogicalScreenHeight Height of the entire image in pixels, minus 1. If the height is greater than 65535 pixels, this should be specified as 65535. Note that even though most screens are 200 lines high, we recommend storing the height as 200 rather than 201 because many graphics editors count from 0 rather than 1. For example, if the logical screen height is 40000 pixels tall, this field should contain 39999.
    #   GCTFlag             Global Color Table Flag. If set to true, specifies that there is a Global Color Table immediately following this sub-block. Otherwise, the Logical Screen contains no Global Color Table.
    #   ColorResolution     Bits per primary color available. Since each pixel in the Logical Screen Descriptor contains three fields which specify the red, green, and blue intensities, the total number of bits required to specify each color is determined by multiplying the number of bits per primary component by 3. Thus, ColorResolution=15 indicates that the total number of bits for each RGB triplet is 48 bits, or six bytes. Legal values are between 1 and 8 inclusive, although some viewers may have limits less than this. Usually these limits lie somewhere near 5 bits per component.
    #   SortFlag            Determines the way the local color table should be sorted. When false, the table should be sorted least significant bit first. When true, the table should be sorted most significant bit first.
    #   SizeOfGlobalColorTable	Size of the Global Color Table. Only present if GCTFlag is true. Should be a multiple of 3 because the size is measured in bytes. Legal values are between 2 and 768 inclusive. Some viewers may allow larger sizes, however, and it is recommended to use the largest legal value whenever possible.
    #   BackgroundColorIndex Index number of the color in the Global Color Table to be used as the background color for the entire Logical Screen. Value should be in range 0..(2^(SizeOfGlobalColorTable)-1) - 1.
    #   PixelAspectRatio    Aspect ratio of the smallest unit in a pixel. Every pixel is made up of identical blocks called pellets. Each pellet may contain several different colors, depending upon how many colors are needed to represent the colors in the source image. In order to reduce storage requirements and improve performance, picture elements containing identical colors should be combined into longer runs before compression. To determine the dimensions of such a run, a Pellet Size field value of n+1 refers to a block consisting of n consecutive pellets in the same location.
    #   GlobalColorTable     24-bit Global Color Table. If present, then SizeOfGlobalColorTable contains its value in bytes. Within the Global Color Table, each entry consists of one byte for the Blue intensity, followed by one byte for the Green intensity, and finally one byte for the Red intensity. The first color table entry corresponds to index number 0, the second to index number 1, etc.
    # Graphics Control Extension Block Format:
    #   Byte1         Label for Graphic Control Extension Block. Should always equal 0x21.
    #   Byte2         Block size. Always equals 0xF.
    #   Byte3         Reserved for transparancy information. Should always be set to 0.
    #   Byte4         Type flag. If value equals 0, then don't change frame settings. If value equals 1, then change frame settings. Meaningful only if application specifies Wait Time field after delay time.
    #   Byte5         Delay time. Specifies how much time to show current graphic before displaying next graphic in sequence. Unit is 1/100th seconds. Value should be in range 0..65535.
    #   Byte6         Transparency Index. Specifies what transparent color to make fully transparent. Value should be in range 0..(2^(SizeofGlobalColorTable)-1) - 1.
    #   Byte7         Input control flags. These flags define whether user input outside the graphic area should affect the execution of the animation. Bit 0: User Input Flag. Set this flag if you want user input to affect the animation. Clear this flag if you want animation to proceed unimpeded by user actions. Bit 1: Disposal Method. See section 2.9.8 for details about disposing methods.
    #   Byte8         Transparent Color Index. Used as input argument to restore previous contents underneath the current object. Value should be in range 0..(2^(SizeofGlobalColorTable)-1) -1.
    #   Application Extension Blocks (1+) The spec says there could be any amount of these. Not sure why, given that there doesn't seem to be any other useful ones.
    # Image Descriptor block format:
    #   Byte1          Specifies beginning of image descriptor block. Should always equal 0x2C.
    #   Byte2          Left position of image. Should always be set to 0.
    #   Byte3          Top position of image. Should always be set to 0.
    #   Byte4          Width of image. Value should be in range 0..65535.
    #   Byte5          Height of image. Value should be in range 0..65535.
    #   Byte6          Size of Local Color Table. If image does NOT use local color table, value should be set to 0. Otherwise, value should be in range 1..255.
    #   Byte7          Packed Field. First two bits denote the image disposal method. Other four bits determine the transparency of each pixel.
    #   Byte8          LZW Minimum Code Size. Used during encoding of compressed image data stream. Must be a valid LZW code size, either 2 through 12.
    #   Image Data        LZW Compressed Image Data Stream.
    # Plain Text Extension Block Format:
    #   Byte1          Label for Plain Text Extension BLOCK. Should always equal 0x21.
    #   Byte2          Block size. Should always equal 9.
    #   Byte3          Grid left offset. Number of character cells to the right of the left edge of the grid. Should always equal 0.
    #   Byte4          Grid top offset. Number of character cells down from the top row of the grid. Should always equal 0.
    #   Byte5          Grid width. Width of the text grid, in characters. Value should be in range 0..65535.
    #   Byte6          Grid height. Height of the text grid, in characters. Value should be in range 0..65535.
    #   Byte7          Cell width. Width of character cell, in pixels. Value should be in range 0..255.
    #   Byte8          Cell height. Height of character cell, in pixels. Value should be in range 0..255.
    #   Byte9          Foreground color index. Index number of desired text foreground color within global color table. Value should be in range 0..(2^(SizeOfGlobalColorTable)-1) -1.
    #   Byte10         Background color index. Index number of desired text backgrounf color within global color table. Value should be in range 0..(2^(SizeOfGlobalColorTable)-1) -1.
    # Application Extension Block:
    #   Byte1          Label for application extension block. Should always equal 0xFF.
    #   Byte2          Block size. Should always equal 11.
    #   Byte3          Application ID. Identifies an application that created this GIF