#!/usr/bin/env python3

import sys
import random
import argparse


class InvalidHeaderException (Exception):

    def __init__(self, message="Invalid Header"):
        self.message = message
        super().__init__(self.message)



class InvalidChunkException (Exception):

    def __init__(self, message="Invalid Chunk"):
        self.message = message
        super().__init__(self.message)


class GifParser:

    # Inverse of the ASCII table - maps each ascii code to it's char value
    INV_ASCII_TABLE = bytes([x ^ 0xFF for x in range(256)])


    def __init__(self, fobj):

        # Read entire object from binary file
        self._raw_bytes = fobj.read()
        
        # Get offset information
        self.version = None
        self.width = None
        self.height = None
        self.flags = None
        self.bgcol = None
        self.aspect = None
        self.offset = {
                'header': len(b'GIF89a'),
                'lsd':    13,
                }
        
        self.__parse_header()
        self.__parse_global_color_table()
        
    
    def __repr__(self):
    
        s = "GifPaser[v{}, w={}, h={}, flags={}, bgcol={}, aspect={}]".format(
                hex(self.version),
                hex(self.width), 
                hex(self.height), 
                hex(self.flags), 
                hex(self.bgcol),
                hex(self.aspect))
    
        return s
    
    
    ##
    ## Private Methods
    ##
    
    @staticmethod
    def _ascii_to_int(a):
        '''Convert an array of three ascii characters to an integer'''
        if not isinstance(a, bytes) or len(a)!= 3:
            raise ValueError("Expected 3-byte string")
            
        return sum((b << i * 8) for i, b in enumerate(a))
    

    def __unpack_bcd(self, bytes):
        '''Unpack a BCD encoded int from a series of bytes'''

        # Check encoding is valid
        val = self._ascii_to_int(bytes[:3])
        if val < 100*100: 
            return val
        
        raise InvalidChunkException("Invalid chunk encoding [{}].".format(hex(val)))


    def __parse_header(self):
        '''Parse header structure'''
        if type(self._raw_bytes) == str:
            self._raw_bytes = self._raw_bytes.encode('latin-1')
            
        # Verify we have enough room for magic number and version 
        nbytes = min(len(self._raw_bytes), 7)
        if self._raw_bytes[:nbytes]!= b"GIF89a":
            raise InvalidHeaderException()
            
        # Set Version
        self.version = ord(self._raw_bytes[6])
            
        
        # Parse Logical Screen Descriptor
        offs = self.offset['lsd']
        lsd = self._raw_bytes[offs : offs+7]
        self.width, self.height, packed_flags, self.bgcol = \
                struct.unpack("<HHBBB", lsd)
                
        self.flags = dict(
            sorted({
                   'sortflag': bool(packed_flags & 0x80),
                    'colorres': packed_flags & 0x70, 
                    'globalct': bool(packed_flags & 0x08),
                    }.items(), key=lambda t:t[0]))
        
        
        # Parse Global Color Table
        if self.flags["globalct"]:
            size = 3*(2**(self.flags["colorres"] + 1))
            
            coltab = self._raw_bytes[offs + 7 : offs + size + 7]
                
            self.bgcol = ord(coltab[-1:])
            self.aspect = self._ascii_to_int(coltab[size - 4 : size])

            del coltab[-1:]

    
    def __parse_image_descriptor(self, offs):
        '''Parse Image Descriptor at given offset.'''
        idat = self._raw_bytes[offs : offs + 10]
        left, top, width, height, packed_flags =\
                struct.unpack("<HHHHB", idat)

        return dict(
                sorted({
                        'left':   left, 
                        'top':    top,
                        'width':  width,
                        'height': height,
                        'interf': bool(packed_flags & 0x40),
                       'sort':   bool(packed_flags & 0x20),
                        'localc': bool(packed_flags & 0x80),
                        'lctsize':((packed_flags >> 3) & 0x7),
                        }.items()))
            
    
            
    def __parse_graphics_control_block(self, offs):
        gcb = self._raw_bytes[offs : offs + 8]
        flags, self.delay, transp, terminator = \
                struct.unpack("<BBBB", gcb)


        self.duration = ((self.delay & 0xFF) << 8 | (self.delay & 0xFF00) >> 8) / 100;

        self.transparent = False
        self.transparency = []

        if transp == 0:
            self.transparent = True
        elif transp > 0:
            self.transparent = False
            self.transparency = [(transp & 0x0F), (transp & 0xF0)]



    def __parse_graphic_content(self, offs, graphic_ext):
        '''Parse Graphic Content block at given offset'''
        
        data = self._raw_bytes[offs : ]

        
        blocks = list() 

        while len(data) > 0:


            if len(data) >= 13:

                blksize, blktype = struct.unpack('<BB', data[:2])
                
                if blktype == 0x2C:

                    imgdesc = self.__parse_image_descriptor(offs)
                    offs += 10
                    blocks.append(('id',imgdesc))
                    
                elif blktype == 0x21:
                    
                    if len(data) >= 20:

                        ext, label = struct.unpack("BB", data[2:4])
                        
                        if ext == 0xF9:
                            self.__parse_graphics_control_block(offs)
                            blocks.append(("gcb", self))
                            offs += 8
                            
                        else:
                            
                            if ext == 0xFE:
                                imgdesc = self.__parse_image_descriptor(offs)
                                
                                offs += 10
                                blocks.append(("ie", imgdesc))

                            elif ext == 0xFC:

                                size = struct.unpack("<B", b'\xff'[0])[0] 
                                offs += 1
                                blocks.append(("gc", {'compressed': data[3: size+3]}))
                                offs += size                        

                        offs += blksize - 2
                        
                    else:
                        break
                    
                else:
                    print("[!] Found unknown block:", hex(blktype))
                    exit(-1)
                    
            else:
                break
            

        return blocks
                
        
    def get_frames(self):
        '''Return frames based on parsed information'''
        
        frameset = set()
        current_frame_offsets = {}
                             
        for idx, (tag, payload) in enumerate(self.__parse_graphic_content()):

            if tag == "id":                
                current_frame_offsets["current"] = {"start":idx}
              
            elif tag == "gcb":
                current_frame_offsets["current"]["end"] = idx + 1
                        
            elif tag == "ie":
                current_frame_offsets["current"]["end"] = idx + 1
                              
                # Assign the frame offsets dictionary to the frameset tuple
                frameset.add(tuple(sorted(current_frame_offsets["current"].items())))     
                
                # Reset offset record
                current_frame_offsets["current"] = {"start":idx}
            
        return frameset
    

##
## Public API Functions
##
    
def verify_file(path):
    '''Verify that file exists and can be opened by this program'''
    
    try:
        with open(path,'rb'):
            pass
            
    except FileNotFoundError as e:
        print(e)
        exit(-1)
        
    return path
    
    
def parse_args():
    parser = argparse.ArgumentParser(\
            description="Generate a random gif using Python.")
    parser.add_argument('-o','--output', help="Output file name.", default='/tmp/.py.gif')
    parser.add_argument('-r','--rand', type=str, help="Use /dev/<name> to generate randomness.")
    args = vars(parser.parse_args())
    
    global RANDOMNESS
    RANDOMNESS = '/dev/' + args['rand'] if args['rand'] else '/dev/urandom'
    
    global OUTPUTFILE 
    OUTPUTFILE = verify_file(args['output']) 
    
    return args


def main():
   
    # Grab command line arguments    
    args = parse_args()
   
    # Generate random gif
    with open('/dev/random', 'rb') as rng,\
            open(OUTPUTFILE, 'wb+') as out:
 
        generate_random_gif(rng, out)
 
 
    # TODO: Open generated file and split into multiple files based on
    #       the content of the Graphics Control Block. This requires
    #       implementing some kind of image processing library like PIL

    
if __name__ == "__main__":