#!/usr/bin/env python3

import random
from io import BytesIO
from typing import BinaryIO

# Generates a random GIF file with random length and color depths in `out`.
def generate_random_gif(
        out: BinaryIO
    ) -> None:

    width  = random.randrange(1024)
    height = random.randrange(1024)

    # We want to have enough colors that the color palette can be compressed
    # using a simple LZW compression scheme. The maximum size of our palette
    # thus becomes unbounded from there onwards. Therefore we need to use a very
    # high value for this upper bound.
    maxColors = pow(2, 16) - 1

    # Write Header
    out.write(b'GIF87a')
    
    lzwCodeLen   = randint(9, 12)
    minCodeSize  = len(str(maxColors)) + 1
    out.write(struct.pack('!HHB', width, height, minCodeSize))

    # Number of Global Color Table entries (can't compress well if greater than 256)...
    globalColorTableEntriesCount = random.randrange(min(maxColors, 256))
    out.write(globalColorTableEntriesCount.to_bytes(1, 'little'))
    for _ in range(globalColorTableEntriesCount):
        r = random.randint(0, 255)
        g = random.randint(0, 255)
        b = random.randint(0, 255)
        a = random.randint(0, 255)

        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))
        out.write(a.to_bytes(1, 'little'))

    #... followed by an optional Local Color Table
    localColorTablePresentFlag = random.choice([True, False])
    out.write(localColorTablePresentFlag.to_bytes(1, 'little'))
    if not localColorTablePresentFlag:
        return

    localColorTableEntriesCount = random.randrange(min(maxColors, 256))
    out.write(localColorTableEntriesCount.to_bytes(1, 'little'))
    for _ in range(localColorTableEntriesCount):
        r = random.randint(0, 255)
        g = random.randint(0, 255)
        b = random.randint(0, 255)
        a = random.randint(0, 255)

        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))
        out.write(a.to_bytes(1, 'little'))

    imageDescriptorFlags = [
             True,                          # Left position present flag
             True,                          # Top position present flag
             False,                         # Width present flag
             False,                         # Height present flag
             False,                         # Interlace flag
             False,                         # Sort flag

             # Reserved flags
             True,                          # Reserved flag
             True]                          # Reserved flag
    imageDescriptorFlags = struct.pack(
            '!8?', *imageDescriptorFlags)
    out.write(imageDescriptorFlags)
    # If left or top are present they take up 2 bytes each
    # Else they don't take space at all
    if imageDescriptorFlags[0]:
        out.write(width.to_bytes(2, 'little'))
    if imageDescriptorFlags[1]:
        out.write(height.to_bytes(2, 'little'))

    # Frame Control Block
    frameControlBlock = struct.pack(
            '!BBBHB',
            0x2C,             # Image Separator Character
            0,                # Left Position
            0,                # Top Position
            width,            # Width
            height           ) # Height
    out.write(frameControlBlock)

    # LZW minimum code size
    out.write(lzwCodeLen.to_bytes(1, 'little'))

    out.flush()

    # Generate data stream
    datastreamBuffer = BytesIO()
    encoder = LzwEncoder(datastreamBuffer, lzwCodeLen)

    # Number of frames as specified by the user
    numFrames = random.randrange(256)
    for _ in range(numFrames):
        # Number of colors...
        numColors      = random.randrange(maxColors+1)
        receiverStream = encoder.encode(
                ''.join('{0:#0{1}X}'.format(
                        random.randint(0, 0xFFFFFFFFFFFFFFF),
                        18)[2:]
                    for i in range(numColors)))

        #...followed by their RGB values
        for c in receiverStream:
            r = random.randint(0, 255)
            g = random.randint(0, 255)
            b = random.randint(0, 255)
            a = random.randint(0, 255)

            out.write(c.to_bytes(1, 'little'))
            out.write(r.to_bytes(1, 'little'))
            out.write(g.to_bytes(1, 'little'))
            out.write(b.to_bytes(1, 'little'))
            out.write(a.to_bytes(1, 'little'))

        fcb = struct.pack(
                '!BBHBB',
                0x2C,     # Image Separator Character
                0xf9,     # Graphic Control Extension Label
                4,        # Length of block
                *([0]*2)) # No disposal method, no delay time
        out.write(fcb)

        ddd = struct.pack(
                '!BBBBBB',
                *([0]*6)) # No transparent color given
        out.write(ddd)

    # Close output after writing the data
    out.close()

