#!/usr/bin/env python3

### Random GIF File Generator

import sys
import os
import time
import math
from typing import (BinaryIO, Optional)

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# Returns True on success and False otherwise.
def generate_random_gif(rng: BinaryIO,
                        out: BinaryIO) -> bool:

    # Read in the first 6 bytes to get width & height of image.
    try:
        w, h = struct.unpack('HH', rng.read(2*2))
    except Exception as e:
        print('Error reading size:', str(e), file=sys.stderr);
        return False
    
    # Read in one byte for color resolution.
    cr = rng.read(1)[0] % 9
    if cr == 7 or cr == 8:
        cr += 1
    
    # Read in one byte for global flags.
    gf = rng.read(1)[0]
    bgcflag = gf & 1      # Background Color Flag
    gfcflag = (gf >> 1) & 1   # Global Color Table Flag
    lctsize = (gf >> 2) & 4    # Local Color Table Size
    interlace = (gf >> 5) & 4  # Interlace flag
    sortflag = (gf >> 6) & 2   # Sort Flag
    reserved = (gf >> 7) & 2   # Reserved bit

    # Read in the remaining control blocks until we hit an end block.
    while True:

        # Check that this is actually a valid block type.
        bt = ord(rng.read(1))
        if not (bt >= 0x20 and bt <= 0xff):
            break
        
        # End Block
        elif bt == 0x3B:
            pass

        # Comment Block
        elif bt == 0xFE :

            # Get length of comment string.
            clen = int.from_bytes(
                    rng.read(2), byteorder="little")
            
            # Skip over the rest of the comment string immediately. 
            comment = rng.read(clen)
            
        # Plain Text Extension
        elif bt == 0x01:
                
            # Get a random language code to write into
            langcode = rng.read(1)
            
            # Next two byes are plain-text subblock size.
            ptlens = rng.read(2)
            
            # Last four bytes indicate how many loops. 
            textloops = rng.read(4)
                
        # Application Extension
        elif bt == 0xFF:

            appid = rng.read(1)
            alen = rng.read(1)
            data = rng.read(alen[0])
                    
        else:
            continue
            
    # Default background color
    bgcolor = [0xF7, 0xF7, 0xF7]
        
    # Generate palette with at most 256 colors. 
    rgb_max = max([w, h]) * 2**cr
    palsize = min((rgb_max // 3) - 1, 256)
    
    # Determine whether to use local or global color table.
    ct = None
    if gfcflag!= 0:
        # Use global color table. 
        ct = []
        for i in range(palsize+1):
            r,g,b = rng.read(3)
            ct.append([r,g,b])
    else:
        # Use local color table. 
        lctsize = lctsize + 1
        lct = []
        for i in range(lctsize):
            lct.append([
                rng.read(1)[0],
                rng.read(1)[0],
                rng.read(1)[0]])
            
    # TODO This compress should be factored out somewhere.
    def compress():
        """Compress our data stream."""
    
        # Prepare initial frame state variables. 
        sof = 0                   # Start of frame marker has not been sent yet.
        pos = 0                   # Current position in data buffer.
        prevbyte = 0              # Value previously read from input stream.
        bytecount = 0             # Number of Uncompressed Bytes since last ByteCode.
        gotcode = False           # A flag indicating whether we have received any codes.
    
        # Output buffer variable; stores compressed data.
        outputbuffer = b''
    
        while True:
    
            # If we haven't sent a start of frame yet, do so now.  
            if not sof:
                pos = 0
                buf = b"\x21\xF9\x04" \
                      b"\x00"             \
                      b"\x00"             \
                      b"\x00"             \
                      b"\x00\x00"         \
                      b"\x0F\x00"
            
                written = len(buf)
                outputbuffer = outputbuffer + buf
    
                sof = True
    
            # Read 1 byte from input buffer. 
            try: 
                curval = rng.read(1)[pos] % 256
            except IndexError:
                break
        
            # Count the number of sequential bytes equaling 'curval'.
            numequal = 0
            while curval == prevbyte:
                numequal += 1
                pos += 1
                
                # Break loop when we've reached end of input. 
                if pos > len(data): 
                    break
                    
                try: 
                    curval = next(data)[pos] % 256
                except StopIteration:
                    break
            
            # Send RLE if applicable.
            if numequal >= 3:
                buf = b""
                buf = buf + bytes([numequal-3])   # Write number of repeats minus 3.
                buf = buf + bytes([prevbyte])     # Output the repeated value.
                written = len(buf)
                outputbuffer = outputbuffer + buf
                bytecount = 0                 # Reset byte counter.
                gotcode = True
            else:
                bytecount += numequal            # Increment byte count. 
                gotcode = True                    
    
                # When we've accumulated enough data, emit it.  
                if bytecount >= 255:
                
                    # Flush the entire state machine and send the bytes. 
                    pos -= (bytecount - 1)
                    buf = b""                   
                    
                    # Emit output buffer's current contents.
                    if len(outputbuffer) > 1: 
                        buf = buf + bytes([255])
                        buf = buf + outputbuffer[:len(outputbuffer)]
                        written = len(outputbuffer)+1
                        
                    # Emit remainder of previous chunk but no more than 255 characters.
                    prevchunk = outputbuffer[-1:]
                    if len(prevchunk) < 255: 
                        buf = buf + bytes([len(prevchunk)])+prevchunk
                        written = len(written)+len(prevchunk)+1
                    
                    # Reset state machine. 
                    bytecount = 0
                    prevbyte = 0
                    outputbuffer = b''
            
                    # Write RLE output to final output buffer. 
                    outputbuffer = outputbuffer + buf
                 
                # Update variables. 
                prevbyte = curval 
    
            # Exit loop when we're done processing all input data. 
            if pos > len(data): 
                break
            
        # Now send End Of Stream signal. 
        eos = b'\x00\x3B'
        outputbuffer = outputbuffer + eos
        written = len(outputbuffer)
       
        # Return the written data size in bytes. 
        return outputbuffer