#!/usr/bin/env python3
"""Generate random GIF files."""


import struct
import sys
from hashlib import sha256
from pathlib import Path
from random import randint, seed as rndseed


# Defines the default name for the output GIFs and the size of each frame
DEFAULT_FILENAME: str = 'output-{}.gif'
FRAME_SIZE:        tuple[int]  = (100, 100)   # in pixels



# The main function
if __name__ == '__main__':

    # Parse command line arguments
    try:
        maxframes=(int)(sys.argv[1])
    except Exception:
        print('Argument must be an integer!')
        exit(-1)


    # Seed the random number generator
    rndseed()

    # Create output directory if it doesn't exist yet
    outdir = Path('./output/')
    outdir.mkdir(parents=True, exist_ok=True)
    

    for i in range(maxframes):

        with open('/dev/urandom', mode='rb') as rng,\
                open(outdir / DEFAULT_FILENAME.format(i+1), 'wb') as out:

            # Generate only one image per hash to avoid duplicates
            imghash = sha256(rng.read())
            
            # Write file header
            write_fileio_struct(out, struct.pack("<8B",
                71,   # "G"
                73,   # "I"
                70,   # "F"
                56,   # version
                57,   # minor version
                97,   # width
                97))  # height


            # Write frames
            nbframes = randint(1, 10)
            for j in range(nbframes):

                # Select color for this frame
                r = randint(0, 255)
                g = randint(0, 255)
                b = randint(0, 255)


                # Write the current frame's lwz info block
                write_fileio_struct(
                    out, struct.pack('<2H', FRAME_SIZE[0], FRAME_SIZE[1]))
                
                # We don't have any special flags or delay value so we pad them
                # out by zeroes
                write_fileio_struct(out, struct.pack('<BBB',
                    0x00,       # No disposal flag
                    0x00,       # User input is not required
                    0x00))      # Delay time: no delay


                # Write palette information
                write_fileio_struct(
                    out, struct.pack('<BBB', r, g, b))

                # Write actual frame content
                write_framecontent(out, len(imghash.digest()))


        print("Generated {} frames".format(str(i)))


# Utility functions go below here

# Writes a binary chunk and its length prefix to a file stream
def write_fileio_struct(outstream: BinaryIO, barray):
    
    outstream.write(len(barray).to_bytes(4, byteorder="big"))
    outstream.write(barray)


# Creates a random frame that contains only black pixels
def write_framecontent(fout: BinaryIO, numpixels: int):

    pixeldata = bytes([0x00]*numpixels*3)
    fout.write(pixeldata)