#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The GIF has 1024x768 resolution and is generated from
# pixels ranging in [0..255].
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):

    # Generate a GIF file with the following structure:
    # - Header (always "GIF89a")
    # - Logical Screen Descriptor
    #   - Width  : 1024 px
    #   - Height : 768 px
    #   - Global Colour Table  : RGB palette of all {0..255} values
    #     (i.e., no compression)
    # - Graphics Control Extension
    #   - Disposal method : No disposal specified
    #   - User input      : False
    #   - Transparent color index: None
    # - Image descriptor:
    #   - Left offset    : 0 px
    #   - Top offset     : 0 px
    #   - Width          : 1024 px
    #   - Height         : 768 px
    #   - Local colour table flag : False
    #   - Interlace flag        : False
    #   - Sorting flag          : False
    #   - Size of local colours : 0 bytes
    # - Uncompressed image data. All pixels are random numbers
    #   ranging between {0.. 255}. The size of this data is
    #   inferred from the previous elements above.

    # First, we write the header
    out.write(b"GIF89a")

    # Next, we write the logical screen descriptor. We use an ungranular gct
    # (i.e., one containing just two colors), so that it fits in the first frame
    # of our animation. This way, we don't have to manually update the image
    # descriptor every time we add another frame to the GIF.
    #
    # Note that the minimal canvas width here is 3px because we want to be able
    # to pack a number of pixels into each byte used for encoding (i.e.,
    # pixel indices).
    lsd = [
        0x0A,                 # Canvas width       : 1024 px
        0xD0,                 # Canvas height      : 768 px 
        0x01,                 # GCT                : True
        0x02,                 # Pixel aspect ratio : 1:1

        # GCT block
            0xFF, 0x00,             # Color 0: Red
            0x00, 0xFF,             # Color 1: Blue

            0x00                    # End of GCT
    ]
    out.write(struct.pack("<{}B".format(len(lsd)), *lsd))


    # Now let's write the graphics control extension
    gfxctl = [
        0xF9,                 # Extension introducer: Graphic extension
        0x04,                 # Block length         :
         0x01,               # Frame label          :
         0x00,               # Null separator

        0x04,                 # Image left offset    : 0 px
        0x00,                 # Image top offset     : 0 px
        0x01,                 # Image width          : 1024 px
        0x00,                 # Image height         : 768 px
        0x00                  # Packed fields        : None set
    ]
    out.write(struct.pack("={}B".format(len(gfxctl)), *gfxctl))

    # Next comes the image descriptor. Again, we'll create frames where we can
    # pack as many pixels as possible into each byte. Thus, in this case, the
    # minimum canvas width will be 7 instead of 3.
    imgdesc = [
        0x2C,                 # Image Separator
        0x00,                 # Left offset           : 0 px
        0x00,                 # Top offset            : 0 px
        0x01,                 # Image width           : 1024 px
        0x00,                 # Image height          : 768 px
        0x00,                 # Packed fields         : None set
        0x00                  # Local Colours count   : none
    ]
    out.write(struct.pack("={}B".format(len(imgdesc)), *imgdesc))


    # Finally, the image itself. The size of this data will depend on
    # the previous entries. In this example, we assume that there'd be at least
    # 256 images packed in the same GIF (so that they occupy a single byte),
    # which is true in this specific case. 
    #
    # Since the GCT contains only 2 colors (black & white), we need to make sure
    # that the range of random numbers used to encode the pixels does not fall
    # outside of their expected bounds ([0..2]). Otherwise, we might end up using
    # invalid colors...
    pix_count = 1024*768
    # Each byte encodes 8 pixels. Compute how many full bytes we're going to fill
    nfullbytes = pix_count // 8
    nbytes = nfullbytes+1 if ((pix_count % 8)!= 0) else nfullbytes
    
    # Write the actual data
    data = []
    while len(data) < nbytes:
        data += rng.randint(-1, 3, size=min(nbytes-len(data), 8)).tolist()
    out.write(struct.pack('=B', *(c&0xff for c in data)))

