#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write header to output (always 1st bytes of file)
    headarr = bytearray([
            0x47, 0x49, 0x46])        # GIF
    version = rng.read(3)              #Version is always '89a', we don't need it
    headarr += version                  # Write the generated version at the end of the header
    headarr[8] |= 0b1 << 5             # Set global color table flag
    # TODO: Make colortable configurable
    headarr += bytearray((2**10 - 2)*3)     # Color table, we use all colors in full color range
    headarr[10] |= 0b1                   # Set the size of the pixel's minimum unit (becomes 2^10)
    headarr[10] |= 0b1 << 3             # Set for LZW compression
    out.write(headarr)                  # Write complete header

    # Now write the image data...
    w, h = {
        1: [80, 60],
        2: [100, 100],
        3: [120, 120],
        4: [160, 160],
        5: [200, 200]
    }[int.from_bytes(rng.read(1), byteorder="big")]   # Get width and height according to the chosen scale factor
    lencap = int.from_bytes(rng.read(1), byteorder="big")    # Get length cap in bytes (1-255)
    lencode = rng.read(lencap)                              # Read the data from the RNG (max lencap bytes)
    nbitscode = len(lencode)//w//h                          # Find out how many bits are required per code part (min. 2-10)

    if not 2 <= nbitscode <= 10: raise ValueError("nbitscode must be between 2 and 10!")

    out.write(struct.pack("<HHB", w, h, nbitscode))         # Write frame dimensions and the nr of bits for each code part
    out.write(lencode)                                      # Write compressed image data