#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generate an infinite stream of random bytes from /dev/urandom.
# This can be used to generate arbitrary length data without storing it in memory.
def get_random_stream():
    return open('/dev/urandom', 'rb')


# Generate a single chunk of image data based on `rng`.
def generate_chunk(
    width: int, height: int, rng: BinaryIO):
    # Append one bit at a time until we reach the target size
    pixeldata = bytearray()

    for y in range(height):
        for x in range(width):

            if len(pixeldata) > 0 and (len(pixeldata) % 2 == 1):
                pixeldata += bytes([int(str(bool(rng.read(1))), 2)])

            pixeldata += bytes([int(str(bool(rng.read(1))), 2)])

        # Pad each line with zeroes for even alignment
        while len(pixeldata) % 2!= 0:
            pixeldata += bytes([0])


    assert(len(pixeldata) <= (width * height))
    return pixeldata


# Write the palette as a square with dimensions `palsize`.
# We use this for both color and background palettes.
def write_palette(
    palsize: int, numcolors: int, rng: BinaryIO, out: BinaryIO):

    colors = [None] * ((numcolors ** 2) * 3)

    for i in range(numcolors ** 2):
        r = rng.read(1)[0] & ~7
        g = rng.read(1)[0] & ~7
        b = rng.read(1)[0] & ~3
        colors[i*3:(i+1)*3] = (r, g, b)

    out.write(struct.pack('B' * (numcolors**2 * 3), *colors))



# A plain black frame
def write_plain_frame(out: BinaryIO):
  out.write((b'\x21\xF9\x04\x00\x0D\x00\x0A\x00'))
  out.write((b'\x00\x2C\x00\x00\x00\x00\x00\x00'))

  # Frame control block
  out.write(b'\x00' * 8);


# Writes the global palette to the output.
def write_global_palette(
    palsize: int, numcolors: int, rng: BinaryIO, out: BinaryIO):

    out.write(b'GCE')
    out.write(b'\0' * 3)    # Background Color Index
    out.write(b'\xFF' * 3)   # Pixel Aspect Ratio

    out.write(struct.pack('<HHH', palsize, numcolors, 3))

    write_palette(palsize, numcolors, rng, out)





# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The generated file
# is valid and encodes all frames that fit within the given constraints.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        maxframes: int=500, minsize: int=(64*64), maxsize: int=(256*256)):

    framecount = 0
    cursize = 64*64;
    firstframe = True
    
    while framecount < maxframes or not firstframe:
        firstframe = False

        # Set up the current frame
        w = h = None

        while w is None:
            w = int.from_bytes(
                    rng.read(1), byteorder='little', signed=False)
            
            if w == 0: continue
            elif w >= 33: continue
            else: w += 1

        while h is None:
            h = int.from_bytes(
                    rng.read(1), byteorder='little', signed=False)
            if h == 0: continue
            elif h >= 33: continue
            else: h += 1


        while cursize > maxsize:

            ratio = round(cursize / (w * h))
            w //= ratio
            h //= ratio

            print("Rounding down to {}x{}".format(w, h))


        # Just don't bother if the frame won't have any pixels
        if w * h == 0: 
            print("Not enough room for another frame")
            break


        cursize = w * h
        print("Frame size: {}x{}".format(w, h))

        out.write(b'GIF89a')
        out.write((w - 1).to_bytes(1,'little'))        # Width
        out.write((h - 1).to_bytes(1,'little'))        # Height
        out.write(b'\xf7')                              # Flags: Global Color Table Enable
        out.write(b'\x00')                              # Background Color Index
        out.write(b'\xff')                              # Pixel Aspect Ratio
        
        # LZW minimum code size
        lzwmincodesize = 2

        # Global Palette
        numcolors = 2
        palsize = 1 << lzwmincodesize

        write_global_palette(palsize, numcolors, rng, out)

        # Image Descriptor
        out.write(b'!')
        out.write(b'\xfe')                              # Left Column
        out.write(b'\x00')                              # Top Row
        out.write(((w-1) & 0xFE).to_bytes(1,'little'))  # Width
        out.write(((h-1) & 0xFE).to_bytes(1,'little'))  # Height
        out.write(b'\x00')                              # Flags: Local Color Table Enabled
        out.write((lzwmincodesize).to_bytes(1,'little'))      # Minimum Code Size

        write_palette(palsize, numcolors, rng, out)

        # Image Data
        # TODO: Choose between LZW and just raw data encoding
        pixeldata = generate_chunk(w, h, rng)
        #pixeldata = bytes([42]*(w*h))
        out.write(pixeldata)
        
        # End of Image
        out.write(b';')

        framecount += 1



    # Ancillary chunks
    write_plain_frame(out)
        
    # Set Delay Time to 0 milliseconds
    out.write(b"!\xf9\x04")                    # Application Extension
    out.write(b"\0\0")                         # Block Size
    out.write(b"\0\0")                         # Loop Count

    # Plain Background
    out.write(b"!\xfc\x04")                    # Application Extension
    out.write(b"\0\0")                         # Block Size
    out.write(b"\0\0\0\0")                     # RGB

    # Force end with EOF
    out.write(b'EOF')