#!/usr/bin/env python3

import gzip
import zlib
from io import BytesIO
from typing import AnyStr, Union

# Generates a random GIF file in `out`. If `compress` is True (the default),
# then it will be compressed with `zlib`; if False, not at all; and if None,
# *some* of them will be compressed but compressibility depends on the RNG state.
def generate_random_gif(
        rng: BinaryIO, out: AnyStr, *,
        compress: bool=True
    ) -> None:

    # Generate the uncompressed data
    num_frames: int  = ((rng.read(1)[0] & 7) + 1) << 2
    width:     int  = 1 << (rng.read(1)[0] % 5)
    height:    int  = 1 << (rng.read(1)[0] % 5)
    color_table = [i for i in range(2**((rng.read(1)[0] & 7) + 1))]
    frames: list[bytes] = []
    for frame_num in range(num_frames):

        x = y = 0
        while x < width and y < height:

            # Determine next command based on `x`, `y`
            cmd = rng.read(1)[0] \
                    if x == 0 or y == 0 else rng.read(1)[0] // 4

            if cmd < 1:
                x += min(width - x, rng.read(1)[0])
            elif cmd < 2:
                y += min(height - y, rng.read(1)[0])
            elif cmd < 3:
                x -= min(max(x, 0), rng.read(1)[0])
            elif cmd < 4:
                y -= min(max(y, 0), rng.read(1)[0])
            elif cmd < 5:

                # Get the length of the line from the following byte
                ln = rng.read(1)[0]

                # Randomly decide whether to use the current color table
                if len(color_table) > 1 and rng.read(1)[0] % 5!= 1:
                    color = rng.read(1)[0] % len(color_table)
                else:
                    color = 0xff

                # Write as many pixels as we need
                out_byte = bytes([color])
                for _ in range(ln):

                    # Determine how much to shift the pixel off by
                    sft = rng.read(1)[0] // 16

                    # Read the next byte
                    nxt = rng.read(1)[0] >> sft

                    # Apply XOR masking and add the result to our output buffer
                    out_byte += (nxt ^ (color >> sft)).to_bytes(
                        1,
                        'little', signed=(sft & 1))

                # Write this frame to the output buffer
                frames[-1] += out_byte

    # Compress it now that everything's built up
    if isinstance(out, str):
        out = open(out, 'wb')

    # Start the output stream
    gif = BytesIO()
    try:

        # First write some metadata before any compression takes place
        gif.write('GIF89a'.encode())

        # In order to make sure that this works, we'll first put each frame
        # within its own sub-block and then combine these together into one big
        # block when we're done (to prevent splitting between the blocks).
        subblocks: list[bytes] = []
        for frame in frames:
            subblocks.append(b'\x2c\x00' +
                             width.to_bytes(
                                2,
                                'little', signed=False
                            ))
            subblocks.append(height.to_bytes(
                2,
                'little', signed=False
            ))
            subblocks.append(b'\x87' + b','.join([
                c.to_bytes(
                    1, 'little', signed=False) for c in color_table
            ]))
            subblocks.append(frame)

        # Now write this all out
        for sb in subblocks[:-1]:
            gif.write(sb)
            gif.write(b';')
        gif.write(subblocks[-1])

        # The final piece should be the application extension
        appext = b'\xf9\x04'
        j = 0
        while len(appext) <= 256:
            j = ((j * 256) + rng.read(1)[0]) & 0xfffff
            tmp = j.to_bytes(3, 'big')
            appext += tmp
            if tmp[:2] == b'\x00\x00': break
        appext += b'\x00' + b'\xfe'

        # Enqueue this one last time too
        gif.write(appext)

        # And finally close the GIF stream so it can be written
        buf = gif.getbuffer()

        # If requested, compress this whole thing!
        if compress == True:
            out.write(gzip.compress(buf))
        elif compress == False:
            out.write(buf)
        else:
            if rng.read(1)[0] % 2 == 0:
                out.write(gzip.compress(buf))
            else:
                out.write(buf)

    except Exception:
        raise

