#!/usr/bin/env python3

import struct
from typing import BinaryIO


#'magic' header for GIF files. Bytes 6 and 7 encode
# the color resolution (1 -> 256 colors).
HEADER = b'\x47\x49\x46\x38\x39\x61'

# Sizes in bytes of various fields in a GIF image
BLOCKSIZE = {
    "signature": 6, 
    "version": 3,  
    "width": 2,     
    "height": 2,    
    "flags": 1,     
    "bgcolorindex": 1,
    "pixelaspectratio": 1     # not used with most images
}
    
# Flag bits indicating that pixels are interleaved
INTERLACED = [b"\x00", b"\x40"]

# The maximum size of an uncompressed palette entry is 16 bits. This must be
# followed by two zero bytes to ensure palettes have even length.
MAXCOLORDEPTH = 1 << 16

# Generate a single frame for a random-colored image. Returns a tuple containing
# - An image descriptor block, which describes this image's position on the screen
#   and its transparent color index
# - A graphics control extension, which tells the decoder how to play this image
#   when it appears
# - A local color table, which contains one or more RGB tuples where each item
#   consists of three unsigned bytes (i.e., 24 bit values)
# - The pixel data stream itself. Its size depends on whether or not the image uses
#   transparency at all 

def make_frame():

    width = height = 32
    
    flags = INTERLACED[0]
    bgcolorindex = 0xFF

    colordepth = MAXCOLORDEPTH
    
    globalcolortablesize = 2 ** colordepth
    colortable = [(rng.randrange(256), rng.randrange(256), rng.randrange(256)) 
                  for _ in range(globalcolortablesize)] 
    
    return (
        # Image Descriptor Block 
        HEADER + \
        width.to_bytes(BLOCKSIZE["width"], byteorder="little") + \
        height.to_bytes(BLOCKSIZE["height"], byteorder="little") + \
        flags + \
        bgcolorindex.to_bytes(BLOCKSIZE["bgcolorindex"], byteorder="little") + \
        BLOCKSIZE["pixelaspectratio"].to_bytes(BLOCKSIZE['pixelaspectratio'], 
             byteorder="little"),

        # Graphics Control Extension
        HEADER + \
        b"\xf9" + \
        b"\x04" + \
        0.0.to_bytes(4, byteorder='little', signed=True) + \
        0.0.to_bytes(4, byteorder='little', signed=True) + \
        0.0.to_bytes(4, byteorder='little', signed=True) + \
        b"\xff" + \
        0.0.to_bytes(2, byteorder="little")

        # Local Color Table
        HEADER + \
        globalcolortablesize.to_bytes(BLOCKSIZE["bgcolorindex"], byteorder="little") + \
        b"".join([struct.pack("<BBB", r, g, b) for r,g,b in colortable])
        
        # Pixel Data Stream
        b"",
    )
