#!/usr/bin/env python3

import os
import sys
import time
import secrets

from PIL import Image


# Generate an image of size (width x height). It has
# around 1 / n_colors distinct colors, where n_colors is either
# provided or chosen by the algorithm such that pixels are
# not too small and there is at least one pixel in each region.
def make_image(
        width: int, height: int, n_colors=None
    ) -> Image.Image:

    # If we do not have enough colors, try again with
    # more colors
    if n_colors is None or n_colors > 256:

        while True:

            # First count how many unique colors would be needed to display this
            pixs_per_color = width * height // max(n_colors, 1)
            n_colors = min((2 ** 24) // pixs_per_color - 1, 256)

            im = Image.new('RGB', (width, height))
            px = im.load()

            for i in range(0, len(px)):
                px[i] = [
                    *(secrets.randbelow(n_colors),)*3,
                ]

            uniq = set([tuple(p) for p in im.getdata()])

            print('[+] Generating {}x{} with {} colors...'.format(
                    width, height, n_colors))
            print('[+]', len(uniq), 'unique colors were used.')
            if len(uniq) >= 256: break


    else:
        
        assert n_colors <= 256, f'{n_colors} exceeds maximum value allowed.'

        im = Image.new('RGB', (width, height))
        px = im.load()

        for i in range(0, len(px)):
            px[i] = [
                *(secrets.randbelow(255),)*3,
            ]
    
    return im



class GifError(Exception): pass



# Writes a RGB565 buffer as a GIF stream to `out`. This can be read by e.g.,
# https://www.wesnoth.org/wiki/Testbed/gif_test.php
def write_rgb565_as_gif(
        inb: bytes, out: BinaryIO
    ):
        
    MAXVAL = 2**16-1
    MAXLEN = 2**16//2

    if len(inb) % 2!= 0: raise ValueError('input must be divisible by two')
    if len(inb) > MAXLEN: 
        raise ValueError('{} exceeds maximum length ({})'.format(len(inb), MAXLEN))

    header = [
        71, 73, 70, 56, 57, 
        97, 0, 2, 0, 0, 
        0, 0, 0
    ]

    def rgb565_to_byte(*args):
        r, g, b = args
        combined = ((r & 0xF8) << 8) | \
                   ((g & 0xFC) << 3) | \
                   ((b & 0xF8) >> 3)
        return struct.pack('<H', combined)[0]

    out.write(bytes(header))
    for i in range(0, len(inb)//2):
        x, y = inb[2*i:2*i+2]
        out.write(struct.pack('<B', rgb565_to_byte(x, y)))


def main():
    if '-h' in sys.argv or '--help' in sys.argv:
        print('''
        usage:./random_gif.py [-h | --help] [filename]
          writes a random gif file to "filename"
        ''')
        exit(0)

    filename = 'random.gif'
    if len(sys.argv) == 2:
        filename = sys.argv[1]
    elif len(sys.argv) > 2:
        raise GifError('invalid command line arguments')

    nframes = 100
    fps = 10
    tottime = float(fps * nframes)

    t = time.monotonic()
    rng = open('/dev/urandom', mode='rb')

    widths = list(range(200, 100, -5))
    heights = [max(w//2, 10) for w in widths]

    ncols = 2**(len(widths) - 1).bit_length()

    frames = []
    for _ in range(0, abs(nframes)):
        frame = []
        for w, h in zip(widths, heights):
            frame += [(make_image(w, h, ncols))]
        frames.append(frame)

    dt = time.monotonic() - t
    print('[+] Generated {} frames in {:.3f} seconds ({:.3f} FPS)'.format(
            nframes, dt, nframes / dt))

    t = time.monotonic()
    durations = [tottime]*abs(nframes)
    durations[-1] = tottime*(dt / tottime)

    images = []
    previmg = None
    for i in range(0, len(frames)):
        for j in range(0, len(frames[i])):
            img = frames[i][j].copy()
            if previmg is not None:
                img.paste(previmg, mask=previmg)
            images.append(img)
            previmg = img
        
        images[-1].save(
            '/tmp/{}_{:03d}.png'.format(filename, i),
            format='PNG', 
            save_all=True,
            append_images=[img for img in reversed(images[:-1])],
            optimize=False,
            duration=durations[i],
            loop=0)

    
    for i in range(0, len(images)):
        img = images[i]
        # Convert to RGB565
        buf = img.convert('RGB').tobytes()
        write_rgb565_as_gif(buf, open('../{}.gif'.format(filename),'wb'))
        os.unlink('/tmp/{}_{:03d}.png'.format(filename, i))



    dt = time.monotonic() - t
    print('[+] Wrote all {} frames in {:.3f} seconds ({:.3f} FPS)'.format(
            nframes, dt, nframes / dt))


if __name__=='__main__':
    try:
        main()
    except Exception as e:
        print(e)
