#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into outfile using the
# random number generator rng (e.g., /dev/urandom).
# The generated GIF has dimensions of width x height, where 1 <= width and
# height <= 2048. If width or height is not specified, a dimension will be chosen
# at random between min_width and max_width, and similarly for height.
def generate_random_gif(
        rng: BinaryIO, outfile: BinaryIO,
        *, width=None, height=None,
        min_width=64, max_width=512,
        min_height=64, max_height=512
    ):

    # Choose a random dimension for width if none was provided
    dimlen = 1
    if width == None:
        width = rng.randint(min_width, max_width)
    else:
        assert isinstance(width,int)
        assert width > 0
        dimlen += 1

    # same for height
    if height == None:
        height = rng.randint(min_height, max_height)
    else:
        assert isinstance(height,int)
        assert height > 0
        dimlen += 1

    dims = [width & 0xff] * dimlen
    dims[0] |= 0x80 << 8   # GCTF bit set
    dims[-1] |= 7          # color resolution
    dims.append((height-1) & 0xff)
    dims[-2] |= ((height-1) >> 8) << 8    # GCTF bit set

    header = b'GIF89a' + bytes([dimlen]) + \
             struct.pack('>{}B'.format(1+dimlen), *dims)

    numcolors = rng.randint(2,2**12-1)     # TODO: is this right?
    palette = []
    for _ in range(numcolors):
        r = rng.randint(0,255)           # TODO: include alpha?
        g = rng.randint(0,255)
        b = rng.randint(0,255)
        palette.extend([b,g,r])

    imgdata = []

    # Each frame contains an LZW code table with exactly <numcodes> entries.
    # We want to have at least one zero-length entry followed by up to sixteen
    # nonzero entries (for portability reasons, see section 19 of the spec).
    # To achieve that goal, we pick a random value for each of these parameters.
    numcodes = rng.randint(1,16)
    clearcode = rng.randint(0,2**numcodes-1)
    endofblockcode = rng.randint(0,2**numcodes-1)
    eoi = rng.randint(0,2**numcodes-1)

    lzwbits = numcodes.bit_length()       # at most 4 bits
    bufsize = 2**(lzwbits+1)              # required buffer size
    bufbits = bufsize - 1                 # how many extra bits are needed?

    # write the LZW code table to the output image
    imgdata.extend([(clearcode<<lzwbits)|bufbits,(endofblockcode<<lzwbits)|0])
    for i in range(1,numcodes):
        imgdata.append(((i)<<lzwbits)+i)

    # generate the compressed image data
    bitsused = len(imgdata)*2 + bufbits
    while True:

        # add bufbits to the image data until it's full
        while bitsused // 8 < bufsize:

            # read two bytes from the input stream
            word = tuple(
                    int(k) for k in
                    hex(int.from_bytes(
                        rng.read(2), byteorder='big'))[2:])

            # check if we need more codes
            if word[:2]!= (0,0):

                if word[0] >= numcodes:

                    # store a new code
                    newlen = word[0] - numcodes + 1
                    if newlen == 1:         # length change!
                        imgdata.append((word[0]<<lzwbits)|bufbits)
                    elif newlen == 17:      # length change again!
                        imgdata.append((word[0]<<lzwbits)|newlen)
                    elif newlen == 18:      # length change last time!
                        imgdata.append((word[0]<<lzwbits)|newlen)
                    else:                   # normal length-change case!
                        imgdata.append((word[0]<<lzwbits)|newlen)
                    numcodes += newlen
                else:
                    # just use an existing code
                    pass

                # try to match the second byte as well
                ref = 1
                if word[ref] >= numcodes:    # get ready for triple matches!
                    ref += 1
                if word[ref] >= numcodes:
                    ref += 1
                if word[ref] >= numcodes:
                    raise Exception("this should never happen!")

                # now we're really at it...
                if (tuple(imgdata[-2::-2]),word[1]) in words:
                    imgdata.pop(); imgdata.pop()     # drop the pair
                    imgdata.append((((word[1]+numcodes)<<lzwbits|
                                     (((word[0]-numcodes)%19)+1))))
                else:
                    if word[ref] >= numcodes:    # do we have a triple match?

                        trip = (words[(tuple(imgdata[-3::-2]),
                                       word[2])],
                                words[(tuple(imgdata[-2::-2]),
                                       word[1])],
                                words[(tuple(imgdata[-2::-2]),
                                       word[0])])

                        triplen = sum(trip)-numcodes # yes, so this must be three!

                        if triplen > 12:
                            raise Exception("this should never happen either")

                        firstbyte = (((trip[0]+numcodes)<<lzwbits)) | \
                                    ((triplen+1) % 19)

                        if triplen == 0:                  # special encoding for triples!
                            imgdata += [(firstbyte<<2)]
                        else:
                            thirdbyte = (((trip[2]+numcodes)<<lzwbits)) | \
                                        ((trip[1]-numcodes+1) % 19)

                            imgdata += [firstbyte,thirdbyte]
                    else:                           # nope, only a pair
                        imgdata.append((((word[1]+numcodes)<<lzwbits|
                                         (((word[0]-numcodes)%19)+1))))

                # update our count on what bits we've used
                bitsused += 2*lzwbits+1

            else:

                # the first two bytes were both zero; add an EOB code instead
                imgdata.append(endofblockcode<<(lzwbits+1))
                break        # skip the next part; everything else is done

        # produce final bits of the stream
        val = int.from_bytes(rng.read(1),'big')
        bitsused -= 8
        while bitsused < 0:
            val <<= (-bitsused)
            val |= int(rng.read(1))
            bitsused += 8

        # stop when we hit the EOI code
        if val == eoi:
            break


    # compress the image data
    scanline_size = ((width - 1) // 8) + 1
    compstream = b''
    for line in range(height):

        # initialize values according to the GIMP decoder source
        offset = 0                      # start of uncompressed data
        start = 1                       # number of pixels read after previous code
        cur = codes = 2**lzwbits         # current code
        prev = endofblockcode            # previous code

        def emit():

            # add an unused trailing zero if necessary -- why?
            if start!= offset:
                compstream += bytes([start])

            return cur, codes


        # go through the pixels and encode them in LZW compression
        imgdata.insert(0,prev)             # fill initial buffer
        while True:
            index = (val>>(8-bitsused))&0xff
            prev = imgdata[offset]         # grab another pixel
            imgdata[offset] = cur          # remember its code
            offset += 1                    # advance past it
            bitsused -= lzwbits            # remove some bits

            # look up the new code in the dictionary
            try:                          # TODO: handle errors better
                cur = words[cur,index]
            except KeyError:

                # let's make things faster later by making sure there's always room
                if offset == len(imgdata):
                    imgdata.append(-1)

                # don't forget about the old stuff too
                while offset >= bufsize:
                    compstream += bytes(imgdata[start:bufsize])
                    del imgdata[start:bufsize]
                    start = 0
                    offset -= bufsize

                # find the closest match to the new code
                bestdist = float('inf')        # infinity distance
                bestcode = 0
                for cand in imgdata[start:]:

                    dist = abs(cand-(cur>>lzwbits))+abs(index-(cur&bufmask))
                    if dist < bestdist:
                        bestcode = cand
                        bestdist =