#!/usr/bin/env python3

import os, sys
import struct
from math import ceil, log2
from typing import BinaryIO

# 64 bytes of palette information with an end-of-file block
palette_data = b'\x00\x01\xff\x00\xff\xe7\x1c\xf5' * (ceil((256+1)/8)) \
               + b'\x00'

def gen_bitmask() -> list[int]:
    mask=[0] * 8

    # Generate first bit mask for palette size code
    while True:
        if len(mask) == 8 and mask[-1]==1: break

        k = sum([mask[i]*pow(2, i) for i in range(len(mask))])
        mask.append(((k & 2) >> 1)^(k & 1))

    # Generate remaining masks
    return [sum([m*pow(2, i) for i, m in enumerate(reversed(mask[:n]))])
        for n in range(2, len(mask)+1)]

def write_palette(out:BinaryIO):
    '''Write our standard 256 color palette to `out`.'''
    out.write(palette_data)

def write_frame(width:int=256, height:int=256, bgcol:int=-1, fgcol:int=-1,
        delay:int=0, trans:bool=False, disposal:int=None, lct:bool=True,
        interlace:bool=True, sort:int=1, out:BinaryIO=sys.stdout):
    '''Generate a frame entry in a GIF containing just solid pixels.'''
    if not (1 <= width**2 <= 2 ** 12):
        raise ValueError('Invalid image dimensions.')
    w = int((width - 1).to_bytes(2,'little'))
    h = int((height - 1).to_bytes(2,'little'))
    
    # Set up local colour table by default but update it later
    lct_size = 2 << ((w + h).bit_length()-1) # Size is nearest power of two
    if lct_size < 7: lct_size = 7 # Do at least one byte per entry

    lct_size = min(lct_size, 256) # Limit palette size since we use only 8 bits
    lct = lct_size > 0 # If there are no entries we can remove the LCT entirely
    
    # Determine if we have transparency
    has_trans   = (fgcol >= 0 or bgcol >= 0)
    has_bg      = bgcol!= None;    has_fg = fgcol!= None
    has_trans   = has_trans or has_bg or has_fg
             
    # We're going to assume that all control flags have been set correctly
    flagbyte = 0b00000000
    flagbyte |= 0b00000001 if interlace else 0
    flagbyte |= 0b00000010 if has_trans else 0
    flagbyte |= 0b00000100 if lct else 0
    flagbyte |= 0b00010000 if disposal else 0
    flagbyte |= 0b10000000 if sorted else 0

    # Write each field individually
    out.write(flagbyte.to_bytes(1,'little'))
    out.write(delay.to_bytes(1,'little'))
    out.write(0.to_bytes(1,'little')) # Transparent pixel index value
    out.write(w.to_bytes(2,'little'))
    out.write(h.to_bytes(2,'little'))

    # If we have transparency, write the background colour as well
    if has_trans: 
        if bgcol<0: bgcol = 0;
        elif bgcol>255: bgcol = 255
        out.write(bgcol.to_bytes(1,'little'))
    
    # Write the LCT (local colour table) which contains the foreground colour 
    # as its final entry. This means transparent GIFs will work without having to
    # do anything special when assembling the frames.
    if lct:
        write_palette(out)
        out.write(fgcol.to_bytes(1,'little'))
        out.seek(max(out.tell(), lct_size+1)) # Pad out to full length

    # Finally write the image itself
    num_pixels = width * height
    packed_bits = []
    packer = '>' if w>=h else '<'
    for i in range(num_pixels):
        packed_bits.extend([fgcol&255, fgcol>>8])
    out.write(struct.pack(packer+'B'*len(packed_bits), *packed_bits))

def generate_random_gif(rng: BinaryIO, out: BinaryIO,
        width:int=256, height:int=256,
        bgcol:int=-1, fgcol:int=-1, delay:int=0,
        trans:bool=False, lct:bool=True, init_disposal:int=None,
        interlace:bool=True, sort:int=1, repeat:int=0,
        ):
    '''Write a complete valid GIF file from scratch into `out`. Random numbers 
    are read from `rng`.'''

    # Default values
    if bgcol==-1: bgcol = 255
    if fgcol==-1: fgcol = 0

    # Generate palette data if necessary
    if out._mode!='rb':
        write_palette(out)

    # Grab info about this stream so we know where to seek after closing the file
    initialpos = out.tell()
    newpos = lambda p : max(initialpos+p, out.seek(0, 2))

    # Write headers and logical screen descriptor
    out.write(b'GIF89a')
    global_cols = 2**log2(out.seek(0, 2)).to_bytes(1, 'little')
    out.seek(-1,1)
    out.write(global_cols)
    out.write(b'\x00\x00')
    out.write(width.to_bytes(2, 'little'))
    out.write(height.to_bytes(2, 'little'))

    # Write application extension blocks for extra info
    exttype = b'NETSCAPE2.0'
    appdata  = b'\x03\x01\x00\x00'
    apploop  = os.environ['LOOP'].to_bytes(2, 'little')
    exthdr   = b''
    exthdr  += len(exttype).to_bytes(1, 'little')
    exthdr  += exttype
    exthdr  += len(appdata).to_bytes(1, 'little')
    exthdr  += appdata
    exthdr  += len(apploop).to_bytes(1, 'little')
    exthdr  += apploop
    out.write(exthdr)

    # Start writing actual images
    for i in range(repeat+1):
        out.write(b';')
        startblock = out.tell();
        
        # Add frame
        write_frame(width, height, bgcol, fgcol,
                delay, trans, init_disposal, lct, interlace, sort, out)

        # Backtrack to rewrite frame count on last iteration
        if i==repeat: continue

        # Seek back over the ending character and overwrite it with image separator
        out.seek(startblock);
        out.write(b',')

    # Update loop counter to match
    out.seek(newpos(-1)-1)
    out.write(os.environ['LOOP'].to_bytes(2, 'little'))