#!/usr/bin/env python3

"""
Generates a random GIF file of size 0 to 2^16 bytes and writes it to stdout.
The GIF should be valid (according to https://www.w3.org/Graphics/GIF/spec-gif89a.txt).
"""

import sys
import struct


class RandomGifGenerator(object):

    def __init__(self, rng=None):
        if rng is None:
            self._rng = open('/dev/urandom', 'rb')
        else:
            assert isinstance(rng, BinaryIO)
            self._rng = rng

        # According to the spec: "A Screen Descriptor specifies image attributes"
        self.screendescriptor_width, self.screendescriptor_height = [
                x for x in range(1, 17)]

        # According to the spec: "Each image must begin with an Image
        #   Descriptor, which gives its width and height."
        self.imagedescriptor_width, self.imagedescriptor_height = [
                x for x in range(1, 17)]


    @staticmethod
    def get_max_size(): return pow(2, 16) - 1


    def generate(self, size):
        assert size < RandomGifGenerator.get_max_size()
        
        print('Generating {} bytes'.format(size))
        header = b'GIF89a'
        descriptortype = b'\x2c'
        screendescriptor = StructBuilder().width(self.screendescriptor_width)\
               .height(self.screendescriptor_height)\
               .build()
        imagedescriptor = \
                StructBuilder()\
                   .left(0).top(0)\
                   .width(self.imagedescriptor_width)\
                   .height(self.imagedescriptor_height)\
                   .packed_flag('\x08')\
                   .packed_localcoloralphabitmask(b'\xff')\
                   .build()
        data = self._rng.read(size)
        payload = descriptortype + len(data).to_bytes(1, 'big') + data
        
        return header + screendescriptor + imagedescriptor + payload


class StructBuilder(object):

    # The first argument is named'self' just because we're lazy about
    # having to type super().__init__() every time.
    def __init__(self):
        pass

    
    def build(self):
        fmt = ''
        data = []
        for k in sorted(vars(self)):
            v = getattr(self, k)

            try:
                fmt += v[0]
            except TypeError as e:
                raise Exception(
                        '`{}` has invalid field format {}'.format(k, str(v))) from e

            try:
                data.append(v[1])
            except IndexError as e:
                raise Exception('{} does not have enough arguments'.format(k)) from e

        try:
            ret = struct.pack(fmt, *data)
        except struct.error as e:
            raise Exception('Invalid format string or values ({})!'.format(str((fmt, data)))) from e

        return ret


    def left(self, value):
        return ('<H', value)


    def top(self, value):
        return ('<H', value)
    
    
    def width(self, value):
        return ('<H', value)


    def height(self, value):
        return ('<H', value)
    

    def packed_flag(self, flag):
        """Must be one of "\xF7", "\xFE" or "\xFF"."""
        assert flag in ['\xf7', '\xfe', '\xff']
        return ('B', ord(flag))
        
    
    def transparentindex(self, index):
        return ('B', index)


    def interlacemode(self, mode):
        return ('B', mode)


    def localcolortableflag(self, flag):
        assert flag in ['\x00', '\xc0']
        return ('B', ord(flag))


    def gctsize(self, size):
        """Must be one of "\x00", "\x01", "\x02"..."\xFF"."""
        assert 0 <= ord(size) <= 0xff
        return ('B', ord(size))


    def sortflag(self, flag):
        assert flag in ['\x00', '\x08']
        return ('B', ord(flag))


    def colorresolution(self, resolution):
        return ('B', resolution)


    def globalcolortablesize(self, tablesize):
        return ('B', tablesize)


    def backgroundcolorindex(self, bgcolor):
        return ('B', bgcolor)


    def pixelaspectratio(self, ratio):
        return ('B', ratio)


    def packed_localcoloralphabitmask(self, mask):
        """Must be one of "\x01", "\x02"..."FF"."""
        assert 0x01 <= ord(mask) <= 0xff
        return ('B', ord(mask))



if __name__ == '__main__':
    assert len(sys.argv) >= 2
    size = int(sys.argv[1], base=0)
    gen = RandomGifGenerator()
    result = gen.generate(size)
    sys.stdout.buffer.write(result)