#!/usr/bin/env python3

import binascii
from typing import BinaryIO


# Produces an ascii representation of a bytestring as a string.
def p_raw(s): return ''.join([f'{i:02x}' for i in s])


# A Header consists of 7 bytes:
# 'GIF87a' or 'GIF89a', then a width and height (each uint16le), then 0,
# followed by ',' then a single color table entry (RGB triple).
def create_header():
    return [b'GIF89a',
            # TODO: make it possible to have different sizes
            b'\xff\xff\x00\x00',       # 160x120
            b'\x00\x00\x00\x00',         # No global palette
            b'\x2c',             # Extension block introducer
            b'\x00',             # Block size
            0x00, 0x00,      # Width x Height (uint16le)
            0b00100000,                       # Flags
            ] + list(range(0, 0x0f))          # Color table entries (RGB triples)
                                            # NOTE: There are actually only 15 colors available

# The Palette extension is required because we're using transparency
def create_palette(has_transparency=False):
    extlen = 0x10 if has_transparency else 0x0e

    res = []
    res += [b'\x21']                        # Extension label
    res += [extlen]                         # Extension length
    res += [ord('P'), ord('A')]             # Type identifier
    res += [4]                              # Number of values per pixel
    res += [(int)(not not has_transparency)] # Transparency flag
    res += [(int)(not not has_transparency)] # Sort flag
    res += range(0x00, 0x0f)                # Actual colors (RGB triples)

    return res



# A Graphics Control Extension is optional. It's used to show or hide
# a given frame (or part thereof).
def create_graphicsctl(has_transparency=False):
    extlen = 0x04 if has_transparency else 0x03

    res = []
    res += [b'\x21']                # Extension label
    res += [extlen]                 # Extension length
    res += [ord('F'), ord('9')]     # Type identifier
    res += [0x04 | bool(has_transparency)]  # Flags & transparency flag
    res += [0x00] + ([0xff] * 3)    # Delay time (in ms)
                                    # 1st value should be 0 but isn't taken into account anyway
                                    # Color index (optional)

    return res


# An Image Descriptor tells how many pixels wide each row of the image is,
# as well as how long that row is supposed to be. If the image is interlaced,
# this also tells us which rows are interlaced vs non-interlaced.
def create_imagedesc(width, height, has_transparency=False, interlace=True):
    return [b'\x2C'                     # Image descriptor marker
          ,width & 0xFF                # Left position of image
          ,height & 0xFF               # Top position of image
          ,((width >>  8) & 0xF ) | (((height >> 8) & 0xF ) << 4) # Header Colors
          ,0x00                        # LZW minimum code size
          ,0                          # Lowest bit means no local color table
          ,(((height & 0xFF) > 0) >> 7) |
              ((bool(has_transparency) ^ bool(interlace)) << 7)
          ,(height & 0xFF)            # Height of image
          ,(width >> 8)               # Width of image
           ]




# For lack of a better name... This function generates a GIF file;
# the output can be piped to a file via "python3 generate_random_gif.py >../a.gif".
# By default, the generated result will be 160x120 pixels large with no local colour
# table and without transparent colours. These can be changed via command line arguments.
if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()

    parser.add_argument('-W', '--width', type=int, help="Width", metavar='<pixels>', default=160)
    parser.add_argument('-H', '--height', type=int, help="Height", metavar='<pixels>', default=120)
    parser.add_argument('--no-local-colourtable', action='store_true', help="Disable the use of a local colour table")
    parser.add_argument('--with-transparency', action='store_true',
                        help="Enable the use of a transparency flag in the graphics control extension")
    args = parser.parse_args()


    # Create our initial GIF and write it to stdout
    gif = create_header()

    # Write the Image Descriptor
    gif += create_imagedesc(args.width, args.height,
                            has_transparency=args.with_transparency,
                            interlace=not args.no_local_colourtable
                           )



    # Add the palette extension
    if not args.no_local_colourtable:
        gif += create_palette(has_transparency=args.with_transparency)

    # Add the graphic control extension
    if args.with_transparency:
        gif += create_graphicsctl()

    # Print the resulting "bytecode"
    print("".join(chr(i) for i in gif))