#!/usr/bin/env python3

import sys
import os
import argparse
import secrets
import math


# Generates a random GIF using the given RNG and writes it to the output stream
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Read in the header (which is static for this format)
    head = 'GIF89a'.encode('ascii')
    header = bytes()
    while len(header) < 6:
        header += next(rng).to_bytes(1, byteorder='big')

    if not header.startswith(head):
        raise ValueError('Invalid magic number at start of file')

    # Read in the global color table length
    gctlen = int.from_bytes(next(rng).to_bytes(1, byteorder='big'),
                            signed=False, byteorder='big') & 0x7F

    # Process the global color table
    gct = list()
    if gctlen > 0:
        for _ in range(gctlen+1):
            r = next(rng) >> 2
            g = ((next(rng) << 4) |
                 (next(rng) >> 4)) >> 2
            b = ((next(rng) << 2) |
                 (next(rng) >> 6)) >> 2
            gct.append((r, g, b))


    # Process the logical screen descriptor
    lsd = [next(rng) for _ in range(7)]
    width = (((lsd[5] & 0xFF) << 8) |
             lsd[4])
    height = (((lsd[7] & 0xFF) << 8) |
              lsd[6])
    gctused = (~lsd[8] & 0b00011000) >> 3
    bgidx = ((~lsd[8] & 0b00000111) << 3) | \
            ((lsd[9] & 0b11100000) >> 5);
    aspect = ((lsd[9] & 0b00011111) << 1)

    # Write the LSD back to the stream so we can calculate offsets as needed
    out.write(bytes(lsd));

    # Calculate offsets
    imagedescriptor_offset = 13;
    lzwmincodesize = 2
    count = 2**(lzwmincodesize + 1);
    clearcode = 1 << (lzwmincodesize - 1);
    endofinfo = clearcode + 1;
    eoifound = False

    # Dithering matrix used to simulate an 3-bit color palette
    dithmat = [[0, 8, 2, 10],
               [12, 4, 14, 6],
               [3, 11, 1, 9],
               [15, 7, 13, 5]]

    # The raster line counter
    rowcount = 0

    # Output the header again since LZW uses the header
    out.write(header)
    out.write(((gctlen).to_bytes(1,'big'))[0]);

    # Main generation loop
    while True:

        # Back up one byte because LZW codes are always preceded by their own
        # code value and terminated with a 0xff byte which needs to be skipped.
        # This only applies after we have already written a single frame.
        if rowcount == 0:
            lastbyte = b'\xff'
        else:
            lastbyte = (rowcount-1).to_bytes(1, 'big')


        if rowcount >= height:
            break

        imgdata = []
        lzwmincodesize = 2

        while True:

            # Get the chunk size from the input stream
            chunksz = next(rng) ^ 0xFFFF
            if chunksz == 0xFFFE or chunksz == 0xFFFF:
                continue

            # If the first bit is set then that means there's a leading zero
            # that has to be added to the chunksz calculation.
            if chunksz!= 0:
                chunksz |= 0b1000000000000000


            # Read in the chunk
            chunk = rng.read(chunksz)

            # Each chunk must terminate with either a 0 or ffff
            highbyte = chunk[-2:]
            lowbyte = chunk[-1:]

            if lowbyte!= b'\x00':
                raise Exception("Invalid chunk termination")

            if highbyte!= b'\xff\xfe' and highbyte!= b'\xff\xff':
                raise Exception("Invalid chunk termination")

            # Process the chunk
            for i in range(1, len(chunk)):

                # Get current and previous pixel values
                currpx = tuple(imgdata[-1][i:])
                prevpx = tuple(imgdata[-1][:i]) if i>0 else (0,0,0)

                # Determine the new pixel value based on the difference between
                # the two pixels and the dither matrix
                newpx = [(currpx[j]+dithmat[(prevpx[j]-currpx[j])%4])
                         % 2**3 for j in range(3)]


                # Write the new pixel to the output stream
                data = bytes([(newpx[0]<<5)+(newpx[1]<<2)+newpx[2]])
                imgdata.append(data)


            if highbyte == b'\xff\xfe':
                break



        # Append an empty byte to the end
        imgdata.append(b'')

        # Try to find the EOI marker
        offs = 0
        while True:

            # Find all instances of the EOI sequence
            idx = imgdata.find(b'\xff\xff', offs)

            # Either write down where we found it or exit
            if idx>=0:
                offs = idx+2
            else:
                imgdata.insert(offs, b'\xff\xff')
                break

        # Create the LZW dictionary
        dict_ = {}
        for (i, v) in enumerate([tuple(v) for v in imgdata]):
            dict_[v] = i

        # Convert the indexed image data to raw binary form
        binimg = []
        for c in imgdata:
            val = dict_[c];
            if val<count:
                binimg.append(val)
            elif val==clearcode:
                lzwmincodesize = 2
                count = 2**(lzwmincodesize + 1);
                binimg.append(clearcode)
            elif val==endofinfo:
                print("End-of-information reached!")
                return;
            else:
                binimg.extend([dict_[tuple(
                    buf[i-lzwmincodesize:i])] for i in range(
                        len(buf)-lzwmincodesize)])
        binimg = bytes(binimg)




        # Exit early if we are done encoding
        if eoifound:
            break

        # Write the local color table descriptor
        out.write((((gctused*3)).to_bytes(1,'big'))[0])
        out.write((((bgidx).to_bytes(1,'big'))[0]))

        out.write(width.to_bytes(2,'big'));
        out.write(height.to_bytes(2,'big'));

        # Write the image descriptor
        out.write((imagedescriptor_offset&0xFF).to_bytes(1,'big'))
        out.write(((imagedescriptor_offset>>8)&0xFF).to_bytes(1,'big'))
        out.write((width.to_bytes(2,'big')))
        out.write((height.to_bytes(2,'big')))
        out.write((8).to_bytes(1,'big'))
        out.write((0).to_bytes(1,'big'))    # Local color table flag
        out.write(bytes([aspect]))            # Aspect ratio
        out.write(bytes([0]))                 # No interlace, etc...
        out.write(bytes([0]))                 # No extra flags

        # Adjust the offset so that the next frame starts where this one ends
        imagedescriptor_offset += 10 + math.ceil(len(binimg)/12)*12

        # Reset row counter
        rowcount+=1



# Command line arguments
parser = argparse.ArgumentParser(description='Generate a random GIF.')
parser.add_argument('-o,--output','-O', dest='outfile', type=str, default='',
                    help="Output filename")
parser.add_argument('-n,--number', '-N', dest='numframes',type=int, default=-1,
                    help="Number of frames to produce")
args = parser.parse_args();


if args.outfile=='':
    outfile = sys.stdout.buffer
else:
    outfile = open(args.outfile, mode='wb')

if args.numframes==-1:
    numframes = float('+inf');
elif args.numframes<=0:
    raise ValueError(f"Number of frames cannot be <=0 ({args.numframes})"); 

with open('/dev/urandom', mode='rb') as rng:
    
    # Generate the file
    try:
        while numframes>0:
            
            if numframes!=float('+inf'):
                numframes-=1

            generate_random_