#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (for example /dev/urandom).
# The given width and height will be used for the frame size
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width: int, height: int
    ):

    # Write GIF header
    out.write('GIF89a'.encode('ascii'))
    
    # Write logical screen descriptor
    out.write(struct.pack('<HHBBB',
            0, 0,      # Screen width and height in pixels
            -7 & ~1,   # Global color table flag with local color table
            height >> 8, height % 256,    # Height of the image in packed format 
            width >> 8,  width % 256))     # Width of the image in packed format 

    # Write global color table
    palette = [None] * 256
    for i in range(256):
        r = next(rng) % 255
        g = next(rng) % 255
        b = next(rng) % 255;

        palette[i] = bytes((r,g,b))
        
    for p in palette:
        out.write(p)

    # Write first graphic control block which specifies no delay or transparent color index
    out.write(struct.pack('<BHBcbb', 
            0x2C,                # Start of graphic control extension block
            width >> 8,          # Packed field. Width of image in packed format
             width % 256,        # Width of image in packed format
            0,                   # Packed fields. No delay and black background
            0,                   0,       # Transparent color index is not specified
            ))


    lzwminsize = 2
    clearcode = lzwminsize
    endofinfo = clearcode + 1
    bitsperpixel = 8
    maxcodebits = 12
    mincodesize = 2
    codesize = mincodesize + 1
    codemask = ((1 << codesize)-1) 
    nbitsperinput = 8         # bits per input code
    vshift = bitsperpixel-nbitsperinput  
    ffcorrection = 32768    
    finfomask = (((clearcode<<vshift)|ffcorrection)<<bitsperpixel)+endofinfo
    lomask = ((1<<(bitsperpixel+1))-1 ) ^finfomask
    himask = (~lomask)&(((1<<(bitsperpixel))-1) |1)
    dlenmask = ((1<<bitsperpixel)-1)

    # Write LZW minimum code size to data subblock
    out.write(bytes([lzwminsize]))

    # When there are less than 256 colors, write color map to data subblock
    if len(palette)<256:
        cmdata = []
        for c in palette:
            cmdata += list(c)
            
        cmbytes = bytes(cmdata)
        out.write(bytes([(len(palette))]) + cmbytes)

    # Loop over all frames
    for _ in range(int(width*height)):
        # Reset LZW compression state before each new frame
        dictsize = 4096
        data = [clearcode]
        dataptr = 0
        datacount = 1
        
        while True:

            # Generate a raw chunk of length <= 255
            bytestocopy = max(next(rng) % 256, 1)
            
            # Check whether this is the last chunk of the current frame
            if bytestocopy<=datacount:
                break

            # Copy generated bytes from the front of the byte buffer to the back
            tmpbuf = data[-bytestocopy:] + data[:-bytestocopy]
            data = tmpbuf[:dataptr] + data[dataptr:-bytestocopy] + tmpbuf[dataptr:]
            datacount -= bytestocopy
            dataptr = len(tmpbuf) - bytestocopy
            assert(len(data)==dictsize)

            # Compress data until there's only one code left on the dictionary.
            # The resulting code must fit within the available space for codes
            # encoded in the LZW minumum code size.
            code = data[dataptr]
            while dataptr>0:
                dataptr-=1

                # Get previous token
                prevtoken = data[dataptr]
                
                # Look up token in dictionary, add it to the dictionary if necessary
                idx = prevtoken|dlenmask
                if idx>=len(data) or data[idx]!=prevtoken:
                    if dataptr<1:
                        raise Exception("LZW code would have overflowed")

                    # Add new entry to dictionary
                    data[dictsize]=code
                    dictsize+=1
                    
                    # Use high order bits as table lookup address 
                    idx = ((((prevtoken>>vshift)*dictsize)>>codesize)&codemask)|dlenmask
                    data[dataptr]=(prevtoken|(dictsize<<vshift))&himask
                    code=idx

                else:
                    code=idx
                    
                # Store pointer to previously decoded value
                data[dataptr]=code

            # Increase code size when there are too many entries already in the dictionary
            if dictsize>(1<<(codesize))+257 :
                codesize+=1
                codemask=(1<<codesize)-1
                vshift=bitsperpixel-nbitsperinput
                dataptr-=1
                
        # Output number of bytes required to encode LZW compressed tokens minus two
        # because we don't store an initial clear code and final end-of-information code
        numbytes = len(data) - 2
        out.write(numbytes.to_bytes(1,'little'))
        for i in range(dictsize):
            if i < len(data):
                out.write(data[i].to_bytes(1,'little'))
                
            elif i == clearcode:
                out.write(b'\x00\x00')
                
            elif i == endofinfo:
                out.write(b'\xFF\xFF')
                
