#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (provided as an open binary file).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Generate the header & comment block
    header : bytes = 'GIF89a'.encode('ascii')   # GIF signature
    for _ in range(4):                          # Reserved fields (4x 0)
        header += struct.pack('B', 0)           # Each field is a single byte
    width = rng.randint(16,255)                 # Random canvas size
    height = rng.randint(16,255)                # Interpretation of LZW minimum code size
    depth = max([width,height]) // 4 + 1        # LZW minimum code size
    bgcolor = rng.randint(0,7)                  # Background color index
    aspectratio = rng.choice((0b0000,    # No Aspect Ratio
                            0b0001,     # 4:3
                            0b0010,     # 16:9
                            0b0011))    # 2:1
    flags : bytes = bin(flags)[2:].zfill(8)       # Some other flags
    gctsize : bytes = str(depth - 1).encode()   # GCT size minus one
    ext_hdrs : bytes = bin(ext_hdrs)[2:]         # Should be empty
    while len(ext_hdrs) < 6:                     # Pad with zeroes to length 6
        ext_hdrs += '0'                        # each extension block has
                                                    # exactly sixteen bits
    interlace = rng.randint(0,1)               # Interlacing flag
    raw_data : bytes = ''                       # Output buffer
    palette : list[tuple] = []                  # Color table
    for i in range(1 << depth):
        r = rng.randint(0,255)                   # RGB triples are stored on little-endian machines
        g = rng.randint(0,255)
        b = rng.randint(0,255)
        # Pack them and add to output buffer
        raw_data += struct.pack('<BBB',r,g,b)
        palette.append((r//8,g//8,b//8))          # Add RGB triples to palette
    # Generate the logical screen descriptor
    lsd_hdr : bytes = struct.pack('BBHHBBBB',
                                width&255,(width>>8)&255,
                                height&255,(height>>8)&255,
                                1<<depth-1,             # Global palette size
                                bgndcolor=bgcolor,
                                aspectration=aspectratio,
                                flags=flags,
                                        gct=gctsize,
                                            ext=ext_hdrs[:6],
                                                lsd=0xF9,  # Terminator
                                interlace=interlace)
    lsd_dat : bytes = CFF_DICT(palette)              # Write colormap dictionary to logical screen descriptor
    # Calculate global color table size and locater (end of logical screen descriptor + 4)
    gct : tuple = ('GCT',
                    struct.pack('B', depth),
                        struct.pack('<H',len(raw_data)))
    # Generate image descriptor
    img_desc : tuple = ('IMG',
                        struct.pack('BBHHBBBHB',
                                    0,0,
                                    width,height,
                                    1<<depth+1,
                                        1<<(depth+1)-1,# Local palettte size
                                        interlace,
                                                0,))
    # Write the entire GIF file
    out.write(header)                                   # Header
    out.write(struct.pack('<L',int(datalen)))            # Data length
    if len(comment)>0:                                  # Comment section
        out.write(struct.pack('<B',len(comment)+2))      # Length of comment plus terminator
        out.write(comment)                              # Comment string
        out.write(struct.pack('<B',0))                  # Terminator
    else:                                               # Empty comment
        out.write(struct.pack('<BB',0,0))               # Zero length
    if len(lsd_hdr) > 0xF9:                             # Large data blocks must start with a label byte
        out.write(struct.pack('<B',0xF9))               # Label byte
        out.write(lsd_hdr[:0xF9])                      # Logical screen descriptor
        rem = len(lsd_hdr[:0xF9]) % 2                   # If remainder not zero pad
        if rem!= 0:                                     # with zeroes so that length is even
            out.write(b'\x00\x00')
        out.write(lsd_hdr[-rem:])                       # Rest of logical screen descriptor
    elif len(lsd_hdr) == 0xF9:                         # Small data blocks may be unlabeled
        out.write(lsd_hdr)                               # Just write it directly
    else:                                               # It's just plain small data
        out.write(struct.pack('<B%ds'%len(lsd_hdr),0x2C,lsd_hdr))
    if len(lsd_dat) > 0xFF:                              # Repeat the same process for the subblocks
        out.write(struct.pack('<B',0xFF))
        out.write(lsd_dat[:0xFF])
        rem = len(lsd_dat[:0xFF]) % 2
        if rem!= 0:
            out.write(b'\x00\x00')
        out.write(lsd_dat[-rem:])
    else:                                               # In this case there isn't any padding required
        out.write(struct.pack('<B%ds'%len(lsd_dat),0x2C,lsd_dat))
    out.write(struct.pack('<B',0x3B))        # Write out final terminator