#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The
# output is guaranteed to be in order with increasing
# frame numbers (i.e., no duplicate frames).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write 1st block
    write_chunk('IHDR', rng, out, 10)
    
    color_table_size = 2 ** int.from_bytes(
            rng.read(1), byteorder='little') - 1
    bytes_per_pixel = max(1, min(color_table_size, 2)) * 3
    local_color_table_flag = bool(int.from_bytes(
            rng.read(1), byteorder='little'))
    interlace_flag = bool(int.from_bytes(
            rng.read(1), byteorder='little'))
    sort_flag = bool(int.from_bytes(rng.read(1),
                                     byteorder='little'))
    global_color_table_flag = not local_color_table_flag and \
                              not interlace_flag
    palette_resolution = rng.read(1)[0] & 7
    sorted_palette_flag = True if sort_flag else False
    size_of_global_color_table = pow(2, (6 + palette_resolution))
    num_colors_in_local_color_table = pow(
            2, (4 + palette_resolution)) / bytes_per_pixel
    transparency_flag = bool(int.from_bytes(
            rng.read(1), byteorder='little'))
    size_of_transparency_index_field = 4

    # Write 2nd block (optional - depends on global_color_table_flag)
    if global_color_table_flag:
        assert(not interlace_flag)

        colors = [make_color() for _ in range(
                len(size_of_global_color_table))]
        
        while any([c == make_transparent_color()
                   for c in colors]):
            pass
            
        write_chunk('gctc', rng, out, 
                    size_of_global_color_table*3+1)
    
        for c in colors:
            write_rgb(c, out)
            
    # Write 3rd block (always exists)
    write_chunk('PLTE', rng, out, 
                color_table_size*(5//3)*num_colors_in_local_color_table)
        
    for i in range((color_table_size //
                    num_colors_in_local_color_table)):
        for j in range(len(
                        size_of_local_color_table)):

            for k in range(num_colors_in_local_color_table):
                c = make_color()

                if c == make_transparent_color():
                    continue
                    
                write_rgb(c, out)
                
    # Write 4th block (always exists)
    write_chunk('IDAT', rng, out, 10)
    write_no_data(rng, out)

    loop_count = rng.read(1)[0]

    # Write 5th block (optional)
    if transparency_flag:
        chunk_type = 'tRNS'
        size = len(size_of_transparency_index_field)
        write_chunk('tRNS', rng, out, size + 10)

        transparency_index = rng.read(1)[0] % \
                             color_table_size

        if transparency_index == 0:
            transparency_index += 1

        write_number(transparency_index, out)

    # Write 6th block (optional)
    if loop_count > 0:
        chunk_type = 'loop'
        write_chunk('loop', rng, out, 10)

        write_number(loop_count, out)


# Write an RGB triple to `out`.
def write_rgb(rgb: tuple[int], 
              out: BinaryIO
              ) -> None:
    for v in rgb:
        binval = v & 0xff

        out.write(struct.pack('<B', binval))
        
# Write a chunk of length `length` starting with `prefix`
# under the name `chunk_type`, whose raw contents are generated by
# reading from `rng`, to `out`.
def write_chunk(chunk_type: str, 
                rng: BinaryIO,
                out: BinaryIO,
                length: int) -> None:
    chunk_size = struct.pack('<H', length | 0x8000)
    end_mark = b'\x00'

    out.write(chunk_type.encode())
    out.write(chunk_size)
    out.write(end_mark)
    
# Write some arbitrary non-zero data to `out` using the random generator
# `rng`.
def write_no_data(rng: BinaryIO, 
                  out: BinaryIO
                  ) -> None:
    data = rng.read(10)
    out.write(data)
    
# Write a 1-byte integer at least as large as n to `out`
def write_number(n: int, 
                 out: BinaryIO
                 ) -> None:
    bn = abs(n)
    extremum = ((bn >> 7) << 7) + 0x80
    signbit = 0x80 if n < 0 else 0x00
    value = bn if bn <= extremum else (extremum + 0x80)

    out.write(value.to_bytes(1, byteorder='big'))