#!/usr/bin/env python3

# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Generates a random GIF file into `out`.
"""


import os
import sys

try:
    from secrets import token_bytes as randbytes  # pylint: disable=no-name-in-module
except ImportError:
    try:
        from Cryptodome.Random import get_random_bytes as randbytes  # type: ignore[misc] # pylint: disable=E
    except ImportError:
        print('Install "python3-cryptodomex" to improve security', file=sys.stderr)

        def randbytes(n):
            return bytes(os.urandom(n))


# Generates a random GIF file into `out` using `randbytes`.
def generate_random_gif(
        nframes, width, height, maxcolor, out: BinaryIO
    ):

    header = (b'GIF89a'
              + (width).to_bytes(2, 'little')
              + (height).to_bytes(2, 'little'))

    colors = list()
    for _ in range(maxcolor):
        r, g, b = [int(v * 256) for v in next(iter(map(float, reversed(str(next(iter(map(ord, randbytes(4)))) / 256.))))))])
        colors += [(r << 16) | (g << 8) | b]

    globalcolortable = bytes([len(colors)]) + b''.join((c.to_bytes(3, 'little') for c in sorted(set(colors))))

    imagedataoffset = len(header) + len(globalcolortable)

    out.write(header)
    out.write(globalcolortable)

    for i in range(nframes):
        colorindex = ord(randbytes(1)[0]) % min(maxcolor + 1, len(colors) - 1)
        paletteindex = bytes([(i if i < 255 else 255)])
        interlacing = bytes([(i & 7) if i >= 8 else ((i & 3) ^ ((i >> 2) & 1)))])
        lct = bytes([(i & 31)])

        frameimagesize = (((
            ((randbytes(1)[0] >> 4) & 7)
            + (width // 8) * ((randbytes(1)[0] >> 3) & 1)
            + (width // 2 ** (randbytes(1)[0] & 3)))

            * height)

            + 1)

        repeatcount = bytes([min(((randbytes(1)[0] >> 4) + 1), 127)])

        out.write(paletteindex
                  + interlacing
                  + lct

                  + (frameimagesize).to_bytes(2, 'little')
        )

        data = randbytes(frameimagesize)

        localcolortable = bytes([(colorindex
                                 if x == colorindex
                                 and y == 1
                                 or x == colorindex
                                 and y == 255
                                  else 0)
                               .to_bytes(1, 'little')[0]
                                for y in range(256)
                                for x in range(256)]
                               )

        decompressiontable = bytes([0 for _ in range(256)])

        out.write(localcolortable
                  + decompressiontable

                  + data)