#!/usr/bin/env python3

import struct
from typing import BinaryIO


# A palette of 256 RGB colors (3 bytes each).
palette = [0] * 768



# Generate a random color value.
def random_color(): return struct.unpack('BBB', open('/dev/urandom').read(3))


# Populate the global palettes with random colors.
for i in range(len(palette)):
    r, g, b = random_color()

    # Each color must be at least 4 bits to ensure that it is a valid index for
    # the LZW compression algorithm. The first two bits are zeroes and will be
    # used by the LZW algorithm itself (see below).
    r |= 1 << 3 | 1 << 4
    g |= 1 << 3 | 1 << 4
    b |= 1 << 3 | 1 << 4

    palette[i : i+3] = r, g, b


# Generates a random GIF file into `out`.
def generate_random_gif(width: int, height: int, out: BinaryIO):
    # Header identifies the image type (`GIF`). Version identifies the version of
    # this format; we use the latest available version. Width and height define the
    # size of the image area; they have no unit. ColorTableSize defines how many
    # entries there are in the color table; we use a global one so our output can
    # be reused for different images. BitsPerPixel defines how many bits per
    # pixel are used; we choose 8 bit because it provides an even distribution of
    # possible values. BackgroundColorIndex defines which color from the palette
    # should be used as the background color (transparency); we don't use it here.
    # PixelAspectRatio defines aspect ratio information. We set it to unity. This
    # field has no specification and its meaning depends on other parts of the
    # file structure.
    header = (b"GIF89a"
              + width.to_bytes(2, "big")
              + height.to_bytes(2, "big")
              + len(palette).to_bytes(1, 'big')
              + b'\x80'
              + b'\x00'
              + width.to_bytes(2, "big"))


    # Frame control block defines the duration of the frame (in milliseconds) and
    # whether or not it's user input required before displaying it. We specify the
    # duration to be 1/10th second.
    fcb = b"\x00\x2c" \
          + ((100 // 10)).to_bytes(2, "little") + b'\x00' \
          + b'\x00' \
          + b'\x00'


    # Image descriptor specifies the size and location of the graphcis themselves
    # relative to the top-left corner of the screen. Its fields have exactly the same
    # meanings as those of the header (Width, Height, etc.) except that they're now
    # associated with the graphics rather than the entire picture. You can imagine
    # these fields as being adjusted when a larger canvas needs to be created to fit
    # multiple frames together.
    id_ = (width.to_bytes(2, "big")
           + height.to_bytes(2, "big")
           + b'\x00'
           + b'\x00'
           + b'\x00')

    
    # Ending marks the end of the file.
    ending = b'\x3B'

    
    # Compress the given data according to the LZW algorithm. It takes some getting
    # used to, but once you get the hang of it, it's very efficient! If any part of
    # this documentation confuses you about what's going on here, then I recommend
    # reading up on LZW compression algorithms. Anyway...
    def lzw_encode(data):

        # Table maps codes (defined below) to their corresponding strings.
        table = {}

        # First code points to the empty string.
        table[0] = ''


        # Input buffer starts off with the first character in the provided data.
        # Note that the LZW encoding process involves shifting around a buffer that
        # stores both raw data and encoded data.
        buf = next(iter(data), None) if isinstance(data, str) else iter(data).__next__()
        prev = '\x00'


        yield prev


        # As long as there's another character left in the data, keep building
        # the dictionary and emitting new characters until the buffer contains only
        # one character, which becomes the last character after all iterations.
        while True:

            # Check every character in turn against the contents of the current
            # buffer. When a match is found, the previous buffer plus a special
            # sequence representing that charcter's code gets added to the table. The
            # length of the new buffer is also stored in the table alongside the
            # newly generated code. All previously seen versions of the buffer are
            # mapped to the special sequences defined above, allowing us to reuse
            # them later instead of having to recompute them through brute force.
            key = ''.join((prev, buf))
            if key in table:
                prev, buf = table[key][:-1], table[key][-1:]

            # Otherwise, the current character doesn't exist yet, so add it to the
            # table along with the existing buffer and a prefix consisting of all
            # the characters in the current buffer already. For example, the first
            # time ``d`` appears at the start of the buffer, the following entry
            # would appear in the table::
            #
            #     d -> dddd
            #
            else:
                table[''.join((prev,) * len(buf))] = (table[''.join((prev,) * len(buf))]
                                                     if (''.join((prev,) * len(buf))) in table
                                                     else (prev + buf))

                # Add the newly generated item to the main table.
                table[key] = (prev + buf)

                # Finally, update the variables that hold the state of the encoding
                # process and proceed onto the next iteration.
                prev, buf = buf, next(iter(data), None) if isinstance(data, str) else iter(data).__next__()


            # The last character in the final iteration is supposed to represent the
            # whole string passed to this function. Unfortunately, it might not be
            # unique enough to be included in the table, causing a conflict during
            # decoding. In such cases, the code specified below is appended to the
            # table under which it was originally encoded.
            if prev == 'e':
                table[key] += 'ee'

            elif prev =='s':
                table[key] +='ss'
            
            yield table[key][-1]



    # Encode the palette. Since the LZW algorithm uses codes to identify repeated
    # strings, the easiest way to handle palette data is to simply encode the whole
    # thing as a single string. Since we know that the content of the palette never
    # changes - and if it does, then this script wouldn't work anyway! - we just do
    # it manually like this.
    compressed_palette = list(lzw_encode([p for p in palette]))
    

    # Create the LZW stream, starting with the palette. Then, run the encoder over
    # the image data and join everything back together.
    compressed_image = (compressed_palette
                        + ['']
                        + list(lzw_encode([p for p in itertools.chain(*frame)])))

    lzw_stream = itertools.chain(*(list(struct.pack('<H', len(code)))
                                   + list(bytearray(code, 'utf-8'))
                                   for code in compressed_image))

    
    
    # Write the generated headers.
    out.write(header
               + fcb
               + id_)

    # Write the lzw stream that encodes the image.
    out.write(b"".join(lzw_stream))

    # Finally, write the ending tag.
    out.write(ending)