#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The
# dimensions of the generated image can be specified 
# with `width` and `height`.
def generate_random_gif(
        rng: BinaryIO, 
        out: BinaryIO, width=1024, height=768):

    # Generate an ID block
    idblock = ''.join([chr(i) for i in range(33)]) 
    out.write(idblock.encode())

    # Write out screen descriptor
    w = width//256
    h = height//256
    
    sd = bytes('!', 'utf-8') \
         + bytes((w,h))           \
         + struct.pack('<BBB',     \
             (width%256),(height%256), 0x08)

    out.write(sd)
        
    # Select size of color table to use
    if ((width+height)<=(3*256)): 
        ncolor = 1<<int(((width+height)/3)).bit_length()
    else: 
        ncolor = 1<<(int(((width+height)/3)).bit_length()+1)

    nbits = max(1, int(((ncolor*3)/8)).bit_length())
    palsize = (nbits+(7>>nbits))&~(7>>nbits)

    # Write palette for true color images
    p = [b'\xFF'*palsize] * ncolor
    out.write(struct.pack('<B',1 << nbits)   )
    out.write(bytes().join(p))

    # Write graphics control extension
    gce = b''
    gce += bytes('!\xF9\x04','utf-8')    \
           + b'\x08\x00\x00'            \
           + struct.pack('<H',             \
               (((100<<8)+100)<<8)+(100)) \
           + b'\x00' 

    out.write(gce)

    # Write local color table
    lct = []
    for _ in range(ncolor):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        lct.append(bytes([r,g,b]))

    out.write(lct[0])

    # Setup LZW compression parameters
    GCTF = False      # Global Color Table Flag
    LCTF = True       # Local Color Table Flag 
    InterlaceFlag = '\x00'
    SortFlag = '\x00'
    ReservedField = '\x00\x00'

    # Write graphic control label
    gclabel = b''
    gclabel+= bytes('!\xFC','utf-8')        \
              + bytes((LCTF,(InterlaceFlag+SortFlag+ReservedField)))

    out.write(gclabel)

    # Write out subblocks containing compressed data        
    encode_data = bytes((width,height))         \
                 + b''.join(lct)                 \
                 + b''.join([chr(i).encode() for i in range(256)]) 
    codesize = bitlen(len(encode_data)+1)
    mincodesize=3
    dictsize = len(encode_data)+1               
    out.write(bytes((mincodesize,dictsize,codesize)))
    
    x, y, ix, iy = 0, 0, 0, 0

    while (y<height):   
        lzwbuf = encode_data[iy:ix]
        ix = iy+1
        if (ix>=len(encode_data)): 
            break 

        c = lzwbuf[-1:]
        s = lzwbuf
        
        while(True):
            prev = s[:-1]#[:-1]

            try:
                offset = dict.index(prev+c)
                break
                
            except ValueError: 
                pass
            
            if not s : 
                c = chr(0).encode()
                continue
        
            del dict[-1]
            dict.insert(0,s)
            s = prev+c
            c = chr(int.from_bytes(s[:1],byteorder='big')).encode()
            
        code = dict.index(s)+1
        
        dict.insert(0,s)
        dictsize += 1
       
        # Prepare next iteration
        if (ix==len(encode_data)-1 or code>=(1<<(codesize-1))):
            buf = bytes((code,))                          # If code is full length it may get truncated on split below
            buf+= b'\x00'                                 # Terminate by sending 0 as last byte when done
            buf+= c                                        # Append last char to stream

            if (ix!=len(encode_data)-1):
                buf+= (chr(offset+1)+b'\x00').split(b'\x00')[1] # Split at offset and remove first char

            out.write(buf)                                # Send buffer

            d = 2**codesize - 1                           # Reset dictionary index
            e = 0                                          # Reset extra bits count
            codesize += 1                                  # Increase code size

        elif (ix!=(len(encode_data)-1)):
            buf = bytes((code,))          # Code
            buf+= chr(e&0x07|d<<3).encode() # Extra Bits and Index
            dict[d]=s                     # Add LZW string to dictionary
            e += 1                        # Increment ExtraBits counter
            
            if (e == (8 >> codesize)):                   # Check if we need more bytes for code
                out.write(buf)                            # Send buffer
                d = 2**codesize - 1                       # Reset dictionary index
                e = 0                                      # Reset extra bits count

                buf = b'\x00'+c                            # Set terminator character and append previous char to stream
                
                if (ix!=len(encode_data)-1):
                    buf+= (chr(offset+1)+b'\x00').split(b'\x00')[1] # Split at offset and remove first char

                out.write(buf)                             # Flush buffer
                
        if (ix!=len(encode_data)-1):            
            numbytes = len(lzwbuf)-1
            if (numbytes>=0): 
                for j in reversed(range(0,numbytes)):
                    if (j==(numbytes-1)):
                        d=lzwbuf[j]+1
                        e=8-numbytes
                    else:
                        d=lzwbuf[j]+2**(e+1)
                        e=0
            if (ix==(len(encode_data)-1)):
                d=lzwbuf[numbytes]+1
                e=8-numbytes
                    
        yc = y//(height//8)
        xc = x%width
        yc = y%8
        h = (yc*(height // 8))+xc

        iy = (y % (height // 8)) * (width)  
        ix = iy + (y % (height // 8) + 1)

        x += 1
 
    # End of Image
    ei = bytes('\x00','utf-8')
    out.write(ei)
