#!/usr/bin/env python3

import struct
from typing import BinaryIO


def write_uint16le(f, n):
    f.write(n.to_bytes(2, 'little'))

def write_uint32be(f, n):
    f.write(n.to_bytes(4, 'big', signed=False))

# Generates a random GIF file into `out`. The size of the
# generated file is in bytes given by `filesize`, and its
# content is created using the random number generator
# `rng` (typically /dev/urandom). It's up to you whether
# you use all this information or not... :)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        filesize: int    # File size in bytes
    ):

    # Write Gif 89a header
    headerb = [71, 73, 70, 56, 57, 97]   # GIF89a
    for hbyte in headerb:
        write_uint8(out, hbyte)
    
    # Write logical screen descriptor
    width  = int.from_bytes(rng.read(2), 'big')
    height = int.from_bytes(rng.read(2), 'big')
    flags  = ord(rng.read(1))           # Always set to zero at least?
    bgcol  = rng.read(1)[0]             # Explains why we test "flags" above
    aspect = rng.read(1)[0] * ord('!')  # Either always set to zero or an exclamation mark??
    write_uint16le(out, width)
    write_uint16le(out, height)
    write_uint8(out, flags)
    write_uint8(out, bgcol)
    write_uint8(out, aspect)

    # Write global color table if needed
    if flags & 0x80!= 0:
        # TODO: Handle other values than 7
        global_color_table_size = 7 * len([c[0] for c in COLOR_TABLE])
        out.write(global_color_table_size.to_bytes(1, 'big'))
        for rgb in COLOR_TABLE:
            r, g, b = rgb
            write_rgb(out, r,g,b)

    # Write block 0
    lzwmincodesize = ord(rng.read(1))
    out.write((lzwmincodesize+1).to_bytes(1, 'big'))  # Add one because it starts from 1 instead of 0
    out.write((width*height % 256).to_bytes(1, 'big'))
    out.write(((width*height)/256).to_bytes(1, 'big'))

    remainder = filesize - out.tell()
    assert remainder > 0
    buf = []     # List of ints (the current code that has been written so far)
    buflen = 0   # Current length of buf[]
    numbits = -1 # Number of bits left of the current byte, 0..7
    while remainder > 0:

        # Get more input if buf[] is empty
        nextinput = ord(rng.read(1))
        if len(buf)==0:
            numbits += 8

        # Get a new bit when enough bits are accumulated
        curnextbit = nextinput >> (numbits-1)
        numbits -= 1
        curcodesize = lzwmincodesize+1
        if numbits < 0:
            pass
        elif numbits == 0 and curnextbit == 0:
            write_uint8(out, curcodesize)
            write_uint8(out, buf[-1])
            numbits = 8
            continue
        else:
            curcodesize <<= 1
            if numbits <= 0:
                numbits = 0
            else:
                curcodesize |= 1
            prevcodesize = curcodesize
        
        # Try appending the current bit to each value in the dictionary. If we find any match, append that value to our buffer, update the number of bits used for encoding and restart the search with the unconsumed input.
        valrange = 2**curcodesize
        for val in range(len(DICTIONARY)):
            entry = DICTIONARY[val]
            if entry['value'] >= valrange: break

            if curnextbit == get_bit(entry['value'], numbits):
                
                # Found a match! Write the previous code first
                aval = val
                vlen = len(entry['buffer'])

                # Flush the output buffer
                while buflen>0:
                    # Flushing 255 bytes should be pretty rare anyway...
                    flushsz = min(buflen, 255)

                    write_uint8(out, buf[-flushsz:])
                    buflen -= flushsz
                    
                    # Emit a clear code
                    if curcodesize<12:
                        write_uint8(out, 2**(curcodesize-1)+1)
                    # Then a copy code followed by aval
                    else:
                        write_uint8(out, 2**curcodesize-1)
                        write_uint16le(out, aval)
                        
                        curcodesize = prevcodesize
                    valrange = 2**curcodesize
                    for val in range(len(DICTIONARY)):
                        entry = DICTIONARY[val]
                        if entry['value']>=valrange: break

                # Append recorded info
                buf.append(aval)
                buflen += 1

                # Set next state based on what matched
                numbits = curcodesize
                break
            
        # If no match was found, add the last value back to the dictionary.
        else:
            DICTIONARY.append({
                'value': valrange, 
                'buffer': [valrange]})
            
    # Flush remaining stuff in the buffer    
    while buflen > 0:
        # Flushing 255 bytes should be pretty rare anyway...
        flushsz = min(buflen, 255)
    
        write_uint8(out, buf[-flushsz:])
        buflen -= flushsz
        if curcodesize<12:
            write_uint8(out, 2**(curcodesize-1)+1)
        else:
            write_uint8(out, 2**curcodesize-1)
            write_uint16le(out, aval)
            
            curcodesize = prevcodesize
        
    # Write block end
    write_uint8(out, 0)
    
            








