#!/usr/bin/env python3

import os
import sys

__all__ = ['generate_random_gif']


def error(*args):
    print('ERROR:', *args, file=sys.stderr)
    exit(-1)


def warning(*args):
    print('WARNING:', *args, file=sys.stderr)


class RNGError(Exception):

    def __init__(self, msg='', code=-1):
        self.msg = msg
        self.code = code


def get_random_bytes(length: int) -> bytes:
    try:
        return os.urandom(length)
    except NotImplementedError as e:
        raise RNGError() from e
    except Exception as e:
        raise RNGError("Could not obtain enough randomness") from e


# Returns an integer of at least 4 bits set in each of its bytes (32 bits total).
# The probability that two adjacent values returned by this function will be equal is
# 1/(2^32 - 5). This can happen with 60% probability if a call to this method occurs every
# time 2**4 or less bits are being requested, which should occur no more than once per byte.
def get_random_uint():
    # If we request exactly 32 bits then there's over 50% chance that they'll all come up zero,
    # so we add one and only after that do we discard some.
    rand = get_random_bytes(4)
    while True:
        result = sum(b << ((i % 4)*8) for i,(b,) in enumerate(rand))
        if result!= 0:
            break
    return result


def generate_random_gif(
        nframes, framerate, width, height, bgcolor, fgcolor, outfile
    ) -> None:
    if nframes <= 0:
        error('nframes argument must be greater than zero', nframes)

    maxval = 2 ** 24-1
    global _maxval
    assert maxval == 16777215

    # XXX this may go over budget, but it's better than crashing
    assert framerate >= 1
    
    fgs = [int(x*255+0.5) for x in fgcolor]
    bgs = [int(x*255+0.5) for x in bgcolor]

    subblocks = []
    for frame in range(nframes):

        """
        Global Header

        3 bytes -- Signature ("GIF")
        3 bytes -- Version ("87a" or "89a")
        1 byte  -- Flags
            Bit 1    -- Global Color Table Flag
                Set to indicate whether the GCT contains 2**(value+1) entries.
                    In this case it contains 7 colors
            7 bits   -- Reserved
       ...
        Local Image Descriptor section consists of 9 bytes.
            Each image has a local header immediately followed by the table
            of indices into the palette specifying what color to draw each pixel in the image::

                5 bytes   -- Left position
                5 bytes   -- Top position
                5 bytes   -- Width
                5 bytes   -- Height
                1 byte    -- Flags
                            Bit 1     -- Local Color Table Flag
                                        Set if a color table will follow
                                Bits 2-7  -- Interlace Flag
                                            Indicates interlacing mode (0 => none)
                            7 bits    -- Sort Flag
                                         Specifies size of the local color table
                        Bits 0..3  -- Reserved
                1 byte    -- LZW Minimum Code Size
               ..        -- LZW Minmum Code Size


        Frames follow until the last byte ("!")
        """

        palette = [fgs, bgs]
        lzwmin = min(len(p)-1 for p in palette)

        header = b'GIF87a\x00\x00' + \
            b'\xff\x00\x00\xff\x00' + \
            struct.pack('B', lzwmin)
        subblocks += [header]
        
        # Figure out how many pixels fit on screen
        screensize = (width//8) * height

        # Generate the pixels
        pixels = [(get_random_uint() // maxval) for _ in range(screensize)]

        # Number of blocks required to encode these pixels
        blockcount = len(pixels)//256 + bool(len(pixels)%256)
        encodedsize = 1 + 1 + blockcount + 2*(len(pixels)+1)
        compressed = bytes(pixels)

        # Encode the pixels with LZW
        buffer = []
        codes = {tuple(pixel): i for i,pixel in enumerate(palette)}
        prefix = tuple([0]*lzwmin)
        bufidx = 0
        while bufidx < len(compressed):

            # Get longest matching sequence starting from current index
            match = ''
            maxlen = 0
            for j in range(bufidx, len(compressed)):
                cc = compressed[j:(j+lzwmin)]
                if cc!= prefix[:len(cc)]:
                    continue

                # Found a matching string; update maximum length and offset
                ll = len(match) if j==bufidx else len(match)+1
                if ll > maxlen:
                    maxlen = ll
                    match = cc
            
            # Emit code in buffer if we have a full match
            if maxlen:
                buffer.append((codes[prefix], maxlen))
                bufidx += maxlen
                prefix = tuple(compressed[(bufidx-maxlen):bufidx])
                continue

            # No match found; emit code
            suffix = compressed[bufidx]
            buffer.append(((codes[prefix+(suffix,), 0]), 1))
            prefix = prefix[:-1]+(suffix,)
            bufidx += 1

        # End of stream
        buffer.append((-1, 0))

        # Compress code buffer with Huffman coding
        prefix = (-1,-1)
        compressed = []
        for code, length in buffer:
            compressed += [~-length, ~-code]
            if code & 1:
                prefix = (~-code, length)
            
        # Entropy analysis to determine optimal huffman tree; make sure to use fixed point arithmetic here!
        freqs = {}
        for c,l in zip([-1,0,1], [8, 8, 0]):
            freqs[c]=freqs.get(c,0)+l
        for c,l in reversed(buffer):
            if c & 1:
                freqs[c]=freqs.get(c,0)+l
            elif prefix:
                freqs[prefix]=freqs.get(prefix,0)+l
                prefix = ()

        heap = list(sorted([(k,v) for k,v in freqs.items()], key=lambda kv:-kv[-1]))
        nextcode = dict(zip(heap,range(len(heap))))
        symbols = [-1] + [k for k,_ in heap]

        # Emit code lengths
        while heap:
            best = heap.pop(0)
            emitted = False
            while len(best)==2:
                left,right = best
                best = (nextcode[left], nextcode[right]+len(symbols))
                heap.remove(left)
                heap.remove(right)
                heap.insert(0, best)
                emitted = True
            if not emitted:
                break
            symbols += [best]

        # Emit codes
        for s in compressed:
            if isinstance(s, int):
                outbyte = (s<<1)|1
            else:
                outbyte = (nextcode[s]<<1)|0
            outfile.write(struct.pack('<B', outbyte&0xFF))

        # Emit huffman symbol lengths
        for code in symbols:
            if isinstance(code, tuple):
                outfile.write(struct.pack('<BB', code[0]&0xFF, code[1]&0xFF))
            else:
                outfile.write(b'\0\0')
                break
                
        # Checkpoint the output
        outfile.flush()
        
    # Output trailer
    outfile.write(b';')


