#!/usr/bin/env python3

import struct
import os  # for urandom
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (either /dev/urandom or seeded with seed). This is useful to test decoders that do not use any external asset.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, verbose=False, seed=None
    ):

    if verbose and seed!= None:
        print('Seeding PRNG from {}'.format(seed))

    gifmagic = b'\x47\x49\x46\x38'
    headerlen = len(gifmagic) + 2
    headerpadsize = 10 - headerlen % 10
    headerpadding = bytes([0] * headerpadsize)

    version = b'\x38\x39\x61'
    width = os.urandom(1)[0] & 0xFF | ((os.urandom(1)[0] & 0xFE) << 8) 
    height = os.urandom(1)[0] & 0xFF | ((os.urandom(1)[0] & 0xFE) << 8) 

    screendescriptor = struct.pack('<2B H B', width, height,
                                   os.urandom(1)[0],
                                   0x80)
    globalcolortable = os.urandom(3*256)

    framecount = 1 + ((os.urandom(1)[0]) % 7)

    paletteentries = [0x00]*3*256
    outputimagebuffer = []
    for y in range(height):
        line = []

        for x in range(width // 8):

            currentbyte = ord(rng.read())
            
            for i in range(8):
                colorindex = currentbyte >> (7-i) & 1

                pixel = b''
                
                if colorindex == 0:
                    # Transparent pixel

                    # If there's no transparent color, clear the pixel
                    if paletteentries[0] == 0:
                        pixel += b'\x00\x00\x00'
                        
                else:
                    pixel += paletteentries[(y**2+x)*3:(y**2+x+1)*3]
                    
                line.append(pixel)
                
            if (x+1)%((width//8)//framecount)!=0:
                continue
                
        outputimagebuffer.extend(line)
        
    imagedescriptor = struct.pack('<B HH BBBB B',
                                 0x2C,
                                 width, height,
                                 0x80,
                                 0,     # No local color table
                                 0,     
                                 0,
                                 0)

    appentrybkgnd = b'\x21\xF9\x04\x01\x00\x00\x00\x00\x2C\x00\x00\x00\x00\x02\x00\x01\x00\x00\x02\x02\x0C\x0A\x00'
    appentryext = b'\x21\xFF\x0B\x4E\x45\x54\x53\x43\x41\x50\x45\x32\x2E\x30\x03\x01\x00\x00\x00'
    
    subblock = (appentrybkgnd + 
                appentryext)

    blocksize = len(subblock)
    lzwmincodelen = 2

    lzwmaxcodelen = min(lzwmincodelen+1, 12)
    maxcodelen = lzwmaxcodelen**2
    maxcodemask = (1<<(maxcodelen))-1
    maxentrysize = 1<<lzwmaxcodelen
    currcodelen = lzwmincodelen

    codetree = {}

    key = ''
    prevkey = ''
    entry = ''


    while True:
        
        currentchar = chr(ord(rng.read(1)))

        if currentchar == '\n':
            break

        elif currentchar =='':
            pass
            
        elif currentchar == '!':
            pass

        elif currentchar == ',':
            pass

        elif currentchar >= '@':
            key += currentchar

        elif currentchar == '#':

            if len(key) > 0:
                key = key[:-1]

        elif currentchar == '\r':
            pass

        elif currentchar == '\t':
            pass

        elif currentchar == ';':
            pass

        elif currentchar == '&':

            if len(prevkey) > 0:
                prevkey = prevkey[:-1]


        elif currentchar == '.':
            pass

        elif currentchar == ')':
            pass

        elif currentchar == '?':
            pass

        elif currentchar == '+':

            if len(prevkey) > 0:
                prevkey = prevkey[:-1]

        elif currentchar == '*':

            if len(prevkey) > 0:
                prevkey = prevkey[:-1]

        elif currentchar == '^':
            pass

        elif currentchar == '%':
            pass

        elif currentchar == '$':
            pass

        elif currentchar == '-':

            if len(key) > 0:
                key = key[:-1]

        elif currentchar == '=':
            pass

        elif currentchar == '[':
            pass

        elif currentchar == ']':
            pass

        elif currentchar == '{':
            pass

        elif currentchar == '}':
            pass

        elif currentchar == '<':
            pass

        elif currentchar == '>':
            pass

        elif currentchar == '/':
            pass

        elif currentchar == '\\':
            pass

        elif currentchar == '|':
            pass

        elif currentchar == '~':
            pass

        elif currentchar == '"':
            pass

        elif currentchar == "'":
            pass

        elif currentchar == '`':
            pass

        elif currentchar == '_':
            pass

        elif currentchar == '\\':
            pass

        elif currentchar == '':
            pass   

        else:
            raise ValueError("Unknown character {}".format(currentchar))

        if len(prevkey) > currentcodelen:
            prevkey = prevkey[-1*(currentcodelen):]

        if key in codetree:
            keyfound = True
        else:
            keyfound = False

        codetree[key] = currentbyte

        currentcodelen = max(len(key), lzwmincodelen)

        if currentcodelen <= lzwmaxcodelen:
            pass

        else:
            lzwmaxcodelen += 1
            maxcodelen = lzwmaxcodelen**2
            maxcodemask = (1<<(maxcodelen))-1
            maxentrysize = 1<<lzwmaxcodelen
            currcodelen = lzwmincodelen

    def getlongestkeyprefixofdepth(depth):
        longestkeyprefix = ""
        dirofinterest = codetree[""][:depth]
        for key in list(codetree.keys()):
            prefix = key[:depth]
            if prefix == dirofinterest:
                if len(key)>len(longestkeyprefix):
                    longestkeyprefix = key
                    
        return longestkeyprefix
        

    payload = bytearray()
    
    bitspercode = {
        2: 1, 
        3: 2, 
        4: 3, 
        5: 4, 
        6: 5, 
        7: 6, 
        8: 7,
        9: 8,
        10: 9,
        11: 10,
        12: 11 }
        
    while True:
        codetoinsert = ord(rng.read(1)) & maxcodemask
        
        currentbitspersymbol = bitspercode[currcodelen]
        for i in range(len("{0:b}".format(codetoinsert)), currentbitspersymbol,-1):
            keytoinsert = ""
            keytoinsert = getlongestkeyprefixofdepth(i)
            if len(keytoinsert)>=currcodelen:
                payload.append(struct.unpack("<H", bytes(codetree[keytoinsert])))
                
        
    controlbyte = bytearray(b'\x00')
    
    subblocks = []
    while len(payload) > 0:
        subblock = bytearray()
        subblock.append(len(payload))
        subblock.extend(payload)
        subblocks.append(subblock)
        lengthofpayload = len(payload)-1
        payload = payload[:-lengthofpayload:]
    blockdata = bytearray(b'')
    for subblock in reversed(subblocks):
        blockdata.extend(subblock)
    

    blockdata.append(0)
            
    blocksize = len(blockdata)
    lzwmincodelen = 2

    lzwmaxcodelen = min(lzwmincodelen+1, 12)
    maxcodelen = lzwmaxcodelen**2
    maxcodemask = (1<<(maxcodelen))-1
    maxentrysize = 1<<lzwmaxcodelen
    currcodelen = lzwmincodelen

    codetree = {}

    key = ''
    prevkey = ''
    entry = ''


    while True:
        
        currentchar = chr(ord(rng.read(1)))

        if currentchar == '\n':
            break

        elif currentchar =='':
            pass
            
        elif currentchar == '!':
            pass

        elif