#!/usr/bin/env python3

import io
import os
import sys
import struct
from enum import Enum
from collections import namedtuple, deque
from typing import List, Optional, Callable


class RNG:
    def read(self, size):
        raise NotImplementedError()

    def randrange(self, n):
        bytesize = (n.bit_length() // 8) + 1 # Round up to nearest byte
        result = self.read(bytesize)[0] & ((2 ** bits - 1))

        if not isinstance(n, int):
            return result % n
        else:
            return result

    def choice(self, xs):
        return xs[self.randrange(len(xs))]

    def weighted_choice(self, choices):
        total = sum(w for c, w in choices)
        r = self.randrange(total)
        upto = 0
        for c, w in choices:
            if upto + w > r:
                return c
            upto += w
        assert False, "Shouldn't get here"


class DistinctRangeRNG(RNG):
    def __init__(self, ranges: List[str]):
        self._ranges = []
        self._rangesmap = {}
        offset = 0
        for s in ranges:
            try:
                lo, hi = map(int, s.split('-'))
                length = hi + 1 - lo
                offset = max(offset, hi)

                self._ranges.append((offset, length))
                self._rangesmap[(offset, length)] = True

            except ValueError:
                print("Invalid range: {}".format(s))


    def read(self, size=None):
        if size is None or size < 1:
            size = 1

        buf = bytearray()
        while len(buf) < size:
            offset, length = self.choice(list(self._rangesmap.keys()))
            buf.extend([offset])
            for i in range(1, length):
                self._rangesmap[(offset+i, length-i)] = True

        return buf[:size]


class SentinelValue(Enum):
    EOL = 0


class ByteReaderRNG(RNG):
    def __init__(self, f):
        self._f = f

    def _no_eof(self, *args, **kwargs):
        v = self._f(*args, **kwargs)
        if v == '': 
            raise EOFError()
        
        return ord(v)
    
    def tell(self):
        return self._f.tell()

    def seek(self, off, whence=io.SEEK_SET):
        return self._f.seek(off, whence)

    def skip(self, n):
        current = self.tell()
        self.seek(current + n)

    def readline(self):
        line = io.StringIO()
        while True:
            char = chr(self.read())
            line.write(char)
            
            if char == '\n': break

        return line.getvalue().encode('utf-8')

    def read(self, size=None):
        if size is None or size < 1:
            size = 1
            
        buf = bytearray()
        while len(buf) < size:
            c = self._no_eof(1)
            if c == '\n'.encode():
                break
                
            buf.extend([c])

        return buf[:size]

    def peek(self, atmost=1):
        original = self.tell()
        value = self.read(atmost)
        self.seek(original)
        return value

    def consume(self, value, case_sensitive=True):        
        if not value.startswith(b'\n'):
            first = value[0]
            target = self.peek()
            if not case_sensitive and type(target) == str: 
                target = target.lower()
            elif case_sensitive and type(target)!= str:
                target = chr(target).lower()
            if target!= first:
                raise AssertionError()

        consumed = io.BytesIO()
        while True:
            target = self.peek()
            if len(value) <= 0:
                break

            if not case_sensitive and type(target) == str: 
                target = target.lower()
            elif case_sensitive and type(target)!= str:
                target = chr(target).lower()

            consumed.write(self.read(1))

            expected = value[0].decode()
            if target!= expected:
                print(consumed.getvalue(), expected, target)
                raise AssertionError("{}!= {}".format(expected, target))

            value = value[1:]
        

    def find_next(self, needle):
        pos = 0
        basepos = self.tell()
        while True:            
            if self.read(len(needle)) == needle:
                self.seek(basepos)
                return pos

            pos += 1

class Config(namedtuple('_Config', [
       'version',
        'width', 
        'height', 
        'background', 
        'gct_size', 
        'bgindex', 
        'pixelaspectratio', 
    ])):
    @staticmethod
    def from_file(f: ByteReaderRNG) -> '_Config':
        version = f.readline()
        f.consume(b'Logical Screen Descriptor')

        width = f.read(4)
        height = f.read(4)
        background = f.read(1)
        gct_size = f.read(1)
        bgindex = f.read(1)
        pixelaspectratio = f.read(1)


        return Config(**{
           'version': version,
            'width': struct.unpack(">H", width)[0],
            'height': struct.unpack(">H", height)[0],
            'background': background[0],
            'gct_size': gct_size[0],
            'bgindex': bgindex[0],
            'pixelaspectratio': pixelaspectratio[0],
        })



class Frame(object):
    def __init__(self, config: Config, data: list):
        self.config = config
        self.data = data
        
    def write(self, f: BinaryIO):
        f.write(self.config.version)

        f.write(struct.pack(">H", self.config.width))
        f.write(struct.pack(">H", self.config.height))

        f.write(struct.pack("B", self.config.background))
        f.write(struct.pack("B", self.config.gct_size))
        f.write(struct.pack("B", self.config.bgindex))
        f.write(struct.pack("B", self.config.pixelaspectratio))

        table_size = 3 * pow(2, self.config.gct_size + 1)
        table = [0] * table_size
        lastcolour = [None, sentinels.EOL]
        for index, colour in enumerate(self.data):
            f.write(struct.pack("B", colour))
        
            if colour >= 0xF7: continue

            if colour in lastcolour:
                colour_index = table[lastcolour.index(colour)]
            else:
                colour_index = len(table)
                table.extend([sentinels.EOL] * (colour_index - len(table)))
                
                table[lastcolour[-1]] = colour_index
                lastcolour.append(colour)
                
        if len(table) < table_size:
            table.extend([sentinels.EOL] * (table_size - len(table)))

        for x in filter(lambda t: type(t)!= SentinelValue, table):
            f.write(struct.pack("BBB", x >> 16, (x >> 8) & 0xFF, x & 0xFF))


class ColorTable(deque):
    def __init__(self, elements=()):
        super().__init__(elements)
        
        for element in elements:
            if element >= 0xF7: continue
            self.appendleft(element)

    def extend(self, other):
        for o in other:
            self.append(o)
            
    def append(self, element):
        if element >= 0xF7: 
           super().append(element)
           return
        
        if element in set(self):
            return
        
        super().append(element)

    def popleft(self):
        ret = super().popleft()
        if ret == sentinels.EOL:
            raise StopIteration()
        
        return ret

    def empty(self):
        return all(map(lambda e: e == sentinels.EOL, iter(self)))
        
class GifGeneratorRNG(RNG):
    def __init__(self, f: ByteReaderRNG):
        self._f = f

    def _genframe(self):
        frame_config = Config.from_file(self._f)
        print("[*] Frame config:", frame_config)

        colorids = ColorTable([])
        lzw_buffer = []
        block_type = None
    
        while True:
            block_type = self._f.read(1)[0]
            print("Block type: ", hex(block_type))
    
            if block_type == 0x2C:
                # Image descriptor
                img_left = struct.unpack("<h", self._f.read(2))[0]
                img_top = struct.unpack("<h", self._f.read(2))[0]
                img_width = struct.unpack("<h