#!/usr/bin/env python3

import math
import os
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`. This function is deterministic and should be
# called for each generation to ensure that there are no duplicate images in the
# output stream.
def generate_random_gif(
        rng: BinaryIO,
        num_entries: int,
        max_data_per_block: float,
        out: BinaryIO
    ) -> None:

    # First part of GIF uses 7 bytes which we can't control easily since it's
    # fixed at 0x1ce2e514 (47416436). We can change the version though so our
    # first entry will be 0x1cf2e514 instead.
    header = b'\x47\x49\x46\x38\x39\x61\x14\xcf'
    out.write(header)


    def pack_entry():

        # Each frame has an initial byte that designates whether or not this is
        # a local color table. Since we're just blindly copying bytes from /dev/
        # urandom, it isn't guaranteed that each block starts with a local color
        # table but here we go anyway because it doesn't really matter to us. The
        # second bit denotes if a transparent color index exists. But again, this
        # doesn't *really* matter either. All we need is some more data!
        packed_entry = b'\x2c\x00'

        # Get the random length of data per frame. Then get bytes necessary to
        # represent said length as 1-byte value.
        num_data_bytes = math.ceil(
                rng.randint(
                    *[
                        0,
                        int((num_entries - 1)*max_data_per_block)]) /
                256)
        data_len = rng.randint(0, min([
            2**(num_data_bytes * 8),
            int((num_entries - 1)*max_data_per_block)]))
        data_len_bytes = data_len.to_bytes(
                num_data_bytes, 'big', signed=False)

        # Write the actual data to the output stream now.
        out.write(data_len_bytes)
        out.write(rng.read(int(data_len)))

        packed_entry += data_len_bytes


        # Use LZW compression to reduce overall size drastically. We don't care
        # about compression ratios much right now, so lets just use Python's built
        # in implementation of a LZW compressor. By default, this algorithm uses
         # a table size of 4096.
        # TODO: Figure out how to configure the algorithm properly. I'm assuming
        #       that you can specify the code size, but don't know how exactly.
        lzw_compresser = \
            zlib.compressobj(wbits=-zlib.MAX_WBITS)
        compressed_data = lzw_compresser.compress(packed_entry[1:])
        compressed_data += lzw_compresser.flush()

        return len(compressed_data)



    # Loop through the entries and write them to the output stream.
    total_length = sum([
        pack_entry()
        for _ in range(num_entries)
    ])

    # Add the rest of the GIF header at the end of writing all frames.
    # TODO: Should add other portions of the GIF header as well?
    palette_size = 1 << bits_per_pixel
    colour_table_offset = 12
    image_data_offset = colour_table_offset + palette_size * 3
    global_colour_table_flag = 0b10000000
    local_colour_table_flag = 0b00000000
    reserved = 0b00000000
    background_color_index = 0
    aspect_ratio = 0

    # Formatting above was gathered from https://www.matthewflickinger.com/lab/whatsinagif/bits_and_bytes.asp
    gif_trailer = struct.pack(
        '<BBBBBHHHBBxxx',
        palette_size,
        colour_table_offset,
        global_colour_table_flag,
        background_color_index,
        aspect_ratio,
        total_length,
        image_data_offset,
        reserved
    )
    out.write(gif_trailer)

