#!/usr/bin/env python3

import os
import sys
import zlib
from math import floor, ceil
from pathlib import Path
from typing import BinaryIO, List

# Generate a random 256 color palette to be used in
# the `pallette` field of the global colour table
def gen_palette() -> bytes:
    colors = [
        0x000000, # Black
        0x00FF00, # Green
        0xFF0000, # Red
        0x00FFFF, # Cyan
        0xFFFF00, # Yellow
        0xFF00FF, # Magenta
        0xFFFFFF, # White
    ]

    return (b''.join([struct.pack('<BBB', *divmod(c, 256)) for c in colors]))


# Generate an LZW decoder table that is capable of decoding
# all possible lzw codes with a minimum code size of 4 bits
def gen_lzw_decoder_table() -> bytes:
    dec = []
    # The first two entries are not needed as they indicate
    # no encoding at all and small codes
    dec += [(i, i) for i in range(1, 5)]
    # The next 8 entries encode one byte into themselves
    dec += [(i+1, i) for i in range(3)]
    dec += [(i<<8|j, j|(1<<(7-i))) for i in range(4) for j in range(2**i)]
    # All other values are encoded by adding their highest order bit
    # to the value preceding it until no more bits can be added without
    # overflowing or reaching zero, then appending the resulting
    # integer to the list.
    # This works because the integers generated represent codes less than
    # their prefix number minus one so there will never be any overflows.
    def gen_bits():
        while True:
            for n in range(1, 5):
                yield from reversed([(n << k, k) for k in range(min(n, 8))])

            for m in range(4, -1, -1):
                for n in range(m, 5):
                    for k in range(min((256 >> m)+1, 2**(n-(m)))):
                        yield ((n+(k>>n)), k)


    for val, bits in gen_bits():
        if len(dec) >= pow(2, bits):
            break

        dec += [(val, val)]

    assert len(dec) == pow(2, max(len(t[1].to_bytes(1,'little'))*8 for t in dec))
    real_size = min(len(dec), 4095)
    dec = sorted(dec[:real_size], key=lambda x: x[1])

    return (b''.join(([struct.pack('BB', h, l) for _, v in dec] +
                     [[0, 0]]*(4096-len(dec)))))

# Decodes a string compressed using the given decoder table
def decode_lzw(
        s: str,
        decoder_table: bytes
    ) -> bytes:
    d = dict(zip(*map(lambda e: e, zip(*(iter(range(256)), iter(decoder_table))))))
    dec = [d[int.from_bytes(s[0:i+1], 'little')]
           for i in range(len(s))]
    return (b''.join(tuple(e.to_bytes(1, 'little') for e in dec[:-1])))


def gen_image_data(
        rng: BinaryIO, depth: int
    )-> bytes:
    data = [0]*ceil(pow(2, depth)*4/3)
    for i in range(1, len(data)):
        r = rng.read(1)
        if r == b'\xff':
            continue


        if r in (b'\x00', b'\x01'):
            s = {
                1: 2,
                2: 4,
                3: 8
            }[r]
            l = rng.read(1)[0]
            for _ in range(l+1):
                data[i] <<= s
                data[i] |= rng.read(1)[0] & ~(-1<<s)
                i += 1
        elif r[0] & 0xe0 == 0xc0:
            for _ in range(r[0]&~0xe0+3):
                data[i] <<= 8
                data[i] |= rng.read(1)[0]
                i += 1
        else:
            pos = (sum(v<<(8*(i-1)) for i, v in enumerate(reversed(r)))
                   ^ sum(v<<(8*(i-1)) for i, v in enumerate(reversed(data[-(r&0x1f):]))))
            for _ in range(((r&0xf)+1) % 0x20):
                data[i] = data[(pos+i)%len(data)]
                i += 1


    return (zlib.compress(b''.join(struct.unpack('BBBB', *data)),
                         9))



# Generate a single frame image block from the given parameters with /dev/urandom
def gen_frame_block(
        width: int, height: int,
        local_color_depth: int, global_color_depth: int,
        interlace: bool, sort_flag: bool,
        transparent_index: int, delay: int,
        frames: int, loops: int,
        dispose: int, blend: int,
        decoder_table: bytes,
        rgb: set=(
            0x000000, # black
            0x00FF00, # green
            0xFF0000, # red
            0x00FFFF, # cyan
            0xFFFF00, # yellow
            0xFF00FF, # magenta
            0xFFFFFF, # white
        ),
        rng: BinaryIO=os.fdopen(sys.stdin.fileno(), "rb")
    )-> bytes:
    params = [width, height, local_color_depth,
              global_color_depth, interlace, sort_flag,
              0, 1, 0, 0, 0, 0, 0, transparent_index,
              delay, frames, loops, dispose, blend]

    header = (~sum(params)&0xffff).to_bytes(2,"big")
    params = tuple(param.to_bytes(2,'big') for param in params)

    gct_size = (global_color_depth==1)*7 + (global_color_depth==2)*4 + (global_color_depth>2)*pow(2, global_color_depth)
    gct = gen_palette()
    img_data = gen_image_data(rng, local_color_depth)

    lzw_img = decode_lzw(img_data, decoder_table)

    return (header + b''.join(params)+gct_size.to_bytes(1,"big")+gct+b''+len(lzw_img).to_bytes(2,"big")+lzw_img)




# Generates a GIF89a body containing `frames` frames with the specified parameters,
# plus the appropriate number of empty frames according to the `loops` parameter.
# The frame data comes from `rng`, which defaults to sys.stdin
def gen_gif_body(
        width: int, height: int,
        local_color_depth: int, global_color_depth: int,
        interlace: bool, sort_flag: bool,
        transparent_index: int, delay: int,
        frames: int, loops: int,
        dispose: int, blend: int,
        rgb: set=(
            0x000000, # black
            0x00FF00, # green
            0xFF0000, # red
            0x00FFFF, # cyan
            0xFFFF00, # yellow
            0xFF00FF, # magenta
            0xFFFFFF, # white
        ),
        rng: BinaryIO=os.fdopen(sys.stdin.fileno(),"rb")
    ) -> bytes:
    encoder_table = gen_lzw_decoder_table()
    if global_color_depth > 8:
        raise Exception("Color depth must be between 1 and 8, inclusive.")

    fblen = 7 + global_color_depth*3
    body = [gen_frame_block(width,
                           height,
                           local_color_depth,
                           global_color_depth,
                           interlace,
                           sort_flag,
                           transparent_index,
                           delay,
                           frames,
                           0,
                           dispose,
                           blend,
                           encoder_table,
                           rgb,
                           rng) for _ in range(frames)]
    body *= loops
    bgcolor = None if bgcol < 0 else bytes([bgcol])
    crlf     = b"\x0D\x0A"
    hdr      = b"GIF89a"+width.to_bytes(2,"little")+height.to_bytes(2,"little")+b"!\xF7"
    lsd      = bytes([local_color_depth])+b"\x00