#!/usr/bin/env python3

import io
import struct
from typing import NoReturn


def _get_random_data(rng: BinaryIO, size: int) -> bytes:
    return rng.read(size)


def _generate_header(bkgd_color_idx: int) -> bytes:
    """
    Generate the GIF header according to the given background color index

    The first 6 bytes of an animated GIF are always 'GIF87a', but can be
    changed to 'GIF89a'. These two values identify an image as being either
    87a or 89a compatible (i.e., they indicate which version of the spec was
    used for encoding this file). If you want your animation to look the same
    on all systems, use 89a. This is also what most web browsers will display
    your image in. Note that GIF87a has some limitations—for example, it only
    supports a maximum of 256 colors per frame. Some older viewers may not
    support them at all.
    """
    # Version
    ver = b'\x47\x49\x46\x38\x39\x61'

    # Logical Screen Descriptor
    lsd = b''
    lsd += struct.pack(">HH", width=0, height=0)       # (width,height)    - Whole screen area
    lsd += b'\xF0'                                     # Packed bits      - Global Color Table Flag and Size
                                                #                  - Global Color Table exists;
                                                #                      Color Resolution = 7
                                                #                      Sort Flag = False
                                                #                      Size of Global Color Table = 2^(n+1); n = 0..7
    lsd += b'\xFF'                                     # Background Color Index

    # Application Extension Block (extension for mouse pointer)
    # NOTE: Ignoring any GCE due to being unneeded here
    appext = b''
    appext += b'\x21\xFF\x0B'                         # Extension Introducer   - Ext label = Graphic Control Label
    appext += b'\x4D\x45\x54\x55\x53\x42\x4F\x4F\x4B'     # Application Identifier
    appext += b'\x03'                                 # Sub-block length        - 3 bytes long
    appext += b'\x01'                                 # Byte 1:             | 1111 1100 : 111 color resolution bits
    appext += b'\x00'                                 # Byte 2:             | 0000 0000 : sort flag = False
    appext += b'\x00'                                 # Byte 3:             | 0000 0000 : color table size = 7
    appext += b'\x3B'                                 # Byte 4:             |
    extapp = len(appext)                                # App extension length

    # Graphics Control Extension block
    gce = b''
    gce += b'\x21\xF9\x04'                           # Extension introducer
    gce += b'\x08'                                   # Flags                   - Reserved (must be zero)
    gce += b'\x00'                                   # Reserved                - also must be zero
    gce += struct.pack(">H", delay=int(10))           # Delay Time              - 1/100 s
    gce += b'\x00'                                   # Transparent Color Index
    gce += b'\x00'                                   # Terminator


    gcedata = len(gce)                                  # Length of graphics control extension data

    # Image Descriptor
    imgdesctable = b''
    imgdesctable += b'\x2C'                             # Image Separator
    imgdesctable += struct.pack(">HHHHH",           # Image Left Column
        left=0, top=0, width=800, height=600, packed=0)

    # Local Color Table
    clrtab = b''
    clrtab += b'\xFF\x00\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'
    clrtablen = len(clrtab)                            # Local color table length

    # Start outputing the final result
    res = b''
    res += ver                                         # Version string          - "GIF89a"
    res += lsd                                         # Logical Screen Descriptor
    res += appext                                      # Application Extension Block
    res += gce                                         # Graphics Control Extension
    res += imgdesctable                                # Image Descriptor
    res += clrtab                                      # Local Color Table
    return res


def _generate_pixel_art(rng: BinaryIO):
    dat = []                                          # List of pixel rows
    for i in range(height):
        rowstr = ''                                    # String representing the current row
        for j in range(width // 8 * 8):                 # For every column in the row
            pix = _get_random_data(rng, 1)[0]         # Get one byte from /dev/urandom
            for k in reversed(range(8)):               # For every bit in the column
                if ((pix >> k) & 1):                    # Check whether bit at position K is set
                    rowstr += '*'                       # If so, add a star to represent black
                else:
                    rowstr +=''                       # Otherwise, add spaces to represent white
            rowstr += '\t'                             # Add tabulation

        # Remove trailing tabs from completed row, then append it to the list
        dat.append(rowstr[:-1])                        # Remove trailing tabs
    return dat                                         # Return generated pixel rows


if __name__ == "__main__":
    width = 800                                        # Width of the image
    height = 600                                       # Height of the image

    with open('/tmp/testimg.gif', 'wb+') as outf:
        rng = io.open('/dev/urandom', mode='rb')        # Open access to /dev/urandom
        outf.write(_generate_header(0xFFFF))            # Write GIF header contents
        dat = _generate_pixel_art(rng)                  # Generate pixel art from random data
        for line in dat:                               # For each line in the pixel art...
            print(line + '\n')                          # Print it out
            outf.write((line + '\n').encode('utf8'))   # And write it to the GIF