#!/usr/bin/env python3

import itertools
import struct
from typing import BinaryIO

"""
Generates a random GIF file into `out` by generating and stacking
the 24-bit colors in the order from `rng` (an open stream to /dev/urandom).

The first frame contains only one color with alpha set to zero. Each subsequent
frame will contain two more colors that make up the full palette of 256 colors.
The last frame has all colors filled in so it can be used as a loop count.
"""
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Set up state variables for generating frames.
    color_count = 1 << 8       # Number of unique RGB triples we have available
    num_colors  = 0            # Colors in the current frame
    next_color  = iter(range(color_count))   # The index of the next color to use
                                            # in this frame
    
    # Sets up a list of boundaries for each new frame's color palette
    # Also adds an extra boundary at the end which is just the size of
    # the entire 256-element palette
    boundaries = [num_colors]
    while True:
        try:
            num_colors += sum(itertools.islice(next(next_color), len(boundaries)))
        except StopIteration:
            break

        boundaries.append(num_colors)


    # Write the header, then loop over each frame
    width    = 4 * max([x - y for x,y in zip(boundaries[1:], range(len(boundaries)-1))])
    height   = 4 * max([x - y for x,y in zip(boundaries[::], range(width+1))])
    pixellen = 4 * width * height // 2
    out.write(struct.pack('<6sHHBBB',
                         b'GIF87a',          # Header
                         0x100,             # Logical screen width
                         height,            # Logical screen height
                         0xF7,              # Packed field bits
                         7,                 # Background color index
                         0                    # Pixel aspect ratio
                         ))
    for i, bound in enumerate(boundaries[:-1]):
        out.write(struct.pack('B',                   # Frame control
                              (i & 0xf7) | (not i >> 3 & 1),
                              0x00,               # Delay time
                              0x00,               # Image left position
                              0x00                # Image top position
                              ))
        out.write(b'\x00\x00')                       # Width and height

        if i == 0:                                    # First frame has black background
            out.write(struct.pack('<BBBxB',           # Black color
                                  0,
                                  0,
                                  0,
                                  0                         
                                  ))
        else:                                         # Other frames are random colors
            for j in range(bound, boundaries[i + 1]):
                entry = struct.unpack("<BBB", rng.read(3))[::-1]
                out.write(entry)                     # RGB values

                # Add an opaque alpha value to the last three bytes
                if not j % (pixellen * 2):
                    out.write(b'\xff')

        out.write((height - 1) & 0xff)               # LZW minimum code size
        out.write(b"\x00")                            # No local colors
        lzwcodesize = 8                                # Code size starts here

        # Compute codes and write them to frame
        def encode_byte(c: int):
            """Encodes a single byte into its LZW representation."""

            nonlocal lzwcodesize

            if c < 1 << lzwcodesize:                  # If bit length is less than
                return (c, 8)                         # the previous, prepend a 0

            elif c < 1 << lzwcodesize + 1:            # Otherwise, prepend a 1 and
                if lzwcodesize < 12:                 # increment the bitlength
                    lzwcodesize += 1
                return ((1 << lzwcodesize) + (c - (1 << lzwcodesize)), lzwcodesize + 1)
            
            raise ValueError("Code too large!")     # We should never hit this

        def get_lzwcode(c: int):
            """Given a char, returns its corresponding LZW code pair."""
        
            if c in lzwdict:
                return lzwdict[c]

            lzwdict[c] = encode_byte(c)
            return lzwdict[c]

        lzwdict = {0:(0,1)}                             # Initialize dictionary
        for j in range(bound, boundaries[i + 1]):      # Encode pixels sequentially 
            encoded = tuple(get_lzwcode(pixel & 0xFF)[0]
                            for pixel in out.getvalue()[j*3:(j+1)*3][::-1])

            out.write(encoded[0].to_bytes(1,'little'))
            if encoded[1]:                              # Only write second part if needed
                out.write((encoded[0]>>8)+128)

