#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write header to file
    out.write('GIF89a'.encode())
    # Generate colors and color table
    color_table = []
    for _ in range(0x100):
        color = [int(i * 256) & 0xff for i in next(rng).to_bytes(3,'big')]
        color_table += bytes([color[0], color[1]])
        if sum(color[:2]) > 384:
            break
    # Add padding until we reach 768 bytes
    while len(color_table) < (768 - 256*3):
        color_table += bytes([0,0,0])
    # Write color table length
    out.write((len(color_table)).to_bytes(1, 'little'))
    # Write color table
    out.write(color_table)

    # Get image dimensions as pixels
    width = int(next(rng)) % 10 + 1
    height = int(next(rng)) % 10 + 1

    # Define LZW minimum code size
    lzwmincodesize = min(max(width//3+1,height//3+1)+1,12)

    # Write standard screen descriptor
    out.write((width).to_bytes(2,'little')
                  + (height).to_bytes(2,'little')
                  + ((lzwmincodesize-1)<<4
                                              | 0b0000
                                                ).to_bytes(1,'little')
                )

    # Write global color table flag specifying one color table
    # and no transparency information
    out.write(b'\x87')

    # Write background color index
    out.write(b'\x00')

    # Write pixel aspect ratio
    out.write(b'\x00\x00')

    # Write graphics control block with delay=0 and transparent color=0x00
    out.write((0).to_bytes(2,'little')
                    + (0).to_bytes(1,'little')
                      + (0).to_bytes(1,'little')
                        + (0).to_bytes(1,'little')
                          + (0).to_bytes(1,'little')
                            + (0x00).to_bytes(1,'little')
                              + (0x00).to_bytes(1,'little')
                                + (0).to_bytes(1,'little')
                                  + (0).to_bytes(1,'little'))

    print("Wrote graphic control block")


    # Write each row of the image's color map indices
    for y in range(height):
        for x in range(width):
            out.write(((int(next(rng))*2)%255).to_bytes(1,'little'))

    # Write end of image block
    out.write(b';')

    print("Wrote color map indices")



    # Generate LZW dictionary and write it to file
    dictsize = 2**(lzwmincodesize+1)
    dict = {b'\x0':n-1 for n,_ in enumerate(color_table)}

    prev = b''
    written = False
    while True:

        # Read two bytes from input stream
        curr = next(rng).to_bytes(1, 'big')
        nexts = next(rng).to_bytes(1, 'big')

        # If both are zero, we're done!
        if not curr \
           or not nexts \
           :
            return

        # See if we can find an entry in our dictionary that starts with these bytes
        key = curr+nexts
        if key in dict:

            # We found a match, so let's add this pair to the dictionary now
            dict[key+curr] = len(dict)

            # Next time through the loop, continue reading pairs of bytes wherever possible
            curr = key

        elif prev+curr in dict:

            # Try the current pair again but backwards
            nexts = prev+curr
            prev, curr = curr, nexts

            # We didn't get anything new, but we did overwrite something, so we'll
            # need to write what we just wrote back to the output stream.
            written = True



        # Now either curr is a character we know how to handle, or it represents "end of string"
        if curr == b'\x0':

            # Flush whatever was left over in our buffer 
            if len(prev)!=  0:
                print("\tWriting", prev)

                # This isn't actually part of the spec since it implied we'd be able
                # to compress runs longer than 255 characters, but after some experimentation 
                # I've determined this doesn't work in practice whatsoever. Even though the
                # rest of the code works fine with arbitrary lengths, the first value in the run
                # must be in the dictionary.
                if len(prev) > 255:

                    # So we'll try to write the maximum chunk that will fit -- this might
                    # result in us writing more than we should, but at least it won't 
                    # fail because we don't have enough space in the dictionary.
                    #
                    # To do this, we'll iterate every possible position in the previous sequence
                    # and take the maximal number of characters that can fit before overflowing.
                    best_run_length = 0
                    for pos in range(1, len(prev)):

                        # The maximum number of characters we could possibly place would be given by the difference between 
                        # the remaining capacity of the buffer, minus the amount required to represent the jump to the start of 
                        # the buffer itself, which has already been taken care of.
                        buffer_rem = 255 - len(prev)
                        numchars = min(pos, buffer_rem)

                        # How far in the dictionary does this sequence start? 
                        dictstart = dict[prev[:numchars]]

                        # Since we can hardly fit any more characters here, go ahead and write it without compression
                        out.write(struct.pack('<BB',
                                             dictstart+1,
                                         numchars ))

                        # Delete all the characters that were successfully encoded on the way down
                        del dict[prev[:numchars]]

                        # Update our best guess about what the longest run we can encode in this block was
                        best_run_length = max(best_run_length, len(prev)-numchars)



                    # If there's still stuff left in the buffer that can't fit in the buffer, flush it now
                    if len(prev) >= best_run_length:
                        remainder = prev[best_run_length:]
                        out.write(struct.pack('<BB', dict['\x0']+1, len(remainder)))
                        out.write(remainder)

                        # And delete it from the dictionary
                        del dict[remainder]

                else:

                    out.write(struct.pack('<BB',
                                         dict[prev]+1,
                                     len(prev)))


            # Signal finished encoding
            out.write(b'\xfe')
            return


        # Otherwise, write the character to the output stream, and update our stored copy 
        out.write(curr)
        prev = curr

        # Also make sure the new character shows up in the dictionary
        dict[prev] = len(dict)

