#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header is 6 bytes long.
    # The first three are fixed (0x47, 0x49, 0x46).
    # The next one indicates the version of the format; for this exercise it's always 0x38 (the latest version at the time).
    # The last two indicate the width and height of the image in pixel.
    header = b'\x47\x49\x46\x38\x39\x3a'.ljust(6)

    # Width and height are unsigned shorts little endian, so we need to convert them from an integer (int -> bytes[2])
    # https://docs.python.org/3/library/struct.html#format-characters
    widthheight = struct.pack('<H', int.from_bytes(rng.read(2), byteorder='little'))

    # Now comes another fixed value that is not used in this exercise but should be present if more frames will follow
    palettesize = b'\xf9\xff'

    # And finally some color information
    bgcolor = b'\xfe\xff'
    aspectratio = b'\x01\x01'

    # Generate the end of file block. It has its length up to four bytes with the least significant bits set to zero.
    # The length is calculated by subtracting the current position from the start address of the block plus 16 (length of EOI block).
    eofblocklen = b'\x00' * ((5 + len(widthheight)) % 2 ** 8 - 1)[::-1]

    # Write everything!
    out.write(header)
    out.write(widthheight)
    out.write(palettesize)
    out.write(bgcolor)
    out.write(aspectratio)
    out.write(eofblocklen + b'\x3b')