#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO, Optional


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF has an average color distance of `-127`.
# If `seed` is not None then it will be used as the seed for /dev/urandom.
# Use this to have repeatable output!
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, seed: Optional[int] = None):

    header = b'GIF89a'
    width = 640
    height = 480

    if seed is not None:
        # re-initialize rng with the given seed
        rng.seek(0, os.SEEK_SET)
        rng.truncate()
        rng.write(struct.pack('I', seed))

    # Write out the header.
    out.write(header)

    # Write out the logical screen descriptor.
    out.write(struct.pack('<HHBBB',
         width, height,      # Screen Width and Height
         0xF0 | 0x0C,          # Packed field (Color Table Size)
         b'\x00'[0], b'\x00'[0]))    # Background Color Index & Pixel Aspect Ratio
    # TODO: x/y offsets?
    # TODO: global color table size?

    # Write out the backgroundColor
    out.write(b'\xFF\xFF\xFF')

    # Write out the pixel aspect ratio
    out.write(b'\x00\x00')


    # Open up our buffer to store the palette chunk.
    buf = bytearray()

    # Generate the global color table.
    for _ in range(0, pow(2, 5)):   # black, white
        buf += rng.read(3)         # red, green, blue

    # Write out the palette chunk.
    out.write(buf[:pow(2, 5)])     # only write our first "actual" palette entry
    buf = buf[len(buf)//2:]        # remove what we wrote from the buffer

    # Initialize local color table with the first actual palette entry
    localColorTable = [None]*pow(2, 5)
    localColorTable[0] = bytes([ buf[i*3+0], buf[i*3+1], buf[i*3+2]])

    while len(localColorTable) < pow(2, 4)*255:

        # Create new RGB value based on the last one.
        newRgbValue = list(struct.unpack("BB", rng.read(2)))
        newRgbValue[0] = max(newRgbValue[0]-127, 0)
        newRgbValue[1] = min(newRgbValue[1]+127, 255)

        prevIndex = None
        for i in range(0, len(localColorTable)):

            if all(c == newRgbValue[j] for j, c in enumerate(localColorTable[i])):
                prevIndex = i
                break

        assert prevIndex is not None          # should never happen...
        localColorTable.append(bytes([ newRgbValue[0], newRgbValue[1], newRgbValue[2]] ))


    # Write out the global color table.
    for i in range(0, len(localColorTable)):
        out.write(localColorTable[i])

    # Write out graphics control extension
    out.write((b'\x21\xF9\x04').hex().encode('utf-8'))

    # Write out application extension
    out.write((b'\x21\xFF\x0BNETSCAPE2.0').hex().encode('utf-8'))
    out.write((b'\x03\x01\x00\x00\x00').hex().encode('utf-8'))

    # Write out image descriptor
    out.write((b'\x2C\x00\x00\x00').hex().encode('utf-8'))

    # Write out the Image Data.
    imageData = []
    predictedImageData = []
    for y in range(height):

        # Create new RGB value based on the last one.
        r = rng.read(2)[0]
        g = rng.read(2)[0]
        b = rng.read(2)[0]
        rgbValue = [r, g, b]
        rgbValue[0] = max(rgbValue[0]-127, 0)
        rgbValue[1] = min(rgbValue[1]+127, 255)

        index = -1;
        for lci in range(0, 64-1):
            if all(c == rgbValue[j] for j, c in enumerate(localColorTable[(lci*256)+r])):
                index = lci * 256 + r
                break

        assert(index!= -1)       # Should never happen...

        nextR = (((rgbValue[0]-1)&0xff)<<8)|(((rgbValue[0]-1+1)&0xff00)>>8)
        nextG = ((rgbValue[1]&0xff)<<8)|((rgbValue[1]&0xff00)>>8)
        nextB = ((rgbValue[2]&0xff)<<8)|((rgbValue[2]&0xff00)>>8)

        # Write out the packed LZW block.
        lzwLen = 0

        codeMap = dict()           # Maps indicies (ints) to LZ codes (tuples of ints)
        codeReverseMap = {}        # Maps LZ codes back to indicies (ints)

        # Codes are tuples of length 2 where each element's bits represent a single character.
        currentCode = tuple(0x00 for _ in range(0, 2))

        # Symbols are integers that map to something else via codeMap.
        symbolList = set([currentCode])

        # How many times does the currentCode have been seen before?
        codeCounts = { currentCode : 1 }


        # Prepare our initial run of pixels
        sliceStart = rng.read(1)[0]
        sliceEnd = rng.read(1)[0]
        numPixels = (sliceEnd - sliceStart) + 1

        # Build the first scanline
        for i in range(numPixels):

            # Predicted next pixel
            pred = predictedImageData[-1][nextX] << 8 | \
                    predictedImageData[-1][nextY]

            # Read in next pixel
            cur = y * width + sliceStart + i

            # Write the predicted pixel
            predictedImageData.append(pred >> 8)

            # Update LZW codes
            currentCode = (currentCode[1], (cur&0xff)-1)
            if currentCode in codeMap:

                codeMap.__setitem__(
                        codeMap[currentCode] + currentCode[0],
                        (codeMap[currentCode] + currentCode[0])[1:])

                codeReverseMap["".join(map(chr, (codeMap[currentCode] + currentCode[0])))[:-1]] = cur


            elif "".join(map(chr, currentCode)).replace("\x00", "") in codeReverseMap:

                # Check if we need to add more bits to the code
                if any( codeMap[key].count > 2**12 - 1 for key in codeMap ):

                    # Get longest string from codeMap
                    longKey = max(
                            codeMap, key=lambda key: len(codeMap[key]), default="")

                    # Remove shortest string from symbolList
                    shortString = sorted(symbolList, key=str)[0]
                    symbolList -= {shortString}


                    codeMap.__setitem__(longKey + "\x00", ("\x00", codeMap[longKey][1:]))
                    codeReverseMap["".join(map(chr, (codeMap[longKey] + ["\x00"])))[:-1]] = cur

                    # Add current code to maps
                    codeMap.__setitem__(longKey + currentCode[0],
                                        (codeMap[longKey] + [currentCode[0]]))

                    codeReverseMap["".join(map(chr, (codeMap[longKey] + [currentCode[0]])))[:-1]] = cur

            else:

                # Add current code to maps
                codeMap.__setitem__(currentCode,
                                    ([currentCode[1]], currentCode[0]))

                codeReverseMap["".join(map(chr, (codeMap[currentCode][:-1] +
                                                    [currentCode[0]]))).replace("\x00",'')] = cur

            # Update codeCount
            if currentCode in codeCounts:
                codeCounts[currentCode] += 1
            else:
                codeCounts[currentCode] = 1

            nextX = 0 if (cur % width == 0) else cur % width - 1

            # Next pixel's Y coordinate is just the same as ours
            nextY = y
        print(f"{100*(y/height)}%")


