#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# - width and height are specified in pixels (2 bytes each)
#   with a maximum of 0xffff.
# - color depth is fixed to 8 bits per channel (not configurable yet)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, 
        height: int, width: int
    ) -> None:

    # Verify that we can actually fit the image inside an 8MiB buffer. 
    assert height * width <= 8 * 1048576

    # Header
    out.write('GIF89a'.encode())
    # Logical screen descriptor
    out.write(struct.pack('<H', width))
    out.write(struct.pack('<H', height))
    out.write('\x87\x00\x00'.encode())
    
    # Global Color Table
    for i in range(2**8):
        out.write((i,)+tuple([int(c*255) for c in [v % 0.6] for v in rng.read(3)])+b'\x00')

    # Image Descriptor
    # Left-to-right ordering
    out.write(b'\x2C')
    out.write(b'\x00\x00')
    out.write(b'\x00')
    out.write(b'\x00')
    out.write(struct.pack("<HH",width,height))
    # Fixed color table size
    out.write(b'\x08')
    # No interlacing
    out.write(b'\x00')

    # The actual pixels!
    while True:
        chunklength = min(rng.read(1)[0], 255)
        if not chunklength: break
        out.write(chunklength.to_bytes(1,'little'))
        out.write(chunklength.to_bytes(1,'little'))
        out.write(rng.read(chunklength*3))