#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file with an arbitrary amount of frames
# into `out` using the 16-bit frame counter as its duration in ms
# and the internal clock granularity as its frame delay time (10ms).
# If exactly one frame is generated, it will be displayed forever.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, frame_count: int=5):

    # Define some constants that we'll need later on
    HEADERSIZE = len('GIF87a')
    FRAMEINFOSIZE = 4
    COLORPALETTESIZE = 2 ** 2 * 3
    DATASIZE = 2 ** 8 // 2 - 2
    TRAILERSIZE = 2
    MINFRAMEDELAYTIME = 2
    MAXFRAMEDELAYTIME = 255

    # Write the Gif Header
    out.write(b"GIF87a")

    # Generate the LZW mincode table
    lzwmincodesize = _generate_lzw_mincode_table(out)
    print("LZW minimum code size:", lzwmincodesize)

    # Write Color Palettes for each Frame
    colorpalettesize = COLORPALETTESIZE
    for i in range(frame_count):
        out.write(_get_colorpalette())

    # Write the Logical Screen Descriptor
    screendescriptor = bytes([
        0x2c,            # Image Separator
        0,               # Left pos
        0,               # Top Pos
        width,           # Width
        0,               # Height
        0b1000,          # Local color map flag
        0,               # Interlace Flag
        0,               # Sorting Flag
        0b11 << 4 |      # Size of GCT
        lzwmincodesize   # Minimum Code Size
    ])
    out.write(screendescriptor)

    # Construct each Frame's Graphic Control Extension
    for i in range(frame_count):

        # The disposal method tells us how to display this frame next time
        disposalmethod = rng.read(1)[0] % 4
        print("Disposal Method", disposalmethod)

        # The input flag determines whether or not a transparency color exists
        hasinputflag = bool(rng.read(1)) & True
        print("Has Input Flag", hasinputflag)

        if disposalmethod == 2:

            # Set background color index to transparent
            # if the Disposal Method was Restore to Background
            backgroundindex = 0

        else:

            # Use a random index from our available colors
            backgroundindex = rng.read(1)[0] % maxcolors

            while backgroundindex >= maxcolors:
                backgroundindex -= 1

        # Get the delay time between displaying this frame and the next
        frameratemoderator = min(MAXFRAMEDELAYTIME, rng.read(1)[0])
        print("Frame Rate Moderator", frameratemoderator)

        # Calculate the actual milliseconds between displaying this frame
        # and the next by multiplying the moderator with the global clock
        # granularity (ie divide 100 with the length of the moderator)
        # Also add a constant value that ensures the frame delay would always
        # be at least MINFRAMEDELAYTIME units long (since we're dividing by
        # 100 which can lead to values < 10)
        framerate = (MINFRAMEDELAYTIME
                    + (100 / frameratemoderator)
                   )

        # Use a random loop count if the Random Loop Count Flag is set
        # This prevents infinite looping in case the user wants to save
        # his animation...
        randloopnum = None
        if bool(rng.read(1)):
            randloopnum = rng.read(1)[0] % 256
            print("Random Loop Count", randloopnum)

        # Create the Graphic Control Extension Block
        gce = bytes([
            0x21,              # Encoded Application Extension Marker

            0xf9,              # Application Extension Label
            0x04,              # Application Extension Block Size
            disposalmethod,    # Disposal Method
            (hasinputflag << 7) | ((backgroundindex & 0xff)),     # User Input Flag
            (framerate & 0xff),             # Transparent Color Index
            ((framerate >> 8) & 0xff),       # Delay Low
            (((width*height) // DATASIZE) & 0xff),   # Transparency Count
            0                    # Transparent Color Index
        ] + [randloopnum] * hasinputflag) # Repeat the last byte until
                                            # the block ends

        # Finally write this frame's information to the output stream
        out.write(gce)

    # Write Global Color Table trailer
    out.write(_get_trailer()[:DATASIZE+1])

    # Finally write all the remaining data to the output stream
    out.write(_get_trailer()[DATASIZE:])
