#!/usr/bin/env python3
"""Generate a random animated GIF file."""

import struct


HEADLEN = 6      # bytes in Gif Header
DATALENBYTES = 3 # bytes in Data Length Indicator


def gen_data():
    """Generate some garbage bytes for our animation frames to animate."""

    # For fun I'll just generate the same amount of "noise" each frame
    with open('/dev/urandom', 'rb') as f:
        return '[{}]'.format(', '.join('0x{:02x}'.format(b)
                                       for b in f.read(15)))


def main() -> None:
    """Write an animated gif consisting of ASCII art images and random
        pixel values.

        Each frame has the following structure:

         * 6-byte header (magic string and version info)
         * 3-byte integer indicating length of image data field
         * N-byte binary image data field containing some garbled
           characters representing pixels
    """

    with open("/tmp/rand.gif", mode="wb") as outfile:
        with open("/dev/urandom", mode="rb") as rfile:

            # Write Gif Header
            headstr = struct.pack('<6s3B',
                                   b'GIF87a',
                                   0xf7, 0, 0,     # Frame size is defaulted
                                   ord('!'))    # Extension Block follows...

            # Write this crazy value to specify how many colors should be used
            # in the global color table (the palette). We use 0xfff which is
            # way too high but we could also specify anything up to 256 here.
            palsize = 0xfff
            outfile.write(headstr +
                          struct.pack("<HBB",
                                      palsize,        # Size of Palette
                                      0,              # Background Color Index
                                      0))             # Pixel aspect ratio

            # Write Netscape Application Extension Block
            appextstr = struct.pack("BccB",
                                    0x21,           # Introducer
                                    0xff,           # Ext Label
                                    b"\x0bNETSCAPE" \
                                    b"2.0",         # App ID
                                    3)              # Subblock length

            subblklen = 3
            appextsubblk = struct.pack("<BBB",
                                       1,               # Loop count (forever)
                                        0,               # Unknown
                                        subblklen)       # Sub-block size

            # Write Comment Extension Block
            commentextstr = struct.pack('<BccB',
                                         0x21,                 # Introducer
                                         0xfe,                 # Ext label
                                         b'\x01\x08Comment',   # Text ID
                                         4)                    # Subblock length

            commenttext = b'This is a comment block'
            commentextsubblk = struct.pack('<{}s'.format(len(commenttext)),
                                           commenttext)

            # Write graphic control extension
            grfctrlstr = struct.pack('<BcccHBHBH',
                                     0x21,                 # Introducer
                                     0xf9,                 # Graphic Control Ext Label
                                     b'\x04',              # Flags - transparent
                                     b'\x00',              # Transparent Color Index
                                     0x00,                  # Delay time
                                     0x00,                  # Undefined
                                     0x00)                  # Terminator

            # Write Logical Screen Descriptor
            lssdescstr = struct.pack('>HHBBBHHHB',
                                     640,                # Width
                                     480,                # Height
                                     8,                  # Bit depth
                                     0,                  # Color table size
                                     0,                  # Sort flag
                                     0,                  # Global Color Table Flag
                                     7,                  # Background Color Index
                                     0)                  # Pixel Aspect Ratio

            # Now we write all these blocks to the output file
            outfile.write(appextstr + appextsubblk +
                          commentextstr + commentextsubblk +
                          grfctrlstr +
                          lssdescstr)


            # The first thing we do is make sure that we alternate between two
            # different objects we want to put on top of the image
            dataobj = list([gen_data(), gen_data()])
            dataobjiter = iter(dataobj)

            # Next we will write the image in multiple frames so it can be animated
            while True:
                dataobjsize = len(','.join(dataobjiter).encode())

                with open('/dev/urandom', 'rb') as infile:
                    imgdat = infile.read(dataobjsize)
                    # The last byte in the GIF file specifies the LZW Minimum Code Size
                    # 2 gives us better compression than 1 or even 3
                    mincodesize = 2
                    datlenstr = dataobjsize.to_bytes((DATAILENBYTES // 2) + 1,
                                                     byteorder='little')

                    outfile.write(struct.pack('<BH',
                                               dataobjsize,
                                               mincodesize)
                                  + datlenstr.ljust(DATAILENBYTES))
                    outfile.write(imgdat)



if __name__ == '__main__':
    main()