#! /usr/bin/python3


import os
import sys

import argparse

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('filesize', type=int,)
    args = parser.parse_args()

    if not 0 < args.filesize <= (1 << 24):
        print("File size must be between 1 and 16 MB")
        sys.exit(-1)

    with open("/dev/urandom", 'rb') as f:

        # Header of all GIFs starts with "GIF87a" or "GIF89a". The first
        # two bytes define the version of the format. We will use version
        # 87 since it seems to the most widely supported by browsers
        # https://en.wikipedia.org/wiki/Graphics_Interchange_Format#Versions
        magic_number = b'GIF87a'

        # LSB is set for global color table. Since we are creating a plain RGB
        # image without any fancy features like transparency, we don't need this
        # feature. Thus we always have to set the flag at least once.
        # See https://www.w3schools.com/tags/canvas_createimagedata.asp
        flags = '\x01\x00'

        # The logical screen descriptor contains information about the entire
        # canvas the graphics commands are drawn on.
        width = str.encode(str(os.get_terminal_size().columns).zfill(3))
        height = str.encode(str(os.get_terminal_size().lines-1).zfill(3))
        # Screen descriptors specify the width and height of a canvas in pixels.
        # Values up to 65535 may be used. The upper bound is due to historical
        # limitations; larger images can be downsampled (and scaled back up) when
        # displayed. As an example, a 200x200 pixel image is stored in a single
        # 800 byte block; the 200x200 region of interest is specified within that
        # block with its lower left corner (at offset 0x0A - 10 decimal) and its
        # dimensions (at offset 0x0C - 12 decimal).
        screen_desc = \
                width + \
                height + \
                b'\xF7\xFF' + \
                flags + \
                b'\x00' + \
                b'\x00'

        # After the screen descriptor comes a local color table containing
        # palette entries. This allows palettes >= 256 colors. Since we only want
        # to demonstrate how terminal rendering works, we only store one entry which
        # defines a pure black color (#000000)
        num_colors = b'\x01\x00'   # Number of colors in the palette
        gct = b'\x00\x00\x00'      # Actual color data, 24-bit RGB for black

        # Finally follows the image data itself, starting after the GCE. It is
        # simply made of a series of blocks called frames. Each frame begins withe
        # Image Descriptor, followed by the actual image data. The image data is
        # stored as a raster of pixels, where each pixel is defined by its
        # corresponding color from the local color table. To make things easier
        # we just fill the whole image with pixels forming a rectangle.
        frame_width = str.encode(str((args.filesize // 3) //
                                    os.get_terminal_size().columns).zfill(3))
        frame_height = str.encode(str(((args.filesize // 3) %
                                     os.get_terminal_size().columns)+1).zfill(3))
        # Even though each pixel uses exactly one byte to represent its
        # color index, there might exist unused color indexes. In order to
        # calculate an absolute position within the bitmap we divide the
        # length of our data stream by three to increase the color depth of
        # the resulting image by factor three.
        imd = \
                frame_width + \
                frame_height + \
                b'\xF7\xFF' + \
                flags + \
                b'\x00' + \
                b'\x00'


        # Calculate and add the local color table to the generated binary data
        payload_cnt = len(magic_number) + \
                      len(screen_desc) + \
                      1 + len(num_colors) + len(gct) + \
                      len(imd) * 3 + \
                      len(b'\x00' * args.filesize)
        print("Payload count:", hex(payload_cnt), "(expected)",
              hex(len(magic_number) +
                  len(screen_desc) +
                  1 + len(num_colors) + len(gct)))

        # Create a GIF datastream according to the format specification found here:
        # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
        with open("test.gif", 'wb+') as o:

            # Write the fixed part of the binary input
            o.write(magic_number)        # Magic number
            o.write(struct.pack("<H", payload_cnt))       # Payload count
            o.write(screen_desc)         # Screen descritor
            o.write(b'#')                # Extension introducer
            o.write(num_colors)          # Block label
            o.write(gct)                 # Color table bytes...

            #...followed by the 3 image data blocks
            for i in range(3):

                o.write(b',')            # Extension introducer
                o.write(imd)             # Image descriptor
                o.write(b'\x00' * args.filesize)     # Pixel data

            # Add some trailing zeroes to align the file's end properly
            padding = b'\x00' * ((1 << 24)-o.tell())
            assert o.tell() == (1 << 24), ("Output file must be exactly "
                                           "{0} bytes long!".format(hex(padding)))
            o.write(padding)