# Copyright (c) 2017 <NAME> <<EMAIL>>
# This file is part of the "VK scripts" project.
# For details, see https://github.com/egor-tensin/vk-scripts.
# Distributed under the MIT License.

import sys


class RandomGifGeneratorError(Exception):
    def __init__(self, message):
        super().__init__(message)


# Generates a random GIF file with the given width and height
# from /dev/urandom.
def main():
    if len(sys.argv)!= 4:
        raise RandomGifGeneratorError('usage: {} WIDTH HEIGHT OUTFILE'.format(
            sys.argv[0]))

    try:
        width = int(sys.argv[1])
        height = int(sys.argv[2])
    except ValueError as e:
        raise RandomGifGeneratorError(str(e))

    if not (width > 0 and height > 0):
        msg = 'both width ({}) and height ({}) must be positive integers' \
             .format(width, height)
        raise RandomGifGeneratorError(msg)

    outfile = open(sys.argv[3], mode='wb')

    # The first three bytes are always "GIF".
    outfile.write(b'\x47\x49\x46')

    # Then follows a 16 bit unsigned integer denoting the denominator of
    # the original frame rate divisor. Usually it's either 5, 10 or 15 fps.
    # See http://www.matthewflickinger.com/lab/whatsinagif/bits_and_headers.asp.
    outfile.write(struct.pack('<H', 10))

    # Next comes an unsigned char denoting the color depth:
    #    - 0 means no global color table;
    #    - 1 means that each pixel is a single-byte value, corresponding to one
    #      of the palettes below;
    #    - 2 means that each pixel is a two-byte value, corresponding to one of
    #      the palettes below;
    #    - 3 means each pixel is a 3-byte RGB pixel, with no global palette.
    # We use only 256 colors so we don't need a global color table.
    outfile.write(b'\x2C')

    # Width and height in pixels.
    outfile.write(struct.pack('<HH', width, height))

    # First byte is packed binary:
    #     - bits 0..2: disposal method (none)
    #       0 -> do nothing (use previous frame);
    #       1 -> replace background with transparency color (from GCT);
    #       2 -> restore to background color (from GCT).
    #       Note that you may also choose to start with none before starting this
    #       animation, but then another option becomes available by which you can
    #       show this frame after having restored it to its initial state.
    # So we just pick zero for now.
    outfile.write(b'\x00')

    # The second byte is packed binary:
    #     - bits 0..2: user input flag (set)
    #       0 -> no keyboard input allowed;
    #       1 -> allow key input;
    #       2 -> reserved for future use.
    #     - bit 3: transparent color flag (unset)
    #       0 -> no transparancy used;
    #       1 -> transparant color index follows.
    #     - bit 4..7: delay time in units of 1/100 seconds.
    # We don't let users interact with our animations so we set both flags to 0,
    # and set the delay to 0 ms.
    outfile.write(b'\x00')

    # Finally, there should be exactly 256*3=768 null bytes which represent the
    # entire palette, although we only use 256 distinct colors. Each color entry
    # consists of three bytes: red, green and blue components, so each palette
    # entry takes up 3*256=768 bytes. To generate some non-repeating random
    # colors quickly I'm generating them using something like this:
    #   >>> import urandom as r
    #   >>> [r.randint(0, 255) for _ in range(256 * 3)]
    # Or, more compactly:
    #   >>> [i for i in range(0, 256 * 3)]
    # And finally, writing everything at once:
    outfile.write(bytearray([
        0xDE, 0xAD, 0xBEEF,
        0xCA, 0xFE, 0xBA,
        0xFF, 0x00, 0xFB,
        0xEF, 0xBE, 0xAD,
     ] * 256))

    # Now comes an unsigned char denoting the LZW minimum code size, usually
    # between 2 and 9 inclusive. Since we already know how many colors we're
    # going to have in our GIF we might as well use all of them.
    outfile.write(b'\xEC')

    # Then come some blocks containing compressed image data. Each block starts
    # with a 1-byte prefix telling us how large they are. Following that there's
    # a bitmap of width*height bits, where every bit corresponds to a single
    # pixel in the image. Pixels are encoded over 8 contiguous horizontal pixels
    # in a row until their end, and then proceeding vertically downwards in
    # subsequent rows until the bottom is reached.
    sequence = []
    while True:
        length = rng.randint(0, 256)

        if len(sequence) >= 256 - length:
            break

        sequence += list(range(length + 1))

    print(len(sequence))
    assert False


    while True:
        length = rng.randint(0, 256)

        if len(sequence) >= 256 - length:
            break

        sequence += list(range(length + 1))

        outfile.write(bytes((length,)) + sequence[-length:])

    outfile.close()
