#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and writes it to stdout.
def generate_random(out:BinaryIO, rng:BinaryIO):
    # Header

    out.write("GIF89a".encode())  # magic code (6 bytes)
    width = int.from_bytes(rng.read(1), byteorder="big") % 256 + 1
    height = int.from_bytes(rng.read(1), byteorder="big") % 256 + 1

    out.write((width).to_bytes(2, byteorder="little"))
    out.write((height).to_bytes(2, byteorder="little"))

    # Logical screen descriptor, color table size 0
    out.write("\x87\x00".encode())

    # Image descriptor
    left = 0
    top = 0
    width = 0xff & ((int.from_bytes(rng.read(1), "big") >> 4) |
                    (int.from_bytes(rng.read(1), "big") << 4))
    if not width or width == 0x100:
        return False
    height = 0xff & ((int.from_bytes(rng.read(1), "big") >> 4) |
                     (int.from_bytes(rng.read(1), "big") << 4))
    if not height or height == 0x100:
        return False
    
    packedByte = int.from_bytes(rng.read(1),"big")
    flags = {
        0b000000 : bool(packedByte&0b00000001), 
        0b000001 : bool(packedByte&0b00000010), 
        0b000010 : bool(packedByte&0b00000100),
        0b000100 : bool(packedByte&0b00001000),
        0b001000 : bool(packedByte&0b00010000),
        0b010000 : bool(packedByte&0b00100000),
        0b100000 : bool(packedByte&0b01000000)
    }
    print(flags)
    lctFlag = flags[0] 
    interlaceFlag = flags[1] 
    sortFlag = flags[2]  
    reservedByte = flags[3] 
    lctSize = None
    if not lctFlag: 
        lctSize = 0
    else:    
        lctSize = [
            2**i for i in range(1, 9)
            if ((lctSize:=int.from_bytes(rng.read(1),byteorder='big'))!=0)
        ][0] 

    localColorTableFlag = flags[4] 
         
        
    backgroundIndex = int.from_bytes(rng.read(1), byteorder="big") % \
                      len(COLORS)
    pixelAspectRatio = int.from_bytes(rng.read(1), byteorder="big")
    

    out.write(((left).to_bytes(2, byteorder="little")))
    out.write(((top).to_bytes(2, byteorder="little")))
    out.write((((width - 1) // 0x10) + 1).to_bytes(1, byteorder="little"))
    out.write((((height - 1) // 0x10) + 1).to_bytes(1, byteorder="little"))

    
    packedByte = (((backgroundIndex & 0xF) << 7)
                  | ((pixelAspectRatio & 0x7) << 4) 
                  | ((lctSize & 0x7)))

    out.write((packedByte).to_bytes(1, byteorder="little"))


    if localColorTableFlag:  
        for x in COLORS[:lctSize]: 
            r, g, b = x // 256**(2), (x % 256**(2))/256, x % 256 
            out.write(struct.pack('BBB',r,g,b)) 


    LZWMinCodeSize = int.from_bytes(rng.read(1),"big") 
    out.write(LZWMinCodeSize.to_bytes(1,"little"))



    # Block info for image data section
    numBlocks = int.from_bytes(rng.read(1),"big")
    out.write(numBlocks.to_bytes(1,"little"))


    # Image Data Section
    while True:

        chunk = int.from_bytes(rng.read(1),"big")
        
        if chunk == 0xff:

            chunk = int.from_bytes(rng.read(1),"big")
            if chunk == 0x00:
                break
            elif (chunk >= 0x01 and chunk <= 0x0f): 
                lengthBytes = []
                while True: 
                    lengthBytes += [(chunk)] 
                    chunk = int.from_bytes(rng.read(1),"big")

                    if chunk & 0x80!= 0x80: 
                        break

                length = int("".join([str(s) for s in lengthBytes]),2)


                subBlocks = b""  
                while length > 0: 
                    subBlockLength = min(length,(2**8)-1) 
                    subBlocks += b"\x00" * subBlockLength  
                    length -= subBlockLength
                
                out.write(subBlocks)
                
            elif chunk == 0x11: 
                pass
            
            elif chunk == 0x12: 
                while True: 
                    chunk = int.from_bytes(rng.read(1),"big")
                    if chunk == 0x00:
                        break
                    
            elif chunk == 0x1c:
                pass
            
            elif chunk == 0xfe: 
                continue

            else: 
                raise Exception("Chunk unknown: {}".format(hex(chunk))) 

        elif chunk < 0x1F:
            lengthBytes = []
            while True: 
                lengthBytes += [(chunk)] 
                chunk = int.from_bytes(rng.read(1),"big")
                if chunk & 0x80!= 0x80: 
                    break

            length = int("".join([str(s) for s in lengthBytes]),2)


            subBlocks = b""  
            while length > 0: 
                subBlockLength = min(length,(2**8)-1) 
                subBlocks += b"\x00" * subBlockLength  
                length -= subBlockLength
            
            out.write(subBlocks)
            
        else: 
            subBlocks = "" 
            while len(subBlocks) < (2**8-1): 
                char = chr(chunk+33)
                if char == "!": 
                    pass 
                subBlocks += char

                chunk = int.from_bytes(rng.read(1),"big")
            
            out.write(subBlocks.encode()) 
            
                
    return True


if