#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the binary data in `in`.


def main(args):
    if len(args)!= 2:
        print("Usage: gen_random.py <in> <out>")

    with open(args[0], 'rb') as inp:

        def readint():
            return int.from_bytes(inp.read(4), byteorder='little', signed=False)

        magic = inp.read(6).decode('ascii', errors='ignore').strip()
        assert magic == "GIF89a" or magic == "GIF87a", f"'{magic}' is not a valid gif!"

        num_frames = readint()
        width = readint()
        height = readint()

        out_frames = []

        for i in range(num_frames):

            # Skip image descriptor block, just skip it all
            inp.seek(10)

            while True:
                lsb = inp.read(1)[0]

                size = (lsb & 0x0f) + 1
                typ = ((lsb & 0xf0) >> 4) | (inp.read(1)[0] << 4)

                if typ == 0x08:
                    break
                else:
                    for _ in range(size):
                        inp.read(1)

            num_colors = readint()

            global_color_table_flag = bool(inp.read(1))
            sorted_color_table_flag = bool(inp.read(1))
            color_resolution = inp.read(1)[0] + 1
            gct_sort_method = inp.read(1)[0]

            background_color = inp.read(1)
            pixel_aspect_ratio = inp.read(1)

            out_frame = b""

            if global_color_table_flag:

                for i in range(pow(2, color_resolution)):

                    r = inp.read(1)
                    g = inp.read(1)
                    b = inp.read(1)

                    out_frame += r + g + b

            elif sorted_color_table_flag and not global_color_table_flag:
                raise Exception("Sorry, I don't support this yet")

            else:
                raise Exception("No idea what to do here!")

            left = readint() - 1
            top = readint() - 1
            width = readint()
            height = readint()

            packed_bits = False
            lzw_min_code_size = inp.read(1)[0]

            # Use inline LZW decoding from https://www.matthewflickinger.com/lab/whatsinagif/lzw_image_data.asp
            prev_pixel = bytes([background_color[0]]) * pow(2, color_resolution)

            dict_size = pow(2, lzw_min_code_size + 1) + 2

            decode_buffer = [b"\xff"] * dict_size

            nxt_bit = 8
            cur_byte = 0
            cur_mask = 2 ** nxt_bit - 1
            bitbuf = 0
            bufptr = 0

            code_len = None
            buf_complete = False
            while not buf_complete:
                if not code_len:
                    bits = inp.read(1)[0]

                    extra_code_bits = None
                    if bits == 0:

                        # Read the next byte as many times as needed until we get one that isn't part of the repeat sequence
                        bits = 255

                        repeat_start_val = 257
                        while (bits >= repeat_start_val
                               and repeat_start_val <= dict_size):
                            count = repeat_start_val

                            while count > 0:
                                value = inp.read(1)[0]

                                for j in range(value + 1):
                                    decode_buffer[j + dict_size] = decode_buffer[j]

                                count -= value + 1
                            dict_size += (repeat_start_val - 257)

                            bits = inp.read(1)[0]

                        extra_code_bits = inp.read(1)[0]

                        code_len = (bits) + 1

                    else:
                        code_len = (bits) + 1

                for _ in range(nxt_bit, code_len):

                    if not packed_bits:
                        bitbuf |= (inp.read(1)[0]) << bufptr
                        bufptr += 8
                    else:
                        bitbuf |= (inp.read(1)[0] & 0x7f) << bufptr
                        bufptr += 7

                    mask = ~(~0 << bufptr)
                    nxt_bit = min((max(0, bufptr - 8)),
                                  max_code_len)

                    if nxt_bit == 0:
                        packed_bits = True
                        continue
                    else:
                        packed_bits = False

                    new_bits = (((bitbuf) & mask) >> bufptr) << nxt_bit
                    bitbuf &= ~(((~0) << nxt_bit) ^ (~0))

                    if not packed_bits:
                        bitbuf |= new_bits
                    else:
                        bitbuf |= new_bits & 0xFF

                cur_byte >>= code_len
                cur_byte <<= code_len

                index = (cur_byte | bitbuf) & cur_mask

                out_pixels = bytes([])
                out_pixels += prev_pixel

                val = decode_buffer[index]

                while val!= b'\xff':
                    out_pixels += decode_buffer[index][1:]

                    val = decode_buffer[index]

                    cur_byte >>= code_len
                    cur_byte <<= code_len

                    index = (cur_byte | bitbuf) & cur_mask

                current_pix = tuple(prev_pixel)[:3]

                for c in out_pixels:
                    if len(current_pix) == 1:
                        current_pix = (c,) + current_pix
                    else:
                        current_pix = (c,) + current_pix[:-1]

                prev_pixel = struct.pack("<BBB", *current_pix)

            out_frame += prev_pixel
            out_frame += b"\x00" * 12

            out_frames.append(out_frame)

    with open(args[1], 'wb') as outp:
        outp.write(struct.pack('<6sHHHBB',
                                b'GIF89a',
                                4 * len(out_frames), width, height, 0x80, 0, 0))

        outp.write(b"\x21\xF9\x04\x05"
                   + struct.pack(">xxxxxxBxxxBBBHBxxx",
                     6*width, 6*height, 0x80, 0x00, 0, 0x00))

        for frame in out_frames:
            outp.write(frame)


if