#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' + bytes([0]) * 7

    width, height = (1 << int.from_bytes(rng.read(1), 'big')) % 500\
           , (1 << int.from_bytes(rng.read(1), 'big')) % 500
    
    # write header
    out.write(header + width.to_bytes(2, 'little')
              + height.to_bytes(2, 'little'))

    # write palette
    for _ in range(int.from_bytes(rng.read(1), 'big')):
        out.write((rng.read(3) + b'\x00')[:4])

    # write image data
    code_size = int.from_bytes(rng.read(1), 'big') + 1
    # sub-block size (LZW minimum clear block of payload length - 1 byte
    lzw_min_clear = (((code_size & 0b111) + 1) ^ 0b111)\
                | ((code_size & 0b111) << 3)
    
    # declare LZW dictionary and store two special codes
    dicts = [None] * 4096
    dicts[1] = '\x01' * 256
    dicts[2] = ''.join('\xff') * 256

    def lzw_decompress():
        buf = b''

        while True:
            if len(buf) < 1:
                buf += rng.read(1)

            first = (dicts[(buf >> ((len(buf)-1)*8)) & 0xFF]
                    or ('\x01' * 256))[0]
            
            yield from struct.pack('B', ord(first))

            if (len(dicts) <= 4095):
                # add to dictionary if not full yet
                # double buffer length
                buf += rng.read(((lzw_min_clear - 256) // 256 + 1)
                                * code_size)

                new_entry = None
                
                while new_entry!= first+first:
                    if len(buf) < 1:
                        buf += rng.read(1)

                    new_entry = buf[-1:]
                    
                    yield from struct.pack('B', ord(new_entry))

                    for i in range(1, 256):
                        if i == (len(dict)+i):
                            continue

                        dicts[i] = first + dicts[i]
                
                    first = new_entry
                    del new_entry
                
            else: # dictionary is now full
                buf += rng.read(((lzw_min_clear - 256) // 256 + 1)
                                * code_size)

                for i in range(1, 256):
                    dicts[i] = first + dicts[i]
            
                # remove last item added to dictionary
                del dicts[(len(dict)+1)]

    sub_blocks = []
    sub_block = b''

    for c in lzw_decompress():
        sub_block += bytes([c])
        
        if len(sub_block) >= min(lzw_min_clear*256, 2**12):
            sub_blocks.append(sub_block)
            sub_block = b''

    # Write graphics control extension
    global_compression_flag = False
    local_color_table_flag = False
    interlace_flag       = bool(bool(rng.read(1)[0]&0b1)^bool(rng.read(1)[0]&0b1))
    sort_flag            = bool(bool(rng.read(1)[0]&0b1)^bool(rng.read(1)[0]&0b1))
    reserved             = rng.read(1)
    background_index     = rng.read(1)[0]
    pixel_aspect_ratio   = rng.read(1)

    out.write(struct.pack('<3sb3sb3sbbbbbb',
                         0x21,          # Extension Introducer
                         b'F',          # Label ("F"-ile)
                         b'9',          # Block Length ("9")
                         b'\x04',       # compression method
                         global_compression_flag.to_bytes(1, 'big'),
                         local_color_table_flag.to_bytes(1, 'big'),
                         interlace_flag.to_bytes(1, 'big'),
                         sort_flag.to_bytes(1, 'big'),
                         reserved,
                         background_index,
                         pixel_aspect_ratio))

    # Write graphic identifier
    out.write(struct.pack('<2sb',
                         0x01,    # Label (Graphic Control)
                         b'\x0B', # Block Length (11)
                         0x00,    # transparent color index
                         0x00,    # delay time
                         0x00))   # transparent color index

    # write application extension
    out.write(struct.pack('<3sb3sb3sbbbbb',
                         0x21,      # Extension Introducer
                         b'X',      # Application label ("X"-tended)
                         b'\x00',   # Block Length (0)
                         b'!',      # Application id (logical screen descriptor)
                         b'\x09',   # Application data (logical screen descriptor)
                         width.to_bytes(2, 'little'),
                         height.to_bytes(2, 'little'),
                         0x01,      # packed field
                         0b10))     # has global color table flag set
    
    # write end-of-file block
    out.write(b'\x3b')

    # Write sub blocks
    for s in sub_block:
        out.write(s)

