#!/usr/bin/env python3

import sys
import struct
import binascii
from typing import BinaryIO


def generate_gif(
        width: int=256, height:int=256, color_depth:int=1, is_color:bool=False,
        global_palette:list=None, frames:list=[], outfile:str=""
    ) -> bytes:

    # Header
    version = 0x89
    widthL = (width & 0xFF).to_bytes(1, 'little', signed=False)
    heightL = (height & 0xFF).to_bytes(1, 'little', signed=False)
    flags = ((is_color<<7)|(color_depth&7)).to_bytes(1, 'little', signed=False)
    header = bytes([version]) + bytes([widthL[0]]) + bytes([heightL[0]]) + flags


    output = []

    if len(global_palette)>0 and not is_color:
        raise ValueError("A palette can only be specified for an indexed color image")

    # Global Color Table
    if len(global_palette)>0:     
        numColors = min(len(global_palette)//3, 2**color_depth)
        palette = [b'\x00\x00\x00'] * numColors * 3

        for i in range(numColors):
            c = map(lambda x : max(min(round(x*0xFF), 0xFF), 0), global_palette[(i*3):((i+1)*3)])
            palette[3*i:(i+1)*3] = list(map(lambda x: x.to_bytes(1,'big'), c))

        output += [(0xFE, b''.join(palette))]

    else: 
        output += [(0xF7, bytes())]

    frame_offsets = []
    photodataoffset = 0
    
    for f in reversed(frames):
        
        left = (f['left']//256).to_bytes(1,"little",signed=True)
        top = (f['top']//256).to_bytes(1,"little",signed=True)
        widthL = (f['width'] & 0xFF).to_bytes(1,"little",signed=False)
        heightL = (f['height'] & 0xFF).to_bytes(1,"little",signed=False)
        flags = (((not f['transparent']) & 1)<<7 | 
                 (f['interlaced'] & 1)<<6 |
                 (f['sortflag'] & 1)<<5 | 
                 0 <<4 |
                 0 <<3 | 
                 0 <<2 |
                 0 <<1 |
                 0).to_bytes(1,"little",signed=False)
        delayN = (f['delay'] // 10).to_bytes(1,"little",signed=False)
        delayD = (f['delay'] %  10).to_bytes(1,"little",signed=False)
        transp = (f['transparentColorIndex']).to_bytes(1,"little",signed=False)

        buffer_size = sum(list(map(len,[left,top,widthL,heightL,flags, delayN,delayD])))
        
        block = bytearray()
        block += left
        block += top
        block += widthL
        block += heightL
        block += flags
        block += delayN
        block += delayD

        padding = b"\x00"*(buffer_size - len(block))
        block += padding
        
        framedataoffset = photodataoffset + buffer_size
        block += (framedataoffset).to_bytes(2,"little",signed=False) 

        lzwcodelength = f['lzwcodelength']
        lzwcodelength = (lzwcodelength).to_bytes(1,"little",signed=False)
        block += lzwcodelength

        data = bytearray()
        for row in f['rows']:

            code = f['lzwtable'][row[:lzwcodelength]] 
            data += binascii.unhexlify('{:0{}X}'.format(code,2*lzwcodelength))
            
            while len(row)>lzwcodelength:
                # Find longest match
                bestmatch = ''
                for i in range(len(row)-lzwcodelength-1,-1,-1):
                    test = row[i:(i+lzwcodelength)]
                    if test == bestmatch or test in f['lzwtable']:
                        break

                # If we found no matching substring, add the entire unmatched string to our data stream
                if i==-1:
                    data += binascii.unhexlify('{:0{}X}'.format(f['lzwtable'][row],2*len(row)))                    
                    row = ""

                    continue
                else:
                    # We found at least one matching substring
                    
                    # Add this substring's corresponding index value from the LZW table to our data stream
                    data += binascii.unhexlify('{:0{}X}'.format(f['lzwtable'][test],2*lzwcodelength))

                    # Save this match as the next part of the LZW code length 
                    lzwcodelength = len(bestmatch)+1

                    # And remove it from the input string
                    row = row[(i+lzwcodelength):]

        # Write the final pixel data to our output array
        block += data
        photodataoffset += len(block)
        


        frame_offsets += [photodataoffset]
        output += [(255, block)]
        
    endofimage = b""
    endofimage+= (b'\x3B').to_bytes(1, "little", signed=False)
    endofimage+= (b'\x00').to_bytes(1, "little", signed=False)
    output += [(255,endofimage)]

    output = sorted(output,key=lambda x: x[0])

    offset = 0
    for o in output:
        offset += len(o[1])

    size = b''
    size += (width).to_bytes(2,"little", signed=False)    
    size += (height).to_bytes(2,"little", signed=False)    
    size += (len(output)).to_bytes(1,"little", signed=False)  
    size += (size[0]).to_bytes(1,"little", signed=False)  

    for offs in frame_offsets:
        size += (offs).to_bytes(2,"little", signed=False)    

    size += (photodataoffset).to_bytes(3,"little", signed=False) 

    return header+size+b''.join(list(map( lambda x: x[1], output )))
    
