#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the least amount of bytes possible. The problem with
# writing a GIF is that only some fields have variable length, but they always start at the same
# byte. This program will try to write random values in those fields and see which ones are valid.
# It generates a 1x1 pixel image with as few bytes as possible for each frame type. This can then be
# used by other programs to generate a minimal GIF (e.g., strip all frames, set image size to 0).

# Different types of frames:
# - Application extension
# - Comment extension
# - Control extension
# - Graphic control extension
# - Plain text extension

# https://www.fileformat.info/format/gif/corion-blocks.htm

# Fields marked with (*) can be randomly generated. All others are fixed.
#
# Byte   Description                               Notes
#    0   | Signature "GIF"                        Fixed
#    3   | Version "89a"                          Fixed
#    6   | Width                                  (*)
#    8   | Height                                 (*)
#   10   | Flags for global color table           7 bits
#       | Background color index                  Not supported
#       | Pixel aspect ratio                      Fixed
#   12   | Global color map size                  1 bit
#   13   | Global color map sort flag             1 bit
#       | Global color map data                   Variable, 3 * (2 ** ((Global Color Map Size + 1) × 3))
# ????   | Block terminator                       0x00
# ????   | Image separator                        0x2C
#   14   | Local color map size                   Same as global
#   15   | Local color map sort flag              Same as global
#   16   | Local color map data                   Variable, 3 * (2 ** ((Local Color Map Size + 1) × 3))
#   18   | X position of image                    (*)
#   20   | Y position of image                    (*)
#   22   | Width of image                         Same as global
#   24   | Height of image                        Same as global
#   26   | Packed fields                          9 bits
#       | Interlace flag                          Not supported
#       | Sort flag                               Not supported
#       | Reserved                                Must be zero
#       | Local color table size                  Same as local
# ????   | LZW Minimum Code Size                  0x08
# ????   | Block terminator                       0x00
# n+0   | Block terminator                       0x3B

def main() -> None:
    out = open("test_input.gif", "wb")

    # Header
    out.write(struct.pack("<3siHHBB", b"GIF", b"89a", 1, 1, 0, 0))

    out.write(b"\x21\xf9\x04\x04\x00\x00\x00\x00\x0c\x00\x00")

    # Image descriptor
    out.write(struct.pack("<HBBBBB", 0, 0, 1, 1, 0, 0))

    # Graphics control extension
    out.write(b"\x21\xf9\x04\x04\x00\x00\x00\x00\x0c\x00\x00")

    # Raster data
    out.write(b"\x00\x00")

    # Terminators
    out.write(b"\x00\x00\x3B")
