#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the `NxM` dimensions.
def generate_random_gif(out: BinaryIO, width: int, height: int) -> None:
  # Don't forget to use mode="wb" when opening for binary write!
  out.write(b"GIF89a")

  # The header is fixed-size, so we can just generate it ahead of time.
  # Note that we don't need to apply any transformations -- we can simply
  # write all data as-is (e.g., bytes vs shorts -- where would the transformation
  # occur?)
  header = [
    "0x{:<2X}{}".format(*struct.unpack("BB", b"H")),   # Screen Width
    "0x{:<2X}{}".format(*struct.unpack("BB", b"V")),   # Screen Height
    "0xFF{}\x00".format("0xF4"),                       # Global Colour Table Flag + Size
    16 * "\x00"                                        # Background Color Index
  ]
  header[0] = bytes([int(header[0].split()[-1], 16)]).decode("ascii")
  header[1] = bytes([int(header[1].split()[-1], 16)]).decode("ascii")
  header[2] = bytes([int(header[2].split()[-1][2:], 16)]).decode("ascii")
  out.writelines((*map(bytes.fromhex, header),))

  # We'll be writing a series of graphics control blocks followed by an image block
  gc_block = [
    "0x{:<2X}{}".format(*struct.unpack("BB", b"\xF9\x04")),         # Block Size
    "0x{:<2X}{}".format(*struct.unpack("B", b"\x04")),              # Disposal Method
    "0x{:<2X}{}".format(*struct.unpack("B", b"\xF0")),              # Transparent Color Flag
    "0x{:<2X}{}".format(*struct.unpack("H", b"\x05")),              # Delay Time
    "0x{:<2X}{}".format(*struct.unpack("H", b"\xFE")),              # Transparent Color Index
  ]
  gc_block = (*map(bytes.fromhex, gc_block),)

  img_block = [
    "0x{:<2X}{}".format(*struct.unpack("BB", b"\x2C")),             # Image Separator
    "0x{:<2X}{}".format(*struct.unpack("BB", b"\x00\x00")),         # X Position
    "0x{:<2X}{}".format(*struct.unpack("BB", b"\x00\x00")),         # Y Position
    "0x{:<2X}{}".format(*struct.unpack("BB", b"\xF8\x08")),         # Width
    "0x{:<2X}{}".format(*struct.unpack("BB", b"\xF8\x08")),         # Height
    "0x{:<2X}{}".format(*struct.unpack("BB", b"\x00")),             # Local Colour Table Flag
    "0x{:<2X}{}".format(*struct.unpack("B", b"\x7C")),              # Interlaced Flag
    "0x{:<2X}{}".format(*struct.unpack("B", b"\x07")),              # Sort Flag
    "0x{:<2X}{}".format(*struct.unpack("B", b"\xFF")),              # Size of Local Colour Table
    "0x{:<2X}{}".format(*struct.unpack("B", b"\xFF"))               # Extra Trailer Bytes
  ]
  img_block = (*map(bytes.fromhex, img_block),)

  # We will now loop through each pixel and update the values in our blocks.
  for i in range(width):
    print("\rProgress: {} / {}".format(i, width), end='')
    for j in range(height):
      value = struct.unpack("<I", struct.pack("<BBBB", *generate_random_color()))
      img_block += struct.unpack("<B", struct.pack("<B", *value))

  # We must include an application extension block at the end of the file.
  appext_block = [
    "0x{:<2X}{}".format(*struct.unpack("BB", b"\x21\xFF")),     # Application Extension Introducer
    "0x{:<2X}{}".format(*struct.unpack("BB", b"\x0B\x00")),     # Application Identifier Length
    "NETSCAPE2.0",                                               # Application Identifier
    "0x{:<2X}{}".format(*struct.unpack("BB", b"\x03\x01")),     # Application Data Sub-Block ID
    "0x{:<2X}{}".format(*struct.unpack("B", b"\x01")),          # Loop Count
    "0x{:<2X}{}".format(*struct.unpack("B", b"\x00"))           # Block Terminator
  ]
  appext_block = (*map(bytes.fromhex, appext_block),)

  # Write out all blocks
  out.writelines((gc_block, img_block, appext_block))
  print()

# Returns a randomly generated color in RGB format
def generate_random_color():
  return [struct.unpack("B", struct.pack("B", i))[0] for i in range(3)]
