#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate(out: BinaryIO, rng: BinaryIO):
    # Generate the 6-byte header
    out.write(b"GIF87a")                     # Magic bytes
    out.write(struct.pack(">H", 0xff))        # Logical screen descriptor
    out.write(b"\x00\x00\xff\xff")            # Background color index
    out.write(b"\x00")                        # Aspect ratio

    # Generate an arbitrary number of image blocks
    while True:
        flags = int.from_bytes(
            rng.read(1), byteorder="big")
        if (flags & 0b11111000) != 0b11111000:
            break

        left = rng.read(1)[0]
        top = rng.read(1)[0]
        width = rng.read(1)[0] + 1
        height = rng.read(1)[0] + 1
        out.write(bytes([left, top, width, height]))
        if (flags >> 5 & 0b111) == 0b100:      # Color table size
            # Local color table
            for _ in range(2 ** ((flags >> 2 & 0b111)+1)):
                out.write(rng.read(3))         # ARGB value
        else:
            out.write(b"\x00")                 # No local color table

        # Image data header
        flags |= 0b1 << 4                      # LZW minimum code size
        out.write(bytes([flags]))              # Zero-initialized LZW dictionary

        # Generate 256 bytes per block
        for i in range(256):
            if i % 255 == 0 and i != 0:
                out.write(bytes([0x00]))       # Block terminator
                continue

            out.write(rng.read(1))             # Single pixel color index

    # Write the trailer
    out.write(b";")                            # GIF trailer; mandatory

# Reads up to 256 bytes from a GIF file at random positions `in`,
# and randomizes their colors by changing them to black.
# This is slightly faster than `generate()` because we don't need
# to create any new files or read their headers.
def mutate(inp: BinaryIO, rng: BinaryIO):
    pos = 0
    while True:
        try:
            inp.seek(pos, 0)
            start_block = struct.unpack("B", inp.read(1))[0]
            if   start_block == 0x2c:          # Image separator
                inp.seek(-1, 1)
                img_len = struct.unpack(">I", inp.read(3))[0]
                inp.seek(9, 1)                 # Skip over image properties
                for i in range((img_len - 10) // 255 + 1):
                    inp.seek(i * 255, 1)        # Seek to next block
                    inp.read(1)                 # Skip block separator
                    for j in range(255):
                        col = inp.read(1)       # Get a pixel
                        col[0] ^= 0xFF           # XOR it with white
                        inp.seek(-1, 1)          # Seek back to modify it
                        inp.write(col)
            elif start_block == 0x3b:          # EOF
                return
            else:                              # Extension block
                ext_type = inp.read(1)[0]
                len = struct.unpack(">H", inp.read(2))[0]
                inp.seek(len, 1)               # Ignore extension data
        except OSError as e:
            if e.errno == 22:
                break
            raise e

        pos += 1

class ArgumentParserError(Exception): pass

if __name__ == "__main__":
    import sys

    try:
        argc = len(sys.argv)
        if argc not in [3, 4]:
            raise ArgumentParserError()
        mode = sys.argv[1].lower()
        if mode not in ["gen", "mut"]:
            raise ArgumentParserError()
        infile = open(sys.argv[2], "rb")
        outfile = open(sys.argv[3], "wb")
        if argc == 4:
            rng = open(sys.argv[4], "rb")
        else:
            rng = open("/dev/urandom", "rb")

        if mode == "gen":
            generate(outfile, rng)
        else:
            mutate(infile, rng)

    except IndexError:
        print("Usage: %s gen <input file> <output file> [<random seed>]" % \
            sys.argv[0])
        print("\nThis program generates a GIF file using a random number" \
            " generation source.")
        print("The optional random seed is used instead of /dev/urandom" \
            " when given.\n")

        print("%s mut <input file> <output file> <random seed>" % \
            sys.argv[0])
        print("\nThis program modifies the input file in place, by replacing" \
            " every nonzero pixel with a randomly chosen color.")
        print("To make sure the resulting output is still valid, it uses" \
            " the same random numbers as the first command.")

    except IOError as e:
        if e.args[0] == 2:
            print("File '%s' does not exist." % e.filename)
        else:
            raise e

    except ArgumentParserError:
        print("Invalid arguments. See usage above.")

    except Exception as e:
        print("Could not run program: %s" % str(e))

    finally:
        if "infile" in locals():
            infile.close()
        if "outfile" in locals():
            outfile.close()
        if "rng" in locals():
            rng.close()
        # We could just close stdin, but that seems a bit too clever
        osbuffer = getattr(sys.stdin, 'buffer', None)
        if callable(getattr(osbuffer, 'close', None)):
            osbuffer.close()
        else:
            stdin = getattr(sys.stdin, 'fileno', None)
            if callable(stdin):
                os.close(stdin())
        osbuffer = getattr(sys.stdout, 'buffer', None)
        if callable(getattr(osbuffer, 'flush', None)):
            osbuffer.flush()
        else:
            stdout = getattr(sys.stdout, 'fileno', None)
            if callable(stdout):
                os.fsync(stdout())
        osbuffer = getattr(sys.stderr, 'buffer', None)
        if callable(getattr(osbuffer, 'flush', None)):
            osbuffer.flush()
        else:
            stderr = getattr(sys.stderr, 'fileno', None)
            if callable(st                rng.read(1), byteorder="big")
