#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # 63 is the smallest size that is accepted by Ubuntu's GIF viewer
    w = min(int.from_bytes(rng.read(2), byteorder="big"), 63) + 1
    h = min(int.from_bytes(rng.read(2), byteorder="big"), 63) + 1

    # Total number of frames
    f = int.from_bytes(rng.read(1), byteorder="big") % (h * w)

    # Settings for each frame
    d = rng.read(1) != b"\x00"
    i = rng.read(1)[0] & 0b111

    # Palette size
    p = 2**((rng.read(1)[0] & 0b111) + 1)

    if d:
        # We will treat the "local color table" as a subpalette
        # to the global one, so that we can re-use colors easily
        c = max(rng.read(1)[0], 1)
        f += 1
    else:
        # No local color table at all so use all available slots
        c = p

    out.write(struct.pack("B", 71))            # GIF89a signature
    out.write(struct.pack("BBHBBB",          # Logical Screen Descriptor
        0b10000000 | w - 1,                   # Global Color Table Flag
        h - 1,                                # Global Size
        0b000,                                # Color Resolution
        7,                                    # Sort flag
        0,                                    # Global Color Table Size
        0))                                   # Background Color Index

    out.write(struct.pack("<{}B".format(3 * p), *[                      # Global Color Table
            int.from_bytes(rng.read(3), byteorder="big") for _ in range(p)]))

    out.write(struct.pack("!BBB",           # Graphic Control Extension
        0xF9,                               # Block Size
        4 << 2 | 3,                         # Flags
        0xFF & int.from_bytes(rng.read(1), byteorder="big")))       # Delay Time

    # Raster Data
    for y in range(h):
        for x in range(w):
            # If there are more frames than pixels and this one has been used before
            # just skip it
            while f < w * h and (y * w + x) >= f:
                f += 1
                rng.read(1), byteorder="big")
