#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    header += bytearray((
        0x1f, 0x4b,
        0xff, 0xff, # width
        0xff, 0xff, # height
        0xff,       # background color index
        0x00,       # default aspect ratio
        0x00,       # flags
    ))
    out.write(header)

    # 256 colors
    for i in range (0x100):
        out.write(struct.pack('BBB', *[rng.read(1)[0] for _ in range(3)]))

    out.write(bytearray([0x2c])) # image separator
    out.write(bytearray((
        0xff, 0xff, # left offset
        0xff, 0xff, # top offset
        0xff, 0xff, # width
        0xff, 0xff, # height
        0x00,       # local color table flag
        0x00,       # interlaced flag
        0xff,       # sort flag
        0xff,       # size of local color table
        rng.read(1)[0],       # background color index
        0x00,           # pixel aspect ratio
    )))

    # Each frame consists of 7 fields
    #   - graphic control extension
    #   - image descriptor
    #   - local color table
    #   - image data
    #   - terminator
    #   - plain text grid
    #   - application extension

    nframes = rng.read(1)[0] % 0xf
    ngrids = rng.read(1)[0] % 0xf

    while True:
        blocklen = int.from_bytes(rng.read(1), 'big')
        if not blocklen: break

        chunk = bytearray()
        while len(chunk) < blocklen:
            tmp = rng.read(blocklen-len(chunk))
            assert tmp is not None and len(tmp) > 0
            chunk += tmp

        out.write(chunk)

    out.write(bytearray([0x3b])) # trailer

if __name__ == "__main__":
    from random import Random
    with open("/dev/urandom", "rb") as urandom:
        rng = Random(urandom.read(32))

    f = open("random_" + str(rng.getrandbits(32)), "w+b")
    generate_random_gif(urandom, f)
    f.close()

    loopvar = rng.getrandbits(8) & 0xff
    if loopvar == 0xff:
        badval = 0
    else:
        badval = loopvar + 1

    print