#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    width = int.from_bytes(
            rng.read(2), byteorder='little') & 0x7fff
    height = int.from_bytes(
            rng.read(2), byteorder='little') & 0x7fff

    # We want to make sure that we don't accidentally create an image too big!
    while True:
       if width < 1 or height < 1:
           continue

       break

    color_count = int.from_bytes(
                     rng.read(1), byteorder='little') & 0xff

    flags = rng.read(1)[0]

    background_color_index = rng.read(1)[0] & 0xff

    pixel_aspect_ratio = rng.read(1)[0] & 0xff

    out.write(header + 
                (width // 256).to_bytes(1, 'little') + 
                ((width % 256)).to_bytes(1, 'little') + 
                (height // 256).to_bytes(1, 'little') +
                ((height % 256)).to_bytes(1, 'little') + 
                (color_count).to_bytes(1, 'little') + 
                (flags).to_bytes(1, 'little') + 
                (background_color_index).to_bytes(1, 'little') +  
                (pixel_aspect_ratio).to_bytes(1, 'little'))

    palette = []
    for j in range(int(pow(2, (color_count+8)))):

        red = rng.read(1)[0] & 0xff
        green = rng.read(1)[0] & 0xff
        blue = rng.read(1)[0] & 0xff
        
        out.write(b'\xff' + 
                    (red).to_bytes(1, 'little') + 
                    (green).to_bytes(1, 'little') + 
                    (blue).to_bytes(1, 'little'))

    # Write transparent color index
    out.write(b'\x00')

    # Graphics control extension
    out.write(b'\x21\xf9\x04\x01\x00\x00' + 
        (rng.read(3)).to_bytes(1, 'little'))

    # Netscape animation extension
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01' + 
               (512).to_bytes(2, 'little'))

    # logical screen descriptor

    global_color_table_flag_mask = 0x80
    sort_flag_mask = 0x08
    color_resolution_mask = 0x07
    global_color_table_size_mask = 0x70

    flags = rng.read(1)[0] & 0xff

    global_color_table_sort_indicator = (flags & sort_flag_mask) == 0

    color_resolution = (flags & color_resolution_mask) + 1

    background = rng.read(1)[0]

    pixel_aspect_ratio = rng.read(1)[0]

    global_color_table_size = \
                    1 << ((flags & global_color_table_size_mask) + 1)

    size_of_global_color_table = global_color_table_size * 3

    palette = [struct.unpack('BBB', rng.read(3)) for _ in range(size_of_global_color_table)]

    out.write(b'\x2c' + 
                (width // 256).to_bytes(1, 'little') + 
                ((width % 256)).to_bytes(1, 'little') + 
                (height // 256).to_bytes(1, 'little') +
                ((height % 256)).to_bytes(1, 'little') + 
                (flags).to_bytes(1, 'little') + 
                (background).to_bytes(1, 'little') +  
                (pixel_aspect_ratio).to_bytes(1, 'little'))

    # optional global color table
    for j in range(int(pow(2, (color_count+8)))):

        red = rng.read(1)[0] & 0xff
        green = rng.read(1)[0] & 0xff
        blue = rng.read(1)[0] & 0xff
        
        out.write(b'\xff' + 
                    (red).to_bytes(1, 'little') + 
                    (green).to_bytes(1, 'little') + 
                    (blue).to_bytes(1, 'little'))

    # Image Data

    code_size = 10

    code_size_byte = code_size.to_bytes(1,"little")

    first_byte = 0x80 | code_size

    second_byte = 0x00|0x00

    datalen = width*height

    # clear code flag
    databytes = 0

    # read a single byte from file
    def get_bit():
        nonlocal databytes
        out.write(code_size_byte + 
                  bytes([first_byte]) +
                  bytes([second_byte]))


    # write frame data

    for i in range((datalen)):

        start = 0
        end = 16

        val = int.from_bytes(
            rng.read(end - start), 
            byteorder="little")%256

        out.write(((val).to_bytes(
            min(end - start, 1), "little")))
