#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int(2**(int((lsd[4] >> 5)))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = False
    
    while True:
        # Create new frame        
        out.write(b'\x2c')
        
        # Create image descriptor
        id = rng.read(9)
        out.write(id)

        # Read imgdata len, get buffer position to set pointer to later
        bufposition = out.tell()
        imgdatalen = int.from_bytes(rng.read(2), 'big')
        out.seek(bufposition + 1)

        # Get image data
        data = rng.read(imgdatalen)
            
        # Write image data block
        out.write(b'\x2C')
        out.write(struct.pack('<H', imgdatalen))
        out.write(data)

        # Read graphic control extension
        gcxtype, gcxblocksize, gcxflags = struct.unpack("<BBB", rng.read(3))
        if gcxtype == 0xf9:
            disposalmethod = gcxflags & 0b00000011
            if disposalmethod != 3:
                prevframe = False
        else:
            raise ValueError("Unsupported graphics control extension"
                             " type: {}\n".format(gcxtype))

        # 0 - None, 1 - Do not dispose, 2 - Restore to background,
        # 3 - Restore to previous
        dispose = [None, lambda:True, lambda:False,
                   lambda:prevframe][disposalmethod]
      
        # Read image code
        ic = rng.read(1)
        assert ic == b"\x21"
        ic = rng.read(2)
        assert ic in (b"F9", b"FE")
        
        if ic == b"F9":
          codelen = int.from_bytes(rng.read(1), 'big')
          appext = rng.read(codelen-1)
          out.write(appext)
          
          # 99 is the end of the loop indicator for an APPE
          if appext == b"99":
              break
        else:
          trailext = rng.read(1)
          out.write(trailext)
          break
        
        # Get data to restore to
        if not prevframe:
            prevframe=True
            image = data[:]
        else:
            image = prevframe[:]

        # Fill image with background color
        if dispose():
            fillcolor = rng.read(3)
            data = bytearray(fillcolor*image.count(b'\x00'))
            data[::3] = image
        else:
            data = bytearray(image)
        data = bytes(data)
        
        # Replace image data with fixed one
        out.seek(bufposition+1)
        out.write(struct.pack("<H", len(data)))
        out.seek(bufposition+3)
        out.write(
                  bytes(data))

        # Save as previous frame because we might need it later
        prevframe = out.getbuffer()[bufposition:].tobytes()