#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int(2**(int((lsd[4] >> 5)))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = False
    
    while True:
        # Create new frame        
        out.write(b'\x2c')
        
        # Create image descriptor
        id = rng.read(9)
        out.write(id)

        # Get number of bytes to read based on bits in ID
        imgdatalenbits = int(id[8])&0x7
        imgdatalen = int(id[4])*pow(2,imgdatalenbits+1)+imgdatalenbits+1
        codelen = pow(2,(id[6]&7))+1

        # Get dispose method, if any
        dispose = -1 if (id[8]&8)==0 else id[8]&7

        if len(prevframe)!= 0 and \
           imgdatalen < len(prevframe) and \
           (dispose==-1 or dispose<=2):
            out.write(prevframe[:imgdatalen])

        elif dispose == 2:
            out.write(prevframe[-imgdatalen:])

        else:
            out.write(b'\0'*imgdatalen)
            
            
        # If we don't yet know how much image data we need per line,
        # get that info now by reading the first row of pixels
        if codelen == -1:
            out.seek(-1, 1)

            # If we've already generated an entire line, just use that data
            if imgdatalenbits <= 6:
                codelen = 2**imgdatalenbits
            # Otherwise, generate more data until this many bytes are available
            else:
                codelen = imgdatalen-(codelen*(codelen & 1))

            # Go back to where we started writing
            out.seek(-codelen-2, 1)

        # Add code size
        out.write(bytes([codelen]))

        # Write all of our pixel data
        out.write(rng.read(codelen))

        # Add terminating data
        out.write(struct.pack('BB', 0, 0))

        # Retain the last line to use for disposal
        prevframe += struct.unpack("B"*imgdatalen, rng.read(imgdatalen))
