#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int(2**(int((lsd[4] >> 5)))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = False

    while True:
        # Read in one byte at a time until we find an image descriptor
        blockhead = rng.read(1)
        
        # End of file
        if not blockhead:
            break
            
        # Image Descriptor
        elif blockhead == b'\x2c':
            idblock = rng.read(9)
            out.write(idblock)

            # Read image data length but don't write anywhere yet
            imagedatablocklen = int.from_bytes((rng.read(2)), 'big') + 1

            # Read code size but don't use yet
            codelen = int.from_bytes((rng.read(1)), 'big') + 1

            # Read disposal method but don't use yet
            dispose = int.from_bytes((rng.read(1)), 'big') & 7
  
            # Copy unused flags over
            out.write(b'\x00\x00')
      
            # Copy image data as long as we expect more blocks
            while imagedatablocklen > 0:
                blocksize = min([imagedatablocklen, 256])
                
                # Write image data
                out.write(rng.read(blocksize))
              
                # Decrement by blocksize
                imagedatablocklen -= blocksize
             
            # Write terminator for this frame
            out.write(b';')
          
            # Keep track of image data length
            imgdatalen += imagedatablocklen
         
        # Extension Block
        elif block