#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# header defined in RFC 4791 (https://datatracker.ietf.org/doc/html/rfc4791)
def generate(out: BinaryIO):
    out.write('GIF89a'.encode())
    out.write(struct.pack("<HH", 6, 5)) # Logical Screen Width and Height
    out.write([0xFF] * 7 + [0])         # Global Color Table
    out.write('\0\0\xF9\0\0\x2C'.encode())   # Graphics Control Extension
    out.write(struct.pack("<IH", 0, 1))     # Disposal Method, Transparent Color Index

    for _ in range(32):
        color = bytes([int(i) for i in bin(int.from_bytes(out.read(1), 'big'))[::-1]])
        out.seek(-2, 1)
        out.write(color)
        out.truncate()
        out.flush()

if __name__ == '__main__':
    from PIL import Image
    import sys

    if len(sys.argv) != 3:
        print("Usage: %s input output" % sys.argv[0], file=sys.stderr)
        exit(1)
    img = Image.open(sys.argv[1]).convert('RGB')
    imgsz = img.size
    with open(sys.argv[2], "wb") as fp:
        generate(fp)

        frame = bytearray((imgsz[0]*imgsz[1]))
        for x in range(imgsz[0]):
            for y in range(imgsz[1]):
                pixel = img.getpixel((x,y))
                if pixel == (0,0,0,255):
                    frame[(y*imgsz[0]+x)] = -1
        if len(frame) < 2**16:
            # We need at least two bytes to write (for the length field), so we
            # just skip this frame if it's too small.
            fp.write(struct.pack("<B", len(frame)))
            fp.write(frame)

        fp.close()
        exit(0)

        # This is where you'd do something clever like send the IHDR chunk first
        # then use LZW compression on the image data, but I don't feel like doing
        # all that :)
        ihdr = b'\x49\x48\x44\x52' + \
               chr(imgsz[0]>>8).encode() + \
               chr(imgsz[0]&0xff).encode() + \
               chr(imgsz[1]>>8).encode() + \
               chr(imgsz[1]&0xff).encode() + \
               b'\x08\x06'
        data = bytearray()
        for y in range(imgsz[1]):
            for x in range(imgsz[0]):
                pixel = img.getpixel((x,y))
                if pixel == (0,0,0,255):
                    data += b'\x00'
                else:
                    data += struct.pack("<BBB", *(tuple(map(ord, "#"+''.join("%02x"%c for c in pixel))))[1:])
        datalen = len(data)
        footer = b'\xAE\xCE\x1C\xE9\x00\x00'
        fp.write(ihdr +
           datalen.to_bytes(2, 'big') +
           data +
           footer)
