#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, size: int):
    # https://en.wikipedia.org/wiki/GIF#File_structure
    
    version = b'89a'                      # 6 bytes
    width = rng.getrandbits(16).to_bytes(2, 'big')   # 4 bytes
    height = rng.getrandbits(16).to_bytes(2, 'big')  # 4 bytes
    
    flags = rng.getrandbits(8).to_bytes(1, 'big')   # 1 byte
    color_table = [ rng.getrandbits(8).to_bytes(1, 'big') for i in range(0, 3 * (1 << ((flags >> 4) + 1))) ]
    image_data = [ rng.getrandbits(8).to_bytes(1, 'big') for _ in range(size - 7) ]
    
    # Write the header fields
    out.write(version)
    out.write(width)
    out.write(height)
    out.write(flags)
    out.writelines([color_table[i] for i in range(len(color_table))])
    
    # First frame
    frame_time = rng.getrandbits(16).to_bytes(2, 'big')  # 2 bytes
    first_frame_introducer = b'\x2c\x00'                # 2 bytes
    
    left = rng.getrandbits(16).to_bytes(2, 'big')       # 2 bytes
    top = rng.getrandbits(16).to_bytes(2, 'big')        # 2 bytes
    right = rng.getrandbits(16).to_bytes(2, 'big')      # 2 bytes
    bottom = rng.getrandbits(16).to_bytes(2, 'big')     # 2 bytes
    
    local_flags = rng.getrandbits(8).to_bytes(1, 'big') # 1 byte
    _, extension_code, graphic_control_extension = parse_local_flags(local_flags)
    
    # Frame delay time | Transparent colour index | Block terminator
    out.write(struct.pack('>HB', frame_time, rng.getrandbits(8).to_bytes(1, 'big')[0]))
    if extension_code != None:
        # Color table size | Extension introducer
        out.write(b'!\xF9' + struct.pack('BB', (graphic_control_extension<<1)+1, 4))
        
        # Graphic control label | Reserved | Disposal method | User input flag | Transparency flag
        out.write(struct.pack('BBBB?', 0xF9, 0x04, graphic_control_extension&3, False, graphic_control_extension>>3 == 0))
        
        # Delay time | Transparent colour index
        out.write(struct.pack('>HB', frame_time, 0))
    out.write(first_frame_introducer)
    out.write(left)
    out.write(top)
    out.write(right)
    out.write(bottom)
    out.write(local_flags)
