#!/usr/bin/env python3

import sys
import os
os.chdir(sys.path[0])

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) for up to `max_len` bytes.
def generate(rng:BinaryIO, out:BinaryIO, max_len:int):
    if max_len < 14: return # GIF header is at least 14 bytes long.
    
    def randstr():
        ret = ""
        while True:
            c = rng.read(1)
            assert len(c) == 1, "unexpected EOF"
            c = int.from_bytes(c, byteorder="big") % 256
            
            if c == ord('!'): break
            elif c in [ord('#'), ord('%')]: continue
            
            ret += chr(c)
        
        assert ret != "", "unexpected GIF terminator!"
        return ret
    
    # Write GIF87a magic.
    out.write("GIF87a".encode())
    
    # Write GIF header (14 bytes).
    width, height = rng.read(2)
    width = struct.unpack(">H", width)[0] + 1
    height = struct.unpack(">H", height)[0] + 1
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)
    
    out.write(width.to_bytes(2, byteorder="big"))   # Width
    out.write(height.to_bytes(2, byteorder="big"))  # Height
    out.write(background_color_index)               # Background color index
    out.write(pixel_aspect_ratio)                   # Pixel aspect ratio
    
    # Read initial code size value and write it as first byte of LZW data.
    lzw_data = b""
    code_size = rng.read(1)
    lzw_data = struct.pack("<B", int.from_bytes(code_size, byteorder="big"))[-1:]
    
    # Write a series of blocks until we're done or we reach the maximum length.
    current_block_length = 0
    block_type = None
    
    while current_block_length <= max_len - 7:
        # Select block type based on weighted probability (1% chance).
        i = rng.read(1)
        assert len(i) == 1, "unexpected EOF"
        if int.from_bytes(i, byteorder="little") & 0xFF >= 97:
            # Image block
            if block_type == "comment":
                # Double-image comment hack to push the frame count up by one.
                out.write(b"\x2C\x00\x00\x00\x00\x00\x00")
            
            else:
                # Regular image block.
                out.write(b"\x2C\x00\x00\x00\x00\x00\x00")
            
            # Write 2-byte x,y offset. We only use the bottom 12 bits though.
            offsets = rng.read(2)
            offset_x = struct.unpack(">H", offsets)[0] % 4096
            offset_y = struct.unpack(">H", offsets)[1] % 4096
            
            out.write((offset_x & 0x0F).to_bytes(1, byteorder="big"))
            out.write((offset_y & 0x0F).to_bytes(1, byteorder="big"))
            
            # Write width and height values. Both are set to 1 pixel unless
            # they have been initialized already.
            w, h = width, height
            if block_type == "image":
                w, h = rng.read(2)
                w = struct.unpack(">H", w)[0] + 1
                h = struct.unpack(">H", h)[0] + 1
            
            w &= 0x0F
            h &= 0x0F
            out.write(w.to_bytes(1, byteorder="big"))
            out.write(h.to_bytes(1, byteorder="big"))
            
            # Set current block length. If this is the first time through, we
            # subtract the header bytes since that was part of the previous
            # block.
            if block_type == None:
                current_block_length = w * h + 7
                block_type = "image"
            
            else:
                current_block_length = w * h
        
        else:
            # Disposal, extension, or plain comment block.
            # This is either 0x03, 0x21, or 0xFE.
            i = rng.read(1)
            assert len(i) == 1, "unexpected EOF"
            block_type = str.format("{0:02X}", int.from_bytes(i, byteorder="big"))
            
            # Disposal block. This is a 1-byte block.
            if block_type == "03":
                dispose = rng.read(1)
                out.write(dispose)
                current_block_length = 1
            
            # Extension block. Then we read two bytes with the label ID.
            # Comments get parsed like regular comments since there isn't an
            # NETSCAPE block yet. Everything else gets treated as an animation
            # extension block with a few frames.
            elif block_type == "21":
                sub_block_id = rng.read(2)
                sub_block_id = struct.unpack(">H", sub_block_id)[0]
                
                if sub_block_id == 0x0100:
                    # Animation control extension block. Parse the first
                    # four bytes; these tell us how many times we should loop
                    # the frames.
                    num_loops = rng.read(4)
                    num_loops = struct.unpack(">I", num_loops)[0]
                    out.write(num_loops.to_bytes(4, byteorder="big"))
                    current_block_length = 4
                
                elif sub_block_id == 0xFF0B:
                    # Application extension block. Read the next two bytes and
                    # look for "NETSCAPE2.0".
                    id_string = rng.read(2)
                    assert struct.unpack(">H", id_string)[0] == 0x0B
                    id_string = rng.read(12)
                    assert id_string[:11].decode() == "NETSCAPE2.0", \
                        "missing NETSCAPE2.0 application identifier"
                    
                    # Write out the rest. It looks like a two-byte size field
                    # followed by "Netscape" and then the version.
                    app_extension = rng.read(2)
                    app_extension += id_string[11:14]
                    app_extension += id_string[14:]
                    
                    out.write(app_extension)
                    current_block_length = 26
                
                elif sub_block_id == 0x01FF:
                    # Terminator block. Just ignore this.
                    current_block_length = 0
                
                else:
                    # Some other kind of extension block.
                    # TODO: Handle other kinds of extension blocks.
                    print("Unsupported extension block!")
                    exit(-1)
            
            # Regular comment block.
            elif block_type == "FE":
                # To avoid having a zero-length Block Size field when converting
                # from GIF to BMP, we add a single null character.
                comment = randstr().encode() + b'\x00'
                comment_len = len(comment).to_bytes(1, byteorder="big")
                out.write(comment_len)