#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int(2**(int((lsd[4] >> 5)))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = False

    bufposition = out.tell()
    while True:
        bufposition = out.tell()
        
        # Skip over app extension block
        extflag = ord(rng.read(1))
        if extflag == 249:
          rng.read(extflag+ord(rng.read(1)))
          
          # if we hit an end of stream marker, stop reading
          if extflag == 0xff:
              break
            
          continue
      
        # Otherwise this is an image block
        out.seek(-1, 1)
        out.write(b'\x2c')

        # Useful block sizes
        blocksize = 1<<(ord(rng.read(1))+1)
        blockwidth = 1<<(ord(rng.read(1))+1)
        
        # Write out the size information for this frame
        out.write(struct.pack('<HHBB',
                               blocksize,
                               blocksize,
                               ord(rng.read(1)),
                               ord(rng.read(1))))

        # Write out some bits of metadata about this frame
        out.write(rng.read(4))

        # Set up the expected codelen
        imgdatalen = blocksize*(blocksize+int(prevframe is not None)) - 1

        if codelen != imgdatalen:
            out.seek(1, 1)

            # If we have no codelen yet, we must find the next one
            if codelen is None:
                while True:
                    n = 1 << (imgdatalenbits-6)

                    x, y = ord(rng.read(1)), ord(rng.read(1))
                    
                    if x >= n or y >= n:
                        break

                codelen = n
                
            out.seek(1, 1)

        # Now create the actual pixel data
        data = []

        for i in range(ceil(imgdatalen/(codelen*2)-1)):
            clen = rng.read(1)*256 + ord(rng.read(1))
            data += [clen]*clen

        out.write(bytes([codelen]*ceil(imgdatalen/(codelen*2)-1)) +
                  bytes(data))

        # Save as previous frame because we might need it later
        prevframe = out.getbuffer()[bufposition:].tobytes()