#!/usr/bin/env python3

import struct
from typing import BinaryIO


def get_first_byte(stream: BinaryIO) -> int:
    """Returns the first byte in a stream and moves the stream pointer"""
    return stream.read(1)[0]


def get_byte(stream: BinaryIO, leftmost=False) -> int:
    """Return next byte in stream and move the stream pointer"""
    if not stream:
        raise EOFError("Cannot read any more from stream!")
    if leftmost:  # untested
        pos = stream.tell()
        try:
            stream.seek(-1, 2)
            byte = stream.read(1)
            stream.seek(pos, 0)
        except OSError:
            byte = None
        else:
            return byte[0]
        finally:
            stream.seek(pos, 0)
    return stream.read(1)[0]


def get_int8(stream: BinaryIO) -> int:
    """Reads an unsigned number represented by one byte"""
    b = get_byte(stream)
    return b & 0xFF


def get_uint8(stream: BinaryIO) -> int:
    """Reads a signed number represented by one byte"""
    b = get_byte(stream)
    if b > 127:
        return -(b ^ 0xff) - 1
    return b


def get_int16(stream: BinaryIO):
    """Reads a signed number represented by two bytes"""
    val = get_byte(stream, True) << 8 | get_byte(stream)
    return val & 0xFFFF


def get_int32(stream: BinaryIO):
    """Reads a signed number represented by four bytes"""
    val = (get_byte(stream, True) << 24 |
           get_byte(stream, True) << 16 |
           get_byte(stream, True) << 8 |
           get_byte(stream))
    return val & 0xFFFFFFFF


def get_uint32(stream: BinaryIO):
    """Reads an unsigned number represented by four bytes"""
    val = (get_byte(stream, True) << 24 |
           get_byte(stream, True) << 16 |
           get_byte(stream, True) << 8 |
           get_byte(stream))
    return val & 0xFFFFFFFF


def get_float(stream: BinaryIO):
    """Reads a float with single precision"""
    return struct.unpack('f', get_bytes(stream, 4))[0]


def get_double(stream: BinaryIO):
    """Reads a float with double precision"""
    return struct.unpack('d', get_bytes(stream, 8))[0]


def get_string(stream: BinaryIO, length: int) -> str:
    """Reads a string of given length"""
    result = []
    for _ in range(length):
        cntrl = get_uint8(stream)
        if cntrl < 32 or cntrl >= 127:
            break
        result.append(chr(cntrl))
    return ''.join(result).rstrip('\0')


def get_nullterminated_string(stream: BinaryIO):
    """Reads a null-terminated string of character codes up to 127"""
    result = ''
    while True:
        char = chr(get_uint8(stream)).encode()
        result += char
        if char == '\0':
            return result.decode().rstrip('\0')


def get_variable_type(stream: BinaryIO):
    """Reads variable type information from header"""
    result = get_uint8(stream)
    if result < 5:
        return "bool", 1
    elif result == 5:
        return "char", 1
    elif result == 6:
        return "uchar", 1
    elif result == 7:
        return "short", 2
    elif result == 8:
        return "ushort", 2
    elif result == 9:
        return "int", 4
    elif result == 10:
        return "uint", 4
    elif result == 11:
        return "long", 8
    elif result == 12:
        return "ulong", 8
    elif result == 13:
        return "float", 4
    elif result == 14:
        return "double", 8
    else:
        return "other", 1


def skip(stream: BinaryIO, count: int):
    """Skips over `count` bytes in the stream"""
    stream.seek(count, 1)


def skip_whitespace(stream: BinaryIO):
    """Skips white spaces in stream"""
    while True:
        byte = stream.read(1)
        if byte != b' ' and byte != b'\t' and byte != b'\n' and byte != b'\r':
            stream.seek(-1, 1)
            break


def get_bytes(stream: BinaryIO, num_bytes: int) -> bytes:
    """Reads n