
import json
from uuid import uuid4

import pytest

from jwkest.jws import JWS
from oic.oic import BEARER_AUTHENTICATION
from oic.oic import message
from oic.utils.authn.client import ClientSecretBasic
from oic.utils.authn.client import MutualTLS
from oic.utils.authn.client import PrivateKeyJWT
from oic.utils.authn.client import client_secret_post
from oic.utils.authn.client import double_encrypt
from oic.utils.authn.client import encrypt
from oic.utils.authn.client import http_b64encode
from oic.utils.authn.client import make_openidreq
from oic.utils.authn.client import mutual_tls_http_info
from oic.utils.authn.client import own_discovery
from oic.utils.authn.client import unpack_signed_jwt
from oic.utils.authn.client import utc_time_sans_frac
from oic.utils.authn.client import verify_signed_token
from oic.utils.authn.client import verify_signed_response
from oic.utils.authn.client import verify_signed_response_or_assertion
from oic.utils.authn.client import wire_format
from oic.utils.authn.client import wrap_in_jws
from oic.utils.authz import ImplicitGrantNoCORS
from oic.utils.http_util import Response

__author__ = 'rohe0002'
KEYSPEC = [
    {"type": "RSA", "use": ["sig"]},
    {"type": "EC", "crv": "P-256", "use": ["sig"]}]
BASEDIR = os.path.abspath(os.path.dirname(__file__))


class TestClientAuthn:
    @pytest.fixture(autouse=True)
    def setup_env(self):
        self._old_servers = oic.oic.SERVER_ENVIRONMENT
        oic.oic.SERVER_ENVIRONMENT = {
            "http://example.com/": {
                'provider_info': {},
                'keys': KEYSPEC}}

    def teardown_method(self):
        oic.oic.SERVER_ENVIRONMENT = self._old_servers

    def test_verify_signed_response(self, keyjar):
        issuer = "https://example.org/"
        signer = JWS(keyjar[issuer])
        req = message.AuthorizationResponse().from_dict({
            "user": "foobar"})
        resp = signer.sign_and_receive(req.to_urlencoded(), issuer)
        msg = verify_signed_response(resp, keyjar)

        assert _eq(msg["user"], "foobar")

    def test_verify_signed_response_no_key(self, keyjar):
        with pytest.raises(ValueError):
            _ = verify_signed_response("foo", keyjar)

    def test_verify_signed_response_wrong_key(self, keyjar):
        issuer = "https://example.org/another"
        signer = JWS(keyjar[issuer], alg="RS256")
        req = message.AuthorizationResponse().from_dict({
            "user": "foobar"})
        resp = signer.sign_and_receive(req.to_urlencoded())
        with pytest.raises(ValueError):
            verify_signed_response(resp, keyjar)

    def test_verify_signed_response_multiple_alg(self):
        srv_keyjar = KeyJar()
        srv_keys = KEYSPEC[:]
        for key in srv_keys:
            if key['type'] == "EC":
                key['crv'] = "P-384"
        srv_keyjar.add_symmetric("", "hemligt")
        srv_keyjar[None] = []
        for spec in srv_keys:
            kp = new_rsa_key(**spec)
            srv_keyjar[None].append(kp)

        # TODO There is a bug here somewhere that needs to be fixed
        # move the first two lines of this function above
        # srv_keyjar = KeyJar()
        # srv_keyjar.add_symmetric("", "hemligt")
        _clients = {
            'https://localhost:8087/as_none': {
                "client_id": "client1",
                "redirect_uris": ["https://localhost:8087/as_none"]},
            'https://localhost:8087/as_private': {
                "client_id": "client2",
                "redirect_uris": ["http://localhost:8087/as_private"]}
        }
        srv_conf = {
            'base_url': 'https://localhost:8087/',
            'behaviour': {
                'backchannel_logout': {
                    'logout_uri': 'https://localhost:8087/bc_logout'}}}
        svc = Server(srv_conf, srv_keyjar, CRYPTSUPPORT)
        # cl_keyjar = KeyJar()
        # cl_keyjar.add_symmetric("", "hemligt")
        # cl_keyjar.imp = []  # empty implicit key list
        cli = Authorization(svc, **_clients)
        # _clients = {}
        # for base, cnf in SERVER_INFO["client"].items():
        #     _clients[base] = Client(cnf["client_id"], None, [],
        #                             redirect_uris=[cnf["redirect_uris"][0]],
        #                             behavior={'request_object_signing_alg': ''})
        # svc = Server(SERVER_INFO['server_info'], srv_keyjar)
        # cli = AS(clients=_clients, conf=CLIENT_INFO)
        authn = ImplicitGrantNoCORS(cli, srv_keyjar, CLIENT_CONFIG)

        args = {
            'client_id': 'client2',
            'scope': ['profile']}
        session = {'state': 'id'}
        url, body, h_args = authn.create_authn_request(
            state="id", request_args=args, session_info=session)
        res = authz_cb(urlparse(url).query)
        atr = authn.parse_authz(res, session, s_param='code')
        assert set(atr.keys()) == {'access_token', 'expires_in', 'token_type',
                                   'state'}

    def test_wrap_in_jws(self):
        data = '{"user":"foobar"}'
        header = {'alg': 'HS256'}
        key = "secret"
        signed_response = wrap_in_jws(header, data, key)
        msg = unpack_signed_jwt(signed_response, key)

        assert _eq(msg["user"], "foobar")

    def test_unpack_signed_jwt(self):
        signed_response = "eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9." \
                          "eyJoZWxsbyI6ICJ3b3JsZCJ9."	\
                          "tvH8VfTee-b-_dBqKAQFgw4YLe0owXZtosIjqQn_hM"
        key = "secret"
        msg = unpack_signed_jwt(signed_response, key)

        assert _eq(msg["user"], "foobar")

    def test_make_openidreq(self):
        endpoint = "http://localhost:8087/as"
        issuer = "https://example.com/"
        scope = 'blah'
        req = make_openidreq(endpoint, issuer, scope)
        assert _eq(req["scope"], 'openid blah')
        assert scope not in req["scope"]

        req = make_openidreq(endpoint, issuer, scope, prompt='login')
        assert _eq(                data + footer)
