#!/usr/bin/env python3

import struct, subprocess, textwrap, sys
from collections import namedtuple

# Reads the MPC8560 DDR memory map
def read_map(file):
    f = open(file)

    # check if loaded PIB has correct module name
    s = ("PIB Resource" in line) for line in f.readlines()
    qlc_name = next(line[14:27] for line in f.readlines() if "PIB Resource" in line)
    qlc_ver =  next(line[50:] for line in f.readlines() if "Version:" in line)

    # check if loaded PIB is compatible with this script version
    if not (qlc_name == "DDR MAPS" and qlc_ver >= "02.01.00"):
        print("ERROR: Incompatible FLASH MAPS PIB")
        exit(-1)

    mem_types = [line[9:].split(' ')[:-1][:-1] for line in f.readlines() if "DDR Dimm Type" in line][0]
    num_channels = len([i for i in range(len(mem_types)) if mem_types[i] != ''])

    # find PIB resource starting line
    pib = [i + 1 for i in range(len(f.readlines())) if "PIB Resource" in f.readlines()[i]]

    # find byte offset to start of banking information
    first_bank = min((pib[i+1] - pib[i]) * 8 + 8 for i in range(num_channels))

    banks = []
    for channel in range(num_channels):
        f.seek(first_bank + channel*4)
        banks.append(struct.unpack("<I", f.read(4)))

    return banks

# calculate a hash from a sequence of bytes
def hash(data):
    h = 0x55555555
    for b in data:
        h += (h << 5) + b

    return h & 0xffffffff

# Find the address of the RICH HMAC
def get_hmac(banks):
    magic = 0x1e1e1e1e
    hmac_short = 0xcffeecff
    hmac_long = 0xe0e0dead
    hmac_addr = None

    for addr, data in enumerate(banks):
        dhash = hash(data)

        if addr % 0x400000 == 0 and dhash == magic:
            for index in range(min(len(data), 0x20)):
                if data[index] == hmac_short:
                    if data[-1] == hmac_long:
                        hmac_addr = addr

            break

        elif addr % 0x400000 == 0:
            continue

    if hmac_addr is None:
        print("Error: Can't find RICH HMAC!")
        exit(-1)

    return hmac_addr

# Find the magic key stored at the bottom of each page
def get_key(banks, hmac_addr):
    key = {k : v for k, v in zip(['A', 'B', 'C'], (0, 0, 2))}
    page = int(hmac_addr / 0x1000)
    bank = int(hmac_addr / 0x200000)

    # increment page until we find the last valid one
    while True:
        key['A'] -= 1
        if key['A'] < 0:
            # switch to B page
            key['B'] -= 1
            if key['B'] < 0:
                # switch to C page
                key['C'] -= 1
                if key['C'] < 0:
                    print("Error: Can't find KEY PAGE!")
                    exit(-1)

            key['A'] = 0xf

        # check if page is valid
        if (key['A'] << 12 | key['B'] << 10 | key['C'] << 8) >> 8 == page:
            break

    # convert the magic page into an actual page
    key_page = key['A'] << 12 | key['B'] << 10 | key['C'] << 8

    # build the key from the PIB
    key = [ord(banks[(key_page * 0x1000 + i)//0x200000][i%0x1000]) ^ ord('\xa') for i in range(0x10)]

    # append the HMAC
    key += banks[(hmac_addr // 0x1000)][(hmac_addr % 0x1000) + 0xe:10]

    # append the page number
    key += [key_page]

    return key

# create a single padding block used to pad short blocks
def make_padding():
    block = list(range(0xa, 0x1a))
    block[0] = 0xd
    block[4] = 0xe
    block[5] = 0x7
    block[6] = 0xe
    block[7] = 0xf
    block.extend(range(0x15, 0x20))

    return bytes(block)

# Decrypts a single block
def decrypt(ctr, block, key):
    xor_out = [0] * 16
    add_in = [0] * 16
    shift_mask = [0] * 16
    sub_mask = [0] * 16

    aes_key = [(ord(key[0]) << 24) | (ord(key[1]) << 16) | (ord(key[2]) << 8) | ord(key[3]), \
               (ord(key[4]) << 24) | (ord(key[5]) << 16) | (ord(key[6]) << 8) | ord(key[7]), \
               (ord(key[8]) << 24) | (ord(key[9]) << 16) | (ord(key[10]) << 8) | ord(key[11]), \
               (ord(key[12]) << 24) | (ord(key[13]) << 16) | (ord(key[14]) << 8) | ord(key[15]), \
               (ord(key[16]) << 24) | (ord(key[17]) << 16) | (ord(key[18]) << 8) | ord(key[19]), \
               (ord(key[20]) << 24) | (ord(key[21]) << 16) | (ord(key[22]) << 8) | ord(key[23]), \
               (ord(key[24]) << 24) | (ord(key[25]) << 16) | (ord(key[26]) << 8) | ord(key[27]), \
               (ord(key[28]) << 24) | (ord(key[29]) << 16) | (ord(key[30]) << 8) | ord(key[31]), \
               (ord(key[32]) << 24) | (ord(key[33]) << 16) | (ord(key[34]) << 8) | ord(key[35]), \
               (ord(key[36]) << 24) | (ord(key[37]) << 16) | (ord(key[38]) << 8) | ord(key[39]), \
               (ord(key[40]) << 24) | (ord(key[41]) << 16) | (ord(key[42]) << 8) | ord(key[43]), \
               (ord(key[44]) << 24) | (ord(key[45]) << 16) | (ord(key[46]) << 8) | ord(key[47]), \
               (ord(key[48]) << 24) | (ord(key[49]) << 16) | (ord(key[50]) << 8