#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan 21 09:45:07 2020

@author: mtageld
"""
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  8 16:46:30 2018

@author: mtageld
"""
import os
from collections import defaultdict, Counter
import pandas as pd
import numpy as np
from sklearn.metrics import f1_score
import matplotlib.pyplot as plt

from histomicstk.annotations_and_masks.annotation_and_mask_utils import (rle_encoding_to_full_mask,
                                                                         annotation_crops_as_list)
from histomicstk.annotations_and_masks.annotation_and_mask_creation_utils import convert_binary_mask_to_polygons

#from histomicstk.annotations_and_masks.masks_to_boxes_and_items import masks_to_tables
from histomicstk.preprocessing.color_deconvolution import color_deconvolution
from histomicstk.annotations_and_masks.annotation_and_mask_processing import load_nuclei_instance_annos
from histomicstk.annotations_and_masks.annotation_and_mask_utils \
    import get_contours_minrectangle, compute_rotated_bbox, get_image_crop_paste_loc
from histomicstk.annotations_and_masks.annotation_and_mask_io import save_annotation
from histomicstk.annotations_and_masks.annotation_and_mask_management import \
    (get_mpi_morphology_annotation_files, \
     find_matching_candidates, \
     match_candidates, apply_new_matches, \
     update_old_annotations, save_old_annotations, remove_unused_annotations)
from histomicstk.features.local_features import compute_local_histogram
from histomicstk.features.global_features import compute_global_statistics
from histomicstk.features.texture_descriptors import compute_glcm_matrix

plt.rcParams['svg.fonttype'] = 'none' #remove this line for non-svg files
SMALL_SIZE = 14
MEDIUM_SIZE = 18
BIGGER_SIZE = 20

plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
plt.rc('axes', titlesize=SMALL_SIZE)     # fontsize of the axes title
plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title
plt.rcParams["svg.hashsalt"]=0

def _check_input(partitions):
    """Ensures that partitions are a list of strings in which each string is
       either "train", "val", or "test".

       Parameters
       ----------
        partitions : list of strings
            List of partitions to consider when creating annotations.

       Returns
       -------
       None
    """

    valid_values = ["train", "val", "test"]

    if not type(partitions) == list:
        raise ValueError("Must provide list of partitions.")

    for p in partitions:
        if not p in valid_values:
            raise ValueError("Invalid value in partitions.")

def _num_valid_annotations(dataset_dir, partitions):
    """Returns the number of nuclei instances in the provided partitions.

       Parameters
       ----------
       dataset_dir : string
           Path to directory containing the full dataset
       partitions : list of strings
           List of partitions to consider when counting annotations.

       Returns
       -------
       num_valid : integer
           The total number of nuclei instances in the provided partitions.
    """

    # check input
    _check_input(partitions)

    # initialize count
    count = 0

    # loop over all of our partitions
    for p in partitions:

        # get file names for this partition
        file_names = get_mpi_morphology_annotation_files(p, dataset_dir)

        # count valid annotations from these files
        for name in file_names:

            # get annotation
            anno = load_nuclei_instance_annos(name)[0]

            # add to count
            count += len([a for a in anno["annotations"] if len(a["bbox"])==4])

    return count

def _count_total_annotations(dataset_dir, partitions):
    """Counts the total number of nuclei instances in the provided partitions
       without filtering out instances with instances without bounding boxes,
       polygons, RLEs, etc. This function should only be used to verify that
       the counts reported by ``_num_valid_annotations`` are correct. For
       example, it is possible to have an image where all of the nuclei are
       somehow invalid but there is still some noise in the annotation file.

       In that case, we want to know how many total instances were annotated.
       However, this function will exclude those instances by counting only
       those instances that pass all checks defined in the ``_filter_anno``
       function.

       Parameters
       ----------
       dataset_dir : string
           Path to directory containing the full dataset
       partitions : list of strings
           List of partitions to consider when counting annotations.

       Returns
       -------
       num_total : integer
           The total number of nuclei instances in the provided partitions,
           regardless of whether they passed validation.
    """

    # check input
    _check_input(partitions)

    # initialize count
    count = 0

    # loop over all of our partitions
    for p in partitions:

        # get file names for this partition
        file_names = get_mpi_morphology_annotation_files(p, dataset_dir)

        # count valid annotations from these files
        for name in file_names:

            # get annotation
            anno = load_nuclei_instance_annos(name)[0]

            # add to count
            count += len(anno["annotations"])

    return count

def _compute_stats(dataset_dir, partitions, include_background=False):
    """Compute statistics about the nuclei annotations.

       ..note::
           These statistics may be slightly different than what is reported in
           the original paper because we are computing them differently.
           Specifically, instead of using ``pillow`` with its ``PILImage.mode``
           property, we use ``matplotlib``, which has slightly different
           behavior. See the discussion here: https://github.com/python-pillow/Pillow/issues/1004

       ..note::
           We also restrict ourselves to considering grayscale images in the
           original paper (grayscale *or* binary). Therefore, to make things
           simple, we ignore any annotations associated with images that do not
           have one of the following modes:
               - L (grayscale)
               - P (segmentation)
               - 1 (binary)

       Parameters
       ----------
       dataset_dir : string
           Path to directory containing the full dataset
       partitions : list of strings
           List of partitions to consider when counting annotations.
       include_background : bool
           Whether to include the background in the computation. By default,
           the background is excluded.

       Returns
       -------
       stats : dict
           Dictionary containing information about the nuclei annotations.
    """

    # check input
    _check_input(partitions)

    # initialize stats
    stats = {
        'count': {},
        'area': {},
        'perimeter': {};
        'major_axis': {};
        'minor_axis': {};
        'solidity': {};
        'compactness': {};
        'ratio': {};
        'extent': {};
        'equivalent_