#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Read the width and height of the image as 16-bit signed integers in little endian order
    w = struct.unpack('<h', rng.read(2))[0] + 1
    h = struct.unpack('<h', rng.read(2))[0] + 1

    print("Applying width %d and height %d." % (w, h))

    # Write the size of the image
    out.write(struct.pack("<HH", w, h))

    # Add the color table (one entry for each possible color, i.e., 256 colors)
    bytes_per_color = 3
    num_entries = 2**bytes_per_color
    num_colors = 2**bytes_per_color

    palette_size = num_entries * bytes_per_color
    out.write((palette_size).to_bytes(1, 'little'))
    for _ in range(num_colors):
        out.write(rng.read(bytes_per_color))

    # We don't have any transparency, so set the transparent color index to -1
    out.write((-1).to_bytes(1, 'little'))

    # Set the background color to be black
    out.write((0).to_bytes(1, 'little'))

    # Add an application extension
    block_length = 11
    app_ext_block_id = 0xFF
    app_identification = "abc"
    appauth_code = ""
    data = app_identification + appauth_code + "".join([chr(rng.read(1)[0]) for _ in range(block_length - len(app_identification) - len(appauth_code))])
    ext_block_length = len(data)
    out.write(b'\x21')
    out.write(b'\xff')
    out.write(struct.pack("<H", ext_block_length))
    out.write(data.encode())

    # Frame Loop Count Block
    out.write(b'\x21')
    out.write(b'\xF9')
    out.write(b'\x04\x04')
    out.write(b'\x01\x00')
    out.write(b'\x00')

    # Graphics Control Extension
    gce_block_list = [ord(rng.read(1))]
    while gce_block_list[-1] & 0b10000000 != 0:
        gce_block_list.append(ord(rng.read(1)))

    gce_block = struct.pack("<B", sum(gce_block_list))

    out.write(b'\x21')
    out.write(b'\xf9')
    out.write(b'\x04')
    out.write(gce_block)

    # Image Descriptor
    descriptor = (0).to_bytes(1, 'little')
    descriptor += (0).to_bytes(2, 'little')
    descriptor += (w).to_bytes(2, 'little')
    descriptor += (h).to_bytes(2, 'little')
    descriptor += (7).to_bytes(1, 'little')
    out.write(descriptor)

    # LZW Minimum Code Size
    lzw_min_code_size = ord(rng.read(1))
    out.write(lzw_min_code_size.to_bytes(1, 'little'))

    # Compressed Data Stream
    compressed_stream = rng.read()
    out.write(compressed_stream)

    # Trailer
    out.write(b';