#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int(2**(int((lsd[4] >> 5)))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = False
    
    while True:
        # Create new frame        
        out.write(b'\x2c')
        
        # Create image descriptor
        id = rng.read(9)
        out.write(id)

        # Get number of bytes to read based on bits in ID
        imgdatalenbits = int(id[8])&0x7
        imgdatalen = int(id[4])*pow(2,imgdatalenbits+1)+imgdatalenbits+1
        codelen = pow(2,(id[6]&7))+1

        # Check whether this is an animation or not
        bufposition = out.tell()

        # Write background index as first byte
        bgindex = rng.read(1)
        out.write(bgindex)

        # Set disposal method
        dispose = (ord(bgindex)>>3)&0x7;

        # Set user input flag
        uiflag = bool(ord(id[7])&0x40)

        # Set transparent color flag
        tcflag = bool(ord(id[7])&0x10)
        
        # Operation codes - set to true when found
        has_transparency = False      # full transparency
        has_transcolor    = False      # partial transparency
        has_userinput     = False      # need 1 more byte afterwards
        has_disposal      = False      # needs one extra block afterwards
        has_interlaceflag = False      # just call us recursively

        # See what options there are
        flags = ord(id[7]) & 0x3f

        # UNDEFINED = 0   # bit 7 clear
        # INTERLACEFLAG = 1 # bit 5 set
        if bool(flags&0x20):
            has_interlaceflag=True

        # DISPOSALMETHOD = 2 # bits 2-4
        if bool(flags&0x18):
            has_disposal=True        
            
        # USERINPUT = 3 # bit 1 set
        if bool(flags&0x02):
            has_userinput=True

        # TRANSPARENCY = 4 # bit 1 clear
        if bool(bool(flags&(~0x02)) == bool(uiflag)):
            has_transparency=True

        # TRANSPARENTCOLOR = 5 # bit 0 set
        if ((not has_transparency or tcflag) 
            and bool(bool(flags&(~0x01))!= bool(tcflag))):
            
            has_transcolor=True


        # Some errors that could happen... We'll deal with them
        # after all the other fields are done writing for this frame
        if imgdatalen % codelen!= 0:
            raise ValueError("Image data length cannot be devided by "
                             "the code length!")

        if has_transcolor and has_transparency:
            raise ValueError("Transparency and transparent colors both"
                            "found at the same time")

        # Start writing optional blocks immediately after the frame
        # is written
        futurestart = out.tell()
        futureend = futurestart + \
                    ceil(((has_transcolor and tcflag)
                        or (has_userinput and len(prevframe))) 
                        / codelen)-1

        # Write local color table if needed
        if has_transcolor or has_transparency:
            lcgtlen = 2**((id[8]>>3)&0x7)

            # If we don't have a local color table yet
            # we need to start writing it over where
            # we wrote the background index last
            if not hasattr(lcgtlen,'__len__'):
                xsize = int(id[4])
                ysize = abs(int(id[5]))

                if ysize > xsize:
                    xsize,ysize = ysize, xsize
                
                lcgtlen=(pow(2,floor(log2(max(xsize,ysize)*3)))))*3

            # Write the local color table
            lcgtdir = rng.read(1)
            out.write(lcgtdir)
            locclrtab = rng.read(lcgtlen)
            out.write(locclrtab)

        """ 
        38th byte is the NETSCAPE APPLICATION EXTENSION which contains 
        information about the application used to make the gif. This 
        field can be ignored but most applications will ignore unknown
        blocks so I'm going to generate some garbage here and write over
        the old application extension block
        """
        if hasattr(lcgtlen,'__len__'):
            netscapeversion = b"\x0bNETSCAPE2.0\x03\x01"
            appext = b"\xff\x0b" + \
                     struct.pack('>H', len(netscapeversion)) + \
                     netscapeversion
                        
            out.seek(-2, 1)
            out.write(appext)

        
        # Write graphic control block if needed
        if has_disposal or has_transparency or has_userinput:            
            # Write graphic control label
            out.write(b'\xF9')
        
            # Write block size
            out.write(b'\x04')
        
            # Write delay time
            dtime = rng.read(2)
            out.write(dtime)
        
            # Write transparent color index
            if has_transcolor or has_transparency:
                colidx = rng.read(1)
                out.write(colidx)
        
            else:
                out.write(b'\x00')
        
            # Write block terminator
            out.write(b'\x00')

            # There's always going to be space left for these two blocks
            futureend -= 2

        # Go back and write the correct block sizes
        out.seek(futurestart, 0)
        futuresize = futureend - futurestart

        out.write(bytes([(futuresize>>(i*8))&0xFF for i in reversed(range(4))]
                       ))

        # Go back and write the correct block terminators
        out.seek(futureend, 0)
        out.write(b'\x00')

        # Go back and write next frame marker (if applicable)
        out.seek(out.tell()-1, 0)

        # Break if we're not making any more frames
        if imgdatalen==0:
            break

        elif prevframe:
            # Find maximum occurence of current pixel
            maxcount = max(*map(lambda tup:
                              prevframe[:tup[0]].count(tup[1]),
                              enumerate(prevframe[-codelen:])))

            # Find next occurance of current pixel
            for n in reversed(range(1,maxcount)):
                codelen = n
                
            out.seek(1, 1)

        # Now create the actual pixel data
        data = []

        for i in range(ceil(imgdatalen/(codelen*2)-1)):
            clen = rng.read(1)*256 + ord(rng.read(1))
            data += [clen]*clen

        out.write(bytes([codelen]*ceil(imgdatalen/(codelen*2)-1)) +
                  bytes(data))

        # Save as previous frame because we might need it later
        prevframe = out.getbuffer()[bufposition:].tobytes()