#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big') & 0xfff

    print('header',header,'datalen',datalen)

    # Write the header to our generated gif
    for char in header:
        out.write(char.to_bytes(1, 'little'))
    
    # This is just a magic number
    out.write((b'\x47\x49\x46\x38').to_bytes(4, 'little'))

    # There are more fields we could do here,
    # but this is enough for now.
    width = ((int.from_bytes(
                rng.read(2), byteorder='big') % 50) + 1).to_bytes(
                    2, 'little')
    height = (int.from_bytes(
                    rng.read(2), byteorder='big') % 50) + 1).to_bytes(
                        2, 'little')
    out.write(width+height)

    # Add an infinite color table of length 768 bytes
    # TODO: randomize it somehow?
    add_color_table(out)

    # LZW minimum code size
    lzwminsize = struct.pack('<B', int.from_bytes(
                        rng.read(1), byteorder='big') & 0xff)
    out.write(lzwminsize)

    # Now comes the image data! In theory there can be any amount of them, and
    # each one can have any size. However, with this program we'll only use one.
    # It will consist of all pixels in sequence, encoded as LZW (because I'm too lazy
    # to implement Huffman encoding)
    imgdata = encode_pixels(generate_pixel_sequence())
    imgdata = len(imgdata).to_bytes(2, 'little') + imgdata
    frame = b',' + (b'!' * 11) + b';' + imgdata

    # The image descriptor has some interesting fields such as disposal method
    # and delay time. For now they're set to 0 so that browser doesn't mess up the picture.
    desc = b''

    for char in frame + desc:
        out.write(char.to_bytes(1, 'little'))


# Writes the global color palette of all colors that we know about.
# This was chosen by picking "the" most common color from the color palettes
# of various popular websites, plus black and white.
def add_color_table(file: BinaryIO):
    palette = [
            # Color 0 - Black
            0, 0, 0,

            # Color 1 - White
            255, 255, 255,

            # Color 2 - Transparent background
            255, 0, 255,

            # Color 3 - Red
            255, 0, 0,

            # Color 4 - Green
            0, 255, 0,

            # Color 5 - Blue
            0, 0, 255]

    # Even though we specify this as a PNG format animation, browsers seem to
    # like to make their own version of the palette. We'll just write what's on
    # our minds, not necessarily what's actually used in reality
    localcolortable = b'\x00\x00\x00\xFF\xFF\xFF\x00\x00\xFF\xFF\x00\x00\x00\x00' \
                      b'\x00\xFF\x00\x00\xFF\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                      b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                      b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                      b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                      b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                      b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'

    colortable = b'\x00\x00\x00\xFF\xFF\xFF\x00\x00\x00\x00\x00\x00' + localcolortable
    colortablelen = struct.pack(
            '<B', len(colortable))[:1]
    file.write(colortablelen)
    file.write(colortable)


# An iterator which generates a sequence of distinct tuples of integer pixel values.
# The sequence will contain exactly 2**24 distinct integers; if you need more or less than
# that then either apply some filter function before calling list() or modify this function.
def generate_pixel_sequence():
    webcolors = []
    
    # Read the list of HTML color names from a text file
    with open('./webcolors.txt', mode='r') as f:
        lines = f.readlines()
        for line in lines:
            # Just read the first value from every line (we don't need anything else)
            val = int(line[1:11], base=16)
            webcolors.append(val)

    for col in webcolors:
        yield col
        
    # Generate a few additional fake colors
    for i in range(1, 250):
        r = int(((i*21)%256)/2)
        g = int(((i*17)%256)/4)
        b = int(((i*13)%256)*2)

        yield r << 16 | g << 8 | b



# Turns the given sequence of integer RGB triplets into a binary string suitable for embedding
# inside an animated GIF file.
def encode_pixels(seq):
    pixstr = b''

    for col in seq:
        pixstr += (struct.pack('<I', col)[-3:])

    return pixstr