#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    lsd_fields = [
        'width', 'height',
        'flags', 'bg_color',
        'aspect_ratio', '\0\0']

    for i in range(6):
        # Convert each 1-byte field from binary to an integer
        if len(lsd[i]) == 1:
            lsd_fields[i] = int.from_bytes(lsd[i], byteorder='little')

        else:
            raise ValueError('Invalid LSD value ({})'.format(repr(lsd)))
    
    width      = lsd_fields[0]
    height     = lsd_fields[1]
    flags      = lsd_fields[2]
    bg_color   = lsd_fields[3]
    aspect_rat = lsd_fields[4]

    # Write LSD fields to output
    for f in lsd_fields[:-1]:
        out.write(struct.pack('<H', f))

    # Get global color table
    gct = b''
    num_gct_entries = ((flags & 0x07) + 1) ** 2
    if flags & 0x80!= 0:
        # Global color table present
        gct += rng.read(num_gct_entries * 3)

        if num_gct_entries > 0xFF:
            raise Exception('Too many GCT entries')
        
        out.write(struct.pack('B', num_gct_entries))
        out.write(gct[:num_gct_entries*3])

    else:
        # No global color table; write transparency flag
        out.write(b'\x00')

    # Get the rest of the LSD
    while len(lsd) < 7 + num_gct_entries * 3:
        lsd += rng.read((len(lsd)-7)/3 + 1)

    lsd = lsd[7+num_gct_entries*3:]

    other_lsd_fields = ['bg_color', 'pixel_aspect_ratio', '\0']

    if len(lsd) >= 2:
        # Parse pixel aspect ratio
        pixel_aspect_ratio = lsd[0] * 256 + lsd[1]
        other_lsd_fields[1] = pixel_aspect_ratio

    if len(lsd) >= 4:
        # Parse background color index
        bg_color_index = lsd[0] | (lsd[1] << 8)
        other_lsd_fields[0] = bg_color_index

    # Write remaining LSD fields to output
    for f in other_lsd_fields:
        out.write(struct.pack('<H', f))

    # Add frames to output
    gce_label = b'\x21\xF9\x04\x05'
    gce_block_size = b'\x00\x0F\x00'
    gce_flags = b'\x00\x00' # no transparancy, user input not allowed
    gce_delay_time = b'\x00\x00' # default delay time
    gce_transparent_idx = b'\x00\x00' # none transparent by default
    gce_terminator = b'\x00'

    while True:
        # New frame label
        out.write(b'\x21\xF9\x04\x05')
        out.write(gce_flags)
        out.write(gce_delay_time)
        out.write(gce_transparent_idx)
        out.write(gce_terminator)

        # Block size
        out.write(gce_block_size)

        # Frame data
        lzw_min_code_size = rng.read(1)[0]
        out.write(struct.pack('B', lzw_min_code_size))

        lzw_max_bits = 12 + lzw_min_code_size
        min_length = 2 ** lzw_min_code_size
        max_length = 2 ** lzw_max_bits - 1
        code_table = {}
        next_code = min_length
        prev_code = None
        block = rng.read(next_code).tobytes()

        while next_code <= max_length or len(block) > 0:
            try:
                cur_code = ord(block[0])*256 + ord(block[1])

            except IndexError as e :
                # End of block reached with incomplete code
                break
            
            code_size = find_code_size(cur_code)
            code = cur_code >> (lzw_max_bits - code_size)

            try:
                next_char = decompress_code(code, code_table)

                if prev_code is not None:
                    code_table[prev_code].append(next_char)
                
                else:
                    code_table[prev_code] = []
                    code_table[prev_code].append(next_char)
                    
                prev_code = cur_code
                block = block[2:]

            except KeyError as e:
                # Key error means we need more bytes!
                extra_bytes = 2**code_size - len(block)
                new_bytes = rng.read(extra_bytes).tobytes()
                block += new_bytes
            
            except AssertionError as e:
                print("Error encountered:")
                print(("Code Size:", code_size, "Cur Code:", cur_code))
                raise e
            
        clear_code = 1 << lzw_min_code_size
        end_of_information_code = clear_code + 1

        compressed_codes = {
            0: chr(clear_code),
            255: chr(end_of_information_code)}

        for k, v in code_table.items():
            assert(k is not None)
            compressed_codes[k] = ''.join(v + [chr(k)])

        encoded_compressed_data = compress_data(
            compressed_codes, lzw_min_code_size)

        out.write(encoded_compressed_data)

        # Trailer
        trailer = 0
        while trailer == 0:
            trailer = rng.read(1)[0]
        
        if trailer == 0x3B:
            # Exit loop
            break

        elif trailer == 0x2C:
            # Reset state variables
            pass

        elif trailer == 0x21:
            # Extension blocks
            extension_type = rng.read(1)[0]

            if extension_type == 0xF9:
                # Graphic control extensions
                delay_time = rng.read(2)
                transparent_color_flag = rng.read(1)[0]
                num_additional_colors = 0
                
                if transparent_color_flag!= 0:
                    first_additional_color_index |= rng.read(1)[0]

                gce_flags = \
                    transparent_color_flag | \
                    0x04 if animation_control else 0x00
                
                gce_delay_time = delay_time
                gce_transparent_idx = transparent_color_flag
            
    # Write GIF trailer (0x3B)
    out.write(b'\x3B')


def find_code_size(value):
    """Finds the smallest code size that can represent this code"""
    # We only have up to 6 bits, so we don't need to go further 
    if value < 0x40:
        return 1

    if value < 0x4000:
        return 2

    if value < 0x400000:
        return 3

    if value < 0x40000000:
        return 4
        
    return 5
    

def decompress_code(code, code_table):
    """Decompresses a given LZW code to its corresponding character sequence"""
    try:
        return code_table[code][0]

    except KeyError:
        raise KeyError('{} was not found in dictionary.'.format(code))

    except TypeError:
        raise TypeError('Dictionary must be a dictionary.')
        
    
def compress_data(code_dict, initial_size=None):
    """Compresses a string according to the given code dictionary."""
    if initial_size is None:
        # Find the smallest key in the dictionary...
        initial_size = sorted(list(set([int(k)**2 for k \
            in list(code_dict.keys())])))[0]

    initial_size = str(initial_size)

    result = ''
    last_key = None
    buffer = ''
    bitstring = ''

    while True:
        if buffer!= '':
            if buffer[-1] in code_dict[last_key]:
                # Partial match
                buffer = buffer + bitstring
                bitstring = ''
                
            try:
                result += code_dict[buffer][0]
                last_key = int(buffer)**2
                buffer = ''

            except KeyError:
                # Buffer has been completed but doesn't exist in dict,
                # so make it the last partial match instead
                pass
                
        if len(bitstring) >= initial_size:
            result += code_dict[int(bitstring)]
            break

        try:
            bitstring += bin(ord(input()))[2:].z