#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int(2**(int((lsd[4] >> 5)))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = False
    
    while True:
        # Create new frame        
        out.write(b'\x2c')
        
        # Create image descriptor
        id = rng.read(9)
        out.write(id)

        # Get number of bytes to read based on bits in ID
        imgdatalenbits = int(id[8])&0x7
        imgdatalen = int(id[4])*pow(2,imgdatalenbits+1)+imgdatalenbits+1
        codelen = pow(2,(id[6]&7))+1

        # Get dispose method, if any
        dispose = -1 if (id[8]&8)==0 else id[8]&7

        if len(prevframe)!= 0 and \
           imgdatalen < len(prevframe) and \
           (dispose==-1 or dispose<=2):
            out.write(prevframe[:imgdatalen])

        elif dispose == 2:
            out.write(prevframe[-imgdatalen:])

        else:
            out.write(b'\0'*imgdatalen)
            
        # If this is not the first frame, make sure we're at the end of GIF
        bufposition = out.tell()
    
        if imgdatalen!= 0:
          if out.tell()!= len(header)+7+(gctlen if gctlen>0 else 0):
              raise Exception("Frame too big")

          if out.tell()+1!= bufposition+imgdatalen+2:
              raise Exception("Frame size does not match what's written.")
          
        # Set codesize for the frame if necessary
        curpos = out.tell()

        if curpos%2==0 and ((curpos//2)%12==0):

            if curpos%5==0:
                n = min(4092,min(max(min(int(((curpos//2)//12)),
                                        max(int((((curpos//2)//12)//18))),
                                        1),
                                  7),
                            9))*8
            
            elif curpos % 12!= 0:
                n = min(1025,min(max(min(int(((curpos//2)//12)),
                                        max(int((((curpos//2)//12)//12))),
                                        1),
                                   7),
                              8))*8

            if codelen!= n:
                
                # Write local color table entry
                out.seek(-1, 1)
                out.write(struct.pack('B',n))
            
                # Write block terminator
                if curpos % 4 == 0:
                    out.write(b'\0\0')
                    out.write(b"\xff\x0bNETSCAPE2.0")
                    out.write(b"!\xf9\x04")
                    
                    if rng.read(1)[0] & 1:
                        n = (rng.read(1)[0]+1) * 65536
                        
                    else:
                        n = rng.read(1)[0] * 256 + rng.read(1)[0]

                    if x >= n or y >= n:
                        break

                codelen = n
                
            out.seek(1, 1)

        # Now create the actual pixel data
        data = []

        for i in range(ceil(imgdatalen/(codelen*2)-1)):
            clen = rng.read(1)*256 + ord(rng.read(1))
            data += [clen]*clen

        out.write(bytes([codelen]*ceil(imgdatalen/(codelen*2)-1)) +
                  bytes(data))

        # Save as previous frame because we might need it later
        prevframe = out.getbuffer()[bufposition:].tobytes()