#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Create the logical screen descriptor (LSD), read from /dev/urandom
    lsd = bytes()
    for i in range(2):
        lsd += struct.pack('B', int.from_bytes(rng.read(1), byteorder='big'))
    
    # Write LSD to output
    out.write(lsd) 

    # Read the global color table
    gct = bytes()
    for i in range(int.from_bytes(lsd[0:1], byteorder='little')):
        color = bytes()
        for j in range(3):
            color += struct.pack('B', int.from_bytes(rng.read(1), byteorder='big'))
        
        gct += color

    # Write GCT to output
    out.write(gct)

    # Set up the local color table to be empty
    lct = b''

    # Setup the block terminator flag
    subblock_flag = False
    end_of_data_flag = True

    while not end_of_data_flag:

        # Get one byte from /dev/urandom - this will determine what we do next
        instruction = int.from_bytes(rng.read(1), byteorder='big')

         # Check if this is an extension block or not
        if instruction >= 0x21 and instruction <= 0x2F:

            # Skip reading until we get our 0x00 terminator
            count = 1
            temp_instruction = int.from_bytes(rng.read(count), byteorder='big')
            while temp_instruction!= 0x00:
                temp_instruction = int.from_bytes(
                    rng.read(count), 
                    byteorder='big'
                )

            # Add it to the buffer and clear the subblock flag
            lct += rng.read(count+1) 
            subblock_flag = False

        elif instruction == 0x2C:
            
            # This is a graphic control extension
            # Skip the first two fields, then grab the rest of them
            gc = bytes() + \
                  rng.read(16)[2:]

            # Append the graphics control data to the LCT
            lct += gc

            # Clear the subblock flag as well
            subblock_flag = False


        elif instruction == 0x3B:
    
            # We've reached the end-of-file marker; break the loop!
            end_of_data_flag = False
        
            
        else:

            # If this isn't any sort of block we care about, just keep going
            continue

        # Skip past the 0x00 terminator that ends each block
        lct += rng.read(1) 

        # Now we know all the data we need for that block has been parsed

        # Write the length of the current block 
        out.write(struct.pack('<H', len(lct)))

        # Write the data itself
        out.write(lct)

        # Reset the contents of the data buffer
        lct = ''

        # Clear the subblock flag again
        subblock_flag = False

