#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    width = int.from_bytes(
            rng.read(2), byteorder='little') & 0x7fff
    height = int.from_bytes(
            rng.read(2), byteorder='little') & 0x7fff

    # We want to make sure that we don't accidentally create an image too big!
    while True:
       if width < 1 or height < 1:
           continue

       break

    print('width', width,
          'height', height)

    datalen = width * height // 4

    first_byte = 0b111 << 5 | 0b11111 << 2
    second_byte = 0b111 << 5 | 0b11111 << 2

    third_byte = (datalen >> 11) & 0xFF

    fourth_byte = (datalen >> 19) & 0xFF

    # Write pre-frame bytes  

    out.write(header +
              width.to_bytes(2, 'little') + 
              height.to_bytes(2, 'little') + 
              bytes([first_byte]) +
              bytes([second_byte]))


    # Write color table
    
    colortable = [
            0xDE, 0xBF, 0xBD, 
            0x6D, 0x6B, 0xC6, 
            0xDB, 0xE4, 0xF2]

    for i in range(len(colortable)):

        val = colortable[i]

        write_rgb(
                out,
                0, 
                i, 
                0, 
                1, 
                val, 
                False, 
                False)


    # Write LZW min code size and data length

    lzwminsize = 2
    lenbits = ((datalen % 7) + 1)
    lzwgrowthfactor = lenbits - 1

    lencode = 3 + lzwgrowthfactor
    endofinfo = 28

    out.write((lzwminsize).to_bytes(1,"little"))
    out.write((lenbits).to_bytes(1,"little"))

    # Write the rest of the frame data

    for _ in range(datalen):
        
        v = rng.read(1)[0]

        if v == ord('\n'):

            raise Exception('newline found')
    
        elif v == ord(','):

            raise Exception('comma found')

        elif v >= 32 and v <= 126:
            
            write_rgb(
                    out, 
                    width, 
                    0, 
                    datalen, 
                    lzwgrowthfactor, 
                    ord(chr(v)), 
                    False, 
                    False)


        else:
        
            write_rgb(
                    out,
                    0, 
                    0, 
                    datalen, 
                    lzwgrowthfactor, 
                    v, 
                    True, 
                    False)

    return


def write_rgb(
        out: BinaryIO,
        width: int, 
        left: int, 
        top: int, 
        grow: int, 
        val: int, 
        is_LZW: bool, 
        has_come_back: bool):

    right = left + 1
    bottom = top + 1

    mask = 0
    shift = 7

    start = top * width + left

    end = (bottom * width + right)

    while start < end:

        mask += (1 << shift)

        if not has_come_back:

            shift -= 1

            if shift < 0:

                shift = 7

                bit = (val & mask) > 0

                val >>= shift

                if is_LZW and bit!= has_come_back:

                    # Write LZW reset code
                    
                    write_rgb(
                            out, 
                            width, 
                            left, 
                            top, 
                            grow,
                            256, 
                            True, 
                            bit)

                else:

                    out.write(((bit > 0)).to_bytes(1,'little'))

                    if not bit:
                        mask ^= 0xFF

        has_come_back = bit



        out.write(((val).to_bytes(
            min(end - start, 1), "little")))
