#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big') & 0xfff

    print('header',header,'datalen',datalen)

    # Frame header
    out.write(header)
    # Image dimensions
    width = (int.from_bytes(
            rng.read(2), byteorder='big')) & 0xffff
    height = (int.from_bytes(
            rng.read(2), byteorder='big')) & 0xffff
    out.write((width).to_bytes(2, 'big'))
    out.write((height).to_bytes(2, 'big'))
    
    # Global color table exists? Yep!
    global_color_table_exists = bool(((
        int.from_bytes(rng.read(1),byteorder='big')) >> 7) & 1)
    print('global_color_table_exists?',global_color_table_exists)
    if not global_color_table_exists:
        return
    
    # Color resolution of 256 colors. Makes things easier.
    out.write(b'\x87')
    out.write(b'\xff\xff\xff\xff')
    out.write(b'\x00')

    size_of_global_color_table = ((
        int.from_bytes(rng.read(1), byteorder='big')))
    print('size_of_global_color_table',size_of_global_color_table)
    for _ in range(size_of_global_color_table):
        r = int.from_bytes(rng.read(1),'big') % 2**8
        g = int.from_bytes(rng.read(1),'big') % 2**8
        b = int.from_bytes(rng.read(1),'big') % 2**8
        
        out.write((r).to_bytes(1, 'big'))
        out.write((g).to_bytes(1, 'big'))
        out.write((b).to_bytes(1, 'big'))

    # As above; this is always true for RNG data we've seen so far
    local_color_table_exists = True
    while local_color_table_exists:

        local_color_table_exists = bool(((
            int.from_bytes(
                rng.read(1),
                byteorder='big')) >> 4) & 1)

        print('local color table exists?',local_color_table_exists)
        if not local_color_table_exists: continue
        
        num_colors_in_table = int.from_bytes(rng.read(1),'big')
        print(num_colors_in_table,'colors')

        for _ in range(num_colors_in_table+1):
            r = int.from_bytes(rng.read(1),'big') % 2**8
            g = int.from_bytes(rng.read(1),'big') % 2**8
            b = int.from_bytes(rng.read(1),'big') % 2**8

            out.write((r).to_bytes(1,'big'))
            out.write((g).to_bytes(1,'big'))
            out.write((b).to_bytes(1,'big'))
        
        
    # Set background color index to first one in global color table
    bg_color_index = 0
    out.write(bg_color_index.to_bytes(1, 'big'))

    print('background color index:',bg_color_index)

    
    # Pixel aspect ratio - fixed at square pixels (aspect ratio=1).
    pixel_aspect_ratio = 0
    out.write(pixel_aspect_ratio.to_bytes(1, 'big'))


    # Now comes the frame control block (FCB)
    # NOTE: The code below assumes that there's only *one*
    # interlaced image and no LZW compression. 
    # Also, "graphics control extension" field has been left off, because it doesn't seem necessary...
    disposal_method = [
            0, # unspecified
            1, # do nothing 
            2, # restore to background color
            3] # restore to previous
    user_input_reqd = False
    transparent_color_flag = False
    delay_time = 100
    trans_index = None
    graphics_control_extension_block = bytes() + \
        (disposal_method[0]).to_bytes(1, 'big') + \
        (user_input_reqd).to_bytes(1, 'big') + \
        (transparent_color_flag).to_bytes(1, 'big') + \
        (delay_time).to_bytes(2, 'big') + \
        ((trans_index or 0)).to_bytes(1, 'big') 

    out.write(graphics_control_extension_block)
    out.write(b'\x00') # Block terminator

    # Image descriptor
    left = 0
    top = 0
    width = (int.from_bytes(
            rng.read(2),
            byteorder='big')) & 0xffff
    height = (int.from_bytes(
            rng.read(2),
            byteorder='big')) & 0xffff
    packed_fields = b'\x00' # fields unused
    out.write(left.to_bytes(2,'big') + 
        top.to_bytes(2,'big') + 
        width.to_bytes(2,'big') + 
        height.to_bytes(2,'big') + 
        packed_fields)
    
    # Local color table follows, i.e., palette for each row of pixels
    # We don't use these. Instead, we'll just re-use the global ones.
    size_of_local_color_table = 0
    out.write(size_of_local_color_table.to_bytes(1,'big'))

    # LZW minimum code size
    lzw_min_code_size = 8
    out.write([lzw_min_code_size])

    # Code length codes
    max_code_length = min(max(lzw_min_code_size+1, 2), 12)
    for i in range(max_code_length + 1):
        code_len = min(i+1, 12)
        out.write([(1 << code_len)-1].to_bytes(1,'big'))

    # Actual compressed image data

    # First, write LZW dictionary values back as raw data
    # since they're never actually used by decoders
    lzw_dictionary_values = [chr(x) for x in range(256)]
    for val in lzw_dictionary_values[:-1]:
        out.write(val.encode())
    dict_end = ''.join(lzw_dictionary_values)[-1:]
    out.write(dict_end.encode())

    # Next, construct the LZW string we'll compress
    lzw_string = ""

    # TODO hacky way to get our hands on enough bytes to be able
    # to generate an uncompressed image. This would be unnecessary if
    # we had a tool that could convert images from GIF into BMP format.
    # Unfortunately, I'm too lazy to find such a thing right now :(
    out_data = []
    expected_bytes = (width * height * 3) // 2
    while len(out_data) < expected_bytes:
        out_data += list(rng.read(expected_bytes))

    # TODO same hack here -- strip the last two bits of every byte
    # because of what appears to be a bug in the Go decoder. Like with
    # the encoder, you shouldn't strip them away. In fact, you should
    # be using them as part of your bitstream.
    out_data = [(x & ~(0b11)) for x in out_data]

    for f in out_data:
        if isinstance(f, int):
            v = chr(f)
        else:
            raise TypeError("Unexpected type")
            
        lzw_string += v

    # Write initial code (first 8 bits)
    current_code = ''
    lzw_string = current_code + lzw_string
    current_code = lzw_string[:8]
    lzw_string = lzw_string[8:]

    out.write(current_code.encode())

    # Maximum 12-bit value possible given # available bits, minus 1
    max_value = (1<<lzw_min_code_size)-1

    # Table of lengths of all strings encountered so far
    code_lengths = {}

    # Print some details about how the algorithm works
    print('Image dimensions ({},{}), data length {}, LZW minimum {}'.format(
        width, height, len(out_data), lzw_min_code_size))
    print('Maximum value {}, codes {}, dictionary size {}'.format(
        hex(max_value), hex(1<<(lzw_min_code_size)), len(lzw_dictionary_values)))

    def print_state():
        print('Current code {!r} with length {}'.format(
            current_code, len(current_code)))
        print('# of distinct codes seen:', len(code_lengths))
        print('# of entries in dictionary:', len(lzw_dictionary_values))
        print('Dictionary:', ", ".join(lzw_dictionary_values))
        print('Code lengths:', ", ".join(['{:02d}'.format(x) for x in code_lengths]))

    # Repeat until entire image has been read
    while lzw_string!= '':
        # Try to match current symbol against dictionary entry
        next_symbol = lzw_string[0]
        substrings = ['']
        for c in lzw_dictionary_values: 
            if c == current_code + next_symbol:
                substrings = [[c]]
                break
            elif c.startswith(current_code + next_symbol): 
                substrings.append([c])
        # If we couldn