#!/usr/bin/env python3

import sys
import os
os.chdir(sys.path[0])

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Write the magic string (header) to the output stream
    out.write(b'GIF89a')

    # Generate some data for the body
    datalen = int.from_bytes(rng.read(1), byteorder='big', signed=False)
    print("Data length:", datalen)
    xcoord = int.from_bytes(rng.read(2), byteorder='big', signed=True)
    ycoord = int.from_bytes(rng.read(2), byteorder='big', signed=True)
    lcmd = b'\x21\xf9\x04' + bytes([
        ((xcoord & 0xff00) >> 8) | 0x80,
        (xcoord & 0x00ff) | 0x80,
        ((ycoord & 0xff00) >> 8) | 0x80,
        (ycoord & 0x00ff) | 0x80
    ])
    eocmd = b'\x00'

    # Create the color table and add it to the output
    coltablelen = int.from_bytes(rng.read(1), byteorder='big', signed=False)
    coltables = []
    for i in range(coltablelen):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 0xd0
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 0xd0
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 0xd0
        coltables += [r << 16 | g << 8 | b]
    for c in coltables:
        out.write((c).to_bytes(length=1, byteorder='big'))

    # Add the local color map command
    out.write(lcmd)

    # Generate compressed image data by repeating the following sequence of
    # operations until we have generated enough bytes:
    dnum = 0
    datatotal = 0
    while True:
        if datalen >= datatotal + len(eocmd) * (dnum+1):
            out.write(eocmd)
            if datalen <= datatotal:
                break

            dnum += 1
            datatotal += len(eocmd)

        # Generate an operation code
        op = rng.read(1)[0] % 5

        # Do something based on the value of "op"
        if op == 0:
            # Copy next bytes from input to output
            bytes_to_copy = min(
                0xfe - rng.read(1)[0],
                datalen - (datatotal + len(eocmd))
            )
            data = rng.read(bytes_to_copy)
            out.write((bytes_to_copy).to_bytes(length=1, byteorder='big'))
            out.write(data[:bytes_to_copy-1])
            datatotal += bytes_to_copy
        elif op == 1 or op == 2:
            # Repeat previous pixel this many times
            repeat_count = 2 + rng.read(1)[0] % 7
            out.write(((repeat_count-2) & 0x0f).to_bytes(length=1, byteorder='big'))
            if op == 2:
                out.write((((repeat_count-2) & 0x70) >> 4).to_bytes(length=1, byteorder='big'))
            else:
                out.write(b'\x00\x00')
            
            datatotal += 2
            
        elif op == 3:
            # Change palette colors
            numcolors = 2 + rng.read(1)[0] % 7
            assert numcolors < 8
            if numcolors > 0:
                datatotal += 2
                out.write(numcolors.to_bytes(length=1, byteorder='big'))
                out.write((op<<3|numcolors-1).to_bytes(length=1, byteorder='big'))

                pixels = list(struct.unpack('>'+'H'*datalen,
                    rng.read(datalen*2)))
                
                for _ in range(numcolors):
                    randcoloridx = rng.read(1)[0] % len(pixels)
                    newval = pixels[randcoloridx]

                    out.write(newval.to_bytes(length=2, byteorder='big'))
                    
                    del pixels[randcoloridx]
        elif op == 4:
            pass
        else:
            raise ValueError('Invalid opcode!')


        
    
    # Add terminator block to mark end of file
    #: EndOfFile = b'\x3b'
    #: BlockSize = b'\x00\x00'
    #: ZeroBlock = b'\x00\x00'
    #: TerminatorBlock = b''
    #: TerminatorBlock += BlockSize
    #: TerminatorBlock += EndOfFile
    #: TerminatorBlock += ZeroBlock
    #: TerminatorBlock += b'\x3a'    # Application Extension marker
    #: TerminatorBlock += b'\x0b'    # Size of application extension block
    #: TerminatorBlock += b'\x47'    # Netscape application identifier block size
    #: TerminatorBlock += b'NETSCAPE'
    #: TerminatorBlock += b'2.0'     # Version number of NETSCAPE
    #: TerminatorBlock += b'\x03'    # Compression flag
    #: TerminatorBlock += b'\x01'    # Loop count
    #: out.write(TerminatorBlock)

    
if