#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    
    # Prepend header to output
    out.write(b'GIF89a')

    
    ### First Image Descriptor ###

    # Create lsd for this frame
    lsd = struct.pack("<HHBbb", 0, 0,
                      0xf0 | rng.read(1)[0],
                      0xff&rng.read(1)[0],
                      rng.read(1)[0])

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int(2**(int((lsd[4] >> 5)))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = False
    
    while True:
        # Create new frame        
        out.write(b'\x2c')
        
        # Create image descriptor
        id = rng.read(9)
        out.write(id)

        # Get number of bytes to read based on bits in ID
        imgdatalenbits = int(id[8])&0x7
        imgdatalen = int(id[4])*pow(2,imgdatalenbits+1)+imgdatalenbits+1
        codelen = pow(2,(id[6]&7))+1

        # Get dispose method, if any
        dispose = -1 if (id[8]&8)==0 else id[8]&7

        if len(prevframe)!= 0 and \
           imgdatalen < len(prevframe) and \
           (dispose==-1 or dispose<=2):
            out.write(prevframe[:imgdatalen])

        elif dispose == 2:
            out.write(prevframe[-imgdatalen:])

        else:
            out.write(b'\0'*imgdatalen)
            
            
        # If we don't yet know how much image data we need per line,
        # get that info now by reading the first row of pixels
        if codelen == -1:
            out.seek(-1, 1)

            # If we've already generated an entire line, just use that data
            if imgdatalenbits <= 6:
                codelen = 2**imgdatalenbits
            # Otherwise, generate more data until this many bytes are available
            else:
                while True:
                    n = 1 << (imgdatalenbits-6)

                    x, y = ord(rng.read(1)), ord(rng.read(1))
                    
                    if x >= n or y >= n:
                        break

                codelen = n
                
            out.seek(1, 1)

        # Now create the actual pixel data
        data = []

        for i in range(ceil(imgdatalen/(codelen*2)-1)):
            clen = rng.read(1)*256 + ord(rng.read(1))
            data += [clen]*clen

        out.write(bytes([codelen]*ceil(imgdatalen/(codelen*2)-1)) +
                  bytes(data))

        # Save as previous frame because we might need it later
        prevframe = out.getbuffer()[bufposition:].tobytes()