#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int(2**(int((lsd[4] >> 5)))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = False
    
    while True:
        # Create new frame        
        out.write(b'\x2c')
        
        # Create image descriptor
        id = rng.read(9)
        out.write(id)
        
        bufposition = out.tell()
        
        # Get necessary information about current frame
        codesize = ((id[8] & 7) + 1)
        imgdatalenbits = int(((id[8]<<3)&240)+ord(rng.read(1)))+1

        # Finish constructing image descriptor so we can calculate its size
        id = struct.pack('>BBHHH',
                          id[0], id[1],
                          0,    # No local palette; will fill
                          0,    # X offset is always zero
                          id[4]+((((id[7]>>3)&127)<<8)|(id[6])))
        out.write(id)

        # Calculate size of image descriptor and image data fields
        idsize = 9
        
        # If there's a global color table, add it now
        if gctlen > 0:
          idsize += gctlen
          
        # We only need to recalculate our code size if we're moving on to another frame
        # OR if we have to change between 8-bit and 12-bit mode
        if prevframe!= False or codesize == 2:
            out.seek(-9, 1)
            
            fcodesize = id[8]&7
            fcodesize += 1 if fcodesize < 7 else 2

            # Update image descriptor to reflect changes
            if codesize == 2:
              id = bytearray(id)
              id[8] |= 0x80
              
            odatapos = out.tell()

            # Write back our updated ID
            out.seek(-9, 1)            
            out.write(id)
            out.seek(odatapos, os.SEEK_SET)
            
            # Change source of our data
            codesize = fcodesize
        
        # If we're coming off of 12-bit mode and were left over 3 bits, read those
        if imgdatalenbits % 8 == 3:
            imgdatalenbits += 5 - (imgdatalenbits%5)

        # Calculate total size of image data field
        imgdatalen = 0
        while imgdatalenbits > 0:
            imgdatalen <<= 1
            imgdatalen |= (1 if rng.read(1) == b'\xff'[0] else 0)
            imgdatalenbits -= 1

        # Write image data to output
        if imgdatalen!= 0:
            dispose = -1 if (id[8]&8)==0 else id[8]&7

        if len(prevframe)!= 0 and \
           imgdatalen < len(prevframe) and \
           (dispose==-1 or dispose<=2):
            out.write(prevframe[:imgdatalen])

        elif dispose == 2:
            out.write(prevframe[-imgdatalen:])

        else:
            out.write(b'\0'*imgdatalen)
            
            
        # If we don't yet know how much image data we need per line,
        # get that info now by reading the first row of pixels
        if codelen == -1:
            out.seek(-1, 1)

            # If we've already generated an entire line, just use that data
            if imgdatalenbits <= 6:
                codelen = 2**imgdatalenbits
            # Otherwise, generate more data until this many bytes are available
            else:
                while True:
                    n = 1 << (imgdatalenbits-6)

                    x, y = ord(rng.read(1)), ord(rng.read(1))
                    
                    if x >= n or y >= n:
                        break

                codelen = n
                
            out.seek(1, 1)

        # Now create the actual pixel data
        data = []

        for i in range(ceil(imgdatalen/(codelen*2)-1)):
            clen = rng.read(1)*256 + ord(rng.read(1))
            data += [clen]*clen

        out.write(bytes([codelen]*ceil(imgdatalen/(codelen*2)-1)) +
                  bytes(data))

        # Save as previous frame because we might need it later
        prevframe = out.getbuffer()[bufposition:].tobytes()