#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random(out: BinaryIO, rng: BinaryIO):
    # Preamble of a gif header
    out.write('GIF89a'.encode("ascii"))

    # Logical screen descriptor is fixed size and has
    # width 1280 pixels, height 768 pixels, and a color depth of 8 bits per pixel
    width = 1280
    height = 768
    bitdepth = 8
    globalcolorcount = (2 ** bitdepth)-1
    localcolorcount = 0
    sortedcolorcount = 0
    bgcolorindex = 0xff
    aspectratiocode = 0b1011

    packedscreendescriptor = \
        f'{width}' +\
        f'{height}' +\
        f'|{bitdepth}00000' +\
        '10' +\
        f'{sortedcolorcount}' +\
        f'{bgcolorindex}' +\
        f'{aspectratiocode}'

    # Write logical screen descriptor to file
    out.write(struct.pack('<6sBBBBB', *packedscreendescriptor.split()))
    
    # Global color table is empty in this example
    out.write((globalcolorcount*3)*'\x00')
    
    # Calculate how many bytes are needed for each row by finding the remainder 
    # when we divide the image's byte length with the height first then multiplying it
    # by two because each pixel takes up two bytes
    rowbytes = ((width*bitdepth)//8)+(((width*bitdepth)%8)>0)*1
    print(f"Row Bytes: {rowbytes}")

    # Write the chunk for the graphic control extension
    packedgraphiccontrol = \
        "0021"\
        "00fc"\
        f"{rowbytes}"\
        f"{width}"\
        f"{height}"\
        "00"\
        "00"\
        "00"\
        "00"\

    # Write graphics control extention to file
    out.write(struct.pack("<4sb", *(packedgraphiccontrol[i] for i in range(len(packedgraphiccontrol)))))
    
    # Finally write the image data at its normal offset after all extensions so that we can just
    # calculate where the frame starts based on the current position in the file
    startoffset = out.tell()
    fullrows = height//256+(((height%256)>0)*1)
    partialrows = fullrows*((width*bitdepth)%8)>0
    fullrowpixels = width*fullrows
    partialrowpixels = (((width*bitdepth)%8)>0)*((((width*bitdepth)%8)/8)+1)

    if (width % 8 == 0):
        wordsperrow = width//8
    else: 
        wordsperrow = 1+(width//8)
    
    print(f"Full Rows: {fullrows}, Partial Rows: {partialrows}")
    print(f"Total Pixels: {width*height}")
    print(f"Total Words Per Row: {wordsperrow}")
    print(f"Bytes Per Word: {(width*bitdepth)//8}")
    print(f"Full Row Pixels: {fullrowpixels}, Partial Row Pixels: {partialrowpixels}")
    print(f"Start Offset: {startoffset}")
        
    for _ in range(fullrows):
        # Each row consists of an unsigned short indicating the number of words in that row followed by the
        # actual row itself which is made up of one or more bytes depending on the the total number of rows
        # and the color depth. The last row may have padding before its end if the total number of rows isn't
        # evenly divisible by ten
        rowwords = wordsperrow
        rowdata = [int(rng.read(1).hex(), base=16) & 0xff for _ in range(rowwords)]
        rowpad = [0]*(((rowwords*8)-(width*bitdepth))//8)

        # Write row data to file
        out.write(
            struct.pack(
                "<H"+str(rowwords)+"B"+str(len(rowpad))+"B", 
                len(rowwords), 
                *rowdata,
                *rowpad
            )
        )

    for _ in range(partialrows):
        row = []
        
        for __ in range(partialrowpixels):
            # If there was any leftover space left over from the previous loop
            # fill it with zeros until we reach our final pixel count
            row += [(int(rng.read(1).hex(), base=16) & 0xff)]

            # If it's time to hit the end of a word
            if (len(row)==((width*bitdepth)//8)):

                # Pack the whole word as big endian
                out.write(struct.pack(">H", int((''.join([f"{val:02X}" for val in reversed(row)])),base=16)))
                
                # Reset the row
                row = []
        # Fill partial row with zeroes until we've reached a full word
        while True:
            # Check if we got enough room in a single word 
            if len(row)==((width*bitdepth)//8):
                break

            row += [0]

        # Still need to pad?
        missingbits = partialrowpixels - (len(row)*8 - (width*bitdepth)%8)
            
        if not missingbits==0:            
            row += [*bytearray.fromhex('00'*missingbits)]

        # Pack final word as big endian    
        out.write(struct.pack(">H", int((''.join([f"{val:02X}" for val in reversed(row)])),base=16)))

    # Go back and update the offsets in the image data chunks so they point to their correct location
