#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    width = int.from_bytes(
            rng.read(2), byteorder='little') & 0x7fff
    height = int.from_bytes(
            rng.read(2), byteorder='little') & 0x7fff

    # We want to make sure that we don't accidentally create an image too big!
    while True:
       if width < 1 or height < 1:
           continue

       break

    print('width', width,
          'height', height)

    datalen = width * height // 4

    first_byte = 0b111 << 5 | 0b11111 << 2
    second_byte = 0b111 << 5 | 0b11111 << 2

    third_byte = (datalen >> 11) & 0xFF

    fourth_byte = (datalen >> 19) & 0xFF

    # Write pre-frame bytes  

    out.write(header +
              width.to_bytes(2, 'little') + 
              height.to_bytes(2, 'little') + 
              bytes([first_byte]) +
              bytes([second_byte]))

# Writes the actual data for the frame onto `out`, which is assumed to be
# at the right position already.
# The frame's size depends on its palette (`palette`) and transparent color
# index (`transparent`), as well as the dimensions of the canvas it will appear in
def write_framedata(
        out: BinaryIO,
        palette: list[int],
        transparent: int,
        canvas_width: int, 
        canvas_height: int):
    
    # Compute minimum code length for the LZW compression algorithm

    mincodesize = max(len('{0:#0{1}x}'.format(len(palette), 6)), 2)
    

    # Write these values into our "logical screen descriptor"

    out.write((canvas_width).to_bytes(
            2, 
            byteorder='little'))
    out.write((canvas_height).to_bytes(
            2, 
            byteorder='little'))

    out.write((((mincodesize - 1) << 4) | 0).to_bytes(
            1, 
            byteorder='little'))


    # Write out the colors used by this frame,
    # plus the background color

    for i in range(0, len(palette)):
        r = g = b = ((palette[i] >> 16) & 0xff)
        
        if not i == transparent:

            out.write((r).to_bytes(
                    1, 
                    byteorder='little'))
            
            out.write((g).to_bytes(
                    1, 
                    byteorder='little'))
            
            out.write((b).to_bytes(
                    1, 
                    byteorder='little'))

        else:

            out.write((0).to_bytes(
                    1, 
                    byteorder='little'))
            
            out.write((0).to_bytes(
                    1, 
                    byteorder='little'))
            
            out.write((0).to_bytes(
                    1, 
                    byteorder='little'))


    # Write out the "local graphic control block",
    # which specifies the delay time for the animation
    # and whether or not there is any transparency set up
    # for this frame


    out.write(0.to_bytes(
            1, 
            byteorder='little'))
    
    out.write(0.to_bytes(
            1, 
            byteorder='little'))
    
        
    out.write(((transparent!= None)).to_bytes(
            1, 
            byteorder='little'))
    
    out.write(0.to_bytes(
            1, 
            byteorder='little'))
    
    
    # Finally, compress all pixel data via LZW compression
    ## First, compute the compressed value for each pixel,
    ## where the compressed value is the index into
    ## the palette array of the first color found
    ## so far that matches the pixel being encoded.
    intermediates = []

    intermediate: dict[tuple[str,...], str] = {}

    prevseq = ()

    for y in range(0, canvas_height):

        seq = ()

        for x in range(0, canvas_width):
            key = tuple()
            compvalue = ''
            
            # Get the current RGB values from our original bitmap
            r = (ord(rng.read(1)))
            g = (ord(rng.read(1)))
            b = (ord(rng.read(1)))
    
            while True:

                rgbval = '{0}{1}{2}'.format(
                    hex(r)[2:],
                    hex(g)[2:],
                    hex(b)[2:])


                newcomp = ''
                
                try:

                    newcomp = intermediate[(prevseq)] + rgbval
                    
                except KeyError:

                    pass
                
                
                if newcomp in palette:
                    
                    key += (rgbval,)

                    compvalue = str(
                        palette.index(
                            newcomp))

                    
                elif newcomp not in palette and \
                        len(intermediate[(prevseq)]) > 0:
        
                    intermediate[(prevseq)] = newcomp

                    intermediate[(key+ (rgbval,))] = newcomp[0:-1] + rgbval
                    
                    intermediate[(newcomp+(rgbval,))] = newcomp

                    compvalue = str(
                        palette.index(
                            newcomp[0:-1] + rgbval))
                    

                elif newcomp not in palette and \
                        len(intermediate[(prevseq)]) <= 0:

                    key += (rgbval,)

                    intermediate[(key + ())] = rgbval

                    compvalue = str(
                        palette.index(
                            rgbval))


                if compvalue!= '':


                    seq += (compvalue,)

                    break

                prevseq = (key + ())
        
        intermediates += [seq]

    
    ## Second, produce the final compressed stream
    ## by writing each 12-bit binary value of each entry
    ## in the list above, bit-packed along with an end-of-line marker
    indices = []

    for line in intermediates:

        for idx in line:

            indices += [(idx,)]

            
    output = []

    for i in range(0, len(indices) + 1):

        try:
        
            current_entry = indices[i][0]

            current_entry_num = 1
                
            for j in range(i + 1, min(i + 1 + 12, len(indices))):

                if indices[j][0] == current_entry:

                    current_entry_num += 1
                    
                else:
                    break

            output += ['{0:#0{1}x}'.format(current_entry_num, 4)]

            output += [current_entry]
            
        except IndexError:
            
            break
            
    compstream = ''.join(output)    

    #print("compressed stream")
    #print(compstream)

    # Now encode this stream using 
    # the LZW compression format extension,
    # which should be pretty intuitive

    def escape_sequence(c: str) -> str:

        return '\\x' + c



    out.write(escape_sequence('21').encode())
    out.write(escape_sequence('f9').encode())
    out.write(escape_sequence('04').encode())
    out.write((mincodesize).to_bytes(
            1, 
            byteorder= 'little'))

    out.write(bytes.fromhex(compstream))





if