#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int(2**(int((lsd[4] >> 5)))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = False
    
    while True:
        # Create new frame        
        out.write(b'\x2c')
        
        # Create image descriptor
        id = rng.read(9)
        out.write(id)

        # Get number of bytes to read based on bits in ID
        imgdatalenbits = int(id[8])&0x7
        imgdatalen = int(id[4])*pow(2,imgdatalenbits+1)+imgdatalenbits+1
        codelen = pow(2,(id[6]&7))+1

        # Get dispose method, if any
        dispose = -1 if (id[8]&8)==0 else id[8]&7

        if len(prevframe)!= 0 and \
           imgdatalen < len(prevframe) and \
           (dispose==-1 or dispose<=2):
            out.write(prevframe[:imgdatalen])

        elif dispose == 2:
            out.write(prevframe[-imgdatalen:])

        else:
            out.write(b'\0'*imgdatalen)
            
            
        # If we don't yet know how much image data we need per line,
        # get that info now by reading the first row of pixels
        if codelen == -1:
            out.seek(-1, 1)

            while not codelen is not None:
                rdat = out.read(1)

                if len(rdat)!= 0:
                    code = struct.unpack("<B", rdat)[0]

                    if ((code>>7)==0):
                        codelen = code+1

                    elif (((code>>7)&1) == 1):

                        # This means there are an odd number
                        # of bytes left in the current raster line 
                        # then the end of this line will indicate
                        # the length of the next raster line
                        if (code&0x7f) == 0:
                            rdat += out.read(1)

                            if len(rdat)!= 0:
                                code = struct.unpack("<B", rdat)[0]
                                codelen = code + 1
                                
                    elif ((code>>7)&1) == 0:

                        # This means all remaining bytes in this
                        # raster line indicate the length
                        rdat += out.read(code&0x7F)
                        
                        if len(rdat)!= 0:
                            codelen = sum([struct.unpack("B", c)[0]+1 for c in rdat])
                            

                    continue
                
                break
            
            # Seek back to where we were before
            out.seek(imgdatalen-len(rdat), 1)
                
        # Count up how many lines we'll decode
        numlines = imgdatalen//(codelen*width)      

        # Set size of entire frame
        width = height = numlines  

        # Decode the image (the loop body isn't _that_ hot)
        for line in range(numlines):
          # Seek to right spot in raster
          seekto = len(b'\x2C')+(line*(codelen*width))
          out.seek(seekto, 1)
          
          # Decode one horizontal line at a time
          for col in range(width):
              # Read next byte as palette index
              dat = out.read(1)
              
              if len(dat)!= 0:
                  pixval = struct.unpack('<B', dat)[0]

                  # If value is zero but there's no more bytes read, 
                  # ignore this pixel
                  if pixval == 0:
                      dat += out.read(1)
                      
                      if len(dat) == 0:
                          break
                        
                  # Write the pixel value
                  out.write(b'\xff\x00\xFF\x00'*pixval)
          
        # Save off what we just encoded so we can dispose later
        prevframe = out.getvalue()

        # Reset our buffer to empty state
        out.truncate(out.tell())        
