#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    global LZW_CODE_SIZE

    # Header
    magic = bytes([0x47, 0x49, 0x46])
    version = bytes([0x38, 0x39, 0x61])
    width = bytes([rng.read(1)[0] % 256])
    height = bytes([rng.read(1)[0] % 256])
    flags = bytes([0b1100])
    background = [rng.read(1)[0] for _ in range(3)]
    foreground = [rng.read(1)[0] for _ in range(3)]

    comment = b"X\xb8\x8d\xa1@D\x9e\xc9\xbe7\xfca?\xee\xd0\xe6\xef\xff\xfe\xfd\xfb\xfc\xdb\xec\xde\xea\xf4\xfd\xfa\xf8\xed\xf1\xf3\xf7\xf4\xf8\xfa\xfe\xfe\xf6\xf2\xf2\xf8\xfa\xf9\xfa\xf9\xfb\xfa\xfb\xfc\xfb\xfc\xfe\xfe\xfe\xff"
    plain_text = b""

    block_size = int.from_bytes((int.from_bytes(version)+4).to_bytes(1, "big"), byteorder="little")
    ApplicationExtension = (
            (b"\x21\xFF\x0B") + 
            ("NETSCAPE2.0".encode("ascii")) +
            (block_size-len("\x0bNETSCAPE2.0")-3).to_bytes(1, "big")+
            "\x03\x01".encode('ascii') +
            "\x00".encode('ascii')
        )
    
    out.write(magic+version+width+height+flags+background+foreground+comment+ApplicationExtension+plain_text)

    # Graphics Control Extension
    graphics_control = (
            b'\x21' + 
            b'\xF9' + 
            4.to_bytes(1,"big")+
            b'\x04'+
            1.to_bytes(3,"big")+
            bytes([(rng.read(1)[0] // 2)]) +
            b'\x00'
        )

    data_subblocks = []

    while True:
        subblock = rng.read(LZW_CODE_SIZE//8)

        if len(subblock)!= LZW_CODE_SIZE//8 or (((~subblock[0])+1)&0xff) == 0: break
        
        data_subblocks.append(subblock)
    
    lzwmincodesize = int.from_bytes(((~data_subblocks[-1][0])+1).to_bytes(1,'big'),byteorder='little')
    lzwdatablock = (lzwmincodesize<<3 | ((LZW_CODE_SIZE - lzwmincodesize)//(8))-1).to_bytes(1,'big')

    i = 0 
    while True and i < len(data_subblocks):
        subblock = data_subblocks[i]
        if len(subblock) >= 255 and not subblock[:3] == b'\x00\x00\x00':
            print(f"Warning: Size is too large for graphic control extension ({len(subblock)} bytes)")
            break 
        elif 255 <= len(subblock) <= 63296-255*3:
            subblock += b'\x00'*(255-(len(subblock)%255==0))+subblock[::(-1)*(len(subblock)%255!=0)]
            assert (len(subblock)-255)%3 == 0

            chunks = [(len(subblock), subblock[:255])]
            subblock = subblock[255:]
            
            flag = False
            while subblock:
                chunk = subblock[:63296-255]
                
                if not flag:
                    chunks.append((len(chunk), chunk))
                    subblock = subblock[63296-255:]
                else: 
                    chunks[-1] = (chunks[-1][0]+len(chunk), chunks[-1][1]+chunk)
                    subblock = subblock[len(chunk):]

                flag ^= True
        
           # chunks = list(map(lambda x:(x[0],x[1].replace(b'\xff', b'\xff\xff')))[::-1])
            chunks = list(filter(lambda x:not any(y == b'\xff' or y == b'\x00' for y in x[1]), chunks))[::-1]
            gce_data = b''
            for size, chunk in chunks: 
                gce_data+=size.to_bytes(2,"big")
                gce_data+=chunk
            data_length = sum(map(lambda x:x[0],chunks))
            data_length += len(gce_data)
        
            break
            
        index_size = min(max((((len(subblock))//255+1)*255)+2, 255), 64)

        lzwindex = subblock[:index_size].replace(b'\xff',b'\xff\xff').replace(b'\x00',b'\x00\x00')
        lzwdata = subblock[index_size:].replace(b'\xff',b'\xff\xff').replace(b'\x00',b'\x00\x00')

        out.write(graphics_control+lzwdatablock+bytes([len(lzwindex)])+lzwindex+bytes([len(lzwdata)])+lzwdata)
        i += 1

