#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    width = int.from_bytes(rng.read(2), byteorder='little') % 640 + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % 480 + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    background_color_index = int.from_bytes(rng.read(1), byteorder='little')
    pixel_aspect_ratio = int.from_bytes(rng.read(1), byteorder='little') & 0xff

    # Write header to output
    out.write(header)

    # Write logical screen descriptor to output
    out.write(struct.pack("<HHB",
            width,
            height,
            (flags[0] |
             (pixel_aspect_ratio << 8))
            ))

    # Write global color table to output - assume there are no
    # global color entries (by setting size to zero) and write all
    # pixels in the image with color index 0x0f which indicates an
    # unfilled pixel
    out.write(b'\x00\x00\x00\xf0')

    while True:

        # Read block type from input stream
        block_type = bytes([int.from_bytes(rng.read(1), byteorder='little')])

        if block_type == b'\x2c':

            # Write graphic control extension for this frame
            out.write(block_type)
            duration = int.from_bytes(rng.read(2), byteorder='little') + 1
            out.write(duration.to_bytes(2,'little'))
            transparent_color = \
                int.from_bytes(rng.read(1), byteorder='little')
            transcolor_flag = bytes([int.from_bytes(rng.read(1),
                                                     byteorder='little')])
            disposal_method = int.from_bytes(rng.read(1),
                                             byteorder='little') & 0xff
            user_input_flag = bytes([int.from_bytes(rng.read(1),
                                                    byteorder='little')])
            out.write((transparent_color).to_bytes(1, 'little'))
            out.write((transcolor_flag +
                      ((disposal_method       )& 0x7).to_bytes(1,'little') + 
                      ((user_input_flag >>  3)& 0x1).to_bytes(1,'little') +
                      ((user_input_flag >>  5)& 0x1).to_bytes(1,'little') +
                      ((user_input_flag >>  7)& 0x1).to_bytes(1,'little')))
            
            # Write local color table to output - assume it has one entry
            # (r,g,b)=(0,0,0) but set the flag so that we can still use
            # paletted images
            out.write(b'\x81\x00\x00\x00\x00\xff\xff\xff')
        
        elif block_type == b'\x21':

            # Read sub-block type from input stream
            sub_block_type = bytes([int.from_bytes(rng.read(1),
                                                   byteorder='little')])

            if sub_block_type == b'\xf9':

                # Write graphic control extension terminator block
                out.write(sub_block_type)
                return

            else:
                
                raise ValueError("Unknown sub-block type")

        elif block_type == b'\x3b':

            # Write end of file marker
            out.write(block_type)
            return

        else:

            raise ValueError("Unknown block type")

