#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big') & 0xfff

    print('header',header,'datalen',datalen)

    # Frame header
    out.write(header)
    # Image dimensions (200 x 100)
    out.write((50 + 6).to_bytes(2, 'little'))   # width  : 50 pixels
    out.write((40 + 7).to_bytes(2, 'little'))   # height : 40 pixels
    # Packed field:
    #   1 - Global color table flag set
    #   1 - Color resolution is 7 bits per pixel
    #      so that this value is 0b111 = 7
    #   1 - Sorting flag to specify that the local color map should be sorted by decreasing frequency of use
    #   1 - Size of global colors, must be 7 or less if last field bit is set; here we have none
    #       => 0b000 -> no global palette
    #   0 - Size of local colors, must be 7 or less; size is 7 because our local pallette always contains just one color
    #       => 0b000 -> local color depth is 7 bits per pixel
    #   0 - Reserved for future extensions
    #   0 - Background color index
    #       => 0b00000
    #   0 - Pixel aspect ratio, 0 = 1x1
    #       => 0b0000
    out.write(int.to_bytes(0b11100000 | ((7 // 3) << 4), length=1, byteorder='little'))
    # Write color table with one entry
    #     Red component, green compoent and blue are all equal to 0xa0
    out.write(b'\x00' * 7)
    # LZW minimum code size (bits)
    out.write(mincodesize := 2.to_bytes(1, 'little'))
    # Write graphic control extension block
    write_graphiccontrolblock(out)
    # Write image descriptor
    write_imagedescriptor(out)

    timechunkoffset = len(header) + 4 + 1   # 7 bytes in total
    datastartpos = timechunkoffset + 11        # start after the time chunk end mark '!'. Next byte will be the first data byte.
    while True:
        lzwbuf = [0] * 257                  # buffer containing codes of previously written substrings

        i = 0                              # running index into input string s

        codelen = mincodesize              # starting code size (in bits)
        oldstring = ''                     # oldest member of the dictionary
        prevcodepos = None                 # position of previous substring's code in output stream

        def addsubtring(c):
            nonlocal i, codelen, oldstring, prevcodepos

            i += 1                         # advance read pointer
            oldstring += chr(c)            # append new character at end of current substring

            if not (i % (1 << codelen)):   # indicates string has become longer than previous
                return                    # because codelen was many times a power of two before

        try:                                # main loop
            while True:                     # do... until EOF
                d = next(rng)               # get the next digit from /dev/urandom

                if d < 0xA0:                # ASCII char
                    yield from addsubtring(d)
                else:                       # otherwise, d is between 0xA0 and 0xFF; it represents a backreference

                    # Here we'll assume that d is a single-digit decimal integer 
                    # representing the length of a string which matches some earlier substring
                    n = int.from_bytes([d], byteorder='little')
                    
                    # If n > codelen, it means that we've encountered a reference to an unseen substring
                    # This can happen when the random number generator returns numbers greater than 0xFF,
                    # since only a maximum of three bytes may fit within a single byte.
                    # We can detect such cases by checking whether the most significant byte of our backreference
                    # matches what would appear in the dictionary if we had seen more than 255 strings of any given length.
                    if n > codelen and d == ord(oldstring[n]):
                        raise RuntimeError("Encountered backreference to unknown substring")

                    for j in range(n):
                        yield from addsubtring(ord(oldstring[-j]))
                
                # We haven't yet checked whether n is a valid backreference against existing strings in dictionary.
                # If it turns out to be invalid, then dictionaries will grow too large,
                # causing us to reach a point where adding a new string exceeds the maximum
                # storage space available inside the LZW compression scheme. In other words, we'll have reached
                # the limtation of the length of the LZW coding itself!
                if len(lzwbuf) <= len(oldstring):
                    if traceoutput:
                        with open('/tmp/%04x' % len(lzwbuf), 'ab') as f:
                            f.write(struct.pack('<H',len(oldstring)))

                    # As a fallback strategy, we "reset" the LZW algorithm to its initial state
                    # (= reset codelength to 9, clear dictionary, set up initial dictionary values)
                    codelen = mincodesize
                    lzwbuf = [0] * 257          # contains strings, indexed by their lengths
                    oldstring = ''             # current substring being processed
                    prevcodepos = None         # position of prevsubstring's code in output stream
                
                    yield from addsubtring(0x21)           # Escaped char 0x21 == '!'
                    yield from addsubtring(0xff)           # Escaped char 0xFF == '?'
                    yield from addsubption(0x00)           # Zero-length string terminator
                    continue
            
                # At this point, we know that n is a valid backreference against existing strings 
                # in the dictionary, but we still need to decide how exactly we're going to represent it.
                # The simplest way to do that is to simply emit a code corresponding to the length
                # of the old substring. However, there's another option: we could also encode the
                # actual matching substring. That takes less space on disk, obviously, but can potentially
                # result in bigger compressed files. It depends on both the compressibility of the original data
                # and the size of the resulting GIF file. On average, GIF images tend to get larger over time,
                # whereas PNG images tend to remain much smaller. So to make best use of both kinds of files,
                # we choose to have a random choice: sometimes we interpret backreferences
                # as encoding positions in dictionary (for which we don't need to store anything);
                # other times we actually construct encoded copies of them in the output stream.
                if rng.randint(0,1) == 0:
                    yield from addsubtring(lzwbuf[len(oldstring)])
                elif prevcodepos!= None:
                    yield from addsubtring(prevcodepos)
                else:
                    yield from addsubtring(0x00)           # zero-length string terminating code

                # Update the LZW dictionary
                lzwbuf.append(lzwbuf[len(oldstring)]+1)

                # Advance code size counter
                if i >= (1<<codelen)-1 and codelen < 12:
                    codelen += 1
        
        except StopIteration:                   # end of source file
            break

    # Write endmark
    out.write(b'\x3b\x00')


def write_graphiccontrolblock(out: BinaryIO):
    pass

def write_imagedescriptor(out: BinaryIO):
    pass


