#!/usr/bin/env python3

import sys
import os
os.chdir(sys.path[0])

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Write the magic string (header) to the output stream
    out.write(b'GIF89a')

    # Generate some data for the body
    datalen = int.from_bytes(
            rng.read(2), 'big')
    data = rng.read(datalen)
'''
    Byte 1:
        0x07 -> Global Color Table Flag set
        0x00 -> No color resolution sub-block
        0x00 -> Sorting flag is not used since there's no local color table
    Byte 2:
        Screen width in left-most position of screen descriptor block
    Byte 3:
        Screen height in next byte (right-most position of screen desc.) + 1 
    Bytes 4-5:
        Zeroes as this field has been deprecated and should be ignored by readers
    Byte 6:
        Aspect ratio byte, expressed as an aspect ration numerator divided by 
        its denominator. For example, "4" means it's a square image.
    Byte 7:
        Background color index
    Byte 8:
        Pixel aspect ratio
'''
    screen_size = [int.from_bytes(rng.read(1), 'big'), int.from_bytes(rng.read(1), 'big')]
    background_index = int.from_bytes(rng.read(1), 'big')

    # Read the rest of the header, which we'll just ignore
    header_end = int.from_bytes(rng.read(1), 'big')
    while header_end!= ord(';'):
        header_end = int.from_bytes(rng.read(1), 'big')
    
    color_palette_entries = 2**int.from_bytes(rng.read(1), 'big')
    global_color_table = []
    for i in range(color_palette_entries):
        if int.from_bytes(rng.read(1), 'big') == 255:
            global_color_table += [
                None] * 3
        else:
            # Build RGB tuple from bytes
            global_color_table += [(
                    int.from_bytes(rng.read(1), 'big'), 
                    int.from_bytes(rng.read(1), 'big'), 
                    int.from_bytes(rng.read(1), 'big'))]    
    print(f'Color palette entries: {color_palette_entries}, global color table size: {len(global_color_table)}')

    # Generate a frame at 10ms delay, with LZW compression enabled
    out.write(b'\x2C\xFB\x0A')            # Start of Frame Block
    out.write(screen_size[0].to_bytes(2,'big'))   # Left-most position of screen descriptor block
    out.write((screen_size[1]-1).to_bytes(2,'big'))   # Right-most position of screen desc. +1
    out.write(b'\x00\x00')                # Zeroes
    out.write(b'\x80\x33')                # Aspect ratio
    out.write(background_index.to_bytes(1,'big'))    # Background color index
    out.write(b'\x00')                    # Pixel aspect ratio
    out.write(data[:255])                # Image descriptor (left-pos, top-pos) & LZW min code size
    out.flush()


    '''
    # If the global color table flag is set, then write the global color table
    if len(global_color_table) > 0:
        out.write(b'\xFF\x0B')        # Application Extension
        out.write(b'\x01')          # Sub-block size
        out.write(b'')              # App name (NETSCAPE)
        out.write(b'\x03')          # Extension size
        out.write(b'\x01')          # Function code (get info)
        out.write(b'\x00')          # Terminator

        out.write(struct.pack('<H', len(global_color_table)))      # Number of colors
        for entry in global_color_table:
            out.write(entry[0].to_bytes(2,'little'))                 # R
            out.write(entry[1].to_bytes(2,'little'))                 # G
            out.write(entry[2].to_bytes(2,'little'))                 # B
            out.write(b'\x00')                                       # Unused

        out.write(b'\x00')                                          # End of extension

    # Now allocate the local color table, if any
    local_color_table = []
    for i in range(color_palette_entries):
        if int.from_bytes(rng.read(1), 'big') == 255:
            local_color_table += [
                None] * 3
        else:
            # Build RGB tuple from bytes
            local_color_table += [(
                    int.from_bytes(rng.read(1), 'big'), 
                    int.from_bytes(rng.read(1), 'big'), 
                    int.from_bytes(rng.read(1), 'big'))]
'''


