#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int(2**(int((lsd[4] >> 5)))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = False
    
    while True:
        # Create new frame        
        out.write(b'\x2c')

        # Frame header has two bytes of flags followed by either an
        # image descriptor or a graphic control extension
        hdr = rng.read(10)
        
        # Write flags byte to output
        out.write(hdr[:1])
        
        # If there is no previous frame...
        if not prevframe:
            # Set the delay time
            dlytime = rng.read(2)

            # Write the delay time to the output stream
            out.write(dlytime)
            
            # If the next bytes in our header are zeroes then this isn't 
            # actually an image descriptor but instead a special case where
            # we need to use a graphic control block so let's try that
            if all([
                i == 0 for i in hdr[2:]
            ]):
              # We want to store how many pixels there will be minus one
              # because there must be at least one pixel per frame
              imgdatalen = rng.read(1)[0] - 1

              # Then we read from the input stream again until the first 
              # non-zero value is found after reading the next 1 byte which
              # should correspond to the code size (how many bits do you
              # need to represent the largest possible index?)
              codelen = None
              while not isinstance(codelen, int):
                  codelen = rng.read(1)[0]

                  # Check that the character was actually a digit
                  assert str(chr(codelen)).isdigit()
              
              # Lastly we read the rest of the graphic control flag section
              extflags = rng.read(1)[0]

              # At this point the graphics control byte can contain values
              # from 0 to 7 inclusive where bit 0 indicates whether transparency
              # should be used when displaying this frame and disposal method
              # defines what should happen when the image goes offscreen
              dispose = extflags & 7 

              # Finally we write our entire control byte back out to the 
              # output stream
              out.write(struct.pack('B', extflags))

        else:
          # Store the amount of data written for this frame so far
          imgdatalen += len(prevframe)

          # Re-use the same delay time as last time!
          out.seek(-2, 1)
          
          # Also re-use the exact same graphic control flag configuration 
          # as well as the already stored image descriptor length
          out.write(extflags)
          
        # Did we run out of memory?
        assert imgdatalen < 2 ** codelen

        # Now we can read out the actual pixel data
        imgdata = rng.read(imgdatalen)

        # Finally we can write these pixels to the output
        out.write(imgdata)

        # Update the total byte count so we know exactly how long to 
        # expect the following frame to be
        out.seek(3, 1)
        out.write(struct.pack('<H', len(imgdata)))

        # Keep track of the current frame type since later on we'll
        # need to write different headers between images vs. animation loops
        prevframe = imgdata

        # Is this the end of the gif?
        if len(imgdata)!= imgdatalen + 1: break
