#!/usr/bin/env python3

import struct
from typing import BinaryIO
if __name__=="__main__":
    from common import b,u,print_blocks_hex
else:
    from.common import *


def xorshift_stream_generate(rng:BinaryIO):
    """ Generate random data using xortshift stream cypher"""
    while True:

        # Write a single random byte
        yield int.from_bytes((rng.read(1)),byteorder="big")



def xorshift_generate(seed:int=None)->bytes:
    """ Generate a random sequence of bytes by xoring each byte with 
        the next one generated by an algorithm that is indistinguishable 
        and almost as good as the one used in the C standard library 
    """

    if seed!= None:
        raise NotImplementedError()
    
    # The counter starts at zero so we can have the first
    # value be anything
    ctr=0

    # For the time being we just read from /dev/urandom but this could easily
    # be replaced with any other source such as Cryptographic Random Number Generators
    # or a hardware RNG
    with open("/dev/random","rb") as rng:        
        
        for _ in range(2**20):

            ctr = ((ctr << 8) & 0xffffffffffffffff) ^ rng.read(1)[0]
            yield ctr & 0xff



# http://csrc.nist.gov/publications/detail/fips/140/1/final
def rc4_generate():
    raise NotImplementedError()


def lfsr_generate():
    raise NotImplementedError()


def mersenne_twister_generate():
    """ Return a pseudorandom number generator based on the Mersenne Twister
    algorithm via
    https://en.wikipedia.org/wiki/Mersenne_Twister

    This version uses the MT19937 Mersenne Twister Algorithm with a word length
    of 32 bits and a period of 2^19937-1.
    """
    MODULUS = (2 ** 19937)-1    

    def twist(mt_state):
        for i in range(624):
            
            y = mt_state[i] & (1<<31)
            y ^= y >> 1
    
            if not (y&1):
                y ^= 0x9908B0DF
                
            y ^= mt_state[(i+1)%624] 
            y ^= y << 1
            
            if y & 0xFFFFFFFF >= MODULUS:            
                y ^= MODULUS

            mt_state[i] = mt_state[(i+397)%624] ^ y
    
        return y


    state=[0]*624
    yield 0
    
    for n in range(MODULUS):
        k = n // 624
        s = state[(n%624)]
        state[(n%624)] = twist(state[:]+[s])
        yield (state[(k%624)] * 0x41C64E6D + k) & 0xFFFFFFFF


def crc32_generate(crc_init:int=0):
    """ A generator which yields a CRC32 with initial value crc_init (defaulting to 0)

    We use zlib's crc32 implementation because it is fast
    """
    import zlib
    yield crc_init
    crc = zlib.crc32(b(""),crc_init) & 0xffffffff
    while True:
        val = crc
        crc = zlib.crc32(struct.pack('<I',val),crc_init) & 0xffffffff
        yield val



if __name__=='__main__':

    TEST_LEN=1000000
    TEST_SEED=17*19
    rng=xorshift_generate(TEST_SEED)


    # First generate a bunch of random bytes
    # We'll use this to check our output later
    testvals = list(range(TEST_LEN))
    for _ in range(len(testvals)):
        testvals[_]=next(rng)

    
    # Create a file that contains all of those random numbers
    # but skip the last two ones since they are unusable due to padding
    with open('testdata','wb') as out:
        for t in testvals[:-2]:
            out.write(
                ((int.from_bytes(
                    rng.read(1), byteorder="big") % 256) + 1).to_bytes(1,"big"))

    # Now write the loop variable
    # -1 is 0xff
    #  0 is 0x00
    #  1 is 0x01
    # etc
    loopvar = (
            (int.from_bytes(
                rng.read(1), byteorder="big") % 5)
            | 0b10000000)
    print("loopvar",loopvar)
    if loopvar == 0xff:
        badval = 0
    else:
        badval = loopvar + 1

    print