#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# Rock Paper Scissors Lizard Spock game.
def generate_random(out: BinaryIO) -> None:
    # Write header section
    data = bytearray()

    # Logical screen width and height in pixels (each)
    # 640x480 is standard for screens today
    w, h = [struct.pack('<H', x) for x in (640, 480)]

    flags = b'\x87\x69'   # Screen descriptor + global color table
    backgroundcolorindex = struct.pack('B', 0)     # Background color index
    pixelaspectratio = b'\x0A'                   # Pixel aspect ratio (=10)

    data += flags + w + h + \
            backgroundcolorindex + pixelaspectratio

    # Global color table
    # Just use one global color with 255 red, green and blue values
    rgb = b'\xFF\xFF\xFF'
    transparantflag = b'\x00'
    gctsize = b'\x01';

    data += gctsize + rgb + rgb + rgb + transparantflag

    out.write(b'GIF8')             # Header string (always "GIF8" or "GIF9")
    out.write((b'\x00\x3B').ljust(6))    # File size - 11 bytes
    out.write(data)               # Image descriptor block

    # Frames loop - we always write exactly one frame
    for i in range(1):
        subblockcount = struct.pack('<B', 3)           # Extension code 0f

        # Graphics control block
        graphicscontrollabel = struct.pack('<B', 0xf9)      # Block label (with length of 4 bytes)

        flags = b'\x89\xA4'                           # Flags (we set only the 5th bit here to tell if it's user input)
        delaytime = struct.pack('<H', 100)              # Delay time between frames in hundredths of seconds
        transparentcolorindex = struct.pack('<B', 0)    # Index of the transparent color
        terminator = struct.pack('B', 0xff)            # Byte that marks end of extension blocks

        data = subblockcount + graphicscontrollabel + \
                flags + delaytime + transparentcolorindex + terminator

        out.write(subblockcount)                  # Sub-block count
        out.write(graphicscontrollabel)           # Graphics control block start code

        out.write(flags + delaytime)               # Graphic control block contents

        # Plain text block starts
        plaintextlabel = struct.pack('<B', 0x01)        # Label for plain text metadata block (with 1 byte length)

        disposalmethod = b'\x00'                        # Disposal method (none)
        userinputflag = b'\x00'                         # User input flag (no action on click)
        maxcodesize = b'\x0B'                           # Maximum code size

        data += plaintextlabel + \
                disposalmethod + userinputflag + maxcodesize + terminator

        out.write(plaintextlabel);                    # Plain text block start code
        out.write(disposalmethod + userinputflag + maxcodesize)    # Plain text block contents
        out.write(terminator);                         # Terminating byte for plain text block (extension block type 1)

        # Application extension block starts
        applicationextlabel = struct.pack('<B', 0x02)       # Label for application-specific metadata block (with 1 byte length)
        appid = b'NETSCAPE2.0'                           # Application identifier
        authcode = b'IS<PASSWORD>'                       # Authentication code

        data += applicationextlabel + len(appid).to_bytes(1,'big') + appid + \
            struct.pack('<B', len(authcode)-1+19) + authcode + terminator;

        out.write(applicationextlabel);                 # Application extension block start code
        out.write(len(appid).to_bytes(1,'big') + appid)         # Application extension block contents
        out.write(struct.pack('<B', len(authcode)-1+19))   # Size field
        out.write(authcode)                             # Authentication code
        out.write(terminator);                          # Terminating byte for application extension block (extension block type 2)

        # Image descriptor block begins
        imagedescriptorlabel = struct.pack('<B', 0x2c)          # Label for image description metadata block (with 1 byte length)
        leftcorner = struct.pack('<h', 0)                     # Left corner position
        topcorner = struct.pack('<h', 0)                      # Top corner position
        width = struct.pack('<h', 640)                        # Width
        height = struct.pack('<h', 480)                       # Height
        lctused = struct.pack('<B', 1)                        # Local color table flag
        interlacemode = b'\x00'                              # Interlace mode
        sortorder = b'\x00'                                  # Sort order
        localcolortablesize = b'\x02'                        # Local color table size

        data += imagedescriptorlabel + leftcorner + topcorner + width + height + lctused + interlacemode + sortorder + localcolortablesize + rgb + rgb + terminator;

        out.write(imagedescriptorlabel);                                # Image descriptor block start code
        out.write(leftcorner + topcorner + width + height + lctused + interlacemode + sortorder + localcolortablesize)                # Image descriptor block content
        out.write(rgb + rgb)                                             # Color table entries
        out.write(terminator);                                           # Terminating byte for image descriptor block (extension block type 0)

        # The remainder of this sequence of writes will be compressed image data.
        # We can just write a bunch of repetitions of the same character.
        # This gives us all sorts of different images over time!
        # It's really neat :)
        out.write(b'R' * ((w[0] * h[0]) // 2), 640*480//2)

    # Trailer
    out.write(b';')                                 # GIF file trailer
    out.write(struct.pack('<B', (((w[0] * h[0]) % 256)+1)))    # Final zero-byte - because compression introduced an extra byte per row of image data...
    out.write(struct.pack('<B', 0))                       # Another terminating zero-byte at the very end of the file

if __name__ == '__main__':
    with open('./output.gif','wb') as f:
        generate_random(f)

    print("Wrote output to./output.gif");
    print("Image dimensions are {}x{}".format(*struct.unpack("<HH",open("./output.gif","rb").read()[6:10])[::-1]))
    exit();

# convert output.gif -crop 128x128+0+0 output.png && feh output.png && rm output.gif && rm output.png &
# ffmpeg -r 30 -i output.%d.png -y movie.mp4
# ffmpeg -framerate 30 -pattern_type glob -i '*.png' -vcodec libx264 -pix_fmt yuv420p out.mp4

# http://www.matthewflickinger.com/lab/whatsinagif/bits_and_bytes.asp
# http://giflib.sourceforge.net/whatsinagif/index.html

# https://stackoverflow.com/questions/3000664/how-do-you-embed-a-custom-palette-into-an-animated-gif

# http://giflib.sourceforge.net/whatsinagif/index.html
# https://ezgif.com/optimize

# http://giflib.sourceforge.net/whatsinagif/bits_and_bytes.html
# https://www.matthewflickinger.com/lab/whatsinagif/bits_and_bytes.asp
# https://www.matthewflickinger.com/lab/whatsinagif/tag_blocks.asp
# https://en.wikipedia.org/wiki/Graphics_Interchange_Format
# https://www.fileformat.info/format/gif/egff.htm
# https://web.archive.org/web/20170810070943/http://pages.cpsc.ucalgary.ca/~jacobs/Courses/cpsc217/W10/code/notes/Gifinternals.html
# https://www.ntu.edu.sg/home/ehchua/programming/java/JavaDataInputOutputStream.html
# https://computing.llnl.gov/tutorials/parallel_comp/#Message_Passing_Interface
# https://docs.python.org/3/library/struct.html
# https://en.wikibooks.org/wiki/Python_Programming/Binary_Files
# https://en.wikipedia.org/wiki/LZW
# https://en.                        4, 'little'))

