#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        width : int, 
        height : int, 
        numcolors : int,  
        numframes: int,
        out: BinaryIO
    ):

    # Header
    version = "GIF89a".encode("ascii")
    out.write(version)
    out.write(struct.pack('<HHHBBBBB', 0xA1B2C3D4, 5678, 
        7,    # Logical Screen Width
        7,    # Logical Screen Height
        0xF7  | 0x00 | (numcolors << 4),   # Flags
        0,    # Background Color Index
        0,    # Pixel Aspect Ratio
    ))

    # Global Colors Table
    if numcolors > 0:

        for i in range(numcolors):

            red = int.from_bytes(
                    rng.read(1), 
                    byteorder="little")
            green = int.from_bytes(
                    rng.read(1), 
                    byteorder="little")
            blue = int.from_bytes(
                    rng.read(1), 
                    byteorder="little")

            print('color',i,'rgb',red,green,blue)
            
            out.write(red.to_bytes(
                    1, 'little'))
            out.write(green.to_bytes(
                    1, 'little'))
            out.write(blue.to_bytes(
                    1, 'little'))


    # Loop for each frame    
    for f in range(numframes):
        
        # Image Descriptor Block
        if numcolors > 0 and False:
            flags = 0x2C | 0x00 | (numcolors << 4)
        else:    
            flags = 0x0C | 0x00 | (numcolors << 4)
        out.write(struct.pack('<HHHHBBBB', 
            0xA1B2C3D4,
            0x0001, 
            0x0001 + f *width*height, 
            width,
            height,
            0xF7,  
            f,
            flags))


        # LZW Compression Method       
        lzwsize=True
       
        if lzwsize == True:
            subblocklength = struct.pack('B', 
                0xFE | ((f+1)%2**7))
            out.write(subblocklength)
        else:
            subblocklength = b'\xFF'
            out.write(subblocklength)
    
        # Image Data Block        
        table = []
        codesize = 8
        clearcode = (1 << codesize)
        endofinformation = clearcode + 1
        lastcode = None
        bitstream = ''  
            
        pixeldata = bytearray()
        for y in range(height):
            for x in range(width):
    
                r = int.from_bytes(
                        rng.read(1), 
                        byteorder='little')
                g = int.from_bytes(
                        rng.read(1), 
                        byteorder='little')
                b = int.from_bytes(
                        rng.read(1), 
                        byteorder='little')
    
                rgb = '{:02X}{:02X}{:02X}'.format(r,g,b)
                
                index = len(table)
            
                if not rgb in table:
                    table.append(rgb)
                    
                pixeldata += struct.pack('<B',index)
                
                #print(x,y,rgb,len(table)-clearcode-lastcode-1)
                
        print('table',table[:min(10,len(table))])            
    
        # Generate the LZW compression codes
        bitsleft = 0;
        dictbitstring = ""
        code = -1
        output = [None] * (endofinformation)

        for c in sorted([int(c,16) for c in table]):
    
            while True:
            
                codesize += 1
    
                maskbits = 1 << codesize
                noclearmaskbits = ~(maskbits >> 1)
    
                code &= noclearmaskbits

                if code < clearcode or c < (dictbitstring + dictbitstring[0]).__hash__():
                    
                    break
                           
                code -= clearcode
                
            if code >= clearcode:
                dictbitstring+=table[code]
                output[code]=dictbitstring
                code = len(output) - 1
        
            bitsleft += codesize
    
            while bitsleft >= 8:
                bitstr = '{:08b}'.format(code & 255)
                
                bitstream+= bitstr[-1:]
                bitsleft-=1
                if bitsleft>0:
                    code >>= 1 
                else:
                    break
    
    
        # Output the first byte from the stream    
        out.write((code).to_bytes(
                1, 'little'))
    
        # Output the rest of the bytes as long words
        if lzwsize == True:
            numlongwords = ((len(pixeldata)+3)//4)*2
            numlongwords-=1
            out.write(numlongwords.to_bytes(
                        2, 'little'))
        else:
            numlongwords = 0
            out.write(numlongwords.to_bytes(
                        2, 'little'))
     
        numlongs = ((len(pixeldata)+3)>>2)
        for m in range(numlongs):
            
            val = pixeldata[(m<<2)] \
                  + pixeldata[(m<<2)+1] \
                  + pixeldata[(m<<2)+2] \
                  + pixeldata[(m<<2)+3]
    
            val2 = int(val,16)
            out.write(val2.to_bytes(
                    4, 'little'))

        if lzwsize == False:
            return

        # Code Subblocks
        maxbits = 12
        while True:
            if len(bitstream)==0:
                break
    
            if len(bitstream)<maxbits:
                amount = len(bitstream)
            else:
                amount = maxbits
            subblocklength = struct.pack('B', 
                0xFC | (((amount+3)/4)<<2))
            out.write(subblocklength)
    
            numlongs = ((((amount+3)//4)*2)-1)
            for m in range(numlongs):

                val = int.from_bytes(
                        rng.read(4), 
                        byteorder='little')

                print('pixeldata',m,val)
                
                out.write(val.to_bytes(
                        4, 'little'))

