#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # We'll need to store the position after the image data so we can write its CRC later on
        pos = out.tell()

        # Handle case where there's no image data
        if not hasimgdata:
            continue

        imgdata = rng.read(size)
        out.write(imgdata)

        '''
        Update frame with their respective CRCs by moving backwards over 3 bytes at a time until we reach the start of
       the block (which is now the beginning of the previous field)
        I'm sure there's a more efficient way to do this but it works fine for these purposes
        This algorithm doesn't account for blocks that span two or more frames, as well as any other special cases
        It also assumes that all image data follows immediately after an LZW code block
        '''
        currpos = pos + len(imgdata) 
        while currpos >= pos:

            # Get current chunk of three bytes relative to current position
            chunk = f.read(3)[::-1]
            
            # Convert them to a list of integers 
            intchunk = [ord(c) for c in chunk]
            
            # Calculate the CRC based off the current byte order
            crcval = binascii.crc_hqx(bytes([intchunk[0]]), 0) ^ \
                     binascii.crc_hqx(bytes([intchunk[1]]), crcval) ^ \
                     binascii.crc_hqx(bytes([intchunk[2]]), crcval)

            # Write the calculated CRC backwards to the original location, then step backward one byte
            f.seek(-3, 1)
            f.write(struct.pack('<H', crcval))
            currpos -= 1