#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) that is of size at most
# `size`. The GIF has an uncompressed color table and uses a local colour
# map to encode frames.
def generate_random_gif(
        rng:     BinaryIO,
        out:     BinaryIO,
        size:    int
    ) -> bool:
    
    header = b'GIF89a'

    if size < len(header): return False

    # Header
    out.write(header); size -= len(header)

    # Logical screen descriptor
    width = min((int.from_bytes(rng.read(2), byteorder='big')),size//4)
    height = min((int.from_bytes(rng.read(2), byteorder='big')),size//4)
    out.write(width);         size -= 2
    out.write(height);        size -= 2
    flags = int.from_bytes(rng.read(1),'big'); size -= 1
    flags &= ~(0x8 | 0x70)   # unset global colour table
    out.write(flags.to_bytes(1,'big')); size-=1
    bgcolorindex = int.from_bytes(rng.read(1),'big'); size -= 1
    aspect = int.from_bytes(rng.read(1),'big'); size -= 1
    out.write(bgcolorindex.to_bytes(1,'big')); size-=1
    out.write(aspect.to_bytes(1,'big')); size-=1

    if size <= 0: return True

    # Global colour table, we will just use all ones here
    gctsize = max(1<<((flags&7)+1),(1<<(flags>>4)&0xff))
    out.write(((gctsize*3)-1).to_bytes(1,'big')); size -= 1;
    bytesPerEntry = 1 << (((flags >> 4) & 0xf) + 1)
    for _ in range(0,gctsize):
        r,g,b = [int.from_bytes(rng.read(bytesPerEntry),'big')
                 for j in range(0,3)]

        out.write(r.to_bytes(bytesPerEntry)); size-=bytesPerEntry
        out.write(g.to_bytes(bytesPerEntry)) ;size-=bytesPerEntry
        out.write(b.to_bytes(bytesPerEntry)) ;size-=bytesPerEntry
        
    # Loop control block
    loopcount = int.from_bytes(rng.read(2),'big')
    loopblock = int.from_bytes(rng.read(2),'big')
    out.write(loopcount.to_bytes(2,'big')); size -= 2
    out.write(loopblock.to_bytes(2,'big')); size -= 2

    while size > 0: 
        # Image descriptor
        leftpos   = int.from_bytes(rng.read(2),'big')
        upperpos  = int.from_bytes(rng.read(2),'big')
        width     = min((int.from_bytes(rng.read(2),'big')),(size-1)//4)
        height    = min((int.from_bytes(rng.read(2),'big')),(size-1)/4)
        flags     = int.from_bytes(rng.read(1),'big')
        
        lctsize = max(1<<((flags & 7) + 1),(1<<(flags >> 4) & 0xff))
        interlaced = (flags >> 6 == 1)
        sortflag = (flags >> 3 == 1)
        reserved = flags & 7

        out.write(leftpos.to_bytes(2,'big'))       ; size -= 2  
        out.write(upperpos.to_bytes(2,'big'))      ; size -= 2   
        out.write(width.to_bytes(2,'big'))         ; size -= 2   
        out.write(height.to_bytes(2,'big'))        ; size -= 2   
        out.write(flags.to_bytes(1,'big'))         ; size -= 1   
    
        if reserved!= 0 or lctsize!= 0: 
            raise ValueError("Reserved bits set or Local Colour Table present - bad format")
        
        # Local colour table, we will just use all ones here
        bytesPerEntry = 1 << (((flags >> 4) & 0xf) + 1)
        for _ in range(0,lctsize):
            r,g,b = [int.from_bytes(rng.read(bytesPerEntry),'big')
                     for j in range(0,3)]
    
            out.write(r.to_bytes(bytesPerEntry)); size -= bytesPerEntry  
            out.write(g.to_bytes(bytesPerEntry)) ; size -= bytesPerEntry   
            out.write(b.to_bytes(bytesPerEntry)) ; size -= bytesPerEntry   
            
        # LZW minimum code size
        mincodesize = int.from_bytes(rng.read(1),'big')        
        out.write(mincodesize.to_bytes(1,'big'))    ; size -= 1   
                
        # LZW encoded pixels
        remainingSize = size
        while remainingSize > 0:
            datum = int.from_bytes(rng.read(1),'big')            
            out.write(datum.to_bytes(1,'big'))

            # Check whether this is a clear code
            if datum == 0: continue
            
            # Clear bit indicates the end of a run
            for j in reversed(range(0,8)):
                if datum & (1 << j) == 0: break
                
            if j >= mincodesize: # End of a run
                out.write((0).to_bytes(1,'big')); remainingSize += -1

                bytesToWrite = min(j+1,remainingSize)
                assert bytesToWrite == remainingSize
                
                # Write as many bytes as possible until we get under
                # the threshold
                for k in range(0,bytesToWrite): 
                    out.write(
                            rng.read(max(1,(remainingSize+255)//256)))
                    
                remainingSize = 0
            else:                 # New code introduced
                # We need one more bit for our value
                value = (datum & ~(-1 << (j+1))) << 8
                mask = (-1) << (8-(j+1))
                shift = 0
                while mask & datum!= 0:
                    out.write(value.to_bytes(1,'big'))

                    datum >>= (j+1)
                    value |= (datum & ((mask ^ (-1)) | shift)) << 8
                    shift = (shift + (j+1)) & 0xff
                    mask <<= (j+1)
                    
                    remainingSize = max(0,remainingSize-1)
            
    return True