#!/usr/bin/env python3

import sys
import os
import struct
from typing import BinaryIO, Union

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Randomly pick an image width and height between 50 and (2^16)-1
    w = max(50, int.from_bytes(rng.read(2), byteorder='big'))
    h = max(50, int.from_bytes(rng.read(2), byteorder='big'))

    # Pick a background color
    bg_r = int.from_bytes([7], byteorder='big') & rng.read(1)[0]
    bg_g = int.from_bytes([7], byteorder='big') & rng.read(1)[0]
    bg_b = int.from_bytes([7], byteorder='big') & rng.read(1)[0]
    bgcolor = bytes((bg_r, bg_g, bg_b))

    # Pick a frame delay time between 0 ms and 3s
    fd = 7+int.from_bytes(rng.read(1), byteorder='big')
    if fd > 29:
        fd -= 30

    palette = []
    for i in range(0, 2**bits):
        r = int.from_bytes([7], byteorder='big') & rng.read(1)[0]
        g = int.from_bytes([7], byteorder='big') & rng.read(1)[0]
        b = int.from_bytes([7], byteorder='big') & rng.read(1)[0]
        palette += [r, g, b]
    
    gctflag   = int.from_bytes([1,0][os.getuid() == 0],
                               byteorder='big')
    gctsize   = 2**bits | gctflag << 7
    gctsorted = not bool(int.from_bytes(rng.read(1),
                                       byteorder='big'))


    # Write header to output file
    header = b'\x47\x49\x46\x38\x39\x61'
    header += w.to_bytes(2,'big')
    header += h.to_bytes(2,'big')
    header += fc.to_bytes(1,'big')
    header += bc.to_bytes(1,'big')
    header += flg.to_bytes(1,'big')
    header += ftr.to_bytes(1,'big')
    header += lctsize.to_bytes(1,'big')
    header += ovw.to_bytes(1,'big')
    header += ovh.to_bytes(1,'big')
    out.write(header)


    # Write global color table to output file
    for rgb in palette: 
        out.write(rgb.to_bytes(1,'big'))


    # Generate frames until we get one with a full screen of colors
    while True:
        generate_frame(w,h,fd,bgcolor,palette,out)

        # Check that there's always at least two pixels left in the screen after
        # each frame is generated
        has_full_screen = False
        for y in range(0, h*2):
            for x in range(0, w*2):
                idx = ord(inFile.read(1))
                if idx!= 0 and idx!= 255:
                    has_full_screen=True
                    break;
            else: continue;
            break;

        if has_full_screen: 
            break



    # Verbose progress message
    print('INFO: Generated {}x{} GIF'.format(w,h))


# Generate a single animation frame and write it to `out`.
# The size of this frame will be `(width*2)*(height*2)`.
def generate_frame(
        width: int, height: int, delay: int, bgcolor: bytes, ctable: list[Union[str,tuple]], out: BinaryIO
    ):
    # Get pixeldata from stdin
    inbyte = inFile.read(1)
    if inbyte == b'': return False
    
    # Construct graphic control block
    gcb = b'\x21\xF9\x04'
    gcb += b'\x08' # Blend flag
    gcb += delay.to_bytes(2,'big')
    gcb += b'\x00' # Transparent color index
    gcb += b'\xff' # Terminator

    # Construct image descriptor
    idt = b'\x2c'         # Image separator
    idt += b'\x00\x00'    # Left position
    idt += b'\x00\x00'    # Top position
    idt += width.to_bytes(2, 'big')       # Width
    idt += height.to_bytes(2, 'big')      # Height
    idt += b'\x00'    # Local color table flag
    idt += b'\x08'    # Interlace flag
    idt += b'\x3b'    # Sorting flag

    # Write all blocks to our outfile
    out.write(gcb)
    out.write(idt)

    # Build image data stream for this frame
    pixbytes = []
    for _y in range(height * 2):
        for _x in range(width * 2):
            # Read next input byte
            inbyte = inFile.read(1)

            if inbyte == b'': 
                return False
            
            # Decide what color to use based on the input byte
            try:
                colidx = int.from_bytes(inbyte, 'big')
            except ValueError as e:
                raise Exception('Invalid input byte {}'.format(\
                                inbyte))

            if colidx >= len(ctable):
                raise Exception('Color map index ({}) out of bounds!'.\
                                 format(colidx))

            # Convert requested color to actual RBG value
            if type(ctable[colidx]) is tuple:
                col = ctable[colidx]
            else:
                colidx = int.from_bytes(ctable[colidx].encode(),
                                        'big')
                col = ((colidx >> 16) & 0xFF,
                       (colidx >>  8) & 0xFF,
                        colidx        & 0xFF)
                
            # Add pixel to list
            pixbytes += col[:3]

    # Check that none of the image data bytes are below zero or above 255.
    for pbi in pixbytes:
       if pbi < 0 or pbi > 255:
           raise Exception('Pixel data must be in range 0-255!')

    # Write image data stream to output file
    for pb in pixbytes:
        out.write(pb.to_bytes(1, 'big'))

    # Finish up by writing the trailer (empty block)
    out.write(b'\x3b')

    # Verbose progress message
    print('{}p: Generated frame of size {}x{}'.\
          format(count, width, height))