#!/usr/bin/env python3

import random
import os
import sys
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # Maximum size in bytes of the LZW dictionary
    MAXDICTSIZE = 4096

    def mask(b): return (b & 1)*255 + ((~b)&1)*(256-255)
    
    # The global color table is 7 entries long
    colors = [mask(i) for i in range(7)]


    def packbits():
        '''
        Packs bits from the `bits` list to a byte and appends it 
        to the second list of integers (`data`).

        See https://en.wikipedia.org/wiki/LZW
        '''
        
        if len(bits) == 0: return False
        while True:
            try:
                c = bits[0]
                del bits[0]
                b1 = min(c//2, 255)
                b2 = max(c - b1*2, 0)

                data[-1] |= b1 << curbitcount
                curbitcount += 8
                
                if curbitcount >= 8:
                    data.append(b2)
                    curbitcount -= 8
                    
                continue
            except IndexError as e:
                break
        return True
    

    def packcolors():
        '''
        Appends each color to an integer in `data`.
        '''
    
        for col in colors:
            data.append((col >> 16) & 255)
            data.append((col >>  8) & 255)
            data.append((col      ) & 255)
    
    
    def compresslzw() -> bool:
        '''
        Compresses the data by performing Lempel-Ziv-Welch encoding on the input
        stream.
        Returns false iff we are fully deterministic and have no more output.
        '''
        
        # lzw dict stores tuples (len+code)
        # where len is the length of the prefix string matching this code
        lzwdict = [{'': ''}] * MAXDICTSIZE 

        curprefix = ''
        foundcode = False
        curcode = 258
        lastchar = '\x00\x00\x00\x00'

        global bits
        n = 1
        while not foundcode or len(curprefix) > 0:
            nextchar = ''

            if curprefix!= '':
                nextchar = curprefix[:n]
            
            else:    
                nextchar = str(lastchar)
            
            if nextchar[-1:]!= '\x00\x00\x00\x00':
                foundcode = False
                n += 1
                continue        

            nextlen = len(nextchar)
            
            try: 
                cut = lzwdict.index({'len': nextlen, 'code': nextchar})
            except ValueError as e:
                # Not previously seen character; add entry to dict
                lzwdict[curcode+1]['len'] = nextlen
                lzwdict[curcode+1]['code'] = nextchar
                curprefix = nextchar
                curcode += 1
                assert curcode < MAXDICTSIZE
                continue
        
            foundcode = True

            if cut == 256:
                # New input word was empty
                cut = 0
                lastchar = nextchar
            
            elif n <= nextlen:
                readchars = ''.join([
                        lzwdict[cut]['code'][i] 
                        for i in range(n)])
        
                n = n + 1
            
            
            else:
                readchars = lzwdict[cut]['code']
                n = 1
                old = cut


                while nextchar!= lzwdict[old]['code']:
                    
                    
                    if lzwdict[old]['len'] == MAXDICTSIZE:
                        print('Reached maximum dictionary size! Aborting.',
                              file=sys.stderr)
                        sys.exit(-1)

                    newlen = lzwdict[old]['len'] + 1
                    news = lzwdict[old]['code'][:-1]
                    news = news + lzwdict[old]['code'][-1]
                    lzwdict[old]['len'] = newlen
                    lzwdict[old]['code'] = news

                    old = cut
                
                    cut = lzwdict.index({'len': newlen, 'code': news})
                lastchar = lzwdict[cut]['code']
                assert n <= nextlen


            bits.extend([(cut-258)+256])


    data = []
    bits = []
    curbitcount = 0

    colorused = set([])

    width = int.from_bytes(rng.read(1), 'little', signed=False) 
    height = int.from_bytes(rng.read(1), 'little', signed=False)

    # Global Color Table
    gctsize = int.from_bytes(rng.read(1), 'little')
    packcolors()
    for _ in range(gctsize-7):
        r = int.from_bytes(rng.read(1), 'little')
        g = int.from_bytes(rng.read(1), 'little')
        b = int.from_bytes(rng.read(1), 'little')
        colormask = mask(r) | mask(g) | mask(b)
        colors.append(colormask)
        colorused.add(colormask)

    # Image Descriptor
    left = int.from_bytes(rng.read(2), 'little')
    top = int.from_bytes(rng.read(2), 'little')
    imagewidth = int.from_bytes(rng.read(2), 'little')
    imageheight = int.from_bytes(rng.read(2), 'little')
    localcolorsize = int.from_bytes(rng.read(1))

    lzwmincodesize = int.from_bytes(rng.read(1))
    # TODO : Handle minimum code sizes greater than 12
    assert lzwmincodesize <= 12
    # Add null terminator
    data.append('\x00\x00\x00\x00')
    # Add clear code
    data.append('\xff\xff')
    # Add end code
    data.append('\x00\x00\x00\x01')
    # Add the first block 0x00
    data.append(0x00)

    # Read the whole image
    bitmap = rng.read(imagewidth*(imageheight%2**16))
    packedbitmap = []

    # First bit is 1 means pixel is set
    # First bit is 0 means pixel is unset
    for byte in bitmap:
        for bit in range(8):
            val = ord(byte) >> bit & 1
            packedbitmap.append(val)


    # Write the LZW encoded data
    compresslzw()

    # Set up the final part of the LZW compression format
    # We need to pad one more time
    finish = b'\x00\x00\x00\x00'
    finish += b'\xff\xfd\x00\x00\xff\xfe'

    # Append our final words
    while len(finish) % 8 > 0:               finish += b'\x00'
    data.append(int(''.join(['{0:08b}'.format(ord(f))
                            for f in finish]), 
                    2).to_bytes(len(finish)//8,'little'))

    # Append the ending character
    data.append(0x01)
    # Pad with zeroes until we reach 256 byte blocks
    while len(data)%256:                    data.append(0)     

    # Make sure that all colors present in the used GCT
    # are actually present in the actual palette
    for el in colorused:
        if colors.count(el) == 0:
            print("Color {} not present after all...".format(hex(el)),
                  file=sys.stderr)
            sys.exit(-1)

    gif = b''    
    # Header
    gif += "GIF89a".encode('ascii')
    gif += (len(data)).to_bytes(2, 'little')
    gif += b''.join([d.to_bytes(1, 'little') for d in data])
    gif += b';'

    out.write(gif)


