#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    # header is always fixed size 
    header = b'GIF89a'
    out.write(header) 

    # Next three bytes are Image Descriptor, which
    # contains width and height information. This info
    # comes from /dev/urandom as well.
    # Note that these values cannot be >65535 pixels.

    # Make sure we don't have zero sized images since some programs seem to crash
    # on those upon loading.
    maxwidth = min((int.from_bytes(rng.read(3),'big')&0xff)+1, 256)
    maxheight=min((int.from_bytes(rng.read(3),'big')&0xff)+1, 256)
    out.write((maxwidth).to_bytes(3,'big'))
    out.write((maxheight).to_bytes(3,'big'))

    # We can choose either truecolor or palette based images; this information
    # also comes from /dev/urandom
    hascolortable=(int.from_bytes(rng.read(1),'big')&0x80)>>7
    out.write(((hascolortable<<7)|0).to_bytes(1,'big'))

    #  Bit flags indicate to skip certain parts of the image descriptor. Since 
    #  we already know what width and height our image will be, there's no point
    #  in skipping it later down the line when writing pixel information.
    numinterlaced=bool(int.from_bytes(rng.read(1),'big')&0x40)
    lefttoright= bool(int.from_bytes(rng.read(1),'big')&0x20)
    interlflag   =(int.from_bytes(rng.read(1),'big')&0x80) | ((numinterlaced)<<4) | ((lefttoright)<<5);
    out.write(((interlflag)<<8).to_bytes(1,'big'))

    # As specified by the dissertation, the meaning of the background color flag
    # depends on whether or not a color table is used. If the image does use a
    # colortable, setting this flag means the pixel at position [0][0] should be
    # ignored because its value doesn't matter.
    bgcolorused = bool(int.from_bytes(rng.read(1),'big') & 0x10)
    out.write(((bgcolorused << 4)).to_bytes(1,'big'))

    # The rest of the header is filled with whatever garbage we've got left over
    # after reading these last five bytes. Call that garbage the global colour map.
    globalcolormapsize= int.from_bytes(rng.read(1),byteorder='big')
    out.write((globalcolormapsize).to_bytes(1,'big'))
    for _ in range(pow(2,globalcolormapsize+1)):
        val=int.from_bytes(rng.read(3),'big')
        out.write((val).to_bytes(3,'big'))

    # Finally, let's get to the fun stuff! Pixel data is written bitwise. One byte
    # represents eight bits worth of luminosity values -- black is represented by a
    # 0 and white is represented by a 255. Each byte ends up representing 2^8 
    # different colours. A full byte therefore means we can only handle an 
    # intensity value between 0 and 255. Therefore, given a set of two jpg pixels,
    # the first pixel's green channel information will be lost. To compensate, a
    # new format is introduced here wherein we follow each RGB triplet with another
    # byte containing all luminosity information instead of just one. For example,
    # the string 00FF00FF represents an RBG pixel whose red component was 0 and green
    # component was 255 while the blue component was 255. In this system, a darker
    # shade of gray would result in a bit being set high in the fourth nibble.
    # Likewise, brighter shades would represent low bits.
    # The way this works is you treat your pixel data as it were a sequence of bytes
    # followed by its complement in reverse order, and convert that information into chunks
    # of eight bits. Accessing the pairs in order ensures the correct ordering of the pixel
    # data.

    # There's still a bug somewhere I need to iron out but for now, here goes...
    raise Exception("Somehow, I didn't implement this correctly.")