#!/usr/local/bin/python3
"""Generate a random GIF file."""

import argparse
import os
import random
import sys


def main():
  args = parse_args()

  with open('/dev/urandom', 'rb') as urandom:
    with open(args.output, 'wb') as dest:
      generate_random_gif(urandom, dest)


def parse_args():
  parser = argparse.ArgumentParser(description='Generates a random GIF file.')
  parser.add_argument('--seed',
                      '-s',
                      help=('Random seed. If unspecified, uses entropy '
                            '(e.g. /dev/urandom). See "man 3 srandom"'),
                      default=None)
    
  parser.add_argument('--output', 
                      '-o', 
                      help="Output filename", 
                      default="rand.gif")
  
  return parser.parse_args()


def seed_rng(rng, seed):
  """Seeds the RNG given an optional string or int seed"""
  if isinstance(seed, str):
    rng.seed(os.urandom(len(seed)))
    rng.seed(seed)
  elif isinstance(seed, bytes):
    rng.seed(seed)
  else:
    print("Error: invalid seed:", seed, file=sys.stderr)
    exit(-1);


def generate_random_gif(rng, out):
  """Generates a random GIF file with a given binary random number generator.

  Args:
    rng: A binary random number generator supporting.read().
    out: An open file handle that supports writing. The returned GIF will be
         written here.
  """  
  # Generate the header and add to output
  header = b'GIF89a'
  out.write(header)

  # Get the logical screen descriptor (LSD) field
  lsd = rng.read(7)
  
  # Add LSD to output
  out.write(lsd)

  # Read global color table length from LSD
  gctlen = pow(2, (lsd[4] >> 5)) * 3;

  # Add GCT to output if we have it
  if gctlen > 0:
    gct = rng.read(gctlen)
    out.write(gct)
  
  while True:
    # Read next field byte for block type
    fieldbyte = chr(ord(rng.read(1)));
    
    # Check if this is the last block
    if (fieldbyte[-1] == '\x00'):
      break
      
    # Get size of block
    size = ord(rng.read(1));

    # Check if there's any image data in this block
    hasimgdata = ((fieldbyte[-1] >> 5) % 3) is not 0; 

    # Write fields back to stream
    out.write(struct.pack('=b', ord(fieldbyte[:-1]) | 2));
    out.write(struct.pack('=b', ord(chr(size))));

    # Handle case where there's no image data
    if not hasimgdata:
      continue

    imgdata = rng.read(size)
    out.write(imgdata)


if __name__ == '__main__':
  main()