#!/usr/bin/env python3

import struct
from math import ceil
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bool:
    try:
        # Generate the header and add to output
        header = b'GIF89a'
        out.write(header)

        # Get the logical screen descriptor (LSD) field
        lsd = rng.read(7)

        # Add LSD to output
        out.write(lsd)

        # Read global color table length from LSD
        gctlen = int(2**(int((lsd[4] >> 5)))) * 3

        # Add GCT to output if we have it
        if gctlen > 0:
          gct = rng.read(gctlen)
          out.write(gct)
        
        # Make sure we're starting fresh with each frame
        bufsize = out.tell()+1<<16
        bufposition = out.tell()
        prevframe = bytearray()
        while not out.closed:
            # Create new frame        
            out.write(b'\x2c')
            
            # Create image descriptor
            id = rng.read(9)
            out.write(id)

            # Get number of bytes to read based on bits in ID
            imgdatalenbits = id[8] & 0x7
            imgdatalen = ((id[4] & 0x7F) << 8 |
                          (id[5] & 0xFF)     )
            codelen = [(id[6] & 0x7),
                       (id[6] & 0x7)]
                    
            # Get dispose method, if any
            dispose = -1 if (id[8]&8)==0 else id[8]&7

            if len(prevframe)!= 0 and \
               imgdatalen < len(prevframe) and \
               (dispose==-1 or dispose<=2):
                out.write(prevframe[:imgdatalen])

            elif dispose == 2:
                out.write(prevframe[-imgdatalen:])

            else:
                out.write(bytearray(b'\0'*imgdatalen))


            # If we don't yet know how much image data we need per line,
            # get that info now by reading the first row of pixels
            if codelen[1] == -1:
                out.seek(-1, 1)

                # If we've already generated an entire line, just use that data
                if imgdatalenbits <= 6:
                    codelen[1] = 2**imgdatalenbits
                # Otherwise, generate more data until this many bytes are
                # available
                else:
                    while True:
                        n = 1 << (imgdatalenbits-6)

                        x, y = ord(rng.read(1)), ord(rng.read(1))
                        
                        if x >= n or y >= n:
                            break

                    codelen[1] = n
                    
                out.seek(1, 1)

            # Now create the actual pixel data
            colors = {}
            
            data = bytearray()
            dpos = 0
            while dpos < imgdatalen:
                pos = int(dpos/(codelen[1]*2))
                
                clen = (colors.setdefault(ord(rng.read(1)),
                                           len(colors)))
                data += bytearray([(clen>>8)&255,
                                   clen&255]
                                  )*clen
                dpos += clen*codelen[1]*2
                

            out.write(bytearray([codelen[0]]*ceil(imgdatalen/(
                                         codelen[1]*2))
                                ))
            out.write(data)

            # Save as previous frame because we might need it later
            prevframe = out.getbuffer()[bufposition:].tobytes()
            yield True

    except IOError as e:
        print('Error:'+ str(e), flush=True)
        return False