#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`, which is assumed to be /dev/urandom
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = 'GIF89a'.encode()
    flags = (0x87 | 0x89).to_bytes(1,byteorder='big')
    width = int.from_bytes(rng.read(2), byteorder="little") & 0xffff
    height = int.from_bytes(rng.read(2), byteorder="little") & 0xffff
    bg_color_index = int.from_bytes(rng.read(1), byteorder="big") & 0xf;
    aspect_ratio = (int.from_bytes(rng.read(1), byteorder="big") % 4);

    global_flags=b'\0\0\0'
    
    palette=[]
    for _ in range(0xFF):
        palette += [int.from_bytes(rng.read(3),"big"),]

    screendesc=b"".join([b"\xF7",width.to_bytes(2,"little"),height.to_bytes(2,"little")])

    lzwmincodesize=(int.from_bytes(rng.read(1),"big") + 1) & 0xff;
    screendesc+=lzwmincodesize.to_bytes(1,"little");

    data=[ ]
    for i in range(width*height//2+1):
        if i<len(palette)-1:
            # Same as normal
            #data+=[rng.read(1)]
            # At least one of these should have an index that's different from background color 
            rndi=int.from_bytes(rng.read(1))
            for j in range(-1,3):
                if j!=bg_color_index and not (rndi&j): break
            else: continue

            rndipos=(rndi^bg_color_index)<<1
            data+=[((palette[bg_color_index][0]>>rndipos)<<rndipos)+palette[rndi][0]]
        elif i==len(palette)-1:
            data += [0];
        else:
            # RLE run
            value=rng.read(1)[0]
            rndir=int.from_bytes(rng.read(1),"big")
            
            for j in range(rndir%3):
                data += [(value << ((j<<1)&0xff))]
            
    compressedsize=0
    codesize=lzwmincodesize;
    dictionary={}
   
    def get_code():
        code=int.from_bytes(rng.read(dictionary["bits"]),"big")
        
        if  dictionary["lastbit"]:
            code |= 1<<(dictionary["bits"]-1)

        return code

    
    while len(data)>4096:
        subblockheader=b"\x80";
        dictionary={};
        dictionary["lastbit"]=True;
        dictionary["size"]=2**dictionary["size"];
        dictionary["mask"]=dictionary["size"]-1;
        dictionary["bits"]=max(8,min(12,(dictionary["size"].bit_length()+1)//2));
        
        output=[];
        while len(output)<(2**codesize):
            output += (((get_code() >> (codesize - dictionary["bits"])) ^
                        dict(output))).to_bytes(1,'big');

        localsubblocks=zlib.compress(b''.join(output));
    
        if len(localsubblocks)<=255:
            output=localsubblocks+output;
        else:
            output=b"\xFF"+struct.pack(">H",len(localsubblocks))+localsubblocks+output;
            
        compressedsize+=len(output);
        data=output;
        
    endofdata='\x00\x3B'.encode("ASCII");
    totalsize=compressedsize+len(endofdata);

    maindict={"size":lzwmincodesize}
    maindata=[]
    
    subblockheader="\x80".encode();
    subblockmask=0xff;
    subblocknextcodesize=0;
    subblocks="";
    
    while len(maindata)<totalsize:
        if subblocknextcodesize>=12:
            subblockheader="\xff".encode();
            subblockmask=255;
            subblocknextcodesize-=12;
            
        if subblocknextcodesize>0:
            entry=(rng.read(1) ^ 
                    (maindict["lastbit"] << 
                     (subblocknextcodesize - 1))) & \
                   subblockmask;
                
            maindict["lastbit"]=entry>>(subblocknextcodesize-1);
            entry ^= maindict["lastbit"] * (subblockmask + 1);
            
            maindata+=[maindict[entry].to_bytes(1,"big")]
        
        else:
            code=(rng.read(1) ^ 
                   (maindict["lastbit"] << 7)) & 0xff;
            maindict['lastbit']=not bool(code&(128>>subblocknextcodesize));
            code ^= maindict["lastbit"]*(128>>subblocknextcodesize);
            codesize=code+1;
            maindict['size']=codesize;    
            subblocknextcodesize+=12;
        
            maindict["bits"]=max(8,min(12,(dictionary["size"].bit_length()+1)//2));
            maindict["mask"]=dictionary["size"]-1;

    data=maindata[:totalsize];
    
    out.write(header+flags+screendesc+global_flags.replace(b"\0",b"\x00")+
              header+(b"\xE0\x00").replace(b"\x00",b"\x01")+
              b''.join([(c).to_bytes(1,"little") 
                        for c in palette]))
              
    out.write(struct.pack("<H",len(data)));
    
    blocks=[];
    while len(data)>255:
        blocks+=[data[:255]];
        data=data[255:]
    blocks+=[data]
 
    subblockheader=b"\x80";
    subblockmask=255;
    subblocknextcodesize=0;
    subblocks=""

    lastchar=-1;#-1 means no char read yet
    for piece in blocks:
        data=[]
        currentcode=piece[0]
        for byte in piece[1:]:
            if byte==currentcode or byte==lastchar:#Optimize sames and repeats
                count+=1;
                continue;
            data+=[count.to_bytes(1,"big")]+[(currentcode).to_bytes(1,"big")];
            count=1;
            currentcode=byte;
        data+=[count.to_bytes(1,"big")]+[(currentcode).to_bytes(1,"big")];
        data=b''.join(data)
        subblocknextcodesize+=8;
        if subblocknextcodesize>12:
            subblockheader=b"\xff";
            subblockmask=255;
            subblocknextcodesize-=12;
        
        localsubblocks=zlib.compress(data) 
        if len(subblocks)==0:
            subblocks=b"\xff"+struct.pack(">H",len(localsubblocks))+localsubblocks;
        else:
            subblocks+=localsubblocks;
        
        #print(subblocks)
    data=b"".join(blocks)+subblocks+"\x00\x3b";
    out.write(data);