#!/usr/bin/env python3

from contextlib import closing
import struct
import tempfile
from typing import BinaryIO


class GifImage:
    """A class representing a gif image"""
    
    def __init__(self,
                 width=None,
                 height=None,
                 palette_size=256,
                 loops=0,
                 duration=0,
                 bg_colour=(0, 0, 0),
                 fg_colour=(255, 255, 255)) -> None:
        
        self._palette_size = palette_size
        self._loops = loops
        self._duration = duration
        self._bg_colour = tuple(bg_colour)[:3]
        self._fg_colour = tuple(fg_colour)[:3]
        
        self._frames = []
        
        self._initialise_image(
                width=width, 
                height=height)
        
        
    @classmethod
    def from_file(cls, path):
        """Constructs a GifImage object from a gif file on disk."""

        with open(path, 'rb') as f:
            return cls.from_stream(f)
        
        
    @classmethod
    def from_stream(cls, stream):  
        """Constructs a GifImage object from a binary stream."""
  
        img = cls()
        
        curr_frame = 0
        
        while True:
        
            try :
                block_type = next(stream).decode("utf-8").strip('\0')
                
            except UnicodeDecodeError:
                raise ValueError("Invalid GIF image.")
            
            
            if block_type == '!':
                
                # extension block
                
                ext_block_size = struct.unpack('<H', stream.read(2))[0]
                
                extension_id = stream.read(1)[0]

                if extension_id == 0xF9:
                    
                    # graphics control extension

                    unpacked = struct.unpack("<BBHB",
                                             stream.read(6))
                    img._duration = unpacked[2]
                    img._loops = unpacked[3]
                    
                elif extension_id == 0xFF:
                    
                    # application extension

                    app_ext_label = stream.read(1)[0]
                    
                    ext_str = ""
                    
                    c = ''
                    while ord(c)!=0 and len(ext_str)<ext_block_size+1:
                        ext_str += str(ord(c))
                        c = chr(int.from_bytes(stream.read(1),'big'))
                        
                    if app_ext_label == 0x87 and \
                            ext_str[-3:] == 'NETSCAPE2.0':
                        
                        packed = struct.pack("<BBBBB",
                                             ord('N'),
                                             ord('E'),
                                             ord('T'),
                                             ord('S'),
                                             ord('C')) 
                        stream.read(1)
                        if packed == ext_str[:-3]:
                            
                            sub_block_size = \
                                    struct.unpack("<B",
                                                   stream.read(1))[0]                               

                            if sub_block_size >= 3:
                                flags = stream.read(sub_block_size)[0]
                                
                                loop_count = stream.read(1)[0]
                                img._loops = ((loop_count % 2**8) << 8) + \
                                           (loop_count % 2**8)
                                sub_block_rest = sub_block_size - 3
                                
                                if flags!= 1 and not sub_block_rest==0: 
                                    pass

            elif block_type == ',': # image descriptor

                left, top, width, height = \
                        struct.unpack("<HHHH",
                                     stream.read(8))
                
                local_color_table_flag,\
                        interlace_flag,\
                        sort_flag,\
                        reserved = \
                        struct.unpack("<BBBB",
                                     stream.read(4))     
                    
                lct_size =  2**(local_color_table_flag+1)      
                size = 3*lct_size
                
                global_color_table=\
                        [struct.unpack("BBB",
                                        stream.read(3))
                         for _ in range(size)]


                # size is now set!            
                
                img._set_global_colourmap(global_color_table)
                
                img._create_new_frame(
                        frame_number=curr_frame,
                        top_left=(top, left),
                        dimensions=(height, width))

                decompressed_data =\
                        list(img._decompress_raw_data(stream))

                img._add_pixels(
                        pixels=decompressed_data,
                        row=len(img._frames)-1)
                
                curr_frame+=1
                
            else:       
                raise ValueError("Invalid GIF image.")
                
        return img
    
    
    def add_frame(self, 
                  pixels)->None:
        """Adds another frame of this animation."""
        
        assert isinstance(pixels[0], list)
        self._ensure_frame_exists()
        self._add_pixels(pixels, len(self._frames)-1)
        
                    
    def remove_frame(self, 
                     index=None)->None:
                 
        num_frames = len(self._frames)
            
        if index is not None:
            del self._frames[index]
            
        elif num_frames > 1:
            del self._frames[-1]
            
            
    def get_next_delay(self) -> int:
        """Returns the delay (in centi seconds) of the next frame."""
        
        return self._duration
    
    
    def save(self, filename, **kwargs):
        """Write this GifImage to a file named filename."""
        
        with open(filename, 'wb') as f:
            self.save_stream(f, **kwargs)
            
        
    def save_stream(self, 
                    target_stream, 
                    **kwargs):
        """Write this GifImage to a binary stream."""
        
        images = []
        
        try:
            from PIL import Image
        except ImportError:
            raise ImportError("PILLOW required!")

        tmpfiles = []
        
        frames = iter(self._frames)
        
        last_frame = self._frames[-1]
        
        reference_image = Image.new(mode="P",
                                    size=(last_frame['dimensions'][0],
                                          last_frame['dimensions'][1]),
                                    color=self._palette_size-1)
                                        
        current_palette = {}
        for i,(r,g,b) in enumerate(reference_image.getcolors()):            
            current_palette[(r, g, b)] = i+1

        reference_image.putpalette([*current_palette.keys()])
        
        
        if kwargs.pop('background','auto')=='transparent'\
                and 'bg_colour' in kwargs:
            kwargs={'transparency':0} 
        else:    
            kwargs={}
            
            
        for i,frame in enumerate(frames):        
            
            im = Image.new(mode="P",
                           size=(frame['dimensions'][0],
                                  frame['dimensions'][1]),
                           color=self._palette_size-1)
                            
            im.putpalette([*current_palette.keys()])           
                             
            for x in range(im.width):
                for y in range(im.height):
                    pixel = frame['pixels'].pixel_at(x,y)
                    
                    im.putpixel((x, y), 
                                current_palette.setdefault(tuple(pixel),
                                                          self._palette_size-1))
                      
                         
            imageio_object =\
                    Image.open(im)\
                   .convert(mode='RGB', **kwargs)


            images.append(imageio_object)
                    
        images[0].save(\
                  fp=target_stream,
                  format='GIF',
                  append_images=images[1:],
                  save_all=True,
                  transparency=kwargs.get('transparency', None),
                  disposal=kwargs.get('disposal', None),
                  optimize=False,
                  loop=self._loops,
                  duration=self._duration,
                  comments=[])