#!/usr/bin/env python3

import collections
import itertools
import struct
from typing import BinaryIO


# Represents a bitmap in RGB565 format
class BitmapRGB565():

    # Constructs a new bitmap from an iterator of pixels
    def __init__(self, size):

        self._size  = tuple(reversed(tuple(size)))

        self._pixels = [[0 for x in range(d)]
                        for d in self._size]


    @property
    def size(self): return self._size


    # Returns the color (r, g, b) at the pixel location (x, y)
    def getpixel(self, pos):
        x, _, y, _ = list(pos)[:4]
        if not any([p >= s or p < 0
                    for p, s in zip((x, y), self.size)]):

            return [
                ((p & 0xF800) >> 11)*255/31,     # r
                (((p & 0x7E0) >> 5))*255/63,   # g
                ((p & 0x1F)) *255/31            # b
                ]

        raise IndexError("Pixel does not exist")


    def putpixel(self, pos, c):
        if len(c)!= 3:
            raise ValueError("Color must be three components long")

        x, _, y, _ = list(pos)[:4]
        p = ((int(round(c[0]*31)/255) << 11) |
             (int(round(c[1]*63)/255) << 5) |
              int(round(c[2]*31)/255) )


        if not any([p >= s or p < 0
                    for p, s in zip((x, y), self.size)]):

            self._pixels[y][x] = p
            return True

        raise IndexError("Pixel does not exist")




class ColorTableBuilder():

    """
        Creates and manages color tables that can be used by LZW encoding.

        A color table is simply a dictionary mapping
        integer keys to tuples corresponding to colors:

        {
            0 : (0xFF, 0x00, 0xFF),    # Blue
            1 : (0xFF, 0xFF, 0x00),    # Green
            2 : (0x00, 0x00, 0x00),    # Black
           ...                      # More colors...
        }

    """

    def __init__(self):
        self.__table = {}



    # Adds a color to be encoded with LZW compression into the color table
    # If no index was provided then one will be automatically assigned
    def add_color(self, rgb, idx=None):
        if idx == None:
            idx = max(self.__table.keys())+1

        elif isinstance(idx, collections.Hashable):
            try:
                v = iter(idx)

                if all([isinstance(_, (int, float, str))
                        for _ in v]):

                    idx = set(itertools.chain(*v))


                else:
                    raise TypeError()


            except TypeError:
                pass



        if not isinstance(rgb, collections.Hashable):
            rgb = frozenset(rgb)


        self.__table[idx] = rgb
        return idx


    # Returns the total number of colors added to this table so far
    def num_colors(self):
        return sum([True for _ in self.__table])


    # Removes all previously-added colors
    def clear(self):
        self.__table.clear()



class LZWEncoder():

    # Makes a new encoder object, which will encode pixels
    # from input stream `source`, into output stream `dest`.
    # `width` and height denote the dimensions of the image being written.
    # `colortable` should be an instance of ColorTableBuilder
    def __init__(self, source, dest, width, height, colortable):

        self._source       = source
        self._dest         = dest
        self._width        = width
        self._height       = height
        self._ctbl         = colortable
        self._next_pix     = lambda : self._next_pixel()
        self._prev_pix     = None

        self._max_code_len = 1
        while pow(2, self._max_code_len+1)-1 <= len(self._ctbl):
            self._max_code_len += 1

        print(f"Maximum code length: {self._max_code_len}")




    # Reads the next pixel from the input stream
    def _next_pixel(self):
        pixdata = bytes(self._source.read(2))
        if len(pixdata) == 0:
            return None


        r, g, b = bytearray(struct.unpack('>H', pixdata))[0]>>8
        return (r,g,b)


    # Reads a series of pixels from the input stream until it has read enough
    # for a full row (i.e., width*height pixels). This method returns a two-
    # element tuple containing the number of pixels actually read as well as
    # the actual row of pixels that were read, each pixel represented as a 3-
    # component RGB tuple.
    def _get_row(self, numpix):
        row = []

        for i in range(numpix):
            pxltr = self._next_pix()

            if pxltr == None:
                break

            row.append(pxltr)

        return row, len(row)


    # Encodes a single row of pixels into the destination buffer
    # The row argument should be a list representing the row;
    # Each entry in the list represents a different pixel.
    def _encode_row(self, row):
        enc    = LZWCodewordEncoder(self._dest)
        bufptr = BitBuffer(self._dest)

        dictsz = len(self._ctbl)

        prev_px = None
        cur_cw  = enc.create_codeword(self._ctbl, 0)

        for pxltr in row:
            cw = enc.create_codeword(self._ctbl,
                                     self._ctbl.add_color(pxltr))

            cur_cw.add_pair(prev_px, cw)
            if len(cur_cw) > self._max_code_len \
               or len(dictsz)+len(enc.dict) > 2**12-2:

                enc.flush_buffer(bufptr)
                enc.reset_dictionary(dictsz)

                cur_cw = enc.create_codeword(self._ctbl,
                                            self._ctbl.add_color(pxltr))

            prev_px = pxltr

        enc.flush_entry(bufptr, cur_cw, prev_px)
        enc.flush_buffer(bufptr)


    # Writes an entire frame worth of compressed pixels
    # into the destination buffer. A frame consists of N rows of pixels
    # where N equals the product of the frame's width and height.
    def _encode_frame(self):
        dictsz = len(self._ctbl)
        enc    = LZWCodewordEncoder(self._dest)
        bufptr = BitBuffer(self._dest)

        # Encode first row
        row, numpix = self._get_row(self._width)
        assert(numpix > 0)

        self._encode_row(row)


        while numpix > 0:
            row, numpix = self._get_row(min(numpix,
                                             self._width*(self._height-1)))
            if len(row) == 0: continue

            self._encode_row(row)

        enc.flush_buffer(bufptr)
        enc.reset_dictionary(dictsz)


    # Actually encodes the contents of the image into the output stream.
    def run(self):
        assert(self._width > 0)
        assert(self._height > 0)

        self._encode_frame()

        while True:
            row, numpix = self._get_row(self._width)
            if numpix == 0:
                break

            self._encode_row(row)





class LZWCodewordEncoder():

    # Initializes a new LZW code word encoder, whose compressed bits will
    # be written to the supplied binary stream
    # `dictsz` denotes the initial size of the codebook
    def __init__(self, deststream):
        self._dict      = dict([(i, chr(i))
                                for i in range(256)])
        self._dest      = deststream
        self._lastkey   = -1


    # Returns the current dictionary associated with this encoder
    @property
    def dict(self): return self._dict


    # Resets the internal dictionary to its original state, including clearing
    # all entries
    # `dictsz` specifies how many colors are initially present in the dictionary
    def reset_dictionary(self, dictsz):
        del self._dict
        self._dict = dict([(i, chr(i))
                           for i in range(dictsz)])

        self._lastkey = -1


    # Flushes the bit buffer to the output stream, writing everything up
    # to the last complete code word found in the buffer
    def flush_buffer(self, buf):
        if buf.bitpos > 3:
            buf.flushbits(1)