#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (e.g., /dev/urandom).  The generated image is always of the same size; it has no global color table and all frames are identical.
#
# Note that this is not intended to be cryptographically secure. For example, if two different RNGs generate the same sequence for pixels in the first frame, those pixels will appear exactly equal. Also note that because there is only one global color table per GIF, this means that the colors used by each frame must also be unique within the global color table. That is, you can't use more than seven distinct colors in your animation's global color table. This also means that the full palette resolution (256 colors) is never actually utilized.
def generate_random_gif(
        rng: BinaryIO, width: int, height: int, out: BinaryIO
    ):

    # Common values for the main image descriptor block byte arrays
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt (pages 10-14)
    GRAPHICS_CONTROL_BLOCK = bytes([
            0x21, 0xF9,         # extension introducer & label
            7,                  # data length
            0x01,               # graphics control label
            0,                   # flags (no transparancy, etc.)
            0,                   # delay x 10ms
            0])                  # transparent color index
    IMAGE_DESCRIPTOR_BLOCK = bytes([
            0x2C,                # image separator
            0,                   # left position
            0,                   # top left position
            width % 256,        # image width low byte
            width // 256,        # image width high byte
            height % 256,       # image height low byte
            height // 256,       # image height high byte
            0,                   # local color table flag (0 = none)
            0                    # interlace flag (0 = no)
            ])

    FRAME_DATA = [b'\xFF\x0B']
    FRAME_DATA += list('{num:<{width}d}'.format(
                num=i, width=len(FRAME_DATA[0])).encode()
            for i in range(height))

    # Construct the binary output based on the above blocks
    out.write(b'GIF89a')                           # GIF signature
    out.write(struct.pack('<H', 10+len(GRAPHICS_CONTROL_BLOCK)+len(IMAGE_DESCRIPTOR_BLOCK)))           # Logical screen descriptor
    out.write((width//256).to_bytes(1,'little'))   # Screen width low byte
    out.write((width%256).to_bytes(1,'little'))    # Screen width high byte
    out.write((height//256).to_bytes(1,'little'))  # Screen height low byte
    out.write((height%256).to_bytes(1,'little'))   # Screen height high byte
    out.write(b'\xf0')                             # Global color table information
    out.write(b'\x00\x00')                         # Background color index
    out.write(b'\xff\xff')                         # Pixel aspect ratio
    out.write(GRAPHICS_CONTROL_BLOCK)             # Graphical control block
    out.write(IMAGE_DESCRIPTOR_BLOCK)             # Image descriptor block
    out.writelines(FRAME_DATA)

