#!/usr/local/bin/python3

import sys
import math
import os
import random as rand
from PIL import Image


# Converts an XPM array to a raw pixel buffer
def xpm_to_raw(xpmarray : str):
  lines = xpmarray.split('\n')

  columncount = int(lines[1][6:])
  rowcount    = len(lines)-2

  striped_p = []

  for i in range(columncount*rowcount):
    striped_p += [[], [], [[]]]
  
  for line in lines[2:-1]:
    vals = list(map(lambda c: 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'.index(c), line[:columncount]))
  
    for i in range(columncount):
      val = vals[i]

      if   val <     3: striped_p[i*3 +    0].append(val)
      elif val >=     3 and val <=    12: striped_p[i*3 +    1].append(val-3)
      else:                   striped_p[i*3 + 1+rand.randint(0, 2)].append(val - 12)

  # Convert to RGB triples
  p = []
  for colset in striped_p:
    maxval = max(colset + [-1])

    r = sum([255*(v>0)*math.sin(-math.pi*2/3*float(v)/maxval) for v in colset])
    g = sum([255*(v>0)*math.cos(-math.pi*2/3*float(v)/maxval) for v in colset])
    b = 127*(maxval==-1)+sum([-64*(v>=0)*(v<=(maxval//2 or 1))*(2+(2*((v&1)==1)-(v&2)//2))*math.sin(-math.pi*2/3*float(v)/maxval) for v in colset])

    p += [(r//3, g//3, b//3)]


  bytespp = 4
  width = columncount
  height = rowcount

  # Output pixels as packed bgra
  rows = []
  for y in reversed(range(height)):
    rowbytes = []
    for x in range(width):
      px = p[(y*width+x) % len(p)]
      
      rowbytes += px[2:]
      rowbytes += px[:2]

    rows += [b'\0'+b'\0\0'.join([(px%256).to_bytes(1,'little') for px in rowbytes])]

  return b''.join(rows)

# Get an indexed image that can be used directly with GIFs
def get_indexed_image():
  # Load the source images
  sand = open("sand_dune.xpm").read()
  grass = open("grass.xpm").read()
  snow = open("snowflake_white.xpm").read()

  # Convert them to raw pixel arrays
  pixels_sand = xpm_to_raw(sand)
  pixels_grass = xpm_to_raw(grass)
  pixels_snow = xpm_to_raw(snow)

  # Merge all of them together into one big pixmap
  pixels = [None]*len(pixels_sand)
  for i,p in enumerate(pixels):
    if   i % len(pixels_grass) < len(pixels_sand):       pixels[i] = pixels_sand[i % len(pixels_sand)]
    elif i % len(pixels_grass) < len(pixels_sand) + len(pixels_grass): pixels[i] = pixels_grass[i % len(pixels_grass)]
    else:                                                  pixels[i] = pixels_snow[i % len(pixels_snow)]

  # Make sure our indices cover all colors in the final product
  assert((set(pixels_sand[:3*len(pixels)])|set(pixels_grass[:3*len(pixels)])) == set(pixels[:3*len(pixels)]))

  # Create an image object
  im = Image.new('RGB', (len(pixels)//3, 1))

  # Iterate through each pixel and assign it an index based on their appearance order in the original array
  for ix, pixel in enumerate(im.getdata()): 
    im.putpixel((ix%(len(pixels)//3), 0), tuple(pixels[3*ix:3*(ix+1)]))

  im.show()

  # After getting the palette computed, convert the image into an indexed format
  # This makes things even more complicated when generating frames because now we actually need to store a 
  # color palette in each frame...
  pal = im.convert('P').palette
  rgbpal = {tuple([pal[i*3+j] for j in range(3)]):i for i in range(1 << pal.bits)}
  

  newpixels = ["\xff"]
  for ix, pixel in enumerate(im.getdata()): 
      newpixels += [rgbpal[pixel]]
    
  print(bytes(newpixels)[1:])



if __name__=='__main__':
  get_indexed_image()