#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generate a random GIF file whose size is between minsize and maxsize
# in bytes, using the random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, minsize: int, maxsize: int, out: BinaryIO
    ) -> None:

    # GIF Header (always uses LZW compression with 256 colors)
    head = b'\x47\x49\x46\x38\x39\x61'
    head += minsize.to_bytes(2, byteorder='big')
    
    # Canvas Width (1-32 pixels; default=8 for compatibility)
    w = rng.randint(1, 32) or 8
    # Canvas Height (1-32 pixels; default=8 for compatibility)
    h = rng.randint(1, 32) or 8
    canvas = f';{w}x{h}'

    # Global Color Table Size (1-256; default=256)
    color_table_size = rng.randint(1, 256) or 256
    palette = ''
    
    if (color_table_size > 2):
        # Palette as hex string
        for i in range((color_table_size - 1)):
          r = rng.randint(0, 255) >> 6 << 5 | rng.randint(0, 255) >> 6 << 2 | rng.randint(0, 255) >> 6
          g = rng.randint(0, 255) >> 5 << 5 | rng.randint(0, 255) >> 5 << 2 | rng.randint(0, 255) >> 5
          b = rng.randint(0, 255) >> 5 << 5 | rng.randint(0, 255) >> 5 << 2 | rng.randint(0, 255) >> 5
          palette += ('{:02x}{:02x}{:02x}'.format(r,g,b))

        # Build global color table
        palette += '{:02x}{:02x}{:02x}'.format(0,0,0)
        palette = ['{:03d};'.format(int('0x'+palette[i:i+3])) for i in range(0,len(palette),3)]
        palette = ''.join(palette)

    else :
      palette = ';0;'

    extensions = ''
    loops = ';0'

    # Image Descriptor
    id = []
    # Left Position of Image (1-32 pixels; default=0)
    left = rng.randint(1, 32) or 0
    id.append(('{};').format(left))
    # Up Position of Image (1-32 pixels; default=0)
    up = rng.randint(1, 32) or 0
    id.append(('{};').format(up))
    # Image Width (1-32 pixels; default=10 for compatibility)
    width = rng.randint(1, 32) or 10
    id.append(('{};').format(width))
    # Image Height (1-32 pixels; default=10 for compatibility)
    height = rng.randint(1, 32) or 10
    id.append(('{};').format(height))
    # Local Color Table Flag (0: no local color table; 1: has local color table)
    id.append(('0;'))
    # Interlace Flag (0: no interlacing; 1: adam7 interlacing)
    id.append(('0;'))
    # Sort Flag (0: not sorted; 1: sorted)
    id.append(('0;'))
    # Reserved Field (must be set to zero)
    id.append(('0;'))
    image_descriptor = ''.join(id) 

    # Data Sub-Block
    # The first byte indicates the code size for the compression algorithm used before encoding.
    # (1-9 bits binary integer followed by the code size value itself)
    while True:
        lzw_code_size = rng.randint(1, 9)
        if ((2**lzw_code_size) >= color_table_size):
            break
        
    data_subblock = ';' + str(lzw_code_size) + ';'

    # Encode lzw code using color values from the global color table previously generated 
    # and lzw code size from previous loop
    if (color_table_size == 2):
        encoded_data = data_subblock + (str(len(data))).encode() * len(data);
    elif (color_table_size > 2):
      encoded_data = palette + data_subblock + encode_lz77(data, color_table_size, lzw_code_size)

    # Write GIF file content
    out.write('{0}{1}{2}{3}{4}{5}\n'.format(head, canvas, palette, extensions, loops, image_descriptor).encode())
    out.write('\x2c'.encode())
    out.write(encoded_data)

def encode_lz77(data: bytes, color_table_size: int, lzw_code_size: int)->bytes:
  encodedData = '';

  dictionary = {}
  dictSize = 2 ** lzw_code_size
  clearCode = 1 << lzw_code_size  
  endOfInformation = clearCode + 1
  prevChar=''
  
  def addDictEntry():
    if (prevChar!= ''):
      entry = prevChar + char
      dictionary[entry] = dictSize
      encodedData += bin(dictionary[entry])[2:].zfill(lzw_code_size)#str((dictionary[entry])).zfill(lzw_code_size)
      dictSize+=1
      
  for char in data: 
      if (char not in dictionary) and (dictSize < color_table_size): 
        addDictEntry()
        
      prevChar = list(filter(lambda x: (x)[0] == char, dictionary))[-1];

      encodedData += bin(dictionary[prevChar + char])[2:]#str(((dictionary[prevChar + char])))

  addDictEntry();
    
  return encodedData