#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (e.g., /dev/urandom).
# The size of the image is given by `width`, `height`.
def generate_random_gif(
        rng: BinaryIO, width: int, height: int, out: BinaryIO
    ) -> None:
        
    def write_logical_screen(): 
        out.write(b'\x21\xF9\x04')    # logical screen descriptor
        out.write((
            0 |     # global color table flag 
            2 << 3 | # each pixel is two bits per color component
            7      # no color resolution specified
        ).to_bytes(1,"little"))

        out.write((
            0 |   # sort flag
            width >> 8 & 255, 
            width & 255
        ))

        out.write((
            0 |  
            height >> 8 & 255, 
            height & 255
        ))
        
        out.write(b"\x00")    # background color index
        out.write(b"\x00")    # pixel aspect ratio

    def write_global_color_table():
        out.write(b'\x21\xFF\x0B')    # application extension
        out.write(b'NETSCAPE2.0')    # app ext label
        out.write(b'\x03\x01')        # data subblock length

        out.write(
            b'2.0'.ljust(16,"\x00").encode("ascii")) # junk data

        out.write(b'\x00')            # end app ext subblocks

    def write_image_descriptor():
        out.write(b'\x2C')    # image descriptor
        out.write((
            0 +                # left
            0 +                # top
            width - 1          | # right
            height - 1         << 4, # bottom
            0                  | # local color table flag
            int(bool(0))       << 7, # interlace flag
            0                  |
            int(bool(0))       << 1, # sorted flag
            0                  |
            0                   |
            0                   << 2
        ).to_bytes(1, "little"))

        out.write((
            0                      | # image lzw min code size
            len(palette)           << 4
        ).to_bytes(1, "little"))


    palette = [
        (int.from_bytes(rng.read(1),"little") % 256)
        for _ in range(2**11)]

    transparency_index = palette[0]

    def write_local_color_table() -> None:
        out.write(b'\x21\xFF\x03')    # application extension
        out.write(b"0")              # app ext label
        out.write(b"\x0C")             # data sub block length
        out.extend([p for p,_ in palette])
        out.write(b"\x00")              # end app ext sub blocks

    def write_graphic_control_extension() -> None:
        out.write(b'\x21\xF9\x04')    # graphic control extension
        out.write((
            0                       |     # disposal method
            int(bool(transparency_index)) << 2 |     # user input
            0                       |
            0                       |
            0                        << 3 |
            len(palette)                     << 4  
        ).to_bytes(1, 'little'))

        out.write((
            100              # delay time
        ).to_bytes(2, 'little'))


        out.write(b"\x00")               # transparent color index
        out.write(b"\x00")               # terminator


    delaytime = 100
    
    # Header
    out.write(b"GIF89a")   
    write_logical_screen() 
    write_global_color_table()
    write_local_color_table()

    # Frames
    write_graphic_control_extension()
    write_image_descriptor();

    for rownum in range(0,height):
        for colnum in range(0, width):
            out.write(palette[(int.from_bytes(rng.read(1),'little') %
                              len(palette))][::-1])

            
    