#!/usr/bin/env python3

import io
import struct
from typing import BinaryIO


def _generate_color():
    return bytes([i * 4 for i in range(3)]) + \
           bytes([(i // 7) if (((i+1)%4==0)) else 200
                 for i in range(63)])


def _get_linebuf() -> bytearray:
    result = bytearray()
    for line in [io.BytesIO(b'\x01' +
                            color + (struct.pack("B", q)))
                 for q in range(256)]:
        result += line.read()

    assert len(result) == 256*11

    return result


class FrameWriter:
    def __init__(self, f: BinaryIO):
        self._f = f

        self._colors = [_generate_color() for x in range(256)]

        self._linebufs = [_get_linebuf(),
                          _get_linebuf()]

    def get_frame_data(self) -> bytes:
        frames = []

        for y in range(0, 150):
            currline = self._linebufs[y&1]

            output = b""
            for x in range(0, 160):
                value = currline[(x//8) |
                                  (y<<3)]

                output += self._colors[value]

            frames.append(output)


        buf = frames[0]
        for frame in frames[1:]:
            buf += frame[:3] + \
                   frame[-2:]

        # Set size of image
        buf = buf[:-2] + \
              len(frames).to_bytes(2, byteorder='big') + \
              buf[-2:]
        
        return buf


    def addline(self, line: str):
        pos = 0
        while pos < len(line):
            value = ord(line[pos])
            
            # Write all but the last two pixels on this scanline
            for px in range(0, 158, 8):
                pixval = (value >> (px%8)) & 0xff
                
                self._linebufs[pos&1][px|pos>>3] &= ~0xff
                self._linebufs[pos&1][px|pos>>3] |= pixval
                    
            self._linebufs[pos&1][-2] = value & 0xf

            self._linebufs[pos&1][-1] = value >> 4
            

            #print(pos,"  ", end="")
            #for c in self._linebufs[pos&1][:].hex():
            #    print("\\x"+c,end="")
            #print("")
            pos += 1
            
        #print("-----\n")
        pass

    def finish(self):
        pass
        



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    
    framewriter = FrameWriter(out)

    framenum = -1

    # Generate the first frame
    framenumber_bytecount = 2

    while True:
        char = chr(rng.read(1)[0])
        if not (char >= '@'):
            break
        elif char == ';':
            framewriter.finish()
            break

        if framenum!= -1 and framenum > 0:
            out.write((len(framewriter.get_frame_data())).to_bytes(2, 'big'))

        framenum+=1

        # Write it as an incremental update to allow
        # us to play back frames.
        framewriter.addline(char)

