#!/usr/bin/env python3

import struct
from math import ceil
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Initialise frame state variables so they can be used inside loops
    bufposition = out.tell()
    lsd = None
    gct = bytearray()
    gctlen = 0
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = bytearray()
        
    def write_block(bstr):
        nonlocal bufposition
        
        bufposition = out.tell()
        out.write(struct.pack('>B', len(bstr)))
        out.write(bstr)
        
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)
    
    try:
        while True:
            # Create new frame        
            out.write(b'\x2c')
            
            # Create image descriptor
            id = rng.read(9)
            out.write(id)

            # Get number of bytes to read based on bits in ID
            imgdatalenbits = int(id[8])&0x7
            imgdatalen = int(id[4])*pow(2,imgdatalenbits+1)+imgdatalenbits+1
            codelen = pow(2,(id[6]&7))+1

            # Get dispose method, if any
            dispose = -1 if (id[8]&8)==0 else id[8]&7

            if len(prevframe)!= 0 and \
               imgdatalen < len(prevframe) and \
               (dispose==-1 or dispose<=2):
                write_block(prevframe[:imgdatalen])

            elif dispose == 2:
                write_block(prevframe[-imgdatalen:])

            else:
                write_block(b'\0'*imgdatalen)
                
                
            # If we don't yet know how much image data we need per line,
            # get that info now by reading the first row of pixels
            if codelen == -1:
                out.seek(-1, 1)

                # If we've already generated an entire line, just use that data
                if imgdatalenbits <= 6:
                    codelen = 2**imgdatalenbits
                # Otherwise, generate more data until this many bytes are available
                else:
                    while True:
                        n = 1 << (imgdatalenbits-6)

                        x, y = ord(rng.read(1)), ord(rng.read(1))
                        
                        if x >= n or y >= n:
                            break

                    codelen = n
                    
                out.seek(1, 1)

            # Now create the actual pixel data
            data = []

            for i in range(ceil(imgdatalen/(codelen*2)-1)):
                clen = rng.read(1)*256 + ord(rng.read(1))
                data += [clen]*clen

            d = struct.pack('>%dB'%len(data), *[codelen]*ceil(imgdatalen/(codelen*2)-1)+list(data))
            write_block(d)

            # Save as previous frame because we might need it later
            prevframe = out.getbuffer()[bufposition:].tobytes()