#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Reads an unsigned 16-bit little endian integer
def read_uint16le(f : BinaryIO) -> int:
  bytes = f.read(2)

  return struct.unpack('<H', bytes)[0];
        

# Reads an unsigned 8-bit integer
def read_byte(f : BinaryIO) -> int:
  bytes = f.read(1)
  
  assert len(bytes) > 0;

  return struct.unpack('<B', bytes)[0];
  
      
# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The generated
# file will be exact match copy of the original specified by
# `inp`. Anything that doesn't need to be changed is copied as
# is. If `loopcount`!= None then the loop count field of the
# header will also be set accordingly.
def generate_random_gif(
        inp: str, 
        rng: BinaryIO,
        out: BinaryIO,
        loopcount: int = None
    ) -> None:
    with open(inp, 'rb') as inpstream:
      # Copy up until first null character (the terminator)
      buf = bytearray()

      while True:
        c = inpstream.read(1);
        if (c == b'\x00'): 
          break;

        buf += c
      
      out.write(buf)

      # Discard the block after the terminator
      inpstream.read(1)

      # Read next field byte for block type
      fieldbyte = read_byte(inpstream) ;
    
      # Set flags bit to indicate there are no errors
      fieldbyte |= 2 

      # Add flags byte now before we start writing other parts
      out.write(struct.pack('=B', fieldbyte));

      # Read next byte which describes the structure of all subsequent blocks
      fieldbyte = read_byte(inpstream);

      # Copy rest of header except for LSD and GCT since these get handled separately
      out.write(struct.pack('=B', fieldbyte));

      # Read Logical Screen Descriptor (LSD)
      lsd = inpstream.read(7) ;

      # Modify color resolution value specially
      lsd = list(lsd)
      lsd[-1] &= ~(1<<0)  # Don't use local color table
      lsd[-1] |= 1<<7      # Global color table exists 
      lsd = bytes(lsd)

      # Add LSD to output
      out.write(lsd)

      # Read global color table length from LSD
      gctlen = 2**(int((lsd[4] >> 5))) * 3

      # Save position so we can skip over the GCT later on
      origpos = inpstream.tell() 
  
      # Add GCT to output if we have it
      if gctlen > 0:
        inpstream.seek(origpos + lsd.__len__())
        gct = inpstream.read(gctlen)
        out.write(gct)
        
      # Skip over the GCT (and I'm assuming the RST we skipped originally)
      inpstream.seek(origpos+lsd.__len__()+gctlen+1)

      # Read first image block…
      while True:
        # Read next field byte for block type
        fieldbyte = read_byte(inpstream) ;
      
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = read_byte(inpstream) 
    
        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Add fields back to stream (but make sure we unflag them again)
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # Handle case where there's no image data
        if not hasimgdata:
          continue

        imgdata = inpstream.read(size)
        out.write(imgdata)