#!/usr/bin/env python3

import collections
import io
import os
import sys
import time

# Generates a random GIF file in memory and returns it as bytes
def generate_random_gif() -> bytes:
    # Read 1024 bits from /dev/urandom
    with open('/dev/urandom', mode='rb') as f:
        header = b'GIF89a'

        datalen_bits = f.read(2)
        datalen = struct.unpack('!H', datalen_bits)[0]
        if not (5 <= datalen <= 65535):
            raise ValueError("Invalid image data length")

        data = f.read(datalen)
        checksum = sum([i for i in bytearray(data)]) % 256

        frames = 7
        delay = 100 // frames

        palette_type = 0x08 | 0x80      # Global color table, sorted, size 256
        gct_length = 7                   # Number of colors in global color table
        background_index = None          # No transparent color
        aspect_ratio = 0                 # Ignored by old GIF viewers
        gct_bits = collections.deque()   # Global color table pixels

        gct_bits.append(palette_type)
        gct_bits.append(gct_length)
        for _ in range(gct_length):     # Initialize to black
            gct_bits.extend((0, 0, 0))

        gct_bits[background_index or 0] |= 0b10000000    # Set color index to opaque

        gct_bytes = [sum([(c << (7 - n)) & 0b10000000
                         for c in reversed(bytearray(gct_bits))])
                     for n in range(8)]
        gct_bytes[-1] |= 0b10000000                        # Last bit is always on

        frame_header = struct.pack('<6sHHBBBHB',
                                   b'GIF89a',           # Signature
                                   0,                  # Width
                                   0,                  # Height
                                   gct_bytes[0],       # Local color table flag
                                   0,                  # Background color index
                                   0,                  # Pixel aspect ratio
                                   len(data),         # Data block length
                                   0                   # LZW minimum code size
                                   )
        subblocks = []
        def append(*args):
            nonlocal subblocks

            subblock = b''
            for arg in args[:-1]:
                if not isinstance(arg, list):
                    arg = [arg]

                subblock += bytes(arg)
            subblock += b'\x00'.join(map(bytes, args[-1]))        # Add LZW dictionary entries
            subblocks.append(subblock)                           # Save for later

        first_frame = True
        for i, d in enumerate(data):
            append([[0], [[gct_bytes[d]], [checksum]]])              # Append local color table entry

            lzwminsize = max(len(data) - ((i+1)*3)+2,
                             min(max(len(data)-(((i+1)*3)), 12), 12)) # Use 12 if too small
            last_byte = -1                                            # Code to repeat until next reset
            while data:                                               # Ensure there are any remaining bytes to encode
                lzwminsize -= 1                                       # Decrement lzwminsize each pass
                chunk = bytearray()                                   # Buffer for encoded data
                while len(chunk) < lzwminsize:                       # Keep encoding till we get enough
                    current_byte = last_byte                         # Save the byte to be repeated
                    current_code = lzwminsize - 1                    # Reduce the code length by one

                    try:                                              # Attempt to read another byte
                        new_byte = data.popleft()                     # Remove it from our sequence
                    except IndexError:                                # If it's empty, we're done
                        break                                        # Break out of this loop

                    code_len = len(chunk)                              # Length of previous chunk
                    next_codes = {}                                    # Absorbed codes
                    try:                                              # Try to access this byte's codes
                        codes = next_codes[current_byte]               # We don't need to create a new dict
                    except KeyError:                                  # if there isn't already one
                        codes = {new_byte: current_code}               # Create an initial dict
                        next_codes[current_byte] = codes                # Store it back so we can use it again
                    else:                                             # Otherwise...
                        try:                                          #...try to find our code
                            code = codes[last_byte]                    # Find the code that was last used
                        except KeyError:                              # When no matching code exists
                            code = current_code                        # It must have been a new code, so reuse it

                        try:                                          # Try to add new_byte to existing codes
                            codes[new_byte] = code                      # Replace the repeating code
                        except KeyError:                              # Otherwise,
                            code = current_code                        # Reuse it instead
                        finally:                                       # Regardless,
                            chunk.append(code)                         # Encode it and continue

                            if len(chunk) == lzwminsize:             # If we've hit lzwMinSize
                                break                                    # Stop compressing now

                    last_byte = new_byte                               # Remember how many times this byte has been encountered

                if first_frame:
                    append(b'\x2C\x00',
                           frame_header[:-3],    # Don't include data size and LZW minimum size
                           subblocks[:2],        # First frame contains only the palette and graphics control extension blocks
                           gct_bytes[1:],        # Skip the background color index
                           )
                    del subblocks[:]
                    first_frame = False
                elif i!= 0:
                    append(b'\x21\xF9\x04', struct.pack('>B', delay))

                append([b'\x21\xFF\x0B', subblocks[1]])                  # Transparency flag, application extension ID, app data size, and subblocks
                del subblocks[:2]                                         # Delete two elements from subblocks; they just contain the trailer

    return b''.join(subblocks)

