#!/usr/bin/env python3

import sys
import os
os.chdir(sys.path[0])

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        palette: bool=True, transparency: bool=False
    ):
    # Write the magic string (header) to the output stream
    out.write(b'GIF89a')

    # Get width and height from user input
    print('Enter image dimensions:')
    w = int(input('Width: '))
    h = int(input('Height: '))
    
    # Write screen descriptor to output stream
    out.write((w).to_bytes(2, byteorder='big'))
    out.write((h).to_bytes(2, byteorder='big'))
    out.write(b'\x87\x00')
    if palette:
        out.write(b'\xff')
    elif transparency:
        out.write(b'\x00')
    else:
        raise ValueError('Neither transparency nor palette flag set.')

    # Generate some global color table entries
    gctlen = 2 << 5
    gct = []
    for i in range(gctlen):
        r = rng.read(1)[0] & ~0xe0 | 0x40
        g = rng.read(1)[0] & ~0xe0 | 0x40
        b = rng.read(1)[0] & ~0xc0 | 0x80
        gct += [r, g, b]

    # Write the global color table to the output stream
    out.write(struct.pack(f'>{gctlen}B', *gct))

    # Start writing frames
    numcolbits = rng.read(1)[0]
    bgidx = rng.read(1)[0]
    pixelaspectratio = 0
    lct = []
    for i in range(numcolbits):
        lct.append(bgidx)
    if not palette:
        transparentcoloridx = None
        for i in range(numcolbits):
            lct.append(transparentcoloridx)
    else:
        for i in range(2 ** numcolbits - 1):
            idx = rng.read(1)[0]
            lct.append(idx)

        # If no transparent color is chosen before this point, choose one now
        if transparentcoloridx is None:
            transparentcoloridx = rng.read(1)[0]
        lct[transparentcoloridx] |= 0x01

    # Write local color table to output stream
    out.write(struct.pack(f'<{len(lct)}B', *lct))

    # Keep track of where the previous frame ended
    lastrectpos = out.tell()

    # Write frames until enough space has been written to the file
    rectlen = ((16 * (((w + 15) // 16))) +
               (16 * ((h + 15) / 16))) * 2
    while out.tell() < out.seek(0, 2) - rectlen:
        # Get duration and left offset
        dur = rng.read(2)
        pos = rng.read(2)

        # Write graphics control block to output stream
        out.write(dur)
        out.write(pos)
        if palette:
            out.write(b'\x0c')
        elif transparency:
            out.write(b'\xf8')
        else:
            raise ValueError('Neither transparency nor palette flag set.')
        
        # Determine size of current frame's rectangle in pixels
        xdim = max(rng.read(2), 1)
        ydim = max(rng.read(2), 1)

        # Switch modes based on whether the next frame should have the same palette or not
        if palette!= rng.readbit():
            palette = not palette

            # Get new list of local colors
            lct = []
            
            # Choose background color index 
            if palette:
                bgidx = rng.read(1)[0]
                lct.append(bgidx)
            else:
                lct.append(-1)

            # Add additional local color indices as necessary
            for _ in range(2 ** numcolbits - 1):
                if palette:
                    idx = rng.read(1)[0]
                else:
                    idx = -1
                lct.append(idx)

            # Write local color table to output stream
            out.write(struct.pack(f'<{len(lct)}B', *lct))

        # Write actual bitmap data
        for i in range(xdim * ydim):
            runlen = rng.read(1)[0] & 0x7F
            if runlen >= 128:
                temp = rng.read(1)[0]
                while temp > 128:
                    runlen += temp & 0x7F
                    temp = rng.read(1)[0]
                out.write(runlen.to_bytes(1, byteorder='big') +
                          temp.to_bytes(1, byteorder='big'))
            else:
                out.write(runlen.to_bytes(1, byteorder='big'))
            alphacoloridx = rng.read(1)[0]
            if transparency:
                if alphacoloridx == transparentcoloridx:
                    alphacoloridx = 0
                else:
                    alphacoloridx |= 0x01
            out.write(alphacoloridx.to_bytes(1, byteorder='big'))
            
    # Repeat the last frame with its own duration and position information
    out.seek(lastrectpos)
    dur = rng.read(2)
    pos = rng.read(2)
    out.write(dur)
    out.write(pos)