#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file of length `datalen`, with
# width and height in pixels given by `width` and `height`.
# The RGB values are generated randomly with /dev/urandom.
# Writes output to `out` as bytes.
def generate_random_gif(
        out: BinaryIO,
        width: int, height: int,
        datalen: int):

    # Write magic string
    out.write('GIF87a'.encode())
    
    # Determine maximum color value for each bit depth
    maxvals = {
        8 : (2 ** 8 - 1),
        4 : (2 ** 4 - 1),
        2 : (2 ** 2 - 1)}
    if any([x >= v for x,v in zip((width, height, datalen),
                                    [2**15, 2**15, 2**24])]):
        raise ValueError("File size too large")

    # Calculate LZW code lengths from minimum code length
    lzwminlen = lambda n: 2 * ((n + 1) // 256) + 1

    # Generate integer for output index calculation
    def idx(i, j, k): return i * width * height +\
                                j * height + k

    # Write palette
    palsize = 3 * maxvals[8]
    out.write(((palsize << 4) | 1).to_bytes(1, byteorder='little'))
    r, g, b = [struct.pack('<H',
                           rng().randint(0,maxval))
               for maxval in maxvals.values()]
    out.write(r * g * b)

    # Write logical screen descriptor
    out.write(width.to_bytes(2, byteorder='little'))
    out.write(height.to_bytes(2, byteorder='little'))
    out.write((lzwminlen(datalen)).to_bytes(1, byteorder='little'))
    out.write(b'\x87')

    # Write global color table structure
    out.write(((maxvals[8] << 4) | 1).to_bytes(1, byteorder='little'))
    out.write(b'\x00')

    # Write image descriptor
    out.write(posX.to_bytes(2, byteorder='little'))
    out.write(posY.to_bytes(2, byteorder='little'))
    out.write(width.to_bytes(2, byteorder='little'))
    out.write(height.to_bytes(2, byteorder='little'))
    out.write((lzwminlen(datalen)).to_bytes(1, byteorder='little'))
    out.write(b'\x87')

    # Define the LZW dictionary
    dictsize = 2**LZWMINLEN
    dict = list(range(dictsize+1))
    dict[0] = None
    dict[-1] = 0

    # Encode input data into LZW stream
    pos = 0
    prev = None
    while pos < datalen:

        # Initialize new buffer for current block
        buf = []
        
        # Iterate until we have enough bytes to encode one datum at least 
        while True:
            
            # Read next chunk from input
            chunk = input[pos:(pos + CHUNKSIZE)]

            # Increase position in input up until it is
            # divisible by CHUNKSIZE
            pos += len(chunk) % CHUNKSIZE

            # Get codes starting with previous character
            try:
                codes = [(prev, dict.index(chunk[:i]))
                        for i in range(1,
                                         min(CHUNKSIZE + 1,
                                             len(chunk)+1))]

                # If there's exactly one code then that's the final result
                if len(codes) == 1:
                    break
                
                # Otherwise iterate through all codes and add them
                # to the buffer; when the buffer exceeds BUFSIZE, or 
                # we've read every possible code, we flush the buffer
                # onto the output.
                else:

                    # Add first two-byte sequence to buffer
                    buf.append(256 + codes[0][1])
                    
                    # Iterate over the rest of the codes and add their
                    # indices to the buffer if they're smaller than 
                    # 256, otherwise add the appropriate multiple of
                    # 256 to the first byte and write the second byte
                    # directly to the buffer.
                    for code in codes[1:-1]:
                        
                        if code[1] > 255:
                            buf.append(code[1] - dictsize)
                        elif code[1] < 256:
                            buf.append(code[0])
                            buf.append(code[1])
                    
                    # Check if last element is larger than 255
                    if codes[-1][1] > 255:

                        # If so, write the proper multiple of 256 to 
                        # the buffer
                        buf.append(codes[-1][1] - dictsize)
                        
                        # Flush buffer to output
                        out.write(bytes([(buf[i] >> 8) & 0xFF
                                        for i in range(len(buf))]))

                    # If not, write it directly onto the end of the buffer
                    else:
                        buf.append(codes[-1][0])
                        buf.append(codes[-1][1])
                    
                    # Flush full buffer to output (if necessary)
                    if len(buf) == 2*BUFSIZE:
                        out.write(bytes([(buf[i]>>8)&0xFF
                                            for i in range(2*BUFSIZE)]))
                        buf = []

            except ValueError: pass

            # If we don't get to here then the loop hasn't been broken
            # yet--so there must be no more codes starting with this 
            # string. So append it to the dictionary. Note that we only
            # store strings up to CHUNKSIZE characters long.
            dict.append(chunk[:CHUNKSIZE])

            # Don't forget to flush what remains on the buffer!
            if len(buf)>0:
                out.write(bytes([(buf[i]>>8)&0xFF
                                     for i in range(len(buf))]))

            # Set previous char to first char of current string
            prev = chunk[0]

            # Keep going if there are still more chunks left
            if pos<datalen: continue
            
            # Or break if everything was written to the buffer already
            break

        # Write code 128 to output to indicate "end of information"
        # This also effectively ends the frame.
        out.write(b'\x80')
        
    # Write empty "trailer"
    out.write(b'\x3B\x00')
    