#!/usr/bin/env python3

import struct
from math import ceil
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). Also returns buffer position at which
# the end of GIF data is located (`endpos`).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> (int, int):
    pos = out.tell()

    # Generate the header and add to output
    out.write(b'GIF89a')

    # Get the logical screen descriptor (LSD) field
    out.write(rng.read(7))

    # Read global color table length from LSD
    gctlen = int(2**(int((lsd[4] >> 5)))) * 3

    def mkcolor(): return tuple(ord(rng.read(1)) for _ in range(3))

    # Add global color table to output
    if gctlen > 0:
        gct = [mkcolor() for _ in range(gctlen)]
        out.write(b"".join(chr(p>>8)+chr(p) for p in sum(gct, ())))
        
    bufposition = out.tell()

    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = False
    
    while True:
        # Create new frame        
        out.write(b'\x2c')
        
        # Create image descriptor
        id = bytearray(rng.read(9))

        codesize = id[8]#&0x7
        id[8]=codesize=min(max(2, codesize), 8)

        # Set local inline flag bit; since we always generate our own GIFs,
        # we can save space by marking this so the decoder doesn't need to 
        # read past the whole thing
        id[8] |= 2

        # Set interlace flag bit
        id[8] &= ~128

        # Write modified image descriptor record back
        out.write(id)
        

        # Get number of bytes to read based on bits in ID
        imgdatalenbits = min(((id[8]//4)+1)*3, 12)
        imgdatalen = int(id[4])*pow(2,imgdatalenbits+1)+imgdatalenbits+1
        codelen = 2**(id[6]&7)+1

        # Get dispose method, if any
        dispose = -1 if (id[8]&8)==0 else id[8]&7

        #print("ID", " ".join("%d" % d for d in id), "imgdatalen:", imgdatalen, "codelen:", codelen)

        if not prevframe or len(prevframe)<imgdatalen or (dispose!=2 and dispose>1):
            out.write("\0"*imgdatalen)
        else:
            data = "\0"*imgdatalen
            # Seek back to beginning of frame data 
            out.seek(-imgdatalen, 1)
            if dispose == 2:
                data = prevframe[-imgdatalen:]
            
            out.write(data)

        # Begin writing image data
        odatapos = out.tell()
        
        # Now create the actual pixel data
        data = bytearray()

        for n in range(0, self.imgdatalen, 2**self.codelen):
            clen = rng.read(1)*256 + ord(rng.read(1))
            data += chr((n//256)%codelen) + chr(n%codelen)
            data += [chr(rng.read(1)) for _ in range(clen)]

        
        out.write(data)

        # Save as previous frame because we might need it later
        prevframe = out.getbuffer()[bufposition:].tobytes()