#!/usr/bin/env python3

import os
import sys
import math
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The width and height are given by 
# `width` and `height`, respectively, but additional options can be provided as keyword arguments.
def generate_random_gif(
    rng: BinaryIO, 
    out: BinaryIO,
    width: int,
    height: int,
    **kwargs
):
    # Supported options with default values
    frame_count = kwargs["frame_count"] or 16
    bgcolor = kwargs["bgcolor"] or [0x4d, 0xb2, 0xf7]
    fgcolor = kwargs["fgcolor"] or [0x9b, 0xdf, 0xff]

    # Validate inputs
    assert isinstance(frame_count, int), "The 'frame_count' option must be an integer"
    assert len(bgcolor)==3 and all([isinstance(c, int) for c in bgcolor]) and min(bgcolor)>=0 and max(bgcolor)<256, f"'bgcolor' value does not make sense ({repr(bgcolor)})"
    assert len(fgcolor)==3 and all([isinstance(c, int) for c in fgcolor]) and min(fgcolor)>=0 and max(fgcolor)<256, f"'fgcolor' value does not make sense ({repr(fgcolor)})"


    # Generate the header and add to output
    header = bytes("GIF89a", encoding="ascii")
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # First byte of LSD determines the size of the global color table
    palette_bitcnt = (lsd[4] >> 5)

    # Global color table length in bits
    gctlen = 2**palette_bitcnt*3

    # Logical screen descriptor flags
    flags = {"has_color_table":bool(((lsd[4]>>7)&1)),
             "sorted_global_color_table":bool(((lsd[5]>>7)&1)), 
             "global_color_table_is_sorted":"| GCT sorted" if flags["has_color_table"] else "",
             "color_resolution_bits":(lsd[7]),
             "has_global_color_table": False}

    # If the application uses its own color table instead of the one defined by the 
    # graphics control extension
    flags["has_local_color_table"]=not flags['has_global_color_table']

    # Set up local color table based on whether it's available or we need it
    flags['local_color_table']=None
    if flags["has_local_color_table"]:
       flags['local_color_table']=[[0x00, 0x00, 0x00],
                                   [0xFF, 0xFF, 0xFF]] 

    # Update LSD with custom colors and write to stream
    bgcolortableindex=(flags['has_local_color_table']+1)*3 if flags['has_color_table']==False else None
    lsd = list(lsd[:4])+[len(flags['local_color_table'])<<5]+[flags["color_resolution_bits"]] + \
          [(lsd[7]&~7)|bgcolortableindex]+list(lsd[8:])
    out.write(bytes(lsd))

    # Add GCT to output if we have it
    if flags["has_global_color_table"]:
      gct = rng.read(gctlen)
      out.write(gct)
      
    # Construct graphic control extension header
    gc_ext = struct.Struct(">BBBBBBHB")
    gc_ext_fields = {
        "blocktype": 0xF9, 
        "blocksize": 4,
        "disposalmethod": 2,   # Background disposal method
        "userinputflag": 0,    # No user input required
        "transparentcoloridx": 0,
        "delaytime": 0,        # No delay
    }
    for key in ("transparentcoloridx","delaytime"):
      val = gc_ext_fields[key]
      if val<0 or val>255:
         raise RuntimeError(f"{val} does not fit in uint8")
     
    gc_ext_packed = bytes(gc_ext.pack(**gc_ext_fields))

    # Construct logical screen descriptor extension header
    ext_screen_descriptor = struct.Struct(">BBBHBBBHH")
    ext_screen_descriptor_fields = {
        "blocktype": 0xFF,              # Application extensions
        "blocksize": 19,                # Block size doesn't include header
        "applicationid": 0x01,          # Graphics Control Extension
        "applicationauthcode": 0x00.     # No auth code used
    }

    # Update with the correct size of the extended headers
    total_gce_length = (max(1, palette_bitcnt)+1)*3+(gc_ext_fields["blocksize"])+1
    ext_screen_descriptor_fields["blocksize"]=total_gce_length-1
    ext_screen_descriptor_packed = bytes(ext_screen_descriptor.pack(*ext_screen_descriptor_fields.values()))

    # Write the updated LSD
    out.write(ext_screen_descriptor_packed)

    # Pad out the rest of the LSD to fit the GCE
    pad_length = total_gce_length - len(lsd)-len(gc_ext_packed)
    out.write(os.urandom(pad_length))
    
    # Finally, write the GCE
    out.write(gc_ext_packed)

    # Construct image descriptor
    img_descriptor = struct.Struct(">BBHHHHBBH")
    img_descriptor_fields={
        "left": 0,
        "top": 0,
        "width": width,
        "height": height,
        "flags": 0x88,     # Has neither local nor global color tables
        "interlace": 0,    
        "sortflag": 0,      # Flag indicates that the image is not sorted
        "reserved": 0       # Reserved flag
    }
    img_desc_packed = bytes(img_descriptor.pack(*img_descriptor_fields.values()))
    out.write(img_desc_packed)

    # Now go through each frame
    for iFrame in range(min(frame_count,1)):
        # Indices of pixels in the current frame
        indices=[]

        # Extract bitmap, if needed
        bitmap=None
        if flags["has_local_color_table"]:
           numcolors = (1 << (flags["color_resolution_bits"]))
           palette_bitcnt = int(math.log(numcolors)/math.log(2))
    
           # Load maximum colors
           color_table = [[0]*3]*numcolors
           for iColor in range(3*(1<<(palette_bitcnt))):
               rgb = tuple(rng.read(1)[0] for _ in range(3))
               idxs = [iColor//3,(iColor%3)+(3*numcolors)]
               color_table[idxs[0]][idxs[1]]=rgb

           # Write color table to stream
           color_table_packed = bytearray()
           for row in range(numcolors):
              for col in range(3):
                color_table_packed+=color_table[row][col].to_bytes(1,'big')           
           out.write(bytes(color_table_packed))
        
        # Write background color index to stream
        out.write(struct.pack('>B',len(indices)%(2**flags["color_resolution_bits"])))

        # Fill the first row
        for x in range(width):
            if flags["has_local_color_table"]:
               index = rng.randint(0,len(color_table)-1)
            else:
               index = rng.randint(0,len(flags["global_color_table"])-1)

            indices.append(index)

        # Write initial rows
        for y in range(height):
            out.write(struct.pack('>B',len(indices)%(2**flags["color_resolution_bits"])))
            for x in range(width):
                out.write(struct.pack('>B', indices[(y)*(width)+(x)]))
                
        # Send the reset command
        out.write(struct.pack('>B', 0))

