from io import BytesIO
from random import randrange


# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom)
def generate_random_gif(
        width=None, height=None, bg_color=None, fg_color=None, num_frames=100, loop=False, duration=50, frame_delay=50,
        seed=42, out=BytesIO()
    ) -> BytesIO:
    if width is None and height is None: # Make it square by default
        sidelen = 1 << randrange(7, 12+1)
        width = sidelen
        height = sidelen

    if bg_color is None: # Default bg color is black
        bg_color = '#000000'

    if fg_color is None: # Default fg color is white
        fg_color = '#ffffff'


    header = [
        71, 73, 70, # Signature
        0x0b, # Version
        0x00, # Color bit depth
        0x00, # Background color index
        0x00 | (int(bg_color[1:], 16)<<4),
        0x00 | ((int(bg_color[3:], 16)>>4)&0x0f) # Foreground color index
    ]

    global palette
    palette = [0] * 4**6 # 256 colors: 0-255
    for i in range(2**14):
        r = randrange(0,1024)
        g = randrange(0,1024)
        b = randrange(0,1024)

        if i < len(palette):
            palette[i] = [r>>4, g>>4]
            palette.append([r&0xf,g&0xf]) # Duplicate every other entry to make it quicker later on; this matches how LZW compression works as well
        else:
            break # Full palette reached

    palette =([[r,g]]*int((0xffff)/2))[:0xfffffff]
    print('Palette:', palette);
    assert all(len(e)==2 for e in palette)
    palette += [[0,0]]*(0xfffffff - len(palette))
    assert all(len(e)==2 for e in palette)

    header += [(width & 0xff),
               (width >> 8),
               (height & 0xff),
               (height >> 8)]

    for f in range(num_frames):
        header += [
            0, 0, # Left, top coordinates
            0, 0, # Width, height of the image

            int(duration / 10),   # Frame duration in ms
            int(frame_delay / 10), # Delay between frames in ms

            0b00100010 | (loop<<7)    # BPP = 1, sorting table = 0, interlace = 0, use local colour map
        ];


    def lzw_compress():
        codesize = 12 # Code size starts at max possible value
        lzwcode = 2**codesize + 2**codesize - 1 # Start with CLEAR code
        clearcode = lzwcode # Save current CLEAR code so we can reuse it
        endofinfocode = lzwcode + 1 # End of information code

        bitsleftinpixel = 8 # How many more pixels do we need to send before we can flush them to the output?
        pixelbuffer = [] # Buffer our pixels until they're ready to be sent out
        lastcolorsent = [] # When writing out the LZW dictionary we'll have to set a flag that tells us whether or not
                           # these are the first instances of a given run of codes

        def writebits(numbits, bitstream):
            nonlocal bitsleftinpixel
            while numbits > 0:
                if bitsleftinpixel == 0:
                    yield from pixelbuffer
                    bitsleftinpixel = 8

                chunksize = min(numbits, bitsleftinpixel)
                chunk = bitstream & ((1<<(chunksize)) - 1)
                bitstream >>= chunksize
                bitsleftinpixel -= chunksize
                numbits -= chunksize
                yield chunk


        def writecolor(color):
            if color not in dictsentsofar:
                if color == '':
                    return # We don't want empty strings stored in the string table
                elif len(dictsentsofar) <= 65535:
                    yield from writebits(bitsperindex+1, 0x0001) # Literal mode 1 (the same values for each byte)
                    yield from writebyte(color[-1], 1) # Write one extra byte since the previous byte was repeated
                else:
                    raise ValueError('Dictionary full!')

                dictsentsofar.add(color)


        def writebyte(bytevalue, repeatcount=0):
            if bytevalue >= 0:
                yield from writebits(repeatcount, bytevalue)
            else: # Repeat the last character N times instead of just that single character
                yield from writebits(-bytevalue, lastcolorsent[-2][0])



        for x in range(width):
            for y in reversed(range(height)):
                rgb = bytes([0]*3)
                for c in ['red', 'green', 'blue']:
                    rgb[0] |= getattr(fg_color,c)*(1<<(7-y%8))*2**(c=='blue')
                    rgb[1] |= getattr(fg_color,c)*(1<<(7-(y+1)%8))*2**(c=='red')

                pixel = rgb + b'\x00' # Add an alpha channel so we can draw transparent pixels

                if any(p!= b'\x00\x00\x00\x00' for p in lastcolorsent): # If there's a pixel with transparency
                    colormatch = next((d for d in dictsentsofar if d[-1] == pixel), '') # Check if the new pixel has been seen before
                    if colormatch == '':
                        matchfound = False

                        if lzwcode > 0xfff:
                            yield from writebits(1, True) # Flush literal buffer
                            yield from writecolor('')

                            lzwcode = (clearcode + 1)*2 + 1 # Reset dictionary so we can start over again!

                            yield from writebits(1, True) # Use CLEAR code instead of SET code because our color dictionary is empty now
                            yield from writecode(lzwcode)
                            continue


                        # Try to find matching substring at very high offsets, then look behind to see if there's any better matches at lower offsets too
                        hashval = sum([(i+j)*(ord(c)^i*j) for i, j, c in zip(*zip(*(reversed(((16*p)+(j%16),(j//16), c) for j, p in enumerate(pixel))))*3)]) % 0xffffffffffffffff
                        startoffset = minmax(0, min((hashval%(0xffffffffffffffff//0xfffffff))[::-1]) // 0xfffffff) # Offset up to which we check backwards to see if there's anything better yet
                        bestmatchstart = startoffset
                        bestmatchlen = 1
                        bestmatchend = ''

                        for offset in range(startoffset, 0, -1):
                            # Here we reconstruct the previous match from the previous iteration, but starting one byte before where it started
                            prevmatch = bestmatchend[bestmatchstart:] + ''.join(chr(lastcolorsent[(i+offset)%len(lastcolorsent)][0]^i) for i in lastcolorsent[(bestmatchstart+offset)%len(lastcolorsent)])
                            curmatch = prevmatch[:-1] + chr(prevmatch[-1] ^ ord(pixel[0]))
                            if curmatch in dictsentsofar:
                                if len(curmatch) > bestmatchlen:
                                    bestmatchstart = offset
                                    bestmatchlen = len(curmatch)
                                    bestmatchend = curmatch


                        # If we got here then we haven't found any matches whatsoever...
                        if bestmatchlen == 1: #...or if our best match is only one letter long
                            yield from writebyte(ord(pixel[0]), 0)
                            yield from writecolor(pixel)

                            if lzwcode < 0xfff:
                                lzwcode += 1
                            else:
                                yield from writebits(1, False) # Flush literal buffer

                                # Next time around look forward this long
                                delta = randrange(minmatchlength+1, maxmatchlength+1)
                                lzwcode = (clearcode + 2)**delta + 1
                                yield from writebits(1, True) # Use SET code instead of CLEAR code beacuse we already filled our dictionary
                                yield from writecode(lzwcode)
                                continue

                        elif bestmatchlen > 1:
                            # We need to send enough bits per pixel so that when we subtract the current length back off the next length will still fit inside of it
                            bitsperpixel -= delta = abs(bitsperpixel - log2(bestmatchlen))

                            yield from writebits(1, False) # Literal-mode switch used to turn on/off LZW encoding
                            yield from writebits(bitsperindex-delta, 0) # Index left shifted by X bits, plus 1

                    else:
                        yield from writebyte(-len(colormatch)+