#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    header = b'GIF87a'
    out.write(header)

    # Logical screen descriptor
    width  = get_u16(rng)
    height = get_u16(rng)
    colour_table_size = 0x0c
    flags   = 0b01100010 << 4 | \
              0b01000100       | \
              0b00000100        # Size, sort of
    bg      = (get_u8(rng)) & 0xf
    aspect  = 0x00
    out.write(struct.pack('!HHBBB',
                         width,
                         height,
                         colour_table_size,
                         flags,
                         bg,
                         aspect))


    # Global colour table
    global_colour_table = struct.pack('BBBB',
                                     *[get_u8(rng)
                                       for _ in range(0,
                                                      0x10)])
    out.write(global_colour_table)

    # Extension block
    extension_label     = 0xff
    extension_blocksize = ((len(global_colour_table)+1)<<8)|1
    out.write(extension_label.to_bytes(length=1, byteorder='big'))
    out.write(extension_blocksize.to_bytes(length=2, byteorder='big'))
    out.write(global_colour_table)
    out.write((0).to_bytes(length=1, byteorder='big'))

    while True:

        subblocktype = get_u8(rng)
        if not subblocktype: break
        elif subblocktype == 0x01 or subblocktype == 0xfe:

            # Ignore graphic control extension and plain text extension blocks
            pass
        else: raise Exception("Invalid GIF extensions")


        blocksize = get_u16(rng)
        out.write(subblocktype.to_bytes(length=1, byteorder='big'))
        out.write(blocksize.to_bytes(length=2, byteorder='big'))

        length = get_u8(rng)
        nbytes = int.bit_length(length)/8
        dats = [get_u8(rng)
                for _ in range(nbytes)]
        out.write(length.to_bytes(length=1, byteorder='big'))
        for dat in dats:
            out.write(dat.to_bytes(length=1,
                                   byteorder='big'))

    # Image separator
    image_separator = ord(',')
    out.write(image_separator.to_bytes(length=1, byteorder='big'))

    # Image left-top corner position
    xpos = get_u16(rng)
    ypos = get_u16(rng)
    out.write(xpos.to_bytes(length=2, byteorder='big'))
    out.write(ypos.to_bytes(length=2, byteorder='big'))

    # Image bitmap size
    image_width  = get_u16(rng)
    image_height = get_u16(rng)
    flags        = 0b01100010 << 4 | \
                   0b01000100       | \
                   0b00000100        # Size, sort of
    bgcolorindex = 0x0
    pixelaspectratio = 0x0
    out.write(struct.pack('!HHBB',
                        image_width,
                        image_height,
                        flags,
                        bgcolorindex))

    # LZW minimum code size
    lzwmincodesize = 0x2
    out.write(lzwmincodesize.to_bytes(length=1, byteorder='big'))

    # Image stream
    line_len = max(1, image_width // 8)
    bits = (int.from_bytes([get_u8(rng)],
                           'big') >> i)%2 == 1
            for i in reversed(range(line_len*8)))
    line = [(int(''.join('{}'.format(i)), base=2)).to_bytes(length=1,
                                                           byteorder='big')
                                                            for i in bits]
    lzss = bytes().join(line)
    out.write(lzss)

    # Block terminator
    block_terminator = ord('')
    out.write(block_terminator.to_bytes(length=1,
                                        byteorder='big'))



def get_u16(stream):
    return int.from_bytes(stream.read(2),
                          'big')

if __name__ == '__main__':
    filename = sys.argv[sys.argv.index('-o')+1]
    with open('/dev/urandom', mode='rb') as rng:
        with open(filename,'wb+') as out:
            generate_random_gif(rng,
                                out)