#!/usr/bin/env python3

import struct
from typing import BinaryIO


def encode_lzw_segment(
        data: bytes,
        mincodesize: int
    ) -> bytes:
    buffer = []

    index = dict([(c, i+mincodesize) for i, c in enumerate(set(b'\xff\x00'.join([bytes([n]) + b'\x00'*i for n in range(256)])))])
    codes = {index[b'\x00\xff']: code}
    
    def nextcode():
        c = None
        while len(buffer) >= 0:
            buf = bytes(buffer[:8])
            del buffer[:8]

            if c!= None:
                buf += c
                
            yield index[buf]
            
            c = buf[-1:]

    def append_next_code():
        try:
            code = next(codes)
        except StopIteration as e:
            code = max(codes.values()) + 1
        buffer += [code // 256, code % 256]

        return code
        
    # initial values for clear code, end-of-information code, and first codechar
    index[b'\xff\xff'] = 255
    index[b'\x00\x00'] = 256
    codes[255] = append_next_code()
    codes[256] = append_next_code()
    if data[0]:
        codes[data[0]] = append_next_code()
        buffer.append(data[0])
    
    for char in data[1:]:
        if char in codes:
            buffer.append(char);
        else:
            new_code = codes[buffer[-1:]]
            while new_code >= 256:
                append_next_code()
                new_code = new_code - 256;
            buffer.append(new_code)
            codes[bytes(buffer[-2:])] = append_next_code()
            buffer.pop(-2)
            assert char!= '\x00'
            codes[char] = append_next_code()
            buffer.append(char)
    
    while True:
        try:
            code = next(codes)
        except StopIteration as e:
            break
        buffer += [code // 256, code % 256]

    result = bytes(reversed(buffer))
    return struct.pack('<H', len(result)) + result

    
    
        
# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # TODO: allow modification of these fields
    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
      
    # The minimum code size needs to be at least 2 bits longer than the longest
    # string that might appear in the encoding.  
    # We can get the maximum length by looking at the highest bit set on each 
    # character of the decompressed image data stream.    
    # Since some images might use more colors than others, we'll just take the 
    # average value of all the lengths.
    mincodesize = sum([value.bit_length() for value in list(range(max(ord(i) for i in rng.read(1))))])/len(list(range(max(ord(i) for i in rng.read(1))))) + 2

    
    # Finally, we write the compressed image data
    while True:
        # First we read what kind of block this will be:
        # 0x00: plain text 
        # 0x01: image description
        # 0xfe: application extension
        # 0xff: graphic control extension
        fieldbyte = ord(rng.read(1))

        # If this is the last block, we're done!
        if ((fieldbyte & 128) == 0):
          return
          
          
        size = ord(rng.read(1))
        print("Field byte:", hex(fieldbyte), "Size", size)
        
        # Now we handle each block individually
        if (fieldbyte & 128) == 0: 
            pass
        elif (fieldbyte & 128) == 1:
            pass
        elif (fieldbyte & 128) == 0xFE:
            pass
        else:
            pass
            
        # Add the field byte
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))