#!/usr/bin/env python3


# Generates a random GIF file into `out` using the
# random number generator `rng` (defaulting to /dev/urandom).
def generate_random_gif(
        rng: BinaryIO=None, out: BinaryIO=None, datalen: int=None
    ) -> None:

    """Generates a random GIF file into `out`.

    Args:
        rng (BinaryIO, optional): Random Number Generator object; default to 
            standard input if not provided. Defaults to None.
        out (BinaryIO, optional): [description]. Defaults to None.
        datalen (int, optional): Length of data portion of GIF file in bytes;
            defaults to 1kB (1024B). Defaults to None.

    Returns:
        None: If no output stream is specified it prints the resulting GIF to 
        stdout instead.
    """


    def _get_random_frame() -> bytes:        
        return (b'\x2c\x00'+
                (rng.read(datalen))).ljust(data_size+2, b'\x00')
        
    if rng == None or out == None:
        from os import urandom as rng
        from sys import stdout as out

    if datalen == None:
        datalen = 1024
            
    assert(datalen <= 0xffff)

    header = b'GIF89a'
    data_size = datalen + 2

    frame_count = 10
    xpos = ypos = 0

    # Write pre-frame bytes  
    out.write(header)
    for i in range(len(header)):
        out.write((i ^ 0xff).to_bytes(1,"big"))
    
    # Write LZW encoder (subblock) 
    subblock_id = len(header) // 2
    lzw_header = b'\xf9\x04\x00\x00\x0f'       # id, len, min code size 
    lzw_terminator = b'\x00'                    # end of compression marker 
    lzw_encoded = b''                           # final encoded result (LZW output)
    
    prev_char = -1                              # previous LZW character (-1 means EOF)
    charcode = []                               # list of LZW characters encountered 
    bitstring = ''                              # current bit string being built 
    
    # Output LZW header and primary packet
    out.write(subblock_id.to_bytes(1, 'little'))    # Subblock ID 
    out.write(len(lzw_header).to_bytes(1, 'little'))      # Subblock length
    out.write(lzw_header)                             # Header 
    out.write((((data_size << 4) |
                  0x01) & ~0xF).to_bytes(1, 'little'))     # Packed field   
    out.write(((xpos & 0xFF) |
             ((ypos & 0xFF) << 8)).to_bytes(2, 'little'))    # Image position 
    out.write((0).to_bytes(1, 'little'))                     # Delay time 
    out.write(b'\x00\x00')                         # Transparent color index 

    # Generate all frames and compress them via LZW encoding
    for j in range(frame_count):

        # Encode the image with LZW compression 
        for ch in _get_random_frame():

            if ord(ch)-1 < 0 or ord(ch)-1 > 255:
                continue
            
            # Add new character to list of encountered chars 
            if chr(ord(ch)) not in charcode:
                charcode += chr(ord(ch)), 
                next_char = len(charcode)-1
                bitstring += '{:08b}'.format(next_char)[:7]
                prev_char = next_char
            
            # Handle one unused literal character at start of sequence 
            elif (prev_char==-1 and 
                  ord(ch)==charcode[0]):
                
                tmp = bin(int('{:<08s}'.format(bitstring)+
                              format(ord(ch), '08b')[7:])
                         ).replace("0b", "")

                tmp = ('{:>08s}'.format(tmp)[1:])[-8:]
                bitstring = ''.join([chr(int(i,2)) for i in tmp])
                prev_char = ord(ch)
                     
            else:

                # Find longest prefix substring in string table 
                s = charcode[:]
                cmpstr = bitstring

                res = ""
                lth =  0
                while len(cmpstr)>0: 
                    substr = cmpstr[:8]
                    pos = s.index(substr)
                    s = s[pos+1:]
                    cmpstr = cmpstr[pos+1:]
                    res += str(pos)
            
                # Append suffix to last LZW output bit sequence 
                lzw_encoded += chr(int(res+'{:<07s}'.format(bitstring)[:8],2))
                    
                # Build up new bit string from rest of byte 
                cmpstrsuff = '{:08b}'.format(ord(ch))[1:]
                cmpstr = '1' + cmpstrsuff[:-1] 
                cmpstr += '{:07s}'.format(cmpstrsuff[-1:]) 

                # Force minimum number of bits used by LZW encoding 
                lth = max(len(cmpstr)//8, 2)
                bitstring = '{:0<{}b}'.format(int(cmpstr,2),lth*8)
                prev_char = ord(ch)
            
        # Terminate LZW encoding 
        lzw_encoded += lzw_terminator
        
        # Encode terminator subpacket 
        out.write(subblock_id.to_bytes(1, 'little'))   
        out.write(len(lzw_encoded).to_bytes(1, 'little'))     
        out.write(lzw_encoded)          

        # Reset LZW vars 
        lzw_encoded = b''
        charcode = [' ', '\n']  # force these two characters 
        bitstring = ''          # clear bit buffer 

        # Write delay time subpacket  
        out.write(subblock_id.to_bytes(1, 'little'))     
        out.write(2.to_bytes(1, 'little'))            
        out.write((j*10).to_bytes(1, 'little'))        



    # Write trailer 
    out.write(b'\x3b')