#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The special value in Gif Header Format (LZW Compression):
    #   - Size of Global Color Table (LZW minimum code size)
    #   - Background color index (>= 0 and less than Number Of Colors)
    #   - Pixel aspect ratio
    # Note that it's actually specified as 1 byte little endian integer!
    gctsize = int.from_bytes(
            rng.read(1), 'little') % 4
    backgroundindex = int.from_bytes(
            rng.read(1), 'little') % 256
    pixelaspectratio = 0

    # No LZW compression support is planned for now, so we use no lzwmincodesize
    lzwmincodesize = 1

    # We can't specify more than two colors in global color table:
    numcolors = min([gctsize, 2])

    # According to spec, Application Extension must be present with the
    # following subblock:
    #   0x01: NETSCAPE2.0
    #   0xFF: Block Terminator
    #   Length: 0x0B
    #   Sub Data:
    #      0x03: Application Extension Flag
    #      0x01: Sub-Block ID
    #      0x00: Repeat count
    #      0x00: Block terminator
    #      Sub block contents:
    #         0x01: Loop count
    appext = b'\x21\xFF\x0B\x00NETSCAPE2.0\x03\x01\x00\x00\x00\x00,\x00\x00'
    loopcount = int.from_bytes(
            rng.read(1), 'big') % 256

    # Image Descriptor:
    #       0x2C: Image Left Position
    #       (Width of image minus one)
    #       0x00: Image Top Position
    #       (Height of image minus one)
    #       gctsize: Local Color Table Flag
    #         0: Do not use local color table
    #         > 0: Use local color table
    #            numcolors: Size of Local Color Table
    width = int.from_bytes(
            rng.read(2), 'big') & ~1
    height = int.from_bytes(
            rng.read(2), 'big') & ~1
    localcolorflag = bool(int.from_bytes(
            rng.read(1), 'little'))
    if localcolorflag:
        localcolortablesize = max([numcolors, 1])
    else:
        localcolortablesize = 0

    # Graphic Control extension
    #    0x21: Introducer
    #    0xF9: Label
    #    0x04: Size of graphic control block
    #    packed field:
    #          1: Reserved
    #          1: disposal method: leave as is or restore previous
    #          1: User input flag: don't accept user input
    #          1: Transparent Color Flag: transparent color is used
    #           7: reserved for future extensions
    #   delaytime: Amount of time to wait before displaying current frame,
    #             in hundredths of seconds
    #   transindex: Index of transparent color
    #               If 1-bit palette, then low bit represents transparency
    #               If has alpha channel, then this should always be set to zero.
    introducer = b'\x21\xF9\x04'
    sizeofgraphiccontrolblock = bytes([
            ((
                1 << 7 |
                1 << 6 |
                1 << 5 |
                1 << 4
            ) ^ (0 << 3)) |
            (((not localcolorflag) and 1) << 2) |
            ((bool(backgroundindex) and 1) << 1) |
            uintbitsneededforvalue(delaytime)])
    graphicsfield = bytes([(disposalmethod << 2) ^ \
                           (((not loopcount) and 1) << 1) ^
                            displaytransparentcolor])
    delaytime = int.from_bytes(
            rng.read(2), 'big') * 10
    transindex = int.from_bytes(
            rng.read(1), 'little') % 256


    # Packed fields:
    #   1: Size of Global Color Table
    #   3: Background Color Index
    #   1: Pixel Aspect Ratio
    #   1: Sort Flag
    sortflag = bool(int.from_bytes(
            rng.read(1), 'little'))

    # Write header
    out.write('GIF89a'.encode())

    # Write Logical screen descriptor
    out.write((width & 0xff).to_bytes(1, 'big'))
    out.write(((height >> 8) & 0xff).to_bytes(1, 'big'))
    out.write((height & 0xff).to_bytes(1, 'big'))
    out.write((gctsize).to_bytes(1, 'little'))
    #[1, 4][localcolorflag].to_bytes(1, 'little')
    out.write((((1 << 7) if (sortflag == True) else 0) |
               (pixelaspectratio & 0x7f)).to_bytes(1, 'little'))

    # Write color table
    for i in range(gctsize+1):
        out.write((i&0xff).to_bytes(1,'big'))
        out.write((i&0xff).to_bytes(1,'big'))
        out.write((i&0xff).to_bytes(1,'big'))

    # Write application extension
    out.write(appext)
    
    # Write Image Descriptor
    out.write('\x2c'.encode())
    out.write((width & 0xff).to_bytes(1, 'big'))
    out.write(((height >> 8) & 0xff).to_bytes(1, 'big'))
    out.write((height & 0xff).to_bytes(1, 'big'))
    out.write((localcolorflag<<7).to_bytes(1, 'little') +
              localcolortablesize.to_bytes(1, 'little'))

    # Write graphics control extension
    out.write(introducer)
    out.write(sizeofgraphiccontrolblock)
    out.write(graphicsfield)
    out.write((delaytime>>8).to_bytes(1, 'big'))
    out.write((delaytime&0xff).to_bytes(1, 'big'))
    out.write((transindex).to_bytes(1, 'little'))
    
    # Write plain text extension
    out.write('\x21\x00\x3B'.encode() + "This is what an unknown subblock looks like".ljust(12, '\0').encode())

    # Write image data
    # TODO: implement LZW compression
    out.write('.'.ljust(width, '\x00').encode() * height)

    # Write trailer
    out.write("\x3B".encode())


# Counts how many bits are needed to store given value
def uintbitsneededforvalue(number: int) -> int:
    assert number >= 0
    return 0 if number==0 else \
           (number.bit_length()-1)+uintbitsneededforvalue(number>>>(number.bit_length()-1))