#!/usr/bin/env python3

import sys
import os
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        width : int, 
        height : int,
        frames : int,
        lctsize : int,
        dithering : bool,
        interlacing : bool,
        progressive : bool,
        colors : list[tuple],
        out: BinaryIO):


    # Generate LZW table based on the given color count
   
    lzwtable = []
    lastindex = len(colors)-1
    baseidx = 0
    clearcode = 1 << lctsize
    endofinfo = clearcode+1
    nextcode = maxcode = clearcode + (1<<(lctsize+1)) - 1

    for i in range(clearcode):
        lzwtable.append([i])

    while True:
        if len(lzwtable)<nextcode:
            lzwtable.append([])
        
        idx = rng.read(1)[0] if not dithering else idx ^ ord(rng.read(1)[0]) % 2
        sublist = [baseidx, idx]
        lzwtable[lastindex].extend(sublist)

        baseidx = lastindex
        lastindex += 1
        nextcode+=1

        if nextcode > maxcode:
            raise Exception("Code overflow")

        if nextcode >= (1<<lctsize):
            break

    # Return LZW compression ratio
    def compressratio():
        return sum(len(item) for item in lzwtable)/len(lzwtable)


    header = b'GIF89a'
    nbcol = len(colors)

    # Calculate size of global colour table
    gctsize=nbcol*3 // 8 * (1+(nbcol*3)%8)
    
    # Determine color table size
    if lctsize==0:
        lctsize=gctsize
    elif lctsize>gctsize:
        print("Error: Color Table Overflow")
        exit()
        
    # Generate image size and buffer size
    xdim=(width+7)//8
    ydim=(height+7)//8
    
    bufsize = ((xdim+1)*ydim)*frames
    imsize = (((xdim+1)+3)*ydim)+((xdim+1)+3)+bufsize

    # Calculate palette size
    plteentrysize = 3
    palsize = 3*(plteentrysize*nbcol)

    # Output global headers
    out.write(header + struct.pack("<HHBBB", width, height, 0xf7, 0x00 | 0b0000 | 
            0b0000 | 0b0000 | 
            interlacing << 7 |
            progressive << 6 |
            0b00001 << 4 |
            dithering << 3 |
            lctsize-1))

    # Global Color Table
    for col in colors:
        out.write(struct.pack("<BBB", *(col)))

    # Image Descriptor
    out.write(struct.pack("<BBBBBHHH", 0x2C, 0x00, 0, 0, width, height, 0, lctsize-1))

    # Local Color Table
    for col in colors[:lctsize]:
        out.write(struct.pack("<BBB", *(col)))

    # Graphic Control Extension
    out.write(struct.pack("<B3sBBBHB", 
        0x21, b"GCE",
        0x00,
        0x00 | (0b0000 << 3) | (dithering << 2) | (lctsize-1),
        0,
        0,
        0xFF,
        0))
     
    # Gif Software Extensions
    out.write(struct.pack("<B3sBBB",
        0x21, b"NETSCAPE2.0", 
        0x03, 
        0x01, 
        0xff))

    # Application Extensions
    out.write(struct.pack("<B3sBBB",
        0x21, b"XMP Data XMP\x00", 
        0x0D, 
        0x0A, 
        0x50, 
        0x4E, 
        0x47, 
        0x0A, 
        0x1A, 
        0x0A))

    # Trailer Marker
    out.write(struct.pack("<3B", 0x3B, 0x00, 0x00));

    # Write pre-frame bytes  
    offset = 0;
    pos = 0;
    total = frames * (xdim+1)
    for f in range(frames):
        out.write(struct.pack(">HHB", offset, xdim, 0x00|(f&1)<<7|((f&2)>>1)<<6|
                 (f&4)<<5|((f&8)>>2)<<4|((f&16)>>3)<<3|((f&32)>>4)<<2|((f&64)>>5)<<1|
                 (f&128)<<3))
        pos += (xdim+1);
        offset += 2;

    # Write frame data

    print("Compress Ratio:", compressratio())

    for n in range(total):

        if n%progression == 0:
            currentcolor = rng.read(1)[0] % len(colors)
            
        col = colors[currentcolor];

        if n%(xdim+1) == 0:
            out.write(struct.pack("<BHBBB", 0x21, 0xF9, 4, 0x0b, 0x00, 0x00))

        if n<=offset:
            out.write(struct.pack("<B", (rng.read(1))[0]))
        else:
            value = lzwtable[prevvalue][rng.read(1)[0]]
            prevvalue = value
            out.write(struct.pack("<B", value))
            
    return compressratio()