#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # Generate the header and add to output
    out.write(b"GIF89a")

    # Generate the logical screen descriptor
    out.write(generate_logical_screen_descriptor())

    # Generate the global color table, if necessary
    if rng.read(1)[0] % 2:
        out.write(generate_global_color_table())

    while True:
        # Generate a graphic control extension (we don't use most of its
        # functionality anyway), then write it to our output
        out.write(generate_graphic_control_extension())

        # Generate an image descriptor and write it to our output
        out.write(generate_image_descriptor())

        # Generate the rest of the image data, then write it to our output
        out.write(generate_image_data())

# Generates a random logical screen descriptor
def generate_logical_screen_descriptor():
    return bytes([
        64 + rng.randrange(64),      # Width [1-32000]
        rng.randrange(64),            # Height [1-32000]
        1 << rng.randrange(3),        # Packed field; local color table follows
        rng.randrange(8),             # Color resolution [1-8]
        bool(rng.randint(0,1)),       # Global color table flag
        rng.randrange(7)+1,           # Background color index
        rng.randrange(7)+1            # Pixel aspect ratio [1-7/128], 0 means ignore
    ])
    
# Generates a random global color table
def generate_global_color_table():
    colors = []
    numcolors = 2**(bool(rng.randint(0,1))+1+rng.randrange(4))
    for i in range(numcolors):
        r = lambda x : max(min(x, 255), 0)
        colors += [(r(rng.randint(0,255)), r(rng.randint(0,255)), r(rng.randint(0,255)))]
        
    return bytes(sum([[c] + list(map(lambda x : int(max(0, min(x, 255))), c))]
                     for c in colors, [])
                 )[:(3*numcolors)-1]

# Generates a random graphic control extension
def generate_graphic_control_extension():
    return bytes([
        0x21,              # Extension introducer
        0xf9,              # Graphic control label
        4,                 # Block size
        rng.randint(0,255),# Disposal method
        rng.randint(0,255),# User input flag
        0                  # Transparent color index
    ])

# Generates a random image descriptor
def generate_image_descriptor():
    return bytes([
        0x2c,                # Image separator
        0,                   # Left position
        0,                   # Top position
        rng.randint(1,32000),# Width
        rng.randint(1,32000),# Height
        1<< rng.randint(3),   # Packed field
        8&~rng.randint(1),    # LZW minimum code size
        0                    # No local color table
    ])

# Generates some random image data
def generate_image_data():
    return bytes(rng.randrange(256) for _ in range(rng.randint(1, 1024)))