#!/usr/bin/env python3

import sys
import os
os.chdir(sys.path[0])

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    
    header = bytearray()
    # Header format: 
    # [magic]
    # [width][height]
    # [flags]
    # [bgcolor]
    header += b'GIF89a'
    width = rng.read(2)
    height = rng.read(2)
    flags = bytearray([rng.read(1)[0]])
    bgcolor = rng.read(1)
    print('Size', width, height)
    print('Flags:', bin(int.from_bytes(flags)), end='\n\n')
    flags += b'\x87\x00'    # Have a global color table, but not interlaced
    header += width + height + flags + bgcolor

    # Write the generated header to the output stream
    out.write(header)

    # Generate some data for the body
    datalen = int.from_bytes(
            rng.read(2), 'big')
    data = rng.read(datalen)
    # TODO: Add encoding subblock

    # Pixel aspect ratio
    pixelaspectratio = bytearray([0, 0])
    out.write(pixelaspectratio)

    # Color table format:
    # [size]
    # [color entries...]
    colortable = bytearray([pow(2, i) for i in range(12)])
    out.write(bytearray([(len(colortable))]))   # Size is just the length of this array, so no need to parse it from the input stream
    for entry in colortable:
        out.write(entry.to_bytes(4, 'little'))

    # Plain text extension format:
    # [extension introducer]
    # [label block size]
    # [label block]
    # [text block size]
    # [text block]
    extintro = b'\x21'        # Extension Introducer
    plaintextlabel = b'NETSCAPE2.0'      # This label means that this is an AOL spin-off of the gifsicle program, who used NETSCAPE2.0 as their application name
    plaintextlabel += b'\x00' * ((15-(len(plaintextlabel)))%8)     # Padding
    plaintextblocksizesize = b'\x0B'
    plaintextblocksizelittleendian = plaintextblocksizesize + b'\x00' * 3   # Size is literally just 11, which represents the minimum size required by NETSCAPE2.0
    plaintextblocksizemsg = plaintextblocksizelittleendian         # We can also put the size here, because it's already encoded
    plaintextblock = b'\x03' + b'\x01'       # Looping
    plaintextblock += b'\x00' * 3             # Padding
    plaintextmsg = plaintextblocksizemsg + plaintextblock
    plaintextext = extintro + plaintextblocksizesize + plaintextlabel + plaintextmsg
    out.write(plaintextext)

    # Application extension format:
    # [application introducer]
    # [application extension block size]
    # [application identifier]
    # [application authentication code]
    appintro = b'\xFF'        # Application introducer
    appidentifier = b'NETSCAPE2.0'      # This is NETSCAPE2.0, used by both AOL spinoffs and other programs like GIMP to identify different kinds of extensions
    appauthenticationcode = b'\x01' + b'\x00' * 3           # Authentication code is set to 0x01, which means that it's B&W, where 0x00 is grayscale and 0x01 is colour mode. Data after should be ignored
    appauthcodemsg = bytearray()
    appauthcodemsg += appauthenticationcode
    appauthcodemsg += b'\x00'                    # End subblock mark
    appauthcodemsg += b'\x00' * 3               # Placeholder for future extension blocks
    appauthcodemsg += b'\x00' * 2               # The first two bytes are reserved for application specific use; we leave them empty
    appextensionsubblock = b'\x01'                 # Block size
    appextensionsubblock += appidentifier       # Identifier
    appextensionsubblock += len(appauthcodemsg).to_bytes(2, 'big')              # Subblock size
    appextensionsubblock += appauthcodemsg       # Actual message inside the extension block
    appext = appintro + appextensionsubblock
    out.write(appext)

    paddinglen = 16 - len(data) % 16
    if paddinglen == 0:
        padding = b''
    else:
        padding = rng.read(paddinglen)
    out.write(data + padding)