#!/usr/bin/env python3

import random
from math import ceil
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and write to output
    header = ''.join(['GIF', '8', '9', chr(ord('a') + random.randrange(6))])
    out.write(header.encode())

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Write LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int(2 ** ((lsd[4] & 0x7) + 1))*3

    # If there's a global color table, add it to output
    if gctlen > 0:
      gct = bytes([(random.randrange(256),random.randrange(256),random.randrange(256)) for _ in range(gctlen)])
      out.write(gct)
    
    # We'll make multiple passes through the frames
    bufposition = out.tell()

    def write_frame():
        nonlocal bufposition

        # Remember where we started writing so far
        bufstart = out.tell()

        # Create new frame
        out.write('\x2c'.encode())

        # Create image descriptor
        id = bytearray(9)

        # Set width between 1 and 64 inclusive
        id[4] = min(max(round(random.gauss(32, 8)), 1), 64)

        # Set height between 1 and 64 inclusive
        id[5] = min(max(round(random.gauss(32, 8)), 1), 64)

        # Choose whether to use local color table or not
        if gctlen > 0 and random.randint(0, 1)!= 0:
          # We do want a local color table; choose length between 2 and 7 inclusive
          id[6] |= random.choice([n for n in range(2, 8)]).to_bytes(1, 'big')[0]

          # Then choose palette entry 0 to be transparent
          id[8] &= ~(1<<7)
          
          # Chose background color index between 0 and 255 inclusive
          id[8] |= random.randrange(256).to_bytes(1, 'big')[0]
        else:
          # Nope! There will be no local color table used
          id[6] &= ~0b11

        # How many bytes?
        bytesperline = max(((id[4]+1)<<1), (id[5]+1)) + (((id[6]&7)+1)<<1) + 1

        # Add number of bytes needed for frame.
        out.write(struct.pack('<B', bytesperline))

        # Add local color table position, if applicable
        if (id[i]&0x7)>0:
          out.write(struct.pack('<H', bufstart+13+(id[6]&7)+1))

        # Figure out what code length to use
        if (id[i]&0x7)<2:
          # Not indexed color mode so all codes of size 1
          codesize = 1
        else:
          codesize = random.choice([n for n in range(2, (id[i]&0x7)+1)])
          
        # And set it
        out.write(struct.pack('<B', codesize))

        # Get ready to start adding pixel data
        out.write(bytes(bytesperline))

        # Keep track of how many bytes in total 
        accumulator = 0

        # Randomize the code length of the image data
        # This is helpful when you need fewer colors but they should
        # still be distinctive from adjacent indices
        codes = [(chr(random.randrange(256)).encode(), codesize)
                 for i in range(ceil(bytesperline /
                                         (codesize + 1)))]

        # Add code lengths to their respective places at end of frame
        for code, clen in codes[:-1]:
            accumulator += clen
            out.seek(-accumulator, 1)
            
            out.write(struct.pack("<B", clen))

            # Move back one character since we overshot earlier
            out.seek(-1, 1)

        # Add last code separately
        clen = sum([len(code) for code, csize in codes])
        accumulator += clen
        out.seek(-accumulator, 1)
        
        out.write(struct.pack("<B", clen))
        out.seek(-1, 1)

        # Write out the pixel data itself
        for code, csize in codes:
            out.write(code)

        return out.getbuffer()[bufstart:out.tell()].tobytes()


    # At most three frames; after that, repeat from beginning
    for framecount in range(random.randrange(3, 10)):
        # Write current frame
        frame = write_frame()

        # Write additional frames from here
        positions = list(reversed([-pos for pos in find_all(frame)]))

        for i in range(random.randrange(10)):
            # Seek to location of previous frame
            out.seek(positions[(framecount-1)%len(positions)], 0)

            # Write a frame here
            frame = write_frame()