#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (either /dev/urandom or /dev/random)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Generate a valid GIF header
    header = get_valid_gif_header()
    out.write(header)
    
    # Get the length of the main image data
    datalen = get_random_image_size(rng)
    print('Image length:', datalen)
    
    # Generate the image data and add it to the output
    data = gen_random_image(rng, datalen)
    out.write(data)
    
    # Add padding bytes if necessary to achieve a multiple of four bits
    padlen = 4 - ((datalen + 1) % 4)
    out.write(padlen * b'\x00')

    # Write the trailer byte
    out.write(b'\x3B')
    
# Returns a valid initial frame descriptor, as described in section 19 of
# https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def get_valid_gif_header():
    width = b'\x0A'
    height = b'\x1E'
    return b'GIF89a\x0A\x1E\x87\xA0\x0F'

# Gets an integer in range [0; 2^(16)-1], inclusive on both ends
# The result will be uniformly distributed across the entire possible range
def get_random_image_size(rng):
    return int.from_bytes(rng.read(2), 'big') & 0xFFFF 

# Generates an uncompressed bitmap image of size `width` by `height`, with each pixel
# randomly generated from [0; 255].
# The resulting stream is guaranteed to start with a valid initial frame descriptor.
def gen_random_image(rng, width, height=None):
    if height == None:
        height = width
        
    lzwmincodesize = 2
    blocksize = min(lzwmincodesize + 1, 1 << lzwmincodesize)
    clearcode = 1 << lzwmincodesize
    endofinformation = clearcode + 1
    nextcode = clearcode + 2
    
    # Set up LZW dictionary
    dictbytepos = 258
    dictbits = 9
    maxdictsize = 1 << dictbits
    firstbitmask = 1 << (blocksize - 1)
    lastbitmask = ~firstbitmask
    dictsize = 258
    dict = {i : i for i in range(256)}
    stack = []
    stacktop = 0
    outbuf = []
    bitcount = 0
    subblocklen = 0
    
    # Function to emit a single code word, encoded using the given number of bits
    def emit_code(codeword, nbits):
        nonlocal subblocklen
    
        # Emit enough zeroes so that we have room for at least one more literal byte
        while bitcount >= 8:
            outbuf.append(0)
            bitcount -= 8
        
        # Emit all but the last full byte
        while nbits > 8:
            outbuf[-1] |= codeword >> (nbits - 8)
            codeword <<= 8
            nbits -= 8
            bitcount += 8
            
        # If there are any remaining bits, they must be part of this last byte
        if nbits!= 0:
            assert bitcount + nbits <= 8
            
            outbuf[-1] |= codeword << (8 - bitcount - nbits)
            bitcount += nbits
            
    # Function to emit the current contents of the bit buffer, starting at
    # position `startpos`.
    def flush_bitbuffer(startpos):
        nonlocal bitcount

        if startpos < 0:
            raise ValueError('Bit count cannot go negative!')
        elif startpos == 0:            
            return
                    
        # Flush the rest of the buffer
        while bitcount >= 8:
            outbuf.append((bitbuf >> (bitcount - 8)) & 0xFF)
            bitcount -= 8
            
        # Fill up the final incomplete byte with zeroes
        if bitcount!= 0:
            outbuf.append((bitbuf << (8 - bitcount)) & 0xFF)
            bitcount = 0
            
    # Function to emit the complete content of the bit buffer
    def flush_bitbuffer_full():
        flush_bitbuffer(-1)

    # Helper function for writing a variable-length value
    def write_varlen(n):        
        if n < 0:
            raise ValueError('Negative numbers not allowed')
        
        # Split the value until we can fit everything inside just eight bits
        keepgoing = True
        while keepgoing:
            val, keepgoing = divmod(n, 256)
            yield val
                
            if not keepgoing:
                break
        
            yield 0
            
    # Main loop which precomputes the codes for every code length, then outputs 
    # the actual compressed data
    for codelen in range(lzwmincodesize, 10): 
        prefixes = {}
                
        # Compute the prefix codes for all codes with the current code length
        code = 0
        while code < dictsize:
            # For each code...

            # Iterate through its bits from MSB to LSB, checking how many of them match 
            # against existing codes whose values have already been emitted.
            # We use the property of powers of two to check these positions quickly.
            mask = 1 << (codelen - 1)
            while mask!= 0:
                # If the relevant bit matches what's expected...

                #... try to fetch the corresponding code from our table; 
                # if such a code does not exist, create a new entry
                key = (prefixes[code] if code in prefixes else code, 
                        mask | dictbytepos)
                if key not in dict:
                    dict[key] = nextcode
                    
                    # Increase the number of bits used to encode dictionary indices 
                    # until the maximum size has been reached
                    dictbytepos += 1
                    if dictbytepos == maxdictsize:
                        dictbytepos = 0
                
                nextcode += 1
                    
                # Advance past this bit
                code <<= 1
                mask >>= 1
            
                # Update list of prefixes
                prefixes = {k[0]: v 
                            for k, v in dict.items() 
                                if (k[1] << (maxdictsize - 1) 
                                    & maxdictsize
                                    == code
                                )}
                
            # No matter whether or not it matched anything, this code may now be used
            # as a prefix itself
            prefixes[code] = code    
        
        # Once we've found all code prefixes for a particular code length, emit the
        # codes themselves
        # Important note: since we don't know ahead of time the total amount of data 
        # produced by this call, we need to ensure that no extra information gets lost.
        # Our strategy here is to use a bit buffer; whenever a bit buffer fills up, 
        # we send it along as soon as possible. The most recent bit in the buffer is 
        # never shifted out of the buffer immediately after being sent, because otherwise
        # we might lose some data in case the very last bit was a zero. Therefore,
        # before filling up the buffer, make sure that the emitted data actually contains 
        # a set bit somewhere.      
        for i in range(endofinformation + 1):
            codeword = dict[(prefixes[i] if i in prefixes else i,
                            dictbytepos)]
            
            # Check whether the codeword requires more than eight bits
            while codeword >= 256**(8 - bitcount):
                # If so, shift it down to only include the lower 8-bit chunk
                codeword &= 0xFF**8 - 1
                
                # Send it to the bit buffer
                emit_code(codeword, 8 - bitcount)
                
                # Reset the codeword so that it doesn't contain the upper bits anymore
                codeword = 0
            
            # This means that the codeword fits entirely within the bitbuffer
            emit_code(codeword, codelen)
            
            # Shift the bitbuffer forward if necessary
            if bitcount + codelen > 8:
                bitbuf |= codeword << (codelen - (8 - bitcount))
                flush_bitbuffer(bitcount + codelen - 8)
            else:
                bitbuf |= codeword << (8 - bitcount - codelen)
                bitcount += codelen
            
        # At this point our bitbuffer should've filled up, so let's flush it
        flush_bitbuffer_full()
        
        # Output the subblock delimiter byte
        outbuf.append(0)  
        
        # Output the length of the previous block (in bytes)
        packedsubblocklen = tuple(reversed([*write_varlen(subblocklen)]))
        for x in packedsubblocklen:
            outbuf.append(x)
        
        # Move the data to the output buffer and reset counters
        subblocklen = 0
        outbuf += outbuf[:]
        del outbuf[:]
        
        # Prepare for decoding the next RLE run
        prevpixval = None
        rlerun = 0
    
    # Compress the image data    
    # This is done in chunks of 256 pixels to limit memory usage
    for y in range