#!/usr/bin/env python3
"""
Generates a random GIF file into a filestream `out` 
using the random number generator `rng`.

Example usage with /dev/urandom :

    from rgifgen import generate_random_gif
    import os

    with open('/dev/urandom','rb') as rng:
        with open('horriblefilename','wb') as horriblefile:

            generate_random_gif(rng, horriblefile)

This produces an 8kb GIF which is always animated. It's funny!
"""

import struct
from typing import BinaryIO

# generates a random GIF into `out`, 
# consuming the random value stream `rng`.
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    ### Header

    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), 
            byteorder='big') & 0xfff

    out.write(header + (datalen).to_bytes(
            3, 'big'))
    
    # write global colour table

    clrslen = ((int.from_bytes(
            rng.read(1),'big'))//3)*4

    print("clrs len",clrslen)

    for i in range(clrslen // 3):
        for j in range(3):
            colbyte = rng.read(1)[0]
            out.write((colbyte).to_bytes(
                1, 'big'))

         
    ### Frame Data

    frame = {}

    frame['left']   = int.from_bytes(
            rng.read(2), byteorder='little')
    frame['top']    = int.from_bytes(
            rng.read(2), byteorder='little')
    frame['width']  = int.from_bytes(
            rng.read(2), byteorder='little')
    frame['height'] = int.from_bytes(
            rng.read(2), byteorder='little')

    lctsize = None

    while True:

        ordinance = int.from_bytes(
                rng.read(1), 
                byteorder='little')[0]
        
        if ordinance >= 0x00 and ordinance <= 0x0f:
        
            # Ordinal 0x00 through 0x7e are reserved for future use.
            raise NotImplementedError()
            
        elif ordinance>= 0x80 and ordinance<= 0xfe:
        
            # Bit 8 is set so this is a local color table

            lctsize = (ordinance & 7)+1

            for i in range(lctsize * 3):
                for j in range(3):

                    colbyte = int.from_bytes(
                            rng.read(1), 
                            byteorder='big')[0]
                    
                    out.write((colbyte).to_bytes(
                            1, 'big'))


        elif ordinance== 0xff:

            # Ending block
            break

        else:

            # This will never happen with any realty since we just start with a clear block
            # but I am keeping it here just in case something unexpected happens...

            raise ValueError()


    # write graphic control extension
    # Two blocks follow after

    out.write(((0x21).to_bytes(
            1, 'big'))+b'\xf9'+\
              (((2).to_bytes(
                  1,
                  byteorder='little'))+\
               (0x00.to_bytes(
                  1, 
                  byteorder='little'))+\
               (0x00.to_bytes(
                  1, 
                  byteorder='little')))+\
              (0x00.to_bytes(6)))

    # Write the last two bytes again twice because reasons

    out.write((((2).to_bytes(
                  1, 
                  byteorder='little'))+\
               (0x00.to_bytes(
                  1, 
                  byteorder='little'))+\
               (0x00.to_bytes(
                  1, 
                  byteorder='little')))+\
              (0x00.to_bytes(6))*2)
        
    # Now comes the image descriptor

    out.write((((0x2c).to_bytes(
                  1, 
                  byteorder='big'))+\
               frame['left'].to_bytes(
                  2, 
                  byteorder='little'))+\
              (frame['top'].to_bytes(
                  2, 
                  byteorder='little'))+\
              (frame['width'].to_bytes(
                  2, 
                  byteorder='little'))+\
              (frame['height'].to_bytes(
                  2, 
                  byteorder='little'))+\
              (0x00.to_bytes(
                  1, 
                  byteorder='little'))+\
              (0x00.to_bytes(
                  1, 
                  byteorder='little')))

    # Next up, the LZW code size and packed field

    out.write((((8).to_bytes(
                  1, 
                  byteorder='little'))+\
               (0x00.to_bytes(
                  3, 
                  byteorder='little'))+\
               (0x00.to_bytes(
                  1, 
                  byteorder='little'))))

    # Finally, the encoded pixels themselves

    # The first code is read in already as an unsigned char
    # and stored in a variable called "code". Then the loop
    # iterates until end-of-frame is encountered.

    pixelstring = [0]*frame['width']*frame['height']

    numcodes = 2**bitsofcode

    for y in range(frame['height']):
        for x in range(frame['width']):

            # If the current pixel was decoded before being written,
            # we need to recode it now. Otherwise, we can just keep going.

            if not isinstance(pixelstring[y*frame['width']+x], int):

                maxbits = bitsofcode
                p = pixelstring[y*frame['width']+x][0]

                while p > symbols[-maxbits]:
                    maxbits -= 1

                code = p>>maxbits

                for i in range(1<<maxbits):
                    pixelstring[y*frame['width']+x][i+(bitsofcode-maxbits)]=None
                    
                while p>symbols[-maxbits]:
                    pixelstring[y*frame['width']+x][code+(bitsofcode-maxbits)]=p%2**(bitsofcode-maxbits)
                    code+=1
                    p/=2**(bitsofcode-maxbits)

            # In either case, we update our decoded pixel string by writing
            # the next two bits of the code into the relevant position.

            pix = pixelstring[y*frame['width']+x]
            pix[0] = pix[1]
            pix[1] = pix[2]    
            pix[2]= ((ordinance & 0xc0)>>6)<<2 | \
                   ((ordinance & 0x30)>>4)<<4 | \
                   ((ordinance & 0x0c)>>2)<<6 | \
                   ((ordinance & 0x03))


            while(pix[1]==None):

                try:
                    pix[1]=pix[2]
                    pix[2]=ord(rng.read(1))[0]
                except IndexError:
                    # End-of-frame reached.
                    break

    # Once that finishes, we simply write all those values to the file.

    buffer = []

    for i in range(frame['height']):
        for j in range(frame['width']):
            try:
                buffer.append(str(pixelstring[(i*frame['width'])+j]).encode())
            except TypeError:
                pass

    out.writelines(buffer)

    ## Trailer

    out.write(((0x3b).to_bytes(
            1, 'big')))