#!/usr/bin/env python3

import io
import random
import struct


# Generates a random GIF file in RAM and returns it as a binary stream
def generate_random_gif() -> bytes:
    gif = bytearray()

    # Write the version string
    gif += b'GIF89a'

    width = random.randint(50, 100)
    height = random.randint(50, 100)

    # Make the header contain the dimensions of the animation frame
    gif += (width).to_bytes(2, 'little', signed=False)
    gif += (height).to_bytes(2, 'little', signed=False)

    # Flags: global color table included at the end of the image
    flags = 0b10000 | 0b00000 | 0b00000

    # Add flag bits to the start of each byte
    for i in range(7):
        mask = 0x01 << i
        if ((mask & flags)!= 0):
            gif += b'\x64'
        else:
            gif += b'\x00'

    # A single pixel value that does not exist in the color palette
    transparent_color = random.choice(range(256))
    gif += transparent_color.to_bytes(
        1, byteorder='little', signed=False)

    # No global color table is used; use this color in every place
    gif += b'\xff\xff\xff'


    def make_subblock(data: bytes):
        len_ = len(data)

        if len_ <= 255:
            return b'\xff' + len_.to_bytes(1, 'big') + data
        elif len_ <= 65535:
            return b'\xfe' + len_.to_bytes(2, 'big') + data
        raise ValueError('Subblocks must be between two and sixty-five thousand bytes long')

    lzw_mincodesize = random.randint(2, 8)
    subblocks = [make_subblock(struct.pack('<H', lzw_mincodesize))]

    while True:
        # This destroys most of the information in the LZW compression algorithm because
        # they are always followed by pure black, so we call it once per block instead of
        # once per pixel. We also have more control over how many colors are actually
        # generated here compared to in the loop below where we iterate through each row
        # and column.
        if random.randint(0, 1) == 1:
            continue

        # The number of different colors used in this frame
        ncolors = random.randint(
            2, min(lzw_mincodesize+1,
                    256 - len(subblocks)))

        # Generate some colors and add them to our color map
        newcolors = []
        for _ in range(ncolors):
            r = random.getrandbits(8)
            g = random.getrandbits(8)
            b = random.getrandbits(8)
            newcolors.append((r,g,b,))

            length = len(newcolors)
            if length > 256:
                break
            subblocks.append(struct.pack('<BBB', *newcolors[-1]))

        # If there are only 1 or 2 colors, then don't bother encoding it either
        if ncolors < 3:
            continue

        subblocklen = sum([len(_) for _ in subblocks]) // 256
        gif += subblocklen.to_bytes(1, 'big')
        for sblk in subblocks[:256]:
            gif += sblk


        # Encode each pixel with a variable size LZW code representing its RGB values
        last_R = None
        last_G = None
        last_B = None
        pixels_left = width*height
        while pixels_left >= 256:
            R = random.getrandbits(8)
            G = random.getrandbits(8)
            B = random.getrandbits(8)

            if (last_R, last_G, last_B) == (None, None, None) \
                    or (last_R, last_G, last_B) == (R, G, B):
                continue

            lzw_code = max(
                1, next(_ for _, c in enumerate(reversed(newcolors))
                        if c==(R, G, B)))
            gif += struct.pack('<H', lzw_code)
            pixels_left -= 1


            # Update the last pixel pair to include the new pairs
            # Use only the least significant byte since the higher ones
            # "leak" into the lower bytes of the LZW codes
            last_R = (last_R >> 8) ^ R
            last_G = (last_G >> 8) ^ G
            last_B = (last_B >> 8) ^ B
        if pixels_left == 0:
            break

        # Store any remaining pixels in their own subblock
        subblocks = [(rgb >> shift).to_bytes(
                            1,
                            byteorder='little',
                            signed=(shift==0))
                      for rgb in (last_R, last_G, last_B)]

        subblocklen = sum([len(_) for _ in subblocks]) // 256
        gif += subblocklen.to_bytes(1, 'big')
        for sblk in subblocks[max((pixels_left-len(subblocks)), 0):]:
            gif += sblk
    gif += b'\x00'

    return bytes(gif)


