#!/usr/bin/env python3

import io
import math
import struct
from copy import deepcopy
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bool:
    buf = bytearray()
    bufsize = 2**15

    # Generate the header and add to output
    tmp = rng.read(6)
    buf+=tmp
    out.write(tmp)

    width = 2
    height = 2

    if ord(tmp[0])!= 0x47 or ord(tmp[1])!= 0x49 or ord(tmp[2])!= 0x46 or\
       ord(tmp[3])!= 0x38 or ord(tmp[4])!= 0x39 or ord(tmp[5])!= 0x61:
        return False

    tmp = rng.read(7)
    buf+=tmp
    out.write(tmp)
    width = 2<<ord(tmp[0])+ord(tmp[1])
    height = 2<<ord(tmp[2])+ord(tmp[3])

    # Add LSD to output
    out.write(struct.pack('B', 1))
    out.write(struct.pack('B', 0))
    out.write(struct.pack('<H', width))
    out.write(struct.pack('<H', height))

    # Get global color table length
    gctlen = ((ord(rng.read(1)) & 0x07) + 1) * 3

    # Add GCT to output if we have one        
    if gctlen > 0:
      out.write(deepcopy(rng.read(gctlen)))

    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = None

    # Loop until there's enough data to fill up our buffer
    while sum(map(lambda f:f[0]+5, filter(
        lambda f:f[1], map(lambda _:nextFrame(_,width), range(height//2))))) < bufsize:
        nextFrame(out, width)
        
    print('{} frames'.format(math.floor(sum(map(lambda f:f[0]/f[1], filter(
        lambda f:f[1], map(lambda _:nextFrame(_,width), range(height//2))))) / bufsize)))




        

# Generates the next frame and returns whether or not it was successful
def nextFrame(
        out: BinaryIO, 
        width : int
    ) -> bool:
    # Create new frame        
    out.write(b'\x2c')
    
    # Create image descriptor
    out.write(struct.pack("B", 0)) 
    out.write(struct.pack("B", 0)) 

    # Add canvas width and height to image descriptor
    out.write(struct.pack("<H", width))
    out.write(struct.pack("<H", 1))

    # Get number of bytes to read based on bits in ID
    imgdatalenbits = pow(2,ord(rng.read(1)))
    imgdatalen = int(width)*(pow(2,imgdatalenbits+1))+imgdatalenbits+1
    codelen = pow(2,(ord(rng.read(1))&7))+1
    
    # Get dispose method, if any
    dispose = -1 if (ord(rng.read(1))&8)==0 else ord(rng.read(1))&7

    if isinstance(prevframe, list) and \
       imgdatalen < len(prevframe) and \
       (dispose==-1 or dispose<=2):
        out.write(prevframe[:imgdatalen])

    elif isinstance(prevframe, list) and dispose == 2:
        out.write(prevframe[-imgdatalen:])

    else:
        out.write(b'\0'*imgdatalen)
            
    # If we don't yet know how much image data we need per line,
    # get that info now by reading the first row of pixels
    if codelen == -1:
        out.seek(-1, 1)

        # If we've already generated an entire line, just use that data
        if imgdatalenbits <= 6:
            codelen = 2**imgdatalenbits
        # Otherwise, generate more data until this many bytes are available
        else:
            while True:
                n = 1 << (imgdatalenbits-6)

                x, y = ord(rng.read(1)), ord(rng.read(1))
                
                if x >= n or y >= n:
                    break

            codelen = n
                
        out.seek(1, 1)

    # Now create the actual pixel data
    data = []

    for i in range(math.ceil(imgdatalen/(codelen*2)-1)):
        clen = ord(rng.read(1))*256 + ord(rng.read(1))
        data += [clen]*clen

    out.write(bytes([codelen]*math.ceil(imgdatalen/(codelen*2)-1)) +
              bytes(data))

    # Save as previous frame because we might need it later
    return True


    