#!/usr/bin/env python3

import sys
import argparse
from io import FileIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: FileIO, 
        width: int=None, 
        height: int=None, 
        colors: dict={
            0: [0x00, 0x00, 0x00], 
            255: [0xff, 0xff, 0xff]}):
    
    colorTableSize = len(colors) * 3

    if not width and not height:
        raise ValueError("Either `width` or `height` must be provided!")

    elif not width:
        width = int(height / 2)

    else:
        height = int(width * 2)

    # We also need to ensure that our new dimensions 
    # are not bigger than the original ones, since you can't have more pixels 
    # on your screen then there are values available in the palette!
    while True:
       if width < 1 or height < 1:
           continue

       break

    header = b"GIF89a"
    duration = 0x00
    backgroundIndex = 0x00
    pixelAspectRatio = 0x00

    globalColorTableFlag = 0x01
    sortFlag = 0x00
    isInterlace = 0x00
    localColorTableFlag = 0x01

    backgroundColor = colors[backgroundIndex]
    aspectRatioByte = (((pixelAspectRatio & 0xf) << 4) | ((pixelAspectRatio & 0xf0) >> 4))

    logicalScreenDescriptor = header + \
                              width.to_bytes(2, 'little') +\
                              height.to_bytes(2, 'little') +\
                              0x80.to_bytes(1, 'little') +\
                              (colorTableSize ^ 0xff).to_bytes(1,'big') +\
                              backgroundIndex.to_bytes(1, 'little') +\
                              aspectRatioByte.to_bytes(1, 'little')

    applicationExtension = b"\x21\xFF\x0BNETSCAPE2.0"
    applicationIdentifier = b"<EMAIL>"
    applicationAuthCode = b'\x03\x01'

    netscapeBlock = applicationIdentifier + applicationAuthCode 

    commentExtension = b"\x21\xFE" + (len(netscapeBlock)).to_bytes(1,"little") + netscapeBlock

    plainTextExtension = applicationExtension + \
                         length.to_bytes(1,"little") + duration.to_bytes(1,"little") + \
                         nameLength.to_bytes(1,"little") + \
                         b"Plain Text Extension Block!" 

    gceHeader = (0x21).to_bytes(1, "big") + (0xF9).to_bytes(1, "big") + (0x04).to_bytes(1, "big")
    disposalMethod = int(0x00).to_bytes(1, "big")

    userInput = int(0x00).to_bytes(1, "big")

    transparentColorFlag = int(0x00).to_bytes(1, "big")
    delayTime = int(0x0d).to_bytes(2, "little")
    transparentColorIndex = int(0x0f).to_bytes(1, "big")

    blockTerminator = (int(0x00)).to_bytes(1, "big")

    graphicControlExtension = gceHeader + disposalMethod + \
                              userInput + transparentColorFlag + \
                              delayTime + transparentColorIndex + \
                              blockTerminator

    plainTextBlock = plainTextExtension + \
                     blockTerminator

    buffer = bytearray()

    buffer += logicalScreenDescriptor

    buffer += graphicControlExtension

    buffer += plainTextBlock

    buffer += b";"

    for y in range(height):
         for x in range(width//16):

            start = 0
            end = 16

            for i in range(16):
                val = int.from_bytes(
                    rng.read(end - start), 
                    byteorder="little")

                index = val % colorTableSize 

                buffer.append(index)
                

    outlist = []
    for key in colors:
        value = colors[key]
        
        outlist += value
        
    flattened = list(flatten(outlist))

    palette = bytearray()

    # Convert the color table from RGB to GRB format
    for i in range(0, len(outlist), 3):
    
        palette += struct.pack('<BBB',
                                flattened[i+2], 
                                flattened[i+1], 
                                flattened[i])


    trailersize = (len(palette)+1).to_bytes(1, "little")

    trailer = b""
    trailer += trailersize + b"\x00" + palette + b"\x00" 
    
    totalsize = len(buffer)+len(trailer)
    size = totalsize.to_bytes(2, "little")
    buffer = size + buffer + trailer

    return buffer


    
if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Generate a GIF with randomly colored pixels.")

    parser.add_argument('-w', '--width', help="Image Width", type=int)
    parser.add_argument('-H', '--height', help="Image Height", type=int)

    args = parser.parse_args()

    if not sys.stdin.isatty():
        inputfile = sys.stdin.buffer
    else:
        inputfile = open("/dev/urandom", mode="rb")

    outputfile = open("output.gif", mode="wb+")

    data = generate_random_gif(inputfile, **vars(args))
    
    outputfile.write(data)

    

