#!/usr/bin/env python3

import struct
from typing import BinaryIO

class Color:
  def __init__(self, r:int=0, g:int=0, b:int=0):
    self.r = r
    self.g = g
    self.b = b

  def toBytes(self) -> bytes:
      return bytes([
          self.r,
          self.g,
          self.b])

class Block:
  def __init__(self, blocktype:str="", data:bytes=None):
    self.blocktype = blocktype
    self.data      = data if data!= None else []

  @staticmethod
  def fromFile(f:BinaryIO, blocktype:str="") -> "Block":

      length = int.from_bytes(
              f.read(1), byteorder="big")
      
      newblock = Block(blocktype)
      while len(newblock.data) < length:
          chunklen = min(length - len(newblock.data), 2048)

          # Read and append a chunk of this size
          newblock.data += f.read(chunklen)
          
      return newblock

  def toBytes(self) -> bytes:
      assert type(self.blocktype) == str
      if len(self.data) > 255:
        raise Exception("No block is longer than 255!")
      return (
          len(self.data).to_bytes(1,byteorder="big") + 
          bytes(self.blocktype,encoding="utf-8") +
          self.data)
  
  def toBytesWithLen(self) -> bytes:
      return (
          (len(self.toBytes())+1).to_bytes(1,byteorder="big") + 
          self.toBytes()
      )


class ImageDescriptor:
  def __init__(self):
    pass

  @staticmethod
  def fromFile(f:BinaryIO) -> 'ImageDescriptor':
      '''
      :returns an image descriptor with pixel info read from f
      '''
      id = ImageDescriptor()
    
      id.left       = int.from_bytes(
                      f.read(2), 
                      byteorder="little"
                  )
      id.top        = int.from_bytes(
                      f.read(2), 
                      byteorder="little"
                  )
      id.width      = int.from_bytes(
                      f.read(2), 
                      byteorder="little"
                  )
      id.height     = int.from_bytes(
                      f.read(2), 
                      byteorder="little"
                  )
      fieldByte    = f.read(1)[0]
  
      id.localColorTableFlag = bool(fieldByte >> 7)
      id.interlaceFlag       = bool((fieldByte>>6)&1)
      id.sortFlag            = bool((fieldByte>>5)&1)
      id.reserved            = bool((fieldByte>>4)&1)
      id.lctSize             = 2**(fieldByte&0xf)
      id.lct = [
              Color(*struct.unpack("<BBB",f.read(3))[:3]) 
              for _ in range(id.lctSize)]
  
      id.imgDataStart = f.tell()
  
      return id

  def getImageDataLength(self) -> int:
      '''
      Returns the size of the image data as determined by the 
      LZW decoder. This may be different to the actual amount 
      of data in the file due to compression artifacts or padding.
      Use.getImageData() to read all data.
      '''
      imgdatapos = self.imgDataStart
      xpos = ypos = 0
      lzwbuffer = {} # lookup buffer can hold up to 4097 entries
      nextcode = 0 # first code in the table
      oldpixel = None
      imglen = 0 # how many bytes we have read so far
      rowstart = True
      bitsread = 9 # how many bits left to read
      maxbits = 12 # maximum bits allowed per scanline

      while True:
          c = ord(self.imageData[imgdatapos])
          imgdatapos+=1
          imglen+=1
          if not rowstart:
              if c >= len(self.lct):
                  raise Exception("Pixel index too large.")
    
              pxcol = self.lct[c]

              f.seek(-1, io.SEEK_CUR)
              f.write(pxcol.toBytes())
    
              # If it's the same pixel repeatedly, use run length encoding!
              if oldpixel==None or pxcol!=oldpixel:
                  f.seek(-1,io.SEEK_CUR)
                  f.write(pxcol.toBytes()[:1])
              
              oldpixel = pxcol
          
          if c == 0:
              break;

          elif c <= 12:
              for i in range(c+1):
                  pixel = self.lct[i]
            
                  f.seek(-1, io.SEEK_CUR)
                  f.write(pixel.toBytes()[:1])

                  if rowstart:
                      imglen+=1
            
          elif c == 12:
              f.seek(-1, io.SEEK_CUR)
              f.write(b'\x00\xff\xff')
              if rowstart:
                  imglen+=3
                    
          elif c == 13:
              f.seek(-1, io.SEEK_CUR)
              f.write(b'\x00\x00\xff')
              if rowstart:
                  imglen+=3
                      
          elif c == 14:
              f.seek(-1, io.SEEK_CUR)
              f.write(b'\x00\x00\x00')
              if rowstart:
                  imglen+=3
              
          elif c == 15:
              f.seek(-1, io.SEEK_CUR)
              f.write(b'\xff\x00\x00')
              if rowstart:
                  imglen+=3
        
          else: # We got an encoded value
              val = c
              bitsneeded = 9 # How many bits needed for this code?
    
              # Get the correct code
              for i in reversed(range(1,17)):
                  if val&(1<<i):
                      val -= 1<<i
                      val |= ~((~0)<<i)
                      bitsneeded = i
                      break
                  
              mask = (1<<(maxbits-bitsneeded))-1
              val &= mask
              
              s = ""
              if val == nextcode: # Repeat code from last time
                  decode = lzwbuffer[nextcode]
              else: 
                  decode = chr(val)+chr(nextcode)
                  nextcode = val
              lzwbuffer[nextcode]=decode
              msbits = [(nextcode>>(n-1))&1 for n in reversed(range(bitsneeded))]
              decoded = list(map(lambda bit:bit=='1',msbits))
              for bit in decoded:
                  if rowstart:
                      imglen+=1
                  s+="0"+hex(bit)[2:]
                
              f.seek(-len(s)//2, io.SEEK_CUR)
              f.write(bytes.fromhex(s))
              
          rowstart = False
          if bitsread < maxbits:
              bitsread+=bitsneeded
          else:
              bitsread = 0
              rowstart = True
  
      return imglen


  def getImagePixels(self) -> bytes:
      '''
      Gets the pixel values of the image without decoding
      the compression scheme used.
      '''
      pixels = b''
      imgdatapos = self.imgDataStart
      xpos = ypos = 0
      lzwbuffer = {} # lookup buffer can hold up to 4097 entries
      nextcode = 0 # first code in the table
      oldpixel = None
      imglen = 0 # how many bytes we have read so far
      rowstart = True
      bitsread = 9 # how many bits left to read
      maxbits = 12 # maximum bits allowed per scanline

      while True:
          c = ord(self.imageData[imgdatapos])
          imgdatapos+=1
          imglen+=1
          if not rowstart:
              if c >= len(self.lct):
                  raise Exception("Pixel index too large.")
    
              pxcol = self.lct[c]
              pixels += pxcol.toBytes()

              # If it's the same pixel repeatedly, use run length encoding!
              if oldpixel==None or pxcol!=oldpixel:
                  pixels += pxcol.toBytes()
              
              oldpixel = pxcol
          
          if c == 0:
              break;

          elif c <= 12:
              for i in range(c+1):
                  pixel = self.lct[i]
            
                  pixels += pixel.toBytes()

                  if rowstart:
                      imglen+=1
            
          elif c == 12:
              pixels += b'\x00\xff\xff'
              if rowstart:
                  imglen+=3
                    
          elif c == 13:
              pixels += b'\x00\x00\xff'
              if rowstart:
                  imglen+=3
                      
          elif c == 14:
              pixels += b'\x00\x00\x00'
              if rowstart:
                  imglen+=3
              
          elif c == 15:
              pixels += b'\xff\x00\x00'
              if rowstart:
                  imglen+=3
        
          else: # We got an encoded value
              val = c