#!/usr/bin/env python3

import sys
import os
os.chdir(sys.path[0])

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The width is determined by
# `width`, and the height will be an integer between minheight and maxheight.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width: int, minheight: int, maxheight: int
    ):
    # Write the magic string (header) to the output stream
    out.write(b'GIF89a')

    # Determine how many frames we need
    nframes = ((width * minheight)//maxheight)+((width*(minheight%maxheight))>0)
    print("Generating {} frame{}...".format(nframes, "" if nframes==1 else "s"))

    # For each frame...
    for i in range(0, nframes):

        # Compute the size of this frame's palette
        palsize = (((i+1)*(i+2))//2)-1;

        # Write the frame header info to the output stream
        out.write(struct.pack('<HHB',
                width & 0xffff,
                (height := maxheight if i < (nframes-1) else (width*minheight)%maxheight) & 0xffff,
                8 | (0<<7)))

        # Mark this as a local color table
        out.write(b'\x04')

        # Write this frame's palette to the output stream
        out.write(b''.join([rng.read(3) for _ in range(palsize)]))

        # Write the LZW minimum code size
        out.write(b'\x02')

        # Start packing with 0xAC
        currentcode = 0xac

        # Figure out the maximum number of data blocks that can fit per scan line
        datablocks = width // 8

        # Build up the LZW table with all single byte prefixes
        lzwtable = {byte:(byte,) for byte in range(256)}

        # Pack 1 block from every scanline
        for y in range(height):

            # Clear any remaining data blocks on oddly sized lines
            if not datablocks:
                datablocks += 1

            # Initialize the buffer list of packed values
            buf = []

            # Write the initial clear code to the buffer
            buf.append(currentcode)

            # For each data block in the row...
            for x in range(datablocks):

                # Read one more byte than needed so we can check if we've hit EOI or
                # if we should pad our buffer to multiples of 12 bits
                data = [rng.read(1)[0] for _ in range(1+x)]

                # If we have enough data then decode it
                while len(buf)+(len(data)*8) > 12:

                    # Get the encoded value
                    code = 0
                    for d in reversed(data[:12]):
                        code <<= 8
                        code |= d

                    # Attempt to use this decoded value in the input buffer
                    try:

                        # Append these characters to the buffer
                        buf.extend(lzwtable[code])

                        # Remove them from the input before continuing!
                        del data[:len(lzwtable[code])]
                    except KeyError:
                        pass

                    # Try again on the same data
                    continue


                # If we're at end of image then flush the buffer and return
                if not data:
                    break

                # Grab a new prefix byte
                prefix = data[-1]

                # Update the LZW table
                lzwtable.update({nextcode: (prefix,)+lzwtable[prefix]})

                # Set the next available code
                nextcode = 0 if len(lzwtable)==pow(2,8) else pow(2,8)-len(lzwtable)

                # Add the character to the buffer
                buf.append(prefix)

                # Encode over the rest of the data buffer
                for d in data[:-1]:
                    code <<= 8
                    code |= d
                    buf.append(code)

            # Wrap the buffer around when we've reached the end
            buf = buf[:]
            buf.append(currentcode)

            # Encode the final buffer contents
            buf = b''.join([(len(val)<<4)|val for val in buf])

            # Write this line's packed pixels to the output stream
            out.write(buf)



