#!/usr/bin/env python3

import argparse
from math import ceil
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). Returns
# the total size of the GIF file written.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width: int=None, height: int=None
    ) -> int:

    # First byte is always GIF identifier "G"
    GIF_ID = b'G'
    
    HEADER_SIZE = 12  # Bytes in minimal GIF header

    FRAME_HEADER_SIZE = 10  # Bytes in frame header
    PIXEL_BLOCK_HEADER_SIZE = 2  # Minimum bytes in pixel block header
    MINIMUM_PIXELS_PER_LINE = 1  # Minimum pixels per line must be at least one

    # The maximum data that can appear in a single pixel block;
    # all other blocks will start with this value followed by
    # a random sequence of lesser values
    MAX_CODELEN = 2 ** 8

    # Maximum decoded image width
    MAX_WIDTH = 32000

    # Maximum decoded image height
    MAX_HEIGHT = 32000

    # Default amount of lines of clear code
    CLEAR_CODE_SIZE = 12

    # Size of LZW minimum code stream buffer
    LZW_BUFFER_SIZE = 2 ** 12

    # Amount of default LZW initial codestream buffer fill
    DEFAULT_FILL_AMOUNT = 2 ** 14

    def bit_length(i): return 0 if not i else 1 + bit_length(i >> 1)

    def next_power_of_two(i): return 2**(bit_length(i) - 1)

    assert CLEAR_CODE_SIZE <= MAX_CODELEN

    # Initialize variables for decoding process
    bufsize = LZW_BUFFER_SIZE
    fillval = 0
    fillamt = DEFAULT_FILL_AMOUNT
    eoi = False

    # Start out with no decoded image dimensions
    w = h = -1

    # Generate the header and add to output
    header = GIF_ID + b'I' + b'\x00\x00\x00\x00\x00' + b'\xff' + b'\xff\xff'
    out.write(header)

    # Read global color table length from command arguments
    numcolors = args.numcolors

    # Add GCT to output if we have it
    if numcolors > 0:
        gct = rng.read(numcolors * COLOR_DEPTH // BITS_IN_BYTE)
        out.write(gct)

    # Make sure we're starting fresh with each frame
    imgdatalen = 0
    codelen = None
    dispose = None
    prevframe = ''

    stack = ['']
    dict = {CLEAR_CODE_SIZE + i : chr(i) for i in range(CLEAR_CODE_SIZE)}
    maxkey = CLEAR_CODE_SIZE

    # When giving the user two dimensions, use them both
    if type(width) in (list, tuple):
        width = width[0], width[1]
        height = height[0], height[1]

    # If only one dimension was given, assume square image
    elif width and not height:
        height = width

    # If no dimension was given, choose a random one
    if not width and not height:
        width = rng.randint(MINIMUM_IMAGE_LENGTH,
                            min(MAX_WIDTH,
                                next_power_of_two(rng.getrandbits(16))))

        height = rng.randint(MINIMUM_IMAGE_LENGTH,
                             min(MAX_HEIGHT,
                                 next_power_of_two(rng.getrandbits(16))))

    # Encode random image starting at top-left corner. We encode our images
    # bottom up instead of top down so that they'll look like this when
    # displayed properly: https://imgur.com/a/OVvKZKu. Note that since our
    # encoding algorithm starts with the most significant bits, those are
    # shown first when encoded in binary.
    xpos = ypos = 0
    while xpos < width:
        while ypos < height:
            color = rng.read(COLOR_DEPTH // BITS_IN_BYTE)
            
            if xpos!= 0:
                # We want to make sure we send enough space between lines
                # so that we never send more than 255 colors per line.
                # Since we encode our images bottom up, we have to keep track
                # of the amount of space left before we reach the line's end.
                bufrem = (max(0, bufsize - len(stack)))
                if bufrem == 0:
                    out.write(bytes([clear]))
                    
                elif bufrem < CLEAR_CODE_SIZE + MAX_CODELEN + 2:
                    for j in range(bufrem):
                        out.write(bytes([clear]))
                        
                # Fill remaining space with incrementing values, which is good
                # for testing purposes. It should produce very repetitive
                # outputs!
                #for j in range(bufrem):
                #    out.write(bytes([dict[''.join((str(x), str(ypos)))]]))
                
                # Reset buffer's size back to its normal value.
                bufsize = LZW_BUFFER_SIZE
                    
            # Send line clear code
            out.write(bytes([clear]))
            
            # Flush buffer content
            if len(stack) > 0:
                out.write(''.join(stack).encode())
            del stack[:]
            
            # We also need to signify the EOI code at the end of every line
            out.write(bytes([eoi]))
            
            # Increment current horizontal position
            xpos += 1

        # Reached end of line, reset vertical position counter
        ypos += 1
        xpos = 0
        
    # Return size of whole GIF file written to output
    return out.tell()