#!/usr/bin/env python3

import os
import sys
import random
import datetime as dt

gif_header = bytes([
  int('0x47', base=16),
  int('0x49', base=16),
  int('0x46', base=16),
  89,
  97,
  103])


# Generates a random GIF file into `out` using the
# random number generator `rng`.
def generate_random_gif(rng: object, out: object): 
    """Generate a random animated GIF given a pseudo-random source"""    
    # Write header
    print("Generating header...")
    out.write(gif_header)

    # Print some statistics about how far through generation we are
    def report_progress():
        percent = round((pos / numbytes)*100, 3)
        sys.stdout.flush()
        sys.stdout.write("\r{:0>5}% complete".format(percent))

    # Get the current date and time
    now = dt.datetime.now().timetuple()

    # year (1970–2069)
    gif_year = bytes([
        now.tm_year - 1970 + 1,
        0 ])
        
    # month (January = 1; December = 12)
    gif_month = bytes([
        now.tm_mon, 
        now.tm_mday ])

    # hour (0-23)
    gif_hour = bytes([
        now.tm_hour, 
        0 ])

    # minute (0-59)
    gif_minute = bytes([
        now.tm_min, 
        0 ])

    # second (0-59)
    gif_second = bytes([
        now.tm_sec, 
        0 ])

    # delay (in hundredths of a second)
    delay = random.randint(5, 15)

    # color table size (2^(N+1)) x 3 entries
    color_table_size = 2 ** random.randint(1, 4) * 3

    # color resolution
    color_resolution = random.choice([8, 16, 24, 32])

    # sort flag
    sort = random.randint(0, 1)

    # global palette flag
    gcflag = random.randint(0, 1)

    # background color index
    bgindex = random.randint(0, 2 ** color_resolution - 1) 

    # pixel aspect ratio
    pixaspectratio = random.randint(0, 2)

    # pack header fields together
    lsd = gif_year + \
             gif_month + \
             gif_hour + \
             gif_minute + \
             gif_second + \
             bytes([color_table_size,
                   (color_resolution - 1 << 5) + 
                   (sort << 3) + 
                    gcflag]) + \
             bytes([(bgindex & 255)]) + \
             bytes([(pixaspectratio & 255)])

    # write LSD
    print("Writing logical screen descriptor...")
    out.write(lsd)

    # Global color table
    gct = [b'\x00\x00\xFF']*color_table_size

    # Local color table
    lct = []

    # If there is one, add local color table
    if len(gct) > 0:
        print("Writing global color table...")
        out.write(bytes(gct))

    # Image block data
    imgblocks = []

    # Define subtypes
    graphics_control     = bytes([0xF9, 4]) + \
                           bytes([delay,
                                  0b11101100,
                                  bgindex]) + \
                            bytes([0x00])
            
    application_extension = bytes([0xFF, 0x0B,
                                   0x4A, 0x46, 
                                   0x49, 0x46, 
                                   0x00, 0x01, 
                                   0x01, 0x00,
                                   0x00]) + \
                            str(random.randint(-1000, 1000)).encode() + \
                            bytes([0x00])

    plaintext            = "Hello World!".encode()

    # Figure out which blocks we will use
    blocks = {}
    for i in range(15):
        blocks['application_extension'] = False
        blocks['plaintext']             = False
        blocks['graphics_control']      = False

    # Randomly choose what blocks we're going to put in the GIF file
    for c in list(blocks.keys()):        
        if random.randint(0, 1):
            blocks[c] = True

    # Send chosen blocks to output file
    pos = 0
    numbytes = max(map(len, blocks.keys()))
    for name, choice in blocks.items():
        progress = ""
        if (pos % 5 == 0):
           progress = "\n"

        if choice: 
            print(f"{progress}Sending {name}")

            # Get contents of block
            if name == 'application_extension':
                content = application_extension
            elif name == 'plaintext':
                content = plaintext
            else: 
                content = graphics_control

            # Create new block
            block = bytes([])
            
            # Record position within GIF
            saved_position = out.tell()
                
            # Go back to start of block
            out.seek(saved_position + 2, os.SEEK_SET)
                   
            # Write length of block
            block += struct.pack('B', len(content))

            # Write contents of block
            block += content
        
            # Seek forward to end of block
            out.seek(saved_position + len(block),os.SEEK_SET)

            # Write block to file
            out.write(block)

        report_progress()
            
        pos += 1

    print("")