#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # TODO use more than 4 colors and use them all
    palette = bytes([
                0xFF, 0x00, 0x00,   # black
                0xCC, 0x00, 0x66,   # brown
                0x66, 0x00, 0x66,   # red
                0xFF, 0x66, 0x00,   # orange
                0x00, 0x66, 0x66,   # green
                0x99, 0x33, 0x33,   # tan
                0x00, 0x00, 0x99,   # blue
                0x99, 0x00, 0x00    # purple
                ]) * (7 // 4)
    
    background = 0
    width      = rng.read(1)[0] % 512 + 2
    height     = rng.read(1)[0] % 512 + 2

    frames = []
    for _ in range(int.from_bytes(rng.read(1), byteorder='big')):

        xblocks = width//8
        if width%8 > 0:
            xblocks += 1
            
        yblocks = height//8
        if height%8 > 0:
            yblocks += 1
        
        bitmap = bytes([(rng.read(1)[0] & 1) << i |
                        (rng.read(1)[0] & 1) << j for i in range(yblocks*xblocks)])

        delay = rng.read(1)[0]
        transparent = bool(rng.read(1)[0])
        disposal_method = rng.read(1)[0]
        user_input = bool(rng.read(1)[0])

        frame = datasize = b''
        exts = bytes()
        if not (transparent or user_input):
            data = b''

            for yblock in range(yblocks):

                y = yblock * 8

                for xblock in range(xblocks-1,-1,-1):

                    x = xblock * 8
                    blocksize = min((x+8)-width,(y+8)-height)
                    
                    subbitmaps = []
                    dctab      = bytes()
                    actab      = bytes()
                    numcolours = 0
                    lzwmincodesize = 3
                    
                    while True:

                        # Split into chunks of size cbsize
                        chunklen = rng.read(1)[0]
                        chunk = rng.read(chunklen)
                        
                        chunktype = chunk[0] & 0xF0
                        chunksize = chunk[0] & 0x0F
                        chunk = chunk[1:]

                        # Extension block
                        if chunktype == 0x20:
                            blockid = chunk[:3]

                            # Graphic control extension
                            if blockid == b'GCE':
                                flags = ord(chunk[3])

                                delay = struct.unpack('!H', chunk[4:6])[0]
                                transparent = flags&0b0001!= 0
                                
                                disposal_method = ord(chunk[6])
                                user_input = flags&0b0010!= 0
                            
                            elif blockid == b'COM':
                                pass

                            else:
                                raise Exception("Unknown extension")

                        # Image descriptor
                        elif chunktype == 0x2C:                    
                            left = ord(chunk[0]) | ((ord(chunk[1]) & 0xF0)<<4)
                            top  = ord(chunk[1]) & 0x0F | (((ord(chunk[2]) & 0xF0)<<4))
                            width  = ord(chunk[2]) & 0x0F | (((ord(chunk[3]) & 0xC0)<<2))
                            height = ord(chunk[3]) & 0x0F
                            interlace = bool(ord(chunk[4]) & 0b10000000)
                            sort      = bool(ord(chunk[4]) & 0b01000000)
                            reserved  = bool(ord(chunk[4]) & 0b00100000)
                            lctflag   = bool(ord(chunk[4]) & 0b00010000)

                            if reserved:
                                raise Exception("Reserved bit set")

                            if sort:
                                raise Exception("Local colour table sorting is unsupported")

                            if lctflag:
                                raise Exception("Interlace flag unsupported")
                                    
                            lctsize = ord(chunk[5]) & 0x0F

                            # Decompress the packet
                            bitmapoffset = (top*(width+1)) + left
                            newsubbitmap, next_lzwmincodesize = decompressLZWSlowly(
                                    chunk[6:], chunk[5], lzwmincodesize, bitmap[bitmapoffset:])
                            
                            subbitmaps.append(newsubbitmap)
                            
                            # Create a local colour table from the first block
                            if not lctflag and lctsize!= 0:
                                lct = chunk[6:6+(3*lctsize)]
                                lctdict = {tuple(lct[i:i+3]): i for i in range(0,len(lct),3)}
                                actab += bytes([i for _,i in sorted(lctdict.items(), key=lambda kv:-kv[1])][:2**lctsize])
                                numcolours = 2**lctsize
                            
                        # End of image
                        elif chunktype == 0x3B: 
                            break
                        
                    
                        if chunktype >= 0x20 and chunktype <= 0x3B:
                            continue

                        else:
                            raise Exception("Unknown data type {}".format(hex(chunktype)))


                    bitsizes = [max(subbitmap.count(0),subbitmap.count(1)) 
                                for subbitmap in subbitmaps]                    
                    maxnumcodes = max(*bitsizes)*2
                    codesdict = {}            
                    codestoplevels = []
                    lzwmincodesize = next_lzwmincodesize
                    
                    if maxnumcodes < (1<<next_lzwmincodesize)+1:
                        next_lzwmincodesize -= 1
                        continue
                    
                    
                    # Create the dictionary and stoplevel values        
                    for subbitmap in subbitmaps:
                        numcoded = 1<<(lzwmincodesize+1) - 1
                        numuncoded = numcoded + 1
                        for colorindex in range(len(actab)):
                            
                            codestoplevel = False
                            
                            if colorindex >= (1<<lzwmincodesize):
                                if left!=0:
                                    prevpixel = bitmap[(top*(left+width))+left-1]
                                elif top!=0:
                                    prevpixel = bitmap[((top-1)*(left+width))+right]
                                else:
                                    prevpixel = 0
                                codestoplevel = tuple(prevpixel) == colorindex
                            codestoplevels.append(codestoplevel)
                            
                            if codestoplevel:
                                dictvalue = numcoded 
                                numcoded += 1
                            else:
                                dictvalue = numuncoded
                                numuncoded += 1
                            
                            codesdict[colorindex] = dictvalue

                    # Flatten the list of all subbitmaps
                    frame += bytes(sum([[colorindex]*size 
                        for colorindex, size in zip(actab,bitsizes)],[]))
                    
                    # Write the remaing LZW blocks
                    def unflatten(colorsused):
                        numcodes = sum(sizes)
                        codesdictlist = [(colorindex, abscode) 
                                        for sizes, colorindex, abscode in
                                        zip(bitsizes, colorsused,
                                            reversed(range(1,2**(lzwmincodesize+1))))]
                        codesdict = dict(codesdictlist)
                        
                        for subbitmap, codestoplevel in zip(subbitmaps, codestoplevels):
                            if codestoplevel:
                                yield from subbitmap

                            else:
                                peek = iter(subbitmap).__next__
                                prevpix = (-1,-1,-1)

                                currentpos = 0
                                while currentpos < len(subbitmap):

                                    if peek()!=0 and peek()!=1:
                                        raise Exception("Invalid pixel value at index "
                                                        "{}".format(currentpos))

                                    bitstr = ''
                                    while bitstr=='':

                                        currentbyte = peek()
                                        currentpos+=1
                                        
                                        if codestoplevels[currentbyte]:
                                            yield from subbitmap[currentpos:(currentpos+bitpos)]
                                            
                                            if currentpos == len(subbitmap):
                                                return

                                            yield prevpix[currentbyte]
                                            bitpos = 0
                                            prevpix = (-1,-1,-1)

                                        elif prevpix==(-1,-1,-1):
                                            prevpix = (
                                                    subbitmap[currentpos]&1,
                                                    subbitmap[currentpos]&2,
                                                    subbitmap[currentpos]&4)

                                        else:
                                            bitstr = bin(peek())[::-1][3:].zfill(bpp)

                                    
                                    currentpos += bitpos
                                    
                                    try:
                                        nextidx = codes