#!/usr/bin/env python3

import struct
from typing import BinaryIO
from string import hexdigits
from pprint import pprint

def gen_rand_color():

    return [int(hexdigits[int(c)],base=16) 
            for c in ''.join(['00','ff'])]

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):


    def get_keystroke() -> str:
        char = chr(rng.read(1)[0])
        
        if not char == '\x1b':

            return char

        elif char == '\x1b' and len(char) > 1:

            char += chr(rng.read(1)[0])

            if char == '[M':
                return {
                    'up': 'u',
                    'down':'d',
                    'right':'r',
                    'left':'l'}[''.join([chr(i) 
                                        for i in rng.read(3)])]
                

            elif char == '[A':
                return 'u'

            elif char == '[B':
                return 'd'

            elif char == '[C':
                return 'r'

            elif char == '[D':
                return 'l'

        else:
            raise Exception("invalid keypress")

    

    color_map = list([(gen_rand_color(),'') 
                      for _ in range(256)])
    

    # set up screen descriptor
    scrn_desc = bytearray(struct.pack('>HHBBBHBHH',
        0, 
        0, 
        ord(' '), 
        ord('X'), 
        7, 
        0xf0 | 0x1f, 
        0, 
        0, 
        0, 
        0))

    # update the palette with new colors

    pal_byte = []

    for col,coldesc in enumerate(scrn_desc[:7]):

        if type(col) is tuple:
            color = col
            desc = coldesc

        else:
            color = color_map[col][0]
            desc = coldesc

        pal_byte += color

        print('pal[{0}] {1}'.format(
            col, hex(desc)))

    scrn_desc[7:] = pal_byte

    # update width based on window size, 
    # otherwise this will be constant

    winsize =  os.get_terminal_size((80,25))
    scrn_desc[0:2] = [winsize[0],
                      winsize[1]]

    print('scrn_desc',scrn_desc)
    print('width',winsize[0])
    print('height',winsize[1])



    # clear variables etc.
    disp_ctrl = 0
    gct_size = None
    gct = []
    lct_flag = False
    lct_size = None
    lct = {}
    img_desc = []
    img_data = ''
    last_line = ''
    pixel_buffer = []
    line_counter = 0
    show = True
        
    # Read in keystrokes until we hit escape
    while True:
    
        try:
            char = get_keystroke()
        except KeyboardInterrupt:
            print('\n\ngoodbye\n\n')
            break

        if char == '':
            continue

        elif char == 'q':
            print('\n\ngoodbye\n\n')
            break

        elif char == 'p':
            pass

        # toggle display ctrl flags
        elif char =='':
            disp_ctrl ^= 0x20

        elif char == '+':
            disp_ctrl ^= 0x10

        elif char == '-':
            disp_ctrl ^= 0x08

        elif char == 'o':
            disp_ctrl ^= 0x04

        elif char == '*':
            disp_ctrl ^= 0x02

        elif char == '!':
            disp_ctrl ^= 0x01

        # handle global color table
        elif char == '%':
        
            gct_size = int(input('gct_size?'))

            print('reading gct...')

            for j in range(gct_size):
                
                rgb = [int(x)
                       for x in input().split()]

                gct += rgb

                color_map[j] = ((rgb,
                                 '?'),
                                '')


            print('done!')


        # handle local color tables
        elif char == '&':
        
            lct_flag = bool(input('[y|N]?'))

            if lct_flag:
                lct_size = int(input('lct_size?'))

                print('reading lct...')

                for j in range(lct_size):
                    
                    rgb = [int(x)
                           for x in input().split()]

                    lct[(j//3)*4+j%3] = ((rgb,
                                          '?'),
                                         '')


                print('done!')

        
        # create image description    
        elif char == '#':
        
            img_desc = [int(x)
                        for x in input().split()]

            print('creating image ',img_desc)

            final_pixels = []

            for y in range(img_desc[-1]):
            
                row = []

                for x in range(img_desc[0]):

                    row.append(color_map[0])

                final_pixels.append(row)
                
            img_data = final_pixels



        # handle drawing commands
        elif char == '@':
        
            x = int(input('x coord?'))
            y = int(input('y coord?'))
            color = int(input('color index?'))

            painted = False

            for rect in pixel_buffer:


                if all([rect[0][0] < x <= rect[2][0], 
                        rect[0][1] < y <= rect[2][1]]):
                    
                    print('painting at ({},{})...'.format(
                            x, 
                            y))
                    img_data[y][x] = color_map[color]
                    painted = True

            if not painted:
                pixel_buffer.append(((x,y),
                                    (x+1,
                                     y),
                                    (x+1,
                                     y+1),
                                    (x,
                                     y+1)))

        elif char == '$':

            saved = img_data[:]

            for pbuf in pixel_buffer:

                for k in range(*pbuf):
                    img_data[k[1]][k[0]] = color_map[saved[k[1]][k[0]]]

            del pixel_buffer[:]

            print('\ndone!\n')

        elif char == '^':
            if savefile!= '':
                print('saving img_data to {}'.format(savefile))
                pickle.dump(img_data,open(savefile,'wb'))

        elif char == '>':
            savefile = input('filename?')

        elif char == '<':
            loadfile = input('filename?')
            if loadfile == '':
                img_data = [['?' for _ in range(img_desc[0])] for _ in range(img_desc[1])]
            else:
                print('loading img_data from {}'.format(loadfile))
                img_data = pickle.load(open(loadfile,'rb'))
            
        elif char == '~':
            img_data = [[('white',
                          ''),
                         ('black',
                          ''),
                         ('red',
                          ''),
                         ('green',
                          '')],
                        [('yellow',
                          ''),
                         ('blue',
                          ''),
                         ('purple',
                          ''),
                         ('cyan',
                          '')]]

        elif char == '_':

            draw_rows = []
            maxlen = 0
            for idx,pix in enumerate(img_data):
                for pixidx,px in enumerate(pix):
                    cur_str = px[1]
                    if len(cur_str)>maxlen:
                        maxlen = len(cur_str)
                    draw_rows.append('#{0}:{1}{2} '.format(
                            idx*2+1,
                           ''*(maxlen-len(cur_str)),
                            cur_str))
            draw_lines = ['|{:^4}|{:^'+str(maxlen)+'s}|'.format(*line) 
                          for line in zip(range(1,
                                                 len(draw_rows)+1),
                                          draw_rows)]
            draw = '|\n'.join(draw_lines)
            print(draw+'|',
                  end='\r')



