#!/usr/bin/env python3


import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Pick a width and height
    w = rng.read(1)[0]
    h = rng.read(1)[0]
    # Pick an RGB color
    r = rng.read(1)[0]
    g = rng.read(1)[0]
    b = rng.read(1)[0]

    # Now that we have our parameters, let's create some GIF data
    # We're going to make a 1 second, 1 frame animation with one pixel
    # The frame will be the size of our chosen window
    # The background color will be black (R:0 G:0 B:0)
    # The only drawn object will be this one pixel
    # This frame would contain only one single byte (the color)
    # To make it repeat infinitely, we will set the Gif Loop Variable to
    # something that makes sense when doing it manually (i.e. not 0 or ff)
    # Otherwise, we will use the special value "ff" which means infinite
    header = [
        71, 73, 70, 56, 57, 97, 
        48, 0, 0, 0, 0, 0, 0, 255, 0, 0, 0, 0,
        0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 44, 0, 0, 0, 0, 1, 0, 1, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 44, 0, 0, 0, 0, 1, 0, 1,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
        ]
    # Add bytes to header to represent the screen size
    header += [(w >> 8) & 0xFF, w & 0xFF] 
    header += [(h >> 8) & 0xFF, h & 0xFF] 

    # Add bytes to header to represent the palette
    header += [r,g,b,0,0]

    # We finally reach the actual data block!
    pixdata = [0, 0, 0, b]

    bodylen = len(pixdata) + 11
    framep1 = [17, 73, 73, 42,
                0, 0, 
                0, bodylen // 256, 
                bodylen & 0xFF,
                33, 0,
                0, 0,
                0, 0,
                255, 0,
                0, 0,
                0, 0,
                255, 0,
                255, 0,
                255, 0,
                255, 0,
                0]
    framep2 = [
            0, w & 0xFF, 
            (w >> 8) & 0xFF, 
            0, h & 0xFF, 
            (h >> 8) & 0xFF,
            # Pixels per millimeter (coord units)
            100, 0, 0, 0,
            0, 0, 0, 0,
            0, 0, 0, 0,
            0, 0,
            # Graphics Control Extension
            249, 0,
            11,
                
                # Value of delay time
                0, 10,

                # Transparent Color Index
                badval,

            # Image Descriptor
            255, 255,
            255,
            
            0, 0,

            0, 0,

            0, 0,

            0, 0,

            0, 0,

            0, 0,

        ]

    out.write(bytes(header))
    out.write(bytes(framep1))
    out.write(bytes(pixdata))
    out.write(bytes(framep2))

    return


