#!/usr/bin/env python3

import sys
import os
os.chdir(sys.path[0])

import struct
from typing import BinaryIO

def generate_random_file(
        rng: BinaryIO, size: int, out: BinaryIO
    ) -> None:
    bufsize = min(4096, size)
    while True:
        chunk = rng.read(min(bufsize, size))

        if not chunk or (len(chunk) < bufsize):
            break
        
        assert len(chunk) <= bufsize

        out.write(chunk)
        size -= len(chunk)


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Write the magic string (header) to the output stream
    out.write(b'GIF89a')

    # Generate some data for the body
    data = bytearray()
    
    # Create the header
    width = rng.randint(1, 10000)
    height = rng.randint(1, 10000)

    colorbits = [2 ** i for i in reversed(range(7))]
    globalcolorcount = rng.choice([True] * 5 + [False])
    localcolorcount = rng.randint(1 << colorbits[-1], 1 << max(colorbits)) | globalcolorcount
    bgcolori = rng.randint(0, localcolorcount - 1)

    gctflag = bool(rng.getrandbits(1))
    bgci = rng.randint(0, localcolorcount - 1) if not gctflag else 0
    pixelaspectratio = rng.randint(0, 10) // 10

    flags = []
    flags += [(bool(rng.getrandbits(1)), 'gctsort', False)]
    if sum(c for f, _, c in flags) > 1:
        raise ValueError('Too many yes-no choices!')
    flags += [(not globalcolorcount, 'globalcolortable', False)]
    flags += [(bool(rng.getrandbits(1)), 'interlace', False)]
    flags += [(bool(rng.getrandbits(1)), 'localcolortable', False)]
    if sum(c for f, _, c in flags) > 1:
        raise ValueError('Too many yes-no choices!')

    lctSizes = [
        None,
        1<<colorbits[-1],
        1<<(colorbits[0]+rng.randint(-2, 3)),      # ±2
        1<<(colorbits[1]+rng.randint(-2, 3)),      # ±2
        1<<(colorbits[2]+rng.randint(-2, 3)),      # ±2
        1<<(colorbits[3]+rng.randint(-2, 3)),      # ±2
        1<<(colorbits[4]+rng.randint(-1, 2)),      # ±1
        1<<(colorbits[5]+rng.randint(-1, 2)),      # ±1
        1<<(colorbits[6]+rng.randint(-1, 2)),      # ±1
        2**rng.randint(7, 12),                      # 2^x
        2**rng.randint(12, 15)+rng.randint(1, 8),   # 2^x + 1..8
        2**rng.randint(15, 18)+rng.randint(1, 4),   # 2^x + 1..4
        2**rng.randint(18, 21)+rng.randint(1, 2)    # 2^x + 1..2
    ]
    lctSize = rng.choice([None] + list(filter(lambda x : x >= localcolorcount, lctSizes)))

    flags += [(lctSize is not None, 'haslocalcolortablesize', False)]
    flags += [(bgci!= 0, 'hasbackgroundcolorindex', False)]
    flags += [(pixelaspectratio!= 0, 'haspixelaspectratio', False)]

    packedflags = 0
    for flag, idstr, default in flags:
        if not flag: continue
        print('Setting ', idstr)
        packedflags |= 2**(flags.index((flag, idstr, default))+1)


    body = bytearray()
    body += width.to_bytes(1, 'little')              # Width
    body += height.to_bytes(1, 'little')             # Height
    body += packedflags.to_bytes(1, 'little')         # Flags
    body += b'\x00\x00'                               # Background Color Index     / None
    body += b'\x00\x00'                               # Pixel Aspect Ratio        / None
    body += bgci.to_bytes(1, 'little')                # Background Color Index    / Global Color Table index
    body += pixelaspectratio.to_bytes(1, 'little')    # Pixel Aspect Ratio        / N/A

    if globalcolorcount:
        for _ in range(localcolorcount):
            r = rng.getrandbits(8)
            g = rng.getrandbits(8)
            b = rng.getrandbits(8)

            body += r.to_bytes(1, 'little')            # Red
            body += g.to_bytes(1, 'little')            # Green
            body += b.to_bytes(1, 'little')            # Blue
            
    elif lctSize is not None:
        for i in range(lctSize):
            r = rng.getrandbits(8)
            g = rng.getrandbits(8)
            b = rng.getrandbits(8)
            
            body += r.to_bytes(1, 'little')            # Red
            body += g.to_bytes(1, 'little')            # Green
            body += b.to_bytes(1, 'little')            # Blue
    
    else:
        raise RuntimeError('BUG: this should never happen!')

    print('\nGenerated GIF:')
    print(body)

    out.write(body)


