#! /usr/bin/python3
"""Generates a random GIF file using the given RNG."""


import argparse
import os
import sys
from typing import Iterable, Tuple

import tqdm
from PIL import Image, UnidentifiedImageError
from byteparsing import get_byte_at_offset


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(dest='outputfile',
                        help='Output Gif File Name.')
    parser.add_argument('-n', '--num-frames', type=int, default=10**7,
                        dest='numframes', metavar='N',
                        help='Number of frames to output [Default : 10^7].')
    parser.add_argument('--seed', nargs='+', type=int, default=[],
                        dest='seeds', metavar='SEED',
                        help='RNG seed value or range')
    args = parser.parse_args()


    if not (os.path.exists('.git')):
        print("This script must be run from the root directory of the project.")
        return -1
    elif len(str(args.numframes)) > 10: # Arbitrary limit for num frames
        raise ValueError("--num-frames is too large!")
    else:
        try:
            with open(f'temp/{args.outputfile}', mode='wb') as out:
                maxblocksize=(1<<16)
                blocksize=max(min(args.numframes//4,(1<<24)),
                              (((1<<23)//(args.numframes+1))))
                print(f"Generating {args.numframes} frames...")

                seeds = []
                if any([x<0 for x in args.seeds]):
                    startval=-abs(min(args.seeds))*0.8
                else:
                    startval=min(args.seeds)*1.2
                stopval=max(args.seeds)*1.2

                while min(stopval,startval) < maxblocksize*10 and \
                      max(stopval,startval)>0:

                    if startval == stopval: break
                    value=None
                    if abs(startval)>=abs(stopval):
                        value=startval
                        startval+=1
                        seeds.append(-value)
                    else:
                        value=stopval
                        stopval-=1
                        seeds.append(value)

                    print('\tSeeds:',seeds)


                totalblocks={v:(args.numframes*v)//(1<<23)+1
                             for v in set(seeds)}
                print("\tTotal blocks:",totalblocks)
                for seed in tqdm.tqdm(set(seeds)):
                    currentframe=0
                    for _ in tqdm.trange(currentframe,
                                         currentframe+totalblocks[seed],
                                         unit='Frame'):

                        im = None
                        try:
                            im = next(get_random_image())
                        except UnidentifiedImageError:
                            pass


                        if im is None: continue
                        im.save(out,
                                save_all=True,
                                append_images=[im] *
                                        (1+(int.from_bytes(
                                            get_byte_at_offset(
                                                rng,
                                                0,
                                                offset=b'\xf8\xff\xe8\x01'),
                                            byteorder='big')& 0x07)),
                                duration=(int.from_bytes(
                                    get_byte_at_offset(
                                        rng,
                                        0,
                                        offset=b'\xf4\xff\xe4\x01'),
                                    byteorder='big')& 0x07 + 1),
                                loop=0,)
                    currentframe += totalblocks[seed]

            print(f"Done! Output file location:\n\t{out.name}")

        except KeyboardInterrupt:
            print("Keyboard interruption detected.\nExiting program...")



def get_random_image() -> Iterable[Tuple[int, int]]:
    """Yield an infinite, random image coordinate pair.

    order is preserved between iterations so this function could be used by
    multiple processes simultaneously.
    """
    count = 0
    while True:
        img = Image.new('RGBA',
                        size=(int.from_bytes(
                            rng.read(length=2), byteorder='big')),
                        color=(int.from_bytes(
                            rng.read(length=1), byteorder='big'),
                               int.from_bytes(
                                   rng.read(length=1), byteorder='big'),
                               int.from_bytes(
                                   rng.read(length=1), byteorder='big')))


        yield (count,img)
        count += 1





