#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The generated
# GIF will contain only the static elements as defined by
# the specification. One or more frames are added with each frame
# containing exactly one rectangle filled with a single color index.
# Frames may be repeated multiple times throughout the animation.
def generate_static_framed_gif(
        rng: BinaryIO, out: BinaryIO, width: int, height: int
    ):
    # Generate the header and add to output
    header = bytes([
        0x47, 0x49, 0x46, 0x38, 0x39, 0x61, # "GIF89a"
        0x00,                                 # Version
        0x00, 0x00,                           # Width 
        0x00, 0x00,                           # Height
        0xf0,                                 # Flags: 11 bits set to 1
        0xff,                                 # Background Color Index 
        0x00                                  # Pixel Aspect Ratio
    ])
    out.write(header)

    ext_graphic_ctl = bytes([
        0x21, 0xF9, 0x04,                   # Extension Block Marker + Length
        0x00,                                 # 0th bit indicates transparent flag off
        0x01,                                 # 1st bit indicates graphics control field present
        0x00,                                # reserved
        0x00, 0x00,                           # Graphic Control Field
        0x00, 0x00,                           # X position
        0x00, 0x00,                           # Y position
        0x00, 0x00                            # Width
    ])
    out.write(ext_graphic_ctl)

    # Get the logical screen descriptor (LSD) field
    lsd = bytes([
        0x00, 0x00,                         # Screen Width
        0x00, 0x00,                         # Screen Height
        0x08,                               # Bit depth: 2^8 pixel colors (i.e., 256)
        0x00,                               # Color resolution: not specified (i.e., real value is 1<<bitdepth)
        0x00,                               # Sorting method: not significant
        0x00,                               # GCT Flag: use global color table
        0xe0                                # GCT Size: 2^(n+1), i.e., 1<<(1+(8-1))
    ])
    out.write(lsd)

    # Add GCT to output if we have it
    gct = bytes()
    black = [0x00, 0x00, 0x00]
    white = [0xff, 0xff, 0xff]
    ncolors = 1 << (8 - 1)
    for i in range(0, ncolors - 1):
        ii = i % 6
        if ii < 3:
            gct += bytes(black)
        else:
            gct += bytes(white)
    out.write(gct)

    ext_comment = bytes([
        0x21, 0xFE,                          # Extension Block Marker + Length
        0x00, len("Generated by genstaticframes") # Comment
    ])
    comment = bytes("Generated by genstaticframes", 'utf8')
    out.write(ext_comment)
    out.write(comment)
    
    ext_plain_text = bytes([
        0x21, 0xFF,                          # Extension Block Marker + Length
        0x0b,                                # Label Block Type
        0x0c, 0x00,                          # Text Grid Left
        0x0c, 0x00,                          # Text Grid Top
        0x00,                                # Text Grid Cell Width
        0x00,                                # Text Grid Cell Height
        0x00, 0x00, 0x00,                    # Cell Text Color RGB
        0x00, 0x00, 0x00                     # Cell Background Color RGB
    ])
    out.write(ext_plain_text)


    def get_next_color():
        yield 0x01

    bgcolorind = next(get_next_color())

    # Add an empty frame to force decoder to read 
    # Logical Screen Descriptor first before continuing
    # This prevents us from having to parse up to start
    # of Image Data Block, which takes time sometimes.
    empty_rgb_table = bytes([(bgcolorind << 4) | bgcolorind]) * 128;
    empt_img = bytes([0x2C, 0x00,            # Image Separator
                       0x00, 0x00,           # Image Left Position
                        0x00, 0x00,           # Image Top Position
                        0x00, 0x00,           # Image Width
                        0x00, 0x00,           # Image Height
                       0x08,                 # LZW Minimum Code Size
                       bgcolorind << 4 | bgcolorind]);     # Global Color Table Indicator and Background Color Index
    out.write(empt_img);
    out.write(empty_rgb_table);
    out.flush();

    # Cut off at half normal video refresh rate (60 fps) so that playback
    # can be viewed quickly without waiting too long.
    cutofftime = 1 << 31 // 2

    # Frame count is reset when > 65536
    cntr = 0

    # Add frames until we reach max num frames
    while true:
       # We don't want the counter and timestamp byte to overflow, so just
       # increment instead of doing modulo math. 
       cntr += 1

       # Get timestamp byte
       ts = cntr & 0xff;

       # Timestamp is encoded as difference between current iteration
       # and previous iterations timestamp minus offset of zero. For now, assume
       # clock starts on zero.
       prevts = ts - 1;

       # Encode timestamp byte as difference
       encenc = 0;
       if prevts!= 0 :
           if prevts >= ts:
               encenc |= 0x80;
           encenc |= abs(prevts - ts);

           # If encoding results in negative, then invert bits. 
           if encenc & 0x80:
              encenc ^= 0xff; 

       # Write timestamp byte.
       out.write(bytes([encenc]))

       # Write graphic extension block
       ext_graphic_ctl = bytes([
            0x21, 0xF9, 0x04,   # Extension Block Marker + Length
            0x00,               # 0th bit indicates transparent flag off
            0x01,               # 1st bit indicates graphics control field present
            0x00,              # reserved
            0x00, 0x00         # Graphic Control Field
       ]);
       # TODO Maybe change delay?
       out.write(ext_graphic_ctl)

       # Write x, y positions
       out.write(struct.pack("<H", 0))
       out.write(struct.pack("<H", 0))

       # Write width, height
       out.write(struct.pack("<H", width))
       out.write(struct.pack("<H", height))

       # Write background color indicator
       out.write(bytes([bgcolorind]))

       # Write rgb palette entry for background color
       out.write(struct.pack('<BBB', *(black*3)));

       # Write image separator character
       out.write(bytes([0x2C]));

       # Write left, top coordinates
       out.write(struct.pack("<H", 0))
       out.write(struct.pack("<H", 0))

       # Write width, height
       out.write(struct.pack("<H", width))
       out.write(struct.pack("<H", height))

       # Write LZW minimum code size
       # TODO Should probably check that this is always 8 for our purposes...
       out.write(bytes([0x08]))

       # Write global color table indicator and background color index
       out.write(bytes([bgcolorind << 4 | bgcolorind]))

       # Populate rgb table
       rgbTable = []
       for _ in range(0, 128):
           colind = next(get_next_color())
           rb = colind % 3
           gb = (((colind / 3) ^ colind) % 3) * 2
           bb = (gb >> 1) ^ gb
           bg = (bb ^ colind) % 3
           rgbTable.append((rb, gb, bb, bg));

       rgbTable = bytes(sum(rgbTable, ()))

       # Write actual rgb values
       out.write(rgbTable)

       # Write image end marker
       out.write(bytes([0x00]))

       # Flush all of the above to disk immediately so that the
       # image is guaranteed to be displayed when we reach here.
       out.flush()

       # Terminate generation after requested number of frames has been reached.
       if cntr >= cutofftime:
           break