#!/usr/bin/env python3

import struct
from typing import BinaryIO


def generate_random_lzw_image(
        rng: BinaryIO, num_lines: int=100,
        min_line_len: int=50, max_line_len: int=100
    ) -> bytes:

    # LZW decoder state
    clearcode  : int = 1 <<  1
    endofinput : int = 1 <<  2
    eoi        : int = 1 <<  3
    tables     : int = 1 <<  4
    errcodes   : int = 1 <<  5

    tablesizebits     =   6
    nextsubblock      =   2**tablesizebits
    bitnum            =   0
    code              = -1
    prev               = -1
    output            = []

    blocktype          = rng.read(1)[0] & ~tables

    if (blocktype == tables):

        # Table size according to specification
        # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
        tablesize = (rng.read(1)[0] |
                          ((rng.read(1)[0])<<8)) + 1

        lentable = [rng.read(1)[0] for _ in range(
                min(tablesize+1, 2 ** tablesizebits))]
        disptable = [[None]*768 for _ in range(
                2 * (max(lentable)+1))]

        while len(output) < 2*min(tablesize+1, 2 ** tablesizebits):

            # First half of decode table
            for i in range(min(tablesize+1, 2 ** tablesizebits)):

                buf = bytearray([
                        (disptable[prev][i] >> 2) & 0b11111111,
                        (disptable[prev][i] << 6) & 0b11111111
                    ])

                # Add new symbol with length from
                # previous decode table entry
                for j in range((buf[-1] & 0b11000000)>>6):

                    c = rng.read(1)[0]
                    if not isinstance(c, int):
                        raise TypeError('Expected integer read.')
                    buf += bytes([c])

                # Finalize
                disptable[nextsubblock][i]=len(output)
                output += buf

            # Second half of decode table
            for i in range(min(tablesize+1, 2 ** tablesizebits)):

                buf = bytearray([(
                        disptable[(prev+(
                            pow(2,
                                tablesizebits))) %
                                    (pow(2, tablesizebits)+1)][i]>>8)&
                                0b11111111,
                                disptable[(prev+(
                                    pow(2,
                                        tablesizebits)))%
                                            (pow(2,tablesizebits)+1)][i]&
                                        0b11111111])

                # Add new symbol with length from
                # previous decode table entry
                for j in range((buf[-1] & 0b11000000)>>6):

                    c = rng.read(1)[0]
                    if not isinstance(c, int):
                        raise TypeError('Expected integer read.')
                    buf += bytes([c])

                # Finalize
                disptable[(nextsubblock+(
                                pow(2, tablesizebits))) %
                           (2*(max(lentable)+1))]=[len(output)]
                output += buf


            if tablesize < 2**tablesizebits and \
               tablesize >= (2**(tablesizebits-1)-1):

                # Use quick mode (no clear codes or EOF yet)
                clearcode = (-(1<<tablesizebits))
                endofinput = -(1<<(tablesizebits-1))
                tables |= 1<<(bitnum)

            elif tablesize <= (2**(tablesizebits-1)-1):

                # Use normal mode
                clearcode = -(1<<tablesizebits)
                endofinput = clearcode
                tables |= 1<<(bitnum)


    while True:

        # Clear codes
        if code == clearcode and blocktype!= tables:

            code = -1
            prev = -1


        # End Of Information Codes
        if code == endofinput and blocktype!= tables:
            break

        # Decode from current code
        if code > -1:

            # Check whether there is enough space left in
            # buffer for this code
            if len(output) >= disptable[prev][code]:
                raise ValueError("Decode error.")

            # Copy next segment from decode table
            output += output[:disptable[prev][code]-len(output)]


        # New code
        while True:

            if blocktype == tables:
                # Get next bitslice from input stream
                if bitnum>0:

                    val = ord(rng.read(1))
                    mask = 1<<(7-(bitnum))
                  
                    while mask!=0x100:

                        # Read more bitslices when we run out
                        if ((val&mask)==0) and (bitnum<8):

                            # More data available?
                            try:

                                # Read one more byte
                                val |= ord(rng.read(1)) <<\
                                       (bitnum+1)

                            except TypeError:

                                # No => zero padding
                                pass


                        mask >>= 1
                        bitnum-=1


                else:

                    # We are done reading already
                    val = ord(rng.read(1))
                    bitnum = 7


                if val==eoi:
                    continue

                # Determine what type of code we're dealing with
                if val <= tablesize:

                    # Decode Table Entry
                    code = val

                elif val <= nextsubblock:

                    # Previous entry's entry
                    code = (disptable[prev][val-tablesize])

                elif val <= nextsubblock+(pow(2,
                                                   tablesizebits)):

                    # Skip past extra subblocks
                    code = (disptable[((prev+(
                                pow(2,tablesizebits)))
                                      %(pow(2,tablesizebits)+1))[
                                              val-nextsubblock]]
                    [val-nextsubblock-tablesize])
                    
                    
                else:

                    print("Error")
                    exit()
                
            else:
            
                # Get next 8 bits as code
                code = ord(rng.read(1))

            # If this code is valid...
            if (((-clearcode<=code<=endofinput) or
                 (0<=code<=tablesize)) and 
                 code > prev):

                break

            else:

                code = -1
        

    return bytes(output)
