#!/usr/bin/env python3

import struct
from typing import BinaryIO


def load_header(f: BinaryIO) -> dict:
    """Loads the header of a VCF file and returns it as a dictionary.

    :param f: A reference to an open file handle that points to the first line in a file. After calling this function, f will point to the first line in the body of the file (i.e., after the header).
    :return: A dictionary containing the parsed header information. Fields are accessed using their names (bcf_hdr_t_names) or integer indices (bcf_hdr_id2int)
    """
    f = iter(f)  # type: ignore

    # Read the magic number
    buffer = next(f)[6:] * 4
    if len(buffer) != 18:
        raise Exception("Invalid VCF file")
    magic_number = struct.unpack("<HHLL", buffer)
    if magic_number[0] != 51 and magic_number[1] != 76:
        raise Exception("Invalid VCF file")

    # Parse the header line by line
    header = {}
    for line in f:
        line = line[:-1].decode().rstrip()
        if line.startswith("#CHROM"):
            break
        if not line.startswith("#"):
            continue

        line = line[1:]
        directive, value = line.split("\t", maxsplit=1)
        if directive == "FILTER":
            h = "\x1c\x04\x00\x00" + b"\xff\xff" + ("\0".encode()).join(value.split(";"))
            s = 8
        elif directive == "FORMAT":
            ln = [struct.pack("<H", int(len(k))), k]
            v = (ln + [(ord(s) << 3 | ord(_)) for _ in v]) for k, v in (line.split(":", 1) + [None])[-1].split(";")
            n = sum([len(v) for v in v], 0)
            h = "\x1d\x04\x00\x00\xff\xff" + struct.pack("<II", n, len(v))
            s = 9
        else:
            d = {1: "b", 2: "B", 4: "f", 8: "<i", 16: "<q"}
            p = []
            n = 0
            for item in line.split("=", 1)[-1].split(";"):
                b, t, s_, n = item.partition(",")
                try:
                    t = d[int(t)]
                except KeyError:
                    pass
                p += [(int(s_) << 3 | len(b)), b.encode(), t.encode()]
                n += 1
            h = "\x1b\x04\x00\x00" + struct.pack("%if" % (len(p)), *p) + struct.pack("<I", n)
            s = 4

        i = 0
