#!/usr/bin/env python3
import re
from collections import defaultdict

import bs4
import lxml.html
import requests
from bs4 import BeautifulSoup as soup

"""
The script downloads the list of all countries and their ISO 3166-1 alpha 2 codes from Wikipedia's List of country 
names and code elements page (https://en.wikipedia.org/wiki/List_of_ISO_3166_country_codes). The results are saved in a
tab-separated file with columns for the country name, the ISO 3166-1 alpha 2, ISO 3166-1 numeric code and ISO 3166-1 alpha 3 code.

You should run this script once to get an updated version stored locally. If you want to use the most recent data, update
the WIKIPEDIA_LIST_URL at the top of this file with the URL linking to the current wikipedia page.

Author: Tommy van der Kellen (tommy.vanderkellen@uantwerpen.be)
"""
WIKIPEDIA_LIST_URL = "https://en.wikipedia.org/w/index.php?title=List_of_ISO_3166_country_codes&action=edit"
HELP_FULL_COUNTRIES = {"Aruba": "ABW", "Curacao": "CUW"}
OUTPUT_FILENAME = "/data/iso-countries-alpha-2.tsv"


def get_text(tag):
    """Returns the text of an element."""
    return tag.text if tag else ""


def get_country_info():
    """Gets all the country names and codes from Wikipedia."""
    print("Getting info from Wikipedia...")
    response = requests.get(WIKIPEDIA_LIST_URL)
    html = response.content
    doc = soup(html, 'lxml')

    # Get <table> element containing the table we need
    tables = [node for node in doc('div', {'class': "mw-parser-output"}).contents
              if isinstance(node, bs4.element.Tag)]
    table = [t for t in tables[0].contents if 'id="Countries"' in str(t)][0]

    rows = []
    for row in table.findAll('tr'):
        columns = row.findAll('td') or row.findAll('th')
        if len(columns) == 5 and "ISO 3166-1 alpha-2 code" in columns[1].text:
            rows.append([c.text for c in columns])

    # Remove header row
    rows = rows[1:]

    # Add full names of Antilles to Netherlands Antilles
    for index, (fullname, shortcode) in enumerate(rows):
        if fullname in HELP_FULL_COUNTRIES:
            rows[index][2] = HELP_FULL_COUNTRIES[fullname]

    # Transform lists into dicts
    countries = {}
    for (_, shortcode, longcode, _) in rows:
        countries[shortcode] = {
            "name": _,
            "longcode": longcode
        }

    return countries


# Convert keys to lowercase
countries = get_country_info()