#!/usr/bin/env python3

from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and `.`,
# `o`, `0`, and space as the only characters.
def generate(nframes: int, out: BinaryIO):
    from math import log
    from gif_core import write_header, write_trailer
    from gif_core import write_app_ext, write_loop_count
    from gif_core import write_image
    from gif_core import write_graphic_control

    # Set the width and height of the display area to be 1 pixel.
    # This makes it very easy to fill the image with characters.
    size = 1

    fps = 24
    # The 10th frame is also the end of the animation.
    delay = round(log(10 / (fps * nframes)) / log(2)) + 1
    print('Delay is ' + str(delay))

    transparent_index = 0
    write_header(size, size, out)
    for i in range(nframes + 1):
        if i == 0 or i % 10 == 0:
            # To make sure every frame has at least one byte, we use
            # a repeating pattern. If we didn't, then this wouldn't work.
            raw = b'\x8c\x59\x76\xc1\xf1\xe7\xa5\x7a\xd5\x6b\x3e\xa9\xbf\x94\x9c'
        else:
            # At each iteration, we double the size of the
            # repeating pattern by adding four copies of itself.
            raw += raw[:4]

        # We can use a color table that contains only white and black.
        table = bytes(
            [0xff, 0xff, 0xff,  # White
             0,   0,   0   ]     # Black
        )

        # First we need to write an application extension block.
        # This tells other programs how many frames there are.
        write_app_ext(raw, out)

        # Each frame should have its own graphic control extension.
        # This tells viewers what character to use as the transparency index.
        write_graphic_control(
            transparent=True,
            ndisposal=0,
            transparent_index=transparent_index,
            delay=delay,
            out=out
        )

        # Then, we write the actual image data.
        # All of the pixels are just the same color: the background color.
        write_image(table, raw[0], (0, 0), size, size, out)