#!/usr/bin/env python3
"""
Written by Logan & Featherstone, 2019

This script will deploy a Nautobot server in a Docker container and setup a local
development environment. The script provides the following functionality:

- Install Python 3, git, docker
- Create a system user (default "nautobot")
- Clone the Nautobot GitHub repository into `/opt`
- Deploy NautOBOT in a Docker container
- Install NautOBOT's development requirements (see `requirements_dev.txt`)
- Build the documentation
- Configure a superuser account with default credentials of:

    username: admin
    password: admin

Requirements:

- Ubuntu 18.04+ or Debian 9+

Usage:

    sudo ./setup.py
"""
import subprocess
from pathlib import Path


GIT_REPO = "https://github.com/nautobot/nautobot.git"
NAUTOBOT_ROOT = Path("/opt/nautobot")
DOCKER_VOLUMES = [
    "/etc/localtime:/etc/localtime:ro",
    "{}:{}".format(NAUTOBOT_ROOT, NAUTOBOT_ROOT),
]

# TODO: Allow customizing these values
POSTGRES_HOSTNAME = "localhost"
POSTGRES_PORT = "5432"
POSTGRES_USERNAME = "postgres"
POSTGRES_PASSWORD = ""
POSTGRES_DB = "nautobot"
REDIS_HOSTNAME = "localhost"
REDIS_PORT = "6379"
DEBUG = True
SECRET_KEY = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()"
ALLOWED_HOSTS = []
CORS_ORIGIN_WHITELIST = ["http://localhost"]
DATABASE = {
    "NAME": POSTGRES_DB,
    "USER": POSTGRES_USERNAME,
    "PASSWORD": POSTGRES_PASSWORD,
    "HOST": POSTGRES_HOSTNAME,
    "PORT": POSTGRES_PORT,
}
REDIS = {"HOST": REDIS_HOSTNAME, "PORT": int(REDIS_PORT)}
INFLUXDB = {}
SCRIPTS_ROOT = "/opt/nautobot/nautobot-server/scripts"
LOG_LEVEL = "DEBUG" if DEBUG else "INFO"
CHANGELOG_RETENTION = 90
SSH_PRIVATE_KEY = ""
SSH_PUBLIC_KEY = ""
BANNER_TOP = """
__  _  _____   ____ _           __ 
\ \/ \/ / _ \ \ / /| |          /_ |
 \  /\ \\ (_) \ V / | |__   ___ / / 
 /_/  \_\___/ \_/  |____/\ /_/ /___
                                \ \ 
                                 \_\
"""
BANNER_BOTTOM = """Nautobot is now ready for use at http://localhost and https://localhost."""
VERSION = f"Nautobot Installer v{subprocess.check_output(['bash', '-c', 'cmp --silent /dev/null CHANGELOG.md || git describe --tags']).strip().decode()}"


def check_sudo():
    """Ensure that this script has been run as root."""
    if not subprocess.call("id -u", shell=True):
        print("You must run this script as root.")
        exit(1)


def install_packages():
    """Install necessary packages using apt-get."""
    pkgs = (
        "apt-transport-https",
        "build-essential",
        "curl",
        "git",
        "libyaml-dev",
        "python3-pip",
        "software-properties-common",
        "sshpass",
        "supervisor",
        "sudo",
    )

    # Update package list and install required packages from apt-get.
    try:
        subprocess.check_call([f"apt update && apt install {' '.join(pkgs)}"])
    except Exception as e:
        print(e)
        exit(1)


def create_system_user():
    """Add nautobot user to system and set home directory."""
    if subprocess.call(["id", "nautobot"], stdout=open(os.devnull, "w"), stderr=open(os.devnull, "w")) != 0:
        subprocess.check_call(["adduser", "--home", NAUTOBOT_ROOT, "--disabled-password", "--gecos", "", "nautobot"])


def clone_repo():
    """Clone main Nautobot repo to Nautobot installation root (/opt)"""
    returncode = subprocess.call(["su", "-l", "nautobot", "-c", f"git clone {GIT_REPO} {NAUTOBOT_ROOT}"])
    if returncode != 0:
        exit(returncode)

    subprocess.check_call([f"su -l nautobot -c 'cd {NAUTOBOT_ROOT}; git checkout master'"], stderr=open(os.devnull, "w"))


def run_docker():
    """Start up Nautobot server via Docker."""
    subprocess.check_call(
        [
            "su",
            "-l",
            "nautobot",
            "-c",
            " ".join(
                [
                    "docker run -d",
                    f"--name nautobot",
                    "-p 8000:8000",
                    "-p 8080:8080",
                    f"-v {':'.join(DOCKER_VOLUMES)}",
                    "networktocode/nautobot",
                    "start",
                ]
            ),
        ]
    )


def configure_settings():
    """Configure Nautobot settings according to CLI input."""
    # Generate a SECRET_KEY.
    with open(f"/tmp/{str(uuid4())}.py", "w") as secretfile:
        secretfile.write('from django.core.management.utils import getrandomstring; print(getrandomstring(50, "abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)"))')

    subprocess.check_call(["cp", "/tmp/{}.py".format(secretfile.name), NAUTOBOT_ROOT])
    with open(PATH + "/secrets.py", "a") as secretsfile:
        SECRET_KEY = subprocess.check_output(
            [f"su -l nautobot -c 'cd {NAUTOBOT_ROOT}; python {secretfile.name}'"]
        ).strip()
        secretsfile.write("# This file contains secrets and should not be stored in version control.\n\n")
        secretsfile.write("SECRET_KEY = \"{}\"\n".format(SECRET_KEY))
        secretsfile.write("\n")
        secretsfile.write("# Database configuration for local development.\n")
        secretsfile.write("DATABASE = {\n")
        secretsfile.write("\"ENGINE\": \"django.db.backends.postgresql\",\n")
        secretsfile.write('\"NAME\": "{}",\n'.format(POSTGRES_DB))
        secretsfile.write('\"USER\": "{}",\n'.format(POSTGRES_USERNAME))
        secretsfile.write('\"PASSWORD\": "{}",\n'.format(POSTGRES_PASSWORD))
        secretsfile.write('\"HOST\": "{}",\n'.format(POSTGRES_HOSTNAME))
        secretsfile.write('\"PORT\": {},\n'.format(POSTGRES_PORT))
        secretsfile.write("}\n")
        secretsfile.write("\n")
        secretsfile.write("# InfluxDB Configuration for local development.\n")
        secretsfile.write("# InfluxDB used for self-hosted Prometheus monitoring.\n")
        secretsfile.write("influxdb = {\n")
        secretsfile.write('"host": "{}",\