#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header (8 bytes): magic 0x47 0x49 0x46 ("GIF"), version 1.0
    header = bytearray([
        0x47, 0x49, 0x46, 0x38,  0x39, 0x61,  0x2c, 0x00
    ])

    # Logical screen descriptor (7 bytes:  width 10, height 10, packed field)
    lsd = struct.pack(">HHBBB", 10, 10,
                        0b11000010,    # Global color table present
                        0,             # Background index
                        0               # Pixel aspect ratio
    )

    out.write(header + lsd)


    palette = []     # 768-color RGB palette in RGBA order; used for transparency
    for _ in range(0, 256 * 4, 4):
        r = ord(rng.read(1)) % 256
        g = ord(rng.read(1)) % 256
        b = ord(rng.read(1)) % 256

        palette += [r] + [g] + [b] + [0xFF]    # last byte is alpha channel

    global_palette = bytearray()          # global color table
    for i in range(0, len(palette), 4):
        rgb = palette[i : i+3]              # skip A component
        global_palette += rgb                # add R, G and B components to global table

    # Output global palette to output stream
    out.write(global_palette)

    delay = 20      # milliseconds per frame
    transparent = 0 # pixel index of transparent background

    # Graphics control extension block (8 bytes). Used both for frames and the trailer
    graphics_control = struct.pack(">BBBBH",
                                     0x21,         # Extension introducer (always!)
                                     ord('f'),     # Graphic control label ("f")
                                     ord('a'),     # Version ("a")
                                     0x04,         # Flags (contains trans index bit)
                                     delay,        # Delay between updates
                                     transparent
    )

    # Frames
    frames = 0
    while True:
        # Start of image
        out.write(struct.pack(">BB",
                                      0x2C,           # Image separator
                                      0xF9            # Size of following data
                                    ))

        imgdata = bytearray()       # binary representation of current frame
        lzwmincodesize = 2**ord(rng.read(1))   # LZW minimum code size

        subblock = bytearray()      # next compressed chunk of pixels
        # Generate each chunk of pixels until we reach an "end" marker
        while subblock[-2:]!= b"\x00\x3B":
            # Clearing frame? If so, fill this chunk with solid black (#000000)
            if not frames or bool(int(rng.read(1), base=2)):
                subblock.extend(bytearray([0]) * 12)

            else:

                numbytes = int(rng.read(1), base=2) + 1
                color = int(rng.read(numbytes), base=2)
                subblock.append(lzwmincodesize - 1)

                # Repeat previous color
                if color == 0:
                    pass

                # Reuse previously defined LZW line code
                elif color <= len(lzwcodes) - 1:

                    def getbits(n):
                        return lambda c: ((c >> n) & 0x1) ^ ((c << (8 - n)) & 0x80)

                    bits = getbits(len(lzwcodes[color]))
                    copybits = getbits(len(lzwcodes[color]) - 1)

                    mask = [copybits(c) for c in lzwcodes[color]]
                    for i in range(0, len(mask)):
                        v = mask[i]

                        for j in reversed(range(0, 8)):
                            if subblock[-1] & 0x80 > 0:
                                subblock[-1] &= ~0x80

                                if mask[(i - 1) % len(mask)] == 1:
                                    subblock[-1] ^= 0x40

                            mask[(i - 1) % len(mask)] = v
                            v = copybits(v)

                #...or define new one
                else:
                    lzwcodes.append([subblock[-1]])
                    subblock.append(color - len(lzwcodes) + 1)


            if len(subblock) > 255:
                imgdata += bytes([(len(subblock)-1) % 256]) + subblock[:-1]
                subblock = subblock[-1:]


        frames += 1

        # End of image
        out.write(imgdata + bytes([0]) + graphics_control)



if