#!/usr/bin/env python3

from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate(out: BinaryIO):
    # Header contains 6 bytes, of which first 3 are fixed
    header = [0x47, 0x49, 0x46]

    # The rest is defined as follows (Byte offset : Description, size in bits):
    # Byte  0 : Signature (GIF)                          3*8 = 24 bit
    # Bytes 1-2: Version (87a or 89a)                    8 * 8 = 64 bit
    # Byte  3 : Image Width                             1 * 16 = 16 bit
    # Byte  5 : Image Height                            1 * 16 = 16 bit
    # Byte  7 : Global Color Table Flag                 1 *  8 =  8 bit
    # Byte  8 : Background Color Index                  1 *  8 =  8 bit
    # Byte  9 : Pixel Aspect Ratio                      1 *  8 =  8 bit
    # Bytes 10-12: GCT Size                              8 * {GCT_size} bit
    #         GCT_size = 7 + C*(3*8 + 1*8)                 = 7 + C*(24+8)=256
    #         So GCT_size might be either 0, 7, 13,..., up to 256. Note that C
    #         can also be zero, if there's no color table at all (e.g., for a
    #         monochrome image).
    # Next comes the LZW minimum code size (LZW_min), 1 byte value.
    # Then comes the images' data. There's an optional block after this one --
    # Application Extension Block, except when there's only global color table and
    # it does not have transparency. In such case there's just plain Graphics
    # Control Block with transparent flag set.
    # Finally, we're done with the image and come to the end of file.
    # All these fields together take TBD bytes.

    width = rng() % (1 << 16)               # 2 bytes
    height = rng() % (1 << 16)              # 2 bytes
    background = rng() % (1 << 8)          # 1 byte
    gctFlag = rng() & 1 == 1                # 1 byte
    lzwMinCodeSize = rng() % (1 << 8)       # 1 byte
    gctSize = min((width * height // 64)+1, rng() % ((width * height) >> 8))    # 1 byte

    ratioNumerator = max([width, height])
    ratioDenominator = min([width, height])
    aspectRatio = round(ratioNumerator / ratioDenominator) - 1     # 1 byte

    out.write(bytes(header))
    out.write(b'\x00\x00')      # version
    out.write(width.to_bytes(2, 'little'))
    out.write(height.to_bytes(2, 'little'))
    out.write(gctFlag.to_bytes(1, 'little'))
    out.write(background.to_bytes(1, 'little'))
    out.write(aspectRatio.to_bytes(1, 'little'))
    out.write(gctSize.to_bytes(1, 'little'))

    return 11                            # length of the header
