#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    width = 640
    height = 480
    
    # Write header
    out.write(header)

    # Write logical screen descriptor (always 7 bytes long)
    packed_dims = struct.pack('>HHB', width, height, 0x72)
    out.write(packed_dims)

    palette = [
        0x00, 0x00, 0x00,   # black
        0xff, 0xff, 0xff    # white
    ]

    # Write global color table
    for c in range(len(palette)):
        out.write(bytes([palette[c]]))


    frame_counter = 15
    while True:

        # Write graphic control block
        packed_delays = struct.pack('<HBB', 5 * frame_counter + 1, 0x70, 0)
        out.write(packed_delays)
        
        # Write local color table
        if frame_counter < len(palette):
            out.write(bytes([palette[frame_counter]]))

        # Write image descriptor
        packed_dims = struct.pack('<HHB', \
                width, height, 0,     # left/top position
                0,                     # no global palette
                0                      # no LZW compression
        )
        out.write(packed_dims)
        
        
        cur_row = []
        stream_pos = out.tell()

        # Write uncompressed raster data
        for i in range(0, width*(height-1), width):
            r = rng.read(width // 8)

            # Pad to even length so that we can use byte stuffing
            l = len(r) 
            if not l % 2 == 0:
                r += b'\x00'
            
            cur_row = list(reversed(cur_row))
            row = r+b''.join([b'\x00'+x for x in cur_row])
            out.write(row)

            # Byte stuff every other scanline
            if ((i//width)%2!= 0): # only do this on odd lines
                num_stuffed = 0

                for k in range(l):
                    bitval = ord(r[k : k+1])
                    msb = bool((bitval & 0xc0)!= 0)
                    lsb = bool(bitval & 0x01)

                    if msb and lsb:
                        num_stuffed += 1
                        continue
                    
                    elif msb or lsb:

                        if num_stuffed > 0:
                            out.seek(-num_stuffed - 1, 1)
                            out.write(struct.pack("<B", 0x00))
                            out.write(b'\x00')

                            stream_pos -= num_stuffed + 1
                        break
                    
                    else:
                        pass
                
                for j in range(l-num_stuffed):
                    out.write(r[j : j+1])

            cur_row = row[:width]


        # Go back to write terminating code at end of last line
        out.seek(stream_pos)
        out.write(b'\x00\x00') 
        out.flush()
            
        if frame_counter <= 0:
            return
        
        frame_counter -= 1