#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and width/height.
def generate(out: BinaryIO, rng: BinaryIO, width: int, height: int):
    # Write the magic identifiers.
    out.write("GIF89a".encode())

    # Determine the colors used in this image.
    palette = [0] * 768
    for i in range(256):
        r = next(rng) & 0xFF;
        g = next(rng) & 0xFF;
        b = next(rng) & 0xFF;

        palette[i*3 + 0] = r << 2;
        palette[i*3 + 1] = g << 2;
        palette[i*3 + 2] = b << 2;
    
    # Write the global header (logical screen descriptor).
    out.write(struct.pack("<HHBBB", width, height, 8, 7, 0))

    # Write the color table.
    out.write(bytes(palette))

    # Compress all frames with no delay between them.
    num_frames = min(width // 4, height // 4);
    for _ in range(num_frames):
        write_frame(out, rng, width, height)


# Generates one frame of the given dimensions at the end of.
def write_frame(out: BinaryIO, rng: BinaryIO, width: int, height: int):
    # Set the minimum LZW code size to 8 bits.
    out.write(b"\x21\xF9\x04")

    # Get the pixel data from /dev/urandom.
    compressed_pixels = compress(get_pixels(rng), rng)

    # Write the extension block which tells how large the uncompressed 
    # pixels will be when decompressed.
    out.write(struct.pack("<BHBB", 0x21, 0xFE, len(compressed_pixels) >> 8, len(compressed_pixels) % 256));

    # Write the uncompressed pixels themselves, but padded with an extra 0 bit since the 
    # last byte has a special meaning.
    out.write((compressed_pixels + bytes([0]))[:len(compressed_pixels)]);

    # Finally, write the trailer value of 0.
    out.write(b"\x3B");


# Returns a bytes object representing the binary encoding of the PNG compression scheme.
def compress(pixels: bytes, rng: BinaryIO) -> bytes:
    # Keep track of what the longest match was so far.
    prev = None

    out = []
    while True:
        curr = pixels[next(rng) % len(pixels)] if pixels else -1
        
        # If we're not in a run, start looking for a new run.
        if prev is None:
            matches = []

            # We'll look ahead by eight characters on each side until finding 
            # something that matches.
            for offset in range(-8, 8+1):
                m = pixels[(offset + next(rng)) % len(pixels)]

                # Only keep it as a potential match if it's different than 
                # the current character itself, and only if at least half 
                # of its bits are shared with the previously matched string.
                if m!= curr and ((prev == m or abs(ord(curr)-ord(m)) <= 32) and (abs(ord(curr)-ord(m)) >= 4)):
                    matches.append(m)
            
            # No matches found? Just break out of here.
            if len(matches) == 0:
                break
            
            # Otherwise choose based on the frequency of letters starting with these.
            freqs = {}
            for c in set(matches):
                key = chr(c)[-1]
                
                if key not in freqs:
                    freqs[key] = 0
                freqs[key] += matches.count(c)
            
            # Sort by frequency and pick whichever letter starts a sequence most often.
            best = sorted([(freqs[k], k) for k in freqs])[::-1][0][1]
            prev = ord(best)
                
            # Add the initial prefix value to the output stream.
            out.extend([0xE0 | ord(best)])
        
        # If we've already generated some bytes before then add any of those 
        # repetitions back to our list first before checking whether there's more input.
        rep_count = 0
        if len(out) > 0:
            rep_count = out[-1] & 0xF
    
        for i in range(rep_count + 1, 0xFFFF):
            
            # Check if the same byte occurs again n times in the remaining input.
            def check(n, text):
                return text[i:] == ([text[i]] * n)
            
            # Try three different runs of four bytes to see if they result in equal output.
            if check(4, pixels) and check(4, pixels[(i + next(rng)) % len(pixels):]):
                out.pop()
                out.extend([0xF0 | (((i - rep_count) & 0xF00) >> 8), (i - rep_count) & 0xFF])
                break
        elif i == rep_count + 1:
            # If the maximum length hasn't been reached yet but no match can be made, just stop.
            break
        
    return bytes(out)
                    
        
    
# Reads'size' bytes from /dev/urandom and returns
# them as an integer.
def get_random(rng: BinaryIO, size: int) -> int:
    return sum([next(rng) << (i * 8) for i in reversed(range(size))])

    
# Returns a list of pixels in a row-major format. The total number of pixels 
# returned will always have a power of two in both dimension like 16, 
 # 32.. etc.
def get_pixels(rng: BinaryIO) -> str:
    # Choose a random offset for upper left corner.
    x = y = get_random(rng, 11) & ~1

    # Generate the pixels line by line. For now only support RGB mode.
    pixels = ""
    while len(pixels) < 1024:
        # Keep generating new lines until we have enough.
        while len(pixels) < 64*(x//8)*(y//8) + 1024:
            # Determine the number of consecutive pixels that will have the same colour.
            run = max(int(get_random(rng, 2)), 2)

            # Determine the various values of R, G and B for the entire run.
            r = get_random(rng, run + 1)# % 256
            g = get_random(rng, run + 1)# % 256
            b = get_random(rng, run + 1)# % 256

            # Store the pixels.
            pixels += "".join([chr(r), chr(g), chr(b)])
            
        # Fill up the rest of the current line.           
        run = (min(((1024 - len(pixels))/3)/((x//8)+1), 63))
        if run > 0:
            pixels += "".join([chr(get_random(rng, 1)) for _ in range(run)])
        
        # Skip downwards by one line.
        y += 1
 
    return pixels[:-1];



if