#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the provided parameters.
def write_random(
    out: BinaryIO, 
    width = 400, height = 250, num_frames = 16, delay = 10, global_color_table = True):

    # Header
    out.write('GIF89a'.encode())
    out.write((width).to_bytes(2, byteorder='little'))
    out.write((height).to_bytes(2, byteorder='little'))
    
    if global_color_table:
        out.write('\x87\x00')       # Global color table with 256 entries

    else:
        out.write('\xf7\x00')       # No global color table

    # Loop control extension (delays are in centiseconds)
    out.write('!\xf9\x04')           # Application extension block
    out.write('NETSCAPE2.0\x03\x01') # NETSCAPE2.0 identifier and sub-block length
    out.write((delay*num_frames + 1).to_bytes(2, byteorder='little'))   # Frame count
    out.write(b'\x00')              # Byte to indicate looping forever
    out.write(b';')                 # End of block

    # Graphics control extension
    out.write(','.encode())         # Application extension block
    out.write('\xf9\x04')           # Block size
    out.write('\x04')               # Flags - transparency enabled, user input disabled
    out.write((delay).to_bytes(2, byteorder='little'))        # Delay between frames in hundredths of a second
    out.write((int(input("Enter background color index (default=255):\n") or "255")).to_bytes(1, byteorder='little'))    # Transparent color index
    out.write('\x00')               # End of block
    for i in range(num_frames):

        frame_size = width * height // 8  
        bits_per_pixel = len([n for n in range(frame_size) if ((i & (1<<n)) > 0)]) 
        print(bits_per_pixel)
        
        # Image descriptor
        out.write(','.encode())         # Application extension block
        out.write((width//256).to_bytes(1, byteorder='big'))     # Left position
        out.write((width%256).to_bytes(1, byteorder='big'))      # Left position
        out.write((height//256).to_bytes(1, byteorder='big'))    # Top position
        out.write((height%256).to_bytes(1, byteorder='big'))     # Top position
        out.write((bits_per_pixel).to_bytes(1, byteorder='big')) # Width x Height
        out.write('\x00')               # Packed field; local color table flag is set to 0
        # Image data
        num_bytes = int((((width+1)*bits_per_pixel + 7)//8))   # Number of bytes needed to store image
        row = []
        for j in reversed(range(height)):
            pixel_array = [n for n in range(frame_size) if (((j << bits_per_pixel) | i) & (1<<n))] 
            row += ([0] * (num_bytes - len(row))) + pixel_array 
        assert len(row) == num_bytes * height
        for k in range(len(row)//num_bytes):
            b = sum([((p+k)%2)**i for p in row[num_bytes*k : num_bytes*(k+1)]]) % 2**bits_per_pixel
            out.write(((b << (8 - bits_per_pixel)) | (ord('\x00') >> bits_per_pixel)).to_bytes(1, byteorder='big'))

