#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)
#     print('lsd', bytes(lsd).hex())
    width  = int.from_bytes(lsd[0:2], 'little') + 1
    height = int.from_bytes(lsd[2:4], 'little') + 1
#     print('width', width, 'height', height)
    packed_byte = lsd[4]
    flags = packed_byte & 0x7

    # Add LSD field to output
    out.write(lsd[:6])

    if (flags == 0):
        out.write(b'\x00\x3B')
    else:
        ext = rng.read(1)[0]

        while True:
            # End of extensions
            if (ext == 0x00):
                break

            # Plain text extension
            elif (ext == 0xF9 or ext == 0xFE):
                size = rng.read(1)[0]

                if ((size!= 12) and
                    (size!= 12+int((packed_byte >> 3)&1))
                   ):
                    continue

                subblock = bytearray()
                
                for i in range(0, 1+int(((packed_byte>>3)&1)*8), 1):
                    subblock += rng.read(1)

                out.write(struct.pack('<cBBBHB',
                                    ext.to_bytes(1, 'big'),
                                    1+len([i for i in subblock]),
                                    *subblock
                                   ))

            # Graphic control extension
            elif (ext == 0xF9):
                size = rng.read(1)[0]

                if (size!= 4):
                    continue

                out.write(ext.to_bytes(1, 'big'))
                out.write(struct.pack('<BBBB', *rng.read(4)))
                out.write(b'\x00')

            # Application extension
            elif (ext == 0xFF):
                appid = rng.read(1)[0]

                if (appid not in [0xF9]):
                    continue

                size = rng.read(1)[0]

                if (size!= 4):
                    continue

                out.write(ext.to_bytes(1, 'big'))
                out.write(appid.to_bytes(1, 'big'))
                out.write(struct.pack('<I', rng.read(4)))
                out.write(b'\x00\x00')

            else:
                size = rng.read(1)[0]
                out.write(ext.to_bytes(1, 'big'))
                out.write(size.to_bytes(1, 'big'))
                out.write(rng.read(size))

            # Read next extension block
            try:
                ext = rng.read(1)[0]
            except IndexError:
                ext = None


    # Set up offset table
    data_offsets = []
    
    for y in range(height-1,-1,-1):
        line = bytearray()
        
        for x in range(width*3):
            line += rng.read(1)
            
        data_offsets.append(out.tell())
        out.write(line)

    # Write colour table
    palette = bytearray()

    for _ in range(256*3):
        palette += rng.read(1)

    out.write(palette)

    # Write code section
    num_codes = 2**bits
    clear   = 1 << bits - 1
    end     = 1 << bits
    codes   = [rng.read(1) for i in range(num_codes)]
    values  = []
        
    for y in range(height-1,-1,-1):
        line = ''

        for x in range(width):
            packed = pack_color(rng.read(3))
            
            if packed == clear:
                line += str(clear)
                values.append('')
                
            elif len(values[-1]) < skip_minimum:
                line += str(clear)
                values.append('')
                values[-2] += codes[int.from_bytes(packed,'big')]
            
            elif any(i in values[-1] for i in backref_codes):
                index = int.from_bytes(packed, 'big')
                marker = min(backref_codes, key=lambda k: values[-1].find(k))

                # If after the marker we find another occurance of this marker
                # then it's an escaped character
                if values[-1][values[-1].find(marker)+len(marker)] == marker:
                    # This is an escape sequence
                    line += str(clear)
                    values.append('')

                    # Append escaped char (as string) to previous value
                    values[-2] += codes[index]
                    
                else:
                    # This is a normal reference
                    line += str(clear)
                    values.append('')
                    
                    ref_index = values[-1].rfind(marker)+len(marker)
                    values[-2] += values[-1][ref_index:]
                    values[-1] = values[-1][:ref_index] + values[-1][marker:]
                    
            else:
                # If no matches are found just append to last value
                index = int.from_bytes(packed, 'big')
                values[-1] += codes[index]

                line += ''.join(['{:>3}'.format(str(v))
                                for v in values[-1]])
                
        encoded_data = encode_string(line, codes)
        out.write(encoded_data)


    # Image descriptor
    out.write(struct.pack('<HHHBB',
                         width      % 0x10000,
                         height//256,
                         0,
                         0,
                         0
                        ))

    # Graphics control extension with transparent color flag set
    out.write(b'\x21\xf9\x04\x01\x00\x00\x00\xff\xff')

    # Ending image descriptor
    out.write(b';')

    # Codes dictionary
    out.write(struct.pack('<B',len(codes)))
    out.write(b''.join(codes))

    # Data offsets
    out.write(struct.pack('<B', len(data_offsets)))

    for doffset in data_offsets:
        out.write(doffset.to_bytes(2, 'big'))
        
    

