#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given PRNG and with some fixed parameters, as
# specified below. See https://www.w3.org/Graphics/GIF/spec-gif89a.txt for details on these params.
def main(rng: BinaryIO, out: BinaryIO):

    # Header section fields (see spec).
    width  =  256                   # "logical screen width" => actual size of image in pixels
    height =    1                    # "logical screen height" => actual size of image in pixels
    fp     = b'\x00'                 # background color index => transparent black
    fmt    = b'\xF7\xFF\xFF'         # pixel aspect ratio; actual size is width * height / 256 => 256 bytes per row

    headerlen = 13                     # length of the header section without delay durations
    nframes   =   4                  # number of frames
    duration  = [ 1 ] * nframes       # frame delays in milliseconds

    # Write all relevant values to a single bytearray first so we can compute offsets later.
    header_buf = bytearray()
    header_buf += width.to_bytes(2, byteorder='little')        # logical screen width
    header_buf += height.to_bytes(2, byteorder='little')       # logical screen height
    header_buf += fp.to_bytes(1, byteorder='little')             # background color index
    header_buf += fmt.to_bytes(4, byteorder='little')    # pixel aspect ratio
    headerpad = 1 << (sizeof(int)*8 - len(header)) % sizeof(int)          # align header len to int boundary
    header_buf += headerpad.to_bytes(2, byteorder='little', signed=False)      # padding between header and subsequent blocks
    for d in duration:
        header_buf += d.to_bytes(2, byteorder='little')           # frame delays

    # Compute offset from end of header block to beginning of data block (after frame 1),
    # which is just after the last written frame's duration value.
    frame1offset = header_buf[2*nframes] + header_buf[2*nframes+1]*256 + headerpad + 2*(nframes-1)

    # Pad header up to a multiple of four bytes long.
    headerpad = 4 - len(header_buf) % 4
    if headerpad == 4:
        headerpad = 0
    header_buf += b'\0'*headerpad

    # Write entire header in one go.
    assert(len(header_buf)%4 == 0)
    header = binascii.hexlify(header_buf)
    datalen = rng.readinto(bytearray())
    print("Header hexdump:")
    dumpbin(header)

    # Write frame data repeatedly.
    for i in range(nframes):
        # First write the frame header that consists of the above-computed frame offset and frame duration.
        startslice = max(frame1offset, len(header)+i*(2+datalen))
        headerslice = header + startslice.to_bytes(2, byteorder='little') + duration[i].to_bytes(2, byteorder='little')

        # Align header slice at word boundary.
        headerslice_pad = 3 -(len(headerslice)-len(header))%4
        if headerslice_pad == 3:
            headerslice_pad = 0
        headerslice += b'\0' * headerslice_pad

        data = rng.read(datalen)
    out.write(header + data)

