#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Write header (6 bytes for this specification)
    out.write('GIF89a'.encode("ascii"))

    # Write logical screen descriptor information - width 720, height 480, global color table of size 16 and color resolution of 8 bits per pixel
    out.write(struct.pack("<HHBBB", 720, 480, 16, 8))

    # Write color table (16 entries consisting of R G B values each byte wide)
    for i in range(16):
        r = next(rng) % 256
        g = next(rng) % 256
        b = next(rng) % 256

        out.write(bytes([r,g,b]))

    # Write image data block as a plain black frame
    LZWMinimumCodeSize = 2
    Width = 720
    Height = 480
    ColorTableBitsPerPixel = 8
    GlobalColorTableFlag = True
    SortingFlag = False
    InterlaceFlag = False
    
    LogicalScreenDescriptorBlock = "{0} {1} {2}".format(LZWMinimumCodeSize,
                                                        struct.pack(">B", GlobalColorTableFlag << 7 | SortingFlag << 6 | InterlaceFlag).hex(),
                                                        struct.pack(">H",
                                                                    struct.calcsize("<{0}B".format((Width*Height)/2)),
                                                                    0,
                                                                    0,
                                                                    ColorTableBitsPerPixel)).strip().replace(" ", "").upper()

    ImageDataBlock = "0 {0} 0 {1} {2}".format(LogicalScreenDescriptorBlock,
                                            "{0} {1} ".format(0, Width),
                                            "{0} {1} ".format(0, Height)).strip().replace(" ", "")

    ExtensionBlock = ""

    TrailerBlock = "{0}{1}3B".format(ImageDataBlock,
                                     ExtensionBlock).strip().replace(" ", "")

    DataSubblock = "".join("{0:02x}".format(next(rng) % 256)
                           for _ in range(int(((Width * Height) / 2))))
    SubblocksCount = len(DataSubblock)//256 + 1 if int(((Width * Height) / 2))%256!= 0 else len(DataSubblock)//256
    Size = str(len(DataSubblock)+10+SubblocksCount-1).zfill(4)
    
    out.write(("".join([Size,
                        "{0:02x}".format(LZWMinimumCodeSize),
                        LogicalScreenDescriptorBlock,
                        ImageDataBlock,
                        "{0:#0{1}x}".format(len(TrailerBlock)*3//2,"x"),
                        "{0:#0{1}x}".format(len(TrailerBlock)*3//2,"x"),
                        TrailerBlock])).encode("ascii")
              )

    out.write("02FE".encode())
    out.write(struct.pack("<H", 10))

    AppExtensionID = "".join(["{0:02x}".format(_[0]) for _ in [
        (
            255,
            Identifier
        ),
        ((Identifier >> 8) & 255,
         Identifier & 255)]]).strip().replace(" ", "")

    ApplicationIdentifier = "{0:#0{1}x}".format(
        subblocksSize+12,"x").strip().replace(" ","")

    ApplicationAuthenticationCode = "{0:#0{1}x}".format(
        appDataLength-appsubblocksLength-14,"x").strip().replace(" ","")

    BlockSize = "{0:02x}".format(len(ApplicationIdentifier))

    out.write("F9".encode("ascii"))
    out.write(BlockSize.encode("ascii"))
    out.write(AppExtensionID.encode("ascii"))
    out.write(ApplicationIdentifier.encode("ascii"))
    out.write(ApplicationAuthenticationCode.encode("ascii"))

    Subblocks = []
    for i in range(SubblocksCount):
        Subblocks.append(DataSubblock[i*256:(i+1)*256])

    for sb in Subblocks[:-1]:
        out.write("{0:02x}".format(sb).encode("ascii"))
        out.write("FF".encode("ascii"))

    out.write("{0:02x}".format(Subblocks[-1]).encode("ascii"))

    # End of File
    out.write(";".encode("ascii"))

