#!/usr/bin/env python3

import struct
from typing import BinaryIO

# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    while len(data)%2 == 1: # odd datalen is invalid
        print(f"odd datalen: {len(data)}, retrying")
        rng.seek(-1, 1)
        continue


    controlblocks = (b'\x2C\x00', b'\x21\xF9\x04\x05\x00\x00', b'\x3B')
    transparencyflag = False
    
    for cb in controlblocks:
        if cb not in data:
            break
        
        if cb[0] == 0x21 and cb[-1] == 0x00:
            print("transparency block found")
            transparencyflag = True
            break

        print("skipping controlblock", bytes([cb[0]]))
        i = data.index(cb)+len(cb)
        data = data[:i]+data[data.index(controlblocks[(controlblocks.index(cb)+1)%len(controlblocks)], i):]

    # check that the file begins with `header`
    assert data.startswith(header)
    out.write(data)

    palette = [b'\xFF'*3]*256 # standard palette
    if transparencyflag:
        tindex = min((data.find(b'\x21\xF9\x04\x01'), data.find(b'\x21\xFE\x00')))+7 # asserts that one of them exists
        print("transparency index at",tindex)
        palettenum = data[tindex] 
        palette[palettenum*3:(palettenum+1)*3] = b'\x00'*3 

    # read all frames from the end until a non-frame marker is reached
    prevmarker = None 
    while data[-3:]!= b';;':
        frametypebyte = data[-1 : -4:-1][0] % 8 # last byte contains frame type information
        assert frametypebyte < 5, "unknown frame type identifier"
        frametype = ['graphics', 'extension', 'application extension', 'comment extension', 'plain text'][frametypebyte]
        print("reading", frametype,"frame...")
        
        # determine how many color registers are expected to be present
        # extensions may change this number
        colors = 2**(frametypebyte+1)
        # default image descriptor contains global color table flag
        gct = bool(int.from_bytes(
                data[-7 : -4], byteorder='little'))
        lctsize = 3*(colors**gct)-1
        localcolortablelength = len(data) - lctsize - 16 # length of the entire local color table (+16 for graphic control field + 16 for LZW minimum code size)
        
        # read the data corresponding to an individual frame from the end
        gifframe = data[:-localcolortablelength]
        data = data[:-len(gifframe)] 

        print("Local color table:")
        for i in range(0,lctsize//3,3):
            color = data[i : i+3] 
            palette[struct.unpack('>H', color)[0]] = color
            print(hex(struct.unpack('>H', color)[0]),":",color.decode())
            
        # local color table ends now, so we can remove it from our buffer
        data = data[lctsize:]

        # determine the maximum possible LZW minimum code size based on the max color value
        # and the passed color table
        lzwmincodesize = ((palette.index(max(palette)) & ~1)+1)//2
        lzwmincodesize = (lzwmincodesize<<1)|bool(lzwmincodesize%5)
        print(f"LZW minimum code size is {lzwmincodesize}")

        # skip the next two bytes since they store LZW minimum code size
        gifframe = gifframe[:-2] 

        # make sure the first byte is either \x00 or \x01
        # else, its only padding between frames
        assert gifframe[0] <= 1
        gifframe = gifframe[1:]

        outputstream = []
        while gifframe:
            prefix = gifframe[0]
            if prefix == 0: 
                outputstream += [prefix]
                gifframe = gifframe[1:]
                
            elif prefix <= 0xF:
                outputstream += [prefix]
                gifframe = gifframe[1:]
                
            elif prefix >= 0x10 and prefix <= 0x1F:
                outputstream += inputstream[outputstream[-1]] * (((prefix & 0x7) << 8) | gifframe[1])
                gifframe = gifframe[2:]

            elif prefix == 0xE:
                compressedbytes = gifframe[1:]
                uncompressedstream = []

                while compressedbytes:
                    prefix = compressedbytes[0] 
                    if prefix == 0:
                        uncompressedstream += [prefix]
                        compressedbytes = compressedbytes[1:]
                        
                    elif prefix <= 0xF:
                        uncompressedstream += [prefix]
                        compressedbytes = compressedbytes[1:]

                    elif prefix >= 0x10 and prefix <= 0x1F:
                        uncompressedstream += uncompressedstream[uncompressedstream[-1]] * (((prefix & 0x7)<<8) | compressedbytes[1]) 
                        compressedbytes = compressedbytes[2:]
                    
                    elif prefix == 0xE:
                        raise NotImplementedError()
                    
                    else:
                        raise ValueError(f"invalid prefix {hex(prefix)}")
                
                uncompressedstream = list(map(lambda x: palette[x], uncompressedstream))
                outputstream += uncompressedstream
            
            else:
                raise NotImplementedError()
                
        outputbytes = bytes(outputstream)
        # LZW compression may have left a trailing null byte
        if outputbytes[-1]==0:
            outputbytes = outputbytes[:-1]

        print("Compression done.")
        outputlen = len(outputbytes)

        print("Adding extra blocks...")
        gctrl = (
            0x21,              # start of the graphics control field
            *(0xF9,             # label identifying as graphice control extention
             0x04,),           # size of the graphics control field
            *((1<<5)|colors),  # size of the color table
            *(0x00,)            # no transparancy in use
        )
        gctrl = bytes(gctrl)
        print(*gctrl)
       
        gce = (
            0x2C,              # start of the graphic field
            *(outputlen.to_bytes(3,'big')),   # data size
            *(0x00, 0x00, 0x00),                  # offset of the image
            *(0x00,)                              # transparent colour flag
        )
        gce = bytes(gce)
        print(*gce)

        ctrlfinish = (
            0x3B                                               # end of the frame
           ,*(lzwmincodesize.to_bytes(1,'big'))               # LZW minimum code size
           ,*((lzwmincodesize//8).to_bytes(1,'big'))         # LZW initial code size
           ,*(outputbytes)                                    # compressed data
        )
        ctrlfinish = bytes(ctrlfinish)
        print(*ctrlfinish)
        out.write(gctrl);out.write(gce);out.write(ctrlfinish)
        prevmarker = data[-3:]

