#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # gif spec says maximum size for width and height are 255 pixels
    # - but it's not clear what happens if you exceed this limit?
    max_width = 255
    max_height = 255

    # number of frames in the gif file
    num_frames = 0x1000

    # specify truecolor mode with rgb color resolution; width and height must be <= 255
    screendescr = [
        0x47,          # signature
        0x49,          
        0x46,          
        ord('8'),       
        0xe0 | (num_frames >> 8 & 0xff),    # logical screen descriptor 
        0xe0 | (num_frames & 0xff),       # total number of frames *must* be <= 256 bytes
        0x00,                           # local color table size 
        max_width&0xff,                # width in pixels 
        max_height>>8 & 0xff,          # height in pixels 
        0xf7                        # packed fields 
    ]

    print(f"screen descr: {screendescr}")
    out.write(bytes([el for el in screendescr]))

    # write global color table (all black)
    colorTable = []
    for i in range(max_width):
        colorTable += [0x00, 0x00, 0x00]
    
    sizeOfColorTable = len(colorTable) // 3
    # we don't have any extra frames or extensions
    nBytesToWrite = sizeOfColorTable + 1
    print(f"number of bytes to write in global color table: {nBytesToWrite}")
    out.write(struct.pack(f"<{nBytesToWrite}B", *(colorTable)))

    # write some frames
    buffer = bytearray()
    for _frameId in range(num_frames):

        # each frame has two blocks: 
        #   - graphic control extension block
        #   - image data block 

        # first one contains info about how large our image will be
        # actually written later on
        # - animation duration is set to zero
        # - transparent color index is also set to zero
        # since transparency isn't used, all other fields should be 0
        graphicsCtrlBlock = [
            0x21,                     # extension introducer 
            0xf9,                     
            ord('&'),                 
                num_frames >> 8 & 0xff,    
                 num_frames & 0xff,     
             0x00,                    # background color index 
             0x00,                   # pixel aspect ratio 
            ord('{')                 # start of image data subblock 
        ]
        out.write(bytes([el for el in graphicsCtrlBlock]))
        
        # construct image data block
        # - choose next image's left corner from the random
        #   number generator
        currLeftCorner = rng.read(1)[0] % max_width
        currTopCorner = rng.read(1)[0] % max_height

        # - specify top-left corner and dimensions of current 
        #   frame
        imageDataBlock = [
            0x2c,                       # image separator
            currLeftCorner,             # left corner
            currTopCorner,              # top corner
            
            # desired image dimension
            dynWidth=rng.read(1)[0],   # width
            dynHeight=rng.read(1)[0]&0x3f,# height

            # flag frame disposal method (no) and user input required before continuing
            flags=ord(b'\x00'),

            # no transparancy in use, so no transparency color specified 
            transColIdx=ord('\x00'),
            
            # end of image data subblock
            ord('}')  
        ]
        print(f"image data block: {imageDataBlock}")
        out.write(bytes([el for el in imageDataBlock]))


        # append actual image data to the buffer
        sampleSize = int((dynWidth*dynHeight)/8)
        # fill each row with zeros
        buffer += [0]*sampleSize*(currHeight := min(max_height, dynHeight))
       
        # generate random samples
        for sample in range(sampleSize):
            sampleByte = None
            while True:
                newSampleByte = rng.read(1)[0]

                # make sure we're not reusing the same value twice 
                # - e.g. prevent the case when all bytes would be 0x00
                if not (newSampleByte == sampleByte and sampleByte!= 0):
                    break
            
            sampleByte = newSampleByte
            for bitOffset in range(8):
                # if chosen bit is set, flip corresponding pixel
                # in the currently constructed image, otherwise keep it constant
                
                # NOTE: I tried to check if the bit was already flipped
                # in the current image, which resulted in strange artifacts -
                # maybe because the original code just made sure to set the bit correctly
                # every time... But anyway, the current solution seems to work well enough
                
                if ((sampleByte >> bitOffset) & 0x1) == 1:
                    idx = (((currTopCorner+bitOffset)%max_height)*max_width
                        + ((currLeftCorner + 8*sample)%max_width))

                    # flip only if the previous pixel is still empty
                    if buffer[idx] == 0:
                        buffer[idx] ^= 0xff

        '''
        # show generated random images
        img = Image.frombytes("P", (max_width, max_height), buffer)
        img.show()
        '''

        # finally, write the constructed frame into output file
        out.write(buffer[:int(((dynWidth*dynHeight)/8)*min(max_height, dynHeight))])
        buffer = buffer[(int(((dynWidth*dynHeight)/8)*min(max_height, dynHeight))):]

