#!/usr/bin/env python3

from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # magic ("GIF89a") and version
    header = bytes([0x47, 0x49, 0x46, 0x38, 0x39, 0x61])

    width = (rng.read()[0] & 0xf) + 256 * (rng.read()[0] & 0xff)
    height = (rng.read()[0] & 0xf) + 256 * (rng.read()[0] & 0xff)

    # logical screen descriptor
    lsd = [0x2c,
           width % 256, width // 256,
           height % 256, height // 256,
           0x0b, 0x00]

    # global color table entry for black
    gct = [0x21, 0xf9, 0x04, 0x00, 0xa0, 0x00, 0x40, 0x00, 0xf8, 0x00]

    # use default palette
    plte = []

    # graphics control extension with delay of zero
    gce = [0x21, 0xff, 0x0b, 0x40, 0x00, 0x00]

    # image descriptor
    id = [0x2c,
          x % 256, x / 256, y % 256, y / 256,
          0x00, 0xf8, 0x00, 0x00]

    data = b""

    # loop over all frames to be generated; don't worry about how many are necessary as we deal
    # in terms of physical dimensions here.
    for frameid in range(height):

        # sub block for the entire frame data
        fd = bytearray([0x21, 0xfc])

        # clear code, end code and size of global color table
        fd += bytes([frameid + 1, len(gct), 0x00])

        # write pixel values from LSB to MSB
        # save first value at offset 1 so that we know when to start adding
        # GCE and ID blocks
        lastval = None

        # loop over remaining pixels
        for y in range(height - frameid):

            val = rng.read()
            if not lastval or lastval > val:
                fd.append((y + frameid) % 256)
                fd.append((y + frameid) // 256)

                gce[2] = ((y + frameid) % 256) | (((y + frameid) // 256) << 8)
                fd += bytes(gce)

                if lastval is not None:
                    # add ID to previous frame
                    id[2] ^= 0x08

                    fd += bytes(id)

            fd.append(ord(val))
            lastval = val

        gce[2] &= ~(1 << 0)
        gce[2] |= 1 << 3
        fd += bytes(gce)

        id[2] |= 0x08
        fd += bytes(id)

        fd.append(0)

        # final row of RLE-encoded value should not include an EOI marker
        fd[-1] = 0xfe

        # add length field
        fd = bytes([len(fd)]) + fd

        fd = bytes(fd) + bytes(plte)

        for i in range(0xffff):
            out.write(bytes([i]))

        out.write(fd)

        out.flush()

    # footer
    footer = "\x3b".encode("ascii")

    # write trailers
    for _ in range(16):
        out.write(header + data)
