#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the `settings`.
def generate(out: BinaryIO, width: int = 1024, height: int = 768):
    # Header.
    # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

    out.write("GIF".encode())

    # Screen descriptor.
    # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

    screen_descriptor = bytearray()

    # Width and height.
    width_byte_sequence = struct.pack("<H", width)
    for i in range(2):
        screen_descriptor += bytes([width_byte_sequence[i]])

    height_byte_sequence = struct.pack(">H", height)
    for i in range(2):
        screen_descriptor += bytes([height_byte_sequence[i + len(width_byte_sequence)]])

    # Global color table flag set to true with 256 colors (https://en.wikipedia.org/wiki/GIF).
    global_color_table_flag = b"\x87"
    background_index          = b"\x00"
    pixel_aspect_ratio        = b"\x00"

    screen_descriptor += global_color_table_flag + background_index + pixel_aspect_ratio

    out.write(screen_descriptor)

    # Image descriptor.
    # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

    image_descriptor = bytearray()

    left   = 0
    top    = 0
    delay  = 100
    loops  = 0xFF

    # Left.
    left_byte_sequence = struct.pack("<h", left)
    for i in range(2):
        image_descriptor += bytes([left_byte_sequence[i]])

    # Top.
    top_byte_sequence = struct.pack("<h", top)
    for i in range(2):
        image_descriptor += bytes([top_byte_sequence[i + len(left_byte_sequence)]])

    # Width and height.
    width_byte_sequence = struct.pack("<H", width)
    for i in range(2):
        image_descriptor += bytes([width_byte_sequence[i]])

    height_byte_sequence = struct.pack(">H", height)
    for i in range(2):
        image_descriptor += bytes([height_byte_sequence[i + len(width_byte_sequence)]])

    # LZW minimum code size.
    lzw_minimum_code_size                = b"\x08"
    local_color_table_flag               = b"\x00"
    interlace_flag                       = b"\x00"
    sort_flag                            = b"\x00"
    reserved                             = b"\x00" * 3
    local_color_table_size_exponent      = b"\x00"
    local_color_table_sort_method        = b"\x00"
    local_color_table_dynamic_resolution = b"\x00"
    local_color_table_size               = b"\x00"
    background_index                     = b"\x00"
    pixel_aspect_ratio                   = b"\x00"

    image_descriptor += lzw_minimum_code_size + local_color_table_flag + \
                         interlace_flag + sort_flag + reserved + \
                         local_color_table_size_exponent + local_color_table_sort_method + \
                         local_color_table_dynamic_resolution + local_color_table_size + \
                         background_index + pixel_aspect_ratio

    out.write(image_descriptor)


    # Color table.
    # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

    color_table = bytearray()

    for red_channel in [b"\x00"] * 256 + [b"\xff"]:
        for green_channel in [b"\x00"] * 256 + [b"\xff"]:
            for blue_channel in [b"\x00"] * 256 + [b"\xff"]:
                color_table += bytes([red_channel, green_channel, blue_channel])

    out.write(bytes([len(color_table)]))
    out.write(color_table)

    # Image data.
    # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

    # Disposal method == 0 means no disposal specified.
    dispose_method = b"\x00"

    # User input flag == false.
    user_input_flag = b"\x00"

    # Transparent color index == None or -1.
    transparent_color_index = b"\x00\x00"

    graphic_control_extension = dispose_method + user_input_flag + transparent_color_index

    block_size       = b"\x04"
    frame_count_byte = b"\x00"
    subblock         = graphic_control_extension + block_size + frame_count_byte

    extension_int_id      = b"\xF9"
    extension_label_size  = b"\x04"
    label                = graphics_control_extension_label + extension_label_size

    application_identifier              = "NETSCAPE2.0".encode()
    application_auth_code_size           = b"\x03"
    application_auth_code               = b"\x01\x02\x03"
    subblock                          = subblock + extension_int_id + extension_label_size + label + \
                                        application_identifier + application_auth_code_size + application_auth_code

    end_of_subblocks_size = b"\x00"

    out.write(subblock)
    out.write(end_of_subblocks_size)

    # Extension introducer.
    # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

    out.write(b"\x2c")

    # Textual data blocks.
    # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

    textual_data_blocks = b""

    textual_data_blocks += b"\x01"                  # Block size.
    textual_data_blocks += b"\x0d"                  # Label size.
    textual_data_blocks += b"\x00"                  # Background color index.
    textual_data_blocks += b"\x00"                  # Pixel aspect ratio.
    textual_data_blocks += b"\x00"                  # GCT flag.
    textual_data_blocks += b"\x03"                  # Size of the palette.
    textual_data_blocks += b"\x0e\x60\xf0\xe0\xc0\xa0\x80\x60\x40\x20\x00" # RGB triplets.
    textual_data_blocks += b"\x00"                  # Block terminator.

    textual_data_blocks += b"\x01"                  # Block size.
    textual_data_blocks += b"\x0c"                  # Label size.
    textual_data_blocks += b"\x00"                  # Text grid left position.
    textual_data_blocks += b"\x00"                  # Text grid top position.
    textual_data_blocks += b"\x00"                  # Text grid width.
    textual_data_blocks += b"\x00"                  # Text grid height.
    textual_data_blocks += b"\x00"                  # Character cell width.
    textual_data_blocks += b"\x00"                  # Character cell height.
    textual_data_blocks += b"\x00"                  # Text foreground color index.
    textual_data_blocks += b"\x00"                  # Text background color index.
    textual_data_blocks += b"\x00"                  # Block terminator.

    end_of_textual_data_blocks = b"\x00"             # Block terminator.

    out.write(textual_data_blocks)
    out.write(end_of_textual_data_blocks)

    # Graphics control extension.
    # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

    duration                      = b"\x00\x00"     # Duration in milliseconds.
    transparent_color_index       = b"\x00\x00"     # Default transparent color index.
    block_size                    = b"\x04"         # Block size.
    frame_count_byte              = b"\x00"         # Frame count.
    subblock                      = duration + transparent_color_index + block_size + frame_count_byte
    extension_int_id              = b"\xFE"         # Extension introducer ID.
    extension_label_size          = b"\x00"        