#!/usr/bin/env python3

import struct
from typing import BinaryIO

# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')

    # size of 1024 bytes is the max we can encode using LZW compression
    if datalen > (1 << 10):
        raise ValueError('The gif you requested is too big to be encoded in a single frame.')
    
    out.write(header + datalen.to_bytes(3, byteorder='big'))
    out.write(b'\x87\x00\x00')

    lzw_stream = compress(datalen, rng)
    for chunk in pack(lzw_stream):
        out.write(chunk)


def compress(size: int, rng: BinaryIO):
    """
    The GIF format uses LZW compression to reduce the amount
    of data that needs to be transferred on the wire. This function
    implements the algorithm used by the GIF format specification. 
    """
    dictionary = {}
    previous = b''
    while True:

        try:
            current = rng.read(1)[0]

            key = previous + current.to_bytes(1, 'big')
            if len(key) == 1 or not key in dictionary:
                output = dictionary[previous].to_bytes(1, byteorder='big')

                # add new character to dictionary
                dictionary[key] = len(dictionary)
                yield output 
                previous = current.to_bytes(1, byteorder='big')

            else:
                # use cached value from dictionary
                previous = key[:2]
        
        except OSError as e:
            return 


def pack(data):
    """
    The LZW compression algorithm outputs all codes of maximum length
    12 bits, values larger than this are split into multiple chunks
    with two extra bits reserved per chunk to allow for the transfer
    of large values over the network. 

    Packing and unpacking takes care of padding these values correctly.
    """
    
    # start with a single-byte codeword
    pending = [data.__next__()]
    for c in data:

        pending.append(c)

        if len(pending) >= 2**12:
            
            # pad each chunk with an extra bit
            packed = []
            mask = 0b11
            for n in range((len(pending)-1)//mask+1):
                
                # first level of extra bits
                msb = ((n // ( 1 <<  2)) % 2 & (~mask)).to_bytes(1, 'big')
                mid = ((n // ( 1 <<  1)) % 2 & (~mask)).to_bytes(1, 'big')
                lsb = ((n // ( 1 <<  0)) % 2 & (~mask)).to_bytes(1, 'big')

                packed += [msb | (mid << 1) | (lsb << 2)]

                # second level of extra bits
                msb = ((n // ( 1 <<  6)) % 2 & (~mask)).to_bytes(1, 'big')
                mid = ((n // ( 1 <<  5)) % 2 & (~mask)).to_bytes(1, 'big')
                lsb = ((n // ( 1 <<  4)) % 2 & (~mask)).to_bytes(1, 'big')

                packed += [msb | (mid << 1) | (lsb << 2)]

                # third level of extra bits
                msb = ((n // ( 1 << 10)) % 2 & (~mask)).to_bytes(1, 'big')
                mid = ((n // ( 1 <<  9)) % 2 & (~mask)).to_bytes(1, 'big')
                lsb = ((n // ( 1 <<  8)) % 2 & (~mask)).to_bytes(1, 'big')

                packed += [msb | (mid << 1) | (lsb << 2)]

            # yield chunks
            for p in packed[:-1]:
                yield p
                
            # last chunk doesn't have any extra bits so just yield it
            yield packed[-1]

            # reset our list of pending bytes
            del pending[:]

    # flush remaining buffer before returning
    if len(pending)!= 1:
        packed = []
        mask = 0b11
        for n in range(((len(pending)-1)//mask)+1):

            # first level of extra bits
            msb = ((n // ( 1 <<  2)) % 2).to_bytes(1, 'big')
            mid = ((n // ( 1 <<  1)) % 2).to_bytes(1, 'big')
            lsb = ((n // ( 1 <<  0)) % 2).to_bytes(1, 'big')

            packed += [msb | (mid << 1) | (lsb << 2)]

            # second level of extra bits
            msb = ((n // ( 1 <<  6)) % 2).to_bytes(1, 'big')
            mid = ((n // ( 1 <<  5)) % 2).to_bytes(1, 'big')
            lsb = ((n // ( 1 <<  4)) % 2).to_bytes(1, 'big')

            packed += [msb | (mid << 1) | (lsb << 2)]

            # third level of extra bits
            msb = ((n // ( 1 << 10)) % 2 ).to_bytes(1, 'big')
            mid = ((n // ( 1 <<  9)) % 2 ).to_bytes(1, 'big')
            lsb = ((n // ( 1 <<  8)) % 2 ).to_bytes(1, 'big')

            packed += [msb | (mid << 1) | (lsb << 2)]

        # yield final chunk
        yield packed[-1]
    
    # write terminating chunk
    yield b'\xff\x00'
