#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    version = b'\x01\x00'

    # 24 bits of global color table
    palette = [struct.pack('<BBB', *rng.read(3)) for _ in range(256)]

    # We can have any height > 0
    height = int.from_bytes(rng.read(1), 'big') % (2**8-1) + 1

    width = int.from_bytes(rng.read(1), 'big') % (2**8-1) + 1

    # LZW min code size is always 4 bits
    lzw_min_code_size = bytes([int(rng.read(1).hex(), base=16) & 0b00001111])

    packed_header = header \
                   + b''.join((version,
                               height.to_bytes(1, byteorder='big'),
                               width.to_bytes(1, byteorder='big'))) \
                   + b'' \
                   + lzw_min_code_size \
                   + b'\xFF' \
                   + (palette[255] * len(palette)).ljust(768, b'\x00').lstrip(b'\x00')
    
    # Write header
    out.write(packed_header)

    # Write palette
    out.writelines(palette)


    # Create data stream with all sizes that are < 255
    block_sizes = []
    while True:
        size = rng.read(1)[0]

        if size == 0x00 or size >= 0xff:
            break
        
        block_sizes.append(size)
    
    # Write image descriptor
    out.write(b'\x2C\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
    out.seek(-2, 1) 
    out.write(len(block_sizes).to_bytes(1, byteorder='big'))

    # Extend the block sizes until it contains an amount of elements divisible by three
    # This makes sure we can fill some of our bits in the next step
    while len(block_sizes) % 3!= 0:
         random = rng.read(1)[0]

         # Make sure the last value is not zero
         while random == 0x00 and len(block_sizes)!= 1:
             random = rng.read(1)[0]
         
         block_sizes.append(random)

    # Split the downloaded values into chunks of three which will be added to the output
    chunked_blocks = list(zip(*[iter(block_sizes)]*3))

    # Add each element of the first part
    for chunk in chunked_blocks[:-1]:
        out.write(chunk[-1].to_bytes(1, byteorder='big'))

    # Now add one more value at the end of the last group
    # If the total amount of blocks was odd this will simply overwrite the previous value
    # Otherwise it should already be set correctly
    out.write(chunked_blocks[-1][-2].to_bytes(1, byteorder='big'))

    # Add padding zeros until the length of all blocks is divisible by two
    while len(block_sizes) % 2!= 0:
        block_sizes.append(0x00)
   
    # Add the block size values as little endian integers
    for i in range(0, len(block_sizes), 2):
        value = int(''.join(['{:02X}'.format(c) for c in reversed(block_sizes[i:i+2])]), base=16)
        out.write(value.to_bytes(2, byteorder='little'))

    # Fill remaining space with random data
    while out.tell() < 13:
        out.write(rng.read(1))

    out.close()

