#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big') & 0xfff

    # Write the header
    out.write(header)
    print(f"Datalen: {datalen}")

    # Generate and write the rest of the GIF data
    for i in range(datalen):
        ptype = rng.read(1)[0]

        if ptype == 0x2C:         # Image Descriptor (extension block)
            pass
        elif ptype == 0x21:       # Extension Block
            ext_label = rng.read(1)[0]

            if ext_label == 0xf9:
                # Graphics Control Extension
                gc_blocksize = rng.read(1)[0]

                if gc_blocksize!= 4:
                    raise Exception("Invalid GC extension size")
                
                flags = rng.read(1)[0]
                delay = int.from_bytes(
                        rng.read(2), byteorder='little', signed=False)
                transparent = rng.read(1)[0]
                
                skip = 0
                while not is_leftover_zeroes(rng):
                    skip += 1
            
                out.seek(-skip - 1, 1)
                
            else:
                raise Exception("What kind of extension are you doing?")
    
        elif ptype == 0x3b:       # Trailer
            break
        
        else:                     # Other reserved blocks
            continue

        # Recording this type lets us know how to process it later on
        out.write(struct.pack('B', ptype))

        # Copy over the data from `in` to `out`
        copy_until_zero(rng, out)
    
    out.flush()

def is_leftover_zeroes(datastream: BinaryIO) -> bool:
    leftovers = datastream.read(1024)
    
    if len(leftovers) < 1024:
        leftovers = leftovers + \
                bytes([datastream.read(max(1, 1024-len(leftovers)))])

    assert isinstance(leftovers[0], int), f"{leftovers[:5]}"
    leftmostzeros = 0
    for idx, v in enumerate(leftovers):
        if v == 0x00:
            leftmostzeros = max(leftmostzeros, idx+1)
        else:
            break
        
    datastream.seek((-leftmostzeros), 1)
    return any((l==0 for l in leftovers))
    
def copy_until_zero(src: BinaryIO, dst: BinaryIO) -> None:
    while not is_leftover_zeroes(src):
        byteval = src.read(1)[0]
        print(byteval)
        dst.write(struct.pack('B', byteval))
    
    return True

