#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Image block
    image_block  = b'!' * 64
    width        = rng.read(1)[0] % 90 + 10   # Can't be too narrow or we'll get an error code (or maybe just overrun?)
    height       = rng.read(1)[0] % 56 + 10   # Might as well limit ourselves to 24-bit screen resolutions for now?
    aspect_ratio = [ 4 << i & int.from_bytes(rng.read(1), byteorder='big') for i in range(3)]
    palette      = bytes([ i | j           for _ in range(2 ** aspect_ratio[2])
                                                            for i,j in zip((7<<5)|int.from_bytes(rng.read(1), byteorder='big'), (7<<5)|int.from_bytes(rng.read(1), byteorder='big')) ])
    table_size   = len(palette).to_bytes(1, 'little', signed=False)
    color_depth  = (aspect_ratio[0]*aspect_ratio[1]).to_bytes(1, 'little', signed=False)
    image_descriptor = []
    image_descriptor += ((width - 1).to_bytes(2, 'little', signed=False))
    image_descriptor += ((height - 1).to_bytes(2, 'little', signed=False))
    image_descriptor += (((~(width-1)) & 0xFFFF).to_bytes(2, 'little', signed=False))    # Complement of above two values
    image_descriptor += (((~(height-1)) & 0xFFFF).to_bytes(2, 'little', signed=False))   # Complement of above two values
    image_descriptor += color_depth
    image_descriptor += b'\x80\x00'                                                   # LZW minimum size code length
    image_descriptor += b'\xFF\xFB'                                                   # Use second local colour table
    image_descriptor += b'\x00\x00'                                                   # No disposal method
    image_descriptor += b'\x00'                                                       # Ignore delay time and transparent colour index fields
    image_descriptor = bytes(image_descriptor)

    # Graphics control extension block
    gce_block     = b''
    gce_block    += b'\x21'                                                        # Extension Introducer
    gce_block    += b'\xF9'                                                        # Graphic Control Label
    gce_block    += b'\x04'                                                        # Block Size
    gce_block    += b'\x00'                                                        # Reserved (set to zero)
    gce_block    += b'\x00'                                                        # Flags
    gce_block    += b'\x00'                                                        # Transparent Colour Index
    gce_block    += b'\x00\x00'                                                    # Delay Time
    gce_block    += b'\x00'                                                        # Transparent Colour Number

    start = out.tell()

    out.write(header)
    out.write(struct.pack('<h', width))
    out.write(struct.pack('<h', height))
    out.write(b'\xf0')
    out.write(palette[:2**aspect_ratio[2]])
    out.write(table_size+color_depth)
    out.write(image_descriptor)

    global_index = 0
    while True:
        byte = ord(rng.read(1))

        if not byte:
            break
        
        if byte!= 0xFF:
            global_index |= byte << 8
            continue

        # Write previous pixel data if necessary
        if global_index:
            out.write(((global_index >> 8)&0xFF).to_bytes(1,'big',signed=True))
            out.write(((global_index     )&0xFF).to_bytes(1,'big',signed=True))

            global_index = 0
            
        byte = ord(rng.read(1))

        if byte == 0x21: # Extension
            subbyte = ord(rng.read(1))
            
            # Plain text
            if subbyte == 0x11:
                out.write(subbyte.to_bytes(1,'big',signed=False))    # Extention type

                length = rng.read(1)[0]%11 + 1                       # This doesn't seed correctly -- why??

                out.write(length.to_bytes(1,'big',signed=False))
                
                string = bytes([i|j for _ in range(length)
                                            for i,j in zip(rng.read(1),(7<<5)|int.from_bytes(rng.read(1),'big'))])
                                
                out.write(string)
                
            else:
                raise ValueError('Unhandled extension type: {}'.format(hex(subbyte)))
                
        elif byte <= 0x1F: 
            pass # Do nothing with this byte; it's an escape sequence that should probably be handled by the decoder but isn't usually any use to us here anyway so whatever
            
        elif byte == 0x2C: # Image Descriptor
            out.seek(-start,-1)
            out.write('\x21'.encode()) # Re-add graphic control block prefix
            out.write(gce_block)
            out.write(struct.pack('<h', width))
            out.write(struct.pack('<h', height))
            out.write(b'\xf0')
            out.write(palette[:2**aspect_ratio[2]])
            out.write(table_size+color_depth)
            out.write(image_descriptor)
        else:
            raise ValueError("Unknown byte sequence! {:02X}".format(ord(rng.read(1))))
