#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(rng.read(2), byteorder='big')

    globalcolormapentrysize = rng.read(1)
    if len(globalcolormapentrysize): # EOF check
        globalcolormapentrysize = ord(globalcolormapentrysize[0])

    if globalcolormapentrysize!= 7 and globalcolormapentrysize!= 9:
        raise RuntimeError("Invalid color map entry size")

    globalcolormaplength = int.from_bytes(rng.read(3), byteorder='big')
    
    palette = []
    for _ in range(globalcolormaplength+1):
        rgb = [int(_) % 0x100] * globalcolormapentrysize
        palentry = bytes([(entry // (0x20 + i)) & 0xff for i, entry in enumerate(rgb)])

        if not any((palentry == old for old in palette)):
            palette.append(palentry)
    
    out.write(header)
    out.write(struct.pack('>H', datalen-69))
    out.write(b'\x21\xF9\x04')
    out.write(b'\x00'*5)
    out.write(struct.pack('BB', 0x80 | globalcolormapentrysize, 0))
    out.write(b'\xFF'*7)
    out.write(struct.pack(">HH", 0x100, globalcolormaplength))
    out.write(b"".join(palette))


    while True:
        blocktype = rng.read(1)[0]
        
        if blocktype == 0x2C:

            leftpos = int.from_bytes(rng.read(2), byteorder="little")
            toppos = int.from_bytes(rng.read(2), byteorder="little")
            width = int.from_bytes(rng.read(2), byteorder="little")
            height = int.from_bytes(rng.read(2), byteorder="little")

            packedbyte = rng.read(1)[0]
            
            localcolormapflag = bool(packedbyte&0x80)
            interlaceflag = bool(packedbyte&0x40)
            sortflag = bool(packedbyte&0x20)
            haslocalcolormap = bool(packedbyte&0x10)
            
            if sortflag or localcolormapflag or haslocalcolormap or interlaceflag:
                continue
                
            lctsize = max(2**(((packedbyte&0x07)<<1)+1), 3)
            encodedimagedata = rng.read(width*height)
            
            totalbytesperline = ((lctsize//8)*width)+min(lctsize//8, width%8)
            padbytesperline = (-totalbytesperline)%4
            leadingpaddingbytes = -leftpos%4
            trailingpaddingbytes = padbytesperline-(padbytesperline-trailingpaddingbytes)%4

            content = struct.pack("<HHHHB", leftpos, toppos, width, height, 0x80|lctsize)
            content += encodedimagedata
            content += b"\x00"*(leadingpaddingbytes+(width*height))
            content += b"\x00"*(trailingpaddingbytes)

            out.write(bytes([blocktype]))
            out.write(content)
            
            
        elif blocktype == 0x3B:
            break

