#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return False
        
        # Get block size
        blocksize = int(((fieldbyte << 8) | ord(rng.read(1)))) + 1

        # Block types are as follows:
        #   0x00 - Extension Introducer
        #   0x2C - Image Descriptor
        #   0x21 - Plain Text Extension
        #   0xFF - Application Extension
        #   0xFE - Comment Extension
        #   0xF9 - Graphic Control Extension
        # We only support generating plain text extensions since they're easy!
        if (ord(rng.read(1))!= 0x1B or
            ord(rng.read(1))!= 0x01):

            print('Invalid GIF block')
            exit(-1)

        # If this was not an image descriptor block, just skip ahead...
        if (fieldbyte!= 0x2C):
            
            # Skip past the data of invalid blocks
            rng.seek(blocksize-1, 1)
            continue
            
        # Create the image descriptor header
        idh = rng.read(9)

        # Write image descriptor header to output
        out.write(idh)

        # Read local color table length
        loccltlen = 2**(int((idh[8] >> 5)))*3

        # Read local color table if it exists
        if loccltlen > 0:
            locclt = rng.read(loccltlen)
            out.write(locclt)

        # Now read the image data by pixel
        imgdata = []
        palentries = []
        i = 0

        while i < blocksize-1:
          numclridx = ord(rng.read(1))

          if numclridx == 0:
              # Zero indicates end of image data
              break
            
          clridx = rng.read(numclridx)
          
          numimgidx = ord(rng.read(1))
          imgidx = rng.read(numimgidx)

          # Add all idx values to palette list so that we can find unrepeated ones later
          palentries += [c for c in clridx]

          # Ensure we don't exceed maximum valid index
          if max(palentries)-1 >= loccltlen-1:
            raise ValueError("Too many different colors in generated GIF!")

          # Convert palette indices to RGB tuples and save them to our image data array
          imgdata += [(
                tuple([
                    ord(locclt[(i*3)+k]) for k in range(3)])
                  ) for i in clridx]

        # Find unique palette entries and convert to binary string
        uniquepalentires = set(palentries+[-1])
        uniquepal = ''
        for p in uniquepalentires:
            uniquepal += struct.pack('<BBB', *(p+(0,))[:3])

        # Print number of unique colors found inside image
        print(f"Image contains {len(uniquepalentires)} unique colors")

        # Create image data subblocks with pointers to first entry of each color in palette
        datsubblks = [[len(uniquepal)]]
        lastidx = None
        for d in imgdata:
          if d!= lastidx:
              datsubblks.append([uniquepal.index(struct.pack('<BBB', *(d+(0,))[:3])), len(datsubblks)])
          else:
              datsubblks[-1][1]+=1

          lastidx = d
          
        # Turn list of lists into bytes object
        datsubblks = bytes([b for s in datsubblks for b in s])+b'\x00'+b'\xff'
        
        # And write the packed image data to the file
        out.write(bytes([len(datsubblks)]))
        out.write(datsubblks)

