#!/usr/bin/env python3
"""Generate an rst file with a table containing the header and all types, verifying the output of `get_type` (in `src\rust_python_types.rs`)."""
import os
from typing import List


def main():
    """Run main function."""

    # Remove previous file if any:
    try:
        os.remove("header-and-types.rst")
    except FileNotFoundError:
        pass  # No need to remove the file since it does not exist yet

    with open(os.path.join('doc','source', 'header-and-types.rst'), "w", encoding='utf-8') as out_file:

        print('''=================
Header & Types
=================''',
              file=out_file)

        for type in get_type_list():
            print(format_row(type),
                  file=out_file)


def format_row(type):
    return f'''{type['header']}
 .. code-block:: c

      PyTypeObject {type["name"]} = {{
          PyObject_HEAD_INIT(&PyType_Type)
          0,                                  /*ob_size*/
          "{type['name']}",                   /*tp_name*/
          sizeof({type['name']}),              /* tp_basicsize */
          0,                                  /* tp_itemsize */
          (destructor){type['destructor']},   /* tp_dealloc */
          0,                                  /* tp_print */
          0,                                  /* tp_getattr */
          0,                                  /* tp_setattr */
          0,                                  /* reserved */
          0,                                  /* tp_repr */
          0,                                  /* tp_as_number */
          0,                                  /* tp_as_sequence */
          0,                                  /* tp_as_mapping */
          0,                                  /* tp_hash  */
          0,                                  /* tp_call */
          0,                                  /* tp_str */
          0,                                  /* tp_getattro */
          0,                                  /* tp_setattro */
          0,                                  /* tp_as_buffer */
          Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /* tp_flags */
          "{type['type']}",                    /* tp_doc */
          0                                   /* tp_traverse */
      }};'''


def gen_destroy_function(types_: List[dict]) -> str:
    return '''void RustPythonTypes_destroy()
{
%s}
''' % '\n'.join([f'    destroy_{t["name"]}_type();' for t in types_])


def get_type_list():
    """Get the list of types from src/rust_python_types.h using ClangTooling."""
    from clang.cindex import Index
    index = Index.create()
    tu = index.parse('include\\rust_python_types.h', options=['-x', 'c++'])

    types = []
    for node in [node
                 for translation_unit in tu.cursor.get_children()
                 for node in translation_unit.cursor.get_children()
                 ]:
        if node.kind == node.STRUCT_DECL or node.kind == node.CLASS_DECL:

            name = None
            destructor = None
            type = None
            for child in node.get_children():
                if child.kind == child.FIELD_DECL and \
                        len(child.displayname) > 12 and \
                        child.displayname[:12] == '_Py_DeallocFn':
                    destructor = child.spelling

                elif child.kind == child.VAR_DECL and child.spelling == 'tp_name':

                    # The variable has this signature: char *tp_name;
                    field_declaration = next((field
                                            for field in child.get_children()
                                            if field.kind == field.TYPE_REF),
                                           None)

                    # We need to find the declaration that corresponds to the field declaration.
                    decscope = field_declaration.extent.start.file.tu.cursor
                    vardec = decscope.find_lexical_parent(lambda x:
                                                        x.kind ==
                                                        x.VAR_DECL).displayname

                    assert vardec is not None, "Could not find tp_name"
                    name = bytes(vardec, 'ascii').decode('unicode_escape') + 'Object'


                elif child.kind == child.CONST_DECL and \
                        ("tp_basicsize" in child.spelling
                         or "tp_itemsize" in child.spelling) :


                    vardec = ''.join(child.displayname
                                     .split('_')[:-2])  # tp_basicsize => basicsize

                    if vardec == 'basicsize' and child.type.kind!= child.INDEX:

                        declstartline = child.extent.start.line - node.extent.start.line
                        newdecl = child.type.get_canonical().spelling
                        if declstartline < 5:
                            newdecl += ';'

                        lines = ''
                        currentline = declstartline - 1
                        while True:

                            line = next(node.extent.start.file.tu.get_tokens(),
                                        None)

                            if line is None or line.location.line >= declstartline:
                                break
                            else:
                                lines += line.spelling + "\n"

                                if line.location.line == currentline:
                                    currentline -= 1

                        lines = lines[:-len("\n")]


                        oldlines = lines
                        start = 0
                        end = len(oldlines)

                        while True:
                            stop = oldlines.find('\n', start)
                            if stop == -1:
                                break

                            def starts_with(string, prefix):
                                string = string.lstrip().split(
                                   '')[0].rstrip(';')
                                return string.endswith(prefix)

                            if starts_with(oldlines[:stop], '#define'):
                                start = stop+1
                            elif starts_with(oldlines[:stop], '}'):
                                # Skip everything until we are after the struct definition
                                end = stop
                                break
                            else:
                                raise ValueError(
                                    "Clang could not parse the tp_* constants.")

                        lines = oldlines[:end]
                        lines = lines.replace('%d', '%ld')
                        lines = lines.replace(', ', ',\n        ')
                        lines = lines.replace('{', '{\n        ')
                        lines = lines.replace('[', '[\n        ')

                        lines = lines.replace('#ifdef __cplusplus', '// #ifdef __cplusplus')
                        lines = lines.replace('__attribute__((visibility("default"))) inline void* operator new (unsigned long size)',
                                              '// __attribute__((visibility("default"))) inline void* operator new (unsigned long size)')

                        newdecl = lines + '\n' + newdecl

                        newdecl = newdecl.replace(
                            '%ld', 'int')  # TODO: Make some checks on this.
                    else:
                        newdecl = ''.join(child.type
                                        .get_canonical()
                                        .spelling.split(' '))
                        if newdecl[-1] == ';':
                            newdecl = newdecl[:-1] + ';\n'

                    lines = ''
                    currentline = child.extent.start.line - 1
                    while True:

                        line = next(node.extent.start.file.tu.get_tokens(),
                                    None)

                        if line is None or line.location.line > child.extent.end.line:
                            break
                        else:
                            lines += line.spelling + "\n"

                            if line.location.line == currentline:
                                currentline -= 1

                    newdecl = lines[:-len("\n")] + newdecl

                    newdecl = newdecl.replace('PyLongObject', 'long')
                    newdecl = newdecl.replace('PyObject', 'object')
                    newdecl = newdecl.replace('.ob_base', '')

                    newdecl = reindent(newdecl, 4*'')
                    newdecl = newdecl.replace('      // #ifdef __cplusplus',
                                                '#ifdef __cplusplus')

                    newdecl = newdecl.replace(",\n#endif//__cplusplus", ");")

                    newdecl = newdecl.replace("%ld", "%d")

                    newdecl = newdecl.replace("#ifdef __cplusplus", "// #ifdef __cplusplus")

                    newdecl = newdecl.replace("(int)", "(long)")

                    newdecl = newdecl.strip() + "\n"

                    # Check that the expression can be compiled by just adding the missing semicolon
                    compile(f"{newdecl};", "<tmp>", "exec")



                    lastpart = child.spelling.split("_")[-1]
                    if lastpart == "size":
                        part1 = ''.join(child.spelling
                                      .split("_")[:-2])  # tp_basicsize => basicsize
                    else:
                        part1 = child.spelling.rpartition('tp_')[0][:-1]

                    assert part1 in ('basicsize',
                                     'itemsize'), "Invalid constant"

                    if lastpart == "size":
                        newdecl = newdecl.replace(lastpart,'sizeof(%s)'
                                                  % (bytes(node.spelling,
                                                           'ascii').decode(
                                                      'unicode_escape')))
                    else:
                        newdecl = newdecl.replace(lastpart, '%s_%s()'
                                                  % (bytes(node.spelling,
                                                           'ascii').decode(
                                                      'unicode_escape'),
                                                     part1))

                    newdecl = reindent(newdecl, 6*' ')

                    newdecl = newdecl.replace(
                        '.tp_basicsize = ', '.tp_basicsize = %d,\n            ')
                    newdecl = new