#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng):
    # Write header
    out.write(b'GIF89a')

    width = 0x20 + rng.randint(16, 48);
    height = 0x20 + rng.randint(16, 48);


    # Logical Screen Descriptor (LSD)
    out.write(struct.pack('<HHBBB',
        width,  # Width
        height,  # Height
        0xf0 | rng.randint(0,7),  # LCT (Local Color Table) flag with one of 2**3 values and the reserved bit set to zero
        0,  # GCT (Global Color Table) flag; must be set ot zero for proper interpretation of image data
        ))
    
    # Background Color Index
    bgcolidx = rng.randint(0, 255)
 
    # Pixel aspect ratio
    pixel_aspect_ratio = 0 # no-scaling

    # Global color table
    # Only one entry required since all colors are unique
    global_color_table_entry_size = 3      # RGB triples
    global_color_table_len = 1 << ((bgcolidx & 0x7) + 1)   # The correct length in bytes is determined by the number of entries in the CT
    out.write((global_color_table_entry_size * b'\xff').ljust(global_color_table_len))    # Fill with 0xFFs

    

    # Application Extension block
    appextlabel = 'NETSCAPE'.encode('ascii')
    # Block size excluding subfields
    extblocksize = len('\x0b\xff\x0b') + \
            len(appextlabel) + \
            2                     # Application identifier subfield
    # Subfield size excluding data
    identsubsize = len('\x08\xff') + \
            len(appextlabel)      # Application identifier field including label
    # Data content
    netscape_looping_value = 1
    # Actual block size
    blocktotalsize = extblocksize + 3 + identsubsize + 1 + len(netscape_looping_value) 
    out.write(struct.pack(">BB", 0x21, blocktotalsize))                    # Extension Introducer and extension label
    out.write(appextlabel)                                                 # Application Identifier Label
    out.write(struct.pack("<BH", 3, len(netscape_looping_value)))            # Subfield code & size
    out.write(b"NETSCAPE")                                                  # Netscape application ID string
    out.write(struct.pack("B", netscape_looping_value[0]))                   # Loop value byte

    # Graphics control extension
    gce_flags = 0x04                      # Transparent colour flag
    transparent_colour_idx = 0           # Placeholder
    delaytime = int(((rng.randint(100, 500)/100) ** 2) * 10)             # Delay time in hundredths of seconds
    disposal_method = 0                  # No special treatment
    userinputflag = False                # User input not expected
    transp_colour_idx_packed = struct.pack("<BBBH",
                                                gce_flags,
                                                userinputflag,
                                                disposal_method,
                                                delaytime
                                            )
    trns_block_size = len("\x21\xf9\x04") + len(transp_colour_idx_packed)
    gce_content = "\x21\xf9\x04" + transp_colour_idx_packed + "!"*delaytime

    out.write(struct.pack(">BB", 0x21, trns_block_size))        # Extension Introducer and graphics control extension label
    out.write(gce_content.encode())                            # Block size, packed fields and delay duration    
    out.write(struct.pack("<B", transparent_colour_idx))         # Trans parent index

    # Plain Text Extension block
    txtseqnum = 0x01
    txthorizontalalignment = 0     # Left alignment
    textverticalalignment = 0       # Top alignment
    fontid = 0                       # Default font
    charwidth = width // 20         # Approx. width per character
    charheight = 8                   # Approx. height per character
    fgcindex = 0                     # Black foreground
    bgcindex = 0                     # White background
    csri = 0                         # Not compressed
    icli = 0                         # Not interlace
    ctbi = 0                         # Not sorted
    tte_block_size = len("\x21\x00\x0b") + len(txtseqnum.to_bytes(1,'big')) + \
                    len(txthorizontalalignment.to_bytes(1,'big')) + len(textverticalalignment.to_bytes(1,'big')) + \
                    len(fontid.to_bytes(1,'big')) + len(charwidth.to_bytes(2,'big')) + len(charheight.to_bytes(2,'big')) + \
                    len(fgcindex.to_bytes(1,'big')) + len(bgcindex.to_bytes(1,'big')) + len(csri.to_bytes(1,'big')) + \
                    len(icli.to_bytes(1,'big')) + len(ctbi.to_bytes(1,'big'))
    tte_content = "\x21\x00\x0b" + txtseqnum.to_bytes(1,'big') + \
                  txthorizontalalignment.to_bytes(1,'big') + textverticalalignment.to_bytes(1,'big') + \
                  fontid.to_bytes(1,'big') + charwidth.to_bytes(2,'big') + charheight.to_bytes(2,'big') + \
                  fgcindex.to_bytes(1,'big') + bgcindex.to_bytes(1,'big') + csri.to_bytes(1,'big') + \
                  icli.to_bytes(1,'big')+ ctbi.to_bytes(1,'big')

    out.write(struct.pack(">BB", 0x21, tte_block_size))                              # Extension introducer and type label
    out.write(tte_content.encode())                                                # Rest of packet
    

    # Graphic Control Extension block
    gce_flags = 0x00 | 0x01                 # Disposal method: leave graphic unchanged after drawing update
    transparent_colour_idx = 0            # Placeholder
    delaytime = 0                          # No need to animate!
    disposal_method = 0                    # No special treatment
    userinputflag = True                  # Expect user input
    transp_colour_idx_packed = struct.pack("<BBBH",
                                             gce_flags,
                                             userinputflag,
                                             disposal_method,
                                             delaytime
                                         )
    trns_block_size = len("\x21\xf9\x04") + len(transp_colour_idx_packed)
    gce_content = "\x21\xf9\x04" + transp_colour_idx_packed + "!"*(delaytime//33)+ chr(0xfe).encode()

    out.write(struct.pack(">BB", 0x21, trns_block_size))        # Extension Introducer and graphics control extension label
    out.write(gce_content.encode())           # Block size, packed fields and delay duration

    # Image descriptor
    lzwmincodesize = 0x01               # Minimum code size to represent data
    imgleftposition = 0
    imgtopposition = 0
    imgwidth = width
    imgheight = height
    local_color_table_flag = 0          # Local color table follows
    interlace_flags = 0                 # Non interlaced
    sort_flags = 0                      # Not sorted
    int_packed = struct.pack("BBBB",
                                  lzwmincodesize,
                                  imgleftposition,
                                  imgtopposition,
                                  imgwidth
                             )
    lzwmincodesize_int_packed = "\x01\x00\x00" + int_packed.decode().ljust(3,"\x00").ljust(11," ")
    out.write(lzwmincodesize_int_packed.encode())                    # Packed options field
    out.write(chr(imgheight).encode())                              # Image height
    out.write(chr(interlace_flags).encode())                        # Interlace flag

    if local_color_table_flag == 0:                           # Local color table follows
        return

    local_color_table_entry_size = 3      # RGB triplets
    local_color_table_len = 1 << ((bgcolidx & 0x7) + 1)   # The correct length in bytes is determined by the number of entries in the CT
    out.write((local_color_table_entry_size * b'\xff').ljust(local_color_table_len))    # Fill with 0xFFs

