#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # "Magic" header bytes and version identifier
    data = b'GIF89a\0'

    # This is just one frame with no color map. Each bit of this byte is the delay
    # for that particular pixel (in hundredths of a second).
    # 1 means show 256ms or so before moving on to the next.
    colorsPerPixel = 7
    paletteBits = colorsPerPixel * 3
    delayBytes = int((paletteBits + 7)/8)
    delays = [b'\x01']*delayBytes

    # The size of each block in the raster bitmap
    blockSize = 1 << colorsPerPixel
    width = height = 4096 // blockSize
    packedByteCount = int(((width//colorsPerPixel)+1)*height / 8);
    packedBitmapBytes = rng.read(packedByteCount)
    bitmapBytes = []
    
    index = 0;
    def readNextByte():
        nonlocal index
        if index == len(packedBitmapBytes):
            return None
            
        v = ord(packedBitmapBytes[index])
        index += 1
        return v
    
    while True:
        i = 0
        
        currentBlock = 0
        while True:
            bits = 0
            
            # Read 3bits at a time until we have exactly 'colorsPerPixel' 1s in a row
            while True:
                c = readNextByte()
                
                if not c:
                    break
                    
                bits <<= 8
                bits |= c

                if not bits & 7:
                    break

            bits >>= (bits&7) - 1;
            
            currentBlock <<= colorsPerPixel
            currentBlock |= bits & ((1<<colorsPerPixel)-1);
            i += colorsPerPixel

            if i >= width:
                break;

        if readNextByte()!= 0:
            raise ValueError("Invalid padding")

        count = i >> colorsPerPixel
        while count > 0:
            bitmapBytes.append(currentBlock & 255)
            currentBlock >>= 8
            count -= 1

        if i == width*height:
            break


    # Add blocks to the output
    totalLength = \
        17 + # Header length \
        256*(len('Plain text')+1) + \
        len(data) + \
        13 + # First image descriptor length \
             len(paletteBits)*3 + \
                 len(delays) + \
                     len(bitmapBytes) + \
                         len(b",\xff\xff")

    print("Total data length:", totalLength)

    out.write(struct.pack("<L", totalLength))
    out.write(data)
    
    # Plain text extension
    writeShortString(out, "\x01plain text")
    writeByte(out, 1) # Extension label
    writeByte(out, 0xFF) # Application extension
    writeShortString(out, "NETSCAPE2.0") # Application name
    writeShortString(out, "http://ns.adobe.com/xap/1.0/") # Application URL
    writeByte(out, 3) # Block size
    writeShortBE(out, 1) # Subblock size
    writeShortBE(out, 2) # Loop count 0=repeat forever
    writeByte(out, 0) # Terminator
    
    # Image descriptor
    writeByte(out, 0x2C) # Image separator
    writeShortBE(out, width)
    writeShortBE(out, height)
    writeByte(out, 0) # Packed field: local color table flag
    writeByte(out, paletteBits) # Palette size
    for _ in range(int(paletteBits/8)):
        writeByte(out, 0) # RGB values
    
    out.write(bytes(delays)); # Delay times
    out.write(bytes(bitmapBytes)) # Image data
    out.write(b',\xff\xff') # Trailer


def writeShortString(f: BinaryIO, s: str):
    f.write(s.encode())
    f.write(b"\0");
    
def writeByte(f: BinaryIO, n: int):
    assert n <= 255
    f.write(struct.pack("B", n));
    
def writeShortBE(f: BinaryIO, n: int):
    assert n < 2**16
    f.write(struct.pack(">H", n));
