#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # LZW min code size byte is 1 (we can change this later if we want)
    lzw_mincodesize = 1

    width = int.from_bytes(rng.read(2), 'big') % 256 + 1
    height = int.from_bytes(rng.read(2), 'big') % 256 + 1

    globalcolortable_present = bool(int.from_bytes(rng.read(1), 'big'))
    colorresolutionbits = int.from_bytes(rng.read(1), 'big') & 0b111

    backgroundindex = int.from_bytes(rng.read(1), 'big')
    pixelaspectratio = int.from_bytes(rng.read(1), 'big')

    imagedescriptor = bytes([
        0x47, 0x49,    # "GI" in ASCII
        0x46, 0x38, 0x39, 0x61  # "89a" in ASCII
    ]) + struct.pack('BBBBB', width, height,
        (backgroundindex << 4) |
            ((globalcolortable_present << 3) |
                (lzw_mincodesize)),
            1 << colorresolutionbits,
            pixelaspectratio)

    # Only global color table for now!
    palette = [(i * 157) % 256 for i in range(256)]
    globalcolortable = bytes([(len(palette))]) + \
                        bytes(sum(([b, g, r]
                                   for [r, g, b] in map(lambda v: bytes([v]),
                                                         palette)), []))

    frames = []
    numframes = int.from_bytes(rng.read(1), 'big') % 10 + 1
    for frame in range(numframes):
        # Frame delay time is between 1 and 9 inclusive
        duration = int.from_bytes(rng.read(1), 'big') % 9 + 1

        labelbyte = rng.read(1)[0] & 0b11
        sublabelbyte = None
        if labelbyte == 0b00:
            # Plain Text Extension
            labelbyte |= int.from_bytes(rng.read(1), 'big') << 2
            sublabelbyte = rng.read(1)[0] & 0b11111
        elif labelbyte >= 0b11:
            labelbyte |= (int.from_bytes(rng.read(1), 'big') << 2) | \
                         (int.from_bytes(rng.read(1), 'big') << 4)
            sublabelbyte = rng.read(1)[0] & 0b111

            if not sublabelbyte:
                pass
            elif sublabelbyte <= 0b00111:
                pass
            else:
                assert False, f"Unknown sublabel {sublabelbyte}"

        if labelbyte < len(palette):
            bgcolor = palette[labelbyte]
        else:
            bgcolor = 0

        uselocalcolortable = bool(int.from_bytes(rng.read(1), 'big'))
        localcolorindices = []
        localcolorpalette = []

        if uselocalcolortable:
            # Local color table begins with its length followed by RGB triplets
            localcolorpalettelength = int.from_bytes(rng.read(1), 'big') >> 2
            localcolorpalette = [[int.from_bytes(rng.read(1), 'big'),
                                  int.from_bytes(rng.read(1), 'big'),
                                  int.from_bytes(rng.read(1), 'big')]
                                 for _ in range(localcolorpalettelength)]
            localcolorindices = [rng.read(1)[0] % len(localcolorpalette)]

        # Prefix end character
        data = [0xf7]
        for _ in range((width * height)):
            palidx = rng.read(1)[0] % len(localcolorindices)
            while (palidx > 0xff - len(localcolorindices)):
                data.append(0xfe)
                palidx -= 0xff - len(localcolorindices)
            data.append(localcolorindices[palidx])

        data += [
            0x3B  # Trailer
        ]
        frames.append({
            "duration": duration,
            "data": data,
            "uselocalcolortable": uselocalcolortable,
            "localcolorindices": localcolorindices,
            "bgcolor": bgcolor
        })

    gif = header + imagedescriptor + globalcolortable
    for frame in frames[:1]:
        gif += bytes([frame["duration"]]) + struct.pack('BBBBBBB',
            0, 0, 0, 0, bgColorByte1(frame["bgcolor"]),
            bgColorByte2(frame["bgcolor"])) + bytes(frame["data"])
    for frame in frames[1:]:
        positionindicator = 0
        if frame["uselocalcolortable"]:
            positionindicator |= 0b1000
        if frame["localcolorindices"][0]!= 0:
            positionindicator |= 0b0100
        if frame["bgcolor"]:
            positionindicator |= 0b0010
        positionindicator |= 0b0001

        gif += bytes([positionindicator]) + bytes([frame["duration"]]) + \
               struct.pack('BBBBBBB',
                    len(frame["localcolorindices"]) - 1,
                    0, 0, 0, bgColorByte1(frame["bgcolor"]),
                    bgColorByte2(frame["bgcolor"])) + \
               bytes(frame["data"]) + bytes(frame["localcolorindices"])
    gif += b';'

    return out.write(gif)

