#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big') & 0xfff
    framecount = rng.read(1)[0] & 0xff

    print("Generating", "2**" + str(datalen),
          "byte data blocks and", framecount, "frames.")
    # Write main GIF header
    out.write(header)
    # Write global (logical screen descriptor)
    out.write(struct.pack("<hhHHHbbbbb",
                          *([datalen]*7)))
    # Write palette in background color block
    out.write(b'\x00' * ((7+3)*4))
    
    for i in range(framecount):
        out.write(struct.pack("<BBBBB", *[
                0x2c,
                # size
                0x00, 0x00,
                # left, top, width, height
                0x00, 0x00, *(rng.read(4)),
                # local color table flag
                # 16 bit colors
                (((rng.read(1)[0] % 5)+1) << 4) | 0xf]))
        
        # Write image data (compressed)
        xors = [i for i in rng.read(datalen)]

        # Initialize LZW dictionary with initial codewords
        dictsize = 2<<((rng.read(1)[0]%3)+1)
        dict = dictsize*[None]
        dict[dictsize-1] = tuple(xors[:dictsize])
        dictnum = len(dict)-1
        lzwmincodelen = len(str(dictsize-1))
        dictindex = {tuple(xors[:x]):x-1 for x in range(lzwmincodelen,
                                                       lzwmincodelen+dictsize)}

        encbuf = []
        # Get first codeword from current buffer
        buf = list(rng.read(1))[0] & 0xff
        dictindexkey = tuple([buf])    
        while True:
            if not dictindexkey in dictindex:
                break

            # If this is a new codeword, get it from our buffer
            if not dictindexkey[-1] == dictsize - 1:
                buf |= rng.read(1)[0]&0xff
            
            # And add to dictionary
            dict.append(list(dictindexkey) + xors[(dictsize*buf)>>8])
            dictnum += 1
            dictindexkey = dictindex[dictindexkey]
            dictindex[tuple(dictindexkey+(buf,))]=dictnum-1
            # And append to encoded buffer
            encbuf.append(dictindexkey)
            del dictindexkey[-1]

        # We're done with encoding; compress buffer
        out.write(encbuf[0][0].to_bytes(1,'big'))
        encsize = 1 + len(encbuf) * (1 if len(encbuf)>0 else 0)
        while len(encbuf) > 1:
            # Encode next value as difference against previous
            prevval = encbuf[:-1][-(1 if len(encbuf) <= 6 else 3)][0]
            encbuf[-1] -= prevval
            # Output length of delta (base one)
            out.write(((prevval.bit_length()-1).bit_length()+1
                       ).to_bytes((prevval.bit_length()//8)+1,'big'))
            # Output deltas themselves
            out.write(sum(([int('{:0>8b}'.format(diff)[::-1],
                               base=2)]*(8 if prevval>=0 else 7)
                           for diff in encbuf[-1]),[]))
            encsize += sum((7 if val<0 else 8
                            for val in encbuf[-1]))
            del encbuf[-1]
            
        # Append null terminator to end of compressed stream
        out.write(b"\x00")
        encsize += 1

        # Flush remaining output
        out.flush()

        # Write subblock footer for this frame
        out.write(struct.pack('<B',
                              encsize.bit_length().to_bytes(
                                          1,"big").pop()))
        # Write delay time
        out.write(rng.read(2))
        # Write transparent color index for this frame (if any)
        out.write((rng.read(1) if rng.read(1)==b'\x01' else
                  b'\x00'))
        # Write terminator for this subblock
        out.write(b';')
        
