#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
    out: BinaryIO,
    rng: BinaryIO,
    width: int,
    height: int,
    color_depth: int = 8
):

    # We need to know how big the palette is. Since we have a lot of colors,
    # it's easier just to make one that covers them all and then write-over
    # some colors with another. This also makes it easy for us to figure out what
    # our two palettes are.
    num_colors = (2 ** color_depth)**3 - 1

    # Our color palette, where each entry starts with its index in the palette
    # followed by three RGB values.
    palette = bytearray()

    for i in range(num_colors):
        r = next(rng) % 256
        g = next(rng) % 256
        b = next(rng) % 256

        palette += bytes((i,)) + struct.pack('BBBB', r, g, b, 0)

    # Write out the actual image data. Each frame begins with its image descriptor
    # which tells us how large its local color table is as well as its size. The
    # first frame doesn't have any custom global color table so we set it to zero.
    # Afterwards, every frame has a new color table generated at random from the full
    # palette we calculated earlier.
    for y in range(height // 4):
        for x in range(width // 4):
            out.write(b'\x2c')

            # Generate a new local color table, but reuse colors between frames
            # to make sure they're actually different this time around.
            if x == 0 and y == 0:
                local_color_table_size = 0
            else:
                local_color_table_size = len(palette) // 4 - 1

            out.write(struct.pack('B', local_color_table_size))
            out.write(struct.pack('<HHBBBBB',
                0, 0, 0,
                0, 0, 0))

            out.write(bytes(local_color_table_size * 3))

            # Now write out an image descriptor for this frame. Note that we never do
            # interlacing here since we don't even bother calculating the offsets or anything
            # like that.
            out.write(struct.pack('<HHHHHBB',
                0, 0,
                x*4, y*4,
                width, height,
                False,
                0xBF))


    out.write(palette)
