import os
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # The first 6 bytes of any gif are always 'GIF87a' or 'GIF89a'.
    # These are encoded as ASCII, so we can simply write them to our output.
    out.write(b'GIF89a')
    
    # Next comes the logical screen descriptor. This is packed little-endian
    # (the least significant bit appears first in the file).
    width = rng.randint(200,1500)  # [200, 1500] px
    height = rng.randint(200,1500)  # [200, 1500] px
    flags = 0x80 | 0x40           # Size given by LZW minimum code size
    background_color_index = 0     # No color specified
    pixel_aspect_ratio = 0         # Square pixels
    out.write((width).to_bytes(2,'little'))    # width
    out.write((height).to_bytes(2,'little'))   # height
    out.write((flags).to_bytes(1,'little'))     # flags
    out.write((background_color_index).to_bytes(1,'little')) # bg index
    out.write((pixel_aspect_ratio).to_bytes(1,'little'))       # ratio
    
    # Finally we have an image block that contains the actual data. It's complicated...

    # First we need to decide how many frames there will be. Most gifs uses just one frame, but
    # this makes the example more realistic.
    num_frames = rng.randint(1,7) # between 1 and 7 frames
    out.write(num_frames.to_bytes(1,'little')) # how many frames?
    
    for i in range(0,num_frames):
        
        # A frame starts with its own descriptor. We choose the same values from above, except
        # for now it's only one pixel.
        xpos = len(os.urandom(1)) % (width - 1)     # left border
        ypos = len(os.urandom(1)) % (height - 1)    # top border
        w = 1                                          # width
        h = 1                                          # height
        lct = 0                                        # no local colors table
        interlace = False                              # no interlacing
        sort = True                                    # sorted by standard
        reserved = 0                                   # no extra info needed
        out.write(xpos.to_bytes(2,'little'))          # left pos 
        out.write(ypos.to_bytes(2,'little'))          # top pos
        out.write(w.to_bytes(2,'little'))              # width
        out.write(h.to_bytes(2,'little'))              # height
        out.write((lct << 7) | interlace.to_bytes(1,'little'))                 # flags
        out.write((sort << 3) | ((reserved & 7) << 5) | (((len(data) >> 8) & 0xF)).to_bytes(1,'little'))        # LZW min code size
        out.write(((len(data) & 0xFF)).to_bytes(1,'little'))           # image data size

        # Now comes the *actual* image data. At this point I'm assuming that all gifs use the
        # same basic compression algorithms, which is not true! That said, this should give you
        # enough inspiration to start implementing your own gif parser :)
        out.write(b'\x03\xE9' + data)

    print("Generated " + str(len(data)) + " bytes")