#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO,
        output_file: str,
        limit: bytes=b'',
        width: int=500,
        height: int=400):

    # Build a GIF animation header according to the specs from wikipedia : https://en.wikipedia.org/wiki/Graphics_Interchange_Format
    def build_animation_header():

        # Header Signature (6 chars)
        sig = 'GIF87a'.encode('ascii')
        
        # LogicalScreen Descriptor Block
        lsd = struct.pack('<H', 1 << 12 | width % 256) # Width
        lsd += struct.pack('<H', 1 << 12 | height % 256) # Height
        lsd += struct.pack('B', 0xF7 & ~1<<7) # Flags : Global color table present and 1 to 8 bits per pixel
        lsd += struct.pack('B', 0xFF) # Background Color Index
        lsd += struct.pack('B', 0x0) # Pixel Aspect Ratio
        
        # Graphics Control Extension Block
        gceb = struct.pack('B', 0x21) # Extension Introducer
        gceb += struct.pack('B', 0xF9) # Graphics Control Label
        gceb += struct.pack('B', 0x04) # Block Size
        gceb += struct.pack('B', 0x0) # Packed Fields - Reserved
        gceb += struct.pack('B', 0xFF) # Delay Time in 1/100 of seconds
        gceb += struct.pack('B', 0xFE) # Transparent Color Index
        gceb += struct.pack('B', 0x0) # Terminator
        
        # Image Descriptor block
        idb = struct.pack('<B', 0x2C) # Image Separator
        idb += struct.pack('<H', 1 << 12 | width % 256) # Image Left Position
        idb += struct.pack('<H', 1 << 12 | height % 256) # Image Top Position
        idb += struct.pack('<H', 1 << 12 | width % 256) # Image Width
        idb += struct.pack('<H', 1 << 12 | height % 256) # Image Height
        idb += struct.pack('B', 0x0) # Flags : Local color table present but no interlace or sort
        idb += struct.pack('B', 0xF7 & ~1<<7) # LZW Minimum Code Size
        
        return (sig, lsd, gceb, idb)


    (headerSignature, logicalScreenDescriptorBlock, graphicsControlExtensionBlock, imageDescriptorBlock) = \
         list(map((lambda x: struct.unpack('{}s'.format(len(x)), x)[0]), build_animation_header()))

    with open(output_file, "wb") as f:
        f.write(headerSignature) # Write the signature at the beginning of the file
        f.write(logicalScreenDescriptorBlock) # Write the logical screen descriptor
        f.write(graphicsControlExtensionBlock) # Write an empty graphic control extension
        f.write(imageDescriptorBlock) # Write an empty image descriptor

        # Write the image's local color palette
        for i in range(0, 2**(int.from_bytes(logicalScreenDescriptorBlock[12], byteorder="big"))) : 
            rgb = os.urandom(3)

            if not len(limit):
                pass
            elif rgb == limit:
                pass
            else: 
                continue
            
            f.write(rgb)
            
        while True: 
            data = rng.read(1) # Read one byte at a time until end of stream is reached
            
            # Check if we got the separator before the header
            if data == b'\x2C': 
                break
            
            f.write(data)