#!/usr/bin/env python3

import sys
import os
import struct
import shutil
import datetime
from pathlib import Path
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def main():
    # Open up our input and output files for reading / writing
    try:
        rnd = open("/dev/urandom", "rb")
        with open("input", 'wb') as f_in:
            with open("output", 'wb') as f_out:
                # Generate the header and add to output
                header = b'GIF89a'
                f_out.write(header)

                # Get the logical screen descriptor (LSD) field
                lsd = getbytes(rnd, 7)

                # Add LSD to output
                f_out.write(lsd)

                # Read global color table length from LSD
                gctlen = 2 ** (int((lsd[4] >> 5))) * 3

                # Add GCT to output if we have it
                if gctlen > 0:
                    gct = getbytes(rnd, gctlen)
                    f_out.write(gct)
                
                while True:

                    # Read next field byte for block type
                    fieldbyte = int(getbytes(rnd, 1)[0])
                
                    # Check if this is the last block
                    if ((fieldbyte & 128) == 0):
                      break;
                
                    # Get size of block
                    size = int(ord(getbytes(rnd, 1)[0]))

                    # Check if there's any image data in this block
                    hasimgdata = bool(((fieldbyte>>5)&3)>0) 

                    # Write fields back to stream
                    write_bytes(f_out, [fieldbyte | 2], [size])

                    # If there's no image data we're done here
                    if not hasimgdata:
                        continue

                    # Write the local color table if necessary
                    numcolors = size // 3 - 1
                    if numcolors > 0:
                        colortable = getbytes(rnd, numcolors * 3)
                        write_bytes(f_out, [], colortable)

                    # Write the actual image data
                    imgdata = getbytes(rnd, size-1)
                    write_bytes(f_out, [], imgdata)
        
    finally:
        del rnd

def getbytes(io, nBytes):
  buf = io.read(nBytes)
  if len(buf)!= nBytes:
    print("Not enough bytes!")
    exit(-1)

  return buf

def write_bytes(io, fields, contents):

  for contentChunk in chunks(contents, 255):
    chunkSize = len(contentChunk) + 1

    fieldbyte = []
    fieldbyte += [chunkSize << 1]
    fieldbyte += [(chunkSize >> 7)]

    io.write(bytes([fieldbyte[0]]))
    io.write(bytes([fieldbyte[1]]))

    if len(contentChunk) > 0:
      io.write(contentChunk)