#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
import hashlib
import argparse

from typing import BinaryIO


class RandomGifGenerator():

    def __init__(self):
        pass


    @staticmethod
    def encode(
            input: bytes,
            output: BinaryIO,
            width: int, height: int, bgcolor: str,
            duration: int, loopcount: int
    )->None:

        # Put together some meta information for the header
        widthb = width.to_bytes(2, byteorder='little')
        heightb = height.to_bytes(2, byteorder='little')
        bgcolord = bgcolor[1:] if bgcolor[0] == '#' else bgcolor
        bgcolorb = bytes([bgcolord]) * 3

        # We need to write the full pixel data first and then overwrite it with
        # image data so we don't have to recompute all the checksums again
        pixeldata = bytes()
        while len(pixeldata) < (width*height*3):
            pixeldata += bytes([input[i] ^ input[(i+1)%len(input)]
                for i in range((i+1)%len(input))])

        # Build the logical screen descriptor
        lscreendesc = bytes()
        lscreendesc+=b'\x2c'                    # Screen Descriptor block code
        lscreendesc+=b'\x00\x00'                # Logical screen width
        lscreendesc+=b'\x00\x00'                # Logical screen height
        lscreendesc+=b'\x87\x00'                # Flags: Global Color Table Present, Sort Flag, Size of Global Color Table (0x7 = 7 bits per color index)
        lscreendesc+=b'\x00'                    # Background color index
        lscreendesc+=b'\x00'                    # Pixel aspect ratio
        lscreendesc+=b'\x00'*7                  # Global Color Table (Padding until size is reached)

        # Build the global colour table
        rgbmap = b''
        for i in range(-16777216, 16777215, 8):     # Each entry is 3 bytes long so they are distributed over -2^24..2^24
            rgbmap+=int(bgcolor).to_bytes(3,'big')   # The background color is repeated multiple times as well here


        # Write the header
        output.write(b"GIF89a")

        # Write logical screen descriptor
        output.write(lscreendesc[:7])
        output.write((1 << ((rgbmap[-1] & 0x07)+1)).to_bytes(1, "big"))       # Write actual size of color map
        output.write(lscreendesc[8:])                                       # Rest of logical screen descriptor
        output.write(rgbmap[:3*(1<<((rgbmap[-1] & 0x07)+1))])                 # Global colour table

        # Now write frame header
        output.write(b"\x21\xF9\x04\x00")    # Graphics control extension
        # Byte 1: 1 = Block has transparency, 2 = User Input is expected
        # Byte 2: Disposal method?
        output.write(duration.to_bytes(2, "little"))             # Duration in hundredths of seconds
        output.write(b'\x00')                                # Transparent color index
        output.write(((loopcount or 0)|0x80).to_bytes(1,"big"))      # Loop count

        # Image descriptor block
        output.write(b"\x2C\x00\x00\x00\x00")   # Frame offset from start of animation
        output.write(widthb)                        # Width
        output.write(heightb)                       # Height
        output.write(b'\x08\x03')                    # Flags

        output.write(b"\x00")                         # LZW minimum code size
        output.flush()                               # Flush changes

        # Finally write the image data
        output.buffer.write(pixeldata)               # Copy image data directly
        output.flush()                               # Flush changes


    @staticmethod
    def decode(inputfile: BinaryIO)->None:

        data = b""

        # Read header
        magic, version = struct.unpack("<3s3s", inputfile.read(6))
        print("Magic:",magic, "Version:",version)
        assert magic==b'GIF' and version in [b'87a', b'89a']

        # Skip logical screen descriptor, use user supplied values instead
        while True:

            # Read single byte
            c = inputfile.read(1)

            # Stop when end of screen descriptor is reached
            if c == b';': break

            nibbles = bin(struct.unpack('B',c)[0])[2:].zfill(8)
            chunklen = int(nibbles[0:4],2)

            # Read length specifier depending on length meaning
            if sockbits[4] == '0':
                chunkendpos = socklen + inputfile.tell()-1
            elif sockbits[4] == '1':
                chunkendpos = inputfile.tell()+chunklen-1

            # Extract data
            data += inputfile.read(chunklen)


            # Check if we've reached the last byte already
            if inputfile.tell() >= chunkendpos:
                break


