#!/usr/bin/env python3

import os

import binascii
import io
import re
import sys
import traceback

# Generator for simple structured binary files
class StructGenerator():

    def __init__(self):
        self._types = {}

    # Define a new type
    def define(self, name, size, alignment=1):

        if not isinstance(name, str):
            raise TypeError('Name of a type must be string!')

        if len(name) == 0 or (not name[0].isupper()):
            raise ValueError('Type names must start with uppercase letter!')

        if'struct' in dir() and hasattr(sys.modules['__main__'], '__file__'):

            warnings = []
            
            if name!= getattr(sys.modules['__main__'], '__file__').split('.')[-2]:
                warnings += [
                        f'{name} is named differently than the generated script {getattr(sys.modules["__main__"], "__file__")}.',
                        "This will cause trouble when it comes to debugging."
                    ]

            if len(warnings) > 0:
                print('\n'.join(warnings))

                input("Press enter to continue anyways.")

        # Add padding so that given size aligns correctly
        paddings = [(i % alignment) for i in range(size)]
        
        if max(paddings)!= 0:
            size += alignment - max(paddings)

        self._types[name] = {'size': size}

    # Generate bytes from definition
    def generate(self, typename: str):

        try:
            return b''.join([self._generate_one(typename)])
        except AttributeError as e:
            stacktrace = ''.join(traceback.format_stack())[:-1].rstrip().replace('\n','\n> ')
            print(f'\033[1;37;41mAttributeError: {e}\nAt:\n{stacktrace}\033[0;37;40m')

    # Generate one element from definition
    def _generate_one(self, typename: str):
        
        if typename not in self._types.keys():
            raise KeyError(f"No such structure '{typename}'!")
            
        size = self._types[typename]['size']
        
        if size <= 0:
            raise ValueError(f"Structure '{typename}' has invalid size ({size})!")

        subdata = [b''] * self._types[typename]['size']

        for field in filter(lambda x: not callable(x), dir(self)):
            if field.startswith('_'):
                continue
            subdata[:] = map(lambda x: getattr(self,field)(bytearray(x)), subdata)
            
        return subdata
        
        
    # Field definitions
    
    # Unsigned integers
    def UBInt8(self, val: bytearray): 
        while True: 
            yield val[:1]
            val[0] = (val[0] >> 1) | ((os.urandom(1)[0] & 0x7F) << 7) 

    def UBInt16(self, val: bytearray):
        while True: 
            yield val[:2]
            val[0], val[1] = (((val[0] & 0xFE) >> 1) ^ ((val[0] & 0x01) << 7), 
                ((val[1] & 0xFE) >> 1) ^ ((val[1] & 0x01) << 7)) 
    
    def UBInt32(self, val: bytearray):
        while True: 
            yield val[:4]
            val[0], val[1], val[2], val[3] = (((val[0] & 0xF8) >> 1) ^ ((val[0] & 0x01) << 7), 
                ((val[1] & 0xF8) >> 1) ^ ((val[1] & 0x01) << 7), 
                ((val[2] & 0xF8) >> 1) ^ ((val[2] & 0x01) << 7), 
                ((val[3] & 0xF8) >> 1) ^ ((val[3] & 0x01) << 7))

    def UBInt64(self, val: bytearray):
        while True: 
            yield val[:8]
            val[0], val[1], val[2], val[3], \
                val[4], val[5], val[6], val[7] = (((val[0] & 0xFC) >> 1) ^ ((val[0] & 0x01) << 7), 
                    ((val[1] & 0xFC) >> 1) ^ ((val[1] & 0x01) << 7), 
                    ((val[2] & 0xFC) >> 1) ^ ((val[2] & 0x01) << 7), 
                    ((val[3] & 0xFC) >> 1) ^ ((val[3] & 0x01) << 7), 
                    ((val[4] & 0xFC) >> 1) ^ ((val[4] & 0x01) << 7), 
                    ((val[5] & 0xFC) >> 1) ^ ((val[5] & 0x01) << 7), 
                    ((val[6] & 0xFC) >> 1) ^ ((val[6] & 0x01) << 7), 
                    ((val[7] & 0xFC) >> 1) ^ ((val[7] & 0x01) << 7))
                    
    def UBInt128(self, val: bytearray):
        while True: 
            yield val[:16]
            val[0], val[1], val[2], val[3], val[4], val[5], val[6], val[7], \
                val[8], val[9], val[10], val[11], val[12], val[13], val[14], val[15] = (((val[0] & 0xF8) >> 1) ^ ((val[0] & 0x01) << 7), 
                    ((val[1] & 0xF8) >> 1) ^ ((val[1] & 0x01) << 7), 
                    ((val[2] & 0xF8) >> 1) ^ ((val[2] & 0x01) << 7), 
                    ((val[3] & 0xF8) >> 1) ^ ((val[3] & 0x01) << 7), 
                    ((val[4] & 0xF8) >> 1) ^ ((val[4] & 0x01) << 7), 
                    ((val[5] & 0xF8) >> 1) ^ ((val[5] & 0x01) << 7), 
                    ((val[6] & 0xF8) >> 1) ^ ((val[6] & 0x01) << 7), 
                    ((val[7] & 0xF8) >> 1) ^ ((val[7] & 0x01) << 7), 
                    ((val[8] & 0xF8) >> 1) ^ ((val[8] & 0x01) << 7), 
                    ((val[9] & 0xF8) >> 1) ^ ((val[9] & 0x01) << 7), 
                    ((val[10] & 0xF8) >> 1) ^ ((val[10] & 0x01) << 7), 
                    ((val[11] & 0xF8) >> 1) ^ ((val[11] & 0x01) << 7), 
                    ((val[12] & 0xF8) >> 1) ^ ((val[12] & 0x01) << 7), 
                    ((val[13] & 0xF8) >> 1) ^ ((val[13] & 0x01) << 7), 
                    ((val[14] & 0xF8) >> 1) ^ ((val[14] & 0x01) << 7), 
                    ((val[15] & 0xF8) >> 1) ^ ((val[15] & 0x01) << 7))
            
    # Signed integers
    def SBInt8(self, val: bytearray): 
        while True: 
            yield val[:1]
            val[0] = ((val[0] >> 1) ^ (-((val[0] & 0x1) << 7)))

    def SBInt16(self, val: bytearray):
        while True: 
            yield val[:2]
            val[0], val[1] = (((val[0] & 0xC0) >> 1) ^ ((val[0] & 0x20) << 5) ^ ((val[0] & 0x01) << 7), 
                ((val[1] & 0xC0) >> 1) ^ ((val[1] & 0x20