#!/usr/bin/env python3

import io

from PIL import Image, ImageDraw  # type: ignore[import]

from typing import BinaryIO


# Generates a random GIF file into `out` using the random number generator `rng` (/dev/urandom).
# The image is generated with a random background color and has randomly placed pixels of that same
# color. Each frame should have one pixel more than the previous (e.g., first frame has at least two
# pixels of its chosen color in it; second frame has three pixels of its color etc.).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width: int=200, height: int=150):

    def make_image():
        img = Image.new('RGB', size=(width, height))

        bg_color = tuple([int(x * 256 / 255) for x in list(rng.read(3))])
        draw = ImageDraw.Draw(img)

        num_pixels = max((width - 4) * (height - 4), 1)
        for _ in range(num_pixels):

            pos = [int(i * 256 / 255) for i in list(rng.read(2))]
            draw.point((*pos, bg_color))

        return img


    def make_frame():
        img = make_image()
        buf = io.BytesIO()
        img.save(buf, format="gif")
        out.write(buf.getvalue())

    # Write GIF animation headers.
    out.write(b'GIF89a')
    loop_count = None
    duration_ms = 75

    head = bytes([(loop_count or 1)]) \
           + bytes([(duration_ms & 0xff00) >> 8]) \
           + bytes([duration_ms & 0xff]) \
           + bytes([width % 256]) \
           + bytes([width // 256]) \
           + bytes([height % 256]) \
           + bytes([height // 256])

    out.write(head)

    # Generate frames until we've written as many as there are degrees in a circle.
    while True:
        make_frame()