#!/usr/bin/env python3

import struct
from typing import BinaryIO


# A list containing all bytes with values between 0x21 and 0xff inclusive
valid_chars = [i for i in range(0x21, 0xff + 1)]

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int((lsd[4] >> 5))

    # Add GCT to output if we have it
    if gctlen >= 0:

      # Calculate how many entries will be needed
      gctneed = max([2**int(gctlen), 1])

      # We need the global color table, so read it in
      gctb = rng.read(3 * gctneed)

      # Add the actual global color table to output
      out.buffer.write(bytes([(gctneed << 5)]))
      out.buffer.write(gctb[:3 * gctneed])
      
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))


        # Calculate number of pixels in image
        numpix = abs(size)//3 - 1

        # Pick some colors!
        pix = []
        while len(pix)<numpix:
            pix += [
                valid_chars[ord(rng.read(1))] % 16,
                valid_chars[ord(rng.read(1))] % 16,
                ]

        # Add the local color table to output
        colortable = bytes(3 * len(pix))
        for i in range(len(pix)):
            colortable[[0, 1, 2][i]] |= pix[i]
        out.buffer.write(bytes(colortable))

        # Write the actual image data
        imgdata = rng.read(size-1)
        out.write(imgdata)