#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO

from util import mktemp


# Generates a random GIF file into `out`. The size of the generated
# file is determined by reading from `/dev/urandom`, and the content
# is determined by reading from `file_in` (if this option is not used,
# then the content will be all zeroes).
def generate_random_gif(
        file_in: str,
        rng: BinaryIO,
        out: BinaryIO,
        delay_factor: float = None,
        num_frames: int = None,
        palette_size: int = None
    ):

    # Read the input file to get image dimensions
    with open(file_in, mode='rb') as src:

        # Read the first part of the header
        header = src.read(6)
        if len(header)!= 6 or header[:6]!= b'\x47\x49\x46\x38\x39\x61':
            raise ValueError('Invalid GIF header!')
        
        # Skip the rest of the header
        src.seek(src.tell() + 5, 1)
        
        # Get width & height in bytes and unpack them
        wdim_b = src.read(2)
        hdim_b = src.read(2)
        width, height = struct.unpack('<HH', wdim_b + hdim_b)
    
    def make_colormap():
        colormap = [None] * 256 # initialize colormap
        for i in range(256):   # read colors from palette
            color_entry = rng.read(3)    # byte-encoded RGB triple
            colormap[i] = tuple([c//2 for c in color_entry])
        return colormap

    def make_framedata():
        '''Read image data from src'''
        framedata = []         # list of pixels per frame
        while True:
            block = src.read(width*height)     # get next chunk of pixel data
            if len(block) == 0: break          # if we're at EOF, stop

            encoded_pixels = []                 # decoded pixels as tuples
            for i in range(len(block)):
                x = i // height                  # row index
                y = i - x * height               # column index

                red, green, blue = block[i], block[i+width*height], block[i+(width*height)*2]
                
                if delay_factor:
                    yield (red, green, blue), ((y, x), (delay_factor * (num_frames)))
                else:
                    yield (red, green, blue), ((y, x))
            
    def make_palette(colormap):
        palette = []           # palette as uint8s
        for r,g,b in colormap:
            palette += [(r // 2**4 << 4) |
                        (g // 2**4 << 0) | 
                        (b // 2**4)]
        return palette

    # Generate a temporary file to hold the output palettized frames
    temppath = mktemp('.gif')
    with open(temppath, mode='wb') as tmpf, \
         open(file_in, mode='rb') as src:

        # Make the color map that maps "color indices" (single bytes)
        # to colors themselves
        cmap = make_colormap()

        # Figure out how many frames there will be. Note: this was copied verbatim
        # from https://github.com/phoboslab/jsmpeg-vnc/blob/master/tools/gifoverlay.py
        num_frames = num_frames or max(int((src.seek(0, os.SEEK_END)/width/height)), 1)
        num_frames = min(num_frames, 100)
        src.seek(0, os.SEEK_SET)

        # If no palette has been specified, use ours
        palette_size = palette_size or 256
        preamble = b''
        if palette_size < 256:      # only write color table if necessary
            assert palette_size > 0 and palette_size <= 256
            
            # Write global colour table to temporary file
            preamble += b','.join([struct.pack('BBB', *cmap[i]) \
                                   for i in range(palette_size)]) \
                          .ljust(3*(palette_size)+1, b'\x00')
        else:
            print(f'WARNING: Palette size {palette_size} is too large! Falling back to default value.')

        # Write individual frames to temporary file
        for f in range(num_frames):
            # Construct each frame's header, consisting of an LZW codesize
            # byte followed by a series of subblocks, which contains
            # information about the compression scheme used on the pixel data
            hdr = b';'                    # start of frame subblock
            hdr += struct.pack('B', 12)    # LZW minimum clear code size

            # construct the actual frame data, consisting of raw RGB triples
            frame = bytearray()
            for rgb, coord in make_framedata():
                frame += struct.pack('<BBB', *rgb)

            # encode the frame data using LZW
            lzwdata = preamble.decode('ascii').encode('ascii')
            lzwdata += hdr.decode('ascii').encode('ascii')
            compressed = ''.join(['%02X' % x for x in compress(frame, method=LZW)])
            lzwdata += ','.join(compressed[i:(i+255)]
                               ).lstrip(',').rstrip(',').encode('ascii')
            lzwdata += ';'.encode('ascii')

            # Add it to the temporary file before moving onto the next one
            tmpf.write(lzwdata)

    # Now copy the contents of the temporary file over to `out`
    with open(temppath, mode='rb') as tempfile:
        out.write(tempfile.read())
        
    # Clean up after ourselves
    os.remove(temppath)