#!/usr/bin/env python3

import struct
from typing import BinaryIO


# The first 6 bytes of any GIF file
HEADER_MAGIC = b'\x47\x49\x46\x38\x39\x61'

# Maximum length for each field in the GIFTags section
MAXLEN = {
        0x000b: 'extension',
        0xfffe: 'comment',
        0xffff: 'application'}

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
#
# Some maximum lengths are enforced to ensure that the
# output is valid (and not just humongous).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Output the header and the logical screen descriptor as they have no
    # side effects on the content inside the file and can be patched later.
    out.write(HEADER_MAGIC)     # Header magic
    out.write((2 << 5 | colordepth()).to_bytes(1, byteorder='big'))    # Logical Screen Descriptor block size
    out.write(width().to_bytes(2, byteorder='big'))                   # Screen width
    out.write(height().to_bytes(2, byteorder='big'))                  # Screen height
    out.write(0b10000000.to_bytes(1, byteorder='big'))                # Packed fields: Global Color Table flag set
    out.write(colordepth().to_bytes(1, byteorder='big'))               # Bits per pixel
    out.write(0x00.to_bytes(1, byteorder='big'))                       # Background color index
    out.write(0x00.to_bytes(1, byteorder='big'))                       # Pixel aspect ratio


    # Write a randomly generated global color table
    write_global_color_table(out)

    # Generate a palette with one entry per pixel. This has side-effects due to the 
    # size of the Image Data section being fixed by the logical screen descriptor.
    patchable_palette = [get_pixel_palette()
                         for _ in range(screenarea())]

    # Write an empty image data section, which will be filled up later
    out.write((0).to_bytes(1, byteorder='little'))        # Byte size of raw data
    out.write(((0).to_bytes(1, byteorder='little')))       # LZW minimum code size
    dummyimagedata = out.tell()                          # Store starting offset of image data
    out.seek(dummyimagedata + 1)                         # Seek past the size descriptor

    
    # Define what's left over from the logical screen descriptor (the actual content
    # of this part is irrelevant since we rewrite it entirely):
    # - The "global" color table
    # - All other tags present
    leftover_bytes = (1<<15)*2+1+1+1+1                     

    # Now comes the fun part where we actually fill up the GIF file with tags and 
    # their contents until there's nothing more to add
    
    while leftover_bytes > 0:
        
        # Choose the next tag type at random, but ensuring that Tag Type Null 
        # doesn't appear (this would break the decoding process)
        next_tagtype = get_next_valid_tagtype(leftover_bytes)

        # Enforce some maximum lengths for certain types of tags
        if next_tagtype == 0xfff9:
            maxlen = MAXLEN[next_tagtype][:-1]          # Remove the s from extension
        else:
            maxlen = MAXLEN[next_tagtype]

        # Create some fake contents for this tag
        fakecontents = create_fake_tagvalue(next_tagtype, maxlen)

        # Compute how many bytes the value occupies
        if   next_tagtype in {0x01ff, 0x02ff}:      # Plain textual values
            num_value_bytes = len(fakecontents)
        elif next_tagtype in {0xfe00, 0xfe01}:    # Application extensions
            appidsize = rng.randint(1, 8)           # Random application id size
            num_value_bytes = len(fakecontents)+appidsize
        else:                                       # Other tags
            num_value_bytes = len(fakecontents)


        # Now let's hack our way through the remaining bytes

        # First, pack the number of bytes that comprise the entire tag including its
        # tag type and actual size. We only want to account for the full length here
        # so we mask away all lower order bits
        num_total_bytes = ((num_value_bytes & 0x7f )
                           | (((num_value_bytes >> 7) & 0x7f) << 7)
                           | (((num_value_bytes >> 14) & 0x3f) << 14))
        
        # Secondly, determine how many bytes we still need to allocate for this tag
        # but don't count those that come before us (i.e. the one written above)
        num_remaining_bytes = leftover_bytes - num_total_bytes - 1

        # And finally, compute the actual length of the tag value, taking care of cases
        # like compressed images whose length measure needs an extra byte
        if   next_tagtype == 0x01ff:                 # Compressed image
            num_value_bytes = min(num_remaining_bytes, math.ceil(num_value_bytes / 126) + 1)
        elif next_tagtype == 0x02ff:                 # Plain textual value
            pass                                      # No special restrictions apply
        elif next_tagtype in {0xfe00, 0xfe01}:       # Application extensions
            num_value_bytes = min(num_remaining_bytes-appidsize,
                                  num_value_bytes)
        else:                                         # Other tags
            raise ValueError('Invalid tag type!')

        # Update the amount of bytes left untouched
        leftover_bytes -= num_total_bytes + num_value_bytes

        # Finally, write down the tag itself now knowing both the total length and the
        # actual length of the payload
        out.write(struct.pack('<HH',
                              next_tagtype,
                              num_total_bytes))
        out.write(struct.pack('{}B'.format(num_value_bytes), *fakecontents[:num_value_bytes]))


    # Fix up the size of the image data section
    imgdata_size = out.tell() - (dummyimagedata + 1)         # Current position minus start pos
    out.seek(dummyimagedata)                                 # Go back to beginning of image data
    out.write(imgdata_size.to_bytes(1,byteorder='little'))    # Patch the size indicator



def colordepth():
    return rng.getrandbits(3)             # Random value in [0,7]

def width():
    w = rng.randint(1, 0xffff)            # A width between 1 and 65535 pixels
    return w                              # Screens are limited to 65535x65535px

def height():
    h = rng.randint(1, 0xffff)            # As above
    return h 

def screenarea():
    return width()*height()

def numpixels():
    return pow(2, colordepth()+1)-1

def get_pixel_palette():
    r = rng.getrandbits(8)
    g = rng.getrandbits(8)
    b = rng.getrandbits(8)
    return (r,g,b)

def get_next_valid_tagtype(leftover_bytes):
    possible_types = [t for t in MAXLEN
                      if leftover_bytes >= MAXLEN[t]]
    assert len(possible_types)!= 0              # shouldn't happen because we're generating
                                                # GIF files that fit within the bounds we impose
    next_tagtype = rng.choice(possible_types)
    return next_tagtype

def create_fake_tagvalue(next_tagtype, maxlen):

    # These tags have a prefixed four-character ASCII string as their value
    prefixtags = {
            0x01ff: 'IDAT',
            0x02ff: 'COMM',
            0xfe00: 'NETSCAPE2.0',
            0xfe01: 'XMPDATA'}

    # If the currently chosen tag is a plain text value then we simply pick a random
    # binary string of a given maximum length
    if next_tagtype in {0x01ff, 0x02ff}:      
        val = rng.randbytes(rng.randint(1,maxlen))
        
    # Or else, if the currently chosen tag is one of the prefixed ones then we generate
    # a pseudo-random value consisting of the correct length plus that prefix encoded as
    # ASCII
    elif next_tagtype in prefixtags:            
            val = prefixtags[next_tagtype].encode(encoding='ascii') + \
                  rng.randbytes(rng.randint(len(prefixtags[next_tagtype])+1,
                                           maxlen