#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # First byte is always 'G'
    for i in range(7):
        out.write(b'\x46')

    # Next two bytes are the logical screen width and height (ie 64 x 48)
    out.write(struct.pack('<HH', *generate_dimensions(rng)))

    globalcolortablesize = ord(rng.read(1)) & 0xFF
    localcolortablesize  = ord(rng.read(1)) & 0xFF

    print("global colortable size:", globalcolortablesize)
    print("local colortable size: ", localcolortablesize)

    # Global color table
    for _ in range(2 << globalcolortablesize):
        out.write(struct.pack('BBB',
                *(ord(rng.read(1)) & 0xFF for _ in range(3))))

    rowcount = int.from_bytes(
            rng.read(2), 'big', signed=False
        ) % 512 + 1 # <= 512 rows

    # Loop count
    loopcount = ord(rng.read(1)) & 0xFF
    if loopcount == 0:
        loopcount += int.from_bytes(
                rng.read(2), 'big', signed=False
            ) + 1 # 0 means infinite loop

    # Image descriptor
    left   = 0 
    top    = 0
    width  = 64 - 1 # Force to something small
    height = 48 - 1 # Force to something small
    packed = ((left & 0xFF) |
              (((top >> 8) & 0xF) << 4) | 
              (((width >> 8) & 0xF) << 8) | 
              (((height >> 8) & 0xF) << 12)
             )

    lctflag = False
    interlaceflag = bool(int.from_bytes(rng.read(1), "little"))
    sortflag      = bool(int.from_bytes(rng.read(1), "little"))
    reserved      = bool(int.from_bytes(rng.read(1), "little"))
    lctsize       = ord(rng.read(1)) & 0xFF
    lctpresent    = [bool(int.from_bytes(rng.read(1), "little"))
                     for _ in range(lctsize)]
    
    print("interlace flag", interlaceflag)
    print("sortflag     ", sortflag)
    print("reserved     ", reserved)
    print("lct present  ", lctpresent)

    imagedata = []
    while len(imagedata) < rowcount*width:
        charcode = int.from_bytes(rng.read(1), "little")

        if charcode == 0x00:
            break
        elif charcode > 0xFE:
            continue

        if not lctpresent[charcode]:
            continue
        
        output = chr(0x00).encode()

        # LZW minimum code size
        minbits = ord(rng.read(1)) & 0xFF

        inputstring = None
        if charcode < 0xE0:
            inputlength = (charcode & 0x1F) + 1
            inputstring = chr((charcode&0xE0)<<3) + rng.read(inputlength-1)

            assert set(inputstring) >= {chr(c) for c in range(0x20)}

            output += inputstring
        else:
            inputstring = ''
            while True:
                inputval = int.from_bytes(
                        rng.read(1), "little"
                    ) & 0xFFFF

                if inputval == 0xFFFE or inputval == 0xFFFF:
                    break
                
                inputstring += str(inputval).zfill(minbits)

        print("char code:", hex(charcode))
        print("LZW min bit size:", minbits)
        print("input string length:", len(inputstring))
        
        dictionary = {}
        dictindex = 2**minbits
        
        # Add special markers at beginning of inputstream
        dictionary['<EOL>'] = '</>'
        dictionary['<EOI>'] = '<EOI>'
        index = 0
        for c in "<>":
            dictionary[c] = f"<{str(index).zfill(minbits)}>"
            index += 1

        index = 0
        while index!= len(inputstring):
            firstletter = ""
            
            # Check for match with our index already added
            try:
                compressedindex = dictionary[inputstring[firstletter]]
                assert '<' in compressedindex
            except KeyError:
                pass
            else:
                secondletter = inputstring[index+1 : index+2]
                if len(secondletter):
                    secondletter = secondletter[0]

                    try:
                        compressedindex = dictionary[
                                compressedindex + secondletter
                            ]
                        if '<' in compressedindex:
                            raise KeyError()

                        index += 2
                    except KeyError:
                        pass
        
            # If no match found then add it normally 
            if not len(compressedindex):
                dictionary[inputstring[firstletter]] = (
                        index.to_bytes(minbits//8, "big").hex().upper()
                    )
                compressedindex = (
                        compressedindex[: -(len(secondletter)+1)] 
                        + secondletter
                    ) 
                index += 1

            imagedata.append(compressedindex)
        
        print("output length:", len(output))

    for n in sorted([int(i, base=16) for i in list(set(imagedata))]):
        print(n,"=>",dictionary[n])
    
    assert len(imagedata)*8 % 2 == 0


    # Write out image data
    out.write(((packed).to_bytes(2,'little')))
    out.write((((rowcount).to_bytes(2,'little'))))
    if interlaceflag:
        out.write((((0).to_bytes(2,'little'))))
    else:
        out.write((((loopcount).to_bytes(2,'little'))))

    print("Image descriptor:")    
    print("\tLeft:",left,"\tTop:",top,"\tWidth:",width,"\tHeight:",height)
    print("\tpacked:",hex(packed))

    for v in filter(lambda x: type(x)==str, imagedata):
        if v == '<EOL>':
            out.write(int.from_bytes(rng.read(1),'little').to_bytes(1,'little'))
        elif v == '<EOI>':
            out.write(int.from_bytes(rng.read(1),'little').to_bytes(1,'little'))
        else:
            out.write((((0x80 | int(v,base=16)).to_bytes(1,'little'))))

# We can only have dimensions which are multiples of 2
def generate_dimensions(rng):
    return tuple(sorted({
        2 ** int.from_bytes(rng.read(1),"big"), 
        2 ** int.from_bytes(rng.read(1),"big")}
                        , reverse=True))
