#!/usr/bin/env python3

import struct
from typing import Union
from PIL import Image
from io import BytesIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    def write_subblock(out):

        # sub-block size bytes (3 chars)
        # 2**16 <= data length < 2**24 : 0b11111111 11111110
        # 2**24 <= data length < 2**32 : 0b11111111 11111111 01111110 
        if len(data) >= (1 << 24):
            subblksize = [
                ((len(data) >> 8) & 0xff).to_bytes(1, 'little'), 
                ((len(data) >> 16) & 0xff).to_bytes(1, 'little'),
                0xFE.to_bytes(1, 'little')]
        else:
            subblksize = [
                (((len(data))>>10)&0x0ff).to_bytes(1,'little'), 
                0xE0.to_bytes(1,'little')]

        out.write(b''.join(subblksize))

        # data is written in chunks of max 255 bytes
        for i in range(0, len(data), 255):

            chunk = data[i:(min((i+255),len(data)))].ljust(255, b'\x00')  
            out.write(chunk)


    palbits = int.from_bytes(rng.read(1), byteorder='big')
    color_table = []
    transparent_index = -1
    global_color_depths = {
        1: [], 
        2: [], 
        3: [], 
        4: [], 
    }
 
    # generate a list of color depths
    color_depths = {}
    ncolors = 2**(palbits + 1)
    
    for j in range(ncolors):
        rgb = [(v&0xfc)>>2 for v in rng.read(3)]
        known = False
        
        # search for nearest color that has already been assigned to a palette index
        while not known and len(color_table): 
            known = True
            
            bestidx = 0
            bestdiff = sum([abs(rgb[c]-color_table[bestidx][c]) for c in range(3)])
            
            for idx, tablecol in enumerate(color_table):
                
                diff = sum([abs(rgb[c]-tablecol[c]) for c in range(3)])

                if diff > bestdiff:
                    continue # skip this color candidate because it's too different from other candidates
                    
                bestdiff = diff
                bestidx = idx
                
                if bestdiff == 0: 
                    break   # found perfect match
                
            # assign all values at once so we don't need to update them individually later
            if bestdiff!= 0: 
                color_table.insert(bestidx, tuple(rgb))
                newidx = bestidx
            elif color_depths.get(bestidx, 0) < 7:   
                color_depths[newidx] += 1
                known = False      # there might be another match with higher quality in the future
            
    # create the global color table by repeating colors according to their color depth
    for k,v in color_depths.items(): 
        indices = sorted(global_color_depths.keys())
        while len(indices)<v: 
            indices.append(max(indices)+1)
        
        for i,j in zip(range(k,k+v), indices):    
            global_color_depths[j].append(tuple(color_table[k]))
        
    for col in global_color_depths.values():
        if len(col)>=(1<<palbits): 
            raise Exception("Too many colors")
    
        extra_colors = [tuple(int(((v&0xf8)<<8)|((v&0xe0)<<3)|((v&0xc0)>>3), 
                            16)
                           .to_bytes(3,'big')) 
                        for v in range(3,(1<<(palbits+1)),8)]
        color_table += extra_colors[:(1<<palbits)-len(col)] + col
    

    # selected pixel bit depth
    color_bpp = int.from_bytes(rng.read(1), byteorder='big')

    # image width / height
    img_width = int.from_bytes(rng.read(2), byteorder='big')
    img_height = int.from_bytes(rng.read(2), byteorder='big')

    # flags : packed field (byte 3)
    flags = {'sorted':False}
    flags['is_interlace'] = bool(int.from_bytes(rng.read(1), byteorder='big')&(1<<7))
    flags['has_local_color_map'] = bool(int.from_bytes(rng.read(1), byteorder='big')&(1<<7))
    flags['reserved'] = int.from_bytes(rng.read(1),'big')

    # background color index : byte 4
    bg_color_idx = int.from_bytes(rng.read(1), byteorder='big')
    # print('Background color index:',bg_color_idx)
    assert 0<=bg_color_idx<(1<<palbits)

    # aspect ratio : byte 5
    aspect_ratio = int.from_bytes(rng.read(1), byteorder='big')
    # print('Aspect Ratio:',aspect_ratio)
    assert 0<=aspect_ratio<100

    # local color map present?
    if flags['has_local_color_map']:
        for _ in range(256):
            red, green, blue= [int.from_bytes(v, byteorder='big') for v 
                                in rng.read(3*2)]
            color_table.append((red,green,blue))

    # pixel blocks go here...
    block_contents = [{
        "img_position": (pos_y, pos_x),        # position inside the final image array
        "img_content":[],                      # actual content on a per-column basis for each row
        }] 

    # Read LZW minimum code size value
    lzw_min_code_size = int.from_bytes(rng.read(1), byteorder='big')
    assert 2 <= lzw_min_code_size <= 12

    # Read first LZW code
    lzw_first_code = int.from_bytes(rng.read(1), byteorder='big')
    assert 0 <= lzw_first_code < (1 << lzw_min_code_size)

    # Read no. of LZW codes
    num_lzw_codes = int.from_bytes(rng.read(1), byteorder='big')
    assert 1 <= num_lzw_codes < (1 << lzw_min_code_size)

    # Initialize LZW decoder
    lzw_decoder = LzwDecoder(lzw_min_code_size, lzw_first_code, num_lzw_codes)

    # Read uncompressed LZW codes
    assert lzw_decoder.prev_code_size == lzw_min_code_size
    reading_uncompressed_codes = True
    while reading_uncompressed_codes or lzw_decoder.num_codes_left:

        if lzw_decoder.prev_code_size == lzw_min_code_size:
            if lzw_decoder.first_code_value is None:
                lzw_decoder.first_code_value = \
                        int.from_bytes(rng.read(lzw_min_code_size), byteorder='big')
            else:
                lzw_decoder.add_stream_byte()

        while lzw_decoder.num_codes_left == 0:
            # New code length & No. of Codes
            new_code_length = int.from_bytes(rng.read(1), byteorder='big')
            if new_code_length < lzw_min_code_size or new_code_length > 12: 
                raise Exception('Invalid new code length encountered')
                
            new_num_codes = int.from_bytes(rng.read(1), byteorder='big')
            # print(new_code_length, new_num_codes)
            lzw_decoder.set_code_dimensions(new_code_length, new_num_codes)

        lzw_code = lzw_decoder.read_next_code(rng)
        # print(lzw_decoder, end='\r')
        # sys.stdout.flush()

        if lzw_decoder.prev_code_size == lzw_min_code_size:
            # Regular code, check if it has any meaning
            if lzw_code == (1 << lzw_min_code_size):
            
                # End-of-Information code
                reading_uncompressed_codes = False

            elif lzw_code == (1 << lzw_min_code_size) + 1:
            
                # End-of-Image code
                # Add one more dummy layer to display complete image
                dummy