#!/usr/bin/env python3
# Generates a random gif file with the same format as this one
# from the RNG. The only difference being it's randomly filled
# frame data.
import struct
from typing import BinaryIO


def make_framedata(rng: BinaryIO):

    # Generate a block containing ints between 0 and 255
    def genintblock():
        while True:
            yield rng.randint(0, 255)

    return bytes((next(genintblock()) for i in range(10)))



def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Write the header
    out.write(b'GIF89a\x00\x00\x00\xff\xf9')
    
    # Write the logical screen descriptor
    width = rng.randrange(64, 1024+1, step=8)
    height = rng.randrange(64, 1024+1, step=8)
    flags = (width & 0b11111111) | ((height >> 8) << 7)
    out.write(struct.pack('>HHBBB', width, height,
                         flags, 0x00, 0x00))
    print(f'{width},{height}')

    # Write the global color table
    # Each byte represents a single color value
    # No need to worry about the two reserved bytes
    colortable = [rng.randint(0, 255) for _ in range(3*256)]
    out.write(bytes(colortable[:]))

    # Write the first graphic control extension
    # We always write 0x00 on any reserved fields
    out.write(b'\x21\xf9\x04\x00\x00\x00\x00\x00\x00')

    # Write application extension
    appextlen = 11
    out.write(b'\x21\xff\x0b' + f"{appextlen:02X}".encode()
              + 'NETSCAPE2.0'.encode("ascii") + '\x03'
              + b'\x01\x00\x00')

    # Write the comment extension
    commlen = rng.randrange(1, 32)
    out.write(b"\x21\xfe" + f"{commlen:02X}".encode() +
                rng.read(commlen - 1))

    # Frame info
    delaytime = rng.randrange(1, 65536)
    disposemethod = rng.randrange(0, 4)
    transindex = rng.randrange(0, 256)
    out.write(b'\x21\xf9\x04' + f"{delaytime:04X}{disposemethod}"
             .encode() + f"{transindex:02X}".encode() + b'\x00')

    # Image descriptor
    left = rng.randrange(0, width//8)
    top = rng.randrange(0, height//8)
    width = rng.randrange(1, min(left+9, width-left)+1)
    height = rng.randrange(1, min(top+9, height-top)+1)
    flags = 0b10000000 | (rng.randrange(0, 4) << 2)
    out.write(struct.pack(">HHBB", left, top, width, height))

    # Image data
    datlen = rng.randrange(1, 128)
    out.write(make_framedata(rng) * ((datlen+255)//256))