#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
  def _write_block(typ: int, data: bytes) -> None:
    """Writes an arbitrary GIF block"""
    out.write(struct.pack('=BB', typ, len(data)))
    out.write(data)

  # First write the header
  _write_block(b'GIF89a'[0], b'')

  # Now get the logical screen descriptor (LSD) field
  lsd = rng.read(7)

  # Add LSD to output
  _write_block(b'\x2c', lsd)
  
  # Read global color table length from LSD
  gctlen = 2**((lsd[4] >> 5)) * 3

  # Add GCT to output if we have it
  if gctlen > 0:
    _write_block(b'\x21\xf9\x04', bytes([
        0x08, 0x00,    # sub-block size
        0x0b,          # flags
        gctlen        # palette size
    ]) + rng.read(gctlen))

  while True:

    # Read next field byte for block type
    fieldbyte = ord(rng.read(1))
    
    # Check if this is the last block
    if ((fieldbyte & 128) == 0):
      break
  
    # Get size of block
    size = ord(rng.read(1))

    # Check if there's any image data in this block
    hasimgdata = bool(((fieldbyte>>5)&3)>0) 

    # Write fields back to stream
    _write_block(ord(fieldbyte | 2), bytes([fieldbyte&64|size]))

    # If there's no image data we're done here
    if not hasimgdata:
      continue

    # Write the local color table if necessary
    numcolors = size//3 - 1
    if numcolors > 0:
      colortable = rng.read(numcolors*3)
      _write_block(b'\x21\xff', bytes([
          0x09,           # block size
          0x20,           # app extension code
          numcolors       # colors
      ] + list(colortable)))

    # Write the actual image data
    imgdata = rng.read(size-1)
    _write_block(b'\x21\xff', bytes([
      0x08,      # block size
      0x21,      # graphics control
      0x00,      # delay time
      0x00,      # transparancy index
      0xff,      # block terminator
      0x0b,      # left position
      (rng.read(1)[0])%256,   # top position
      (rng.read(1)[0]&0xfe)|0x01,# width (low bit set)
      (rng.read(1)[0]),         # height
      *(rng.read(min(size-1, 255))), # image data
      *[0]*(255-(size-1))         # padding
  ]))