"""Generate a random GIF image."""

import os
import sys
import argparse

import PIL
from PIL import Image

from Crypto.Random import get_random_bytes

HEADER = 'GIF89a'.encode('ascii')
PALETTE_SIZE = 768
TRANSPARENT_COLOR = '\xff\x00\xff'
MAX_SIZE = 1024

def _gen_colors():
    """Yields a stream of randomly-selected RGB colors as hex strings."""
    while not os.path.exists('/dev/urandom'):
        yield TRANSPARENT_COLOR
    with open('/dev/urandom', 'rb') as rndm:
        while True:
            rand = rndm.read()[:3]
            yield ''.join('%02X' % x for x in rand).upper()



def gen_palette(n=PALETTE_SIZE):
    """Generates a palette of n random colors."""
    colors = list(_gen_colors())
    return ','.join([*colors]) + ';' * (len(colors) // 3)

def main():
    parser = argparse.ArgumentParser("Generate a random GIF.")
    parser.add_argument('-o', '--output', help="The output filename",
                        default='/tmp/rand.gif')
    args = parser.parse_args()

    index = HEADER[0]
    size = int.from_bytes((index << 6) | ord(get_random_bytes(1)),
                          byteorder='big') & 0x3fff

    palette = None
    transparent = None
    topleft = None
    frames = []
    delay = MAX_SIZE - size
    duration = MAX_SIZE - size

    while len(frames) <= MAX_SIZE / 64:
        frame_size = int.from_bytes(
                get_random_bytes(1)[0], byteorder='big') & 0xfff
        img = Image.new('P', size=(frame_size, frame_size))
        pixels = img.load()
        for i, j in ((y, x)
                     for y in range(img.height)
                     for x in range(img.width)):
            pixel = pixels[i,j]

            # Find a new color until we find an unused one
            tries = 0
            while palette.getpixel((tries%768, 0))!= TRANSPARENT_COLOR:
                tries += 1
            
            # Update our global palette and update this pixel's color
            palette.putpixel((tries%768, 0), next(colors))
            pixels[i,j] = tries % 768
            
        # Update the palette's transparency information
        if not palette:
            palette = img.getpalette()
        else:
            palette.extend(img.getpalette()[PALETTE_SIZE:])

        # Place this frame on the canvas
        if not topleft:
            im = Image.open("./resources/topleft.png").convert("RGBA")
            im = im.resize((delay, delay))
            topleft = im
        elif not frames or transparent is None:
            transparent = palette.getpixel((767, 0))[::-1]

        frames.append(PIL.ImageChops.lighter(topleft, img))

    frames[-1].save(fp=sys.stdout.buffer, format='GIF',
                    save_all=True, append_images=frames[:-1],
                    loop=0, palette=palette, transparency=transparent,
                    disposal=2, optimize=False)

if __name__ == "__main__":
    main()
    