#!/usr/bin/env python3

import struct
import secrets
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The GIF's dimensions are determined by `width` and `height`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width: int, height: int):

    # Dimensions must be positive integers between 0 and 256 inclusive.
    if (not isinstance(width, int) or not isinstance(height, int)):
        raise ValueError('Width and Height parameters must be of type integer.')
    elif width < 1 or width > 256 or \
         height < 1 or height > 256:

        raise ValueError('Illegal value for Width or Height parameter.'
                         'Must be an integer between 1 and 256 inclusively.')

    secret_gen = secrets.token_bytes(
            4 * ((width * height) // 8))
    secret = []
    idx = 0
    while idx < len(secret_gen):
        secret += [struct.unpack("=I",
                                 secret_gen[idx:idx+4])[0]]
        idx += 4

    # Write the fixed parts of the header.
    out.write(b'GIF89a')
    # Write the logical screen descriptor.
    out.write(struct.pack('>HHBBBB',
                           width, height,
                           0xf7,  # Global Color Table Flag: True
                           None, 0x00,  # Background color index: None
                           None   # Pixel aspect ratio: None
                          ))

    # Write the global color table
    out.write(struct.pack(">BBB", 0xff, 0x00, 0x00) * 256)

    # Write the local image header.
    out.write(struct.pack('>BHBHHBB',
                          0x2c,  # Image separator code
                          None, # Left position is not specified to avoid
                                # errors when displaying in some browsers.
                          0,    # Top position is set later.
                          width, # Image width.
                          height,# Image height.
                          0x08,  # Local color table flag: True
                          0      # LZW minimum size.
                         )
              )


    # We will now fill the rest of the frame with pixels from our randomly
    # generated pseudo-binary data. Each bit is one pixel. We set this up so we
    # can write 4 bits at a time without introducing any new bytes until we hit
    # the end of each byte. This means we need to do some special stuff when
    # reading or writing lsb first.

    bitpos = -1;
    bytenum = 0;

    # Start by skipping forward enough that we're on a byte boundary.
    while (bitpos!= 7):
        next_byte = out.read(1)
        bytenum = ord(next_byte)
        bitpos = 7
        while (bytenum & 1 << bitpos == 0):
           bitpos -= 1

    while (True):
       # Set the current top row.
       out.seek(-len(secret_gen)+ \
                     (bytenum * 8) + \
                      bitpos//8
                   )

       # For each pixel in the row...
       for i in range(width):

           # Read 2 more bytes from the input stream.
           next_byte = out.read(1);
           next_byte = out.read(1)
           next_byte = out.read(1)
           next_byte = out.read(1)

           # Extract the appropriate bit from these bytes.
           bitnum = 7 - bitpos
           mask = 1 << bitnum
           if(ord(next_byte) & mask):
               # The bit was high, draw black.
               out.write(b'\x00\x00')
           else:
               # The bit was low, draw white.
               out.write(b'\xff\xff');

           # If we've just written all the bits in the byte, move onto the
           # next byte.
           if (bitpos == 0):
               # Move back up two bytes since we had read four extra bytes.
               out.seek(-2, 1)

               next_byte = out.read(1);
               next_byte = out.read(1)

               # Update the byte number to reflect the change.
               bytenum += 1
               bitpos = 7

           # Decrement the bit position, and if it goes negative incrementing
           # the byte position indicates moving to the previous byte.
           bitpos -= 1
           if (bitpos < 0):
               bitpos = 7
           bytenum += 1

  