#!/usr/bin/env python3

import struct
from typing import BinaryIO, Union

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width=2**10-5, height=2**10-7,
        colors=2**6, bgcolor=2**0):

    # Check if we have enough bytes to generate this image
    if (width * height) > ((2 ** 32) - 59): raise ValueError("Image too large") 

    def write_word(value: int, wordsize: int = None):
        if not isinstance(value, int):     value = ord(value)
        if   wordsize == 2:                 return value.to_bytes(2, 'little', signed=False)
        elif wordsize == 4:                 return value.to_bytes(4, 'little', signed=False)
        else:                               return value.to_bytes((value.bit_length() + 7) // 8, 'little', signed=False)

    # Header
    out.write('GIF89a'.encode())
    
    # Logical screen descriptor
    logical_screen = struct.pack('<HHBBBH', width, height,
                                bits_per_pixel=(colors*bits_per_channel).to_bytes(1, 'little'), 
                                background=bgcolor, show_grid=b'\x00', packed=b'\x00')
    out.write(logical_screen)

    # Global color table
    for i in range(colors):
        color_table += struct.pack("<BBB", *(rng.randint(0, 255)
                                             for _ in range(3)))

    global_color_table = struct.pack("<B", len(color_table))
    global_color_table += color_table

    # Extension block
    ext_label = b"\xF9"                       # Application extension
    ext_size  = b"\x04"                       # Size of application extension data
    ext_app   = b"\xFF\xFE"                   # Application identifier
    ext_ajvws = [b"", b""]                    # Authentication code
    if rng.randrange(2):                     # Color resolution flag
        cr_size    = b"\x04"                   # Size of color resolution sub-block
        cr         = struct.pack("<BBBB", *[rng.randint(1, 8)
                                                for _ in range(4)])
    else:                                   # No color resolution
        cr_size    = b"\x01"                   # Size but no sub-blocks
        cr         = b""                      # empty string
    ext_data  = ext_app + cr + ext_ajvws[0]   # Combine all sub-blocks
    ext       = ext_label + ext_size          # Create label and size only
    ext      += ext_data                     # Add sub-blocks
    block_len = len(ext)                     # Block is always this long
    out.write(struct.pack("<BBH", block_len >> 8,
                            block_len & 255, 0xf9))
    out.write(ext)                           # Write all blocks from here on
    out.write(global_color_table)            # Write the color table
 
    # Image descriptor
    image_descriptor = struct.pack("<HHBBBBBHB",
                                   left=0, top=0,
                                   width=width, height=height,
                                   local_map_flag=1, interlace_flag=0, sort_flag=0,
                                   reserved=0, packed=0)
    out.write(image_descriptor)              # Write the image descriptor

    # Table based encoder
    # Get the pixel generator
    def get_pixels(): 
        while True: 
            yield tuple([rng.randint(0, 255)] * channels)
    pixels = iter(get_pixels())

    local_color_depth = dict(enumerate(("00","01","10"), start=1))
    lzw_mincode_size = 8

    buffer = []                          # Buffer with previous tokens
    prev   = next(pixels)                # The first token
    mincode = maxcode = 2 << lzw_mincode_size
    clear   = struct.pack(">H", 2 << lzw_mincode_size)
    out.write(clear)                     # Prepend the clear command

    last = len(local_color_depth)-1
    count = 0
    for channel in zip(*pixels):
        for c in reversed(channel[:-1]):
            buffer.append(c)

        val = buffer[last:] + list(channel[-1:])

        bitfield = "".join("-" if x==y else "+" for x, y in zip(prev, val))
        flush = False
        while len(bitfield) >= prefix_mask+2:
            # Check if we should "flush" anyway
            if flush or trailer_mask <= bitfield < trailer_mincode:
                out.write(clear)
                flush = True

            # We can fit one more token
            if maxcode < MAXCODE:
                maxcode <<= 1
                out.write(struct.pack(">H", (prefix_mask<<5) |
                                          suffix_mask | (maxcode - MINCODE)))

                print(f"\n{count}: {hex(val)} [{len(buffer)+1}]",
                      f"({len(bitfield)}-{prefix_mask})",
                      f"{hex(maxcode)}, {hex(MINCODE)}:", 
                      f"[{bitfield}]==>", end="")

                offset = lzw_mincode_size + \
                         (maxcode - MINCODE)*lzw_mincode_size
                prev = tuple(out.getbuffer()[offset:offset+lzw_mincode_size])
                
                print(f"{hex(prev)}\n", end='')

                maxcode += 1
                continue
            
            # We need to create a new table entry
            if mincode < MAXCODE:
                mincode <<= 1
                lzw_mincode_size += 1
                out.write(struct.pack(">HH",
                                      (prefix_mask << 5) |
                                      suffix_mask |
                                      (MAXCODE - MINCODE),
                                      0xffff ^ (MAXCODE - MINCODE)))

                print(f"\n{count}: {hex(val)} ({len(bitfield)})",
                      f"=> {hex(newcode)} => {[hex(x) for x in sorted(table)]}")

                MAXCODE *= 2
                MINCODE *= 2
                local_color_depth[lzw_mincode_size] = (
                    local_color_depth[lzw_mincode_size-1][:-1]+str(
                        local_color_depth[lzw_mincode_size-1][-1])+str(1)).zfill(3)

                MINCODES[lzw_mincode_size] = mincode
                MAXCODES[lzw_mincode_size] = maxcode

                prefix_mask >>= 1
                suffix_mask += 1
                continue

            # Can't encode any more tokens
            break
        
        # If we are within the block and there's room
        if len(bitfield) < prefix_mask+2:
            buffer = val

        prev = val

        if len(buffer) == last+1:        # Flush when there is exactly 1 token
            out.write(clear)             # Reset stream
            flush = True

        count += 1

    if flush == True:
        out.write(clear)                 # Trailing clear-command 
