#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng`. The maximum data length is controlled by the
# parameter `maxlen`, which defaults to 65535.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, maxlen: int = 65535
    ):
    # Generate header and first block
    header = bytes((
        0x47, 0x49, 0x46, 0x38, 0x39, 0x61
    ))
    global_color_table_flag = False
    
    # Make up a random logical screen descriptor
    width = int.from_bytes(
        rng.read(2),
        byteorder='big',
        signed=False
    ) % (2**10 - 1) + 1

    height = int.from_bytes(
        rng.read(2),
        byteorder='big',
        signed=False
    ) % (2**10 - 1) + 1

    gcd1 = int.from_bytes(
        rng.read(1),
        byteorder='big',
        signed=False
    ) & ~(1 << 7) | min(width >> 7, 1)

    gcd2 = int.from_bytes(
        rng.read(1),
        byteorder='big',
        signed=False
    ) & ~(1 << 7) | min(height >> 7, 1)

    bgcolorindex = int.from_bytes(
        rng.read(1),
        byteorder='big',
        signed=False
    ) % 256

    pixelaspectratio = 0

    lsd = struct.pack('>HHBBB', width, height, gcd1, gcd2, bgcolorindex) \
            + bytes((pixelaspectratio,))

    # Write out LSD, GCD, then application extension for Netscape loop count
    out.write(lsd)
    if global_color_table_flag == True:
        color_resolution = 1        # Use 2^n colors in palette
        sorted_flag = True           # Sort colors by frequency
        sizeOfGlobalColorTable = 7   # 2^(sizeOfGlobalColorTable + 1) entries

        bgColorIndex = int.from_bytes(
            rng.read(1),
            byteorder='big',
            signed=False
        )
        
        aspectRatio = 0

        gct = []        
        for i in range(0, (1<<sizeOfGlobalColorTable)):
            r = int.from_bytes(
                rng.read(1), 
                byteorder='big',
                signed=False
            )

            g = int.from_bytes(
                rng.read(1),
                byteorder='big',
                signed=False
            )
            
            b = int.from_bytes(
                rng.read(1),
                byteorder='big',
                signed=False
            )

            gct += [r,g,b]

        lzwMinCodeSize = 2    # Minimum code size for encoding
        out.write(struct.pack('BBBB',
            0xFF,       # AppExtension introducer
            0x0B,       # Application Extension Block Size
            0xF9,       # Extention Code (Application Extension)
            0x04,       # Length of sub-block; must be at least 12
            1,          # Netscape Looping extension label
            len(lsd)+len(gct)+12))             # Sub-block payload length
        out.write(bytearray([
            lzwMinCodeSize,     # Data sub-block ID
            bgColorIndex]))    # Background color index
        out.write(bytes(gct).ljust(2**(sizeOfGlobalColorTable+1)*3, b'\x00')) 
    else: 
        pass
    netscape_loop_count = int.from_bytes(
        rng.read(3),
        byteorder='big',
        signed=False
    )

    out.write(netscape_loop_count.to_bytes(3,'big'))

    # Now encode all frames with no interframe compression and LZW minimum
    # code sizes between 2 and 12 bits long
    numframes = int.from_bytes(
        rng.read(2),
        byteorder='big',
        signed=False
    ) % 1024 + 1
    for f in range(numframes):
        frame_dims = []
        local_color_table_flag = bool(int.from_bytes(
            rng.read(1), 
            byteorder='big',
            signed=True
        ))

        # Frame offsets are from start of block not header or previous frame
        offset_of_local_color_table = None
        if local_color_table_flag == True:
            sizeOfLocalColorTable = int.from_bytes(
                rng.read(1),
                byteorder='big',
                signed=False
            ) % 3 + 1
            offset_of_local_color_table = int.from_bytes(
                rng.read(2),
                byteorder='big',
                signed=False
            ) % 65535


        leftEdge = int.from_bytes(
            rng.read(2),
            byteorder='big',
            signed=False
        ) % (2**10 - 1) + 1

        topEdge = int.from_bytes(
            rng.read(2),
            byteorder='big',
            signed=False
        ) % (2**10 - 1) + 1

        rightEdge = int.from_bytes(
            rng.read(2),
            byteorder='big',
            signed=False
        ) % (2**10 - 1) + 1
            
        bottomEdge = int.from_bytes(
            rng.read(2),
            byteorder='big',
            signed=False
        ) % (2**10 - 1) + 1

        fcd1 = int.from_bytes(
            rng.read(1),
            byteorder='big',
            signed=False
        ) & ~(1 << 7) | min((rightEdge - leftEdge)>>7, 1)

        fcd2 = int.from_bytes(
            rng.read(1),
            byteorder='big',
            signed=False
        ) & ~(1 << 7) | min((bottomEdge - topEdge)>>7, 1)

        duration = int.from_bytes(
            rng.read(2),
            byteorder='big',
            signed=False
        ) % (2**10 - 1) + 1


        disposalMethod = int.from_bytes(
            rng.read(1),
            byteorder='big',
            signed=False
        ) % 8

        userInputFlag = int.from_bytes(
            rng.read(1),
            byteorder='big',
            signed=False
        ) % 2

        transparencyGivenFlag = int.from_bytes(
            rng.read(1),
            byteorder='big',
            signed=False
        ) % 2

        transparentColorIndex = None
        if transparencyGivenFlag == 1:
            transparentColorIndex = int.from_bytes(
                rng.read(1),
                byteorder='big',
                signed=False
            ) % 256

        out.write(bytes((
            0x21,         # Extension introducer
            0xFE,         # Graphic Control Label
            0x04,         # Graphic Control Block Size
            0x08,         # Packed field for Disposal Method
                             #      User Input Flag
                             #      Transparency Given Flag
            int(((disposalMethod&~7)<<2)|userInputFlag|(transparencyGivenFlag<<1)),
            ((duration>>0)&0xf)*10+(duration>>8),
            ((duration>>4)&0xf)*10+(duration>>12),
            (((transparentColorIndex>>0)&0x1f)<<3)|((transparentColorIndex>>5)&0x7),
            (((transparentColorIndex>>3)&0x3f)<<2)|((transparentColorIndex>>11)&0x3),
            ((transparentColorIndex>>6)&0x3f)))
        )
        if local_color_table_flag == True:
            # Local color table; we just use whatever was generated
            # above but check that it's not too big...
            assert offset_of_local_color_table <= len(lsd)\
                   + len(gct) + len(frame_dims*3) + 1,\
                    "LCT offset greater than end of frame"
            ct = lsd[offset_of_local_color_table:]\
                .ljust(1<<(sizeOfLocalColorTable+1)*3, b'\x00')[:1<<(sizeOfLocalColorTable+1)*3]
            out.write(bytes((sizeOfLocalColorTable,)))
            out.write(ct)

        out.write(bytes((
            leftEdge & 0xff,
            leftEdge >> 8,
            topEdge & 0xff,
            topEdge >> 8,
            fcd1,
            fcd2,
            ((rightEdge-leftEdge) & 0xff),
            ((rightEdge-leftEdge) >> 8),
            ((bottomEdge-topEdge) & 0xff),
            ((bottomEdge-topEdge) >> 8)))
        )