#!/usr/bin/env python3

import binascii
import os
import random as rand
import sys
import time


def write_uint8(out: BinaryIO, val):
    if not isinstance(val, int):
        raise TypeError('Not an unsigned integer')

    if val >= pow(2, 8):
        raise ValueError('Value is too large to fit in one byte')

    return out.write(struct.pack('<B', val))


def write_uint32(out: BinaryIO, val):
    if not isinstance(val, int):
        raise TypeError('Not an unsigned integer')

    try:
        out.write(struct.pack('<I', val))
    except OverflowError:
        raise ValueError('Too big for uint32!')



# Makes sure that each pixel in frame `frame` has
# the same color at every frame step.
# `frame` must be a list or tuple of lists
# (each sublist should contain exactly three elements).
# `step` specifies what percentage of frames will have this effect.
#   e.g.: if step == 5 then 5% of all frames will get modified;
#          if step == 100 then no frames will get modified
def add_color_uniformity(frame, step):
    num_frames = len(frame)
    pixels_per_frame = len(frame[0])

    if not (isinstance(frame, (tuple, list)) and \
            isinstance(frame[0], (tuple, list)) and \
            isinstance(frame[0][0], int)):
        raise TypeError('Frame structure does not match expectations!')

    if step > 100 or step < 1:
        raise ValueError('Step argument must be between 1 and 100 inclusive.')

    print('\nAdding uniformity to given frame...')
    start_time = time.time()


    for i in range(num_frames):

        if i!= 0 and i % round((float)(1 / float(step))) == 0:

            prev_frame = frame[i - 1]

            if i >= num_frames//2 and \
               len(prev_frame) == pixels_per_frame:

                ind = rand.randrange(len(prev_frame))
                r = prev_frame[ind][0]
                g = prev_frame[ind][1]
                b = prev_frame[ind][2]

                for j in range(pixels_per_frame):
                    frame[i][j] = (r, g, b)


        else:
            continue

    end_time = time.time()
    duration = end_time - start_time
    print('Done! Took {}s\n'.format(round(duration, 3)))




# Pads frame with black pixels so that it's always rectangle-shaped.
# The padding value can be specified through the optional parameter `fill`.
def pad_to_square(frame, fill=(0, 0, 0)):

    width = max([p[0] for p in frame])
    height = max([p[1] for p in frame])

    wdiff = width - len(frame[0])
    hdiff = height - len(frame)

    if not ((isinstance(wdiff, int) and wdiff >= 0) and
            (isinstance(hdiff, int) and hdiff >= 0)):
        raise RuntimeError('Unable to calculate diff values!')

    if wdiff == 0 and hdiff == 0:
        return



    print('\nPadding given frame to square size...')
    start_time = time.time()


    while wdiff > 0 or hdiff > 0:

        if hdiff > 0:
            frame.append([(0, 0, 0)] * width)
            hdiff -= 1

        elif wdiff > 0:
            for line in frame:
                line += [(0, 0, 0)]
            wdiff -= 1




    end_time = time.time()
    duration = end_time - start_time
    print('Done! Took {}s\n'.format(round(duration, 3)))





# Converts frame to RLE in order to reduce its size.
# This function works only on rectangular frames.
def convert_to_RLE(frame):

    if any(any(x) for x in frame) and \
       type(frame[0][0]) not in [int, bool]:
        print('[WARN] Frame contains nonzero colors AND is not '
             'rectangular.\n       Will skip RLE conversion.',
              file=sys.stderr)
        return

    if not isinstance(frame, (tuple, list)):
        raise TypeError('Given frame is not a list nor a tuple!')


    print('\nConverting given frame to run-length encoding...')
    start_time = time.time()

    # First, we need to detect which consecutive lines are identical.
    # To do this, we'll compare each line to the previous one. If they're
    # equal, we'll mark them with the special "repeat" flag.
    repeat_flag = False
    new_frame = []

    for i in range(len(frame)):

        line = frame[i]

        if i == 0:
            new_frame.append(line)
            continue

        prev_line = frame[i - 1]

        if line == prev_line:
            repeat_flag = True
            continue

        else:
            if repeat_flag:
                # We've just finished a set of repeating lines.
                # Let's store the first one and the count.
                repcount = 1
                cur_rep_line = prev_line

                while cur_rep_line == prev_line:

                    # Check whether there are more repetitions...
                    next_index = i + repcount

                    #...if yes, increase the counter by one...
                    if next_index < len(frame):
                        repcount += 1

                        #...and also update the currently comparing element.
                        cur_rep_line = frame[next_index]


                    else:
                        break


                # We reached the end without finding another occurrence.
                # So let's stop and process the last sequence found manually.
                i += repcount - 1


            # Now that current line is different from the previous one,
            # check whether our temporary flag was raised.
            if repeat_flag:
                # Store the repeated line alongside the count.
                new_frame[-1].extend([cur_rep_line, repcount & 0xFF])

            else:
                new_frame.append(line)

            repeat_flag = False


    # At this point, the whole frame has been processed and stored under
    # variable `new_frame`. However, some zeroes might still remain.
    # Let's remove these excess zeros and store the result under final name.
    final_frame = []

    # I'm also storing the length difference here.
    # (Difference between original and reduced sizes.)
    dlen = 0

    for line in new_frame:

        new_line = []

        for pxl in line:

            nxt = pxl[:]

            # Is this pixel actually empty?
            if sum(pxl) == 0:
                dlen += 1

            else:
                new_line.append(nxt)



        # Skip adding white lines altogether.
        if new_line:
            final_frame.append(new_line)



    # Add the magic code at the beginning and the end of the frame.
    final_frame[:0] = [[[255, 255]]]
    final_frame.append([[255, 256]])



    end_time = time.time()
    duration = end_time - start_time
    print('Done! Took {}s\n'.format(round(duration, 3)))

    print('Original and reduced lengths:',
          '{}px'.format(len(frame)),
          '-',
          '{}px'.format(len(final_frame)))

    print('Bytes saved:', sys.getsizeof(frame)-sys.getsizeof(final_frame),
          '=',
          '{:.1f}%'.format(
          100 - (((float)(sys.getsizeof(final_frame))) / \
                  ((float)(sys.getsizeof(frame))))*100)
         )

    return final_frame




def compress_frame(frame):
    new_frame = convert_to_RLE(frame)
    padded = pad_to_square(new_frame)
    add_color_uniformity(padded, 5)

    return padded




def create_gif(outpath, frames):

    # In order to save space efficiently, the entire image is divided into
    # several blocks called "graphics control blocks". One block per frame.
    GCB_size = 8 * len(frames)                         # Block size

    # And once again, since blocks consume space, we'd like to find out how
    # many blocks are required for the image itself.
    num_blocks = len(frames)                           # Number of blocks

    # To determine the size of the final image, we must use the LZW scheme.
    # The uncompressed size of our image will thus be calculated according to
    # the below formula written in natural language:
    #
    #     compressed_size = size_of_all_colors      +
    #                       size_of_LZW_codes    +
    #                       total_uncompressed_data
    #
    # Therefore, we only need the sum of all colors.
   