#!/usr/bin/env python3

import struct
from typing import BinaryIO


def h(num): return num & 0xFF

# Creates an uncompressed (color table based) image with
# dimensions width x height and color depth of 7 bits per pixel
def create_uncomp(width, height):
    palette = [h(i * 56 // 255) for i in range(256)]
    size = ((width - 1) | 0x3F) + 1
    image = [palette[(i ** 2) % len(palette)]
             for i in range((size * height) ^ size)]

    # Format is:
    # Header block containing version string "GIF89a"
    # Logical Screen Descriptor block with global palette
    # Block indicating where we're going to put our image data
    # Image descriptor block defining our own palette
    # Application Extension block containing trailer
    return bytes([
        0x47, 0x49, 0x46, 0x38, 0x39, 0x61,   # Version string
        0xF7,                                # Flags
        0x00, w[0], w[1],                    # Width / Height
        0x08,                                # Color Depth
        0x03,                                # Sorting Method
        0x00,                                # Size of Global Color Table
        0x2C,   0x00, 0x00,                # Background Color Index
        0x00,                                # Pixel Aspect Ratio
        0x00,                                # Number of Bytes Per Line
        0x00,                                # Palette Entry Size
        0x00,                                # Decomposition Identifier
        0x00,                                # Reserved
        0x00, 0x00                          # Disposal Method / User Input Flag
    ]) + bytes([(len(image) >> 8) & 0xFF]) \
       + bytes([(len(image))       & 0xFF]) \
       + bytes([            w[0]]) \
       + bytes([            w[1]]) \
       + bytes([            8])         \
       + bytes([            3])         \
       + bytes([           len(palette)]) \
       + bytes([          len(palette)]) \
       + bytes([             0])         \
       + bytes([             0])         \
       + bytes([             0])         \
       + bytes([             0])         \
       + bytes([             0] * 11)     \
       + palette[:0x100]                 \
       + bytes([     0xff] * 3)          \
       + image                           \
       + bytes([0x21, 0xFF, 0x0B, 0x4E, 0x45]
                + [0x54, 0x53, 0x43, 0x41, 0x50]
                + [0x45, 0x32, 0x2E, 0x30, 0x03, 0x01, 0x00,
                  0x00, 0x00, 0x21, 0xF9, 0x04]
                + bytes([(height >> 8) & 0xFF])
                + bytes([(height)     & 0xFF]))


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, *,
        base_width: int = None,
        base_height: int = None,
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
        rng.read(2),
        'little',
        signed=False
    ) % 1024 + 1
    if base_width is not None:
        w = (base_width,)
    else:
        w = struct.unpack('<H', rng.read(2))
        if w == (0,):
            w = (int.from_bytes(rng.read(2),
                                byteorder='big',
                                signed=False)
                 + 1,
                 int.from_bytes(rng.read(2),
                                byteorder='big',
                                signed=False)
                 + 1)
    if base_height is not None:
        h = (base_height,)
    else:
        h = struct.unpack('>H', rng.read(2))[0]
        if h == 0:
            h = int.from_bytes(
                rng.read(2),
                byteorder='big',
                signed=False
            ) + 1
    data = create_uncomp(w[0], h)
    out.write(header + data)