#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates an n-bit binary number and returns it as bytes
def bintobytes(n):
    return format(n,'b').zfill(8).encode()

# Converts a list of numbers to a byte string
def packnumbers(*args):
    #print('Packing', args)
    output = []
    for arg in args:
        if isinstance(arg, tuple):
            output += [struct.pack('<I', arg[0])]
            output += [bintobytes(arg[1])]
        elif isinstance(arg, str) or len(arg) < 4:
            packed_str = ''
            while len(packed_str) % 4!= 0:
                packed_str += chr(0)
            if isinstance(arg, str):
                output += [packed_str[:len(arg)].encode()]
                output += [arg]
            else:
                output += [arg]
        else:
            raise Exception('Unsupported type'+ repr(type(arg)))

    #print('Packed', len(output))
    return b''.join(output)


# Writes a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        size=500*500, frames=50, palette_size=768, colordepth=8,
        background=(0x00, 0x00, 0x00), output='out.gif',
        rng='/dev/urandom'):
    with open('/tmp/test', 'wb+') as tmp:
        generated_data = b''

        # Header
        generated_data += b"GIF89a"     # Header
        generated_data += b"\x00\x00"   # Width placeholder
        generated_data += b"\x00\x00"   # Height placeholder
        generated_data += b"\xF0"       # Flags - global colour table
        generated_data += b"\x00"       # Background colour index (palette)
        generated_data += b"\x00"       # Pixel aspect ratio
        
        # Logical screen descriptor
        width = round(size **.5 // 2)*2
        height = round((width * 2)/3)
        print("Width:", width)
        print("Height:", height)
        
        generated_data += packnumbers(
            6,                        # Size
            (width, height),          # Dimensions
            b'\x87\x00',               # Color depth & colour resolution
            b'\x00',                   # GCT sort flag
            )
            
        # Global colour table        
        gct = []
        for i in range(0, palette_size):
            gct += [(background[0],
                     background[1],
                     background[2])]
        gct += [(255, 255, 255)] * (palette_size - len(gct))
        generated_data += packnumbers(
            3*(palette_size+1),         # Size 
            *gct                       # Palette entries
            )
    
        # Graphic control extension
        generated_data += packnumbers(
            4,                          # Extension introducer
            b'\xF9',                    # Graphic control label
            b'\x04',                    # Block size
            b'\x00',                    # Flags
            0xFFFF                      # Delay time
            )
        generated_data += packnumbers(
            3,                         # Size of block
            b'\xFF\x00\x00',            # End of image
            )
    
        # Image descriptor
        generated_data += packnumbers(
            4,                           # Image separator
            width//2,                    # Left position
            height//2,                   # Top position
            width//2,                    # Width
            height//2,                   # Height
            0x08,                        # LZW minimum code size
            )
    
        # Data subblocks
        rawimage = []
        for frame in range(frames):
            rawimage += [os.urandom(frame%10)]
        lzwcodes = dict()
        prefix = (1,)
        v = None
        codes = {prefix}

        def getcode():
            if not codes:
                nextcode = max(lzwcodes)+1
                print('Adding code ', nextcode, 'to dictionary')
                newcodes = set()
                newcodes |= {(nextcode,) + c
                            for c in codes}

                lzwcodes[(v, prefix)] = nextcode
                
                codes = newcodes
                prefix += (v,)
            return min(codes)

        for frame in rawimage:
            for byte in frame:
                if (byte, prefix) in lzwcodes:
                    v = lzwcodes[(byte, prefix)]
                else:
                    if v is not None:
                        try:
                            code = getcode()
                        except ValueError:
                            pass # Do nothing
                        
                            
        # Finish using last code
        if True:
            code = getcode()

        
        print(generated_data == b'')
        
        print(sum([len(f) for f in rawimage]),
              sum([sys.getsizeof(f) for f in rawimage]))
        exit()


