#!/usr/bin/env python3

from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (`/dev/urandom`)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header (fixed)
    hdr = b"GIF89a"
    out.write(hdr)

    # Logical screen width and height
    w = 2**int.from_bytes(
            rng.read(1), "big") % 65000   # [0x0000, 0xffff]
    h = 2**int.from_bytes(
            rng.read(1), "big") % 65000   # [0x0000, 0xffff]
    out.write((w).to_bytes(2, 'big'))
    out.write((h).to_bytes(2, 'big'))

    # Color table size
    ctsize = 7 << int.from_bytes(
                rng.read(1), "big") & 7    # [0x00, 0x07]
    print("ctsize", ctsize)            # For debugging only
    out.write(((ctsize+1)*3).to_bytes(1, 'big'))


    # Zero-terminator of color table
    out.write(b'\x00')

    palette = []
    for i in range(4**(ctsize+1)):

        # R-value is in the first three bytes
        r = int.from_bytes(rng.read(1), "big") & 0xff
        g = int.from_bytes(rng.read(1), "big") & 0xff
        b = int.from_bytes(rng.read(1), "big") & 0xff
        palette += [(r,g,b)]                  # Add to palette list
        out.write(r.to_bytes(1,'big'))       # Write red component
        out.write(g.to_bytes(1,'big'))       # Write green component
        out.write(b.to_bytes(1,'big'))       # Write blue component

    # Size of global color table
    out.write((len(palette)).to_bytes(1,"big"))

    # Background color index
    bgindex = len(palette)-1      # Use last color as background color
    out.write(bgindex.to_bytes(1,"big"))

    # Pixel aspect ratio
    paratio = int.from_bytes(rng.read(1),"big")&0xf     # [0x00, 0x0f]
    out.write(paratio.to_bytes(1,"big"))

    # Image descriptor (fixed)
    imgdesc = b"\x2c\x00\x00\x00"
    out.write(imgdesc)

    # Image left offset
    xoff = int.from_bytes(rng.read(2),"big")      # [0x0000, 0xffff]
    out.write(xoff.to_bytes(2,"big"))

    # Image top offset
    yoff = int.from_bytes(rng.read(2),"big")      # [0x0000, 0xffff]
    out.write(yoff.to_bytes(2,"big"))

    # Width of current frame
    w = 2**int.from_bytes(rng.read(2),"big")%10000        # [0x0000, 0xffff]
    out.write(w.to_bytes(2,"big"))
    if not w : w=2**int.from_bytes(rng.read(2),"big")%10000        # [0x0000, 0xffff]
    print('w', w)

    # Height of current frame
    h = 2**int.from_bytes(rng.read(2),"big")%10000        # [0x0000, 0xffff]
    out.write(h.to_bytes(2,"big"))
    if not h: h = 2**int.from_bytes(rng.read(2),"big")%10000       # [0x0000, 0xffff]
    print('h', h)

    # LZW minimum code size used
    lzws = int.from_bytes(rng.read(1),"big")&7       # [0x00, 0x07]
    out.write(lzws.to_bytes(1,"big"))

    # Transparency flag
    transflag = bool(int.from_bytes(rng.read(1),"big")&0x1)
    out.write(transflag.to_bytes(1,"big"))

    # User input flag
    inpflag = bool(int.from_bytes(rng.read(1),"big")&0x1)
    out.write(inpflag.to_bytes(1,"big"))

    # Delay time
    delaytime = int.from_bytes(rng.read(2),"big")           # [0x0000, 0xffff]
    out.write(delaytime.to_bytes(2,"big"))
    noframes = max([0, int(delaytime)])
    print('delaytime', delaytime, 'noframes', noframes)

    # Transparent color index
    tcolidx = -1 if not transflag else len(palette)-1    # Use last if transparent, use None otherwise
    out.write(tcolidx.to_bytes(1,"big"))

    # Loop count
    loopcount = 0
    out.write(loopcount.to_bytes(2,"big"))


    # Data subblocks
    for f in range(noframes):

        # Block size indicator
        blksizeind = int.from_bytes(
                    rng.read(1),"big") & 0x7f         # [0x00, 0x7f]
        out.write(blksizeind.to_bytes(1,"big"))
        print('blksizeind', blksizeind)

        # Prefix codes
        prefixcodes = {}              # Maps from prefixes to suffixes
        for i in range(8):

            # The following two lines could be replaced with:
            # prefixbyte = int.from_bytes(rng.read(1),"big") & 0x7f
            prefixbyte = int(i) ^ rng.read(1)[0] & 0x7f # The & operator returns an int

            # Check if this prefix has been seen already
            if prefixbyte in prefixcodes:
                continue

            # Generate a suffix
            suffixbyte = int.from_bytes(
                        rng.read(1),"big") & 0x7f             # [0x00, 0x7f]
            if not suffixbyte or suffixbyte == 0xff:          # Exclude null terminator
                continue                                      # and repeat values
            while suffixbyte > 0x80:                          # Repeat until below 0x80

                # Read next byte
                extra = int.from_bytes(
                            rng.read(1),"big")                   # [0x00, 0xff]

                # Concatenate it with previously read byte
                suffixbyte ^= ((extra & 0xc0)<<2) | \
                                (extra & 0x3f)                    # [0x00, 0x3f]

            # Update dictionary mapping
            prefixcodes[prefixbyte] = suffixbyte


        # Null terminator
        postfixcode = 0
        out.write(postfixcode.to_bytes(1,"big"))
        print('postfixcode', postfixcode)


        # LZW encoded pixel data
        pixeldata = ""               # Accumulate all pixels here
        while True:                  # Continue until we find 2nd null

            # Get new prefix
            prefixbyte = int.from_bytes(
                        rng.read(1),"big") & 0x7f                 # [0x00, 0x7f]
            print('prefixbyte:', hex(prefixbyte))


            # Fetch corresponding postfix
            if prefixbyte!= 0 or postfixcode==0x7f:
                try:
                    postfixcode = prefixcodes[prefixbyte]    # This should work
                except KeyError:                        # but just to make sure...
                    raise Exception('Cannot map prefix '+str(hex(prefixbyte))+
                                    ', please report this bug at:\nhttps://github.com/DennisMitchell/generativepy/issues/')
            else:                                       # If the resulting sequence ends
                break                                   # then exit loop early


            # Translate to actual symbol
            c = chr(0x21 + min(postfixcode, 0x1f))      # [0x21, 0x3a], excluding full block character
            print("c:", c)
            pixeldata+=c                               # Append to output string

        # Finally write the whole subblock
        out.write(pixeldata.encode())


    # Write end of image marker
    trailer = b';'
    out.write(trailer)