#!/usr/bin/env python3

import os
import sys
import zlib
from random import Random
from typing import BinaryIO

def write_uint16(v: int, f: BinaryIO):
    assert v >= 0 and v <= (2**16-1)
    f.write(struct.pack('>H', v))

def write_int32(v: int, f: BinaryIO):
    # Note that this handles negative values too!
    f.write(struct.pack('>i', v & ((1<<32)-1)))

def make_imagedata(w: int, h: int, rng: Random, pixeldata=None):
    if pixeldata is None:
        arr = []
        for i in range(h*w):
            r, g, b = [x % 256 for x in rng.randrange(765)]
            arr += [r | g<<8 | b<<16]
        return bytes(arr)

    arr = [pixeldata[i * w:(i+1) * w]
           for i in range(h)][::-1]
    return b"".join([b"".join(row) for row in arr])

def make_palette(rng: Random, palsize: int):
    ans = [(rng.randrange(256), rng.randrange(256), rng.randrange(256))
           for _ in range(palsize)]
    return ans

def apply_palette(pixeldata, palette):
    def mapfn(p):
        r, g, b = p // 256**2, (p // 256) % 256, p % 256
        r, g, b = palette[r][0], palette[g][1], palette[b][2]
        return r|g<<8|b<<16
    new_pixeldata = []
    for pixel in pixeldata:
        new_pixeldata += [mapfn(pixel)]
    return bytes(new_pixeldata)

def apply_dithering(pixeldata):
    def dither():
        r = [0]*3
        g = [0]*3
        b = [0]*3

        # The idea here of generating noisy indices
        # and then just applying them to colors works
        # well enough for a balanced result even with
        # simplistic dithers like these
        for j in range(4):

            # First go through each color component independently...
            for i in range(3):
                r[i] = yield from ditherone(r[i], g[i], b[i], 0, 8 - j)

            for i in range(3):
                g[i] = yield from ditherone(r[i+1], g[i+1], b[i+1], 1, 8 - j)

            for i in range(3):
                b[i] = yield from ditherone(r[i+2], g[i+2], b[i+2], 2, 8 - j)

    def ditherone(a, b, c, idx, shift):
        error = a + b + c
        if error > 0:
            yield (idx << 2) + 3

            # This line gives an explicit bias towards shifting towards
            # zero. It's equivalent to (error >> 3). This should have no effect
            # on the final output due to rounding errors anyway.
            errscale = ((error + 4) >> 3) + 1
        else:
            yield (idx << 2) + 2
            errscale = (-error) + 1

        error -= errscale * 8
        if error < 0:
            error += 8;
            yield (idx << 2) + 2
        elif error > 0:
            yield (idx << 2) + 3
            error -= 8
        yield (idx << 2) + 1

        answer = min(max(round((error + errscale / 2) / errscale), 0), 255)
        return answer

    pixels = list(pixeldata)
    it = dither()
    while True:
        try:
            pixval = next(it);
            val = pixels.pop(0)
            val = (pixval << 24) + (val & 0xffffff)
            pixels.append(val)
        except StopIteration as e:
            break

    print("len(pixels)", len(pixels), "len(pixeldata)", len(pixeldata))
    return bytes(pixels)

def encode_image_into_frame(
        imagedata: bytes, width: int, height: int, palettesize: int,
        lzwmincodesize: int, interlaced: bool
    ):
    frames = []
    maxcolors = {
        1: 2,
        2: 4,
        4: 16,
        8: 256,
        16: 2 ** 12,
    }

    palsize = 2 ** palettesize
    colorsadded = 1
    while colorsadded < palsize:
        colorsadded *= 2
    numcolorbits = round(math.log2(colorsadded)).to_bytes(
            1, byteorder="little")
    #print("numcolorbits", numcolorbits, palettesize)
    anycolor = False
    newpixeldata = []
    for p in imagedata:
        if not anycolor and p!= 0xff:
            anycolor = True
        if anycolor or p == 0xff:
            newpixeldata += [p]
    imagedata = bytes(newpixeldata)
    #print("original len:", len(imagedata), "new len:", len(newpixeldata))
    imagesize = len(imagedata)
    codesize = 8
    codemask = 2 ** codesize - 1
    #print("palettesize:", palettesize, "maxcolors:", maxcolors[palettesize])
    table = {}
    clearcode, endofinfo = 0, 1
    codes = set([clearcode, endofinfo])
    resettable = True
    currpixpos = 0
    remaining = ""
    prevvals = []
    while currpixpos < imagesize:
        if resettable:
            table = {}
            # We can always add one more value before we need to
            # increase code size again.
            if len(codes) < maxcolors[palettesize]:
                codes.add(currpixpos)
                resettable = False

        # This is a bit silly but maybe worth it for slightly better compression
        if currpixpos - len(prevvals) >= 16 and \
          all(x in codes for x in prevvals[-16:]):
            prevvals += [currpixpos]
            #print("Found a run starting at", currpixpos)


        if resettable:
            s = imagedata[currpixpos:currpixpos+(1<<(codesize-1))]
            found = s in table
        else:
            found = False
        if not found:
            #print("Doing something...")
            if remaining!= "":
                chunkvals = [remaining[:(1<<codesize)]]
                if currpixpos + len(remaining) < imagesize:
                    chunkvals += [remaining[(1<<codesize):]]

                #print("chunkvals", chunkvals)
                for cv in chunkvals:
                    if cv in table:
                        #print("already had", cv)
                        continue

                    #print("Adding", cv)
                    codes.add(cv)
                    table[cv] = len(codes) - 1


            # Do some kind of binary search
            halfway = (1<<(codesize-1))+1
            lower = 0
            upper = currpixpos + len(imagedata) - 1
            #print("lower", lower, "upper", upper, "halfway", halfway)
            found = False
            while lower <= upper:
                midpoint = math.floor((lower+upper)/2)


                s = imagedata[midpoint:midpoint+(1<<(codesize-1))]
                #print("s", s)
                if s in table:
                    found = True
                    midpoint += len(s)
                    break
                elif s < imagedata[lower:lower+(1<<(codesize-1))]:
                    upper = midpoint-1
                elif s > imagedata[upper-(1<<(codesize-1)):upper]:
                    lower = midpoint+1
            #print("found", found)
            if not found:
                raise RuntimeError("Could not find a suitable spot.")



        chunk = imagedata[currpixpos:midpoint]
        remaining = imagedata[midpoint:]

        #print("chunk", chunk, "chunkval", table[chunk])
        code = table[chunk]
        encoded = code.to_bytes(
                1, byteorder="little"
            )
        #print("encoded", encoded)
        currcodesize = 0
        while code >= 2**(currcodesize + 1):
            currcodesize += 1
        if currcodesize > codesize or len(codes) == 1<<currcodesize:
            encoded = b"\x01"+encoded

        currpixpos += len(chunk)
        if currpixpos == imagesize and remaining == "":
            # Don't bother adding padding