#!/usr/bin/env python3

import struct
from typing import BinaryIO

def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical Screen Descriptor
    width  : int   = _get_small_endian_int24(rng)
    height : int   = _get_small_endian_int24(rng)
    flags  : bytes = rng.read(1)
    background: int    = _get_byte_as_int(flags[0]) & 0x7
    aspect: int          = _get_small_endian_int8(rng)
    out.write(_get_small_endian_int16(width))
    out.write(_get_small_endian_int16(height))
    out.write(b'\xf7') # Global Color Table Flag, size of global color table (0 = no gct; not used in this case)
    out.write(b'\x00\x00') # Background color index
    out.write(b'\x00')     # Pixel Aspect Ratio (not used for GIFs)

    # Image Descriptor block
    # left position is always zero on my screen
    out.write(b'\x00\x00')
    out.write(_get_small_endian_int16(height))
    out.write(b'\x00\x00')
    out.write(b'\x00')

    # Local Color Table
    numcolors: int        = _get_byte_as_int(rng.read(1)[0]) & 0x7f
    colors: list          = []
    
    for _ in range(numcolors):
        rgbcolor = rng.read(3)
        colors.append(rgbcolor)

    _write_gct(out, colors)

    # Graphics Control Block
    out.write(b'\x2c')           # Extension Introducer
    out.write(b'\x00')           # Sub-block Length (always even)
    out.write(b'\xf9')           # Application identifier
    out.write(b'\x04')           # Block Size
    duration                   = _get_small_endian_int16(rng)
    transparent: bool          = False
    disposalmethod: int        = (_get_byte_as_int(rng.read(1)[0]) >> 5) & 0x7
    userinput: bool            = _get_byte_as_int(rng.read(1)[0]) & 0x1!= 0
    out.write(duration.to_bytes(2, 'little'))
    out.write((transparent | (((disposalmethod << 3) & 0xe0) | ((userinput << 2) & 0x1c))).to_bytes(1, 'little'))
    out.write(b'\xff')           # Transparent Color Index

    # Image Data are written here

    # Block Terminator
    out.write(b';')


def _get_byte_as_int(byte: bytes) -> int:
    return int.from_bytes(
            byte, byteorder='big', signed=False)

def _get_small_endian_int16(buffer: BinaryIO) -> int:
    return _get_byte_as_int(
            buffer.read(2))

def _get_small_endian_int24(buffer: BinaryIO) -> int:
    return _get_byte_as_int(
            buffer.read(3)) * 0x100 + \
           _get_byte_as_int(
               buffer.read(1))

def _get_small_endian_int8(buffer: BinaryIO) -> int:
    return _get_byte_as_int(
            buffer.read(1))
        
def _write_gct(out: BinaryIO, colors: list) -> None:
    out.write(len(colors).to_bytes(1, 'little'))
    for color in colors:
        out.write(color)