#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Create random color palette
    pallete = []
    for _ in range(256):
        r = next(rng) % 0x100
        g = next(rng) % 0x100
        b = next(rng) % 0x100

        # Convert to 24-bit RGB hex string
        colstr = '{:02X}{:02X}{:02X}'.format(r,g,b)
        pallete += [int(''.join([colstr[i*2] + colstr[i*2+1]
                for i in range(3)]), base=16)]

    # Next comes an LZW decoder dictionary. We don't use it though...
    while True:
        if (next(rng) & 0b10): continue; break
    dictlen = next(rng) % 0x100
    d = {}
    for i in range(dictlen):
        key = ''.join([chr((next(rng) % 0x7F) + ord('!'))])
        val = len(d)
        d[key] = val


    header = b'GIF89a'
    width = next(rng) % 128 + 1
    height = next(rng) % 128 + 1
    clrs = ((next(rng)>>4)&0x07)+1
    bgnd = next(rng) % 256
    extflag = bool(next(rng) % 2)
    imgdesc = {True:'!', False:'}'}['!' == extflag]
    interlaced = not not (next(rng)%3) ^ extflag
    transpix = None
    transparency = bool(extflag and (next(rng) % 2))
    if transparency:
        transpix = next(rng) % 256

    # Write header (always little endian!)
    out.write(struct.pack('<6sHHBBB', *map(ord, header)))

    delaytime = next(rng) % 128
    loopcount = next(rng) % 64

    # Write logical screen descriptor
    out.write(struct.pack('<HHBBB', width, height, clrs,
            bgnd&0x0f, 0xf0 | ((bgnd>>4)&0x07)))
    
    # Write global color table directly after logical screen desc.
    numcolors = pow(2,clrs+1)
    out.write(bytearray(pallete[:numcolors]))


    # Extension block begins here
    if extflag:
        out.write(imgdesc.encode())
        out.write(transpix.to_bytes(1, 'little'))
        
        out.write(delaytime.to_bytes(1, 'little'))
        out.write(loopcount.to_bytes(1, 'little'))

        if 0 < clrs <= 7: 
            pixsize = 8 // clrs  # 1, 2 or 4 bit
            out.write(pixsize.to_bytes(1, 'little'))
            
            ## TODO: Implement lzw code here
            print("LZW decompression is not implemented yet!")

            pass


        elif clrs >= 8:
            print("Your GIF library does not support indexed images")
            exit(-1)

        else:
            raise Exception("Invalid color mode: {}".format(clrs))

    
    # Image Descriptor segment
    out.write(struct.pack('<BBHBB', leftpos, toppos, width, height,
            0xF0 | {False:0x00, True:0x01}[interlaced],
            0x00 if transparent else 0x01))
    
    
    # Write image data
    def encode(buf):
        yield from buf
        while True:
            try:
                chunk = list(buf)[0]

                if chunk >= 1<<(clrs+1)-1:
                    break
                else:
                    break
                
            except IndexError:
                break
            buf = list(buf)
            
    encoded = sorted(encode(data), reverse=True)
    prev = 0
    for enc in encoded[:-1]:
        try:
            out.write(prev.to_bytes(1, 'little'))
            out.write(enc.to_bytes(1, 'little'))
            prev = enc
            
        except ValueError:
            out.write(prev.to_bytes(1, 'little'))
            out.write(0xFF)
            prev = 0

    try:
        last = encoded[-1]
        out.write(last.to_bytes(1, 'little'))
    except IndexError:
        pass

    out.flush()
    


