#!/usr/bin/env python3

import functools
import io
import os
import sys
import struct
from typing import BinaryIO

class RandomGifGeneratorOptions():

    def __init__(self):

        self._maxdata = None
        self._minframedelayms = None
        self._maxframedelayms = None
        self._width = None
        self._height = None


    @property
    def maxdata(self):
        return self._maxdata
    
    @maxdata.setter
    def maxdata(self, value):
        if not isinstance(value, int):
            raise TypeError()
        elif value < 1 or value > (pow(2, 16)-1):
            raise ValueError()
        else:
            self._maxdata = value


    @property
    def minframedelayms(self):
        return self._minframedelayms
    
    @minframedelayms.setter
    def minframedelayms(self, value):
        if not isinstance(value, int):
            raise TypeError()
        elif value < 0 or value >= pow(2, 32):
            raise ValueError()
        else:
            self._minframedelayms = value


    @property
    def maxframedelayms(self):
        return self._maxframedelayms
    
    @maxframedelayms.setter
    def maxframedelayms(self, value):
        if not isinstance(value, int):
            raise TypeError()
        elif value <= 0 or value > pow(2, 32):
            raise ValueError()
        elif value < self.minframedelayms:
            raise ValueError('Max frame delay must be greater than minimum.')
        else:
            self._maxframedelayms = value


    @property
    def width(self):
        return self._width
    
    @width.setter
    def width(self, value):
        if not isinstance(value, int):
            raise TypeError()
        elif value < 1 or value > 65535:
            raise ValueError()
        else:
            self._width = value


    @property
    def height(self):
        return self._height
    
    @height.setter
    def height(self, value):
        if not isinstance(value, int):
            raise TypeError()
        elif value < 1 or value > 65535:
            raise ValueError()
        else:
            self._height = value


# Returns the lsb of x as an unsigned byte
def lsb(x: bytes) -> int:
    return ord(struct.pack('<B', x[0])) & 1

# Returns the msb of x as an unsigned byte
def msb(x: bytes) -> int:
    return ord(struct.pack('<B', x[0])) >> 7

# Converts a series of bits to its equivalent integer representation
# in big-endian order
def bitstr_to_intbe(bitstring: str) -> int: 
    num = 0
    for c in reversed(bitstring): 
        num <<= 1
        num |= int(c)
    return num
    
# Converts a series of bits to its equivalent integer representation
# in little-endian order
def bitstr_to_intle(bitstring: str) -> int:    
    num = 0
    for c in bitstring: 
        num >>= 1
        num |= int(c) << 7
    return num
    

# Encodes the given image data in a GIF file at the given path
def encode_image(path: str, imgdata: list[list[bool]], opt: RandomGifGeneratorOptions):
    
    # Writes the given sequence of color indices into the output stream
    def write_colorindices(stream: BinaryIO, colorsused: dict[tuple[int, int, int], int]):
        
        # The current number of distinct RGB tuples used by this frame
        ndistinctcolors = len(colorsused)
        
        # The current LZW code length (as a power of two)
        currcodelength = 9
        
        # A dictionary mapping each 256-color palette entry to a unique index
        indices = {}
        
        # A dictionary mapping each LZW prefix to either a unique prefix code,
        # indicating that it is a valid prefix, or -1, indicating that it points
        # to an invalid code
        codes = {b"": -1}
        
        # The largest prefix key from the previous iteration; will be initialized
        # with the empty string below
        prevprefixkey = b""
        
        # A variable representing the next available code
        availcode = 0
        
        # Write initial control block
        stream.write((ndistinctcolors).to_bytes(1, "big"))
        stream.write(((currcodelength+1).bit_length()).to_bytes(1, "big"))
        stream.write(b"\x80")
        
        
        # Iterate over all values in the image
        for yoffset, row in enumerate(imgdata):
            
            # The first run contains only one color, and does not use any 
            # additional bits
            stream.write(b"\x00")
            
            # The RLE portion starts after that
            while True:
                hasmore = False
                
                # As long as we have not reached the last column on this row...
                for xoffset, px in enumerate(row[:len(row)-(yoffset == len(imgdata)-1)]):
                    
                    # Get the appropriate RGB tuple
                    rgb = tuple([px] * 3)
                    
                    # If the pixel has already been assigned a code...
                    if rgb in indices:
                        
                        # Use that
                        stream.write(indices[rgb].to_bytes(1, "big"))
                        
                    # Otherwise, if there are fewer than 2**currcodelength 
                    # colors left unassigned...
                    elif len(colorsused) < pow(2, currcodelength):
                        
                        # Assign the new color a new code
                        indices[rgb] = len(colorsused)
                        colorsused[rgb] = len(colorsused)
                        
                        # Indicate that there are still more
                        hasmore = True
                        
                        # Append the new color's code to the prefix
                        stream.write(
                            ((availcode>>8)&255).to_bytes(
                                1,
                                "big"
                            ))
                        stream.write(
                            (availcode&255).to_bytes(
                                1,
                                "big"
                            ))
                        
                        # Increment the available code counter
                        availcode += 1
                        
                    # Finally, if we're out of space in our current code table...
                    elif len(colorsused) == pow(2, currcodelength):
                        
                        # Find the longest prefix up to what we've seen so far 
                        bestmatchindex = 0
                        bestmatchlength = 0
                        for i in range(256):
                            
                            # Skip empty prefixes
                            if not codes[(prevprefixkey+i).decode("utf-8")] == -1:
                                
                                # Check how many characters match up to here
                                matches = sum(
                                    [
                                        e==v 
                                        for v, e in zip(
                                            (prevprefixkey+i),
                                            
                                            # This should be the same size as 
                                            # prevprefixkey since we just looked 
                                            # through every possible character
                                            prevprefixkey
                                        )
                                    ]
                                )
                                
                                # If this match is better than the current best 
                                # match found thus far...
                                if matches > bestmatchlength:
                                    
                                    # Update the best match information
                                    bestmatchindex = i
                                    bestmatchlength = matches
                                    
                        # If no prefix was better than anything we have seen 
                        # before then...
                        if bestmatchlength == 0:
                            
                            # Break out of the loop! We cannot possibly compress 
                            # further 
                            break
                            
                        # Get the code corresponding to this prefix
                        bestmatchcode = codes[(prevprefixkey+(bestmatchindex)).decode("utf-8")]
                        
                        # Write out these bytes, which indicate that we need to 
                        # backtrack to find the most recent occurrence of the 
                        # longest matching prefix    
                        stream.write(
                            (((bestmatchcode>>8)+1)//256).to_bytes(
                                1,
                                "big"
                            ))
                        stream.write(
                            (((bestmatchcode>>8)+1)%256).to_bytes(
                                1,
                                "big"
                            ))

                        # Find the second-longest prefix
                        secondbestmatchindex = 0
                        secondbestmatchlength = 0
                        for i in range(256):
                            
                            # Only consider non-empty prefixes
                            if not codes[(prevprefixkey+i).decode("utf-8")] == -1:
                                
                                # Compare it to the previous best match
                                matches = sum(
                                    [
                                        e==v 
                                        for v, e in zip(
                                            (prevprefixkey+i),
                                            
                                            # Same idea as above
                                            prevprefixkey
                                        )
                                    ]
                                )
                                
                                # If there's a tie between this prefix and the 
                                # previously best match, keep track of both
                                if matches == bestmatchlength:
                                    
                                    # If this prefix is better than the 
                                    # currently recorded second best match
                                    if matches > secondbestmatchlength:
                                        
                                        # Update the record
                                        secondbestmatchindex = i
                                        secondbestmatchlength = matches
                                        
                        # Write out the two remaining bytes
                        stream.write(
                            (secondbestmatchindex//256).to_bytes(
                                1,
                                "big"
                            ))
                        stream.write(
                            (secondbestmatchindex%256).to_bytes(
                                1,
                                "big"
                            ))
                    
                        # Save the second-longest prefix as the new current prefix
                        prevprefixkey += chr(bestmatchindex).encode("utf-8") + \
                                          chr(se