#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Random Gif Generator (RGG)
# Copyright (c) 2016 <NAME> <<EMAIL>>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import struct


class RGBABitmap:

    # Color Table size must be a power of two between 256 and 4096 (inclusive).
    def __init__(self):

        self._width       = None
        self._height      = None

        self._global_ctbl = True # Should we write out global color table?
        self._lctbl_size  = False # Is local color table size set?
        self._lctbl       = False # Local color table exists
        self._lctbl_size  = None   # Size of local color table (between 2 and 7)
        self._data        = []     # Data for each frame


    @property
    def width(self):
        return self._width - 1 if self._width else None

    @width.setter
    def width(self, value):
        assert isinstance(value, int) and value > 1
        self._width = value

    @property
    def height(self):
        return self._height - 1 if self._height else None

    @height.setter
    def height(self, value):
        assert isinstance(value, int) and value > 1
        self._height = value

    @property
    def bitmap(self):
        return self._bitmap

    @bitmap.setter
    def bitmap(self, value):
        self._bitmap = value

    def _set_color_table_section(self, ctbl):
        
        i = len(ctbl) // 3

        while not ((i == 2) ^ (i == 4) ^ (i == 8)):

            i /= 2

        if type(i)!= int:
            raise Exception('Color Table Size must be equal to { 2^N } where N is an integer')

        # Set local color table bit shorthand flag.
        self._lctbl_size = bool((i / 2))

        # Set local color table section flag depending on length of color table.
        self._lctbl = bool(((len(ctbl) * 8) - 1 >= 3) & (((len(ctble) * 8) - 1 <= 768)))

    def add_frame(
            self, 
            bitmap : list, 
            lctbl  : bytes    = None,  
            delay  : int      = 100, 
            dispos : str      = 'none'):
    
        if len(bitmap)!= self.width * self.height:
            raise Exception("Bitmap size is incorrect")

        self._data += [
                dict(
                    bitmap = bitmap, 
                    lctbl  = lctbl, 
                    delay  = delay, 
                    dispos = dispos) ]

    def export(self, fpath : str):
    
        with open(fpath, mode="wb") as fh:
    
            # Write Header Section
            fh.write('GIF89a'.encode())

            width  = self.width
            height = self.height
            
            # Logical Screen Descriptor
            fh.write(struct.pack('<HHBBB', 
                width,                                # Canvas Width
                height,                               # Canvas Height
                7,                                    # Bit Depth 
                0,                                    # Global Color Table Flag
                0                                     # Background Color Index
                ))

            # NULL Block
            fh.write('\x00\x00'.encode()) # Pixel Aspect Ratio

            # Application Extension
            fh.write(','.encode()) # Extension Introducer
            fh.write('NETSCAPE2.0'.encode()) # Extension Label
            fh.write('\x0b'.encode()) # Sub-block Size (11 bytes)
            fh.write(struct.pack('<BBHBB', 
                3,                                 # Application Identifier
                1,                                 # Application Authentication Code
                len(self._data) * 2,               # Number of Repetitions
                0                                  # Block Terminator
                ))

            # Graphics Control Extension
            fh.write(','.encode()) # Extension Introducer
            fh.write('XGRAPHIC'.encode()) # Extension Label
            fh.write('\x04'.encode()) # Sub-block Size (4 bytes)
            fh.write(struct.pack('<BHB', 
                4,                             # Block Size (4 bytes)
                delay,                         # Delay Time
                0                              # Transparent Color Index
                ))

            # Comment Extension
            fh.write(','.encode()) # Extension Introducer
            fh.write('COMMENT'.encode()) # Extension Label
            fh.write('\x00'.encode()) # Sub-block Size (0 bytes)
            
            # Plain Text Extension
            fh.write(','.encode()) # Extension Introducer
            fh.write('PLAIN TEXT'.encode()) # Extension Label
            fh.write('\x00'.encode()) # Sub-block Size (0 bytes)
        
            # Loop Forever
            loop_forever_ext = ','.encode() + 'LOOPEXT'.encode() + '\x00'.encode()
            fh.write(loop_forever_ext)            

            frames = []

            # If global color table present then write it here.
            if self._global_ctbl:
                
                # Global Color Table Section
                fh.write(','.encode())

                gctbl_section = '\x00\xff'.encode() * 3 # Red Green Blue
                fh.write(gctbl_section)


            # Write color table section
            if self._lctbl_size:
        
                # Color Table Size Section
                fh.write(','.encode())
                fh.write(('!').encode())
                fh.write(str((int(math.log(self._lctbl_size ** 2, 2)) - 1)).encode())
            
            elif self._lctbl:

                # Local Color Table Flag
                fh.write(','.encode())
                fh.write(('|').encode())
                
            # Write Image Descriptor Section
            frame_number = 0
            for d in self._data:
    
                frame_delay     = d['delay']              # Frame Delay Time
                frame_dispos    = d['dispos'].upper()[0]  # Disposal Method
                frame_img_left  = 0                       # Left Position
                frame_img_top   = 0                       # Top Position
                frame_img_width = width                   # Width
                frame_img_height= height                  # Height
                frame_local_ctbl= self._lctbl            # Use Local Color Table Flag
    
                fh.write(','.encode()) # Extension Introducer
                fh.write('!'.encode()) # Extension Label
                fh.write('\x00'.encode()) # Sub-block Size (0 bytes)
    
                if not frame_local_ctbl:
                    
                    # Write global color table
                    fh.write(','.encode())

                    gctbl_section = ''.join([
                        chr(0xFF), 
                        chr(0xFF), 
                        chr(0xFF)])
    

                    fh.write(gctbl_section)
    
                # Write image descriptor block
                fh.write(struct.pack('<BBHHHHHH',
                    0x2C,                    # Image Separator
                    frame_img_left,          # Left Position (pixels)
                    frame_img_top,           # Top Position (pixels)
                    frame_img_width,         # Width (pixels)
                    frame_img_height,        # Height (pixels)
                    (bool(frame_local_ctbl))*8,# Local Color Table Flag (+ bitwise mask)
                    frame_dispos,            # Disposal Method
                    min(max(frame_delay, 2), 63)# Frame Delay Time
                    ))

    
                # Write LZW compression
                numcodes = 2**(min(max(d['lzwcode'], 2), 12)-1)
                maxbits = 2**numcodes
                    
                clearcode = 1<<numcodes
                endofinfo = clearcode+1
                nextcode = maxbits+1
            
                codesize = int(math.ceil(math.log(maxbits)/math.log(2)))
                usebitstream = [0]*(4 + (maxbits+1)*2)
            
                # Clear code
                usebitstream[0] = 1                                        
                for j in range(numcodes):
                    usebitstream[j*2+1] = chr(j % 256)                 
    
                count = 0