#!/usr/bin/env python3
"""
Random GIF generator
====================
Generates a random GIF file into `out`. This is useful to test that
readers can handle malformed files even if they are not fully compliant.
"""


__author__    = "<NAME>"
__copyright__ = "Copyright © 2017 <NAME>"
__license__   = "MIT"

import sys
import os
import math
import struct
import argparse
from typing import BinaryIO, Iterator, Union, Tuple, AnyStr

def _iter_chunks(stream: BinaryIO, chunksize: int)\
        -> Iterator[AnyStr]:

    while True:
        # Read in chunks of size at most MAXLEN bytes..
        thischunk = stream.read(chunksize)

        yield thischunk

        if len(thischunk)!= chunksize: break



def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c','--chunk-size', type=int,\
                        default=4*10**5, help="""Size (in bytes) for \
the chunks read from /dev/urandom when generating the initial GIF frame""")
    return parser.parse_args()




def main():

    args = parse_args()

    # Random access to /dev/urandom
    urand = open('/dev/urandom', 'rb')


    # Header for the generated GIF file.
    header = b'GIF89a'

    # The following values were hand-picked from some tests
    width = 360
    height = 180

    flags = (
            1 << 7 | # Global Color Table Flag set
            7       # Pixel aspect ratio = 1:1
    ).to_bytes(1, byteorder='little', signed=False)



    # Build up the palette
    color_table = [b'\x00\x00\x00'] * 256
    for i in range(len(color_table)):
        color_table[i] = b'\xff' * 3

    # Build up the image descriptor and output it
    local_flags = ((width & 0xFF)              .to_bytes(1,'big'))
    local_flags += ((height >> 8)              .to_bytes(1,'big'))
    local_flags += (((height) & 0xFF)          .to_bytes(1,'big'))
    local_flags += (8                         .to_bytes(1,'big'))      # Local color table flag
    local_flags += (0                         .to_bytes(1,'big'))      # Interlace == no
    local_flags += (0                         .to_bytes(1,'big'))      # Sorting == no
    local_flags += (0                         .to_bytes(1,'big'))      # Reserved bits used as transparancy value

    image_descriptor = b''.join([
        b'!' +                  # Image separator
        b'ScreenDescriptor',    # Block label
        flags,                  # Flags
        width                  .to_bytes(2,'little'),     # Width
        ((height) >> 8)        .to_bytes(1,'little'),     # Height
        b'\x08',                # Bit depth (8bits per pixel)
        0                      .to_bytes(1,'little'),
        b'',                    # Default LZW minimum size
        0                      .to_bytes(2,'little'),
        ] + list(map(lambda x: x, color_table)) + [local_flags])

    print("Writing an image with dimensions {}x{}".format(width,height))


    with open('rand.gif', 'wb+') as outfile:
        outfile.write(header)                         # Write header
        outfile.write((width).to_bytes(2,'little'))   # Write screen descriptor
        outfile.write(((height)>>8).to_bytes(1,'little'))
        outfile.write((((height)&0xFF)).to_bytes(1,'little'))
        for chunk in _iter_chunks(urand, args.chunk_size):
            outfile.write(image_descriptor)            # Write image descriptor
            outfile.write(chunk)                        # Write the randomly-generated pixels!
        outfile.seek(-1, os.SEEK_END)                 # Seek back by one byte
        outfile.truncate()                            # Truncate the last byte we wrote
    exit(0)
if __name__=='__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass