#!/usr/bin/env python3

import struct
from pathlib import Path
from typing import BinaryIO

IMAGE_WIDTH = 512
IMAGE_HEIGHT = 512
COLORMAP_SIZE = (4 ** 8)
DATA_SIZE = IMAGE_WIDTH * IMAGE_HEIGHT // 64
XPOS = 0
YPOS = 0


# Returns a 7 bit encoded color from the given sequence and offset
def encodeColor(i: int):
    if i < COLORMAP_SIZE - 1:
        return bytes([i])
    elif i == COLORMAP_SIZE - 1:
        return b'\x00'
    else:
        raise ValueError("Invalid color")
        

# Encodes the given string as a series of length-prefixed characters.
def putString(s: str):
    print("Encoding", s)
    for c in map(ord, list(s)):
        yield struct.pack('B', len(c))
        yield struct.pack('B', c)



# Constructs a palette with all adjacent colors
palette = [encodeColor(x) for x in range(COLORMAP_SIZE)]
print(f"Palette size {len(palette)}")

with open('/dev/urandom', 'rb') as rng:

    # First write one frame containing only the palette
    buffer = BytesIO()

    # Global headers
    buffer.write(b'GIF89a')
    buffer.write(struct.pack('BBBxxx', 1, 0, 0))
    
    # Logical Screen Descriptor
    buffer.write(struct.pack('<HHHHBB',
        IMAGE_WIDTH,
        IMAGE_HEIGHT,
        1,
        0,
        0,
        0))
        
    # Color Table
    for p in palette:
        buffer.write(p)
            
    # Image Descriptor
    buffer.write(struct.pack("<BBBHHHH", 
        0x2C,
        0,
        0, 
        XPOS,
        YPOS,
        DATA_SIZE,
        0))

    # Ending marker
    buffer.write(b';')


    # Now construct some frames at random
    while True:

        # Select number of bits to randomize from
        nbits = min((rng.read(1)[0] & 0xF) + 1, 32)

        # Read bits from the random stream
        value = int.from_bytes(
                rng.read(nbits//8+1), 
                byteorder='big')
        
        # Generate the mask
        mask = sum(
                2**bit 
                for bit in range(
                    nbits,
                    32)) - sum(
                        2**(bit-nbits)
                        for bit 
                        in range(
                            0,
                            nbits)
                      )

        # Apply it on the value
        maskedValue = (value & ~mask) | ((value << nbits >> nbits) & mask) 

        # Write this new image frame to the buffer
        buffer.seek(-1, SEEK_END)
        buffer.write(struct.pack(">H", maskedValue))
        buffer.truncate()


        path = f'test{randrange(1<<32)}.gif'
        print(path)
        try: 
            path.touch()
        except FileExistsError:
            pass
        with open(path, mode='wb') as fd:    
            fd.write(buffer.getbuffer())
