#!/usr/bin/env python3

import os
import sys
import time
import random
import argparse
import struct
from typing import BinaryIO


def main():
  argparser = argparse.ArgumentParser()

  argparser.add_argument("--seed", help="Provide your own seed")
  argparser.add_argument("outputfile", nargs='?',
                         default="./out.gif",
                         help="File to write the generated image")
  args = argparser.parse_args()
  
  if args.seed:
    seed = int(args.seed)
  else:
    seed = None
    
  gen_rand_gif(seed, sys.stdout, filename=args.outputfile)
    

# Generates a random GIF file to stdout or to the given `filename`.
# Uses the random number generator with the given seed (`seed`, /dev/urandom by default).
def gen_rand_gif(seed, destination_stream, filename=None):
  if filename:
    print(f"Generating {filename}...")
    dest_stream = open(filename, "wb+")
  elif hasattr(destination_stream, 'buffer'):
    dest_stream = destination_stream
  else:
    raise TypeError("The given destination is neither a writable binary stream nor a filename.")

  if seed!= None:
    print(f"Using provided seed {seed}.")
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)

  # Generate the header and add to output
  header = b'GIF89a'
  dest_stream.write(header)

  # Get the logical screen descriptor (LSD) field
  dimension = [random.randint(16,255),
               random.randint(16,255)]
  lsd = bytes([
      0x2c,                    # Image separator 
      dimension[0],            # Width 
      dimension[1],            # Height 
      0xf7 | (random.randint(1,2)*8)])    # Local colour table flag + interlace flag 
  dest_stream.write(lsd)

  # Randomly choose whether to include a global colour table
  do_global_colourtable = random.choice([True, False])
  if do_global_colourtable:
    # Pick a colour table length between 2^n and 2^(n+1)-1 
    # where n is uniformly selected at random
    gclen = min(pow(2, random.randint(1,7)), pow(2, random.randint(1,7))+1)
    gclengthbytesize = max(1, floor(log(gclen, 2)/8)+1)
    gcfulllength = 2 ** (gclen * 3)

    # We only support truecolor so RGB will always be 0
    gcte = (b'\xff\xff\xff\x80')[:gclengthbytesize]
    gc = os.urandom(gcfulllength)
  
    entry_format = f'<{gclengthbytesize}s'
    gcte += struct.pack('<H', gclen)
    dest_stream.write(gc)

  # Now start writing blocks until we see the terminator block
  extensionblockwritten = False

  while True:

      # Generate a new block
      blocktype = random.randint(1,7)
      blocksize = random.randint(10,125)
      
      # First bit indicates whether we've got an image, we don't
      # but that doesn't matter for now

      # Second bit indicates whether the block ends on a byte boundary
      endsofblockboundary = random.choice([False, True])

      # Third bit indicates whether we have a global colour table
      useglobalcolortable = random.choice([False, True])
  
      # Fourth bit indicates whether we have a local colour table
      uselocalcolortable = random.choice([False, True])

      # Fifth bit is reserved and must be set to 0
      isreservedsettozero = 0 
  
      # Sixth bit indicates whether the image uses transparency
      havetransparency = random.choice([False, True])

      # Seventh bit indicates whether the image has user input
      # In practice we just ignore all user input features so we'll never set this
      doesimagehaveuserinput = False
    
      # Eighth bit indicates whether the image needs sorting
      hassorting = False

      # Combine bits into fieldbyte
      fieldbyte = \
                (((not useglobalcolortable)<<7)|
                 ((not uselocalcolortable)<<6)|
                 ((isreservedsettozero)<<5)| 
                 ((havetransparency)<<4)| 
                 ((doesimagehaveuserinput)<<3)|
                 ((hassorting)<<2)) 
    
      # Get padding size based on byte boundary setting
      padsizeforboundarycheck = (-blocksize)%8;
    
      # Make sure we're aligned correctly on byte boundaries even without
      # trailing zeros since some viewers might choke on images with bad byte alignment
      if padsizeforboundarycheck>0:
        blocksize+=padsizeforboundarycheck
    
      fieldbyte |= len(gc)
      
      block = bytes([
                  fieldbyte,      # Field byte 
                  blocksize
          ])
      
      # Write block header
      dest_stream.write(block)
  
      # If the block has a local colortable then add one as well
      if uselocalcolortable:
         entry_format = f"<{(blocksize-1)//3}s"
         entries = os.urandom(blocksize-1)
      else:
         entries = os.urandom(entry_format)
        
      # Use the same format when we read them back later
      dest_stream.write(entries)    

      # Check if this is the last block
      if ((fieldbyte & 128) == 0):
        # We want to insert an empty graphics control and 
        # plain text extension after every single frame otherwise
        # these are ignored by most viewers
        if not extensionblockwritten:
            dest_stream.write(b'!\x00\x00\x00\x00,\x00\x00\x00')
            dest_stream.write(b'Comment\x00\x00\x00\x00')
            dest_stream.write(b'Generated by https://github.com/asweigart/gif-generator\n')
            dest_stream.write(b'https://www.asweigart.com/\n')

            extensionblockwritten = True
            
        break
    
  
