#!/usr/bin/env python3

import argparse
import struct
from typing import BinaryIO


# Generates a random GIF file with the given size (`size`) and color depth (to use in the palette) (`colordepth`), into `out`.
def generate_random_gif(
        colors: int, width: int, height: int, frames: int, background: int,
        out: BinaryIO, progress: bool=False):

    # TODO make this more efficient by reading from /dev/urandom once and only as much as needed to fill up the GIF image buffer
    def random_block():
        return bytes([next(rng) for _ in range(0, 4)])


    palette = [background]
    while len(palette) < pow(colors, 2):
        current = next(rng)

        if not any((c[0] - c[1]) % colors == 0 for c in zip(palette[-colors:], palette)):
            continue
    
        palette.append(current)

    # The first byte is an unsigned integer that specifies the version of the GIF format.
    # This value should always be 0x87, which is what we are generating.
    # The second byte specifies whether or not there is global color table flag set. 
    # If it's set then the lower three bits indicate how many times the background color index can
    # be re-used before another color index must be used instead.
    # Otherwise, the least significant bit indicates whether a local color table follows.
    # For now, we're keeping the two flags at zero, so we just write 0x00.
    out.write(b'\x47\x49\x46\x38\x39\x61\x00\x21')
    # Next comes the logical screen descriptor block, which contains information about the overall dimensions
    # of the image followed by its color table. We need to calculate the length of this based on our parameters
    # once everything is done.
    #   Size of Logical Screen Descriptor Block: 7 bytes
    #   Color Table Entry Length:              1 byte
    #   Global Color Table Flag:               1 byte
    #   Background Color Index:                1 byte
    lsdlen = 7 + 1 + 1 + 1
    # Next come the global color table (if applicable). It's specified by the logical screen descriptor block.
    # It needs one entry for each unique color in the image plus one extra to specify transparency.
    # So if we have n colors, then it has n+1 entries. We don't know yet how many colors we'll need since
    # we may find a duplicate when trying to add one. But because it starts out empty and ends up filled completely
    # we can calculate how long it is right here.
    gctlen = ((len(palette)-1) * 3) + 1
    # Finally, we have the remaining parts of the logical screen descriptor. It takes 5 bytes total;
    #    Width:                  2 bytes
    #    Height:                 2 bytes
    #    Pixel Aspect Ratio:     1 byte
    #    Global Color Table Size: 1 byte (same thing as above, but different name)
    lsdelen = 5
    # Now we know enough to determine the length of the entire logical screen descriptor, which is basically
    # the sum of all these lengths combined.
    lsdlen += gctlen
    lsdlen += lsdelen
    lsd = struct.pack('<BHHBBBBB', lsdlen-69, width, height, 0x00, 0x00, 0x00, 0xFF, len(palette)-1)
    out.write(lsd)
    
    # Write the palette contents to the output stream
    palette += [(0xff, 0xff, 0xff)]*(pow(2, colordepth)-len(palette))
    for rgb in palette:
        out.write(rgb[0])
        out.write(rgb[1])
        out.write(rgb[2])

    for fnum in range(frames):
        frame_size = 0
        
        # Frame Control Flags
        # We're setting the "Transparency" flag here, which means that any pixels with transparent colors
        # will be set to fully opaque black rather than whatever they were originally.
        # Since we've generated a black-ish background already, this should result in us getting a solid
        # black square instead of some sort of checkerboard pattern, which is cool.
        # BUGBUG this doesn't seem to work correctly on GIMP though, I think maybe they have a bug? Need to verify.
        # TODO try setting the interlace flag too, see if that changes anything?
        fcflags          = 0x09
        fcflags_byte     = struct.pack('B', fcflags)[0]
        frame_size       += 1
        # Image Descriptor
        # Next comes the image descriptor block.
        # The first byte describes the left position of the image inside the canvas:
        idleft            = 0x00
        idleft_byte       = struct.pack('B', idleft)[0]
        frame_size       += 1
        # The second byte describes the top position of the image inside the canvas:
        idtop             = 0x00
        idtop_byte        = struct.pack('B', idtop)[0]
        frame_size       += 1
        # The third byte describes the width of the image:
        idwidth           = width >> 8
        idwidth_byte      = struct.pack('B', idwidth)[0]
        frame_size       += 1
        # The fourth byte describes the height of the image:
        idheight          = height & 0xFF
        idheight_byte     = struct.pack('B', idheight)[0]
        frame_size       += 1
        # The fifth byte describes the color depth of the image:
        idcolordepth      = colordepth
        idcolordepth_byte = struct.pack('B', idcolordepth)[0]
        frame_size       += 1
        # NOTE: This field is actually unused, but is required anyway according to the spec!
        _                 = 0x00
        _byte            = struct.pack('B', _)[0]
        frame_size       += 1
        # Next comes the LZW minimum code size...
        lzws              = 0x08
        lzws_byte         = struct.pack('B', lzws)[0]
        frame_size       += 1
        #...followed by the image data block itself.
        lzwdata           = get_lzw_image()
        lzwdata_length    = len(lzwdata)
        frame_size       += lzwdata_length
        fd = struct.pack('<BbbbbBBBBB{}s'.format(lzwdata_length),
                         fcflags_byte, idleft_byte, idtop_byte,
                         idwidth_byte, idheight_byte, idcolordepth_byte, _,
                         lzws_byte, frame_size, lzwdata)
        out.write(fd)
        out.write(lzwdata)
        pinfo("Wrote frame {}/{}".format(fnum+1, frames))

    # Mark this as a loopable animation, likely won't really work
    # Well, kinda works, you'd just have to play around with the delay time and the frame count for your particular case
    numloops = 0xFFFF
    out.seek(10)
    out.write(struct.pack("<H", numloops))

    # Set both delays to 10ms (1/100 seconds)
    # You might want to change this depending on what you do afterwards in terms of editing software support
    delaytime = 0x0A
    out.seek(13)
    out.write(struct.pack("<B", delaytime))
    out.seek(14)
    out.write(struct.pack("<B", delaytime))



# Returns the LZW encoded image data for the RGB image described by the parameters below.
# Colors: is an array of tuples representing distinct colors in the form "(R, G, B)" where R, G, and B are integers between 0 and 255 inclusive.
# Width: the width of the image in pixels.
# Height: the height of the image in pixels.
# Color Depth: the color depth of the returned image data. This controls the maximal number of unique colors that can be stored per pixel.
def get_lzw_image(colors=[(255, 255, 255)], width=0, height=0, colordepth=8):
    MAXCOLORS    = 2 ** colordepth
    DICTIONARY   = []
    IMAGEDATA    = bytearray()

    # Add all the distinct colors to the dictionary initially.
    # The last tuple element is True since the colors have been added to the dictionary.
    for i in range(MAXCOLORS):
        red   = colors[i][0]
        green = colors[i][1]
        blue  = colors[i][2]
        tup   = (red, green, blue, True)
        DICTIONARY.append(tup)

    pos  = 0
    val  = None

    while pos <= (height*width):
        # Get the next color from the image data
        r = next(rng) % colors.__len__()
        g = next(rng) % colors.__len__()
        b