#!/usr/bin/env python3

from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
# This will be truncated to exactly 2048 bytes before being written as it is not clear if
# that's useful or not
def generate_random_gif(
        rng: BinaryIO = None, out: BinaryIO = None, max_size=2048):

    # Generate a random giffer
    def gen():

        header = b"GIF89a"
        frame_count = rng.getrandbits(16)     # Number of frames in animation
        width       = rng.getrandbits(16) - 1  # Screen size (in pixels)
        height      = rng.getrandbits(16) - 1  # Screen size (in pixels)
        color_table = [
                rng.getrandbits(8) for _ in range(2 ** int(log(frame_count, 2)))]
        bg_color    = rng.getrandbits(8)        # Background colour index
        aspect      = rng.getrandbits(8) + 1   # Pixel aspect ratio (non-square pixels!)

        yield from header.to_bytes(6, "little")
        yield from frame_count.to_bytes(2, "little")
        yield from width.to_bytes(2, "little")
        yield from height.to_bytes(2, "little")
        yield from len(color_table).to_bytes(1, "little")
        yield from bg_color.to_bytes(1, "little")
        yield from aspect.to_bytes(1, "little")
        yield from b"".join([b'\x00\x00' + c for c in color_table])

        # Now we build up each frame followed by a terminator so don't bother wasting space

        # The time delay between frames
        delay = rng.getrandbits(16) * 10 // 100
        yield from delay.to_bytes(2, 'little')

        flags = 0x80 | rng.getrandbits(7)
        yield from flags.to_bytes(1, 'little')

        # Frame width and height are stored seperately due to their variable nature
        yield from width.to_bytes(2, 'little')
        yield from height.to_bytes(2, 'little')

        # LZW min code size
        lzw_code_size = rng.randint(2, 8)
        yield from lzw_code_size.to_bytes(1, 'little')

        # Put all the image into one big string
        img = []
        while True:

            # Get a random length for a chunk
            split_len = rng.randint(2, 255)

            # If we can't fit another split point at this location then stop here
            if len(img) + split_len > max_size:
                break

            # Otherwise add some more data
            img += [(rng.getrandbits(8)).to_bytes(1, "little")]

        # Add an end symbol for the stream
        img.append(b"\xff")

        # Now write the whole lot into the output
        yield from b"".join(img)


    # Write the entire result to our destination
    if out is not None:
        out.write(b"".join(gen()))
    return b"".join(gen())
