#!/usr/bin/env python3

import struct
from typing import (BinaryIO, Tuple)

# Generate a 24-bit RGB color from three bytes
def rgb(r1: int, g1: int, b1:int):
    r0 = (r1 & 0x1F) << 3
    g0 = (g1 & 0x1F) << 3
    b0 = (b1 & 0x1F) << 3

    return ((r0 | g0 | b0) & 0xFF, )

# Generates a random GIF file with the given dimensions and
# palette entry count using the provided random number generator
def generate_random_gif(
        rng: BinaryIO, width: int, height: int, num_colors: int,
        out: BinaryIO
    ):
    # Set up the basic GIF metadata
    out.write(b'GIF89a')
    # Write the logical screen descriptor block
    out.write(struct.pack('<HHBBBBB', width, height, 7, 0, 0, 0))
    
    # Select the appropriate color table size for the declared number of colors
    if num_colors <= 2**1:
        global colortablesize    
        colortablesize = 1
    elif num_colors <= 2**2:
        colortablesize = 2
    elif num_colors <= 2**3:
        colortablesize = 3
    else : 
        assert False, 'Too many colors!'

    # Write some color entries to represent the desired palette
    for i in range(num_colors):
        colorbyte = rng.read(colortablesize)

        if len(colorbyte) < colortablesize:
            # Pad with zeroes as necessary...
            otherbytes = bytes([0] * (colortablesize - len(colorbyte)))
            colorbyte += otherbytes
        
        out.write(rgb(*colorbyte))
        
    # Calculate the data length of the image portion
    num_pixels = width*height
    datalen = num_pixels//2
    if num_pixels % 2!= 0:
        datalen += 1

    out.write(struct.pack("<B", colortablesize * 3))
    out.write((struct.pack("<H", datalen)).ljust(3, b'\0'))

    # Now write the actual image data!
    while True:
        chunk_len = min(datalen, 256)
        out.write(struct.pack("B", chunk_len))

        # Encode pixels in chunks of two bytes
        for _ in range(chunk_len // 2): 
            blk = []
            
            for _ in range(2):
                # Weirdly, we need to encode each pixel pair with its red value 
                # first before moving on to the green value. This is because 
                # otherwise there are more instances where this code will go 
                # out of bounds when trying to select a random color index.
                # It's probably not an optimal solution but it works well enough
                # for now.
                blk.append(rng.read(1)[0])

            missingpixcount = pixcount = len(blk[0])//3

            for j in range(missingpixcount):
                idx = rng.read(1)[0]
                pixelpart = blk[idx][j * 3:]

                for k in range(3):
                    blk[k].append(pixelpart[k])
                
            encodedblock = b''
            for part in blk:
                val = int.from_bytes(part[:3], byteorder="big")
                encodedblock += struct.pack('>I', val)


            out.write(encodedblock)
            
            datalen -= chunk_len

            if datalen == 0: break