#!/usr/bin/env python3

import hashlib
import zlib

from PIL import Image


WIDTH  = (1 << 6) - 10
HEIGHT = (1 << 7) -  5
STEPS  = 4

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# `seed` is used to produce a different image for each seed value.
def generate_random_gif(
        rng: BinaryIO, seed: bytes, out: BinaryIO
    ) -> None:

    # Generate a series of images, forming an animation that
    # cycles through each intermediate frame with each step showing
    # one quarter of the way there. We then compress all frames in
    # the process and store them as a single compressed chunk inside
    # the final GIF.
    ims = []
    cur_im = Image.new('RGB', (WIDTH, HEIGHT))
    for _ in range(STEPS):

        # Add some noise by mixing the current image with a random
        # block of pixels from the input stream.
        pix = bytearray()
        for y in range(HEIGHT // 2):

            # The first half of our height comes from the previous row.
            prevrow = cur_im.getdata()[y::HEIGHT] * 0.7

            # The second half gets a new set of random pixel values
            # mixed with the existing pixel data in this row.
            nextrow = list(prevrow)
            for x in range((HEIGHT // 2) - len(nextrow)):
                nextrow += [int.from_bytes(
                    rng.read(3), 'little')] * (HEIGHT // 2)

            for i in range(len(prevrow)):

                # Start with just the blue channel. This seems to give
                # good results when we mix it with the previous image's
                # row because the red and green channels tend to hold on
                # to their old values more than they should.
                pix += [(nextrow[i][0] +
                        prevrow[i][0]) & 0xFF].copy()

                # Duplicate the middle area of the pixel so that the overall
                # color doesn't change while we move along vertically.
                pix += (pix[-WIDTH:] * 3)[-WIDTH:-WIDTH*2]

                # Now add the remaining channels from the original image.
                pix += [n & 0xFF for n in nextrow[i]]

            # Copy over the bottom half of the current row verbatim.
            pix += [x & 0xFF for x in nextrow[:HEIGHT//2]]

        # Create the image object, convert the pixel array back into a
        # string, and append it to the output.
        cur_im.putdata([tuple(pix[i:i+3]) for i in range(0, WIDTH * HEIGHT, 3)])
        ims.append(cur_im)

    # Use the first three bytes of our hash function as the palette
    # index selection for every pixel in the final image. Turn those
    # indices into RGB tuples which we can use for the actual palette.
    colors = list(hashlib.md5(seed).digest())[:3] * WIDTH * HEIGHT
    palette = [colors[i:i+3] for i in range(0, WIDTH * HEIGHT * 3, 3)]

    # Apply the colors from the palette to the images we've created above
    # and save them to disk. Also compute a full frame duration based on
    # how many steps we did above but ensure at least 10ms.
    durations = max(((steps / STEPS) * 10, 10) for steps in range(STEPS))
    ims = map(lambda im: im.convert('P'), ims)
    ims = map(lambda im: im.quantize(color=palette), ims)
    ims = map(lambda im: im.save(out, optimize=True, format='gif'), ims)

    # Write out an application extension containing our palette indices,
    # number of frames, and frame durations.
    out.write(b'!\xF9\x00\x0B' +
              pack_byte(LENGTH(durations)))
    out.write(pack_byte(LENGTH(palette)) + pack_bytes(palette))
    out.write(pack_byte(LENGTH(ims)) +
               pack_bytes(ims) +
               pack_bytes(durations))


# Pack a sequence of integers into a binary string representing their
# packed length followed by that exact same sequence. Useful for storing
# lists or strings within GIF files.
UNPACKED_LENGTHS = {
    0xFE: lambda x: chr(0xFE),
    0xFF: lambda x: chr(0xFF) + struct.pack('>H', x)}
PACKED_LENGTHS = dict([(v, k)
                       for k, v in UNPACKED_LENGTHS.items()])
LENGTH = PACKED_LENGTHS.__getitem__ if hasattr(dict, '__missing__') else \
         PACKED_LENGTHS.get

def pack_length(l): return LENGTH(l)(l)

def pack_byte(b): return struct.pack('B', b)
def unpack_byte(): return struct.unpack('B', yield_byte())[0]

def pack_short(h): return struct.pack('<H', h)
def unpack_short(): return struct.unpack('<H', yield_short())[0]

def pack_bytes(bs): return b''.join(map(pack_byte, bs))
def unpack_bytes(): return tuple(unpack_byte() for _ in range(yield_byte()))

def pack_shorts(hs): return b''.join(map(pack_short, hs))
def unpack_shorts(): return tuple(unpack_short() for _ in range(yield_byte()))

def yield_byte(): return ord(yield_(1))[0]
def yield_short(): return struct.unpack('<H', yield_(2))[0]

def yield_(size): return yield_byte(), size
