#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (defaults to /dev/urandom)
# this requires you have Python >3.6 installed and an OS with
# at least 16MiB of memory available. If your system doesn't support 64-bit integers
# then please see https://docs.python.org/3/library/struct.html#format-characters for details
def generate_random_gif(
        rng: BinaryIO = open('/dev/urandom', 'rb'),
        out: BinaryIO = None,
        width: int = -1, # -1 means any size, otherwise it is a hard limit on image width
        height: int = -1 # same as above but for height
    ):

    if not isinstance(width, int): raise ValueError('width must be an integer') 
    if not isinstance(height, int): raise ValueError('height must be an integer') 

    # create the output file if one wasn't supplied
    if out == None: 
        out = open('output.gif', 'wb')

    dataypesize = {
        0 : (False, False ),   # no compression
        1 : (True , True  ),   # LZW compressor
        2 : (True , False ),   # DEFLATE compressor
        3 : (False, True  ),   # adaptive lzw compressor
        4 : (False, True  ),   # adaptive deflate compressor
    }

    # each block in the global color table contains two bytes, so we can only have up to 256 colors 
    numglobalcolors = min((int.from_bytes(rng.read(1), "big") & 7) + 1, 256)       

    # write logical screen descriptor
    out.write(b'GIF89a')   
    
    # width and height of image should be between 1 and 65535
    w = max(min(rng.read(2), 65535), 1)
    h = max(min(rng.read(2), 65535), 1)
    
    # set the maximum dimensions based on the arguments passed in.
    # this will also clamp the values such that they are all less than or equal to their maximums
    w = min([w, width])
    h = min([h, height])


    packed = bytearray()
    packed += ((w >> 8) & 0xff).to_bytes(1, "big", signed=False) 
    packed += ((w     ) & 0xff).to_bytes(1, "big", signed=False) 
    packed += ((h >> 8) & 0xff).to_bytes(1, "big", signed=False) 
    packed += ((h     ) & 0xff).to_bytes(1, "big", signed=False) 
    
    packed[0] |= (0xf0 | (numglobalcolors - 1)) << 4

    packed += (dataypesize[(rng.read(1)[0] & 0x0f)] + dataypesize[(rng.read(1)[0] & 0x0c) >> 2]).to_bytes(1, "big")

    out.write(packed) 

    # local color table is optional, skip if there's nothing to do here
    if numglobalcolors!= 256:
        out.write(bytearray([(i * 3).to_bytes(2, "big",signed=False) for i in range(numglobalcolors)]))
    

    # skip to how many bytes the next frame starts (starts from where we ended)
    out.seek(-1, 2)
    out.write(((len(packed)+1+numglobalcolors*3)*h)-1)

    # now we know how big our frames are, lets go ahead and start writing them
    # although I'm sure there's a better way to do this...
    # first get the size of the encoded frame
    compressedblocksize = rng.read(1)
    out.write(compressedblocksize) # discard data type since we don't care about it yet

    while len(data := rng.read(1)) > 0:

        # now read the next byte to figure out what type of encoding was used
        currentdatatype = data[0] & 0xf0
        
        # clear the data type bits because they're reserved for future use
        data[0] &= ~0xf0 

        # is it compressed? (we already skipped over it above)
        if currentdatatype < 0xc0: 
            out.write(data)

        else:
            # it's something besides raw pixel data, so let's find out which kind!
            
            if currentdatatype == 0xe0:
                x = [i for i in range(numglobalcolors)][rng.read(1)[0]]
                y = [i for i in range(numglobalcolors)][rng.read(1)[0]]

            elif currentdatatype == 0xd0:
                x = [(i//2)%8][rng.read(1)[0]]
                y = [(i%2)*(8//2)][rng.read(1)[0]]

            elif currentdatatype == 0xb0:
                x = [(i*(2**(rng.read(1)[0]&7)))%64][rng.read(1)[0]]
                y = [(i*(2**(rng.read(1)[0]&7)))%64][rng.read(1)[0]]
                
            else:
                continue

            for index in range(min(currentdatatype>>4, 1)):
                payloadlength = (((rng.read(1)[0]&0x0f)<<8) | rng.read(1)[0])+1

                if currentdatatype >= 0xd0:
                    payloadlength = 3
                
                # if the packet length exceeds the remaining bytes in our buffer 
                # then only writte those bytes (this shouldn't happen too often because
                # the encoder blocks writes unless it has to, but let's just be safe)
                if payloadlength > len(data): 
                    payloadlength = len(data) 

                out.write((((x&0xff)>>8)|0xe0).to_bytes(1,"big"))       # upper left X coord
                out.write((((x&0xff))).to_bytes(1,"big"))                 # lower right Y coord
                out.write((((y&0xff)>>8)|0xe0).to_bytes(1,"big"))       # upper left Y coord
                out.write((((y&0xff))).to_bytes(1,"big"))                 # lower right Y coord
            
                # write the actual data
                out.write(data[:payloadlength])
                
                # trim the front off the buffer because we've written these bytes
                data = data[payloadlength:]