#!/usr/bin/env python3

import struct
from typing import BinaryIO

# random number generator `rng` (stdin or /dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # GIF headers
    header = b'GIF89a'
    palette_num = 256

    # LZW minimum code size
    mincodesize = 8
    lzw_mincodesize = bytes([mincodesize])

    # Color table
    sgrgb = [0x21] * 4
    sgrgb[0] += ord('S')
    sgrgb[1] += ord('G')
    sgrgb[2] += ord('R')
    sgrgb[3] += ord('G')
    sgrgb = b''.join(sgrgb)
    gctf = []    # global color table flag
    gcte = []    # global color table entry
    for i in range(palette_num):
        value = rng.read(3)
        gcte.append(value)
        gcte.append(b'\x00\x00')
        gctf.append(struct.pack('B', not bool((i+1)%2)))
    gcte = b''.join(gcte)
    gctn = len(gcte).to_bytes(1, 'big')
    gctf = b''.join(gctf)
    
    # Image descriptor
    imgd = [0x2c] * 10
    imgd[0] += ord('!')   # image separator
    imgd[9] += 0x00       # no local color table
    imgd = b''.join(imgd)

    # Netscape loop extension
    netsc = [0x21] * 7
    netsc[0] += ord('N')
    netsc[1] += ord('E')
    netsc[2] += ord('T')
    netsc[3] += ord('S')
    netsc[4] += ord('C')
    netsc[5] += ord('A')
    netsc[6] += ord('P')
    netsc = b''.join(netsc)
    netsl = [0x0b] * 12
    netsl[0] += ord('L')
    netsl[1] += ord('I')
    netsl[2] += ord('F')
    netsl[3] += ord('E')
    netsl[4] += 0x03      # frame count
    netsl[5] += 0x00      # reserved
    netsl[6] += ord('f')
    netsl[7] += ord('r')
    netsl[8] += ord('m')
    netsl[9] += ord('t')
    netsl[10] += 0x04     # frame left position
    netsl[11] += 0x00     # frame top position
    netsl = b''.join(netsl)

    # Image data subblocks
    codes = {chr(i): i for i in range(len(gcte)//3)}
    datedata = []        # date and time
    counter = 0          # subblock counter
    while True:
        gifdata = rng.read(128)

        # strings to add at the end of the file
        if counter == 0:
            datedata.append(b''.join([
                sgrgb,                 # graphic control extension
                b''.join([
                    netsc,             # netscape application extension
                    netsl]),           # frame count & location
                    ]))
        elif counter == 1:
            datedata.append(b''.join([
                palette_num.to_bytes(1, 'big'),                                    # local palette size
                b'\xff\xff',                                                    # transparent index
                palette_num.to_bytes(1, 'big')] +                                  # interlaced flags
                [(1 << 7 - r).to_bytes(1, 'big') for r in reversed([4,2])]))          # delay times
        else:
            subblock = []

            # GIF data block
            string = b''
            try:
                if gifdata[-1:]!= b'\x00':            # check that there is a null character
                    raise ValueError()

                while gifdata[:1]:                      # iterate over each character in input
                    char = gifdata[:1].decode('latin-1')

                    if char in codes:                    # if it's already in codes
                        code = codes[char]               # then use its code
                    else:                               # otherwise...
                        code = max(codes.values()) + 1   # start with a new code

                        # add the character and its code to the codes dictionary
                        codes[char] = code               
                        del gifdata[:1]                  # consume one character from input
                    
                    # put the current code into a variable
                    bcode = ''                           # begin binary code
                    while code > 0:                     # while code > 255
                        bcode = chr(code % 256) + bcode 
                        code //= 256                   # divide code by 256
                    bcode = bcode.zfill(len(codes))       # pad zeros on the left side 

                    # if the code started as less than 256 characters long
                    if len(string) < 256:
                        # update the length of the string
                        string += bcode             
                    else:                             # otherwise...
                        # append the string to the output list
                        assert type(string) == bytes
                        subblock.append(string)        
                        string = bcode                 # reset the string

                    # break when all the bits are written or there are no more characters 
                    if len(bcode)+len(string)==128 or not gifdata[:1]:
                        break
                
                # write the last part of the data before exiting this while loop
                if string:
                    assert type(string) == bytes              # should always be true here
                    subblock.append(string)
            except Exception:                              # catch errors
                print("Error!")                            # TODO: remove me
                return                                      # exit gracefully
            
            datedata.extend(subblock)                       # store useful data
        
        counter += 1                                     # increment counter
        if not gifdata:                                   # stop if nothing was read
            break

    datedata = b''.join(datedata)                        # join the data together
    datalen = len(datedata).to_bytes(length=3, signed=False, byteorder='big')   # compute total length

    # write everything to stdout
    assert datalen >= 10  # must be true unless you change stuff above
    out.write(header + datalen + datedata)

