#!/usr/bin/env python3

import struct
from typing import (BinaryIO, Final, Literal, Optional)

# Returns a byte string containing an encoded integer with
# 1-4 bytes depending on its value and how many bytes it
# will take to encode it.
def encode_int(num: int, maxbits: int):
    assert num >= 0

    if maxbits <= 8: return num.to_bytes(maxbits//8, 'big')
    
    res : list[int] = []
    while True:
        part = num & ((1 << maxbits)-1)

        # If num fits in this part of the encoding, we're done!
        if not (part == 0 or
                part == 1 and len(res) > 0 and res[-1] == 0xff):
            break
        
        res.append((part >> (maxbits-8)) | 0x80)
        num >>= 8

    return bytes(reversed(res))


# Decode an encoded integer back into its original form
# given what maximum bits were used for the encoding
def decode_int(data: bytes, maxbits: int) -> int:
    assert maxbits >= 1

    res : int = 0
    shift : int = 0
    for i, byte in enumerate(data):
        if i < len(data) - 1:
            res |= (byte & ~0x7f) << shift
        else:
            res |= (byte & ~0x7f) 

        shift += 8
        if byte & 0x80!= 0: continue
            
        if maxbits == 8*len(data):
            return res
        elif maxbits < 8*i+8:
            return (-(-res >> (shift-(8*(i+1))))) & ((1<<maxbits)-1)
    raise ValueError("Not enough valid data")



# Returns a byte string containing an encoded unsigned fixed point number.
# It doesn't have to be the same length as the input, but it must never be larger, either.
# The output is always big endian. 
def encode_ufixed(num: float, frac_bits: int, max_value: Optional[float]=None) -> bytes:
    assert frac_bits >= 0
    assert num >= 0
    if max_value is None: 
        max_value = (1<<(frac_bits*8))-1
    assert num <= max_value

    ints = [0]*((len(num.as_integer_ratio())-1)*2)
    base = 1<<(frac_bits*8)
    base_roundup = math.ceil(base)

    fpart = abs(num-(num//base)*base) * (base_roundup**2)
    ipart = round(abs(num)//base)*(base_roundup)

    ints[::2] = reversed([int(i) for i in str(ipart).split('.')])
    ints[1::2] = [int(c) for c in format(int(fpart), "0" + str(frac_bits) + "b")]
    ints.reverse()

    overflow : bool = False
    for i, v in enumerate(ints):
        if overflow: 
            ints[i] = 0
        elif v >= base_roundup:
            overflow = True
            ints[i] = min(v, max_value)+1

    # Handle fractional part overflowing
    if n := sum([(b&~0xff)>>8 for b in ints]):
        print("[WARNING] Fraction overflowed by", n, "bytes")

    return b"".join([struct.pack(">B", v) for v in ints])
        

# Decode a ufixed from its representation in memory
def decode_ufixed(numstr: bytes, frac_bits: int) -> float:
    nums = struct.unpack('>'+'B'*len(numstr), numstr)
    res : float = 0
    for i, v in enumerate(nums):
        res += (v/(1<<(frac_bits*8)))
        if i%2==1: res /= 256
    return res


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width: int=320, height: int=240, loopcount: int=-1
    ):
    # Header
    out.write(b'\x47\x49\x46\x38\x39\x61')

    palette_size_pow2 = 8    #(log2 size of color table)
    delay_denominator = 10   #(delays are measured in 100th's of seconds)
    trailer_offset = 6       #(trailer offset in animation control block; can be calculated later)
    framesize = width * height

    # Logical Screen Descriptor
    lsd = struct.pack("<HHBBB", width, height, palette_size_pow2, 0, 0)
    out.write(lsd)

    # Global Color Table
    global_color_table = [struct.pack('<BBBB', *(rng.randint(0, 255), )*4)
                          for _ in range(1<<palette_size_pow2)] 
    gct_packed = b''.join(global_color_table)
    out.write(gct_packed)

    # Frame Control Block
    fc = struct.pack("<HBBBBB", framesize, 0x2c, 0, 0, 0, 0)     # no local color table
    out.write(fc)

    # Image Descriptor
    id = struct.pack("<HHHHB", 0, 0, width, height, palette_size_pow2)
    out.write(id)

    # Local Color Table
    lct = struct.pack('<BBBBBB', *[rng.randint(0, 255), ]*6)        # use global color table
    out.write(lct)

    # Graphic Control Extension blocks
    if loopcount!= -1:
        gce = b"\x21\xf9\x04" + struct.pack("<BBBBB",
                      6,                   # extension introducer
                      loopcount % 256,    # loop count
                      loopcount // 256,
                      trailer_offset % 256,# terminator
                      trailer_offset // 256
                  )
        out.write(gce)

    # Now comes the image data. This is a simple black and white RGB image.
    out.write(b''.join([global_color_table[(rng.getrandbits(1)<<2)|(rng.getrandbits(1)<<1)]
                        for _ in range(framesize)]))

    # Now write the animation metadata at the end of the file. An animation consists of a series
    # of images that are shown one after another without interruption. The duration of each image
    # depends on whether there was a GCE block before showing the next image. The last image in an
    # animation needs to specify a zero delay so that the current image stays visible until you click
    # through to the next. So it goes...
    #   | delay of previous image | "image data" | delay |...
    # Looping animations make this even more complicated since they need to loop around when finished.
    # To handle these cases I'm just going to add some extra frames with nonzero delays at the start
    # and end of the animation. (I don't know why.)
    extra_frames : tuple[tuple[int, int]] = ()
    if loopcount!= -1:
        extra_frames = (
              (loopcount+1)*delay_denominator, width * height
          ), (
              0, width * height 
          )

    del delay_denominator
    assert all(d >= 0 for d,_ in extra_frames)

    delay_encsize = lambda x : int(math.log(x) / math.log(2))+1

    # Calculate offsets within the file where to put GCE blocks. There should be one per frame, plus
    # two more for the animation control block and the ending zero-duration frame.
    gce_offsets : list[int] = [-1] + [(out.tell()+i*(trailer_offset+(delay_encsize(d))*framesize)
                                      +sum(e*framesize
                                          for e,_ in extra_frames[:i])
                                      )
                                     for i,(d,_) in enumerate(extra_frames+[(0,-1)])
                                     ]
    trailer_offset = gce_offsets[-1]+delay_encsize(-1)*framesize
    gce_offsets.remove(-1)

    # Write Animation Control Block
    ac = struct.pack("<BBHB", 0x21, 0xff, 12, trailer_offset)      # Application Extension Introducer
    ac += struct.pack("<BH", 11, 0)                                # Application ID
    ac += struct.pack("<BBBB",
                  0xfd,               # no application data
                  palette_size_pow2, # color table size                 
                  framesize,         # LZW minimum code size (aka terminating code)
                  0                 # background color index 
              )
    for i,(_,d) in enumerate(extra_frames):