#!/usr/bin/python3

import argparse
import os

from random import Random


class GifGeneratorError(Exception): pass


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('--seed', '-s', metavar='SEED',
                        help="The seed for the PRNG used.")

    return parser.parse_args()


def main():
    args = parse_args()

    if not args.seed:
        raise GifGeneratorError("You must specify --seed/-s")

    # Create our own prng so we can set its seed and make sure it's deterministic
    myrng = Random()
    myrng.seed(args.seed)

    with open('out.gif', 'wb+') as f:
        # The first byte of the header contains version information
        magic_number = 57

        header = bytes([magic_number])
        outlen = len(header)

        # Then comes the logical screen width & height
        outwidth = (1024 * myrng.randint(1, 4)) // 4
        outheight = ((1024 * myrng.randint(1, 4))) // 4
        outbpl = outwidth * 3        # Bytes per line
        outsize = 256                # We only support one color palette

        # Next comes the packed field which describes global properties of this
        # image such as background color index etc.
        bgcolidx = myrng.randint(0, 256 - 1)     # Background Color Index
        aspect = 4                            # Pixel Aspect ratio
        gctflag = False                       # Global Color Table Flag
        interlaced = True                     # Interlace Flag
        lctsize = 0                           # Local Color Table Size

        outfield = ((bgcolidx << 7) |
                    (aspect << 4) |
                    (gctflag << 3) |
                    0b011)                  # Packed Field

        outfield |= ((interlaced << 7) |
                    (lctsize << 0)).to_bytes(1, "big", signed=False)[0]   # Pack again.. :)
        header += outfield.to_bytes(1, "big", signed=False)

        outlen += 1

        # Finally comes the initial global color table which is repeated for all frames in the animation
        # There are two bits set in each entry in the global color table. For most images they
        # should both be zero but just to be safe let's initialize them correctly too...
        # NOTE: In reality there's no need to repeat these entries since we don't actually use them!
        colors = [myrng.randint(0x00, 0xFF).to_bytes(1, "big", signed=False)
                  for _i in range(0, pow(2, lctsize+1))]
        palettelen = sum((len(_) for _ in colors))
        header += palettelen.to_bytes(1, "big", signed=False)     # Palette size
        header += b'\x00\x00'                                     # Reserved

        outlen += 2
        for c in colors:
            header += c

        # Now we add more bytes equal to the previously calculated number of bytes remaining
        outlen += 1022 - outlen
        header += b'\x00' * outlen

        print("Writing the following header:")
        print(header)

        f.write(header)

        # We're done writing the header so now we start adding the actual content
        seq_num = None    # Sequence Number
        imgcnt = 0        # Image counter

        while True:
            # First off is the sequence number which identifies the current frame within this animation
            if seq_num == None:
                seq_num = myrng.randint(1, 0xFFFF)

            # Op code, fixed value 22 (this is the Start Of Frame op)
            f.write(struct.pack('<B', 22))

            # Then comes the LZW minimum code size. Set on init to default of 8-bits
            mincodesize = 8
            f.write(struct.pack("<B", mincodesize))

            # After the opcode and the minimum code size we have the dimensions of this frame
            # Here I'm assuming a maximum of 256 cols in X and Y
            xdim = myrng.randint(1, 256)
            ydim = myrng.randint(1, 256)

            # Now follows the packed fields
            useref = True              # Has a local color table been referenced?
            interlaced = bool(myrng.getrandbits(1))      # Do we want this frame to be interlaced?
            sortflag = myrng.getrandbits(1)         # Color table sorted flag
            reserved = True             # Reserved bit, always true
            sizeloccolortable = myrng.randint(0, 8)    # How many unique colors do we contain in the color table
            lctflag = myrng.getrandbits(1)           # Is there a local color table present in the frame?

            packfields = (((useref << 7) |
                           (interlaced << 6) |
                           (sortflag << 5) |
                           0b00000) |
                          (reserved << 4) |
                          sizeloccolortable |
                          (lctflag << 0)).to_bytes(1, "big", signed=False)[0]

            if sizeloccolortable > 0:
                loccolors = [[myrng.randint(0, 0xFFFFFF)]
                             for i in range(pow(2, sizeloccolortable+1))]

                # Write our local color table before continuing
                writelen = 2 << (sizeloccolortable + 1)
                writeval = sum([(l[0][0] << 16) |
                                (l[0][1] << 8) |
                                (l[0][2]) for l in loccolors])

                f.write(writelen.to_bytes(1, "big"))                        # Local color table size
                f.write(((writeval >> 56) & 0xFF).to_bytes(1, "big"))       # R
                f.write(((writeval >> 48) & 0xFF).to_bytes(1, "big"))       # G
                f.write(((writeval >> 40) & 0xFF).to_bytes(1, "big"))       # B

            f.write(packfields.to_bytes(1, "big"))                          # Packed Fields

            # Now comes the coordinates of the area on the canvas where we want the picture to live
            left = myrng.randint(0, outwidth - xdim)
            top = myrng.randint(0, outheight - ydim)
            f.write(left.to_bytes(2, "big", signed=False))                   # Left
            f.write(top.to_bytes(2, "big", signed=False))                    # Top
            f.write(xdim.to_bytes(2, "big", signed=False))                    # Width
            f.write(ydim.to_bytes(2, "big", signed=False))                    # Height

            # Then follow the delay time and the transparent color index
            delaytime = myrng.randint(0, 100)
            transcolindex = myrng.randint(0, 256)
            f.write(delaytime.to_bytes(1, "big", signed=False))               # Delay Time
            f.write(transcolindex.to_bytes(1, "big", signed=False))            # Transparent Color Index

            # Finally follow the LZW code portion of the image descriptor including the EOI statement
            maxcode = pow(2, mincodesize+1)-1
            eoltoken = pow(2, mincodesize+1)
            output = []

            # Initialize the buffer for reading codes from the input stream
            buffer = []

            def readcode():
                try:
                    return buffer.pop(0)
                except IndexError:
                    bufchar = f.read(1)

                    if bufchar == b"":
                        return -1
                    elif bufchar == b'\x00':
                        raise GifGeneratorError("Invalid file format!")
                    
                    # If the character is higher than what was expected then we ignore everything else until EOI
                    if ord(bufchar) >= eoltoken:
                        return ord(bufchar) - eoltoken

                    charval = ord(bufchar)
                    highbit = 0b10000000 >> (mincodesize-1)
                    codemask = (highbit << 1) | 0b00000001
                    lowbits = ord(f.read(1))
                    if lowbits!= 0 or highbit!= 0b10000000:
                        buffer += [(lowbits & codemask)]
                        charval = (charval << 8) | lowbits & (~codemask)
                        
                    return charval
            
            # Read the next code from the stream and check if it matches anything special
            token = readcode()
            if token < 0: break                           # End of file reached, exit loop
    
            # Output the first code directly
            output += [token]

            # Now keep processing tokens until we hit their end
            while token!= eol