#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Write the logical screen descriptor
    width = int.from_bytes(
            rng.read(2), 'big', signed=False
        ) % (2**16-1) + 1
    height = int.from_bytes(
            rng.read(2), 'big', signed=False
        ) % (2**16-1) + 1
    global_color_table = int.from_bytes(
            rng.read(1), 'big', signed=False
        ) & 0x80 > 0
    background_index = int.from_bytes(
            rng.read(1), 'big', signed=False
        ) % (global_color_table and len(COLORS) or 255+1)

    flags = bytearray([
        0x80 |        # Global color table flag
        (len(COLORS)-1 if global_color_table else 0)
    ])
    out.write(header + bytes(flags) + struct.pack('>H', width)+struct.pack('>H', height))


    if not global_color_table:
        # Write local color table
        numColors = min((width*height)//5, 256)

        # Create an array of colors with no duplicates (and at least one entry)
        uniqueColors = set()
        while True:
            r = max(0, min(int.from_bytes(
                rng.read(1), "big", signed=False
            ), 255))
            g = max(0, min(int.from_bytes(
                rng.read(1), "big", signed=False
            ), 255))
            b = max(0, min(int.from_bytes(
                rng.read(1), "big", signed=False
            ), 255))

            newColor = [r,g,b]
            if tuple(newColor) in uniqueColors:
                continue
            uniqueColors.add(tuple(newColor))
            COLORS += newColor
            break

        # Write the color table to the output file
        for i in range(numColors):
            out.write(COLORS[i])



    # Write the image descriptor
    left = int.from_bytes(
            rng.read(2), 'big', signed=False
        ) % ((2**16-1)-(width+1))+1
    top = int.from_bytes(
            rng.read(2), 'big', signed=False
        ) % ((2**16-1)-(height+1))+1
    flip = bool(int.from_bytes(
            rng.read(1), 'big', signed=False
        ) & 0x40 > 0)
    delayTime = int.from_bytes(
            rng.read(2), 'big', signed=False
        ) % (2**16-1)
    transparentIndex = int.from_bytes(
            rng.read(1), 'big', signed=False
        ) % (global_color_table and len(COLORS) or 255+1)
    interlaceFlag = bool(int.from_bytes(
            rng.read(1), 'big', signed=False
        ) & 0x20 > 0)
    
    flags = bytearray([(not flip)<<7, (interlaceFlag)<<6])
    out.write(bytes(flags) + struct.pack('>H', left)+struct.pack('>H', top))
    out.write(struct.pack('B', delayTime))
    out.write(bytes([transparentIndex]))
    
    # Write compressed LZW data
    lzwMinCodeSize = int.from_bytes(
            rng.read(1), 'big', signed=False
        ) % 8+1 # 2^n+1 where n is 0-7
    codeSize = lzwMinCodeSize
    clearCode = 1 << codeSize
    endCode = clearCode + 1
    nextCode = endCode + 1
    freeCode = nextCode + 1
    bitmask = (2 ** codeSize) - 1
    maskDelta = pow(2, 13) // nextCode
    currentDataByte = 0
    currentBitsLeftInThisByte = 8

    def flushCurrentByte():
        nonlocal currentDataByte, currentBitsLeftInThisByte
        out.write(bytes([currentDataByte]))
        currentBitsLeftInThisByte = 8
        currentDataByte = 0

    def addDataToOutputBytes(byte):
        nonlocal currentDataByte, currentBitsLeftInThisByte
        currentDataByte |= byte << (8 - currentBitsLeftInThisByte)
        currentBitsLeftInThisByte -= 8

        while (currentBitsLeftInThisByte < 8):
            flushCurrentByte()
    

    def emit(item):
        nonlocal nextCode, bitmask, maskDelta, codeSize, freeCode, \
                    addDataToOutputBytes, flushCurrentByte
        
        if item == CLEAR:
            addDataToOutputBytes(clearCode >> 8)
            addDataToOutputBytes(clearCode & 255)
            codeSize = lzwMinCodeSize
            bitmask = (2 ** codeSize) - 1
            freeCode = nextCode + 2
            return

        elif item == END: 
            addDataToOutputBytes(endCode >> 8)
            addDataToOutputBytes(endCode & 255)
            flushCurrentByte()
            return
        
        elif freeCode >= (1<<codeSize):
            addDataToOutputBytes(CLEAR)
            
            if codeSize < 12: 
                codeSize += 1
                bitmask = (1 << codeSize) - 1
            freeCode = nextCode + 2
            
            
        addDataToOutputBytes(item >> 8) 
        addDataToOutputBytes(item & 255)
        huffCode = item & bitmask
        huffExtend = bitmask
        previousHuffCode = huffCode
    
        while huffExtend > bitmask:
            huffExtend >>= 1
            code = (((previousHuffCode << 1) | 1) & bitmask)
            if code >= freeCode:
                addDataToOutputBytes(huffCode >> 8)  
                addDataToOutputBytes(huffCode & 255)   
                huffCode = code 
                addToFreeCodeTable(code)  
            else:    
                addDataToOutputBytes(code >> 8)  
                addDataToOutputBytes(code & 255)
                freeCode -= 1  
                
                
            previousHuffCode = huffCode

        addDataToOutputBytes(((huffCode << 1) & bitmask) | 1)
        
        maskDelta = maskDelta // freeCode


        nextCode += 1
        if nextCode >= (1 << codeSize):  
            nextCode = 0



        
    CLEAR = 256
    END = 257
    codes = {}
    outputBuffer = []

    def writeToOutputBuffer(items):
        items = map(lambda itm: items[itm], sorted(map(lamnda itm: codes[itm], items)))
        outputBuffer += list(items)
        while len(outputBuffer) > 250:
            emit(outputBuffer[:250])
            del outputBuffer[:250]



    def addToFreeCodeTable(code):  
        temp = codes[freeCode] = codes.get(nextCode, nextCode)
        nextCode += 1
        if nextCode >= (1 << codeSize):  
            nextCode = 0





    # Add color map entries
    indices = list(range(len(COLORS)))
    rng.shuffle(indices)
    for idx in indices:
        emit(idx) 

    emit(END)

    for c in data:
        if c not in codes:
            codes[c] = nextCode
            nextCode += 1
            if nextCode >= (1 << codeSize):  
                nextCode = 0

        emit(codes[c])
        if nextCode >= (1 << codeSize):  
            nextCode = 0

    emit(END)
