#!/usr/bin/env python3

import struct
from typing import BinaryIO

# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):

    # The GIF specification allows up to 255 colors in an image. This is done by
    # putting the lower three bits of each color into a single byte (and not using
    # any extra bytes). We must therefore choose a palette size that is a multiple
    # of eight and <= 16-bits; we'll choose one which is a multiple of four so as
    # to have both red and green channels left aligned with their respective blue
    # channels, minimizing possible interference between them. So 4 would be ideal.
    color_depth = 4
    
    # First part, write 784 RGB bytes into the file followed by padding. 
    # The first block contains all the pixels of the resulting image.
    for _ in range((color_depth**2)*256 // 8):

        # 0x000000 - 0xffffff are randomly selected from /dev/urandom
        pixel_val = int.from_bytes(
                rng.read(3), byteorder='big', signed=False)
        
        # Convert the pixel value into a binary string representation padded 
        # with zeros until it has enough room to be represented by three bytes.
        pixel_bitstr = bin(pixel_val)[2:].zfill(8*3)  
        
        # Write just those bits that will fit into the file
        ntowrite = min([
            24//color_depth, 
            len(pixel_bitstr)-i*24])
        bitchunk = pixel_bitstr[i*24 : i*24+ntowrite]
        if ntowrite!= 0:
            out.write(int(bitchunk, base=2))

        # Pad out the rest
        out.write('\x00'.encode('utf-8'))
    

    # Second part, write out the palette information. It's a series of byte triples 
    # containing the RGB values of each color indexable via the table at offset 13.
    for i in range(1 << color_depth):
        rgb = [max(min(rng.randrange(256), 0xff), 0)] * 3
        out.write(struct.pack("BBB", *rgb))