#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Generate a header with width and height between [5-7]x[5-7].
    width = (int.from_bytes(rng.read(1), "big") & 0b011111) - 2 
    if not width:
        width = 1
    
    height = (int.from_bytes(rng.read(1), "big") & 0b011111) - 2;
    if not height:
        height = 1;
        
    screenwidth = width * 10
    screenheight = height * 10

    pixelcount = width * height

    out.write(b"GIF89a");
    out.write((screenwidth).to_bytes(2, byteorder="little"));
    out.write((screenheight).to_bytes(2, byteorder="little"));
    
    output = []

    colormapdata = bytes([
        0, 0, 0,    # Background Color Index
        0xFF, 0, 0, # Red
        0, 0xFF, 0, # Green
        0, 0, 0xFF   # Blue
    ])
   
    # Write graphic control extension
    out.write(b'\x21\xF9\x04');
    out.write(b'\x08');
    out.write(((pixelcount)&0xff00)>>8);
    out.write(((pixelcount)&0xff));
    out.write(b'\x00');
    out.write(b';');
    
    # Write dimensions in logical order (top to bottom rather than left to right).
    out.write(b"\x21\x00\x0B\x6E\x6F\x6E\x65\x2D\x67\x61\x6C\x70\x68\x61\x21\x0A");
    
    # Save local colour table as global one
    out.write(b'\x21\xFC')
    out.write(len(colormapdata).to_bytes(1, byteorder='little'))
    out.write(colormapdata)
    
    # Write palette:
    for i in range(0, len(colormapdata), 3):
        out.write(b"\x00\x00") # Padding
    
    # Write graphics control block
    out.write(b"\x21\xF9\x04\x08")
    out.write((((pixelcount)&0xff00))+1)
    out.write((((pixelcount)&0xff)))
    out.write(b"\x00\x0f")
    
    
    imagedescriptorblocksize = ((width+1)*height)+3

    # Loop until we have written enough pixels to fill up
    # each line of the canvas
    while True:

        # Write image descriptor
        out.write(
            b"\x2c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00" +
            (((imagedescriptorblocksize)&0xff00)>>8).to_bytes(1, byteorder="little"))
        out.write(
            ((imagedescriptorblocksize)&0xff).to_bytes(1, byteorder="little"))
        
        # Write a single colour index -- this is redundant but it makes
        # viewers happier because they don't need to look for an
        # associated GCT entry (although we do provide one above so even
        # better!)
        out.write(b"\x00\x00")

        rowsize = width*(1+2)+1
        scanline = []
    
        # Read enough bytes from the input stream to fill the current scanline
        while len(scanline) < width*2+1:
            scanline += list(rng.read(rowsize))
        
        # The last two bytes represent how many of those bytes are part of the
        # scanline before the next extension chunk begins -- set them to zero since
        # there isn't another image inside us!
        scanline[-2:]=[0, 0];
        
        # Output the scanline
        out.write(bytearray(scanline))

        if len(scanline) >= width*2+1:
            break
