#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO


# Constants for generating random GIFs
HEADER = b'GIF89a'
WIDTH  = (0x05, 0x1f, 0xff) # 72 - 168 pixels
HEIGHT = (0x05, 0x1f, 0xff) # 43 - 127 lines
COLORMAP = {'static': True} # Use a static colormap rather than lzw compression
TRANSPARENT = False
TRAILER = [0x3b]



# Generate a random GIF to the given stream and return its size in bytes
def make_random_gif(rng: BinaryIO, out: BinaryIO):

    if COLORMAP['static']:
        colormap = _make_color_table()
    else:
        raise NotImplementedError

    # Write the header
    out.write(bytes(HEADER))

    width = rng.read(1)[0] % len(WIDTH) + WIDTH[0]
    height = rng.read(1)[0] % len(HEIGHT) + HEIGHT[0]

    out.write(struct.pack('<HHB', width, height, 0xf7 << 4 | 0x01))

    # Write the color table
    if COLORMAP['static']:
        for byte in colormap:
            out.write(bytearray([byte]))

    # Write global flags
    gflags = 0x80 # Global color table is present
    if TRANSPARENT:
         gflags |= 0x01
    out.write(gflags.to_bytes(1, 'big'))

    # Write loop control block
    out.write(bytes([*reversed((0x00, 0x00)), *reversed((width, height))] + TRAILER))

    # Write local flags
    lflags = 0xe0 & width >> 3
    lflags += 0xe0 & height >> 3
    if not COLORMAP['static'] or TRANSPARENT:
        lflags += 0x01
    out.write(lflags.to_bytes(1, 'big'))

    # Write palette
    if COLORMAP['static']:
        out.write(b'\x00')     # Background color index
        out.write(b'\xff\xff\xff') # Default foreground color

    # Write image data
    while True:

        chunksize = rng.read(1)[0]
        if chunksize == 0x00:
            break

        ltype = rng.read(1)[0]
        if ltype == 0x2c:

            pass
            # Write graphics control block

        elif ltype == 0x21:
            
            pass
            # Write extension blocks

        else:

            pass
            # Write plain data
        
        chunkend = rng.read(1)[0]
        assert chunkend == 0x00
    
    return out.tell()



# Make a new color table for use with the global RGB color map
def _make_color_table():
    # A list of all the possible RGB colors that can be used as an index
    rgbcolors = []
    for r in range(256):
        for g in range(256):
            for b in range(256):
                rgbcolors.append((r, g, b))

    # Shuffle them up so they don't always land in order
    rng = os.urandom(len(rgbcolors)).hex().encode('utf-8')
    shufflelist = list(range(len(rgbcolors)))
    for i in reversed(shufflelist):
        j = ord(rng[i])
        shufflelist[i], shufflelist[j] = shufflelist[j], shufflelist[i] 
    
    # Randomly pick some colors from the shuffled list
    ncolors = ord(rng[-1]) % max(len(rgbcolors)//10, 1)
    shortlist = sorted(rng[:ncolors].decode())
    suggestions = {chr(v >> 4).lower()+chr(v & 0x0f).lower()\
            : None for v in range(*map(lambda c: ord(c)*16, shortlist))}

    colorlist = []
    for i in shufflelist:
        color = '#'+''.join('%02x'%c for c in rgbcolors[i][:3]).upper()
        colorlist.append(suggestions.setdefault(color, chr(ord('A')+i%26)))

    # Return as bytes
    return bytes(''.join(colorlist), encoding='ascii')