#!/usr/bin/env python3

import os, sys
import struct
from typing import BinaryIO
import argparse



def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('out', type=argparse.FileType('wb'))
    args = parser.parse_args()

    try:
        with open('/dev/urandom', 'rb') as f:
            ok = generate_random_gif(
                    f, args.out
                )
    except PermissionError:
        errmsg = "Cannot access /dev/urandom!\n"\
                 + "(Try running this script as sudo)"
        raise PermissionError(errmsg) from None
        
    if not ok:
        print("Could not generate random GIF!")
        sys.exit(1)



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Magic string
    magic = 'GIF89a'.encode('ascii')
    out.write(magic)

    # 7-bit packed integer representing screen width (w)
    w = rng.read(1)[0]
    packstr = '>Bxxx'
    out.write(struct.pack(packstr, w))

    # 7-bit packed integer representing screen height (h)
    h = rng.read(1)[0]
    packstr = '>Bxxx'
    out.write(struct.pack(packstr, h))

    # Packed little endian unsigned short (2 bytes)
    # storing the global color table size (ctsize)
    ctsize = rng.read(1)[0] | ((rng.read(1)[0]) << 8)
    packstr = '<xxH'
    out.write(struct.pack(packstr, ctsize))

    # Global Color Table
    if ctsize > 0:
        for _ in range(pow(2, ctsize)):
            r, g, b = rng.read(3)
            out.write(struct.pack('<BBBxxx', r, g, b))

    # Extension introducer
    # Graphics Control Extension (0xf9) label
    extlbl = '\x21\xff\x0b'.encode('ascii')
    out.write(extlbl)

    # Block Size (# bytes until end of block)
    bs = b'\x46'   # 46 bytes long
    out.write(bs)

    # Packed bit fields - 1 byte
    flags = rng.read(1)[0]
    # 1st bit sets transparent color flag
    tcf = bool((flags >> 7) & 1)
    # Next three bits set disposal method
    dm = (flags >> 2) & 7 
    # Next three bits set user input flag
    uif = (flags >> 1) & 1
    # Last bit indicates transparency is given by pixel
    tp = flags & 1 

    # Delay time (2 bytes)
    dly = rng.read(2)
    out.write(dly)

    # Transparent color index (1 byte)
    # If transparancy is given by pixel, then no value should be included
    tcidx = rng.read(1)
    if tp ^ tcf:
        out.write(tcidx)

    out.write('\x00\x00'.encode('ascii'))    # Block terminator


    # Image descriptor label
    imglbl = '\x2c'.encode('ascii')
    out.write(imglbl)

    # Image left offset byte (1 byte)
    loffs = rng.read(1)
    out.write(loffs)

    # Image top offset byte (1 byte)
    toffs = rng.read(1)    
    out.write(toffs)

    # Image width byte (1 byte)
    imw = rng.read(1)    
    out.write(imw)

    # Image height byte (1 byte)
    imh = rng.read(1)   
    out.write(imh)

    # Packed bit field - 1 byte
    ibge = rng.read(1)[0] 
    lct = bool((ibge >> 7) & 1)      # Local color table flag
    interlace = bool((ibge >> 6) & 1)     # Interlace flag
    sort = (ibge >> 3) & 7           # Sort flag - ignored here
    reserved = (ibge >> 1) & 3        # Reserved bits - set to zero
    lgmsk = ibge & 1                 # Local color mask - ignored here

    # LZW min code size - 1 byte
    lcsz = rng.read(1)[0]
    packstr = '<Bxxxx'
    out.write(struct.pack(packstr, lcsz))

    # LZW max code size - 1 byte
    mcsz = rng.read(1)[0]
    packstr = '<Bxxxx'
    out.write(struct.pack(packstr, mcsz))

    # Read the remaining image data
    imdata = rng.read(int.from_bytes(imw))*int.from_bytes(imh)
    out.write(imdata)

    # Pack EOF
    trailer = b'\x3b\x00'       # File terminator
    out.write(trailer)

    return True



