#!/usr/bin/env python3

import random
import itertools


# Generates a random sequence of state values that will eventually be written
# as a GIF (or similar graphics format). The RNG object given should produce
# bytes uniformly at random over the interval [0, maxvalue]. This function
# generates up to maxblocks blocks before returning. Each block may contain one
# or more images; each image contains up to maximages frames, which are
# concatenated together with each other. The endianness argument determines how
# many bits are used per byte when writing numbers.
def generate_random_seq(maxblocks, maximages, rng, endianness='little'):

  # Define helper functions
  def chunks(src, n):
    """ Yield successive n-sized chunks from src list.""" 
    srclist = iter(src)
    while True:
      chunk = tuple(itertools.islice(srclist, n))
      if len(chunk)==n: yield chunk
      else: break
  
  def write_fields(*args):
    outstream.write(bytes([sum(args)]))
    outstream.write(bytes([len(args)+1]))
    
  def write_blocktype(blocktype):
    if blocktype!= 'last':
      write_fields(1 << 7)
    elif blocktype == 'last':
      write_fields(0x00)
    else:
      raise ValueError("Invalid Block Type")

  def choose_blocktype():
    if random.random() < 0.1:
      return "last"
    else: 
      return None

  def write_blocklength(blocknumber):
    # TODO: Maybe try different lengths?
    outstream.write(bytes([random.randint(16, 256 + 16)]))

  def choose_first_length():
    if random.random() < 0.5:
      return False
    else:
      return True
  
  validincrements = {
    3 : [(0,0),(-1,-1),(0,-1)], 
    4:  [(0,0),(-1,-1),(0,-1),(1,-1)] 
  }
  
  incrementbase = {
    3:(3,2,1),
    4:(8,4,2,1)    
  }
  
  def getincrement(length):
    inc = validincrements[length][random.randrange(0,validincrements[length])]
    return incrementbase[length][inc[0]],incrementbase[length][inc[1]]
  
  
  # Set default parameters
  blocktype = choose_blocktype()
 
  while blocktype!= 'last':
    blocknumber += 1
    firstlength = choose_first_length()
    length = 1
  
    if firstlength:
      if random.random() <= 0.1:
        length += int(min(maximages+1, pow(3, random.uniform(1, 10))))
      else:
        length += int(pow(2, random.uniform(1, 10)))
    else:
      incX,incY = getincrement(length)
            
      while length < min(maximages+1, pow(3, random.uniform(1, 10))): 
        length += 1
        incXprev,incYprev=incX,incY
        incX,incY = getincrement(length)
    
    write_blocktype(blocktype)    
    write_blocklength(blocknumber)
    
    for i in range(length):

      # Choose whether to introduce an interlace frame
      hasinterlacedframe = False
      if blocktype == 'image':
        if random.random() <= 0.5: 
          hasinterlacedframe = True
      
      # Process interlace frame
      if hasinterlacedframe:
        pass

      # Write next frame
      outstream.write(bytes([1]))

      # Write left offset
      outstream.write(bytes([0]))

      # Write top offset
      outstream.write(bytes([0]))

      # Write width
      outstream.write(bytes([width << 2]))
  
      # Write height
      outstream.write(bytes([height << 2]))
    
      # Write packed field
      packedfield = 0x00
          
      if random.random() <= 0.1:
        packedfield |= 0b0100
      
      if hasinterlacedframe:
        packedfield |= 0b1000
        
      outstream.write(bytes([packedfield]))
  
      # Initialize LZW code size and dictionary
      LZWcodesize = min(8, maxbits // 3 - 1)
      dictsize = 1<<LZWcodesize
      LZWdict = {}     
      
      # Initialize bit buffer and bit counter
      bitbuffer = 0
      bitcounter = 0 

      # Write the compressed data to stream
      while True:
          
        # Check if we need to refill the dictionary
        if len(LZWdict) >= dictsize:
          
          # Refill dictionary by adding all possible symbols
        
          codelen = min(8, maxbits // 3 - 1)
          dictstartindex = 1<<(codelen-1)
              
          for value in range(dictstartindex, dictsize):
            LZWdict[value] = bytes([(value >> 8)]) \
                           + bytes([value % 256])
                    
          # Set new code size based on current dictionary size
          LZWcodesize = min(8, maxbits // 3 - 1)
          dictsize = 1<<LZWcodesize
                    
        # Choose the length of the next word
        if bitcounter == 0:
          nextwordlen = rng.randint(1, LZWcodesize+1)
                
        # Fill the bit buffer until we have enough bits        
        while bitcounter < nextwordlen:
          # Fills evenly so the result has equal probability as normal randint
          bitbuffer <<= 1
          bitbuffer |= rng.randint(0, 1)
          bitcounter += 1
        
        # Pick a symbol to read from the bit buffer         
        symstartpos = bitbuffer>>nextwordlen-(LZWcodesize-nextwordlen+1)
        symendpos = symstartpos + LZWcodesize-nextwordlen+1
        sym = bitbuffer & (0xff>>(8-symendpos))
        
        # Extract the next word from the bit buffer
        bitbuffer ^= sym << (bitcounter-(LZWcodesize-nextwordlen+1))
        bitcounter -= (LZWcodesize-nextwordlen+1)
            
        # Use existing symbol
        if sym in LZWdict:
          outstream.write(LZWdict[sym])
        
        # Otherwise construct a new one  
        else:
          newsymbol = [symstartpos ^ 0xff, symendpos ^ 0xff]
            
          for i in range(LZWcodesize-nextwordlen+1):
            character = rng.randint(0, 255) 
            newsymbol.append(character)
          
          outstream.write(bytes(newsymbol))

          # Update the dictionary
          LZWdict[sym]=bytes(newsymbol)
          
      # Flush the remaining bit buffer contents      
      while bitcounter>0:
        outstream.write(bytes([(bitbuffer & 1) ^ 0xff]))
        bitbuffer >>= 1
        bitcounter -= 1

    # Do something with the result
    blocktype = choose_blocktype()