from enum import IntEnum, unique
import math

@unique
class Header(IntEnum):
    """Enumerates all parts of the GIF header."""

    SIGNATURE = 0x47494658 # XFI\x00
    VERSION   = 0x2A31           # *!
    WIDTH     = 0x212C           #!,\
    HEIGHT    = 0x212D           #!-
    WHITEPALETTE = 0x21FF      #!\xFF (not in spec)
    LOCALCOLORMAP = 0x2100      #!\00 (not in spec)
    INTERLACEFLAG = 0x2108       #![\n
    GRAPHICCONTROLEXTENSION = 0x21F9  #!\xF9 (not in spec)
    APPPLICATIONEXTENTION = 0x21FF  #!\xFF (not in spec)
    IMAGEDESCRIPTOR = 0x2C         #,
    GRAPHICCONTROLEXTENSIONLABEL = 0x22 # "
    BLOCKTERMINATOR = 0x3B          # ;
    APPEXTENSIONLABEL = 0xFF01      # \xFF\x01
    NETSCAPECLIENTAPPLICATION = 0x1B
    NETSCAPEBLOCK = 0x03
    LOOPCOUNT = 0x01
    NETSCAPENUMBER = 0X214E5343 # NS\x43\x4E\x50\x1
    NETSCAPELOOPEXITFLAG = 0x3B # ';'
    NETSCAPEEND = 0x00 # '\x00'
    COLORDEPTH = 0x2102 #!\x08
    BACKGROUNDCOLORINDEX = 0x2109 #!\t
    STARTPOSITION = 0x210A #!\n
    ENDINGOFFSET = 0x210B #!\v
    
# Replaces all bytes including and past `pos` with
# `filler`.
# Returns True if no replacements were made. Otherwise
# returns False.
def _compress_data(
        data: memoryview, pos: int, filler=b'\x00', left=None):
    
    if not isinstance(data[pos], type(filler)): return True
    
    if left is None: 
        left = len(data) - pos
        
    # Go backwards until we find something other than
    # `filler`, or reach zero length. Then replace the 
    # entire range between start and i inclusive with
    # `filler`. If this would do nothing then do nothing. 
    for i in reversed(range(left+1)):
        if data[i]!= filler: break
        
    else: return True
        
    del data[i:pos+1]
    return False

def _make_compressed_data(
        rng: BinaryIO, maxlen: int, residue=b'\x00'):
    
    datalen = min(maxlen, rng.readable() // 3)
    data = rng.read(datalen*3)
    
    while len(data) > maxlen:
        
        # First try to compress from end to beginning.
        # If that doesn't work, compress from front to
        # end.
        compressed = _compress_data(
                memoryview(data), 
                len(data)-1, residue=residue)

        if compressed:
            
            # Compressing from back didn't help so now 
            # we need to compress from front.
            _compress_data(
                    memoryview(data), 
                    0, residue=residue)
            
        datalen -= 1
        assert datalen >= 0
        data = rng.read(datalen*3)
        
    assert len(data) == maxlen
    return data

# Makes an uncompressed image descriptor block of size
# `size` with top left corner at `(x,y)` and color map
# located at offset `offset` in the palette. The rest of
# the fields are filled in with default values.
# Returns the bytes object representing these fields as a
# 1-dimensional array of unsigned integers.
def make_uncompressed_imagedescriptorblock(
        x:int, y:int, size:int, offset:int, palsize=0):
    
    ret = []
    ret += [Header.IMAGEDESCRIPTOR.value]
    ret += [0x00, 0x00, 0x00, size % 256]
    ret += [0x00, (size >> 8) % 256]
    ret += [(x << 4) & 0xF0 | ((y >> 8) % 0x0F)]
    ret += [y % 0xFF]
    ret += [0x00];
    ret += [(palsize<<4) | 0x3] # local color table size / sort flag
    ret += [offset % 256]
    ret += [(offset >> 8)% 256]
    return bytes([b for t in ret for b in t])
    

# Makes a compressed image descriptor block of size `size`
# with top left corner at `xy` and encoded data that will be
# written after it. The rest of the fields are filled in with
# default values.
def make_compressed_imagedescriptorblock(
        xy: tuple, size: int, encodedsize: int):

    ret = []
    ret += [Header.IMAGEDESCRIPTOR.value]
    ret += [0x00,0x00,0x00, (size%256)]
    ret += [0x00,((size>>8)%256)]
    ret += (((xy[0]%256)<<4)|(((xy[1]>>8)&0x0F)))
    ret += ([xy[1]&0xFF])
    ret += [0x00];
    ret += [encodedsize | 0x3]
    ret += [0x00,0x00]
    ret += [0x00, 0x00]
    return bytes([b for t in ret for b in t])
    

# Makes a graphic control extension block that clears the screen
# to black before displaying the next frame.
def make_graphiccontrol():
    return bytes([
        Header.GRAPHICCONTROLEXTENSIONLABEL.value,
        0x08, # sizeof block
        0x00, # reserved one byte
        0x00, # frame disposal method
        0x00,0x00, # delay time
        0x00,0x00, # transparent color index
        0x03, # block terminator
    ])

# Attempts to encode a graphics control extension followed by a raw
# raster image using LZW compression. If the encoding fails because
# the resulting image exceeds MAXSIZE bytes, attempts are made using
# compression starting at offsets from the beginning of the data and 
# towards the end. This search continues until either MAXATTEMPTS have
# been tried without success OR the data has been successfully decoded
# OR the maximum possible size has been reached. In the last case, the 
# function returns (False, None). In the first two cases, the returned value
# is a pair where the second element is non-None iff the image was successfully
# decoded.
MAXATTEMPTS=10
#MAXSIZE=math.inf
MAXSIZE=50000
def _encode_lzw(
        rng: BinaryIO, width: int, height: int,
        maxiter=-1, mininterval=1, verbose=True
    ) -> tuple:

    if maxiter < 0:
        maxiter = width*height//mininterval
    
    assert maxiter > 0
    assert mininterval <= maxiter
    
    # Try to make a graphic control extension first...
    try:
        gcrange = rng.tell(), rng.seek(0,2)
        gcstart = rng.tell()
        rng.write(make_graphiccontrol())
        rng.seek(gcrange[0],0)
    
    except IOError:
    
        # Compression failed on making the graphic control
        # label but we don't really care since there's no
        # pixels anyway so just set gcstart to 0.
        gcstart = 0
    
    # Construct empty arrays to store the offsets of each
    # valid location in the header and the corresponding
    # pixel counts.
    validmap = [gcstart]
    pixelcount = [0]
    
    # Make sure the encoder is initially at the beginning
    # of the data stream
    rng.seek(0)
    
    for attempt in range(MAXATTEMPTS):
        
        # Seek back to the original position of the GC
        # extension. We want the encoder to start reading
        # the raster image right after the opaque part of
        # the extension.
        rng.seek(gcrange[0],0)
        
        # Write the raster image descriptors for every frame
        # to the header. Also build up a list mapping which
        # header locations correspond to which frames.
        validframes = []
        
        # These variables keep track of where we are in the
        # image and how many pixels there are
        count=width*height
        x=y=0
        
        # The headers of the previous image must be padded
        # with 4 bytes to achieve a multiple of four. We'll
        # need to know when that happens to avoid padding it
       