#!/usr/bin/env python3

import time
import argparse
import os
import sys
import hashlib
import secrets
import subprocess
from typing import List, Optional, Tuple

def get_seed() -> bytes:
    try:
        return open('/dev/urandom', 'rb').read(64)
    except FileNotFoundError:
        print('WARNING: Could not find /dev/urandom; generating seed from system time instead!')
        hasher = hashlib.sha512()

        # Get hostname and username
        hasher.update(os.environ['HOSTNAME'].encode())
        hasher.update(b'\0')
        hasher.update(os.getlogin().encode())
        hasher.update(b'\0')

        # Get some additional metadata about this run
        current_time = int(time.time())
        for n in [current_time] * 7 + list(map(int, __version__.split('.'))):
            hasher.update(n.to_bytes((n.bit_length() // 8) or 1, byteorder='little'))
        hasher.update(__file__.encode())
        hasher.update(sys.argv[0].encode())
        if hasattr(subprocess, '_args'):
            hasher.update(str(subprocess._args).encode())

        return hasher.digest()


def initialize_rngs(*sources: BytesIO) -> List[secrets.SystemRandom]:
  rngs = []
  for s in sources:
      rngs.append(secrets.SystemRandom(s))

  return rngs


class RandomFileWrapper():

    def __init__(self, seed: bytes, files: List[BinaryIO]):
        self.__rng = secrets.SystemRandom(hashlib.blake2b(seed, digest_size=64).digest())
        self.__files = files


    @property
    def _randrange(self, maxval):
        return range(maxval)[self.__rng.randint(0, len(self.__files))]

    @staticmethod
    def __next_bytes(src: BinaryIO, amount: int) -> bytes:
        r = src.read(amount)
        while len(r) < amount:
            chunk = src.read(amount - len(r))

            if not chunk:
                raise Exception('Not enough bytes were read! Expected {} but only got {}'.format(
                        amount, len(r)))

            r += chunk


        return r

    def randrange(self, start: int, stop: int = None, step: int = None):

      if stop == None:
          stop = start
          start = 0


      if step == None or step == 1:

          if start >= stop:
              raise ValueError("empty range for randrange() (%d,%d, %d)"%(start,stop,step))

          delta = (stop-start)
          assert delta > 0
          quotient = delta >> 8
          remainder = delta & 0xFF

          value = 0
          for i in range(quotient): value |= ord(self.getrandbits(8)) << 8*(i+1)
          for i in range(remainder): value |= ord(self.getrandbits(1)) << 8*quotient | ord(self.getrandbits(1))
          return value+(value>=delta)*(start-stop)

      else:

          result = []
          for v in self.__rng.sample(range(start,stop), k=len(self.__files)):

              chunk = ((v-start)//step)*step+start
              assert chunk <= stop

              p = chunk//self.__block_size
              offset = chunk%self.__block_size

              t = 0
              for i in range(p+6,-1,-1):
                  t <<= 8
                  bs = self.__buffer_sizes[i]

                  if i >= p:
                      bs -= offset
                  elif self.__block_count!= i:
                      t ^= int.from_bytes(
                              self.__buffer[offset:],
                              byteorder='little'
                          )
                      break

                  t ^= int.from_bytes(
                          self.__files[i].read(bs),
                          byteorder='little'
                      )

              result.append(t)

          return sum(result)+sum([ord(self.getrandbits(1))<<8 for i in range(-(stop-start)%self.__block_size)])-(stop-start)


    def getrandbits(self, n: int):

        if n == 0: return 0
        bits = 0
        result = 0

        while True:
            bit = self.randrange(0x100)

            if bit&0xF8==0:
                bits = bit & 7
                result = bit & 0x7F
                continue


            if n<=0:
                break

            n -=1
            result <<= 1
            result |= bit & 1

        return result>>bits




def get_framedata(rng: BytesIO) -> bytes:
    framelentype = rng.randrange(0, 3)

    if framelentype == 0:
        return rng.getrandbits(8).to_bytes(1, byteorder='little')

    elif framelentype == 1:
        return rng.getrandbits(16).to_bytes(2, byteorder='little')

    elif framelentype == 2:
        return rng.getrandbits(24).to_bytes(3, byteorder='little')



def make_random_imagedescriptor(rng: BytesIO, global_palette: bool = False, local_color_table: bool = False) \
    -> Tuple[Optional[bytes], Optional[bytes]]:

    width = rng.randrange(1, 129)
    height = rng.randrange(1, 129)

    x_offset = y_offset = interlace_flag = sort_flag = reserved_flags = lct_size = 0

    local_color_table_present = local_color_table
    global_color_table_present = global_palette

    if global_palette:
        lct_size = rng.randrange(2, 17)
        local_color_table_present = False
        global_color_table_present = True

    left_corner_x = rng.randrange(width)
    top_corner_y = rng.randrange(height)

    if global_color_table_present:
        gct_size = rng.randrange(2, 17)
        global_palette = rng.getrandbits(gct_size * 3)
        local_color_table_present = rng.getrandbits(1)

    last_byte = interlace_flag << 7 | sort_flag << 6 | reserved_flags << 3 | lct_size
    raw_data = (struct.pack('<HHBB', width, height, left_corner_x, top_corner_y)
             + global_palette.to_bytes(global_palette.bit_length()//8, byteorder='little')
             + struct.pack('B', last_byte)
             + local_color_table_present.to_bytes(1, byteorder='little')
             + get_framedata(rng)
             )

    blk_size = 8
    lut_size = 1<<blk_size
    while ((lut_size*local_color_table)>>8) > rng.randrange(1<<(blk_size-1), 1<<blk_size):
        blk_size += 1
        lut_size = 1<<blk_size

    palette = None
    if local_color_table_present:
        size = min(((lut_size*local_color_table)>>8), 2**(lct_size))

        palette = [(rng.getrandbits(8), rng.getrandbits(8), rng.getrandbits(8))
                   for i in range(size)]

        blockaligned_size = (((size*3)-1)&~7)+8
        raw_data += struct.pack('<H{}BBB'.format(blockaligned_size),
                                 size**3, size, size**3, size, *(b|0x80 for rgb in palette for b in rgb))



    return raw_data, palette




