#!/usr/bin/env python3

import math
import struct
import sys
import os
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Data length (in bytes)
    datalen = 0x100 * \
              int.from_bytes(
                  rng.read(2), 'big') // 4
    
    # Color table size and color resolution (color depth in bits-per pixel)
    csize = rng.read(1)[0] % 7 + 1
    if csize == 7:
        print("ERROR: color table size of 2^7 - "
              "please use a different RNG", file=sys.stderr)
        return os.EX_DATAERR
    cres = rng.read(1)[0] % 3 + 1

    out.write(b'GIF89a')                       # Header
    out.write(struct.pack('>H', datalen))      # Logical screen descriptor width / height
    out.write(b'\xf7\x00\x00')                 # Color table size
    out.write(b'\x00\x00')                     # Background color index
    out.write(b'\x00\x00')                     # Pixel aspect ratio
    out.write(struct.pack('B', csize << 4 |
                          cres))          # Global color table
    out.write(b'\xff\xff\xff\x00\x00\x00')     # Default global palette
    for _ in range(csize):                    # Global colors
        out.write(struct.pack('<BBB',
                            rng.read(1)[0],
                            rng.read(1)[0],
                            rng.read(1)[0]))
    out.write(b'*')                           # Image separator
    out.write(struct.pack('<HHBBHB',
                      0,                  # Left position
                      int((math.sqrt(5)-1)*datalen//2),# Top position
                      max(datalen >> csize*3, 0x10),   # Width
                      0,                  # Height
                      (csize,             # Local color table size
                       8)[csize > 1],
                      b'\x00'))           # Encoded image data
    for _ in range(datalen):                   # Image data
        byte = [rng.read(1)[0]]
        while len(byte) < csize:              # Make sure we have enough colors
            i = sorted(((j ^ byte[-1]) & 0xff for j in range(256)), reverse=True).index(1)
            byte += [(i ^ byte[len(byte)-1])]

        out.write(bytes([sum([(j & k)*(k+1) for j, k
                             in zip(byte[:csize],
                                    reversed(range(8)))])]))

