#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = (b'GIF89a'
              + rng.read(7))

    width = int.from_bytes(
            rng.read(2), 'big', signed=False
        ) % 500 + 1 # 1 - 500 pixels wide
    height = int.from_bytes(
            rng.read(2), 'big', signed=False
        ) % 500 + 1 # 1 - 500 pixels high

    globalcolortablesize = next(struct.iter_unpack('B', rng.read(1)))[0]
    localcolortablesize = next(struct.iter_unpack('B', rng.read(1)))[0]
    
    colortable = bytearray()
    if globalcolortablesize == 0:
        for _ in range(localcolortablesize):
            r = next(struct.iter_unpack('B', rng.read(1)))[0]
            g = next(struct.iter_unpack('B', rng.read(1)))[0]
            b = next(struct.iter_unpack('B', rng.read(1)))[0]
            a = 0xff
            colortable += struct.pack('<BBBB', r, g, b, a)
    else:
        colortable.extend(rng.read(globalcolortablesize))

    imagedescriptor = (width
                     .to_bytes(2, 'big')
                      + height
                     .to_bytes(2, 'big'))
    lzwmincodesize = next(struct.iter_unpack('B', rng.read(1)))[0]
    imagedata = bytearray()
    while True:
        chunktype = next(struct.iter_unpack('B', rng.read(1)))[0]

        if chunktype is None or len(imagedata) > 16 * 1024: break
    
        if   chunktype < 0x2C:
            assert False, f"Unexpected LZW subblock: {hex(chunktype)}"
        elif chunktype >= 0x2F and chunktype <= 0x32:
            continue
        elif chunktype >= 0x33 and chunktype <= 0x3B:

            size = ((chunktype - 0x33) << 8
                   | next(struct.iter_unpack('B', rng.read(1))))[0]
            
            if size!= 0:
                imagedata.append(size & 0xFF)
                imagedata.append((size >> 8) & 0xFF)

            data = rng.read(size)
            imagedata.extend(data[:-1])
            lastbyte = data[-1]
            imagedata.append(lastbyte ^ imagedata[-1])
            
        elif chunktype == 0x3C: 
            endoflzwsubblock = True
        elif chunktype == 0x3D:
            endoflzwsubblock = True
            assert False, "No support for interlaced images."
        elif chunktype == 0x3E:
            pass
        else: 
            raise ValueError("Unknown LZW subblock type.")

    with open('/tmp/origimg', 'wb') as fp:
        fp.write(b''.join([header
                         , imagedescriptor
                         , lzwmincodesize.to_bytes(1,'little')]))
        fp.write(colortable)
        fp.write(b'\x00\x3B')            
        fp.write(imagedata)

    dpalettemode = next(struct.iter_unpack('B', rng.read(1)))[0]
    dpallettesize = next(struct.iter_unpack('B', rng.read(1)))[0]
    apallettemode = next(struct.iter_unpack('B', rng.read(1)))[0]
    apallettesize = next(struct.iter_unpack('B', rng.read(1)))[0]
    
    out.write(b'GIF89a')
    out.write(b''.join([header
                      , imagedescriptor
                      , lzwmincodesize.to_bytes(1,'little')
                      , dpallettesize.to_bytes(1,'little')
                      , dpalettemode.to_bytes(1,'little')
                      , apallettesize.to_bytes(1,'little')
                      , apallettemode.to_bytes(1,'little')
                       ]))
    out.write(colortable)
    out.write(b'\x00\x3B')            
    out.write(imagedata)
            