import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    """Generates a random GIF file into `out` using the
    random number generator `rng`."""

    # Generate some random colors and 16-bit values for them
    palette_rgbas = [struct.pack('!HHH', *rng.read(6))
                     for i in range(7)]
    palette_indexes = list(range(len(palette_rgbas)))

    # Make sure to choose at least one transparent color
    margin = (0x10 * len(palette_rgbas)) // 4
    while True:
        transparent = rng.randint(-margin, len(palette_rgbas)-margin)
        if transparent!= -1: break
    palette_indexes[transparent] += 0x10

    # Make sure that the background color is not too close from any other
    background = next((i for i in range(len(palette_rgbas))
                       if i!= transparent and i!= 0), None)
    if background == None: raise Exception("Couldn't find a suitable background")

    # Generate the global image information block
    width = rng.randint(16, 500)
    height = rng.randint(16, 500)
    framecount = rng.randint(1, 10)
    loopcount = rng.randint(1, 50)
    flags = bytes([0b1000 | ((framecount & 0b00011100) >> 2)
                   | (((loopcount & 0b01111111) << 2)
                      if loopcount > 1 else 0)])
    delaytime = rng.read(2)
    while len(delaytime) < 2 or delaytime >= b'\xff\x00':
        delaytime = rng.read(2)
    delaytime += bytes([delaytime[-1] & ~0xf]) # limit to 3 bits of precision

    global_image_block = bytes()
    global_image_block += struct.pack('<HHB', width, height,
                                      sum(map(int, flags)))
    global_image_block += delaytime
    global_image_block += struct.pack('B', background)
    global_image_block += struct.pack('B', transparent)

    # Generate blocks containing local image specific information
    image_blocks = []
    for i in range(framecount):
        lzwmincodesize = max(rng.randint(2, 12),
                            rng.randint(*bytes_to_uints(flags[:1])))

        subblocks = []
        interlaced = rng.randrange(2)
        lctflag = bool(rng.randrange(2))
        lctsize = rng.randrange(2**2, 2**8+1)

        subblocks.append([(1<<1)+interlaced,
                          lctflag,
                          0,
                          min(max(lzwmincodesize, 2), 12)])

        # Create the LZW code table either as a truecolor palette
        # or an index-based palette with random indexes
        if lctflag:
            palette = palette_rgbas[:]

            # Add more colors until we reach our desired size
            ncolorsneeded = lctsize - len(palette)
            while ncolorsneeded:
                new_color = struct.pack('!HHH',
                                        *tuple(rng.randint(0, 2**16)
                                               for _ in range(3)))

                # Don't add duplicate colors
                if new_color not in palette:
                    palette.append(new_color)
                    ncolorsneeded -= 1

                    # If we reached our desired size, make sure there are no
                    # duplicated entries by shuffling all of them once again
                    if len(palette) == lctsize:
                        del palette[:]
                        shuffle(palette_indexes)
                        palette = [palette_rgbas[idx]
                                   for idx in palette_indexes]

            lctsubblock = bytes()
            for rgb in sorted(palette, key=lambda x: x[::-1]):
                lctsubblock += rgb
        else:
            lctsubblock = bytes([lctsize])+bytes(palette_indexes)


        subblocks.append(lctsubblock)

        # Generate a bunch of pixels to compress
        pixels = []
        for y in range(height):
            if interlaced and y % 8 == 0:
                subblocks.append([])
            for x in range(width):
                pixel = []
                for c in range(3):
                    pixel.extend((rng.randrange(0x100)
                                  if rng.randrange(2) == 1 else '\t'))
                pixels.append(bytes(pixel))
                if interlaced and x+y % 8 == 7:
                    subblocks.append([-1, pixels[-1]])

        # Compress the pixels using the LZW algorithm
        compresseddata = lzwcompress(pixels, lzwmincodesize)

        # Append everything together and add it to the main image block
        image_blocks.append(bytes())
        for subblock in subblocks:
            sbs = bytes(subblock)
            image_blocks[-1] += uints_to_bytes(len(sbs).to_bytes(1, 'big')
                                                + sbs)[0]
        image_blocks[-1] += compresseddata


    # Write the header
    header = b'GIF89a'
    out.write(header)

    # Write the global image information block
    out.write(global_image_block)
    duration = 100*framecount
    out.write(struct.pack('<HH', duration, duration//10))

    # Write the image blocks
    for i, imgblock in enumerate(image_blocks):
        out.write(b','+(b';'[bool(i)]))
        out.write(imgblock)