import struct
from typing import BinaryIO

# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # GIF signature (6 bytes) and version string (3 bytes) are written as is.
    out.write(b'GIF89a')

    # Screen descriptor block (7 bytes):
    #  - width: the image's width in pixels (1..2^15-1). Must be even if height > 0!
    #  - height: the image's height in pixels (1..2^15-1)
    #    Height of 0 signifies that the image has no defined height yet;
    #    this should only happen for a Doodle or an animation with variable frame sizes.
    #  - packed field: bitfield comprised of bits ordered from lsb to msb:
    #    - global color table flag: 1 means that a global color table follows,
    #      otherwise there isn't one.
    #    - color resolution: indicates the number of colors we can choose from.
    #      1, 2, 4, 8 indicate up to 2^(N+1)-1 distinct colors, respectively.
    #    - sort flag: set iff every local color table is sorted (has no effect on the
    #      global color table)
    #    - size of global colormap: must be <= 2^(N+1); used only when global color
    #      table exists.
    #  - background index: index of the background color within the global palette
    #  - pixel aspect ratio: must be >= 0; typically 0 unless you're doing a non-square
    #     tiling pattern, in which case it's usually 1/4, 1/2 or 3/4.
    screen_descriptor = struct.pack('>HHBBB',
            0xFF, # width
            0xAB, # height
            0,    # global color table flag
            2,    # color resolution: 2 bits per entry
            0,    # sort flag
            0     # 2 ^ (color resolution + 1) entries in the global color table
    )
    out.write(screen_descriptor)

    # Global color table section, if any (3 bytes total):
    #  - length of global color table, in bytes (= 3 * 2^(N+1)) where N=2.
    #  - always present, regardless of whether we have a color table.
    #    The first two bytes specify the number of colours in the color table,
    #    minus 1 (so, 00 = 2 colours, 01 = 4 etc.).
    #    If the global colour table contains more than 256 colours then this field is
    #    not present and no global color table follows but instead (in the Color Map
    #    Extension Block) each image uses its own local colour table.
    global_colour_table = [i % 256 for i in range(0x100)]
    out.write(struct.pack('B', len(global_colour_table)))
    for r, g, b in zip(*[iter(global_colour_table)]*3):
        out.write(bytes((r,g,b)))

    # Graphics control extension subblock goes here:
    #  - blocksize: size of the graphics control extension block (always 4 bytes)
    #  - flags: packed field, order from lsb to msb:
    #      - reserved bit: must be zero
    #      - disposal method: how to dispose of the previous frame before rendering
    #        the new one: 0 = keep old area, 1 = restore to bg color, 2 = clear to transparent,
    #        3 = revert to prev. frame.
    #      - user input: if set, the graphic should pause after drawing this frame
    #        until some event occurs. If unset, the graphic plays smoothly
    #      - transparent color flag: if set, the alpha channel affects transparency.
    #        Otherwise alpha does nothing.
    #  - delay time: number of hundredths of a second before continuing with the next frame
    #                   0 => default value of 100
    #  - transparent color index: index into global palette indicating what color will
    #               be treated as transparent during display (if transparent color flag was set)
    graphics_control_extension = struct.pack('BBHBBBB',
            4,          # block size
            0,          # transparent color flag & KEEP DISPOSE ALWAYS
            0,          50,# delay time
            0xCD        # transparent color index
    )
    out.write(graphics_control_extension)

    # Image descriptor block (10 bytes):
    #  - left: position at which to place this image relatively to top-left corner of page
    #  - top: ditto
    #  - width: width of this image in pixels
    #  - height: height of this image in pixles
    #  - local colour table: boolean specifying whether this image uses a local or
    #                        global colour table. True => use local. False => use global.
    #  - interlace: whether the rows of the image should be read top-down rather than left-right
    #  - sort flag: whether ths local colour table is sorted. Only relevant if we're using
    #               a local colour table. Irrelevant if we're using a global colour table.
    #  - reserved: must be zero
    #  - LZW minimum code size: smallest size of compressed codes to write, in bits.
    #                          Typically between 8 and 12 inclusive.
    image_descriptor = struct.pack('>HHHHBB',
            0x00,  0x00, # left
            0x00,  0x00, # top
            0xFF,  0xAB, # width
            0xFF,  0xAB, # height
            0,      0, # no local colour table
            0,      0, # no interlacing
            0x08     # LZW min code size : 8 bits
    )
    out.write(image_descriptor)


    # Frame:
    #  - Table of LZW lengths:
    #     A packed series of bytes beginning with a single byte determining the initial
    #     code size: EITHER a fixed number of bits or a variable number of bits.
    #     Then come either fixed numbers of bits OR pairs of bits indicating how many
    #     subsequent elements in the list are identical to the previous element.
    #     Once again, this one is accompanied by a special byte to mark the end of the table.    
    def make_table():
        yield from (
                0xC4, # initial code size 12
                0xB0, # number of following bytes giving actual length values
                0x0A, # the length value
                0x0A, # another length value
                *[0xF9] * 3, # 12 repetitions of 0xF9
                0xFE, # end marker of table
        )

    #  - Raw LZW compressed bitmap data:
    #     This comes in blocks of different types, all preceded immediately by a type number:
    #     1 => plain text, followed by a stream of bytes representing characters.
    #            Each character gets compressed to a corresponding sequence of codes according
    #            to the current dictionary. In practice, the dictionary just keeps track of
    #            the last 12 codes (or a few more depending on the initial code size), so the
    #            most common codes are actually repeated frequently enough in practice to
    #            encode them efficiently. So basically, we get a bunch of "AA" sequences,
    #            then a bunch of "BB",... Then we start getting "AAA", "AAB",...,
    #            "AAAA", "AAAB"...etc.
    #     2 => compressing the above stream of bytes back into shorter sequences of bytes.
    #            We use the same dictionary so only changes in compression are recorded.
    #     3 => EOF, no bytes follow except possibly padding.
    #     4 => application extension. No bytes follow except those needed to complete the block.
    #            This usually includes a comment about who created the file ("Created by GIMP")
    #     5 => image description. Same as appext but generally longer and less informative.
    def make_frame_data():
        # First frame data starts with a table of LZW lengths (see above function).
        yield from make_table()

        # Next, the actual frame data begins. It consists of a bunch of chunks of various
        # types whose meaning is specified in the preceding paragraph.

        # Plain text chunk:
        #  - a single byte indicating what type of data we're looking at. Here it's 1, which
        #    stands for "plain text".
        #  - a series of bytes representing characters. These characters are encoded according
        #    to our dictionary mentioned earlier.
        #    Since we've got plenty of dictionary available already due to the amount of text
        #    being the same throughout frames, it turns out that these characters are usually
        #    already represented in our dictionary and don't need to be newly added to it.
        #    Every once in a while though, we do add new stuff to our dictinoary to represent
        #    a unique character. Usually 2 or 3 bytes long, when they can't be