"""
Generate a random GIF from /dev/urandom.
"""

import struct
import sys
from argparse import ArgumentParser
from typing import BinaryIO


def parse_args() -> str:
    parser = ArgumentParser("generate-random-gif")

    # Generate a default name if none is provided
    if len(sys.argv) == 1:
        return "random.gif"

    # Allow passing an argument to specify output filename.
    parser.add_argument("-o", "--output", type=str, help="set destination for generated image (default: %(default)s)")
    ns = parser.parse_args()

    if not ns.output:
        raise RuntimeError('No output specified.')
    
    return ns.output


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Randomly select one of 6 colors
    color_indexes = [i for i in range(0, 7)]
    color_count = len(color_indexes) - 1
    current_color = next(rng) & 0x3f

    # Set initial background color and repeat count
    out.write("\x47\x49\x46\x38\x39\x61\x00\x1e".encode())
    out.write(struct.pack("<H", 255))   # width = 255 pixels
    out.write(struct.pack("<H", 255))   # height = 255 pixels

    # Number of logical palette entries (7 in this case)
    out.write("\x07\x00".encode())

    # 3 RGB triples (each 3 bytes)
    out.write((b"\xff\x00\x00", b"\x00\xff\x00", b"\x00\x00\xff", b"\xff\x00\xff", b"\x00\xff\xff", b"\xff\xff\x00", b"\xffff\xff").get(current_color, b"\x00\x00\x00"))

    # Background index, aspect ratio byte, and global color table flag
    out.write(b"\x00\x00\x3f")

    # Start generating data
    f1 = False
    f2 = False
    while True:
        v = next(rng)
        
        # Avoid generating too many frames in quick succession.
        if f2:
            f1 = True
            
        if f1:
            f2 = True
            out.write(struct.pack("!BB", 0x2c, 0x00))
            out.write(b"\x00\x00\x00\x00")
            out.write(struct.pack("<B", current_color))

            # Pick a new color at random
            # This could be optimized by eliminating the colors already used.
            new_color = color_indexes[v % color_count]
            if current_color != new_color:
                out.write(struct.pack("<B", new_color))
                current_color = new_color
                
        else:
            # The first frame is full black, so we can just write a list of random numbers.
            out.write(struct.