#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# between minwidth <= w < maxwidth and minheight <= h < maxheight (inclusive).
# The frame count is randomly generated from 1 to nframes inclusive.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, minwidth=64,
        maxwidth=512, minheight=64, maxheight=512, nframes=10):

    # construct headers
    
    def pack(*args):
        return struct.pack("!" + "".join([x for x in args if x != None]), *args)
    
    assert minwidth % 2 == 0
    assert minheight % 2 == 0
    assert maxwidth % 2 == 0
    assert maxwidth >= minwidth
    assert maxheight % 2 == 0
    assert maxheight >= minheight
    
    out.write(b'GIF89a')
    screen = pack('H', minwidth)
    out.write(screen)
    out.write(screen)
    
    # animation control block
    # has fields for background color index, aspect ratio, sort flag, and
    # global color table flag
    acblock = pack('B', ((minwidth >> 7 * 2 - 1) & 0xff),
                   ((maxwidth >> 7 * 2 - 1) & 0xff))
    acblock += pack('B', ((minheight >> 7 * 2 - 1) & 0xff),
                    ((maxheight >> 7 * 2 - 1) & 0xff))
    acblock += pack('B', nframes - 1)
    acblock += b'\x00'
    out.write(acblock)
    
    # logical screen descriptor
    # has fields for aspect ratio, sort flag, and global color table flag
    lsd = pack('B', 0, 0, 0, 0)
    out.write(lsd)
    
    # global color table
    # contains 256 colors with the same RGB values
    ct = []
    for i in range(256):
        ct.append(struct.pack("BBB", i, i, i))
        
    out.writelines(ct)
    
    # write frames
    for i in range(nframes):
        # frame header
        fh = pack('H', minwidth)
        out.write(fh)
        out.write(fh)
        
        # graphics control extension block
        gcb = pack('B', 0xf9, 4, 0, 0)
        out.write(gcb)
        
        # write data
        for j in range(minwidth // 2):
            out.write(rng.read(minheight // 2))
            
        # trailer
        out.write(b';')
    