#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generate a random GIF given an input stream of bytes to use for randomness and output stream to write to.
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    # Create the header
    header = b"GIF89a"

    # Get the length of the image from the RNG
    width  = (int.from_bytes(rng.read(1), 'big') >> 4) + 1
    height = ((int.from_bytes(rng.read(1), 'big') & 0xf) << 8) | int.from_bytes(rng.read(1), 'big')
    print("Generating {}x{} GIF".format(width,height))

    # Write it all together
    out.write(header)
    out.write(struct.pack('>H', width))
    out.write(struct.pack('>H', height))

    # Background color index is just zero
    out.write((b'\x00'))
    out.write((b'\x00\xFF\xFF\xFF'))

    # LZW minimum size == 2
    mincodesize = 2
    out.write(mincodesize.to_bytes(1,'little'))
    
    # Image flags are packed 1 bit each with as follows:
    #   1 - Color table sort flag
    #   2-4 - GCT size (LZW minimum size)
    #   5   - Local colors sorted flag
    #   6   - Interlace flag (NOT USED)
    #   7   - Sort flag (NOT USED)
    #   8   - Reserved for future use

    # We set interlace off, color table sorted on, local colors sorted off,
    # and everything else off that we don't care about.
    gctsortflag = 1
    lctsortedflag = 0
    interlaced = False
    sortflag = 0
    reserved = 0
    flags = gctsortflag | (mincodesize << 2) | lctsortedflag
    out.write((flags).to_bytes(1,'little'))

    out.write(b"\xff") # End of graphic control extension block

    # Graphic control extension parameters
    gceextsize = 4      # Always 4 bytes long
    bboxexists = False  # Never
    disposalmethod = 0  # No action
    userinput = 0       # User not present
    transparencygiven = 0 # Not used
    delaytime = 0       # Not used
    transparentcolorindex = 0 # Not used

    gceextensionblock = \
         gceextsize.to_bytes(1,"little") + \
         bboxexists.to_bytes(1,"little") + \
         disposalmethod.to_bytes(1,"little") + \
         userinput.to_bytes(1,"little") + \
         transparencygiven.to_bytes(1,"little") + \
         delaytime.to_bytes(2,"little",signed=True) + \
         transparentcolorindex.to_bytes(1,"little")

    out.write(gceextensionblock)

    # Global color table
    out.write(b"\x00\xff\x00\xff\x00\xff")
    out.write(b"\xff\x00\xff\x00\x00\xff")
    out.write(b"\x00\xff\x00\x00\xff\x00")
    out.write(b"\xff\xff\x00\x00\xff\x00")
    out.write(b"\xff\xff\x00\xff\x00\xff")
    out.write(b"\xff\x00\xff\xff\x00\xff")
    out.write(b"\xff\x00\x00\xff\xff\x00")
    out.write(b"\xff\xff\xff\xff\xff\xff")

    out.write(b"\xff\x00\x00\x00\x00\xff")
    out.write(b"\xff\xff\x00\x00\x00\xff")
    out.write(b"\xff\xff\xff\x00\x00\xff")
    out.write(b"\xff\x00\xff\xff\x00\xff")
    out.write(b"\xff\xff\xff\xff\x00\xff")
    out.write(b"\xff\xff\x00\xff\xff\xff")
    out.write(b"\xff\x00\xff\xff\xff\xff")

    # Image descriptor block
    idb = \
         b"\x2c" + \
         struct.pack('<H', width) + \
         struct.pack("<H", height) + \
         b"\x00" + \
         b"\x00"

    out.write(idb)

    # Local color map
    out.write(b'\x00\xff\x00\xff\x00\xff')
    out.write(b'\xff\x00\xff\x00\x00\xff')
    out.write(b'\x00\xff\x00\x00\xff\x00')
    out.write(b'\xff\xff\x00\x00\xff\x00')
    out.write(b'\xff\xff\x00\xff\x00\xff')
    out.write(b'\xff\x00\xff\xff\x00\xff')
    out.write(b'\xff\x00\x00\xff\xff\x00')
    out.write(b'\xff\xff\xff\xff\xff\xff')

    # Add the data
    pixelcount = 0
    while True:
        nextbyte = rng.read(1)
        if len(nextbyte) == 0:
            break
        byte = int.from_bytes(nextbyte, "big")
        low  = byte >> 4
        high = byte & 0xf
        if low != high:
            assert(low < 0xF)
            assert(high < 0xF)
            out.write(low.to_bytes(1, 'l