#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# between minwidth <= w < maxwidth and minheight <= h < maxheight (inclusive).
# The frame count is randomly generated from 1 to nframes inclusive.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, minwidth=64,
        maxwidth=512, minheight=64, maxheight=512, nframes=10):

    # construct headers
    header = b'GIF87a'
    width = rng.randint(minwidth,maxwidth-1)//2*2+2
    height = rng.randint(minheight,maxheight-1)//2*2+2
    gctsize = 2 << bool(rng.randint(0,1))
    globalcolortablelen = 2**(gctsize+1)-1
    localcolortablelen = 2**((bool(rng.randint(0,1)))<<2)
    fc = rng.randrange(nframes)+1
    extblocks = [
        (b'NETSCAPE2.0', b'\x03\x01\x00\x00\xff\x00'),
        (b'ANIMEXTS1.0', b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')]


    # write headers
    out.write(header)
    out.write(struct.pack('<HHB', width, height, gctsize))
    
    # write global color table
    for _ in range(globalcolortablelen):
        out.write(bytes([rng.getrandbits(8),]*3))

    # write frames
    for _ in range(fc):
        
        out.write(b';')
        
        # write graphic control extension block
        out.write(b'\xF9\x04\x08\x00\x00'.replace(b'\x08', bytes([rng.randrange(10)])))

        if not extblocks[0][1]:
            out.write(extblocks[0][1])
            
        # write image descriptor
        out.write(b',' + b'!\x00\x00\x00\x00'.replace(b'!', bytes([rng.randrange(10)])))
        
        # write local color table
        for _ in range(localcolortablelen):
            out.write(bytes([rng.getrandbits(8),]*3))
        
        # write pixel data
        for y in range(height):
            pxdata = []
            
            for x in range(width):
                pxdata += [rng.getrandbits(8)]
                
            out.write(b''.join([bytes([pxdata[i], pxdata[i+1], pxdata[