#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:        BinaryIO=None,
        width:      int    = 1024,
        height:     int    = 768,
        colordepth: int    = 8,
        bgcolor:    int    = None,
        outfile:    str    = 'output.gif',
    ) -> None:

    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    blackcode = bytes([bgcolor & (255**2)])
    whitecode = bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])
    
    with open(outfile,'wb') as f:

        # The GIF header
        version = b"89a"
        screendescriptor = struct.pack('>HHBBB',
                width,
                height,
                0xF0 | colordepth,
                0,
                0,
        )
        
        # Write the header and descriptor block
        print("Writing the header")
        f.write(b"GIF"+version+screendescriptor)

        # Build the color lookup table
        print("Building the color map")
        colortable = []
        for i in range((2**colordepth)-1):
            x = (i*(bgcolor**(1/(256.0))))%(256**3)
            colortable += [struct.pack('>BBB',
                    int(((x>>16)&255)**1.5),
                    int(((x>>8)&255)**1.5),
                    int((x&255)**1.5),
                ),
            ]
        colortable += [blackcode]
        colortable += [whitecode]

        # We can now write the palette.
        print("Writing the palette")
        f.write(b"".join(colortable))

        # Now we need to write the image data itself...
        print("Writing the image")
        f.write(b"\x2C\x00\x00\x00\x00\x00\x00\x00\xFF")   # Image descriptor

        for y in range(height):
            row = bytearray()
            for x in range(width):
                try:
                    code = struct.unpack(f'<B{colordepth}x',
                            colortable[rng.read(1)[0]])[0]
                except Exception:
                    code = rng.read(1)[0]
                row.append