#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' + bytes([0]) * 7

    width, height = (1 << int.from_bytes(rng.read(1), 'big')) % 500\
           , (1 << int.from_bytes(rng.read(1), 'big')) % 500
    
    # write header
    out.write(header + width.to_bytes(2, 'little')
              + height.to_bytes(2, 'little'))

    # write palette
    for _ in range(int.from_bytes(rng.read(1), 'big')):
        out.write((rng.read(3) + b'\x00')[:4])

    # write image data
    code_size = int.from_bytes(rng.read(1), 'big') + 1
    # sub-block size (LZW minimum clear block of payload length - 1 byte
    lzw_min_clear = (((code_size & 0b111) + 1) ^ 0b111)\
                | ((code_size & 0b111) << 3)
    
    # declare LZW dictionary and store two special codes
    lzw_dictionary = {(i+1).to_bytes(2, 'little'): \
            i.to_bytes(1, 'big') for i in range(0, 512)}
    lzw_dictionary[b'\x00'] = b''
    lzw_dictionary[b'\xFF'] = b'\xFF'
    
    sub_block = [lzw_min_clear.to_bytes(1, 'big')]
    while True:
        i = int.from_bytes(rng.read(2), 'little')
        if i not in lzw_dictionary:
            k = (i & 0xff).to_bytes(1, 'big')
            
            x = (i >> 8).to_bytes(1, 'big')
            y = lzw_dictionary[k] + lzw_dictionary[x]
            while len(y) > 255:
                sub_block[-1] += lzw_dictionary[y[:255]]
                y = y[255:]
            lzw_dictionary[k] = y
            
            if len(sub_block[-1]) >= 255:
                sub_block.append(lzw_min_clear.to_bytes(1, 'big'))
        
        else:
            sub_block[-1] += lzw_dictionary[i.to_bytes(2, 'little')]
        
        if len(sub_block[-1]) >= 255:
            sub_block.append(l zw_min_clear.to_bytes(1, 'big'))

        # end of image
        if i == 0:
            break

    # Flags
    global_compression_flag = 0
    local_color_table_flag = 0
    interlace_flag = 0
    sort_flag = 0
    reserved = 0
    background_index = 0
    pixel_aspect_ratio = 0
    
    # Write graphic control extension
    out.write(struct.pack('<3sb3sb3sbbbbbb',
                         0x21,          # Extension Introducer
                         b'F',          # Label ("F"-ile)
                         b'9',          # Block Length ("9")
                         b'\x04',       # compression method
                         global_compression_flag.to_bytes(1, 'big'),
                         local_color_table_flag.to_bytes(1, 'big'),
                         interlace_flag.to_bytes(1, 'big'),
                         sort_flag.to_bytes(1, 'big'),
                         reserved,
                         background_index,
                         pixel_aspect_ratio))

    # Write graphic identifier
    out.write(struct.pack('<2sb',
                         0x01,    # Label (Graphic Control)
                         b'\x0B', # Block Length (11)
                         0x00,    # transparent color index
                         0x00,    # delay time
                         0x00))   # transparent color index

    # write application extension
    out.write(struct.pack('<3sb3sb3sbbbbb',
                         0x21,      # Extension Introducer
                         b'X',      # Application label ("X"-tended)
                         b'\x00',   # Block Length (0)
                         b'!',      # Application id (logical screen descriptor)
                         b'\x09',   # Application data (logical screen descriptor)
                         width.to_bytes(2, 'little'),
                         height.to_bytes(2, 'little'),
                         0x01,      # packed field
                         0b10))     # has global color table flag set
    
    # write end-of-file block
    out.write(b'\x3b')

    # Write sub blocks
    for s in sub_block:
        out.write(s)

