
import struct
from .gif import GIFError

def encode_graphic(sub_block):
    """Encodes a sequence of sub blocks into an animated gif.

    Parameters
    ----------
    sub_block : list[bytes]
        A sequence of bytes representing the contents of each sub block.

    Returns
    -------
    bytes
        Encoded GIF file as bytes.

    Raises
    ------
    ValueError
        If too many sub blocks are supplied.
    GIFError
        If the dimensions or number of colors do not match.
    """
    
    if len(sub_block) > 65535:
        raise ValueError("Too many sub blocks.")

    # Extract information from first sub block
    _, type_id, width, height, depth, *other = struct.unpack('<3sbxx2HHH', sub_block[0])
    if type_id != 0x02 or other != []:
        raise GIFError

    num_colors = 2 ** depth

    # Check that all sub blocks have same dimension and color depth
    for s in sub_block:
        t, type_id, w, h, d, *other = struct.unpack('<3sbxx2HHH', s)
        if t != 0x02 or type_id != 0x02 or other != []:
            raise GIFError
        if w != width or h != height or d != depth:
            raise GIFError

    # Set up boolean flags for the logical screen descriptor
    global_compression_flag = False
    local_color_table_flag = True
    interlace_flag = True
    sort_flag = True
    reserved = b"\x00\x00"
    background_index = 0
    pixel_aspect_ratio = b"\x00"
    
    # Create output stream
    out = bytearray()
    
    # Write header
    out.extend(b'GIF87a')

    # Write logical screen descriptor
    out.write(struct.pack('<3sBBB2s',
                         b'F',          # Label ("F"-ile)
                         b'9',          # Block Length ("9")
                         b'\x04',       # compression method
                         global_compression_flag.to_bytes(1, 'big'),
                         local_color_table_flag.to_bytes(1, 'big'),
                         interlace_flag.to_bytes(1, 'big'),
                         sort_flag.to_bytes(1, 'big'),
                         reserved,
                         background_index,
                         pixel_aspect_ratio))

    # Write graphic identifier
    out.write(struct.pack('<2sb',
                         0x01,    # Label (Graphic Control)
                         b'\x0B', # Block Length (11)
                         0x00,    # transparent color index
                         0x00,    # delay time
                         0x00))   # transparent color index

    # write application extension
    out.write(struct.pack('<3sb3sb3sbbbbb',
                         0x21,      # Extension Introducer
                         b'X',      # Application label ("X"-tended)
                         b'\x00',   # Block Length (0)
                         b'!',      # Application id (logical screen descriptor)
                         b'\x09',   # Application data (logical screen descriptor)
                         width.to_bytes(2, 'little'),
                         height.to_bytes(2, 'little'),
                         0x01,      # packed field
                         0b10))     # has global color table flag set
    
    # write end-of-file block
    out.write(b'\x3b')

    # Write sub blocks
    for s in sub_block:
        out.write(s)

