#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generate a random GIF given an input stream of bytes to use for randomness and output stream to write to.
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    # Create the header
    header = b"GIF89a"

    # Set the screen dimensions to something ridiculously high because it makes things easier
    # NOTE: This could be optimized to make the height smaller if the aspect ratio is not square!
    width = 500
    height = 500
    wh = width * height
    header += width.to_bytes(2, 'little')
    header += height.to_bytes(2, 'little')

    # Read in our random values for the background and foreground colors from the file we were passed
    bgcolor = int.from_bytes(rng.read(3), byteorder='big')
    fgcolor = int.from_bytes(rng.read(3), byteorder='big')

    # Write the background and foreground colors
    header += bgcolor.to_bytes(1, 'big', signed=False)
    header += fgcolor.to_bytes(1, 'big', signed=False)

    # Go ahead and start writing the result
    out.write(header)

    # The next few bytes are the control block to specify characteristics of the image.
    # See https://www.fileformat.info/format/gif/corion.htm for specs on the different bits
    # and their meanings.

    # First we want to decide how many bits to use when compressing each pixel
    # There should be at least one bit per channel and every other bit after that
    maxbits = 1 + len(struct.pack("L", bgcolor)) + len(struct.pack("L", fgcolor))
    mincodesize = 2 ** (maxbits+1) // (bgcolor+fgcolor+2)

    # Next we have the img flags which are composed of 4 1-bit integers:
    #   2-4 - GCT size (LZW minimum size)
    #   5   - Local colors sorted flag
    #   6   - Interlace flag (NOT USED)
    #   7   - Sort flag (NOT USED)
    #   8   - Reserved for future use

    # We set interlace off, color table sorted on, local colors sorted off,
    # and everything else off that we don't care about.
    gctsortflag = 1
    lctsortedflag = 0
    interlaced = False
    sortflag = 0
    reserved = 0
    imgflags = (gctsortflag << 7) + (lctsortedflag << 6) \
                + (interlaced << 5)     + (sortflag << 4) \
                + (reserved << 3)       + mincodesize
    out.write(imgflags.to_bytes(1, 'little'))


    # Skip the rest for now because I'm gonna need to learn more about this format later.

    return