#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# 
# The file will have 10 frames in it, with 5 pixels of
# transparent background, and a variable-sized image.
def generate_gif(out :BinaryIO, rng):
    # GIF header (6 bytes)
    out.write(b'GIF89a')
    
    # Logical screen descriptor (7 bytes)
    # Width: 255 pixels + 1 (LSB = minimum). Note that this
    # is for a single frame. It will be multiplied by the number
    # of frames later on.
    width_lsb = rng.randint(1, 255)
    out.write(struct.pack('<BBB', 255, width_lsb, 1))
    
    # Height: 255 pixels + 1 (LSB = minimum).
    height = rng.randint(1, 255)
    out.write(struct.pack('<BB', height, 1))
    
    # Background color index
    out.write(struct.pack('B', 0xFF & rng.randint(0, 255)))
    
    # Pixel aspect ratio (unused)
    out.write(b'\0')
    
    # Global Color Table Flag (incomplete)
    out.write(b'\x80')
    
    # Color resolution
    out.write(b'\x07')
    
    # Sort flag, size
    out.write(b'\x00\x00')
    
    # Number of colors
    out.write(b'\xF8')
    
    # Background color
    out.write(b'\0')
    
    # Palette
    palette = b''
    for i in range(0, 4 * 3 - 1):
        palette += struct.pack('B', 0xFF & rng.randint(0, 255))
    out.write(palette)
    
    # Application Extension block
    # Block length, label "NETSCAPE"
    out.write(struct.pack('<BB', 0x0B, ord('N')))
    out.write(b'ETSCAPE')
    # Block subtype: loop count extension
    out.write(struct.pack('<H', 0x03))
    # Block length, Netscape loop count (forever)
    out.write(struct.pack('<HH', 0x01, 0x00))
    
    # Image Descriptor block 0 (10 bytes)
    # Separator, left position, top position
    out.write(b';\0\0\0,\0\0\0')
    
    # Width: 255 pixels + 1 (LSB)
    out.write(struct.pack('<BB', 255, width_lsb))
    
    # Height: 255 pixels + 1 (LSB)
    out.write(struct.pack('<BB', height, 1))
    
    # Local Color Map Flag, Interlace Flag
    out.write(b'\x00\x08')
    
    # Minimum code size
    out.write(b'\x02')
    
    # LZW dictionary size
    out.write(b'\x04')
    
    # Bits per pixel:
    bits_per_pixel = rng.randint(2, 8)
    
    # Frame data
    frame_data = [ord('\x00'), ord('!')]
    x = 0
    y = 0
    bit_stream = []
    byte_length = bits_per_pixel // 8
    if bits_per_pixel % 8 != 0:
        byte_length += 1
    while True:
        # If we need to start a new byte...
        if len(bit_stream) == bits_per_pixel:
            # Start a new byte
            frame_data.append((sum(bit_stream) << (8 - bits_per_pixel)) & 0xFF)
            
            # Reset the bit stream
            bit_stream = []
        
        # Read a bit from RNG
        bit = 1 & rng.getrandbits(1)
        
        # Add the bit to the bit stream
        bit_stream.append(bit)
        
        # Increment X or Y counter depending on current bit, if necessary
        if (0b101 & x) == 0:
            # First bit of an even-numbered byte; increment X
            x += 1
        elif (0b101 & y) == 0:
            # First bit of an odd-numbered byte; increment Y
            y += 1
        else:
            # Not first bit of either an even- or odd-numbered byte;
            # advance both X and Y
            x += 1
            y += 1
        
        # Check that we haven't reached one of the edges
        if x >= 255:
            x = 0
        if y >= height:
            break
    
    # Finish up last byte
