#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# between minwidth <= w < maxwidth and minheight <= h < maxheight (inclusive).
# The frame count is randomly generated from 1 to nframes inclusive.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, minwidth=64,
        maxwidth=512, minheight=64, maxheight=512, nframes=10):

    # construct headers
    header = b'GIF87a'
    width = rng.randint(minwidth,maxwidth-1)//2*2+2
    height = rng.randint(minheight,maxheight-1)//2*2+2
    gctsize = 2 << bool(rng.randint(0,1))
    globalcolortablelen = 2**(gctsize+1)-1
    localcolortablelen = 2**((bool(rng.randint(0,1)))<<2)
    fc = rng.randrange(nframes)+1
    extblocks = [
        (b'NETSCAPE2.0', b'\x03\x01\x00\x00\xff\x00'),
        (b'ANIMEXTS1.0', b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')]


    # write headers
    out.write(header)
    out.write(struct.pack('<HHB', width, height, gctsize))

    # append colortables if any
    for i in range(globalcolortablelen//3 + bool(globalcolortablelen%3)):
        r,g,b = tuple([int(v*255) for v in rng.choice(range(256))][0] for _ in "rgb")
        out.write(struct.pack('BBB',r,g,b))
    
    # append frames, blocks and pixels
    for i in range(fc):
        out.write(b';')
        
        blocktype = bool(rng.randint(0,1))
        extidx = rng.randint(0, len(extblocks)-1)
        if blocktype == 0:
            # graphic control extension
            out.write(b'\xF9')
            out.write(struct.pack('B', 8 | bool(rng.randint(0,1))))
            out.write(struct.pack('B', rng.randint(0,7)))
            out.write(struct.pack('<H', rng.randint(0,2**16-1)))
        else:
            # application extension
            out.write(b'\xFF')
            out.write(extblocks[extidx][0])
            out.write(b'\x0B')
            
            out.write(extblocks[extidx][1])
            out.flush()
