#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:        BinaryIO=None,
        width:      int    = 1024,
        height:     int    = 768,
        colordepth: int    = 8,
        bgcolor:    int    = None,
        outfile:    str    = 'output.gif',
    ) -> None:

    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    blackcode = bytes([bgcolor & (255**2)])
    whitecode = bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])
    
    with open(outfile,'wb') as f:

        # The GIF header
        version = b"89a"
        screendescriptor = struct.pack('>HHBBB',
                width,
                height,
                0xF0 | colordepth,
                0,
                0,
        )
        
        # Initialize the global color table
        palette = []
        for i in range(3):
            palette += [
                b'\xFF' + (blackcode if i==0 else (whitecode if i==2 else b'\x00\x00'))
            ]

        # Write the header and screen descriptor
        f.write(version+b"NETSCAPE2.0"+screendescriptor+b"\x00")

        # Write the graphic control extension
        imageheader = (
            b'\x2C'+struct.pack(">H",   # Image position
                width-1)+struct.pack(">H",
                height-1)+b'\x00'        # No local colormap
        )
        gifcontrol = b'\x21\xF9\x04\x04\x00\x00\x00'
        f.write(imageheader+gifcontrol+b'\xFF')

        # Write the palette
        f.write(b'\x3B'+"".join(palette))

        # Loop over all pixels, writing each one at a time
        pixels = bytearray((width*height)//3)
        while True:
            
            # Create a new pixel value
            newvalue = rng.read(3) if rng is not None \
                else bytes(