#!/usr/bin/env python3

import struct
from typing import List, Tuple, Union
from collections import defaultdict
import argparse
import os
import re

_REGION_MARKER = "__LLVM_LIT_REGION_"

def main(args):
  with open(args.inputfile) as f:
    lines = f.readlines()
  regions = parse_regions(lines)
  region_split = split_region(regions, args.split_on_header)
  if args.print_region_names:
    for name in get_region_names(regions):
      print(name)
  else:
    write_region_files([filename for (filename,_) in region_split], [content for (_, content) in region_split])

def write_region_files(filenames :List[str], contents :List[str]):
  assert len(contents) == len(filenames), "Must have same number of files and file contents"
  for i, filename in enumerate(filenames):
    with open(filename, 'w') as f:
        f.write(contents[i])

def parse_regions(lines_in :List[str]) -> List[Tuple[int, int]]:
  # removed the extra newline that was added by readlines
  lines = [(line[:-1] if line[-1] == "\n" else line) for line in lines_in]
  
  current_region = None
  regions = []

  def add_current_region():
    nonlocal current_region
    if current_region is not None:
      regions.append((current_region[0], len(lines)))
      current_region = None
  for linenum, line in enumerate(lines):
    match = re.match(_REGION_MARKER + "(.*)", line)
    if match:
      add_current_region()
      current_region = (linenum+1, match.group(1))
    elif current_region is not None:
      pass
    else:
      add_current_region()
  add_current_region()
  return regions


def split_region\
  (regions :List[Tuple[int, int]], split_on_header :bool)\
    -> List[Tuple[str, str]]:
  """Split a region into multiple regions based on whether they start with headers"""
  splits = []
  for region in regions:
    filetext = '\n'.join(lines[region[0]:region[1]])

    first_line_is_header = False
    rest_of_file = filetext
    while True:
      if split_on_header or not first_line_is_header:
        header_match = re.search('^==={1,2}', rest_of_file)
        if header_match:
          header_end_offset = header_match.span()[1]
          
          splits.append((region[2], rest_of_file[:header_end_offset]))
          rest_of_file = rest_of_file[header_end_offset:]
          first_line_is_header = True
          continue
        
      splits.append((region[2], rest_of_file))
      break
    
  return splits

def get_region_names(regions :List[Tuple[int, int]]) -> List[str]:
  names