#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' + bytes([0]) * 7

    width, height = (1 << int.from_bytes(rng.read(1), 'big')) % 500\
           , (1 << int.from_bytes(rng.read(1), 'big')) % 500
    
    # write header
    out.write(header + width.to_bytes(2, 'little')
              + height.to_bytes(2, 'little'))

    # write palette
    for _ in range(int.from_bytes(rng.read(1), 'big')):
        out.write((rng.read(3) + b'\x00')[:4])

    # write image data
    code_size = int.from_bytes(rng.read(1), 'big') + 1
    # sub-block size (LZW minimum clear block of payload length - 1 byte
    lzw_min_clear = (((code_size & 0b111) + 1) ^ 0b111)\
                | ((code_size & 0b111) << 3)
    
    # declare LZW dictionary and store two special codes
    dicts = [None] * 4096
    dicts[1] = '\x01' * 256
    dicts[2] = ''.join('\xff') * 256

    def lzw_decompress():
        buf = b''
        while True:
            c = rng.read(1)[0]
            
            if c == 0: break # EOF
            
            elif c <= len(dicts):
                yield from dicts[c]
                continue
                
            elif c > len(dicts):
                raise RuntimeError('Dictionary overflow @ index {}'
                                   .format(c))
        
    def lzw_encode(inp):
        if not isinstance(inp, bytes):
            inp = str(inp).encode()
        
        for i in range(len(dicts)):
            current_len = len(dicts[i])
            if dicts[i].endswith(inp):
                return i + 256 if i < 256 else i
            elif current_len >= 256:
                if current_len == code_size + 1:
                    del dicts[i][:-1]
                else:
                    dicts.insert(i, dicts[i][:-1])
                break
            elif current_len == code_size:
                del dicts[-1]
                break
        else:
            raise RuntimeError('Dictionary full')
        
        inp = inp + bytes([0])[:1 + (current_len < 256)]
        dicts.append(inp)
            
        return current_len + 256 if current_len < 256 else current_len
        
    # write graphics control extension with transparent color index 0
    gce_payload = struct.pack("<BBHHB", 0xf9, 0x04, 0, 0, 0, 0)
    out.write('\x21' + '\xF9' + gce_payload)

    iters = rng.read(1)[0]
    for _ in range(iters):
        buf = []
        while len(buf) % l