#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    header += struct.pack('>H', rng.randint(200, 500))
    header += struct.pack('>H', rng.randint(200, 500))

    # color table (RGB triplets)
    ctsize = rng.randint(1, 7) * 3
    for _ in range(ctsize):
        header += bytes([rng.randint(0, 255)])

    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen) + b'\x00\x2c'

    while len(data) < 4096*1024 - 1:

        sublen = rng.randint(
                min(
                    255,
                    max(
                        1,
                        datalen // rng.randint(
                            2,
                            6))),
                255) | 0xfe

        if rng.random() >.15:

            # raw image data
            blocktype = ord('!')

            blklen = rng.randint(
                    max(sublen//2, 1),
                    min(sublen, 255)) |\
                    0xff << 8

            # no interlacing
            flags = 0b11100000 & rng.randint(0xf0, 0xff)

            if flags == 0x20:

                # subblock length is not specified
                blklen |= len(data) & 0xff

            elif flags!= 0x00:

                # TODO: add custom subblocks with graphic control extension
                raise NotImplementedError()
            
            data += struct.pack('<BB', blocktype, blklen >> 8)
            data += bytes([(blklen & ~0xff) >> 8])
            data += bytes([flags])
            data += bytes([datalen & 0xff])
            data += bytes([(datalen >> 8) & 0xff])
            data += bytes([(datalen >> 16) & 0xff])
            data += data[-sublen:]

        else:

            subdata = data[max(-datalen-1, -255)-1 : \
                            min(datalen+255, len(data))]

            if rng.random() >.8:
                
                # use local colour table
                lctsize = rng.randint(1, 7) * 3
                
                for i in range(lctsize):

                    if i >= ctsize:

                        # new colors start at 0
                        break
                    
                    r = ((header[i] or 0)
                        + rng.randint(-255, 255)) % 256

                    g = ((header[ctsize+i] or 0)
                        + rng.randint(-255, 255)) % 256

                    b = ((header[ctsize*2+i] or 0)
                        + rng.randint(-255, 255)) % 256

                    header += bytes([r])
                    header += bytes([g])
                    header += bytes([b])

                header += bytes([0]) * (3*(7-lctsize)%0xC)

                lctflag = 0x80

                lctsize = (lctsize)

                if lctsize <= 1:

                    lctsize = 0


                lctsize |= 0b111 << 4

                imghdr = struct.pack('<BBBHHB', 0x2C,
                        lctsize, lctflag|imgtype, xpos, ypos, 0b0)


            else:

                # don't use local colour table
                imghdr = struct.pack('<BBBHHB', 0x2c,
                        ctsize, 0x00, xpos, ypos, 0b0)

            imghdr += bytes([imgseq])

            data += imghdr
            data += subdata
            imgseq += 1

        datalen = len(data)
    
    combined = header + data
    print("{:d} bytes generated".format(len(combined)))
    
    out.write(combined)

