#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generate a random GIF given an input stream of bytes to use for randomness and output stream to write to.
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    # Create the header
    header = b"GIF89a"

    # Get the length of the image from the RNG
    width  = (int.from_bytes(rng.read(1), 'big') >> 4) + 1
    height = ((int.from_bytes(rng.read(1), 'big') & 0xf) << 8) | int.from_bytes(rng.read(1), 'big')
    print("Generating {}x{} GIF".format(width,height))

    # Write it all together
    out.write(header)
    out.write(struct.pack('>H', width))
    out.write(struct.pack('>H', height))

    # Background color index is just zero
    out.write((b'\x00'))
    out.write((b'\x00\xFF\xFF\xFF'))

    # LZW minimum size == 2
    mincodesize = 2
    out.write(mincodesize.to_bytes(1,'little'))
    
    # Image flags are packed 1 bit each with as follows:
    #   1 - Color table sort flag
    #   2-4 - GCT size (LZW minimum size)
    #   5   - Local colors sorted flag
    #   6   - Interlace flag (NOT USED)
    #   7   - Sort flag (NOT USED)
    #   8   - Reserved for future use
#    print("\nImage Flags:")
#    imageflags = b"\x80"  # Color table sorting ON
#    imageflags += b"\xe0"  # LZW minimum size == 12
#    imageflags += b"\xc0"  # Local colors sorted OFF (not used in this case)
#    out.write(imageflags)
        
    # Global color table size -- we only have one entry here so set it to that
    gct_size = 1
    out.write(((gct_size & 0b0001111) << 4).to_bytes(1,'big'))

    # Set up local color table too if needed
    lct_size = None
    if gct_size!= 1:
        lct_size = max(mincodesize+1,gct_size+1)
        out.write(((lct_size & 0b0001111) << 4).to_bytes(1,'big'))

    # Write the palette
    for i in range(pow(2,gct_size)):
        out.write((i % 0xff).to_bytes(1,'big'))
        out.write((i // 0xff ).to_bytes(1,'big'))
        out.write((i * i % 0x3f).to_bytes(1,'big'))
#        print("{:#04x} {:#04x}".format(i%0xffff,(i*i)%0xffff))

    # Write the code table
    code_table = []
    for i in range(max(2**mincodesize,2**(lct_size-1)), pow(2,lct_size)):

        # Generate a code word by writing a byte unless it's already present
        codeword = [ ]
        while True:
            newbyte = rng.read(1)[0]
            if len(set(codeword) ^ {newbyte}) > 0 or len(codeword) >= mincodesize:
                break

            # Add it to our code word list
            codeword.append(newbyte)

        # Sanity check on our generated value
        assert len(codeword) <= mincodesize

        # Make sure none of these codes overlap with any other
        assert not any([any([xx==yy for xx in code]) for yy in code_table])

        # We can encode this as either a single byte or two bytes depending on what's necessary
        while len(codeword) < mincodesize:
            codeword.insert(-1, 0)

        # Convert the array of bytes into a number
        codeword = sum([(c<<(8*(len(codeword)-ii))) for ii, c in enumerate(codeword)])

        # Save off our value
        code_table.append(codeword)

    # Now dump the lookup table into binary
    for code in code_table:
        # First byte is how many msb are valid, rest are the actual msb bytes
        numvalidbits = min(8,len(code))
        out.write((((numvalidbits&0b0001111)<<8)|(code>>8)).to_bytes(2,'big'))
        out.write((code & 0xff).to_bytes(1,'big'))


    # Image descriptor block
    out.write(b'\x2c' )
    # Left position
    xpos = 0
    out.write(xpos.to_bytes(2,'little'))
    # Top position
    ypos = 0
    out.write(ypos.to_bytes(2,'little'))
    # Width
    out.write(width.to_bytes(2,'little'))
    # Height
    out.write(height.to_bytes(2,'little'))
    # Size of local color table (if used, always 1<<lct_size if global size is less than mincodesize)
    if lct_size:
        out.write(((lct_size & 0b0001111) << 4).to_bytes(1,'big'))
    else:
        out.write((gct_size & 0b0001111).to_bytes(1,'big'))
    # Image flags again
    imageflags = b"\xe0"  # LZW minimum size == 12
    imageflags += b"\x00"  # No interlace
    imageflags += b"\x00"  # No sort
    out.write(imageflags)

    # Handle ending character(s) left over... pad until next multiple of 256 (since we're done with byte writes now)
    charcount = 0
    while True:
        charcount += 1
        out.write(rng.read(1))
        if charcount % 256 == 0:
            break
        
    
if