#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:        BinaryIO=None,
        width:      int    = 1024,
        height:     int    = 768,
        colordepth: int    = 8,
        bgcolor:    int    = None,
        outfile:    str    = 'output.gif',
    ) -> None:

    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    blackcode = bytes([bgcolor & (255**2)])
    whitecode = bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])
    
    with open(outfile,'wb') as f:
    
        # First things first - write out some headers!
        version = b'89a'                                    # GIF format version
        screendescriptor = struct.pack('<HHBBBBBHHH',
                width%2<<7 | height%2<<6,                   # Screen dimensions
                0,                                           # Packed field - reserved bits set to zero.
                colordepth,                                  # Color resolution - no more than this.
                0,                                           # GCT present flag.
                0,                                           # Sort flag.
                0,                                           # Size of global colour table.
                2**(int(colordepth / 3)),                     # Background color index.
                0,                                           # Pixel aspect ratio.
                0,
        )
        
        # Write the header and descriptor block
        print("Writing the header")
        f.write(b"GIF"+version+screendescriptor)

        # Build the color lookup table
        print("Building the color map")
        colortable = []
        for i in range((2**colordepth)-1):
            x = (i*(bgcolor**(1/(256.0))))%(256**3)
            colortable += [struct.pack('>BBB',
                    int(((x>>16)&255)**1.5),
                    int(((x>>8)&255)**1.5),
                    int((x&255)**1.5),
                ),
            ]
        colortable += [blackcode]
        colortable += [whitecode]

        # We can now write the palette.
        print("Writing the palette")
        f.write(b"".join(colortable))

        # Now we need to write the image data itself...
        print("Writing the image")
        f.write(b"\x2C\x00\x00\x00\x00\x00\x00\x00\xFF")   # Image descriptor
        f.write(b"!\xF9\x04\x01\x00\x00\x00,\x00\x00\x00") # Graphics control extension
        f.write(b"\x00\x01\x00\x21\xFE\x09D\x0B\xCF\x00") # Application Extension
        f.write(b"\x0BNETSCAPE2.0\x03\x01")                  # Application Data
        f.write(b"\x00\x00\x00!")                           # Trailer
        print("\nDone.")