#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# 
# The resulting file will have exactly $size bytes of
# data following the GifHeader.
def generate_random_gif(out: BinaryIO, rng: BinaryIO, size: int):
    # Header
    out.write(b'GIF89a')

    # Logical Screen Descriptor (dimensions = 2 x 2)
    out.write(struct.pack('<HHBBB',
        0x1E, 0x02,
        0x0B,     # Width / height bit depths are fixed to this value
        0x02,
        0xFF      # Global color table flag = True; no global color table set
    ))
    
    # Image Descriptor (first frame)
    out.write(bytes([
        ord('!'),       # Image separator
        ord('?'),       # Left = 0 | Top = 0
        ord('&'),       # Width = 2 | Height = 2
        0x7D,           # Local color table flag = False; no local color table specified
        0,              # LZW minimum code size parameter; will be dynamically generated by encoder
    ]))

    # Still-image data (all black pixels in first frame)
    out.write(bytes([
        0x2C,            # Extension introducer
        ord('%'),         # Graphics control block
        4,               # Block size = 4
        b'\x00\x00',     # Delay time = 0
        b'\x00',          # Transparent color index is not given
        ord(','),         # Terminator
    ]))
    for _ in range(4 * 4):
        out.write(b'\x00')
    
    # Image descriptor (second frame)
    out.write(bytes([
        ord('!'),       # Image separator
        ord('@'),       # Left = 0 | Top = 0
        ord('A'),       # Width = A | Height = A
        0x7D,           # Local color table flag = False; no local color table specified
        0,              # LZW minimum code size parameter; will be dynamically generated by encoder
    ]))

    # Still-image data (all white pixels in second frame)
    out.write(bytes([
        0x2C,            # Extension introducer
        ord('%'),         # Graphics control block
        4,               # Block size = 4
        b'\x00\x00',     # Delay time = 0
        b'\x00',          # Transparent color index is not given
        ord(','),         # Terminator
    ]))
    for _ in range(ord('A')*ord('A')):
        out.write(b'\xFF')


    # Padding
    while len(out.getbuffer()) < size + out.tell():
        byte = rng.read(1)[0] % 256
        out.write(byte.to_bytes(1, 'little'))
    
