#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # LZW compression
    lzwlen = (datalen + 2).to_bytes(3, byteorder='big', signed=False)
    print('lzwlen:', lzwlen)
    compressed = compress(data)
    
    # Determine values for the control block
    num_colors = len({*compressed})
    if num_colors > 128:
        raise Exception("Too many colors")

    bitsize = ((num_colors - 1).bit_length() | 7)+1
    palette = [(0,0,0)] * num_colors


    data = [compressed[i] << 1 for i in range(len(compressed))]

    # Write the file
    write_chunk(header, out)
    write_chunk(b'IHDR',
                out,
                b''
                b'\x01\x00'    # Screen width
                b'\x01\x00'    # Screen height
                 b'\xf1'        # GCT: yes
                              # Sort flag: no
                              # Color resolution: 7 bits
                              # Global color table: present
                              # Background color index: 0
               ,
                b'\x00'       # Pixel aspect ratio
               )
    
    write_chunk(b'PLTE', *[
            *(struct.pack('BBB', r, g, b) for (r,g,b) in palette)])

    write_chunk(b'tRNS', *[
            *([struct.pack('BBB', 0, 0, 0)]
             if num_colors == 1 else []) ])

    sz = get_code_size(len(data))
    write_chunk(
            b'IDAT',
            out,
            lzwlen+compressed
           )
    write_chunk(
            'IEND', 
            out,
           )



def read_chunk(fd):
    chunk_len = fd.read(4)
    return fd.read(int.from_bytes(chunk_len, byteorder='big'))


def write_chunk(
        id: bytes, 
        fd: BinaryIO, 
        *chunks
    ):
    fd.write((len(id)).to_bytes(2, byteorder='little'))
    fd.write(id) 
    
    for chunk in chunks:
        fd.write((len(chunk)).to_bytes(4, byteorder='big'))
        fd.write(chunk)
    

    
# Compress the given array of 8-bit integers using LZW algorithm
def compress(arr):
    code_sizes = { 
            2**n : n 
            for n in range(8)}
    
    prefix = arr[0].to_bytes(byteorder='big', length=1, signed=False)
    codes = dict([(prefix, 256)])
    current_code_size = min(code_sizes)

    next_code = 257
    result = []
    
    def add(c):
        nonlocal next_code

        result += c.to_bytes(current_code_size // 8,
                             byteorder='big',
                             signed=False)
        
        curr_code_size = code_sizes[next_code] \
                                if next_code < max(code_sizes) else None

        
        while curr_code_size is not None and \
              len(result)*curr_code_size >= 2**(curr_code_size):

            curr_code_size -= 1
            
        if curr_code_size!= current_code_size:
            
            old_codes = list(sorted(codes.items(), key=(lambda x:
                    -(len(x[0])*(code_sizes[x[1]]//8)))))[:4096]
            new_codes = {}
            for k,(v, _) in enumerate(old_codes):
                
                prefix = v[:-1] 
                suffix = v[-1:][0]

                if (new := codes.get(prefix+(suffix,))):
                    new_codes[(k, prefix, suffix)] = new
                
            new_codes = sorted(list(new_codes.values()),
                               key=(lambda n:
                                    -(len(n)*(code_sizes[n]//8))))[:4095]


            for n in new_codes:
                codes[n] = next_code 
                next_code+= 1

            code_sizes[next_code] = curr_code_size
        
    
    add(256)
        
    for ch in arr[1:]:
        word = struct.pack('B', ch)
        
        if (code:=codes.get(prefix+word)):
            prefix += word
        elif next_code <= (maxval := 2**current_code_size)-1:
            codes[prefix] = next_code
            add(next_code);
            prefix = word
            next_code += 1
        else:
            prefix += word[0]
        
        if next_code > maxval:
            current_code_size += 1
            
    # Output final symbols    
    while  len(prefix)>1 or next_code==257:
        if next_code<= (maxval:= 2**current_code_size)-1:
            codes[prefix] = next_code
            add(next_code)
        else:
            add(256)
        prefix = prefix[0]
        if next_code > maxval:
            current_code_size+= 1

        
            
    return result


def decompress(arr):
    pass

if