#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' + bytes([0]) * 7

    width, height = (1 << int.from_bytes(rng.read(1), 'big')) % 500\
           , (1 << int.from_bytes(rng.read(1), 'big')) % 500
    
    # write header
    out.write(header + width.to_bytes(2, 'little')
              + height.to_bytes(2, 'little'))

    # write palette
    for _ in range(int.from_bytes(rng.read(1), 'big')):
        out.write((rng.read(3) + b'\x00')[:4])

    # write image data
    code_size = int.from_bytes(rng.read(1), 'big') + 1
    # sub-block size (LZW minimum clear block of payload length - 1 byte
    lzw_min_clear = (((code_size & 0b111) + 1) ^ 0b111)\
                | ((code_size & 0b111) << 3)
    
    # declare LZW dictionary and store two special codes
    dicts = [None] * 4096
    dicts[1] = '\x01' * 256
    dicts[2] = ''.join('\xff') * 256

    def lzw_decompress():
        buf = b''

        while True:
            if len(buf) < 1:
                buf += rng.read(1)

            first = (dicts[(buf >> ((len(buf)-1)*8)) & 0xFF]
                    or ('\x01' * 256))[0]
            
            yield from struct.pack('B', ord(first))

            if (len(dicts) <= 4095):
                # add to dictionary if not full yet
                # double buffer length
                buf += rng.read(((lzw_min_clear - 256) // 256 + 1)
                                * code_size)

                new_entry = None
                
                while new_entry!= first+first:
                    if len(buf) < 1:
                        buf += rng.read(1)

                    new_entry = buf[-1:]
                    
                    yield from struct.pack('B', ord(new_entry))

                    for i in range(1, 256):
                        if i == (len(dict)+i):
                            continue

                        dicts[i] = first + dicts[i]
                
                    first = new_entry
                    del new_entry
                
            else: # dictionary is now full
                buf += rng.read(((lzw_min_clear - 256) // 256 + 1)
                                * code_size)

                for i in range(1, 256):
                    dicts[i] = first + dicts[i]
            
                # remove last item added to dictionary
                del dicts[(len(dict)+1)]

    sub_blocks = []
    sub_block = b''

    for c in lzw_decompress():
        sub_block += bytes([c])
        
        if len(sub_block) >= min(lzw_min_clear*256, 2**12):
            sub_blocks.append(sub_block)
            sub_block = b''

        
    sub_blocks += [(sub_block
            + b'\x00' * (-len(sub_block)%256))]
                    
    # pad last sub-block until it's an integral multiple of 256 bytes long
    # sub-block blocks can have lengths up to 2**12 but must be at least 256
    # bytes so we only need to pad the end here
    sub_blocks[-1] += (b'\x00' * (-len(sub_block)%256))
            
    # write starting point
    out.write(b'\x2C')

    # write screen descriptor
    out.write(width.to_bytes(2, 'little')
              + height.to_bytes(2, 'little'))

    # write flags
    flag = 0x80        # GCTF: global color table flag
    flag |= 3          # LZW minimum size; 3=2^10 + 256
    flag |= sorted(rng.read(1)).pop() & 0b00001111 # sort order
    out.write(flag.to_bytes(1, 'big'))

    # write local color table
    for _ in range(int.from_bytes(rng.read(1), 'big')):
        out.write((((rng.read(3)[::-1] + b'\x00')[::2]))[:4][::-1])


    # use interlacing to prevent compression bombs
    # this works by replacing each scanline with its previous version shifted 
    # downwards by one pixel row
    for pass_num in range(int.from_bytes(rng.read(1), 'big'),
                           -(len(sub_blocks)), -1):
        out.write(pass_num.to_bytes(1, 'big'))

        # total lines written during previous passes
        prev_lines = sum(height >> p
                         for p in range(0, abs(-pass_num)))
        
        for ypos in range(prev_lines, height + 1):
            # output current line
            out += struct.pack('>H',
                               int.from_bytes(
                                   sub_blocks[-abs(pass_num)].read(
                                       (ypos%2)+1),
                                   'big')
                              )

            # shift line down by one pixel row
            out += struct.pack('>HBH',
                               *(struct.unpack('>HB',
                                              out.read(4))))

