#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:        BinaryIO=None,
        width:      int    = 1024,
        height:     int    = 768,
        out:        str    ="image.gif",
    ) -> None:

    # If no RNG is given, use /dev/urandom. It's not as good
    # but it's better than nothing.
    if rng == None:
        rng = open("/dev/urandom", "rb")

    try:
        with open(out, 'wb') as f:

            # The first three bytes are always GIF89a
            f.write(b"GIF89a")
            while True:

                # First byte contains the flags; for simplicity I just set all of them
                gifheader = b'\x21\xf9\x04' + bytes([
                    rng.read(1)[0],       # Aspect ratio
                    0 |                        # Sorting flag 
                    0,                         # Color table size
                    0,                         # Background color index
                    0,                         # Pixel aspect ratio
                ])

                # Next is the screen descriptor which specifies the global colors and dimensions
                gifheader += struct.pack("<HHBBBBB",
                    width, height,          # Screen dimensions
                    0,                      # Global color resolution
                    0,                      # Background color index
                    0,                      # Pixel aspect ratio
                    0                       # GCT field terminator
                )

                # Now comes the actual frames (we're only generating one)
                framedata = struct.pack('<BBHBBBB',
                    0x2C,                     # Image separator
                    0x00,                     # Left position
                    0x00,                     # Top position
                    width, height,            # Width & height
                    0x01,                     # Flags: local color table
                    0x00,                     # LZW minimum code size
                    0x21, 0xFE                # Block end marker
                )

                # Now comes an optional color table; this time there will be two entries
                # that specify black and white as foreground and background color respectively
                blockdata = bytes([
                    2,                              # Size
                    0, 0, 0, 0,                     # Color 0: Black
                    255, 255, 255,                   # Color 1: White
                ])
                while len(blockdata) < ((len(framedata)-1)*8):
                    # Here you can add other colors! This particular GIF needs at least three
                    # more, but you have to adjust the last byte of the previous entry accordingly
                    blockdata += b"\x00"*3+struct.pack('c'*3,
                        *(rng.read(1) * 3)           # Generate a RGB triple
                    )
                
                # Write everything!
                f.write(gifheader)
                f.write(framedata)
                f.write(b"\x00")             # Frame delay in 10 ms units
                f.write(blockdata)
                f.write(b";")               # End of image block
    
    except Exception as e:
        raise SystemExit("[!] Error: Couldn't generate the GIF file: %s"%str(e).strip())


if __name__ == "__main__":

    # We want to get the seed from the user, or use urandom by default
    while True:
        seed = input(">> Random seed? ")
        if len(seed)!= 0: break
        else:              seed = "/dev/urandom"
    
    
    # Then, create a file-like object for the given seed 
    try:
        rng = open(seed, mode="rb")
    except FileNotFoundError:
        print("[!] Warning - couldn't find the specified seed :(\n>> Using \"/dev/urandom\" instead...")
        rng = open("/dev/urandom", mode="rb")


    try:
        
        # Create the output file name; we'll ask the user about their desired filename later
        outfilename = ""
        
        # Make sure the file doesn't already exist before actually doing anything
        while True:
            
            outfilename = input(">> Output file name? (%s) "%outfilename)
            if len(outfilename) == 0: outfilename = "image.gif"

            if check_outputfile(outfilename): break
            else:                          print("[-] Error: Output file \"%s\" exists!"%outfilename)

    finally:
        rng.close()                       # Don't forget to close the file even if we exit early
    

    # Now let's go through opening the output file for writing and begin writing the GIF thingy
    try:
        with open(outfilename, "wb") as f:
            generate_random_gif(rng=f)

    except KeyboardInterrupt:
        pass


# =============================[ CODE THAT NEEDS TO BE ADJUSTED FOR YOUR PARTICULAR TASK ]===========================

# Check whether an output file would overwrite something important.
def check_outputfile(outfile: str) -> bool:
    return outfile!= "/etc/passwd" and \
           outfile!= "/etc/shadow" and \
           outfile!= "/etc/group" and \
           outfile!= "/"



# ===================================[ TESTING STUFF ]======================================

# For testing purposes, here's how to make a GIF animation that displays a mandelbrot fractal.
"""
# Code for the mandelbrot fractal.
# Requires numpy.
import math
import cmath
import numpy

def mandelbrot(width, height, cx=-0.75, cy=0., scale=7./3.):

    # First, let's create our map where we'll store each pixel's complex number value
    # Each pixel has coordinates x=i*scale-0.75, y=j*scale

    array = []
    for j in range(height):
        array.append([])
        for i in range(width):
            array[height - j - 1].append(-2.)
    
    # Secondly, let's calculate our values according to the mandelbrot formula
    def iterate(c):
        z = c
        for _ in range(256):         # Iterate up to 255 times
            if abs(z) > 2.:         # If we exceed 2, stop iterating
                break
            z = z**2 + c            # Otherwise, update our value
        return float(_)            # Return the iteration count as floating point number
    
    # Thirdly, put all those numbers in the map
    for j in range(height):
        for i in range(width):
            array[j][i] = iterate(cx + i*(scale - 0.75) + scale*(1j)*(cy+j*(scale)))
    
    return numpy.array(array)

# Let's save some space and turn the above matrix into a binary string
def binarize(matrix):
    m = "".join(map(lambda row: ''.join(["{0:{fill}{align}8b}".format(int(x), fill='0', align='center') for x in row]), matrix))
    return m[:m.index("0")]

# Finally, let's export the result into a GIF file.
def export(matrix, filename, duration=100):
    with open(filename, "wb") as f:

        # First off, GIF files start with "GIF89a".
        f.write(b"GIF89a")
 
        # Then follows the logical screen descriptor. 
        # It tells us the dimensions of the matrix and the color depth per pixel used.
        w = len(matrix[0])
        h = len(matrix)
        f.write(bytes([
            0x21, 0xF9, 0x04,             # Magic Number
            0x08,                         # Version
            0x00,                         # No color table
            0x00,                        (w//256)%256, w%256,  # Width
            0x00,                        (h//256)%256, h%256,  # Height 
            0x00,                         # Local colour table
            0x08,                         # Bits per pixel
            0x00                          # No sorting required
        ]))

        # Then come the frames themselves. Each frame consists of a graphic
        # control block followed by the image data itself.
        f.write(bytes([
            ord(","),                               # Introducer
            0x00,                                  # Start position X
            0x00,                                  # Start position Y
            0x00,                                 (w//256)%256, w%256,  # Dimension
            0x00,                                 (h//256)%256, h%256,  
            ((duration//256)>>1)+1, duration%256,    0x00,       0x00,     # Delay time, pixel aspect ratio
            0x00                                    # Transparent color index
        ]))

        # After the image data are the blocks containing the compressed image.
        # We use the LZW compression algorithm to compress every row into 256 different codes.
        # Note that it won't work well if a row repeats itself over and over again.
        m = "\x00".encode().join(map(lambda r: ("\0"+r+"\xff").encode(), binarize(matrix).split("\n")))
        print(len(m))
        m = "\x0F".encode()+lzw_compress(m)+"\0".encode()
        print(len(m))
        f.write(b"".join([(len(m)//256)%256,(len                    int((x&255)**1.5),
                ),
            ]
        colortable += [blackcode]
        colortable += [whitecode]

        # We can now write the palette.
        print("Writing the palette")
        f.write(b"".join(colortable))

        # Now we need to write the image data itself...
        print("Writing the image")
        f.write(b"\x2C\x00\x00\x00\x00\x00\x00\x00\xFF")   # Image descriptor
        f.write(b"!\xF9\x04\x01\x00\x00\x00,\x00\x00\x00") # Graphics control extension
        f.write(b"\x00\x01\x00\x21\xFE\x09D\x0B\xCF\x00") # Application Extension
        f.write(b"\x0BNETSCAPE2.0\x03\x01")                  # Application Data
        f.write(b"\x00\x00\x00!")                           # Trailer
        print("\nDone.")