#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write banner (6 bytes) and logical screen descriptor (7 bytes).
    out.write(header + struct.pack('<HHBBB',
            1024, 512, 8, 0, 0))

    # Write palette (768 bytes) in color map format 0x03.
    for i in range(0, 768): out.write(b'\x03' +
            bytes([i//3%2 << 7 |
                    i//9%2 << 6 |
                    i//36%2 << 5]))

    # Write image descriptor block (10 bytes) with local colour table.
    out.write(b'\x2c\x00\x00\x00\x00\x00\x00\x00\xff')

    # Write graphic control extension (8 bytes) that does not loop forever.
    out.write(b'\x21\xf9\x04\x04\x4e\x01' +
            struct.pack('!H', round((rng.read(1)[0] & 3)*1.5)))
    
    # Write plain text extension (19 bytes) of size 128 containing some junk.
    out.write(b'\x21\xff\x0b\x49\x45\x54\x46\x38\x39\x61\x01\x00\x01\x00' +
            b'\xce\x00\x00\x00\x00\x00\x0b\x49\x45\x54\x46\x20\x38\x39\x61' +
            b'\x20\x20\x20\x20\x20\x20\x20\x20\x20\x00\x00\x00\x00\x00')
    
    # Write application extension (21 bytes) with application identifier "NETSCAPE".
    out.write(b'\x21\xff\x0b\x4e\x45\x54\x53\x43\x41\x50\x45\x32\x2e\x30\x03' +
            b'\x01\x00\x00\x00')

    # Encode GIF delay time as little-endian unsigned short integer.
    delay = 8*int.from_bytes(rng.read(1), byteorder='big') + 1
    out.write(struct.pack('!H', delay))

    # Repeat the GIF data until we reach 1024 bytes, which is what img2gif enforces.
    while len(data) < 1024:
        tryagain = False

        if len(data) > 0:
            # Repeated blocks have to be at least two characters long, so prepend an empty one.
            if len(data) <= 1:
                data = b'\x00' + data

            # Repeat it by repeating its last character until we reach another byte or hit 1024 in total length.
            data += data[-1:]
            while len(data) % 2!= 0 and \
                   len(data) < 1024:

                newlen = min(len(data)+1, 1024)

                # If we are about to overflow our defined length, don't repeat this byte any further.
                if data[newlen-1] == data[-1]: break
                
                else: data += data[-1:]
        
        elif len(data) == 0:
            # Append zeroes until we reached 1024 bytes again.
            tryagain = True
            while len(data) < 1024: data += b'\x00'
        
        if tryagain: continue

        # We got no new data because either there was enough room left or nothing could be repeated anymore.
        raise OverflowError("Could not make valid gif")
    
    out.write(data)
