"""
Generate a random GIF from /dev/urandom.
"""

import struct
import sys
from argparse import ArgumentParser
from typing import BinaryIO


def parse_args() -> str:
    parser = ArgumentParser("generate-random-gif")

    # Generate a default name if none is provided
    if len(sys.argv) == 1:
        return "random.gif"

    # Allow passing an argument to specify output filename.
    parser.add_argument("-o", "--output", type=str, help="set destination for generated image (default: %(default)s)")
    ns = parser.parse_args()

    if not ns.output:
        raise RuntimeError('No output specified.')
    
    return ns.output


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, wsize: int = 640, hsize: int = 480, nframes: int = 59):
    # Header frame.
    header = [
        71, 73, 70,  56,  57,  97,  24,   0,  33,
          0,   0,   0,    0,   0,   0,   2,   0,  64,
      0, 0, 8]

    header += bytes([wsize % 256]) + bytes([int(hsize / 256)])
    header[4:6] = list(struct.pack("<H", sum(header[4:6])))

    out.write(bytes(header))

    color_table = [0xFFFFFF, 0xFF00FF, 0x00FFFF, 0x0000FF, 0xFFFF00,
                   0xFF0000, 0x00FF00, 0x000000] * 16

    out.write(bytes(color_table))

    lct_flag = True if max(len(color_table), 2) <= 2**8 else False
    gct_flag = True if len(color_table) > 2**8 else False

    flags = b"\x8C\xFB"
    flags += (b"\x0F" if lct_flag else b"\xE0")
    flags += (b"\x00" if gct_flag else b"\xE0")

    out.write(flags)

    position_byte = bytearray([
        ((0 << 7) | ((wsize >> 7) & 0x7f)), 
        (wsize & 0xff), 
        ((0 << 7) | ((hsize >> 7) & 0x7f)), 
        (hsize & 0xff) 
    ])

    out.write(position_byte)

    packed_byte = min((min(wsize, hsize) // 2, 32)) - 1
    lzw_min = 2 ** packed_byte

    sublzw_code_length = []
    for _ in range(nframes):

        sublzw_code_length.append(bytes([(packed_byte + 1)]))

        clear_code = 1 << packed_byte
        end_of_info_code = clear_code + 1
        sublzw_code_length[-1] += bytes([clear_code, end_of_info_code])

        for i in range(lzw_min):
            sublzw_code_length[-1] += bytes([i, 0])

        lc = bytes([end_of_info_code])
        bc = bytes([0])

        while len(lc) < 128 and len(bc) < 128:
            lc += bytes([len(sublzw_code_length[-1])])

            for c in sublzw_code_length[-1]:
                bc += bytes([c])

                left_shifted = (((len(sublzw_code_length[-1])) + 1) & 0xff) << packed_byte

                count = 0
                tmp = ord(bc[count])
                while tmp >= left_shifted:
                    count += 1

                    tc = (tmp >> (8 - (left_shifted & 0xff))) & 0xff
                    bc += bytes([tc ^ sublzw_code_length[-1][count]])

                    tmp ^= tc
                print(ord(bc[count]), len(bc), count)
            
            remainder = (len(bc) - 255) if len(bc) > 255 else 0
            if remainder!= 0:
                sublzw_code_length[-1] += bc[:remainder + 256]
                bc = bc[remainder:]
        
        print(len(sublzw_code_length[-1]), len(lc), len(bc))


if