#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' + bytes([0]) * 7

    width, height = (1 << int.from_bytes(rng.read(1), 'big')) % 500\
           , (1 << int.from_bytes(rng.read(1), 'big')) % 500
    
    # write header
    out.write(header + width.to_bytes(2, 'little')
              + height.to_bytes(2, 'little'))
    
    # write palette
    for i in range(0, 768):
        out.write((rng.read(1)[0] & ~i).to_bytes(1, 'little'))
        
    # LZW minimum code size
    out.write(b'\x01')
    
    # image descriptor block
    out.write(b'\x2c' + width.to_bytes(2, 'little')
              + height.to_bytes(2, 'little')
              + b'\x00\x00\x00\x00')
    
    # LZW compression, 8 bits per code chunk
    lzw_minimum_code_size = 9
    # input data buffer
    buf = []
    # output codes buffer
    outcodes = [lzw_minimum_code_size+1]
    # previous code
    prev_code = None
    while True:
        # new byte to read?
        if len(buf) == 0:
            buf.append(rng.read(1))
            
            if not buf[0]:
                break
            
        # get current byte
        cur_byte = buf[-1][0]
        
        # is this byte less than 128?
        if not cur_byte >= 128:
            buf.pop()
            outcodes.append(cur_byte)

        else:
            # last character was an encoded value; split previous code
            if prev_code and prev_code < lzw_minimum_code_size:
                prefix = 2 ** prev_code
                suffix = cur_byte - prefix

                prefix_bitarray = '{:0{}b}'.format(prefix-1,
                        lzw_minimum_code_size)
                
                # write it as two codes
                for bit in reversed(prefix_bitarray):
                    bitint = 1 if bit == '1' else 0

                    outcodes.append(prev_code*2
                                  + bitint)
                    
                outcodes.append(suffix)

            elif prev_code!= None and \
                 prev_code > lzw_minimum_code_size:
                # just append the current byte
                outcodes.append(cur_byte)
                
            else: # first byte of sequence is special case
                outcodes.append(cur_byte)
            
        # check if we're ready for a code value
        if len(outcodes) >= 2**lzw_minimum_code_size:
            code = ''.join(['{:08b}'.format(entry)
                            for entry in outcodes[:2]])
            
            out.write(struct.pack('>H', int(code, base=2)))
        
            del outcodes[:2]
            prev_code = None
                
        else:
            prev_code = outcodes[-1]
