#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:        BinaryIO=None,
        width:      int    = 1024,
        height:     int    = 768,
        colordepth: int    = 8,
        bgcolor:    int    = None,
        outfile:    str    = 'output.gif',
    ) -> None:

    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    blackcode = bytes([bgcolor & (255**2)])
    whitecode = bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])
    
    with open(outfile,'wb') as f:

        # The GIF header
        version = b"89a"
        screendescriptor = struct.pack('>HHBBB',
                width,
                height,
                0xF0 | colordepth,
                0,
                0,
        )
        
        # Write the header and descriptor block
        print("Writing the header")
        f.write(b"GIF"+version+screendescriptor)

        # Build the color lookup table
        print("Building the color map")
        colortable = []
        for i in range((2**colordepth)-1):
            x = (i*(bgcolor**(1/(256.0))))%(256**3)
            colortable += [struct.pack('>BBB',
                    int(((x>>16)&255)**1.5),
                    int(((x>>8)&255)**1.5),
                    int((x&255)**1.5),
                ),
            ]
        colortable += [blackcode]
        colortable += [whitecode]

        # We can now write the palette.
        print("Writing the palette")
        f.write(b"".join(colortable))

        # Since this is random noise anyway, no packed data required
        datablock = bytearray()

        # Frame durations of (1/fps) seconds in milliseconds
        framedelays = {
            1: 1000//1,
            2: 1000//2,
            4: 1000//4,
            8: 1000//8,
            16: 1000//16,
            32: 1000//32,
            64: 1000//64,
            128: 1000//128,
            256: 1000//256,
        }

        # Pages are limited to one bit per pixel so we have to use the loop
        # construct. Instead of packing everything into an indexed format like
        # other formats do, each page consists of one bit per pixel plus two
        # extra bits of padding on either side. They're only needed because the
        # encoder needs some way of stopping the compression prematurely when it
        # reaches the end of data.
        def encode_page(data):

            bytecount = len(data)*8
            datalenlo = bytecount % 256
            datalenhi = (bytecount >> 8) & 255
            return "\x21\xF9\x04\x01\x00" + \
                   chr(datalenlo)+chr(datalenhi)+"!" + \
                  ''.join(map('{0:08b}'.format, data))+ \
                   ";;"

        while True:
            
            # Each frame starts with a graphic control block. This tells the
            # decoder how long the frame should be displayed for before the
            # next frame runs. It also allows the decoder to skip over this
            # frame if it's not wanted.
            displaytime = min(framedelays)
            disposalmethod = ord(rng.read(1))%2
            userinputflag = 0
            transparentindex = 0xff
            f.write(bytes([ord('\x21'), ord('\xF9'), 4])) 
            f.write(struct.pack('<BBHB',
                    0x01,              # Block size
                    disposalmethod,    # Disposal method
                    0,                 # User input flag
                    displaytime,       # Display duration
                ))
            f.write(bytes([transparentindex]+[userinputflag]*3))

            # Next comes the image descriptor block. There's a lot of stuff here
            # but for us the most important part is the LZW minimum code size.
            # According to the spec, it must be between 2 and 12 inclusive. If
            # the requested value is outside that range, then we pick a random
            # value within that range instead.
            lzwminsize = max(2,min(int(rng.read(1)),12))
            print("LZW minimum code size:",lzwminsize,"bits")
            f.write(bytes([ord('\x2C'),
                            0,          # Left position
                            width-1,    # Right position
                            height-1,]))# Bottom position
            f.write(bytes([
                        lzwminsize,# Minimum code size
                        0,         # Maximum code size
                    ]))

            # Finally we write out the compressed data.
            compresseddata = ""
            while True:

                # First we need to determine whether another bit is going to be
                # encoded in this run; this happens when the current run would
                # exceed the maximum number of bits allowed by the minimum code
                # size for this frame. Note that this means we can never reach
                # more than 2^lzwminsize for a single frame.
                segsize = 2**lzwminsize - 1
                if len(compresseddata)>segsize:

                    # When encoding the data, we use space to store both the new
                    # data segment and the old data segment concatenated together.
                    # In order to extract just the last chunk from the string, we
                    # rotate the string back by a certain amount, where the number
                    # of characters rotated by is equal to the length of the
                    # current data segment. Afterwards, we remove all the spaces
                    # that were added to make the transformation work.
                    seglenstr = '{0:0'+str(lzwminsize)+'b}'
                    segs = list(filter(lambda x: x!=' ',compresseddata[-segsize:]))
                    segs[0] = segs[0][-(len(segs[1])+1):]
                    
                    # Finally, we append any remaining data after rotation to the
                    # empty string which will become the next segment. Anything
                    # left over after the first 128 segments is stored until the
                    # next time around the loop.
                    leftover = ''
                    if len(segs) == 129 and len(compresseddata) >= 2**12:

                        # Here we reassemble the leftovers from earlier.
                        leftover = compresseddata[-2**12:]
                        
                        # To put things back as they were originally, we need to
                        # undo the rotations we did previously. For every character
                        # rotated, we apply its inverse rotation to get what was
                        # originally at the beginning of that character.
                        leftover = [''.join(list(' '*rot))[::-1]
                                    for rot in range(len(leftover))]

                        # With the original data unrolled, we can concatenate them
                        # back together.
                        segs[-1] = "".join(leftover)
                        segs.pop(-2)

                        # And finally we can use the entire thing again.
                        leftover = "".join(segs)
                        segs = ['']
                    
                    # At this point, we've got a list of data chunks of length
                    # less than 2^lzwminsize. By inserting these back into the
                    # data stream, we can keep track of their lengths without
                    # having to manually compute them later.
                    compresseddata = compresseddata[:-segsize].strip()+seglenstr.format(len(segs[-1]))+' '.join(segs)+' '+leftover
                
                else:

                    # Otherwise, we read the next bit from the RNG.
                    seg = int(rng.read(1))
                    segstr = '{0:08b}'.format(seg)[::-1]

                    # If there isn't enough room left in the current run, we need
                    # to start a new one.
                    if len(compresseddata)+(len(segstr)-1) > 2**lzwminsize-1:
                        break

                    # Otherwise, we add the new bit onto the end of the current
                    # run and continue.
                    compresseddata += segstr

            # As mentioned above, the encoder needs to know how much data it has
            # still to come and wants the data split into pages if necessary. So
            # let's do that now since we've already finished constructing the
            # compressed data.
            pages = [encode_page(chunk) for chunk in compresseddata.split()]
            for page in pages[:129]:
                f.write(bytearray(page,encoding='utf-8'))

        # Now we need to write the image data itself...
        print("Writing the image")
        f.write(b"\x2C\x00\x00\x00\x00\x00\x00\x00\xFF")   # Image descriptor
        f.write(b"!\xF9\x04\x01\x00\x00\x00,\x00\x00\x00") # Graphics control extension
        f.write(b"\x00\x01\x00\x21\xFE\x09D\x0B\xCF\x00") # Application Extension
        f.write(b"\x0BNETSCAPE2.0\x03\x01")                  # Application Data
        f.write(b"\x00\x00\x00!")                           # Trailer
        print("\nDone.")