#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    header += struct.pack('>H', rng.randint(200, 500))
    header += struct.pack('>H', rng.randint(200, 500))

    # color table (RGB triplets)
    ctsize = rng.randint(1, 7) * 3
    for _ in range(ctsize):
        header += bytes([rng.randint(0, 255)])

    # background is black
    header += b'\x00\x00\x00'

    # pixel aspect ratio
    header += b'\x00'

    # image descriptor
    idsize = rng.choice((0, 4, 10)) + rng.randint(0, 3)
    if idsize > 0:
        header += b'\x2c'
        for _ in range(idsize - 1):
            header += bytes([rng.randint(0, 255)])

    # image data subblock
    iblocks = []
    while True:

        blocktype = rng.choice((0b00, 0b01, 0b10, 0b11))
        blocklen = rng.randint(1, 256)
        blockdata = bytearray()
        for _ in range(blocklen):
            blockdata.append(rng.randint(0, 255))
        
        block = bytearray()
        block.extend(struct.pack('<BB', blocktype, blocklen))
        block.extend(blockdata)

        if len(iblocks) == 0 or rng.getrandbits(1):

            # use local colour table
            if len(block) < blocklen + 1:
                continue
            
            if block[0] & 0x01!= 0:

                cti = rng.randint(0, ctsize // 3 - 1)
                lcti = ((block[0] >> 2) % 7) * 3
                
                if ctsize <= 64:
                    clen = rng.choice((0, 1, 2, 3))
                elif 64 < ctsize <= 128:
                    clen = rng.choice((0, 1, 2, 3, 4, 5, 6, 7))
                else:
                    clen = rng.choice((0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                                        10, 11, 12, 13, 14, 15, 16))
                    
                block[clen+1] |= cti << 3 | lcti
                
            else:

                # no LZW compression
                block[1] &= ~0x80
                
        iblocks.append(bytes(block))

        if rng.getstate()[1][0] >= 1000: break
    
    # terminator
    iblocks[-1] = bytes([0])
    
    out.write(header)
    for block in iblocks:
        out.write(block)