#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:        BinaryIO=None,
        width:      int    = 1024,
        height:     int    = 768,
        colordepth: int    = 8,
        bgcolor:    int    = None,
        outfile:    str    = 'output.gif',
    ) -> None:

    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    blackcode = bytes([bgcolor & (255**2)])
    whitecode = bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])
    
    with open(outfile,'wb') as f:

        # The GIF header
        version = b"89a"
        screendescriptor = struct.pack('>HHBBB',
                width,
                height,
                0xF0 | colordepth,
                0,
                0,
        )
        
        # Write the header and descriptor block
        print("Writing the header")
        f.write(b"GIF"+version+screendescriptor)

        # Build the color lookup table
        print("Building the color map")
        colortable = []
        for i in range((2**colordepth)-1):
            x = (i*(bgcolor**(1/(256.0))))%(256**3)
            colortable += [struct.pack('>BBB',
                    int(((x>>16)&255)**1.5),
                    int(((x>>8)&255)**1.5),
                    int((x&255)**1.5),
                ),]
        colortable[0] = blackcode
        colortable[-1] = whitecode

        # Dump it to the output file
        print("Writing the color map")
        f.write(bytes(colortable))
        
        # Generate the image data stream. First we write out
        # the logical screen descriptor again, then the actual
        # image data blocks.
        print("Generating image data")
        f.write(b"\x2c"+screendescriptor)
        
        # Calculate where the image should start so the GIF viewer doesn't lag
        imgstart = f.tell() + len(blackcode)

        # We're going to use delta encoding from now on - this allows us 
        # to compress the data relatively well while still making it look
        # pretty. The first frame is an entirely opaque background anyway.
        prevframeline = bytearray([255]*width)
        frames = []

        # Each frame consists of a set of subblocks starting with a 
        # control code. This loops generates all the different types
        # of frame information to put inside each frame block.
        nframes = min(int(rng.read(1)[0]/10)+1,20)
        print(f"Writing {nframes} frames!")

        for _ in range(nframes):

            # Frame discriptor
            delay = max(rng.read(1)[0],1)*50
            
            lsd = struct.pack('<HBB',delay,0,0xF9)
            f.write(lsd)

            # LZW compression!
            # Reference: http://www.matthewflickinger.com/lab/whatsinagif/lzw_image_data.asp

            # Set up the initial dictionary for the compression algorithm
            dictsize = 2**colordepth-2
            dictionary = {}
            for i in range(dictsize):
                dictionary[chr(i)] = chr(i)

            # Loop through each line in the frame and compress them individually
            for y in range(height):

                currentframeline = bytearray(prevframeline)

                # Grab two pixels at once here because otherwise they're
                # too tightly coupled. At least, I think that's why?
                pixelpair = rng.read(2)
                for x in range(0,len(pixelpair),2):
                    
                    # Get the color code from these two pixels
                    px = tuple(reversed(currentframeline[x:(x+2)]))
                    colorcode = (px[0]<<16)+(px[1]<<8)+rng.read(1)[0]

                    # If their difference isn't already in the dictionary, add it
                    if colorcode not in dictionary:
                        dictionary[(px[0]<<16)+(px[1]<<8)+rng.read(1)[0]] = \
                                chr(dictsize)

                        if dictsize == 255:
                            raise Exception('Ran out of colors!')
                        
                        dictsize+=1

                    # Add the new character onto the end of the buffer. It'll
                    # probably be one char long.
                    currentframeline += bytes([dictionary[(px[0]<<16)+(px[1]<<8)+rng.read(1)[0]][0]])
                
                # Now we've written the entire line down, we can update what 
                # the previous line was
                prevframeline = currentframeline

                # Before writing this line, we need to have encoded it to the 
                # smallest possible size. To do this, we have to sort the list
                # so that it decreases in length until the entire thing fits.
                compressedline = ''
                while True:
                    best = ''
                    for string in dictionary.values():
                        if string=='' or best=='': 
                            continue
                        elif ''.join(sorted(string))<best: 
                            best = ''.join(sorted(string))
                        
                    if best!= '':
                        compressedline+=best
                        del dictionary[best]
                    else: break
                compressedline = bytes([len(compressedline)]) + compressedline.encode()

                # Write the result to the output file
                f.write(compressedline)

            # Close off the frame by putting an "end of data" marker 
            # after it. Just like in the animation control extension...
            f.write(struct.pack('<B', 0))

        # Finally, close off the file using the appropriate trailer
        # information. Note: we ignore most of the properties and
        # just include the bare minimum needed to make it animate properly.
        print("Closing the file.")
        f.seek(imgstart)
        f.write(b'\xff\x21\xf9\x04'+struct.pack('<BBBB',width//8,0,0,0))
        f.write(b'\x00,\x00\x00\x00')

