#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# between minwidth <= w < maxwidth and minheight <= h < maxheight (inclusive).
# The frame count is randomly generated from 1 to nframes inclusive.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, minwidth=64,
        maxwidth=512, minheight=64, maxheight=512, nframes=10):

    # construct headers
    header = b'GIF87a'
    width = rng.randint(minwidth,maxwidth-1)//2*2+2
    height = rng.randint(minheight,maxheight-1)//2*2+2
    gctsize = 2 << bool(rng.randint(0,1))
    globalcolortablelen = 2**(gctsize+1)-1
    localcolortablelen = 2**((bool(rng.randint(0,1)))<<2)
    fc = rng.randrange(nframes)+1
    extblocks = [
        (b'NETSCAPE2.0', b'\x03\x01\x00\x00\xff\x00'),
        (b'ANIMEXTS1.0', b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')]


    # write headers
    out.write(header)
    out.write(struct.pack('<HHB', width, height, gctsize))

    # append colortables if any
    for i in range(globalcolortablelen//3 + bool(globalcolortablelen%3)):
        r,g,b = tuple([int(v*255) for v in rng.choice(range(256))][0] for _ in "rgb")
        out.write(struct.pack('BBB',r,g,b))
    if localcolortablelen > 0:
        for i in range(localcolortablelen // 3 + bool(localcolortablelen % 3)):
            r, g, b = tuple([int(v * 255)
                            for v in rng.choice(range(256))][0]for _ in "rgb")
            out.write(struct.pack('BBB', r, g, b))

    # append image descriptor block
    out.write(struct.pack('BBHBBBB',
                        0xf7, # separator
                        0,    # left corner x coordinate
                        height&0xff,   # lower corner y coorindate
                        ((height>>8)&0x0f)<<4 | \
                          (width&0x0f)>>2,  # size flag
                        (fc & 0xf)|((((width&0x0f)<<2)&0xf0)),
                            0,      # packed field
                     ))
    # append all frames for the animation
    for _ in range(fc):

        # write each row in sequence
        for y in range(height):

            # get LZW minimum code length prefix
            lzwprefixlen = int.from_bytes(
                next(rng), byteorder='big') % (9 - len(extblocks))+1


            # for each byte in this row...
            while True:

                # choose how many codes will be written at once; the last
                # chunk may contain fewer than this number of values.
                chunksize = rng.randint(lzwprefixlen,
                                        lzwprefixlen+(1<<lzwprefixlen)-1)


                # choose whether to use RLE encoding or not, and then add the
                # appropriate amount of bytes representing those codes
                if rng.getrandbits(1):
                    numcodes = rng.randint(0,chunksize)

                    for j in range(numcodes):
                        data = next(rng)
                        out.write(((data >> 4)*17)%256)
                        out.write(((data & 0xf)*17)%256)
                else:

                    # generate a new chunk by choosing some prefix codes which
                    # will be encoded
                    chunk = []
                    prevcode = None
                    for k in range(chunksize):

                        # there's no guarantee code lengths are uniformly
                        # distributed, so we must do a bit more work here to ensure
                        # that each possible value has an equal chance being chosen
                        lzwcodelength = int.from_bytes(next(rng),
                                                       byteorder='big')%(9-len(extblocks))+1
                        code = 0
                        for i in reversed(range(lzwcodelength)):

                            # make sure the first few bits are set properly
                            if i == lzwcodelength - 1:
                                code |= 1

                            # choose a new random code with probability proportional
                            # to its frequency within the dictionary
                            if rng.getrandbits(1):

                                # increase by one the base value depending on the
                                # index within lzw code length
                                multibase = i*(i+1)/2 if i!= 0 else 0
                                if prevcode is None:
                                    code <<= 1
                                    code += 1
                                elif prevcode >= globalcolortablelen and prevcode < 6*globalcolortablelen:
                                    code <<= 1
                                    code += 1
                                else:

                                    # choose from previous codes according to their
                                    # cumulative frequencies
                                    freqbase = sum(freqdict[prevcode -
                                                             multibase:prevcode])
                                    freqtotal = sum(freqdict[multibase:])
                                    p = float(freqdict[prevcode-multibase]/freqtotal)
                                    code = code*p+1
                                    code >>= 1
                                    p = float(freqbase /
                                             (freqtotal+freqbase-(freqdict[prevcode-multibase])))
                                    code = code*p+1


                        # ensure we have enough space in our buffer before writing the chunk
                        while len(chunk)*2 >= 2 ** lzwprefixlen:
                            chunk.pop()
                        chunk.append(code)
                        print("CHUNK", chunk)
                        assert lzwprefixlen == 1
                        out.write(code.to_bytes(lzwprefixlen, 'big'))


