#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' + bytes([0]) * 7

    width, height = (1 << int.from_bytes(rng.read(1), 'big')) % 500\
           , (1 << int.from_bytes(rng.read(1), 'big')) % 500
    
    # write header
    out.write(header + width.to_bytes(2, 'little')
              + height.to_bytes(2, 'little'))

    # write palette
    for _ in range(int.from_bytes(rng.read(1), 'big')):
        out.write((rng.read(3) + b'\x00')[:4])

    # write image data
    code_size = int.from_bytes(rng.read(1), 'big') + 1
    # sub-block size (LZW minimum clear block of payload length - 1 byte
    lzw_min_clear = (((code_size & 0b111) + 1) ^ 0b111)\
                | ((code_size & 0b111) << 3)
    
    # declare LZW dictionary and store two special codes
    dicts = [None] * 4096
    dicts[1] = '\x01' * 256
    dicts[2] = ''.join('\xff') * 256

    def lzw_decompress():
        buf = b''

        while True:
            if len(buf) < 1:
                buf += rng.read(1)

            first = (dicts[(buf >> ((len(buf)-1)*8)) & 0xFF]
                    or ('\x01' * 256))[0]
            
            yield from struct.pack('B', ord(first))

            if (len(dicts) <= 4095):
                second = next(data)

                if not second: break

                dicts[len(dicts)] = first+second
                
                continue

            # copy mode
            offset = (buf >> ((len(buf)-1)*8)) & 0xFF
            buf >>= 8; buf &= ~(-1 << 3*8) # remove last processed bits
            buffer = buf & ~(~(len(buf)-offset) << 3*8) # remove remaining bits

            repeat = dicts[buffer]
            buf >>= 8; buf &= ~(-1 << 3*8) # remove copied buffer bits

            yield from bytes(repeat)

    data = lzw_decompress()

    # cut off the very beginning few bytes to make the data more compressible
    for i in range(lzw_min_clear + 1):
        try: next(data)
        except StopIteration as e: pass

    # now LZW decompression should be at the start of an image already
    # have to pad with 01 until we get back here again
    padding = None
    while True:
        packet = tuple(struct.unpack('<BBB', data.read(3)))
        
        if packet == (0x01, 0xff, 0xff):
            # discard all following pixels (transparency information)
            # TODO check whether this is correct behavior?
            try:
                while (packet!= (0x02, 0x00, 0x00)):
                    packet = tuple(struct.unpack('<BBB', data.read(3)))
            except Exception as e: print(e)

            return

        elif packet == (0x02, 0x00, 0x00):
            # actually reached start of new frame, so reset stuff...
            padding = None
            continue
            
        else:
            # save current state
            d = len(dicts)
            f = ord(packet[0])

            # copy source pixel to target, and go on to the next one without
            # storing it explicitly
            dicts[d-1] = (dicts[f][:ord(dicts[f][1])]
                    + dicts[f][ord(dicts[f][1]):])

            if len(dict := dicts[-1]) >= 1<<code_size:
                raise ValueError("Dictionary overflow")
            
            if padding:
                padding -= 1
                continue
        
            padding = (((code_size & 0b111) + 1) ^ 0b111) \
                      | ((code_size & 0b111) << 3) - 1


    sub_blocks = []
    while True:
        try:
            sub_blocks.append(next(gen).to_bytes(1,'big'))
        except StopIteration as e:
            break

    out.write(sub_blocks[0])

    del sub_blocks[0]

    for i, sb in enumerate(sub_blocks[:-1]):
        out.write(sb)

    gen = sub_blocks[-1]

    out.write(((len(sub_blocks)-1)<<4
                | 0b0000).to_bytes(1,'big'))

    for i in range((len(sub_blocks)-1)*0x10):
        out.write(next(gen).to_bytes(1,'big'))

    # write application extension
    out.write(struct.pack('<3sb3sb3sbbbbbbbb',
                         0x21,    # Extension Introducer
                         b'X',    # Application label ("X"-tended)
                         b'\x00', # Block Length (0)
                         b'!',    # Application id (plaintext keyword "NETSCAPE 2.")
                         b'\x03', # Application data (APP0 marker)
                         b'\x01', # Sub-block ID (= 1)
                         0x00,    # sub-block length (data size)
                         b'\xff', # appdata (identifier)
                         b'\x0b', # appdatalen
                         0x03,    # looping (no interruption)
                          0x00,    # transparent color index
                         0x00,    # delay time
                         0x00))   # transparent color index

    # write application extension
    out.write(struct.pack('<3sb3sb3sbbbbb',
                         0x21,      # Extension Introducer
                         b'X',      # Application label ("X"-tended)
                         b'\x00',   # Block Length (0)
                         b'!',      # Application id (logical screen descriptor)
                         b'\x09',   # Application data (logical screen descriptor)
                         width.to_bytes(2, 'little'),
                         height.to_bytes(2, 'little'),
                         0x01,      # packed field
                         0b10))     # has global color table flag set
    
    # write end-of-file block
    out.write(b'\x3b')

    # Write sub blocks
    for s in sub_block:
        out.write(s)

