#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# between minwidth <= w < maxwidth and minheight <= h < maxheight (inclusive).
# The frame count is randomly generated from 1 to nframes inclusive.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, minwidth=64,
        maxwidth=512, minheight=64, maxheight=512, nframes=10):

    # construct headers
    header = b'GIF87a'
    width = rng.randint(minwidth,maxwidth-1)//2*2+2
    height = rng.randint(minheight,maxheight-1)//2*2+2
    color_map_size = 0x00
    bg_color_index = 0x00
    pixel_aspect_ratio = 0x00
    gct = bytes([0xFF]) + bytes([bg_color_index])*3+\
          bytes([pixel_aspect_ratio])+bytes([gcten[rng.randrange(len(gcten))]]*3)

    # assemble screen descriptor block
    screen_descriptor = struct.pack('<HHBBB',
                                   width, height,
                                   color_map_size, gcten[rng.randrange(len(gcten))],
                                   0b11110000 | rng.randrange(nframes))

    # assemble global color table block
    global_color_table = bytes()

    for i in range(256):
        r = c = b = 0

        while r == c == b:
            r = rng.randint(0,255)
            c = rng.randint(0,255)
            b = rng.randint(0,255)

        global_color_table += bytes((r,c,b,))

    # write out the GIF file
    out.write(header);
    out.write(screen_descriptor)
    if len(global_color_table) > 1:
        out.write(global_color_table)
    else:
        out.write(b'\x00\x00\x00')

    # generate frames
    for f in range(nframes):
        lzw_minimum_code_size = rng.randint(2,9)
        for i in range(lzw_minimum_code_size):
            out.write(bytes([rng.randrange(2**lzw_minimum_code_size)]))

        runlengths = [rng.randrange(2**lzw_minimum_code_size)]*(2**lzw_minimum_code_size)+\
                     [rng.randrange(1,2**lzw_minimum_code_size) \
                      for _ in range(int(2**lzw_minimum_code_size))]
        
        out.write(bytes([(lzw_minimum_code_size << 4)+rng.randrange(2**lzw_minimum_code_size),
                         *(runlength - 1)\
                             for runlength in runlengths]))


        subblock_data = None
        lzw_minimum_code_size = 0;
        while True:

            if subblock_data is not None:
                if rng.getrandbits(1):
                    break

                subblocks = [(subframe << 1) & ~0xff |
                             ((rng.getrandbits(7) +
                              (subframe >> 8 * int((rng.getrandbits(1))))
                              )&0xff)
                             for subframe in subblock_data]
                out.write(bytes(subblocks))


                subblock_data = None


            newsubframe_data = []
            cursubframe_data = []
            try:
                while rng.getrandbits(1):

                    firstbyte = rng.getrandbits(8)
                    if firstbyte!= 0x00 and firstbyte!= 0xff:
                        cursubframe_data += bytes([firstbyte])

                    cursubframe_data += bytes([rng.getrandbits(8)])



                subblock_data = b''
                while True:
                    
                    numbytes = rng.randint(1,5)
                    subblock_data += bytes([rng.getrandbits(8)
                                             for _ in range(numbytes)])


                    if rng.getrandbits(1):
                        break
            except IndexError: pass
            
            out.write(cursubframe_data)



