#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:        BinaryIO=None,
        width:      int    = 1024,
        height:     int    = 768,
        colordepth: int    = 8,
        bgcolor:    int    = None,
        outfile:    str    = 'output.gif',
    ) -> None:

    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    blackcode = bytes([bgcolor & (255**2)])
    whitecode = bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])
    
    with open(outfile,'wb') as f:

        # The GIF header
        version = b"89a"
        screendescriptor = struct.pack('>HHBBB',
                width,
                height,
                0xF0 | colordepth,
                0,
                0,
        )
        
        # Write the header and descriptor block
        print("Writing the header")
        f.write(b"GIF"+version+screendescriptor)

        # Build the color lookup table
        print("Building the color map")
        colortable = []
        for i in range((2**colordepth)-1):
            x = (i*(bgcolor**(1/(256.0))))%(256**3)
            colortable += [struct.pack('>BBB',
                    int(((x>>16)&255)**1.5),
                    int(((x>>8)&255)**1.5),
                    int((x&255)**1.5),
                ),]

        # Write out the color table
        print("Writing out the color table")
        for c in colortable:
            f.write(c)


        # Set up the image descriptor
        lzwminsize = 2

        globalcolormap = True
        interlaceflag = False
        sortflag = False
        localcolortableflag = False
        reservedflags = 0
        

        # Write out the data subblocks of each frame in turn
        print("Writing the frames:")
        for i in range(int(width*height/(256**3))+1):

            # If this is the first frame we need to write its header
            if i == 0:

                # Color table size - one byte, set above
                colortablesize = len(colortable)

                # Global color table flag - one bit, 1 indicates it exists
                globalcolormaphookup = globalcolormap<<7
                f.write(bytes([globalcolormaphookup]))

                # Interlacing flag - one bit, 1 means no interlacing (no sorting needed)
                # bits 0-2 are reserved, must all be zero
                interlaceflaghookup = (interlaceflag<<6)|reservedflags

                # Sorting flag - one bit, 1 means pixel code size larger than min (no sorting needed)
                # bits 0-3 are reserved, must all be zero
                sortflaghookup = (sortflag<<5)|reservedflags

                # Size of LZW min code length - three bits, value + 1
                lzwminsizehookup = (lzwminsize-1)<<2|reservedflags

                # Local color table flag - one bit, say yes (we have one!)
                localcolortableflaghookup = localcolortableflag<<7

                # Bits 0-6 of the flags, then the size of the color table 
                frameheader = bytes([
                        0xE0 |
                        interlaceflaghookup | 
                        sortflaghookup |
                        lzwminsizehookup
                ]) + bytes([colortablesize]) 

                # Write the frame header
                f.write(frameheader)

                # Write out the color table again (this time it's actually the palette)
                for c in colortable:
                    f.write(c)
                
                # This is the first frame so set up the static background
                fill = bytes([blackcode[i%len(blackcode)]])

            else:
                fill = bytes([whitecode[i%len(whitecode)]])
            
            # Go through the pixels and encode them via 
            # variable bit encoding (LZW compression)
            print("\tFrame %d"%i)
            imagename="image_%d.png"%i
            img = Image.new('RGB', (width,height), bg_color=(fill,))
            draw = ImageDraw.Draw(img)
            try:
                os.remove(imagename)
            except OSError:
                pass
            img.save(imagename)
            gifencode(img,f,colordict={tuple():blackcode})
            assert(os.path.exists(imagename))
            os.remove(imagename)

