#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# between minwidth <= w < maxwidth and minheight <= h < maxheight (inclusive).
# The frame count is randomly generated from 1 to nframes inclusive.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, minwidth=64,
        maxwidth=512, minheight=64, maxheight=512, nframes=10):

    # construct headers
    header = b'GIF87a'
    width = rng.randint(minwidth,maxwidth-1)//2*2+2
    height = rng.randint(minheight,maxheight-1)//2*2+2
    globalcolortablesize = 7
    localcolortablesize = 0
    loopcount = 0
    bgindex = 0
    aspectrationnumerator = 1
    aspectrationdenominator = 1
    screendescriptor = bytes([
        0x48|globalcolortablesize,     # size of color table
            0 |                        # sort flag
            int((width>>8)&0xff),      # width high byte
            int(width&0xff),            # width low byte
            0 |                        # size of pixel data high bit
            int((height>>8)&0xff),     # height high byte
            int(height&0xff)           # height low byte
    ]) + struct.pack('<BBBB', 0xf9, 0,    # background index & color resolution
                     loopcount,         # number of loops
                     bgindex) + \
                 struct.pack('>HH',   # pixel aspect ratio
                    aspectrationnumerator,aspectrationdenominator) + \
                 3*(b'\0')               # padding for the rest of the block

    # write headers to file
    out.write(header)
    out.write(screendescriptor)
    
    # output global server palette; it's useless but let's at least get some
    # variation
    for i in range(0xa9, 0x1f9, 0xa9):
        rgb = [i&(j^7) for j in range(0,256)]
        out.write(bytes(rgb))

    # output frames
    for _ in range(rng.randrange(nframes)+1):

        # RGB pixels come first, so we can use the minimal LZW code length of
        # 3 bits per color value
        lzwcodelength = 3
        
        gce = False
    
        # create each frame with a different random color palette
        colors = []
        for i in range(rng.randrange(16)):

            rr = rng.randrange(256)
            gg = rng.randrange(256)
            bb = rng.randrange(256)
            colors += [rr,gg,bb]

            out.write(struct.pack("BBB",
                rr,gg,bb))

        # construct a graphic control extension as well
        if not gce:
            
            transparent = True
            delaytime = 0
            transparentindex = None
            for x in range(len(colors)):

                if rng.getrandbits(1):

                    if rng.getrandbits(1):

                        transparentindex = len(colors)-1
                        break
            
            out.write(struct.pack('>BBHB',
                0x21,          # type: Graphic Control Extension
                0xf9,          # size
                0<<7|                 # no disposal method 
                1<<6|                 # user input is required 
                int(transparent)<<5|   # transparency flag
                delaytime,        # delay time
                transparentindex or 0  # transparent color index
            ))

            gce = True
            
        # output image descriptor
        interlaceflag = bool(rng.getrandbits(1))
        sortedflag = bool(rng.getrandbits(1))
        localcolortableflag = False
        out.write(bytes([
            0x2c,              # image separator
            0 |                # left position high bit
            0,                 # left position low byte
            0 |                # top position high bit
            0,                 # top position low byte
                int((width>>8)&0xff),       # width high byte
                int(width&0xff),             # width low byte
                int((height>>8)&0xff),      # height high byte
                int(height&0xff),            # height low byte
                0 |                         # local color table flag
                int(localcolortablesize)   # local color table size
        ]))

        # output local color table
        if localcolortableflag:

            # sort flag needs to be set before writing local color table size
            out.write(sortedflag.to_bytes(1,'big'))

            out.write(struct.pack('B',localcolortablesize))
            colorsizes = [int(round(c**(1./3)))//8 for c in colors[:localcolortablesize]]
            for colorsize in colorsizes:
                for idx in range(colorsize): 
                    out.write(bytes([(idx%4)*0xaa]))
        
        # output packed fields (LZW minimum clear code, LZW maximum clear code,
        # LZW end of information code, LZW clear code)
        out.write(struct.pack('BBBB', 
            0b00000000,                              # packed field
            1 << lzwcodelength,                      # minimum code size
            0b00000000,                              # reserved
            0b00000111                               # clear code
        ))

        # uncompressed subblocks consist of one or more "sub-blocks" which
        # contain a single compression dictionary followed by a series of codes
        # within this dictionary. Each sub-block ends with an "end of information" 
        # code. When an uncompressed sub-block runs out of space, another must be 
        # started, even if the previous one still has room remaining.
        
        # we'll start by just adding all the pixels (rather than trying to optimise
        # them) and then compressing what we have afterward

        # initialize encoder. note that we set its initial dictionary directly here
        # rather than letting it pick a good one automatically because we want to
        # make sure our encoder doesn't ever choose a local colour table based on
        # the frame's local colour table values; these should never be used during
        # decoding.
        encoder = lzw.EncoderDictSize()
        encoder.initial_dict[tuple(colors)] = tuple(range(len(colors)))

        # add the pixels to the encoder
        for y in range(height):
            for x in range(width):
                
                r = rng.randrange(256)
                g = rng.randrange(256)
                b = rng.randrange(256)
                decodervalue = (r,g,b)
                encoderinputval = struct.unpack('BBB',bytes([r,g,b]))
                
                encoder._add_data(decodervalue, encoderinputval)

        # finish encoding the last byte        
        encoder._flush(encoder.eof())

        # extract encoded data into separate chunks
        encchunks = encoder.encoded().split(encoder.eof()[0])[:-1]

        # now we need to convert the compressed stream from our encoder back to
        # a raw stream containing only the values of the entries stored in the
        # last chunk of the encoder's list. 
        # note that this requires inserting a couple of filling zeroes to account
        # for the fact that some integer types may require multiple bytes to encode
        filledencchunks = [[encoder.eof()] + list(chunk) + [-1,-1,-1] 
                                for chunk in encchunks]
        strippedencchunks = [(chunk[-1],) for chunk in filledencchunks]

        # write the final sub-block
        for sbc in strippedencchunks: 
            out.write(b','.join([struct.pack('>H',s) for s in sbc]))
        out.write(-1)

