#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:        BinaryIO=None,
        width:      int    = 1024,
        height:     int    = 768,
        colordepth: int    = 8,
        bgcolor:    int    = None,
        outfile:    str    = 'output.gif',
    ) -> None:

    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    blackcode = bytes([bgcolor & (255**2)])
    whitecode = bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])
    
    with open(outfile,'wb') as f:

        # The GIF header
        version = b"89a"
        screendescriptor = struct.pack('>HHBBB',
                width,
                height,
                0xF0 | colordepth,
                0,
                0,
        )
        
        # Write the header and descriptor block
        print("Writing the header")
        f.write(b"GIF"+version+screendescriptor)

        # Build the color lookup table
        print("Building the color map")
        colortable = []
        for i in range((2**colordepth)-1):
            x = (i*(bgcolor**(1/(256.0))))%(256**3)
            colortable += [struct.pack('>BBB',
                    int(((x>>16)&255)**1.5),
                    int(((x>>8)&255)**1.5),
                    int((x&255)**1.5),
                ),]


        localcolormapsize = len(colortable)+1
        localbits = int(localcolormapsize**.5)
        sizeoflocalpalette = max(min(int(4096//(localcolormapsize)), 256), 1)
        globalcolormapsize = min((((2**colordepth)-1)*sizeoflocalpalette), 256)
        globalbits = int(globalcolormapsize**.5)
        

        # Write out the global palette header
        print("Writing the globals")
        bgindex = 0
        bgflag = False
        if bgcolor == 0xffffff:
            bgindex = 1
        elif bgcolor == 0x000000:
            bgindex = 2
        else:
            bgflag = True
            

        
        f.write(bytes([
                globalcolormapsize,                   # Size of the global colormap
                globalbits,                            # Bits per index in the global colormap
                0x80,                                   # Global palette initialized flag
                0x00 if bgflag else 0x01,              # Background color is either fixed or has transparency
                ] + ([bgindex] if bgflag else []) + [   # Color index to use for background
                    0xFF,                                 # Pixel aspect ratio
                ]))

        # Write out the global palette entries
        print("Writing the globals...")
        for entry in colortable:
            f.write(entry)

        
        
        # Write the lzw dictionary
        print("Writing LZW dictionary")
        # This section is just the code from the wikipedia article on LZW
        dictlen = 256
        dictionary = {chr(i):[chr(i)] for i in range(dictlen)}
        dictionary['\0']=[]
        stackptr = '\0'

        f.write(stackptr.encode())
        for character in "abcdefg":
            charlist = dictionary[character][:]
            for prefix in stackptr:
                dictionary[prefix].append(charlist[-1]+prefix)
            dictionary[character]=charlist

            f.write((character+'\0').encode())
            f.write(charlist[-1].encode())

        # Write the image descriptor struct
        print("Writing frame...")
        f.write(bytes([
                0x2C,         # Image separator
                width % 256,  # Left position of image data
                int(width/256),  # Top position of image data
                (width % 256)/8,  # Width of image area
                (int(width/256))/8,  # Height of image area
                0,             # Flags
                0,             # Background color index
                0,             # LZW minimum code size
                ]))
        try:
            while True:
                cellsize = ord(rng.read(1))

                for _ in range(cellsize-1):
                    f.write(dictionary[stackptr[-1]][ord(rng.read(1))])
                    
                character='\0'
                bestkey=''
                while chr(stackptr[-1])!= character:
                    key = stackptr[:-1]+chr(stackptr[-1])+character
                    if key in dictionary:
                        bestkey+=character
                        character='\0'
                        stackptr = key
                    else:
                        break

                    
                
                f.write(bestkey[(len(stackptr)-1):].encode() if stackptr!='\0' else '\0'.encode())
                f.write(dictionary[stackptr][ord(rng.read(1))] if len(stackptr)>0 else 'a'.encode())


            
        except Exception as e:
            pass

        # Write out the final byte of the stream
        print("Finishing up...")
        f.write('\0'.encode())
