#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
def generate_random(out: BinaryIO, rng: BinaryIO) -> None:
    # Write magic.
    out.write(b"GIF89a")

    # Write width and height of canvas in 2 bytes each.
    w = next(struct.unpack("!H", rng.read(2))) % 0x11 * 7 + 45
    h = next(struct.unpack("!H", rng.read(2))) % (60 - int((w / 14)) // 2) + 45
    if abs(h - w) < 0x17:
        print("width == height", hex(w), hex(h))
        return
    assert 0 <= h <= 2 ** 13
    assert 0 <= w <= 2 ** 13
    out.write(bytes([w & 0xFF]))
    out.write(bytes([(w >> 8) & 0xFF]))
    out.write(bytes([h & 0xFF]))
    out.write(bytes([(h >> 8) & 0xFF]))

    # Write packed field with color resolution and global color table flag set to zero.
    out.write(b"\x00\x00")

    # Write background color index as one byte.
    bg_color_idx = next(struct.unpack("B", rng.read(1)))
    while not ((bg_color_idx == 0 or bg_color_idx >= len(colors))
               and sum(colors[bg_color_idx]) > 0):
        bg_color_idx = next(struct.unpack("B", rng.read(1)))
    out.write(bytes([bg_color_idx]))

    # Write pixel aspect ratio which is fixed at 4:3.
    out.write(b"\x00")

    # Write flags for local colour table.
    local_colour_table_flag = True
    while not (local_colour_table_flag is False
               or local_colour_table_flag is True):
        local_colour_table_flag = bool(next(
            struct.unpack("B", rng.read(1))))
    out.write(b"\x01" if local_colour_table_flag else b"\x00")

    # If we use a global colour table write it here.
    palette = []
    if local_colour_table_flag:
        for _ in range(0x11):
            rgb = [
                next(struct.unpack("B", rng.read(1))) % 0x64,
                next(struct.unpack("B", rng.read(1))) % 0x64,
                next(struct.unpack("B", rng.read(1))) % 0x64]

            # Check that this colour hasn't already been used before. We want
            # to have 0x11 different colours available but it's also important
            # to keep them diverse enough so they can be easily distinguished by
            # human eyes when rendered in an image viewer. This check ensures
            # that no two colours are too similar.
            found_duplicate = False
            duplicate_i = 0
            for i, old in enumerate(palette):
                dist = sum(((rgb[j] - old[j]) ** 2)
                           for j in range(len(old))) **.5

                if dist < 0x16:
                    found_duplicate = True
                    duplicate_i = i
                    break

            if not found_duplicate:
                palette += [[rgb]]
            elif rng.read(1)[0] & 0xF!= 0:
                palette[duplicate_i] += [[rgb]]
            else:
                palette[duplicate_i] += \
                        [[int(round(.5*(rgb[j]+palette[duplicate_i][0][j])))
                          for j in range(3)]
                         ]

        colors += [sum(col, []) for col in zip(*palette)][-1:]

        # Pad the global colour table up until 0x11 entries long.
        num_colours = len(colors)
        assert num_colours <= 0x11
        padding = [(0, 0, 0)] * (0x11 - num_colours)
        colors += padding[:-(num_colours + len(padding))]

        # Write size of global colour table as 2 bytes. Make sure that it's always big endian.
        out.write(bytes([(num_colours * 3) & 0xFF]))
        out.write(bytes([((num_colours * 3) >> 8) & 0xFF]))

        for idx, c in enumerate(colors[:-1]):
            out.write(bytes([c[0]]))
            out.write(bytes([c[1]]))
            out.write(bytes([c[2]]))

    # Now start writing frames.
    for frame_index in range(0x1F+1):
        delay = next(struct.unpack("!H", rng.read(2))) % 0x7FFF
        if delay == 0:
            continue

        # Write header specifying the area on screen where this frame will be displayed.
        x = next(struct.unpack("!H", rng.read(2))) % (w//2)
        y = next(struct.unpack("!H", rng.read(2))) % (h//2)
        out.write(bytes([(x << 4) | ((y & 0xF) ^ 0xF)]))
        out.write(bytes([y >> 4]))

        # Write flags indicating whether there is transparency information present.
        has_transparency = bool(rng.read(1)[0] & 0x1)
        has_local_color_table = bool(rng.read(1)[0] & 0x80)
        interlace = bool(rng.read(1)[0] & 0x40)
        sort = bool(rng.read(1)[0] & 0x20)
        reserved = rng.read(1)[0]
        lct_size = 0
        extra_background = False
        transparent_color_index = 0
        if has_transparency:
            transparent_color_index = next(struct.unpack("B", rng.read(1)))
            if transparent_color_index == bg_color_idx:
                extra_background = True

        has_global_color_table = False
        if has_local_color_table:
            lct_size = 0x11
        else:
            has_global_color_table = True
            lct_size = max(lct_size, 0x11)

        # Write fields such as size of local colour table and subblocks of the image data.
        out.write(bytes([(lct_size << 4) | ((has_transparency << 2) & 0xC)]))
        out.write(b"\x00")
        if has_transparency:
            out.write(bytes([transparent_color_index]))
        else:
            out.write(b"\x00")

        # Write more stuff.
        out.write(bytes([delay & 0xFF]))
        out.write(bytes([(delay >> 8) & 0xFF]))
        out.write(b"\x00\x00")      # No disposal specified.
        out.write(bytes([0]))         # No user input expected.

        # Finally write the image data itself.
        mincodesize = 0x1
        imgdata = b""
        string_start = out.tell() + 1
        length = 0   # Number of bytes written.
        while length < w*h*4:
            if rng.read(1)[0] & 1:
                runlength = next(struct.unpack("B", rng.read(1)))
                if runlength > 0:
                    newdata = b"\x00"*runlength
                    if rng.read(1)[0] & 1:
                        newdata = b"\xFF"*runlength

                    imgdata += newdata
                    length += runlength
                    continue

            imgdata += rng.read(1)
            length += 1

        string_end = out.tell()


        # Calculate minimum code size needed for LZW compression later.
        if len(imgdata) < 0x80:
            mincodesize = 0x1
        elif len(imgdata) < 0x100:
            mincodesize = 0x2
        elif len(imgdata) < 0x200:
            mincodesize = 0x3
        elif len(imgdata) < 0x400:
            mincodesize = 0x4
        elif len(imgdata) < 0x800:
            mincodesize = 0x5
        elif len(imgdata) < 0x1000:
            mincodesize = 0x6
        elif len(imgdata) < 0x2000:
            mincodesize = 0x7
        elif len(imgdata) < 0x4000:
            mincodesize = 0x8
        elif len(imgdata) < 0x8000:
            mincodesize = 0x9
        elif len(imgdata) < 0x10000:
            mincodesize = 