#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:        BinaryIO=None,
        width:      int    = 1024,
        height:     int    = 768,
        colordepth: int    = 8,
        bgcolor:    int    = None,
        outfile:    str    = 'output.gif',
    ) -> None:

    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    blackcode = bytes([bgcolor & (255**2)])
    whitecode = bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])
    
    with open(outfile,'wb') as f:

        # The GIF header
        version = b"89a"
        screendescriptor = struct.pack('>HHBBB',
                width,
                height,
                0xF0 | colordepth,
                0,
                0,
        )
        
        # Write the header and descriptor block
        print("Writing the header")
        f.write(b"GIF"+version+screendescriptor)

        # Build the color lookup table
        print("Building the color map")
        colortable = []
        for i in range((2**colordepth)-1):
            x = (i*(bgcolor**(1/(256.0))))%(256**3)
            colortable += [struct.pack('>BBB',
                    int(((x>>16)&255)**1.5),
                    int(((x>>8)&255)**1.5),
                    int((x&255)**1.5),
                ),
            ]
        colortable += [blackcode]
        colortable += [whitecode]

        # We can now write the palette.
        print("Writing the palette")
        f.write(b"".join(colortable))

        # Now we need to write the image data itself...

        # Get some pixel values.
        print("Generating pixels...")
        pixels = []
        while len(pixels)<width*height:
            r = next(rng) % 256
            g = next(rng) % 256
            b = next(rng) % 256

            px = struct.pack('>BBB',r,g,b)
            
            if any(px==c for c in colortable):
                continue
                
            pixels += [px]

        print("Writing pixels...")    
        for rowstart in range(0,len(pixels),width):
            # Write the local image descriptor.
            f.write(bytes([
                0x2C,
                ord(","),
                chr(rowstart%256),chr(rowstart//256),
                chr(width%256),chr(width//256),
                0,
                0,
            ]))
            # Now write out each of those lines
            for i in range(height):
                f.write(pixels[i*width+rowstart])
            
        # And finally we write an end record.
        f.write(b"\x3B")


if