#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' + bytes([0]) * 7

    width, height = (1 << int.from_bytes(rng.read(1), 'big')) % 500\
           , (1 << int.from_bytes(rng.read(1), 'big')) % 500
    
    # write header
    out.write(header + width.to_bytes(2, 'little')
              + height.to_bytes(2, 'little'))

    # write palette
    for _ in range(int.from_bytes(rng.read(1), 'big')):
        out.write((rng.read(3) + b'\x00')[:4])

    # write image data
    code_size = int.from_bytes(rng.read(1), 'big') + 1
    # sub-block size (LZW minimum clear block of payload length - 1 byte
    lzw_min_clear = (((code_size & 0b111) + 1) ^ 0b111)\
                | ((code_size & 0b111) << 3)
    
    # declare LZW dictionary and store two special codes
    dicts = [None] * 4096
    dicts[1] = '\x01' * 256
    dicts[2] = ''.join('\xff') * 256

    def lzw_decompress():
        buf = b''

        while True:
            if len(buf) < 1:
                buf += rng.read(1)

            first = (dicts[(buf >> ((len(buf)-1)*8)) & 0xFF]
                    or ('\x01' * 256))[0]
            
            yield from struct.pack('B', ord(first))

            if (len(dicts) <= 4095):
                # add to dictionary if not full yet
                # double buffer length
                buf += rng.read(((lzw_min_clear - 256) // 256 + 1)
                                * code_size)

                new_entry = None
                
                while new_entry!= first+first:
                    if len(buf) < 1:
                        buf += rng.read(1)

                    new_entry = buf[-1:]
                    
                    yield from struct.pack('B', ord(new_entry))

                    for i in range(1, 256):
                        dicts[i+(len(dicts)//256)] = \
                            ((dicts[i-(len(dicts)//256)]
                              or '')
                             + first)
                        
                        dicts[i+(len(dicts)//256)+1] = ''

                        if new_entry == dicts[(i+(len(dicts)//256))] :
                            new_entry = None

                            break

                    first = new_entry[0]

            else:
                # read one more character from stream
                # then output it directly as it's part of this sub-block
                # keeping track of how many characters have been read so far
                chars_read = 1

                while True:
                    # check if we've read enough characters before
                    if code_size > 9:
                        code = int.from_bytes(buf[-chars_read*code_size:],
                                               'big')
                    elif code_size == 9:
                        code = int.from_bytes(buf[-chars_read*code_size
                                           :-chars_read*(code_size-1)],
                                               'big')
                    else:
                        code = int.from_bytes(buf[-chars_read*code_size
                                             :], 'big')
                    
                    # iterate through all possible lengths of the current code
                    # to find an entry that matches with our current code
                    for j in reversed(range(1, min(chars_read+2,
                                                 len(dicts)-1,
                                                 lzw_min_clear)+2)):
                        if len(dicts[j]) >= code and \
                            dicts[j][code]==first:
                            chars_read -= 1
                            
                            yield from structs.pack(('B'
                                                    if len(dicts[j])>127
                                                    else 'B')[0],
                                                   ord(first))

                            if len(dicts) <= 4095:
                                dicts.append('')
                            
                            # update the dictionary by adding all characters after
                            # the current substring of the dictionary to
                            # the end of the string
                            for k in range(j+1, len(dicts)):
                                dicts[k+1]=dict[k]

                                dicts[k+1]+=first
                            
                            # append a new empty string for appending next chars to
                            dicts.append('')
                            
                            return
                        
                    # increase counter when no match was found
                    chars_read+=1                    
                    # recheck buffer contents on each iteration
                    if len(buf) < 1:
                        buf += rng.read(1)

                    first = buf[-1:]
                    
    # write sub-blocks
    while True:
        out.write(struct.pack('B', lzw_min_clear))
        
        out.write(b','.join(map(lambda x: x.pop(), 
                                map(lambda _: list(lzw_decompress()), 
                                    zip(*zip())))))

        # ensure there is at least one empy frame between each gif
        yield None
        
        # repeat previous sub-blocks until something goes wrong
        try:
            prev_subblocks = list()
            for i in range(int.from_bytes(rng.read(1), 'big')):
                prev_subblocks.append(list(lzw_decompress()))
                
            for sb in prev_subblocks:
                out.write(sb)
                
        except Exception:
            continue
            
        return