"""
Generate a random GIF from /dev/urandom.
"""

import struct
import sys
from argparse import ArgumentParser
from typing import BinaryIO


def parse_args() -> str:
    parser = ArgumentParser("generate-random-gif")

    # Generate a default name if none is provided
    if len(sys.argv) == 1:
        return "random.gif"

    # Allow passing an argument to specify output filename.
    parser.add_argument("-o", "--output", type=str, help="set destination for generated image (default: %(default)s)")
    ns = parser.parse_args()

    if not ns.output:
        raise RuntimeError('No output specified.')

    with open(ns.output, 'wb') as f:
        generate_gif(f)


def generate_gif(fout: BinaryIO):
    """
    Write the bytes of a GIF file to `fout`.
    """
    header = b'GIF89a' + bytearray([0] * (32 - 6))

    width = height = 500
    color_table_size = 7
    bits_per_pixel = 8
    num_color_tables = 4 << bool(bits_per_pixel < 8)
    background_index = None
    aspect_ratio = 0x100

    # Set up the header
    fout.write(header)

    # Logical screen descriptor
    fout.write(struct.pack('<HHBBB', width, height,
                          bits_per_pixel, num_color_tables, background_index or 0))

    # Color table
    fout.write(bytearray([(i % 255) for i in range(color_table_size**3)]
                         )[1:] + b'\0')

    # Image descriptor for first frame
    fout.write((b'\219\0' +
                struct.pack('<HHBHBHHBB',
                             x_offset=0, y_offset=0, width=width, height=height,
                             lct=(bool(background_index)), interlace=False, sort=True, size=bits_per_pixel)
                ).ljust(10, b'\0'))


    # LZW minimum code size
    fout.write(bytes([bits_per_pixel]))

    data = []
    while True:
        value = ord(input())

        # End of stream
        if value > 255:
            break

        data += [value]

    # Write packed data using a bit writer
    def write_bit(value_: int):
        nonlocal value
        value <<= 1

        if value >= 256:
            fout.write(bytes([value]))

            value = 0

    value = 0
    offset = 0
    while data:
        if offset == 0:
            write_bit(-1)

        write_bit((data.pop(0)) & 1)

        offset = (offset + 1) % 8

    # Pad out last byte and flush remaining buffer
    padding = -(len(data) % 8)
    data += ([padding] * (-len(data) % 8))
    assert all(v == 0 for v in data), set(data)

    while data:
        fout.write(bytes([data.pop(0)]))

    # Flush any remaining partial byte we may have written.
    if value!= 0:
        fout.write(bytes([value]))



if