#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generate a random GIF given an input stream of bytes to use for randomness and output stream to write to.
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    # Create the header
    header = b"GIF89a"

    # Get the length of the image from the RNG
    width  = (int.from_bytes(rng.read(1), 'big') >> 4) + 1
    height = ((int.from_bytes(rng.read(1), 'big') & 0xf) << 8) | int.from_bytes(rng.read(1), 'big')
    print("Generating {}x{} GIF".format(width,height))

    # Write it all together
    out.write(header)
    out.write(struct.pack('>H', width))
    out.write(struct.pack('>H', height))
    out.write(b'\xf7\x00') # Global color table with one entry

    # Build the palette
    current_color   = bytearray()
    colors          = []
    color_index     = -1

    while len(current_color)<3 or any([c!= current_color[0] or c!= current_color[1] or c!= current_color[2] for c in current_color]):
        if len(current_color)==3:
            color_index += 1

            if len(colors) <= color_index:
                colors.append((
                    current_color[0],
                    current_color[1],
                    current_color[2]))

        current_color = bytearray([
            rng.read(1)[0],
            rng.read(1)[0],
            rng.read(1)[0]])

    # Write it
    for i, color in enumerate(colors):
        assert isinstance(color, tuple)
        assert len(color) == 3
        rgb = list(map(lambda x : x>>3<<5|x>>2<<3|x>>3&0b111, color))
        out.write(struct.pack('<BBB', *rgb ))

    # Keep track of where we are writing data so we can build the graphic control extension later on
    pos = out.tell()


    # Write the frames
    for i in range(width*height//64):
        size    = rng.read(1)[0]&0xff
        code    = (size>>3)&0x07
        flag    =(size>>7)&0x01
        size   &=0x7f

        # Clearing the screen? Then set the background color as our frame
        if code==0:
            bg_idx = rng.read(1)[0] % len(colors)
            out.write(bytes([bg_idx+1])) # Faster than using get()
            continue


        # LZW stuff
        buf        = bytearray()
        prev       = None
        lookup     = [ None ]*(1<<code)+[ prev ]
        cur        = []
        buffer     = [[-1,-1]]*4

        while len(buf)<size:
            if not prev is None: # Do we have something from before?
                cur = []

                if len(buffer)>0:
                    for j in reversed(range(len(buffer))):
                        idx=lookup.index(tuple(buffer[:j+1])+tuple(cur))+1

                        if idx < 2**(1+code):
                            break

                    cur      += buffer[:j+1]
                    del buffer[:j+1]


                # If we didn't find anything, just add more bytes
                if len(cur)<=0:
                    cur.append(prev)
                    cur+=list(rng.read(1))

                lookup.append(tuple(cur))


            # We only store the last two entries because they're always going to be the same thing anyway!
            if len(cur)>2:
                buf.append(len(lookup)-1)

            print(flag, code, hex(prev), cur, end=' ')
            prev = cur[-1]+1 if len(cur)>0 else None
            if flag!=0:
                buffer = [
                    rng.read(1)[0],
                    rng.read(1)[0],
                    rng.read(1)[0],
                    rng.read(1)[0]]
            print(buffer)

        print()







    # Write global trailers
    gct_pos           = out.tell()-pos
    out.seek(gct_pos+4)
    out.write(struct.pack(">HHB",
        len(colors),
        0,
        0))

    end_of_file_pos   = out.tell()
    out.seek(end_of_file_pos-2) # Write trailer size at previous position
    out.write(struct.pack(">H", end_of_file_pos-pos))
