#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# but not larger than the maximum values allowed by the specification
def gen_random_gif(rng: 'Generator[int,None,None]',
                   fc: int,
                   out: BinaryIO):
    assert(fc <= 256)
    
    # append file header
    out.write(b'GIF89a')

    # set up logical screen descriptor
    max_dim = 1<<12
    width = rng.randrange(max_dim//2,max_dim*2)
    height = rng.randrange(max_dim//2,max_dim*2)
    flags = 0b10000000 | (rng.randint(0,2)<<2)
    bgcolor = rng.getrandbits(8)
    aspectratio = 0
    out.write(struct.pack('<HHBBB',
                          width,         # image width
                          height,        # image height
                          0x87,          # pixel size
                          flags,         # flags
                          bgcolor,       # background color index
                          aspectratio))  # pixel aspect ratio

    # initialize global color table with 256 colors
    for i in range(fc):
        if rng.randint(0,2)==0:
            r = rng.getrandbits(8)
            g = rng.getrandbits(8)
            b = rng.getrandbits(8)
        else:
            r = g = b = 0
        out.write(struct.pack('BBB',r,g,b))

    extblocks = []

    # add an application extension
    appcmd = rng.getrandbits(8)
    appflags = rng.getrandbits(8)
    appsize = rng.randrange(0,65535)
    applength = appsize + len(b'\x00\xfe') + 1
    appdata = [appcmd] + [(appflags>>(8*i))%256 for i in range(3)] + \
              [applength%(2**16),applitghgt%(2**16)]
    appdatalen = sum(map(len,appdata))+(appdatalen<0x100)
    appdatalenbytes = bytes([(appdatalen>>8)%256,(appdatalen<<8)%256])
    extblocks.append(('NETSCAPE2.0',
                      bytes([len(applength)+appdatalen-2])+appdatalenbytes+\
                      bytes(appdata)+'\x00\xfe'))

    # TODO: other extensions

    # append graphics control block
    delaytime = rng.randint(0,65535) # frame duration is 1/100th sec
    transparentindex = rng.randint(0,fc) if fc > 0 else 0
    disposalmethod = rng.randint(0,3) 
    out.write(struct.pack('BBBHHB',0x21,0xf9,
                          4,
                          0<<1|disposalmethod,
                          delaytime,
                          0<<1|transparentindex))

    # append graphic control label
    out.write(b'\x00')

    # initialize local color table
    for i in range(fc):
        if rng.randint(0,2)==0:
            r = rng.getrandbits(8)
            g = rng.getrandbits(8)
            b = rng.getrandbits(8)
        else:
            r = g = b = 0
                out.write(struct.pack('BBB', r, g, b))

    # append image descriptor block
    out.write(struct.pack('BBHBBBB',
                        0xf7, # separator
                        0,    # left corner x coordinate
                        height&0xff,   # lower corner y coorindate
                        ((height>>8)&0x0f)<<4 | \
                          (width&0x0f)>>2,  # size flag
                        (fc & 0xf)|((((width&0x0f)<<2)&0xf0)),
                            0,
                            0))               # no local color table

    # write image data blocks
    lastbyte = None
    for j in range(fc):
        for i in range(((width+7)/8)*height):
            byte = rng.getrandbits(8)%256
            if lastbyte!= None:
                while rng.randint(0,2)==0:
                    byte |= lastbyte
            lastbyte = byte ^ (i==0) ^ (((j == 0) or
                                        (j == (fc - 1))) << 7)
            out.write(struct.pack('B',byte))

    # copy and extend extension blocks
    for name,data in extblocks:
        data += bytes([sum(map(ord,name))+len(data)])*(8-(len(data)%8))
        out.write(data)
