#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and returns it.
def generate_gif(seed: int, out: BinaryIO):
    rng = struct.unpack('<Q', seed.to_bytes(8, 'little'))[0]

    # Header
    out.write(b'GIF89a')  # file signature
    out.write((256 + rng % 11).to_bytes(1, 'little'))  # width and height (max=256)
    out.write(b'\x87\x00')  # flags: global color table flag set, sorted colors flag not set
    out.write(((rng & ((1 << 24)-1)) | (((rng & ~((1<<24)-1))) >> 24)).to_bytes(3, 'little'))  # background color (random)
    out.write(b'\xff\x00')  # pixel aspect ratio

    # Color table
    palette = [i for i in range(256)]
    for i in range(len(palette)):
        next_idx = (i+rng)%len(palette)

        r = palette[i] & 0xff; r ^= palette[next_idx] & 0xff
        g = (palette[i]>>8) & 0xff; g ^= (palette[next_idx]>>8) & 0xff
        b = (palette[i]>>16) & 0xff; b ^= (palette[next_idx]>>16) & 0xff

        palette[i] = (r<<16)|(g<<8)|b

    for c in reversed(palette):
        out.write(c.to_bytes(1, 'little'))

    # Image data
    for _ in range(256//2*256//2):
        if rng%2 == 0:
            # RLE compression: two pixels with same color
            out.write(struct.pack('B', rng&0xff))
            out.write(struct.pack('B', 0))

            out.write(struct.pack('B', (rng>>(1*8))&0xff))
            out.write(struct.pack('B', (rng>>(2*8))&0xff))
            out.write(struct.pack('B', (rng>>(3*8))&0xff))
        else:
            # no RLE compression: four different pixels
            out.write(struct.pack('BBBB', rng&0xff, (rng>>(1*8))&0xff, (rng>>(2*8))&0xff, (rng>>(3*8))&0xff))

    return bytearray(out.getvalue())

