#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# between minwidth <= w < maxwidth and minheight <= h < maxheight (inclusive).
# The frame count is randomly generated from 1 to nframes inclusive.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, minwidth=64,
        maxwidth=512, minheight=64, maxheight=512, nframes=10):

    # construct headers
    header = b'GIF87a'
    width = rng.randint(minwidth,maxwidth-1)//2*2+2
    height = rng.randint(minheight,maxheight-1)//2*2+2
    gctsize = 2 << bool(rng.randint(0,1))
    globalcolortablelen = 2**(gctsize+1)-1
    localcolortablelen = 2**((bool(rng.randint(0,1)))<<2)
    fc = rng.randrange(nframes)+1
    extblocks = [
        (b'NETSCAPE2.0', b'\x03\x01\x00\x00\xff\x00'),
        (b'ANIMEXTS1.0', b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')]


    # write headers
    out.write(header)
    out.write(struct.pack('<HHB', width, height, gctsize))

    # append colortables if any
    for i in range(globalcolortablelen//3 + bool(globalcolortablelen%3)):
        r,g,b = tuple([int(v*255) for v in rng.choice(range(256))][0] for _ in "rgb")
        out.write(struct.pack('BBB',r,g,b))
    if localcolortablelen > 0:
        for i in range(localcolortablelen // 3 + bool(localcolortablelen % 3)):
            r, g, b = tuple([int(v * 255) for v in rng.choice(range(256))][0] for _ in "rgb")
            out.write(struct.pack('BBB', r, g, b))

    # write lsd
    out.write(bytes([(fc&0xff),(fc>>8)]))

    # write blocks
    for blockid, blockdata in extblocks:
        blocksize = len(blockdata)
        out.write(struct.pack("<BB", 0x21, blockid[0]))
        out.write(struct.pack("<BBBB", blocksize & 0xff, (blocksize >> 8) & 0xff,
                             ((blocksize >> 16) & 0xf) | (((~blocksize)&0xf)<<4),
                             0))
        out.write(blockdata)
    for fnum in range(fc):
        # image descriptor
        out.write(b"\x2c")
        left = fnum * width
        top = fnum * height
        out.write(struct.pack("<HHHH", left,top,width,height))

        flag = int("1"*bool(fnum)==str(int(True)),2)
        out.write(bytes([(flag<<7)|9])) # lzw minimum size=9
        out.write(b"\x00") # no palette used

        #image data
        for y in range(height):
            for x in range(width):
                pixel = rng.randrange(0,256)
                out.write(bytes([pixel]))


if