#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generate a random GIF given an input stream of bytes to use for randomness and output stream to write to.
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    # Create the header
    header = b"GIF89a"

    # Get the length of the image from the RNG
    width: int = struct.unpack('>H', rng.read(2))[0] % 320 + 1
    height: int = struct.unpack('>H', rng.read(2))[0] % 240 + 1
    maxcolorbits: int = rng.read(1)[0] & 0x07 # Mask off high bits which aren't used
    globalcolortablesize: int = 2**(maxcolorbits+1)
    mincodesize: int = rng.read(1)[0] & 0x07

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    out.write((width).to_bytes(2, "big"))
    out.write((height).to_bytes(2, "big"))
    out.write(((globalcolortablesize >> 3) | ((mincodesize & 0x07) << 4)).to_bytes(1, 'little'))
    out.write((0).to_bytes(1, 'little'))

    # We'll skip the background color field since it's not important
    bgcolorindex = 0xFF
    bgcolor = bytearray([bgcolorindex])
    out.write(bgcolor)

    # Likewise, ignore pixel aspect ratio
    pixelaspectratio = 0b00111000
    out.write(pixelaspectratio.to_bytes(1, 'little'))

    # Now write the global color table
    for _ in range(globalcolortablesize):

        red = float(rng.read(1)) / 255.0
        green = float(rng.read(1)) / 255.0
        blue = float(rng.read(1)) / 255.0
        alpha = float(rng.read(1)) / 255.0

        # BGRA formulation is correct
        colorrep = [blue*255.0, green*255.0, red*255.0, alpha * 255.0]
        repbytes = bytearray(int(c+.5) for c in colorrep)
        out.write(repbytes)

    # The application extension to configure loop count
    appext = b'\xfe'
    appid = b'NETSCAPE2.0'
    authcode = b''
    junk = rng.read(1)
    loopcount = 1
    extlen = len(appid)+1+len(authcode)+1+(junk[0]&0x0F)
    subblock = bytearray([
        extlen,
        0x0B,
        ord("N"), ord("E"), ord("T"), ord("S"), ord("C"), ord("A"), ord("P"),
        ord("E"), ord("."), ord("0")
    ])
    if len(authcode)!= 0:
        pass
    out.write(appext)
    out.write(subblock)

    # And then the actual NETSCAPE block containing the loop count
    netscapeext = b'\x01'
    netscapeblock = bytearray([loopcount, 0])
    out.write(netscapeext)
    out.write(netscapeblock)

    # Finally! That's all the blocks in the image! Next up: graphics control block

    # The graphic control block looks like so:
    # uint8_t             BlockSize;         /* Size of the block */
    # uint8_t             PackedFields;      /* Flags to indicate what fields follow. Should be 0. */
    # uint16_t            DelayTime;         /* Number of milliseconds after drawing new frame before displaying next one. */
    # uint8_t             TransparentColorIndex; /* Index number of transparent color if any. Set to 0-2^n-1 where n is bit depth minus 1. */
    # uint8_t             Terminator          /* Must always be equal to 0. */

    # Image flags are packed 1 bit each with as follows:
    #   1 - Color table sort flag
    #   2-4 - GCT size (LZW minimum size)
    #   5   - Local colors sorted flag
    #   6   - Interlace flag (NOT USED)
    #   7   - Sort flag (NOT USED)
    #   8   - Reserved for future use

    # We set interlace off, color table sorted on, local colors sorted off,
    # and everything else off that we don't care about.
    gctsortflag = 1
    lctsortedflag = 0
    interlaced = False
    sortflag = 0
    reserved = 0
    imgflags = (gctsortflag << 7) + (lctsortedflag << 6) \
                + (interlaced << 5)     + (sortflag << 4) \
                + (reserved << 3)       + mincodesize
    out.write(imgflags.to_bytes(1, 'little'))


    # Skip the rest for now because I'm gonna need to learn more about this format later.

    return