#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:        BinaryIO=None,
        width:      int    = 1024,
        height:     int    = 768,
        colordepth: int    = 8,
        bgcolor:    int    = None,
        outfile:    str    = 'output.gif',
    ) -> None:

    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    blackcode = bytes([bgcolor & (255**2)])
    redcode = bytes([(bgcolor>>16) & 255])
    greencode = bytes([(bgcolor>>8) & 255])
    bluecode = bytes([bgColor & 255])

    colortable = [blackcode+redcode+greencode+bluecode] + \
                 [(bytes([i])+bytes([j])+bytes([k])+bytes([l]))
                  for i in range(256)
                  for j in range(256)
                  for k in range(256)
                  for l in range(256)]
    
    bytecount = 10       # The header size
    bytecount += len(colortable) * 4

    for y in range(height-1):
        for x in range(width//8):
            b = 0xff
            bitnum = 7
            while b!= 0 and bitnum >= 0:
                b = rng.read(1)[0]
                sbyte = bytes([b])

                if sbyte[0] == 0:
                    pass
                elif sbyte[0] <= 127:
                    bytecount += 1
                else:
                    bytecount += 2
                    
                bitnum -= 1

    f = open(outfile,"wb")

    with f:
        print("Writing the initial header")
        f.write(struct.pack("<2sBBBBB", "GIF", ord('8'), 1, 0, 0, 0))
        
        print("Writing the comment")
        f.write(b"; Generated by gifsicle")

        print("Writing the logical screen descriptor")
        f.write(struct.pack('<HHHHBB', width, height, colordepth,
                           bytecount % 256 ** 2, bytecount // 256 ** 2,
                           0xf7))

        print("Writing the palette")
        f.write(b"".join(colortable))

        # Now we need to write the image data itself...
        print("Writing the image")
        f.write(b"\x2C\x00\x00\x00\x00\x00\x00\x00\xFF")   # Image descriptor
        f.write(b"!\xF9\x04\x01\x00\x00\x00,\x00\x00\x00") # Graphics control extension
        f.write(b"\x00\x01\x00\x21\xFE\x09D\x0B\xCF\x00") # Application Extension
        f.write(b"\x0BNETSCAPE2.0\x03\x01")                  # Application Data
        f.write(b"\x00\x00\x00!")                           # Trailer
        print("\nDone.")