"""
Generate a random GIF from /dev/urandom.
"""

import pathlib
import struct
import sys
from argparse import ArgumentParser
from typing import BinaryIO


def parse_args() -> pathlib.Path:
    """Parse arguments."""
    parser = ArgumentParser("generate-random-gif")

    # Generate a default name if none is provided
    if len(sys.argv) == 1:
        return pathlib.Path("random.gif")

    # Allow passing an argument to specify output filename.
    parser.add_argument("-o", "--output", type=pathlib.Path, help="set destination for generated image (default: %(default)s)")
    ns = parser.parse_args()

    if not ns.output:
        raise RuntimeError('No output specified.')
    
    return ns.output


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, numbytes: int
    ):
    header = bytes([
        0x47, 
        0x49, 
        0x46, 
        0x38, 
        0x39, 
        0x61]
    ) 
    datalen = min(numbytes, 2**16 - 1)
    data = rng.read(datalen)
    out.write(header + struct.pack(">H", datalen))
    out.write(data[:datalen // 2]) # Write first half of data, then second half
    # This ensures that we generate a valid GIF by filling any extra bits with null bytes.
    out.write((b'\x00' * ((datalen & 1) ^ 1))) 


if __name__ == "__main__":
    outfilename = parse_args()
    print(f"Outputting randomly generated GIF at '{outfilename}'.")
    with open('/dev/urandom', 'rb') as rng:
        with open(outfilename, 'wb') as out:
            numframes = 5 
            framesize = 16384
            for _ in range(numframes):
                generate_random_gif(rng, out, framesize)