#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# between minwidth <= w < maxwidth and minheight <= h < maxheight (inclusive).
# The frame count is randomly generated from 1 to nframes inclusive.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, minwidth=64,
        maxwidth=512, minheight=64, maxheight=512, nframes=10):

    def randbytes(): return bytes(rng.randrange(256)
                                  for _ in range(rng.randrange(8)+1))

    # construct headers
    header = b'GIF87a'
    width = rng.randint(minwidth,maxwidth-1)//2*2+2
    height = rng.randint(minheight,maxheight-1)//2*2+2
    gctsize = 2 << bool(rng.randint(0,1))
    globalcolortablelen = 2**(gctsize+1)-1
    localcolortablelen = 2**((bool(rng.randint(0,1)))<<2)
    fc = rng.randrange(nframes)+1
    extblocks = [
        (b'NETSCAPE2.0', b'\x03\x01\x00\x00\xff\x00'),
        (b'ANIMEXTS1.0', b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')]


    # write headers
    out.write(header)
    out.write(struct.pack('<HHB', width, height, gctsize))

    # append colortables if any
    for i in range(globalcolortablelen//3 + bool(globalcolortablelen%3)):
        out.write(randbytes())
    if localcolortablelen > 0:
        for i in range(localcolortablelen // 3 + bool(localcolortablelen % 3)):
            out.write(randbytes())

    # append image descriptor block
    lzwmincodesize = rng.randrange(9)+1;
    out.write(struct.pack('BBHBBBB',
                        0xf7, # separator
                        0,    # left corner x coordinate
                        height&0xff,   # lower corner y coorindate
                        ((height>>8)&0x0f)<<4 | \
                          (width&0x0f)>>2,  # size flag
                        lzwmincodesize|(lzwmincodesize<<4),
                            0,
                            0))               # no local color table

    # init LZW encoder
    clearcode = 1<<(lzwmincodesize-1)
    endofinfocode = clearcode+1
    nextcode = clearcode+2
    entab = {chr(c): c for c in list(range(32,119))}
    entab['>'] = 120      ;entab['<'] = 121
    entab['^'] = 122      ;entab['_'] = 123
    entab['|'] = 124      ;entab['\\'] = 125
    entab['{'] = 126     ;entab['}'] = 127
    detab = dict([(y,z) for z,y in entab.items()])
    tab = {}
    code = []
    buf = ''
    
    # write image data blocks
    lastbyte = None
    for j in range(fc):

        for i in range(((width+7)/8)*height):

            byte = rng.getrandbits(8)%256
            
            # update LZW compressor state
            if len(buf)>8 or byte!=lastbyte:
                
                buf+=chr(byte)

                if byte not in entab:

                    key = ''.join(reversed(buf))
                    
                    try:

                        # use existing entry
                        out.write(bytes([detab[key]]))

                    except KeyError:
                    
                        # register new entry
                        out.write(bytes([nextcode]))
                        
                        # add new entry to tables
                        tab[key] = nextcode
                        entab[key] = nextcode
                        detab[nextcode]=key
                        
                        nextcode+=1
                        if nextcode==(2<<lzwmincodesize):
                            
                            # reset coder
                            nextcode = clearcode+2
                            lzwmincodesize+=1
                            tab = {}
                            entab = {chr(c): c for c in
                                     list(range(32,119))}
                            entab['>'] = 120; entab['<'] = 121 
                            entab['^'] = 122; entab['_'] = 123
                            entab['|'] = 124; entab['\\'] = 125
                            entab['{'] = 126; entab['}'] = 127
                            
                    
                else:
                
                    tmp = chr(entab[buf[-1]])
                    
                    if len(buf)<2 or tmp!= buf[-2]:
                        out.write(bytes([entab[tmp]]))

                buf=''
                
            lastbyte = byte ^ (i==0) ^ (((j == 0) or
                                    (j == (fc - 1))) << 7)
            
    # flush buffer
    code=[clearcode]+code
    while len(code)<4096:   code.append(code[-1])
                    
    # write final LZW encoded block
    out.write(bytes([endofinfocode])+
              bytes([len(code)]))+bytes(code)

    # copy and extend extension blocks
    for name,data in extblocks:
        data += bytes([sum(map(ord,name))+len(data)])*(8-(len(data)%8))
        out.write(data)

