#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated file has the following properties:
# - no animation duration
# - color tables of 1 to 7 colors
# - LZW compression with a minimum code size of at least two
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    header += struct.pack('>H', rng.randint(200, 500))
    header += struct.pack('>H', rng.randint(200, 500))

    # color table (RGB triplets)
    numcolors = rng.randint(1, 7)
    for i in range(numcolors):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        header += struct.pack('<BBB', r, g, b)

    # palette disposal method is none
    header += b'\x00\x00'

    # palette updated flag set
    header += b'\xFF'

    # graphics control extension exists
    header += b'\x21\xF9\x04'

    # delay time value is zero
    header += b'\x00' * 2

    # transparent color index is unset (-1 => not set)
    header += b'\x00'

    # end of graphic control block
    header += b'\x00'

    # application extension containing comment "P"
    header += b'\x21\xFF\x0B'
    header += b'Comment'.ljust(11, '\x00').encode()

    # image descriptor
    header += b'\x2C'

    # left position is zero
    header += b'\x00' * 2
    # top position is zero
    header += b'\x00' * 2
    # width and height are 16 bits each
    header += struct.pack('>HH',
                          rng.randint(int(header[6] * header[7]),
                                       int(header[6] * header[7]) + 200),
                          rng.randint(int(header[8] * header[9]),
                                       int(header[8] * header[9]) + 200))
    # local color table flag is clear
    header += b'\x00'
    # interlace flag cleared
    header += b'\x00'
    # sort flag cleared
    header += b'\x00'
    # palette size is one entry per pixel minus one
    # (so if there's only one color we have one palette)
    header += b'\x01'
    # background color index is zero
    header += b'\x00'

    # local color table follows
    for i in range(numcolors):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        header += struct.pack('<BBB', r, g, b)

    datalen = rng.randint(1, 2 ** 16) // 2
    compressed = []
    clearcode = 2 ** 8
    endcode = 2 ** 9
    # first character is delimiter which doesn't need
    # compression
    compressed.append(-1)
    previous_char = -1
    while len(compressed) < datalen:
        char = rng.randint(0, 2 ** 16) // 2
        # follow repeated characters with repetition count
        # of 'one more than the last occurence of that character'
        if char == previous_char:
            repcount = min((2 ** 12) - 1,
                           1 + compressed.count(previous_char))
            compressed.append(repcount)
        else:
            compressed.append(char)
            previous_char = char

        # write escape codes every 256 characters or so
        if len(compressed) % 256 == 0:
            compressed.append(clearcode)
            compressed.append(endcode)

    # now compress it all
    lzwmincodelen = max(2,
                        (numcolors - 1).bit_length())
    def outputbits(nbits, nshifted):
        nonlocal compressed
        while True:
            if nbits <= 8:
                yield (nshifted >> ((-nbits) & 7)).to_bytes(
                    ceildiv(nbits, 8), byteorder='big')
                break

            chunklen = min(ceildiv(nbits, 8),
                           len(compressed) - len(outputbits.buf))
            outputbits.buf += compressed[:chunklen].to_bytes(
                chunklen, byteorder='little')
            del compressed[:chunklen]

            yield outputbits.buf[-ceildiv(nbits, 8):]
            nbits -= chunklen * 8
            nshifted &= (1 << nbits) - 1

    outputbits.buf = bytearray()
    bitstream = b''.join(outputbits(8, c)
                         for c in [lzwmincodelen + 1]
                         )
    bitstream += b''.join(outputbits(12, c)
                          for c in compressed[:-2]
                          )
    compressedlasttwo = compressed[-2:]
    bitstream += b''.join(outputbits(8, c)
                          for c in compressedlasttwo
                          )
    compressedlastfour = compressed[-4:-2]
    bitstream += b''.join(outputbits(8, c)
                          for c in compressedlastfour
                          )

    # compute final length
    clcl = ceillog2(len(compressed) + 2)
    bitstreamlen = sum(ceildiv(c, clcl)
                       for c in [numcolors * 3,
                                 lzwmincodelen + 1,
                                 12 + (len(compressed) - 2) * clcl,
                                 8 * len(compressedlasttwo)])

    # concatenate everything together
    header += b'%s%s' %(struct.pack('>L', bitstreamlen + 1)[1:],
                        bitstream)
    out.write(header)


def ceildiv(dividend, divisor):
    return dividend // divisor + bool(dividend % divisor)

def ceillog2(x):
    assert x > 0
    return x.bit_length() - 1