#!/usr/bin/env python3

import random
import struct
from typing import BinaryIO


# Generate an integer in the range [low, high].
def randint(low: int,high: int)->int:
    return random.randint(low,high-1)


# Generates a random RGB pixel value.
def gencol() -> int:
    rtn = 0
    rtn |= randint(0,256)**2
    rtn |= randint(0,256)<<8
    rtn |= randint(0,256)<<16
    return rtn


# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:           BinaryIO=None,
        width:         int     = 1024,
        height:        int     = 768,
        colordepth:    int     = 8,
        bgcolor:       int     = None,
        frames:        int     = -1,
        duration:      float   =.05,
        outfile:       str     = "output.gif",
        transparent:   bool    = False,
    ):
    """Generate a random GIF animation."""


    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor<0 or bgcolor>(255**3):
        bgcolor = 0xBFBFBF ^ ((width*height)%(255**3))

    blackcode = bytes([bgcolor&(255**2)])
    whitecode = bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])
    
    with open(outfile,"wb") as f:

        # The GIF header
        version = b"89a"
        screendescriptor = struct.pack(">HHBBB",
                width,
                height,
                0xF0|colordepth,
                0,
                0,
        )
        
        # Write the header and descriptor block
        print("Writing the header")
        f.write(b"GIF"+version+screendescriptor)

        # Build the color lookup table
        print("Building the color map")
        colortable = []
        for _ in range((2**colordepth)-1):

            c = gencol()
            
            while c==bgcolor:                
                c = gencol()


            colortable += [(bytes([c&(255**2)]))]
            colortable += [(bytes([(c>>8)&255,(c>>16)&255]))]

        
        colortable += [blackcode]
        colortable += [whitecode]

        # We can now write the palette.
        print("Writing the palette")
        f.write(b"".join(colortable))


        # Some basic variables which we'll use later
        frames = min(-frames,256)
        delaytime = max(.01,.001+(duration/100)*100)
        transindex = -1

        print("Writing the frame information")
        imdat = bytearray()
        for i in range(frames + (frames % 2)):
            bgcolor = 0xBFBFBF ^ (((i*256) % ((255*256)**3)))
            bgcode = [0]+[bgcolor&(255^2)**2][:-1]+[(bgcolor>>8)&255,(bgcolor>>16)&255]
            code = bytes(bgcode+[0]*(3*(2**colordepth-len(bgcode))))

            islastframe = True if i == (frames-1) else False

            datsize = len(imdat)
            lzwminsize = 2**(colordepth+1)+2
            datasize = datsize+lzwminsize
            chunkdata = b""

            bitmask = 0x80>>(datasize%8)

            # LZW minimum size plus one byte of colour info per row
            lineinit = bytes([lzwminsize//256,lzwminsize%256])+bytes(range(2**colordepth))
            lineinit += lineinit[:lzwminsize]

            imdat += code

            # Gather all the pixels together for this frame
            pxls = bytearray()
            for y in range(height):

                firstbyte = 0
                lastbyte = 0

                # Allocate enough space to read in the next scanline's worth 
                # of data
                totalline = [0xff] * (datasize+2)
                
                for x in range(width):

                    c = randint(0,2**colordepth)
                    
                    if c==(2**colordepth)-1:
                        c-=1
                        
                    if transparent and randint(1,10)==1 and c!=transindex:
                        
                        transindex = c

                        if c>=len(colortable):
                            raise Exception("Transparent index is invalid!")
                            
                        totalline[-(2**colordepth)+c]=0

                    pxls += colortable[c]


                    if len(pxls)>=(datasize+2):


                        # Grab the first part of what we just added 
                        if bitmask!= 0x01:
                            firstbyte = ord(pxls[-datasize-1])
                            mask = ~bitmask
                            firstbyte &= mask

                        # replace the part we want tweaked by moving it over
                        lastbyte = ord(pxls[-(datasize)])
                        lastbyte >>= (8-bitmask)
                        lastbyte <<= (bitmask*-1)

                        numbits = len(pxls)-(datasize+1)

                        #Move everything back 1 byte so we can fill out 
                        #the final byte at the end
                        pxls[:-numbits] = pxls[numbits:]
                        del pxls[-numbits:]

                        #We are always going to have the same item on the end
                        #and it must be a literal, so reset its code 
                        #to zeroes
                        lastbyte |= 0x00

                        if firstbyte >= 0x100:
                            chunkdata += chr(firstbyte)
                        elif len(chunkdata)!=0:
                            chunkdata+=chr(0)

                        chunkdata += chr(lastbyte)
                        
                        # Fill in any extra bits
                        totalline[-(datasize+2)] = firstbyte
                        totalline[-datasize] = lastbyte

                        # Insert the rest of the bytes:
                        # the new colour info that was added earlier, 
                        # and then whatever we still need from pxls
                        totalline[:-numbits] = pxls[datasize+2:(datasize+2)+(datasize+2)]

                        # Reset the array after shuffling around stuff
                        del pxls[:]
                        pxls += totalline
            

                # If there are no more pixels left:
                if len(chunkdata)==0:
                    break


                # First gather up all the data remaining
                while len(chunkdata)<lzwminsize:
                    chunkdata+=chr(0)
                    
                # Then add the column initiation codes needed
                while len(chunkdata)<datasize:

                    chunkdata+=lineinit


                # Add the length of the data that we wrote to the stream
                chunkdata += chr(max(8-(datasize%8),1))+chr(lzwminsize-1)
                imdat += chunkdata
                chunkdata = ""


            extblock = bytearray()
            extblock += b"\x21\xF9\x04\x00"
            extblock += bytes([delaytime//256,delaytime%256])
            extblock += b"\x00"
            extblock += b"\x2C\x00\x00\x00\x00\x00\x00\x00\x00"
            if transparent:
                extblock += b"\x03\x00\x00\x00"
            else:
                extblock += b"\x00\x00\x00\x00"
            extblock += b"\x00\x00\x00\x00"

            if islastframe:
                extblock += b"\x00"
                
            imdat += extblock
                    

        # Now we need to write the image data itself...
        print("Writing the image")
        f.write(b"\x2C\x00\x00\x00\x00\x00\x00\x00\xFF")   # Image descriptor
        f.write(b"!\xF9\x04\x01\x00\x00\x00,\x00\x00\x00") # Graphics control extension
        f.write(b"\x00\x01\x00\x21\xFE\x09D\x0B\xCF\x00") # Application Extension
        f.write(b"\x0BNETSCAPE2.0\x03\x01")                  # Application Data
        f.write(b"\x00\x00\x00!")                           # Trailer
        f.write(imdat)