#!/usr/bin/env python3

import sys
import os
import binascii
from typing import Optional
from typing import List
from typing import BinaryIO
from typing import Iterable


# An exception thrown when the input is invalid.
class InvalidInputError(Exception): pass


# A namespace for all supported color table entries.
ColorTableEntry = int


# The maximum value of any ColorTableEntry.
COLOR_TABLE_ENTRY_MAX = 255


# All possible extension blocks.
EXTENSIONS = {
    0x01: 'Application Extension',
    0xff: 'Comment Extension',
}


# All length fields that are known and can be passed to read_length().
LENGTH_FIELDS = [
    'blocksize',
    'width',
    'height',
    'color-table-entries (global)',
    'color-resolution',
   'sort-flag',
    'color-table-entries (local)',
    'bg-color-index',
    'pixel-aspect-ratio',
    'delay-time',
    'transparent-color-index',
    'loopcount',
]


# Headers that contain specific fields with special handling.
HEADER_HANDLING = {
    # This field should not exist at this point in time.
    b'\x2c': ['subblocks'],

    # This field has a variable length and needs special handling.
    b'\x21\xf9': ['extension-intention', 'blocksize'],

    # This field has a variable length and needs special handling.
    b'\x21\xfe': ['application-identifier', 'authentication-code'],
}


# Headers we don't care about because they have no payload after them.
HEADER_EXEMPTIONS = [b'\x3b']


# Default values for some global headers.
DEFAULTS = {
    'gct-flag': True,
    'gce-flag': False,
    'lct-flag': False,
    'interlace-flag': False,
   'sorted-color-table-flag': True,
    'background-color-index': None,
    'pixel-aspect-ratio': 0,
}



# If `i` isn't an integer or doesn't fit inside two bytes, throw an
# InvalidInputError exception.
def check_integer(
        i: int
    ) -> None:
    if len(struct.pack('>H', i))!= 2:
        raise InvalidInputError("This field requires exactly 2 bytes.")


# Returns whether or not the given binary string represents a valid GIF
# frame.
def is_valid_frame(s: bytes) -> bool:
    return s[:6].lower() == b"gif87a" \
           or s[:6].lower() == b"gif89a"


# Writes the provided subblocks to the output stream as part of an extension block.
def write_subblocks(
        subblocks: list[bytes], f: BinaryIO
    ) -> None:
    for subblock in subblocks:
        if isinstance(subblock, str):
            subblock = subblock.encode('utf-8')

        check_integer(len(subblock))
        f.write(struct.pack(">B", len(subblock)))
        f.write(subblock)


# Reads a specified amount of bytes from the reader. Throws an
# InvalidInputError if there aren't enough bytes available.
def read_bytes(r: BinaryIO, n: int) -> bytes:
    bs = r.read(n)
    if len(bs) < n:
        raise InvalidInputError("Not enough bytes left")

    return bs


# Decodes a single length field by reading its length and then immediately
# reading those amount of bytes from the reader.
def decode_lengthfield(
        name: str,
        minval: int, maxval: int,
        default: Optional[int]=None,
        r: Optional[BinaryIO]=None,
        rs: Optional[Iterable[str]]=None
    ) -> int:
    if default is None:
        assert r is not None
        assert rs is not None
        default = DEFAULTS.get(rs[-1], -1)

    check_integer(minval)
    check_integer(maxval)
    check_integer(default)

    if r is not None:
        bs = read_bytes(r, 1)[0]
    else:
        bs = min(default, maxval)

    if not (minval <= bs <= maxval):
        raise InvalidInputError(f"{name} must be between "
                                f"{minval}..{maxval}")

    return bs


# Reads a length field and returns it. Does so using one of the pre-defined
# lengths or by decoding another length field beforehand.
def read_length(
        name: str,
        minval: int, maxval: int,
        default: Optional[int]=None,
        r: Optional[BinaryIO]=None,
        rs: Optional[list[str]]=None
    ) -> Optional[int]:
    # This field contains either the literal integer itself, or a reference to
    # another length field which will be decoded first.
    if name in LENGTH_FIELDS:
        val = decode_lengthfield(
                name,
                minval, maxval,
                default,
                r, rs)

        # We need to keep track of the current state when reading these kind of
        # fields, because otherwise we wouldn't know where to find referenced
        # length fields later on.
        if r is not None:
            rs = [*rs[:-1], name]
        elif rs is not None:
            rs = [name]

        return val

    # Interpret a previously encoded length field.
    if r is not None:
        for field in LENGTH_FIELDS:
            if field in rs and field!= name:
                return decode_lengthfield(
                        field,
                        0, COLOR_TABLE_ENTRY_MAX,
                        -1,
                        r)

    # Returning None instead of throwing an error here means it's up to the calling function
    # to handle this case appropriately.
    return None




# Parses the GIF image header starting from the beginning of `r`. Returns
# a tuple containing the following fields extracted form the header:
#    - width (in pixels)
#    - height (in pixels)
#    - gct-flag
#    - lct-flag
#    - interlace-flag
#    - sorted-color-table-flag
#    - background-color-index
#    - pixel-aspect-ratio
# If the GIF image's header is incomplete or incorrect, throws an
# InvalidInputError exception.
def parse_image_header(r: BinaryIO) -> dict[str, int | bool]:
    bs = read_bytes(r, 6).lower()

    if bs[:6]!= b"gif87a" and bs[:6]!= b"gif89a":
        raise InvalidInputError(f"Invalid GIF signature ({bs!r})")


    # Header flags are stored in the last three bits of each byte of the
    # logical screen descriptor.
    buf = []
    while True:
        byte = read_bytes(r, 1)[0]
        buf += [(byte >> bit & 1) << index
                for index, bit in enumerate([7, 6, 5])]

        if byte & 0b1000_0000 == 0:
            break

    gct_flag   = bool(buf[0]) # Global Color Table Flag
    lct_flag   = bool(buf[1]) # Local Color Table Flag
    interlace  = bool(buf[2]) # Interlace Flag
    sorted     = bool(buf[3]) # Sort Flag

    # Each field except the size of the Logical Screen Descriptor is represented
    # by a single byte. These fields always come right after the GIF signature
    # (6 bytes long), so the first thing we do here is skip over these bytes.
    read_bytes(r, 3)

    # Finally, we're able to start parsing our headers. First up is the size of
    # the Logical Screen Descriptor again, but now it's not used anymore since
    # all other fields are preceded by their respective length field.
    ls_size = read_length("Screen Descriptor Size",
                          4, 4,
                          default=-1,
                          r=r)
    if ls_size is not None:
        read_bytes(r, ls_size)

    # Width
    width = read_length("Width",
                        1, COLOR_TABLE_ENTRY_MAX,
                        default=-1,
                        r=r, rs=["width"])

    # Height
    height = read_length("Height",
                           1, COLOR_TABLE_ENTRY_MAX,
                           default=-1,
                           r=r, rs=["height"])



    # Image Descriptor flags are stored in the same way as the Logical Screen
    # Descriptor ones, except that the last two bits denote the local color
    # table flag.
    buf = []
    while True:
        byte = read_bytes(r, 1)[0]
        buf += [(byte >> bit & 1) << index
                for index, bit in enumerate([7, 6, 5])]

        if byte & 0b1000_0000 == 0:
            break

    lct_flag |= bool(buf[1]) # Local Color Table Flag
    inter