#!/usr/bin/env python3


import sys
import time
import struct
from typing import List, Tuple, BinaryIO

# Reads from /dev/urandom until it has enough values to return `num_values`.
# Returns a tuple of those values.
def read_entropy(
        num_values: int
    ) -> Tuple[int]:
    entropy : List[int] = []
    while len(entropy) < num_values:
        entropy += list(
                struct.unpack('>I',
                    open('/dev/urandom').read(4)))
        time.sleep(0.1) # sleep for 100 ms before calling urandom again
    del entropy[:num_values]
    return tuple(entropy)

def generate_rand_color() -> str:
    r, g, b = [hex(x)[-1].upper() for x in read_entropy(3)]
    return '#' + ''.join((r, g, b))
    

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated colors will be unique within 8 frames of eachother.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'\x47\x49\x46\x38\x39\x61'

    # width and height are 16 bits each
    header += struct.pack('>H', rng.randint(200, 500))
    header += struct.pack('>H', rng.randint(200, 500))

    global_colors = {} 

    # color table (RGB triplets)
    ctsize = rng.randint(1, 7) * 3


    for i in range(ctsize // 3):

        newColor = generate_rand_color()
        
        if not any(global_colors[c][1] == newColor for c in global_colors):
            if len(global_colors) >= 2**i - 1: break
            else:
                rgb = tuple(map(lambda x: int(x, base=16), ('%s%s%s' % newColor[-6:-4], '%s%s%s' % newColor[-4:-2], '%s%s%s' % newColor[-2:])))                
                
                global_colors[(i+1)**2] = (rgb, newColor)

                header += bytes([(i+1)**2])
                header += bytes(rgb)



    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + data)
        
if __name__ == '__main__':
    
    try:
        with open(sys.argv[1], 'wb+') as f:
            print("Generating random gif...")
            generate_random_gif(open('/dev/urandom'), f)
            print("Finished generating!")
    except IndexError:
        print("No output name specified.")