#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# between minwidth <= w < maxwidth and minheight <= h < maxheight (inclusive).
# The frame count is randomly generated from 1 to nframes inclusive.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, minwidth=64,
        maxwidth=512, minheight=64, maxheight=512, nframes=10):

    # construct headers
    header = b'GIF87a'
    width = rng.randint(minwidth,maxwidth-1)//2*2+2
    height = rng.randint(minheight,maxheight-1)//2*2+2
    gctsize = 2 << bool(rng.randint(0,1))
    globalcolortablelen = 2**(gctsize+1)-1
    localcolortablelen = 2**((bool(rng.randint(0,1)))<<2)
    fc = rng.randrange(nframes)+1
    extblocks = [
        (b'NETSCAPE2.0', b'\x03\x01\x00\x00\xff\x00'),
        (b'ANIMEXTS1.0', b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')]


    # write headers
    out.write(header)
    out.write(struct.pack('<HHB', width, height, gctsize))

    # apped colortables if any
    for i in range(globalcolortablelen//3 + bool(globalcolortablelen%3)):
        r,g,b = tuple([int(v*255) for v in rng.choice(range(256))][0] for _ in "rgb")
        out.write(struct.pack('BBB',r,g,b))
    if localcolortablelen > 0:
        for i in range(localcolortablelen // 3 + bool(localcolortablelen % 3)):
            r, g, b = tuple([int(v * 255)
                            for v in rng.choice(range(256))][0]for _ in "rgb")
            out.write(struct.pack('BBB', r, g, b))

    # build image descriptor block
    imgdescblock = []
    imgdescblock.append(0xf7)            # separator
    imgdescblock.extend([(width >> 8), (width & 0xff)])
    imgdescblock.extend([(height >> 8), (height & 0xff)])
    flags = ((fc & 0xf)|((((width&0x0f)<<2)&0xf0)))
    imgdescblock.append(flags)
    imgdescblock.extend([(flags & 0x3c)>>2, (flags & 0xc0)>>6])
    imgdescblock.extend([(height >> 8), (height & 0xff)])
    out.write(bytes(imgdescblock))

    # Write image data blocks
    # For each row, we take the first bit from each byte plus the previous one,
    # but with the topmost row having an additional 1 added to it, and the bottom
    # row having an additional 0 added. Then after every column, we rotate the bits.
    # This results in stripes of ones and zeros appearing as long as there is actual
    # data there, at least up until the point where we reach vertical repetition.
    for j in range(fc):
        for i in range(((width+7)/8)*height):
            byte = rng.getrandbits(8)%256
            if i>0 and i<((width+7)/8)*(height-1):
                prevbit = byte & 1!= 0
                xorbit = False
                for k in range(8)[::-1]:
                    newbit = rng.randint(0,3)==0
                    byte |= newbit<<(k+1)
                    xorbit ^= newbit
                byte ^= prevbit^xorbit
            else:
                byte ^= ((j == 0) or (j == (fc - 1))) << 7

            out.write(struct.pack('B',byte))

        # We have reached vertical repetition!
        if not (j == (fc - 1)):
            out.write(imgdescblock[:-1]+[0])
            out.seek(-9,1)           # move our position back 9 bytes



    # copy and extend extension blocks
    for name,data in extblocks:
        data += bytes([sum(map(ord,name))+len(data)])*(8-(len(data)%8))
        out.write(data)