#!/usr/bin/env python3
"""
Generate a random GIF from /dev/urandom.
"""

import os
import struct
import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import BinaryIO


def main():
    outfile = parse_args()
    
    with open("/dev/urandom", 'rb') as rng,\
         open(outfile, 'wb') as out:
        generate_random_gif(rng, out)


def parse_args() -> str:
    
    def checkpath(path):
        pathobj = Path(path)

        if not pathobj.parent.exists():
            raise ValueError('The parent directory of the output must exist!')
        
        return path
        
    parser = ArgumentParser("generate-random-gif")

    # Generate a default name if none is provided
    if len(sys.argv) == 1:
        return "/tmp/random.gif"

    # Allow passing an argument to specify output filepath.
    parser.add_argument(
        "-o", 
        "--output", 
        type=checkpath, 
        default="/tmp/random.gif", 
        dest="output",  
        help="set destination for generated image (default: %(default)s)"
    )
    args = parser.parse_args()

    return args.output


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Read the header and save it for later use.
    header = rng.read(6)
    
    # Verify that this looks like a valid LZW encoded GIF by checking magic numbers in the header
    # and ensure the color table size is zero.
    assert header[0] == 0x47 or "Invalid GIF header!"
    assert header[1] == 0x49 or "Invalid GIF header!"
    assert header[2] == 0x46 or "Invalid GIF header!"
    assert header[3] == 0x38 or "Invalid GIF header!"
    assert header[4] == 0x39 or "Invalid GIF header!"
    assert header[5] == 0x61 or "Invalid GIF header!"

    # The first three bytes after the signature are used to store the size of the rest of the data chunk
    datalen = int.from_bytes(
        rng.read(3), byteorder='big', signed=False)
    print(f"Datalen: {datalen} bytes.")

    # The next part should be the logical screen descriptor
    lsd = rng.read(7)
    #print(lsd.hex())

    width = int.from_bytes(
        lsd[:2], byteorder='big', signed=False)
    height = int.from_bytes(
        lsd[2:4], byteorder='big', signed=False)
    flags = int.from_bytes(
        lsd[4:], byteorder='big', signed=False)
    bgcoloridx = None
    if ((flags & 0x80)):
        bgcoloridx = lsd[6]
    print(f"Width={width}, Height={height}. Flags={flags}. BGCOLIDX={bgcoloridx}")

    # This should be the global color table
    gct = b''
    bitdepth = flags & 0x7
    palettebits = bitdepth * 3
    numcolors = 2 ** bitdepth
    for i in range(numcolors):
        gct += struct.pack(">BBB", *list(map(int, rng.read(palettebits))))
    
    # This should be the remainder of the image data
    imgdata = rng.read(datalen - len(gct)-7)
    
    # Write everything to the final file!
    out.write(header)
    out.write(struct.pack(">H", datalen+len(gct)))
    out.write(lsd)
    out.write(gct)
    out.write(imgdata)
    

if __name__=="__main__":
    main()