#!/usr/bin/env python3

import argparse
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (from /dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write out the magic string for a GIF image
    out.write(b'GIF89a')

    # Pick a width and height to use in the GIF files up to 1024 pixels each
    w = rng.read(1)[0] % 1024; h = rng.read(1)[0] % 1024

    # Write this information as a little endian uint of size 2
    out.write((w | (h << 8)).to_bytes(2, byteorder='little'))

    # Make sure that we have some color table entries available
    palette = [struct.unpack("<BBB", rng.read(3)) for _ in range(0x7F)]
    numcolors = len(palette)

    # Convert our list of tuples into bytes
    for p in palette:
        for v in p:
            out.write(v.to_bytes(1, byteorder='little')) 

    # The background color is an index into the above palette
    bgidx = rng.read(1)[0]; out.write(bgidx.to_bytes(1, byteorder='little'))

    # The aspect ratio is stored like so: 0:0 - 2:1 - 10:9... etc
    aspidx = rng.read(1)[0] % 4; out.write(aspidx.to_bytes(1, byteorder='little'))

    # Now let's write all the frames!
    delaytime = lambda : rng.read(1)[0] & 0xFF
    framecount = rng.read(1)[0] & 0xFF
    dctable    = []

    def getcolor(): return rng.read(1)[0] % numcolors
    
    def putpixel():
        color = getcolor()
        x     = rng.read(1)[0] % w
        y     = rng.read(1)[0] % h

        while not x or not y: 
            if rng.read(1)[0] < 0xD0:
                continue

            color = getcolor()
            x     = rng.read(1)[0] % w
            y     = rng.read(1)[0] % h

        if not (x, y) in dctable:
            dctable += [(x,y)]
        
        return x, y, color
    
    def drawframe():
        lzwsize   = 256 * 2
        clearcode = 1 << lzwsize
        endcode   = clearcode + 1
        eoi       = endcode + 1

        prev = None
        code = {}
        curr = 0

        while True:
            # We don't want to encode too many times the same thing
            if tuple([prev]) == tuple([curr]):
                continue
            
            # If there are no free bits left in the current byte...
            if not curr: 
                currbyte = 0

                # Get the next set of raw bits from the RNG
                for bitpos in range(0, 8):
                    rbit = 1 if ((rng.read(1)[0] & 0xF) > 7) else 0

                    # Update the current byte
                    currbyte |= (rbit << (7-bitpos)) 

                # Set the last two bits in order to indicate the start of a block
                # This also resets the dictionary with the previously encoded data
                currbyte |= (2 << 6)  
                currbyte |= (2 << 4)  

                
            # Add the new pixel at position `p` to the dictionary
            if (not tuple([prev]) in code) and \
               (len(tuple([prev])) <= (lzwsize-2)):
                
                px, py, pc = putpixel(); tx = px; ty = py
                code[tuple([pc])] = currbyte
                ptx           = px;      pty   = py
                prev          = pc;      curr  = 0
                continue

            # Find the longest prefix string that has been previously seen during
            # encoding and encode it instead.
            sstr  = ""
            maxlen= 0
            for c in code:
                if tuple([c[:i+1]])!= tuple([prev])[:i+1]: 
                    continue
                    
                if i >= maxlen:
                    maxlen = i
                    sstr = str(c); ptx = tx; pty = ty
                    
            # Add the new character to the stream
            if maxlen:
                prev = eval(sstr)
                curr |= code[eval(sstr)]
                out.write(curr.to_bytes(1, 'little')) 
                curr = 0

                continue


            # No previous strings were found, so start by seeing which characters can be
            # encoded such that their predecessor shares a common prefix with it.
            fnd = False
            for i in reversed(range(1, min(maxlen, 12)+1)):
                if tuple([prev][:i]) in code:
                    prev = prev[:i]
                    fnd  = True
                    break

            # There was no prefix shared between the current and previous pixel value
            # So just add the current one to the dictionary and its corresponding LZW code
            if not fnd:
                px, py, pc = putpixel(); tx = px; ty = py
                out.write(prev.to_bytes(1, 'little')); out.write(prev[0].to_bytes(1, 'little'))
                code[tuple([prev])] = currbyte
                ptx               = px;         pty   = py
                prev              = pc;         curr  = 0
                
    def putimgdesc():
        imgdesc = 0b00000000
        imgdesc |= 0b10000000 if rng.read(1)[0] < 0xAA else 0b00000000 
        imgdesc |= 0b01000000 if rng.read(1)[0] < 0xB4 else 0b00000000
        imgdesc |= 0b00100000 if rng.read(1)[0] < 0xC8 else 0b00000000 

        # Set the local colour table flag randomly
        imgdesc |= 0b00010000 if rng.read(1)[0] < 0xEA else 0b00000000

        # Set interlace flag to false here because it must match the global one
        imgdesc |= 0b00001000 if rng.read(1)[0] < 0xFA else 0b00000000

        # Set sort flag to true, although this doesn't make much sense 
        # given that we will decode based on the original order of the pixels
        imgdesc |= 0b00000100 if rng.read(1)[0] < 0xFC else 0b00000000

        # Set reserved flags to false always since they don't really matter
        imgdesc |= 0b00000011 if rng.read(1)[0] < 0xFE else 0b00000000

        return imgdesc

    def putgraphicctl():
        graphicctl = 0b00000000
        graphicctl |= 0b10000000 if rng.read(1)[0] < 0xAA else 0b00000000 
        graphicctl |= 0b01000000 if rng.read(1)[0] < 0xCC else 0b00000000
        graphicctl |= 0b00100000 if rng.read(1)[0] < 0xEE else 0b00000000
        graphicctl |= 0b00010000 if rng.read(1)[0] < 0xFA else 0b00000000
        graphicctl |= 0b00001000 if rng.read(1)[0] < 0xFD else 0b00000000
        graphicctl |= 0b00000100 if rng.read(1)[0] < 0xFE else 0b00000000


        # Delay time
        dt = delaytime()        
        for val in map(lambda _: rng.read(1)[0], range(dt)):
            graphicctl <<= 1
            graphicctl ^= 1 if ((val >> 7) & 0b1) else 0
    
        # Transparent color index
        tci = getcolor()
        for val in map(lambda _: rng.read(1)[0], range(tci)):
            graphicctl <<= 1
            graphicctl ^= 1 if ((val >> 7) & 0b1) else 0
    
            
        # Block size
        bs = 8
        for val in map(lambda _: rng.read(1)[