#!/usr/bin/env python3
'''Generate some random GIFs to test the decoder.'''
import random
import struct
from typing import BinaryIO, List


# A byte-string representing the GIF signature and version string.
signatureandversion = "GIF89a".encode()

# Given an integer value within the given maximum, return it as two single
# bytes.
def twobytevalue(v:int, maxval:int)->bytes:
    assert v<maxval
    return struct.pack("<BB", v%256, v//256)

# Generate a random uint16.
def randuint16()->int:
    return random.randint(0, 2**16-1)

# Generate a random uint8.
def randuint8()->int:
    return random.randint(0, 2**8-1)

# Return a random number of entries from the given list. This is done by first
# selecting a level of depth based on the size of the list, then randomly choosing
# values at that level until we have enough..
def randlistentries(l:List[int])->List[int]:
    depth = len(l)
    if depth == 0:
        raise ValueError("Empty list provided when trying to choose entries")
    elif depth <= 32:
        numentries = random.randint(0, depth)
    else:
        numentries = random.randint(0, min(depth, 32))

    chosenindices = set()
    while len(chosenindices)!= numentries:
        index = abs(hash(str(random.getrandbits(64)))) % depth
        chosenindices.add(index)

    return sorted([ l[x] for x in chosenindices ])

# Turn a list of integers into their two-byte representations.
def twobytesvalues(l:List[int])->bytes:
    return bytes(twobytevalue(x, 2**16-1) for x in l)


# Generates a random GIF file into `out`.
def generate_random_gif(
        out:        BinaryIO,
        width:      int    = 1024,
        height:     int    = 768,
        colordepth: int    = 8,
        bgcolor:    int    = None,
        numframes:  int    = 10,
    ) -> None:

    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    # Randomly determine how many frames are going to be generated.
    framecount = randuint16() % numframes + 1

    # Create all of the frames and collect them together so they can be written last.
    frames = []
    for framenum in range(framecount):
        f = Frame(width, height, colordepth, bgcolor)
        frames.append(f)

    # The GIF header
    out.write(signatureandversion)

    # Screen descriptor
    out.write(twobytevalue(width, 2**16-1))
    out.write(twobytevalue(height, 2**16-1))
    out.write(struct.pack('<BBB',
            0xF0 | colordepth,             # Color depth (high bit indicates global color table)
            0,                             # Color resolution
            0                              # Sort flag
        ))
        
    # Global color table
    globalcolortable = [struct.pack('>BBBB',
            *(c&(2**8-1)//(256**x)) for c,x in zip(bgcolor,range(3,-1,-1)))
        for _ in range(2**(colordepth))]
    out.write(b"".join(globalcolortable))

    # Loop information
    loopcount = randuint16() % (numframes+1)+1
    out.write(twobytevalue(loopcount, 2**16-1))
    out.write(twobytevalue(randuint16(), 2**16-1))

    # Images!
    for f in frames:
        f.write(out)



class Frame():
    '''A class which represents a single frame within a GIF file.'''


    def __init__(self, width:int, height:int, colordepth:int, bgcolor:int):
        self._width       = width
        self._height      = height
        self._colordepth  = colordepth
        self._bgcolor     = bgcolor
        self._imagedata   = bytearray(width*height)


    def write(self, out:BinaryIO):

        # Remember where this image begins before writing any pixel data.
        startofimage = out.tell()

        # First we'll get a full list of pixels (including transparent ones)
        # that should appear in the final image.
        pixellength = self._width * (self._height - 1) + self._height // 2
        rawpixels = [ self._bgcolor for _ in range(pixellength)]
        rawpixels[::2],rawpixels[1::2] = list(zip(*[
                ( random.randint(0,self._colordepth-1),
                  random.randint(0,255),
                  random.randint(0,255),
                  random.randint(0,255),
                ) for _ in range(len(rawpixels)//2)]))

        # Next we'll select exactly what indices should actually be non-transparent.
        maskedpixels = randlistentries(rawpixels)

        # And finally mask away those pixels that aren't supposed to be there.
        for p in range(pixellength):
            if rawpixels[p]!= self._bgcolor and p not in maskedpixels:
                rawpixels[p] &= ~(256**self._colordepth-1)

        
        # Finally construct the actual image data section.
        interlaceoffset = 0
        for y in range(self._height):

            # If it's in the odd row, we need to change the offset value.
            if y&1:
                interlaceoffset -= 1

            for x in range(self._width):

                _skipnext = False

                # Determine the position in the image (for LZW compression)
                posinimage = (((interlaceoffset)+(y//2)*self._width+(x//2))*16
                        +((y&1)<<2)|((x&1)<<1))

                # Get the current pixel's value and its location inside the output stream.
                currentvalue = rawpixels[(y*self._width)+x] 
                curloc = out.tell() 

                # Figure out whether we're going to do runlength encoding...
                rlecount = 1
                if posinimage >= 2:

                    trypos   = posinimage-(rlecount<<1)
                    prevvalue = rawpixels[trypos]

                    while (trypos > 0
                            and rawpixels[trypos] == prevvalue 
                            and rlecount < 63):

                        trypos -= 2
                        rlecount += 2

                #...or just write out the actual pixel data.
                if rlecount < 2:
                    out.write(struct.pack(">H", currentvalue))

                else: 
                    out.write(struct.pack(">HB",
                            0xFC00 | (rlecount-2),
                            currentvalue
                        ))


                # Check to see if the RLE code was part of a longer run that needed to
                # be skipped over in order to make sure the stream was valid.
                leftoveroffset = (rlecount-2) << 1
                while leftoveroffset > 0:
                    
                    if leftoveroffset >= 2:
                        
                        trypos = posinimage-leftoveroffset
                        nextvalue = rawpixels[trypos]

                        while (trypos > 0
                                and nextvalue == prevvalue 
                                and leftoveroffset < 63):

                            trypos -= 2
                            leftoveroffset += 2


                    out.seek(-2,1)                
                    out.write(struct.pack(">H", nextvalue))

                    posinimage = trypos
                    prevvalue = nextvalue
                    leftoveroffset -= 2



        # Once everything has been written, we need to fill in some extra animation
        # headers since we don't know yet how long each one will take.
        endofimage = out.tell()
        out.seek(startofimage + 10)
        out.write(b'\x21\xFF')                                   # Graphic Control Extension
        out.write(b'\x0B\x4E\x45\x54\x53\x43\x41\x50\x45\x32\x2E\x30') # NETSCAPE2.0
        out.write(b'\x03')                                        # Version
        out.write(b'\x01')                                        # Loop count
        out.write(b'\x00\x00')                                   # Length of subblock
        out.write(b'\x3B')                                        # End of block
        out.write(b'\x2C')                                        # Image Descriptor
        out.write(b'\x00\x00')                                   # Left Position 
        out.write(b'\x00\x00