#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generate a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# between `minwidth` <= w < `maxwidth` and `minheight` <= h < `maxheight` 
# (inclusive). The frame count is randomly generated from 1 to `nframes` 
# (inclusive), and each pixel's value is chosen randomly from an array containing 
# all possible values. The color bits per pixel are randomly chosen from `[1,2,4]`.
def generate_random_gif(
        rng: BinaryIO, 
        out: BinaryIO, 
        minwidth=64,
        maxwidth=512, 
        minheight=64, 
        maxheight=512, 
        nframes=10) -> None:
    
    # Construct headers
    header = b'GIF89a'
    width = rng.randint(minwidth, maxwidth - 1) // 2 * 2 + 2
    height = rng.randint(minheight, maxheight - 1) // 2 * 2 + 2
    depth = rng.randint(1, 8)
    color_bit_depth = rng.randint(1, 4)
    globalcolor_table_len = 2 ** (depth + 1) - 1
    background_index = rng.randint(0, globalcolor_table_len)
    aspect_ratio = rng.randint(0, 8)

    # Write headers
    out.write(header) 
    out.write(struct.pack('<HHB', 
            width, height, depth)) 

    # Append any color tables
    color_table = []
    for i in range(globalcolor_table_len // 3 + bool(globalcolor_table_len % 3)):
        r, g, b = next(rng)[:3]
        out.write(struct.pack('<BBB', r, g, b))
        color_table.append((r, g, b))
    if localcolor_table_len > 0:
        raise RuntimeError("not implemented yet");
 
    # Write frames
    for f in range(nframes):

        # Construct image descriptor
        out.write(bytes([
            0x2C,              # Image Descriptor Block
            0x00,     
            (width >> 8) & 0xFF,         # Width MSByte
            width & 0xFF,     # Width LSB
            (height >> 8) & 0XFF,        # Height MSByte
            height & 0xFF,    # Height LSB
            color_bit_depth, 
            0]))      
        
        # Write pixels to output file
        for i in range(width * height):

            # Get pixel value
            r, g, b = color_table[rng.randint(0, len(color_table))]
            
            # Write pixel to output stream
            out.write(bytes([(r >> 5 << 5) | (g >> 5 << 2) | (b >> 6)])) 

        out.write(bytes([0]))          # Zero-length trailer

    # Finish file with netscape application extension
    out.write(bytes([
        0x21,                    # Extension Introducer
        0xff,                    # AppExt Label
        0x0b,                    # Size of following data
        0x4e, 0x65, 0x74, 0x53,  # NetScape title string
        0x63, 0x72, 0x65, 0x65,  # End of NetScape label
        0x6e, 0x45, 0x78, 0x74,  # Application ID
        0x72, 0x61, 0x2d, 0x31, 
        0x2e, 0x30,                # End of appid
        0x03, 0x01, 0x00, 0x00,  # Subblock length
        0x00,                     # Loop subblock
        0xff,                     # Trailinng zero
        0x0b,                    # Size of following data
        0x4c, 0x69, 0x6E, 0x6B,  # Link string
        0x2D, 0x65, 0x76, 0x65, 
        0x6e, 0x74, 0x73, 0x00]))

    return 
