#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generate a random GIF given an input stream of bytes to use for randomness and output stream to write to.
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO,
        frames=10,
        delay=25,
        transparentcolorindex=-1,
        disposalmethod=2):

    # Create the header
    header = b"GIF89a"

    # Get the length of the image from the RNG
    width  = (int.from_bytes(rng.read(1), 'big') >> 4) + 1
    height = ((int.from_bytes(rng.read(1), 'big') & 0xf) << 8) | int.from_bytes(rng.read(1), 'big')
    print("Generating {}x{} GIF".format(width,height))

    # The only real difference between the first frame and subsequent frames is that there's no background colour at the start so the previous frame should be used.
    # So let's draw our first frame separately.
    # The initial value of previousframe will be None until after the first iteration of the loop below.
    previousframe = None

    # Write it all together
    out.write(header)
    out.write(struct.pack('>H', width))
    out.write(struct.pack('>H', height))

    # LZW minimum size == 2
    mincodesize = 2
    out.write(mincodesize.to_bytes(1,'little'))
    
    # TODO(...): Move this stuff into a function or class since it'll likely get reused multiple times in different places.

    # Image flags are packed 1 bit each with as follows:
    #   1 - Color table sort flag
    #   2-4 - GCT size (LZW minimum size)
    #   5   - Local colors sorted flag
    #   6   - Interlace flag (NOT USED)
    #   7   - Sort flag (NOT USED)
    #   8   - Reserved for future use

    
    # TODO(...): Figure out how to have variable flags such that they can be changed between frames if necessary.

    # Set interlace off, color table sorted on, local colors sorted off,
    # and everything else off that we don't care about.
    gctsortflag = 1
    lctsortedflag = 0
    interlaced = False
    sortflag = 0
    reserved = 0
    imgflags = (gctsortflag << 7) + (lctsortedflag << 6) \
                + (interlaced << 5)     + (sortflag << 4) \
                + (reserved << 3)       + mincodesize
                

    # Finally, let's go back to drawing the actual gif
    for i in range(frames+1):

        # If this isn't the first frame then use the previously generated frame as the basis for what this one draws on top of.
        if not i == 0:
            out.write(previousframe[1])
            out.write(previousframe[2])

            # Get the last two bytes written (the end of the frame). That's where we will insert the LZW codesize change.
            pos = out.tell() - 2;

            # Seek to the right place and patch up the data 
            out.seek(pos);
            out.write(imgflags.to_bytes(1, 'little'))

        # Now put together the palette and screen descriptor. For simplicity's sake we won't bother creating a sub-block for the palette other than literally just writing its contents down here.

        # Do something similar for the image descriptor. Just hardcoded values for now.
        desc = bytearray([
            0x2C,
            *((width).to_bytes(2,"big")),
            *((height).to_bytes(2,"big")),
            *(delay.to_bytes(2,"little")),
            *((transparentcolorindex).to_bytes(1,"little")),
            *disposalmethod.to_bytes(1,"little"),
            0])
        
        # Write these two sections along with the local color table section (for ease of inserting them when we know where they belong).
        out.write(desc)
        out.write(bytearray([len(palette)]));
        out.write(palette)

        blocktype = b"\xF9"
        blocklength = len(lzwcompressedimage)+1
        userinputbyte = 0
        out.write(_make_subblock(blocktype, blocklength, userinputbyte, lzwcompressedimage))

        # Save this off as the current frame so we can use it to make future frames based on it.
        previousframe = [out.tell(), desc, _make_subblock]

    pass