#!/usr/bin/python3

import os
import random
import struct
from collections import namedtuple


## Color utility functions ##

Color = namedtuple('Color',['r','g','b'])

def rgb2int(r,g,b): return r<<(16)+g+(256<<8)+b

def lerp(a,b,t):
    """Linearly interpolate between two values."""
    return a + t * (b-a)

def clamp(v,low,high):
    """Clamp v to [low,high]."""
    assert low <= high
    return min(max(low,v),high)



## Generating GIFs ##

VERSION = "89a"
HEADER_SIZE = 6

class ImageDataEncoder(object):

    def __init__(self,fp,*args,**kwargs):
        self._fp = fp
        super().__init__(*args,**kwargs)

    def _write_bytes(self,bs):
        nwritten = self._fp.write(bs)
        if len(bs)!= nwritten: raise Exception()

    def write_image_header(self,width,height,bpp):
        if bpp == 8:
            self._write_bytes(b'\x2c') # Image separator
        elif bpp == 4:
            self._write_bytes(b'\x22') # Image separator
        else:
            raise ValueError("'bpp' must equal either 8 or 4")
            
        self._write_bytes(b'\x00\x00\x00\x00\x00') # Left position - Top Position - Width - Height
    
        global VERSION
        self._write_bytes(VERSION.encode())           # Version string

        self._write_bytes(b'\xff')                     # LZW minimum size code
        self._write_bytes(struct.pack(">H",bpp))       # LZW maximum size code

    def write_palette(self,palette):
        self._write_bytes(b"".join(palette))

    def write_data(self,data):
        self._write_bytes(b"".join(data))


class GIFFrameWriter(ImageDataEncoder):

    def __init__(self,fp,*args,**kwargs):
        self._fp = fp
        self._lzwminsize = kwargs['lzwminsize']

        # Construct the global colour table from our list of colors.
        self._globaltable = bytearray()
        for c in kwargs['colors']:
            self._globaltable += struct.pack('>BBB',c[0],c[1],c[2])

        # Generate the palette which maps indices back to actual colours.
        self._localtable = {}
        for i,(c,_) in enumerate(zip(*kwargs['colors'])):
            self._localtable[i] = c

        super().__init__(fp,*args,**kwargs)


    def _write_graphic_control(self):
        disposalmethod = 2
        userinputflag  = 0
        transparentidx = 0

        # Add a graphic control block
        self._write_bytes('\x21\xf9\x04')          # Extension introducer and label
        self._write_bytes(struct.pack('BB',
            0xf0|disposalmethod,userinputflag)) # Rest of block

        # Finally add the value which indicates transparency.
        self._write_bytes(struct.pack('>H',transparentidx))


    def write_frame(self,pixels,delay):
        
        import io
        import PIL.Image        
        imgio = io.BytesIO()
        im = PIL.Image.new('RGB',(len(pixels[0]),len(pixels)))
        for y in range(im.height):
            im.putdata([tuple(p) for p in pixels[:,y]])
        im.save(imgio,'gif')

        # Generate an encoded LZW stream
        encoder = Encoder(imgio.getbuffer(),
            maxbits=12,
            gmemlim=-1,
            glevel=6,
            clearcode=1 << self._lzwminsize,
            endcode=1 << self._lzwminsize + 1,
            tabprefix=''.join(['{:02X}'.format(_) for _ in range(3)]))

        encstream = encoder.encode().tobytes()

        # Remove the first CLEAR code and replace it by the local gct.
        encstream = encstream[1:]
        encstream[encstream == 255] = len(self._localtable)

        # Add the local gct, then the remainder of the frame.
        encstream = encstream[:1]+[self._localtable[_] for _ in encstream[1:]]

        # At this point, we have all the elements necessary to construct our
        # graphics control block.
        self._write_graphics_control()
        self._write_bytes(struct.pack('<H',delay))            # Frame delay
        self._write_bytes(b'\x00')                             # Transparent index
        self._write_bytes(struct.pack('>H',len(encstream)+1)) # Block size
        self._write_bytes(struct.pack('B',len(self._localtable)+1))    # Local colour table entries
        self._write_bytes(encstream)                            # Compressed data (now just raw data)
    
