#!/usr/bin/env python3

"""
Generate a random GIF from /dev/urandom.
"""

import os
import struct
import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import BinaryIO


def parse_args() -> str:
    parser = ArgumentParser("generate-random-gif")

    # Generate a default name if none is provided
    if len(sys.argv) == 1:
        return "random.gif"

    # Allow passing an argument to specify output filename.
    parser.add_argument(
        "-o", "--output", type=Path, 
        help=(
            "set destination for generated image " 
            "(default: %(default)s)"
        )
    )
    ns = parser.parse_args()

    if not ns.output:
        return f"{os.getcwd()}/random.gif"
    
    return ns.output


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = bytes([
        0x47, 
        0x49, 
        0x46, 
        0x38, 
        0x39, 
        0x61,
    ]) 
    
    width = 256
    height = 256
    color_depth = 256 # max of RGB values
    global_color_table = True
    local_color_table = False
    
    screen_width = width.to_bytes(2, 'little')
    screen_height = height.to_bytes(2, 'little')
    field_size = b'\xFF\x0B' # Indicates no compression and sorting required.
    bg_color = b'\x00\x00\x00'
    pixel_aspect = b'\xFF'
    
    flags = b''
    if global_color_table:
        flags += b'\x87'
        color_table_size = ((color_depth * 3) + 1).to_bytes(1, 'little')
        flags += color_table_size
    else:
        flags += b'\x00'
        
    if local_color_table:
        flags += b'\x0C'
        color_table_size = (((color_depth - 1) * 3) + 1).to_bytes(1, 'little')
        flags += color_table_size
    else:
        flags += b'\x00'
    
    gct_sorted = sorted(range(int.from_bytes(bg_color, 'little')), reverse=True)
    
    palette = []
    for i in range(gct_sorted[0], gct_sorted[-1] + 1):
        rgb = tuple(i % 256 for _ in range(3))
        palette.extend(rgb)
    
    pallete_entry = lambda c: b'%c%c%c' % (c[0], c[1], c[2])
    
    gct_entries = [pallete_entry(palette[(j*3):((j+1)*3)]) for j in gct_sorted[:-1]]
    gct_entries.append(b'')
    gct_body = b''.join(gct_entries)
    gct_bodylen = len(gct_body).to_bytes(2, 'little')
    
    # Image descriptor.
    image_descriptor = \
        bg_color + \
        screen_width + \
        screen_height + \
        field_size + \
        flags + \
        gct_bodylen
    idlen = len(image_descriptor)
    
    # Image data.
    bodylen = width * height * 3
    frame_control_block = b'\x2C' + b'\xFF'*(idlen-2) + bodylen.to_bytes(2, 'little')
    lzwmincodesize = b'\x08'
    pixels = [b'\xFF'] * bodylen
    
    start_of_image = b'\x21\xF9\x04\x00\x00\x00\x00\x00' + \
                     b'\x2C\x00\x00\x00\x00\x01\x00\x01\x00\x00' + \
                     b'\x02\x02\x4C\x01\x00\x3B'

    
    # Header fields.
    header = [
        header, #  7 bytes: Header signature and version.
        screen_width, # 2 bytes: Screen width in pixels.
        screen_height, # 2 bytes: Screen heigth in pixels.
        field_size, # 2 bytes: Size of each table entry in bits.
        flags, # 1 bit: Global Color Table flag.
        flags, # 3 bits: Reserved
        flags, # 3 bits: Sorting method.
        flags, # 1 bit: Local Color Table flag.
        flags, # 2 bit: Interlace flag.
        flags, # 3 bits: Reserved
        
        gct_bodylen, # 2 bytes: Length of GCT.
        gct_body, # GCT # Bytes: Global Color Table entries.
        idlen, # 1 byte: LZW minimum code size.
        frame_control_block, # Frame control block length & contents.
        # Pixels follow...
        # Start of image chunk.
        start_of_image,
    ]
    
    for chunk in header:
        out.write(chunk)
    
    # Write pixels to final stream as they are generated.
    #for p in pixels:
    #    out.write(p)
    



