#!/usr/bin/python3


def getn(f, n):
    """
    Read exactly N bytes from a binary stream and return them as an integer

    Parameters
    ----------
    f : BinaryIO
        The source to read from
    n : int
        How many bytes should be read

    Returns
    -------
    result : int
        An integer containing all those bytes in big endian format
    """
    assert 0 < n <= 4
    x = struct.unpack('>' + 'B'*n, f.read(n))[0]
    if x == (1 << (8*n - 1)): # 2's complement negative
        x -= 1 << (8*n-1)      # interpretation for unsigned values
    return x


def u8(f):
    """
    Get one unsigned byte value from a binary stream

    Parameters
    ----------
    f : BinaryIO
        The source to read from

    Returns
    -------
    result : int
        A positive integer representing the next byte of the input
    """
    return getn(f, 1)


def s8(f):
    """
    Get one signed byte value from a binary stream

    Parameters
    ----------
    f : BinaryIO
        The source to read from

    Returns
    -------
    result : int
        An integer representing the next byte of the input
    """
    return u8(f) - 128


def u16(f):
    """
    Get two unsigned bytes interpreted as a 16 bit value from a binary stream

    Parameters
    ----------
    f : BinaryIO
        The source to read from

    Returns
    -------
    result : int
        An integer representing the next two bytes of the input
    """
    return getn(f, 2)


def u32(f):
    """
    Get four unsigned bytes interpreted as a 32 bit value from a binary stream

    Parameters
    ----------
    f : BinaryIO
        The source to read from

    Returns
    -------
    result : int
        An integer representing the next four bytes of the input
    """
    return getn(f, 4)


def encode_u16(x):
    """
    Encode an integer as two consecutive bytes according to the encoding defined by the specification

    Parameters
    ----------
    x : int
        An integer with absolute value less than 2^16

    Returns
    -------
    result : list[int]
        Two integers that represent the encoding of this number
    """
    assert abs(x) < (1<<16)
    lower = x & 0xff
    higher = (x >> 8) & 0xff
    return [lower, higher]


def encode_u32(x):
    """
    Encode an integer as four consecutive bytes according to the encoding defined by the specification

    Parameters
    ----------
    x : int
        An integer with absolute value less than 2^32

    Returns
    -------
    result : list[int]
        Four integers that represent the encoding of this number
    """
    assert abs(x) < (1<<32)
    lower = x & 0xff
    x >>=  8
    higher = x & 0xff
    x >>= 8
    third = x & 0xff
    fourth = (x >> 8) & 0xff
    return [lower, higher, third, fourth]


class Encoder():
    """Utility class used when encoding a GIF"""
    def __init__(self, output):
        self._output = output

        # GIF encodes colors as RGB triplets, so each pixel has three values
        self._color_buffer = []

    # Write a single byte to the file
    def write_byte(self, x):
        # TODO: Make sure we never write more than 255 bytes at once?
        self._output.write(struct.pack("=B", x))

    # Write a series of bytes to the file
    def write_data(self, data):
        self._output.write(bytes(data))

    # Add a color entry to the buffer. When writing a block of pixels,
    # these entries will be written first.
    def add_color_entry(self, r, g, b):
        self._color_buffer += [r, g, b]

    # Write any buffered colors to the file
    def flush_colors(self):
        while len(self._color_buffer) > 0:
            c = self._color_buffer[:3]

            # We are only allowed to have up to 256 palette entries
            # If there aren't enough colors, pad the rest with zeros
            available_colors = min(len(c)/3, 255)
            remaining_colors = max(len(c)/3 - 255, 0)
            self.write_byte(available_colors)
            self.write_data(c[:3 * available_colors])
            self.write_data([0]*3*remaining_colors)

            del self._color_buffer[:3*available_colors]


    # Set size information about the image in the file
    def set_image_info(self, width, height):
        self.add_color_entry(*encode_u16(width)[::-1])
        self.flush_colors()

        self.add_color_entry(*(height & 0xffff).to_bytes(2, "little"))
        self.add_color_entry((height >> 16) & 0xff)
        self.flush_colors()



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    encoder = Encoder(out)
    encoder.set_image_info(getn(rng, 2), getn(rng, 2))