#!/usr/bin/env python3
"""
Generate a random GIF from /dev/urandom.
"""

import os
from argparse import ArgumentParser
from typing import BinaryIO


def parse_args():
    parser = ArgumentParser("generate-random-gif")

    # Generate a default name if none is provided
    if len(sys.argv) == 1:
        return ("random.gif", "/dev/urandom")

    # Allow passing an argument to specify output filename and source of randomness.
    parser.add_argument("-o", "--output", type=str,
                        help="set destination for generated image (default: %(default)s)", required=True)
    parser.add_argument("--source", type=str,
                        help="set source of randomness (default: %(default)s)", default="/dev/urandom")
    args = parser.parse_args()

    if not os.path.isfile(args.source):
        raise ValueError(
            f'Source file "{args.source}" does not exist or is not accessible.'
        )

    if not args.output:
        raise ValueError('No output specified.')
    
    return (args.output, args.source)


# Generates a random GIF file into `out` using the
# random number generator `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    size = os.stat(rng.name).st_size
    header = b'\x47\x49\x46\x38\x39\x61'
    headersize = len(header)
    framesizelength = headersize + 2
    framesize = framesizelength + size
    # Write the entire header so that we can overwrite it later on with our own
    # frame size information. This allows us to write the appropriate amount of
    # chunk data before writing the actual data at the end.
    out.write(b'\x47\x49\x46\x38\x39\x61')
    # The first two bytes of the body are the data length. We need this value to be correct in order
    # for the browser to correctly interpret the stream as a valid GIF and render it.
    out.write((framesize).to_bytes(length=2, byteorder='big'))
    # Write the rest of the initial header
    out.write(b'\x87\x00\x00\x00\xff\x00,\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;')
    # Now just copy over everything from our original reader to the final writer until EOF.
    while True:
        buff = rng.read(512)
        if not buff: break
        out.write(buff)


