#!/usr/bin/env python3

import math
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:       BinaryIO=None,
        width:     int    = 1024,
        height:    int    = 768,
        colordepth:int    = 8,
        bgcolor:   int    = None,
        outfile:   str    = 'output.gif',
    ) -> None:

    def rgb_to_luma(r:int,g:int,b:int)->float:
        return.299 * r +.587 * g +.114 * b

    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolors = [(rgb_to_luma(*[c/255.0 for c in struct.unpack('>BBB',bytes([d]))]), d) for d in range(256)]
        bgcolor = sorted(bgcolors)[-1][1]^((width*height)%(255**3))

    blackcode = bytes([bgcolor & (255**2)])
    whitecode = bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])
    
    with open(outfile,'wb') as f:

        # The GIF header
        version = b"89a"
        screendescriptor = struct.pack('>HHBBB',
                width,
                height,
                0xF0 | colordepth,
                0,
                0,
        )
        
        # Write the header and descriptor block
        print("Writing the header")
        f.write(b"GIF"+version+screendescriptor)

        # Build the color lookup table
        print("Building the color map...")
        codesize = calc_codesize(len(set([bytes([i]) for i in range(256)])))
        mincodesize = max(min(calc_codesize(len(set([bytes([i]) for i in range(256)]))<<(i-1)), 12) for i in range(9,1,-1))
        colortable = []
        luma = {}
        lumasorted = set()
        for i in range(2**mincodesize):

            # Generate a pseudo-random color
            x = (((i//(2**mincodesize))*width)+(i%2**(mincodesize-1)))*(bgcolor**(1/(256.0))) % (256**3)
            
            # Convert it to RGB
            R = int(((x>>16)&255)**1.5)
            G = int(((x>>8)&255)**1.5)
            B = int((x&255)**1.5)
            
            # Get some information about its brightness
            L = round(rgb_to_luma(R,G,B)*255)
            if L in luma:
                luma[L].append(R)
                luma[L].append(G)
                luma[L].append(B)
            else:
                luma[L]=[R,G,B]
                
            if L not in lumasorted:
                lumasorted.add(L)
                
        while len(lumasorted)<2**mincodesize:
            lumasorted.add(max(lumasorted)+1)
            
        for i in range((2**mincodesize)-1):
            x = lumasorted.__next__()
            colortable += [struct.pack('>BBB',
                    int(((x>>16)&255)**1.5),
                    int(((x>>8)&255)**1.5),
                    int((x&255)**1.5),
                ),
            ]
        colortable += [blackcode]
        colortable += [whitecode]
        
        print("Found {0} colors of size {1}".format(len(colortable),mincodesize))
        print("The background is close enough to one of them ({0})".format(hex(bgcolor)))
        print("Writing the palette...")
        f.write(b"".join(colortable))

        # Now we need to write the image data itself...
        print("Writing the image")
        f.write(b"\x2C\x00\x00\x00\x00\x00\x00\x00\xFF")   # Image descriptor
        f.write(b"!\xF9\x04\x01\x00\x00\x00,\x00\x00\x00") # Graphics control extension
        f.write(b"\x00\x01\x00\x21\xFE\x09D\x0B\xCF\x00") # Application Extension
        f.write(b"\x0BNETSCAPE2.0\x03\x01")                  # Application Data
        f.write(b"\x00\x00\x00!")                           # Trailer
        print("\nDone.")