#!/usr/bin/env python3

import os
import sys
import select
import time
import string
import struct
from typing import BinaryIO


# Creates random data in `length` bytes and writes it to `out`.
# Uses the random number generator `rng` (default: /dev/urandom).
def create_rnd_data(
        length: int, out: BinaryIO = None,
        rng: BinaryIO = open('/dev/urandom', 'rb')) -> bytes:

    if not out:
        return rng.read(length)

    chunksize = 1024 * 64 # arbitrary

    for i in range((length // chunksize) + 1):
        out.write(
                rng.read(min(chunksize,
                        length - i*chunksize)))

        if i % 5 == 0:
            print('.', end='', flush=True)

    print()



# Reads from stdin and writes it as an uncompressed
# LZW-encoded GIF image to stdout. The decoding can be
# done by gifsicle --unoptimize <file> or other tools.
class RandomGifGenerator:

    def __init__(self):
        self._fdin = sys.stdin.buffer
        self._fdout = sys.stdout.buffer


    def _create_lzw_image(self, fdin: BinaryIO,
            nframes: int = 1024) -> bytes:


        # LZW decoder state
        MAXBITS = 12
        FCT = 1 << MAXBITS
        FLG = 1 << (MAXBITS+1)

        INIT = 0b10101000 # initial code size
        REQSTD = 0b00000000 # minimum allowed code size (don't change!)
        EOF = FLG | 0 # end of input marker code


        # decompression lookup table
        lut = [None] * (nframes + nframes//4 + 2)
        fno = 0

        # read optional header
        hdr = b''
        while True:
            buf = bytearray(fdin.read(1))

            if len(buf)!= 1:
                raise Exception('unexpected EOF in file header')

            char = buf[0]
            hdr += char
            if char == 0x00: break


        # read logical screen descriptor
        buf = fdin.read(7)
        if len(buf)!= 7:
            raise Exception('unexpected EOF in logical screen descriptor')

        (w, h, resv, bgcol, pixelaspect) = \
                struct.unpack('<HHBBB', buf)


        # write header
        hdr += b'\x21\xf9\x04\x01\x00\x01\x00\x2c\x00\x00\x00\x00\xff\x00;\n'
        hdr += w.to_bytes(2, byteorder='little')
        hdr += h.to_bytes(2, byteorder='little')
        hdr += b'\x87'

        # This is where we'll store our output frames
        imgs = []

        # Do this until all frames are generated!
        while fno < nframes:


            # Read next frame from standard input
            curimg = b''
            bpp = 8
            while True:

                ctrlcode = ord(fdin.read(1)[0])


                # decode previous codes
                if (ctrlcode & 0x80)!= 0:
                    seq = [ctrlcode ^ 0xFF]

                    while True:

                        ctrlbyte = fdin.read(1)[0]


                        # reading more bits
                        if (ctrlbyte & 0x80)!= 0:
                            seq.append(
                                ((ctrlcode << 8) ^
                                    (ctrlbyte ^ 0xFF)).to_bytes(
                                        bpp>>3, byteorder='big')[::-1])

                            continue

                        # get final sequence value
                        val = (((ctrlcode<<8)^(ctrlbyte^0xFF))&FCT)+1
                        seq.extend([lut[(i-val)%len(lut)]
                                    for i in reversed(seq)])

                        entry = b'\x00'*bpp
                        for ch in seq[:-1]:
                            entry = lut[int.from_bytes(ch,
                                                byteorder='big')] + entry
                            lut.insert(-1, entry)

                        # update dictionary
                        lut.insert(-1, seq[-1])

                        if len(entry) > bpp:
                            bpp <<= 1

                        break


                else:

                    # read literal byte
                    seq = [(ctrlcode).to_bytes(bpp>>3,
                                            byteorder='big')[::-1]]


                    while True:

                        ctrlbyte = fdin.read(1)[0]


                        # reading more bits
                        if (ctrlbyte & 0x80)!= 0:
                            seq.append(
                                ((ctrlcode << 8) ^
                                    (ctrlbyte ^ 0xFF)).to_bytes(
                                        bpp>>3, byteorder='big')[::-1])


                            continue


                        # read final sequence character
                        seq.append(((ctrlcode << 8) ^
                                     (ctrlbyte ^ 0xFF)).to_bytes(
                                        bpp>>3, byteorder='big')[::-1])

                        entry = b'\x00'*bpp
                        for ch in seq[:-1]:
                            entry = lut[int.from_bytes(ch,
                                                byteorder='big')] + entry
                            lut.insert(-1, entry)

                        if len(entry) > bpp:
                            bpp <<= 1

                        break


                # Add sequence characters to current frame buffer
                curimg += b''.join(seq)


                # If this was the last frame, wrap up!
                if len(curimg) >= w*h*(bpp>>3):
                    imgs.append(curimg[:w*h*(bpp>>3)])
                    fno += 1
                    break

            # If there's no room left on the screen, then just stop here...
            if len(imgs) > nframes:
                break



        data = bytearray()


        # Create new GIF palette with less than 256 colors
        depth = max(2, min(8, bpp))
        pal = ([0]*depth)*3

        # Initialize GIF stream
        data += b'GIF89a'
        data += b'\x00\x00\x00\x00'
        datsizes = []

        # Write global color table
        data += (-3*(1<<(depth+1))).to_bytes(1, byteorder='big')
        data += b'\x0C\x00\x00'
        data += b'\x00'
        data += b'\x00\x00\x00\x00'


        # Write image headers
        for img in imgs:

            width = w; height = h
            xoff = yoff = 0

            # Compress image using LZW algorithm
            dictsize = 2**INIT
            initindex= len(lut)-1
            nextcode = len(lut)-1 + 1
            prevcode = 0
            bitcount = INIT+1
            dat = []
            offset = 0
            eoi = False
            clearflag = True
            eof = False

            while not eoi and not eof:

                # Find longest match starting at this point
                matchlen = 0
                matchoffs = 0
                for i in range(nextcode, len(lut)):
                    lutslot = lut[i]
                    if lutslot.startswith(img[offset:]):

                        # Match found! Output its prefix
                        if matchlen < len(lutslot)-1:
                            matchlen = len(lutslot)-1
                            matchoffs = i

                # Add any remaining characters to the dictionary as literals
                if matchlen == 0:
                    add = list(img[offset:])
                elif matchlen < len(add := img[offset:offset+matchlen]):
                    pass
                else:
                    add = []

                dictsize += matchlen + 1
                lut.insert(-1, img[offset:offset+add])

                # Pad short matches to full byte length
                if matchlen < 255:
                    add = [lut[prevcode][:-1]+b'\x00']*(255-matchlen)

                # Start encoding the frame
                bit = b'%c%c' % (clearflag|bitcount-REQSTD, prevcode)

                clearflag = False

                prevcode = nextcode
                outcode = prevcode
                nextcode = prevcode + 1

                while True:

                    # Collect compressed data
                    if eoi or len(dat)<32 or \
                            len(','.join(dat))+(len(bit)+7)//8 > 255:

                        dat.append(bit)
                        bit = ''
                        datstr = ','.join(dat)

                        while len(datstr)>255:
                            data += chr(255).encode('ascii')
                            datsizes.append(255)
                            datstr = datstr[255:]

                        data += chr(len(datstr)).encode('ascii')
                        datsizes.append(len(datstr))

                        data += datstr.encode('ascii')
                        dat = []

                    # Get next matching character
                    try:
                        cand = lut[nextcode]
                    except IndexError:
                        # End of compression? Okay,