#!/usr/bin/env python3
"""Generates a random animated gif with animextension support."""

import struct
from typing import BinaryIO


class AniFrame:
    """An individual frame in an animation."""

    def __init__(self, bpp, width, height,
                 literal_count, color_table_length, palette_size,
                 lzw_code_lengths):

        self._bpp = bpp
        self._width = width
        self._height = height

        self._literal_count = literal_count
        self._color_table_length = color_table_length
        self._palette_size = palette_size
        self._lzw_code_lengths = lzw_code_lengths


class AniHeader:
    """The header of an ANIMATED GIF file."""

    def __init__(self, version="89a", width=0, height=0,
                 flags=[], background=None, aspect_ratio=None):

        assert isinstance(version, str), type(version)
        assert len(version) >= 3, version
        assert all(c.isalpha() for c in version[0]), repr(version)
        assert all(c.isdigit() for c in version[1:-1]), repr(version)
        assert version[-1].isalpha(), repr(version)

        self.version = version
        self.width = width
        self.height = height
        self.flags = sorted(set(flags))
        self.background = background
        self.aspect_ratio = aspect_ratio


class AniExtension:
    """A single ANIMATED GIF EXTENSION block."""

    def __init__(self, label="", data=bytes()):

        if not data:
            raise ValueError("Data must have some content!")

        self.label = label
        self.data = data


class AniReader:
    """Reads ANIMATED GIF files into memory."""

    def __init__(self, source):
        """Creates an AniReader reading from source.

        Source can be either a filename or a file-like object.
        """

        try:
            self._file = open(source,"rb")
        except TypeError as e:
            if "not a" in str(e):
                self._file = source
            else:
                raise

        try:

            self._read_signature()
            self._read_logical_screen_descriptor()
            self._read_global_color_table()
            self._read_looping_block()
            self._read_extensions()

            # Check that we've reached EOF
            self._check_eof()

        finally:
            if hasattr(self,'_file') and self._file!= source:
                self._file.close()


    @property
    def signature(self):
        return self._signature

    @property
    def logical_screen_descriptor(self):
        return self._logical_screen_descriptor

    @property
    def global_color_table(self):
        return self._global_color_table

    @property
    def looping_block(self):
        return self._looping_block

    @property
    def extensions(self):
        return list(sorted(self._extensions, key=lambda x: x.label))

    def frames(self):
        """Returns a generator which yields each frame in turn."""
        if 'gce' in map(lambda x: x.label, self.extensions):
            raise NotImplementedError("Interframe GCEs not implemented.")
        elif self.global_color_table:
            yield self._build_image()
        else:
            raise Exception("No global color table found! Frame generation impossible.")


    def close(self):
        self._file.close()


    def _read_signature(self):

        # Read first three characters only
        sig = ""
        while True:
            try:
                char = chr(self._file.read(1)[0])
            except IndexError:
                break
            if char.isupper():
                sig += char
            else:
                self._file.seek(-1,1)
                break

        # Check signature validity
        if not ('G' == ord(sig[0]) and
                'I' == ord(sig[1]) and
                'F' == ord(sig[2])):
            raise Exception("Not a valid GIF file")

        self._signature = sig


    def _read_logical_screen_descriptor(self):

        # Read bytes
        b = self._file.read(7)

        # Unpack fields
        self._logical_screen_descriptor = AniHeader(*struct.unpack("<3sHHBBBxxxxxx", b))

        # Sanity check
        if not self._logical_screen_descriptor.version.startswith("8"):
            raise Exception("Only GIF87a supported!")

        if not 0 < self._logical_screen_descriptor.width < 65536:
            raise Exception("Invalid width value:",repr(self._logical_screen_descriptor.width))

        if not 0 < self._logical_screen_descriptor.height < 65536:
            raise Exception("Invalid height value:",repr(self._logical_screen_descriptor.height))

        if self._logical_screen_descriptor.flags:
            print("WARNING: unknown behaviour for these flags:",
                  " ".join(self._logical_screen_descriptor.flags))

    def _read_global_color_table(self):

        if self._logical_screen_descriptor.flags:
            field = self._logical_screen_descriptor.flags[0]
            bpp = int(field[0])
            ifield = int(field[1:])
        else:
            field = None
            bpp = None

        if bpp:

            # Calculate length of global color table
            rgb_sizes = {8:(3,), 4:(1,2), 2:(1,1,1)}[bpp]
            rgb_bytes = sum(rgb_sizes)
            ifield -= rgb_bytes
            if ifield % rgb_bytes:
                raise Exception("Malformed field", field)
            length = ifield // rgb_bytes


        else:
            length = False

        self._global_color_table = []

        if length:

            # Read entire global colour table
            self._global_color_table = [
                struct.unpack("BBB", self._file.read(3))
                for i in range(length)]

            # Sanity check -- do all the colours work?
            for r,g,b in self._global_color_table:
                if not (0 <= r < 256 and 0<=g<256 and 0<=b<256):
                    raise Exception("Non-colour entry:", r,g,b)


    def _read_looping_block(self):

        # Read magic word
        loopmagic = self._file.read(6)
        if not loopmagic == b"\x21\xFF\x0B\x4E\x45\x54":
            raise Exception("Looping block missing?")

        # Read rest of 18 bytes
        b = self._file.read(18)
        self._looping_block = dict(zip(("delay_time","transparent_index"),
                                struct.unpack("<HBxxxBBB",b)))


    def _read_extension(self):

        # Read functional code byte
        funcbyte = ord(self._file.read(1)[0])
        if not (funcbyte in [0x01, 0xfe]):
            raise Exception("Invalid extension code: {}".format(hex(funcbyte)))

        # This should never happen but let's make sure...
        if funcbyte == 0xFE:
            print("WARNING: skipping unrecognized Extension Block!")
            return

        # Read the label
        label = "".join(chr(self._file.read(1)[0])
                        for i in range(1,9))[:-1]

        # Skip byte if needed
        nb = ord(self._file.read(1)[0])
        if not nb:
            pass
        elif nb > 9:
            raise Exception()
        else:
            self._file.seek(nb-9, 1)

        # Read the data until next separator or end-of-file
        data = b""
        cont = True
        while cont:
            try:
                char = self._file.read(1)[0]
            except IndexError:
                break
            if char == 0x2C:
                cont = False
            else:
                data += bytes([char])

        # Add an extension object
        self._extensions.append(AniExtension(label,data))



    def _read_extensions(self):

        self._extensions=[]
        while True:
            try:
                self._read_extension()
            except Exception as e:
                print("Exception during read of extension:\n",str(e)) 
                try:
                    self._check_eof()
                except Exception as e:
                    print("(Also exception during checking for eof):\n", str(e)) 
                    continue

                break


    def _check_eof(self):

        pos = self._file.tell()
        char = self._file.read(1)
        self._file.seek(pos, 0)
        if char:
            raise Exception("{} bytes remaining.".format(len(char)))


    def _build_image(self):

        # Determine BPP
        if self._logical_screen_descriptor.flags:
            field = self._logical_screen_descriptor.flags[0]
            bpp = int(field[0])
        else:
            field = None
            bpp = None

        # Set up defaults
        fn = lambda : {}

        # Figure out what