#!/usr/bin/env python3

import struct
from typing import List, Tuple

from util import randint


def write_byte(value: int) -> bytes: return (chr(value)).encode()


def write_short(value: int) -> bytes: return struct.pack(">H", value)
    

def write_long(value: int) -> bytes: return struct.pack(">L", value)
    

def write_string(s: str) -> bytes: return s.encode("ascii")


class LogicalScreenDescriptor(object):

    def __init__(self):

        self._width = randint(20, 512)
        self._height = randint(20, 512)
        self._flags = [randint(0, 1) for _ in range(8)]


    @property
    def width(self) -> int:
        return self._width
    

    @property
    def height(self) -> int:
        return self._height
    

    @property
    def flags(self) -> List[int]:
        return self._flags


    def serialize(self) -> bytes:
    
        result = b""
        
        result += write_short(self.width)                     # ScreenWidth
        result += write_short(self.height)                    # ScreenHeight
        result += write_byte((self.flags[0] << 7) |           # Packed fields
                            (self.flags[1] << 6) | 
                            (self.flags[2] << 5) | 
                            (self.flags[3] << 4) | 
                            (self.flags[4] << 3) | 
                            (self.flags[5] << 2) | 
                            (self.flags[6] << 1) | 
                            (self.flags[7]))                        
        result += write_byte(randint(0, 255))                 # Background Color Index
        result += write_byte(randint(0, 255))                 # Pixel Aspect Ratio

        return result


class GlobalColorTable(object):

    def __init__(self):

        self._num_colors = randint(0, 256)

        if self._num_colors == 0:
        
            # Do not use global color table at all
            self._table = []
            return

        else:

            # Generate global colours
            self._table = [(randint(0, 255), randint(0, 255), randint(0, 255))
                for _ in range(self._num_colors)]


    @property
    def num_colors(self) -> int:
        return self._num_colors


    @property
    def table(self) -> List[Tuple[int]]:
        return self._table
    
    
    def serialize(self) -> bytes:

        result = b""

        if len(self._table) > 0:

            result += write_byte(1 << 7 | self.num_colors)     # Table Size

            for r, g, b in self.table:

                result += write_byte(r)                       # Red
                result += write_byte(g)                       # Green
                result += write_byte(b)                       # Blue
            
        return result

    
    
class ImageDescriptor(object):

    def __init__(self):
    
        self._left = randint(-32768, 32768) // 2
        self._top = randint(-32768, 32768) // 2
        self._width  = randint(0, 512) 
        self._height = randint(0, 512) 
        self._flags = [randint(0, 1) for _ in range(8)]


    @property
    def left(self) -> int:
        return self._left


    @property
    def top(self) -> int:
        return self._top

        
        
    @property
    def width(self) -> int:
        return self._width
    

    @property
    def height(self) -> int:
        return self._height

    
    @property
    def flags(self) -> List[int]:
        return self._flags

        
        
    def serialize(self) -> bytes:
    
        result = b""
        
        result += write_short(self.left)                      # Left
        result += write_short(self.top)                       # Top
        result += write_short(self.width)                     # Width
        result += write_short(self.height)                    # Height
        result += write_byte((self.flags[0] << 7) |           # Packed fields
                            (self.flags[1] << 6) | 
                            (self.flags[2] << 5) | 
                            (self.flags[3] << 4) | 
                            (self.flags[4] << 3) |  
                            (self.flags[5] << 2) | 
                            (self.flags[6] << 1) | 
                            (self.flags[7]))                        

        return result
    
    
    
class GraphicControlExtension(object):

    def __init__(self):

        self._flags = [randint(0, 1) for _ in range(8)]
        self._delay = randint(0, 65536)
        self._transparent_idx = randint(0, 255)
        

    @property
    def flags(self):
        return self._flags


    @property
    def delay(self):
        return self._delay


    @property
    def transparent_idx(self):
        return self._transparent_idx


    def serialize(self):

        result = b""
        
        result += b"\x21"                                     # Extension Introducer
        result += b"\xf9"                                     # Graphic Control Label
        result += b"\x4"                                      # Block Size
        result += b"\x0"                                       # Flags

        if self.flags[0] == 0:                              # Transparent Color Flag
            
            result += write_byte(self.transparent_idx)        # Transparent Color Index

        result += write_short(self.delay)                     # Delay Time
        result += b"\xff"                                     # Transparent Color Index


        return result
            
                
class PlainTextExtension(object):

    def __init__(self):

        self._text_left = randint(-32768, 32768) // 2
        self._text_top = randint(-32768, 32768) // 2
        self._grid_left = randint(-32768, 32768) // 2
        self._grid_top = randint(-32768, 32768) // 2
        self._cell_width = randint(0, 256)
        self._cell_height = randint(0, 256)
        self._foreground_idx = randint(0, 255)
        self._background_idx = randint(0, 255)
        

    @property
    def text_left(self):
        return self._text_left


    @property
    def text_top(self):
        return self._text_top


    @property
    def grid_left(self):
        return self._grid_left


    @property
    def grid_top(self):
        return self._grid_top


    @property
    def cell_width(self):
        return self._cell_width


    @property
    def cell_height(self):
        return self._cell_height


    @property
    def foreground_idx(self):
        return self._foreground_idx


    @property
    def background_idx(self):
        return self._background_idx


    def serialize(self):

        result = b""

        result += b"\x21"                                     # Extension Introducer
        result += b"\x01"                                     # PLAINTEXT Introducer
        result += b"\x0c"                                     # Block Size

        result += write_short(self.text_left)                 # Text Grid Left
        result += write_short(self.text_top)                  # Text Grid Top
        result += write_short(self.grid_left)                 # Grid Cell Left
        result += write_short(self.grid_top)                  # Grid Cell Top
        result += write_byte(self.cell_width)                  # Character Width
        result += write_byte(self.cell_height)                 # Character Height
        result += write_byte(self.foreground_idx)              # Foreground Color Index
        result += write_byte(self.background_idx)              # Background Color Index

        return result
        
        
class ApplicationExtension(object):

    def __init__(self):

        self._application_id = "\x01NETSCAPE2.0".encode("ascii")
        self._sub_block_size = 3
        self._loop_count = randint(0, 100)
        

    @property
    def application_id(self):
        return self._application_id


    @property
    def sub_block_size(self):
        return self._sub_block_size


    @property
    def loop_count(self):
        return self._loop_count


    def serialize(self):

        result = b""

        result += b"\x21"                                     # Extension Introducer
        result += b"\xff"                                     # Application extension label
        result += b"\x0b"                                     # Block Size
        result += self.application_id