#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom() by default)
# with an optional prefix that is repeated at least once in the output.
# 
# The resulting file will have exactly $size bytes of
# data following the GifHeader.
def generate_random_gif(
        size: int,
        rng: BinaryIO = os.urandom,
        repeatable: bool = False,
        prefix: bytes = b'',
        out: BinaryIO = sys.stdout.buffer
    ) -> None:

    if repeatable:
        assert len(prefix)!= 0

    headlen = 6 + len(prefix)
    payloadlen = size - headlen

    # Use the first three bytes as datalen instead of a fixed value. This makes it
    # easier to test for false positives.
    datalen = int.from_bytes((bytearray([headlen]) +
                              rng.read(3)),
                             byteorder='big')
    
    # Write header and data
    assert datalen >= size 
    out.write(b'\x47\x49\x46\x38\x39\x61') # 'GIF89a'
    out.write(struct.pack('>B', headlen-1))     # PrefixSize
    out.write(struct.pack('>B', payloadlen&255)) # LZWMinCodeSize
    out.write(struct.pack('>H', 0))               # Width
    out.write(struct.pack('>H', 0))               # Height
    out.write(struct.pack('>B', 0))               # Flags
    out.write(prefix)                             # Prefix
    out.write(rng.read(payloadlen))               # Payload