#!/usr/bin/env python3

import struct
from typing import BinaryIO


RNG = None


def set_rng(new_rng: BinaryIO):
    """Set this module's RNG."""
    assert isinstance(new_rng, BinaryIO), 'RNG must be an IO object.'
    global RNG
    RNG = new_rng


def generate_random_gif(
        out: BinaryIO, minwidth=64,
        maxwidth=512, minheight=64, maxheight=512, nframes=10):
    """Generates a random GIF image in binary form to `out`, using
    modulus_rng.py's RNG. The width and height are randomly generated such
    that they are multiples of two between minwidth <= w < maxwidth and
    minheight <= h < maxheight (inclusive); the frame count is randomly
    generated from 1 to nframes inclusive. Note that it may not actually
    make sense to have both minwidth and maxwidth!= w, since there won't
     be enough room for all the pixels."""

    assert len(extblocks) >= 2, 'need at least one ANIM extension block'
    assert len({block[0]: True for block in extblocks}) == len(extblocks), \
           f"Only one {block[0]} allowed."

    # construct headers
    header = b'GIF89a'
    width = RNG.randint(minwidth, maxwidth - 1) // 2 * 2 + 2
    height = RNG.randint(minheight, maxheight - 1) // 2 * 2 + 2
    gctsize = 2 << bool(RNG.randint(0, 1))
    globalcolortablelen = 2 ** (gctsize + 1) - 1
    localcolortablelen = 2 ** ((bool(RNG.randint(0, 1))) << 2)
    fc = RNG.randrange(nframes) + 1

    # write headers
    out.write(header)
    out.write(struct.pack('<HHB', width, height, gctsize))

    # append colortables if any
    for i in range(globalcolortablelen // 3 + bool(globalcolortablelen % 3)):
        r, g, b = tuple([int(v * 255) for v in RNG.choice(range(256))]
                       [0] for _ in "rgb")
        out.write(struct.pack('BBB', r, g, b))
    if localcolortablelen > 0:
        for i in range(localcolortablelen // 3 + bool(localcolortablelen % 3)):
            r, g, b = tuple([int(v * 255)
                            for v in RNG.choice(range(256))][0]
                           for _ in "rgb")
            out.write(struct.pack('BBB', r, g, b))

    # append image descriptor block
    out.write(struct.pack('BBHBBBB',
                         0xf7,                   # separator
                         0,                      # left corner x coordinate
                         height & 0xff,          # lower corner y coordinate
                         ((height >> 8) & 0x0f) << 4 | \
                          (width & 0x0f) >> 2,     # size flag
                         (fc & 0xf)|((((width&0x0f)<<2)&0xf0)),
                            0,
                                 0))             # no local color table

    # write image data blocks
    lastbyte = None
    for j in range(fc):
        for i in range(((width + 7) / 8) * height):
            byte = RNG.getrandbits(8) % 256
            if lastbyte is not None:
                while RNG.randint(0, 2) == 0:
                    byte |= lastbyte
            lastbyte = byte ^ (i == 0) ^ (((j == 0) or
                                          (j == (fc - 1))) << 7)
            out.write(struct.pack('B', byte))

    # copy and extend extension blocks
    for name, data in extblocks:
        data += bytes(
            [sum(map(ord, name)) +
             len(data)] * (8 - (len(data) % 8)))

        out.write(data)