#!/usr/bin/env python3

import gzip
import io
import struct
from typing import BinaryIO

# Generates a random GIF file in a format suitable for Web browsers
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # The first 10 bytes of a GIF are fixed.
    out.write(b'\x47\x49\x46\x38\x39\x61\x2c\x00')

    # Read an unsigned short from `rng`.
    def read_ushort():
        return int.from_bytes(
                rng.read(2), byteorder='little', signed=False
            )
    
    height = read_ushort()
    width = read_ushort()
    print("Image size:", (width, height))
    
    # Write the LZW minimum codesize.
    mincodesize = 2**read_ushort()
    if not (mincodesize >= 2 and mincodesize <= 12):
        raise ValueError(f"Invalid minimum codesize {mincodesize}")
    print("Minimum codesize", mincodesize)
    assert out.tell() == 10
    out.seek(-1, io.SEEK_CUR) # Rewind to write this value.
    out.write(struct.pack('B', mincodesize))

    # Write the image descriptor block.
    dsfhdr = b'\x2C'
    xpos = ypos = 0
    dsf = f"{dsfhdr}{xpos:02X}{ypos:02X}" \
          f"\x00\x00{width:02X}{height:02X}".encode()
    crc = (0).to_bytes(4, byteorder='little', signed=False)
    dsf += crc
    out.write(dsf)

    # Generate some random pixel data.
    lzwdata = []
    while len(lzwdata) < width * height:
        codelen = read_ushort() & 0xFFF # zero-padded code length

        if codelen > 0 and codelen!= 12:
            # Non-empty packet with non-special code length.
            # This is just raw pixel data.

            # We only care about the last two bytes, but we'll read them all.
            # The zeroes aren't important because they will be encoded as their
            # own codes.
            pixels = rng.read(codelen - 2)
    
            lzwdata.append((codelen - 2).to_bytes(2, byteorder="big"))
            lzwdata.append(pixels)
        
        elif codelen == 0:
            continue # No need to encode empty packets
            
        else:
            # Special case; we have a special code.
            
            
            if codelen == 12:
                # A 'clear table' code.
                
                # Clear the compression table because it's getting old.
                # It doesn't matter what values we set here; they're never used.
                comptable = [
                    b'', # Empty string represents no prefix at all.
                    1 + 3,
                    b'\x00\xff\xff\xff',
                    b'\x00\x00\xfe\xff',
                    b'\x00\x00\x00\xff',
                    b'\x00\x00\x00\x00',
                    ]

                # The next few codes are literal ASCII control characters.
                for byte in (
                        b'\xf9\x04', 
                        b'\xfa\x04', 
                        b'\xfb\x04', 
                        b'\xfc\x04'):
                    lzwdata.append(byte)

            elif codelen == 14:
                # An 'end of information' code.
                break
            
    # Now that we've generated our LZW data, let's pack it up into blocks.
    # Each block can contain up to 255 codes.
    lzwblocks = []
    currblock = None
    for i, code in enumerate(lzwdata):
        if currblock is None or i % 255 == 0:
            if currblock is not None:
                lzwblocks.append(currblock)
                
            currblock = struct.pack('>H', len(code))
        
        currblock += code
        
    if currblock is not None:
        lzwblocks.append(currblock)

    # Write each block out.
    for block in lzwblocks:
        out.write(block)



