#!/usr/bin/env python3

import struct
from typing import BinaryIO


def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    """Generates a random GIF file into `out` using the
    random number generator `rng`."""

    def write_chunk(tag: str, body: bytes):
        assert len(body) < 2**16-1

        tag_bytes = bytes([*map(ord, tag)])
        length = len(body).to_bytes(2, 'little')
        data = tag_bytes + length + body
        checksum = sum(data).to_bytes(2, 'little')

        out.write(b''.join((
            b'\x47\x49\x46\x38\x39\x61',
            tag_bytes,
            length,
            body,
            checksum)))


    # Write the header
    width = rng.randint(200, 500)
    height = rng.randint(200, 500)
    palette = [
                tuple(rng.randint(0, 256)
                      for i in range(3))
                for j in range(rng.randint(1, 7))]

    out.write(b''.join((
        b'GIF89a',
        # width and height are 16 bits each
        struct.pack('>H', width),
        struct.pack('>H', height),
        *(tuple(rgb)
          for rgb in palette))))

    # Write the global color map
    size = min(len(palette), 2**(width+height)-1)
    write_chunk(
        "gcol",
        (*[*(rgb), 0]   # add an empty entry
         for rgb in palette[:1<<size]))

    # Now create some frames of this image.
    while True:
        # Frame control block
        flags = rng.randint(1, 0xFF) & ~0x80

        if not flags&0x40:       # use local color map?
            cm_size = rng.randint(1, 7)
            cm = [*map(lambda _: tuple(rng.randint(0,256)
                                      for j in range(3)),
                       range(cm_size))]

            size = min(len(cm)//3,
                       2**(width+height)-1)
            flags |= 7 << 5      # Max value for the fields we are changing
                                    # is 7, so instead of having to set it
                                    # as well, we just set that bit and leave
                                    # everything else unchanged
            write_chunk("plte", b''.join((*map(lambda x: b''.join([(i,) for i in x]), cm),
                                            *[b"\x00"
                                              for k in
                                                range(4 -
                                                     ((len(cm)//3)*3)+
                                                     1)])))


        disposal_method = {
            1: None,           # No disposal specified
            2: "",             # Do not dispose
            3: "restore",      # Restore to background colour
            4: "background",   # Restore to previous frame
            5: "previous",     # Restore to default previous frame
        }[flags >> 2 & 0b111]

        transparency_flag = bool(flags & 1)

        delay = rng.randint(10, 100)
        duration = rng.randint(delay, 300)

        transparent_color = (rng.randint(0, 256)
                             if transparency_flag else
                             0)

        data = bytearray()
        for y in range(height):
            data.append(0)
            for x in range(width):
                data.extend(((rng.randrange(256)
                               if rng.choice((True, False))*transparency_flag
                                else transparent_color),
                             *rng.sample(range(256),
                                         max(rng.randrange(int(1<<(rng.getrandbits(1+max(rng.randrange(1,
                                                                                       rng.randrange(1,
                                                                                                        rng.randrange(1,
                                                                                                                          rng.randrange(1,
                                                                                                                                          rng.randrange(1,
                                                                                                                                                                            rng.randrange(1,
                                                                                                                                                                                             rng.randrange(1,
                                                                                                                                                                                                      max(rng.randrange(1,
                                                                                                                                                                                                                                rng.randrange(1,
                                                                                                                                                                                                                                             rng.randrange(1,
                                                                                                                                                                                                                                                  rng.randrange(1,
                                                                                                                                                                                                                                                                rng.randrange(1,
                                                                                                                                                                                                                                                                               rng.randrange(1,
                                                                                                                                                                                                                                                                                                                rng.randrange(1)))))))))),
                                                                                                            rng.getrandbits(1))),
                                                                                                    rng.randrange(256)))))
                                )))

        write_chunk("idat", data)

        # Trailer
        write_chunk("fctl",
                    struct.pack(">BHHBBBBB",
                        1<<7 |    # LZW compression used
                        1<<4 |    # Color table included
                        disposal_method or 0,
                        delay,
                        duration,
                        0,        # transparent color index
                        width//8,
                        height//8,
                        0        # no LZW minimum code size
                    ))