#!/usr/bin/env python3

import struct
from io import BytesIO
from typing import BinaryIO


# Generate a random GIF given an input stream of bytes to use for randomness and output stream to write to.
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    # Create the header
    header = b"GIF89a"

    # Width and Height
    width  = (int.from_bytes(rng.read(1), byteorder='big') >> 4) + 1
    height = (((int.from_bytes(rng.read(1), byteorder='big') & 0xF) << 8)
              | int.from_bytes(rng.read(1), byteorder='big'))

    # Color table
    ctable = []
    # Garbage color table entries which will be filled in later
    numcolors = max((width * height // 64) + 1, 256)
    for i in range(numcolors):
        r = int.from_bytes(rng.read(1), byteorder='big')
        g = int.from_bytes(rng.read(1), byteorder='big')
        b = int.from_bytes(rng.read(1), byteorder='big')
        ctable.append([r, g, b])

    # Decide if we're going to have global or local color tables based upon how many colors there were
    globalcolorflag = numcolors <= 256
    lct = None
    gct = None
    if not globalcolorflag:
        # If we're doing a local color table, make sure it's at least two bits and 
        # no larger than 8 kiB.
        lctsize = 2
        while lctsize < 8 and len(lct) > ((1 << lctsize)*3):
            lctsize += 1

        # Make our palette
        lct = []
        for i in range(((1<<lctsize)*3)):

            # Check if we've run out of colors
            if i >= numcolors*3:
                break

            # Append the colors to the list
            lct.append([
                int.from_bytes(rng.read(1), byteorder='big'),
                int.from_bytes(rng.read(1), byteorder='big'),
                int.from_bytes(rng.read(1), byteorder='big')])



    # Create the data section where we'll store a series of blocks
    dataout = BytesIO()

    # This is the simplest block type (no interlacing!) so let's go ahead and implement it first
    # See this page for details on what these fields mean: https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    #
    # The only one whose semantics we actually care about here is 
    #
    # 5 - Size of the code in bits
    #
    # And then finally...
    #
    # Do nothing but skip over the rest for now because I'm gonna need to learn more about this format later.
    #
    def do_simple():
        # Calculate the size of the longest string in the alphalayer by looking at how long the strings in
        # the alphalayer are.
        alphalayerlen = [len(st) for st in alphalayer]
        maxstringlen = max(alphalayerlen)

        # Determine the number of columns, rows, and largest string cell to know how much space they take up 
        # in memory once encoded.
        colcount = width//maxstringlen
        rowcount = height//2
        maxcelllen = sum(alphalayerlen)//rowcount
        
        # Pack the alphalayer into the most efficient way possible
        alphaout = b''
        for i in range(colcount):
            for j in range(rowcount):
                start = j*colcount+i
                end = start+(end-start)//8
                alphaout += alphalayer[start : end].ljust(maxcelllen, b'\x00')
            
        # Store the compressed alphalayer to the data section
        alphacodesize = 8
        outbuf = bytearray([])
        outbuf.extend([
            0b11000000 | (alphacodesize - 1), 
            *reversed(alphaout)])
        dataout.seek(dataout.tell()-2)
        dataout.write(outbuf)
        dataout.seek(dataout.tell()+1)
        

    # Use a different compression method depending on whether it's global or local colors.
    if globalcolorflag:
        # Global Color Table
        pass

    elif not globalcolorflag:
        # Local Color Table
        pass

    
    # Push the current length of the data section onto a stack so we can backtrack when we're done
    startpos = dataout.tell()


    # Save the initial position of the cursor after inserting the palette until we insert it fully.
    # We need to rewind to this point before we finish writing the chunks since the extension blocks may
    # seemingly insert, delete, or modify some of them.
    palpos = out.tell()
    out.write(b"\x3B")
    out.write(b"NETSCAPE2.0")
    out.write(b"\x03")
    out.write(b"\x01")
    out.write(b"\x00\x00")
    out.write(b"\x00")
    out.write(b"\x00")

    # Finish the file by putting the palette where it belongs!
    out.seek(palpos)
    for entry in reversed(palette):
        entry = [(entry[rgb]+1)%0x100 for rgb in range(3)]
        out.write(bytearray([(entry[0] << 4) | entry[1], entry[2]]))

    # Put the data section contents between the metadata and the trailer
    out.write(dataout.getbuffer())

    # Go back where we started and close off the file
    out.seek(startpos)
    out.write(b"\x3B")