#!/usr/bin/env python3

import os
import sys


def write_short(data: bytes):
    assert len(data) == 2
    return data[0] * (1<<8) + data[1]


def read_short(data: bytes):
    assert len(data) >= 2
    return int.from_bytes(data[:2], 'big')


def main():
    if '-h' in sys.argv or '--help' in sys.argv:
        print('Usage: {} <nframes> <size>'.format(sys.argv[0]))

    with open('/dev/urandom', 'rb') as randfile:

        # Read image dimensions from command-line arguments
        imgw = int(sys.argv[1]) if len(sys.argv) > 1 else None
        imgh = int(sys.argv[2]) if len(sys.argv) > 2 else None

        if not imgw and not imgh:
            raise ValueError("Please provide at least one argument")

        framecount = 0
        while True:

            # Generate output filename based on how many frames we've written
            fn = '{:04d}.gif'.format(framecount)
            
            # Write to stdout instead of a file
            if fn == '-':
                f = sys.stdout.buffer
            else:
                f = open(fn, mode='wb')

            # Generate next GIF animation frame
            try:
                generate_next_animation_frame(randfile, f, imgw, imgh)

                # Stop after generating nframes animation frames
                framecount += 1
                if framecount == numframes:
                    break
                
                # Flush so that user can see some progress
                sys.stdout.flush()
            finally:            
                f.close()


def generate_next_animation_frame(
        randfile: BinaryIO, outfile: BinaryIO,
        width: int=None, height: int=None
    ) -> tuple:
    """Generates a single frame for an animated GIF image"""
    
    # Optional: We want to have a fixed size for our images
    if not width and not height:
        # Choose random number between [60x60; 256x256]
        wmin, hmin = 60, 60
        wmax, hmax = 256, 256
        dims = randfile.read(1)[0] % ((wmax - wmin) * (hmax - hmin))
        
        width = wmin + (dims // (hmax - hmin))
        height = hmin + (dims % (wmax - wmin))


    def chunkwriter(chunkid, content):
        length = len(content).to_bytes(2, 'little')
        flagbyte = b'\x80'   # LZW compression
        outfile.write(chunkid + length + flagbyte + content)
        
    
    # Header fields
    version = b'89a'         # Fixed value
    logicalscreenwidth = width.to_bytes(2, 'little')     # Image width
    logicalscreenheight = height.to_bytes(2, 'little')    # Image height
    packedfield = bytes([b'\x70'])      # No global colour table used
    backgroundcolourindex = bytes([b'\xf'])        # Transparent color index
    pixelaspectratio = b'\x00'          # Pixel aspect ratio
    gctflag = bytes([b'\x00'])           # Don't use global colour map
    gctsize = b'\x00\x00'               # No global colour table
    gct = b''                           # Global colour table is empty

    # Each row uses the same local colour table
    lct_size = b'\xff\x00'                # Maximum size for GCT
    lct_entries = b''                    # Local colour table entries are empty
    lct_flags = b'\x00'                  # All entries are 24 bit RGB values

    # Write image descriptor to file
    outfile.write(version
                   + b'\x00\x00'                       # Size (fill later)
                   + logicalscreenwidth
                   + logicalscreenheight
                   + packedfield
                     + backgroundcolourindex
                     + pixelaspectratio
                   + gctflag
                       + gctsize
                         + gct
                   + lct_size
                       + lct_entries
                       + lct_flags)
    

    # Number of loops
    loops = b'\x01\x00'                   # Loop once per frame

    # First image descriptor has no LZW compression
    firstdescriptor = b'\x2c'              # Start of image descriptor block
    descriptor = b'\x2c'                  # Start of image descriptor block
    codesize = bytes([b'\x20'])           # Use default LZW code size value
    outfile.write(firstdescriptor
                  + b'\x00\x00'                 # Size (fill later)
                  + logicalscreenwidth
                  + logicalscreenheight
                  + packedfield
                    + backgroundcolourindex
                    + pixelaspectratio
                  + gctflag
                      + gctsize
                        + gct
                  + lct_size
                    + lct_entries
                    + lct_flags
                  + codesize)
    
    
    # Text extension blocks
    exttext = b'\x01\xf9\x0bNETSCAPE2.0\x03\x01' \
              b'\x00\x00\x00'            # Extension block identifier
    textextlength = len(loops + exttext).to_bytes(2, 'little')
    outfile.write(b'\x21' + textextlength + exttext)
    

    # Application extension block
    appblock = b'\x01\xfe' + b'\x00\x0b' \
               b'NETSCAPE2.0'            # Application name
    appblocklength = len(appblock).to_bytes(2, 'little')
    junkstring = b'foobar'              # Some arbitrary string
    netscapeext = b'\x03' + b'\x01' + len(junkstring).to_bytes(2, 'little') \
                  + junkstring             # Netscape application extension
    ntscflag = b'\x01'                   # Is this NTSC?
    loopext = b'\x01'+loops              # Loop count
    outfile.write(b'\x21' + appblocklength + appblock
                   + netscapeext+ntscflag+loopext)
    
        
    # Graphics control extension block
    graphicsextension = b'\x01'       \
                        + b'\x0f'      \
                        + b'\x00\x0e' \
                          + b'\x00\x00' \
                            + b'\x00\x00' \
                              + b'\x00\x00'

    
    disposalmethod = b'\x00'          \
                     | b'\x01' << 2 \
                     | b'\x02' << 4
    userinputflag = b'\x00'          \
                    | b'\x01' << 1 \
                    | b'\x02' << 2 \| b'\x03' << 3 \
                    | b'\x04' << 4 \
                    | b'\x05' << 5 \
                    | b'\x06' << 6 \
                    | b'\x07' << 7
    transparencylevel = b'\x00'
    delaytime = b'\x00\x00'
    transparentcolorindex = b'\x00'   # Unused
    terminator = b'\x00'              # End of graphics control extension block
    
    outfile.write(graphicsextension
                 + disposalmethod
                 + userinputflag
                 + transparencylevel
                 + delaytime
                 + transparentcolorindex
                 + terminator)

    # Plain Subblocks
    plainsubblockstart = b'\x21'       \
                         + b'\xff'     \
                         + b'\x0b'     \
                           + b'subblocks:\x00'
    
    # Plain subblock 1
    plainsubblockdata = os.urandom((len(plainsubblockstart)-5)//2)
    outfile.write(plainsubblockstart+plainsubblockdata)
    
    
    # Plain subblock 2
    plainsubblockdata = os.urandom((len(plainsubblockstart)-5)//2)
    outfile.write(plainsubblockstart+plainsubblockdata)
    
    
    # Plain subblock 3
    plainsubblockdata = os.urandom((len(plainsubblockstart)-5)//2)
    outfile.write(plainsubblockstart+plainsubblockdata)
    
    
    # Plain subblock 4
    plainsubblockdata = os.urandom((len(plainsubblockstart)-5)//2)
    outfile.write(plainsubblockstart+plainsubblockdata)
    
    
    # Plain subblock 5
    plainsubblockdata = os.urandom((len(plainsubblockstart)-5)//2)
    outfile.write(plainsubblockstart+plainsubblockdata)
    

    # Crunchy Blocks
    crunchysubblockstart = b'\x21'      \
                          + b'\xff'    \
                          + b'\x0b'    \
                             + b'crunchy blocks:\x01'
    
    # Crunchy subblock