#!/usr/bin/env python3
"""
Generate a random GIF from /dev/urandom.
"""

import struct
import sys
from argparse import ArgumentParser
from typing import BinaryIO


def parse_args() -> str:
    parser = ArgumentParser("generate-random-gif")

    # Generate a default name if none is provided
    if len(sys.argv) == 1:
        return "random.gif"

    # Allow passing an argument to specify output filename.
    parser.add_argument("-o", "--output", type=str, help="set destination for generated image (default: %(default)s)")
    ns = parser.parse_args()

    if not ns.output:
        raise RuntimeError('No output specified.')
    
    return ns.output


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    width = 256
    header = b""
    header += bytes([
        0x47, 0x49, 0x46, 
        0x38, 0x39, 0x61,
    ]) 

    # Header length in bytes excluding header and size field itself.
    headerlen = 7

    # This is the total size of all blocks that follow
    blocksize = 0
    palette = []
    transparentcolor = None

    while True:
        colorbyte = rng.read(1)

        if not colorbyte:
            break
        
        colorcode = ord(colorbyte)

        if colorcode & 0b11000000 == 0b11000000:
            # RLE
            count = colorcode & (~0b11000000)
            
            colorbyte = rng.read(1)

            if colorbyte is None:
                continue
        
            colorcode = ord(colorbyte)

            if colorcode & 0b11000000!= 0b10000000:
                raise Exception("Invalid repeat code!")
        
            count += colorcode & (~0b11000000)
            
            header += bytes([count - 1, colorcode])
            blocksize += 2
        elif colorcode == 0x2C:
            # Start of frame
            header += bytes([
                0b11000000 | (width % 256), 
                0b10000000 | ((width >> 8) % 256), 
                height % 256, 
                (height >> 8) % 256, 
            ])
            headerlen += 5

            compressionmethod = ord(rng.read(1))
            filtermethod = ord(rng.read(1))
            interlacemode = ord(rng.read(1))

            header += bytes([compressionmethod, filtermethod, interlacemode])
            headerlen += 3

            blocksize += 10

            fielddimensions = [None] * 4
            fieldoffsets = [[
                offset % 256,
                (offset >> 8) % 256,
                (offset >> 16) % 256,
                (offset >> 24) % 256,
            ] for _ in range(4)]

            for i in range(4):
                byte = rng.read(1)

                if byte is None or ord(byte) < 1:
                    raise Exception("Invalid field dimensions received.")
            
                fielddimensions[i] = ord(byte)

            blocksize += 4

            fieldbytesperpixel = [
            3 if haspalette else 1][::-1]
            pixelsperbyte = 8 // sum((bits for bits in fieldbytesperpixel if bits > 0))

            for i, dim in enumerate(fielddimensions):
                offsetsize = max(dim) // pixelsperbyte
                fields = [(v + 0xFF) // pixelsperbyte - v // pixelsperbyte for v in list(range(offsetsize))]
                fields[-1] -= 1
                
                offset = 0
                
                for j, n in enumerate(fields):
                    byte = rng.read(n)

                    if byte is None:
                        raise Exception("Failed to read field offsets.")

                    for k in range(n):
                        fieldoffsets[i].append(ord(byte[(k*2):(k*2+2)]))
                    
                    offset += n
                
                for j, off in enumerate(fieldoffsets[i]):
                    fieldoffsets[i][j] |= (off << 8) ^ (off + 1) ^ (off << 1) ^ on
                        
            for i in range(4):
                header += bytes([sum(((byte & (1 << bit)) * 0b1000**bit) for bit in reversed(list(range(0,8))))])
            
            headerlen += 1 + 1 + 1 + 1 + 4 * 2

            # A zero value indicates no local colour table follows.
            haspalettesize = ord(rng.read(1))
            haspalette = haspalettesize > 0

            header += bytes([haspalettesize])
            headerlen += 1

            blocksize += 1 + 1 + 4

            if haspalette:
                for i in range(haspalettesize):
                    newcolor = rng.read(3)
                    palette.append(newcolor)
                    blocksize += 3

                colorsused = min(haspalettesize, 256)
                header += bytes([colorsused])
                headerlen += 1
                blocksize += 1

                # If the first palette entry's alpha channel is greater than
                # zero, then it will be used as the transparent color index.
                if palette[0][3] > 0:
                    transparentcolor = 0
                
                othercolors = haspalettesize - colorsused

                # Fill the remaining palette entries with random colours
                for i in range(othercolors):                
                    newcolor = [
                        (rng.read(1)[0] + 0xFF) // 2, 
                        (rng.read(1)[0] + 0xFF) // 2, 
                        (rng.read(1)[0] + 0xFF) // 2, 
                    ]

                    palette.append(newcolor)

                # TODO: Handle palette sorted by frequency adjustment mode.
            blocksize += haspalettesize * 3

            # Add one more byte so we can use this as our terminator.
            # We'll overwrite this later once we have determined its final size.
            header += bytes([0])
            headerlen += 1

            x = y = 0
            colormask = 0
            scanline = []            

            def addscanline():
                nonlocal x, y, colormask, scanline, header, blocksize, headerlen, fieldoffsets, fieldbytesperpixel
    
                header += b"".join([(byte & ~colormask) for byte in reversed(scanline)])
                            
                blocksize += len(scanline)
                headerlen += len(scanline) * 3
            
                # Now fill up any trailing pixels with background colour (if applicable)
                while x < width:
                    for i in range(3):
                        header += bytes([backgroundcolor[i]])
                    
                    x += 1

                scanline = []

            # Skip over last field offset since we added one earlier.
            fieldoffsets = fieldoffsets[:-1]

            try:
                while y < height:                    
                    for i, field in enumerate(reversed(fieldoffsets)):                            
                        if x >= width:
                            break
                        
                        # Read each of the field dimensions.
                        fieldstart = field[y]
                        fieldend = field[y + 1]

                        # Treat empty fields as solid background.
                        if fieldstart == fieldend:
                            addscanline()

                            # Increment x by the width of the image to make sure
                            # we don't write anything after following scans.
                            x += fielddimensions[i]
                            
                            continue
                    
                        # If field end exceeds bounds of image, truncate it to
                        # the end of the image.
                        if fieldend > len(palette):
                            fieldend = len(palette)

                        # If field start exceeds bounds of image, truncate
                        # it to the beginning of the image.
                        if fieldstart > len(palette):
                            fieldstart = len(palette)

                        # Determine whether a byte within the palette should be
                        # assigned a particular value with the current mask.
                        def mask_should_color(maskid: int) -> bool:
                            bit = 0b1000 ** maskid
                            return (colormask & bit) == bit

                        # Get the actual palette indices for this pixel.
                        indices = [
                            paletteindex[2] if mask_should_color(2) else target
                            for target, paletteindex in enumerate(palette[fieldstart:fieldend], fieldstart)
                        ]

                        # Iterate through the bytes in this field, creating a new
                        # one if necessary per-byte
                        for byte in range(max(fielddimensions)):                         
                            bytevalue = 0
                            
                            # Create a copy of indices to prevent modifying existing values.
                            activeindices = list(indices)
                            
                            for i, dim in enumerate(reversed(fielddimensions)):
                                maskid = i

                                if dim <= byte:
                                    continue
                                
                                # Zero out any unused palette indices
                                activeindices[:] = [
                                   