#!/usr/bin/env python3

import struct
from typing import BinaryIO

transparent_color = b"\x00\xFF\xFF\xFF"


# Generate a GIF given an input stream of bytes to use for randomness and output stream to write to.
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO,

        width=None,      # Width of the frame
        height=None,     # Height of the frame
        colordepth=None  # Bit depth
        ):

    # If not provided, determine dimensions by using the max of our seed bytes
    if not width or not height:
        width  = (max(int.from_bytes(rng.read(1), 'big'), 1) >> 4) + 1
        height = (((max(int.from_bytes(rng.read(1), 'big'), 1) & 0xf) << 8)
                  | max(int.from_bytes(rng.read(1), 'big'), 1))

    # If no bit depth is specified, pick one based on the image's size
    if not colordepth:
        colordepth = {
            2**2 : 2,
            2**3 : 3,
            2**4 : 6}[width * height]

    # Finally, write the animated GIF header
    out.write(b'GIF89a')

    # Write the logical screen descriptor
    out.write(struct.pack('<HHBBB',
                         width,         # Logical screen width in pixels
                         height,        # Logical screen height in pixels
                         0xF0,          # Packed field byte, see below.
                         0x00,          # Global color table entries, must be zero since no global colortable exists yet.
                         0x00           # Background color index, useless since there isn't a global colortoab yet.
                        ))

    # Write the global color table
    out.write(struct.pack('{}B'.format(3 ** colordepth),
                      *[0xFF]*(3**colordepth)))
    
    # Write the netscape application extension
    out.write(b"\x21\xff\x0bNETSCAPE2.0")
    out.write(struct.pack("<HB", 3, 1))
    out.write(b"\x03\x01")
    out.write(b"\x00")

    # Write the "plaintext" application extension required before transparent colors may appear
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    out.write(b"\x21\x01\x00")

    # Write the graphics control extension block specifying how the animation should continue after the last frame
    out.write(b"\x21\xF9\x04")
    out.write(b"\x0A")                 # Delay time in hundredths of seconds
    out.write(b"\x00")                # Transparent color index, again useless until we have a colortable
    out.write(b"\x00")                # Block terminator
    out.write(b"\x00")

    
    # Write the entire image data
    for y in range(height):
        for x in range(width):
            out.write(rng.read(1))
        # EOB blocks must be padded with zeroes so that when they're decoded their lengths match up
        out.write(b'\x00'*(((1+width)//8)*int((y%8!=0)*(y%8)+((1+width)%8))))
        
    # Write the trailer
    out.write(b";")