#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# between minwidth <= w < maxwidth and minheight <= h < maxheight (inclusive).
# If animated is True, then a comment will be added into the file which will
# trigger an animation with the given number of frames. Otherwise, it will use
# minimal colors so that the output is at most 9919 bytes long. The maxframes
# parameter sets the maximum number of frames to choose; defaults to 10.
#
# Note that this function does not guarantee that each frame's palette is unique
# across all frames, but different palettes are allowed within a single frame.
# This could result in some animations being much larger than others because of
# duplicated palettes.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, minwidth=64,
        maxwidth=512, minheight=64, maxheight=512, maxframes=None, animated=True):

    def get_color():
        return int(rng.choice(range(256))*255), int(rng.choice(range(256))*255), int(rng.choice(range(256))*255)

    if not animated:
        maxframes = 1
    elif maxframes is None:
        maxframes = 10

    # construct headers
    header = b'GIF87a'
    width = rng.randint(minwidth,maxwidth-1)//2*2+2
    height = rng.randint(minheight,maxheight-1)//2*2+2
    globalcolortablelen = 2<<(bool(rng.randint(0,1)))
    localcolortablelen = 2**((bool(rng.randint(0,1)))<<2)
    fc = rng.randrange(maxframes)+1
    extblocks = []

    # write header and screen descriptor
    out.write(header)
    out.write(struct.pack('<HHB', width, height, globalcolortablelen))
    
    # add comment containing the number of frames if we're making an animated GIF
    if fc>1:
        extblocks.append(('NETSCAPE2.0', '\x03\x01\x00\x00\xff\x00'))
    else:
        extblocks.append(("NETSCAPE2.0", "\x03\x01\x00\x00\x00\x00"))

    # build palettes
    globalpalette = list()
    localpalette = dict()
    for i in range(globalcolortablelen//3 + bool(globalcolortablelen%3)):
        c = get_color()
        globalpalette += c
        out.write(struct.pack('BBB',c[0],c[1],c[2]))
    for i in range(localcolortablelen // 3 + bool(localcolortablelen % 3)):
        c = get_color()
        localpalette[(i%2==0)] = c
        out.write(struct.pack('BBB',c[0],c[1],c[2]))

    # write image descriptor block
    out.write(struct.pack('BBHBBBB',
                        0xf7,#separator
                        0,    #left corner x coordinate
                        height&0xff,   #lower corner y coorindate
                        ((height>>8)&0x0f)<<4 | \
                          (width&0x0f)>>2,  #size flag
                        (fc & 0xf)|((((width&0x0f)<<2)&0xf0)),
                            0,
                            0))            #no local color table

    # write image data blocks
    lastbyte = None
    for j in range(fc):

        bgindex = rng.randrange(globalcolortablelen//3 +
                                bool(globalcolortablelen%3))
        bgcolor = globalpalette[bgindex*3:(bgindex+1)*3]

        # add background color as first pixel
        out.write(struct.pack('BBB', bgcolor[0], bgcolor[1], bgcolor[2]))

        for i in range(((width+7)/8)*(height-1)):
            byte = rng.getrandbits(8)%256

            # check whether this pixel uses the local or global palette
            if i%(2*((width+7)//8))<2:
                paltype = 'local'
                pallen = localcolortablelen
                pal = localpalette[(i//(2*((width+7)//8))) % 2]
            else:
                paltype = 'global'
                pallen = globalcolortablelen
                palindex = (byte >> bool(rng.randint(0,1))) % (pallen//3 +
                                                             bool(pallen%3))
                pal = globalpalette[palindex*3:(palindex+1)*3]

            # apply palette lookups
            byte &= ~(1 << bool(rng.randint(0,1)))       # erase current pixel
            byte ^= (rng.randrange(256)>=(0 if byte else bgcolor[bool(rng.randint(0,1))] +
                                          pal[(0 if byte else 1)]))<<bool(rng.randint(0,1))
                                                     # set new pixel
            out.write(struct.pack('B',byte))


        # add trailing padding byte unless this is the final frame
        if j!= (fc-1):
            out.write(struct.pack("B",rng.randrange(256)))

    # copy and extend extension blocks
    for name,data in extblocks:
        data += bytes([sum(map(ord,name))+len(data)])*(8-(len(data)%8))
        out.write(data)