#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:             BinaryIO=None,
        width:           int    = 1024,
        height:          int    = 768,
        colordepth:      int    = 8,
        bgcolor:         int    = None,
        delay:           float  =.05,
        outfile:         str    = 'output.gif',
    ) -> None:

    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    blackcode = bytes([bgcolor & (255**2)])
    whitecode = bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])
    
    with open(outfile,'wb') as f:

        # The GIF header
        version = b"89a"
        screendescriptor = struct.pack('>HHBBB',
                width,
                height,
                0xF0 | colordepth,
                0,
                0,
        )
        
        # Write the header and descriptor block
        print("Writing the header")
        f.write(b"GIF"+version+screendescriptor)

        # Build the color lookup table
        print("Building the color map")
        colortable = []
        for i in range((2**colordepth)-1):
            x = (i*((bgcolor**(1/(256.0)))+.5))%int(256**3)
            colortable += [struct.pack('>BBB',
                    int(((x>>16)&255)**1.5),
                    int(((x>>8)&255)**1.5),
                    int((x&255)**1.5),
                ),
            ]
        colortable += [blackcode]
        colortable += [whitecode]

        # We can now write the palette.
        print("Writing the palette")
        f.write(b"".join(colortable))

        # Now we need to write the image data itself...
        print("Writing the image")
        f.write(b"\x2C\x00\x00\x00\x00\x00\x00\x00\xFF")   # Image descriptor
        f.write(b"!\xF9\x04\x01\x00\x00\x00,\x00\x00\x00") # Graphics control extension
        f.write(b"\x00\x01\x00\x21\xFE\x09D\x0B\xCF\x00") # Application Extension
        f.write(b"\x0BNETSCAPE2.0\x03\x01")                  # Application Data
        f.write(b"\x00\x00\x00!")                           # Trailer

        # Now it's time to write out the frame information.
        print("Writing frames...")
        while True:

            print(".",end="")
            
            tsize = rng.read(1)[0]&0xff
            lctflag = bool(rng.read(1)[0]&0x80)
            interlaceflag = bool(rng.read(1)[0]&0x40)
            sortflag = bool(rng.read(1)[0]&0x20)
            gctsize = rng.read(1)[0]&0xff

            if gctsize!= len(colortable):
                raise RuntimeError("Invalid GCT size.")

            f.write(bytes([tsize]))                 # Image separator size
            f.write(bytes([lctflag]))               # LCT flag
            f.write(bytes([interlaceflag]))         # Interlace flag
            f.write(bytes([sortflag]))              # Sort flag
            f.write(bytes([gctsize]))               # Size of global color table
            f.write(bytes(rng.read(gctsize)))       # Global color table
            f.write(bytes([0]))                     # Background color index
            f.write(bytes([0]))                     # Pixel aspect ratio
        
            # LZW minimum code size
            f.write(bytes([8]))
            # Block size
            f.write(bytes([0x2c]))
            # Subblock size
            subbs = 0xffff-rng.read(2)
            f.write(subbs.to_bytes(2,'big'))

            data = bytearray()
            while len(data)<subbs:

                b = rng.read(1)[0]&0xff
                if b==0:
                    continue

                elif b<=127:
                    data+=bytes([b]+rng.read(b))
                
                else:

                    slen = 2**b+rng.read(b)
                    lb = rng.read(1)[0]&0xff

                    ndata = data[-lb:] + data[:-lb] + bytes([data[lb]])
                    
                    while len(ndata)>slen:
                        ndata = bytes([
                            (
                                ndata[(~k)&0xff]*msb 
                                + ndata[(k&0xff)+1]*lsb
                            )&0xff 
                            for k in range(len(ndata))]
                        )
                        
                    data = ndata
                    
            f.write(data[:subbs])

            f.write(b"!\xF9\x04")                   # Graphics control extension
            f.write(delay.to_bytes(2,"little"))    # Delay between frames 
            f.write(b"\x00")                         # Transparent color index
            f.write(b"\x00")                         # Terminator