"""
Generate a random GIF from /dev/urandom.
"""

import os
import struct
import sys
from argparse import ArgumentParser
from typing import BinaryIO


# Writes a GIF palette containing a single entry of index `i`.
def write_palette(i: int, out: BinaryIO):
    r = g = b = i * 256 // 256
    out.write(struct.pack("<BBB", r & 0xFF, g & 0xFF, b & 0xFF))


# Write a frame with the given dimensions and color table from `input`
# to `output`. This function only writes one complete frame worth of
# data - it does NOT loop on itself.
def write_frame(
        input: BinaryIO,
        width: int, height: int,
        numcolors: int, colorsused: int,
        delay: int, transindex: int, interlaced: bool,
        out: BinaryIO,
    ):
    # Write the initial screen descriptor
    out.write(b"\x2c")           # Extension introducer
    out.write(b"\x00\x00")       # Block size
    out.write(b"\xf9\x04")       # Graphics control label
    out.write(b"\xff\xff\xff\x00") # All bits are reserved
    out.write(b"\x00\x00\x00\x00") # No transparancy
    out.write(width.to_bytes(2, byteorder="little"))   # Width in pixels
    out.write(height.to_bytes(2, byteorder="little"))  # Height in pixels

    # Write the local color table.
    assert numcolors <= 256
    for _ in range(numcolors):
        write_palette((transindex + _) % numcolors, out)

    
    out.write(b"\xfe\x00")             # Image separator block
    out.write(b"\x00")                 # no LZW minimum code size
    out.write(delay.to_bytes(2, byteorder="little"))  # Delay timer

    lctflag = bytes([(interlaced << 7) | ((colorsused & 7) << 4)])

    out.write(struct.pack(
        "<BBBBB",
        0x2c,                       # Start of frame marker and length
        (numcolors & 0xF) |        # Number of colors used
        (lctflag[0] & ~0x0F),       # Global flag byte.
        0x00 |                     # No sort flags
        (lctflag[0] & 0x0f),         # Size of local color table 
        0x00 |                     # No background color index
    ))                                

    def __next_byte():
        ret = input.read(1)
        while len(ret)!= 1 or ret[0] == 0:
            try:
                ret = input.read(1)     # Zeroes indicate end of stream
            except KeyboardInterrupt as e:
                exit(-1)

        return ret


    # Write the actual pixel data until we encounter another zero byte, which
    # indicates compressed data begins and ends.
    while True:
        c = __next_byte()[0]
        if c < 0x80:
            out.write(bytes((c,)))
        else:                                  # Discrete pixel run mode

            # Read the next chunk's parameters.
            n = abs(int(c - 0x80)) + 1          # Number of pixels in this chunk
            if n >= 64:
                r = ord(__next_byte()[0])      # Next byte
                n += 63 + (r & 0x3F)           
            
            if c > 0x80:                        # For negative runs, read more
                r = ord(__next_byte()[0])      # the following byte to get sign
                n += (-1 if (r & 0x80) else 1)*((((r & 0x7F) << 8) | ord(__next_byte()[0]))+1)
                
            # Finally, write the packet info
            out.write(bytes((0xFC, n)))        # Run length packet
            for _ in range(n):
                r = ord(__next_byte()[0])      # Pixel value
                out.write(bytes((r,)))                 
                
        
        # Stop when the previous was complete.
        if c == 0: break;
            
    

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    w = h = int.from_bytes(
        rng.read(2), byteorder='big', signed=False)
    print(f"Dimensions: {w} x {h}")
    out.write(b"GIF89a")
    globalcolortablebits = 0
    globalcolortablesize = 0
    flag = rng.read(1)[0]
    print(f"Flasg: {hex(flag)}")
    n = 0
    while flag & (1<<n)!= 0:
        n += 1
    print(f"Number of global color tables: {n}")
    globalcolortablesize = 2**n
    globalcolortablebits = [None]*n
    lctsize = 2**((globalcolortablesize >> 1) + 1)
    for i in range(n):
        globalcolortablebits[i] = rng.read(1)[0] & 0x07
    print(f"Global Color Table Bits: {[hex(e) for e in globalcolortablebits]}")
    lzwmincodesize = rng.read(1)[0] + 1
    print(f"LZW Min Code Size: {lzwmincodesize}")

    out.write(w.to_bytes(2, 'little'))    # width in px
    out.write(h.to_bytes(2, 'little'))    # height in px

    # The logical screen descriptor specifies a packed field where bit N is set
    # if the associated value is present or should be valid. Fields marked with
    # (*) must always have these values.
    #
    # Screen width          : wxh (short form) or WxH (long form) (*)
    # Packed fields         : 0MSTTAAABBB
    #  M = GCT style        : 1 - global color table follows, 0 - no
    #                          (see below for description)
    #  S = Sort order       : 1 - use global color table sorted by decreasing
    #                           size, 0 - use global color table sorted by
    #                           increasing size
    #  T = Color resolution : 1 - 1..256, 0 - 2..256
    #  A = Background color : 1 - use bgColorIndex, 0 - don't
    #  B = pixel aspect ratio: 1 - square pixels, 0 - not (4:3 or 1:1)
    #  R = Reserved         : Must be 0
    #  I = GCT size         : 1 - 2..256, 0 - 2..3
    out.write(struct.pack('<BBBBBBB',
        0x21,                                   # extension introducer
        b'!',                                    # application extension
        b'\xb0',                                 # subblock length
        b'\x00',                                 # NETSCAPE2.0 identifier
        b'\x03',                                 # Flags Index
        b'\x01',                                 # Netscape Application Version
        b'\x01\x00',                             # Subblock size (2 bytes)
        b'\x3B'                                  # Terminator
    ))


    # Write the global color table
    for i in range(n):
        out.write(struct.pack('<BBB',
            *(range(globalcolortablebits[i]+1))),               # RGB channels
        )

    # Write the graphics control extension.
    out.write(b"\x21\xF9\x04")                   # Extension Introducer
    out.write(b"\xff\xff\xff\x00")               # Block size
    out.write(b"\x00\x00\x00\x00")               # All bits are reserved
    out.write(b"\x00\x00\x00\x00")               # No transparancy
    out.write(w.to_bytes(2, "little"))          # Width in pixels
    out.write(h.to_bytes(2, "little"))          # Height in pixels
    out.write(b"\x00")                           # Delay timer
    out.write(b"\x00")                           # Transparent color index
    out.write(b"\x00")                           # Block terminator
    # End of graphics control block

    # Write the netscape comment extension that specifies the repeat count. We
    # arbitrarily choose 10 loops instead of 0 since it makes the animation
    # easier to watch.
    out.write(b"\x21\xFE\x0BNETSCAPE2.0\x31\x10\xFF\x00")


    #