#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO
from enum import Enum, auto, unique
from PIL import Image

# http://www.matthewflickinger.com/lab/whatsinagif/bits_and_bytes.asp
class ColorTableFlag(Enum):
    GCT_NONE = 0,
    GCT_COLORRESOLUTION = auto() # 1 bit
    GCT_SORTED = auto()         # 1 bit
    GCT_INTERLACED = auto()     # 1 bit

    def to_byte(self): return bytes([ self.value ])

class ImageDescriptorFlags(Enum):
    IMAGE_LEFT_TO_RIGHT = 0b0
    IMAGE_RIGHT_TO_LEFT = auto() | 0b1
    IMAGE_HAS_LOCAL_COLOURTABLE = auto() << 2 | 0b0
    IMAGE_INTERLACED = auto() << 2 | 0b1
    OVERSIZE_FOR_SCREEN = auto() << 4 | 0b0
    LOCAL_COLORREFS_PRESENT = auto() << 4 | 0b1

    @staticmethod
    def from_data(data):
        if len(data)!= 1:
            raise Exception("Image descriptor has invalid length")

        flags = ImageDescriptorFlags(data[0])

        left_to_right = bool(flags & ImageDescriptorFlags.IMAGE_LEFT_TO_RIGHT == ImageDescriptorFlags.IMAGE_LEFT_TO_RIGHT)
        interlaced = bool(flags & ImageDescriptorFlags.IMAGE_INTERLACED == ImageDescriptorFlags.IMAGE_INTERLACED)

        return (left_to_right, interlaced)



@unique
class FrameControlBlock(Enum):
    LABELS_NOT_USED = 0x00
    TRANSPARENT_PIXELS_ARE_INDEXES = auto() | 0b0
    TRANSFORM_DATA_IS_AN_INTERMEDIATE_FRAME_IN_THE_ANIMATION = auto() | 0b1
    TRANSPARENT_COLOR_IS_INDEXED = auto()<<1|0b0
    USED_INTERLACE_MODE_IS_SPECIFIED = auto()<<2|0b0
    SORTING_OF_COLOR_TABLE_IS_NEEDED = auto()<<3|0b0
    COLOR_TABLE_IS_INTENDED_AS_GLOBAL_COLOR_TABLE = auto()<<4|0b0
    GLOBAL_COLOR_TABLE_IS_INTENDED_AS_BACKGROUND_COLOR = auto()<<5|0b0
    RESERVED = auto()<<6|auto()<<7|0b0


    def to_bytes(self):
        return bytes([ self.value ])

@unique
class DisposalMethod(Enum):
    NO_DISPOSAL = 0
    DO_NOT_DISPOSE = auto()
    RESTORE_BACKGROUND = auto()
    RESTORE_PREVIOUS = auto()

    def to_bytes(self):
        return bytes([ self.value ])

def create_image_descriptor_block(
        width: int, height: int, color_table: list, global_color_table: list, frame_control_block: FrameControlBlock, transparent_index=None, disposal_method=DisposalMethod.NO_DISPOSAL
    ) -> bytes:

    block = [ ]

    image_left_to_right, interlaced = ImageDescriptorFlags.from_data(frame_control_block.to_bytes())
    interlaced_modifier = { True : " ", False : "-" }[interlaced]

    # TODO: set x and y position correctly based on the screen size
    color_resolution = 7
    sort_flag = True

    x_position, y_position = (0, 0)
    width &= 0xffff
    height &= 0xffff

    local_color_table_flag = None
    if not global_color_table is None or len(global_color_table) > 0:
        local_color_table_flag = ColorTableFlag.GCT_NONE
    elif not color_table is None or len(color_table) > 0:
        local_color_table_flag = ColorTableFlag.GCT_COLORRESOLUTION | \
                ColorTableFlag.GCT_SORTED | \
                ColorTableFlag.GCT_INTERLACED
    else:
        local_color_table_flag = ColorTableFlag.GCT_NONE

    packed_local_colorref_size = ((len(color_table) - 1).bit_length()+1) // 8 if (color_table is not None and len(color_table)) > 0 else 0
    bits_per_pixel = 1 if color_table is None or len(color_table) < 2**packed_local_colorref_size else packed_local_colorref_size+1

    packed_local_colorref_size |= (local_color_table_flag.to_byte()[0] << 4)

    packed_sorted_flag = 0x01 if sort_flag else 0x00

    packed_reserved_flag = 0x00

    flags = packed_local_colorref_size | packed_sorted_flag | packed_reserved_flag

    packed_width = (width % 2**(4 + packed_local_colorref_size)).to_bytes(3, 'little')
    packed_height = (height % 2**(4 + packed_local_colorref_size)).to_bytes(3, 'little')

    packed_width += (width >> (4 + packed_local_colorref_size)).to_bytes(1, 'little')
    packed_height += (height >> (4 + packed_local_colorref_size)).to_bytes(1, 'little')

    block += [
        ord('!'), ord(','),
        0xf9, # graphics control extension label
        4,    # block size

        *frame_control_block.to_bytes(),

        0xff, # application extension label
        0x0b, # block size
        ord('A'), ord('N'), ord('I'), ord('M'), ord('E'), ord('X'), ord('T'), ord('_'),
        *(disposal_method.to_bytes()), # dispose method

        0x00, # reserved
        0x04, # Block size of graphic control sub-block

         *packed_width,
         *packed_height,

         *((transparent_index&0xff).to_bytes(1,'little')), # transparent index as byte

     ]

    if local_color_table_flag in [ColorTableFlag.GCT_COLORRESOLUTION, ColorTableFlag.GCT_FIXED]:
        block += [
            0x0f, # block size
            0x00, # background colour index

            *color_table[1].to_bytes(3,'big'), # first palette entry

            0x00 # block terminator
        ]

    # This is just a dummy. No idea what this is supposed to be for.
    block += [
        0x00 # block terminator
    ]

    return bytes(block)

GIF_HEADER = b"GIF89a"

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def make_random_gif(
        rng: BinaryIO, out: BinaryIO,
        num_frames: int = 10, duration: int = 200, loop_count:int = 1,
        width: int = 32, height: int = 32,
        color_palette: list = []
    ):



    out.write(GIF_HEADER)


    # Write LZW prefix table
    out.write(struct.pack("<BBB", 0x2c,0, 0))

    # Write logical screen descriptor
    out.write(struct.pack("<HHHBB", width, height, 0x80, 0, 0x00))

    # write color table
    if len(color_palette) > 0:
        field_values = []
        field_values += [(len(color_palette)-1).to_bytes(1,'little')]
        field_values += [0]*3
        field_values += [0]*3*len(color_palette)
        out.write(b"".join(field_values))

    # Make frames now
    for i in range(num_frames):
        image_file_path = f"/tmp/{i}.png"
        img = Image.new('RGB', (width, height), color=(0,0,0))


        pixels = img.load()   # create the pixel map
        pixdata = img.getdata()

        colors = sorted({ p for _,_,p in pixdata })[:256]
        print(colors)

        dithered = {}
        for i, col in enumerate(img.convert('P').getcolors()):
            dithered[(col[0],col[1])] = colors[i]

        new_pixels = []
        for j in range(height):
            for k in range(width):
                r,g,b = pixels[k,j]
                new_pixels.append(dithered.get((r,g), (0,0,