#!/usr/bin/env python3

import struct
from os import urandom as randsource
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (default is /dev/urandom).
# `width`, `height`: size of the image in pixels.
# `has_trailer`: if set to True appends a trailer block; otherwise writes only
#                the image and terminator blocks.
def generate_random_gif(
        rng=randsource, out:BinaryIO, width=1024, height=512, has_trailer=True
    ) -> None:

    def write(data):
        assert len(data) % 4 == 0,\
                f"len({data}) should be divisible by four."
        while data:
             chunksize = min(len(data), maxsize)
             #print("chunk:", chunksize)
             out.write(struct.pack('>I', chunksize))
             out.write(data[:chunksize])
             data = data[chunksize:]

    header = b'\x47\x49\x46\x38\x39\x61'
    logical_screen_descriptor = \
            struct.pack('<6sHHbb',
                    *([b'', width, height] + [int(not has_trailer)]))
    global_color_table_entry = bytes(3)*256*3
    local_color_table_entry = bytes(3)*256*3
    graphics_control_extension =  \
            struct.pack('<B3sbBBBHB',
                    0xf9, *[b'']*3 + [0x00, 0xff, 0x00, 0x00, 0x00])
    image_descriptor = struct.pack('<BHBBHHHBB', 
            0x2c, width & 0xffff, width >> 16, 
            height & 0xffff, height >> 8, 0, 0, 8);
    application_extention = struct.pack('<BBBBBI', 
            0x21, 0xff, 0x0b, 0x4e, 0x45, 0x54, 0x53, 0x43, 0x41, 0x50,
            0x45)
    comment_block = struct.pack('<BBBBBI', 
            0xfe, 0x01, 0xfc, 'S'[0], 'O'[0], 'M'[0], 'E'[0],
            0x00, 0x00, 0x00) + b'Meow meow!'
    plain_text_extension = struct.pack('<BBBBBBBIIBBI', 
            0x21, 0xf9, 0x04, 0x04, 0x00, 0x10, 0x00, 0x10, 0x00, 0x00,
            0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00) 
    sequence_delimiter = struct.pack('<3B', 
            0x3b, 0x00, 0x00);
    maxsize = 2**sizeof_size - 16

    out.write(header)
    out.write(logical_screen_descriptor)
    write((global_color_table_entry 
        + local_color_table_entry
        + graphics_control_extension
        + image_descriptor
        + data_subblocks
        + application_extention
        + comment_block
        + plain_text_extension
        + sequence_delimiter
       ))

    if not has_trailer: return
    
    # The termination code must appear at most once per file.
    # Since it occupies exactly 13 bytes, we can append it
    # next time this function is called even after another termintation
    # code or any other extension has already been appended.
    # (We could choose to have a separate buffer for just the last but one bytes
    # that don't fit in an integer-aligned chunk.) 
    out.write(bytes(maxsize - out.tell()))
    out.write(b'\x3b\x00\x00')