#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF image in binary form to `out`, using `/dev/urandom`.
# The width and height are randomly generated such that they are multiples of 2
# between minwidth <= w < maxwidth and minheight <= h < maxheight (inclusive).
# The frame count is randomly generated from 1 to nframes inclusive.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, minwidth=64,
        maxwidth=512, minheight=64, maxheight=512, nframes=10):

    # construct headers
    header = b'GIF89a'
    width = rng.randint(minwidth,maxwidth-1)//2*2+2
    height = rng.randint(minheight,maxheight-1)//2*2+2
    backgroundcoloridx = rng.randint(0,255)
    aspectratio = rng.choice([(b'\x00', b'\x00'),
                              (b'\x40', b'\x21'),
                              (b'\x20', b'\x21'),])
    gctsize = 2 << bool(rng.randint(0,1))
    globalcolortablelen = 2**(gctsize+1)-1
    localcolortablelen = 2**((bool(rng.randint(0,1)))<<2)
    fc = rng.randrange(nframes)+1
    extblocks = [
        (b'NETSCAPE2.0', b'\x03\x01\x00\x00\xff\x00'),
        (b'ANIMEXTS1.0', b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')]


    # write headers
    out.write(header)
    out.write(aspectratio[0])
    out.write(struct.pack('<BBBBB',
                           width&0xff,  # screen width low byte
                           (width&0x7f00)>>8,  # screen width high byte
                           height&0xff,  # screen height low byte
                           (height&0x7f00)>>8,  # screen height high byte
                           bgcoloridx))  # background color index

    # append colortables if any
    for i in range(globalcolortablelen//3 + bool(globalcolortablelen%3)):
        r,g,b = tuple([int(v*255) for v in rng.choice(range(256))][0] for _ in "rgb")
        out.write(struct.pack('BBB',r,g,b))
    if localcolortablelen > 0:
        for i in range(localcolortablelen // 3 + bool(localcolortablelen % 3)):
            r, g, b = tuple([int(v * 255)
                            for v in rng.choice(range(256))][0]for _ in "rgb")
            out.write(struct.pack('BBB', r, g, b))

    # append image descriptor block
    out.write(struct.pack('BBHBBBBB',
                        0x2c,        # image separator
                        0,           # left corner x coordinate
                        height&0xff, # lower corner y coorindate
                        ((height>>8)&0x0f)<<4 | \
                          (width&0x0f)>>2,       # size flag
                        (fc & 0xf),          # number of frames
                        0,                   # local color table flag
                        bgcoloridx,         # background color idx
                        0))                  # reserved

    # write image data blocks
    lastbyte = None
    for j in range(fc):
        for i in range(((width+7)/8)*height):
            byte = rng.getrandbits(8)%256
            if lastbyte!= None:
                while rng.randint(0,2)==0:
                    byte |= lastbyte
            lastbyte = byte ^ (i==0) ^ (((j == 0) or
                                        (j == (fc - 1))) << 7)
            out.write(struct.pack('B',byte))

    # copy and extend extension blocks
    for name,data in extblocks:
        data += bytes([sum(map(ord,name))+len(data)])*(8-(len(data)%8))
        out.write(data)