#!/usr/bin/env python3


"""Generates a random GIF file into `out`. By default it uses `/dev/urandom`, but you can also provide an alternative RNG."""

RANDOM_FILE = '/dev/urandom'
MAX_WIDTH   = 1000   # max image width
MAX_HEIGHT  = 1000   # max image height
COLOR_TABLE_SIZE = 1 << 4    # size of the color table (max is 1<<8-1=255)
DATA_LENGTH      = 0x100000 # length of the content chunk (max is FFFF0000h)

import sys
import os
from datetime import datetime as dt
from typing import Optional, BinaryIO
import struct


def get_random_file() -> str:
    return os.environ.get('RNG', RANDOM_FILE)


def main():
    with open(sys.argv[1], 'wb') as f:
        g = RandomGifGenerator(f, get_random_file(), MAX_WIDTH, MAX_HEIGHT, COLOR_TABLE_SIZE, DATA_LENGTH)
        g.generate()


class RandomGifGenerator:
    """Generate a random GIF file to `out`."""


    def __init__(self, out: BinaryIO,
                 rng: Optional[str] = None,
                 max_width: int = MAX_WIDTH, max_height: int = MAX_HEIGHT,
                 num_colors: int = COLOR_TABLE_SIZE, data_length: int = DATA_LENGTH
                ) -> None:

        self._out = out
        if not rng or not os.path.exists(rng):
            raise Exception("Invalid RNG path")
        else:
            self._rng = open(rng, "rb")

        self._max_width  = max_width
        self._max_height = max_height
        self._num_colors = num_colors
        self._data_length = data_length


    def write(self, data: bytes)->None:
        self._out.write(data)


    def generate(self) -> None:
        self.write(b"GIF89a")
        self.write(struct.pack(">H", self._make_int(min=2 ** 1, max=self._max_width)))
        self.write(struct.pack(">H", self._make_int(min=2 ** 1, max=self._max_height)))


        # TODO better color generation (currently just RGB values from /dev/urandom)
        self.write(b"\xff\xff\xff" * self._num_colors);

        self.write((self._data_length).to_bytes(3, byteorder="big"))
        while True:
            b = self._rng.read(1)
            self.write(b)

            if len(b) == 0: break



        



    @staticmethod
    def _make_int(*, min : int, max: int)->int:
        return struct.unpack("<I", os.urandom(4))[0] % ((max - min) + 1) + min

