#!/usr/bin/env python3

import struct
from typing import BinaryIO
import argparse


# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
def generate_random_gif(
        rng:        BinaryIO=None,
        width:      int    = 1024,
        height:     int    = 768,
        colordepth: int    = 8,
        bgcolor:    int    = None,
        outfile:    str    = 'output.gif',
    ) -> None:


    # Set up the base colors that will be used to build our palette.
    if bgcolor is None or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    if not isinstance(bgcolor,bytes):
        bgcolor = bytes([bgcolor & (255**2)])
        bgcolor += bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])

    with open(outfile,'wb') as f:


        # The GIF header
        version = b"89a"
        screendescriptor = struct.pack('>HHBBB',
                width,
                height,
                0xF0 | colordepth,
                0,
                0,
        )
        
        # Write the header and descriptor block
        print("Writing the header")
        f.write(b"GIF"+version+screendescriptor)

        # Build the color lookup table
        print("Building the color map")
        colortable = []
        for i in range((2**colordepth)-1):
            x = (int(i*(bgcolor[0]*bgcolor[1]**(1/(256.0)))))%256**3
            colortable += [struct.pack('>BBB',
                    int(((x>>16)&255)**1.5),
                    int(((x>>8)&255)**1.5),
                    int((x&255)**1.5),
                ),
            ]
        colortable += [bgcolor]

        # We can now write the palette.
        print("Writing the palette")
        f.write(b"".join(colortable))

        # Now we need to write the image data itself...
        print("Writing the image")
        f.write(b"\x2C\x00\x00\x00\x00\x00\x00\x00\xFF")   # Image descriptor
        f.write(b"!\xF9\x04\x01\x00\x00\x00,\x00\x00\x00") # Graphics control extension
        f.write(b"\x00\x01\x00\x21\xFE\x09D\x0B\xCF\x00") # Application Extension
        f.write(b"\x0BNETSCAPE2.0\x03\x01")                  # Application Data
        f.write(b"\x00\x00\x00!")                           # Trailer
        print("\nDone.")



def main() -> None:

    parser = argparse.ArgumentParser("")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-s", "--seed", type=int, help="The seed to use for generation.")
    group.add_argument("--rand", action='store_true',help="Use the system RNG instead of --seed.")

    parser.add_argument("-o","--outfile",type=str,help="Set the output filename.",default="out.gif")
    parser.add_argument("-w","--width",type=int,help="Set the output width.",default=1024)
    parser.add_argument("-H","--height",type=int,help="Set the output height.",default=768)
    parser.add_argument("-d","--depth",type=int,help="Set the output color depth. (1-8)",default=8)
    args = parser.parse_args()

    if args.seed is None and not args.rand:
        raise ValueError("Please specify either -r/-R or an explicit --seed argument." % sys.argv[0])

    rng = None
    if args.rand:
        try:
            from os import urandom as rngsource
        except ImportError as ex:
            raise NotImplementedError("System RNG not implemented for your platform." % sys.argv[0]) from ex
        else:
            def rng(): return rngsource(1)[0]
        
    generate_random_gif(rng,width=args.width,height=args.height,colordepth=args.depth,outfile=args.outfile)
    



