#!/usr/bin/env python3

import sys
import binascii
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, max_width=200, max_height=150
    ) -> None:

    # Write 6 bytes of magic and version information
    out.write(b"GIF89a".encode('ASCII'))

    # Determine width and height in pixels for the image
    width      = (rng.randint(1,max_width)) & 0xFFFF
    height     = (rng.randint(1,max_height)) & 0xFFFF
    size       = ((width * height) >> 3) & 0xFFFFFFFF
    palette    = []

    if width > 0x7FFF or height > 0x7FFF:
       raise ValueError("Image is too big!")

    # Write out the global colour table
    num_colours   = min([int((rng.randint(1,255))&0xFF), 256]) & 0xFF
    out.write(((num_colours << 4) | 1).to_bytes(1,"little"))
    for i in range(0,num_colours):
        r   = (rng.randint(0,255)&0xFF)
        g   = (rng.randint(0,255)&0xFF)
        b   = (rng.randint(0,255)&0xFF)

        palette += [r,g,b]

    out.write(bytes(palette))
    out.write([(size%0xFF), ((size>>8)%0xFF)])


    # LZW compress the image data
    lzw_codebook = dict()
    lzw_dict = dict()
    lzw_next_code = 258
    output = []

    while len(lzw_dict) < 4096:
        next_byte = rng.read(1)[0]
        if next_byte not in lzw_dict:
           lzw_dict[next_byte] = lzw_next_code
           lzw_next_code += 1
    
    for x in range (0,size+1):
        cur_byte = -1;
        try:
            cur_byte = rng.read(1)[0]
        except Exception as e:
            print(e,file=sys.stderr)
        
        lzw_input = False
        if (cur_byte >= 0xD8 and cur_byte <= 0xFB) or \
                (cur_byte >= 0xFC and cur_byte <= 0xFE):

            if cur_byte == 0xD9:
               break
            
            pass_through = True

        elif cur_byte in lzw_dict:

            lzw_input = cur_byte

            if lzw_next_code < 4096:
                lzw_codebook[lzw_next_code] = lzw_input + \
                        [(lzw_dict[lzw_input][len(lzw_dict[lzw_input])-1])<<8]
                lzw_next_code += 1


        else:
           lzw_dict = dict() 
           lzw_next_code = 258
           continue

        while len(lzw_dict) < 4096:
            #print("Len ",len(lzw_dict)," Cur code ", lzw_input)
            if lzw_input in lzw_dict:
               lzw_input = lzw_dict[lzw_input].copy()
               if lzw_next_code < 4096:
                  lzw_codebook[lzw_next_code] = lzw_input + \
                          [(lzw_dict[lzw_input][len(lzw_dict[lzw_input])-1])<<8]
                  lzw_next_code += 1
            else:
               break 

        lzw_dict[(lzw_input)] = lzw_input + \
                                  [(lzw_dict[lzw_input][len(lzw_dict[lzw_input])-1])<<8]

        if lzw_next_code < 4096: 
            lzw_codebook[lzw_next_code] = lzw_input + \
                    [(lzw_dict[lzw_input][len(lzw_dict[lzw_input])-1])<<8]
            lzw_next_code += 1


        if lzw_next_code == 4096:
             lzw_codebook[lzw_next_code] = b'\xff\xff'
             lzw_codebook[lzw_next_code+1] = b'\xfe\x01'
             output += lzw_codebook[lzw_input]
             break

        output += lzw_dict[lzw_input]
     
    lzw_output = output
    lzw_output_final = b""
    last_value = 0xFFFE
    print(lzw_output)
    for c,v in enumerate(lzw_output): 
        if v == 0xFFFE:# or (c+1)==len(lzw_output):
            lzw_output_final += b"\xFF\x00"
            #print(f"{c}:{v} END")
            last_value = 0xFFFE
        else:
            #print(f"{c}:{v}")
            if v == 0xFFFD:
                lzw_output_final += b"\xFE"
                #print(f"END")
                last_value = 0xFFFE
            elif v == 0xFFFC:
                print(f"{c}:{v}")
                lzw_output_final += b"\xFD"        
                #print(f"END")
                last_value = 0xFFFE
            elif v!= 0xFFFA: 
                lzw_output_final += struct.pack("<H",last_value)
                last_value = v
            else:
                lzw_output_final += v
            
    print(lzw_output_final)
         
    print(len(lzw_output))
    print(len(lzw_output_final))
        
    if last_value!= 0xFFFE:
        lzw_output_final+=struct.pack("<H",last_value)



    # Write out the block denoting that it's compressed with LZW
    out.write(b'\x2C\x00\x00\x00\xF9\x04\x00\x00')
    out.write((((width * height) >> 3) & 0xFFF).to_bytes(2,'little'))

    # Write out the actual image
    out.write(b'\x08')
    out.write(b'\x02')
    out.write(width.to_bytes(2,'little') +
              height.to_bytes(2,'little'))
    out.write((0).to_bytes(1,'little'))
    out.write(lzw_output_final)
    out.write(b'\x00')
     



