#!/usr/bin/env python3

import argparse
import os
import sys
import struct
from typing import BinaryIO


# This function generates a random GIF file into `out` using the
# random number generator `rng`. If no `rng` is specified then it uses
# `/dev/urandom`, which should work on most systems.
def generate_random_gif(
        rng:        BinaryIO=None,
        width:      int    = 1024,
        height:     int    = 768,
        colordepth: int    = 8,
        bgcolor:    int    = None,
        outfile:    str    = "output.gif",
    ) -> None:

    try:
        from tqdm import trange
    except ImportError:
        def trange(_:int=None,_:str=None)->range: return range(0)

    # Set up the base colors that will be used to build our palette.
    if not bgcolor or bgcolor < 0 or bgcolor > 255**3:
        bgcolor = 0xFFFFFF ^ ((width*height)%(255**3))

    blackcode = bytes([bgcolor & (255**2)])
    whitecode = bytes([(bgcolor>>8)&255,(bgcolor>>16)&255])
    
    with open(outfile,'wb') as out:

        # The GIF header
        print("Generating GIF header...")
        version = b"89a"
        screendescriptor = struct.pack(">HHBBB",
                width,
                height,
                0xF0 | colordepth,
                0,
                0,
        )
        
        print("Writing header and descriptor block")
        out.write(b"GIF"+version+screendescriptor)

        # Build the color lookup table
        print("Building color table...",end="")
        colortable = []
        for i in trange((2**colordepth)-1):
            x = (i*(bgcolor**(1/(256.0))))%(256**3)
            colortable += [struct.pack(">BBB",
                    int(((x>>16)&255)**1.5),
                    int(((x>>8)&255)**1.5),
                    int((x&255)**1.5),
                ),
            ]
        colortable += [blackcode]
        colortable += [whitecode]
        print(" Done.")

        print(f"Writing {len(colortable)} colors to Gif... ",end="")
        out.write(b"".join(colortable))
        print("Done.")


        # Now we need to write the image data itself...
        print("Writing image information...",end="")
        out.write(b"\x2C\x00\x00\x00\x00\x00\x00\x00\xFF")   # Image descriptor
        out.write(b"!\xF9\x04\x01\x00\x00\x00,\x00\x00\x00") # Graphics Control Extension
        out.write(b"\x00\x01\x00\x21\xFE\x09D\x0B\xCF\x00") # Application Extension
        out.write(b"\x0BNETSCAPE2.0\x03\x01")                  # Application Data
        out.write(b"\x00\x00\x00!")                           # Trailer
        print(f"Wrote image. ({width}x{height})")