#!/usr/bin/env python

"""
Generate a random GIF from /dev/urandom.
"""

import struct
import sys
from argparse import ArgumentParser
from typing import BinaryIO


def parse_args() -> str:
    """Parse CLI arguments."""

    parser = ArgumentParser("generate-random-gif")

    # Allow passing an argument to specify output filename.
    parser.add_argument("-o", "--output", type=str,
                        help="set destination for generated image (default: %(default)s)")
    ns = parser.parse_args()

    if not ns.output:
        raise RuntimeError('No output specified')

    return ns.output
    

# Generates a random GIF file with size `width x height`.
# Uses `rng` and writes result in `out`.
def generate_random_gif(
        width: int, height: int, rng: BinaryIO, out: BinaryIO
    ) -> None:
    # Header of GIF files - see https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    header = bytes([
        0x47, 0x49, 0x46, 0x38, 0x39, 0x61,   # Signature
        0x00, 0x20,                         # Logical screen descriptor
        0x00, 0x00,                         # Global color table length
        0xFF,                               # Background color index
        0x00                                # Pixel aspect ratio
    ])

    # Compute size of local color table (always zero here since we are encoding in grayscale.)
    gctsize = b""
    offbits = b""
    lsd = header[6:] + gctsize + offbits + struct.pack(">HHB", width, height, 8)[::-1]

    # Write logical screen descriptor followed by global color table
    out.write(lsd)


    # Size of the remaining data block after logical screen descriptor and global color table
    bodylen = struct.pack("<I", width * height)
    num_colors = struct.pack("<H", 256)
    palette = [struct.pack("<BBB", c, c, c) for c in range(256)]
    palletesize = struct.pack("<B", len(palette))

    cols = b""
    for i in range(len(palette)):
        cols += palette[i]

    # Local color table
    lct = b"\x00\x00" + num_colors[::-1] + palletesize + cols

    alphasize = struct.pack("<B", 0)

    # Image descriptor
    idd = b"\x2C" + b"\x00\x00" + b"\x00\x00" + b"\x00\x00" + b"\x00\x00" + b"\x00" + alphasize

    # LZW minimum code size determined by C standard.
    mincodesize = b"\xF0"


    # Table which maps each byte to its prefix code.
    prefixmap = [[p] for p in range(256)]

    # Frequency of each prefix code.
    freqs = []
    for idx, v in enumerate(prefixmap):
        freqs.append((idx+1)*8)

    minfreq = sorted(freqs)[0]

    lzwminsize = b"\xf0"

    # Data section contains a sequence of sub-blocks. Each subblock starts with one or more control codes.
    # The first code determines what kind of other codes follow it. See spec.
    datasize = b"\xff\xfe"
    databytes = [struct.pack("<BB", symb, code) for code, symb in enumerate(range(256))]
    datasection = b''.join(databytes)

    # Trailer of GIF images, always same as below.
    trailer = b'\x3b'    

    # Join sections together
    body = datasize + datasection + lzwminsize

    body = body + \
           (b"".join(struct.pack("<B", c) for c in line)) + \
            lzwminsize + \
             trailer

    out.write(body)


    

