#!/usr/bin/env python3

import io
import os
import sys
import tempfile
import unittest
import unittest.mock

import colorama
import termcolor

import make_test_suite
import palint
import run_tests



class TestRandomGifGenerator(unittest.TestCase):


    def setUp(self) -> None:

        self._tempdir = tempfile.TemporaryDirectory()
        self._path = f'{self._tempdir}/test.gif'


    def tearDown(self) -> None:

        self._tempdir.cleanup()

    
    @property
    def path(self) -> str:
        
        return self._path
    

    def test_generate_random_file(self) -> None:
    
        with open('/dev/urandom', mode='rb') as infp:

            with open(self.path, 'wb') as outfp:

                size = 2048
                generate_random_gif(
                    size=size, 
                    rng=infp, 
                    out=outfp
                )

        self.assertEqual(os.stat(self.path).st_size, size + 25)
        
        with open(self.path, mode='rb') as infile:
        
            self.assertFalse(palint.is_binary(infile))
            
            header_data = infile.read(6)
            self.assertEqual(header_data, b'GIF89a')

            # Data Len
            infile.seek(6, os.SEEK_CUR) 
            self.assertEqual(infile.read(1)[0], 2)

            # Frame Size
            infile.seek(1, os.SEEK_CUR)
            frame_width = int.from_bytes(infile.read(2), byteorder='little')
            frame_height = int.from_bytes(infile.read(2), byteorder='little')

            # Color Table
            infile.seek(2, os.SEEK_CUR)
            palette_table_len = int.from_bytes(infile.read(1), byteorder='little') * 3
            infile.seek(palette_table_len, os.SEEK_CUR)

            # Image Descriptor
            infile.seek(9, os.SEEK_CUR)
            self.assertEqual(frame_width, int.from_bytes(infile.read(2), byteorder='little'))
            self.assertEqual(frame_height, int.from_bytes(infile.read(2), byteorder='little'))

            # Local Color Table Flag
            infile.seek(1, os.SEEK_CUR)
            local_color_table_flag = infile.read(1)[0]

            # Reserved
            infile.seek(1, os.SEEK_CUR)

            self.assertTrue((local_color_table_flag & (1 << 7))) # Global Color Table Flag

            # LZW Min Code Size
            infile.seek(1, os.SEEK_CUR)
            lzw_min_code_size = infile.read(1)[0]
            
            infile.seek(-lzw_min_code_size - 1, os.SEEK_END)
            compressed_image_data = infile.read()[::-1][:-lzw_min_code_size]

        with open(self.path[:-4] + '.dat', mode='wb') as datfile:

            datfile.write(compressed_image_data)

        output = io.BytesIO()
        result = run_tests.run([sys.executable, '-m','make_test_suite'], stdout=output)
        self.assertEqual(result, 0)

        testsuite = output.getvalue().decode()
        nr_assertions = sum(line.startswith('assert True == ') for line in testsuite.splitlines())
        self.assertEqual(nr_assertions, len(compressed_image_data))


def main():

    try:
        colorama.init()
        unittest.main(verbosity=2, exit=False)
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()