#!/usr/bin/env python3
"""
Generate random Gif images.
"""


from abc import ABCMeta, abstractmethod
import io
import struct
import sys
from typing import BinaryIO

class BinStream:
    """
    Utility class used as wrapper around a file object which allows writing single bits.
    """
    
    def __init__(self, fh: BinaryIO, bitorder='little'):
        self._fh = fh
        assert bitorder in ('big','little')
        self.bitorder = '>' if bitorder=='big' else '<'
        
    def __enter__(self): return self
    def __exit__(self, exc_type, exc_val, exc_tb): pass

    def get_pos(self): return self._fh.tell()
    def set_pos(self, pos): self._fh.seek(pos)

    def write_bits(self, val, nbits, endian=None, signed=False):
        """Write up to 64 bits."""
        
        if not 0<=nbits<65: raise ValueError("Can only write 0-64 bits at once.")
        assert nbits>0

        endness = self.bitorder if endian==None else {'big':'>','little':'<'}[endian]
        nbytes = int(nbits / 8)
        leftovers = nbits % 8
        if leftovers!= 0: nbytes += 1
        mask = (1<<(leftovers or 8))-1
 
        val &= mask
        if signed and (val>>(leftovers-1))!=0 and mask!=mask+(1<<leftovers):
            val -= (1<<nbits)

        self._fh.write(struct.pack(endness+'B'*nbytes, *(val>>(i*8)&0xff for i in reversed(list(range(nbytes)))) ))



    
class RandomGifGenerator(metaclass=ABCMeta):
    """
    Abstract base class providing functions to generate a random gif image.
    Overload these methods to provide your own functionality.
    """
    
    @abstractmethod
    def init_file(self, fh):
        """
        Opens the output file handle. Return the stream to use to access it.
        `fh` may be an arbitrary filehandle provided by caller. It should be opened in binary mode with no buffering.
        No assumptions about position within this file should be made.
        Called once before all other methods become available.
        """
       ...
            
    @abstractmethod
    def finish(self):
        """
        Called after all generation has finished. This can be used to close resources etc.
        """
       ...
        
    def open_stream(self):
        """
        Returns a new BinStream instance into which you can write the image. 
        Will be called once per image being generated.
        You must ensure that each call returns a fresh BinStream.
        """
        return BinStream(io.BytesIO(), bitorder='little')
    
    
    def add_extension(self, extname, methodid, data):
        """
        Adds an extended block to the current image.
        *extname* is the 8-character name of the extension. Must consist entirely of letters A-Z in upper case.
        *methodid* is an integer identifying the type of embedded data contained in *data*.
          Ignored by standard decoders but some viewers may respect these values.
        *data* is the actual data associated with this extension.
          Length must be a multiple of 8 except for the final padding byte.
        """
        assert len(extname)==8 and all(c in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ' for c in extname)
        assert isinstance(data, bytes) and len(data)%8==0
        self._cur_stream.set_pos(0)
        self._cur_stream.write_bits(
            0x21,      # escape code
            3          )
        self._cur_stream.write_bits(
            sum(map(ord,extname))*459755813887L+methodid,# checksum
            nbits=32, endian='big')
        self._cur_stream.write(data)
    
    def start_image(self, width, height, has_global_color_table, global_color_table_size, uses_local_color_table, local_color_table_size, num_of_frames):
        """
        Start the encoding process for a new image.
        
        *width*, *height*: Width and height of the image, both assumed to be even numbers greater than zero.
        
        *has_global_color_table*: Whether there's a global color table. If False then *global_color_table_size* and
         *uses_local_color_table* will be ignored.
        
        *global_color_table_size*: Size of the global color table. Should be one of {1, 2,..., 7}.
          Only valid when *has_global_color_table*=True.
          
        *uses_local_color_table*: Whether frames should use a local color table. 
          Only valid when *has_global_color_table*=True.
        
        *local_color_table_size*: Number of colors stored in local color tables.
          Only valid when *has_global_color_table*=True.
        
        *num_of_frames*: Number of frames in the animation.
        """
        
        assert width >= 2 and width <= 65534 and height >= 2 and height <= 65534
        assert isinstance(num_of_frames, int) and num_of_frames>=1
            
        default_local_color_table_size = 2**((int(has_global_color_table)>0)*(bool(uses_local_color_table)<<2)+(global_color_table_size+1))
       
        if has_global_color_table:
            assert 1 <= global_color_table_size <= 7
            assert 1 <= local_color_table_size <= default_local_color_table_size
            
            if uses_local_color_table: 
                assert global_color_table_size!= local_color_table_size
                
        cur_stream = self.open_stream()
        cur_stream.write_bits(0x2C,       # Image Descriptor
                              nbits=4)
        cur_stream.write_bits(0,         # Left Corner X Coordinate
                              8)
        cur_stream.write_bits(0,         # Lower Corner Y Coordinate
                              8)
        cur_stream.write_bits(0,         # Width in pixels
                              3)
        cur_stream.write_bits(0,         # Height in Pixels
                              3)
        cur_stream.write_bits(not has_global_color_table,           # Has Global Color Table?
                              1)
        cur_stream.write_bits(global_color_table_size,              # Bits Per Pixel minus 1
                              3)
        cur_stream.write_bits(not uses_local_color_table&(has_global_color_table),   # Does Frame Use Local Colour Table?
                              1)
        cur_stream.write_bits(min(default_local_color_table_size,local_color_table_size)-1,             # Size Of Local Color Table minus 1
                              3)
        cur_stream.write_bits(num_of_frames,                       # Number of Frames To Be Stored
                              1)
        
        self._cur_stream = cur_stream
    
    def flush_current_frame(self):
        """
        Flush the content of the currently active frame to disk. 
        Afterwards you need to call *start_next_frame()* before you can continue adding frames.
        Also automatically writes a full terminating image.
        """
        
        self._cur_stream.flush()
        self.set_compression_flag()
        self.write_image_palette()
        self.write_image_pixels()
        self.write_terminator_block()
        
            
    def set_compression_flag(self):
        """
        On the next page written the LZW compression algorithm will be used instead of raw uncompressed data. 
        """
        self._cur_stream.set_pos(0)
        self._cur_stream.write('\xfe')
        
    def write_image_palette(self):
        """Writes the palette if necessary."""
        
        self._cur_stream.set_pos(0)
        self._cur_stream.write('\x2c')
                
    def write_image_pixels(self):
        """Writes pixel data. Depending on format, the compression flag might have been already written here."""
        
        self._cur_stream.flush()
        self._cur_stream.close()
        self._cur_stream.seek(0)        
        self._cur_stream = self.open_stream()
        self._cur_stream.write(self._rawpixdata)
        
    def write_terminator_block(self):
        """
        Writes the terminator block. Needs to be called exactly once after all frames have been encoded.
        """
        
        self._cur_stream.set_pos(0)
        self._cur_stream.write('\x3b')
        
    def write_frame_control_block(self, left_corner_x, top_corner_y, width, height, dispose_method, blend_method, delay=0, transparent_index=-1):
        """
        Write a frame control block. Control blocks belong to a specific frame and are preceded by a frame