#!/usr/bin/env python3

import io
import math
import struct
from typing import Tuple, Optional


# Global constant for maximum possible LZW dictionary size
MAXDICTSIZE = 4096

# Global constants representing the various fields in the GIF file spec.
HEADER_SIZE        = 6
LOGICALSCREENDESC  = 7
IMAGEDESC          = 9
APPEXT             = 11
NETSCAPECTRLBLOCK  = 13
NETSCAPENEXTBLK    = 14
GRAPHICCTLBLOCK    = 15
IMGDATA            = 16
TRAILER            = 17



# Generates a new GIF by copying frames from `source` into `destination`, injecting text metadata at random intervals while doing so.
def generate_random_gif(
        source: BinaryIO,
        destination: BinaryIO,
        framecount: int,
        textdelays: Tuple[Optional[float],...]):

    # Store the last position where data was written, in case we want to read back from here again during generation.
    lastpos = len(destination.getbuffer())

    # Make sure there's enough space left for any of this stuff
    destination.seek(lastpos+framecount*trailerlen()+(len(textdelays)-1)*appextlen())

    # Set up some empty stores to keep track of what's been generated thus far
    imagesgenerated = []
    textsinserted = [False] * len(textdelays)
    framenumbers  = list(range(0,framecount))

    validkeyframe = True

    try:

        # Loop through every frame in the animation
        while len(imagesgenerated) < framecount:

            # If the last frame had no keyframe, then the next one won't either
            if not validkeyframe:
                validkeyframe = False

                # Keep looping until we encounter a keyframe or run out of frames
                while not validkeyframe and len(imagesgenerated) < framecount:
                    print("No keyframes found; skipping ahead")

                    # Read the current frame number
                    currframe = int.from_bytes(source.read(1),'little')
                    if source.read(1)!=b'':
                        print("Invalid frame encountered")
                        raise Exception()

                    # Check whether it's a keyframe or not
                    if (currframe&1)==0:
                        print("Frame {:d} was not a keyframe.".format(currframe>>1))

                        # Don't increment the counter since we didn't actually insert anything!
                        framenumbers.append(framenumbers[-1]+1)

                        continue

                    # Mark this as a keyframe and keep going
                    validkeyframe = True


            # Case 1: The previous frame did have a keyframe but we've already injected the necessary delay for its display time. Just copy the old frame over silently
            if validkeyframe and textsinserted[len(textsinserted)-1]:
                print("\nCase 1: Injected delay time for frame {:d}".format(framenumbers[-1]))

                # Copy over the raw frame from before the text injection
                destination.write(source.read())

                # Add another frame to our internal store of "generated" frames
                imagesgenerated.append(source.tell()-1)


                # Now we're done processing this frame, so move on to the next one
                continue


            # Case 2: This is the first frame and doesn't have a keyframe; skip it entirely
            elif validkeyframe==False and len(imagesgenerated)==0:
                print("\nCase 2: First frame has no keyframe.")
                
                # Skip forward past the whole frame
                destination.seek(destination.tell()+trailerlen())

                # Increment the frame number manually
                framenumbers.insert(0,0)
                continue


            # Case 3: An invalid frame was detected earlier; make sure the code gets triggered.
            assert False


        # Finished generating the entirety of the image without running out of frames or bad frames; do some final checks.
        print("Finished parsing GIF successfully!")

        # Update the application extension block after the last frame in the animation 
        update_appext(destination, 
                    appextlen(),
                    len(imagesgenerated)+1, 
                    "".join([str(t) if t is not None else '' for t in textdelays]), 
                    0,
                    0,
                    0)

        # Seek back to the end and chop off the excess bits from the end of the animation
        destination.truncate()


    except AssertionError:
        print("Detected unexpected error; rolling back changes to output GIF...")

        # Remove the latest frame from the lists and seek backwards to overwrite it.
        exportedimages = [f for f in imagesgenerated[:-1]]
        trailerstart = len(exportedimages)*trailerlen()+lastpos

        # Rollback the final assembly of the application extension block
        appextblock = ""
        for i in range(-1,-len(textdelays)-1,-1):
            if textdelays[i]==None:
                break

            appextblock += str(textdelays[i])+"\0"

        appextblock+="\0\0\0\0"
        rollback_bytes(destination, appextlen(), appextblock)

        # Truncate the GIF to remove everything added beyond this point...
        destination.truncate(trailerstart)
        destination.flush()

        # Move everything back one place (the beginning of the newly truncated section)
        rollback_bytes(destination, lastpos-(destination.tell()), io.BytesIO().getvalue())

        # Close the destination file handle
        destination.close()
        
        raise
        
    
    finally:
        pass




# Reverts writes made to `stream` by shifting them backward by `amount`.
def rollback_bytes(stream: BinaryIO, amount: int, backupdata: bytes):

    # Get the starting position in the stream to roll back to
    startpos = stream.tell()-amount

    # Shift everything left by one byte and replace whatever used to be there
    stream.seek(startpos)
    stream.write(backupdata)
    stream.seek(startpos)







#### Utility functions ####


# Calculates the length of a standard trailer excluding the IEND chunk.
def trailerlen():
    return HEADER_SIZE+LOGICALSCREENDESC+APPEXT+NETSCAPECTRLBLOCK+NETSCAPENEXTBLK+GRAPHICCTLBLOCK+IMGDATA





# Updates the contents of the specified application extension block in the provided GIF binary.
def update_appext(out: BinaryIO, 
            appextsize: int, 
            numframes: int,
            textdelays: str, 
            loopcount: int=0, 
            bgcolorindex: int=-1, 
            pixelaspectratio: float=0.0):

    # Check that the requested app ext size makes sense
    assert appextsize==appextlen()

    # Go straight to the start of the application extension block
    out.seek(-appextsize, 2)

    # Store the current position in the stream
    start = out.tell()

    # Write the frame count
    out.write(numframes.to_bytes(2,"little"))

    # Write the background color index
    if bgcolorindex>=0:
        out.write(bgcolorindex.to_bytes(1,"little"))
    else:
        out.write(b"\xff")

    # Write the pixel aspect ratio
    if pixelaspectratio>0.0:
        out.write(math.ceil(pixelaspectratio).to_bytes(1,"little"))
    else:
        out.write(b"\xff")

    # Pad the rest of the block with null characters
    out.write(b"\0"*10)

    # Write the actual text delays
    for t in textdelays.split("\0"):
        if t!='':
            out.write(int(round(float(t))).to_bytes(2,"little"))

    # Write the loop count
    out.write(loopcount.to_bytes(2,"little"))

    # Work out how many additional bytes were needed based on the input string
    paddingneeded = 0
    for t in textdelays.split("\0"):
        # Round the delay value down
        v = round(float(t)/1000)

        # Calculate how many bytes of storage would be required by this entry
        if v<256:
            paddingneeded+=1
        elif v<65536:
            paddingneeded+=2
        else:
            paddingneeded+=4

    # Pad the remaining bytes using null characters
    out.write(b"\0"*paddingneeded)

    # Write the correct length of the block and return to the end of the stream
    out.write(appextlen().to_bytes(2,"little"))
    out.seek(start)







