#!/usr/bin/env python3

import struct
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The output is guaranteed to be valid, but it may not have any visible content
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    w = int.from_bytes(rng.read(2), byteorder='big')
    h = int.from_bytes(rng.read(2), byteorder='big')
    header += struct.pack('>H', w)
    header += struct.pack('>H', h)

    # color table (RGB triplets)
    if w == h == 1:
        raise ValueError("Invalid image size")
    elif w > h:
        ctlen = min((w+h-1)*3//4, 7*3)
    else:
        ctlen = max(((w+h+1)//2)*3//4, 7*3)
    while len(header) < 10 + ctlen:
        header += bytes([rng.randint(0, 255)])

    # data block length -- 1-byte integer (MSB first order)
    rndblocksize = rng.randrange(0, 7)
    if rndblocksize!= 0:
        rndblocksize |= 0b10000000
    header += bytes([rndblocksize])

    # LZW minimum code size - 1-byte integer (MSB first order)
    lzwmincodesize = rng.randrange(2, 9)
    header += bytes([lzwmincodesize])

    # Data section: up to 2^16 bytes of raw data
    rnddatalen = rng.randrange(0, pow(2, 16)-1)
    if rnddatalen!= 0:
        rnddatalen |= 0b1000000000000000
    header += rnddatalen.to_bytes(length=2, byteorder='big')
    
    # Fill with zeroes until end of block
    padding = 2*(pow(2, 16)-1) - rnddatalen
    header += b'\x00'*padding

    # Write everything
    out.write(header)
    out.write(rng.read(rnddatalen))

