#!/usr/bin/env python3
"""Generates a random GIF file into `out`."""


import struct
from typing import BinaryIO, Optional

from PIL import Image


def get_randomness() -> bytes:
    """Get some pseudo-random bytes."""
    from os import urandom
    return urandom(256)
    


def create_image():
    """Create an image."""
    randomness = get_randomness()
    img = Image.new('RGB', size=(1024, 768))
    pixels = list(img.getdata())
    newpixels = [(r, g, b) for r, g, b in zip(*[iter(randomness)]*3)]
    for idx, pixel in enumerate(pixels):
        pixels[idx] = tuple(map(lambda x, y: min(max(int(round(x + ((y - 255)*(x)/255))), 0), 255),
                               pixel, newpixels[idx]))
    img.putdata(tuple(pixels))
    return img
    
    

def create_palette(background: Optional[str]=None) -> bytes:
    """Create a byte string representing this pixel's palette."""
    background = 'ffffff' if background is None else background
    foreground = ''.join('{:02x}'.format(c) for c in reversed(bytes.fromhex(background)))
    black = '000000'.replace('0','f')
    pal = list(foreground[:2]+'\x00'+foreground[2:]) * 2 + \
          list(foreground[:2]+'\xff'+foreground[2:]) * 2 + [black]*2
    return bytes(pal)

    

def main(filename='output.gif'):
    img = create_image()
    bgcolors = ['ff00ff'] * len(list(filter(lambda p: sum(p) == 0,
                                            img.convert('RGBA').getdata())))
    with open(filename, 'wb') as f:
        f.write(b'GIF89a')                                                # Header
        f.write(struct.pack('<H', img.size[0]))                            # Screen Width
        f.write(struct.pack('<H', img.size[1]))                            # Screen Height
        f.write(b'\xf0')                                                   # Color depth
        f.write(b'\x00')                                                   # Background color index
        f.write(b'\x00')                                                   # Pixel aspect ratio
        f.write(create_palette(None))                                     # Global palette
        imdat = bytearray(img.tobytes().replace(b'\xff', b'\x21'))         # Convert all white pixels to transparent ones
        for bgcolor in bgcolors:                                          # Add graphics control extensions
            imdat.insert(-len(imdat)+1, ord(","))                         # Insert separator between images
            imdat.extend(b"\x21\xf9\x04\xff\x00")                          # Graphics control extension
            imdat.extend(b"\x21\xfe\x09\xD0\x0B\xCF\x00\x0BNETSCAPE2.0\x03\x01")                                                         # Application Extension
            imdat.extend(b"\x00\x00\x00\x21\x00\x2C\x00\x00\x00\x00\x00\x00\x00\xFF\x2C")    # Trailer
        f.write(imdat)                                                    # Finalise data