#!/usr/bin/env python3

"""
Generate a random GIF from /dev/urandom.
"""

import pathlib
import struct
import sys
from argparse import ArgumentParser, FileType
from typing import BinaryIO


def resolve_path(p: str):
    """Ensure that path `p` exists."""
    try:
        p = pathlib.Path(p)

        if not p.exists():
            raise ValueError(f"{p} does not exist.")
        
        elif hasattr(p, 'is_dir') and p.is_dir():
            raise ValueError(f'{p} is a directory.')
        
        else: 
            return p.resolve()

    except OSError as e:
        msg = f'The following error occurred trying to resolve "{p}"'
        raise EnvironmentError(msg) from e


def parse_args():
    parser = ArgumentParser("generate-random-gif")

    # Generate a default name if none is provided
    if len(sys.argv) == 1:
        return resolve_path("random.gif"), None

    # Allow passing an argument to specify output filename.
    parser.add_argument("-o", "--output", type=FileType('wb'), 
                        help="set destination for generated image (default: %(default)s)",
                        metavar="<file>")
    ns = parser.parse_args()

    if not ns.output:
        raise RuntimeError('No output specified.')
    
    return resolve_path(ns.output.name), ns.output.fileno()
    

# Generates a random GIF file in `outfile`.
def generate_random_gif(outfile: BinaryIO, outfilefd: int):
    with open("/dev/urandom", mode='rb') as rng,\
         open("/dev/stdout", mode='ab+') as out:
        header = bytes([0x47, 0x49, 0x46, 0x38, 0x39, 0x61]) 
        datalen = int.from_bytes(
            rng.read(3), byteorder='big', signed=False)
        print(f"[FD {outfilefd}] Datalen: {datalen}", flush=True)
        data = rng.read(datalen)
        out.write(header + struct.pack(">H", datalen))
        out.write(data)
        out.flush()
        os.fsync(outfilefd)

    