#!/usr/bin/env python3

import struct
import sys

from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):

    # The first three bytes are always "GIF". Read them
    # to determine how many more bytes we need to read for
    # the rest of the file and assign them to `datalen`.
    header = b'GIF'
    datalen = 0
    headerlen = len(header)
    try:

        while True:

            chunk = rng.read(1)
            if chunk == '':
                break;            # EOF was reached

            elif not (header[:headerlen-1] <
                    chunk[0:headerlen-1]):

                pass              # Skip until we get to the start
            else:                  # We're at the beginning now

                if chunk[-1:] == b'\x2c':
                    datalen += 1   # This is a frame. Add one to `datalen`.
                    continue        # Continue processing this loop iteration.

                elif chunk[-1:] in [b'\x21', b'\x2e']:

                    datalen += 1   # This is an application extension or graphic control extension.
                                    # Add one to `datalen`.
                    continue

                elif chunk[-1:] in [b'\x3b', b'\x24']:
                    datalen += 1   # This is the end marker or image separator.
                                    # Add one to `datalen`.
                    break           # Exit processing this loop iteration immediately.
                
                raise ValueError("Unexpected value {chunk} found.".format(
                        chunk=chunk))
    
    except KeyboardInterrupt as e:
        
        print()
        exit(-1)
                
    finally:
    
        # Read and output the contents of the file.
        data = rng.read(datalen)
        out.write(header + data)


if __name__=='__main__':

    with open('/dev/urandom', 'rb') as urand:
        generate_random_gif(urand, sys.stdout.buffer)
