#!/usr/bin/env python3
"""A generator of GIFs with predictable content."""


import struct
from typing import BinaryIO

MAXWIDTH = 1599
MAXHEIGHT = 1599


def generate_random_gif(rng: BinaryIO, out: BinaryIO,
                        minwidth: int = 32, maxwidth: int = MAXWIDTH,
                        minheight: int = 32, maxheight: int = MAXHEIGHT,
                        nframes: int = 10) -> None:
  """Generates a random GIF image in ``binary`` format to ``out``, using ``rng``.

  Args:
      rng: A stream producing pseudo-random bytes.
      out: Where to write the result.
      minwidth: Minimum width constraint on frames.
      maxwidth: Maximum width constraint on frames.
      minheight: Minimum height constraint on frames.
      maxheight: Maximum height constraint on frames.
      nframes: Number of frames.
  """
  # construct headers
  header = b'GIF87a'
  width = rng.randint(minwidth, maxwidth - 1) // 2 * 2 + 2
  height = rng.randint(minheight, maxheight - 1) // 2 * 2 + 2
  gctsize = 2 << bool(rng.randint(0, 1))
  globalcolortablelen = 2 ** (gctsize + 1) - 1
  localcolortablelen = 2 ** ((bool(rng.randint(0, 1))) << 2)
  fc = rng.randrange(nframes) + 1
  extblocks = [(b'NETSCAPE2.0', b'\x03\x01\x00\x00\xff\x00'),
               (b'ANIMEXTS1.0', b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')]


  # write headers
  out.write(header)
  out.write(struct.pack('<HHB', width, height, gctsize))

  # append colortables if any
  for i in range(globalcolortablelen//3 + bool(globalcolortablelen%3)):
      r,g,b = tuple(int(v*255) for v in rng.choice(range(256)).to_bytes(length=3, byteorder="little"))
      out.write(struct.pack('BBB', r,g,b))
  if localcolortablelen > 0:
      for i in range(localcolortablelen // 3 + bool(localcolortablelen % 3)):
          r, g, b = tuple(int(v * 255) for v in rng.choice(range(256)).to_bytes(length=3, byteorder="little"))
          out.write(struct.pack('BBB', r, g, b))

  # append image descriptor block
  out.write(struct.pack('BBHBBBB',
                      0xf7, # separator
                      0,    # left corner x coordinate
                      height&0xff,   # lower corner y coordinate
                      ((height>>8)&0x0f)<<4 | \
                        (width&0x0f)>>2,  # size flag
                      (fc & 0xf)|((((width&0x0f)<<2)&0xf0)),
                          0,
                          0))               # no local color table

  # write image data blocks
  lastbyte = None
  for _ in range(fc):
      for lineindex in range(((width+7)//8)*height):
          byte = rng.getrandbits(8)%256
          if lastbyte!= None:
              while rng.randint(0,2)==0:
                  byte |= lastbyte
          lastbyte = byte ^ (lineindex==0) ^ ((lastbyte >> 7) & 0xff)
          out.write(struct.pack('B',byte))

  # copy and extend extension blocks
  for name,data in extblocks:
      data += bytes([sum(map(ord,name))+len(data)])*(8-(len(data)%8))
      out.write(data)