#!/usr/bin/env python3

import sys
import struct
from typing import (BinaryIO, Iterator)


class Generator(object):

    # The length in bits of each "subblock" of the generated value.
    # This is equivalent to an integer being represented using the
    # smallest possible size where all the integers fall within their
    # respective range. Eg. If range is [1..8] then this would be 3, as
     # it's enough for values up to 7 and we'll need one extra bit to
    # represent that last number.
    SUBBLOCKLEN:int = 24

    def __init__(self, minval:float=0.0, maxval:float=1.0):
        assert isinstance(minval, float)
        assert isinstance(maxval, float)

        if minval >= maxval:
            raise Exception("Invalid bounds")
        self._min:float = minval
        self._max:float = maxval

    @property
    def range(self) -> tuple[float]:
        return (self._min, self._max)

    def num_of_bits(self, val:float) -> int:
        """
        Returns the appropriate number of bits required to fit ``val``.

        Takes into account the minimum allowable value and whether or not
        the value should be rounded off to the nearest whole number.
        """
        if (not self._is_int(val)) or (self._should_round(val)):
            return self.__class__.SUBBLOCKLEN - 1
        else:
            return self.__class__.SUBBLOCKLEN


    def _is_int(self, val:float)->bool:
        return (val % 1 == 0)

    def _should_round(self, val:float)->bool:
        if self._min < 0:
            try:
                negval = round(-val) * (-1)
                if negval < self._min:
                    return True
            except OverflowError:
                pass
        return False

    def _pack_val(self, val:float, nbits:int) -> bytes:
        """
        Packs ``val`` into a binary string using the format specified by
         ``nbits``
        """
        return struct.pack('>I',
                          int((val / ((2 ** nbits) - 1)) * (2 ** 32)))[:nbits // 8]



# Given an iterable stream yielding bytes, unpack them according to the given
# generator object
def unpack(gen:Generator, inStream:Iterator[bytes]) -> Iterator[float]:
    while True:
        block = next(inStream)
        if len(block)!= gen.SUBBLOCKLEN // 8:
            break
        yield gen._unpack_block(block)


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()

    parser.add_argument('--minval', '-m',
                        type=float, default=0.0, dest="min",
                        help="Minimum value.")

    parser.add_argument('--maxval', '-M',
                        type=float, default=1.0, dest="max",
                        help="Maximum value.")

    args = parser.parse_args()

    g = Generator(args.min, args.max)

    with open(sys.argv[1], 'wb') as fOut, \
            open("/dev/urandom", "rb") as fIn:
        generate_random_gif(fIn, fOut, g)