#!/usr/bin/env python3

import os
import sys
import time
import struct
from enum import IntEnum
from hashlib import blake2s
from typing import Tuple, Union, Any, Optional, List, Dict, Iterable, Callable

from PIL import Image
from tqdm import tqdm

from santamatcher.datastructs import Team, MatchResult


class FrameFormat(IntEnum):
    RGB332             = 0
    RGB444             = 1
    RGB555             = 2
    ARGB6666           = 9
    RGBA6666           = 10
    ARGB8888           = 13
    RGBA8888           = 14



def read_team_colors() -> List[Tuple[str,Tuple[int,int,int]]]:
    """Read team colors from disk"""

    colorsfile = "assets/color-codes.txt"
    if not os.path.isfile(colorsfile):
        raise RuntimeError("{} does not exist! Run santamatcher/tools/generate-map-colors.py".format(colorsfile))

    validchars = bytearray([i for i in range(ord('A'), ord('Z')+1)])

    # Load in the file data
    with open(colorsfile,"rb") as f:

        # Verify first line is valid
        firstline = []
        while True:
            c = f.read(1).decode()
            if len(c)==0 or c=='\n': break;
            elif not c in validchars:
                raise ValueError("Invalid character '{}' in {}".format(c,colorsfile))

            firstline += [c]

        firstline = "".join(firstline)

        if "\t" in firstline:
            separator = '\t'
        elif "," in firstline:
            separator = ','
        else:
            raise ValueError("Could not find delimiter in {}, must be comma or tab separated.".format(colorsfile))

        # Read in the map color codes
        print("Reading team colors...")
        teams : List[Tuple[str,Tuple[int,int,int]]] = []
        for line in f.readlines():
            line = line.strip().decode().split(separator)
            teams += [(line[0], tuple([int(c,base=16) for c in line[1:]]))]

        return teams
    

def gen_match_frames(
        matchresults      : List[MatchResult],
        framespersecond   : float,
        resolution        : Tuple[int, int]) -> List['Tuple[Image.Image,float]']:
    """Generate all frames for a match given results and desired framerate."""

    # Divide the duration by the number of frames per second to get how long each frame should last
    seconds_perframe = 1 / framespersecond

    # Then round up to the nearest whole second so we have enough information when making the palette
    numpixelsperframe = resolution[0]*resolution[1] * 3
    maxframeduration = int(round(seconds_perframe*numpixelsperframe)/numpixelsperframe)

    # Now we can go through every single result and create a new frame every time we encounter a new one
    prevteam = None
    starttime = endtime = matchresults[0].timestamp
    currentframe : Image.Image = Image.new("RGB", resolution)
    frames : List['Tuple[Image.Image,float]'] = []

    # First, let's add any initial text
    draw = ImageDraw.Draw(currentframe)
    draw.text((0, 0 ), "Starting at {}".format(starttime.strftime("%I:%M %p")), fill="white")
    draw.text((0, 14), "{} has two goals".format(prevteam), fill="white")

    for resulthash in matchresults:

        if resulthash.team!= prevteam:
            
            # Add new frames between different teams
            elapsedtime = resulthash.timestamp - prevteam.lastresult.timestamp
            numframes = int(elapsedtime.total_seconds() / seconds_perframe)

            for _ in range(numframes):

                # Add a new frame
                frames += [(currentframe,endtime)]
                endtime += datetime.timedelta(seconds=maxframeduration)
                
                # New starting timestamp for next frame
                starttime = endtime
                
        # Update previous team info
        prevteam = resulthash

    # Finally, add an ending message for the final team
    draw = ImageDraw.Draw(currentframe)
    draw.text((0, 0 ), "Ending at {}".format(endtime.strftime("%I:%M %p")), fill="white")
    draw.text((0, 14), "{} has two goals".format(resulthash.team), fill="white")

    frames += [(currentframe,endtime)]

    return frames


def make_palette(images : List['Tuple[Image.Image,float]']) -> List[int]:
    """Take in a list of images and build a global palette based on their unique pixels."""

    # START NEW PALETTE
    palette : List[int] = []

    # For each image...
    for image, _ in tqdm(images, desc="Building palette"):
    
        # Find the unique pixels
        pxls = set([(r,g,b) for r,g,b in image.getdata()])
        
        # And add them to our palette if they aren't already there
        for p in pxls:
            if not p in palette:
                palette += [p]

    # END OF PALETTE MARKER
    palette += [2**24]

    return palette

def encode_image(im : Image) -> bytes:
    """Encode an image into GIF-friendly pixel data using minimal size LZW compression"""

    # First, we need to decide on a static dictionary size
    dictsize = 2  # Minimum required

    # Decode all the pixels so we can count them
    pxls = im.getdata()

    # Count the number of unique colors used across all frames
    colorsused = {pxl for pxl in pxls}
    colorsusedcount = len(colorsused)
    
    if colorsusedcount > dictsize:
        # If we're over then try to increase the dictionary size until it fits
        while colorsusedcount + 1 >= 2**(dictsize+1):
            dictsize+=1
        
    #print("Dictionary size:",dictsize)

    # Set up a new encoder instance
    enc = pylzw.encoder.LZWEncoder(width=8,dict_size=dictsize)

    # Initialize the buffer with the appropriate header
    buf = bytearray([FrameFormat.RGB555, dictsize])

    # Encode each pixel using our LZW encoder
    for pxl in tqdm(pxls, desc="Encoding"):
        rgb555 = (((pxl[0]&0xF8)>>3)<<10) + (((pxl[1]&0xF8)>>3)<<5) + ((pxl[2]&0xF8)>>3)
        buf += enc.encode(rgb555)[0]

    return bytes(buf)

def write_image_blocks(fout : BinaryIO, blocks : List[bytes]):
    """Write the content of the given blocklist as a series of sub-blocks"""

    # Iterate over the blocks and write out their sizes and contents
    for block in blocks:
        fout.write(len(block).to_bytes(1, 'big'))
        fout.write(block)
    
    
    
def write_logical_screen_descriptor(
        fout:          BinaryIO, 
        screensize:    Tuple[int,int],
        colortableloc: int):
    """Write the logical screen descriptor block"""

    # Logical Screen Descriptor
    # 0  1  2  3  4  5  6  7 
    # -------------------------------------------
    # x  x  x  Width Height Global Color Table Flag
    # x  x  x  Sort Flag Size of Global Color Table
    # RESERVED         Background Color Index
    # RESERVED         Pixel Aspect Ratio

    # Get the width/height values
    width, height = screensize

    # The global color table flag
    gctflag = 1 # Yes please :)

    # The background color (index 0)
    bgcoloridx = 0 # White?

    # The pixel aspect ratio (1:1)
    parval = 0

    # Pack it all together
    lsdblock = bytearray((width&0xff,(height>>8)&0xff,height&0xff))
    lsdblock += gctflag.to_bytes(1,'little')
    lsdblock += (colortableloc&0xff).to_bytes(1,'little')
    lsdblock += (bgcoloridx&0xff).to_bytes(1,'little')
    lsdblock += (parval&0xff).to_bytes(1,'little')
    
    # Write it out
    fout.write(lsdblock)

    

def write_global_color_table(fout: BinaryIO, colors : List[int]):
    """Write a global color table to the given binary output stream"""

    # Global Color Table is stored like so:
    # 0  1  2  3  4  5  6  7
    # -------------------------------
    # RED GREEN BLUE TRANSPARENCY FLAG
    # FIRST PIXEL IN TABLE
    # NEXT PIXEL IN TABLE

    # Write out the transparency flag (1 = transparant)
    fout.write(