#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generate a random GIF given an input stream of bytes to use for randomness and output stream to write to.
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    # Create the header
    header = b"GIF89a"

    # Get the length of the image from the RNG
    width  = (int.from_bytes(rng.read(1), 'big') >> 4) + 1
    height = ((int.from_bytes(rng.read(1), 'big') & 0xf) << 8) | int.from_bytes(rng.read(1), 'big')
    print("Generating {}x{} GIF".format(width,height))

    # Write it all together
    out.write(header)
    out.write(struct.pack('<HHB', width, height, 0xfc))

    # Background color index is just zero
    out.write((b'\x00'))
    out.write((b'\x00\xFF\xFF\xFF'))

    # LZW minimum size == 2
    mincodesize = 2
    out.write(mincodesize.to_bytes(1,'little'))
    
    # Palette contains one black color
    palette = [bytearray([0x00]), bytearray([0xff, 0x00, 0x00])]
    out.write(struct.pack('<B', len(palette)))
    for c in palette:
        out.write(c)

    # Global color table (GCT) is used; no images or other blocks use the global table except for the
    # first frame's palette data which is split into two non-global tables if necessary.
    gctused = True
    lctused = False
    gcteof = 0
    lcteof = 0
    gctsz = []
    lctsz = []

    curpos = out.tell()
    while not gcteof:
        blockid = int.from_bytes(rng.read(1),'big')
        if blockid!= 0x2C:
            break

        hsize = int.from_bytes(rng.read(1), 'big')
        left = hsize
        while left > 0:
            subblocklen = int.from_bytes(rng.read(1), 'big')
            left -= subblocklen+1

            if subblocklen & 1:
                raise Exception('Unexpected odd sized subblocks!')
            
            subblockdata = rng.read(subblocklen//2)
            left -= len(subblockdata)
            subblockcrc = int.from_bytes(rng.read(2), 'big')
        
            if blockid == 0x2C:
                if gctused:
                    if subblockid[0] < 0xF9:
                        continue

                    elif subblockid[0] == 0xFE:
                        gcteof = 0xfe

                    else:
                        gctsz += [(subblockid[0]&0xF)]

                else:
                    if subblockid[0] < 0xF9:
                        lctused = True
                    
                    elif subblockid[0] == 0xFE:
                        lcteof = 0xfe

                    else:
                        lctsz += [(subblockid[0]&0xF)]

        if gcteof:
            pass
        
        if lcteof:
            pass
        

    # Image flags are packed 1 bit each with as follows:
    #   1 - Color table sort flag
    #       0 - No global color table is present
    #       1 - The first frame's palette should be moved into the global color table instead
    #          since it is unsorted
    #   2-4 - GCT size (LZW minimum size)
    #   5   - Local colors sorted flag
    #       0 - Colors in the palette are in order of appearance
    #       1 - Colors in the palette are ordered according to frequency
    #   6   - Interlace flag (NOT USED)
    #       0 - Not interlaced
    #       1 - Interlaced
    #   7   - Sort flag (NOT USED)
    #       0 - Images will not be sorted
    #       1 - Images will be sorted
    #   8   - Reserved for future use

    # We set interlace off, color table sorted on, local colors sorted off,
    # and everything else off that we don't care about.
    gctsortflag = 1
    lctsortedflag = 0
    interlaced = False
    sortflag = 0
    reserved = 0
    imgflags = (gctsortflag << 7) + (lctsortedflag << 6) \
                + (interlaced << 5)     + (sortflag << 4) \
                + (reserved << 3)       + mincodesize
    out.write(imgflags.to_bytes(1, 'little'))

    return