#!/usr/bin/env python3

import struct
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (defaults to /dev/urandom;
# if an argument is provided it will be interpreted as a filename).
# The `--maxlen` option can be used to limit length of generated files.
if __name__ == '__main__':
    import argparse

    genlen = 1024 * 1024 # default max length for generation in bytes
    parser = argparse.ArgumentParser()
    parser.add_argument('-l', '--maxlen', type=int, help='Maximum length '
                        '(in bytes) that may be generated.')
    parser.add_argument('output', nargs='?', help='Output path. Defaults '
                        'to stdout if not specified.', default='/dev/stdout')
    args = parser.parse_args()
    if args.maxlen and args.maxlen < 65535+26:
        print('error: maximum length must be at least 65535+26 (= 267914)',
                file=sys.stderr)
        exit(1)
    elif args.maxlen:
        genlen = args.maxlen - len('GIF89a')

    if args.output!= '/dev/stdout':
        with open(args.output, 'wb') as f:
            generate_random_gif(f)
    else:
        generate_random_gif(sys.stdout.buffer)


# Generates a random GIF file into `out`.
def generate_random_gif(out):
    MAXLEN = 65535 # maximum length for non-data content in bytes
    VERSION = b'89a'
    FIELD_COUNT = b'\x00\x00' # 0 fields
    DATAFIELD_STRUCT = '>HHB' # big endian short unsigned integer followed by
                                 # a single byte
    DATAFIELD_ID = b'\xff\xfe' # application extension field always uses this
                               # ID
    APPLICATION_ID = b'NETSCAPE2.0' # our application ID is NETSCAPE2.0
    BLIP_BLOCK = b'\x21\xf9\x04\x00\x00\x00\x00\x00' \
                 b'\x00\x2c\x00\x00\x00\x00\x00\x01\x00' \
                 b'\x01\x00\x00\x02\x02D\x01\x00;'
    END_BLOCK = b';'
    
    datafield_length = len(DATAFIELD_STRUCT)
    # generates three values for each iteration:
    #   - field id (always 0xfffe)
    #   - block size (size of datafield)
    #   - datafield structure packed together as binary string
    def genfields():
        while True:
            yield ((yield DATAFIELD_ID),
                   (yield datafield_length),
                   (yield struct.pack(DATAFIELD_STRUCT)))

    gen = genfields()
    next(gen) # advance first field
    try:
        lastblockendpos = 0

        while True:
            pos = out.tell()

            # end if we've exceeded MAXLEN or there's nothing more to
            # do from our side
            if pos >= genlen+lastblockendpos:
                break
                
            # get field information
            fid = next(gen) # tuple containing field info
            out.write(fid[0]) # write field identifier
            
            # prepare the rest of the datafield
            bs = fid[1] 
            ds = fid[2]
            dl = bs & (~0b1111) # we don't have any flags set

            # determine the remaining space after the datafield
            remspace = genlen-(pos+bs)-len(END_BLOCK)
            assert remspace > 0
            
            # either pack some data, which ends up being exactly
            # the amount of remaining space or fill the rest with nulls
            if dl <= remspace:
                out.write((dl&0xFF).to_bytes(1, byteorder='big')) 
                out.seek(remspace-dl, 1)
                out.write(ds[-dl:])
                out.write(END_BLOCK)

                # update position tracker
                lastblockendpos = out.tell()+1+dl
            
            # short field, no need to bother about padding
            elif dl < 0b10000:
                nulbuf = [0]*(remspace//dl)*dl
                out.write(((dl|0b10000)&0xFF).to_bytes(1, byteorder='big')) 
                out.seek(-dl, 1)
                out.write(nulbuf)
                out.write(END_BLOCK)
                
                # update position tracker
                lastblockendpos = out.tell()+1+(remspace//dl)*dl
            
            # long field, needs additional null-padding
            else:
                ll = dl-0b10000
                padlen = (-ll)%remspace
                nulbuf = [0]*(padlen // dl)*dl
                out.write(((dl|0b10000)&0xFF).to_bytes(1, byteorder='big')) 

                # seek back so that we can write our padding
                out.seek(-dl, 1)
                out.write(BLIP_BLOCK[:-1]+struct.pack('>H', ll))
                out.write(nulbuf)
                out.write(END_BLOCK)

                # update position tracker
                lastblockendpos = out.tell()+1+padlen

    except StopIteration:
        pass