"""
Generate a random GIF from /dev/urandom.
"""

import io
import os
import struct
import sys
from argparse import ArgumentParser
from typing import BinaryIO

def parse_args() -> str:
    """Parse arguments and set up sensible defaults."""
    parser = ArgumentParser("generate-random-gif")

    # Generate a default name if none is provided
    if len(sys.argv) == 1:
        return "random.gif"

    # Allow passing an argument to specify output filename.
    parser.add_argument("-o", "--output", type=str, help="set destination for generated image (default: %(default)s)")
    ns = parser.parse_args()

    if not ns.output:
        raise RuntimeError('No output specified.')
    
    return ns.output

def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, size : int
    ) -> None:
    """Generates a random GIF file into `out` using the
    random number generator `rng` (/dev/urandom).
    """
    header = b'\x47\x49\x46\x38\x39\x61'
    width = height = size
    color_table = [b'\x2c\xff\xff\xff']*256
    bgcolor = b'\xFF\xFF\xFF'
    pixel_aspect = b'\x00'
    
    extension = [b'\x21\xF9\x04\x00\x00\x00\x00\x2C',
                 b'\x00\x00\x00\x00\x01\x00\x01\x00',
                 b'\x00\x00\x00\x00\x00',
                 b'\x3B']
    
    appext = []
    appext += [struct.pack("<BB", ord(a) & 0xFF, ord(b) & 0xFF)]
    appext += [os.urandom(size//10)]
    applen = len(appext[1]) + len(appext[2])
    appext = [struct.pack("<HH", 0xFFE0, applen)] + appext
    appext = b"".join(appext)
    
    plain_text = b""    
    
    lzwmincodesize = 2
 
    cont = True
    while cont:
        codesize = rng.read(1)[0] % 2**lzwmincodesize + lzwmincodesize
        cont = bool(rng.read(1)[0] // 2**codesize)
        
        plain_text += (
            plain_text[-1:] * 
            (codesize - 1)
        ).encode("utf-8")

        plain_text += (
            chr((
                rng.read(1)[0] // 
                (2**(8 - codesize)) 
            ))
        ).encode("utf-8")

    text = b"".join([
        b"\x00\x3b", 
        plain_text[:int(len(plain_text)/2)], 
        b"\x00\x3b", 
        plain_text[int(len(plain_text)/2):],
        ])
            
    comment = b"\x00Comment " + os.urandom(size//10)
        
    application = b"\x09Netscape Looping Extension" 

    graphic_control = b''
    delay = rng.read(1)[0] % 255
    transparent = 0x00FF00
    aspect = 0x00
    graphic_control = b"\xf9\x04" + \
                       struct.pack("!B",
                               0 |
                               0<<1 |
                               0<<2 |
                               (transparent!= 0)<<3 |
                               (delay!= 0)<<4 |
                               0<<5) + \
                       struct.pack("!BBB", 
                                delay,
                                transparent >> 16,
                                transparent >> 8) + \
                       struct.pack("!B", aspect)
    
    gce = b"\x21\xF9\x04" + graphics_control_extension + b"\x00" 
    
    block = [header,
             struct.pack(">H", width),
             struct.pack(">H", height), 
             b"\x87",
             struct.pack(">B", color_table), 
             struct.pack(">BBB", bgcolor[0], bgcolor[1], bgcolor[2]),
             b"\x00",
             ]

    blocks_left = len(block)
    assert blocks_left > 0
    while blocks_left > 0:
        chunksz = min(blocks_left, size-10)//3+1
        chunk = block[:chunksz]
        chunk = b"".join([
            struct.pack(">Hh",
                         blocks_left-chunksz,
                         0xFFFF),
            b"".join(chunk)])
        chunk = zlib.compress(chunk)
        out.write(chunk)
        blocks_left -= chunksz


    lastbgcolor = None
    lastpixel_aspect = None
    nframes = 0
    
    try:
        while True:
            frame_x = rng.read(1)[0] % (width-1)+1
            frame_y = rng.read(1)[0] % (height-1)+1
            
            #if lastbgcolor is None or lastpixel_aspect is None:
            #    lastbgcolor = bgcolor
            #    lastpixel_aspect = pixel_aspect
                
            nframes += 1

            out.write(b"\x21\xFF\x0B")
            out.write(struct.pack(">HHHHHH",
                                  frame_x,
                                  frame_y,
                                  0,
                                  0,
                                  0,
                                  nframes))
            out.write(graphic_control)
            out.write(b"\x2C")
            out.write(struct.pack(">HHHHHH",
                                  frame_x,
                                  frame_y,
                                  0,
                                  0,
                                  0,
                                  nframes))
            out.write(struct.pack(">BBB",
                                 r,
                                 g,
                                 b))              
                          
    except KeyboardInterrupt:
        pass
    

def main():    
    args = parse_args()

    rng = open("/dev/urandom", "rb")

    with open(args, 'wb') as f:
        generate_random_gif(rng, f, 512)

