#!/usr/local/bin/python3

# This is a simple tool designed to make creating animated gifs from scratch
# easier than any current alternatives I've found on GitHub which would require
# editing an existing GIF manually afterwards.
#
# It's written in pure Python so it should run anywhere Python runs; however,
# it has only been tested on Linux systems.

import io
import os
import struct
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser


class RandomGifGenerator(object):


    def __init__(self, *, outstream:BinaryIO=None, width:int=1024, height:int=768, depth:int=8, bgcolor:int=None):

        self._width = width
        self._height = height
        self._depth = depth
        self._bgcolor = bgcolor
        self._colortable = []

        # Set up the base colors that will be used to build our palette.
        if not self._bgcolor or self._bgcolor<0 or self._bgcolor>255**3:
            self._bgcolor = 0xFFFFFF^((self._width*self._height)%(255**3))

        self._blackcode = bytes([self._bgcolor&(255**2)])
        self._whitecode = bytes([(self._bgcolor>>8)&255,(self._bgcolor>>16)&255])

        # Open the output stream
        if isinstance(outstream,str):

            try:

                self._outstream = open(outstream,"wb")
        
            except Exception as e:
                
                raise RuntimeError("Couldn't create {0}: '{1}'".format(outstream,e))

        else:
            
            self._outstream = outstream


    @property
    def width(self) -> int: return self._width


    @property
    def height(self) -> int: return self._height


    @property
    def depth(self) -> int: return self._depth


    @property
    def bgcolour(self) -> int: return self._bgcolor
    

    @property
    def colours(self) -> int: return len(self._colortable)

    
    # Generate a new colour code by picking one at random
    def _picknewcolour(self)->bytes:
    
        candidate = bytes([int(round(x * 0.5 + y * 0.5)) for x,y in zip(*[iter(os.urandom(3))] * 2)])

        while candidate == self._blackcode or candidate == self._whitecode or candidate in self._colortable:        
            candidate = bytes([int(round(x * 0.5 + y * 0.5)) for x,y in zip(*[iter(os.urandom(3))] * 2)])
        
        return candidate
    
    
    # Add a new colour to the colour table
    def _addnewcolour(self, c:bytes) -> None:
    
        if c!= self._blackcode and c!= self._whitecode and c not in self._colortable:
            self._colortable.append(c)



    # Pick a new colour for each pixel within this rectangle. Will use any already set pixels where possible
    # Returns True if successful, False otherwise
    def fillrect(self, left:int, top:int, right:int, bottom:int) -> bool:
    
        try:
    
            # Do some bounds checking
            if left >= right: return False
            if top >= bottom: return False
            if left < 0: left = 0
            if top < 0: top = 0
            if right > self._width: right = self._width
            if bottom > self._height: bottom = self._height

            totalpix = (right - left)*(bottom - top)
            pixdone = 0


            for y in range(top, bottom-1,-1):
            	for x in range(left, right):
                    
                    pixdone+=1

                    # Find what color we're going to write here...
                    currentcode = self._blackcode
                    if (x+(y*self._width))%2==1:
                        currentcode = self._whitecode
                        
        			if currentcode in self._colortable:
        				# Already have this code, just write it
        				self._outstream.write(currentcode)
        			else:
        				# Need to pick a new code, first check there isn't any overlap with an existing colour
        				while True:
        					candidate = self._picknewcolour()
        					if candidate not in self._colortable: break

        				self._colortable.append(candidate)
        				self._outstream.write(candidate)
        				print('\r{0:.2f}%'.format(float(pixdone)/totalpix*100.),end='')


        finally:
            
            print('')
            return True



    # Output the GIF header
    def emitheader(self) -> None:
    
        version = b'89a'
        screendescriptor = struct.pack(">HHBB", self._width, self._height, 0xF0|self._depth, 0, 0,)
        self._outstream.write(b'\x47'+version+screendescriptor)


    # Build the colour table 
    def emitcolortable(self) -> None:
    
        for colour in self._colortable[:-2]:
    	    self._outstream.write(colour)
    

    	self._outstream.write(self._blackcode)
    	self._outstream.write(self._whitecode)




    # Emit the graphics control block
    def emitgce(self, loopcount=-1) -> None:
    
        applicationdata = b''

        extensions = [{
            "label": b'NETSCAPE2.0',
            "data": struct.pack("<HH",3,loopcount),
        },]

        self._outstream.write(b'\x21\xF9\x04\x00')
        self._outstream.write(struct.pack('<BBBBI',len(extensions),0,0,0,0))

        for ext in extensions:
            label = ext["label"]
            size = len(ext["data"])
            self._outstream.write(struct.pack('<BBBB',size,0,0,0))
            self._outstream.write(label)
            self._outstream.write(struct.pack("<H",size-11))
            self._outstream.write(ext['data'])

        self._outstream.write(b'\x00')


    # Emit the terminator of the GIF file
    def emittrailer(self) -> None:

        self._outstream.write(b";")



def main():

    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,description="A simple utility to generate Gif images containing randomly generated colors.",prog='creategif')
    parser.add_argument('-w','--width',type=int,default=1024,help="Width of the image in pixels.")
    parser.add_argument('-h','--height',type=int,default=768,help="Height of the image in pixels.")
    parser.add_argument('-d','--depth',type=int,default=8,choices=[1,2,4,8],help="Colour depth; how many colours are available in the palette.")
    parser.add_argument('--loop',type=int,default=-1,help="How many times the animation should loop.")
    parser.add_argument('--outfile',type=str,default='output.gif',help="Output filename")
    args = parser.parse_args()

    if args.depth == 1:
        bgcolor = 0xFFFFFF
    elif args.depth == 2:
        bgcolor = 0
    else:
        bgcolor = None
        
    gen = RandomGifGenerator(outstream=args.outfile, width=args.width, height=args.height, depth=args.depth, bgcolor=bgcolor)

    gen.emitheader()
    gen.fillrect(0,0,gen.width//2,gen.height)
    gen.emitcolortable()
    gen.emitgce(loopcount=args.loop)
    gen.fillrect(gen.width//2,0,gen.width,gen.height)
    gen.emittrailer()
    
    




if __name__=='__main__':
    main()