#! /usr/bin/python3

# Copyright © 2019 <NAME>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Generate a random GIF image."""


import io
import sys
import math
import os
import struct
from itertools import chain
from functools import reduce


# shift bit length to make it an even multiple of eight
def nextbyteboundary(length): return (length+7)//8*8

def floorpow2(n): return int(math.floor(math.log2(n)))


# packs an integer into two's complement big endian byte representation with the given number of bits
def pkeb(value : int, width : int ) -> bytes:
    scale = 2 ** width
    assert value >= -(scale//2) and value < scale//2,\
           f"{value} not representable on {width}-bit signed"
    if value < 0: value += scale
    res = []
    while True:
        byte = value // 256
        if len(res)>0 or byte!= 0:
            res = [byte]+res
        else: break
        value -= byte*256
    # remove unnecessary most significant zeroes
    while (len(res) > 1) and (res[-1]==0): del res[-1]
    # correct sign bit
    if len(res)>0: res[0] &= (1<<(8-len(res)%8))-1
    return bytes(chain(*(iter([(value >> (8*i)) & 0xFF]), res)))


# converts integer to little endian byte represenation
def ileb(value: int) -> bytes:
    res = []
    while value > 0:
        res = [(value & 0xFF)]+res
        value >>= 8
    return bytes(res)



# constructs an uncompressed lzw data subblock
def packsubblock(lzwdata : list ):
    buf = iter(lzwdata[:])
    yieldileb = lambda val : ileb(val)
    def group(gen, n=None):
        if n is None:
            n = 1
        assert n>=1
        acc = list()
        try:
            while True:
                acc.append(next(gen()))
                if len(acc)==n:
                    yield acc; acc=[]
        except StopIteration: pass
        if len(acc)>0: yield acc
    # parse lzw dictionary
    dictbuf = iter(pkeb(v, 12) for k,v in enumerate(lzwdata[:-1]))
    dict_last_key = -1
    dict_last_val = lzwdata[-1]
    # read a single token stream
    def stream():
        while True:
            # check if we have enough space in output buffer
            if len(buf)<10:
                yield from buf
                buf = iter([])
            # fetch the next dictionary entry
            key = next(dictbuf)
            if isinstance(key,list): raise ValueError("unexpected end of input")
            if dict_last_key<0:
                dict_last_key = key
                dict_last_val = next(buf)
            elif key!=dict_last_key:
                yield dict_last_key
                dict_last_key = key
                dict_last_val = next(buf)
            # yield the dictionary entry as LEB12
            yieldileb(dict_last_val)
    # build the compressed token sequence
    seq = list(group(stream(),4))
    if len(seq)==0: return b''
    seq[-1].extend(buf)
    # prepare output buffer
    out = io.BytesIO()
    # first item is the size of the entire subblock minus data size
    out.write(pkeb(-len(pkeb(*lzwdata))*len(seq), 3))
    # then compress each chunk separately
    for chunk in seq:
        # first chunk contains only the data size
        # subsequent chunks contain the mask for the compression type
        if chunk==seq[0]: continue
        out.write(pkeb(reduce(lambda a,b: a|b,chunk[:3],0), 3)[::-1])
        del chunk[:3]
        # create and write the packed data
        out.write(bytes(
            sum(len(item) for item in chunk) +
            (k==len(chunk)-1)*(len(chunk)*len(chunk[-1])+(len(chunk)-1)*8),
            ))
        for item in chunk: out.write(pkeb(*item))
    return out.getbuffer()





###########
#        #
# GENERATE A RANDOM GIF IMAGE INTO FILE `OUT`
#        #
###########




class RandomGifGenerator:

    def __init__(self, seed=os.urandom, maxtotalelements=10000,
                 minwidth=64, maxwidth=512, minheight=64, maxheight=512,
                 nframes=10, colormapper=(
                     lambda c : tuple(c*255 for c in
                                 map(lambda e :
                                     round(abs(e)**(1/2)),
                                     rng.choice(range(256))
                                    )*rng.choice([-1,1])))
               , backgroundcolor = (0,0,0)):

        self.rng                = seed
        self.background         = backgroundcolor
        self.maxelems           = maxtotalelements
        self.minw               = minwidth
        self.maxw               = maxwidth
        self.minh               = minheight
        self.maxh               = maxheight
        self.nfr                = nframes
        self.colormapper        = colormapper
        self.currentframe       = 0



    def getminmaxdim(self):
        minw = self.minw
        minh = self.minh
        maxw = self.maxw
        maxh = self.maxh
        if self.currentframe==0:
            minw *= 2
            maxw *= 2
            minh *= 2
            maxh *= 2
        return minw, maxw, minh, maxh




    def genheader(self):

        self.width              = self.rng.randint(*self.getminmaxdim())//2*2
        self.height             = self.rng.randint(*self.getminmaxdim())//2*2

        self.aspectratio        = self.rng.uniform(
            1, min(self.width,self.height)/max(self.width,self.height))

        self._progress          = float('inf')
        self._progressupdate    = False
        self.percentdone        = 0.0


        self.out.write(b'GIF89a')
        self.out.write(ileb(self.width))
        self.out.write(ileb(self.height))



        bgcolor = self.background
        bgcolmapped = self.colormapper(bgcolor)

        self.out.write(bytes(chain(*(iter([bgcolmapped[0]]),
                                      iter([bgcolmapped[1]]),
                                      iter([bgcolmapped[2]])))))

        self.out.write(b'\x00')                    # GCT offset
        self.out.write(b'\xf7')                    # image descriptor block code
        self.out.write(b'\x00')                    # LZW minimum code size
        self.out.write(pkeb(0, 3))                 # top left corner pos
        self.out.write(pkeb(0, 3))                 # size flags
        self.out.write(pkeb(0, 1))                 # no local colors used
        self.out.write(b'\0\0\0')                  # LZW table clear code
        self.out.write(b'\xe0\xaa\xa9\xa8\xc0\xa0\xa0\xa0\xae\xaa\xad\xac')






#######
#     #
# MAIN FUNCTION
#     #
######




if __name__=='__main__':

    rng = RandomGifGenerator()

    with open("/tmp/test.gif","wb") as out:


        rng.genheader()      # write file header

        # fill image by adding new frames until there aren't enough pixels anymore
        pixcount = rng.width*rng.height

        # start lzw with