#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(5) + bytes([2])

    # Add the LSD to our output
    out.write(lsd)

    # Get the global color table flag bit
    gctflag = lsd[4] & 0x80

    # If this is true, read the global color table
    if gctflag != 0:
      numcolors = (1 << ((lsd[4] & 7)+1))
      colortable = rng.read(numcolors*3)
      out.write(colortable)

    # Read the contents of our GIF
    while True:
      blocktype = rng.read(1)[0]
      size = rng.read(1)[0]
      
      # Check for terminator
      if blocktype == 0x3B:
        break

      # Otherwise write out the block type and size
      out.write(struct.pack('BB', blocktype, size))

      # Then get our block based on block type
      if blocktype == 0xFE:
        comment = rng.read(size-1)
        out.write(comment)
        
        # If the comment has a zero byte at the end, make sure to strip it off
        if comment[-1] == 0:
            out.seek(-1, 1)
            
      elif blocktype == 0xF9: # Graphics Control Extension (GCE) Block
        gce = rng.read(6)
        out.write(gce)
        
        # Find the color table flag in our GCE and set whether to use local color tables
        useLocalColorTable = (gce[4]&0x80)/256 > 0
                    
        # Read the image data
        if useLocalColorTable:
          colortable = rng.read(numcolors*3)
          out.write(colortable)

          # Skip over the rest of the colors in our palette since we only support
          # 256
          leftovercolorbytes = max((numcolors - 256)*3, 0)
          if leftovercolorbytes > 0:
              rng.seek(leftovercolorbytes, 1)
              
        # Write the actual image data
        imgdata = rng.read(size-1)
        out.write(imgdata)