#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep 29 10:42:26 2021

@author: alina
"""
import numpy as np
from copy import deepcopy
import json
import warnings
warnings.filterwarnings('ignore',category=DeprecationWarning)

from . import io_utils
from . import communication_measurements
from .communication_measurements import generate_comm_matrix, comm_to_json, comm_matrix_from_json, comm_matrices_from_dict
from .io_utils import load_config, save_obj, load_obj
from .smoothing import filtered_by_exposure, smoothed_data

def get_PFS(gm_dic, config):
    """
    Obtain PFS from GM data and saving smoothed version of the GM data as a 
    dictionary with the time stamps as keys

    Parameters
    ----------
    gm_dic : dict
        Dictionary containing the gait magnitude (Gm) measurements from the 
        subject
    config : YAML
        Configuration file with the parameters for the analysis.
        
    
    Returns
    -------
    smoothed_pfs : list
        List of the smoothed PFS.

    """
    gm = [] #Smoothed GM
    pfs = [] #Smoothed PFS
    maxs = [] #Maximum GM in each trial
    mins = [] #Minimum GM in each trial

    for key in gm_dic:
        if config['verbose'] == True: print("Trial name is: ",key)
        datamatrix = np.array(gm_dic[key])
        if config['preprocess']['filter_window_length'] != None:
            datamatrix = filtered_by_exposure(datamatrix, config['preprocess'])
        
        smoothed_GM = smoothed_data(datamatrix, config['preprocess'],
                                   config['verbose'])
        maxs.append(np.max(smoothed_GM))
        mins.append(np.min(smoothed_GM))
        gm.append(smoothed_GM)
        
        timestamps = np.arange(0,(len(smoothed_GM)*config['preprocess']['dt']), config['preprocess']['dt'])
        trialdict = {timestamps[i]:float(smoothed_GM[i]) for i in range(0,len(timestamps))}
        try:
            pfs[-1].update(trialdict)
        except IndexError:
            pass
            
        pfs.append(deepcopy(trialdict))

    return pfs

def get_IPF_profile(ipf,PFS,config, RMS = False):
    """
    Returns IPF profile
    Parameters
    ----------
    ipf : array_like
        Array containing the IPF values
    PFS : list
        List of the smoothed PFS
    config : YAML
        Configuration file with the parameters for the analysis.
    RMS : Boolean
        If True returns the RMS value instead of the average
    Returns
    -------
    Ipf_list : list
        List of the IPF profiles

    """
    Ipf_list = []
    if len(PFS)>1:
        for n in range(1,len(PFS)):
            try:
                if RMS==False:
                    Ipf_list.append((ipf/np.mean([PFS[n][timestamp] for timestamp in PFS[n]])))
                else:
                    Ipf_list.append(np.sqrt(ipf**2/(np.std([PFS[n][timestamp] for timestamp in PFS[n]])**2+config['noise_variance']**2)))
            except ZeroDivisionError:
                continue
    elif len(PFS)==1:
        for key in PFS[0]:
            try:
                if RMS==False:
                    Ipf_list.append((ipf/PFS[0][key]))
                else:
                    Ipf_list.append(np.sqrt(ipf**2/(np.std([PFS[n][key] for n in PFS])**2+config['noise_variance']**2)))
            except ZeroDivisionError:
                continue
    return Ipf_list


def compute_mean_squared_error(mean_value, mean_square_value):
    """
    Computes the mean squared error between two vectors by the formula
    MSE = ((sum((x_i - y_i)^2)/N)
    
    Where x_i and y_i are the first and second elements of the vectors respectively.
    
    Parameters
    ----------
    mean_value : array_like
        Vector containing one set of values.
    mean_square_value : array_like
        Vector containing another set of values.
    
    Returns
    -------
    float
        Mean squared error between both sets of values.
    """
    N = len(mean_value)
    sum_of_differences = 0
    for i in range(0,N):
        sum_of_differences += (mean_value[i] - mean_square_value[i]) ** 2
    return sum_of_differences / N


def noise_correlation_analysis(PFS, config, RMS = False):
    """
    Performs a noise correlation analysis for a set of trials, given the 
    configurations in the configuration file.

    Parameters
    ----------
    PFS : list of dictionaries
        List containing the PFS values as dictionaries.
    config : YAML
        Configuration file with the parameters for the analysis.
    RMS : boolean
        If True the standard deviation used to calculate the noise covariance 
        matrix will be the arithmetic (RMS) root of the square of the variance
        plus the noise variance. Otherwise, the variance is going to be 
        calculated using the sample variance.

    Returns
    -------
    icovariance or noise_covariance : numpy matrix
        Noise covariance matrix

    """
    
    n_samples = [len(ps) for ps in PFS]
    
    N = np.sum(n_samples)
    tau_bar = np.zeros((N,N))

    for i in range(0, len(PFS)):
        for j in range(0, len(PFS)):
            time_stamps = sorted(PFS[i].keys() & PFS[j].keys()) #& performs intersection operation
            index_i = [PFS[i][ts] for ts in time_stamps]
            index_j = [PFS[j][ts] for ts in time_stamps]
            if RMS==True:
                stddev = np.sqrt(np.var(index_i)+np.var(index_j)) + config['noise_variance']
            else:
                stddev = np.std(index_i)+np.std(index_j) + config['noise_variance']
            tau_bar[i][j] = np.exp(-int(time_stamps[0])/(2*stddev**2)) * int(time_stamps[1]) / stddev**2
            tau_bar[j][i] = tau_bar[i][j]

    icovariance = 1 / np.linalg.det(tau_bar) * np.linalg.inv(tau_bar)

    return icovariance


def add_trial(subject_dir, trial = 'Trial_4'):
    """
    Adds one trial to the dictionary that contains all the data for a specific 
    subject.

    Parameters
    ----------
    subject_dir : string
        Path to the directory where the subject's folder is located.
    trial : string
        Name of the trial to be added to the dictionary. The default trial 
        will be 'Trial_4'.

    Returns
    -------
    Nothing

    """

    configname = 'MANOVA.yaml'
    gm_name = 'gait_magnitude.csv'
    ema_name = 'emg.csv'
    ipf_name = 'imu_position.csv'
    iro_name = 'imu_rotation.csv'
    emg_name = 'force_plate.csv