#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # If Global Color Table exists, write it into output
    if gctlen > 0:
        out.write(rng.read(gctlen))

    # Start background color index
    bgci = int.from_bytes(rng.read(1), 'big') & 0xFF

    # Add background color index to output
    out.write(bgci.to_bytes(1,'big'))

    # Pixel aspect ratio - unused, just read one byte
    out.write(rng.read(1))

    # Image Separator
    image_separator = b'\x2C'
    out.write(image_separator)

    # We don't need min codes for RLE decoding, so we skip them
    rng.read(1)

    # Let's not try to decode an image that is too large or small.
    max_width = 1 << 16
    max_height = 1<<16
    width, height = struct.unpack('<HH', rng.read(4))
    assert width < max_width
    assert height < max_height

    # Write image dimensions to output
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))

    # Local Color Table Flag
    local_color_table_flag = bool(int.from_bytes(rng.read(1), 'big') & 0x80)

    # Interlace Flag
    interlace_flag = bool(int.from_bytes(rng.read(1), 'big') & 0x40)

    # Sort Flag
    sort_flag = bool(int.from_bytes(rng.read(1), 'big') & 0x20)

    # Reserved
    reserved = bool(int.from_bytes(rng.read(1), 'big') & 0x08)

    # Size of Local Color Table
    size_of_local_color_table = int.from_bytes(rng.read(1), 'big') & 0x07

    # Skip the rest of the fields in the LCT (Color Map Specification)
    rng.read(3)

    # Create local color table
    local_palette = []
    if local_color_table_flag == True:
        # Loop over each palette entry
        for i in range(size_of_local_color_table + 1):
            # Get color component values
            red   = int.from_bytes(rng.read(1), 'big') & 0xFF
            green = int.from_bytes(rng.read(1), 'big') & 0xFF
            blue  = int.from_bytes(rng.read(1), 'big') & 0xFF

            # Calculate color value
            color_value = (red << 16) | (green << 8) | blue

            # Append color value to palette
            local_palette.append(color_value)

    # Number of bits per code in compression block
    nbpc = int.from_bytes(rng.read(1), 'big') & 0xF8

    # Skip bitmap data block offset field since we are going to start writing after it
    rng.read(1)

    # Initialize pixel
    pixels = [0] * width * height

    # Initialize current pixel x position
    xpos = 0

    # Initialize current pixel y position
    ypos = 0

    # Initialize available bit count
    availible_bits_count = 0

    # Initialize current_code
    current_code = 0

    # Initialize code size
    code_size = nbpc + 1

    # Loop until either end of block is reached
    while True:
        # Check if we have enough availible bits
        if availible_bits_count >= code_size:
            # Get new code based on availible bits
            current_code = availible_bits & ((1 << code_size) - 1)
            
            # Remove used bits from availible bits
            availible_bits >>= code_size

            # Reduce availible bit count by code size
            availible_bits_count -= code_size

            # Is this the end of the block?
            if current_code == 0x00:
                break

            # Is this clear code?
            elif current_code == 0x01:
                # Start with new code size
                code_size = nbpc + 1
                
                # Reset availible bits
                availible_bits = 0
                availible_bits_count = 0
                
                # Set initial code size
                current_codesize = old_current_code_size = code_size + 1
            
            # Is this end of information code?
            elif current_code == 0x02:
                break

            # Is this new code?
            else:
                # Do we need to resize code table?
                if next_code >= 4096:
                    # Double the size of the table
                    code_size += 1
                
                # Add another code
                next_code += 1

                # Is this the first code?
                if old_current_code == None:
                    # No previous code
                    availible_bits = current_code
                    availible_bits_count = code_size
                    old_availible_bits = availible_bits
                    old_availible_bits_count = availible_bits_count
                    continue
        
                # Append new code to prefix table
                prefixes.append(old_current_code)
                suffixes.append(current_code)

                # Set the code value for the new code to be equal to the
                # concatenation of its prefix code with its suffix code
                values[next_code] = values[old_current_code] + values[current_code]

                # Reset old code size
                old_current_code_size = current_code_size

            # Determine the current_code, based on the old_current_code and its suffix code
            old_current_code = current_code
            current_code = suffixes[-1]

            # Is old_current_code referring to the last string in the table?
            if len(suffixes) <= old_current_code:
                # Then there is no more such code. Break the loop
                break;

            # Check availible bits count
            if availible_bits_count > 0:
                # Go back to previous availible bits and availible bits count
                availible_bits = old_availible_bits
                availible_bits_count = old_availible_bits_count

            # Check for clean up suffix table entries
            if next_code >= 4096:
                # Clean up the table
                for i in range(next_code, 511+1):
                    # Remove code by setting it to none
                    values[i] = None

                # And reset the table
                next_code = 512
            
            # Increment old code
            old_current_code += 1

        # Make sure that at least one bit is availible
        if availible_bits_count == 0:
            availible_bits = rng.read(1)[0]
            availible_bits_count = 8

        # Decrease availible bits count
        availible_bits_count -= 1

        # If we have not finished reading all bytes and this is the last pixel
        if ypos == height-1 and xpos == width-1:
            # If some bits remain
            if availible_bits_count != 0:
                # Append a dummy byte
                availible_bits |= 0x100
                availible_bits_count += 1

        # If the current position is within the bounds of the image
        if xpos < width and ypos < height:
            if local_color_table_flag == False:
                # Use the global palette value
                palette_index = current_code % (len(global_palette))
            else:
                # Use the local palette value
                palette_index = current_code % (size_of_local_color_table + 1)

            # If the palette index is valid
            if palette_index >= 0 and palette_index < len(local_palette):
                #