#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(5) + bytes([2])

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have one
    if gctlen > 0:
      gct = rng.read(gctlen)

      # If we don't get an even number of colors then we need padding to make sure
      # all resulting images contain the same amount of bytes
      outbytesleft = 256 - len(gct)
      if outbytesleft % 3!= 0:
        out.write(b'\x00'*(outbytesleft%3))
      
      out.write(gct)
      
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(rng.read(1))
        
        # We currently only handle up to 256 so we can use it directly as the total
        # number of colors
        if (fieldbyte & 7) == 4:
          
          # Read the total number of colors
          numcolors = 2**(int((fieldbyte>>2)&7)+1)
          colortable = rng.read(numcolors*3)
          out.write(colortable)

          # Skip over the rest of the colors in our palette since we only support
          # 256
          leftovercolorbytes = max((numcolors - 256)*3, 0)
          if leftovercolorbytes > 0:
              rng.seek(leftovercolorbytes, 1)
              
        # Write the actual image data
        imgdata = rng.read(size-1)
        out.write(imgdata)