#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)
#     print('lsd', bytes(lsd).hex())
    width  = int.from_bytes(lsd[0:2], 'little') + 1
    height = int.from_bytes(lsd[2:4], 'little') + 1
#     print('width', width, 'height', height)
    packed_byte = lsd[4]
    flags = packed_byte & 0x7

    # Add LSD field to output
    out.write(lsd[:6])

    if (flags == 0):
        out.write(b'\x00\x3B')
    else:
        ext = rng.read(1)[0]

        while True:
            # End of extensions
            if (ext == 0x00):
                break

            # Plain text extension
            elif (ext == 0xF9 or ext == 0xFE):
                size = rng.read(1)[0]

                if ((size!= 12) and
                    (size!= 12+int((packed_byte >> 3)&1))
                   ):
                    continue

                subblock = bytearray()
                
                for i in range(0, 1+int(((packed_byte>>3)&1)*8), 1):
                    subblock += rng.read(1)

                out.write(struct.pack('<cBBBHB',
                                    ext.to_bytes(1, 'big'),
                                    1+len([i for i in subblock]),
                                    *subblock
                                   ))

            # Graphic control extension
            elif (ext == 0xF9):
                size = rng.read(1)[0]

                if (size!= 4):
                    continue

                out.write(ext.to_bytes(1, 'big'))
                out.write(struct.pack('<BBBB', *rng.read(4)))
                out.write(b'\x00')

            # Application extension
            elif (ext == 0xFF):
                appid = rng.read(1)[0]

                if (appid not in [0xF9]):
                    continue

                size = rng.read(1)[0]

                if (size!= 4):
                    continue

                out.write(ext.to_bytes(1, 'big'))
                out.write(appid.to_bytes(1, 'big'))
                out.write(struct.pack('<I', rng.read(4)))
                out.write(b'\x00\x00')

            else:
                size = rng.read(1)[0]
                out.write(ext.to_bytes(1, 'big'))
                out.write(size.to_bytes(1, 'big'))
                out.write(rng.read(size))

            # Read next extension block
            try:
                ext = rng.read(1)[0]
            except IndexError:
                ext = None


    # Set up offset table
    data_offsets = []
    
    for y in range(height-1,-1,-1):
        line = bytearray()
        
        for x in range(0, width*3, 3):
            line.append(rng.read(1)[0])
            line.append(rng.read(1)[0])
            line.append(rng.read(1)[0])
            
        data_offsets.insert(0, len(out.getbuffer()))
        out.write(line)

    # Pad to multiple of 255 with 0xFFs
    # This is just how it works atm
    padding = -len(out.getbuffer())%255

    if padding > 1:
        out.write(bytearray(padding-1))
        padding = 1

    # Add the trailer
    out.write(struct