#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
        rng.read(2), byteorder='little', signed=False
    ) + 1
    if datalen == 0:
        raise ValueError('GIF data length can\'t be zero')

    delay = 65535 - int.from_bytes(rng.read(2), byteorder='big', signed=True) % 65535
    if delay < 1:
        print("Delay too short")
        exit()
        
    # Transparent color
    transparent_color = int.from_bytes(
            rng.read(1), byteorder='big', signed=False
    ) % 256

    out.write(b'\x47\x49\x46\x38\x39\x61\x00\x00' + datalen.to_bytes(2, byteorder='little'))

    block = b''
    for i in range(datalen):
        block += rng.read(1)
        if len(block) >= 10:
            try:
                (
                    extcode, blocksize, x, y, width, height, flags, *rest
                ) = struct.unpack('<BHHHBBB4s', block[:10])
                block = block[10:]
                if width > 0 and height > 0:
                    offset = ((y + height - 1) % 256) << 8 \
                          | (x % 256)
                    (loops, terminator) = struct.unpack('BB', rest)
                    if loops != terminator or terminator != 0:
                        raise ValueError(
                            'Invalid loop count'
                        )
                    transparent = bool((flags & 1) != 0)
                    interlace = bool((flags >> 1 & 1) != 0)
                    localcolortable = bool((flags >> 7 & 1) != 0)
                    globalcolortable = not localcolortable
                    tablebits = 1 + ((localcolortable or globalcolortable) and 7 or 0)
                    colordelay = 1 / float(tablebits ** 2)
                    repeatdelay = [
                        (1, 0), (2, delay // 2), (3, delay // third)
                    ][(interlace, i)]

                    colors = []
                    for _ in range(3 ** tablebits):
                        colors.append(int.from_bytes(
                            rng.read(3), byteorder='big', signed=False
                        ))

                    if transparent:
                        colors.remove(transparent_color)
