#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)
#     print('lsd', bytes(lsd).hex())
    width  = int.from_bytes(lsd[0:2], 'little') + 1
    height = int.from_bytes(lsd[2:4], 'little') + 1
#     print('width', width, 'height', height)
    packed_byte = lsd[4]
    flags = packed_byte & 0x7

    # Add LSD field to output
    out.write(lsd[:6])

    if (flags == 0):
        out.write(b'\x00\x3B')
    else:
        ext = rng.read(1)[0]

        while True:
            if not ext: break

            length = rng.read(1)[0]
#             print('ext', byte.from_bytearray(length))
            data = rng.read(length+1)
#             print('data', byte.from_bytearray(data))
            out.write(bytearray([ext]))
            out.write(bytearray([length]))
            out.write(data)
            
            ext = rng.read(1)[0]
            if not ext:
                break

#         print()

    # Color table fields
    out.write(b'\xFF\xFF\xFF')
    out.write(bytearray((0,)*252))

    # Pixel data
    for _ in range(width*height):
        pixel = rng.read(1)[0]
        out.write(pixel.to_bytes(1, 'big'))

    # Trailing data
    out.write(b'\x3B')


# Encodes a string using a given codebook
def encode_string(
        string: str, codebook: list, bitorder='msbfirst'):

    encoded_data = ''
    bits_per_codeword = len(codebook[0]).bit_length()

    words = [int(s, 2) for s in zip(*[iter(string)]*(bits_per_codeword))]
#     print(words)

    for word in words:
        if bitorder == "msbfirst":
            code = bin(word)[2:].zfill(bits_per_codeword)
        elif bitorder == "lsbfirst":
            code = bin(word)[2:].zfill(bits_per_codeword)[::-1]
        encoded_data += codebook[int(code,2)]
#         print("word", word)
#         print("code", code)
#         print("encoded_data", encoded_data)
#         print()

    return encoded_data


# Generates a random LZW image from a stream of raw pixel data
def generate_random_lzw_image(
        rng: BinaryIO, out: BinaryIO, 
        width: int, height: int, 
        codesize: int, init_codes: int):

    # Rewind random source to start at beginning
    rng.seek(0)

    # Set up clear code and end code
    clear = 1 << codesize
    end   = clear+1

    # Read in first pixel value as initial code
    pixel = rng.read(1)[0]
    saved_pixels = [pixel]

    # Read in next pixel value
    pixel = rng.read(1)[0]
    pixels = [pixel]

    # Create empty dictionary
    codes = []

    # Generate unique code for each terminal symbol
    for i in range(init_codes):
        codes.append(list(saved_pixels[-1:])[0].to_bytes(1, 'big'))

        # Add both newcodes to the list of saved codes
        saved_pixels.append(pixels[-1])
        pixels.append(pixels[-1])

    codes.extend(tuple(map(lambda x: x.to_bytes(1, 'big'), pixels)))

    # Generate random suffix codes
    suffixes = []
    for i in range(2**codesize - init_codes - 2):
        suffixes.append(rng.read(1)[0].to_bytes(1, 'big'))

    codes.extend(suffixes)

    # Create a list of data offset addresses
    data_offsets = [0] * width * height


    ##############
    # ENCODING LOGIC

    data_offset = 0

    # For each column in the image
    for col in range(width):
        # For each row
        for row in range(-height//2+1, height//2):
            if row > 0:
                # Move past existing offset address
                out.write(b'\x00\x00')
                data_offset += 2

            # For each pixel in the current row
            for pix in range(col, width):
                # Save the current offset into the array
                data_offsets[row*width + pix] = data_offset

                # Check whether we can create a new code or not
                if pix<width-1 \
                and ((pix==col) | (pix%(width//10)==0)):

                    # New code gets inserted before the end code
                    codes.insert(end, codes.pop()+codes.pop())
                    end+=1


                # Encode the current row
                encoded_data = encode_string(
                                        ''.join(str(pixel) 
                                                for pixel in pixels),
                                        codes
                                       )
                # Write the encoded data to output
                out.write(encoded_data.encode('iso-8859-1'))

                # Update encoder state
                if pix >= width-1:
                    # Need to write an ending code
                    pixels = [end]*2
                else:
                    pixels = pixels[-1:] + [pixel]

                # Update the list of pixels
                if pix < width-1:
                    # Shift old pixels down by one
                    pixels[-2:] = pixels[:-1]

                # Move to the next data offset
                data_offset += len(encoded_data.encode('iso-8859-1'))


################
# DECIPHERING LOGIC

    def decode_block(
            block: bytes, codes):

        values = ['']
        line = ''

        for byte in block:
            packed = byte.to_bytes(1, 'big')
#             print("packed", packed)
#             print(type(packed), len(packed), packed.hex(), packed.decode('utf-8'))
#             print(type(packed), len(packed), packed.bit_length(), packed.to_bytes(2, 'big').hex())
#             print("packed", packed.to_bytes(2, 'big').hex())
            clear = '\x00\x00'[packed==b'\x00']
            code = int(str(packed), 16)

            if code == clear:
                values.append('')
            
            elif code == end:
                values.append('')
                values[-2] += codes[int.from_bytes(packed,'big')]
            
            elif any(i in values[-1] for i in backref_codes):
                index = int.from_bytes(packed, 'big')
                marker = min(backref_codes, key=lambda k: values[-1].find(k))

                # If after the marker we find another occurance of this marker
                # then it's an escaped character
                if values[-1][values[-1].find(marker)+len(marker)] == marker:
                    # This is an escape sequence
                    line += str(clear)
                    values.append('')

                    # Append escaped char (as string) to previous value
                    values[-2] += codes[index]
                    
                else:
                    # This is a normal reference
                    line += str(clear)
                    values.append('')
                    
                    ref_index = values[-1].rfind(marker)+len(marker)
                    values[-2] += values[-1][ref_index:]
                    values[-1] = values[-1][:ref_index] + values[-1][marker:]
                    
            else:
                # If no matches are found just append to last value
                index = int.from_bytes(packed, 'big')
                values[-1] += codes[index]

                line += ''.join(['{:>3}'.format(str(v))
                                for v in values[-1]])
                
        encoded_data = encode_string(line, codes)
        out.write(encoded_data)


    # Image descriptor
    out.write(struct.pack('<HHHBB',
                         width      % 0x10000,
                         height//256,
                         0,
                         0,
                         0
                        ))

    # Graphics control extension with transparent color flag set
    out.write(b'\x21\xf9\x04\x01\x00\x00\x00\xff\xff')

    # Ending image descriptor
    out.write(b';')

    # Codes dictionary
    out.write(struct.pack('<B',len(codes)))
    out.write(b''.join(codes))

    # Data offsets
    out.write(struct.pack('<B', len(data_offsets)))

    for doffset in data_offsets:
        out.write(doffset.to_bytes(2, 'big'))
        
    

