#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)
#     print('lsd', bytes(lsd).hex())
    width  = int.from_bytes(lsd[0:2], 'little') + 1
    height = int.from_bytes(lsd[2:4], 'little') + 1
#     print('width', width, 'height', height)
    packed_byte = lsd[4]
    flags = packed_byte & 0x7

    # Add LSD field to output
    out.write(lsd[:6])

    if (flags == 0):
        background_color = packed_byte >> 4
        pixel_aspect_ratio = 0
    elif (flags == 1):
        background_color = packed_byte >> 5
        pixel_aspect_ratio = packed_byte
    else:
        raise ValueError("Unknown flag!")
    out.write((packed_byte,))

    ## Background color
    out.write(struct.pack('BBB', *background_color))

    # Global Color Table
    gct_flag = bool((flags >> 3) & 0b1)

    if gct_flag:
        gct = [0, 0, 0]
        for i in range(0, 768, 3):
            gct[0] = int(256 * rng.read(1)[0] / 255)
            gct[1] = gct[0]
            gct[2] = gct[0]
            out.write(gct)

    # Image Descriptors
    image_descriptors = []
    num_images_in_file = rng.read(1)[0] % 10 + 1
    for _ in range(num_images_in_file):
        id_field = bytearray([0x2c])
        size = rng.read(3)
        x_pos, y_pos = struct.unpack("<H", size[0:2])
        size = ((size[2] << 3) | 0b111110)
        size += 1
        id_field.extend(size)
        id_field[3] |= 0b11111
        out.write(id_field)

        # Image Data
        line_len = (width + 3) // 4 * 3
        skip = 0

        while True:
            if rng.read(1)[0]:
                break

            is_transparent = rng.read(1)[0] & 0b1
            skip = rng.read(1)[0]
            if not skip or skip > (line_len - 8):
                skip = line_len - 8
                out.write(b'\x00\x00\x00\x00\x00\x00\x00\x00')
                continue
            count = rng.read(1)[0]
            if not count:
                count = line_len - x_pos
            if count > line_len - 2:
                count = line_len - 2
            if count < 2:
                break
            
            block_type = rng.read(1)[0]
            if block_