#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(5) + bytes([2])

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have one
    if gctlen > 0:
      gct = rng.read(gctlen)

      # If we don't get an even number of colors then we need padding to make sure
      # all resulting images contain the same amount of bytes for their 'data' sections.
      pad = gctlen % 2

      if pad == 1:
          gct += rng.read(1)

      out.write(gct)

    # Get image dimensions and write to output
    width, height = struct.unpack("<HB", rng.read(3))
    out.write(width.to_bytes(2, "little"))
    out.write(height.to_bytes(2, "little"))

    # Start writing the 'data' section. First we get the Packet Size Byte (PSB). This is
    # used to group blocks together based on how much data they contain. We set it so
    # that we will only ever have one such block in our data section.
    psb = int(rng.read(1)).to_bytes(1, "little")
    out.write(psb)

    # Now we get the pixel data for this image. One byte per pixel!
    pixels = rng.read(width*height)

    # Add the pixel data to the 'data' section along with any required padding
    img_data = pixels + rng.read(pad)
    out.write(img_data)

# Generates a random GIF file called 'random.gif'. Uses the same RNG as
# the reference implementation
with open('random.gif', 'w+b') as out:
  with open('/dev/urandom', 'rb')