#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO) -> None:
    # Magic header and version string. Version 89a is the most recent as of this writing
    out.write(b"GIF89a")
    
    # Height/width fields. These are actually UINT16s but we're lazy so let's just use UINT8s instead
    height = rng.read(1)[0]
    width = rng.read(1)[0]
    out.write(struct.pack("BB", width, height))
    
    # Backgroud color index field. This is a single-byte value
    background_color_index = rng.read(1)[0]
    out.write(background_color_index)
    
    # Aspect ratio. Unused, so all we need to do here is write one byte with zeroed bits
    aspect_ratio_bits = 0
    out.write(aspect_ratio_bits & 0xFF)
    
    # Block terminator byte (always 0)
    block_terminator = 0
    out.write(block_terminator)
    
    # Image header. Only one image per file, so you know what comes next!
    out.write(b"!")
    
    # Left/top fields. Once again, we can get away with UINT8s since they'll never need more than two bytes
    left = rng.read(1)[0]
    top = rng.read(1)[0]
    out.write(struct.pack("BB", left, top))
    
    # Width/height fields. These will be UINT16s, since they could potentially go beyond the size of an unsigned byte
    width = rng.read(2)
    height = rng.read(2)
    out.write(width + height)
    
    # Local color table specification bitfield
    local_color_table_specification = rng.read(1)[0]
    out.write(local_color_table_specification)
    
    # Number of tables in the global color table. Since we aren't using a global color table, we set this to zero
    reserved_zero = 0
    out.write(reserved_zero)
    
    # Size of each color table entry (in bytes). We want 4 bytes per color entry, or 32 bits total
    sort_flag = 0
    size_of_color_table_entry = 4
    out.write((sort_flag << 7) + ((size_of_color_table_entry - 1) << 4))
    
    # Color table. We'll skip over this for now because its values don't matter much
    # If we wanted to fill it in later on, we would make sure that each color tuple is
    # 256 colors long, then loop through each pixel's red value, green value, blue value
    # in order to output RGB data like this:
    # for red in range(256):
    #     for green in range(256):
    #         for blue in range(256):
    #             out.write(struct.pack("B", red))
    #             out.write(struct.pack("B", green))
    #             out.write(struct.pack("B", blue))
    pass
    
    # LZW minimum code size, preceded by an LZW extension block identifier byte
    lzw_extension_identifier = 0xF9
    out.write(lzw_extension_identifier)
    lzw_minimum_code_size = rng.read(1)[0]
    out.write(lzw_minimum_code_size)
    
    # Block terminator byte (again)
    block_terminator = 0
    out.write(block_terminator)
    
    # Applicaiton extension block. This holds info about which application created this file
    application_extension = 0xFF
    out.write(application_extension)
    application_extension_length = 11
    out.write(struct.pack("B", application_extension_length))
    application_identifier = b"NGIF"
    out.write(application_identifier)
    authentication_code = b"\x00\x00\x00\x00"
    out.write(authentication_code)
    
    # Application name
    application_name = "GIF_GEN".encode('utf-8')
    out.write(struct.pack("BB", len(application_name), 0x00))
    out.write(application_name)
    
    # Block terminator byte (yet again)
    block_terminator = 0
    out.write(block_terminator)
    
    # Graphic control extension
    graphic_control_extension = 0xF9
    out.write(graphic_control_extension)
    graphic_control_extension_length = 4
    out.write(struct.pack("B", graphic_control_extension_length))
    disposal_method = 0
    user_input = False
    transparent_color_flag = True
    out.write(struct.pack("BB", disposal_method | (user_input << 2) | (transparent_color_flag << 1), background_color_index))
    
    # Delay time
    delay_time = 10
    out.write(struct.pack("B", delay_time & 0xFF))
    
    # Transparency index field (same as background color index field because we didn't specify anything else)
    out.write(background_color_index)
    
    # Block terminator byte
    block_terminator = 0
    out.write(block_terminator)
    
    # Image descriptor
    image_descriptor = 0x2C
    out.write(image_descriptor)
    
    # Left/top/width/height fields
    out.write(struct.pack("BB", left, top))
    out.write(width + height)
    
    # LCT flag (not using a local color table)
    local_color_table_specification = 0
    out.write(local_color_table_specification)
    
    # Code size
    sort_flag = 0
    code_size = 4
    out.write(struct.pack("BB", sort_flag << 7, code_size - 1))
    
    # Data subblock. Just one byte, so we'll have a single code length of 1
    LZW_END_OF_INFO_CODE = 0xFE
    out.write(LZW_END_OF_INFO_CODE & 0xFF)
    
    # Block terminator
    block_terminator = 0
    out.write(block_