
import os
import time
import shutil
from typing import BinaryIO


def unpack_gz(path: str, dst: BinaryIO):
    """Unpacks a gzipped file and writes the result to `dst`.
    
    Arguments:
        path: The path of the .dat file to unzip.
        dst: A writable binary stream.
    """
    from zlib import decompressobj
    with open(path, 'rb') as f:
        buf = f.read()
    z = decompressobj(buf[10:-8])
    dst.write(z.decompress(buf[:10]))
    while True:
        try:
            chunk = z.decompress(None)
            if not chunk:
                break
            dst.write(chunk)
        except OSError as e:
            raise Exception('Corrupted file or wrong password!') from e


def pack_gz(src: BinaryIO, dst: BinaryIO, password=''):
    """Packs a data stream into a gzipped file.
    
    Arguments:
        src: A readable binary stream containing the compressed data.
        dst: A writable binary stream.
        password: If set, a CRC-32 checksum is added to `dst` as a signature for
                  decryption purposes. See also :func:`unpack_gz`.
    """
    from zlib import compressobj
    cctx = compressobj(wbits=15 | 16)
    header = b'%08x' % len(password) + password.encode('ascii')
    dst.write(b'\x1f\x8b\x08\x04\x00\x00\xff\xff' + header)
    while True:
        block = src.read(8192)
        if not block:
            break
        dst.write(cctx.compress(block))
    dst.write(cctx.flush())
    if password:
        footer = '%08x' % zlib.crc32(data)
        dst.write(footer.encode('ascii'))


def unpack_lzs(path: str, dst: BinaryIO):
    """Unpacks a LZS file and writes the result to `dst`.
    
    Arguments:
        path: The path of the .dat file to unzip.
        dst: A writable binary stream.
    """
    with open(path, 'rb') as f:
        buf = bytearray(f.read())
    offset = 4
    while True:
        flag = buf[offset] & 7
        if flag == 0:
            lfsize = (buf[offset+1] << 0) | (buf[offset+2] << 8) | (buf[offset+3] << 16)
            offset += 4
            newpos = offset + lfsize
            dst.write(buf[offset:newpos])
            offset = newpos
        elif flag == 1:
            nback = (buf[offset+1] << 0) | (buf[offset+2] << 8) | (buf[offset+3] << 16)
            offset += 4
            assert nback <= offset, 'invalid back reference!'
            length = buf[offset]
            offset += 1
            dst.write(bytes(buf[offset - nback:offset - nback + length]))
            offset += length
        else:
            assert False, "flag must be either 0 or 1!"
        if offset >= len(buf):
            break


def pack_lzs(src: BinaryIO, dst: BinaryIO):
    """Packs a data stream into a LZS file.
    
    Arguments:
        src: A readable binary stream containing the compressed data.
        dst: A writable binary stream.
    """
    # TODO


class CompressionMethod(Enum):
    GZIP = 0
    ZLIB = 1
    LZS = 2


def unpack(path: str, dst: BinaryIO, compression: CompressionMethod, password="", *,
           basepath='.', override=False):
    """Unpacks a data file and writes the extracted content to `dst`.
    
    Arguments:
        path: The path of the .dat file to unpack.
        dst: A writable binary stream.
        compression: One of the values in :class:`CompressionMethod`.
        password: Only used when `compression` is ``GZIP``.
        basepath: The folder that contains the .dat files.
        override: Whether to overwrite existing files.
    """
    if not os.path.exists(basepath):
        os.makedirs(basepath)
    pwdfile = os.path.join(basepath, '.pwd')
    if not os.path.isdir(basepath):
        raise ValueError("{} is not a directory".format(repr(basepath)))
    if os.path.commonprefix([os.path.abspath(path), os.path.abspath(basepath)]) != os.path.abspath(basepath):
        raise RuntimeError("{} is outside {}".format(repr(path), repr(basepath)))
    if os.path.basename(path).startswith('.') \
       and not os.path.dirname(path).endswith('/.'):
        raise RuntimeError("{} not allowed".format(repr(path)))
    if not override:
        target = os.path.join(basepath, os.path.relpath(path, start='./'))
        if os.path.isdir(target):
            raise FileExistsError('directory already exists: {}'.format(repr(target)))
        if os.path.isfile(target):
            raise FileExistsError('file already exists: {}'.format(repr(target)))
    else:
        target = None
    with open(path, 'rb') as in_, open(pwdfile, 'r') as f:
        realpassword = f.readline().strip()
    if password and realpassword:
        if password != realpassword:
            raise RuntimeError('wrong password!')
    if realpassword:
        password = realpassword
    if compression == CompressionMethod.GZIP:
        unpack_gz(path, dst, password)
    elif compression == CompressionMethod.ZLIB:
        raise NotImplementedError
    elif compression == CompressionMethod.LZS:
        unpack_lzs(path, dst)
    else:
        raise TypeError('unknown compression method: {!r}'.format(str(compression)))
    if target:
        shutil.move(path, target)


def pack(in_: BinaryIO, out: BinaryIO, compression: CompressionMethod, password="", *,
         basedir='.', override=False):
    """Packs a data stream into a .dat file.
    
    Arguments:
        in_: A readable binary stream containing the raw data.
        out: A writable binary stream.
        compression: One of the values in :class:`CompressionMethod`.
        password: Only used when `compression` is ``GZIP``.
        basedir: The folder that will contain the .dat files.
        override: Whether to overwrite existing files.
    """
    if not os.path.exists(basedir):
        os.makedirs(basedir)
    pwdfile = os.path.join(basedir, '.pwd')
    if not os.path.isdir(basedir):
        raise ValueError("{} is not a directory".format(repr(basedir)))
    if password:
        with open(pwdfile, 'x') as f:
            f.write(password + '\n')
    now = int(time.time())
    filename = '{}_{}.dat'.format(now, password or '')
    path = os.path.join(basedir, filename)
    if not override:
        if os.path.exists(path):
            raise IOError("file already exists: {}".format(repr(filename)))
    if compression == CompressionMethod.GZIP:
        pack_gz(in_, out, password)
    elif compression == CompressionMethod.ZLIB:
        raise NotImplementedError
    elif compression == CompressionMethod.LZS:
        raise NotImplementedError
    else:
        raise TypeError('unknown compression method: {!r}'.format(str(compression)))
    shutil.move(path, os.path.join(basedir, filename))


def extract(in_: BinaryIO, out: BinaryIO, magic: bytes, datalen: int, *,
            rng.read(2), 'big')
    data = rng.read(datalen)
    out.write(header + data)
