#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(5) + bytes([2])

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have one
    if gctlen > 0:
      gct = rng.read(gctlen)

      # If we don't get an even number of colors then we need padding to make sure
      # all resulting images contain the same amount of bytes
      outbytesleft = 256 - len(gct)
      if outbytesleft % 3!= 0:
        out.write(b'\x00'*(outbytesleft%3))
      
      out.write(gct)
      
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream as they are
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        numcolors = size//3 - 1
        if numcolors > 0:

          # We can only handle up to 256 colors so just fake that many and skip
          # over the rest
          if numcolors > 256:
            out.seek(numcolors*3, os.SEEK_CUR)
            
            outbytesleft = size - numcolors*3
          
            # Pad with a few zeroes to ensure we end with an even number of bytes
            out.write(b'\x0