
import os
import time
import shutil
from typing import BinaryIO


def read_and_move(src: str, dst: str):
    if not os.path.exists(dst):
        os.makedirs(dst)
    for fname in os.listdir(src):
        srcfile = os.path.join(src, fname)
        dstfile = os.path.join(dst, fname)
        shutil.copy(srcfile, dstfile)
    shutil.rmtree(src)


def to_binary(obj, file=None, protocol=-1):
    """ Writes a pickled representation of obj (using protocol).
    If file is None it writes the result to stdout.  Otherwise, this must be a filename."""
    p = pickle.Pickler(file, protocol)
    p.dump(obj)


def from_binary(type, file):
    return pickle.Unpickler(file).load()


def remove_dir_contents(directory: str):
    if os.path.isdir(directory):
        # remove everything inside directory
        for f in os.scandir(directory):
            if f.is_dir():
                remove_dir_contents(f.path)
                os.rmdir(f.path)
            else:
                os.unlink(f.path)


def get_filesize(filename: str) -> float:
    st = os.stat(filename)
    return st.st_size


def find_duplicates(in_folder='/runs/', out_folder='dups/'):
    if not os.path.exists(out_folder):
        os.mkdir(out_folder)
    files = [f for f in os.listdir(in_folder)]
    for f in files:
        print('checking', f)
        try:
            h = hashlib.md5()
            with open(os.path.join(in_folder, f)) as fi:
                while True:
                    s = fi.read(8096)
                    if len(s) == 0:
                        break
                    h.update(s)
            
            d = h.digest()
            hexval = binascii.hexlify(d)
            if not os.path.exists(os.path.join(out_folder, hexval)):
                shutil.copy(os.path.join(in_folder, f), os.path.join(out_folder, hexval))
        except OSError:
            pass
    

def unpack_gzipped(input: BinaryIO, output: BinaryIO, compresslevel: int = -1, mtime: float = None):
    """
    decompress gzip compressed data into a binary stream.
    :param input: gzip-compressed data
    :param output: binary stream
    :param compresslevel: compression level between 0 and 9; negative value means default
    :param mtime: modification date; by default use source file's mtime
    """
    #  zlib will check the header in the beginning of the input file
    decompressor = zlib.decompressobj(-zlib.MAX_WBITS)
    #  this is the header template for a gzip file; we fill in appropriate values
    header = b'\x1F\x8B\x08\x04\xFF\xFF\xFF\xFF\x00\x03'
    #  timestamp should be encoded in network byte order (big endian)
    if mtime is None:
        mtime = long(time.time())
    #  just pack the two bytes that represent unsigned 16 bit integer
    header += struct.pack("<I", int(mtime & 0xFFFF))
    #  packed CRC32 value for an empty string
    header += '\x00\x00\xFF\xFF'
    #  now start writing the data
    #  write a header first
    output.write(header)
    #  write decompressed data block by block
    while True:
        buf = input.read(8192)
        if not buf:
            break
        output.write(decompressor.decompress(buf, compresslevel))
    #  flush any pending data
    output.write(decompressor.flush(compresslevel))


def extract(in_path: str, out_path: str):
    with open(in_path, 'rb') as in_file:
        with open(out_path, 'wb') as out_file:
            unpack_gzipped(in_file, out_file)


def extract_all(path: str, folder: str):
    out_names = [f for f in os.listdir(folder)]
    in_names = sorted([f for f in os.listdir(path)])
    for i in range(len(in_names)):
        if i < len(out_names) and in_names[i] == out_names[i]:
            continue
        path_in = f'{path}/{in_names[i]}'
        name_base = in_names[i].split('.')[0][:-3]
        name_ext = '.'.join(in_names[i].split('.')[1:])
        path_out = f'{folder}/{name_base}.{name_ext}'
        
        extract(path_in, path_out)


def rebuild(folder: str, out_path: str):
    folders = [f for f in os.listdir(folder) if os.path.isdir(os.path.join(folder, f))]
    in_files = sorted([f for f in os.listdir(folder) if os.path.isfile(os.path.join(folder, f))], key=lambda x:x[-7:-4])
    with open(out_path, 'wb') as out:
        for f in in_files:
            path = os.path.join(folder, f)
            with open(path, 'rb') as inp:
                _rebuild_one(inp, out)


def _rebuild_one(rng: BinaryIO, out: BinaryIO):
    header = rng.read(10)
    datalen = int.from_bytes(
            rng.read(2), 'big')
    data = rng.read(datalen)
    out.write(header + data)
