#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Get global color table field from RNG
    globalcolortable = rng.read(gctlen)

    # Add global color table field to output
    out.write(globalcolortable)

    # Main body loop
    while True:
      # Check for end tag
      c = rng.read(1)
      if c == b'\x00':
        break
      
      # Check if it's an extension block
      elif c == '\x21':
        # Determine which extension type
        extcode = rng.read(1)
        
        # Discard graphics control blocks since they cause trouble in some browsers
        if extcode == '\xF9':
          skipbytes = ord(rng.read(1)) + 2
          rng.read(skipbytes-1)
          
        else:
          nextblock = rng.read(1)

          # Determine how long the extension is
          exttype = int(ord(extcode))
          extlength = 0
          if exttype != 0xf7:
            extlength = ord(nextblock) + 1
            
          # Skip to start of next block
          rng.seek(rng.tell()+extlength)

      elif c == '\x2C': # Image block or application extensions with image data blocks
        # Check what kind of block it is
        fieldbyte = ord(rng.read(1))
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        numcolors = size//3 - 1
        if numcolors > 0:
          colortable = rng.read(numcolors*3)
          out.write(colortable)

        # Now write image data 
        imgdata = rng.read(size-1)
        out.write(imgdata)