#!/usr/bin/env python3
'''
    Generate a JSON file that describes all of the commands and events.
'''
import json
import re
from collections import OrderedDict, namedtuple

def parse_desc(desc):
    '''Parse a command or event description and extract it's arguments.

        Args: desc (str) - The command or event description to parse

        Returns: The argument descriptor for this command or event
                     args (dict<str, str>) - Map from arg name to type string

    '''
    # First, find the argument pattern in the documentation
    match = None
    if '*Arguments:' in desc:
        match = re.search('\*Arguments:\*\n\n([^;]*)\n', desc)
    elif '- Arguments' in desc:
        match = re.search('- Arguments:\s*\n\s*`(.*)`', desc, re.DOTALL | re.MULTILINE)
    else:
        raise Exception("Unable to determine how to parse command description")

    if not match:
        return dict()

    arg_text = match[1]
    arg_list = []
    # Then split apart the argument text by commas
    parts = [part.strip() for part in arg_text.split(',')]
    for part in parts:
        # Finally, each argument can be further broken into name and type
        name, _, typename  = part.partition(': ')
        arg_list.append((name, typename))

    return {arg.lower(): typename for arg, _ in arg_list}

def generate_json(json_filename='command-and-event-descriptors.json'):
    '''Generate the JSON file from the command and event names found in
       xcbproto library files.
    '''
    cmd_data = OrderedDict({})
    evt_data = OrderedDict({})

    # Iterate over every message defined in an XCB protocol file
    for header_file in ['xproto.xml', 'xtest.xml']:
        with open('../include/' + header_file, 'r') as f:
            in_messages = False
            for line in f:
                # We care only about the Messages block
                if '<Messages>' in line:
                    in_messages = True
                elif '</Messages>' in line:
                    break
                elif not in_messages:
                    continue

                # <message> tags contain the information we need
                match = re.match('<message opcode="(\d+)" name="(.*?)" since="(.*?)">', line)
                if match is None:
                    continue

                # For command messages, use the opcode as the key
                if 'request' in match.group(2).lower():
                    if match.group(1) not in cmd_data:
                        cmd_data[match.group(1)] = {'opcode': int(match.group(1)), 'name': match.group(2)}
                    data['args'] = parse_desc(match.group(0))
                    continue

                # For events, use the name as the key
                if match.group(2)[-5:] != "Event" \
                   or 'notify' in match.group(2).lower() \
                   or 'window' in match.group(2).lower():
                    continue

                evt_data[match.group(2)] = {'opcode': int(match.group(1)), 'name': match.group(2