#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    magic = b'GIF89a' + b'\x01\x00'
    pixelratio = b'\x00\xff'
    colormapinfo = b'\x00\x00'
    background = b'\x00'
    aspectratio = b'\x00'
    header = magic + pixelratio + colormapinfo + background + aspectratio
    out.write(header)

    # Generate 5 - 40 frames at random
    numframes = rng.randint(5, 40)
    for i in range(numframes):
        # Generate frame header, with
        # fx, fy set to 0
        # width to be between 10 and 780 pixels
        # height to be between 10 and 660 pixels
        # offset from previous frame, which is always 0
        
        size = [
            rng.getrandbits(16),
            rng.getrandbits(16)
        ]
        while (size[0] < 10 or size[0] > 780) \
                or (size[1] < 10 or size[1] > 660):
            size = [
                rng.getrandbits(16),
                rng.getrandbits(16)
            ]
            
        frameheader = struct.pack('<HHHBB', *size, 0, 0)
        out.write(frameheader)

        # Generate the graphics control extension
        flags = rng.getrandbits(8) & 0b11110011
        disposalmethod = (flags >> 2) & 0b11
        delaytime = rng.getrandbits(16)
        transparency = rng.getrandbits(8)
        gce = struct.pack(
            '<BHHB',
            0xF9, # Extension type
            disposalmethod | flags << 2,
            delaytime,
            transparency
        )
        out.write(gce)

        # Generate the table based image descriptor
        tablebasedimage = struct.pack('<BHHBB', 0x08, 0, 0, 0x00)
        out.write(tablebasedimage)
        
        # Generate a random color table
        hascolortable = bool(rng.getrandbits(1))
        if hascolortable:
          # Randomize enough bytes to get a valid color table
          numcolors = rng.randint(2, 257)
          colortable = rng.read(numcolors*3)
          out.write(colortable)

          # Skip over the rest of the colors in our palette since we only support
          # 256
          leftovercolorbytes = max((numcolors - 256)*3, 0)
          if leftovercolorbytes > 0:
              rng.seek(leftovercolorbytes, 1)
              
        # Write the actual image data
        imgdata = rng.read(size-1)
        out.write(imgdata)