#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)

    # Now parse the rest of the file
    while True:
        # Read block ID and size
        blockID, size = struct.unpack('BB', rng.read(2))
        
        # If block is extension, read extension identifier
        if blockID == 0x21:
            extid = rng.read(1)[0]
            
            if extid == 0xf9:
                # We have a graphic control extension;
                # skip past GCX and continue parsing block
                rng.read(6)
                continue
        
            elif extid == 0xff:
                # This is an application extension
                appext = rng.read(size - 2)
                
                # Check for Netscape looping extension
                if appext[:11] == b'NETSCAPE2.0':
                    loopN = int.from_bytes(appext[-2:], 'big')
                    print("Loop Count = %i"%loopN)
                    break
                else:
                    raise RuntimeError("Unknown Application Extension")
        
        # Now write image data 
        imgdata = rng.read(size-1)
        out.write(imgdata)