#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)
    width = int.from_bytes(lsd[0:2], byteorder='big')
    height = int.from_bytes(lsd[2:4], byteorder='big')
    flags = lsd[4]
    bgcolorindex = lsd[5]
    aspectratio = lsd[6]
    
    # Add to struction by packing it as big endian 
    # (we need an array for that!)
    lsdfields = [
        *width.to_bytes(2, 'be'),
        *height.to_bytes(2, 'be'),
        flags,
        bgcolorindex,
        aspectratio
    ]

    # Pack the LSD with little endian byte order so it can be written
    # directly to our outstream
    packedlsd = struct.pack('<HHBBB', *lsdfields)
    out.write(packedlsd)

    # Set current palette entry index to 0
    paletenum = 0

    while True:
      # Read field byte from input 
      fieldbyte = rng.read(1)[0]

      # Start a new graphic control extension block
      if (fieldbyte==0x21):

        # Write field byte back again because we need two bytes        
        out.write(struct.pack('=B', fieldbyte))

        # Read block size 
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        numcolors = size//3 - 1
        if numcolors > 0:
          colortable = rng.read(numcolors*3)
          out.write(colortable)

        # Now write image data 
        imgdata = rng.read(size-1)
        out.write(imgdata)