#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)
#     print('lsd', bytes(lsd).hex())
    width  = int.from_bytes(lsd[0:2], 'little') + 1
    height = int.from_bytes(lsd[2:4], 'little') + 1
#     print('width', width, 'height', height)
    packed_byte = lsd[4]
    flags = packed_byte & 0x7

    # Add LSD field to output
    out.write(lsd[:6])

    if (flags == 0):
        out.write(b'\x00\x3B')
    else:
        ext = rng.read(1)[0]

        while True:
            # End of extensions
            if (ext == 0x00):
                break

            # Plain text extension
            elif (ext == 0xF9 or ext == 0xFE):
                size = rng.read(1)[0]

                if (size!= 12):
                    raise Exception('Invalid plain text block length'+
                                    '{}.'.format(size))
                
                out.write(bytes([ext]))
                out.write(rng.read(11))

            # Application extension
            elif (ext == 0xFF):
                out.write(bytes([ext]))
                app_id = rng.read(11)
                out.write(app_id)
            
                size = rng.read(1)[0]
                out.write(size.to_bytes(1, 'big'))

                block = rng.read(size)
                out.write(block)

            else:
                raise Exception("Bad block type '{}'".format(ext))
        
            ext = rng.read(1)[0]
        
        # Block terminator
        out.write(b'\x00')
    
    # Prepare loop extension fields
    packed_byte = rng.read(1)[0]
    has_transparency = bool((packed_byte >> 3) & 0x1)
    interlaced       = bool((packed_byte >> 6) & 0x1)
    sorted           = bool((packed_byte >> 7) & 0x1)

    # Loop extension
    out.write(bytes([0x21, 0xFF, 0x0B,
                     0x4E, 0x45, 0x54, 
                     0x53, 0x43, 0x41, 
                     0x50, 0x45, 0x32, 
                     0x2E, 0x30, 0x03, 
                     0x01, packed_byte]))

    # Application specific data subblocks
    out.write(bytes([0x04, 0x01, 0x00, 0x01]))
    delay = rng.read(2)
    out.write(delay)
    out.write(b'\x00\x00\x00')

    # Other application specific data subblocks
    app_id_length = rng.read(1)
    app_id        = rng.read(ord(app_id_length)-1)
    out.write(app_id_length+app_id+b'\x00')

    out.write(bytes([0x00]))
    

    # Create palette entry cache
    palmap = {}

    def get_palentry(rgb):
        rgbkey = tuple(rgb)

        if not rgbkey in palmap:
            palmap[rgbkey] = len(palmap)

        return palmap[rgbkey]

    # Encode alphabet for data encoding
    def prepare_alphabet():
        num_chars = 256

        # Build our own custom alphabets by first selecting a few colors from
        # the palette as markers and then choosing where these markers can
        # appear. We then use the remaining 252 characters to represent how 
        # many times each marker appears
        
        marker_colors = [get_palentry(rng.read(3)) for _ in range(4)]

        # Find all possible variations on the marker colors
        marker_variations = []
        for r1 in range(0,256,4):
            for g1 in range(0,256,4):
                for b1 in range(0,256,4):
                    for r2 in range(0,256,4):
                        for g2 in range(0,256,4):
                            for b2 in range(0,256,4):
                                variation = [((r1+rr)%256,(g1+gg)%256,(b1+bb)%256)\
                                             for rr, gg, bb in \
                                             zip(range(-64,64),
                                                 range(-64,64),
                                                 range(-64,64))]

                                variation = [(get_palentry(_v),(r2+rr)%256,\
                                              (g2+gg)%256,(b2+bb)%256)\
                                             for _v,_vr,_vg,_vb in\
                                             zip(variation,
                                                  range(-64,64),
                                                  range(-64,64),
                                                  range(-64,64))]
                                
                                marker_variations.append(variation)


        seen_markers = set()
        used_indexes = [[] for _ in range(len(marker_variations))]
        
        for variation in marker_variations:
            for midx, mc in enumerate(variation):
                idx = marker_colors.index(mc[0])
                idx //= 4
                
                if (not idx in seen_markers):
                    used_indexes[midx].append(idx)
                    seen_markers |= {idx}

        
        # Generate the actual alphabet
        characters = [''] * num_chars
        for midx, used_indexes in enumerate(used_indexes):
            for _unused in range(len(characters)-1):
                rand_idx = rng.read(1)[0]%len(used_indexes)
                chr = chr(used_indexes[rand_idx]*4 + midx*4);

                characters[_unused] = chr
                used_indexes.pop(rand_idx)
            
        return characters
            

    def encode_string(_str, alphabet):
        encoded_str = b''
        mdict = dict([(c, alphabet.index(c)) for c in alphabet[:-1]])
        seq_start_idx = None

        for idx, char in enumerate(_str):
            try:
                enc = mdict[char]
            except KeyError:
                pass
            else:
                if not seq_start_idx:
                    encoded_str += enc.to_bytes(1, 'big')

                continue

            if seq_start_idx:
                encoded_str += ((idx-seq_start_idx)*255 + enc).to_bytes(2, 'big')

                seq_start_idx = None
            else:
                seq_start_idx = idx

        if seq_start_idx:
            encoded_str += ((len(_str)-seq_start_idx)*255).to_bytes(2, 'big')

        return encoded_str

    def pack_color(rgb):
        r,g,b = rgb;
        return bytearray(((b<<4)|r,(g<<4)|r)).hex().encode('ascii');

    def chunkify(lst, n):
        return [lst[i::n] for i in range(n)]

    def iter_chunks(iterable, n):
        lst = list(iterable)
        return chunkify(lst, n)


    # Set up skip minimum so that trailing zeros don't slow us down too much
    skip_minimum = max(min(height//16, width//4),max(width,height)//200)

    # Backreferences are limited to skipping 3 frames back at most
    backref_codes = [code.decode('utf-8') for code in iter_chunks(prepare_alphabet(), 3)][::-1]

    alphabet = prepare_alphabet();

    # Main body of generation
    num_colortables = (1<<(bits-1))
    colortable_size = num_colortables*(3+(has_transparency))*3//8

    # Calculate number of data chunks required
    data_chunk_size = colortable_size + max(colortable_size//skip_minimum,1)
    data_offsets = []

    for chunk_number in range(height//data_chunk_size+1):
        offset = len(out.getvalue())
        dchunk = rng.read(data_chunk_size)
        data_offsets.append(offset)
    
        # Invert every other column
        indices = [*range(colortable_size//3,0,-3)] + \
                  [*range(colortable_size//3,colortable_size,3)]

        dchunk = bytes([dchunk[indices[i]] for i in range(len(dchunk))])

        # Chunk parity check
        checksum = sum(iter_chunks(dchunk,3))%(2**bits)
        checksum = [checksum>>bits-i&0x1 for i in range(bits)][::-1]

        enc_checksum = []
        
        for idx, cbit in enumerate(checksum):
            try:
                enc_checksum.append(alphabet[(cbit<<bits-idx)&0xFF])
            except IndexError:
                break

        checksum = b''.join(enc_checksum)
        assert len(checksum)==len(enc_checksum)

        out.write(dchunk)
        out.write(checksum)

    # Palette
    palette = b''
    for idx,[_,rgb,_] in enumerate(sorted(palmap.items())):
        palette += get_palentry(rng.read(3))

    out.write(palette)

    # Write code section
    num_codes = 2**bits
    clear   = 1 << bits - 1
    end     = 1 << bits
    codes   = [rng.read(1) for i in range(num_codes)]
    values  = []
        
    for y in range(height-1,-1,-1):
        line = ''

        for x in range(width):
            packed = pack_color(rng.read(3))
            
            if packed == clear:
                line += str(clear)
                values.append('')
                
            elif len(values[-1]) < skip_minimum:
                line += str(clear)
                values.append('')
                values[-2] += codes[int.from_bytes(packed,'big')]
            
            elif any(i in values[-1] for i in backref_codes):
                index = int.from_bytes(packed, 'big')
                marker = min(backref_codes, key=lambda k: values[-1].find(k))

                # If after the marker we find another occurance of this marker
                # then it's an escaped character
                if values[-1][values[-1].find(marker)+len(marker)] == marker:
                    # This is an escape sequence
                    line += str(clear)
                    values.append('')

                    # Append escaped char (as string) to previous value
                    values[-2] += codes[index]
                    
                else:
                    # This is a normal reference
                    line += str(clear)
                    values.append('')
                    
                    ref_index = values[-1].rfind(marker)+len(marker)
                    values[-2] += values[-1][ref_index:]
                    values[-1] = values[-1][:ref_index] + values[-1][marker:]
                    
            else:
                # If no matches are found just append to last value
                index = int.from_bytes(packed, 'big')
                values[-1] += codes[index]

                line += ''.join(['{:>3}'.format(str(v))
                                for v in values[-1]])
                
        encoded_data = encode_string(line, codes)
        out.write(encoded_data)


    # Image descriptor
    out.write(struct.pack('<HHHBB',
                         width      % 0x10000,
                         height//256,
                         0,
                         0,
                         0
                        ))

    # Graphics control extension with transparent color flag set
    out.write(b'\x21\xf9\x04\x01\x00\x00\x00\xff\xff')

    # Ending image descriptor
    out.write(b';')

    # Codes dictionary
    out.write(struct.pack('<B',len(codes)))
    out.write(b''.join(codes))

    # Data offsets
    out.write(struct.pack('<B', len(data_offsets)))

    for doffset in data_offsets:
        out.write(doffset.to_bytes(2, 'big'))
        
    

