#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the files in gif_files as frames.
def randomize(out: BinaryIO, *gif_files):
    # Header size is 6 bytes for signature and 7 bytes for version string.
    header_size = len('GIF89a') + len('Version:') + 1

    # Compute total size of all data before we start writing anything to the output.
    initial_data_size = sum([len(file) for file in gif_files]) - header_size
    final_data_size = 0x52494447 - initial_data_size
    assert final_data_size % (4*2) == 0, 'final_data_size must be divisible by 4'

    # Write out the new headers based on the information in `gifs`.
    # TODO(yann): Actually do this right per spec
    # Since our gifs are likely not going to change much, it's safe enough.
    # We just need to copy the original header except that the width/height needs
    # to be multiplied by two since each frame will actually have a larger image.
    for filename in gif_files:
        with open(filename, 'rb') as f:
            f.seek(header_size+2)
            width, height = struct.unpack('<HH', f.read(4))
            actual_width, actual_height = width * 2, height * 2

            # Original image block:
            #     [=] Image descriptor U+2C
            #     9B [Width]   2xB [Height]      5B [Packed Fields]         B [LZW min code size]
            out.write(b'\x2c\x00\x00' + width.to_bytes(2,'little'))
            out.write(height.to_bytes(2,'little') + b'\x00\x00\xff\xf9')
            
            # Data block
            while True:
                data = f.read(1024)

                if not data or len(data) < 1024:
                    break
                
                out.write(data)
            
            # Trailer block:
            #     [;] Block terminator U+3B
            #     0B [Block Size]  1xB [Terminator]
            out.write(b';\x00')
    
    # Now add padding until we reach the end of the file.
    # This ensures we don't accidentally write beyond the EOF and avoid overflowing.
    for i in range((final_data_size // (4*2))-1):
        out.write(b'\x2c\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xf9;\x00\x00')

