#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
        rng.read(2),
        'little', signed=False
    )

    globalcolorspace = bool(int.from_bytes(
        rng.read(1)[0:1],
        'little', signed=True
    ))

    width = int.from_bytes(
        rng.read(2),
        'little', signed=False
    )

    height = int.from_bytes(
        rng.read(2),
        'little', signed=False
    )

    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    if width > 65535 or height > 65535:
        print('[!] Invalid size')
        exit(-1)
    
    loopcounter = int.from_bytes(
        rng.read(1),
        'little', signed=True
    )

    colorbits = {
        1 : (2 ** 1, [b'\x01']),
        2 : (2 ** 2, [b'\x02']),
        4 : (2 ** 4, [b'\x04']),
        8 : (2 ** 8, []),
    }[int.from_bytes(
            rng.read(1),
            'little', signed=False
        )]

    bgindex = int.from_bytes(
        rng.read(1),
        'little', signed=False
    )

    aspectratio = int.from_bytes(
        rng.read(1),
        'big', signed=False
    )

    extensionblock = True
    while extensionblock:
        exttype = rng.read(1)[0:1]

        if exttype == b'\xfe':
            terminator = rng.read(1)[0:1]

            if not ((terminator == b'\x00')):
                print("[-] Unterminated block")
                break
            
            extensionblock = False
        
        elif exttype == b'\xf9':
            passcode = rng.read(1)[0:1]
            flags = rng.read(1)[0:1]
            delaytime = int.from_bytes(
                rng.read(2),
                'little', signed=False
             )
            
            transparentflag = flags & 1
            inputflag = flags >> 1 & 1
            disposalmethod = flags >> 2 | flags << 6 & 7

            transparencyindex = rng.read(1)[0:1]
            terminator = rng.read(1)[0:1]
            
            if not ((terminator == b'\x00')):
                print("[-] Unterminated block")
                break

            continue
        
        else:
            # Skip data blocks of unknown type
            length = int.from_bytes(
                rng.read(2), 
                "little", signed=False
            )

            skipcount = 0
            for _ in range(length):
                byte = rng.read(1)[0:1]

                if byte!= b'\x00':
                    skipcount += 1
                    
                    if skipcount >= len(exttype + terminator + byte):
                        break
                
                elif skipcount == 0 and \
                     exttype + terminator + byte == b'\x0b\xff\x04\x0c\x00\x00':
                     skipcount += 1

            continue


    palettesize = int.from_bytes(
        rng.read(2), 
        "little", signed=False 
    )

    if palettesize % 3!= 0:
        print('[-] Incorrect palette size')

    rgbcolors = []
    for i in range(palettesize // 3):
        r = int.from_bytes(
            rng.read(1), 
            "little", signed=False 
        )

        g = int.from_bytes(
            rng.read(1),
            "little", signed=False 
        )

        b = int.from_bytes(
            rng.read(1), 
            "little", signed=False 
        )

        if r < 0 or g < 0 or b < 0:
            print("[!] Negative RGB value encountered")

            r -= max(r, -255)
            g -= max(g, -255)
            b -= max(b, -255)

        
        if r > 255:
            r = 255

        if g > 255:
            g = 255

        if b > 255:
            b = 255


        rgbcolors.append((r, g, b))

    
    imagedata = b''
    for y in range(height):
        for x in range(width):
            pixelindex = int.from_bytes(
                rng.read(1), 
                "little", signed=False
            )

            if pixelindex > len(rgbcolors):
                print('[!] Pixel index is too large')
                continue
            
            imagedata += bytes([pixelindex])
    

    trailer = b'\x00\x3B'
    out.write(header)
    out.write(struct.pack('<H', datalen))
    out.write(globalcolorspace.to_bytes(1, byteorder='little'))
    out.write(width.to_bytes(2, byteorder='little'))
    out.write(height.to_bytes(2, byteorder='little'))
    out.write(loopcounter.to_bytes(1, byteorder='little'))
    out.write(colorbits[0].to_bytes(1, byteorder='little'))
    out.write(bgindex.to_bytes(1, byteorder='little'))
    out.write(aspectratio.to_bytes(1, byteorder='big'))
    out.write(extensionblock.to_bytes(1, byteorder='little'))
    out.write(transparencyindex)
    out.write(terminator)
    out.write(palettesize.to_bytes(2, byteorder='little'))
    for col in rgbcolors:
        out.write(col[0].to_bytes(1, byteorder='little'))
        out.write(col[1].to_bytes(1, byteorder='little'))
        out.write(col[2].to_bytes(1, byteorder='little'))
    out.write(imagedata)
    out.write(trailer)

