
import os
import time
import shutil
from typing import BinaryIO

# random number generator `rng` (/dev/urandom)
rng: BinaryIO
# path to output file
outpath: str


def setup() -> None:
    """
    Setup the module-level variables by parsing command line arguments, and open both stdin & stdout files as binary
    streams in read / write mode (rb / wb).

    Global variables modified:
        - rng: reference to a stream for reading from `/dev/urandom`.
        - outpath: absolute pathname of the generated file.
    """
    global rng, outpath
    parser = argparse.ArgumentParser(description='Generate n bytes of random data.')
    parser.add_argument('-n', '--size', type=int, required=True,
                        help='The number of bytes that will be written to the specified output file.', metavar='number')
    parser.add_argument('output', nargs='?', default='/dev/stdout',
                        help="""The output location where the random data should be stored. This is a string containing
                                either an absolute or relative pathname (relative paths are relative to the current working
                                directory). If this argument is omitted, then the standard output device is used instead.""",
                        metavar='pathname')
    args = parser.parse_args()

    # Get the absolute path name for the output file. The damn `shutil.copyfileobj()` requires one...
    outpath = os.path.abspath(args.output)
    if not outpath:
        print("Invalid output filename", file=sys.stderr)
        sys.exit(-1)
    
    try:
        rngf = open('/dev/urandom', 'rb')
    except OSError as err:
        print("{}".format(err), file=sys.stderr)
        sys.exit(-1)
    else:
        with rngf:
            rng = rngf

    return
    
def run() -> None:
    """Write randomly generated data to the user selected output."""
    with open(outpath, "wb") as out:
        generate_data(out)
    return


def teardown():
    """Close all file descriptors opened during program execution."""
    rng.close()
    return

if __name__ == "__main__":
    setup()
    t0 = time.time()
    run()
    dt = time.time() - t0
    teardown()
    print("\nTotal elapsed time: {:.5f}s\n".format(dt))



def generate_data(out):
    """
    Generate random data and write it into file `out`.

    Arguments:
        - out: A writable file descriptor.
    """
    header = int.to_bytes((0xA4A3B5C6, 0x987601FEDCBA9876),
                          length=16, byteorder='big')   # header magic numbers
    datalen = int.from_bytes(
            rng.read(2), 'big')
    data = rng.read(datalen)
    out.write(header + data)
