#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 2 bytes of length of block (N).
#   3. N bytes of zero padding to align to 8 byte blocks.
#   4. 1 byte of version ("\x07").
#   5. 3 bytes of width & height of image in little endian order.
#   6. 1 byte of flags which set LZW minimum code size to 11 and has no global color table.
#   7. 2 bytes of 768 zeros for palette data.
#   8. 2 bytes that are both black, which is the gloabl color index to use when drawing each pixel's color index from the local color table.
#   9. A total of 257 pixels, with the first pixel being at position (0, 0) top left, continuing onwards counter clockwise. Each pixel consists of two consecutive bytes containing its X & Y coordinate followed by an additional byte indicating what color to draw it with. The last byte of the sequence contains a special value 0 which indicates that there are no more pixels to come after this one.
def generate_gif(width: int = 100, height: int = 100, out: BinaryIO = None):
    # If no output stream was given just create one in memory.
    if not out:
        out = BytesIO()

    # Write the gif header.
    out.write("GIF89a")

    # Calculate the block length and write it as a short into the stream.
    out.write(struct.pack("<H", 0))

    # Write some extra zeroes to make sure our total bytes per frame will be divisible
    # by 8.
    out.write(bytes(b"\x00" * 12))

    # Write the logical screen descriptor.
    out.write(struct.pack("<BBHHHH", 7, 0, width, height, 0, 0))

    # Write the global color table with all colors set to black.
    out.write(bytes([0] * 768))

    # Set the local color table flag.
    out.write(struct.pack("<B", 0b10000000))

    # Write the starting point of the graphics control extension block. This points to where we will later put the delay time for the animation. We can ignore most of these fields since we're just going to go directly from frame to frame without doing any complex animations or anything like that.
    out.write(struct.pack("<BB", 0b1010, b""))

    # Write the ending point of the graphics control extension block.
    out.write(struct.pack("<BBB", b"", 0, b""))

    # Set the graphic control extension block flags to 0.
    out.write(struct.pack("<B", 0))

    # Write out the application extension block. This tells the browser how many frames
    # are in the animation so that it knows when to wrap back around to frame 0 again.
    out.write(bytes([
        0x21, 0xFF,           # Extension type (application)
        0x0B,                 # Block size in bytes (11 bytes below)

        ord('A'), ord('N'),   # Application identifier (ASCII "AN")
        0x03,                 # Sub-block size in bytes (3 bytes below)

        0b11111000 | 1        # Flags (don't use local colour tables, has background color, sort flag, interlace flag)
                              # Delay time in milliseconds (still have no idea why you'd ever want to change this but ok.)
                              0, 0,    0]))

    # For each row, write out the packed byte indicating the start of a new line.
    out.writelines((b"\x00,"))

    # Keep track of the current x position on this row and iterate through every possible column.
    line_start = True
    x = 0
    while x < width:
        # At the beginning of the row, indicate the start of a line.
        if line_start:
            line_start = False

            out.write(struct.pack('<BB', x % 256, y))

        else:
            # Otherwise write out another pixel along with specifying that there are no more pixels coming down the line.
            out.write(struct.pack('<B', 0))

        # Increment the X position.
        x += 1


    return out

