#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(5) + bytes([2])

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have one
    if gctlen > 0:
      gct = rng.read(gctlen)

      out.write(gct)

    # Read next byte which is the starting point for variable portions of the
    # file
    firstbyte = int(ord(rng.read(1)))
    
    # Iterate through blocks until you reach end of file
    while True:

      # We exit when we see the last two bits set
      if firstbyte & 7 == 7:
        return

      # Add the first byte with the second bit set to signify it's extension
      out.write(struct.pack('=B', firstbyte | 0x20))
      
      # Get the length of the following extension type
      exttypelen = int(ord(rng.read(1)))

      # We also exit when these two bits indicate the last block
      if exttypelen == 0:
        return

      # Add the next byte with the third bit set to signify it's an application 
      # extension
      out.write(struct.pack('=B', exttypelen | 0x1F))

      # The next exttype will be a plain text extension
      out.write(b'CE')

      # Add the length of the string 'Netscape2.0'
      out.write(struct.pack('=B', len(b'NETSCAPE2.0')))

      # Finally write the actual string itself
      out.write(b'NETSCAPE2.0')

      # Now read the application ID length field which should always be three
      appidlen = int(ord(rng.read(1)))

      # If it isn't throw error
      assert appidlen == 3

      # Write the application ID
      out.write(b'\x03\x01\x00')

      # Next read the subblock
      sblocklen = int(ord(rng.read(1)))

      # Exit when we see the last two bits set
      if sblocklen == 0:
        break

      # Add the final byte with the fourth bit set
      out.write(struct.pack('=B',sblocklen|0x80))

      # Get the delay time
      delaytime = [0]*3

      # Iterate each value in delay time
      for i in range(3):
        
        # This loop runs three times because delays are stored as BCD values 
        # where each digit takes four bytes
        highfour = int(ord(rng.read(1)))
        
        lowfour = int(ord(rng.read(1)))

        delaytimestr = ("%d" % ((highfour<<4)|lowfour)).encode()

        delaytime[i]=delaytimestr
        
      # Convert the list of strings to a single long binary string
      delaytimesstr = b"".join(delaytime)

      # Write this delay time to output
      out.write(delaytimesstr)
      
      
      # While true loop to fill out the subblocks
      while True:
        sblocklen = int(ord(rng.read(1)))

        # If the final two bits aren't set then we're at the end
        if sblocklen & 0b11!= 0b11:
          
          # Reset the most significant bit to zero
          sblocklen &= ~0b10000000

          # Send the same byte if we're on our last iteration else send the 
          # byte with the final bit set
          if sblocklen <= 0xff:
            
            out.write(struct.pack('=B',sblocklen))

            # Break out of inner while loop now that we've found the end of 
            # the subblocks
            break
            
          else:
            out.write(struct.pack('=B',sblocklen&~0b10000000))
            
        # Else we still need more subblocks
        elif sblocklen < 0xFF:
        
          # Set the last two bits to zero
          sblocklen &= ~0b11

          # Send the length byte along with the most significant bit set again
          # meaning it's another subblock
          out.write(struct.pack('=B',sblocklen | 0b10000000))
          

      # Advance index to start of next block
      firstbyte = int(ord(rng.read(1)))
      
    # Now we're ready to begin reading the image data
    while True:

      # Get the first byte which contains information about the type and length
      # of the current block
      firstbyte = int(ord(rng.read(1)))

      # Extract the two different flags
      hasimgflag = bool((firstbyte >> 7) & 1)

      hascmapflag = bool((firstbyte >> 6) & 1)
      
      # Extract field byte
      fieldbyte = firstbyte & ~(hasimgflag << 7 | hascmapflag << 6)
      
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream as they are
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        numcolors = size//3 - 1
        if numcolors > 0:

          # We can only handle up to 256 colors so just fake that many and skip
          # over the rest
          if numcolors > 256:
            numcolors = 256

          colortable = rng.read(numcolors*3)
          out.write(colortable)

          # Skip over the rest of the colors in our palette since we only support
          # 256
          leftovercolorbytes = max((numcolors - 256)*3, 0)
          if leftovercolorbytes > 0:
              rng.seek(leftovercolorbytes, 1)
              
        # Write the actual image data
        imgdata = rng.read(size-1)
        out.write(imgdata)