#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big') & 0xfff
	# (and we don't want too big of an image)

    # frame duration is [1..255] seconds
    dur = int.from_bytes(
            rng.read(1), byteorder='big') % 256 + 1

	# (and we also don't want it to be too small)

	# NOTE : We're not really generating a valid GIF, but I'm sure there some tools that can do so :)
    out.write(header)
    out.write(int.to_bytes(datalen+7, length=3,
                           byteorder='big', signed=False))
    out.write(b'\x87\x00\x00')
    out.write(dur.to_bytes(length=1,
                          byteorder='big', signed=False))
    
    reserved = bytes([
            0xFF,    # No color table
            0xF7,    # No interlacing
            0x00     # No sort flag
        ])
    out.write(reserved)
    global_color_table_size = bytes([
            0x00,    # Size of color table is one byte long
            0x00,    # Since we use the global color table
        ])
    full_screen = bytes([
            0x00,    # Width and height are in pixels
            0x00,
            0x00,    # Global colour table should be used
        ])
    background_color_index = bytes([
            0x00,    # Set no transparent pixel
        ])
    pxaspectratio = bytes([
            0x00,    # Pixel aspect ratio numerator is zero for square pixel
        ])
    out.write(global_color_table_size)
    out.write(full_screen)
    out.write(background_color_index)
    out.write(pxaspectratio)
    lzwmincodelength = int.to_bytes(
            9, length=1, byteorder='big', signed=False)   # LZW minimum code size
    out.write(lzwmincodelength)
	
    # Write palette colours
    palettecolours = [
            (0xff, 0xff, 0xff),   # White / black
            (0x00, 0x00, 0x00),
            (0xff, 0x00, 0x00),   # Red
            (0x00, 0x00, 0xff),   # Blue
            (0x00, 0xff, 0x00),   # Green
        ]
    for pc in palettecolours:
        color = bytes([(pc[0]),
                       (pc[1]),
                       (pc[2])])
        out.write(color)

    # Write local color table
    localcolortablesize = bytes([
            0x00,    # Length of this chunk is always zero
        ])
    out.write(localcolortablesize)

    # Write images data block
    imagedescriptor = bytes([
            0x2C,    # Image separator block
            0x00,    # Left position is zero
            0x00,    # Top position is zero
            0x12,    # Width (in pixels)
            0x12,    # Height (in pixels)
            0x08,    # Local color table will follow
        ])
    out.write(imagedescriptor)

    out.write(palettecolours[4])    # Transparent color index

    imagedata = []
    for i in range(datalen):
        if i < len(palettecolours)-1:
            pcidx = int.from_bytes(
                    rng.read(1), byteorder='big') % len(palettecolours) - 1
            imagedata += [(pcidx).to_bytes(length=1,
                                           byteorder='big', signed=False)]
        else:
            imagedata += ((i-len(palettecolours)+1)%256).to_bytes(length=1,
                                                                 byteorder='big', signed=False)
    # Write compressed data with LZW compression method
    out.write(struct.pack("B", 0b11000000 | len(imagedata)))

    lzwdict = {}	# Dictionary of strings -> integers
    dictionarysize = 256	# Number of entries currently stored in our dictionary
    clearcode = 256
    eoi = 257

    currentstring = b''
    while True:
		# First, we add any new characters found in input stream to our dictionary
        nextchar = rng.read(1)
        if nextchar == b'': break
        
        try:
            lzwdict[currentstring].append(nextchar)
            continue
        except KeyError: 
            pass

        if dictionarysize < 4095:
            lzwdict[currentstring] = [[nextchar], dictionarysize]
            dictionarysize += 1
            currentstring = b''
        else:
			# If we need more space than what's allowed by our bitstream, then 
			# we just force EOI character right now since we don't have enough room
			# to encode it properly
            imagedata += [eoi.to_bytes(length=1,
                                       byteorder='big', signed=False)]
            break

		# Now, we update our string
        currentstring += nextchar
		
		# And see whether we can find a matching prefix in our dictionary

        # NOTE : In case of match we must *always* take shorter encoding 
        #        compared to the previous sequence (even when longer one would actually fit).
        #        This is why we iterate over all possible prefixes, stopping once we get a match
        #        which could potentially be encoded as a longer character.
        for i in reversed(range(dictionarysize)):
            lastchars = lzwdict[currentstring][0][:i]
            
            try:
                if b''.join(lastchars) in lzwdict:
                    imagedata += lzwdict[b''.join(lastchars)][1].to_bytes(
                            length=1, byteorder='big', signed=False
                        )

                    # Update our string accordingly
                    currentstring = lzwdict[currentstring][0][i:]
                    
                    break
            except ValueError: 
                pass
		#... or end up with EOI character
        imagedata += [eoi.to_bytes(length=1,
                                   byteorder='big', signed=False)]
		
	# Finally, we write our compressed data
    for idata in imagedata:
        out.write(idata)