#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 2 bytes of length of block (N).
#   3. N bytes of block.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(2), byteorder='little', signed=False
    )
    data = rng.read(datalen)
    out.write(header + data)


# Reads the given binary stream until the first occurance of the second argument,
# and returns that segment as a new binary string object without the second argument.
def read_until(
        fio: BinaryIO, delim: bytes
    )-> bytes:

    # Read segments until we find the delimiter character.
    seg = fio.read(1)
    segs = [seg]
    while not seg == delim:

        # Read another byte if we're at EOF.
        try:
            seg = fio.read(1)
        except IOError:
            break
        segs.append(seg)

    return b''.join(segs[:-1])


# Generates a random PNG image in chunked encoding.
# Returns the size of each chunk.
def generate_random_png(
        rng: BinaryIO, out: BinaryIO,
        maxlen = 50, maxwidth = 600, maxheight = 600, verbose = False):
    
    def writechunk(name, content):
        namelen = len(content)
        out.write(
                struct.pack('>I', namelen) \
                + name.encode() \
                + struct.pack('>I', ~namelen & 0xffffffff) \
                + content \
                + struct.pack('>I', namelen ^ 0xffffffff)
        )

    writechunk(b'IHDR', struct.pack(">IIIBBBBB",
        1 << rng.randint(1,7), 
        1 << rng.randint(1,7), 
        rng.randint(0,maxwidth-1), 
        rng.randint(0,maxheight-1), 
        8, 2, 0))

    for _ in range(rng.randint(0,2)):
        writechunk(b'PLTE', 
                b'\x00\x00\xff' * rng.randint(0, 256) \
                + b'\x00\x00\x00')

    compr = 'z'+str(rng.randint(1,9))
    writechunk(b'tEXt', 
            b"Comment=" \
            + str(rng.randint(0, 2**32)).encode())


    lzwmincodesize = rng.randint(1, 9)
    datasize = rng.randint(2, maxlen+1)
    data = bytes([rng.randint(0,255) for i in range(datasize)])
    writechunk(compr.encode(), 
            chr(lzwmincodesize).encode()\
            + data[1:] + b'\x00\x00')

    writechunk(b'IDAT', 
            b"\x00" + data[1:])

    writechunk(b'IEND', b'')

    if verbose:
       print("Generated {} chunks".format(len(chunks)))

    return sum(map(len, chunks))+8

