#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(5) + bytes([2])

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have one
    if gctlen > 0:
      gct = rng.read(gctlen)

      # If we don't get an even number of colors then we need padding to make sure
      # all resulting images contain the same amount of bytes
      outbytesleft = 256 - len(gct)
      if outbytesleft % 3!= 0:
        out.write(b'\x00'*(outbytesleft%3))
      
      out.write(gct)
      
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        blocksize = int(((fieldbyte & 7) << 8 | rng.read(1)[0]))
        print("Read BlockSize", blocksize)

        # I'm not really sure what this field is but it always seems to be zeroes
        compressedcode = rng.read(1)
        
        # Write other fields to output as-is
        out.write(fieldbyte.to_bytes(1,'big')+blocksize.to_bytes(1,'big'))

        # Check if this was a graphics control block        
        if (((fieldbyte ^ int.from_bytes(compressedcode,'big'))&64)==64):

            # Read in the following field
            numfields = int.from_bytes(rng.read(1),'big')
            
            # We only care about a very few things here so just read them all
            # out but ignore some of them
            flags = [False]*numfields
            transparentcolormask = None
            delaytime = None
            transparentcolorindex = None
            terminator = False
            disposalmethod = None

            for i in range(0,numfields):
                fieldtype = int(ord(rng.read(1)))

                # Read data depending on field type                
                if fieldtype == 0xF9:
                    flags[i]=True
                    duration = int.from_bytes(rng.read(2),'big')
                    
                    terminator=duration==0

                    delaytime=duration*10
                
                elif fieldtype == 0xFC or fieldtype == 0xFD:
                    flags[i]=True
                    transparantcolortableindex = int.from_bytes(rng.read(1),'big')

                else:
                    flags[i]=True                    
                    subdata = rng.read(fieldtype)            
                
                    
            
            # Only write full blocks to output            
            if numfields < 8 and not terminator:
              out.write(flags[0].to_bytes(1,'big'))
              out.write(flags[1].to_bytes(1,'big'))
              out.write(flags[2].to_bytes(1,'big'))
              out.write(flags[3].to_bytes(1,'big'))

              # Field 4
              if not flags[3]:
                out.write(transparentcolormask.to_bytes(1,'big'))
                out.write(delaytime.to_bytes(2,'big'))
              
              # Field 5
              if not flags[4]:
                out.write(transparentcolorindex.to_bytes(1,'big'))
              
              # Field 6
              if not flags[5]:
                out.write(disposalmethod)
              
              # Field 7
              if not flags[6]:
                out.write(terminator.to_bytes(1,'big'))
                     

        else:
            # We didn't read a graphic control block so read the next field byte
            nextfield = int(ord(rng.read(1)))
            if nextfield!= 0:
                print('New field byte',nextfield)

            # Now we actually do something with non-graphic control blocks
            # See if we're reading an image block
            if fieldbyte == 4:
                # These fields are the width, height, and local color table length
                widthandheight = rng.read(2)+rng.read(2)
                lctlen = 2**(int((rng.read(1)[0] >> 5))+1)*3

                # Write those values directly to our output stream
                out.write(widthandheight)
                out.write(rng.read(lctlen))

                # Now we can simply read the pixel data
                pxs = []
                while sum([pxs.count(c) for c in list(range(0,maxval))]):
                  px = rng.read(1)[0]

                  # In order to reduce the quantization error add mask 128 to pixels
                  # that would otherwise be rounded down (which is how they should be displayed for normal viewing)
                  if px <= maxval-128:
                    pxs += [px+128]
                  else:
                    pxs += [px]
              
                out.write(bytes(pxs).decode())
                
            else:
                raise Exception('Unknown block type')


if