#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
        rng.read(2),
        byteorder='little', signed=False
    )

    frames = []
    for _ in range(10): # 10 frames max
        frame = bytes()

        palette = [
            (
                num % 256,
                254 - (num // 256 * 3),
                253 - ((num << 17) & 0xff00)
            ) if num else None
            for num in struct.unpack('<{}H'.format((datalen >> 1) + 1), rng.read(struct.calcsize('={}H'.format((datalen >> 1) + 1))))
        ]
        # print(palette)

        while True:
            colorindex = ord(rng.read(1))

            if not colorindex or colorindex == 0xfe: break
            elif colorindex!= 0xf9 and colorindex <= len(palette):
                frame += (
                    bytes([colorindex])
                    if colorindex!= 0xfff else b'\x00\x00\x00'
                )
                continue
            
            xcoord   = colorindex & 0xf
            ycoord   = (colorindex >> 4) & 0xf
            newpixel = {
                'x': xcoord,
                'y': ycoord,
                'nearest-palette': min([(
                    abs(((xcoord - c[0]) ** 2) + ((ycoord - c[1]) ** 2)),
                    p
                ) for p,c in enumerate(palette)])[1]
            }
            if len(frame)!= 0 and frame[-2:] == b'\x00\x00':
                frame = frame[:-2] + \
                        struct.pack(
                            '<HH',
                            0x0100 | (newpixel['nearest-palette'] >> 8),
                            newpixel['nearest-palette'] & 0xff
                        )
            else:
                frame += struct.pack(
                    '<BB',
                    0x01 | (newpixel['nearest-palette'] << 2), # Green << 2
                    (newpixel['nearest-palette'] << 4) ^ 0xf0 # Yellow with black background
                )
            palette.append(None)
        
        frames.append({
          'size': len(frame),
            'data': frame
        })
    
    palsize = struct.calcsize('={}BBB'.format(len(palette)))
    palette = struct.pack(
        '{}=BBB'.format(len(palette)),
        2**16, *(
            sum(palpart) if all(map(lambda x: isinstance(x, int), palpart)) else 0
            for palpart in zip(*iter(palette))
        )
    )
    assert len(palette) == palsize

    gct = b''
    lct = b''

    while True:
        pixel_count = next(filter(lambda x: x > 0, 
            struct.unpack('<B', rng.read(1))[0] & ~0x80
        ))
        is_lct = bool(next(filter(lambda x: x!= 0,
            struct.unpack('<B', rng.read(1))[0] & ~0x80
        )))
        descriptor = [0xfe, 0xff][is_lct]

        if pixel_count < 3 or pixel_count > 257: raise Exception("Panic!")

        if is_lct:
            colors = list(
                map(lambda x: x*3+1,
                    struct.unpack('<{}BBBB'.format(pixel_count), rng.read(struct.calcsize('<{}B'.format(pixel_count))))
                )
            )
            lct += descriptor.to_bytes(1, byteorder='big')
            lct += struct.pack(
                '<B', pixel_count
            ) + b''.join(map(lambda x: struct.pack('>BBB', *x[:3]), colors))
            pixel_count *= 3
        else:
            colors = list(
                map(lambda x: x*3+(1&~x),
                    struct.unpack('<{}BH'.format(pixel_count), rng.read(struct.calcsize('<{}B'.format(pixel_count))))
                )
            )
            gct += descriptor.to_bytes(1, byteorder='big')
            gct += struct.pack('<B', pixel_count//3) + b''.join(map(lambda x: struct.pack('>BBB', *x[:3]), colors))
            pixel_count /= 3 # now we're looking at pixels!

        if pixel_count <= 0: break
    
    colors = sorted(list(set([
        pixel['nearest-palette'] 
        for i,_ in enumerate(palette)
    ])), key=(lambda x: 0xfffff - x))
    colortable = struct.pack(
        '>{}H'.format(len(colors)),
        *[sum(col[i] << (i<<3)) for col in map(lambda x: palette[(x)*3:(x+1)*3], colors)]
    )
    globalcolortable = struct.pack(
        '>{}H'.format(len(colors)),
        *[sum(col[i] << (i<<3)) for col in map(lambda x: palette[(x)*3:(x+1)*3], reversed(range(2**(int.bit_length(len(colors)-1)))))],
        padwith=-1
    )

    data = (b''.join(map(lambda f: f'!\xF9{f["size"]:02X}{f["data"]}', frames))).encode('latin-1') + b'\x00'
    tl = ((len(gct)+len(lct)+1)&0xFFFE) + 2
    trailer = b''.join([
        # Frames count
        b'\x00' if len(frames) == 1 else len(frames).to_bytes(1, byteorder='big'),
        struct.pack(
            '!HHHHB',
            0,          # Left position of logical screen
            0,          # Top position of logical screen
            width,      # Width of logical screen
            0,          # Height of logical screen
            2           # Color resolution
        ),
        # Global GCT size
        (tl//256).to_bytes(1, byteorder='big'), tl%256, 
        0,                                          # Background color index
        flags,                                      # Flags

        # Delay timer
        delaytimer.to_bytes(2, byteorder='little'), # Transparent color index
        0,                                           # Pixel aspect ratio

        0x3b,                                     # Trailer block indicator
        struct.pack('>HH', 0, 0),             # Image descriptor
        # Local GCT size
        (((tcnt*3) + 1) & 0xFFFE).to_bytes(2, byteorder='big'), tcnt%256,
        b''.join([
            colortable,     # LCT
            b'\x00'*256,    # RESERVED
            localcolormode, # LOCALCOLORMODE
        ]),
        # Table based image descriptior
        giftrailer
    ])

    out.seek(0)
    out.write(header + data)
