#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)
#     print('lsd', bytes(lsd).hex())
    _, width, height = [int.from_bytes(lsd[i:(i+2)], "little")
                       for i in range(0, len(lsd), 2)]
    
    # Add LSD to the output
    out.write(lsd)

    # Define initial position code
    position_code = {
        0:'0',
        1:'1',
        2:'2',
        3:'3',
        4:'4',
        5:'5',
        6:'6',
        7:'7',
        8:'8',
        9:'9',
        10:'A',
        11:'B',
        12:'C',
        13:'D',
        14:'E',
        15:'F'
    }

    # Define initial graphics control extension code
    graphic_control_extension_code = {}

    # Define initial application extension code
    application_extension_code = {
        0xF9 : 'NETSCAPE'
    }

    # Define initial comment extension code
    comment_extension_code = {}

    # Define initial plain text extension code
    plaintext_extension_code = {}

    # Define initial frame control block code
    frame_code = []

    # Define how many frames should be included in the gif
    num_frames = 10

    # Create a list containing all of these codes
    codes = [*position_code,
             *graphic_control_extension_code,
             *application_extension_code,
             *comment_extension_code,
             *plaintext_extension_code,
             *frame_code
            ]

    def encode_string(text: str, dict_values):

        encoded_string = b''

        index = -1
        
        for c in text:

            if not c in dict_values:
                encoded_string += b'_' + \
                    chr(ord('_')).encode() + chr(ord(c)).encode()
                
            else:
                
                index += 1
                
                while dict_values[index].find(c)!= 0:
                    
                    index += 1
                    
                encoded_string += chr(dict_values[index][0]).encode()
            
        return encoded_string
    
    # Read through the rest of the file
    data_offsets = []
    lines = ['']*num_frames
    for _ in range(num_frames):
        data_offset = rng.tell()

        # The first byte is always zero
        out.write(b'0')

        # Size of the compressed sub-block
        size = int.from_bytes(rng.read(1),'little')

        # Check if this is the last frame
        is_last = bool(size & 1<<7)
        
        size &= ~(1<<7)

        # Write the size to the stream
        out.write(size.to_bytes(1,'little'))
        
        
        # Decode the string into its original format
        # This also gets rid of any extra information that might have been added to the end of the line
        decoded_line = ''
        while True:
            
            value = ord(rng.read(1))

            if value == 0xFF or value == 0xFE:
                                
                break
            
            elif value >= 0x20 and value <= 0x7e:
                
                decoded_line += chr(value)
                
        line = decoded_line[:width].strip()
        lines[_] = line
                
        data_offsets.append(data_offset)
        
        # Encode the line back into it's original format
        if len(decoded_line) > width:
        
            values = []

            index = 0
    
            while index < len(line):

                values.append([])
            
                while index < len(line):

                    found = False
                
                    for i in range(1,min([len(line)-index,6])):

                        try:
                            values[-1].append(codes.index(line[index:index+i],
                                                           index))
                            
                            index += i

                            found = True
                            break
                            
                        except ValueError:
                            pass

                    if not found:
                        
                        values[-1].append(codes.index("_"))
                        
                        index += 1

                        values[-1] += codes[index]

                line += ''.join(['{:>3}'.format(str(v))
                                for v in values[-1]])
                
        encoded_data = encode_string(line, codes)
        out.write(encoded_data)


    # Image descriptor
    out.write(struct.pack('<HHHBB',
                         width      % 0x10000,
                         height//256,
                         0,
                         0,
                         0
                        ))

    # Graphics control extension with transparent color flag set
    out.write(b'\x21\xf9\x04\x01\x00\x00\x00\xff\xff')

    # Ending image descriptor
    out.write(b';')

    # Codes dictionary
    out.write(struct.pack('<B',len(codes)))
    out.write(b''.join(codes))

    # Data offsets
    out.write(struct.pack('<B', len(data_offsets)))

    for doffset in data_offsets:
        out.write(doffset.to_bytes(2, 'big'))
        
    

