#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    width = 10 + int.from_bytes(rng.read(4), 'big') % 5

    height = 10 + int.from_bytes(rng.read(4), 'big') % 5

    print('width: ', width)
    print('height: ', height)

    color_table_size = 2 ** (8 - ((int.from_bytes(rng.read(1), 'big')) & 7))

    print('color table size:', color_table_size)

    is_global_color_table = bool((int.from_bytes(rng.read(1), 'big')) & 1)

    print('has global color table:', is_global_color_table)

    bits_per_pixel = {
        0x87: 1,
        0x09: 2,
        0x0A: 4,
        0x0B: 8
    }[(int.from_bytes(rng.read(1), 'big'))]

    print('bits per pixel:', bits_per_pixel)

    background_index = int.from_bytes(rng.read(1), 'big') % color_table_size

    print('background index:', background_index)

    aspect_ratio = {
        0x00: 'no',
        0x04:'square',
        0x0C: '4:3',
        0x18: 'letterbox'
    }[(int.from_bytes(rng.read(1), 'big'))]

    print('aspect ratio:', aspect_ratio)
    
    # Write header
    out.write(b'GIF89a')

    # Write logical screen descriptor
    out.write(struct.pack('<HHBBBBB', 
        0xFFFF & width, 0xFFFF & height, 
        bits_per_pixel, background_index, 
        color_table_size > 0,
        0x00 if not is_global_color_table else 0xF7,
        0x00 if is_global_color_table and color_table_size < 2**8 else background_index
    ))

    if not is_global_color_table:
        
        # Write global color table
        for i in range(0, color_table_size):
            out.write(struct.pack('<BBBB',
                * [0xFF & rng.read(1) for _ in range(0, 3)]
            ) + b'\x00')

    else:

        # Ensure there are at least two colors to use as transparency
        while color_table_size <= 1:

            color_table_size += 1

            generate_random_gif(rng, out)
        
    # TODO: Use LZW compression
    # TODO: Use interlacing
    # TODO: Use local color tables
    # TODO: Use animated GIFs
    # TODO: Use transparency

    # Write image descriptor
    out.write(struct.pack('<HHHHHBB', 
        0, 0, 0xFFFF & width, 0xFFFF & height, 
        0,
        0x00 | ((is_global_color_table or color_table_size < 2**8) << 7),
        0
    ))

    data = []

    for y in range(0, height):
    
        for x in range(0, width):
        
            data.append({
                1: [(bool((i >> x) & 1)), True][:]
               , 
                2: [(i >> (2*y + x)) & 0b11, True][:]
               ,
                4: [(i >> (4*y + x)) & 0b1111, True][:]
               ,
                8: [(i >> (8*y + x)) & 0b11111111, True][:]
            }[bits_per_pixel])

    for i in range(0, len(data)):
        bg_clr = rng.read(1)
        px_clr = bg_clr if data[i][0] else random.choice([
            c for j, c in enumerate(colors) 
            if j!= background_index
        ])

        out.write(bg_clr)

        out.write(px_clr)

    write_trailer(out)

def write_trailer(out: BinaryIO):

    out.write('\x3b\x00'.encode())

