#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted image,
# so it doesn't try to encode too much data itself (e.g., it does not have
# any palette).


def dump_chunk(out: BinaryIO,
               id: bytes,
               data: bytearray):
    out.write(struct.pack('>B', len(data)))  # length field
    out.write(id)                              # chunk type ID
    out.write(bytes(data))                     # chunk payload


def generate_gif(rng: object,
                 width: int = 1920,
                 height: int = 1080,
                 framecount=64,
                 fuzzlen=0x100000,
                 out: BinaryIO = None,
                 ):

    lo_width = width & 0xFF
    hi_width = (width >> 8) & 0xFF

    lo_height = height & 0xFF
    hi_height = (height >> 8) & 0xFF

    header = bytearray([
        71,             # signature [GIF]
        58,             # version
        lo_width,       # logical screen width
        hi_width,
        lo_height,      # logical screen height
        hi_height,
        0xF7 | 0x80,    # packed fields -- global color table bit
                         # set and sort flag bits off
    ])                                  # end of header

    # The spec says minimum size for global color table is three colors, but
    # some apps will crash if there aren't at least two...
    colors = bytearray([
        0x3F, 0xFF, 0xC0,     # black
        0x3F, 0xCF, 0xA0,     # white
    ])

    dump_chunk(out, b'IHDR', header + colors)   # insert the IHDR chunk

    blocksize = rng.randint(0x100, 0xFFFFFFF)        # pick a random blocksize
    dump_chunk(out, b'acTL', struct.pack('>HHH',
                                          framecount,
                                          0,         # no repeats
                                          ))          # insert the acTL chunk

    out.write(b'\xE0\x00')                         # application extension
    out.write(b'sample\0')
    out.write(b'\xFF\x00\x00')                      # terminator

    # One or more animation frames follow; start with an empty frame
    frameheader = bytearray([0])                   # single pixel in the top-left corner

    lastbyte = 0
    nblocks = fuzzlen // blocksize                  # number of blocks to write
    for _ in range(nblocks):                        # main loop
        # build up this frame one block at a time

        lastbyte += (blocksize - sum(frameheader)) % blocksize + 1

        pixels = []                                # collect all pixel data
        while lastbyte < blocksize:                 # repeat until we've filled our block
            r = rng.randint(0x00, 0xFF)            # get next byte from the RNG
            pixels.append(r)                       # add it to the list

            lastbyte += 1                          # increment our progress counter

        codelen = min(2**int((lastbyte*8)/nblocks),
                      max(min(pixels),
                          max(pixels)))              # determine how many codes should be used

        # now convert each byte into a lzw code
        codes = {}                                 # dictionary mapping input values to output codes
        count = 0                                  # current number of codes used
        clearcode = count                           # first reserved code value
        endofdata = count+1                         # second reserved code value
        prevkey = ''                               # previous key (text string)
        prevval = 0                                # previous code (integer)

        for p in sorted(set(pixels)):               # unique-ify and sort the pixels
            k = p                                   # new key
            v = codes.get(k, None)                 # check if we already got a code for this value

            if v is None:                          # new key
                codes[k] = v = count                # map this value to its new code

                count += 1                          # advance the number of codes used

            elif v == endofdata:                    # special case: after previous run, reset code tables
                clearcode = 1 << 2
                        # first code must be <= 4095 according to the spec
                clearcode <<= 1

                endofdata = clearcode + 1           # second code must be > than the first

                        # skip over the reserved code values since they shouldn't appear again
                codes[clearcode] = clearcode
                codes[endofdata] = endofdata

                count += 2                          # advance the number of codes used

            else:                                   # duplicate key
                del codes[k]                        # delete old entry

            assert p!= clearcode and \
                   p!= endofdata,'reserved code detected!'


            prevkey = k                             # update the compressor state variables
            prevval = v

        codes[prevkey] = endofdata                  # always terminate compression with EOD marker

        # generate the compressed byte stream
        def genbitstream():
            yield clearcode                         # initial clear code
            yield endofdata                         # final end of data marker

            x = 0                                    # starting prefix value
            prefix = 0                               # prefix (lookahead) value
            offset = 0                               # shift offset
            pos = 0                                  # current read position

            while True:                              # infinite loop!
                x |= ((pixels[pos]-offset)>>1)&~prefix  # mask out unused bits
                pos += 1                              # advance to next byte

                if pos >= len(pixels):               # if we hit EOF
                    break

                t = (((pixels[pos]-offset)<<7)-prefix) & ~prefix  # higher bit and complement it
                x |= t                                  # append to existing value
                prefix >>= 1                          # move prefix down to lowest two bits

                if prefix == 0:                       # when prefix is zero, emit what we have
                    yield codes[(x&0xFF)+offset]*2     # use lookup table based on prefix

                    offset += 0x100                    # advance the code offset
                    x &= ~t                             # remove added high bit
                    prefix = (~x) + 1                  # make room for lower two bits

            yield clearcode                         # another clear code before ending

        pixelbytes = []
        for bx in genbitstream():
            pair = (bx>> 8) & 0xFF
            pixelbytes.extend((pair,))
            pair = bx & 0xFF
            pixelbytes.extend((pair,))


        frameheader.extend(pixelbytes)              # extend the frame header


    # Write out the completed frame followed by a trailer consisting of just
    # enough padding and then the GIF EOF chunk in order to satisfy the gif parser.
    dump_chunk(out, b'IDAT', frameheader)
    extra = fuzzlen - len(frameheader)
    out.write(b'\xFF\xFF' + b'\0'*extra + b'\x3B')

