
import os
import time
import shutil
from typing import BinaryIO

# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, duration: int = 10, background: str = 'white',
        width: int = 640, height: int = 352, frames: int = 10) -> None:
    """Generate a GIF with the specified parameters from an RNG."""

    # write header (logical screen descriptor) to buffer
    # logical screen:
    #   7 bit color resolution
    #   0 - global color table
    #   bits per pixel is 8 (color palette index size)
    #   0 - sort order of colors
    #   width in pixels
    #   height in pixels
    #   flags are set as follows:
    #       - has global color table flag (0x80) set
    #       - no interlace flag (0x40) not set
    #       - no sorted color table flag (0x20) not set
    #       - no local color table flag (0x10) not set
    gif = bytearray()
    gif.append('GIF89a'.encode()[0])         # signature "GIF"
    gif.append((width & 0xFF).to_bytes(1, byteorder='little')[0])      # width low endian
    gif.append(((width >> 8) & 0xFF).to_bytes(1, byteorder='little')[0])  # width high endian
    gif.append((height & 0xFF).to_bytes(1, byteorder='little')[0])     # heigth low endian
    gif.append(((height >> 8) & 0xFF).to_bytes(1, byteorder='little')[0]) # heigth high endian
    gif.append('\xF7'.encode()[0])           # flags packed into one byte
    gif.extend(b'\x00\x00')                  # background color index and pixel aspect ratio
    # write color table that will be used for all images
    for _ in range(0xFF):            # for each entry in color table
        gif.extend(rng.read(3))     # read RGB value

    # write image descriptors to file
    # frame:
    #   0x2C indicates image separator
    #   x coord low endian
    #   y coord low endian
    #   image width low endian
    #   image height low endian
    #   flags (bit packed):
    #       - local color table is present (0x80)
    #       - interlacing (0x40) not required for this image
    #       - lzw compression (0x20) is used by default
    #       0x10 reserved for future use
    #       0x08 sort order for local color table not important
    #       0x04 transparent color is given (0x00 means none)
    #       0x02 user input (not implemented yet)
    #       0x01 reserved for future use
    #   local color table size is 7 (0-128 entries)
    #   LZW min code size (log base two of actual dictionary size plus one)
    #   image data block follows (byte count variable)
    for i in range(frames):
        gif.extend([
            0x2C,                          # image separator
            ((i % width) & 0xFF),          # x coordinate low endian
            (((i % width) >> 8) & 0xFF),  # x coordinate high endian
            ((i // width) & 0xFF),         # y coordinate low endian
            (((i // width) >> 8) & 0xFF), # y coordinate high endian
            ((width & 0xFF)),              # image width low endian
            (((width >> 8) & 0xFF)),      # image width high endian
            ((height & 0xFF)),             # image height low endian
            (((height >> 8) & 0xFF)),     # image height high endian
            '\xF7'.encode()[0],            # flags packed into one byte
            0x07                           # local color table size; 2^(size+1)=7
        ])

        # compute frame duration
        nxt = i + 1 if i < frames else 0
        delaytime = round(duration * 100 / (frames if frames > 0 else 1))
        # add graphic control extension
        gif += bytes([(
            0x21,                       # extension introducer
            0xf9,                       # graphic control label
            0x04,                       # block size
            0x01 |                        # reserved
                (0 if nxt == 0 else
                0x08 if nxt >= frames else
                0x04 if nxt <= 0xFF else
                0x02) |                      # disposal method
                (0x01 << 3),                 # user input (ignored)?
            (delaytime >> 8) & 0xff,     # delay low byte
            delaytime & 0xff,            # delay high byte
            0x00                         # transparency color index
        )])
        # add image data subblock
        data = [background] * width * height
        gif.extend(data)


    # write trailer to file
    gif.extend(b'!')
    gif.append(((len(gif) & 0xFF)).to_bytes(1, byteorder='little')[0])
    gif.append(((len(gif) >> 8) & 0xFF).to_bytes(1, byteorder='little')[0])
    out.write(gif)


if