#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If we need an image data field...
        if hasimgdata:

          # Build an image block with appropriate fields set
          imageblockfields = [
            int(hasimgdata),
            0,
            # Number of loops must be at least two
            max(ord(rng.read(1)), 2),
            # No interlace
          !(ord(rng.read(1))),
            # No local colors
            False,
            # No sort flag
          !!((rng.read(1)>>7)&1)
          ]
          
          # Pack them up as binary
          restofimageblock = bytes([sum([(2**i)*b
              for i,b in enumerate(imageblockfields)])])
          out.write(restofimageblock)


        # Finally get the actual colors used
        numcolors = size//3 - 1
        if numcolors > 0:
          colortable = rng.read(numcolors*3)
          out.write(colortable)

        # Now write image data 
        imgdata = rng.read(size-1)
        out.write(imgdata)