#!/usr/bin/env python3

import struct
from typing import BinaryIO, Callable


def consume_varint(stream: BinaryIO) -> int:
    """Read a varint from the stream"""

    val = 0
    i = 0
    while True:
        byte = int.from_bytes(stream.read(1), "big")

        # If we've reached the end of our buffer then stop trying to read more data
        if not byte or (byte & 0x80 == 0):
            break

        # Mask off any bits above 7 and add them in as our value
        byte &= ~0x80
        val += byte << (7 * i)
        i += 1

    return val


# Convenience function for reading raw bytes given an expected size
def _read_exactly(stream: BinaryIO, count: int) -> bytes:
    buf = b""
    while len(buf) < count:
        new = stream.read(count - len(buf))
        assert new, f"Expected {count} bytes but got only {len(buf)}!"
        buf += new
    assert len(buf) == count, f"Expected {count} bytes but got {len(buf)}"
    return buf


def decode_kvpairs(stream: BinaryIO, func: Callable[[], None]) -> None:
    """Given a file handle that contains key-value pairs encoded using the
    `encode_kvpair` function, call the provided closure on each pair until reaching EOF."""

    while True:
        try:
            kvlength = consume_varint(stream)

            key = _read_exactly(stream, kvlength).decode("utf-8")
            value = _read_exactly(stream, consume_varint(stream)).decode("utf-8")

            func()
        except AssertionError:
            print(
                "Invalid KVPair encountered! Buffer might be truncated... continuing anyway..."
            )
            break

    pass


def encode_kvpair(key: str, value: str) -> bytes:
    """Encode a single k-v pair into a bytestring suitable for use with the kvpair
    encoder."""

    return (
        struct.pack(">H", len(key)) + key.encode("utf-8")
        + struct.pack(">H", len(value)) + value.encode("utf-8")
    )
