#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#       Ref: https://www.w3.org/Graphics/GIF/spec-gif89a.txt
#      - The first three bytes are always fixed to this value, and
#        identify the file as being in gif format.
#      - The fourth byte indicates the version of the format, it's usually
#        0x75. This script uses version 0x76 for demonstration purposes.
#  2. 3 bytes which represent the logical screen descriptor block length.
#     It should be set to 0xF7 (hex) or 0b11110111 (binary).
#  3. 1 byte representing the color resolution (bits per pixel) of the image.
#     As stated above, we use an example with only one bit per pixel.
#  4. 1 byte representing color table information. In our case there is no
#     global colour table so it's just set to 0x00.
#  5. 1 byte indicating that ther are no global colour table entries, since
#     they are not present at all.
#  6. 2 bytes that represent the dimensions of the output image. We can pick
#     any RGB(A) images size but here we choose 100 x 100 pixels.
#  7. 1 byte denoting that we will be displaying the transparent background
#     for each pixel.
#  8. 1 byte which represents how many loops are required to display the
#     animation infinitely. Since we don't want to halt execution after
#     displaying the first frame, we explicitly state that it should take
#     1 loop / infinite times.
#  9. 1 byte which indicates the amount of bits used per pixel. Here
#     we picked 24 bits per pixel to save some space.
# 10. 1 byte that signifies if a local colour table is present or not.
#     Since we already have a global colour table, this byte is set to 0x00.
# 11. 1 byte that signifies the termination of the logical screen descriptor.
#
# After the logical screen descriptor comes the actual image data. Due to
# its simplicity, we can store the whole payload of the image directly as a
# string literal inside Rust code! That would look something like this:
# ```rust
# const IMAGE_DATA : & [ u8 ] = b "\x00\x00\xff\x00\xff\xff\x00\x00\x00\xff\x00" ;
#...
# out[ 13 ] = IMAGE_DATA [.. ]; // Don't forget offset 13 because of logical
#                                screens descriptors.
# ```
def generate_random_image(out: BinaryIO):
    # Magic Number ("GIF89a").
    out.write(struct.pack('>L', 0x464947f7))

    # Logical Screen Descriptor Block Length.
    out.write(struct.pack('B', 0xf7))

    # Color Resolution.
    out.write(struct.pack('B', 0x76))

    # Color Table Information.
    out.write(struct.pack('B', 0x00))

    # Global Colour Table Entries Count.
    out.write(struct.pack('B', 0x00))

    # Image Width/Height.
    out.write(struct.pack('<LL', 0x0064, 0x0064))

    # Background Color Index.
    out.write(struct.pack('B', 0x00))

    # Pixel Aspect Ratio (Packed Format). We can safely ignore this.
    out.write(struct.pack('B', 0x00))

    # Global Colour Table. Not Applicable here as there isn't one.
    raw_data = []
    for _ in range(0x00ff + 0x00):
        raw_data += list([
            0x00,
            0x00,
            0xff,
            0x00,
            0xff,
            0xff
            ])

    out.write(bytes(raw_data[:0x00]))

