#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(5) + bytes([2])

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlength = lsd[4] << 1;
    
    # Check for an application extension block present
    appflag = int.from_bytes(rng.read(1), "little") & 1
    hasappextblock = False
    if not appflag == 0:
      hasappextblock = True

      # Jump ahead to next byte after app extension
      rng.seek(1, 1)
      
      # Skip four bytes because we don't care about these fields
      rng.seek(4, 1)

      # After this app extension block is read then later on we can get back to reading
      # normal image blocks
      
    # Just write the GCE which signals end of the header
    out.write(b'\x21\xFF\x0B')

    # Length indicator is 11 plus the length of any application extension block
    lenind = (11).to_bytes(1, 'little')
    if hasappextblock:
       lenind += (11+1+4).to_bytes(1, 'little') 
    else:
       lenind += (11).to_bytes(1, 'little') 

    out.write(lenind)

    # 76 51 62 38 39 61
    # Set background index
    bgind = rng.read(1)
    out.write(bgind)
    # Pixel aspect ratio
    paxr = rng.read(1)
    out.write(paxr)
    # GCT flag
    gctf = rng.read(1)
    out.write(gctf)
    # Sort flag
    sftf = rng.read(1)
    out.write(sftf)
    # Global Color Table Size
    gcts = rng.read(1)
    out.write(gcts)

    # If there's no application block still write one to signal end of LSD
    if not hasappextblock:
      out.write(b'\x21\xF9\x04')
      out.write(b'\x01\x00')
      out.write(b'\x00')
      out.write(b'\x00')

    # Write image block
    out.write(b'\x2C')

    # Image left position within frame as unsigned short value
    imleftpos = rng.read(2)
    out.write(imleftpos)

    # Image top position within frame as unsigned short value
    imtoppos = rng.read(2)
    out.write(imtoppos)

    # Image width as unsigned short value
    imwidth = rng.read(2)
    out.write(imwidth)

    # Image height as unsigned short value
    imheight = rng.read(2)
    out.write(imheight)

    # Frame flags
    frflags = rng.read(1)
    out.write(frflags)

    # Delay time in milliseconds as two bytes
    delaytimems = rng.read(2)
    out.write(delaytimems)

    # Transparent color index as single byte
    tci = rng.read(1)
    out.write(tci)

    # Block terminator
    out.write(b'\x00')


    # Write palette here if needed
    size = 3*(2**int.from_bytes(gcts,'little'))
    if size!= 0 and int.from_bytes(gctf,'little'):

        # Write the palette block
        out.write(b'\x2F')
        
        # The size of the palette minus one
        numcolors = size//3 - 1
        if numcolors > 0:

          # We can only handle up to 256 colors so just fake that many and skip
          # over the rest
          if numcolors > 256:
            numcolors = 256

          colortable = rng.read(numcolors*3)
          out.write(colortable)

          # Skip over the rest of the colors in our palette since we only support
          # 256
          leftovercolorbytes = max((numcolors - 256)*3, 0)
          if leftovercolorbytes > 0:
              rng.seek(leftovercolorbytes, 1)
              
        # Write the actual image data
        imgdata = rng.read(size-1)
        out.write(imgdata)