#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(rng.read(1), byteorder='big') & 0x7F
    while datalen == 0x7f:
        datalen += int.from_bytes(rng.read(1), byteorder='big')

    width = int.from_bytes(rng.read(1), byteorder='big', signed=True) + 1
    height = int.from_bytes(rng.read(1), byteorder='big', signed=True) + 1

    flags = rng.read(1)[0]
    if not (flags & 1): raise Exception('Image must have local color table.')

    bytes_per_entry = ((flags >> 3) + 1) // 8 # Number of bits per palette index

    for i in range(256):
        red   = rng.read(bytes_per_entry)[-1] * (2 ** (i % bytes_per_entry))
        green = rng.read(bytes_per_entry)[-1] * (2 ** (((i//bytes_per_entry)%bytes_per_entry)))
        blue  = rng.read(bytes_per_entry)[-1] * (2 ** (((i//(3*bytes_per_entry))%bytes_per_entry)))

        out.write(struct.pack('BBB', red, green, blue))
    
    # Application Extension Block
    app = b'\x21\xFF\x0B'
    app += b'ISNETS'

    # Write application extension block size
    out.write((len(app)).to_bytes(2, 'little'))
    out.write(app)

    # Graphics Control Extension Block
    gce = b''

    # Flags field
    gce += b'\x04'

    delay = rng.read(1)[0]*0.01 + 10 # Add between 10 and 30ms to the frame delay
    gce += int(delay).to_bytes(2, 'little', signed=False)
    gce += b'\x00'

    trans = rng.read(1)[0]
    gce += trans.to_bytes(1, 'little', signed=False)

    # Application Extension Block Terminator
    gce += b'\x00'


    # Image Descriptor
    idb = ''
    subfield = b''
   
    # Left position
    left = -width*(rng.read(1)[0]/256) # Integer division rounds downwards
    subfield += int(left+1).to_bytes(2, 'little', signed=False)
     
    # Top position
    top = -height*(rng.read(1)[0]/256) # Integer division rounds downwards
    subfield += int(top+1).to_bytes(2, 'little', signed=False)
        
    # Width
    w = max(width + 2*abs(left) - 1, 1)
    subfield += int(w).to_bytes(2, 'little', signed=False)
    
    # Height
    h = max(height + 2*abs(top) - 1, 1)
    subfield += int(h).to_bytes(2, 'little', signed=False)
    
    # LZW minimum code size
    lzw = rng.read(1)[0]&7 + 1
    subfield += lzw.to_bytes(1, 'little', signed=False)
    
    # Field terminator
    subfield += b'\x00'

    idb += len(subfield).to_bytes(1, 'little')
    idb += subfield

    # Now we write this image descriptor
    out.write(idb)

    # Finally, write compressed data itself
    lzw_data = b''
    num_bits = [
        8,      # Empty
        1, 2, 3, 4,    # For codes 0 through 15, the next n bits form the actual entry value
        0, 1, 2, 3,     # Codes 16 through 23 are disabmbiguted 0-3 repetitions of previous code
        0,             # Code 24 is disabmbiguous empty string (and end-of-stream marker)
        1, 2, 3, 4      # Codes 25 through 30 are disabmbiguous 1-5 repetitions of previous code
     ]

    def get_bits(byte: int, offset: int, amount: int)->int:
        return (byte>>offset)&((1<<amount)-1)

    while True:
        datum = rng.read(1)[0] 

        for _ in range(num_bits[get_bits(datum, 0, 3)]):
            lzw_data += chr(get_bits(datum, 0, 8)).encode()

        if get_bits(datum, 3, 3) == 7: continue

        num_bytes = {
            2: 1,
            3: 2,
            4: 3,
            5: 4
        }[get_bits(datum, 3, 3)+1]
        
        extra = int.from_bytes([rng.read(1)[0] for _ in range(num_bytes)], byteorder='big', signed=False)
        lzw_data += extra.to_bytes(num_bytes, byteorder='big', signed=False)

        if get_bits(datum, 0, 3) == 0: break
    

    out.write(lzw_data)