#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        seed: int,
        width: int, height: int, numframes: int,
        bgcolor: tuple[int, int, int],
        fgcolor: tuple[int, int, int],
        lumrange: tuple[int, int]=(0x00, 0xFF),
        crange: tuple[int, int]=(0x80, 0xBF),
        cglobal: bool=True,
        out: BinaryIO):

    # Initialize pseudorandom number generator with a known seed
    from random import Random
    rng = Random()
    rng.seed(seed)

    header = \
            b'GIF89a' +\
            (width).to_bytes(2, 'little') + \
            (height).to_bytes(2, 'little') + \
             b'\x87\x00\x00\x00\x00\xff'

    framedescs = []
    for i in range(numframes):

        # Frame descriptor
        fd = bytearray(b'\x2c\x00\x00\x00')

        if not cglobal:
            # Pick foreground color
            fgc = [rng.randint(*lumrange)] * 3
            # Set CRT effect flag and pick background palette index
            fgc.append((rng.randrange(64)+1)*0x80 | rng.randrange(63))

            # Set global table flag and pick foregroud palette index
            fgc[3] |= 0x20 | rng.randrange(63)

            fd += bytes(fgc)
        else:
            # Pick foreground color
            fd += bytes([rng.randint(*lumrange)])*3
            # Pick background palette index
            fd.append((rng.randrange(64)+1)*0x80 | rng.randrange(63))

            # Global table flag set by default


        fd.extend(
                #(rng.randint(*crange)).to_bytes(1, 'big'),
                (rng.randint(*crange) & ~0xF8).to_bytes(1, 'big'))

        for j in range(i):
            offset = len(framedescs[-j])-1
            assert(offset < 65536)
            fd[offset] |= (((offset - 7) // 3) & 0x07) << 3

        framedescs.append(fd[:2])


    pos = sum([(len(fd)+1)//2 for fd in framedescs])
    datalen = (pos+1) * 3   # 3 colors per pixel

    # Write initial image descriptor
    out.write(\
            header + \
            b'\x21\xf9\x04\x00' + \
            b'\x00\x00' + \
            b'\x00' + \
            (len(framedescs)-1).to_bytes(1, 'little', signed=False) + \
            bytes(sum(((len(fd)+1)//2 for fd in framedescs),[])))

    # Write frame descriptors
    for fd in framedescs[:-1]:
        out.write(fd + b'\x00'*((len(fd)+1)%2))

    # Write last frame descriptor
    fd = framedescs[-1]
    out.write(fd + b'\x00'*(len(fd)%2))

    # Write dummy data
    while datalen > 0:
        datalen = min(
                     datalen,
                     8192
                     ) % 1024 + 1
    data = rng.read(datalen)
    out.write(header + data)
