#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    while True:
      fieldbyte = struct.unpack('>B', rng.read(1))[0]

      if (fieldbyte&0x3f)==0x2c: # Image separator
        break
    
    gifheader = bytearray()

    # Append every field from the original header except LZW min code size
    for i in range(6):
      gifheader += bytes([fieldbyte])
      fieldbyte = ord(rng.read(1))
      
      if i!=4:
        gifheader += bytes([fieldbyte])

    assert len(gifheader)==10

    # Add fixed-length GIF header 
    out.write(bytes("GIF89a", "ascii")) 

    # Write generated header values
    out.write(gifheader)
    out.flush()

    # Determine how many loops will be in the animation
    loopcount = rng.read(1)

    # Write that value
    out.write(loopcount)
    out.flush()

    # Randomly decide whether or not to include an application extension in the file
    appextpresent = bool((ord(rng.read(1))&7)!=0)

    # Write that value
    out.write(b"\x82\xFF" if appextpresent else b"\x00") 

    # Write remaining extensions
    while appextpresent:
        exttype = rng.read(1)
        out.write(exttype)
        size = int.from_bytes(rng.read(1), 'little')
        
        # Write the whole extension at once; don't drain the source buffer because it might be needed later on
        out.write(rng.read(size+1))

        appextpresent = bool((ord(rng.read(1))&7)!=0 )
        out.write(b"\x82\xFF" if appextpresent else b"\x00") 
    
    # Start writing blocks of image data until none are left
    while True:
        fieldbyte = struct.unpack('>B', rng.read(1))[0]

        if ((fieldbyte&0x3F)==0x2C):
            pass
        elif ((fieldbyte&0x3F)==0x21):
            return
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        numcolors = size//3 - 1
        if numcolors > 0:
          colortable = rng.read(numcolors*3)
          out.write(colortable)

        # Now write image data 
        imgdata = rng.read(size-1)
        out.write(imgdata)