#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header, version = gen_header(rng)
    write_gif_data(out, *header)

    # Generate the logical screen descriptor and add to output
    lsd, bgcol, pxwid, pthei, gctflag, gctsize = gen_lsd(rng)
    write_gif_data(out, *lsd)

    # Only use global color table if flag is set (0x80)
    numcols = 2**(gctsize+1) if gctflag else 0
    gct = [gen_rgb(rng) for i in range(numcols)]
    write_gif_colors(out, gct, gctflag == GCT.TrueColorGlobal)

    # Generate graphic control extension
    ext = [b'GCE', b'\x04']

    # Flags are all zero except delay (bit 0), no disposal method (bit 3-5)
    # and user input is not required (bit 7)
    flags = 0b01 << 6 | rng.read(1)[0] & ~(0b111 << 5)
    ext += struct.pack('<BBBHB', flags, bgcol, 0, 0)

    # Write graphic control block and graphics control extensions
    write_gif_data(out, *(b''.join(ext)))

    # Get how many frames should be in this file based on repeat count
    reps = rng.read(1)[0] + 1
    frames = []
    while len(frames) < reps:

        # Generate image descriptor
        imd, left, top, width, height, localpal = gen_imd(rng, lsd[3], pxwid)
        write_gif_data(out, *imd)

        # Image data
        imgdata = bytearray(width*height*(pxwid//8))
        for i in range(len(imgdata)):
            ind = rng.read(1)[0] % localpal or 1
            imgdata[i] = ind
        write_gif_data(out, bytes(imgdata))
        
        # LZW compression of image data with minimum code size 3
        comp = compress(imgdata)
        assert len(comp) <= 2**11 - 2
        write_gif_data(out, b'\x03', comp)

        # Add current frame to list so we can track repeat count
        frames.append((left, top, width, height, imd[-1]))

    lastdesc, lastpal = None, None
    for left, top, width, height, palent in frames:
        # Generate local palette
        write_gif_colors(out, [gct[i-1] for i in palent])

        # Determine where to write the next EOF marker
        chunklen = 0
        eofpos = out.tell()
        currloc = out.tell()
        while True:

            # Generate maximal length blocks and update location
            chunklen += 255
            write_gif_data(out, *[b'\xFF']*chunklen, b';')
            
            # Decide whether we're done generating additional chunks
            pos = out.tell()
            written = pos - eofpos
            if written >= 1024:
                break
                
            # If necessary, go back and try again at a different position
            if chunklen > 255:
                out.seek(currloc)            
                write_gif_data(out, b'\xFF', b';')
                chunklen = 0
        
    # Now that the file has been generated successfully,
    # finalize the actual GIF by writing final stop byte
    write_gif_data(out, b';')
    
# Compresses a string using LZW algorithm
# Returns uncompressed data with new minimum code size
def compress(data):
    
    # Build symbol dictionary and reverse mapping from index->symbol
    symbols = {chr(i).encode('ascii'): bytes([i]) for i in range(256)}
    rmap = [' ']*256
    rmap[0] = chr(0)
    bits = 9

    # Do the first pass through data to build up initial dictionary
    prev = None
    cur = data[:2]
    for c in data[2:]:
        if cur in symbols:
            cur = symbols[cur]+bytes([c])
        elif prev in symbols:
            symbols[cur] = symbols[prev]+bytes([symbols[prev][0]])+bytes([c])
        else:
            raise Exception("Error within compression")
            
        prev = cur
        cur = cur[-1:]
            
    # Start building compressed data
    res = b''    
    ent = ''
    for d in data:
        d = bytes([d])
        if d in symbols:        
            ent += rmap[d[0]]
            if len(ent) >= 1<<bits:
                symind = int(rmap.index(ent[0]), base=2)
                res += struct.pack("<BB", bits, symind)

                bits += 1
                if bits == 12:
                    res += b'\xFF\x0F'
                    bits = 9
                
                symstr = str(symind)
                while symind >= len(rmap):
                    symstr = " "+symstr
                    symind -= 2**(bits-1)-1
                rmap.append(ent[0]+symstr)
            prev = symbols[d]
        else:            
            symind = int(rmap.index(ent[0]), base=2)
            res += struct.pack("<BB", bits, symind)

            bits += 1
            if bits == 12:
                res += b'\xFF\x0F'
                bits = 9
                    
            res += symbols[prev]
            prev = d
            ent = ''
            
                
    # Handle end case where there's trailing data        
    if ent:        
        symind = int(rmap.index(ent[0]), base=2)
        res += struct.pack("<BB", bits, symind)

        bits += 1
        if bits == 12:
            res += b'\xFF\x0F'
            bits = 9
                
        res += symbols[prev]     

    return res    
    
# Writes a single sequence of bytes to an open binary stream
def write_gif_data(*args):
    for arg in args:
        if isinstance(arg, bytes):
            out.write(arg)
            continue
        for item in arg:
            if isinstance(item, bytes):
                out.write(item)
                continue
            out.write(struct.pack('B', item))
        
# Writes GIF colors to a file buffer
def write_gif_colors(buf: BinaryIO, rgbarr: list, localcolor: bool):
    buf.write(b'\xFE')                                # Application extension
    buf.write(b'NETSCAPE2.0')                         # Extension name
    buf.write(b'\x03\x01')                            # Size of block
    buf.write(b'\x01')                                # Value of subblock
    buf.write(b'')                                    # No comment
    buf.write(b'\xFF')                                # Block terminator

    buf.write(b'\xFF')                                # Image separator
    buf.write(b'\x0B' if localcolor else b'\xF9')    # Size of block
    buf.write(b'\x00')                                # Left
    buf.write(b'\x00')                                # Top
    buf.write(b'\x00')                                # Width
    buf.write(b'\x00')                                # Height
    if localcolor:
        buf.write(b'\x00')                            # Local color map
    else:
        buf.write(b'\xFF')                            # Interlace bit
        buf.write(b'\x09')                            # Default GCT size
        buf.write(b'\x01\x01\x01')                    # Black RGB
        buf.write(b'\x00\x00\x00')                    # White RGB
        
    for rgb in rgbarr:
        buf.write(rgb)                                  # Color
    buf.write(b'\x00')                                # End of image

    
# Generates a header tuple consisting of signature and version
def gen_header(rng):
    sig = rng.read(6)
    ver = sig[-1]
    sig[-1] = ord('f')
    return (sig, ver)
    
    
# Generates a logical screen descriptor tuple consisting of its various fields
# as well as the background color and pixel aspect ratio.
# The dimensions of the logical screen depend on the given aspect ratios
# for vertical vs horizontal, though it will also choose the closest possible
# solution.
def gen_lsd(rng):
    whids = rng.read(2)
    w = (((whids[0]&7)+1)<<2) + ((whids[0] >> 3)&1)
    h = 1<<(whids[1] + 1)
   
    gctflag = bool(rng.read(1)[0] & 0x80)
    bgcol = rng.read(1)[0]
    pxwid = rng.read(1)[0]&7
    pthei = 1<<(rng.read(1)[0] + 1)
    gctsize = rng.read(1)[0]&(~7)
    
    # The GIF format specifies 213dpi, which means that the lowest