#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:5])
    print('.', end='')

    # Keep track of the byte position at the start of each frame's
    # descriptor so we know where to go back and write the
    # frame duration in the appropriate place when writing the rest
    # of the file incrementally
    frame_descriptors = [5]

    while len(frame_descriptors) > 0:

        # Add this descriptor's location to our list; essentially,
        # this will build up a stack of those that need to be updated later
        frame_descriptors.append(out.tell())

        # Read everything but the canvas width for now. We'll come back
        # at once to write it with the correct size.
        flags = ord(rng.read(1))
        transparencycolor = ord(rng.read(1))
        xpos = int.from_bytes(rng.read(2), byteorder='little')
        ypos = int.from_bytes(rng.read(2), byteorder='little')
        width = ord(rng.read(1))
        height = ord(rng.read(1))

        # The third byte contains both the local colour map flag and its
        # length. Since bitwise operations aren't available on single-
        # character strings in Python, we just treat them as integers.
        paletteflag = (flags & 0b00100000) >> 5
        colormaplen = ((flags & 0b00011111) << 8) | ord(rng.read(1))

        # In the case of a local colour map being used, copy over it too
        if paletteflag!= 0:

            palette = []
            for _ in range(colormaplen):
                r, g, b = struct.unpack('<BBB', rng.read(3))
                palette += [r, g, b]

            out.write(bytes([paletteflag]))
            out.write((colormaplen).to_bytes(4, 'little'))
            out.write(bytes(colormaplen*3))

            # Also, store the index of the first global color, since we're about
            # to overwrite the entirety of the existing one below
            startidx = 6 + (colormaplen * 3)

        else:
            # Otherwise, write out only the local colour map's length
            out.write(((flags&0b00011111)<<8)|ord(rng.read(1)))


        # Now let's do the same thing for the image data
        datalenfield = rng.read(1)
        packedfields = bool(int.from_bytes(rng.read(1),'little'))
        lctflag = int.from_bytes(rng.read(1), 'little')
        interlacedflag = int.from_bytes(rng.read(1), 'little')
        last_byte = 19 - out.tell()

        # Update the location of the new descriptor based on what we know
        # of its contents. For some reason, there are always two more than
        # the actual amount of bytes expected after the length field, or
        # else nothing ends up getting drawn
        expected_extra_bytes = 2 - ((last_byte % 8)+1)
        frame_descriptors[-1] -= expected_extra_bytes

        # If not using the LZW compression method, then we have to skip
        # over the image data itself
        if packedfields: continue

        # Read and discard the compressed length field
        lzwlen = int.from_bytes(rng.read(2), 'little')
        lzw = rng.read(lzwlen)

        # Next comes the LZW minimum code size field
        mincodesize = int.from_bytes(rng.read(1), 'little')

        # Finally, the remaining bits of the image descriptor contain the delay time.
        # A zero indicates "no disposal specified", meaning that whatever was shown
        # previously should remain visible until something new appears onscreen again.
        # We'll fix that in a moment.
        delaytime = int.from_bytes(rng.read(2), 'little')

        # Lastly, there may be extra bytes until the next byte aligned boundary. We
        # need to seek past these to get to the true beginning of the next descriptor.
        assert abs(expected_extra_bytes) <= 1,'remaining extra bytes should fit within one byte!'
        if expected_extra_bytes >= 0:
            restofdescriptor = rng.read(expected_extra_bytes)
        else:
            restofdescriptor = ''

        # All done! Let's move onto the next descriptor...

        # Writing the proper values for this frame requires knowing how many bytes were
        # consumed by each part of the previous descriptor, plus how many bytes had to
        # be padded to stay byte-aligned. If any of these numbers exceed 256, then they
        # overflow into subsequent fields, increasing their total length by another byte!
        old_fields = [
            4,                           # Data length field
            1,                           # Field terminator byte
            len(restofdescriptor),       # Padding
            1,                           # Local colour map indicator
            colormaplen if paletteflag else 0, # Palette size
            1,                           # Packed fields identifier
            1,                           # LZW flag
            lzwlen,                      # Compressed image data length
            mincodesize,                 # Minimum code size
            last_byte//8,                # Delay time value
            ]
        new_fields = [
            0,                          # Data length field (will be updated later)
            1,                          # Field terminator byte
            len(restofdescriptor)%8,     # Padding
            1 if paletteflag==0 else paletteflag, # Local colour map indicator
            0,                          # Palette size (will be updated later)
            0,                          # Packed fields identifier (#TODO)
            1,                          # LZW flag
            len(lzw),                   # Compressed image data length
            mincodesize,                # Minimum code size
            0,                          # Delay time value (will be updated later)
            ]
        overflow_count = sum([(oldval > 255) + (newval>255)
                              for oldval, newval in zip(old_fields[:-1],
                                                        new_fields[1:-1])])

        # So now we need to figure out a strategy for how to update all the values both
        # to prevent overflows and to keep the rest of the GIF file relatively sane. One idea
        # would be to calculate the maximum possible value for each field and loop through every
        # possibility in turn, updating each respective field appropriately. However, this is
        # effectively O(n^k) over all the fields, which is too slow to even complete for
        # k = 9. Instead, I'm going to try a different approach:
        #   Step 1: Find out how many completely new descriptors there will ultimately be, accounting
        #           for the fact that the original containing the global colour table might be
        #           replaced entirely by a new one without affecting the overall structure of the file.
        #   Step 2: Recursively walk down from the top level descriptor, replacing each descriptor with
        #           the largest possible values possible that still preserve the overall structure of the file.
        #           Each level of recursion corresponds to a type of descriptor, i.e. the extension block
        #           has no sub-descriptors, but a graphic control extension does.
        #   Step 3: Once I've reached the bottom level of the tree, I've found the optimal solution to
        #           the problem! At that point, I can simply use my recursive function call stack
        #           to replace all the fields with the correct amounts of data.

        ### Step 1
        # First, let's find out how many new descriptors we're potentially creating by walking
        # upwards through the tree as far as we can.
        # NOTE: It's important to note that calling `.read()` advances the current read
        # pointer of the stream by however many bytes you asked it to advance by. To avoid
        # accidentally reading ahead, store the starting read pointer and reset it afterwards.
        out_start = out.tell()
        num_new_descriptors = 1
        while True:
            try:
                datalenfield = int.from_bytes(
                        rng.peek(1), byteorder='little')
                if datalenfield < 128:
                    break
                datalenfield &= 0b00111111

                out.seek(out_start + datalenfield)

                if len(rng.peek(1)) > 0:
                    raise StopIteration()
                else:
                    num_new_descriptors += 1

            except StopIteration:
                out.seek(out_start)
                break

        ### Step 2
        def update_fields(idx):
            nonlocal new_fields, old_fields, overflow_count
            nonlocal frame_descriptors, f, curr_descriptor

            # Based on the type of block contained in this descriptor,
            # recurse either once to handle its own nested frames, or twice
            # to handle its parent.
            curr_descriptor = frame_descriptors[idx]
            out.seek(curr_descriptor)
            old_field = old_fields[idx]
            new_field = new_fields[idx]
            f = lambda x : x
            if idx == 0:        # Header
                pass

            elif idx == 1:      # Graphic Control Extension Block
                pass

            elif idx == 2:      # Image Descriptor
                paletteflag =    print(f'The generated gif is {last_byte} bytes long.')

    # For now we'll hardcode the number of loops to use, since we don't
    # yet know how much space remains between the final descriptor and EOF
    numloops = 0xC0

    paddinglen = (numloops * 3).to_bytes(size=2, byteorder='little')

    # Write the leftover bytes before the last descriptor
    print('Writing remaining bytes', end='', flush=True)
    out.seek(-1, 2)
    out.write(paddinglen)
    print('.', end='', flush=True)

    # Write the last descriptor!
    out.write(b'\x3B')

    # And pad up until the end of the file, which is easy enough because
    # we've already written exactly datalen bytes before ending.
    print('Padding global color table', end='', flush=True)
    while True:
        if out.tell() == datalen+1: break
        out.write(rng.read(1))
    print('. Done!', flush=True)

