#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' + bytes([
        72, 84, 50, # width/height
        (1 << 6),    # has color table
        0,           # background index
        0            # pixel aspect ratio
    ])

    numcolors = min(max(int.from_bytes(rng.read(1), 'big') >> 1 & 7, 2), 63)
    
    palette = [None] * numcolors
    for i in range(numcolors):
        palentry = bytearray()

        red = int.from_bytes(rng.read(1), 'big') | 1
        green = int.from_bytes(rng.read(1), 'big') | 1
        blue = int.from_bytes(rng.read(1), 'big') | 1

        palred = max(min(((red - 1) // 33) << 5, 255), 0)
        palgreen = max(min(((green - 1) // 33) << 5, 255), 0)
        palblue = max(min(((blue - 1) // 33) << 5, 255), 0)

        palentry += struct.pack('<B', 255 - palred)
        palentry += struct.pack('<B', 255 - palgreen)
        palentry += struct.pack('<B', 255 - palblue)
        
        palette[i] = bytes(palentry[:])
        
    data = []
    while True:
        control = rng.read(1)[0] & 0xCF

        if not control or len(data) >= ((1 << 12) - 1):
            break
        
        if control == 0xF9 or control == 0xFE:
            continue
            
        dlen = rng.read(1)[0] + 1
        if dlen > len(palette):
            dlen = len(palette)
        data.append((control, dlen))
        out.write(struct.pack('>BB', control, dlen-1))
        out.write(b'\x00\xFF\x00')
        mask = None
        if control == 0xE0:
            mask = rng.read(dlen - 1)
        elif control == 0xFC:
            count = int.from_bytes(rng.read(3), 'little')
            mask = [int.from_bytes(rng.read(1), 'little')
                    for _ in range(count)]
        else:
            mask = list(reversed([(
                rng.read(1)[0] ^ (~index & 0xFF)
            ) for (index,) in data]))
        out.write(bytes(mask).ljust(data[-1][1], b'\x00'))
            
    out.seek(-1, 1)
    out.write(bytes([len(palette)]))
    out.write(b''.join(palette))
    for control, length in data:
        out.write(b'\x00' * length)

    out.write(b';*')

if