#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
        rng.read(2),
        'little',
        signed=True
    ) % 1024 + 1

    # Write header to output file and then encoded data
    out.write(header)
    encode_graphics_control_block(rng, out)
    encode_image_descriptor(rng, out)
    encode_color_table(rng, out)
    encode_pixel_data(datalen, rng, out)


# Helper function that encodes Image Descriptor Block (see below)
def encode_image_descriptor(rng, out):
    descriptor = bytearray([
        0x2c,
        *rng.read(7),
    ])
    color_table_size = int.from_bytes(
        rng.read(1), 
        'little'
    ) & 0b11000000 >> 6
    local_color_table_flag = int.from_bytes(
        rng.read(1), 
        'little'
    ) & 0b00100000 > 0
    interlace_flag = int.from_bytes(
        rng.read(1), 
        'little'
    ) & 0b00010000 > 0
    sort_flag = int.from_bytes(
        rng.read(1), 
        'little'
    ) & 0b00001000 > 0
    reserved = int.from_bytes(
        rng.read(2), 
        'big'
    )
    lcx, lcy = tuple(int.from_bytes(rng.read(2), 'little') for _ in range(2))
    global_color_table_flag = bool(local_color_table_flag ^
                                 int.from_bytes(
                                     rng.read(1),
                                     'little'
                                 ))
    graphics_control_flag = bool(global_color_table_flag ^
                                int.from_bytes(
                                    rng.read(1), 
                                    'little' 
                                ))
    
    if not global_color_table_flag:
        raise RuntimeError('Local color table is currently unsupported!')

    descriptor[9] |= local_color_table_flag << 5 | \
                     interlace_flag << 4 | \
                     sort_flag << 3 | \
                     reserved 

    descriptor[10], descriptor[11] = lcx % 2**12, lcy % 2**12
    descriptor[-1] = 0
    out.write(descriptor)



# Helper function that encodes Graphics Control Block (see below)
def encode_graphics_control_block(rng, out):
    block = bytearray([
        0x21,
        0xf9, 
        0x04,
        *(rng.read(3)), 
    ])
    transparancy_flag = int.from_bytes(
        rng.read(1), 
        'little'
    ) & 0b10000000 > 0

    user_input_flag = int.from_bytes(
        rng.read(1), 
        'little'
    ) & 0b01000000 > 0

    delay_time = [
        int.from_bytes(rng.read(1), 'little'),
        int.from_bytes(rng.read(1), 'little'),
    ]
    
    transparent_index = None
    if transparancy_flag == True:
        transparent_index = int.from_bytes(
            rng.read(1), 
            'little'
        )
        block += bytes([(transparent_index)])

    block[4] |= transparancy_flag << 2 | user_input_flag 
    
    # Reserved field set to zero by default
    block[5] &= ~0b11100000 
    block[5] |= (delay_time[0] // 10) & 0b00011100
    block[6] |= ((delay_time[0] % 10) << 4) | ((delay_time[1] & 0b00001111) << 2)
    if len(delay_time) >= 3:
        block[7] |= (delay_time[1] >> 6) & 0b00000011
        block[8] |= delay_time[2] & 0b00001111

    out.write(block)


def encode_color_table(rng, out):
    """
    Writes the Color Table to the given output stream

    The length of the Color Table must be between 0 and
    2^(number of bits per pixel). It's recommended to have at least 
    2^n colors where n depends on the value of the 
    bit depth used in the image descriptor. This implementation only supports 
    maximum of 256 colors
    """
    num_colors = int.from_bytes(
        rng.read(3), 
        'little',
        signed=False
    ) % 2**7 + 1
    
    color_table = []
    for i in range(num_colors):
        r = int.from_bytes(rng.read(1), 'little')
        g = int.from_bytes(rng.read(1), 'little')
        b = int.from_bytes(rng.read(1), 'little')

        color_table.append((*struct.pack('<BBB', r, g, b)))

    out.write(b"".join(color_table))


def encode_pixel_data(datalen, rng, out):
    """
    Encodes image data based on its palette size.

    Currently this function assumes there exists a single
    local color table with 2^n entries. Where n is calculated as follows:

    1) If the local color table flag in the image descriptor
       is one, the number of entries equals 2^(bit depth - 1);
    2) Otherwise, it should contain exactly 2^n entries, where 
       n is the bit-depth in the image descriptor.

    Data format after encoding:

    ```
    BYTE LZW min code size
    {
      subblock header
      {
        subblocks
      }*
    }
    ```

    Subblock Header structure:

    ```
    BYTE 1110 LZW compression method
           |   DC length             |  AC length
           --------------------------|-----------
    BIT 7-6 |   2                   |   2
    BIT 5   |  Subblock has EOI marker|  Subblock type
    ```

    The subblock types are defined as follows:

    11 - uncompressed
    00 - compressed with fixed Huffman codes
    01 - compressed with dynamic Huffman codes without end-of-information code
    10 - reserved for future use
    """

    # TODO support multiple images
    data = bytearray()
    while len(data) < datalen:
        # Decode LZW minimum code size from current byte
        lzwmincodesize = int.from_bytes(
            rng.read(1), "little"
        ) 

        # Generate Huffmans tree
        huffmantree = {}
        
        # NOTE: As defined in section 10 of the spec,
        #       we currently assume the minimal code size
        #       11, meaning maximal 12 bits are being used
        huffmantree[lzwmincodesize - 1] = genhuffmantree(
            12, lzwmincodesize - 1
        )
        
        color_table_size = 2 ** (lzwmincodesize - 1)
        mask = 2 ** color_table_size - 1

        # Decode compressed chunk from input
        compressedchunk = bytearray([])
        while True:
            nextbyte = rng.read(1)[0]

            # Check if we've reached end of information marker
            if nextbyte & 0x80 > 0:
                break
            
            # Append decoded byte to chunk
            compressedchunk.extend(decodebyte(nextbyte, huffmantree))


        # Perform LZW decompression algorithm on chunk
        # NOTE: We can't just apply the algorithm directly here because
        #       some values used by the algorithm need to be compared against
        #       other chunks, so we don't want them to be affected by the
        #       currently processed chunk.
        data.extend(decompresschunk(
                            list(range(color_table_size)), 
                            compressedchunk
                        ))


    out.write(data[:datalen])




# Decodes an encoded byte according to the supplied Huffman tree
def decodebyte(byte: int, tree: dict) -> list:
    # Extract the first two bits from the byte
    decodedlist = []
    lastbyte = False
    while not lastbyte:
        # Ignore first three bits (for now)
        byte <<= 3
        # Get the next two bits
        tmpbyte = byte >> 8
        # Mask those two bits off
        byte &= ~(0b1111 << 8)

        # Handle special case where last byte was read
        if len(decodedlist)!= 0:
            # Inverse binary representation of index of previous byte in tree
            invprevbyteidx = bin(tree[lastbyte][tmpbyte])[::-1].zfill(8)
            # Add the inverse binary representation to decoded list
            decodedlist += invprevbyteidx[1:]
        else:
            decodedlist.append(str(tmpbyte))

        try:
            lastbyte = tree[lastbyte][tmpbyte]
        except KeyError:
            return decodedlist

    return decodedlist 


# Generate a Huffman tree whose leaves 
#