#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 4 bytes for the size of the entire header (7 bytes) plus
#       the size of the logical screen descriptor (7 bytes).
#       0x2C = 44, which is the length of this section in bytes.
#   3. 4 bytes set to the value 1 since only one image exists. The
#       minimum possible value was found by experimenting with different
#       values and determining that if there are more than one block it must be > 1.
#   4. 4 bytes set equal to the width of the image from 50-100 pixels wide.
#   5. 4 bytes set equal to the height of the image from 50-100 pixels high.
#   6. 1 byte set as 0xF6 indicating that a color table should be used.
#   7. 1 byte set as 0xA7 indicating a global color table should exist.
#       0x3B = 59, which is the length of the palette data stored in bytes.
#   8. 24 * 4 bytes representing all colors within the global palette.
#       Each color is represented as an RGB triple where each component has been
#       compressed down to 8 bits per channel. This results in 24 total bits.
#      - Colors were chosen randomly between black and white, where
#         R, G, B totals around 156; however any individual component can
#         exceed this value, so it's not truly random.
def gen_random(out: BinaryIO):
    out.write("GIF89a".encode('ascii'))

    # Calculate size of gifs header & color map according to spec
    # https://www.w3resource.com/graphics/gif/format/gif-header-format-specification.php
    magic_number = b'GIF89a\x00'
    sizeOfHeader = len(magic_number) + 4 + 4 + 4 + 4 + 1 + 1 + 24*4 

    # Descriptor of the logical screen contains width, height,
    # color resolution, sort order, flags, and size of the global color table
    logicalScreenDescriptor = b'\x2c\x00\x00\x00\x00\x05\x00\x3b'

    # Write sections to output
    out.write(struct.pack("<I", sizeOfHeader))
    out.write(logicalScreenDescriptor)

    numImagesInGif = 1 # Must be greater than zero?
    out.write(struct.pack("<I", numImagesInGif))

    # Width and height values - anything less than or equal to 0 will cause
    # issues with browsers displaying stuff incorrectly. 
    width,height = randrange(50, 101),randrange(50,101)
    out.write(struct.pack("<HH",width, height))
    
    # Flag indicates whether or not a local color table follows and how large it is
    flagPrefixByte = 0xf6 | 0xa7 << 5
    out.write(bytes([flagPrefixByte]))
    
    # Size field of the color table - if there is no global color table then this value
    # will be 0 until we implement support to add our own custom ones later on.
    out.write(b"\x3b")
    
    # Write out random rgb triples that sum up to 24 bits each. This means that each
    # part of the color can have a max value of 255.
    for _ in range(255):
        r = randrange(0, 256)
        g = randrange(0, 256)
        b = randrange(0, 256)

        rPacked = r
        gPacked = g << 8
        bPacked = b << 16
        
        packedRGB = rPacked | gPacked | bPacked

        out.write(struct.pack(">I",packedRGB))

