#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
        rng.read(2),
        'little',
        signed=True,
    )

    # Write the GIF magic and data length (0x01) to the file
    print("Writing 6 bytes of GIF header", end='')
    out.write(header + datalen.to_bytes(2, byteorder='big'))
    print(", done")


    # The next two bytes are probably the width & height but we'll ignore them for now
    # See https://en.wikipedia.org/wiki/Image_file_formats#Graphics_Interchange_Format_(GIF)
    print("Skipping first 4 bytes in Gif Header (width & height)", end='')
    out.write((0).to_bytes(2, byteorder='little') * 4)
    print(", done")

    # Compression flag is set to True
    compression = [True]   # Type: List[bool]
    print("Adding image descriptor byte with compression flag set to true.", end='')
    out.write(compression[0].to_bytes(1, byteorder='big'))
    print(", done")

    # We don't need many colors so we'll just use default palette values
    pal = [
        0xFFFFFF,    # White
        0xFF00FF,    # Magenta
        0x00FFFF,    # Cyan
        0x007FFF,    # Light blue
        0x00FF7F,    # Teal
        0x00FF00,    # Green
        0xFFFF00,    # Yellow
        0xFF7F00,    # Orange
        0xFF0000,    # Red
        0x7F0000     # Dark red
    ]

    # Print color table entries
    print("Generating or overwriting Color Table:", end='')
    for i in range(len(pal)):
        out.write(
            ((pal[i]) >> 16 & 0xff).to_bytes(1, byteorder='big') + 
            ((pal[i]) >> 8 & 0xff).to_bytes(1, byteorder='big') + 
            ((pal[i]) & 0xff).to_bytes(1, byteorder='big'),
        )
    
    print(", done")

    lzwmincodesize = 2;
    print("Overwriting LZW Min Code Size:", end='')
    out.write(lzwmincodesize.to_bytes(1, byteorder='big'))
    print(", done")

    # Start building the actual image data
    imagedata = []             # Type: List[int]
    lzwhistreetable = [[]]     # type: List[List[Tuple]]
    lastpix = -1               # Last pixel was...?
    numrepetitions = 0         # Number of identical pixels since last change
    longestchainindex = 0      # Index of longest chain in lzwhistreetable
    pixbuf = [-1,-1,-1,-1]     # Buffer of pixel data
    previouslength = 1         # Length of previous encoded line segment
    bufindex = 0               # Current index of buffer
    outputbyte = None          # Byte currently being written
    nextoutputbyteiszero = False
    isliteral = True           # Is current sequence literal (not reference)?
 
    # Image will be a solid 5-pixel block of whites
    while len(imagedata) < 5*datalen:

        if not isliteral:

            # If it's not a literal read the length from the lzw dictionary
            # TODO: Make sure this works for longer strings than one character!
            
            if longestchainindex >= len(lzwhistreetable):
                raise Exception("Error reading data!")

            # This is our new output byte
            outputbyte = pixbuf[longestchainindex][0]

            # Check if it should be followed by an extra zero bit
            if longestchainindex == lzwhistreetable[longestchainindex]:
                nextoutputbyteiszero = True
                
            else:

                # Update index of longest chain in lzwhistreetable
                longestchainindex += 1

                # Find the corresponding tuple entry in the lzw dictionary
                tup = list(filter(lambda x : x[1] == longestchainindex, lzwhistreetable[-1]))
                
                # Tuple must have exactly one element 
                if len(tup)!= 1:
                    raise Exception("Tuple lookup failed.")
                
                tup = tup[0];
                
                # Extract the length from the tuple
                dat = tup[0]

                # Don't write zeros within non-zero sequences, instead add to the number of repetitions
                if outputbyte == 0 and dat > 0:

                    numrepetitions += dat
                    
                    # Set outputbyte to the amount of repetitions following
                    outputbyte = numrepetitions
                    
                    # We're no longer writing literals after all :)
                    isliteral = False
                    
                elif outputbyte == dat:
                
                    while outputbyte > 0xff:
                        bufindex -= 1
                        imagedata.append(((outputbyte>>8)&0xff))
                        outputbyte &= 0xff

                    imagedata.append(outputbyte)

                    # Reset all state variables for the next loop iteration
                    numrepetitions = 0
                    longestchainindex = 0
                    longestchainindex = 0
                    lastpix = -1
                    previouslength = 1
                    longeschainindex = 0
                    pixbuf = [-1,-1,-1,-1]
                    bufindex = 0
                    outputbyte = None
                    nextoutputbyteiszero = False
                    isliteral = True
                
        elif lastpix!= -1:

            # Add last literal to lzwhistreetable
            if not lzwhistreetable[-1]:
                lzwhistreetable[-1] = [(lastpix)]
                
            elif tuple([lastpix]) not in lzwhistreetable[-1]:
                lzwhistreetable[-1].insert(-1, (lastpix))
            
            # If there were more than 8 pixels with same value then the repeat run has ended
            # which means that the previous length needs to be stored in the dictionary.
            if numrepetitions > 0:
                
                # Create new tuple entry in lzwhistreetable
                lzwhistreetable[-1] += [(numrepetitions, previouslength)]

                # Store the length of the repeated sequence at position in lzwhistreetable
                previouslength = numrepetitions + 1

                # Zero out the number of repetitions
                numrepetitions = 0

            # Get some fresh bits
            bits = rng.read(1)[0] << 8 | rng.read(1)[0]

            # Now get the length of the sequence from those bits
            dat = extractbits(bits, 0, 3)
            # Check if we got a long run of white pixels (first few cases)
            if dat <= 4:
                previouslength = dat + 1
            
            # Else check if we got a long run of black pixels
            elif dat == 12:
                previouslength = 13

            # Else just grab whatever nonsense length it is
            else:
                previouslength = dat + 13
            
            # Grab another bit
            bits >>= 4

            # And finally, check if there's a clear code (0xE0) after this sequence
            if bits & 0xe0 == 0xe0:
            
                # reset all state variables for the next loop iteration
                numrepetitions = 0
                longestchainindex = 0
                longestchainindex = 0
                lastpix = -1
                previouslength = 1
                longestchainindex = 0
                pixbuf = [-1,-1,-1,-1]
                bufindex = 0
                outputbyte = None
                nextoutputbyteiszero = False
                continue

            # Set lastpix again before starting the loop again
            lastpix = pixbuf[0]

        else:

            # Read first pixel in sequence
            pixbuf[0] = rng.read(1)[0]

            # Figure out length of the new literal sequence (aka how many pixels are the same)
            bits = rng.read(1)[0] << 8 | rng.read(1)[0]
            dat = extractbits(bits, 0, 3)
            # Check if we got a long run of white pixels (first few cases)
            if dat <= 4:
                previouslength = dat + 1
            
            # Else check if we got a long run of black pixels
            elif dat == 12:
                previouslength = 13

            # Else just grab whatever nonsense length it is
            else:
                previouslength = dat + 13
        
            # Grab another bit
            bits >>= 4

            # And finally, check if there's a clear code (0xE0) after this sequence
            if bits & 0xe0 == 0xe0:
            
                # reset all state variables for the next loop iteration
                numrepetitions = 0
                longestchainindex = 0
                longestchainindex = 0
                lastpix = -1
                previouslength = 1
                longestchainindex = 0
                pixbuf = [-1,-1,-1,-1]
                bufindex = 0
                outputbyte = None
                nextoutputbyteiszero = False
                continue

            # Set lastpix
            lastpix = pixbuf[0]
            
        # If the sequence isn't going to be too short, make it longer
        if previouslength <= 12:
            previouslength += 1
        
        # Loop through the whole sequence
        for j in range(previouslength):

            # Fill up buffer until full
            if bufindex >= 4:

                # Add buffer contents to the image
                for k in range(4):