#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
        ) -> None:

    # Header (always 6 bytes long, magic + version)
    out.write("GIF89a".encode())

    # Logical screen descriptor (always 7 bytes long;
    # width, height of the image are specified later on)
    out.write((0).to_bytes(2, "little"))               # Screen Width
    out.write((0).to_bytes(2, "little"))               # Screen Height
    out.write((0b00001110).to_bytes(1, "little"))     # Packed field
    out.write(b'\x00')                                 # Background Color Index
    out.write(b'\x00')                                 # Pixel Aspect Ratio

    # Graphics Control Extension Block
    # Always starts with an extension header (always 3 bytes long);
    # then comes the packed field of length 4 bytes and
    # finally ends with the application identifier, version and flags
    out.write(b"\x21\xf9\x04")                         # Extension Header
    out.write((0b00000000 |                        # Packed Field
               int.from_bytes(
                   rng.read(1), byteorder="big")).to_bytes(1, "big"))

    # Application Extension block
    out.write(b"\x21\xff\x0bNETSCAPE2.0")             # Extension Header
    out.write(b"\x03\x01")                             # Sub-block length
    out.write(
            b"WAPU\0".replace(
                b"W",
                b"\xaa").replace(
                b"A",
                b"\xbb").replace(
                b"P",
                b"\xcc").replace(
                b"U",
                b"\xdd"))                            # Application Identifier
    out.write(b"\x00")                                 # Application Auth Code
    out.write(b"\x00")                                 # Sub-block End

    # Image Descriptor / Frame Descriptor
    # (always start with the left offset of x = y = 0)
    out.write((0).to_bytes(2, "little"))               # Left Position
    out.write((0).to_bytes(2, "little"))               # Top Position
    out.write((0).to_bytes(2, "little"))               # Width
    out.write((0).to_bytes(2, "little"))               # Height
    out.write(((0b00000000 |                       # Flags
               0b10000000) &                      # Local Color Table Flag
              (~0b00000100)).to_bytes(1, "little"))   # LZW Minimum Size
    out.write(b'\x00')                                 # Background Color Index
    out.write(b'\x00')                                 # Pixel Aspect Ratio

    # Image Data - we'll fill this up later on
    datalen_bytes = [ ]
    while True:
        curbyte = rng.read(1)

        if len(curbyte)!= 1 or ord(curbyte[0]) == 0x00:

            break

        else:
            datalen_bytes += [ curbyte ]

    datalen = 0
    for byte in datalen_bytes:
        datalen = (datalen << 8) ^ byte[0]
    for i in range(datalen):
        out.write(
                ((int.from_bytes(
                    rng.read(1), byteorder="big") %
                 256)).to_bytes(1, "big"))

