#!/usr/bin/env python3

import os
import sys
import struct
import numpy as np
from typing import BinaryIO


# Generates a random GIF file into `out`. The outputted GIF is made up
# of the following components:
#   - 3 color blocks repeated three times, each with a different palette index.
#       This makes sure that we can verify that the decoder handles partial or
#       duplicate frames correctly.
#   - A single image descriptor block containing a local colour table where
#       green and red are equal to one another but blue is randomly set for
#       each pixel in the image.
#   - An uncompressed bitmap containing two colours, black and white, each
#       occuring once per row, and alternating between them.
def generate_random_gif(rng: BinaryIO, out: BinaryIO):

    # Generate the global colour table.
    gctsize = 7
    gctdata = [0] * ((gctsize + 1) ** 3)
    rgb = [(x << 5) | x for x in range(gctsize)]
    for i in range(len(rgb)):
        gctdata[i * 3 + 1] = rgb[(i >> 6)][0]
        gctdata[i * 3 + 2] = rgb[(i >> 6)][1]
        gctdata[i * 3 + 3] = rgb[(i >> 6)][2]

    # Write the GIF file header.
    outfile.write('GIF89a'.encode())

    # Write the GCT block for all frame images.
    write_block(outfile, '!H', len(gctdata))
    write_block(outfile, 'B'*len(gctdata), gctdata)

    # Create an LZW-coded image.
    height = width = 5
    indices = np.empty((height, width), dtype=np.uint8)
    indices[:] = list(range(2))
    lzw_table = {tuple(indices.flatten().tolist()): 1}
    outfile.write('\x2c\x00\x00\x00'
                  '\x00\x00\x00,\x00\x00\x00'.encode("utf-8"))
    write_block(outfile, '!HHHH',
                0b0000000000000000,
                width, height)
    write_block(outfile, '!B', 0b00000000)
    write_lzw_code(outfile, indices, lzw_table)

    # Repeat the above process for every frame of animation.
    nframes = 3
    for _ in range(nframes):

        # Write the GCT block for this specific frame.
        write_block(outfile, '!H', len(gctdata))
        write_block(outfile, 'B'*len(gctdata), gctdata)

        # Write the frame.
        outfile.write('\x2c\x00\x00\x00'
                      '\x00\x00\x00,\x00\x00\x00'.encode("utf-8"))
        write_block(outfile, '!HHHH',
                    0b0000000000000000,
                    width, height)
        write_block(outfile, '!B', 0b00000000)
        write_lzw_code(outfile, indices, lzw_table)