#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.getrandom default).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate some random data to make up the GIF image
    gifdata = bytearray()

    # Add a GIF magic string
    gifdata += b"GIF87a"

    # Add a screen descriptor
    width = ord(os.urandom(1)) % 640 + 1
    height = ord(os.urandom(1)) % 480 + 1
    flags = 0x88 | ((ord(os.urandom(1)) & 2) << 5) \
            | (((ord(os.urandom(1)) & 2) ^ 3) << 7)
    bgcol = bytes([ord(os.urandom(1)) for i in range(3)])
    aspect = 0x20 # "SQUARE"
    gifdata += struct.pack('<HHBBB', width, height, flags, bgcol[0],
                            bgcol[1], bgcol[2])
    gifdata += struct.pack('BB', aspect, 0)


    # Construct a palette consisting of 256 RGB triplets chosen randomly from
    # the set {0,...,255}^3 then write it as a series of packed BGR bytes
    paldata = [struct.pack("BBB", *os.urandom(3)) for _ in range(256)]
    gifdata += b"".join(paldata)

    # Add one or more graphics control blocks; see also RFC 2389
    nframes = ord(os.urandom(1)) % 4 + 1
    delay = ord(os.urandom(1)) % 10 + 1
    transparency = bool(ord(os.urandom(1)))
    if transparency:
        transindex = ord(os.urandom(1)) >> 4
        flags |= 0x01
    else:
        transindex = None
    globalcol = bytes([ord(os.urandom(1)) for i in range(3)])
    for frame in range(nframes):
        gifdata += (b'\x21\xF9\x04') + struct.pack("<BHB", flags,
                                                delay*10, globalcol[0]<<1,
                                                globalcol[1]<<1, globalcol[2]<<1)
                                               + (transindex!= None)*struct.pack("B", transindex)

        # If there is a local colour table, add that too
        if not transparency and (flags&0x80):
            gifdata += b"\x2C\x00\x00" + b"\x00".join((os.urandom(1)
                                                       for i in range(2**((flags&0x7)+1))))
            flags &= ~0x80

        # Write a plain LZW-compressed chunk with no subblock continuation
        lzwminsize = 2**(1+(flags>>3)&7)
        lzwmaxsize = min(lzwminsize+4095,len(data))
        lzwsizes = tuple((i,j) for j in range(lzwminsize,lzwmaxsize+1)
                        for i in range(lzwminsize,j) if len(range(i,j)) <= 12)
        lzwsizes = sorted((abs(len(data)-(j*(i//j))),i,j) for i,j in lzwsizes)[:3]
        gifdata += b"\x21\x00\x3B" + struct.pack("<BBH", *(lzwsizes[0]))
        gifdata += b"\x00\x01" + bytes(bytearray(range(lzwsizes[0][1])))

        # Then a single graphic block containing all remaining data
        gifdata += b"\x21\xFF\x0B"
        gifdata += bytes(bytearray(range(lzwsizes[-1][1],len(data))))

    # Finally output our entire payload
    out.write(gifdata)