#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(rng.read(1))
        print("Block:", fieldbyte, hex(size), "Has Image Data?", bin(bool(((fieldbyte>>5)&3)))) 

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # Calculate how many colors are needed
        numcolors = int(size)//3 - 1 if hasimgdata else None


        # If there's no image data we're done here
        if not hasimgdata:

          # If the last frame had an extension block or was completely empty,
          # then we can clear the palette 
          if fieldbyte!= 0x2C or size < 10:
            numcolors = 0
           noop = bytes([0])
           out.write(noop*(gctlen+numcolors))
           
          # Otherwise we just keep the existing palette
          else:
            # How many colors do we need?
            palette_size = (int(lsd[6]) + 1)*3

            # Allocate space for that many palette entries and dump them
            palette = [0]*palette_size
            for i in range(palette_size):
              palette[i] = rng.read(1)[0] 
            out.write(palette)
            
          # Done with block
          continue
        

        # Write the local color table if necessary
        if numcolors > 0:
          colortable = rng.read(numcolors*3)
          out.write(colortable)

        # Now write image data 
        imgdata = rng.read(size-1)
        out.write(imgdata)