#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)
    
    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7) + bytes([2])

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = int(ord(lsd[4])*pow(2, ord(lsd[5]))) * 3

    # Add GCT to output if we have one
    if gctlen > 0:
      gct = rng.read(gctlen)

      # If we don't get an even number of colors then we need padding to make sure
      # all resulting images contain the same amount of bytes
      outbytesleft = pow(2, ord(lsd[5])) * 3 - len(gct)
      if outbytesleft % 3!= 0:
        out.write(b'\x00'*(outbytesleft%3))
      
      out.write(gct)
      
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream as they are
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        numcolors = size//3 - 1
        if numcolors > 0:

          # We can only handle up to 256 colors so just fake that many and skip
          # over the rest
          if numcolors > 256:
            numcolors = 256
          
          colortable = []
          for i in range(numcolors):

            # Color bytes are represented with a bit depth of 8 bits per channel 
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            
            # Make sure the channels aren't too dark or too bright otherwise it'll be unreadable
            if r < 32 or r > 223 or g < 32 or g > 223 or b < 32 or b > 223:
              r = min(max(r, 32), 223)
              g = min(max(g, 32), 223)
              b = min(max(b, 32), 223)

            colortable.append(r)
            colortable.append(g)
            colortable.append(b)
            
          # Pad to full byte
          if len(colortable)%3!= 0:
            colortable += [0]*(3-(len(colortable)%3))

          assert(len(colortable)%3 == 0)

          colortable = bytes(colortable)

          out.write(colortable)

          # Skip over the rest of the colors in our palette since we only support
          # 256
          leftovercolorbytes = max((numcolors - pow(2, ord(lsd[5])) )*3, 0)
          if leftovercolorbytes > 0:
              rng.seek(leftovercolorbytes, 1)
                  
          
        # Write the actual image data
        imgdata = rng.read(size-1)
        out.write(imgdata)