#!usr/bin/env python3
"""
For reference, see
https://www.w3.org/Graphics/GIF/spec-gif89a.txt

This script generates a random GIF file using the provided RNG object
as its source of randomness to ensure the generated file contains no
vulnerable input. The resulting GIF file should still trigger the vuln,
but the exact location may vary depending on how well your RNG works.
"""

import argparse
import sys
import struct
from typing import Tuple, BinaryIO

def parse_args():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--seed', type=int, default=None,
                        help="A seed value")
    return parser.parse_args()


def write_chunk(out: BinaryIO, label: str, contents: bytes):
    out.write(struct.pack('<sH', label.encode(), len(contents)))
    out.write(contents)


def write_chunk_with_label(out: BinaryIO, label: str, chunkname: str,
                           contents: bytes):
    write_chunk(out, label, struct.pack('<s*', chunkname.encode()) + contents)


def generate_logical_screen(width: int, height: int):
    return struct.pack("<HHBBB", width,
                       height,
                       # Background colour index; 0 = black
                       background_colour := 0,
                       pixel_aspect_ratio := 0,
                       # Global colour table only has one entry right now
                       # We add another entry later when writing palette
                       palette_size := 1 << background_colour,
                       # Colour resolution is used for both RGB and alpha components
                       # It can take any nonzero value, but it must be less than or
                       # equal to the bit depth, which is fixed at 8 in a logical screen
                       colour_resolution := min(
                               max((palette_size - 1).bit_length(), 1), 10
                           ))



def generate_global_colour_table(background_palette_entry: int):
    return struct.pack("<BBBB", *[i//2**8 for i in [
            0x00FF00FF, 0xFF00FFFF,  0xFFFF00FF, 0xFFFFFFFF,
            0xFF0000FF, 0xFFFFFF00,  0xFF00FF00, 0xFFFF0000,
            0xFF00FFFF, 0xFF0000FF,  0xFFFFFF00, 0x00FFFFFF,
            0xFFFFFF00, 0xFFFF00FF,  0xFF00FFFF, 0xFF00FF00]]) \
           + struct.pack("B"*3*background_palette_entry, *(0 for _ in range(3*background_palette_entry)))


def generate_frame_info(width: int, height: int, local_colour_table: bool,
                        dispose_method: int, transparency_flag: bool,
                        user_input_required: bool, delay_time: float):

    xoffset = yoffset = 0
    return struct.pack('<HHBBBBBHBHBB', width, height, 0xF7,
                       0, 1<<transparency_flag, 8,
                       0, 0, 0 if local_colour_table else 1,
                       dispose_method | user_input_required << 2 | transparency_flag << 3,
                       delay_time*100, width, height,
                       xoffset, yoffset)


def generate_image_descriptor(local_colour_table: bool,
                              left_position: int, top_position: int,
                              width: int, height: int):

    return struct.pack('<BHHHHBB', 0x2C,
                       0, left_position, top_position, width, height,
                       0, 0, not local_colour_table) \
           + struct.pack('<L', left_position * top_position)


def generate_block(rng: BinaryIO, chunktype: str) -> bytes:
    size = struct.unpack('<H', rng.read(2))[0]
    return f'{chunktype}:{size}'.encode() + rng.read(size)


def generate_data(rng: BinaryIO):
    yield from [generate_block(rng, 'logicalscreen'),
                generate_logical_screen(*struct.unpack('<hhbbbbb',
                                                     rng.read(9))),
                generate_global_colour_table(
                    struct.unpack('B', rng.read(1))[0])]

    for _ in range(257):
        yield from [generate_block(rng, 'imagedescriptor'),
                    generate_image_descriptor(
                        struct.unpack('?', rng.read(1))[0],
                        *struct.unpack('<HHHHBB', rng.read(9))),
                    generate_block(rng, 'palette')]

    for _ in range(257):
        yield from [generate_block(rng, 'imagecontents'),
                    generate_data(rng)]


def main():
    args = parse_args()

    try:
        rng = open('/dev/urandom', mode='rb') if args.seed is None \
              else open(args.seed, mode='rb')
    except FileNotFoundError:
        raise RuntimeError('Could not find /dev/urandom; please specify --seed')

    # Magic numbers

    header = b"GIF89a"
    version = b"\x00"
    background_colour = b"\xFF"  # Black
    for _ in range(3):
        yield version[0]

    dim_h, dim_w = struct.pack(">HH", 100, 200)
    for val in dim_h: yield val
    for val in dim_w: yield val

    packed_flags = struct.pack("<B", 0xD7)  # Global Color Table flag
    for val in packed_flags: yield val

    bg_color_index = b"\x00"

    pixel_aspect_ratio = b"\x00"

    # Logical Screen Descriptor
    yield from header
    yield from dim_h
    yield from dim_w
    yield from packed_flags
    yield from background_colour
    yield from pixel_aspect_ratio
    yield from bg_color_index

    # Global Color Table
    for idx in range(256):
        rgb = struct.pack(">I", idx*idx*idx)[1:]
        yield rgb[0]
        yield rgb[1]
        yield rgb[2]

    for block in generate_data(rng):
        for c in block:
            yield c


if __name__ == '__main__':
    try:
        with open(sys.argv[-1], "wb+") as fp:
            fp.writelines(main())
    except IndexError:
        print(__file__, ': error: missing filename parameter')