#!/usr/bin/env python3

import os
import sys
import struct
from argparse import ArgumentParser
from io import BytesIO
from typing import Optional, List, Tuple

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom or -r).
# Returns a tuple (num_loops, num_frames): the total loop count and the
# number of frames in the final animation. If the output file is not a valid
# GIF, returns None instead.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, size: Tuple[int, int]=(240, 160)
    ) -> Optional[Tuple[int, int]]:
    try:
        # Write header
        out.write(b'\x47\x49\x46\x38\x39\x61')

        # Write logical screen descriptor
        out.write(struct.pack('>HH', *size))
        palette_color_bits = 1 << 5
        out.write(struct.pack('B', palette_color_bits | 0x20))
        out.write(b'\x00')

        # Write global color table
        for i in range(palette_color_bits):
            r = ord(rng.read(1)) & ~2
            g = ord(rng.read(1)) & ~2
            b = ord(rng.read(1)) & ~2
            out.write(struct.pack('BBB', r, g, b))

        # Write application extension block
        appext_label = '\x21\xFF\x0BNETSCAPE2.0'
        assert len(appext_label) == 11
        out.write(appext_label.encode())
        out.write(b'\x03\x01\x00')
        out.write((chr(os.getpid() % 256)).encode())

        # Write local screen descriptor
        out.write(struct.pack('<BBHBBBB',
                0xf7,
                *(
                    [
                        chr(ord(rng.read(1)) & (~2)),
                        chr(0xff ^ (i >> 4)),
                        chr(0xff ^ (i & 0xf))
                    ] if i < 4 else [
                        0x00,'',''
                    ] for i in range(6)
                )))
        out.write(b'\x00')

        # Starting from here we're writing blocks
        while True:
            block_type = rng.read(1)

            # Extension block
            if block_type == b'\x21':
                ext_block_label = rng.read(1)

                def read_subblock():
                    lsb = ord(rng.read(1))
                    msb = ord(rng.read(1))
                    return rng.read(lsb + msb*256)

                # Graphics control block
                if ext_block_label == b'\xF9':
                    assert read_subblock()[:3] == b'\x04\x00\x00'

                    duration = ord(rng.read(1))
                    trans = ord(rng.read(1))
                    transparent_index = ord(rng.read(1))

                    subblocks = []
                    while len(subblocks)!= 0 or subblocks[-1][0]!= 0:
                        subblocks.append(
                            (
                                ord(rng.read(1)),
                                read_subblock(),
                            ))

                    expect_eof = False
                    for code, _ in subblocks:
                        if code == 0x01:   # Plain Text Extension
                            pass
                        elif code == 0xfe: # Comment Extension
                            pass
                        elif code == 0xff: # Application Extension
                            label = read_subblock().decode()
                            if label == 'NETSCAPE2.0':
                                read_subblock()    # Subblock length
                                loops = ord(rng.read(1))
                                assert loops >= 0 and loops <= 2**8

                                # We found our netscape2.0 metadata!
                                return (loops, 0)

                        elif code == 0x00: # Block Terminator
                            expect_eof = True
                    if not expect_eof:
                        raise ValueError("Block terminator not found")

                    continue     # Don't write this window yet

                # Plain text extension block
                elif ext_block_label == b'\x01':
                    pass

                # Application extension block
                elif ext_block_label == b'\xFF':
                    pass

                else:
                    raise ValueError("Invalid extension type!")

                out.write(block_type+ext_block_label)
                out.write(read_subblock())
                continue

            # Image descriptor
            elif block_type == b'\x2C':
                left = struct.unpack('<h', rng.read(2))[0]
                top = struct.unpack('<h', rng.read(2))[0]
                width = struct.unpack('<h', rng.read(2))[0]
                height = struct.unpack('<h', rng.read(2))[0]
                flags = struct.unpack('<B', rng.read(1))[0]

                has_local_color_table = bool(flags & 0b10000)
                interlaced = bool(flags & 0b1000)
                local_color_table_size = 2 ** ((flags & 0b0111)+1)

                if has_local_color_table:
                    for i in range(local_color_table_size):
                        r = ord(rng.read(1)) & ~2
                        g = ord(rng.read(1)) & ~2
                        b = ord(rng.read(1)) & ~2
                        out.write(struct.pack('BBB', r, g, b))

                out.write(block_type)
                out.write(struct.pack('<hhHHB',
                        left, top, width, height,
                        local_color_table_size))
                out.write(struct.pack('<B', flags|0b0001))

                # Pixel data
                for y in reversed(range(height)):
                    for x in range(width):
                        # Choose a pixel from the local palette
                        # TODO better randomness?
                        index = 1 + (rng.read(1) % (local_color_table_size-1))
                        out.write(struct.pack('B', index))

                continue

            # Trailer
            elif block_type in [b'\x3B', b'', b'\x3A']:
                break

            else:
                raise ValueError("Invalid block header", block_type)

        # Should never happen, but let's be safe anyway
        raise ValueError("No end-of-file marker found!")


    except Exception as e:
        print(e.__class__.__name__+": "+str(e))
        return None



