#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
  write_gif_header(out)
  
  while True:
    gif_blocktype = read_next_block_type(rng);
    
    # Check if this is the last block
    if gif_blocktype == None:
      break;
      
    writenextblock(out, gif_blocktype, rng);
    
def write_gif_header(outstream: BinaryIO):
  header = b'GIF89a';

  outstream.write(header)
  
def read_next_block_type(instream : BinaryIO)->bytes:
  # Read the first 2 bytes of the current block type which specifies 
  # whether it's compressed or not
  # If the first bit is set then it's compressed otherwise it isn't
  # https://www.w3resource.com/graphics/gif/format/gif-file-format.php
  # Bits 7 and 6 of the block type identify how large it is. The other bits specify what kind of information follows it.
  # Figure out the size of the block based upon those two bits
  blocksizeindicator = instream.read(1)[0];
    
  if (((blocksizeindicator & 0b10000000)>>7)!= 0):
    # There must be more than 256 blocks due to above rule
    assert False, "Block too big! Can't do > 256 because of limitation of 256 colors"
      
  if (((blocksizeindicator & 0b01000000)>>6)!= 0):
    # This header is in the form [<<flag, highbit, lowbit]] where 
    # flag is set to indicate that this is the last block
    # And the high bit and low bit together represent the size of the block
    bytelengthofsizeindicator = instream.read(1)
    return instream.read(ord(bytelengthofsizeindicator))
  
  else:
    # This header is in the form [[<<flag, size]]] where 
    # flag is not set to indicate that this is the last block
    # And size represents the size of the block
    return instream.read(1)
    
    
def writenextblock(outstream : BinaryIO,current_block_type : bytes, instream : BinaryIO ) -> int:
  # Get the block type
  fieldbyte = current_block_type[0];
    
  # Set to indicate that this is NOT the last block
  fieldbyte &= ~0b10000000;
  
  # Write the new header with the flag now unset
  outstream.write(struct.pack('=B',fieldbyte));
  
  # Write the size of the next block depending on flags
  block_length_indicator = current_block_type[0] & 0b01000000;
  if (block_length_indicator == 0):    
    # Just write the size value
    outstream.write(current_block_type);
    
  elif (block_length_indicator == 1):
    # First byte indicates the number of bytes for the second byte
    outstream.write(current_block_type[:1]);
    next_byte = ord(instream.read(1));
    outstream.write(struct.pack('=B',next_byte));
    
  # If there was some actual content in this block then write it     
  inblockcontent = block_length_indicator+1 < len(current_block_type)
  if inblockcontent:
    outstream.write(current_block_type[block_length_indicator+1:])
    