#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
#   1. 6 bytes of GIF magic number "GIF87a" or "GIF89a".
#   2. 2 bytes for the width of the frame in pixels.
#   3. 2 bytes for the height of the frame in pixels.
#   4. Byte specifying how many colors are used by this GIF. It may be one of {0x00, 0x01, 0xfe}.
#       a. If value is less than 0xFE then it means that there's only one color table available and the entire
         #    image can be represented using that color palette.
#       b. Value of zero means no global color table is needed.
#       c. Otherwise, an application must retrieve the Global Color Table from the Logical Screen Descriptor
#          of the first Image Descriptor after the logical screen descriptor (before the first Image Descriptor use
#          Local Color Table).
#   5. 2-byte delay time for each frame. This value represents the wait time before displaying the next frame,
#       so the delay time is half of what you might expect. e.g., if Frames per second (Fps) is set to 10 Hz,
#       then a Delay Time would have its MSB set as  0 and LSB set as 5. Note that the unit for Delay Time is
#       1/100th of a second. In other words, it is measured in hundredths of seconds. For example, 2 frames
#       at FPS of 10 should be displayed every two tens of seconds.
#   6. 1 byte for the pixel aspect ratio (pixel aspect ratio: PixAspectRatio): always set to 0x00 so it has been
#       ignored by most software.
#   7. The global color map. You will see these in a lot of files since they allow for color reduction without having
#       to drop any frames. They are not strictly necessary but some programs do expect them. The reasoning is simple.
#       Each bit in the color map specifies whether or not the corresponding entry in the red, green, blue, and alpha
#       channels will be applied to the current pixel in the frame. Since we are generating random data, there very well
#       could be duplicate colors between frames which would mean that we could just use less bits and save space. But
#       I'm just going to make sure the entries in the colormap aren't all unique because this makes things simpler too!
#       A better solution would be to store the RGB values multiplied by a constant instead of storing a color index. We
#       don't need to worry about calculating this constant since we know the size of the palette in advance. So here we go...
#       Let the global color table be GCT. Then let CR, CG, CB, CA be the four channel components of the color to add to the
#       GCT immediately after the last one was added. Let GA be 0xFF (i.e. opaque) unless otherwise specified. Then the
#       new entry to add to the GCT is:
#           GCT[newlen] := (CR << 3 | CG << 2 | CB << 1 | GA) & 0xFCFC; /* where newlen is (# of cols already + 1) */
#       That is, take the new color specified in RGBA with 8 bits per channel and shift it over three times to put the
#       five least significant bits into the correct position in the global color table. Now mask off everything except
#       12 bits leaving our new 12-bit entry in the correct place.
#   8. All the image data.
#       a. N bytes of image descriptor.
#       b. LZW min code size (LZWminsize).
#       c. N bytes of compressed image data.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    header = b'GIF87a' # Must start with either 'GIF87a' or 'GIF89a'.
    out.write(header)

    # Logical Screen Descriptor
    # Width of frame in pixles
    w = rng.read(2)
    out.write(w)

    # Height of frame in pixles
    h = rng.read(2)
    out.write(h)

    # How many colors that are used by this GIF? One of {0x01, 0xFE}.
    #   a. If value is less than 0xFE then it means that there's only one color table available and the entire
     #      image can be represented using that color palette.
    #   b. Value of zero means no global color table is needed.
    ncol = int.from_bytes(rng.read(1), byteorder='big')
    assert((ncol == 0) or (ncol < 0xFE))
    out.write(ncol.to_bytes(1, byteorder='big'))

    # Pixel aspect ratio. Always set to 0x00 so it has been ignored by most software.
    out.write(b'\0')

    # Global color map. You will see these in a lot of files since they allow for color reduction without having
    # to drop any frames. They are not strictly necessary but some programs do expect them. The reasoning is simple.
    # Each bit in the color map specifies whether or not the corresponding entry in the red, green, blue, and alpha
    # channels will be applied to the current pixel in the frame. Since we are generating random data, there very well
    # could be duplicate colors between frames which would mean that we could just use less bits and save space. But
    # I'm just going to make sure the entries in the colormap aren't all unique because this makes things simpler too!
    # A better solution would be to store the RGB values multiplied by a constant instead of storing a color index. We
    # don't need to worry about calculating this constant since we know the size of the palette in advance. So here we go...
    # Let the global color table be GCT. Then let CR, CG, CB, CA be the four channel components of the color to add to the
    # GCT immediately after the last one was added. Let GA be 0xFF (i.e. opaque) unless otherwise specified. Then the
    # new entry to add to the GCT is:
    #       GCT[newlen] := (CR << 3 | CG << 2 | CB << 1 | GA) & 0xFCFC; /* where newlen is (# of cols already + 1) */
    # That is, take the new color specified in RGBA with 8 bits per channel and shift it over three times to put the
    # five least significant bits into the correct position in the global color table. Now mask off everything except
    # 12 bits leaving our new 12-bit entry in the correct place.

    # Number of colors
    numcols = None
    while True: # Keep trying until we get a non-duplicate color
        numcols = int.from_bytes(rng.read(1), byteorder='big')

        if numcols <= len(set()): break

    if numcols > 255: raise Exception("Too many colors")
    
    # Add the colors to the global color table
    gct = []
    for _ in range(numcols):
        r, g, b = [int(_) * 8 for _ in rng.read(3)]
        a = 0xff
        val = ((r >> 3) | (g >> 2) | (b >> 1) | a) & 0xfcfc
        gct += [(val >> 8), (val & 0xff)]

    # Write the global color table
    out.write(struct.pack('B'*len(gct), *gct))

    # All the image data.
    #    a. N bytes of image descriptor.
    #       i. Left Position
    #       ii. Top Position
    #       iii. Width
    #       iv. Height
    #       v. Flags
    #         -> 0x20 -> local color table
    #         -> 0x40 -> interlace
    #         -> 0x80 -> sort flag
    #         -> 0x10 -> reserved
    #       vi. Index of the local color table (if flags indicate that there is one)
    imgdesc = lambda posx, posy, width, height, lctid : \
             posx | (posy << 8) | (width << 12) | (height << 16) | (lctid << 24)
    imgextent = lambda x, y, w, h : (x, y, w - x, h - y)
    imgextent87a = lambda x, y, w, h : (x//8, y//8, (x // 8) + w, (y // 8) + h)

    for frameidx in range(0, 250): ## FIXME: HACK! REMOVE THIS LINE WHEN DONE TESTING!!!!!!
        print(frameidx)
    
        # Image descriptor
        posx = rng.