import os
import time
import shutil
from typing import BinaryIO

# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # GIF87a or GIF89a
    magic = bytes([int('0x{}'.format(rng.read(1).hex()), 16)]) * 3

    # logical screen descriptor
    width = int.from_bytes(rng.read(2), 'little', signed=False) % (2 << 4)
    height = int.from_bytes(rng.read(2), 'little', signed=False) % (2 << 5)

    globalcolorcount = [
            0b000,     #  2^n    + 1 colors
            0b001,     #  2^(n+1)+ 1 colors
            0b010,     #  2^(n+2)+ 1 colors
            0b100      # >= 2^(n+3)+ 1 colors
        ][((width - 1) // 8) + ((height - 1) // 8)]

    colorcount = {
        2 ** n + 1 for n in range(globalcolorcount.bit_length()) if (
                2**(n+1)+1 <= width*height and
                len({(r, g, b)
                    for r in range(2**n+1)
                        for g in range(2**n+1)
                            for b in range(2**n+1)}) ==
                    2**(n+1)+1
        )}

    bgcolorindex = int.from_bytes(rng.read(1), 'little', signed=False)
    pixelaspectratio = b'\x00\x00'

    # application extension block
    appext = b'\x21\xFF\x0B\x4E\x45\x54\x53\x43\x41\x50\x45\x32\x2E\x30\x03\x01\x00\x00' \
        + b'\x00'
    appext += len(appext).to_bytes(2, byteorder='big', signed=False)
    appext += b'\x00'*(len(appext)%2)

    # graphics control extension block
    graphicsext = b'\x21\xF9\x04' \
        + b'\x00'*3

    # plain text extension block
    plaintedext = b'\x21' \
        + b'\x00'*3 \
        + b'\x01\x01\x00\x00' \
        + b'\xFE\x00\x00\x00' \
        + b'w\x00o\x00r\x00l\x00d\x00!\x00'

    extblocks = []
    if any([(
            appext[i]!= 0 for i in range(len(appext))
        ), (
            bgcol := bgcolorindex, len(colorlist := list(sorted(colorcount))) > 1,
            not all([(bgcol == bgcol_,) for bgcol_ in colorlist[:-1]])
        )]):

        # add an application extension block
        if appext[0]:

            # replace a single null-byte with a nonzero byte
            while rng.read(1): pass
            appext = bytes([rng.read(1)[0]]) + appext[1:]
            while len(appext)%2: appext += b'\x00'

            # pad to even length
            while len(appext)%2: appext += b'\x00'

        extblocks = [appext]

    # image descriptor
    imgext = b'\x2C' + bytes([0]) \
        + pos := int.from_bytes(rng.read(2), 'little', signed=False) \
        + int.from_bytes(rng.read(2), 'little', signed=False) \
        + int.from_bytes(rng.read(1), 'little', signed=False) \
        + int.from_bytes(rng.read(1), 'little', signed=False) \
        + int.from_bytes(rng.read(1), 'little', signed=False) \
        + ord(magic[-1:]) & ~0xf0 | 0x20 # noninterlaced LZW compression

    datastream = b''

    maxpos = sum(map(lambda c: 2**c + 1, [
        p.bit_length() // 8 for p in (*imgext[:2],)
    ]))

    while pos < maxpos:
        if (len(datastream) >= 2 << 10): break # maximum image size is 2^10

        if pix := rng.read(1):
            datastream += pix
            continue

        if code := int.from_bytes(rng.read(1), 'little'):
            datastream += bytes([code])
            continue

        if code := int.from_bytes(rng.read(2), 'little'):
            datastream += bytes([code])
            continue

        if code := int.from_bytes(rng.read(3), 'little'):
            datastream += bytes([code])
            continue

    # end of file
    endofdata = b'\x3B'
    outbuffer = b''.join((
        magic,
        width.to_bytes(2, 'little'),
        height.to_bytes(2, 'little'),
        globalcolorcount.to_bytes(1, 'little'),
        colorcount | set(),
        bgcolorindex.to_bytes(1, 'little'),
        pixelaspectratio,
        segs := len(extblocks)*2 + 1 + len(graphicsext) + len(plaintedext) + len(imgext) + len(datastream) + len(endofdata) if extblocks else [],
        *(len(b).to_bytes(2, 'big') for b in extblocks),
        graphicsext,
        plaintedext,
        imgext,
        datastream,
        endofdata
    ))
    assert len(outbuffer)==sum(segs), "Buffer lengths don't match!"
    out.write(outbuffer)