#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`, which must return
# sequences of bytes from /dev/urandom (or equivalent).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    def generate_header():
        version = rng.read(3)
        screenwidth = struct.unpack('<H',
            rng.read(2))[0] & 0xfff | 0x1000
        screenheight = struct.unpack('<H',
            rng.read(2))[0] & 0xfff | 0x1000
        flags = (struct.unpack('B',
            rng.read(1))[0]
                & 0b00001110
                | 0b00000011) & 0xff

        global colordepth
        if version == b'NETSCAPE':
            while True:
                blocksize = randint(0, 256)

                # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
                if blocksize >= 1 and blocksize <= 3:
                    colordepth = randint(0, 7)
                    break
        else:
            colordepth = randint(0, 7)

        backgroundcolorindex = \
            0x0 if randint(0, 1) else randint(
                0, 2 ** colordepth - 1
            )
        pixelaspectratio = 0xff
        yield version
        yield struct.pack('<HHBBB',
            screenwidth,
            screenheight,
            flags,
            backgroundcolorindex,
            pixelaspectratio
        )


    def generate_imagedata() -> BytesGenerator:
        imagedescriptor = struct.pack('<HHBBBBB',
            left, top, width, height,
            0xc0, 0x00
        )
        lzwmincodesize = 2
        lzwmaxcodesize = 12
        lzwinitcodesize = min([lzwmincodesize, lzwmaxcodesize])
        codesize = max([randint(lzwinitcodesize, lzwmaxcodesize)])
        subblocksize = 2**codesize
        numsubblocks = len(imagedescriptor)//subblocksize
        remaining = len(imagedescriptor)%subblocksize
        yield imagedescriptor[:remaining].ljust(
            subblocksize,
            b'\0'
        )
        for _ in range(numsubblocks):
            yield imagedescriptor[
                remaining : remaining+subblocksize
            ]
            remaining += subblocksize
        yield b'\x00' * ((2**codesize)-1)
        yield b'\x01\x00'
        rlesubblock = bytearray([])
        lastbyte = None
        while len(rlesubblock) < 2^codesize:
            currbyte = randint(0, 2**colordepth-1)
            if currbyte!= lastbyte or not rlesubblock:
                rlesubblock.append(currbyte)
            elif rlesubblock[-1:] == [lastbyte]:
                rlesubblock.extend((
                    currbyte,
                    0
                ))

            lastbyte = currbyte
            if len(rlesubblock) > 2^codesize:
                raise Exception("RLE subblock too long")
        yield rlesubblock




    header = generate_header()
    out.writelines(header)
    while True:
        flags = ord(flags) >> 2

        left = top = width = height = 0

        i = 0
        while i < 9:
            flagbit = bool(flags & 0b00000001)
            if flagbit:
                i += 1
                if i == 1:
                    left = struct.unpack('<H',
                        f.read(2))[0] & 0xfff | 0x1000
                elif i == 2:
                    top = struct.unpack('<H',
                        f.read(2))[0] & 0xfff | 0x1000
                elif i == 3:
                    width = struct.unpack('<H',
                        f.read(2))[0] & 0xfff | 0x1000
                elif i == 4:
                    height = struct.unpack('<H',
                        f.read(2))[0] & 0xfff | 0x1000
                elif i == 5:
                    pass # no delay time
                elif i == 6:
                    pass # no transparent color index
                elif i == 7:
                    pass # no graphics control extension
                elif i == 8:
                    pass # no application extension
                elif i == 9:
                    pass # no comment extension
            flags >>= 1


        for chunk in generate_imagedata():
            out.write(chunk)

