#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Determine whether an application extension should be added
    addappext = False
    if rng.randint(0,1)==1:
        addappext = True

    # Determine how many frames will go in the animation
    numframes = rng.randint(0,6) + 1
    
    # Generate the global color table
    colors = [rng.randint(0x00, 0xFF) for _ in range(2 ** int((lsd[4] >> 5)))]
    bgndidx = rng.randint(0, len(colors)-1)

    # Pad the color table with black so that its length is a power of two 
    while len(colors)<pow(2, int(lsd[4])&7):
        colors.append(0x00)
        
    # Randomly permute the colors to make them look more interesting
    rng.shuffle(colors)

    # Determine which color(s) to use as transparent pixels
    transpcolors = []
    if addappext:
        while len(transpcolors)<numframes+1:
            idx = rng.randint(0, len(colors)-1)
            if not (colors[idx]==0x00 or colors[idx]==bgndidx):
                transpcolors.append(colors[idx])

    # Add GCE to output if we have one
    if addappext:
        appextheader = b'\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00'
        appexthdrsize = len(appextheader)+3
        framecountbytes = numframes.to_bytes(2,'little')[::-1]
        subblock = bytes([
            0x0b,           # Subblock label
            0x03,           # Application ID
            0x01,           # Subblock size
            0x00,           # Loop count indicator
            0x00,
            0x00            # Block terminator
        ])    
        appextbody = framecountbytes+subblock
        applen = len(appextbody)+appexthdrsize
        appextdata = appextheader[::-1]+applen.to_bytes(1,'big')[::-1]+appextbody
        out.write(appextdata)


    # Add GCT to output if we have it
    gctlen = pow(2,(int(lsd[4]&7)))
    if gctlen>0:
      gct = bytes([
          0xF7,           # Extension introducer
          0x00,           # Label
          0x00,           # Size
          ]+[colors[i] for i in range(gctlen)]+[0x00]*(4*(2**int(lsd[4]&7)))
        )
      out.write(gct)
    

    def writeframe():
        # Read next field byte for block type
        fieldbyte = rng.randint(0,255)
    
        # Check if this is the last block
        if ~fieldbyte & 128:
          raise StopIteration()
    
        # Get size of block
        size = rng.randint(0,255)

        # Check if there's any image data in this block
        hasimgdata = ~(fieldbyte&128) & 3 > 0
        
        # Write fields back to stream
        hdr = fieldbyte | 2
        out.write(hdr.to_bytes(1, 'big'))
        out.write(size.to_bytes(1, 'big'))

        # If there's no image data we're done here
        if not hasimgdata:
          pass

        # Write the local color table if necessary
        numcolors = size//3 - 1
        if numcolors > 0:
          colortable = bytes([])
          for i in range(min(numcolors, gctlen)):
              rgblit = colors[i].to_bytes(1, 'big')
              rgblit += colors[(i+1)%gctlen].to_bytes(
                  1,
                  'big')
              rgblit += colors[(i+2)%gctlen].to_bytes(
                  1,
                  'big')
              colortable+=rgblit[::-1]

          for i in range(max(numcolors, gctlen)-(numcolors%2), gctlen):
              rgblit = colors[i].to_bytes(1, 'big')
              rgblit += 0x00.to_bytes(1, 'big')
              rgblit += 0x00.to_bytes(1, 'big')
              colortable+=rgblit[::-1]
              
          pad = bytes([0x00]*((4*((colortable)//4))-(colortable)))
          colortable += pad

          out.write(colortable)


        # Now write image data 
        imgdata = bytes([])
        if hasimgdata:
          for i in range(size-1):
              rbit = rng.randint(0,1)
              gbit = rng.randint(0,1)
              bbit = rng.randint(0,1)

              rgbbits = rbit<<2|gbit<<1|bbit 
              imgdata += rgbbits.to_bytes(1,'big')
              
              if len(transpcolors)>0:
                  tbit = rng.randint(0,1)
                  idx = rng.randint(0, len(transpcolors)-1)

                  trnsbits = tbit << 1
                  trnsbits |= colors[transpcolors[idx]] 

                  imgdata += trnsbits.to_bytes(1,'big')
          out.write(imgdata)

    try:
        # Keep going until enough frames are written  
        for i in range(numframes):
            writeframe()
    except StopIteration:
        # Do nothing
        pass