import struct
import os

from typing import BinaryIO


class RandomGifGenerator:

    def __init__(self):
        self.__rng_fd = open('/dev/urandom')

        # The first two bytes of any valid gif are its magic number
        # and the length in bytes of it's subblocks. We can use this to
        # determine if we have enough bytes for another block.
        self.__header_len = len('GIF89a')

    @property
    def is_eof(self) -> bool:
        return not self.__rng_fd or self.__rng_fd.closed

    def close(self) -> None:
        try:
            assert isinstance(self.__rng_fd, BinaryIO)
            self.__rng_fd.close()
        except AssertionError:
            pass

    def read_block(self) -> tuple[int, bytes]:
        buf_header = self.__rng_fd.read(self.__header_len)

        if self.is_eof:
            raise EOFError("no more blocks left")
        elif len(buf_header) < self.__header_len:
            raise IOError("ran out of bytes while reading Gif block header")
        
        # Read size field from header. This tells us how many bytes long
        # our buffer will be.
        buf_size = struct.unpack('<H', buf_header[:2])[0]

        # Skip past the rest of the header since it takes up too much space
        self.__rng_fd.seek(self.__header_len - 2, os.SEEK_CUR)

        # Read actual image data depending on what was written inside the
        # header earlier. Make sure that there actually exist that many
        # bytes remaining before throwing an error.
        buf_data = self.__rng_fd.read(buf_size)

        if self.is_eof and len(buf_data)!= buf_size:
            raise IOError(
                f"ran out of bytes after reading {len(buf_data)} bytes."
                'Expected {}.'.format(buf_size)
            )

        return buf_size, buf_data

    def generate_random_gif(self, fd: BinaryIO):
        fd.write(b'GIF89a')

        while True:
            block_size, block_content = self.read_block()

            fd.write(struct.pack('<H', block_size))
            fd.write(block_content)