#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (default to /dev/urandom).
#
# The generated file is somewhat generic with some color table and palette
# but no graphics control extention nor local color tables. It does not have any
# other subblocks than those required below the Logical Screen Descriptor as
# defined in the spec.
def generate_random_gif(rng: BinaryIO = None, out: BinaryIO = None):
    if rng is None:
        rng = open('/dev/urandom', mode='rb')

    # Header
    out.write('GIF89a'.encode())
    width = struct.pack('<H', int.from_bytes(rng.read(2), byteorder='little'))
    height = struct.pack('<H', int.from_bytes(rng.read(2), byteorder='little'))
    bitdepth = struct.pack('B', ord(rng.read(1)) & 7)
    flags = bytes([ord(rng.read(1)) | 0x80])
    background = struct.pack('B', ord(rng.read(1)))
    aspect_ratio = struct.pack('B', ord(rng.read(1)))

    # Local Color Table
    ncolors = ((int.from_bytes(rng.read(1), byteorder='little') & 0x07) +
               1 << (ord(rng.read(1)) & 0x07)) * 3
    lct = bytes()
    for i in range(ncolors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        lct += bytes([r,g,b])

    # Image Descriptor
    out.write(width + height + bitdepth + flags
              + background + aspect_ratio)

    # Graphics Control Extension - Plain Black Frame
    out.write((0x21).to_bytes(length=1, byteorder='little'))
    out.write(b'\xF9')
    out.write(struct.pack('>HHBBBHB',
                          0, 0,    # Left/Top
                          width[0],   # Width
                          height[0],     # Height
                          0x00,         # Flags
                          b'\x00\x00',       # Delay Time
                          0xff           # Transparent Index
                         ))
    out.write(b';')

    # Plain Block
    out.write(lct)

    # Application Extension
    out.write(((0xFF).to_bytes(length=1, byteorder='little')))
    out.write((11).to_bytes(length=1, byteorder='little'))
    out.write("NETSCAPE".encode())
    out.write((3).to_bytes(length=1, byteorder='little'))
    out.write((1).to_bytes(length=1, byteorder='little'))
    out.write((0).to_bytes(length=2, byteorder='little'))
    out.write(";".encode())

    # Trailer
    out.write(b"")
