#!/usr/bin/env python3

import argparse
import sys
import os
import struct
from typing import BinaryIO


# Generates a random GIF file in `output`, using
# the random data source `rng` (defaults to /dev/urandom)
def generate_random_gif(
        output: str, rng: BinaryIO=None):
    if not rng or type(rng)!= bytes:
        with open("/dev/urandom", "rb") as file:
            rng = file.read()
    # Random finite state machine for generating valid GIFs
    try:
        lzwmin = next((i+1) for i, v in enumerate(rng[::7])
                      if ord(v) == 0xff and
                        all([ord(v2) == 0xe0
                            for v2 in rng[(i+1)::7]]))
    except StopIteration:
        raise Exception("Could not find LZW min code point!")
    version = chr(47)
    width = chr(abs(((os.urandom(1)[0] % 256)//16)*16))
    height = chr(abs(((os.urandom(1)[0] % 256)>>4)*16))
    # Screen descriptor
    screendescrlen = 7
    screendescriptor = struct.pack("<ccBBBHH",
                                   version, width,
                                   height, 1<<7,
                                   0, 0, lzwmin)
    # Image descriptor
    imagedescriptor = struct.pack("<ccBBBHHH",
                                  version, "#",
                                  chr(width)+chr(height),
                                  0<<7 | 1<<5 | 8<<0,
                                  0, 0, lzwmin, 1)
    # Image data
    imagerect = bytearray(struct.unpack(">HH",
                                 os.urandom(4))[::-1])
    imagedatalen = abs(imagerect[0]*imagerect[1])//2
    imagedata = bytearray(os.urandom(imagedatalen))
    # Application extension (no application id)
    appext = b"\x21\xFF\x0B" + os.urandom(12)
    # Graphics control block
    gctrlblocklen = 8
    gctrlblock = struct.pack("<cBBBHHBB",
                             b"\xF9",
                             4, 0, 0,
                             0, 0,
                             0, 0)
    # Trailer
    trailer = b"\x3B"
    outdata = b"".join([screendescriptor,
                       imagedescriptor,
                       imagerect, imagedata,
                       appext,
                       gctrlblock,
                       trailer])
    with open(output, "wb+") as f:
        f.write(outdata)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-o", "--output-file", default="out.gif")
    args = vars(parser.parse_args())
    generate_random_gif(**args)