#!/usr/bin/env python3

import struct
from typing import BinaryIO
from pathlib import Path
import sys

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # Now we're going to write 256 descriptors at once
    # Using this loop means we avoid making multiple calls to read()
    for i in range(257):
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

        # TODO: Should experiment with variations on these values too
        #       see https://github.com/decompme/decompilation-guide/issues/47
        out.write(b'\x2C\x00\x00\x00\x00\x00\x00\xFB\x00\x00\x00')

        for j in range(i):
            # Iterate over every other descriptor, writing each
            # one to disk

            # For each descriptor we'll randomly choose some number of
            # bytes from the input stream, then copy those to the output
            # stream.
            buf = rng.read(int.from_bytes(
                    rng.read(1), 'little'))
            out.write(buf)

        # Lastly, we can add any extra data to the end of this frame.
        # Note that this might cause us to exceed the maximum size of a GIF
        # (which is what led to CVE-2017-16232).
        # If this happens we'll just append nulls to fill the remainder
        # of the frame.
        rem = datalen-(j*i)+1
        if rem!= 0:
            buf = rng.read(rem)
            out.write(buf)



    last_byte = out.tell() - 1
    # At the moment this code doesn't support appending arbitrary
    # bytes onto the output so we'll have to manually compute where
    # our data will start relative to the rest of the file
    print(f'The generated gif is {last_byte} bytes long.')

    # For now we'll hardcode the number of loops to use, since we don't
    # yet know how much space remains between the final descriptor and EOF
    numloops = 0xC0

    paddinglen = (numloops * 3).to_bytes(size=2, byteorder='little')

    # Write the leftover bytes before the last descriptor
    print('Writing remaining bytes', end='', flush=True)
    out.seek(-1, 2)
    out.write(paddinglen)
    print('.', end='', flush=True)

    # Write the last descriptor!
    out.write(b'\x3B')

    # And pad up until the end of the file, which is easy enough because
    # we've already written exactly datalen bytes before ending.
    print('Padding global color table', end='', flush=True)
    while True:
        if out.tell() == datalen+1: break
        out.write(rng.read(1))
    print('. Done!', flush=True)



if __name__=='__main__':
    with open('/dev/urandom', 'rb') as f:
        generate_random_gif(f, sys.stdout.buffer)