#!/usr/bin/env python3

import sys
import os
import argparse
import struct
from typing import BinaryIO


# Generates a random GIF file with maximum width `maxwidth`, max height
# `maxheight`, and with up to `maxcolors` colors into `out` using the
# random number generator `rng` (/dev/urandom). Returns true on success, false otherwise.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, maxwidth: int, maxheight: int, maxcolors: int
    ) -> bool:
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          break
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
            continue

        # Don't generate images larger than MAXWIDTH or MAXHEIGHT
        if maxwidth >= 0xFFFF or maxheight >= 0xFFFF:
            raise ValueError("Invalid maxwidth or maxheight")
        elif size >= 65536:
            raise ValueError("Image too large")
        else:
            width = min(max(1 + (rng.read(1)[0] % maxwidth), 1), size/3)
            height = min(max(1 + (rng.read(1)[0] % maxheight), 1), size/3)

        # Add an extra frame if we finished the last one early
        if width <= 0 or height <= 0:
            break;

        # Generate palette
        numcolors = min(min(maxcolors+1,(size // 3)-1), 2**((lsd[4]&7)+1)*3)
        palettemasks = [
            i | 1 << 7
            for i in range(numcolors)]    
        palmap = [x for x in rng.read(numcolors)][::-1]
        palette = bytes([palmap[i] for mask in palettemasks for i in range(3) 
            if mask&1!= 0])

        # Write the local color table if necessary
        if len(palette) > 0:
          out.write(palette)

        # Now write image data 
        imgdata = rng.read(width*height)
        out.write(imgdata)
        padding = b'\x00\x00\x00';
        if width%8>0:
          out.write(padding*(8-(width%8)))
        if len(imgdata)%8>0:
          out.write(padding*((len(imgdata)//8+1)*8 - len(imgdata)))

    return True


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-w', '--maxwidth', dest='maxwidth', required=False, default=-1, type=int, help='Maximum width')
    parser.add_argument('-H', '--maxheight', dest='maxheight', required=False, default=-1, type=int, help='Maximum height')
    parser.add_argument('-c', '--maxcolors', dest='maxcolors', required=False, default=-1, type=int, help='Maximum number of colors')
    args = parser.parse_args()

    rng = open('/dev/urandom','rb')     # use /dev/urandom as our RNG
    outfile = open('test.gif','wb')    # write.gif files into current directory

    try:
      succeeded = generate_random_gif(rng,outfile,
                                   args.maxwidth,args.maxheight,args.maxcolors);
      print ("Generated", "succeeded" if succeeded else "failed");
      exit(not succeeded)
    except Exception as e:
      print (e.__class__.__name__, ':', e)
      exit(-1)
