#!/usr/bin/env python3
"""
Random GIF Generator 

Generates a random GIF file into `out` using the
random number generator `rng` (/dev/urandom).
"""
__author__ = "<NAME>"
 
import math
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    """
    Generates a random GIF file into `out` using the
    random number generator `rng` (/dev/urandom).

    Parameters
     ----------
    rng : BinaryIO
       The random number source to use for generating
       the GIF file.
    out : BinaryIO 
       The output file which will contain the generated
       GIF file
    Returns 
    -------
    None
    """

    # Read in all of hte bytes from /dev/urandom and then convert them
    # into an array of integers ordered big endian
    data = [int.from_bytes(b, byteorder="big") for b in iter(lambda : rng.read(1), b"")]

    # We're gonna need an index so we can reference each bit easiy during encoding
    idx = 0;
        
    # Header is GIFF 89a plus some extra 4 bytes of data length
    header = b'GIF89a'
    out.write(header + len(data).to_bytes(3,"big"))
            
    while True:
            
        # Frames are composed of a single field header
        # followed by the actual data
        
        # First read in the block size       
        frame_size = get_block_size(idx, data);

        if frame_size == -1:
            break;
                
        # Write it as a fixed width integer in big endian format
        out.write((frame_size).to_bytes(2,"big"));

        for _ in range(math.floor(frame_size / 8)):
            
            # Get a byte worth of bits from our randomly generated integer
            # pointer idx keeps track of how many remaining bits we have
            # left, and as such when idx reaches zero we've exhausted 
            # the input data stream  
            byte = extract_byte(idx, data);

            # Encode the byte into its corresponding GIF escape sequence
            # This means setting the first bit high in every byte but the last            
            out.write((((~(byte >> idx)) << 7) | (byte >> idx+1))).to_bytes(1,"big");

            idx += 8;
                
def extract_byte(idx : int, data : list[int]):
    """
    Extracts a single byte of data at position <idx> within the array <data>.

    Parameters 
     ----------
    idx : int 
        Index where we should start extracting
    data : List[int]
        Array containing the raw data to extract from
        
     Returns 
    -------
    An integer representing one byte of extracted data
    """
    result = 0x0;
    
    # Go through each of the bits in this byte until either there's no
    # more data or we've extracted enough
    while not(idx >= len(data)) and result <= 0xff: 
        result <<= 1;
        result |= ((data[idx // 8] >> (~idx & 0x7)) & 0x1);
        idx += 1; 
    
    return result;    
    

def get_block_size(idx : int, data : list[int]):
    """
    Reads a block size from the given <data> starting at position <idx>

    Parameters 
     ----------
    idx : int 
        Index pointing at the current point in the input stream
    data : List[int]
        Raw data to parse
        
     Returns 
    -------
    A positive integer indicating the size of a block of compressed data    
    """
    while idx < len(data):
    
        # Check that the block starts with a clear code                
        value = (((data[idx // 8]) >> ~idx & 0x7) & 0x1);
        if value!= 0:
        
            # Block marked clear code, so go ahead and read what follows        
            sizebits = (((data[(idx // 8)+1]) >> ~((idx+1) & 0x7) & 0x1) << 1) \
                       | (((data[(idx // 8)+2]) >> ~((idx+2) & 0x7) & 0x1));
            
            size = int.from_bytes([value, sizebits], byteorder="big", signed=False);
                
            return size;
                    
        else:
            # Block was not cleared so increment the index and keep on looking
            idx+= 1;
            
    # If we got here we ran out of data before finding our next frame so return -1
    return -1;