#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rnd.get_uint8() << 7 | rnd.get_uint8()
    
    # Set version to 0x03 which means the full range from 
    # [0..65535]. This should be compatible with most decoders.
    # The first bit indicates this is a "logical" screen; a fully
    # visible screen. The second bit indicates whether a global
    # color table follows. Disabling the global color table is rarely used
    # but still technically valid according to the spec.
    lsd |= 0x30

    # Bit 3 and 4 set the pixel aspect ratio to 1:1. In other words,
    # each pixel in the frame counts as two vertical resolution pixels.
    # Note that while the spec does permit nonstandard values here,
    # many GIF decoders do not honor these settings anyway.
    lsd |= 0xc
   
    # Set bit 7 to indicate a fixed color table size of 256 entries,
    # rather than the usual variable number of entries.
    lsd |= 0x80 

    # Add LSD to output
    out.write(bytes([lsd]))

    # Read global color table length from LSD
    gctlen = int(pow(2,(lsd & 7))+1)*(lsd&0x80)>>7

    # Add GCT to output if we have one
    if gctlen > 0:
      gct = []
      i = 0
      while i < gctlen:
        r = rnd.get_uint8()
        g = rnd.get_uint8()
        b = rnd.get_uint8()
        gct += [(r,g,b)]
        i+=3
      
      # If we don't get an even number of colors then we need padding to make sure
      # all resulting images contain the same amount of bytes
      outbytesleft = 256 - len(gct)
      if outbytesleft % 3!= 0:
        gct += [(0,0,0)]*outbytesleft
      out.write(gct)

    while True:

        # Read next field byte for block type
        fieldbyte = ord(rnd.get_uint8())
       
        # Check if this is the last block
        if ((fieldbyte & 128)!= 0):
            fieldbyte ^= 128
            
        # Get size of block
        size = rnd.get_uint8()
 
        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0)
  
        # Write fields back to stream as they are
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
            continue
    
        # Write the local color table if necessary
        numcolors = size//3 - 1
        if numcolors > 0:

            # We can only handle up to 256 colors so just fake that many and skip
            # over the rest
            if numcolors > 256:
                numcolors = 256
    
            colortable = []
            i = 0
            while i < numcolors:
                r = rnd.get_uint8()
                g = rnd.get_uint8()
                b = rnd.get_uint8()
                colortable += [(r,g,b)]
                i+=1
        
            # Write GCE block
            out.write(struct.pack('=BBBBB', 
                                   0xF7,
                                   0x00,
                                   ord(chr(numcolors)),
                                   0x00,
                                   0x00
                                  )
                     )
            
            # Write color table
            out.write(colortable)
      
        # Write graphics control extension if necessary
        if size >= 4:
           d = rnd.get_uint16()
           disposalmethod = d & 0x1c
           userinputflag   = d & 0x20
           transparentidx  = d & 0xf

           e = rnd.get_uint8()
           delaytime       = e & 0xff
           transparencyleft = e >> 8
           transparency    = rnd.get_uint8()

        else:
            # These parameters aren't important because the client will always 
            # use their own defaults if the server doesn't send them. But it must
            # conform to GCE format if it does send them. A zero value is invalid
            # according to the spec but nonetheless many clients expect it when
            # sending blindly
            delaytime = 1
            transparencyleft = 0
            transparency = 0

        # Write the actual image data
        imgdatalen = size-1
        imgdata = []
        i = 0
        while i < imgdatalen:
            imgdata += [rnd.get_uint8()]
            i+=1
          
        # Write out last part of IDAT block
        out.write(struct.pack('=BBBB',
                               0xFE,
                               0x00,
                               delaytime >> 8,
                               delaytime & 0xff
                              ))
        
        # Write out Netscape extenstion
        out.write(struct.pack('=BBBBBBB',
                               0x21,
                               0xFF,
                               0x0B,
                               0x4E,
                               0x45,
                               0x54,
                               0x53,
                               0x43,
                               0x41,
                               0x50,
                               0x45,
                               ord('2'))
                 )
        
        # Write out application extension
        out.write(struct.pack('=BBBBBBB',
                               0x21,
                               0xFF,
                               0x0B,
                               0x4F,
                               0x70,
                               0x74,
                               0x69,
                               0x6F,
                               0x6E,
                               0x73,
                               0x20,
                               0x63,
                               ord(','),
                               ord(',')
                              ))
        
        # Write out NETSCAPE2.0 token
        out.write(struct.pack('=BBBBBBBBBBB',
                               0x3B,
                               0x03,
                               0x03,
                               0x00,
                               0x00,
                               0x00,
                               0x00,
                               0x00,
                               0xFF,
                               0x00,
                               0x00,
                               0x00
                              ))