#!/usr/bin/env python3

import os
import sys
import math

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)
    lsd[4] |= 2
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have one
    if gctlen > 0:
      gct = rng.read(gctlen)

      # If we don't get an even number of colors then we need padding to make sure
      # all resulting images contain the same amount of bytes
      outbytesleft = 256 - len(gct)
      if outbytesleft % 3!= 0:
        out.write(b'\x00'*(outbytesleft%3))

      out.write(gct)
      
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          break
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream as they are
        out.write(struct.pack('<BB', fieldbyte|2, size))

        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        numcolors = min(math.ceil(size / 3), 256) - 1

        # We can only handle up to 256 colors so just fake that many and skip
        # over the rest
        if numcolors > 256:
            numcolors = 256
            
        colortable = []
        for _ in range(numcolors):
            colortable += [
                rng.read(1),
                rng.read(1),
                rng.read(1),
            ]

        out.write(bytes(colortable[:]))
          
        # Write the actual image data
        imgdata = rng.read(size-1)
        out.write(imgdata)