import os
import time
import shutil
from typing import BinaryIO


def _generate_random_gif(
        rng: BinaryIO, out: BinaryIO, length=10**7
    ) -> None:
    """Generate a random GIF of specified size in bytes using the provided
    binary stream as the source of entropy."""

    # 4 bytes: signature ("GIF")
    out.write(b"GIF")
    
    # 3 bytes: version code (always "87a", for GIF87a)
    out.write(b"87a")
    
    # 2 bytes: logical screen descriptor block
    #   - 1 byte: global color table flag (1 bit)
    #   - 3 bits: color resolution (3 bits): 7 colors
    #   - 1 bit: sort flag (1 bit): false, no need to sort palette entries.
    #   - 3 bits: global color table size (3 bits): 7 colors
    #       => we have 2^7==128 colors available
    out.write(int.to_bytes(
        0b01000100,
        1, 'little'))
    
    # 1 byte: background color index
    #     => unused so it is always set to zero
    out.write(b'\x00')
    
    # 1 byte: pixel aspect ratio
    #     => we're using square pixels so it's fixed at 0xF0 (~6.66:1)
    out.write(b'\xF0')
    
    # remaining bytes: image descriptor block
    #   1 byte: left position of the image
    #        => we use width 0 and don't actually display anything
    #   1 byte: top position of the image
    #        => we use height 0 and don't actually display anything
    #   1 byte: width of the image
    #        => this is set randomly between 1px and 512px
    #         (so that GIFs are never too large and the OS can read them quickly)
    #   1 byte: height of the image
    #        => same thing here : limited to either 1px or 512px
    #   1 byte: local color table flag
    #        => false because we already defined our own local color table earlier
    #   1 byte: interlace method
    #        => only 0 is valid value (no interlacing)
    #   1 byte: reserved flags
    #        => all zeros
    #   1 byte: LZW minimum size
    #        => not needed since we won't compress the generated GIF
    out.write(b'\x00\x00' \
      + (os.urandom(1)[0] % min(length//1024+1, 512))\
            .to_bytes(1, 'little') \
      +(os.urandom(1)[0] % min(length//1024+1, 512))\
            .to_bytes(1, 'little') \
      + b"\x08\x00\x00")
    
    # 1 byte: image data subblock identifier
    out.write(b',')
    
    # write frame data according to first frame information
    #   1 byte: LZW minimum repetition count
    #           we don't perform any compression so it's just 0
    #   1 byte: LZW maximum repetition count
    #           again, we don't do compression so it's also 0
    #   1 char: end-of-subblock character
    #           the value FF marks an EOI in GIF images
    #   2 bytes: start of compressed data
    #           there's none in our case since we don't do compression
    repeated_color = False
    length -= len(out.getvalue())
    while length > 0:
        if length >= 256:
            out.write(b'\xFF\x00\xFF')
            length -= 255
        elif length == 1:
            out.write(b'\xFF\x01')
            break
        else:
            while True:
                repeat = max(
                    os.urandom(1)[0]%min(length+1, 256),
                    1)
            
                out.write((repeat-1).to_bytes(1, 'little'))
                length -= repeat

                if length <= 0:
                    break
        
    # 1 byte: terminator
    out.write(b';')
    

def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    _generate_random_gif(rng, out)
    
