#!/usr/bin/env python3

import os
import sys
import time
import pickle
import argparse
import traceback
import resource
import signal
import shutil
import tempfile
import subprocess as sp

from typing import BinaryIO

# Program name
PROGRAM = "gif"

# Default filename of generated files
DEFAULT_OUTFILE = f"{PROGRAM}-out.gif"

# Determines the content of each pixel based on the seed
def get_pixel_value(seed):
    return ord('%' if seed >=.4 else '.')

# Generates a random GIF file into `f` using the
# random number generator `rng`, with size max_width x max_height
# and frames duration ms apart
def generate_random_gif(
        rng: BinaryIO, f: BinaryIO,
        max_width: int, max_height: int,
        duration_ms=75
    ) -> None:

    # Emit header
    f.write("GIF89a".encode())

    # Determine screen width and height
    d = {}
    while True:
        w = rng.randint(0,max_width+1)
        h = rng.randint(0,max_height+1)
        if (w * h) > 50000: continue

        print('{}x{}'.format(w,h))
        break;

    d['width'] = w
    d['height'] = h

    # Indicate that there is no global color table
    f.write((b'\x08\x00').ljust(6," "))

    # Set color resolution to 256
    f.write((b'\xff\x00').ljust(6," "))

    # No background color index
    f.write(b"\x00\x00")

    # No aspect ratio specifier
    f.write(b"\x00\x00")

    # Size of global color table
    # We are not using one...
    f.write((b'\x00\x00').ljust(6," "))

    # Write per-image info

    # Image separator
    f.write(b"\x2c\x00")

    # Left corner specified by 2 bytes
    f.write(((d['width']).to_bytes(2, "little")).ljust(6," "))

    # Top corner specified by 2 bytes
    f.write(((d['height']).to_bytes(2, "little")).ljust(6," "))

    # Frame duration
    f.write((((duration_ms//10)*10).to_bytes(2, "little")).ljust(6," "))

    # Packed field: local color table flag set, interlace off, sort left->right
    # 0b10000000 == bit 7 set indicates presence of color table 
    f.write(b"\x40\x00")

    # Background color index
    f.write(b"\x00\x00")

    # Pixel Aspect Ratio - we don't use any!
    f.write(b"\x00\x00")

    # Number of bits used to specify each pixel value
    f.write(b"\x00\x00")

    # End of image descriptor
    f.write(b"\x3b")

    # Data sub-block starts here
    f.write(b"\x21\xf9\x04\x00")
    f.write(b"\x00\x00")
    f.write(b"\x00\x00")
    f.write(b"\x00\x50\x02\x0e\x00\x00\x00\x00")

    # Emit pixels
    for y in range(d["height"]):
        for c in range(3):
            p = [get_pixel_value(rng.random()) for _x in range(d["width"])] 
            blocksize = len(p)//256
            for k in range(blocksize):
              chunk = p[k*256:(k+1)*256]
              f.write(("{:02X}".format(len(chunk))).encode()) 
              f.write(b","+("".join(["{:02X}".format(i) for i in chunk])).encode()) 

