#!/usr/bin/env python3

import os
import sys
import math
import argparse
import tempfile
import shutil

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # 
    body_length = 0
    descriptor_count = 257
    max_descriptor_size = 10   # bytes
    min_descriptor_size = 6    # bytes
    fixed_descriptor_size = 10 # bytes

    # Determine the size of each descriptor
    # They must all be equal to or smaller than this value
    descriptor_body_length = \
        descriptor_count*fixed_descriptor_size \
        //max_descriptor_size*max_descriptor_size
    
    # How many times does the framedata section need to be repeated?
    repetitions = descriptor_body_length//descriptor_body_length
    
    # Last frame needs to have some additional padding
    extra_padding = descriptor_body_length%descriptor_body_length

    # Allocate memory for the entire file contents
    contents = bytearray(repetitions*(descriptor_body_length
                            +extra_padding)+extra_padding)
    
    # Generate the entire file first
    # FIXME: Offset from descriptor start should also depend on the index
    for i in range(repetitions):
        offs = i*descriptor_body_length
        
        contents[offs]      = ord(',')           # ','
        contents[offs+1]    = 0x2c              #,
        contents[offs+2]    = 0x00              # 
        contents[offs+3]    = 0x00              # Descriptor offset
        contents[offs+4]    = 0x00              #  
        contents[offs+5]    = 0x00              # Frame delay
        contents[offs+6]    = 0xfB              # LZW minimum code size
        contents[offs+7]    = 0x00              # 
        contents[offs+8]    = 0x03              # Image width (4 pixels)
        contents[offs+9]    = 0xe0              # Image height (64 pixels)
        
        # Make sure sizes are valid before writing them
        assert len(contents[offs:]) <= max_descriptor_size
        assert len(contents[offs:]) >= min_descriptor_size
    
        # Copy the data
        contents[offs+10:] = rng.read(len(contents[offs+10:]))
        
        # Keep track of the total body size
        body_length += len(contents[offs:])
        
    # Finalize the file
    # Write the length of the file and file type identifier
    contents[0]     = ord('G')             # Identifier 'G'
    contents[1]     = ord('I')             # Identifier 'I'
    contents[2]     = ord('F')             # Identifier 'F'
    contents[3]     = ord('8')             # Version string '8'
    contents[4]     = ord('9')             # Version string '9'
    contents[5]     = ord('a')             # Version string 'a'
    
    body_length_bytes = body_length.to_bytes(
            3, byteorder='big')
    
    contents[:3] = body_length_bytes
    
    
    # Write the whole thing to disk
    out.write(contents)
    
    
    
    
    
if __name__=='__main__':
    parser = argparse.ArgumentParser(description='')
    parser.add_argument('-o','--output','-O','--OUTPUT',
                        help='Output path for the file '
                             '(default: stdout)',
                        default=None)
    args = parser.parse_args()
    
    # Open our input file for reading
    f_in = open('/dev/urandom','rb')
    
    # Open our output file for writing
    if args.output==None:
        f_out = sys.stdout.buffer
    else:
        f_out = open(args.output,'wb')
    
    try:
        generate_random_gif(
                f_in, f_out
            )
    except KeyboardInterrupt:
        pass