#!/usr/bin/env python3

import argparse
import os
import struct
from math import floor, sqrt
from pathlib import Path
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 2 bytes of length of block (N).
#   3. N bytes of block.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # First block.
    metadata_block_length = 7
    graphics_control_extension_block_length = 8
    plain_text_extension_block_length = 0
    application_extension_block_length = 11
    comments_block_length = 0
    signature_block_length = 5
    palette_block_length = 3 * 768 + 1
    logical_screen_descriptor_block_length = 7
    color_table_block_length = 3 * 768 + 1
    image_data_block_length = 255 * 32 * 3
    frame_count = 1
    screen_width = 512
    screen_height = 32
    bits_per_pixel = 8
    background_color_index = 0
    aspect_ratio = 0
    extensions = [b'\x21\xff\x0bNETSCAPE2.0']
    loop_counter = 0
    trailer = b''
    
    output = bytearray()
    output += struct.pack('>cH', header[0].encode(), len(output))

    # Metadata extension sub-blocks.
    output += struct.pack('<B', metadata_block_length)           # Sub-block length. 
    output += struct.pack('<BB', 0, 5)                           # Extension introducer & label.
    output += struct.pack('<BBBBBBB', *map(len, extensions))      # Block lengths for each sub-block.
    for extension in extensions:                                 
        output += struct.pack(">BB", *bytearray(extension)[::-1]) # Sub-block content.
    output += struct.pack("<BBBBBBB", 
                            *map(len, [
                                b'',                                # Plain text extension. 
                                b'\x01',                           # Application extension.
                                b'',                                # Comments extension.
                                b'\x00\x00\x00\x00',               # Signature extension.
                            ])
                        )                                          # Block lengths for each sub-block.

    # Graphics control extension sub-blocks.
    output += struct.pack('<B', graphics_control_extension_block_length)   # Sub-block length. 
    output += struct.pack('<BB', 0, 4)                                   # Extension introducer & label.
    output += struct.pack('<B', 4)                                       # Block length.
    output += struct.pack('<BBBB',
                            0xf9 | 0b10000000,                         # Packed field where flags are set to 1 and reserved bit is set to 0.
                            0,                                      # Duration divisor (always zero).
                            0,                                      # Loop counter start point (always zero).
                            loop_counter                            # Common loop counter terminator value. 
                        )                                                   # Sub-block content.

    # Plain text extension sub-blocks.
    if plain_text_extension_block_length!= 0:
        output += struct.pack('<B', plain_text_extension_block_length)        # Sub-block length. 
        output += struct.pack('<BB', 0, 1)                                     # Extension introducer & label.
        output += struct.pack('<BBBB', 
                                    plain_text_extension_block_length - 2,       # Block size excluding label and length bytes.
                                    0,                                           # Text grid left position (always zero).
                                    0,                                           # Text grid top position (always zero).
                                    len("Comment"),                              # Number of characters in comment string.
                                )                                                      # Sub-block content.
        output += map(ord, "Comment").__iter__()                               # Comment string.
        while len(output) % 16!= 0:                                            # Pad out to even multiple of 16 bytes.
            output.append(0)
     
    # Application extension sub-blocks.
    if application_extension_block_length!= 0:
        output += struct.pack('<B', application_extension_block_length)         # Sub-block length. 
        output += struct.pack('<BB', 0, 0xff)                                  # Extension introducer & label.
        output += struct.pack('<BBBB', 
                                        application_extension_block_length - 2,     # Block size excluding label and length bytes.
                                        0x5e,                                      # Application identifier offset from 0xe0.
                                        0x5f,                                      # Application authentication code (always zero).
                                        len("\x01")                                # Length of application specific data.
                                  )                                                       # Sub-block content.  
        output += struct.pack('<BB', 1, ord('\x01'))                             # Application specific data.

    # Comments extension sub-blocks.
    if comments_block_length!= 0:
        output += struct.pack('<B', comments_block_length)                    # Sub-block length. 
        output += struct.pack('<BB', 0, 0xfe)                                  # Extension introducer & label.
        output += struct.pack('<BBBB', 
                                        comments_block_length - 2,                 # Block size excluding label and length bytes.
                                        0x5d,                                      # Label for this type of extension (always zero).
                                        0x5d,                                      # Label for this type of extension (always zero).
                                        len("\x01"),                               # Comment string's length.
                                     )                                                  # Sub-block content.
        output += map(ord, "\x01").__iter__()                                    # The comment string itself.

    # Signature extension sub-blocks.
    if signature_block_length!= 0:
        output += struct.pack('<B', signature_block_length)                  # Sub-block length. 
        output += struct.pack('<BB', 0, 0x37)                                  # Extension introducer & label.
        output += struct.pack('<BBBB', 
                                        signature_block_length - 2,                # Block size excluding label and length bytes.
                                        0x78, 0xae, 0xc6,                          # Version of signing algorithm used (always 1377662606).
                                        0xaf, 0xae, 0xce                     
                                 )                                              # Sub-block content.

    # Palette sub-blocks.
    output += struct.pack('<B', palette_block_length)                       # Sub-block length. 
    output += struct.pack('<BB', 0x2C, 0)                                  # Image separator & label.
    output += struct.pack('<BBBBHHHBB',
                                palette_block_length - 6,                   # Block size excluding label, length, and separator bytes.
                                768,                                      # Integer designating how many colors in the global color table.
                                768 >> 8,                                 # Global color table sorted by intensity (0=black, 255=white).
                                768 >> 16,                                # Global color table sorted by red green blue (0-255).
                                0,                                        # Reserved for future use (must be set to null (0)).
                                0,                                        # Background color index; 0 means no background color needed.
                                aspect_ratio,                             # Pixel Aspect Ratio (aspect ratio of pixel width vs. height).
                                background_color_index,                   # Background color index, or transparent if none wanted.
                                0                                         # Terminator.
                           )
   
    # Logical Screen Descriptor sub-blocks.
    output += struct.pack('<B', logical_screen_descriptor_block_length)    # Sub-block length. 
    output += struct.pack('<BB', 0x2A, 0)                                  # Image separator & label.
    output += struct.pack('<BBBBHHB',
                                logical_screen_descriptor_block_length - 6,   # Block size excluding label, length, and separator bytes.
                                screen_width,                                 # Logical width of display area.
                                screen_height >> 8,                           # Logical height of display area (in units of 2^8).
                                screen_height,                              # Logical height of display area (in units of 2^8).
                                0,                                           # Global color table flag (1=global color table follows, 0=no table defined).
                                0,                                           # Color resolution (bits per primary color): 1=1, 2=2,... 16=16, etc.
                                0,                                           # Sorting method (0=not sorted, 1=sorted)
                                0,                                           # Size of GCT (either 0 or 2^(value+1)): 0=2^n entries, 1=2^(n+1) entries,..., 7=2^(n+8) entries.
                                bits_per_pixel                               # Bits per sample (1, 2, 3, 4, 8, or 16).
                          )                                                   # Sub-block content

    # Color Table sub-blocks.
    output += struct.pack('<B', color_table_