"""Generate a random gif from /dev/urandom and write it to stdout."""

import sys
import struct
import os
import platform

def main():
    """Entry point for this program"""
    if not hasattr(os,'sysconf'):
        raise RuntimeError('This is only supported on Unix-like systems')
    elif platform.system() == 'Windows':
        raise RuntimeError('This is currently unsupported on Windows')
    else:
        bufsize = os.sysconf("SC_PAGESIZE") * (512 // os.sysconf("SC_PAGE_SIZE"))
        with open('/dev/urandom', 'rb', buffering=bufsize) as urand:
            generate_random_gif(urand, sys.stdout.buffer)
    return 0

def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    """Generate a random gif into `out`, reading bytes from `rng`."""
    # see https://www.w3.org/Graphics/GIF/spec-gif89a.txt section 16 for more info
    
    # first we make our header, which consists of 7 fields:
    # - signature (4 bytes)
    # - logical screen descriptor (7 bytes):
    #     - (1 byte) 0x2c (image separator)
    #     - (1 byte) x-origin coordinate (most significant 2 bits are ignored?)
    #     - (1 byte) y-origin coordinate (least siginificant bit is ignored?)
    #     - (1 byte) width in pixels
    #     - (1 byte) height in pixels
    #     - (1 byte) color table size; legal values are 0, 1, or 2^n for n>=128
    #       representing 2^(n+1) colors
    #       (some software may reject files that don't have an explicit color table
    #       because they're weird)
    #       (we'll always use 256-color tables so that's what we set here)
    #       [this is actually stored backwards wrt spec]
    #     - (1 byte) background color index
    #           (foreground color will be whatever color comes before it after every frame)
    #   total length is thus 10 bytes
    global _color_table
    _color_table = create_color_table()
    bg_color_index = len(_color_table) - 1

    header = (b'GIF89a'
              + b'\x00\x2c\x00\x00'
              b'\x00\xff\xfe\x00\x04\x00\x00\x00\x00\xff\xff'
             )


    # next up is our global color table, encoded as a palette
    # each entry takes 3 bytes (index + rgb triple)
    # the entire block is preceded by a single byte indicating how many entries there are
    # since we have exactly one entry right now, we just manually specify its size
    palette_block = b'\x01' \
                    + b''.join([
                        _color_table[i].to_bytes(
                            3, byteorder='little', signed=False)
                        for i in range(len(_color_table))])

    # then we have frames! We'll leave them empty so that they can later be filled
    # in with real data
    num_frames = 100
    frame_blocks = []
    for i in range(num_frames):
        frame_blocks.append(b'\x21\xf9\x04\x00\x00\x00\x00\x2c'
                            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                           )

    # encode all our blocks into their appropriate format
    header = struct.pack('>7sBBBBB',
                         header,
                         2 ** ((len(_color_table)-1).bit_length()),
                         *(bg_color_index,)                        
                        )
    palette_block = b'\x01' \
                    + b''.join([
                        _color_table[i].to_bytes(
                            3, byteorder='little', signed=False)
                        for i in range(min(len(_color_table), 256))])
    body = b''.join([(rle_compress(palette_block), rle_compress(block))
                     for block in frame_blocks])

    # finally we concatenate everything together and write it out
    written = out.write(header + body)
    print(f"written {written} bytes", file=sys.stderr)
    
def create_color_table() -> list:
    """Create a default 256-entry GIF color table."""
    # This uses a very dumb algorithm. It could probably be optimized, but who cares?
    # For explanation of the algorithm used here, see:
    # http://dissectinganimatedgifs.com/blog/tag/algorithm/
    color_table = []
    red = blue = green = current = step = 0
    middle_value = divide_if = 0
    while True:
        if middle_value!= 0:
            for i in range(middle_value):
                color_table += [current]
            current += step

        color_table += [current]
        
        if current > 255:
            break
        current += step

        if divide_if % 3!= 0:
            continue

        red += 1
        if red == 255:
            red -= 255
            blue += 1
        if blue == 255:
            blue -= 255
            green += 1
        if green == 255:
            green -= 255
            step *= -1

            if step < 0:
                middle_value = abs(step)
                step = 1
                
                if divide_if % 2!= 0:
                    divide_if += 1
            
    assert len(color_table) == min(2**((len(color_table)+1)/3), 256)
        
    return color_table

def rle_compress(data) -> bytes:
    """RLE compress some data (an array of ints)."""
    result = bytearray()
    last_val = -1
    count = 0
    max_count = 255
    for val in data:
        if val == last_val and count < max_count:
            count += 1
            continue
        else:
            count = 0
        if count >= 3:
            result.extend((last_val, 0, count)) 
        elif count == 2:
            result.extend((last_val, 0, 2*last_val)) 
            pass
        elif count == 1:
            result.extend(((last_val,)*3)) 
        else:
            result.extend(([val]*3))
        
        last_val = val
        count = 0
    return bytes(result)

    
