#!/usr/bin/env python3

import sys
import os
import random
import time

from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # TODO: get a better name than "rng"
    # TODO: make sure that rng open on /dev/urandom and that we handle errors correctly
    assert isinstance(rng, int), 'Can only use integer as RNG seed!'


    # Set up initial state variables

    # Initial value of field byte 
    fieldbyte = 0x80      # The first bit must be set to indicate more blocks are coming 

    # Padding bytes before LSD
    paddingbytes = [
        0x33, 
        0x66, 
        0xCC, 0x66, 
        0xFF, 0xFF]

    # Color table length (in bytes)
    ctablelength = 0       # No color table yet, but will be determined by other bits

    # Local color table provided?
    haslocctble = False    # No local color table initially

    # Transparent index
    transindex = None      # No transparency at all

    # Background index
    bgindex = None         # Whole background visible

    # Aspect ratio
    aspectratio = 0        # 4 by 3 square pixels (don't care since user can overwrite later)

    # Image positioning values
    xpos = 0               # Placeholder for now
    ypos = 0               # Placeholder for now

    # Image width and height
    imwidth = 0            # Will be determined by LZW code size
    imheight = 0           # Will be determined by LZW code size

    # LZW minimum code size
    mincodesize = 0        # Will be determined by user input or randomly chosen

    # LZW dictionary of codes to indices
    dictcodes = {}

    # LZW current pixel stack
    pixstack = []          # Stack that holds pixel values until they're written to an array

    # LZW current prefix code
    prefixcode = ''        # Prefix codes are stored as strings in a dictionary with indices

    # LZW next available code index
    nextindex = 0          # Will be incremented continuously

    # LZW previous code in dictionary
    prevdictcode = ''      # This contains the key (prefix code + current character) for the currently processed code


    # Output header
    print("Outputting header...", end='', flush=True)
    out.write(b'GIF89a')
    print("\t\u2714", end='')

    # Output fixed first three blocks 
    # according to spec
    print("Outputting fixed blocks...", end='', flush=True)
    for p in range(0, len(paddingbytes)):
      out.write(struct.pack('=B', paddingbytes[p]))

    # Build LZW dictionary from first block
    dictcodes[b'\x00'] = 0
    nextindex += 1
    prevdictcode = b'\x00'
    print("\t\u2714", end='')

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Save LSD info as our own variables
    gctlen = 2**(int((lsd[4] >> 5))) * 3   # Bits 5+5 for size, times three because we're storing RGB values
    bgcoloridx = lsd[-1]                     # Last byte is background color

    print("\nLogical Screen Descriptor:", end='\n\n')
    print("\tResolution:\t\t{width}X{height}".format(width=(2**lsd[4]), height=(2**lsd[5])))
    print("\tColor Table Size:\t{size} colors ({bytes} bytes)".format(size=gctlen, bytes=gctlen//3))
    print("\tBackground Index:\t#{bgcolor}".format(bgcolor=hex(bgcoloridx)[2:].zfill(2).upper()))

    # Add GCT to output if we have it
    if gctlen > 0:
      
      # Since LZW code sizes are specified by the user, we need to determine how much space 
      # we'll need for each color palette entry based off what code size was selected

      # First check if there's a specific code size supplied by the user
      if args.maxcodesize!= None:

        maxcodesize = args.maxcodesize # We will ignore anything larger than this

        # Use the smallest possible code size needed to store the amount of colors requested in the GCT
        requiredcodesize = math.ceil(math.log2(gctlen/3))
        mincodesize = min(requiredcodesize, maxcodesize)

      else:

        # Go through every possible code size in case one would work
        for i in reversed(range(1, gctlen//3+1)):

          # Calculate the maximum color palette entries that could fit within a single 
          # block using the given code size
          maxentriespossible = 2**i 

          # Is this a valid code size?
          if gctlen <= maxentriespossible*3:
            maxcodesize = i     # Valid code size found! Use it!
            break


      # Display result for user reference
      print("\nMaximum Code Size:\t{size}".format(size=maxcodesize))
      print("\tMaximum Entries Possible:\t{maxentries}".format(maxentries=maxentriespossible))

      # Store LZW data to GCT variable so we know where to put it when building our LZW dictionary
      gct = lsd[:-1]                 # Strip reserved portion above code size
      gct += struct.pack('>H', mincodesize)  # Append new code size specification
      gct += rng.read(gctlen-2-mincodesize) # Append remainder of color table

      # Add GCT to output if we have it
      print("\nWriting Global Color Table...\t", end='')  
      out.write(gct)
      print('\u2714', end='\n\n')
      print("Global Color Table:")

      # Print contents of GCT for debugging purposes
      for i in range(0, gctlen//3):
        idxstr = "#{idx:{}.{}x}".format(i+1, 2).format(i)
        rgbstr = "(R:{r}, G:{g}, B:{b})".format(r=rgb[i][0], g=rgb[i][1], b=rgb[i][2])
        print("{idx}: {rgb}".format(idx=idxstr, rgb=rgbstr))



    # Initialize helper arrays for easier color lookup and manipulation
    r = []
    g = []
    b = []
    rgb = []
    # rgbpop = []

    # Populate helper arrays with values from our GCT
    for i in range(0, gctlen//3):
      r.append(ord(gct[(i*3)+1]))
      g.append(ord(gct[(i*3)+2]))
      b.append(ord(gct[(i*3)+3]))
      rgb.append([r[i], g[i], b[i]])


    print("Building LZW Dictionary...")

    # While the GCT isn't finished, keep adding elements to the dictionary
    while (nextindex < pow(2, mincodesize)):

      # Randomly select two distinct colors to form the next element
      colorchoiceidx = random.randrange(0, gctlen//3)
      colorchoicestring = chr(colorchoiceidx).encode()

      # Make sure it's not already been added to our dictionary
      if colorchoicestring in dictcodes.keys():
        continue

      # Add to dictionary
      dictcodes[colorchoicestring] = nextindex
      nextindex += 1

      # For testing purposes, let us see what we've been doing thus far
      print("New LZW Element:\t'{string}' (#{idx}): {rgb}".format(string=colorchoicestring.decode(), idx=colorchoiceidx, rgb=rgb[colorchoiceidx]))

      # Did we just pick up a different color than the last one?
      if colorchoiceidx!= 0 and colorchoiceidx >= prevdictcode[0]:
        # Yes, we did. Update the prefix code
        prefixcode += colorchoicestring
      elif colorchoiceidx == 0 and 0 in dictcodes.values():
        pass
      else:
        # Nope, it's a repeat pattern. Reset the dictionary prefix code
        prefixcode = ''.join([chr(k).encode() for k in sorted(list(set(prevdictcode)))])

      # Keep track of the most recent sequence of characters included in 
      # the current element (the prefix code plus the current character)
      prevdictcode = prefixcode + colorchoicestring.decode()


    # Finished constructing the LZW dictionary; now begin generating pixels
    print("Generating Pixels...")

    # Keep looping until we reach the end of the GCT
    while (nextindex < pow(2, mincodesize)):
    
      # Choose another color to form the next element
      colorchoiceidx = random.randrange(0, gctlen//3)
      colorchoicestring = chr(colorchoiceidx).encode()

      # Make sure it's not already been