#!/usr/bin/env python3

import bisect
import functools
import io
import math
import os
import sys


# max is exclusive
@functools.lru_cache(maxsize=None)
def randrange(min_: int, max_: int):
    return min_ + (os.getrandom()[0] % (max_-min_))



@functools.lru_cache(maxsize=None)
def get_palette():
    return [
            bytes([r, g, b]) for r in range(256) for g in range(256)
                for b in range(256) if sum(map(abs, (r,g,b))) > 384
            ]




@functools.lru_cache(maxsize=None)
def get_lzw_compression_table() -> tuple[list[int], dict]:

    # The first element of each entry is the prefix code; the second one contains
    # the length and index of the table to be used by LZW decompressor after the
    # current byte.
    #
    # Note that there are at most 257 entries because we need one extra slot
    # to allow the literal characters to use them as their own prefixes when
    # prefixed with FF 00. This way we don't have to do any special case to deal
    # with them.
    compressionTable = [(i,) for i in range(257)]

    # 0xFF 0x00 means "the last code in the dictionary", so it's not strictly part
    # of the standard but we add it anyway for convenience
    compressionTable += [
        (bytes([c]), (len(prefixes)+1, len(prefixes) - c)) for c in range(256) 
            for prefixes in (
                    filter(
                        lambda p: p[-1] == c or p[-1] == 0xFF, 
                        compressionTable[:-1]
                    ),
                    []
                )
            ]
    return compressionTable, {p[0][-1]: p[1:] for p in compressionTable}




def lzwwrite(data: bytes, out: BinaryIO):

    # We know that GIFs can't support more than 2^12 bits, since it's coded on two bytes only
    assert(sum(map(lambda x: 1 << x, range(2*9+1))) <= 2**12 * 2)


    def writebyte(b):

        if isinstance(b, str):
            raise TypeError('Expected an integer argument')
        
        while True:

            # If the high bit is set then the next byte will determine how many
            # additional bytes follow this one. There is always at least one byte following the
            # initial one. A zero value means just the next byte will be used.
            if (b & 128)!= 0:

                # The maximum number of additional bytes is 127. Technically not all of those
                # are necessary unless the highest bit of the byte is also set. But let's keep
                # things simple :)
                nbBytes = min((b & 127), 127) 
                b >>= 7
                out.write(bytes([(nbBytes | 128)]))
                
                for _ in range(nbBytes):
                    out.write(bytes([b]))
            else:
                break
            
    
    try:
        appendToBuffer = out.buffer.write
    except AttributeError:
        appendToBuffer = out.write
        
    lzwCompressionTable, reverseLzwCompressionTable = get_lzw_compression_table()
    palette = get_palette()
    bufferOffset = 0
    runningBytes = []

    
    def flushBuffer():
        nonlocal bufferOffset
            
        if bufferOffset >= 2:
            
            # Buffer offset is measured in words (4 bytes). Since we're compressing with
            # words instead of bytes, we need to convert from words to bytes here.
            nWordsInBuffer = (bufferOffset // 2)
            compressedBytes = bytearray()

            # We'll work with paths of up to 127 bytes long each. Each path corresponds
            # to a sequence of pixels to copy inside the image. 
            # It starts with an indicator byte indicating how many of these paths exist
            # in the current run. We start with a zero indicator byte.
            out.write(bytes([0]))
                
            for i in range(nWordsInBuffer):
                    
                word = ((runningBytes[i*2] << 8) | runningBytes[i*2 + 1])
                            
                # For each complete pixel path read from the buffer...
                for j in range(word >> 10, (-(-word>>10))):
                        
                    #...find its matching prefix among the already computed ones 
                    # and append the corresponding suffix to our current run.
                    suffix = j ^ (word & 0x3ff)

                    # find the length of the prefix and check if it's enough to fit it 
                    # in a single byte. If yes, emit it right away. Otherwise start the encoding
                    # procedure.
                    prefix = None
                    
                    for k in range(127, -1, -1):
                        possiblePrefix = lzwCompressionTable[k]
                        
                        if possiblePrefix[-1] == suffix:
                            prefix = possiblePrefix[:k]
                            break

                    if prefix is None:

                        # if there's no such prefix, store what we saw till now as the prefix
                        # for later reuse. That requires us to split the existing prefix into 
                        # chunks of 127 values, which we do by storing the remainder as a new
                        # prefix itself
                        prefixLen = len(compressedBytes)//127 + 1
                        curIdx = len(lzwCompressionTable)
                        prefixRem = prefixLen - 1
                        compressedBytes.extend(prefixLen.to_bytes(1,'big'))
                    
                        for i in reversed(range(prefixLen)):
                            
                            chunk = prefix[(i*127) : (i+1)*127]
                            chunk.append(suffix)

                            if chunk[-1] == 0xFF:
                                chunk[-1] = curIdx
                                
                                lzwCompressionTable.append(tuple(chunk)) 
                                compressionEntry = (curIdx, (prefixRem, curIdx))
                                reverseLzwCompressionTable[compressionEntry] = compressionEntry
                                curIdx += 1
                                prefixRem -= 1
                                
                            else:
                                compressionEntry = (curIdx, (len(chunk)-1, curIdx))
                                reverseLzwCompressionTable[compressionEntry] = compressionEntry
                                lzwCompressionTable.append(tuple(chunk)) 
                                curIdx += 1
                        
                        # reset the counter and continue processing the remaining part of the input
                        # as usual
                        packedSuffix = (suffix << 10)
                        out.write(bytes([packedSuffix >> 8, packedSuffix & 0xff]))
                        break

                    else:

                        # Append the decoded prefix to the current run
                        out.write(prefix) 

                        # And make sure that the previous word was copied to the output before starting
                        # looking for another word
                        out.write(bytes([*(reversed(runningBytes[-2:]))]))
                        
                # After every iteration, move the data we've processed to the beginning of the array
                # to free some space
                del runningBytes[:2]
                
            
    appendToBuffer(struct.pack('<B', 0x21))   # Extension introducer
    appendToBuffer(struct.pack('<B', 0xf9))   # Gif Graphic Control Block
    appendToBuffer(struct.pack('<BB', 0x04, 0x04))    # Packed fields: user input disabled, transparent color
    appendToBuffer(struct.pack('<B', 10))      # Delay time = 100ms
    appendToBuffer(struct.pack('<B', 0))       # Transparent color
    appendToBuffer(struct.pack('<B', 0))       # End of block marker
    

    # Figure out how big each frame will be
    width = 2 ** (randrange(1, 4))
    height = 2 ** (randrange(1, 4)) 

    # Compute how many bits are required to encode the smallest value in the image
    smallestValue = min(width,height)
    bitsPerPixel = round(math.log2(smallestValue))

    # Write the screen descriptor byte
    appendToBuffer(
        struct.pack(
            '<BHHBBBHB', 
            0x2c,
            width,
            height,
            bitsPerPixel,
            0,     # Color resolution - not important
            0,     # Sort flag - not important
            0,     # Size of GCT
            0      # Block terminator
        )
    )
    

    # Also figure out how many tiles per line we should have
    tileWidthBits = randrange(bitsPerPixel+1, 8) 
    # Figure out the layout of tiles in the output image
    # There might be unused tiles after the row ends if we're dealing with odd widths
    # So, we choose a probability of having zero trailing tiles based on the value of the
    # number of trailing tiles versus the total number of pixels in the line
    trailingTilesProbability = float(((tileWidthBits - (width%tileWidthBits))+1)/(tileWidthBits+1))
    if trailingTilesProbability <.5:
        trailingTilesProbability *= -.5
        trailingTilesProbability += 1
    else:
        trailingTilesProbability = -(trailingTilesProbability-.5)**2
    numTrailingTiles = min(round(trailingTilesProbability*tile