import struct


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # First create a new empty GIF file and write its contents to `out`,
    # we'll overwrite this later with our own headers

    # The first byte is always "G", the next three bytes are always "IFF"
    # (Interchange File Format) followed by the version of the format
    # in ASCII ("87a" or "89a"). This gives us a total of five bytes for
    # the initial GIF header
    gif_header = b'GIF89a'

    # Next two bytes designate how many bytes there will be after the
    # GIF header but before any image data. We specified that we want one
    # kilobyte for now so it's as simple as reading those two bytes here
    # and storing them in memory
    framelen = struct.unpack('<H', rng.read(2))[0] & 0xfff

    # Now we can put together our complete header including all these fields
    gif_file = gif_header + struct.pack(
            '<HHB',
            0x000b,              # width
            0x000c,              # height
            0xf6,                # global color table flag, sort flag
            ) + bytearray([
                0xff,             # background color index
                0x0b,             # pixel aspect ratio numerator
                0x07,             # pixel aspect ratio denominator
                ])                  # flags for left-to-right sorting, etc

    # Our `framelen` is the length of the actual image bits, not counting
    # the header. Thus we need to figure out where in the file the actual
    # image starts and then calculate the proper value for the starting
    # offset
    start_offset = len(gif_file) - 1

    # Now we can actually add some more information, like the palette size
    # and some pixels!
    gif_file += struct.pack('<BBBBBBBBB',
        0x5d,         # header for local color table
        0x00, 0xe6,   # palette size
        0xb8,         # rgb triplet for black
        0xa5,
        0xcf,
        0x44,
        0xda,         # rgb triplet for white
        0xc0,
        0xd7,
        0x5e          # rgb triplet for red
        )

    # Finally, we can fill up the rest of the image space with an appropriate
    # amount of random bytes. In this case however I'm just going to set it
    # to zeroes since it doesn't really matter what they're filled with as long
    # as they're valid pixels
    gif_file += rng.read(start_offset + framelen).ljust(framelen, b'\x00')

    # And finally, write our completed header to the output stream
    # (and hopefully close it!)
    out.write(gif_file)