# Generates a random GIF file. Randomness is provided by `/dev/urandom`. This script was used to create
# https://github.com/phildini/logjam-honeypot/blob/master/test/static/random-gifs/64x64.gif

import os
import sys
import time


def main():
    with open('/dev/urandom', mode='rb') as urandom:
        for i in range(int(sys.argv[1])):
            ensure_dir()

            filename = f'{i}.gif'
            print('Generating', filename)
            with open(filename, 'wb') as output:
                # TODO: use a better RNG?
                generate_random_gif(urandom, output)

                # Wait before generating next gif (in case we are being rate limited)
                time.sleep(1)


def ensure_dir():
    directory = './random-gifs/'
    if not os.path.exists(directory):
        os.makedirs(directory)


def generate_random_gif(rng, outfile):
    header = b'GIF89a'
    datalen = int.from_bytes(rng.read(2), byteorder='big')
    data = rng.read(datalen)
    framecount = int.from_bytes(rng.read(1), byteorder='big')
    palette_size = pow(2, framecount+1)
    background = rng.read(1)[0] % palette_size
    aspect_ratio = rng.read(1)[0] & 7
    has_trailer = False
    loops = 1

    # Remaining bytes of gif spec: http://giflib.sourceforge.net/whatsinagif/index.html
    flags = b'\x00\x00\x51\xF9\x04'    # 11111111 00000000 01011001 11111001
    color_resolution = b'\x7C'           # binary: 01111100
    sort_flag = b'\x00'                  # binary: 00000000
    global_color_table = b'\xFF\xFE'     # binary: 11111111 11111110
    table_size = (palette_size * 3 - 1).to_bytes(byteorder="big", length=1)   # 0 to 255 inclusive

    trailer = b''
    if has_trailer:
        # 0000TTTT SSRR
        delaytime = rng.read(2)         # 0 to 65535 milliseconds
        disposalmethod = rng.read(1)    # 0 to 3, see below
        userinput = rng.read(1)         # Not really sure what this means

        trailer += delaytime + disposalmethod + userinput


    outfile.write(bytes([len(header)]))       # LSD
    outfile.write(header)                     # Header string
    outfile.write(datalen.to_bytes(length=2, byteorder='big'))            # File size in bytes
    outfile.write((framecount).to_bytes(length=1, byteorder='big'))      # Frame count
    outfile.write(flags)                      # Flags
    outfile.write(background.to_bytes(length=1, byteorder='big'))       # Background index
    outfile.write(aspect_ratio.to_bytes(length=1, byteorder='big'))     # Aspect ratio
    outfile.write(width.to_bytes(length=1, byteorder='big'))             # Width
    outfile.write(height.to_bytes(length=1, byteorder='big'))            # Height
    outfile.write(table_size)                 # Global Color Table Size
    outfile.write(global_color_table)         # Global Color Table
    outfile.write(trailer)                    # Trailer


if __name__ == "__main__":
    main()