#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 2 bytes of length of block (N).
#   3. N bytes of block.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Write GIF header. We always use version 89a because it is easy to parse.
    out.write(b"GIF89a")

    # First we write out the LZW minimum size. This value isn't strictly
    # required for correctness but can help with performance.
    lzw_minimum_size = rng.read(1)[0] % 16
    if lzw_minimum_size == 0:
        # Use a non-zero default value so that decoders don't get confused by an
        # empty table and will instead fall back to their own heuristic.
        lzw_minimum_size = 1
    out.write((lzw_minimum_size).to_bytes(1, 'big'))

    # Next we write the logical screen descriptor. The spec says that all fields are big endian.
    width = rng.read(2)
    height = rng.read(2)
    flags = rng.read(1)
    background_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)
    out.write(width + height + flags + background_color_index + pixel_aspect_ratio)

    # Now we write out the global color table. Color tables are optional in the spec but there
    # isn't a good way to omit one when generating random files. So this code writes one even when
    # the user didn't ask for one.
    num_global_colors = ((flags[0] >> 7) & 1) | (((flags[0] << 1) & 0x03) << 1)
    global_color_table = [rng.read(3) for i in range(num_global_colors)]
    colors = sorted(global_color_table)
    for c in colors[:-1]:
        out.write(c)
    out.write(int.to_bytes(len(colors), 1, byteorder="big"))

    # Finally we write out the image descriptor which contains dimensions as well as other metadata
    # such as local color table usage. Again, since the spec explicitly allows them to be omitted lets
    # just always emit these values.
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    local_color_table_flag = rng.read(1)
    interlace_flag = rng.read(1)
    sort_flag = rng.read(1)
    reserved_bits = rng.read(1)
    local_color_table_size = rng.read(1)

    def write_block(s):
        s = bytearray(s)
        datasize = len(s)

        # Markers take up two bytes at the start of blocks. They consist of two parts:
        #   * 1 bit indicating whether or not the block should be compressed.
        #       In practice this is ignored by most software.
        #   * 2 bits indicating what kind of compression algorithm was used.
        #       Here we say that the block was uncompressed.
        marker = 0b00
        out.write(marker.to_bytes(1,'big'))

        # Next comes the size of the data that follows the marker. We emit this field
        # without any transformations.
        out.write(datasize.to_bytes(1,'big'))

        # Next comes the actual data itself.
        out.write(s)

    # The simplest valid frame consists of a single null byte.
    write_block("")