#!/usr/bin/env python3

import io
import random
import os
import sys
import argparse

import struct


class RNG:
  def __init__(self, seed):
    self._rng = random.Random()
    self._rng.seed(seed)

  @property
  def rng(self):
    return self._rng
  
  def read(self, n):
    buf = []
    i = 0
    
    while i < n:
      val = self._rng.randint(0, 255)
      buf.append(val)
      i += 1
    
    return bytes(buf)
    

DEFAULT_SEED = '42'


def main():
    parser = argparse.ArgumentParser(description='Generate some randomly generated GIF files')
    parser.add_argument('-o', dest='outdir', default='./out/', required=False, help="Directory where output should be placed")
    parser.add_argument('-n', '--number', type=int, default=1, help="How many files do you want?")
    parser.add_argument('--maxframes', type=int, default=-1, help="The maximum numbers of frames per image (-1 means unlimited)")
    parser.add_argument('--width', type=int, default=640, help="Width of each frame")
    parser.add_argument('--height', type=int, default=480, help="Height of each frame")
    parser.add_argument('--fps', type=float, default=8., help="FPS rate at which the video should play")
    parser.add_argument('--seed', default=DEFAULT_SEED, help="Seed to use when generating the different frames")


    args = parser.parse_args()

    # Make sure the directory exists
    try:
        os.makedirs(os.path.abspath(args.outdir), exist_ok=True)
    except Exception as e:
        print("Error creating output dir: {}".format(e), file=sys.stderr)
        exit(-1)

    # Open the RNG with given seed
    rng = RNG(args.seed)

    # Generate N number of images
    for idx in range(args.number):
        filename = "random{}.gif".format(idx)
        filepath = "{}/{}".format(args.outdir, filename)
        
        f = open(filepath, mode='wb+')
        write_random_image(rng, f, args.maxframes, args.width, args.height, args.fps)
        f.flush()
        f.close

def write_random_image(rng:RNG, f:BinaryIO, maxframes: int, width: int, height: int, fps: float)->None:
    # Write the header now
    f.write(b'GIF89a')

    # Write Logical Screen Descriptor
    lsd = io.BytesIO()
    lsd.write(b'\x00\x00\x00')         # Width
    lsd.write(b'\x00\x00\x00')         # Height
    lsd.write(b'\x07\x00')              # Bits Per Pixel
    lsd.write(b'\x80')                  # Background Color Index
    lsd.write(b'\xFF')                  # Pixel Aspect Ratio
    lsd.write(b'\x00')                  # Global Color Table Flag
    # Size of Global Color Table, must be power of two between [2..7]
    lsd.write(b'\x05')
    f.write(lsd.getvalue())

    # The whole image will be stored in its own unique color palette
    # Create it!
    globalPalette = create_palette(rng)

    # Now write the created palette
    f.write(globalPalette)

    currentFrameIdx = 0
    while maxframes==-1 or currentFrameIdx<maxframes: 
        write_frame(f,rng,currentFrameIdx,width,height,globalPalette)
        currentFrameIdx+=1
    
def create_palette(rng: RNG) -> bytes: 
    # Define a list of all possible RGB values
    rgbvalues = [[i, j, k] for i in range(256) for j in range(256) for k in range(256)] 
    # Shuffle them shuffling around their positions at every iteration
    rng.rng.shuffle(rgbvalues) 

    # Convert it to packed binary
    result = struct.pack(str(len(rgbvalues))+'B B B',*[v[0] for v in rgbvalues])
    result += struct.pack(str(len(rgbvalues))+'B B B',*[v[1] for v in rgbvalues])
    result += struct.pack(str(len(rgbvalues))+'B B B',*[v[2] for v in rgbvalues]) 
    
    return result 

def write_frame(f:BinaryIO, rng:RNG, index: int, width: int, height: int, globalPalette: bytes):
    # Reset the frame buffer
    fb = io.BytesIO()

    # Write the control block
    ctrlblk = io.BytesIO()
    ctrlblk.write(b'\x2C')             # Extension Introducer
    ctrlblk.write(b'\x00')             # Frame Label
    ctrlblk.write(b'\x00\x00\x00')     # Left Offset
    ctrlblk.write(b'\x00\x00\x00')     # Top Offset
    ctrlblk.write(struct.pack('<HH', width, height))          # Width x Height
    ctrlblk.write(b'\x00')             # Packed Field
    ctrlblk.write(b'\x00\x00\x00\x00') # No Local Palette

    # Write it to the frame buffer
    fb.write(ctrlblk.getvalue())

    # Calculate how much space we've got left in the frame, including the palette
    availablePixels = width*height*3 - len(globalPalette)

    # How many pixels left after writing the colors of the newly added pixel?
    # This will tell us what to write afterwards
    totalAvailablePixelsAfterThis = availablePixels - 3
    
    # Write LOCAL COLOR TABLE first:    
    # How many colors can we fit inside the remaining space in the image?
     
    # First calculate how big our local palette would be and how many entries we'll need
    numColorsInPalette = min(availablePixels // 3, 2**(ord(globalPalette[-2]) & 0x07)+1)

    # Keep track of how many more pixels we can write later
    numPixelsLeftToWriteAfterLocalPalette = totalAvailablePixelsAfterThis - numColorsInPalette*3
    
    # Write the SPLIT POINT
    splitpoint = numPixelsLeftToWriteAfterLocalPalette
    if numPixelsLeftToWriteAfterLocalPalette <= 0:   # Overflow, no SPLIT POINT
      splitpoint=0
    elif numPixelsLeftToWriteAfterLocalPalette >= 255: # Underflow, full SPLIT POINT
      splitpoint=255
    else:                                         # Regular case, somewhere in between
      pass
    fb.write(splitpoint.to_bytes(1,'big'))

    # Write the GLOBAL COLOR TABLE
    fb.write(globalPalette)
                          
    # Placeholder for Image Data Block
    imagedatablkpos = fb.tell()
    fb.write(b'\x00'*numPixelsLeftToWriteAfterLocalPalette)

    # Write END OF IMAGE block
    endimgblock = io.BytesIO()
    endimgblock.write(b'\x3B')       # Trailer Marker
    endimgblock.write(b'\x00')       # Filler Byte
    fb.write(endimgblock.getvalue()) 

    # Write graphics control extension
    graphicscontrolextension = io.BytesIO()
    graphicscontrolextension.write(b'\x21')           # Extension Introducer
    graphicscontrolextension.write(b'\xF9')           # Graphics Control Extension Label
    graphicscontrolextension.write(b'\x04')           # Block Size
    # Disposal method => Restore to background color (see http://www.matthewflickinger.com/lab/whatsinagif/bits_and_bytes.asp )
    graphicscontrolextension.write(b'\x08')  
    graphicscontrolextension.write(index.to_bytes(2,'little'))      # Number of loops, set to 0 for infinite looping
    graphicscontrolextension.write(b'\x00')           # Trailer (filler byte)
 
    # Write Plain Text Extension
    plainTextExtensionData = io.BytesIO()
    plainTextExtensionData.write(b'\x01\xFE\x00\x0BNETSCAPE2.0\x03\x01\x00\x00\x00')
    plainTextExtensionData.write(fps.to_bytes(2,"little"))        # FPS
    plainTextExtensionData.write(b'\x00\x00\x00')                 # Trailer (filler bytes)

    # Write these extensions right before the image data
    f.seek(fb.tell())
    f.write(graphicscontrolextension.getvalue())
    f.write(plainTextExtensionData.getvalue())

    f.seek(imagedatablk