#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
COLORSPACE = b'\x2c\x00'        # Color table subblock identifier
COLORLEN   = b'\xf1'             # Number of colors to allocate
                                # (not used in real gifs)
LOOPOFFSET = b'\xff\xff'         # Loop offset, big endian
LOOPS      = b'\x00\xc0'          # Number of times to loop
TRANSP     = b'\x00'              # Background color index
IMAGEDESC  = COLORSPACE + COLORLEN + LOOPOFFSET + LOOPS \
             + TRANSP                        # Image descriptor label
ENDL       = b'\x3b'               # File terminator

# Returns a string describing both sides of the equality check
# for the format check, e.g.:
#   $FF => not equal to $00
#    $F0 == $F0
def describe_block(block: str):
    if len(block)>1: return f'${block[0]:02X} == ${block[1]:02X}'
    else:           return f'$FF!= ${block[0]:02X}'




# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # Generate the color table
    colormasks = [int(rng.read(1)[0] / 8) for _ in range(7)]
    colsizes = [1 << i for i in reversed(range(7))]

    palette = []
    for sz in colsizes:
        palette += [(int(rng.read(sz)[0]),
                     int(rng.read(sz)[0]),
                     int(rng.read(sz)[0]))]

    sizestr = ''.join([f'{colmasks},{sz}'.encode('utf-8')
                       for colmasks, sz in zip(colormasks, colsizes)])
    colorstr = b''.join([struct.pack('<BBB', c[0], c[1], c[2])
                         for c in palette])

    # Now let's append some garbage to make sure everything runs smoothly
    # Otherwise we'd run out of bytes from the RNG when writing the last
    # descriptor or the global color table.
    unpadded = out.tell() % 256 
    print(f'Unpadded bytes required: {unpadded}')

    # Write the image descriptor
    descriptor = IMAGEDESC + sizestr + colorstr
    out.write(descriptor)

    # Pad up until the next even number of bytes
    while bool((out.tell()+1) & 0x01):
        out.write(rng.read(1))

    # Now we're going to write 256 descriptors at once
    # Using this loop means we avoid making multiple calls to read()
    for _ in range(257):
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

        # TODO: Should experiment with variations on these values too
        #       see https://github.com/decompme/decompilation-guide/issues/47
        newdesc = struct.pack("<BB", 0x2C, 0x00)
        newdesc += rng.read(8)
        out.write(newdesc)
        print(describe_block(rng.read(2)), end=', ', flush=True)

    # Write the last descriptor!
    out.write(b'\x3B')
    print('\nThe generated gif is '
          f'{out.tell()} bytes long ({datalen} expected)')

    