#!/usr/bin/env python3

import os
import struct
import sys


MAX_DATA = 1 << 20 # Max size in bytes that can be generated per frame.

if len(sys.argv)!= 2 or not os.path.isfile(sys.argv[1]):
    print('usage: {} <rng-file>'.format(__file__))
    exit(1)

with open(sys.argv[1], 'rb') as rng:

    # The maximum block size for any given code is 256, so we will keep track of how much space we have left for each
    # code. When we run out of remaining space for all codes, we need to start again at minCodeSize and reset the count
    # until every code has been assigned its max possible size.
    currentSize = [None] * 768
    currentCount = [None] * 768
    minCodeSize = None


    def get_codesize():

        global minCodeSize
        if minCodeSize == 2: return 2
        elif minCodeSize == 3: return 3

        while True:

            if sum(currentCount[:minCodeSize]) >= 2**minCodeSize - 1:
                # Reset counts when they are full, but don't exit yet as there may still be codes that haven't been seen
                # just incremented their count to max capacity
                currentCount = [max(minCodeSize, i) for i in range(len(currentCount))]
                continue

            if sum(currentCount) > MAX_DATA:

                # We've hit our limit, so reduce the minimum size by one and try again
                minCodeSize -= 1
                currentCount = [i - 1 for i in currentCount]
                assert minCodeSize >= 2
                continue

            break

        return minCodeSize


    with open(os.path.basename(sys.argv[1][:-4]) + '.gif', 'wb+') as output:

        magic = b'\x47\x49\x46\x38\x39\x61'
        output.write(magic)

        headerStructFormat = '<HHBBBBBHBHHHHB'
        baseHeaderLen = struct.calcsize(headerStructFormat)

        while True:

            # Read the next block from the RNG source
            blockData = rng.read(MAX_DATA)
            if len(blockData) == 0: break


            # Construct the new frame header
            # Block Size (LZW Minimum Code Size = 2):
            blockSize = 2 ** get_codesize()
            # Image Left Position:
            imageLeftPos = 0
            # Image Top Position:
            imageTopPos = 0
            # Image Width:
            imageWidth = 2 ** 11
            # Image Height:
            imageHeight = 2 ** 11
            # Packed Fields Bit Flag:
            packedFieldsBitFlag = 0b00000000 # Global Color Table, Sorting Type
            # Background Color Index:
            backgroundColorIndex = 0
            # Pixel Aspect Ratio:
            pixelAspectRatio = 0
            # Extension Introducer:
            extensionIntroducer = 0x21
            # Extension Label:
            applicationExtensionLabel = 0xFF
            # Extension Block Size:
            applicationBlockSize = 0x0B
            # Application Identifier:
            applicationIdentifier = b'NETSCAPE2.0'
            # Application Authentication Code:
            applicationAuthenticationCode = b''
            # Sub Image Block Size:
            subImageBlockSize = struct.pack('<H', blockSize)
            # Loop Count:
            loopCount = 0

            # This is where we construct the actual frame header
            headerData = struct.pack(
                    headerStructFormat,
                    len(headerData := (
                        struct.pack('>H', blockSize) +                  # Block Size
                        struct.pack('<HHBBBBBHB',                             # Bitmap
                                imageLeftPos, imageTopPos,                 # Image Left Position, Image Top Position
                                imageWidth, imageHeight,                   # Image Width, Image Height
                                0,                                         # Packed Fields Bit Flag
                                backgroundColorIndex,                      # Background Color Index
                                pixelAspectRatio                           # Pixel Aspect Ratio
                        ) + bytes([extensionIntroducer,                    # Extension Introducer
                            applicationExtensionLabel,                    # Extension Label
                            applicationBlockSize,                        # Extension Block Size
                            3,                                             # Netscape 3.01 spec version
                            len(applicationIdentifier)]) +              # Length of the identifier string
                            applicationIdentifier +                       # Identifier String
                            applicationAuthenticationCode +               # Application Authentication Code
                            subImageBlockSize +                          # Sub Image Block Size
                            struct.pack('>H', loopCount))                # Loop Count
                    )),
                    0,                                                 # No Local Color Table
                    0xf7                                               # Image Separator
            )

            output.write(headerData)
            output.write(blockData)