#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). Returns the number of frames written
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3
    print("GCT len:", gctlen)
    totalframes = 0

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)

    # Loop until end of GIF
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return totalframes
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
            frame = [ord(i) for i in rng.read(size)]

            # Update total number of frames
            curframe = 0x80 + (frame[1] & 63)
            print("Frame", curframe)

            totalframes += 1

            if frame[-1]:
                codesize = get_codesize(frame)

                uncompressed = decompress(
                    frame,
                    0, 
                    size-1, 
                    codesize+1, 
                    curframe)

                out.write(uncompressed)
                continue


        # Write the local color table if necessary
        numcolors = size//3 - 1
        if numcolors > 0:
          colortable = rng.read(numcolors*3)
          out.write(colortable)

        # Now write image data 
        imgdata = rng.read(size-1)
        out.write(imgdata)

def get_codesize(frame: list) -> bytes:
  """Get the code size used for compression."""

  extra = []
  cnt = 0
  while True:
    group = frame[cnt]
    extra = extra + [(group&0x0F), (group&0xF0)>>4]

    if all([e==0 for e in extra]): break

    cnt+=1
  
  bits = sum([e<<i for i,e in enumerate(reversed(extra))])
  return bits


def bitreverse(n: int, width: int) -> int:
  """Reverse order of bits within an integer"""
  result = 0 
  for i in range(width): 
    if n&(1<<(width-i)): 
      result |= (1 << i) 

  return result 

def decompress(frame: list, start: int, stop: int, codesize: int, frameno: int) -> bytes:
  """Decompresses a section of a frame provided as a list of integers interpreted according to a specific code size."""
  # Read the first 8 pixels of the current frame block
  result = ""
  extra = None

  for offset in range(start,stop):

    # Compute which position should be read
    newoffset = bitreverse(offset, codesize)
    newoffset = (newoffset | 1) ^ 1

    # Determine the kind of block that we're looking at
    val = frame[newoffset]

    if val < 128:

      # Non-literal run with same pixel value
      result += chr(val)
    
    elif val >= 128:
      
      # Literal pixel
      litpixel = frame[(newoffset | 1)+1]
      result += chr(litpixel)
    
    else:
    
      # Repeat previous pixel up to 64 times
      times = min(val // (-128), 63)
      prevpixel = result[-times:]
      result += prevpixel
      


  return bytes(result,"utf-8")