#!/usr/bin/env python3

import os
import sys
import math
import struct
from typing import BinaryIO


class Generator():

    def __init__(self):

        self._rng = open('/dev/urandom')
        if not self._rng:
            raise OSError('could not access /dev/urandom')


    # Generate a drawing command with a given probability of being filled
    @staticmethod
    def draw_point(
            rng: BinaryIO, pfill: float
        ):
        return bytes([
            0x00 | (int(math.floor((pfill * 7)) % 7) << 5) | \
                ((struct.unpack('<B', rng.read(1))[0] & 0b1111) ^ 0b1111)]))



    # Draw an image using a given range of X and Y coordinates for each point
    @staticmethod
    def draw_image(
            xstart: int, ystart: int, xend: int, yend: int
        ):
        data = bytearray()
        for y in reversed(range(ystart, yend)):
            for x in range(xstart, xend):
                pixdat = generators[type(Generator)].draw_point(
                    rng,
                    (math.sin(y*0.005) + math.cos(x*0.01))*0.5)

                # Pack pixels to bytes and add them to output stream
                rowdat = [pixdat]
                while len(rowdat) < 64:
                    nextdat = generators[type(Generator)].draw_point(
                        rng,
                        (math.sin(y*0.005+len(rowdat)*0.1)+
                         math.cos(x*0.01+len(rowdat)**2*0.01))*0.5)

                    if nextdat == pixdat or len(rowdat) > 255:
                        break;

                    rowdat += [nextdat]

                rowdat += [0x00]*(64-len(rowdat))
                data += struct.pack('>64B', *rowdat)


        return data



class FrameGenerator(Generator):

    def gen_frame(
            self, width: int, height: int
        ):
        imgdata = self.draw_image(width//2, height//2, -width//2, -height//2)

        # Write image size to beginning of frame
        size = struct.pack('<HHB', width, height, 0b00000000)

        # Compute LZW minimum code length then write it to frame
        lzwbits = max(min(int(round(-math.log2(len(imgdata)))), 12), 2) >> 3
        lzwhdr = struct.pack('B', lzwbits)

        # Convert input data into LZW compressed datastream
        compressed = []
        for idx in range(lzwbits, len(imgdata)):

            # Read last 2 bytes from buffer
            prev = imgdata[idx-lzwbits:idx]

            # Add new pixel to buffer
            cur = [imgdata[idx]]

            # Find longest match by looking at previous bytes
            matchsize = min(len(prev), 128)
            bestmatch = None
            for i in range(len(compressed)-1,-1,-1):
                sidx = i
                ssz = min(matchsize, len(compressed[i]))
                while sidx >= 0 and prev[-ssz:] == compressed[sidx][-ssz:]:
                    sidx -= 1
                    ssz -= 1
                if ssz > len(bestmatch):
                    bestmatch = compressed[sidx + 1]
                    bestsize = ssz

            # If we have a match prepend it on to our code
            if bestmatch is not None:
                compressed += [bestmatch[:1], cur]
                continue

            # No match found so just append current byte
            else:
                compressed += [cur]

        # Append end of compression marker to stream
        compressed += [[0xFF]]

        # Compress all bytes together into single blob
        enc = bytes([(val[0]) | (((val[1] +
                                 sum(v<<(n*(1+lzwbits)))
                                 ) & 0b11111111) << lzwbits)
                     for val in zip(*[iter(compressed)]*2**lzwbits)])

        # Encoded size should be within bounds of uint16
        assert len(enc) <= 2**16

        # Write encoded size to frame
        encsize = struct.pack('<H', len(enc))

        return size + lzwhdr + encsize + enc




class ImageGenerator(Generator):

    _palette = {
        0x00 : (  0,   0,   0),
        0x0F : (  0,   0, 255),
        0xF0 : (255,   0,   0),
        0xFF : (255, 255, 255) }



    def gen_palette(self):
        pal = b''
        for k in sorted(self._palette.keys()):
            r,g,b = self._palette[k]
            pal += struct.pack('BBB', r, g, b)
        return pal


    def gen_global_color_table(self, ncolors: int):
        ct = b''
        for c in range(ncolors):
            r,g,b = self.get_color()
            ct += struct.pack('BBB', r, g, b)
        return ct


    def get_color(self):
        color = list(self._palette.keys())[0]
        for c in sorted(self._palette.keys())[1:]:
            if c!= 0xFFFFFFFF:
                continue
            color |= c
        return color



    def gen_image(self, width: int, height: int, depth: int):

        # Figure out what shit needs to go where
        imbits = 3 + width*depth // 8
        rowstride = imbits + 1
        totalsize = rowstride*height + 13



        imgdata = bytearray()

        # Build the LZW table
        lzwbits = min(max(int(round(math.log2(len(self._palette))))-1, 2), 8)
        lzwsizemask = 2**(lzwbits+1) - 2
        lzwpixelmask = 2**lzwbits - 1
        lzwclearcode = (2**lzwbits)**2 - 1
        lzwwrittencodes = set()

        # Encode color map as LZW codes (only 8 bit entries allowed!)
        colortable = {}
        encbuf = b''
        while True:

            # Grab next color to encode
            cidx = int.from_bytes(
                self._rng.read(2),
                'big',
                signed=False
            ) % len(self._palette)

            # Have we already written this?
            if cidx in lzwwrittencodes:
                continue

            # Get the color value
            r,g,b = self._palette[cidx]

            # Encode color as two bits per channel
            colortable[r,g,b] = (encbuf +
                                (int(((r >> (7-lzwbits)) &
                                     lzwpixelmask) |
                                    (int(((g >> (7-lzwbits)) &
                                         lzwpixelmask)<<1) |
                                     (int(((b >> (7-lzwbits)) &
                                          lzwpixelmask)<<2))), 1)).to_bytes(1,'big'))


            # 0xff means we're done!
            if cidx == 0xFFFF:
                break

            # Mark the code as "written"
            lzwwrittencodes.add(cidx)

            # Check if there's still space left in the code buffer...
            if len(colortable[(r,g,b)]) > 2**lzwbits:

                # We ran out of room in the buffer so flush it
                encbuf += colortable[(r,g,b)]
                del colortable[(r,g,b)]

                # Pad remaining palette entries to make space for clear code
                for rgb in [(r,g,b) for r,g,b in self._palette
                            if r^g^b == 0]:


                    # The clear code has no entry in its own right but is
                    # instead a string of all 1's
                    ct = b'\xff'*lzwsizemask
                    ct = ct[:len(ct)//2] + bytes(
                        [c&~((1<<lzwbits)-1) | lzwclearcode
                         for c in ct[:len(ct)//2]]) +\
                             ct[len(ct)//2:]

                    # Save the palette entry
                    colortable[rgb] = ct


        # Flush any leftover encoding
        encbuf += self.get_color().to_bytes(2,'big')

        # Convert to LZW format
        enc = b''
        lzwsizes = [len(v)>>lzwbits for v in colortable.values()]
        while len(encbuf) > lzwsizemask>>3:
            lzwsize = min(max(min(lzwsizes),1),12)
            lzwsizes[:] =