#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, outstream: BinaryIO
    ) -> None:
    # Write magic bytes
    outstream.write(b'GIF89a')

    # Parse the global screen descriptor
    lsd_bits = bin(ord(rng.read(1)[0]))[2:] + '000000000000000'[::-1]
    lsd = [ int(lsd_bits[i : i+8], base=2)
            for i in range(0, len(lsd_bits), 8)] 
    width, height, gclrlo, gclrlc = \
         lsd[0], lsd[1], lsd[3]+8**5, min(max(lsd[4]+1, 12), 256)-1
    gclr = bytes([ v for _ in range(gclrlo+1)
                   for v in list(rng.read(3))[:gclrlc] ])

    # Write the global screen descriptor
    outstream.write(bytes([
        lsd[0],           # Width
        lsd[1],           # Height
        lsd[2],           # Packed field; Global Color Table
                          # Size 1-8 bits; sort flag (1);
                          # Color resolution 1-8 bits;
                          # GCF Flag (1); Sorting method (3)
                          
        lsd[3]&0x07,      # Background color index
        lsd[3]&0xf8,      # Pixel aspect ratio
        lsd[4]            # Global Color Table length
    ]))
    outstream.write(gclr)

    # Create buffer for color indexes
    buff = []
    colors = set()

    # Number of frames is 2**GCF
    nframes = 2 ** max(min(height // 2, 8), 1)

    # We save a frame every time we get a new color
    def flush():
        nonlocal buff, colors
        if not buff:
            return
        colors |= { k for sublist in buff
                    for k in sublist }
        ncols = len(colors)
        rawframe = b''
        for row in buff:
            rawframe += bytes([(ncols-len(row))*b'\xff']
                             + sorted(row)
                              + [ncols])
        ssz = len(rawframe).to_bytes(2, byteorder='big')
        lzssframe = b''.join([ssz]
                            + [(len(rawframe)+1)//256]
                            + [_ for _ in range(
                                0, len(rawframe), 256)])
        lzssframe += bytes([
            0xe0|((len(lzssframe)>>8)&0x1f),   # Control bit
            (len(lzssframe)&0xff)^0xff        # Checksum
        ])
        lzssframe += rawframe
        sz = len(lzssframe).to_bytes(2, byteorder='big')
        outstream.write(b''.join([sz, lzssframe]))

    for _ in range(nframes):
        for y in range(height):
            for x in range(width):
                buff.append({ord(rng.read(1)[0]):None})
        flush()
        del buff[:]

    # Last frame is empty so we don't need to write anything
    flush()