#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`. The logical screen descriptor
# provides the width and height of the generated image. The content
# produced by the generator may be either static or animated; setting
# repeat to true will cause the generator to produce an animation where
# each frame is [framecount] frames long.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        w: int, h: int, repeat: bool = False,
        framecount: int = None
    ):
    
    # Calculate image area
    imarea = w*h

    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Handle repeat and framecount values
    if repeat and framecount is None:
        raise RuntimeError("repeat must also specify framecount")
        
    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7) + bytes([(w&255), (w>>8), (h&255), (h>>8), 0])

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))+1) * 3

    # Add GCT to output if we have one
    if gctlen > 0:
        gct = rng.read(gctlen)

        # If we don't get an even number of colors then we need padding to make sure
        # all resulting images contain the same amount of bytes
        outbytesleft = 256 - len(gct)
        if outbytesleft % 3!= 0:
            out.write(b'\x00' * (outbytesleft % 3))

        out.write(gct)

    # Loop through frames until we've exhausted the RNG source 
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))

        # Check if this is the last block
        if ((fieldbyte & 128) == 0):

            # There are two ways to terminate animations:
            # 1) If we were asked to loop endlessly then simply write another 
            #    extension block containing the application extension as per spec
            # 2) If we were given a specific frame count then we set the T flag on
            #     the final graphic control block
            if repeat:
                print("Writing loop terminator", file=sys.stderr)

                out.write(b"\x21\xFF\x0B")
                out.write(b"NETSCAPE2.0")
                out.write(b"\x03\x01")
                out.write(b"\x00\x00")
                
            elif framecount is not None:
                
                print("Writing framecount termination code", file=sys.stderr)

                # Seek backwards to overwrite the frame count value 
                f.seek(-1, 1) 

                # Overwrite the flag byte with the new state
                flags = ord(f.read(1)) | 128
                out.write(bytes([flags]))
        
            else:

                # Warn about unterminated animations
                print("WARNING: Generated GIF contains unterminated animation!", 
                      "It should be terminated properly but you should verify it.",
                      file=sys.stderr)

            break
            
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte >> 5) & 3) > 0)

        # Write fields back to stream as they are
        out.write(struct.pack('=B', fieldbyte & ~2 | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
            continue

        # Write the local color table if necessary
        numcolors = size // 3 - 1
        if numcolors > 0:

            # We can only handle up to 256 colors so just fake that many and skip
            # over the rest
            if numcolors > 256:
                numcolors = 256

            colortable = rng.read(numcolors * 3)
            out.write(colortable)

            # Skip over the rest of the colors in our palette since we only support
            # 256
            leftovercolorbytes = max((numcolors - 256) * 3, 0)
            if leftovercolorbytes > 0:
                rng.seek(leftovercolorbytes, 1)

        # Write the variable part of graphics control block
        out.write(b'\x00')
        delaytime = rng.read(2)
        out.write(delaytime)
        transindex = rng.read(1) 
        out.write(transindex)
        
        # Write the actual image data
        imgdata = rng.read(size - 1)
        out.write(imgdata)

