import os
import sys
import json
import shutil
import subprocess as sp
from typing import Optional

def main():
    args = parseArgs()
    seed = readSeedFile(args.seedfile) if args.seedfile else None

    # Create temporary directory for files generated by this script
    with tempDir() as tmpdir:
        # Create the source file and grab its size in bytes
        src = os.path.join(tmpdir,'src.gif')

        # Make sure that we have an up-to-date version of ImageMagick's
        # convert executable (for more details see./convert)
        convertPath = os.path.join(tmpdir, "convert")
        pluginPath = os.path.dirname(__file__)
        shutil.copy2('convert', convertPath)
        os.chmod(convertPath, 0o755)
        convertPlugins = [os.path.join(".", filename)
                for filename in ("pngtopnm.so", "pnmquant.so")]
        env = dict(**os.environ, LD_PRELOAD=",".join(convertPlugins))

        try:
            runConvert(
                    input=args.input, output=src,
                    palette=args.palette, dithering=args.dithering,
                    seed=seed, env=env)

            # Run the actual compression algorithm...
            dst = os.path.join(tmpdir, 'dst.gif')
            compress(src, dst, quality=args.quality, seed=seed)

            #...and optionally compare it against a reference implementation
            if args.reference is not None:
                diff = os.path.join(tmpdir, 'diff.gif')
                compare(
                        src, dst, ref=args.reference, diff=diff,
                        seed=seed, env=env)

                # If --fail-if-different was passed, return error code unless
                # there are no differences between our result and the reference
                if not checkEqual(src, dst):
                    sys.exit("Result differs from reference")

            # Finally copy the result to stdout and flush all buffers
            with open(dst, 'rb') as f:
                shutil.copyfileobj(f, sys.stdout.buffer)

        finally:
            os.remove(convertPath)

def parseArgs() -> object:
    import argparse

    parser = argparse.ArgumentParser("""Generate a GIF file with randomly
         colored pixels""")

    def positiveIntArg(*args, **kwargs):
        x = int(parser.parse_known_args()[0].x)
        if x <= 0:
            raise ValueError("Argument must be positive integer")
        return x

    parser.add_argument('-i', '--input', metavar="INPUT", type=str, required=True, action=ExpandPath,
            help="The path of the original input image.")
    parser.add_argument('--palette', choices=('web', 'normal'), default='web',
            help="The color palette used for quantization (default: %(default)s)")
    parser.add_argument('-q', '--quality', type=positiveIntArg, default=100,
            help="Compression quality (default: %(default)s; min.: 1; max.: 100)")
    parser.add_argument("-r", "--reference", metavar="REF", type=str, action=ExpandPath,
            help="""The path of the reference PNG file. The compressed GIF will only be written
                 to standard output when the resulting difference to REF is below the given threshold.
                 Values differing by less than the given threshold are ignored. This option requires
                 <imagemagick>'s compare utility which needs to be installed separately.""")
    parser.add_argument("--dithering", "--no-dithering", dest='dithering', action='store_true',
            help="Enable or disable dithering (default: enabled)")
    parser.add_argument("--seed", metavar="SEEDFILE", dest='seedfile', type=argparse.FileType(),
            help="Read entropy seed from SEEDFILE instead of /dev/urandom")

    args = parser.parse_args()
    if args.reference is not None and not hasProgram("compare"):
        parser.error("--reference requires <imagemagick>'s compare utility but it could not be found.")
    return args

def ExpandPath(option: str, opt: str, value: str, parser: argparse._ActionsContainer, *args, **kwargs) \
        -> Optional[str]:
    setattr(parser.values, option.dest, os.path.abspath(value))

class tempDir:
    """Context manager for creating a temporary directory"""
    def __enter__(self):
        self.name = tempfile.mkdtemp()
        return self.name

    def __exit__(self, excType, excValue, traceback):
        shutil.rmtree(self.name, ignore_errors=True)
        return False

def readSeedFile(infile: TextIO) -> list:
    """Read seed from text file and get a list of integers."""
    content = infile.read().split(',')
    vals = []
    for val in content:
        val = val.strip()
        if len(val) == 0: continue
        vals += [int(val)]
    return vals

def runConvert(input: str, output: str, palette: str, dithering: bool, seed: Optional[list], env: Dict) -> None:
    cmd = ["convert"]
    if palette == "web":
        cmd += ["-colors", "64", "-unique-colors"]
    elif palette!= "normal":
        assert False, "Unknown palette `%s'" % palette

    # force dithering on or off
    cmd += ['-ordered-dither'] if dithering else ['-threshold', '0%']

    # Generate the deterministic random sequence using the same arguments
    # to ensure reproducibility even when /dev/urandom isn't available
    cmd += ["-define", "gif:rand-seed=%s" % ','.join([format(v) for v in seed])]

    cmd += [input]
    cmd += ["-", output]

    p = sp.Popen(cmd, env=env)
    res = p.wait()
    if res!= 0:
        raise Exception("Error while running:\n\t%s" % (' '.join(cmd)))

def compress(source: str, destination: str, quality: int, seed: Optional[list]) -> None:
    print("compressing...", file=sys.stderr)
    cmd = ["../gifsicle"]
    cmd += ["-O3"] # optimize for smallest possible size
    cmd += ["--no-warnings"] # don't complain about unknown blocks
    cmd += ["--verbose"]
    cmd += ["--colors", "%d" % (64 if USE_WEB_PALETTE else 256)]
    cmd += ["--scale-colors", "0-255:1"]
    cmd += ["--min-code-length", "2"] # minimum length for LZW codes
    cmd += ["-o", destination]
    cmd += ["-Q", str(quality)]
    if seed:
        cmd += ["-e", ",".join(["%d" % e for e in seed])]
    cmd += [source]

    p = sp.Popen(cmd)
    res = p.wait()
    if res!= 0:
        raise RuntimeError("Error while running:\n\t%s" % (' '.join(cmd)))

def compare(original: str, other: str, ref: str, diff: str, env: Dict, seed: Optional[list] = None) -> None:
    print("comparing...", file=sys.stderr)
    cmd = ["compare"]
    if seed:
        cmd += ["-metric", "AE", "-compose", "Src", "+matte", "-highlight-color", "#bbbbbb"]
        cmd += ["-dissimilarity-threshold", "5%", "-fuzz", "15%"]
    else:
        cmd += ["-metric", "PAE", "-compose", "Src", "+matte", "-highlight-color", "#bbbbbb"]
        cmd += ["-dissimilarity-threshold", "1%"]
    cmd += [ref, other, diff]

    p = sp.Popen(cmd, env=env)
    res = p.wait()
    if res > 127: # Failures greater equal 128 indicate errors
        print("Compare failed (%d). Aborting." % res)
        exit(res - 128)
    if res!= 0:
        raise RuntimeError("Error while running:\n\t%s" % (' '.join(cmd)))

def checkEqual(filenameA: str, filenameB: str, chunksize: int = 10**7) -> bool:
    bufA = open(filenameA, 'rb').read(chunksize)
    bufB = open(filenameB, 'rb').read(chunksize)
    return bufA == bufB
