#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Write the header; we won't worry about the extra
    # 7 bytes or any other parameters since they aren't needed
    # for triggering the vulnerability anyway.
    print('Writing the header', end='', flush=True)
    out.write(b'GIF89a\x00\x00\x00')
    print(' done.', flush=True)

    # Figure out how many bytes are left over from each chunk
    # and use them for the next one.
    def write_descriptor():
        nonlocal rng
        nonlocal out

        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >=10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)
        label_and_length = b'\x2C\x00\x00\x00\x00\x00\x00\xFB\x00\x00\x00'

        max_bytes = len(label_and_length)+10
        chunk_remaining = min((max_bytes-out.tell())//2, 256*5)


        if chunk_remaining <= 0: return False
        else:
            chunk_remaining -= 1   # Leave room for the \0 after the chunk

            print(f'Writing a descriptor ({chunk_remaining})', end='', flush=True)
            out.write(label_and_length+rng.read(chunk_remaining))
            out.write(b'\0')    # Always ends in null terminator
            print('.', flush=True)

            return True

    # Keep writing descriptors until something goes wrong
    while write_descriptor(): pass

    # Write the last descriptor!
    out.write(b'\x3B')

    # Now write the global color table
    # This is kinda silly because it has no effect on anything, but it might
    # still trigger some heuristics about what decompilers do when reading
    # invalid files?
    print('Generating global color table...', end='', flush=True)
    out.write(struct.pack('<BBB', *(rng.read(1)[0] % 256 for _ in range(3))))
    print('done!')

    # Add padding to reach eof
    expected_size = out.tell()
    print(f'Total size of generated gif should be {expected_size}')

    padding_len = (expected_size // 2**16) + 1
    padding = [rng.read(1) for _ in range(padding_len)][::-1]
    padding_len = ((expected_size % 2**16)*256+ord(padding[-1])).to_bytes(2, 'little')
    
    print(f"Writing padding section ({padding_len})", end='', flush=True)
    padding[-1:] += padding_len
    out.write(b''.join(padding))
    print('.')