#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    if rng is None or out is None:
        raise ValueError('Invalid arguments!')
    
    # Header / Version Bytes
    version = rng.read(4)
    out.write(version)

    # Logical Screen Descriptor
    width = rng.read(2)[::-1]
    height = rng.read(2)[::-1]
    packed = bytes([
        0xF7 | (int.from_bytes(width, 'little') > 0xFF), 
        *rng.read(2), 
        0x80 | ((int.from_bytes(height, 'little') >> 8) & 0x07), 
        *rng.read(3)])
    out.write(packed)

    # Global Color Table
    palette = [
            *(rng.read(3) for _ in range(
                0x1 << int.from_bytes(
                    packed[7], 'little')))]
    out.write(bytearray(palette))

    # Graphic Control Extension
    packed = bytearray([*rng.read(5)])
    packed[0] |= 0xF7
    packed[4] |= 0x04 
    out.write(packed)

    # Image Descriptor
    packed = [
        0xFE, 
        *rng.read(4), 
        0x80 | ((int.from_bytes(width, 'little') >> 8) & 0x07), 
        *rng.read(3)]
    out.write(bytes(packed))

    # Plain Text Extension
    text = bytearray()
    while True:
        char = rng.read(1)
        if ord(char) == 0: break
        else: text += bytearray(char)
    extlen = text.__len__() % 8
    text += b'\x00' * (8 - extlen)
    out.write((extlen + 3).to_bytes(1, 'big') +
            text)

    # Image Data
    datalen = int.from_bytes(
            rng.read(2), 'big')
    data = rng.read(datalen)
    padding = 0xFF & (-datalen)
    out.write(data + padding.to_bytes(padding, 'big'))

    # Trailer
    trailer = int.from_bytes(
            rng.read(1), 'big').__str__.zfill(3)
    out.write(trailer.encode())

