#!/usr/bin/env python3

import math
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Convert the field into a dictionary so we can easily access individual parts
    lsddict = { 'width': int.from_bytes(lsd[:2], "little"),
                'height': int.from_bytes(lsd[2:4], "little"),
                'flags': int.from_bytes(b'\x00'+lsd[-1:], "little"),
                'bgcoloridx': None if (lsddict['flags']&128)==0 else int.from_bytes(rng.read(1), "little") }

    print(lsddict)

    # Now convert dict to bytes
    lsd = struct.pack('<H', lsddict["width"]) + \
          struct.pack('<H', lsddict["height"]) + \
          struct.pack('<B', lsddict["flags"] & ~64) + \
          b'\xff\xff' # Background color idx

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
            continue

        # Iterate over all images to get their width/heights
        width = int.from_bytes(rng.read(1), "big")
        height = int.from_bytes(rng.read(1), "big")

        # Calculate number of colors needed in image
        numplanes = int.from_bytes(rng.read(1), "big")
        numcolors = max(numplanes+1, math.ceil(math.log2(width*height)))

        # Create a new palette by adding the existing one with extra entries if needed
        palette = [None]*numcolors
        for i in range(max(len(palette), len(lsddict["palette"]))): 
            try:
              palette[i] = lsddict["palette"][i]
            except IndexError: 
              pass
        
        # Construct an image
        image = [] 
        for c in range(numplanes): 
          for y in range(height): 
            row = []
            for x in range(width): 
              row += [c*((y%(len(palette)+2))+1)]
            image.append(row)
        
        # Write the constructed image back to our output
        for plane in image:  
            for cell in plane:
                 out.write(palette[cell])