#!/usr/bin/env python3

import os
import sys

import struct
from typing import BinaryIO


# This function converts a binary string to a more readable hex representation
def str2hex(s : object ) -> object:
  result = ""

  for i in range(len(s)):
    result += "%02X" % s[i]
    if i < len(s)-1:
        result += " "
  
  return result


# Converts a single byte to a two character hexadecimal representation
def formatHexByte(b: int) -> str:
   return f"{(b&0xF0)>>4:x}{b&0x0F:x}"


def rgbToIndexedColor(r : int, g: int, b : int) -> int:
  r &= 0xFF
  g &= 0xFF
  b &= 0xFF

  bestMatchIndex = 0

  minimumDistanceSquared = float("inf")

  for index in range(len(_COLOR_TABLE)):
    cr = _COLOR_TABLE[index][0]
    cg = _COLOR_TABLE[index][1]
    cb = _COLOR_TABLE[index][2]

    dr = abs(cr - r)
    dg = abs(cg - g)
    db = abs(cb - b)

    dMin = min(dr,dg,db) ** 2 
    if dMin < minimumDistanceSquared:
      minimumDistanceSquared = dMin 
      bestMatchIndex = index 
  return bestMatchIndex 

# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7) 

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have one
    if gctlen > 0:
      gct = rng.read(gctlen)

      # If we don't get an even number of colors then we need padding to make sure
      # all resulting images contain the same amount of bytes
      outbytesleft = 256 - len(gct)
      if outbytesleft % 3!= 0:
        out.write(b'\x00'*(outbytesleft%3))
      
      out.write(gct)
    
    # Write the first frame with all the colors from the global color table
    # and randomized delay times
    frame = b"\x2C\x00\x00\x00\x00\x01\x00\x01\x00\x00\x00\x00\x00\x00" \
           + struct.pack('<H', 1 << randint(0,math.ceil(math.log2(gctlen))))  

    out.write(frame)


    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream as they are
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        numcolors = size//3 - 1
        if numcolors > 0:

          # We can only handle up to 256 colors so just fake that many and skip
          # over the rest
          if numcolors > 256:
            numcolors = 256

          colortable = rng.read(numcolors*3)
          out.write(colortable)

          # Skip over the rest of the colors in our palette since we only support
          # 256
          leftovercolorbytes = max((numcolors - 256)*3, 0)
          if leftovercolorbytes > 0:
              rng.seek(leftovercolorbytes, 1)
              
        # Write the actual image data
        imgdata = rng.read(size-1)
        out.write(imgdata)