#!/usr/bin/env python3

import argparse
import os
from pathlib import Path
from contextlib import ExitStack
from tempfile import TemporaryDirectory
from typing import BinaryIO

from PIL import Image

# Writes an entire GIF animation into `outfile`, which is either specified as a path or '-' for stdout
def writeGif(anim, outfile):
    if isinstance(outfile, str):
        with open(outfile, 'wb') as f:
            anim.save(f, save_all=True)
    else:
        raise NotImplementedError("writing directly to binary streams isn't implemented yet")

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--out', default='-', help="path to output.gif (or - for stdout)")
    parser.add_argument('width_px', type=int, nargs='+', help="width in pixels")
    args = parser.parse_args()

    assert len(args.width_px) > 1

    width_range = min(args.width_px), max(args.width_px)

    with ExitStack() as stack:

        # create temporary directory for storing temp files
        tmpdir = stack.enter_context(TemporaryDirectory())

        # set up filenames -- one per frame, plus a palette and map (.plte &.trns)
        filenames = [
                Path(tmpdir).joinpath(f'{i}.png')
                for i in range(1, len(args.width_px)+1)]
        palFilename = Path(tmpdir).joinpath('pal.png')
        transparencyMapFilename = Path(tmpdir).joinpath('trans.png')
        
        # create frames
        for width, filename in zip(args.width_px, filenames):
            img = Image.new('RGB', size=(width, width))

            # randomly color each pixel
            pix = img.load()
            for y in range(img.height):
                for x in range(img.width):
                    r = ord(os.urandom(1))
                    g = ord(os.urandom(1))
                    b = ord(os.urandom(1))

                    pix[x,y] = (r, g, b)
                    
            # save it to a file
            img.save(filename)
    
        # make some rainbow colors
        palletteImg = Image.new('P', size=(256, 1))
        pallettePixels = palletteImg.load()
        hueStepPerPixel = round((1 / len(filenames)) * 360)
        hueDirection = -1
        for i in range(palletteImg.width):
            h = (hueDirection * i) % 360
            col = tuple([round(v*255) for v in ImageColor.getrgb(f'hsv({h}, 1.0, 0.7)')])
            pallettePixels[i, 0] = col

            if h < 5 or h >= 355:
                hueDirection *= -1


        # use final image as the initial background
        im = Image.open(filenames[-1])
        background = Image.new('RGB', size=im.size, color=im.getpixel((0,0)))

        # make the alpha channel transparent wherever there's no change from the previous frame
        transmasks = []
        lastFrame = None
        for filename in reversed(filenames[:-1]):
            currentFrame = Image.open(filename)
            mask = ImageChops.difference(currentFrame, lastFrame)
            lastFrame = currentFrame
            
            # convert to grayscale
            mask = mask.convert('L')
        
            # get all non-transparent pixels
            maskArray = list(set([(x,y) for x in range(mask.width)
                                  for y in range(mask.height)
                                  if mask.getpixel((x,y))!= 0]))

            # get first coordinate to have changed from previous frame
            if not maskArray: continue
            x1,y1,*rest = sorted(maskArray)[0]
            
            # expand the area around that coordinate by half its length on all sides
            x1 -= round(mask.width // 2)
            y1 -= round(mask.height// 2)
            x2 = x1 + mask.width
            y2 = y1 + mask.height
            if x1 < 0:
                x2 += abs(x1)
                x1 = 0
            elif x2 > im.size[0]:
                x1 -= x2 - im.size[0]
                x2 = im.size[0]
                
            if y1 < 0:
                y2 += abs(y1)
                y1 = 0
            elif y2 > im.size[0]:
                y1 -= y2 - im.size[0]
                y2 = im.size[0]

            # add the new expansion region to our alpha masks
            transmasks.append(((x1,y1),(x2,y2)))
            
        # paste images together while composing their respective alphamasks
        prevImage = background
        curTransMaskPos = 0
        for transmask in reversed(transmasks):
            x1,y1 = transmask[0]
            x2,y2 = transmask[1]
            curAlphaMask = Image.new('L', size=im.size, color=None)
            curAlphaMask.paste(prevImage.crop((x1,y1,x2,y2)), box=(x1,y1))
            curAlphaMask.putalpha(255)
            curImage = Image.open(filenames[curTransMaskPos+1]).resize(background.size)
            background.paste(curImage, mask=curAlphaMask)
            prevImage = curImage
            curTransMaskPos += 1

        # apply the pallette and transparency map
        background = background.quantize(palette=palletteImg)
        background.putpalette(palletteImg.getpalette(), "RGBA")
        background.save(transparencyMapFilename, transparency='alphamap')
        background.save(palFilename, format='PNG')

        # write to.gif
        writeGif(background, args.out)