#!/usr/bin/env python3

import io
import sys
import math
import time
import warnings
import argparse
import struct

from PIL import Image


# Iteratively generates frames for a animated GIF that are as far from each other in time
# as possible subject to given constraints and attempts to pass some simple heuristics on color
# usage. Frames will be written into `out`.
class RandomGifGenerator():

    def __init__(self):
        self._rng = None
        self._width = -1
        self._height = -1


    # Determines the block size based on internal state and returns it.
    def _get_blocksize(self):
        return max(math.floor((self._rng.randint(0x10000 // 50)) / 64) * 64, 2 ** 7)


    # Writes an image into `out`, while ensuring that there is space between consecutive images by calling `_ensure_space()`.
    def write_image(self, out: BinaryIO):

        # Ensure we have required space before writing anything.
        self._ensure_space()

        # Write out frame duration information, and fill up with zeroes until end of current frame.
        dur = struct.pack('<HH',
                self._rng.randint(0xf1f),
                0
        )
        out.write(dur + b'\x00' * ((self._width * self._height * 6) % self._framesize))

        # We are going to use this `Image` instance to access RGB pixel values. However, we do not want
        # it to close the underlying file when done, so we set its `fp` field to `None`. More info here:
        # https://pillow.readthedocs.io/en/stable/reference/open_files.html#exclusive-access
        img = Image.new('RGB',
                (self._width, self._height),
                'white'
        )
        img.fp = None
        
        # Now let's draw something! There can be various approaches, but I decided to go for one which uses only
        # very limited subset of available colors, thus decreasing the amount of pixels which require extra bits per pixel.
        table = [
                hex(color)[2:]
                for color in list(img.convert('P').getpalette())
        ]
        maxlen = len(max(table, key=lambda x: len(x)))

        whiteindex = table.index('fff') if 'fff' in table else -1
        blackindex = table.index('000') if '000' in table else -1
        index = -1
        while True:

            # Try randomly generating either a single pixel or multiple ones within a box.
            num = self._rng.choice([0, 1, 1])
            for i in range(num):

                # Generate coordinate pair. `randint()` method used directly because we need integers.
                ycoord = self._rng.randint(0, self._height - 1)
                xcoord = self._rng.randint(0, self._width - 1)

                # Check whether the resulting pixel is within bounds and has been already assigned. If not, assign it.
                # Note that we compare against the string representation of the tuple instead of the tuple itself since
                # tuples are mutable objects and cannot be hashed.
                if (ycoord, xcoord)!= (-1, -1) and str(
                        (ycoord, xcoord)
                    ) not in map(str, zip(*[
                            f.getdata()[int(i * self._width + j)]
                            for i in range(self._height)
                            for j in range(self._width)
                    ])):

                    # We need to avoid drawing too many transparent colors. Thus, whenever we encounter a transparent
                    # pixel, we should check whether there exists at least one more non-transparent one nearby to ensure
                    # that they form a connected group. Otherwise, we need to force ourselves to change the previous
                    # pixel to a different value. This could lead to infinite loop due to unavoidable race condition,
                    # so whenever the desired behavior cannot be achieved after 100 iterations, we raise an exception.
                    nontransparency_count = sum([(
                                True if c == 255
                                    else False
                        ) for c in img.getpixel((xcoord, ycoord))]
                    )
                    
                    limit = self._rng.randrange(
                            1, min(nontransparency_count + 1, 101)
                    )
                    attempts = 0
                    while nontransparency_count > limit and attempts < 100:

                        # Select new coordinates and check again.
                        ycoord = self._rng.randint(0, self._height - 1)
                        xcoord = self._rng.randint(0, self._width - 1)
                        nontransparency_count = sum([(
                                        True if c == 255
                                            else False
                                ) for c in img.getpixel((xcoord, ycoord))]
                        )
                        attempts += 1

                    assert attempts < 100, 'Could not get enough non-transparent neighbors.'

                    # Generate random palette index except those which correspond to transparency (`index >= 0`).
                    index = self._rng.randrange(
                            whiteindex if index >= 0
                                else self._rng.randrange(blackindex + 1, 256)
                    )

                    # Assign the chosen color to our currently selected pixel.
                    img.putpixel((xcoord, ycoord), index)

            # Convert generated image into the binary format expected by GIF and store them into the output stream.
            img.save(out, format='gif', optimize=True)
            
            # Finally, if no pixels were drawn (which means that the entire image consists of
            # fully transparent pixels) we immediately sleep for a bit and move to next iteration.
            if num == 0:
                
                # Sleeping for a fraction of a second prevents us from having extremely short
                # animation sequences where most frames contain nothing but background.
                time.sleep(
                    1e-6 + 1e-7 * float(self._rng.rng.getrandbits(64))
                )
                continue

            break
    

    # Ensures that there is at least `self._framesize` bytes remaining in the provided `stream`. Note that this does not mean
    # that any further writes would succeed; it just makes sure that there will be at least `self._framesize` bytes left afterwards.
    def _ensure_space(self):
        pos = self._rng.tell()
        delta = self._framesize - (pos - self._lastwrite)
        if delta <= 0:
            return

        buf = io.BytesIO()
        buf.seek(-delta, io.SEEK_END)
        buf.write(b''* delta)
        buf.seek(0)

        self._rng.seek(pos)
        self._rng.readinto(buf)
        self._rng.seek(pos)

    def _parseargs(self):
        parser = argparse.ArgumentParser(description='')
        parser.add_argument('--output', '-o', help='output file name')
        parser.add_argument('--seed', '-s', type=int, default=-1, help='PRNG seed')
        args = parser.parse_args()
        return args

    
    def run(self, width: int, height: int, prngseed: int, outfile):

        # Open the output file.
        with open(outfile, mode='wb') as outputfile:

            self._rng = io.BytesIO()
            
            # Seed PRNG.
            if prngseed == -1:
                prngseed = int.from_bytes(time.monotonic().to_bytes(length=8, byteorder='big'), byteorder='big')
            print('[+] Using seed {0}'.format(repr(prngseed)), flush=True)
            self._rng.seek(self._rng.truncate(0))
            self._rng.write(struct.pack('<Q', prngseed))

            # Set dimensions.
            self._width = width
            self._height = height

            # Determine starting point for first frame.
            self._framesize = \
                    math.ceil(((width * height * 6) + 2) / pow(2, math.ceil(math.log2(self._get_blocksize())))) * self._get_blocksize()

            # Firstly, try to find exact positions for subsequent frames such that they are spaced by less than 10 ms
            # and still fit into the total GIF size constraint (or as close as possible). If this fails, fallback to
            # approximate position generation.
            spacing = 0
            lastspacing = 0
            failed = False
            attemptcount = 0
            solutionfound = False
            while not solutionfound:
                
                # Generate 1000 samples and keep track of their respective positions.
                samplepositions = []
                for i in range(1000):
    
                    # Because of the overhead introduced by `write_image()`, we add padding for first 200 frames.
                    # Thus, if we fail to find proper spacing after 200 attempts, we know that the initial positions
                    # where chosen badly. In order to not waste time repeating the same search over and over, we
                    # stop right away and assume that the user screwed up the settings.
                    if attemptcount >= 200:
                        failed = True
                        break

                    # Go to the start of the next frame.
                    self._last