#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF is only as large as `max_size`.
def generate_random_gif(
        rng: BinaryIO, max_size: int, out: BinaryIO
    ) -> None:
    header_struct = struct.Struct('>4s3B')
    buf = bytearray()

    while True:
        # Read a length field from the input and use it to determine how many
        # bytes should be read for this record's content.
        rle_block = rng.read(1)
        if len(rle_block)!= 1 or (ord(rle_block) & ~0x7F):
            raise Exception("invalid RLE block in input stream")

        # If high bit set make this a special case RLE run containing either zero
        # characters or one character repeated multiple times.
        if ord(rle_block[0]) >= 0x80:
            remaining = ord(rle_block[0]) - 0x7F
            n_pixels = rng.read(remaining)
            if len(n_pixels) < remaining:
                raise EOFError

            count = sum([1 for x in n_pixels])

            # Write the XOR'd block to the output buffer
            buf += bytes([(count ^ c) for c in range(256)])
            continue

        # Not a special case RLE block so we have just a normal block.
        n_pixels = rng.read(1)[0]  # first byte of header contains size of block

        # Check whether we are at the end of our allotted space in the output.
        if len(buf)+n_pixels > max_size:
            break

        # Add the data to our buffer by copying a chunk of the /dev/urandom source.
        buf += rng.read(n_pixels)

    # write out a valid header
    out.write(b'\x47\x49\x46\x38\x39\x61')
    out.write(header_struct.pack(*header_struct.size*[0]))

    # write out all of our buffered data
    out.write(buf[:max_size])