#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # First lets make sure that the global color table is properly addressed
    globalcolorsperrow = 8  # From https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    globalcolors = [None]*(globalcolorsperrow**2)
    for i, entry in enumerate(globalcolors):
        if entry is None:
            r = rng.read(1)[0] % 0xFF
            g = rng.read(1)[0] % 0xFF
            b = rng.read(1)[0] % 0xFF

            globalcolors[i] = (r,g,b)

    print('Writing global colors table...', end='')
    for rgb in globalcolors:
        out.write(struct.pack('<BBB',*rgb))
    print('done!')

    # Now we're going to write N descriptors at once
    MAX_DESCRIPTORS = 256

    # Using this loop means we avoid making multiple calls to read()
    for i in range(MAX_DESCRIPTORS):
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)
        entriesconsumed = len([e for e in globalcolors if e is not None])
        datatoreadthisround = min(
                max((datalen-entriesconsumed) // (MAX_DESCRIPTORS-i),
                    1),
                0xFD)


        # TODO: All of the fields need work, they're kind of cruft right now,
        #       see https://github.com/decompme/decompilation-guide/issues/47
        try:
            datafield = rng.read(datatoreadthisround)
        except Exception:
            raise ValueError("Not enough valid data left") from None

        if i < MAX_DESCRIPTORS-1:
            subblocksleft = MAX_DESCRIPTORS-i-1
            # Pad any remaining bytes to trigger an early exit of the decoder.
            # It won't decode them because they aren't part of a full animation,
            # but it should still stop parsing sooner than later.
            if subblocksleft-1 <= datatoreadthisround:
                datafield += b'\x00'*subblocksleft*(0xFE+1)

        out.write(b'\x2C\x00\x00\x00\x00\x00\x00\xFB\x00\x00\x00')
        out.write(datafield)



    last_byte = out.tell() - 1
    # At the moment this code doesn't support appending arbitrary
    # bytes onto the output so we'll have to manually compute where
    # our data will start relative to the rest of the file
    print(f'The generated gif is {last_byte} bytes long.')

    # For now we'll hardcode the number of loops to use, since we don't
    # yet know how much space remains between the final descriptor and EOF
    numloops = 0xC0

    paddinglen = (numloops * 3).to_bytes(size=2, byteorder='little')

    # Write the leftover bytes before the last descriptor
    print('Writing remaining bytes', end='', flush=True)
    out.seek(-1, 2)
    out.write(paddinglen)
    print('.', end='', flush=True)

    # Write the last descriptor!
    out.write(b'\x3B')

    # And pad up until the end of the file, which is easy enough because
    # we've already written exactly datalen bytes before ending.
    print('Padding global color table', end='', flush=True)
    while True:
        if out.tell() == datalen+1: break
        out.write(rng.read(1))
    print('. Done!', flush=True)