#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out`. Randomness comes from
# the provided random number generator (`rng`) which must be a readable
# binary file handle (usually something like `/dev/urandom`). The
# `width`, `height`, `duration`, and `loops` parameters are used as
# part of generating the GIF file format.
def generate_random_gif(
        rng: BinaryIO, width: int, height: int, duration: int, loops: int,
        out: BinaryIO
    ):
    # Initial byte sequence
    hdr = b'GIF89a' + \
            bytes([
                # Logical Screen Descriptor
                0x2c |        # Flags
                0 |           # Color Table
                0 |           # Sort Method
                0,            # Background Color Index
                0,            # Pixel Aspect Ratio
                0,            # Image Left Offset
                0,            # Image Top Offset
                width,        # Width
                height],      # Height
            )

    # Pack up an integer bitfield from 0..7 with each segment providing
    # one bit that describes how many times larger than 8 each component 
    # will be scaled during display ([3 bits][3 bits][3 bits]).
    pixelscalerflags = [
        0b100 << i | 
        0b010 << j | 
        0b001 << k
        for i in range(3)
        for j in range(i+1, 3)
        for k in range(j+1, 3)
    ]

    # Sample a pixel scaler flag at random
    def samplepixelscaleflag():
        return rng.choice(pixelscalerflags)

    # Append pixel scaler flags to initial byte array
    hdr += bytes([samplepixelscaleflag()] * 3)

    # Add palette colors to the end of the byte array if needed
    palen = 2 ** ((hdr[-3] >> 5) + 1) * 3
    if palen > 0:
        # Read all available RGB values from the random number generator
        rgbs = [rng.read(3) for _ in range(palen)]
        # Add them to the end of the header
        hdr += b''.join(rgbs)

    # Append initial loop count value to header
    hdr += struct.pack('>H', loops)

    # Append delay time to header
    hdr += struct.pack('>H', duration)

    # Write header to target file
    out.write(hdr)

    # Create a buffer to store the current frame
    curframe = bytearray()
    while True:

        # Iterate over the three control blocks defined by the standard
        for blktype in ('imagedesc','graphiccontrol','applicationext'):

            # Start off by writing the block identifier byte
            out.write(bytes([0x21]))
            
            # Then get the size of the rest of the block...
            sz = 1
            while True:
              val = ord(rng.read(1))

              #...and append it to our running total
              sz <<= 7
              sz |= val & 0x7f
              
              # Break once the low order bit indicates the end of 
              # the block
              if val < 128:
                  break

            # Write the full size to the output
            out.write(struct.pack('>H', sz))

            # Continue reading until the block ends or we reach 
            # the application extension block
            while True:
              val = ord(rng.read(1))
              out.write(bytes([val]))

              # End when the first non-zero byte occurs since 
              # we only care about the second byte after the start
              # of the block - it contains the block type
              if val!= 0:
                break
              
            # Check if we've reached the application extension block
            if blktype=='applicationext':
              continue

            # Otherwise read the contents of the block and write them
            # to the final file
            cbuf = rng.read(sz-1)
            out.write(cbuf)