#!/usr/bin/env python3

import argparse
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, size: int
    ):

    # Add magic numbers to the beginning of the file
    out.write(b'GIF89a')

    # Create each frame and write it to our output file
    for i in range(size):
        # Generate a random palette entry
        rgb = [
            rng.read(1)[0] & 0x3f | 0xc0,
            rng.read(1)[0],
            rng.read(1)[0]
        ]

        # Write color table
        out.write(struct.pack('BBBxxx', *rgb))

        # Determine how large this frame will be
        width = rng.read(1)[0] & 0x7f + 1
        height = (rng.read(1)[0] & 0xff ^ 0xff) + 1
        control = bytes([
            0xf9, # Image separator
            0x04, # Left position
            0x00, # Top position
            0x00, # Width
            0x00, # Height
            rng.read(1)[0] & 0x07 | 0x80, # LZW min code size
            0x00   # No local color map
        ])

        # Write canvas descriptor
        out.write(struct.pack('<HHHBB', width, height, 0x8765, 0, 0))
        
        # Write image descriptor
        out.write(struct.pack('<HHBxxx', 0, 0, 0, 0, 0))
        
        # Write graphic control extension block
        out.write(control)

        # Generate image data by applying a crude algorithm
        # that generates a bunch of black pixels where n is
        # equal to the length of the data string. lzw() returns 
        # the minimum code size needed to represent the image.
        def lzw():
            codes = {
                b'' : 0xffff,
                0xffff : -1,
                bytearray((i,)) : i 
            }

            curcode = []
            prev = None
            while True:
                if len(curcode) == 0:
                    c = rng.read(1)
                else: 
                    c = curcode[-1:]

                if not c or prev not in codes: return max(codes.values())+1
                
                prev = c[0]
                prevc = codes[prev]
                nc = rng.read(1)[0] & 0x7f
                curcode = bytearray(nc)
                pos = 0

                if nc > 0:
                    curcode[pos] |= 0x80

                    for _ in range(nc-1):
                        if len(curcode) == pos+1:
                            curcode += bytearray((c[0],))
                        else: 
                            curcode[pos+1] |= rng.read(1)[0] & 0x7f

                        pos -= ~pos&1

                if prevc >= 0:
                    if prevc < 256 and nc!= 0:
                        codes[curcode[:]] = prevc
                    elif prevc not in codes:
                        codes[prevc, curcode[:]] = len(codes)//2
                    else:
                        codes[len(codes)] = prevc
                        
                yield (prevc << 8 | nc).to_bytes(2, 'little')
                
        # Generate lzw encoded pixel data
        data = bytes().join(lzw())

        # Write lzw encoded data
        out.write(data)

        # Zero pad until next multiple of 256
        if len(data)%256:
            out.write(bytearray(((256-(len(data)%256))%256,)))

    # Write EOF marker
    out.write(b';')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Generate a GIF with a random palette')
    parser.add_argument('-o','--output', default='rand.gif', type=argparse.FileType('wb'), help='output filename')
    parser.add_argument('--size', default=1, type=int, help="number of frames")
    
    args = vars(parser.parse_args())
    
    generate_random_gif(open('/dev/urandom', 'rb'), **args)