#! /usr/bin/env python3

import argparse
import os
import sys
from enum import IntEnum, auto


class Header(IntEnum):
    SCREEN_WIDTH = 768 # pixels
    SCREEN_HEIGHT = 512 # pixels
    COLOR_BITS = 1 << 7 | 1 << 4
    COLOR_TABLE = False
    INTERLACED = False
    GCT_SIZE = 7 # entries

    def __str__(self):
        return self._name_.lower().replace('_', '-')


HEADER = bytearray([
    0x47, 0x49, 0x46, 0x38,  # magic string
    0x39,                    # version of.gif spec
    0x61,                    # screen width in bytes (LSB first)
    0xff, 0x00,              # screen width as LSB-MSB
    0x00, 0x00,              # screen width as MSB-LSB
    0x00,                    # screen height in bytes (LSB first)
    0xc2,                    # screen height as LSB-MSB
    0x00, 0x00,              # screen height as MSB-LSB

    0xe0, 0x00                # color table size (in bytes)
])

FOOTER = bytearray([
    0x00,                   # background index
    0x00,                   # pixel aspect ratio
    0x3b                    # end of image marker
])

# A packed bitfield configures how to pack the resulting colors
class PackedBitField(object):
    BITS = {
        1:   0b1,
        2:   0b11,
        3:   0b111,
        4:   0b1111,
        5:   0b11111,
        6:   0b111111,
        7:   0b1111111,
        8:   0b11111111
    }


    @classmethod
    def create1(cls, d, x=0):
        """Creates an instance of this class with 1 bit"""
        return cls(d, [(w, x) for w in range(0, 9)])

    @classmethod
    def create2(cls, d, x, y):
        """Creates an instance of this class with 2 bits"""
        return cls(d, [(w, x) for w in range(0, 8)] + [(y, 0)])

    @classmethod
    def create3(cls, d, x, y, z):
        """Creates an instance of this class with 3 bits"""
        return cls(d, [(w, x) for w in range(0, 7)] +
                    [(z, 2), (y, 0)])

    @classmethod
    def create4(cls, d, x, y, z, t):
        """Creates an instance of this class with 4 bits"""
        return cls(d, [(w, x) for w in range(0, 6)] + 
                    [(t, 3), (z, 2), (y, 0)])

    @classmethod
    def create5(cls, d, x, y, z, t, u):
        """Creates an instance of this class with 5 bits"""
        return cls(d, [(w, x) for w in range(0, 5)] + 
                    [(u, 4), (t, 3), (z, 2), (y, 0)])
        
    @classmethod
    def create6(cls, d, x, y, z, t, u, v):
        """Creates an instance of this class with 6 bits"""
        return cls(d, [(w, x) for w in range(0, 4)] + 
                    [(v, 5), (u, 4), (t, 3), (z, 2), (y, 0)])

    @classmethod
    def create7(cls, d, x, y, z, t, u, v, w):
        """Creates an instance of this class with 7 bits"""
        return cls(d, [(w, x) for w in range(0, 3)] + 
                    [(w, 6), (v, 5), (u, 4), (t, 3), (z, 2), (y, 0)])

    @classmethod
    def create8(cls, d, *args):
        """Creates an instance of this class with 8 bits"""
        return cls(d, list(zip(*[iter(range(8))] * 2)))
    
    def __init__(self, _dict_, pairs):
        if not all(len(p) == 2 and isinstance(p[0], int)
                   and p[0] >= 0 and p[0] <= 7 for p in pairs):
            raise ValueError("Each pair should be a two element tuple "
                             "with the second value between 0 and 7")

        tot = sum(map(lambda p: pow(2, p[0]), pairs))
        if not len(_dict_) == tot or any(map(lambda k: not hasattr(_dict_, str(k)),
                                              range(tot))):
            raise ValueError("Length of dict must match total number of "
                             "defined bits")

        self.__pairs__ = sorted([(w, f) for (f, w) in pairs])
        self.__mask__ = reduce((lambda m, pp:
                                (m >> 1) - ((pp[1] & m & 0b1) << 7)),
                               reversed(self.__pairs__),
                               0b11111111)
        self.__dict__ = _dict_
    
    def set_bit_order(self, order):
        """Sets the bit ordering of this bitfield
        
        The ordering should be a sequence that represents 
        the desired output order of each bit within this field"""
        new_pairs = []
        i = 0
        while True:
            try:
                o = next(order)
                old_pair = self.__pairs__[i]
                new_pairs.append((old_pair[0], o))
                i += 1
            except StopIteration:
                break

            if i > 7:
                raise ValueError("Ordering is too long!")    

        new_pairs.extend(self.__pairs__[i:])
        self.__pairs__ = new_pairs

    def get_pack_string(self):
        fields = ''
        for _, mask in self.__pairs__:
            fields += '{:0>8}'.format('{:b}'.format(mask)[::-1])
        return '<' + fields

    def get_unpacked(self, r, g, b):
        return self.__dict__.get(
                sum(m << s for (_, m), s
                    in zip(self.__pairs__, (r, g, b))),
                0)

    def unpack(self, lst):
        result = []
        for e in lst:
            r,g,b,_ = struct.unpack('<BBBB', e)
            result.append(self.get_unpacked(r, g, b))
        return result        

    def pack(self, lst):
        fmt = self.get_pack_string()
        return (struct.pack(fmt, *(e for l in lst
                                  for e in self.unpack(l))))
    

PACKING = PackedBitField.create8({
    0x000000: 0b111,
    0x0000FF: 0b111,
    0xFF0000: 0b111,
    0xFFFF00: 0b111,
    0xFFFFFF: 0b111,
    0x00FF00: 0b1101,
    0x00FFFF: 0b1101,
    0xFF00FF: 0b1101,
    0xFFFFFF: 0b1101,
    0x0000FF: 0b1100,
    0x00FF00: 0b1100,
    0x00FFFF: 0b1100,
    0xFF0000: 0b1100,
    0xFF00FF: 0b1100,
    0xFFFF00: 0b1100,
    0xFFFFFF: 0b1100,
    0x0000FF: 0b1011,
    0xFFFFFFFF: 0b1011,
    0xFF00FF: 0b1011,
    0x00FF00: 0b1010,
    0xFFFF00: 0b1010,
    0xFFFFFF: 0b1010,
    0x0000FF: 0b1001,
    0x00FF00: 0b1001,
    0xFFFFFFFF: 0b1001,
    0xFF0000: 0b1001,
    0xFFFF00: 0b1001,
    0xFF00FF: 0b1000,
    0x00FF00: 