#!/usr/bin/env python3

import math
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 2 bytes of length of global color table.
#       - If zero, indicates no global color table should be used.
#       - Otherwise, contains the size of the color table in entries.
#           - Every entry is either a single-byte value or an RGB triple.
#   3. Global color table.
#       1. Globally specified colors that can be shared between images in a given
#          animation. The first byte of this block specifies how many colors are
#          being defined here. A special case is when this field's first byte is
#          0, which means there is no global color table available. Note that if
#          there is a local color table definition as part of the
#          corresponding image descriptor then it must precede any global color
#          table and its size must match up to the one found in the header;
#          otherwise, all data after the global color table will be ignored.
#   4. Blocks of image data.
#       1. Each frame has at most two blocks associated with it:
#           1. Image descriptor. Specifies location on the screen for each frame by
#              defining the top left corner position + width / height offsets for
#              drawing the graphic, whether or not local color tables are present,
#              and other properties related to interlacing & transparency.
#           2. Local color table. Same format as the global color table above. This
#              defines the color mappings used for non-local color indices in the
#              next image sub-block.
#       2. An end marker appears last once the whole animation sequence finishes.
#          It consists of only the GIF terminator code ("!" character) followed by
#          three null bytes.
#
# See https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, *,
        repeat: bool = True, max_frame_count: float = None
    ):

    # Header
    ## Magic number
    out.write(b'GIF89a')


    ### Logical Screen Descriptor
    ## Size
    w = int.from_bytes(rng.read(2), 'little', signed=False)
    h = int.from_bytes(rng.read(2), 'little', signed=False)
    ## Has global color table?
    gctsize = rng.read(1)[0]
    assert gctsize <= 256

    # Field Control Flags
    ffc0 = False
    ffc1 = False
    ffc2 = False
    ffc3 = False
    ffc4 = False
    ffc5 = False
    ffc6 = True   # User input (not done yet!)
    ffc7 = False
    bitmask = [ffc0, ffc1, ffc2, ffc3, ffc4, ffc5, ffc6, ffc7]
    flags = sum([int(bit) << i for i, bit in enumerate(bitmask)])
    out.write((flags).to_bytes(1, byteorder='little'))

    # Background Color Index
    out.write(''.encode())         # Default transparent background.

    # Pixel Aspect Ratio
    aspectratio = rng.read(1)[0]  # default ratio = 1:1
    out.write(aspectratio.to_bytes(1, byteorder='little'))

    ### Global Color Table
    ## Colors
    numcolors = min(gctsize, 256)
    clist = [rng.read(3) for _ in range(numcolors)]
    out.write(clist[0])             # First color is always black.
    if len(clist) > 1:
        for pixel in clist[1:]:
            out.write((pixel+pixel).decode())

    ### Frames
    while True:

        ### Image Descriptor
        ## Left Position
        xpos = int.from_bytes(
                rng.read(2), byteorder='little', signed=False
        )
        ## Top Position
        ypos = int.from_bytes(
                rng.read(2), byteorder='little', signed=False
        )
        ## Width
        imgwidth = int.from_bytes(
                rng.read(2), byteorder='little', signed=False
        )
        ## Height
        imgheight = int.from_bytes(
                rng.read(2), byteorder='little', signed=False
        )


        ## Field Control Flags
        ffc0 = False     # LZW Minimum Code Size
        ffc1 = True      # Interlace flag
        ffc2 = False     # Sort flag
        ffc3 = False     # Reserved
        ffc4 = False     # Reserved
        ffc5 = True      # Transparency Flag
        ffc6 = False     # Delayed Loading
        ffc7 = False     # Reserved
        bitmask = [ffc0, ffc1, ffc2, ffc3, ffc4, ffc5, ffc6, ffc7]
        flags = sum([int(bit) << i for i, bit in enumerate(bitmask)])
        out.write((flags).to_bytes(1, byteorder='little'))

        ## Indexed Color Table
        ## Number of Entries
        tablelen = int.from_bytes(
                rng.read(1), byteorder='little', signed=False
        )
        ## RGB triplets
        ct = []
        for _ in range(tablelen):
            rgbtriplet = tuple(
                    rng.read(1)[0] for _ in range(3)
            )
            ct += list(rgbtriplet)
        out.write(chr(tablelen))
        for pixel in ct[:256]:
            out.write(chr(pixel))

        ## LZW minimum code size
        lzwmincodesize = rng.read(1)[0]
        out.write((lzwmincodesize).to_bytes(1, byteorder='little'))

        ### Image Data
        ## Image LZW codes
        # Read lzwmincodesize bytes from the input stream.
        # Let S be the set of these bytes concatenated together.
        # Let k be the first byte read from the stream.
        # Let s be the empty string.
        # While we have not reached the end of the data...
        #     let m = the next byte in the input stream.
        #     Add m to the end of S.
        #     If s and S both begin with the same byte k, do the following:
        #         Write k to the output stream.
        #         Remove k from the beginning of S and from the stream just read.
        #         Set k to the next byte in the input stream.
        #         Append the value of k to S.
        #         Repeat step 2 until S does not begin with k.
        #     Else if S begins with a prefix of k...
        #         Write k to the output stream.
        #         Prepend k to S.
        #         Set k to the first byte of S.
        #         Repeat steps 4 through 6.
        #     EndIf
        # EndWhile
        buf = rng.read(lzwmincodesize)
        S = ''.join(map(chr, buf))
        k = chr(buf[0])
        s = ''

        # These values were taken from the C program in the spec.
        MAXBITS = 12
        HSIZE = 5003
        hashsize = 997
        bitmask = (1 << MAXBITS)-1
        free_ent = 0

        ent = ''
        switch = {}
        offset = 0
        eoi = chr(0x00)
        clear = chr(0x01)
        eof = chr(0x02)
        oldcode = None

        def newcodes(n_newcodes):
            global ent, offset, hsize, codesize, lowoffset, highoffset

            offset -= n_newcodes
            lowoffset = -1

            if offset < 0:
                offset += (1<<MAXBITS)

            while highoffset < offset:
                highoffset += 1<<bitmask
                bitmask <<= 1

                if bitmask >= (1<<MAXBITS):
                    bitmask = (1<<MAXBITS)-1

                    hshift = ((oldcode << 3) ^ oldcode) >> 13
                    codesize += 1
                    hsize *= 2

                    tmp = [None]*hsize
                    for item in switch.items():
                        tmp[item[0]] = item[1][:]
                    switch = dict([(i, tmp[i]) for i in range(hsize)])

                code = (((highoffset << clc) ^ lowoffset) % hsize)

                switch[code][-1] = (oldcode, tblsize)
                if code in switch[code]:
                    lst = switch[code]
                    pos = 0
                    while pos < lst.__len__():
                        prev = lst[pos]

                        if prev == oldcode:
                            del lst[pos]
                            break
                        else:
                            pos += 1
                    pos = lst.__len__()
                    lst.append(-1)
                    lst[pos] = oldcode

               