#!/usr/bin/env python3

import os
import sys
from typing import BinaryIO

# Generates a random GIF file into `out`, pulling bytes from stdin
def main():
  # Don't allow arbitrary files as arguments
  infile = sys.stdin.buffer

  # Open an output file if specified or use stdout
  outfile = None
  if len(sys.argv) > 1:
    outfile = open(sys.argv[1], 'wb')
    print('Writing to %s...' % sys.argv[1])
  else:
    outfile = sys.stdout.buffer
    
  generate_random_gif(infile, outfile)
  exit(0)
  
# Generates a random GIF file into `out` using the
# random number generator `rng` (`in`)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = read_bytes(rng, 7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have one
    if gctlen > 0:
      gct = read_bytes(rng, gctlen)
      out.write(gct)
    
    while True:

        # Read next field byte for block type
        fieldbyte = ord(read_bytes(rng, 1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(read_bytes(rng, 1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        numcolors = size//3 - 1
        if numcolors > 0:
          colortable = read_bytes(rng, numcolors*3)
          out.write(colortable)

        # Now write image data 
        imgdata = read_bytes(rng, size-1)
        out.write(imgdata)
          
def read_bytes(rng: BinaryIO, n: int)-> bytes:
  buf = []
  while len(buf) < n:
    chunk = os.read(rng.fileno(), n)
    if not chunk: break;
    buf += list(chunk)
  return bytes(buf[:n])