#!/usr/bin/env python3

import math
import os
import time
import struct
from typing import BinaryIO

class RandomGifGenerator:
    def __init__(self, seed=None, maxframetime=0.05):
        self._rng = None
        self.seed(seed)

        minframetime = 0.05
        assert maxframetime <= 1 and maxframetime >= minframetime
        self.maxframetime = maxframetime
        self.minframetime = minframetime
        self.delaytable = [int(math.floor((maxframetime - minframetime)/i))*10 for i in range(1,256)]
        self.colortable = [(int(c*255)&0xFF) for c in [
            0x00 / 255., 0x54 / 255., 0xA8 / 255., 0xFF / 255.,
            0x0D / 255., 0x3F / 255., 0x7E / 255., 0xCC / 255.,
            0x3A / 255., 0x69 / 255., 0xAA / 255., 0xE8 / 255.,
            0x76 / 255., 0xBA / 255., 0xF9 / 255., 0xFF / 255.]
        ]

    @property
    def rng(self):
        return self._rng or SeedSource().rng

    def seed(self, seed=None):
        """If no seed is provided, one will automatically be determined."""
        self._rng = time.Random(seed)

    def generate_framedata(self):
        """Generates a new framedata field for a single frame"""

        width = 32
        height = 32

        decoderowlen = ((width+7)//8)*height

        comprowlen = (((width+7)//8)+1)
        rowlen = (decoderowlen+1)*(3//2)+comprowlen+1


        framedata = bytearray([0] * rowlen)
        framedata[0] = 0x21 # Extension introducer
        framedata[1] = 0xFF # Application extension
        framedata[2] = 0x0B # Block size
        framedata[3] = ord('S')
        framedata[4] = ord('G')
        framedata[5] = ord('I')
        framedata[6] = ord('F')
        framedata[7] = ord('X')
        framedata[8] = 8
        framedata[9] = 1
        framedata[10] = 0x09
        framedata[11] = 0x03
        framedata[12] = len(self.colortable)%256
        framedata[13] = len(self.colortable)//256
        for c in self.colortable:
            framedata.append(c)
        
        framedata.extend([0x2C, 0, 0, 0, 0,
                          0,            # Left
                          0,            # Top
                          width%256,    # Width
                          width//256,   # Height
                          0             # LZW minimum code size
                         ])
        delay = self.delaytable[self.rng.randrange(len(self.delaytable))]
        framedata[-2] = delay % 256
        framedata[-1] = delay // 256

        mask = 0b00001111
        # Note that it's important to switch from top->bottom in each inner loop
        # otherwise some decoders might not interpret the frame correctly
        for y in reversed(range(height)):
            framedata.append(mask)
            for x in range(width):
                bit = bool(self.rng.getrandbits(1))
                framedata.append(bit)

            # Mask out any unused bits above the first line
            framedata += ([False]*(decoderowlen-x))

        # Our encoding reference has various errors!
        # For example, it claims that everything beyond this point should be 0s,
        # but apparently it needs to contain the actual encoding, plus whatever
        # extra garbage needed to pad it out to rowlen bytes. So we need to
        # calculate the proper size and add bytes to make sure they fit
        FDATASIZE = len(framedata)-1
        compdatasize = rowlen-FDATASIZE
        assert compdatasize < 2**16

        # Fill in compression data
        # In theory our compression data could include junk data that would get
        # discarded later when being unpacked by the decoder, so we need to fill
        # the entire area (minus header) with zeros before packing it
        framedata += [0]*max(0,(compdatasize-3))
        lzwsize = struct.pack('<H', compdatasize)
        framedata[14:16] = lzwsize
        framedata += [0]*(rowlen-len(framedata))

        return framedata



# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width: int = 32, height: int = 32, nframes: int = 256,
        maxframetime: float = 0.05
    ):
    
    gen = RandomGifGenerator(maxframetime=maxframetime)

    # Set up the header
    header = b'GIF89a'
    datalen = 10 + (nframes*(13+(width+7)//8)*(3//2)+(width+7)//8 +1)*height
    data = rng.read(datalen-(len(header)))
    out.write(header + data)

    # Now we're going to write nframes frames at once
    # Using this loop means we avoid making multiple calls to read()
    for f in range(nframes):
        framedata = gen.generate_framedata()
        out.write(framedata)