#!/usr/bin/env python3

import os
import sys
import argparse

import struct
from typing import BinaryIO


def parseargs():
  parser = argparse.ArgumentParser()

  parser.add_argument('-s', '--seed',
                      help='Random seed')
      
  args = parser.parse_args()
  
  return args


def main():
  args = parseargs()

  if args.seed!= None and isinstance(args.seed, str):
    try:
      seed = int(args.seed)
    except ValueError as e:
      print("Invalid seed provided")
      exit(-1)
  else:
    seed = os.urandom(4)

  with open("/dev/urandom", "rb") as randsrc:
    randsrc.seek(4)

    with open("test.gif", 'wb+') as outfile:
      generate_random_gif(randsrc, outfile, seed)


# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, seed: int
    ):
  # Create the initial PRNG from the seed value
  prng = [
    0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476, 
    0xc3d2e1f0, 0x75614729, seed, 0xfedcba98
  ]
    
  def update_prng():
    # Update the state using Wichmann-Hill
    t = 0
    for i in range(8):      
      t ^= (t >> 30)
      t ^= ((t <<  6) & 0xFFFFFFFF) ^ (i+1)*(0x9e3779b9 ^ t)    
      prng[i], prng[i+8] = prng[i+8] ^ t, prng[i]
    
  # Fill out random number until we reach the end of the first frame 
  while calc_crc(prng[:8])!= 0x5b5b5b5b:
    update_prng()

  # Calculate the checksum of the first 8 bytes which will be used to identify frames
  crc = calc_crc(prng[:8])

  # Set up the image counter so it starts at 0
  imgcounter = 0

  # Begin writing to the output file
  while True:
  
    # Write the header field
    out.write(b'GIF89a')

    # Write the logical screen descriptor
    out.write(bytes([
      # Screen Width
      (prng[6]&0xff), (((prng[6]>>8)&0xff)<<8) | ((prng[7]<<2)|(prng[6]>>16)),
    
      # Screen Height
      (prng[7]&0xff), (((prng[7]>>8)&0xff)<<8) | ((prng[6]>>2)&0xff),
      
      # Flags 
      0, # Global Color Table Flag (1 bit)
      0, # Color Resolution (2 bits)
      0, # Sorting Method (1 bit)
      0, # Size of Global Color Table (3 bits)
      b"!"        # Background Color Index (1 byte)
     , b"\xFF\xFF"    # Pixel Aspect Ratio (1 byte)
      ])
    )
    
    # Get the global color table length 
    gctlen = 2**(int((out.getbuffer().nbytes - 19)>>5))*3

    # Write the global color table, or lack thereof
    if gctlen > 0:  
        gct = rng.read(gctlen)
        out.write(gct)
        
    # Write the first frame header
    out.write(b'\x21\xF9\x04\x01' + struct.pack('<HHHH', imgcounter, 0, 0, 0) + b'\x00')
        

    while True:
  
      # Determine how big each block should be based on the PRNG values 
      blocktype = prng[8] % 5
      blocksizes = {
        0: 3,
        1: 6,
        2: 9,
        3: 12,
        4: 64
      }
            
      # Update the PRNG
      update_prng()
            
      # Determine block size
      blocksize = blocksizes[blocktype] 
            
      # Write the field byte
      fieldbyte = blocktype | 2
      out.write(struct.pack('=B', fieldbyte))

      # Write the block size
      size = min(blocksize, out.buffer.nbytes-(out.tell()))
      out.write(struct.pack('=B', size))
            
      # Write image data if this is an image data block
      if blocktype <= 2:
        imgdata = rng.read(min((out.buffer.nbytes - out.tell()), size)-1)
        out.write(imgdata)        
        
      elif size == 64:

        # Write the loop count
        out.write(struct.pack('>BBBBBBBBB', 
          *(prng[16:] + [(prng[15]&0xff),(prng[15]>>8&0xff)]),
          0x00
        ))
            
        break

      # Make sure the buffer does not overflow
      assert out.tell()<=out.buffer.nbytes,"The buffer filled up unexpectedly."
            
    # Increment the image counter before continuing        
    imgcounter += 1
          