#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
#   Header
#       - 4 bytes of GIF magic number "GIF89a".
#   Logical Screen Descriptor
#       - 2 bytes of width in pixels as unsigned short.
#       - 2 bytes of height in pixels as unsigned short.
#   Extension Block
#       - 2 bytes for label indicating extension type.
#       - 2 bytes specifying length field.
#       - The remainder is the actual data that makes up this block.
#           - 1 byte for flag indicating whether it's safe to loop.
#               If set to zero then the whole animation will be played once.
#           - 2 bytes of delay time between frames as unsigned shorts.
#   Image Descriptors
#       - There are 0x7fff blocks of these.
#          - A single-byte even indicates an image separator rather than one
#            containing an image descriptor. This means we can't put any more
#            together at least until we figure out how exactly those fields work.
#       - Each image descriptor consists of multiple parts:
#           - Flags and size. Each bit is either 1 or 0, where 1 means
#             that there exists information after each part for the specified part.
#             7654 3210
#             ---- ----
#             xxxx xxxx xxxxxxxx Size in bytes for this particular frame.
#             xx xxxxxxxxx Interlace flag.
#                 x    1   Animation is interlaced.
#                 xxx  000 Background color index.
#                 xxx xxx  00 Foreground color index.
#             xx      x   Sort flag.
#                 x    1   Local colors are sorted by usage frequency.
#                 xx  00    Reserved bits. Must be zero.
#             x       x   Reserved bit.
#                 x    0   Default LZW min code size.
#                 x    1   Specified LZW min code size.
#                 xx  00    Reserved bits. Must be zero.
#           - 0-255 bytes size. Optional graphic control extension describing
#             the effect of the preceding frame on the current frame.
#           - Immediately followed by another size, optional local color table.
#           - Immediately followed by a nonzero sized graphics rendering block.
#
# From http://giflib.sourceforge.net/whatsinagif/index.html with some minor tweaks.
def generate_random_gif(rng: BinaryIO):
    # Generate header and logical screen descriptor.
    header = b'GIF89a'
    width = 0xffff ^ rng.getrandbits(16) & 0xffff
    height = 0xffff ^ rng.getrandbits(16) & 0xffff
    yield header

    # Write logical screen descriptor.
    desc = struct.pack('<HHB', width, height, 8)
    yield desc

    # Write global color table for background color indices 0 through 7.
    count = 0
    while count < 8:
        yield bytes([count]) * 3
        count += 1;

    # TODO: Add local color tables here. These should probably not have the same
    # palette as the global ones.

    # Write rest of image descriptors.
    done = False
    while not done:

        # Determine flags and size for this image descriptor.
        # Note: We're only writing the first two entries of FLAGS so far since
        # they're both reserved.
        flags = [0] * 9
        flags[2] |= 1 << 7 if rng.randint(0, 1) == 1 else 0
        flags[1] |= 1 << 7 if rng.randint(0, 1) == 1 else 0
        flags[1] |= rng.randint(0, 7) << 4
        size = rng.randint(0, 255)
        
        flags = bytes(flags[:2])
        
        yield flags + bytes([size])

        # Check if we need to write a graphic control extension.
        # That happens every other frame because that's what I used when making
        # my test images.
        if size!= 0 and rng.randint(0, 1) == 1:

            # Write the graphic control extension.
            # Control extension type should always be 0xf9.
            exttype = bytes([0xf9])

            # Use a random delay time from 0 to 65535.
            delaytime = struct.pack('<H', rng.randint(0, 65535))
            
            extsize = len(delaytime) + 1
            extdata = delaytime + bytes([0b0])

            yield exttype + bytes([extsize]) + extdata
            
        # TODO: Add local color tables here too. We don't have to worry about them
        # being the exact same colors as the global ones because they'll show the
        # same pattern over itself.

        # Skip padding byte.
        if (size % 2) == 1:
            yield bytes([0]);

        # Write graphics rendering block.
        # First check if we want to end the image.
        if size == 0:
            done = True
            break

        lzwmincodesize = rng.choice((1, 2, 3, 4))
        yield bytes([lzwmincodesize])

        # Convert the size to chunks of 256 bytes representing the max possible
        # chunk size for EOI at 255.
        chunks = []
        while size > 0:
            chunklen = size // 255
            if chunklen == 0:
                chunklen = 1
            chunks.append(chunklen)
            size -= chunklen * 255

        output = None
        for i in range(chunks[-1]):
            # Calculate the chunk size.
            chunklen = chunks[-1]
            if i!= 0:
                chunklen -= 1
                
            # Generate output starting with a single byte expressing the chunk size.
            if i == 0:
                output = bytes([], 'ascii')
            else:
                output += bytes(' ', 'ascii')
            output += str(hex(chunklen)).upper()[2:]

            # Now generate a bunch of random values ranging from 0 to 255 that represent
            # the next byte for our compression scheme.
            for j in range(chunklen):
                loopflag = rng.randint(0, 1) == 0
                foreidx = rng.randint(0, 7)

                if rng.randint(0, 1) == 1:
                    comprsize = rng.randint(0, 2**lzwmincodesize-1)

                    # Codes smaller than this get encoded directly, larger codes
                    # are stored using RLE.
                    mincomprsize = 1
                    
                    # RLE encoding consists of two bytes. One specifies the number
                    # of times the given value has been repeated in a row. The second
                    # contains the value that was repeated.
                    output += f'{str(hex(mincomprsize))[2:]}' \
                            + f'{str(hex(comprsize))[2:]}' 
                elif rng.randint(0, 1) == 1:
                    comprsize = rng.randint(0, 255)
                    output += f'{str(hex(comprsize))[2:]}'
                else:
                    comprsize = rng.randint(0, 255)
                    output += f'{str(hex(comprsize)[2:])}'
                
                # Append the compressed block to the output string to be written later.
                output += '\n'

                # TODO: For now just generate zeros but eventually do something better.
                output += ''.join(['\x00'] * comprsize)

        output += 'eoi'

        # Note: Don't forget to include the trailing newline character!
        yield bytes(output, 'ascii') + bytes('\n', 'ascii')