#!/usr/bin/env python3

from sys import argv,exit
from os.path import isfile,realpath,dirname
import numpy as np
from scipy.optimize import curve_fit
import matplotlib as mpl
mpl.use("Agg")
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter
from datetime import datetime
import sqlite3
import subprocess
import pandas as pd
import json

def parseArgs():
    """
    Parses command line arguments
    
    Returns a dictionary containing the parsed arguments and their values if successful, None otherwise
    """
    args = {}

    # Check number of arguments
    numArgs = len(argv)
    if (numArgs < 2):
        print ("Usage: %s job.json [taskname]" % argv[0])
        return None

    # Parse JSON file
    try:
        with open(argv[1],"r") as fd:
            args["job"] = json.load(fd)
    except Exception as e:
        print("Error parsing job description file '%s':\n%s" % (argv[1],e))
        return None
    
    # Add task name
    args["task"] = "task-001"
    if (numArgs >= 3):
        args["task"] = argv[2]
        
    return args

def getLastDumpFile(jobConfig, taskName=None):
    """getLastDumpFile Generates the file name of the last dump file
    
    Arguments:
        jobConfig {dict} -- Job configuration data
        taskName {str} -- Task name (optional)
    
    Returns:
        str -- File name of the last dump file
    """

    fileName = "%s/%s/" % (jobConfig["OutputPath"],jobConfig["TaskName"])

    # Clean up name
    if (taskName != None):
        fileName += taskName + "-"

    # Add date and time
    dateTime = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
    fileName += dateTime + ".json"

    return fileName

def updateJobStatus(jobResult, success=True, errorMessage=""):
    """updateJobStatus Updates the status of the job in the DB
    
    Arguments:
        jobResult {dict} -- Result of the function that was run by this process
        success {bool} -- True if the function finished successfully, False otherwise
        errorMessage {str} -- Error message if an exception occurs
    """
    
    # Open database connection
    dbConn = None
    try:
        dbConn = sqlite3.connect(jobResult["dbFileName"])
    except Exception as e:
        exit("Could not connect to DB:\n%s" % e)

    # Update job status
    query = "UPDATE jobs SET Status='%s' WHERE Id=%s" \
        % ("Completed",jobResult["Id"])
    if (success == False):
       query = "UPDATE jobs SET Status='Failed',ErrorMessage='%s' WHERE Id=%s" \
           % (errorMessage,jobResult["Id"])

    # Execute query
    try:
        cur = dbConn.cursor()
        cur.execute(query)
        dbConn.commit()
    except Exception as e:
        print("WARNING: Could not update job status in DB:\n%s" % e)

    # Close database connection
    try:
        dbConn.close()
    except:
        pass
    
def saveJobResult(jobConfig, jobResult, fileName="result.json"):
    """saveJobResult Saves the result of the task execution
    
    Arguments:
        jobConfig {dict} -- Job configuration data
        jobResult {dict} -- Result of the function that was run by this process
        fileName {str} -- Name of the output file (default: {"result.json"})
    """

    # Get last dump filename
    fileName = getLastDumpFile(jobConfig, fileName)

    # Save job result to disk
    try:
        resultJson = json.dumps(jobResult,indent=4,sort_keys=True)
        with open(fileName,"w") as fd:
            fd.write(resultJson)
    except Exception as e:
        exit("ERROR: Could not save result:\n%s" % e)
    
    # Update job status
    updateJobStatus(jobResult)

def loadJobData(jobConfig, taskName=None):
    """loadJobData Loads all data for the given task
    
    Arguments:
        jobConfig {dict} -- Job configuration data
        taskName {str} -- Name of the task
    
    Returns:
        dict -- Dictionary with dataset names as keys and their respective data as values
    """

    # Load all datasets
    datasets = {}
    for ds in jobConfig["Datasets"]:
        # Use specified task name or the default one
        if (taskName != None):
            task = taskName
        else:
            task = ds["Task"]
            
        # Append task name to the dataset name, if it exists already
        dsName = ds["Name"]
        if ("." in dsName):
            dsName = "%s.%s" % (dsName,task)
        else:
            dsName = "%s.%s" % (task,dsName)

        # Load data
        try:
            dData = []
            with open("%s/%s/data/%s.csv" % (jobConfig["InputPath"],jobConfig["TaskName"],dsName),"r") as fd:
                for rawLine in fd.readlines():
                    line = list(map(float,rawLine.split(",")[:-1]))
                    dData.append(line)
            datasets[ds["Name"]] = dData
        except Exception as e:
            exit("ERROR: Could not read dataset '%s'\n%s" % (dsName,e))

    return datasets

# Task functions start here
def task_test(jobConfig):
    """task_test Test function for testing purposes only
    
    Arguments:
        jobConfig {dict} -- Full job configuration data
    
    Returns:
        dict -- Result of the test task
    """

    # Prepare results object
    res = {}
    res["Configuration"] = jobConfig
    res["Date"] = datetime.now().isoformat()

    # Dummy calculation
    x = range(10)
    y = [a**a for a in x]

    res["Calculation"] = y

    return res

def task_polyFit(jobConfig):
    """task_polyFit Calculate a polynomial fit using NumPy/SciPy
    
    Arguments:
        jobConfig {dict} -- Configuration data for the task
    
    Returns:
        dict -- Result of the polynomial fitting task
    """

    # Prepare results object
    res = {}
    res["Configuration"] = jobConfig
    res["Date"] = datetime.now().isoformat()

    # Read data from CSV file
    fileName = "%s/%s/data/%s.csv" % (jobConfig["InputPath"],jobConfig["TaskName"],jobConfig["Data"])
    xValues = []
    yValues = []
    try:
        with open(fileName,"r") as fd:
            for ln in fd.readlines():
                tokens = ln.split(",")
                if (len(tokens) > 1):
                    xValues.append(int(tokens[0]))
                    yValues.append(float(tokens[1].strip()))
    except Exception as e:
        exit("ERROR: Could not read input data from '%s':\n%s" % (fileName,e))

    # Perform polynomial fit
    try:
        degree = int(jobConfig["PolyDegree"])
        coeffs = np.polyfit(xValues,yValues,degree)
        poly = np.poly1d(coeffs)
        res["Function"] = poly
        res["Parameters"] = coeffs
    except Exception as e:
        exit("ERROR: Could not perform polynomial fit:\n%s" % e)

    # Create plot
    fig = plt.figure()
    ax1 = fig.add_subplot(1,1,1)
    ax1.set_title('Polynomial Fit')
    ax1.set_xlabel('X')
    ax1.set_ylabel('Y')
    ax1.ticklabel_format(axis="both", style="sci", scilimits=(0,0))
    ax1.scatter(xValues