#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_gif(out: BinaryIO, rng: BinaryIO):
    # magic number / header (6 bytes; 3 byte signature + data version 87a)
    out.write((b"GIF87a"))

    # logical screen descriptor (7 bytes);  1-2 width, 3-4 height, then
    # 5 flags[packed field]
    x = int.from_bytes(rng.read(2), "big") % 0xFFFF
    y = int.from_bytes(rng.read(2), "big") % 0xFFFF
    f = int.from_bytes(rng.read(1), "big") % (1 << 8)
    
    out.write((struct.pack(">H", x)))
    out.write((struct.pack(">H", y)))
    out.write((struct.pack("B", f)))

    # background color index (1 byte)
    b = int.from_bytes(rng.read(1), "big") % (1 << 8)
    
    out.write((struct.pack("B", b)))

    # pixel aspect ratio (1 byte) -- we don't care about this
    p = int.from_bytes(rng.read(1), "big") % (1 << 8)
    
    out.write((struct.pack("B", p)))

    # global color table flag (1 bit in packed field of next block)
    g = bool(int.from_bytes(rng.read(1), "big") & (1 << 7))

    # color resolution (3 bits in packed field of next block)
    c = int.from_bytes(rng.read(1), "big") & ((1 << 3)-1)

    # sort flag (1 bit in packed field of next block)
    s = bool(int.from_bytes(rng.read(1), "big") & (1 << 0))

    # size of global color table (3 bits in packed field of next block)
    n = int.from_bytes(rng.read(1), "big") & ((1 << 3)-1)

    # background color index (1 byte)
    b = int.from_bytes(rng.read(1), "big") % (1 << 8)

    # pixel aspect ratio (1 byte) -- we don't care about this
    p = int.from_bytes(rng.read(1), "big") % (1 << 8)

    # we have a global color table, so let's write it
    for i in range(0, n+1):
        # read a random RGB color
        r = int.from_bytes(rng.read(1), "big") % (1<<8)
        g = int.from_bytes(rng.read(1), "big") % (1<<8)
        b = int.from_bytes(rng.read(1), "big") % (1<<8)
        
        # print it to the output stream as a 24bit value
        out.write((struct.pack("BBB", r, g, b)))
    
    while True:
        # image separator (1 byte)
        out.write((struct.pack("B", 0x2c)))

        # image left position (2 bytes)
        x = int.from_bytes(rng.read(2), "big") % 0xFFFF
        out.write((struct.pack(">H", x)))

        # image top position (2 bytes)
        y = int.from_bytes(rng.read(2), "big") % 0xFFFF
        out.write((struct.pack(">H", y)))

        # image width (2 bytes)
        w = int.from_bytes(rng.read(2), "big") % 0xFFFF
        out.write((struct.pack(">H", w)))

        # image height (2 bytes)
        h = int.from_bytes(rng.read(2), "big") % 0xFFFF
        out.write((struct.pack(">H", h)))

        # local LZW minimum code size (1 byte)
        m = int.from_bytes(rng.read(1), "big") % (1 << 8)
        
        # global color table flag (1 bit in packed field of next block)
        g = bool(int.from_bytes(rng.read(1), "big") & (1 << 7))

        # color resolution (3 bits in packed field of next block)
        c = int.from_bytes(rng.read(1), "big") & ((1 << 3)-1)

        # sort flag (1 bit in packed field of next block)
        s = bool(int.from_bytes(rng.read(1), "big") & (1 << 0))

        # size of local color table (3 bits in packed field of next block)
        n = int.from_bytes(rng.read(1), "big") & ((1 << 3)-1)

        # background color index (1 byte)
        b = int.from_bytes(rng.read(1), "big") % (1 << 8)

        # pixel aspect ratio (1 byte) -- we don't care about this
        p = int.from_bytes(rng.read(1), "big") % (1 << 8)

        if not g:
            # no global color table, so let's write one
            
            # we currently only use one color for the entire GIF,
            # which is black, so we can skip over the whole section
            pass
                
        else:
            # global color table exists, so let's write it
            for i in range(0, n+1):
                # read a random RGB color
                r = int.from_bytes(rng.read(1), "big") % (1<<8)
                g = int.from_bytes(rng.read(1), "big") % (1<<8)
                b = int.from_bytes(rng.read(1), "big") % (1<<8)
                
                # print it to the output stream as a 24bit value
                out.write((struct.pack("BBB", r, g, b)))
        
        t = int.from_bytes(rng.read(1), "big") % (1 << 8)
        
        # loop counter
        while t != 0:
            k = int.from_bytes(rng.read(1), "big") % (1 << 8)
            lzw = []
            cs = k
            cc = 0
            ce = 0
            try:
                while cc < t:
                    e = int.from_bytes(rng.read(1), "big") % (1 << 8)
                    lzw.append(e)
                    cc += 1
                    if e == 0:
                        break
