#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():
    def __init__(self, output: BinaryIO):
        self.output: BinaryIO = output

        # Header (first 6 bytes)
        self.output.write(b'GIF87a')
        # Logical screen descriptor (next 7 bytes)
        self.output.write(struct.pack('<2H', 100, 100))
        # Global color map (next 7 bytes)
        self.output.write(bytes([0]*7 + [255]))
        # Application extension block (4 bytes)
        self.output.write(bytes([0] * 4))
        # Image data start code (first byte)
        self.output.write(bytes([0x2C]))
        # Image descriptor (next 9 bytes)
        self.output.write(struct.pack("<3H", 0, 0, 0))
        # Local color map (next 3 bytes)
        self.output.write(bytes([0] * 3))

    @staticmethod
    def next_digit() -> int:
        return ord(os.urandom(1).decode()) % 10

    @staticmethod
    def random_color() -> int:
        return ord(os.urandom(1).decode()) % 256

    @staticmethod
    def random_byte() -> int:
        return ord(os.urandom(1).decode()) % 256

    @staticmethod
    def random_sign() -> int:
        if os.urandom(1)[0] & 0x80 == 0:
            return -1
        else:
            return 1

    def write_constant(self, length: int, value: int):
        for i in range(length):
            self.output.write(bytes([value])[0].to_bytes(1, "big"))
            if i != length-1:
                self.output.write(bytes([RandomGifGenerator.random_byte()])[0].to_bytes(1, "big"))

    def write_random(self, length: int):
        for i in range(length):
            self.output.write(bytes([RandomGifGenerator.random_byte()])[0].to_bytes(1, "big"))
            if i != length-1:
                self.output.write(bytes([RandomGifGenerator.random_byte()])[0].to_bytes(1, "big"))

    def generate_block(self):
        # Block header
        self.write_constant(1, 0x21)
        
        block_type = RandomGifGenerator.next_digit()
        print("Block type is", block_type)
        if block_type == 1 or block_type == 2:
            # Extension block
            
            # Block length
            self.write_constant(1, 0xFF)

            if block_type == 1:
                # Plaintext extension
                
                # Extension label
                self.write_constant(1, 0xF9)

                # Text grid rows
                text_row_count = RandomGifGenerator.next_digit() + 1
                print("Text row count is", text_row_count)
                self.write_constant(1, text_row_count)

                # Text grid columns per row
                for i in range(text_row_count):
                    self.write_constant(1, RandomGifGenerator.next_digit()+1)

                # Color table flag
                self.write_constant(1, 0)

                # Background color index
                self.write_constant(1, RandomGifGenerator.next_digit())

                # Pixel aspect ratio
                self.write_constant(1, 0)

                # Text rows
                for i in range(text_row_count):
                    for j in range(RandomGifGenerator.next_digit()+1):
                        self.write_constant(1, RandomGifGenerator.next_digit())
                    
                # Block terminator
                self.write_constant(1, 0)
            elif block_type == 2:
                # Graphic control extension
                
                # Block length
                self.write_constant(1, 4)
                
                # Extension label
                self.write_constant(1, 0xF9)
                
                flags = 0x01 | (RandomGifGenerator.next_digit() << 1)
                if flags > 2**7 - 1:
                    raise Exception("Flags out of bounds")
                print("Flags are", flags)
                self.write_constant(1, flags)
                
                delay = 0
                while True:
                    d = RandomGifGenerator.next_digit()
                    delay *= 10
                    delay += d
                    if not (d >= 0 and d < 10):
                        break
                delay *= 10
                if (delay >> 16) > 2 ** 8 - 1:
                    raise Exception("Delay out of bounds")
                print("Delay is", delay)
                self.write_constant(2, delay)
                
                transparency = RandomGifGenerator.next_digit()
                print("Transparency is", transparency)
                self.write_constant(1, transparency)
                
                term = RandomGifGenerator.random_byte()
                print("Terminating with byte", term)
                self.write_constant(1, term)
        elif block_type == 3:
            # Comment extension
            
            # Block length
            comment_length = RandomGifGenerator.next_digit() + 1
            self.write_constant(1, comment_length + 1)
            
            # Extension label
            self.write_constant(1, 0xFE)
            
            # Comment contents
            self.write_random(comment_length)
            
            # Termination
            self.write_constant(1, 0)
        else:
            # Unknown block
            print("Unknown block type encountered")
            self.write_random(RandomGifGenerator.next_digit() + 1)

        # Terminating zero
        self.output.write(bytes([0]))

try:
    if len(sys.argv) > 2:
        raise Exception("Too many arguments")
    elif len(sys.argv) == 1:
        with open(DEFAULT_OUT