#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given pseudorandom number generator in binary format
def build_gif(out: BinaryIO, prng: str):
    header = b'GIF89a'  # File identifier
    width = 1024       # Image width (pixels)
    height = 768       # Image height (pixels)
    bytesPerPixel = 3  # Number of color components per pixel (red, green, blue)

    # Pixel aspect ratio - default to 1:1
    pixelAspectRatio = 0x0100
    
    # Global color table flag - true (1) for global colors
    gifGlobalColorTableFlag = True

    # Color resolution - one byte of color depth (bits per component)
    gifColorResolution = 0b111

    # Sort flag - unused
    gifSortFlag = False

    # Size of global color table - log2(size)
    nColorsInGlobalColorTable = 0b00111

    # Background color index - index of background color in global color table
    gifBackgroundColorIndex = 0xff

    # Pixel aspect ratio - ignored by most modern browsers, so set to 1:1
    gifPixelsPerUnitVert = 0x01
    gifPixelsPerUnitHorz = 0x01

    # Default to black and white
    gifBitsPerPrimaryTrans = 0
    gifBitsPerSecondaryTrans = 0

    # Unlimited delay time
    gifDelayTime = 0
    gifBlockTerminator = 0

    # Set up loop count field
    loopCountShort = 0
    loopCountLong = 0

    gifHeaderStructureDefinition = '!BBHBBBHHBBHB'
    gifDataStructureDefinition = '!BBBB'

    # Write GIF header structure
    header += struct.pack(
        gifHeaderStructureDefinition,
        
        # Screen descriptor fields
        width,
        height,
        pixelAspectRatio,

        # Packed fields
        (
            gifGlobalColorTableFlag << 7 |
            gifColorResolution << 4 |
            gifSortFlag << 3 |
            nColorsInGlobalColorTable
        ),

        gifBackgroundColorIndex,
        gifPixelsPerUnitVert,
        gifPixelsPerUnitHorz,

        # Packed fields
        (
            gifBitsPerPrimaryTrans << 4 |
            gifBitsPerSecondaryTrans
        ),

        gifDelayTime,
        gifBlockTerminator,

        # Loop count fields
        loopCountShort,
        loopCountLong,
    )

    # Write GIF data structure
    data = struct.pack(gifDataStructureDefinition, *[int(prng[i], 2) for i in range(len(prng))])
    out.write(header + data)
