#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


def main() -> int:

    if len(sys.argv) > 2:
        print('Usage: {} [filename]'.format(sys.path[0]))
        return 1
    elif len(sys.argv) == 2:
        fname = sys.argv[1]
    else:
        fname = DEFAULT_OUTPUT

    try:
        output = open(fname, mode='wb')
    except Exception as e:
        print("Couldn't open " + str(e))
        return 1

    rg = RandomGifGenerator(output=output)
    rg.generate()
    output.close()

    return 0


def generate_image(fn: BinaryIO) -> None:

    """
    Generate a random gif file from a randomly seeded source file.

    Arguments
    ---------
    fn : BinaryIO
      File containing seed data.

    Returns
    -------
    Image file with the following format:
      8-bit GIF Header 
      Logical Screen Descriptor
      Global Color Table
      Image Descriptor
      Local Color Table
      Image Data Block
    """
    
    rng = fn

    # GIF Header
    header = b'GIF89a'

    # Logical Screen Descriptor
    w = rng.read(1)                # width
    h = rng.read(1)                # height
    header += w + h

    # Color Table size
    ctsz = rng.read(1)[0] * 3
    for _ in range(ctsz):
        header += bytes([rng.read(1)[0]])

    # Block Size (Data Sub-Block within Data Stream)
    bs = rng.read(1)[0]         # block size is an unsigned char
    subblock = rng.read(bs - 1)  # discard sub-block marker
    header += bytes([bs])        # add block size back again
    data = rng.read(bs)          # get actual image data
    out.write(header + data)     # write it all together
    
