#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, size=20, count=8):
    # First we need to decide how many colors to use based on the bit depth.
    color_count = 1 << 8 // count
    # Now we can write the header for our GIF. We don't need to be accurate about
    # what these numbers mean, but it should look like a valid GIF file.
    print("GIF89a", end="", file=out)
    # Width and height are both set at 20
    print(struct.pack("<HH", size, size), end="", file=out)
    # Background color is black, global color table flags are set
    print("\x00\x00\xff\xff", end="", file=out)
    # Color count and block size are both set to 7 (we only have a few colors)
    print("\x07\x00", end="", file=out)
    # Write out each of the colors in turn. Note that we take them from the top
    # of the palette so they come out scrambled. This makes the output appear more
    # interesting.
    for i in range(color_count - 1, -1, -1):
        print(struct.pack("<BBB", rng.read(1)[0],
                          rng.read(1)[0], rng.read(1)[0]),
              end="", file=out)
    # Now we want to write an image descriptor block. This tells applications
    # where this frame starts so they can draw it. The x/y position is always
    # zero, meaning the image will always start at the top-left corner. The
    # width and height match the entire picture size.
    print(b"\x2c\x00\x00\x00\x00\x00\xff\xff\xff\x00", end="", file=out)
    # We're going to need a bit mask for the image data because it needs to be
    # repeated over multiple lines. It's just all ones so we repeat each pixel.
    mask = b"".join([chr(i).encode('ascii') * (size // 8 + 1) for i in [
        0b11111111] * count])[:size]
    # Finally we output the image data. Each line repeats itself multiple times.
    # So when you save this as a PNG, it looks like the original image repeated
    # multiple times over. You may notice some artifacting due to integer scaling.
    for i in range(count):
        d = "\x4c%s" % ((mask * (i + 1))[-size:]).decode('ascii')
        print((d.encode('ascii') if hasattr(d, 'encode') else d) * count,
              end="", file=out)
    print(b'\x3b', end='', file=out)
