#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the following heuristics:
#   1. The first 6 bytes of the file are the magic number "GIF87a".
#   2. The next byte is the width in pixels (random).
#   3. The next byte is the height in pixels (random).
#   4. The next byte is the number of colors present in the color table. One
#      of the first two possibilities:
#       * 0x00: no global color table, so all the images have their own.
#       * 0x03: one color table with 3 entries, which is used by all the
#               frames. Each entry has its red, green, and blue components set
#               to the same value, which should be between [0, 255].
#   5. The next byte is the amount of bits per pixel. Must be 8 for index-based
#      color schemes like this one.
#   6. The next two bytes contain some data we don't care about too much. It's
#      always zeroes, but it represents the amount of time before the animation
#      starts playing in units of hundredths of seconds.
#   7. The next byte is the background color. It must be an index into the
#      global color table.
#   8. There are two more bytes that we also ignore because we don't care about
#      the aspect ratio; they represent how many horizontal pixels there will be
#      per vertical pixel.
#
# From here on, the data is divided into chunks.
#
#   9. First chunk:
#       1. A single byte representing the size of the following byte array.
#          This value must be 0xfe since there are no extensions.
#       2. Next come two bytes representing the delay between this frame and the
#          previous one in units of hundredths of seconds.
#       3. The next byte is the image code, which must be 0xff.
#       4. The last byte is the transparent color index. It must be valid, so
#          it can't be 0 or 255.
#  10. Second chunk:
#       1. The location of the top left corner of the bounding box.
#       2. The size of the bounding box.
#  11. Third chunk:
#       * The header of the local color table.
#       * The content of the local color table.
#  12. Fourth chunk:
#       * LZW-encoded image data, starting from the top left corner of the
#         bounding box.
def generate(out: BinaryIO):
    out.write(b"GIF87a")
    out.write(struct.pack("BB", randint(), randint()))
    out.write(b"\x00\x00")

    colors = randint() % 2 + 1
    out.write(bytes([colors]))
    if colors == 0:
        out.write(b"\x08")
    else:
        out.write(
            b"""\
             \x03
             \x00\xc0\xff"""
        )

    out.write(b"\x08")
    out.write(b"\x00\x00\x00")
    out.write(struct.pack("B", randint() & ~1))
    out.write(b"\x00\x00")

    # First frame
    out.write(b"\x21\xf9\x04\x03\x00\x00\x00,\x00\x00\x00\x2c")
    out.write(b"\x00\x00\x00\x00\x00!\xff\x00,")
    out.write(b"\x00\x01\x00\x01\x00\x00")
    out.write(bytes([randint()]) * 3)
    out.write(b",\x00\x00\x00!")
    out.write(b"\xff\xff\xff\x00")
    out.write(b"\x2c")
    out.write(b"\x00\x00\x00\x00\x01\x00\x01\x00\x00")
    out.write(b"\x00")
    out.write(bytes([randint()]) * 3)
    out.write(b"\x21\xf9\x04\x03\x01\x00\x03\x00,\x00\x00\x00\x2c")
    out.write(b"\x00\x00\x00\x00\x00!\xff\x00,")
    out.write(b"\x00\x01\x00\x01\x00\x00")
    out.write(bytes([randint()]) * 3)
    out.write(b",\x00\x00\x00!")
    out.write(b"\xff\xff\xff\x00")
    out.write(b"\x3b")

    # Intervals between balls, all overlaid onto each other
    out.write(b"\x21\xfe\x04\x01\x00\x00\x00\x02\x00\x00\xff\xff\xff\xff")
    out.write(b"\x00\x00\x00\xcc\xce\xff\xcb\x00\x00\x00\x00")
    out.write(b"\x00\x01\x00\x01\x00\x00")
    out.write(bytes([randint()]) * 3)
    out.write(b"," + b"\x00\x00\x00!" + b"\xff\xff\xff\x00")
    for i in range(15):
        out.write(
            b"\x00"
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + b"\x00"
            + bytes[randint()]
            + bytes[randint()]
            + b"\x00"
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + b"\x00"
            + bytes[randint()]
            + bytes[randint()]
            + b"\x00"
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + b"\xff\x00"
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + b"\x01\x00"
            + bytes[randint()]
            + bytes[randint()]
            + b"\x00"
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + bytes[randint()]
            + b"\x