#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # If we're being called with no arguments, use standard in and out
    if out == None or rng == None:
        from sys import stdin as rng
        from sys import stdout as out
    
    # The first 6 bytes of any GIF are always identical. We don't care about
    # them at all; just skip over them.
    rng.read(6)

    # Now read the size header. This is 2 bytes for width and height, followed
    # by 1 byte for the colour depth (always 8 bits per pixel).
    width, height = struct.unpack('<HH', rng.read(4))

    # We can't write to a string because it's a fixed-size object, so let's
    # do our own buffering instead.
    image = [b'\0' * (width*height)]

    def write_buffered():
        # Write the buffered data to the real output stream.
        out.write(b''.join(image))
        # Reset the buffer.
        del image[:]
        
    # Read through each block in turn
    while True:
        # First byte gives the type of block. Block types have values between 0x00
        # and 0xff, inclusive.
        btype = ord(rng.read(1))

        # A value of 0x2c means "interlaced graphic". It also encodes the LZW
        # compressed pixel data.
        if btype == 0x2C:
            # Skip the next 4 bytes, which we already read above.
            rng.read(4)
            
            index = 0
            buf = [0]
            buflen = 1
            max_buflen = 5000
            chunk = 0

            # Keep reading until we reach an empty code (value 0xFF), which we
            # treat as EOF.
            while True:
                # Read the current bit/byte.
                bytelist = []
                # Read another byte.
                val = ord(rng.read(1))
                # Split into least and most significant bits.
                vallsb = val % 2
                valmsb = val // 2
                
                # Concatenate bit/byte value onto the end of the buffer
                buf.append(valmsb)
                buflen += 1
                # Write any full chunks to disk
                while (chunk + 1) * 8 >= buflen:
                    # Get the chunk starting at this point
                    chunkval = struct.unpack('B', bytes(buf[index+chunk*8 : index+(chunk+1)*8]))[0];
                    # Add this chunk to the image data
                    image.append([chunkval])
                    # Increment the chunk offset
                    chunk += 1
                    
                # If the buffer is getting too long, start writing it out.
                if buflen > max_buflen:
                    # Write out the buffer until we get back down to half the maximum length.
                    while buflen > max_buflen / 2:
                        write_buffered()
                    # Shrink the buffer so that we're only keeping the last half of it around.
                    # We do this by creating a new list containing just the last half of the
                    # original, and copying that across.
                    image.extend(bytes(buf[-max_buflen//2:]))
                    del buf[:]
                    buflen -= max_buflen - max_buflen // 2
                    index = 0
                    chunk = 0

                # If we hit an empty code (value 0xFF), we must have finished decoding the pixels.
                if val == 0xFF:
                    break

                # Create a dictionary entry for the two-byte sequence just read out.
                newdictentry = buf[index:] + [valmsb]
                # Check if there's an existing entry that starts with the same thing as what
                # we just read out.
                match = False
                while len(newdictentry) <= 12:
                    olddictentry = bytes(newdictentry)
                    # Find the position of the matching prefix in the dictionary.
                    pos = [i for i, x in enumerate(dictdata) if x.startswith(olddictentry)]
                    if len(pos) != 1:
                        break
                    else:
                        pos = pos[0]
                        
                    dictdata.insert(pos+1, newdictentry)
                    index += newdictentry[1:] == dictdata[pos][1:]
                    newdictentry = [0]*(newdictentry[0]+1) + newdictentry[1:]
                    if index >= len(dictdata):
                        # We found a match! Replace the last part of the buffer with the
                        # rest of the entry.
                        chunk = (index - 1) % (max_buflen//8)
                        buf = dictdata[index][index:]
                        buflen = len(buf)
                        index = 0
                        match = True
                        break

                # If there was no exact match, create a new one.
                if not match:
                    dictdata.append(newdictentry)

                # If we've got enough entries, remove some obsolete ones.
                if len(dictdata) > 4097:
                    dictdata.pop(0)

            # There should be padding at the end of the final byte.
            if buflen & 7 != 0:
                raise Exception("Padding required")
            write_buffered()

        # Otherwise, if the block type is anywhere between 0x01 and 0x80, they
        # contain metadata related to the image. Skip these blocks.
        elif btype >= 0x01 and btype <= 0x80:
            rng.read(struct.unpack("<H", rng.read(2))[0])
