#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    palette_size = 2 ** (int.from_bytes(rng.read(1), 'big'))
    background_palette_index = int.from_bytes(rng.read(1), 'big')

    num_frames = 0
    while True:
        try:
            frame_duration = int.from_bytes(rng.read(1), 'big')
        except EOFError:
            break
        if not frame_duration > 0:
            continue

        total_frame_delay = 0
        for i in range(frame_duration):
            delay = int.from_bytes(rng.read(1), 'big')
            total_frame_delay += delay
            if delay == 0xff:
                break
        else:
            raise RuntimeError("Loop contains only FFs")

        # XXXXX Gif frames have a max size of 65535 bytes.
        # Theoretically this can be extended to use multiple
        # extensions with more data blocks, but it's probably
        # not worth trying to exhaustively handle all cases.
        extra_data_len = int.from_bytes(rng.read(1), 'big')
        assert extra_data_len <= 0xff
        
        block_type = int.from_bytes(rng.read(1), 'big')
        assert block_type == 0xfe # graphic control extension
        block_length = int.from_bytes(rng.read(1), 'big')
        flags = int.from_bytes(rng.read(1), 'big')
        delay = int.from_bytes(rng.read(2), 'little')
        
        palette_info_block_i = 0
        while True:
            try:
                block_type = int.from_bytes(rng.read(1), 'big')
            except EOFError:
                break
            
            if block_type == 0xfe: # graphic control extension
                # skip for now
                block_length = int.from_bytes(rng.read(1), 'big')
                skip = 3 + block_length - 1
                rng.seek(skip, 1)
                
                if palette_info_block_i < len(palette_info):
                    palette_info_block_i += 1
                    continue
                else:
                    raise ValueError("Unexpected graphic control extension "
                                     "in palette info")
            elif block_type != 0xf9: # application extension
                # Unexpected extension type
                assert False

            block_length = int.from_bytes(rng.read(1), 'big')

            if palette_info_block_i >= len(palette_info):
                raise ValueError("Too many palette info blocks ({})"
                                 .format(palette_info_block_i))
            palette_info[palette_info_block_i].append(block_length)
            for _ in range(block_length):
                byte = int.from_bytes(rng.read(1), 'big')
                palette_info[palette_info_block_i][-1] <<= 8
                palette_info[palette_info_block_i][-1] |= byte & 0xff

            if palette_info_block_i < len(palette_info):
                palette_info_block_i += 1

        bytes_remaining = 2 * (image_width * image_height)
        def read_next_byte():
            """Returns next byte from rng."""
            nonlocal bytes_remaining
            bytes_remaining -= 1
            return int.from_bytes(rng.read(1), 'big')

        lzw_minimum_code_size = rng.read(1)[0] # TODO
        assert lzw_minimum_code_size <= 2**bit_depth

        rng.reset()
        d = dict()
        code_mask = (1<<lzw_minimum_code_size)-1
        code_shift = bit_depth+1
        clear_code = 1<<bit_depth
        end_of_information = clear_code + 1
        output = 0
        running_state = 0
        stack = []

        block_id = 0
        while bytes_remaining > 0:
            if block_id:
                block_id -= 1
            else:
                block_sz = rng.read(1)[0]
                block_id = block_sz
            if block_id > 0:
                bytes_remaining -= 1
                continue

            if running_state == 0:
                d = dict()
                code_mask = (1<<lzw_minimum_code_size)-1
                code_shift = bit_depth+1
                clear_code = 1<<bit_depth
                end_of_information = clear_code + 1
                output = 0
                running_state = 1
            elif running_state == 1:
                if output:
                    stack.append((output, len(stack)))
                    output = 0
                if code_mask == 0xffff:
                    code_mask += 1
                    code_shift += 1
                code = read_next_byte()
                if not code:
                    raise RuntimeError("Invalid state 1")
                while code <= code_mask:
                    if code == clear_code:
                        code_mask = (1<<lzw_minimum_code_size)-1
                        code_shift = bit_depth+1
                        clear_code = 1<<bit_depth
                        end_of_information = clear_code + 1
                        output = 0
                        break
                    elif code == end_of_information:
                        running_state = 2
                        break
                    elif not code in d:
                        # Failure
                        running_state = 0
                        break
                    else:
                        key, value = d[code]
                        # Build up current string character by character
                        if isinstance(value, tuple):
                            new_key = (key+(value[0],), value[1])
                            d[code_mask+1] = new_key
                            value = key[-1:]
                        else:
                            value = chr(value)
                            
                        output = ord(value)+(output<<8)
                        code = read_next_byte()
                        if code == end_of_information:
                            running_state = 2
                            break
                        elif code > code_mask:
                            break

            if running_state == 2:
                if not output or not stack:
                    raise RuntimeError("Output underflow")
                ok = True
                while ok:
                    last_entry = stack.pop()
                    ok = bool(last_entry[1])
                    if ok:
                        output = last_entry[0]+(output<<8)
                        if stack:
                            entry = stack.pop()
                            stack.append((entry[0]+(chr(output&0xff),), output>>8))
                            output >>= 16
                        else:
                            data = [(output&0xff)]
                            ok = False
                stack.append(((chr(output&0xff),), output>>8))
                running_state = 1

        assert bytes_remaining == 0
        assert running_state == 2

        # Write GIF header
        out.write(header)

        # Write logical screen descriptor
        out.write(struct.pack('<HHBBB',
            image_width,
            image_height,
            0, # global color table flag
            bit_depth-1, # bit depth minus 1
            background_color))

        # Write global color table
        for entry in color_table:
            out.write(struct.pack('<BBB', *entry))

        # Write palette information extension
        palette_ext_block_start = b'\x21\xff\x0b'
        palette_ext_block_end = b'\x00'
        palette_ext_frame_delay = b'\xF9'
        palette_ext_frame_delay += struct.pack('<H', delay)
        palette_ext_frame_delay += b'\xFF'
        palette_ext_trans