#!/usr/bin/env python3

import struct
from typing import List, Tuple, Optional, Callable

import rospy
import sensor_msgs.msg as msgs
import nav_msgs.srv as srv
from geometry_msgs.msg import PointStamped  # For testing.
import tf2_ros
import tf2_geometry_msgs
from std_srvs.srv import Empty

from .common import is_close, get_transform, quat_to_euler, euler_to_quat

# A small timeout for the transform lookups.
TRANSFORM_TIMEOUT = rospy.Duration(0.1)

# Publishers and subscribers to the ROS topic system. The publisher is for
# sending the pointcloud out to other nodes. The subcriber is so this node can
# receive updates to its (x, y, z) position relative to the frame specified in
# `set_frame()`.
pcl_pub: Optional[rospy.Publisher] = None
tf_sub: Optional[rospy.Subscriber] = None


def _publish_pointcloud(data: bytes):
    """Publishes a point cloud with information about obstacles."""
    pcs: Optional[msgs.PointCloud2] = None
    try:
        pcs = msgs.convert_point_cloud2_to_message(
            data, stamp=rospy.Time().now(), frame_id=get_transform().child_frame_id)
    except Exception as e:
        rospy.logwarn("Failed to convert {} to message type: {}".format(repr(data), repr(e)))
    if pcs is not None:
        pcl_pub.publish(pcs)


def set_cloud_topic(name: str):
    """Sets the name of the topic to be published."""
    global pcl_pub
    pcl_pub = rospy.Publisher(name, msgs.PointCloud2, queue_size=10)


def set_pose_update_callback(func: Callable[[float, float, float], None]) -> bool:
    """Sets the callback function that gets called when the robot's pose changes.

    Returns whether or not the callback was successfully registered.
    This will fail if another callback has already been registered."""
    def cb(msg: PointStamped):
        func(msg.point.x, msg.point.y, msg.point.z)
    global tf_sub
    if tf_sub is None or not tf_sub.is_registered():
        tf_sub = rospy.Subscriber('/map', PointStamped, cb)
        return True
    return False


def clear_pose_update_callback() -> bool:
    """Un-sets the callback that was used to notify the client about the robot's
    pose change."""
    if tf_sub is not None and tf_sub.is_registered():
        tf_sub.unregister()
        tf_sub = None
        return True
    return False


def get_transform(timeout: Optional[float] = TRANSFORM_TIMEOUT) -> geometry_msgs.TransformStamped:
    """Returns the transform from the frame specified in `set_frame()` to the
    ROS TF frame.

    If no frame is specified, the transform defaults to the ROS TF frame.

    By default, this call waits up to one second before giving up trying to find
    the requested transform. To check your code against the simulator, you can
    use negative values to wait forever.
    """
    return tf2_ros.Buffer().lookup_transform(
        'base_footprint', get_frame(), timeout, TRANSFORM_TIMEOUT)


def get_frame() -> str:
    """Gets the current ROS TF frame.

    Setting the ROS TF frame is done with the service call `/set_frame`.
    """
    return rospy.get_param('~frame') or '/odom'


def set_frame(frame: str, timeout: Optional[float] = TRANSFORM_TIMEOUT) -> bool:
    """Sets the name of the ROS TF frame to use for all future transform calls.

    Note that this does not actually do anything separate from calling into the
    standard TF2 system like any other user would. This API is mostly useful for
    the Simulated Proximity Sensor class.
    """
    assert timeout is None or timeout > 0.0
    rospy.set_param('~frame', frame)
    return get_transform(timeout).header.stamp != rospy.Time(0)


def is_collision_topic_available(topicName: str) -> bool:
    """Checks whether or not the given collision topic exists in the current ROS
    graph."""
    assert topicName.startswith("/")
    _, _, tt = rospy.get_master().lookupTopic(topicName)
    return len(tt) > 0

