#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # The size of the GIF header. See https://en.wikipedia.org/wiki/GIF#File_format
    const_header = b"GIF89a"

    # This is a Netscape 2.0 Application Extension (NSEA).
    # It is mandatory for animated GIFs. More info at
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    nsea = (b"\x21\xFF\x0BNETSCAPE2.0\x03\x01", 0)

    # Set GCE to None so we can detect when we need to write it later on.
    gce = None
    palette = []
    pixels_written = False

    while True:
        id = int.from_bytes(rng.read(1), 'big') % 65537

        if id == 65534:
            # This is an EOF block
            break
        
        elif id == 65535:
            # This is a pause for this amount of units
            delay = int.from_bytes(rng.read(2), 'big') * 10
            
            # Set GCE to None so we can detect when we need to write it later on.
            gce = None

        else:
            # This is a graphic control extension (GCE).
            # The first octet must be 0xf9.
            if not int.from_bytes(rng.read(1), 'big') & 0xF9:
                continue
                
            # Delay time is in 10ms units.
            # If 0, use the previous value.
            delay = int.from_bytes(rng.read(2), 'big' ) or nsea[1]
            
            # The last byte defines whether transparency is being used and what color index represents transparent pixels.
            flags = int.from_bytes(rng.read(1), 'big')
            
            # If the least significant bit of the third byte is set to 1, then all subsequent pixels will be transparent until
            # either another GCE block is encountered with the same value for the transparent color index, or the next image
            # descriptor.
            if flags & 1:
                # Set GCE to None so we can detect when we need to write it later on.
                gce = None
            else:
                gce = ((flags >> 2) & 0xff, int.from_bytes(rng.read(1), 'big'))

            # Skip any remaining bytes, including a terminator.
            rng.read(int.from_bytes(rng.read(1), 'big'))
            
        # Read the image dimensions
        width = int.from_bytes(rng.read(2), 'big') + 1
        height = int.from_bytes(rng.read(2), 'big') + 1

        # Read the local color table flag. If 1, there is a local color table.
        local_color_table_flag = bool(int.from_bytes(rng.read(1), 'big'))
        
        # If there is no local color table, we reuse the global one from above
        if not local_color_table_flag:
            local_palette = []
        else:
            # Otherwise, read the color table length - 1
            # because we add the background color below.
            entries = 1 << (int.from_bytes(rng.read(1), 'big'))

            # We add the background color ourselves here because we might not have written this out earlier.
            local_palette = [(0, 0, 0)]

            for _ in range(entries):
                red = rng.read(1)[0]
                green = rng.read(1)[0]
                blue = rng.read(1)[0]

                local_palette.append((red, green, blue))

        # Set the global color table to our local color table
        palette = local_palette

        # Read the interlacing flag
        interlace_flag = bool(int.from_bytes(rng.read(1), 'big'))

        # Finally, write the GIF header, size, etc.
        # Note that the entirety of this is generated from random data...
        out.write(const_header +
                  width.to_bytes(2, 'big', signed=False) +
                  height.to_bytes(2, 'big', signed=False) +
                  b'\x87\xcc\x00\x00' +
                  str.encode(f"{delay}ms") +
                  b'\x00'*6 +
                  (nsea[0]+struct.pack('<H', delay)).ljust(18, b'\x00') +
                  b'\x2c' +
                  b'\x00'*3 +
                  len(local_palette).to_bytes(1, 'big', signed=False) +
                  b'\x00')
        for color in local_palette:
            out.write(bytearray(color))

        # Write a frame.
        # There are two types of frames: uncompressed and compressed.
        method = rng.read(1)
        if method != b"\x01":
            out.write(method+b"\x00"+rng.read(width*height)+b"\x00")
        else:
            out.write(method)
