#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
        
        with open('/dev/urandom', mode='rb') as rng,\
             open('header.gif', mode='rb') as header,\
             open('palette.gif', mode='w+b') as palette:

            # read the header bytes and write them to the palette file
            header_bytes = bytearray(struct.unpack('<784B', header.read()))
            palette.write(header_bytes)
        
            # read from /dev/urandom until we get enough bytes
            while len(header_bytes) < 256 * 3:
                header_bytes += bytearray(rng.read())
    
            # insert into the palette file
            for b in range(0, 256*3, 3):

                rgb = [
                    chr((header_bytes[b] & 0b11111) << 3 | (header_bytes[(b+1)] >> 5)), 
                    chr(((header_bytes[(b+1)] & 0b11111) << 3 | (header_bytes[(b+2)] >> 5)))
                ]
                
                palette.seek(b + 9)
                palette.write(rgb[0].encode() + rgb[1].encode())
                
            
            # make a gif palette image using the contents of the palette file
            palette.seek(0)
            palette_bytes = bytearray(palette.read().replace('\n', '').strip())
        
            # create an empty GIF according to the header
            gif : BinaryIO
            gif = open(self.output, "wb")

            gif.write(header_bytes[:6])    # Logical screen width         - unsigned short   big endian
            gif.write(header_bytes[6:8])   # Logical screen height        - unsigned short   big endian
            gif.write(header_bytes[8:10])  # Packed field                 - packed fields    
            gif.write(header_bytes[10:11]) # Background color index       - fixed palette entry 
            gif.write(header_bytes[11:12]) # Pixel aspect ratio           - byte             no transformation

            gif.write(palette_bytes)      # Global color table          - RGB triples      sorted by importance
            gif.flush()

            # Frames --------------------------------------------------------
            frame_count = 0
            frame_size = int.from_bytes(header_bytes[10:12], byteorder='big', signed=False)

            while True:

                frame = rng.read(frame_size)

                if not frame or len(frame)!= frame_size: break

                # print(f"Frame #{frame_count}")

                lzw_mincodesize = struct.pack("<B", next(iter([b for b in frame])))
                data_subblock = frame[len(lzw_mincodesize):-1]
                end_of_image = bytes([(int(data_subblock[-1])*2)])
                crc_value = frame[-1:]

                size = ((len(lzw_mincodesize) +\
                         len(data_subblock)) // 256) +\
                      (((len(lzw_mincodesize) +\
                         len(data_subblock)) % 256) > 0)

                blocktype = 0x2C.to_bytes(length=1, byteorder="big", signed=False)
                subblocksize = str(size).zfill(2).encode("ascii")

                # Image descriptor ---------------------------------------------
                position = 0

                img_left = 0
                img_top = 0
                img_width = 100
                img_height = 100
                img_control = 0b00000000

                gctsize = header_bytes[10] & 0b00001111
                lctsize = max(gctsize, 2**(next(iter([b for b in frame]))+1))
                interlace = (header_bytes[8] & 0b10000000) >> 7
                sortflag = (header_bytes[8] & 0b01000000) >> 6
                
                imgdesc = struct.pack("<HHBBB",
                                      img_left,
                                      img_top,
                                      img_width,
                                      img_height,
                                      img_control)

                # Local color table                                         
                lctwhite = chr(0xFF).encode() + chr(0xFF).encode() + chr(0xFF).encode()
                lctblack = chr(0x00).encode() + chr(0x00).encode() + chr(0x00).encode()
                lctred = chr(0xFF).encode() + chr(0x00).encode() + chr(0x00).encode()
                lctgreen = chr(0x00).encode() + chr(0xFF).encode() + chr(0x00).encode()
                lctblue = chr(0x00).encode() + chr(0x00).encode() + chr(0xFF).encode()
                lctyellow = chr(0xFF).encode() + chr(0xFF).encode() + chr(0x00).encode()
                lctorange = chr(0xFF).encode() + chr(0xA5).encode() + chr(0x00).encode()
                lctpurple = chr(0x80).encode() + chr(0x00).encode() + chr(0x80).encode()

                colors = {
                    2**i : lctwhite, 
                    2**i+1 : lctblack, 
                    2**i+2 : lctred, 
                    2**i+3 : lctgreen, 
                    2**i+4 : lctblue, 
                    2**i+5 : lctyellow, 
                    2**i+6 : lctorange, 
                    2**i+7 : lctpurple
                }

                localcolortable = b"" 
                for i in range (lctsize):

                    if i < 9:
                        localcolortable += colors[2**i]

                        if i == 8:
                            continue
                    
                    localcolortable += random.choice(colors.values()).decode()
                    
                # Sub blocks ---------------------------------------------------
                # LZW minimum code size                                      
                lzwmsb = lzw_mincodesize

                subblock = lzwmsb + data_subblock + end_of_image + crc_value
                length = len(subblock)
                size = ((len(subblock))/256) + (len(subblock)%256>0)

                # Application extension ----------------------------------------
                app_extension_label = b'\x21\xF9\x04'
                # Netscape application comment                               
                netscapeappcomment = b'\x01\x00\x00\x0A'
                # Loop count subblock                                        
                loopcount = b'\x00'
                # Block terminator                                          
                blockterm = b'\x00' 

                layerandblocksize = \
                    struct.pack(">BBH", 
                        blocktype[position], subblocksizelsb[position], length)[position:position+1]+\
                    (size.to_bytes(length=1, byteorder="little", signed=False))[position:position+1]

                appext = \
                    struct.pack(">BBH", 
                        blocktype[position], subblocksizelsb[position], len(netscapeappcomment)+len(loopcount)+len(blockterm))[position:position+1]+\
                    netscapeappcomment+\
                    loopcount+\
                    blockterm
                
                # Graphics control extension ------------------------------------
                graphext = \
                    struct.pack(">BBBBBBHB", 
                                blocktype[position],
                                subblocksizelsb[position],
                                0x0b,
                                0,
                                0,
                                0,
                                0,
                                0,
                                1)\
                                [position:position+1]+\
                    (interlace.to_bytes(length=1, byteorder="big", signed=False))[position:position+1]+\
                    (sortflag.to_bytes(length=1, byteorder="big", signed=False))[position:position+1]+\
                    b"\x00"+\
                    (gctsize.to_bytes(length=1, byteorder="big", signed=False))[position:position+1]
                
                gif.write(layerandblocksize)
                gif.write(imgdesc)
                gif.write(localcolortable)

                if position == 0: 
                    # Extension blocks are optional but must be at least one block long. 
                    gif.write(appext)
                    gif.write(graphext)

                gif.write(subblock)

                frame_count += 1


            # Trailer -------------------------------------------------------
            trailer = b"\x3B"
            gif.write(trailer)

                
            gif.close()


if