#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    # This should always be zero for our purposes
    screenwidth = 0
    screenheight = int.from_bytes(
            rng.read(2), byteorder='big')

    blocktype = int.from_bytes(
            rng.read(1), byteorder='big') | (
                int.from_bytes(
                    rng.read(1), byteorder='big') << 8)

    if blocktype!= 768:
        raise ValueError('Invalid GIF: expected application extension')

    extlabel = b'\x21\xFF\x0BNETSCAPE2.0'
    extblocksize = len(extlabel) + 3
    extdata = bytes([3]) + b'\x00' * 254
    extbody = extlabel + struct.pack('<H', extblocksize) + extdata

    framelen = 10 + extblocksize + 1
    framecount = 1
    loopcount = 0
    frames = []

    while True:

        aspectratio = 0
        colorres = 10
        pixelaspect = 0

        framedelaytime = int.from_bytes(
                rng.read(2), byteorder='big')
        displxoffset = int.from_bytes(
                rng.read(2), byteorder='big')
        displyoffset = int.from_bytes(
                rng.read(2), byteorder='big')
        width = int.from_bytes(
                rng.read(2), byteorder='big')
        height = int.from_bytes(
                rng.read(2), byteorder='big')
        
        localcolortable = [0] * ((
            2**(int(colorres & 7) + 1))**3)
        globalcolortable = [0] * 256

        bgindex = None
        for i in range(0, 256):
            localcolortable[i*3+0] = \
                int.from_bytes(
                    rng.read(1), byteorder='big')
            localcolortable[i*3+1] = \
                int.from_bytes(
                    rng.read(1), byteorder='big')
            localcolortable[i*3+2] = \
                int.from_bytes(
                    rng.read(1), byteorder='big')

            if localcolortable[i*3 : i*3+3] == [0]*3:

                # Background index is determined by the first entry that is all zeros
                if not bgindex and len(localcolortable)!= i:
                    bgindex = i

            elif any([c > 255 for c in localcolortable[i*3 : i*3+3]]):

                # Global colortable is populated with entries from local
                # colortable that have values above 255, as required to make
                # every value up to at least 256

                for j in range(0, min((
                    2**(int(colorres & 7) + 1))**3 - i, (
                        len(globalcolortable) - len(localcolortable)))):

                    ind = len(localcolortable) + j 
                    globalcolortable[ind] = localcolortable[ind*3+0]
                    globalcolortable[ind+1] = localcolortable[ind*3+1]
                    globalcolortable[ind+2] = localcolortable[ind*3+2]


        lzwmincodesize = int.from_bytes(
                rng.read(1), byteorder='big')

        lzwcodesize = lzwmincodesize + 1
        # We'll represent LZW codes as tuples of size two, so we can easily
        # identify when we need to add more bits to an already-full tuple
        lzwcodes = [(b'', 256)]
        lzwpixmap = []

        for row in range(0, height):
            lzrwunencodedrow = []
            rawpixmap = []
            
            for col in range(0, width):
                codestr = ''
                while len(lzwcode := lzwcodes[-1]) < lzwcodesize:
                    codebyte = ord(
                            rng.read(1).decode())
                    lzwcodes.append(
                            (lzwcode[0] + chr(codebyte % 256),
                             codebyte // 256))
                    codestr += chr(codebyte)
                
                pixval = lzwcode[1]
                rawpixmap.append(pixval)

                lzwcodes[:] = [l for l in lzwcodes if l[1]!= pixval or l[1]]
                
                lzrwunencodedrow.append(codestr)
        
            lzwpixmap.append(lzrwunencodedrow)

        data = b''
        data += struct.pack(
                '<HHBBBBBHBHH',
                width,
                height,
                0xf9, # Block type 249
                0xff, # Extension label
                6, # Block length
                *(
                    bitfield := [
                        0, # No transparency
                        0, # Sorted colors
                        0, # Reserved field
                        0, # No glitter
                        0, # User input flag
                    ]
                )
            )
        data += struct.pack('<BBBBBBBBH', *bitfield, lzwmincodesize, 0)
        data += extbody
        data += struct.pack('<HH', 0x2C00, framelen)
        data += struct.pack('<HHHHHHHB',
                width,
                height,
                displxoffset,
                displyoffset,
                localcolortable[bgindex*3],
                localcolortable[(bgindex+1)*3],
                localcolortable[(bgindex+2)*3],
                0,
                0xf9)
        data += struct.pack('<B', lzwmincodesize + 1)
        for row in lzwpixmap:
            for item in row:
                data += item.encode()
 
        data += b'\x3B'
        data += struct.pack('<HBH', 0x21F9, 4, framelen - 2)
        data += struct.pack('<L', framedelaytime * 10)
        data += struct.pack('<BB', 0, loopcount)
        data += struct.pack('<B', 0)

        frames.append({
          'screenwidth': screenwidth,'screenheight': screenheight, 
            'width': width, 'height': height, 'rawpixmap': rawpixmap})
    
        if framecount >= datalen: break
        framecount += 1

    out.write(header + data)
