#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):
        self.output = DEFAULT_OUTPUT if not output else str(output)


    @staticmethod
    def help() -> None:

        print('''Usage: python3 {} [OPTIONS]
Randomly generates a GIF with given options.
Options are:
  --help      Prints this message.
  --output    Specify name of generated GIF file
              Default: "{}"'''.format(__file__, DEFAULT_OUTPUT))



    def run(self):
        
        try:
            args = iter(sys.argv[1:])

            while arg := next(args):

                if   arg == '--help':
                    RandomGifGenerator.help()
                    break

                elif arg == '--output':
                    self.output = str(next(args))
                
                else:
                    raise ValueError('Invalid option.')


        except Exception as e:
            print(e)
            RandomGifGenerator.help()
            exit(-1)


        finally:
            
            # Use /dev/urandom to create a new random number seed.
            # This makes sure that calling this script repeatedly will
            # always give different results.
            with os.popen('/dev/urandom', 'rb') as rng,\
                 open(self.output,    mode='wb') as out:

            return generate_random_gif(
                    rng, out
                )



# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# correspond to the information found here:
# https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Signature
    header = b'GIF89a'
    
    # Width and Height
    w = rng.read(2)   # width is 16 bit
    h = rng.read(2)   # height is 16 bit
    header += w + h

    # Color Table size
    ctsz = rng.read(1)[0] * 3
    for _ in range(ctsz):
        header += bytes([rng.read(1)[0]])

    # Block Size (Data Sub-Block within Data Stream)
    bs = rng.read(1)[0]         # block size is an unsigned char
    subblock = rng.read(bs - 1)  # discard sub-block marker
    header += bytes([bs])        # add block size back again
    data = rng.read(bs)          # get actual image data
    out.write(header + data)     # write it all together
    
