#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    gif_header = bytearray()
    gif_header += header.encode('ascii')\
                  + bytes([image_width >> 8,
                           image_width & 0xFF])\
                  + bytes([image_height >> 8,
                           image_height & 0xFF])

    palette = [byte for byte in rng.read((3 * 256))]
    color_table_entry = lambda index: (palette[(index << 1)] << 8)\
                                     | (palette[(index << 1) + 1])
    color_table_size = len(palette) // 3

    local_color_table_flag = ord('L') if color_table_size > 7 else 0x00
    table_descriptor = (local_color_table_flag   |\
                        ((color_table_size - 1) >> 1) |\
                        0x80).to_bytes(1, 'little')


    # construct global color table
    global_color_table = []
    for i in range(0, (color_table_size ** 2), 3):

        entry = palette[i] << 16 | palette[i+1] << 8 | palette[i+2]

        # duplicate colors twice to save space
        for j in range(0, int(((color_table_size / 3)**2)*2)):
            global_color_table.append(entry)

    end_of_file = b';'.encode('ascii')


    # write all blocks upfront
    data = b''
    data += gif_header
    data += b'\xF7' # application extension block label
    data += bytes([0x0B]) # length of block
    data += b"NETSCAPE2.0" # app name
    data += b"\x03" # app version
    data += b"\x01" # sub-block size
    data += b"\x00" # loop count
    data += b'\x00' # terminator
    data += table_descriptor
    data += global_color_table
    data += end_of_file

    # add one frame per iteration
    while True:

        # control block
        data += b'\x2C' # image separator
        data += struct.pack('<HHBBBBB',
                            image_width, image_height,\
                            0b00000,     # no local color table, interlace off
                            0x00,        # default background color index
                            0x00,        # pixel aspect ratio
                            0x00         # no disposal specified
                          )


        # add transparent black as last color in color
        # table so it is available immediately
        transparency_idx = color_table_size**2 // 3
        palette[-1] = 0x00
        palette[-2] = 0x00
        palette[-3] = 0x00

        for y in range(image_height):

            # RLE encoding
            run_length = 0
            prev_pixel = None

            for x in range(image_width):

                # remap pixels
                pixel = ord(rng.read(1)[0]) % 256


                # first pixel as literal
                if not prev_pixel or prev_pixel == pixel:
                    data += pixel.to_bytes(1, 'little')

                    # reset run-length counter when we have a new pixel
                    run_length = 1
                    prev_pixel = pixel

                # next pixels as repeater
                elif run_length < 0xFF:
                    data += (run_length - 1).to_bytes(1, 'little')
                    data += pixel.to_bytes(1, 'little')


                    # reset run-length counter with new pixel
                    run_length = 1
                    prev_pixel = pixel

                # maximum repetition reached, store this
                # pixel directly and trigger flushing
                else:
                    data += b'\x00'
                    data += pixel.to_bytes(1, 'little')


            # flush remaining literals
            if run_length >= 2:
                data += (run_length - 1).to_bytes(1, 'little')



        # restore original palette that was modified
        del palette[-1:]

        out.write(header + data)
