#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Script to end a period of time in the database."""
import argparse
from datetime import date, timedelta

from pytz import timezone

from wanderu.aws import get_dynamodb_table
from wanderu.config import load_config


def main(config=None):
    """Ends a period of time for all non-ended time slots ending before today's
    date.

    :param config: Configuration dictionary containing'region',
                   'dynamodb_table' and 'time_zone'. If not provided this will be
                   loaded from ~/.wanderu/config or /etc/wanderu/config depending on whether running as root user or not (see `load_config()`).
    :type config: dict
    """
    if config is None:
        config = load_config()
    dt = datetime.now() - timedelta(days=1) # yesterday at midnight
    ay, am, ad = str(dt).split(" ")[:3]
    ay = int(ay)
    am = int(am) + 24 * (int(ad)-1)
    enddate = date(year=ay, month=am, day=1).strftime("%Y-%m-%d")
    table = get_dynamodb_table(config['dynamodb_table'], region_name=config['region'])
    response = table.scan(Limit=500)
    data = response["Items"]
    while "LastEvaluatedKey" in response:
        print("Continuing scan...")
        response = table.scan(ExclusiveStartKey=response["LastEvaluatedKey"], Limit=500)
        data += response["Items"]
    print("{} items found...".format(len(data)))
    with table.batch_writer() as batch:
        for item in data:
            try:
                tzinfo = timezone(item["timezone"])
            except KeyError:
                continue

            e_start_time = item["endtime"].astimezone(tzinfo)
            s_start_time = item["starttime"].astimezone(tzinfo)
            start_day = s_start_time.isoweekday() % 7
            end_day = e_start_time.isoweekday() % 7


            if ((e_start_time < dt)
                    and (e_start_time.hour >= 21
                         or e_start_time.hour <= 9)):
                print(f"{item['id']}: {e_start_time} -> ended.")
                batch.delete_item(Key={"id": item["id"]})

                # Don't use the old hours slot if it has no entries now that we've deleted!
                res = table.query(
                        IndexName="hours-gsi",
                        KeyConditionExpression="#id = :i AND #year = :y AND #month = :m AND #day = :d",
                        ExpressionAttributeNames={
                            "#id":"id",
                            "#year":"year",
                            "#month":"month",
                            "#day":"day"
                        },
                        ExpressionAttributeValues={
                            ":i":{"S":item["id"]},
                            ":y":{"N":str(currenttime.year)},
                            ":m":{"N":str(currenttime.month)},
                            ":d":{"N":str(currenttime.day)}
                        })
                if len(res["Items"]) == 0:
                    print("\tDeleting hours entry...")
                    table.delete_item(Key={
                        "id": "{}/{}/{}/hours".format(str(currenttime.year),
                                                     str(currenttime.month),
                                                     str(currenttime.day)),
                        "year": {"N": str(currenttime.year)},
                        "month": {"N": str(currenttime.month)},
                        "day": {"N": str(currenttime.day)},
                    })

            elif s_start_time > dt: # pragma: no cover
                break

            else:
                print(f"{item['id']}: {e_start_time} -> continuing.")
                batch.update_item(Key={
                                "id": item["id"]
                                }, UpdateExpression='REMOVE hours[\'{}:{}\']'.format(
                                    end_day, e_start_time.hour))



if