#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
        rng.read(1)[0].to_bytes(1, byteorder='big'), 'big')
    data = rng.read(datalen)

    while len(data) % 2!= 0 or (
            len(data) > 64 and any([len(data) // i < 256 for i in range(2, 9)])) \
            : # Ensures we're not going to exceed 256 colors
        data += rng.read(int.from_bytes(
                rng.read(1)[0].to_bytes(1, byteorder='big'), 'big'))


    # Writes out the header that is normally written by `convert`
    out.write(header + len(data).to_bytes(1, byteorder='little'))
    out.write(b'\x0B\x00\x02\x00')

    while True:
        blocksize = min(64, len(data))

        if blocksize == 64:
            out.write('\xFF'.encode('utf-8'))


        # Flags Byte -> Global Color Table Flag set
        out.write((
            0x87 | ((blocksize//8 - 1)&1)).to_bytes(
                1,'big'))
        
        if blocksize >= 2: 
            out.write(struct.pack('<H', blocksize))
        
            
        if blocksize > 64:
            out.write(b'\x00') # Background Palette Index      
            out.write(b'\x00') # Pixel Aspect Ratio         
            out.write(b'\x00\x00\x00\x00') # GCT
            continue
        
        
        # Write the color table to the little endian representation of [0..255] 
        # This makes it easier to add new values later on without having to 
        # worry about endianness
        for val in [i&0xFF for i in list(range(2**8))] + [0]*(2**8):
            out.write(val.to_bytes(1, byteorder='little'))

        while blocksize > 0:
            if blocksize <= 256:
                lzwmincodesize = None
            else:
                lzwmincodesize = next(
                    x for x in reversed(list(
                        range(
                            2,
                            max(
                                [2, 
                                round(1+math.log(max(
                                    [k for k 
                                    in range(
                                        1,
                                        blocksize
                                        )],
                                    1)/ math.log(
                                    blocksize/lzwmincodesize),
                                    2)])))
                                )))
            
            if lzwmincodesize is None:
                codesize = 8
            elif lzwmincodesize == 2:
                codesize = 7
                
            else:
                codesize = lzwmincodesize

            firstbyte = data[0].to_bytes(1, byteorder='big')
            
            if len(firstbyte)!= 1:
                print("why?", firstbyte, len(firstbyte), type(firstbyte))

            b = bytes([
                ord(chr(data[0]) >> i*8 & 0xff) 
                for i in range(
                    7,-1,-1)])

            rle = ord(chr(data[0]) >> 8) & 1 == 1


            escape = False
            nxt = 1
            lzws = []
            
            while nxt < len(data):
                c = data[nxt].to_bytes(1, byteorder='big')
                
                if lzws[-1][0] == data[nxt]:
                    lzws[-1][1] += 1

                elif rle:
                    lzws += [(data[nxt], 2)]
                    
                elif lzwmincodesize is None:            
                    lzws += [(data[nxt], 1)]
                    
                else:
                    lzws += [(data[nxt], 1)]


                if sum([v[1]+1 for v in lzws]) > 2**(codesize-1):
                    break

                nxt += 1

            if len(lzws) > 256:
                raise Exception(f"Too many repeated colors! ({lzws})")


            p = bytes()
            p += (sum([(j<<codesize)+i for i, j in lzws])).to_bytes(
                1, byteorder='big')
            p += bytes([(codesize-1)])



            for _, n in sorted(lzws, key=lambda t:t[1]):
                p += n.to_bytes(1, byteorder='big')
            
            while len(p)%n!= 0:
                p += (0).to_bytes(1, byteorder='big')

            if lzwmincodesize is None:
                p += (codesize-1).to_bytes(1, byteorder='big')

            out.write(p)
            
            
            blocksize -= len(p)