#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag, etc.)
    for _ in range(7): rng.read(2)

    # read the local palette as well
    gct_size = int((image_width * image_height) / 0x100) + 1

    if gct_size == 0: return False
    
    # how many colors to include?
    total_colors = 0xFF + 1
    if gct_size < 0xF: total_colors = 0x1 << gct_size

    # build up the palette
    lct = bytearray()
    for color in range(total_colors):

        # R
        r = rng.read(1)[0] & 0xFF

        # G
        g = rng.read(1)[0] & 0xFF

        # B
        b = rng.read(1)[0] & 0xFF
        
        # flags
        flags = rng.read(1)[0] & 0xFF

        # add the individual entry
        lct += bytes([r, g, b])


    # write the header
    out.write(header)

    # write the width and height of the image
    out.write(struct.pack('<HH', image_width, image_height))

    # screen descriptor with no color table, 
    # 1 bit per pixel monochrome
    out.write(b'\xf7\x00\x00')

    # graphics control extension
    out.write(b'\x21')
    out.write(b'\xff')
    out.write(b'\x0b')
    out.write(b'\x4e')
    out.write(b'\x45')
    out.write(b'\x54')
    out.write(b'\x53')
    out.write(b'\x43')
    out.write(b'\x41')
    out.write(b'\x50')
    out.write(b'\x45')
    out.write(b'\x32')
    out.write(b'\x2e')
    out.write(b'\x30')

    # use transparent black on all frames
    out.write(b'\x00')
    out.write(b'\x00')

    # delay between each frame is one hundredth of a second 
    # or since we're using an old version of GIF89b this should
    # be zeroed here too!
    out.write(b'\x00')
    out.write(b'\x00')

    # end graphic control block
    out.write(b';')

    # image descriptor block: left = 0, top = 0
    out.write(bytes([0x2c, 0x00]))

    # width and height again like above
    out.write(struct.pack('<HH', image_width, image_height))

    # no LZW compression used
    out.write(b'\x00')

    # finally, the actual image data
    for row in range(image_height):
        # last parameter says "this is the last line"
        out.write(
            b'\x00' + 
            lct[row % len(lct)]*image_width +
            b'\x00' 
        )

    # done.
    return True

