#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and the
# screen descriptor data `scrn_desc`. Coordinates in the `scrn_desc` are given as top-left is (0, 0).
def gen(out: BinaryIO, rng: BinaryIO, scrn_desc):
    # Header
    out.write(b'GIF89a')

    width = scrn_desc[2] - scrn_desc[0] + 1
    height = scrn_desc[3] - scrn_desc[1] + 1

    # Logical screen descriptior with global color map of size 7 bits per pixel.
    out.write(struct.pack('<HHBBB',
        width,     # Screen width
        height,    # Screen height
        0x87,      # Flags: Global Color Map, Sort Colors, Size of Global Color Map = 7
        0,         # Background color index
        0,         # Pixel aspect ratio
        7))        # Size of global color table

    # Application extension for the NETSCAPE application to define repeat count of frames.
    appExtLabel = b'\x21\xFF\x0BNETSCAPE2.0'
    appExtData = b'\x03\x01\x00\x00\x00' + bytes([int((width * height / 15)**0.5)])
    out.write(bytes([len(appExtLabel)]))
    out.write(appExtLabel)
    out.write(bytes([(len(appExtData) + 1), len(appExtData)]))
    out.write(appExtData)
    
    # Graphic control label and block
    out.write(b'\x21\xF9\x04')
    out.write(struct.pack('BBH',
        0x08,          # Extension size
        0x01,          # Packed field: disposal method, user input required
        0x00))         # Frame delay time in units of 1/100 seconds.
    out.write(b';')
    
    # Image descriptor and local palette
    out.write(b',' + bytes([0, 0, height % 256]) + bytes([height // 256]))
    for i in range(7**2):
        clr = struct.unpack('BBBB', rng.read(4))
        out.write(bytes([clr[i] & ~0xE0 | ((clr[i] >> 5) & 0xE0)]));

    for y in reversed(range(height)):
        out.write(b'\x21\xFE\x00')
        out.write(b',')
        
        move = False

        for x in range(width):
            if move == True:
                out.write(b'+')
            else:
                out.write(b'-')

            byte = int(rng.read(1)[0] / 2)
            
            out.write(struct.pack('BB', byte >> 4, byte & 0xF))

            move = not move