#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # The first three bytes of a valid GIF header.
    signature = [0x47, 0x49, 0x46]

    # A valid logical screen descriptor that says it's 1025 pixels wide and tall
    # with a depth of 8 bits per pixel.
    lsd = [
        0b1000_0000 | (1025 >> 8),       # Width, high byte
        0b1000_0000 | (1025 & 0xff),     # Width, low byte
        0b1000_0000 | (1 << 5) | (1025 >> 8),   # Height, high byte
        0b1000_0000 | (1 << 5) | (1025 & 0xff), # Height, high byte

        0b1000_0000 | (8),                # Bit depth

        # No color table specified, so we'll use one from the palette instead.
        0b1000_0000                        
    ]

    # The Global Color Table section containing 256 RGB entries each.
    gct_size = b'\xFF' + (struct.pack('B', 256)) * 3
    gct_content = []

    # Generate 256 RGB colors in the range 0-255 inclusive.
    for i in range(256):
        r = next(rng) % 256
        g = next(rng) % 256
        b = next(rng) % 256
        gct_content += [(r << 16) | (g << 8) | b]

    # The Image Descriptor section which contains information about this frame.
    idt = [
        # The left offset of the image. Since there is no horizontal spacing between frames
        # then we just don't care what this is set to yet.
        0,

        # The top offset of the image. Again since there will be no vertical spacing we can
        # leave this as zero.
        0,

        # Width, high bit unset because width should always be less than 2**12 but let's not make sure 
        # either by setting it here and removing all other bits later on too.
        0b1000_0000 | (1025 >> 8),        

        # Width, low bit (LSB). See above comment. 
        0b1000_0000 | (1025 & 0xff),     

        # Height, same comments apply here.
        0b1000_0000 | ((1 << 5) | 1025 >> 8), 

        # Height, LSB.
        0b1000_0000 | (1 << 5) | (1025 & 0xff),   

        # Bits Per Pixel, same as other sections.
        8,                
    ]

    num_lzw_codes = {
        2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7, 8: 8, 9: 9, 10: 10, 12: 12, 14: 14, 16: 16, 18: 18, 20: 20, 22: 22, 24: 24, 26: 26, 28: 28, 30: 30, 32: 32
    } 

    # Now that everything has been generated we can start writing to our output stream. 
    # First off write the header.
    out.write(bytes(signature))
    
    # Write the Logical Screen Descriptor
    out.write(bytes(lsd))

    # Write GCT
    out.write(gct_size)
    out.write(bytes(gct_content))

    # Write Image Descriptor
    out.write(bytes(idt))

    # The minimum and maximum code lengths respectively. We only need to specify these if they aren't equal!
    min_code_len = max_code_len = None

    # This loop gets repeated until our animation has finished.
    while True:
        # The first byte defines how many bytes are used to store the length of the subblock.
        size = next(rng) % 128
        
        # If the highest bit of the first byte isn't set then len is stored in a single byte, otherwise you have
        # to read an additional byte from the input stream.
        if size & 0b1000_0000 == 0:
            len_in_bytes = 1
        else: 
            len_in_bytes = size - 0b1100_0000

        # Read the actual length value now.
        len = int.from_bytes(rng.read(len_in_bytes), 'big')
            
        # From this point onwards we're going to ignore the size field completely so we set it back to 8.
        size = 8

        # Next up comes the block type byte. This tells us exactly what the following part means.
        kind = rng.read(1)[0]

        if kind == 0x00:
            break
        elif kind == 0x2C or kind == 0x21:
            pass
        else:
            raise RuntimeError("unknown data block")

        # Read the image data itself.
        imgdata = rng.read((len_in_bytes)*2)

        # Subblocks must end on byte boundaries and also cannot contain any zeroes.
        while imgdata[-1]!= 0:
            imgdata += rng.read(1)

        # When processing SubBlocks we do not know whether their contents use the minimum or maximum possible code lengths.
        # So when encountering such blocks we need special handling. However, due to the way we've written the main code
        # we never actually reach into the SubBlock processor at all. Instead we process the main code and check if the
        # previous operation was a literal string literal. In case it was then we append its code length to the array below.
        # At the very least however we reserve room for it even though it might turn out to be useless.
        code_lengths = []

        # To simplify things we decode the entire image data at once. While this produces duplicate code blocks every few
        # bytes this doesn't matter much as we won't do anything special with them anyway. Anyway it makes the decompression easier...
        # This process consists of two parts. First we use LZW decoding to obtain the decoded image data. Then after this we go through
        # again and look for specific sequences of codes that tell us some extra metadata that we want to extract.
        # Algorithmically speaking this could probably be done more efficiently but this implementation is hopefully simple enough
        # unless anyone finds a really obscure edge case that we haven't considered yet! ;)
        def process():

            # Initialize empty current_seq. We initialize it to a dummy value here so that we can always access it without having
            # to worry if it exists or not.
            current_seq = 0xFF
            
            # This list stores all the code lengths which are present in this compressed image data. Note that we never consider
            # the first code length when looking for repetitions since it's always 0.
            distinct_code_lengths = []

            # Decode the image data and keep track of the sequence currently being processed.
            for chunk in lzw_decode(imgdata):
                # If we encounter a new sequence then update the distinct_code_lengths list accordingly.
                if chunk > current_seq:
                    distinct_code_lengths.append(chunk)

                yield distinct_code_lengths
                
                # Update the last known sequence.
                current_seq = chunk 

        # This is where we process the decoded image data.
        image = list(process())

        # For the rest of the metadata extraction we look for certain patterns in the distinct_code_lens list.
        # This is pretty ugly and horrible programming practice but oh well... It works!
        # TODO: find better solution?
        def get_min_max_code_length(distinct_code_lengths):
            nonlocal min_code_len, max_code_len
        
            assert isinstance(distinct_code_lengths, list)

            counts = {}

            for code_len in distinct_code_lengths:
                try:
                    count = counts[code_len]
                except KeyError:
                    count = 0
                    
                counts[code_len] = count + 1

            max_num = float('-inf')
            min_num = float('+inf')

            for key, val in counts.items():
                if (val / len(distinct_code_lengths)) >=.95:
                    if key < min_num:
                        min_num = key

                    if key > max_num:
                        max_num = key

            return (max_num, min_num)

        # Look for the magic code that tells us the minimum code length to use.
        # Technically there might be multiple copies of this pattern in the image but my guess would be
        # that most images only have a single occurrence of it. Still, I'm not entirely sure.
        # Should we ever come across an example where there are multiple occurrences we