#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_gif(out: BinaryIO, rng):
    # Write magic string and header
    out.write(b"GIF87a\x02\x4a")

    w = 16 * int((rng.read(1)[0] & 5) + 2)
    h = 16 * (int((rng.read(1)[0] & 5)) + 2)
    bpp = ((rng.read(1)[0] & 7) + 1) * 8

    datalen = w*h // (bpp / 8) - 2
    if datalen < 0:
        return
    datastr = struct.pack('<HHH', w, h, bpp)
    out.write(datastr);

    # Write color table
    for i in range(2**(bpp//8)):
        out.write(rng.read(3))

    x = (-w | 0xFF)*bpp // 8
    y = (-h | 0xFF)*bpp // 8
    framehdr = bytes([
        0x2c,# Image separator
        0,       # Left position
        0,       # Top position
        w-1,     # Width
        h-1,     # Height
        bpp,     # Bits per pixel
        0,       # Flags
        x&0xff,  # LZW minimum code size
        x>>8,    # LZW minimum code size
        y&0xff,  # Data subblock size
        y>>8])   # Data subblock size
    out.write(framehdr)
    while len(data := rng.read(y+2)) > 0:
        out.write(data)
        datalen -= len(data)-2
        if datalen <= 0:
            break
    out.seek(-1, 1)
    out.write(bytes([0]))

if __name__ == "__main__":
    from os import urandom as rng
    with open("random.gif", "wb") as f:
        generate_gif(f, rng)

"""
The official specification can be found at https://www.w3.org/Graphics/GIF/spec-gif89a.txt
The following references are useful to know about the GIF format:
  http://weblog.boiteaklou.fr/?p=876
  http://giflib.sourceforge.net/whatsinagif/index.html
"""


def check_magic(reader:BinaryIO):
    """
    Reads the first three bytes of a gif file, returns ``True`` iff it is valid.
    Raises an exception on error or EOF.
    """
    try:
        assert tuple(reader.read(3)) == ('G','I','F')
        version = ord(reader.read(1))

        assert reader.read(1) == b'8'
        assert str(reader.read(1), 'ascii').lower() in ['7', '9']

        reader.seek(-1, 1)  # Restore cursor so we can read more safely later
        return True
    except AssertionError:
        raise ValueError('Invalid GIF file')
    finally:
        # Try to restore cursor even if there was an exception
        reader.seek(-3, 1)


def validate_image(fileobj:BinaryIO):
    """
    Checks whether a given image file object contains valid GIF image data.
    Returns ``None``, raises ValidationError otherwise.
    If you're unsure whether the provided file-like object supports seeking,
    use :func:`validate_image_buffer`.
    This function does not have access to filenames when validating buffer objects.
    """
    try:
        check_magic(fileobj)

        width, height, depth = struct.unpack("<HHB", fileobj.read(5))

        _, left, top, right, bottom, flags, *_ = \
                struct.unpack("<HBHHBB", fileobj.read(9))


        if flags & 0x80:
            # Interlaced images aren't supported yet
            raise NotImplementedError(
                    "Interlaced GIFs are unsupported.")

        if _!= 0:
            # Colormaps aren't supported yet
            raise NotImplementedError(
                    "Colormapped GIFs are unsupported.")

        if left!= 0 or top!= 0:
            # Images whose origin isn't at upper-left aren't supported yet
            raise NotImplementedError(
                    "Images whose origin isn\'t at upper-left aren\'t implemented.")

        if width == 0 or height == 0:
            # Zero sized images aren't supported
            raise ValueError("Zero-sized images are invalid.")

        if left >= right or top >= bottom:
            # Invalid crop area
            raise ValueError("Image's crop area lies outside its bounds")

        if depth < 1 or depth > 8:
            # Unsupported bitdepth?
            raise ValueError(
                    "Unsupported bitdepth {}.".format(depth))


    except TypeError:
        # Not enough arguments supplied by unpack
        raise ValueError(
                "Too few bytes available to parse image dimensions.")

    else:
        # No errors encountered, all good!
        pass


def validate_image_buffer(buf:bytes):
    """
    Convenience wrapper around :func:`validate_image`, taking a
    filename instead of a file object..
    The returned value is identical to that of :func:`validate_image`.
    """
    with BytesIO(buf) as bufio:
        return validate_image(bufio)


def convert_size(size, original=(1, 1)):
    """
    Converts the specified size to a new aspect ratio.
    Returns a tuple containing the converted size.
    The default aspect ratio used is 4:3, which means that width will always be
    4/3 times larger than height.
    """
    hw = float(original[0])/float(original[1])
    return round(size[0]/hw), round(size[1]*hw)



# GifReader class stolen shamelessly from
# http://stackoverflow.com/questions/7551607/how-to-get-all-the-frames-of-a-gif-using-python
class GifReader():
    """
    A simple GIF frames extractor, designed to provide a single iterator
    over the frames contained within a GIF file.
    Passes all keyword arguments to PIL's ``ImageFile.Parser()``.
    Intended usage::
      >>> from PIL import Image
      >>> img = Image.open("test.gif")
      >>> gif = iter(GifReader(img))
    """

    def __init__(self, file, **kwargs):
        self._idx = None
        self._nframes = 0
        p = ImageFile.Parser(**kwargs)
        p.feed(file.read())
        self._frames = p.close()[0].n_frames
        file.seek(0)
        p = ImageFile.Parser(**kwargs)
        self._frameinfo = []
        while 1:
            idx = p.tell()
            try:
                p.feed(file.read(1024))
            except Exception:
                # End of file
                break

            # Append frame info
            self._frameinfo += [(idx, p.image)]

        del p

    def __len__(self):
        return self._frames

    def getinfo(self, n):
        """
        Return the :py:meth:`~PIL.GifImagePlugin.GifImageFile.n`-th frame information of the GIF image.
        Returned value is a tuple containing offset and frame duration.
        Offset is absolute file position starting from beginning of file. Duration is measured in milliseconds.
        """
        if isinstance(n, slice):
            return [self.getinfo(i) for i in range(*n.indices(len(self)))]
        elif n < 0:
            n += len(self)
        if n < 0 or n >= len(self):
            raise IndexError("Index %d out of range."%n)
        return list(self._frameinfo[n])

    @staticmethod
    def _dispose_opcodes(image):
        """
        Iterator over disposal method codes yielding tuples containing image rectangle coordinates and
        disposal method code respectively.
        """
        opcodes = {
            0x00: 'No disposal specified.',
            0x01: 'Leave canvas untouched.',
            0x02: 'Restore to background color.',
            0x03: 'Restore to previous.'}

        def toByte(byte):
            """Convert integer to signed char."""
            return struct.unpack('b', byte)[0]

        data = image.info['background']
        r, g, b = toByte(data[0]), toByte(data[1]), toByte(data[2])
        cblack = (r << 16) + (g << 8) + b

        # Grab graphic control block
        loc, count = image.tile[0][2], image.tile[0][3]
        blocks = image.tobytes().split(b'\x00'[0])[-count:-1]

        for blk in blocks:
            blk = blk[:loc]
            flags = blk[-1]
            fgnd = blk[3:6]

            # Get transparency color index
            t = ((fgnd[1]<<8)+fgnd[2])&0xffff

            # Convert transparent color to RGB
            tr, tg, tb = toByte(t >> 8), toByte(t & 255), toByte(t >> 16)


            # Check if this is a local color change block
            if            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + data)
