#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given pseudorandom number generator in binary format
def generate(out: BinaryIO, rng):
    out.write(b"GIF89a")

    header_bytes = bytearray()
    # Screen width and height (small values for testing)
    header_bytes += struct.pack("HH", 16, 16)
    # Global color table size, no global color table used
    header_bytes += b"\x00\x00"

    out.write(header_bytes)

    while True:
        # Frame duration in centiseconds (small value for testing)
        duration_bytes = bytearray([0, 20])

        left_bytes = bytearray()
        top_bytes = bytearray()
        packed_width_height = struct.pack("<HHH", 0, 0, 16)

        local_color_table_size = 0
        local_color_table_bytes = bytes([])

        lzw_mincodesize = 255
        pixels = [[0] * 16 for _ in range(16)]
        blocktype = next(rng) % 4 + 1

        if blocktype == 1 or blocktype == 2:
            local_color_table_size = 7

            palette = []
            for i in range(local_color_table_size + 1):
                palette.append((next(rng), next(rng), next(rng)))

            local_color_table_bytes = b"".join([struct.pack("BBB", r, g, b)
                                                for r, g, b in palette[1:]])

        elif blocktype == 3:
            local_color_table_size = 1 << next(rng) & 7

            palette = []
            for i in range(local_color_table_size + 1):
                palette.append((next(rng), next(rng), next(rng)))

            local_color_table_bytes = b"".join([struct.pack("BBB", r, g, b)
                                                for r, g, b in palette[1:]])

        else:
            pass

        if blocktype >= 2:
            # LZW minimum code size
            lzw_mincodesize = next(rng) % 4 + 2

        data = [pixels]

        compressed_data = b""
        lzw_clearcode = 1 << lzw_mincodesize
        num_written = 0

        while len(compressed_data) < 256:
            pixel = [(i // 16, i % 16)
                     for i in range(len(pixels)) if not pixels[i]]

            if not pixel:
                break

            x, y = pixel[next(rng) % len(pixel)]
            pixels[y * 16 + x] ^= 1

            start = ord(blocktype - 1) << 1 | 1
            end = ((lzw_mincodesize - 1 + 12) >> 4) + 1
            
            # Since we have some data already written,
            # use that as our first code instead of clear code
            first_code = (num_written > 0) << 1 | (1 ^ bool(start))
            current_code = None
            codes = {}

            for i in range(start, end):
                index = 1 << (lzw_mincodesize - 1 - i*(lzw_mincodesize-1)//12)

                if current_code is None:
                    current_code = index

                    if current_code!= lzw_clearcode:
                        compressed_data += struct.pack("B",
                                                        current_code - 1)

                        num_written += 1
                else:
                    prev_code = current_code

                    if current_code & index:
                        current_code |= ~index
                    else:
                        current_code &= index

                    new_bit = int(current_code == lzw_clearcode)
                    
                    if current_code in codes:
                        pos = codes[current_code][0]
                        codes[current_code].insert(
                            0, new_bit |
                            (pos % (1 << lzw_mincodesize - 1) + 1) << lzw_mincodesize - 1)
                    else:
                        codes[current_code] = [-new_bit,
                                               len(compressed_data)+1,
                                               (prev_code & index)!= 0]

                    if current_code!= lzw_clearcode:
                        compressed_data += struct.pack("B",
                                                        current_code - 1)

                        num_written += 1

            assert len(codes) <= (1 << lzw_mincodesize)

        out.write(duration_bytes
                 + local_color_table_size.to_bytes(1, "little")
                 + local_color_table_bytes
                 + lzw_mincodesize.to_bytes(1, "little")
                 + compressed_data)
        
        yield

