#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):

    # Write header
    head = [b"GIF", b"89a"]
    for i in range(7):
        head.append(struct.pack("B", int.from_bytes(rng.read(1), "big")))
    out.write(head[0] + b"".join([len(x).to_bytes(1, byteorder="big")
                                 for x in head]))
    offs = 6 + sum([len(x) for x in head]) // 255

    # Get palette and color table size (int. from bytes, big endian)
    palsize = int.from_bytes(rng.read(1), "big")
    ctsize = 3 * pow(2, palsize + 1) - 1

    # Generate global color table
    gct_b = []
    if palsize > 0:
        for _ in range(palsize):
            r = math.floor((rng.read(1)[0]/255)*255)
            g = math.floor((rng.read(1)[0]/255)*255)
            b = math.floor((rng.read(1)[0]/255)*255)

            gct_b += [r,g,b]

        while len(gct_b) % 3!= 0:
            r = 0
            g = 0
            b = 0

            gct_b += [r,g,b]

        gct = [int.from_bytes(x, "little")
               for x in zip(*(iter(gct_b),) * 3)]

        # Set up local color table
        lct = [[r,g,b]
              for r in range(256)
              for g in range(256)
              for b in range(256)]
        randmap = list(range(256))

        # Randomize color map
        for i in range(len(randmap)):
            tmp = randmap[(i+42)%len(randmap)]
            randmap[i], randmap[tmp] = randmap[tmp], randmap[i]


        # Generate local color table
        lct_b = ""
        for i in randmap[:ctsize]:
            lct_b += "".join(["{0:08b}".format(x)
                              for x in lct[i]])

        # Write global color table
        gct_b = b"".join([x.to_bytes(1,"big")
                          for x in gct])
        while len(gct_b) % 3!= 0:
            gct_b += "\x00".encode('ascii')
        while len(gct_b) < pow(2, palsize + 1) * 3:
            gct_b += ("\xff".encode('ascii'))*3
        out.write(gct_b)

        # Write local color table
        lct_b = b"".join([x.to_bytes(1,"big")
                         for x in lct])
        while len(lct_b) % 3!= 0:
            lct_b += "\x00".encode('ascii')
        while len(lct_b) < ctsize * 3:
            lct_b += ("\xff".encode('ascii'))*3
        out.write(lct_b)

        lct_entry = 3*(2**(palsize+1)-1)+3

    else :
        lct_entry = 0
        gct_b = None

    # Write image descriptor block
    imd = [0, 0, 0, 0, 0]
    for i in range(4):
        imd[i] = rng.read(1)[0]
    imd[1] |= ((rng.read(1)[0]&0x3)<<8) | (imd[1]&0xfc)
    imd = [x.to_bytes(1,"little")
           for x in imd]
    
    out.write(b"".join(imd))
    offs+= 8

    # Write LZW code size 
    code_size = rng.read(1)[0]&0x0F
    out.write(code_size.to_bytes(1,"little"))
    offs += 1

    # Calculate no of data blocks
    datablocks = rng.read(1)[0]&0xFF
    out.write(datablocks.to_bytes(1,"little"))
    offs += 1

    # Write LZW data
    min_code_size = 1+min(((2**code_size)//255),(256//(2**code_size)))
    max_code_val = (2**(code_size+1))-1
    bitsize = {
        1:[],
        2:	["{:02b}", "{:04b}"],
        3: [":03b]", ":06b]",
        4: ["{:04b}", "{:08b}"],
        5:[ "{:05b}", "{:010b}" ],
        6: [ "{:06b}", "{:012b}" ]
    }[min_code_size]

    codes = dict()
    codes_list = []
    cur_pos = 0;
    next_clear = 2**max_code_val;
    last_char=None;
    chars=[]

    def appendToChar(c):
       nonlocal char_count;

       c1=c&0xF0
       c2=(c<<4)&0xF0
       c1|=last_char&0x0F
       c2|=(last_char>>4)&0x0F

       s=""
       for i in range(2,-1,-1):
           e=bitsize[i].format(c1>>i*4)
           s+="".join(e)
       d=""
       for j in range(2,-1,-1):
           f=bitsize[j].format(c2>>j*4)
           d+="".join(f)
       return s+d


    for current_block in range(datablocks):
        if current_block<datablocks-1:
            toWrite=rng.read(1)[0]+256
        elif current_block==datablocks-1:
            toWrite=-1

        while True:
            curr_bits=[]
            mask_num=128
            for num in range(9):
                if toWrite &mask_num>0:
                    curr_bits.append(1)

                else:
                    curr_bits.append(0)
                mask_num>>=1
            if not bits:break
            if lzw_code is not None:
                pass
            if len(curr_bits)==9:
                if lzw_code is None:
                    lzw_code=cur_pos
                    codes_list.append([])

                


                last_lzw_code = codes_list[-1][-1]
                
                codes[current_block]=codes.get(current_block,{})
                codes[current_block][lzw_code]=appendToChar(last_lzw_code)



                bits=[0]*12
                bits[0]=curr_bits[0];
                bits[1]=curr_bits[1];
                bits[2]=curr_bits[2];

                del curr_bits[0:3]
                bits[3]=curr_bits[0]
                bits[4]=curr_bits[1]

                del curr_bits[0:2]
                bits[5]=curr_bits[0]
                bits[6]=curr_bits[1]
                

                del curr_bits[0:2]
                bits[7]=curr_bits[0]
                bits[8]=curr_bits[1]

                del curr_bits[0:2]
            


            if len(bits)>12:
                print("ERROR HERE!")
            w = "".join(["{:01b}".format(x) for x in bits])
            
            out.write(bytearray([int(w[i:i+8],2) for i in range(0,len(w),8)]))
            chars.extend(w);

            cur_pos+=1
            if cur_pos>=next_clear:
                break

        if current_block!=datablocks-1:
            out.write(b"\x00")
            cur_pos+=1


    out.flush();

    # Write trailer
    offs = set_chunk(out, offs, "NETSCAPE2.0", b"\x03\x01\x00\x00\x00")
    offs = set_chunk(out, offs, "XMP Data XMP", b"<?xpacket begin=\"\" id=\"W5M0MpCehiHzreSzNTczkc9d\"?>\<x:xmpmeta xmlns:x=\"adobe:ns:meta/\"><rdf:RDF xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"><rdf:Description rdf:about=\"\"\txmp:CreatorTool=\"Adobe Photoshop CC 2014 (Windows)\"></rdf:Description></rdf:RDF></x:xmpmeta><?xpacket end=\"w\"?>\x00");
    offs