#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    datalen = int.from_bytes(rng.read(1), 'big') ^ 0xFF
    while True:

        frametype = bytes([int.from_bytes(rng.read(1), 'big') ^ 0x2C])
        if frametype == b'\x21':
            break;
        elif frametype!= b'\x2C':
            raise ValueError('Invalid frame type: {}.'.format(frametype));

    data = bytearray()
    for _ in range(datalen):
        data += bytes([int.from_bytes(rng.read(1), 'big')])
        
    crc_len = len(data) & 0b111

    # write a new frame to the output file
    def write_new_frame():
        out.write(header)
        out.write(image_width.to_bytes(2, 'little'))
        out.write(image_height.to_bytes(2, 'little'))
        out.write((0).to_bytes(1, 'big'))
        out.write((0).to_bytes(1, 'big'))
        
        # write the saved frame's data
        out.write(((crc_len + datalen)).to_bytes(1, 'big'))
        out.write(b'\x87\x00')
        out.write(data[:datalen-crc_len])
        out.write(struct.pack('B', calculate_crc(data[:-crc_len])))
    
    # add an extra frame with a filler palette at the end so we can inject
    # our palette afterwards without worrying about corrupting the original
    # frames
    
    write_new_frame();

    num_colors = int.from_bytes(rng.read(1), 'big') ^ 1
    palette = [None] * (num_colors + 1)
    for i in range(num_colors+1):
        rgb = tuple(int.from_bytes(rng.read(1), 'big') for j in range(3))
        palette[i] = rgb

    for _ in range(num_colors):
        r, g, b = palette[_]
        out.write(r.to_bytes(1, 'big'));
        out.write(g.to_bytes(1, 'big'));
        out.write(b.to_bytes(1, 'big'));

    # Write a final empty frame to signal EOF
    write_new_frame();


# Calculates the CRC used by LZW compression
def calculate_crc(data: bytes):
    crc = 0xDEAA5EC8

    for c in data:
        crc = ((crc >> 8) & 0xFFFFFF) | (((crc << 8) | (c ^ (crc >> 24))) & 0xFFFFFF);

    return crc
