#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Header
    header = bytearray([
        0x47, 0x49, 0x46, 0x38,
        0x39, 0x61, 0x01, 0x2c,
        0x00, 0xff, 0xff, 0xff,
        0xff, 0xff, 0xff, 0x21,
        0xf9, 0x04, 0x05, 0x0a,
        0x00, 0x00, 0x00, 0x2c,
        0x00, 0x00, 0x00, 0x00,
        0x01, 0x00, 0x01, 0x00,
        0x00, 0x02, 0x01, 0x44,
        0x00, 0x3b])

    out.write(header)
    
    num_frames = int(struct.unpack("<Q", rng.read(8))[0] % 10 + 1) # range[1,10]
    for _ in range(num_frames):
        frame_size = (int(struct.unpack("<Q", rng.read(8))[0] % 200 +1)) * num_frames

        # Image Descriptor Block
        out.write(bytearray([
            0x21, 0xFF, 0x0B,
            ((frame_size >> 24) & 0xff),
            ((frame_size >> 16) & 0xff),
            ((frame_size >>  8) & 0xff),
            ((frame_size      ) & 0xff)]))
        
        # Graphics Control Extension
        out.write(bytearray([
            0x21, 0xF9, 0x04, 0x05, 
            ((int(struct.unpack("<Q", rng.read(8))[0] % 100) << 4) | 
             int(struct.unpack("<Q", rng.read(8))[0] % 16)), # delay between [0..16] and duration [0..100]
            0x00]))

        # Plain Text Extension
        out.write(bytearray([
            0x21, 0x01, 0x0f, 0x00,
            0x01, 0x00]))
    
        # Application Extension
        block_size = int(struct.unpack("<Q", rng.read(8))[0] % 1000 +  1)
        data_size = int(struct.unpack("<Q", rng.read(8))[0] % 1000 +  1)
        app_ext_block = bytearray()
        while len(app_ext_block) < block_size:
            chunk_size = min((block_size - len(app_ext_block)), data_size)
            
            ext_chunk = b"\x0b" + \
                        bytes(range(len(app_ext_block))) + \
                        bytes(range(chunk_size)) + \
                        b"\x00\x00"

            app_ext_block += ext_chunk[:chunk_size-1]

        out.write(bytes([
            0x21, 0xFF, 0x0B,
            *(app_ext_block)]))

        # Comment Extension
        comment_size = int(struct.unpack("<Q", rng.read(8))[0] % 500 +  1)
        comment_ext = gen_comment(comment_size, rng)
        out.write(bytes([
            0x21, 0xFE, 0x0B,
            *(comment_ext)]))


        # Image Data
        out.write(gen_image(frame_size, rng))


def gen_comment(size: int, rng: BinaryIO) -> bytearray:
    return bytearray([(size >> 24) & 0xff,
                      (size >> 16) & 0xff,
                      (size >>  8) & 0xff,
                       size        ]) + \
           bytes(rng.read(size))

def gen_image(size: int, rng: BinaryIO) -> bytearray:
    return bytearray([(size >> 24) & 0xff,
                      (size >> 16) & 0xff,
                      (size >>  8) & 0xff,
                       size        ]) + \
           bytes(rng.read(size))


