#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # 4 bytes of magic
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Use a big enough LZW codesize to ensure that every bit is used in all
    # situations (the specification says 257-510.)
    lzwmincodesize = 6
    lzwmaxcodesize = 12
    lzwsize = rng.randint(lzwmincodesize, lzwmaxcodesize)

    # 1 byte per color palette entry
    colors = min((2 ** lzwsize)+1, len(data))
    # Ensure there's always at least one transparent color
    if colors == 1 and not isinstance(colors, int):
        raise Exception()
    pallen = colors*3+3 # at least 3 colors + local color table size
    pal = [b'\x00\xFF\x00'] * colors
    while True:
        if len(pal) < colors:
            new_color = list(
                rng.read(3))
            for i in range(3):
                    new_color[i] ^= pal[-1][i]

            pal += [new_color]

        elif len(pal) > colors:
            del pal[rng.randrange(len(pal))]

        else:
            break


    print('generating gif with {} colors'.format(colors))

    # TODO: write image block headers

    # Write the rest of the header
    out.write(header)
    # Screen descriptor
    out.write(struct.pack('<HHHHB',
            1,         # width
            1,         # height
            0xf0 | ((lzwsize - lzwmincodesize)//2),        # packed fields
            0,         # background color index
            0          # pixel aspect ratio
        ))

    # Global color table
    global_ct = b''.join([struct.pack('<BBB', *c) for c in reversed(pal)])
    out.write(global_ct[:3])    # size of global color table in bytes, minus 3
    out.write(b'\x00')           # reserved color table entry
    out.write(b'\x01')           # sort flag
    out.write(global_ct[3:])     # global color table entries

    # Image descriptor
    out.write(struct.pack('<HHHHBB',
            0,             # left position
            0,             # top position
            1,             # width
            1,             # height
            0,             # flags
            (lzwsize - lzwmincodesize)//2       # packed fields
        ))

    # Local color table
    out.write(local_ct := b''.join([struct.pack('<BBB', *c) for c in reversed(pal)]),)
    out.write(local_ct[:3])
    out.write(b'\x00')   # reserved color table entry
    out.write(b'\x01')   # sort flag
    out.write(local_ct[3:])

    # Image data
    out.write(data)


if