#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def gen_random(out: BinaryIO, rng: BinaryIO):
    magic = 0x47494638.to_bytes(4, 'big') # "GIF8" in big-endian
    out.write(magic)

    width = int((rng.read(1)[0] % 256) * 16 + (rng.read(1)[0] % 16))
    height = int((rng.read(1)[0] % 256) * 16 + (rng.read(1)[0] % 16))
    
    flags = bytes([
        ((rng.read(1)[0] & 1) << 7), 
        ((rng.read(1)[0] & 1) << 6), 
        0b00010000 | ((rng.read(1)[0] & 3) << 4),
        0b00000000
    ])

    global_color_table_size = 2 ** (((flags[0] >> 3) & 7) + 1)
    background = ord(rng.read(1)[0])
    pixel_aspect_ratio = rng.read(1)[0] & 7
    if pixel_aspect_ratio == 0: pixel_aspect_ratio = 10

    app_extension_block = bytes()
    while len(app_extension_block) < 12:
        app_extension_block += b'\xFF'
        app_extension_block += chr(len(app_extension_block)).encode('ascii')

    for i in range(global_color_table_size):
        color_r = rng.read(1)[0] & 0xFE
        color_g = rng.read(1)[0] & 0xFE
        color_b = rng.read(1)[0] & 0xFE

        # last byte is zeroed out to signify no transparency
        if i!= background:
            rle_count = rng.read(1)[0]
            for j in range(rle_count):
                app_extension_block += bytes([
                    color_r, 
                    color_g, 
                    color_b, 
                    0x00, 
                ])

    application_extensiion_identifier = \
            0xFF.to_bytes(1, 'little')\
            + b'XMP' + b'.' + b'R'
    application_extensio_authentication = \
                0x00.to_bytes(1, 'little')\
                + 0x00.to_bytes(1, 'little') 
    application_extensio_application_data = \
                    0x01.to_bytes(1, 'little')\
                    + 0x00.to_bytes(1, 'little')\
                    + 0x0C.to_bytes(1, 'little')\
                    + b'svg:width' + b'\x00' + chr(width).encode('ascii')\
                    + b'svg:height' + b'\x00' + chr(height).encode('ascii')\
                    + b'svg:version' + b'\x00' + b'1.1'
    
    out.write(struct.pack('<H', len(app_extension_block) + 19))
    out.write(b'!')
    out.write(app_extension_block)
    out.write(application_extensiion_identifier)
    out.write(application_extensio_authentication)
    out.write(application_extensio_application_data)

    # create empty local color table and save it at offset 13
    local_color_table_size = 2**(((flags[0] >> 6) & 7) + 1)
    local_color_table = bytes()
    for i in range(local_color_table_size):
        color_r = rng.read(1)[0] & 0xFE
        color_g = rng.read(1)[0] & 0xFE
        color_b = rng.read(1)[0] & 0xFE
        
        local_color_table += bytes([
            color_r, 
            color_g, 
            color_b, 
            0x00, 
        ])

    out.write(local_color_table)

    image_descriptor = bytes([
        (0b10000000 | ((rng.read(1)[0] & 0x0F) << 4)), 
        (rng.read(1)[0] & 0xF0) | ((rng.read(1)[0] & 0x0F) >> 4),
        rng.read(1)[0],
        rng.read(1)[0],
        0x00, 
        0x00, 
        0x00, 
    ])

    image_lzw_min_code_size = rng.read(1)[0] & 0x07 + 8 

    subblock_data = bytes([image_lzw_min_code_size, 0])

    color_table_size = 2 ** (((image_descriptor[0] >> 3) & 7) + 1)
    lzw_first_code_size = ((image_descriptor[1] >> 3) & 7) + 8 
    
    lzw_clear_code = 1 << lzw_first_code_size
    lzw_endof_information_code = lzw_clear_code + 1
    lzw_clear_subblock_header = bytes([
        (lzw_first_code_size - 8).to_bytes(1, "big")[0],
        0x00, 
    ])

    lzw_subblocks = list([])
    
    def add_lzw_codes():
        nonlocal subblock_data
        curr_code_size = lzw_first_code_size
        bits_remaining = 16
        code = []

        while True:
            if not bits_remaining:
                bits_remaining = 16

                if len(subblock_data) >= 255:
                    lzw_subblocks.append(subblock_data)
                    subblock_data = bytes([
                        (curr_code_size - 8).to_bytes(1, "big")[0],
                        0x00, 
                    ])
                else:
                    subblock_data += (curr_code_size - 8).to_bytes(1, "big")

            code.insert(0, (rng.read(1)[0] & (2**bits_remaining - 1)))
            
            remainder -= (2**bits_remaining // curr_code_size)
            if not remainder: break

            bits_remaining -= curr_code_size

        return code[:-1]
    
    full_subblock = False

    while not full_subblock:
        full_subblock = bool(rng.read(1)[0] & 1)
        num_additional_subblocks = rng.read(1)[0] & 7
        for _ in range(num_additional_subblocks):
            nonlocal subblock_data
            subblock_data += rng.read(1)

    lzw_codes = set(add_lzw_codes())
    assert lzw_codes <= {0, 1, 2}

    lzw_encoded_pixels = [2, 1][int(str(background))]
    while lzw_endof_information_code not in lzw_codes:
        lzw_codes |= {
            3,
        }
        code = tuple(reversed(list(add_lzw_codes())))
        print("lzw_encoded_pixels", lzw_encoded_pixels, end=' ')
        if code[:len(lzw_encoded_pixels)] == lzw_encoded_pixels:
            print("\t\tpixels", "".join(map(lambda x: str(x) + ",", code)))
            lzw_encoded_pixels = code[-1:]
            continue

        assert code[:len(lzw_encoded_pixels)] > lzw_encoded_pixels
        print("")
        lzw_encoded_pixels = code[-1:]
        lzw_codes |= set(code)


    lzw_subblocks.append(b'\x00')

    # write all subblocks to output file
    for index, subblock in enumerate(lzw_subblocks):
        extension_block_type = b''
        if index == 0:
            extension_block_type = b'!X'
        elif index == len(lzw_subblocks) - 1:
            extension_block_type = b';;'
        else:
            extension_block_type = b','
        
        out.write(struct.pack('<H', len(subblock) + 1))
        out.write(extension_block_type)
        out.write(subblock)

    # Write final image descriptor block
    data = bytes([
        image_descriptor[0],
        image_descriptor[1],
        image_descriptor[2],
        image_descriptor[3],
    ])
    data += (width % 256).to_bytes(1, 'big')
    data += (width // 256).to_bytes(1, 'big')
    data += (height % 256).to_bytes(1, 'big