#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))[:2]

    color_table_size = int(struct.unpack('B', rng.read(1))[0]) // 3
    global_color_table = bytearray()

    for i in range(0, (color_table_size * 3)):

        r = rng.read(1)[0] % 0x100
        g = rng.read(1)[0] % 0x100
        b = rng.read(1)[0] & 0b11110000 | rng.read(1)[0] & 0b00001111

        global_color_table += bytes([r, g, b])

    data = bytearray()


    while True:
        block_type = ord(struct.unpack('B', rng.read(1))[0])


        if block_type == 0x2C:
            frame_left_pos = struct.unpack('<H', rng.read(2))[0]
            frame_top_pos = struct.unpack('<H', rng.read(2))[0]

            frame_width = struct.unpack('<H', rng.read(2))[0]
            frame_height = struct.unpack('<H', rng.read(2))[0]

            flags = ord(struct.unpack('B', rng.read(1))[0])


            background_color_index = rng.read(1)
            pixel_aspect_ratio = rng.read(1)

            local_color_table_flag = bool(flags & 0b10000000)
            interlace_flag = bool(flags & 0b01000000)
            sort_flag = bool(flags & 0b00100000)
            reserved_bits_per_pixel = flags & 0b00011100

            lzw_min_code_size = struct.unpack('B', rng.read(1))[0]

            image_data_size = 0
            data_block_offset = len(data) + 2;

            while not data[-1]:
                data += bytearray([ord(rng.read(1))])
                image_data_size += 1

            data[data_block_offset - 2] |= image_data_size & 0xFF
            data[data_block_offset - 1] |= image_data_size >> 8

            if local_color_table_flag:

                local_color_table_size = ord(
                    struct.unpack('B', rng.read(1))[0]
                )

                local_color_table = bytearray()

                for _ in range(local_color_table_size):

                    r = rng.read(1)[0] % 0x100
                    g = rng.read(1)[0] % 0x100
                    b = rng.read(1)[0] & 0b11110000 | rng.read(
                        1)[0] & 0b00001111

                    local_color_table += bytes([r, g, b])

            break

        elif block_type == 0x21:

            sub_block_type = ord(struct.unpack('B', rng.read(1))[0])


            if sub_block_type == 0xf9:

                block_descriptor = ord(struct.unpack('B', rng.read(1))[0])


                delay_time = struct.unpack('<H', rng.read(2))[0]
                transparent_color_index = [
                    None,
                    struct.unpack('B', rng.read(1))[0],
                    struct.unpack('<H', rng.read(2))[0]
                ][bool(block_descriptor & 0b100)]
                terminator = struct.unpack('B', rng.read(1))[0]

                assert terminator == 0

                continue

    end_of_file_marker = rng.read(1)
    assert end_of_file_marker == b'\x3B'[0]

    assert block_type == 0x2c

    if local_color_table_flag:
        data[6] &= ~0b10000000 | ((local_color_table_size - 1) << 4)

    out.write(header)
    out.write(image_width.to_bytes(2, 'little'))
    out.write(image_height.to_bytes(2, 'little'))
    out.write(global_color_table)

    out.write(bytes([
        0x2c,
        frame_left_pos.to_bytes(2, 'little')[0],
        frame_top_pos.to_bytes(2, 'little')[0],
        frame_width.to_bytes(2, 'little')[0],
        frame_height.to_bytes(2, 'little')[0],
        (
            0b10000000 |   # color table flag
            (lzw_min_code_size - 1) << 4 |  # LZW min code size
            (int(interlace_flag) << 5) |     # interlace flag
            (int(sort_flag) << 6)           # sort flag
        ),
        background_color_index,
        pixel_aspect_ratio,
        *(local_color_table or [])
    ]))

    out.write(sorted(set((
        *transparent_color_index,
        *[i for i in range(256)],

        value.to_bytes(1, 'little') for value in data
    ))))


if