#!/usr/bin/env python3

import struct
from typing import BinaryIO

with open('data/kantanai.raw', 'rb') as f:
    with open('data/kantanai-patched.raw', 'wb') as g:
        patch(f, g)

def patch(inp: BinaryIO, out: BinaryIO):

    # Read header to check if file is kantanai or not.

    inp.seek(0x28)  # Seek to magic number
    magicnumber = inp.read(5).decode()
    # print("Magic number:", magicnumber)
    
    if (magicnumber!= "Kant"):   # If the first four bytes are NOT Kant
        return                    # Then we don't need to do anything special.

    # Then do our custom thing here...

    # First read the original header and store it in a variable.
    inp.seek(0x00)           # Seek back to beginning of file 
    header_orig = inp.read(4) # Just read until magic number 
    inp.seek(0x28)           # Seek to magic number again       
    header_new = inp.read(4)  # Store remaining header too 

    # Now write new header that points to an empty cadenza table.
    out.write(header_orig+b'\x6B\x00\x50\x7C'+header_new)
    # The patcher will then copy the entire raw file from 0x00 onwards into 0x6B7C9D ('Cadenza' function)

    # Patch the randomizer seed at offset 0x1A2FDB4 to be correct for the patched cadenza table.
    inp.seek(0x1A2FDB4)               # Seek to this offset
    inp.read(4)                       # Read 4 bytes so next uint can be read correctly
    randomizerseed = int.from_bytes( # Convert 4 bytes to unsigned int 
        inp.read(4),                   # Read these 4 bytes
        byteorder='little')            # Little endian
    
    #print("Randomizer seed before patch:", hex(randomizerseed)) 
    
    randomizerseed -= 0x6B             # Subtract 0x6B from randomizerseed  
    randomizerseed += 0x6B7CE7         # Add another 0x6B7CE7 to compensate 
                                        # for the size change due to inserting
                                        # a Cadenza function at 0x6B 
    
    
    #print("Randomizer seed after patch:", hex(randomizerseed))    
    out.seek(0x1A2FDB4)                # Seek back to same location                               
    new_seed = randomizerseed.to_bytes(
        4,                              # Store 4 bytes
        'little')                      # As little endian value
    out.write(new_seed)                # Write those 4 bytes at current position
    
    
    # Finally, append the rest of the file contents to output
    inp.seek(0x6B7CF0)                 # Seek to end of patches applied
    rng = inp                           # Rename so code is clearer
    while True:                        
        data = rng.read(16 * 1024)     # Read chunks of 16KB
        out.write(data)                # And append them to output

        if len(data) < 16*1024:         # Break when no more data was read
            break


    # Now just copy whatever's left over from the original binary to the patch.
    inp.seek(rng.tell())                # Go to end of what has been copied already
    while True:                          # Copy only the non-patch data
        data = inp.read(16 * 1024)      # Read 16KB at a time
        out.write(data)                 # Append each chunk to the patch file

        if len(data) < 16*1024:          # Break when no more data was read
            break




# This script doesn't work anymore because I changed how the game handles its
# randomizer seed. Instead of making a separate patch table like in Kenjiro
# Ryuugoroshi's project, this method uses offsets inside the game to overwrite
# the relevant values without needing to insert any functions.



    
def write_bytearray(out: BinaryIO, array: bytearray):
    out.write(struct.pack(">i", len(array)))
    out.write(array)


# Header format:
# [magic] ([len]) ([size]) [content]
MAGIC = b"CAFE"
HEADER_FORMAT = ">4s i"
HEADER_SIZE = struct.calcsize(HEADER_FORMAT)


def write_string(out: BinaryIO, string: str):
    encoded = string.encode()
    write_bytearray(out, encoded)


def make_patch(out: BinaryIO, targetname: str, contentstr: str):
    header = MAGIC + struct.pack(">I", HEADER_SIZE +
                                 len(targetname)+len(contentstr)) + targetname + contentstr
    write_bytearray(out, header)
    

def process_rngfile(infilename: str, outfilename: str):
    with open(infilename, "rb") as inf:
        with open(outfilename, "wb") as ouf:
            process_rngstream(inf, ouf)


def process_rngstream(inp: BinaryIO, out: BinaryIO):
    rng = inp
    header_format = '>4s i'
    header_size = struct.calcsize(header_format)
    magic = rng.read(4)
    assert magic == MAGIC, f"{magic}!= {MAGIC}"
    header = rng.read(header_size)
    _, datalen = struct.unpack(header_format, header)
    while datalen > 0:
        subheader = rng.read(struct.calcsize(">i"))
        _, sub_datalen = struct.unpack(">i", subheader)
        data = rng.read(sub_datalen)
        out.write(subheader + data)

        datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + data)
