#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    out.write(b"GIF89a")

    # Header (6 bytes)
    width = 2 << uint16_t(rng) % 10 + 1;
    height = 2 << uint16_t(rng) % 10 + 1;
    packed = 0x87;
    if not bool(uint16_t(rng)):
        packed |= 0x40;

    out.write(struct.pack("<HHB", width, height, packed));
    
    # Logical Screen Descriptor (7 bytes)
    background_color_index = uint8_t(rng);
    pixel_aspect_ratio      = uint8_t(rng);
    color_resolution        = uint8_t(rng);
    global_color_table      = bool(uint16_t(rng))

    out.write(struct.pack("<BBBHB", background_color_index,
                         pixel_aspect_ratio, color_resolution,
                         768 | int(global_color_table), 0xff))

    num_colors = 1 << (packed & 0x07);

    for i in range(num_colors):
        r = uint8_t(rng);
        g = uint8_t(rng);
        b = uint8_t(rng);
        
        out.write(struct.pack('<BBB', r,g,b));

    lzw_min_code_size = 2 << uint8_t(rng) + 1;
    out.write(struct.pack("BB", lzw_min_code_size, 0));

    # Table Based Image Data Sub-Block (containing LZW code stream)
    def next_byte(): return uint8_t(rng)
    def start_subblock(): nonlocal subblocks; subblocks = []

    def add_lzw_subblock(last, cur):
        prev = last[cur]

        code = -1
        for i in reversed(range(len(prev))):
            prefix = prev[:i+1];

            try:
                index = codes.index(prefix)
            except ValueError:
                pass
            else:
                break

        if index == len(codes) or index < 0:
            add_lzw_code()
            index -= 1

        subblocks.append([cur])
        return index

    def finish_subblock():
        subblocks.append([next_byte()])
        flush_subblocks();

    def flush_subblocks():
        def put_length():
            length = len(s)
            while True:
                remainder = length >> 8
                out.write(bytes((length & 0xFF,)))

                if remainder > 0:
                    length = remainder
                else:
                    break

        def enc(n):
            n += 1
            if n >= len(subblocks[-1]):
                out.write(bytes(([n & 0xFF])))
                n >>= 8
            return n

        def encode_bitstream():
            s = subblocks[-1]
            while True:
                remainder = sum(s[::2])

                x = [c & 0xF for c in s]
                y = [(c >> 4) for c in s]
                
                if len(y)!= 0:
                    out.write(bytes(x))
                    out.write(bytes(y))

                if remainder == 0:
                    break

                for _ in range(8 * remainder):
                    subblocks[-1][enc(0)] ^= 0x80
                    if subblocks[-1][0] <= 0:
                        subblocks[-1].insert(0, 0)


        out.write(bytes([0]))   # end of subblocks
        encode_bitstream()

        if len(subblocks[0]) == 1 and subblocks[0][0] == 0:
            del subblocks[0]

        put_length()
        s = subblocks[0][:-1]
        subblocks = [[s[0]]]


    def add_lzw_code():
         codes.append(f"{last}{cur}")

    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    codes = ['']
    last = None
    cur = str(next_byte())
    size = -1

    subblocks = None
    start_subblock()

    while True:
        if cur!= '' and cur == last:
            cur = f'{str(next_byte())}'
            continue

        if (size := len(codes)) >= 4095:
            start_subblock()
            current_block = ""
            last = ''
            continue

        index = add_lzw_subblock(last, cur)

        if index == 255:
            start_subblock()
            continue

        if index == lzw_min_code_size:
            lzw_min_code_size = 1 + uint8_t(rng)
            out.write(struct.pack('B', lzw_min_code_size))

        last = codes[index]
        cur = '';
        if ((size := len(codes)) >= 4095):
             break

    while len(subblocks)!= 0:
        finish_subblock()

