#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'
	width  = int.from_bytes(rng.read(1), 'big')  & 0xFF
	height = int.from_bytes(rng.read(1), 'big')  & 0xFF

	out.write(header + width.to_bytes(1,'little'))
	out.write(height.to_bytes(1,'little'))
	out.write((int.from_bytes(rng.read(1),'big')  & 0x7F).to_bytes(1,'little')) # Background color index
	out.write(b'\x00\x00') # Pixel aspect ratio

	palette = [
		b'\x00\x00\x00',
		b'\xff\x00\x00',
		b'\x00\xff\x00',
		b'\x00\x00\xff',
		b'\xff\xff\xff']

	for i in range(len(palette)):
		r = int.from_bytes(rng.read(1),'big') % 256;
		g = int.from_bytes(rng.read(1),'big') % 256;
		b = int.from_bytes(rng.read(1),'big') % 256;
		
		palette[i] += (r.to_bytes(1,'little') + g.to_bytes(1,'little') + b.to_bytes(1,'little'));

	out.write(struct.pack('%dB' % len(palette * 3), *sum(([c for c in x] for x in palette), [])))
	
	flags = bytearray([
		0x2C,                   # Image separator
		0,                      # Left corner x coordinate
		0,                      # Left corner y coordinate
		width,                  # Width of image
		height,                 # Height of image
		0x00                    # GCT used and local colour table
	])

	out.write(flags);

	lzwMinCodeSize = 2 if height > 1 else 1;

	if lzwMinCodeSize == 2:
		out.write(b'\x0A');                # Minimum code size
	else:
		out.write(b'\x09');

	plainText = bytes()
	compressedData = bytes()

	blockLength = ((height + 1) // 2)
	for j in range(blockLength):

		if lzwMinCodeSize < 2 or j < 2:

			symbol = ord('!');
			plainText += symbol.to_bytes(1, "big")
			compressedData += (lzwMinCodeSize - 1).to_bytes(1,"big");

		if lzwMinCodeSize <= blockLength:

			symbol = ord("!") + blockLength;
			plainText += symbol.to_bytes(1, "big")
			compressedData += (1 << 7).to_bytes(1,"big");

			for k in reversed(range(j*2+1)):
				symbol = ord("!") + blockLength - k

				if k!= j*2:
					plainText += symbol.to_bytes(1, "big")
					compressedData += (1 << 7 | 1 << 6).to_bytes(1,"big");
				else:
					compressedData += (1 << 7).to_bytes(1,"big");


		if j == 0:

			for i in range(((width + 1) // 2) * height):
				symbol = int.from_bytes(rng.read(1),"big") % 256
				
				plainText += symbol.to_bytes(1, "big")
				compressedData += (symbol >> 4).to_bytes(1,"big");
				compressedData += (((symbol & 0x0F) << 4) | (symbol >> 8)).to_bytes(1,"big");

		else:
			
			for i in range((((width + 1) // 2) * 2)-1):
				symbol = plainText[-1] ^ plainText[-2];
				plainText += symbol.to_bytes(1, "big")
				compressedData += (symbol >> 4).to_bytes(1,"big");
				compressedData += (((symbol & 0x0F) << 4) | (symbol >> 8)).to_bytes(1,"big");


	sizePlainText = len(plainText);
	packedPlainText = struct.pack('%dB' % len(plainText), *list(plainText))

	@functools.lru_cache(maxsize=None)	
	def getLZWCodeSizeAndValueFromPacked(idx: int) -> tuple:
		
		codeByte    = packedCompressedData[idx]
		value       = None
		nextIdx     = idx + 1

		if not codeByte & 0x80:
			value      = codeByte
			nextIdx    = idx + 1
		elif codeByte & 0x40:
			value = codeByte & 0x3f;

			while nextIdx < sizeCompressedData and \
				  (value := (value << 7) | (packedCompressedData[nextIdx] & 0x7F)) >= 0x80:
				nextIdx += 1

		return (nextIdx-idx, value)

	sizeCompressedData = len(compressedData);
	packedCompressedData = struct.pack('%dB' % len(compressedData), *list(compressedData))

	frameControlFlags = 0x41  # Transparency at pixel removed from frame
	frameControlFlags |= 0x08 # Use global colour table instead local colour table
	frameControlFlags |= (lzwMinCodeSize-1) << 4 # LZW minimum code size

	for h in range(height//2):
		for w in range(((width + 1) // 2)*2):

			key = h*w

			value2 = key % 256;
			value1 = (h + key // 256) % 256

			xOffset, value = getLZWCodeSizeAndValueFromPacked(0)
			yOffset, _    = getLZWCodeSizeAndValueFromPacked(xOffset)

			controlFlag = (frameControlFlags & 0xFE).to_bytes(1,"big")
			controlFlag += controlFlag[::-1]

			out.write(controlFlag)

			out.write(value1.to_bytes(1,"big"))
			out.write(value2.to_bytes(1,"big"))
			out.write(yOffset.to_bytes(1,"big"))
			out.write(xOffset.to_bytes(1,"big"))


	endOfFile = b'\x3B';
	out.write(endOfFile)
