#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The
# argument `resolution` is in pixels (horizontal by
# vertical). The colors of the generated image are
# saturation-value pairs selected from the set [S, V]
# where S and V are taken to be integers uniformly
# distributed on [0, 255].
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, resolution: tuple[int, int], svs_range: tuple[int, int]
    ):
    def read_color():
        return bytes([
            round((svs_range[0]+rng.read(1)[0]) / 2 ** 7 * 255),
            round((svs_range[1]+rng.read(1)[0]) / 2 ** 7 * 255),
            0 # constant black gradient
        ])

    header = b'GIF89a'
    width, height = resolution
    colortable = b''
    for i in range(2**(width > height)):
        colortable += read_color()
    palettenumcolors = len(colortable)//3 # 256**3 should be more than enough...
    imagedescriptor = bytearray([
        # left position
        0, 0,

        # top position - can be negative
        min(0, -(height*8)),

        0b10000 | (width > height << 7) | ((palettenumcolors&0xf)<<4),
        (width%256), (height%256)
    ])

    out.write(header)
    out.write(imagedescriptor)
    out.write(colortable)

    remlen = lambda x : max(x-(1+len(header)+len(imagedescriptor)+len(colortable)), 0)

    while True:
        chunksize = min(remlen(0x10000 + 12), 0xfff)

        if not chunksize: break

        out.write(chunksize.to_bytes(2, 'big'))
        out.write(b'\x87\x00')
        out.write(rng.read(chunksize))

    out.write(bytes([0]))