import sys

import struct
import argparse

from typing import BinaryIO


class GifHeader:
    def __init__(self):

        self.magic        = None   # (always "GIF")
        self.version      = None   # (always 0x07)
        self.width        = None   # in pixels
        self.height       = None   # in pixels
        self.flags        = None   # bitfield, see below

    @staticmethod
    def _parse_int16le(data: bytearray): return int.from_bytes(data[0:2], 'little', signed=False)
    @staticmethod
    def _parse_uint16be(data: bytearray): return int.from_bytes(data[0:2], 'big',    signed=False)
    @staticmethod
    def _parse_uint8(data: bytearray):     return int.from_bytes(data[0:1], 'big',    signed=False)
    
    def parse(self, data: bytearray):

        if not len(data) >= 6: raise ValueError("Not enough data to read the header!")
        
        self.magic        = GifHeader._parse_uint8(data[0:1])       == ord('G') and \
                            GifHeader._parse_uint8(data[1:2])       == ord('I') and \
                            GifHeader._parse_uint8(data[2:3])       == ord('F')
        self.version      = GifHeader._parse_uint8(data[3:4])       == 0x07
        width, height       = GifHeader._parse_uint16be(data[4:6]), \
                              GifHeader._parse_uint16be(data[6:8])
        self.flags        = GifHeader._parse_uint8(data[8:9])       | \
                            ((GifHeader._parse_uint8(data[9:10])) << 8) | \
                            ((GifHeader._parse_uint8(data[10:11]))<< 16) | \
                            ((GifHeader._parse_uint8(data[11:12]))<< 24)
        self.width        = width
        self.height       = height

    def dump(self): print(f"[{hex(self.magic)}] {self.width} x {self.height}")


class GifChunk:
    def __init__(self):
    
        self.id                  = None   # 32 bits tag
        self.remaining           = None   # length left for this chunk in bytes
        self.data                = None   # raw binary stream
        self.subchunks           = []     # list of sub-chunks

    TAG_IMAGE_SEPARATOR          = [0x2C, ]
    
    TAG_APPLICATION             = [0xFF, 0xFB, ]
    TAG_APPLICATION_NETSCAPE    = [0x53, 0xFE, ]
    TAG_APPLICATION_XMP         = [0xFF, 0xE1, ]
    TAG_COMMENT                 = [0xFF, 0xFE, ]
    TAG_GRAPHIC_CONTROL_EXTENT  = [0xF9, ]
    TAG_PLAIN_TEXT              = [0x01, 0x00, ]
    TAG_END_OF_FILE             = [0x3B, ]

    def parse(self, data: bytearray):

        assert type(data) is bytearray or type(data) is memoryview

        if len(data) < 10: 
            # Not enough data to be valid... just bail!
            return

        self.id            = tuple([GifHeader._parse_uint8(data[i:i+1]) for i in range(0, 4)])
        self.remaining     = GifHeader._parse_uint16le(data[4:6]) 
        self.data          = data[6:(6+self.remaining)]

        offset = 6 + self.remaining
        while True:

            if len(data) <= offset: break

            chunk = GifChunk()
            
            try:
                chunk.parse(data[(offset):((offset)+chunk.remaining+10)])
            except AssertionError as e: 
                print(e.args)
                continue
                
            self.subchunks.append(chunk)
            offset += (len(chunk.data))

    def dump(self, prefix=""): 

        print(f"{prefix}[ -{'-'.join([hex(c)[2:] for c in self.id])}- ({len(self.data)}) ] ")
        print(f"{prefix}\tRemaining : {self.remaining}")

        for chunk in self.subchunks: chunk.dump(f"\t\t")


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = GifHeader()
    chunks = []

    # Read full header first
    data = bytearray(rng.read(13)) 
    header.parse(data) 
    chunks.append(GifChunk())
    chunks[-1].parse(data)

    # Then process other chunks until we get something that looks like an image separator
    while not any([chunks[-1].id == chunk for chunk in GifChunk.TAG_IMAGE_SEPARATOR]):

        data = bytearray(rng.read(13)) 
        if len(data)!= 13: break
        chunks.append(GifChunk())
        chunks[-1].parse(data)

    # We got ourselves an image separator! Let's write it to disk now

    # First let's construct our own version of the header based on what we parsed from the input file
    header = GifHeader()
    header.magic = 0x47
    header.version = 0x07
    header.width = chunks[-1].subchunks[0].data[0]
    header.height = chunks[-1].subchunks[0].data[1]
    header.flags = 0x0800
    data = bytearray(struct.pack("<ccHHBBBB", 
                                b'G',b'I',
                                header.width, header.height,
                                header.flags & 0xff, 
                                0,0,0,
                                ))
    # Now let's add all those chunks to the output file
    for chunk in chunks:
        data += chunk.data

    out.write(data)


