#!/usr/bin/env python3


import struct

from PIL import Image
from itertools import zip_longest
from typing import List, Tuple




def grouper(n, iterable):
    args = [iter(iterable)] * n
    return ([e for e in t if not isinstance(e, str)] for t in zip_longest(*args))



# Generates a random animated GIF file with colors
# sampled from `color_source`, which is an iterator over
# RGB triplets. `image_size` is a tuple specifying the
# dimensions of all frames. The file is written to `out`.
def generate_random_animated_gif(
        color_source, *,
        image_size: Tuple[int], delay=200, loopcount=None,
        out: BinaryIO
    ):

    def write_graphic_control_extension():

        flags = 0b1 << 0   # disposal method 0
        flags |= 0b1 << 3  # user input

        duration = bytes([delay])

        block = bytes((
            0x21,    # extension introducer
            0xF9,    # graphic control label
            0x04,    # block size
            0x08,    # packed field

            flags,      # flags
            *duration,  # frame delay

            0,          # transparent color index
            0           # block terminator
        ))


        out.write(block)


    def write_image_descriptor(width, height):

        flags = 0b11100000 | 0b10000000         # local color table flag set, interlace and sort disabled
        flags |= 0b111 << 5                      # bits per pixel is 8
        flags |= 0b11 << 7                       # no transparency

        block = bytes((
            width & 0xFF,                   # image left position
            (height >> 8) & 0xFF,           # image top position
            height & 0xFF,                  # image height low byte
            (height >> 8) & 0xFF,           # image height high byte
            flags,                          # packed field
            0                               # lzw minimum code size
        ))

        out.write(block)



    # write out the header
    out.write(b'GIF89a')

    # the logical screen descriptor
    # we'll hardcode the background color to black
    # but leave everything else as zero
    canvas_width, canvas_height = image_size

    screen_descriptor = bytes((
        0b11100000 | ((canvas_width // 256) & 0b00000111),     # packed field
        canvas_width % 256,                                     # screen width
        (canvas_height // 256) & 0b00000111,                    # screen height
        canvas_height % 256,                                    # bit depth
        0b11111100,                                             # GCT color resolution
        0b00000000                                              # sort flag
    ))
    out.write(screen_descriptor)

    # global color table
    palette = []
    for i, rgb in enumerate(grouper(3, color_source)):
        r, g, b = map(lambda x: round(x*255), rgb)
        palette += [r,g,b]             # add triplet to palette

    palette += list(range(len(palette)//3, 2**16))       # pad the rest of the palete with black
    assert len(palette) == 2**16                         # make sure our palette has full coverage

    out.write(struct.pack('>H', len(palette)))            # global color table length (in entries)
    out.write(bytes(bytearray(palette)))                 # global color table

    # now start writing the image data

    # first up, a single graphics control extension
    write_graphic_control_extension()

    # then a single image descriptor
    write_image_descriptor(canvas_width, canvas_height)

    # finally the image itself
    # since this is an animated GIF, we need to write out the whole thing repeatedly
    # until we hit the loop count limit
    while True:

        img = Image.new("P", image_size)
        pixdata = img.load()

        for y in range(img.size[1]):
            for x in range(img.size[0]):
                col = next(color_source)              # get a new color sample

                # normalize the channel values (they should already be between 0 and 1)
                # and convert them to integers in the required range
                r, g, b = map(lambda c: min(max(c*255, 0), 255), col[:3])


                idx = img.putpixel((x,y), (r<<16)|(g<<8)|b)



        # back-to-front, convert the image to raw pixels
        pixdata = list(img.getdata())

        # break it down into groups of three bytes
        # because these are RGB triplets
        pixdata = list(zip_longest(*(pixdata[i::3] for i in range(3))))
        pixdata = sum(map(list, pixdata), [])

        out.write(bytes(bytearray(pixdata)))
        yield img                                            # pass along the current image

        if loopcount is None or loopcount > 0:
            loopcount -= 1
        else:
            break

    # end of gif; write final stuff
    out.write(bytes([
        0x3B                                                # terminator character
    ]))