#!/usr/bin/env python3

import os
import sys
import time
import math
import argparse
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (`/dev/urandom`)
def generate_random_gif(
        args: argparse.Namespace
    ) -> None:
    with open('/dev/urandom', 'rb') as rng:

        # We write in 256 bytes blocks so that their size is constant
        # (otherwise, we would need to keep track of how much space was
        # currently written for each block).
        buffer = bytearray()

        # Write the HEADER
        buffer += b'GIF89a'

        # Write the logical screen descriptor
        width  = args.width if args.width else randint(4, 1024)
        height = args.height if args.height else randint(4, 1024)
        color_table_size = min((math.log2(args.color_count) - 1), 7) + 1 if args.color_count else 0
        background_color = randint(0, 2 ** ((color_table_size * 3))) | 0x80
        aspect_ratio = 0 # Ignored for now
        packed = (
            ((color_table_size & 0b111) << 4) |  
            ((background_color >> 16)      ) | 
            (((aspect_ratio   ) & 0b11) << 8))        
        buffer += struct.pack('!HHBBBB', 
            width, 
            height,
            0xF7, 
            0,        # Global Color Table flag and sort flag should be set to
                     # ignore local color table and use global one instead
            packed    # Pack all the info together
        )

        # Write the image descriptor
        buffer += struct.pack('!BBBHHBBBBB',
            0x2C,  # Image separator
            0,     # Left position of the image inside the canvas
            width, # Width of the image on the canvas
            height,# Height of the image on the canvas
            0,     # No local color table index specified
            False, # LZW compression
            0,     # No interlace specification
            0,     # No filter specified
            0      # Default delay time is used
        )

        # If a custom palette has been provided, then it is used
        # instead of generating a new one
        if len(args.palette):            
            buffer += args.palette

        # Otherwise, we create a new palette from scratch
        else:

            # Generate the colors to be stored in the global color table
            color_table_colors = [
                (
                    i[0],       # red
                    i[1],       # green
                    i[2]        # blue
                ) for i in [ tuple(randint(0, 255) for j in range(3)) \
                        for i in range(int(2 ** color_table_size))]
            ]         
            
            # Add them to the output stream
            for color in color_table_colors:                
                buffer += (color[0]).to_bytes(1, byteorder='big')
                buffer += (color[1]).to_bytes(1, byteorder='big') 
                buffer += (color[2]).to_bytes(1, byteorder='big') 

        # Get the count of frames to be generated 
        frame_count = args.frames if args.frames else randint(1, 100)
        
        # Start writing the data section
        starttime = time.time()
        while len(buffer) < (frame_count * width * height * 3):

            # Use the next byte from the random generator to decide whether or not
            # we should add a pixel into our current buffer
            val = ord(rng.read(1)) % 0xFF
            
            # Get the current length of the buffer
            buflen = len(buffer)            
                        
            # Check if this value can still be added into the buffer
            if buflen >= (frame_count * width * height * 3):

                print("Buffer overflowed")
                break

            # If yes, then do it
            elif val <= args.density:

                # Test if the buffer contains enough space left for another pixel
                if buflen > (frame_count * width * height * 3):

                    # Pop off some pixels until there are no more than 
                    # two empty lines at the end of the buffer                    
                    while True:

                        # Try popping
                        try:
                            del buffer[-3:]                        

                        except IndexError:

                            # If we get an error because the buffer cannot be
                            # decreased by three elements, then stop trying to
                            # remove these pixels                            
                            break
                        
                        # Stop removing pixels when we have exactly two empty 
                        # lines at the bottom of the buffer
                        lasttwoemptylines = True
                        for l in range(-3,-(width*3)-1, -(width*3)):
                            if sum(buffer[l:(l+3)])!= 0:
                                lasttwoemptylines = False

                        if lasttwoemptylines == True:
                            break
                
                # Append the pixel
                buffer += b'\xFF\x00\x00'
                
            # If not, skip adding anything
        out.write(buffer)        
            
        # Calculate duration and speed
        duration = round(time.time() - starttime, 2)
        speed = "{:.2f} frames per second".format(
            float(frame_count) / float(duration))        
                
        # Display result in the terminal
        print("Generated {} frames ({}) in {:.2f} seconds".format(
            frame_count,
            speed,
            duration
        ))
                
