#!/usr/bin/env python3

"""Generate random GIFs

This script will generate a random GIF image according
to the GIF specification version 89a. It will produce
a random color table and also use the full palette of
the available colors."""

__author__      = "<NAME>"


# TODO allow custom output name
# TODO clean up CLI usage message formatting
# TODO improve help messages
# TODO support multiple files? or only one input file?

# All of these are defined on page 15 of spec v89a
magic                = b'GIF89a'
width                = bytearray([])
height               = bytearray([])
bits                 = None
background           = None
aspect_ratio         = None
color_table_size     = None
global_color_table   = []
blocks               = {}     # dict of block types by block size
image_data           = bytearray()
block_mode           = ''

# Gif color resolution, bits per pixel
colors       = {
    1 : [2],
    2 : [4],
    4 : [8],
    8 : [16, 24, 32]
}

# Available global color entries
color_table  = list(range(2**24))

# Global color entry recycling bin
recycled     = set()


def parse_arguments():
    parser = argparse.ArgumentParser(description='Generate random GIF images.')
    parser.add_argument('--output', default=DEFAULT_OUTPUT)
    return parser.parse_args()


def invalid_parameters(**kwargs):
    pass


def build_header():
    """Build the GIF header from the previously generated variables

    This function builds the GIF header from the above meta-variables, 
    including the constructed global color table, which was stored in reverse
    order so that we can simply pop off values when they become needed.
    """
    current_block = blocks[current_block_size()]
    init_block(current_block)
    current_block['header'] = magic[:] + \
                              width[:2]            + \
                              height[:2]           + \
                              bytes([color_table_size])
    while len(current_block['entries']) < current_block['header'][7]:
        new_entry = bytes([next(recycled),
                           next(recycled),
                           next(recycled)])
        current_block['entries'].append(new_entry)
    header = current_block['header'][:]
    [header.extend(_) for _ in current_block['entries']]
    return header


def init_block(block_type=''):
    """Initialize a block type's variables
    
    A dictionary of various variables corresponding to each
    block type should be made before this initialization step. 
    If none exists yet, then make some basic ones. Return the
    newly initialized block
    """
    if block_type == '':
        global blocks
        blocks = {
            13 : {},
            12 : {}
        }
        keys = sorted(list(blocks.keys()))
        block_type = blocks[keys[-1]]
    else: 
        assert isinstance(block_type, dict)
    for key in ['code', 'header', 'entries']:
        if key not in block_type.keys():
            block_type[key] = []
    return block_type


def ensure_correct_size(num):
    """Ensure a given integer has two bytes"""
    num_bytes = int.to_bytes(num, length=2, byteorder='little', signed=False)
    return num_bytes


def current_block_size():
    return bytes([len(build_header())+1])


def gen_random_color():
    """Get a random RGB value"""
    r = random.randint(0, 255)
    g = random.randint(0, 255)
    b = random.randint(0, 255)
    rgb = bytes([(r << 5 | r >> 3), 
                 (g << 5 | g >> 3),
                 (b << 5 | b >> 3)])
    if rgb not in color_table:
        color_table.append(rgb)
    return rgb


def main():

    if not os.path.exists('images/'):
        os.makedirs('images/')

    args = parse_arguments()
    output = args.output

    try:
        f = open(output, 'wb')

        # Write the header
        print('Generating {}'.format(output))
        f.write(magic)
        f.write(ensure_correct_size(width))
        f.write(ensure_correct_size(height))
        f.write(bits)
        f.write(background)
        f.write(aspect_ratio)
        f.write(ensure_correct_size(color_table_size))
        
        # Build global color table
        for i in range(color_table_size):
           f.write(gen_random_color()) 

        # Write out initial image data
        f.write(current_block_size())
        f.write(bytearray([]))

        # Write the rest of the file
        while True:
            chunk = f.read(BLOCK_SIZE)
            if not chunk: break
            f.write(chunk)
        f.close()

    except Exception as e:
        print("Couldn't create {}".format(output))
        print("{}".format(e))


if __name__ == '__main__':
    main()