#!/usr/bin/env python3

import sys
import os
import random
import math
import argparse


# Generates a random GIF file at `path`.
def generate_random_gif(
        path: str, size=1024*16, color_depth=7, max_colors=256):

    # Color depth must be in {1, 2, 4, 8}
    if not (color_depth == 1 or color_depth == 2 or \
                color_depth == 4 or color_depth == 8):

        print('Invalid color depth {}'.format(color_depth))
        return False;

    # Data block must be large enough to hold every color for each pixel.
    min_data_block = ((size ** 2 * color_depth) // 8) + 1

    # Create the directory if it doesn't already exist
    dir_name = os.path.dirname(os.path.abspath(path))
    if len(dir_name) > 0 and not os.path.exists(dir_name):
        try:
            os.makedirs(dir_name)
        except OSError as e:
            pass
    
    with open(path, 'wb') as f:
        width = height = int(math.sqrt(size))
        
        palette = [i % 256 for i in range(max_colors)]
        image = []
        n_bits = width * height * color_depth
        n_bytes = (n_bits + 7) // 8

        # Generate random data
        data = bytes([random.randint(0, 2**8 - 1)
                        for _ in range(min_data_block)])
        while len(data) < n_bytes:
            data += bytes([random.randint(0, 2**8 - 1)
                            for _ in range(min_data_block)])
        data = data[:n_bytes]

        # Decode the pixels from the binary data
        pixels = []
        p = 0     # Current index of the pixel array
        for v in data:

            # For each bit in this byte...
            for bit in '{0:0>8b}'.format(v):

                # If it's set then add that color to our list of pixels
                if bit == "1":
                    pixels.append(palette[p % len(palette)])
                else:
                    pixels.append(None)
                
                # Move to the next pixel
                p += 1

        # Write the GIF file
        f.write(b'GIF89a')         # File magic
        f.write(b'\x00\x00\x00\x00')   # Screen descriptor (not used)
        f.write(f'{width}\x00'.encode())    # Width
        f.write(f'{height}\x00'.encode())   # Height

        # Set global color table
        f.write((len(palette)).to_bytes(1, 'little'))   # Size of global table
        for c in palette:                          # Write all colors
            f.write(c.to_bytes(1, 'little', signed=True))

        # Automatic animation flag
        f.write(b'\x80\x00')               # Flags
        
        # Image descriptor
        f.write(b'\x2C\x00\x00\x00\x00\x00,\x00\x00\x00')        # Left corner x, y
        f.write(b'\x00\x00\x00\x00\x00\x00')                        # No local table

        # LZW minimum code size
        lzw_code_size = color_depth + 1
        f.write(lzw_code_size.to_bytes(1, 'little'))

        # Block length
        lzw_data_len = 0
        f.write(b'\xFF\x0B')           # Application extension block
        f.write(b'NETSCAPE2.0\x03')   # Extension Label
        f.write(b'\x01\x00')           # Sub-block Version
        f.write(b'\x00\x00\x00')       # Unknown
        f.write(b'\xFF\x00')           # Application Block

        # Loop count sub-block type
        f.write(b'\x01')               # Netscape loop count identifier
        f.write(b'\x00\x00\x00')       # Loop count (forever)

        # Encode the LZW compression bits
        # This is an implementation of the LZW algorithm, which needs some
        # explanation.
        # 
        # The LZW algo works by taking chunks of bits and progressively building up
        # a lookup dictionary mapping those chunks to shorter chunks. It starts off
        # with one byte per chunk and builds up to four bytes per chunk until only
        # a single byte remains after eight consecutive chunks. In other words,
        # once we have enough unique values that they won't fit into two bytes,
        # there are no more opportunities to build up a shorter representation.
        # 
        # Once we hit the maximum value that can be represented by its own byte,
        # we start adding null bytes to the front so that we can always represent
        # any possible value.
        # 
        # The format of the compressed data is as follows:
        # 
        #   [NULL BYTE ("No Code")] | [INDEX OF PARENT CODE]
        #
        # A short example should illustrate how this algorithm works better than I can:
        # ```
        # Suppose you have a code size of three bits. That means there will be
        # a total of 2^3 = 8 codes available. Also suppose we have these five strings:
        #   XYZ      -> 000
        #   YZX      -> 001
        #   ZXY      -> 010
        #   WXYZ     -> 011
        #   XYZW     -> 100
        # We could encode this with a dictionary like this:
        #       "" : 000
        #       X  : 001
        #       YZ : 010
        #       ZXW: 100
        #       WY : 011
        # Since 000 has multiple children but none of them are valid yet, we store NULLs:
        #   NULL : 000
        #   X    : 001
        #   YZ   : 010
        #   ZXW  : 100
        #   WY   : 011
        # Now let's pretend we get a new string:
        #   ABCDEFGHIJKLMNOPQRSTUVWXYZ -> 10110000110000000100110011
        # We first split it into its constituent parts:
        #   10110000110000000100110011
        #   1011 -> 1011
        #           0000 -> 0000
        #         110000 -> 110000
        #       00000100 -> 0100
        #     100110011 -> 100110011
        # Now consider what happens when we search through the dictionary for prefixes
        # of this stream. Let's say we're looking for "AB".
        # Notice that in order to find "AB", we need to look at the dictionary entries
        # underneath AB. So, we see that the prefix "A" points to no entry.
        # Then we follow the pointer for "A" and see that it points to "AB". But,
        # since "ABCDE" is longer than "AB", we keep searching recursively. However,
        # since now we look for "BC", we realize that "BC" exists directly under "AB".
        # We then make a note of what we found along the way.
        # Finally, we end up finding a match for BC under "AB", so we update our table like so:
        #  ...
        #   ABCD: 0000
        # After repeating this process until we've exhausted the entire string, we
        # now have the following values stored in the dictionary:
        # NOTE: WE ONLY STORE THE LAST FOUR VALUES BECAUSE THIS IS AN EXAMPLE. IN REAL
        # USE YOU SHOULD STORE ALL 2^CODESIZE ENTRIES.
        #   ABC : 0000
        #   BCDC: 0001
        #   CDDE: 0010
        #   DEEF: 0011
        #   EFFG: 0100
        #   FGHI: 0101
        #   GHIJ: 0110
        #   HIJK: 0111
        #   JKLM: 1000
        #   KLMN: 1001
        #   LMNO: 1010
        #   MNOP: 1011
        # So now, if we come across another string "AC