#!/usr/bin/env python3

import os
import struct
import sys

from typing import BinaryIO


# The GIF header only has 6 bytes (per section 21 of the spec).
# It is therefore not suitable to be used as template for writing arbitrary
# data from /dev/urandom without modifications -- we need to know it's size
# in advance and handle overflows manually. Luckily, though, this can be done
# very easily by reading ahead!

# We read just enough data from the input file (`infp`) so that we can determine
# how much data to write to the output file (`outfp`). If there are less than
# 6 bytes of input, we raise an exception. If there are more, we simply stop
# reading and write them to the output file immediately. This is OK because all
# valid headers are 6+ bytes long.

# Returns a tuple containing the GIF header, and its length in bytes
def get_gif_header(
        infp: BinaryIO
    ) -> [bytes, int]:

    # Read first three bytes
    possible_header = infp.read(3)
    if len(possible_header) < 3:
        raise Exception('Not enough data available from /dev/urandom!')

    # Determine whether these are indeed the magic numbers of our target GIF
    if possible_header == b'\x47\x49\x46':
        return ([possible_header], 3)
    elif possible_header[0]!= ord('#'):
        return ([b'#', *possible_header], 4)

    # Else we have found a comment; seek back one byte
    infp.seek(-1, os.SEEK_CUR)
    
    # Read next two bytes to determine block size
    blocksize = infp.read(2)
    if len(blocksize)!= 2 or blocksize[:1]!= b'!':
        raise Exception('Invalid GIF header detected!')
    blocksize = int(blocksize[1:], base=16)

    # Read next (at least one) full block (to avoid edge cases when we hit EOF)
    data = infp.read(min(blocksize, 6))
    if len(data) < 6:
        raise Exception('Not enough data available from /dev/urandom!')

    # Return the header
    return ([*possible_header, *blocksize.to_bytes(2, 'big'), *data],
            7 + min(len(data), blocksize - 6))



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header_template, header_length = get_gif_header(rng)

    header = []
    while True:

        # Copy up to header_length bytes of the original header
        header += header_template[:header_length]
        
        # Read until a newline character appears
        line = []
        char = None
        while char!= '\n' and char is not None:
            try:
                char = chr(ord(rng.read(1)))
                line.append(char)

            except OSError:
                break

        # Update header with data read from the RNG
        header += ''.join(line)

        # Write header to the output file
        outfp.write(''.join(header).encode())

        # Reset variables
        header[:] = []
        header_length = 0



