#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    
    # We want to use this frame as many times as possible
    img_frame_data = bytes()

    GCT_size = 0b00100000                     # Global Color Table Flag set
    sorted_colortable = False                  # Sort Flag set to false (not required by spec)
    color_resolution = 0b000100               # Color Resolution field is one bit (not required by spec)
    num_bits_per_pixel = 0b0110              # Bits per pixel for default GCT * 256 (not required by spec)
    num_colors_in_palette = 0                 # No colors in palette since we don't have a GCT
    background_color_index = None             # Not relevant since there's no GCT
    pixel_aspect_ratio = 0                    # Only values greater than zero are allowed

    # Build up the GIF header so we can place it at the start of our output stream
    header = b'\x47\x49\x46\x38\x39\x61'                 # Signature and version
    header += struct.pack("<H", image_width)          # Image Width
    header += struct.pack("<H", image_height)         # Image Height
    header += bytes([                                       # Packed Fields
        GCT_size | num_bits_per_pixel,                   # Global Color Table Size
        0,                                              # Background Color Index
        pixel_aspect_ratio                              # Pixel Aspect Ratio
    ])

    # Write the header to the output stream
    out.write(header)


    # Generate the global color table if needed
    global_color_table = []
    if len(global_color_table):

        # If the size of the global color table exceeds 2^n 
        # then we need more than one block
        while True:

            current_block_size = 2 ** (num_bits_per_pixel+1)
            num_blocks_needed = len(global_color_table) // current_block_size
            
            # Create an empty block
            current_block = bytearray([])

            # If we're not going to exceed the maximum entries 
            # per block then we can fit all the color entries
            # into one long block
            if num_blocks_needed == 0 or \
                num_blocks_needed % 2!= 0:

                max_entries_for_this_block = 2**num_bits_per_pixel
                for entry_index in range(len(global_color_table)):

                    red, green, blue = \
                        tuple(
                            global_color_table[entry_index])
                    rgbas = [red >> 4, 
                            green >> 4, 
                            blue >> 4]

                    packed_rgb = sum([(v << i*4) for v,i 
                                        in zip(rgbas,range(3,-1,-1))
                                        ])

                    current_block += struct.pack("B",packed_rgb)


                num_blocks_to_write = 1
            else:
                pass


            num_blocks_written = 0

            # Write the block to the output
            out.write(current_block)



    # We now know how big our data section will be
    # So lets write the appropriate length value after 
    # the global color table
    num_color_entries_in_gct = len(global_color_table)
    gce = bytes([                                        # Graphic Control Extension
        num_color_entries_in_gct & 0xFF,                # Block Size
        ((num_color_entries_in_gct >> 8)& 0xFF),        # Block Size
        ((num_color_entries_in_gct >> 16)& 0xFF),       # Block Size
        background_color_index& 0xFF,                  # Background Color Index
        (((background_color_index>>8) & 0b1111)<<4),   # Global Color Table Indicator
        (((background_color_index>>8) & 0b11110000)>>4) | (sorted_colortable<<3)|((color_resolution<<7)),            # Flags
        num_bits_per_pixel,                             # LZW Minimum Code Size
    ])
    print(f"GCE: {gce}")
    out.write(gce)




if __name__=='__main__':

    with open('/dev/urandom','rb') as f:

        try:
            input_file_path = sys.argv[1]
            with open(input_file_path,'wb') as o:
                generate_random_gif(f,o)
        except IndexError:    
            generate_random_gif(f,sys.stdout.buffer)