#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The frame header is 7 bytes long then follows by an image descriptor
    # and finally the image itself (the animation control block is ignored).
    # See https://www.matthewflickinger.com/lab/whatsinagif/bits_and_bytes.asp
    def write_frame() -> None:

        # Frame header
        # Offset | Length   | Description
        # ------ | -------- | ---------------------------------
        #   0    |    3     | Header ("GIF")
        #   3    |    3     | Version ("87a" or "89a")
        #   6    |    2     | Whitespace to pad up to next byte boundary
        offset = len(data)
        data += b'\x47\x49\x46' # GIF signature and version

        # Image descriptor
        # Offset | Length   | Description
        # ------ | -------- | -------------------------------
        #   0    |    1     | Image separator ("!")
        #   1    |    1     | Packed field:
        #                    * Bit 0:     Local color table flag
        #                                Set if local color table exists
        #                    * Bits 1..3: Reserved for future use
        #                    * Bit 4:     Interlace flag
        #                                Set to zero when not interlaced
        #                    * Bit 5:     Sort flag
        #                                Set to zero when entries are not sorted
        #                    * Bit 6:     Reserved for future use
        #                    * Bit 7:     Size of global color table
        #                                1 - 2 colors
        #                                2 - 4 colors
        #                                3 - 8 colors
        #                                4 - 16 colors
        #                                5 - 32 colors
        #                                6 - 64 colors
        #                                7 - 128 colors
        #   2    |    2     | Screen width in pixels (lower nybble)
        #   4    |    2     | Screen height in pixels (upper nybble)
        #   6    |    1     | Upper bit of bits per pixel
        #                        Valid values are between 0 and 7 inclusive
        size = rng.randint(32, 128) // 8
        data += b'\x21\xf9' + \
            len(struct.pack('=L', size << 4)).to_bytes(length=1, byteorder='little') + \
            struct.pack('<H', rng.randint(2, 4)) + \
            3.to_bytes(length=1, byteorder='little')

        # Color table
        # This section only appears if there's at least one color
        # It should be ordered such that the first element contains all the
        # black shades followed by the white ones so it can later be used as a LUT
        # to map arbitrary RGB tuples into colormap indices
        num_colors = min((size / 8), 16)
        palette = []
        for _ in range(num_colors):
            r, g, b = rng.randint(
                0, 255), rng.randint(0, 255), rng.randint(0, 255)

            while any([r == c[0] and g == c[1] and b == c[2]
                    for c in palette]):
                r, g, b = rng.randint(
                    0, 255), rng.randint(0, 255), rng.randint(0, 255)

            palette += [(r,g,b)]

        assert(len(palette) <= 2**(size / 8))

        if size < 128:
            keys = list(range(len(palette)))
            val = lambda key: palette[key][0:3]
        else:
            keys = [i*256+j for i in range(256) for j in range(256)]
            val = lambda key: palette[(key % 256)*256//256 +
                                     key//256][0:3]

        lookupTable = {}
        for idx, key in enumerate(keys):
            lookupTable[val(key)] = idx

        outpal = [lookupTable[c]
                  for c in zip(*[[p]*3 for p in palette])]

        rgbres = size // 8
        if size >= 128:
            rgbres -= 1

        assert(len(outpal) == rgbres ** 3)

        outpal = b''
        for i in range(rgbres**3):
            outpal += struct.pack("<B", outpal[i])
        out.write(outpal)


        # LZW minimum code size
        data += rng.read(1)[0].to_bytes(1, byteorder='little')

        # Image data
        # Every non-zero value in the compressed data stream represents
        # the index of the corresponding entry from the colormap above
        values = set().union(*(tuple(p)
                              for p in palette)).intersection(set(range(len(palette))))
        lzwstream = tuple([(rng.choice(values), idx)
                          for idx in range(len(palette))]).__iter__()

        lastval = next(lzwstream)[0]
        output = {lastval}
        while True:
            try:
                val, idx = next(lzwstream)
            except StopIteration: break
            if val in output: continue
            data += (output[-max(len(output)-1, 1)]+idx).to_bytes(
                4, byteorder='little')[1:]
            output.add(val)


    data = b''

    # Write the entire gif at once because we have no idea how much memory
    # it will take nor do we want to make this script depend on very large files
    # Instead let's assume that every single frame takes about ten bytes and
    # that our input comes from /dev/urandom which never gives us more than ~3MiB
    # of data in a single read
    for _ in range(int(3e6)):
        write_frame()

    out.write(data)

