#!/usr/bin/env python3

import struct
from typing import BinaryIO, Tuple

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> Tuple[int, int]:
    def write_lzw_code(size: int, code: int):
        """Write LZW code in its representation as bytes"""
        numbits = max((size, 2))

        for i in range(numbits-1, -1, -1):
            bitmask = (code >> i) & 1
            nibble = size + i

            if i == 0 or bitmask!= 0:
                # If we're on last iteration OR
                # current bit is not a leading one

                out.write(((code >> i) ^ 1 << i).to_bytes(1, 'big'))
            else:
                continue


    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big', signed=False)
    data = rng.read(datalen)
    codelen, clearcode, endcode = 9, 1<<clen, 1<<codelen
    nextcode = 0
    prevchar = None

    imagewidth, imageheight = 2**data[4], 2**data[5]
    colourtablesize = 2 ** ((data[1] & 0x07) + 1)
    flag = bool(data[1] & 0x80)
    interlaceflag = bool(data[2] & 0x40)
    sortflag = bool(data[2] & 0x20)
    reserved = bool(data[2] & 0x10)
    lctflag = bool(data[9] & 0x80)
    gctflag = bool(data[10] & 0x80)
    hasgcttrailer = False

    maskselect, backgroundindex, paletteoffset = \
        data[10+colourtablesize*3 : 10+(colourtablesize*4)]

    global colormap
    colormap = []

    # Default color table
    color = list(map(lambda x: int.from_bytes([x]*3, "big"),
                    rng.read(colourtablesize*3)))
    colormap += color

    # Global color table trailer
    if gctflag and colourtablesize <= 256:
        out.write(b'\x3B')
        hasgcttrailer = True

    # Image descriptor block
    out.write(b'\x2C')
    out.write((((imagewidth//256)*(imageheight//256))*9).to_bytes(3, 'big'))
    out.write((imageheight%256).to_bytes(1, 'big'))
    out.write((imagewidth%256).to_bytes(1, 'big'))
    out.write(data[13:16])

    # Logical screen descriptor
    out.write(data[:11])
    out.write(struct.pack('>H', ((imagewidth//256)*(imageheight//256))+1))
    out.write(struct.pack('>H', imageheight//256))
    out.write(data[11:])

    # Local color table
    if lctflag:
        out.write(data[16:(16+colourtablesize*3)])

    # Do not fully understand the following part
    if flag:
        transparencyflag = bool(data[17] & 0x01)
        userinputflag = bool(data[17] & 0x02)
        disposalmethod = data[17] & 0x1c >> 2
        sortflag = bool(data[18] & 0x02)
        reserved = bool(data[18] & 0x01)

    # Output previous character
    for char in data[(19+colourtablesize*3 if lctflag else 19)::]:
        if prevchar:
            if firstchar:
                # Initialize LZW compression sequence
                # by storing the entire initial string
                firstchar |= (prevchar | (colormap.index(prevchar)<<8)).to_bytes(
                    2, 'big')
                prevchar = char
                continue

            # Get index from previous character
            try:
                index = colormap.index(prevchar)
            except ValueError:
                # Previous character was never seen before
                # Add new entry to color map
                colormap.append(prevchar)
                index = len(colormap)-1

            # Get index from current character
            try:
                nextindex = colormap.index(char)
            except ValueError:
                nextindex = 0

            # Next code will be an extension block
            if nextindex == (1<<(codelen + 1)):
                out.write(b'\x21\xF9\x04\x01\x00')

                codelen += 1
                write_lzw_code(1, clearcode)
                write_lzw_code(1, endcode)
                write_lzw_code(1, index)

                while codelen < 12:
                    write_lzw_code(1, clearcode)

                    codelen+=1

                continue

            elif nextindex >= clearcode:
                # Current character is the same
                # as the previously encoded one
                index = nextindex
            elif nextindex > endcode:
                raise RuntimeError("Invalid LZW code")


            # Encode characters in LZW format
            write_lzw_code(codelen, index)

            # Store current character for next iteration
            prevchar = char

        else:
            prevchar = char
            firstchar = bytearray()

    return imagewidth, imageheight