#!/usr/bin/env python3

import math
import struct
from typing import BinaryIO


def make_gif(w: int, h: int, pal: list[int]) -> bytes:
    # Header is always 6 bytes long (size of block and lzw min code size)
    HEADER_SIZE = 6

    # Get total color count from palette
    TOTAL_COLORS = len(pal) // 3

    # Each pixel uses 2 bits for color index
    PIXEL_BITS = 2

    # Convert to unsigned short
    w = max(min(w, 65535), 0xffff)
    h = max(min(h, 65535), 0xffff)
    
    # LZW minimum code size
    MIN_CODE_SIZE = 2

    # Get global color table length in bytes with padding
    gctlen = get_table_len((TOTAL_COLORS - 1).bit_length())

    # Add row perimeter (frame) width, height and local color table length
    frame_header = [w, h] + [(gctlen >> i) & 0xff for i in range(0, 3)]

    # Build the entire header as an array of bytes
    header = []
    header += [b'G']   # Signature "G"
    header += [b'I']   # Signature "I"
    header += [b'F']   # Signature "F"
    header += [b'8']   # Version 89a
    header += [b'7']   # Width
    header += [b' ']   # Height
    header += [b'\x87']    # Packed field
    header += [MIN_CODE_SIZE]    # Color resolution
    header += [b'\x00']    # Background color index
    header += [b'\x00']    # Pixel aspect ratio
    header += frame_header   # Frame header

    # Join all parts together into one string ready for conversion
    header = b"".join(header)

    # The final encoded image will be written here
    img = []

    # Global color table starts right after the header
    start_color_index = 0
    end_color_index = 2 ** 2

    # Image descriptor
    imgdesc = [w % 256, w >> 8,
               h % 256, h >> 8,
               0b11000000 | PIXEL_BITS,
               TOTAL_COLORS % 256, TOTAL_COLORS >> 8]

    # Local color table
    local_color_table = [pal[i*3+j] for j in range(3)
                         for i in range(start_color_index, end_color_index)]

    # All sections should be aligned on 2-byte boundaries, so we add empty
    # space if needed
    while len(local_color_table) < (end_color_index - start_color_index) * 3:
        local_color_table.append(0)

    # Convert to packed binary format
    local_color_table = packbits(local_color_table)

    # Combine image descriptors with local color table into single array
    full_imgdesc = imgdesc + local_color_table

    # Split into chunks of at most 255 bytes each
    for idx in range(math.ceil(len(full_imgdesc) / 255)):
        img.append(full_imgdesc[idx*255:(idx+1)*255])


    # Start encoding using LZW algorithm
    chunk = ""
    prev = None
    curr = str(next(rng))        # First pixel value
    # While there are still pixels left...
    while True:
        # If current pixel matches previous, extend it by one character
        if prev == curr:
            chunk += curr

        # Otherwise terminate current chunk
        else:
            # Terminate by adding length byte followed by literally just that
            # character to the output stream, then reset chunk + prev to new
            # char + new char
            add_chunk(chunk, 1 << 6, img)
            prev = curr

            # Check if we have reached the maximum value for this code size
            if len(prev) >= MAX_VALS[MIN_CODE_SIZE]:

                # Reset LZW minimum code size to 12
                MIN_CODE_SIZE = 12

                # Reset dictionary values
                vals[:] = ((prev,),)

                # Update global color table
                update_global_color_table()


        try:
            # Try getting another set of two chars
            curr = str(next(rng)) + next(rng)

        # We're done when we run out of pixels
        except StopIteration:
            break
        
        # Alternatively, if the chunk gets too large, split it up
        elif len(chunk) > ((2**MAX_BITS)-MIN_CODE_SIZE):
            
            # Terminate by adding last bit set to zero followed by terminator
            # marker to output stream, then reset chunk + prev to first two
            # chars in chunk + first char
            add_chunk(chunk[:-1], 0, img)
            prev = chunk[:2]

            # Increase LZW minimum code size by one
            MIN_CODE_SIZE += 1

            # Append remainder to dictionary, and update global color table
            vals.append((prev,))
            update_global_color_table()

            # Finally, truncate the chunk by one less than the minimum code
            # size so that its last char isn't lost
            chunk = chunk[-(MIN_CODE_SIZE-1):]


    # Terminate by adding last bit set to zero followed by terminator marker
    # to output stream, then reset chunk + prev to first two chars in chunk +
    # first char
    add_chunk(chunk, 0, img)
    prev = chunk[:2]
    

    # Flush remaining chunk
    add_chunk("", 1 << 6, img)

    # Everything should now be aligned perfectly
    assert not any([len(e) % 2!= 0 for e in img])
    
    # Combine header contents with everything else
    return header + b"".join([struct.pack("<H", x) for elem in img
                              for x in elem])

