#!/usr/bin/env python3

import struct
from typing import BinaryIO
import argparse

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width=256, height=256
    ):
    # Header for GIF files (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)
    header = bytearray([
        71, 73, 70,  56,  57,  65,  38,  00, 40])

    # Write header and screen descriptor to output
    out.write(header)
    out.write((width).to_bytes(2, byteorder="little"))
    out.write((height).to_bytes(2, byteorder="little"))
    out.write(b"\x87\x00") # Global color table flag set
    out.write(b"\x00") # Background color index
    out.write(b"\x00") # Pixel aspect ratio

    # Generate global color palette entries for random colors
    colors = [struct.unpack("BBB", rng.read(3)) for i in range(256)]
    for r, g, b in colors:
        out.write(bytearray([r,g,b]))

    # Get all pixel values on image (in order of appearance)
    pixels = []
    for y in range(height):
        row = []

        # For each row we have one bit per column indicating if there is an
        # animation frame or not.
        if rng.read(1)[0] % 5 == 0:
            # If this is odd then we use black as our background color so that
            # it appears to be transparent instead of having frames that clear
            # everything behind them with black.
            bg = (0, 0, 0)
        else:
            # Otherwise we choose a random color from our color palette. This
            # allows us to create multiple different shapes via overlapping
            # frames.
            bg = colors[int.from_bytes(
                rng.read(1), "little")]

        for x in range(width):
            if rng.read(1)[0] % 5 == 0:
                # Use the current background color since we are doing nothing here
                row += list(bg)
            else:
                # Choose another random color
                fg = colors[int.from_bytes(
                    rng.read(1), "little")]

                # Work out what proportion of the pixel should be changed by using
                # a coin flip. We need to consider both the foreground and
                # background color to ensure we don't always just change either
                # to the same value. Even if we do this you can still see tiny
                # variations between frames due to the way the GIF format works.
                p = rng.read(1)[0]/256
                row += [(fg[i] * p + bg[i] * (
                    1 - p)).to_bytes(
                        1, byteorder="little")[0]
                        for i in range(3)]


        # Now take these RGB bytes and write them to the output stream as raw
        # data.
        out.write(bytearray(row))


if __name__=="__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-o", "--output", type=argparse.FileType('wb'), default="-")
    args = parser.parse_args();
    generate_random_gif(open("/dev/urandom","rb"), args.output);