#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        minlen: int=2**14, maxlen: int=2**28,
        repeats=True
    ) -> None:
    # Generate a valid GIF header from binary data
    def gen_header() -> bytes:
        return (
            b"GIF89a" +
            struct.pack(">L", rng.randint(minlen, maxlen)) +

            # Use only the global color table; no local tables are needed
            b"\x00\x00" +

            # No background color; use transparency instead
            b"\x0f")

    if repeats:
        # Generate unlimited repetitions; this makes it
        # easier to detect that there is any actually
        # useful data in the image afterwards.
        repeat_count = b"\xff"

        # Set the last two bits of the delay time to 0
        delaytime = b"\xf0"
    else:
        # Limit the repetition count to something reasonable
        repeat_count = rng.choice([b"\x01", b"\xfe"])

        # Set the last bit of the delay time to 1
        randdelay = rng.randint(0, 31)
        delaytime = bytes([(randdelay | 0x80)])

    # Select a random position index from which to begin reading
    start = rng.randint(0, max(maxlen-1, 1))
    posidx = rng.randint(start, maxlen+1)

    # Read until the given position index has been reached
    data = b""
    while True:
        # Read at most 256 bytes for each iteration
        chunk = rng.read(max(minlen, 256))

        if not chunk or len(chunk)+len(data) >= posidx:
            break
        data += chunk

    # The last chunk might be shorter than 256, so pad it properly
    if len(chunk) < 256:
        data += chunk + bytes(256 - len(chunk))

    # Determine an appropriate amount of padding to avoid
    # accidentally reading past our desired end point
    padlen = posidx % 256
    if padlen == 0:
        padlen = 256

    # Add the necessary amounts of padding before the image data
    data += bytes(padlen) + b"!"*repeat_count + delaytime

    # Determine whether we should switch the first three bits of
    # the LZW minimum code size field from one state to another.
    lzwbits = rng.choice([1, 2, 3, 4, 5, 6, 7, 8, 2<<3, 3<<3], p=[0.06, 0.06, 0.06, 0.06, 0.06, 0.06, 0.06, 0.06, 0.1, 0.04])

    # Insert a new field into the header
    head_end = data.find(b"!")
    orig_head = data[:head_end]
    new_head = (orig_head[:5] +
                bytes(((lzwbits >> 3) & 0x07) + 0x10, "latin1") +
                orig_head[-3:])
    data = new_head + data[head_end:]

    # Reinsert the original header without the LZW bits value
    out.write(gen_header())

    # Write the remaining data to the output stream
    out.write(data)