from __future__ import annotations
from enum import Enum


class GraphicsControlExtensionType(Enum):
    # Transparent color index used in this frame
    transparent_color_index = 0
    # Delay time for this frame (in ms)
    delay_time = 1
    # Transparency flag; whether to use transparency or not in this frame
    transparent_flag = 2
    # User input flag; whether the user can interact with this frame through input devices
    user_input_flag = 3


# Struct that represents the Graphics Control Extension block of an animated GIF
# See https://www.w3schools.com/app/img_displaygif.asp#gif-animation-control-block
struct_graphics_control_extension = struct.Struct('<B5xBB')


# Struct that represents the Image Descriptor block of an animated GIF
# See https://www.w3schools.com/app/img_displaygif.asp#image-descriptor
struct_image_descriptor = struct.Struct('<BBH2BHB')


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The output will always have at least two frames.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, max_num_frames: int = 256
    ):

    max_width = min(max_num_frames, 32768)
    max_height = min(int((8192**2) / (max_width * max_width)), 32768)
    
    num_frames = int.from_bytes(rng.read(1), byteorder='little', signed=False) % max_num_frames + 2
    out.write('GIF89a'.encode())

    # Write global header
    bytes_per_pixel = 1 # all pixel format are monochrome
    background_color_index = 0

    
    buffer = [None] * 256
    counter = 0
    while counter < num_frames:
        
        if counter == 0:
            left = 0
            top = 0
            width = 1
            height = 1
        else:
            
            # Width and Height must be between 0 and 65535
            width = int.from_bytes(
                rng.read(1), 
                byteorder='little', 
                signed=False
            ) % max_width + 1
            height = int.from_bytes(
                rng.read(1), 
                byteorder='little', 
                signed=False
            ) % max_height + 1

            left = int.from_bytes(
                rng.read(1), 
                byteorder='little', 
                signed=False
            ) % ((max_width - width))
            top = int.from_bytes(
                rng.read(1), 
                byteorder='little', 
                signed=False
            ) % ((max_height - height))
            
        is_local_palette = bool(int.from_bytes(rng.read(1), byteorder='little', signed=False) % 2)

        # List of colors; each item has three components R, G, B
        local_palette = []
        for _ in range(2 ** bytes_per_pixel):
            local_palette.append([
                rng.read(1)[0], 
                rng.read(1)[0], 
                rng.read(1)[0]
            ])
        
        lzwminsize = len(data) + 1
        subblocks = []
        num_subblocks = int.from_bytes(
            rng.read(1), 
            byteorder='little', 
            signed=False
        ) % 128 + 1
        for i in range(num_subblocks):
            chunklen = int.from_bytes(
                rng.read(1), 
                byteorder='little', 
                signed=False
            ) % 256
            chunk = list(rng.read(chunklen))[::-1]
            print(f'{i} : {chunk}')
            subblocks.extend([(len(chunk)).to_bytes(1, 'big')])
            subblocks.extend(chunk[::-1])

        
        frame_end = False
        # Write image descriptor
        flags = 0x20 | 0x1c
        if counter > 0:
            flags |= 0x80
        elif isinstance(transparent_color_index, int):
            flags |= 0xf0

        out.write(struct_image_descriptor.pack(
            num_frames,
            flags,
            left, top,
            width,height,
            bytes_per_pixel,
            background_color_index,
            0
        ))

        # Write graphics control extension
        if counter > 0:
            out.write(struct_graphics_control_extension.pack(
                4, 
                0, 
                0,
                0
            ))

        # Write local palette
        if is_local_palette:
            out.write((2 ** bytes_per_pixel).to_bytes(1, byteorder='little'))
            for rgb in local_palette:
                out.write(b'\x00\x00\x00')
                for c in rgb:
                    out.write(c.to_bytes(1, byteorder='little'))
        
        # Write image data
        out.write(lzwminsize.to_bytes(1, byteorder='little'))

        out.write(len(subblocks).to_bytes(1, byteorder='little'))
        out.write(b"".join(subblocks))
        out.write(b"\x00")

        # Check if we need to end the animation
        if counter >= num_frames - 1:
            frame_end = True

        counter += 1