#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):


    def gen_byte():
        return rng.read(1)[0]

    def gen_short():
        short = 0x100 * gen_byte()
        short += gen_byte()
        return short

    def gen_long():
        long = 0x1000000 * gen_byte()
        long += 0x10000 * gen_byte()
        long += 0x100 * gen_byte()
        long += gen_byte()
        return long

    framecount = gen_byte() & 0x7f
    print('frame count:', framecount)
    out.write(b'GIF89a')
    out.write((2).to_bytes(1, 'big'))
    out.write(gen_short().to_bytes(2, 'big'))
    out.write((0).to_bytes(2, 'big'))

    for i in range(framecount):

        print('frame', hex(i))
        
        localcolortablesize = (
                gen_byte() & 0x07 | ((gen_byte() & 0x01) << 8)) + 1
        print('local colortable size:', localcolortablesize)        
        out.write(localcolortablesize.to_bytes(1, 'big'))

        globalcolortableflag = False;
        if gen_byte() % 2 == 1:
            globalcolortableflag = True
            print('global color table flag set')
        else:
            print('global color table flag not set')
            
        width = gen_short()
        height = gen_short()
        print('width:', width)
        print('height:', height)
        out.write(width.to_bytes(2, 'big'))
        out.write(height.to_bytes(2, 'big'))
        
        packedfield = gen_byte()
        interlacedbit = bool((packedfield >> 7) & 0x01)
        print('interlaced bit:', interlacedbit)
        sortbit = bool((packedfield >> 6) & 0x01)
        print('sort bit:', sortbit)
        reservedbits = (packedfield >> 3) & 0x07
        print('reserved bits:', reservedbits)
        haspaletteinfo = bool(packedfield & 0x01)
        print('has palette info?', haspaletteinfo)
        out.write(((0x80 if interlacedbit is True else 0x00)
                   | (0x40 if sortbit is True else 0x00)
                   | (reservedbits & 0x07)).to_bytes(1, 'big'))
    
        lctsize = None 
        if haspaletteinfo:

            lctsize = (
                    gen_byte() & 0x07 | ((gen_byte() & 0x01) << 8)) + 1
            
            print('local color table size:', lctsize)
            out.write(lctsize.to_bytes(1, 'big'))
        
            if lctsize > 0 and lctsize <= 2**2:

                lctsortedflag = False
                if gen_byte() % 2 == 1:
                    lctsortedflag = True
                    print('local color table sorted flag set')
                else:
                    print('local color table sorted flag not set')
    
                lctintencityflag = [
                        (gen_byte() & 0x01)!= 0
                        for _ in range(3)]
                print('local color table intensity flags:')
                for j in range(len(lctintencityflag)):
                    print('\tlctintencityflag[{}]={}'.format(
                            j, str(lctintencityflag)))
                    
                gctsortedflag = False
                if gen_byte() % 2 == 1:
                    gctsortedflag = True
                    print('global color table sorted flag set')
                else:
                    print('global color table sorted flag not set')
                
                gctintencityflag = [
                        (gen_byte() & 0x01)!= 0
                        for _ in range(3)]
                print('global color table intensity flags:')
                for j in range(len(gctintencityflag)):
                    print('\tgctintencityflag[{}]={}'.format(
                            j, str(gctintencityflag)))

                out.write(struct.pack('<B', lctsortedflag))
                lctintencityflagbits = sum([
                        1*int(j) for j in lctintencityflag])
                out.write(
                        lctintencityflagbits.to_bytes(1, 'big'))
                out.write(struct.pack('<B', gctsortedflag))
                gctintencityflagbits = sum([
                        1*int(j) for j in gctintencityflag])
                out.write(
                        gctintencityflagbits.to_bytes(1, 'big'))
                    
        elif localcolortablesize > 1: # hack to get around broken GIFs
            raise Exception("No palette but LCTS>0")

        localcolormaskingflag = False
        if gen_byte() % 2 == 1:
            localcolormaskingflag = True
            print('local colormasking flag set')
        else:
            print('local colormasking flag not set')
            
        locallayerblockingflag = False
        if gen_byte() % 2 == 1:
            locallayerblockingflag = True
            print('local layer blocking flag set')
        else:
            print('local layer blocking flag not set')
                
        disposalmethod = gen_byte() & 0x07
        print('disposal method:', disposalmethod)
        userinputflag = bool(gen_byte())
        print('user input flag:', userinputflag)
        transparencylevel = gen_byte() & 0xf
        print('transparent level:', transparencylevel)
        out.write(((disposalmethod << 2)
                   | (0x01 if userinputflag is True else 0x00)
                   | (transparencylevel & 0x0f)).to_bytes(1, 'big'))

        delaytime = gen_short()
        transparentcolorindex = gen_byte()
        extralegth = gen_byte()
        print('delay time:', delaytime)
        print('transparent color index:', transparentcolorindex)
        print('extra length:', extralegth)
        out.write(delaytime.to_bytes(2, 'big'))
        out.write(transparentcolorindex.to_bytes(1, 'big'))
        out.write(extralegth.to_bytes(1, 'big'))

        if extralegth > 0:
            extradata = []
            while len(extradata) < extralegth:
                b = gen_byte()
                extradata.append(b)
                out.write(b)
            print('extra data:', bytes(extradata))
            assert len(extradata)==extralegth
            
        if localcolormaskingflag is True:
            print('local colormasking flag set')
            localcolormaskdata = []
            localcolorreferencebits = sum([
                    (gen_byte()&0xff)<<(k*8) 
                    for k in reversed(range(3))]);
            print('local color reference bits:', localcolorreferencebits)
            localcolormaskdata.extend([(localcolorreferencebits>>i)&0xff
                                       for i in range(0, 24, 8)]);
            localcolormaskdata.reverse();
            print('local colormaskdata:', bytes(localcolormaskdata));
            1/0
        else:
            print('local colormasking flag not set');            
        
        if localcolortablesize > 1 or (not globalcolortableflag is True
                                      and lctsize is not None):
            print('local colortable present')
            localcolortableentries = [[None]*3
                                     for _ in range(2**localcolortablesize)];
            if localcolortablesize >= 0:
                for i in range(2**localcolortablesize):                
                    rgb = [(gen_byte() & 0xff)
                           for _ in range(3)];                    
                    localcolortableentries[i] = rgb
                    out.write(rgb);
            else:
                1/0
                
            print('local colortable entries:')
            for i in range(2**localcolortablesize):
                print('\tlocalcolortableentry[{}]: {}'.format(
                        i, bytes(localcolortableentries[i])))
                
        if lctsize is not None:
            pass

        if globalcolortableflag is True:
            print('global colortable present')
            globalcolortableentries = [[None]*3
                                        for _ in range(2**globalcolortablesize)];
            if globalcolortablesize >= 0:
                for i in range(2**globalcolortablesize):                
                        rgb = [(gen_byte() & 0xff)
                               