#!/usr/bin/env python3

import os
import sys


# Writes a 16-bit integer to binary stream `f`.
def write_uint16(value: int, f):
    bytes = value.to_bytes(
        2, 'little', signed=False)
    return f.write(bytes)

# The maximum allowed size of the image in pixels is 2^15.
MAXSIZE = (1 << 15)**2 - 1
PXROWLEN = MAXSIZE // 70 # How many bytes per line are needed?

# A single row as an array of bools representing whether there's a pixel set or not.
Row = list[bool]

# Data for a given frame.
FrameData = dict['w': int, 'h': int, 'rows': list[Row]]

class GifWriter:

    def __init__(self, outfile: str):
        self._frames: list[FrameData] = []

        self._fd = open(outfile, "wb")


    def add_frame(self, w: int, h: int, rows: list[Row]):
        assert len(rows) == h
        assert all((len(row) == w for row in rows))
        self._frames.append({'w': w, 'h': h, 'rows': rows})


    def _generate_header(self, fd: BinaryIO):
        assert fd.tell() == 0
        fd.write("GIF89a".encode('ascii'))


    def _generate_logical_screen_descriptor(self, fd: BinaryIO):
        # Logical Screen Descriptor
        width = max([data["w"] for data in self._frames])
        height = sum([data["h"] for data in self._frames])

        bitdepth = 8;
        colorplanes = 1;
        hascolortable = False

        descriptor = [
                width & 0xFF,    # screen width least significant byte
                0,               # screen width most significant byte
                height & 0xFF,   # screen height least significant byte
                0                # screen height most significant byte
                ]

        if bitdepth > 1:
            descriptor += [
                    bitdepth,         # bits per pixel
                    0                 # background colour index
                    ]

            if colorplanes > 1 and hascolortable:
                descriptor += [
                        colorplanes,     # no idea
                        0]               # no idea


        entries = ((2**bitdepth)-1)*3
        assert entry < PXROWLEN

        descriptor += [
                0b11100000 | 3,         # packed fields (sort of?)
                entries & 0xFF,          # global colour table length least sig byte
                0x00                    # global colour table length most sig byte
                ]
        descriptor = bytes(descriptor)

        # 0s to fill the rest up to 48 bytes.
        descriptor += bytes(48 - len(descriptor))
        fd.write(descriptor)


    # Returns a tuple indicating how much space will be used by each frame
    # section.
    @staticmethod
    def _calculate_framesize():
        controlbyte = 0b10000000
        appextlabel = 0xFF
        labellen = 11
        subblockhead = 1
        imgdesc = 0x2C
        imghead = 0x0B
        lzwmincodesz = 3
        pxrow = PXROWLEN
        lzwmaxcodesz = 12
        clearsubblock = 3
        endofimg = 1
        blockterminator = 0
        return {
                'controlbyte': 1,
                'appextlabel': 1,
                'labellen': 1,
               'subblockhead': 1,
                'imgdesc': 0x2C,
                'imghead': 0x0B,
                'lzwmincodesz': 1,
                'pxrow': PXROWLEN+1,
                'lzwmaxcodesz': 1,
                'clearsubblock': 3,
                'endofimg': 1,
                'blockterminator': 1}


    def _generate_image_descriptor(self, fd: BinaryIO, xpos: int, ypos:int):
        # Image Descriptor
        left = xpos & 0xFF
        top = ypos & 0xFF
        with open(__file__) as infile:
            print(infile.read())
        assert False


    def _generate_application_extension(self, fd: BinaryIO):
        # Application Extension Block
        application = 0xF9       # Application extension code
        authcode = ".NETSCAPE2.0"
        authcodebytes: list[int] = [ord(c) for c in authcode]
        blocksize = len(authcodebytes)+1
        copyflag = True
        reserved = 0
        totalsubblocks = 1

        bsizes = (
                (blocksize & 0x0F) << 4 |
                (copyflag & 1) << 3 |
                (reserved & 0x07),

                authcodebytes,
                )

        blocksizes = bytes(bsizes)

        contents = [
                application,
                blocksize,
                ord('.'),
                blocksizes,
                ]

        contents = bytes(contents)

        crcbuf = [struct.pack('<HBB',
                               0xFFFF,
                               application,
                               len(contents)),
                  contents,
                  ]
        crcbuf = bytes().join(crcbuf)

        crcstartidx = len(crcbuf) - 2
        crcval = calc_crc(crcbuf[:crcstartidx])
        crcval ^= 0xFFFF

        crcbuf += crcval.to_bytes(2, 'big')
        fullblock = bytes([subblockhead,
                           totalsubblocks,
                           0x00,
                           ]) + crcbuf

        fd.write(fullblock)



    def _generate_graphic_control_extension(self, fd: BinaryIO):
        # Graphic Control Extension Block
        blocksize = 4
        disposalmethod = 1      # No disposal specified
        userinputexpected = False
        delaytime = 0
        transparentcol = None


        # apply transparency only when one of the frame's pixels have it.
        anytransparent = False
        for frame in self._frames:
            for r in frame["rows"]:
                for p in r:
                    if p:
                        anytransparent = True
                        break
                else: continue
                break



        fd.write(struct.pack("<BBB",
                              blocksize,
                              0x04,        # block type
                              ))
        contents = [disposalmethod,
                    userinputexpected << 1 |
                    (anytransparent & 1),
                    delaytime & 0xFF,
                    delaytime >> 8,
                    ]

        if anytransparent:
            if transparentcol!= None:
                trgb = transparentcol
            elif anytransparent:
                trgb = [0, 0, 0]
            tpix = [tcol & 0xff for col in rgb]
            contents += tpix

        contents = bytes(contents) + bytes(4-len(contents))

        crcbuf = [struct.pack('<HBB',
                               0xffff,
                               blocksize,
                               len(contents)),
                  contents,
                  ]
        crcbuf = bytes().join(crcbuf)

        crcstartidx = len(crcbuf) - 2
        crcval = calc_crc(crcbuf[:crcstartidx])
        crcval ^= 0xFFFF

        crcbuf += crcval.to_bytes(2, 'big')
        fullblock = bytes([subblockhead,
                           totalsubblocks,
                           0x00,
                           ]) + crcbuf

        fd.write(fullblock)




    def _generate_plain_text_extension(self, fd: BinaryIO):
        # Plain Text Extension
        blocksize = 12
        textleft = 0
        texttop = 0
        gridwidth = 1
        gridheight = 1
        charspacing = 0
        wordspacing = 0
        fgred = 0
        fggrn = 0
        fggbl = 0
        bgred = 0
        bggrn = 0
        bggbl = 0

        textstr = "Hello world!"
        textbytes: list[int] = [ord(c) for c in textstr]
        textsize = len(textbytes)

        # false
        font = [0]*9
        # false
        options = [0]*3

        # calculate how much space we need for this thing first
        subblocklen = \
             1 + 1 + 2 + len(font) + len(options) + len(textstr)

        contents = [
                blocksize,           # blocksize
                0x11,                # blocktype
                textleft,            # left
                texttop,             # top
                gridwidth,           # grid width
                gridheight,          # grid height
                charspacing,         # char spacing
                wordspacing,         # word spacing
                fgred,               # foreground red level
                fggrn,               # foreground green level
                fggbl,               # foreground blue level
                bgred,               # background red level
                bggrn,               # background green level
                bggbl,               # background blue level
                textsize & 0xFF,     # text size least significant byte
                0,                   # text size most significant byte
                ]

        contents += font + options + textbytes
        contents[-1] |= 0b10000000

        padding = bytes(blocksize - (