#!/usr/bin/env python3

import struct
from typing import BinaryIO


class GifHeader(object):

    """
    Represents a GIF header for an image of size (width x height).
    See https://www.w3.org/Graphics/GIF/spec-gif89a.txt for details on how to
    interpret these values.
    """

    def __init__(self, width: int, height: int) -> None:

        self._width = width
        self._height = height
        self._colormap_size = -1 # let's assume there is no colormap
        self._global_table_flag = False
        self._bgcolor_index = 0
        self._pixel_aspect_ratio = 0


    @property
    def width(self) -> int:
        return self._width

    @property
    def height(self) -> int:
        return self._height

    @property
    def color_table_size(self) -> int:
        if self._colormap_size == -1:
            raise Exception('Color table not set')
        else:
            return 2 << (self._colormap_size+1)

    @property
    def global_colormap_flag(self) -> bool:
        return self._global_table_flag

    @property
    def bg_color_index(self) -> int:
        return self._bgcolor_index

    @property
    def pixel_aspect_ratio(self) -> float:
        if self._pixel_aspect_ratio!= 0 and \
           self._pixel_aspect_ratio % 10!= 0:
            raise ValueError('Pixel aspect ratio must be multiple of 10')
        
        return self._pixel_aspect_ratio / 10



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        header: GifHeader,
        out: BinaryIO
    ):
    width, height = header.width, header.height

    # Write header fields from specified
    header_fields = [
        width, height,
        header.color_table_size,
        header.global_colormap_flag,
        header.bg_color_index,
        header.pixel_aspect_ratio
    ]

    header += header_fields
    
    for field in header_fields:
        field_data = struct.pack('<B', field)
        out.write(field_data)