#!/usr/bin/env python3

import os, sys, logging
import struct
from pathlib import Path
from typing import List, Tuple

class RandomGifGenerator():
    def __init__(self, size: Tuple[int]):
        self._rng = open('/dev/urandom', 'rb')

        # Make sure we stitch together an integer amount of pixels
        assert size[0] % 2 == 0 and size[1] % 2 == 0
        
        self._width, self._height = list(map(lambda x: x // 2, size))
    
    @property
    def width(self) -> int: return self._width
    @property
    def height(self) -> int: return self._height

    def generate(self, path: str):
        '''Generate a random gif at <path>, which must not already exist.'''
        assert not os.path.exists(path), f'{path} exists!'

        with open(path, 'wb') as fd: 
            self.__generate(fd)
    
    def __generate(self, out: BinaryIO):
        # Write header
        out.write(b'GIF89a')

        # Generate a random color palette from the RNG
        palbuf = []
        for _ in range(2**10):
            r, g, b = tuple(
                map(
                    lambda x: max(min(round((i / 255)**0.4) * 8, 7), 0), 
                    self._rng.read(3)))

            palbuf += [r << 4 | g, b]

        out.write(len(palbuf).to_bytes(2, "little"))
        for pixel in zip(*[iter(palbuf)] * 2):
            out.write(pixel[0].to_bytes(
                1, byteorder="little", signed=False))
            out.write(pixel[1].to_bytes(
                1, byteorder="little", signed=False))

        # Write each line of pixels
        for y in range(self._height):
            buf = [struct.unpack('>H', self._rng.read(2))[0]]
            for _ in range(self._width - 1):
                data = self._rng.read(1)[0]

                # Set high bit on every second byte because that's how gifs work
                if (_ + 1) % 2 == 0:
                    data |= 0x80
                
                buf += [data]
            
            buf += [None]
            out.write(bytes([*filter(None, buf)]))
