#!/usr/bin/env python3

import struct
from typing import BinaryIO


PALETTE_COLORS = [0, 1, 2]

# Generates a random GIF file into `out` using the
# random number generator `rng`, which should be /dev/urandom.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:


    # Format is as follows (all numbers LSB):
    # - Header ("GIF87a" or "GIF89a")
    # - Logical screen descriptor
    #   - Width in pixels, LSB first
    #   - Height in pixels, LSB first
    #   - Flags
    #       * Size of global color palette (1 << n colors).  Anything else would
    #           signal to use the local color table instead.
    #       * Sort flag
    #       * Color resolution (1 == 7 bits per pixel)
    #       * Global color table flag (set this if we have a global palette)
    #   - Background color index
    #   - Pixel aspect ratio (always set to zero for now)
    #
    # Then come either an image descriptor chunk or a graphics control extension
    # with a single frame delay and transparent color set to our background
    # color.
    #
    # Finally comes the image data itself.  The actual length here can be found by
    # subtracting however much space it took up on disk from the total size that was
    # specified at the start of the file.
    header = b'GIF89b'

    # LSB-first width, height, flags
    sdesc = struct.pack('<HHB',
            500, 500,
            0x80 | len(PALETTE_COLORS) # Global color table, sort
            | 0x40                   # Resolution bit
            | 0x08                   # Global color table present
            | len(PALETTE_COLORS))    # How many colors do we have?

    # Then 1 byte each for background index, pixel aspect ratio
    sdesc += b'\x0f\x00'

    # Now we finally put together all the parts...
    full_spec = header + sdesc

    # Our color table
    ct = bytes([
        0xff, 0xff, 0xff, # White
        0xf0, 0xf0, 0xf0, # Light gray
        0xe0, 0xe0, 0xe0]) # Gray

    # Data chunk to follow later on
    data_chunk = b''

    # And raw image data!  Let's just generate some random lines.
    num_lines = 500 // 8 
    for lineid in range(num_lines):
        data_chunk += struct.pack('BBB',
                PALETTE_COLORS[lineid % len(PALETTE_COLORS)],
                0,
                *[PALETTE_COLORS[(lineid+colid)%len(PALETTE_COLORS)]
                    for colid in range(8)]) 

    # Ok, let's pad out until we are aligned to a 4-byte boundary...
    while len(data_chunk) % 4!= 0:
        data_chunk += '\x00'
    
    # OK, what's our final size going to be when written to disk?
    total_size = len(full_spec) \
            + len(ct) \
            + 4      \
            + len(data_chunk)
    print('Total size:', total_size, 'bytes')

    # Now write everything out to our output stream
    out.write(struct.pack('<L', total_size))     # Write size
    out.write(full_spec)                          # Write spec
    out.write(ct)                                 # Write color table
    out.write(b'\x00\x00')                        # Frame delay
    out.write(struct.pack('BB', 
            0xfe,        # Extension type
            len(data_chunk)))                  # Extension body size
    out.write(data_chunk)                         # Image data