#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    width  = get_random_int(256, 4096)
    height = get_random_int(256, 4096)

    blocktypes = [
        33, # Img Descriptor
        44, # Global Color Table
        255 # Application Extension
    ]

    # Create img descriptor
    descriptor = bytes([width % 256]) \
                + bytes([(height >> 8) & 0xff]) \
                + bytes([height & 0xff]) \
                + bytes([get_random_int(0, 7),           # bg color index
                         bool(get_random_bool()),      # local colors?
                         0x00,                          # sort flag...
                         0xf9 | (get_random_bool() << 1)]) # interlace flag
                                                                  # and size of lct

    global_color_table_size = get_random_int(1, 9)
    global_color_table = None
    if not global_color_table_size == 0:
        global_color_table = bytes([global_color_table_size * 3]) \
                            + rng.read((global_color_table_size - 1) * 3)
    
    # Write all blocks to output 
    while True:
        blocktype = get_random_element(blocktypes)
        
        # Image Descriptor Block 
        if blocktype is 33:
            out.write(descriptor[0:9])
            continue

        # Global Color Table Block
        elif blocktype is 44: 
            if global_color_table is None:
                print("Global Color table should be present")

            out.write(global_color_table)
            continue

        # Application Extension Block
        else:
            application_identifier = bytes([get_random_byte(),
                                            get_random_byte()])

            application_authcode    = bytes([get_random_byte(),
                                            get_random_byte(),
                                            get_random_byte()])
            
            # This is actually useless because it's up to the app to read
            # that many bytes from stdin before loading the image
            num_application_data_blocks   = get_random_int(1, 5)
            application_data = []
            for i in range(num_application_data_blocks): 
                application_data += bytes([get_random_byte()])
                
            application_data = bytes(application_data)

            applength = len(application_data).to_bytes(2, byteorder='big', signed=False)
            out.write(applength +
                      application_identifier + 
                      application_authcode + 
                      application_data)
            break
            


def get_random_element(population):
    return population[get_random_int(0, len(population)-1)]
    
def get_random_int(lowerbound, upperbound):
    return lowerbound + ((upperbound-lowerbound+1)*struct.unpack('>I', rng.read(4))[0] // (2**32))

def get_random_bool():
    return get_random_int(0,1) > 0 

def get_random_byte():
    return get_random_int(0,255) 
    