#!/usr/bin/env python3

import sys
import io
import binascii
import argparse


# Generate a random binary object of length n
def randbytes(n):
    return bytes([ord('x') for _ in range(n)])


# Generate an array of n pointers to data chunks
def ptrs(data: list[bytes], n: int) -> bytearray:

    # Write the pointer table as initial zeros
    p = [0] * (n + 1)

    # Append the data and build up the pointer table
    p += sum((len(x), i) for i, x in enumerate(data))

    assert len(p) == n+1
    
    return bytearray().join([struct.pack('<I', v) for v in p])


# Construct a GIF header from its pieces
def header(width=512, height=512, frames=-1, color_table=[0]):

    # Color resolution is 7 bits per pixel indexed with no transparency
    cr = 7 << 5

    # Global color map present; global color count specified by color_table
    gcp = 0x80 | len(color_table)

    flags = ord('\x87')        # No local color map, interlace, etc.

    hdr = bytearray()          # Header starts empty

    # Size block
    hdr.extend([
        6,              # Block size in bytes
        width & 0xff,   # Width lo-byte
        width >> 8,     # Width hi-byte
        height & 0xff,  # Height lo-byte
        height >> 8,    # Height hi-byte
        cr|gcp,         # Flags
    ])

    if frames > -1:

        # Loop control extension, endlessly repeat
        hdr.extend([
            0x21,   # Extension introducer
            ord('L'), ord('O'), ord('P'),             # Label "LOP" == loop control
            8,      # Block size in bytes
            0b00000001,       # Loops forever
            frame & 0xff,     # Frame delay lo-byte
            frame >> 8,       # Frame delay hi-byte
        ])

    hdr.append(0)            # Trailer
    hdr.extend(color_table)  # Color table
    hdr.append(flags)        # Image descriptor
    hdr.extend([0, 0, 0])    # X,Y position

    return hdr


# Main function
if __name__ == "__main__":

    parser = argparse.ArgumentParser() 
    parser.add_argument('-w', '--width', help='Image width', default=512, type=int)
    parser.add_argument('-H', '--height', help='Image height', default=512, type=int)
    args = parser.parse_args()

    # Create some fake image data
    imgdata = [randbytes(i>>1 & 0xfffff) for i in range(1<<24)]

    # Now we have all of our raw data we can write it out
    buf = io.BytesIO()

    buf.write(header())                      # Header
    buf.write(ptrs(imgdata, 1))               # Pointer table
    buf.write(imgdata[0])                    # First chunk of image data
    buf.seek(0)

    # Read back what was written
    print(" ".join(["{:02x}".format(v) for v in buf.read()]).upper(),
          file=sys.stderr)
