from typing import BinaryIO


# Generates a random GIF file into `out` and returns its size in bytes
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> int:

    # Generate the header
    w = h = 2 ** (int.from_bytes(rng.read(1), 'little') & 7)
    flags = {
        "has_color_table": bool(
            int.from_bytes(rng.read(1), 'little') & 16
        ),
        "is_interlaced": bool(
            int.from_bytes(rng.read(1), 'little') & 64
        )
    }
    has_gct = flags["has_color_table"]
    is_il   = flags["is_interlaced"]
    
    if not has_gct:
        gct_size = 0
    elif has_gct:
        gct_size = 2**((int.from_bytes(rng.read(1), 'little'))&7) + 1

    header = [b"GIF",
              # Logical screen descriptor
              0x2c,                    # separator: image descriptor
              ord('!'),                # image left position
              ord(','),                # image top position
              b'\0\0',                 # no width
              b'\0\0',                 # no height
              chr(0xf0 | ((8-1)&7)),  # pixel depth
              0,                       # color table bits
              0,                       # sort flag
              # Global color table
              *[0xff]*gct_size]        # opaque white


    # Write the header to the output stream
    for b in header[0]:
        out.write(struct.pack("B", b))
    for i, v in enumerate(header[-1]):
        out.write(v.to_bytes(1, 'little'))

    # Write all other fields one by one
    out.write(w.to_bytes(2, 'little'))
    out.write(h.to_bytes(2, 'little'))

    # Compute the total amount of pixels
    n = w*h

     # LZW minimum code size
    min_code_size = 2 ** ((int.from_bytes(rng.read(1), 'little')))

    # LZW maximum code size
    max_code_size = 2 ** ((int.from_bytes(rng.read(1), 'little')))

    # The actual LZW minimum code size must be within the range
    assert 2 <= min_code_size <= 12, f"{min_code_size} is invalid!"
    assert min_code_size < max_code_size, f"{min_code_size} > {max_code_size}"

    # Initialise the LZW code size with the provided value or determine it
    # from the pixel count otherwise. We can't use a fixed lzw code size as we
    # want to have at least three different colors used for each block.
    if max_code_size - min_code_size >= n / 3:
        lzw_code_size = min_code_size
    else:
        while True:
            lzw_code_size += 1
            n_blocks = n // 2**(lzw_code_size+1)
            tmp_n     = sum([
                2**i for i in
                range(lzw_code_size+1)])

            if tmp_n == n:
                break

        print(f"Allocated {lzw_code_size} octets per color")


    # Image descriptor
    # Left offset
    out.write(ord('!').to_bytes(1, 'little'))
    # Top offset
    out.write(ord(',').to_bytes(1, 'little'))
    # Width
    out.write(w   .to_bytes(2, 'little'))
    # Height
    out.write(h   .to_bytes(2, 'little'))
    # Flags
    out.write(((0xf7 & ~((not has_gct)<<5)) | (((8-1)&7)<<5)).to_bytes(1,'little'))
    # Sort flag
    out.write(0      .to_bytes(1, 'little'))
    # Local color table
    if has_gct:
        for j in range(0, 3*(2**((8-1)&7)), 3):
            out.write(*[(j >> i) & 2**7 - 1
                      for i in reversed(range(0, 24, 8))]
                     )

    # Minimum code size
    out.write(lzw_code_size      .to_bytes(1, 'little'))
    # Image data subblock
    out.write(b'\xfe')
    # Block size
    out.write((lzw_code_size + 1).to_bytes(1, 'little'))

    # Populate the whole LZW buffer with values
    keys = list()
    val_set = set()
    val = None
    for i in range(2**lzw_code_size):
        key = rng.read(1)[0]
        if len(keys) == 2**lzw_code_size:
            keys.clear()
        keys.append(key)
        val = rng.read(1)[0]
        while val in val_set:
            val = rng.read(1)[0]
        val_set.add(val)

    assert False, ""

    return