#!/usr/bin/env python3

import sys
import os
import io
import random
import struct


DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    @staticmethod
    def randint() -> int:
        '''
        Returns a psuedorandom integer between
        0x000000 and 0xffffff
        '''
        return random.randint(0, 255**3 - 1)

    
    @staticmethod
    def color() -> list[int]:
        '''
        Returns a psuedorandom RGB color code as a three element list of ints
        '''
        return [RandomGifGenerator.randint(), 
                RandomGifGenerator.randint(),
                RandomGifGenerator.randint()]


    @staticmethod
    def bits(*args: tuple[int]):
        '''
        Converts integers into their binary representation in big endian order.
        Each argument can be either decimal or hexadecimal format.
        '''
        result = []

        for arg in args:
            if type(arg) == str:
                dec = int(arg, base=16)
            elif type(arg) == int:
                dec = arg
            
            binstr = ''
            while True:
                binstr = f'{dec % 2}{binstr}'
                dec //= 2

                if dec < 1: break
                
            result.append(f'{binstr:{8}}')
            
        return ''.join(reversed(result)).encode('ascii')


    @staticmethod
    def generate_colorspace() -> bytearray:
        ''' 
        Generates a color space consisting of 768 colors and returns its raw form.
        '''

        cs = bytearray([])
        for i in range(768):
            cs.extend(struct.pack("BBB", *RandomGifGenerator.color()))

        return cs

    
    def generate_random_gif(self):
        
        with open(self.output, "wb") as out:

            return generate_random_gif(out)


# Generates a random GIF file into `out`.
# The parameters correspond to the information found here:
# https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(
        out: BinaryIO
    ):

    # Signature
    out.write(b"GIF89a")

    # Width and Height
    w, h = RandomGifGenerator.randint(), RandomGifGenerator.randint()
    out.write(RandomGifGenerator.bits(w,h))

    # Color Table size
    out.write(RandomGifGenerator.bits())

    # LZW Minimum Code Size
    lzw_min_code_size = RandomGifGenerator.randint() & 0b1111
    out.write(RandomGifGenerator.bits(lzw_min_code_size))

    # Graphics Control Extension
    gce = RandomGifGenerator.bits(
        RandomGifGenerator.randint(),      # disposal method
        RandomGifGenerator.randint(),      # user input flag
        RandomGifGenerator.randint(),      # delay time
        RandomGifGenerator.randint(),      # transparent color index
        RandomGifGenerator.randint()       # block terminator
    )
    out.write(gce)

    # Plain Text Extension
    pte = RandomGifGenerator.bits(
        len(pte := chr(RandomGifGenerator.randint())),             # text length
        *(pte:=RandomGifGenerator.bits(*(pte:=chr(c))) for c in pte),  # text characters
        RandomGifGenerator.randint()                                 # block terminator 
    )
    out.write(pte)

    # Application Extension
    ae = RandomGifGenerator.bits(
        len(ae := chr(RandomGifGenerator.randint())),                 # application ID
        *(ae:=RandomGifGenerator.bits(*(ae:=chr(c))) for c in pte),      # application data
        RandomGifGenerator.randint()                                     # block terminator 
    )
    out.write(ae)

    # Comment Extension
    ce = RandomGifGenerator.bits(
        len(ce := chr(RandomGifGenerator.randint())),                 # comment length
        *(ce:=RandomGifGenerator.bits(*(ce:=chr(c))) for c in pte),      # comment characters
        RandomGifGenerator.randint()                                     # block terminator 
    )
    out.write(ce)

    # Image Descriptor (starts new frame)
    id = RandomGifGenerator.bits(
        w,                                                      # left position
        h,                                                      # top position
        len({RandomGifGenerator.randint()} | set()),               # width
        len({RandomGifGenerator.randint()} | set()),               # height
        RandomGifGenerator.randint() & 0b111,                     # local color table flag
        RandomGifGenerator.randint() & 0b1111111,                # interlace flag
        0,                                                      # sort flag
        0,                                                      # reserved (packed fields)
        len({RandomGifGenerator.randint()} | set()),               # global color table flag
        0                                                       # block terminator
    )
    out.write(id)

    # Color Table
    ct = RandomGifGenerator.generate_colorspace()
    out.write(ct)

    # Image Data
    imgdata = io.BytesIO()
    gs = io.BytesIO(gs := '')
    print(f'Color Space Length: {len(ct)}')
    while True:
        nxt = next((i for i in ct), None)
        idx = random.choice(range(len(ct))) if nxt is None \
              else ct[:idx].index(nxt)
        
        gs += struct.pack("<B", idx)
        imgdata.write(struct.pack("B", idx))

        if len(gs) >= 255: 
            out.write(bytes([255]))
            out.write(gs[:-1])
            gs = io.BytesIO(gs[-1:])

    out.write(bytes([gs.getvalue()[0]]))
    out.write(imgdata.getbuffer())