#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given pseudorandom number generator in binary format
def generate_random_gif(rng: BinaryIO, out: BinaryIO):

    # Writes an unsigned 1-byte to the output (0x00... 0xFF)
    def write_uchar():
        byte = rng.read(1)[0] % 256
        out.write(struct.pack('B', byte))

    # Writes an unsigned 2-byte integer to the output
    def write_uint16():
        uint16 = rng.read(2)
        out.write(struct.pack('>H', uint16[0]*256+uint16[1]))

    # Writes a color table entry and returns its index
    def write_colorentry():

        trns = bool(rng.read(1)[0])    # Transparency flag
        if not trns: write_uchar()     # Red component   (0x00.. 0xFF)
        else:        write_uchar()     # Reserved field  (0x00)
        if not trns: write_uchar()     # Green component (0x00.. 0xFF)
        else:        write_uchar()     # Reserved field  (0x00)
        if not trns: write_uchar()     # Blue component  (0x00.. 0xFF)
        else:        write_uchar()     # Reserved field  (0x00)
        return len(colors)              # Return index of this color

    width  = rng.read(2)               # Width  of image in pixels
    height = rng.read(2)               # Height of image in pixels
    colors = []                       # Entries of the global color table

    # Write screen descriptor block
    write_uchar()                     # Screen width  in pixel
    write_uchar()                     # Screen height in pixel
    out.write(b'\x87\x00')             # Flags: Global Color Table Flag set
                                       #       Sort Flag not set
                                       #       Size of Color Table logarithmically scaled (ie. 2^n bytes long)
                                       #       Background color index is unused
                                       #       Pixel aspect ratio is unspecified (ie. pixel square)
    num_global_color_entries = rng.read(1)[0] & 0xF  # Number of entries in color table that should be used
                                                   # Rest are reserved for future use by extensions
                                                   
    # Generate global color table
    for i in range(num_global_color_entries):
        colors += [write_colorentry()]
        
    # Write local color table
    num_local_color_entries = rng.read(1)[0] & 0xF  # Number of entries in color table that should be used
                                                      # Rest are ignored
    for i in range(num_local_color_entries):
        write_colorentry()
    
    # Write image descriptor block
    write_uchar()                              # Left position
    write_uchar()                              # Top position 
    write_uchar()                              # Image width
    write_uchar()                              # Image height
    out.write(b'\x87')                          # Local color table flag
    out.write(struct.pack('B', len(colors)))    # Index of current color palette table from which to draw pixels
    out.write(b'\x00')                          # LZW minimum code size
    
    # Prepare compression stream
    lzwminsize = rng.read(1)[0]                # Minimum code size for encoded strings (must be between 2 and 12)
    dictsize = 2**lzwminsize                   # Dictionary size (number of possible characters)
    clearcode = 1 << lzwminsize                # Clear code
    endofinfo = clearcode + 1                  # End of information code
    nextcode = clearcode + 2                   # Next available code (starts after clearcode)
    lastchar = None                            # Last character written to the dictionary
    stringbuf = ''                             # Buffer containing uncompressed data before flushing it to compressed form
    # Initialize dictionary with first two codes (clear and end of info symbols)
    dict = ['']*(dictsize+1)
    dict[clearcode] = 'CLEARCODE'
    dict[endofinfo] = 'ENDOFINFO'

    # Loop over all rows of the image
    for row in range(height[0]):
        
        # Reinitialize dictionary on every row change
        dict = ['']*(dictsize+1)
        dict[clearcode] = 'CLEARCODE'
        dict[endofinfo] = 'ENDOFINFO'
        
        # Create new string buffer
        stringbuf = ''

        # Loop over each column of the row
        for col in range(width[0]):

            # If there's still data left in the buffer, just pick randomly one bit from it
            if len(stringbuf) > 0:
                charbit = rng.read(1)[0] % 2
                
            
            # Otherwise generate a new symbol
            else:

                # Pick random bit for first test
                charbit = rng.read(1)[0] % 2
                
                # Terminate when we reached maximum code length (12 bits)
                while charbit == 0:
                    charbit = rng.read(1)[0] % 2
                    
                    
                # Check if we're going to encode a clear or end of information code
                if (nextcode <= dictsize 
                        and ((lastchar is None
                                and dict[nextcode][0]==str(charbit))
                            or (lastchar is not None 
                                and lastchar==dict[nextcode][0]))):

                    # Encode clear code instead of last character
                    stringbuf += str((clearcode >> (len(stringbuf)-1))*2 | charbit)

                    # Reset encoder state
                    lastchar=None
                    nextcode+=1
                    continue

                # Check if we need to begin encoding a new character
                elif nextcode < dictsize \
                        and (dict[nextcode]=='CLEARCODE' 
                            or dict[nextcode]=='ENDOFINFO'):
                    
                    # Encode clear code instead of last character
                    stringbuf += str((clearcode >> (len(stringbuf)-1))*2 | charbit)

                    # Reset encoder state
                    lastchar=None
                    nextcode+=1


                # We don't need to do anything special; simply construct a new character and add it to the buffer
                else:
                    
                    # Add random bit to the string buffer
                    stringbuf += str(charbit)


                    # Allocate a new codeword if we can't fit more characters in the dictionary
                    if nextcode >= dictsize:
                        
                        # If we've already encountered a character then assign it as prefix of newly generated word
                        if lastchar is not None:
                            
                            dict[nextcode] = dict[lastchar]+dict[charbit]
                            
                        else:

                            dict[nextcode] = dict[charbit]
                        
                        # Increase next free dictionary index
                        nextcode+=1

                        # Throw away oldest entry if necessary
                        if nextcode > dictsize:
                            del dict[1]
                        
                    # Remember currently processed character as "last"
                    lastchar = charbit
                    
            # Flush the current buffer contents whenever we reach the maximum
            # codeword length or target codeword size
            if nextcode >= 2**(lzwminsize+1) or (len(stringbuf)+1) >= (1<<(lzwminsize)):
            
                # Compressed string is always padded with zeroes until next power of two
                padamount = -(len(stringbuf).bit_length()+1)%8
                if padamount!= 0:
                    stringbuf += ''.join([chr(0)]*padamount)
                
                # Reverse bit order of every octet (required by specification)
                stringbuf = stringbuf.encode()[::-1].decode()
                
                # Write the length of the compressed string
                write_uint16()
                
                # Write actual string content
                out.write(bytes(int(stringbuf[i:i+8], base=2) for i in range(0, len(stringbuf), 8)))
                
                # Reset buffer and encoder state
                stringbuf=''
                nextcode=clearcode+2
                lastchar=None


    # Flush remaining parts of the encoder state
    if len(stringbuf)>0:
    
        # Compressed string is always padded with zeroes until next power of two
        padamount = -(len(stringbuf).bit_length()+1)%8
        if padamount!= 0:
            stringbuf += ''.join([chr(0)]*padamount)
        
        # Reverse bit order of every octet (required by specification)
        stringbuf = stringbuf.encode()[::-1].decode()
        
        # Write the length of the compressed string
        write_uint16()
        
        # Write actual string content
        out.write(bytes(int(stringbuf[i:i+8], base=2) for i in range(0, len(stringbuf), 8)))
        
    
    # Ending marker
    out.write(b'\x00')