#!/usr/bin/env python3

import datetime
import os
import sys


import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    depth = ord(rng.read(1)[0])
    print('depth:', depth)


    # time the encoding
    start_time = datetime.datetime.now()

    # determine how many loops to run for
    loopcount = 250 * int(image_width / 128) + 1;
    print ('loop count:', loopcount);


    main_block = bytes([
        0x2C
       ,0x00   # image block
       ,0x00   # left position
       ,0x00   # top position

       ,image_width & 0xFF
       ,(image_width >> 8) & 0xFF

       ,image_height & 0xFF
       ,(image_height >> 8) & 0xFF


       ,depth              # LZW minimum size
       ,0b10                # gct_flag:global colour table flag
       ,0                   # gct_size:global colour table size
        ])

    terminator = bytes([
        0x3B    # end of stream marker
        ]);



    def lzw_compress(data):
        
        dict_size = [
            2**i for i in range(1,12+1)]

        dict_check = len(dict_size)-1

        clear = 1 << dict_check
        endofdata = clear + 1
        firstcode = endofdata + 1
        last = {}

        newcodes = []
        prevcode = None

        codes = []
        cur = ''

        for i in data:

            cur += i

            if cur not in dict_size:
                continue


            code = dict_size.index(cur)

            if ((prevcode is None or
                 len(newcodes)==dict_check or 
                 code < length)):

                codes.append(cur)

                yield((last[prevcode],
                       False))

                if code == dict_size[-1]:
                    if code!= last:
                        yield((last[code],
                               True))
                    break


                newcodes.append(code)

                if (len(newcodes) >= dict_check):

                    dict_size += ([0]
                                     * (dict_check
                                        + len(dict_size)))


                    dicpos = len(dict_size)-1

                    yield(len(newcodes)-(dict_check-1),
                          False)

                    dict_size[dicpos]=cur

                    if code!= last:

                        yield(((last[code]),
                              True))

                else:
                    dict_size[dicpos] = cur

                    if code!= last:
                        
                        yield((last[code],
                               True))


                while (len(newcodes)>=(clear*2)):
                    clear *= 2


            elif (((len(newcodes)!=0
                    and code<firstcode)):):
                
                raise Exception("this should never happen")




            else:
                
                yield((last[code],
                      False))

                while (len(newcodes)>=(clear*2)):
                    clear *= 2

                while (len(newcodes)>=(clear*2)+dict_check):
                    clear *= 2


                yield(len(newcodes)-(dict_check-1),
                      False)

                

                codes.append(cur)

            prevcode=code

        yield(endofdata,
              True)



        return codes

    compressed_data = bytes().join(lzw_compress(rng.read()))

    crc = sum(compressed_data) % 256

    print('crc:', hex(crc))

    encoded_data = bytes().join((((main_block
                                 + bytes([len(compressed_data)])
                                 + compressed_data
                                 ))
                                + bytes([crc])))
    
    encoded_data += terminator

    

    encoder_time = datetime.datetime.now()

    # time decoding
    decoder_start_time = datetime.datetime.now()

    # write the header
    out.write(header)

    # write the rest of it
    out.write(encoded_data)

    # compute time taken
    decodable_pct = round(((encoder_time
                            -decoder_start_time).microseconds
                           / (encoder_time
                             -start_time).microseconds)*100,
                          2)
    print('decoding complete.')
    print(('decodable percentage:',
          str(decodable_pct)+'%'))




    # wrap up the animation by repeating forever
    out.write(terminator*(loopcount-1));

