#!/usr/bin/env python3

import os
import sys
from typing import BinaryIO


# Generates a random GIF file to `out` using the specified
# RNG, which should be either /dev/urandom or an instance of
# SystemRandom (see https://docs.python.org/3/library/random.html).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width=500, height=500, black_white=False
    ) -> None:

    # Write the magic bytes
    out.write('GIF87a'.encode())

    # Write a bunch of empty blocks for initialisation
    for _ in range(10):
        out.write(struct.pack('<B', 0))

    # Write screen descriptor block
    if not black_white:
        flags = 0x80 | 0x70 | 0x40        # Default graphics control, no interlacing
        packed_flags = [
            ((width & 0xff00) >> 8),      # Width
            (width & 0x00ff),              # Width
            ((height & 0xff00) >> 8),     # Height
            (height & 0x00ff),             # Height
            0xf,                          # Color resolution
            0,                            # No background color index
            0,                            # Pixel aspect ratio
        ]

        # Write flag byte
        packed_flags[0] |= len(packed_flags)-1
        out.write(bytes([packed_flags[0]]))

        # Write remaining bytes as single block
        out.write(bytes(packed_flags[1:]))

    else:
        # Backwards compatibility: write the same image twice with different colors
        out.write(b'\x00\x00')   # White
        out.write(b'\xFF\xFF')   # Black

    # Write LZW code size
    out.write(b'\xe0')         # Minimum code size

    # Set up two buffers of appropriate length and write them both at once
    # Thanks to <NAME> for pointing this out via StackOverflow
    pixels = []
    packstring = '<{}B'.format((width*height)//8)    # Packs all bits into one byte
    while True:
        c = rng.read(1)[0]
        bytes_to_read = min((c+1)*4, len(data))
        chunk = data[:bytes_to_read]
        data = data[(c+1)*4:]           # Remove from buffer what we just read
        pixels += list(chunk)           # Add to pixel list
        if len(pixels)==len(data)*8//6: break    # Finish when we have enough pixels
    packed_pixels = struct.pack(packstring,*pixels)
    out.write(packed_pixels)
    
    # EOF
    out.write(b';;')
