#!/usr/bin/env python3

import os
import struct
from typing import IO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom)
def generate_random_gif(
        rng: IO[bytes] = open('/dev/urandom', mode='rb'), out: IO[bytes] = None
    ):

    def randint() -> bytes:
        return rng.read(1)[0].to_bytes(1, 'big')
    
    if out is None:
        out = open(os.path.expanduser('~/random-gif-%d.gif' % os.getpid()),
                   mode='wb')
        
    # Header information
    out.write(b'GIF89a')                        # Signature
    w, h = [randint() for _ in range(2)]        # Width and height
    clr = [ord(randint()) for _ in range(6)]   # Global color table
    out.write((w+h).to_bytes(2, 'little'))    # Image block size
    out.write((0<<7)+((clr[5]&0x07)<<4)+(clr[1]>>4))      # Flags
    out.write(w)                               # Width
    out.write(h)                               # Height
    out.write(b'\x87')                         # Pixel aspect ratio

    bg = ord(randint()) | ((ord(randint())&0x07)<<8)       # Background color
    out.write(bg.to_bytes(2, 'little'))                  # Logical screen
    out.write(clr[:3])                                    # Global color table
    out.write(b'\xff\xf9')                                # Code size
    out.write(b'\0'*12)                                   # Extra info

    # Local image descriptor
    lzwmincodesize = ord(randint())                       # Minimum code size
    out.write((((lzwmincodesize+1)<<5)|((w//128)+1)).to_bytes(1,'little'))
    out.write(((h//128)&0xff).to_bytes(1,'little'))
    out.write(b'\0\0')                                     # Interlace flags
    out.write(b'\0\0')                                     # Sorting flag
    out.write(b'\0\0')                                     # Reserved fields
    out.write(b'\0\0')                                     # Size of LZW subblock

    # Start of "image" data (LZW compressed data)
    data = b''
    while True:
        d = rng.read(1)
        if not d or len(data) >= 256:
            break

        ch = ord(d)
        if ch > lzwmincodesize:
            data += data[-ch:]
        else:
            data += d
        data += randint()
        bs = min(len(data)-1, lzwmincodesize)
        
        # Write control character and length to output stream
        out.write(struct.pack('<BB', 1+(bs==255)*2, bs))
        # Write LZW subblock with reversed bit order and from last to first
        out.write(bytearray([
            0xFF - (c >> i & 0x01)
            for c in data[-bs:]
            for i in range(8)
        ]))

    out.close()