#!/usr/bin/env python3

import random
import struct
from typing import BinaryIO


# The preamble of a valid GIF file (header and logical screen descriptor)
preamble = bytes([
    0x47, 0x49, 0x46, # 'GFi', magic signature
    0x38, 0x39, 0x61, # version string
    0x03,             # Logical Screen Descriptor block follows

    0x2c, 0x00,       # 4-bit color depth, no global palette
    0x00, 0x00, 0x00, # Upper left corner X=0, Y=0
    0xff, 0x00, 0x00, # Width is 255 pixels, height is infinite
    0x00              # Greyscale output
])



# A single frame contains image data for a single duration
frame = [
    0x21, 0xf9, 0x04, 0x00, 0xa0, 0xff, # Image descriptor with delay time of 1 second
    0x2c, 0x00, 0x00,                   # Logical screen descriptor,
    0x00, 0x00,                         # No global palette
    0x00,                               # No LZW compression
    0xff                                # Frame ends here
]




# Appends the given integer to the specified variable length integer list
def appendVarLenInt(out: list[int], val: int):
    if not -0xFF < val < 0xFF: raise ValueError('Value too big.')

    while True:
        val, mod = divmod(val, 0xFF)

        out.append((mod))

        if val == 0: break

    return out

# Converts an unsigned short value into a byte pair that can be packed
# as little endian
def uint16LEBytes(val: int) -> bytes:
    return struct.pack('<H', val)

# Returns a byte sequence containing a valid GIF file with one empty frame
def makeValidEmptyGif() -> bytes:
    out = []
    
    # Append the preamble
    out += preamble
    # Append the frame object
    out += frame


    # Make sure we end on a byte boundary
    while len(out) % 8!= 0:
        out.append(0x00)
    

    # Add the data size field at offset 6
    dataSizeOffset = 6
    out[dataSizeOffset : dataSizeOffset+3] = \
                    uint16LEBytes(len(out)-dataSizeOffset)

    return bytes(out)

# Generates a random GIF file into `out` using the
# given pseudorandom number generator in binary format
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    out.write(makeValidEmptyGif())
    pass