#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO
from dataclasses import dataclass
from PIL import Image

@dataclass
class BlockInfo:
    size: int
    offset: int
    
    @property
    def end_offset(self):
        return self.start_offset + self.size

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    BLOCKSIZE=8*1024
    MAXBLOCKCOUNT=None

    # Read entire contents of input file as bytes
    infile = open(sys.argv[1], "rb")
    inbuffer = infile.read()
    infile.close()
    
    im = Image.open(inbuffer)

    # Determine pixel width and height from image
    # NOTE!! This assumes that all images have the same dimensions!
    PIXELWIDTH = im.width // 8
    PIXELHEIGHT = im.height

    palette = list(im.getpalette()[::3])

    # Create color table for this image by rearranging RGB values 
    # from palette so they are ordered into a more-standard format
    colortable = []
    for i in range(len(palette)//3):
        r, g, b = palette[(i*3)], palette[(i*3)+1], palette[(i*3)+2]
        colortable += [(r << 2)|(g >> 6),(g << 4)|(b >> 4)]

    # Header consists of:
    ##     Signature: GIFFORMAT
    ##     Version:   89a or 87a
    ##     Width:     Pixel Width divided by 8
    ##     Height:    Pixel Height
    ##     Color Table Size: Number of colors in color table; always 256
    ##     Background Index: Index used to represent transparent pixels
    ##     Pixel Aspect Ratio: Always set to 0
    ##     Global Color Table: 256 triplets representing the color table.
    ##         Each entry is stored as three bytes, red, green, blue respectively.
    ## First two bytes should be set to \x47\x49 after reading signature.

    # Calculate size of global color table plus background index value
    tablesize = len(colortable)
    backind = 1
    headsz = 2+3+1+tablesize*3

    # Initialise block information array
    info = []

    # Check how many blocks we need to read
    currentoffset = headsz
    blockcount = 0
    while True:
        # Work out next block's offset
        newoffset = currentoffset + BLOCKSIZE

        # Stop when we've reached the last block or did enough already
        if newoffset >= len(inbuffer): break
        elif (MAXBLOCKCOUNT!= None) and (newoffset < len(inbuffer)-BLOCKSIZE): 
            if blockcount == MAXBLOCKCOUNT:
                break
            else:
                pass
        
        # Append new block information to block information list
        info.append(BlockInfo(currentoffset))
        
        # Increment our counters
        blockcount += 1
        currentoffset = newoffset

    # Get the first field of the header, which will be the version string
    version = "\x87"
    if len("".join([chr(_) for _ in info])) <= 2**16: 
        version = "\x89"

    # Construct the header
    header = "GIF%s%s\x00\x00\x00<\x00\xff\x00,\x00@\x00;\x00\x00\x00\x00\x00" % (version, "".join([chr(_) for _ in info]).encode("hex"))
    print(header.decode('hex')[0:headsz].encode('hex'))

    # Write header and length of image to output
    out.write(header.decode('hex') + len(inbuffer).to_bytes(3, 'little'))
    
    
    blocknumber = 0
    for _ in info:
        # Read next block and write it to output
        block = inbuffer[_:]
        if not blocknumber:
            # If it's the first block then also write
            # the color table and background index to the start
            lookup = b''
            for i in range(PIXELHEIGHT//2):
                lookup += struct.pack("B",backind)
                lookup += struct.pack("B",backind)

            out.write(lookup)
        out.write(block)

        # Update counters
        blocknumber += 1


