#!/usr/bin/env python3

"""
Copyright © 2021 <NAME>
Licensed under the MIT license. See LICENSE file for details.
"""

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_WIDTH      = 512
DEFAULT_HEIGHT     = 256
DEFAULT_COLORS     = 2**8
DEFAULT_BGCOLORIDX = 0
DEFAULT_FGCOLORIDX = 1
DEFAULT_DPI        = 72
DEFAULT_OUTPUT     = 'random.gif'

GIF_HEADER_OFFSET = {
    0x0A : ('width',           '<H'),
    0x0C : ('height',          '<H'),
    0x0E : ('colorTableSize',  '<B'),
    0x0F : ('backgroundColorIndex', '<B'),
    0x10 : ('pixelAspectRatio',       '?'),

    0x20 : ()
}

GIF_BLOCK_SIZE = [
    0x02,
    0x03
]




class RandomGifGenerator():

    def __init__(
            self,
            width=None,
            height=None,
            colors=None,
            bgColorIdx=None, 
            fgColorIdx=None,
            dpi=None,
            output=None
        ):


        self._info = {}

        if width is None:
            self._info['width'] = DEFAULT_WIDTH
            
        elif isinstance(width, int):
            if width <= 0 or width > 2**16 - 1:
                raise ValueError("Invalid parameter")
            
            self._info['width'] = width

        
        if height is None:
            self._info['height'] = DEFAULT_HEIGHT
            
        elif isinstance(height, int):
            if height <= 0 or height > 2**16 - 1:
                raise ValueError("Invalid parameter")

            self._info['height'] = height
        
        
        if colors is None:
            self._info['colors'] = DEFAULT_COLORS
            
        elif isinstance(colors, int):
            if colors <= 0 or colors > 2**8 - 1:
                raise ValueError("Invalid parameter")

            self._info['colors'] = colors


        if fgColorIdx is None:
            self._info['fgColorIdx'] = DEFAULT_FGCOLORIDX
            
        elif isinstance(fgColorIdx, int):
            if fgColorIdx <= 0 or fgColorIdx >= self._info['colors']:
                raise ValueError("Invalid parameter")

            self._info['fgColorIdx'] = fgColorIdx



        if bgColorIdx is None:
            self._info['bgColorIdx'] = DEFAULT_BGCOLORIDX
            
        elif isinstance(bgColorIdx, int):
            if bgColorIdx < 0 or bgColorIdx >= self._info['colors']:
                raise ValueError("Invalid parameter")

            self._info['bgColorIdx'] = bgColorIdx


        

        if dpi is None:
            self._info['dpi'] = DEFAULT_DPI

        elif isinstance(dpi, int):
            if dpi < 0 or dpi > 2**16 - 1:
                raise ValueError("Invalid parameter")

            self._info['dpi'] = dpi


        if not output:
            self._info['output'] = DEFAULT_OUTPUT
        else:
            self._info['output'] = output



    @property
    def info(self):
        return self._info


    def generate_random_gif(self):
        
        with open('/dev/urandom', mode='rb') as rng,\
             open(self._info['name'],    mode='wb') as out:

            return generate_random_gif(
                    rng, 
                    out, 
                    **self._info
                )



# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# correspond to the information found here:
# https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(
        rng: BinaryIO, 
        out: BinaryIO, 

        name=DEFAULT_OUTPUT,
        width=DEFAULT_WIDTH,
        height=DEFAULT_HEIGHT,
        colors=DEFAULT_COLORS,
        bgColorIdx=DEFAULT_BGCOLORIDX,
        fgColorIdx=DEFAULT_FGCOLORIDX,
        dpi=DEFAULT_DPI
    ):

    
    # First create and write the Header
    _writeHeader(
            rng, out, 
            width, height, colors,
            bgColorIdx, fgColorIdx,
            dpi
        )

    # Next we need to specify the Color Table
    # which is simply an array of RGB triplets
    # Each entry must be distinct (not checked!!!)
    colorTable = []
    while len(colorTable) < colors:
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        colorTable.append((r,g,b))

    # now we can write the table into the output stream
    # each triplet is stored as 3 bytes
    for colTriplet in colorTable:
        for val in colTriplet:
            out.write(bytes([val]))

    # The parameter `name` is used only to calculate some
    # properties such as the size of the stream etc...
    # This should be changed later to use the appropriate 
    # methods to obtain this kind of information from the OS
    dataBlockSize = calcDataBlockSize(name)
    frameDuration = 1 / float(fps) * 1e+3


    # Finally make up some frames
    currentFrame = 0
    while True:

        # Keep track of the amount of bytes written so far
        byteCount = 0

        # Write another Frame Marker into the stream
        out.write(b'\x21')
        byteCount += 1

        # Then write the Extension Introducer (Label Byte)
        # followed by the Graphics Control Label
        out.write(b'\xF9')
        out.write(b'\x04')
        byteCount += 2

        # Now comes the block size
        out.write(dataBlockSize)
        byteCount += 1

        # Delay Time & Transparent Color Index
        delayTime = int(frameDuration).to_bytes(2,'little')
        out.write(delayTime)                           # Delay time
        out.write(bytes([transparentColorIndex]))      # Transparent color index
        byteCount += 3                                 # The two above counts for three bytes

        # Now comes the Image Descriptor
        out.write(b'\x2c')                              # image separator
        out.write(imageLeftPos.to_bytes(2, 'little'))   # left pos
        out.write(imageTopPos.to_bytes(2, 'little'))    # top pos
        out.write(imageWidth.to_bytes(2, 'little'))     # image width
        out.write(imageHeight.to_bytes(2, 'little'))    # image height
        out.write(bytearray([0x87 | lctSize]))          # flags
        out.write(lzwMinCodeSize.to_bytes(1, 'little')) # LZW min code size (always zero?)
        byteCount += 11                                # Everything but the image data counts for eleven bytes

        # Write the Image Data
        imgPixels = [(0, transparentColorIndex)]*imageWidth*(imageHeight//2)
        for i in range(imageHeight // 2):
            for j in range(imageWidth):

                p0 = rng.read(1)[0]
                p1 = rng.read(1)[0]
                px = (p0 << 4) | p1
                
                imgPixels[i*imageWidth + j] = (px, colorTable[px][0], colorTable[px][1], colorTable[px][2])

        for pixel in imgPixels:
            out.write(struct.pack('<BBBB', pixel[0], pixel[1], pixel[2], pixel[3]))
            byteCount += 4

        # Now comes the block terminator
        out.write(b'\x00')
        byteCount += 1

        assert byteCount == dataBlockSize, "The stream got corrupted"

        # Increase current frame number
        currentFrame += 1
        print(f"Frame #{currentFrame}")



# Calculates the appropriate Data Block Size depending on
# the length of the provided filename `name`
def calcDataBlockSize(name):

    # In case we want to store the image data in multiples blocks
    # then we have to divide the total size of the stream by n
    # where n denotes the amount of blocks that our structure has
    # divided into
    # To avoid problems due to rounding errors we simply choose
    # the largest possible value that still allows us to fit the
    # stream into one single block without exceeding its capacity
    sizes = [0x3fff, 0x1fff, 0xffff, 0x7fff]
    for size in sizes:

        if len(name) < size:
            return size

    raise Exception('Cannot find suitable Data Block Size!')




# Tries to identify the correct header layout by checking
# the second character of every line until one matches the
# signature
# Returns the offset at which the first line begins
def _findHeaderLayout(stream):

    for offs in range(len(stream)):

        try:
            testChar = chr(stream[offs]).lower()

            if testChar!= '#' \
               and testChar!= '/'\
               and testChar!= '\n':

                continue

        except Unicode