#!/usr/bin/python3

import os
import sys


class Error(Exception):
    pass



class RandomNumberGenerator:
    def __init__(self, path=None):
        self._path = path if (path is not None) else '/dev/urandom'


    @staticmethod
    def _chunk(size):
        return size & ~0x7F | 0x80


    ## Read and validate the 12-byte header
    #  @throw IOError If file ends unexpectedly or does not match format
    #                 as defined in https://tools.ietf.org/html/rfc2361
    def read_header(self, f):

        header = f.read(13)  # 12 plus null terminator
        if len(header)!= 13:
            raise IOError('Unexpected end of file.')


        signature = header[0]
        version   = header[1]

        if signature!= ord('G'):
            raise IOError('Invalid GIF signature.')

        if version!= ord('8'):
            raise IOError('Only GIFFormatVersion 8 supported at this time.')

        width        = header[5] << 8 | header[4]     # MSB first
        height       = header[7] << 8 | header[6]     # MSB first
        bitsPerPixel = header[9]                     # No color table
        bgIndex      = header[10]                    # Background Index
        aspectRatio  = header[11]

        if ((width <= 0)           or
            (height <= 0)          or
            (bitsPerPixel < 2)     or
            (bgIndex >= 256)):

            raise IOError(
                    'Invalid header parameters {} {}'.format(
                        width, height))

        if aspectRatio == 0:    # No aspect ratio should be zero
            pass
        elif aspectRatio > 0:   # Value is invalid
            print('Warning: Invalid aspect ratio {}.'.format(aspectRatio))
        else:                   # Negative value means no aspect ratio specified
            print('Warning: Aspect ratio unspecified.\n')



    def read_data(self, f, width, height, numBytesPerRow):
        rowLen = numBytesPerRow * width
        bufLen = rowLen + height

        buf = [b'\x00'] * bufLen

        for i in range(bufLen - 1):
            if i % rowLen == 0:
                valByte = f.read(1)[0];

                if i + rowLen < bufLen:
                    # copy last pixel to next line
                    buf[i+rowLen] = buf[i+rowLen-numBytesPerRow]

                # set all pixels in current line from right to left
                for j in reversed(range(i, min(i+min(rowLen, (valByte >> 3)), bufLen))):
                    buf[j] = chr((valByte & 0x7) ^ 0x7).encode()

                    valByte >>= 3;

        assert len(buf) == rowLen + height

        return buf[:rowLen] + buf[-height:]









## Generates a random GIF file into `out` using the
#  random number generator `rng`.
#
#  Note that this will rewind the output stream before writing anything.
#
#  @param rng An object supporting a.read method
#             which returns an array of up to N bytes
#             where N <= 2^16.
#
#  @param out A binary I/O object supporting write access
def generate_random_gif(rng, out):

    out.seek(0)
    out.truncate()

    header = b'GIF89a'

    # Write out GIF header
    out.write(header)


    # Get dimensions of image (2^(8*4) pixels by default)
    width        = rng.read(2)
    h            = rng.read(2)
    bpp          = rng.read(1)
    backgroundId = rng.read(1)
    aspectRatio  = rng.read(1)
    screenBlock  = (
        width[0], width[1], h[0], h[1],
        bpp,
        0,                      # flags field
        backgroundId,
        aspectRatio,)

    # 12 bytes total
    screenBlock += b'\x00'*3

    # TODO: Should we support color tables?

    out.write(struct.pack('<BBBBBBBBB', *screenBlock))


    # Get repeat count (currently unused)
    repCount = rng.read(2)

    # No app extension blocks
    extSize = struct.pack('<B', 0)

    # Null separator
    sep = '\x00'.encode()

    # Write full-sized application extension block
    out.write(
                struct.pack('<Bxxxxxx', 0xFF) +
                extSize +
                struct.pack('<BB',
                            0xF9, 0x04)    +  # Application Extension Block ID
                                # and size
                sep +              # Name field (null terminated string)
                sep +              # Authentication code field (null term.)
                b'\x00\x00\x00'    )  # Subblock size (always even?)

    out.write('\x00'.encode()) # Stop subblocks with null separator


    # Get palette information (how many entries)
    palEntries = rng.read(1)

    # Palette size can't exceed 256 colors
    if palEntries[0] > 0 and palEntries[0] < 256:
        # Not sure what to do about this...
        out.write(palEntries)

        # 3 bytes per entry RGB
        out.write(b'\x00' * (palEntries[0]*3));
    else:
        out.write(b'\x00');


    # Get color resolution
    colRes = rng.read(1)

    # Get sort flag
    sortedFlag = rng.read(1)

    infoBlock = (colRes[0], sortedFlag[0])

    # 2 bytes total
    out.write(struct.pack('<BB', *infoBlock))


    # Write global color table
    for _ in range(0, 256):
        out.write(rng.read(3))




    # Now comes the fun part! :D
    while True:
        # Get control block type
        cbType = rng.read(1)

        # End of block
        if cbType[0] == 0x00:
            break

        if cbType[0] in [
                0x2C, 0x21]:


            imgWidth = rng.read(2)
            imgHeight = rng.read(2)
            lctFlag = rng.read(1)

            # TODO: Handle LCT flag properly

            xOffset = rng.read(2)
            yOffset = rng.read(2)

            # Set local color table size
            lctSize = rng.read(1)

            # Sort flag doesn't exist
            skip = rng.read(1)

            interlaceFlag = rng.read(1)

            # Size of graphic control block in bytes
            blockLength = rng.read(1)

            # Ignore block length
            gcData = rng.read(len(blockLength))

            # Image data should be placed after graphics control block
            data = []

            while True:
                # Get block op code
                blkCode = rng.read(1)

                # Trailer
                if blkCode[0] == 0x2C:
                    break

                if blkCode[0] == 0x21:
                    #TODO: Parse the rest of these blocks
                    # This isn't the most important thing so let's move on
                    continue


                # We need a minimum of two bytes to specify a block length
                if len(blkCode) == 1:
                    blkLen = rng.read(1)
                else:
                    blkLen = blkCode
                    blkCode = blkCode[:-1]


                # Plain old uncompressed data
                if blkCode[0] == 0x20:
                    data += rng.read(int.from_bytes(blkLen, "big"))

                # Plain text
                elif blkCode[0] == 0x01:
                    # TODO: Check bounds here
                    txtData = rng.read(int.from_bytes(blkLen,"big"))
                    txtDict = {'text': txtData}
                    out.write(str(txtDict).encode())


                # Graphics control extension
                elif blkCode[0] == 0xF9:
                    out.write(blkCode);
                    out.write(struct.pack('<H', int.from_bytes(blockLength,'big')))

                    # Unused fields
                    4 * rng.read(1)
                    rng.read(1)


                    # Disposal method
                    dispMethod = rng.read(1)

                    # User input flag
                    usrInputFlg = rng.read(1)

                    4 * rng.read(1)
                    rng.read(1)


                    out.write(dispMethod)
                    out.write(usrInputFlg)



                # Comment extension
                elif blkCode[0] == 0xFE:
                    comData = rng.read(int.from_bytes(blkLen,'big'))
                    comDict = {
                            'comment': comData,
                            }
                    out.write(str(comDict).encode())


                # Plain text extension
                elif blkCode[0] == 0x01:
                    pln