#!/usr/bin/env python3

"""A script to generate random GIF files."""


import argparse
import io
import itertools
import subprocess as subp
import sys

from PyQt5.QtCore import QByteArray
from PyQt5.QtGui import qRgb



# Generates a random GIF file into `out` using the
# random number generator `rng`. An optional palette can be used for coloring
# the image by passing it in as an arg. The default is to use pallette #0 from
# gifsicle's list of schemes: http://www.lcdf.org/~eddietwo/gifsicle/schemes.html
def generate_random_gif(
        rng: io.BytesIO, out: io.BytesIO, palette=None
    ) -> None:

    if not palette:
        scheme = 'xterm-256color'
        palette = (subp.check_output([
            '/Applications/GIF Brewery 3.app/Contents/Resources/gifbrewery',
            '-n',
            '--list-schemes',
            scheme]).decode('utf-8')).split('\n')[1].strip()

        palette = [int(col*255) for col in
                    subp.check_output([
                        '/Applications/GIF Brewery 3.app/Contents/Resources/'
                            'gifbrewery',
                        '-n',
                        '--list-colors',
                        palette]).decode('utf-8').split()]


    header = bytearray((0x47, 0x49, 0x46, 0x38, 0x39, 0x61))

    # size of global color table
    header += bytearray((
        0x00, 0xFF,    # width and height
        0xF7,           # flags - no local colors, interlace, sort, has GCT
        0x00, 0x00      # background color index & pixel aspect ratio
    ))

    numcols = len(palette)//3
    # the actual global color table comes right after this
    header += bytearray((numcols,))

    for i, palentry in enumerate(itertools.islice(palette, 0, numcols)):
        # RGB pallete entries are all given as one byte each
        red   = (palentry >> 16) & 0xFF
        green = (palentry >>  8) & 0xFF
        blue  = (palentry      ) & 0xFF
        entry = bytearray((red, green, blue))

        header[13] += 1   # increase count of global entries that follow
        header.extend(entry)     # add the new entry

    out.write(bytes(header))         # write header

    numpixels = int(((out.getbuffer().nbytes - header.__len__()) / 3)+1)
    print("num pixels", numpixels)

    # create a qt array for holding these pixels
    pixarray = QByteArray()
    pixarray.allocate(numpixels*10)

    for _ in range(numpixels):
        # choose a random color index as the current pixel
        coloridx = rng.read(1)[0] % numcols

        # convert it to a qt color format which is just rgb triplets
        rgb = QByteArray([
            *[chr(c) for c in
              list(struct.unpack('<BBB',
                                  bytes((coloridx,)*3)))],
            0])
        pixarray.append(rgb)

    out.write(pixarray)        # append pixel data

    # write terminating string
    out.write(b';')








