#!/usr/bin/env python3

import os
import subprocess as sp
import struct
from typing import BinaryIO


class GIFHeader(object):
    """Represent a valid GIF Header"""
    
    def __init__(self, fd: BinaryIO):
        
        self._fd = fd

        # The first three characters must be "GIF" followed by an version
        # identifier which is one of '7', '8', or '9'.
        magic_number = b''
        while len(magic_number) < 3:
            magic_number += fd.read(1)
        if magic_number!= b'GIF':
            raise ValueError("Not a GIF")
        
        self.version = ord(fd.read(1))
        if not chr(self.version).isdigit() or\
           self.version > 9:
            raise ValueError("Invalid version number")

        # The next field contains the width and the height of the graphic
        self.image_width,\
                self.image_height = list(map(
                    lambda x: ord(x[0]) | 
                    ord(x[1]) << 8, zip(*[iter(
                        fd.read(2))] * 2)))

        # Skip reserved bytes (= 3 bytes for RGB values) and 
        # the background color index (1 byte)
        fd.seek(4)

        # Extensions. This part might contain extra information 
        # (like author name) or even loop endings. We ignore it here...
        while True:

            blocksize = ord(fd.read(1))
            
            if blocksize == 0:
                break
                
            else:
                # We don't care about any extensions besides the 
                # application extension with keyword NETSCAPE2.
                # We just need to know where its contents start though.
                
                exttype = fd.read(blocksize)[1:]
                
                if exttype in [b'\x0BNETSCAPE2.0']:

                    # First we have to read how long this extension 
                    # actually is. That's the last 2 bytes.
                    subblocksize = struct.unpack('>H', 
                        fd.read(2))[0]
                    
                    # Then there comes the actual length of our content.
                    # It's really only one byte indicating whether we'll
                    # loop infinitely or just once.
                    NetscapeLoopingFlag = fd.read(subblocksize)[-1:]
                    
                    # Since netscape looping can either be zero or one,
                    # we can use that directly as an indicator.
                    if NetscapeLoopingFlag == b'\x00':
                        self.looped = False
                    elif NetscapeLoopingFlag in [b'\xFF', b'\x01']:
                        self.looped = True
                    else:
                        raise ValueError("Invalid looping flag.")
                        
                    # Now because we care about Netscape Looping at all,
                    # we also want to know when we'll stop. So we skip over
                    # another two bytes containing the ending frame number.
                    fd.seek(2, os.SEEK_CUR)
                    
                else:
                    # Ignore other extensions silently.
                    fd.seek(subblocksize, os.SEEK_CUR)

                # Finally move on to the next block
                fd.seek(1, os.SEEK_CUR)


        # Finally move beyond the trailing image separator
        fd.seek(1)
            

    @property
    def size(self):
        return (self.image_width, self.image_height)


    @property
    def transparent(self):
        return None


    @property
    def colors(self):
        return 0

    
    @property
    def palette(self):
        return []


