#!/usr/bin/env python3

import os
import sys
import time
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    def next_int() -> bytes:
        return rng.read(4)

    def write_value(val: bytes):
        out.write(struct.pack('>B', val))

    def flush():
        out.flush()


    # Header
    out.write(b'GIF89a')
    # Logical screen descriptor
    write_value((next_int()[0] & 0xF7) | 0x08)    # Screen width
    write_value(next_int()[0] & 0xFF)                # Screen height
    out.write(struct.pack('>B', 0xf0))               # Color table size: None
    write_value(b'\xff')                             # Background color index: None
    write_value(b'\x00')                             # Pixel aspect ratio: Default

    # Global color table
    for _ in range(0xff):                            # Each color is either black or white
        out.write(b'\xff\xff\xff\xc0')              # Black/White color entry

    # Image descriptor
    write_value(b'\x2c')                             # Image separator
    write_value(next_int())                          # Left position
    write_value(next_int())                          # Top position
    write_value(next_int())                          # Width
    write_value(next_int())                          # Height
    out.write(b'\x00')                               # Local color table flag: False
    out.write(b'\x08')                               # LZW minimum code size
    out.write(b'\x01')                               # Image block starts here

    content = []                                     # Buffer for data until we get something useful
    codeSize = 2                                    # The initial code size of an LZW compressed image
    clearCode = 2**codeSize - 1                      # Value that indicates an end-of-stream marker
    eoiMarker = 2**codeSize                         # End-of-image marker

    while True:                                      # Generate infinite stream of data
        value = next_int()                           # Read one byte from /dev/urandom

        if value == b'\x00':                         # Got a null byte
            continue                                # Discard it
        elif value == b'\xff':                       # Got a new frame start sequence
            print(f"CODESIZE={codeSize}, CLEAR={clearCode}, EOI={eoiMarker}")

            flush()                                  # Write all buffered contents
            write_value(content[-1:])                 # Write last byte with its most significant bit set
            write_value(b'\x00')                     # Block terminator
            yield content[:-1]                       # Return everything up to this point

            out.write(b'\x2c')                       # Image separator
            write_value(next_int())                  # Left position
            write_value(next_int())                  # Top position
            write_value(next_int())                  # Width
            write_value(next_int())                  # Height
            out.write(b'\x00')                       # Local color table flag: False
            out.write(codeSize.to_bytes(1, 'little')) # Code size for compression
            out.write(b'\x01')                       # Image block starts here

            content = list(b'\xfe')                   # Reset buffer and add a RasterStart marker
        elif not content:                            # Ignore anything before the first raster start
            continue
        elif content[-1:] == [value]:                # If previous byte was equal to current, repeat it
            content += value                         # Just adding another instance increases the frequency counter
        else:
            content += value                         # Add normal byte to the buffer
            while len(content) >= codeSize:
                chunk = content[:codeSize]           # Grab a full chunk

                try:
                    rep = chunk.index(chunk[-1], 0, codeSize - 1) # Find repeating marker (+1 because of index=0)
                except ValueError:
                    # Repeat chunk has no repeated section!
                    rep = 2**codeSize - 1             # Indicate no repeated section ("Clear code")
                finally:
                    tmp = chunk[:]                    # Make a copy

                    if rep < clearCode:
                        # Repeated part is small enough to fit within the chunk itself
                        tmp = tmp[:rep] + [(tmp[-1])] # Replace repeating part by its own value
                    elif rep <= eoiMarker:
                        # Repeated part is too large to fit within the chunk itself; store it separately
                        tmp = tmp[:rep] + [tmp[-1]]     # Store only last byte as repeated part

                        # Split chunk at the repeating part so we can keep track of it
                        origLen = len(content)          # Original length of buffer
                        content = content[origLen - origLen // abs(rep - clearCode)] \
                                + content[origLen:]       # Cut off original part of buffer, keep appendix
                    else:
                        # We have a bug somewhere...
                        raise Exception("Invalid repeat length!")
                
                    out.write(str(len(tmp)).encode('ascii')) # Number of pixels in this chunk
                    out.write(struct.pack('>B', codeSize))   # Maximum code size used in this chunk
                    out.write(struct.pack('>H', rep))        # The actual value corresponding to the stored code size
                    
                    for c in tmp:                        # Copy chunk to output
                        out.write(c.to_bytes(1, 'little'))

                    del content[:codeSize]              # Remove fully copied chunk
