#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given pseudorandom number generator in binary format
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Read 10 bytes for the GIF header
    raw = rng.read(10)
    assert len(raw) == 10
    # Gather all of the fields from the header
    (
        signature, version, width, height, 
        gctflag, bgindex, aspect
    ) = struct.unpack('6s3B5x', raw)


    if signature!= b'GIF89a':
        print("Invalid signature")
        exit()

    if not (width > 0 and width < 2**16):
        print("Bad width")
        exit()
    
    if not (height > 0 and height < 2**16):
        print("Bad height")
        exit()

    if not (aspect >= 0 and aspect <= 240):
        print("Bad aspect ratio")
        exit()

    # Write back the modified header to the stream
    out.write(raw)

    buf = []
    counter = 0
    while True:

        # Determine the next bit
        bit = -((counter & 7) + 1) >> 3
        value = ord(rng.read(1))

        # Check if we are at the end of the chunk
        if not ((value >> bit) & 1):

            # We are done with this frame
            break
        
        # Collect up to 12 bits
        while True:
        
            # Make sure there is still space in our buffer
            if len(buf) <= 12:
                buf += [None] * 12
            
            # If the next bit is set... 
            if (value >> bit) & 1:

                # Get another byte or as many as necessary 
                newbyte = ord(rng.read(1))
                nbits = min(len(buf), 8-bit)
                
                # Set them all
                for i in range(nbits):
                    j = (i+bit)%8
                    buf[j] = newbyte>>(7-i)&1
                
                # Break when done
                if nbits == 8:
                    break
            
            else:
                break

        # Output the first 8 bytes of the data block
        out.write(struct.pack('<HBBBBBHB',
            0, 
            0,  
            0xFF, 
            0, 
            0,  
            len(buf)-1,
            *buf[:8]))

        # Increase the counter
        counter += len(buf)

        # Reset the buffer
        buf = [*buf[8:]]

    
    # Calculate the length of the image data
    datalen = counter//8

    # Add the image descriptor
    out.write(struct.pack('>HHHHBBBxxx', 
        0,
        0, 
        width, 
        height, 
        0, 
        0, 
        0))

    # Copy over the remainder of the data
    while counter % 8:
        counter -= 1
        byte = ord(rng.read(1))
        out.write(struct.pack('Bxxx',
            byte>>counter&1))
        datalen -= 1