#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# The maximum number of bytes to read from /dev/urandom at once.
MAX_READSIZE = 1024 * 64 # 64 KiB


# Converts an integer to little-endian and returns it as bytes
def lei_to_bytearray(x):
    return x.to_bytes((x.bit_length() + 7) // 8, 'little')


# Generates a random GIF file into `out` using `/dev/urandom`.
# Optionally print every n lines of output (defaults to 50).
def generate_random_gif(
        out: BinaryIO,
        linecount=None
    ):

    if not linecount is None:
        assert type(linecount) == int

    # Write the magic string "GIF" first
    out.write('GIF'.encode())

    # Then write a version number
    out.write('\x89'.encode())


    # Generate a frame size in bytes but don't exceed MAX_READSIZE
    with open('/dev/urandom', mode='rb') as urand:

        # Get frame dimensions width and height
        while True:

            # Read 3 bytes for the size in bytes
            dimstr = urand.read(3)

            width = struct.unpack('<B', dimstr[0])[0] % 100
            height = struct.unpack('<B', dimstr[1])[0] % 100

            # If both are less than or equal to 2 then we have valid values
            if width > 2 and height > 2:
                break


        # Also get background color index
        bgcolor = urand.read(1)[0] % 256

        # Now that we're sure everything went well, write them down
        out.write(dimstr)
        out.write(struct.pack('<B', bgcolor))



    # Number of frames
    numframes = 10

    # Repeat this block until there's enough space left for all the
    # frames in the animation
    while True:

        # Get the image descriptor length which should be 10 bytes long
        idlen = struct.unpack('<B', urand.read(1))[0] % 10
        if len(idlen) < 10:
            idlen += 10

        # Image descriptor raw data
        iddata = urand.read(idlen - 1)

        # And now the actual pixels
        imgdat = urand.read(width*height)

        # Check how much more space we need
        spacereq = ((numframes+1)*idlen) + (width*height)*(numframes)
        remaining = os.fstat(out.fileno()).st_size - out.tell()
        if spacereq <= remaining:
            break


    # Finally, write each frame one by one
    for fidx in range(numframes):

        # Start with the image descriptor
        if idlen >= 10:
            out.write(b'\x2C\x00')
        else:
            out.write(b'\x2C')

        # Then add the raw ID data
        out.write(iddata)

        # And finally write the pixel data
        out.write(imgdat)



    # Calculate the end of file offset based on previous calculations
    eofoffset = out.tell()

    # Go back right after the last frame
    out.seek(-numframes*(idlen+(width*height)), SEEK_CUR)

    # Now calculate the duration between frames
    durations = [struct.unpack('<H', urand.read(2))[0] % 256
                 for i in range(numframes)]


    # And now write the entire GCE extension
    gceext = b'\x21\xF9\x04\x08' \
             + lei_to_bytearray(durations[0]) \
             + b'\x00'

    extblock = urand.read(2)
    while extblock!= b'\x00':

        extsize = ord(extblock[0]) * 256 + ord(extblock[1])

        gceext += extblock + extsize * b'\xff'

        extblock = urand.read(2)

    gceext += b'\x00'

    # Finally, go back again to the beginning and insert the GCE extension
    out.seek(eofoffset);
    out.write(gceext)


