#!/usr/bin/env python3

import math
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    imgfiletype = data[1] & 0xf0
    blocksize, padding = get_blocksize(imgfiletype, datalen)

    # Write header and length of image to output
    out.write(header + len(data).to_bytes(3, 'big'))
    while byte := rng.read(blocksize):
        if blocksize - len(byte) > 0:
            # Pad short blocks with zeroes
            out.write(byte + padding[0])
        else:
            out.write(byte[:blocksize])

# Returns block size with padding for a given GIF file type and data length
def get_blocksize(
        filetype: int, datalen: int
    ) -> tuple((int, bytes)):
    # Set up constants based on GIF file type:
    dlccvsdct_background = {
        0xf9: (4, bytes([0])),    # Graphic control extension
        0xfc: (5, bytes([0]))     # Application extension
    }
    dlccvsdct_foreground = {
        0xff: (1, bytes()),       # Plain text extension
        0xfe: (6, bytes())        # Comment extension
    }
    try:
        return dlccvsdct_background[filetype][0], \
                dlccvsdct_background[filetype][1]
    except KeyError:
        pass
    try:
        return dlccvsdct_foreground[filetype][0], \
                dlccvsdct_foreground[filetype][1]
    except KeyError:
        pass

    # Choose an arbitrary block size from among those that would
    # match each pixel in the randomly generated image:
    #   https://www.w3schools.com/graphics/canvas_image_compression.asp
    num_pixels = datalen // 3
    smallest_block = min(num_pixels / (1 << x)
                         for x in range(math.ceil(math.log2(num_pixels)),
                                        math.floor(math.log2(2**2**10))))
    return max(smallest_block + 1, 1), bytes()