#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width=120, height=100
    ):

    # Header ( 6 bytes )
    header = b'\x47\x49\x46\x38\x39\x61'

    # Logical screen descriptor ( 7 bytes )
    lsd = struct.pack('>HHBBB', width, height,
            8, 7, 0b10000000 | 0b1 << 5 | 0b1<<6)

    # Global color table ( 3 bytes )
    gct = b'\xFF\x00\xFF\xFF\x00'
    
    # Image Descriptor ( 10 bytes ) 
    idd = struct.pack('>BBHBBBBB',
            0x2C,      # image separator
            0b00000000,    # left pos 
            0b00000000,    # top pos 
            width,
            height,
            0b10000000,     # local colour table flag  
            0b00000000,    # interlace flag
            0b00000000,    # sort flag
            0b00000000)    # repeat count 

    # Size for the pixels in this frame
    pixel_data_size = width * height

    # Pixels is an array of size equal to the product of the input sizes
    pixels = [int.from_bytes(rng.read(1), byteorder='big')
              for _ in range(pixel_data_size)]

    # Output stream order (this may not be correct)
    imgstream = b''
    for x in range(width):
        for y in range(height // 8):
            bitplane = []
            for xx in range(8):
                if pixels[y*8*width+xx*width+x] & (1 << (y % 8)):
                    bitplane.append(255)
                else:
                    bitplane.append(0)
            imgstream += struct.pack("B"*len(bitplane), *bitplane)

    # Build and write all frames
    frames = 3
    for i in range(frames):

        # Frame control block ( 8 bytes )
        fcb = struct.pack('>BBBBBB',
                0b00000000,         # reserved 
                0b10000000,         # disposal method: clear
                0b00000000,         # user input (no wait)
                0b00000000,         # transparent colour index is absent
                0b00000000,         # no transparancy
                0b00000000)         # end of frame

        out.write(header       # Gif89a
                  + lsd        # logical screen descriptor
                  + gct        # global color table 
                  + idd        # image descriptor
                  + fcb        # frame control block
                  + imgstream)   # image data 
        