#!/usr/bin/env python3

import os
from typing import BinaryIO

# Generates a random GIF file with `width` by `height` pixels to `out`.
# Uses `/dev/urandom` as the source of entropy.
def generate_random_gif(
        width: int, height: int, out: BinaryIO
    ) -> None:
    if not (0 < width and  0 < height): raise ValueError('Invalid size')

    # We use a fixed RGBA palette since we can't really control it anyway
    palette = bytes([
        0xff, 0x00, 0x00,   0x00, 0xff, 0x00,   0x00, 0x00, 0xff,   0xff, 0xff, 0x00,
        0x55, 0x55, 0x55,   0xaaaa, 0xaa, 0xaa,   0xff, 0xff, 0xff
    ])

    numcolors = len(palette)//3

    bgindex = 7 // numcolors * 3
    palette = bytes([*palette[:bgindex], 0]) + \
              bytes([*(palette[bgindex+1:] + palette[:bgindex+1])])

    ext_hdrs = [b'\xf9\x04\x01\x00',
                b'\xfe\x00\x3b']
    
    hdrlen = sum((len(ext) for ext in ext_hdrs)) + 10
    gfxctl = [0x2c,
              0 | ((numcolors-1)<<4),   
              0,
              0,
              0,
              0,
              0]
    
    gctsize = 1 << (gfxctl[1]&0x07)
    background = gfxctl[6] >> 4
    transparent = gfxctl[6] & 0x0f

    out.write(struct.pack('<6sHHBBBBBBBHB',
                          b'GIF89a',
                          10 + hdrlen,       # Logical screen descriptor length
                               width,         # Screen Width
                               0,             # Screen Height
                               *gfxctl,
                              gctsize,        # Global color table size
                                 8,            # LZW minimum code size
                                  1,            # Has palette
                                    bgindex,      # Background index
                                      numcolors,    # Number of colors
                                      0))           # Block terminator

    for ext in ext_hdrs:
        out.write(ext + b'\x3b')

    blocksize = max(2**bpp//8, 9)     # Minimum block size is 9
    blockmask = blocksize - 1        # Mask to get last byte address
    rowsize = (width*bpp + 7)//8     # Actual pixel data written
    datapos = out.tell() + hdrlen

    for y in range(height):
        bgcolor = palette[(background + y*gctsize)%256][0]

        pos = (datapos + (((y % 8)*rowsize) + (y//8)*blocksize) + 1) & ~blockmask
        out.seek(pos)
        
        currentpos = out.tell()
        out.write(((bgcolor >> i) << 1) | 1 for i in reversed(range(bpp)))

        if transparent!= 0:
            transparency = palette[transparent][0] ^ bgcolor

            out.seek(currentpos + 1)
            out.write(((transparency >> i) << 1) |
                      (palette[transparent][i//2] >> 7) 
                      for i in reversed(range(bpp)))
    
        out.write(bytes([0]))
    
    out.write(b';')

