#!/usr/bin/env python3

import math
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (or /dev/urandom if None).
# If `min_data_length` is not none it will be used as minimum for data length
# and max_data_length would be calculated accordingly.
def generate_random_gif(
        rng=None, min_data_length=None, max_data_length=None, out=None):

    # Default to the standard output stream if no other one was given
    if out == None:
        out = open('/dev/stdout', 'wb')

    if rng == None:
        from os import urandom

        rng = lambda n: urandom(n)

    write = lambda x: out.write(x)

    def w2p(l): return l & 0xff

    # Header
    magic = "GIF89a".encode("ascii")
    write(magic)

    # Logical screen descriptor
    width = 1 << 6 | 7 + rng(1)[0] % 5
    height = 1 << 6 | 7 + rng(1)[0] % 5
    flags = (1 << 7) | ((1 << 6) * (rng(1)[0] % 2))
    palette_size = [
        8, 4, 2, 1][math.floor((rng(1)[0] % 100) / 25)]
    background_index = rng(1)[0] & 0xe
    pixel_aspect_ratio = {
        8: 0, 4: (rng(1)[0] & 0xc), 2: (rng(1)[0] & 0xf0), 1: 0xF}[palette_size]
    disposal_method = 0
    user_input = 0
    global_color_table = 1
    sort_flag = 0
    size = 4
    write(struct.pack("<HHBHB", w2p(width),
                       w2p(height), flags, w2p(background_index),
                       w2p(pixel_aspect_ratio)))

    extensions = []

    if global_color_table == 1:
        color_table = [(1, 1, 1), (2, 2, 2), (3, 3, 3), (4, 4, 4),
                       (5, 5, 5), (6, 6, 6), (7, 7, 7), (8, 8, 8),
                       (9, 9, 9), (10, 10, 10),
                       (11, 11, 11), (12, 12, 12), (13, 13, 13), (14, 14, 14),
                       (15, 15, 15), (16, 16, 16)]
        size += len(color_table) * 3
        write(b"\xfe\x00")
        write(struct.pack("<HBBBBB", w2p(size - 7), 0xf7,
                          palette_size, sort_flag, 0, user_input))
        for i in range(len(color_table)):
            r, g, b = color_table[i]
            write(bytearray([r, g, b]))

    extensions.append({
        "label": "\xfe", "size": w2p(size - 7), "ext_start": out.tell(),
        "sub-blocks": [{"label": "", "size": w2p(user_input)}]})
    write(b"\x21\xf9")
    write(struct.pack("<BHBBBB", 4, 0, w2p(disposal_method),
                      user_input, 0))

    if min_data_length!= None:
        min_data_length = \
            2**int(math.log(min_data_length)/math.log(2)**2)*2

    if max_data_length == None and min_data_length == None:
        max_data_length = [2**3, 2**5, 2**7, 2**11,
                           2**13, 2**16, 2**19, 2**21][rng(1)[0] % 8]

    elif max_data_length == None:
        max_data_length = min_data_length*4

    elif min_data_length == None:
        min_data_length = max_data_length//4

    if min_data_length > max_data_length or min_data_length < 0:
        raise Exception("Invalid parameters.")

    if min_data_length <= 2**11:
        data_length = min_data_length

    else:
        data_length = \
            int(
                min_data_length**(
                    (max_data_length / min_data_length)**(
                        (rng(1)[0]%1000)/1000+1e-6)))

    while data_length > 2**11:
        subblock = {}
        size = 2**11
        subblock["label"] = ""
        subblock["size"] = size
        subblocks = {"": subblock}
        block_start = out.tell()
        # Prepare the size of the first buffer
        write(b"\x2c")
        write(struct.pack(">H", w2p(size)))
        subblock["ext_start"] = block_start
        subblock["sub-blocks"] = subblocks
        data_length -= size

    subblock = {}
    size = data_length
    subblock["label"] = ""
    subblock["size"] = size
    subblocks = {"": subblock}
    block_start = out.tell()
    # Prepare the size of the first buffer
    write(b"\x2c")
    write(struct.pack(">H", w2p(data_length)))
    subblock["ext_start"] = block_start
    subblock["sub-blocks"] = subblocks

    # Block end
    write(b";")

    # Write extensions
    current_extension_offset = out.tell()
    for e in extensions:
        write("\x21" + e["label"].encode("ascii"))
        write(struct.pack("<BH", e["size"], 0))
        extension_end = out.tell()
        e = e["sub-blocks"][""]
        out.seek(current_extension_offset +
                 e["ext_start"])
        write(struct.pack("<BH", e["size"], 0))
        out.seek(extension_end)
        current_extension_offset = extension_end

    print(f"Resulting file size: {(out.tell()-8)/1024:.2f}KB")
