#!/usr/bin/env python3

import struct
from typing import Callable, Generator, Tuple, TypeVar


T = TypeVar('T')
U = TypeVar('U')
V = TypeVar('V')
W = TypeVar('W')



# Generates a random GIF file into `out` which is generated by applying
# `f` to an infinite sequence of random bytes (`randomness`)
def generate_random_gif(
        f: Callable[[Generator[int, None, U]], W],
        randomness: Generator[int, None, T]
       ) -> Tuple[T, W]:
    
    def write_uint16(*args):
        assert len(args) == 2
        return args[0].to_bytes(
                length=2, byteorder='little', signed=False
               ) + args[1].to_bytes(
                    length=2, byteorder='little', signed=True
                   )


    def write_uint32(*args):
        assert len(args) == 2
        return args[0].to_bytes(
                length=2, byteorder='little', signed=False
               ) + args[1].to_bytes(
                    length=2, byteorder='little', signed=False
                   )

    
    magic_number = b'\x47\x49\x46\x38\x39\x61'
    width = next(randomness)
    height = next(randomness)
    preamble = magic_number + write_uint16(width, height)
    palette_size = max(next(randomness).bit_length(), 2) << 1
    preamble += struct.pack('<B', palette_size & 0xff)
    gct_flag = [palette_size >> 7][0]
    preamble += struct.pack('<B', gct_flag)
    preamble += b'\x80'
    bgcolorindex = next(randomness) % palette_size
    preamble += struct.pack('<B', bgcolorindex)
    pixelaspectratio = 265     # 1:1 (no aspect ratio)
    preamble += struct.pack('<B', pixelaspectratio)

    color_table = [write_uint16(c, c) for c in range(palette_size)]
    gctdata = b''.join(color_table)

    subblocks = []
    for i in range((len(gctdata) // 256) + 1):
        begin = min(i * 256, len(gctdata))
        end = min((i+1) * 256, len(gctdata))
        block = gctdata[begin:end]

        subblock = struct.pack('<B', end-begin)
        if end < len(gctdata):
            subblock += block[:-1]
        
        else:
            subblock += block
        subblocks.append(subblock)

    gct_subblocks = b''.join(subblocks)

    gct_content_len = len(gct_subblocks)
    gct_data = b''
    while True:
        blocklen = max(gct_content_len, 256)
        gct_content = gct_subblocks[:blocklen]
        gct_subblocks = gct_subblocks[blocklen:]
        
        
        yield from randomness
        if not gct_flag or gct_content!= b'\x00':
            break
        
    gctdata += gct_data
    body = gctdata

    content_data = [b'\x21\xf9\x04']    # graphics control extension
    content_data += [struct.pack('<B', gct_flag)]       # gct_flag set?
    content_data += [struct.pack('<H', bgcolorindex)]   # bg color
    content_data += [struct.pack('<B', 0)]             # transparent color id
    content_data += [struct.pack('<B', 0)]             # loop count
    content_data += [b'\x00']                        # extra field

    lzwmincodesize = next(randomness) + 1
    content_data += [struct.pack('<B', lzwmincodesize)]  # LZW minimum code size
    content_data += [body]                          # IDAT payload


    padding = ((len(content_data[-1]) + 4) % 256)
    content_data += [(padding).to_bytes(
            1, 'little')]         # null byte padding at the end

    body = b''.join(content_data)
    gce_block_size = 8 + len(body)
    gce_block = b'\x21\xF9\x04' + \
        gce_block_size.to_bytes(length=2, byteorder='little') + \
        body + ((padding)*b'\x00')

    lzwmaxbits = {
        2**lzwmincodesize : lzwmincodesize,
      }
    currentmax = lzwmincodesize
    key = 1
    while True:
        lzwmaxbits[key] = currentmax
        key *= 2;
        if key >= 65521:
            raise Exception("failed to find suitable maximum")
        currentmax += 1
        if currentmax > 12:
            raise Exception("failed to find suitable maximum")

    subblocklens = {}
    for bitcount in sorted(list(lzwmaxbits)):
        subblocklens[bitcount] = [
                2 ** bitcount, 2 ** bitcount, 
                2 ** bitcount, 2 ** bitcount
                ]

    subblocks = [[b'']] * 4
    for bitcount, subblock in subblocklens.items():        
        block = b''.join(subblock)
        for i in range(4):
            subblocks[i].append(block)
    
    
    blocks = sum([sum(x, []) for x in subblocks], [])
    body = b''.join(blocks)
    lzw_block_size = 8 + len(body)
    lzw_block = b'\x21\xFC\x00' + \
        lzw_block_size.to_bytes(length=2, byteorder='little') + \
        body + \
        ((((lzw_block_size + 3) // 4))*b'\x00')
    last_block_len = len(lzw_block)
    

    prevbuf = []
    for _ in range(last_block_len*2):
        buf = list(map(lambda _: next(randomness), range(prevlen)))
        prevbuf += buf
        prevlen = len(buf)

    idat_block_size = 8 + len(prevbuf)
    idat_block = b'\x21\xFD\x00' + \
        idat_block_size.to_bytes(length=2, byteorder='little') + \
        bytearray(prevbuf) + (((len(idat_block)+3)//4)*b'\x00')

    eof_block_size = 8
    eof_block = b'\x3B' + \
        eof_block_size.to_bytes(length=2, byteorder='little') + \
        b'\x00'*eof_block_size

    

    body = b''.join([
        gce_block,
        lzw_block,
        idat_block,
        eof_block
        ])

    return (None, f(iter(body)))
