#!/usr/bin/env python3

"""Random GIF Generator

This script generates a random GIF file which can be used for testing
purposes or simply fun :).
If no arguments are passed then the generated GIF will have the default name
"random.gif". Otherwise you can decide what name you want your new GIF file
to have by passing that as the first argument.
"""

import random
import argparse
import os
import sys
import struct

RNG = '/dev/urandom'           # source of entropy
DEFAULT_OUTPUT = 'random.gif'
DEFAULT_WIDTH = 50             # dimensions must be positive integers
DEFAULT_HEIGHT = 50            # larger images might take long time to render!


class RandomGifGenerator():

    def __init__(self, output=None, width=None, height=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output

        if not width:
            self.width = DEFAULT_WIDTH
        elif int(width) <= 0:
            self.width = DEFAULT_WIDTH
        else:
            self.width = int(width)

        if not height:
            self.height = DEFAULT_HEIGHT
        elif int(height) <= 0:
            self.height = DEFAULT_HEIGHT
        else:
            self.height = int(height)


    def generate_random_gif(self):
        
        with open(RNG, mode='rb') as rng,\
             open(self.output,    mode='wb') as out:

            return generate_random_gif(
                    rng, out, self.width, self.height
                )



# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# correspond to the information found here:
# https://www.w3.org/Graphics/GIF/spec-gif89a.txt
# Parameters:
#   rng -- binary input stream for random data
#   out -- binary output stream for writing our GIF file
#   width -- desired width of the resulting image
#   height -- desired height of the resulting image
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, 
        width: int, height: int
    ):

    # Signature
    header = b'GIF89a'
    
    # Width and Height
    w = struct.pack('<H', width)               # little endian
    h = struct.pack('>H', height)              # big endian
    header += w + h                            # concatenate them

    # Color Table size
    color_table_size = None
    while True:
        remaining_bits = 7 & random.getrandbits(3)
        color_table_size = remaining_bits + 1
        bytes_needed = 3 * pow(2,color_table_size+1) + 1
        if len(bytes_needed) == 1:
            break

    print("generating color table...")
    colors = []
    for i in range(pow(2,color_table_size)):
        r = g = b = 0
        bits_left = 8
        k = random.getrandbits(3)
        for j in range(k):
            r |= ((random.getrandbits(1) << bits_left)) 
            g |= ((random.getrandbits(1) << bits_left))
            b |= ((random.getrandbits(1) << bits_left)) 
            bits_left -= 1
        colors.append((r,g,b))
        header += struct.pack('<BBB', r, g, b)
    # Header is now complete. Write it into the output buffer
    out.write(header)

    # Block Size (Data Sub-Block within Data Stream)
    bs = random.randint(4,128)                 # block size between 4 and 128
    subblock = bytearray()                      # initialize empty byte array
    subblock.extend(struct.pack("<B",bs))      # store block size on its own
    subblock[0] ^= 0x80                         # make sure block terminator set
    out.write(subblock)                        # write it to the output buffer

    # Image Descriptor
    print("generating image descriptor")
    id = bytearray()
    id.extend(struct.pack('<BBHHHB',
        0x2C,                                  # separator
        0,                                     # left position of image
        0,                                     # top position of image
        width,                                 # width of image
        height,                                # height of image
        0                                      # no local color table
    ))
    id[0] ^= 0x80                              # make sure image descriptor
    out.write(id)                               # write canvas to output buffer

    # Local Colour Table
    lct = []                                   # local colour table
    for i in range(pow(2,remaining_bits)+1):  
        lct.append(colors[i])                    # take next color from list
        out.write(struct.pack('<BBB',*lct[-1]))  # write it to the output buffer
    lct_length = len(lct)                       # save how many colours there 
    print("generated {} entries for LCT.".format(len(lct)))  

    # Get the length of the compressed image data
    print("computing total decompressed image length...")
    decompressed_image_length = 0
    nx = ny = cx = cy = 0                   # current x and y coordinates
    lx = ly = clen = 0                     # last x and y coordinates
    while True:

        # These are the "subblocks" for the image data
        # (for more info see spec section 12)
        if random.choice((True,False)):                             # EOI
            subblock = bytearray()
            subblock.extend(struct.pack(">B",0x00))                # EOI
        else:                                                         # Graphic Raster Info
            nxt = bool(random.getrandbits(1))                      # NXT
            subblock = bytearray()
            if nxt:                                                   # get code
                cx &= ~1                                              # clear lowest bit
                cx |= random.getrandbits(1)                          # choose X
            else:                                                    # use previous value 
                cx = lx                                               # copy last coordinate
            if nxt:                                                   # get code
                cy &= ~1                                              # clear lowest bit
                cy |= random.getrandbits(1)                          # choose Y
            else:                                                    # use previous value  
                cy = ly                                               # copy last coordinate
            subblock.extend(struct.pack(">BBBB",cx,cy,nxt,nxt))     # concatinate code
            lx = cx                                                  # remember last sequence
            ly = cy
            while not nxt:                                           # repeat until end of line
                cx &= ~1                                              # clear lowest bit
                cx |= random.getrandbits(1)                          # choose X
                cy &= ~1                                              # clear lowest bit
                cy |= random.getrandbits(1)                          # choose Y
                subblock.extend(struct.pack(">BB",cx,cy))            # concatinate code
                nxt = bool(random.getrandbits(1))                    # get next code
                lx = cx                                              # remember last sequence
                ly = cy
            subblock[0] ^= 0x80                                       # set bit to indicate graphic data

        out.write(subblock)                                          # write subblock to output file
        subblock_length = len(subblock)                              # get its length
        decompressed_image_length += subblock_length                  # add up overall length

        # Switching to a new frame starts when we hit the end
        # of this frame's decompression
        if nxt:                                                       # if nxt set, start a new frame
            break                                                      # exit loop and get another frame
        
    print("done computing decompression length.")
    print("{} bytes after compression.".format(decompressed_image_length))

    # Now we can fill the rest of the output buffer with zeros
    # so we don't run off the end of the file accidentally
    padding = bytearray()
    padding.extend(b'\x00'*(699-(decompressed_image_length%699)))   # pad to multiple of 699
    out.write(padding)

    # Finally read the rest of the uncompressed image data
    print("reading pixels from /dev/urandom:")
    pixel_data = b''
    chunk_size = 10**6                                            # read chunk by chunk
    num_chunks = decompressed_image_length//chunk_size + 1          # compute number of chunks needed
    for chunk in range(num_chunks):
        if chunk < num_chunks-1:                                    # not yet done reading?
            pixel_data += rng.read(chunk_size)                      # read at most 1 million bytes at once
        else:                                                       # last chunk?
            adjusted_chunk_size = decompressed_image_length % chunk_size   # compute remaining bytes to read
            pixel_data += rng.read(adjusted_chunk_size)             # read whatever is left over to finish
        print("\t{}/{}...".format(chunk,num_chunks),end='\r')       # show progress

    print("")                                                        # newline
    out.write(pixel_data)                                            # write the final chunk to our output buf
