#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# given pseudorandom number generator in binary format
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Header
    head_ver  = b'\x47\x49\x46'      # Signature and version string
    width     = rng.read(2)         # Screen size
    height    = rng.read(2)
    color_map = b'\xff\xf9\x0b'      # Grayscale palette - start index (1), length (15)
    pixel_as  = b'\x00'              # Pixel aspect ratio
    gct_flag  = b'\x80'              # Global colour table flag
    gct_size  = bytes([rng.randrange(2**3)])   # GCT size

    for i in range(int.from_bytes(gct_size)):
        r = bytes([rng.randrange(2**6)])
        g = bytes([rng.randrange(2**6)])
        b = bytes([rng.randrange(2**6)])
        gct_data += r + g + b
    
    c = {
        1 : '255',
        2 : '255',
        3 : '255',
        4 : '255',
        5 : '255',
        6 : '255',
        7 : '255',
        8 : '255',
        9 : '255',
        10: '255',
        11: '255',
        12: '255',
        13: '255',
        14: '255',
        15: '255'
    }

    # Logical screen descriptor
    lsd       = head_ver             \
             + width                \
             + height               \
             + color_map            \
             + pixel_as             \
             + gct_flag             \
             + gct_size             \
             + gct_data

    # Actual image descriptor
    img_width   = height
    img_height  = width
    img_left    = b'\x00\x00'        # Image offset from left edge of logical screen
    img_top     = b'\x00\x00'        # Image offset from top edge of logical screen
    field_color = b'\x00\x00'        # Background colour index
    packed      = b'\x08'            # Packed field with LZW compression flag on
    img_lzw     = b'\xff\xff'        # Size of LZW compressed subblock
    img_size    = img_width          \
                + img_height         \
                + img_left           \
                + img_top            \
                + field_color        \
                + packed             \
                + img_lzw


    # Graphics control extension block
    gce_label      = b'\x21\xf9\x04'   # Extension label
    gce_header     = b'\x0b'           # Block size
    gce_transp     = b'\x00'           # Transparency index
    gce_usertrans  = b'\x00'           # User defined transparent
    gce_disposal   = b'\x01'           # Disposal method
    gce_delaytime  = b'\x0b'           # Delay time in 1/100 sec
    gce_transparent= b'\x00'#000000ff'    # transparent colour
    gce_terminator = b'\x00'#000000ff'    # Block terminator

    gce_extention = gce_label      \
                 + gce_header     \
                 + gce_transp     \
                 + gce_usertrans  \
                 + gce_disposal   \
                 + gce_delaytime  \
                 + gce_transparent\
                 + gce_terminator

    # Comment extension block
    com_label = b'\x21\xfe'          # Extension label
    com_text  = b"It's not rocket science!"

    comment = com_label + len(com_text).to_bytes(1, byteorder="big") + com_text

    # Plain text extension block
    txt_label  = b'\x21\x01'         # Extension label
    txt_header = b'\x0d'            # Block size
    txt_xpos   = b'\x00\x00'        # Text position x-coordinate
    txt_ypos   = b'\x00\x00'        # Text position y-coordinate
    txt_grid   = b'\x00\x00'        # Grid cell size
    txt_fgcol  = b'\xff\x00\x00'    # Foreground color
    txt_bgcol  = b'\xff\x00\x00'    # Background color
    txt_text   = b'Rocket science!'

    plain_text = b''
    plain_text += txt_label   
    plain_text += txt_header  
    plain_text += txt_xpos    
    plain_text += txt_ypos    
    plain_text += txt_grid    
    plain_text += txt_fgcol   
    plain_text += txt_bgcol   
    plain_text += txt_text    

    # Application extention block
    app_label  = b'\x21\xff'         # Extension label
    app_header = b'\x0b'            # Block size
    app_id     = b'NETSCAPE2.0'     # Application ID
    app_proto  = b'\x03'            # Application auth code
    app_flag   = b'\x01'            # Application flags
    app_loops  = bytes([rng.read(1)[0] | 0x80])  # Loop counter
    app_coderun= bytes([rng.read(1)[0]])  # Unused byte
    app_end    = b'\x00'#000000ff      # Ending byte sequence

    application = app_label      \
               + app_header     \
               + app_id         \
               + app_proto      \
               + app_flag       \
               + app_loops      \
               + app_coderun    \
               + app_end

    # Actual image data
    rows  = []                    # List to hold individual row buffers
    buffs = {}                   # Dictionary that holds the buffer metadata
    frame = bytes()              # Buffer accumulator

    for _ in range(int.from_bytes(img_width)):

        bg_colind = rng.read(1)[0] & 0x0f
        fg_colind = (rng.read(1)[0] & 0x0f ) | 0x10
        
        for _ in range((rng.read(1)[0]&0x0F)):

            if rng.read(1)[0]:

                if rng.read(1)[0]:

                    pixbuf = bytes([(3*(bg_colind&0x0F))|128])
                    buffs[len(buffs)] = {'size':1,'type':'raw'}
                    
                else:

                    pixbuf = bytes([(3*(fg_colind&0x0F))])
                    buffs[len(buffs)] = {'size':1,'type':'raw'}
            
            else:
                
                num_rep = rng.read(1)[0]%3+1
                pixbuf  = bytes([(num_rep|(3*(fg_colind&0x0F)))]) 
                buffs[len(buffs)] = {'size':num_rep,'type':'repeat'}
                
            frame += pixbuf

        rows += [frame]
        buffs[len(buffs)] = {'size':len(frame),'type':'row'}
        frame = bytes()

    gif_data = b''

    for i in sorted(buffs):

        size = buffs[i]['size'] 
        type = buffs[i]['type'] 

        if type == 'raw':

            while True:

                if size > 255:

                    chunk = bytes([255])+rows[i][:255].ljust(255,b'\x00')
                    del rows[i][:255]
                    size -= 255

                elif size <= 255:
                    
                    chunk = bytes([size])+rows[i][:size].ljust(size,b'\x00')
                    del rows[i][:size]
                    break

                gif_data += chunk

        elif type == 'row':

            gif_data += rows[i]

        elif type =='repeat':

            gif_data+= rows[i][:-1]+bytes([rows[i][-1]]*size)
            

    # Footer 
    foot_subhead = b'\x3b'  # GIF trailer 
    foot_end     = b'\x00'#000000ff
    end          = foot_subhead + foot_end

    gif_data += application 
    gif_data += plain_text 

    gif_file = lsd + img_size + gif_data + gce_extention + comment +