#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (any object that 
# implements bytes() interface)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    
    # magic numbers and "version" (always 87a)
    out.write(b'\x47\x49\x46\x38\x37\x61')

    # width as two bytes in LSB order
    for i in range(2):
        out.write((rng.read(1))[0])

    # height as two bytes in LSB order
    for i in range(2):
        out.write((rng.read(1))[0])

    # flags
    for flag in [ord('L'), ord('A'), ord('S'), ord(',')]:
        out.write(flag.to_bytes(1, byteorder="little"))

    # background color index
    out.write(((rng.read(1))[0] % 256).to_bytes(
        1, byteorder='little'))

    # pixel aspect ratio 
    # (default to no pixel aspect ratio information)
    out.write(b'\x00')

    # giff comment extension
    length = rng.read(1)
    if len(length) == 0: 
        raise Exception("Empty Comment")
    if len(length) > 255:
        raise Exception("Comment too long")
    out.write(length[0])
    comment = rng.read(int(length))
    if len(comment)!= int(length):
        raise Exception("Unable to read enough comment")
    out.write(comment)

    # application extension
    app = (
            b"\x21\xFF\x0B\x47\x49\x46\x38\x37\x61" + 
            b"\x00\x00\x00\x00\x21\xF9\x04\x05\x00\x05" + 
            b"\x00\x00\x00\x2C\x00\x00\x00\x00\x01\x00" + 
            b"\x01\x00\x00\x02\x02\x44\x01\x00\x3B"
          ).ljust(100, '\x00')
    print(len(app))
    out.write(len(app).to_bytes(1, byteorder="little"))
    out.write(app)

    # graphic control extension
    gc = (
            b"\x21\xF9\x04\x05\x00\x00\x00\x00\x2C" + 
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02" + 
            b"\x02\x44\x01\x00\x3B"
          ).ljust(100, '\x00')
    out.write(len(gc).to_bytes(1, byteorder="little"))
    out.write(gc)

    # Image Descriptor block
    idb = (
            b"\x2C\x00\x00\x00\x00\x01\x00\x01\x00\x00" + 
            b"\x02\x02\x44\x01\x00\x3B\x00"
          ).ljust(100, '\x00')
    out.write(idb)


    datalen = int.from_bytes(
        rng.read(1), byteorder='little')[0] * 1024 

    data = []
    while len(data) < datalen:
        data += rng.read(1024)

    out.write(data[:datalen])

