#!/usr/bin/env python3

import struct
from os import urandom
from sys import stderr
from typing import BinaryIO


# Same as generate_random_gif but without touching /dev/urandom
def generate_pseudo_random_gif(
        seed: int, datalen: int, out: BinaryIO, verbose=False
    ) -> None:

    # Use this instead of os.urandom to make sure we can reproduce results later
    rng = urandom if seed == -1 else iter(struct.pack('>I', seed))

    def writechunk(*args):
        if verbose:
            print(' '.join(hex(x)[2:] for x in args), file=stderr)

        chunkdata = struct.pack(*args)
        out.write(b''.join([
            (len(chunkdata)).to_bytes(1, byteorder='little'),
            chunkdata ]))


    writechunk(0x56)           # version

    width = 8192 // height     # optimal size
    assert datalen % width == 0

    # The logical screen descriptor block is 7 bytes long:
    #   - 3 bytes: always 'GIF87a' or 'GIF89a' depending on the version
    #   - 3 bytes: screen width and height
    #   - one byte: color depth, must be 8
    #   - one byte: global color table flag. If set, there's a local colour map
    ssd = [ 0x47, 0x49, 0x46,
             width & 0xff, (width >> 8) & 0xff, height & 0xff, (height >> 8) & 0xff,
             8,              # bit per pixel
             1               # with local color table
           ]                                                                             # noqa E127
    writechunk(*ssd)

    # Next comes the global color table, which contains all possible colours that are
    # actually used anywhere within this frame. We're just going to pretend there is no
    # global color table here by putting an empty palette.
    palletesize = 1 << 8
    writechunk(palletesize * 3)
    writechunk(*(iter([])))    # empty palette

    # The next section is how many frames this animation has. Since we don't have any
    # animations yet, this should always be 1.
    writechunk(1)

    # This section describes the two images contained in the first frame
    #
    # It starts with a single block describing each image
    nimages = 2
    lzwmincodesize = 2
    writechunk(nimages*2, lzwmincodesize)

    # Then it has one extra length byte before each following sub-block containing its own payloads
    imgdesc = [
        0,         # left position
        0,         # top position
        width,     # image width
        height,    # image height
        0          # flags, none specified so this whole byte is zero
    ]
    for _ in range(nimages):
        writechunk(*(imgdesc + [lzwmincodesize]))

    # Finally it has the blocks that contain the actual uncompressed image data. We're going
    # to put two identical ones now.
    for i in range(nimages):
        lzwmincodesize = 2
        rgbdata = [(i%datalen) ^ ord(next(rng)) for j in range(datalen)]
        writechunk(
                len(rgbdata).to_bytes(1, byteorder='little') +      # LZW min code size
                len(rgbdata).to_bytes(1, byteorder='little'))       # Sub-block length

        compresseddata = compress(rgbdata, lzwmincodesize, verbose=verbose)

        writechunk(compresseddata[1:])                          # skip first byte because we know what it is already

        # Note that we only need to specify the last byte of these lengths once since they will
        # never change anymore after this point. However, it doesn't hurt to do so and makes
        # our output more predictable.
