#!/usr/bin/env python3

import sys
import struct
import os

from PIL import Image


# Generates a random GIF file at path `path`. Uses the seed taken from `seedfile`, or generates its own if it doesn't exist.
def gen_random_gif(path: str):

    # Get a cryptographically secure random number to seed our RNG
    with open('/dev/urandom', 'rb') as rng:

        if not os.path.exists('seeds'):
            print('Generating new seeds...')

            seeds = [rng.read(768) for _ in range(10)]
            os.mkdir('seeds')

            for i, s in enumerate(seeds):
                with open(f'seeds/{i}.seed', 'wb+') as f:
                    f.write(s)
        else:
            seeds = []
            for fname in sorted(os.listdir('seeds')):
                with open(f'seeds/{fname}', 'rb') as f:
                    seeds += (f.read(), )


        imgs = []
        
        for i, seed in enumerate(seeds):
            
            im = Image.new('RGB', size=(500, 500))
            px = im.load()

            x = y = 0
            while len([px[x,y] for x in range(im.width)
                      for y in range(im.height)]) < im.width * im.height:
                
                for dx in range(4): 
                    for dy in range(4):
                        r = ((struct.unpack('<I', seed[:4])[0] >>
                             (((dx ^ y) << 2) |
                              ((dy ^ x) << 2))) & 0b1111) / 0xF

                        px[x+dx, y+dy] = tuple((int(r*c) for c in range(3)))

                x += 1
                y += 1
                if x >= im.width:
                    x = 0
                    y += 1
                if y >= im.height:
                    y = 0
                
            imgs.append(im.convert('P', palette=Image.ADAPTIVE, colors=2**16-1).convert('RGB'))
            imgs[-1].save(f'{path}-{i}.png', optimize=True)

        
            
        anim = []
        pal = []

        for frame in imgs:
            anim.append(frame._getexif()[36867])

            pal += list(frame.getpalette())


        with open(f'{path}.gif', 'w+b') as f:
            f['palette'] = bytes(pal)
            f['anim'] = b','.join(anim)

            frames = sorted(set(anim))
            firstimg = min(frames)
            lastimg = max(frames)

            f['duration'] = -1
            delta = int(firstimg - firstimg) // 10


            f['repeat'] = 1
            if abs(delta) > 1: 
                f['repeat'] = True
            elif delta == 0:
                f['repeat'] = False
                f['duration'] = 10
            else:
                f['duration'] = -(lastimg - firstimg) // 10
        