#!/usr/bin/env python3

import struct
import os
from typing import BinaryIO


HEADER_LEN = len('GIF89a')
GLOBAL_COLOR_TABLE_OFFSET = HEADER_LEN+3


def write_header(
        writer: BinaryIO,
        width: int=0x200, height: int=0x100):
    writer.write(b'GIF89a')
    writer.write(struct.pack('<HHB', width, height, 0xf7))
    return GLOBAL_COLOR_TABLE_OFFSET


def write_lsd(writer: BinaryIO):
    pass


def write_gct(writer: BinaryIO):
    pass


def read_section(reader: BinaryIO, section_length: int):
    print("Reading section", reader.tell(), "of length", section_length)
    if section_length == 0:
        raise EOFError()
    
    data = reader.read(section_length)
    print("Read bytes:", data)
    return data



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Write header
    offset = write_header(out)

    while True:

        secthdr = rng.read(1)
        if secthdr is None or secthdr[0] in (0x3b, 0x21):

            # This is an extension block or terminator block
            print ("Encountered extension or terminator")
            
            try:
                ext_type = ord(secthdr[0])
            except TypeError:
                ext_type = secthdr[0]

            # Skip blocks with invalid types
            if not 0x00 <= ext_type < 0xff: break


            # Get the section length
            ext_len = read_section(
                    rng,
                    struct.unpack('<BB', rng.read(2))[0]+1)[-1]
            print ("Extension type:", ext_type)
            
            # Handle special cases for the application extension
            if ext_type == 0xfe:

                # Application extensions are followed by a
                # subblock containing its name (in plain text)
                subblocks = []
                while ext_len > 0:

                    subblock_len = read_section(
                            rng,
                            struct.unpack('<BB', rng.read(2))[0]+1)[-1]
                    
                    subblocks.append(
                        read_section(rng,subblock_len))
                    ext_len -= subblock_len
                
                appname = ''.join([chr(c) for c in subblocks])
                print ("Application name:", appname)
                
            else:
                
                # Other kinds of extensions are just skipped entirely
                while ext_len > 0: 
                    ext_len -= read_section(rng,ext_len)[-1]
                    


        elif secthdr[0] & 0xf0 == 0xc0:
        
            # The first byte of any graphic control label must be zero
            assert not secthdr[0]

            # Read graphics control block contents
            gctrlbl = [ord(i) for i in read_section(rng,8)]
            print ("Graphics control block:", gctrlbl)
        
        else:
    
            # We've encountered an unrecognized part of the GIF format!
            print ("Unrecognized section", secthdr, reader.tell())


    # Write the trailing 'trailer' marker
    out.write(b'\x3b')