#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_WIDTH      = 500
DEFAULT_HEIGHT     = 500
DEFAULT_COLORDEPTH = 4  # Bits per pixel / byte
DEFAULT_OUTPUT     = 'random.gif'


class RandomGifGenerator():

    def __init__(self,
                 width       = DEFAULT_WIDTH, 
                 height      = DEFAULT_HEIGHT,
                 colordepth  = DEFAULT_COLORDEPTH, 
                 output      = None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output

        self._width       = width 
        self._height      = height
        self._colordepth  = colordepth


    @property
    def width(self):
        return self._width


    @property
    def height(self):
        return self._height

    
    @property
    def colordepth(self):
        return self._colordepth
    

    @colordepth.setter
    def colordepth(self, value):
        """Ensures that the new colordepth can be written to 1 byte."""
        if isinstance(value, int):
            pass
        elif isinstance(value, float):
            value = round(value)
        elif isinstance(value, str):
            try: 
                value = int(value)
            except ValueError:
                raise TypeError('Could not cast {} ({}) to integer.'.format(
                        repr(value), type(value))) from ValueError
        else:
            raise TypeError('Could not cast {} ({}) to integer.'.format(
                    repr(value), type(value))) from TypeError

        if not 1 <= value < pow(2, 8):
            msg = "Color depth must be between [1..2^8)" \
                  ", however was set to '{}'.".format(value)
            raise ValueError(msg)
        else:
            self._colordepth = value


    def generate_random_gif(self):
        
        with open('/dev/urandom', mode='rb') as rng,\
             open(self.output,    mode='wb') as out:
            
            return generate_random(
                rng, out,
                self.width, self.height, self.colordepth
            )



# Generate the entire GIF file given by the arguments
# TODO: Implement optional progress report
def generate_random(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int, colordepth: int
    ):

    check_dimensions(width, height)
    check_colordepth(colordepth)

    # Signature
    header = b'GIF89a'

    # Width and Height
    w = convert_to_bytes(width,           length=2)   # width is 16 bit
    h = convert_to_bytes(height,          length=2)   # height is 16 bit
    header += w + h

    # Color Table size
    ctsz = colordepth * 3 
    for _ in range(ctsz):
        header += bytes([rng.read(1)[0]])

    # Block Size (Data Sub-Block within Data Stream)
    bs = rng.read(1)[0]                                         # block size is an unsigned char
    subblock = rng.read(bs - 1)                                  # discard sub-block marker
    header += bytes([bs])                                        # add block size back again
    pixels = convert_to_bytes(colordepth, count=(width*height)) # create a list of pixels
    data = bytes(pixels[i % len(pixels)] for i in range((bs - 1)*8)) + subblock # create image data stream
    out.write(header + data)                                     # write it all together
    
