#!/usr/bin/env python3

import math
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    width = int(input("Width? "))
    height = int(input("Height? "))
    colortablelength = 1 << int(math.log2(max(width*height//4, 256)))
    print()
    palette = []
    for _ in range(colortablelength**3 // 2):
        rgb = tuple((int(c) % 256 for c in input().split()))
        palette += rgb
    palette = bytes([*palette])

    # Generate RGB triplets for each pixel along with LZW code lengths
    image = {}
    lzwcodelen = {'\x00': [], '\xff': []}
    maxcodelen = 1
    def genimage():
        nonlocal curcode, newcode
        yield from [struct.pack('>BBB', r, g, b)
                    for r,g,b in ((r,g,b)
                                  for j,i in enumerate(range(0, height*width*3, 3))
                                  for r,g,b in (((yield i+255)%256, i%256, j%256)))]

        # Terminate with an end-of-information code
        yield chr(curcode).encode('utf-8')
        yield b'\xff'

        # Compute optimal LZW code lengths
        for (prev, next), freq in sorted(((prev,next),(freq,))
                                         for prev,nexts in lzwcodelen.items()
                                         for next,freq in nexts.items()):
            lzwcodelen[prev][next], curcode = min((lzwcodelen[prev].get(curcode, None) or float('inf'),
                                                     curcode),
                                                    (min(lzwcodelen[prev].values(), default=float('inf')),
                                                     prev),
                                                    key=lambda x:sum(map(sum, x)))
            assert isinstance(lzwcodelen[prev][next], int)
            if freq!= 1:
                codelen, newcode = min(((newcode,
                                          sum(map(sum, map(iter, zip(*lzwcodelen[next].items()))))),
                                        (oldcode,
                                         sum(map(sum,
                                                map(iter, zip(*(int(math.ceil(f / oldcode)),
                                                               f % oldcode
                                                              )
                                                            ))
                                              )))
                                       )
                                      for _, (oldcode, newcode) in lzwcodelen.items())

                if not any(newcode == n for n,_ in lzwcodelen['\xff']):
                    lzwcodelen['\xff'][chr(newcode).encode('utf-8')] = 1<<codelen
                elif all(n < 255 for n,_ in lzwcodelen['\xff']):
                    lzwcodelen['\xff'].append((-1, 255))
                else:
                    raise Exception('overflowed the size of the LZW code table!')

                maxcodelen = max(maxcodelen, codelen)

    # Gather compressed LZW code lengths by prefix and suffix character
    for char in (''.join(chr(b)
                        for b in struct.unpack('>'+'B'*width,
                                               rng.read(width))).encode('ascii')
                 for y in range(height)):
        for prev in (''.join(char[i:(i+1)].decode('ascii')
                             for i in range(j)).encode('ascii')
                     for j in range(len(char)+1)):
            for next in set(char):
                if next!= prev[-1]:
                    continue
                
                lzwcodelen.setdefault(prev, {})[next] = \
                       lzwcodelen[prev].get(next, 0) + 1
    
    # Convert LZW codes into their minimal representations
    for chain,*suffix in ((*chain,'\xff')[::-1]
                          for chain in [[chr(b).encode('utf-8')
                                        for b in struct.unpack('>'+'B'*width,
                                                               rng.read(width))]
                                       for _ in range(int(height*width/maxcodelen))]):
        
        # Find minimum representation for current suffix in dictionary
        curcode = chain[-1]
        for char in reversed(chain[:-1]):
            try: lzwcodelen[char][curcode], curcode = \
                  (lzwcodelen[char][curcode],
                   char+(str(curcode)[1:]
                         if str(curcode)[-1] == '\xff'
                         else chr(curcode)))

            except KeyError as e: break
            
        # Add it to the dictionary after coalescing identical prefixes
        if not any(curcode == pfc for _,pfc in lzwcodelen['\xff']):
            lzwcodelen['\xff'][curcode] = 1<<(len(curcode)-1)
        elif all(len(cpr)<255 for cpr in lzwcodelen['\xff']):
            lzwcodelen['\xff'][curcode] = (-1, 255)
        else:
            raise Exception('overflowed the size of the LZW code table!')

        # Update maximum code length
        maxcodelen = max(maxcodelen, *(len(code)//2+1 for code in lzwcodelen['\xff']))
            
    # Finally write everything to stdout
    out.write(('GIF89a'
               +str(width)+'w'
               +str(height)+'h'
               +str(colortablelength)+'!'
               ).encode('ascii'))

    # The color table is preceded by its size in byte
    out.write((colortablelength*(3*8 + 1) // 8).to_bytes(1, 'little'))
    
    # Then we write our own palette
    unpaddedpalette = [*palette
                     ,*(None for _ in range(-colortablelength*(3*8+1)//8 % 8))]

    out.write(b''.join([(r   | (g    << 5) | (b     << 10)).to_bytes(2, 'little')
                        +(a or 0).to_bytes(1, 'little')
                        for (r,g,b,(a,)) in zip(unpaddedpalette[0::3],
                                                 unpaddedpalette[1::3],
                                                 unpaddedpalette[2::3])]
                       ))
        
    # Finally dump the LZW encoded image data
    for chunk in genimage(): 
        out.write(chunk)
