#!/usr/bin/env python3

import os
import sys
from typing import BinaryIO

import click


@click.command()
@click.argument('outfile', type=str) # TODO: support stdin as outfile?
@click.option('-l', '--length', 'length', default=-1, help='How long should it be?')
def main(**kwargs):
    """Generate a random GIF."""

    with open('/dev/urandom', 'rb') as f:
        generate_random_gif(f, kwargs['outfile'], length=kwargs['length'])


# Generates a random GIF file into `out` using the
# random number generator from `rng`.
def generate_random_gif(
        rng: BinaryIO, out: str, *, length: int
    ):
    if not isinstance(length, (int, float)):
        raise TypeError("length must be an integer")

    # Check that we don't overflow when multiplying by 16
    max_length = (1<<32)-1 * (16//7) // 8 - len(b"GIF89a")

    if -1 < length <= max_length:
        pass
    elif length == -1 and os.path.exists(out):
        # Just read it to check its size
        with open(out, "rb") as f:
            f.seek(-max_length-len(b"GIF89a"), os.SEEK_END)
            actual_length = f.tell()

        if actual_length!= -1+max_length:
            print(actual_length, file=sys.stderr)
        else:
            print("Length of existing output is too short.", file=sys.stderr)

        return

        #raise ValueError("Length of existing output is too short.")
    else:
        raise ValueError("Invalid length value", length)

    # The maximum length, in bytes, minus the GIF's header
    if length <= 0:
        datalen = max_length
    else:
        datalen = min(max_length, length*(16//7))

    # Write the header
    with open(out, "wb") as f:
        f.write(b"GIF89a")

        while True:

            if datalen >= 2**24:
                blocksize = 2**24

                # This will always be true since we have handled
                # all possible cases above
                assert datalen % blocksize == 0, datalen % blocksize

            else:
                blocksize = datalen

            try:
                data = rng.read(blocksize)
            except Exception as e:
                if datalen < 2**24:
                    raise AssertionError(
                        "Could only read {}/{} bytes.".format(
                            datalen, max_length), file=sys.stderr)
                else:
                    raise RuntimeError("""\
There was an unexpected error reading from /dev/urandom; please \
file a bug report including the traceback below!""", e, file=sys.stderr)

            f.write(data)

            datalen -= len(data)

            # Break on end of the file or after writing enough data
            if not data or datalen == 0:
                break