#!/usr/bin/env python3

import sys
import struct
from typing import Optional

# Generates a random GIF file with an invalid screen descriptor
# into `out`. If `size` is not None, it should be a tuple (width, height).
# Otherwise, the screen dimensions will come from /dev/urandom.
def generate_invalid_screen_descriptor_gif(
        seed: bytes, size: Optional[tuple] = None, out: BinaryIO = sys.stdout.buffer
    ):

    header = b'GIF89a'
    if size is None:
        width, height = struct.unpack('<HH', seed[:4])
    else:
        width, height = map(int, size)


    # Make sure that width * height < 256**2, otherwise we're going to overflow when
    # calculating the palette
    assert width*height<256**2

    paldata = [0]*((width+height)*3)
    for i in range(len(paldata)):
        print(i)
        rng = struct.unpack('B', bytes([seed[i % len(seed)]]))[0]
        paldata[i] = rng%256

    paldata = bytearray(paldata)

    # Logical screen descriptor format:
    #     Bytes   Description        Value
    #     -----------------------------------
    #     0-1     Screen width       Unsigned 16 bit integer
    #     2-3     Screen height      Unsigned 16 bit integer
    #     4       Packed field: Global Color Table Flag    Bit 0 1=global colors have been
    #                                                                defined, 0=not
    #     5       Color Resolution   Unsigned 8 bit integer
    #                                      Number of bits per primary color
    #                 Valid values: 1, 2, 4, 8
    #     6       Sort Flag          Bit 0 1=global color table sorted by importance, 0=unsorted
    #     7       Size of global colortable    Unsigned 8 bit integer
    #                                         2^(value+1)
    sdesc_packed = width & 0x00FF | ((height & 0x00FF)<<8)
    sdesc = struct.pack('<BHBBBBB',
                        sdesc_packed,
                        0,                # Background color index
                        7,                # Pixel aspect ratio
                        0,                # GCTF == 0 -> no global colors
                        0,                # Colour res
                        0,                # Sort flag
                        len(paldata)-1,   # Length of global color table
                       )

    trailer = struct.pack('<BBB',
                               0x3B,         # Trailer block identifier
                               ord(';'),     # Block terminator
                               0             # No comment
                           )

    frame_table = struct.pack('<cBBHBHBHBHBBBBBBBBBBBBBBBBB',
                                b',',
                                    0b11000000,
                                    0,
                                    0,
                                    width,
                                    height,
                                    width//256,
                                    width&0xFF,
                                    height//256,
                                    height&0xFF,
                                    0b11000000,
                                    0,
                                    0,
                                    7,
                                    0,
                                    0,
                                    7,
                                    0,
                                    0,
                                   )

    data = b''

    for y in range(height):
        line = []
        for x in range(width):
            line += [0b11000000|((y&(1<<7))>>(7-x))]

        data += bytes(line)

    out.write(header + sdesc + paldata + frame_table + data + trailer)

