#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO
from random import Random

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    
    def _generate_extension() -> bytes:
        exttype = rng.randint(0, 14)

        if exttype == 0:    # Plain text extension
            return struct.pack('>BBB',
                0xFF, ord('t'), ord('X')) \
                + b'\x0C' + rng.randrange(1 << 16).to_bytes(2, 'little')

        elif exttype in range(1..=8):   # Graphic control extension
            return struct.pack('>BBBBB',
                0xFF, ord('g'), ord('c'),
                4, ord('\x00')
            ) + rng.randint(0, 1<<8-1).to_bytes(1, 'little') \
                + rng.randint(0, 1<<8-1).to_bytes(1, 'little') \
                + rng.randint(0, 1<<8-1).to_bytes(1, 'little')

        else:        # Unknown type; skip for now
            return b''
        
    def _generate_image():    
        width  = rng.randint(1, 10)
        height = rng.randint(height, 2**16 // width)
        
        localcolortable = rng.choice([True, False])
        interlaceflag = rng.choice([False, True])
        sortflag = rng.choice([False, True])

        colorresolution = 8
        globalcolorcount = 2**(rng.randint(0, 8))
        backgroundindex = rng.randint(0, globalcolorcount-1)
        pixelaspectratio = 256 * rng.randint(0,15)

        localcolorcount = 2 ** ((rng.randint(0, 8)))

        if not localcolortable:
            globalcolortable = []
            
            for i in range(globalcolorcount):
                rgb = [rng.randint(0, 255),
                    rng.randint(0, 255), 
                    rng.randint(0, 255)]
                globalcolortable += rgb

            localcolortable = globalcolortable

        else:    
            globalcolortable = []

        lzwmincodesize = rng.randint(0, 12)
    
        imagedescriptor = struct.pack('<BHHBBBB',
            0x2C, 
            x * y, 
            (localcolortable << 7 
                | interlaceflag << 6 
                | sortflag << 5 
                | int(log2(width))+1)[0], 
            (localcolortable << 7 
                | interlaceflag << 6 
                | sortflag << 5 
                | int(log2(height))+1)[0], 
            lzwmincodesize, 
            backgroundindex, 
            pixelaspectratio >> 8
        ) 

        return imagedescriptor + globalcolortable


    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    blocksize, padding = 256 * (data[1] & 0x07) + 1, [b'\x00']

    # Write header and length of image to output
    out.write(header + len(data).to_bytes(3, 'big'))
    while byte := rng.read(blocksize):
        if blocksize - len(byte) > 0:
            # Pad short blocks with zeroes
            out.write(byte + padding[0])
        else:
            out.write(byte[:blocksize])
