#!/usr/bin/env python3

import sys
import enum
import os
import pathlib
import struct
from typing import Union


class _State(enum.Enum):
    HEADER = 0
    DATA = 1


class RandomGifGenerator():

    def __init__(self,
                 seedfile=os.devnull,
                 length=None,
                 output=sys.stdout.buffer,
                 ext=None):

        if isinstance(output, str):
            self._outfp = open(output, "wb")
        elif isinstance(output, (pathlib.Path)):
            self._outfp = open(str(output), "wb")
        else:
            self._outfp = output
        
        if not os.path.exists(seedfile):
            raise Exception("Seedfile \"{}\" does not exist.".format(
                seedfile))
            
        with open(seedfile, mode="rb") as fp:
            entropy = fp.read()
            entropylen = len(entropy)

            # Length is validated in bytes; print warning and set to
            # entropylen if it's invalid
            try:
                self._length = length or entropylen
                assert type(self._length) == int
                assert self._length >= 0
                assert self._length <= entropylen
            except AssertionError as e:
                warn = ("Length must be an integer between zero and {}. Using"
                        "{} instead.")
                warn = warn.format(entropylen, entropylen)
                print(warn, file=sys.stderr)
                self._length = entropylen

        self._ext = ext or ".gif"
        assert type(self._ext) == str

    
    def generate(self):
        state = _State.HEADER
        bufsize = 4096
        datalen = None
        readpos = 0
        while True:
            # If we've reached the end of the buffer, reset the buffer and
            # start reading from the beginning again.
            if readpos > bufsize - datalen - 1:
                state = _State.HEADER
                readpos = 0

            if state == _State.HEADER:

                # Print the Gif Header
                self._outfp.write(b'\x47\x49\x46\x38\x39\x61')

                # Get a uint8 datatype for the random length
                randlen = struct.unpack('>B',
                                        self._randgen().read())[0] % self._length
                
                # Write that length to the output
                self._outfp.write(struct.pack('>B', randlen))
                state = _State.DATA
            
            elif state == _State.DATA:

                # We don't need to write any more data after this point so we
                # can break out of the loop early if we're at the right spot 
                if readpos >= randlen-2:
                    return
            
                # Get a series of uint8s and pack them together
                chunk = struct.pack('>'+'B'*randlen,
                            *[struct.unpack('B', self._randgen().read())[0]
                                for i in range(randlen)])
                self._outfp.write(chunk)


    @staticmethod
    def _get_rng(entropy):
        return lambda n: entropy[:n]
    

    def _randgen(self):
        entropy = os.urandom(4096)
        return RandomGifGenerator._get_rng(entropy)(self._length)


