#!/usr/bin/env python3

import io
import struct
from enum import Enum
from os import urandom


class BlockType(Enum):
    EXTENSION = 0x21
    IMAGE_DESC = 0x2c

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if self is BlockType.EXTENSION:
            return 'Extension'
        elif self is BlockType.IMAGE_DESC:
            return 'Image descriptor'

        raise KeyError('Unknown block type')


class ExtensionBlock(object):
    identifier = None  # type: bytes

    def __init__(self):
        pass

    @staticmethod
    def from_data():
        return NotImplemented

    def write_to(self, buf: io.BytesIO):
        buf.write(BlockType.EXTENSION.value.to_bytes(
            1, byteorder='little'
        ))
        buf.write(len(self.identifier).to_bytes(
            1, byteorder='little'
        ))
        buf.write(self.identifier)


class GraphicControlExtention(ExtensionBlock):
    identifier = b'\xf9'
    label = True
    transparent = False
    delay = 0
    transcolor = (0, 0, 0)

    def __init__(self):
        super().__init__()

        self._buf = io.BytesIO()
    
    def set_label(self):
        self.label = True
        return self

    def unset_label(self):
        self.label = False
        return self
 
    def set_transparent(self, color=(0, 0, 0)):
        self.transparent = True
        self.transcolor = tuple(int(x) for x in color)
        
        return self

    def use_delay(self, milliseconds):
        assert milliseconds <= 65535

        self.delay = milliseconds 
        return self

    def write_to(self, buf: io.BytesIO):
        super().write_to(buf)
        buf.write(((0b11 << 7) |
                   (0b10 if self.transparent else 0b00) |
                    ((not self.label) << 3)).to_bytes(
                        1, byteorder='little'))
        buf.write((self.delay).to_bytes(
            2, byteorder='little'
        ))
        buf.write(struct.pack('<BBB', *self.transcolor))
        
        
class PlainTextExtention(ExtensionBlock):
    identifier = b'\x01'
    _text = ""

    def __init__(self):
        super().__init__()

    def set_text(self, text=""):
        self._text = text
        return self

    def write_to(self, buf: io.BytesIO):
        super().write_to(buf)

        buf.write(struct.pack("<BH", len(self._text), 0))
        buf.write(self._text.encode())
        buf.seek(-1, io.SEEK_CUR)
        buf.write(b'\x00')
    

class ImageDescriptor(object):
    left = 0
    top = 0
    width = 0
    height = 0
    local_color_table = False
    interlaced = False
    sort = 0

    def __init__(self):
        pass

    @property
    def has_local_colortable(self):
        return bool(self.local_color_table)

    @has_local_colortable.setter
    def has_local_colortable(self, value):
        self.local_color_table = bool(value)

    def write_to(self, buf: io.BytesIO):
        buf.write(BlockType.IMAGE_DESC.value.to_bytes(
            1, byteorder='little'
        ))
        pos = buf.tell()
        buf.seek(pos + 8)
        buf.write(self.left.to_bytes(2, byteorder='little'))
        buf.write(self.top.to_bytes(2, byteorder='little'))
        buf.write(self.width.to_bytes(2, byteorder='little'))
        buf.write(self.height.to_bytes(2, byteorder='little'))
        buf.write([(self.interlaced << 7) |
                  (self.sort << 6) |
                  (0 << 4) |
                  0].to_bytes(
                      1, byteorder='little'
                      ))
        buf.write([0b10 if self.local_color_table else 0]
                .to_bytes(1, byteorder='little'))
        buf.seek(pos - 1)
        buf.write((0x00).to_bytes(1, byteorder='little'))


def generate_gif(out: io.BytesIO, imagedescriptor, exts=[]):
    buf = io.BytesIO()
    buf.write("GIF89a".encode())

    pos = buf.tell()

    buf.seek(pos+4)

    imagedescriptor.write_to(buf)

    buf.seek(pos-1)
    size = buf.tell()-pos+imagedescriptor.width*imagedescriptor.height

    buf.seek(pos-1)
    buf.write(size.to_bytes(1, byteorder="little"))

    for ex in exts:
        ex.write_to(buf)

    buf.seek(pos+4)
    out.write(buf.getbuffer())


