#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    
    # The GIF format stores every pixel as an RGB triplet, which is
    # convenient for us since we can just treat it like that rather than
    # worry about bit depths and such. However, each row needs to be padded to
    # multiples of three in order not to confuse this encoding with it being
    # an image encoded in 24-bit color

    width = (1 << 7) + int.from_bytes(rng.read(1), 'big')
    height = (1 << 7) + int.from_bytes(rng.read(1), 'big')
    out.write((width).to_bytes(length=2, byteorder='little'))
    out.write((height).to_bytes(length=2, byteorder='little'))

    palettesize = ((width * height - 1) // 510) * 510 + 61
    palettelist = []
    while len(palettelist) < palettesize:
        palettelist += [int.from_bytes(rng.read(1), 'big')]

    out.write(len(palettelist).to_bytes(length=1, byteorder='little'))
    for p in range(palettesize):
        out.write(palettelist[p].to_bytes(length=1, byteorder='little'))
    
    out.write(b'\x00') # background index
    out.write(b'\x00\x00') # aspect ratio

    # LZW compression

    widthBytes = width * 3
    lzwData = []
    lastByteValue = None 
    lzwBuffer = None
    clearCode = 1 << 11
    endOfInformationCode = 1 << 12

    firstIteration = True
    while True:

        if firstIteration:

            lzwStartBits = rng.read(1)[0] & ~7 # strip away spare bits
            lzwStartMask = ~(~0 << min(lzwStartBits+1, 9)) >> 1

            numBlocks = min((2**lzwStartBits)-1,
                            len(lzwData)//lzwStartMask+1)
            
            lzwBlockIndices = list()
            for i in range(numBlocks):
                blockOffset = lzwStartMask*i
                startSliceIndex = max(blockOffset,
                                      len(lzwData)-(lzwStartMask*(i+1)))
                
                lzwBlockIndices += [(startSliceIndex,
                                    startSliceIndex+(lzwStartMask-1))]
            lzwBuffer = lzwData[:lzwBlockIndices[-1][1]]

        else:
            lzwBuffer += [rng.read(1)]
        
        currentByteVal = lzwBuffer[-1][0]
        if currentByteVal == 0:
            break
        
        matchLen = 0
        bufferIdx = lzwBuffer[-1][0]
        if lastByteValue!= None:
            while True:

                try:
                    matchLen += 1
                    if lzwData[bufferIdx - (lastByteValue + 1)] \
                      != lzwBuffer[-matchLen]:
                        raise ValueError()
                    
                    if matchLen >= 12 or \
                       lzwBuffer <= lzwBuffer[-matchLen]:
                        
                        lzwData += lzwBuffer[-matchLen:]

                        controlCode = 1 << 10 
                        if matchLen > 2:
                            if matchLen % 2!= 0:
                                print('OH NOES', matchLen)
                            
                            controlCode |= (((matchLen//2)%8)+2) ^ 8

                            if matchLen >= 12: 
                                controlCode |= (1<<11)
                                controlCode <<= 8

                        for c in reversed(range(min(12, matchLen))):
                            lzwData += [(controlCode | 
                                        lzwBuffer[-c-(2 if matchLen>12 else 0)])]

                        break
                
                except IndexError:

                    break
                finally:

                    bufferIdx += 1

        lzwData += [currentByteVal]
        lzwBuffer[:-1] = lzwBuffer[-1:]

        lastByteValue = currentByteVal
        
        firstIteration = False

    lzwEndCodes = [clearCode, endOfInformationCode]
    lzwEndCodes += [e for e 
                    in set([lzwData[i]+lzwData[j] 
                            for i in range(len(lzwData))
                            for j in range(i+2, len(lzwData))])] 
    
    lzwDictSize = len(set(lzwData + lzwEndCodes))
    out.write(((lzwDictSize-1)<<4)+lzwStartBits)
    
    lzwCurrentValue = None
    remainingRunLength = 0
    while len(lzwData)>0:
        nextLZWCode = lzwData.pop(0)
        runOnLastRead = False
        if remainingRunLength > 0:
            if remainingRunLength > 12 or \
               nextLZWCode!= lzwCurrentValue:
                

                if remainingRunLength > 12:
                    out.write((((remainingRunLength//2)%8)+2)^8)
                    out.write(1<<(11))
                    
                out.write((1 << (remainingRunLength%8))|nextLZWCode)   
                remainingRunLength -= 1 
                continue 
            else:
                remainingRunLength += 1 
                runOnLastRead = True

        
        # We've either read a new value here, or hit the end of the run
        # Let's see what we can write now
        while len(lzwData) > 0:
            if (lzwData[0] not in lzwData[1:]) or \
               (runOnLastRead and 
               remainingRunLength == 12):
                break
            
        
            runOnLastRead = False
        
            lzwDictSize = len(set(lzwData + lzwEndCodes))
            codeSizeBits = (lzwDictSize-1).bit_length()-1
            
            if remainingRunLength > 0 or \
               nextLZWCode!= lzwCurrentValue or \
               (codeSizeBits > (lzwStartBits-1)):

                if remainingRunLength > 12:
                    out.write((((remainingRunLength//2)%8)+2)^8)
                    out.write(1<<(11))
                
                if (nextLZWCode == lzwCurrentValue):
                    out.write((1 << (remainingRunLength%8))+
                               (codeSizeBits&7)|0x80)  
                elif remainingRunLength > 0:
                    out.write((1 << (remainingRunLength%8))|
                               (codeSizeBits&7))  
            
                out.write(codeSizeBits>>3)

                out.write(max([clearCode, endOfInformationCode]) |
                          (1 << (codeSizeBits-1)) |
                          0x80 |
                             (nextLZWCode << 1)) 

                remainingRunLength = 0

            else:
                remainingRunLength += 1 
                
        
        # If we didn't get a chance to find any more runs, then we should
        # output our current one before moving on
        if remainingRunLength > 0:
            
            if remainingRunLength > 12:
                out.write((((remainingRunLength//2)%8)+2)^8)
                out.write(1<<(11))
                
            out.write((1 << (remainingRunLength%8))|nextLZWCode)   
            remainingRunLength -= 1 
        else:
            lzwCurrentValue = nextLZWCode
        
    out.write(endOfInformationCode<<2)
    