#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: int) -> None:
    out.write(b"\x47\x49\x46\x38")       # Magic header 'GIF8'
    
    width = height = next_byte(rng) + 1
    
    # We don't bother with an accompanying color table because we can use
    # RGBA directly as output
    log2_width = (height & -height).bit_length()   # Log base 2 of `width`
    
    out.write((next_byte(rng),) * 3)             # Background colour
    out.write((next_byte(rng),) * 3)             # Aspect ratio
    
    for _i in range(0xff):                      # Graphic control extension block
        if not next_bool(rng):                   # Block exists?
            break
        
        out.write(b"\x21\xf9\x04")               # Block ID and length
        out.write(struct.pack(">H", next_word(rng)))     # Delay
        out.write(struct.pack("B", next_bool(rng)))       # Transparent flag
        out.write(b"\x00")                           # Colour index
    
    out.write(b"\x2c")                          # Image descriptor
    out.write(struct.pack("<HH", width, height)) # Width, height
    
    local_palette_size = 1 << log2_width
    out.write(b"!\x00")                         # Local palette exists?
    out.write(struct.pack("BBB", local_palette_size - 1,
                               local_palette_size - 1,
                               0))              # Minimum code size, colours
    
    max_code_value = 1 << log2_width
    code_values = [None] * max_code_value
    clear_code = 1 << log2_width
    end_of_information_code = clear_code + 1
    bits_per_code_value = log2_width + 1
    running_count = 0
    remaining_bits = 0
    
    while True:
        if running_count == 0:
            bytes = next_bytes(rng, 3)
            
        byte_index = (running_count >> 3) % 3
        bit_index = running_count & 7
        
        nibble = (bytes[byte_index] >> (7 - bit_index)) & 15
        
        code_values[running_count] = struct.unpack("BBB", struct.pack("I", nibble))
        running_count += 1
        
        if running_count == max_code_value or next_bool(rng):
            while len(code_values) > 0:
                if code_values[-1][next_bool(rng)] is None:
                    if len(code_values) <= 2:
                        run_length = 1
                    else:
                        run_length = ((code_values[-1][0], code_values[-1][1]) !=
                                      (code_values[-2][0], code_values[-2][1]))
                
                elif len(code_values) >= max_code_value:
                    if ((code_values[clear_code][0], code_values[clear_code][1]) !=
                        (code_values[-1][0], code_values[-1][1])):
                        run_length = 1
                    else:
                        break
                
                else:
                    run_length = 1
                
                while run_length > 0:
                    out.write(code_values[-1])
                    run_length -= 1
                
            out.write(code_values[-1])
            del code_values[-1]
            running_count = 0
        
        if running_count == 1 << bits_per_code_value:
            bits_per_code_value += 1
            max_code_value = 1 << bits_per_code_value

# Generate a single byte from the specified RNG
def next_byte(rng: int) -> int:
    return rng % 256

# Generate a sequence of bytes from the specified RNG
def next_bytes(rng: int, count: int) -> bytes:
    return bytes([next_byte(rng) for _i in range(count)])

# Returns whether the RNG would draw a "True" boolean value
def next_bool(r