#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # Parse the image descriptors
    while True:
        buf = rng.read(7)
        
        # End of stream?
        if len(buf) < 7:
            break

        # Read the descriptor type
        ubyte = buf[0]
        assert (ubyte & 0xF2) == 0
        dsc_type = (ubyte >> 5) | 0b100

        # Read the image data size and write it to the output
        # We read it from the the random stream which has a size
        # that is a multiple of four, so no need to pad the size
        width = int.from_bytes(buf[1:3], 'big')
        height = int.from_bytes(buf[3:5], 'big')
        imgsz = int.from_bytes(buf[5:7], 'big')
        out.write(buf[:4])

        # If we are not writing yet, or the image size is 0, skip the image
        if imgsz == 0: continue

        # Read the delay time and write it
        wait = int.from_bytes(rng.read(2), 'big')
        out.write(struct.pack('!H', wait))

        # Read the transparent color index and write it
        transpix = rng.read(1)[0]
        out.write(transpix)

        # Ensure that the image is at least as large as its bounds
        realimgsz = ((width * height * 3) + 3) &~ 3
        if imgsz > realimgsz:
            raise Exception('Image size too small!')
            
        # Handle simple image data
        if dsc_type == 0b101:
            print('Writing the frame data for frame {}'.format(
                    dsc_type), end='... ', flush=True)
            out.write(data[:imgsz])
            print