#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # Now we're going to write 256 descriptors at once
    # Using this loop means we avoid making multiple calls to read()
    for _ in range(257):
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

        # TODO: Should experiment with variations on these values too
        #       see https://github.com/decompme/decompilation-guide/issues/47
        out.write(b'\x2C\x00\x00\x00\x00\x00\x00\xFB\x00\x00\x00')


    last_byte = out.tell() - 1
    # At the moment this code doesn't support appending arbitrary
    # bytes onto the output so we'll have to manually compute where
    # our data will start relative to the rest of the file
    print(f'The generated gif is {last_byte} bytes long.')
    print(f'Running the fuzz test...', end='', flush=True)

    try:
        while True:
            new_data = rng.read(1)
            if not new_data:
                break
            # Keep track of how many bytes we wrote
            bytes_written = last_byte + 1
            out.write(new_data)
    except EOFError:
        pass

    # Construct an IEND record and append it
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    # ; The Image Descriptor
    # ;   The Image Descriptor contains fixed size fields.
    # ;   It also contains a variable sized Local Color Table (LCT).
    # ;   If the LCT field is present (see below) then it immediately
    # ;   follows the Image Descriptor.

    # We don't need to bother setting the dimensions or the flags;
    # they're set when creating the original file and will remain unchanged
    out.write(
            b'\x21\xff\x0bIEND\xae\x42\x60\x82'
            )
    out.flush()
    print('done!')


with open('/dev/null', 'rb+') as out