#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    last_byte = 0
    while len(data) - last_byte > 10:
        # Locate the next descriptor
        index = data.find(b';', last_byte)
        if index < 0 or (index + 10) >= len(data):
            raise RuntimeError('Descriptor not found')

        # Read the first 10 bytes of the descriptor
        descriptor = data[index:index+10]

        # Check whether the first 3 bytes are '!FN', indicating
        # a FRAME_DATA descriptor
        if descriptor[0:3] == b'!FN':
            print("Found frame descriptor, overwriting the following", end='')
            print(descriptor[3:-1], end='')

            # Overwrite with random payload
            payload_len = rng.read(1)[0]
            print('with {} bytes'.format(payload_len))
            out.write(struct.pack('<BBHIB', *[rng.read(1)[0] for _ in range(5)]))
            
        else:
            print("Copying bytes", end='')
            print(descriptor[:10], end='')
            out.write(descriptor[:10])
        
        # Skip the previous descriptor and start from the first byte after it
        last_byte = index + 10

    # Copy the remaining bytes, which don't contain descriptors
    if last_byte != len(data):
        print('Copying remaining bytes ', end='', flush=True)
        out.write(data[last_byte:])
        print('done.', flush=True)

    # Finally add the trailer to finish off the output.
    # Normally this would be part of the application logic
    # when writing a gif but since this is intended only
    # for demonstration purposes we do it here directly.
    # See https://github.com/decompme/decompilation-guide/issues/46
    print('Adding the trailer and EOF', end='', flush=True)
    # 2 bytes for signature
    # 1 byte for version
    # 1 byte for flags
    # 2 bytes for background color index
    # 2 bytes for pixel aspect ratio
    out.write(b'\x3B\x00\x00\x00')
    print(' done', flush=True)

