#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Compute the global color table field size
    global_color_table_size = 1 << ((lsd[4] >> 7) + 1)

    # Write the logical screen descriptor and add to output
    out.write(lsd)

    # Read in background color and add to output
    bgc = rng.read(1)
    out.write(bgc)

    # Generate the pixel aspect ratio field
    papf = rng.read(1)

    # Write the pixel aspect ratio field and add to output
    out.write(papf)

    # Generate the global color table and add to output
    for i in range(global_color_table_size):
        col = rng.read(3)
        out.write(col)
    assert len(out.getvalue()) == 0x26

    # Go until we have written at least one graphic control extension block
    while True:
        gceb = rng.read(5)
        
        if not gceb or gceb[0] != 0xf9:
            raise ValueError("Need GCEB before image.")
            
        # Add GCEB to the output
        out.write(gceb)
        out.write(rng.read(gceb[2]))

        image_blocks = []
        while True:
            image_block = rng.read(1)
            if not image_block or image_block == b'\x00':
                break
            else:
                image_blocks.append(image_block)
                
        # Compute the LZW code size
        lzw_code_size = struct.unpack('>B', image_blocks[1][0])[0]

        # Write the first two bytes of the image block to the output
        out.write(image_blocks[0])
        out.write(image_blocks[1][0])

        # Write all of the rest of the blocks to the output
        for image_block in image_blocks[2:]:
            out.write(image_block