#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    for i in range(len(header)):
        out.write(bytes([*header[i].to_bytes(1, 'big'), *data[:2]]))
        data = data[2:]

    # Write every frame. Each frame starts with an image
    # descriptor (an 11-byte sequence) followed by image data.
    # The LSB of the delay time field is `0`, so the frames will have
    # no delay between them.
    while len(data):
        print('Writing frame', end=' ', flush=True)
        
        # We write frame data preceded by an image descriptor
        imgdesc = b'\x2C\x00\x00\x00\x00\x00\xFF\x0B'

        # All frames begin with an image descriptor
        out.write(imgdesc)

        # The first bytes of the descriptor are the width and height of
        # the frame in pixels
        w = struct.unpack(">H", data[:2])[0]
        h = struct.unpack(">H", data[2:4])[0]
        print(f'({w}x{h})', end='', flush=True)

        # Write remaining frame data (minus the local color table)
        out.write(data[4:])

        # Shrink our buffer to just the rest of the frame data minus
        # the local color table.
        data = rng.read(datalen - len(imgdesc))
        print(', done!', end='', flush=True)
    # Write the last descriptor!
    out.write(b'\x3B')

    # And pad up until the end of the file, which is easy enough because
    # we've already written exactly datalen bytes before ending.
    print('Padding global color table', end='', flush=True)
    while True:
        if out.tell() == datalen+1: break
        out.write(rng.read(1))
    print('. Done!', flush=True)

