#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Mar 26 17:50:44 2017

@author: Sebastian Mujica

Takes a CSV file with the columns 'name' and 'country', as well as a folder of
image files, and writes a JSON file of the same name as the input file but in
the output directory. Also creates a parallel list of filenames for use in the
web interface.
"""
import csv
from os import path, walk, makedirs
import shutil
import sys
import json

__version__ = "0.0.1"

def main():
    if not (len(sys.argv) == 3):
        print("Usage: {} <input_csv> <output_dir>\n".format(__file__))
        sys.exit(1)
    
    global LOGGER
    LOGGER = open('./{}'.format(path.basename(__file__).rstrip('.py')), 'w+')
    print ("Logging to ./{}\n".format(path.basename(__file__)))
    
    read_directory = sys.argv[1]
    write_directory = sys.argv[2]
    output_filename = "{}.json".format(path.splitext(path.basename(read_directory))[0])
        
    # Read people from CSV into a dictionary.
    reader = csv.DictReader(open(read_directory))
    
    # Make sure there is a matching filename for each person.
    images_by_person = dict()
    for row in reader:
        filename = find_image_for_person(row['name'], row['country'])
        if not filename or len(filename) <= 6:
            print("Could not find image for {}\t{}".format(row['name'], row['country']))
            continue
            
        row['imgFilename'] = filename
        row['imgPath'] = "/images/people/{}/{}".format(row['id'], filename)
                  
        try:
            images_by_person[row['country']].append([row['name'].lower(), row['id']])
        except KeyError:
            images_by_person[row['country']] = [[row['name'].lower(), row['id']]]
              
    # Write out to the new JSON file.
    with open("{}/{}".format(write_directory, output_filename), 'w+') as fp:
        json.dump(reader, fp)
    
    # Write out a list of country names and their images.
    with open("{}/countries.txt".format(write_directory), 'w+') as countries:
        for country in images_by_person:
            #print (country + "\t" + str(images_by_person[country]))
            countries.write("%s\t%s\n" % (country, str(images_by_person[country]).replace("\'", "")))
    
    LOGGER.close()
    return

def find_image_for_person(name, country):
    """
    Searches through the image directory and returns a good match for the 
    given name & country combination. First try with full name, then just 
    first initial and last name; and finally, just last name.
    
    Arguments
    =========
    name : string
        Full name of the person whose images are being searched.
    country : string
        The country where this person lives.
    """
    base_path = "../data/stanford"
    img_path = None
    
    # Try to get an exact image match.
    img_path = look_in_dir(base_path+"/"+country, name)
    if img_path:
        print("Found exact image match!")
        return img_path
    
    # If that didn't work, try just the last name.
    img_path = look_in_dir(base_path+"/"+country, name.split(" ")[-1])
    if img_path:
        print("Found last name match.")
        return img_path
    
    # If that didn't work either, give up and just take a random image.
    img_path = rand_select(base_path+"/"+country)
    return img_path
    
def lookup_names(matches):
    """
    Returns a tuple of two lists that contain the name of the first person in 
    the matches list, and the last person respectively, since it seems that 
    the number of images per person may vary by country.
    
    Argument
    ========
    matches : string[]
        List of strings that are the names of the people we want to search
        through the image directories for.
    
    Return
    ======
    first_name : string[]
        List containing the names of the first person in the list.
    last_name : string[]
        List containing the names of the last person in the list.
    """
    first_name = []
    last_name = []
    
    for p in matches:
        if len(first_name) == 0:
            first_name.append(p)
        else:
            last_name.append(p[-1])
    
    return (first_name, last_name)

def look_in_dir(search_dir, name):
    """
    Recursively searches through all subdirectories under the given search dir
    for a file with the provided name.
    
    Arguments
    =========
    search_dir : string
        Path to the starting directory to begin searching.
    name : string
        Filename we are looking for inside of the search_dir.
    
    Returns
    =======
    img_path : string
        Full system path to the file name passed as a parameter.
    """
    img_path = None
    
    for root, dirs, files in walk(search_dir):
        for file in files:
            if file.startswith(name):
                img_path = file
                break
        if img_path:
            break
            
    return img_path

def rand_select(search_dir):
    """
    Returns a randomly selected file from the given directory. Doesn't look at
    any subdirectories recursively.
    
    Arguments
    =========
    search_dir : string
        Directory path in which to select a file randomly.
    
    Returns
    =======
    img_path : string
        Name of a randomly selected file.
    """
    import random
    img_list = next(walk(search_dir), (None, None, []))[-1]
    if len(