#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 1 byte for width and height (`width`, `height`) in
#       LSB order (e.g., width = 0x12, height = 0x34).
#   3. 256 x 3 bytes of color table. Each entry is an RGB
#       triplet with each channel taking up 1 byte. The
#       colors are just repeated versions of the values
#       0x00 to 0xFF.
#   4. 1 byte for the block label `%b`.
#   5. 1 byte for the code size `size`.
#   6. `size` bytes of compressed data.
def generate_gif(out: BinaryIO, width: int, height: int) -> None:
    # Magic number
    out.write(struct.pack('<4s', b'GIF89a'))
    
    # Logical screen descriptor
    ls_dscr = struct.pack('<BBB', width & 0xff, height & 0xff, 0x7f | 0x80)
    out.write(ls_dscr)
    
    # Background color
    bg_color = 0x00
    
    # Global color table
    global_ctable = []
    for i in range(256):
        ctable = struct.pack('<B', i|i<<8|i<<16)
        
        out.write(ctable)
        
        global_ctable += [ctable]
    
    # Image block
    img_block = struct.pack('<B', '%')
    out.write(img_block)
    
    # Image descriptor
    descriptor = struct.pack('<BBHH', (0x80 << 1) | 0xf0, 0, 0, 0x100)
    out.write(descriptor)
    
    # Local color table flag
    local_ctable = True
    if local_ctable:
        lc_flag = 0x80
    else:
        lc_flag = 0x00
    
    lc_table = struct.pack('BB', lc_flag|7, 0)
    out.write(lc_table)
    
    # Codesize
    codesize = 8
    out.write(struct.pack('B', codesize))
    
    # Data
    # Taken from original gif
    data = b'\x80\xa0\xbd\x00\x80\xa0\xb3\x00\x80\xa0\xba\x00\x80\xa0\xbb\x00\x80\xa0\xbc\x00\x80\xa0\xbf\x00\x80\xa0\xc0\x00\x80\xa0\xc1\x00\x80\xa0\xc2\x00\x80\xa0\xc3\x00\x80\xa0\xc4\x00\x80\xa0\xc5\x00\x80\xa0\xc6\x00\x80\xa0\xc7\x00\x80\xa0\xc8\x00\x80\xa0\xc9\x00\x80\xa0\xca\x00\x80\xa0\xcb\x00\x80\xa0\xcc\x00\x80\xa0\xcd\x00\x80\xa0\xce\x00\x80\xa0\xcf\x00\x80\xa0\xd0\x00\x80\xa0\xd1\x00\x80\xa0\xd2\x00\x80\xa0\xd3\x00\x80\xa0\xd4\x00\x80\xa0\xd5\x00\x80\xa0\xd6\x00\x80\xa0\xd7\x00\x80\xa0\xd8\x00\x80\xa0\xd9\x00\x80\xa0\xda\x00\x80\xa0\xdb\x00\x80\xa0\xdc\x00\x80\xa0\xdd\x00\x80\xa0\xde\x00\x80\xa0\xdf\x00\x80\xa1\xe0\x00\x80\xa1\xe1\x00\x80\xa1\xe2\x00\x80\xa1\xe3\x00\x80\xa1\xe4\x00\x80\xa1\xe5\x00\x80\xa1\xe6\x00\x80\xa1\xe7\x00\x80\xa1\xe8\x00\x80\xa1\xe9\x00\x80\xa1\xea\x00\x80\xa1\xeb\x00\x80\xa1\xec\x00\x80\xa1\xed\x00\x80\xa1\xee\x00\x80\xa1\xef\x00\x80\xa1\xf0\x00\x80\xa1\xf1\x00\x80\xa1\xf2\x00\x80\xa1\xf3\x00\x80\xa1\xf4\x00\x80\xa1\xf5\x00\x80\xa1\xf6\x00\x80\xa1\xf7\x00\x80\xa1\xf8\x00\x80\xa1\xf9\x00\x80\xa1\xfa\x00\x80\xa1\xfb\x00\x80\xa1\xfc\x00\x80\xa1\xfd\x00\x80\xa1\xfe\x00\x80\xa1\xff\x00\x80\xa2\x80\x00"r\xe0<\x80\xa2\x81\x00"\n\xe0<\x80\xa2\x82\x00"\t\xe0<\x80\xa2\x83\x00"\f\xe0<\x80\xa2\x84\x00"\r\xe0<\x80\xa2\x85\x00"\x00\xe0<\x80\xa2\x86\x00"\x01\xe0<\x80\xa2\x87\x00"\x02\xe0<\x80\xa2\x88\x00"\x03\xe0<\x80\xa2\x89\x00"\x04\xe0<\x80\xa2\x8a\x00"\x05\xe0<\x80\xa2\x8b\x00"\x06\xe0<\x80\xa2\x8c\x00"\x07\xe0<\x80\xa2\x8d\x00"\b\xe0<\x80\xa2\x8e\x00"\t\xe0<\x80\xa2\x8f\x00"\n\xe0<\x80\xa2\x90\x00"\v\xe0<\x80\xa2\x91\x00"\f\xe0<\x80\xa2\x92\x00"\r\xe0<\x80\xa2\x93\x00!r\xe0<\x80\xa2\x94\x00! \xe0<\x80\xa2\x95\x00!"\xe0<\x80\xa2\x96\x00!"\xe0