#!/usr/bin/env python3

import os
import struct
from hashlib import sha1
from binascii import unhexlify

from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

def pbkdf2(password, salt, iterations = 100000):
    kdf = PBKDF2HMAC(
            algorithm = hashes.SHA256(),
            length = 4,
            salt = unhexlify(salt),
            iterations = iterations,
            backend = crypto_default_backend())
    return str(kdf.derive(unhexlify(password)))

def xor_bytes(a, b):
    return bytes([c^d for c, d in zip(a,b)])

def generate_key(password, salt, bits):
    return pbkdf2(password, salt, bits//8*4).encode("hex")

def encrypt(msg, key):
    ciphertext = []
    for i in range(len(msg)):
        ciphertext.append(chr((ord(msg[i]) ^ ord(key[i % len(key)]))))
    return "".join(ciphertext)

# This function is from https://stackoverflow.com/questions/21015182/python-fastest-way-to-gzip-a-file
def compress_data(output, input):
    f = open(input, 'rb')
    data = f.read()
    f.close()
    with gzip.open(output, 'wb', compresslevel=9) as fo:
        fo.write(data)

if __name__ == '__main__':
    import sys
    import random
    import os
    import math
    import mmap
    import zlib
    import gzip
    import time

    if len(sys.argv) != 4 and len(sys.argv) != 5:
        print("Usage: {} <input> <output> <password> [salt]".format(os.path.basename(__file__)))
        exit(-1)

    inputF = sys.argv[1]
    outputF = sys.argv[2]
    password = sys.argv[3].strip('\n').strip('/')

    salt = sys.argv[4] if len(sys.argv) == 5 else time.time().hex()
    key = generate_key(password, salt, 128)
    
    rng = open("/dev/urandom", "rb")
    print('Generating a header...')
    magic = rng.read(3)
    width, height = map(int, struct.unpack("<BB", rng.read(2)))
    background, pixel_aspect = struct.unpack("<BB", rng.read(2))
    flags = ord(rng.read(1))
    delay = struct.unpack("<H", rng.read(2))[0]
    transparent = ord(rng.read(1))
    terminator = ord(rng.read(1))
    header = struct.pack("<3sBBBBI", *[magic, width, height, background, pixel_aspect, flags, delay, transparent, terminator])
    data = rng.read(width*height)

    datalen = len(data)

    print('Writing the header...', end='', flush=True)
    out = open(outputF, "wb")
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # Now we're going to write 256 descriptors at once
    # Using this loop means we avoid making multiple calls to read()
    for _ in range(257):
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)
        out.write(struct.pack('<H', 0x10))
        out.write(rng.read(0x10))
    # Write the last descriptor!
    out.write(b'\x3B')

    # And pad up until the end of the file, which is easy enough because
    # we've already written exactly datalen bytes before ending.
    print('Padding global color table', end='', flush=True)
    while True:
        if out.tell() == datalen+1: break
        out.write(rng.read(1))
    print('. Done!', flush=True)

