#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # The GIF's header
    # See https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    out.write(b"GIF89a")

    # The Logical Screen Descriptor
    width = rng.read(2)
    height = rng.read(2)
    rng.seek(-1, 1) # Seek back one byte
    packed = (int.from_bytes(rng.read(1), "big"), int.from_bytes(width, "little"))
    global_color_table_flag = (packed & (1 << 7)) != 0
    color_resolution = ((packed >> 4) & 7) + 1
    sort_flag = (packed >> 3) & 1
    global_color_table_size = 2 ** ((packed & 7) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    out.write(struct.pack("!HHBBB", *[x for x in [width, height, global_color_table_flag, color_resolution, sort_flag, global_color_table_size, background_color_index, pixel_aspect_ratio]]))

    # Global Color Table
    if global_color_table_flag:
        out.write(b"\xFF\xFF\xFF")

    # Application Extension Block
    app_extension_block_length = b"\x0C" # Length of this block
    app_identifier = "NETSCAPE2.0"
    app_authentication_code = b"\x03\x01"
    block_size = b"\x01"
    loop_count = rng.read(2)
    rng.seek(-1, 1)
    out.write(app_extension_block_length)
    out.write(app_identifier.encode())
    out.write(app_authentication_code)
    out.write(block_size)
    out.write(loop_count)

    # Image Separator
    out.write(b";")

    # Graphic Control Extension Block
    gce_block_length = b"\x04" # Length of this block
    reserved = b"\x00"         # Reserved; should be zero
    disposal_method = rng.read(1)[0] & 0b111
    user_input_flag = (rng.read(1)[0] & 0b100) >> 2
    transparent_color_flag = (rng.read(1)[0] & 0b10) >> 1
    delay_time = rng.read(2)
    rng.seek(-1, 1)
    out.write(gce_block_length)
    out.write(reserved)
    out.write(struct.pack("!BB", disposal_method, user_input_flag | transparent_color_flag))
    out.write(delay_time)
    out.write(b"\x00") # Transparent Color Index

    # Image Descriptor
    left_pos = rng.read(2)
    top_pos = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    rng.seek(-5, 1) # Seek back five bytes
    packed = (int.from_bytes(rng.read(1), "big"),
              int.from_bytes(left_pos, "little"),
              int.from_bytes(top_pos, "little"))
    local_colortable_flag = (packed & (1 << 7)) != 0
    interlace_flag = (packed & (1 << 6)) != 0
    sorted_flag = (packed & (1 << 5)) != 0
    reserved_bit = (packed & (1 << 4)) != 0
    size_of_local_colortable = (packed & 7) + 1
    if not reserved_bit:
        raise Exception()
    out.write(struct.pack("!HHHBB", *[x for x in [left_pos, top_pos, width, height, local_