#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   - Each frame's header consists of:
#       2. 4-byte size field (unsigned, big endian).
#       3. 1 byte introducer ("!").
#       4. 1 byte label ("G").
#       5. 1 byte separator (",").
#       6. 3 bytes reserved for future use (0).
#       7. 1 byte extension introducer (";").
#       8. 1 byte label ("F").
#       9. 1 byte separator (",").
#      10. 4 bytes frame delay in milliseconds (big endian).
#      11. 1 byte terminator ("\x00").
#     - Each frame's data is filled with zeroes (`0x00`) to make its size divisible by 16.
#   - The last frame ends with an additional block of null bytes from which a new frame can be started.
def gen_random(out: BinaryIO) -> None:
    out.write("GIF89a".encode())

    while True:
        # Read and discard the next frame, but check if we have reached EOF first.
        try:
            size = int.from_bytes(out.read(4), "big")

            assert out.read(1) == b"!"
            assert out.read(1) == b"G"
            assert out.read(1) == b","
            assert out.read(3) == b"\x00\x00\x00"
            assert out.read(1) == b";"
            assert out.read(1) == b"F"
            assert out.read(1) == b","
            assert out.read(4) == struct.pack(">I", 1 + (size // 2))[:4]
            assert out.read(1) == b"\x00"

            # Make sure that the entire frame has been read before continuing.
            assert len(out.read(1)) == 0

        except AssertionError:
            break

        # Write data until it fits into a multiple of 16.
        written = 0
        while written % 16!= 0:
            out.write(b"\x00")
            written += 1


if