#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # 1024 should be enough to make all bitmaps in the gif.
    frame_count = 16

    width = height = 8 * (frame_count // 5 + 1)
    global_color_table = [struct.pack('BBB', i, i, i) for i in range(256)]

    local_color_tables = []
    for _ in range(frame_count):
        palette = [
            global_color_table[rng.read(1)[0] % len(global_color_table)],
            global_color_table[rng.read(1)[0] % len(global_color_table)],
            global_color_table[rng.read(1)[0] % len(global_color_table)]
        ]

        local_color_tables += [palette]


    out.write(b'GIF89a')
    write_logical_screen_descriptor(width, height, out)
    write_loop_control(out)
    write_application_extension(out)

    data = bytearray()

    for i in range(frame_count):
        write_image_descriptor(width, height, False, 0, out)
        write_local_color_table(*local_color_tables[i], out)
        write_graphics_control_extention(out)
        compress_frame(data, out)

    write_trailer(len(data), out)


def write_file_header(out: BinaryIO):
    out.write(b"GIF87a")


def write_logical_screen_descriptor(
        width: int, height: int, out: BinaryIO):
    bits_per_pixel = 1
    global_color_table_size = 0
    color_resolution = 0xff
    sort_flag = b'\x00'
    global_color_table = b''
    background_color_index =.background_color_index
    pixel_aspect_ratio = 0xff

    descriptor = struct.pack('<HHBBB',
                             width, height, bits_per_pixel,
                             global_color_table_size, color_resolution) \
                 + sort_flag \
                 + global_color_table \
                 + bytes([background_color_index]) \
                 + bytes([pixel_aspect_ratio])

    out.write(descriptor)


def write_loop_control(out: BinaryIO):
    loop_count =.loop_count
    out.write(bytes([
        0x21, 0xFF, 0x0B,  # Extension Introducer and Label
        0x00,             # Block Size
        0x01,             # Loop Control Flag
        0x00, 0x00        # Loop Count Lo, Hi
    ]))


def write_graphics_control_extention(out: BinaryIO):
    delay =.delay * 10  # Frames per second
    out.write(bytes([
        0x21, 0xF9,      # Extension Introducer and Label
        0x04,             # Block Size
        0x05,             # Graphics Control Extension Flags
        0x0A,             # Delay Time Lo
        0x00,             # Delay Time Hi
        0x00              # Transparent Color Index
    ]))


def write_local_color_table(r: str, g: str, b: str, out: BinaryIO):
    out.write(bytes([
        0x2C,                   # Image Separator
        0x00, 0x00, 0x00,       # Image Left Position Lo, Mid, Hi
        0x00, 0x00, 0x00,       # Image Top Position Lo, Mid, Hi
        0x08, 0x00,             # Image Width Lo, Hi
        0x08, 0x00              # Image Height Lo, Hi
    ]))



def write_image_descriptor(
        width: int, height: int, has_local_color_table: bool, lzw_min_code_size: int, out: BinaryIO):
    table_size = 0 if not has_local_color_table else 7
    interlace = True
    sort_flag = 0
    local_color_table = b''

    descriptor = struct.pack("<BBH",
                             0x2c,  # Image Seperator
                             table_size,
                             0x00
                            )\
                  + bytes([int(interlace) << 7 | sort_flag]) \
                  + struct.pack("<BBBBBBB",
                                width & 0xff,
                                (width >> 8) & 0xff,
                                height & 0xff,
                                (height >> 8) & 0xff,
                                0x08 | lzw_min_code_size,
                                0x00
                               )

    print(" ".join(["{:02X}".format(_) for _ in descriptor]))
    out.write(descriptor)

def write_application_extension(out: BinaryIO):
    app = b""
    out.write(app)

def write_comment_extension(out: BinaryIO):
    comment = b""
    out.write(comment)


def write_plain_text_extension(out: BinaryIO):
    text = b""
    title = b""
    author = b""
    copyright = b""
    out.write(text)
    out.write(title)
    out.write(author)
    out.write(copyright)


def write_graphic_control_label(out: BinaryIO):
    block_size = 0x00
    label = 0xf9
    flags = 0x04
    delay_time = 0x0a
    transparent_idx = 0x0
    terminator = 0x00

    out.write(bytes([block_size, label, flags, delay_time]))
    out.write(bytes([transparent_idx, terminator]))


def write_trailer(lzw_stream_length: int, out: BinaryIO):
    stream = b""  # TODO
    out.write(bytes([
        0x3B,   # Frame Separator
        0x00    # File Terminator
    ]))
