#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # Now we're going to write 256 descriptors at once
    # Using this loop means we avoid making multiple calls to read()
    for _ in range(257):
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

        out.write(b'\x2C\x00\x10')

        # Actually writing the descriptor body from the random input
        pixel_width = ord(rng.read(1))
        pixel_height = ord(rng.read(1))

        # Two bits for each color but only need to support black & white
        has_color_table = bool(ord(rng.read(1)))
        interlaced = bool(ord(rng.read(1)))
        sorted_color_table = bool(ord(rng.read(1)))

        # There is no transparency flag, instead
        # the number of colors used in the palette is duplicated
        # as the sort of transparent color index
        ncolors = ord(rng.read(1))

        if not has_color_table or sorted_color_table:
            raise ValueError("Has color table but is not sorted")
        
        out.write((pixel_width).to_bytes(1,'little'))
        out.write((pixel_height).to_bytes(1,'little'))
        # 2 bytes for pixel aspect ratio
    out.write(b'\x3B\x00\x00\x00')
    print(' done', flush=True)

