#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the given parameters.
def generate_random(width: int, height: int, color_depth: int = 8, out: BinaryIO = None):
    # Header (6 bytes)
    header = b"GIF89a"

    # Logical screen descriptor (7 bytes)
    descriptor = bytearray()
    descriptor += width.to_bytes(2, "little")
    descriptor += height.to_bytes(2, "little")
    # Flags: global colour table and no sort flag
    flags = 0b10000000 | ((color_depth - 1) << 4).to_bytes(1, "big")
    descriptor += flags + b"\x00\x00"
    
    # Global color table (3 * 2^(n+1)) (4 bytes each)
    color_table = []
    for i in range((1 << color_depth)):
        r = g = b = 0
        if i == 0:
            r = b = 255
        elif i % 3!= 0:
            if not i % 3:
                g = 255
            else:
                r = g = 255
        color_table.append([r, g, b])
    colors = [int(i / 2**(color_depth-1), 2) for i in sum(color_table, [])]
    color_data = b"".join([(c & 0xFF).to_bytes(1, "big") for c in colors])

    out.write(header)       # Write header data to file
    out.write(descriptor)   # Write descriptor data to file
    out.write(color_data)   # Write color table data to file

    # Image block (13 bytes)
    image_block = bytearray()
    # Left position of image
    image_block.extend(b"\x00\x00\x00")
    # Top position of image
    image_block.extend(b"\x00\x00\x00")
    # Width of the frame
    image_block.extend(width.to_bytes(2, "little")) 
    # Height of the frame
    image_block.extend(height.to_bytes(2, "little"))  
    # Flags: packed fields present
    image_flags = 0b10000000 | 0b00000011     # Packed field size is 8 bits
    image_block.append(image_flags)
    # Background colour index
    image_block.append(colors[0].to_bytes(1, "big"))
    # Pixel aspect ratio
    image_block.append(b'\x00')
    out.write(image_block)      # Write image block data to file

    # Graphics control extension (10 bytes)
    graphics_control = b"\x21\xF9\x04\x01\x00\x0F" \
                       b"\x00"          # Block terminator
    out.write(graphics_control)    # Write graphics control data to file

    # Application extensions (11 bytes)
    application_extensions = b"\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00\x00" \
                             b"\x00\x00\x00"         # Block terminator
    out.write(application_extensions)    # Write application data to file

    # Loop counter block (7 bytes)
    loop_counter_block = b"\x21\xFF\x00\x0BNETSCAPE2.0\x03\x01\x00\x00\x00" \
                         b"\x00\x00\x00"            # Block terminator
    out.write(loop_counter_block)    # Write loop count block to file

    # Image data (LZW encoded)
    lzw_bits = max(color_depth, 2)
    clear_code = (1 << lzw_bits)
    end_of_information_code = clear_code + 1

    # Mapping from codes to pixel values
    dictionary = {clear_code: [], end_of_information_code: []}
    for i in range((1<<lzw_bits)-1):
        dictionary[(i+2)] = [(i>>1) & 1, (i>>2) & 1][::-1] \
                            + list(dictionary[i][:-(len(dictionary[i]) // 2)])

    # Start with an empty code
    current_code = clear_code
    stream = bytearray()
    for i in range(width*height):
        # Add new code to LZW stream
        stream.extend(dictionary[current_code])

        # Output pixel value based on code
        if len(stream) > 0:
            out.write(struct.pack("!%dB" % lzw_bits,
                                 *(sum(stream[-lzw_bits:], ()))))

            if current_code < end_of_information_code \
                    and len(dictionary[current_code]) >= (1<<(lzw_bits-1)):
                new_entry = tuple(0 if x==1 else 255 for x in dictionary[current_code])

                # Check if we need to increase the dictionary size
                if len(dictionary)==(1<<lzw_bits):
                    print("Dictionary full!")

                    # Increase the number of bits used by the lookup table
                    lzw_bits += 1
                    clear_code = (1 << lzw_bits)
                    end_of_information_code = clear_code + 1
                    
                    # Update the mapping from code to pixels
                    dictionary.update({clear_code: [], end_of_information_code: []})
                    dictionary.update({k: v+(new_entry[:-(len(v)//2)], )
                                        for k, v in dictionary.items() 
                                        if isinstance(k, int)})
                
                # Add new entry to dictionary
                next_index = len(dictionary)+1
                dictionary[next_index] = new_entry

                # Set current code and replace old code with new one
                old_code = current_code
                current_code = next_index
        
        # If current code has been replaced, reset stream
        while len(stream) <= 0 or len(dictionary[current_code]) == 0:
            stream = bytearray() 
            current_code = clear_code
        
    termination_byte = struct.pack('!H', b'\x3B')
    out.write(termination_byte)

