#!/usr/bin/env python3

import struct
from typing import BinaryIO

# The target GIF starts with 'GIF89a' BE followed by a screen descriptor
# and then an application extension containing the bare minimum for any valid
# gif file as well as a color table starting at offset 765 in the file.
with open('myanim.gif', mode='r+b') as infile:
    # Read all bytes from input
    gifdata = infile.read()

    # Find location of last image descriptor from first byte after screen desc
    loc = gifdata[4]+1 + gifdata[(5*4)+4] + 1

    # Extract the color table section into a separate buffer
    ncolors = len(gifdata)//3
    colortable = b''.join([gifdata[i:i+3]
                           for i in range(loc-ncolors, loc)])

    # Now re-create the output file but only using what remains after the
    # original images are completed
    with open('out.gif', mode='wb') as outfile:
        outfile.write(b'GIF89a')

        # Copy over all remaining bytes until the last image description block
        # but include a new screen descriptor that points back to the position
        # the old one pointed to
        outfile.write(gifdata[:loc])
        outfile.write((loc//4).to_bytes(size=2, byteorder='little'))

        # Write the rest of the output which is just the full contents of the
        # old file minus everything up through the last image descriptor
        outfile.write(gifdata[loc:])

    # Now write the color table
    with open('color.tbl', mode='wb') as ctablefile:
        ctablefile.write(colortable)


print('\nFinished!')

# Now let's add some more animated pixels!
with open('out.gif', mode='ab+') as appendfile:
    # We can also read files like normal binary files if needed
    with open('additionalpix.gif', mode='rb') as pixfile:
        # This skips over the header of the file
        pixdata = pixfile.read()[10:]

    # Seek forward to the correct point in the output file
    appendfile.seek(appendfile.tell())

    # Write the additional pixel data
    appendfile.write(pixdata)

    # Now we need to find the address of the beginning of the data chunk
    # for the last frame so we can pad it appropriately
    last_byte = out.tell() - 1
    # At the moment this code doesn't support appending arbitrary
    # bytes onto the output so we'll have to manually compute where
    # our data will start relative to the rest of the file
    print(f'The generated gif is {last_byte} bytes long.')

    # For now we'll hardcode the number of loops to use, since we don't
    # yet know how much space remains between the final descriptor and EOF
    numloops = 0xC0

    paddinglen = (numloops * 3).to_bytes(size=2, byteorder='little')

    # Write the leftover bytes before the last descriptor
    print('Writing remaining bytes', end='', flush=True)
    out.seek(-1, 2)
    out.write(paddinglen)
    out.write(struct.pack('<B', numloops))
    out.flush()
