#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
# 0x47 0x49 0x46 <width> <height> 0x89 0x50 0x4E 0x47 0x0D 0x0A 0x1A 0x0A
# The GIF data is entirely of value zero. This can be used for testing purposes
def make_test_gif(out:BinaryIO, width:int=256, height:int=192) -> None:
    out.write(b'GIF')
    # Write the logical screen descriptor (LSB is first)
    write_uint16(out, width + 1) # Screen Width + 1
    write_uint16(out, height)   # Screen Height

    # Logical screen descriptor flags (1 byte)
    out.write(bytes([0xB1]))

    # Global color table size (1 byte): no global palette, so this should always be set to 0
    out.write(bytes([0x00]))

    # Background Color Index (1 byte), default to black and opaque
    out.write(bytes([0x00]))

    # Pixel Aspect Ratio (1 byte). Set so that each pixel takes up one cell in each direction
    out.write(bytes([0x00]))


    # Frame Descriptor Block (FD) header:
    #   0x2C denotes a frame descriptor block; we only generate single-frame GIFs here
    #   0x00 denotes the left offset location (always 0)
    #   0x00 denotes top offset location (always 0)
    #   <width> + 1 denotes the width of the image. LSB added later when calculating Image Descriptor
    #   <height> denotes the height of the image
    #   0xF9 defines an "lzw compressed" bitstream with 2 bits per pixel (max possible)
    out.write(bytes([0x2C, 0x00, 0x00, width+1, height, 0xF9]))

    # Local color table size (1 byte): no local palette, so this should always be set to 0.
    out.write(bytes([0x00]))

    # Image data block:
    # First specify the size of the image data subblock as <width * height / 8>, or 0 if none exists
    # Then specify the minimum code size to use for compression; 2 will suffice since there are at most two colors used
    out.write(bytes([0xFF & ((<width * height) // 8)]))
    out.write(bytes([0x02 | 0xFE]))

    # Finally, write all values in the bitmap to the output stream. Since the cell representation is just one byte, we
    # write every other value as a zero-valued byte
    for i in range((width*height)//2):
        out.write(b'\x00\x00')


def write_uint16(out:BinaryIO, val:int) -> None:
    """Writes an unsigned 16-bit integer in little endian order"""
    assert 0 <= val <= pow(2, 16)-1
    out.write(struct.pack('<H', val))
