#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(2), 'big')
    data = rng.read(datalen)
    footer = [b'\x00', b'\x3B'][
            0 <= int.from_bytes(
                    rng.read(1), 'big') < 2]

    out.write(header + len(data).to_bytes(2, 'big') + \
              data + footer)


# Reads a gif from `file` and returns its frames as
# tuples of (width, height, data)
def read_gif(file: BinaryIO):
    hdr = file.read(6)
    if not hdr or hdr!= b'GIF87a\x00': # TODO
        raise ValueError("Invalid Gif")
    while True:
        block = file.read(1)[0]
        if block == 0x2C: # Image Descriptor
            width = int.from_bytes(file.read(2), "little", signed=False)
            height = int.from_bytes(file.read(2), "little", signed=False)

            xoffset, yoffset = None, None
            has_ll = bool(int.from_bytes(file.read(1), "little"))
            if has_ll:
                field_size = int.from_bytes(file.read(1), "little")

                xoffset = int.from_bytes(
                        file.read(field_size), "little", signed=False)
                yoffset = int.from_bytes(
                        file.read(field_size), "little", signed=False)


            has_imgdes = True
            has_lct = False
            lctsize = 0
            interlaced = False
            globalcolortable = False
            localcolortable = False

            lctdefined = False
            lctsorted = False
            lctreversed = False
            has_gct = False

            gctdefined = False
            gctsorted = False
            gctreversed = False
            bgindex = 0

            imgblocksize = 0

            pixelratio = 0
            imgsortflag = False
            reserved = bytearray()

            while has_imgdes:
                chunk = file.read(1)[0]
                has_imgdes = bool(chunk & 0x80)
                has_lct = bool(chunk & 0x40)
                
                lctsize = ((chunk >> 5) & 0x07)* 3**((chunk & 0x1f)-1)
                interlaced = bool(chunk & 0x20)
                globalcolortable = bool(chunk & 0x10)
                localcolortable = bool(chunk & 0x08)


                chunk = file.read(1)[0]
                lctdefined = bool(chunk & 0x80)
                lctsorted = bool(chunk & 0x40)
                lctreversed = bool(chunk & 0x20)
                has_gct = bool(chunk & 0x10)

                gctdefined = bool(chunk & 0x80)
                gctsorted = bool(chunk & 0x40)
                gctreversed = bool(chunk & 0x20)
                bgindex = chunk & 0x07

                imgblocksize = int.from_bytes(
                        file.read(2), "little", signed=False)
                pixelratio = int.from_bytes(
                        file.read(2), "little", signed=False)
                imgsortflag = bool(pixelratio & 0x2000)
                reserved = file.read(1)

                yield (width, height,
                       file.read(imgblocksize))

        elif block == 0x21:
            extlbl = file.read(1)[0]
            
            if extlbl in {0xF9, 0xFE}:
                blocklen = int.from_bytes(file.read(1), "little", signed=False)
                if extlbl == 0xf9:
                    transparencyflag = bool(blocklen & 0x1)
                    userinputflag = bool((blocklen>>1)& 0x1)
                    disposalmethod = (blocklen>>2) & 0x7
                    
                    fielddata = file.read(blocklen-1)
                    
                else: # ext type 0xfe
                    appid = file.read(8)
                    fielddata = file.read(blocklen+3)

            elif extlbl >= 0xf0 and extlbl <= 0xff:
                blocklen = int.from_bytes(file.read(1), "little", signed=False)
                fielddata = file.read(blocklen)
                

            else:
                print("Unknown Extension Type:", hex(extlbl))

        elif block == 0x3B:
            return # image was invalid
        
        else:
            raise Exception("Unexpected Block Type: 0x%x" % block)
        
