#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def write_gif(out: BinaryIO, rng):
    header = b'GIF89a\x02\x00\x01\x00'

    out.write(header)

    # Write logical screen descriptor
    width = 5 * (4 + rng.randint(0xff))
    height = rng.randint(0, 999) % 7
    sbits = 1 << rng.randint(0b10000, 0b11111) & 0xffff
    sbgcolor = rng.randint(0x000000, 0xffffff) & sbits
    sebgindex = rng.randint(0x00, 0xff) if sbits & 0b1 else None
    spixelaspectratio = rng.choice([None] + list(range(1, 16)))


    # Logical Screen Descriptor
    out.write(struct.pack('>H', width))
    out.write(struct.pack('>H', height))
    packed = [sbits      ,
              sbgcolor    ,
              sebgindex    ]

    for i in range((sbits >> 4) - len(packed)):
        packed += [rng.choice([-1, None]),
                   rng.randint(0x000000, 0xffffff)]

    for p in zip(*[iter(packed)]*2):
        out.write(struct.pack('BB', *p))

    if spixelaspectratio is not None:
        out.write(bytes([spixelaspectratio]))

    # Global Color Table
    sctsize = min(2 ** ((sbits&0b1100)>>2), 256)
    for _ in range(sctsize):
        r, g, b = tuple(map(lambda x: x / 0xff,
                            map(lambda x: int(round(rng.random() * x)),
                                [2**i for i in reversed(range(8))])))

        out.write(struct.pack('BBB',
                              int(r * 0xff),
                              int(g * 0xff),
                              int(b * 0xff)))

    # Color Table Size and Sort Flag
    out.write(bytes([(len(packed)-1)&0xff,
                     (len(packed)-1) > 1<<((sbits&0b1100)>>2)]))

    # Background Color Index
    bcgcolor = rng.randint(0x00, 0xff) if sbits & 0b10 else 0xf9
    out.write(bytes([bcgcolor]))

    # Pixel Aspect Ratio
    if spixelaspectratio is not None:
        palettesize = max(min(((spixelaspectratio+1)**2) // 2, 256), 1)
        out.write(bytes([palettesize]))

    # Image Descriptor
    out.write(bytes([0x2c,
                     0x00,      # Left
                     int(width//3)%width,   # Top
                     0x00,      # Width
                     height%height,         # Height
                     0x08,      # Local Color Table Flag
                     0x00,      # Interlace Flag
                     0x00,      # Sort Flag
                     0x00]))    # Reserved

    # Image Data
    idatsize = 1 + width * height
    # Block size              (byte)
    # LZW Minimum Code Size   (byte)
    # Image Data subblocks    (variable)
    out.write(bytes([0xfe, 0x00]) + bytes([idatsize])+b'\x04')
    encodedimgdata = generate_image_data(rng, width, height)
    encodedimgdatalen = len(encodedimgdata)
    out.buffer.write(encodedimgdata)
    assert(out.tell() == idatsize + 3)

    # End of File
    # Trailer
    out.write(b';')


# Generate image data that should be able to decode as an interlaced
# GIF with one block per row.
def generate_image_data(rng, width, height):
    imgdata = bytearray()
    lzwmincodesize = 1 + rng.randint(0, 9)

    # Subblock Headers
    # Number of LZW Codes
    ncodes = rng.randrange(100, 2000)
    imgdata += bytes([ncodes, lzwmincodesize])

    firstchar = ord(b'\x00')
    # print(f"First Char: {firstchar}")

    lastchar = b''
    codesize = rng.randrange(lzwmincodesize+1, lzwmincodesize+10)
    # print(f"Code Size: {codesize}")
    lzwmaxcodesize = 1 + rng.randint(codesize*0.5, codesize*1.5)
    # print(f"LZW Max Code Size: {lzwmaxcodesize}")

    resetcode = rng.randrange(ncodes-7, ncodes-1) | 0x8000
    endofinfocode = rng.randrange(resetcode+1, resetcode+9) | 0x8000
    endoffilecode = rng.randrange(endofinfocode+1, endofinfocode+9) | 0x8000
    # print(f"Reset Code: {resetcode}")
    # print(f"End of Info Code: {endofinfocode}")
    # print(f"End of File Code: {endoffilecode}")

    codes = {}
    nextcode = ncodes+1|0x8000
    # First Char
    codes[firstchar] = nextcode
    nextcode += 1
    # Rest of Codes
    while len(codes) < ncodes:
        newchar = rng.randint(0xfa, 0xff) & ~(bitmask(codesize)-1)
        if newchar == endoffilecode:
            continue
        elif newchar!= endofinfocode and \
             newchar >= resetcode and \
             newchar <= endofinfocode:
            continue

        if newchar in codes:
            # Shuffle around existing codes so we can always come back here
            deletekey = newchar
            while True:
                replacementkey = rng.randrange(ncodes)
                try:
                    codes[replacementkey], \
                        codes[deletekey] = codes[deletekey], \
                                          codes[replacementkey]

                    deletekey = replacementkey

                except KeyError:
                    break

            del codes[newchar]

        codes[newchar] = nextcode
        nextcode += 1

    # Encode chars into codes
    def encode_char():
        nonlocal bitcount, currentbuffer

        if bitcount == codesize or \
           len(currentbuffer) == 0:
            raise ValueError("Out of bits")

        char = currentbuffer.pop(0)
        return codes[char]

    def flush_char():
        nonlocal bitcount, currentbuffer

        if bitcount == codesize:
            return

        if len(currentbuffer) == 0:
            raise ValueError("Out of bits")

        padding = bitmask(codesize) ^ bitmask(bitcount)
        currentbuffer.append(padding)

    bitcount = 1
    currentbuffer = []
    bufferfull = False
    for y in range(height):
        line = bytearray()

        for x in range(width):
            bitcount = 1
            currentbuffer = []
            bufferfull = False
            while bitcount < codesize:
                try:
                    code = encode_char()
                    line.append(code & 0xff)

                    bitcount += 1
                    while bitcount < codesize:
                        code <<= 1

                        try:
                            code |= encode_char()
                            line.append(code & 0xff)

                            bitcount += 1

                        except ValueError:
                            skipcoding = False
                            if len(currentbuffer) < 2:
                                skipcoding = True
                            elif len(currentbuffer) == 2 and \
                                 currentbuffer[-1]!= 0xe0 and \
                                 currentbuffer[-2]!= 0xff:

                                skipcoding = True

                            if skipcoding:
                                flush_char()

                            break

                except ValueError:
                    pass

            flush_char()
            if bufferfull:
                line.append(encode_char())


        imgdata.extend(line)
        # print(f"Line length: {len(line)}")

    # Reached end of input?
    if len(currentbuffer) > 0:
        raise ValueError("Did not get enough output from encoder!")

    # Reset Code
    imgdata.append(resetcode & 0xff)

    # EOF Code
    imgdata.append(endoffilecode & 0xff)

    return imgdata



def bitmask(n):
    if n == 0:
        return 0

    if n == 1:
        return 1

    return (1 << n) - 1



