#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Read global color table if present
    if gctlen != 0:
      gct = rng.read(gctlen)
      out.write(gct)

    while True:
        blocktype = rng.read(1)[0]
        if blocktype == 0x3b: # stop marker
            break
        elif blocktype != 0x2c: # unknown block type
            continue

        # Read the rest of the extension
        extheader = rng.read(6)
        out.write(extheader)
        
        # Parse extension header
        hasgct   = (extheader[0] & 0x80) != 0
        hasimgdata = (extheader[0] & 0x08) != 0
        size    = int(struct.unpack(">H", extheader[2:])[0]) + 1
        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        if hasgct:
          gctsize     = int(struct.unpack("B", rng.read(1))[0])
          gct         = []
          for i in range(gctsize):
             rgb                  = []
             for j in range(3):
                rgb.append(ord(rng.read(1)))
             gct.append(rgb)
          out.write(bytes([gctsize]))
          for item in gct:
              out.write(bytes([(item[i // 3] >> ((2 - i % 3)*8)) & 255
                               for i in range(9)]))

        # Determine duration
        # TODO: This is actually an unsigned integer but it seems like
        #       the maximum that can be stored in a byte is 255 so just use this?
        # dur = abs(float(rng.read(1))) / float(255)
        dur = ord(rng.read(1))

        # Read LZW min code size
        lzws = rng.read(1)
        out.write(lzws)

        # Read subblocks until end marker or max size reached
        lzwmax = 2**11-1
        lzwhash = [None]*lzwmax
        lzwval = None
        lzwcnt = 1
        lzwmincodesize = int(lzws)
        lzwnextcode = int(1 << lzwmincodesize)
        oldcode = None
        img = []
        idx = 0
        while True:
            # Read next byte of compressed data
            cbyte = rng.read(1)

            # End of stream
            if len(cbyte) < 1: 
                break
            
            # Unpack byte as little endian
            cbyte = int(struct.unpack("<B", cbyte)[0])

            # For each bit starting with msb
            for msbidx in range(7,-1,-1):

                # Shift the previously read bits left to get the current msb
                val = cbyte >> msbidx
                
                # If there's no new information in this msb then skip
                if val == 0:
                    continue

                # A new value is available now!
                # If we have settled on a previous value before then we should write
                # the pair as a code and reset our hash table. Otherwise store it in case
                # we need it again when another valid one is found later.
                if val <= lzwcnt:

                    # A zero means its the last value in the table and should be written 
                    # unconditionally
                    if val == 0:
                        # Output the tuple
                        img.append(tuple(lzwhash[:idx+1]))

                        # Reset the hash table
                        lzwhash = [None]*lzwmax 
                        lzwcnt = 1
                        lzwval = None
                        lzwnextcode = int(1 << lzwmincodesize)
                        oldcode = None
                    else:
                        # Write existing value
                        # Output the tuple
                        img.append(tuple(lzwhash[:idx+1]))
                        
                        # Set up hash table entry
                        lzwhash[idx] = lzwhash[val-1]
                        
                        # Adjust count
                        lzwcnt -= 1
                        lzwval = val
                else:
                    # The value is undefined at this point!
                    # If we have settled on a previous value before then we should emit it
                    # followed by the special "end of information" sequence which is always
                    # the same...
                    if lzwval is not None:
                        # Output the tuple
                        img.append(tuple(lzwhash[:idx+1]))

                        # Special end of info code
                        img += [(oldcode,)]
                        
                        # Reset the hash table
                        lzwhash = [None]*lzwmax 
                        lzwcnt = 1
                        lzwval = None
                        lzwnextcode = int(1 << lzwmincodesize)
                        oldcode = None
                        
                    #... otherwise we look ahead to see what comes next
                    lookahead = rng.read(1)
                    if len(lookahead) < 1:
                        raise Exception('Expected more bytes')
                    
                    # Unpack next byte
                    lookahead = int(struct.unpack('<B', lookahead)[0])
                    
                    # If it is larger than the minimum code size then it must be part of
                    # a new code so we just store it alongside whatever else was already
                    # received
                    if lookahead > lzwmincodesize:

                        # Store value
                        lzwhash[idx] = lookahead
                        
                        # Increase count and adjust min code size
                        lzwcnt += 1
                        if lzwcnt >= lzwnextcode:
                            lzwnextcode <<= 1
                            lzwmincodesize += 1
                        
                        # Look for upper limit of huffman codes
                        if lzwval is None or lzwval >= lzwnextcode: 
                            lzwval = lookahead
                        else:
                            # See if we should write any tuples yet
                            while lzwval is not None:
                                if lzwcnt <= lzwmax:
                                    img.append(tuple(lzwhash[:idx+1]))

                                # Remove substituted value
                                lzwhash[idx] = lzwhash[lzwval-1]
                                
                                # Decrease count
                                lzwcnt -= 1
                                lzwval = lzwhash[idx]

                            # Now just update hash table with the looked-up value
                            lzwhash[idx] = lookahead
                            
                        # Keep track of the newly inserted value
                        oldcode = lookahead 
                    else:
                        # We don't know anything about this value except that it fits within
                        # the minimum code size so we mark it as such
                        # Increase count and adjust min code size
                        lzwcnt += 1
                        if lzwcnt >= lzwnextcode:
                            lzwnextcode <<= 1
                            lzwmincodesize += 1
                        
                        # Update stored value
                        lzwhash[idx] = lookahead
                        
                        # Clear lookup value
                        lzwval = None
                    
                print("V:", lzwval, "E:", lzwhash[idx], 
                    "C:", lzwcnt, "N:", lzwnextcode, "M:", lzwmincodesize)
                idx+=1
            
        # Image separator blocks are not allowed to contain more than 
        # 255 bytes, so adjust accordingly
        while len(img) > 255:
            del img[round(len(img)/2)-10:(round(len(img)/2)+10)]

        # Write total number of bytes used for this frame
        # Not including the ending 0xff byte
        out.write(bytes([len(img)]))

        # For each tuple in the image
        for tup in img:
            # Write the tuple's length
            # Make sure length is never bigger than 255 so this wouldn't go negative
            # Also make sure it isn't smaller tha 2 because that would be silly
            tupnum = round(((tup[-1]-2)/255)**2)-1
            tupnum = tupnum if tupnum > 0 else 0
            out.write(bytes([tupnum]))
            
            # Write actual values
            for item in tup[:-1]:
                out.write(bytes([item]))
                
        # Write duration
        out.write(bytes([dur]))
