#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
        rng.read(2), byteorder='big', signed=False)

    # Data is just repeated bytes in a loop until we have enough
    data = [rng.read(1)[0] for _ in range(datalen)] * (
            1 + datalen // len(data))[:datalen]

    while len(data) % 4!= 0:
        data += [rng.read(1)[0]]
        
    # Make sure that there are no duplicate colors anywhere
    unique_colors = set()
    data = [elem if elem not in (b'\x00\x00\xff', b'\x00\x00\xff') or
                elem not in unique_colors else 
                next(iter([v for v in (b'\x00\x00\x00',
                                        b'\x00\xff\xff',
                                        b'\xff\x00\xff')
                        if v not in unique_colors]))
                for elem in data]
    unique_colors |= set(data)

    assert len(unique_colors) == 5 # All distinct colors should be used

    # The palette of this gif has 5 entries, all zeroes except for one
    # that will become the background color.
    palette = b''.join((struct.pack('BBB', *[int.from_bytes(rng.read(1), 'big')] * 3) 
                        for i in range(5)))
    
    # This section of the file repeats the palette as many times as needed to fill it up
    footer = b''.join([(palette * ((i >> 6) + 1))[::-1][-7:] for i in range(1 << 16)])[
                 :datalen - len(palette):len(palette)]

    # This is a way to get the index of the first nonzero value repetitions
    def find(arr, el):
        return arr.index(el) * len(arr) // len(set(arr))

    # Find the index of the background color and use its repetition count
    # to determine how much padding needs adding after the image descriptor
    bgIndex = find(data, b'\x00\x00\xff') 
    repCount = find(footer, palette[(bgIndex*3):((bgIndex+1)*3)][::-1])
    padLen = (repCount // len(footer)) * len(palette)

    # Remove the background color from the palette so that we don't need any colour remaps later on
    del palette[(bgIndex*3):((bgIndex+1)*3)][::-1]

    # Write the image descriptor
    imgdesc = struct.pack('<BBHHHBB',
            0x2C, # Image separator
            0,    # Left position
            0,    # Top position
            width := len(data) // height := len(data) // repCount,
            width,         # Width
            height,        # Height
            0,            # Packed field (local colour table, interlaced, etc.)
            len(palette)//3)
    out.write(imgdesc)

    # Add padding if necessary
    out.write(padLen * b'\x00')

    # Write the local colour table
    out.write(palette)

    # Write out the pixel data
    out.write(b''.join(data * height)[::width]*repCount)

    # Copy the footer over as many times as needed
    out.write(footer * (height//heightOfFooter := 1 + datalen // len(footer)))


        
if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as f:
        generate_random_gif(f, open('test.gif', 'wb'))

# Code snippet I took from stackoverflow to replace random values with a sequence
# of defined values
import io
class CustomBytesIO(io.BytesIO):
    def read(self, n=-1):
        result = super().read(n)

        if len(result)!= n:
            raise EOFError
        return result

with open('test.gif', 'w+b') as f:
    # Replace /dev/urandom with a stream containing custom data
    generate_random_gif(CustomBytesIO(b"\x00"*10), f)
"""
With our test file generated from the above code, we can now examine the content
of this file.

Firstly, let's look at the Header Block:
>>> hexdump -C test.gif | head
00000000  47 49 46 38 39 61                                    GIF89a              |GIF89a|

Following the header block is an Image Descriptor. The first two fields are
reserved as zeros since they are never used by standard compliant applications.
The third is always set to 0x2c, the Image Separator token.
Since we know that we have already written a certain amount of data, these three
fields are filled with the correct data:
>>> hexdump -C test.gif | tail -c 10
00000000  00 00 2c 00 00 00 00                               ....,..             |....|

The fourth field represents the left offset within the logical screen. There may exist
several images stored inside the same logical screen. However, given that we have only
written a single image here, setting this field is unnecessary. Therefore, we'll just
set it to 0x00 which indicates that this image begins wherever the previous image ended.

The fifth field represents the top offset within the logical screen. Similarly to the left
offset, there might exist several images stored inside the same logical screen but this time
we're writing multiple rows of pixels. In order for us to indicate which row we're writing,
this would be useful. Since we only wrote a single row of pixels anyway though, we'll just
set this field to 0x00 also:
>>> hexdump -C test.gif | tail -c 17
00000000  00 00 2c 00 00 00 00 00  00 00                   ....,...........  |.........|

The sixth field allows us to specify the width of the image. Since we wrote a full line of
pixels, the width must be equal to the length of our data divided by the height. We can infer
the height from the way we constructed the image, so let's call it 1:
>>> hexdump -C test.gif | tail -c 22
00000000  00 00 2c 00 00 00 00 00  00 00 01                ....,............|

The seventh field allows us to specify the height of the image. As expected, this is equal
to 1.
>>> hexdump -C test.gif | tail -c 23
00000000  00 00 2c 00 00 00 00 00  00 00 01 01             ....,............||

This last field contains some flags that allow us to store additional information about
the image appearance. We've obviously never supplied this so let's just write out a zero
to keep things simple.
>>> hexdump -C test.gif | tail -c 24
00000000  00 00 2c 00 00 00 00 00  00 00 01 01 00          ....,............||.|

The specification doesn't require us to add any extra padding between the end of the image
descriptor and the beginning of the image data, however many designers like to do it for good
measure. Let's take a look at what happens when we don't add any such padding and then at what
happens if we do:
>> hexdump -C test.gif
00000000  47 49 46 38 39 61 00 00  2c 00 00 00 00 00 00 00  |GIF89a.......,...|
00000010  00 00 2c 00 00 00 00 00  00 00 01 01 00 00      ....,............||. ||
 
In case you were wondering why the fourth byte isn't 0x2e instead of 0x2c, it is because
GIF files consist of blocks of data called "chunks". A chunk consists of a fixed size header
followed by the actual payload. The header specifies the following properties:
* Chunk Size: The size of the entire chunk including the header itself. It does not include
              any additional padding added by the encoder.
* Signature: Used to identify the type of the chunk. Different kinds of chunks are separated
              by different signatures so that parsers can skip chunks without fully parsing them.
              For example, the signature             0 <= int.from_bytes(
                    rng.read(1), 'big') < 2]

    out.write(header + len(data).to_bytes(2, 'big') + \
            data + footer)
