#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 2 bytes of length of block (N).
#       N is either 0x7D or 0xFF, representing an unlimited size or a limited one respectively.
#   3. if N == 0x7D:
#       - 1 byte for the minimum code size to use. This should be randomly chosen from [2, 9],
#         inclusive.
#     else:
#       - no data present in this case
#   4. X * Y blocks:
#       4-byte header:
#           - 1 bit indicating whether there's another image after this one or not.
#           - 3 bits are reserved and must always have value 0b000.
#           - 3 bits represent the color table flag (CTF):
#               0b000 - No local colour map exists. LZW compression is used instead.
#               0b001 - A global colour map exists that applies to all images within this file.
#             The palette and its size will come later.
#               0b100 - Uses an indexed colour map, which has already been defined.
#               0b101 - Reserved.
#               0b110 - Reserved.
#               0b111 - All colours are transparent; only a single pixel of colour index 0 is
#                       stored. However, it is still necessary to specify a valid palette
#                       with at least one entry.
#            - Lower 5 bits indicates the number of bits per colour component:
#                0b00000 - Greyscale.
#                0b00001 - Colourmap. Each pixel uses one byte of storage for both red, green and
#                           blue components.
#                0b00010 - Truecolour. Each pixel uses three bytes of storage for red, green and
#                           blue components.
#                0b00100 - Black/white. Each pixel is only one bit wide.
#              This value can also take the values 0b11111, though they don't indicate anything;
#              for compatibility purposes we're going to treat them as truecolour anyway.
#       1-byte LZW minimum code size. Only required when CTF!= 0b000 ("no local colour map").
#       3-byte RGB colour for each colour in colourtable:
#           If greyscale, then only R component is stored.
#           If colourmap, then R, G & B components are stored.
#           Otherwise, all components are stored.
#       6-bit indices of pixels
def generate_random_gif(out: BinaryIO) -> None:
    # write magic number
    out.write("GIF89a".encode())

    # determine block size to use
    blocksize = 0xFF + 0x7D
    
    lzwmincodesize = 2
    numcolours = 7
    colourbits = { 1 : 0b00000,
                   3 : 0b00010,
                   7 : 0b00100 }.get(numcolours, 0b00001)
    colourmapflag = 0b000 # "No local colour map"
    backgroundcolourindex = 0
    transparentcolorindex = 0
    gct = [[i*255//numcolours for i in range((colourbits+7)//8)]]

    # write leading control block
    bits = []
    appendBits = lambda b, n=1: bits.extend([int(x) for x in '{:0{}b}'.format(b,n)])
    appendBit = lambda b: appendBits(b, 1)
    appendNumericalBits = lambda b, n: appendBits(b, n)
    def flush():
        while len(bits) >= 8:
            out.write(struct.pack('B', int(''.join(['1' if b else '0' for b in bits[-8:]]), 2)))
            del bits[-8:]
        assert len(bits) < 8
    appendNumericalBits(blocksize, 7) # blocksize
    if blocksize == 0x7D:
        appendNumericalBits(lzwmincodesize, 3) # mincodesize
    flush()
    
    # write image descriptor block
    appendingimages = False
    interlaced = False
    sortcolours = False
    for y in range(numcolours):
        for x in range(numcolours):
            bit = ((y*numcolours)+x) % 2
            appendBit(bit)
        appendBit(0)
        appendingimages |= bit
    appendNumericalBits(colourmapflag |
                        colourbits,
                        3) # colormap flag
    if colourmapflag == 0b000:
        appendNumericalBits(lzwmincodesize, 3) # mincodesize
    appendNumericalBits(backgroundcolourindex, 8) # background colour index
    appendNumericalBits(transparentcolorindex, 8) # transparent color index
    width = height = min(numcolours, 255)
    packedwidthheight = ((width << 8) & 0xFF00) | \
                        (height & 0xFF)
    appendNumericalBits(packedwidthheight, 16) # dimensions
    appendNumericalBits(0, 3) # left offset
    appendNumericalBits(0, 3) # top offset
    appendNumericalBits(0, 1) # width ratio & flags
    appendNumericalBits(0, 1) # height ratio & flags
    appendNumericalBits(interlaced, 1) # interlace flag
    appendNumericalBits(sortcolours, 1) # sort flag
    flush()

    # write palette
    if colourmapflag == 0b000:
        raise Exception() # TODO
    elif colourmapflag == 0b010:
        raise Exception() # TODO
    elif colourmapflag == 0b100:
        raise Exception() # TODO

    # write compressor tables
    if colourmapflag == 0b000:
        raise Exception() # TODO
    elif colourmapflag == 0b010:
        raise Exception() # TODO
    elif colourmapflag == 0b100:
        raise Exception() # TODO
        
    # write image data
    amagicnumber = 0
    abitsremaininginthisbyte = 0
    abitssincelastcontrolbyteoutput = 0
    lastacode = 0b111
    codesseensincelastclear = set()
    globalcolourmapentrycount = 0
    globalcolourmap = {}
    packpixelbytes = lambda r,g,b:(r<<16)|(g<<8)|b
    unpackpixelbytes = lambda rgb:(rgb>>16&0xff, (rgb>>8)&0xff, rgb&0xff)

    previousimagedata = None
    for y in range(height):
        rowcodebook = {}
        for x in range(width):
            newpixeldata = tuple([int(((x*(y%2))+y)//2==p)*255 for p in range(numcolours)])

            polypixels = [(newpixeldata[j], newpixeldata[j+1], newpixeldata[j+2])
                          for j in range(0, len(newpixeldata)-2, 3)]
            
            encodedpixels = sum([(polypixels[i][c]*(len(polypixels)**i))
                                 for c in range(3)
                                 for i in reversed(range(3))],
                                [])
                
            mainpolypixel = newpixeldata[:3]
            code = rowcodebook.setdefault(mainpolypixel,
                                         hex(globalcolourmapentrycount))
            globalcolourmapentrycount += 1
                    
            ac = None
            rc = None
            for i in range(len(encodedpixels)):
                v = encodedpixels[i]

                if isinstance(v, str):
                    rc = v
                else:
                    ac = ('{:0{}b}'.format(v, 6))[::-1]
                
                if ac is not None:
                    for _ in range(abitssincelastcontrolbyteoutput):
                        flush()

                    abitssincelastcontrolbyteoutput = 0
                    
                    if amagicnumber > 0:
                        flush()

                    amagicnumber <<= 1
                    if ac[0]:
                        amagicnumber |= 1
                    abitsremaininginthisbyte = 6
                    continue
                    
                if rc is not None:
                    if rc in rowcodebook:
                        code = rowcodebook[rc]
                    else:
                        print(rowcodebook)
                        raise Exception('{} ({}) was never seen before in this line!'.format(rc, rowcodebook[rc]))
                
                    if rc not in codesseensincelastclear:
                        output = int(rc, 16)
                        
                        while abitsremaininginthisbyte <= 6:
                            flush()
                            
                            amagicnumber <<= 1
                            abitsremaininginthisbyte -= 1

                        amagicnumber <<= abitsremaininginthisbyte
                        amagicnumber |= (output >> (abitsremaininginthisbyte-(6-allbitsremainingtooutput)))
                        
                        abitsremaininginthisbyte -= (6-allbitsremainingtooutput)
                        flush()
                            
                        output &= (1<<(6-allbitsremainingtooutput)) - 1
                        if not abitsremaininginthisbyte:
                           