#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
    
        # Break apart fieldbyte with bit masks
        fbi = (fieldbyte & 64)!= 0
        fcfi = (fieldbyte & 32)!= 0
        frle = (fieldbyte & 16)!= 0
        fcce = (fieldbyte & 8)!= 0
        fbci = (fieldbyte & 4)!= 0
        flzw = (fieldbyte & 2)!= 0
        fqoi = (fieldbyte & 1)!= 0


        # If there's no FBI, next byte must be background index; add that to output buffer
        if not fbi:
            bgidx = ord(rng.read(1))
            out.write(struct.pack('B', bgidx))
        

        # Next byte must be pixel aspect ratio, so just add one to output buffer
        pastratio = ord(rng.read(1)) + 1
        out.write(struct.pack('B', pastratio))
        
        

        # Get the size of image data in bytes
        datalen = read_size(rng)

        
        # Handle plain-text extension blocks
        if fcfi:

            # Write 1 or more plain text subblocks to output stream
            for i in range(datalen+1):
                out.write(rng.read(1))

            continue


            
        # Otherwise, handle an image

        # If there's no FCCE, next byte must be local palette size
        if not fcce:
            lps = ord(rng.read(1))
            

            # Add this many bytes to output buffer
            lpt = rng.read(lps*3)
            out.write(lpt)



        # If there's no FQOI, next byte must be local palette refinement method
        if not fqoi:
            lpqm = ord(rng.read(1))


        # Next two bytes are reserved for future use
        resv = rng.read(2)
        
        
        # Now for the tedious part: getting all the LZW codes
        
        # Create empty array to store codes in
        codes = []
        
        # Set initial code sizes based on LZW minimum allowed codesizes per specification
        mincodes = [
            2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 
            16, 20, 24, 28, 32, 40, 48, 56, 64, 
            80, 96, 112, 128, 160, 192, 224, 256 
        ]

        # Read first code from input as initial code
        curcode = ord(rng.read(1))
        j = 0
        
        # Fill codes array with initialized values based on minimum code size
        while len(codes) < mincodes[(curcode >> 8)]:
            codes.append(j)
            j += 1

        # Initialize clear code entry and end code entries
        clrcod = 0xFFFF
        encod = 0xFFFE
        badcod = 0xFFFD
        rsvcod = 0xFFFC

        # Read remainder of codes in 12-bit chunks until EOF
        while len(codes) < 4096:
            
            # Get next chunk from input
            numbytes = mincodes[(curcode >> 8)]
            
            # Large enough for 12 bits?
            if numbytes >= 2:
                highbyte = int(((ord(rng.read(1)) << 4) | (curcode >> 4)))
            else:
                highbyte = -1
                
            lowbyte = int((((ord(rng.read(1)) & 0xF) << 8) | curcode))

            # Combine them together for new code value
            newcode = (((highbyte << 8) | lowbyte) if highbyte > 0 else lowbyte)
            
            # Make sure key doesn't already exist before adding it
            if newcode not in codes:
                codes.append(newcode)

            # Update current code value being processed    
            curcode = newcode


        # Sort codes by frequency usage, starting with most frequent
        sortedcodes = list(zip(*sorted(enumerate(codes), key=lambda x : -x[1])))[0]

        # Find where the clear code would fit in the order
        clrcodfnd = sortedcodes.index(clrcod) if clrcod in sortedcodes else None

        # The clear code gets its own spot because it always has to go at the start of the GIF
        while clrcodfnd > 0:
            temp = codes[clrcodfnd]
            del codes[clrcodfnd]
            sortedcodes.insert(0, temp)
            clrcodfnd -= 1

        
        
        # Find where the end code would fit in the order
        encodfnd = sortedcodes.index(encod) if encod in sortedcodes else None

        # The end code gets its own spot because it always has to go at the end of the GIF
        while encodfnd < (len(codes)-1):
            temp = codes[encodfnd]
            del codes[encodfnd]
            sortedcodes.insert(-1, temp)
            encodfnd += 1


        # Find where the bad-code replacement code would fit in the order
        badcodfnd = sortedcodes.index(badcod) if badcod in sortedcodes else None

        # Same deal here -- all bad-data codes assumed to be identical
        while badcodfnd < (len(codes) - 1):
            temp = codes[badcodfnd]
            del codes[badcodfnd]
            sortedcodes.insert(-1, temp)
            badcodfnd += 1


        # Find where the reserved-code replacement code would fit in the order
        rsvcodfnd = sortedcodes.index(rsvcod) if rsvcod in sortedcodes else None

        # Same deal here -- reserved-data codes assumed to be identical
        while rsvcodfnd < (len(codes) - 1):
            temp = codes[rsvcodfnd]
            del codes[rsvcodfnd]
            sortedcodes.insert(-1, temp)
            rsvcodfnd += 1


        # Remove any duplicate codes
        uniquecodes = [c for c in set(codes)]

        # Then insert back each unique code in their original place
        for c in sorted(uniquecodes, reverse = False):
            codes.insert(sortedcodes.index(c), c)

            
        
        # At this point, codes[] contains our complete code mapping
        # We can now begin writing the LZW compression codes to the file

        # Codes are written backwards, least significant digit first
        # But we want the most significant first, so swap them around
        codes = [(c >> 8) | ((c & 0x00FF)<<8) for c in codes]
        codes.reverse()


        # Write EOF code to output stream (if necessary)
        eofcod = codes[-1]
        write_code(eofcod, out)


        # While we still have data left to write...
        while len(codes)!= 0:
            # Write next code to output 
            codenum = write_code(codes[0], out)

            # Delete code we wrote from list
            del codes[0]


            # Calculate what was supposedly the prefix of this code
            # This will either be the previous valid code we added,
            # OR the clear code that just got inserted above (clrcd)
            try:
                if (codenum!= 0):
                    prevalcode = codes[0]
                else:
                    prevalcode = clrcod
                        
            except IndexError:
                pass
            
            

            # Find index of prefix code within list of codes
            try:
                prevalcodfnd = codes.index(prevalcode)
            except ValueError:
                prevalcodfnd = -1

            # If that code actually exists in the list of codes...
            if prevalcodfnd!= -1:
                
                # Replace it in the list with the actual code we just wrote
                codes[prevalcodfnd] = ((prevalcodfnd << 8) | codenum)
                

       

        # Pad final frame to 2 byte multiple
        padding = 2 - (out.tell() % 2)
        if padding == 2:
            padding = 0
        padbytes = rng.read(padding)
        out.write(padbytes)




# Given a byte count and a binary reader object, reads n bytes
# of data and returns them as an integer.
def read_size(
        rng: BinaryIO
    ) -> int:
    return (ord(rng.read(1)) << 8) | ord(rng.read(1))





# Given a code number, writes it to the file
# Returns the number of bytes used to represent the code
def write_code(
        code: int, 
        out: BinaryIO
    ):

    # Shift MSB down to position
    msbshift = 8


    # Pull off the leading byte
    leadbyte = 255 & (code >> msbshift)

    

    # while the upper 8 bits aren't null
    while leadbyte > 255:
        
        # Mask off the lower bit and write it to output
        out.write(struct.pack('B', (leadbyte & 254)))

        

        # Shift everything up to make room for another byte
        code =