#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    color_table = bytes([255] * 768)

    # version of GIMP used to create this gif (probably 0)
    version = ord(rng.read(1))
    out.write(version.to_bytes(1, byteorder='little'))

    width = int.from_bytes(rng.read(2), 'big') // 24
    height = int.from_bytes(rng.read(2), 'big')

    lct_flag = bool(int.from_bytes(rng.read(1), 'big'))

    if not lct_flag:
        out.write(color_table) 

    gct_flag = bool(int.from_bytes(rng.read(1), 'big'))
    bgcolor = int.from_bytes(rng.read(1), 'big')

    pixel_info = bytes([(width & 0x7F) | ((height & 0x7F) << 7)])
    out.write(pixel_info)
    
    local_color_table = None
    if gct_flag:
        entry_size = 3 ** int.from_bytes(rng.read(1), 'big')

        num_entries = 2 ** (entry_size - 1)
        
        local_color_table = [
            int.from_bytes(rng.read(3), 'big') for _ in range(num_entries)]

        local_color_table += [(0, 0, 0)] * \
            max((2 ** (entry_size * 3) - len(local_color_table)), 0)

        out.write(b'\xF7')
        out.write(struct.pack('B', entry_size))
        out.write(bytes(local_color_table[i: i+1]) for i in range(num_entries))

    else:
        out.write(bgcolor.to_bytes(1, byteorder='little'))
    

    flags = bytes()
    interlace_flag = bool(int.from_bytes(rng.read(1), "big"))
    sorted_flag = bool(int.from_bytes(rng.read(1), "big"))
    reserved_bits = rng.read(2)
    background_color_index = int.from_bytes(rng.read(1), "big")
    pixel_aspect_ratio = int.from_bytes(rng.read(1), "big")
    global_flags = bytes([
        0, 
        interlace_flag | (sorted_flag << 1), 
        0,
        0 if lct_flag or gct_flag else 1
    ])

    size = 0
    while True:
        frame = []
        size += 1
        code = int.from_bytes(rng.read(1), "big")
        flags += code.to_bytes(1, byteorder="little", signed=False)

        if code == 0xFF and len(frame):
            break

        if code == 0x2C:
            # image descriptor
            left = int.from_bytes(rng.read(2), "big")
            top = int.from_bytes(rng.read(2), "big")
            width = int.from_bytes(rng.read(2), "big") // 24
            height = int.from_bytes(rng.read(2), "big")

            local_color_table_flag = bool(
                int.from_bytes(rng.read(1), "big"))
            
            lct_size = None

            if local_color_table_flag:
                lct_size = 2 ** (int.from_bytes(rng.read(1), "big") + 1)

                frame.append(lct_size.to_bytes(1, "big"))
                frame += list(bytes(local_color_table[:2**lct_size]))
            else:
                pass

            local_gct_flag = bool(int.from_bytes(rng.read(1), "big"))
            bgnd_index = int.from_bytes(rng.read(1), "big") 
            pixel_aspect_ratio = int.from_bytes(rng.read(1), "big") 

            frame += [top.to_bytes(2, byteorder="big"),
                      left.to_bytes(2, byteorder="big"), 
                      width.to_bytes(2, byteorder="big"),
                      height.to_bytes(2, byteorder="big"),
                      bytes([
                          local_color_table_flag | (
                              local_gct_flag << 1),
                          0])]
        
            if local_gct_flag:
                frame += list(
                    bytes(local_color_table[-(2**(lct_size * 3)):]))

            size += sum(map(lambda x: len(x), frame))
            
        elif code == 0x21:
            # extension block
            ext_label = rng.read(1)

            if ext_label == b'\xf9':
                # graphic control extension
                label = ext_label
                size += len(ext_label)

                transposition_flag = bool(int.from_bytes(
                    rng.read(1), "big"))
                user_input_flag = bool(int.from_bytes(
                    rng.read(1), "big"))
                disposal_method = int.from_bytes(
                    rng.read(1), "big")
                
                delay_time = int.from_bytes(rng.read(2), "big") 
                transparent_color_flag = bool(int.from_bytes(
                    rng.read(1), "big"))

                index = None
                
                if transparent_color_flag:
                    index = int.from_bytes(rng.read(1), "big")

                    if index >= 2**(
                        entry_size * 3) and not local_color_table:
                        raise ValueError("invalid transparent color index")

                size += sum([
                    1, 
                    1,
                    1,
                    2,
                    1, 
                    index is not None
                ])
                
                frame.append(label)
                frame.append(transposition_flag.to_bytes(1, "big"))
                frame.append(user_input_flag.to_bytes(1, "big"))
                frame.append([disposal_method, delay_time %
                            0x100, delay_time // 0x100])
                frame.append(transparent_color_flag.to_bytes(1, "big"))
                frame.append(index.to_bytes(1, "big") if index is not None else b'')
                
                
            else:
                raise NotImplementedError("extension blocks aren't implemented :(")


        elif code!= 0x2C:
            raise AssertionError(code)


    data = b''.join(frame)

    out.write(header + data)
    print(f"generated {len(data)}-byte image")
