#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Read global color table if present
    if gctlen != 0:
      gct = rng.read(gctlen)
      out.write(gct)

    while True:
        blocktype = rng.read(1)[0]
        if blocktype == 0x3b: # stop marker
            break
        elif blocktype != 0x2c: # unknown block type
            continue

        # Write the image separator byte `0x2C`
        out.write(bytes([blocktype]))
        
        # Now we can read the rest of the fields in this record
        ihdr = rng.read(9)
        width = int(((ihdr[0] << 8) | ihdr[1]) & 0x3FFF) + 1
        height = int((((ihdr[2] << 8)) | ihdr[3])) + 1
        flags = ihdr[8] # Flags are used for local color tables

        # Check if there is a local color table, or not
        if flags & 0x80!= 0:
          size = int(flags & 0x7f)+1
          localtable = rng.read(size*3)
          out.write(localtable)

        # Size of code block
        size = ((width+7)//8)*(height) 
        numcolors = size//3 - 1
        if numcolors > 0:
          colortable = rng.read(numcolors*3)
          out.write(colortable)

        # Now write image data 
        imgdata = rng.read(size-1)
        out.write(imgdata)