#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
    rng: 'random.Random',
    width: int = -1, height: int = -1, bpp: int = 8, numframes: int = 500):

    global colors
    if (width == -1 or height == -1):
        width = rng.randint(1,799)
        height = rng.randint(1,599)
    
    colors=rng.randint(2 ** min(bpp, 7), 2 ** max(bpp, 7) + 64 )

    outfile="{}.gif".format(numframes)

    with open(outfile, "wb") as f:

        generate_file(f, numframes, width, height, bpp, rng)

# Generate the GIF file from scratch
def generate_file(
    out: BinaryIO, 
    numframes:int=-1, width:int=-1, height:int=-1, bpp:int=-1, rnd:object=None ):


    global colors
    if rnd==None:
        rnd = random.SystemRandom()

    if (width == -1 or height == -1):
        width = get_rndint(out, rnd, 1, 799)
        height = get_rndint(out, rnd, 1, 599)
        
    if not numframes:
        numframes = get_rndint(out, rnd, 1, 30)
    else:
        numframes = numframes-1

    if not bpp:
        bpp = get_rndint(out, rnd, 1, 8)

    
    print("Generating {} frames at size {},{} with {} bits per pixel.".format(numframes+1, width,height,bpp))
    header = bytearray([
        0x47, 0x49, 0x46, 0x38, 
        0x39, 0x61, # GIFFIHDR
        0x00, 0x00, 0x00, 0x00, # WIDTH and HEIGHT fields
        *int2bytes(width), 
        *int2bytes(height), 
        0xf0 | bpp, 0x00])

    paddinglen = bytearray([])
    for _ in range((4-(len(header)%4))%4):
        paddinglen += bytearray([get_rndint(out, rnd, 0, 255)])
    paddinglen *= ((len(header)+len(paddinglen))+3)&~3
    header+=paddinglen
    out.write(header)

    color_table = []
    colors >>= 1 #< CMYK mode!

    while len(color_table)<colors*3:
        r = [get_rndint(out, rnd, 0, 255)] * 3
        if all(v!=r[0] for v in r):
            break

    color_table += r * (colors - len(color_table)//3)

    while len(color_table)>colors*3:
        del color_table[-1]

    assert len(color_table)==colors*3
    for rgb in color_table:
        out.write(bytearray([rgb]))

    dispose_method = 0
    for i in range(numframes+1):
        
        delaytime = get_rndint(out, rnd, 0, 32767)
        d = bytearray([(delaytime>>8)&0xff,(delaytime&0xff)])

        left   = get_rndint(out, rnd, 0, width-1)     
        top    = get_rndint(out, rnd, 0, height-1)    
        width  = get_rndint(out, rnd, 1, width-left)  
        height = get_rndint(out, rnd, 1, height-top) 

        packed_info = 0xe0 | dispose_method << 2

        lzwminsize = get_rndint(out, rnd, 2, 255)
        interlace = get_rndbool(out, rnd)
        localpal = get_rndbool(out, rnd)

        out.write(bytearray([
            0x2c, 
            *d,
            *(int2bytes(left)),
            *(int2bytes(top)),
            *(int2bytes(width)),
            *(int2bytes(height)),
            packed_info,
            lzwminsize ]))

        if localpal:
            
            for l in range(colors*3):
                out.write(bytearray([get_rndint(out, rnd, 0, 255)]))


        if interlace:

            triples = [
                (top//8)*width*(i+1), 
                (top//8)*(width*(numframes-i))+(width//8),
                0]
            for t in reversed(triples):

                for j in range(t % width, width*(height//8)):
                    out.write(bytearray([get_rndint(out, rnd, 0, 255)]))
                
        else:
        
            for yp in range(height // 8):
            
                for xp in range(width // 8):

                    for line in range(8):
                        out.write(bytearray([get_rndint(out, rnd, 0, 255)]))
                    
        out.write(bytearray([0x00]))

    out.write(b'\x3b')
    out.flush()


# Returns an integer between minval incl., maxval excl.
# This is just here for easier mocking elsewhere
def get_rndint(out, rnd, minval, maxval):
    return rnd.randint(minval,maxval-1)


# Returns either true or false based on SystemRandom().randrange(0,2)
# This is just here for easier mocking elsewhere     
def get_rndbool(out, rnd):
    return rnd.randrange(0,2)

if __name__=="__main__":
    import os
    import sys
    import random

    seed=0xdeadbeef
    if "-seed" in sys.argv:
        seed=sys.argv[sys.argv.index("-seed")+1]

    try:
        seed=int(os.path.basename(seed),16)
    except ValueError:
        pass

    try:
        seed = int(seed,0)
    except ValueError:
        seed = hash(str(seed)) & 0xffffffff
    print(seed)
    rnd = random.Random(seed)

    with open(sys.argv[1],'bw+') as out:
        generate_file(out, None,None,-1,-1, rnd)



# Dumps the binary data representing this object so we can read it back later
def dumpobj():
  return raw


def loadobj(data):
  obj = cls.__new__(cls)
  obj._init_from_raw_data(data)
  return obj


# Length encoded integers are represented by one to four bytes. The first byte contains
# the information necessary to decode how many more bytes follow. If msb is set in the high order
# bit then another byte follows which is multiplied with 128 and added to the result thus building
# up a representation of length until the most significant bit is not set.
class LengthedInt:

  @staticmethod
  def from_stream(stream):
    """Read a length encoded integer starting from stream"""
    count = ord(stream.read(1))
    n = count & 0x7F
    if n == count:
      return n
      
    mult = 128
    val = 0
    while True:
      next = ord(stream.read(1))
      count = next
      val += (count & 0x7F) * mult
      if not count & 0x80:
          break
        
      mult *= 128
    return val      
    
  @classmethod
  def write_to_stream(self, stream, value):
    """Write a length encoded integer starting from stream"""
    buf = ArrayBuffer()
    while True:
      buf.append(((value & ~0x7F) >> 7) | 0x80)
      if (value <= 0x7F):
          break
      value &= 0x7F
      value <<= 7
    if value > 0:
       buf.append(value & 0xFF)
    stream.write(buf.as_string())
  

    # Write the leftover bytes before the last descriptor
    print('Writing remaining bytes', end='', flush=True)
    out.seek(-1, 2)
    out.write(paddinglen)
    out.write(struct.pack('<B', numloops))
    out.flush()
