#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # Write magic bytes
    out.write("GIF89a".encode('ascii'))

    # Write screen descriptor (logical screen width = 640px,
    # logical screen height = 480px, color depth = 16bit (565),
    # background color index = None, aspect ratio = No)
    out.write(struct.pack("<HHBBB",  640,    # Logical Screen Width, px
                            480,    # Logical Screen Height, px
                            12,     # Color Depth -> RGB16
                            0b0,    # Background Color Index
                            0b10))  # Pixel Aspect Ratio - 1:1

    # Write global color table with entries ranging from 0 to 255 inclusive
    for i in range(0x1000 * 3 // 4):
        out.write(bytes([rng.read(1)[0] % 0xff]))
    
    # Write Control Block (Graphics Control Extension, sub-block data length = 4, no transperency, delay time = 1 second)
    out.write("\x21\xF9\x04\x01\x00")

    # Write Image Descriptor (Separator)
    out.write("\x2C")

    # Write Image Data Sub-Block (Image Left Position, Top Position, Width, Height)
    out.write(struct.pack("<HHHH",  rng.read(2)[0],       # Image Left Position
                                rng.read(2)[0],       # Image Top Position
                                640 if rng.read(1)[0] else 1 + rng.read(1)[0],     # Image Width
                                480 if rng.read(1)[0] else 1 + rng.read(1)[0]))    # Image Height

    # Write LZW Minimum Code Size byte
    lzwMinCodeSize = rng.read(1)[0] << 3 | rng.read(1)[0] >> 5          # Get first and third bits of first byte
                                                                         # and concat them to get minimum LZSS code size

    # Write local color table with all zeroes (however, this is not strictly necessary as we don't use it)
    for _ in range(0x1000//3):
        out.write('\x00'*3)

    # Write image block data: LZW encoding of all white pixels (whose encoding requires at least one bit because 
    # value 0 does not occur in our custom color table)
    lzwTable = [0]*0x1000                                                 # Initialize empty LZW table
    currByte = b"\xFF"                                                    # Current byte being encoded
                                                                         # NOTE: This must be the first byte due to assertion below!!
    while True:                                                          # Loop until EOF
        nextByte = rng.read(1)                                            # Read 1 byte randomly

        assert len(nextByte) == 1                                         # Only allowed when reading nonzero amount of bytes!!!
        
        if len(lzwTable) < 0x100 or len(currByte)+len(nextByte) >= 256: # Check whether we need to add new 
                                                                    # LZW Table entry
            lzwTable += (1<<lzwMinCodeSize)*[None]                       # Extend current LZW table by minimal required amount
            lzwMinCodeSize += 1                                           # Increase minumum LZWS code size accordingly

        if currByte!= b'\x00':                                            # Add new entry only if last byte was nonzero
            lzwTable[-(ord(currByte) ^ ord(nextByte) & ~ord(currByte))] = currByte+nextByte     
                                                                            # Calculate hash

        if (ord(currByte)+ord(nextByte)&~ord(currByte)) >= lzwMinCodeSize: # We did not have enough bits yet, so continue with decoding
            currByte += nextByte                                          # Append next byte to current byte
        elif lzwTable[currByte][0]:                                       # If nextByte starts a new sequence already in table
            currByte = lzwTable[currByte]                                 # Set current byte equal to currently decoded sequence
        else:                                                            # Else we had enough bits, but the next sequence doesn't start here
            out.write((lzwTable[currByte]+nextByte[:1]).to_bytes(length=2,byteorder='big',signed=False))
                                                                            # Write compressed pixel
            del lzwTable[(ord(currByte)^ord(nextByte)&~ord(currByte))][0]     # Delete already used entry from LZW table
            
            if currByte in lzwTable:                                      # If there are any other sequences starting with same byte
                lzwTable[currByte] += nextByte                            # Concatenate that byte with an existing sequence
                                                                            
            break                                                         # Stop loop

    # Write trailing block
    out.write(";".encode('ascii'))