#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
    out: BinaryIO,
    rng: BinaryIO,
):

    # Write header and logical screen descriptor
    # See https://en.wikipedia.org/wiki/Graphics_Interchange_Format#Logical_screen_descriptor for details of each field
    width = 512 + int((struct.unpack('<B', rng.read(1))[0] >> 4) * (8*16))
    height = 512 + int(((struct.unpack('<B', rng.read(1))[0] & 15) << 4) * (8*16))

    print("Generating {}x{} gif".format(width, height), flush=True)

    out.write('GIF89a'.encode())
    out.write(b'\xF7\x00')                   # Pixel aspect ratio - square pixels
    out.write(struct.pack("<H", width))     # Logical screen width in px
    out.write(struct.pack("<H", height))    # Logical screen height in px
    out.write(b'\x00')                       # Global color table size
    out.write(b'\x00')                       # Background color index
    out.write(b'\x00')                       # Pixel aspect ratio

    # Encode image data as RGB triplets where each component is in range [0-255]
    while True:
        r = ord(rng.read(1)[0]) % 256
        g = ord(rng.read(1)[0]) % 256
        b = ord(rng.read(1)[0]) % 256

        # End of stream means we're done!
        if not len(rng.read(1)):
            break

        # LZW minimum code size
        codesize = min(max(r, g, b) + 1, 12)

        # Write the local image descriptor (row offset = 0, no lzw compression, interlacing disabled)
        out.write(struct.pack("<HHBBB", 0, 0, codesize, 0b11000000 | ((g & 0xf0) > 0xE0), 0b11111100 | ((b>>3) & 0xff)))

        # Data sub-block with one byte indicating the bitmap starting with R, then G, then B.
        out.write(b"\x00" + bytes([r, g, b]))

    out.write(";")

