#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # Now we're going to write 256 descriptors at once
    # Using this loop means we avoid making multiple calls to read()
    for _ in range(257):
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

        # TODO: Should experiment with variations on these values too
        #       see https://github.com/decompme/decompilation-guide/issues/47
        out.write(b'\x2C\x00\x00\x00\x00\x00\x00\xFB\x00\x00\x00')


    last_byte = out.tell() - 1
    # At the moment this code doesn't support appending arbitrary amounts
    assert len(data[last_byte:]) <= 256 * 13
    if len(data[last_byte:])!= 0:
        # If the data[last_byte:] isn't empty adjust
        # what's already been written according to its size
        out.seek(-len(data[last_byte:]), 1)

        # And then append whatever was left from the beginning
        print("Appending", len(data[last_byte:]), 'bytes...', end='', flush=True)
        out.write(data[last_byte:])
        print('done.', flush=True)

    # Finally add the trailer to finish off the output.
    # Normally this would be part of the application logic
    # when writing a gif but since this is intended only
    # for demonstration purposes we do it here directly.
    # See https://github.com/decompme/decompilation-guide/issues/46
    print('Adding the trailer and EOF', end='', flush=True)
    # 2 bytes for signature
    # 1 byte for version
    # 1 byte for flags
    # 7 bits for background color index
    bgindex = rng.read(1)[0] & 0b00000111
    out.write(struct.pack('<6sBBBHB',
                         b'GIFTrailer',
                         0x00,           # Version number?
                         0x00,           # Flags
                         0xFF >> bgindex,# Background color index
                         # 0xFFFF,        # Pixel aspect ratio
                         # 0,             # Frame count
                         0,              # Frame count
                         0               # Offset to frames
                        ))
    # 1 byte for terminator
    out.write(b'\x3B\x00\x00\x00')
    print(' done', flush=True)

