#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    width = next(rng) % 16 + 70 # between [70, 85] inclusive
    height = next(rng) % 24 + 10 # between [10, 33] inclusive

    palette_size = (next(rng) & 0x07) + 1
    if palette_size == 1:
        # case for one-color palettes is always transparent
        palette_size += 1

    background = next(rng)
    # two least significant bits are G and B colors respectively
    foreground = ((next(rng) >> 2) & 0x0F) | \
                 (((next(rng) << 6) >> 2) & 0xF0)

    flags = bytearray([
        0b00101000, # packed field of LZW min code size (3),
                    # transparency color index (transparent=palette 0)
        0b00000000, # reserved fields set to zero
        palette_size - 1,
        background,
        foreground])

    # append control block with specified header and data length
    out.write(struct.pack('<6s2B', b'GIF89a', *flags))

    # append global color table which consists simply of three bytes per entry
    # since we're only using one-byte RGB values, we just use the same value
    # every time
    for i in range(3):
        out.write(struct.pack('BBB',
            *(next(rng) for _ in range(palette_size))))

    # append logical screen descriptor
    out.write(struct.pack('<2H', width, height))
    
    # append application extension to specify that there's no color table
    # (used as an optimization by some parsers such as ImageMagick)
    out.write(struct.pack('<2B', 0xFF, 0xCE))
    out.write(bytes((0x00,) * 11))

    # append graphic control extension to disable interlacing and
    # explicitly specify that background should be used instead of transparent
    # when displaying this frame on top of another frame
    out.write(struct.pack('<2B', 0xFF, 0x21))
    out.write(struct.pack('<2B', 0b10110000, 0x00))

    # append image separator
    out.write(b'\x2C')

    # append local color table specification
    out.write(struct.pack('<3B', 0b00000001, palette_size - 1, background))

    # insert plain text subblock after its definition
    out.seek(-1, io.SEEK_CUR)
    plain_text_block_position = out.tell()
    out.write(struct.pack('<BB', 0x01, 0x00))
    out.write(b"\n\t")

    written = 0
    while True:
        symbol = next(rng)

        if not symbol: continue # nul character terminates string

        if symbol > palette_size:
            # symbol does not exist yet so it must be larger than the highest
            # symbol currently defined, hence it's added to the dictionary

            # in order to achieve compression, symbols are grouped together
            # into chunks called "substrings"
            dictionary_index = len(dictionary) - 1
            if not dictionary or dictionary[-1][0]!= symbol:
                dictionary.append((symbol, []))

            substring = dictionary[dictionary_index][1]
            chunk_length = max(min(len(substring), 127),
                               next(rng) % 128)
            
            substring[:] = [(dictionary_index, chunk_length)]

            chunk = (symbol >= 2**chunk_length)
            chunk |= -(symbol < 2**(chunk_length-1)) << 7

            out.write(bytes((chunk)))
            written += 1
        else:
            # symbol already exists in our batch of symbols; find its position
            # within the list of existing dictionaries (requires linear search)
            found = False
            for pos, substring in enumerate(dictionary):
                if substring[0] == symbol:
                    found = True

                    chunk_length = max(min(len(substring[1]), 127),
                                       next(rng) % 128)
                    
                    substring[1].append((pos, chunk_length))

                    chunk = (symbol >= 2**chunk_length)
                    chunk |= -(symbol < 2**(chunk_length-1)) << 7

                    out.write(bytes((chunk)))
                    written += 1

                    break

            assert found, "Should have found a matching symbol!"

        if written % 127 == 0:
            # check whether end of block has been reached -- in LZW, each
            # new substring can start at any point where the last substring
            # ended -- so simply testing whether we've written more than
            # 127 characters means we may be in the middle of a new substring
            out.write(b"\n\t")
   