#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `fout` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 2 bytes of length of block (N).
#   3. N bytes of block.
def generate_random_gif(
        rng: BinaryIO, fout: BinaryIO
    ) -> None:
    fout.write(b"GIF89a")

    # Find an admissible length for the main image block.
    while True:
        try:
            imglen = rng.randint(50, 600)

            tmp = rng.tell()
            rng.seek(-imglen - 2, 1)

            datalen = rng.read(2)
            rng.seek(tmp)

        except Exception as e:
            print("could not seek to end", e)
            raise

        if not (struct.unpack("<H", datalen)[0] >= imglen):
            break

    fout.write(struct.pack('<H', imglen))

    # Generate the actual main image block with given length.
    data = rng.read(imglen)
    fout.write(data)

    # Add one or more extension blocks with random lengths and contents.
    num_exts = rng.randint(0, 5)
    for _ in range(num_exts):
        ext_type = rng.choice([
            0xf9,   # Graphic control extension
            0xff,   # Application extension
            0xfe,   # Comment extension
        ])

        # The size can be at most 255 bytes.
        if ext_type == 0x0f:  # Plain text extension
            txtlen = rng.randint(0, 255)
            txtlen = min(txtlen, (rng.size - rng.tell()))
            txt = rng.read(txtlen)

            fout.write(struct.pack('BBB',
                0x21,
                ext_type,
                len(txt) | 0x80
            ))
            fout.write(txt)
        
        else:
            # Extension types 0xf9 or 0xff have two subblocks. One is fixed-length, 
            # whereas the other has a single byte length followed by that many bytes.
            ext_id_subblock = [0x0b] \
                               + list(range(0xf7, 0x100)) \
                               + ([0x00] * 3)
            
            ext_id = rng.randint(0x0b, 0xff)

            subblock_type = ext_id_subblock[ext_id]
            subblock_len = rng.randint(0, 255) if subblock_type in (0x0b, 0x0c) \
                           else rng.randint(0, 255) & ~0x80

            fout.write(struct.pack('BBBB', 
                0x21, 
                ext_type, 
                ext_id, 
                subblock_type
            ))
            fout.write(subblock_len.to_bytes(1, 'big')
                       if subblock_type!= 0x0b else b'')
            fout.write((b'\x00\x00'
                        if subblock_type in (0x0b, 0x0c) 
                        else b'').ljust(subblock_len, b'\xff'))