#!/usr/bin/env python3


from PIL import Image as pillowImage

from typing import Optional, Tuple, Union
from io import BytesIO

# Generates a random image into `byteio` using the
# random number generator `rng` (default /dev/urandom)
def generate_random_image(
        rng: BinaryIO=None, byteio: BinaryIO=None,
        w: int=500, h: int=500, colorspace: str='RGB', mode: str='P'
    ) -> None:

    if rng is not None and type(rng)!= bytes or \
       byteio is not None and type(byteio)!= BytesIO:
        raise TypeError('Invalid arguments; either pass in the function to call '\
                        '(for example from open("/dev/urandom", "rb") or an instance of '\
                        f'the class {BytesIO.__name__}.')


    if rng is None:
        rng = BytesIO() # Use the default rng for now...
        rng.seek(0)

    if byteio is None:
        byteio = BytesIO() #...and output directly to memory

    modes = ['1', 'L', 'LA', 'P', 'PA', 'RGB', 'RGBA']

    if colorspace == 'RGB':
        if mode not in modes[:7]:
            raise ValueError(f"Invalid mode '{mode}'; only valid options are "\
                             f"{', '.join([\'%s\'' % s for s in modes[:7]])}")

        img = pillowImage.new(colorspace, (w, h))

        for y in range(h):
            for x in range(w):
                r = rng.read(1)[0] >> 4 | rng.read(
                    1)[0] << 4 & 0xF0
                g = rng.read(1)[0] >> 4 | rng.read(
                    1)[0] << 4 & 0xF0
                b = rng.read(1)[0] >> 4 | rng.read(
                    1)[0] << 4 & 0xF0

                rgbvalue = tuple((r, g, b))[modes.index(mode)]

                img.putpixel((x, y), rgbvalue)

    elif colorspace == 'RGBA':
        if mode not in modes[2:]:
            raise ValueError("Can only generate RGBA images with modes LLA or RGB")

        img = pillowImage.new(colorspace, (w, h))

        for y in range(h):
            for x in range(w):
                r = rng.read(1)[0] >> 4 | rng.read(
                    1)[0] << 4 & 0xF0
                g = rng.read(1)[0] >> 4 | rng.read(
                    1)[0] << 4 & 0xF0
                b = rng.read(1)[0] >> 4 | rng.read(
                    1)[0] << 4 & 0xF0

                if mode == 'LLA':
                    luma = rng.read(1)[0] >> 4 | rng.read(
                        1)[0] << 4 & 0xF0

                    rgbavalue = tuple(
                            (luma, luma, luma, r, g, b))[modes.index(mode) - 2]
                else:
                    a = rng.read(1)[0] >> 4 | rng.read(
                        1)[0] << 4 & 0xF0

                    rgbavalue = tuple((r, g, b, a))[modes.index(mode)-2]

                img.putpixel((x, y), rgbavalue)

    elif colorspace == 'Grayscale':
        if mode not in modes[1:3]:
            raise ValueError("Only modes LA and L are supported.")

        img = pillowImage.new(colorspace, (w, h))

        for y in range(h):
            for x in range(w):
                value = rng.read(1)[0] >> 4 | rng.read(
                    1)[0] << 4 & 0xF0
                
                img.putpixel((x, y), value)

    elif colorspace == 'Paletted':
        if mode!= 'P':
            raise ValueError("Mode can only be P when generating paletted images.")

        palette = []
        
        while True:
            v = rng.read(1)

            if v == b'\xFF':
                break
            
            palette += list(v * 3)

        img = pillowImage.new(colorspace, (w, h))

        for y in range(h):
            for x in range(w):
                index = rng.read(1)[0]

                pixel = 0
                for i in range(len(palette)):
                    if index == i // 3:
                        pixel |= 0b1 << i % 3

                img.putpixel((x, y), pixel)
        
    else:
        raise NotImplementedError(f'Colorspace "{colorspace}" is not implemented.')
    
    img.save(byteio, format='PNG')