#! /usr/bin/env python3
"""
Generate random pixel GIF files for testing purposes

Usage:
   ./generate_random_gifs N > somefile.gif
where N is an integer.  It will write that many
(smallish, <4k) images to stdout in their own GIF format
"""


import sys
import random
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Generate the header bytes
    width = random.randint(10, 256)
    height = random.randint(10, 256)
    
    # Create the palette and output it as the first part of the header
    pal = [struct.pack('BBB', *rng.read(3))
           for _ in range (768)]
    header = (b'GIF89a'
              + (width).to_bytes(1, 'big')
              + (height).to_bytes(1, 'big'))

    # Compute the size of the global color table
    # which must be rounded up to nearest multiple of 3
    numcolors = len({tuple(pal[x])
                      for x in range(768)}) // 3
    if numcolors!= 2 << max((numcolors - 1).bit_length(), 0):
        raise RuntimeError("Can't handle this many colors??")
    numcolors += 1

    # Now we can use the palette
    # Put the rest of the header together
    header += ((numcolors*3).to_bytes(1, 'little')
               + b'\0\0')

    out.write(header)

    # Write out all the image blocks
    for row in range(height):

        # Markers start with ','
        out.write(b',')
        lzwflag = False
    
        # First comes the local color table
        lctsize = min(max(numcolors//2**lzwflag, 1), 256)
        lct = [pal[i] for i in sorted(
                set(random.sample(range(
                    256), k=lctsize)))[:lctsize]]

        # Then comes the LZW minimum code size
        out.write(((lzwflag << 7) | lctsize).to_bytes(1, 'little'))
        
        # The colors are packed as raw RGB triplets
        out.writelines(lct)
        nbits = 8 - lctsize
        while lctsize:
            
            # If we've run out of bits, get more from the RNG stream
            if not nbits:
                nbits = 8
                c = ord(rng.read(1))
                pixels = []

            # Store this bit and update our count
            pixels.append((c >> nbits-1) & 1)
            nbits -= 1


            # When we have enough bits back together, emit them
            if not nbits:

                bits = sum(p<<n for p,n in zip(pixels, reversed(range(8))))
                if bits == 0xFFF:
                    out.write(b'\xFF\xFF')
                elif bits >= 256:
                    out.write(b'\xFE')
                    out.write(bits.to_bytes(2, 'little'))
                else:
                    out.write(bits.to_bytes(1, 'little'))
                    
                pixels = []
                


        # Next come the remaining pixels
        for col in range(width):
            # 0 means transparent
            if bool(int(rng.read(1))):
                out.write(b'\xFC')
            else:
                out.write(b'\x00')


    out.write(b';')
    return None

def main() -> None:

    myargs = sys.argv[1:]
    if '-h' in myargs or '--help' in myargs:
        print(__doc__)
        return None

    try:
        count = int(myargs[0])
    except Exception:
        print("Please specify how many GIFs you want.")        
        print(__doc__)
        return None

    # Set the seed based on current time so we'll get
    # always same sequence everytime this script runs
    SEED = str(abs(hash(str(time.time()))))
    random.seed(SEED)

    for idx in range(count):
        print("\tGenerating", end='')
        generate_random_gif(sys.stdin.buffer,
                            sys.stdout.buffer)                        
                            


