#!/usr/bin/env python3

import struct
import os
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). 
# Returns true iff the generated image was valid according
# to https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(
        rng: BinaryIO, 
        fout: BinaryIO,
        max_width : int = 640,
        max_height : int = 480,
        verbose : bool = False)->bool:


    # First some utility functions

    def read_logical_screen():
        """
        Reads the Logical Screen Descriptor

        Returns tuple containing width, height, background colour index, flag
        indicating presence of global color table, and flag indicating
        interlace
        """
        # Read the logical screen descriptor
        lsd = rng.read(7)

        # Extract the width and height from the descriptor
        w = 2 ** ((lsd[0] & 7) + 1)*(lsd[1])
        h = 2 ** ((lsd[0] >> 4) + 1)*(lsd[2])

        assert w <= max_width, "Image too wide"
        assert h <= max_height, "Image too tall"
        bgcoloridx = lsd[-1]
        useglobalcolortable = bool((lsd[0] >> 7) & 1)
        isinterlaced = bool((lsd[0] >> 6) & 1)

        return w, h, bgcoloridx, useglobalcolortable, isinterlaced

    def write_logical_screen(
            width, 
            height, 
            bgcoloridx, 
            useglobalcolortable,
            isinterlaced):
        """
        Writes the Logical Screen Descriptor with given parameters
        """

        # Pack up width and height values as an unsigned short
        w = width % 256
        h = height % 256
        wh = struct.pack('<H', 2**((math.log(max(width, height))/math.log(2)-1)*8+4)+w)
        lsd = struct.pack("=BBBBB", (((math.log(max(width, height))/math.log(2)-1)<<4)|(h&0xf)),
                         w,
                         0b000000, # No aspect ratio defined
                         (1<<7 if useglobalcolortable else 0),
                         (1<<6 if isinterlaced else 0)) + wh
        assert len(lsd)==7
        fout.write(lsd)

    def get_gctlength():
        """
        Gets the length of the global color table based on the LSD info
        """
        lsd = fout.tell()
        print("LSD at ", lsd)

        offset = io.seek(-1, 1)
        io.seek(offset)
        lsdbytes = io.read(7)

        assert len(lsdbytes) == 7

        # Unpack the LSD
        flags, wh1, wh2 = struct.unpack("<BBBB", lsdbytes[:4]), struct.unpack("<HH", lsdbytes[4:])
        print("Flags={}; WH1={}, WH2={}".format(*flags))

        # Calculate the start of the palette
        bgndindex = lsdbytes[-1]
        print("Background index=",bgndindex)

        return 2**(flags[0]&7 + 1)*(wh1 + wh2[0]+1)

    def write_gct():
        """
        Writes the Global Color Table if one is required
        """
        gct = None
        colors = get_gctlength() // 3
        if colors!= 0:

            # Generate a random palette
            #!!! Note that these should be checked within the legal range specified by the standard
            gct = [
               0xffffff*int(os.urandom(1)[0]/255)
               for i in range(get_gctlength())
           ]

            # Translate the palette indices to RGB tuples 
            gct_rgb = [
               (gct[i], 
                gct[i+1], 
                gct[i+2]) 
               for i in range(0, len(gct), 3)]

            # Convert all colours to BGRA tuples
            gct_bgra = [(c[2], c[1], c[0], 255) for c in gct_rgb]

            # Pack the values and append to the output
            gct_packed = bytes([(c[0] << 16) | (c[1] << 8) | c[2] | (c[3] << 24) 
                                for c in gct_bgra])

            # Write the packed values
            fout.write(gct_packed)

    def skip_to_next_imagedescription():
        """
        Skips to the next image description block
        """
        current = fout.tell()

        while True:
            fieldbyte = ord(rng.read(1))
            if ((fieldbyte & 128)!= 0 or
                    (fieldbyte & 1)!=1 or 
                    (fieldbyte & 2)==0 or
                    (fieldbyte & 32)==32):
                break

        newpos = fout.tell()
        fout.seek(current)
        return newpos-current
        
    # Start generating the image

    # Add header to the file
    fout.write(b'GIF89a')

    # Read the logical screen descriptor
    width, height, bgcoloridx, \
        useglobalcolortable, \
        isinterlaced = read_logical_screen()

    # TODO: Handle interlacing properly
    assert not isinterlaced,"Interlaced images are not supported yet!"

    # Write the LSD
    write_logical_screen(width, 
                        height,
                        bgcoloridx,
                        useglobalcolortable,
                        isinterlaced)

    # Get the length of the palette if present
    if useglobalcolortable:
        palletesize = get_gctlength()
        # Skip past the color table        
        fout.seek(palletesize, 1)

    # Eat blocks until we find one that doesn't meet our requirements
    position = skip_to_next_imagedescription()

    while position>0:
        print("Skipping {} bytes".format(position))
        fout.seek(position, 1)
        position = skip_to_next_imagedescription()

    # Create image data
    imgdataraw=[int(os.urandom(1)[0]/255) for x in range(width*height)]
    imgdata=[x%256 for x in imgdataraw]

    # Write image data section to the file
    # Set the first two bits of field byte to custom 
    # value so that it works with all browsers
    fieldbyte = imgdata[0]|(imgdata[1]<<8)
    fieldbyte |= 192|16     # Image separator | Local colour table
    fout.write(struct.pack(">B", fieldbyte))
    fout.write(struct.pack(">B", 8))    # Data length is always 8
    fout.write(bytes(imgdata))         # Actual data follows