#!/usr/bin/env python3
import sys
import os
import argparse
import random
import struct


MAX_WIDTH  = 32
MAX_HEIGHT = MAX_WIDTH
NUM_FRAMES = 10
RANDOM_SEED = None

COLOR_TABLE_LENGTHS = [
    2**n for n in range(1,9)
]

EXTENSION_CODE_BLOCKS = {
      0xf9 : "graphic control",
      0xfe : "comment",
      0xff : "application"
}

APP_EXTENSION_CODES = {
      0x01 : "netscape application extension",
      0xff : "application extension"
}


class RNG:
  def __init__(self):
    self._state = (None,)

  def _get_state(self):
    return self.__state

  state = property(_get_state)

  def randbytes(self, count):
    """Return some random bytes."""
    raise NotImplementedError

  @classmethod
  def new(cls, seed=None):
    """Construct a RNG with a given seed"""
    self = cls()
    
    initial_state = bytearray([
        0xb4, 0x8f, 0x0a, 0xcf, 0xef, 0x8b, 
        0x74, 0xc3, 0x1e, 0xa3, 0x8d, 0xac, 
        0xbb, 0xd6, 0x1e, 0x23, 0xfc, 0xdb, 
        0xbe, 0xca, 0x0b, 0x95, 0xec, 0xbc, 
        0xc7, 0x1b, 0xa7, 0x8d, 0x6c, 0xdf,
        0x86, 0x53, 0xdd, 0x73, 0x72, 0xe4, 
    ])

    if seed is not None:
      assert len(seed) >= len(initial_state), "Seed must be at least %d bytes long."%len(initial_state)

      # Hash the seed so that it takes longer to reverse than just generating it
      seed = hash(str(seed)).to_bytes(length=len(initial_state)*2, byteorder="big")
      
      # Expand the seed so that it uses all of our bits
      seed += int(hash("expand me!").hexdigest(), base=16).to_bytes(length=len(initial_state)-len(seed), byteorder="big")


    # Use the seed to appropriately initialize the PRNG
    self._state = tuple(
        ((initial_state[i] ^ seed[i]) & 0xff) for i in range(len(initial_state))
    )
  
    return self

  def init_from(self, other_rng):
    assert isinstance(other_rng, RNG)
    self._state = other_rng._state[:]

  def read(self, count):
    result = []
    while count > 0:
      v = self.randbytes(min(count, 20))
      result.append(v)
      count -= len(v)

    if len(result) == 1:
      return result[0]
    else:
      return b"".join(result)



class RandomBinRNG(RNG):
  def __init__(self, seed):
    super().__init__()
    self.__bitbuffer = 0
    self.__bitcount = 0
    if seed is None:
      seed = str(os.environ.get('RANDOM')).encode()
    if isinstance(seed, str):
      seed = seed.encode()

    self.init_from(RNG.new(seed))
  
  def _get_bitbuffer(self):
    return self.__bitbuffer

  bitbuffer = property(_get_bitbuffer)

  def _get_bitcount(self):
    return self.__bitcount

  bitcount = property(_get_bitcount)

  def init_from(self, other_rng):
    super().init_from(other_rng)
    
  def randbytes(self, count):
    buf = bytearray()
    while count > 0: 
      remaining = 8 - self.bitcount
      producing = min(remaining, count)
      count -= producing
  
      self.bitbuffer >>= producing
      self.bitbuffer <<= remaining
      self.bitbuffer |= (self.state[0] >> 7)
      self.bitcount += producing
  
      if self.bitcount >= 8:
        self.state = self.state[1:]
        self.bitcount -= 8
  
      if self.bitcount < 8:
        self.state = self.state[-1:]
  
  
      val = self.bitbuffer | ((8-producing)<<self.bitcount)
      buf.extend(((val>>offset)&0xff) for offset in reversed(range(producing)))
  
    return bytes(buf)

class RandomGenRNG(RandomBinRNG):
  def init_from(self, other_rng):
    super().init_from(other_rng)

  def randbytes(self, count):
    buf = bytearray()
    while count > 0:
      remaining = 32 - self.bitcount
      producing = min(remaining, count)
      count -= producing

      self.bitbuffer >>= producing
      self.bitbuffer <<= remaining
      self.bitbuffer |= self.state[0] & 0xFF
      self.bitcount += producing

      if self.bitcount >= 32:
        self.state = self.state[1:]
        self.bitcount -= 32

      if self.bitcount < 32:
        self.state = self.state[-1:]


      val = self.bitbuffer | ((32-producing)<<self.bitcount)
      buf.extend(((val>>offset)&0xff) for offset in reversed(range(producing)))

    return bytes(buf)


class RandomFileRNG(RNG):
  def __init__(self, filename):
    super().__init__()
    f = open(filename,'rb')
    self.init_from(RNG.new(f.read()))

  def init_from(self, other_rng):
    super().init_from(other_rng)

  def randbytes(self, count):
    return os.urandom(count)




def get_random_color():
  return bytes([
      random.getrandbits(8),
      random.getrandbits(8),
      random.getrandbits(8)
  ])

def get_random_color_table():
  ctlen = COLOR_TABLE_LENGTHS[
    random.randint(0, len(COLOR_TABLE_LENGTHS)-1)
  ]
  colors = []
  for x in range(ctlen):
    colors.append(get_random_color())
  return bytes(colors)
  

def get_random_extension():
  extcode = random.randint(0, 255)
  exthdr = bytes([
      0x21, # start marker
      extcode, # extension code (LSB)
      0, # extension code (MSB) 
      0, # size (LSB)
      0 # size (MSB) 
  ])

  if extcode in EXTENSION_CODE_BLOCKS:
    exthdr[3],exthdr[4] = struct.pack(">H", random.randint(0, 2**16))

  # Add more custom behavior based on extensions
  if extcode == 0xf9:
    # graphic control extension
    pass
  elif extcode == 0xff:
    # application extension
    appcode = random.randint(0, 255)
    applen = random.randint(0, 2**16)
    exthdr[5], exthdr[6] = struct.pack(">H", appcode)
    exthdr[3], exthdr[4] = struct.pack(">H", applen+1) 

    if appcode in APP_EXTENSION_CODES:
      print("\t\t%s"%APP_EXTENSION_CODES[appcode])


  valueblock = bytes([])

  # Insert something customized depending on the extension
  if extcode == 0xf9:
    delaytime = random.randint(0, 255)
    transparentindex = random.randint(0, 255)
    disposalmethod = random.randint(0,3)

    flags = delaytime | (transparentindex << 8) | (disposalmethod<<10)
    exthdr[7] = flags
    # TODO: handle transparency info from the spec
    # print("\t\tTransparency index:", transparentindex)
    # print("\t\tDelay time:", delaytime)
    # print("\t\tDisposal method:", disposalmethods[disposalmethod])
    # exthdr[8:12] = transparentcolor
  elif extcode == 0xfe:
    commentsize = random.randint(0, 2**16)
    comment = bytes([(random.randint(0, 255)) for x in range(commentsize)])

    # Update the header fields
    exthdr[3],exthdr[4] = struct.pack(">H", len(comment)+1)
    valueblock = comment
  elif extcode == 0xff:
    if appcode ==