#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # Generate the header and add to output
    header   = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3
    numcolors = int(gctlen / 3)

    # Read global color table if present
    if gctlen!= 0:
      gct = rng.read(gctlen)
      out.write(gct)

    extblocks = [] # extensions go in their own blocks
                   # since they are optional and could
                   # contain arbitrary amounts of data 
    while True:
        blocktype = rng.read(1)[0]
        if blocktype == 0x3b: # stop marker
            break
        elif blocktype!= 0x2c: # unknown block type
            continue

        # Read the rest of the extension
        extheader = rng.read(6)
        out.write(extheader)
        
        # Parse extension header
        hasgct   = bool(extheader[0] & 0x80)
        hasimgdata = bool(extheader[0] & 0x08)
        size    = int(struct.unpack(">H", extheader[2:])[0]) + 1
        # If there's no image data we're done here
        if not hasimgdata:
          extblocks.append({
              "hasgct":      False, 
              "hasimgdata":  True,
              "size":        size}) 
          continue
 
        # Write the local color table if necessary
        if not hasgct:
            numcolors = size//3 - 1
            if numcolors > 0:
                colortable = rng.read(numcolors*3)
                out.write(colortable)
        else:
            gclen = 2 ** ((extheader[4] & 0x07) + 1) * 3
            gct = rng.read(gclen)
            out.write(gct)

        # Now write image data 
        imgdata = rng.read(size-1)
        out.write(imgdata)

        extblocks.append({"hasgct": hasgct,
                          "hasimgdata": hasimgdata,
                          "size": size}) 

    print("Generated {} with {} colors".format(len(extblocks), numcolors))