#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    while True:
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

        # TODO: Should experiment with variations on these values too
        #       see https://github.com/decompme/decompilation-guide/issues/47
        chunksize = min(rng.read(1)[0] % 16, 10)
        out.write(chunksize.to_bytes(1, 'little'))


        # Drawing Area Width
        val = rng.read(2)
        width = struct.unpack('<H', val)[0] & 0x3FFF
        out.write(width.to_bytes(2, 'little'))

        # Drawing area height
        val = rng.read(2)
        height = struct.unpack('<H', val)[0] & 0x3FFF
        out.write(height.to_bytes(2, 'little'))



        palette = []
        for i in range(1 << ((val >> 4) & 0x7)):
            palette += [tuple([
                rng.read(1)[0], 
                rng.read(1)[0], 
                rng.read(1)[0]])]

        # Simply copy over x / y as is
        out.write((val & 0xf).to_bytes(1, 'little'))
        out.write(val.to_bytes(1, 'little'))
        
        pbitdepth = ((palette[0][0] | palette[0][1] | palette[0][2]) == 0) 
        out.write((pbitdepth << 7) | len(palette))

        # Let's assume they're all paletted...
        out.write(b'\xFF\xDE')
        out.write(((len(palette)//3)-1).to_bytes(1, 'little'))
        for entry in palette:
            if not palletebitdepth:
                out.write((entry[0]).to_bytes(1, 'big'))
            else:
                out.write(entry[0].to_bytes(2, 'big'))
            
            if not palletebitdepth:
                out.write((entry[1]).to_bytes(1, 'big'))
            else:
                out.write(entry[1].to_bytes(2, 'big'))

            if not palletebitdepth:
                out.write((entry[2]).to_bytes(1, 'big'))
            else:
                out.write(entry[2].to_bytes(2, 'big'))

        lzwmincodesize = rng.read(1)[0]
        out.write(lzwmincodesize.to_bytes(1, 'little'))

        # Calculate how many bits our maxval will need        
        bitsperpixel = {
                8: 1,
                4: 2, 
                2: 4}[max(v[0] for v in palette)]

        
        outframedata = int.from_bytes(rng.read(4), byteorder='big')
        outframesize = bitsperpixel // 8
        # LZW minimum code size is stored in one byte, thus only 256 possible
        # codes. So we need ceil(log(len(colors)/256)), or log2(len(colors)/256)
        # For now I'm assuming there aren't any images where the number of colors
        # exceeds 1<<26, which fits well within 3 bytes.
        # See for example https://stackoverflow.com/questions/30117903/what-is-the-maximum-number-of-colours-in-an-lzw-compressed-image
        lzwmaxbits = 3

        out.write(outframedata >> 8*outframesize)
        out.write(outframedata & 2**8*outframesize-1)
        out.write((((lzwmincodesize+1) << 8)+lzwmaxbits).to_bytes(1, 'little'))

        break
    
    last_byte = out.tell() - 1
    # At the moment this code doesn't support appending arbitrary amounts
    assert len(data[last_byte:]) <= 256 * 13
    if len(data[last_byte:])!= 0:
        # If the data[last_byte:] isn't empty adjust
        # what's already been written according to its size
        out.seek(-len(data[last_byte:]), 1)

        # And then append whatever was left from the beginning
        print("Appending", len(data[last_byte:]), 'bytes...', end='', flush=True)
        out.write(data[last_byte:])
        print('done.', flush=True)

    # Finally add the trailer to finish off the output.
    # Normally this would be part of the application logic
    # when writing a gif but since this is intended only
    # for demonstration purposes we do it here directly.
    # See https://github.com/decompme/decompilation-guide/issues/46
    print('Adding the trailer and EOF', end='', flush=True)
    # 2 bytes for signature
    # 1 byte for version
    # 1 byte for flags
    # 2 bytes for background color index
    # 2 bytes for pixel aspect ratio
    out.write(b'\x3B\x00\x00\x00')
    print(' done', flush=True)