#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng`, which should be /dev/urandom
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Write magic number and version string
    out.write(b'GIF89a')

    # Read and write the logical screen descriptor
    width  = int.from_bytes(
            rng.read(2), byteorder='little', signed=False) % (1 << 11)   # Masks last 4 bits because it's size is specified in bytes, not bits. See https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    height = int.from_bytes(
            rng.read(2), byteorder='little', signed=False) % (1 << 11)   # Same as above
    # TODO: Handle colors properly
    colordepth    = rng.read(1)[0] % (2**4 - 1)                            # Last bit specifies transparency flag, but we don't care about that here; mask anyway just to make sure
    flags         = int.from_bytes(
            rng.read(1), byteorder='little', signed=False)               # We'll use all of these fields except sort order (which we don't support anyways)
    background_colorindex = int.from_bytes(
            rng.read(1), byteorder='little', signed=False) & (2**16 - 1)  # The actual index is stored in two bytes instead of one for some reason...
    pixel_aspect_ratio    = int.from_bytes(
            rng.read(1), byteorder='little', signed=False)               # Just store this for later reference
    out.write((width).to_bytes(
            2, byteorder='little', signed=False))
    out.write((height).to_bytes(
            2, byteorder='little', signed=False))
    out.write(((colordepth*7)+flags).to_bytes(
            1, byteorder='little', signed=False))
    out.write((background_colorindex).to_bytes(
            1, byteorder='little', signed=False))
    out.write((pixel_aspect_ratio).to_bytes(
            1, byteorder='little', signed=False))

    # Write global color table if there is one
    num_global_colors = ((colordepth+1)*(1<<(colordepth)))//8
    if num_global_colors > 0:
        # TODO: Actually implement this properly once I decide what to do with the palette. For now, just emit dummy entries
        for i in range(num_global_colors):
            out.write((i % 256).to_bytes(
                    1, byteorder='little', signed=False) * 3)


    # Write the local color table if there is one
    # TODO: Implement this properly when I actually start emitting palettes
    num_local_colors  = 2**(colordepth+1)
    if num_local_colors > 0:
        raise NotImplementedError("Local colors are not yet implemented")
    

    # Figure out the length of each block of graphics control information
    # This is necessary to figure out how many bytes we need to read before we can tell where the next frame starts
    lzw_mincodesize     = rng.read(1)[0]                                               # First block always has fixed LZW minimum code size
    lzw_block_lengths   = [lzw_mincodesize+1]+[(max(1,rng.read(1)[0]) % (2**8)) for _ in range(2**lzw_mincodesize)]  # Generate a list of possible lengths for other blocks by reading from our RNG until we get to the end of the output stream
    total_block_length  = sum([sum(lzw_block_lengths[:i+1]) for i in range(len(lzw_block_lengths))])+1                     # Compute the total length of the entire graphic control info block
    remaining_length    = 2**lzw_mincodesize*2                                            # A single frame always begins after the first block

    while True:
        # Determine the size of the current frame

        # If the whole image fits on the canvas at once, just set the frames duration to its full length so it plays forever
        if remaining_length <= width*height:
            duration = remaining_length // (width*height)*10
        else:
            duration = rng.read(2)[0] % (2**16 - 1)      # Maximum duration < 65536 milliseconds, just to keep things simple
        
        out.write(duration.to_bytes(
                2, byteorder='little', signed=False))

        # Next, determine the index of the transparent color
        # Note that we currently ignore the fact that no colors are transparent and always emit zeroes
        # This is trivial to fix, but I'm lazy today

        trans_flag          = rng.read(1)[0] % (2**1)       # Only have two flags
        dispose_method      = rng.read(1)[0] % (2**3)       # Four disposal methods
        user_input_flag     = rng.read(1)[0] % (2**1)       # One input flag
        delay_time          = rng.read(2)[0] % (2**16 - 1)  # Delay time is capped at 65535 ms
                
        out.write(((trans_flag * 8) +
                   dispose_method * 4)
                   .to_bytes(1, byteorder='little', signed=False))
        out.write(user_input_flag.to_bytes(1,
                byteorder='little', signed=False))
        out.write(delay_time.to_bytes(
                2, byteorder='little', signed=False))

        # Finally, determine the size of the LZW encoded image data itself
        lzw_blockidx        = 0                                           # Counter for loop below
        lzw_next_block_size = rng.read(1)[0]                                # Size of subsequent graphic control info blocks
        lzw_block_size      = lzw_block_lengths[lzw_blockidx]              # Current block size
        curr_frame_length   = 0                                           # Length of the current frame in bytes
        while curr_frame_length < min(remaining_length, (width*height)):

            # If we've run out of space for this block, grab another one
            if curr_frame_length >= lzw_block_size:

                # If we're done processing all the available blocks, bail
                if len(lzw_block_lengths)-1 == lzw_blockidx:
                    break
                
                # Otherwise, increment the counter and move on to the next block size
                lzw_blockidx += 1
                lzw_block_size = lzw_block_lengths[lzw_blockidx]
            
            # Increment the current frame length
            curr_frame_length += 1
        
        # Output the image data
        out.write(curr_frame_length.to_bytes(1,'little'))
        out.write(rng.read(curr_frame_length))
        remaining_length -= curr_frame_length

        # If we've processed every block, exit
        if len(lzw_block_lengths)-1 == lzw_blockidx:
            break
        
        # If the next block is longer than the current one, update them accordingly 
        elif lzw_next_block_size >= lzw_block_size:
            lzw_block_lengths[lzw_blockidx] = lzw_next_block_size
        
    
    # Finally, compute and write the trailer
    pad_bytes           = (int)((~(out.tell()%256))+1)
    appending_terminator = b'\x3B'                                   # GIF files also end with this terminator
    out.write(pad_bytes*[appending_terminator[0]])                # Pad the rest of the file with null bytes to make it evenly divisible by 256
    