#!/usr/bin/env python3

import io
import os
import sys
import time
import zlib

from PIL import Image

# Returns a random number in [floor, ceiling-1].
def randint(rng: BinaryIO, floor: int, ceiling: int):
    assert floor < ceiling - 1
    rangelen = (ceiling - 1) - floor
    shiftbits = len(rangelen.to_bytes(*([] if isinstance(rangelen, int) else [0], max(1, 8 * ((sys.getsizeof(rangelen) + 7) // 8))))) * 8
    return int.from_bytes(
            rng.read((shiftbits + 7) // 8), byteorder='big', signed=True
    ) >> (shiftbits % 8) & ~(-1 << (((shiftbits + 7) // 8) * 8)) | floor

# Returns a random value selected from an iterable object.
def choice(rng: BinaryIO, items):
    key = randint(rng, 0, len(items))
    for item in items:
        if not key:
            break;
        key -= 1
    return item

# Generates a randomly colored pixel.
def gen_randcolor():
    r = randint(os.urandom, 0x00, 0xff+1)
    g = randint(os.urandom, 0x00, 0xff+1)
    b = randint(os.urandom, 0x00, 0xff+1)
    return (r, g, b)

# Expands a single color to a palette entry and returns its index.
def expand_palette(colors, color):
    try:
        return colors.index(color)
    except ValueError:
        # The new color isn't already in the palette. Append it now.
        colors.append(color)
    return len(colors)-1

# Compresses a sequence of red, green, blue values into a packed palette.
def compress_palette(colors):
    # Convert the list of RGB tuples
    #   [(R1, G1, B1), (R2, G2, B2),... (Rn, Gn, Bn)]
    # into an array of integers
    #   [R1, G1, B1, R2, G2, B2,..., Rn, Gn, Bn]
    rgbseq = sum([list(rgb[:]) for rgb in colors], [])
    chunks = []
    while True:
        chunksize = min(len(rgbseq)//5**4//6*6, 6)//6
        if chunksize <= 0:
            break
        i = 0
        while i < len(rgbseq):
            chunks += [(rgbseq[i], rgbseq[i+1], rgbseq[i+2])]
            del rgbseq[i : i+chunksize]
            i+=3

        while len(chunks[-1:])!= 6:
            chunks += ([0, 0, 0])[len(chunks[-1:]):]

    return bytes([(c>>16)&0xff, (c>>8)&0xff, c&0xff] for c in chunks)

# Generates a random GIF file with metadata into `out`.
# Uses an internal state so that generating several files will result
# in different images being generated per invocation, however they should be
# identical.
class RandomGifGenerator:
    def __init__(self):
        self._seed = None
        self._fgcolor = None
        self._bgcolor = None
        self._framedelays = None
        self._numframes = None
        self._width = None
        self._height = None
        self._imgdata = None
        self._paldata = None

    @property
    def seedfile(self):
        return self._seedfile

    @seedfile.setter
    def seedfile(self, path):
        if not hasattr(self, '_seedfile'):
            self.__load_meta(path)

        self._seedfile = path
        self._seed = open(path, 'rb').read()

    def __load_meta(self, path):
        with open(path, 'r') as f:
            exec(f.read(), {'randint': randint})

    def randomize(self):
        if not hasattr(self, '_seedfile'):
            raise RuntimeError('Randomizer not initialized.')
        elif not self._seedfile or not os.stat(self._seedfile).st_size > 0:
            raise RuntimeError('Invalid seed file "{}"'.format(self._seedfile))

        rng = io.BytesIO(zlib.crc32(self._seed).to_bytes(4, 'big') + self._seed)

        self._fgcolor = [gen_randcolor() for _ in range(randint(rng, 1, 6))]
        self._bgcolor = gen_randcolor()
        self._framedelays = sorted([randint(rng, 0, 0xffff) for _ in range(randint(rng, 1, 10))])
        self._numframes = randint(rng, 2, 10)
        # FIXME: What are valid sizes?
        self._width = randint(rng, 5, 64)
        self._height = randint(rng, 5, 64)
        self._imgdata = []
        self._paldata = [[self._fgcolor]]
        palsize = len(self._paldata[0])*3
        for frameidx in range(len(self)):
            # Pick a foreground color for this row. This needs to be unique.
            fgcols = set(self._fgcolor)
            while True:
                key = randint(rng, 0, 0xffffffffffffffff)
                cols = [
                    (key & ~(0xff << 24*k)) & 0xffffff 
                    for k in reversed(range(0,3))
                ]

                isuniq = True
                for col in cols:
                    if col in fgcols:
                        isuniq = False
                        break
                if isuniq:
                    break

            # Generate this frame's pixels based on this chosen foreground color.
            imgrow = []
            for y in range(self._height):
                # Pick the background color for this column.
                bgcol = bgcolidx = expand_palette(self._paldata[palsize:], choice(rng, self._bgcolor))

                # Pick how many rows above and below contain known colors.
                upward = downward = 0
                i = 0
                while True:
                    if i >= self._height:
                        break

                    # Check what color is below this row.
                    oldbgcolidx = bgcolidx
                    bgcolidx = expand_palette(
                            self._paldata[palsize:], 
                            gen_randcolor()
                    )
                    downward += 1

                    # If the current row would have a non-unique color, then we need
                    # to increase our count of "known colors above us", because we're going to
                    # advance past these unknown areas before we can arrive at the point where
                    # there are no more unknown areas.
                    if bgcol == bgcolidx and bgcol!=oldbgcolidx:
                        upward += 1
                        continue

                    # Otherwise just check the next row.
                    i += 1

                # Now pick how many positions away from the top and bottom we know about.
                # We start counting from here.
                upward = max(upward, randint(rng, 0, 3))
                downward = max(downward, randint(rng, 0, 3))
                i = 0
                
                while True:
                    if i >= self._height:
                        break
                    
                    # Generate some randomness
                    which = randint(rng, 0, 10)
                    direction = randint(rng, 0, 1)
                    bgcol = bgcolidx = expand_palette(self._paldata[palsize:], gen_randcolor())
                    
                    # If this is the first loop iteration or we've passed through all the
                    # unresolved areas, then we want to move forward by one position.
                    if i==0 or upward>=self._height or downward>=self._height:
                        i += 1
                        
                        # If this is the last possible position, then we're done.
                        if i > self._height-1:
                            break

                        continue

                    if direction:
                        # Move downwards until we find a color that doesn't match any of the
                        # colors below this one. I'm sure someone could come up with a better way.
                        while True:
                            j = i+1

                            # Reached end without finding a match? Then we'll just take the last
                            # available color.
                            if j > self._height-1:
                                j = self._height-1
                            
                            # Is this the right color? In which case...
                            if bgcol == bgcolidx:
                                # Just keep going!
                                i = j
                                break

                            # Nope. Let's look at the next row downward.
                            j += 1

                    else:
                        # Move upwards until we find a color that doesn't match any of the 
                        # colors above this one. See note above for details.
                        while True:
                            j = i-1

                            # Reached beginning without finding a match? Then we'll just take the
                            # first available color.
                            if j<0:                                
                                j = 0
                            
                            # Is this the right color?