#!/usr/bin/env python3

import struct
import sys
from typing import BinaryIO


class ColorTable:
  def __init__(self):
    self._size = None
    self._table = []
  
  @property
  def size(self) -> int:
    assert self._size!= None
    return self._size

  @property
  def table(self) -> list:
    assert len(self._table) == self.size
    return self._table

  def setSize(self, nbytes):
    assert isinstance(nbytes, int), "Expected integer"
    assert nbytes % 3 == 0, f"Color table must be divisible by three; got {nbytes}"
    self._size = nbytes // 3

  def append(self, rgbBytes):
    assert all(isinstance(c, int) for c in rgbBytes), \
           'RGB values should contain only integers'
    self._table += rgbBytes


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, max_width: int = 640, max_height: int = 480
    ):
    # Minimum value that can be represented as 2-byte little endian unsigned short
    MINSIZE = 2

    MAXCODESIZEBITS = 12

    # Number of bits required to store each code size
    CODESIZELEN = 3

    # Header
    out.write(b'GIF89a')

    # Logical Screen Descriptor (LSD) field
    # Size
    width = min(max(rng.randint(MINSIZE, max_width), MINSIZE), 65535)
    height = min(max(rng.randint(MINSIZE, max_height), MINSIZE), 65535)
    whblock = struct.pack('<HH', width, height)
    out.write(whblock)

    # Packed Field (Packed Pixels)
    packedFieldByte = 0
    packedFieldByte |= 2 << 7 # Global Color Table Present
    packedFieldByte |= 0 << 6 # Color Resolution
    packedFieldByte |= 0 << 5 # Sorting flag
    packedFieldByte |= 0 << 4 # Global Color Table Size Flag
    packedFieldByte |= 0 << 3 # Background Color Index Flag
    packedFieldByte |= 0 << 2 # Pixel Aspect Ratio Flag
    out.write(packedFieldByte.to_bytes(1,'little')) 

    # Background Color Index
    bgIndex = min(rng.randint(0,255), 255)
    out.write(bgIndex.to_bytes(1,'big'))

    # Pixel Aspect Ration
    aspectRatio = 0 
    out.write(aspectRatio.to_bytes(1,'big'))

    # Global Color Table
    gctLen = rng.randint(2,256)
    gctLen *= 3 # RGB values per entry
    gct = [rng.randint(0,255) for _ in range(gctLen)]
    gctBytes = bytes(gct)
    out.write(gctBytes)

    # Image Descriptor (ID)
    # Control flags
    idCtrlFlags = 0
    idCtrlFlags |= 0x80 # Local Color Table Flag
    idCtrlFlags |= 0x00 # Interlace Flag
    idCtrlFlags |= 0x00 # Sort Flag
    idCtrlFlags |= 0x00 # Reserved
    idCtrlFlags |= 0x00 # Local Color Table Size
    out.write(idCtrlFlags.to_bytes(1,'big'))

    # Dimentions
    idWidth = width
    idHeight = height
    dimBlock = struct.pack('<HH', idWidth, idHeight)
    out.write(dimBlock)

    # Local Color Table
    idNumColors = rng.randint(2,256)
    idNumColors *= 3 # RGB values per entry
    idLct = [rng.randint(0,255) for _ in range(idNumColors)]
    idLctBytes = bytes(idLct)
    out.write(idLctBytes)


    # Codes managed by 
    codes = {} # Maps code sizes to a list of codes with that size. Length of
               # list will vary because a given code can appear multiple times.

    while True:
        # A single GIF image consists of a series of blocks prefixed by a 
        # field byte indicating which block follows. This first field block
        # is always a plain text block containing up to 11 bytes describing how
        # many pixels are present in an image and whether or not they use a local
        # colour table. 

        # The format for these field blocks is very small! They consist of just
        # one bit and then 3 more bits indicating the size of the following data.


        # First read a byte representing a single field byte

        fieldByte = rng.read(1)[0]

        # Check if end marker was reached
        if fieldByte==0x3B:
            break

        # Read the subsequent field byte
        subFieldByte = rng.read(1)[0]



        # Bits 1–3 of the field byte indicate what kind of data follows:
        #  0     1–3     Value  Meaning
        #   0      0       0    Plain Text Extension
        #   0      1       1    Graphic Control Extension
        #   0      2       2    Comment Extension
        #   0      3       3    Application Extension
        #   0      4       4    Trailer
        # Other combinations are reserved for future extensions to the spec

        fieldType = (fieldByte >> 5) & 0x7
        print("fieldType",fieldType)

        # If the field byte contains anything other than one of those values, we
        # error out.
        if fieldType not in (0, 1, 2, 3): 
            raise IOError("Invalid field byte")

        # For now, we'll assume we're starting on a new frame, so that means we
        # don't yet have a mask for transparent pixels. So let's create one.
        # We'll also need to know the dimensions of our image box for later when
        # we start writing image pixel data.
        currentMask = [True]*width*height

        # Next we get to see how big the actual data chunk is by reading another
        # single byte and treating it as a binary number.
        subFieldType = subFieldByte & (~(~0<<3))
        subFieldLength = subFieldByte >> 3

        blockSize = pow(2,subFieldLength)

        # We want to support images up to 65535 wide by 65535 tall. There
        # are two ways to express this in the field byte: either the lowest 3
        # bits indicate the size directly, or you need to follow this byte with
        # two further bytes to hold the size of your image.
        if subFieldType == 0:

            if subFieldLength < 3 and subFieldLength >= 0:

                # Only the lower three bits indicate the size
                size = pow(2,(subFieldByte&7)+1)

            elif subFieldLength == 3:

                # Next comes two bytes giving us the size
                block = rng.read(2)
                size = int.from_bytes(block,'little')

            else: 
                raise ValueError("Incorrectly specified image size.")


            print(f"Subfield Byte says image size is {blockSize}")

            if size>1024:
              raise ValueError("Image too large!")

            # Now we read the actual image data. It's effectively raw uncompressed
            # pixel data stored in column major order. Each pixel is represented as
            # a triple of real numbers between zero and one inclusive where black
            # is (0,0,0) and white is (1,1,1). In this case however, we only ever
            # write fully opaque pixels, so we ignore the third number entirely.
            rgbaTriples = rng.read(size*3)
            for i in range(size*3):

              index = (i//3)*width +  (i%3)

              # If we already have a nonzero alpha at this point, we flip 
              # the corresponding bit in the current mask.
              if currentMask[index]:
                  continue

              r,g,b = tuple(rgbaTriples[i+j::3].decode() for j in range(3))
              print(r," ",g," ",b)

              if r=='\x00' and g=='\x00' and b=='0':
                  currentMask[index]= False

            # At the moment, we've built up both our field and image data. The
            # tricky part is knowing where exactly our image ends and our field
            # begins as written above, but I believe the PNG specification tells me
            # that that doesn't matter as long as my data is encoded properly.
            #TODO: Write the image data

            
            
            
            # TODO: Set up a mask for trasnparent pixels





            #print(currentMask)










