#!/usr/bin/env python3

import array
import ctypes
import sys
import struct


class VGIF(ctypes.Structure):

    _fields_ = [('signature', (ctypes.c_byte * 6).in_dll(sys, '_signature')),
                ('version', ctypes.c_char),
                ('width', ctypes.c_ushort),
                ('height', ctypes.c_ushort)]

    def __str__(self):

        return'signature:{:s} version:{:s} width:{:d} height{:d}'.format(
                    self.signature.decode(), self.version.decode('ascii'),
                    self.width, self.height)


class BlockHeader(ctypes.Structure):
    
    # There is no way to determine at build time how many colors there are in
    # the image--not even using a C preprocessor. The best we can do here is set
    # an arbitrary maximum value for now and then make sure we don't go over it
    MAX_NUM_COLORS = 256

    _fields_ = [('block_size', ctypes.c_ubyte), 
                ('packed_flag', ctypes.c_ubyte), 
                ('background_color_index', ctypes.c_ubyte), 
                ('aspect_ratio', ctypes.c_ubyte), 
                ('palette_colors', (ctypes.c_ubyte * 3 * MAX_NUM_COLORS))]

    def __init__(self, block_size=None, packed_flag=None, 
                 background_color_index=None, aspect_ratio=None):
        
        if not isinstance(block_size, type(None)):

            setattr(self, 'block_size', block_size)

        if not isinstance(packed_flag, type(None)):

            setattr(self, 'packed_flag', packed_flag)

        if not isinstance(background_color_index, type(None)):
            
            setattr(self, 'background_color_index', 
                    background_color_index)

        if not isinstance(aspect_ratio, type(None)):
        
            setattr(self, 'aspect_ratio', aspect_ratio)

    def __str__(self):

        return '{:s}(size={:d}, packed={:#04X})'.format(
                super().__str__(), self.block_size, self.packed_flag)

    @property 
    def has_global_color_table(self):

        return bool(self.packed_flag & 0b1000)


    @property 
    def sort_flag(self):

        return ((self.packed_flag >> 3) & 0b1) == 1


    @property 
    def size_of_local_color_table(self):

        return 2 ** (((self.packed_flag >> 7) & 0b111) + 1)


    @property         
    def transparent_color_flag(self):

        return (self.packed_flag & 1)!= 0


    @staticmethod
    def get_block_size():

        return (BlockHeader._length_)


class ImageDescriptor(ctypes.Structure):

    _fields_ = [('left_position', ctypes.c_ushort),
                ('top_position', ctypes.c_ushort),
                ('width', ctypes.c_ushort),
                ('height', ctypes.c_ushort),
                ('packed_flags', ctypes.c_ubyte)]

    def __init__(self, left_position=None, top_position=None,
                 width=None, height=None, packed_flags=None):

        if not isinstance(left_position, type(None)):

            setattr(self, 'left_position', left_position)

        if not isinstance(top_position, type(None)):

            setattr(self, 'top_position', top_position)

        if not isinstance(width, type(None)):

            setattr(self, 'width', width)

        if not isinstance(height, type(None)):

            setattr(self, 'height', height)

        if not isinstance(packed_flags, type(None)):

            setattr(self, 'packed_flags', packed_flags)

    def __str__(self):

        return '{:s}(L={:d}, T={:d}, W={:d}, H={:d})'.format(
                super().__str__(), self.left_position, self.top_position,
                self.width, self.height)        

    @property    
    def local_color_table_flag(self):

        return bool((self.packed_flags >> 7) & 1)


    @property
    def interlace_flag(self):

        return bool((self.packed_flags >> 6) & 1)


    @property         
    def sort_flag(self):

        return bool((self.packed_flags >> 5) & 1)


    @property        
    def reserved_bits(self):

        return self.packed_flags & 0b11111000


class GraphicsControlExtension(ctypes.Structure):

    _fields_ = [('block_size', ctypes.c_ubyte),
                ('packed_flags', ctypes.c_ubyte),
                ('delay_time', ctypes.c_ushort),
                ('transparent_color_index', ctypes.c_ubyte),
                ('terminator', ctypes.c_ubyte)]

    def __init__(self, delay_time=None, transparent_color_index=None):

        if not isinstance(delay_time, type(None)):

            setattr(self, 'delay_time', delay_time)

        if not isinstance(transparent_color_index, type(None)):

            setattr(self, 'transparent_color_index', 
                    transparent_color_index)

    @property
    def disposal_method(self):

        return (self.packed_flags >> 2) & 0b111


    @property
    def user_input_flag(self):

        return (self.packed_flags >> 1) & 1 == 1


    @property
    def transparency_flag(self):

        return bool((self.packed_flags) & 1)



class PlainTextExtension(ctypes.Structure):

    _fields_ = [('block_size', ctypes.c_ubyte),
                ('text_grid_left', ctypes.c_ubyte),
                ('text_grid_top', ctypes.c_ubyte),
                ('text_grid_width', ctypes.c_ubyte),
                ('text_grid_height', ctypes.c_ubyte),
                ('cell_width', ctypes.c_ubyte),
                ('cell_height', ctypes.c_ubyte),
                ('foreground_color_index', ctypes.c_ubyte),
                ('background_color_index', ctypes.c_ubyte),
                ('subblocks', ctypes.POINTER(ctypes.c_ubyte)),
                ('num_subblocks', ctypes.c_long)]

    
class ApplicationExtension(ctypes.Structure):

    _fields_ = [('block_size', ctypes.c_ubyte),
                ('application_id', (ctypes.c_ubyte * 8)),
                ('auth_code', ctypes.POINTER(ctypes.c_ubyte)),
                ('num_auth_codes', ctypes.c_long)]


class CommentExtension(ctypes.Structure):

    _fields_ = [('block_size', ctypes.c_ubyte),
                ('comment', ctypes.POINTER(ctypes.c_ubyte)),
                ('num_chars', ctypes.c_long)]
    

class GraphicControlLabel(ctypes.Structure):

    _fields_ = [('label', (ctypes.c_ubyte * 3))]

    def __init__(self, label):

        ctypes.memmove(ctypes.addressof(self), ctypes.create_string_buffer(label),
                       len(label))


class FileTrailer(ctypes.Structure):

    _fields_ = [('signature', ctypes.c_uint)]

    def __init__(self):

        super().__setattr__('signature', b'GIF')


def create_random_graphic_control_extension():

    extension = GraphicsControlExtension()
    num_subblocks = randint(1, 4)
    subblocks = [randint(0, 0xFF) for i in range(num_subblocks)]
    terminator = randrange(0, 256 - sum(subblocks)-1)
    extension.__setattr__('block_size', len(subblocks) + 4)
    extension.__setattr__('packed_flags', 
                          0
                          | (choice([False, True]) << 2)
                          | (choice([False, True]) << 1)
                          | choice([False, True]))
    extension.__setattr__('delay_time', randint(0, 65535))
    extension.__setattr__('transparent_color_index', randint(0, 255))
    extension.__setattr__('subblocks', (ctypes.c_ubyte * num_subblocks)(*subblocks))
    extension.__setattr__('num_subblocks', num_subblocks)
    extension.__setattr__('terminator', terminator)
    return extension


def create_random_image_descriptor():

    descriptor = ImageDescriptor()
    descriptor.__setattr__('left_position', randint(0, 65535))
    descriptor.__setattr__('top_position', randint(0, 65535))
    descriptor.__setattr__('width', randint(0, 65535))
    descriptor.__setattr__('height', randint(0, 65535))
    descriptor.__setattr__('packed_flags', 
                            choice([