#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)


    # First write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)


    # Now we're going to write 256 descriptors at once
    for i in range(258 - len(data)):
        # To make sure our padding isn't too small or large,
        # ensure that each individual descriptor has at least one
        # extra byte in it.
        blocklen = rng.randint(1, 0xfffe)

        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)
        label = 0x10 | ((blocklen >> 8) & 0xf)

        out.write(struct.pack('<H', label))
        if blocklen >= 0x10:
            out.write((blocklen).to_bytes(2, 'little'))
        else:
            # Tack on a null byte when necessary
            out.write(((blocklen << 8) & 0xffffff00).to_bytes(3, 'little'))

        # Add some data if present
        sizeleft = blocklen - 10
        if sizeleft > 0:
            out.write(rng.read(sizeleft))

        # Fill with zeros so the total sum of sizes matches up
        out.write([0] * (-(-blocklen // 4)))

    # Finally, finish it off!
    out.write(b'\x3B')

    # And pad up until the end of the file, which is easy enough because
    # we've already written exactly datalen bytes before ending.
    print('Padding global color table', end='', flush=True)
    while True:
        if out.tell() == datalen+1: break
        out.write(rng.read(1))
    print('. Done!', flush=True)