#!/usr/bin/env python3

import math
import struct
from typing import BinaryIO


# Generates `nframes` frames of an animated gif with dimension `width`x`height`.
# The random number generator used to generate the contents is `rng`, which should be
# /dev/urandom or a similar source of randomness.
def generate_random_animated_gif(
        nframes: int, width: int, height: int, rng: BinaryIO, out: BinaryIO
    ) -> None:

  # Generate the header
  header = b"GIF89a"
  out.write(header)
  
  # Get a 6-byte Logical Screen Descriptor (LSD) field
  lsd = rng.read(6)

  # Set gctlen to zero since we don't use one
  lsdbytes = list(lsd)
  lsdbytes[4] &= ~0b111 << 5
  lsd = bytes(lsdbytes)

  # Write LSD to output
  out.write(lsd)

  def get_color() -> tuple:
    r = max(min(rng.read(1)[0], 255), 0)
    g = max(min(rng.read(1)[0], 255), 0)
    b = max(min(rng.read(1)[0], 255), 0)
    return r, g, b

  # Calculate how many colors we need based on desired dimensions
  colorcount = width * height
  if colorcount >= 2**12:
    raise ValueError("Too much color")
    
  # Generate global color table
  globalcolortable = [get_color() for _ in range(colorcount)]
  gctlen = len(globalcolortable)*3
  out.write(bytes(sum([c+(0,) for c in globalcolortable], ())[:gctlen]))

  # Prepare to write first frame by writing image data
  repeatflag = False
  interlaceflag = False

  # Generate each frame
  # We only support non-interlaced images right now
  for iframe in range(nframes):    

    # First bit is reserved
    flagbyte = 0

    # Next three bits are the code size for the local color table, so just set that to 0
    flagbyte |= 0 << 5
    
    if interlaceflag:
      flagbyte |= 1 << 6
      
    if repeatflag:
      flagbyte |= 1 << 7
        
    # Write frame flags to output
    out.write(struct.pack('<B', flagbyte))
    
    # Calculate the size of the local color table
    colortablelen = math.ceil(math.log2(len(globalcolortable))) + 1

    # Pack tables into a single buffer
    # Since PIL doesn't care about the difference between RGB and RGBA channels
    # we can pretend all pixels have alpha channel = 255 and pack everything as RGB
    localcolortable = sum([[r,g,b]+[255] 
      for (r,g,b) in globalcolortable][:2**(colortablelen+1)], [])
    
    # Write the packed local color table to output
    out.write(bytes(localcolortable))

    # Write image data to output
    alphalen = width * height
    alphabuffer = [255]*alphalen
    rgbbuffer = [(r,g,b) for (r,g,b) in zip(*[iter(alphabuffer)] * 3)]
    pixeldata = sum([(p)+(q,) for p, q in zip(rgbbuffer, [alphabuffer])], ())
    out.write(bytes(pixeldata))