#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor. It also shows off some
# new features from Python 3.8 like f-strings :)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # Now we're going to write 256 descriptors at once
    # Using this loop means we avoid making multiple calls to read()
    for _ in range(257):
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

        # TODO: Should experiment with variations on these values too
        #       see https://github.com/decompme/decompilation-guide/issues/47
        frametype = rng.read(1)[0] & 0xf
        if frametype == 0xe:
            subblock = rng.read(1)[0]
            paletteidx = subblock >> 4
            haslocalcolortable = bool(subblock & 0x8)

            numpalcols = 2 << ((subblock & 0x7) + 1)


            desc = fr'''
                \nImage Descriptor
                    Frame type: {frametype} ({'undefined extension'})
                    Palette index: {paletteidx} {' (has local color table)' if haslocalcolortable else ''}.
                    Number of colors: {numpalcols}
                '''
        elif frametype == 0xf:
            desc = fr'''
                Image Descriptor
                    Frame type: {frametype} ({'application extension'})
                '''
        else:
            bitfield = rng.read(1)[0]
            isinterlaced = not (bitfield & 0x40)
            controlflags = bitfield & 0x1e
            paletteidx = bitfield >> 3

            lctexist = bool((controlflags & 0x8)!= 0)
            interlaceflag = bool((controlflags & 0x4)!= 0)
            sortflag = bool((controlflags & 0x2)!= 0)
            reservedbits = bool((controlflags & 0x1)!= 0)

            desc = fr'''
                Image Descriptor
                    Frame type: {frametype} ('{chr(frametype+0x30)}')
                    Interlaced: {"yes" if isinterlaced else "no"}
                    Local Color Table Exists: {lctexist}
                    Interlace flag: {interlaceflag}
                    Sort flag: {sortflag}
                    Reserved bits: {reservedbits}

                '''

        out.write(fr'''
            \n\n{desc}\n
            Raw Data:\n
            {hexdump(rng.read(10))}'''.encode())



    last_byte = out.tell()
    # At the moment this code doesn't support appending arbitrary
    # bytes onto the output so we'll have to manually compute where
    # our data will start relative to the rest of the file
    print(f'The generated gif is {last_byte} bytes long.')

    # For now we'll hardcode the number of loops to use, since we don't
    # yet know how much space remains between the final descriptor and EOF
    numloops = 0xC0

    paddinglen = (numloops * 3).to_bytes(size=2, byteorder='little')

    # Write the leftover bytes before the last descriptor
    print('Writing remaining bytes', end='', flush=True)
    out.seek(-1, 2)
    out.write(paddinglen)
    out.write(struct.pack('<B', numloops))
    out.flush()

