#!/usr/bin/env python3

import argparse
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 2 bytes of length of block (N).
#   3. N bytes of block.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, args: argparse.Namespace
    ):
    header = b'GIF89a'
    nframes = args.nframes or -1

    while True:
        if not nframes: break

        # NOTE: The first frame is always used as the control frame when
        # converting to an animated GIF with ImageMagick. This means that we
        # need to write a valid animation from the start in order for it to be
        # correctly displayed.
        nframes -= 1

        # Write the main image descriptor.
        width  = int.from_bytes(rng.read(2), byteorder='big')
        height = int.from_bytes(rng.read(2), byteorder='big')
        depth  = next(rng) & 0b11 | ((next(rng) >> 7) << 1)
        size   = int.from_bytes(rng.read(1), byteorder='big')
        packed = int.from_bytes(rng.read(size), byteorder='big')
        lctsize = 2**(packed & 0b111)
        interlace = bool(packed & 0x40)
        sorted = bool((packed & 0b100000000000) >> 12)
        gctflag = bool(packed & 0b100)
        bgindex = int(packed & 0b11111)
        pixelaspectratio = int.from_bytes(
                rng.read(1), byteorder='big'
        )
        out.write(struct.pack('>HHBBBHB', width, height, depth,
                0xf7, size, packed, bgindex, pixelaspectratio))

        # Generate a color table.
        ctable = list()
        for _ in range(lctsize):
            r = min(max(int.from_bytes(rng.read(1), byteorder='big'), 0), 255)
            g = min(max(int.from_bytes(rng.read(1), byteorder='big'), 0), 255)
            b = min(max(int.from_bytes(rng.read(1), byteorder='big'), 0), 255)
            ctable += [r, g, b]
        assert len(ctable) == lctsize * 3
        edatapos = out.tell()
        for i in range(len(ctable)):
            out.write(struct.pack('<BBB', ctable[i*3+0],
                    ctable[i*3+1], ctable[i*3+2]))

        # Write the graphics control extension.
        transpos = rng.tell()
        delay = max(int.from_bytes(rng.read(2), byteorder='big'), 0)
        transidx = int.from_bytes(rng.read(1), byteorder='big') % lctsize
        disposalmethod = next(rng) & 0b111
        userinput = False
        transparentcolor = None
        if transidx < lctsize and disposalmethod == 2:
            userinput = bool(disposalmethod & 0b100)
            if userinput:
                transparentcolor = rng.read(3)
            else:
                transparentcolor = ctable[transidx*3:(transidx+1)*3]
        edata = struct.pack('<BBHBB', 4, disposalmethod, int(delay / 10),
                int(transpos / 2), transidx if not userinput else 0)
        if not userinput:
            edata += bytes([lctsize])
            for rgb in transparentcolor:
                edata += bytes([rgb])
        edata += b'\0\0'
        edatalen = len(edata)
        out.write(struct.pack('<BBBB', 33, int(edatalen/256), edatalen%256, 0xf9))
        out.write(edata)

        # Write the image data.
        imgdata = b''
        idxpos = out.tell()
        idxtail = idxpos + 3 * width * height
        for y in range(height):
            for x in range(width):
                pixel = int.from_bytes(rng.read(1), byteorder='big')
                pixdat = b''

                if depth <= 4:
                    pos = pixel * 3

                    # Add a RLE-encoded run of pixels up until this point if
                    # either there are no more runs left to encode or doing so
                    # would save space.
                    if imgdata and (len(imgdata) >= 2 or
                        ord(imgdata[-1:])!= ord(pixel) or
                            ord(imgdata[-2:-1])!= ord(pixel)):

                        # If we're at the end of the line, flip back to the top
                        # left corner without changing the current position.
                        if x == width - 1:
                            x = 0; y = 0

                            # We also want to flip the bit order of all bytes
                            # here because they represent offsets relative to
                            # the previous one rather than absolute values.
                            tmp = ''
                            for j in range(0, len(imgdata)-1, 2):
                                tmp += chr(ord(
                                    imgdata[(j+1)%len(imgdata)]
                                        ^ 0xFF
                                ))
                            imgdata = tmp

                        imgdata += str(pixel).encode('latin-1')
                        imgdata += '\x00'.encode('latin-1')

                    elif len(imgdata) > 2:
                        imgdata += '\x01'.encode('latin-1')
                        imgdata += str(pixel).encode('latin-1')
                    else:
                        imgdata += str(pixel).encode('latin-1')

                elif depth <= 8:
                    imgdata += str(pixel).encode('latin-1')

                else:
                    # Depending on the type of color table used by this frame,
                    # some colors may have been replaced with palette indices
                    # stored earlier in the file.
                    if gctflag:
                        pixdat = struct.pack('<B', pixel)
                    else:
                        pixdat = struct.pack('<BBB', *(ctable[
                            (pixel % (lctsize))*3 :
                            (((pixel % (lctsize))+1))*3
                        ]))
                imgdata += pixdat

            # If the row isn't a multiple of two bytes long, pad it accordingly.
            rem = width % 2
            if rem: imgdata += b'\x00'*(rem - 1)

        idxdatapos = out.tell()
        out.seek(idxpos)
        out.write(struct.pack('<BBBB', 0, 0, 0, 0))
        out.seek(idxtail)
        out.write(struct.pack('<BBBB', 0, 0, 0, 0))

        out.seek(idxdatapos)
        out.write(struct.pack('<BBBB', int(idxpos/256), idxpos%256,
                int(idxtail/256), idxtail%256))
        out.write(imgdata)

        # Write the local colour table again.
        if gctflag:
            out.seek(edatapos)
            for i in range(len(ctable)):
                out.write(struct.pack('<BBB', ctable[i*3+0],
                        ctable[i*3+1], ctable[i*3+2]))

        # Write the terminator.
        out.write(b';')

    print("Wrote {} frames.".format(-nframes))


