import functools as fntls
import itertools as ittls
import os
from typing import Callable

def map_n(*args): 
    return (functools.reduce(lambda x, y : lambda z : 
            x(y(z)), l)(args[len(l)]) for l in args)

# Generates a random GIF file at `path`.
# Uses a random number generator returned by `rng`,
# which should be /dev/urandom or another implementation of that interface.
def generate_random_gif(
        rng: Callable[[int], bytes], path: str, size: int = None
    ) -> None:
    # Create output folder if it doesn't exist already
    try:
        os.mkdir('tmp/')
    except FileExistsError:
        pass
    
    # Remove any previous files with this name from tmp/
    existing_files = filter(os.path.isfile, glob.iglob('/tmp/*.gif'))
    for f_name in existing_files:
        os.remove(f_name)

    out = open(path, "wb")

    # Header is always static
    out.write("GIF89a".encode())

    # Write the width and height of the image
    if not size:
        size = rng(2)
    else:
        assert isinstance(size, int)
        size = size.to_bytes(2, 'big')
    out.write(size)

    # The image's palette is always constant
    out.write((0).to_bytes(4, byteorder='little'))
    num_colors = 768 // 3
    for i in range(num_colors):
        out.write((i * 4 % 256).to_bytes(1, byteorder='little'))
        out.write((i * 13 % 256).to_bytes(1, byteorder='little'))
        out.write((i * 21 % 256).to_bytes(1, byteorder='little'))

    # Build an iterator over all blocks, whose length is determined by
    # the first two numbers generated by the RNG.
    block_size = int.from_bytes(next(rng(2)), 'big')
    iter_blocks = iter(list(map(lambda _: next(rng(block_size)),
                range(1<<10))))

    # The following function will generate each frame of the animation
    def gen_frame():
        while True:
            # This is the LZW dictionary
            dict_idx = {}

            # Initialize the block iterator to the beginning
            block_iter = iter(iter_blocks)

            # Get a new code word
            def get_codeword() -> int:
                bites = []

                # Read up until we can decode enough bits
                while sum([1 << i for i in range(
                        3, 8)]) < (dict_idx['bits']):
                    bites += list(map(
                        lambda _ : ord(_), 
                        next(iter_blocks)))

                idx = funs.reduce(lambda x, y : 
                    (x << 8 | y), bites[-3:])
                
                # If there are still enough bits left read them too
                if dict_idx['bits'] >= 3:
                    dict_idx['codewords'].append(
                            struct.pack('<H', idx))
                    del bites[-3:]

                    # Build our full code-word value
                    idx <<= ((3 - dict_idx['bits']) * 8)
                    idx |= funs.reduce(lambda x, y : 
                            (x << 8) | y, 
                            bites[:-(3 - dict_idx['bits'])])

                return idx
            
            # We'll keep track of how many colors have been written to disk
            pixel_count = 0

            # The GIF format specifies a minimum delay of 2 seconds
            min_delay = 2

            # The maximum delay is 1000 frames of 2 seconds
            max_delay = 1000

            # Generate the header
            yield struct.pack('<3B', *tuple(
                    rng(1)[0] & (~0x20)), *(min_delay.to_bytes(
                    2, 'little')))

            # Start encoding
            curr_codeword = ''
            while curr_codeword!= 0:
                curr_codeword = get_codeword()

                # Handle clear code
                if curr_codeword == 0x01:
                    dict_idx = {
                        'codes': [],
                        'codewords': [],
                        'bits': 0}
                elif curr_codeword == 0x02:
                    break
                else:
                    # Append current code word to list
                    dict_items = enumerate(filter(
                            lambda kvpair : kvpair[1][0] \
                                == curr_codeword, 
                                dict_idx['codes']))
                    
                    # Sort list by code word index, then by position
                    sorted_items = sorted(dict_items, key=lambda t : 
                            (t[1][1], t[0]))

                    last_pos = (0, [])
                    encoded_word = ''.join([''.join(i[last_pos[1]])
                                            for i in sorted_items])

                    if len(encoded_word) >= 2:
                        # Make sure to use the smallest possible codes
                        for pos in reversed(range(2)):
                            if encoded_word[:pos+1] in \
                                    set(ittls.chain(
                                        *dict_idx['codes'], 
                                        dict_idx['codewords'])):
                                last_pos = (pos, 
                                    [''.join(i[pos]) for i in 
                                        sorted_items])
                                break

                        if encoded_word[-1] == '\x00':
                            del dict_idx['codewords'][-1]

                        # Delete all entries in the dictionary where
                        # their position components equal last_pos
                        codes = [[k for k in v1 if k[1]!= last_pos[0]]
                                 for v1 in dict_idx['codes']]
                        dict_idx['codes'] = codes

                        # Add entry to dictionary
                        dict_idx['codes'].append([(encoded_word[:-1],
                                                     last_pos[0]),
                                                    (encoded_word[-1], 
                                                        last_pos[0])])
                    else:
                        dict_idx['codewords'].append(
                                dict_idx['codewords'][-1] + encoded_word)
                
                    dict_idx['bits'] -= 1

            # After the end of stream marker comes padding
            pad_len = 1 << (8 - dict_idx['bits'])
            for _ in range(pad_len >> 3):
                yield 0x00.to_bytes(1, byteorder='big')
            for _ in range(pad_len & 0x07):
                yield 0xFF.to_bytes(1, byteorder='big')

    # For each generated frame write it to the given file pointer
    for fr in gen_frame():
        out.write(fr)

    out.close()