#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
#
# This uses an experimental probabilistic approach rather than an exact one in order
# to avoid accidentally generating a valid Gif without the intended content. The approximate
# probability of the algorithm is as follows:
#
# First, for every byte that gets written subsequently to the output it is checked whether that
# byte indicates the start of a new frame or a gif extension.
#
# For frames the probability of each subblock is:
#     1st Extension Block: 1/10
#     Other Extensions:   1/(1<<e-1)/10 where e>=1
#     Image Data:         1/(1<<(s+1)*q) where s=number of colors, q=lzw min code size
#     Graphics Control Ext.: 1/25
#     Trailer:            1/1
#
# For extensions the probability of different subblocks are:
#     Function code = 1/8
#     Extension Label = 1/4
#     Subblock length = 1/4
#     Subblock data = 1/<subblock length>
#     Trailer = 1/1
#
# This should guarantee that all valid legal outputs have approximately this distribution.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Store the current state of the system so we can reset later on
    old_pos = rng.tell()
    old_buf = rng.getbuffer().tobytes()

    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2 ** (int((lsd[4] >> 5)) + 1) * 3

    # Read global color table if present
    if gctlen!= 0:
        gct = rng.read(gctlen)
        out.write(gct)

    # TODO: Handle other images?
    while True:
        # Sample the next block
        blocktype = rng.read(1)[0]
        if blocktype == 0x3b: # stop marker
            break
        elif blocktype!= 0x2c: # unknown block type
            continue

        # Sample an image from the remaining list
        hasgct       = bool(rng.randint(0, 9) >= 1)
        hasimgdata   = bool(rng.randint(0, 9) >= 1)
        numcolors    = rng.randint(2, 255) // 8
        lzwmincodesz = rng.randint(2, 255) % 8
        size         = (numcolors * 3 + 1) + \
                       ((numcolors * 3 + 1) << lzwmincodesz) + \
                       (hasgct        << 6) + \
                       25              + \
                       1
        # Now sample how many times we'll be writing the image bytes
        niters = rng.randint(1, 3)

        # Write the full image description header to output
        idh = (0xf7 << 8 | rng.randint(10, 11)) \
             .to_bytes(1, 'big') + \
              0.to_bytes(1, 'big') + \
             (size           ).to_bytes(2, 'big') + \
             (lzwmincodesz   ).to_bytes(1, 'big') + \
              hasgct.to_bytes(1, 'big')
        out.write(idh)

        # Write the local color table if necessary
        numcolors = size // 3 - 1
        if numcolors > 0:
            colortable = rng.read(numcolors * 3)
            out.write(colortable)

        # Now write image data
        while niters > 0:
            niters -= 1

            # 1. Check if we've reached a known extension
            i = rng.tell()
            if i + 2 < len(old_buf):
                if ord(old_buf[i : i + 1]) == 0x21:
                    if ord(old_buf[i + 1 : i + 2]) >= 0xb2: 
                        # We must be looking at a graphics control block!

                        # Sample a function code
                        fcc = rng.randint(0, 7)
                        
                        if   fcc == 0:
                            # Plain text extension

                            # Skip past label
                            i += 2
                            while ord(old_buf[i : i + 1])!= 0:
                                i += 1
                            
                            # Skip past subblock length
                            i += 1
                            while ord(old_buf[i : i + 1])!= 0:
                                i += 1                            
                            
                            # Skip past subblock data                        
                            i += 1
                            while i < len(old_buf):
                                if ord(old_buf[i : i + 1]) == 0x00:
                                    # Stop when we find terminator
                                    break
                                else:
                                    i += 1
                            
                        elif fcc == 1:
                            # Application extension

                            # Sample a label
                            lbl = rng.randint(0, 3)
                            lbl |= (lbl << 8) 
                            lbl.to_bytes(2, 'big', out)
                            
                            # Skip past subblock length
                            i += 1
                            while ord(old_buf[i : i + 1])!= 0:
                                i += 1                            
                            
                            # Skip past subblock data                        
                            i += 1
                            while i < len(old_buf):
                                if ord(old_buf[i : i + 1]) == 0x00:
                                    # Stop when we find terminator
                                    break
                                else:
                                    i += 1
                                    
                        else:
                            raise ValueError("Invalid function code")
                        
                        # That's enough for now; get back to main loop
                        continue
                        
                    else:
                        # So far we only know about graphic control blocks
                        pass

                elif ord(old_buf[i : i + 1]) == 0x2C: 
                    # We've reached a frame header

                    # Compute image size
                    imsz = size - 1
                    
                    pdata = None
                    if imsz > 0:
                        # It seems like some image metadata will come with it...
                        md = [ord(b)]
                        k = max([imsz>>ii for ii in range(0, 10)])
                        pdata = []
                        while len(md) + sum(pdata) < imsz and len(pdata) <= k:
                            pdata.append(rng.randint(1, 1<<25))
                            md.extend(padata)
                        assert len(md) + sum(pdata) == imsz
                    
                        # Go back to beginning of buffer
                        rng.seek(old_pos)

                        # Write the metadata section
                        out.write(bytes(md))
                    
                    # Return to start of frame body
                    rng.seek(-sum(pdata)-1, 1)
            
            # Otherwise just emit raw data but make sure its actually within bounds
            chunk = rng.read(max([rng.randint(1, 1<<14), imsz]))
            out.write(chunk)
    
    # Reset the stream back to before the call was made
    rng.seek(old_pos)