#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # TODO: Default to a larger palette or allow it as argument?
    palette = [b'\xff\x00\x00',
               b'\x00\x7f\x00',
               b'\x00\x00\xff']

    # The first three bytes of the input are ignored 
    # and replaced with a fixed value (`GIF89a`)
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big') & 0xfff


    # Write pre-frame bytes  

    out.write(header + (datalen).to_bytes(
            3, 'big')) 

    # Construct color table by repeating colors until its full
    colorTable = palette * 256
    while len(colorTable)<(2**12):
        colorTable += palette

    # Write the constructed color table
    if len(colorTable)>=(2**12):
        raise Error("palette too large!")
    out.write(bytearray([len(colorTable)]));
    out.write(bytearray(colorTable))

    # Start graphics control block
    out.write((1).to_bytes(1,"little"));   # block size
    out.write(((datalen+4)<<1).to_bytes(2,"little"));    # delay
    out.write(b"\x00");                         # transparent index
    out.write(b"\x00");                         # terminator

    # Write image descriptor
    out.write(((datalen+4)<<1|128).to_bytes(2,"little")); # left offset
    out.write(((datalen+4)<<1|128).to_bytes(2,"little")); # top offset  
    out.write(((datalen+4)*8).to_bytes(2,"little"));         # width     
    out.write(((datalen+4)*8).to_bytes(2,"little"));            # height  
    out.write(b"\x08");                                         # local color map flag
    out.write(b"\x00");                                           # interlace flag
    out.write(b"\x00");                                           # sort flag
    out.write(b"\x00");                                           # reserved
    out.write(b"\x00\xF6")                                        # plain LZW min code size

    # Write code dictionary
    dictsize = 2**(min(12,max(8,datalen))+1);
    out.write(b'\xFF');          # end of information sequence
    out.write(dictsize.to_bytes(1,"big"));     # code table length
    for i in range(dictsize):
        val = int.from_bytes(rng.read(1),"big")
        cnt = int.from_bytes(rng.read(1),"big")

        out.write(val.to_bytes(1,"big"));
        out.write(cnt.to_bytes(1,"big"));

    # Write compressed image data    
    for i in range(datalen):
        out.write(
                ((int.from_bytes(
                    rng.read(1), byteorder="big") %
                 256)).to_bytes(1, "big"))

    # Write trailer
    out.write(b"!\3\xE5\xBF\xAE")