#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
#
# The first three bytes of the output are formatted as an integer specifying
# the length of the actual image (in bytes). This bytes are not part of the
# image itself. Thus, the resulting GIF is not valid by the spec and should be
# only used for testing purposes!
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width=120, height=40
    ) -> None:

    # Build up the header manually so we can write it at the beginning of the
    # output file (which will later also contain our randomly generated pixel
    # data.)
    header = \
        b'\x47\x49\x46' +  \
        b'\x38\x01\xf1' +  \
        b'\x00\x00\xff\x00' +  \
        b'\x00\x00\x00\xff' +  \
        b'\xe0\x00\x00\xff'

    # Write the header to the file now that we know how long it is.
    out.write(header)
    
    # Generate a bunch of pixels with the given size using random numbers from
    # /dev/urandom.
    pixels = []
    for _ in range(width*height):
        pixels += [int.from_bytes(rng.read(1),'big')]
    
    # We have finished generating the pixels, so convert them back into binary
    # form. A GIF file consists of sequential chunks where each chunk contains
    # one or more rows of pixels. An entire frame therefore represents multiple
    # chunks with each chunk containing some number of lines. To find the amount
    # of chunks we need for this particular frame, we calculate the minimum
    # number of pixels that fit into one line (since each line has to be padded
    # to whole byte boundaries.)
    num_chunks = len(pixels)//(width//8*height)
    if len(pixels)%(width//8*height)!=0:
        raise Exception("not enough pixels!")

    # Now do the same thing again but for lines instead of colors.
    num_lines = min((num_chunks*(width//8)),height)*(width//8)
    if len(pixels)%min((num_chunks*(width//8)),height)*width//8!=0:
        raise Exception("not enough pixels!")
 
    # Now we know how many chunks and lines to make. Start writing pixel data.
    # First we pad all remaining lines to full lines
    i = 0
    while i < num_lines:
        j = 0
        while j < (width//8):
            out.write(struct.pack('B',pixels[i+j]))
            j+=1
        while j < width//8:
            out.write(b'\x00')
            j+=1
        i+=width//8

    # After that we add empty lines until we reach the desired height
    while i < height * (width//8):
        j = 0
        while j < (width//8):
            out.write(b'\x00')
            j+=1
        i+=width//8
        
    # Finally, we can signal to the reader how long the image will actually be    
    out.seek(0)
    out.write(len(pixels).to_bytes(3,'big'))

    print(f"generated {len(pixels)}-pixel image ({width} x {height})")