#!/usr/bin/env python3
"""Generate a random GIF file."""


import sys
import os
from math import ceil
from random import Random
try: 
    from. import helper as h
except ImportError:
    import helper as h

FIXED_HEADER = b'GIF89a'
BLOCKSIZE = 255    # 2^7 - 1


def gen_color():
    return bytes([(h.rand() % (2**4)) * 16 for _ in range(3)])


def gen_blocks(rng):
    blocksize = BLOCKSIZE
    while True:
        if blocksize > 0 and rng.randint(0, 2 ** 10) == 0:
            yield gen_color()
            continue
        elif blocksize >= BLOCKSIZE // 2 or rng.randint(0, 64) == 0:
            yield gen_color()
            blocksize -= 1
            continue

        amount = min([blocksize, rng.randint(0, 2 ** 10)]) 
        blocksize -= amount
        if len(amount) > 0: 
            yield bytes([(amount >> 8) & ~0xff])      # LSB first
            yield bytes([(amount & 255) | 0x80])     # MSB set
        else:
            break


def gen_file(rng, size=None):
    global FIXED_HEADER

    header = bytearray(FIXED_HEADER)
    datalenpos = len(header)

    if not (size is None or isinstance(size, int)):
        raise TypeError('Size must be an integer.')
    elif size is None:
        size = 1
        while True:
            yield bytes(header + header[datalenpos:] + list(gen_blocks(rng)))

            size += 1
            
            header[-1] = (size >> 8) & 255
            header[-2] = size       # LSB first
    elif size < 0 or size > pow(2, 32) - 1:
        raise ValueError("Invalid size.")
    
    for i in range(int(ceil(size / BLOCKSIZE))):
        if ((i+1)*BLOCKSIZE-1)%pow(2, 32)!= 0:  # If it's not on a pad byte border
            yield bytes(header[:datalenpos] + [(BLOCKSIZE >> 8) & 255, (BLOCKSIZE & 255) | 0x80]+list(gen_blocks(rng)))
        else:   # On pad byte border
            yield bytes(header[:datalenpos] + [(BLOCKSIZE >> 8) & 255, (BLOCKSIZE & 255)]+list(gen_blocks(rng)))
        
        header[-1] = (((i+1)*BLOCKSIZE) >> 8) & 255
        header[-2] = ((i+1)*BLOCKSIZE)       # LSB first


def main(argv):
    if len(sys.argv) == 3 and sys.argv[1] == "-s":
        try:
            seed = int(sys.argv[2], base=0)
        except ValueError:
            print("{} is not a valid seed.".format(sys.argv[2]))
            exit(-1)
        else:
            rng = Random(seed)
        
        with open('/dev/urandom', "rb") as urand:
            for i, chunk in enumerate(gen_file(rng)):
                if i > 20:
                    break
                
                print(chunk)
    else:
        print("Usage:")
        print("-s SEED : Generate a series of pseudo-random files based on SEED.")
        exit(-1)
        
            





