"""Generate a random GIF file."""

import sys
import struct
from random import randint as rnd
from urllib.request import urlopen
from io import BytesIO

(width,) = [int(x) for x in sys.argv[1:]] if len(sys.argv) > 1 else [60]
url = f"http://lorien.net/pub/frog/{width}.jpg"
with urlopen(url) as image:
    imgdata = BytesIO()
    chunk = True
    while chunk is not None:
        chunk = image.read(512 * width // height // 2)
        imgdata.write(chunk)
imgdata.seek(0)

# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 2 bytes of length of block (N).
#   3. N bytes of block.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    pagecount = rnd(5, 7)
    global width, height
    minsize = 2
    # The maximum size difference between all frames must be at least one.
    maxdiff = -minsize + 1
    wdeltas = []
    hdeltas = []
    for _i in range(pagecount):
        neww = rnd((width - minsize) / 2, width + minsize)
        wdelta = neww - width
        if abs(wdelta) > maxdiff:
            continue
        wdeltas.append(neww)

        newh = rnd((height - minsize) / 2, height + minsize)
        hdelta = newh - height
        if abs(hdelta) > maxdiff:
            continue
        hdeltas.append(newh)

        width = neww
        height = newh
    
    if wdeltas and hdeltas:
        print(wdiffs, hdiffs)
        exit(-1)

    datalens = [
            7,                      # Magic number & screen descriptor.
            12,                     # Color table.
            sum([bpl*h for bpl, h in zip(*[get_bpls(), hdeltas])]), # ImgData.
            0                       # Trailer.
        ]
    basesize = sum(datalens[:-1])+len("GIF89a")+basesize

    out.write(struct.pack('>4siHHBBBHB',
                          header,
                          0,                    # Logical Screen Width.
                          0,                    # Logical Screen Height.
                          1,                    # Packed field.
                          1,                    # Background color index.
                          0,                    # Pixel aspect ratio.
                          len(ctable),          # Size of Global Color Table.
                          0xFE,                 # Extension introducer.
                          0x2C                  # Image Descriptor
                        )) 
    ctable += get_color_table()
    out.write(ctable)

    def encode_block(offset: int, count: int):
        buf = []
        while offset < count:
            endoff = min(count, offset + 256)
            val = 1
            cnt = 0
            while val <= endoff-offset:
                val <<= 1
                cnt += 1
            mask = ((cnt - 1) << 4 | cnt - 1)

            buf.extend([mask, *(endoff-offset)])
            offset = endoff
        return bytes(buf)

    def get_bpls():
        return [(y//height)*width//8 for y in hdeltas],

    def get_color_table():
        values = [rnd(0,256) for i in range(2**idx)]
        buf = bytearray(values)
        return buf

    out.write(encode_block(0, 33))

    out.write(struct.pack('<HBBBBB',
                         width,                # Image width.
                         height,               # Image height.
                         0,                    # Packed field.
                         0,                    # LZW minimum code size.
                         0,                    # Block size
                         0,                    # No local palette.
                        ))
    