#!/usr/bin/env python3

import struct
from typing import BinaryIO
import os

# Generates a random GIF file into `out` using the
# given RNG as source of randomness.
#
# Note that some graphic viewers have very poor performance
# when displaying very large GIFs (e.g. tries to download them etc.)
# So you should pick a reasonable amount of colors such that each
# frame contains ~128 pixels or less. You could also use a tool like gifsicle
# to optimize your final result further.
#
# When generating many frames at once make sure to call out.flush()
# frequently so that the OS has time to write to disk without blocking
# for a long period of time which will cause weird side effects.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width:int=256, height:int=256, max_colors:int=256, frames:int=1000
    ):

    # Generate palette
    palette = [
        tuple([rng.read(1)[0] for i in range(3)])
        for j in range(max_colors)]

    # Generate pixel data
    pixel_data = []
    for y in range(height):
        row = [palette[rng.read(1)[0]] for x in range(width)]
        pixel_data += row

    width_bytes = ((width*3+1)//2)
    min_lzw = b''
    for color in palette:
         min_lzw += color + b'\xFF'
    min_lzw += b'\x00\x00'

    print('Using minimum LZW dictionary of %i elements.'%len(min_lzw))

    # Write the header
    print('Writing the header...', end='', flush=True)
    header = b'GIF89a'
    datalen = len(pixel_data) + 2 + 13 # 2 bytes for color table, 13 bytes per descriptor
    out.write(header + struct.pack('>HHH', width, height, datalen))
    print('done.')

    # Write global color table
    print('Writing the color table...', end='', flush=True)
    out.write((max_colors//3)*3*b'\xFF')
    print('done.', flush=True)

    # Write frames
    while frames > 0:
        print('Generating a new frame...', end='', flush=True)
        out.write(struct.pack('<BBBBB',
            0x2c,    # Label
            0x00,    # Left Offset
            0x00,    # Top offset
            0xff,    # Width, 255 indicates "full screen"
            0xff,    # Height, ditto
            ))


        lzwbuf = bytearray(min_lzw)
        nextchar = 128
        prevcode = None
        def update_min_lzw():
            nonlocal nextchar
            lzwbuf.extend(prevcode)
            lzwbuf.append(nextchar)

            # Can't exceed 255
            if len(lzwbuf) >= 256:
                return False
            else:
                nextchar += 1
                return True

        enc = lambda c: lzwbuf[(c & 0xFF)+128] | (((~c)&0xFF+1)<<8)

        for y in range(0, height, 2):
            for x in range(0, width, 2):

                encoder = lambda c : enc(c)
                pixel = pixel_data[(y*width+x)]

                success = update_min_lzw()
                code = encoder(pixel)


                while not success and code < 0xffff:
                    prevcode = code

                    success = update_min_lzw()
                    code = encoder(pixel) + 0x100

                pixel_pair = list(reversed([(pixel >> bit) & 0b1111 for bit in range(12,-1,-4)]))
                for p in pixel_pair:
                   success = update_min_lzw()
                   code = encoder(p) + 0x100


        # Finish up LZW buffer
        lzwbuf.extend(enc(prevcode))
        lzwbuf.extend(b'\x00')

        # Figure out how much space the compressed text takes
        # NOTE: It might be possible to compress it more efficiently
        #       than this, but I haven't found anything obviously better yet...
        compressed_size = (len(lzwbuf)-128)*2+1
        print('Compressed data size:', compressed_size)
        if compressed_size >= 0x10000:
            raise Exception('LZW compressed data exceeds maximum buffer size!')

        # Last 0x100 bytes used by previous frame
        if frames == 1:
            lzwbuf.extend(b'\x00'*0x100)
        else:
            lzwbuf.extend(b'\x00'*compressed_size)

        # Add the compressed text
        out.write(bytes(lzwbuf))

        # Prepend the chunk size to allow skipping forward
        # to the next frame quickly
        out.seek(-compressed_size-2, 1)
        out.write(struct.pack('<H', compressed_size))

        print('done.\n', flush=True)
        frames -= 1

    # Finally add the trailer to finish off the output.
    # Normally this would be part of the application logic
    # when writing a gif but since this is intended only
    # for demonstration purposes we do it here directly.
    print('Adding the trailer and EOF...')
    out.seek(-1, 1)
    out.write(b';')
    print('Done!')