#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out`, with the given
# width and height of `width`*`height`. If `use_palette` is True (default False) then the color palette will be generated for you as well; otherwise, it's up to you to provide one
def generate_random_gif(
        size: tuple[int, int],
        rng: BinaryIO, 
        out: BinaryIO,
        use_palette=False):
    # Step 1: Generate Header
    # ====================================================================
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    # 
    # The Image File Header contains an initial "magic number" (signature), followed by three unsigned bytes indicating the overall logical screen dimensions of the image. This is always fixed at 78 x 54 pixels when written in the simplest form of the graphics interchange format.
    # 
    # The first byte indicates the low order bits of the fixed value of 78H.
    # 
    # The second byte indicated the high order bits of the fixed value of 78H.
    # 
    # The third byte specifies the logical screen width in the range from 2 to 65535 inclusive.
    # 
    # The fourth byte specifies the logical screen height in the same range as the last one.
    # 
    # Since the bytes specifying the logical screen dimensions are only two each, the possible ranges for these fields are thus limited to 2 to 65535.
    # 
    # The next field in the header identifies the global color table used to define colors within the image. In simplest terms, this field accepts either zero or one of the following values, which correspond respectively to no global table or one global table consisting of 256 entries:
    # 
    #    Value       Meaning
    #   --------------------------
    #       0         No global table
    #       1         256-entry global table
    # 
    # Next come eight bytes reserved for future expansion.
    # 
    # Finally comes a single flag byte that signals whether there should be an application extension introducing the NETSCAPE Application Extension to further control how the animation is displayed.
    # 
    #   Bit      Description
    #    ------   ---------------------------------------------
    #     bit 7   Global Color Table Flag
    #           Set if a global color table has been defined
    #           Clear if not
    #     other bits Reserved for future use
    # 
    # Note: in the current version of the specification, the application of global color tables is limited to displaying the images on devices capable of displaying more than 256 colors.
    # 
    # Possible values here include:
    # 
    # Byte       Meaning
    # ------------- -----------------------------------------------------------
    # B0 - B7     Signature (always equals 0x67452F31 ("GIF"))
    # B8 - B9     Logical Screen Width (LZW minimum size)
    # BA - BB     Logical Screen Height (LZW minimum size)
    # BC - BD     Pixel Aspect Ratio
    # BE          Background Color Index
    # BF          Pixel Aspect Ratio
    # C0          GCT Size (1 means that the following 256 bytes contain the actual color table)
    # C1 - FF     Color Table Data (if present)
    # 
    # 
    # For our purposes, we'll just hardcode all of those things as boring as possible
    header = b'GIF89a'

    # Step 1: Write Header
    # ====================================================================
    width, height = size
    out.write(header + width.to_bytes(2, 'little'))
    out.write(height.to_bytes(2, 'little'))
    # We're gonna assume that we don't have any background colour set
    out.write((0).to_bytes(1, 'little')) 

    # Step 2: Generate Palette
    # ====================================================================
    # A palette is *just* an array of RGB triples. Each component of
    # each triple is an integer between 0 and 255 (inclusive).
    palette = []
    # If the user wants us to make one...
    if use_palette:
        #...then let's generate a pallette based on the randomly selected colours
        count = rng.randint(1, 256)

        print("Generating palette...")
        # Fill the palette with some random colours
        for i in range(count):
            r = rng.randint(0, 255) 
            g = rng.randint(0, 255) 
            b = rng.randint(0, 255) 
            palette.append([r,g,b])

    else:
        # Otherwise, empty out the palette
        palette = []

    # Finally, write the palette!
    print("Writing palette...")
    # First off, see what the length of the palette should be
    palsize = min(len(palette), 256)
    # Then write the size
    out.write(palsize.to_bytes(1, 'little'))
    # Now loop through and write all the colours
    for col in palette:
        out.write((*col, ))

    print("Palette size:", palsize)

    # Step 3: Generate Image Descriptor
    # ====================================================================
    # After the image descriptor comes a Local Color Table (LCT), which defines a local color table for the purposes of processing the image data stream. Again, there may be no LCT stored in the image. If such a field is set, its size is determined by the next byte after that field, which is the size of the LCT.
    # 
    # Following the LCT, comes a Graphic Control Extension block that can signal the disposal method to employ, how much of a delay to place before displaying the frame, etcetera.
    # 
    # Lastly, follows the image data stream compressed via the LZW algorithm, using the preceding LCT to decompress the data.
    idlen = ((bool(len(palette)) << 7) | (
        0b00 << 4) | (
        0b1111 << 0)).to_bytes(1,'little')

    # Step 4: Write everything else
    # ====================================================================
    # For now we'll go ahead and hardcode a complete blank
    # graphic control extension. I did this using GIMP.
    ext = b'\x21\xF9\x04\x08\x00\x00\x00\x00\xFF\x00\xFE\x00';

    # We'll also hardcode a full image descriptor
    imdesc = b'\x2C\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'

    out.write(idlen); # ID Len goes here
    out.write(ext); # Extensions go here
    out.write(imdesc+b'\x00'); # Image descriptor goes here
    
    # Let's do something super simple here. Just pick a random
    # number and write it out as binary :)
    num = rng.randint(0, 1<<48)
    out.write(num.to_bytes(6, 'little'));


    # Step 5: And that's it?
    # ====================================================================
    # Well... kinda.
    # 
    # We still need to append the GIF trailer, but we haven't really touched the IO object yet so it gets ugly quickly :(
    # 
    # So we'll just close the original outfile and open another one
    # to write Trailer to instead.
    print("Closing file descriptors")
    out.close()
    print("Writing trailer", end='')
    with open(out.name, mode='ab') as f:
        f.seek(0, 2)
        pos = f.tell();
        print(" at position", hex(pos));
        
        trailer = b'\x3B'
        f.write(trailer);
