#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file with a given width in pixels (`width`)
# and height in pixels (`height`). The generated GIF will be written to
# the binary stream `out`. This generator uses the `/dev/urandom` as its
# RNG source so only works on Linux systems for now.
def generate_random_gif(
        *,
        width: int,
        height: int,
        out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Calculate some values
    flags          = bools2bitflags([False, False, True, False, True, True, False, False])
    bgcolor        = 0xff00ff
    aspectratio    = 0xfffff
    transparentidx = 0xeeeeeeee

    # Create the logical screen descriptor (LSD) field according to
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    lsd = bytearray()
    lsd += struct.pack('B',      width & 0xff)           # Screen width low byte
    lsd += struct.pack('B', (width >> 8) & 0xff)         # Screen width high byte
    lsd += struct.pack('B',      height & 0xff)          # Screen height low byte
    lsd += struct.pack('B', (height >> 8) & 0xff)       # Screen height high byte
    lsd += bytes([flags])                              # Flags
    lsd += bytes([(bgcolor     ) & 0xff])             # Background color index
    lsd += bytes([(bgcolor >> 8) & 0xff])             # Background color index
    lsd += bytes([(bgcolor >> 16)& 0xff])             # Background color index
    lsd += bytes([(aspectratio     ) & 0xff])         # Aspect ratio
    lsd += bytes([(aspectratio >> 8) & 0xff])         # Aspect ratio
    lsd += bytes([transparentidx & 0xff])             # Transparent color idx
    lsd += bytes([(transparentidx >> 8) & 0xff])       # Transparent color idx
    lsd += bytes([(transparentidx >> 16) & 0xff])     # Transparent color idx
    lsd += bytearray(b'\x00')                         # Terminator
    assert len(lsd) == 7

    # Add LSD to output
    out.write(bytes(lsd))

    # Global colors
    palette = [
        0x000000, 0x0000ff, 0x00ff00, 0x00ffff, 0xff0000, 0xff00ff, 0xffff00, 0xffffff, 
        0x555555, 0x5555aa, 0x55aa55, 0x55aaaa, 0xaa5555, 0xaa55aa, 0xaaaa55, 0xaaaaaa,
        0x000000, 0x111111, 0x222222, 0x333333, 0x444444, 0x555555, 0x666666, 0x777777,
        0x888888, 0x999999, 0xaaaaaaaa, 0xbbbbbbbb, 0xcccccccc, 0xdddddddd, 0xeeeeeeee, 0xffffffff
    ]

    # Write global color table
    out.write(bytearray([len(palette)]))
    for i in range(len(palette)):
        c = palette[i]
        out.write(bytes([(c      ) & 0xff]))              # Color code
        out.write(bytes([(c >> 8 ) & 0xff]))              # Color code
        out.write(bytes([(c >> 16) & 0xff]))              # Color code
    
    # Buffer for current frame
    buffer = bytearray();

    # Draw initial shape
    buffer += drawline(0,           0,           2);
    buffer += drawline(width // 4, 0,           3);
    buffer += drawline(width // 2, 0,           4);
    buffer += drawline(3 * width // 4, 0,           5);
    buffer += drawline(width - 1, 0,           6);

    # Draw second shape
    buffer += drawline(width // 2, height / 2, 7);
    buffer += drawline(width - 1,  height - 1, 8);
    buffer += drawline(0,          height - 1, 9);


    def _finalize():
        # Encode image data
        encoded = encode_image_data(buffer)

        # Length of image data in bytes
        datalen = len(encoded)

        # Write image data
        out.write(bytes([0x2C]))                   # Image separator
        out.write(bytes([0,0,    0,0]))                # Left
        out.write(bytes([0,0,    width % 256, 
                          (width >> 8) % 256]))               # Width
        out.write(bytes([0,0,    height % 256, 
                         (height >> 8) % 256]))               # Height
        out.write(bytes([0xf0 | ((datalen % 8) << 4),
                    (((datalen % 256 ** 2) // 8) % 256), 
                    (((datalen % 256 ** 3) // 256) % 256)]))               # Size in bytes
        out.write(bytes([0,      0]))                 # No interlace
        out.write(bytes([0,      0]))                 # Sort flag
        out.write(bytes([0,0,    0,0]))                # Local color table offset
        out.write(bytes(encoded))                    # Image data
    
    _finalize()

    # Write trailing image separator
    out.write(bytes([0x3B])) # End of images



    def drawline(x0, y0, pixel):
        buffer = bytearray()
        while x0 < width:
            buffer.append(pixel)
            x0 += 1
        return buffer

    def encode_image_data(buffer):
        encoded = []
        curbyte = 0
        bytestowrite = min(width, 255)

        while bytestowrite >= 2:
            window = buffer[-bytestowrite:]

            bit1 = window[0] & 0x01;
            bit2 = window[0] & 0x02;
            bit3 = window[0] & 0x04;
            bit4 = window[0] & 0x08;
            bit5 = window[0] & 0x10;
            bit6 = window[0] & 0x20;
            bit7 = window[0] & 0x40;
            bit8 = window[0] & 0x80;

            for i in range(1, bytestowrite):
                bit1 |= (window[i] & 0x01) << i;
                bit2 |= (window[i] & 0x02) << i;
                bit3 |= (window[i] & 0x04) << i;
                bit4 |= (window[i] & 0x08) << i;
                bit5 |= (window[i] & 0x10) << i;
                bit6 |= (window[i] & 0x20) << i;
                bit7 |= (window[i] & 0x40) << i;
                bit8 |= (window[i] & 0x80) << i;
            
            curbyte |= (bit1 << 7)
            curbyte |= (bit2 << 6)
            curbyte |= (bit3 << 5)
            curbyte |= (bit4 << 4)
            curbyte |= (bit5 << 3)
            curbyte |= (bit6 << 2)
            curbyte |= (bit7 << 1)
            curbyte |= (bit8 << 0)

            encoded.append(curbyte)

            del buffer[-bytestowrite:]

            bytestowrite = min(width, 255)
            curbyte = 0
        if bytestowrite == 1:
            curbyte |= (bit1 << 7)
            curbyte |= (bit2 << 6)
            curbyte |= (bit3 << 5)
            curbyte |= (bit4 << 4)
            curbyte |= (bit5 << 3)
            curbyte |= (bit6 << 2)
            curbyte |= (bit7 << 1)
            curbyte |= (bit8 << 0)
            encoded.append(curbyte)

        return encoded


    