#!/usr/bin/env python3

import sys
import struct
from math import floor, log2
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int,
        nframes: int = None):

    def rnd(): return int.from_bytes(rng.read(1), 'little')
    def ln2(x): return round(log2(x))

    # Seed the RNG with the desired dimensions and frame count
    seed = [width&255, width>>8,
            height&255, height>>8,
            0, 0,          # No background color
            0b10000010,    # Global Color Table + Interlace | Sorted
            0]             # Default pixel aspect ratio
    seed += list(range(1+nframes or 64))
    for s in seed:
        out.write(struct.pack('<B',s))

    # Generate an initial global color table
    colors = [[rnd(),rnd(),rnd()] for i in range(1<<ln2(min(width,height))+1)]
    out.write(bytes([*map(sum,zip(*colors)),*colors]))


    # Set up a codebook for the compression algorithm
    cb = []
    for i in range(6):
        cb.append([[rnd()%256,(rnd()&255)>>1]
                   for j in range(pow(2,i + 1)-1)])
    cb.append([])

    # Each frame will be generated as follows:
    #   Select a palette
    #   Select an LZW mincode (from first entry above)
    #   For each scanline:
    #     Select a new LZW mincode (from previous one)
    #     Choose the correct number of pixels per line
    #     Output the pixels
    while nframes is None or nframes > 0:

        palidx = rnd() % len(cb[-1])
        pal = cb[-1][palidx]

        lzwmin = rnd() % pow(2,6)
        if lzwmin==(len(cb)):
            lzwmin = rnd() % (len(cb)-1)

        lzw = [chr(x) for x in pal + [*cb,[]]]
        codebook = [[]]*pow(2,6)+[*cb]
        prevcode = None

        linesize = 1 + rnd() % (width // len(codebook))

        # Write the line counts, interlaced
        lineoffsets = [linesize]*floor(width / linesize)*height

        # Alternate between odd and even offsets unless
        # every other offset is zero
        for idx in range(1,len(lineoffsets),2):
            if sum(lineoffsets[:idx-1])!= 0:
                lineoffsets[idx] = 0


        # Handle the final partial line if necessary
        if (width % linesize)!= 0:

            off = 0
            while off < height:

                # Select a current LZW mincode and make sure that
                # it isn't equal to the previous value chosen
                curnode = lzwmin + rnd()%(pow(2,6)-lzwmin-1)
                while curnode==prevcode:
                    curnode = lzwmin + rnd()%(pow(2,6)-lzwmin-1)
                
                # Select a number of the next n bits to use
                # to determine the length of the upcoming string
                strlenbits = rnd() % 9; 
                strlensrc = rnd()
                strlenmask = 2**strlenbits-1
                strlen = 1 + (strlensrc & strlenmask)
                strlensrc >>= strlenbits
        
                # Use the selected length information to form a string
                curstr = ''.join(lzw[(curnode+k)%len(lzw)]
                                 for k in range(strlen))

                # Look up the matching LZW node index
                idx = codebook.index(curstr)
                
                # Rearrange LZW nodes after the match for easier access
                tail = lzw[idx:] + lzw[:idx]
                del lzw[:]
                lzw += tail


                # Remove outdated codebook entries
                while len(codebook) >= idx + 1:
                    remstr = codebook.pop()

                    # Skip entries that don't need updating
                    if remstr!=tail[-1]:
                        lzw = [remstr,*lzw]

                # Insert new string at start so that the most recently
                # used strings are closer together towards the front
                # of the array
                lzw = [curstr,*lzw]

                # Save the just-used node as the previous LZW code
                prevcode = curnode

                # Determine how many pixels to write on this line
                pixcount = lineoffsets[off]
                if pixcount == 0: 
                    continue
                elif pixcount > len(curstr):
                    pixcount = len(curstr)

                # Output the pixels
                out.write(chr(1 << 7 | (pixcount - 1)).encode())
                out.write(curstr[:pixcount].encode())

                # Update our line offset state
                lineoffsets[off] -= pixcount
                off += 1


        # Write the rest of the frame
        off = 0
        while off < height:
    
            # Select a current LZW mincode and make sure that
            # it isn't equal to the previous value chosen
            curnode = lzwmin + rnd()%(pow(2,6)-lzwmin-1)
            while curnode==prevcode:
                curnode = lzwmin + rnd()%(pow(2,6)-lzwmin-1)
            
            # Select a number of the next n bits to use
            # to determine the length of the upcoming string
            strlenbits = rnd() % 9; 
            strlensrc = rnd()
            strlenmask = 2**strlenbits-1
            strlen = 1 + (strlensrc & strlenmask)
            strlensrc >>= strlenbits
    
            # Use the selected length information to form a string
            curstr = ''.join(lzw[(curnode+k)%len(lzw)]
                             for k in range(strlen))
    
            # Look up the matching LZW node index
            idx = codebook.index(curstr)
        
            # Rearrange LZW nodes after the match for easier access
            tail = lzw[idx:] + lzw[:idx]
            del lzw[:]
            lzw += tail
    
    
            # Remove outdated codebook entries
            while len(codebook) >= idx + 1:
                remstr = codebook.pop()
    
                # Skip entries that don't need updating
                if remstr!=tail[-1]:
                    lzw = [remstr,*lzw]
    
            # Insert new string at start so that the most recently
            # used strings are closer together towards the front
            # of the array
            lzw = [curstr,*lzw]
    
            # Save the just-used node as the previous LZW code
            prevcode = curnode
        
            # Determine how many pixels to write on this line
            pixcount = linesize
            if lineoffsets[off]==0: 
                continue
            elif lineoffsets[off]<len(curstr):
                pixcount = lineoffsets[off]
            else:
                print("fuck")
                sys.exit(-1)
    
            # Output the pixels
            out.write(chr(1 << 7 | (pixcount - 1)).encode())
            out.write(curstr[:pixcount].encode())
    
            # Update our line offset state
            lineoffsets[off] -= pixcount
            off += 1

        nframes -= 1