#!/usr/bin/env python3

"""Generates a simple binary file containing garbage data."""

import argparse
import os
import sys
import tempfile
from typing import BinaryIO

def generate_random_binaryfile(path, size, seed):
    """Generates a random binary file with the given path
       and returns the path to it.

       The contents of the file will be generated based on the value
       of `seed`. A random number generator (RNG) will be created with
       the same initial state used to initialize the RNG for this call
       which can be used to get reproducible results across runs. The
       RNG is used to generate the output file from a sequence of
       unpredictable bytes.

       Contained within the output file will also be a single unsigned
       integer representing the used seed.

       Args:
           path: Where to put the generated file
           size: How many bytes to generate
           seed: Seed for the RNG used to create the output file

       Returns: Path to the generated file
    """
    tmpdir = tempfile.TemporaryDirectory()
    outfile = open(os.path.join(tmpdir.name, "dgb"), "wb")
    rng = RandomState(seed)

    outfile.write("# Generated from seed {}.".format(seed).encode())
    while outfile.tell() < size:
        if rng.getrandbits(16):
            # Write two pieces of unpredictable data
            outfile.write(rng.tobytes(1))
            outfile.write(rng.tobytes(1))
        else:
            # Or just one piece
            outfile.write(rng.tobytes(1))

    outfile.seek(0)
    seedint = int.from_bytes(outfile.read(), byteorder="little", signed=False)
    outfile.close()

    os.rename(outfile.name, path)
    return path, seedint


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--size", "-s", type=int, default=1 << 24,
                        help="Size of the generated binary file.")
    parser.add_argument("--output", "-o", required=True,
                        help="Where to write the generated binary file.")
    args = parser.parse_args()

    outfile, seedint = generate_random_binaryfile(
        args.output,
        args.size,
        os.urandom(8)[::-1],
    )
    with open(outfile, "rb+") as f:
        f.seek(0, os.SEEK_END)
        f.write(seedint.to_bytes(8, byteorder="little"))