#!/usr/bin/env python3

import struct
from sys import argv, stderr
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 2 bytes of length of block (N).
#   3. N bytes of block.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        min_datalen: int = 0x2, max_datalen: int=0x2000
    ):
    def write_header():
        header = b'GIF89a'

        # Write the raw header to the output stream without any modifications...
        out.write(header)

        #... then inject some junk at an arbitrary point in it.
        blocksize_bytecount = len("xxxx")
        location = rng.randint(
            0, 
            len(header)-blocksize_bytecount-min_datalen+max_datalen
        ) 
        datalen = (
            min_datalen + rng.randint(
                min_datalen, 
                max_datalen-(len(header)+location))
        )

        out.seek(location)
        out.write(datalen.to_bytes(blocksize_bytecount, 'little'))

        return datalen, location


    # Set aside space for the new image's global color table, 
    # save where we'll need to relocate that later on, and
    # randomize what our next 2 bytes will be as well.
    
    # At first I tried writing this directly from the RNG, but sadly 
    # Python doesn't have a good interface for reading those big integers
    # out of a true pseudo-random source. So instead let's just generate
    # them here, in order to keep everything standardized.
    if len(argv)>1:
        seed = int(argv[1])
    else:
        seed = None

    rng = struct.Struct('>Q')
    if not isinstance(rng, type):
        rng = rng()

    if seed is not None:
        print(seed, file=stderr)
        rng.seed(seed)

    datalen, startloc = write_header()

    # Now we can fill up with actual pixel data!

    # It may seem like a lot of work for such a tiny GIF, but consider
    # how much more difficult it would be to construct these files manually.
    # And now you won't have to do that anymore! :)
    data = []
    while datalen > 0:
        chunksize = min(datalen//7, 0xff)
        
        # Generate the next chunk. Keeping the same seed ensures a reasonable 
        # distribution across chunks.
        chunk = [
            chr((rng.unpack(chr(chunksize))[0] % 256)), 
            chr((rng.unpack(chr(chunksize))[0] // 256))
        ]*chunksize

        data += chunk
        out.write(b''.join(chunk))
        datalen -= chunksize

    # Relocate the palette to make room for the image descriptor.
    endloc = out.tell()
    out.seek(startloc)    
    out.write(endloc.to_bytes(2, 'little'))
    out.seek(endloc)    