#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Write the header
    header_fields = [b'GIF89a',
                     rng.read(4)]
    out.write(b''.join(header_fields))

    # Now we're going to write N descriptors with arbitrary
    # payloads at once, where N >= 256. We do things a bit more
    # manually than you might expect because there are a lot of
    # constraints on what makes sense in a GIF file format, like
    # needing at least one block of each type and disallowing
    # certain sequences.
    #
    # Note also that we don't keep track of or enforce any kind of order
    # between blocks of different types, since it doesn't matter -- only
    # that they exist.
    numblocks = rng.randint(256, 1024)
    offsets = []
    for i in range(numblocks):
        btype = rng.choice([b'\x2C', b'\x21'])

        if btype == b'\x21':
            subtype = rng.choice([
                b'\xF9', b'\x01\x00', b'\xFE\x00\x01', b'\xFF'])

            if subtype == b'\xF9':
                fields = [subtype,
                          rng.read(1),
                          rng.read(3)]
            elif subtype == b'\xFF':
                fields = [subtype]

                ext_len = rng.randint(5, 128) - len(fields[0])
                fields += [rng.read(ext_len)]
        else:
            fields = [btype]

            w = rng.randint(1, 512)
            h = rng.randint(1, 512)
            fields += list(struct.pack('<HH', w, h))

            lctflag = bool(rng.getrandbits(1))
            fields += [b'\x87'] if not lctflag else [b'\x00']
    
            localcolortablesize = rng.randint(2**field& ~1 | 1 for field in
                                             ['lctdepth', 'lctsize'])
            fields += [localcolortablesize]

            fields += [rng.read(localcolortablesize*3)]

            interlaceflag = bool(rng.getrandbits(1))
            fields += [b'\x40' if interlaceflag else b'\x00']

            sortflag = bool(rng.getrandbits(1))
            fields += [b'\x00' if sortflag else b'\xF0']
            
            fields += [b'\x00'] * 3

            start = out.tell()
            out.seek(start + sum(map(lambda x: len(x), fields)))
            offsets.append((start, out.tell()))

            fields = map(lambda f: len(f).to_bytes(1, byteorder='little'),
                         fields)
        
        for field in fields[:-1]:
            out.write(field)
            
        eof_len = 255 - sum(map(lambda x: len(x), fields))
        out.write(eof_len.to_bytes(1, byteorder='little'))
        
    out.seek(-1, 2)
    out.write(offsets[-1][1].to_bytes(1, byteorder='little'))
    
    # And pad up until the end of the file, which is easy enough because
    # we've already written exactly datalen bytes before ending.
    print('Padding global color table...', end='', flush=True)
    while True:
        if out.tell() == offsets[0][1]: break
        out.write(rng.read(1))
    print('\rDone padding!                      ', flush=True)