#!/usr/bin/env python3

import struct
from typing import BinaryIO
import argparse


BLOCKTYPE_EXT = b'\x21'
BLOCKTYPE_IMG = b'\x2C'
BLOCKTYPE_END = b'\x3B'

EXTENSION_TIMESTAMP = b'\xF9\x04'
EXTENSION_NETSCAPE = b'\xFF\x0B'

# Extension subtypes for NETSCAPE and TIMESTAMP
SUBTYPE_NETSCAPE_EXT = bytes([0x01, 0x00])
SUBTYPE_NETSCAPE_AUTH = bytes([0x03, 0x01, ord('N'), ord('E'), ord('T')])
SUBTYPE_TIMESTAMP     = bytes([0x01])




# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width=None, height=None
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)
    
    # Figure out the screen dimensions if they weren't specified as arguments
    if width is None or height is None:       
      w = 2 ** ((lsd[4] >> 5) + 1)
      h = 2 ** (((lsd[4] & 7) << 3) + ((lsd[5] >> 5) + 1))

      width = max(width, w)
      height = max(height, h)
    
    print("Width:", width)
    print("Height:", height)


    # Modify LSD to match our desired screen dimensions
    lsd = list(lsd)
    lsd[5] &= ~0x7
    lsd[5] |= min((height-1).bit_length() - 1, 7);
    lsd[1] = (width % 2**8).to_bytes(1, 'big')
    lsd[2] = (width // 2**8).to_bytes(1, 'big')
    lsd[3] = (height % 2**8).to_bytes(1, 'big')
    lsd[4] = (height // 2**8).to_bytes(1, 'big')
    lsd = bytes(lsd)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2 ** (int((lsd[4] >> 5))+1) * 3

    # Read global color table if present
    if gctlen!= 0:
      gct = rng.read(gctlen)
      out.write(gct)

    while True:
        blocktype = rng.read(1)[0]
        if blocktype in [0x3b, 0xff]: # stop or application extension marker
            break
        elif blocktype not in [0x21, 0x2c]: # unknown block type
            continue

        # Read the rest of the extension
        extheader = rng.read(6)
        out.write(extheader)
        
        # Parse extension header
        hasgct   = bool(extheader[0] & 0x80)
        hasimgdata = bool(extheader[0] & 0x08)
        size      = int(struct.unpack(">H", extheader[2:])[0])+1
        subtype   = extheader[:2]
       
        # Handle NETSCAPE extensions specially since it can be used as an oracle
        # Bail on these because their last argument depends on a secret input
        if subtype == EXTENSION_NETSCAPE:

          # Parse netscape extension 
          argsize = int(struct.unpack(">H", extheader[-2:]))+2
          nsc = extheader[argsize:-2]
          
          # Print some info about this extension
          print("-" * 50)
          print("Netscape Extension:")
          print("\tType:", end=' ')
          print({
            SUBTYPE_NETSCAPE_EXT : "Extension introducer", 
            SUBTYPE_NETSCAPE_AUTH : "Authentication code"
          }.get(nsc[:2], "Unknown"))
          print("\tSize:", argsize)
          print("\tData:", nsc[2:], sep='')      
          
          continue

         # Otherwise just bail out as we don't know what the data is
         else:
           pass

        # if there isn't any image data then we are done with this GIF
        if not hasimgdata:
          continue

        # Print some info about this extension
        print("-" * 50)
        print("Extension:")
        print("\tType:", end=' ')
        print({
          EXTENSION_TIMESTAMP : "Timestamp", 
          EXTENSION_NETSCAPE : "Netscape"
        }.get(subtype, "Unknown"))
        print("\tHas Global Color Table?", hasgct)
        print("\tIs Image Data Present?", hasimgdata)
        print("\tSize:", size)

        # Write the local color table if necessary
        colorsused = size//3 - 1
        if hasgct and colorsused <= len(gct):
          numcolors = { v : k for k,v in enumerate(gct) }
          colortable = []
          for i in range(colorsused):
              colortable += list(gct[i*3:(i+1)*3])
              out.write(gct[i*3:(i+1)*3])
          colortable = bytes(colortable)
          del numcolors
          
        # Now write image data 
        imgdata = rng.read(size-1)
        out.write(imgdata)