#!/usr/bin/env python3

import struct
from typing import BinaryIO


def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    """Generates a random GIF file into `out`.

    Uses the given RNG as source of randomness.
    See `generate_random_gif()` docstring for details."""
    # First thing we need to do is assemble the header
    hdr = [
        0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x01, 0x00, 0x01, 0x00, 0x80, 0x00]
    
    # Randomize the logical screen width and heights
    rng.seek(4)
    w = rng.read(2)
    h = rng.read(2)
    hdr += list(w)
    hdr += list(h)

    # We've got a static color table so overwrite its size
    hdr[11] = 0xFC

    # And add another selector which tells us what the background color
    # should be. Again we want it to be randomly chosen from any valid value
    # (see the `hdr` array declaration above)
    hdr += [0x00, 0x00] + list(rng.read(3))

    # And finally, randomize the number of frames we'll have
    rng.seek(12)
    nframes = rng.read(1)[0] % 50 + 2
    
    
    # Write everything that we assembled so far
    out.write(bytes(hdr))

    # Now let's append a bunch of frame descriptors. These work like this:
    #  1. Label
    #  2. Block Size - 10
    #     Block Data (Size including Block Size and Label)
    # So when you're constructing your own file you'd have to make sure to 
    # account for the extra four bytes needed for the label & block size. 
    # You can change the identifier to anything >= 0xF9 (block labels for 
    # application extensions) or the program will crash due to trying to 
    # process garbage memory.

    for _ in range(nframes):
        # We need a way to identify this block uniquely while still 
        # following the rules about identifiers
        lbl = 0x2C
        
        # How many bytes does the data section have?
        sz = 10 + rng.read(1)[0] % 10
        
        # Offset of the next frame's Block Header
        off = out.tell() + sz + 1
        
        
        # Append the label and computed length
        out.write(lbl.to_bytes(length=1, byteorder='big'))
        out.write((sz-10).to_bytes(length=1, byteorder='big'))


        # And write the actual frame data
        # Here I'm assuming that the only time we ever deal with an image
        # whose bit depth isn't one of 1, 2, 4, or 8 bits per pixel is if
        # the mode is LZW compression. In that case I'm generating a random
        # palette consisting of colors in RGB format.
        palette = []
        has_palette = False
        bgnd_color_idx = None
        while not has_palette:
            # Generate eight random bytes
            rand = rng.read(1)[0].to_bytes(length=8, byteorder='big')

            # If they're all zeros then break because we have a full palette
            if 0 == sum(rand):
                break
            
            # Otherwise convert them to their corresponding rgb tuple
            rgb = tuple([int(_c) for _c in
                        [rand[_i : (_i+1)]
                         for _i in range(0, len(rand), 1)]
                    ])
            palette.append(rgb)

        # Do we need to insert a palette into the file?
        if palette!= []:
            out.write(bytes([(len(palette) << 4) | 1]))
            for c in palette:
                for i in c:
                    out.write(i.to_bytes(1, 'big'))
            has_palette = True
            
            
        # Choose a random background color index. We'll assume we won't have
        # images without palettes for simplicity sake
        if has_palette:
            bgnd_color_idx = rng.read(1)[0] % len(palette)
            out.write(bgnd_color_idx.to_bytes(1, 'big'))
    
        
        # Write pixels in column order starting at top left
        for y in range(h):
            # Each row starts with a subblock label
            out.write(0x21.to_bytes(1, 'big'))
            # Then comes the compressed bitmap info
            sz = rng.read(1)[0] % 10
            out.write(sz.to_bytes(1, 'big'))

            # Next comes either the local color table or the transparent
            # color index depending on whether we had a palette or not 
            if has_palette:
                out.write(((len(palette)-1)<<4)|1)
                pass
            else:
                out.write(rng.read(1)[0].to_bytes(1, 'big'))


            # Finally comes the image data itself. It consists of
            # a run of blocks with each block representing a number of
            # consecutive pixels of the same color.
            curr_color = None
            prev_color = None

            while ((y < h)):

                # A new color can either come from the local color table 
                # if we had one, or it can also come from a previous block
                # in the case where the current block uses delta encoding
                
                # Use a different method to choose a new color for higher
                # entropy
                #new_color = rng.read(1)[0] % (len(palette)+1)
                new_color = rng.read(1)[0] % (2**max(2,sz))

                if new_color < len(palette):
                    
                    # Always pick a color from the palette
                    curr_color = new_color

                    if prev_color is None:

                        # This was the first iteration so naturally
                        # nothing needs to happen. Just write the initial 
                        # color index
                        out.write(curr_color.to_bytes(1, 'big'))
                        continue

                    # Compute the difference from the previous color 
                    diff = curr_color ^ prev_color 

                    # Use 1 byte to encode small differences
                    # (these are encoded identically to the DEFLATE spec)
                    if diff < 0x80:
                        out.write(diff.to_bytes(1,'big'))
                    elif diff <= 0xFF:
                        out.write(
                            (0x80|len([_d for _d in 
                                [_i for _i,_v in enumerate(
                                    reversed(
                                        bin(diff)[2:]
                                           .replace("0", " ")
                                           .replace("1","X")[:-1]
                                                )
                                        )
                                ] if _v=="X"])
                            ).to_bytes(1,"big"))
                        
                        out.write(
                            (((1<<(len([_d for _d in 
                                    [_i for _i,_v in enumerate(
                                        reversed(bin(diff)[2:])
                                    )
                                ] if _v=="X"])))-1)&diff
                            ).to_bytes(1,"big"))
                    else:
                        raise Exception(f"Invalid difference! {diff}")

                    # Also store the color in case we get delta-encoded later
                    prev_color = curr_color
                    
                else:
                
                    # If no palette was present, this must be to change the
                    # transparent color index. As such we should never get
                    # here unless we have already written some palette data
                    assert has_palette
                    
                    # Set the new color to the selected index
                    curr_color = bgnd_color_idx

                    # Write a single byte representing 0x00 followed by
                    # the desired index
                    out.write((0x00 ^ curr_color).to_bytes(1,'big'))


                # To keep track of the number of pixels we've already processed
                sz -= 1

                # Read enough bytes to construct a complete pixel
                pix_data = rng.read(sz) if sz > 0 else b''
                
                # Unpack pixels until we run out of data
                while len(pix_data) >= 2**(sz-1):
                    d = pix_data[-2**sz:][:2**(sz-1)]
                    out.write(d)
                    pix_data = pix_data[2**(sz-1):]
                    
                
                # Consume whatever left over data
                if pix_data!= b'':
                    out.write(pix_data)

                    
                # Check if we've finished writing pixels
                if sz==0:
                    break

                
            
            # Once we finish processing rows then move on to the next descriptor
            out.write(off.to_bytes(1, 'big'))
            
            




    # Done!