#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # Now we're going to write up to MAX_DESCRIPTORS descriptors at once
    MAX_DESCRIPTORS = 256
    for i in range(MAX_DESCRIPTORS+1):
        # The label says 10 but there are actually three more
        # bytes after the label indicating the length of
        # the block. We know it's always greater than 10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

        # TODO: Should experiment with variations on these values too
        #       see https://github.com/decompme/decompilation-guide/issues/47
        out.write(b'\x2C\x00\x00\x00\x00\x00\x00\xFB\x00\x00\x00')


        # Once you've reached our maximum number of descriptors break out
        if not i < MAX_DESCRIPTORS:
            break


        # However now let's check if we need to insert any extra bytes between them
        max_extra_size = 256*13 - (out.tell()-1)


        if max_extra_size == 0:
            # No room for anything else!
            continue

        # How much more space can we allocate?
        remaining_space = len(data)-i*13

        # Fewer than 13 bytes left to write so put everything in place
        if remaining_space < 13:
            extra_size = remaining_space

            # Reset position back to the start of where we wrote the previous one
            out.seek(-13, 1)

        elif remaining_space < max_extra_size:
            # We have enough leftover space but not even enough for another descriptor
            extra_size = remaining_space % 13

        else:
            # Alright, we should fill up to 256*13 bytes with randomness
            extra_size = min([max_extra_size, remaining_space])

            # Reset position back to the start of where we wrote the previous one
            out.seek((extra_size%13)+(13*(i)), 1)


        # Append the extra data
        extra_data = data[(i*13)+1:(i*13)+1+(extra_size)]
        print("Appending {} bytes...".format(len(extra_data)))#, end='', flush=True)
        out.write(extra_data)
        print('done.')


    # Finally add the trailer to finish off the output.
    # Normally this would be part of the application logic
    # when writing a gif but since this is intended only
    # for demonstration purposes we do it here directly.
    # See https://github.com/decompme/decompilation-guide/issues/46
    print('Adding the trailer and EOF', end='', flush=True)
    # 2 bytes for signature
    # 1 byte for version
    # 1 byte for flags
    # 2 bytes for background color index
    # 2 bytes for pixel aspect ratio
    out.write(b'\x3B\x00\x00\x00')
    print(' done')


