#!/usr/bin/env python3

import io
import random
import timeit

# Generates a random GIF file into `out` using the
# random number generator `rng`. The LSD can optionally be
# specified by providing `lsd`, otherwise one will be generated based on the
# given width / height. If `has_global_color_table` is set, then an additional
# GCT will be generated that matches the provided data.
def generate_random_gif(
        rng: io.RawIOBase, 
        out: io.BufferedWriter, 
        lsd=None, 
        has_global_color_table=False,
        global_color_table=b''
    ) -> None:

    starttime = timeit.default_timer()

    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)
   
    # Get the logical screen descriptor (LSD) field
    if lsd is None:
        lsd = bytearray(
            [
                ord('0'),  # Screen width (2 bytes)
                0x00,

                ord('0'),  # Screen height (2 bytes)
                0x00,

                0x08,      # Bits per pixel
                0xFF       # Background color index (-1)
            ] + 
            [0x00]*3   # Global color table
        )
        lsd[0] |= random.randint(0,63)//8 << 5
        lsd[1] |= random.randint(0,63)%8 << 5
       
        out.write(bytes(lsd))
    else:
        assert len(lsd)==7
        out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3
    
    # Add GCT to output if we have it
    if has_global_color_table or gctlen > 0:
        if len(global_color_table)==0:
            global_color_table = bytes([
                random.randrange(256)
                for i in range(gctlen)
            ])
        out.write(global_color_table)

    elapsed = timeit.default_timer() - starttime 
    print("Generated header in {:.1f} ms".format(elapsed*1000))

    prevx = prevy = 0
    while True:
        with open("/dev/null", "wb") as devnull:
            
            # Construct block data to write
            blocktype = random.choice([0x2C, 0x21])

            flags = 0
            flags &= ~3<<5
            flags |= random.randint(0, 3) << 5

            gceflag = False
            if blocktype==0x21:
              flags &= ~(1<<7)

              command = random.choice(["NETSCAPE2.0","COMMENT"])
              if command=="NETSCAPE2.0":
                  subcommand = random.choice([
                      bytes([0x31]), 
                      bytes([0x30])
                  ])
                  data = subcommand + \
                    int(1000*(random.random())).to_bytes(2,'little')
              elif command=="COMMENT":
                  datalen = random.randint(10, 100)
                  data = b"".join([
                      bytes([random.getrandbits(8)])
                      for _ in range(datalen)
                  ])

              gceflag = data.startswith(b"NETSCAPE2.0")
              if gceflag:
                  flags |= 1<<7
              
              fieldbyte = blocktype | flags 
              data_size = len(data)+1
            else:
                fieldbyte = blocktype 
                data_size = 2**random.randint(0,12)-1+1

                x = y = 0
                if prevx>0 or prevy>0:
                    maxdiff = min(prevx,prevy) // 2
                    x = max(-maxdiff,min(
                        maxdiff+1, 
                        random.randint(-maxdiff, maxdiff)
                    ))

                    maxdiff = min(
                        320-prevx, 
                        240-prevy 
                    ) // 2 

                    y = max(-maxdiff,min(
                        maxdiff+1, 
                        random.randint(-maxdiff, maxdiff)
                    ))
                
                prevx += x
                prevy += y

                flags &= ~(3<<7)
                flags |= random.randint(0, 3) << 7

                flags &= ~(7<<2)
                flags |= random.randint(1, 7) << 2

                xincrem = bytes([(x&1)*255,(x>>1)*255])    
                yincrem = bytes([(y&1)*255,(y>>1)*255])

                w = h = 0
                while w<=0 or h<=0:
                    w = random.randint(1, 100)
                    h = random.randint(1, 100)
                wh = [(w&1)*255,(h&1)*255,(w>>1)*255,(h>>1)*255]
    
                data = b"\x00" + \
                       xincrem + \
                       yincrem + \
                       bytes(wh) 


                # We don't want empty images
                if sum(wh)<4:
                    continue
                    
                if random.random()>0.5:
                    flags |= 1<<6
                    data+=xincrem
                else:
                    flags |= 1<<5
                    data+=yincrem

            outstream = io.BytesIO()
            outstream.write(bytes([fieldbyte]))
            outstream.write(bytes([data_size]))
            outstream.write(data)
            databytes = outstream.getvalue()
        
            out.write(databytes)

            if gceflag:
                break