#!/usr/bin/env python3

import sys
import os
import random
import math
import hashlib
import argparse
import struct


class RandomGifGenerator():
  def __init__(self):
    self._rng = None

  @property
  def rng(self):
    return self._rng
  
  @rng.setter
  def rng(self, value):
    if self._rng is not None:
      raise AttributeError("RNG may only be set once")
    self._rng = value

  def _write_byte(self, byteval):
    if isinstance(byteval, str):
      byteval = ord(byteval)
    self.rng.write(bytes([byteval]))
    
  def _initialize_color_table(self, gctlen=None):
    if gctlen is None:
      gctlen = random.choice([2**i for i in range(3)])
    gctlen *= 3
    self._write_byte(0xf7)
    self._write_byte(gctlen)
    
    colors = []
    sizemask = (gctlen // 3)-1
    for i in range(sizemask+1):
      r = random.randint(0, 255)
      g = random.randint(0, 255)
      b = random.randint(0, 255)

      colors += [r, g, b]
      
    self.rng.write(bytes(colors))
      
  def generate_random_gif(self, filename=""):
    with open("/dev/urandom", "rb") as f:
      self.rng = f
      
      gif = bytes()

      # Generate the header and add to output
      gif += b"GIF89a"
    
      # Get the logical screen descriptor (LSD) field
      lsd = self.rng.read(7)
  
      # Add LSD to output
      gif += lsd

      # Read global color table length from LSD
      gctlen = 2**(int((lsd[4] >> 5))) * 3

      # Read global color table if present
      if gctlen!= 0:
        gif += b"\x2c\x00\x00" # Start of frame
        gif += self._create_extension_block(True, False, gctlen)
        gif += self._initialize_color_table(gctlen)

      while True:
        blocktype = self.rng.read(1)[0]
        if blocktype == 0x3b: # stop marker
          break
        elif blocktype!= 0x2c: # unknown block type
          continue

        # Read the rest of the extension
        extheader = self.rng.read(6)
        gif += extheader

        # Parse extension header
        hasgct   = (extheader[0] & 0x80)!= 0
        hasimgdata = (extheader[0] & 0x08)!= 0
        size      = int(struct.unpack(">H", extheader[2:])[0]) + 1
        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        numcolors = size//3 - 1
        if numcolors > 0:
          gif += self._create_extension_block(False, hasgct, size)
          
          if hasgct:
            gif += self._initialize_color_table(numcolors)

          else:
            colors = []
            
            for i in range(numcolors):
              r = random.randint(0, 255)
              g = random.randint(0, 255)
              b = random.randint(0, 255)
              
              colors += [r, g, b]
              
            gif += bytes(colors)
            
            
        # Now write image data 
        imgheader = self._create_image_descriptor()
        gif += imgheader

        imgdata = self.rng.read(size-1)
        gif += imgdata

      
    if len(filename) == 0:      
      filename = "/tmp/{}".format(hashlib.sha256(gif).hexdigest())
    print("Writing to {}".format(filename), flush=True)
    with open(filename, "wb") as f:    
      f.write(gif)

  # This will create an image descriptor based on the current state of
  # our RNG because this isn't currently handled by any of the other classes
  def _create_image_descriptor(self, width=None, height=None):
    gd, gm = self.get_global_flags()

    x_pos = 0
    y_pos = 0
    width = max(width or (self.rng.read(2) + 1), 2)
    height = max(height or (self.rng.read(2) + 1), 2)
    left = x_pos
    top = y_pos
    flags = self._pack_bytes(gd << 4 | gm)
    interlace = random.random() <=.25
    sortflag = random.getrandbits(1)
    palette_size = min(((1<<self.rng.read(1)+1)*3), 256)
    background_color = self.rng.read(1)
    pixel_aspect_ratio = 0 # TODO implement
    image_descriptor = bytes([left, top, width-1, height-1, flags[0], flags[1], \
                      sortflag, palette_size, background_color, \
                      pixel_aspect_ratio])
    return image_descriptor


  def get_global_flags(self):
    if bool(random.getrandbits(1)):
      return self._pack_bytes(random.getrandbits(3)), 0
    return 0, random.getrandbits(3)
  
  def _pack_bytes(self, val):
    return [(val>>8)&0xff,(val&0xff)]


  def _create_extension_block(self, hasglob = False, hasloc = False, size=0):
    flags = ((hasglob << 7) | (hasloc << 6)) & 0b11100000
    size = size % pow(2, 16) 
    return bytes([0x21, flags, size % 256, size >> 8])
      
if __name__ == "__main__":
  parser = argparse.ArgumentParser()  
  parser.add_argument("--count", help="Write COUNT files instead of just one.", default=1, type=int)
  args = parser.parse_args()
  
  gen = RandomGifGenerator()
  gen.generate_random_gif()