#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Generate a random GIF file."""
from __future__ import division

import os
import sys
from io import BytesIO
from threading import Timer

# Use curses if available on the system...
try:
    from curses import ascii
    USE_CURSES = False
except ImportError:
    try:
        import msvcrt
        USE_CURSES = False
    except ImportError:
        USE_CURSES = True
        
if not USE_CURSES or len(sys.argv) < 2:
    if '-h' in sys.argv or '--help' in sys.argv:
        print("""Random Gif Generator
Generates a random GIF file with the given size limit. If no argument was provided,
it will prompt interactively for the width and height of the generated GIF.

If you want to generate a deterministically reproducible output, feed the PRNG
with the command line arguments `--seed 42` and `--quality 1`.""")

        exit()
    elif ('--seed' not in sys.argv
          or '--quality' not in sys.argv):
        print("You need to provide --seed and --quality")
        exit(-1)
        
    SEED = None
    QUALITY = None
    
    args = [arg for arg in sys.argv[1:] if '=' in arg]
    for arg in args:
        key, value = arg.split('=', maxsplit=1)
        
        assert(key in ['--seed', '--quality'])
        
        if key == '--seed':
            SEED = int(value)
        else:
            QUALITY = float(value)
            
    sys.argv[:] = [arg for arg in sys.argv if '=' not in arg]
    
else:
    SEED = None
    QUALITY = 0.2

if len(sys.argv) >= 2:
    WIDTH, HEIGHT = map(int, sys.argv[1:])
else:
    WIDTH, HEIGHT = get_size()

MAX_SIZE = min(*os.get_terminal_size(), MAX_WIDTH, MAX_HEIGHT) * ((1 << 31)-1) // SIZE

# In particular, we shouldn't trigger integer overflows due
# to unbounded arrays when generating large images
assert((WIDTH*HEIGHT)//QUALITY <= MAX_SIZE)

# For now, hardcode these values since they're really simple anyway
BACKGROUND        = (  0,   0,   0)
FRAME             = (255, 255, 255)
COLOR     = {
   'red'       : (255,   0,   0),
    'green'     : (  0, 255,   0),
    'blue'      : (  0,   0, 255),
    'yellow'    : (255, 255,   0),
    'cyan'      : (  0, 255, 255),
   'magenta'   : (255,   0, 255),
}

# These quantization tables are used when converting from RGB
# colors to indexed color
QUANT          = [[],
                  [(  0,   0,   0)],
                  [(  4,   4,   4)],
                  [(  8,   8,   8)],
                  [( 12,  12,  12)],
                  [( 16,  16,  16)],
                  [( 20,  20,  20)],
                  [( 24,  24,  24)],
                  [( 28,  28,  28)],
                  [( 32,  32,  32)],
                  [( 36,  36,  36)],
                  [( 40,  40,  40)],
                  [( 44,  44,  44)],
                  [( 48,  48,  48)],
                  [( 52,  52,  52)],
                  [( 56,  56,  56)],
                  [( 60,  60,  60)],
                  [( 64,  64,  64)],
                  [( 68,  68,  68)],
                  [( 72,  72,  72)],
                  [( 76,  76,  76)],
                  [( 80,  80,  80)],
                  [( 84,  84,  84)],
                  [( 88,  88,  88)],
                  [( 92,  92,  92)],
                  [( 96,  96,  96)],
                  [(100, 100, 100)],
                  [(104, 104, 104)],
                  [(108, 108, 108)],
                  [(112, 112, 112)],
                  [(116, 116, 116)],
                  [(120, 120, 120)],
                  [(124, 124, 124)],
                  [(128, 128, 128)],
                  [(132, 132, 132)],
                  [(136, 136, 136)],
                  [(140, 140, 140)],
                  [(144, 144, 144)],
                  [(148, 148, 148)],
                  [(152, 152, 152)],
                  [(156, 156, 156)],
                  [(160, 160, 160)],
                  [(164, 164, 164)],
                  [(168, 168, 168)],
                  [(172, 172, 172)],
                  [(176, 176, 176)],
                  [(180, 180, 180)],
                  [(184, 184, 184)],
                  [(188, 188, 188)],
                  [(192, 192, 192)],
                  [(196, 196, 196)],
                  [(200, 200, 200)],
                  [(204, 204, 204)],
                  [(208, 208, 208)],
                  [(212, 212, 212)],
                  [(216, 216, 216)],
                  [(220, 220, 220)],
                  [(224, 224, 224)],
                  [(228, 228, 228)],
                  [(232, 232, 232)],
                  [(236, 236, 236)],
                  [(240, 240, 240)],
                  [(244, 244, 244)],
                  [(248, 248, 248)],
                 ]

def rgb2index(color: tuple):
    """Converts an RGB color to a palette index."""
    r, g, b = color
    luminosity = (r * 299 + g * 587 + b * 114) / 1000
    return next(i for i, q in enumerate(QUANT) if luminosity >= q[-1][-1])[1]

def greyscale():
    """Returns a list of greylevels (indexed) that correspond
    to the PALETTE above."""
    grayscale = []
    for c in QUANT:
        indexes = [rgb2index(c)] * 256
        grayscale += indexes
    return grayscale

PALETTE         = greyscale()

# Constants below follow the specification found at https://www.w3.org/Graphics/GIF/spec-gif89a.txt
HEADER          = b'GIF89a'
LOGICALSCREENDESCRIPTOR = '<HHHHBBBBB'
GLOBALCOLORTABLE = b''
BITMAPINFOHEADER = '<HHBBBBBBBBII'
IMAGESEPARATOR  = b','
APPLICATIONEXTENSION = b'\x21\xFF'
NETSCAPEBLOCK   = b'\x01\x00\x0B'
APPNETSCAPEEXT  = b'\x01\x00\x0B'
ANIMEXTENTION   = b'\x01\x00\x00\x00\x00'
GRAPHICCONTROLBLOCK = b'\xF9'
COMMENTEXTENSION = b'\xFE'
PLAINTEXTEXTENSION = b'\x01'
ENDMARKER       = b';'



def get_size():
    """Prompts for the desired resolution of the GIF."""
    def