#!/usr/bin/env python3

import os
import sys
import json
import base64
import hashlib
import shutil
import subprocess
import tempfile
import textwrap
from defusedxml import ElementTree as ET


class GifGen():

    def __init__(self):

        self._data = []

        # Set default values here for local testing
        self._galois_fields = {
            "prime": 71,
            "degree":  5,
            "coefBits": 6,
            "roundSize": 2 ** (5 * 6 // 2 - 2),
            "modsize": 2 ** ((71 / 5) - 1)
        }


    def run(self, options={}):

        print("Starting Galois field generation")
        self.__generate_matrix()
        self.__generate_code()
        self.__run_avr_gcc()

        return True


    @property
    def _config_file(self):
        return 'galois-config.json'


    @property
    def _random_in(self):
        return './random-inputs/' if os.path.exists('./random-inputs/') else '/tmp/random-inputs/'


    @property
    def _temp_dir(self):
        return './temp/' if os.path.exists('temp/') else '/tmp/temp/'


    @property
    def _output_dir(self):
        return './output/' if os.path.exists('output/') else '/tmp/output/'


    @property
    def _parse_args(self):
        args = {}
        argv = sys.argv[:]
        if not '--' in argv:
            return False
        index = argv.index('--')
        for i in range(len(argv)):
            arg = argv[i].split('=')
            try:
                option = arg[0]
                value = arg[1]
            except IndexError:
                break

            if option == '-b':
                args['bits'] = int(value)
            elif option == '-f':
                args['field'] = int(value)
            elif option == '-r':
                args['rounds'] = int(value)
            elif option == '-p':
                args['poly_degree'] = int(value)
            elif option == '-c':
                args['num_coefficients'] = int(value)
            elif option == '-m':
                args['modsize'] = int(value)


        return args



    def __generate_code(self):
        print("Writing code to disk...")

        preamble = \
            '#include <stdint.h>\n\n' + \
            '// DO NOT MODIFY THIS FILE DIRECTLY!\n' + \
            '// This file was automatically generated by galois-field-generator.\n' + \
            '// To modify this file please refer to https://github.com/vlab-cs-ucsb/quacky/\n\n'


        with open('{}galois.h'.format(self._temp_dir), 'w+') as f:
            f.writelines([preamble])
            f.writelines(['uint{}_t poly_rounded_vals[] = {{\n'.format(self._galois_fields['coefBits'])])
            f.writelines(textwrap.indent(', '.join((str(v) for v in self._data)), '\t'))
            f.writelines(["};\n"])

        with open('{}galois-defines.h'.format(self._temp_dir), 'w+') as f:
            f.writelines(['#define FIELD {}\n'.format(self._galois_fields['prime'])])
            f.writelines(['#define DEGREE {}\n'.format(self._galois_fields['degree'])])
            f.writelines(['#define COEFFICIENTS {}\n'.format(self._galois_fields['coefBits'])])
            f.writelines(['#define ROUNDINGSIZE {}\n'.format(self._galois_fields['roundSize'])])
            f.writelines(['#define MODULUSSIZE {}\n'.format(self._galois_fields['modsize'])])




    def __generate_matrix(self):
        
        parsedArgs = self._parse_args

        configFilePath = ''
        if parsedArgs and ('field' in parsedArgs or
                           'poly_degree' in parsedArgs or
                           'num_coefficients' in parsedArgs):
            
            """This is what will be written to the config JSON file"""
            newConfig = dict({
                    "prime":            "", 
                    "degree":           "", 
                    "coefBits":         "", 
                    "roundingSize":     "", 
                    "modulusSize" :     "" 
                })
        
            if 'field' in parsedArgs:
                
                # Generate a prime from the given field size
                newConfig["prime"] = hex(int(parsedArgs['field']))
            
                # We can calculate the degree of the polynomial directly based off
                # of the prime given
                newConfig["degree"] = str(int(newConfig["prime"], 16).bit_length())
            
            
            elif 'poly_degree' in parsedArgs and 'num_coefficients' in parsedArgs:
            
                # Generate a prime that's large enough to make sure we have enough bits
                # to represent all coefficients
                coefficientBits = 0
                coeffDegree = int(parsedArgs['poly_degree'])
                while True:
                    coefficientBits += 2 * coeffDegree
                    newConfig["prime"] = hex((coeffDegree << coefficientBits))
                
                    # Make sure it's possible to fit the given amount of coefficients
                    # within the bit length of the chosen prime
                    modulusSize = int(newConfig["prime"], 16) >> coeffDegree
                    maxCoefficientBits = min([(modulusSize.bit_length()),
                                             2 * coeffDegree - 1 ])
                    if coefficientBits >= maxCoefficientBits:
                        break

                # Update the new dictionary with our newly calculated parameters
                newConfig["degree"] = str(coeffDegree)
                newConfig["coefBits"] = str(coefficientBits)


            elif 'poly_degree' in parsedArgs:
                raise Exception("Need to specify num_coefficients when specifying poly_degree.")


            elif 'num_coefficients' in parsedArgs:
                raise Exception("Need to specify poly_degree when specifying num_coefficients.")


            # Add rounding sizes so we know how big each element needs to be
            newConfig["roundingSize"] = "{0:#0{1}X}".format(((int(newConfig["prime"], 16) >> int(newConfig["coefBits"])) - 1) & (~((int(newConfig["prime"], 16) >> (int(newConfig["coefBits"])*2)))),6)
            newConfig["modulusSize"] = "{0:#0{1}X}".format((((int(newConfig["prime"], 16) >> int(newConfig["coefBits"])) - 1) + 1) & (~((int(newConfig["prime"], 16) >> (int(newConfig["coefBits"])+1)))),6)

            

            """ Check to see if there already has been a configuration file created for this run """
            configFileCreated = False
            tempDirCreated = False
            for root, dirs, files in os.walk('.'):
                for name in files:
                    
                    if name!= self._config_file:
                        continue

                    configFileCreated = True

                    configFilePath = os.path.join(root,name)
                    
                    """ If there isn't an existing config file, create one """
                    if not configFileCreated:
                        
                        """ Create temporary directory for storing configs """
                        tempDirCreated = True
                        configFileName = '{}{}'.format(self._temp_dir, name)

                        outputFileName = '{}{}'.format(self._output_dir, name)
                        if os.path.isdir(self._output_dir):
                            shutil.rmtree(self._output_dir)
                    
                        shutil.copytree('.', self._temp_dir )

                    
                        if os.path.isfile(configFileName):

                            """ Get SHA1 hash of current file """
                            shaOneHasher = hashlib.sha1()
                            with open("{}/{}".format(root, fileName), mode='rb') as f:
                                while True:
                                    buffer = f.read(65536)
                                    
                                    if not buffer:
                                        break

                                    shaOneHasher.update(buffer)
                            
                            
                            """ Compare to the new file's SHA1 hash stored in the config file """ 
                            oldShaOneHash = None
                            with open(configFileName, 'r') as f:

                                configJson = json.load(f)
                                oldShaOneHash = configJson["sha1"]

                            if oldShaOneHash == shaOneHasher.hexdigest():
                                print("Found existing copy of {}".format(fileName))
                                continue


                        """ Write contents to a new config file """ 
                        with open(configFileName, 'w') as f:                    
                            json.dump({"sha1": shaOneHasher.hexdigest(),
                                       "config": newConfig},
                                      f, indent=4)

    
                        """ Copy over the config file to output dir now that we've finished writing it """
                        shutil.move(configFileName, outputFileName)


                    """ Write contents of config file to header file """
                    with open(configFilePath,"r") as read_file:
                        configData = json.load(read_file)["config"]

                    with open('{}galois-config.h'.format(self._temp_dir),'w') as write_file:
                        for key, value in self.__dict__.items():
                            write_file.write('#define'+ key +'' +