#!/usr/bin/env python3

import hashlib
import os
import struct

from typing import BinaryIO


# Seed the PRNG with MD5 hashing of the user's username
def seed_prng():
  username = os.getlogin()
  md5digest = hashlib.md5(username.encode())

  seed = int.from_bytes(md5digest.digest(), "big") % (2 ** 64)

  return struct.unpack(">Q", struct.pack("I", seed))[0]


# Generates a random GIF file into `out` using the
# random number generator `rng`, which should be an instance of /dev/urandom
def generate_random_gif(
        rng: BinaryIO, outfile: str
    ):
  
  # Constants
  OUTFILE = open(outfile, 'wb')
  BLOCKSIZE = 2**16 // 3

  # Create a new file that starts with the standard GIF header
  write_chunk(OUTFILE, b"GIF89a", b"", False)

  # Read the first chunk from our input stream
  ctype, clen, content = read_next_chunk(rng)


  while len(content) < clen:
    # The first bit denotes whether or not the global palette is present
    gctpresent = (ctype&0xf0)==0x20
    # The second bit denotes whether or not there is a local palette
    lctpresent = (ctype&0x0f==0x09)


    # First 8 bits are reserved for future use so we'll just zero them
    ctype = ctypes & ~0xff

    # Next 3 bits represent the number of colors in our global palette
    cpcnt = (ctypes>>5) & 7

    if gctpresent:
        # Next 8 bits are reserved for future use too so let's zero them
        ctype &= ~(128|3)

        # Calculate how long the global palette will be
        gclength = cpcnt * 3


        # Read the global palette from our input stream
        content += rng.read(gclength)

    
    if lctpresent:
      # Read the local colour table from our input stream
      lctlength = contents[1]

      # Skip the 8-bit LCT size field
      contents = contents[2:]
      
      # Read the actual LCT data
      ct = contents[:lctlength]
    
      # Print the LCT to stdout
      print(*tuple(f'{c:02X}' for c in ct), sep=" ")

      # Skip over the LCT
      contents = contents[lctlength:]
    
    
    # Read the remainder of the animation and store it in memory as JPEG
    jpeg = bytes()
    while len(jpeg) < clen-BLOCKSIZE:
        # Each block has its own chunk header consisting of four bytes
        _, _, partialjpeg = read_next_chunk(rng)

        # We want all the JPEGs in the same order they came in so append each one 
        # without its header
        jpeg += partialjpeg[4:]


    # Write the full originalized chunk alongside the modified JPEG
    write_chunk(outfile, ctype, jpeg, False)
  
  
  