#!/usr/bin/env python3

import argparse
import os
import struct
from collections import deque
from typing import BinaryIO

# Generates a random GIF file into `out`.
# Options include:
#   - Data length
#   - Number of frames
#   - Random colors
# You need a suitable RNG as input. Preferably one with predictable output.
def generate_random_gif(**kwargs):
    # These parameters control the number of frames and their size.
    # Control the entropy of the random seeding by modifying these values.
    minframelen = kwargs['minframelen']         # Smallest possible frame
    maxframelen = kwargs['maxframelen']         # Largest possible frame
    nframes     = kwargs['nframes']             # How many frames?
    framelens   = [rng.randrange(minframelen,    # Generate some lengths
                maxframelen)
                    for i in range(nframes)]       # <-- Keep changing this

    # Gather the palette
    palette = []
    for i in range(256):
        r = rng.randint(0, 256) & ~1
        g = rng.randint(0, 256) & ~1
        b = rng.randint(0, 256) & ~1
        palette.append((r,g,b))
    
    # At this point, "palette" should be full. Let's create a list of frames
    # containing lists of colors, each color from the palette picked
    # uniformly at random.
    frames = []
    for l in framelens:
        d = []
        while len(d)<l:
            d += palette[rng.randint(0, 256)]
        frames.append(deque(d[:l]))

    # After having built the frames, let's build the actual file. First the
    # header...
    header = b'GIF89a'
    datalen = len(frames)*sum(map(len, frames))+10      # Don't forget initial 10
    data = struct.pack('<3sHH',
                        header[:-1],           # No version ID
                        0<<5|1<<4|0<<3|0<<2|0<<1|0,
                        datalen)

    # Then we'll append the palette. It will have 256 entries, each of them
    # composed of 3 bytes (R,G,B). That's 768 bytes total.
    data += struct.pack('<'+'BBB'*256, *sum(palette,[]))

    # Next comes the image descriptor. For now we assume a fixed pixel width
    # of 1 and a height proportional to the number of lines. I chose not to
    # make height customizable since it doesn't really affect anything except
    # possibly compression efficiency.
    data += struct.pack('<2BH',
                        0x2c,                   # Label, '!'
                        0,                      # Left position; unused
                        sum(framelens)//256,    # Top position
                        0)                     # Width of 1px

    # Finally come the frames themselves. Each frame starts with a graphic
    # control extension. By default it has no transparancy and the delay time
    # is 10ms.
    data += struct.pack('<2BHBB',
                        0x21,              # Extension label, '!'
                        0xf9,              # Graphic control label, 'f'
                        4,                 # Length of the following fields
                        0b00000001,        # Flags
                        0)                 # Delay 10ms
    # Then comes the actual frame pixels. Since we don't support transparency
    # yet we can simply copy it as-is verbatim. Padded with zeros where needed.
    for f in map(lambda x: ''.join(x)+('\x00'*
                            ((-len(x)&7)>>3)),
                frames):
        data += f.encode('latin-1')

    # And padding
    print('Padding global color table', end='', flush=True)
    while True:
        if len(data)==datalen: break
        data += chr(rng.randrange(0,256)).encode('latin-1')
    print('. Done!')

    # And finally, we write everything to disk
    filename = kwargs['filename']
    with open(filename, 'wb') as fp:
        fp.write(data)
    print("File created!")

