#!/usr/bin/env python3

import struct
from typing import BinaryIO, List


# Parses the data in `stream`, returns it as bytes
# Returns an error message on failure or None otherwise
def parse_stream(stream: BinaryIO):
  result = bytearray()

  try:
    while True:
      nextbyte = stream.read(1)
      if len(nextbyte) < 1:
        return "Unexpected end of input"
      
      opcode = ord(nextbyte)

      if opcode <= 0x2f or 0x34 <= opcode < 0xff:
        raise Exception("Opcode invalid")
      
      if opcode == 0x21: # graphics control
        pass # TODO
      elif opcode == 0x2c: # image frame
        yield result
        result = bytearray()
      else:
        # Code below handles all other opcodes
        if opcode >= 0x30 and opcode <= 0x33:
          # Code 00-07 is reserved for future use, skipping them now...
          if opcode == 0x30:
              nbytes = 3
          elif opcode == 0x31:
              nbytes = 7
          elif opcode == 0x32:
              nbytes = 1
          elif opcode == 0x33:
              nbytes = 1

          if nbytes > 0:
            padding = stream.read(nbytes+1)

            if len(padding) < 1:
                return "Unexpected end of input"
            
            opcode = ord(padding[:1])
            if opcode!= 0x00:
                return f"Expected {hex(opcode)} when expecting {nbytes} bytes of garbage"

        count = stream.read(1)
        if len(count) < 1:
            return "Unexpected end of input"
        count = ord(count)

        bytestocopy = count
        while bytestocopy > 0:
          readchunk = min(bytestocopy, 16384)
          
          chunk = stream.read(readchunk)
          if len(chunk) < 1:
              return "Unexpected end of input"

          result += chunk
          bytestocopy -= readchunk
  except Exception as e:
    print(e)
    return str(e)

  yield result


class GifWriter():
  def __init__(self, outfile: BinaryIO):
    self.outfile = outfile


  def close(self):
      self.outfile.close()


  def new_frame(self, width: int, height: int, transparentcolor: int=-1):
    # Write subblocks
    wsubblocks = [width % 256, width // 256]
    hsubblocks = [height % 256, height // 256]

    # Write header
    self.outfile.write([
      0x2C, # image separator
      sum([wsubblocks[-1], wsubblocks[-2]]) & 0xFF,
      sum([hsubblocks[-1], hsubblocks[-2]]) & 0xFF,
      0,     # left position
      0,     # top position
      wsubblocks[0],
      wsubblocks[1],
      hsubblocks[0],
      hsubblocks[1],
      0      # no flags set
      ])


    if transparentcolor!= -1:
      self._write_graphiccontrol(transparentcolor=transparentcolor)


  @staticmethod
  def _pack_subblock(subblockid: int, value: List[int]):
    if isinstance(value, list):
      value = bytes(value)
    
    checksum = sum(value)
    packet = [
      subblockid, 
      len(value),
      value,
      checksum % 256,
      ]

    return packet
  
  def _write_graphiccontrol(self, disposalmethod: int=0, userinputflag: bool=False, delaytime: int=0, transparentcolor: int=0):
    packed = self.__class__._pack_subblock(0xf9, [disposalmethod, userinputflag, (delaytime//60)%256, delaytime%256, transparentcolor, 0])
    self.outfile.write(packed)



  @staticmethod
  def _write_imagedescriptor(left: int, top: int, width: int, height: int, lctused: bool, interlaceflag: bool, sortflag: bool, localcolortablesize: int=-1):
    assert localcolortablesize == -1 or ((localcolortablesize % 256) == localcolortablesize and (localcolortablesize // 256) == localcolortablesize), 'Expecting localcolortablesize to be either -1 or multiple of 256'

    packed = [
      0x2c,        # image separator
      left&0xff,
      left>>8,
      top&0xff,
      top>>8,
      width&0xff,
      height&0xff,
      0 | (localcolortablesize << 4),
      interlaceflag<<6 | sortflag<<3 | lctused,
    ] 

    return packed

  @staticmethod
  def convert_rgb_to_pallete(rgb: List[int]): 
    palleteentry = []
    palleteentry.append(((rgb[0]&0xF8)>>3)<<6)
    palleteentry.append(((rgb[1]&0xFC)>>2)<<3)
    palleteentry.append(((rgb[2]&0xF8)>>3))
    return palleteentry