#!/usr/bin/env python3

import io
import struct
from typing import Iterator


# The header of the generated files
HEADER = (b"GIF87a", b"GIF89a")

# A list containing all possible frame durations in ms
# and their block size
DURATIONS = ((0, 0), (33, 1), (67, 1), (100, 1),
             (134, 1), (150, 2), (200, 2), (250, 2),
             (300, 2), (500, 2), (1000, 2))

# Make sure that every color has its own bitfields on LZW encoding
COLORS = [(2 ** i - 1, 1 << i) for i in range(9)]


# Use a constant for the maximum size to be able to
# display it nicely with printf()
MAX_SIZE = max([d[1] for d in DURATIONS]) * 8


class PseudoRNG:
    def __init__(self):
        self._state = 0

    # Generate a byte from /dev/urandom seeded by XORing
    # the current state with the new one
    def _generate_byte(self) -> bytes:
        ret = self._state ^ (self._state >> 9) & 0xff
        self._state ^= ret
        return ret.to_bytes(1, "little")

    # Read `size` bytes from /dev/urandom
    def read(self, size:int)-> bytes:
        buf = b""
        while len(buf)!= size:
            buf += self._generate_byte()
        return buf

    # Return an integer between min and max
    def randint(self, min:int, max:int) -> int:
        sizebits = None
        if min >= 0:
            sizebits = (max+1-min).bit_length()
        else:
            sizebits = (-min).bit_length()

        nbytes = sizebits // 8 + bool(sizebits % 8)
        mask = pow(2, sizebits) - 1

        buf = self.read(nbytes)
        val = int.from_bytes(buf, "little")
        val &= mask

        ret = val % (max - min + 1) + min

        assert min <= ret <= max,\
                f"{ret} not between [{min}, {max}]"\
               .encode("utf8")
        return ret


    # Put two bytes together as a short and swap them
    @staticmethod
    def combine_short(high:bytes, low:bytes,
                     signed=False)->int:
        high, low = int.from_bytes(high, "big"),\
                    int.from_bytes(low, "big")
        val = (high << 8) | low
        if signed and val > MAX_SIZE//2:
            val -= MAX_SIZE
        return val



# Create a dictionary which can store values up to `size` bits
# See https://stackoverflow.com/a/34241075/867223
def create_dict(size: int) -> dict:
    """Create a dictionary with at least ``size`` keys."""
    return {-1: -1, *(i : i for i in range(pow(2, size)))}




def gen_colors(rng: PseudoRNG, num_colors: int)->list:
    colors = []
    for _ in range(num_colors):
        gct_size = rng.randint(2, 4)
        gct = []
        for _ in range(gct_size*3):
            val = rng.randint(0, 255)
            gct.append(val.to_bytes(1, "little"))
        gct.extend((0xFF,)*(7-len(gct)%7))
        colors.append(bytes(gct))
    return colors



# Returns the extension block used when generating LZW frames
def get_lzw_extension(image_width:int,
                      image_height:int,
                      loop_count:int=-1)->bytes:
    global COLORS
    # Find the first valid combination of Colors and Bits Per Pixel
    # such that we can fit within our bounds without overflowing
    # This is unlikely but still possible
    for c, bpp in COLORS:
        if image_width*(image_height*((bpp+7)//8)) <= MAX_SIZE:
            break
    return b"\xF9\x04" + \
           image_width.to_bytes(2, "little")+\
           image_height.to_bytes(2, "little")+\
           b"\x80" + c.to_bytes(1,"little")+\
           b"\x00" + bpp.to_bytes(1, "little")\
           + ((loop_count.to_bytes(1, "little")
              ).encode('iso-8859-1') if loop_count>=0
               else b"")





# Returns the application extension with the given identifier
# and the given subblocks
def get_application_ext(identifier: str,
                        subblocks: list) -> bytes:
    #! = endianess, H = unsigned short, B = char
    # s = string (terminated with NUL), p = pascal style string
    ext_str = "!H4s{0}s".format(len(subblocks)+1)

    id_len = len(identifier)
    ext_len = sum(map(lambda x: len(x), subblocks))+id_len+2
    subblock_lens = map(lambda x: len(x),"!" +\
                          ",".join(["p"]*len(subblocks)).encode()\
                         )

    total_len = reduce(lambda acc, elem: acc+elem,
                       subblock_lens,
                       ext_len)

    return ext_str.format(total_len).\
          encode()+struct.pack("!H",
                              id_len)\
                         +identifier.encode()+(
                            b"".join(subblocks))






# Yields the blocks of the given LZW encoded image
def gen_lzwframe(in_: BinaryIO,
                 lzw_table: dict,
                 width:int, height:int,
                 background_index:int=None,
                 transparent_index:int=None,
                 delay:int=-1,
                 dispose:int=-1)->Iterator[bytes]:
    global HEADER, DURATIONS

    duration_ind, duration_mult = next(\
            filter(lambda t: t[1]>delay,
                   enumerate(DURATIONS)), (None,None))
    if duration_ind == None:
        duration_ind, duration_mult = 0, 2

    def _get_block():
        # Read the next code from `in_`
        code = in_.read(1)[0]

        # If there are no more codes then yield the last block
        if code == 0:
            yield b"\x00"
            return

        key = ""
        while True:
            prefix = key[:-1].encode() if key else b''

            # Check whether this code appears in the table
            # If yes yield the corresponding value in the table
            # otherwise yield the prefix followed by the new code itself
            val = lzw_table.get(prefix + code.to_bytes(1,'little'))
            if val!=None:
                key+=code.to_bytes(1,'little').decode("utf8")
                yield val.to_bytes(1, 'little')
                break
            elif len(key)==0 or key[-1:] in lzw_table:
                yield prefix + code.to_bytes(1,'little')
                break
            else:
                key+=code.to_bytes(1,'little').decode("utf8")


    # Generate the graphics control extension
    gce = lambda: b"\x21\xF9\x04"+\
                  b"\x08"+\
                  (((duration_ind//duration_mult).to_bytes(1,"little")
                    )).encode()+"\x00"*3+\
                  ((dispose.to_bytes(1,"little")
                   )).encode()+"\x00"*3+\
                  "\xF0" if transparent_index is not None else b"\x00"

    # Header and Color Table
    imghdr = HEADER[rng.randint(0,1)]+b"\x00"
    colortable = gen_colors(rng, 2**bpp)-1

    # Build the lzw dictionary
    lzw_dict = {(i+1).to_bytes(1,'little'):colortable[c][1]
                for c, i in enumerate(range(1<<bpp))}

    # Build the actual lzw image
    lzw_image = gce()+imghdr+b''.join(_get_block())

    # Get Application Extension
    app_ext = get_application_ext("NETSCAPE2.0",\
                                  ["\x01","\x0F",
                                   "\x00\x00\x00",
                                   "#BLEND"])



    totallen = len(lzw_image) + len(app_ext)
    yield b"\x2C"+\
          (width.to_bytes(2