#!/usr/bin/env python3
## \file          generate-random-gifs.py
# \brief         Generate random GIF files given RNG input

from io import BytesIO
class LimitedRandomBytesIO(BytesIO):
    """A class that acts like BytesIO but reads from an additional source"""

    def __init__(self, maxsize: int, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.__maxsize = maxsize

    # This is how you read bytes from it!
    def read(self, size=-1):
        if size == -1 or len(self.getvalue()) < self.__maxsize:
            return super().read()
        else:
            return super().read(min((len(self.getvalue()), size)))


# Generates a random GIF file into `out` using the
# following format:
#   Header:
#     1. 4 bytes of GIF magic number "GIF89a".
#   Logical Screen Descriptor:
#     2. Unsigned short width in pixels (>=0 and <=65535).
#        NOTE: No where does this value need to be actually used for anything...
#     3. Unsigned short height in pixels (>=0 and <=65535).
#        NOTE: No where does this value need to be actually used for anything...
#     4. Packed field:
#           Bit 1   Size of global color table. If set, there is a local color table
#                   immediately afterwards. If not set, the image has no color table.
#           Bits 2-7 Reserved for future use; current must be zero.
#     5. Byte pixel aspect ratio. The value is really only significant if the
#        global color table flag is set. We set it to something valid regardless.
#       NOTE: We don't support transparency so we don't care about setting the bit
#            declaring presence of a color table which could just always be false. 
#  Trailer:
#     See above for the header except with less data.
def generate_random_gif(rng: BinaryIO, out: BinaryIO, minwidth=1, maxwidth=100, minheight=1, maxheight=100, mindepth=1, maxdepth=8):
    header = b'GIF89a'
    
    width = rng.randint(minwidth, maxwidth)
    height = rng.randint(minheight, maxheight)
    depth = rng.randint(mindepth, maxdepth)

    lsd = struct.pack('<HHBxB', width, height, ((depth & 0x07) << 4) | 0xF7)

    out.write(header)
    out.write(lsd)