#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, imgwidth=2048, imgheight=1536
    ):

    # Header is always 6 bytes long in GIF files (7 if an extension block begins with 'NETSCAPE'):
    header = b'GIF89a'

    # The Image Descriptor section contains some metadata about how to display the image
    id_bitfield = bytes([imgwidth%256]) + \
                  bytes([(imgwidth >> 8)%256]) +\
                  bytes([(imgwidth >> 16)%256]) +\
                  bytes([imgheight%256]) +\
                  bytes([(imgheight >> 8)%256]) +\
                  bytes([int('10', base=2)]) + \
                  bytes(['\x00'])           # No global color table used and local color table not present
    # ID block size is always 7 bytes; the first byte of that field indicates how many bytes are left until the end of the file
    id_blocksize = len(id_bitfield).to_bytes(1, byteorder='big') + b'\x00\x00'
    # Afterwards there's the actual Image Descriptor data
    id = id_blocksize + id_bitfield
    
    ext_blockheader = b'\x21\xff\x0b'       # Extension Block Header
    netscape_header = b'NETSCAPE2.0'      # NETSCAPE Application Extension marker followed by version string
    loop_count = b'\x03\x01'                # Loop forever next frame after this one
    end_ext_block = b'\x00'                 # End the block with null terminator
    ext_block = ext_blockheader + netscape_header + loop_count + end_ext_block

    # Next comes the LZW compression data for our pseudorandom image
    lzwmincodesize = b'\x02'               # Minimum code size == 2 bits (the lowest allowed value)
    lzwtablesize = bytes([imgwidth*imgheight])+b'\x00'   # Number of symbols in the table == width * height pixels
    lzwhuffmancode = b'\xd8\xe0\x00\xf0\xc0'    # Huffman codes generated from Floyd-Steinberg dithering algorithm
    lzwworkbufferlen = b'\x02'              # Work buffer length is only two bytes long -> just use fixed length decoding here

    # Since we don't have any real LZW compressed data yet, let's just encode
    # a single white pixel as a demonstration
    uncompressed = b'\x01\x01\xff\xff\xff'

    # Total length of all subsections so far
    totalsize = len(lzwmincodesize + lzwtablesize + lzwhuffmancode + lzwworkbufferlen + uncompressed)
    # We need to fit the whole thing into an unsigned short integer to specify it later on
    assert totalsize <= 2**16 - 1 
    # We also write a special empty character at the beginning of every chunk because of the way GIFs work
    chunk = b'\x00'                   # Empty prefix character
    chunk += totalsize.to_bytes(2, byteorder='little')         # Size of the chunk without the initial prefix character
    chunk += b'\xfe\x00'             # Special "subblock separator"
    chunk += lzwmincodesize          # Min code size subsection
    chunk += lzwtablesize            # Table size subsection
    chunk += lzwhuffmancode          # Actual image data
    chunk += b'\x00\x3b'              # Separator between image data blocks and application extensions
    chunk += ext_block               # Application Extension Chunk

    # Finally, put everything together
    outbuf = header + id + chunk
    datalen = len(outbuf)
    out.write(outbuf)
    print(f"generated {datalen} bytes (including headers)")