#!/usr/bin/env python3

import zlib, struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # First write the GIF magic signature
    out.write(b'GIF89a')

    # Then get the logical screen descriptor (LSD). This contains three
    # important pieces of information that are needed by the rest of the
    # file; frame width, height, and color depth.
    #
    # The first two bytes specify the width as the most significant bit-pair
    # followed by the least significant bit pair. Since our images will be tiny,
    # these values are fixed at 64px each
    lsd = [0x40, 0x64, 0x00,
          0x64, 0x00, 0xf7, 0x00]
    out.write(bytes(lsd))

    # Now read the global color table length which is specified as follows:
    #   exp-12 : 0 <= exp < 14 (hence the value + 12). Indicates how many bits 
    #           must be set in the following byte
    #           0b110 => 1 << 12 = 4K colors
    #           0b1110 => 1 << 14 = 16K colors etc.
    #   nbitsc : Number of bits per color channel in global color table
    #   gct    : Global color table or empty list if none exists
    gctlen = 2**((lsd[4] >> 5) + 12) * 3
    gct = []
    for i in range(gctlen // 3):
      r = int(ord(rng.read(1))/16)*16
      g = int(ord(rng.read(1))/16)*16
      b = int(ord(rng.read(1))/16)*16

      pixels = r|g|b
      gct += [(r, g, b)]


    # Finally output the global color table if one was generated
    if len(gct) > 0:
      out.write(b''.join(map(lambda x: bytes([x]), map(sum, zip(*gct)))))

    # Now write all the blocks for each frame, starting with an Image Block
    # Indicator (blocktype=2C), then the image data to follow.
    # Each frame is terminated by a "trailer" block (end of file indicator
    # byte=3B)
    fbi = b'\x2c\xff\x00\x00\x00\x00'
    out.write(fbi)

    # As mentioned above, start off with an initial code size where
    # every repetition results in a single byte.
    # This means that subsequent codes can only appear once.
    maxcodebits = 4
    initial_codesize = 2 ** maxcodebits
    maxcodes = initial_codesize + 1
    exponents = tuple(range(maxcodebits, -1, -1))

    # Bits used so far for current code
    curcodebits = maxcodebits

    # Buffer storing uncompressed bytes.
    buf = []

    # Encoder lookup tables (see RFC1951 section 3.2.3)
    litlenlut = []
    distlut = []

    # For now just make sure everything decompresses correctly
    def compress():
      chunk = bytes(buf)
      compressed = []
      pos = 0

      while pos!= len(chunk):
        # Find longest match
        longestmatch = ''
        repstart = 0
        repend = 0
        for repstart in range(pos):
          searchstr = (
              chunk[pos:]
              .split(chunk[repstart:], 1)[0])

          repend = repstart+len(searchstr)-1

          if len(longestmatch)<len(searchstr):
            longestmatch = searchstr

        # If we found something...
        if len(longestmatch) >= 3:
          # Output the copy mode marker
          compressed.append(b'\xfc')
          # Output match offset
          compressed.append(bytes([(repend&255),(repend>>8)]))
          # Output match length minus 3
          compressed.append(bytes([len(longestmatch)-3]))
          # Increment buffer pointer accordingly
          pos += (len(longestmatch)+1)

        else:
          # No match was found, output literal byte
          compressed.extend(
              ([b'\xfe'] 
                + [bytes([chunk[pos]])][
                    chunk[pos] > 255
                ] 
              ))

          pos+=1

      # We've made it through the whole compression process and have
      # accumulated enough literals to fill up the entire buffer. Time to 
      # flush!
      compressed.insert(0, bytes([initial_codesize]))
      
      # Make sure to only use what we need
      compressed = compressed[:curcodebits]

      return b''.join(list(filter(bool, compressed))).ljust(
        2**curcodebits, b'\x00')

    running = True
    while running:
      # Create a new huffman tree
      codetree, codes = {}, {}
        
      # Read the next byte.
      # If EOF encountered before getting here, exit loop
      try:
        b = ord(rng.read(1))
      except:
        break
    
      # Initialize lengths dictionary based on previous byte
      if b!=0:
        lens = {
            1:(1, 0),
            2:(2, 0),
            3:(3, 0),
            4:(0, 1),
            5:(0, 2),
            6:(0, 3),
        }
      else:
        # End of file reached
        running = False
        break

      # Loop until everything has been decoded
      while len(lens)!=0:
        # Determine best match
        maxprob = 0
        maxcode = ('','')
        for k,v in lens.items():
          prob = (
              3+(k<6)*5
              *(v[0]+1)*(v[1]+1)
              /codelength)

          if prob>maxprob:
            maxprob = prob
            maxcode = (k, v)
            
        # Move node to front of queue
        node = k
        del(lens[node])
          
        # Calculate where to insert it such that its children nodes are 
        # always less than the parent node
        kold, vold = maxcode
        k, v = kold, vold
        i = (k<<2)+v[0]
        j = (k<<2)+v[1]
        while i in codetree:
          i-=1
        while j in codetree:
          j-=1
        codetree[j]=i
        codes[(k,v)]=(i,j)
              
        # Update child nodes and their lengths
        if v[0]<3:
          lens[(k<<2)+v[0]] = (v[0]+1,0)
        if v[1]<3:
          lens[(k<<2)+v[1]] = (v[1]+1,0)
      
      # Encode symbol
      sym = codetree.get(-1,-1)
      code = [sym]
      while 1:
        i,j = codes[sym]
        sib = code[-1]
        if sib<=i:
          code.extend([sib,j])
        elif sib>=j:
          code.extend([i,sib])
        else:
          raise Exception('this should never happen')
        if sym==0:break
        sym = codetree.get(sym-1,-1)

      if len(code)>maxcodebits:
        # Flush buffer with largest codebits between old and new
        diff = len(code)-maxcodebits
        buf.extend(compress())
        buf.pop()
        curcodebits -= diff
        maxcodebits = len(code)

      # Output code
      buf.append(code.pop(0))
      for n in reversed(code):
        buf.append(n >> curcodebits)
        buf.append(n % pow(2, curcodebits))
        
    # Flush buffer
    buf.extend(compress())
  
    # Output remainder of buffer
    out.write(b''.join(map(lambda x: bytes([x]), buf)))