#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO


# The maximum length of a block in bytes;
# we're generating 2**num_blocks long data blocks
MAX_BLOCK_LEN = (1 << 40)
    
# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom);
# this will be written as sequential blocks of
# up to MAX_BLOCK_LEN each.
def generate_random_gif(
        num_blocks: int, rng: BinaryIO, out: BinaryIO):

    # Write the "logical screen descriptor", which normally
    # contains color table information and dimensions but not here
    screen_desc = b'\x2c\0\0\0\0,\0\0\0\7\0\0\0\0\0\0\x80\xff\xff\xff'
    out.write(screen_desc)
    
    # Generate `num_blocks` large chunks of random bytes and write them
    for i in range(num_blocks):
        
        # How many bytes should be generated next?
        data_len = min(
                MAX_BLOCK_LEN,
                2 ** (i+6))
            
        # Generate that chunk!
        data = rng.read(int(data_len))

        # OK now output it; I've found no official documentation on how long
        # it should take to encode the length field so we'll just try different
        # lengths until we find one that outputs successfully...
        while True:

            # Try encoding two byte length first, then four byte length 
            if len(data) < 255:

                # Encode length as two byte unsigned integer
                fmt = f'>B{len(data)}s'
                encoded_len = struct.pack(fmt, len(data), data)
                
                try:

                    # If this succeeds we can exit the loop
                    out.write(encoded_len)
                    break

                except Exception as e:

                    pass
            
            else:

                # Encode length as four byte unsigned integer
                fmt = f'>I{len(data)}s'
                encoded_len = struct.pack(
                        fmt, 
                        len(data), 
                        data)
                
                try:

                    # If this succeeds we can exit the loop
                    out.write(encoded_len)
                    break

                except Exception as e:

                    pass
                    
            raise e
            
    print(f"generated {num_blocks} blocks")

    
if __name__ == "__main__":
    
    # Check command line arguments
    if len(sys.argv)!= 2 or \
       type(int(sys.argv[1])) is not int:
    
        print("Usage:")
        print(" $./generate-random-gif.py NUMBLOCKS")
        print()
        print("where NUMBLOCKS is the desired number of data blocks")
        sys.exit(1)
        
    # Read argument from command line
    num_blocks = int(sys.argv[1])
        
    with open("/dev/urandom", 'rb') as rng:
    
        with open("./tmpfile", 'wb') as out:
        
            generate_random_gif(
                    num_blocks=num_blocks,
                    rng=rng,
                    out=out)
