#! /usr/bin/env python3

import os
import sys
import string
import struct
from pathlib import Path
from typing import BinaryIO


class RandomGifGenerator():
    def __init__(self, targetfile: str):
        self._targetfile = targetfile
        self._rng        = None

    @staticmethod
    def get_suitable_name(targetdir: str)->str:
        candidate_names = []
        for fname in os.listdir(targetdir):
            basefname,_ = os.path.splitext(fname)

            if not basefname[0] == 'r':
                continue

            try:
                index = int(basefname[1:])
            except ValueError:
                continue
            
            if len(candidate_names) <= index:

                max_index = 0
                for candname in candidate_names:
                    _,candidx = os.path.splitext(candname)

                    if candidx < max_index:
                        continue
                    
                    max_index = candidx
                
                candidate_names += [None]*(max_index+1-len(candidate_names))
                
            candidate_names[index] = fname
        
        while True:
            name = ''.join([
                choice(string.ascii_letters) 
                for i in range(8)])

            if any(not x for x in candidate_names):
                return f'{name}.gif'

            if candidate_names[-1].startswith(name):
                continue
            
            break
    
    def randomize_output(self):
        if isinstance(self._targetfile, str):
            exitstatus = os.system(f'randomize --input={self._targetfile} > {self._targetfile}')
        else: 
            raise NotImplementedError("Unsupported type")


        if exitstatus!= 0:
            print(f"Randomization failed ({exitstatus})")
            # In case randomize fails we still want to create a working gif
            # So let's just put garbage content in the header instead...
            # Sorry :P
            pass
    
    def initialize_rng(self):
        self._rng = open('/dev/urandom','rb')

    def shutdown_rng(self):
        self._rng.close()

    def generate(self):
        if not hasattr(self,'_rng'):
            raise RuntimeError("_rng has not been initialized!")

        print("Creating a random gif",end='')
        sys.stdout.flush()

        generate_random_gif(
                self._rng,
                self._targetfile)
        
        print(f". Done! Outputted to \"{os.path.basename(self._targetfile)}\".")


            
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    print('Writing the header', end='', flush=True)
    out.write(header + data[:10])
    print(' done.', flush=True)

    # Now we're going to write 256 descriptors at once
    # Using this loop means we avoid making multiple calls to read()
    for _ in range(257):
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

        # TODO: Should experiment with variations on these values too
        #       see https://github.com/decompme/decompilation-guide/issues/47
        out.write(b'\x2C\x00\x00\x00\x00\x00\x00\xFB\x00\x00\x00')


    last_byte = out.tell() - 1
    # At the moment this code doesn't support appending arbitrary
    # bytes onto the output so we'll have to manually compute where
    # our data will start relative to the rest of the file
    print(f'The generated gif is {last_byte} bytes long.')

    # For now we'll hardcode the number of loops to use, since we don't
    # yet know how much space remains between the final descriptor and EOF
    numloops = 0xC0

    paddinglen = (numloops * 3).to_bytes(size=2, byteorder='little')

    # Write the leftover bytes before the last descriptor
    print('Writing remaining bytes', end='', flush=True)
    out.seek(-1, 2)
    out.write(paddinglen)
    out.write(struct.pack('<B', numloops))
    out.flush()