#!/usr/bin/env python3

import struct
from os import urandom as randbytes
from sys import stderr
from typing import Optional, Union, BinaryIO

class RNG:
    def __init__(self):
        self._rng = None # type: Optional[BinaryIO]

        try:
            self._rng = open('/dev/urandom', 'rb')
        except FileNotFoundError:
            pass

        if not self._rng:
            print('warning: could not find /dev/urandom',
                file=stderr
            )

            self._rng = _FakeRandom()


    def read(self, n: int) -> bytes:
        return self._rng.read(n)



class _FakeRandom:
    def __init__(self):
        self.__pos = 0


    def read(self, count: int) -> bytes:
        result = []

        while len(result) < count:
            new_byte = ord('A') + ((self.__pos >> 6) & 0x1F)
            self.__pos += 1

            result.append(new_byte)


        return bytes(result[:count])



# Generates a random GIF file into `out` using the
# following format:
#   1. 4 bytes of GIF magic number "GIF89a".
#   2. 2 bytes of length of block (N).
#   3. N bytes of block.
def generate_random_gif(
        rng: RNG, out: BinaryIO, size_range=(5, 50),
        max_frames=None
    ):
    header = b'GIF89a'
    datalen = None # type: Union[int, None]
    pixels = [] # type: List[bool]
    max_width = 0 # type: int
    frame = True # type: bool

    num_frames = 0;
    while True:
        height = rng.randint(size_range[0], size_range[1])
        width = rng.randint(size_range[0], size_range[1])
        
        if width > max_width:
            max_width = width

        if max_frames is not None and num_frames >= max_frames:
            break

        num_frames += 1

        for i in range(height * width // 7):
            randbyte = rng.read(1)[0] % 7
            bits = [randbyte & 1!= 0] * 7
            pixels += bits


    # Create image descriptor block.
    color_table_flag = False
    local_color_table_size = 0
    sort_flag = False
    interlace_flag = False

    idatbl = (
        0b0000 | \
             (color_table_flag << 7) | \
             (local_color_table_size << 4) | \
             (sort_flag << 3) | \
             (interlace_flag << 2)
    ).to_bytes(1, 'big')

    idatalen = len(max(pixels)).to_bytes(1, 'big')
    image_left = max_width.to_bytes(2, 'little')
    image_top = 0.to_bytes(2, 'little')
    image_width = max_width.to_bytes(2, 'little')
    image_height = height.to_bytes(2, 'little')

    image_descriptor = (
        idatbl + idatalen + image_left + image_top + \
        image_width + image_height
    )


    # TODO: handle transparency blocks.
    data = (
        0x00, # background index
        0x00, # pixel aspect ratio

        0xF9, 0x04, # graphics control extension label
        0x0B, # graphic control extension size
        0x04, # transparent color flag
        0x00, 0x00, # delay time
        0xff, # disposal method
        0x00, # user input flag

        0x01, # graphic control label
        0x00, # graphic control size
        0x00,# 0x00, # top left X coord
        0x00,# 0x00, # top left Y coord
        0x00,# 0x00, # width
        0x00,# 0x00, # height
        0x00,# 0x00, # palette update flag

        0x2C, # image separator
        0x00, # image separator size
        0x00, # left position
        0x00, # top position
        0x00, # width
        0x00, # height
        0x00, # local colour table flags
        0x00, # interlace flag
        0x00, # sorted colour table flag
        0x00, # reserved bit
        0x00, # LZW minimum code size
        0x00, # basic graphemes
        0x00, # basic grapheme size

        *(b'\xFF\x00' if p else b'\x00\x00' for p in pixels),

        *[ # Plain Block Introducer Label
            0x21, # label
            0x00, # plain block introducer size

            59, # loop control extension label
            4, # loop control extension size
            0x01, # loop repetitions
            0xff, 0x00, # loop terminator
        ]
    )



    # Create color table.
    lctblsiz = 0
    rgbtbl = pack('<HHH', 0, 0, 0) 

    colortable = lctblsiz.to_bytes(1, 'big') + rgbtbl

    clrlstlen = 3 * (lctblsiz+1).to_bytes(1, 'big')

    colorlst = (
        0xf7, # application extension label
        0x09, # application extension size
        b'NETSCAPE', # application identifier
        3, # application auth code size
        b'\x01', # application auth code data
        
        0xfe, # comment extension label
        0x00, # comment extension size
    )

    out.write(header + len(colortable).to_bytes(2, 'little'))
    out.write(image_descriptor + len(colortable).to_bytes(1, 'little'))
    

    global_hdr = ( # Global Header
        b'GIF89a', # signature
        b'\x00', # version
        32, # logical screen width
        32, # logial screen height
        0xf7, # color resolution
        0x00, # background color index
        0x00, # pixel aspect ratio
        3, # global color table present
        0x00, # background color index
        0x00, # pixel aspect ratio

        # Start of Local Color Table
        0xff, # color resolution
        0x00, # background color index
        0x00, # pixel aspect ratio
        3, # no global color table present
        0x00      # no local color table present
        #[*rgbtbl for _ in range((3 ** lctblsiz))]
    )

    hdrlen = sum([len(i) for i in global_hdr]) + crc(global_hdr)

    out.write(clrlstlen)
    out.write(colortable)
    out.write(clrlstlen)
    out.write(colorlst)
