#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, num_descriptors:int = 100
    ):
    header = b'GIF89a'
    data = b''
    for i in range(num_descriptors):
        datalen = int.from_bytes(
                rng.read(3), byteorder='big')
        descriptor_body = rng.read(datalen)

        # Write the descriptor body which includes the
        # label, block size and everything else.
        # We don't need to keep track of the position of the cursor or anything like that right now.
        out.write(descriptor_body)
        data += descriptor_body

    # Now we're going to write a graphic control extension in the form
    # of 0x21f9 followed immediately by the entirety of the "graphic control" section of the spec
    # See https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    out.write(b'\x21\xf9\x04')
    data += b'\x21\xf9\x04'

    # Setting the delay time to max value
    # See  https://www.w3.org/Graphics/GIF/spec-gif89a.txt#Page%2018
    # There are also optional fields such as transparent colors etc
    # But they aren't really relevant to our case since we aren't doing any processing afterwards anyway
    out.write(struct.pack('<H', 0xffff))
    data += struct.pack('<H', 0xffff)

    # Now comes the actual image data which will contain garbage data
    width = int.from_bytes(rng.read(2), byteorder='little')
    height = int.from_bytes(rng.read(2), byteorder='little')
    depth = 8
    color_table_size = ((2 << depth) ** 3 - 1)//2
    out.write(width.to_bytes(2,'little'))
    out.write(height.to_bytes(2,'little'))
    data += width.to_bytes(2,'little') + height.to_bytes(2,'little')
    out.write((depth).to_bytes(1, 'little'))
    out.write((index_color_flag).to_bytes(1, 'little'))
    out.write((interlace_flag).to_bytes(1, 'little'))
    out.write((sort_flag).to_bytes(1, 'little'))
    out.write((reserved_bits).to_bytes(1, 'little'))
    out.write(color_table_size.to_bytes(1, 'little'))
    data += (depth).to_bytes(1, 'little')
    data += (index_color_flag).to_bytes(1, 'little')
    data += (interlace_flag).to_bytes(1, 'little')
    data += (sort_flag).to_bytes(1, 'little')
    data += (reserved_bits).to_bytes(1, 'little')
    data += color_table_size.to_bytes(1, 'little')

    # Finally append the garbage data
    data += rng.read(width*height//8)
    
    # Now go back and replace the amount of random data
    # we generated previously. We start at -2 instead of -1 because
    # we have recently read one byte and want to overwrite that
    # TODO: It might be slightly better to have this next line check that
    #       we didn't go beyond the bounds of the stream?
    out.seek(-(len(data)-2), 1)
    out.write(len(data).to_bytes(2, 'little'))
    out.write(data[-1:])

    # Finally add the trailer to finish off the output.
    # Normally this would be part of the application logic
    # when writing a gif but since this is intended only
    # for demonstration purposes we do it here directly.
    # See https://github.com/decompme/decompilation-guide/issues/46
    print('Adding the trailer and EOF', end='', flush=True)
    # 2 bytes for signature
    # 1 byte for version
    # 1 byte for flags
    # 2 bytes for background color index
    # 2 bytes for pixel aspect ratio
    out.write(b'\x3B\x00\x00\x00')
    print(' done', flush=True)