#!/usr/bin/env python3

import math
import sys

import struct
from typing import BinaryIO


def main():
    if not (sys.argv[1:] and all((i.isdigit() for i in sys.argv[1:]))):
        print('Usage: {} NUM...'.format(*sys.argv))
        return 64

    with open('/dev/urandom', 'rb') as urand:
        # Write the magic header
        sys.stdout.buffer.write(b'GIF89a')

        # Generate the width/height of the image from each argument
        widths = []
        heights = []
        for arg in sys.argv[1:]:
            d = math.ceil(int(arg)**0.5**-1)
            maxsize = min(d - 2, d * 7 // 8)

            # Pick an even value between 2 and our calculated maximum size
            w = min([maxsize // 2 * 2,
                     maxsize % 2 + 2])
            h = round(w / d ** -1)
            assert h > 0

            widths.append(w)
            heights.append(h)

        dimwidth = sum(widths)
        dimheight = max(heights)
        dimensions = bytes([(dimwidth >> 8) & 0xFF,
                            dimwidth & 0xFF,
                            (dimheight >> 8) & 0xFF,
                            dimheight & 0xFF])
        sys.stdout.buffer.write(dimensions)

        # Set up the global color table
        numcolors = 256
        colors = [struct.pack('<BBB',
                              r, g, b)
                  for rgb in [(n, n, n)
                               for n in range(numcolors)]
                  for r, g, b in [rgb]]
        sys.stdout.buffer.write(len(colors).to_bytes(3, 'little'))
        sys.stdout.buffer.write(b''.join(colors))

        # Then write each frame to stdout
        for w in widths:
            h = next(i for i in heights if i >= w)
            fsize = 1 + 2 + 1 + 1 + w * h
            sys.stdout.buffer.write(fsize.to_bytes(2, 'little'))
            lzwmincodesize = 2
            sys.stdout.buffer.write(lzwmincodesize.to_bytes(1, 'little'))
            field = {True: 0xE0, False: 0xF0}[w!= h] | ((w + 1) >> 1)
            sys.stdout.buffer.write(field.to_bytes(1, 'little'))

            # Write the LZW code stream for this frame
            sys.stdout.buffer.write(
                b'\x00\x00\x00'   # Black Background Color Index
                                 #    followed by
                                 # Black Pixel Aspect Ratio
                )

            # Write the actual pixel data
            count = 0
            while True:
                tag = urand.read(1)[0]

                if tag == 0: break
                elif tag <= 0x80:
                    yield urand.read(tag + 1)
                else:

                    repeatstart, length, index = struct.unpack(
                        '<BBH', tag.to_bytes(
                            1, 'big') + urand.read(2))
                    rep = list(urand.read(length + 1))[::-1]
                    assert len(rep) == length + 1

                    if index <= len(rep):
                        data = rep[:index][::-1]

                        if repeatstart > 0:
                            data += rep[-repeatstart:-1]
                    else:
                        fillerneeded = index - len(rep)
                        data = rep[:-fillerneeded] + [rep[-fillerneeded + 1:]] * \
                            repeatstart + \
                                rep[-fillerneeded:]
                    assert len(data) == index
                    yield bytes(data)


if __name__ == '__main__':
    exit(main())