#!/usr/bin/env python3

import os
import struct
import sys

from argparse import ArgumentParser


# Generates a random animated GIF file with no runtime dependencies.
class RandomGifGenerator:

  def __init__(self, args):
    self._width = args.width
    self._height = args.height
    self._fps = args.fps
    self._stride = 0
    self._color_depth = None
    self._palette = []
    self._global_color_table = {}
    self._frame_buffer = []

    # Initialize palette entries where possible
    if args.monochrome or len(args.palette)==0:
      for i in range(2**args.bpp):
        r = int(((i/(2**args.bpp))*255)+0.5)
        g = int(((i/(2**args.bpp))*255)+0.5)
        b = int(((i/(2**args.bpp))*255)+0.5)
        entry = [r, g, b]
        self._palette.append(entry)

      self._color_depth = 8
      self._global_color_table['entries'] = [b'\x00'*3]*6
    else:
      if len(args.palette)%3!= 0:
        print("Palette must be divisible by three")
        exit(-1)

      self._color_depth = 7
      numpalentries = len(args.palette)//3

      for i in range(numpalentries):
        rgb = list(args.palette[i*3:(i+1)*3])
        entry = [rgb[2], rgb[1], rgb[0]]
        self._palette.append(entry)

      self._global_color_table['entries'] = [bytes(args.palette)]*4


  @property
  def width(self):
    return self._width

  @property
  def height(self):
    return self._height

  def write_header(self, fd):
    """Write the GIF header."""
    # Header string and version.
    fd.write(b"GIF89a")

    # Logical Screen Descriptor.
    fd.write(struct.pack('<H', self.width))      # Width in pixels
    fd.write(struct.pack('<H', self.height))     # Height in pixels
    fd.write(struct.pack('<B', int(0xf7)))       # Global color depth / packed field
    fd.write(struct.pack('<B', int(0)))         # Background color index
    fd.write(struct.pack('<B', int(7)))         # Pixel aspect ratio.
    fd.write(struct.pack('<B', int(0)))         # Single-image GIF.
    fd.write(struct.pack('<B', int(0)))         # Default black background.


  def write_color_table(self, fd):
    """Write the global color table."""

    fd.write(struct.pack('<B', int(0)))           # Extension code for application extension.
    fd.write(struct.pack('<B', int(0)))           # Subblock length indicator.
    fd.write(struct.pack('<B', int(0)))           # Application identifier (NETSCAPE).
    fd.write(struct.pack('<B', int(3)))           # Subblock length indicator.
    fd.write(struct.pack('<B', int(1)))           # Data sub-block label.
    fd.write(struct.pack('<B', int(self._fps)))   # Frames per second.
    fd.write(struct.pack('<B', int(0)))           # No loop extension present.

    # Write the color table itself as an application extension.
    fd.write(struct.pack('<B', int(0xff)))        # Extension introducer.
    fd.write(struct.pack('<B', int(0xff)))        # Graphic control extension.
    fd.write(struct.pack('<B', int(0)))           # Block size.
    fd.write(struct.pack('<B', int(0)))           # Flag indicating transparency.
    fd.write(struct.pack('<H', int(self._color_depth)))    # Transparent color index.
    fd.write(struct.pack('<B', int(0)))           # Block terminator.
    fd.write(struct.pack('<B', int(0)))           # Image separator.

    # Write the global color table prefix.
    fd.write(struct.pack('<B', int(0xfe)))          # Extension introducer.
    fd.write(struct.pack('<B', int(0x00)))          # Label for local color table.
    fd.write(struct.pack('<B', int(int(math.log(float(len(self._palette))/3), 2)))))            # Log base 2 of number of colors.

    # Write the global color table.
    fd.write(b"".join([struct.pack("<BBB", x[0],x[1],x[2]) for x in self._palette]))



  def write_graphic_control_extension(self, frame_no, fd):

    fd.write(struct.pack('<B', int(0xf9)))                      # Extension introducer.
    fd.write(struct.pack('<B', int(4)))                        # Block size.
    fd.write(struct.pack('<B', int(0)))                        # Flags.
    fd.write(struct.pack('<H', int(0)))                        # Delay in centiseconds.
    fd.write(struct.pack('<B', int((frame_no + 1) % 256)))    # Transparency index.
    fd.write(struct.pack('<B', int(0)))                        # Terminator.


  def write_image_separator(self, fd):
    """Write the image separator."""
    fd.write(struct.pack('<B', int(0x2c)))


  def write_local_color_table(self, frame_no, fd):
    """Write the local color table"""

    numcolors = math.pow(2, self._color_depth) - 1
    fd.write(struct.pack('<B', int(0)))                       # Extension introducer.
    fd.write(struct.pack('<B', int(0xe0)))                     # Label for local color table.
    fd.write(struct.pack('<B', int(int(math.log(float(numcolors)/3), 2)))))             # Log base 2 of number of colors.
    fd.write(struct.pack('<B', int(0)))                       # Local color table follows.


    for j in range(numcolors):
      r = self._frame_buffer[frame_no][j]['red']
      g = self._frame_buffer[frame_no][j]['green']
      b = self._frame_buffer[frame_no][j]['blue']
      fd.write(struct.pack('<BBB', r, g, b))                 # Color value.



parser = ArgumentParser()
parser.add_argument('-o', '--output', type=str, default='output.gif')
parser.add_argument('--width', required=True, type=int)
parser.add_argument('--height', required=True, type=int)
parser.add_argument('--fps', required=False, type=int, default=10)
parser.add_argument('--bpp', type=int, choices=[1, 2, 4, 8], default=1)
parser.add_argument('--monochrome', action="store_true")
parser.add_argument('--palette', nargs='+', default=[], help='RGB triplets; e.g., "FF FF 00 00 00 00".')

