#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # First we need to figure out how many frames we have to insert
    # Right now I'm assuming that even if you get lucky enough to find
    # exactly one suitable spot to corrupt another descriptor, inserting
    # 256 more won't change the result significantly. If not...well..let's fix it later I guess.
    # Also note that I made some assumptions about the format of the
    # underlying format: specifically that each image descriptor has 10
    # extra bytes, which makes things easier. If that changes then this
    # needs to also change.
    
    # First, let's check how many descriptors there already are
    # TODO: Make this work with files that aren't animated
    numdescs = sum([chunk[0] == b'\x2C'[0] for chunk in iter(lambda: rng.read(1), b'')])
    print(f'Found {numdescs} existing descriptors')

    # There should be no more than 256 more available
    # Note that if we ran out of randomness or something else went wrong,
    # this may fail badly
    assert((numdescs+256) < 256)

    # Next up, let's find the first place that would fit in the file
    # based on a very simple heuristic rule. It's not perfect because
    # it will leave a few bytes empty, but it should do okay for most cases
    # With the magic numbers here (`labellen`) we're checking whether
    # we can safely put a new descriptor right after the current one without
    # running over any labels. Basically we're looking for a payload like:
    # | size of block   | label            | frame data        |

    # And `labellen` is computed from that.
    def next_frame():
        
        # First let's pick a random number of bytes that might be valid
        # TODO: Figure out what the limits are on the number of valid
        #       bytes per descriptor. Assuming that it's below ~10k seems
        #       reasonable though?
        numentries = rng.randint(1, 0x1F-10)
            
        # Now let's create that many entries!
        entries = []
        while len(entries) < numentries:
            labellen = rng.randint(1, 0xF)
            size = rng.randint(1, 0xFF)
            
            # After figuring out the size lets add one entry
            # Note that each entry includes its own length byte
            entries += [
                    labellen,
                    size,
                    rng.read(size)]

        return entries

    # Keep looking until we find one that fits in the file
    print('Finding acceptable location for the new descriptor...')
    attempts = 0
    while True:
        attempts += 1
        print('.', end='', flush=True)
        candidates = {}
        while len(candidates) < 256:

            # Create a couple frames
            f1 = next_frame()
            f2 = next_frame()

            """
            These checks are meant to ensure that the candidate
            insertion point is safe to insert a new descriptor
            somewhere within.

            Basically we want to look for a payload like:
                | descriptor     | descriptor          |

                Where "descriptor" contains either the entirety
                of a single frame descriptor, or part of it
                (i.e., the first half or second half depending on
                how it's oriented)

                It is only safe to insert a new descriptor inside
                a descriptor if both of those conditions hold:

                1. Neither of them contain any label bytes at
                   their start
                2. They both have at least one full entry
            """
            if ((any(entry[0] >= 10 for entry in f1) or any(entry[0] >= 10 for entry in f2)) and \
                (all(entry[0]+entry[1]+10 <= 0x10 for entry in f1) and all(entry[0]+entry[1]+10 <= 0x10 for entry in f2))) :
                
                # Check if they overlap. If they do, keep looking.
                if not any([f1[-1][0]+f1[-1][1]+10!= f2[0][0],
                            f2[-1][0]+f2[-1][1]+10!= f1[0][0]]):
                        continue
            
                # Ok! Looks good. Insert them into the dictionary
                key = f'{rng.read(1)}-{rng.read(1)}'.encode() 
                candidates[key] = {'f1': f1, 'f2': f2}


        print('')
        for k,_ in candidates.items():
            print(k)

        print(f'Generated {attempts} different candidates...', end='')

        # Pick a random candidate
        i = rng.randrange(0, len(candidates))
        cand = list(candidates.keys())[i]

        # Let's check whether the candidate is ok and construct a
        # synthetic input stream
        descloc = struct.unpack("<H", cand)[0] & 0x1FFF

        # Seek backwards to the beginning of the frame data section
        # For sanity sake we'll seek back to the first frame's label;
        # if that's present we can skip it otherwise we have to go all the way back
        out.seek(descloc)
        label = out.read(1)
        if label!= '\x2C'[0]:
            out.seek(descloc-1)
        assert(out.read(1) == '\x2C'[0])

        # The 10th byte of the label indicates the length of the whole block
        labellen = struct.unpack("B", out.read(1))[0] 
        assert(labellen % 0x10 == 0)
        out.seek(descloc-(labellen%0x10)+1)

        # Write the new descriptor sequence
        seq = candidates[cand]['f1'] + candidates[cand]['f2']
        for entry in seq: 
            l,s,*d = entry 
            out.write(l.to_bytes(length=1, byteorder='little'))
            out.write(s.to_bytes(length=1, byteorder='little'))
            out.write(d)

        # This flag is true when the last descriptor was inserted
        if len(seq) == 0x10: break
        elif len(seq) == 0x11: 
            # We were close -- try again
            pass
        else: raise RuntimeError("Encountered unexpected number of entries")

    # That should've worked, except that it could've created
    # an invalid file entirely. So let's check that everything looks correct
    last_byte = out.tell()-1
    assert(last_byte >= 0)

    # Find the EOF marker
    eof = None
    out.seek(-1, 2)
    while out.read(1) == b'\x3B':
        eof = out.tell() - 1

    assert(eof > 0)

    # TODO: Are there other possible problems?
    #       I would imagine that sometimes we'd run out of space
    #       entirely, which would probably mean creating an invalid
    #       file...but does that exist?

    print(f'The generated gif is {last_byte} bytes long.')

