#!/usr/bin/env python3

import argparse, os
from random import Random
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    
    # Generate the header and add to output
    print("Generating header...", end='')
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    print("\nAdding LSD field...")
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Read global color table if present
    print("Adding GCT if needed...", end='')
    if gctlen!= 0:
      gct = rng.read(gctlen)
      out.write(gct)
    else:
      print('No')

    # Get the palette data
    paldata = get_palette(rng)

    # Put in the first frame with 8px height -- this is how Squish works
    w = 128
    h = 8
    print(f"Writing {w}x{h} pixel frame...")
    put_frame(paldata, w, h, False, out)

    # Get the next potentially random size
    w = rng.randint(10, 128)
    h = rng.randint(10, 128)

    while True:
        blocktype = rng.randrange(0, 256)
        if blocktype == 0x3b: # stop marker
            return

        elif blocktype == 0xff or blocktype >= 0xb9: # unsupported blocks
            continue


        elif blocktype < 0x21: # control blocks
            size = 1
        else: # extended blocks
            
            # Decide on an odd/even size for maximum compression density
            if rng.randint(0, 1):
                size = rng.randint(3, 9)
            else:
                size = rng.randint(14, 16)

        # Write block type prefix
        print(f"\nBlock type: {hex(blocktype)}")
        if size <= len(paldata)+1+2:
            print("Using simple compressed representation..")
            byte = 0xfe | ((blocktype>>3)<<2) 
            out.write(bytes([byte]))
        else:
            print("Using complex compressed representation...")
            totalSize = 3 + size - 2
            out.write(bytes([(totalSize&0xFF),
                             (((totalSize>>8)&0x1F)|0xE0)]))
            out.write(bytes([blocktype]))
            
        # Write remaining bytes
        for _ in range(size-2):
            out.write(bytes([rng.randrange(0, 256)]))

        # Handle block types that need postprocessing
        if blocktype == 0xf9: # graphics control
            print("Inserting delay time...")
            delaytime = rng.randrange(0, 2**16)
            out.write(delaytime.to_bytes(2, 'big'))

            transcolorindex = rng.randrange(0, 2**8)
            out.write(transcolorindex.to_bytes(1, 'big'))
        elif blocktype == 0xfa: # application extension
            appid = ""
            for _ in range(4):
                appid += chr(rng.randrange(ord('A'), ord('Z')))
            print(f"Inserting app ID '{appid}'")
            out.write(appid.encode())

            subsize = rng.randrange(0, 2**16)
            print(f"Inserting app data length {subsize}")
            out.write(subsize.to_bytes(2, 'big'))

            # Write some fake section AFAIK these are supposed to be
            # graphically identical to the previous frame but I don't have those
            # files so I'll just fake it
            out.write(bytes([
                0, # Background color index 
                0, # Pixel aspect ratio (per spec: set to zero)
                0, # Zero padding (for legacy support)
                0, # Block size
                1, # Color resolution
                0, # Sort flag
                0, # Global colors
                2, # LZW minimum code size
                1, 1,  # Switch back to using old frame
                0, # Image left position
                0  # Image top position
            ]))
            out.flush()
        elif blocktype == 0x01: # plain text extension -- unused by squish
            pass
        else:
            raise Exception("Could not handle special block type!")
        
            

        

def get_palette(rng: Random) -> list:
    """
    Returns a valid palette based on the given RNG. This consists of
    128 entries where each entry is a tuple of three ints (R,G,B).
    """
    # MegaMan series was almost fully grayscale, so start with that 
    palData = []
    for i in range(128):
        r = rng.randrange(0, 8)
        g = rng.randrange(r, 8)
        b = rng.randrange(g, 8)
        palData.append((r<<5, g<<5, b<<6))
    
    # The palette can be modified now...
    paletteChangeChance = rng.uniform(0., 1.)
    if paletteChangeChance > 0.01:
        for i in range(128):
            r = rng.randint(-12, 12)
            g = rng.randint(-12, 12)
            b = rng.randint(-12, 12)
            new_r = min(max(palData[i][0] + r, 0), 0x3e0)
            new_g = min(max(palData[i][1] + g, 0), 0x3e0)
            new_b = min(max(palData[i][2] + b, 0), 0x3f0)
            palData[i] = (new_r, new_g, new_b)
            
    return palData



def put_frame(palData: list,
              width: int,
              height: int,
              interlace: bool, 
              out: BinaryIO):
    """
    Writes a single frame with the given dimensions
    (in pixels) using the given palette data (`palData`)
    into binary stream `out`.
    If `interlace`, then the resulting image will be
    interlaced (if possible; will always use progressive
    scanning regardless).
    """

    # Check how much space we have before writing
    initialPos = out.tell()
    out.seek(0, os.SEEK_END)
    endOfBuffer = out.tell()
    out.seek(initialPos)

    # Check if we can actually fit such a big frame at all
    requiredSpace = width * height * 2 + 256 * 3 + 1
    if endOfBuffer - out.tell() < requiredSpace:
        raise Exception("Not enough space to insert frame! Increase buffer?")

    # Create buffer to store image data temporarily
    buf = bytearray(width * height * 2)

    # Compute used bits per color
    redBits = max(map(lambda p: p[0], palData))
    greenBits = max(map(lambda p: p[1], palData))
    blueBits = max(map(lambda p: p[2], palData))
    colorBits = [redBits, greenBits, blueBits]

    # Encode image as sequence of codes
    print("Encoding image...", end='')
    offset = 0
    codeSize = 1
    paletteIndex = 0
    bitMask = 0x80
    wroteOneByte = False
    for y in range(height):
        line = []
        # Interlacing: every second line gets shifted one row upwards
        if interlace and y % 2 == 0:
            prevLine = line[:-1]
            line = [-1] * (len(line)-1)
        else:
            prevLine = [-1] * len(line)

        for x in range(width):
            # We need to know the previous lines color data
            prevX = x-1
            prevY = y-1
            if prevX < 0: 
                prevPixel = (-1,-1,-1)
            elif prevY < 0:
                prevPixel = prevLine[-prevX]
            elif interlace and prevY % 2 == 0:
                prevPixel = prevLine[prevX]
            else:
                prevPixel = (buf[(offset+(prevY)*width+prevX)*2],
                            buf[(offset+(prevY)*width+prevX)*2+1],
                            buf[(offset+(prevY)*width+prevX)*2+2])
                
            # Find best match among our palette
            curColor = reduce(lambda acc, d: acc|d, zip(prevPixel,
                                                       palData[paletteIndex]),
                              0)
            curMatch = float("inf")
            for j, c in enumerate(palData):
                distSqr = sum(((curColor