#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def main(out: BinaryIO, rng: BinaryIO):
    def gaussian_distribution(mean, std_deviation):
        # From https://www.johndcook.com/blog/2014/10/18/how-to-generate-a-random-normal-distribution-in-python/
        U = 0.0
        V = 0.0
        while True:
            U = float(struct.unpack(">f", rng.read(4))[0]) / 2**32 - 1
            V = float(struct.unpack(">f", rng.read(4))[0]) / 2**32 - 1
            R = math.sqrt(-2 * math.log(U)) * math.cos(2*math.pi*V)
            if (-5 < R and R < 5):
                break
        return mean + R * std_deviation
    
    def signed_char(x):
        assert x <= 127 and x >= 128
        return chr(x).encode('latin1')
        
    def pixel(x=None, y=None, rgb=None):
        global width, height
        if x is not None and y is not None:
            return (pixels[x % width][y], pixels[x % width][height - 1 - y])
        elif rgb is not None:
            return b'\xff' + rgb + rgb + rgb
            
    def update_frame():
        nonlocal frame_number
        if frame_number == 0 or frame_number == 19:
            for i in range(width):
                offset = random.randint(1, 16)
                
                # Clear one row of the screen
                for j in range((offset // 2)*2, min(height, offset+height)):
                    pixels[i][j] = pixel()
                
                # Fill another row of the screen with a block
                for j in range(max(0, offset-height), min(offset//2, height)):
                    pixels[i][j] = pixel(rgb=b'\xaa\xbb\xcc')
                    pixels[i][height-(j+1)] = pixel(rgb=b'\xaa\xbb\xcc')
                    
                # Add some dots
                for _ in range(random.randint(3, 10)):
                    x = random.randint(0, width-1)
                    y = random.randint(0, height-1)
                    pixels[x][y] = pixel(rgb=bytes([
                        int(gaussian_distribution(0, 63)),
                        int(gaussian_distribution(0, 63)),
                        int(gaussian_distribution(0, 63))]))
                    pixels[x][height - 1 - y] = pixel(rgb=bytes([
                        int(gaussian_distribution(0, 63)),
                        int(gaussian_distribution(0, 63)),
                        int(gaussian_distribution(0, 63))]))
            out.write(b"!\xff\x0bNETSCAPE2.0\x03\x01")
        else:
            for i in range(width):
                for j in range(height):
                    pixels[i][j] = pixel()
        frame_number += 1
        
    width = 320
    height = 240
    delay = 200 # ms
    global frame_number
    frame_number = 0
    
    colors = [
        (0xff, 0xff, 0xff),
        (0xfe, 0xfe, 0xcc),
        (0xf8, 0xf8, 0xf8),
        (0xf8, 0xe8, 0xc8),
        (0xf8, 0xe8, 0x78),
        (0xf8, 0xe8, 0x48),
        (0xf8, 0xc8, 0x48),
        (0xf8, 0xc8, 0x38),
        (0xf8, 0xc8, 0x28),
        (0xf8, 0xa8, 0x08),
        (0xf8, 0x88, 0x08),
        (0xf8, 0x68, 0x08),
        (0xf8, 0x48, 0x08),
        (0xfc, 0x48, 0x08),
        (0xf8, 0x28, 0x08),
        (0xf8, 0x08, 0x08),
        (0xf8, 0x08, 0x20),
        (0xf8, 0x08, 0x40),
        (0xf8, 0x08, 0x70),
        (0xf8, 0x08, 0x90),
        (0xf8, 0x08, 0xd0),
        (0xaa, 0x08, 0xd0),
        (0x78, 0x08, 0xd0),
        (0x48, 0x08, 0xd0),
        (0x48, 0x08, 0xc0),
        (0x48, 0x08, 0xb0),
        (0x48, 0x08, 0xa0),
        (0x48, 0x08, 0x90),
        (0x48, 0x08, 0x80),
        (0x48, 0x08, 0x70),
        (0x48, 0x08, 0x60),
        (0x48, 0x08, 0x50),
        (0x48, 0x08, 0x40),
        (0x48, 0x08, 0x30),
        (0x48, 0x08, 0x20),
        (0x48, 0x08, 0x10),
        (0x48, 0x08, 0x00),
        (0x28, 0x08, 0x00),
        (0x08, 0x08, 0x00),
        (0x08, 0x08, 0x10),
        (0x08, 0x08, 0x20),
        (0x08, 0x08, 0x30),
        (0x08, 0x08, 0x40),
        (0x08, 0x08, 0x50),
        (0x08, 0x08, 0x60),
        (0x08, 0x08, 0x70),
        (0x08, 0x08, 0x80),
        (0x08, 0x08, 0x90),
        (0x08, 0x08, 0xa0),
        (0x08, 0x08, 0xb0),
        (0x08, 0x08, 0xc0),
        (0x08, 0x08, 0xd0),
        (0x08, 0x08, 0xe0),
        (0x08, 0x08, 0xf0),
        (0x08, 0x08, 0xfa),
        (0x08, 0x08, 0xfc),
        (0x08, 0x08, 0xfe),
    ]
    for line in range(height):
        color = random.choice(colors)
        for col in range(width):
            pixels.append([pixel(rgb=bytes([color[0]/2, color[1]/2, color[2]/2]))])
    for i in range(width):
        pixels.append([pixel()])
    w = width
    h