#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
        
        with open('/dev/urandom', mode='rb') as rng,\
             open(self.output,    mode='wb') as out:

            return generate_random_gif(
                    rng, out
                )



# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# are read from different sources to ensure that they're valid.
def generate_random_gif(rng:BinaryIO, out:BinaryIO) -> None:
    
    # Read in the first 6 bytes of data from the RNG and check them for
    # sanity (they should be the ASCII codes for "GIF87a" or "GIF89a").
    data = bytes(struct.unpack('<BBBBBB', rng.read(6)))

    magic = data[:6]
    version = b''.join([data[0], data[3]])

    if (magic != b'GIF89a') and (magic != b'GIF87a'):
        raise ValueError("Invalid GIF file.")

    if (version == b'GIF87a'):
        print("\nWARNING: Outputting a GIF87a file.\n\tThis may not work properly!")
    
    # Write the header to the output file.
    out.write(data)

    # Read in the Logical Screen Descriptor fields.
    width  = struct.unpack("<H", rng.read(2))[0] + 1
    height = struct.unpack("<H", rng.read(2))[0] + 1
    color_table_size = int((ord(rng.read(1)[0]) & 0x7) * 3) + 1
    background_color_index = ord(rng.read(1)[0])
    aspect_ratio = ord(rng.read(1)[0])

    # Write the Logical Screen Descriptor to the output file.
    lsd = struct.pack(
        "<HBHHBB",
        width,  # screen width
        b'\00',  # packed field - ignored
        height,  # screen height
        b'\00',  # global color table flag - false
        b'\01',  # color resolution - indexed
        b'\04',  # sort flag - true
        color_table_size  # size of global color table
    )

    out.write(lsd)

    # Read in the Background Color Index.
    bgci = bytes(struct.unpack("<BB", rng.read(2)))

    # Write the Background Color Index to the output file.
    out.write(bgci)

    # Loop over each pixel on the grid.
    for _ in range(width*height):

        # Read in one byte of data from the RNG, which will contain the
        # information about this pixel.
        pixel_info = struct.unpack("<B", rng.read(1))[0]

        # Break the byte up into its individual components.
        local_color_table   = bool(pixel_info & (1 << 7))
        interlace            = bool(pixel_info & (1 << 6))
        sorted               = bool(pixel_info & (1 << 5))
        reserved             = bool(pixel_info & (1 << 4))
        disposal             = pixel_info >> 3
        user_input           = bool(pixel_info & (1 << 2))
        transparent          = bool(pixel_info & (1 << 1))
        local_color_table_size = int((pixel_info & 0x7)*3)+1

        if local_color_table:
            colormap_bytes_per_entry = 3
            colormap_length = local_color_table_size
        else:
            colormap_bytes_per_entry = color_table_size
            colormap_length = 0

        # If we need to write 256 bytes, we can do it in bulk by splicing
        # all 256 bytes together and writing the result at once; otherwise
        # we need to loop over the entire thing and write out each pixel.
        if colormap_length > 0:

            colormap_raw = rng.read(colormap_bytes_per_entry * colormap_length)

            colormap_entries = []

            for i in range(colormap_length):
                colormap_entries.append(list(bytearray(colormap_raw[(i*3):(i*3+3)])))

            # Sort our list of colors according to their index bytes.
            if sorted:

                colormap_sorted = [[] for i in range(colormap_length)]

                for entry in colormap_entries:
                    colormap_sorted[entry[3]] = entry

                colormap_entries = colormap_sorted

            # Convert our list into an array of bytes.
            output = b''
            for entry in colormap_entries:
                output += bytes(entry)

            # Now write out our color map.
            out.write(output)

        if background_color_index >= colormap_length:
            background_color_index %= colormap_length

        # Generate a random delay time.
        delay_time = bytes(struct.pack("<H", struct.unpack("<H", rng.read(2))[0]))

        # Read in any additional fields that may have been provided.
        while True:

            # First, read in the next line of data from the RNG.
            block_type = ord(rng.read(1)[0])

            # If the type is a plain text extension...
            if block_type == 1:

                text_grid_left = ord(rng.read(1)[0])
                text_grid_top = ord(rng.read(1)[0])
                text_grid_width = ord(rng.read(1)[0])
                text_grid_height = ord(rng.read(1)[0])
                cell_width = ord(rng.read(1)[0])
                cell_height = ord(rng.read(1)[0])
                text_foreground_color_index = ord(rng.read(1)[0])
                text_background_color_index = ord(rng.read(1)[0])
                padding = b'\00'

                # Output the block header.
                out.write(bytes([block_type,text_grid_left,text_grid_top,text_grid_width,text_grid_height,cell_width,cell_height,text_foreground_color_index, text_background_color_index]))

                # Read in the text string and output it to the file.
                for _ in range(text_grid_height*text_grid_width):
                    out.write(padding + rng.read(1))

            # If the type is a comment extension...
            elif block_type == 2:
                
                length = ord(rng.read(1)[0])
                comment_string = bytes(rng.read(length))

                out.write(bytes([block_type,length]) + comment_string)

            # If the type is an application extension...
            elif block_type == 3:
                
                app_name_length = ord(rng.read(1)[0])
                app_name = bytes(rng.read(app_name_length))
                app_data = bytes(rng.read(0xffff))

                out.write(bytes([block_type,app_name_length]) + app_name + app_data)

            # If the type is a graphics control extension...
            elif block_type==4:

                gc_flags = ord(rng.read(1)[0])
                gc_delay_time = bytes(struct.pack("<H", struct.unpack("<H", rng.read(2))[0]))

                gc_transparent_color_index = ord(rng.read(1)[0])

                out.write(bytes([block_type,gc_flags,gc_delay_time[0],gc_delay_time[1],gc_transparent_color_index]))

            # If the type is a frame image...
            elif block_type==5:

                # Store the fact that this is the end of the GIF by adding the
                # terminator to our output file.
                out.write(b'\x3b')

                # Exit the loop that reads extensions.
                break

            # If the type is an unknown block type, raise an exception.
            else:
                raise Exception("Unknown block type: "+str(block_type))


# Main function. Reads commandline arguments and generates a GIF file.
def main() -> None:
