#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + datalen.to_bytes(3, 'big'))
    out.write(data)

# Generates a random JPEG file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_jpeg(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int
    ):
    SOF_TABLE = {
        256: ((1, 1, 2), (7, 2, 1)),
        128: ((2, 1, 3), (9, 2, 1))
    }
    DCTSIZE = 8
    HUFFMAN_DESCRIPTOR = 0xC4
    START_OF_IMAGE = 0xD8
    END_OF_IMAGE = 0xD9
    PREDICTOR = [0] * 64
    def _build_huffman_table(
            bits: bytes, values: bytes,
            maxcode: int, maxbits: int
        ) -> bytes:
        """
        Builds an huffman table for encoding/decoding
        JPEG image data according to section A.2 and
        A.3 of the JFIF spec. The table is given in
        bits of length up to 16 (the maximum possible).
        Special rules are applied to ensure that smaller codes
        have higher precedence than larger ones.
        @param bits: Code lengths, from smallest to largest
        @param values: Corresponding code words
        @return: Bytes corresponding to the huffman table
                as defined by the JPEG standard
        """
        LENS_INDEX = {}
        VALUES_INDEX = {}

        # Assign code lengths, codes, and code value indices
        lencount = [0] * 16
        valptr = [0] * 16
        for i in range(len(bits)):
            nbits = bits[i]
            hval = values[i]
            if nbits != 0:
                lencount[nbits] += 1
                LENS_INDEX[(nbits, hval)] = i
                VALUES_INDEX[hval] = i
        # Count the number of codes of each length
        j = 0
        for i in range(1, 17):
            lens[j] = lencount[i]
            j += 1
        # Assign numerical values to all codes
        k = 0
        m = 0
        for bitnum in range(1, 17):
            incr = 1 << (bitnum - 1)
            while lens[k]:
                vals[m] = k
                m += 1
                k += incr
                lens[k] -= 1
            k = 0
        # Fill in codes and tables
        for bitnum in range(1, 17):
            incr = 1 << bitnum
            fill = (incr // 2)
            k = vals[fill]
            while fill < 64:
                if bitnum <= 10:
                    # Small codes
                    codesize = bitnum
                    top = (maxcode[bitnum-1] + 1)
                    bot = maxcode[bitnum-1] + 1 + lens[k]
                    delta = (top - bot) // (lens[k] + 1)
                    i = bot
                    for j in range(1, lens[k]+1):
                        htab[i] = k
                        itab[k] = i
                        i += delta
                elif bitnum == 11:
                    # Large codes
                    codesize = 10
                    top = (maxcode[codesize-1] + 1)
                    bot = maxcode[codesize-1] + 1 + lens[k]
                    delta = (top - bot) // (lens[k] + 1)
                    i = bot
                    for j in range(1, lens[k]+1):
                        htab[i] = k
                        itab[k] = i
                        i += delta
                    # Special case for the first few codes
                    codesize = bitnum
                    top = (maxcode[codesize-1] + 1)
                    bot = maxcode[codesize-1] + 1 + lens[k]
                    delta = (top - bot) / 16
                    for j in range(1, 16+1):
                        htab[bot+j] = k
                        itab[k] = bot+j
                elif bitnum >= 12:
                    # Even more large codes
                    codesize = 10
                    top = (maxcode[codesize-1] + 1)
                    bot = maxcode[codesize-1] + 1 + lens[k]
                    delta = (top - bot) // (lens[k] + 1)
                    i = bot
                    for j in range(1, lens[k]+1):
                        htab[i] = k
                        itab[k] = i
                        i += delta
                k += 1
                fill = fill + incr
        return bytes(htab) + bytes(itab)

    MIN_HUFFBITS = 1
    MAX_HUFFBITS = 12
    MAX_HUFFCOUNTS = 256
    padding = [b'\xff\xff\xff']
    # Table to store huffman codes by bit length
    maxcode = [0] * (MAX_HUFFBITS + 1)
    huffman_values = []
    huffman_bits = []
    huffman_tables = []
    # Write header
    out.write(struct.pack('!cc', b'\xff\xd8', b'\xff\xe0')) 
    out.write(struct.pack('!bbI', 16, 0, width*height*3))
    out.write(struct.pack('!HH', width, height))
    out.write(struct.pack('!c', b'\x02'))
    out.write(struct.pack('!cc', b'\xff\xdb', b'\x00\x43'))
    out.write(struct.pack('!c', b'\x03'))
    out.write(struct.pack('!bbb', 2, 1, 3))
    out.write(struct.pack('!bb', 1, 1))
    out.write(struct.pack('!c', b'\xff\xc0'))
    out.write(struct.pack('!cc', b'\x00', b'\x11'))
    out.write(struct.pack('!b', 3))
    out.write(struct.pack('!b', 1))
    out.write(struct.pack('!c', b'\x00'))
    out.write(struct.pack('!c', b'\x02'))
    out.write(struct.pack('!b', 240))
    out.write(struct.pack('!c', b'\x01'))
    out.write(struct.pack('!b', 179))
    out.write(struct.pack('!c', b'\x01'))
    out.write(struct.pack('!b', 160))
    out.write(struct.pack('!c', b'\x00'))
    out.write(struct.pack('!c', b'\x02'))
    out.write(struct.pack('!b', 144))
    out.write(struct.pack('!c', b'\x01'))
    out.write(struct.pack('!b', 128))
    out.write(struct.pack('!c', b'\x00'))
    out.write(struct.pack('!c', b'\x03'))
    out.write(struct.pack('!bb', 1, 1))
    out.write(struct.pack('!cc', b'\xff\xc4', b'\x00\x1f'))
    out.write(struct.pack('!cc', b'\x00', b'\x00'))
    out.write(struct.pack('!c', b'\x01'))
    out.write(struct.pack('!c', b'\x05'))
    out.write(struct.pack('!c', b'\x12'))
    out.write(struct.pack('!c', b'\x00'))
    out.write(struct.pack('!c', b'\x00'))
    out.write(struct.pack('!c', b'\x00'))
    out.write(struct.pack('!c',        if blocksize - len(byte) > 0:
            # Pad short blocks with zeroes
            out.write(byte + padding[0])
        else:
            out.write(byte[:blocksize])

