#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jul 12 16:47:05 2020

@author: mzimmer
"""

import numpy as np
import pandas as pd
from scipy import optimize
import matplotlib.pyplot as plt
from tqdm import tqdm
tqdm.pandas()
import seaborn as sns
import multiprocessing as mp
import itertools

#%%
def DeltaPsi(psi, alpha, beta):
    return (alpha * psi + beta)
    
#%%
params_daily = {'a': [19, 3], 'b': 3}

def loglikelihood_daily(theta, alpha, beta, psi, phi):
    # want to calculate the sum of squares error between each data point and its predicted value
    # consider using broadcasting for speedup
    theta_hat = DeltaPsi(psi, theta[0], theta[1])
    psi_hat = DeltaPsi(phi, theta_hat, 0)
    return np.sum((np.log(theta_hat) - np.log(psi_hat))**2)
    
def loglikelihood_daily_wrapper(args):
    """Wrapper function for use with multiprocessing."""
    return loglikelihood_daily(*args)
    
def get_ML_predictor_daily(data, phi=None):
    """This function fits a regression model to the input data. The model is a linear combination
       of the predictors. The coefficients are found by maximizing the likelihood function.  
       
       Input
           data  : pandas DataFrame object containing the daily returns
                   of the portfolio.  Required columns: ['delta', 'phi']
            phi : float; initial estimate for coefficient phi
               Defaults to 1/(1+mean(data['delta']))     
         Output
             theta_hat : a tuple containing the best estimates of alpha and beta that fit the data
             ML_error  : the mean squared error of the best fitting line
             
    """
    if phi == None:
        phi = 1 / (1 + data['delta'].mean())
    args = [(params_daily, alpha, beta, data['delta'], phi)
                for alpha, beta in it.product(range(-20, 20), repeat=2)]
    pool = mp.Pool(mp.cpu_count())
    results = list(tqdm(pool.imap_unordered(loglikelihood_daily_wrapper, args), total=len(args)))
    pool.close()
    pool.join()
    theta_hat = params_daily[np.argmax(results)]
    return theta_hat

#%%
def DeltaAlphaBeta(alpha, beta, gamma, delta):
    return gamma * (alpha * delta + beta)

    
params_cumulative = {'g': [1, 0.003], 'a': 0, 'b': 0}

def loglikelihood_cumulative(theta, gamma, delta, alpha, beta):
    # want to calculate the sum of squares error between each data point and its predicted value
    # consider using broadcasting for speedup
    return np.sum((DeltaAlphaBeta(gamma, theta[0], theta[1], delta) - DeltaAlphaBeta(gamma, alpha, beta, delta))**2)
    
def loglikelihood_cumulative_wrapper(args):
    """Wrapper function for use with multiprocessing."""
    return loglikelihood_cumulative(*args)
    
def get_ML_predictor_cumulative(data, phi=None):
    """This function fits a regression model to the input data. The model is a linear combination
       of the predictors. The coefficients are found by maximizing the likelihood function.  
       
       Input
           data  : pandas DataFrame object containing the cumulative returns
                   of the portfolio.  Required columns: ['alpha', 'beta', 'delta']
            phi : float; initial estimate for coefficient phi
               Defaults to 1/(1+mean(data['delta']))     
         Output
             theta_hat : a tuple containing the best estimates of alpha and beta that fit the data
             ML_error  : the mean squared error of the best fitting line
             
    """
    if phi == None:
        phi = 1 / (1 + data['delta'].mean())
    args = [(params_cumulative, gamma, delta, data['alpha'], data['beta'])
                for gamma, delta in it.product(np.linspace(1e-2, 1e1, 100), repeat=2)]
    pool = mp.Pool(mp.cpu_count())
    results = list(tqdm(pool.imap_unordered(loglikelihood_cumulative_wrapper, args), total=len(args)))
    pool.close()
    pool.join()
    theta_hat = params_cumulative[np.argmin(results)]
    return theta_hat
    
#%%
def plot_regression_residuals(x, y, x_hat, color='k'):
    """Plots the residuals (y - y_hat) over time, where y_hat is the predicted values and y is the actual values"""
    fig = plt.figure(figsize=(10,4))
    axes = fig.add_subplot(111)
    axes.scatter(x, y, c=color)
    axes.plot([x.min(), x.max()], [y.mean(), y.mean()], 'r--')
    axes.set_title('Residual Plot')
    axes.set_xlabel('Time')
    axes.set_ylabel('Returns')
    plt.show()
    
#%%