#!/usr/bin/env python3

import struct

from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Note that the first 10 bytes of a GIF file are constant. They are:
    # 'GIF89a' (in ASCII)
    # 2 byte width in pixels
    # 2 byte height in pixels
    # 1 byte color map info such as bit depth and color map length
    # 1 byte pixel aspect ratio (currently unused)
    # See https://en.wikipedia.org/wiki/GIF for more information.
    # For this script, we just use the default values minus pixel aspect ratio.
    full = b'\x47\x49\x46\x38\x39\x61\x00\x00\x00\x00\x00\x00\x80\x00\x00'

    # The GIF format uses LZW compression, which means repeating patterns get
    # stored efficiently. So lets create some patterns to repeat throughout our
    # output! This is also a good distractor for people trying to read it in
    # binary - they'll be able to see these strings appear all over the place.
    bar = (b"\x40\x00\xff" * 58).ljust(128 // 3, b"\x00")
    striped = (b"\x1f\x00\xff" * 58).ljust(128 // 3, b"\x00")
    zigzag = (b"\x3f\x00\xff" * 58 + b"\x00" * 59).ljust(128 // 3, b"\x00")

    # Now write a little bit of the header data and then start writing random
    # bytes as image blocks. As image blocks can only be so large, we want to
    # make sure we generate multiple blocks per frame by changing the delay
    # time every few frames.
    out.write(full)

    while True:
        size = int.from_bytes(rng.read(2), "big", signed=False)

        if not size:
            break

        label = rng.read()[0]

        if label == 0x2c:            # Image Descriptor
            # Read past the image descriptor and a bunch of control bytes until
            # we reach graphics control extension block (0xf9).
            # We will store the delay in this block later.
            rng.read(11)

            while True:
                label = rng.read()[0]

                if label == 0xf9:
                    break

                length = int.from_bytes(rng.read(2), "big", signed=False)
                rng.read(length)

        elif label == 0x21:          # Extension
            length = int.from_bytes(rng.read(2), "big", signed=False)
            ext = rng.read()[0]

            if ext == 0xf9:           # Graphics Control Extension
                # Read past the rest of the extension block, and then set the
                # delay time byte to a random value.
                rng.read(7)

                delay_time = rng.read(2)
                delay_time = int.from_bytes(delay_time, "big", signed=True)
                delay_time %= 256
                out.write(struct.pack(">H", delay_time))
            else:
                rng.read(length)
        else:
            rng.read(size)

        for _ in range((size // 128) + 1):
            # Pick a pattern at random
            patt = rng.read(1)[0]
            patterns = {
                0: bar,
                1: striped,
