#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):
        self._output = DEFAULT_OUTPUT if not output else output
        self._outfile = open(self._output, "wb")

    def _write_header(self):
        """Write the GIF87a header to the output file.

        https://en.wikipedia.org/wiki/GIF#GIF_bitmap_structure
        """
        header_data = b"GIF87a" + \
            bytes([0x01]) + \
            bytes([0x00]) + \
            bytes('RANDOM'.encode()) + \
            bytes('RANDOM'.encode()) + \
            b"\xff\x00;" + \
            b"\x2c\x00\x00\x00\x00\x01\x00\x01\xff\x21\xf9\x04\x05\x00\x00\xff"
        self._outfile.write(header_data)

    def generate(self, width: int, height: int, length: int):
        """Generate a random GIF of given width and height with specified length."""
        print("[+] Starting ...", flush=True)
        # Write the GIF header
        self._write_header()

        for i in range(length):
            print("\r[+] Generating frame %d/%d..." % (i+1, length), end="", flush=True)
            junk_data = random_junk(width*height)
            self._write_frame(width, height, junk_data)

        print("\n[+] Done! Output at %s" % self._output)

    def _write_frame(self, width: int, height: int, data: bytearray):
        """Write a single GIF frame into the outfile.
        
        https://en.wikipedia.org/wiki/GIF#Image_block
        """
        # Write an image block
        self._outfile.write(bytes((0x2C)))
        # Image position & size
        self._outfile.write(struct.pack("<HH", 0, 0))
        self._outfile.write(struct.pack("<HH", width, height))
        # Packed fields
        self._outfile.write(bytes((0x00)))
        # Local color table indicator, color resolution, sort flag and size
        self._outfile.write(bytes((0x00)))
        # LZW min code size
        self._outfile.write(bytes((0x08)))
        # Write image data
        for chunk in chunks(data, 1024):
            self._outfile.write(chunk)
        # Add zero-byte terminator, I don't know what it does but it's here lol
        self._outfile.write(bytes((0x00)))

    def close(self):
        self._outfile.close()


def random_junk(size:int, seed:int = None) -> bytearray:
    """Return a random array of n bytes, using a given seed or the system time as default."""
    rand = Random()
    if not seed:
        seed = rand.randrange(0, INT_MAX)
    rand.seed(seed)
    return rand.bytes(size)


def chunks(l: list