#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' # GIF magic header
    rng.readinto(header[6:]) # Read image dimensions from /dev/urandom
    out.write(header)

    screen_width = struct.unpack('<H', header[4:6])[0] + 1
    screen_height = (struct.unpack('<H', header[6:8])[0] // 256) % 255 + 1

    # Write color table
    for i in range(7):
        pixel_rgb = struct.unpack('<I', rng.read(4))[0] * 255 // 0xFFFFFF
        if i == 0:
            pixel_rgb |= 0b11 << 24 # Transparent background color
        elif i > 1 and pixel_rgb & 0xFF0000 != 0:
            pixel_rgb = 0xF80000 # Red
        else:
            pixel_rgb &= ~pixel_rgb >> 16 # Copy green to blue
        out.write(bytes([i, pixel_rgb]))

    gct_size = 0x2B
    rndtab_size = 0
    gif_term = bytearray([0x3B])

    while True:
        if screen_width < 1 or screen_height < 1: break

        gct_flag = 0
        lsd = [gct_flag, rndtab_size, screen_width - 1, screen_height - 1, 0, 4, 0]
        out.write(bytearray(lsd))

        # Randomize local color table (but not transparent!)
        rng.readinto(out.buffer[:gct_size + 4])

        image_data = bytearray()
        image_left = screen_width * screen_height
        while image_left > 0:
            code_length = min(image_left, rng.read(1)[0] + 1)
            image_left -= code_length
            image_data += b'\x00' + struct.pack('B', code_length - 1)
            
            fill_color = 1
            if rng.read(1)[0]:
                new_code = rng.read(1)[0] + 2
                if rng.read(1)[0]:
                    new_code += rng.read(1)[0] + 2**2
                fill_color = new_code
            image_data += bytes((fill_color,))
        
        # Write image descriptor...
        idd = bytearray([0x2C, 0, 0, screen_width - 1, screen_height - 1, gct_flag << 2])