#!/usr/bin/env python3
"""Testing script of CANopenNode for testing Flexray."""
import time
from can import Notifier
from can.interface import Bus
from canopen import objectdictionary, network, symbolic
import logging

import sys
sys.path.insert(0, '../..')  # allow importing from parent directory
import flexray_driver as fd # pylint: disable-msg=E0401

# Setup logging to file - see previous section for more details
LOG_FILENAME = "flexray_test.log"
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                    datefmt='%m-%d %H:%M',
                    filename=LOG_FILENAME,
                    filemode='w')
# define a Handler which writes INFO messages or higher to the sys.stderr
console = logging.StreamHandler()
console.setLevel(logging.INFO)
# set a format which is simpler for console use
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
# tell the handler to use this format
console.setFormatter(formatter)
# add the handler to the root logger
logging.getLogger('').addHandler(console)

def main():
    """Main function"""
    # Define network parameters
    nodeid = 50
    baudrate = 100000
    heartbeat = 0xA0
    bitrate = symbolic.BitRate.Kbps100
    channel = 'can0' if len(sys.argv) < 2 else sys.argv[1]
    bus = Bus(channel, bustype='socketcan')

    # Create object dictionary and load it with nodes information
    od = objectdictionary.ObjectDictionary()
    od.load('/etc/canopen/nodes.yaml')

    # Add vendor specific objects and update sdo size
    device = fd.FlexRayDevice(nodeid)
    device.init(od)

    # Create CANopenNetwork and pass the object dictionary
    network = network.Network(fd.FlexRay(), od)
    # Add SDO client to be able to configure node via SDOs
    network.add_client_sdo(device.sdo_client)
    network.add_server_sdo(device.sdo_server)
    # Add Emergency consumer to handle emergencystop, reset, etc.
    network.add_consumer(_EmergencyConsumer())

    # Configure heartbeat producer and start it
    hbprod = network.heartbeat.add_producer(device.hb_prod)
    hbprod.start()

    notifier = Notifier(bus, [network], timeout=0)
    print("notifier started")

    while True:
        try:
            # Without timeout parameter notifier.poll() blocks until something receives or sends
            notifier.poll(timeout=1)
        except KeyboardInterrupt:
            break
    notifier.stop()
