#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    # The first block has to be at least 2 bytes long
    if not datalen: datalen += 1
    # And it can't end with 0
    while (datalen % 256 == 0):
        datalen += 1
    data = rng.read(datalen)
    trailer = b'\x3B'

    out.write(header + struct.pack('<H', len(data)) + data + trailer)

if __name__ == '__main__':
    import os, sys
    from random import Random
    argv = sys.argv[1:]
    if len(argv) > 1 and argv[0] in ('-h', '--help'):
        print(__doc__)
        exit()
    if len(argv) >= 1:
        filename = argv[0]
    else:
        filename = '-'
    with open('/dev/urandom', 'rb') as rng, \
         open(filename, 'wb') as out:
        generate_random_gif(rng, out)

"""
This is a silly script which generates large .gif files by reading
large amounts of random data off /dev/urandom. The file will be
upwards of 4GBs unless you get an error when writing.
"""
\end{code}

Comment: I suspect that the OS buffering layer may be causing intermittent write errors instead of an OOM.  You could try setting `PYTHONUNBUFFERED=1`.

Answer: [Python 3.7](https://docs.python.org/3/whatsnew/3.7.html) added support for buffered writes of arbitrary size. This means you should no longer need to split up writes yourself. In Python 3.7+ you would use [`file.buffer.write()`](https://docs.python.org/3/library/io.html?highlight=buffer#io.BufferedWriter.write) to write arbitrarily large chunks without having to split them up yourself. Note that this method returns the number of bytes written, so if it's not equal to what you gave it, then you know you've hit some limit somewhere along the way.

Note: It sounds like you have some additional constraints on file size. The main limitation here is the available RAM on the system, as well as the maximum possible length of a single call to [`os.write()`](https://docs.python.org/3/library/os.html?highlight=write#os.write). So start trying arbitrarily large values until your program fails to complete successfully.

\begin{blockquote}

Write up to n bytes from string s, where n is either the length of s or
  the value specified by bufsize. A short write may occur if the I/O
  device or operating system does not allow the full operation to be
  done atomically.
\end{blockquote}

For example, say you want to write 8MB of data to a file. Then you can do something like this (using pseudo code):

\begin{code}
chunk_size = 8 * 1024 ** 2
with open("temp.txt", "w") as fp:
    chunk = ... # read a chunk of data
    while True:
        f_written = fp.buffer.write(chunk)
        if f_written == chunk_size: break

        chunk = read_remaining(fp, chunk_size - f_written)
        f_written += fp.buffer.write(chunk)
        if f_written < chunk_size: break
\end{code}

The above code snippet assumes that `read_remaining()` reads and returns the remaining data necessary to fill a chunk of data completely.

Comment: Could you perhaps show how to adapt my script to take advantage of this? I don't see how one would split up the calls automatically; maybe I just don't understand how this works. I know how to manually check whether the call succeeded in the trivial case, but I don't understand how the API is supposed to work in the general case (and whether the returned value is redundant or useful?)

Comment: @JanusTroelsen I updated my answer slightly. Please let me know if this makes sense.

Comment: What you're describing looks great, but I still don't get it. Is the idea that `fp.buffer.write` takes care of splitting up into blocks for me?

Answer: The solution was to catch the exception and try again with a smaller blocksize, stopping once I got down to a nonzero return value. Here's the new version:

\begin{code}
#!/usr/bin/env python3

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    # The first block has to be at least 2 bytes long
    if not datalen: datalen += 1
    # And it can't end with 0
    while (datalen % 256 == 0):
        datalen += 1
    data = rng.read(datalen)
    trailer = b'\x3B'

    out.write(header + struct.pack('<H', len(data)) + data + trailer)

if __name__ == '__main__':
    import os, sys
    from random import Random
    argv = sys.argv[1:]
    if len(argv) > 1 and argv[0] in ('-h', '--help'):
        print(__doc__)
        exit()
    if len(argv) >= 1:
        filename = argv[0]
    else:
        filename = '-'
    with open('/dev/urandom', 'rb') as rng, \
         open(filename, 'wb') as out:
        blocksize = 8*2**20
        while True:
            try:
                byte = rng.read(blocksize)
            out.write(byte[:blocksize])

