#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 21 16:54:06 2017

@author: paul.vandelay
"""
import os
from collections import Counter, defaultdict
from functools import partial
from multiprocessing import Pool, cpu_count

import numpy as np
import pandas as pd
from tqdm import tqdm

def get_classes(labels):
    """Get a list of the unique classes in `labels`."""
    return sorted({l[1] for l in labels})
    
def get_mismatches(same_label=True):
    """Get a list of mismatches based off ground truth and predicted label."""
    mismatches = []
    if same_label:
        # The ground truth and predicted labels match
        mismatches += [(1, 1), (0, 0)]
        
    mismatches += [(0, 1), (1, 0)]
    
    return mismatches

def confusion_matrix(labels, priors, mismatches=[], normalize=False):
    mismatches = get_mismatches() if len(mismatches) == 0 else mismatches
    
    df = pd.DataFrame(columns=['predicted', 'true'], index=get_classes(labels))
    df.index.name = 'class'
    
    matches = pd.Series([0]*len(priors)*len(df.index), index=pd.MultiIndex.from_product((
        ['match'] * len(df.index),
        df.index)))
    mismatches = pd.Series([0]*len(priors)*len(df.index), index=pd.MultiIndex.from_product((
        ['misclassified'] * len(df.index),
        df.index)))
    
    if not normalize:
        counts = pd.concat([matches, mismatches])
    else:
        counts = pd.concat([matches, mismatches]) / len(labels)
    
    return counts

    
def confusion_metrics(confusion_vals):
    """Get confusion metrics from a confusion matrix."""
    return {
        'precision': precision(*confusion_vals),
        'recall': recall(*confusion_vals),
        'f1': f1(*confusion_vals)
    }
    
def precision(tp, fn):
    return tp/(tp + fn)

def recall(tp, fp):
    return tp/(tp + fp)

def f1(prec, rec):
    return 2*(prec * rec)/(prec + rec)

def apply_to_cols(func, cols):
    """Apply function to columns of DataFrame."""
    series_list = [func(col) for col in cols]
    
    new_cols = [pd.Series(data=series.values, 
                          name=' '.join(new_col_name).lower(),
                          dtype=np.float64) 
                for new_col_name, series in zip(zip(cols[0].name, func.__name__),
                                            series_list)]
    
    return pd.concat(new_cols, axis=1)

def calc_confusion_metrics(counts):
    """Calculate all confusion metrics from a confusion matrix."""
    cols = counts.columns.levels[0][:2]
    rows = counts.index.tolist()
    true_cols = ['true_{}'.format(c) for c in rows]
    pred_cols = ['predicted_{}'.format(c) for c in rows]
    
    true_df = pd.DataFrame(
        data={true_cols[i]:counts[rows[i]] for i in range(len(rows))},
        index=rows)

    pred_df = pd.DataFrame(
        data={pred_cols[i]:counts[cols[i]] for i in range(len(rows))},
        index=rows)
    
    confusion_sets = pd.concat([true_df, pred_df], axis=1)
    
    conf_metrics = apply_to_cols(confusion_metrics, confusion_sets.columns)
    
    return pd.concat([conf_metrics, confusion_sets], axis=1)

def build_confusion_table(labels, priors, mismatches=[], normalize=False):
    """Build a confusion table and calculate all metrics."""
    counts = confusion_matrix(labels, priors, mismatches, normalize)
    return calc_confusion_metrics(countes)

def conditional_entropy_binned(values, bins=None):
    """Given values, calculate their entropy over a set of binned values."""
    hist, bin_edges = np.histogram(values, bins=bins)
    p = hist/sum(hist)
    H = sum(-p*np.log2(p)+(1-p)*np.log2(1-p))
    return H
    
def conditional_entropy(X_train, y_train, bins=None):
    """Take the conditional entropy of X wrt y."""
    entropies = []
    train_pairs = dict(Counter(zip(y_train, X_train)))
    for class_, pairs in train_pairs.items():
        x_vals, _ = zip(*pairs)
        entropies += [conditional_entropy_binned(x_vals, bins)]
    return np.mean(entropies)

def mutual_information(X_train, y_train, n_estimators=1):
    """Compute the mutual information between two variables."""
    class EntropyPairMutualInformationEstimate:
        def __init__(self, estimator, name="mutual_info_score", **params):
            self.estimator = estimator
            
            self.name = name
            self._parameters = params
            
            self._fit_attributes = ["mutual_info_attr_" + str(i+1) for i in range(n_estimators)]
            
            self.estimator_type = "mixture"
        
            self.MI = None
        
        def fit(self, X, y=None):
            mi_steps = []
            for i in range(n_estimators):
                est = clone(self.estimator)
                est_ = est.fit(X, y)
                mi_step_ = getattr(est_, self.name)(**self._parameters)
                mi_steps += [mi_step_]
            
            self.MI = reduce(lambda x, y: x+y, mi_steps)
            return self
        
        def score(self, X=None, y=None):
            return self.MI
        
        @property
        def feature_importances_(self):
            return self.MI
        
        @property
        def coef_(self):
            return self.MI
        
        @property
        def intercept_(self):
            return self.MI
        
    return EntropyPairMutualInformationEstimate().fit(X_train, y_train).score()

def outlier_removal(df, percentile=99.9):
    """Remove an outlier column from a dataframe."""
    outliers = df[df > df.quantile(percentile)]
    return df[~df.isin(outliers)]

def plot_features(X_train, y_train, figsize=(6, 8), ncols=3):
    """Plot the features in X_train."""
    plt.figure(figsize=figsize)
    
    sns.set(style="ticks")
    sns.pairplot(X_train, hue=y_train, diag_kind="kde", markers=["o", "D"])
    
    row = int(np.floor(len(X_train.columns)/ncols))
    col = len(X_train.columns) % ncols
    
    ax1 = plt.subplot(row, col, 1)
    ax1.set_title("Individual Features\n", fontweight='bold')
    
    for i, c in enumerate(X_train):
        ax = plt.subplot(row, col, i+2, sharey=ax1)
        plt.scatter(y_train, X_train[c], alpha=.8, edgecolors='none')
        plt.xlim(.69, .71)
        plt.xticks([],