#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
    
        with open(f'./{self.output}', 'wb') as fh:
            generate_random_gif(open('/dev/urandom'), fh)


# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# correspond to the information found here:
# https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(inp:BinaryIO, out:BinaryIO):
    # First 6 bytes are signature + version
    assert inp.read(6) == b'GIF87a'

    # Read width and height (little endian)
    width, height = struct.unpack('<HH', inp.read(4))

    # Color table size is the next byte
    color_table_size = inp.read(1)[0]

    # Background color index, pixel aspect ratio,
    # and global flags are the next three bytes
    background_color_index, _, _ = struct.unpack('BBB', inp.read(3))

    # Write out header
    out.write(b'GIF89a')                             # Signature
    out.write(struct.pack('<HH', width, height))     # Width, height
    out.write(bytes([color_table_size]))              # Color table size
    out.write(bytes([0]))                            # Background color
    out.write(bytes([0]))                            # Pixel aspect ratio
    out.write(bytes([0xFF | 0x40]))                  # Global flags

    # Generate color table
    for _ in range(2**color_table_size):
       r, g, b = [bytearray([int(i * 256 / 255)]) for i in struct.unpack('fff', inp.read(12))]
       out.write(b'\x00\x00\x00' + r + g + b)

    # Generate image data blocks
    while True:
        # Block type, which can be one of:
        # - 0x2C -> Image Data
        # - 0x3B -> Trailer (ends the GIF)
        block_type = inp.read(1)[0]
        
        if block_type == 0x3B:
           break
        elif block_type != 0x2C:
          print("Bad block type!")
          exit(-1)
  
        # Read left/top position and width/height (little endian)
        ltpwh = inp.read(4)

        # Read local flags, which can be:
        # bit 0: interlace flag
        # bit 1: sort flag
        # bit 2: reserved
        # bit 3: local color table flag
        # bit 4: 2=4 bits, 3=8 bits per palette entry
        x = int(ltpwh[0])
        y = int(ltpwh[1])
        w = int(ltpwh[2])
        h = int(ltpwh[3])
        flags = inp.read(1)[0]

        if flags & 0x80:
           print("Interlaced images not supported.")
           exit(-1)

        # Local color table flag
        if flags & 0x8:
           # Read local color table
           num_colors = 2 ** ((flags & 7) + 1)
           for _ in range(num_colors):
               r, g, b = [struct.unpack('bb', inp.read(2))[0]]
               out.write(bytes([r, g, b]))

        # Read image descriptor (LZW code size)
        code_size = inp.read(1)[0]

        # Initialize decoder
        clear_code = (1 << code_size)
        eoi = clear_code + 1
        max_code = clear_code + 2
        old_code = None
        prefix = None
        suffix = None
        table = {}

        for c in range(clear_code, eoi):
            table[c] = [c]

        # LZW decode
        while True:
            code = (max_code if len(table) >= max_code else max_code + 1)

            # If code is greater than the maximum possible, this is EOI
            if code > 4096:
                break

            # Read another byte from the stream and append it to our current code
            buffer = inp.read(1)
            code = struct.unpack('>H', buffer + struct.pack('B', code))[0]
            entry = []

            if code <= max_code:
                entry = table.get(code).copy()

            # Add first code to the list
            if not old_code:
                old_code = code
                continue

            # Check for last code
            if code == eoi:
                break

            # Append suffix to the entry
            if code <= max_code:

                # Get previous code entry and append it to the new entry
                prev_entry = table.get(old_code)
                entry += prev_entry[-1:]
                if prefix:
                    entry += prev_entry[:-1]

                # Set the prefix code to the new code
                prefix = old_code

            # Set the suffix code
            suffix = prev_entry[-1]

            # Add the entry
            table[max_code] = entry
            max_code += 1

            # If we have more than 4096 codes, set the code size and reset
            if max_code > 4096:
                out.write(bytes([code_size + 1]))
                max_code = clear_code + 2
                code_size += 1

            old_code = code

        # Write out data subblocks
        out.write(b'\x00' + struct.pack('BB', x, y))
        out.write(b'\x00' + struct.pack('BB', w, h))
        out.write(b'\x08')
        for i in range((w * h)):
            out.write(bytes([background_color_index]))
        out.write(b'\x00')
        out.write(b'\x03')

