#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting to fuzz
    rng.read(7)

    palette_size = int(rng.read(1).hex(), base=16)

    color_table = []
    for i in range(palette_size * 3):
        color_table.append(int(rng.read(1).hex(), base=16))

    # block terminator
    rng.read(1)

    local_color_table_flag = int(rng.read(1).hex(), base=16)
    interlace_flag = int(rng.read(1).hex(), base=16)
    sort = int(rng.read(1).hex(), base=16)
    global_color_table_size = (2 ** ((sort >> 0x05) + 1)) - 1

    if local_color_table_flag != 0:
        raise Exception(f'local color table flag is {local_color_table_flag}')

    if interlace_flag != 0:
        raise Exception(f'interlace flag is {interlace_flag}')

    if sort != 0:
        raise Exception(f'sort flag is {sort}')

    background_color_index = int(rng.read(1).hex(), base=16)

    pixel_aspect_ratio = int(rng.read(1).hex(), base=16)

    # skip fields that aren't particularly interesting to fuzz
    rng.read(6)

    out.write(header)

    # write out the size field
    out.write(struct.pack('<HH', image_width, image_height))
    out.write(b'\xF8\x00')
    out.write(bytes([palette_size]))

    # generate the color table
    for i in range(palette_size):
        out.write(struct.pack('BBB', *[c // 2 for c in color_table[i*3:(i+1)*3]]))

    # block terminator
    out.write(b'\x00')

    # set up the graphic control extension
    gce_flag = 0xF9
    gce_block_size = 0x04
    gce_delay = int(rng.read(1).hex(), base=16)
    gce_transparent_index = rng.read(1)[0] & 0xFF

    out.write(struct.pack('BBBB', gce_flag, gce_block_size, gce_delay, gce_transparent_index))

    # write out the frame
    lsd_flag = 0x21
    lsd_block_size = 0x0A
    lsd_flags = 0x00 | (local_color_table_flag << 7) | (interlace_flag << 6) | (sort << 0x05)

    out.write(struct.pack('BBBB