#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# provided RNG. The size of the generated image is determined by
# the first four bytes read from the RNG (which are discarded).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    width, height = struct.unpack("<HH", rng.read(4))
    
    # Gif header, see https://en.wikipedia.org/wiki/GIF#Fixed_header
    header = b'\x47\x49\x46\x38\x00\x00\x00\x00'       \
            + struct.pack('<HB', width, height)        \
            + b'\xF7\x00\x00'                          \
            + b'\x2c'                                  \
            + b'\x00\x00\x00\x00'                      \
            + b'\x01'                                  \
            + b'\x00'                                  \
            + b'\x2c'                                  \
            + b'\x00\x00\x00\x00'
            
    # Random palette, 256 x 3 bytes per entry in little-endian order
    palette = [rng.read(3) for _ in range(256)]
    palette_length = len(palette) * 3 - 1
    
    # Image descriptor block, 11 byte header plus one color table
    idb = b'\x2c'                                      \
         + b'\x00\x00\x00\x00'                          \
         + b'\x00'                                     \
         + b'\x01'                                     \
         + struct.pack('B', palette_length)           \
         + b''.join(palette)

    # Delay times array, delay time set to zero
    dta = b'\x00'*256

    # Color table length minus one
    ctlm1 = struct.pack('B', palette_length - 1)

    # One-byte transparent colour index
    transparency = rng.read(1)

    # Block terminator
    block_terminator = b'\x00'

    # Read a single pixel and convert it to an LSB-first byte string
    def get_next_pixel_lsb():
        return rng.read(1)[0] & 1

    # Read 8 bits at once and interpret as a byte array
    def get_next_byte():
        return int.from_bytes([get_next_pixel_lsb()], 'little')

    # Read 3 bytes at once for the last three pixels of a row
    def get_last_three_pixels():
        return sum((get_next_byte() << i * 8) for i in range(3)).to_bytes(3, 'big')

    # Write the supplied number of bytes as a row of pixels
    def write_row(size):
        # If there are no pixels left over after reading the full row, use that
        if size % 3 == 0:
            for _ in range(size // 3):
                out.write(struct.pack('BBB', *[get_next_byte() for _ in range(3)]))
        else:
            # Otherwise, write an incomplete row with padding bits
            first = size // 3
            pad = size - first * 3
            row = sum((get_next_byte() << i * 8) for i in range(first))
            out.write(row.to_bytes(first, 'big'))
            out.write(get_last_three_pixels()[:pad])

    # Encode the row, stripe and interlace information into a packed byte
    def encode_line_info(row, stripe, interlaced):
        return ((interlaced << 6) | (stripe << 3) | (row & 0b111))

    # Repeat the following process until we reach the end of our grid
    while width > 0 or height > 0:
        # Decide whether to use interlacing
        interlaced = bool(height % 2)
        
        # Write the IDB if this is a new frame
        if height % 8 == 1 or not interlaced:
            out.write(idb)
            out.write(ctlm1 + transparency + block_terminator)
        
        # Calculate how many rows there should be on this frame
        num_rows = min(min(width, height), 8)

        # If the image is too tall, swap the width and height
        if height > width:
            num_rows = min(num_rows, ((width+3)*(height//4)) // (width+1))
            num_rows += num_rows % 4
            for stripe in range(num_rows // 4):
                for row in range(4):
                    # Write the line info byte
                    out.write(encode_line_info(row, stripe, False).to_bytes(1, 'big'))
                    write_row(width)
                    write_row(width)
            height -= num_rows
            width, height = height, width
        else:
            for stripe in range(num_rows // 4):
                for row in range(4):
                    # Write the line info byte
                    out.write(encode_line_info(row, stripe, interlaced).to_bytes(1, 'big'))
                    write_row(width)
                    write_row(width)
            height -= num_rows


# Generates a random gif file with size 's' (in bytes)
def gen_gif(rng: BinaryIO, s: int):
    out.write(header + data)
