#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
        
        with open('/dev/urandom', mode='rb') as rng,\
             open(self.output,    mode='wb') as out:

            return generate_random_gif(
                    rng, out
                )



# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# are read from different sources to ensure that they're valid.
def generate_random_gif(rng: BinaryIO, out: BinaryIO) -> None:
    
    # Header
    header = b''
    header += b'GIF89a'            # File signature and version
    # Color Table size
    ctsz = rng.read(1)[0] * 3
    for _ in range(ctsz):
        header += bytes([rng.read(1)[0]])
    out.write(header)             # write the header we've just generated

    # Image info
    blocksize = rng.read(1)[0]   # Block size
    color_table_size = rng.read(1)[0] % 7;
    code_size = rng.read(1)[0]   # Bits per pixel
    width = rng.read(2)          # Width
    height = rng.read(2)         # Height
    flags = rng.read(1)          # Flags (we only really care about interlacing)

    # Background Color Index
    bgci = 0
    if rng.read(1) != bytes([0xF0]):
        bgci = rng.read(1)       # Background Color Index
    else:
        print('Background color set to black')

    # Palette Size
    ps = rng.read(1)[0] & 0x7
    if ps == 0x7:
        ps = rng.read(1)[0]      # Can be extended up to 256 colors

    # Now let's construct a proper image descriptor
    idesc = bytearray()
    idesc += bytearray([0x2C])   # Start of image descriptor
    idesc += width              # Image width
    idesc += height             # Image height
    idesc += bytearray([flags[0]] | 0x80) # Interlaced image bit on
    idesc += bytearray([ps])    # Only support palette sizes <= 8 bits
    idesc += bgci               # Background color index
    idesc += bytearray([0xFF])  # Aspect ratio - square pixels
    out.write(bytes(idesc))      # Write image descriptor

    # Get the LZW dictionary
    dictionary = rng.read(blocksize)
    # Get the initial compressed stream
    compstream = rng.read(blocksize)
    # We haven't decoded anything yet
    decompbuf = bytearray()
    # Add the clear code
    dictionary.append(0x00)
    # Increment the code length
    codelen = 9
    # Decode the compressed stream into the buffer
    while True:
        # Start at the beginning of the buffer
        i = 0
        while i < len(compstream):
            # If we have an invalid code
            if compstream[i] > len(dictionary):
                break
            # Otherwise, expand the code
            try:
                decompbuf.append(dictionary.pop(compstream[i]))
            except IndexError:
                break
            # Advance our position
            i += 1
            # If we've gotten a complete code
            if compstream[i-1] == 0xFF:
                # Clear the dictionary
                dictionary = bytearray()
                # Reset the code length to its minimum value
                codelen = 9
                # Stop scanning the compressed stream
                break
            # Enlarge the dictionary
            elif len(dictionary) >= 1 << codelen:
                # Double code size
                codelen += 1
        # Check whether we've finished decoding
        if i == len(compstream):
            break
        # Otherwise, keep going
        # Remove the largest possible prefixes from the dictionary
        d = dict(enumerate(dictionary))
        keys = list(d.keys())
        values = list(d.values())
        newentries = []
        for e in range(1<<codelen):
            key = rng.read(1)[0]           # Get a random byte
            if key == 0xFF:               # Break if FF is read
                break
            k = keys.index(key)           # Look for it in the dictionary
            newentry = values[k] + values[key]  # Add them together
            newentries.insert(0, newentry) # Insert entry into front of list
        dictionary.extend(newentries)     # Extend the dictionary
    
    # Convert the decompression buffer to a string
    buf = str(decompbuf).encode('ascii')
    # Compute the checksum of the decompressed buffer
    cs = 256**4 - sum([ord(c)-1 for c in buf])%256**4
    # Write the decompressed image data
    out.write(buf);
    # Write the checksum
    out.write(struct.pack('<I',cs));

    # Finally, write the EOF trailer
    out.write(bytearray([0x3B]))     # EOF trailer



# A more direct implementation using files
def main(infile: str, outfile: str) -> None:
    
    with open(infile, mode='rb') as rng,\
         open(outfile,mode='wb')  as out:

        return generate_random_gif(
                rng, out
            )



# Test program
# Usage: python3 randimgfy.py [-o outfile] inputfile
"""
if __name__ == "__main__":
    if len(sys.argv) == 1 or '-h' in sys.argv \
            or '--help' in sys.argv or '/?' in sys.argv:
        usage()

    # Parse arguments
    args = [arg for arg in sys.argv if arg not in ['-o','--out']]
    infile = args[0]
    outfile = DEFAULT_OUTPUT if len(args) <= 1 else args[1]

    # Execute function
    main(infile, outfile)
"""



# Additionally, the following is a modification of the above that
# adds random padding, so that the resulting file might be compatible
# with modern browsers such as Chrome.



# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# are read from different sources to ensure that they're valid.
# This one has extra code at the end that will hopefully make
# the browser think it's a valid image.
def generate_random_gif_with_thumbnail(rng: BinaryIO, out: BinaryIO) -> None:
    
    # Original image
    generate_random_gif(rng, out)
    
    # Thumbnail image
    # Construct a simple application extension block
    # with the thumbnail dimensions
    header = bytearray()
    header += bytearray([0x21, 0xff])  # Application extension block identifier
    header += bytearray([0x0b])        # length of rest of block
    header += bytearray("NETSCAPE")    # Netscape animator identifying string
    header += bytearray([3])           # sub-block type
    header += bytearray([1])           # loop count
    header += bytearray([0])           # terminating zero
    bs = rng.read(2)                   # get block size
    header += bytes([bs])        # add block size back again
    data = rng.read(bs)          # get actual image data
    out.write(header + data)     # write it all together
    
