#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'
PACKED_TYPE = 0x874F


def convert(infile: str, outfile=None):
    if not isinstance(infile, (BinaryIO, bytes)):
        infile = open(infile, mode='rb')

    def _read(nbytes=-1) -> bytes or None:
        return infile.read() if nbytes == -1 else infile.read(nbytes)

    def _write(*data) -> int:
        outstr = b''
        for v in data:
            outstr += v if isinstance(v, bytes) else bytes([v])

        return outfile.write(outstr)

    try:
        img_size = int.from_bytes(_read(4), byteorder='little', signed=False)
        packed_type = int.from_bytes(_read(2), byteorder='little', signed=False)
        desc_len = int.from_bytes(_read(), byteorder='little', signed=False)

        # print(f"img_size={img_size} desc_len={desc_len}")

        _read(desc_len + 1)
        if packed_type!= PACKED_TYPE:
            raise ValueError('Unknown GIF type!')

        _write(b'GIF89a')
        while True:
            chunk_hdr = _read(6)

            if not len(chunk_hdr):
                break

            chunk_id = f'{int(hex(ord(chunk_hdr[0]))[2:]):0>2X}' \
                       f'{int(hex(ord(chunk_hdr[1]))[2:]):0>2X}'
            chunk_len = int.from_bytes(chunk_hdr[2:], byteorder='little',
                                       signed=False)
            # print(f"{hex(ord(chunk_hdr[0])), hex(ord(chunk_hdr[1]))}, "
            #       f"chunk_len={chunk_len}")

            if chunk_id == '0A':  # image descriptor
                _write(chunk_hdr[:5])

                lzw_mincodesize = ord(_read())
                # print(f"lzw min code size:{lzw_mincodesize}")

                lzw_maxcodesize = ((1 << lzw_mincodesize) |
                                   ((1 << (lzw_mincodesize-1)) - 1))
                # print(f"lzw max code size:{lzw_maxcodesize}")

                lzw_codesize = lzw_mincodesize+1
                lzw_codemask = (1 << lzw_codesize)-1
                lzw_clearcode = (lzw_clearcode := 1)
                lzw_endcode = (lzw_endcode := lzw_clearcode + 1)
                lzw_nextcode = lzw_endcode + 1

                lzw_table = [x for x in range(lzw_maxcodesize)]
                lzw_index = 0
                lzw_curcode = nextbyte = _read()

                lzw_prevbuf = (_read(lzw_maxcodesize//2)+
                              [0]*((lzw_maxcodesize//2)+
                                      (lzw_maxcodesize & 1)))

                lzw_outputbuffer = []
                lzw_outputbufferpos = 0

                img_buff = []
                while lzw_nextcode <= lzw_maxcodesize:
                    while lzw_curcode >= lzw_clearcode and \
                            lzw_curcode < lzw_endcode:
                        lzw_currentprev, *lzw_currentprevptr = lzw_prevbuf

                        if lzw_currentprev is lzw_endcode:
                            lzw_outputbuffer.append(
                                lzw_prevbuf[-1] ^
                                lzw_prevbuf[0])

                            lzw_outputbufferpos += 1
                            continue

                        if lzw_index > 0:
                            lzw_index -= 1
                        elif lzw_nextcode >= lzw_maxcodesize:
                            break

                        lzw_prevbuf = lzw_currentprevptr

                    if lzw_nextcode >= lzw_maxcodesize:
                        break

                    if lzw_index < 0:
                        lzw_index = 0
                        lzw_curcode = nextbyte
                        nextbyte = _read()

                    lzw_oldcode = lzw_curcode
                    lzw_curcode = ((nextbyte >> (8-(lzw_index+lzw_codesize))) &
                                   lzw_codemask)
                    lzw_index += lzw_codesize

                    if (lzw_curcode!= lzw_nextcode) and \
                            (((lzw_oldcode = lzw_table[lzw_curcode]) <
                                  lzw_clearcode) or
                             ((lzw_oldcode >= lzw_nextcode) and
                                 (lzw_nextcode < lzw_maxcodesize))):
                        lzw_oldcode = lzw_nextcode

                    _img_buff = lzw_table[lzw_oldcode]
                    img_buff.extend(_img_buff)

                    _write(*_img_buff)
                    lzw_outputbuffer.append(_img_buff[0])
                    lzw_outputbufferpos += 1

                    if lzw_index == 0:
                        nextbyte = _read()
                        lzw_nextcode, lzw_oldcode = (
                            lzw_curcode, lzw_currentprev)

                        if lzw_curcode == lzw_endcode:
                            _img_buff = lzw_table[lzw_oldcode][0]
                            img_buff.extend(_img_buff)

                            _write(*_img_buff)
                            lzw_outputbuffer.append(_img_buff[0])
                            lzw_outputbufferpos += 1

                        if lzw_oldcode == lzw_clearcode:
                            lzw_curcode = lzw_endcode
                            lzw_nextcode = lzw_endcode + 1
                        elif lzw_oldcode == lzw_endcode:
                            lzw_curcode = lzw_clearcode
                            lzw_nextcode = lzw_clearcode + 1
                        else:
                            lzw_nextcode += 1

                        lzw_prevbuf[lzw_nextcode //
                                    (lzw_maxcodesize//2) %
                                    (lzw_maxcodesize & 1)] = lzw_oldcode

                    lzw_table[lzw_nextcode] = img_buff[-lzw_maxcodesize:]
                    lzw_nextcode += 1

                while lzw_outputbufferpos > 0:
                    chunk_sz = min(lzw_maxcodesize, lzw_outputbufferpos)
                    # print("chunk sz:", chunk_sz)
                    _write(struct.pack("<B", chunk_sz))
                    _write(b'\x00\x00')

                    _write(b''.join(map(lambda x: bytes([x]),
                                        lzw_outputbuffer[:chunk_sz])))
                    del lzw_outputbuffer[:chunk_sz]
                    lzw_outputbufferpos -= chunk_sz

                _write(struct.pack(">H", lzw_curcode))
                _write(struct.pack(">H", lzw_nextcode))
                _write(struct.pack(">H", lzw_maxcodesize))
                _write(struct.pack(">H", lzw_codesize))
                _write(b'\x00')

                _write(b'\xFF\xD9')  # end of frame
                chunk_len = len(img_buff)
                _write(chunk_hdr[:-1]+
                        struct.pack("<L", len(img_buff))+
                        chunk_hdr[-1:])

                chunk_data = b''.join([bytes([(i >> i_off) & 0xff
                                               for i in img_buff])] *
                                      (chunk_len // len(img_buff) + 1))[:chunk_len]

                assert len(chunk_data) == chunk_len
                _write(chunk_data)
                
                # print(f"chunk hdr:{chunk_hdr}")
                # print(f"chunk id:{chunk_id}")
                # print(f"chunk len:{chunk_len}")
                # print(f"chunk data:{chunk_data}")
            else:
                _write(chunk_hdr)
                chunk_len = chunk_len - 6
                if chunk_len:
                    _write(_read(chunk_len))
        _write(b'\x3B')
    except Exception as e:
        print(e)
        exit(-1)
    finally:
        if isinstance(infile, (BinaryIO, bytes)) and \
                hasattr(infile, 'close'):
            infile.close()
        if isinstance(outfile, (BinaryIO, bytes)) and \
                hasattr(outfile, 'close'):
            outfile.close()


if