#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output

        self._rng = open('/dev/urandom', mode='rb')


    def _make_header(self, width, height):
        
        '''
        This function creates the GIF header for a random frame with the given
        dimensions (width and height). It does not check any values apart from 
        the correctness of the parameters.
        '''

        header = b''
        header += b'GIF87a\x20\x20\x00\x00\x00,\x00\x00\x6f\xff'
        header += f'{width}x{height}'.encode('ascii')
        return header


    def generate(self, width, height):

        '''
        This functions generates a file containing a single random animated gif
        frame of the specified dimension. All pixels are set to black on white.
        The generated file is named "random.gif" by default or can be changed
        through the constructor parameter <output>.
        '''

        try:
            out = open(self.output, mode='wb')
        except IOError as err:
            print("Cannot access output file: {}".format(err))
            exit()

        self._generate_frame(out, width, height, True)
        out.close()


    def _generate_frame(self, out: BinaryIO, width: int, height: int, first_frame: bool):

        '''
        To allow for an animation this function repeats the generation logic of
        _make_header(), appending subsequent frames in a loop until the user
        interuptions the program. This is achieved using a generator pattern where
        every call returns another yield statement which contains the header 
        followed by the pixel data. If the <first_frame> flag is set to False, no
        additional header is added before the actual pixel data block.
        '''

        while True:

            if not first_frame:
                header = self._make_header(width, height)
            else:
                header = b''
            
            yield header
            data = bytearray()

            for i in range(width*height):

                # Every pixel has two elements, thus we need each element to hold 
                # one bit (so that the sum equals 1). We then use chr() to convert
                # between ints and bytes since bytes work with more than just ASCII
                elem = self._rng.read(1)[0] % 256
                
                data.append((elem * 4) % 256)
                data.append(((elem * 6) % 256) >> 2 | ((elem ** 2) << 6 & 0xc0))


            # Write the size of the upcoming data chunk (which is equal to 
            # 2 + len(data)), the color table with only black (R=0, G=0, B=0), 
            # global palette, color resolution in bits per pixel, background index
            # being zero and the packed field always being set to 1.
            yield bytes([(len(data) // 2) + 2,
                          0xf7,         # color table: R=0,G=0,B=0 only
                          width, 
                          height,      # logical screen descriptor fields
                          0b00000001])  

            yield bytes(data)
            break

        



        
if __name__ == '__main__':
    
    if len(sys.argv)!= 3:
        print("Usage:\n{} [width] [height]\n".format(__file__))
        exit()

    try:
        width = int(sys.argv[1])
        height = int(sys.argv[2])
    except ValueError as err:
        print("{} is not a valid number.\n".format(err))
        exit()

    try:
        rng = RandomGifGenerator()
        out = open(rng.output, mode='wb')
    except IOError as err:
        print("Cannot access output file: {}".format(err))
        exit()

    gen = rng._generate_frame(out, width, height, True) 

    # read the initial frame
    hdr = next(gen)       # yields header as bytes instance
    pix = next(gen)       # yields pixel array as bytes instance
    out.write(hdr+pix)    # append both to final file

   # now repeat adding new frames forever
    while True:
        try:                
            hdr = next(gen)  
            pix = next(gen)
            out.write(hdr+pix)
            
        except KeyboardInterrupt:
            print("\nExiting...")
            break

    out.close()
    




'''
This version uses less memory but needs to reassemble the header after every frame
to know its size. After creating the following test code I discovered that the previous 
version automatically appended the header to the output file without having to manually
reassemble it:

    while True:
        try:            
            hdr = next(gen)
            pix = next(gen)
            out.write(bytes(hdr)+bytes(pix))
            
        except KeyboardInterrupt:
            print("\nExiting...")
            break

The original version used BytesIO objects instead of open() calls so that they could be 
passed around like generators. That however required me to do some ugly conversion between 
bytes instances and normal ones at some point...
'''













# Below is my original attempt using BytesIO object to simplify the code above

'''
with open('test.gif', mode='wb') as out:

    rng = BytesIO(os.urandom())    # create pseudo-random number generator
    gen = rng._generate_frame(out, 160, 90, True)  

    width = 160
    height = 90

    # read the initial frame
    header = next(gen)           # yields header as bytes instance
    data = next(gen)             # yields pixel array as bytes instance
    out.write(header + data)     # append both to final file

    # now repeat adding new frames forever
    while True:
        try:                    
            header = next(gen)   
            data = next(gen)
            out.write(header+data)

        except KeyboardInterrupt:
            print("\nExiting...")
            break
'''




















































'''
Implementation notes about the old way of generating images.

First approach was reading data from /dev/urandom directly into the image while
creating the header. This turned up in the results to have a very large file
size (~1 GB!) even though there should be little change between frames due to
the fact that the seed is fixed. Thus I reimplemented the process such that
every frame is written to a temporary file and then only moved to the output
when the whole image has been created.

Out of pure curiosity I also tried writing each frame directly to the target
filestream. In contrast to the previous implementation this resulted in a file
of roughly the same size (~1 MB) although there were differences in the
contents. In particular, changing the seed had visible effects.

Finally, I wrote the header to a temporary filestream and then copied everything
together with the help of shutil.copyfileobj(). Again, the output file size
was roughly similar to either the previous implementation or the direct write
approach. However, when comparing the contents it became apparent that the
binary representation was somewhat different. For example, the RGB values of
one frame looked like this:

    \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff\x00\x00\xff\xff
  ...
    \x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff

This means that some of the pixels seem to be mixed together inside the blocks.
I am currently not sure whether this is a problem caused by packing the colors
differently or something else entirely.
'''






# Original method using BytesIO objects (abandoned because it didn't really improve on
# the previous one)
'''
while True:
    try:    
        img = Image.new('RGB', (160, 90), color=(0, 0, 0))

        for y in range(img.height):
            for x in range(img.width):

                r = bin(rng.read(1)[0])[2:]               # read red componet
                r = ('0'*(8-len(r))) + r                  # fill with leading zeros
                g = bin(rng.read(1)[0])[2:]               # read green component
                g = ('0'*(8-len(g))) + g                  # and fill with leading zeros
                b = bin(rng.read(1)[0])[2:]               # read blue component
                b = ('0'*(8-len(b))) + b                  # and fill with leading zeros
                img.putpixel((x,y),(int(r[:8],2)*2**16,
                                    int(g[:8],2)*2**8,
                                    int(b[:8],2)))      # store result

        img.save(out,'GIF')
        
    except KeyboardInterrupt:        
        print("\nExiting...")
        break
'''







    
    

'''
Bit processing experimentation using Python's BitArray module.

Results from converting each colour value individually:

    setitem       ~11 uses
    getvalue      ~14 uses
    pop           ~13 uses
    extend        ~24 uses    bs = rng.read(1)[0]         # block size is an unsigned char
    subblock = rng.read(bs - 1)  # discard sub-block marker
    header += bytes([bs])        # add block size back again
    data = rng.read(bs)          # get actual image data
    out.write(header + data)     # write it all together
    
