#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the [LZW](https://en.wikipedia.org/wiki/Lempel%E2%80%93Ziv%E2%80%93Welch) algorithm and given palette colors in `palette`.  The output is compressed via LZW, but not by any fancy scheme (i.e., no RLE).
def generate_lzw_gif(out: BinaryIO, palette: bytes):
    # Limit of code size in LZW compression.
    MAXCODESIZE = len(palette)-1

    # First byte of header block is magic number; rest are fixed values.
    out.write(b"GIF87a")
    out.write(struct.pack("<HHBBB", 640, 480, 0xF0 | 0x07, 0, 0))

    # Image descriptor block.
    out.write(b"\x2C\x00\x00\x00\x00\x00,\x00\x00\x00\x00\x01\x00\x18\x00\x0F\xFF\xF9")

    lzwflags = {
        1: b'\x00',      # GCTF_GLOBAL_COLORMAP
        0: b'\x80'       # GCTF_LOCAL_COLORMAP
    }[len(palette)]

    out.write((
        b"\x21\xFC"                        # Graphic Control Extension Block
        + b"\x0B"                           # Size of extension data in bytes (11)
        + b"\x0A"                           # Packed field - global color table flag set, disposal method 0, transparent color index 15
        + b"\x00\x00"                       # Delay timer x 10ms
        + (lzwflags if palette else b'')   # Local color map or special bits for LZW encoding
        + ((palette+b'\x00'*255)[:MAXCODESIZE]
            if palette                  # If we're using local color map...
            else b"\x02")                 #...then code size is 2bits; otherwise it's maxed at 12-bit mode
    ))


    # Application comment block.
    out.write(b"\x21\xFE\x0D\x0A\x47\x49\x46\x38\x39\x61")
    out.write("Generated with https://github.com/CleverRaven/Cataclysm-DDA".encode() + b"\x00")

    # Logical screen descriptor block.
    out.write(b"\x2c\x00\xff\x00\x00\x00\x00\xf0\x00\x07\x00,\x00\x00\x00\x00\x01\x00\x18\x00\x0f\xff\xf9")

    # Encode pixels as ints to prevent issues with bit shifts on Python 2.
    # Byte order does need to be big endian here because this block will be at the end of the file before actual pixel data starts.
    out.write(bytearray([0]))

    # Initialize LZW state machine variables.
    clearcode = 1 << MAXCODESIZE              # Code indicating that the next input character should reset the dictionary.
    endofinfo = clearcode + 1                 # Code indicating that there is no more input to be encoded after the current dictionary.
    prevchar = None                          # Previous character read from input stream; used to construct new strings from old ones after performing lookups in the dictionary.
    dictsize = 2                             # Current contents of dictionary. Starts off with two "empty" characters so we don't have to add an extra condition around initialization of prevchar when we decrement dictsize below.
    curcode = None                           # Current string being written out at this point in time.
    curcodesize = 99                         # Starting value of codesize; it'll get overridden once our first character has been processed.
    writepending = False                     # Set whenever a chunk of compressed output needs writing back out even though its length hasn't yet reached granularity.

    def encode():
        """Encodes the current string (`curcode`) into LZW format."""

        nonlocal curcode, curcodesize

        # No-op if empty string. This happens when we reach the maximum size allowed by our current LZW encoding scheme.
        if len(curcode) == 0: return

        # When a previous code was already present in the dictionary, we can just append the last char of the current one onto what we've already put out.
        if prevchar!= None:
            curcode += prevchar[-1]

            # Write it out.
            while True:
                code = int(''.join(['{:08b}'.format(ord(x))[::-1] for x in curcode]), base=2)

                # Check if we shouldn't dump anything right now due to incomplete granularity.
                if writepending:
                    out.write(bytes([(code >> (8*writepending)) & 255]))

                    writepending -= writepending    # Reset back to zero.

                    if writepending > 0: break     # Still pending? Exit loop.


                elif len(curcode)*curcodesize <= 255:
                    # We've got full granularity! Dump the char/string pair.
                    out.write(bytes([len(prevchar)+1, ord(prevchar)]))
                    out.write(bytes([code & 255]))

                    curcode = ""                    # Clear everything.
                    break                           # Done with this loop iteration.

                else:
                    # Not enough granularity left; save some output for later.
                    out.write(bytes([int(dictsize/256), dictsize&255]))
                    out.write(bytes([(code>>8)&255]))        # Only need top half.
                    out.write(bytes([code&255]))               # Only need bottom half.

                    writepending = curcodesize//8         # How many bytes still need to be written out?
                    curcode = ''                            # Clear everything.

                    break                                   # Done with this loop iteration.

        prevchar = curcode                      # Remember the last character processed.
        del curcode                              # Clear the variable so it doesn't persist past this function call.

    try:
        # Open a binary file named `infilename` for reading.
        with open("/dev/urandom", 'rb') as f:
            # Keep track of how many times we exit the inner encode routine without hitting the end of the file.
            # An exception is raised if the while loop never runs, meaning none of the codes were added to the dictionary.
            exhaustioncount = 0


            # Read chunks of random data until EOF.
            while True:
                # Grab a chunk of data.
                c = f.read(1<<18)

                if len(c) < 1<<17: break          # Reached EOF; stop processing.

                # Compress this chunk of data.
                for i in range(len(c)):
                    # Look up encoder code associated with current char.
                    enc = (ord(c[i]) if isinstance(c, str) else c[i]) ^ (dictsize * 256)


                    # Add code to dictionary if necessary.
                    if enc in (clearcode, endofinfo):
                        if dictsize >= MINCODESIZE:
                            curcodesize = dictsize - MINCODESIZE

                            # Update dictionary based on current state.
                            encode()

                        if enc == clearcode:
                            # Reset dictionary contents.
                            dictsize = 2
                            prevchar = None

                            # Start over again with our newly updated context for the next string.
                            continue


                        # End of information marker; we're done adding codes to the dictionary.
                        break



                    # New unique code found; add it to dictionary.
                    if dictsize < MAXCODESIZE:
                        dictsize += 1
                        curcodesize = dictsize - MINCODESIZE

                    # Insert new code into dictionary.
                    curcode = ((chr(enc // 256) if enc>=256 else '')
                                + chr(enc % 256)).encode()

                    # If this code exceeds the granularity limit, flush the buffer.
                    if len(curcode)*curcodesize > 255:
                        encode()


    except Exception as e:
        print("[!] Error encountered during random data generation:\n"+str(e)+"\n"+
                "[i] Skipping to end of LZW compressed GIF...")



    # Flush remaining data from above processing.
    encode()


if