#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = 1<<24
    data = rng.read(datalen)

    # http://giflib.sourceforge.net/whatsinagif/bits_and_bytes.html
    # logical screen descriptor (logical screen area = global color table)
    width = 0x163e
    height = 0x7d5f
    flags = 0b10110100   # color resolution (3 bits), sorted flag is set
    background = 0x00      # background color index (not used here)
    aspect_ratio = 0xf     # pixel aspect ratio is fixed to 1.0

    struct.pack_into('>H', out, 0, version=0x0003, signature=header)
    offset = len(header) + 4

    # application extension block
    out.write(b'\x21\xFF')
    out.write(b'sBIT')
    out.write(b'\x0c')              # subblock size
    out.write(struct.pack('BBB', 0x11, 0x11, 0x11))    # RGB palette
    offset += (1+1+1+1)

    # graphic control extension block
    out.write(b'\x21\xF9')
    out.write(b'\x04')              # subblock size
    out.write(struct.pack('BBBBBB',
        0b11111111 & int((width * height / 2)),        # duration in milliseconds
        0x00,                                           # transparent color index
        0x00,                                           # reserved
        0b11000000 | 0b00000011 if flags&0b1 else 0b11000001,
            0b11111100 if flags&0b100 else 0b11111111,
            0b11111000 if flags&0b1000 else 0b11111111,
                0b11110000 if flags&0b10000 else 0b11111111,
                    0b11100000 if flags&0b100000 else 0b11111111,
                        0b11000000 if flags&0b1000000 else 0b11111111,
                            0b10000000 if flags&0b10000000 else 0b11111111
    ))
    out.write(struct.pack('<B', 0x00))       # end of sequence
    out.write(struct.pack('<B', 0x00))       # no disposal method
    out.write(b'\x00')                      # no transparency information
    offset += (1+1+1+1)+1+1+1+1+1          # only one byte required for short values

    # image descriptor block
    out.write(b'\x2C')                  # image separator
    out.write(struct.pack('<HHHHB',
        width%2**16,                     # left position
        height % 2**16,                 # top position
        width%(2**16*2**8)/2**8,         # width (low byte)
        height%(2**16*2**8)/2**8,       # height (low byte)
        0b00001010                   # local color table follows and contains one entry
    ))
    out.write(struct.pack('<B', 0x00))       # LZW minimum code size
    offset += 1+(2*4)+(1+1)               # 1 byte required for each integer value

    out.write(data[:datalen//4])           # first frame uses half as much space
    offset += datalen // 4

    out.seek(offset)

    out.write(b'\x3B')                       # terminator

