#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        
    else:
        paldata = []

    # calculate how many bytes we need to get back to an even
    # byte position
    pad = ((image_width * 3) + 1) & ~1
    
    # write the terminator
    out.write(b'\x3B')
    
    # write the image data block
    for i in range((pad + 1) // 2):
        
        # image separator
        out.write(header[i])
    
        # field which is not very interesting to us:
        # left, top, width and height of next frame
        out.write(rng.read(2))

        # packed fields
        flags = rng.read(1)[0]
        lctflag = bool(flags & 0b10000000)

        if lctflag:
            
            # local color table flag set so there's no global
            # one either; instead they're stored as a 3-byte RGB triple

            # TODO: what do I do with this?
            pass

        else:
    
            # use the global color table
            assert len(paldata) == ctsize
        
            # put in the correct number of colors from our
            # randomly generated palette
            out.write(bytes([ctsize]))
            out.writelines(paldata[:ctsize])

        # image data starts here
        # first two bytes represent length of LZW compressed data subblock
        compdatalen = rng.read(2)
        out.write(compdatalen)
        
        # remainder is the actual LZW compression data
        compdata = rng.read(int.from_bytes(compdatalen, 'little'))
        out.write(compdata)

    
  