#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
        
        with open('/dev/urandom', mode='rb') as rng,\
             open(self.output,    mode='wb') as out:

            return generate_random_gif(
                    rng, out
                )



# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# are both binary streams (`io.BufferedRandom`).
def generate_random_gif(rng: BinaryIO, out: BinaryIO) -> bool:
    """ Generate a random gif from /dev/urandom and save it to `out`. """
    
    # Header
    header = [0x47, 0x49, 0x46, 0x38] + [0 for _ in range(12)] 
    header[6] += 0b10000000
    header[6] |= 0b00000001     # Global color table
    header[7] |= 0b00000001     # Color resolution
    header[8] |= 0b00000001     # Sorting flag
    header[9] |= 0b00000001     # Size of global color table (256 colors)
    header[10] = 255             # Background color index (transparent)
    header[11] = 0               # Pixel aspect ratio (1:1, so ignore me!)
    out.write(struct.pack('!BBBBHHBBBHB2X', *header))

    width  = struct.unpack('B', rng.read(1))[0] % 32 + 1      # Width should be between [1, 32]
    height = struct.unpack('B', rng.read(1))[0] % 32 + 1      # Height should be between [1, 32]
    pixels = bytearray([ord(rng.read(1)) for _ in range(width*height)]) # Array of pixel values

    colortable = sorted(list(set(pixels)))                      # Get list of unique colors
    palette = b'\xFF\xFF'.join(struct.pack('<H', c) for c in colortable[:256])   # Create color palette
    out.write(palette)                                          # Write color palette to new image
    out.write(bytes(pixels))                                    # And finally write our pixels!

    return True




if __name__ == '__main__':
    gen = RandomGifGenerator()
    if len(sys.argv)!= 1:
      try:
          for i in range(-int(sys.argv[-1]), int(sys.argv[-1])):
              gen.generate_random_gif()
      except Exception as e:
        print("Usage: {} [count]\nGenerate a random gif file.".format(__file__))
        print("Specify count amount.")
    else:
      gen.generate_random_gif()
    