#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
    
        with open(f'./{self.output}', 'wb') as fh:
            generate_random_gif(open('/dev/urandom'), fh)


# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# correspond to the information found here:
# https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Signature
    header = b'GIF89a'
    
    header += bytes([45, 57, 65, 255])   # version numbers
    global color_table_size              # use global variable
    color_table_size = rng.read(1)[0]    # RGB size of color table (1 byte)
    header += bytes([color_table_size, 0])# color table entries (2 bytes)
    bg_color = rng.read(1)[0]            # background index (1 byte)
    header += bytes([bg_color])           # reserved (1 byte)
    loop_flag = rng.read(1)               # has bit flags indicating looping
    header += loop_flag                   
    disposal_method = rng.read(1)         # disposal method on close (1 byte)
    header += disposal_method 
    transparent_flag = 0x01           # tRNS flag set by default
    header += bytes([transparent_flag, 0xff])   # transparent color index
    
    for _ in range(color_table_size):      # color table
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        a = 0xFF                           # opaque
        header += bytes([r, g, b, a])      # each entry is RGBA
        
    width = rng.read(2)                   # image dimensions (2 bytes)
    height = rng.read(2)                  # in pixels (2 bytes)
    pixel_aspect_ratio = 0                # no pixel aspect ratio specified
    header += width                       # add width and height to header
    header += height                      # after converting them into bytes
    header += bytes([pixel_aspect_ratio]) # add final byte in header

    # Graphics Control Extension Block
    graphic_ctrl = b'\x21\xF9\x04\x00\x00\x00\x00\x00\x2C'
    delay_time = rng.read(2)             # Delay time (2 bytes)
    trans_index = 0                      # transparence index defaults to 0
    terminator = b'\x00'                 # end of extention block
    graphic_extn_blk = graphic_ctrl + \
                       delay_time + \
                       bytes([trans_index]) + \
                       terminator       # assemble GCE block
    out.write(graphic_extn_blk)          # write GCE block to output buffer

    # Image Descriptor
    left_pos = 0                         # position off screen from left edge
    top_pos = 0                          # position off screen from top edge
    img_width = int.from_bytes(width, "big")
    img_height = int.from_bytes(height, "big")    
    img_desc = b'\x2c'                   # image separator
    img_interlaced_flag = b'\x00'        # image not interlaced 
    img_sorted_flag = b'\x00'            # image not sorted
    lct_flag = b'\x00'                   # local color table flag
    lct_size = 0                         # no local color table defined
    lct_entries = 0                      # no local color table entries
    img_data = b''                       # image data will be filled later
    descriptor = img_desc + \
                 left_pos + \
                 top_pos + \
                 img_width + \
                 img_height + \
                 img_interlaced_flag + \
                 img_sorted_flag + \
                 lct_flag + \
                 bytes([lct_size, lct_entries]) + \
                 img_data  # assemble ID/img data block
    out.write(descriptor)                # write ID/img data to output buffer

    # Plain Text Extension
    text_extn = b'\x21\x01\x00\x01\x00\x00\x00' # graphics control block
    title_len = len('Random Title') # length of our title string
    title =''.join('{}'.format(title_len))+'!'
    out.write(text_extn+title.encode()) # append title to extension block

    # Application Extension
    app_extn = b'\x21\xFF\x0BNETSCAPE'    # app ID (11 letters)
    app_auth_code = b'\<PASSWORD>'     # auth code (3 bytes)
    app_version = b'\x03\x01'             # version (2 bytes)
    app_loop_flag = b'\x01'               # repeat animation
    netscape = app_extn + \
               app_auth_code + \
               app_version + \
               app_loop_flag # assemble full app extension
    
    app_extn_blk = netscape # append it to the main GIF payload
                        
    out.write(app_extn_blk)  # write full application extension
                            # to output buffer

    # Trailer
    trailer = b'\x3b'                     # trailer
    out.write(trailer)                    # write trailer to output buffer

    # Frames of Color Table
    for i in range(0, color_table_size):  # add color table to frame list
        red = ord(header[i*4+1])
        green = ord(header[i*4+2])
        blue = ord(header[i*4+3])
        alpha = ord(header[i*4+4])
        frame = [red, green, blue, alpha]
        frames.append(frame)

    # build up image data blocks iteratively
    while True:                        # do until we run out of input
        bs = rng.read(1)[0]
        subblock = rng.read(bs - 1)
        
        if len(subblock)!= (bs - 1):
            break
            
        data.append(bs)
        data.extend(subblock)

    # insert it into the main GIF payload
    gif_payload = []
    gif_payload += [len(data)]       # add data block tag to payload
    gif_payload += data              # append data block contents
    gif_payload += [0]               # start new data block
    gif_payload += [0]               # end with single zero byte

    # convert to C-style uint_8 array
    raw_payload = bytes(gif_payload)

    # write the data into the file
    out.write(raw_payload)

    # return success status
    return True



if __name__ == '__main__':
    try:
        outfile = str(sys.argv[1])
        print(outfile)
    except IndexError:
        pass
    rg = RandomGifGenerator(output=outfile)
    rg.generate_random_gif()

        
        bs = rng.read(1)[0]         # block size is an unsigned char
    subblock = rng.read(bs - 1)  # discard sub-block marker
    header += bytes([bs])        # add block size back again
    data = rng.read(bs)          # get actual image data
    out.write(header + data)     # write it all together
    
