#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):

            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]

        # write to output stream
        out.write(header + struct.pack('<HHBB', image_width, 
            image_height, ctsize | 0b10000000, 
            0) + palents)

    else:

        # no palette
        out.write(header + struct.pack('<HH', image_width, 
                image_height))

    # get remaining lzw compressed info from the RNG
    datalen = int.from_bytes(rng.read(2), 'big')
    data = rng.read(datalen)

    # lzw compression
    lzwmincodesize = int.from_bytes([data[0]], 'little')
    
    out.write(lzwmincodesize.to_bytes(1, 'little'))

    table = dict([(entry, entry) 
        for entry in range(2**lzwmincodesize+2)])

    last = None
    bytepos = 1
    while True:
        
        if last is not None:
            
            try:

                code = data[bytepos-1] + data[bytepos]*256 + 1

                out.write((table[last][0]).to_bytes(1,'little'))

            except KeyError:
                
                out.write(((table[last][0] + 1).to_bytes(1,
                    'little')))
                out.write((((table[last][0]) // 256).to_bytes(
                    1, 'little')))
                
            
        try:

            next = table[code]

            # add this one to our dictionary so it can be referred back to later
            table[(last,next)]=code

            last = next
            bytepos+=1

        except KeyError as e:

            # we've run out of codes, let's flush them all out!
            # NOTE: I don't know if this actually matters or not
            for k,v in sorted(list(table.items()), key=lambda x: x[0]):
                out.write(v.to_bytes(1, 'little'))
                
            break
            