#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self._output = DEFAULT_OUTPUT

        else:

            _, ext = os.path.splitext(output)
            
            if '.gif'!= ext:
                raise ValueError('Output file must have extension ".gif".')

            elif os.path.isfile(output):
                raise RuntimeError(f'"{output}" already exists.')

            else:
                self._output = output


    @staticmethod
    def create(prng: BinaryIO, out: str) -> None:
        
        rg = RandomGifGenerator()
        with open(out, 'wb+') as f:
            generate_random_gif(prng, f)


    def run(self) -> None:

        prng = self.__create_prng()
        print("Generating GIF...")
        try:
            self.create(prng, self._output)
            print(f"Generated \"{os.path.basename(self._output)}\".")

        finally:
            pass
        
        
    def __create_prng(self):
        """Create the PRNG."""
        return os.urandom()


# random number generator `rng`. The parameters
# are read from different sources to ensure that they're valid.
def generate_random_gif(rng: BinaryIO, out: BinaryIO) -> None:
    
    # Header
    header = b''
    header += b'GIF89a'            # File signature and version
    # Color Table size
    ctsz = rng.read(1)[0] * 3
    for _ in range(ctsz):
        header += bytes([rng.read(1)[0]])

    # Block Size (Data Sub-Block within Data Stream)
    bs = rng.read(1)[0]         # block size is an unsigned char
    subblock = rng.read(bs - 1)  # discard sub-block marker
    header += bytes([bs])        # add block size back again
    data = rng.read(bs)          # get actual image data
    out.write(header + data)     # write it all together
    
