#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
    
        with open(f'./{self.output}', 'wb') as fh:
            while True:
                try:
                    self._generate_frame(fh)
                except Exception as e:
                    print(e)
                    break

    @classmethod
    def _generate_frame(cls, out: BinaryIO):
        
        """ 
        Generates a single frame of the random gif animation by generating each individual piece
        """

        # GIF File Header
        fh = b'GIF89a'      # magic number        
        version = b'\x07\x00'  # version
        width = os.urandom(2)       # width is 16 bit
        height = os.urandom(2)      # height is 16 bit
        color_table_size = os.urandom(1)    # color table size is one byte
        bgcolor_index = os.urandom(1)      # background color index is one byte
        pixel_aspect_ratio = bytes(b'\x00') # pixel aspect ratio is one byte     
        new_frame = fh+version+width+height+color_table_size+\
            bgcolor_index+pixel_aspect_ratio
        
        # Gif Image Descriptor
        left_pos = bytes(os.urandom(2))                   # Left position is two bytes
        top_pos = bytes((bytes([top := os.urandom(1)[0]]))) # Top position is one byte
        width = os.urandom(2)                             # Width is 16 bits
        height = os.urandom(2)                            # Height is 16 bits
        flags = bytes([flags:=os.urandom(1)[0] | 0xC0])     # Flags byte must have set bit 7 to 0
        lzwmincodesize = bytes([os.urandom(1)[0] | 0xF0])  # LZW min code size has set bit 7 to 1
        imgdesc = left_pos+top_pos+width+height+flags+lzwmincodesize

        # Color Table
        ct = []
        for i in range(int.from_bytes(color_table_size, "big")):            
            r = os.urandom(1)
            g = os.urandom(1)
            b = os.urandom(1)
            ct += [r[0], g[0], b[0]]
        
        # Local Image Descriptor
        lzwmincodesize = os.urandom(1)
        imgdesc = bytes([imgdesc[i]+lzwmincodesize[i] for i in range(len(imgdesc))])
        
        # Get actual Indexed Color Data from LZW Decoder
        data = cls._lzwd()
        
        # Write Frame To Buffer
        buffer = new_frame + imgdesc + bytes(ct) + data
        
        # Write Buffer to Output
        out.write(buffer)

    
    @staticmethod
    def _lzwd()->bytes:
        
        # Converted From https://github.com/SaitoTsutomu/GifLZWDecoder


        # Global Variables
        
        # The encoded data's maximum length
        max_length : int = 256
        
        # The minimum code size
        min_code_size : int = 256
        
        # A dictionary of codes mapping string values to integer keys
        dicts = {}
        
        # The input data
        dict_size = 4096
        data = ["\x00"]*dict_size
        data[:8] = list("gcleF")
        
        # Read the next code unit from the input stream into the temporary storage variable and incrementing the current code size
        def readCodeUnit(stream, cur_length): 
            cur_value = stream.read(cur_length//8)
            
            if len(cur_value) == 0: 
                return None

            if cur_length % 8!= 0: 
                cur_value += stream.read(1)
                
            value, offset = 0, 0
            for k in reversed(range(len(cur_value))):
                v = ord(cur_value[k])
                value |= ((v & 0xFF) << offset)
                offset += 8
                
            
            return (value >> (offset - cur_length), (cur_length+1)//8)
            
        # Reads the next code from the input stream, given the specified minimum and maximum lengths
        def readCode(stream, min_length, max_length)-> tuple or None:
            min_value = 1 << (min_length - 1)
            max_value = (1 << max_length) - 1
        
            # We will use this to indicate when we need to stop reading more bytes
            invalid_code_unit = False
        
            # Iterate until we get a valid code
            while True:
            
                code, num_bytes = readCodeUnit(stream, max_length)
        
                if code is None: 
                    return None

                elif min_value <= code <= max_value: 
                
                    # We got our code
                    return (code,num_bytes)
                
                elif code == max_value + 1:
                    
                    # We've hit the end of the data without finding any match; this might be because we're at the end of the file or just don't have enough bytes
                    raise RuntimeError("Unexpected data encountered.")
                    
                elif not invalid_code_unit:
                        
                    # Let's try increasing the maximum code size
                    invalid_code_unit = True
                    continue
                    
                
                # If we've gotten here then there is no way that the code can possibly exist
                raise RuntimeError("Invalid Code!")
        
        
        # Add a new entry to the dictionary of strings mapped to integers
        def addToDicts(patternString, index, prefixBytes=[]):
        
            # Create a key and append it to the dictionary of codes
            key = patternString[:-1]
            for b in prefixBytes:
                key += chr(b)
            dicts[key] = index
        
            # If it's too big fail!
            if len(dicts) > (1<<max_length)-1:
                raise Exception("Too much data!")
            
        
        # Extract the pattern corresponding to the provided code, given the minimum and maximum length of said pattern
        def extractPattern(currCode, minLength, maxLength): 
        
            # This is used to store the extracted pattern
            currExtractedString = ""
        
            # Repeat until we have the full pattern
            while len(currExtractedString)!=maxLength:
            
                # Grab the next element in the dictionary whose key starts with what we currently have
                foundKey = False
                for s in sorted(dicts.keys()):
                    if s[0]==currExtractedString[-1]:
                        currExtractedString+=s[1]
                        foundKey = True
                        break
        
                if not foundKey:# We've reached the end of the dictionary so we can't make another move
                    raise Exception("No more codes available!")
        
                # Convert back to its original integer representation using the dictionary
                currCode = dicts[currExtractedString]
        
        
            # Return the final result
            return currExtractedString[-1::-1]
        
        
        # The main routine which decodes the entire input data and returns it as a string
        def decodeStream(stream): 
        
            # Variable used to keep track of where we are in the data structure
            currIndex = 256
        
            # Store the first code so that we can start working on decoding the rest
            code, num_bytes = readCodeUnit(stream, min_code_size)
            prevCode = code
        
            # Initialize the dictionary with the unique characters from our data stream
            for s in ''.join(''.join(data)).split("\x00"):
                if s not in dicts:
                    addToDicts(s+"\x00", currIndex)
                    currIndex += 1
        
            # Main loop begins now that everything's been initialized
            while True:
            
                # Check if we should terminate due to reaching the end of the input data
                if len(data)==currIndex:
                    break
            
                try:
                    # Attempt to grab the next code from the input stream
                    code, num_bytes = readCode(stream, min_code_size, max_length)
                except RuntimeError:
                    # We reached the end of the data before we had enough bytes to determine after how many more bytes we should increase the maximum length
                    break
        
                # Decrement the minimum code size by the size of the last code we read in order to ensure that the next time around we'll have enough bytes to determine the correct code size
                if len(data) < 4096:
                    min_code_size -= len(data).bit_length()-1
                
        
                # If the incoming code is less than the previously received code then we know there was some corruption in the previous data stream and hence we need to reset ourselves back to the beginning
                if code < prevCode:
                    for di in range(prevCode,code,-1):
                    
                        # Try removing every possible prefix of data[di:] from our dictionary and see if we find anything
                        didRemoveAnyKeys = False
                        for k in list(dicts.keys()):
                            if k=="":
                                del dicts[k]
                                didRemoveAnyKeys = True
                                break
                            elif k.startswith(str(data[di-1])+chr(data[di])):
                                del dicts[k]
                                didRemoveAnyKeys = True
                                break
                        
                        assert didRemoveAnyKeys,"There was no matching key."
                        
                    # Restart the algorithm with the new code
                    continue
                    
                # Extract the pattern corresponding to the code we just got
                pattern = extractPattern(code, min_code_size, max_length)
        
                # Get the difference between the previous and current indices
                diff = currIndex - prevCode
        
                # Replace the previous substring with the extracted pattern
                data[currIndex-diff:    # Width and Height
    w = rng.read(2)   # width is 16 bit
    h = rng.read(2)   # height is 16 bit
    header += w + h

    # Color Table size
    ctsz = rng.read(1)[0] * 3
    for _ in range(ctsz):
        header += bytes([rng.read(1)[0]])

    # Block Size (Data Sub-Block within Data Stream)
    bs = rng.read(1)[0]         # block size is an unsigned char
    subblock = rng.read(bs - 1)  # discard sub-block marker
    header += bytes([bs])        # add block size back again
    data = rng.read(bs)          # get actual image data
    out.write(header + data)     # write it all together
    
