
from typing import BinaryIO

# Generates a random GIF file into `out` using the given palette. The size of
# each frame is determined by `frame_size`. The number of frames in total
# is determined by `num_frames`, or if that's zero, calculated based on the
# duration of time and framerate.
def make_random_gif(
    out: BinaryIO, 
    palette: Palette,
    duration_ms: int = 1000, 
    num_frames: int = 0, # Or specify this instead for exact control over length (then frame rate will be ignored).
	frame_size: Dimensions = (64, 64), 
	fps: float = 2):
    
    # Compute frame count from duration & fps
    if not num_frames:
        num_frames = round((duration_ms / 1000) * fps)

    # Prepare headers
    header_block = bytes(0b1110 << 4 | 3) + bytearray([0xFF] * (6))
    app_ext_header = b"NETSCAPE2.0\x03\x01\x00\x00"
    graphic_ctl_ext = b"\xF9\x04"
    comment_ext = b"Generated with gif-palette v0.1.1" + (bytes(5) * 8)
    table_based_image_data_looping = b"\x01\x01"
    
    image_descriptor = bytes([
        0b1100 << 4 | palette.bits_per_entry, 
        frame_size[0], frame_size[1], 
        0,      # Left position
        0,      # Top position
        palette.bits_per_entry, 
        0       # LZW minimum code size
    ])

    # Write first header block
    out.write(header_block)

    # Write application extension
    out.write(app_ext_header[:3])
    write_uint16be(out, len(app_ext_header)-3)
    out.write(app_ext_header[3:])

    # Write graphics control extenstion
    out.write(graphic_ctl_ext)
    write_uint16be(out, 4)        # Block Size
    write_uint16be(out, num_frames)   # Loop Count
    write_uint8(out, duration_ms % 256)     # Delay Time
    out.write(table_based_image_data_looping)    # Transparent Color Index
    out.write(b"\x00")           # Terminator

    # Write comment extension
    out.write(comment_ext)

    # Start data block
    out.write(b";")

    # Write global color table
    for entry in palette.as_rgb():
        write_uint8(out, entry[0])
        write_uint8(out, entry[1])
        write_uint8(out, entry[2])

    # Write individual frames
    for _ in range(num_frames):

        # Write image descriptor
        out.write(image_descriptor)

        # Generate some garbage pixels before we start writing RGB values to
        # create "transparency".
        for _ in range(round(math.sqrt(sum(i**2 for i in frame_size)))): 
            pixel = next(palette.make_gif_stream())
            write_uint8(out, pixel[0])
        
        # Write actual RGB values
        stream = palette.make_gif_stream()
        while True:
            try:
                rgb_values = [next(stream) for _ in range(frame_size[0]*frame_size[1])]

                for r, g, b in rgb_values:
                    write_uint8(out, r)
                    write_uint8(out, g)
                    write_uint8(out, b)

            except StopIteration: break

    out.flush()

