#!/usr/bin/env python3

import struct
from typing import BinaryIO

def step_image(
        fps: float=50./12., width: int=640, height: int=480, colors: int=2**12-7
):
    from math import sin
    from PIL import Image

    # Animate a grid of sine waves
    img = Image.new('RGB', (width*colors, height), 'white')
    for x in range(width):
        for y in range(height):
            f = ((x / width) - 0.5) * 2    # [-1,1]
            t = sin(y / height * 2 * 3.1415)   # [0,1]
            c = int((t+f)*colors//2)      # [0,2^12-1]
            img.putpixel((c*width + x, y), (0, 0, 0))
    return img

def write_gif(
        filename: str='wallpaper.gif', frameduration: float=1/50.
):
    ''' Write an animated GIF image to the given file descriptor. '''
    from PIL import Image

    # Compose frames into sequence and save as GIF
    imgseq = []
    for i in range(1000):
        img = step_image()                            # Generate next frame
        imgseq += [img.quantize()]                    # Add color table & pixel data

    imgseq[0].save(filename, format='GIF', append_images=imgseq[1:],
                   save_all=True, duration=frameduration, loop=0)

def embed_gif_in_png(
        pngfile: str, giffile:str, outputfile:str, xywh=(20, 20, 200, 200)
):
    """Embeds the content of `giffile` inside `pngfile`. The coordinates are
    expressed relative to PNG's upper left corner."""
    with open(giffile, 'rb') as gfh:
        with open(pngfile, 'r+b') as pfh:
            read_and_replace_png(pfh, xywh, gfh.read())
    print("Saved {}.".format(outputfile))

def read_and_replace_png(fhandle : BinaryIO, coords, replace):
    def chunklen(i):
        return struct.unpack(">I", fhandle.read(4))[0]

    assert fhandle.read(8) == b"\211PNG\r\n\032\n"
    while True:
        length = chunklen(fhandle)
        type_ = fhandle.read(4).decode()
        if type_ == "IHDR": break

        fhandle.seek(length, 1)
    w, h = struct.unpack(">II", fhandle.read(8))
    old_data = bytearray(w*h*4)
    offset = fhandle.tell()
    fhandle.readinto(old_data)

    nbytes = len(replace) // 4
    new_data = bytearray(w*h*4)

    startx, endx, starty, endy = map(int, coords)
    stride = w*4
    for j in range(-starty, endy):
        srcoff = offset+(j*stride)+startx*4
        dstoff = (h-endy+j)*(stride)+(startx*4)
        new_data[dstoff:(dstoff+nbytes)] = replace[(j+starty)*nbytes:]
        new_data[(dstoff+nbytes):(dstoff+stride)] = \
                old_data[(srcoff+nbytes):(srcoff+stride)]

    fhandle.seek(offset)
    fhandle.write(struct.pack(">I", len(replace)))
    fhandle.write(replace)
    fhandle.seek(offset+len(replace)+8)
    fhandle.write(struct.pack(">I", len(new_data)-len(replace)-8))
    fhandle.write(new_data[len(replace)+8:])

def embed_random_data(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = 1<<24
    data = rng.read(datalen)
    out.write(header + data)
