#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# palette `pal`. The image has dimensions `width` x
# `height`, and uses up to `n_colors` colors from `pal`.
def generate_gif(out: BinaryIO, rng: BinaryIO, pal: bytes, width: int = 1024, height: int = 768, n_colors: int = 256):
    out.write(b'GIF')                             # Magic header
    out.write(struct.pack('<HH', 6912, 0))       # Logical screen descriptor

    # Global color table (palette), with flags for sorting and size. This looks like this in memory:
    # [flags][size][palette]
    #...where flags is a single byte containing b00011111 for some reason?
    out.write(bytes([0x7F | len(pal)]))          # Sorting flag + length of palette
    out.write(pal)                              # Palette itself
    out.write(bytes([0]))                       # Padding byte

    # Image descriptor structure:
    # [left offset][top offset][width][height][flags]
    # Flags are a single byte, where the first bit set indicates
    # that the image should be interlaced. We don't support it here.
    out.write(struct.pack('<HHHHB', 0, 0, width, height, 0))

    # Write an "end" block to mark the end of the image data.
    out.write(b';')

    # Finally, write all the pixels themselves.
    # They're stored as a series of blocks starting with a
    # one-byte code indicating how many pixels follow until we
    # reach the next block:
    # 0 - use another pixel value (which follows) or transparent;
    # 1 - use the same pixel value from the previous frame;
    # 2,..., 255 - use the next pixel value multiple times.
    def get_next():
        return ord(rng.read(1)) % 256

    prev = None
    while True:
        count = 0
        while count == 0:
            count = get_next()

        if prev is not None:
            d = b''
            while count > 0:
                if count < 255:
                    d += struct.pack('BBB', *prev)
                else:
                    p = prev[0] ^ prev[1] ^ prev[2]
                    d += struct.pack('<BBBBB', *(p,) * 5)

                count -= 1

            out.write(struct.pack('B', 1))
            out.write(d)
        elif count!= 1:
            break

        prev = tuple(ord(rng.read(1)) & 0xFF for _ in range(3))
