#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# provided RNG. The size of the generated image is determined by
# the first four bytes read from the RNG (which are discarded).
# If `frame_clearance`, then each frame is preceded with 0x21FF, or an
# extended control block specifying that there should be no delay.


def gen(
    out: BinaryIO, rng: int, width=None, height=None, background=None, frame_clearance=False
):

    # Header and logical screen descriptor
    if not width:
        width = rng % 50 + 50
    if not height:
        height = rng % 50 + 50
    if not background:
        background = rng & 0xFF

    header = "GIF89a"
    hdrsz = len(header)
    version = b"\x00\x3B"
    bgcol = struct.pack("<H", background)
    aspect_ratio = b"\xC0\x00"
    scrn_dscr = struct.pack("<6LBB", 0, width, height, 0b11000000 | 7, bgcol[0], bgcol[1])

    fd = bytearray()
    fd.extend(struct.pack(">H", width))
    fd.extend(struct.pack(">H", height))
    fd.append(0b00100000)

    global_color_table_flag = True
    color_resolution = 4
    sort_method = 1
    gct = []
    palette = [bgcol] * 2**(color_resolution+1)
    for i in range(len(palette)):
        r = ((i >> 0) & 1) * 255
        g = ((i >> 1) & 1) * 255
        b = ((i >> 2) & 1) * 255

        if i == 0:
            r = ((r << 5) & 0xf8) | (((r >> 3) & 7) << 2)
            g = ((g << 5) & 0xf8) | (((g >> 3) & 7) << 2)
            b = ((b << 5) & 0xf8) | (((b >> 3) & 7) << 2)

            palette[i] = bytes([r, g, b])
            continue

        # Don't use the same color twice
        while palette[-1] == bytes([r, g, b]):
            r = (rng >> 8) % 256
            g = (rng >> 16) % 256
            b = (rng >> 24) % 256
        palette[-1] = bytes([r, g, b])

        if len(gct) >= 2**color_resolution:
            global_color_table_flag = False

        if global_color_table_flag:
            r = ((r << 5) & 0xf8) | (((r >> 3) & 7) << 2)
            g = ((g << 5) & 0xf8) | (((g >> 3) & 7) << 2)
            b = ((b << 5) & 0xf8) | (((b >> 3) & 7) << 2)

            gct += [r, g, b]
        else:
            gct = None

        fd.append((i << 1) ^ 2)

    lzw_min_code_size = max(width//height, height//width)+1 if width > height else 0
    lzw_min_code_size = min(max(lzw_min_code_size, 2), 12)
    lzw_min_code_size |= 0b10000000

    fd.append(lzw_min_code_size)

    out.write(b"".join([header, version, scrn_dscr]))

    if global_color_table_flag:
        out.write(bytes([color_resolution, sort_method]))
        out.write(struct.pack("<B", len(gct)//3))
        out.write(bytes(gct))

    out.write(fd)

    code_len = lzw_min_code_size+1
    clear_code = 1 << code_len-1
    eoi_code = clear_code+1

    def append(data):
        nonlocal code_len
        nonlocal pendingbits

        if isinstance(data, str):
            data = data.encode("latin-1")

        bits = list(map(lambda x: (x>>7)&1, data))+[0]
        data = [(x&0x7F)<<pendingbits for x in data]+[(1<<pendingbits)-1]*(-len(data)%code_len)
        pendingdata += sum([[data[j]|data[j+1]] for j in range(0, len(data), 2)], [])
        pendingbits += 1

        if pendingbits < 8:
            return

        flushbits()
        pendingdata = []
        pendingbits = 0

    def flushbits():
        nonlocal pendingdata
        nonlocal pendingbits

        while pendingbits >= 8:
            out.write(bytes([(1<<pendingbits)-1]+pendingdata[:code_len-1][::-1]))
            del pendingdata[:code_len]
            pendingbits -= 8

    pendingdata = []
    pendingbits = 0

    codes = {}
    current_code = -1
    prev_byte = ""

    # Build up LZW table
    for i in range(256):
        s = chr(i)
        codes[s] = [current_code, s]
        current_code += 1
    codes["", [-1]] = [-1]

    pendingbits = 0
    flushbits()
    out.write(struct.pack("<B", clear_code))

    # Generate frames
    while True:
        append("".join(["X"]*rng%4096))

        if frame_clearance:
            out.write(b"\x21\xff")

        # Write LZW tree
        for c in set(codes.values()):
            if c[1]:
                append(c[1])

        # Write image descriptor + graphic control extension
        out.write(struct.pack(">H", 0))  # Image left position
        out.write(struct.pack(">H", 0))  # Image top position
        out.write(struct.pack(">H", width))  # Image width
        out.write(struct.pack(">H", height))  # Image height
        out.write(b"\x00")  # Packed field: local color table flag, interlace, sort method
        out.write(b"\x00")  # Background color index
        out.write(aspect_ratio)  # Aspect ratio

        # Write graphics control extension
        out.write(b"\x21\xF9\x04")  # Extension introducer, graphics control label
        out.write(b"\x08")  # Block size
        out.write(b"\x00")  # Packed field: transparent color flag, user input flag
        out.write(b"\x00")  # Delay time
        out.write(b"\x00")  # Transparent color index
        out.write(b"\x00")  # Block terminator

        # LZW compressor
        fringe = ["", None]
        for c in pendingdata:
            fringe.insert(0, codes[chr(c)])
            while len(fringe) > 1 and type(fringe[0])!= list:
                codes[fringe[0][1]][0] = current_code

                append(fringe[0][1].encode("latin-1"))

                if frame_clearance:
                    out.write(b"\x21\xff")

                # Write LZW tree
                for cc in set(codes.keys()) - {""}:
                    if codes[cc] not in [fringe[0]]:
                        append(cc)

                codes.pop(fringe[0][1])
                fringe.pop(0)

        # Clear codes
        append(clear_code)

        if frame_clearance:
            out.write(b"\x21\xff")

        # Finishing EOI code
        append(eoi_code)

        break
