#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# provided RNG. The size of the generated image is determined by
# the first four bytes read from the RNG (which are discarded).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO) -> None:

    # Read the seed and discard it
    seed = bytearray()
    while len(seed) < 4:
        seed += rng.read(1024)[:4 - len(seed)]
    print("Seed:", repr(bytes(seed)))

    width = int.from_bytes(rng.read(1), "little") * 6 + 79
    height = int.from_bytes(rng.read(1), "little") * 6 + 59
    print("Resolution:", width, "x", height)

    screen = bytearray([
        0b11011010,     # Screen descriptor block
        8,              # Width in pixels minus one
        0,              # Height in scanlines minus one
        0b11100100,     # Flags: colour table included, no sort flag
        255])           # Background colour index
    for i in range(256):
        r = rng.read(1)[0] % 64
        g = rng.read(1)[0] % 64
        b = rng.read(1)[0] % 64

        screen.append((r << 5) | (g << 2) | ((b >> 3) & 0b111))

    out.write(struct.pack("<4sHHBBB",
        b"GIF89a",          # Header
        0,                  # Logical screen width
        height,             # Logical screen height
        0b10000000,         # Flags: global color table present
        0,                  # Colour resolution
        0,                  # GCT size
        ))
    if width > 0xFFFF:
        raise Exception("Too wide!")
    out.write(width.to_bytes(2, "little"))
    out.write(screen[len(struct.pack("<4sHHBBB")):])

    # Image header
    out.write(b"\x2c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")

    # Image data blocks; each has a prefix that determines its length
    block = b""
    n_blocks = 0
    while True:
        opcode = rng.read(1)[0]
        if not block or opcode == 0x3B:
            remaining = (n_blocks*128) - len(block)

            # Block size is at most 0xFFF8 plus 2
            size = min(remaining, 0xFFF8+2)
            assert abs(size-len(block)) >= 1
            
            # Determine whether to use an LZW codeword or literal frame data
            use_codeword = (
                False #rng.read(1)[0]%2 == 0      # 50% chance
                )

            if use_codeword:
                if size > 2:
                    block += rng.read(1)[0].to_bytes(1, "big")
                    size -= 1

                if size > 2:
                    block += rng.read(1)[0].to_bytes(1, "big")
                    size -= 1

                if size <= 2:
                    break

                if size > 0x102:
                    block += rng.read(size - 4)
                else:
                    code = rng.read(1)[0]
                    if size > 2:
                        code |= rng.read(1)[0] << 8

                    # Codewords cannot be larger than 11 bits
                    assert code < 0x800
                    assert code!= 0xFFFE
                    assert code!= 0xFFFF

                    block += code.to_bytes(2, "big")
            else:
                block += rng.read(min(size - len(block) - 2, 0xFFFD))
            n_blocks += 1

            out.write(struct.pack("<H", len(block)+2))

        elif opcode == 0x00 or \
             opcode == 0xFE or \
             opcode == 0xFF:
            continue
        
        else:
            try:
                # Frame data
                instruction = gif_instructions[opcode]
                
                # Get the number of frames before the current frame
                cur_frame = n_blocks // instruction["frames"]

                max_val = 2**instruction["nbits"]-1

                numvals = [cur_frame]
                val_bits = []
                for _ in range(instruction["nbits"]):
                    val = rng.read(1)[0] & instruction["mask"]
                    max_val -= val & instruction["maxval"]
                    
                    numvals.append(val & instruction["maxval"])
                    val >>= instruction["maxval"].bit_length()-instruction["nbits"]
                    val_bits.append(val)

                if max_val > 0:
                    # If there's unused bits, repeat them the entire time
                    if size > instruction["nbits"]:
                        block += instruction["unused"]
                else:
                    if size > len(instruction["prefix"]):
                        block += instruction["prefix"][len(block)-1:]
                size -= len(block)

                for numval, val_bit in zip(reversed(numvals[:-1]), reversed(val_bits)):
                    block += instruction["prefix"][numval][val_bit]
                    size -= 1
                    
                assert size >= 0
                
            except KeyError as e:
                raise Exception(hex(opcode))
            
    out.write(block)
        
    # End marker block
    out.write(b"\x3B\x00")


# Instruction set used by the GIF specification
gif_instructions = {
    0x2C: {    # Define Graphic Control Extension
        "frames": 1,
        "nbits":  3,
        "mask":   0b111,
        "maxval": 0b000,
        "unused": b'\xfe',
        },

    0x21: {    # Define Trailer Marker
        "frames": 1,
        "nbits":  3,
        "mask":   0b111,
        "maxval": 0b000,
        "unused": b'\xfe',
        },

    0x29: {    # Define Disposal Method
        "frames": 1,
        "nbits":  3,
        "mask":   0b111,
        "maxval": 0b000,
        "unused": b'',
        },

    0x3B: {    # End Of File
        "frames": 1,
        "nbits":  3,
        "mask":   0b111,
        "maxval": 0b000,
        "unused": b'',
        },

    0xF9: {    # Graphics Control Extension
        "frames": 1,
        "nbits":  3,
        "mask":   0b111,
        "maxval": 0b000,
        "unused": b'',
        },

    0xFE: {    # Application Extension
        "frames": 1,
        "nbits":  8,
        "mask":   0xFF,
        "maxval": 0xFF,
        "unused": b'',
        },

    0xFF: {    # Comment extension
        "frames": 1,
        "nbits":  8,
        "mask":   0xFF,
        "maxval": 0xFF,
        "unused": b'',
        }
}