#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    footer = b';\n'

    out.write(header + struct.pack('>H', len(data)) + data + footer)


# Main function which reads in an image and splits it recursively into chunks of size <= 256 bytes.
# If the input is longer than that we write the current chunk as-is (including its length prefix).
# Otherwise, we split up until there are no more chunks to be splitted.
# This requires two passes over the original image but simplifies our implementation quite a lot.
def split_image(in_path: str):
    with open(in_path, 'rb') as in_file, open('%s.splitted' % in_path, 'wb') as out_file:

        # Read all bytes at once to avoid reaching EOF while trying to read individual
        # chunks from disk for every pass through this loop.
        content = in_file.read()

        offset = 0
        while True:
            maxlen = min(
                end=len(content)-offset,
                # We don't want to have empty frames because they would never
                # get decoded properly, so make sure each frame contains at least one byte.
                 amount=max(1, 256 - ((len(content)-offset)%256)))

            if maxlen == 0 or offset >= len(content):
                break

            out_file.write(struct.pack('B' * maxlen, *content[offset : offset+maxlen]))
            offset += maxlen



# Merges previously generated images by concatenating them together again.
def merge_images(in_paths: list[str], out_path: str):
    with open(out_path, 'w+b') as out:
        for path in in_paths:
            with open(path, 'r+b') as img:
                buf = img.read()

                # Truncate output file to correct size first, then overwrite existing bytes.
                out.seek(0, 2)
                out.truncate(out.tell()+len(buf))
                out.seek(-len(buf), 2)
                out.write(buf)


# Encodes whole pixels into Netpbm format using the given number of colors per channel.
# Assumes width and height are power of 2.
# Does not perform any dithering.
def encode_ppm_netpbm(
        arr: list[list[int]], w: int, h: int, c: int, out: BinaryIO):

    def ppm_magic():
        return f"P{c}\n{w} {h}".encode("ascii")

    def netpbm_to_bytes():
        if   c == 4: ldtype = "BBBB"
        elif c == 3: ldtype = "BBB"
        elif c == 2: ldtype = "BB"
        elif c == 1: ldtype = "B"

        for y in range(h):
            for x in range(w//8):
                outx = []
                for i in range(8):
                    for j in range(c):
                        outx.append((arr[(y*8+i)*w+(x*8)+i][j])*256//c)
                yield struct.pack(">"+ldtype, *outx)


    out.write(ppm_magic())
    for byte in netpbm_to_bytes():
        if len(byte)!= 1:
            raise ValueError("Invalid PPM encoding!")
        if byte!= b'\x00':
            # Write short blocks with padding
            while len(out.getbuffer()) & 1:
                out.write(b'\x00')
            out.write(byte)
        else:
            # Write long blocks without padding
            out.write(byte)
