#!/usr/bin/python3

import sys
import os.path as path
"""
This script is used to generate a file with variable declarations for the constants defined in C headers using preprocessor directives.
It operates on all header files specified in the input, and parses out all #defines which are declared to take an expression or primitive constant.
The output of this tool can be directly included in any source file that needs access to such variables without having to include explicit type definitions for them.

Example usage: python gen_constants.py linux/include/uapi/linux/if_ether.h linux/include/uapi/linux/socket.h > uapi_constants.c && clang-format -i uapi_constants.c
"""


def check_file(filename):
    if not path.exists(filename) or not path.isfile(filename):
        print("File does not exist:", filename)
        exit(-1)


def parse_file(filename):
    check_file(filename)

    defines = []  # List of triples (name, value, type). Type can either be 'int' or'string'.
    typedefs = {}  # Map from typename -> value. This list is built while parsing, and then processed to create triples for each name->value pair where appropriate.
    aliases = {}  # Map from alias name to real name.

    f = open(filename)
    lines = iter(f.readlines())

    # Find all typedefs first so we later know what values correspond to types instead of names.
    line = next(lines)
    while "struct __sk_buff" in line:
        # We only care about struct __sk_buff since it contains many macros that resolve into other structs.
        line = next(lines)
        while True:
            m = re.match('^typedef (\w+) \w+\[\];', line)
            if m is None:
                break

            # Record this typedef for future use when we encounter expressions.
            t = m[1]
            v = ""
            if t in typedefs:
                assert typedefs[t] == v  # Make sure no typo made here!
            else:
                typedefs[t] = v

        try:
            line = next(lines)
        except StopIteration:
            break

    def get_type(v):
        # There should never be nested types like `struct foo { int x; }`.
        # However, there may be an enum like `enum bar { A };`, whose raw string value would simply be `A` and has no type information.
        # So we special case these strings by looking at their context.
        if v!= "":
            s = v.strip()
            if len(s) >= 2 and ((s[0] == '"' and s[-1] == '"') or (s[0] == "'" and s[-1] == "'")):
                return "string"

        if v == "":
            # If empty string, assume int (most common case).
            return "int"

        # Look for a known type definition. These are the ones already added to the map earlier.
        for k, val in typedefs.items():
            if v == val:
                return k

        # If none found
        raise Exception("Unknown type:" + v)


    # Parse all #define lines now.
    line = next(lines)
    while "#endif" not in line:
        m = re.match('#\s*define\s+([^\s(]+)\s*(.*)', line)
        if m is None:
            line = next(lines)
            continue

        name = m[1].replace("\n", "")
        if "(" in name:
            # Ignore function macros because they require more complicated parsing logic.
            line = next(lines)
            continue

        val = m[2].replace("\n", "").strip()
        # Now look through the value again to see whether its an expression where a new type was introduced.
        for k, v in typedefs.items():
            if v in val:
                val = val.replace(v, k)


        # Now that we have parsed everything out of the value, figure out what kind it is, e.g., primitive constant vs. expression.
        typ = get_type(val)
        entry = [name, val, typ]
        defines.append(entry)

        line = next(lines)

    f.close()

    # For each constant which uses another one, replace the latter's entry entirely with the former's.
    for i in range(len(defines)):
        d = defines[i]
        name = d[0]
        value = d[1]
        typ = d[2]

        # For primitives, nothing to do.
        if typ!= "string":
            defines[i] = d
            continue

        # Non-primitives must go back and modify previous entries.
        parts = value.split('|')
        new_parts = []
        for part in parts:
            part = part.lstrip().rstrip()

            # Handle cases where the struct is unnamed but only used once, eg.'struct foo foo;'
            n = part.count('{')
            nn = part.count('}')
            if n < nn:
                part += '}'* (nn - n)


            # First extract out the part before the struct. At least one character must remain since the struct itself cannot be empty.
            prefix, _, suffix = part.rpartition("{")
            if part == "{":
                # Could happen due to an odd number of braces after the prefix, or something completely different.
                # This is most likely some malformed macro expansion, so just skip it rather than trying to parse it.
                print("Skipping bad macro:", part)
                continue

            if len(prefix) <= 1:
                print("Skipping short entry:", part)
                continue

            # Then find the corresponding define declaration for the given struct.
            found = False
            for j in range(len(defines)):
                d = defines[j]

                # The define might contain non-closed braces, which will throw off the matching later on.
                # So let's change them to closed ones temporarily to simplify things.
                replace_mapping = {'{': '{ ', '}' :'}'}
                p = "".join([replace_mapping[c] if c in replace_mapping else c for c in d[1]])

                if d[2] == "string":
                    match = p.find(prefix)
                    if match!= -1:
                        # We found a suitable match. Replace the original entry with the modified form.
                        del defines[i]

                        # Add the new entry with the proper type for the existing one.
                        new_d = [d[0], d[1][:match]] + ["struct"] + [d[1][match:]]
                        defines.insert(i, new_d)

                        # Also add the alias entry.
                        if d[0] not in aliases:
                            aliases[d[0]] = name

                        # There should be no need to check further matches because our loop breaks early once a match is found.
                        found = True
                        break

                elif d[2] == "struct":
                    if d[0] == prefix:
                        # Found the struct declaration! Let's proceed normally.
                        del defines[i]

                        # Check if there were parts following the structure declaration, and insert those after the newly created entry.
                        if len(suffix) > 0:
                            new_d = copy.deepcopy(d)
                            new_d[1] = suffix

                            if new_d[0] not in aliases:
                                aliases[new_d[0]] = name

                            defines.insert(i, new_d)


                        # Followup with adding the actual alias.
                        defines.insert(i, [name, "", typ])
                        found = True
                        break

            if not found:
                print("Could not resolve reference to", prefix)

        for j in reversed(range(len(defines))):
            d = defines[j]
            if d[0] in aliases:
                del defines[j]




    if len(defines) == 0:
        return

    max_length = max([len(x[0]) for x in defines])
    formatstr = "{{:<{}s}} {{}}\n".format(max_length)

    print("#ifdef __cplusplus")
    print("extern \"C\" {")
    print("#endif /* __cplusplus */")
    print("")

    print("#ifndef UAPI_CONSTANTS")
    print("#define UAPI_CONSTANTS")
    print("")
    print("// This file is generated automatically.")
    print("")
    print("#include <linux/types.h>")
    print("#include <stdbool.h>")
    print("#include <stdint.h>\n")
    print("\n#define STRUCT_DECL(x) _STRUCT_RESERVED_NAME_##x")
    print("_STRUCT_RESERVED_NAME_{}();\n".format(re.sub("[/.]", "_", filename)))
    print("#undef STRUCT_DECL")
    print("")
    print("#endif // UAPI_CONSTANTS");
    print("")

    for entry in defines:
        name = entry[0]
        assert len(name) > 0

        value = entry[1]
        typ = entry[2]

        if typ == "string":
            sizestr = "sizeof({}) - 1".format(name)
            if value == "":
                # Just make it an empty string by default. The compiler should optimize away the unused buffer anyway.
                print(formatstr.format(name, "{}(\"\")".format(sizestr)))
            else:
                s = "\"{}\"".format(value)
                print(formatstr.format(name, "{}({}