#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # 768 comes from the formula in https://stackoverflow.com/questions/14550622/gif-transparency-handling
    numcol = int.from_bytes(rng.read(1), 'big') % (768 + 1) - 1
    colpal = [b'\x00\x00\x00'] * ((numcol + 1) * 3)

    lsz = 2 << len(colpal)
    while True:
        lsz -= 1

        if lsz == 0:
            break

        colpal += rng.read(lsz)


    # Do not use LZW compression for simplicity and because we're too lazy to implement it fully
    header += bytes([0xf9]) + \
              struct.pack('B', 4) + \
              struct.pack('<H', 0xFFFF) + \
              struct.pack('<BBBBB', *(2 ** x - 1 for x in range(7)))

    out.write(header)

    width = height = 1 + int.from_bytes(rng.read(1), 'big') % 40
    out.write(struct.pack('<HHHHBCC',
                          width, height,
                          0,    # No color table or interlacing
                          0, 2, # No local colors, sorted color table
                          0    # Use plain RGB
                        ))

    loop = False
    delay = 100 * rng.read(1)[0]     # Delay between frames
    discr = 1                       # Discretization factor of transparency palette
    imglen = 1                      # Length of each frame's image data section
    clrsz = 1                       # Number of bits per pixel in this frame

    hblen = 1                       # Shortest possible Huffman code length bitstream
    while True:
        val = rng.read(1)             # Read one byte
        if val[0] & 0x80!= 0:        # If first bit is set...
            continue                  #...then we must have run out of headers

        ypos = val[0] & 0x07          # Extract field size and positioning information about y axis
        xpos = (val[0] >> 3) & 0x07
        lct = bool(val[0] & 0x20)     # Is there a local color table?

        clrsz += (clrsz >= 2)         # Increase color depth by 1 on every non-raw frame

        # The only other four values represent the minimum color depths that are required; as such they should never happen
        assert clrsz < 8              # This assertion doesn't really give any useful information since clrsz can be overwritten later on

        haslocclrtab = bool(val[0] & 0x80)       # Does the frame have its own local color table?
        hasinterlace = bool(val[0] & 0x40)      # Does the frame contain an interlaced version?

        lsz = 2 << len(colpal)
        while True:
            lsz -= 1

            if lsz == 0:
                break

            colpal += rng.read(lsz)


        blocktype = None
        while True:
            blocktype = rng.read(1)   # Read another byte
            if blocktype[0]!= 0x2c:
                break                  # We did not read the start of an image descriptor

        imglen = 1                   # Reset frame image length
        while True:                 # Continue reading until an end marker is found
            val = rng.read(1)
            if val[0] & 0x80!= 0:
                continue                # Check for end marker and skip it if present

            hasimgdata = bool(val[0] & 0x40)       # Has image data?
            haslocclrtab = bool(val[0] & 0x20)    # Has local color table? Interestingly enough, this flag does not appear in the previous line

            lsz = 2 << len(colpal)
            while True:
                lsz -= 1

                if lsz == 0:
                    break

                colpal += rng.read(lsz)


            clsz = min(clrsz, 2 ** discr)               # Number of bits used to encode a single color index

            # If size of color palette exceeds limit then switch to raw mode
            if len(colpal) > 255 * clsz:
                ncol = int((255 * clsz / 3) // 8)
                colpal = [0] * (ncol * 3)
                clrsz = 1                                 # Switch back to raw mode
                blocktype = None                         # Break out of the inner while loop
                continue                                  # Jump down to the outer while loop to write current frame to output stream

            if hasimgdata:                                # We reached the beginning of the image data section
                if blocktype[0]!= 0x2c:
                    raise Exception()                     # Image data chunk without preceding image descriptor

                imglen += 1                               # And we just add another byte to our running total


                while True:
                    val = rng.read(1)                    # Read next byte

                    if val[0] & 0x80!= 0:                # Check whether this is an end marker
                        break                              # End markers always occur at the very end of an image data sequence

                    imglen += 1                           # Count length of actual image data
                    hasimgdata = bool(val[0] & 0x40)      # More chunks coming?
                    haslocclrtab = bool(val[0] & 0x20)

                    lsz = 2 << len(colpal)
                    while True:                          # Keep reading longer color tables until finding an end marker
                        lsz -= 1

                        if lsz == 0:
                            break

                        colpal += rng.read(lsz)


                    clsz = min(clrsz, 2 ** discr)           # Number of bits used to encode a single color index

                    if hasimgdata:                        # We're still looking for more image data
                        continue

                    if blocktype[0] == 0x3b:              # We reached the end of compressed image data
                        if haslocclrtab:
                            clsz = 1                         # Switched back to raw mode

                            # We expect no more blocks with this particular combination of flags here but let's check anyway
                            while True:
                                blocktype = rng.read(1)

                                if blocktype[0] == 0x3b:
                                    return

                                if blocktype[0]!= 0x2c:
                                    raise Exception()


                        else:
                            lsz = 2 << len(colpal)
                            while True:
                                lsz -= 1

                                if lsz == 0:
                                    break

                                colpal += rng.read(lsz)


                            clsz = min(clrsz, 2 ** discr)   # Number of bits used to encode a single color index



                        if blocktype[0]!= 0x3b:
                            raise Exception()

                        imglen += 1                           # Add two more bytes to account for the two end markers
                        out.write(struct.pack('<BBBB',
                                              0x21, 0xff, 0x0b,
                                              numcol
                                            )
                                  )

                        lsz = 1                               # Continue counting the length of the data
                        while True:
                            lsz += 1

                            if lsz == 0:
                                break

                            out.write(rng.read(1))

                        out.write(bytes([0]))

                        continue                            # Start writing rest of the remaining image data

                    elif blocktype[0] == 0x2c:            # New image descriptor starts right after the last one!
                        raise Exception()

                if (xpos + 1) * width <= height:
                    pass

                continue                                    # Skip all following checks for now so that we can write current frame to output stream

            elif blocktype[0]!= 0x21:
                raise Exception()                             # After an empty image data block we should find either an application extension or an image descriptor

            # Let's parse the application extension
            appidsize = 1                                   # Size of application identifier
            appidhuffsize = 1                                # Size of encoded application identifier
            while True:
                val = rng.read(1)                            # Read another byte

                if val[0] & 0xc0!= 0:                       # If first three bits are set...
                    raise Exception()                         #...then this was not meant to be part of application identifier

                appidhuffsize += (appidhuffsize!= 1)        # Increment counter on every increment in huffman encoding state
                appidsize += (appidsize!= 1)                # Increment counter on every increment in raw encoding state

                if val[0] & 1!= 0:                           # If second bit is unset...
                    continue                                  #...then this is not the last byte of the application id

                break                                          # Otherwise, this was the last byte of the application id

            if appidhuffsize!= 1:
                raise Exception()                             # If application identifier is encoded using huffman codes, their lengths must be equal to 1

            if appidsize == 11 and rng.read(6).decode().lower() == "netscape":
                # We've found the Netscape Loop Extension