
from typing import BinaryIO

# Generates a random GIF file into `out` using the given generators.
def generate(generators: list[Generator], out: BinaryIO):
    # Write "GIF89a" at the start of the output file.
    out.write("GIF89a".encode())

    color_map = []
    for i in range(256):
        r, g, b = 0, 0, 0

        for gen in generators:
            r += int((i & (1 << 7)) / 128 * next(gen) * 31 + 0x40)
            g += int(((i >> 3) & (1 << 7)) / 128 * next(gen) * 31 + 0x40)
            b += int((((i >> 3) >> 3) & (1 << 7)) / 128 * next(gen) * 31 + 0x40)

            if len(color_map) < 64 and all([r <= 255, g <= 255, b <= 255]):
                color_map.append((r, g, b))

    # The third byte is the number of colors in use.
    out.write(bytes([len(color_map)]))

    # Then we write each entry in the color map to the output file as bytes.
    for rgb in color_map:
        out.write(bytes([rgb[0] - 0x40] + [rgb[1] - 0x40] + [rgb[2] - 0x40]))

    # Then we can calculate how long the image data will be by adding up the lengths of every frame's data.
    length = sum([generator.__len__() for generator in generators])
    
    # Finally, we write the minimum header information necessary to make it a valid GIF file.
    # Details about this header are listed here: https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    # 0b11111111 : 255 : Size in bits of the entire graphic control block
    # 0b10000001 : 1   : Block size (64)
    # 0          : 0   : Reserved
    # 0b1        : 1   : LZW Compression method
    # 1          : 1   : True color image 
    # 0b11       : 3   : Color resolution (0b11 = 72 DPI)
    # 0b11       : 3   : Sort flag
    # 0b1111     : 0xF : Bits per pixel minus one (minus 1 because 1 bit means no compression)
    # 0          : 0   : Background colour index
    # 0          : 0   : Pixel aspect ratio
    out.write(bytes([0xFF, 0x21, 0xF9, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]))

    # We then encode our length value into two bytes so that it fits inside our data block.
    buf = length.to_bytes(2, "little")

    # The first byte is the length of the data block, including its length field itself.
    out.write(bytes([buf[0]]))

    # And the second byte contains whatever was left over from the original number after subtracting 256.
    out.write(bytes([buf[1]]))

    # Now finally we can actually write our actual image data!
    for gen in generators:
        out.write(bytearray(gen))