#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(rng:BinaryIO, out:BinaryIO):
    # Header (6 bytes)
    header = b'GIF89a' + rng.read(7)

    # Logical Screen Descriptor (7 bytes)
    lsd = bytearray()
    lsd += rng.read(1)  # Width in bits
    lsd += rng.read(1)  # Height in pixels
    lsd += b'\x07'       # Packed field (global color table present and sorted)

    # Global Color Table
    gct_size = int.from_bytes(rng.read(1), 'big') & 0b00011111    # Size of global color table
    gct = [rng.read(3) for _ in range(gct_size)]                # Load colors from /dev/urandom
    lsd[4] |= len(gct).to_bytes(1, 'big')[0]                     # Set GCT size to actual size
    lsd += b''.join(gct)                                        # Add them all at once

    # Image Descriptor (10 bytes)
    idt = bytearray()
    idt += b'\x2c'                                           # Image separator
    idt += b'\x00\x00'                                       # Left position
    idt += b'\x00\x00'                                       # Top poistion
    idt += b'\x01\x00'                                       # Width in pixels
    idt += b'\x01\x00'                                       # Height in pixels
    idt += b'\x00'                                           # Packed field (no local color table)

    # Graphics Control Extension (8 bytes)
    gce = bytearray()
    gce += b'\x21'                                             # Introducer
    gce += b'\xf9'                                             # Graphic control label
    gce += b'\x04'                                             # Block length
    gce += b''                                                 # Packed field (unused)
    gce += b'\x00'                                             # Delay time
    gce += b'\xff'                                             # Transparent color index
    gce += b'\xfe'                                             # Terminator

    # Application Extension (11 bytes)
    ae = bytearray()
    ae += b'\x21'                                               # Introducer
    ae += b'\xff'                                               # App extension label
    ae += b'\x0b'                                               # Block length
    ae += b'NETSCAPE2.0'                                        # Application ID
    ae += b'\x03'                                               # Sub-block size
    ae += b'\x01'                                               # Loop sub-block ID
    ae += b'\x00\x00'                                           # Loop count

    # Comment Extension (11 bytes)
    ce = bytearray()
    ce += b'\x21'                                               # Introducer
    ce += b'\xfe'                                               # Comment label
    ce += b'\x0b'                                               # Block length
    ce += b'Generated by urandom.py'                            # Comment string

    # Image data block (variable size)
    idatablock = rng.read(int.from_bytes(lsd[:2], 'big'))      # Read the amount of image data as per LSD

    # Footer (3 bytes)
    footer = b';'                                              # Trailer

    out.write(header)                                          # Write everything in one go
    out.write(lsd)
    out.write(idt)
    out.write(gce)
    out.write(ae)
    out.write(ce)
    out.write(idatablock)
    out.write(footer)


if