#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
        
        with open('/dev/urandom', mode='rb') as rng,\
             open(self.output,    mode='wb') as out:

            return generate_random_gif(
                    rng, out
                )


def generate_random_gif(rng:BinaryIO, out:BinaryIO)->None:
    
    out.write('GIF89a'.encode())   # GIF Header
    screen = b'\x2c\x00\x00\x00'  # Screen Descriptor
    palette = b'\x00'*768        # Global Color Table
    datastream = b''             # Image Data

    while True:
        datastream += _generate_image(rng, out)
    
        # If next byte isn't zero the image was finished
        if int.from_bytes(rng.read(1), "big")!= 0x00:
            break
        
    end = b'\x3b' # GIF Trailer
    out.write(screen + palette + datastream + end)
    
        
def _get_color() -> bytes:
    """Generates a random RGB triple for the color table"""
    
    r = str(hex(int.from_bytes(os.urandom(1),'big')))[2:]
    g = str(hex(int.from_bytes(os.urandom(1),'big')))[2:]
    b = str(hex(int.from_bytes(os.urandom(1),'big')))[2:]
    
    while len(r)<2:
        r = "0" + r
    while len(g)<2:
        g = "0" + g
    while len(b)<2:
        b = "0" + b

    return ("#" + r + g + b).encode()


def _generate_image(rng:BinaryIO, out:BinaryIO)\
                  ->bytes: 
    
    header = b"\x21\xf9\x04"    # Graphics Control Extension
    # Disposal Method => None (No disposal specified)
    dm = rng.read(1)[0] % 5  
    header += bytes([(dm << 2) | 0])  
    # User Input Flag => False 
    ui = rng.read(1)[0] & 1
    header += bytes([ui << 1])     
    # Delay Time => 0 milliseconds 
    dt = [0, 0]
    dt[0] = rng.read(1)[0]
    dt[1] = rng.read(1)[0]
    header += bytes(dt)            # delay time (2 bytes)

    # Transparent Color Index => Not Specified
    tci = [0, 0]
    tci[0] = rng.read(1)[0]       # index itself
    tci[1] = rng.read(1)[0]       # padding byte
    header += bytes(tci)           # transparent color index (2 bytes)

    # Image Left Position
    ilp = rng.read(1)[0]
    header += bytes([ilp])    
    # Image Top Position
    itp = rng.read(1)[0]
    header += bytes([itp])        
    # Image Width
    iw = rng.read(1)[0]
    header += bytes([iw])
    # Image Height
    ih = rng.read(1)[0]
    header += bytes([ih])  

    # Local Color Table? Yes!
    lct = [0, 0]
    lct[0] = rng.read(1)[0]      # number of colors in table
    lct[1] = rng.read(1)[0]      # maximum number of colors
    header += bytes(lct)          # local color table info (2 bytes)

    clrs = []
    for i in range(0, min(2**int(lct[0]), 256)): 
        clrs.append(_get_color())

    out.write(struct.pack(str(len(clrs))+'s', bytes().join(clrs)))
    
    # LZW Minimum Code Size => 2 bits per pixel
    mcs = rng.read(1)[0] % 8
    header += bytes([mcs]) 

    # Image Data
    while len(header)!= 12: 
        header += bytes([rng.read(1)[0]])

    # Block Size (Data Sub-Block within Data Stream)
    bs = rng.read(1)[0]         # block size is an unsigned char
    subblock = rng.read(bs - 1)  # discard sub-block marker
    header += bytes([bs])        # add block size back again
    data = rng.read(bs)          # get actual image data
    out.write(header + data)     # write it all together
    
