#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    blocksize, padding = 256 * (data[1] & 0x07) + 1, [b'\x00']
#    print('Data length:', datalen, 'Block size:', blocksize)

    # Write the header to the output file
    out.write(header)
    
    # Write the logical screen descriptor
    out.write(struct.pack('<HHBBB',
        400,
        300,
        8,
        0,
        0))

    # Use LZW compression for all frames
    lzwmincodelen = 8
    out.write(struct.pack('BB', 
        0xf9, 
        lzwmincodelen))

    # Start of Frame
    out.write(b'\x21\xff\x0bNETSCAPE2.0')
    out.write(b'\x03')
    out.write((2**16 - 1).to_bytes(2, byteorder='big'))
    out.write(b'\x00') # End of sub-block
    delaytime = int.from_bytes(
            rng.read(2), byteorder='big')
    out.write(delaytime.to_bytes(2, byteorder='big'))
    out.write(b'\x01') # Transparent colour index
    out.write(b'\x00') # Block terminator

    # Image Descriptor
    out.write(b'\x2c')
    out.write(struct.pack('<HH',
        0,
        0))
    out.write(struct.pack('<HH',
        400,
        300))
    out.write(b'\x00') # No local colour table
    out.write(b'\x00') # Interlace flag is false
    out.write(chr(lzwmincodelen)) # Minimum code size

    # TODO: Actually do LZW encoding here

    while len(data):
        copycount = min(blocksize, len(data))
        out.write(copycount.to_bytes(1, byteorder='big'))

        if copycount >= blocksize and not len(padding):
            raise ValueError("Too much data!")
        
        out.write(data[:copycount])
        data = data[copycount:]

        padlen = max(0, blocksize - copycount)
        out.write(padding[0] * padlen)
        data += padding[0] * padlen
                
    out.write(b'\x3b')