#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
        rng.read(4), 'little', signed=False
    )

    globalcolorcnt = int.from_bytes(
        rng.read(1), 'little', signed=False
    )
    backgroundidx = int.from_bytes(
        rng.read(1), 'little', signed=False
    )
    aspectratiocode = int.from_bytes(
        rng.read(1), 'little', signed=False
    )
    aspectratioratio = {
        0: b'\x00\x00',  # No units
        1: b'\x55\xc5',   # Pixels per centimeter (2.5)
        2: b'\xaa\x6a',   # Pixels per inch (72)
        3: b'\x00\x00'    # Invalid unit specifier; should raise
    }[aspectratiocode]
    
    imgdescriptorstart = out.tell()
    framewidth = int.from_bytes(
        rng.read(2), 'little', signed=False
    )
    frameheight = int.from_bytes(
        rng.read(2), 'little', signed=False
    )
    bitdepth = int.from_bytes(
        rng.read(1), 'little', signed=False
    )
    colorrescode = int.from_bytes(
        rng.read(3), 'little', signed=False
    )
    colorsused = int.from_bytes(
        rng.read(3), 'little', signed=False
    )

    data = []
    for y in range(frameheight):
        line = []
        for x in range(framewidth):
            paletteidx = int.from_bytes(
                rng.read(1), 'little', signed=False
            )

            r = ord(rng.read(1)) % 256
            g = ord(rng.read(1)) % 256
            b = ord(rng.read(1)) % 256
            line += [r, g, b]

        data += line


    print("Writing data")
    datalen = write_datasection(datalen, data, out)
   
    print("Writing header")
    hdrlen = write_header(
        header, datalen, globalcolorcnt, 
        backgroundidx, aspectratioratio,
        imgdescriptorstart, framewidth, frameheight,
        bitdepth, colorrescode, colorsused, out
    )

    return hdrlen + hdrlen + datalen



def write_header(
        idstr: bytes, datalen: int, gccount: int, bgidx: int, ratio: bytes,
        imgdescstartpos: int, width: int, height: int, depth: int, colres: int,
        cused: int, fd: BinaryIO
    ):
    hdrlength = len(idstr) + \
            4 + 1 + 1 + 1 + 3 + \
            3 + 1 + 3 + \
            4 + 2 + 2 + 1 + 3 + 3 + \
            1 + 24
    assert hdrlength == 13

    fd.seek(hdrlength-13)
    fd.write(struct.pack('<ccBbbBBBHBbbbBBH',
        *idstr,
        datalen&0xff, (datalen>>8)&0xff, (datalen>>16)&0xff,(datalen>>24)&0xff,
        gccount,
        bgidx,
        *ratio,
        *(b'\xf0' * 24)
    ))

    fd.seek(imgdescstartpos+4)
    fd.write(struct.pack('<HHHHBBBBBB',
        width&0xffff, (width>>16)&0xffff,
        height&0xffff, (height>>16)&0xffff,
        depth,
        colres,
        cused,
        0, 0,
        0,
        *([ord(':')] * 24)
    ))

    return hdrlength


def write_datasection(
        datalen: int, data: list, fd: BinaryIO
    ):
    datalen = 0
    for datum in data:
        datalen += len(datum)
    fd.write(struct.pack('>L', datalen))

    fd.write(b'\x2c')
    fd.write(struct.pack('>LHHBB', 
        0, 
        0, 
        len(data[0]), 
        len(data)%256, 
        0, 
        0
    ))

    for row in data:
        fd.write(b'\x2c')
        fd.write(struct.pack('>LHHBB', 
            0,
            0, 
            len(row), 
            len(row)%256, 
            0, 
            0
        ))
        for pixel in row:
            fd.write(struct.pack('BBB', 
                *map(lambda x: max(min(255, x), 0), 
                    pixel)))

    
    fd.write(struct.pack('>B', 0x3b))

    return datalen + 4


# Pads up to blocksize with zeros from /dev/zero and writes it
def pad_and_write(
        rng: BinaryIO, blocksize: int, out: BinaryIO
    ):
    padding = []
    with open('/dev/zero', 'rb') as inf:
        padding = inf.read(-(-blocksize // len(inf.read())))
    
    while byte := rng.read(blocksize):
        if blocksize - len(byte) > 0:
            # Pad short blocks with zeroes
            out.write(byte + padding[0])
        else:
            out.write(byte[:blocksize])

