#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
        
        with open('/dev/urandom', mode='rb') as rng,\
             open(self.output,    mode='wb') as out:
            
            # Generate Animated GIF Header and Global Color Table
            gen_gif_header(out, rng)

            while True:
                try:
                    gen_img_subblock(out, rng)
                except EOFError:
                    break

        print('Random GIF created at "{}"'.format(os.path.realpath(self.output)))


def gen_img_subblock(out: BinaryIO, rng: BinaryIO):
    
    codesize = rng.read(1)[0]    # read LZW minimum code size
    assert len(codesize) == 1
    # write sub-block marker followed by LZW min code size
    out.write(bytes([0x2C])) 
    out.write(codesize)
    bitwidth = 1 << codesize[0]   # bit width of current lzw table entry
    rle = False                   # default to do run length encoding instead of LZW

    # Terminate if we've reached the end of the file or found a loop
    while True:
        token = rng.read(1)[0]    # read next byte from input stream
        if token > bitwidth:      # if we find a non RLE matching token...
            bs = rng.read(token % 256) 
            out.write(bs)          #... write its corresponding value directly
        elif token < bitwidth:     # otherwise if we find a token that matches previous one...
            out.write(bytes([token])) #... just write another instance of the token
        else:                      # if we reach the end of the table itself..
            # Write Disposal Method Indicator
            dmi = rng.read(1)[0] & 0x1F
            pce = dmi & 0x01
            pev = bool((dmi >> 1) & 0x1)
            cev = bool((dmi >> 2) & 0x1)
            avi = ((dmi >> 3) & 0x7) + 1
            out.write(bytes([dmi]))
            # Write Image Left Position
            ilp = int.to_bytes(int.from_bytes(rng.read(2), 'little'), 2, 'big')
            out.write(ilp)
            # Write Image Top Position
            itp = int.to_bytes(int.from_bytes(rng.read(2), 'little'), 2, 'big')
            out.write(itp)
            # Write Width
            wdt = int.to_bytes(int.from_bytes(rng.read(2), 'little'), 2, 'big')
            out.write(wdt)
            # Write Height
            hgt = int.to_bytes(int.from_bytes(rng.read(2), 'little'), 2, 'big')
            out.write(hgt)
            # Write Flags
            flg = rng.read(1)[0]
            # Set Transparency Flag according to generated random number
            if rng.read(1)[0] & 0x1!= 0:
                flg |= 0x01
            # Force Empty Background color flag depending on transparency flag state
            if flg & 0x1:
                flg |= 0xFE
            out.write(bytes([flg]))
            # Write Animation Time
            atm = int.to_bytes(int.from_bytes(rng.read(1), 'little'), 1, 'big')
            out.write(atm)
            
            # If animation time exceeds 255s set it to maximum allowed value
            if atm[0] >= 254:
                atm = bytes([254])
                out.seek(-1, os.SEEK_CUR)       # go back to overwrite flag
                out.write(atm)                  # and write new value
                
            # Write Transparent Color Index if selected above
            if flg & 0x1:
                tci = int.to_bytes(int.from_bytes(rng.read(1), 'little'), 1, 'big')
                out.write(tci)
            # Write Control Extension Block if chosen by a generated random number
            # Check only once whether this is the last frame
            if rng.read(1)[0] & 0x1!= 0:
                pass
            # or start writing Graphic Control Extension Block
            else:
                gcf = bytes()                 # will hold global control flags
                gcf += bytes([0xF9])           # extension block label
                gcf += bytes([4])              # block size of 4 bytes
                
                gcf += bytes([(rng.read(1)[0]) | 0x80])  # set disposal method indicator
                gcf += bytes([rng.read(1)[0]])               # set user input flag
                gcf += bytes([0])                            # transparent color index
                gcf += bytes([rng.read(1)[0]])               # set delay time
                
                out.write(gcf)                          # write graphic control extension block

                # This GIF always uses the entire screen; thus we can ignore scroll position info
                spn = int.to_bytes(int.from_bytes(rng.read(2), 'little'), 2, 'big')
                out.write(spn)                         # skip 2 byte GIF application field
                out.write(bytes([1]))                  # write one more zero to terminate the GCE block
            
            # Now comes the tricky part: repeat pattern until we find a match
            # or hit the GIF's terminator flag
            end_of_frame = False
            while not end_of_frame:
                # Read next encoded token
                token = rng.read(1)[0]
                # Check for RLE encoding termination
                if token == 0:
                    # If so, mark end of frame and bail out immediately
                    end_of_frame = True
                elif token <= bitwidth:
                    # Otherwise check whether this is a match against previously written tokens
                    out.write(bytes([token])) 
                else:                  
                    # If no match was found, check whether we have to switch to RLE encoding
                    if rle:
                        # We already did this, so there must be a coding error here
                        raise RuntimeError('RLE Encoding failed.')
                    # Try switching to RLE encoding now
                    rle = True
                    rlt = token // bitwidth
                    rlp = [token % bitwidth]
                    
                    # Fill first repetition pattern
                    for i in range(rlt):
                        rlp.append(rng.read(1)[0])

                    # Write Repetition Pattern
                    bs = bytes(rlp)
                    assert len(bs) == 2
                    out.write(bs)

                    # Write End Of Image marker if necessary
                    if rng.read(1)[0] & 0x1!= 0:
                        out.write(bytes([0x2c]))
                        out.write(codesize)


def gen_gif_header(out: BinaryIO, rng: BinaryIO):
   
    # First four bytes are file signature "GIF"
    header = bytes([0x47, 0x49, 0x46, 0x38])
    # Version string follows with three digits and letter "a"
    header += b'GIF89a'            # File signature and version
    # Color Table size
    ctsz = rng.read(1)[0] * 3
    for _ in range(ctsz):
        header += bytes([rng.read(1)[0]])

    # Block Size (Data Sub-Block within Data Stream)
    bs = rng.read(1)[0]         # block size is an unsigned char
    subblock = rng.read(bs - 1)  # discard sub-block marker
    header += bytes([bs])        # add block size back again
    data = rng.read(bs)          # get actual image data
    out.write(header + data)     # write it all together
    
