#!/usr/bin/env python3

import struct

from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # 4 bytes of magic and version number.
    out.write(header)

    # Screen descriptor block (SDB).
    width, height = 100, 50
    # 2 bytes for screen color resolution, and then a bit field.
    packed_sdb = struct.pack('<HB6B', width, height, 7, 128, 0, 0)
    # First two bytes are the color table size with it being zero. It seems to be ignored by browsers anyway.
    out.write(packed_sdb)

    # Global color table is optional but recommended in GIF format specification so we will add one here.
    global_color_table = [
        0x00, 0x00, 0xFF,
        0xFF, 0x00, 0x00
    ]
    # Byte signifying how many colors there are.
    local_color_table_size = len(global_color_table) // 3
    # Then comes the first byte of data specifying a local color table.
    out.write((local_color_table_size + 1).to_bytes(1, 'little'))
    # Then comes all the RGB colors one after another. They will always have their red value followed by green
    # and blue values.
    out.write(bytearray(global_color_table))

    frames_count = rng.read(1)[0] % 100
    print('Writing {} frames.'.format(frames_count))

    delay = 25
    disposal = 1

    for i in range(frames_count):

        frame_control_block = [
            # Image separator
            0x2C,

            # Left position
            0x00, 0x00,

            # Top position
            0x00, 0x00,

            # Width
            0xFF & (width >> 8),
            0xFF & width,

            # Height
            0xFF & (height >> 8),
            0xFF & height,

            # Frame delay
            0xFF & (delay >> 8),
            0xFF & delay,

            # Disposal method
            0b01 << 2 | disposal,

            # Transparent color index
            0x00,

            # Include local color table flag.
            0,

            0x00, 0x00
        ]
        out.write(bytes(frame_control_block))

        # LZW minimum code size used to compress each line. This number should be between 2 and 12 inclusive.
        # However almost every encoder uses 2 as the default value and does not specify this parameter at all.
        # See https://stackoverflow.com/questions/43596664/what-is-the-minimum-lzw-code-size-used-by-a-gif for more details.
        lzw_min_code_size = 2
        out.write(struct.pack('<B', lzw_min_code_size))

        print('Compressing...')

        # Compress the image into the buffer until its length is less than or equal to ((local_color_table_size**2) - 1) * 3.
        compressed_image, _ = encode_lzwg_via_huffman(rng, out, i, width, height, local_color_table_size)

        print('Compressed {} bytes to {} bytes'.format(len(compressed_image), len(out.getvalue())))

    end_of_file = [
        0x3B
    ]
    out.write(bytes(end_of_file))

    return True


