#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
        
        with open('/dev/urandom', mode='rb') as rng,\
             open(self.output,    mode='wb') as out:

            return generate_random_gif(
                    rng, out
                    )

    
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    
    header = b''
    screendescriptor = bytearray()

    # Header and Screen Descriptor
    header += bytes([b'G', b'I', b'F'])   # magic number
    version = rng.read(1)[0]       # version of GIF file format
    header += bytes([version])      # also known as Version Number
    header += bytes([b'87a'])       # signature
    screendescriptor[0] = version  # copy gif version into ScrenDescriptor byte array
    header += bytes([screendescriptor[0]])             # width
    width = int.from_bytes(rng.read(2), byteorder='big')  # another short value
    screendescriptor[1] = width & 0xFF                   # add low byte to screen descriptor array
    screendescriptor[2] = width >> 8                     # add high byte to screen descriptor array
    header += bytes([screendescriptor[1],\
                     screendescriptor[2]])
    height = int.from_bytes(rng.read(2), byteorder='big') # another short value
    screendescriptor[3] = height & 0xFF                  # add low byte to screen descriptor array
    screendescriptor[4] = height >> 8                    # add high byte to screen descriptor array
    header += bytes([screendescriptor[3],\
                     screendescriptor[4]])
    flags = rng.read(1)[0]                                       # a single octet
    screendescriptor[-1] = flags                                 # add last byte 
    bgcolorindex = rng.read(1)[0]                                # background color index
    screendescriptor += bytes([bgcolorindex])
    aspectration = rng.read(1)[0]                                # pixel aspect ratio
    screendescriptor += bytes([aspectration])                      # add pixel aspect ratio
    header += bytes([flags, bgcolorindex, aspectration])         # finally add all the above to our header
    out.write(header)                                            # now we can write the header to the stream

    # Global Color Table
    gct = rng.read(3)                                             # read three more bytes
    gcts = gct[0] | (gct[1] << 8) | (gct[2] << 16)                 # convert those 3 bytes into one integer
    header += bytes([(gcts // 3)])                                  # pack that integer into one byte
    out.write(header)                                             

    # Image Descriptor
    imgdes = rng.read(9)                                          # from here on they are all plain binary
    localimgdes = bytearray()                                     # this is how we will construct local image descriptors    
    xposition = int.from_bytes(rng.read(2), byteorder='big')       # first two bytes are horizontal position
    localimgdes[0] = xposition & 0xFF                             # add them to list (low byte first)
    localimgdes[1] = xposition >> 8                               # add them to list (high byte second)
    yposition = int.from_bytes(rng.read(2), byteorder='big')       # follow up with vertical positions
    localimgdes[2] = yposition & 0xFF
    localimgdes[3] = yposition >> 8
    wpixelsize = rng.read(1)[0]                                   # number of pixels per row
    localimgdes[4] = wpixelsize                                    # just add it to our list of parameters
    hpixelsize = rng.read(1)[0]                                   # same thing but these time the other way around
    localimgdes[5] = hpixelsize
    flags = rng.read(1)[0]                                        # one byte of flags
    lf = 0x80                                                      # LZW min code size flag
    lzwmincodesize = 8                                            # minimum code size used by LZW algorithm
    sortflag = 0x00                                                # no sorting
    reservedbit = 0                                                # no reserved bit
    localimgdes += bytes([lf|lzwmincodesize, sortflag, reservedbit])# add params to local image descriptor's array
    interlaceflag = rng.read(1)[0]&0x01                            # one boolean flag
    lctflag = rng.read(1)[0]&0x01                                 # another
    lctsize = rng.read(1)[0]                                      # another one
    lctsortedflag = rng.read(1)[0]&0x01                           # yet another
    localimgdes += bytes([interlaceflag, lctflag, lctsize, lctsortedflag]) # add all these params to our local image descriptor array
    transpcolnum = rng.read(1)[0]
    localimgdes += bytes([transpcolnum])                           # and the transparent color index
    dataoffset = len(localimgdes)+len(screendescriptor)+3           # since we have already written everything into streams, adding up their lengths gives us the offset where image data starts
    localimgdes += dataoffset.to_bytes(2,'little')                # little endian encoding (actually byte order is big endian, but we'll swap bytes anyway when writing)
    header += bytes(localimgdes+screendescriptor)                   # finally concatenate header and local image descriptor to create global image structure
    out.write(header)                                              # and put both into our output stream

    # Local Color Table
    clt = rng.read(3)                                              # read color table
    clts = clt[0] | (clt[1] << 8) | (clt[2] << 16)                  # combine them
    header += bytes([(clts//3)])                                   # pack them into our header
    out.write(header)                                                

    # Color Table size
    ctsz = rng.read(1)[0] * 3
    for _ in range(ctsz):
        header += bytes([rng.read(1)[0]])

    # Block Size (Data Sub-Block within Data Stream)
    bs = rng.read(1)[0]         # block size is an unsigned char
    subblock = rng.read(bs - 1)  # discard sub-block marker
    header += bytes([bs])        # add block size back again
    data = rng.read(bs)          # get actual image data
    out.write(header + data)     # write it all together
    
