#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        while len(palents) < ctsize * 3:

            nbits = min(ctsize * 3 - len(palents) * 8,
                    8)

            val = int.from_bytes(rng.read(1),
                    byteorder='big') >> (8 - nbits)

            palents += struct.pack('%dB' % 
                    nbits, *[val & ((0b1 << i) - 1)
                        for i in reversed(range((nbits+7)//8))]
                    )

        out.write(header + struct.pack('<HH', image_width, 
            0) + palents)

    else:

        # no palette
        out.write(header + struct.pack('<HH', image_width, 
                image_height))

    # get remaining lzw compressed info from the RNG
    datalen = int.from_bytes(rng.read(2), 'big')
    data = rng.read(datalen)
    out.write(header + data)
