""" Generates a random GIF file with arbitrary content """

import os

import numpy as np


class GifGenerator:

    def __init__(self):
        self._seed = None
        self._data = []


    @property
    def seed(self):
        return self._seed


    @seed.setter
    def seed(self, value):

        # Seed can be an integer or it can be a list or array-like object
        if isinstance(value, (int, float)):
            sval = str(int(value))
        else:
            try:
                valarray = np.asarray(value)
            except ValueError:
                raise TypeError("Seed must be either an integer "
                                "or something that can be cast to one")

            # Convert values which cannot be represented in a uint32_t back down
            sval = ''.join((str(int(v)) for v in valarray
                            if (-0x7fffffff - 1 <= v < 0x7fffffff)))


        self._seed = sval


    def generate(self, nbytes):

        # Ensure there's enough data remaining to satisfy this request
        if len(self._data) < nbytes:
            self._fill()

        result, self._data = self._data[:nbytes], self._data[nbytes:]
        return result


    def _fill(self):

        gen = np.random.RandomState(self._seed)
        self._seed += 'e'     # ensure we're not just repeating ourselves!

        # Generate some more random numbers and put them all together as hex strings
        vals = [hex(gen.randint(-2**31, 2**31))[2:] for i in range(6)]
        while len(''.join(vals)) > 0:
            for i in range(6):

                # Every 8 hex characters represents a single byte, so we need
                # to pad the end of our string with zeros until we can successfully
                # convert these to bytes without overflowing their representation
                test_str = ''.join(vals)
                if ((i+1)%8 == 0) and (len(test_str) % 2!= 0):
                    test_str += '0'           # zero padding
                elif len(test_str) >= 2:

                    # Try converting what we have thus far to make sure it works. If
                    # it doesn't, keep adding another character on at the end. This will
                    # eventually cause us to hit a length where the last two hex chars
                    # represent a full byte, at which point we add another random number
                    # before continuing...
                    try:
                        byte = int(test_str[:2], base=16)
                        break
                    except OverflowError:
                        pass

                vals[5-i] = vals[(5-i)-1][:-1]   # pop off extra characters from larger ones
                if len(vals[5-i]) == 0:          #...or add random numbers if needed
                    vals[5-i] = hex(gen.randint(-2**31, 2**31))[2:]

        # Build up the header and pixel blocks
        header = b'GIF89a'
        num_blocks = int(np.ceil(len(vals)/255.))
        pixel_blocks = b''
        for blocknum in range(num_blocks):
            startidx = blocknum*255
            endidx = min((blocknum+1)*255, len(vals))
            pixels = bytes([int(v,base=16) for v in vals[startidx:endidx]])
            pixel_blocks += b'\x2c\x00\x00\x00\xff' + pixels


        # Write out the new data
        self._data += [struct.pack('<3sB3sBHBHHB',
                                   header[i:i+1], ord(header[i+1]),
                                   header[i+2:i+3], ord(header[i+3]),
                                   *[ord(b) for b in bytearray(pixel_blocks)])
                       for i in range(0, len(header), 4)]



def main():

    g = GifGenerator()
    g.generate(os.urandom(3)[0]*2**16)
    print('Seed: {}'.format(g.seed))

    with open('/tmp/random.gif', mode='wb') as f:
        g.generate(os.urandom(3)[0]*2**16)
        f.write(b''.join(g._data))
