#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (`/dev/urandom`)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # Create buffer for header, which consists of its own size,
    # and some other junk we're not going to use right now
    headerbuf = bytearray()
    
    # Read first three bytes of header.  This contains the size
    # of the rest of the header in bytes. We also want to make sure that
    # those are valid ASCII characters
    headerlen = rng.read(3)
    assert set(headerlen) <= {ord('0'), ord('1'), ord('2')} \
                or set(headerlen) == {ord('\x00'), ord('\xff')}
    # Add padding to the beginning of the header so it's always multiple
    # of 256
    headerbuf += b'\x00\x00' + headerlen

    # Read all remaining bytes of the header until we reach the
    # start of the logical screen descriptor
    while True:
        hbyte = rng.read(1)
        if hbyte == b';': break

        # At this point, we only care about the following bytes:
        #     (*) Logical Screen Descriptor
        #         (**) Canvas Width
        #         (**) Canvas Height
        #         (***) Global Color Table Flag
        #         (***) Color Resolution
        #         (***) Sorting Method
        #         (***) Global Color Table Size
        
        # If it isn't one of these, write it as-is to our header
        if hbyte!= b'!':
            headerbuf.append(hbyte[0])
            continue

        # Read the next two bytes.  These describe the type of data
        # that follows them. The first nibble is the block size, and
        # must be at least 4. For example, the value 0x0A means that
        # there will be an additional 10 bytes after these, excluding
        # any padding.
        cmd = rng.read(2)
        blksize = int.from_bytes(cmd[:1], 'big')
        if blksize < 4: raise RuntimeError('Invalid block size')
            
        # Now read the rest of the command.  Depending on what it is,
        # decide how much more data to read before emitting it
        if cmd[1] == 0xF9: pass    # Graphic Control Extension
        elif cmd[1] == 0xFE:       # Comment Extension
            commentlen = blksize - 2

            # Make sure they're actually trying to hide something
            assert commentlen % 2 == 0
            
            # Get rid of NULLs
            commentlen //= 2
            for _ in range(commentlen):
                rndchr = rng.read(1)                
                # Only accept printable and space chars
                assert set(rndchr) <= {
                    chr(c) for c in range(ord(' '), ord('~')+1)}
        elif cmd[1] == 0xFF:       # Application Extension
            appnamelen = int.from_bytes(rng.read(1), 'big')
            appname = rng.read(appnamelen)

            # Ignore all extensions except Netscape Looping Application Extension
            if appname == b'NETSCAPE2.0':

                # First read the application identifier, which should contain
                # the string "NETSCAPE" followed by the version indicator "\x01".
                assert rng.read(9) == b'NETSCAPE2.0\x01'
                
                # Then read the subblock containing the loop counter. Convert this
                # from little endian to big endian (and replace the null terminator
                # with a big endian short integer indicating how many times the GIF 
                # should be looped.)
                loopcount = struct.unpack('<L', b'\x00'+rng.read(2))[0]
                
                # Finally, look for the end of the extension marker (\x3B)
                while rng.read(1)!= b'\x3B': pass
            else:
                # Skip over the entire block body
                rng.seek((blksize - 1) - appnamelen, io.SEEK_CUR)
        elif cmd[1] == 0x2C:       # Image Descriptor
            # Simply skip past the whole thing
            rng.seek((blksize - 1))
        else:
            # Unrecognized control code!
            raise RuntimeError(f'Unsupported code: {hex(int(cmd[1]))}')

        # Append this chunk to our header
        headerbuf += cmd + rng.read(blksize - 2)
        
        
    # Add the string ";\x21\xF9\x04\x00\x00\x00\x00\x00\x00;" to the
    # header to indicate the start of the graphics control extension
    headerbuf += b";\x21\xF9\x04\x08\x00\x00\x00\x00\x00\x00;\r"

    # Construct a dummy palette -- RGB triplets of eight bits each
    # However, since we don't know exactly how large our image will be,
    # just let the library fill in the correct size later instead
    palettesize = ((headerbuf[-2]+1)*3)-1
    palette = []
    for i in range(palettesize):
        palette += [i&0xff, (i>>8)&0xff, (i>>16)&0xff]

    # Fill in the correct size of the pixel data
    datalen = headerbuf[-3] + headerbuf[-2]*headerbuf[-1]
    headerbuf[-3:] = [datalen&0xff, (datalen>>8)&0xff, (datalen>>16)&0xff]
        
    # Write everything to disk
    out.write(headerbuf)
    out.write(palette)
    while datalen>0:
        nwritten = os.write(out.fileno(), rng.read(min(datalen, 2**20)))
        datalen -= nwritten
