#!/usr/bin/env python3

import argparse
import io
import math
import os
import random
import struct
from typing import BinaryIO


DEFAULT_OUTPUT = './random.gif'


def main() -> None:

    
    parser = argparse.ArgumentParser()
    parser.add_argument('--width', '-W', type=int, default=100)
    parser.add_argument('--height', '-H', type=int, default=100)
    parser.add_argument('--color-table-size', '-c', type=int, default=7)
    parser.add_argument('--data-sub-block-size', '-d', type=int, default=4096)
    parser.add_argument('output', nargs='?', metavar='OUTFILE',
                        help="The name of the output file")

    args = parser.parse_args()

    width = args.width or int(input("Width? ")) 
    height = args.height or int(input("Height? ")) 
    colorTableSize = args.color_table_size or int(input("Color table size? ")) 
    dataSubBlockSize = args.data_sub_block_size or int(input("Data sub block size? ")) 

    outputFile = f'{os.path.abspath("./")}{"/" if "/" not in args.output else ""}{args.output}'

    generate_random_gif(outputFile, width, height, colorTableSize, dataSubBlockSize)
    

def generate_random_gif(fileName: str, width: int, height: int, 
                        colorTableSize: int, dataSubBlockSize: int):

    rng = io.BytesIO()
    generate_random_gif_content(rng, width, height, colorTableSize,
                                dataSubBlockSize)

    try:
        with open(fileName, "wb+") as fh:
            fh.seek(0, 0)
            fh.write(rng.getvalue())
    except OSError as e:
        print(e, file=sys.stderr)


def generate_random_gif_content(rng: BinaryIO, width: int, height: int, 
                                colorTableSize: int, dataSubBlockSize: int):

    # Signature
    header = b'GIF89a'
    
    # Width and Height
    w = struct.pack('H', width)    # width is 16 bit
    h = struct.pack('H', height)   # height is 16 bit
    header += w + h

    # Color Table size
    ctsz = colorTableSize * 3
    for i in range(ctsz):
        rgb = []
        while len(rgb) < 3:
            r = random.randint(0, 255)

            # Generate colors that are less likely to be black
            if r <= 20:
                continue
            
            rgb.append(r)
        
        header += struct.pack('BBB', *rgb).ljust(3, b'\x00')

    # Block Size (Data Sub-Block within Data Stream)
    bs = dataSubBlockSize & 0xFF      # block size is an unsigned char
    dsblm = ((dataSubBlockSize >> 8) & 0xFF) << 7
    header += struct.pack('B', bs | dsblm)       # add block size back again
    data = bytearray(struct.pack('BxHxH', 0x2C, width, height))

    numPixels = width * height
    idx = 0
    while True:
        l = min((numPixels - idx), dataSubBlockSize // 3)
        n = f'{l}P3\n'.encode('ascii')
        idx += l
        if idx >= numPixels: break
        data += n

    out = [header + data]

    # Graphics Control Extension
    gce = bytearray(b'\x21\xF9\x04')                          # graphics control label
    gce.append(((dataSubBlockSize // 10) & 0xff))           # delay time
    gce.append(0)                                           # Transparency index
    gce.append(0)                                           # Terminator
    out.append(gce)

    # Image Descriptor
    id = bytearray(b'\x2C')                                 # image descriptor label
    id.extend(struct.pack('<HHBB', width, height,             # dimensions
                          0, 0))                           # no local color table    
    out.append(id)

    # Application Extension (Terminal = False)
    ae = b''
    ae += b'\x21\xFF\x0BNETSCAPE2.0'                       # application identifier
    ae += b'\x03\x01'                                       # block length
    ae += b'\x01'                                           # terminal data stream flag set
    ae += b'\x00\x00'                                       # No comment
    out.append(ae)

    # Application Extension (Terminal = True)
    ae = b''
    ae += b'\x21\xFF\x0BNETSCAPE2.0'                       # application identifier
    ae += b'\x03\x01'                                       # block length
    ae += b'\x00'                                           # terminal data stream flag unset
    ae += b'\x00\x00'                                       # No comment
    out.append(ae)

    out[0] += b'\x3B'                                       # terminator

    assert sum([len(_) for _ in out]) == len(header) + len(data)
    return b"".join(out)

