#!/usr/bin/env python3

import itertools as it
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given pseudorandom number generator in binary format
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Set up the global color table with some basic colors. Ideally we'd have an infinite palette of
    # unique colors to choose from but then that would mean having 256*256*256 possible colors and
    # we can't be bothered.
    colortbl = [
        0x0000FF,      # blue
        0xFF00FF,      # magenta
        0xFFFF00,      # cyan
        0x00FFFF,      # yellow
    ]

    def get_rgb():
        yield from (rng.read(1)[0] % 2 for _ in range(3))

    def make_color():
        return tuple(get_rgb())

    def has_nonzero(*args):
        return any((arg!= 0 for arg in args))

    def mask_last_byte(value):
        return value & ~0x7F

    while True:
        # Get the first 3 bytes to determine how long this block is. The length counts
        # everything including itself so subtracting one will give us its true length.
        ident = rng.read(3)
        if not ident or len(ident) < 3:
            break

        instr_len = ((struct.unpack('>B', ident[0].to_bytes())[0] & 0b01111) << 8) | \
                    struct.unpack('>H', ident[:2])[0] - 1
        
        # Read in the instruction type and figure out which blocksize we're dealing with based on
        # the instruction type
        instr_type = struct.unpack('>B', rng.read(1))[0]
        if instr_type == 0xF9:   # Graphics Control Extension Block
            blk_sz = 6
        elif instr_type == 0xFE: # Comment Extension Block
            blk_sz = ident[-1]+1
        else:                   # Plain old graphics extension block
            blk_sz = 2

        # Make sure that our input stream contains enough data available before we attempt to read
        # something out of it since they'll always be at least one byte less than specified
        # because counting starts at zero
        if len(ident)+blk_sz > instr_len+1:
            continue

        # Now read the actual instruction data and act accordingly depending on what kind of
        # instruction we've encountered
        if instr_type == 0xF9:       # Graphics Control Extension Block
            flags = struct.unpack('>B', rng.read(1))[0]
            delay = struct.unpack('>H', rng.read(2))[0]

            transparent = False
            if flags & 0x01:
                transparent = True
                trans_idx = struct.unpack('>B', rng.read(1))[0]
            
            disposal = 'none'
            if flags & 0b01100000:
                disposal = {
                    1 : 'do_not_dispose',
                    2 :'restore_to_bg',
                    3 :'restore_to_prev',
                }[(flags >> 2) & 0b11]
        
            userinput = False
            if flags & 0b10000000:
                userinput = True
                
            out.write(ident)
            out.write(instr_type.to_bytes(1, byteorder='big'))

            new_flags = 0
            if transparent:        
                new_flags |= 0x01
            if disposal == 'do_not_dispose':
                new_flags |= 0b00000000
            elif disposal =='restore_to_bg':
                new_flags |= 0b00100000
            elif disposal =='restore_to_prev':
                new_flags |= 0b01000000
            if userinput:
                new_flags |= 0b10000000
            out.write(new_flags.to_bytes(1, byteorder='big'))

            if transparent:
                out.write(trans_idx.to_bytes(1, byteorder='big'))
            out.write(delay.to_bytes(2, byteorder='big'))
            
        elif instr_type == 0xFE:     # Comment Extension Block
            comment = rng.read(ident[-1])
            out.write(ident + comment)
        else:                       # Plain old graphics extension block
            sub_blockid = struct.unpack('>B', rng.read(1))[0]
            out.write(ident + (sub_blockid).to_bytes(1, byteorder='big'))
        
        # Figure out whether there's another item left to process
        hasnext = bool((instr_len - ident[-1]) // blk_sz - 1)
        out.write(hasnext.to_bytes(1, byteorder='big'))

        # If there are multiple items remaining we should reset the colors
        # again otherwise things look weird!
        if instr_type == 0xF9:
            colortbl = []
    
    # Construct the final image descriptor block
    imgdesc = 0x2C          # Image Descriptor Block
    imgdesc |= 0x00        # No local color table
    imgwidth = rng.read(2)
    imgheight = rng.read(2)

    # Figure out the global color table size by scanning through all the individual blocks to see
    # where we end up. We do this because we need to know how many entries there are in order to
    # correctly store the real length later when we write out the block
    gct_length = sum([1 if block[0] == 0x2C else 0 for block in it.zip_longest(itertools.repeat(imgdesc, 1000), out)])
    gct_size   = (gct_length**0.5)**2

    # Now that we know the correct GCT size, let's construct the final image descriptor
    imgdesc += (gct_size-1).to_bytes(1, byteorder='big')
    imgdesc += imgwidth
    imgdesc += imgheight
    imgdesc += ((imgdesc + gct_length+1) % 256).to_bytes(1, byteorder='big')

    # Write out the complete image descriptor
    out.write(imgdesc)