from typing import BinaryIO


def lcg_minstd32(seed):
    i = seed

    def next():
        nonlocal i

        # Implementation assumes integers are at least 32 bits wide
        i = (i ^ ((i & 63) | 19)) * 1664525 + 1013904223
        return i >> 32


    return next


# Generates a random GIF file into `out` using the
# linear congruential generator `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    bitstream = []
    while len(bitstream) < 10:
        bitstream += bin(int.from_bytes(
            rng.read(4), byteorder='big', signed=False))[2:]
    for _ in range(7):
        bitstream += '1'

    width = int(''.join((
        bitstream[:11],
        bitstream[11:12] if not bitstream[11] == '0' else '',
        bitstream[-9:-8]), base=2), base=2)
    height = int(''.join((
        bitstream[-9:],
        bitstream[-8:-7] if not bitstream[-8] == '0' else ''), base=2), base=2)
    # Force even dimensions; this is what most software does anyway
    width -= (width % 2)
    height -= (height % 2)
    print("w:{} h:{}".format(width // 2, height // 2))
    write_uint16(out, width // 2)
    write_uint16(out, height // 2)
    color_table_size = min(len(bitstream) - 10, 2 ** read_bits(bitstream, 3) + 1)
    # The first value of each color table entry is black but we don't need to
    # bother setting it since that's the default value.
    palette = [(0x00,) * 3] * color_table_size
    indices = [None] * (width * height)
    index = 0
    while True:
        for y in reversed(range(height)):
            for x in range(width):
                try:
                    indices[index] = read_bits(
                        bitstream, read_bits(bitstream, 3))
                except IndexError as e:
                    break

                if index >= len(indices) or \
                   indices[index] > len(palette):
                    raise ValueError('Invalid index')

                write_byte(out, palette[indices[index]][0])
                write_byte(out, palette[indices[index]][1])
                write_byte(out, palette[indices[index]][2])
                index += 1

            if index >= len(indices):
                break

        if index >= len(indices):
            break

        # Make sure there's enough space left for the new colors
        remaining_colors = max(color_table_size -
                              len(palette), width * height)

        prev_colors = palette[:]
        palette = [None] * color_table_size
        j = 0
        for neighbours, color in enumerate(prev_colors):
            neighbour_colors = set()
            for n in neighbours_1d(neighbours, prev_colors):
                if n!= None and n <= len(prev_colors) - 1:
                    neighbour_colors |= {prev_colors[n]}
            if len(neighbour_colors) == 0:
                palette[j] = color
                j += 1
            elif len(neighbour_colors) == 1:
                palette[j] = list(neighbour_colors)[0]
                j += 1
            else:
                dithered_colors = dither(list(neighbour_colors))
                for dc in dithered_colors:
                    palette[j] = dc
                    j += 1

        if j!= len(palette):
            raise ValueError('Incorrectly generated palette')

        for k in range(remaining_colors):
            palette[k] = pick_color(rng)


        # TODO: Do something here? Maybe make a global palette of all possible
        # pixel values. I think we can achieve this by making our own version
        # of the floyd-steinberg dithering algorithm where the error isn't reset
        # after every pixel...
        pass

    imgdata = bytes([read_byte(bitstream)] * len(indices))
    write_gif_image_data_block(imgdata, out)


# Generator of 1D array indices adjacent to `p`. We assume `p` is always less
# than len(arr).
def neighbours_1d(p, arr):
    yield p - 1
    yield p + 1
    yield p - len(arr)
    yield p + len(arr)
    yield p % len(arr) - 1
    yield p % len(arr) + 1



# Converts an integer to its big endian representation as a string of binary
# digits. This will be twice as long as the length of str(num) because the
# conversion uses two chars per digit.
def num_to_binstr(num: int):
    return ''.join(bin(c)[2:].zfill(8) for c in
               bytearray(num.to_bytes(length=(num.bit_length() // 8) + 1,
                                      byteorder='big')))


def read_bits(bitstream, count: int) -> int:
    value = 0
    for i in range(count):
        value <<= 1
        value |= int(next(bitstream))
    return value


def read_byte(bitstream):
    return read_bits(bitstream, 8)


def write_byte(f, b):
    f.write(b.to_bytes(length=1, byteorder='big'))


def write_uint16(f, v):
    f.write(v.to_bytes(length=2, byteorder='big'))


# Returns three ints from 0-255 representing one RGB color.
def pick_color(rng):
    return tuple(b for b in (rand(rng)(), rand(rng)()
                            , rand(rng)())
                 )


def rand(rng):
    @lcg_minstd32
    def gen(_):
        return next(rng)

    return lambda _: gen(_)


def dither(colors):
    """Performs Floyd-Steinberg dithering on the given palette."""
    errors = [[0, 0, 0]] * len(colors)
    for row, col in itertools.product(reversed(range(len(errors))),
                                       range(len(errors[row]))):
        color = colors[col][::-1]  # Reverse so LSB is leftmost
        r = color[0]
        g = color[1]
        b = color[2]
        errR = int(round(r / 8))
        errG = int(round(g / 8))
        errB = int(round(b / 8))
        rgb = tuple(map(lambda v: round(v)
                      , map(sum, zip(color, errors[row][col]))))
        errors[row+1][col] = [(errR * 7) % 8, (errG * 7) % 8, (errB * 7) % 8]
        yield rgb
