#!/usr/bin/env python3

import os
import random
import struct
from typing import BinaryIO


class RandomGifGenerator():

    def __init__(self):

        self._seed = None


    @property    
    def seed(self) -> str:
        
        return self._seed

    
    @seed.setter
    def seed(self, value: int) -> None:

        self._seed = value
    
    @staticmethod
    def _gen_palette_color(index: int) -> tuple:
        '''
        Generate RGB color triplet, given an integer index
        '''

        # create triple with each component being a multiple of
        # index and add some noise at least three times larger than 
        # this index. This gives us a more interesting palette.
        # Triplets below zero will be clamped to zero, since we're
        # just picking colors at random.
        r = ((index ** 3) % 256) + max([0, index // 2]) + \
                    min([0, (index * (-1)) // 2])
        g = (((index + 7) ** 2) % 256) + max([0, index // 5]) + \
                        min([0, (index * (-1) // 5)])
        b = ((((index + 11) ** 1.5) % 256) + max([0, index // 11]) + 
                    min([0, (index * (-1) // 11)]))

        return (r, g, b)
    

    @staticmethod
    def _gen_lzw_compressed_sequence(rng: BinaryIO, 
        bufsize=None) -> bytes:

        # initial dict
        d = {
            0: '',
            1: '\x01',
            2: '\x00\xff'
        }


        # load entire buffer if no bufsize specified
        if not bufsize:
            bufsize = os.path.getsize(rng.name or '/dev/urandom')

        data = b''
        curr = ''
        prev = ''

        while True:
            
            # pick random next character and append it to current string
            c = chr(int.from_bytes(rng.read(1),'little'))
            nxt = f'{curr}{c}'

            # add new entry to dictionary
            if nxt not in d.values():

                # choose unused key
                k = max(d.keys()) + 2
                
                d[k] = nxt
                curr = c

            else:

                curr = d[[v for v in d
                            if d[v] == nxt][0]]
                prev += c
                


            # generate at most one extra frame per byte
            if (prev!= '') and ((bufsize < len(data)+1) or 
                (c == chr(0xFF))):

                # stream previous two characters as single-byte codes
                outchr = []

                for i in range(-1,-3,-1):

                    try:

                        # encode as LZW code
                        outchr.append(f'\xfe{ord(d[int(prev[i:], 16)])}')
                        
                        del d[int(prev[i:], 16)]

                    except KeyError:

                        continue
                    
                    finally:

                        prev = prev[:i]
                        

                # always append last two bytes as is
                outchr.append('\xfe\xff')

                # update data with encoded sequence
                data += ''.join(outchr).encode('ascii')

                # reset previous two characters
                prev = ''
        

            elif bufsize >= len(data)+1:
            
                break

        return data


    @staticmethod 
    def _compress_image_data(data: str) -> bytes:
                
        # use LZW compression on input image data
        imgdata = []
        dic = {}
        char = data[0]
            
        for i in range(1, len(data)):
        
            if data[i] in dic:
                    
                char = data[i]
                
            else:
    
                imgdata.append((dic[char],))
                dic[char+data[i]] = len(dic)
                char = data[i]
                            
        imgdata.append((dic[char],))
    
        imgdata = b''.join(map(lambda x:(len(x)).to_bytes(1,'big')+x,imgdata))
        
        return imgdata


    def gen_random_gif(self, out: BinaryIO)->None:

        # initialize pseudo-random generator
        rng = open('/dev/urandom','rb')
        random.seed(self._seed)

        # build headers
        out.write(b'GIF89a')
        out.write(struct.pack('<2H', *(self._gen_dimension())))
        out.write(struct.pack('<BBBB', *[
            *self._gen_screendescriptor(), 0b10000000]))
        
        # get palette and bit depth information
        paletteflag = bool(random.randint(0,1))
        palette = []

        if paletteflag:

            # load palette with up to 256 colors
            palettesize = random.randint(2, 256) 
            palette = map(self._gen_palette_color,range(palettesize))

            # generate palette entries
            colortable = bytes()
            for clr in palette:
                colortable += struct.pack('BBB',*clr)

            # attach to output
            out.write(colortable)

        # amount of space left in table
        tableleft = pow(2, random.choice([1,2,3])) - 1 

        # generate image data
        for i in range(tableleft):
            lzw = self._gen_lzw_compressed_sequence(rng)
            out.write(self._compress_image_data(lzw))

        # pad with zeros until end-of-file
        padding = random.randint(pow(2,2)-1,pow(2,12)-1)
        out.write(b'\x00'*padding)