#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    width = int.from_bytes(
                rng.read(1), byteorder='big', signed=False
            ) + 1
    height = int.from_bytes(
                rng.read(1), byteorder='big', signed=False
            ) + 1

    globalcolortableexists = bool.from_bytes(
                rng.read(1), byteorder='big')
    localcolortableexists = False
    interlaceflag = False
    sortflag = True
    reservedflags = (bool.from_bytes(
                    rng.read(1), byteorder='big') << 7) | \
                    bool.from_bytes(
                        rng.read(1), byteorder='big')
    backgroundcolorindex = int.from_bytes(
                        rng.read(1), byteorder='big')
    pixelaspectratio = int.from_bytes(
                        rng.read(1), byteorder='big')
    # Now read the global colormap, which is possibly present
    if globalcolortableexists:
        for _ in range((int)(pow(
                        2, pixelaspectratio+1))):
            out.write(rng.read(3))
    
    # The image descriptor
    out.write(width.to_bytes(2, 'little'))
    out.write(height.to_bytes(2, 'little'))
    localcolortableexists = bool.from_bytes(
                        rng.read(1), byteorder='big')
    interlaceflag = bool.from_bytes(rng.read(1), byteorder='big')
    sortflag = bool.from_bytes(rng.read(1), byteorder='big')
    reservedflags |= ((bool.from_bytes(
                        rng.read(1), byteorder='big') << 7) | \
                            bool.from_bytes(
                                rng.read(1), byteorder='big'))
    backgroundcolorindex = int.from_bytes(
                        rng.read(1), byteorder='big')
    pixelaspectratio = int.from_bytes(
                        rng.read(1), byteorder='big')
    # Now read the local colormap, which is possibly present
    if localcolortableexists:
        for _ in range((int)(pow(
                        2, pixelaspectratio+1))):
            out.write(rng.read(3))

    # We're now ready to write the table of content section
    # We'll build it as we go along instead of at once because
    # we have no idea how many frames there will be...
    datapos = out.tell()
    toclength = 0
    # First, sortflag and gce
    out.write(((sortflag & 0x1) << 3) |
              (((not globalcolortableexists) & 0x1) << 2) |
              0)
    toclength += 1
    # Then all other flags except lce
    out.write((interlaceflag & 0x1) << 7 |
              ((not localcolortableexists) & 0x1) << 6 |
              ((reservedflags >> 6) & 0x3) |
              0)
    toclength += 1
    # Delay time
    out.write(rng.read(2))

    # Transparent color index
    out.write(backgroundcolorindex.to_bytes(1, 'little'))
    toclength += 1
    out.write(pixelaspectratio.to_bytes(1, 'little'))
    toclength += 1
    # Write the offset of the LZW compressed sub-blocks of each frame
    # after this point
    #tocstart = out.tell()
    for i in range(40):
        toclength += 1
        out.write(rng.read(2))
    
    # End of the table of contents
    out.write(b'\x3B')

    totaldatacount = 0
    for i in range(40):
        delaytime = int.from_bytes(rng.read(2),
                                    byteorder='little')
        transptr = int.from_bytes(rng.read(1),
                                  byteorder='little')
        imgleft = width
        imgtop = height
        localcolortableexists = bool.from_bytes(
                                        rng.read(1),
                                        byteorder='big')

        codedsubblocks = []
        # Read until we reach the end of the current frame
        while imgleft!= 0 or imgtop!= 0:
            control = int.from_bytes(
                                       rng.read(
                                           1), byteorder='big')
            
            # Disassemble control bytes
            interlaceflag = (control & 0x40 == 0x40)
            sortedflag = (control & 0x20 == 0x20)

            numofpasses = \
                    1 + int.from_bytes(
                                        rng.read(
                                            1), byteorder='big')*8
            leftcoord = int.from_bytes(
                                        rng.read(
                                            1), byteorder='big')
            topcoord = int.from_bytes(
                                        rng.read(
                                            1), byteorder='big')
            width = int.from_bytes(rng.read(1),
                                   byteorder='big')
            height = int.from_bytes(rng.read(1),
                                    byteorder='big')
            # The meaning of the above values depends on
            # whether the current frame uses a local colourmap
            print(f"Frame {i} contains pixels ({imgleft}, {imgtop}) "
                  f"- ({leftcoord}, {topcoord}), size ({width},{height})")
            localcolortableexists &= bool.from_bytes(
                                                rng.read(
                                                    1),
                                                byteorder='big')
            lctsize = pow(
                            2, int.from_bytes(
                                                 rng.read(
                                                     1), byteorder='big')+1)
            lct = []
            for j in range(lctsize):
                lct.append(rng.read(3))
            lzwmincodesize = rng.read(1)[0]
            #print("Lzw min code size:", lzwmincodesize)
            datacount = 0
            # Let's start by writing the control bytes again so that
            # we can use them later when assembling the final stream
            out.write(control.to_bytes(1,'big'))
            # Next comes the number of sequential passes
            out.write(numofpasses.to_bytes(1,'big'))
            # Then the left and top coordinate
            out.write(leftcoord.to_bytes(1,'big'))
            out.write(topcoord.to_bytes(1,'big'))
            # Then the width and height of the rectangle
            out.write(width.to_bytes(1,'big'))
            out.write(height.to_bytes(1,'big'))
            # Finally, let's add our local colormap flag
            # and the local colormap size
            out.write((localcolortableexists << 7) |
                      (lctsize-1).to_bytes(1, 'big'))
            # And then last but not least, the minimum LZW code size
            out.write(lzwmincodesize.to_bytes(1,'big'))
            # OK! Now we can start reading bits from /dev/urandom
            bitbuffer = ''
            for passnum in range(numofpasses):
                if interlaceflag:
                    imgleft -= (width//8)*8
                    imgtop -= (height//8)*8
                    # Check if we went over the edge
                    if imgleft < 0:
                        print("We went too far horizontally!")
                    elif imgtop < 0:
                        print("We went too far vertically!")
                    if imgleft <= 0 or imgtop <= 0:
                        break

                # Assemble the run length encoded scanline
                data = rng.read(
                                     ((imgleft+7)//8)*
                                                        height +
                                                        height//8 +
                                                        1)
                datacount += len(data)
                
                # Now compress the line!
                complen = zlib.compressobj(wbits=-lzwmincodesize)\
                                .compress(data)
                complen += zlib.compressobj()\
                                .flush(zlib.Z_FINISH)
                compdata = complen[:-4]

                complen = int.from_bytes(complen[-4:],
                                         byteorder='big')
                # And finally, store the compressed data
                out.write(complen.to_bytes(2,'big'))
                out.write(compdata)
                #if interlaceflag:
                    # Remember where to put the row offsets
                #    rowoffsetptrs.append(out.tell())

                    # TODO: Insert the row offsets here
            totaldatacount += datacount
        # Now add a trailer for the frame
        out.write(transptr.to_bytes(1,'little'))
        out.write(delaytime.to_bytes(2,'little'))
        out.write(b'\x00')
    # And finally, write the table of contents, starting
    # right before the first frame
    out.seek(-toclength, 1)
    tocend = out.tell()
    tocstart =