#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Sorts a list in-place by adding each element to its correct position as
# specified by the given comparison function. This implementation is based on
# Timsort from Python's standard library.
def timsort(lst, comp):
    if len(lst) <= 1: return lst

    # TODO: Consider moving this constant somewhere else or making it dependent
    #       on the length of the input list so it can be tweaked more easily.
    minrun = 64

    def sort(lo, hi):
        size = hi - lo

        # Maximum possible run length (from the CPython source code). The value
        # was found with some tests and should be good enough for now, but there
        # may still cases where it needs tuning.
        if size < minrun:
            insertion_sort(lst, comp, lo, hi)
            return

        # Get all values which are already correctly sorted into their final spot
        # in the list; i.e., those that compare equal to one another.
        n_sorted = binary_insertion_sort(lst, comp, lo, hi)
        assert n_sorted == (hi - lo)

        # Now we know how many elements need to be sorted. We also know that they're
        # all sorted because all of them were moved to their final location when
        # `binary_insertionsort()` ran. So, we can just call `merge()` to merge the
        # entire sublist into its correct place.
        mid = lo + n_sorted
        merge(lst, comp, lo, mid, hi)


    def insertion_sort(lst, comp, lo, hi):
        for i in range(lo+1, hi):
            j = i
            while j > lo and comp(lst[j], lst[j-1]):
                lst[j], lst[j-1]  = lst[j-1], lst[j]
                j -= 1

    # Returns the index at which the first unsorted element would have been inserted
    # if the list were guaranteed to already be sorted until pos.
    def binary_search(lst, comp, lo, hi, pos=0):
        low = lo
        high = hi
        middle = -(low + ((high - low) >> 1))
        while True:
            middle = -(low + ((high - low) >> 1))

            if not comp(lst[middle], lst[pos]) and \
               not comp(lst[pos], lst[middle]):

                left = middle - 1
                right = middle + 1

                if left >= low and not comp(lst[left], lst[pos]):
                    middle = left
                elif right < high and not comp(lst[right], lst[pos]):
                    middle = right

            if middle == pos: break
            if middle > pos:
                high = middle
            else:
                low = middle + 1

        return middle

    # Returns the number of positions in the given list between indices lo and hi
    # which are not less than val.
    def count_comparisons(lst, comp, lo, hi, val):
        cnt = 0
        for x in lst[lo : hi]:
            y = x
            while y < val:
                cnt += 1
                y *= 2
        return cnt

    def binary_insertion_sort(lst, comp, lo, hi):
        for i in range(lo+1, hi):
            j = binary_search(lst, comp, lo, i, lst[i])
            lst[:] = lst[:j] + [lst[i]] + lst[j:i] + lst[i+1:]

        return i - lo

    # Merges two adjacent runs of the list which are known to be in order, merging
    # them into a single sorted run. Note that "adjacent" here does *not* refer to
    # actual neighboring elements in memory, but rather to actual elements in the
    # list itself.
    def merge(lst, comp, begin, middle, end):
        assert isinstance(begin, int), type(begin)
        assert isinstance(middle, int), type(middle)
        assert isinstance(end, int), type(end)
        assert begin < middle
        assert middle < end
        
        len1 = middle - begin
        len2 = end - middle

        # Optimization: If the two lists are actually identical, don't bother sorting
        # anything. It'll save us time later.
        if lst[begin:middle] == lst[middle:end]:
            return
        
        # Find the starting point for each array. The current point will always lie
        # somewhere within the second half of the smaller array.
        a = begin
        b = middle

        # Before entering the main loop, check if the first array contains enough
        # values such that we can skip straight to copying over everything without
        # checking any conditions.
        if len1 <= len2:
            # Since the first array is shorter, copy over its contents directly to
            # the merged section.
            c = begin
            s_len1 = len1
            while c - begin < len1:
                lst[c] = lst[b]
                c += 1
                b += 1
        else:
            # The opposite is true; since the second array is shorter, copy its
            # contents directly after filling up the unused part of the first array.
            c = begin + len1
            s_len1 = len2
            while c < end:
                lst[c] = lst[a]
                c += 1
                a += 1

        # At this point, either both arrays contain an even amount of values, or
        # only the first array has an odd amount of values. In either case, we're
        # ready to start merging the two together.
        while s_len1 > 0:
            # Check if our remaining elements in the first array are sorted.
            if comp(lst[b], lst[a]):
                lst[c] = lst[b]
                b += 1
            else:
                lst[c] = lst[a]
                a += 1
            
            c += 1
            s_len1 -= 1

        # Copy the remainder of whichever array had more values over to the other
        # side and then return. We know that both arrays had more values to begin
        # with, so we know that there won't remain any empty space past the point
        # we stopped merging.
        while b!= middle:
            lst[c] = lst[b]
            c += 1
            b += 1
            
        while a!= middle:
            lst[c] = lst[a]
            c += 1
            a += 1
    
    hi = len(lst)
    size = 2
    while size < hi:
        for i in range(0, hi, size * 2):
            # Determine the bounds for the next iteration. This essentially splits
            # the array into chunks of size 2 * size, searching through them
            # individually and then merging them all back together. Note that the
            # outer chunk might contain fewer elements than what you'd expect due
            # to rounding errors.
            lo = max((i // (size * 2)), 0)
            hi = min((((i + size * 2) // size) + 1)*size, hi)
            sort(lo, hi)
        size *= 2
    
            
class GifImage:
    SIZE = {
        1 << 1:     (1 << 7,  1 << 5,   1 << 8), 
        1 << 2:     (1 << 13, 1 << 4,   1 << 8), 
        1 << 3:     (1 << 3,  1 << 5,   1 << 8), 
        1 << 4:     (1 << 8,  1 << 4,   1 << 8), 
        1 << 5:     (1 << 12, 1 << 4,   1 << 8), 
        1 << 6:     (1 << 2,  1 << 6,   1 << 8), 
        1 << 7:     (1 << 10, 1 << 3,   1 << 8), 
        1 << 8:     (1 << 4,  1 << 4,   1 << 8), 
        1 << 9:     (1 << 9,  1 << 3,   1 << 8), 
        1 << 10:    (1 << 8,  1 << 3,   1 << 8), 
        1 << 11:    (1 << 7,  1 << 4,   1 << 8), 
        1 << 12:    (1 << 6,  1 << 5,   1 << 8), 
        1 << 13:    (1 << 5,  1 << 6,   1 << 8), 
        1 << 14:    (1 << 14, 1 << 10,  1 << 8), 
        1 << 15:    (1 << 12, 1 << 11,  1 << 8), 
        }

    
    def __init__(self, width, height):
        self.__width = width
        self.__height = height
        self.__data = bytearray()


    @property
    def width(self):
        return self.__width
    

    @property
    def height(self):
        return self.__height


    def set_pixel(self, x, y, r, g