#!/usr/local/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
    
        with open(f'./{self.output}', 'wb') as fh:
            generate_random_gif(open('/dev/urandom', mode='rb'), fh)


# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# correspond to the information found here:
# https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Signature
    sig = b'GIF89a'

    # Screen Descriptor
    w = rng.read(2)    # width is 16 bit
    h = rng.read(2)    # height is 16 bit
    sf = struct.unpack('<B', rng.read(1))[0] & 7 # 3 bits of packed field
    bg = bytes([struct.unpack('>H', rng.read(2))[0] >> 8]) # background color index
    cf = bytes([struct.unpack('>H', rng.read(2))[0] >> 8]) # global color table flag
    cs = bytes([struct.unpack('>H', rng.read(2))[0] >> 8]) # color resolution
    ce = bytes([struct.unpack('>H', rng.read(2))[0] >> 8]) # sort flag
    pd = bytes([struct.unpack('<B', rng.read(1))[0] >> 3]) # pixel aspect ratio
    sd = sig + w + h + sf + bg + cf + cs + ce + pd

    # Global Color Table
    gct = []
    num_colors = 2 << ((sf & 7) + 1)      # Number of colors in global colortable 
    for i in range(num_colors):           # RGB triplets stored as byte values
        r,g,b = rng.read(3)               # each component is a 1-byte value
        gct.append((r,g,b))
    # Palette consists of GCT followed by local color table
    p = [bytes(i) for i in zip(*[iter(gct)]*3)] 

    # Image Descriptor
    left = h = bytes([rng.read(1)[0]])
    top = w = bytes([rng.read(1)[0]])
    flags = bytes([struct.unpack('<B', rng.read(1))[0]])
    id = sig + left + top + flags

    # LZW min code size
    lzwmincodesize = struct.unpack('<B', rng.read(1))[0]
    
    # Actual image data stream
    img = []
    while True:
        s = struct.unpack('<B', rng.read(1))[0] # next chunk size
        marker = rng.read(s+1)                # include delimiter in chunk size
        if marker == b'\x00':                  # EndOfImage Marker
            break                              # end loop
        elif marker == b'\xff\x00':            # Trailer Marker
            continue                           # skip rest of loop body
        else:                                   # Just some data
            img.append(marker[:-1])            # append to list
            
    # Write everything to final buffer
    out.write(sd + bytes(p).join(img))
    