#!/usr/bin/env python3

# A simple program to create random GIF images which can be used for testing

import sys
import os
import hashlib


class BitStream:

    def __init__(self):
        self._buffer = []

    def _bitstream(self, bytestream: bytes):
        bits = bytearray()

        for byte in bytestream:
            bits += [((byte & (0x80 >> i))!= 0)
                     for i in range(7,-1,-1)]
        
        return bits

    @property
    def binary(self):
        result = [b'\x0'] * len(self._buffer[0]) # Prefix with empty byte

        for item in self._buffer:
            assert len(item) == len(result), "All items must have same length"

            result = [(r << 1 | i) for r,i in zip(result, item)]

        return b''.join([bytes([c]) if c else b'' for c in result])
    
    def push(self, bs):
        assert isinstance(bs, list), "Bitstreams should be lists of bools"
        
        self._buffer.append(bs)

def encode_subblock(name, content):
    identifier = {
        'graphic-control':   b'\xF9',
        'image-descriptor':  b'\x2C',
        'application':       b'\xFF',
        'comment':           b'\xFE',
        'extension':         b'\x21',
        'plain text':        b'\x01',
        }[name]

    blocksize = len(content) + 1
    data = bytes([blocksize])+identifier+content

    crc = hashlib.md5()
    crc.update(data[:-1]) # Exclude checksum byte from CRC so as not to include it twice
    crcdigest = int.from_bytes(crc.digest()[0:2], byteorder='little')
    #print("{}: {}".format(crcdigest, hex(crcdigest)))

    return data+struct.pack('<H', crcdigest)
    
def encode_graphics_control():
    delay = 6 # In 1/100ths of a second
    transparentindex = None
    disposalmethod = 1
    userinputflag = False

    colorinfo = bytes([
        0x80 |                   # Global Color Table Flag
        7                       # Bits per pixel minus one
    ])

    content = bytes([
        ((delay&0b111111)>>0),    # Delay time low part
        (((delay&0b11111111)<<7)|((transparentindex&0b1111111)>>0)),    # Delay time high part and Transparent Index low part
        (((delay&0b111111111111)>>0) if delay > 1023 else 0) | \
      (((transparentindex&0b111111111)<<7)|disposalmethod),              # Disposal method and User Input Flag
        (((transparentindex&0b11111111111111)>>0) if transparentindex > 1023 else 0) | \
          (userinputflag<<7) | \
          0b00000000                 # Transparent Color Index low part or Reserved low part
        ]).ljust(6,'\x00')     # Fill rest of packet with zeroes

    return encode_subblock('graphic-control',colorinfo+content)
    

def encode_imageheader(width, height, localcolortablepresent=False):    
    flags = bytes([
        width>>8,            # Width low byte 
        width&0xFF,          # Height low byte 
        0b11100000 |         # Local Color Table | Interlace | Sort | Reserved
        (height>2**16)-1,    # Size bit field
        localcolortablepresent*0b10000000
    ])

    return encode_subblock("image-descriptor",flags)

def encode_localcolortable(palette):
    assert all([(len(t)==3 and type(t) in (list,tuple)) for t in palette]), "Invalid palette entry."
    
    return encode_subblock('graphic-control',b''.join([bytes(t)+b'\x00' for t in palette]))

def encode_plaintext(text=""):
    """Encode an optional plaintext subblock."""

    return encode_subblock('plain text', text.encode()+b'\x00') if text else b''

def encode_application(appcode="", appdata=""):   
    """Encode application extension containing additional metadata"""
    flags = bytes([
        len(appcode)>0,      # Application Identifier present flag
        0,                    # Authentication Code Present flag
        len(appdata)          # Data Length field 
    ])

    return encode_subblock('application', flags+appcode.encode()+appdata) 

def encode_trailer(numframes, loopforever=True):
    flags = bytes([loopforever*0b01000000])

    # Number of frames is stored in little endian format
    numframesbytes = (numframes%256).to_bytes(length=1, byteorder='little') + \
                    (int(numframes/256)%256).to_bytes(length=1, byteorder='little')

    return encode_subblock('graphic-control'+b'\x00'*12, flags+numframesbytes)

def write_file(filename, filenameprefix='', frametime=None, colordepth=(8,8,8), colormap=(), text='', appcode='', appdata='', loopforever=True, fps=24, obeyaspectratio=False, fillwithbackground=False, generate_random_colors=False, genrowsfunc=None, gencolsfunc=None, w=3, h=6, colormask=[0xffffff]):
    """Write an animated GIF image.

    The animation consists of `n` frames. Each frame has dimensions 
    `width x height`. Each pixel has RGB values between 0 and 2^bitdepth - 1.

    Args:
        filename (str): Name of output file. Should end in '.gif'.
        filenameprefix (str, optional): Prefix to prepend to each generated 
            temporary file during animation. Defaults to ''.
        frametime (float, optional): Time interval between frames measured in 
            1/100ths of a second. If unspecified, then defaults to 1/fps.
        colordepth (Tuple[int, int, int], optional): Maximum value per channel 
            e.g. (8,8,8) for RGB where every color component has 8 bits. Must 
            contain three integers only; no error checking on other arguments. 
            Defaults to (8,8,8).
        colormap (List[List[int]], optional): List of colors specified as RBG tuples.
            Colors should be specified as integer triples between 0 and 2^bitdepth - 1. 
            Defaults to [].
        text (str, optional): Text to display in the first non-transparent frame. 
            Defaults to "".
        appcode (str, optional): Application code string. Defaults to "".
        appdata (str, optional): Application data string.Defaults to "".
        loopforever (bool, optional): Whether to repeat the animation forever. 
            Defaults to True.
        fps (int, optional): Frames per second. Defaults to 24. Note that this limit is not enforced.
        obeyaspectratio (bool, optional): Respect aspect ratio, maintaining original aspect ratio, when resizing frames. Defaults to false.
        fillwithbackground (bool, optional): Fill each new frame with the background color before writing pixels. Defaults to False.
        generate_random_colors (bool, optional): Generate a random colormap for each new frame based on the current color depth. Cannot make use of any pregenerated colormaps. Defaults to False.
        genrowsfunc (function, optional): Function which generates all rows in a single frame. Takes two parameters: a list representing a row of pixels, and the index of the row within the frame itself. Returns nothing. If left blank, will fall back on default function. If generating random colors, the default function makes use of this function. Defaults to None.
        gencolsfunc (function, optional): Function which generates all columns in a single frame. Takes two parameters: a list representing a column of pixels, and the index of the column within the frame itself. Returns nothing. If left blank, will fall back on default function. If generating random colors, the default function makes use of this function. Defaults to None.
        w (int, optional): Default width. This does not affect the size of the generated frame. Defaults to 3.
        h (int, optional): Default height. This does not affect the size of the generated frame. Defaults to 6.
        colormask (List[int], optional): Mask to apply after applying colormap. Useful when exporting a series of frames to another file format like PNG. Defaults to [.FFFFFF].
    """

    assert type(w) is int, f"{w} is invalid input for w"
    assert type(h) is int, f"{h} is invalid input for h"
    assert type(fps) is int, f"{fps} is invalid input for fps"
    assert type(obeyaspectratio)