import itertools as itt


# Generates a random GIF file into `out` using the
# random number generator `rng` (or /dev/urandom). The
# file generated is of size at most `size`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, size: int
    ):

    # To generate the correct length of the global color table, we need to know how many different
    # blocks there will be in the animation and the maximum block size. We use that to calculate
    # the maximum length of the global color table.

    # This gives us the max length for all frames except the last one
    frame_ctls = [2 ** i for i in range(3)]
    max_block_len = sum([i - 1 for i in frame_ctls])

    # Now we can build up the entire color table
    ctls = []
    for ctl in itt.chain(*[itt.repeat(x, y)
                           for x, y in zip(frame_ctls[:-1], frame_ctls[1:])]):

        if len(ctls) >= size or ctl == frame_ctls[-1]:
            break
        
        ctls += [ctl]
    
    nframes = size // (max_block_len + 4)
    nblocks = len(ctls)
    nrepeats = ((nblocks // nframes) + 1) * \
               ([nframes+1] + list(range((nblocks % nframes))))[:nframes]
    block_repeats = [*sorted(zip(ctls, nrepeats))]

    # Now we start writing the bytes
    out.write(b"GIF89a")

    write_byte = lambda buf, v: buf.extend(v.to_bytes(length=1,
                                                     byteorder="big",
                                                     signed=False))
    write_short = lambda buf, v: buf.extend(v.to_bytes(length=2,
                                                      byteorder="big",
                                                      signed=False))
    write_long = lambda buf, v: buf.extend(v.to_bytes(length=4,
                                                     byteorder="big",
                                                     signed=False))
    
    # Write the screen descriptor
    buf = bytearray()
    width = rng.randint(64, 512)   # TODO make this more realistic
    height = rng.randint(32, 512)  # TODO ditto
    
    write_byte(buf, ord('4'))        # Color table flag set
    write_short(buf, width)          # Width
    write_short(buf, height)         # Height
    write_byte(buf, 0xf7)            # Global color table size

    # Table background color index
    write_byte(buf, rng.randint(0, 7))  

    # Local color table size
    write_byte(buf, 0)              

    out.buffer.write(buf)           # Screen descriptor plus global color table

    # Escape code
    write_byte(buf, 0x21)           
    write_byte(buf, 0xff)           
    write_byte(buf, 0x0b)           
    write_byte(buf, 0x4e)           
    write_byte(buf, 0x45)           
    write_byte(buf, 0x54)           
    write_byte(buf, 0x53)           
    write_byte(buf, 0x43)
    write_byte(buf, 0x41)           
    write_byte(buf, 0x50)           
    write_byte(buf, 0x45)           

    while True:                     # Loop over blocks of data

        # Block size (minus escape sequence)
        write_byte(buf, rng.randint(0, min(max_block_len,
                                            size-len(buf)-4))) 

        # Block size marker
        write_byte(buf, 0x2c)       
        if not buf:                   # Escape sequence matches nothing
            continue

        if len(buf) >= size:          # Reached size limit; stop
            break

        # No local color table used on first frame
        write_byte(buf, 0xf9)           
       
        # Delay time
        write_short(buf, rng.randint(0, 0xffff))

        # Frame control flags
        write_byte(buf, 0x08)           

        # Block repeats
        repeat = rng.choice(nrepeats)
        write_byte(buf, repeat)    

     
        for _ in range(repeat):

            # Ratio & palette size; these are ignored by browsers so shouldn't matter
            ratiopal = rng.randint(0, 3)
            write_byte(buf, ratiopal << 4 | ratiopal)
            
            # LZW minimum size
            lzwsizemin = rng.randint(0, 32)
            write_byte(buf, lzwsizemin)   

            # First entry in the color table
            write_byte(buf, rng.randint(0, 2**ratiopal-1))   
            if not buf:                
                continue

            # Data subblock count
            datasubcnt = rng.randint(min(lzwsizemin, 2**(ratiopal&~1)),
                                     min(lzwsizemin*2, 2**(ratiopal)+1))
            predatasublen = len(buf)

            
            # Finally the data subblocks themselves
            for _ in range(datasubcnt):

                # This is the data subblock size field itself
                data_subblock_size = rng.randint(0, 2**8-1)
                write_byte(buf, data_subblock_size)
                
                write_byte(buf, rng.randint(0, 255))      # Subblock index
                write_short(buf, rng.randint(0, 2**8-1))    # Repeat delta
                dsb_len = len(buf) - predatasublen
            
                while len(buf) < predatasublen + data_subblock_size:

                    chunk_type = rng.randint(0, 2)
                    if chunk_type == 0:
                        # Raw literal
                        write_byte(buf, rng.randint(0, 2**8-1))
                        
                    elif chunk_type == 1:
                        # Escaped literal
                        write_byte(buf, rng.randint(0, 2**8-1))
                    
                    else:
                        # End subblock
                        continue
                        
                    dsb_len += 1
                    
                
            predatasublen = len(buf)

            # Trailing end marker
            for _ in range(datasubcnt):
                write_byte(buf, 0x00)            
            
        assert(len(buf) <= size)
        
    out.buffer.write(buf)