#!/usr/bin/env python3

import os
import re
import sys
import click
import io
import pathlib
import struct
from typing import Optional


@click.command()
@click.option('-l', '--length', type=int, default=-1, show_default=True,
              help='The length of the output (in bytes). '
                   '-1 means a random length.')
@click.argument('out')
def main(length,
         out):

    if not isinstance(pathlib.Path(out).parent, pathlib.PurePosixPath):
        raise ValueError("Output filename must be relative to current directory")
    
    # Create the output file right away so we can use it elsewhere
    with open(out, "wb") as f:
        pass

    while True:

        rng = io.BytesIO()

        try:
            generate_random_gif(rng, length, f)
        except Exception as e:
            print(f"{e.__class__.__name__}: {str(e)}",
                  file=sys.stderr)
            continue
        else:
            break


BLACK   = '\033[90m'
RED     = '\033[91m'
GREEN   = '\033[92m'
YELLOW  = '\033[93m'
BLUE    = '\033[94m'
MAGENTA = '\033[95m'
CYAN    = '\033[96m'
WHITE   = '\033[97m'
RESET   = '\033[0m'


# Take a string and return an ANSI color code if it is one; otherwise return None
def get_color(s: str) -> Optional[str]:
    match = re.match(r"^([0-9a-fA-F]{6})$", s)
    if match == None:
        return None
    rgb = [tuple(int(c, base=16) for c in chunk)
           for chunk in zip(*[iter(match.group(1))]*2)]
    codes = list((36 * r + 6 * g + b + 16) for r, g, b in rgb)
    return "\033[{}m".format(";".join(map(str,codes)))


# Generate a sequence of colors that are guaranteed never to repeat themselves
def generate_colors():
    seen = set()
    while True:
        new_color = "#" + "".join("{0:02x}".format(ord(b))
                                  for b in os.urandom(3))
        if new_color not in seen:
            yield new_color
            seen.add(new_color)


COLORS = iter(generate_colors())


# Takes a value between 0 and 1 and returns a corresponding color from the spectrum
def colorize(value):
    assert 0 <= value <= 1
    return next(COLORS)
    

# Prints a PNG representation of the image stored at `fp`
def display_image(fp: BinaryIO):

    width, height, data = read_image(fp)
    maxnum = len(data)
    header = '{}{:>6} {}\n'.format(RED, width * height, RESET)
    content = ''.join((colorize(val / maxnum)
                       for val in data)).replace('\n','\n{}'.format(header))
    print('{}{:>{width}}{}\n{}'.format(CYAN,' '*width*height+':'*(width//2+1),
                                      RESET, content, width=width+6))



# Reads an image from `fp`, returning its size and pixel values
def read_image(fp: BinaryIO) -> tuple[int, int, list[float]]:
    
    def split_into_blocks(stream):
        
        def block():

            header = stream.read(4)
            ext = header[:3]
            length = int.from_bytes(
                header[-3:], byteorder="big")
            
            subpackets = b'' if ext!= b'!' else stream.read(length)

            endcode = subpackets[-1:]
            body = subpackets[:-1].decode('ascii')

            return ext, body, endcode
        
        ext, body, endcode = block()
        packets = [ext + body]
        while endcode!= b'\x00':
            ext, body, endcode = block()
            packets.append(body)
            
        return packets
    
    
    blocks = split_into_blocks(fp)

    def parse_headers(lines):

        headers = {}

        for line in lines:
            key, value = map(lambda x: x.strip(),
                             line.split(':'))

            headers[key] = int(value)

        return headers

    def parse_pixels(lines):
    
        pixels = []

        for line in lines:
    
            vals = list(map(lambda x: float(x)/255,
                            line.split()))
        
            pixels += vals
        
        return pixels


    metadata = ""
    data = ""

    for i in range(len(blocks)):

        if blocks[i][0] == b'!':
            metadata += blocks[i][1]
        elif blocks[i][0] == b';':
            data += blocks[i][1]
        else:
            print(f"Unexpected token '{blocks[i][0]}'")
    
        
    headers = parse_headers([line
                              for line in metadata.split("\n")
                              if ":" in line])
        
    #print(headers["interlace"],
    #      headers['bitdepth'],
    #      headers['colormodel'])

    dim_w = headers["width"]
    dim_h = headers["height"]

    if dim_w > 1 or dim_h > 1:
        pixels = [[pixels[i + j*dim_w]
                    for i in range(dim_w)]
                    for j in range(dim_h)]
    else:
        pixels = [pixels]

    
    return (headers['width'],
            headers['height'],
            parse_pixels(data))



# Takes three RGB tuples and returns their HSL equivalent
# Taken from https://gist.github.com/mathebox/5223567
def rgb_to_hsl(R, G, B):
    
    R /= 255
    G /= 255 
    B /= 255 
    
    Cmax = max(R, G, B);  
    Cmin = min(R, G, B)  
    Delta = Cmax - Cmin

    L =.5 * ((Cmax + Cmin) )
    S = 0 if Delta==0 else Delta/(1-abs(.5*(Cmax+Cmin)-1));  

    if Cmax == R :  
        h = (((G-B)/Delta)%6)
    elif Cmax == G : 
        h = ((B-R)/Delta + 2)
    else:     
        h = ((R-G)/Delta + 4)
    h *= 60
      
    return round(h),round(S*100,1),round(((L+1)*100)/2,1)



# Takes a number on [0, 1], representing a shade of red, green or blue
# Returns the corresponding brighter color code
def lighten(rgb, amount=.5):

    if rgb == RED:
        return '\033[91;1m'
    elif rgb == GREEN:
        return '\033[92;1m'
    elif rgb == BLUE:
        return '\033[94;1m'
    else:
        raise ValueError(f"Invalid color {rgb}")



# Takes a number on [0, 1], representing a shade of red, green or blue
# Returns the corresponding darker color code
def darken(rgb, amount=.5):

    if rgb == RED:
        return '\033[91;2m'
    elif rgb == GREEN:
        return '\033[92;2m'
    elif rgb == BLUE:
        return '\033[94;2m'
    else:
        raise ValueError(f"Invalid color {rgb}")




# Generates a random GIF file into `out` using the
# given pseudorandom number generator in ASCII format
def generate_random_gif(
        rng: BinaryIO, length: int, out: BinaryIO
    ):

    # Generation parameters
    numframes = 2 # Number of frames generated by each epoch
    fps = 3        # Framerate per second
    colors = ["red","green","blue"] # Colors used in generation


    frame = 0
    epoch = 0

    while True:

        print(f"\n{epoch}:", file=sys.stderr)
        
        # Top part
        w = rng.read(1)[0] % 32 + 10
        h = rng.read(1)[0] % 32 + 10
        
        duration = ",".join(map(str,[fps]))
        bgcolor = colors[(rng.read(1)[0] // 3) % 3]

        imgdata = dict([(c, "")
                        for c in colors])


        for y in reversed(range(h)):

            for x in range(w):

                red = rng.read(1)[0] % 2**8
                green = rng.read(1)[0] % 2**8
                blue = rng.read(1)[0]