#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, 
        out: BinaryIO,
        version='89a'
    ):


    header = f'GIF{version}'.encode('utf-8')
    out.write(header)

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))
    
    # get remaining lzw compressed info from the RNG
    datalen = int.from_bytes(rng.read(2), 'big')
    data = rng.read(datalen)

    flags = 0b10000000 # global colour table present

    if version == '89a':
        # skip fields that aren't particularly interesting
        # (the logical screen descriptor - color table size, background
        # color index, pixel aspect ratio, global color table flag,
        # color resolution, sort flag, size of global color table.)
        rng.seek(7)

        # size of the colour table
        ctsize = struct.unpack("<B", rng.read(1))[0]

        if ctsize > 0:
            flags |= 0b10000000 # global colour table present
            
            # write the logical screen descriptor
            out.write(struct.pack(
                '<HHBB',
                image_width, 
                image_height, 
                ctsize | flags,
                0))

            # read in palette values (RGB triples)
            paldata = [struct.pack('BBB', *x) for x in
                [(y % 256, y // 256, y // 256 * 256)
                    for y in range(ctsize)]]

            for i in range(len(paldata)):

                # write out the length byte first
                out.write(struct.pack('B', len(paldata[i])))
                out.write(paldata[i])

    elif version == '87a':
        pass 

    else:
        raise Exception("Unknown GIF version!")

    # write to output stream
    out.write(data)