from typing import BinaryIO
import argparse
import io
import os


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, width: int, height: int, frames: int, outf: str
    ):

    # The first four bytes are always 'GIF89a'.
    # This is an animated GIF with no optional fields except for global palette (256 colors).
    header = b'\x47\x49\x46\x38\x39\x61'

    # Bits 5-7 of the flags field indicate how many bits are used to specify the delay time in units of 10ms.
    # We will use 1 ms so that it's easier to parse and visualize.
    # However this means we can only support delays up to ~6 seconds.
    gct_flag, width_flag, height_flag, local_color_table_flag, interlaced_flag, sort_flag, reserved_flags = \
        bin((1 << 7) | ((7 - log2(10)) << 5)).encode().zfill(8)[1:]

    # Iterate over all images in the animation
    for frame in range(frames):

        # Header block
        # Each image is preceded by a control block defining its properties.
        # Since we want this to be an animation, there needs to be at least one more than one frame.
        if frame == 0:
            flags = gct_flag + width_flag + height_flag + b'\x00' + interlaced_flag + sort_flag + reserved_flags
            duration = b'\x00\x00'
            left = b'\x00\x00'
            top = b'\x00\x00'

            width_block = width.to_bytes(2, byteorder='little')    # Width of each frame
            height_block = height.to_bytes(2, byteorder='little')  # Height of each frame
            out.write(
                b'\x2c' + flags + duration + left + top + width_block + height_block
            )
        else:
            # Frames between the first and last need to repeat their contents from the previous frame
            # Same as above but without any pixels.
            out.write(b'\x21\xf9\x04\x01\x00\x00\xff\xff\xff\xff\x00\x2c\x00\x00\x00\x00\x00\x00\x00\x00')


        # Image data blocks come right after. They contain pixel values which encode the indices of the color table.
        # This is not quite how GIF files work, but it works well enough to let us generate some cool images!
        # This loop iterates over every row in our image (which is just a single line since our images are monochromatic).
        num_pixels = width * height // 8
        for i in range(num_pixels // 255 or 1):
            chunksize = min(255, num_pixels % 255)
            out.write((chunksize).to_bytes(1, byteorder='little'))   # Write the length of this chunk
            out.write(rng.read(chunksize))                            # Write the actual pixel data
            num_pixels -= chunksize



if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--width', type=int, default=128, help="Width of the generated images.")
    parser.add_argument('--height', type=int, default=128, help="Height of the generated images.")
    parser.add_argument('nframes', type=int, nargs='?', default=1,
                        help="The number of frames in the resulting GIF file.")


    args = parser.parse_args()
    rng = open('/dev/urandom', 'rb')
    out = io.BytesIO()
    generate_random_gif(rng, args.width, args.height, args.nframes, out)
    rng.close()
    out.seek(0)
    out_file = open('./random-gif.gif', 'wb')
    out_file.write(out.read())
    out_file.close()