#!/usr/bin/env python3

import gzip
import os
import secrets
import time
import zlib
from typing import BinaryIO


# Generates a randomly-generated GIF file in `out`.
def main():
    # Determine output filename based on current time.
    t = time.time()
    st = time.gmtime(t)

    filename = f'{st[0]:0>4d}-{st[1]:0>2d}-{st[2]:0>2d}' \
               '_{st[3]:0>2d}.{st[4]:0>2d}.{st[5]:0>2d}.gif'


    print('Creating new random GIF file:', filename)

    print('\nDo you want to continue?')
    response = input('[y]/n ')

    if not response or response == 'y':

        # Create a temporary file and write a randomly-generated gif there.
        with open('/tmp/' + filename, mode='wb') as tmpfile:
            generate_random_gif(
                rng=secrets.SystemRandom().getrandbits,
                out=tmpfile
            )


        print('\nCompressing...')
        compressed = compress(path='/tmp/' + filename)

        print('\nWriting compressed GIF file...')

        try:
            with open('../public/' + filename, mode='xb+') as outfile:
                outfile.seek(0)

                # Write the compressed data.
                for chunk in compressed:
                    outfile.write(chunk)

                # Move back to beginning so we can truncate the file.
                outfile.seek(0)

                # Truncate to remove any old data that isn't part of this file.
                outfile.truncate()

            os.remove('/tmp/' + filename)

        except FileExistsError:
            print('Could not write compressed file; does it already exist? You probably need to delete it.')



# Compress a single file at path `path`. Returns an iterable over chunks.
def compress(path):
    with open(path, mode='rb') as infile:
        uncompressed = infile.read()

    compressed = zlib.compressobj(level=6)
    compressed_data = compressed.compress(uncompressed)
    compressed_data += compressed.flush(zlib.Z_FINISH)
    yield compressed_data


# Generates a random GIF file into `out` using the
# provided RNG. The size of the generated image is determined by
# the first four bytes read from the RNG (which are discarded).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
        4 * rng.read(1), byteorder='little', signed=False
    ) & 0xffffff
    data = rng.read(datalen)
    out.write(header + data)
