#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'
HEADER_SIZE = 7   # in bytes


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
        
        with open('/dev/urandom', mode='rb') as rng,\
             open(self.output,    mode='wb') as out:

            return generate_random_gif(
                    rng, out
                )



# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# are read from different sources to ensure that they're valid.
def generate_random_gif(rng: BinaryIO, out: BinaryIO) -> None:
    
    # Header
    header = b''
    header += b'GIF89a'            # File signature and version
    print('File Signature:', repr(header[:6]))
    # Color Table size
    ctsize = rng.read(1)[0]*3
    print("Color table size:", ctsize)
    for i in range(ctsize):
        header += bytes([rng.read(1)[0]])
        
    header += bytes([5])           # TODO: what should this be?

    
    # Global Color Table
    gct = []
    for _ in range(2**(header[7]+1)):
        gct.append((rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]))

    # Image Descriptor
    idata = ''
    idata += bytes([ord(b'\x2c')])  # image separator 0x2C
    idata += bytes([4])             # image descriptor size
    
    # Image Left Position
    xpos = randint(0, 65535) // 2
    idata += pack('<H', xpos)       # position on the horizontal scale
    
    # Image Top Position
    ypos = randint(0, 65535) // 2
    idata += pack('<H', ypos)       # position on the vertical scale
    
    # Image Width
    width = randint(1, 65535) // 2
    idata += pack('<H', width)      # width of image
    
    # Image Height
    height = randint(1, 65535) // 2
    idata += pack('<H', height)     # height of image
    
    # Packed Field
    pf = bin(randint(0, 255))[2:]
    pf = '0'*(8 - len(pf)) + pf
    pfx = [False, False, False,
           False, False, True,   # global color table flag
           False, False, True]    # color resolution & sort order
    pfy = ['00','01','10','11']
    pfdata = ''
    for px, py in zip(pfx, pfy): 
        pfdata += py if px else '0'
    pfdata += '0'                # end of packed field byte
    idata += bytes([int(pfdata, base=2)])    
    

    # Image LZW Min Code Size
    lzwmincodesize = rng.read(1)[0]
    idata += bytes([lzwmincodesize])
    
    # Image Data
    imagedata = pack('{}B'.format(width*height), *[randint(0, 255)
                                                      for x in
                                                      range(width*height)])
    zipped = compress(bytes(imagedata)).split(b'\x00')[1:-1]
    nblocks = len(zipped)
    idata += pack('<B', nblocks)                    # Number of Blocks
    compressed = b''
    for z in zipped[:-1]:                           # Compress the image data
        compressed += bytes([(len(z)+127)%256])+z
    # Add the last block
    compressed += bytes([(len(zipped[-1])+127)%256]) + zipped[-1]
    idata += compressed                             # Append compressed data
    
    
    # Write the GIF file!
    out.write(header+gct+idata)
    