#!/usr/bin/env python
"""
This script generates a pseudo-random animated.GIF image
file based on system entropy. It does this by creating one
or two images that are transparent + white backgrounds and
then overlaying them onto a single frame of solid black background.
"""
__author__ = '<NAME>'

# Copyright 2017 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import functools
import io
import itertools
import logging
import math
import os
import subprocess
import sys
import tempfile
import textwrap
from PIL import ImageDraw, ImageFont
from timeit import default_timer as timer
from typing import BinaryIO, Iterable, Optional

logger = logging.getLogger(__name__)
logging.basicConfig()
logger.setLevel(logging.INFO)

# Default font file name, passed to PIL's ImageFont constructor.
FONT_FILE = '/Library/Fonts/Arial Black.ttf'

# Default font color used when rendering the text string.
TEXT_COLOR = '#ffffff'

# Fallback font for drawing the watermark. This includes the text string
# and also a timestamp.
FALLBACK_WATERMARK_FONT = ImageFont.load_default().font_variant(size=40)

# Maximum number of characters from the input string that we will try drawing
# in the watermark. If there are more than this many, they will be randomly
# truncated before being rendered.
MAX_CHARS_IN_WATERMARK = len('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ ')

# Set of valid colors to choose from when adding a watermark to each image.
VALID_COLORS = [
    '#8A2BE2',  # BlueViolet
    '#A52A2A',  # Brown
    '#DEB887',  # BurlyWood
    '#5F9EA0',  # CadetBlue
    '#7FFF00',  # Chartreuse
    '#D2691E',  # Chocolate
    '#FF7F50',  # Coral
    '#6495ED',  # CornflowerBlue
    '#DC143C',  # Crimson
    '#00FFFF',  # Cyan
    '#00008B',  # DarkBlue
    '#008B8B',  # DarkCyan
    '#B8860B',  # DarkGoldenRod
    '#A9A9A9',  # DarkGray
    '#006400',  # DarkGreen
    '#A9A9A9',  # DarkGrey
    '#BDB76B',  # DarkKhaki
    '#8B008B',  # DarkMagenta
    '#556B2F',  # DarkOliveGreen
    '#FF8C00',  # DarkOrange
    '#9932CC',  # DarkOrchid
    '#8B0000',  # DarkRed
    '#E9967A',  # DarkSalmon
    '#8FBC8F',  # DarkSeaGreen
    '#483D8B',  # DarkSlateBlue
    '#2F4F4F',  # DarkSlateGray
    '#2F4F4F',  # DarkSlateGrey
    '#00CED1',  # DarkTurquoise
    '#9400D3',  # DarkViolet
    '#FF1493',  # DeepPink
    '#00BFFF',  # DeepSkyBlue
    '#696969',  # DimGray
    '#696969',  # DimGrey
    '#1E90FF',  # DodgerBlue
    '#B22222',  # FireBrick
    '#FFFAF0',  # FloralWhite
    '#228B22',  # ForestGreen
    '#FF00FF',  # Fuchsia
    '#DCDCDC',  # Gainsboro
    '#F8F8FF',  # GhostWhite
    '#FFD700',  # Gold
    '#DAA520',  # GoldenRod
    '#808080',  # Gray
    '#008000',  # Green
    '#ADFF2F',  # GreenYellow
    '#808080',  # Grey
    '#F0FFF0',  # HoneyDew
    '#FF69B4',  # HotPink
    '#CD5C5C',  # IndianRed
    '#4B0082',  # Indigo
    '#FFFFF0',  # IVory
    '#F0E68C',  # Khaki
    '#E6E6FA',  # Lavender
    '#FFF0F5',  # LavenderBlush
    '#7CFC00',  # LawnGreen
    '#FFFACD',  # LemonChiffon
    '#ADD8E6',  # LightBlue
    '#F08080',  # LightCoral
    '#E0FFFF',  # LightCyan
    '#FAFAD2',  # LightGoldenRodYellow
    '#D3D3D3',  # LightGray
    '#90EE90',  # LightGreen
    '#D3D3D3',  # LightGrey
    '#FFB6C1',  # LightPink
    '#FFA07A',  # LightSalmon
    '#20B2AA',  # LightSeaGreen
    '#87CEFA',  # LightSkyBlue
    '#778899',  # LightSlateGray
    '#778899',  # LightSlateGrey
    '#B0C4DE',  # LightSteelBlue
    '#FFFFE0',  # LightYellow
    '#00FF00',  # Lime
    '#32CD32',  # LimeGreen
    '#FAF0E6',  # Linen
    '#FF00FF',  # Magenta
    '#800000',  # Maroon
    '#66CDAA',  # MediumAquaMarine
    '#0000CD',  # MediumBlue
    '#BA55D3',  # MediumOrchid
    '#9370DB',  # MediumPurple
    '#3CB371',  # MediumSeaGreen
    '#7B68EE',  # MediumSlateBlue
    '#00FA9A',  # MediumSpringGreen
    '#48D1CC',  # MediumTurquoise
    '#C71585',  # MediumVioletRed
    '#191970',  # MidnightBlue
    '#F5FFFA',  # MintCream
    '#FFE4E1',  # MistyRose
    '#FFE4B5',  # Moccasin
    '#FFDEAD',  # NavajoWhite
    '#000080',  # Navy
    '#FDF5E6',  # OldLace
    '#808000',  # Olive
    '#6B8E23',  # OliveDrab
    '#FFA500',  # Orange
    '#FF4500',  # OrangeRed
    '#DA70D6',  # Orchid
    '#EEE8AA',  # PaleGoldenRod
    '#98FB98',  # PaleGreen
    '#AFEEEE',  # PaleTurquoise
    '#D87093',  # PaleVioletRed
    '#FFEFD5',  # PapayaWhip
    '#FFDAB9',  # PeachPuff
    '#CD853F',  # Peru
    '#FFC0CB',  # Pink
    '#DDA0DD',  # Plum
    '#B0E0E6',  # PowderBlue
    '#800080',  # Purple
    '#FF0000',  # Red
    '#BC8F8F',  # RosyBrown
    '#4169E1',  # RoyalBlue
    '#8B4513',  # SaddleBrown
    '#FA8072',  # Salmon
    '#F4A460',  # SandyBrown
    '#2E8B57',  # SeaGreen
    '#FFF5EE',  # Seashell
    '#A0522D',  # Sienna
    '#C0C0C0',  # Silver
    '#87CEEB',  # SkyBlue