import struct
from typing import BinaryIO


# A random GIF converter that uses the LCG PRNG to choose bits and bytes at
# random from its output stream
class RandomGifConverter():

    # The maximum length of an image is about 16 MB, so we'll keep our images
    # below this threshold. (For a real GIF it's more like ~25MB.)
    # We're not going for anything too small because it takes longer to ensure
    # we have enough bits for our width and height when deciding on the color
    # table size, as well as the time to actually draw them onto the canvas
    MAX_DATALEN = 16 * 10**6

    def __init__(self,
                 rng: BinaryIO):

        self._rng = rng

        # Generate the header and global color table
        self._header()


    # Create the initial set of state variables necessary for generating
    # a valid GIF file from our input RNG
    def _header(self):
        self._out = bytearray()
        self._out += self._generate_signature()
        self._bgindex = self._generate_color_table()
        self._gcindex = self._bgindex
        self._imagecnt = 0


    @property
    def outstream(self) -> bytearray:
        return self._out


    # Generate the signature field, which should be the string "GIF87a" or
    # similar. This doesn't need to worry about supporting other variants, since
    # those don't produce valid files anyway.
    def _generate_signature(self) -> bytes:
        ver = self._rng.read(1)[0]
        if ver == 73:
            sig = b'GIF87a'
        elif ver == 76: 
            sig = b'GIF89a'
        else:
            raise ValueError('Invalid GIF version {}'.format(ver))

        return sig


    # Generate the global color table for the entire animation, returning its
    # index inside the color table.
    #
    # Since this will become the first entry in any frame-local color tables we
    # create later, we can use this as a default background color as well.
    def _generate_color_table(self) -> int:
        color_entries = min(
            2**(int.from_bytes(
                self._rng.read(1), 'big', signed=False) + 1), 256)

        ctdata = bytearray([*range(color_entries)])
        ctdata[0] |= 0b11000000

        self._out += struct.pack('B', len(ctdata))
        self._out += ctdata

        return 0


    # Generate a single frame using the current global color table
    def _generate_frame(self):
        self._imagecnt += 1

        # Choose what color table index to use for the local palette
        self._lcindex = self._generate_color_table()

        # Use the same values between frames unless told otherwise by the caller
        delay    = self._rng.read(2)
        disposal = self._rng.read(1)[0] % 8
        size     = self._calculate_size()
        self._draw_image(delay, disposal, size)


    # Convenience function used to calculate how many pixels to allocate for each
    # dimension
    def _calculate_size(self) -> tuple:
        w = None
        h = None

        while True:
            if w is not None and h is not None and \
               w <= self.MAX_DATALEN and h <= self.MAX_DATALEN:

                break

            tmpw = max(self._rng.read(2), 1)
            if w!= tmpw:
                w = tmpw

            tmph = max(self._rng.read(2), 1)
            if h!= tmph:
                h = tmph

        return ((w, h))
        

    # Draws one frame worth of content onto the canvas
    def _draw_image(self,
                    delay: bytes,
                    disposal: int,
                    size: tuple):

        # Set up the base data for all subsequent drawing commands, including 
        # the resolution fields itself
        self._out += struct.pack('BBBBB',
                                 size[0], size[1],
                                 self._gcindex,      # Background color
                                 self._lcindex,       # Default drawing color
                                 0)                   # LZW minimum code size
        
        # Now add the actual drawing commands. These consist of either literal
        # commands for every pixel (if color count > 2^n) or commands referring
        # back to previous ones to simulate longer runs of repeating ones.
        # With some probability, we might also do a copy command instead.
        # Note that there are some special cases where certain combinations of
        # these commands would result in a larger output than desired due to
        # differing semantics with respect to the way they interact with the
        # control flow operator (clearly the most intense case). However, the
        # resulting effect is still visually pleasing and I doubt it matters
        # much.
        for y in range(size[1]):
            for x in range(size[0]):
                if bool(self._rng.read(1)):
                    self._out += self._generate_copy(y, x)
                else:
                    self._out += self._generate_literal(y, x)


        self._out += struct.pack('<HBB',
                                delay,         # Frame delay
                                disposal | 0b1000,          # Disposal method
                                0)             # Transparent color index



    # Generate a series of copy commands until we reach an end condition such
    # as a different value in the destination buffer or a different color in
    # the source buffer
    def _generate_copy(self,
                       y: int,
                       x: int) -> bytes:

        # If we've already drawn two identical rows without seeing a difference
        # then stop copying
        if y - 2 >= 0 and \
           self._getpixel(y - 2, x) == self._getpixel(y, x):
            
            return b''

        deltay = 0
        deltax = 0

        # Keep track of the current run so far
        currrunlen = [0]
        currcolor = self._getpixel(y, x)

        # Try advancing the cursor forward until we hit a different color or
        # another edge condition
        while True:
            newcolor = self._getpixel(y + deltay, x + deltax)
            if newcolor!= currcolor or deltay > 2 ** 10 or deltax > 2 ** 10:
                
                break

            currrunlen[0] += 1
            deltay += 1
            deltax += 1

        # We may have generated a zero-length run here if we ended on a
        # different color or went over the maximum window size; just encoding
        # empty sequences gets us pretty close
        currrunlen[0] -= 1

        cmdbuf = bytearray()
        cmdbuf += struct.pack('BB',
                              (deltax >> 8) | 0b11000000,  # Horizontal offset
                              (deltay >> 8) | 0b11000000   # Vertical offset
                             )

        cmdbuf[0] |= (currrunlen[0] - 1)        # Run length
        cmdbuf[0] |= (newcolor << 4)            # New color index

        return cmdbuf


    # Generate a series of literal commands until we reach an end condition such
    # as a different color or a different color in the source buffer
    def _generate_literal(self,
                          y: int,
                          x: int) -> bytes:
        
        cmdbuf = bytearray()

        currbyte = 0
        for i in range(self._lcindex + 1):
            currbyte <<= 1
            currbyte |= bool(self._rng.read(1))

        while currbyte < 2 ** (self._lcindex + 1):
            cmdbuf += struct.pack('B', currbyte)

            prevcolor = currbyte
            currbyte = 0
            for i in range(self._lcindex + 1):
                currbyte <<= 1
                currbyte |= bool(self._rng.read(1))

            if prevcolor == currbyte:

                return cmdbuf[:-1]
    
        # In theory this shouldn't happen, but better safe than sorry
        print('[!] No termination sequence found! (try again)')
        exit(-1)


    # Return the color identifier stored at the specified position in the 
    # global color table
    def _getpixel(self,
                  y: int,
                  x: int) -> int:

        return self._bgindex



# Main entry point, normally invoked through the'main' block defined above
def main(*argv, 
         stdin: BinaryIO = open('/dev/urandom'),
         stdout: BinaryIO = open('rnd.gif', 'wb')):
    

    converter = RandomGifConverter(stdin)
    
    try:
        while True:
            converter._generate_frame()
            
    except KeyboardInterrupt:

        pass

    stdout.buffer.write(converter.outstream)
    stdout.flush()