#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# If `nolclr`, don't use a color table.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, nolclr: bool=False
    ):

    out.write(b'GIF89a')

    # width and height are 16 bits each
    w, h = struct.unpack('<HH', rng.read(4))

    # Skip the next few parts of the header since they're not particularly
    # useful right now; we'll come back to them later...
    rng.seek(7)

    # Get the size of our global color table, or zero if there isn't one
    ctsize = int.from_bytes(rng.read(1), 'little')
    if not nolclr and ctsize > 0:

        # The way we represent this is as a list of ints that's padded with zeros
        ctpadd = lambda x: ((x >> 3) + 1) << 3
        gclrtbl = [int.from_bytes(rng.read(1), 'little') for _ in range(ctpadd(ctsize))]

        # Convert these numbers to RGB tuples then to bytes
        rgbconv = lambda x: (((x >> 1 & 0x1F) * 255) >> 3,
                             ((x >> 10 & 0x1F) * 255) >> 3,
                              ((x >> 19 & 0x1F) * 255) >> 3)
        gclrtbp = [struct.pack('BBB', *rgbconv(x))
                   for x in gclrtbl[:min(ctsize*3, len(gclrtbl))]]


        # Write the global color table size, background color index, pixels per
        # unit horizontal and vertical resolution, flags, size of global color
        # table, and finally the actual global color table itself
        out.write(struct.pack('<BBBBHHB',
            ctsize | 0b10000000,
            0, 0, 0, 1<<5, ctsize, *gclrtbp))


    else:
        # no palette
        out.write(struct.pack('<HH', w, h))



    # Get remaining lzw compressed info from the RNG
    compressor = LZWCompressor()

    while True:

        try:

            # Read enough data to fill up an entire block
            data = rng.read(compressor.block_length())

            # Compress it all, then write it out
            out.write(compressor.compress(data))

        except LZWFinishedException: break

