#!/usr/bin/env python3

import sys
import os
import io
import struct
from typing import BinaryIO



DEFAULT_OUTPUT = 'random.gif'
BUFSIZE = 4096


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
    
        with open(f'./{self.output}', 'wb') as fh:
            generate_random_gif(io.BytesIO(), fh)


# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# correspond to the information found here:
# https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Signature
    out.write('G'.encode())
    out.write('I'.encode())
    out.write('F'.encode())
    out.write('8'.encode())
    out.write('9'.encode())
    out.write('a'.encode())

    # Width and Height
    w = rng.read(2)   # width is 16 bit
    h = rng.read(2)   # height is 16 bit
    out.write(w)      # width leftmost byte first
    out.write(h)      # height next

    # Color Table size
    ctsize = rng.read(1)[0]       # color table size byte
    ctsz = ctsize * 3             # color table size in bytes
    n_colors = int((ctsz + 7) / 8) # colors needed in this table
    colorbuf = [0] * n_colors
    for i in reversed(range(n_colors)):
        colorbuf[i] = rng.read(1)[0] << ((i % 8) * 8)
    for b in colorbuf[:ctsz]:
        out.write(struct.pack('<H', b))

    # Blocks
    while True:
        
        r = rng.read(1)[0]           # read a single byte from RNG
        if r == 0x00:                 # check if at end of 
            break                      # image data

        bs = rng.read(r)[0]            # retrieve the block size
        out.write(bytes([bs]))         # add it to the header
        subblock = rng.read(bs - 1)    # discard sub-block marker
        data = rng.read(bs)            # get actual image data
        out.write(data)                # write it all together
        