#!/usr/bin/env python3

import os
import random
import struct
from datetime import timedelta
from timeit import default_timer as timer

from tqdm import trange

# Generates a random GIF file into `out` using the
# random number generator `rng`, which is either `/dev/urandom`, or
# an RNG from urandom-ctypes which is best for CPython
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bytes:

    header = b'GIF89a'
    datalen = random.randrange(1 << 14)
    data = rng.read(datalen)
    crc = _crc32(data) & 0xffff
    out.write(header + datalen.to_bytes(3, 'big'))
    out.write(data)
    return crc


# Generates a random JPEG file into `out` using the
# random number generator `rng`.
def generate_random_jpeg(
        rng: BinaryIO, out: BinaryIO,
        blocksize: int=256
    ) -> bytes:
    
    header = b'\xff\xe0\x00'
    datalen = (random.randrange((1<<16)-blocksize)) % ((1<<16)-len(header))
    padding = [b'\x00'] * blocksize
    out.write(header + datalen.to_bytes(3, 'big'))
    crc = 0
    while byte := rng.read(blocksize):
        if blocksize - len(byte) > 0:
            # Pad short blocks with zeroes
            out.write(byte + padding[0])
        else:
            out.write(byte[:blocksize])
        crc ^= _crc32(byte)
    return crc
    

# Returns the CRC-32 of `data` using the algorithm from
# https://www.zlib.net/manual.html#How-to-compute-the-adler-checksum
def _crc32(data: bytes) -> int:
    
    POLYNOMIAL = 0xedb88320
    crc = 0xffffffff
    for char in data:
        crc ^= char
        for i in range(8):
            if (crc & 1):
                crc = (crc >> 1) ^ POLYNOMIAL
            else:
                crc >>= 1
    return ~crc & 0xffffffff


if __name__ == '__main__':
    n = 7*10**7
    rands = open('/dev/urandom', mode='rb')
    gifs = open('random_gifs.bin', mode='wb+')
    jpegs = open('random_jpegs.bin', mode='wb+')
    
    start = timer()
    crcs = []
    for i in trange(n):
        crcs.append(generate_random_gif(rands, gifs))
    end = timer()

    print('Total Time:', timedelta(seconds=(end-start)))
    print('CRC32:', sum(hex(c)[2:] for c in crcs))