#!/usr/bin/env python3

import struct
from typing import BinaryIO, cast

# Generates a random GIF file into `out` using the
# random number generator `rng` (typically /dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # We'll be writing to this in-memory buffer as we go along
    buf = bytearray()

    def write(data):
        nonlocal buf

        # Write from the start of the buffer and increase it by how much is written
        buf[:len(data)] = data
        return len(data)

    width    : int   = cast(int, struct.unpack('>H', rng.read(2))[0]) % 65536
    height   : int   = cast(int, struct.unpack('>H', rng.read(2))[0]) % 65536
    bitdepth : int   = cast(int, ord(rng.read(1))) & 7
    palettebytes = 1 << (bitdepth+1)

    # Flags variable contains our flags for the image descriptor that we fill in later
    flags = cast(int, ord(rng.read(1)))

    # TODO: Handle color table size > 256
    # The first byte tells us if there's a global color table or not, and what the maximum color index actually is.
    # If the first byte has its highest bit set, then there's no global color table. Only use local ones per frame.
    # Otherwise, read the next byte and everything up until the end of the block represents the lenghth of the color table.
    has_global_color_table = True
    global_color_table_size = max((cast(int, ord(rng.read(1))), 1))
    print(f"Global color table size {global_color_table_size}")
    assert global_color_table_size <= 256, "TODO"


    # Frame control information
    #: Specifies the position of the left corner of the current frame with respect to the canvas origin (top left).
    x_position = cast(int, ord(rng.read(1)))
    y_position = cast(int, ord(rng.read(1)))
    #: Width of the current frame in pixels; must not exceed the image width specified in the logical screen descriptor.
    width = cast(int, ord(rng.read(1))) 
    #: Height of the current frame in pixels; must not exceed the image height specified in the logical screen descriptor.
    height = cast(int, ord(rng.read(1)))
    #: Is the aspect ratio given explicitly?
    #: - If false, interpret both ``width`` and ``height`` as pixel counts.
    #: - If true, compute an aspect ratio from these two values, and make sure your choice of width and height can accommodate it.
    #:
    #: Note that the aspect ratio only affects scaling, so if you have a 3x3 grid of pixels which should all be blue but one is green, you do not need to say it is a 3x3 grid. You just have to make sure the aspect ratio really makes sense.
    #:
    #: Hence why I picked a super weird value here.
    aspect_ratio_explicitly_given = bool(ord(rng.read(1)))
    assert aspect_ratio_explicitly_given == False, 'TODO'
    
    #: Free for user defined bits
    background_color_index = ord(rng.read(1))
    transparent_color_index = ord(rng.read(1))

    #: Should this frame be interlaced rather than progressive?
    is_interlaced = bool(ord(rng.read(1)))

    #: Has a local color table been used? If so, what's the size of the table?
    local_color_table_exists = bool(ord(rng.read(1)))
    local_color_table_size = max((cast(int, ord(rng.read(1))), 1))

    # TODO: handle more complicated things, like disposal methods etc...

    # Make sure that the following section doesn't fit inside a single frame -- since we're doing this randomly, we don't know exactly where we'll land,
    # so make sure we always write at least part of it after each line.
    num_lines_in_image = (height + local_color_table_size)//local_color_table_size
    while num_lines_in_image >= 0:
        for i in range(num_lines_in_image*width):
            buf += bytes([ord(rng.read(1))])
        num_lines_in_image -= 1

