#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file with the given
# dimensions and frame delay into `out` using the
# given pseudorandom number generator in binary format
def generate_random_gif(
        rng: BinaryIO, width: int, height: int, fps: float, out: BinaryIO
    ):

    # Create image palette required by the spec (256 colors).
    palette = []
    for i in range(256):
        r, g, b = [int((rng.read(1)[0] / 255)) * c
                   for c in [0xFF, 0xFF, 0xFF]]
        palette += [r, g, b]

    # Calculate delay time as milliseconds from frames per second
    ms = round(1000 / fps)

    # Write headers. Note that we don't really care about the background color
    # or pixel aspect ratio so use some arbitrary values here.
    header = {
       'signature':    b'GIF89a',
        'width':        width % 0xFFFF,
        'height':       height % 0xFFFF,
        'flags':        0x87,      # RGB colors, sort colors, interlaced
        'palette_size': len(palette),
        'background':   0,
        'pixel_aspect': 0,

        # Frames start after the entire header has been written
        'frames': [],
    }
    out.write(struct.pack('<6sHHBBBHBH', **header))

    # Write the global palette at offset 0x10
    out.write(bytearray(palette))

    # Generate each of the frames
    for y in range(height):
        for x in range(width // 8):
            lsb = rng.read(1)

            for bit, px in enumerate(reversed([bool(b << bit)
                                              for b in reversed(lsb)])):
                if not px:
                    continue

                pxlen = rng.read(1)[0]
                pxdelay = rng.read(1)[0]
                pxcolor = rng.read(1)[0]
                pxdata = rng.read(pxlen)

                # The minimum length is 2 bytes (00) but this breaks our
                # current parser which expects a packet to always be at least 4
                # bytes long anyway so let's just make it 4 then pad later.
                size = 4 - ((pxlen + 1) % 4)
                if size == 4:
                    size = 0

                # Build payload manually since we're building raw bytes
                data = struct.pack('BBBxB',
                                   pxlen + size,
                                   pxdelay,
                                   pxcolor,
                                   ) + pxdata + b'\0' * size

                # Pad to next 4-byte boundary
                padding = (-len(data)) % 4
                data += b'\0' * padding

                data = struct.pack('<HBB', len(data), 0xF9, 0x04) + data

                out.write(data)

        # Write a separator between each row of blocks
        out.write(b'\0\xf9\x00')

