#!/usr/bin/env python3

import io
import sys
import struct
from typing import BinaryIO


# Generate a random GIF file with width and height between the given
# minimums and maximums with the color depth in bits per pixel. All values
# are inclusive!
def random_gif_bytes(
        wmin: int, hmin: int, wmax: int, hmax: int, cpbp: int
    ) -> bytes:
    # 8-bit color table size
    ctsize = 2**(cpbp//8)-1

    # Header for a single frame
    hdr = [b'GIF89a']

    # Logical screen descriptor
    hdr += [struct.pack('<6BHHBBB', *(7, 0, 0, wmax, hmax, 0x00, 0xf7))]

    # Global color table
    hdr += [ctsize*b'\xff\xff\xff\x00']

    # Extension block
    ext = [
        b';', b'!', b'[', b'F', b'9', b'L', b'C', b'=', b'1',
        b',', b'1', b',', b'2', b',', b'0', b',', b';',
        ]
    if wmin < 32 or hmin < 32:
        minsize = max((wmin, hmin)) % 32
        if wmin <= wmax:
            ext[8] = str(minsize).encode('ascii')
        elif hmin <= hmax:
            ext[9] = str(minsize).encode('ascii')
    hdr += [struct.pack('<'+str(len(ext)+1)+'BB',
                         len(ext)//256, len(ext)%256)]
    hdr += ext

    # Image descriptor
    imgdesc = [
        b',', b',', b',', b'.', b'', b',', b',', b',', b'', b',', b',', b'3'
    ]
    if wmin <= wmax:
        imgdesc[6] = str(wmin%256).encode('ascii')
        imgdesc[7] = str(wmin//256).encode('ascii')
    else:
        imgdesc[6] = str(hmin%256).encode('ascii')
        imgdesc[7] = str(hmin//256).encode('ascii')
    if hmin <= hmax:
        imgdesc[8] = str(hmax%256).encode('ascii')
        imgdesc[9] = str(hmax//256).encode('ascii')
    else:
        imgdesc[8] = str(wmax%256).encode('ascii')
        imgdesc[9] = str(wmax//256).encode('ascii')

    # Finish the image descriptor
    hdr += [struct.pack('<'+str(len(imgdesc))+ 'BB',
                        len(imgdesc)//256, len(imgdesc)%256)]
    hdr += imgdesc

    # Determine where to begin writing the image data
    hdrlen = sum([len(b) for b in hdr]) - 1
    startdatapos = hdrlen + 10
    enddatadirpos = startdatapos + len(ctsize)

    # Write the palette data at this position in the output stream
    hdr[-1] = struct.pack('<H', startdatapos)

    # Repeat the last color table as many times as necessary
    dat = ctsize*(ctsize*[b'\xff\xff\xff'])

    # Copy some more bytes from the input until we hit EOF
    while True:
        buf = sys.stdin.buffer.read()
        if not buf: break
        dat += buf[:enddatadirpos-(hdrlen+10)]

    return b''.join(hdr) + dat


if __name__ == '__main__':
    # Read parameters from stdin
    parambuf = io.BytesIO(sys.stdin.buffer.read())

    # Get the desired width, height, and bit depth
    width = int.from_bytes(parambuf.read(2), byteorder='big', signed=False)
    height = int.from_bytes(parambuf.read(2), byteorder='big', signed=False)
    coldepth = int.from_bytes(parambuf.read(1), byteorder='big', signed=False)

    # Generate the GIF
    print(random_gif_bytes(width, height,
                           width, height, coldepth),
          flush=True)