# -*- coding: utf8 -*-

"""
    This module contains a generator class that can create a
    completely random GIF file without any input. The randomness
    comes from `/dev/urandom`, however this class should be able 
    to handle any other source of entropy too.
"""

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
    
        with open('/dev/urandom', mode='rb') as rng,\
             open(self.output,    mode='wb') as out:
    
            return generate_random_gif(
                    rng, out
                )



# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# are read from different sources to ensure that they're valid.
def generate_random_gif(rng: BinaryIO, out: BinaryIO) -> None:
    
    # Header
    header = b''
    header += b'GIF89a'            # File signature and version

    # Color Table size
    color_table_size = int.from_bytes(
            rng.read(1), byteorder='big'
        ) * 3
    header += color_table_size.to_bytes(
        1, byteorder='big'
    )
    for i in range(color_table_size):
        header += rng.read(1)   # discard table entries
    
    # Global Image Descriptor
    global_image_descriptor = b''
    global_image_descriptor += bytes([
    	int.from_bytes(
    	        rng.read(1), byteorder='big'
	        ) & ~(0b11 << 6) | 0x40                 # Flags
	    ])                                             # Disposal Method
    global_image_descriptor += rng.read(3)               # Delay
    global_image_descriptor += bytes([                     # Transparent Color Index
		int.from_bytes(
			rng.read(1), byteorder='big'
		) & ~0b1111                                               # Flags
	])                                                         # Transparent Color Index
    global_image_descriptor += bytes([                      # Left Position
    	int.from_bytes(
    	        rng.read(1), byteorder='big'
	        ) & ~0b1111111                                          # Left Position
    ])                                                          # Left Position
    global_image_descriptor += bytes([                      # Top Position
    	int.from_bytes(
    	        rng.read(1), byteorder='big'
	        ) & ~0b1111111                                          # Top Position
    ])                                                          # Top Position
    global_image_descriptor += bytes([                      # Width
    	int.from_bytes(
    	        rng.read(1), byteorder='big'
	        ) & ~0b1111111                                          # Width
    ])                                                          # Width
    global_image_descriptor += bytes([                      # Height
    	int.from_bytes(
    	        rng.read(1), byteorder='big'
	        ) & ~0b1111111                                          # Height
    ])                                                          # Height

    # Block Size (Data Sub-Block within Data Stream)
    bs = int.from_bytes(
            rng.read(1), byteorder='big'
        )                                                # block size is an unsigned char
    bs -= 2                                              # minus two more
    subblock = rng.read(bs - 1)                          # discard sub-block marker
    bs += 1                                              # increment back again
    global_image_descriptor += bs.to_bytes(             # add block size back again
    	length=1, byteorder='big'
    )
    data = rng.read(bs)                                  # get actual image data

    # Frames
    framecount = rng.read(1)[0] % 5                       # arbitrary 5 frames max
    for f in range(framecount):                           # for each frame...
        # Clear Code
        frame = b''
        frame += rng.read(1)                              # Block Size
        frame += bytes([                                     # Flags
	        int.from_bytes(
	                rng.read(1), byteorder='big'
	            ) & ~(0b1100 << 4) | 0b0011                 # Local Color Table Flag
	        ])
        frame += bytes([                                     # Size
	        1 << int.from_bytes(
	                rng.read(1), byteorder='big'
	            )                                         # LZW Minimum Code Size
	        ])
        while True:                                        # get next code
            x = rng.read(1)[0]                             # read one byte
            if ((x >> 7) & 1) == 0:                        # last code?
                break                                      # exit loop
            elif (((x >> 7) ^ (x >> 6)) & 1)!= 1:           # only 9 bits used!
                raise RuntimeError('Invalid LZW code!')      # error!
            frame += bytes([(x >> 2)])                      # first part
            frame += bytes([                                 # second part
	            (x >> 4) | ((x & 0b11) << 4)
	        ])
        data = frame + data                                # concatenate everything

    out.write(header + global_image_descriptor + data)     # write it all together