#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# provided RNG. The size of the generated image is determined by
# the first four bytes read from the RNG (which are discarded).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header color table index. We're using this to decide whether or not we
    # should emit an application extension block.
    header_cti = next(rng) % 256

    # Application extension block if necessary. This serves two purposes:
    #   1. Add some entropy so that the generator doesn't always generate the same thing
    #      when given identical inputs.
    #   2. "Compress" the original size field in the header to something smaller than it
    #      would have been otherwise without the block.
    data = None
    extb = None
    if header_cti == 0x01:
        extb = b'\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00' + \
               next(rng).to_bytes(2, 'big')
        data = rng.read()

    elif header_cti!= 0xf7:
        data = rng.read(int.from_bytes(
            4 * rng.read(1), byteorder='little', signed=False
        ))

        extblen = next(rng) % max((data.__sizeof__() // 10), 1)
        extb = b'\x21\xff\x0b' + \
            rng.read(extblen) + \
                b'\x00' + rng.read(2) + b'\x03\x01\x00\x00' + \
                    rng.read(1) + rng.read(1)

    header = b''
    header += b'GIF89a'
    header += len(extb).to_bytes(1, 'big')
    header += b'\x00'
    header += b'!'
    header += len(data).__str__().encode('ascii')
    header += b','
    header += extb
    header += b';'
    header += len(data).to_bytes(2, 'little')
    header += b'\x00'

    out.write(header)
    out.write(data)
