#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# Image width and height are limited to be at most 65k x 65k pixels.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Header.
    out.write(b'GIF89a')

    # Logical screen descriptor (see GIF89a spec for details.)
    out.write(struct.pack('>HH', 0xFFFF & randint(0x100, 0x7FFF), 0xFFFF & randint(0x100, 0x7FFF)))
    color_depth = randint(1, 8)
    out.write((color_depth << 4 | color_depth).to_bytes(1, byteorder='big'))
    set_bits = {
        1: 0b00000,
        2: 0b00001,
        3: 0b00011,
        4: 0b00111,
        5: 0b01111,
        6: 0b11111,
        7: 0b11111,
        8: 0b11111,
    }[color_depth]
    out.write(set_bits.to_bytes(1, byteorder='big'))
    out.write(b'\0\0')

    def write_block():
        blockid = randint(0b0000, 0b1111)

        if blockid == 0b0000:
            numfields = randint(0, 255)
            fields = []
            while len(fields) < int(numfields):
                fieldcode = randint(0b0000, 0b1111)

                # Start of image block.
                if fieldcode == 0b0000:
                    fields += [randint(0, 0xFFFFFFFF)] * min([3, int(numfields)])
                elif fieldcode in {
                        0b0001, 0b0010, 0b0011,
                        0b0100, 0b0101, 0b0110,
                        0b0111, 0b1000, 0b1001,
                        }:
                    pass    # Don't need these.
                else:   # 0b1110 or 0b1111
                    fields += [randint(0, 0xFFFFFFFF)]

            assert len(fields) <= 255
            out.write(((len(fields)+1).to_bytes(1, 'big')+bytes(fields)) + bytes(256-len(fields)-1))

        elif blockid == 0b0001:
            subblocks = []
            while True:
                havesubblocklen = bool(randint(0, 1))
                if not havesubblocklen: break
                subblocklen = randint(0, 0xFFF)
                subblocks += [rng.read(subblocklen)]
            out.write(bytes(len(subblocks))+b''.join(subblocks))

        elif blockid >= 0b0010 and blockid <= 0b0111:
            subblocks = []
            while True:
                havesubblocklen = bool(randint(0, 1))
                if not havesubblocklen: break
                subblocklen = randint(0, 0xFFFF)
                subblocks += [rng.read(subblocklen)]
            out.write(bytes([(blockid << 4|len(subblocks)).to_bytes(1, 'big')]) + b''.join(subblocks))
        
        elif blockid == 0b1010:
            numcolors = randint(0x100, 0xFFFFFF)
            palette = [(randint(0, 0xFF), randint(0, 0xFF), randint(0, 0xFF)) for _ in range(numcolors)]
            out.write(bytes([(numcolors >> 16) & 0xFF,
                             (numcolors >>  8) & 0xFF,
                             (numcolors      ) & 0xFF])+
                     b''.join(bytes(c) for c in palette))

        # We don't care about anything else right now.
        elif blockid!= 0b1011:
            pass

    for i in range(65*65):
        write_block()