#!/usr/bin/env python3

import io
import os
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom())
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    def read_be16() -> bytes:
        return rng.read(2)

    def read_le16() -> bytes:
        return rng.read(2)[::-1]

    def read_le32() -> bytes:
        return rng.read(4)[::-1]


    version = "GIF89a"
    width   = read_le16()[1] & 0xFF | ((read_le16()[0]) & 0xFF) << 8
    height  = read_le16()[1] & 0xFF | ((read_le16()[0]) & 0xFF) << 8

    background = rng.read(3)
    foreground = rng.read(3)
    framecount = rng.read(1)[0] & 0xFF
    disposal   = rng.read(1)[0] & 0xFF
    _          = rng.read(1)[0] & 0xFF

    # Write header
    out.write((version + "\x00").encode('utf-8'))
    # Write logical screen descriptor
    out.write(width.to_bytes(2, 'little', signed=False))
    out.write(height.to_bytes(2, 'little', signed=False))
    out.write((0xF7).to_bytes(1, 'little', signed=False))
    # Write global color table
    for i in range(0, 8):
        out.write(((background[i] & 0xFF)).to_bytes(1, 'little', signed=False))
    for i in range(0, 8):
        out.write(((foreground[i] & 0xFF)).to_bytes(1, 'little', signed=False))
    # Write frames
    for f in range(framecount):

        # Write graphic control extension
        out.write("!\xF9".encode('utf-8'))
        out.write(struct.pack("<HBB", 0, disposal, 0))
        # This is an animation delay of 0ms instead of decoding time
        # The actual value does not matter
        out.write("\x00\x00".encode('utf-8'))

        # Write local image descriptor
        out.write(",".encode('utf-8'))
        out.write(width.to_bytes(2, 'little', signed=False))
        out.write(height.to_bytes(2, 'little', signed=False))
        out.write((0x08).to_bytes(1, 'little', signed=False))
        out.write((0x00).to_bytes(1, 'little', signed=False))

        # Write LZW min code size and raster data itself
        out.write(rng.read(2))
        bytesize = int.from_bytes(
                rng.read(1), byteorder='big')
        out.write(rng.read(bytesize))



# Takes a RNG or a list of bytes to use as input.
# If it's a list then a BytesIO will be used.
# Otherwise we assume that it's already a RNG.
def gen_rand_jpeg(src)->io.BytesIO:
    if isinstance(src, io.BytesIO):
        pass
    elif isinstance(src, list):
        src = io.BytesIO(bytes(src))
    else:
        raise TypeError(f'{type(src)} is not supported.')

    jpegbody = []

    # Section 1: SOF0 Huffman Table IDs
    jpegbody += [
        0xC0, 0xD0, 0xE0, 0xF0, 0x01, 0x11, 0x02, 0x12,
        0x21, 0x31, 0x41, 0x03, 0x13, 0x22, 0x32, 0x42,
        0x51, 0x61, 0x71, 0x81, 0x91, 0xA1, 0xB1, 0xC1,
        0xD1, 0xE1, 0xF1, 0x10, 0x20, 0x30, 0x40, 0x50,
        0x60, 0x70, 0x80, 0x90, 0xA0, 0xB0, 0xC0, 0xD0,
        0xE0, 0xF0,
    ]

    # Total length of section 1: 40 bytes (= 0x28) followed by itself. Then,
    # section 2 is one byte long: 0x01. Finally, section 3 has two bytes whose sum is always even:
    # 0xFF 0xD9 (~49151 decimal)
    hufflen    = 40+40+1+2
    jpegbody += ([hufflen]*2)*256

    return io.BytesIO(jpegbody)



