# Copyright (c) 2017 <NAME> <<EMAIL>>
# This file is part of the "VK scripts" project.
# For details, see https://github.com/egor-tensin/vk-scripts.
# Distributed under the MIT License.

import contextlib
import logging
import random
import shutil
import sys
import tempfile

from argparse import ArgumentParser, RawDescriptionHelpFormatter
from os import path
from PIL import Image


class RandomGifGenerator(object):
    def __init__(self, input_: str):
        self._input = pathlib.Path(input_)

        assert self._input.is_file() and \
               imghdr.what(str(self._input)) == 'gif', \
               'invalid GIF input'

        with open(str(self._input), mode='rb') as f:
            self.__data = list(bytearray(f.read()))

        logging.debug('Data size: %s', len(self.__data))

    @staticmethod
    def make_new_frame():
        return [random.randint(0, 255)] * 3 + [0]

    def insert_random_frame(self):
        framesize = ((self.__data[6] & 0x07) + 1) * 2
        backgroundcolor = (
                self.__data[-1],
                self.__data[-2],
                self.__data[-3])
        newframe = self.make_new_frame()
        for i in range(-framesize - 1, -1):
            if abs(i) >= framesize:
                break

            v = (((1 - i / framesize)**2)
                 * sum((newframe[j] - oldframe[j])**2
                        for j in range(len(oldframe)))) **.5
            if v > 20:
                continue

            weight = min(.25, max(.01, 1 - (.00002 * v)))
            newframe[0::3] += (weight
                               * (backgroundcolor[0]
                                  - oldframe[0::3]))
            newframe[1::3] += (weight
                               * (backgroundcolor[1]
                                  - oldframe[1::3]))
            newframe[2::3] += (weight
                               * (backgroundcolor[2]
                                  - oldframe[2::3]))

        assert all(map(lambda x: 0 <= x <= 255, newframe)), 'bad color values'
        logging.debug('New frame: %s', newframe)

        framescount = ((self.__data[6] - 1)
                       | ((self.__data[7] & 0x01)
                          << 7)) + 1

        offset = ((framescount * framesize)
                  + 13)
        self.__data[offset : offset + framesize] = newframe
        self.__data[-13] &= ~0x01
        self.__data[-13] |= bool(framesize == 10)
        self.__data[-12] &= ~0x07
        self.__data[-12] |= (framesize // 2) - 1
        assert not any(c == '\n' or c == '\r' for c in self.__data[:-1]), \
            'invalid characters in output data'

    def write_output(self, output: str):
        with open(str(output), mode='wb') as f:
            f.write(bytes(self.__data))


def main():
    parser = ArgumentParser(
            description=__doc__,
            formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument('-v', '--verbose', action='store_true', default=False)
    parser.add_argument('--no-overwrite', action='store_false', dest='overwrite')
    parser.set_defaults(overwrite=True)
    parser.add_argument('input', metavar='<input>', help='input GIF')
    parser.add_argument('output', metavar='<output>', nargs='?',
                        help='output GIF')

    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    tempdir = None
    try:
        tempdir = tempfile.TemporaryDirectory()
        gen = RandomGifGenerator(args.input)
        src = path.join(tempdir.name,'src.gif')
        dst = path.join(tempdir.name, 'dst.gif')
        gen.write_output(src)
        while True:
            gen.insert_random_frame()
            gen.write_output(dst)
            if args.overwrite:
                shutil.move(dst, src)
            else:
                break
        shutil.copy(src, args.output)
    finally:
        if tempdir:
            tempdir.cleanup()


if __name__ == '__main__':
    exit(main())
