#!/usr/bin/env python3

"""A simple random GIF generator."""

__version__ = "0.2"
__author__ = "<NAME>"

import os
import struct

def main() -> int:
    """Main function run by this script."""

    print("Generating", DEFAULT_OUTPUT)

    # Open a pseudo-random number generator so we don't have to bother with seeding
    try:
        rng = open("/dev/urandom")
    except IOError:
        msg = ("Could not open /dev/urandom.\n"
               "\tThis script will only work on Unix systems.")
        raise OSError(msg)

    # Generate the gif
    try:
        gen_gif(rng, DEFAULT_OUTPUT)
    finally:
        rng.close()

    print("Done!")

    return 0
    

    

def gen_gif(rng : BinaryIO, destfile : str) -> None:
    """Generates a random GIF and stores it at `destfile`, using
       `rng` as the source of randomness for the generation."""

    # Header
    with open(destfile, "w+b") as f:
        # File signature and version
        f.write(struct.pack("<6sHHBBBxxx", b"GIF89a", 768, 512, 0, 10, 2))
        
        # Logical Screen Descriptor
        f.write(struct.pack("!HhhHHBBBBBxxxxxxx",
                40, 512, 512, 1, 2**f.read(1)[0], 
                0, 0,                   # GCTF color resolution
                0,                       # background color index
                False, True,             # pixel aspect ratio
                0                        # transparency flag
            ))

        # Image Descriptor
        f.write(struct.pack("!HHHHHHBxxx",
                0,      # left offset
                0,      # top offset
                512,    # width
                512,    # height
                0,      # flags
                0       # local color table size
            ))

        # Global Color Table
        for x in range(0, 3*256, 3):
            r = rng.read(1)[0] % 256
            g = rng.read(1)[0] % 256
            b = rng.read(1)[0] % 256
            f.write(bytes([r]))
            f.write(bytes([g]))
            f.write(bytes([b]))


        # Graphic Control Extension
        f.write(struct.pack("!BxxxxxxBBH",
                0x21,   # introducer [GRAPHIC CONTROL EXTENSION CODE BLOCK]
                0xF9,   # graphic control label ([APPLICATION SPECIFIC LABEL] in spec)
                4,      # block size 
                False,  # no disposal specified
                False,  # user input traversable
                0,      # delay time & transparent idx
                0x00    # terminator [BLOCK TERMINATOR CODE]
            ))

        # Plain Text Extension
        txt = "".join(chr((rng.read(1)[0] % 256) + ord(' ')) for _ in range(10))
        f.write(struct.pack("!BxxxxxxxxxxBB",
            0x21,           # introducer [TEXT SUB-BLOCK START CODE BLOCK]
            0x0A,           # text extension label
            12,             # block size (fixed value)
            len(txt),       # text length
            ord(' '),       # default character encoding [ASCII] 
            ord('T'),       # font descriptor [TI-Nspire CX]
            ord('e'),       # font descriptor [TI-Nspire CX]
            ord('x'),       # font descriptor [TI-Nspire CX] 
            0                # terminator [BLOCK TERMINATOR CODE]
        ))
        f.write(txt.encode())
        f.write(bytes([0]))                     # terminator [BLOCK TERMINATOR CODE]

        # Application Extension
        f.write(struct.pack("!BxxxxxxxxBB",
            0x21,   # introducer [APPLICATION EXTENSION CODE BLOCK]
            0xFF,   # application extension code [NETSCAPE APPLICATION EXTENSION]
            11,     # block size (fixed value)
            ord('N'),   
            ord('E'),  
            ord('T'),  
            ord('S'),  
            ord('C'),  
            ord('A'),  
            ord('P'),  
            ord('E'),  
            ord(' ') 
        ))
        f.write(struct.pack(">HH",
            3,              # application ID (fixed like above)
            ord('2'.encode()[0]),   # application authentication code [stub]
        ))

        # Comment Extension
        f.write(struct.pack("!BxxxxxxBBB",
            0x21,   # introducer [COMMENT EXTENSION CODE BLOCK]
            0xFE,   # comment extension label
            12,     # block size (fixed value)
            ord('S'),  
            ord('o'),  
            ord('u'),  
            ord('r'),  
            ord('c'),  
            ord('e')  
        ))
        f.write(struct.pack("!H", 1))              # comment string byte length
        f.write(struct.pack("!H", 0))              # comment string termination
        f.write(bytes([0]))                         # terminator [BLOCK TERMINATOR CODE]

        # Trailer
        f.write(struct.pack("!B", 0x3B))           # trailer [TRAILER CODE BLOCK]



        

        


if __name__ == "__main__":
    status = main()
    sys.exit(status)






