#!/usr/bin/python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
    
        with open(f'./{self.output}', 'wb') as fh:
            generate_random_gif(open('/dev/urandom'), fh)


# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# correspond to the information found here:
# https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Signiture
    signature = bytes('GIF89a', encoding='utf-8')
    out.write(signature)

    # Screen descriptor
    screendescriptor = bytearray()

    # Logical Screen Descriptor
    # width is 16 bit
    width = rng.read(2) 
    screendescriptor += width

    # height is 16 bit        
    height = rng.read(2)                  # TODO check order of operations
    screendescriptor += height 

    # Global Colour Table Flag
    globalcolourtableflag = int.from_bytes(rng.read(1), "big")
    screendescriptor += globalcolourtableflag.to_bytes(1, byteorder='big')

    # Sort flag
    sortflag = int.from_bytes(rng.read(1), "big")
    screendescriptor += sortflag.to_bytes(1, byteorder='big')

    # Size of colour table
    colours = int.from_bytes(rng.read(1), "big")
    screendescriptor += colours.to_bytes(1, byteorder='big')

    # Background colour index
    backgroundindex = int.from_bytes(rng.read(1), "big")
    screendescriptor += backgroundindex.to_bytes(1, byteorder='big')

    # Pixel aspect ratio
    pixelratio = int.from_bytes(rng.read(1), "big")
    screendescriptor += pixelratio.to_bytes(1, byteorder='big')

    
    out.write(screendescriptor)             # TODO remove when testing done

    # Image descriptor
    out.write((b'\x2c'))                   # Start of image decriptor

    # Image descriptor
    imagedescriptor = bytearray()

    # Left position
    leftposition = rng.read(2)            # TODO convert to integer
    imagedescriptor += leftposition

    # Top position    
    topposition = rng.read(2)             # TODO convert to integer
    imagedescriptor += topposition

    # Width of image
    widthofimage = rng.read(2)            # TODO convert to integer
    imagedescriptor += widthofimage

    # Height of image
    heightofimage = rng.read(2)           # TODO convert to integer
    imagedescriptor += heightofimage

    # Local colour table flag
    localcolourtableflag = int.from_bytes(rng.read(1), "big")
    imagedescriptor += localcolourtableflag.to_bytes(1, byteorder='big')

    # LZW minimum code size
    lzwminimumcodesize = int.from_bytes(rng.read(1), "big")
    imagedescriptor += lzwminimumcodesize.to_bytes(1, byteorder='big')

    out.write(imagedescriptor)

    # Discard image data sub-blocks until we meet
    # the application extension block signitue
    while True:
        
        # Read the next byte from stream
        databyte = rng.read(1)

        # Check if this is the end of the image data 
        # block by checking if this is the last code    
        # byte which has its most highest bit set to 1        
        if databyte[0] & 0x80 == 0x80: break

        # Write original byte onto the image file
        out.write(databyte)

    # Application Extension Block
    appextensionblock = bytes([
        0x21,                                    # Extension Introducer
        0xff,                                    # Appliation Extension Label
        0x0b,                                    # Block Size
        0x4e, 0x45, 0x54, 0x53, 0x43, 0x41,      # Application Identifier
        0x50, 0x45,                              # Application Authentication Code
        0x31,                                    # Sub-block Size = 1

        0x00                                      # Application Data
    ])

    out.write(appextensionblock)

    # Comment Extension Block
    commentextensionblock = bytes([
        0x21,                                    # Extension Introducer
        0xfe,                                    # Comment Extension Label
        0x0d,                                    # Block Size
        0x4C, 0x6F, 0x67, 0x2D, 0x31,            # Comment Tag
        0x2E, 0x30, 0x2E,                        # Comment Version
        0x31,                                    # Sub-block Size = 1

        0x00
    ])

    out.write(commentextensionblock)

    # Graphics Control Extension Block
    graphicscontrolblock = bytes([
        0x21,                                        # Extension Introducer
        0xf9,                                        # Graphic Control Label
        0x04,                                        # Block Size
        0x08,                                        # Packed Field
        0x00, 0x00,                                  # Delay Time {1/100 sec}
        0x00,                                        # Transparent Index
        0x00                                         # Terminator for sub-blocks
    ])

    out.write(graphicscontrolblock)

    # Plain Text Extension Block
    plaintextblock = bytes([
        0x21,                                        # Extension Introducer
        0x01,                                        # Plain Text Label
        0x09,                                        # Block Size
        0x00, 0x00,                                  # Grid Left Position
        0x00, 0x00,                                  # Grid Top Position
        0x00, 0x00,                                  # Grid Width
        0x00, 0x00,                                  # Grid Height
        0x00,                                        # Cell Width
        0x00                                         # Cell Height
    ])

    out.write(plaintextblock)


    # Ending
    ending = bytes([
        0x3b                                       # File terminator
    ])

    out.write(ending)