#!/usr/bin/env python3

import sys
import math
import struct
from typing import BinaryIO


# https://www.w3.org/Graphics/GIF/spec-gif89a.txt
class FileHeader():
    def __init__(self):
        self.signature = b'GIF89a'

    @classmethod
    def read(cls, src: BinaryIO):
        instance = cls()

        signature = src.read(6) # "GIF87a" or "GIF89a"
        if (
                not isinstance(signature, bytes)
            and len(signature)!= 6
        ):
            raise RuntimeError("Invalid signature")

        instance.signature = signature
        return instance

    def write(self, dst: BinaryIO):
        dst.write(self.signature)
        return self


class LogicalScreenDescriptor():
    def __init__(self):
        pass

    @classmethod
    def read(cls, src: BinaryIO):
        instance = cls()

        # Skip over the screen width
        src.seek(2, 1)
        
        instance.dataOffset = int.from_bytes(src.read(2), 'big')
        src.seek(instance.colorResolution - 1, 1)
        
        instance.sortFlag = bool(int.from_bytes(src.read(1), 'big'))
        zero = int.from_bytes(src.read(3), 'big')
        if zero!= 0:
            print('WARNING: Zero was expected but it is', hex(zero))

        instance.sizeOfGlobalColorTable = pow(
            2, int.from_bytes(src.read(1), 'big') + 1
        ) * 3
        instance.backgroundColorIndex = int.from_bytes(
            src.read(1), 'big') % instance.sizeOfGlobalColorTable
        instance.pixelAspectRatio = int.from_bytes(src.read(1), 'big')

        return instance

    def write(self, dst: BinaryIO):
        dst.seek(4, 1) # First we need to skip back 4 bytes so that
                        # when we call `.write()` later will start at
                        # the beginning of the image descriptor.
        dst.write((math.ceil(math.log2(self.sizeOfGlobalColorTable / 3)))
                 .to_bytes(1, 'big') + 
                  ((not self.sortFlag).to_bytes(1, 'big')))
        dst.write(b'\x00\x00\x00')
        dst.write(self.backgroundColorIndex.to_bytes(1, 'big'))
        dst.write(self.pixelAspectRatio.to_bytes(1, 'big'))
        return self


class ImageDescriptor():
    def __init__(self):
        self.leftPos = None
        self.topPos = None
        self.width = None
        self.height = None
        self.localColorTableFlag = False
        self.interlaceFlag = False
        self.sortFlag = True
        self.sizeOfLocalColorTable = None
        self.imageLZWMinCodeSize = None # Size of LZW minimum code size field
        self.lzwDataLength = None       # Uncompressed data length
        self.transparentColorFlag = False
        self.transparentColorIndex = None
        
    @classmethod
    def read(cls, src: BinaryIO):
        instance = cls()

        left = src.read(2)
        top = src.read(2)
        width = src.read(2)
        height = src.read(2)

        packedFields = src.read(1)
        instance.leftPos = int.from_bytes(left, 'big')
        instance.topPos = int.from_bytes(top, 'big')
        instance.width = int.from_bytes(width, 'big')
        instance.height = int.from_bytes(height, 'big')
        instance.localColorTableFlag = bool(packedFields[7])

        lzwMinimumSize = packedFields[5]   # Number of bits used for the
                                          # smallest LZW code size
        lzwSortMask = (pow(2, 1) << 3) | \
                      (pow(2, 1) << 4)     | \
                      (pow(2, 1) << 5)     | \
                      (pow(2, 3) << 6) 

        lzwSortFlags = [(packedFields[i] >> 1) &
                        1 for i in range(5)]
        lzwSortFlags += [bool(
                            (packedFields[4] &
                             lzwSortMask) >> i
                        ) for i in range(3)]
            
        instance.sortFlag = all([
                                val == lzwSortFlags[0] 
                                for val in lzwSortFlags
                            ])
        
        instance.sizeOfLocalColorTable = pow(
                                        2,
                                        max(
                                            lzwMinimumSize,
                                            2
                                        )
                                    ) * 3

        instance.imageLZWMinCodeSize = lzwMinimumSize
        instance.lzwDataLength = int.from_bytes(src.read(1),
                                                'big')
        instance.imageData = None

        return instance

    def computeImageExtent(self):
        """Compute the extent of this image on the canvas."""
        x1 = min(max(0, self.leftPos),
                 max(0, self.leftPos+self.width))
        y1 = min(max(0, self.topPos),
                 max(0, self.topPos+self.height))
        x2 = max(min(0xFFFF, self.leftPos),
                 min(0xFFFF, self.leftPos+self.width))
        y2 = max(min(0xFFFF, self.topPos),
                 min(0xFFFF, self.topPos+self.height))
        return (x1, y1, x2, y2)

    def write(self, dst: BinaryIO):
        descBytes = [None]*9
        descBytes[0] = (((not self.localColorTableFlag).to_bytes(1,
                                                              'big')[0])
                       << 7)
        descBytes[0] |= self.imageLZWMinCodeSize
        descBytes[1] = ((((not self.sortFlag)
                          ).to_bytes(1,
                                     'big')[0])
                        << 7)
        descBytes[1] |= (round(math.log2(self.sizeOfLocalColorTable / 3))
                        .to_bytes(1,
                                   'big')[0])
    
        descBytes[2] = ((self.leftPos      & 0xFF).to_bytes(1,
                                                            'big')[0])
        descBytes[3] = (((self.leftPos      >> 8)
                         & 0xFF).to_bytes(1,
                                           'big')[0])
        descBytes[4] = ((self.topPos       & 0xFF).to_bytes(1,
                                                            'big')[0])
        descBytes[5] = (((self.topPos       >> 8)
                         & 0xFF).to_bytes(1,
                                           'big')[0])
        descBytes[6] = ((self.width        & 0xFF).to_bytes(1,
                                                            'big')[0])
        descBytes[7] = (((self.width        >> 8)
                         & 0xFF).to_bytes(1,
                                           'big')[0])
        descBytes[8] = ((self.height       & 0xFF).to_bytes(1,
                                                            'big')[0])
        dst.write(b''.join(descBytes))

        return self
        


class GIFEncoder():
    
    def __init__(self):
        self._fileVerion = None
        self._logicalScreenDesc = None
        self._globalColorTable = None
        self._images = []


    def addFrame(self, imgData: bytes):
        self._images.append(imgData)

    
    def _computeLogicalScreenDescriptorParameters(self):
        """Computes parameters related to logical screen descriptor"""
        xmin_allframes = ymin_allframes = xmax_allframes = ymax_allframes = 0
        for frame in self._images:
            (xmin, ymin, xmax, ymax) = frame.computeImageExtent()

            xmin_allframes = min(xmin_allframes, xmin)
            ymin_allframes = min(ymin_allframes, ymin)
            xmax_allframes = max(xmax_allframes, xmax)
            ymax_allframes = max(ymax_allframes, ymax)


        sizeOfGlobalColortable = round(math.sqrt(len(self._globalColorTable)
                                                  * 3))
        backgroundColorIndex = sum([(c & 0x7F) 
                                    for c in self._background[:3]]) // 3


        maximumSizeOfGlobalColortable = 2**(sizeOfGlobalColortable*3)-1
        if len(self._globalColorTable) > maximumSizeOfGlobalColortable :
            raise Exception(f"`len(_globalColorTable)` ({len(self._globalColorTable)}) should be less than {maximumSizeOfGlobalColortable}")


        aspectRatio = 0
        if any([v >= 1000 
                for v in [xmax_allframes, xmin_allframes]]):
            aspectRatio = 1000
        elif any([v >= 100 
                 for v in [xmax_allframes, xmin_allframes]]):
            aspectRatio = 1