#!/usr/bin/env python3
"""Random GIF Generator."""


import argparse
import itertools as it
import logging
import os
import sys
import zlib # for compressing lzw-compressed blocks
from pathlib import Path
from typing import BinaryIO, Tuple

# Globals
DIRECTORIES = []
OUTPUTS = {}

def main():
    """Main function"""


    parser = argparse.ArgumentParser(description='Generate a random collection of GIF files.')
    parser.add_argument('-z', '--zipfile', type=str, help='Path to zip archive containing input files.')
    parser.add_argument('-f', '--files', nargs='+', help='Paths to input files.')
    parser.add_argument('-o', '--output', type=str, default='',
                        help='Output directory where all generated files will be placed.')
    parser.add_argument('--seed', type=int, default=-1, 
                        help='Seed used to initialize the RNG.')
    
    args = vars(parser.parse_args())

    # set up rng seed
    seed = None
    if args['seed'] == -1:
        try:
            with open('/dev/urandom', mode='rb') as f:

                seed = struct.unpack('<L', f.read(4))[0]
                logging.debug("Using seed {:X}".format(seed))

        except Exception:
            raise SystemExit("Could not determine system entropy.")

    else:
        seed = args['seed']
    
    # Process inputs
    process_inputs(args, seed)
    

def process_directory(dirpath):
    """Process a directory recursively."""

    def _process_directory(dirpath):
        
        # Get list of directories
        dirnames = sorted([Path(x).name for x in
            filter(lambda x:
                    Path(x).is_dir(),
                it.chain(*[list((dirpath / d).iterdir()) 
                    for d in DIRECTORIES]))])

        dirpaths = {d: dirpath / d for d in dirnames}
        OUTPUTS |= dirpaths

        # Recursively process subdirectories
        [_process_directory(dirpaths[d]) for d in dirnames]
    
    _process_directory(Path(dirpath))
    



def process_input(filepath, seed):
    """Process an individual input file."""

    output = Path(OUTPUTS[os.path.basename(filepath)])
    output.mkdir(parents=True, exist_ok=True)

    n = 0

    while True:

        if n >= 2**32: break

        n += 1

        filename = "{:08}.gif".format(n)
        filepath = output / filename


        with open(filepath, "wb") as ostream:
            
            generate_random_gif(ostream, seed)

def process_inputs(args, seed):
    """Process inputs from CLI arguments"""

    if args['zipfile']:
        pass
    
    elif args['files']:

        for path in args['files']:
            process_input(path, seed)
    
    elif args['output']:
        
        if not os.path.isdir(args['output']):
    
            sys.exit("Invalid output '{0}'. Output must be a directory."\
                   .format(args['output']))
    
        global DIRECTORIES, OUTPUTS
        DIRECTORIES = ['.']
        OUTPUTS = {'': Path(args['output']).resolve()}

        process_directory('')


def generate_random_gif(
        out: BinaryIO, 
        seed: int
    ) -> None:

    
    def get_bits(num, nbits):

        return num & ((1 << nbits)-1)


    # Set up w, h
    v = get_bits(get_bits(seed, 64), 32)
    w, h = [get_bits(v, 16) for i in range(2)]

    ostream = out.buffer

    # Make sure we're within bounds so that the gifs aren't too big.
    max_size = min(w,h) 
    if max_size < 4: w, h = 4, 4
    elif max_size < 9: w, h = 9, 9
    elif max_size < 16: w, h = 16, 16
    elif max_size < 33: w, h = 33, 33
    elif max_size < 65: w, h = 65, 65
    
    
    # Write header
    ostream.write(b'\x47\x49\x46\x38\x39\x61')
    ostream.write(struct.pack('<HHCC', w, h, 0xF7, 0, 0))

    # Determine global color table size
    cts = get_bits(get_bits(seed, 64), 32) % 16
    cti = get_bits(cts+1, 8)
    ccir = get_bits(cts<<8|cti, 16) | 0b1000000000000000

    # Palette chunk follows header
    if cts > 0:

        # Get palette data
        paldata = [(get_bits(v, 8), get_bits(v, 8), get_bits(v, 8))
            for i in range(cti*3)]
            
        paldata = bytes().join([(struct.pack('BBB', r, g, b)
            for (r,g,b) in paldata)])

        palents = bytes()
        for i in range(len(paldata)//3):

            l = paldata[i*3]
            assert isinstance(l, int) or ord(l) is int

            palents += struct.pack('B', l)
            palents += paldata[(i*3)+1:(i*3)+4]
       
        assert len(palents) == sum([ord(p) for p in palents]) + 1

        ostream.write(struct.pack('<HBB', len(palents), 0,
            0b10000000 | cti))
        ostream.write(palents)
        
    else:

        ostream.write(struct.pack('<HBB', 0, 0, 0))

    # Image Data Chunk
    data = bytes()
    ostream.write(b'\x2C\x00\x00\x00') # image block identifier

    if cts == 0:

        # No palette; just use all black pixels

        data = b'\x00'*w*h
        interlace_flag = 0

    else:

        # Use palette to encode pixels

        prevpix = None

        while len(data)<w*h:

            nextpixidx = chr(get_bits(seed, 8)%cti)
            pixidx = nextpixidx if prevpix is None else prevpix
            prevpix = nextpixidx

            if pixidx!= '\x00':

                col = [struct.unpack('<BBB', paldata[i*3:(i+1)*3])
                    for i in range(cti)][ord(pixidx)]
                
            else:
                
                col = (0xFF, 0xFF, 0xFF) # white

            data += struct.pack('BBB', *col)

    dointerframe = get_bits(seed, 64)&1==1
    interlace_flag = 0x08 if dointerframe else 0x00

    lzwmincodesize = get_bits(get_bits(seed, 64), 32)%9
    lzwmincodesize = 9 if lzwmincodesize<9 else lzwmincodesize

    

    # LZW compression scheme
    # <https://www.matthewflickinger.com/lab/whatsinagif/lzw_image_data.asp>

    lzwsizes = bytes([
        255, 225, 193, 161, 129, 97, 65, 33, 32])

    lzwcodes = dict(enumerate(range(256)))
    lzwcodescur = len(lzwsizes)
    lzwbuf = ''
    lzwdatabuf = b''



    # idat section
    ostream.write(struct.pack('<BB', lzwmincodesize, 
        lzwsizes[lzwmincodesize]|interlace_flag))

    # lzw code buffer
    lzwbuf += data
    lzwbuf = bytes([lzwbuf[i-1] for i in range(len(lzwbuf)+1)])

    assert len(lzwbuf)==256

    # Loop over data generating codes until we've filled lzwbuf
    while any(c!='\x00' for c in lzwbuf[:lzwcodescur]):

        # Generate new lzwcode if lzwcodescur has reached zero
        if lzwcodescur == 0:

            # Select code size based on previous value
            lzwcodesize = get_bits(get_bits(seed, 64), 32)%9
            lzwcodesize = 9 if lzwcodesize<9 else lzwcodesize
            lzwcodescur = lzwcodesize+2

            # If this was our first time through then start by filling buf
            if len(lzwdatabuf)