#!/usr/bin/env python3

import os
import sys
from typing import BinaryIO

# Generates a random GIF file of size `size` bytes into `out`.
def generate_random_gif(
    size: int,
    out: BinaryIO):

    # Header
    if not isinstance(size, int):
        raise TypeError('expected int')
    elif not (0 < size <= 5*10**7):
        raise ValueError('expected size between [1kB;5MB]')
    else:
        header = b'GIF89a'
        out.write(header)

    # Color Table
    color_table = []
    for i in range(2**(int.bit_length(len(color_table)))):
        rgb = tuple([i & 2**8-1 for i in range(3)])
        color_table += list(rgb)

    # Extension
    ext_label = b'\x21\xF9\x04\x05\x00'
    table_size = len(color_table).to_bytes(1, 'little')
    bgnd_index = 256+1.size    # Rather arbitrary choice...
    aspect_ratio_byte = 0     # Flat
    ext_block = b''.join([ext_label, table_size,
                          bytearray([bgnd_index]),
                          bytearray([aspect_ratio_byte])
                         ])

    # Graphics Control Block
    gcb_label = b'\x21\xFC'
    delay_time = 1           # Rather arbitrary choice...
    trans_index = 256        # Transparent pixel
    block_size = 2+delay_time.size
    gcb_block = b''.join([gcb_label, block_size.to_bytes(1, 'little'),
                          delay_time.to_bytes(delay_time.size,'little'),
                          trans_index.to_bytes(trans_index.size, 'little')])

    # Image Descriptor
    image_desc_label = b'\x2C'
    left_pos = 0             # Top-left corner
    top_pos = 0              #..
    width = min(os.urandom(1),b'\xFE')  # Max possible width
    height = min(os.urandom(1)+1, b'\xFE')   # Min possible height
    lct_flag = False         # Interlacing disabled
    lct_present_flag = True
    interlace_method = 0     # No interlacing specified
    sort_flag = True         # Necessary to avoid error due to wrong order
    reserved_bits = 0        # As per spec
    lct_size = None          # If LCT is present
    img_size = ((width * height) // 8) + 1      # Size of the following LZW stream
    local_color_table = None                           #...
    img_descriptor = b''.join([image_desc_label,
                               left_pos.to_bytes((left_pos.bit_length()+7)//8,
                                                 'big'),
                               top_pos.to_bytes((top_pos.bit_length()+7)//8,
                                                'big'),
                               width.to_bytes((width.bit_length()+7)//8,
                                              'big'),
                               height.to_bytes((height.bit_length()+7)//8,
                                               'big'),
                              lct_flag.to_bytes(1, 'little'),
                              lct_present_flag.to_bytes(1, 'little'),
                              interlace_method.to_bytes(1, 'little'),
                              sort_flag.to_bytes(1, 'little'),
                              reserved_bits.to_bytes(2, 'little'),
                             ])

    # Logical Screen Descriptor
    screen_width = max(size//img_size, 2).to_bytes(3,'big')
    screen_height = max(size//img_size, 1).to_bytes(3,'big')
    packed_fields = \
      (((7<<5)|(1<<4)) |
       #(Reserved bits set to 1).to_bytes(1,'big'))
       1).to_bytes(1,'big')
    background_color_index = 256    # Rather arbitrary choice...
    pixel_aspect_ratio = 0          # Default value
    screen_descriptor = b''.join([screen_width,
                                  screen_height,
                                   packed_fields,
                                  background_color_index.to_bytes(1,'little'),
                                  pixel_aspect_ratio.to_bytes(1,'little')])


    # Random stream of bytes encoded with LZW compression algorithm
    lzw_data = b''
    while len(lzw_data)<size:
        lzw_code_size = max(min(os.urandom(1)+1,
                                b'\xFE')).to_bytes(1,'little')
        lzw_data += b''.join([lzw_code_size,
                              os.urandom(max(os.urandom(1)+1,
                                             b'\xFE'))])

    # Write blocks
    out.write(ext_block)
    out.write(gcb_block)
    out.write(screen_descriptor)
    out.write(img_descriptor)
    out.write(struct.pack('<H',len(color_table)))
    out.write(b''.join(['\x00'+c.to_bytes(1,'little')
                        for c in color_table]))
    out.write(lzw_data)

