#!/usr/bin/env python3

import os
import struct

from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, size=None
    ):
    """Generate a GIF of some specified width and height with a given RNG"""
    if not size:
        print("Specify size!")
        return -1

    (width, height) = size
    header = 'GIF89a'.encode('ascii')

    local_palette = [0x00, 0xff] * 256

    global_palette = []
    for i in range(768):
        global_palette.append(local_palette[rng.randint(
                0, len(local_palette)-1)])
    out.write(struct.pack("<3sBBB",
                          header,
                          0,      # Width
                          0,      # Height
                          7       # Color table
                          ))

    # Write palette to file
    psize = len(global_palette) // 3
    paldata = bytearray()
    colordepth = max((width+height)//32, 1)
    while global_palette:
        paldata += bytes([colordepth])
        paldata += bytes([global_palette.pop()])
        paldata += bytes([global_palette.pop()])
        paldata += bytes([global_palette.pop()])

    lzwmincodelen = 3
    lzwmaxcodelen = min(psize, 12)+1

    lzwendofblock = (1 << lzwmaxcodelen) | 1
    lzwclearcode = (1 << lzwmaxcodelen)
    lzwlastcode = None
    table = dict([(i, i) for i in range(lzwclearcode)])
    codes = list(range(lzwclearcode, lzwclearcode+(1<<lzwmincodelen)))
    bitcount = lzwmincodelen+1

    # Data section
    encodingstep = 2**bitcount
    def encode_value(val):
        nonlocal codes
        code = val & ((1<<(bitcount))-1)
        val >>= bitcount

        # If we need more codes
        if len(codes) < encodingstep:
            # Return it as is
            return code

        # Otherwise add an entry to our table/dict
        new_code = codes[-1]+1
        if new_code >= 2**(bitcount+1)-1:
            raise BufferError("No spare codes")
        table[new_code] = val
        codes.append(new_code)
        return new_code


    image = bytearray()
    for y in range(y):
        byt = 0
        for x in range(x):
            imgval = rng.randint(0, 767)

            # Get the LZW encoded value for this position
            vallen = bitcount//3
            lzw = encode_value(imgval)
            lzw |= image[(y*width+x)*vallen:(y*width+x+1)*vallen] << bitcount

            # Add to our image buffer
            encbuf = bytearray()
            while lzw > 255:
                encbuf.extend([lzw&255])
                lzw >>= 8
            encbuf.extend([lzw])
            image.extend(encbuf)

            # Write our next output block
            writebuffer = bytearray()
            if all(b == 0xFF for b in encbuf[:-1]):
                pass    # Do nothing because we're at the end
            elif any(b==255 for b in encbuf):
                # Flush until we've reached the start of a new block
                lastflg = False
                while True:
                    if lastflg or (0xFF not in encbuf):
                        break

                    right = encbuf.index(255)
                    left = right-vallen
                    if left <= 0:
                        left += right
                    else:
                        right = left

                        for pos in range(left, right):
                            rvl = (pos-left)/vallen
                            val = image[(rvl)]

                            # Special case for first pixel
                            if rvl == 0:
                                val <<= bitcount
                                val |= image[:vallen]

                                image[:] = image[vallen:]
                                writebuffer.extend([val])
                            else:
                                val <<= bitcount
                                val |= image[(rvl+1)*vallen:
                                              (rvl+2)*vallen]

                                image[rvl*vallen:
                                      (rvl+1)*vallen] = \
                                        image[(rvl+1)*vallen:
                                              (rvl+2)*vallen]
                                writebuffer.extend([val])

                    lastflg = True


        # Write our output block
        pass



    # Image descriptor
    out.write(bytearray([
        0x2C,         # Image separator
        0x00, 0x00,   # Left
        0x00, 0x00,   # Top
        0x00, width,  # Width
        0x00, height, # Height
        0x80          # Local color table
        ]))

    # Write palette to file
    out.write(paldata)

    # Write compressed data
    out.write(image)