#!/usr/bin/env python3

import struct
from os import urandom
from pathlib import Path


# Generates a random GIF file into `out`. If `palette_path` is provided then it will be used
# as the color palette for the generated frames; otherwise one will be chosen randomly instead.
def generate_random_gif(
        palette_path=None, frame_count=30, 
        out='output.gif', delay=50): 

    # always start with a fresh buffer
    buf = []

    # extract palette data from given path; this should be a.actl file
    if palette_path!= None:
        palette_path = Path(palette_path).resolve().expanduser()
    
        fbuf = open(palette_path, mode='rb').read()

        # make sure we have an even number of RGB tuples before unpacking
        assert len(fbuf) % 3 == 0, "Palette must contain an even number of RGB triplets"
        
        paldata = list(zip(*[iter(fbuf)]*3))

        # pad to 256 entries
        while len(paldata) < 256:
            paldata += [[0, 0, 0]]
        
        actlhdr = struct.pack('<BBBBBBIII', *([
            0, 0xff, 76, 0, 0, 0, 2, 0, len(paldata)*3]) + 
            [int.from_bytes(x,'little') for x in paldata])
    else:
        # choose a random value between 0-255 for all 256 colors
        actlhdr = struct.pack('<BBBBBBII', *[0]*6 + [255] + [urandom(3)])
    
    # add file header
    hdr = b'GIF89a'

    # store the offsets at which each frame begins within the file
    frame_offsets = {}

    # fill in all frames except the last one (which has zero duration)
    for idx in range(frame_count-1):
        buf += [hdr+actlhdr]

        # build a temporary buffer for storing this frame's data
        tmpbuf = []
    
        # set up the logical screen descriptor
        tmpbuf += [struct.pack('<HHBB', frame_count*(idx+1),
            frame_count*(idx+1), 0b10000000, 0)]
    
        # default to LZW compression
        compmeth = 0x80
        compflag = 0

        # create a new local palette, or use the existing one based on whether
        # a palette was passed in via `palette_path`
        local_clrtbl = list(zip(*[iter(urandom(3))] * 3)) if palette_path==None else paldata 

        # write out our globals
        clrtblsz = min(len(local_clrtbl), 256) & ~1
        clrtblsz |= 0b10000000
        tmpbuf += [struct.pack('<BBBB', compmeth, 
            0, delay&255, ((delay >> 8)&255)|compflag)]
    
        # append actual color table
        tmpbuf += [struct.pack('<'+'HB'*len(local_clrtbl),''.join([(len(x)+1).to_bytes(1,'little')]+list(x)))]

        # now that our screen descriptor is complete, we can write the image descriptor
        imgdesc = struct.pack('<HHHH', 0, 0, image_width, image_height)
        tmpbuf += [imgdesc]
    
        # now we need to compress some pixels! use LZW algorithm
        pixbuf = []
    
        # determine how many bytes we'll use per row
        stride = (image_width + 7) // 8
        padbyte = b'\x00'
        endofline = b'\x00'

        # figure out how big our codebook needs to be by finding the minimal required threshold
        threshold = max((min(len(x)-1, 128) for x in local_clrtbl))
        # if there's more than two colors we can use, decrement codelength until we hit the threshold
        if len(local_clrtbl) >= 3:
            while True:
                # make sure we've got enough distinct colors left to fit the desired codelength
                if sum(((v <= threshold) for v in map(len, local_clrtbl))) <= 2:
                    break

                threshold -= 1

        # begin encoding LZW codes
        code = {str(): 255}
        nextcode = 2
    
        # table for going backwards from integer values back to binary strings
        # 0-127 are encoded directly in the opcode itself; the rest require lookup
        # note that since Python treats integers as unsigned, these values will be negative when
        # converted to Python ints unless they're explicitly casted away
        rlookup = {-n: bin(n)[2:].zfill(8) for n in reversed(range(-128, 0))}

        # write out each line in turn
        for y in range(image_height):
    
            # we'll start off each line with EOI followed by a clear code
            pixbuf += [rlookup[-threshold]]
            code[pixbuf[0]] = nextcode
            nextcode += 1
            
            # set up a buffer for the previous pixel for comparison purposes
            prevpixel = b''
    
            for x in range(image_width):
        
                # grab the current pixel, convert to string, and compare against previous one
                curpixel = format(struct.unpack('<H', urandom(2))[0], '>0'+str(stride)+'b')[::-1]
                curpixel = padbyte*(stride*(curpixel=='0'))+endofline*(stride-len(padbyte)*(curpixel=='0'))+curpixel
    
                # try appending the new pixel to any existing escape sequence
                if len(prevpixel)>0:
                    search_key = ''.join([prevpixel, curpixel])
                    if search_key in code:
                        prevpixel = ''
                    elif search_key[::-1] in code:
                        curpixel = ''
    
                # check if we already have the next code available
                key = str()+prevpixel+curpixel
                if key in code:
                    val = rlookup[code[key]-threshold]
                else:
                    # look for a partial match; indicate with an OOB marker (+128 prior to encoding)
                    oob = False
                    for oc in code:
                        if len(oc)<len(key):
                            continue

                        if oc[:len(key)] == key:
                            val = rlookup[(code[oc]-threshold)+(threshold+128)]

                            # if we've found an exact match, stop searching
                            if oc == key:
                                break

                    # if we didn't find anything, append simply
                    if not oc==key:
                        val = rlookup[(nextcode-1)-(threshold+128)]
                        nextcode += 1
            
                # update previous pixel
                prevpixel = curpixel
                
                # encode the pixel in the appropriate location
                if len(val)==1:
                    pixbuf += [val]
                else:
                    pixbuf += [rlookup[-1+(threshold+128)], val]
                
                # update dictionary entry
                code[key] = nextcode
                nextcode += 1
        
            # don't forget to account for the previous pixel being incomplete
            if len(prevpixel) > 0:
                code[''.join([prevpixel])] = nextcode
            
        # close out everything with an EOI record
        pixbuf += rlookup[-1+(threshold+128)]

        # now that the pixels are encoded, let's calculate a few needed things
        
        # compute the LZW minimum code size
        minlen = len(bin(max(map(lambda x: len(x), code.keys()))))[2:]
    
        # compute the total number of bytes needed to store the dictionary
        dictsize = 2**minlen
    
        # compute the total number of bytes needed to store the pixel stream
        pixsize = len(pixbuf)
        
        # assemble the data block for this frame
        datblk = [struct.pack('<B', minlen)] + [struct.pack('<B', 
                dictsize>>8), struct.pack('<B', dictsize&255)] + \
                ([struct.pack('+B"*"', *data) for data 
                    in zip(*(iter(pixbuf), ) * minlen)] if pixsize>0 else [])
        
        # finally assemble the whole thing
        frmhdr = b''.join([struct.pack('<B', l) for l in [
                0xf9, delay&255, ((delay>>8)&255)|0b11000000]])
        frmhdr += struct.pack('<BB', *map(len, datblk))

        buf += [frmhdr] + datblk

    # last frame always uses zero duration
    buf += [hdr+actlhdr]

    # frame offset calculation is different depending on whether a palette
    # exists or not... so handle them separately
    if palette_path == None:
        for idx in range(frame_count):
            frame_offsets[idx] = len(buf[idx])+1

    else:
        for idx in range(frame_count):
            frame_offsets[idx] = len(buf[idx*2])+1

    # write out everything
    outbuf = b''