#! /usr/bin/env python3

import errno
import random
from argparse import ArgumentParser

DEFAULT_WIDTH = 49
DEFAULT_HEIGHT = 50
DEFAULT_COLOR_TABLE_SIZE = 7
DEFAULT_BLOCK_SIZE = 8

TOOL_VERSION="0.1"



def main():

    parser = ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-v", "--version", action='store_true', help="Show version")
    group.add_argument("-o", "--output", nargs='?', default=DEFAULT_OUTPUT, const="",
                       type=str, metavar='FILE', help="Output FILE (default=%s)" % DEFAULT_OUTPUT)     

    args = parser.parse_args()

    if args.version:
        print("Version:", TOOL_VERSION)
        return 0

    try:
        f_out = open(args.output, "wb")
    except OSError as e:
        if e.errno == errno.EACCES or e.errno == errno.EISDIR:
            print("Cannot access '%s': permission denied." % args.output)
        elif e.errno == errno.ENOENT:
            print("File '%s' does not exist!" % args.output)
        elif e.errno == errno.ENOTDIR:
            print("'%s' is not a directory!" % os.path.dirname(args.output))
        else:
            raise
        exit(-1)

    rand_gen = RandomGifGenerator(output=args.output)
    rand_gen.generate_random_gif(width=DEFAULT_WIDTH,
                                 height=DEFAULT_HEIGHT,
                                 ncolors=DEFAULT_COLOR_TABLE_SIZE,
                                 blocksize=DEFAULT_BLOCK_SIZE)


class GifWriter():

    MAGIC_STRING = bytearray('GIF89a'.encode())

    def __init__(self,
                 out_file,
                 width, height,
                 num_colors, bg_color=-1,
                 loop=False, sensitivity=0.0):
        
        assert(num_colors <= 256), \
               'Number of colors must be less than 256.'

        assert(bg_color < num_colors), \
               'Background color index must be smaller than number of colors.'

        self._out_file = out_file
        self._width = int(width)    # already checked by argparser
        self._height = int(height)  # already checked by argparser
        self._bg_color = bg_color
        self._ncolors = num_colors
        self._loop = loop
        self._palette = []

        self._lzw_mincode_size = None
        self._lzw_dc_lookup = {}
        self._lzw_ac_lookup = {}
        self._lzw_clear_code = None
        self._lzw_end_of_info_code = None
        self._lzw_end_of_image_code = None
        self._lzw_reset()


        # Write Header (Logical Screen Descriptor)
        self._write(b''.join((
            self.MAGIC_STRING,                      # signature
            self._byte(self._width),                # logical screen width
            self._byte(self._height),               # logical screen height
            self._byte(int(math.log(pow(2, self._ncolors))) - 1),       # packed field
            self._byte(self._bg_color),             # background color index
            self._byte(self._ncolors),              # pixel aspect ratio
            0x00                                     # global color table
        )))

        # Write Global Color Table (optional)
        for color in range(self._ncolors):
            self._write(struct.pack("<BBB", *(random.getrandbits(8) for i in range(3))))

    
    @staticmethod
    def _byte(value):
        """ Converts value into one byte, clipped at zero boundaries."""
        if isinstance(value, tuple):
            value = sum(map(lambda x : pow(256, x[0])*x[1], enumerate(reversed(value))))
        elif not isinstance(value, int):
            raise TypeError('%r is not an integer!' % value)
        return min(max(0, value & 0xFF), 0xFF)

    def _lzw_reset(self):
        # LZW constants
        self._lzw_clear_code = ord('\x1A')
        self._lzw_end_of_info_code = 256
        self._lzw_end_of_image_code = 257
        self._lzw_mincode_size = 2
        self._lzw_dc_lookup = {
            self._lzw_clear_code: [0]*self._lzw_mincode_size,
            self._lzw_end_of_image_code: [],
        }
        self._lzw_ac_lookup = {
            self._lzw_clear_code: {},
            self._lzw_end_of_image_code: {}
        }

    def _lzw_write_subblock(self, code):
        code &= ~128                  # remove first bit from code
        len_curr_code = 0
        while code >= 128:            # While more codes to come...
            self._write(bytes([(len_curr_code+1)|128])) # set MSB on length byte
            self._write(bytes([code&0xFF]))                  # write code byte
            code >>= 8                 # move to next higher byte
            len_curr_code += 1

    def _lzw_find_longest_prefix(self, curr_char):
        best_match = None
        longest_prefix_length = 0
        for prefix, suffixes in self._lzw_ac_lookup[curr_char].items():

            match_len = len(suffixes)+1
            if match_len > longest_prefix_length:
                longest_prefix_length = match_len
                best_match = prefix

        return (best_match, longest_prefix_length)

    def _lzw_compress(self, bytes_buffer):
        new_codes = list(set(bytes_buffer))
        old_codes = sorted(list(self._lzw_dc_lookup.keys()))
        max_code_index = max(old_codes+new_codes)
        code_index = max_code_index + 1
        for byte in range(max_code_index+1):
            if byte in old_codes:
                self._lzw_dc_lookup[byte] = chr(byte)
                continue
            elif byte in new_codes:
                self._lzw_dc_lookup[byte] = ''
            else:
                continue
            
            longest_prefix = self._lzw_find_longest_prefix(byte)
            for i in range(longest_prefix[1]):
                self._lzw_ac_lookup[longest_prefix[0]][i] = ''
    
            prefix_index = longest_prefix[0]
            for i in reversed(range(longest_prefix[1])):

                code_string = str(chr(byte))+self._lzw_dc_lookup[prefix_index][:-1] 
                prefix_index = prefix_index ^ byte << ((9*i)//8)
                self._lzw_dc_lookup[code_index] = code_string[:-1]
                self._lzw_ac_lookup[prefix_index][i] = code_string
                code_index+=1
        
    def _write_graphics_control_extension(self):
        pass

    def _write_application_extension(self):
        pass
    
    def _write_comment_extension(self):
        pass
    
    def _write_plain_text_extension(self):
        pass

    def _write_trailer(self):
        self._write(b'\x3B')

    def _write_frame(self, pixels):
        lzw_data_stream = bytes()
        for row in pixels:
            lzw_data_stream += bytes([row['index']])
            self._lzw_compress(row['c'])
            
        lzw_data_stream += bytes([
            self._lzw_end_of_image_code
        ])

        # Write Image Descriptor
        self._write(b''.join((
            self._byte(0),                        # left position is unused
            self._byte(0),                        # top position is unused
            self._byte(self._width),              # frame width
            self._byte(self._height),             # frame height
            self._byte(self._lzw_mincode_size-1), # lowest resolution code size
            b'\x00\x00',                         # no local color table
            0x00                                 # interlace flag, sort flag, reserved bits should be zero
        )))

        # Compression method indicator
        self._write(b'\x08')

        # Write Graphic Control Extension
        self._write_graphics_control_extension()

        # Write Plain Text extension
        self._write_plain_text_extension()

        # Write Application Extension
        self._write_application_extension()

        # Write Comment Extension
        self._write_comment_extension()
        

        # LZW MinCodeSize
        self._write(bytes([self._lzw_mincode_size]))

        self._write_subblocks(lzw_data_stream, self._