#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# provided RNG. The size of the generated image is determined by
# the first four bytes read from the RNG (which are discarded).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, codelen: int
    ):

    # We'll write the length as little-endian UINT32
    def write_uint32(out, val):
        out.write((val & 0xff).to_bytes(1, 'little'))
        out.write(((val >> 8) & 0xff).to_bytes(1, 'little'))
        out.write(((val >> 16) & 0xff).to_bytes(1, 'little'))
        out.write(((val >> 24) & 0xff).to_bytes(1, 'little'))

    # We'll write the code len as little-endian UINT16
    def write_uint16(out, val):
        out.write((val & 0xff).to_bytes(1, 'little'))
        out.write(((val >> 8) & 0xff).to_bytes(1, 'little'))

    # First we need to encode the given number of bits in LZW
    # and our dictionary (which needs one extra entry for the
    # final character 0xffff)
    lzwcode = []
    dictsize = pow(2, codelen)
    dict = [None] * (dictsize+1)
    w = ''
    r = rng.read(1)[0] or 1
    while True:
        if r < pow(2, codelen - 7):
            nbits = (r >> 5) + 2
        else:
            nbits = ((r >> 5) & 0x3f) + 3
        
        x = 0
        for i in range(nbits):
            x |= (((r >> 5) & 1) << i)

        r <<= nbits
        e = lzwcode[x] + str(w[-1])
        lzwcode += [e]
        dict[len(lzwcode)-1] = dict[len(w)] + w[-1]
        w = e
        if len(lzwcode) == dictsize:
            break
    
    out.write('GIF89a'.encode())
    write_uint16(out, codelen)     # Logical screen descriptor
    out.write(b'\xf7\x00')         # Global color table
    out.write('\x00\x00'.encode())  # Background color index
    write_uint16(out, 0x1f)        # Pixel aspect ratio
    write_uint32(out, 0xffffffff)  # Image descriptor
    write_uint16(out, 0xd8ff)      # Size of global color table
    out.write(struct.pack('<H', 0))  # Palette index
    out.write(b'\x00\x00\x00')      # Transparent color
    write_uint16(out, 0x000d)      # Block size
    write_uint16(out, 0x0000)      # Block size
    write_uint16(out, 0x0000)      # Grid width
    write_uint16(out, 0x0000)      # Grid height
    out.write(b'\x00\x01\x00\x00\xff\xff\xff\xff')   # Block terminator
    out.write(struct.pack('<H', len(lzwcode)))    # Block size
    maptable = [i for i in range(-1, len(lzwcode)-1)]
    out.write(bytearray([maptable.index(c) for c in lzwcode]))
    out.write(b'\x00\x00')  # Block terminator; end of stream

