#!/usr/bin/env python3

import os
from itertools import cycle

import numpy as np

# Generates a random GIF file into `out` using the
# random number generator `rng`. Use `/dev/urandom`,
# but attempt to seed from userland sources (eg. /dev/arandom on Linux);
# otherwise fall back to the given seed value for testing purposes
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, *,
        seed: bytes=None, globalseed: bool=False
    ):
    blocksize = None
    data = b''
    padding = [] # Padding required by the block alignment rules
    header = b'GIF89a'

    if not seed:
        # Try to get some entropy from the system or other devices
        try:
            seed = bytes([*os.urandom(4)])
        except Exception as e:
            print('WARNING:', e)

        if not seed:
            raise RuntimeError('No suitable source of entropy!')

    # Seed our RNG with the seed value provided to us
    if globalseed:
        os.urandom(len(seed))

    rng = np.random.default_rng(seed)
    while True:
        # Generate some random data
        genlen = max((rng.integers(maxval=5) + 1) * blocksize, blocksize) // 3
        if not genlen: break # Avoid generating an empty slice
        data += bytes(rng.integers(low=0, high=2 ** 8, size=(genlen,), dtype=np.uint8))

        newpadding = min((genlen * 3) % blocksize, blocksize - len(data) % blocksize)
        if newpadding:
            padding += [bytes([(newpadding)])]

            # If we've accumulated enough padding, write it all at once and begin again
            if sum(map(len, padding)) >= blocksize:
                out.write(header + len(data).to_bytes(3, 'big'))

                while any(padding):
                    chunk = next(cycle(iter(filter(lambda x: x!= [], padding))))

                    if len(chunk) <= blocksize:
                        out.write(data[:blocksize-len(chunk)] + chunk)
                        del data[:blocksize-len(chunk)]

                        padding.remove(chunk)
                    elif type(chunk) == str:
                        out.write(data[:blocksize] + chunk.encode())
                        del data[:blocksize]

                        padding.remove(chunk)
                    else:
                        out.write(data[:blocksize])
                        del data[:blocksize]

    # Finish writing everything left over in the buffer
    if data or any(padding):
        while any(padding):
            chunk = next(cycle(iter(filter(lambda x: x!= [], padding))))

            if len(chunk) <= blocksize:
                out.write(data[:blocksize-len(chunk)] + chunk)
                del data[:blocksize-len(chunk)]

                padding.remove(chunk)
            elif type(chunk) == str:
                out.write(data[:blocksize] + chunk.encode())
                del data[:blocksize]

                padding.remove(chunk)
            else:
                out.write(data[:blocksize])
                del data[:blocksize]

