#!/usr/bin/env python3

import sys
import os
import struct
from typing import BinaryIO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output


    def generate_random_gif(self):
        
        with open('/dev/urandom', mode='rb') as rng,\
             open(self.output,    mode='wb') as out:

            return generate_random_gif(
                    rng, out
                )



# Generates an animated, colorful GIF file of a
# predetermined length into `out`, based on the
# random numbers generated by the urandom pseudo-device
# at `rng`.
def generate_random_gif(rng: BinaryIO, out: BinaryIO) -> None:

    # Header
    header = b''
    header += b'GIF89a'                # File Signature and Version
    header += bytes([
        255 & ((rng.read(1)[0] << 1) | 1),      # width
        255 & ((rng.read(1)[0] << 1) | 1),      # height
        rng.read(1)[0],                       # bit depth
        7                                    # 2^7 colors max
    ])
    # Global Color Table
    gctsz = rng.read(1)[0]                   # global color table size
    gct = [bytes([rng.read(1)[0]] * 3)]       # set up GCT
    for _ in range(gctsz - 1):                 # fill up until we have enough colors
        clr = bytes([rng.read(1)[0]] * 3)
        while clr in gct:
            clr = bytes([rng.read(1)[0]] * 3) 
        gct.append(clr)                        # append to our list
    for clr in gct:                           # build up full buffer
        header += clr
    # Image Descriptor
    header += bytes([
        rng.read(1)[0],             # left position
        0,                         # top position
        rng.read(1)[0],             # width
        0b00010000 |               # local color table flag
        rng.read(1)[0]              # interlaced flag and block size
    ])
    # Local Color Table
    lctsz = rng.read(1)[0]                  # local color table size
    lct = []                                 # create empty LCT
    for i in range(lctsz):                   # iterate over entries
        clr = bytes([rng.read(1)[0]] * 3)   # read entry
        while clr in lct or clr == gct[i]:   # check for duplcates
            clr = bytes([rng.read(1)[0]] * 3)  
        lct.append(clr)                      # append unique ones
    for clr in lct:                          # build up full buffer
        header += clr
    # Image Data (Image Data Sub-Blocks within Data Stream)
    bs = rng.read(1)[0]                     # read blocksize
    subblock = rng.read(bs - 1)             # discard sub-block marker
    x = y = 0                               # initial position
    while True:                             # iterate until end of animation
        pos = bytearray([x % 256, y % 256])   # calculate current absolute position
        pos.extend((y >> 8).to_bytes(1, "big"))
        pos.extend((x >> 8).to_bytes(1, "big"))
        imgdata = bytes(pos)                 # position info in first three bytes
        imgdata += bytes([
            0b00000000 |                                   # no palette change
            rng.read(1)[0],                                # local color index
            rng.read(1)[0]                                 # LZW min code size
        ])
        buf = bytearray()                      # create an empty buffer
        while True:                            # loop until we run out of stuff
            buf.append(rng.read(1)[0])          # put next value onto buffer
            if len(buf) > 254: break           # stop when buffer is full
            if buf[-1] >= 255: continue         # ignore last two bits
            if buf[-2] < 128: break             # stop when we reach an EOL
            if buf[-2] >= 255: continue         # ignore last two bits
        nxt = rng.read(1)[0]                    # read next byte
        if nxt <= 1:                            # only allow one zero-length sequence
            nxt += rng.read(1)[0] + 1           # so there's always room to move forward
        imgdata += bytes([nxt])                # add nxt to buffer and done
        imgdata += bytes(buf)
        pos = bytearray([x % 256, y % 256])   
        pos.extend((y >> 8).to_bytes(1, "big"))
        pos.extend((x >> 8).to_bytes(1, "big"))
        out.write(header + imgdata)
        x += 1                                  # increment x; goto next line
        if x!= 2 ** ord(header[6]): continue    # check bounds
        x = y = 0                              # reset x, y after eol
        y += 1                                  # increment y
        if y == 2 ** ord(header[7]): break      # check bounds

