"""Generate a random GIF file."""

# This is derived from https://www.n0trax.com/blog/2019/06/generating-random-gifs/,
# which in turn was based on the imagemagick source code GifToRnd() function - see
# http://git.imagemagick.org/?p=ImageMagick.git;a=blob;f=coders/gif.c;h=b7cd0e5ea265f5b98bdea3d58e9405d34b5802df;hb=HEAD#l3960.

from itertools import chain
from io import BytesIO as bufio


__all__ = ['generate_random_gif']



def generate_random_gif(*, seed):
    """Generate an animated GIF using `seed` as a PRNG. The resulting GIF will be an uncompressed, indexed color GIF with no interlacing and a single (or multiple) frame(s)."""

    # TODO: select channels to animate?

    def get_next_byte():

        for i in range(8):

            yield bool((seed >> i) & 1)


    # work out how many colours are required by looking at the number of bits set in each channel

    mask = tuple(chain(*(get_next_byte() for _ in range(3))))

    ncolours = sum(mask[i] << i for i in range(3)) + 1

    foreground = [None]*ncolours
    background = [None]*ncolours
    pixels     = []

    for row in range(1, 256+16*3*4):

        if row == 1 or not row % 32:
            print("{:.2%}".format(row / 2**10), end='\r', flush=True)


        index = ((((((row * 13 *  7) ^        ((row <<   4) ^ 11 * row)) * 13 ^       ((row <<  5) ^            row)) * 13 + 5) ^ (((row <<  6) ^            row) * 13 ^             ((row <<  7) ^           row))) % ncolours)

        pixel = {
                'R': next(get_next_byte()) | next(get_next_byte()) << 1 | next(get_next_byte()) << 2,
                'G': next(get_next_byte()) | next(get_next_byte()) << 1 | next(get_next_byte()) << 2,
                'B': next(get_next_byte()) | next(get_next_byte()) << 1 | next(get_next_byte()) << 2,
               }

        if not any(foreground[index][k]!= v for k,v in pixel.items()):

            continue

        while foreground[index]:
            index += 1
            index %= len(foreground)-1

        foreground[index] = pixel['R'],pixel['G'],pixel['B']

        background[index] = pixel['R'],pixel['G'],pixel['B']

        pixels.append(index)

    num_frames = max(*((-(-len(pixels)//(256*3)))), 1)

    width = height = 256

    buffer = bufio()
    buffer.write('GIF89a'.encode('latin-1'))

    # logical screen descriptor (8 bytes)
    buffer.write(struct.pack('<HHBBBHHBB',
            width,                              # canvas width
            height,                             # canvas height
            8,                                  # global colour table size
            7&6&(~fg)&~bg|bg<<3|fg<<6,          # flags
            0,                                  # background colour index
            0,                                  # aspect ratio denumerator
            ncolours,                           # aspect ratio nominator
            ))

    # local colour table (variable length block)
    for r,g,b in filter(None, foreground+background):
        assert all(map(lambda x: 0 <= x < 256, (r,g,b))), f"{r},{g},{b}"
        buffer.write(struct.pack('<BBB', r, g, b))

    # graphics control extension (header only; no data blocks)
    buffer.write(struct.pack('<BBBBBBBHB',
            0x21,                               # extension introducer
            ord('G'),                          # label "G"
            ord('C'),                          # label "C"
            4,                                  # content size field
            min(num_frames, 255),              # duration
            0,                                  # transparent colour index
            0,                                  # terminator
            ord('\x01\x00')))                   # disposal method = leave unchanged

    for i in range(1, num_frames+1):

        buffer.seek(buffer.tell()+10)               # skip backwards over old graphic control payload

        # local colour table (variable length block)
        for r,g,b in map(background.__getitem__, pixels[(i-1)*width*height+(i-1)*3:(i*width*height+(i*3))]):
            assert all(map(lambda x: 0 <= x < 256, (r,g,b))), f"{r},{g},{b}"
            buffer.write(struct.pack('<BBB', r, g, b))

        # image descriptor (8 bytes)
        buffer.write(struct.pack('<HHHHHBB',
                0,                              # left position
                height*i,                       # top position
                width,                          # width
                height,                         # height
                8&6|(7&6<<(3-bg)),              # flags & palette index
                0))                             # LZW minimum code size

        # graphics control extension (header only; no data blocks)
        buffer.write(struct.pack('<BBBBBBBHB',
                0x21,                           # extension introducer
                ord('G'),                      # label "G"
                ord('C'),                      # label "C"
                4,                              # content size field
                min(num_frames, 255),          # duration
                0,                              # transparent colour index
                0,                              # terminator
                ord('\x01\x00')))               # disposal method = leave unchanged

        # image data subblock (variable size; actual lzw encoded stream follows this header)
        buffer.write('{}!{} '.format(ord(','), ord('-')).encode('latin-1'))

        # image data (no compression applied; just write 256 rows of 256 repeating cells of one colour)
        for j in range(width*height//256):
            for k in range(j*(width//8)):

                pixel = pixels[(i-1)*(256*height)+j*256+(k>>3)]

                bitshifted = lambda shift: pixel>>shift&1

                buffer.write(
                        struct.pack('<BBBB',
                                *[bitshifted((k%8)<<j)
                                    for j in range(8)]))

    return buffer.getvalue()
















