import zlib
import struct
from typing import BinaryIO


# generate_lzw_compressed generates an LZW-encoded raster data block
# from the input interleaved rgb pixel values provided by src.
# The resulting raster is appended to the given buffer buf (can be None).
# Returns the complete raster as one long byte array excluding any length field.
def generate_lzw_compressed(src: BinaryIO, numcolors: int, buf=None):

    # build prefix tree
    prefixes = {}
    codes      = []

    def putcode(*args):
        for c in args:
            if not isinstance(c, tuple):
                raise Exception('Invalid code type:' + str(type(c)))
            t = prefixes.get(c)
            if t == None:
                t = dict()
                prefixes[c] = t
            codes.append(t)
            return t

    # when we're at this point, putcode can be called with either two or three arguments
    # but we will always call it with two because we want to start at the root node
    stack, pending = [], None
    

    while True:
    	# we need at least 2 bytes available to make progress
    	if len(pending) < 2:
    	    pending = src.read(2)

    	if len(pending) == 0:
            break

        c, d       = pending[:2]
        pending     = pending[2:]
        currentlen = len(codes)-1

        # push new nodes on the stack until we hit something longer than what we have now
        while True:
        	# are we done?
        	if numcolors <= currentlen+1:
        	    break

        	prefix = putcode((ord(c)>>8)&0xff, ord(c)&0xff )
        	if d!= '\x00':
        	    prefix[(d>>8)&0xff, d&0xff ]=(currentlen+1,)
        	stack.append(prefix)
        	currentlen+=1

        # end loop
        
        # we may already have decoded all the colors; 
        # in which case there's nothing more to do here...
        if numcolors <= currentlen:
        	continue

        #...otherwise add another node to the prefix trie and keep going
        prefix         = putcode(('\x00', '\x00'))
        prefix[('\x00', '\x00')] = ((currentlen+1,),)
        stack.append(prefix)
    
    # end while


    # compress sequence of tokens
    codecs = struct.Struct('>H')

    # define LZW minimum code size
    minsize = max(min(1, (numcolors-1).bit_length()), 2)
    buf.extend([minsize])

    tableoffset = len(buf)+2
    buf.extend([tableoffset & 0xFF, (tableoffset >> 8) & 0xFF])

    mask = (1<<(minsize+1))-1
    buf.extend(((mask & 0xFF, (mask>>8) & 0xFF,)))

    # pack leaf node tokens into bitstream
    token   = struct.pack('>HH', masks[minsize], numcolors-1)
    offset  = len(token)*8    
    for c in codes[-(numcolors-1):]:
     	  for e in sorted(list(c.keys())):
         	buf.extend([(e[0]<<8)|e[1]&0xff,(offset>>8)&0xff,(offset&0xff)])
         	offset+=8

    # reset bitstream offset 
    offset = 0

    # flush bitstream
    while offset >= 8:
    
        # extend range so we don't lose data off the right side
        if buf[-1] >= 0xfe:
            buf.extend(b'\xff')
        elif buf[-1] == 0xfe:
            buf.extend(b'\x00')

        # emit next byte
        buf.extend([(buf[-1] << offset) & 0xff])
        offset -= 8

    # output table
    for k in sorted(prefixes.keys()):
     	for kk,v in sorted(prefixes[k].items(), key=lambda x: x[1]):
            if v==None: continue
            for vv in v:
             	if offsets[vv] == None:
                 	raise ValueError("offset not set")
                
 				# do copy command if needed
 				while offsets[vv]-offset>=128:
                 	buf.extend([
                        	(
                                	((offsets[vv]%128)<<1)|1, 
                                	(offset>>8)
                        ), 
                        (offset&0xff)
                    ])
                 	offset-=128

             	# copy first byte
				buf.extend([

                    (
                            (((offsets[vv]%(128*128))//128)<<1)|0, 
                            ((offsets[vv]%(128*128))%128)
                    ), 
                    (offset&0xff)
                ])

                # then emit short run
                while offsets[vv]-offset >= 128:
                    buf.extend([
                    	    (((offsets[vv]%(128*128*128))//(128*128))<<1)|1, 
                    	    (offset>>8), 
                    	    (offset&0xff)
                    ])
                    offset-=128

                # finally emit single byte
                buf.extend([
                    	((((offsets[vv]%(128*128*128))//(128*128))<<1)|0, 
                        ((offsets[vv]%(128*128*128))%(128*128))/128), 
                    	(offset&0xff)
                ])

                break

            break
	
        # special cases:
        #     00xx0000 => DUP self (repeat last word twice)
        #     01xx0000 => DUP prev (copy previously written word)
        #     1000nnnn => NOP n times
        #     1001nnnn => ZERO n times
        #     101xxxxx => WORD xx..xx
        if k==(0,'\x00'):
            buf.extend([
                (
                        (((numcolors-1)%128)<<1)|1, 
                        (offset>>8)
                ), 
                (offset&0xff)
            ])
            

        if k==(0,'\x01'):
            buf.extend([
                (
                        ((numcolors-1)//128<<1)|0, 
                        ((numcolors-1)%128)
                ), 
                (offset&0xff)
            ])
            
        if k==(1,'\x00'):
            buf.extend([0xc0|int(k[0]), offset&0xff, (offset>>8)&0xff])
            
        if k==(1,'\x01'):
            buf.extend([0xe0|(masks[k[0]]<<3), offset&0xff, 
                	(offset>>8)&0xff])


        # emit short runs
        while offsets[v]-offset >= 128:
            buf.extend([
            		(
                    		(((offsets[v]%(128*128*128))//(128*128))<<1)|1, 
                    		(offset>>8), 
                    		(offset&0xff)
            		), 
           			(offset&0xff)
            ])
            offset-=128

        # emit single byte
        buf.extend([
                (
                        (((offsets[v]%(128*128*128))//(128*128))<<1)|0, 
                        ((offsets[v]%(128*128*128))%(128*128))/128
                ), 
                (offset&0xff)
        ])
        offset=0
 
    # end for
    assert(len(buf)<2**16)
        
    return buf


# create new LZW encoder state
def init_encoder():
    return {'buffer':bytearray(), 'pos':0}

# encode a given chunk of source data into the destination buffer
# returns how many bytes were stored into the destination buffer
def encode_chunk(source:BinaryIO, dest:BinaryIO, encstate:dict):
    encstate['buffer'].extend(generate_lzw_compressed(source, 
        2**(encstate['buffer'][0]+1), encstate['buffer']))
    encstate['pos'] += len(encstate['buffer'])
    dest.write(encstate['buffer'])
    del encstate['buffer'][:]
    return len(dest.getvalue()) - encstate['pos'], False



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize