#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# It also writes the palette information into `pal`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, pal: BinaryIO):


    # header is static except for the version string
    header = b'\x47\x49\x46\x38\x39\x61'

    # the rest varies depending on whether there's a palette or not
    # this tuple contains the appropriate parts of the final header
    # as well as the length of part that contains the LZW-compressed
    # image data
    hdrparts = ((0,), (1,), (0,), (0,), (0,), 
        (int(os.environ['PALETTE']) & 1,), (), ())
    
    def join(*args):
        return sum(args, ())
    
    # unpacked form of hdrparts specifies how many elements of the 
    # binary representation of an integer should be stripped off
    # the end before it can be concatenated onto the rest of the header
    hdrnumsizes = (*(len(x) for x in hdrparts),)
    hdrformat = ''.join(('H',)*sum(hdrnumsizes))

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # construct the header based on whether we're going to use
    # a palette or not; note that the version field gets overwritten later
    headersize = len(header) + struct.calcsize(hdrformat)
    header = header[:headersize] + struct.pack(hdrformat, *(
        join((image_width, ), hdrparts[0]),
        join((image_height, ), hdrparts[1]),
        join((int(os.environ['PALETTE']), ), hdrparts[2]),
        join(((int(os.environ['PALETTE']) >> 1) & 1), hdrparts[3]),
        join(((int(os.environ['PALETTE']) >> 2) & 1), hdrparts[4]),
        join((((int(os.environ['PALETTE'])) >> 3) & 1), hdrparts[5])
    ))

    # size of the color table -- includes the initial zero entry!
    tbllen =  pow(2, int(os.environ['PALETTE']))
    if tbllen < 256:
        tbllen |= 0b10000000

    # get data from RNG, possibly including palette entries
    numpals = (tbllen >> 7) + 1 if int(os.environ['PALETTE'])!= 0 else 0
    numdata = tbllen*numpals
    data = rng.read(numdata+2)

    # write to output stream
    out.write(header + data)

    # write out palette (if needed)
    if numpals == 1:
        palentry = struct.pack('BBB', 0, 0, 0)
        # write out the length byte first
        pal.write(struct.pack('B',1)+palentry)

    elif numpals > 1:
        # the zeroth palette entry is already present in the GIF spec, so start at one
        for i in range(1, numpals):

            # create a random RGB value for the new palette entry
            c = rng.read(3)
            
            # write out the length byte first
            pal.write(struct.pack('B',1)+c)