#!/usr/local/bin/python3

import sys
import os
import struct
from typing import BinaryIO

PRINTABLES = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789!\"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~"
DEFAULT_OUTPUT = 'random.gif'




class RandomGifGenerator():

    def __init__(self, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        elif type(output) == str:
                self.output = output
        else:
            raise ValueError('Output must be of type string.')


    def generate_random_gif(self, width, height, numframes=16):
        
        with open('/dev/urandom', mode='rb') as rng,\
             open(self.output,    mode='w') as out:
            
            generate_random_gif(
                    rng, 
                    out,
                    width,
                    height,
                    numframes
                )



# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# are read from different sources to ensure that they're valid.
def generate_random_gif(rng: BinaryIO, out: BinaryIO, width: int,
                        height: int, numframes: int = 16) -> None:
    

    # Obtain palette
    palette = [b'\x00\x00\x00']               # black
    numcolors = len(palette)                 
    changefreq = max(numcolors // numframes,  1) 
    buffer = rng.read(2)                     # two byte flag indicating next action
    while True:                              # keep looping until we encounter a '!' character
        f = ord(buffer[-1:])                  # last byte contains frame control flags
        framecontrol = {                      # decode them here
                0 : 'end',                    # end of frames
                1 : 'ext_introducer',         # application extension
                2 : 'graphics_control',        # graphics control extension
                3 : 'image_descriptor',        # image descriptor
                4 : 'table_based_image',       # table based image data
                5 : 'table_based_extension'     # table based graphic control extension
            }                               
        print(framecontrol[f % 6], end='')   
        if f >= 128:                          # first bit set indicates a comment or plain text
            print()                           # ignore comments 
            break                              
        elif f > 4:                           # any other value except 0 or 128 corresponds to plain text
            buffer = buffer[:-1]              # delete this flag (or not ;))
            length = ord(buffer[-1:])         # last byte gives text length
            buffer = buffer[:-1]              # now cut off the length byte itself
            ascii = ''                        # convert bytes to ASCII
            for i in range(-length, 0):      # iterate over the remaining bytes
                ascii += chr(ord(buffer[i]))
            for character in ascii:           # encode the text
                if character!= '!':          # but ignore the '!' character because it terminates the entire sequence
                    out.write(character)
            break                              
        elif f < 4 and (f % 2 == 0):           # odd values between 4 and 128 correspond to a color entry
            
            buffer = buffer[:-1]              # already read one byte so remove it
            r = ord(buffer[-1:]) % 36         # red component 
            g = ord(buffer[-1:]) % 36         # green component 
            b = ord(buffer[-1:]) % 36         # blue component
            
            # check if new colors need to be added
            if palette.count((r,g,b)) <= changefreq:
                palette += [(r,g,b)]          # adding three bytes (the new rgb tuple) stores them in an array

            print("{:0>2x}{:0>2x}{:0>2x}".format(*palette[-1]), end="")
            buffer = rng.read(2)

        
        elif f == 0:                          # when no more animation is left simply exit
            buffer = rng.read(1)  
            continue                           
        elif f % 6 == 1:                       # if the frame control indicator is either 1 or 2 
            buffer = rng.read(1)               # the next byte will contain an extended block identifier
            exttype = {                         # map these types 
                1 : 'application',               # application extension
                0xF9 : 'graphic_control'         # graphics control extension
            }.get(ord(buffer), '?')             # otherwise use? as fallback
            print("{:<5s} ".format(exttype), end="")
            buffer = rng.read(1)               # useless byte lol
            buffer = rng.read(1)               # another useless byte lol
            print("{:0>2x}".format(ord(buffer)), end="")
            buffer = rng.read(1)               # yet another useless byte lol
            if f >= 128:                      # handle application extensions
                buffer = rng.read(1)           # another useless byte lol
                length = ord(buffer[-1:])     # last byte gives text length
                buffer = buffer[:-1]          # now cut off the length byte itself
                appname = ""                  # convert bytes to ASCII
                for i in range(-length, 0):  # iterate over the remaining bytes
                    appname += chr(ord(buffer[i]))

                appid = ""                    # convert bytes to ASCII
                for i in range(-length, 0):  # iterate over the remaining bytes
                    appid += chr(ord(buffer[i]))
                    
                print(" {} {}".format(appname, appid), end="")
                
            else:                             # handle graphic control flags 
                # 0: transparency flag (1: transparent; 0: opaque)
                T = bool(ord(buffer[-1:]) & 1)  
                # 1-3: delay time (lower four bits)
                D = (ord(buffer[-1:]) >> 1) * 256  
                # 4-7: transparent color index
                TC = ord(buffer[-1:]) >> 4       
                print("\ttransparency {}, delay {}, transcolor {}".format(T,D,TC), end="")
                buffer = rng.read(1)           # yet another useless byte lol
                print(", disposal method {:d}".format(ord(buffer[-1:])), end="\n")



        elif f == 3:                          # read image descriptor flags after each image has been processed

            # 0: local color table
            # 1: interlace flag
            I = bool(ord(buffer[-1:]) & 64)   
            # 2: sorted color table
            S = bool(ord(buffer[-1:]) & 32)   
            # 3: global color table
            G = bool(ord(buffer[-1:]) & 16)   
            # 4-5: reserved for future use
            R1 = ord(buffer[-1:]) >> 4 & 3   
            # 6-7: LZW min code size
            MS = ord(buffer[-1:]) >> 6 & 3  
            print("local colortable :{}|interlace:{}|sorted:{}|global:{}".format(I,S,G,(R1<<2)+MS), end="")
            buffer = rng.read(1)               # another useless byte lol

            xpos = ((ord(buffer[-1:]) << 8) | ord(rng.read(1)[-1:])) & 0xfff
            ypos = ((ord(buffer[-1:]) << 8) | ord(rng.read(1)[-1:])) & 0xfff
            wx = (((ord(buffer[-1:]) << 8) | ord(rng.read(1)[-1:])) & 0xfff)//256 
            wy = (((ord(buffer[-1:]) << 8) | ord(rng.read(1)[-1:])) & 0xfff)//256 

            print("\tx={:0>4},y={:0>4},width={},height={}".format(xpos, ypos, wx, wy), end="")
            buffer = rng.read(1)               # yet another useless byte lol
            print(",LZW min codesize {:.0f}".format((ord(buffer[-1:]) & 0xf)+1), end="\n")
                        
            buffer = rng.read(1)               # wait... what's going on???
            LZWmincodesize = ord(buffer[-1:])  # just take the lowest nybble lol

            buffer = rng.read(1)               # how many bytes do we need for the full line?
            charsinline = ord(buffer[-1:])//LZWmincodesize

            # TODO: implement proper lzw decompression
            # https://stackoverflow.com/questions/40046078/decompressing-lzw-compressed-data-with-python
            print("[{}]".format(','.join(["{:0>2x}".format(b) for b in rng.read(charsinline)])), end=" ")



    # Header
    header = '\x47\x49\x46\x38\x39\x61'   # File signature and version
    header += '{:c}'.format(len(palette)*3).encode()         # Color Table size