#!/usr/bin/env python3

import os
import sys
import struct
import time

from PIL import Image


# Generates a random GIF file into `out` using the
# provided RNG. The width and height of the generated image
# is determined by the command line arguments.
if __name__ == '__main__':

    # Check for proper usage.
    if len(sys.argv[1:])!= 3:
        print('Usage:', sys.argv[0], 'WIDTH HEIGHT OUTFILE')
        exit()
    w, h, outfn = [int(arg) for arg in sys.argv[1:]]
    
    # Select a suitably deterministic PRNG.
    seed = hash((time.time(), os.urandom(256))) \
            | ((w << 16) ^ h)
    rng = struct.Struct('<QQQQQ').pack(*os.urandom(16))
    prng = struct.Struct('<QQQQQ').pack(seed, *(rng[-4:] * 4))

    print(f'seed={hex(seed)}', f'datalen={len(prng)}')
    
    # Generate an image with a pseudorandom palette. Then build
    # a stream containing that palette followed by pixel data.
    img = Image.new("P", (w, h)).convert('RGB')    
    pixdata = list()
    for y in range(h):
        for x in range(w):
            coloridx = ord(img.getpixel((x,y))[0]) % 256
            pixdata += [coloridx]
    pal = bytearray([i for t in zip(*(img.palette.getdata()[1])) for i in t])
    assert len(pal) <= 768        # This should never happen!
    data = pal + bytes(pixdata)

    # Write out the GIF file.
    with open(outfn, "wb") as out:

        # Write the header and image descriptor.
        out.write(b"GIF89a")
        out.write(struct.pack("<HHHHB", w, h, 0xf000, 0, 0))
        
        # Write the logical screen descriptor plus global color table.
        colorcount = min(2 ** w, 2 ** h);
        out.write(struct.pack("<BBBHBB",
                colorcount - 1,
                0,    # No sort flag
                0,    # GCT exists; no other fields to set
                768,  # Global color table size 
                0,    # Background color index
                0     # Pixel aspect ratio
            ))
        out.write(pal)
       
        # Write the loop control block.
        out.write(struct.pack("<HHBBBBBHB",
                 0,          # Loop count
                 0,          # Initial delay 
                 0,          # Transparent color index
                 b'\xff',    # Terminator
                 0,          # Block terminator
                 b'\x21',    # Extension introducer
                 0,
                 4,
                 10         # Text extension label length
             ))
        out.write(b'text\0')   # Plain text label        
        out.write(struct.pack('<BBBBBBB',
              b'\x01',      # Block size
              b'\xff',      # Left margin
              b'\xff',      # Right margin
              b'\x00',      # Font size
              b'\x00',      # Style flags
              0,            # Foreground red component 
              0             # Foreground green component
        ))
        out.write(b'\0')       # Font family name    
        out.write(b'\0')       # Font subtype

        # Write the graphics control block.
        out.write(struct.pack('<BBBBBBBHB',
                 4,           # Block size
                 0,           # Packed field
                 0,           # Delay time 
                 0,           # Transparent color index
                 b'\x00',     # Terminator
                 0,           # Block terminator
                 b'\x21',     # Extension introducer
                 0,
                 5,
                 0   
        ))               # Application extension label
        out.write(b'NETSCAPE2.0');                 
        out.write(struct.pack('<BBBBBBB',
              b'\x03',     # Block size
              b'\x01',     # Subblock size
              b'\0',       # Loop subblock id 
              b'\0',       # No flags
              b'\0',       # Loop count
              b'\0',       # Iterations per second
        ))
        out.write(b';')

        # Write the plain text extension block.
        out.write(struct.pack('<BBBBBBBHB',
                 24,              # Block size
                 b'\x01',         # Extension introducer
                 0,               # Label block size
                 1,               # Plain text label
                 0,               # Block terminator
                 0                # Block terminator
        ));                    
        out.write(b'Created By Python on Unix at ')
        out.write(time.asctime().encode())
        out.write(b'.')
        out.write(struct.pack('<BBBBBBB',
              b'\x00',     # Block size
              b'\x00',     # Sub-block size
              b'\x00',     # Block terminator
              b'\x00',     # Block terminator
              b'\x00',     # Block terminator
              b'\x00'      # Block terminator
        ))

        # Write the application extension block.
        out.write(struct.pack('<BBBBBBBHB',
                 11,               # Block size
                 b'\x01',          # Application extension label
                 b'\x0B',          # Sub-block size
                 b'\x08',          # Block ID string
                 b'\x03',          # Flags
                 b'\0',            # Client block size
                 0,                # Sub-block size
                 0,                # Block terminator
                 0,                # Block terminator
        ));                    
        out.write(b'ICCRGB')      
        out.write(struct.pack('<BBBBBBB',
              b'\x00',     # Block size
              b'\x00',     # Sub-block size
              b'\x00',     # Block terminator
              b'\x00',     # Block terminator
              b'\x00',     # Block terminator
              b'\x00'      # Block terminator
        ))

        # Write the graphics rendering block.
        out.write(struct.pack('<BBBBBBBHB',
                 3,                  # Block size
                 b'\x2C',            # Graphics control label
                 b'\x0A',            # Sub-block size
                 b'\x00',            # Index of first color map entry
                 b'\x00',            # Length of color table
                 b'\0',              # Size of local color table
                 b'\0',              # Block terminator
                 0                   # Block terminator
        ));                    
        out.write(struct.pack('<BBBBBBB',
              b'\x00',     # Block size
              b'\x00',     # Sub-block size
              b'\x00',     # Block terminator
              b'\x00',     # Block terminator
              b'\x00',     # Block terminator
              b'\x00'      # Block terminator
        ))

        # Write the end marker.
        out.write(b';')

        # Write the actual pixels themselves.
        out.write(data[:datalen])