#!/usr/bin/env python3

import itertools as it
import struct
from typing import BinaryIO


TRANSPARENT = (
    0xFFFFFF, # white
    0xFEFEFE, # light grey
    0xFCFCFC, # gainsboro
    0xFFE4E1, # moccasin
    0xFAF0E6, # linen
    0xFAEBD7, # antique white
    0xFDF5E6, # old lace
    0xFFFFF0, # ivory
    0xC0C0C0, # silver
    0x808080, # gray
    0x800000, # red
    0x8B4513, # saddle brown
    0xBDB76B, # dark khaki
    0xA52A2A, # firebrick
    0xD2691E, # chocolate
    0xFFA07A, # light salmon
    0xFFA500, # orange
    0xFFD700, # gold
    0x9ACD32, # yellow green
    0xADFF2F, # green yellow
    0x4682B4, # steel blue
    0x4169E1, # royal blue
    0x0000FF, # deep sky blue
    0x00FFFF, # cyan / aqua
    0xE0FFFF, # lightcyan
    0xDCDCDC, # gainsboro
    0xF0FFF0, # honeydew
    0xE6E6FA, # lavender
    0x808080, # gray
    0x000080, # navy
    0x008080, # teal
    0xA9A9A9, # dimgray
    0x8B0000, # maroon
    0x800080, # purple
    0xFF0000, # red
    0xFF00FF, # fuchsia
    0xFFC0CB, # pink
    0xFF8C00, # darkorange
    0xE9967A, # bisque
    0x66CDAA, # medium aquamarine
    0x483D8B, # darkslateblue
    0xB0C4DE, # lightslategrey
    0xEE82EE, # violet
    0xD8BFD8, # thistle
    0xBA55D3, # medium orchid
    0xAFEEEE, # paleturquoise
    0x40E0D0, # turquoise
    0x00CED1, # darkturquoise
    0x5F9EA0, # cadet blue
    0x008B8B, # dark cyan
    0x008000, # darkgreen
    0x00FA9A, # mediumspringgreen
    0x2F4F4F, # darkslategrey
    0x3CB371, # medium sea green
    0x808080, # gray
    0x008080, # olive
    0x00FFFF, # aqua / cyan
    0xE0FFFF, # lightcyan
    0x808000, # olivedrab
    0x00FF00, # lime
    0x32CD32, # limegreen
    0x000080, # navy
    0x800080, # purple
    0x556B2F, # darkolivegreen
    0x00FF7F, # springgreen
    0x48D1CC, # seagreen
    0x4B0082, # indigo
    0x6A5ACD, # slateblue
    0x0000CD, # medium blue
    0x000080, # midnight blue
    0x00008B, # darkblue
    0xC71585, # mediumvioletred
    0x8B008B, # dark violet
    0x9400D3, # darkorchid
    0x9932CC, # darkmagenta
    0xFF00FF, # fuchsia
    0xFF1493, # deeppink
    0xFF6347, # tomato
    0xFFB6C1, # lightpink
    0xFFC0CB, # hotpink
    0xDB7093, # palevioletred
    0xFFBCD9, # papayawhip
    0xFFDAB9, # blanched almond
    0xFFDEAD, # navajowhite
    0xFFD395, # peachpuff
    0xFFFAF0, # floralwhite
    0xFFFF00, # yellow
    0xFFFFF0, # ivory
    0xFFFFE0, # lightyellow
    0xFFFFFF, # white
    0xF5F5F5, # whitesmoke
    0xFDF5E6, # oldlace
    0xA52A2A, # brown
    0x800000, # maroon
    0x808000, # olive
    0x808080, # grey
    0x000000, # black
    0x0000FF, # blue
    0x008000, # green
    0x008080, # teal
    0x00FFFF, # cyan
    0x000080, # navy
    0x008080, # darkteal
    0x008000, # darkgreen
    0x00FFFFFF, # brightcyan
    0x0000FF, # darkblue
    0x00FF00, # lime
    0x00008B, # darkcyan
    0x00005F, # darkred
    0x008080, # teal
    0x00191970, # midnightred
    0x00FFFF, # aqua
    0x008000, # forestgreen
    0x00FFFF, # cyan
    0x00FF00, # lime
    0x000080, # navy
    0x008080, # darkteal
    0x008000, # darkgreen
    0x00FFFFFF, # bright cyan
    0x00008B, # darkblue
    0x00005F, # darkred
    0x008080, # teal
    0x000080, # midnightblue
    0x008080, # darkteel
    0x00C0C0, # lightsteel
    0x00FF00, # lime
)

def make_palette():
    return list(it.chain((
        0,            # black
        0xFFFFFF,     # white
        *((i << 16 | i << 8 | i for i in range(32)),   # greyscale
          *(i << 8 | 0xff for i in range(16)),       # rgb
          *(0xff << 16 | i << 8 | 0x80 for i in range(16))),

        0xfffffe,    # lightgrey
        0xf0f0f0,    # colors from https://lospec.com/palette-list/space-age
        *((0xff << 8 | i << 16 | i for i in range(8)),        # palette
          *(i << 8 | j for i, j in zip(*zip(*(
              ((c >> shift & mask) << shift
                for _ in range(shift // 2))
             for shift, mask in enumerate(((1 << k) - 1
                 for k in range(24, -1, -8)))))))),


        *[0, u + 0x400000 for u in TRANSPARENT],

        (u << 8 | 0x80
           for t in TRANSPARENT
           for u in (*range(t - 0x400000, t + 0x400000, 0x100),
                     *set(range(
                         max(0x400000, t - 0x400000),
                         min(0x7fffff, t + 0x400000)))
                   )
        ),
    )))



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width=16, height=16
    ):
    palette = bytes([*reversed(make_palette())])


    def write_chunks(**kwargs):
        payload = b''.join(v for _, v in sorted(kwargs.items()))
        size = len(payload).to_bytes(4, 'little', signed=False)
        out.write