#!/usr/bin/env python3

import random
import struct
from array import array
from math import ceil
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generators `rng` (normally /dev/urandom) and
# `randrange`, which normally just returns ints from a list of primes.
def generate_random_gif(
        rng: BinaryIO, randrange: type, out: BinaryIO
    ):

    width, height = randrange(1, 4097), randrange(1, 4097)
    imagecolortable = array('B', map(int, rng.read(randrange(0, min(width, height))*3)))
    globalcolortable = [array('B', rng.read(1)) for _ in range(randrange(0, 257))]
    # TODO: handle interlaced mode
    palette = [(imagecolortable[i], imagecolortable[i+1], imagecolortable[i+2])
                    for i in range(len(imagecolortable))[::3]]

    data = []
    color = randrange(0, len(palette))
    for x in range(height):
        data += [[color] * width]
        for y in range(width)[::-1]:
            color ^= randrange(0, 3)
            if color == 3:
                color ^= 3

    encodedlength = ceil((width*height)/9) + 1
    out.write(struct.pack('<6sHHBBBxB',
              b'GIF89a',
              0,                   # Logical screen dimensions
              0, 0,               # Physical screen dimensions
              0xf7 | 0xc0 & 1<<3,   # Packed fields: use color table, sorted GIF
              len(globalcolortable),    # Size of global color table
              0,                   # Background color index
              0,                   # Pixel aspect ratio
              ))
    
    prefix = {0: bytes([0b11])]}.get(encodedlength-1, bytearray())

    def encode_subblock():
        length = 1 << (encodedlength-1)
        subblocks = []

        for x in range(ceil(width/length)):
            dat = data[(x % height)*width+(x//height)]
            subblocks += dat[:length]
        
        remainder = width%length
        finalsubblock = [*data[-remainder:], *dat[remainder:]] \
                        if remainder else None
            
        for subblock in subblocks[:-1]:
            out.write(prefix + tuple(subblock))
        if finalsubblock:
            out.write(tuple(finalsubblock))

    encode_subblock()
    return palette
    
    
    