#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The maximum color depth is defined by `maxbitdepth`.
# If `bgcolorindex` is not None, it will be used as the default
# graphics control block's background color index field. Otherwise,
# there won't be one.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, maxbitdepth: int=8,
        bgcolorindex: int=None
    ):

    # make sure we have an even bit depth
    assert((maxbitdepth & 1) == 0)

    # create header info
    header = b'GIF89a'

    # the current position is at the beginning of the Logical Screen 
    # Descriptor; since this is fixed-length, just calculate what it should
    # be now based on the specified color depth
    pos = 6 + 3*int(maxbitdepth / 8)
    image_width, image_height = \
        struct.unpack('<HH', 
        *[x for x in rng.read(pos)])
    
    # create palette, if needed
    if bgcolorindex!= None:

        # start with full range of colors starting at black (#000000)
        # (i.e., all zeroes in 24-bit RGB)
        palents = bytearray([0, 0, 0])

        for _ in range(1 << int(maxbitdepth / 8)):
        
            # add another entry to the palette
            # (for 24-bit RGB, it gets added as three 2-byte
            # pairs representing BGR triples, so we need every
            # other value here -- namely, those whose corresponding
            # second and third bytes don't have 2 high order bits set
            # -- because these represent negative numbers.)
            r, g, b = [y for y in rng.read(3)]
            while ((r >= 0xc0 or g >= 0xc0 or b >= 0xc0)):

                print(r,g,b);
                
                r, g, b = [y for y in rng.read(3)]
            
            palents += (r, g, b)
        
        # write out the length byte first
        palents += struct.pack('B', len(palents))

        # write to output stream
        out.write(header + struct.pack('<HHBB', image_width, 
            image_height, len(palents) | 0b10000000, 
            0) + bytes(palents))

    else:

        # no palette
        out.write(header + struct.pack('<HH', image_width, 
                image_height))

    # get remaining lzw compressed info from the RNG
    datalen = int.from_bytes(rng.read(2), 'big')
    data = rng.read(datalen)
    out.write(header + data)