#!/usr/bin/env python

import argparse
import io
import os
import struct
import sys
from typing import IO

DEFAULT_OUTPUT = 'random.gif'


class RandomGifGenerator():

    def __init__(self, args):

        self._args = args

    @property
    def width(self): return self._width

    @property
    def height(self): return self._height

    @property
    def color_table_size(self): return self._color_table_size

    @property
    def block_size(self): return self._block_size

    @property
    def output(self): return self._output

    def generate_random_gif(self):

        with open(f'{self.output}', 'wb') as fh:

            generate_random_gif(
                rng=io.BytesIO(), 
                out=fh, 
                width=self.width,
                height=self.height,
                color_table_size=self.color_table_size,
                block_size=self.block_size
            )


    def prepare_arguments(self):
        '''Prepares arguments.'''

        parser = argparse.ArgumentParser()
        parser.add_argument('--output', '-o', 
            help='Output filename.',
            default=DEFAULT_OUTPUT
        )
        parser.add_argument('width', type=int,
            help='Width of the generated GIF.'
        )
        parser.add_argument('height', type=int,
            help='Height of the generated GIF.'
        )
        parser.add_argument('color_table_size', type=int,
            nargs='?', default=7, choices=(7, 15),
            help='Color table size to use in the generated GIF.\n'
                 'Default: %(default)s\n'
                 '(choices: {7, 15})'
        )
        parser.add_argument('-b','--block-size', type=int,
            nargs='?', default=256,
            help='Size of the data blocks used for generating the GIF.\n'
                 'Default: %(default)s\n'
        )

        args = parser.parse_args()
        
        setattr(self, '_output', os.path.abspath(args.output))
        setattr(self, '_width', args.width & ~0x8000)    # ensure even value
        setattr(self, '_height', args.height & ~0x8000)  # ensure even value
        setattr(self, '_color_table_size', args.color_table_size)
        setattr(self, '_block_size', args.block_size)


# Generates a random GIF file into `out` using the
# random number generator `rng`. The parameters
# correspond to the information found here:
# https://www.w3.org/Graphics/GIF/spec-gif89a.txt
def generate_random_gif(
        rng: IO[bytes], out: IO[bytes], 
        width: int, height: int, 
        color_table_size: int, block_size: int
    ):

    # Signature
    out.write(b'GIF89a')
    
    # Width and Height
    out.write(struct.pack('<HH', width, height))

    ct_len = ((color_table_size << 1) | 1)     # amount of colors in colortable
    ct_len *= 3                                 # each color has RGB components
    assert len(ct_len) <= 255                   # max color table size is 256
    out.write((ct_lent).to_bytes(1, byteorder='big'))

    # Color Table
    for _ in range(color_table_size):
        r = rng.read(1)[0] % 256                 # red component
        g = rng.read(1)[0] % 256                 # green component
        b = rng.read(1)[0] % 256                 # blue component
        out.write(r.to_bytes(1, byteorder='big'))  # write color components to GIF
        out.write(g.to_bytes(1, byteorder='big'))
        out.write(b.to_bytes(1, byteorder='big'))

    # Block Size (Data Sub-Block within Data Stream)
    bs = min(block_size, 255)                    # maximum block size is 256
    out.write((bs).to_bytes(1, byteorder='big'))  # write block size

    # Data Sub-Blocks
    while True:                                  # keep creating blocks until done

        cs = rng.read(bs)                        # read block of data from RNG
        ncs = len(cs)                            # compute length of this block

        out.write(ncs.to_bytes(1, byteorder='big'))                  # write size of current block
        if ncs == 0 or ncs >= bs: break            # end when we reached the end of input or the block size
        out.write(cs)                            # write the rest of the block
    