#!/usr/bin/env python3

import argparse

import PIL.Image

GIF_HEADER_SIZE = 13

# Generates a random GIF file into `out`. It is generated by drawing from
# `/dev/urandom`, then shuffled, quantized and compressed using Pillow's
# `save()` function. The options given in kwargs are passed verbatim to that
# function.
# Refer to http://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#gif for more details about the parameters.
def generate_random_gif(**kwargs):
    with open('/dev/urandom', mode='rb') as rng:

        # Read first block
        header = rng.read(GIF_HEADER_SIZE)
        width, height = struct.unpack('<HH', header[-4:-2])
        
        imgdata = []        
        
        while True:

            # Read next sequence block from /dev/urandom
            seqhdr = rng.read(1)
            
            # Next bit indicates whether this block should be included in the animation
            if not bool(seqhdr[0] & 0b10000000):
                continue
            
            # Read current block size
            currsz = rng.read(1)[0]       
            
            # If not last frame, first two bits indicate interlacing flag and color table usage. We ignore these values unconditionally because we will regenerate them later.
            # Also note that we can't use the decoder here since it doesn't decode entire images; there's no way to get all the frames beforehand.
            # ctabsz = ((currsz >> 3) & 0b0011) + 1
            clrused = (currsz & 0b0010)!= 0
            lctblsz = (currsz & 0b0001) + 1
            pass
            
            # Read command bytes
            cmdbyt = rng.read(1)[0]
            lc = cmdbyt & 0b00001111
            gifcolbits = 2 ** ((cmdbyt >> 4) + 1)
            pass

            # Read LZW minimum code size
            codesz = rng.read(1)[0]
            
            # Read local color table
            ctable = []
            for _ in range(lctblsz):
                rgb = tuple(rng.read(3))
                ctable += list(rgb) + [0]
            assert len(ctable) == pow(gifcolbits+1, 3)

            # Create temporary image to encode data into. This must be larger than any actual data set we might have, otherwise the encoder could complain about overflowing its own buffer.
            tmpimg = PIL.Image.new('P', (width*2,height*2), None)
            tmpimg.putpalette(ctable)

            # Decode data block
            x, y = 0, 0     # Position in decoded image
            lzwbuf = []      # Buffer for currently decoded bytes
            pos = 0          # Current position within bytebuffer
            buf = []         # Buffer for encoded pixels
            lzwnum = []      # Number part of LZW codes
            lzwstr = []      # String part of LZW codes
                            # Note: both these arrays store integer indices instead of strings!
            pass

            # While there are still bytes left to decode...
            while True:

                # Check if we've reached end of block
                if pos >= currsz:
                    break
                
                # Get our code
                code = ord(seqhdr[0]) << 8 | ord(codesz) | ord(rng.read(1)[0])
                
                # If high order bit is set, new code string received. Otherwise, add to existing
                if code & 0x8000:

                    # Add empty string to buffer so we can force decoding of a single character
                    buf += [len(lzwstr)]
                    
                    # Copy old string and convert to integers
                    lzwstr = map(lambda x: ord(x), ''.join(map(chr, lzwbuf)))
                    
                    # Reset buffer and counters
                    del lzwbuf[:]
                    buf.append(code & 0x0FFF)
                    pos += 1
                    
                elif code < len(lzwstr):   # Existing code received
                
                    buf += [code + len(lzwstr)]
                    lzwbuf += [lzwstr[code]]
                    del lzwstr[code]
                    pos += 1
                    
                else:       # New code received
                
                    
                        
                    # Handle special case where we don't receive enough bits to form a complete code
                    if pos == 1 and type(buf[-1]) == str:
                        lzwbuf += [ord(buf[-1])]
                        buf[-1] += chr(code)
                        print(lzwstr, buf)
                        raise Exception("This shouldn't happen")
                    
                    lzwnum.append(pos - 1)
                    lzwstr.append(code & 0xFFF)
                    lzwbuf += [code & 0xFFF]
            
                    # Try to assemble a full code
                    try:
                        
                        packed = sum([lzwnum[i] << (i * 8) for i in reversed(range(len(lzwnum)))])
                        print("New packed value", packed, "(", [hex(c) for c in lzwnum], ")")
                        
                        nxtch = chr(packed)
                        print("Next char:", repr(nxtch))

                        # Add decoded characters to buffer
                        buf += list(map(ord, nxtch))

                        # Delete used codes
                        del lzwnum[:len(lzwnum)-2]
                        del lzwstr[:len(lzwstr)-1]
                        pos -= 2
                        
                        # Reset number array
                        lzwnum.append(-1)
                        
                    except ValueError as e:
                        
                        # print("Num:", lzwnum)
                        # print("Str:", lzwstr)
                        # print("Buf:", buf)
                        raise e
                        
            print("Buffer:", buf)
            
            # Convert pixel order back to right side up
            pixarr = [[int(buf[4*(y+(j%hgt)*hgt)+(i%wid)+(j//hgt)*(wid)]) for i in range(wid)] for j in range(hgt)]
            pixarrrev = [[] for _ in range(width)][::-1]
            print("\nPixels:\n{}".format('\n'.join([' '.join(map(str,[pixarr[j][i] for i in range(width)])) for j in range(height)])))
            print()
                
            # Move buffer values to temporary image
            for p in zip(*pixarrrev):
                for pv in p:
                    tmpimg.putpixel((x,y), pv)
                    x += 1
                x = 0
                y += 1
                        
            imgdata.extend(tmpimg.getdata())
            
        sizex = kwargs['sizex'] or width
        sizey = kwargs['sizey'] or height
        im = PIL.Image.new('RGB', (sizex, sizey), '#FFFFFF')
        print("Data size:", width*height, "=>", sizex*sizey)
        im.putdata(imgdata)
        im.save(output, **kwargs)