#!/usr/bin/env python3

import argparse
import struct
import sys
import timeit

from io import BytesIO

from PIL import Image


DEFAULT_WIDTH      =  256
DEFAULT_HEIGHT     =  256
DEFAULT_NOISE_LEVEL = 17
DEFAULT_COLORMAP   = False
DEFAULT_OUTPUT     = 'random.png'
BACKGROUND         = '#ffffff'

STARTING_COLORS = [
    { 'color': '#4d4d4d',
      'percentage':.2 },
    { 'color': '#595959',
      'percentage':.3 },
    { 'color': '#6b6b6b',
      'percentage':.3 }
]

HEADER = b'\x47\x49\x46\x38\x39\x61'


class RandomImageGenerator():

    def __init__(self, width, height, noise_level, colormap, output=None):

        if not output:
            self.output = DEFAULT_OUTPUT
        else:
            self.output = output

        self.width       = width
        self.height      = height
        self.noise_level = noise_level
        self.colormap    = colormap


    @staticmethod
    def parse_color(color):
    
        try: 
            red, green, blue = map(int, color[1:].split(','))
            return (red, green, blue) 
        except ValueError:
            raise RuntimeError('Invalid color!')
        
        
    def generate_image(self):

        buf = BytesIO()
        im = Image.new("RGB", (self.width, self.height), BACKGROUND)
       
        # Read colors from STARTING_COLORS constant
        pixels    = list(im.getdata())
        max_index = len(pixels)-1
        
        start_colors = []
        for start_color in STARTING_COLORS:
            
            start_colors.append({
                'color':    self.parse_color(start_color['color']),
               'max': int((start_color['percentage']/(1-.sum(map(lambda sc: sc['percentage'], STARTING_COLORS))))*max_index)
                })
            
        # Add noise
        for i in range(len(pixels)):

            pixel = pixels[i]
            
            if i <= sum(map(lambda x: x['max'], start_colors)):
            
                continue

            r = pixel[0]+(pixel[0]/float(self.noise_level))%(256-pixel[0])
            g = pixel[1]+(pixel[1]/float(self.noise_level))%(256-pixel[1])
            b = pixel[2]+(pixel[2]/float(self.noise_level))%(256-pixel[2])
                
            pixels[i] = (r,g,b)


        # Set colormap
        if self.colormap:
            
            colormap = {}
        
            for i in range(len(pixels)):

                pixel = pixels[i]
                
                if str(pixel) not in colormap:

                    colormap[str(pixel)] = self._generate_random_color()


                pixels[i] = tuple(colormap[str(pixel)])

        
        im.putdata(pixels)
        
        # Save to PNG or Gif
        if self.output[-4:] == '.gif':
            
            buf.seek(0)
            im.save(buf, format="GIF")
            buf.seek(0)
            self._write_gif(buf)

        elif self.output[-4:] == '.png':

            im.save(self.output, "PNG")

        else:

            print('Output must be either.gif or.png.')
            exit(-1)


    def _generate_random_color(self):

        red, green, blue = [],[],[]

        while True:

            value = ord(os.urandom(1))%256;

            if (value > 200):

                red.append(value);

            if (value < 100):

                green.append(value);

            if (value > 10):

                blue.append(value);


            if ((len(red)==3) and (len(green)==3) and (len(blue)==3)):

                break;

        return tuple(red+green+blue)


    def _write_gif(self, buffer):

        with open(self.output, 'wb') as f:
    
            f.write(HEADER)
            f.write(buffer.read())



if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('--width',           type=int, help='Width of generated images', default=DEFAULT_WIDTH)
    parser.add_argument('--height',          type=int, help='Height of generated images', default=DEFAULT_HEIGHT)
    parser.add_argument('--noise',           type=int, help='Noise level on image generation', default=DEFAULT_NOISE_LEVEL)
    parser.add_argument('--colormap',        action='store_true', help='Use a color map instead of auto-generated colors', default=DEFAULT_COLORMAP)
    args = vars(parser.parse_args());

    img_gen = RandomImageGenerator( **args )
    img_gen.generate_image();