#!/usr/bin/env python3
"""Generate a random giff."""

import io
import struct
import zlib

from binascii import hexlify as xhex
from itertools import count
from typing import Iterable, Sequence, Tuple, Union


# Gif Header Constants
VERSION     = b'\x01'       # Version (1)
LOOP        = b'\x00'       # Loop Forever
DESCRIPTOR  = {               # Image Descriptor Block
    0b00 : 0b1010,          # Plain Text Extension
    0b01 : 0b1010,          # Graphic Control Extension
    0b10 : 0b1010,          # Comment Extension
    0b11 : 0b1010            # Application Extension
}
IMAGE_DESC  = b'\x2c'       # Start Image Descriptor (Image Seperator)
GRAPHICCTRL = b'\xf9\x04'   # Graphic Control Extension
PLAIN_TEXT  = b'\x01\xfe'   # Plain Text Extension
COMMENT     = b'\xfe\x01'   # Plain Text Extension
APPLICATION = b'\xff\x0b'   # Application Extension
GRAPHICDATA = b'\x21\xf9\x04'
BLOCK_SEP   = b'\x3b'
END         = b'\x3b'       # End File


class ColorTable:
    """A color table object that supports indexing and is iterable."""
    def __init__(self):
        self._colors = []
    
    def __iter__(self):
        return iter(self._colors)
    
    def add(self, r, g, b):
        """Add an RGB tuple to the table and return its index."""

        idx = len(self._colors)
        
        while True:

            if not all((r,g,b)):
                raise ValueError('RGB values must be non-zero.')
            
            elif any(((r & 0xF8) == 0), ((g & 0xFC) == 0),
                     ((b & 0xF8) == 0)):

                raise ValueError("Invalid RGB value '{}'".format((r,g,b)))
            
            else:
                
                r >>= 3
                g >>= 2
                b >>= 3

                self._colors += [
                    r << 5 | g << 2 | b, 
                    r << 6 | g << 3 | b >> 1,
                    r << 7 | g << 5 | b >> 1,
                    r << 7 | g << 5 | b >> 1
                ]
                
                break
                
        return idx


class Interlacing:
    """Interlacing enum class"""
    NONE    = 0
    LINEAR  = 1
    FUZZY   = 2
    DONTCARE= -1


class LogicalScreenDescriptor:
    """The logical screen descriptor block in a GIF file."""
    DESCRIPTOR_STRUCT = '>HHBBBH'
    DESCRIPTOR_SIZE   = struct.calcsize(DESCRIPTOR_STRUCT)

    def __init__(self, width, height, depth=8):
        """Initialize the object using the given parameters.

        Parameters
        ----------
        width : int
            The width of the image in pixels. Must be less than or equal to
            65535, but greater than zero.
        
        height : int
            The height of the image in pixels. Must be less than or equal to
            65535, but greater than zero.

        depth : int
            The bit depth of each pixel. Must be one of 8, 16, or 24. Defaults
            to 8 bits per pixel.
        """
        self.width  = width
        self.height = height
        self.depth  = depth

        self.packed = struct.pack(
            '>' + DESCRIPTOR_STRUCT[1:], 
            *self.__get_parameters())


    @property
    def packed(self):
        """Get the raw binary string containing the structure's data."""
        return self._packed
    

    @packed.setter
    def packed(self, buff):
        if len(buff)!= self.DESCRIPTOR_SIZE:
            raise ValueError('Got {} bytes'.format(len(buff)))
        self._packed = buff
    

    def __get_parameters(self):

        # TODO: Validate parameters

        # Determine type code based on depth
        tcode = 0
        if self.depth <= 8:  tcode |= 0b0001 
        if self.depth > 8:   tcode |= 0b0010 

        return (self.width, self.height, tcode|0b00000000, 
                self.depth, 0, 0)


    def __str__(self):
        """Return a human readable description of the structure."""
        return '\n'.join([
            "Logical Screen Descriptor",
            "\tWidth    : {}".format(self.width),
            "\tHeight   : {}".format(self.height),
            "\tDepth    : {}\n".format(self.depth),
            "\tPacked Bytes:",
            "\t{}".format(', '.join(['{:02X}'.format(i) for i in self.packed]))
        ])


class GraphicsControlExtension:
    """A graphics control extension."""
    EXTENSION_STRUCT = '>BbbbbBBHB'
    EXTENSION_SZIE   = struct.calcsize(EXTENSION_STRUCT)

    DISPOSE_METHODS = {
        0b0000 : 'NONE',
        0b0001 : 'DO_NOT_DISPOSE',
        0b0010 : 'RESTORE_BACKGROUND',
        0b0011 : 'RESTORE_PREVIOUS'
    }

    USERINPUT      = {
        0b0000 : False,
        0b0001 : True 
    }

    REPEAT         = {
        0b000 : None,
        0b001 : 0, 
        0b010 : 1, 
        0b011 : 2, 
        0b100 : 3, 
        0b101 : 4, 
        0b110 : 5, 
        0b111 : 6
    }

    def __init__(
            self, delay:int=100, transparent:bool=False, disposal_method=-1,
            userinput=None, repeats=None):
        """Initialize the object with the given parameters. Any unspecified
        parameters will default to their respective defaults (-1).

        Parameters
        ----------
        delay : int
            How many hundredths of a second to display this frame before moving
            to the next. This must be between 0 and 65535 inclusive. Defaults
            to 100 hundreths of a second.

        transparent : bool
            Whether or not to make this frame transparent. If true, a new entry
            will be added to the global color table for every frame containing
            the specified transparent color, rather then displaying it as black.
            Defaults to false.

        disposal_method : int
            Specifies what to do when finishing displaying the graphic. See the
            GIF spec for details about the different options. Defaults to leave
            the decoder up to them.

        userinput : bool
            Specifies whether the receiver should wait for user input after
            displaying this frame. See the GIF spec for details. Defaults to
            false.

        repeats : int
            Specified how often this frame should repeat itself. A null value
            means infinite repeats. Can take on any of the following values:

            0 -> 0 repetitions
            1 -> forever
            2 -> twice
            3 -> thrice
            4 -> four times
            5 -> five times
            6 -> six times


        To summarize, valid disposal methods are:

        0 : DO NOT HIDE
        1 : RESTORE BACKGROUND COLOR TO PREVIOUS FRAME
        2 : RESTORE BACKGROUND COLOR TO TRANSPARENT COLOR
        3 : RESERVED VALUE


        And valid repeat counts are from 0 to 6 inclusive.
        """

        self.delay           = delay
        self.transparent     = transparent
        self.disposal_method = disposal_method
        self.userinput       = userinput
        self.repeats         = repeats

        self.packed = struct.pack(
            '>' + self.EXTENSION_STRUCT[-1], 
            *self.__get_parameters()
        )

    
    @property
    def packed(self):
        """Get the raw binary string containing the structure's data."""
        return self._packed
    

    @packed.setter
    def packed(self, buff):
        if len(buff)!= self.EXTENSION_SZIE:
            raise ValueError('Got {} bytes'.format(len(buff)))
        self._packed = buff
    

    def __get_parameters(self):

        # Validate parameters
        if not isinstance(self.delay, int):
            raise TypeError('"delay" parameter must be an integer!')
        elif self.delay < 0 or self.delay > 65535:
            raise ValueError('"delay" parameter must be between '
                             '0 and 65535 inclusive')
            
        if self.transparent not in self.USERINPUT:
            raise ValueError('"transparent" parameter may only take boolean '
                             'values!')

        if self.disposal_method not in [-1] + list(self.DISPOSE_METHODS.keys()):
            raise ValueError('"{}" method not supported! Valid disposal methods '
                             ': {}'.format(
                                 self.disposal_method, 
                                 ', '.join(
                                     map(