#!/usr/bin/env python3

import base64
from os import urandom
from typing import BinaryIO

# Generates a random GIF file in `out`. The image will contain no more than
# 256 colors and only include one frame with an unlimited duration. The colors
# are chosen randomly based on 3byte RGB values. The image's color table is set
# to its maximum length so that all colors are used before the next image or
# animation begins.
def generate_random_gif(
        out: BinaryIO
    ):

    # Generate custom palette for each pixel (up to max length)
    def getColorPalette():
        rgb = urandom(3)
        return [rgb[i] | 0x80 for i in range(3)]
    
    # Determine how many frames we want to have (including the current one)
    numFrames = {b'\xff\xff': 0b000,
                 b'\xfe\xff': 0b001,
                 b'\xfc\xff': 0b010,
                 b'\xf8\xff': 0b100}[urandom(2)] << 3

    # Now begin writing the output buffer
    buf = []
    buf += 'GIF89a'.encode('ascii')
    buf += urandom(7)
    buf += struct.pack('<H', 0xffff - len(buf))
    buf += b'NETSCAPE2.0\x03\x01'
    buf += numFrames.to_bytes(1, byteorder='big', signed=True)
    buf += b'\00'
    buf += b'!\xf9\x04\x01\x00\x00'
    buf += b';'

    # Set our color table as full as possible since we don't yet know 
    # which colors we're going to use
    buf += bytes([(len(getColorPalette()))])
    buf += bytes([paletteEntry for entry in map(getColorPalette, range(256))]
                    if len(getColorPalette()) else [])
    
    while True:

        # New image block has this structure: 
        #   - Image separator \x2c
        #   - Left-position (always zero)
        #   - Top-position (always zero)
        #   - Width of image
        #   - Height of image
        #   - Flags: 
        #       - 0x80 sets local color table flag, 0x40
        #         indicates global color table presence
        # 
        # If there is any data remaining after reading these initial fields, then it is
        # assumed to be a local color table to override the previous version
        buf += b','
        buf += b'\00'*2
        
        # Now determine the width and height of the image
        bufLen = 1 + int.from_bytes(urandom(1), byteorder='big', signed=False) % 0x1f
        buf += bufLen.to_bytes(1, byteorder='big', signed=False)
        bufLen -= 1
        bufLen |= 0x80 # indicate there is a local color table
        buf += bufLen.to_bytes(1, byteorder='big', signed=False)
        sz = 1 << (int.from_bytes(
            buf[-2:], byteorder='big', signed=False
        ) + 1)

        # Next comes the time delay until the image should be displayed 
        # (in hundredths of a second)
        buf += urandom(2)

        if not bufLen: 
            break

        # Finally append our pixels
        buf += bytes([paletteEntry for entry in map(getColorPalette, 
                        range((sz ** 2) // bufLen))]
                        if len(getColorPalette()) else [])

    # Add closing tags
    buf += bytes([0]).join([b'\x3b', b'\x00'])

    # Write our complete output
    out.write(base64.encodebytes(bytes(buf)))