#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Encodes an unsigned short to its bytes
def encode_unsigned_short(val):
    return val.to_bytes(2, 'big', signed=False)

# Generates a random GIF file into `out` using the
# given pseudorandom number generator in binary format
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Write the magic number
    out.write('GIF87a'.encode())

    # Write screen descriptor block
    width = (int.from_bytes(rng.read(1),
                     byteorder='big') % 65536) | \
            0x2000    # Make sure it's at least 8 pixels wide
    height = (int.from_bytes(rng.read(1),
                      byteorder='big') % 65536) | \
             0x2000   # Make sure it's at least 8 pixels high
    # Block size is always fixed for this implementation
    out.write((b'\x2c\x00,\xff\x0b\x4e\x45\x54\x53\x43\x41\x50'))
    out.write(width.to_bytes(2, 'big'))
    out.write(height.to_bytes(2, 'big'))

    # Tell how many bytes are used by global color table and set default value
    gctSize = ((int.from_bytes(rng.read(1),
                       byteorder='big') % 3) + 1) * 32
    out.write(gctSize.to_bytes(1, 'big'))
    bgIndex = (int.from_bytes(rng.read(1),
                    byteorder='big') % gctSize)    # Background color index
    pixelAspect = (int.from_bytes(rng.read(1),
                         byteorder='big') % 4)     # Pixel aspect ratio
    # Write it all together
    out.write((bgIndex).to_bytes(1, 'big'))
    out.write((pixelAspect).to_bytes(1, 'big'))
    out.write(b'\x00\x00')      # No local palette entries

    # Write the image descriptor block
    leftPos = int.from_bytes(rng.read(2),
                        byteorder='big') % width
    topPos = int.from_bytes(rng.read(2),
                     byteorder='big') % height
    imgWidth = (int.from_bytes(rng.read(2),
                            byteorder='big') & 0x3FFF) | 1
    imgHeight = (int.from_bytes(rng.read(2),
                             byteorder='big') & 0x3FFF) | 1
    flags = int.from_bytes(rng.read(1),
                  byteorder='big') & 1 | 2 | 4 |  0xC0         # Local color table exists
    if flags & 0x80 == 0x80:    # Interlace flag
        interlaceFlag = b'\x40'
    else:
        interlaceFlag = b''
    lctSize = ((flags & 0x7) + 1) * 32
    lctSizeBytes = lctSize.to_bytes(1, 'big');
    out.write(lctSizeBytes)
    lctBg = (int.from_bytes(rng.read(1),
                   byteorder='big') % lctSize)       # Local background colr idx
    out.write(lctBg.to_bytes(1, 'big'))
    # Write it all together
    out.write(leftPos.to_bytes(2, 'big'))
    out.write(topPos.to_bytes(2, 'big'))
    out.write(imgWidth.to_bytes(2, 'big'))
    out.write(imgHeight.to_bytes(2, 'big'))
    out.write(flags.to_bytes(1, 'big'))
    out.write(((flags >> 6) & 3).to_bytes(1, 'big'))
    out.write(interlaceFlag)
    out.write(b'\x00')          # No sort flag

    # Write the graphic control extension
    out.write(b'\x21\xf9\x04')             # Extension introducer and label
    frameDelay = int.from_bytes(rng.read(2),
                          byteorder='big') & 0xFFFF
    transIndex = (int.from_bytes(rng.read(1),
                       byteorder='big') % gctSize)
    flags = int.from_bytes(rng.read(1),
                byteorder='big') & 1 | 2              # User input or disposal method
    transparentFlag = ''                # Assume no transparency otherwise
    if flags >= 2 and transIndex < gctSize:
            transparentFlag = b'\x01'        # Set flag only when needed
    out.write((((frameDelay << 8) | 0x01) |
               ((transIndex & 0xF) << 4)).to_bytes(2, 'big'))
    out.write(transparentFlag)
    if transparentFlag!= '':
        out.write(transIndex.to_bytes(1, 'big'))
    out.write(b'\x00\x00\x00')                      # End of application extension

    # Write plain text extension to go with our animation
    out.write(b'\x21\x03\x0aHello World!')            # Extension introducer and label
    out.write(b'\x00')                              # End of plain text extension

    # Write the image data subblock
    lzwMinCodeSize = (int.from_bytes(rng.read(1),
                              byteorder='big') % 9) + 1
    out.write(encode_unsigned_short(lzwMinCodeSize))
    compressor = LZWCompressor()
    compressor.compress(rng, out, lzwMinCodeSize);
