#!/usr/bin/env python3

import math
import os
import sys
import struct
import time

from argparse import ArgumentParser


# A simple LZW dictionary implementation which we will use
# as our LZSS compressor. This version is based on this article by <NAME>:
# https://www.johndcook.com/blog/2017/12/03/lzw-in-python/
class Dictionary:

    def __init__(self):

        self.dicts = dict()
        self.maxes = []
        self.nextcode = 258


    def clear(self):

        self.dicts = {
            255 : None,
            256 : '',
        }

        self.maxes = ['']
        self.nextcode = 258


    def haskey(self, key):

        return key in self.dicts


    def addword(self, word, code=None):

        if not code or code == 257:
            
            self.dicts[self.nextcode] = word
            self.maxes.append('')
            code = self.nextcode
            self.nextcode += 1
        
        elif self.haskey(code):
        
            self.dicts[self.nextcode] = word[:0:-1] + self.dicts[code][0]

            if len(self.maxes[-1]) <= len(self.dicts[self.nextcode]):

                self.maxes.append(self.dicts[self.nextcode])

            code = self.nextcode
            self.nextcode += 1


        while True:

            if len(self.maxes) == 1:

                break

            if len(self.dicts[code]) >= len(self.maxes[-1]):
            
                break

            del self.maxes[-1]


        return code

    
    def lzwdecompress(self, data):

        result = ''

        word = self.dicts[255]

        for c in data:

            cc = ord(c)
            wc = ord(word)

            if cc == wc:
                
                result += chr(wc)
                word = word[:-1]

            else:

                result += word + chr(cc)
                word = word + chr(cc)[0]

        return result



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    
    starttime = time.perf_counter()

    dictxml = {}

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    
    rng.seek(7)

    # get the LZW min and max codes
    lzmin = int.from_bytes(rng.read(1), 'little')
    lzmax = int.from_bytes(rng.read(1), 'little')

    #print(f'{lzmin}, {lzmin}')

    # how many colors should there be? (0 means none)
    colortablebits = int.from_bytes(rng.read(1), 'little')

    # now create an XML document representing the image
    imgxml = f'''<image width="{image_width}" height="{image_height}">
  <colorTableBits>{colortablebits}</colorTableBits>
  <lzwMinCode>{lzmin}</lzwMinCode>
  <lzwMaxCode>{lzmax}</lzwMaxCode>
  <frames>\n'''

    # how big do we want the frame buffer to be?
    buffertotalsize = ((image_width * image_height) / 8) + 1

    # read the initial LZW codes
    dic = Dictionary()
    dic.clear()
    
    ncodes = int.from_bytes(rng.read(1), 'little')
    lzcodes = []
    for _ in range(ncodes):
    
        lzcodes.append(int.from_bytes(rng.read(1), 'little'))
        
    for code in lzcodes:
    
        dic.addword('', code)

    # read the rest of the LZW codes until we reach the end of the frame
    dicidx = 0
    while True:

        # grab next byte and check against the minimum range limit
        c = int.from_bytes(rng.read(1), 'little')

        if c >= lzmin:

            # adjust dictionary index so it won't point outside the list
            if dicidx < 0:

                dicidx = 0

            # add new entry to the dictionary
            dicidx = dic.addword(dic.dicts[dicidx], c)

        else:

            # got EOF; first make sure we're back at zero
            assert dicidx == 0

            # put all the parsed codes together as one string
            s = ''.join([dic.dicts[x] for x in lzcodes])

            # add the last character in the dictionary to close things off
            # (so that DEFLATE can actually work correctly with the trailing
            # bytes after the LZFH block)
            s += dic.dicts[dicidx][:0:-1]

            # compress it up
            z = s.encode('zlib_codec')

            # convert to hexstring
            hxstr = ''.join('%02X' % x for x in z)

            # trim off any padding characters
            hxstr = hxstr.rstrip('0').rstrip('.')

            # save it to a dictionary
            dictxml['frame{}'.format(imgxml.count('\n'))] = hxstr

            # done!
            break


    # build XML document
    imgxml += '</frames>'

    # done reading all frames, let's time it
    elapsed1 = round((time.perf_counter() - starttime) * 1000.0)

    # go through these frames backwards so that we can remove them from 
    # the original list if we need to
    for k in reversed(sorted(dictxml.keys())):

        # figure out the total number of chunks 
        chunkstotal = len(dictxml[k].split('.'))

        # how many bytes per chunk?
        spare = image_width * image_height * 1.0 / (chunkstotal * 8)

        # calculate the upper bound for the percentage of unused space
        percentupperbound = 100 - math.floor(((spare **.5) / 2) * 100)

        # figure out how much compression is left over
        leftover = 100 - (percentupperbound % 100)

        # update the dictionary element
        dictxml[k] = '.'.join(list(reversed(
            dictxml[k].split('.')[math.ceil(leftover / 100 * chunkstotal):])))


    elapsed2 = round((time.perf_counter() - starttime) * 1000.0)

    print("Elapsed:", elapsed1, elapsed2 - elapsed1)

    # ok, ready to write everything out...
    # note: threads don't really seem to help here, for some reason...
    out.write('''\
<?xml version="1.0"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml"><head><title>Animated GIF</title></head><body>
''')

    out.write(imgxml)

    out.write('''\
  </image>
</body></html>
''')