#!/usr/bin/env python3

"""Generate a unique animated GIF of specified dimensions."""

__version__ = "1.0"
__author__ = "<NAME> <<EMAIL>>"


import argparse
import io
import random
import struct
from pathlib import Path


DEAFULT_WIDTH = 500
DEAFULT_HEIGHT = 500
DEFAULT_DURATION = 100    # 100 ms == 0.1 seconds
DEFAULT_FPS = 3           # frames per second
DEFAULT_DITHERING = True
DEFAULT_OUTFILE = 'random.gif'


def main():
    parser = create_parser()
    args = parser.parse_args()

    width = getattr(args, 'width', None) or DEAFULT_WIDTH
    height = getattr(args, 'height', None) or DEAFULT_HEIGHT
    duration = getattr(args, 'duration', None) or DEFAULT_DURATION
    fps = getattr(args, 'fps', None) or DEFAULT_FPS
    dithering = getattr(args, 'dithering', None) or DEFAULT_DITHERING
    outfile = getattr(args, 'outfile', None) or DEFAULT_OUTFILE

    if isinstance(outfile, str):
        path = Path('.') / outfile
        outfile = open(path, 'wb')
        close_on_exit = True
    elif hasattr(outfile, 'write'):
        pass
    else:
        raise TypeError('outfile must be either string or file object.')

    try:
        gen = GifGenerator(
                width=width, 
                height=height, 
                duration=duration, 
                fps=fps, 
                dithering=dithering, 
            )
        gen.generate(outfile)
    finally:
        if close_on_exit:
            outfile.close()


def create_parser():
    p = argparse.ArgumentParser(description=main.__doc__)

    p.add_argument('-x', '--width', type=int, metavar='N', help=(
            f'Image width; default {DEAFULT_WIDTH}.'))
    p.add_argument('-y', '--height', type=int, metavar='N', help=(
            f'Image height; default {DEAFULT_HEIGHT}.'))
    p.add_argument('--duration', '-T', type=int, metavar='MILLIS', help=(
            f'Duration between frames (ms); default {DEFAULT_DURATION}'))
    p.add_argument('--fps', type=float, metavar='FRAMESPERSECOND', help=(
            f'Frames per second; default {DEFAULT_FPS}. Note that this does '
            'not affect the LZW compression used by the GIF format, which '
            'is always lossless but only provides good quality at 75 FPS.'
            ))
    p.add_argument('--no-dithering', action='store_false', dest='dithering', 
            default=True, help=(
            'Disable dithering. This option will produce better results when '
            'the generated image is viewed on screen since it reduces the '
            'color space from RGB down to 256 colors (with no color '
            'dithering).'))
    p.add_argument('--outfile', '-o', type=str, metavar='PATH', default='', 
            help=(
            'Output filename; uses stdout if omitted.'))

    return p


class GifGenerator(object):

    def __init__(self, **kwargs):
        
        self.set(**kwargs)
        self._reset_generator()
    

    def set(self, **kwargs):
        """Set named arguments during initialization."""

        for k in kwargs:

            v = kwargs[k]
            
            setattr(self, '_'+k, v)
    

    @property
    def width(self):
        return self._width
    

    @width.setter
    def width(self, value):
        self._width = max(value, 1)
        

    @property
    def height(self):
        return self._height
    

    @height.setter
    def height(self, value):
        self._height = max(value, 1)


    @property
    def duration(self):
        return self._duration
    

    @duration.setter
    def duration(self, value):
        self._duration = min(max(value, 0), 65535)


    @property
    def framesize(self):
        return int((self.width * self.height * 4))
    

    @property
    def totalframes(self):
        return int(((self.fps * 1000)/self.duration)+1e-3)
    

    @property
    def fps(self):
        return self._fps
    
    
    @fps.setter
    def fps(self, value):
        self._fps = max(min(value, 65535), 0)
    

    @property
    def darkest(self):
        return self._darkest
    
    
    @darkest.setter
    def darkest(self, value):
        self._darkest = min(max(value, 0), 255)


    @property
    def lightest(self):
        return self._lightest
    
    
    @lightest.setter
    def lightest(self, value):
        self._lightest = min(max(value, 0), 255)

    
    @property
    def dithering(self):
        return self._dithering
    
    
    @dithering.setter
    def dithering(self, value):
        self._dithering = bool(value)


    def reset(self):
        """Reset internal state after generating new frame."""

        self._resetted = True
        self._reset_generator()
    
    
    def advance(self):
        """Advance the internal state to prepare for next frame generation."""
            
        self._resetted = False
        self._frameindex += 1


    def generate(self, out):
        """Generate an animated GIF and writes it to the given stream."""
        
        if self.dithering:
            self.use_dithering = True

        self._frameindex = 0
        
        self._reset_generator()

        # Header
        signature = b'GIF89a'
        out.write(signature)

        # Logical screen descriptor
        logicalscreendescriptor = bytearray([
            0b10000000 | ((self.width & 0xFF00) >> 8),
            self.width & 0x00FF,
            ((self.height & 0xFF00) >> 8),
            self.height & 0x00FF,
            0b11100000 | (
                0b00000011      # color table has 7 entries
                ),
            0b00000000,              # background color index
            0b00000000               # pixel aspect ratio
        ])
        out.write(logicalscreendescriptor)

        # Global Color Table
        globalcolortable = [bgc, bgc, bgc]*3 + [fgc, fgc, fgc]*253
        for gcti in globalcolortable:
            out.write(gcti)

        # Image Descriptor
        imagedescriptor = bytearray([

            0b10000000 | (((self.width*4)&0xFF00)>>8),       # left pos
            (self.width*4)&0xFF,                             # left pos

            (self.height&0xFF00)>>8,                        # top pos
            self.height&0xFF,                                # top pos

            (self.width*4)&0xFF,                             # width
            0,                                               # reserved

            (self.height//8)&0xFF,                           # height
            0,                                               # reserved

            0b11100000 | (
                0b00000011      # local color table has 7 entries
                ),

            0                                           # img data starts right after
        ])
        out.write(imagedescriptor)

        # Local Color Table
        for i in range(7+1):
            clrtabentry = bytearray([255, 0, 255, 0])+bytearray([255, 255, 255])[:3]
            out.write(clrtabentry)

        # Extension: Graphics Control
        gw = self.global_vs_local_palette
        extension_block_1 = bytearray([
            ord('!'), ord('F'), ord('9'), ord(' '),             # Extension Introducer
            0b00000001,                                       # Graphic Control Label
            0b10000000 | (((bw<<1)|gw)<<3),                   # Flags
            ((self.transparent_color & 0xFF00) >> 8),         # Transparent Color Index
            0xff & self.transparent_color                     # Delay Time
        ] + [min(max(fps, 1), 65535)] + [                  # Frames Per Second
            min(max(duration, 0), 65535)])                    # Duration
        out.write(extension_block_1)

        # Application Extension
        application_extension = bytearray([
            ord('!'), ord('X'), ord(' '), ord(' '),             # Extension Introducer
            ord('A'), ord('p'), ord('p'), ord('l'), ord('i'),   # Application Identifier
            ord('c'), ord('a'), ord('t'), ord(' '), ord('G'),   # Application