#!/usr/bin/env python3

import sys
import argparse
# Add current folder to system path, so that we can find the other modules.
sys.path.insert(0, './')
from lib import shared_functions as sf
from lib import preprocess as prep
from lib import modeling as mdl
import os

def main():
    argparser = argparse.ArgumentParser()
    # Required arguments:
    argparser.add_argument("--data-file", type=str, required=True)
    argparser.add_argument("--output-dir", type=str, required=True)
    # Optional arguments:
    argparser.add_argument("--test-size", default=0.25, help="Proportion of data to withheld for testing.")
    argparser.add_argument("--seed", default=17, help="Random seed.", type=int)
    args = argparser.parse_args()
    
    # Create output directory if it does not exist yet.
    os.makedirs(args.output_dir, exist_ok=True)
        
    # Load dataset and split into train and test sets.
    X, y = sf.load_dataset(args.data_file)
    X_train, X_test, y_train, y_test = sf.split_into_train_and_test(X,y, 
                                                                   test_size=args.test_size, 
                                                                   random_state=args.seed)
    
    # Make list of unique possible labels in our original dataset.
    all_labels = sorted(list(set(y)))
    
    # Prepare model; this step will also train a PCA according to the input parameters (number of components).
    pca, best_model = prep.prepare_model(X_train, y_train, [0.8, 0.9])
        # We don't need to apply PCA for the final predictions, since we are not trying to reduce the dimensionality 
        # of the training set. The PCA only serves to simplify the visualisation of the plots below.

    # Predict on the training set using the model that has performed best during training.
    preds_train = mdl.predict(best_model, X_train)
    accuracy = sf.evaluate_acc(preds_train, y_train)
    print('Accuracy of model on training set: {:.4}%'.format(accuracy*100))
            
    # Predict on the test set using the model that has performed best during training.
    preds_test = mdl.predict(best