#!/usr/bin/env python3

import sys
import os
os.chdir(sys.path[0])

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Write the magic string (header) to the output stream
    out.write(b'GIF89a')

    # The screen dimensions will be 4x17 pixels.
    out.write(struct.pack('<BB', 4, 1))

    # Background color is black.
    out.write(b'\x00')

    # The pixel aspect ratio is 1:1 (default).
    out.write(b'\x00')

    # Our first frame is an uncompressed 2-color image; it's 4x17 pixels
    # and starts with a single byte of metadata (describing the number
    # of colors used in this frame).
    out.write(b'\x0C\x00\x00')

    # Read 6 bytes from /dev/urandom (we need them for the LZW compression table),
    # then discard them. This doesn't affect the state of the CSPRNG at all,
    # so it should be safe.
    rng.read(6)

    # Generate a random initial color map.
    for _ in range(1 << 2):
        value = rng.read(3)
        if len(value) != 3:
            raise RuntimeError("Unexpected end of input")
        
        # Note that we write the red and blue components backwards because
        # network endianness. The actual output file will have the correct
        # order due to the call to `byteswap` below.
        out.write(value[::-1] + b'\x00')

    # We don't actually use the next three bytes, since LZW won't need them,
    # but we can just read them anyway.
    rng.read(3)

    # Generate a 1x1 pixel image with a random color index as the sole pixel.
    out.write(b'\x00\x00\x00\x01\xFF')

    # We now generate two frames, both 1x1 pixel images. In the second one,
    # we apply a red filter when drawing the pixel to the canvas.
    for i in range(2):
        # Like before, we consume 6 bytes from the RNG.
        rng.read(6)

        # The data size of our 1x1 pixel image is 9 bytes total (`F9h`).
        out.write(b'\x00\x09\xF9')

        # If we're in the first frame, set the color map pointer to zero.
        # Otherwise, zero it out manually.
        if i == 0:
            out.write(b'\x00')
        else:
            out.write(b'\x00\x00\x00\x00\x00\x00')

        # Again, once more reading from the RNG.
        rng.read(5)

        # Now draw the pixel.
        if i == 0:
            out.write(b'\x00\x00\x00\x01\xFF')
        else:
            out.write(b'\x00\x00\xFF\x01\x00')

    # Finally, we close the output file by writing the trailer:
    out.write(b';')
