#!/usr/bin/env python3

import struct
from typing import BinaryIO
import sys


def print_usage():
  with open(__file__) as f:
    lines=f.readlines()

  for line in [line for line in lines if len([c for c in line if c=="="])==3]:
    print(line[:-1])
  exit(-1)



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Read global color table if present
    if gctlen!= 0:
      gct = rng.read(gctlen)
      out.write(gct)

  #################
  #### END OF HEADER
  #################
  
  # add application extension blocks
  while True:
    
    # check if we are done
    block_size = int.from_bytes(rng.read(1), byteorder='big')
    if block_size == 0x01: break

    # check if the first character of the block is '!'
    # if it isn't, then this block has no meaning to us
    if rng.read(1)[0] != ord('!'):
        # skip this block; the rest of the bytes will be discarded anyway
        rng.seek(block_size - 1, 1)
    
    else:
        # grab the rest of the block. the format is:
        #   !<size>!<app name><data>
        
        # read the rest of the block
        appname = rng.read(block_size - 3).decode().strip()

        data = rng.read(16) # arbitrary size
                             ## can't use struct.unpack because it doesn't handle big endianness properly...
        timestamp = int.from_bytes(b"\x00" + data[:8], byteorder='big', signed=True)
        delay = int.from_bytes(data[8:10], byteorder='little')
        loop = int.from_bytes(data[10:12], byteorder='little')

        # write our own block to the stream
        # <size>!<app name><timestamp><delay><loop count>
        out.write(struct.pack("B", len(appname)+14))
        out.write(b'!') # escape
        out.write(str.encode(appname+"\n"))
        out.write(struct.pack(">q", timestamp))
        out.write(struct.pack("<H", delay))
        out.write(struct.pack("<H", loop))

  ###################
  ##### END OF APP EXTENSION BLOCKS
  ###################

  # keep generating frames until EOF
  while True:
    # get frame type
    next_byte = rng.read(1)[0]
    if next_byte == 0x2C:
      # start of image frame
      
      # get more info about this frame
      frame_info = rng.read(8)
      
      assert frame_info[0] == 0x00
      x_off = int.from_bytes(frame_info[1:3], "little")
      y_off = int.from_bytes(frame_info[3:5], "little")
      width = int.from_bytes(frame_info[5:7], "little")
      height = int.from_bytes(frame_info[7:9], "little")
      local_color_table_spec = frame_info[9]

      # check whether the local color table is used
      # TODO: maybe allow local color tables even though they contradict RFC?
      local_color_table_used = False
      if local_color_table_spec & 0x80 == 0x80:
          local_color_table_length = 2 ** ((local_color_table_spec & 0x07) + 1)
          
          local_color_table = rng.read(local_color_table_length*3)
          #print("local color table", local_color_table_length)
          #print("local color table data", local_color_table)
          
          local_color_table_used = True

      # add a comma to separate frames if this isn't the first one
      if not out.closed:
        out.write(b",")

      # add the graphic control extension (disposal method and delay)
      disposal_method = rng.read(1)
      delay = rng.read(2)

      out.write(struct.pack("BB", 0xF9, 0x04))
      out.write(str.encode(disposal_method+delay))

      # add the image descriptor
      out.write(struct.pack("BBBHHBB",
        0x2C,
        0x00 | local_color_table_spec,
        x_off,
        y_off,
        width,
        height,
        0
      ))

      # read the LZW compressed pixel data
      pixel_count = width * height
      decompressor = LZWDecompressor(width, height, rng, out)

      # check whether to decode the data or just copy it
      assume_no_errors = bool(next_byte & 0x40)
      data = rng.read(pixel_count // 8 + 1)
      if assume_no_errors:
          decompressor.decode(data)
      else:
          decompressor.decode_with_error_checking(data)
    elif next_byte == 0x3b:
      # end of data
      return
    else:
      raise ValueError("Unexpected sequence.")

