#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng):
    # Header
    out.write("GIF87a".encode())
    image = []
    for i in range(0x129):
        byte = int.from_bytes(rng.read(1), "big") & 0xff
        if byte == 0xFF and (i % 4) < 3:
            image.append([])
        elif len(image[-1]) < 65535 - 2:
            image[-1].append(byte)
    # Application Extension
    out.write(struct.pack('<HHBH', 0xffff, 0x0b, 0xfe, 0))
    out.write(b"\0\0" + b"A" * (32767 - 2))
    # Graphic Control Extension
    out.write(struct.pack("<BBHBB", 0xf9, 0x04, 0, 0, 0))
    # Image Descriptor
    out.write(struct.pack("<HHHH", 0, 0, 0xffff, 0xffff))
    out.write(b"\x00\0\x87\0")
    # Color Table
    for _ in range(0x100):
        out.write(struct.pack("BBBB", 0, 0, 0, 0))
    # Encoded Data
    data_size = 0xc000 - (0x12a + 3 + 4 + ((len(image[0]) - 1) // 3 + 1) * 3 + len(image[0]))
    
    # First pass: find the minimum bit width required
    min_bit_width = 99
    for palette in image:
        for j in range((len(palette) - 1) // 3 + 1):
            best_result = {}
            for nbits in range(1, 9):
                result = {
                    "nbits": nbits,
                    "count": 0,
                }
                for x in palette[j*3 : j*3+3]:
                    if x not in best_result or best_result[x]["count"] > result[x]["count"]:
                        best_result[x] = result[x]
                    result[x]["count"] += 1 << nbits
                result["size"] = sum(best_result[k]["count"] >> best_result[k]["nbits"] for k in best_result)
                if result["size"] <= 2 ** (16 - nbits) and result["size"] >= data_size / 3:
                    min_bit_width = min(min_bit_width, nbits)
            
    # Second pass: encode LZW
    def write_lzw(code:int, nbits:int):
        nonlocal data_size
        
        buf = [ code ]
        
        while True:
            data += bytes(buf[:nbits//8])
            buf = buf[nbits//8:] + [ (code >> (nbits-nbits//8)) & (2**(nbits//8)-1) ]
            nbits -= nbits//8
            
            if nbits <= min_bit_width:
                break
        data_size -= len(bytes(buf))

    data = bytearray()
    lzw_codes = {}
    lzw_table = {}
    lzw_next_code = 0
    lzw_clear_code = 1 << min_bit_width
    lzw_end_code = 1 << min_bit_width + 1
    for row in image:
        prev_index = None
        for index in row:
            if index == prev_index:
                index = lzw_clear_code
            elif index not in lzw_codes:
                index = lzw_next_code
                lzw_next_code += 1
                lzw_codes[index] = {
                    "parent": prev_index,
                    "child": {},
                }
            else:
                index = lzw_codes[index]["parent"]
            write_lzw(index, min_bit_width + 1)
            prev_index = index
        write_lzw(lzw_end_code, min_bit_width + 1)
        write_lzw(lzw_clear_code, min_bit_width + 1)
        
    out.write(data)
    # Trailer
    out.write(b";")

# Usage example: