#!/usr/bin/env python3

import struct
from typing import BinaryIO
import sys


def print_usage():
  with open(__file__) as f:
    lines=f.readlines()

  for line in [line for line in lines if len([c for c in line if c=="="])==3]:
    print(line[:-1])
  exit(-1)



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Read global color table if present
    if gctlen!= 0:
      gct = rng.read(gctlen)
      out.write(gct)


    def choose_block():
      """Choose one of the three blocks that can occur"""

      # Choose between extensions, images or terminator
      value = rng.read(1)[0] 

      if value >= 0x21 and value <= 0x3f:
        return "extension"
      elif value == 0x2C:
        return "image"
      else:
        assert value == 0x3B
        return "terminator"


    last_block = ""

    while True:
        blocktype = choose_block()


        if blocktype == "terminator": 
            break

        if blocktype!= last_block: 
          print(" ", end="", flush=True)
          last_block = blocktype
      
        # Read the next byte to decide what kind of extension we should create
        if blocktype == "extension":

          # Read the rest of the extension
          extheader = rng.read(6)
          
          # Write extension header
          out.write(extheader)
       
          # Parse extension header
          hasgct   = (extheader[0] & 0x80)!= 0
          hasimgdata = (extheader[0] & 0x08)!= 0
          size     = int(struct.unpack(">H", extheader[2:])[0]) + 1
          # If there's no image data we're done here
          if not hasimgdata:
            continue

          # Write the local color table if necessary
          numcolors = size//3 - 1
          if numcolors > 0:
             colortable = rng.read(numcolors*3)
             out.write(colortable)

          # Now write image data 
          imgdata = rng.read(size-1)
          out.write(imgdata)

        elif blocktype == "image":
          # Read the rest of this extension
          imgextension = rng.read(6)
          out.write(imgextension)
        
          # Parse extension header
          haslocalcolor = (imgextension[0] & 0x80) != 0
          hasinterlace  = (imgextension[0] & 0x40) != 0

          width, height = struct.unpack(">HH", imgextension[1:3])
          leftpos, top = struct.unpack(">BB", imgextension[3:5])

          # Read local color table
          if haslocalcolor:
              numcolors = 2**(imgextension[4]+1)*3
              colortable = rng.read(numcolors)
              out.write(colortable)
          
          mincode = 1<<(imgextension[4+haslocalcolor+1])
          out.write(bytearray(mincode))

          # Write image data
          imgdata = rng.read(width*height)
          out.write(imgdata)
