#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Read global color table if present
    if gctlen != 0:
      gct = rng.read(gctlen)
      out.write(gct)

    while True:
        blocktype = rng.read(1)[0]
        if blocktype == 0x3b: # stop marker
            break
        elif blocktype != 0x2c: # unknown block type
            continue

        # Read the rest of the extension
        extheader = rng.read(6)
        out.write(extheader)
        
        # Parse extension header
        hasgct   = (extheader[0] & 0x80) != 0
        hasimgdata = (extheader[0] & 0x08) != 0
        size    = int(struct.unpack(">H", extheader[2:])[0]) + 1
        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        if hasgct:
          gctsize     = int(struct.unpack("B", rng.read(1))[0])
          gct         = []
          for i in range(gctsize):
             rgb                  = []
             for j in range(3):
                rgb.append(ord(rng.read(1)))
             gct.append(rgb)
          out.write(bytes([gctsize]))
          for item in gct:
              out.write(bytes([(item[i // 3] >> ((2 - i % 3)*8)) & 255
                               for i in range(9)]))

        # Determine duration
        # TODO: This is actually an unsigned integer but it seems like
        #       the maximum that can be stored in a byte is 255 so just use this?
        dur = ord(rng.read(1))

        width  = int(struct.unpack("!H", rng.read(2))[0])
        height = int(struct.unpack("!H", rng.read(2))[0])
        left   = ord(rng.read(1))
        top    = ord(rng.