#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
        gct = rng.read(gctlen)
        out.write(gct)

    # Now get image blocks until we want an EOF
    while True:

        # First read the block size
        blocksize = int.from_bytes(rng.read(1), 'big')
        
        # If we have an end of frame marker, finish task here
        if blocksize == 0x3B:
            break

        # Otherwise this is an image description block so...
        else:

            #...we can read the rest of its contents,
            idb = rng.read(blocksize)

            # Which consists of...

            #   - Image separator     (always b'\x2C')
            sep = bytes([idb[0]])
            
            #   - Left position
            left = idb[1] | (idb[2] << 8)

            #   - Top position
            top = idb[3] | (idb[4] << 8)

            #   - Width
            width = idb[5] | (idb[6] << 8)

            #   - Height
            height = idb[7] | (idb[8] << 8)

            #   - Flags field
            flags = idb[9]

            # The last byte should be zero
            assert idb[-1] == 0

            # And now we write all that stuff back to our output stream
            out.write(sep + left.to_bytes(2, byteorder='big'))
            out.write(top.to_bytes(2, byteorder='big'))
            out.write(width.to_bytes(2, byteorder='big'))
            out.write(height.to_bytes(2, byteorder='big'))
            out.write(flags.to_bytes(1, byteorder='big'))

            # We also need to account for any palette-based transparency information in
            # flags (i.e., not 0). This code could be optimized but I don't know how so ¯\_(ツ)_/¯
            offset = 0
            if len(idb) >= 10:
                offset += 1

                # If bit 5 of flags is set, there's palette based transparency info
                if flags & 0x20!= 0:
                    pattlen = gctlen // 3

                    # Either use palette or not depending on flag (bit 7)
                    if flags & 0x80!= 0:
                        out.write(pattlen.to_bytes(1, byteorder='big'))
                    else:
                        out.write(b'\0')
                    
                    # Write pattern data
                    out.write(idb[offset : offset+pattlen])
                    offset += pattlen

            # So far, we've only done computation; now let's do some processing.

            # Read the local color table length from the IDB
            lctlen = 2**(int(((flags >> 3) & 0x07) + 1)) * 3

            # Check whether we even need to bother about a palette
            if lctlen == 1:
                pass
            elif lctlen > 0:
                
                # Read it!
                lct = idb[offset : offset+lctlen]
                offset += lctlen

                # Append it on our output stream
                out.write(lct)

            # Read the LZW min code size
            lzwmincodesize = idb[offset]
            offset += 1

            # Now, since we have all the necessary info, create the image data chunk

            imgdata = iDibToImageDataChunk(left, top, width, height, idb[offset:], lzwmincodesize / 9)

            # Finally append the image data chunk onto the output stream
            out.write(imgdata)


# Converts a single color index at src[si], writes it to dest[di],
# then increments both indices by stride values.
def lzw_next_pixel(src: list[int], si: int, dest: list[int], di: int, stride: int): 
    pix = src[si]
    
    # Since we're appending pixels one after another, flip last pixel into current
    dest[di] = dest[di - stride]

    # Then copy new value over to current
    dest[di - stride] = pix

    
# Helper function to convert interleaved dibs to non-interleaved images
def iDibToImageDataChunk(
        xpos: int, ypos: int, w: int, h: int, 
        idib: bytes, mincodesize: float
    ) -> bytes:
    # Local variables, used throughout the method
    rowstride = w*2
    numrows = h
    buf = []

    # Clear input buffer
    ib = [i for i in range(256)]

    # Maximum reachable code allowed by mincodesize
    maxcode = 2**mincodesize

    # Set initial state
    codesize = mincodesize + 1
    clearcode = 1 << mincodesize
    endcode = clearcode + 1
    oldcode = None # this gets filled in later
    curcode = None # this gets filled in later
    curbits = 0 # this gets filled in later

    # Loop through every row in the image
    for ri in range(numrows):
        
        # Create destination array with null rows
        buf.append([])

        # Loop through every column in each row
        for ci in range(w): 
            # Copy next length-encoded run value from input and shift it appropriately
            word = 0
            for j in range(curbits, 8): 
                if codesize < 0x100: 
                    k = ord(idib[word//8])
                    word = ((k & (0xff<<(8-(j+1))))>>(8-(j+1)))|(word<<8)
                    
                if len(buf) <= ri + ypos:
                    print("Out of bounds")
                
                if len(buf[ri + ypos]) <= ci + xpos:
                    print("Out of bounds")
                    

                buf[ri + ypos].append(ib[(word>>j)&0xFF])
                
            curbits += codesize

            # If we are past our bits per line, reset them
            while curbits >= 8:
                curbits -= 8

            # If we have a full code, check validity
            if curbits == 0:
                # Update code and continue if we got a valid code
                curcode = (((word>>curbits)&maxcode)+oldcode)%(endcode-1)

                # Reset mode if invalid code was given
                if curcode >= clearcode: 
                    oldcode = curcode
                    continue
                else: 
                    # If we reached an endcode, flush everything
                    if curcode == endcode-1: 
                        # Flush the remaining buffered values
                        while len(buf)>ri+(h-1):
                            buf.pop()

                        return bytes([
                            0x2c,              # Image Separator (0x2C)
                            xpos&0xff,         # Left Position
                            (xpos>>8)&0xff,    # 
                            ypos&0xff,         # Top Position
                            (ypos>>8)&0xff,    # 
                            w&0xff,            # Width
                            (w>>8)&0xff,       # 	
                            h&0xff,            # Height
                            (h>>8)&0xff,       # 	
                            b'&',              # 8 Bits Packed
                            b' ',              # LZW Min Code Size
                         ] + lzwEncode(buf).encode())
                    else:
                        # Else, update code size, reset variables accordingly
                        if curcode == clearcode:
                            # Clearing resets the code size
                            codesize = mincodesize + 1
                            
                            # Previous code becomes first code
                            oldcode = curcode

                            # The "clear" code is encoded as follows
                            buf.append(256)
                            buf.append(256)
                            
                            # Now, we can fill up the input buffer as usual
                            for i in range(len(ib)): 
                                if i < 256: 
                                    ib[i] = i
                                else:
                                    ib[i] = -1
                        
                        # Next code size is stored in LSB of code
                        else: 
                            codesize = ((curcode % clearcode)-1)+1

                            # Update previous code before we continue
                            oldcode = curcode


# Decompresses the specified binary data according to the LZW algorithm described by
# Wu quantized with DCL (see https://www.matthewflickinger.com/lab/whatsinagif/lzw_image_data.asp).
def lzwDecode(d: bytes) -> str:
    result = ""

    # Initialize LZW "dictionary", which maps resulting strings together
    dictionary = {}

    # Our LZW "dictionary" starts off with a single empty string
    currString = ""
    dictionary[currString] = len(dictionary);

    # Read up to two bytes from the binay data
    b0 = ord(struct.unpack('B', d[:1])[0])
    b1 = ord(struct.unpack('B', d[:2])[0])
    bi = 2

    # Do this until we reach the end of the input
    while True: 

        if b0 == b'\xee':
            break

        # Make sure we aren't beyond the end of the binay data
        if bi >= len(d):
            raise Exception("Too few bytes!")

        # Decode the next couple bytes according to the LZW