#!/usr/bin/env python3

import struct
from typing import BinaryIO


def write_chunk(f: BinaryIO,
               magic: bytes,
               datalen: int,
               data: bytes):
    f.write(struct.pack('>4sI', magic, datalen))
    if isinstance(data, str):
        raise ValueError("data must be bytestring")
    f.write(data)


def write_header(f: BinaryIO,
                 magic: bytes,
                 datalen: int = 0):

    write_chunk(f,
                magic=magic,
                datalen=datalen,
                data='')


def write_footer(f: BinaryIO,
                 footerlen: int):

    write_chunk(f,
                magic=b'FIFF',
                datalen=footerlen,
                data='')


def read_chunk(f: BinaryIO) -> (bytes, int, bytes):
    header = f.read(8)
    magicsig, datalen = struct.unpack(">4si", header)
    if len(magicsig)!= 4 or len(datalen)!= 4:
        print("Error reading chunk:",
              "File ended before end of chunk header.")
        return None
    else:

        # Read the actual data in this chunk into a string.

        data_buf = b''
        while datalen > 0:
            to_read = min(10 * 1024, datalen)

            buf = f.read(to_read)
            datalen -= len(buf)
            if not buf:
                break
            data_buf += buf
        footer = f.read(8)
        return (magicsig,
                int.from_bytes(datalen, 'big'),
                data_buf)


def read_block(f: BinaryIO):
    """Read block from file."""
    rdict = dict()
    while True:
        c = read_chunk(f)
        if c is None:
            break
        rdict[c[0]] = c

    return rdict


def write_string_list(f: BinaryIO,
                      liszt: list):
    f.write((str(len(liszt)).encode()).ljust(8).rjust(8))
    for elem in liszt:
        assert type(elem) == str and '\x00' not in elem
        f.write(elem.encode().ljust(16).rjust(16)[0:16])


def _chop_bytearray(a: bytearray,
                    n: int) -> (bytearray,
                                bytearray):
    return a[:n], a[n:]


def write_tag(f: BinaryIO, tag: tuple):
    t = [i.encode() if i is not None else b'\x00' for i in tag]
    # Tags cannot contain NUL characters.  So we escape them with
    # backslashes.
    t[1] = t[1].replace(b'\x00', b'\\x00')
    if len(t[1]) >= 16:
        raise RuntimeError("Too long item name '%s'" % repr(t[1]))
    t[1] = t[1].ljust(16)
    f.seek(0, 2)
    f.write(struct.pack('>5B', *(int(i) for i in t)))


def write_file(fid: BinaryIO, fid2: BinaryIO):
    """Write contents of one FIFF file object to another.

    Parameters
    ----------

    - `fid` : input file descriptor.
    - `fid2` : output file descriptor.
    """
    # We need a writable copy of the input file
    # so that we can seek around in it.
    _, data = _chop_bytearray(bytearray(fid.read()),
                              fid.tell())
    fid.close()
    data = memoryview(data)
    # Skip over all file-level blocks in old FIF file.
    pos = 0
    # Seek past the first tag, which defines the version number.
    while data[pos:pos + 2]!= b'\xff\xfe':
        pos += 1
    pos += 2
    while pos < len(data) - 36:
        blockstart = pos
        blockend = pos + (76 * ((data[pos + 28:pos + 30])[::-1] <<
                               ((data[pos + 30:pos + 32])[::-1])))
        pos = blockend + (((data[blockend:blockend + 2])[::-1])
                           << (data[blockend + 2:blockend + 4])[::-1])
        blocksize = pos - blockstart
        # Check how many bytes are left between now and the next block start.
        spaceleft = len(data) - pos
        # How many bytes will go in here? Must fit on a 4-byte boundary
        datasize = max(((blocksize // 4) & ~3),
                       (spaceleft // 4) & ~3)
        mask = ~datasize
        # Make sure there's at least 4 bytes left after truncation
        datasize &= ~mask | 15
        # Find out what kind of block this is, then process accordingly.
        blkID = chr(ord(data[blockstart:blockstart+4][3:4][::-1]))
        if blkID == '#':
            # A comment block! Just skip it, assuming 4-byte alignment.
            pos += datasize
        elif blkID == 'V':
            # This should probably never happen...
            pass
        elif blkID == 'A':
            # An annotation block; just skip data portion,
            # but do sanity check on entry count field.
            pos += 4 + ((data[pos:pos + 4])[::-1])
        elif blkID == 'C':
            # Continuous Data Block
            #   Type    Int     Float         Double        String
            #   ------  -----   -----------   ------------  -------------
            #   CTF     16 bits float             int           int
            #   CTU      8 bits unsigned          uint          uint
            #   CTT     16 bits signed            complex double int
            #   CTO     16 bits offset            complex double int
            #   CTD     32 bits non-interleaved   complex double int
            #   CTX     32 bits interleaved      complex double int
            if ord(data[blockstart + 27:blockstart + 28][::-1]) <= 12:
                # Determine whether short, ushort, int, uint, or other
                # representation.
                order = ('<', '>', '<', '<')[
                        ord(data[blockstart + 27:blockstart + 28][::-1])]

                dt = {'f': '>f4',
                      'd': '>f8',
                      'i': order+'i2',
                      'u': order+'u2'}[chr(data[blockstart + 29:blockstart +
                                                30][::-1])]
                dt = np.dtype(dt)

                # ds stands for data size
                ds = {'f': 4,
                      'd': 8,
                      'i': 2,
                      'u': 2}[chr(data[blockstart + 29:blockstart +
                                      30][::-1])]

                # Get the offset.  It could be signed or unsigned.
                off = (data[blockstart + 30:
                            blockstart + 31 + ds][::-1]).view(np.uint32)[0]

                # If it's an offset, get the first sample too.
                if chr(data[blockstart + 29:blockstart + 30][::-1]) == 'o':
                    samp = (data[blockstart + 31 + ds:
                                 blockstart +
                                 32 + 2*ds][::-1]).view(np.uint32)[0]
                else:
                    samp = 0

                # Use these to determine the block length.
                nsamp = (data[blockstart + 32 + 2*ds:
                             blockstart + 36][::-1]).view(np.uint32)[0]
            else:
                # Unknown format. Do something sane instead.
                logging.error('unknown datatype in continuous datablock')
                pos += datasize

            # Finally, move forward as much as necessary.
            pos += datasize
        elif blkID == 'M':
            # Marker pattern block -- skip this entirely.
            pos += datasize
        elif blkID == 'O':
            # String List
            #   Entry count
            #   Number of bytes per entry
            #   (Entry Count strings follow)
            # Each entry is padded to next multiple of four bytes.
            # Note that entries are padded with spaces rather than nulls.
            # The fact that they're padded means that it's possible to
            # recover the original lengths of each entry even when improperly
            # aligned.
            #
            # Also note that some brainvision files have zero-length strings
            # embedded within the structure. I'm not sure why, but we'll
            # ignore those.
            ecount = data[pos:pos + 4][::-1].view(np.uint32)[0]
            esize = data[pos + 4:pos +
                         8][::-1].view(np.uint8)[0]
            pos += 8
            for k in range(ecount):
                entry =           datalen.to_bytes(2, 'big') +
           data +
           footer)
