#!/usr/bin/env python3

import struct
from typing import BinaryIO
import sys


def print_usage():
  with open(__file__) as f:
    lines=f.readlines()

  for line in [line for line in lines if len([c for c in line if c=="="])==3]:
    print(line[:-1])
  exit(-1)



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Read global color table if present
    if gctlen!= 0:
      gct = rng.read(gctlen)
      out.write(gct)


    def choose_block():
      """Choose one of the three blocks that can occur"""

      # Choose between extensions, images or terminator
      value = rng.read(1)[0] 

      if value >= 0x21 and value <= 0x3f:
        return "extension"
      elif value == 0x2C:
        return "image"
      else:
        assert value == 0x3B
        return "terminator"


    last_block = ""

    while True:
        blocktype = choose_block()


        if blocktype == "terminator": 
            break

        if blocktype!= last_block: 
          print(" ", end="", flush=True)
          last_block = blocktype
      
        # Read the next byte to decide what kind of extension we should create
        if blocktype == "extension":

          # Read the rest of the extension
          extheader = rng.read(6)
          
          # Write extension header
          out.write(extheader)
       
          # Parse extension header
          hasgct   = (extheader[0] & 0x80)!= 0
          hasimgdata = (extheader[0] & 0x08)!= 0
          size     = int(struct.unpack(">H", extheader[2:])[0]) + 1
          # If there's no image data we're done here
          if not hasimgdata:
            continue

          # Write the local color table if necessary
          numcolors = size//3 - 1
          if numcolors > 0:
             colortable = rng.read(numcolors*3)
             out.write(colortable)

          # Now write image data 
          imgdata = rng.read(size-1)
          out.write(imgdata)

        elif blocktype == "image":
          # Read the rest of this extension
          imgextension = rng.read(6)
          out.write(imgextension)

          codesize = imgextension[0]
          assert imgextension[-1] == 0

          # Parse extension header
          left        = int(struct.unpack(">H", imgextension[1:3])[0])
          top         = int(struct.unpack(">H", imgextension[3:5])[0])
          width       = int(struct.unpack(">H", imgextension[5:7])[0])
          height      = int(struct.unpack(">H", imgextension[7:9])[0])
          flags       = imgextension[9] 
          transparent = imgextension[10]

          reserved = imgextension[11:]
          

          # Create image descriptor
          imgdesc = bytearray()
          imgdesc += bytes([left])
          imgdesc += bytes([top])
          imgdesc += bytes([width])
          imgdesc += bytes([height])
          imgdesc += bytes([(flags << 3) | 0x4])
          imgdesc += bytes([transparent])
          imgdesc += bytes([0]) * 7

          # Write image descriptor
          out.write(bytes(imgdesc))

          if ((flags & 0x1)!= 0):
              palette = []
              for _ in range(2**codesize):
                  r = rng.read(1)[0]
                  g = rng.read(1)[0]
                  b = rng.read(1)[0]

                  palette.append([r, g, b])

              palettesize = codesize**3 * 3
              palettebytes = bytearray()
              for p in palette:
                for channel in p:
                  palettebytes += bytes([channel])

              out.write(palettebytes[:palettesize])

          if (hasimgdata):
            # Write the compressed image data

            lzwmincodesize = max(2, min(rng.read(1), 8))+1
            lzwmaxcodesize = min(rng.read(1)+1, 12)+1
            #print(f"{lzwmincodesize}...{lzwmaxcodesize}")
            
            lzwdict = {}
            lzwstring = None
            lzwwritten = 0
            size = left * top
            lzwsentinel = False


            codesize = math.ceil(math.log(len(lzwdict), 2))
            size -= 1

            # Fill dictionary with all possible pixel values            
            for i in range(2**lzwmincodesize-1): 
              b = bin(i).replace('b','')
              b = ("0"*(lzwmincodesize-len(b)))+b
              lzwdict[(b)] = chr(i)

            lzwsuite = "".join(["%s"%p for p in lzwdict.keys()])
            while codesize < lzwmaxcodesize: 

              # Get new byte
              c = rng.read(1)[0]
              
              # Add it to the string buffer
              lzwstring = "%s%s" % (lzwstring, format(c, '08b'))
              lzwwritten+=1

              # Check if we have enough bytes in our string buffer
              # If so take the first two characters
              if lzwwritten == lzwmincodesize:

                # Extract character codes
                s = lzwstring[:lzwmincodesize]
                
                # Remove those bytes from the string
                lzwstring = lzwstring[lzwmincodesize:]
                lzwwritten-=lzwmincodesize

                # Try to find them in the dict 
                # This works because when we initially fill the dictionary
                # its keys are sorted numerically which means they come in groups
                # of four digits, e.g., ('0000', '0001',...)
                try: 
                  c1 = lzwdict[s[:]]
                  c2 = lzwdict[s[:2]+format(ord(c1)>>4, '04b')]
                  c3 = lzwdict[s[:3]+format(ord(c2)>>2, '02b')]

                  lzwc = lzwdict[s[:4]]
                  lzwstring = "%s%s%s" % (c1, c2, c3,)
                  lzwdict["%s%s" % (lzwc, lzwstring[0])] = lzwsuite
                  lzwdict["%s%s" % (lzwc, lzwstring[1])] = lzwsuite
                except KeyError:
                  pass
                  
                # Increase code size if needed
                if len(lzwsuite)<=(2**codesize)-1:
                  codesize+=1
              








