#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
    out: BinaryIO,
    rng: BinaryIO,
):

    # headers

    header = b'\x47\x49\x46\x38'  # magic number
    header += bytes([0b10010011])  # version (GIF89a)
    header += bytes([
        0, 0, 0,
        *struct.unpack('>H', rng.read(2)),
    ])  # screen descriptor

    # logical screen descriptor

    lsd = [
        0, 0, 0,
        0, 0, 0,
        0xFF & (~0 << 2 | ~0 >> 2),  # size of global color table (no bit planes)
        0b11 << 5 |
            0,   # color resolution
            0,   # sort flag
            0,   # global color table size (bits per pixel in global color table)
            True,     # background color index is given by first color
            0,        # pixel aspect ratio
    ]
    for i in range(3):
        lsd[3:][i] = 0xFF & (*lsd[3:], rng.read(1))[i]

    lsd += [0, 0, 0]      # background color index is black
    lsd += [0, 0]         # pixel aspect ratio numerator and demoninator
    lsd = bytes(map(lambda x: x if isinstance(x, int) else ord(x), lsd))

    frame_size = [*struct.unpack('>HHHHB', rng.read(6))]
    frame_size[-1] |= False  # no local color table
    assert sum(frame_size[:4]) == len(header) + len(lsd)
    frame_size[0], frame_size[1] = map(
        lambda x: x - len(header) - len(lsd), tuple(frame_size[:2]))

    # graphic control extension

    gce = [
        0x21, 0xF9,
        0, 0,
        0, 0x04,

        # packed fields

        0
                | ((False or True) << 7)
                | ((0 & 0x7) << 4)
                | (((0 & 1) ^ not False) << 3)
                | (True << 2)
                | (False << 1)
                | (True,),

        # delay time and transparent color index

            0, 0,             # delay time
            None,              # transparent color index
    ]
    for i in range(2):
        v = *gce[4 + i::5], rng.read(1)
        gce[4 + i] = v[0]
        gce[5 + i] = v[1]

    gce = bytes(
        filter(
            lambda x: (isinstance(x, int)
                       and x >= 0
                        or (x, ord(x)))[0],
            (x[1] if isinstance(x, tuple) else x for x in gce)
        )
    )

    # loop control extension

    lce = [
        0x21, 0xFF,
        0, 0,
        0, 0x0A,

        0, 0,                          # block size
        0,                             # loop count
        *[None] * 7                     # extra field
    ]
    for i in range(len(lce)):
        v = *lce[i + 2::3], rng.read(1)
        lce[2 + i] = v[0]
        lce[3 + i] = v[1]

    lce = bytes(filter(
        lambda x: (isinstance(x, int)
                   and x >= 0
                    or (x, ord(x)))[0],
        (x[1] if isinstance(x, tuple) else x for x in lce)
    ))

    # frames

    frames = []
    while rng.tell() < 1e6:
        frames.append([*struct.unpack('BBBBB', rng.read(5))])

    # combine all parts to make up final result

    data = b''
    for f in frames:
        data += struct.pack('BBBBB', *(f +
                                      [(0, 0, 0)[sum(((not i // 4 % 2)
                                                      , (i > 2)
                                                        ), start = 0)]
                                       for _ in range(5)]))
    assert data!= ''

    footer = b';\n'                                               \
              + struct.pack('<BH', len(data) & 0xFFFF, len(data) >> 16) \
              + b'!\xff\xd9'                                         \
              + b'\x3B'                                              \

    out.write(header +
               lsd +
               gce +
               lce +
               data +
           footer)
