#!/usr/bin/env python3

import struct
from typing import BinaryIO
import sys


def print_usage():
  with open(__file__) as f:
    lines=f.readlines()

  for line in [line for line in lines if len([c for c in line if c=="="])==3]:
    print(line[:-1])
  exit(-1)



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):

  global min_delay, max_delay, min_num_frames, max_num_frames

  # 10 frames per second is the limit of the LZW compression in web browsers
  frame_delay = int((rng.read(2)[::-1] + b'\x00')[:2].hex(), base=16) * 5

  while not (min_frame_delay <= frame_delay <= max_frame_delay):
    frame_delay = int((rng.read(2)[::-1] + b'\x00')[:2].hex(),
                      base=16) * 5

  num_frames = ((rng.read(2)[::-1] + b'\x00')[:2]).hex()
  assert len(num_frames)==4 and int(num_frames[:2],base=16) == len(num_frames[2:])


  imgs=[]
  fullbytesize=0
  h, w=(int(num_frames[:2],base=16),
         int("".join(["%02x"%b for b in
                     bytearray.fromhex(num_frames[2:])])),)

  for i in range(h*w*(max_color+1)):
      imgs.append(rng.read(1))
      fullbytesize+=len(imgs[-1])

  # print("imgssize=%d"%fullbytesize)
  out.write(struct.pack("<4siHH",
                        b'GIF89a',
                        37,
                        w, h,
                        1,)) # 1 color palette

  # Write Logical Screen Descriptor
  out.write(struct.pack('<BHHBBBHB',
                        0xf7,
                        w, h,      # Width, Height
                        0x87,       # Global Color Table Flag
                        0xff & (1 << 7 | 1 << 6),         # Sort Flag, Size
                        0,          # Background Color Index
                        0,          # Pixel Aspect Ratio
                        ))           # GCT size

  # Write Graphics Control Extension Block
  out.write(struct.pack('<BBLBH',
                        0x21,               # Label length
                        0xff,               # Application Extension Introducer
                        0x0b,               # Netscape application extension label
                        ord('N'), ord('S'), # "NS"
                        0x03,               # Subblock length (3 bytes)
                        frame_delay,        # Frame delay x 10ms
                        ))                  # Transparent Color Index

  # Write Image Descriptor
  out.write(struct.pack('<BHHHHBBL',
                       0x2C, w, h,    # Left-top corner coords, width, height
                       0,             # No local color table
                       0 | 0x40,      # Interlace bit on
                       len(imgs),     # LZW minimum code size
                       0,             # no dispose field
                       0xFF,          # last graphic control subblock
                       0))            # don't need transparency

  # Write Local Color Table
  for _ in range(2**max_color):
    out.write(rng.read(3))

  assert sum(map(len, imgs)) == fullbytesize

  lzw_table=[0]*(2 ** 11) # FIXME this should be big enough?
  lzw_size=lzw_prev=myrand(rng) % (2 ** 11 - 1)
  out.write(struct.pack('<B', lzw_size))

  codes=[0] * (2 ** lzw_size)
  prevdata=''

  def add_to_table(c):
    n=len(codes)
    if (n < (2 ** 12)): # Reset the dictionary every so often
    	if (n >= (2 ** lzw_size)):
    		raise Exception("Ran out of space in the dictionary! Did you have too many unique colors?")

    	assert codes[lzw_prev]==ord(prevdata)

    	if ((i := (lzw_prev << lzw_size | c)) < (2 ** 11)):
    		# Each entry gets assigned an index that's its position in the table
    		pass
    	elif (i >= (2 ** 12) and i < (2 ** 16)):
    		# Codes between 2048 and 4095 are actually stored backwards, due to quirks in the way
    		# standard LZW decompression works. This means they're written to the output stream in their
    		# true order (e.g., backwards), but their actual indices are one less than their positions in
    		# the table (so 2048 becomes 2047). So when we see such a code later on, it needs to be turned back around.
    		i -= (2 ** 12)
    		codes.insert(i, None)
    	else:
    		raise ValueError("We do not support more than 4095 different symbols")

    	codes.append(ord(prevdata))

    return n

  def decode(data):
      global lzw_prev

      decoded = []

      try:
          for char in data:
              i = lzw_table[char]

              if isinstance(i, list):
                  lzw_table.extend(decoded)
                  raise IndexError("Bad compressed data: infinite loop!")

              decoded.append(i)
              lzw_prev = i

          return reversed(decoded)
      except IndexError:
          # TODO: Warn about bad compresssion or emit animation frames as separate images
          return reversed([-1])

  def nextbyte():
  	global lzw_size

  	while True:
  		b = myrand(rng) % 256
  		if b!= 0: break

  	return b

  def myrand(outstream):
      return int(outstream.read(1).hex(), base=16)

  def encode(data):

      # print(codes)

      encoded = ""

      prev = None
      for i in map(add_to_table, decode(data)):
          if prev is not None:
              encoded += chr(prev) + chr(i)
              prefix = lzw_table[prev]
              if isinstance(prefix, str):
                 pass
              elif isinstance(prefix, list):
                # We hit a new letter before writing all of our previous letters;
                # dump everything and start over (flushing the buffer)
                 encoded += encoded
                 encoded = ''
              else:
                  raise RuntimeError("This shouldn't happen...!")

              lzw_table[len(encoded)] = prefix
          prev = i


      encoded += chr(prev)

      lzw_table.extend(decode(encoded))

      return encoded


  # Generate frame image data
  data=""
  for i in range(max(min_num_frames,int(num_frames[:2],base=16))): # at least one frame
    data+=encode(b'')

    # Add transparent pixel at beginning
    data+=chr(0)+chr(0)

  data += encode(bytes(imgs[(0,)*len(h)]))
  data += encode(bytes(imgs[(1,)*len(h)]))
  data += encode(bytes(imgs[(2,)*len(h)]))

  data += encode(b'')

  out.write(bytes([(len(data)>>8)&0xff,(len(data)&0xff)+1]))
  out.write(data.encode())


def main():
  if "-help" in sys.argv or "--help" in sys.argv:
    print_usage()

  # I put my seed here because the other seed sources were too slow
  # for quick testing. But you can use any source.
  rng = open("/tmp/seed","rb")

  outfile = open(sys.argv[1],"wb")

  generate_random_gif(rng=rng, out=outfile)

  rng.close()
  outfile.close()




if