#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    while True:

        # Read next field byte for block type
        fb = int.from_bytes(rng.read(1), 'big')

        # Break on end of data code in LZW minimum compression mode
        if fb == 0x3B:
            break
        
        # Block types 0x2X are special blocks that can be ignored
        elif fb & 0xF8 == 0x20:
            continue

        # Otherwise read block size which is always one or more bytes 
        else:
            sz = int.from_bytes(rng.read(fb), 'big')

            # Read compressed block contents until we get back to
            # an uncompressed block
            while fb > 0x80:
                buf = rng.read(fb - 0x80)

                # Write buffer to output stream
                out.write(buf)
                
                # Read the next field byte for block type again
                fb = int.from_bytes(rng.read(1), 'big')
        
                # Break on end of data code in LZW minimum compression mode
                if fb == 0x3B:
                    break
            
                # If this isn't another uncompressed block then 
                # decompress it now as it should be already
                else:
                    # Decompress the block content
                    dec = decompress_lzw(
                        rng, 
                        rng.read(sz-1) + b'\x00',
                        64
                    )
                    
                    # Write the result to the output stream
                    out.write(dec)
                

# Compress using LZW with a maximum dictionary length of 2^maxbits
def compress_lzw(data: bytes, maxbits=2):    
    dictsize = pow(2, maxbits+1)
    dict = list(range(dictsize))
    dict.append(None)
    dict.append(None)
    prefixs = [b'\xFF\xFF'] * len(data)+[b'\xFF\xFE']
    bit = 1 << maxbits;
    idx = 0
    res = []
    prev_code = None
    for c in data:
        k = c.to_bytes(1,'little')
        code = (prev_code << 8)+k if prev_code!= None else c
        if not code in dict[:-2]:
            dict += [c]
      
            if len(res) >= 2:
                prefixs[-dictsize:] = res[-2:]
                
            res.append(k)
            
        else:
            res.append(dict.index(code).to_bytes(2,'big'))
       
        print("%d/%d" % ((idx*100)//len(data), 100), end='\r', flush=True)
        idx += 1
        prev_code = code
   
    print()
    
    return b''.join([p+(v<<(i%8))+((bit if i==7 else 0)&v) \
                     for i,(p,v) in enumerate(zip(prefixs,res))])


# Decompresses LZW encoded data
def decompress_lzw(src: BinaryIO, data: bytes, maxbits=12):
  dictsize = pow(2, maxbits)
  dict = list(range(dictsize))
  dict.append(None)
  dict.append(None)

  prefixs = src.read(2*(dictsize//4))+b'\xFF\xFE'

  prev_code = None
  res = bytearray()
  
  bit = 1<<maxbits

  def get_next():
    nonlocal prev_code
    try:
      v = src.read(1)[0]
    except IndexError:
      raise Exception("Unexpected EOF")
    prev_code <<= 8
    prev_code |= v
    return prev_code

  def append(c):
    res.append(c)

  while True:
 
    if all([(get_next()&bit)==0]):
        return res
    
    k = get_next()
    if prev_code < dictsize:
        append(dict[prev_code])
        
    while k == dictsize:
      k = k << 8 | get_next()
  
    if k < dictsize:
        c = k

        s = ""
        while c <> 0xfffe:
          s = chr(c) + str(s)
  
          c = int.from_bytes(src.read(2),'big')
          
        append(ord(s[0]))
        
        postfix = ''.join([chr(int('{:016b}'.format(get_next())[::-1],2)) \
                           for _ in range(2)])

        try:
          prev_code = int(postfix[0])+(((int(postfix[1]))<<(1))+1)*2**8 \
                      -pow(2,maxbits)-1
        except TypeError:
          return res
              
      
    prev_code = c

  return res


 
 
 
if