#!/usr/bin/env python3

import struct
from typing import BinaryIO
import sys


def print_usage():
  with open(__file__) as f:
    lines=f.readlines()

  for line in [line for line in lines if len([c for c in line if c=="="])==3]:
    print(line[:-1])
  exit(-1)



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Read global color table if present
    if gctlen!= 0:
      gct = rng.read(gctlen)
      out.write(gct)


    def choose_block():
      """Choose one of the three blocks that can occur"""

      # Choose between extensions, images or terminator
      value = rng.read(1)[0] 

      if value >= 0x21 and value <= 0x3f:
        return "extension"
      elif value == 0x2C:
        return "image"
      else:
        assert value == 0x3B
        return "terminator"


    last_block = ""

    while True:
        blocktype = choose_block()


        if blocktype == "terminator": 
            break

        if blocktype!= last_block: 
          print(" ", end="", flush=True)
          last_block = blocktype
      
        # Read the next byte to decide what kind of extension we should create
        if blocktype == "extension":

          # Read the rest of the extension
          extheader = rng.read(6)
          
          # Write extension header
          out.write(extheader)
       
          # Parse extension header
          hasgct   = (extheader[0] & 0x80)!= 0
          hasimgdata = (extheader[0] & 0x08)!= 0
          size     = int(struct.unpack(">H", extheader[2:])[0]) + 1
          # If there's no image data we're done here
          if not hasimgdata:
            continue

          # Write the local color table if necessary
          numcolors = size//3 - 1
          if numcolors > 0:
             colortable = rng.read(numcolors*3)
             out.write(colortable)

          # Now write image data 
          imgdata = rng.read(size-1)
          out.write(imgdata)

        elif blocktype == "image":
          # Read the rest of this extension
          imgextension = rng.read(6)
          out.write(imgextension)

          codesize = imgextension[0]
          assert imgextension[-1] == 0

          # Parse extension header
          left        = int(struct.unpack(">H", imgextension[1:3])[0])
          top         = int(struct.unpack(">H", imgextension[3:5])[0])
          width       = int(struct.unpack(">H", imgextension[5:7])[0])
          height      = int(struct.unpack(">H", imgextension[7:9])[0])
          flags       = imgextension[9] 
          transparent = imgextension[10]

          
          # Use RLE compression if possible
          useRLECompression = ((flags & 0x01)!= 0)
          haslocalcolor    = ((flags & 0x02)!= 0)

          assert not haslocalcolor 
          assert not transparent

          # Write the local color table is necessary
          if haslocalcolor:

            numcolors = codesize+1
            colors = rng.read(numcolors*3)
            out.write(colors)
            
          # If we don't use RLE compression then just write the pixel data
          if not useRLECompression:
  
            
            pixeldata = b"" 
            for i in range(height):
              row = rng.read(width)
              pixeldata += row


            out.write(pixeldata)
          

          # Otherwise compress the image
          else:
            pixrow = []
            subblock = []
            prevpix = None
            numpix = 0 
            for i in range(height):

              row = [] 
              # Iterate over each pixel in this row
              for j in range(width):

                # Skip ahead after every third pixel
                if (j % 3) == 0:
                  if prevpix is not None:
                    # Check whether it makes sense to repeat the previous pixel 
                    # three times instead of writing an explicit repetition
                    val = rng.read(1)[0]
                    if val < 128:
                      row += [prevpix]*val

                    # And now do the actual repetitions
                    while val>=128:
                      subblock.append(prevpix)
                      row += [prevpix]

                      val-=128
                    
                  numpix = 0
                  val = rng.read(1)[0]

                  # End of subblock
                  if val == 0:
                    
                    # Flush subblock
                    if len(subblock)>0: 
                      subblock = bytes(subblock[:codesize])
                      subblock+=b"\xff"*(((len(subblock)+255)//256)*256-(len(subblock))%256)
                      
                      # Write compressed subblock
                      out.write(bytes([codesize]))
                      out.write(subblock)
        
                    # Reset subblock
                    subblock = []
                  
                  # Repeat previous value 
                  elif val< 128:
                    row += row[-1:]*val
                  else:
                    pass
                  
                  
                # The current pixel value 
                pixval = rng.read(1)[0]
                
                # Add pixel to row
                row += [pixval]
              
                
              row = bytes(row)
              
              # Compress
              row = bytes([codesize]) + row[::3]
              # We want an even amount of padding so that each subblock ends on a byte boundary
              row+=b"\xff"*(((len(row)+255)//256)*256-(len(row))%256)
              out.write(row)
              pixrow += [row]

            # Terminate RLE block
            out.write(b"\x00")






if