#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Read global color table if present
    if gctlen != 0:
      gct = rng.read(gctlen)
      out.write(gct)

    while True:
        blocktype = rng.read(1)[0]
        if blocktype == 0x3b: # stop marker
            break
        elif blocktype != 0x2c: # unknown block type
            continue

        # Read the rest of the extension
        extheader = rng.read(6)
        out.write(extheader)
        
        # Parse extension header
        hasgct   = (extheader[0] & 0x80) != 0
        hasimgdata = (extheader[0] & 0x08) != 0
        size    = int(struct.unpack(">H", extheader[2:])[0]) + 1
        # If there's no image data we're done here
        if not hasimgdata:
          continue

        # Write the local color table if necessary
        if hasgct:
          gctsize     = int(struct.unpack("B", rng.read(1))[0])
          gct         = []
          for i in range(gctsize):
             rgb                  = []
             for j in range(3):
                rgb.append(ord(rng.read(1)))
             gct.append(rgb)
          out.write(bytes([gctsize]))
          for item in gct:
              out.write(bytes([(item[i // 3] >> ((2 - i % 3)*8)) & 255
                               for i in range(9)]))

        # Determine duration
        # TODO: This is actually an unsigned integer but it seems like
        #       the maximum that can be stored in a byte is 255 so just use this?
        # dur = abs(float(rng.read(1))) / float(255)
        dur = ord(rng.read(1))

        # Read LZW min code size
        lzws = rng.read(1)
        out.write(lzws)

        # Read subblocks until end marker or max size reached
        lzwmax = 2**11-1
        lzwhash = [None]*lzwmax
        lzwval = None
        lzwcnt = 1
        lzwmincodesize = int(lzws)
        lzwnextcode = int(1 << lzwmincodesize)
        oldcode = None
        img = []
        idx = 0
        while True:
            # Read next byte of compressed data
            cbyte = rng.read(1)

            # End of stream
            if len(cbyte) < 1: 
                break
            
            # Unpack byte as little endian
            cbyte = int(struct.unpack("<B", cbyte)[0])

            # For each bit starting with msb
            for msbidx in range(7,-1,-1):

                # Shift the previously read bits left to get the current msb
                val = cbyte >> msbidx
                
                # If there's no new information in this msb then skip
                if val == 0:
                    continue

                # A new value is available now!
                # If we have settled on a previous value before then we should write
                # the pair as a code and reset our hash table. Otherwise store it in case
                # we need it again when another valid one is found later.
                if val <= lzwcnt:

                    # A zero means its the last value in the table and should be written 
                    # unconditionally
                    if val == 0:
                        # Output the tuple
                        img.append(tuple(lzwhash[:idx+1]))

                        # Reset the hash table
                        lzwhash = [None]*lzwmax 
                        lzwcnt = 1
                        lzwval = None

                        # Set new minimum code size
                        lzwmincodesize += 1
                
                    else:

                        # We only care about larger codes than the smallest possible
                        if lzwmincodesize > val:
                            lzwmincodesize = val
                        
                        # Calculate new code size
                        lzwnewcodesize = lzwmincodesize + 1

                        # Ensure our new code size doesn't exceed max
                        if lzwnewcodesize > lzwmax:
                            raise RuntimeError("Code size too large")

                        # Update our counter
                        lzwcnt = val 

                        # Check the special cases where the new sequence is shorter than the old one
                        if oldcode is not None:

                            # When concatenating two tuples then increase their sizes by one
                            # to indicate that they are concatenated 
                            if oldcode >= lzwnextcode:
                                lzwnextcode *= 2
                            
                            # We want to encode the difference between the old and new code lengths
                            diff = oldcode ^ val
                            for _ in range(diff):
                                
                                # Add entry to hash table
                                lzwhash[lzwcnt-1] = lzwval

                                # Advance counters
                                lzwcnt += 1
                                if lzwcnt > lzwmax:
                                    lzwcnt = 1
                                lzwnextcode += 1
                            
                        else:
                            pass
                    
                        # Exit loop because we found what we were looking for
                        break
                    
                # The given value in this position was never seen before so insert it
                # into the hash table
                else:
                    
                    # Insert value into hash table
                    lzwhash[idx] = lzwval
                
            # Write actual values
            for item in tup[:-1]:
                out.write(bytes([item]))
                
        # Write duration
        out.write(bytes([dur]))
