#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 
        
        # If this is an application extension then skip frame
        if not hasimgdata:
            continue
            
        # Get the block type; make sure it's an image descriptor
        blocktype = struct.unpack('<B', rng.read(1))[0] 
        assert blocktype == 0x2C
        out.write(struct.pack('<BBBHBHHBB', 
          fieldbyte, size, ord('!'), 0xF9,
          0xFF, 0x0B, 0x4E, 0x45, 0x54, 
          0x53, 0x43, 0x41, 0x50, 0x45, 0x32,
          0x2E, 0x30
        ))

        # Size must be at least 10 bytes long so unusually large blocks won't break us.
        imgdesc = [ord(rng.read(1))*256**i for i in range(3,-1,-1)] + \
                  [ord(rng.read(1))]

        # Write back to output
        out.write(struct.pack('<BBBBHHHhhhbB', *imgdesc[:10]))