#!/usr/bin/env python3

import sys, os.path as osp
from math import ceil

import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde

sys.path.insert(0, '../')
import utils


def plot_hist(*args, **kwargs):
    N = args[0].size

    plt.figure()
    hist, bins = np.histogram(np.hstack(args), density=True)
    width = 0.7 * (bins[1] - bins[0])
    center = (bins[:-1] + bins[1:]) / 2
    cumsum = np.cumsum(hist)
    plt.bar(center, hist, align='center', width=width)
    plt.plot(center, cumsum, c='r')
    title = kwargs['title'] if 'title' in kwargs else ''
    plt.xlabel('$\mathregular{x}$')
    plt.ylabel('$p(\mathregular{x})$')
    plt.title('%s (N=%d)' % (title, N))



def main():
    
    # parameters of the experiment
    noise_ratio =.15
    n_samples = int(8e4)    # number of samples per class
    d = 2                    # dimensionality
    mu0 = np.array([0., 0.])      # mean vector
    sigma0 = np.eye(d)         # covariance matrix
    pca_n_comp = None          # number of principal components to retain for PCA-based noise model
    use_ml = False             # whether ML or MAP should be used in estimating Gaussian noise
    
    # create a list containing noise models corresponding to different sources of variance in data
    # 1) true generative process with additive noise and known parameter values
    gmom_param_dict = {'mu': mu0[:,None],'sigma': sigma0}
    gprocess = lambda x: utils.generate_gprocess(x, gmom_param_dict)['y']
    noise_models = [utils.GaussianNoiseModel(variance=noise_ratio*sigma0)]
    # 2) real-world observation data: results from simulation experiments
    obsdatafile = osp.join('../obsdata/','simulated_data%d.json' % d)
    obsdata = utils.load_observation_data(obsdatafile, normalize=False)[..., :d]   # use only first d dimensions
    assert(d <= obsdata.shape[-1])
    obsdata = obsdata[:(n_samples//len(noise_models)),...]
    noise_models += [utils.RealWorldNoiseModel(obsdata[..., i]) for i in range(obsdata.shape[-1])]
    # 3) spurious signal injected into the real world observation data
    sigdatafile = osp.join('../sigdata/', '%ddataset_X_%d_%f.npy' % (d, n_samples//len(noise_models), noise_ratio))
    sigdata = np.load(sigdatafile).astype(float)
    if d < len(sigdata[0]):
        sigdata = sigdata[:, :d]
    noise_models += [-utils.RealWorldNoiseModel(sigdata[..., i]) for i in range(sigdata.shape[-1])]
    
    # compare noise models based on observed eigenvalues
    true_eigs = utils.get_true_eigenvalues(gmom_param_dict)
    print('Eigenvalues:\n\tTrue: %s\n\tML: %s\n\tMAP: %s' % tuple(['%.3f'%v for v in true_eigs]))
    evals = []
    kdes = []
    for i, model in enumerate(noise_models):
        
        # sample synthetic data using the current noise model
        synth_data = model.sample_observations(n_samples)
        X = np.atleast_2d(synth_data.flatten())
        
        # compute PCA-based noise model via SVD
        _, evecs, _ = utils.pca_transform(X, n_components=pca_n_comp, svd_solver='full')
        evals.append(abs(evecs @ np.diag(model.eigvals) @ evecs.T))
        kdes.append([])
        for j in range(len(evals)):
            kde = gaussian_kde(evals[j].flatten(), bw_method=.2)
            xx = np.linspace(-6, 6, num=int(ceil(.9*(1+4))))
            nn = [int((.5+(i+.5)*xx.size)/xx.size ) for i in range(xx.size)]
            kdes[i].append(np.asarray([kde.evaluate(xj)
                                        if not sum(nn[i:]*np.diff(kde.dataset)>xj)
                                          and abs(kde.integrate_box([-xj,-xj],[xj,xj])) > 1e-2
                                       else 0 for xj in xx]).reshape(tuple(nx for nx,xxl in zip(nxlist,xx.tolist()))))
    
    
    # plot results
    plt.figure(figsize=(15, 9))
    gridspec = {1: (1, 3), 2: (2, 3)}
    nxlist = [(1,),
              (1, 1),
              (1, 1, 1)]
    axlist = [plt.subplot2grid(gridspec[len(noise_models)], loc, colspan=colspan)
                  for loc, colspan in zip(((0, 0), (0, 1), (1, 0)), nxlist)]
    for i, ax in enumerate(axlist):
        im = ax.imshow(kdes[i][i],
                       origin='lower', extent=[-6, 6, -6, 6], aspect='auto', cmap=plt.cm.viridis)
        for j, kdej in enumerate(kdes[i]):
            if i!= j:
                ax.contour(xx, yy, np.maximum(0,kdej), colors=('C'+str(j),), levels=[1e-2], linewidths=(1,))
        ax.scatter([(true_eigs[i]-mu0[i])/sigma0[i] for i in range(d)],
                   [(true_eigs[i]+mu0[i])/sigma0[i] for i in range(d)], color='w', alpha=1.)
        ax.axis('equal')
        ax.set_title('Noise Model %d' % i)
        ax.set_xlabel('$\mathregular{\lambda_{%d}}$' % i)
        ax.set_ylabel('$\mathregular{\lambda_{%d}}^*$' % i)
    bar_ax = plt.subplot2grid(gridspec[len(noise_models)], (0, 2))
    cbar_ax = fig.add_axes([.91,.3,.03,.4])
    cb = plt.colorbar(im, cax=cbar_ax)
    cb.outline.set_linewidth(.5)
    cb.set_label('$p(\lambda|\\mathbf{x})$', rotation=-90, va="bottom")
    plt.tight_layout()
    
if