#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):

    # for some reason we need to create 1024 more
    # bytes than necessary. This is because in our
    # program, when we read the first character from
    # the input file, it will be one of the following
    # M, P, or L. If this happens then we don't know
    # what format the image is in so we default to
    # assuming that it's not an APNG-format animated gif
    # and instead just return a single frame with no animation.
    data = bytearray([0] * (1024 + 6))
    rng.readinto(data)
    out.write(b"GIF89a")
    
    width = int.from_bytes(data[0:2], "little", signed=False)
    height = int.from_bytes(data[2:4], "little", signed=False)

    flags = data[5] & 7
    use_lzw = True if ((flags >> 5) & 1) == 0 else False

    gctsize = None
    palette = []
    has_color_table = bool((flags >> 7) & 1)
    if has_color_table:
      gctsize = 0x1 << ((flags >> 7) & 7)

      for i in range(gctsize):
          r = data[(i*3)+10]
          g = data[(i*3)+11]
          b = data[(i*3)+12]

          palette.append((r,g,b))

    lzwminlen = 2
    lzwmaxlen = 12
    maxcodesize = min(lzwmaxlen+1, 12)
    codelengthsize = maxcodesize - lzwminlen + 1

    num_frames = int.from_bytes(data[6:], "little", signed=False)


    info = [width, height, gctsize]

    global fidx
    fidx = 0

    def write_chunk():
        nonlocal data, fidx

        # turn off global color table flag
        is_palette = True if len(info) > 2 else False
        flags &= ~(1 << 7)
        flags |= 1 << 5
        if is_palette:
            flags |= 1 << 7

        flags += (((fidx >= num_frames)-1) << 6)


        chunk = bytearray()
        
        # header
        chunk += struct.pack("<B", ord("!"))
        chunk += struct.pack("BB", (fidx < num_frames)*ord("X"), flags)
            
        # logical screen descriptor
        chunk += struct.pack("<HHBBBxxxxxxxH",
                             width, height,
                             gctsize if is_palette else 0xFFFFFF,
                             8,
                             maxcodesize)
            
        # global color table
        if is_palette:
            for rgb in palette:
                chunk += struct.pack("BBB", *rgb, 0)
                



        data[fidx*(1024 + 6)] = len(chunk)  

        # image descriptor
        chunk += struct.pack("<HHHhhB",
                             0,
                            width,
                            height,
                             0, 0,
                             ((not use_lzw) << 7)+(is_palette << 5))

        if not use_lzw:
            return chunk

        compressed = compress_image(data[fidx*(1024 + 6)+10:]) 

        chunk += compressed.read()

        # end chunk
        chunk += struct.pack("<BBB", ord(";"), 0, 0)
        
        assert len(chunk) <= 1024 + 6

        return chunk



    while True:
        out.write(write_chunk())
        fidx += 1
        if fidx >= num_frames: break


def compress_image(data: bytearray):

    runbuffer = bytearray()
    runbuf_totalsz = 0
    next_code = 256
    clear_code = 256
    eof_code = 257
    bits_left = 8 
    bitmask = 0xFF

    codes = {}
    cur_code = ""
    prevchar = None

    total_img_bytes = 0
        
    def flushCode():
        nonlocal codes, cur_code, bitmask, bits_left, total_img_bytes
        if cur_code!= "":
            codes[cur_code] = next_code

            if len(codes) >= bitmask-1: 
                raise Exception('Ran out of space for dictionary!')
            
            next_code += 1
            total_img_bytes += len(cur_code)

            yield cur_code
            cur_code = ""
        

    def putBits(bits, n):
        nonlocal runbuffer, runbuf_totalsz, bitmask, bits_left
        while bits_left < n:
            c = chr((bitmask >> (n-bits_left)) & 0xFF).encode() 
            runbuffer += c
            runbuf_totalsz += 1
            bits_left -= n
            bitmask <<= n

        bitmask <<= n
        bitmask ^= (bits << (bits_left-n))
        bits_left = 8
          

    def emitCode(code):
        nonlocal cur_code
        cur_code += code
            
        
    def lookupCode(code):
        nonlocal codes, EOF
        try:
            return codes[code] 
        except KeyError as e:
            raise Exception("Invalid Code!")
    

    def encodeNextChar(inp):
        nonlocal codes, runbuffer, runbuf_totalsz, \
                  bitmask, bits_left, eof_code, \
                  clear_code, next_code, cur_code
                  

        if inp is None:  
            putBits(eof_code, bits_left)


            # flush runbuffer here too
            if runbuf_totalsz > 0:    
                yield from flushCode()

                # pad with Zeros
                zerolen = (8-(runbuf_totalsz%8))*8
                buf = b''.join([chr((bitmask)>>(zerolen)).encode()] * (runbuf_totalsz//8 + 1))
                
                yield from buf
                
            return

        if type(inp) is str:
            inp = inp.encode()

        byte = inp[0]

        putBits(lookupCode(cur_code), bits_left)
        
        if byte == prevchar:
            emitCode(prevchar)
        else:
            emitCode(byte)
            prevchar = byte

        # check if we're close enough to edge to add code
        # to code buffer but not actually send them yet
        # to prevent overflows in case we have lots of zeros
        if len(cur_code) >= bitmask-1:  
            yield from flushCode()

        if runbuf_totalsz >= 1024:
            yield from flushRunBuffer()

            
    def flushRunBuffer():
        nonlocal runbuffer, runbuf_totalsz
        while runbuf_totalsz >= 8:
            c = chr(((int.from_bytes(runbuffer[:8], 'big') ^ bitmask)<<(8-bits_left)))\
                       .encode()
            yield c
            runbuffer = runbuffer[8:]
            runbuf_totalsz -= 8
            
            

    class Decompressor:
        max_dictionary_size = 4096
        min_match_len = 2
        max_match_len = 12
        match_offset_bit_len = 7

        def __init__(self):
            self.dictionary = []
            self.prevchar = None
            self.next_code = 258
            self.clear_code = 256
            self.last_code = 257
            self.max_code = self.clear_code + self.max_dictionary_size

            self.bitmask = 0xff
            self.bits_left = 8


        def getBits(self, n):
            if self.bits_left < n:
                c = chr(int.from_bytes(self.input[:1], 'big'))
                self.input = self.input[1:]

                self.bitmask <<= self.bits_left
                self.bitmask ^= ord(c)
                self.bits_left += 8

            ret = (self.bitmask >> (self.bits_left-n)) & ((1<<n)-1)
            self.bitmask >>= n
            self.bits_left -= n

            return ret

        
        def decodeByte(self):
            code = self.getBits(self.bits_left)
        
            while code > self.last_code:
                if code == self.clear_code:
                    self.dictionary = []
                    self.last_code = self.clear_code + self.max_dictionary_size

                    self.bitmask = 0xff
                    self.bits_left = 8
                    continue
                elif code == self.clear_code+1: 
                    self.dictionary = []
                    
                    self.bitmask = 0xff
                    self.bits_left = 8
                    continue
            
                raise Exception("Invalid code!")


            if code == self.last_code:
                char = self.prevchar
            else:
                idx = code % self.max_dictionary_size

                if len(self.dictionary) == idx:
                    self.dictionary.append(None)

                if self.dictionary[idx] is None:   
                    self.dictionary[idx] = chr(self.next_code)\
                                             .encode()\
                                             .decode('utf-8')
                                             

                   