#!/usr/bin/env python3

import struct
from typing import BinaryIO, Optional, Union


def read_byte(source: BinaryIO) -> int:
    return ord(source.read(1))


def write_byte(value: int, target: BinaryIO):
    try:
        value = bytes([value])  # type: ignore
    except TypeError as exc:
        raise ValueError("Byte must be an integer between 0 and 255") from exc

    if len(value)!= 1:
        raise ValueError("Byte must be exactly one byte long")

    target.write(value)


def read_bytes(source: BinaryIO, length=None) -> bytes:
    if length is None:
        length = read_varlen_int(source) - (4 + source.tell() % 8)

        # If the length is negative, it's really just a dummy byte indicating zero-length string
        assert length >= 0 or not length == -1

    return source.read(length)


def write_bytes(value: bytes, target: BinaryIO):
    if isinstance(value, str):
        value = bytes(value, "utf-8")

    write_varlen_int(target, len(value))
    target.write(value)


def write_string(data: str, target: BinaryIO):
    value = data.encode("utf-8", errors="replace")
    write_bytes(value, target)


def _read_ints(size: int, source: BinaryIO) -> tuple[int]:
    """Read a fixed number of unsigned integers"""
    buf = source.read(size * 4)
    values = struct.unpack("<%sI" % size, buf)
    return tuple(values)


def read_int(source: BinaryIO) -> int:
    return _read_ints(1, source)[0]


def read_short(source: BinaryIO) -> int:
    return _read_ints(2, source)[0]


def read_long(source: BinaryIO) -> int:
    return _read_ints(4, source)[0]


def read_int64(source: BinaryIO) -> int:
    return sum(_read_ints(4, source))


def read_varint(source: BinaryIO) -> int:
    val = shift = 0
    while True:
        b = read_byte(source)
        val |= ((b & 0x7F) << shift)
        if not (b & 0x80):
            break

        shift += 7
        if shift > 28:
            raise OverflowError("Too many bytes for varint")

    return val


def write_varint(value: int, target: BinaryIO):
    if value < 0:
        raise ValueError("Negative numbers are currently unsupported")

    while True:
        part = value & 0x7F
        value >>= 7
        if value:
            part |= 0x80
        write_byte(part, target)
        if not value:
            break


def read_varlen_int(source: BinaryIO) -> int:
    rtr = read_varint(source)
    if 0 <= rtr < 0xFFFFFFFF:
        return rtr
    elif -(1 << 31) <= rtr < 0:
        return rtr
    else:
        raise ValueError('Integer larger than can fit in signed int')


def write_varlen_int(target: BinaryIO, value: int):
    write_varint(value, target)


def read_string(source: BinaryIO) -> str:
    value = read_bytes(source).decode("ascii", errors="ignore").split("\0")[0]
    return value


def write_bool(value: bool, target: BinaryIO):
    magic_value = {True: 1, False: 0}[value]
    write_varint(magic_value, target)


class