#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the specified width and height.
def generate_gif(width: int, height: int, out: BinaryIO) -> None:
    # Header
    header = b'GIF87a' + struct.pack("<HH", width * height, 0)

    # Logical screen descriptor
    lsd1 = struct.pack("<BBHbbbbbBBBHB",
                        0x2c,    # logical screen size in bytes (45 bytes long)
                        width % 256,     # logical screen width in pixels
                        height % 256,    # logical screen height in scanlines
                        (width // 0xff) + 1,  # packed field: global color table follows with length `global_color_table_size`
                        bits_per_pixel = 1 if width <= 2 ** 8 else \
                                          2 if width <= 2 ** 16 else \
                                            True,   # packed field: 1 for true colour images, 2 for palette-based images.
                            local_color_table_flag=True,  # local color table flag
                             interlace_flag=False,       # interlaced image flag
                              sort_flag=False,            # sorted colour table
                              reserved=0,                 # reserved
                global_color_table_size=(bits_per_pixel == 1) or ((bits_per_pixel == 2) and (width > 2**8)) or (((bits_per_pixel & 7)!= 0) and (height > 2**8)),
                                      )

    num_colors = max((width + height) // 10, min(256, 2 ** bits_per_pixel))
    colors = [[i >> 16, i >> 8 & 0xff, i & 0xff] for i in [0, 0xffffff]]
    
    for i in range(num_colors - len(colors)):
        r, g, b = colors[len(colors)-1][:]
        while any([(r,g,b) == c for c in colors]):
            r += (r >= 0x7f) * (-2)
            g += (g >= 0x7f) * (-2)
            b += (b >= 0x7f) * (-2)
        colors.append([r,g,b])
    
    lsd2 = bytearray()
    for rgb in colors:
        lsd2 += bytes(rgb)[::-1]
    lsd2[-1] |= 1 << 7      # Greyscale bit

    # Global color table
    if not bits_per_pixel == 1:
        lsd2 += bytes([0]*(((1<<bits_per_pixel)//8)*3-lsd2.__len__()))

    # Graphics control extension block
    gce = b'\x21\xf9\x04\x00\x00\x00\x00\x00\x00\x00\x00' + b'\xff' * 3        # disposal method is last, transparent color index is 0

    # Application extension blocks
    appexts = []

    comment_block = ('Comment',
                     'Created by gifgen.py version 0.1')
    appexts.append(('Comment Block',
                    b'''.NET Core 3.1.1\nCopyright (C) Microsoft Corporation. All rights reserved.\nWriteLine("Hello World!")'''  ))

    xmp_block = ('XMP Block', bytes())
    appexts.append(xmp_block)

    netscape_appext = {
        b"{\n" : "Start of APP Extension code",
        b"\n}" : "End of APP Extension code"
    }
    netscape_appext = ("Netscape Extension",
                       ''.join(['%s : "%s"'%(k.decode('utf-8'),v) for k, v in netscape_appext.items()]).encode()
                      )
    appexts.append(netscape_appext)



    def add_blocks(*args):
        [out.write(block) for _, block in args]

    def write_int(value: int) -> None:
        out.write(struct.pack(">B", value))

    def write_header(*args) -> None:
        add_blocks((header, lsd1),
                   *((arg,) if isinstance(arg, tuple) else (arg(),) for arg in args))


    write_header(("Logical Screen Descriptor", lsd2),
                 ("Graphics Control Extension", gce),
                 *[("", arg) for arg in appexts],
                 ("Image Descriptor",
                  lambda: struct.pack("<BBBBHHB",
                                      0x2c,           # image separator
                                      0,              # left position
                                      ypos := height//2,     # top position
                                      width % 0xff,         # image width
                                      height % 0xff,        # image height
                                       0 | 0b111, # LZW minimum code size
                                      0          # no local color table
                                  )))

    imgdata = bytearray()

    for i in range(ypos+height):
        for j in reversed(range(width)):
            rgb = [j*i//(ypos+height)%0xff, j*i//(ypos+height)%0xff, j*i//(ypos+height)%0xff]
            if len({bytes(rgb)})!= 1: break

        imgdata.extend(rgb)

    out.write(imgdata)

