#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header bytes
    out.write('GIF89a'.encode())

    w = 127 + int.from_bytes(rng.read(4), byteorder='big') % (640 - 128)
    h = 127 + int.from_bytes(rng.read(4), byteorder='big') % (480 - 128)

    # Logical screen descriptor
    out.write((w).to_bytes(2, byteorder='little'))
    out.write((h).to_bytes(2, byteorder='little'))
    out.write('\x87\x00\x00'.encode())     # Global color table size of one bit

    out.write('\x00'.encode())            # Background Color Index
    out.write('\x00'.encode())            # Pixel aspect ratio is 1:1

    write_color_table_entry(out, '\x00\x00\xFF', True)   # White background

    write_image_descriptor(out, 0, 0, w, h)

    # Graphics control extension
    out.write('\x21'.encode())
    out.write('\xF9'.encode())           # Extension introducer
    out.write('\x04'.encode())           # Block size
    out.write(((2 << 5) | (1)).to_bytes(1, byteorder='little'))    # Packed fields
    out.write(struct.pack('<H', 1))      # Delay time
    out.write('\x00'.encode())           # Transparent color index
    out.write('\x00'.encode())           # Terminator

    # Image data
    for i in range(0, h * w // 5):
        out.write('\x00' if bool(int.from_bytes(rng.read(1), byteorder='big') & 1) else '\xFF')

    while len(out.getbuffer()) % 2!= 0:
        out.write('\x00'.encode())       # Pad to even length

    palette = b''
    for _ in range(1 << 1):        # Repeat palette twice because we only have two colors
        palette += b'\xFF\x00\xFF'    # Blue, transparent
        palette += b'\xFF\x00\x00'    # Red, transparent

    # Application extension
    out.write('\x21'.encode())
    out.write('\xFF'.encode())           # Extension introducer
    out.write('\x0B'.encode())           # Block size
    out.write(('NETSCAPE2.0').ljust(11).encode())         # Application identifier
    out.write('\x03'.encode())           # Sub-block size
    out.write('\x01'.encode())           # Loop count sub-block id
    out.write(b'\x00\x00')              # No loop

    out.write('\x3B'.encode())           # Trailer


# Writes an image descriptor block starting at offset (`x`, `y`) on the screen and with dimensions `width` and `height`.
def write_image_descriptor(out: BinaryIO, x: int, y: int, width: int, height: int):
    out.write('\x2C'.encode())       # Image separator
    out.write((x).to_bytes(2, byteorder='little'))
    out.write((y).to_bytes(2, byteorder='little'))
    out.write((width).to_bytes(2, byteorder='little'))
    out.write((height).to_bytes(2, byteorder='little'))
    out.write('\x00'.encode())       # Local color table size
    out.write('\x00'.encode())       # Interlace flag set to zero (no interlace)
    out.write('\x08'.encode())       # LZW compression algorithm used


# Writes global color table entry `rgb` followed by its transparency setting `transparent`.
def write_color_table_entry(out: BinaryIO, rgb: str, transparent: bool):
    out.write(rgb[0].encode())       # Red
    out.write(rgb[1].encode())       # Green
    out.write(rgb[2].encode())       # Blue
    out.write((0 if not transparent else 1).to_bytes(1, byteorder='little'))