#!/usr/bin/env python3

import sys
from pathlib import Path
from typing import BinaryIO

# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    width  = 640
    height = 320

    # https://en.wikipedia.org/wiki/LZW
    color_resolution = 1
    background       = (0x00,) * 3
    pixel_aspect     = 0
    frame_delay      = 5         # unit is centiseconds
    num_frames       = 1
    local_color_table = None
    block_size       = 0
    lzw              = False
    interlace        = False

    def get_bits():
        return ((yield from rng.read(1))[0]) & 0b111

    def get_chunk(n):
        while True:
            chunk = yield from rng.read(n)

            if len(chunk)!= n:
                raise Exception("unexpected EOF in RNG stream!")

            for c in chunk:
                yield c
            break

    id   = b'\x00\xFC\xFF\xC9'
    size = bytes(((width >> i) & 0xff, ))+bytes(((height >> i) & 0xff, ) for i in range(0, 7))
    flags = bytes((int(''.join(('1', '0')[(get_bits() >> i) & 1] for i in range(0, 8)), 2), ))

    data = bytearray()
    while len(data) < (width*height):
        code_size = min(max(2, int(get_bits())), 12)

        if not lzw and code_size > 8: continue
        elif lzw and code_size == max(2, int(get_bits())): continue

        # code
        chunk = get_chunk(code_size)
        codes = list()

        next_code = chunk.__next__().__next__()
        codes += [next_code]
        last_code = next_code

        # bitstream that represents the code
        bits = ''

        def write_bits():
            nonlocal bits, data
            data[len(data)-1] |= int(bits[:8], base=2) << len(bits[8:])
            del bits[:8]

        # encoded data
        chunk = get_chunk(1)
        while True:
            try: chunk_char = chunk.__next__()
            except StopIteration: return

            code = chunk_char | \
                   chunk.__next__() << 8 | \
                   chunk.__next__() << 16

            if code >= pow(2, code_size)*12 + 2:
                raise ValueError(
                    f"invalid LZW code ({code}) encountered!"
                )

            if code == 256:
                # clear code
                codes = [i for i in range(pow(2, code_size))]

                write_bits()

            elif code == 257:
                # end of information code
                break

            else:
                # given a sequence of previously encountered symbols,
                # construct a new symbol by fetching the first character in the sequence,
                # then appending each subsequent character in order to the new string.

                if code <= last_code:
                    result = codes[-1-codes[0]]

                    for i in reversed(range(1, len(codes)+1)):
                        if codes[i] == code or (i < len(codes) and codes[i] == -1):
                            result = codes[i-1] ^ result

                            del codes[i:]
                            break

                    assert isinstance(result, int)
                    result %= 2**code_size

                    codes += [-1]*(last_code-len(codes)+1)
                    codes[0] = result

                else:
                    assert code <= pow(2, code_size)-1
                    codes += [code]

                write_bits()

            last_code = code

    # assemble all metadata into an overall buffer
    buf = bytearray([
        0x21, 0xF9, 0x04, *(frame_delay,),            # graphics control extension
        0x00,                                           # no transparancy
        0x2C, 0x00, 0x00,                               # left, top position
        0x00, 0x00,                                      # image dimensions
        *(flags,),                                       # packed field
        *((len(data) >> i) & 0xff for i in range(0, 16)) # size in bytes
    ])

    # finally append our encoded data to the existing buffer
    buf += data

    # send the final product to stdout
    out.write(buf)

    out.write(header + data)
    print(f"generated {len(data)}-byte image")
