#!/usr/bin/env python3
"""
    Generate a GIF file with a random payload from /dev/urandom

    This is useful for testing that things can parse arbitrary GIF files in
    order to mitigate against [CWE-20](https://cwe.mitre.org/data/definitions/20.html).
    
    The generated bytes are output as hex strings. This makes it easy to paste them into C++ code or other languages which have poor support for binary literals (such as Python).
"""

import binascii
import os
import sys


# random number generator `rng` (/dev/urandom)
def make_gif(rng: BinaryIO, width=16, height=16, depth=4, savepath="out.gif"):
    # image descriptor block
    idb = f"{width}x{height},{depth}".encode('utf-8')
    idblen = len(idb).to_bytes(length=1, byteorder='little', signed=False) 

    # physical screen descriptor block
    psd = ((width * height * depth) // 8).to_bytes(length=2, byteorder='little', signed=False)

    # logical screen descriptor block
    lsd = bytes([width % 256]) + \
          bytes([(width >> 8)]) + \
          bytes([height % 256]) + \
          bytes([(height >> 8)]) + \
          b'\xf7\xa0' # Bits per pixel; RGBA32!

    # global color table
    gct = b'\x00' * (depth == 4 and 1 << 24 or
                      depth == 8 and 1 << 16 or
                      0)

    gfxctrl = b'' # no animation
    appext = b''   # no application extension blocks
    trailer = b''  # no trailing bytes

    chunks = [idblen, idb,
              psd, lsd,
              appext, gfxctrl]
    if gct:
        chunks += [gct]
    chunks += [trailer]

    with open(savepath, 'wb') as outfile:
        outfile.write(b''.join([chunk for chunk in chunks]))

        frame_cnt = 0
        while True: 
            # write the next frame
            frame_cnt += 1

            # decide how many bytes of data to read
            if frame_cnt < 3:
                numBytesToRead = 3
            else: 
                numBytesToRead = min((frame_cnt - 2) * 3, 255)
            assert numBytesToRead >= 3
            
            frameData = rng.read(numBytesToRead)
            if not frameData:
                break
        
            # now do a little bit of tricky encoding to get the length right...
            dataLengthByte = (numBytesToRead & 0xff).to_bytes(
                    length=1, byteorder='little', signed=False)
            outfile.write(dataLengthByte + frameData)
        
        assert frame_cnt > 0
        print(f"generated {frame_cnt}-frame image ({hex(frame_cnt)}) at '{savepath}'")

