#!/usr/bin/env python3

import gzip
import io
import multiprocessing as mp
import os
import shutil
import subprocess
import tempfile


THISDIR=os.path.dirname(__file__)

class RandomGifGenerator:
    def __init__(self):
        self._tempdir = None
    
    @property
    def tempdir(self):
        if not self._tempdir:
            self._tempdir = tempfile.TemporaryDirectory()

        return self._tempdir.name

    def generate_random_gif(self, seed:int):
        callargs=[
                "python", f"{THISDIR}/generator.py", 
                "--seed", str(seed),
                "-o", f"{self.tempdir}/test-{seed}.gif"
            ]
        
        print(" ".join(callargs))
        popenresult = subprocess.run(callargs)

        assert popenresult.returncode == 0

        return f"test-{seed}.gif"

    def clean_up(self):
        if self._tempdir:
            self._tempdir.cleanup()

if __name__=='__main__':
    n_workers = 16
    queue = mp.Queue()
    manager = mp.Manager()
    files = manager.dict()


    gen = RandomGifGenerator()
    try:
        with gzip.open('seeds.txt.gz', mode='rt') as seedsfp:

            workers = []
            
            for _ in range(n_workers):

                worker = mp.Process(target=worker_function,
                                    args=(gen,))
                worker.start()
                workers.append(worker)


            while True:
                
                line = seedsfp.readline().strip()

                if not line:
                    break

                seed = int(line)
                filename = gen.generate_random_gif(seed)

                queue.put((seed, filename))

        while any([ w.is_alive() for w in workers ]):

            newfiles = { queue.get() for _ in workers }
            files.update({ fn : None for (_,fn) in newfiles })


    finally:
        gen.clean_up()
    

    if len(files)==0:
        exit(1)

    # Create an index page listing all images
    with open("__index.html", 'w+') as fp:
        for fn in sorted(files.keys()): 
            fp.write(f"<img src=\"{fn}\" />\n")
        