#!/usr/bin/env python3
"""
Given a binary stream of randomness (e.g., /dev/urandom), this script generates
a valid gif file from it and writes that to stdout.
The output is intended for texting purposes only as it contains unusual preamble
data which could potentially be detected by malware scanning applications.
"""


import sys
import os
import argparse
import io
from PIL import Image



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, 
        outstream: BinaryIO,
        scale: bool=False,
        minscale: float=1.0, maxscale:float=1.0):
    size = (400, 400)
    blackpixel = (0, 0, 0)
    whitepixel = (255, 255, 255)
    redpixel = (0xff, 0x00, 0x00)

    img = Image.new("RGB",size,"black")

    screen = {}
    screen["width"] = int.from_bytes(
            rng.read(2), byteorder='little')
    screen["height"] = int.from_bytes(
            rng.read(2), byteorder='little')
    if not scale:
        screen["width"] = 400
        screen["height"] = 400
    else:
        screen["width"],screen["height"] = \
            tuple([x*y for x,y in zip((screen["width"],
                                      screen["height"]),
                                     (minscale,maxscale))])

        tmpimg = img.resize((screen["width"],
                             screen["height"]))

        img = Image.new("RGB",
                         (screen["width"],
                          screen["height"]),"white")
        img.paste(tmpimg,(0,0),mask=tmpimg)


    numcolors = int.from_bytes(
            rng.read(1), byteorder='little') * 256 +\
                 int.from_bytes(
                     rng.read(1), byteorder='little')

    colorsused = [redpixel]
    colorsunused = []

    for _ in range(numcolors - 1):
        color = None
        while True:
            rgb = [(ord(b) >> k) % len(colorsused)\
                   for k in (0,8,16)]

            color = tuple(rgb)
            if color == blackpixel or\
               color == whitepixel or\
               color in colorsused:
                continue

            break
        
        colorsused.append(color)

    # With all colors generated, we can now start generating pixels
    pixelbuf = []
    count = 0
    for y in reversed(range(screen["height"])):
        row = []
        for x in range(screen["width"]):
            color = colorsused[count%len(colorsused)]
            
            rgb = [(c >> k)%256 for c,k in\
                   zip(color,
                       (8,16))]

            row += rgb + [0]

            count+=1

        pixelbuf += row



    def add_subblock(
            data: bytearray, 
            subtypecode: str):
        blocksizestr = ""
        blocksize = len(data).to_bytes(
            2, "little")
        assert len(subtypecode)==1,\
               f"{subtypecode}"
        subtypecodestr = bytes([ord(c) for c\
                                in subtypecode])

        for item in [blocksize, subtypecodestr]:
            blocksizestr+="".join(["{:x}".format(i)\
                                   for i in item])
        
        sumvalue = sum([ord(i)&0xff for i\
                        in blocksizestr]).to_bytes(2,
                                                      'big')
        outstream.write(
                bytes([(sum([ord(v)&0xff for v in\
                            blocksizestr])&0xff)+1]))
        outstream.write(blocksizestr+b"\x00"+sumvalue)
        outstream.write(data)
        
    add_subblock(
            bytes([screen["width"], 
                  screen["height"]]),
            "wh")
    
    add_subblock(
            bytes([0xf7, 0x00]),
            "ce")

    
    add_subblock(
            bytes([0x00, 0x3B]),
            "ap")

    add_subblock(
            bytes([]),
            "in")

    add_subblock(
            bytes([0x08, 0x00]),
            "pa")
    
    add_subblock(
            bytes(pixelbuf),"fd")

    outstream.seek(0)




    