#!/usr/bin/env python3

import struct
from typing import BinaryIO


def make_random_block(rng: BinaryIO, size: int)-> bytes:
    blockdata = []
    while len("".join(hex(i)[2:] for i in blockdata)) < size*2:
        blockdata += [int.from_bytes(rng.read(size),"big")]
    return "".join(hex(i)[2:] for i in blockdata[:size])


def make_color_table(rng: BinaryIO):
    blocks = []
    clim = 2**(rng.read(1)[0] & 7)+1
    for i in range(clim**2):
        rgb = tuple([i%clim]*3) + (0,)
        blocks += list(rgb)
    return bytes(blocks)


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b"GIF89a"
    global_flags = int.from_bytes(rng.read(1),"big")
    print("global flags", hex(global_flags))

    if global_flags&0x80:
        gct_flag=True
        print("GCT flag set:",gct_flag)
        color_table = make_color_table(rng)
        print("Color table:",color_table)
    else:
        gct_flag=False


    # Write header
    out.write(header)


    # Write Logical Screen Descriptor
    w,h = [(int.from_bytes(rng.read(2),"big") & 0xfff)]*2
    pw,ph = [pow(2,i+1)&-2 for i in range(2)]
    background_color = int.from_bytes(rng.read(1),"big")
    pixel_aspectratio = int.from_bytes(rng.read(1),"big")
    print("Logical screen descriptor : {}".format((
            w, h,
            pw-(w>>1), ph-(h>>1),
            0xf&(background_color<<4 | (pixel_aspectratio)))
    ))
    out.write((w).to_bytes(2,"big"))
    out.write((h).to_bytes(2,"big"))
    out.write(((pw)-(w//2)).to_bytes(2,"big"))
    out.write(((ph)-(h//2)).to_bytes(2,"big"))
    out.write((background_color<<4 | pixel_aspectratio).to_bytes(1,"big"))

    # Write global color table, if any
    if gct_flag:
        out.write(len(color_table).to_bytes(1,"big"))
        out.write(color_table)

    # Write graphics control extension
    has_control = bool(global_flags&0x01)
    disposal_method = 0xe&(int.from_bytes(rng.read(1),"big")>>2)
    user_input_flag = bool(int.from_bytes(rng.read(1),"big")&0x02)
    transparent_color_flag = bool(int.from_bytes(rng.read(1),"big")&0x01)
    delay_time = int.from_bytes(make_random_block(rng,2),"big")

    print(("Graphics Control Extension: {}".format(
        (has_control,disposal_method,user_input_flag,transparent_color_flag,
         delay_time))))
    out.write((0x21).to_bytes(1,"big"))
    out.write((0xf9).to_bytes(1,"big"))
    out.write((has_control << 7) |
              (delay_time&0xff00)>>8 |
              (disposal_method)<<2 |
              (transparent_color_flag)<<1 |
              0x01)
    out.write(delay_time.to_bytes(2,"big"))
    out.write((0).to_bytes(1,"big"))

    if transparent_color_flag == True:
        out.write(list(range(1))[0].to_bytes(1,"big"))
    else:
        out.write((255).to_bytes(1,"big"))

    out.write((0x00).to_bytes(1,"big"))

    # Write image descriptor
    lzwmincodelength = pow(2,int.from_bytes(rng.read(1),"big")+1)
    x,y = [(int.from_bytes(rng.read(2),"big") & -2)]*2

    print("""Image descriptor: {}""".format((lzwmincodelength,
                                          x+(w>>1), y+(h>>1),
                                          0,0)))
    out.write((0x2c).to_bytes(1,"big"))
    out.write((lzwmincodelength).to_bytes(1,"big"))
    out.write((x+(w>>1)).to_bytes(2,"big"))
    out.write((y+(h>>1)).to_bytes(2,"big"))
    out.write((0).to_bytes(1,"big"))
    out.write((0).to_bytes(1,"big"))

    # Write compressed data stream
    lzwdict = {b"\x00":b"",b"\xff\xff":b""}
    curbyte = None
    for i in range(2):
        curbyte = int.from_bytes(rng.read(1),"big")
        if not gct_flag or curbyte!= 0xff: break
    while curbyte!= 0x00 and curbyte!= 0xff:
        if curbyte >= lzwmincodelength and \
           curbyte < lzwmincodelength + len(lzwdict.keys()):
            if lzwdict[curbyte]:
                out.write(lzwdict[curbyte].to_bytes(1,"big"))
                nextbyte = lzwdict[-1][:-1]+lzwdict[curbyte][-1:]
                if nextbyte[0]!= "\x00":
                    if len(lzwdict.keys()) <= 255:
                        lzwdict[len(lzwdict.keys())+lzwmincodelength]=nextbyte
                    elif len(lzwdict.keys()) > 255:
                        del lzwdict[max(lzwdict.keys())]
                        lzwdict[len(lzwdict.keys())+lzwmincodelength]=nextbyte
                if curbyte == 255:
                    curbyte = int.from_bytes(rng.read(1),"big")
                continue

        lzwdict[len(lzwdict.keys())+lzwmincodelength]=(
                chr(curbyte%256)+lzwdict[curbyte//256])
        if len(lzwdict.keys()) <= 255:
            lzwdict[len(lzwdict.keys())+lzwmincodelength]=(
                    chr(curbyte%256)+lzwdict[curbyte//256])
        elif len(lzwdict.keys()) > 255:
            del lzwdict[max(lzwdict.keys())]
            lzwdict[len(lzwdict.keys())+lzwmincodelength]=(
                    chr(curbyte%256)+lzwdict[curbyte//256])


        if len(lzwdict.keys()) <= 255:
            out.write(curbyte.to_bytes(1,"big"))
        elif len(lzwdict.keys()) > 255:
            try:
                out.write((-1).to_bytes(1,"big"))
            except OverflowError as e:
                raise Exception("Too much LZW dictionary content!") from e



        nextbyte = lzwdict[-1][:-1]+chr(curbyte%256)

        if nextbyte[0]!= "\x00":
            if len(lzwdict.keys()) <= 255:
                lzwdict[len(lzwdict.keys())+lzwmincodelength]=nextbyte
            else:
                del lzwdict[max(lzwdict.keys())]
                lzwdict[len(lzwdict.keys())+lzwmincodelength]=nextbyte

        if curbyte == 255:
            curbyte = int.from_bytes(rng.read(1),"big")



    # Write data subblock with no subblocks
    out.write((0x00).to_bytes(1,"big"))


