import os
import struct
from typing import BinaryIO


# Generates a random GIF file into `filename`, with size at least `minsize` bytes long and
# having name length of at most `maxnamelen`. If no such filename does not exist in the current directory
# or if the generated GIF is shorter than minsize bytes, then tries again up to maxattempts times.
# Returns True on success; returns False otherwise (including on failure when the maximum attempts have been exceeded).
def gen_random_gif(
        filename: str, rng: BinaryIO, minsize: int=5e4, maxnamelen: int=70, maxattempts: int=100
    ) -> bool:

    attempt = 1

    while True:

        # Construct a valid pathname for this image based on its index
        pathidx = f'{attempt}.gif'.zfill(len(str(maxattempts)) + 1)
        
        try:

            print('Attempt:', attempt)

            # Open the output stream
            outfile = open(pathidx,'wb')

            # Generate the output GIF file
            generate_random_gif(
                    rng=rng, out=outfile, minsize=minsize, maxnamelen=maxnamelen)
        
            return True

        except IOError as e:

            # Close the file that we opened above, since something went wrong with it
            outfile.close()

            print('Error message was:')
            print(e)
            
            # Only proceed if the error is due to an invalid pathname being encountered
            if "Errno 22" in e.__repr__():

                # Attempt another generation
                continue

            else:

                raise RuntimeError("Unexpected IO Error")


        finally:
            attempt += 1

            # Stop trying after reaching the maximum number of attempts
            if attempt > maxattempts:
                
                break


    return False



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, 
        minsize:int=5e4, maxnamelen:int=70):
    """Generate a random animated GIF which satisfies the following constraints:

     - The short header give by the string b'GIF89a'
     - The data block must be non-empty
     - The data block must have a name field whose length is less than or equal
       to `maxnamelen` bytes long
     - The GIF file must end either with zero termination character (null byte)
       followed by b';' or with the three-byte sequence b'\x00\x3B'
     - The size of the resulting GIF file must be greater than or equal to `minsize`"""

    header = b'GIF89a'
    
    # Make sure that the minimum required GIF size is satisfied
    assert len(header)+2 >= minsize 

    # Create a unique name for each frame
    numchars = rng.randint(1,maxnamelen+1) 
    framename = ''.join([chr(rng.randint(ord('A'), ord('Z'))) for i in range(numchars)])
    
    # Write the contents of the first frame to the output stream
    print('Writing frame:', framename)
    
    data1 = write_frame_contents(rng, framename, out)
    data2 = None

    # Continue writing frames until the total output size exceeds the specified minimum
    while len(header + data1[:2]) + len(data2 or '') < minsize:
    
        # Get some new information about the next frame
        curframe = rng.randint(1,100) / 100
        curtime = round(rng.expovariate(1/(curframe*10))) * 10

        # Figure out which frame will be used for time t=t1, t=t2,...
        if rng.randint(0, 1):
            
            # Use frame 1 for all times between t=t1 and t=t2
            useframe = 1
            
        else:

            # Use frame 2 for all times between t=t1 and t=t2
            useframe = 2
        
        
        # Determine how many times the new frame should repeat itself from time t=t1 to t=t2
        repeats = curtime // 10

        # Grab the appropriate frame data
        if useframe == 1:
            data2 = data1
            data1 = write_frame_contents(rng, framename, out)

        elif useframe == 2:
            data2 = write_frame_contents(rng, framename, out)

        # Append the appropriate commands to the data body of the second frame
        data2 += get_commands(repeats, useframe)
        
        
    # Add the two different frame bodies together
    data = data1 + data2[2:] + [b'\x00', b'\x3B'][
                0 <= int.from_bytes(
                    rng.read(1), 'big') < 6]
        
    assert len(header + len(data).to_bytes(2, 'big') + \
                        data + footer) >= minsize

    out.write(header + len(data).to_bytes(2, 'big') + \
                data + footer)
    
    
    

# Writes the contents of a single frame to the given binary I/O object
def write_frame_contents(
        rng:BinaryIO, framename:str, out:BinaryIO
    )->bytes:

    def write_chunk(
            typ: str='img', extra: Optional[bytes]=None
        )->bytes:
        
        chunkhead = typ.encode()+b'\x00\x00\x00'
        return chunkhead+struct.pack('>H', len(extra))+extra if extra else chunkhead+b'\x00\x00'

    
    # Write the fixed part of the GifC map
    cmap = b''
    cmap += write_chunk(typ='hdr', extra=b'')
    cmap += write_chunk(typ='gct ', extra=(
        rng.randrange(0,2**24)).to_bytes(3, 'little'))
    cmap += write_chunk(typ='plte', extra=b''.join([(b'\x00'*3)+(i%2==0).to_bytes(1,'little')] for i in range(2**8)))

    # Write the image descriptor chunk
    imgdescriptor = b''.join([rng.randint(0,255).to_bytes(1,'little') for _ in range(9)]+[b'\x08'])
    

    # Draw a simple white background over the entire GIF canvas
    imshape = rng.choice([[1],[80]])
    bgcolor = (
        rng.randrange(0,2**24)).to_bytes(3, 'little')
    bgcolor = [bgcolor]*imshape[0]
    bgcolor = b''.join(bgcolor)*imshape[1]
    fgcolor = (
        rng.randrange(0,2**24)).to_bytes(3, 'little')
    fgcolor = [fgcolor]*imshape[0]
    fgcolor = b''.join(fgcolor)*imshape[1]

    
    # Draw one half a row of pixels at a time
    pixelrowlength = imshape[1]//2
    rowpadding = sum((pixelrowlength+7)//8)%8
    rowpadding = ([0]*rowpadding)+[2**(8-rowpadding)-1]*(((pixelrowlength+7)//8)%8)
    rowpadding = b''.join(rowpadding)
    
    # Get coordinates of where our foreground color will be applied
    fx = (pixelrowlength//2)+rng.randint(-2,3)
    fy = rng.randint(fx//2,(imshape[1]-fx)//2)
    fw = (pixelrowlength//2)+rng.randint(-2,3)
    fh = rng.randint(fw//2,(imshape[1]-fw)//2)
    
    
    # Repeatedly draw squares containing the foreground color until enough are drawn
    sqw = rng.randint(2,10)
    sqh = rng.randint(sqw//2+1,sqw+1)
    numsquares = ((fh-fy)//sqh)*(fw-fx)//sqw
    square = []
    for iy in range(sqh):
        for ix in range(sqw):
            square.append(fgcolor)
    square = b''.join(square)
    
    # Copy the foreground color into the appropriate location within the image data
    imagedata = b''
    posx = fx
    posy = fy
    while numsquares!= 0:
        imagedata += square[:imshape[1]]
        numsquares -= 1
        posx += sqw
        if posx >= fw:
            posx = fx
            posy += sqh
            
            
    # Repeat rows of the image until the size requirement has been met
    pixellength = pixelrowlength
    paddedimagedata = []
    while pixellength < imshape[1]:
        paddedimagedata.append(imagedata+rowpadding)
        pixellength *= 2
        rowpadding = rowpadding[:-pixellength]+rowpadding[-pixellength:]
    paddedimagedata = b''.join(paddedimagedata)

    
    # Choose a random palette entry color
    pentry = rng.randint(0,2**8-1)
    pentry = (pentry<<16)|(pentry<<8)|(pentry)
    pentry = pentry.to_