#!/usr/bin/env python3

import sys
from pathlib import Path
from typing import BinaryIO


MAX_IMG_LENGTH  = 1024 * 1024 # bytes
MAX_IMG_WIDTH   = 640 // 8     # pixels / byte
MAX_IMG_HEIGHT  = 350          # pixels per line
MIN_IMG_LENGTH  = (MIN_IMG_WIDTH * MIN_IMG_HEIGHT) // 8

def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    def get_img_width() -> int:

        width_raw = rng.read(1)[0]

        return min(
                MAX_IMG_WIDTH,
                2 << (
                    ((width_raw >> 7) & 1) |           # giant block
                    (((width_raw >> 5) & 1) ^ 1))       # small blocks
                )


    img_width      = get_img_width()
    img_height     = max(
            1,
            round((MAX_IMG_LENGTH - MIN_IMG_LENGTH) // img_width)) # never zero height images
    frame_duration = rng.read(3)[::-1]

    # HEADER
    out.write(header)                  # GIF signature
    out.write(b'\x00\x21\xF9\x04')    # screen descriptor (logical screen size is not provided)
    out.write(frame_duration)          # duration of a single frame in milliseconds
                                        # NOTE: this isn't actually used by any reader as it's supposed to be determined from the LZW compression method and we don't use that!
    bg_color_index = 0                 # background color index; we're using white so let's keep that simple by always setting it to '0'
    aspect_ratio   = b'\xE0\x00'        # pixel aspect ratio; default "magic" value for "square pixels"
    out.write(bg_color_index.to_bytes(1, byteorder="little"))
    out.write(aspect_ratio)

    # GRAPHICS CONTROL EXTENSION BLOCK
    out.write(b'\x21\xF9\x04')         # graphics control label
    out.write(b'\x01')                 # block size
    out.write(b'\x00\x06')             # delay between frames
    out.write(b'\xFF')                 # no transparancy
    out.write(b'\x00')                 # terminator

    # IMAGE DESCRIPTOR AND LOCAL COLOR TABLE
    lct_size  = 7                     # local color table size
    lct_sorted= False                 # sorted flag
    lct_repeat= True                  # repeat flag
    out.write(b'\x2C\x00\x00\x00')     # image descriptor
    out.write(lct_size.to_bytes(1, byteorder="little"))
    out.write(b'\x00'*lct_size*3)      # RGB color palette
    out.write(lct_sorted.to_bytes(1, byteorder="little"))
    out.write(lct_repeat.to_bytes(1, byteorder="little"))

    def write_line():
        line_length = img_width // 8          # how many bytes do we need for one line?
        scanlinesize= line_length + 1       # scaneline blocks have an extra "end of line" byte; even if there are none
        interlaceflag= 0                    # only non-interlaced files are supported
        out.write(scanlinesize.to_bytes(1, byteorder="little"))
        out.write(interlaceflag.to_bytes(1, byteorder="little"))

        # RANDOM DATA BOUNDED BY `img_width`
        cur_line  : bytearray = bytearray()
        num_bytes : int = 0
        while num_bytes < img_width or len(cur_line) > 0:

            # fill up current line until its full
            while not(num_bytes < img_width and len(cur_line) >= line_length):
                next_byte = rng.read(1)
                cur_line += next_byte

                # break if we reached end of stream
                if len(next_byte)==0:
                    break

            # write complete lines
            while len(cur_line) == line_length:
                eol = 0x00 if len(cur_line) == line_length else 0xff
                out.write(cur_line[:line_length])
                out.write(eol.to_bytes(1, byteorder="little"))
                del cur_line[0:line_length]
            
            num_bytes += 1

    # WRITE ALL THE IMAGES
    for _ in range(img_height):
        write_line()

    # TERMINATOR
    out.write(b'\x3B')
    print(f"""generated a {img_width}x{img_height}px GIF""")