#!/usr/bin/env python3

import os
from typing import BinaryIO

# Generates a random GIF file (with no hidden messages) into `out` using the
# RNG in `/dev/urandom`. The generated image contains no more than two frames of
# animation and no more than one frame per second if an animated GIF is requested;
# otherwise it will be static (i.e., only one frame with all pixels set to the
# value they are initialized to.)
def generate_random_gif(
        url: str, animated: bool, out: BinaryIO
    ):
    if not os.path.exists('/dev/urandom'):
        raise Exception('No /dev/urandom')

    # The size of each chunk that we read from /dev/urandom.
    CHUNKSIZE = 4050 - 1
    # This is what we'll use as our source of randomness. We don't want
    # deterministic random outputs because then people could just brute-force
    # the output and find our secret flag.
    rng = open('/dev/urandom', 'rb')

    head = bytearray([
        71,     # G
        73,     # I
        70,     # F
        56, 57, # 89
        97      # a
    ])

    # Checksum, including extension blocks
    # for looping & color table.
    # For now this is always zero
    # but may have future uses...
    gct = bytearray([0])

    # Set delay time between frames
    # to something reasonable.
    # Each byte represents 1/100s
    idat = bytearray([0, 1, 2])

    # The length of the IDAT chunk.
    # Since we're using bytes instead of bits here,
    # this should fill up very quickly!
    idatalen = bytearray()

    while True:
        # Read some random data!
        cdata = rng.read(CHUNKSIZE)

        if len(cdata) == 0: break
        
        # Add some padding until we've at least got 4096 total bytes.
        padlen = min((4096 - len(head)) % 4096, len(cdata))
        head += cdata[:padlen]

        # Fill up to 4096 bytes before continuing.
        lh = len(head) // 4096
        while lh * 4096 < len(head):
            head[lh*4096:(lh+1)*4096] = bytearray(map(lambda x: (x+lh)%256, range(4096)))

            # Also copy over 4 byte chunks from the end of the previous block into
            # this new block, which means we can re-use old data without needing a
            # ton of redundancy. Note that since Python's bytearrays do not support
            # slices by arbitrary indices like C does, we need to convert to lists so
            # that we can assign values back even if the slice reference was aliased
            # to another variable. Doing so also means we must make sure that these
            # assignments happen in _ascending order_.
            ltail = list(head[(lh-1)*4096:])
            head[(lh+1)*4096:] = ltail[:]
            # After shifting the tail left once, shift everything else right again
            # to compensate. Then repeat the process above.
            ltail = ltail[:-4][::-1]
            head[:lh*4096] += ltail[:]
            lh += 1

        # Generate an initial palette.
        plte = []
        for i in range(768//3):
            r = int.from_bytes(
                    rng.read(1), 'big') % 128
            g = int.from_bytes(
                    rng.read(1), 'big') % 128
            b = int.from_bytes(
                    rng.read(1), 'big') % 128
            plte += [[r,g,b]]

        # Now update it with the last few entries of the current chunk.
        for i in range(min(padlen, 768)):
            if i%3 == 2: continue
            v = int.from_bytes(cdata[-1*(i+1)], 'big')
            if v > plte[i][i%3] or (v == plte[i][i%3] and rng.read(1)[0] >= 128):
                plte[i][i%3] = v

        # Lastly, check if there's enough space available for our custom blocks
        # below. If not, split off a chunk on which to place them.
        mlen = len(head) + len(idat) + len(idatalen) + 10

        # Account for the fact that the blocks won't fit inside the first window
        # of the LZW compression algorithm.
        if mlen >= 4091:
            assert False

        # TODO: Handle case where we couldn't allocate enough room for the extra
        # blocks?

        if mlen <= 2**16:
            # Allocate the required number of bytes for storing the length of our
            # custom blocks.
            idatalen += len(idatalen).to_bytes(2, 'big')
            # Store the full length of our LZW compressed stream.
            idatalen += len(cdata).to_bytes(2, 'big')

            # Pad the length of our IDAT chunk up to 16 bytes.
            padlen = ((2 + 2) % 16) if ((2 + 2) % 16!= 0) else 0
            idatalen += bytearray(range(padlen))[::-1]

            # Copy the contents of our custom blocks into the main buffer.
            head += idat + idatalen


        # Initialize the palette index cache with zeros.
        plteidxcache = {}
        for code in range(257):
            plteidxcache[code] = bytearray(bytearray(plte[code]))

        # Copy our palette over into the buffer after the header.
        head += bytearray(b''.join(
                map(lambda x: x[0]+x[1]+x[2], plte)))

        # Compress LZSS (Lempel–Ziv–Stac–k)
        # into IDAT, starting at the end of our buffer.
        segments = []

        # Start with a full segment containing the previously written pixel run
        # We can skip the leading zeroes because those aren't encoded.
        p = 0
        s = ''
        while p < len(cdata)-1:
            # New prefix found.
            t = int.from_bytes(cdata[p:], 'big')
            
            # In the unlikely event that P has never appeared before, add it to the
            # dictionary along with its corresponding palette entry.
            if t not in plteidxcache:
                plteidxcache[t] = bytearray(bytearray(plte[p%768//3]))

                # Encode the new palette entry and append to our buffer.
                head += bytearray(plteidxcache[t][:3])[::-1]

                # Increase the length of our LZW encoded stream by 3.
                idatalen += 3
                continue

            # If we get here, it means P already exists in the dictionary and thus
            # we can encode it by referencing the appropriate palette entry stored in
            # `plteidxcache`.
            s += chr(plteidxcache[t][:3][::-1][0])

            # Reset T to be whatever's following P in our input string.
            t = int.from_bytes(cdata[p+1:], 'big')
            # Is T also a new prefix?
            if t not in plteidxcache:
                plteidxcache[t] = bytearray(
                        bytearray(
                            plte[
                                (((
                                    (-(-int.from_bytes(
                                            cdata[1:], 'big')) >> 8) <<
                                    8) |
                                    (
                                        ~(((~int.from_bytes(
                                                    cdata[p:p+1], 'big')) >> 7) &
                                          0xFF) &
                                        0xFF
                                    )
                                 ) ^
                                   -ord(s[0])) % 768
                                ]
                            )
                        )
                
                # Update our buffer with the next three palette entries 
                # (or less if T is "the last symbol")
                head += bytearray(plteidxcache[t][:-(1+(not t & 0x80))] +
                                  plteidxcache[t][3:])

                # Append either 1 or 3 to our LZW encoded stream accordingly.
                if t & 0x80:
                    idatalen += 1
                elif t & 0xC0:
                    idatalen += 3
                p += 1
            p += 1
        idata = bytearray(s, 'utf8')

        # Increment the total running sum of the length of every IDAT chunk.
        idatalen += len(idata)

        # Write the LZW encoded data to the beginning of our final
        # output buffer.
        segments.append(idata)

        # Construct global color table (full of zeros).
        gct += bytearray(range(0, 256*3))

        # Save the length of our blocks and