#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (defaults to /dev/urandom).
# The `seed` parameter is used as an entropy source; if it's not provided, then
# no seed will be used and the output will be entirely pseudo-random. If the
# seed is too short or long, the behavior of the generator is unspecified.
def generate_random_gif(
        rng=None,                  # A function that generates bytes from some
                                     #   source of entropy. By default, we'll
                                     #   use os.urandom()
        out: BinaryIO = None,      # where to write the generated file
        width: int = -1,           # Image width in pixels (-1 means "any")
        height: int = -1,          # Image height in pixels (-1 means any)
        palette_size: int = 1 << 15,# Palette size (max colors, defaults to 2^15)
        num_frames: int = 1        # Number of frames in the animation
  ) -> bytearray:               # Returns the generated content as an array

    # Use urandom by default
    if rng is None:
        rng = os.urandom

    # First try to read the seed from stdin
    if len(sys.argv) > 1:
        seed = sys.argv[1].encode('utf-8')

        # Try to read seed from standard input. Since this is meant for testing,
        # we don't care about exceptions here. We just make sure to print an error
        # message so we know what went wrong.
        try:
            seed += rng(len(sys.stdin))[:7]
        except Exception as ex:
            print("Error while reading seed:", type(ex).__name__, str(ex))

    else:
        # Read the seed directly from the generator
        seed = rng(16)


    # Generate a frame descriptor block. This contains some basic information
    # about each frame, such as its dimensions and color map.
    def generate_frame():
        nonlocal seed

        # Frame control flags
        fc = {
            0x80: False, # No local colour table
            0x40: True,  # Interlaced image
            0x20: True,  # Sorted colour map
        }[int.from_bytes(seed[1], 'little')]
        seed = seed[1:]


        # Frame dimensions
        dim = {}
        w, h = {
            0 : (width,height ),
            1 : (width//2,height //2),
            2 : (width//3,height //3),
            3 : (width*2 //3, height *2 //3),
            4 : (width*3 //4, height *3 //4),
            5 : (width*4 //5, height *4 //5),
            6 : (width*5 //6, height *5 //6),
            7 : (width*6 //7, height *6 //7),
            8 : (width*7 //8, height *7 //8),
            9 : (width*8 //9, height *8 //9),
            10:(width*9 //10,height*9 //10),

            11:(width*10// 9,height*11//10),
            12:(width*11// 8,height*12//11),
            13:(width*12// 7,height*13//12),
            14:(width*13// 6,height*14//13),
            15:(width*14// 5,height*15//14),
            16:(width*15// 4,height*16//15),
            17:(width*16// 3,height*17//16),
            18:(width*17// 2,height*18//17),
            19:(width*18// 1,height*19//18),
            20:(width*19// 0,height*20//19),
        }.get(int.from_bytes(seed[0], 'little'),
              (
                  min((int.from_bytes(seed[:2], 'little') & 0xfff) + 1,
                       width),

                  min((int.from_bytes(seed[2:], 'little') & 0xfff) + 1,
                       height)))
        dim['w'],dim['h'] = w,h
        seed = seed[3:]

        # Color map
        cm = []
        ncols = {
              0: 1<< 1,
              1: 1<< 2,
              2: 1<< 4,
              3: 1<< 8,
              4: 1<< 16}[int.from_bytes(seed[0], 'little')]
        for _ in range(ncols):
            r = rng(1)[0]
            g = rng(1)[0]
            b = rng(1)[0]

            if all([r == gc and
                   g == gb and
                   b == br for gc,gb,br in cm]):

                continue

            cm[:] = [(r,g,b)]+cm
            if len(cm) >= palette_size: break
        seed = seed[1:]

        return (fc, dim, cm)

    # Return a single LZW code based on the next bits from the generator
    def get_code():
        # Grab the next byte
        nb = ord(next(rng))

        # See how many unused bits are left in the current byte
        rem = 8-numrem
        if rem<=nb:
            # Return only the used bits and keep track of them
            numrem -= rem
            return ((nb >> rem) |
                    (ord(next(rng)) << (8-numrem))) & 0xff
        else:
            # Return all but the first bits, which indicate the number of
            # remaining bits in the next byte
            return nb & ((1<<(8-numrem)-1)
                         if numrem!= 0 else 0xff)



    # The amount of LZW codes used per bit length
    lzwbits = {
        1: 2,   2: 3,   3: 4,   4: 5,   5: 6,   6: 7,   7: 8,   8: 9,
        9: 10,  10: 12,  11: 16,  12: 16,  13: 16,  14: 16,  15: 16}

    # Decode one LZW code stream into a pixel buffer.
    def decode_stream(s):
        buf = []
        s.append(0xffff)                 # EOF symbol
        prev = 0                         # Previous symbol

        while True:                      # Loop until EOF
            b = get_code()                # Get a code
            clen = 1                      # Length of this prefix chain
            csym = -1                     # Symbol decoded with this prefix

            # Find the longest matching prefix
            for i in reversed(range(8)):
                sym = (prev << lzwbits[i]) | (b&((1<<lzwbits[i])-1))

                # Check if there is a match for this prefix
                if sym in s:
                    csym = sym
                    clen = i+1

                    # Stop when we find a perfect match
                    if csym == 0xffff:
                        break

            # Add symbols to the buffer
            while clen > 0:
                buf.append(csym)
                prev = csym             # Remember last code
                clen -= 1


            # Remove already decoded codes
            del s[:len(buf)+1]

            yield buf[-1]
    # }}} End of build_pixel_buffer


    # Build the whole file
    # ---------------------

    # Header
    head = bytearray(b'GIF89a')
    #     Version string: "GIF89a"
    #     Width and Height: Indicate canvas size (not necessarily same as
    #                        animated image size)
    #     Flags: Global color table flag, Sort order flag (ignored),
    #            Size of global color table (background color index)
    #     Background color index: Index of background color in color table
    #                            (none); ignored unless global color table flag
    #                            set
    #     Pixel aspect ratio: Aspect ratio - 1:1 (may also be 4:3)
    head += struct.pack('<HHBBBH',
                        width, height,
                        0x80 | 0x40 | 0x07 | 0, # Flags
                        0,                       # Background color index
                        0x100 if width>=height else 0x200)
    out.write(head)

    # Logical screen descriptor (LSD)
    #   Size of global color table minus 1 (one byte)
    #   Color resolution
    #   Sorting method: none
    #   Flags: Global color table exists
    #   Size of initial global color table minus 1 (one byte)
    #   Bits per primary color
    #   Size of secondary color table
    #   TODO: What are these three fields? They're always zeroed out in my files
    out.write(