#!/usr/bin/env python3

import struct
from typing import BinaryIO


class RNG:
  def __init__(self, buffer : bytes):
    self._buffer = buffer
    self._offset = 0

  @property
  def offset(self):
    return self._offset

  def read(self, numbytes : int):
    result = self._buffer[self._offset : self._offset + numbytes]
    self._offset += len(result)
    return result

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Read entire Logical Screen Descriptor section
    lsdbuf = rng.read(7)
    lsdhdr = struct.unpack("!BBHHHH", lsdbuf)

    # Check to make sure that the version flag is correct
    assert lsdhdr[0] == 0xFF, "Header needs to start with FF"
    assert lsdhdr[6] == 0x00, "Size must be zero so far"

    # Set appropriate flags based on input
    lsdhdr_flags = 0b01110001
    lsdhdr_flags |= (((lsdhdr[1]&0b0000111)+1)<<5)
    lsdhdr_flags |= ((len(lsdhdr[4]))<<5)
    lsdhdr_flags |= (bool(lsdhdr[5])<<3)

    # Reconstruct our new LSD structure with proper flags set
    lsdhdr_full = struct.pack("!BBHHHH", 
        lsdhdr[0], 
        lsdhdr_flags,
        lsdhdr[2], 
        lsdhdr[3],  
        lsdhdr[4], 
        lsdhdr[5] 
    )

    # Add LSD to output
    out.write(lsdhdr_full)

    # Read global color table length from LSD
    gctlen = 2**((((lsdhdr[1]>>5)&3)-1)*3+1)

    # Add GCT to output if we have one
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)

    # We now should be at the start of an image section
    assert rng.read(1)==b"\x2C".encode(), \
      f"{repr(rng.read(1))}!= {repr(b'\x2C'.encode())}"

    # Read all remaining image section headers
    imsgf_array = []
    imsgl_array = []
    while True:
      
        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          break
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Append to our list
        imsgf_array.append(fieldbyte)
        imsgl_array.append(size)

        # Skip over image data as necessary
        if hasimgdata:
            rng.read(size)


    # Now go through each image section trying to find valid ones
    for i in range(0, len(imsgf_array)):

        # Copy the original fieldbyte such that we can modify it
        fieldbyte = imsgf_array[i]
        
        # See if this is indeed a graphics control extension
        if (fieldbyte & 0xF0)!= 0x20 or\
           imsgl_array[i] < 4:

          # If these conditions aren't met, just skip this entry
          print(f"Skipping fieldbyte:{hex(fieldbyte)}")
          continue

        # Extract the graphic control information
        gcinfo = struct.unpack("BBBB", rng.read(imsgl_array[i]-1))

        # Check if transparent color index is specified and valid
        transcolidx = gcinfo[3]
        if transcolidx >= 0:

            # Verify that this color index refers to a color actually 
            # present in the color table
            if transcolidx >= gctlen//3:
                print(
                    f"Transparent color index {transcolidx} does not exist. " 
                    f"(Valid colors are 0-{gctlen//3})" 
                )
                continue

            # Check if disposal method is invalid
            if (gcinfo[1] & ~0b111)!= 0:

                # There isn't much point in continuing here since this
                # seems like a really bad image. Just stop here and move
                # onto other options
                print("Invalid disposal method.")
                continue

        else:
          gcinfo = None

        # Determine if this frame will be interlaceable
        interlaced = bool(fieldbyte & 0x40)

        # Finally check if we were able to extract enough info about
        # the current section which would allow us proceeding further
        if all([gcinfo,interlaced]):
            
            # Create our IHDR RGB image section header
            ihdrrgb = bytes()
            ihrdrgn = bytes()

            ihdrrgb_start = out.tell()

            # Construct IHDRRGB LSB
            ihdrrgb_lsb = [
                0, 
                0, 
                0,  
                0,  
                interlaced,    
                0b011111,    # reserved bits
                0,           # background color
                0            # pixel aspect ratio
            ]
        
            ihdrrgb_lsb = map(lambda v: (v&0xFF), ihdrrgb_lsb)
            ihdrrgb_lsb = [*ihdrrgb_lsb]

            ihdrrgb += struct.pack('<BBBBBBB', 
                                 *ihdrrgb_lsb
                             )

            # Construct IHRDRGN LSB
            ihrdrgn_lsb = [
                0,          # left position
                0,          # top position
                1,          # width - initial value ignored by Firefox
                1,          # height - same ^
                0,          # local colour table flag
                0,          # interlace flag - already determined
            	False       # sort flag - always false
            ]
            
            # If we have transparency, use closest match instead
            if not gcinfo is None:
              ihrdrgn_lsb[7] = False
            
            ihrdrgn_lsb = map(lambda v: (v&0xFF), ihrdrgn_lsb)
            ihrdrgn_lsb = [*ihrdrgn_lsb]

            ihrdrgn += struct.pack('<BBBBBBBB', 
                                  *ihrdrgn_lsb
                                )

            # Add both IHDR sections together
            ihdr = ihdrrgb + ihrdrgn

            # Write IHDR to output
            out.write(ihdr)
    
            # Add the rest of the fields afterwards
            out.write(bytes(gcinfo[:-1]))

            # Add the final byte indicating end of image data
            out.write(b"\x00")

            # Retrieve seek pointer relative to start of other header stuff
            iendptr = out.tell()-ihdrrgb_start-len(ihdr)

            # Calculate the IEND header
            iend = bytes()
            iend_start = out.tell()

            # Construct the first part of the IEND header
            iend += b'[;]'

            # Construct the second part of the IEND header
            iend += struct.pack('<B', iendptr)

            # Write IEND to output
            out.write(iend)


            # Truncate our image section data so we don't include it again
            del immsgf_array[:i+1]