#!/usr/bin/env python3

import sys
import os
import argparse
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Define all our constants
    C_LASTBLOCK          = 0x1
    C_HASGLOBALCOLORTBL  = 0xf7
    C_MINIMALIMGDATA     = 0xe0
    C_MAXIMALIMGDATA     = 0xff
    C_HASLOCALCOLORTBL   = 0xfe

    # Generate the header and add to output
    header = bytes([ord('G'), ord('I'), ord('F'), ord('8'),
                    ord('9'), ord('a')])
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)
    
    # Loop until we've reached the end of the
    # animation; when this happens, the loop body
    # will be executed once more
    while True:

        # Read next field byte for block type
        fieldbyte = int(ord(rng.read(1)), base=0)
        
        # Check if this is the last block
        if ((fieldbyte & C_LASTBLOCK)!= 0):
          out.write(bytes([C_LASTBLOCK]))
          return
    
        # Get size of block
        size = int(ord(rng.read(1)), base=0)

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', (fieldbyte|2) % 256))
        out.write(struct.pack('=B', size))

        # Handle case where there's no image data
        if not hasimgdata:
            continue

        imgdatalen = min(max(size, C_MINIMALIMGDATA),
                         C_MAXIMALIMGDATA-9)+9
        imgdata = rng.read(imgdatalen)
        out.write(imgdata[:size])