#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Generate the header and color table
    width = 73; height = 54
    header = bytearray('GIF87a'.encode())\
        + (width).to_bytes(2,'little')\
        + (height).to_bytes(2,'little')
    palette = bytearray([
        0xFF, 0xFF, 0xFF, 
        0xF5, 0xEA, 0xD0, 
        0xC0, 0xB0, 0xA0, 
        0xD5, 0xAA, 0xBD, 
        0xDC, 0x9D, 0xFF, 
        0xF0, 0xEE, 0xCC, 
        0xEC, 0xFA, 0xFF, 
        0xFFFF])
    header += palette
    header[10] |= 0x80        # Transparent background
    header[13] |= 0x07        # Unused global color table flag
    header[13] &= ~0x08       # Sorting by size flag set
    header[13] |= 0x20        # Global color table present flag
    
    # Write the header to the output stream
    assert(len(header) == 13) 
    out.write(header)

    # Compute the color frequency for each RGB channel
    freqs = [0]*768
    while True:
        framebuf = bytes()

        for y in range(height):

            for x in range(width):

                pixel_idx = ((y*width)+x)*3
                r = int.from_bytes(
                        rng.read(1),'big') % 6
                g = int.from_bytes( 
                        rng.read(1),'big') % 6
                b = int.from_bytes(
                        rng.read(1),'big') % 6
                
                if 0 == freqs[(r<<12)+(g<<6)+b]:
                    header += bytearray([
                        0x21, 0xF9, 0x04, 0x08, 0x00, 0x00FF])
                    
                freqs[(r<<12)+(g<<6)+b] += 1
                
                try:
                    framebuf += bytearray([(r << 4) | g, b])
                except IndexError as e:
                    print("Index error")
                    print("r=%d, g=%d, b=%d" % (r,g,b))
                    raise(e)
        
        blocksize = len(framebuf)
        if not blocksize: break   # No more pixels left!
        header += bytearray([blocksize])
        out.write(framebuf)

    # Write the end-of-file appending block
    footer = bytearray([
        0x3B,          # End of image
        0x21, 0xFE,    # Application extension
        0x00, 0x3E,    # Size of application data
        0x09, 0x03,    # Graphic control label
        0x00,          # Block size
        0x00,          # Packed field
        0x00,          # Delay time
        0xFF,          # Transparent color index
        0x00           # Terminator for graphics control
    ])
    out.write(footer)

    return len(header)

