#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file with width WxH into `out` using the
# random number generator `rng` (/dev/urandom). Multiple calls to
# this function should result in independent random images being generated;
# however, each call will produce the same image due to its seed.
def generate_random_gif(
        rng: BinaryIO, w: int, h: int, out: BinaryIO
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    lsd = rng.read(7)

    # Update the width and height in the Logical Screen Descriptor
    lsd = bytes([*bytes(map(ord, lsd[:6])),
                 *struct.pack('<HH', w, h),
                 *bytes(map(ord, lsd[-1:]))])

    # Add LSD to output
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 2**(int((lsd[4] >> 5))) * 3

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = rng.read(gctlen)
      out.write(gct)


    # Track how many times we've seen a frame before. If we see the same ordering twice, stop.
    prev_frame_order = None
    while True:
        # Generate the frame control flags
        framecontrolflags = rng.read(1)
        # Mark this as an interlace frame
        framecontrolflags |= ord('\x40')

        # See how long our frame is. We want at least two pixels high so we can get some transparency.
        frameheight = min(h + 1, max(h // 2, 2))
        # Make sure we're not too big or negative
        framewidth = max(w - 1, 1)

        # Set the delay time to something semi-realistic
        delaytime = rng.read(2)

    
        # Pack up all our information
        frameinfo = [*framecontrolflags,
                     *delaytime,
                     *struct.pack('<HH', framewidth, frameheight)]


        # Get the order index of the frames that appear before us
        frameorder = struct.unpack(">I", rng.read(4))[0] % 8
        if prev_frame_order!= frameorder:
            prev_frame_order = frameorder
        else:
            break

        # Write the first part of our "sub" block
        subframehead = [frameorder | 0x2C] 
        out.write(bytes(subframehead))

        # Write the rest of our frame-info
        out.write(bytes(frameinfo))

        numcols = 2**((framecontrolflags[0]&7)+1)
        paletteentry = rng.read(numcols*3)
        out.write(paletteentry)

        # Write out each line
        for _ in range(frameheight):

            # Do some fun masking tricks on the X coordinate based on the Y position
            xmask = [(i&1)*rowvalue for i, rowvalue in enumerate(rng.read(framewidth))]
            
            # Write out the actual data
            line = bytes(
                map(
                    lambda x:
                    0xFF*(
                        abs(
                            sum(
                                [
                                    weightedcol[y]*xmask[(x+y)%framewidth]
                                    for y,weightedcol in zip(range(framewidth), packet)
                                ]
                            )//packetsum
                        )
                    ),
                    rng.read(framewidth)
                )
            )
            out.write(line)