#!/usr/bin/env python3

import sys
import os
import time
import struct
from pathlib import Path
from typing import BinaryIO


# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # GIF file header (13 bytes total)
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    def make_header() -> bytes:
        return b'GIF89a'

    # Graphics control block (7 bytes total)
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    def make_gce() -> bytes:
        delay = 0x02D0    # 450 ms of display time for one frame
        trans = 0         # no transparency in this example

        buf = [
            0x21,       # extension introducer
            0xF9,       # graphics control label
            0x04,       # size of GCE block
            ]
        
        # packed field flags
        buf += [(trans << 2 | 0x01).to_bytes(1, 'little')]     # bits 0-1 are reserved

        # delay times (2 bytes each)
        buf += [
                ((delay >>  0) & 0xFF).to_bytes(1, 'little'), 
                ((delay >>  8) & 0xFF).to_bytes(1, 'little'),
                ] 

        # transparent color index
        buf += [(trans).to_bytes(1, 'little')]
    
        # terminate with a zero byte to indicate end of extensions
        buf += [b'\0']

        return b''.join(buf)


    # logical screen descriptor (7 bytes total)
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    def make_lsd(width: int=64, height: int=64):
        gct = 0              # global color table flag
        sort = 0             # sorting flag
        gctsize = 0          # color resolution (bits per primary color)
        bgindex = 0          # background color index
        pixelAspect = 0      # pixel aspect ratio = 0

        buf = []
        buf += [('!', chr(width % 256)).encode()]
        buf += [('?', chr(height % 256)).encode()]
        buf += [(chr(gct<<7 | sort<<5
                          | gctsize<<3
                          | bgindex)).encode()]
        buf += [(chr(pixelAspect)).encode()]
        return b''.join(buf)


    # local colour table (3 * n bytes total)
    def make_lct(ncolors: int) -> bytes:
        palette = [] 
        for i in range(ncolors):
            r = next(rng) % 256 
            g = next(rng) % 256 
            b = next(rng) % 256 
            palette += [r, g, b]

        return b''.join([struct.pack('<BBB',*c) for c in palette])

    # TODO: use colors from X11 named colours if they exist somewhere

    width = 64
    height = 64
    ncolors = pow(2, 8 - 3)
    lct = make_lct(ncolors)

    header = make_header()
    gce = make_gce() 
    lsd = make_lsd(width, height) 

    out.write(header + lsd + lct + gce)

    # write actual pixels
    for _y in range(0, height//8):
        row = list(map(lambda x: x%ncolors, map(next, [rng]*(width))))
        out.write(bytes([(row[i] // 8) << 5 |
                        (((row[i] / pow(2, 3)) % 8) << 2) | 
                        ((row[i] % 4) >> 1)
                        for i in range(width)]))
    
    print("done")


