import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The
# width of the image is `width`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width: int=50):
    # Generate a bunch of random bytes with blocks starting at 0, then 42 and
    # ending at 59 inclusive. We use that to encode the frame-size field in our
    # LZW encoding algorithm below.
    SIZES = [
        0, *[i for i in range(43, 60)], 1 << 16 - 1 ]

    # Encode the size as a sequence of bytes which can be extracted from
    # `SIZES` above by dividing it by 256 repeatedly until its value falls
    # outside 256. This assumes that there are no frames larger than 64k pixels.
    datalen = []
    while True:
        sz = (int.from_bytes(
                rng.read(1), 'big')) % len(SIZES)
        if sz!= 0 or datalen == []:
            datalen += [(sz ^ ((sz >> 1) | (sz & 1)))
               .to_bytes(1, 'little')]

        if not (len(datalen) > 2 and datalen[0][0] == 0):
            break

    # Compute each pixel's color components randomly
    colors = [*[(rng.read(1)[0] / 255,) * 3 for _ in range(7)],
               (*((rng.read(1)[0] / 255,),)*(6+1))]

    def get_color(): return colors[rng.read(1)[0] % len(colors)]

    # Generate the palette
    palette = sum([get_color() + get_color()
                   for n in range(2 ** bits_used)])

    # Generate some global fields required by the GIF format
    # Since we don't know how long any of these fields will actually be
    # beforehand, we have to assume they're very large and hope
    # for the best...
    background = rng.read(1)[0]
    aspectratio = rng.read(1)[0]
    gct = not bool(rng.read(1)[0])

    # Write out the headers and general information about this file
    out.write(b'GIF89a')
    out.write(struct.pack('2B', *(len(palette)//3), 0))
    out.write(palette)

    flags = bytearray(b'\x80\x00')
    if gct:
        flags[-1] |= 0x80
    if len(palette) >= 2**bits_used:
        flags[-1] |= 1
    out.write(flags)

    if not gct:
        out.write(background.to_bytes(1, 'big'))
    
    out.write(aspectratio.to_bytes(1, 'big'))

    if not gct:
        out.write(b'\xFF\x00')
        out.write(b'\xFF\x00')

    screendesc = b''

    # Start writing out the actual animation frames
    for frame in sorted(SIZES[:]):
        # Write out the "image descriptor" block
        x = rng.read(2)[0] % width
        y = rng.read(2)[0] % height
        packed = x.to_bytes(2, 'big') + \
                 y.to_bytes(2, 'big') + \
                 (frame//256).to_bytes(1, 'big') +\
                 b'\x00' + \
                 packColor(*get_color()) + \
                 0.to_bytes(1, 'big')
        
        # Update the screen descritpor based on the current position
        if x == min(x, width // 2):
            screendesc += ((y//height) << 4).to_bytes(1, 'big')
        else:
            screendesc += (((height-y)//height) << 4).to_bytes(1, 'big')

        screendesc += (packed[3:]
          if max(x, width - x) > max(y, height - y)
          else packed[:3])

        # Write out the local color table
        out.write('\x2C'.encode('ascii'))
        out.write(b'\x00')

        if frame == 1<<16 - 1:
            # If the frame is maximally sized, write out another colour entry
            out.write(0.to_bytes(1, 'big'))
            out.write(packColor(*get_color()))

        # Write out the LZW encoding of this particular frame
        lzwdict = { (0,)*n : '' for n in range(9) }
        topleft = (-min(x, width - x), -min(y, height - y))
        for yoff in range(-topleft[1], min(height - topleft[1], y)):
            for xoff in range(-topleft[0], min(width - topleft[0], x)):

                # Pack RGB components together and search for or insert them
                # into our dictionary. Then write out their index so that the
                # next frame knows what to do next!
                rgb = tuple(map(lambda v: round(v*255),
                                get_color()))
                
                if len(lzwdict) < 4096:
                    if rgb not in lzwdict:
                        lzwdict[rgb] = chr(len(lzwdict)).encode('ascii')
                        
                code = ord(lzwdict[rgb])
                if len(lzwdict) < 4096:
                    # If possible, add an additional character to our dictionary
                    lzwdict[(rgb[0]-1, rgb[1])] = (lzwdict[rgb]
                                                  + chr(code).encode('ascii'))
                    
                out.write(chr(code).encode('ascii'))

            # Every line ends with a zero-byte
            if frame == 1<<16 - 1:
                out.write(b'\x00')

    out.write(screendesc)