#!/usr/bin/python3

'''Generates a random GIF file using /dev/urandom as source of entropy.'''

import io
import os


def get_bytes(n=None) -> str:
    '''Returns n (or up to len(os.urandom()) or by default all available bytes)
    from /dev/urandom in hexadecimal string format.'''
    with open('/dev/urandom','rb') as f:
        return ''.join('%02x'%b for b in f.read(n)) if n else \
               ''.join('%02x'%b for b in iter(lambda:f.read(1),''))


class GifWriter():
    def __init__(self, filename :str, width :int, height :int,
                 duration :float=.1) -> None:
        self.width   = width
        self.height  = height
        self.duration = duration

        assert isinstance(filename, str), 'filename must be a valid str.'
        
        self._filehandle = open(filename,'wb')
        self._filehandle.write(b'GIF89a')
    
        self._write(4, width, height)
        self._write(1, b'\xff\x1f')
        

    def __enter__(self) -> 'GifWriter':
        return self


    def _write(self,*data) -> None:
        '''Write data to file as big endian integers.'''
        to_write = [
                len(data).to_bytes(1,"big"),
                *map((lambda x:x if type(x)==bytes else 
                      (x).to_bytes((x).bit_length()+7//8,"big")),
                      data)]
        self._filehandle.write(b''.join(to_write))

    def step(self, pixel_values=[]) -> None:
        '''Writes one frame to file. pixel values are given as list in row 
        major order and are encoded on a byte level.'''
        self._write(0x2c)
        self._write(0,0, self.width - 1, self.height - 1)
        self._write(2**(self.duration+1)-1)
        self._write(0x87)
        self._filehandle.write(b''.join([pixel_byte for pixel_row in pixel_values
                                        for pixel_byte in pixel_row]))

    @staticmethod
    def _make_square_image(pixel_values :list, size) -> list:
        '''Pads square pixels to fit into image of size size*size where the
        output is a list of lists.'''
        rows = []
        for rownum in range(len(pixel_values)//size +
                           (1 if len(pixel_values)%size else 0)):
            if not (len(rows)<size): raise ValueError("Too many pixels.")
            
            start       = rownum * size - sum(map(len,rows))
            end         = min(start + size - len(rows), len(pixel_values))

            rows += [[*(pixel_values[i]
                       for i in range(start,end)),
                     *(b'\x00' for j in range(size - (end-start)))]]
            
        return rows
    
    def write_animation(self, pixel_values :list) -> None:
        '''Writes an animation to file consisting of frames generated based on
        pixel values. The pixel values are given as flat array of bytes that
        can have any length but MUST be a multiple value of the square of the
        images height.'''
        if len(pixel_values) == 0: 
            raise ValueError("No pixels specified!")
        elif len(set(map(type, pixel_values)))!= 1:
            raise TypeError("All data must be of same type.")
        else:
            try:
                pixel_rows = self.__class__._make_square_image(
                        pixel_values, 
                        self.width)#int(len(pixel_values)**(.5)))
            except Exception as e:
                raise ValueError("Invalid amount of pixels") from e
        
        while True:
            self.step(pixel_rows[:self.width])
            del pixel_rows[:self.width]
            yield
            
        
    def close(self) -> None:
        self._filehandle.seek(-1,io.SEEK_END)
        self._filehandle.write(b'\x00')

        
