#!/usr/bin/env python3

import sys
import os
os.chdir(sys.path[0])

import struct
from typing import BinaryIO

# Returns an array of 256 pseudorandom bytes generated by xoring the
# previous byte with the current one (mod 256), starting from seed.
def generate_pseudorandom_bytearray(seed):
    result = [seed] * 256
    index = seed + 1
    while True:
        result[index % len(result)] ^= result[(index - 1) % len(result)]
        yield result[(index % len(result))]
        index += 1


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        min_width=4, max_width=None, width_step=1,
        min_height=4, max_height=None, height_step=1        
    ):
    if not isinstance(min_width, int): raise TypeError("min_width must be an integer")
    elif min_width < 4: raise ValueError("min_width must be >= 4")
    
    if not isinstance(max_width, int): raise TypeError("max_width must be an integer or None")
    elif max_width is not None and max_width <= min_width: raise ValueError("max_width must greater than min_width")

    if not isinstance(width_step, int): raise TypeError("width_step must be an integer")
    elif width_step <= 0: raise ValueError("width_step must be > 0")


    if not isinstance(min_height, int): raise TypeError("min_height must be an integer")
    elif min_height < 4: raise ValueError("min_height must be >= 4")
    
    if not isinstance(max_height, int): raise TypeError("max_height must be an integer or None")
    elif max_height is not None and max_height <= min_height: raise ValueError("max_height must greater than min_height")

    if not isinstance(height_step, int): raise TypeError("height_step must be an integer")
    elif height_step <= 0: raise ValueError("height_step must be > 0")



    # Write the magic string (header) to the output stream
    out.write(b'GIF89a')

    # Start generating the image data

    # First, we need to choose a screen size. We use a simple scheme whereby 
    # width increases in powers of two between min_width and (max_width - stepsize)
    # (and similarly, for height). If max_dim is none, then it simply stops at the maximum available.
    def next_pow2(val): return pow(2, val-1).bit_length() 

    width = next_pow2(rng.randint(min_width, 8)) << 1
    while ((not max_width) or width < max_width) and \
            (width & (width_step - 1)):
        width <<= 1

    height = next_pow2(rng.randint(min_height, 8)) << 1
    while ((not max_height) or height < max_height) and \
            (height & (height_step - 1)):
        height <<= 1

    print("Width:", width, "Height:", height)

    # Now write the logical screen descriptor
    out.write((width | height).to_bytes(2, 'little', signed=False))

    
    # Next, we need to write a color table with 256 entries. Each entry consists
    # of three bytes R G B all in range [0, 255]. For now, we just use a linear
    # progression ranging over each channel from 0 to 255. TODO: Perhaps make this
    # more realistic?
    palette = []
    for i in range(0, 256): 
        palette += list(map(lambda v : v // 2, 
            rng.randrange(i*3, i*3+3)))

    # There are various ways to encode RGB values within a single byte. The most
    # common way is as 3 bits per channel. This gives us space for 8 colors
    # (2^3 == 8) but only 24 bit total. Since we represent every color in eight
    # bits, we have 2 pixels per byte which means we can store up to 12 colors
    # (8 * 2 == 16, rounded down to 12). However, since our final product is going to 
    # run on a monochrome display anyway, we might as well just pick either two
    # pair of colors (12 / 2 == 6) and duplicate them so that there's effectively
    # no redundancy. That leaves us with four groups of six colors left over; these
    # will act as background colors for the rest of the animation. Each group has
    # a black background, a white background, and two other colors. Within each color
    # group we also have greyscale palettes, slightly darker and slightly lighter.
    # Finally, there's a few special tinted variations of a few different colors.
    # I picked the following colors as my main foreground palettes:
    fg_pallete = [[  0,   0,   0],
                  [255, 255, 255]]

    # Similarly, here are the backgrounds for each of those colors
    bg_pallete = [[  0,   0,   0],
                  [255, 255, 255]]

    # And these were the secondary ones
    alt_fg_pallete = [[100, 100, 100],
                      [175, 175, 175]]

    # And these were their corresponding backgrounds
    alt_bg_pallete = [[175, 175, 175],
                      [100, 100, 100]]

    # Here are the grayscale palettes for the foreground (primary) colors. These
    # are brighter versions of what you'd get if you made the primary color
    # even more contrasting against white
    fg_grayscale_pallete = [[  0,   0,   0],
                            [ 50,  50,  50],
                            [ 80,  80,  80],
                            [120, 120, 120],
                            [170, 170, 170],
                            [200, 200, 200],
                            [255, 255, 255]]

    # Here are the grayscale palettes for the secondary colors
    alt_fg_grayscale_pallete = [[ 80,  80,  80],
                                [100, 100, 100],
                                [120, 120, 120],
                                [140, 140, 140],
                                [160, 160, 160],
                                [180, 180, 180],
                                [200, 200, 200]]

    # Finally, let's add the special tints. The first 2 represent shades of blue
    # and orange respectively. The second two represent greens - both regular and
    # muted green.
    fg_tinted_pallete = [[  0,   0, 255],
                         [255,   0, 128]]
    alt_fg_tinted_pallete = [[  0, 128, 255],
                             [128, 255, 255]]

    bg_tinted_pallete = [[  0,   0,   0],
                         [255, 255,   0]]
    alt_bg_tinted_pallete = [[  0, 255,   0],
                             [255, 255,   0]]
    # Combine all of the above together
    pallete = [
        palette[:6] + fg_pallete,
        palette[6:12] + bg_pallete,
        palette[12:18] + alt_fg_pallete,
        palette[18:] + alt_bg_pallete,

        fg_pallete + fg_grayscale_pallete,
        bg_pallete + fg_grayscale_pallete,
        alt_fg_pallete + alt_fg_grayscale_pallete,
        alt_bg_pallete + alt_fg_grayscale_pallete,

        fg_pallete + fg_tinted_pallete,
        bg_pallete + bg_tinted_pallete,
        alt_fg_pallete + alt_fg_tinted_pallete,
        alt_bg_pallete + alt_bg_tinted_pallete
    ]
    

    # Save the computed palette to the output file (it's a binary file, dammit!)
   