#!/usr/bin/env python3

import struct
import string
import random
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width:int=400, height:int=400):

    # Header:
    header = "GIF89a".encode('ascii')
    out.write(header)

    # Logical screen descriptior
    size = (width,height)
    colorTableSize = -1
    backgroundColorIndex = 0
    pixelAspectRatio = 0
    graphicsControlExtension = False
    applicationExtension = False
    commentString = "".join([random.choice(string.printable[:-5]) for _ in range((size[0]+size[1])*4)])
    printableCommentString = ''.join([c if c not in '\r\n' else repr(c)[1:-1] for c in commentString])


    logicalScreenDescriptor = bytearray()
    logicalScreenDescriptor += struct.pack('<H', *size)        # Size of screen area
    logicalScreenDescriptor += chr(colorTableSize).encode('latin-1')    # Color table size
    logicalScreenDescriptor += chr(backgroundColorIndex).encode('latin-1')         # Background color index
    logicalScreenDescriptor += chr(pixelAspectRatio).encode('latin-1')             # Pixel aspect ratio

    out.write(logicalScreenDescriptor)


    # Global Color Table
    globalColorTable = bytearray()
    out.write(globalColorTable) # Empty!

    # Image Descriptor
    imageDescriptor = bytearray()

    imageWidth = width
    imageHeight = height
    localColorTableFlag = True
    interlaceFlag = False
    sortFlag = False
    localColorTableSize = 7   # 2^7 == 128
    reservedField = 0
    backgroundColourIndex = 0
    imagePixelRatio = 0

    imageDescriptor += struct.pack('<BHHBBB', localColorTableFlag,
                                  imageLeftPosition,
                                  imageTopPosition,
                                  imageWidth,
                                  imageHeight,
                                  localColorTableSize)

    imageDescriptor += struct.pack('<BBB',      reservedField,
                              backgroundColourIndex,
                              imagePixelRatio)

    out.write(imageDescriptor)




    # Extension Blocks
    extensionBlockLabel = ord('\xF9'.encode('latin-1'))
    graphicControlLabel = ord('\xFE'.encode('latin-1'))
    commentLabel = ord('\xFF'.encode('latin-1'))

    if graphicsControlExtension or applicationExtension:

        # Application Extension Block
        applicationIdentifier = "NETSCAPE2.0".encode("ascii")     # Application identifier
        applicationAuthCode = "\x01\x00".encode('latin-1')       # Application auth code
        blockTerminator = "\x00".encode('latin-1')              # Block terminator


        applicationExtensionSubBlocks = bytearray()
        applicationExtensionSubBlocks += applicationIdentifier
        applicationExtensionSubBlocks += len(applicationIdentifier).to_bytes(1, 'little')          # Subblock length
        applicationExtensionSubBlocks += applicationAuthCode
        applicationExtensionSubBlocks += blockTerminator
        applicationExtensionBlockSize = len(applicationExtensionSubBlocks)

        applicationExtensionHeader = bytearray()
        applicationExtensionHeader += extensionBlockLabel                                 # Extension block label
        applicationExtensionHeader += applicationExtensionBlockSize.to_bytes(1, 'little')    # Extension block size
        applicationExtensionHeader += applicationIdentifier                              # Data subblock label
        applicationExtensionHeader += applicationExtensionBlockSize                         # Data subblock size

        out.write(applicationExtensionHeader)
        out.write(applicationExtensionSubBlocks)

    else:
        pass



    # Comment Extension Block
    asciiCommentString = printableCommentString.encode("ascii")
    nullByte = b"\x00"

    asciiCommentLength = len(asciiCommentString)
    commentSubBlocks = bytearray()
    commentSubBlocks += asciiCommentString
    commentSubBlocks += nullByte
    commentSubBlocks += nullByte
    commentSubBlocks += nullByte

    commentDataSubBlockSize = len(commentSubBlocks)

    commentSubBlockLabel = commentLabel
    commentSubBlockSize = commentDataSubBlockSize+2
    commentHeader = bytearray()
    commentHeader += commentSubBlockLabel
    commentHeader += commentSubBlockSize.to_bytes(1, 'little')
    commentHeader += asciiCommentLength.to_bytes(1, 'little')

    out.write(commentHeader)
    out.write(commentSubBlocks)


    # Graphics Control Extension Block
    delayTime = max(33, min(int(width*height**0.5), 255))
    transparentColourFlag = bool(random.randrange(5)) #% 2)
    userInputFlag = False
    disposalMethod = 0

    graphicsControlExtensionSubblocks = bytearray()
    graphicsControlExtensionSubblocks += delayTime.to_bytes(2,'little')                      # Delay time
    graphicsControlExtensionSubblocks += bytes([transparentColourFlag |
                                                userInputFlag << 1 |
                                                disposalMethod << 2 ])                    # Flags and transparancy colour index
    graphicsControlExtensionSubblocks += b'\x00'

    graphicsControlExtensionBlockSize = len(graphicsControlExtensionSubblocks)+1

    graphicsControlExtensionHeader = bytearray()
    graphicsControlExtensionHeader += extensionBlockLabel                                # Extension block label
    graphicsControlExtensionHeader += graphicsControlExtensionBlockSize.to_bytes(1, 'little')   # Extension block size
    graphicsControlExtensionHeader += graphicControlLabel                               # Data subblock label
    graphicsControlExtensionHeader += graphicsControlExtensionBlockSize                  # Data subblock size

    out.write(graphicsControlExtensionHeader)
    out.write(graphicsControlExtensionSubblocks)



    # Local Colour Table (if present)
    localColorTable = bytes([])
    out.write(localColorTable)



    # Image Data
    encoding = 0            # No LZW compression
    lzwMinimumCodeSize = 2  # Minimum Code Size required by spec

    imageDataSubBlocks = bytearray()
    imageData = b''+"".join(["." for _ in range(max(len(data)//(lzwMinimumCodeSize),width*height)*lzwMinimumCodeSize)]).encode("ascii")[:width*height]*lzwMinimumCodeSize
    print(f"{width} {height}")

    imageDataSubBlocks += struct.pack('<BB', encoding,
                                      lzwMinimumCodeSize)                                   # Encoding method used & minimum code size
    imageDataSubBlocks += imageData                                               # Image data
    imageDataSize = len(imageDataSubBlocks)

    imageDataSubBlockLabel = 0x2C # Image separator
    imageDataSubBlockSize = imageDataSize

    imageDataHeader = bytearray()
    imageDataHeader += imageDataSubBlockLabel
    imageDataHeader += imageDataSubBlockSize.to_bytes(1, 'little')

    out.write(imageDataHeader)
    out.write(imageDataSubBlocks)


    # Trailer
    trailer = bytearray()
    trailer += b"\x3b" # File terminator

    out.write(trailer)
