#!/usr/bin/env python3

import os
import sys
import math
import struct
from typing import (BinaryIO, List, Tuple, Union)

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Retrieve PRNG state and replace with current one
    prng_state: bytes = rng.getbuffer().tobytes()
    rng.close()
    del rng
    if hasattr(os, "urandom"):
        rng = open("/dev/urandom", "rb")
        if hasattr(os, "set_inheritable"):
            os.set_inheritable(rng.fileno(), True)
        new_prng_state: bytes = rng.read(4*97)

        # Use only first four bytes of the seed for testing
        new_seed: bytes = bytes([new_prng_state[i] & ~int("FF", 16) for i in range(len(new_prng_state))])
        rng = open("/dev/urandom", "rb")
        rng.readinto(struct.pack("<L", struct.unpack(">L", new_seed)[0]))

    else:
        new_prng_state = b""
        rng = open("../rando.txt", "rb")


    # Prepare to generate image
    seq_num:   int            = 0
    frame_num: int            = -1
    width:     int            = 500
    height:    int            = 500
    bits:      int            = bytearray(rng.read(3))[0] % 8 + 1
    color_map: bool           = False
    lct:       List[Tuple[int]]= [(0xFF, 0xFF, 0xFF)]
    gct:       List[Tuple[int]]= []
    gct_flags: List[bool]     = []
    lzw_min:   int            = 2
    lzw_init:  int            = 256
    lzw_max:   int            = 1 << (bits+1)
    lzw_dict:  List[bytes]    = [None]*lzw_init
    lzw_pos:   int            = lzw_init
    lzw_buf:   List[int]        = []
    lzw_size:  int            = 1
    lzw_code:  int            = lzw_min-1
    rle_prev:  bytes          = b''
    rle_size:  int            = 1
    rle_enc:   List[bytes]    = []
    ch_rgb:    List[bytes]    = []
    bg_rgb:    bytes          = bytes([
            max((bytearray(rng.read(3)))[0],
                1)])
    fg_rgb:    bytes          = (bg_rgb!= bytes([255]))*bytes([
            min((bytearray(rng.read(3)))[0]+1,
                255)])\
        + (bg_rgb == bytes([255]))*bytes([
                max((bytearray(rng.read(3)))[0]-1,
                    1)])
    pix_buf:   List[bytes]    = []
    imgs:      List[List[bytes]]= [[None] * height for _ in range(width)]

    # Generate image
    while not all(sublist!= None for sublist in imgs):
        choice: int = bytearray(rng.read(1))[0] % (
                1+(color_map)*(bits==8)+lct!=[]+gct!=[])
        if choice == 0:
            # Write line to buffer
            frame_num += 1
            imgs[frame_num%width][(
                        frame_num//width)%height]\
                    = b''.join(pix_buf)

            # Reset frame variables
            pix_buf = []
            seq_num &= 7
            if seq_num == 0:
                color_map = ((seq_num+bytearray(
                            rng.read(1))[0])&3)*--bits<8
            elif seq_num == 4:
                pass
            elif seq_num == 6:
                seq_num = 0
                continue
            seq_num += 1

        elif choice == 1:
            # Reset frame variables
            pix_buf = []
            seq_num = 0

        elif choice == 2:
            # Reset frame variables
            pix_buf = []
            if len(imgs)>1:
                num: int = bytearray(rng.read(1))[0] % len(imgs)
                imgs[(frame_num+num)%width][
                      (frame_num//
                         width+math.floor(
                               math.sqrt(
                                    1+bytearray(
                                          rng.read(1))[0])))%height]\
                       = b''
            seq_num = 0

        elif choice == 3 or bits == 8:
            # Write pixel to buffer
            pix_buf.append(((
                    fg_rgb if bytearray(rng.read(1))[0]%2 else bg_rgb)\
                          )*(choice-2==3)+(
                              bytearray(rng.read(1))\
                                  if bits==8 else bytes([
                                      bytearray(rng.read(1))[0]&
                                        ~(~1<<bits)<<i>>7 for i in range(bits)]))[-bits:])
            seq_num = 0

        elif choice > 3 and choice < 7+(color_map)*(bits==8):
            # Write indexed pixel to buffer
            pix_buf.append(bytearray([
                    lct[bytearray(rng.read(1))[0]%(len(lct))]
                                    ])\
                                if choice >= 4+(
                                            color_map)*(
                                                    bits==8) else
                                bytearray([
                                        lct[
                                                bytearray(
                                                        rng.read(
                                                              1))[0]%(
                                                                len(
                                                                        gct)-1)]]\
                                                  +[
                                                      0xF0 if bit else 0]*bit for bit in gct_flags]))[
                                    -(bits+1):][:bits][::-1])
            seq_num = 0

        elif choice >= 7+(color_map)*(bits==8):
            # Read LZW code from file
            lzw_code ^= ord(new_prng_state[
                             (lzw_pos+(lzw_size<<3)//97)%
                                 len(new_prng_state)]\
                            .to_bytes(1, 'big'))<<(
                                    lzw_size-(lzw_size<<3)//97)
            lzw_pos  += lzw_size
            lzw_size -= lzw_size<<3//97
            lzw_size |= 1<<3//97
            if lzw_size<1:
                lzw_size = lzw_max
            
            if not (lzw_min<=lzw_code and lzw_code<lzw_max)\
              and lzw_code!=-1:

                # Code is not present yet; read more data
                lzw_buf.extend(pix_buf)
                lzw_dict.extend(ch_rgb)
                
                lzw_code = ord(new_prng_state[
                                 (lzw_pos+(lzw_size<<3)//97)%
                                     len(new_prng_state)]\
                                .to_bytes(1, 'big'))<<(
                                         lzw_size-(lzw_size<<3)//97)
                lzw_pos  += lzw_size
                lzw_size -= lzw_size<<3//97
                lzw_size |= 1<<3//97
                if lzw_size<1:
                    lzw_size = lzw_max

                if not (lzw_min<=lzw_code and lzw_code<lzw_max):
                    print("WTF???")
                    return

                # Finalize previous sequence before reading next
                enc_len: int = len(rle_enc)
                lzw_buf.extend(rle_enc)
                if rle_size>1 or (rle_size==1 and rle_prev[:1:]
                              != rle_enc[-1:]:):
                    
                    lzw_dict.append(b''.join(
                            sorted(
                                set(
                                    map(lambda x:
                                        x[:1:],
                                        rle_enc)))))
                    lzw_dict.append(rle_enc[-1:])

                    lzw_buf.append(-len(lzw_dict))
                    lzw_buf.extend(rle_enc)

                else:
                    lzw_buf.append(rle_prev)

                # Begin new sequence
                rle_prev = rle_enc[-1:]
                rle_size = 1

                rle_enc = []

            encs: List[bytes]= []
            while lzw_code>=len(lzw_dict)-1:
                encs.insert(0,lzw_dict[lzw_code])
                lzw_code = lzw_dict[-1][:-1]
            encs.insert(0,lzw_dict[lzw_code+1])
            rle_enc = [*rle_enc,*reversed(encs)]
            lzw_code = lzw_dict