#!/usr/bin/env python3

import sys
import os
import struct
from typing import BinaryIO


class RandomGifGenerator():
    """A class that generates a random GIF file"""
    def __init__(self, seed=None):
        self._r = None

        if seed is None:
            try:
                seedfile = open('/dev/urandom', 'rb')
            except FileNotFoundError as e:
                print("Please install /dev/urandom", file=sys.stderr)
                raise e

            with seedfile:
                self._r = seedfile.read()
        else:
            self._r = bytes(seed)


    @staticmethod
    def _genblock(
            gen: BinaryIO,
            writefunc: callable,
            writeparam: list=[],
            btypechar: int=b'\x08',     # Block type for plain text
        )->bool:
        """Creates a new block on stream `gen` according to its parameters
           Returns whether or not the block should be terminated (`False`)

           Args:
               gen: Stream to create the new block on
               writefunc: Function used to actually write to the stream
                       Should take at least one parameter (the generator object)
                 If additional parameters are needed, they may be passed through
                         writeparam
               btypechar: Byte indicating the block type
        """
        # Error check
        if len(writeparam)==0:
            writeparam=[gen]

        # Create the block type byte
        blockbyte = btypechar | 2       # Set the packed bit, and set the
                                        #   leftmost bit of the block type char
                                        #   to indicate extension blocks

        # Call write function
        writefunc(*writeparam)
        writeparam[-1].flush()         # Flush after each write

        # Terminate if there was an error writing
        if gen.closed:
            return False

        # Begin constructing our own block
        block = [                # List containing all the bytes making up 
            blockbyte        #   this block
        ]

        # Get the size of the block
        sizestring = gen.read(1)
        if len(sizestring)==0:      # Terminate if end-of-stream occurs
            return False
        if sizestring==b'\x00':    # Size can't be zero; force minimum size
            sizestring = b'\x01'

        # Add size string to beginning of block
        block.append(sizestring)

        # Compute the total packet size based on the first character
        # The extra two bytes are for the initial block and size characters respectively
        pktsize = struct.unpack('>H', sizestring)[0]+2+1

        # Keep reading until the whole packet is written
        while pktsize > 0:
            pkt = gen.read(min(pktsize, 65535))
            if len(pkt)==0:          # Terminate if end-of-stream occurs
                return False
            blkpkt = [               # Contains all the bytes making up 
                struct.pack('>B', len(
                    pkt                 # Packet being added to block
                ))                     #   packet
            ] + pkt                    # Add packet length packet, then payload

            # Append this block to our actual block
            block += blkpkt

            # Update packets remaining
            pktsize -= len(blkpkt)+1       # Subtract for size byte

        # Send the completed block over
        writeparam[-1].write(bytes().join(block))
        writeparam[-1].flush()             # Flush after each write

        return True                   # Return true if we're done here


    def generate_random_gif(self, out: BinaryIO)->bool:
        """Generates a randomized gif onto `out`.
           Returns whether or not generation was successful."""

        # Start by generating the header
        if self._r is None:
            # We weren't given anything to read from!
            # Just use some default value instead.
            ret = self._genblock(
                        out, out.write, [b"GIF89a"], 
                        b'\x00' 
                    )
        elif isinstance(self._r, str):
            # This isn't binary but a string...
            # Try converting to bytes
            ret = self._genblock(
                        out, out.write, [
                            self._r.encode('utf-8'), 
                            b"\n"
                        ],
                        b'\x00' 
                    )
        else:
            randbytes = bytes(self._r)
            
            # Create random header
            hdr = []
            for i in range(6):
                hdr.append(randbytes[i])

            # Make sure the first three bytes match 'GIF'
            assert(hdr[:3]==b'GIF')

            ret = self._genblock(
                        out, out.write, [bytes(hdr)],
                        b'\x00' 
                    )

        # If we had some sort of error generating the header earlier, stop
        if not ret:
            return False

        # Now get the Logical Screen Descriptor (LSD) data
        ret = self._genblock(
                    out, out.write, [],
                    b'\x2c'    # LSD block
                )

        # If we had an error getting the LSD, stop
        if not ret:
            return False

        # Now get the Global Color Table
        randbytes = bytes(self._r)

        # Number of colors defined in global color table
        numcolors = 2**((ord(randbytes[4])&0b00000110)>>1)

        globalcolortable = []
        for i in range(numcolors*3):
            c = chr(randbytes[(i%7)+5]).encode('utf-8')
            globalcolortable.append(c[0])
        rgbstring = b"".join(globalcolortable)

        # If there were no colors in the global color table, just pass along
        # what we have right now and let it fail later because it doesn't 
        # contain enough data
        if numcolors!= 0:
            # Pad our RGB values with empty space so the rest of the program knows how far
            # to move forward when parsing them
            padlength = math.ceil(numcolors*3/3)*4 - numcolors*3
            padding = b'\x00'*padlength
        else:
            padding = b''

        numcols = 2**((ord(randbytes[4]&0b00000001)<<2)|(ord(randbytes[5])>>6))
        trnscolpresent = bool(ord(randbytes[4])&0b00000010)
        interlaceflag = bool(ord(randbytes[5])&0b00000100)
        sortedflag = bool(ord(randbytes[5])&0b00001000)

        ret = self._genblock(
                    out, out.write, [rgbstring],
                    b'\x00' 
                )

        # If there was an error getting the global color table, stop
        if not ret:
            return False

        # Now start creating other important parts of the GIF picture, starting with
        # the Image Control Extension
        icceext = [
            b"\x21\xf9\x04",              # ICE label and version number
            b"\x00",                      # No subfields present
            b"\x00",                      # No background color index specified
            b"\xff",                      # Transparent color flag
            b"\x00"                       # No transparent color index
        ]

        ret = self._genblock(
                    out, out.write, [icceext],
                    b'\x00' 
                )

        # If there was an error getting the ICE, stop
        if not ret:
            return False
        
        # Next come Application Extensions
        apext = [
            b"\xFF\xFF"
        ]

        ret = self._genblock(
                    out, out.write, [apext],
                    b'\x00' 
                )

        # If there was an error getting the application extensions, stop
        if not ret:
            return False


        # Finally comes the Graphics Control Extension
        # Let's save off a few things for later
        delaytime = struct.unpack(">H", randbytes[7:9])[0]
        transpflag = bool(randbytes[7]&0b00010000)
        transindex = ord(randbytes[8])
        userinputflag = bool(randbytes[8]&0b00000001)

        gcceext = [
            b"\x21\xF9\x04",                  # Type label and version
            b"\x08",                          # Subfield count
            b"\x00"*2,                        # Reserved (now obsolete?)
            b"\x00" if transpflag else b"\x01",   # Transparency flag
            struct.pack(">H", delaytime),     # Delay time
            b"\x00"*1,                        # Transparent color index
            b"\x00"*1                         # Terminator
        ]

        ret = self._genblock(
                    out, out.write, [gcceext],
                    b'\x00' 
                )

        # If there was an error getting the GCCE, stop
        if not ret:
            return False

        # Well.. technically, nothing could go wrong until now
        # Let's do the image header section
        imgehead = [
            b"\x2c",                          # Image separator
            b"\x00",                          # Left position
            b"\x