#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file with up to 5MB of noise in it at
# `/dev/urandom`.
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # Header (always starts with this string)
    header = b'GIF89a'

    # The finalization byte will be either a zero or a semicolon. We want
    # roughly half zeros and half semicolons.
    numzeros = int(-4e6 * (int.from_bytes(
        rng.read(7), 'big') / int.from_bytes(b'\xff\xff\xff\xff\xff\xff\xff',
                                                'big')))
    numsemic = 4e6 - numzeros
    assert numzeros >= 0
    assert numsemic >= 0
    if numsemic % 2 == 1:
        numsemic += 1

    # Number of bytes that are part of the image data; we need to ensure all
    # the bytes appear somewhere in the image
    numele = min((numzeros // 2) + (numsemic // 2),
                 int(5e6))
    while True:
        wordlen = max(int.bit_length(numele), 1) + 1
        try:
            numwords = int.from_bytes(
                    rng.read(wordlen), 'big')
        except ValueError as e:
            continue

        if numwords > numele:
            continue

        break

    # Actual size of each "word" in the binary file; these words get compressed
    # by the LZW algorithm later on
    wordsize = 8

    # Constructing the actual image's data requires us to know how many bytes
    # are needed for color table entries (either one or three bytes per entry);
    # note that there is always an even number here since we have both RGB
    # pixels and alpha values
    colortableentrysize = max(int.bit_length(numwords*wordsize) //
                           8, 3) & ~1
    assert not (colortableentrysize % 3)

    # This is the actual size of the whole image after compression by setting
    # all transparent pixels to black
    datasize = ((numele*wordsize+1)*numwords)//colortableentrysize


    # Write the file
    out.write(header)
    out.write(datasize.to_bytes(2, 'big'))
    for i in range(numwords):
        word = int.from_bytes(rng.read(wordsize), 'big')
        out.write(((word << i)&0xFF).to_bytes(wordsize, 'little'))

    for _ in range(numzeros//2):
        out.write(b'\x00\x00\x00\x00')
    for _ in range(numsemic//2):
        out.write(b'\x00\x00\x00;')

    out.write(b';')


