#! /usr/bin/env python3

import sys
from pathlib import Path
from typing import IO

# random number generator `rng` (stdin or /dev/urandom)
def read_random_bytes(
        rng: IO[bytes], n_bytes: int
    ) -> bytes:
    result = []

    while len(result) < n_bytes:
        chunk = rng.read(n_bytes - len(result))

        if not chunk:
            raise IOError("end of file reached")
        
        result += chunk
    
    return b''.join(result)


# write at least `minsize` octets into `out`, preferably more
def fill_file_with_random_bytes(
        minsize:     int,
        maxsize:     int,
        size_factor: float, # >= 1.0
        rndstream:   IO[bytes],
        out:         IO[bytes]
    ):
    totalsize = 0

    try:
        while True:
            chunksize = min([
                max([
                    minsize * size_factor ** (i+1),
                    maxsize]) - totalsize,
                4096])

            buf = read_random_bytes(rndstream, chunksize)

            # avoid partial writes by retrying until complete
            while chunksize > 0 and out.write(buf[:chunksize]):
                totalsize += chunksize

                # reduce buffer size to indicate unwritten bytes
                chunksize -= out.tell()
                out.seek(-chunksize, 1)
                
                continue

    except IOError as e:
        pass


def write_header(out: IO[bytes]):
    print('GIF89a', file=out)

    
if __name__ == '__main__':
    outfile_path = 'image.gif'

    if len(sys.argv) > 1:
        outfile_path = sys.argv[1]
        
    with open('/dev/urandom', mode='rb') as urand, \
         open(outfile_path,    mode='wb') as out:
        write_header(out)
    
        fill_file_with_random_bytes(
            256*1024,       # at least... bytes
            2**16,          # but no more than... bytes
            2.0,            # trying to grow by factor... 
            urand,          # read from this stream
            out             # output is written here
        )