#!/usr/bin/env python3

import argparse
import os
import math
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, args: argparse.Namespace
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Get the logical screen descriptor (LSD) field
    max_width = min(args.max_width, 65536) 
    max_height = min(args.max_height, 65536) 

    width      = math.floor(rng.randint(10, max_width)/10)*10;
    height     = math.floor(rng.randint(10, max_height)/10)*10;

    background_color_index   = rnn.randint(0, 255);
    pixel_aspect_ratio       = rng.randint(0, 255);
    flags                    = 0b10000000; # Global Color Table flag set

    lsd = bytes([
        0x00,                   # Screen Width
        0x00,                   # Screen Height
        0xf4 | ((flags & 1) << 2),        # Flags, Background Color Index
        background_color_index,        # Background Color Index
        0x00,                   # Pixel Aspect Ratio
        width&0xff,             # Screen Width
        (width >> 8)           # Screen Width
    ])
    out.write(lsd)

    # Read global color table length from LSD
    gctlen = 3*(2**((pixel_aspect_ratio + flags * 8) % 8))

    # Add GCT to output if we have it
    if gctlen > 0:
      gct = [ rng.randrange(256) for i in range(gctlen)]

      out.write(bytes(gct))
    
    while True:

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128) == 0):
          return
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # Handle case where there's no image data
        if not hasimgdata:
            continue

        frame_lsd = bytes([
            0x00,                   # Left Position
            0x00,                   # Top Position
            0x00,                   # Image Width
            0x00,                   # Image Height
            0x00,                  # LZW Minimum Code Size
        ])
        out.write(frame_lsd)

        imgdata = [ rng.randrange(256) for i in range(size-5)]

        out.write(bytes(imgdata))