#!/usr/bin/env python3

import struct
from typing import BinaryIO


class RandomGifError(RuntimeError): pass


def read_chunk(in_: BinaryIO) -> bytes:
    try:
        chunklen = in_.read(4)
        if len(chunklen)!= 4: raise EOFError()
        (chunklen,) = struct.unpack('<L', chunklen)
        return in_.read(chunklen)
    except:
        raise EOFError('EOF while reading chunk.') from None


def write_chunk(out: BinaryIO, content: bytes) -> None:
    out.write(struct.pack('<L', len(content)))
    out.write(content)


def inject_extension(in_: BinaryIO, out: BinaryIO) -> None:
    # Read the GIF header and version
    hdr = b''
    while not hdr or hdr[-2:]!= b'IH':
        hdr += read_chunk(in_)

    init_screen = hdr[:-7]
    app_ext = False
    ext = ''
    while not ext or ext[:11]!= b'NETSCAPE2.0':
        ext += read_chunk(in_)

        if len(ext) == 13:
            app_ext = True

    subblocks = list()
    if app_ext:
        subblocks += [b'ANIMEXTS1.0']
        subblocks += [b'\x0b\xff\x01\x00']
    else:
        subblocks += [b'XMP Data XMP ']
        subblocks += [b'http://ns.adobe.com/xap/1.0/\n']
    subblocks += ['<rdf:RDF>\n']
    for blkname, blkdata in [('xmlns:xmp',
                             'http://ns.adobe.com/xap/1.0/'),
                            ('xmlns:dc',
                             'http://purl.org/dc/elements/1.1/'),
                            ('xmlns:crs',
                             'http://ns.adobe.com/camera-raw-settings/1.0/')]:
        subblocks += [f'<{blkname}> {blkdata} </{blkname}>\n']
    subblocks += ['</rdf:RDF>'.encode()]
    blocksize = sum((len(blknm)+5+len(blk)) for blknm, blk
                    in zip(['NETSCAPE', 'ANIMEXT'],
                           [['2.0\x03\x01', '2.0\x0b\x01'],
                            subblocks]))

    subblockhead = b'\x21\xFF\x0b'
    subblock = chr((blocksize >> 8)&0xff).encode()\
               +chr(blocksize&0xff).encode()\
               +subblocks[0]\
               +b'\x00'*(13 - len(subblocks[0])+1)\
               +subblocks[1]

    gct_flag = bool(int.from_bytes(init_screen[-5:-2],
                                   byteorder='little'))
    color_table_size = ((2<<bool(gct_flag))+
                         min(abs(int.from_bytes(init_screen[-2:],
                                                byteorder='little')),
                             16)) * 3
    screen_width = int.from_bytes(init_screen[-15:-11],
                                  byteorder='little')
    screen_height = int.from_bytes(init_screen[-11:-7],
                                   byteorder='little')
    background_color_index = int.from_bytes(init_screen[-6:-5],
                                            byteorder='little') & 0xf
    pixel_aspect_ratio = int.from_bytes(init_screen[-4:-3],
                                        byteorder='little') & 0x03
    flags = bytearray([0])
    flags[0] |= 0x80 if gct_flag else 0
    flags[0] |= 0x0c if color_table_size > 2**8 else \
                0x08 if color_table_size > 2**4 else \
                0x04 if color_table_size > 0 else 0
    flags[0] |= 0x02 if pixel_aspect_ratio >= 2 else \
                0x01 if pixel_aspect_ratio >= 1 else 0

    new_hdr=bytes([*init_screen[:11],
                   *(flags),
                   *init_screen[12:]],
                  encoding='latin-1')
    if gct_flag:
        new_hdr+=init_screen[-2:]
    write_chunk(out,new_hdr)


    while len(ext)>0:
        write_chunk(out, ext[:12])
        ext = ext[12:]
    while len(subblockhead)>0:
        write_chunk(out, subblockhead[:6])
        subblockhead = subblockhead[6:]
    while len(subblock)>0:
        write_chunk(out, subblock[:128])
        subblock = subblock[128:]



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(rng: BinaryIO,
                        out: BinaryIO) -> None:
    read_chunk(rng)    # Skip animation label
    print("Input")
    print(rng.getvalue())
    inject_extension(rng, out)