#!/usr/bin/env python3

from PIL import Image
import numpy as np
import math
import random

# Fill an array with random bytes from /dev/urandom
def fill_rng(arr):
    f = open('/dev/urandom', 'rb')
    arr[:] = [int.from_bytes(f.read(), 'big') for _ in range(len(arr))]
    f.close()


def lzw_encode_bytearray(inbuff : bytearray, outbuff : bytearray):
    max_code_size = 12   # bits per code word
    clear_code    = 1 << max_code_size     # end of transmission marker
    stop_code     = clear_code + 1          # end of information marker

    # Data stream globals
    input_bitbuf  = 0                        # buffer for next read bit
    input_bits    = 0                        # counter for unread bits
    output_wordbuf= []                       # decoded output buffer
    cur_output_word= None                      # current working ouput word


    def get_next_input_bit():
        nonlocal input_bitbuf, input_bits

        if input_bits <= 0: 
            input_bitbuf  = inbuff[0]
            del inbuff[0]

            input_bitbuf |= inbuff[0] << 8;
            del inbuff[0]

            input_bits = 16

        input_bits -= 1
        return (input_bitbuf >> input_bits & 1) == 1 


    def write_output_word(val):
        nonlocal cur_output_word
        cpy = list(cur_output_word+[val])
        while len(cpy)>1 and ((cpy[-1] << 1) | cpy[-2]) < 256:
            cpy.pop(-2);
            cpy[-1] <<= 1


        cur_output_word = cpy

        if len(cur_output_word) >= 255:
            outbuff.append(len(cur_output_word))
            for i in cur_output_word:
                outbuff.append(i)
            cur_output_word = None


    def encode_char(c):
        nonlocal cur_output_word
        if not cur_output_word or cur_output_word[0]!= c:
            write_output_word(c)



    clear_code_bitpattern = "0"*max_code_size+"1"
    assert(len(clear_code_bitpattern)==(max_code_size+1)*2)

    # Clear encoding table state
    outbuff += '\x00\xff'.encode('utf-8')

    # Encode the first clear code
    for bp in map(lambda x: format(x,'0'+str(max_code_size*2)+'b'), clear_code_bitpattern):
       for b in bp:
           encode_char((int)(b=='1'))

    index = dict()      # hashmap for the prefix table
    table = ['']        # actual prefix table, initialized with empty strings

    # The main encoding loop
    while True:
        # Read one more character
        if len(index) > 0xffff:
            raise ValueError("Too many characters!")

        c = chr(get_next_input_bit())

        # Update prefix table
        t = ''.join([table[-1],c])
        if t in index:
            table.append(t[:])
            
        else:
            index[t[:]] = len(table)
            table.append(t[:])

        v = index[''.join(table)]

        # Output LZW code
        if v == stop_code:
            break
        
        elif v == clear_code:
            continue

        write_output_word(v)

        # If we have too much stuff in our buffer, flush it
        if len(outbuff) > 512:
            # Send output buffer
            outbuff.append('\x00'.encode('utf-8'))
            yield outbuff
            outbuff = ''
    
    # Flush the remaining queue
    while len(outbuff):
        outbuff.append('\x00'.encode('utf-8'))
        yield outbuff
        outbuff = ''
    

# Writes an image as an indexed color GIF
def make_indexed_color_gif(img: Image.Image, fname: str):
    w, h = img.size
    inbuff = np.zeros(w * h // 4, dtype=np.uint8)             # input byte buffer
    outbuffs = list([])                                      # output byte buffers 
    rgb_vals = {}                                            # rgb values already seen
    colormap = {v:(k//(2**16), k%(2**16)//2**8, k%2**8) 
                for k,v in enumerate(img.convert(mode='RGB').tobytes()[::3])}
                

    # Get the different RGB colors present in the image
    for k in sorted(colormap.keys()):
        rgb = colormap[k]
        if not rgb in rgb_vals:
            rgb_vals[rgb] = k

    
    index_colors = [(r//(2**8), r%(2**8)//2**8, r%2**8)
                    for _, r in rgb_vals.items()]                 # different colors


    # Write the header
    with open(fname+".tmp", mode="wb") as f:
        f.write("\x47\x49\x46\x38\x39\x61".encode('utf-8'))
        print(struct.pack(">HHHBB", width=w, height=h,
                         palette_size=pow(2, math.ceil(math.log(len(index_colors))/math.log(2))),
                         bg_color=0, aspect_ratio=0))

    # Push the palette size at the beginning fo the file
    tmpname = fname+".tmp"
    os.rename(tmpname, fname)

    # Accepted by other programs
    im = img.copy().convert(mode='P', palette=Image.ADAPTIVE, colors=255)
    im.save(tmpname)


