#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO


# Chunks of data, with an optional name and extra fields
class Data:
    def __init__(self,
                 kind: str, body=b'', name='', **extra):

        self.kind = kind
        self.body = body
        
        if len(name) > 0xff:
            raise ValueError("Name too long!")
        self.name = name.encode()

        self._extra = {}
        for key in extra:
            value = extra[key]

            if not isinstance(value, (str, int)):
                raise TypeError("Extra field must be string or integer!")
            
            # Convert to little endian
            if isinstance(value, str):
                value = [ord(c) for c in value]
            elif isinstance(value, float):
                value = list(struct.pack('<f', value))

            if any([len(v)!= 1 for v in value]):
                raise ValueError("Extra field values must be single characters!")

            self._extra[key] = value

    def write_chunks(self, fobj: BinaryIO):
        length = 0

        # Write header chunk
        fobj.write((self.kind +'').encode())
        length += 1 + len(self.kind)  # Extra space for separator

        # Name
        if type(self.name) == str:
            self.name = self.name.encode()
        else:
            self.name = ''.join(chr(c) for c in self.name).encode()
        fobj.write(struct.pack("<B", len(self.name)) +
                   self.name)
        length += 1 + len(self.name)

        # Length
        fobj.write(struct.pack("<H", length))
        length += 2

        # Add extra fields as sub chunks
        for key in sorted(self._extra.keys()):
            value = self._extra[key]

            # Length
            fobj.write((key + "X").encode()) 
            length += 1 + len(key) + 1    # Extra X character

            # Subchunk content
            fobj.write(struct.pack("<L", len(value)))
            for val in value:
                fobj.write(val.to_bytes(length=1, byteorder="little"))
            length += 4
        
        fobj.write("\x00".encode())   # Terminator

        # Body contents
        fobj.write(self.body)
        length += len(self.body)
    
        return length
    

    
# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    
    # Header - made up of the signature ("GIF"), version 0x87a, and width x height 
    header = Data('IHDR', 
                 struct.pack("<HH", 
                             *[random.randint(10, 1000) for _ in range(2)]),
                     imgwidth=",imgheight=")
    
    # Logical screen descriptor - colors used / pixel aspect ratio
    lsd = Data(' Logical Screen Descriptor',
               struct.pack("<BBHBBBB",
                           *[random.randint(1, 255)] * 6),
               colorres="", 
               gctflag="", 
               sortflag="",
               reserved="",
               sizegct="",
               background="",
               pixelaspectratio="")
    
    # Global Color Table
    if lsd.sizegct < 3:
        numcolors = 2 ** lsd.sizegct
        ct = []
        for i in range(numcolors):
            ct.append(struct.pack("<BBB",
                                  *[random.randint(0, 255)] * 3))
    else:
        ct = ""
        
    # Graphics Control Extension
    gce = Data(' Graphics Control Extension',
               struct.pack("<BBHBB",
                           *[random.randint(0, 255)] * 5),
               transpflag="",
               userinputflag="",
               disposalmethod="",
               delaytime=""
               transparentcolorindex=""
               )
    
    # Image Descriptor
    idd = Data(' Image Descriptor',
               struct.pack("<BBHHHHBB",
                           *[random.randint(0, 255)] * 8),
               left="", top="", width="", height="",
               localcolormap="", interlaceflag="",
               sortflag="", reserved=""
               )

    # Plain text extension
    tetx = Data(' Plain Text Extentsion', 
                struct.pack("<BBHBBHB",
                            *[random.randint(0, 255)] * 6),
                txtleft="", txtttop="", txtgridw="", txtgridh="",
                cellw="", celly=""
                )

    # Application Extention
    appext = Data(' Application Extension', 
                  struct.pack("<BBHBB",
                              *[random.randint(0, 255)] * 3),
                  authcode="", appname=""
                  )
    
    # Comment extension
    comment = Data(' Comment Extentsion', 
                   struct.pack("<BBH",
                               *[random.randint(0, 255)] * 2),
                   comment=""
                   )
    
    # Trailer
    trailer = Data(' Trailer', '')


    # Assume we'll succeed...
    success = True
    try:
        # Let's write the header 
        pos = header.write_chunks(out)

        # LSD follows the header
        pos += lsd.write_chunks(out)

        # CT follows LSD
        out.write(ct)
        pos += len(ct)
            
        # GCE follows CT
        pos += gce.write_chunks(out)
                
        # IDD follows GCE
        pos += idd.write_chunks(out)

        # TETX follows IDD
        pos += tetx.write_chunks(out)

        # APPEXT follows TETX
        pos += appext.write_chunks(out)
    
        # COMMENT EXTENTION FOLLOWS APPEXT
        pos += comment.write_chunks(out)
    
        # TRAILER FINALLY
        pos += trailer.write_chunks(out)

        # Now embed the termination code into the last byte
        out.seek(-1, os.SEEK_END)
        out.write("\x3b\x00".encode())

    except Exception as e:
        success = False
        raise