#!/usr/bin/env python3

import struct
from typing import BinaryIO
import math


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The given width
# and height are used as the logical screen dimensions, and then
# each frame until it is declared terminated by an end-of-image block
# (`2b`) is generated with the specified delay and transparency flag.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, w: int, h: int, 
        delay: float =.1, trans: bool = False
    ):
    # Generate the header and add to output
    header = b'GIF89a'
    out.write(header)

    # Create the logical screen descriptor (LSD) entry
    packed = struct.pack('<HBBHHB', w, h, 0x87, 0, 0, 
        2 ** int(math.ceil(math.log(max(w,h), 2))))
    lsddata = bytes([ord(c) for c in packed])

    # Add LSD to output
    out.write(lsddata)

    # Initialise current colour map
    cmap = [0]*(2**int(math.ceil(math.log(max(w,h), 2))))

    # Generate frames until we get an EOI marker
    while True:
        # Select a background color from the initial palette
        bgcolor = rng.randrange(len(cmap))

        # Convert index to RGB component string
        rbgstr = ('{:03b}'*3).format(*(cmap[bgcolor]))

        # Construct graphics control extension header
        gcehdr = b'\x21\xF9\x04' + bytes([6<<5|trans*1]+[delay&0xFF, 
            (delay//256)&0xFF])+rbgstr+b'\x00'

        # Add graphic control extension to output
        out.write(gcehdr)

        # Read next field byte for block type
        fieldbyte = ord(rng.read(1))
        
        # Check if this is the last block
        if ((fieldbyte & 128)!= 0 or fieldbyte==0x2C):
          break
    
        # Get size of block
        size = ord(rng.read(1))

        # Check if there's any image data in this block
        hasimgdata = bool(((fieldbyte>>5)&3)>0) 

        # Write fields back to stream
        out.write(struct.pack('=B', fieldbyte | 2))
        out.write(struct.pack('=B', size))

        # Handle case where there's no image data
        if not hasimgdata:
            continue

        imgdata = rng.read(size)
        out.write(imgdata)