#!/usr/bin/env python3

import sys
import os
os.chdir(sys.path[0])

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None: 
    # Write header bytes
    # Always 69-byte header
    out.write(b'GIF89a')

    # Int-ify the flags so we can make use of them later
    flags = int.from_bytes(rng.read(1), byteorder='big', signed=False)
    
    # Bool representing if the GIF has a color table or not
    colourtable = bool(flags & 0b10000000)

    # Iterate over each block in the animation as long as there are still frames left
    while True:

        # Create the image descriptor; always 24-bit
        out.write(b'\x2C\x00\x00\x00')
        
        # If this frame has a colour table then append the size
        # of it to the image descriptor. Otherwise just write two
        # zeroes since no such value exists
        if colourtable: 
            tablesize = int.from_bytes(
                    rng.read(1), byteorder='big', signed=False
                )

            out.write((tablesize + 7).to_bytes(1, byteorder='big', signed=False))

        else:
            out.write(b'\x00\x00')

        # Parse the fields in the frame control block
        delay = int.from_bytes(rng.read(2), byteorder='big', signed=False)
        disposalmethod = int.from_bytes(rng.read(1), byteorder='big', signed=False)
        userinput = bool(int.from_bytes(rng.read(1), byteorder='big', signed=False))

        # Append the values to the image descriptor
        out.write(delay.to_bytes(2, byteorder='big', signed=False))
        out.write(disposalmethod.to_bytes(1, byteorder='big', signed=False))
        out.write(userinput.to_bytes(1, byteorder='big', signed=False))

        # Create the local colour table by generating three unique RGB triples. This is done 
        # because many programs will only support up until 256 colours anyway. Also note that since
        # these will be written little endian (least significant byte first) they'll need reversing
        # before being appended to the image descriptor
        colourtable = [
                b'\x00\x00\xFF' + rng.read(3), 
                b'\xFF\x00\x00' + rng.read(3), 
                b'\x00\xFF\x00' + rng.read(3)]

        # If the GIFFLAG_COLOURTABLE flag is set, the order of
        # appearence should be reversed
        if flags & 0b10000000:
            colourtable.reverse()

        # Concatenate all the colour table entries together
        colourtable = b''.join(colourtable)

        # Calculate the length of said entries and add
        # that to their respective fields in the image descriptor
        colortablelength = len(colourtable)
        remainingcolortablespace = 256 - colortablelength

        out.write(remainingcolortablespace.to_bytes(1, 'little', signed=False))
        out.write(b'\x00\x00')

        # Write the colour table to the end of the stream
        out.write(colourtable)

        # Read the field telling us how long the LZW codestream is
        lzwcodestreampriority = int.from_bytes(
                rng.read(1), byteorder='big', signed=False)
        lzwcodestreamsize = int.from_bytes(
                rng.read(2), byteorder='big', signed=False)

        # Check if the GIF contains either a plain text extension or
        # an application extension box. Plain text extensions have priority but
        # if neither one is found, then create an application extension with something
        # funnily named to remind everyone that it's a fake
        if ((lzwcodestreampriority == 0x01) or 
            (lzwcodestreampriority == 0xF9)):
            
            # Read the contents of the text extension. According to the spec, this should
            # start right after the type indicator and the block size
            lzwcodestream = rng.read(lzwcodestreamsize)
            
        elif ((lzwcodestreampriority == 0xFE) or 
              (lzwcodestreampriority == 0xFF)):
            
            # Read the contents of the application extension according to the spec
            name = rng.read(8)
            authcode = rng.read(3)
            applicationidentifier = rng.read(11)
            applblocksize = int.from_bytes(
                    rng.read(2), byteorder='big', signed=False)

            lzwcodestream = b''
            
            # If its a plain text extension then read the plain text payload
            if (applicationidentifier == b'tEXt'):

                # Note that we're reading the entirety of the plain text extension block here
                # which means that even though it may not actually contain any plain text content,
                # we'll still be wasting memory space on this variable if it does
                lzwcodestream = rng.read(applblocksize)
                
                # Make sure that our plain text block ends properly
                if (lzwcodestream[-1:]!= b'\x00'): 
                    raise Exception('Plain text block did not end properly!')
                    
            # The amount of metadata contained inside the application block
            # should equal 11 plus the block size itself
            assert len(name) + len(authcode) + len(applblocksize) + \
                   len(applicationidentifier) + len(lzwcodestream) == 26
                
            # Make sure that our application block ends properly
            if (lzwcodestream[-1:]!= b'\x00'): 
                raise Exception('Application block did not end properly!')
                
        else:
            # Define a dummy application identifier
            applicationidentifier = b'NOPE!'

            # As above, we're reading the entirety of the plain text extension block here
            # which means that even though it may not actually contain any plain text content,
            # we'll still be wasting memory space on this variable if it does
            lzwcodestream = b'\x0A\x0B\x0C\x0D' + \
                            rng.read(lzwcodestreamsize - 4)

            # Make sure that our plain text block ends properly
            if (lzwcodestream[-1:]!= b'\x00'): 
                raise Exception('Plain text block did not end properly!')

        # Write the block size and priority level to the beginning of the block
        out.write((lzwcodestreamsize + 1).to_bytes(
                2, 'big', signed=False))
        out.write(lzwcodestreampriority.to_bytes(
                1, 'big', signed=False))

        # Write the actual LZW codestream
        out.write(lzwcodestream)

        # Break out of the loop once we've finished writing all the blocks
        # TODO: Currently there isn't anything preventing the generation of malformed
        #       files where blocks do not follow the standard format. Add sanity checks!
        if flags & 0b01000000: break