#!/usr/bin/env python3

import argparse
import io
import os
import struct
from typing import (BinaryIO, List, Optional, Tuple)


CAPTIONS_SIZE = {
    1: 54,
    2: 73,
    3: 92
}

# Returns true if the byte is not zero or null.
def nonzero(byte: bytes):
    return byte!= b'\x00' and ord(byte[0]) > 1


# Counts the amount of consecutive zeros at the end of a block.
def countTrailingZeros(blockBytes: bytes) -> int:
    i = 0
    while i < len(blockBytes) - 1 and blockBytes[-i-1] == 0:
        i += 1

    # If we have reached the start of the array then all bytes are zeros
    return i if i >= len(blockBytes) else 0


# Removes all consecutive trailing zeros from a block until the length is valid.
def removeTrailingZerosFromBlock(blockBytes: bytes) -> bytes:
    totalLength = CAPTIONS_SIZE[len(blockBytes)]
    i = min(totalLength - 1, blockBytes.index(0)+1)
    return blockBytes[:i].ljust(totalLength, b'\x00')


# Reads a block containing either one or three sub blocks. The first sub block contains 8 bytes in this order:
# * Packed field: This field always has bit 7 set to 1, bits 6-4 are the size of the second block, and bits 3-0
# are the size of the first block. Size of the second block can be 0 to 7 which specifies that there is no second block.
# * Animation control flags: The packed field also sets these flags for each frame. This lets you specify whether an
# animation should loop infinitely on the last frame, whether it plays once forwards, backwards, pingpong, etc.
# * Delay timer: How long after the previous frame should the animation wait before showing the next frame. This value must
# be between 0 and 255 inclusive. Once again, the packed field controls how the delay value is applied; values of 0-255
# mean milliseconds but values of 256+ mean times of 10ths of a second.
# Following that are two lengths of image data that follow as follows:
# * Frame left position: If the first sub block is larger than 12 bytes, this value indicates where the frame is located
# relative to other frames in the animation. A value of 0 means the frame will remain stationary relative to any others.
# Otherwise the frame is assumed to begin at point (0, 0).
# * Frame top position: Similarly to above, this value tells us what pixel the bottom edge of the frame starts on. Again,
# a value of 0 means that the bottom edge will coincide with the bottom edge of the screen.
# Finally comes the actual frame itself. This is simply an uncompressed bitmap so its length cannot exceed 2048 pixels wide
# by 256 pixels high.
# If the third block exists then it encodes a caption message for the end of the GIF. It's just raw text since the caption
# may contain arbitrary characters. Its length may not exceed 79 bytes (19 lines x 80 chars plus the line ending character
# CR) and any unused bytes past 79 must be empty.
# Gif frames are stored as blocks of bytes in little endian format (least significant bit first). The current version only
# supports LZW encoded images and uses the default setting for DCT coefficients.
# https://www.w3schools.com/graphics/lgif_format.asp
def readBlock(inFile: BinaryIO) -> Tuple[bool, bool, int, int, int]:
    packedField = int.from_bytes(inFile.read(1), 'little')
    active = ((packedField & 0x80) >> 7) == 1
    disposalMethod = ((packedField & 0x70) >> 4)
    durationModifier = ((packedField & 0xF))

    numSubBlocks = 1 if disposalMethod == 0 else 3
    delay = None
    width = None
    height = None
    offsetX = 0
    offsetY = 0
    captionData = None

    for _ in range(numSubBlocks):
        blockTypeAndLength = int.from_bytes(inFile.read(1), 'little')

        typeCode = blockTypeAndLength & 0xE0
        length = blockTypeAndLength & 0x1F
        
        if typeCode == 0xC0:
            assert not delay is None

            duration = int.from_bytes(
                    inFile.read(length), 'little')

            delay = 256*delay + duration
            
            continue
        elif typeCode == 0xD0:
            width = int.from_bytes(
                    inFile.read(length), 'little')

            continue
        elif typeCode == 0xE0:
            height = int.from_bytes(
                    inFile.read(length), 'little')

            continue
        elif typeCode == 0xF0:
            offsetX = int.from_bytes(
                    inFile.read(length), 'little')

            continue
        elif typeCode == 0x00:
            offsetY = int.from_bytes(
                    inFile.read(length), 'little')

            continue
        elif typeCode == 0xFE:
            captionData = inFile.read(length)
            break
        else:
            raise ValueError('Invalid block type! %d' % typeCode)

    assert not width is None
    assert not height is None
    
    return (active, disposalMethod, delay, width, height, offsetX, offsetY, captionData)

# Removes padding space from a string.
def stripStringPadding(string: str) -> str:
    s = ''
    pad = False

    for c in string:
        if c == '\x00':
            if not pad:
                s += '\\x00'
            continue
        elif c == '\r':
            s += '\\r'
            pad = True
            continue

        pad = False
        s += c

    return s


# Reads the APNG chunk named `name`.
def readChunk(inFile: BinaryIO, name: str) -> bytes:
    id = inFile.read(4)
    length = int.from_bytes(inFile.read(4), 'big')
    data = inFile.read(length)
    crc32 = inFile.read(4)

    assert len(id) == 4
    assert len(crc32) == 4

    assert id == name.encode()
    assert crc32 == b'\xff\xff\xff\xff'

    return data


# A map of image meta data including a GIF block byte stream.
class ImageMetadata():
    def __init__(self, active: bool, disposalMethod: int, delay: int,
            width: int, height: int, offsetX: int, offsetY: int,
            imageData: bytes, captionData: Optional[bytes]):
        self.__active = active
        self.__disposalMethod = disposalMethod
        self.__delay = delay
        self.__width = width
        self.__height = height
        self.__offsetX = offsetX
        self.__offsetY = offsetY 
        self.__imageData = imageData
        self.__captionData = captionData

    @property
    def active(self):
        return self.__active

    @property
    def disposalMethod(self):
        return self.__disposalMethod

    @property
    def delay(self):
        return self.__delay

    @property
    def width(self):
        return self.__width

    @property
    def height(self):
        return self.__height

    @property
    def offsetX(self):
        return self.__offsetX

    @property
    def offsetY(self):
        return self.__offsetY 

    @property
    def imageData(self):
        return self.__imageData

    @property
    def captionData(self):
        return self.__captionData

    def __str__(self):
        return """
Active:         {}
Disposal Method:   {}
Delay:        {}
Width:          {}
Height:         {}
Offset X:       {}
Offset Y:       {}
Caption Data:   '{}'""".format(
                self.__active, self.__disposalMethod, 
                self.__delay, self.__width, 
                self.__height, self.__offsetX, 
                self.__offsetY, self.__captionData) 


# Parses an APNG file called `filename`, returning a list of images embedded within.
def parseAPNGFile(filename: str) -> List[ImageMetadata]:
    file = open(filename, "rb")

    # Read the main PNG header
    signature = file.read(8)
    assert signature == b"\x89PNG\x0D\x0A\x1A\x0A"

    # Read the IHDR chunk
    ihdrId = file.read(4)
    assert len(ihdrId) == 4
    assert ihdrId == b"IHDR"
    ihdrLen = int.from_bytes(file.read(4), 'big')
    ihdrCRC = file.read(4)