#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng`.
# The size of the generated image is given by `size`, and its color depth
# is given by cdepth (in bits per pixel).
# cdepth must be between 1 and 16 inclusive!
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, size=(400, 400), cdepth=5
    ):
    if not (1 <= cdepth <= 16): raise Exception('unsupported color depth!')

    # Header is always "GIF89a"
    header = b'GIF89a'

    # The logical screen descriptor consists of fixed values followed by two
    # fields in byte order which encodes width and height respectively;
    # these are stored as unsigned shorts big endian.
    lsd = bytearray()
    lsd += 0xF7			# Screen width (max 65535 pixels)
    lsd += 0xFF & ((size[0]) >> 8)    # Screen width (MSB)
    lsd += (((size[0]) & 0xff))      # Screen width (LSB)
    lsd += 0xFF & ((size[1]) >> 8)    # Screen height (MSB)
    lsd += (((size[1]) & 0xff))      # Screen height (LSB)
    lsd += (cdepth - 1).to_bytes(1, 'big')     # Color depth
    lsd += bytes([0b10 | 0b01100   # Global flags
                 ])                   # Background color index (always black)
    lsd += bytes([0b00               # Pixel aspect ratio
                 ])
    assert len(lsd) == 7

    # Finally we have the Graphics Control Extension block with global parameters.
    gce = bytearray([
        0x21,                         # Extension introducer
        0xf9,                         # Graphic control label
        0x04,                         # Block length
        0x00,                       # no disposal specified
        0x00,                        # Transparent color index
        0x00       # No delay given
    ])

    # Now comes the Image Descriptor. It contains the following parameters:
    id = bytearray([
        0x2C,                     # Image separator
        0x00, 0x00,               # Left position
        0x00, 0x00,               # Top position
        0xFF & ((size[0]) >> 8),           # Width (MSB)
        0xFF & (size[0]),             # Width (LSB)
        0xFF & ((size[1]) >> 8),           # Height (MSB)
        0xFF & (size[1])              # Height (LSB)
    ])

    # Finally put it all together
    out.write(header + lsd + gce + id)


    # Now comes the image data...
    # First find the freshly filled GIF file's end offset, so that we can jump there after
    out.flush()
    pos = out.tell()
    out.seek(0, 2)
    endpos = out.tell()
    out.seek(pos + 2, 0)                # Skip past header

    while True:
        if out.tell() >= endpos: break
        for i in range(int((endpos - out.tell()) / 4)):
            b = None
            while b is None or b > 0xbf:
                b = int.from_bytes(
                        rng.read(1), 'big')

            out.write(b.to_bytes(1,'big'))