#!/usr/bin/env python3

import sys
import os
os.chdir(sys.path[0])

import struct
from typing import Generator, Tuple

# Generates a list of frames with width w and height h
def framelist(w:int, h:int) -> Generator[Tuple[int, bytes], None, None]:
    # The initial color table contains black (#000000) as foreground, white (#ffffff) as background
    yield 1 << (7+1), b'\x00\x00\x00\xff\xff\xff' + bytearray([i for i in range(256-6)])

    # Create a new frame.
    frame = []

    # Add each pixel
    for y in range(h):
        frame += [y]
        for x in range(w):
            # Choose a random color from the previous palette
            col = rng.randint(4, len(palette))

            # Choose whether this pixel is transparent or not using probability 1/3
            if col == 4 or rng.randint(0,2)==0:
                # If it's transparent, choose any color except the last one (transparent)
                frame.append(col - 1)
            else:
                # Otherwise use the selected color
                frame.append(col)

        while len(frame)%w!= 0:
            frame.append(-1)

    yield 0, bytes(bytearray(frame))


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, w:int=None, h:int=None
    ):
    global palette, Ratio

    # Determine dimensions
    if w is None:
        w = rng.randint(32, 128)
    if h is None:
        h = rng.randint(16, 128)

    # Calculate the aspect ratio
    Ratio = 1 / (6 * w // h)

    # Write the header to the output stream
    out.write(b'GIF89a')

    # Determine how many frames we'll need
    maxframes = int((rng.randrange(64)+1)*Ratio)
    print("we'll need",maxframes,"frames")

    # Write a bunch of frames to the output file
    nframes = 0
    palette = {}
    trywhile = True
    while trywhile:
        try:
            for frameno, data in framelist(w, h):
                palette.setdefault(frameno, []).extend(data)

                # Write the image descriptor chunk
                out.write(struct.pack("<BBHHHH",
                        ord('!'), # Extension type
                        0xf9    , # Subtype: Graphic control
                        0      , # No flags
                        0      , # X position
                        0      , # Y position
                        w      , # Width
                        h        )) # Height
                out.write(struct.pack("<B", frameno | ((nframes & 1) << 7)))
                nframes += 1

                # Write the graphics control extension block
                out.write(struct.pack("<BBBBBBBHB",
                        249     , # Extension type (graphic control)
                        4       , # Size
                        0       , # Flags
                        0       , # Delay time (unused)
                        0x2c    , # Transparent color index
                        0       , # Block terminator
                        0       , # Block terminator
                        0x2c    , # Block terminator
                        0))       # Block terminator

                # Check if we have enough frames yet
                if nframes >= maxframes:
                    raise StopIteration()

                # Write the image data chunk
                out.write(struct.pack("<BBH",
                        ord(','), # Image separator
                        0x2c   , # LZW code size (= 8bpp)
                        len(data))) # Length of the following data
                out.write(data)
        except StopIteration:
            trywhile = False

    # Now write our palette at the end of the file. We write the length first, then all colors.
    # Note that even though the spec says there must be no duplicates, reality requires us to do so.
    psize = sum([(len(v)-1)*3 + 1 for v in palette.values()]) + 1
    out.write(psize.to_bytes(2,'little') + b'\x00')
    for p in palette.values():
        assert(len(p) > 1)
        p = sorted(set(p[:-1])) + [-1]
        out.write(bytes(bytearray(p)))