#!/usr/bin/env python3

import rnglib

# generates a random GIF file into the open file object `out`.  The
# `length` is in bytes; there's no guarantee that each frame will be
# exactly this long (but it should be longer than any reasonable image
# might be).
def make_image(
        length: int, out: BinaryIO=None, verbose: bool=False):
    if not out:
        out = open('tmp.gif', 'wb')

    gifer = rnglib.SimpleRNG()    # instance of RngFile wrapper for urandom
    blocksize = 255              # maximum bytes per record
                                # note: the spec limits to 254

    version = "GIF89a"           # begin GIF file with magic gif string
    w = 1024                    # generated images are square
    h = 1024
    bgcolor = 0xffffff          # background color: white
    aspect = 0                  # pixel aspect ratio: none

    records = []                # list of byte arrays making up entire file
    nframes = max(int((length + blocksize - 1) / blocksize), 1)   # at least one frame

    def flush():                 # write one or more blocks
        nonlocal records         # use local vars from outer scope
        while len(records) > 0:   # block size limit?
            sub = min(blocksize, len(records[0]))      # truncate frame to buffer size
            buf = records[0][:sub]                     # grab first part of frame
            del records[0][:sub]                      # slice off first part
            if len(buf) == blocksize:                # full frame?
                out.write(buf[:3])                   # write block size & bytes
                out.write(buf[3:])
                continue                               # go around again

            # partial frame: pad and add terminator
            padding = blocksize - len(buf)            # count remaining space left in current block
            buf += bytes([padding]) * padding        # pad it with nulls
            buf += b'\x3b'                           # add trailing terminator
            assert len(buf) == blocksize             # check we got right size
            out.write(buf[:3])                       # write block size & bytes
            out.write(buf[3:])
            break                                     # all done!

    # compute global palette with two colors
    pal = [bgcolor, 0x000000]                          # define palette
    out.write(version.encode())                        # write magic string
    out.write(struct.pack('<HHBBBHB', w,h,             # logical screen descriptor
                         8,            # packed field: global colormap
                         7,
                            bgcolortable,0,0))     # background color index
                                                            # aspect ratio
    out.write(bytes(len(pal)))                         # write palette
    for i in range(0, len(pal)):
        out.write(struct.pack('<H', pal[i]))
    out.write(b'\x00')                                # zero-terminator
    chunk = []                                        # start empty
    for f in range(nframes):
        chunk.append(b'\x2c')                          # graphics control
        chunk.append(struct.pack("<BBHHHBB",
                            0xf9,             # discriminator code
                            4,
                            0,                # delay time (1/10 sec)
                            0,0,               # leave transparent
                            0))               # end graphic control

        chunk.append(b'\x21')                          # extension block introducer
        chunk.append(b'\xff')                          # application extension
        chunk.append(b'\x0b')                          # app extension label
        chunk.append(b'text\x00')                      # app name
        chunk.append(b'\x03')                          # text encoding scheme: UTF-8
        chunk.append(b"\x9B")                          # JIS X 0208 character set
        chunk.append(struct.pack("Bxxx",
                    len(str(f+1))))                  # how many chars to display
                                                    #   (we'll overwrite the last char later)

        chunk.append(str(f+1).encode('utf_8'))[-1:]  # replace final char with frame number
        chunk.append(b'\x00')                          # end label

        lzwmincodesize = 2                             # minimum LZW code size
        chunk.append(b'\x21')                          # extension introducer
        chunk.append(b'\xf9')                          # graphic control label
        chunk.append(b'\x04')                          # graphic control block size
        chunk.append(struct.pack("<BBHHHBB",
                            1+(lzwmincodesize<<1),    # maximum size + flags
                            0,                # no transparancy
                            0,0,               # delay time (1/10sec)
                            0,                # no transparancy
                            0))               # end graphic control
        chunk.append(struct.pack("<BBB",
                            3,                 # 3 bitfields: x, y, and pixel depth
                            ((w - 1) >> 5)-1,  # 3 lsb's of width
                            ((h - 1) >> 5)-1 )) # 3 lsb's of height
        chunk.append(b'\x00')                          # black background
        chunk.append(b'\xff')                          # image separator
        chunk.append(b'\x0b')                          # image label
        # build uncompressed LZW data stream
        lzwdata = gifer.some_bytes(lzwmincodesize + 1)
                                   # guaranteed to have at least one byte after first pass
        while True:                                   # repeat until all codes defined
            try:                                       # try next symbol in table
                val = lzwtable.index(
                        lzwdata, lzwmincodesize)       # get value, or add as new entry
                lzwdata.append(val)                    # add actual value to encoded data
                matchlen = max(matchlengths[val], 1)   # get match length of current symbol
                lzwdata += gifer.some_bytes(matchlen)  # add new node's match values
                continue                              # keep going
            except ValueError:
                break                                  # stop when the table gets full
            finally:                                    # always do this
                lzwtable.append(tuple(lzwdata[-matchlengths[val]:]))
                                                        # reclaim unused code space
        symbits = sum(map(lambda x: bin(x)[2:].count('1'),
                            lzwtable))           # count bits used by symbols
        symbits += lzwmincodesize + 1                # include initial code size + EOI marker
        # Pack LZW data stream into chunks of 255 bytes each.
        # Any leftover bytes are packed into an extra chunk then terminated
        chunk.append(struct.pack('<BBBB',
                                1,            # LZW compression method
                                lzwmincodesize,
                                (symbits//8)-2, # 2 bytes per codepoint
                                    (symbits%8)+1))  # 1 bits required to encode remainder
        while len(lzwdata) > 0xffff:
            chunk.extend(struct.pack("<H", 0xffff) + lzwdata[:0xffff])
            lzwdata = lzwdata[0xffff:]
        else:
            chunk.extend(struct.pack("<H", len(lzwdata)) + lzwdata)
        chunk.append(b'\x00')                          # clear code
        # add a code definition for every unique symbol
        for s, t in enumerate(lzwtable[:-1]):
            chunk.append(struct.pack('<BB',
                                    0xfe,         # symbol definition
                                    len(t)))      # length of match sequence
            chunk.extend(t)                            # append match sequence
        # finally, add the EOI marker
        chunk.append(struct.pack('<BB',
                                0xff,         # EOI marker
                                0x00))
        flush()                                      # write a block group
    # done building LZW stream: pad it and terminate
    if len(chunk) % 2!= 0:
        chunk.append(b'\x00')
    flush()                                          # write last block
    out.flush()                                       # push everything out


if __name__ == "__main__":
    make_image(
        2**20,
        open('/home/cat/Desktop/vga.gif', 'wb'))