from io import BytesIO as StringIO
from typing import Tuple, List


# Generates an entire GIF file in memory and returns it as a string of bytes.
def generate_random_gif() -> Tuple[int, str]:

    # Define constants to be used throughout the program.
    b = lambda x: x if isinstance(x, bytes) else bytes([x])         # Convert integer to byte.
    w = lambda n: int(n).to_bytes(length=n.bit_length(), byteorder='little', signed=False)   # Convert integer to little-endian word (4 bytes).

    # Generate the image data buffer with width=512, height=512.
    testdata = StringIO()
    for y in range(0, 512):
        rowstr = ""
        for x in range(0, 512):
            pix = ((x ^ y) & 0xFF) % 256     # XOR modulus 256. Seems pretty neat!
            rowstr += "%02X" % pix          # Append hex-formatted value.

        # Every row is padded at the end by an extra null character (\x00).
        testdata.write(w((y*512)) + bytes.fromhex(rowstr + "00"))

    # The color map size must be a power of two between 2 and 256 inclusive. We will use 256.
    colormapsize = 7        # Number of colors minus one (i.e., how many entries are there excluding transparency). Must fit within byte.
    cmapbits = max(1, min(colormapsize - 1, 8)).bit_length()     # Bits required to store the number of colors. Always a minimum of 1 bit.
    cmapsize = pow(2, cmapbits)      # Actual number of colors we'll need to store all possible colors.
    assert cmapsize >= colormapsize   # Should always be true.

    # Calculate the image descriptor block. The first field specifies the left offset. I don't know what this means in terms of pixels.
    # The second field defines whether or not there's transparency. If so then we can just set this to 0 since that only applies if
    # we have more than 256 colors. The third field is the top offset. Again, this doesn't make any sense. The fourth field defines
    # the global color table size; that is, if we want to access individual colors other than those available via the palette. This
    # field should be zero unless we are also using transparent colors which requires us to provide a local color table later on.
    imgdescblock = w(0) + \
                   b(0) + \
                   w(0) + \
                   w(cmapsize)       # Global Color Table Size.

    # Construct the Logical Screen Descriptor. Note that we're assuming the screen width and height are both fixed at 512. Also note that
    # local palettes aren't supported but this might change in the future. See http://www.matthewflickinger.com/lab/whatsinagif/lsd.
    # We specify no background color index, pixel aspect ratio = 0, sorting method = 0, global color table flag set to True.
    lsd = b('GIF89a') +                # Signature ('GIF89a').
          b('\x00\x00') +              # Screen Width (LSD).
          b('\x00\x00') +              # Screen Height (LSD).
          b('\xF0') +                  # Packed Field -- no color resolution specified, 8 bits per sample, color table sorted and interlaced.
          imgdescblock                 # Image Descriptor Block.

    # Construct the main color table by specifying each RGB triplet plus the color being transparent. For now we will assume that
    # whatever viewer loads our GIF file will handle transparency properly without us having to do anything special.
    ctable = []    # Main color table.
    for i in range(cmapsize):
        r = (i * 37) % 256
        g = (i * 19) % 256
        b = (i * 13) % 256
        ctable.append(r)
        ctable.append(g)
        ctable.append(b)
        ctable.append(0)                    # Transparency.

    # Now construct the GCT by adding the indicated length prefix followed by the values. Note that the numbers we pass here
    # are actually bytes rather than integers because Python 3 has decided to treat bytes as immutable. So we coerce them back to ints.
    gct = b([cmapbits])[0] + bytes(map(lambda i: i if isinstance(i, int) else ord(i), ctable))

    # Finally, assemble everything together into the final result.
    return len("".join(list(map(lambda x: chr(x) if isinstance(x, int) else x,
                          lsd +                          # Logical Screen Descriptor
                          b(len(gct)) +                   # Length Prefix for Global Color Table Extension
                          gct +                           # Global Color Table Extension
                          b(len(imgdescblock)) +         # Length Prefix for Image Descriptor Extension
                          imgdescblock +                 # Image Descriptor Extension
                          testdata.getvalue())))),     # LZW compressed image data.