#!/usr/bin/env python3


from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). This follows
# the official specification.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # Create a simple utility function that writes bytes to
    # an output stream with padding.
    def writePad(bs: bytes) -> None:
        out.write(b'\x21')
        pad = (4 - len(bs) % 4)%4
        out.write(bs + bytes([pad]*pad))

    # The minimum length of a valid GIF file
    minlen = 6

    # Initialize buffer array
    buf = []

    # Read first chunk until it's bigger than minlen
    while True:
        c = rng.read(minlen+1)
        if len(c) < minlen or ord(c[:3])!= minlen-1:
            raise Exception("GIF too short")
        buf.append(c)
        break

    # Write everything up to first control code to output
    writePad(buf[0][3:])

    # Update status variables
    remaining = int.from_bytes(buf[-1][3:], 'little')
    written   = len(buf[0]) - 3
    seq       = False

    # Write everything after the first control code as long as
    # there are still more blocks
    while remaining > 0:
        # Choose next block type randomly
        typ = rng.read(1)[0] % 4

        # If we've chosen this one before and wrote some data since then
        # switch types once again 
        if typ == 0 and seq:
            typ = 1
        elif typ == 1 and not seq:
            typ = 0
        else:
            seq = False

        # Get next block size
        if typ > 0:
            size = 2 ** (typ + 1)
        else:
            seq = True
            size = 2 ** (-typ + 1)

        # Read block data
        data = b''
        while len(data) < size:
            c = rng.read(max(1, min(remaining, size-len(data))))
            if len(c)==0:
                print("\n\nFinished! We are missing {} more "
                      "blocks".format(-remaining), flush=True)

            # Keep reading until we get enough data
            data += c
            
        # Write it all to output
        writePad(data)

        # Update counters
        written    += len(data)
        remaining -= len(data)