#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out`. If the "seed" argument is given it will be used to seed /dev/urandom, otherwise (or if it's not an integer) it defaults to the current time
def generate_random_gif(
        seed=None, out:BinaryIO=sys.stdout):

    if type(seed) == str and seed.isdigit():
        seed = int(seed)
    elif type(seed)!= int:
        seed = int(os.path.getmtime(__file__))

    with open('/dev/urandom', 'rb') as rng:

        # Header field (always GIFF89a I guess)
        out.write('GIF89a'.encode())

        # Data length. We'll just use a fixed 54 bytes but this could obviously be changed later on
        out.write((len(struct.pack('<HHbbbbBBHB', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0))).to_bytes(2, byteorder='big'))

        # Image descriptor block
        out.write(struct.pack('<HHbbbbBBHB', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0))


        # Fallback palette. This is how most programs do things so let's keep that consistent
        defaultPalette  = [(c[0], c[1], c[2]) for c in [(0, 0, 0),(255, 255, 255)]]
        defaultPalette += [(c[0], c[1], c[2]) for c in [(0, 0, 255),(0, 255, 255),(0, 255, 0),(255, 255, 255),(255, 0, 0),(255, 0, 255),(255, 255, 0),(255, 255, 255),(127, 127, 127),(0, 255, 127)], ]
        
        # Add a background color
        defaultPalette += [(int.from_bytes([c//2], 'little'),
                            int.from_bytes([c//2], 'little'),
                            int.from_bytes([c*2], 'little'))
                           for c in range(5)]

        
        # Add a foreground color
        for _ in range(3):
            for i in range(0x00, 0xff+1):
                defaultPalette += [(i, i, i)]

        
        # Add more fallbacks if necessary, they're going to look ugly though
        while len(defaultPalette) % 256!= 0:
            defaultPalette += [(c[0], c[1], c[2])]



        # Color table. Start by adding less than 256 entries from our default palette.
        colorTableSize = min(max(0, 1<<int.from_bytes(rng.read(1),'big')), len(defaultPalette)//2-1) *2
        
        
        colorTableData = []
        startColorIndex = max(1, 2**int.from_bytes(rng.read(1),'big'))
        endColorIndex =   max(startColorIndex, min(colorTableSize // 2 - 1,
                                                   len(defaultPalette)//2-1))
        colorTableIndexesToUse = sorted(set([(startColorIndex-index)%
                                               (endColorIndex-startColorIndex)+startColorIndex
                                               for index in range(colorTableSize // 2)]))
                                            
        for index in colorTableIndexesToUse:
            
            red   = int.from_bytes(rng.read(1),'big')
            green = int.from_bytes(rng.read(1),'big')
            blue  = int.from_bytes(rng.read(1),'big')

            r = min(red,  255)
            g = min(green,255)
            b = min(blue, 255)
            
            packed = ((r & 0xF8)<<8)|((g & 0xFC)<<3)|((b & 0xF8)>>3)
            
            colorTableData.append(packed)
            try:
                colorTableData.append(packed)
            except TypeError:
                pass

        # Make sure we have exactly 256 entries now that we've added some if needed
        assert len(colorTableData) % 2 == 0 and len(colorTableData) >= 256
            
        # Write all the data to output stream
        out.write((bytes(colorTableData)))
        

        # Transparent color indicator
        transparentFlag = bool(int.from_bytes(rng.read(1),'big'))

        

        # Image descriptor (again)
        width    = int.from_bytes(rng.read(2),'big')
        height   = int.from_bytes(rng.read(2),'big')
        left     = int.from_bytes(rng.read(2),'big')
        top      = int.from_bytes(rng.read(2),'big')
        localColorMapSortFlag = bool(int.from_bytes(rng.read(1),'big'))
        interlaceFlag        = bool(int.from_bytes(rng.read(1),'big'))
        localColorMapSize    = int.from_bytes(rng.read(1),'big')
        
        # LZW minimum code size
        lzwMinimumCodeSize = int.from_bytes(rng.read(1),'big')

        # Make sure it has a valid value
        lzwMinimumCodeSize = max(min(lzwMinimumCodeSize, 12), 2)

        # Make sure that the image dimensions are valid
        width  = max(width, 2)
        height = max(height, 1)
    
        # Bits per pixel
        bitsPerPixel = int.from_bytes(rng.read(1),'big')

        # Number of colors actually used (>= 1)
        numColorsUsed = 2**(bitsPerPixel+1)

        # Packed fields
        packedFields = int.from_bytes(rng.read(1),'big')

        # Get the global color map flag and get rid of the last three unused bits
        globalColorMapFlag = (packedFields>>7)&1
        packedFields &= ~ (1<<7 | 1 << 6 | 1 << 5)



        dataBlocks = []

        # We need at least one data block here
        (nextByte,) = rng.read(1)
        if nextByte!= 0:
            dataBlocks.append(nextByte)

        remainingBytesInBlock = None

        while True:
                
            chunk = rng.read(1)            
            dataBlockSize = chunk[0]

            # At EOF, break out of loop! 
            if not chunk or dataBlockSize==0:
                break


            # End of line code?
            if dataBlockSize == 0xfe:
                
                if remainingBytesInBlock:
                    raise Exception("This shouldn't happen")

                if len(dataBlocks):
                    raise Exception("This should never happen either!")

                remainingBytesInBlock = 8
                continue

            if remainingBytesInBlock:
                dataBlockSize = min(remainingBytesInBlock, dataBlockSize)
                remainingBytesInBlock -= dataBlockSize
            else:
                dataBlockSize = min(0xfff, dataBlockSize)

            # Read data
            dataBlocks.append(chunk[:dataBlockSize+1])



    # Now write the various headers...
    head = [
            0x21,                                  # Extension introducer

            0xff,                                  # Application extension marker

            0x0b,                                  # Block length: 11
            0x08,                                  # Application identifier length (always 8)
            ord('A'),ord('n'),ord('i'),ord('M'),ord('e'),ord('t'),\
            ord('a'),0x00                      ,

            0x00,                                  # Subblock length: 0, no meaning

            0xf9,                                  # Graphics control label
            0x04,                                  # Block length: 4
            0x00,0x00,                             # No transparancy
            0x00,                                  # Delay time: 0
            0x00                                   # Transparent color index: nonexistent
        ]


    # Define the logical screen descriptor
    head += [
            0x2C,                                  # Image separator
            
            0x00,                                  # Left position
            0x00,                                  # Top position
            0x00,0x00,0x00,                        # Width
            0x00,0x00,0x00,                        # Height
            0x08,                                  # Local color table size: 2^(bitsPerPixel+1)
            0x00,                                  # Sort type: none
            0x00,                                  # Global color table flag: no
            0x00,                                  # Background color index: n/a
            0x00,                                  # Pixel aspect ratio: n/a
        ]

    # Add the header and the data blocks to output
    out.write(head)
    for datablock in dataBlocks:
        out.write(datablock)

    print('Done!')