#!/usr/bin/env python3

import struct
from pathlib import Path
from typing import BinaryIO


def write_image_header(
        nframes: int, width: int, height: int, localcolor: bool,
        globalcolor: bool, interlaced: bool, tablesize: int,
        backgroundidx: int, aspectratio: float, colortable: bytes,
        trailer: bytes, outstream: BinaryIO):

    # image descriptor
    imagedescriptor: list[int] = []  # type: ignore
    imagedescriptor += [0x2c, 0x00, 0x00, 0x00]      # separator of images
    if (nframes - 1).bit_length() > 0:
        imagedescriptor += [(nframes - 1).bit_length(), 0x00]
    else:
        imagedescriptor += [0x00, 0x00]

    imagedescriptor += [width & 0xff,
                        ((height >> 8) | 0x80)]          # height
    imagedescriptor += [height & 0xff,
                        0x00,                         # palette color resolution
                        (backgroundidx << 4) |     # background color index
                        (localcolor << 3)  |        # flags
                        (globalcolor << 2),         # flags
                        (((aspectratio * 100) // 256) & 0xff),
                                            0x00]   # aspect ratio
    outstream.write(struct.pack('>B', len(imagedescriptor))
                    + bytes(imagedescriptor)
                    + colortable                  # local palette
                    )


    block: list[int] = []                       # type: ignore
    if not localcolor and not globalcolor:    # no LZW compression needed
        if interlaced == False:                # no interlace
            for _ in range(nframes):
                outstream.write((block+[0x2c])[:11])
                outstream.write(trailer)

        elif interlaced == True:               # interlace
            pass

        return
    
    # lzw compression required
    interlaceflag: int = (interlaced << 7) 
    codesizebits: int = 2
    while tablesize.bit_length() > codesizebits:
        codesizebits += 1
    if tablesize.bit_length() < 0 or tablesize.bit_length() > 12:
        raise ValueError("too many colors in the color table")
        
    for i in range(nframes):                    # frames
        block += [0x21,                                    # extension
                  0xf9,                                    # graphic control
                  interlaceflag |                          # flags
                          0x01 |                        # transparency
                          (codesizebits - 1),            # size of data
                          ]
        block += [10,                                      # delay
                  backgroundidx,                           # transparent idx
                  *(trailer)]                              # terminator
        outstream.write(struct.pack('>BBB',
                                    *block))           # frame control extension
        block = []
            

def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header: bytes = b'\x47\x49\x46\x38\x39\x61'       # GIF89a trailer
    version: str = "GIF89a"
    zero: bytes = b''
    null: bytes = b'\x00'
    gifend: tuple[int, int, int] = (0x3b, 0x00, 0x00) #; trailer
    nframes: int = 1                                # workaround for mypy
    loopcount: int = 0                              # workaround for mypy
    sequencenumber: int = 0                         # workaround for mypy
    localcolors: bool = False                       # workaround for mypy
    globalcolors:bool = False                       # workaround for mypy
    with open('/dev/zero', 'rb') as zeroflo:
        trailer = bytes([*gifend, *zeroflo.read()])
    colorresolution: int = 7                        # minimum value with palette
    tablesize: int = 2**colorresolution              # size of color table
    globalcolorindex: int = tablesize              
    screenwidth: int = 0                            # workaround for mypy
    screenheight: int = 0                           # workaround for mypy
    aspectratio: float = 0.0                        # workaround for mypy
    backgroundindex: int = 0                        # workaround for mypy
    localcolortable: bytes = b""                    # empty local palette
    globalcolortable: bytes = b""                   # empty global palette
    data: bytes = b""                               # raw data
    blocks: dict[str, list[int]] = {}

    def read_next_byte():
        nonlocal data
        d = data[0:1]
        data = data[1:]
        return ord(d) if d!= "" else None


    def read_next_word():
        w = read_next_byte()*256 + read_next_byte()
        return w
        

    def next_data_subblock(outflo: BinaryIO, subnumber: int):
        datablock = blocks["DATA"][subnumber]
        lzwminsize = 2**(datablock[-2] + 1)
        datablock = datablock[:-2]
        
        # initialize dictionaries
        clearcodes: list[int] = [*range(2**lzwminsize)]
        eoi: int = clearcodes[-1] + 1
        codes: dict[tuple[int],int] = {(i,):i for i in clearcodes}
        last: tuple[int] = ()

        # main decoding loop
        while True:

            # output one code
            code: int = datablock[0]
            datablock = datablock[1:]
            
            # special codes
            if code == clearcodes[-1]:                 # clear dictionary
                last = ()
                codes.update({(i,):i for i in clearcodes})
                continue
                
            elif code == eoi:                          # end of image
                assert last!= (), \
                                    f"end of image without start ({outflo})"
                yield last[-1]
                break

            # all other codes are either strings from previous codes
            # or single characters preceded by a zero bit
            if code <= lzwminsize:                     # use string
                seq: tuple[int]= (*last,code)
                if seq in codes:
                    code = codes[seq]
                else:
                    unlzwcode: tuple[int,...] = (*seq, last[-1])
                    codes[(unlzwcode):] = len(codes)+1
                    code = codes[seq]
                    
            else:                                       # literal character 
                code -= lzwminsize
                unlzwcode = (code,)
                last = unlzwcode
                
            
            yield code


            # update sequence for next iteration
            if last!= ():
                unlzwcode = (*last,code)
                codes[(unlzwcode):] = len(codes)+1
            
            
            
    if read_next_byte()!= 0x47:
        raise RuntimeError("not an animated gif!")
        
    while True:
        readchar: chr = read_next_byte().to_bytes(1,'little').decode()
        if readchar is None:
            break

        if readchar == ';':
            break
            
        elif readchar == '!':
            labelname = ''
            while True:
                readchar = read_next_byte().to_bytes(1,'little').decode()
                if readchar is None:
                    raise RuntimeError("unexpected EOF reading form label name")
                elif readchar == '\x00':
                    break

                labelname += readchar

            if labelname == "NETSCAPE":
                loops = read_next_word()
                sleepms = read_next_word()
                readchar = read_next_byte().to_bytes(1,'little').decode()
                if readchar!= ',':
                    raise RuntimeError("\"netscape\" label must be followed by comma!")
                labelname = ''
                while True:
                    readchar = read_next_byte().to_bytes(1,'little').decode()
                    if readchar is None:
                        raise RuntimeError("unexpected EOF reading form netscape info")
                    elif readchar == '\x00':
                        break

                    labelname += readchar
                
                if labelname.lower()!= "loop":
                    raise RuntimeError("\"netscape\" label must contain \"Loop\" keyword!")
                    
                loopcount = loops
        
            else:
                raise NotImplementedError(f"label \"{labelname}\" not implemented")
                
        elif readchar == ',' and not localcolors and not globalcolors:
            sequencenumber += 1
            nframes = max(sequencenumber, nframes)
            blocks.setdefault(labelname,[]).append(sequencenumber)

        elif readchar == '#':
            sequencenumber += 1
            nframes = max(sequencenumber, nframes)
            blocks.setdefault(labelname,[]).append(sequencenumber)
            subblocks = []
            while True:
                readchar = read_next_byte().to_bytes(1,'little').decode()
                if readchar is None:
                    raise RuntimeError("unexpected EOF reading form subblocks")
                elif readchar == '\x00':
                    break
                else:
                    subblocks += [readchar]

            blocktype = ''.join(subblocks[:-1]).upper()
            subblocknumber = int(''.join(subblocks[-