#!/usr/bin/env python3

import sys
import os
os.chdir(sys.path[0])

import struct
from typing import List, BinaryIO


# The list of colors we will use in our generated frames.
COLORS = [
    0x000000, 0xFF6500, 0xFF9F00, 0xFDC700, 0xCCFF00,
    0x9FFF00, 0x6EFF00, 0x3DFF00, 0x00FF01, 0x00EC0C,
    0x00DA4B, 0x00CB9A, 0x00BCE9, 0x00B2FF, 0x00AAFF,
    0x00AAFF, 0x0090FF, 0x006AFE, 0x0039FD, 0x2401FE,
    0x7700FA, 0xBA00EE, 0xFE00E6, 0xFE00D9, 0xFE00BF,
    0xFE009E, 0xFD007D, 0xFC005C, 0xFA003A, 0xF90018,
]



# Converts an integer to its little-endian representation as bytes
def to_le(i: int): return i.to_bytes((i.bit_length()+7)//8, byteorder='little', signed=True)



# A simple class representing one frame of an animated GIF image
class Frame:
    def __init__(self, width: int, height: int, x: int, y: int):

        self.__width: int = width
        self.__height: int = height
        self.__x: int = x
        self.__y: int = y

        # Create the palette
        self.__palette = []
        for color in COLORS:
            r = ((color >> 16) & 0xff) / 255
            g = ((color >>  8) & 0xff) / 255
            b = ((color      ) & 0xff) / 255

            self.__palette += [(r*255,g*255,b*255)] * 3


        # Create the actual pixel data
        self.__pixelData = [[0]*width for _ in range(height)]


    # Sets the value at column `x`, row `y` to be the given color index `index`.
    # This is done by setting all three channels to the same value
    def setPixel(self, x: int, y: int, index: int):
        if not -1 < index < len(COLORS): raise ValueError("Color index out of bounds")

        self.__pixelData[y][x] = index



    # Returns the raw frame data as a binary array suitable for writing directly to memory
    def getFrameData(self)-> List[int]:
        encodedPixels = [bytearray()] * self.__height

        for y in range(self.__height):
            prevCodeLen = None
            litCodeLen = 9
            codeTable = {}
            codesUsed = []
            currentByte = bytearray()


            # For each pixel in this line...
            for x in range(self.__width):
                pix = self.__pixelData[y][x]

                # If we have just started sending pixels, encode the first LZW code. It's a literal.
                if len(currentByte) == 0:
                    # Encode length
                    if prevCodeLen!= None:
                        encodedPixels[y].append(prevCodeLen)

                    # Get index of pixel and store it
                    ix = PIXELS.index(pix)
                    currentByte.extend(struct.pack("<H", ix))


                else:
                    # Encode what we already sent last time
                    encodedPixels[y].extend(currentByte)

                    # Find a matching previous code
                    key = currentByte[-1:] + to_le(pix)
                    try:
                        codeIx = codesUsed.index(key)
                        if litCodeLen >= 12:
                            currentByte.extend([codeIx//256, codeIx%256])
                        elif litCodeLen >= 4:
                            currentByte.extend([codeIx % 256 | 0xE0])
                        else:
                            currentByte.extend([codeIx%16 << 4])

                        prevCodeLen = litCodeLen


                    except ValueError:
                        pass

                    # Nope, need to make a new code
                    if len(codesUsed) > 0xffff:
                        print("Code table full!")
                        exit(-11)

                    codesUsed.append(key)
                    currentByte[:] = []



        return encodedPixels









# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    global FRAME_COUNT, FRAMES, MAX_FRAMES, FRAME_DELAY

    # Write the magic string (header) to the output stream
    out.write(b'GIF89a')



    # Generate a bunch of random sized images
    while True:
        WIDTH = int.from_bytes(
                rng.read(2), 'big')
        HEIGHT = int.from_bytes(
                rng.read(2), 'big')
        
        FRAME_DELAYs = []
        for _ in range(FRAME_COUNT):
            delay = int.from_bytes(
                  rng.read(2), 'big')
            
            FRAME_DELAYs.append(delay)
            
        
        X = int.from_bytes(
              rng.read(2), 'big')
        
        
        Y = int.from_bytes(
              rng.read(2), 'big')
        

        # Create the main frame object
        f = Frame(WIDTH, HEIGHT, X, Y)

        

        
        for _ in range(FRAME_COUNT):
            # Generate some more pixels!
            DATA = int.from_bytes(
                   rng.read(1), 'big')
            for _ in range(DATA):
                # We only support single colors right now
                COLORIX = int.from_bytes(
                              rng.read(1), 'big')
                
                # Set some pixels randomly :)
                for _ in range(MAX_PIXELS // 5):
                    px = int.from_bytes(
                               rng.read(1), 'big')
                    
                    py = int.from_bytes(
                               rng.read(1), 'big')
                    
                    
                    f.setPixel(px,py,COLORIX)




            
        # Add the frame to our list of frames
        FRAMES.append(f)


        # Check if there are enough frames yet
        if len(FRAMES)==MAX_FRAMES: break
        
        
        
        
    # Write the size of the entire animation to the output stream
    out.write(len(FRAMES).to_bytes(2,'big'))


    # Write the individual frame delays next
    for fd in FRAME_DELAYs[:MAX_FRAMES]:
        out.write(fd.to_bytes(2,'big'))




    # Now write the rest of the header. To do so, we need to keep track of where the
    # image data starts and how big it is. Of course we'll still need to adjust these
    # offsets.
    bodyOffset = 13
    for f in FRAMES[:MAX_FRAMES]:
        # Skip over the size fields which we write later
        bodyOffset+=2

        # Then add the screen descriptor
        bodyOffset+=7

        # Finally, add the rest of the header
        bodyOffset+=3

        # Now calculate how far the image data actually ends up
        endOfImageData = bodyOffset+sum(map(lambda s : (s[0]+1)*(s[1]), zip(*f.getFrameData())))

        # Save the offset
        bodyOffset = endOfImageData




    # Write the usual stuff
    out.write(bodyOffset.to_bytes(2,'big'))
    out.write(b'\x2c\x00')
    out.write(b'\x00')
    out.write(b'\x00')




    # Write the image information
    for f in FRAMES[:MAX_FRAMES]:
        # Write the size field
        out.write(((f.__width+1)*f.__height).to_bytes(2,'big'))

        # Write the left and top coordinates
        out.write(f.__x.to_bytes(2,'little'))
        out.write(f.__y.to_bytes(2,'little'))

        # Write the local colour map info
        out.write(b'\x00')
        out.write(b'\x00')
        out.write(b'\x00')

        # Write the interlace bit
        out.write(b'\x00')

        # Write the background color index
        out.write(b'\x00')

        # Write the aspect ratio
        out.write(b'\xff')
        out.write(b'\xff')

        # Now write the image descriptor
        out.write(f.__width.to_bytes(2,'little'))
        out.write(f.__height.to_bytes(2,'little'))

        out.write(b'\x00')
        out.write(b'\x00')

        out.write(b