#!/usr/bin/env python3
"""Generate a random gif file."""

import struct
from typing import BinaryIO


def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    """Writes a random.gif file to `out`.

    The content is generated by drawing bytes from `rng`, which should be
    initialized with /dev/urandom.
    """
    header = b"GIF89a"
    width = bytearray([width])
    height = bytearray([height])
    depth = bytearray([depth])
    bgcolor = b"\xFF\xFF\xFF" * (depth // 8)
    aspect = b"\xE0\x00"
    gctsize = b"\x07" if depth == 8 else b"\x05"

    screendesc = b"".join((
        b"!\xF9\x04",
        b"",
        # size of global color table, minus one
        gctsize,
        bgcolor,
        aspect,
        b"\xF9\x00"))

    imagedescriptor = b"".join((
        b"," + b"\x00\x00",      # left and top position
        width,                   # width and height
        b"\x80",                 # flags - no local color table
        b"\x00",                 # interlace flag
        b"\x00",                 # sort flag
        b"\x00",                 # reserved

        # size of local color table, minus one
        b"\x00" if depth!= 8 else gctsize))

    applicationextblock = b"".join((
        b"\x21\xFF\x0B",        # extension label
        b"NETSCAPE2.0",         # application identifier
        b"\x03\x01",             # application authentication code
        b"\x00\x00"))            # sub-block length (should always be zero)

    commentextblock = b"".join((
        b"\x21\xFE\x0C",        # extension label
        b"comment",             # comment text
        b"This is a test."))    # comment

    trailer = b"\x3B"           # terminator for GIF file

    def write_subblocks(out: BinaryIO,
                        content: bytes,
                        blocklen=255):
        while True:
            chunk = content[:blocklen].ljust(blocklen, b"\x00")
            out.write(struct.pack("BB",
                                 len(chunk),
                                 0))   # no compression
            out.write(chunk)

            content = content[blocklen:]
            if not content or len(content) <= blocklen:
                break


    # Write the GIF file:
    out.write(header)               # Magic Header
    out.write(b"")                  # Logical Screen Descriptor
    out.write(imagedescriptor)      # Image Descriptor
    out.write(screendesc)           # Global Color Table

    write_subblocks(out, applicationextblock[:-1])
    write_subblocks(out, commentextblock[:-1])
    out.write(trailer)              # Trailer