#!/usr/bin/env python3

import os, sys, struct


def write_header(
        out: BinaryIO, width: int, height: int, bits: int
    ) -> None:
    """Write out the header for our generated GIF."""

    # Create a black palette entry with all entries equal to zero
    palette = bytes([0x00]*bits//8*3)

    # The header should look as follows:
    #   Header    - b'GIF89a'
    #   Logical Screen Descriptor - 6 bytes
    #       Size of the whole GIF (including this field)
    #       0x00
    #       Pixel aspect ratio - set to 'none' (i.e., 0)
    #       Width of the screen in pixels
    #       Height of the screen in pixels
    #       Global Color Table Flag - True
    #           Length of the global color table in 8-bit increments
    #             0 means that we don't have one.
    #   Global Color Table
    #       Entries are 3 bytes each, representing an R, G, B value in
    #       RGB format respectively. Set to black (all zeros).
    #   Zero-frame Data Extension Block
    #       This extension block contains exactly two bytes, both of which
    #       are equal to 0x00. See below for more explanation on what these
    #       blocks mean.
    header = [b'GIF89a']
    header += list(struct.pack('<2BH3xBBBBB',
        0xFFF7 + len(palette), 
        0, 
        0, 
        width, 
        height, 
        0b11000000 | 7, 
        7, 
    ))
    header += [palette[:], b'\x00\x21\xF9\x04\x00\x00\x00\x00\x00\x2C\x00\x00']
    out.write(b''.join(header))


def write_graphics_control_extension(
        out: BinaryIO, delay: int, transparent: int
    ) -> None:
    """Write out the graphics control extension for our generated GIF"""
    # We want to use this extension because it allows us to specify how long
    # each frame will be displayed for before switching to the next frame.
    # Also, we can specify whether or not the frame should be treated as
    # transparent when displayed. If the frame is transparent then its
    # background will not be overwritten by subsequent frames but instead the
    # first pixel's color will be used.
    out.write(bytes([
        0x21,
        0xf9,
        0x04,
        0x05,
        delay&255,
        delay>>8,
        0,
        0,
        0,
        0 if transparent else 1
    ]))


def write_image_descriptor(
        out: BinaryIO, x: int, y: int, width: int, height: int, local_color_table: str=''
    ) -> None:
    """Writes the Image Descriptor section of the GIF"""
    # NOTE: It appears that the x and y coordinates refer to the upper left hand corner
    # of the image rather than the lower right corner like everything else does....?
    flags = '\x87' if local_color_table!= '' else '\x00'
    out.write(bytes([
        0x2c,
        x & 255,
        x >> 8,
        y & 255,
        y >> 8,
        width & 255,
        width >> 8,
        height & 255,
        height >> 8,
        0,
        0,
        flags,
    ]))


def write_local_color_table(
        out: BinaryIO, num_entries: int
    ) -> None:
    """Writes the Local Color Table Section of the GIF"""
    out.write(list(struct.pack('<BBBBB',
        num_entries,
        0,
        0,
        7,
        0,
    )))



def create_image_data(
        rng: BinaryIO, width: int, height: int, bits: int, delay: int, transparent: int
    ) -> None:
    """Create the image data for our generated GIF"""
    # NOTE: It appears that the x and y coordinates refer to the upper left hand corner
    # of the image rather than the lower right corner like everything else does....?
    # So we need to specify the dimensions of the images as being bigger than they actually
    # are so that nothing gets cropped off...

    # Use the provided RNG object to generate randomized data for the image. Since we
    # ultimately just want some random bytes, we'll just read up to the appropriate amount
    # of bytes here.
    max_num_bytes = (width * height // 8) + (height * 2)
    image_data = rng.read(max_num_bytes)

    # Now we go through and replace every nonzero byte with the specified transparency index.
    new_image_data = []
    for idx, val in enumerate(image_data):
        if idx < height * 2:
           # We keep the top row at zeros since it's our text description above this row
           new_image_data+=[val]
        elif val == 0:
            new_image_data += [(transparent & 2**bits-1)]
        else: 
            new_image_data += [val]

    # Now we go through again and make sure that every n^th bit is reversed. This ensures
    # that we avoid having artifacts due to anti-aliasing effects.
    def subsample_byte(b):
        result = b << 1
        carryover = (result & 2**bits) >> bits
        result &= ~(2**bits)
        result |= carryover
        return result
    subsampled_image_data = map(subsample_byte, new_image_data)

    # Finally, we flip every other line horizontally so that they're vertically inverted
    flipped_image_data = []
    for i in range(height):
        start = i*width//8
        end = (i+1)*width//8
        flipped_image_data+=reversed(list(subsampled_image_data)[start:end])

    # At this point, we know that the image data we have is valid and ready to be outputted!
    out.write(flipped_image_data)


if __name__=='__main__':
    if len(sys.argv)!= 2:
        print("Usage:./generate_random.py outfile")
        exit()

    outfile = open(sys.argv[1], "wb")

    try:
        # Open /dev/urandom and pass it to generate_random_gif
        inputfile = open("/dev/urandom", "rb")

        seed = inputfile.read(8)
        print("Using seed:", ''.join(['{:02X}'.format(c) for c in seed]))

        rng = os.fdopen(os.dup(inputfile.fileno()))
        generate_random_gif(rng, outfile)

    finally:
        # Close files opened by this script
        outfile.close()
        inputfile.close()