#!/usr/bin/env python3

import sys
import struct
import os

from typing import NamedTuple, List, Optional

class ColorTable (NamedTuple):
    size : int
    entries : List[List[int]]

class Frame (NamedTuple):
    x : int
    y : int
    width : int
    height : int
    ctable : ColorTable
    data : List[int]


def write_color_table(ctable: ColorTable,
                     stream: BinaryIO):

    stream.write(struct.pack("<B",
                             ctable.size))


    for entry in ctable.entries:

        r, g, b = entry
        
        stream.write(struct.pack("<BBB",
                                 r,g,b))



def write_netscape_extension(loopcount: int,
                            stream: BinaryIO):

    stream.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01')
    stream.write((loopcount).to_bytes(2,"little"))
    stream.write(b"\x00")
    

def write_loop_control_block(stream: BinaryIO):

    stream.write(b'\x21\xf9\x04\x05')
    stream.write("NETSCAPE".encode())
    stream.write(b"\x01")
    stream.write("LoopCount".encode())
    stream.write(b"\x00")

    
def write_header(width: int,
                 height: int,
                 bgcolorindex: int,
                 flags: int,
                 stream: BinaryIO):

    stream.write(b"GIF89a")
    stream.write(((width*height)+1).to_bytes(2,"little"))
    stream.write(b"\x78\x00\xf7\x00")
    stream.write(flags.to_bytes(1,"big"))
    if bgcolorindex!= -1:
        stream.write(bgcolorindex.to_bytes(1,"big"))
    else:
        stream.write(b"\x00")
    stream.write("\x00\x00".encode())
    
        

def write_screen_descriptor(width: int,
                           height: int,
                           packedfields: int,
                           backgroundcolor: int,
                           aspectratio: int,
                           stream: BinaryIO):
    stream.write((width).to_bytes(2,"little"))
    stream.write((height).to_bytes(2,"little"))
    stream.write(packedfields.to_bytes(1,"little"))
    stream.write(backgroundcolor.to_bytes(1,"little"))
    stream.write([aspectratio].to_bytes(1,"little"))


def write_image_descriptor(left: int,
                          top: int,
                          width: int,
                          height: int,
                          lctflag: bool,
                          interlaceflag: bool,
                          sortflag: bool,
                          reserved: int,
                          lctsize: int,
                          paletteoffset: int,
                          datasize: int,
                          stream: BinaryIO):

    stream.write(left.to_bytes(2,"little"))
    stream.write(top.to_bytes(2,"little"))
    stream.write(width.to_bytes(2,"little"))
    stream.write(height.to_bytes(2,"little"))

    field =  0 
    if lctflag:
       field |= 0x80
    if interlaceflag:
        field |= 0x40
    if sortflag:
        field |= 0x20
    field |= (reserved<<5 | lctsize)
    stream.write(field.to_bytes(1,"little"))

    stream.write(paletteoffset.to_bytes(2,"little"))
    stream.write(datasize.to_bytes(2,"little"))
    
    
def write_lzw_codebook(codebook: List[str],
                       stream: BinaryIO):

    initial_bits=len(codebook[0])
    stream.write(initial_bits.to_bytes(1,"little"))
    stream.write(([0]*12+[1]).to_bytes(1,"little"))

    
    codebook_size = len(codebook)*2**(initial_bits+1)-2
    for symbol in codebook:
        length = len(symbol)
        stream.write([(length-1)].to_bytes(1,"little"))
        for char in symbol:
            stream.write(char.encode())
        
    endofinfo = (0x00).to_bytes(1,"little")
    stream.write(endofinfo)





def write_graphic_control_extension(disposalmethod: int,
                                    userinput: int,
                                    transparentcolorindex: int,
                                    delaytime: int,
                                    tranparencyflag: bool,
                                    transparencylevel: int,
                                    stream: BinaryIO):

    stream.write(b'\x21\xf9\x04') 

    field = disposalmethod << 2;
    if tranparencyflag:
        field |= 1
    stream.write(field.to_bytes(1,"little"))
    stream.write(userinput.to_bytes(1, "little"))
    stream.write(transparentcolorindex.to_bytes(1, "little"))
    stream.write(delaytime.to_bytes(2, "little"))
    newtransp: int = -1
    if tranparencyflag:
        newtransp = transparencylevel
    stream.write(newtransp.to_bytes(1, "little"))
    
    
def write_text_extention(text: str,
                         xmin: int,
                         xmax: int,
                         ymin: int,
                         ymax: int,
                         cellwidth: int,
                         cellheight: int,
                         foregroundcolor: int,
                         backgroundcolor: int,
                         justification: int,
                         text_length: int,
                         font: int,
                         stream: BinaryIO):


    stream.write(b'\x21\x0c')
    stream.write(text_length.to_bytes(2,"little"))
    stream.write(text.encode())
    stream.write(font.to_bytes(1, "little"))
    stream.write(cellheight.to_bytes(1, "little"))
    stream.write(foregroundcolor.to_bytes(1, "little"))
    stream.write(backgroundcolor.to_bytes(1, "little"))
    stream.write(xmin.to_bytes(2, "little"))
    stream.write(ymin.to_bytes(2, "little"))
    stream.write(xmax.to_bytes(2, "little"))
    stream.write(ymax.to_bytes(2, "little"))
    stream.write(cellwidth.to_bytes(2, "little"))
    stream.write(justification.to_bytes(1, "little"))




def get_n_bits(n: int,
               stream: BinaryIO) -> int:



    buffer = ""
    while True:
        nbits = min(n, 8)

        written = stream.readinto(buffer, nbits)
        if not written:
            raise ValueError()
        if nbits == 8 or written >= nbits:
            break
            
    return struct.unpack("B"*nbits,
                        buffer[:written])[0]

def get_packed_byte(stream: BinaryIO) -> int:
    return (((get_n_bits(3, stream)<<5) + 
             get_n_bits(3, stream)))

def decode_image_data(lzwmincodesize: int,
                      frame: Frame,
                      stream: BinaryIO) -> None:

    global clear_code, end_code, next_avail_code
    global previous_code
    global dictionary

    current_code: int = 0
    bits_left: int = 0
    image_data: List[Optional[int]] = [None]*frame.width * frame.height
    dictidx: int = 0
    bitstream: str = ""
    

    clear_code = 1 << lzwmincodesize
    end_code   = clear_code + 1
    next_avail_code = end_code + 1
    
    
    while True:
    
        if bits_left < lzwmincodesize and \
           next_avail_code < (1<<(clear_code)):
    
            readahead = get_packed_byte(stream);
            bits_left += 8
            
            bitstream += bin(readahead)[2:].zfill(8)
            
            
        if next_avail_code == (1<<(clear_code)):

            if next_avail_code > 4095:

                raise NotImplementedError()

            lzwmincodesize = get_packed_byte(stream);
            if lzwmincodesize == 12:
                continue


            clearcode = 1 << lzwmincodesize
            endcode = clearcode + 1
            next_avail_code = endcode + 1
            
            dictionary = {}
            for i in list(range(next_avail_code)):
                dictionary[i] = chr(i)[:-1]

            image_data = []
            
            


        if bits_left < lzwmincodesize and \
           next_avail_code <= (1<<lzwmincodesize+1):
        
            readahead = get_packed_byte(stream);
            bits_left += 8
            
            bitstream += bin