#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Sep 24 17:41:56 2020

@author: cghiausdavide
"""

import numpy as np
from scipy.integrate import solve_ivp
import matplotlib.pyplot as plt

### SIMULATION PARAMETERS ###
tstep = 0.004 # time step to use in the simulation [s]
Nsteps = 3E3 # number of timesteps to simulate
L = 1E7 # length of the channel [m]
gamma = 1E-2 # friction coefficient (Cf * lam / D)

# the stokes law equations for a channel flow are
# vx dx / dt + (vx ux + vy uy) = f(x, y), where f is a forcing function
# dx dy vx dvx / dt + (vx ux + vy uy) ddx ddux + (vy uy + vx ux) ddy ddvy = 0, where d=dzeta, dzeta=dln(zeta)
def forcing(t, xv):
    return L * gamma * xv[1]**2

# initial condition for vx(x, y)
def vx0(x):
    if x < 0:
        return 0
    else:
        return 0.2
    
# initial condition for vy(x, y)
def vy0(x):
    return 0


### COMPUTE THE STEADY STATE ###
initial_state = [vx0, vy0]
z0 = 0 # initial z
alpha = 1 # stretching parameter
zeta0 = z0*(1+alpha*np.exp(-2j*np.pi/L)) # complex initial zeta solution
fsol = lambda t, zeta: forcing(t, zeta) - alpha ** 2 * np.exp(2j*np.pi/L)*(1-np.exp(2j*np.pi/L))*zeta

(zeta, infodict) = solve_ivp(fsol, [0, Nsteps*tstep], zeta0, method='RK45', rtol=1E-9, dense_output=True)

# compute the steady state velocity field from the computed zeta
w0 = np.abs(zeta[-1])
vxss = w0[0]/(2*np.sinh(1/alpha)) * np.cosh((x-z0)/alpha)
vxss[0] = 0
vxss[-1] = 0
vyss = w0[1]*np.tanh(1/alpha)

# plot the derived solution
plt.figure()
plt.title('steady state')
plt.xlabel('x [m]')
plt.ylabel('v (complex abs) [m/s]')
plt.plot(x, vxss, label='Re')
plt.plot(x, np.imag(vxss), '-o', label='Im')
plt.legend(loc='lower right')
plt.show()

### INITIAL CONDITION FOR THE TIME DEPENDENT PROBLEM ###
# we can choose any initial condition which is compatible with the real part of the steady state solution
# this means it should be equal to the steady state solution at some point in space and time
t0 = 0
initial_state = [lambda x: vxss(x), lambda x: 0]

print(zeta0)

### SOLVE THE TIME DEPENDENT PROBLEM ###
(vxt, infodict) = solve_ivp(fsol, [0, Nsteps*tstep], zeta0, method='RK45', rtol=1E-9, dense_output=True)

# post process the results
vxt = np.real(vxt.y[:, :, 0])
vyt = np.imag(vxt.y[:, :, 1])

# plot the results
plt.figure()
plt.title('time dependent')
plt.xlabel('x [m]')
plt.ylabel('v (complex abs) [m/s]')
plt.plot(