#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + datalen.to_bytes(3, 'big'))
    out.write(data)

# Generates a random JPEG file into `out` using the
# random number genrator `rng` (/dev/urandom).
def generate_random_jpeg(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1024 bytes of binary random data, to be written
    # afterwards as JFIF-formatted image.
    jfif = rng.read(1024)

    # First two bytes are SOI marker (start of image).
    # Last five bytes are EOI marker (end of image).
    jfif = b'\xff\xd8' + jfif + \
           b'\xff\xd9'
    # Write the data to the output.
    out.write(jfif)

# Generates a random PNG file into `out` using the
# random number genrator `rng` (/dev/urandom).
def generate_random_png(
        rng: BinaryIO, out: BinaryIO
    ):
    # The first eight bytes of any PNG image must be
    # this exact string.
    magic = b"\x89PNG\r\n\x1a\n"
    # Then comes a series of chunks consisting of:
    # - length in bytes of chunk data (four bytes)
    # - type of chunk (4 bytes)
    # - chunk data (length bytes)
    # - CRC checksum over chunk type and data (four bytes)
    # We need to write at least three chunks that have
    # some meaningful content.
    # The IDAT chunk contains the actual image data.
    # It is compressed with zlib, but here we just use
    # raw deflate data.
    idat = rng.read(512)
    # The IHDR chunk describes the overall picture.
    ihdr = struct.pack(">IIBBBxxxx", *[rng.read(4) for i in range(6)] )
    # The PLTE chunk stores palette information for
    # indexed color images. Here we don't use it, so
    # its contents don't matter.
    plte = rng.read(768)
    # The IEND chunk marks the end of the image.
    iend = b"\x00\x00\x00\x00IEND\xae\x42\x60\x82"
    # Now assemble these chunks into full PNG image
    # data.
    out.write(magic + ihdr + plte + idat + iend)
