#!/usr/bin/env python3

import struct
from typing import BinaryIO, List, Tuple


def parse_line(line: str) -> Tuple[int, int]:
    """Parse a single line of instructions.

    :param line: The line to be parsed.
        Must contain a list of four 16-bit integers (floating point instructions).

    :return: A tuple containing the four floating point instructions in order.
            If not valid, raise an exception.

    """
    try:
        return tuple([struct.unpack('>H', bytes.fromhex(x))[0] for x in line.split()])
    except struct.error as ex:
        raise ValueError("Invalid hex string") from ex


def parse(input_file: BinaryIO) -> List[Tuple[int, int]]:
    """Parse the given input file.

    :param input_file: The input file.
        Can contain multiple lines of 4 integer instructions each.
        Each instruction must be separated by whitespace characters and
        can optionally contain leading zeros.

    :return: An array containing the parsed instructions.
                The array is sorted using the sort order used in hardware.

    """
    # Create an empty array that will contain the output instructions.
    result = []

    # Iterate over all lines in the input file.
    for line in input_file:
        line = line.strip().lower()

        # Parse the current line.
        parts = [parse_line(line), ]

        # Filter out invalid instructions.
        filtered_parts = [p for p in parts if len(set(p)) != 1]

        # Add each instruction in the current line to the output array.
        result += filtered_parts

    # Sort the instructions using the same sort order as in hardware.
    result = sorted(result, key=lambda