#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(ctsize):
            if i == 0:
                palents += struct.pack('B', *[0] * 3)
            elif i == 1:
                palents += struct.pack('!BHB', *[0] + [int.from_bytes(rng.read(2), 'big'), 0])
            else:
                palents += paldata[i]

        # done with the palette data

        # skip the background color index (we don't use this info)
        rng.seek(1)

        # skip the aspect ratio byte
        rng.seek(1)

        # write out the global color table
        out.write(palents)

    # add metadata header block
    out.write(header + b'\x03\x01' + \
        struct.pack('<HH', image_width, image_height) + \
        b'\x87a' + b'\x00' * 7 + b'\xFF' * 33 + b'\xFF')

    # one frame
    out.write(b'\x2C')

    # offset from left edge of canvas to
    # the current frame
    out.write(struct.pack("<HH", 0, 0))

    out.write(struct.pack("<HH", image_width, image_height))
    out.write(b'\x00')

    # no transparent pixels
    out.write(b'\x00')

    out.write(b'\x21\xF9\x04\x01\x00\x00\x00')
    out.write(b'\x2C')

    # output image data per-byte
    for _ in range(image_width * image_height):

        # little endian
