#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # GIF header consists of "GIF" and 6 version info bytes.
    # We do not care about these values, so fill them with zeros.
    gif_header = b'GIF\x38\x9b\x00\x00\x00'
    out.write(gif_header)

    # ASCII string 'NETSCAPE2.0', also zero padded but followed by a
    # null byte, for a total length of eleven bytes.
    app_ext_header = b'NETSCAPE2.0' + b'\x00' * 7

    # Application extension block consists of an application extension
    # header, then a single loop count byte for the animation (zero),
    # and finally a terminator.
    app_ext_block = struct.pack('<BBI', len(app_ext_header), 0xa1, 0x01) + \
                     app_ext_header + b'\x00'

    # Image descriptor is up to 7 bytes long. The first byte is
    # set to zero, because we only have one graphic control block and
    # one image descriptor following. The remaining six bytes contain
    # the size of the image, which must be divisible by 2 to match the
    # way that BMP files store images as DWORDs. The image width is taken 
    # from the second byte of the image descriptor.
    img_desc = struct.pack("<BBHB", 0x00, 0x0, 0x0, 0x0)[:int((rng.read(1)[0] % 256)*3)]
    #print("img_desc = ",img_desc )
    #print("type(img_desc) = ", type(img_desc))
    print ("len(img_desc) = ", len(img_desc))
   # out.write(img_desc)
    # Graphic control extension block consists of a 1-byte header,
    # 3-byte delay time, transparent color index (always zero),
    # and a block terminator.
    gce = struct.pack('<BHHB', 0x21, 0xf9, 0x04, 0x00)
    out.write(gce)
    # Image data block is a variable-length binary blob of pixel data.
    # It may or may not be in color depending on the graphic control block,
    # and it will use either indexed or full colorspace depending on whether
    # there was an application extension block present. If no application
    # extension block is provided, then this is a fullcolorspace image, and
    # the format is simply raw rgba data. Otherwise it is a palletised image,
    # and the format is RLE compressed.
    if len(app_ext_block) == 0:
        # Write out 1024 bytes of random data.
        data = rng.read(1024)

        # Write out data section of image.
        out.write(data)
    else:
        # Write out 1024 bytes of random data.
        data = rng.read(1024)
    out.write(data)

# Generates a random JPEG file into `out` using the
# random number genrator `rng` (/dev/urandom).
def generate_random_jpeg(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1024 bytes of binary random data, to be written
    # afterwards as JFIF-formatted image.
    jfif = rng.read(1024)

    # First two bytes are SOI marker (start of image).
    # Last five bytes are EOI marker (end of image).
    jfif = b'\xff\xd8' + jfif + \
           b'\xff\xd9' 

    # Write out data section of JFIF.
    out.write(jfif[2:-5])

