#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):

            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]
            
    else:
        
        # no palette; just 0xFF to tell it's unused
        palents = b'\x00\xFF'

    # skip this block
    rng.seek(1)

    # create the graphics control extension block
    gcxtblk = \
        struct.pack('>bbH',
                    0xF9,
                    0x04, # flags
                    0) + \
        struct.pack('B', rng.read(1)[0])

    # create the image data block
    imgedat = f'{image_width}x{image_height}'
    pixdat = ''

    # iterate through each row
    for _ in range(image_height):

        # for each column, pull from our palette
        for _ in range(image_width):
            pixdat += chr(ord(rng.read(1)))

    # the image block is the data
    imgblk = \
        struct.pack('b', len(imgedat)) + \
        imgedat.encode() + b';' + \
        palents + gc