#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + datalen.to_bytes(3, 'big'))
    out.write(data)

# Generates a random JPEG file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_jpeg(
        rng: BinaryIO, out: BinaryIO,
        blocksize: int=256*2048 # 1MiB
        ):
    signature = bytes([0xFF]) + b'JFIF'
    w = h = int.from_bytes(rng.read(2), 'big')
    header = {
            1: b'\xC4\xD8',  # huffman-only baseline
            2: b'\xC4\xDC',  # huffman-only extended sequential
            3: b'\xC4\xDD',  # huffman-only progressive
            4: b'\xC4\xDE',  # huffman-only lossless
            5: b'\xC7\xC0',  # arithmetic coding baseline
            6: b'\xC7\xC1',  # arithmetic coding extended sequential
            7: b'\xC7\xC2',  # arithmetic coding progressive
            8: b'\xC7\xC3',  # arithmetic coding lossless
            9: b'\xC8\xCC',  # JPEG-LS baseline
            10:b'\xC8\xCD',  # JPEG-LS extended sequential
            11:b'\xC8\xCE',  # JPEG-LS progressive
            12:b'\xC8\xCF',  # JPEG-LS lossless
            13:b'\xC9\x00',  # mixed RGB - YCbCr baseline
            14:b'\xC9\x01',  # mixed RGB - YCbCr extended sequential
            15:b'\xC9\x02',  # mixed RGB - YCbCr progressive
            16:b'\xC9\x03',  # mixed RGB - YCbCr lossless
            17:b'\xCA\x00',  # mixed CMYK - YCbCr baseline
            18:b'\xCA\x01',  # mixed CMYK - YCbCr extended sequential
            19:b'\xCA\x02',  # mixed CMYK - YCbCr progressive
            20:b'\xCA\x03',  # mixed CMYK - YCbCr lossless
            21: b'\xCB\x00', # mixed CMYK - YCCK baseline
            22: b'\xCB\x01', # mixed CMYK - YCCK extended sequential
            23: b'\xCB\x02', # mixed CMYK - YCCK progressive
            24: b'\xCB\x03', # mixed CMYK - YCCK lossless
            25: b'\xCD\x00', # mixed HSL - YCCK baseline
            26: b'\xCD\x01', # mixed HSL - YCCK extended sequential
            27: b'\xCD\x02', # mixed HSL - YCCK progressive
            28: b'\xCD\x03', # mixed HSL - YCCK lossless
          }[int.from_bytes(rng.read(1), byteorder='big')]
    out.write(signature)

    def write_marker(marker):
        out.write(struct.pack('>H', marker))

    def write_segment(segment):
        out.write(struct.pack('>H', len(segment)))
        out.write(segment)

    def write_app0():
        segment = b'\x00\x0E\x4A\x46\x49\x46\x00\x01\x01\x00\x01\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        write_marker(0xFFE0)
        write_segment(segment)

    def write_sof():
        segment = (
                b'\x08\x00' +
                struct.pack("BB", w & 0xff , (w >> 8) & 0xff ) +
                struct.pack("BB", h & 0xff , (h >> 8) & 0xff ) +
                0x03).ljust(blocksize, b'\x00')
        write_marker(0xFFD8)
        write_segment(segment)

    def write_eoi():
        write_marker(0xFFD9)

    write_app0()
    write_sof()

    for i in range(512):
        segment = rng.read(blocksize)
        out.write(segment)

    write_eoi()
