#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        # convert them to the format GIF requires, which is {R, G, B}
        palfmt = '<{}B'.format(len(paldata))
        palette = struct.pack(palfmt, *paldata)

        # write out the color table
        out.write(b'\x21\xF9\x04\x01' +
                    struct.pack('<HHB',
                        0, 0, ctsize)[::-1])
        out.write(palette)

    # start timing block
    out.write(b'\x21\xFE\x04\x01\x00\x00\xFF\xFF\xFF\x00')

    # loop forever
    out.write(b'\x00\x01')

    # start image
    out.write(b'\x2C')

    # set left and top offsets
    out.write(struct.pack('<HH', 0, 0))

    # send image dimensions
    out.write(struct.pack('<HH', image_width, image_height))

    # packed field: local color table flag, interlaced flag, sort flag,
    # reserved zero bits, size of local color table
    out.write(b'\x8C')

    # write out the color table again in the same format as before
    out.write(palette)

    # clear code flag set
    out.write(b'\x00')

    # this is the minimum LZW code length for this particular application
    out.write(b'\x02')

    # pixel data start
    out.write(b'\x21')

    # send frame delay here
    out.write(struct.pack('<HHH', 0, 0, 0))

    # end of gif stream marker
    out.write(b';')

# Main program logic follows:

with open('/dev/