#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(n: int, out: BinaryIO):
    HEADER = b"GIF89a"
    BLOCK_SIZE = 64
    
    # Create the header.
    # The first six bytes are guaranteed to be unique (this is not actually true),
    # so we'll use them as an entropy source for the rest of the file.
    # This will also be used as the gif image color palette
    out.write(HEADER)
    out.write(bytes([rng.randint(0x21, 0x7F)] * 5))
    
    # Write the screen descriptor and local color table.
    # Here, the last byte of the screen descriptor (`color depth`) uses all bits,
    # and indicates that no global color table follows.
    out.write(struct.pack("<HHB", n, n, rng.randint(0x00, 0xFF)))
    
    # We will now write blocks of data alternating between graphic and animation control blocks.
    # The block type ID is written at its start, followed by a length field and then the payload.
    while True:
        if rng.getrandbits(1):
            # Graphic Control Extension Block.
            
            # Prepare the extension block.
            # All fields except `duration` can have completely arbitrary values (as long as they fit in their size),
            # but since this is just a RNG test program, we'll set some things to "interesting" values:
            # `transparent color index`, `disposal method` and `user input flag`.
            # `delay time` will be written later, after a call to another function.
            ext = struct.pack('<BB', 0xF9, 0x1) + \
                struct.pack('>H', rng.randint(0x0000, 0xFFFF)) + \
                bytes([rng.randint(0x00, 0xFF)])
                
            # Determine how many bytes should follow before the next block.
            # Since fields other than `delay time` and `block terminator` don't need padding,
            # we add the delay time length to the initial value here.
            remaining_bytes = BLOCK_SIZE - len(ext) - 1
            
            # Add the delay time to the extension block.
            # Genarate a random delay time value, and encode it as an unsigned LEB128 value.
            # Since the maximum possible delay time (in centiseconds) is ~8 minutes (~500s),
            # we'll use a small modulo operation to limit the range of randomness.
            delay_time = rng.randrange(0, 10 ** 5 // 100)
            ext += leb128_encode(delay_time)
            
            # Update `remaining_bytes` accordingly.
            remaining_bytes -= len(leb128_encode(delay_time))
        
            # We may not have enough space left to write any more data into the extension block.
            # If so, write what's there to disk and advance to the next block.
            if remaining_bytes < 0:
                # Pad the remaining bytes with zeroes.
                ext += bytes([rng.randint(0x00, 0xFF)] * (-remaining_bytes))
                out.write(struct.pack('BB', 0x21, len(ext & 0xFF)))
                out.write(ext)
                break
            
            else:
                # Otherwise, proceed to adding whatever is left up to the end of the block size.
                ext += bytes([rng.randint(0x00, 0xFF)] * remaining_bytes)
        
        else:
            # Image Descriptor Block.
            
            # Prepare the descriptor block.
            desc = struct.pack('<BBHH', 0x2C, 0x00, 0, 0) + \
                bytes([rng.randint(0x21, 0xFE)] * 3) + \
                bytes([rng.randint(0x00, 0xFF)] * 4) + \
                bytes([rng.randint(0x00, 0xFF)] * 9)
            remaining_bytes = BLOCK_SIZE - len(desc) - 1
            
            # Fill the rest of the block with random data
            desc += bytes([rng.randint(0x00, 0xFF)] * remaining_bytes)
        
        # Write the block to disk.
        out.write(struct.pack('BB', 0x21, len(desc & 0