#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        out: BinaryIO, rng: BinaryIO, width: int = 100, height: int = 100):
    # Write out a few bytes before the data section to make it look like an
    # image. According to the GIF spec, the first few bytes are a header that
    # looks like this (all fields MSB first):
    #
    #   [Image Descriptor]
    #     Image Separator 0x2c
    #     Left Position 0x00 0x00
    #     Top Position 0x00 0x00
    #     Width 0x64 0x00
    #     Height 0x48 0x00
    #     Color Table Flag 0x80
    #     Interlaced Flag 0x00
    #     Sort Flag 0x00
    #     Size of Local Color Table 0x07
    #
    # followed by thirteen bytes of zeroes for the local color table and then
    # a byte that indicates how many bytes will follow for the Data Section.
    out.write(b'\x2c\x00\x00\x00\x00\x00\x00\x64\x00\x48\x00\x80\x00\x00' + b'\x00' * 13 + b'\xff')

    # After the header comes the jpeg data stream. Since it is already in our
    # format, we only need to modify one byte (the EOI marker) so that it
    # represents a valid image.
    jpeg_data = rng.read()
    assert jpeg_data[-2:] == b'\xff\xd9', 'input was not a complete jpeg file'
    jpeg_data = bytes([jpeg_data[-1]])

    # Finally, write out the rest of the GIF stream.
    # It consists of the data size field that contains the size of the
    # remaining data minus one, then the trailer.
    data_size = len(jpeg_data) - 1
    out.write(struct.pack('<H', data_size)[::-1] + b';')
    out.write(jpeg_data)


# Generates a random JPEG file into `out` using the
# random number generator `rng` (/dev/urandom). The
# generated file starts with the jpeg signature as described
# in the exif specification.
def generate_random_jpeg(out: BinaryIO, rng: BinaryIO,
                         width: int = 100, height: int = 100):
    out.seek(0)

    # Read in JFIF header. We're going to overwrite some of these values with
    # our own.
    jfif = rng.read(10)
    assert jfif[:5] == b'\xff\xc0\xff\xe1JFIF\x00'

    # Overwrite header with new image dimensions.
    jfif[6], jfif[7] = struct.unpack('>HH', struct.pack('>II', width,
        height))[::-1]

    # Write out data section of JFIF.
    out.write(jfif[2:-5])

