#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + datalen.to_bytes(3, 'big'))
    out.write(data)

# Generates a random JPEG file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_jpeg(
        rng: BinaryIO, out: BinaryIO,
        blocksize: int=256
    ):
    header = b'\xff\xe0\x00'
    padding = [b'\x00'] * blocksize
    out.write(header + len(data).to_bytes(3, 'big'))
    while byte := rng.read(blocksize):
        if not byte and not padding:
            break
            
            out.write(byte[:blocksize])

# Generates a ZIP file 
def generate_random_zip(
        rng: BinaryIO, out: BinaryIO,
        maxfilesize: int=1 << 24
        ):
    assert(maxfilesize > 0)
    def generate_random_filename():
        filename = []
        for i in range(16):
            filename.append(chr(int.from_bytes(rng.read(1), byteorder='little') & 0b111111))
        return bytes(''.join(filename), 'ascii').decode()

    # ZIP Central directory structure
    centraldir = []
    # ZIP local File Header format (https://en.wikipedia.org/wiki/Zip_(file_format)#File_headers)
    localfileheader = []
    # ZIP File Data format (https://en.wikipedia.org/wiki/Zip_(file_format)#File_data)
    filedata = []

    num_of_files = int.from_bytes(rng.read(2), byteorder='little') & 0xffff
    fileoffset = 0
    for fidx in range(num_of_files):
        print(f"Generating file {fidx+1}/{num_of_files}...")

        # File name
        filename = generate_random_filename().encode("cp437", errors="replace").decode()
        # System/creation date
        epoch = int.from_bytes(rng.read(4), byteorder='little')
        timestamp = struct.pack("<L", (epoch >> 8)).decode()
        # Modified time
        timedata = struct.pack("<H", (epoch >> 24) | ((epoch & 0xff00) << 8)).decode()
        # Version made by
        version = rng.read(2)
        # Extraction method
        extractionmethod = rng.read(2)
        # General purpose bit flag
        generalpurposeflag = rng.read(2)
        # Compression type
        compressiontype = rng.read(2)
        # Last modified time
        lastmodifiedtime = rng.read(2)
        # Last modified date
        lastmodifieddate = rng.read(2)
        # CRC-32
        crc32 = int.from_bytes(rng.read(4), byteorder='little').to_bytes(4, 'little')
        # Compressed size
        compsize = int.from_bytes(rng.read(4), byteorder='little').to_bytes(4, 'little')
        # Uncompressed size
        uncompsize = int.from_bytes(rng.read(4), byteorder='little').to_bytes(4, 'little')
        filename_len = len(filename).to_bytes(2, 'little')
        extrafield_len = rng.read(2)

        # Write to ZIP Local File Header
        localfileheader += [
                localfileheader_version,
                bitflag,
                time,
                date,
                zip_method,
                lastmodtime,
                lastmoddate,
                crc32,
                comp_sz,
                uncomp_sz,
                filename_len,
                extrafield_len]

        # Write to ZIP Data
        filedata += [
                b'\x50\x4b\x03\x04',
                bitflag,
                time,
                date,
                zip_method,
                lastmodtime,
                lastmoddate,
                crc32,
                comp_sz,
                uncomp_sz,
                filename_len,
                extrafield_len]
        filedata += [filename]
        filedata += [extrafields]
        filedata += [
                b'\x50\x4b\x01\x02',
                b'\x1e',
                b'\x03\x0a\x00',
                len(data)]
        filedata += [data]
        filedata += [
                b'\x50\x4b\x05\x06',
                b'\x00\x00']

        # Write to ZIP Central Directory Structure - https://en.wikipedia.org/wiki/Zip_(file_format)#Central_directory_structure
        centraldir += [
                centraldir_version,
                created_os,
                extracted_os,
                general_purpose_flags,
                compression_method,
                last_modification_time,
                last_modification_date,
                crc32,
                compressed_size,
                uncompressed_size,
                filename_length,
                extra_field_length,
                comment_length,
                start_disk_number,
                internal_file_attributes,
                external_file_attributes,
                offset]

        fileoffset += (
                1024 +
                len(filename) +
                len(extrafields) +
                10 +
                len(data) +
                18)

    # Calculate the size of the central directory
    centraldir_len = sum([len(elem) for elem in centraldir])
    
    # Write to output
    out.write(b'\x50\x4b\x03\x04\x0a\x00\x00\x00\x00\x00')
    out.write(centraldir_len.to_bytes(4, 'little'))
    out.write(fileoffset.to_bytes(4, 'little'))
    out.write(b'\x00\x00')

