#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # The width of the image is to be between 1 and 256 (inclusive) pixels.
    w = rng.read(1)[0] % 257 + 1

    # The height of the image is to be between 1 and 256 (inclusive) pixels.
    h = rng.read(1)[0] % 257 + 1

    # A color table of at least one entry must be present in order for the
    # decompression algorithm to function properly. This table can have any size,
    # though it's recommended that its length not exceed 256. We'll give it up to
    # seven entries so we don't need to worry about padding the GIF header with
    # null bytes.
    ncolbits = max((w * h).bit_length(), 1) - 8
    ncols = 1 << ncolbits
    if ncols > 256:
        raise Exception("Too many colors!")

    coltable = []
    for i in range(ncols):

        # Each color component should be an unsigned byte value from 0-255.
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        a = rng.read(1)[0] or 255

        coltable += [r,g,b,a]

    # Write the GIF header
    out.write(
            b'GIF89a' + \
            struct.pack("<HHBBB", w, h,
                        ncolbits,
                        0,
                        0,
                )
        )

    # Write the local color table
    out.write(bytes([ncols]))
    for c in coltable:
        out.write(struct.pack('<B', c))


    #...and now comes the fun part!
    imagedata = bytearray()
    for xpos in range(w):
        lzwmincodesize = min(rng.read(1), 255)
        lzwmaxcodesize = min(int(lzwmincodesize*1.4)+1, 255)
        lzwsizesize = rng.read(1)[0]+1
        lzwnextcode = 2 ** lzwmincodesize
        lzwstreamsize = int(h / ((float)(lzwmaxcodesize + 1.0) /
                              float(lzwmincodesize + 1.0)))
        
        # LZW code sizes will always be between the minimum and maximum values given above.
        lzwcodesizes = bytes([(rng.read(1)[0] % (lzwmaxcodesize - lzwmincodesize + 1)) + lzwmincodesize])

        lzwcodes = {}
        currentcode = None

        def nextlzwbyte():
            nonlocal lzwcode, currentcode

            while len(lzwcode) < lzwcodessize:

                # If there aren't enough bits left over in this byte for another code,
                # read more data from the stream.
                if (currentcode == None) or (len(currentcode) <= 8):

                    # If we're at EOF, stop reading
                    if len(imagedata) >= lzwstreamsize:
                        break

                    currentcode = struct.unpack('B', imagedata[len(imagedata)-1])[0].to_bytes(1, 'little')
                else:
                    currentcode = bytes([currentcode[(len(currentcode)-8):]])
                
                lzwcode += currentcode


            try: return lzwcode[:lzwsizesize]
            finally:
                lzwcode = lzwcode[lzwsizesize:]


        # Read some LZW codes to populate our dictionary
        while nextlzwbyte()!= bytes([255]):
            lzwcode = nextlzwbyte()
            
            # Parse out the index from the first two bytes, then use the second as
            # the key, and append the rest of the code to the end.
            idx = (lzwcode[0]<<8|lzwcode[1]).to_bytes(2,'big')
            lzwcodes[idx] = lzwcode[2:]

            if lzwnextcode == 4095:
                lzwcodessize = lzwcodessize+1
                lzwnextcode = 2 ** (lzwcodesizes[-1] + 1)
            elif lzwnextcode == 65535:
                lzwcodessize = lzwcodessize+1
                lzwnextcode = 2 ** (lzwcodesizes[-1] + 2)
            lzwnextcode += 1

        assert(len(lzwcodes) == 2**lzwmincodesize)

        # Just because we've started filling up our dictionary doesn't mean we want
        # to actually store all those entries inside our image data stream; we just
        # want to make sure they're referenced somewhere in our dictionary. Therefore,
        # we keep track of which indices are being used in each line and only write
        # them when necessary to save space. Because we know how long these lines 
        # will be beforehand, we'll also load them ahead of time so we don't have to
        # reallocate them every frame.
        lineusedlist = []
        linetable = []
        lineusedmask = 0x80 >> (xpos % 8)
        for ypos in range(h):
            lineused = False
            lastindex = b''
            while True:
                lzwcode = nextlzwbyte()

                # We've reached the end of this line
                if lzwcode == bytes([255]):
                    imagedata += lzwcode
                    break

                index = (lzwcode[0]<<8 | lzwcode[1]).to_bytes(2,'big')

                # If we haven't seen this sequence yet, add it to our main list
                if index not in lzwcodes:
                    lzwcodes[index] = lastindex[0:1] + lzwcode[2:]
                    imagedata += lzwcode
                else:
                    if lineused:

                        # Here we loop until we find a previously encountered window
                        # where this code has already been stored.
                        found = False
                        while (not found):
                            for i in reversed(range(len(linetable))):
                                linetableentry = linetable[i]

                                # Keep looping until we either find the previous
                                # entry...
                                if (index in linetableentry['entries']):

                                    #...or run off the beginning of the list
                                    if (i == 0):
                                        linetable[i]['entries'].append(index)
                                        imagedata += lzwcode
                                        found = True
                                        break

                                    # If we found the previous instance of the 
                                    # matching entry, swap it with the current one,
                                    # and update the reference count on that line.
                                    oldlinetableentry = linetable.pop(i)
                                    linetableentry['entries'].remove(index)
                                    linetableentry['entries'] += [(lastindex[0:1] +
                                                                lzwcodes[index][1:])
                                                               ]
                                    oldlinetableentry['refcount'] -= 1
                                    if oldlinetableentry['refcount'] == 0:
                                        linetable.insert(0,oldlinetableentry)
                                    imagedata += (oldlinetableentry
                                                  ['prevsequence'][0:1] + 
                                                  lzwcodes[index][1:])
                                    found = True
                                    break

                            # Didn't find anything? That means we're making a new 
                            # uniqe entry here; no problem, just continue.
                            if (not found):
                                linetableentry = {
                                        'prevsequence': (lastindex[0:1]
                                                        + lzwcodes[index]),
                                        'entries': [index],
                                       'refcount': 1
                                        }
                                linetable.append(linetableentry)
                                imagedata += lzwcode
                    
                    else:
                        # Append this sequence onto the last sequence in our list
                        imagedata += (lastindex[0:1] + lzwcodes[index])
                    

                # Update the state variables for the next iteration.
                lastindex = index
                lineused |= bool(lineusedmask & ord(lzwcode[2]))
                
            # Save whether or not this entire line was actually used by the decoder
            lineusedlist.append(lineused)
            lineusedmask >>= 1
            if lineusedmask == 0:
                lineusedmask = 0x80

        # Okay, we've finished parsing through every line! Printing out all the info
        # we collected makes it easier to debug later.
        print()
        print('LZW min codesize:', lzwmincodesize)
        print('LZW max codesize:', lzwmaxcodesize)
        print('LZW code size size:', lzwsizesize)
        print('LZW stream size:', lzwstreamsize)
        print('Line used mask:', hex(lineusedmask))
        print('# of Lines:', len(lineusedlist))
        print()
    
        # Now let's go back through again, and insert references to each unique entry
        # we saw in this entire picture.
        for i in reversed(range(len(linetable))):
            linetableentry = linetable[i]
            for j in range(len(lineusedlist)):
                # Only write down the sequences that were used in any way
                if lineusedlist[j]:
                    # Insert a copy of the entire entry every time it's used