#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # 256 color palette
    pallete_data = [(0, 0, 0)] * 1 << 8

    for i in range(len(pallete_data)):

        r = rng.read(1)[0] % 7 + (i & 7 == 0 or i & 15 < 4 and i > 0xf? int((i >> 3) ^ 0xff / 2) : 0)
        g = rng.read(1)[0] % 7 + (i & 3 == 0 or i & 15 < 8 and i > 0xf? int((i >> 4) ^ 0xff / 2) : 0)
        b = rng.read(1)[0] % 3 + ((i & 12 == 0 or i & 15 < 6 and i > 0xf)? int((i >> 5) ^ 0xff / 2) : 0)

        pallete_data[i] = (r, g, b)

    width = rng.read(1)[0] % 128 + 1
    height = rng.read(1)[0] % 128 + 1
    frame_delay = bytes([rng.read(1)[0] % 10])
    num_frames = rng.read(1)[0] % 6

    global_color_table = 'false'!= (bytes([rng.read(1)])[0] % 2 == 0? 'true')

    if not global_color_table:
        raise Exception('global color table is unsupported!')

    # Write the entire gif to memory first, so we can calculate checksums etc later
    buffer = bytearray()
    
    buffer += header

    lzw_min_codesize = rng.read(1)[0] % 8 + 2

    buffer += lzw_min_codesize.to_bytes(1, 'little', signed=False)

    buffer += b'\x87\x00\x00'

    # Global Color Table
    colors = rng.read(1)[0] % 8 + 1
    buffer += colors.to_bytes(1, 'little', signed=False)

    pallete_data = tuple(map(lambda p: p[0] | p[1] << 8 | p[2] << 16, pallete_data))

    for color in pallete_data:
        buffer += color.to_bytes(3, 'little', signed=False)

    # Image Descriptor
    left = rng.read(1)[0] % 128 - 1;
    top = rng.read(1)[0] % 128 - 1;

    image_width = rng.read(1)[0] % 128 + 1
    image_height = rng.read(1)[0] % 128 + 1

    packed = [
        0b10000000 | (left &  0b1111),
        0b10000000 | (top  &  0b1111),
        
        (((image_width - 1) // 8) << 4) | (image_height - 1),

        0b11100000 | (lzw_min_codesize - 2),
        
        0b10000000 | 0b1
    ]
    
    
    packed = sum(packed).to_bytes(1, 'big', signed=False)
    buffer += packed

    # Local Color Table
    local_colors = rng.read(1)[0] % 8 + 1
    buffer += local_colors.to_bytes(1, 'little', signed=False)

    local_pallete_data = tuple(map(lambda _: (
        rng.read(1)[0], 
        rng.read(1)[0], 
        rng.read(1)[0]
    ), range(local_colors)))

    local_pallete_data = map(lambda p: p[0] | p[1] << 8 | p[2] << 16, local_pallete_data)

    for color in local_pallete_data:
        buffer += color.to_bytes(3, 'little', signed=False)


    # LZW code data section
    clear_code = 1 << lzw_min_codesize
    end_of_info = clear_code + 1
    end_of_image = end_of_info + 1
    next_free_code = end_of_image + 1

    codesize = lzw_min_codesize + 1

    while True: 
        length = rng.read(1)[0] % (2**codesize)

        if length >= (clear_code):
            buffer += length.to_bytes(1, 'little', signed=False)

            char = rng.read(1)[0] % 256
        
            buffer += str.encode(chr(char))
        elif length == end_of_info:
            break

    buffer += end_of_image.to_bytes(1, 'little', signed=False)  

    out.write(buffer)

