#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = 1<<24

    # Start with a header and logical screen descriptor
    out.write(header + struct.pack('<HHB',
                                    500, 675, 255))

    # Write an animation control block (no loops!)
    out.write(b'\x21\xF9\x04\x08\x00\x00')

    # Write some color tables
    for i in range(2):

        tablelen = 2<<(i+1)+1
        palette = [struct.pack('BBB',
                                *rng.read(3))
                   for j in range(tablelen)]
        
        lctsize = len(palette).to_bytes(2, 'little')
        lctinfo = bytes([
            0,                      # LCT size - dynamic length
            pow(2, i+1)-1,         # LCT sort flag
            0]                     # LCT entry size
        )

        gctsize = len(palette).to_bytes(2, 'little')
        gctinfo = bytes([
            0,                      # GCT size - dynamic length
            pow(2, i+1)-1          # GCT sort flag
        ])

        out.write((lctsize + lctinfo + gctsize + gctinfo
                       ).replace(b'\x00', b''))
 
        if not all(map(lambda x: len(x)==3, palette)):
            raise ValueError("Invalid colors generated")

        
        out.write(b"".join(reversed(palette)))

    # Write separator byte
    out.write(b'\x2C')

    # Write one graphics control block
    out.write(struct.pack("<BHBHBH",
                          *(0xf9,
                            0, 0,
                            (datalen >> 1)&0xff,
                            ((datalen>>9)|0x8000),
                            0)))

    # Write image descriptor
    out.write(struct.pack("<BHHHHHH",
                          *(0x2c,
                            0, 0,      # X, Y position of top left pixel
                            datalen,    # Width of image
                            0,          # Height of image
                            0,          # Local color table index
                            0           # Interlace bit
                        )))

    # Write data sub-block
    out.write(b"\x02" + datalen*b'A'
                             .replace(b'A',
                                          rng.read(1)[0].to_bytes(1,'big')))

