#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(rng.read(2), 'little') & 0xffff

    # Image Header
    ihdr = [0] * 10
    ihdr[0] = 0x2c                   # image separator
    ihdr[1] = (int.from_bytes(rng.read(4), 'little')) & 0xff
    ihdr[2] = (int.from_bytes(rng.read(4), 'little')) & 0xff
    ihdr[3] = (int.from_bytes(rng.read(4), 'little')) & 0xff
    ihdr[4] = (int.from_bytes(rng.read(4), 'little')) & 0xff
    ihdr[5] = (int.from_bytes(rng.read(4), 'little')) & 0xf       # local colour table exists
    ihdr[6] = ((int.from_bytes(rng.read(1), 'little')) & 0x7f) + 1   # LZW minimum size in bits
    ihdr[7] = (int.from_bytes(rng.read(2), 'little')) & 0xff      # background index
    ihdr[8] = (int.from_bytes(rng.read(2), 'little')) & 0xff      # pixel aspect ratio

    # Global Colour Table
    global_ctable = []
    if ihdr[5]:                       # has local color table
        for i in range(0, pow(2, 1 << ihdr[5])):
            r = (int.from_bytes(rng.read(1), 'little', signed=True)) & 0xff
            g = (int.from_bytes(rng.read(1), 'little', signed=True)) & 0xff
            b = (int.from_bytes(rng.read(1), 'little', signed=True)) & 0xff

            global_ctable += [b, g, r]

        global_ctable *= math.ceil(datalen / len(global_ctable))
    else:                               # no local color table
        ctablelen = (int.from_bytes(rng.read(1), 'little')) & 0xff    # GCT length is only byte!

        for i in range(0, ctablelen):
            r = (int.from_bytes(rng.read(1), 'little', signed=True)) & 0xff
            g = (int.from_bytes(rng.read(1), 'little', signed=True)) & 0xff
            b = (int.from_bytes(rng.read(1), 'little', signed=True)) & 0xff

            global_ctable += [r, g, b]
        
        while len(global_ctable) < pow(2, ctablelen) * 3:
            global_ctable += [(int.from_bytes(rng.read(1), 'little')) & 0xff]
    
    # Frame Descriptor
    fdsz = 10
    fd = [0] * fdsz
    fd[0] = 0x21                     # graphics control extension

    for i in range(fdsz - 1, 1):
        fd[i] = (int.from_bytes(rng.read(1), 'little')) & 0xff

    fd[1] = (int.from_bytes(rng.read(1), 'little')) & 0xff            # separator
    fd[2] = 0x04                                                       # block size

    fd[3] = (int.from_bytes(rng.read(1), 'little')) & 0xff            # delay time low
    fd[4] = (int.from_bytes(rng.read(1), 'little')) & 0xff            # delay time high
    fd[5] = (int.from_bytes(rng.read(1), 'little')) & 0xff            # transparent colour index
    fd[6] = *(struct.pack('B', 0)),                                     # terminator

    fd[7] = 0x2c
    fd[8] = *(struct.pack('HHBBBBBBBHBHHBBBBBHBBBBBHBBBBBH', \
             ihdr[0],                                                      # image descriptor block
             ihdr[1],                                                      # left
             ihdr[2],                                                      # top
             ihdr[3],                                                      # right
             ihdr[4],                                                      # bottom
             0x00,                                                          # not interlaced
             0x08 | ihdr[5],                                               # no sort flag set
             ihdr[6],                                                      # lzw min code size

             ihdr[7],                                                      # bg index
             ihdr[8],                                                      # pixel aspect ratio
             ihdr[1],                                                      # left
             ihdr[2],                                                      # top
             ihdr[3],                                                      # right
             ihdr[4],                                                      # bottom

             0x01,                                                          # transparency flag set
             ihdr[5],                                                      # lzw min code size
             ihdr[8]),                                                     # pixel aspect ratio

    fd[9] = 0

    first_frame = True
    framecount = 0

    while True:
        out.write(header)
        out.write(*(struct.pack('<H', datalen)))
        print("Writing DATALEN")

        for i in range(0, len(ihdr)):
            out.write((struct.pack('B', ihdr[i])))
            print("Writing IHDR["+str(i)+"]")
    
        for i in range(0, len(global_ctable)):
            out.write((global_ctable[i]).to_bytes(1, 'little', signed=False))
            print("Writing GLOBAL_CTABLE["+str(i)+"]")

        for i in range(0, len(fd)):
            out.write((struct.pack('B', fd[i])))
            print("Writing FRAMEDESCRIPTOR["+str(i)+"]")

        if first_frame:
            first_frame = False
        elif framecount == 0:
            break

        framecount -= 1

    return

if