
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Write the header of 64 bytes to out
    write_bytes(b"GIF89a", out)
    
    width = randint(100,255) * 16 + (randint(0,7)+1)
    height = randint(100,255) * 16 + (randint(0,3)+1)

    # Write the Logical Screen Descriptor
    write_byte(width >> 8, out);     # Width high byte
    write_byte(width & 0xFF, out);   # Width low byte
    write_byte(height >> 8, out);    # Height high byte
    write_byte(height & 0xFF, out);  # Height low byte
    # Bit packed field:
    # - Global Color Table Flag is set
    # - Sort Flag is set to Standard RGB
    # - Size of Global Color Table is 0
    # - Background Color Index is 0
    # - Pixel Aspect Ratio is 0x10 (1:1)
    write_byte((0b101 << 4)|(0b0<<3)|(0b0), out);
    # Transparent color index
    write_byte(0xF, out)

    # Write the Image Descriptor
    # - Left Position Byte = 0
    # - Top Position Byte = 0
    # - Width Byte = 1...n with n in {1..255}
    # - Height Byte = 1...m with m in {1..255}
    # - Packed field: Reserved should be 0, and LZW Compression Method is used.
    #                 so bit 4=1, bits 3-2 are 0, and bit 1 is always 1 when not zeroed
    write_byte(0b0 | ((width-1)//16), out) 
    write_byte(((height-1)//16)<<4, out)
    write_byte(0b10<<3|0b0, out)

    palette = [i for i in range(3*256)]
    shuffle(palette, rng)
    for p in palette:
        if type(p)==tuple or type(p)==list:
            write_byte(p[0], out)
            write_byte(p[1], out)
            write_byte(p[2], out)
        else:
            assert False, "Palette element must contain rgb values as either list or tuple!"
    
    # Write size of remaining data (used by size field)
    num_colors = randint(2,16)
    global_color_table_size = 3*(num_colors**2)
    remaining_data_size = global_color_table_size+1;
    remaining_data_size += width*height

    # Write LZW minimum code length (must be > 0 and < 9)
    lzw_min_code_length = max(
        1, min(
            8, 
            randint(2,remaining_data_size//global_color_table_size)
        )
    )
    write_byte(lzw_min_code_length, out)

    # Write LZW maximum code length (must be >= LZWMInCodeLength)
    write_byte(max(lzw_min_code_length+1, randint(lzw_min_code_length+1, 12)), out)

    # Write actual image data block
    colors = [(randint(0,255),randint(0,255),randint(0,255)) for _ in range(num_colors)]
    compress_lzw([
        encode_pixel(color, randint(-16,-4), randint(-16,-4)) 
        for x in range(width) 
        for y in range(height)], 
        lzw_min_code_length, out, colors)
    