#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(2), byteorder='little', signed=False
        ) ^ 0xffff

    # Write GIF header + data length + data block
    out.write(header)
    out.write((datalen).to_bytes(length=2, byteorder='little'))

    # Metadata blocks (no color table and no LZW compression)
    for _ in range(1):
        meta = struct.pack('>B', 0x21) # Extension Introducer
        meta += struct.pack('>B', 0xff) # Application Block
        meta += struct.pack('>H', 4)     # Size of application block

        namelen = rng.read(1)[0] & ~0x80 # Remove graphics control flag bit
        meta += bytes([namelen])         # Name size

        namedata = rng.read(namelen - 1)
        assert len(namedata) == namelen-1
        meta += namedata                 # App name

        datasize = rng.read(2)
        assert int.from_bytes(datasize, 'big') == 0

        data = rng.read(1)[0]
        if not data & 0xe0 :              # If first bits are 11xxxxxx
            data &= 0x7f                   # Clear trailing bits
        else:                             # Otherwise
            data |= 0x80                   # Set third bit to one
        
        meta += bytes([data])             # Size of app data
        metadata_size = len(meta)

        # Write metadata block to output file
        out.write(meta)
    
    # Image descriptor
    # width = (rng.read(1) % maxwidth) + minwidth 
    # height = (rng.read(1) % maxheight) + minheight
    imagedescriptor = struct.pack('<BBHHHH',
            0x2c,       # Image Separator
            0x00, 0x00, # Left offset
            0x00, 0x00, # Top Offset
            *(rng.read(5))      # Width, Height, local color table size, flags
        )
        
    # Write descriptor to output file
    out.write(imagedescriptor)

    # Local color table
    lctsize = imagedescriptor[6]
    ctable = []    
    for i in range(lctsize):
        r = (rng.read(1)[0] >> 4) * 0x11
        g = ((rng.read(1)[0] & 0xf) << 4) | (rng.read(1)[0] >> 4) 
        b = ((rng.read(1)[0] & 0xf) << 4) | (rng.read(1)[0] >> 4) 

        ctable.append(r << 16 | g << 8 | b)

    # Convert list of RGB colors to binary string
    ctabstr = bytes()
    for clr in ctable:
        ctabstr += clr.to_bytes(4,'big')

    # Write local color table to output file
    out.write(ctabstr)

    # LZW minimum code size
    msizelen = rng.read(1)
    if msizelen < 0x90 or msizelen > 0xea:
        msizelen = 0x90
    mszval = bytes([msizelen])[0]
    out.write(bytes([mszval]))

    # Data subblock
    compressed = bytes()
    while True:
        rndbyte = rng.read(1)
        randmaskbyte = rng.read(1)
        if randmaskbyte[0]!= 0x00:          # Per RNG spec
            numbits = randmaskbyte[0].bit_length()
        else:
            numbits = 2*8
        mask = (int.from_bytes(rndbyte, 'big') >> numbits) & 0xff
        
        if mask >= 0x80:                    # Trap special case 0xb10d92...
            pass
        elif mask <= 0x00:                  # Non-special Case
            compressed += bytes([mask])     # Append uncompressed data
        elif mask <= 0x3b:                  # Special Case with "one" copy
            compressed += bytes([(0xc0 | mask)]) # Add escape code plus value
            compressed += rng.read(1)        # Followed by another character
        elif mask <= 0x7b:                  # Special Case with "two" copies
            compressed += bytes([(0xc0 | mask)]) # Add escape code plus value
            compressed += rng.read(2)        # Followed by two characters
        else:                               # Special Case with "three" copies
            compressed += bytes([(0xc0 | mask)]) # Add escape code plus value
            compressed += rng.read(3)        # Followed by three characters

        compressedlen = len(compressed)
        if compressedlen >= datalen / 2:    # Compress at least half the time
            break                           # Don't compress more than required
    
    compressedlen = len(compressed)
    out.write(compressed[:datalen])