#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

        
    background_color_index, pixel_aspect_ratio = \
        struct.unpack('BB', rng.read(2))


    flags='\x00\x00\x07';            # gct sort flag: yes
    if image_height <= 255:          # global color table size required
        flags+= '\x00\x00\x00\xFF';    # lzw minimum code size (unused)
    else:                           # no global color table
        flags += '\xFF\x00\x00\x00'   # lzw minimum code size (unused)
    flags += '\x00\x00'             # No interlace or sorted gct
    
    # Write "logical screen descriptor"
    out.write((header +
               bytes([image_width & 0xff,
                      image_width >> 8,
                      image_height & 0xff,
                      image_height >> 8]))
              )
    # Write "global color table" section, if required
    if image_height > 255:
            for i in range(0, 256):
                out.write(bytes([i,
                                 rng.read(3)[0],        # R
                                 rng.read(3)[1],        # G
                                 rng.read(3)[2]         # B
                                ]));
    
    
    # Write "image descriptor" section
    out.write(bytes([flags[0],
                     flags[1],
                     flags[2],
                     flags[3],
                     image_width & 0xff,
                     image_width >> 8,
                     image_height & 0xff,
                     image_height >> 8])
             );
    # Write "local color table", if present
    out.write(b'\x00');           # local color table flag
            
    # Write actual image data
    LZW_min_code_size = int.from_bytes(rng.read(1), 'big')
    print("LZW min code size:", LZW_min_code_size);
    while True:
        byte = rng.read(1)
        # EOF reached
        if not len(byte): break

        if ord(byte[0]):     # extension block type
            ext_type = rng.read(1).decode()

            if ext_type == '!':       # comment block
                while True:
                    buffer = rng.read(1)
                    if str(buffer)!= b'\x00':
                        continue
                    break
            
            elif ext_type == ',':     # application block 
                app_block_size = int.from_bytes(rng.read(1), 'big')
                app_id = rng.read(8).decode().strip('\x00')

                if app_id == 'NETSCAPE':
                    sub_len = int.from_bytes(rng.read(2), 'big')

                    if sub_len >= 3:
                        app_sub_data = rng.read(sub_len-3)

                        if len(app_sub_data)==3:
                            app_version, loop_count, terminator = \
                                    map(lambda x : int.from_bytes(x, 'big'),
                                         [app_sub_data[:1],
                                          app_sub_data[1:2],
                                          app_sub_data[2:]]);

                            if app_version==1 and loop_count!= 0:
                                raise ValueError("A Netscape Looping Extension Block with version 1 must have a nonzero loop count")

                            if terminator!=0:
                                raise ValueError("A Netscape Looping Extension Block must end with null byte");
                
                else:                       # ignore unknown applications
                    rng.seek(app_block_size - 1, 1)

            
                
            elif ext_type == ';':     # graphics control block
                indx=0;
                while indx < 8:      # read until we don't get another byte 
                    ctrl_byte = rng.read(1);
                    
                    if ctrl_byte == b'\x00': 
                        break;
                    
                    indx+=1
                    
                gc_block_size = int.from_bytes(rng.read(1), 'big')
                
                if gc_block_size >= 4:
                    bitfield = int.from_bytes(rng.read(2), 'big') & 7
                    print("GC bitfield:", bin(bitfield))

                    if bitfield == 0:                 # transparency is on
                        transparent_index = int.from_bytes(rng.read(1), 'big')
                        
                        if gc_block_size >= 5:
                            delay = int.from_bytes(rng.read(2), 'big')
                            
                        if gc_block_size >= 7:
                            trans_size = int.from_bytes(rng.read(1), 'big')
                            
                            if gc_block_size >= 8:
                                skip = rng.read(gc_block_size - 8)
                    
                elif gc_block_size >= 1:
                    skip = rng.read(gc_block_size - 1)
                
            else:                     # ignore other extensions blocks
                ext_len = int.from_bytes(rng.read(1), 'big')
                rng.seek(ext_len - 1, 1)


            
        elif ord(byte):     # plain text block
            txt_block_size = int.from_bytes(rng.read(1), 'big')
            text = rng.read(txt_block_size).decode();



        
        else:               # graphic control label
            ctl_code = int.from_bytes(rng.read(1), 'big')
            sz = 1
            if ctl_code == 0x04:
                sz = 6
            elif ctl_code >= 0x08 and ctl_code <= 0x0f:
                sz = 1+int(((ctl_code&0x07)+1)<<1);
            
            skip = rng.read(sz-1)



        result=[]
        while True:
            buffer = rng.read(1)
            if str(buffer)!= b'\x00':
                result.append(str(buffer))
                continue
            break


        out.write(struct.pack('<B', len(result))) 
        for codepoint in result: 
            out.write(codepoint.encode())
        



    return (image_width, image_height, background_color_index)

