#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + datalen.to_bytes(3, 'big'))
    out.write(data)

# Generates a random JPEG file into `out` using the
# random number genrator `rng` (/dev/urandom).
def generate_random_jpeg(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1024 bytes of binary random data, to be written
    # afterwards as JFIF-formatted image.
    jfif = rng.read(1024)

    # Create the exif IFD tag. It consists of two fields:
    #   * TAGID=256 (byte order marker) and TYPE=2 (short):
    #      - ExifVersionID is 4 bytes long for some reason...
    #    * TAGID=271 (resolution units) and TYPE=1 (byte)
    ifdtag = \
         b'\xFF\xE1\x00\x0E\x45\x78\x69\x66\x00\x00\x49\x49\x2A\x00' \
       + b'\x08\x00\x00\x00\x01\x00\x12\x00\x03\x00\x00\x00\x01\x00' \
       + b'\x01\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
       + b'\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x8E\x9C' \
       + b'\x3B\x00\x00\x00\x00\x41\x43\x44\x20\x46\x49\x46\x46\x00' \
       + b'\x01\x01\x00\x00\x01\x00\x01\x00\x00\x3B'

    # Append our random data at the end of it
    jfif += rng.read((510 - len(jfif)) // 8 * 8)

    # Write all metadata immediately after SOI token
    start_of_image = b'\xff\xd8'
    out.write(start_of_image + ifdtag + jfif)


# Pad size up to next multiple of alignment
def align(size: int, alignment: int):
    return ((size + alignment - 1) // alignment) * alignment

# Use `rng` to create an empty ELF file in memory that has its entrypoint set
# to `entry`. The first section will contain a single loadable segment with
# contents from `code`, and the second one will execute code from the same area
# once loaded. If `text` is specified, it will also be copied there.
def make_elf(
        rng: BinaryIO, entry: int,
        code: BinaryIO, text: BinaryIO|None = None
    ) -> bytearray:
    elf = bytearray()

    # File header
    elf += b'\x7fELF'     # Magic
    elf += b'\x02\x01\x01'   # Class / Data encoding
    elf += b'\x00\x00'       # Version
    elf += b'\x00\x00'       # OS ABI
    elf += b'\x00\x00'       # ABI version
    elf += struct.pack('>H', 256 | 32) # Etype
    elf += b'\x00\x00'       # Machine type (ARM little endian)

    # Program header table
    elf += b'\x02\x00'         # Number of segments
    elf += b'\x00\x00\x00\x00' # Segment header offset
    elf += b'\x00\x00\x00\x00' # Section header offset
    elf += b'\x00\x00\x00\x00' # Flags
    elf += b'\x00\x00\x00\x00' # Initial virtual address
    elf += b'\x00\x00\x00\x00' # Virtual memory size
    elf += b'\x00\x00\x00\x00' # Physical memory size
    elf += b'\x00\x00\x00\x00' # Alignment

    # Segment header[0]
    elf += b'\x01'             # Type: LOAD
    elf += b'\x00\x00\x00\x00' # Offset within the file
    elf += b'\x00\x00\x00\x00' # Virtual addr of the segment
    elf += b'\x00\x00\x00\x00' # Physical addr of the segment
    elf += struct.pack('<Q', align(len(code.getbuffer()), 0x1000))
                            # Size of this segment
    elf += b'\x00\x10\x00\x00' # Virtial addr of the section
    elf += b'\x00\x00\x00\x00' # Physical addr of the section
    elf += b'\x00\x00\x00\x00' # Flags
    elf += b'\x00\x00\x00\x00' # Alignement

    # Segment header[1]
    elf += b'\x01'             # Type: LOAD
    elf += b'\x00\x00\x00\x00' # Offset within the file
    elf += b'\x00\x00\x00\x00' # Virtual addr of the segment
    elf += b'\x00\x00\x00\x00' # Physical addr of the segment
    elf += b'\x00\x00\x00\x00' # Size of this segment
    elf += b'\x00\x00\x00\x00' # Virtial addr of the section
    elf += b'\x00\x00\x00\x00' # Physical addr of the section
    elf += b'\x00\x00\x00\x00' # Flags
    elf += b'\x00\x10\x00\x00' # Alignement

    elf += b'\x00\x00'       # 1st string table index
    elf += b'\x00\x00'       # 2nd string table index

    # Program header table entries filled later on demand by mmap
    elf += b'\x00'*4*4

    # String table
    elf += b'/lib/ld-linux.so.3' # First string: dynamic loader path
    elf += b'\x00'                 # Null terminator

    # Actual program code goes here
    elf += code.getbuffer()

    # Text section content, if any
    if text!= None:
        elf += text.getbuffer()

    # Fill program headers
    elf[52:56] = struct.pack('<L', len(elf))           # Elf header length
    elf[56:60] = struct.pack('<L', len(elf) - 52)     # Program hdr table len
    elf[60:64] = struct.pack('<L', align(len(elf), 0x2000))
                                                        # Program hdrs's aligned

    # Program header table entry [0]
    elf[68:72] = struct.pack('<L', len(elf) - 104)     # Location in file
    elf[72:76] = struct.pack('<L', align(len(elf), 0x2000)
                                                                # Virt addr of seg
    elf[76:80] = struct.pack('<L', align(len(elf), 0x2000)
                                                                # Phys addr of seg
    elf[80:84] = struct.pack('<L', len(efi))            # Size of this segm
    elf[84:88] = struct.pack('<L', 0)                  # Virt addr of sec
    elf[88:92] = struct.pack('<L', 0)                  # Phys addr of sec
    elf[92:96] = struct.pack('<L', 6)                  # Flags

    # Program header table entry [1]
    elf[100:104] = struct.pack('<L', 0)                # Location in file
    elf[104:108] = struct.pack('<L', entry - 0x80400000)
                                                    # Virt addr of the seg
    elf[108:112] = struct.pack('<L', entry - 0x80400000)
                                                    # Phys addr of the seg
    elf[112:116] = struct.pack('<L', len(text or