#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    
    # Header
    width = 1280 * (rng.read(1)[0] & 7) + ((1 << (rng.read(1)[0] % 8)) - 1)
    height = 720 * (rng.read(1)[0] & 7) + ((1 << (rng.read(1)[0] % 8)) - 1)

    print("WIDTH:", width)
    print("HEIGHT:", height)
    global_color_table = bool((rng.read(1)[0]&15)>>4)
    local_color_table = bool((rng.read(1)[0]&15)>>4)
    interlace = False
    sort = True
    color_resolution = 8
    size = 16
    background = 0xFFFFFF
    pixel_aspect_ratio = 0
    gct = b''

    if global_color_table:
        ncolors = 2 << int(size-1)

        for i in range(ncolors):
            r = rng.read(1)[0];
            g = rng.read(1)[0];
            b = rng.read(1)[0];

            if not i:
                transparency = 0xFF

                while transparency == 0 or transparency >= ncolors:
                    transparency = rng.read(1)[0];

            gct += bytes([r,g,b])
        
    else:
        ncolors = 2 << int(size-1)
        
        for i in range(ncolors):
            r = rng.read(1)[0];
            g = rng.read(1)[0];
            b = rng.read(1)[0];
            
            if not i:
                transparency = 0xFF
                
                while transparency == 0 or transparency >= ncolors:
                    transparency = rng.read(1)[0];
                
                gct += bytes([transparency | 0xF0,
                              r,g,b]);
                         
    lct = None
    
    if local_color_table:
        nlct = 2 << int(size-1)
        
        lct = bytearray()
        
        for i in range(nlct):
            r = rng.read(1)[0];
            g = rng.read(1)[0];
            b = rng.read(1)[0];

            lct += bytes([r,g,b]);

        

    # Frame header
    left = 0;
    top = 0;
    width = max(width//10 + rng.read(1)[0], 1);
    height = max(height//10 + rng.read(1)[0], 1);
    flags = 0;
    
    if local_color_table:
        flags |= 1<<9;
    
    if interlace:
        flags |= 1<<8;
    
    if sort:
        flags |= 1<<7;
    
    index_bits = rng.read(1)[0]%3+1;
    border = rng.read(1)[0]%3;

    flags |= index_bits<<5
    flags |= border<<2

    # Image data
    nulls = [0]*9*width + [0xff]*(height*width*((flags&7)+1));

    rng.readinto(nulls);

    lzw = len(nulls).to_bytes(2,'big')
    
    compressed_image_data = bytearray();

    last = '0';

    byte = ''

    for c in nulls:
    
        if last!= str(c):
        
            compressed_image_data += (len(last)*byte).encode('utf-8');
            compressed_image_data += chr(int(last)).encode('utf-8');

            byte = '';
        
        byte += str(c);
        
        last = str(c);
    
    compressed_image_data += (len(last)*byte).encode('utf-8');
    compressed_image_data += chr(int(last)).encode('utf-8');

    end = bytearray.fromhex('3B');

    image_data = bytearray();
    
    for c in compressed_image_data:
        
        if c < 0x21 and c > 0x0f :
            image_data += c.to_bytes(1,'little');
            
        elif c <= 0xf:
            image_data += bytes([0xE0|c]);
            
        else:
            image_data += bytes([(c&0xF0)>>4|0xED,(c&0xF)]);
        

    # Trailer
    trailer_extension_block = 0x01
    trailer_extension_label = ord('.')
    trailer_application_identifier = 'NETSCAPE'
    trailer_application_authcode = '\x03\x01'
    trailer_application_block_length = \
            3 + len(trailer_application_identifier) + len(trailer_application_authcode)
            
    trailer_delay = 0
    trailer_subblocks = 1 + \
              trailer_extension_block + \
              trailer_application_block_length + \
              1 + \
              1 + 2 + \
              1 + 2 + \
              1 + 1 + \
              1 + 1 + \
              1 + 4 + \
              1 + 1 + \
              1 + 1 + 4 + \
              1 + 1 + \
              1 + 4 + \
              1 + 1 + \
              1 + 4;
            
    trailer_extension_block_data = bytearray(
    [0x01,
     (trailer_extension_label & 0x1F),
     (((trailer_extension_label >> 5) +
       (trailer_subblocks    << 3)) & ~0x80)]) + \
      trailer_extension_block.to_bytes(1,'big') + \
      trailer_application_identifier.encode() + \
      trailer_application_authcode.encode() + \
      bytes([trailer_delay//256,
             trailer_delay%256]) + \
      trailer_subblocks.to_bytes(1,'big') + \
      bytes([0x00,
             0x00,
             0x00,
             0x00,
             0x00]) + \
      bytes([0x00,
             0x00,
             0x00,
             0x00,
             0x00]) + \
      bytes([0x1A,
             0x40,
             0xAF,
             0xC8,
             0x00])


    # Write to output stream
    header = bytearray(
    [0x47, 0x49, 0x46, 0x38, 0x37, 0x61, 0x00, 
     0x00, 0x00,
     0x00, 0x00,
     0x00, 0x00,
     0x00, 0x00,
     0x00, 0x00,
     0x00, 0x00,
     0x00, 0x00,
     0x02, 0x00,
     0x1C, 0x00,
     0xFF, 0xFF, 
     0x0B,
     0x4D, 0x45, 0x54, 0x41, 0x2D, 0x49, 0x4E, 0x46, 0x2F, 0x67, 0x69, 0x66, 0x38, 0x37, 0x61, 0x00, 
     0x3B]);
      
    out.write(header)
    out.write(struct.pack('<H', width))
    out.write(struct.pack('<H', height))

    out.write((global_color_table << 7) + (local_color_table << 6) + 
        (interlace << 5) + (sort << 4) + 
          (size-1 << 0))

    out.write(background.to_bytes(3,'big'))
    out.write(pixel_aspect_ratio.to_bytes(1,'little'))

    if global_color_table:
        out.write(struct.pack("<B", ncolors))
        out.write(gct)

    if local_color_table:
        out.write(struct.pack("<B", nlct))
        out.write(lct)

    frame_header = [left,top,width,height,flags]

    for f in frame_header:
        data = bytearray(struct.pack("<H", f))

        out.write(data + padding)
