#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header structure is as follows (see https://www.w3.org/Graphics/GIF/spec-gif89a.txt):
    # ```
    #     typedef struct {
    #         char    Signature[3];
    #         uint16  Version;      /* Must be '87a', '89a' or '89b'. */
    #         GifScreenDescriptor   ScreenDescriptor;
    #     } GifImageFileHeader;
    # 
    #     typedef uint16 ColorTable[256][3]        /* Array of Red, Green, and Blue values for each base color in the image*/
    #     typedef struct {
    #         uint16      Width;            /* Image width in pixels (must be less than 65535). */
    #         uint16      Height;           /* Image height in pixels (must be less than 65535). */
    #         byte        Flags;            /* Global color table flag - 1 bit */
    #         byte        Reserved;
    #         byte        BackgroundIndex;          /* Index of transparent background color in global color table - 8 bits */
    #         byte        PixelAspectRatio;         /* Pixel aspect ratio - 8 bits */
    #         ColorTable  GlobalColorTable;
    #     } GifScreenDescriptor;
    # ```

    # We only care about the first three fields so we don't need to read it
    signature_and_version = rng.read(6)
    
    # Read screen descriptor
    # Flags & Size are not relevant here since we're going to output an error in any case
    raw_screendesc = rng.read(7)
    # Pixel aspect is unused by us because we set the flags field appropriately
    _, width, height, flags, reserved, bgcolorindex, pixelaspectratio = struct.unpack('>HBHHBBB', raw_screendesc)

    if flags > 0x7:
        raise Exception("Flags must be less than 0x8.")
    elif width == 0 or height == 0:
        raise Exception("Width and height cannot both be zero")
    elif bgcolorindex!= 0:
        raise Exception("Background index must be zero when included.")

    # Write GIF header
    print("> Writing GIF header...")
    out.write(header + signature_and_version)
    out.flush()

    # Calculate size
    SIZEOF_COLOR_TABLE = 256*3 * 2
    gifsize = len(signature_and_version) + \
              len(raw_screendesc) + \
              3 * 4 +       # Block length bytes
              SIZEOF_COLOR_TABLE * 2**flags

    # Write new screen desc
    print("> Writing custom screen descriptor...")
    out.write(struct.pack('<H', gifsize))
    out.write(struct.pack('B', width % 0x100))
    out.write(struct.pack('B', height % 0x100))
    out.write(struct.pack('B', flags | 0x80))
    out.write(struct.pack('B', reserved))
    out.write(struct.pack('B', bgcolorindex))
    out.write(struct.pack('B', pixelaspectratio))
    out.flush()

    # Generate color table zeroes
    print("> Generating color table zeros...")
    for _ in range(SIZEOF_COLOR_TABLE // 2):
        out.write(b'\x00\x00')

