#!/usr/bin/env python3

import struct
from typing import BinaryIO

def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
        4 * rng.read(1), byteorder='little', signed=False
    ) & 0xffffff # limit to 2^24-1

    width = int.from_bytes(
        2 * rng.read(1), byteorder='little', signed=False
    )
    height = int.from_bytes(
        2 * rng.read(1), byteorder='little', signed=False
    )
    global_color_table_flag = (
        int.from_bytes(
            rng.read(1), byteorder='little', signed=False
        ) % 2
    ) == 1

    color_resolution = 7 + ((int.from_bytes(
        rng.read(1), byteorder='little', signed=False
    ) % 4) << 5)

    sort_flag = False

    size = bytes([height]) + bytes([width])
    screen_descriptor = size + bytes([color_resolution]) + \
                                    bytes([(sort_flag<<7)|
                                        (global_color_table_flag<<6)]) + \
                                    b'\x00\x00'


    palette = []
    for i in range(pow(2, color_resolution)):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        a = rng.read(1)[0]

        if len([r,g,b,a])!= 4:
            raise AssertionError('Invalid Color Byte')

        palette += [struct.pack('<BBBB', r, g, b, a)]

    background = struct.pack('<H', rng.read(1)[0]+rng.read(1)[0]*256)
    aspect = struct.pack('<B', rng.read(1)[0]&0xe0)

    image_data = b''

    x = y = 0
    while True:
        blocktype = rng.read(1)[0] & 0x7f
        lzwmincodesize = rng.read(1)[0]
        if not lzwmincodesize or lzwmincodesize > 12:
            continue

        max_bits = min(lzwmincodesize+1,8)
        bitcount = max_bits+1

        imgindex = 0


        # get the code length table
        codelenlen = 3
        clt = []
        for _ in range(lzwmincodesize):
            clt += [(cl:=rng.read(1)[0])]
            if cl==128:
                codelenlen = 4

        # build code length table via "dynamic huffman" algorithm
        nextcode = pow(2,max_bits)+2
        ent = 0
        while nextcode < 65536 and ent < lzwmincodesize-1:

            buf = None
            try:
                buf = rng.read(1)
            except Exception as e:
                pass
            
            if buf is not None:
                val = buf[0]

                if val >= 128 :
                    ent+=1

                    if ent >= lzwmincodesize-1:
                        break
                    
                    clt += [(val&0x7F)]
                else:
                    clt += [(val,rng.read(1)[0],
                            rng.read(1)[0],
                            rng.read(1)[0])]
                    ent+=3
                
                bits = val >> 7
                while bits == 0xFF:
                    bits = rng.read(1)[0]
                    codelenlen += 1

                if bits:
                    max_bits = max(bitcount,bits)
                    bitcount = max_bits + 1
                    
                if ent == 1<<max_bits:
                    max_bits += 1

            try:
                buf = rng.read(1)
            except Exception as e:
                pass
            
            if buf is not None:
                val = buf[0]

                if val >= 128 :
                    ent+=1

                    if ent >= lzwmincodesize-1:
                        break
                    
                    clt += [(val&0x7F)]
                else:
                    clt += [(val,rng.read(1)[0],
                            rng.read(1)[0],
                            rng.read(1)[0])]
                    ent+=3
                
                bits = val >> 7
                while bits == 0xFF:
                    bits = rng.read(1)[0]
                    codelenlen += 1
                if bits:
                    max_bits = max(bitcount,bits)
                    bitcount = max_bits + 1
                    
                if ent == 1<<max_bits:
                    max_bits += 1
                    
        num_entries = ent+1
        
        # build decoding table from code length table
        htab = {}
        for k,_ in enumerate(clt[:num_entries]):
            if type(_) == tuple:
                for j in range((i:=_[0])):
                    htab[2**(j+1)-1] = _[1:]
            elif _.isalpha():
                cl = ord(_.lower())-ord('a')+1
                htab[2**_.count('1')] = chr(cl//(2**(cl%2)))
        lenthis = num_entries

        # append pixels until we have an entire row of them
        while y <= height:
            # this whole thing only works for up to 31 rows of pixels total
            scanline = b''
            subblock = b'\x00'

            while True:
                bitbuf = 0
                codelen = -1

                while codelen < 0:
                    bitbuf += (int.from_bytes(rng.read(1),'little',signed=False)) << (-bitcount)
                    bitcount -= 1
                    codelen = clt[lenthis][bitbuf>>bitcount]

                if codelen == 128:
                    break

                if codelen < lzwmincodesize:
                    bitbuf >>= (max_bits-bitcount+codelen)
                    val = htab[(bitbuf>>(c:=codelen))].encode()
                    if c>1:
                        bitbuf |= (int.from_bytes(rng.read(c-1), 'big', signed=False)<<bitcount)
                        bitcount = -(len(htab[(bitbuf>>(c:=codelen))])+c)
                        val += htab[(bitbuf>>(c:=codelen))][:c].encode()
                        bitcount-=c
                    if type(val)==str:
                        val = val.encode()
                    scanline += val
                else:
                    val = scanline[-st:-st+htab[lenthis]]
                    scanline += val


                st = len(val)
                lenthis += 1
                codelen = clt[lenthis]
                

    
            if len(scanline) > datalen / height // 3:
                print('Warning: Data too large!')

            scanline += b'\x00'*(datalen//height//3-len(scanline))

            if y>=height:
                break
            

            if global_color_table_flag:
                gce = int.from_bytes(rng.read(1),'little', signed=True)
                interlace = bool(gce & 0x40)
                local_screen_descriptor = bool(gce & 0x80)
                sorted_color_table = bool(gce & 0x10)
                color_table_size = (gce & 7) + 1

                has_local_color_table = data[p]=="\x01"
                if has_local_color_table:
                    p+=1
                    local_color_table = data[p:p+3*color_table_size]
                    p+=3*color_table_size
            else:
                local_screen_descriptor = has_local_color_table = False


            if local_screen_descriptor:
                scrd = data[p:p+7]
                p+=7

            if has_local_color_table:
                lct=[]
                for i in range(0, len(local_color_table)//3):
                    lct += [struct.unpack("<BBB",local_color_table[i*3:(i+1)*3])]
                lct += [None] * (256-len(lct))

            nxtsublen = rng.read(1)[0]+1
            subblock += bytes([nxtsublen]) + scanline

            if nxtsublen==255:
                subblock += b'\xfe'
                break

            if y<height-1:
                subblock += b'\x00'+bytes([rng.read(1)[0]])
                y+=1
                
            imgindex+=1
            
        if (imgindex+1)%(data["imgtable"][0])==0:
            image_data += subblock
            subblock = b''
            y=0

        if y==height and not imgindex%(data["imgtable"][0]-1):
            image_data += subblock
            subblock = b''
            y=0



    gif = header + datalen.to_bytes(3,'big') + screen_descriptor + \
                              b'\xff\x21\xf9' + background + aspect + \
                              b'\xff\x00'+palette[-1] + \
                              (len(palette)*(1)).to_bytes(1,'big') + b'\x00'.join(palette) + \
                              image_data
    
    out.write(gif)