#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting for this test case
    background    : int   = ord(rng.read(1)[0])
    pixel_aspectn : float = struct.unpack('B', rng.read(1))[0] / 256

    # Let's pretend there is only one color palette
    global_color_table_size = 7
    local_color_table      = b'\x00\x00\x00\x00\x00\x00'


    # Image descriptor block
    image_descriptor                                      =  b''
    image_descriptor     += struct.pack('<BBHHBBBBB',
        0b10000000 | (image_width >> 8),          # left position
        image_width & 0xFF,                      # right position

        ((image_height + 1) >> 8) - 1,           # top position
        (((image_height + 1) & 0xFF) << 8) | 0b11111,

        False,                                    # local color table flag
        0,                                        # interlace flag
        0,                                        # sort flag
        0,                                        # reserved bit for size of gce
        0                                         # size = 1
    )
    image_descriptor                                  += local_color_table
    image_descriptor                                  += bytes([background])
    image_descriptor                                  += struct.pack('<B', round(pixel_aspectn * 256))

    
    
    data               = b''
    frame_count        = 0
    block_type         = None
    block              = b''
    control_byte       = 0

    while True:



        # End of file
        if len(block) == 0 and rng.read(1)!= b'.':
            raise Exception("All frames have been generated but input ended prematurely")


        # Read more blocks until we find an extension block
        if len(block) == 0:
            block_type = ord(rng.read(1))

            control_byte = ord(rng.read(1))

            block_length = ord(rng.read(1))
            block_length |= ord(rng.read(1)) << 8


            block = rng.read(block_length)


        elif control_byte == 0x21: # Extension block
            
            # Block terminator
            if block[0] == 0:
                block                 = b''

                control_byte          = None
                block_length          = None


                continue


            # Graphics Control Extension
            elif block[0] == 0xF9:
                
                block                     = b''
                control_byte              = None
                block_length              = None


                background    = ord(rng.read(1))
                frame_delay   = struct.unpack("<H", rng.read(2))[0]

                frame_count += 1

                continue



            else:
                raise Exception("Unknown extenstion code encountered in frame " + str(frame_count))



        elif control_byte == 0x2C or control_byte == 0x21: # Image Descriptor

            # Insert the image descriptor block
            data             += image_descriptor
            data_size         = len(data)

            # Reset the image descriptor back to default values
            image_descriptor  = bytearray()

            image_descriptor                          += struct.pack('<BBHHBBBBB',
                0b10000000 | (image_width >> 8),      # left position
                image_width & 0xFF,                  # right position

                ((image_height + 1) >> 8) - 1,       # top position
                (((image_height + 1) & 0xFF) << 8) | 0b11111,

                False,                                # local color table flag
                0,                                    # interlace flag
                0,                                    # sort flag
                0,                                    # reserved bit for size of gce
                0                                     # size = 1
            )
            image_descriptor                              += local_color_table
            image_descriptor                              += bytes([background])
            image_descriptor                              += struct.pack('<B', round(pixel_aspectn * 256))

            
            

            ### Compressor Code goes here ###
            data_comp          = ''
            compress_stack     = []

            local_color_table_size = 7
            local_color_table = [
                '000000', 
                'ffffff'
            ]

            while len(data):
                chunk = data[:128]
                del data[:128]

                match_found = False

                for i in range(len(compress_stack)):
                    k = min(len(chunk), len(compress_stack[-i-1]))

                    if chunk[:k] == compress_stack[-i-1][:k]:
                        color_code = str(int(local_color_table_size - i)).zfill(3)

                        while k > 1:
                            surplus = k%64

                            chunk = chunk[(surplus+1):]
                            
                            if surplus == 64:
                                break;

                            k -= surplus


                        length_code = str(surplus).zfill(2)
                        
                        char = f"{color_code}{length_code}"

                        data_comp += char
                        

                        del compress_stack[-i-1][:k]

                        match_found = True
                        break
                    

                if not match_found:
                   lzw_code = '1000'
                   char = ''.join([(str(ord(c))).zfill(3) for c in chunk])
                   data_comp += lzw_code + char 

                   compress_stack.append(chunk)



            print(f"Frame {frame_count} compressed from {len(data)} to {len(data_comp)} (Ratio: {(len(data)/len(data_comp))*100}%)")
            frame_count += 1



            # Write the length indicator for the data sub-block
            data            += struct.pack('<H', len(data_comp) + 2)
            data_size        = len(data)

            # Write the data sub-block
            data            += bytes(['!']) 
            data            += data_comp

            # Write the ending byte for the graphic block
            data            += bytes(['?']) 




    
        else:
           raise Exception(f"Unexpected control byte '{control_byte}' encoutered in frame {frame_count}")
            



    

    gif = header
    gif += struct.pack('<HHHH',
        image_width, 
        image_height, 
        0, 
        0
    )

    gif += data


    out.write(gif)