#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):

            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]

    else:
        palents = b''

    # leave these as zero
    bgindex, pxratio = struct.unpack('BB', rng.read(2))

    # flags
    globalct = bool(int.from_bytes(rng.read(1), 'little'))
    cres = int.from_bytes(rng.read(1), 'little')
    sflag = int.from_bytes(rng.read(1), 'little')

    # no clean way to read/write longs in Python, so just use ints
    lentable = int.from_bytes(rng.read(2), 'big')

    # TODO: handle non-global color tables later on

    # back to frame data now!
    frames = []
    while True:

        frame = dict()

        from_left = int.from_bytes(rng.read(2), 'little')
        from_top = int.from_bytes(rng.read(2), 'little')
        frame['dims'] = (image_width, image_height)
        frame['from'] = (from_left, from_top)

        delay = int.from_bytes(rng.read(2), 'little')
        frame['delay'] = delay

        transcolor = int.from_bytes(rng.read(1), 'little')
        if transcolor == 0xFF:
            transcolor = None
        elif transcolor!= 0x00:
            raise ValueError("Invalid transparent colour")
        frame['transcol'] = transcolor

        disposalmethod = int.from_bytes(rng.read(1), 'little')
        if disposalmethod!= 0 and disposalmethod!= 1:
             raise ValueError("Disposal method must be either 0 or 1")
        frame['dispmthd'] = disposalmethod

        userinput = int.from_bytes(rng.read(1), 'little')
        if userinput!= 0:
            print("Warning: User input flag set!")
        frame['userinpt'] = userinput

        buf = bytes()
        while len(buf)!= 10:
            r = rng.read(1)
            if r == b'\xfe':

                label = rng.read(1)[0];
                print(f"Local image data of type {label} encountered.")

                if label == 0x2C:
                    continue
                raise RuntimeError("Encountered an invalid subblock ID")

            buf += r

        lzwmincodesize = rng.read(1)

        # TODO: implement LZW decoding here later

        # the second two bytes are LZW min code size and the third is
        # one byte reserved
        if lzwmincodesize!= 0x08 or rng.read(1)!= b'\x00':
            raise RuntimeError("LZW not implemented")

        fdatapos = out.tell()
        if lentable == 0:
            # if there's no LZW dictionary at all then just attach it directly
            out.write(palents + data)
        else:
            # otherwise, put them together inside the file!
            out.write(b'\xff\xe0' + bytes([lentable]) + palents)


        # now we need to go back and write some information about where
        # out frame data starts and ends relative to the beginning of
        # the output file
        out.seek(-7, 1)
        out.write(struct.pack('<LL', lentable, fdatapos))
        # return to end of file
        out.seek(0, 2)

        frames.append(frame)

        # check if we're done yet :o
        if rng.read(1) == b'\x3b': break

    for y in reversed(range(image_height)):
        for x in range(image_width):
            
            idx = int.from_bytes(out.read(1), 'little');

            if idx >= len(frames):
                raise IndexError("Frame index out of bounds")

            if frames[idx]['dims'][0] <= x+1:
                raise IndexError("Pixel position out of bounds")

            if frames[idx]['dims'][1] <= y+1:
                raise IndexError("Pixel position out of bounds")

            if frames[idx]['transcol']:
                # if transparent color was specified, then instead of
                # writing the index we need to write the special value
                # 0x40
                out.write(b'\x40')
            else:
                out.seek(-1, 1)
                out.write(bytes([idx]))

    # finally, let's go ahead and add the footer
    out.write(b';')

    # write the loop control block
    # TODO: we don't support looping by default right now; so just
    # throw a warning
    out.write(b'[')
    print("Warning: No animation loop was generated")

    # indicating that any further data after this will be ignored
    out.write(b'\xfe')

    # TODO: we could do something with this; but not now...
    rng.seek((3 << 8) | ctsize)

    datalen = int.from_bytes(rng.read(2), 'big')
    data = rng.read(datalen)
    out.write(header + data)
