#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):

            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]

    else:
        palents = b''

    # colormap is followed by a local color table
    # TODO: we could do something with this; but not now...
    rng.seek((3 << 8) | ctsize)

    datalen = int.from_bytes(rng.read(2), 'big') & ((1 << 12) - 1)
    data = rng.read(datalen + 1)
    
    # Raster control extension block to set delay time
    rcblock = struct.pack("<BBHBBBB",
        0x21, 0xF9, 4,     # Application Extension Block
        0b00000001,        # No transparancy
        0,                 # Delay time in hundredths of seconds
        0,                 0
        )

    lzwmincodesize = min(max(int.from_bytes(rng.read(1), "little"), 2), 12)

    subblocks = []
    encsubblocks = []
    nsavedcodes = 0
    while True:
        
        thislen = 2 ** lzwmincodesize

        # Make sure it's clean
        thisref = None
        while thisref!= b'':
            thisref = data[:thislen].tobytes().strip(b'\xff\x00')

            thislen = thislen + 1

        chunk = data[:thislen]
        data = data[thislen:]

        # I don't really understand how LZW works, so I'm just going
        # to use a very opaque solution 
        encchunk = bytes([nsavedcodes]) + chunk[:-1]

        subblocks.append(encchunk)
        encsubblocks.append(chunk)

        nsavedcodes += 1
        if data == b'': break

    extheader = struct.pack("<BBBBBB",
        0x21, 0xFF,         # Application Extension Block
        11,                # Size of block including itself
        0b10000010,        # Label as Graphic Control Extension
        ord('F'),          ord('I'),      ord('X'),       ord(' '),
        0                  # No sub-block data follows
        )

    # the output gif will have an explicit frame duration 
    # instead of storing all frames on loop
    expframeduration = 0b10000000 | 10

    framecount = len(subblocks)

    out.write(header)
    out.write(struct.pack("<HHBBBBB",
        image_width, image_height, 
        0xF0,              # Global Color Table Flag
        ctsize,            # Color Table Size (number of colors)
        0,                 # Background Color Index
        0,                 # Aspect Ratio
        ))
    out.write(palents)

    formbody = bytes()

    # Loop count of zero means infinite looping
    # (1 means one iteration only).
    # The image data should already be compressed
    # so no need to use LZW compression here.
    formbody += struct.pack("<BBHBBBBBHB",
        0x21,             # Image Descriptor
        0,                # Left position
        image_width,      # Width
        image_height,     # Height
        0,                # Not Local Color TableFlag
        ctsize,           # Color Table Size
        0,                # LZW minimum code size
        expframeduration, # Frame Duration
        framecount,       # Number of Frames
        )

    formbody += extheader

    for subblockidx in range(framecount):
        formbody += struct.pack("B", lzwmincodesize)
        formbody += encsubblocks[subblockidx]

    # End of image marker
    formbody += b'\x3B'

    out.write(formbody)
