#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Number of frames to generate
    num_frames = 50

    # First comes the header
    # Fixed values in bytes [0x47, 0x49, 0x46]
    out.write(b'GIF')
    # Version (byte[3]) - Write "89a"
    out.write(b'\x89\xa')
    # Width & height of image
    width, height = rng.read(2), rng.read(2)
    out.write(width + height)
    # Flag for global color table, set to True
    # Flags & background color index are reserved and should be 0
    flag = b'\x10'
    out.write(flag)
    # Frames per second is fixed at 10
    fps = b'\x0a'
    out.write(fps)
    # Trailer is empty!
    trailer = b''
    out.write(trailer)

    # Now we get into our main loop
    for i in range(num_frames):

        # Get the current frame size, this can vary from frame to frame.
        # The minimum value for each byte will always be 0x00 and the max 0xff
        frame_size = rng.read(2)
        
        # Next up come the LZW minimum code size : 0-12
        lzw_min_code_size = rng.read(1)

        # Then follows the compressed data
        compressed_data = rng.read((i+1)*1000)

        # Now comes one last field where the color table can be specified, but we ignore that
        # since we're not implementing any form of compression here

        # Finally after all those fields there's an optional string indicating the end of the
        # frame with ";"
        terminator = b';'

        # And then the checksum, which is always calculated by XORing all previous bytes together
        checksum = b'\x00'
    
        # Concatenate everything together
        full_frame = \
            frame_size + \
            lzw_min_code_size + \
            compressed_data + \
            terminator + \
            checksum
        

        # Calculate the final length of the frame by adding 1 for each extra null byte
        num_extra_nulls = len([c for c in full_frame if c == 0])

        # Finally write it out as a single block
        print(full_frame[:-1].hex())


if