#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):

            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]

    else:
        palents = b''

    # colormap is followed by a local color table
    # TODO: we could do something with this; but not now...
    rng.seek((3 << 8) | ctsize)

    gif = []

    # frames should have an empty data block
    framedata = b'\x00\x2C\x00\x00\x00!\xF9\x04\x01\n\x00,\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;'

    # repeat forever!
    while True:

        # flush through end of current frame
        gif += [header + 
            struct.pack('<HH', image_width, image_height) + 
            palents +
            # Logical Screen Descriptor plus Global Color Table
            b'\xFF\x7C\x00\x00\x03\x00\x01\x00\x80\xFF\x7C\x00\x00\x00\x07\x00\x00\x00\x00' +
            b'\x01\x00\x01\x00\x00\x02\x02D\x01\x00;\x00\x00' +
            # Graphics Control Extension
            b'\x21\xF9\x04\x01\n\x00\x00\xFF\x00,\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;' +
            framedata +
            # Image Separator
            b'\x2C' +
            # Local Image Descriptor
            struct.pack('<HHBBHHH',
                0, 0, 
                image_width & 0xff, 
                image_height & 0xff, 
                0, 
                0, 
                0, 0),
            b';']
    
        # get next image dimensions
        image_width, image_height = \
            struct.unpack('<HH', rng.read(4))

        # size of the color table
        ctsize = int.from_bytes(rng.read(1), 'little')

        if ctsize > 0:
        
            # read in palette values (RGB triples)
            paldata = [struct.pack('BBB', *x) for x in
                [(y % 256, y // 256, y // 256 * 256)
                    for y in range(ctsize)]]
    
            palents = bytes()
            for i in range(len(paldata)):
    
                # write out the length byte first
                palents += struct.pack('B', len(paldata[i]))
                palents += paldata[i]
        
        else:
    
            # no palette data to include
            palents = b''
        
        # make sure there's room in the output buffer before appending 
        if len(gif[-1]) + len(palents) + 3 >= 65536:
            
            # break up very long frames so we don't overflow the buffersize
            splitpoint = max([framedata.index(b'\x2C'), framedata.index(b'\x00')]) + 1

            gif[-1] += framedata[:splitpoint]
            gif[-1] += b'\x2C'

            framedata = framedata[splitpoint:]

        # append new image descriptor and data
        gif[-1] += b'\x2C' + \
            struct.pack('<HHBBHHH',
                0, 0, 
                image_width & 0xff, 
                image_height & 0xff, 
                0, 
                0, 
                0, 0)

        # insert local color table, if any
        gif[-1] += palents

        # now find a suitable palette entry at random
        palidx = rng.read(1)[0]
        count = 0
        while palidx == 0 or palidx == 255:
            palidx = rng.read(1)[0]
            count += 1
        assert count < 256

        # use the selected palette entry in the chosen frame
        gif[-1] = gif[-1].replace(b'\x00;', 
                    b'\x00\x00\x00' + struct.pack('B', palidx).decode()) + b'\x00;'

        # make sure there's room in the output buffer before appending 
        if len(gif[-1]) + len(framedata) + 3 >= 65536:
                
            # break up very long frames so we don't overflow the buffersize
            gif[-1] += framedata[:max([framedata.index(b'\x2C'), framedata.index(b'\x00')]) + 1]
            gif[-1] += b'\x2C'

            framedata = framedata[max([framedata.index(b'\x2C'), framedata.index(b'\x00')]) + 1:]
            
        # and finally finish the frame off with the actual data block
        gif[-1] += framedata

        # and then go back around again, until we run out of input data
        
