#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# writes it in an optimized way.
def gen_gif(out: BinaryIO, rng: BinaryIO):
    # The header magic numbers
    out.write(b'GIF89a')

    # Width and height of the image
    w = 1280
    h = 720
    out.write(struct.pack('<HH', w, h))

    # Global color table length (256 colors maximum)
    ctlen = b'\xff' * 3 * 256
    out.write(ctlen)
    
    # Image descriptor block size
    out.write(b'\x2c')

    # Left / top position
    out.write(b'\x00\x00')

    # Image width and height
    out.write(struct.pack('BB', *(w % 256), int((w - 1) / 256)))
    out.write(struct.pack('BB', *(h % 256), int((h - 1) / 256)))

    # Packed field: local color table flag set to 0
    out.write(b'\x00')

    # Interlacing flag in lowest bit (set to false).
    out.write(b'\x00')

    # LZW minimum code size
    out.write(b'\xf0')

    # Animation delay time in centiseconds
    delay = 25
    out.write(struct.pack('B', delay));

    # Loop counter for infinite loops (set to 100)
    loopcount = 100
    out.write(struct.pack('B', loopcount))

    # Direct color mode data is after this point...

    def nextbit() -> bool:
        return ord(rng.read(1)) & 1!= 0

    # Minimum code size before using LZW compression -- we'll use LZW with
    # no compression so that's not necessary!

    # Generate pixels by reading the RNG until there are none left.
    while True:
        # If we're on a new line but also at the end of a column, jump back up
        # to the first column right below us (before advancing downwards one).
        if (ord(rng.read(1)) & 1) == 0:
            if i >= w - 1:
                i = 0
                j += 1
            else:
                i += 1

            continue
        
        # Read two bytes from the RNG to get our next pixel value.
        r, g, b = [nextbit(), nextbit(), nextbit()]

        # Write the appropriate byte to our output stream depending on whether
        # or not this pixel uses the global palette.
        if r and g and b:
            continue

        # We've gotten here either because the last three bytes were all zeroes
        # or these four are not all ones; either way they can be written as-is.
        out.write(r.to_bytes(1, 'big') +
                  g.to_bytes(1, 'big') +
                  b.to_bytes(1, 'big'))


