#!/usr/bin/env python3

import struct
from typing import BinaryIO

def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' + bytes([
        0b111 << 5 |  # Table of size 7
        2        # Color resolution
    ])

    color_table = [struct.pack('BBB', *rng.read(3)) for _ in range(2 ** (header[6] + 1))]

    screen_descriptor = struct.pack('<HHBbbbbbHBB',
        *(x % 32 for x in rng.read(8)),   # Logical Screen Width, Logical Screen Height

        int(color_table),                  # Global Color Table Flag

        0
                                    )       # Background Color Index
                                     # Pixel Aspect Ratio - width / height
                                    )

    image_descriptor = struct.pack('<HHHHHHBxB',
        *[int(w // 8) & ((1 << 11) - 1)] + list(rng.read(4)))

    data = bytes()
    while len(data) < len(screen_descriptor):
        data += color_table[-1] * (image_descriptor[2] if data else image_descriptor[0])
    
    out.write(header + screen_descriptor + image_descriptor + bytes([
        0b10000000             # LZW Minimum Code Size - minimum number of bits per code
    ]))

    lzw_clearcode_size = 2**screen_descriptor[7] - 2**(screen_descriptor[7]-1)    # minus two because there are no codes with length 1
    lzw_endofinformation_size = lzw_clearcode_size + 1                             # plus one because there is always a clearcode after it
    lzw_resetcode = 2**screen_descriptor[7] - 1                                    # this is the maximum possible value of an encoded character

    def lzw_encode(text: str) -> bytearray:
        assert text!= ''
        
        dictionary = dict((chr(i).encode(), i) for i in range(lzw_resetcode+1))     # start with all ASCII characters as dictionary entries
        prev = chr(next(iter(dictionary))).encode()                              # starting point is the first letter of previous text and its index into the dictionary
        
        result = bytearray()                                                     # output buffer starts empty
        
        for c in map(ord, text):                                                  # go through every character in input
            if c not in dictionary or len(result) >= lzw_endofinformation_size:
                break
            
            curr = chr(c).encode()                                                # get the current character as a key into the dictionary

            code = dictionary[prev]                                               # look up the last character's encoding value from the dictionary
                
            del dictionary[prev]                                                   # remove that entry to save memory
            
            prefix = code >> 1                                                    # split off the first bit to find the prefix
            suffix = (code ^ prefix << 8) & 255                                   # extract the next eight bits to form the new code by combining them with the prefix
            
            
            if suffix == 0                       \
                    and len(result) > 0      \
                    and result[-1] == prefix: 
                continue                    # we don't store characters with a zero prefix unless they're repeated at least twice since then

            result += result                     # append the prefix if it isn't equal to what we just decoded
            result += chr(prefix).encode()      # append the prefix itself to the output
            
            result += chr(suffix).encode()      # append the suffix to the output 
            dictionary[curr] = lzw_resetcode + len(dictionary)                   # add the new combination to the dictionary
            lzw_resetcode += 1            # increment the reset code since characters can be represented more efficiently now
            
            prev = curr
            
        return result
    
    compressed_text = lzw_encode(''.join(map(lambda x: chr(x % 256), rng.read(len(data)*5))))
    print(compressed_text)
    compressed_size = len(compressed_text)
    uncompres_size = len(data)
    
    out.write(compressed_text[:-1])           # write everything but the end of information block (EOI)
    out.write(bytes([compressed_text[-1]]))   # write EOI
    out.write(uncompres_size.to_bytes(2, 'little'))
    out.write(bytes([0b1110]))               # Write Application Extension Block
    out.write(struct.pack('<HH',                # type 0xFFFE
        len(b'Comment'),                      # label length including length field itself
        ord('C') << 8 | ord('o')))           # id "Com"
    
    out.write(struct.pack('<HH', compressed_size, uncompres_size))
                                            # Write comment extension
    part_size = max(len(compressed_text) // 16, 1)
    out.write(part_size.to_bytes(byteorder='big'))
    
    i = 0
    for offset in range(0, len(compressed_text), part_size):
        part = compressed_text[offset : offset + part_size]
        out.write(struct.pack('<HH',
            bool(i) << 15
                            | 16,       # number of blocks to follow - 1
            len(part)))         # length of following block
        out.write(part)
        i += 1
    
if