#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):
##            print(i*3, ":", paldata[i])
            palents += paldata[i]

        palsize = len(palents).to_bytes(
            1, byteorder='big', signed=False)

        out.write(header + image_width.to_bytes(2,'little') +
            image_height.to_bytes(2,'little')) 

        out.write((b'\xf7\x00' + ctsize.to_bytes(1, 'little')))
        
        out.write(b'\x00')

        out.write((b'\x2c' + b'\xff'*2 +
            image_width.to_bytes(2,'little')+image_height.to_bytes(
                2,'little')+b'\x00'+b'\x00').ljust(19, b'\x00'))

        out.write(b'\x00\x3b')

        out.write(palsize + palents)

    else:

        out.write(header + image_width.to_bytes(2,'little') +
            image_height.to_bytes(2,'little'))
    
        out.write(b'\xf7\x00\x00')
    
    # no "graphic control extension" block here...

    # set transparent color to the first one defined above
    out.write(b'\x21\xf9\x04\x00\x00\x00\x00\x00\x00\x00\x00') 

    # no comment extension blocks
    
    # this is all we need for basic GIFs, just frames with 
    # no delay time and simple graphics data
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00\x00')


    framecount = rng.read(1)[0] % 127 # can't be larger than 127
    for _ in range(framecount):
        duration = rng.read(2)+b'\x00'

        extsize = rng.read(1)[0]%511 # can't be larger than 510
        # "plain text" subblock 
        # (see section 22 of https://www.w3.org/Graphics/GIF/spec-gif89a.txt)
        out.write(b'\x21\x00' + extsize.to_bytes(1,'little') + 
            b'Comment:' + duration + b"\x00")


    # put some actual pages in here
    # ones with only a graphic control block but no 
    # bitmap data
    for _ in range(framecount):
        duration = rng.read(2)+b'\x00' # another 2 bytes to make it easy
                                       # to add the transparency color

        # technically doesn't have to be true, but makes sense :)
        #... set to zero because we don't use any loops or disposal method
        transpflag = b'\x00'

        out.write((b'\x21\xf9' + b'\x04'+duration+transpflag +
                   b'\x00\x00\x00\x00\x00\x00\x00\x00')).ljust(
                       int.from_bytes(
                           struc.pack('BBBBBB'), byteorder='little'),
                       0)



    # provide the five bytes at the tail end of the graphic
    # control block as an indication of how many 
    # additional data blocks will follow before
    # reading exactly this amount from the input
    out.write(struc.pack('BBBBBB', 
                         *(len(bytearray(duration)).to_bytes(
                             1))), byteorder='little'))

