#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# RNG in `rng`. Also writes the palette.
def genrandomgif(rng: BinaryIO, out: BinaryIO):
    header = rng.read(13)

    width = int.from_bytes(rng.read(2), 'big')
    height = int.from_bytes(rng.read(2), 'big')
    flags = int.from_bytes(rng.read(1), 'big')
    globalctflag = (flags >> 7) & 0x1
    gfcrcflag = (flags >> 6) & 0x1
    ctsize = flags & 0xf

    # we ignore these two fields for simplicity
    bpp = int.from_bytes(rng.read(1), 'big')
    bgindex = int.from_bytes(rng.read(1), 'big')

    if globalctflag:

        # TODO: actually support these, but they are pretty rare so who cares?
        palsize = int.from_bytes(rng.read(1), 'big')
        sortedflag = bool(int.from_bytes(rng.read(1), 'big'))

    # there's some optional chunks at the end of the header, we don't handle them right now
    while True:
        chuncktype = rng.read(4)
        chunklen = int.from_bytes(rng.read(4), 'big')

        if chuncktype == b'NETSCAPE':

            subblock = rng.read(chunklen - 4)
            esccode = subblock[:2]
            blocksize = int.from_bytes(subblock[2:], 'little')

            if esccode == b'\0\xff':
                throwaway = rng.read(blocksize - 2)
                continue
        
        elif chuncktype == b'\xfe\x00\xfc':
            
            subblock = rng.read(chunklen - 4)
            appextver = int.from_bytes(subblock[:2], 'big')
            apptype = subblock[2:]

            assert appextver <= 11
            assert apptype!= b'NETSCAPE'
            
            # read and discard any application specific blocks
            continue

        break
    
    # hm, I think we can keep all the other things as-is here!
    out.write(b'GIF89a')
    out.write(struct.pack('<HHBBB', width, height, bpp, 0, 0))

    # image descriptor
    out.write(b'\x2c') 
    # frame delay
    out.write(b'\x00\x00') 

    # palette
    # bit depth determines how many bits to use per pixel
    if max(bpp // 8, 1) > 1:
        paldata = [None for _ in range(2 ** ((bpp // 8) * 2))]
    else:
        paldata = []
        
    for i in range(max(256, 2 ** (bpp // 8))):

        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        a = 0xff

        paldata.append(struct.pack('BBBB', b, g, r, a))


    if paldata:

        ctablesize = 2 ** ((bpp // 8) * 2)
        tablewidth = 2 ** (bpp // 8)

        # ensure that 2**n divides both
        leng = len(paldata) 
        if tablewidth < leng:
            tablewidth *= 2
            tablewidth |= 0x1

        # there must be an even number of colors
        if tablewidth % 2:
            tablewidth -= 1

        # then pad it out to the nearest power of two
        if tablewidth > leng:
            newpal = list()
            for i in range(leng, tablewidth):
                b = rng.read(1)[0]
                g = rng.read(1)[0]
                r = rng.read(1)[0]

                paldata.append(struct.pack('BBBB', r, g, b, 0xff))

        # note: tablewidth is one byte less than what the spec specifies
        # but it seems to make more sense? maybe?
        out.write(struct.pack('BBB', 0x87, tablewidth, ctablesize))
        palents = bytes()

        for i in range(tablewidth):
            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]

    else:
        palents = b''

    # colormap is followed by a local color table
    # TODO: we could do something with this; but not now...
    rng.seek((3 << 8) | ctsize)

    datalen = int.from_bytes(rng.read(2), 'big')
    data = rng.read(datalen)
    out.write(header + data)
