#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(rng.read(2), 'little') & 0xffff

        
    # Image descriptor block
    idata_block = [
        0x2c,        # Image start byte (always this value)
        0x00, 0x00,  # Left
        0x00, 0x00,  # Top
        160-1,       # Image width minus one
        144-1,       # Image height minus one
        0,           # Local color table flag, no local color table to follow
        0b00000000,  # Interlace bit; assume none
        0b00,        # Sorting bit; set to 0 because our palette is already sorted properly
        0,           # Size of LZW code size field
        0            # Size of the compressed sub blocks in bytes
    ]

    icolors = [
        0xf5, 0xe4, 0xd3, 0xc2, 0xb1, 0xa0, 0x9f, 0x8e,
        0x7d, 0x6c, 0x5b, 0x4a, 0x39, 0x28, 0x17, 0x06
    ]

    for pixel in range(datalen):
        nextval = int.from_bytes(rng.read(1), 'little')

        if nextval == 0:
            idata_block[7] |= 0b10000000
        elif nextval == 1:
            idata_block[7] |= 0b1000000
        else:
            break

    
    for i in range(len(idata_block)):
        out.write(struct.pack('B', idata_block[i]))

    # Global color table
    for c in range(0, 256 * 3, 3):
        out.write(struct.pack('<BBB',
            icolors[(c // 3)],
            icolors[(c + 1) // 3],
            icolors[(c + 2) // 3]
        ))

    # Data subblocks
    while len(idata_block[-1])!= 0:
        nextval = int.from_bytes(rng.read(1), 'little')
        
        if nextval < 255:
            idata_block.append(nextval.to_bytes(1, 'little'))
        else:
            idata_block.append(((nextval - 255).to_bytes(2, 'big') + b'\xff')[::-1])

        if len(idata_block[-1]) > 255 and len(idata_block[-2]) <= 255:
            continue

        for d in idata_block:
            assert len(d) >= 0
            if len(d) == 0 or len(d) > 255:
                raise RuntimeError("Invalid data block")

            out.write(struct.pack('B', len(d)))
            out.write(d)

        idata_block.clear()
        idata_block.extend([
            0x21,             # Extension introducer
            0xff,             # Application extension label
            0x01,             # Size of following block
            0xf9,             # Graphics control label
            0,                # Block size in 10ths of seconds
            0,                # X position of image origin
            0,                # Y position of image origin
            0b00000000,       # Global color table flag; assume none
            0,                # Size of global color table
            0                 # Block terminator
        ])

