#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):

    header = b'GIF89a'

    # get some gif info from the RNG
    width_bytes = rng.read(1)[0] & 0b11111
    descriptor_byte = rng.read(1)[0]
    status_flag = descriptor_byte >> 7
    gct_flag = (descriptor_byte >> 5 & 1)!= 0
    color_res = (descriptor_byte & 0b111) / 32
    sorted_flag = ((descriptor_byte & 0b100000) >> 6)!= 0
    interlace_flag = ((descriptor_byte & 0b1000000) >> 7)!= 0
    colormap_size = rng.read(1)[0] + 1
    background_color = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    # we only support certain features right now...
    assert not interlace_flag
    if status_flag == 1 or pixel_aspect_ratio!= 0:
        raise AssertionError("unsupported feature in graphics control block")

    # get color table size and palette index data from the RNG
    ctsize = (colormap_size**2 - 1).bit_length() // 8 + 1
    palettedata = []
    for i in range(colormap_size):
        palettedata += [rng.read(3)] * 2**(ctsize*8//3)

    # construct global color table
    palents = b''
    for i in range(len(palettedata)): 
        r,g,b = palettedata[i][0], palettedata[i][1], palettedata[i][2]
        palents += bytes([r,g,b])
    palents *= (2**ctsize)//3

    # determine how many repeats of each frame is needed based on height; 
    # since we don't support interlacing we can just add enough frames to fill
    # that space
    num_repeats = round((image_height+height)%height) // height

    # write out all repeats to the output stream
    for _ in range(num_repeats):

        # read new image information from the RNG
        image_width = rng.read(width_bytes)
        image_height = 0x00
        
        while image_height < height:

            image_height |= rng.read(1)[0] << (width_bytes*8-8)

        # read LZW codesize data from the RNG
        clen = (rng.read(1)[0]+1)**2
        codesizes = list(reversed([(rng.read(1)[0]**2)-1 for i in range(clen)]))

        # get rest of the data from the RNG
        clens = [-1]+list(range(0xff))
        codelengths = [(clens[-codesizes[i]]+1)+clens[-codesizes[i+1]] for i in range(clen)][::-1]
        data = rng.read(sum(codelengths))


        def get_lzw_string():
            
            def unpackbits(stream, n=None):
                res = 0
                for bitno in range(n or len(stream)*8 - 1, -1, -1):
                    res |= (stream.read(1)[0] >> bitno & 1) << bitno
                return res
    
            repcount = {v:k for k, v in enumerate(clens)}

            def decode_string(stream):

                bins = reversed(unpackbits(stream, n=(max(clens)+1)).to_bytes((max(clens)+1)>>3, 'little'))
    
                currlength = clens[-1]
                decoded = ''

                while True:

                    chunk = next(bins)

                    if chunk >= currlength:
                        decoded += chr(chunk)
                        break
                    
                    chunk <<= max(clens) - currlength
                    chunk += next(bins)

                    decoded += chr(repcount[chunk])
                
                return decoded

            string = ''
            while True:
                try:
                    string += decode_string(stream=BinaryIO(data)); data = b''
                except StopIteration: 
                    data += rng.read(1); continue
                else: break

            return string

        # get the entire data as a single lzw encoded string
        lzws = get_lzw_string(); del data

        # split it up by scanline and convert each line to rgb byte values
        lines = [lzws[i::image_width].encode('ascii') for i in range(image_width)]; del lzws

        imgdat = b''
    
        # convert lzw encoded string back into raw pixels
        for line in lines:

            row = b''
            prev = None
    
            bits_left = min(prev, clens[(prev*prev)-1] if prev is not None else 0)
            idxshift = [int(((curr*curr)-1)/2)+1 for curr in clens]
    
            # loop through the current scanline's every 8 bits
            for byte in map(lambda x: int.from_bytes(x, 'big'), line):
            
            
                # shift previous value left until we have enough bits left
                while bits_left <= 8 - bits_left:
        
                    prev >>= idxshift[bits_left]; bits_left += idxshift[bits_left]
                    
                # grab the portion of this byte that fits within our current context
                valpart = byte >> (8-bits_left)
                byte &= ~(~(0xFF) << bits_left) 

                # set curval to either the previous value or the initial value
                curval = prev or valpart

                # iterate over the rest of the bits in the byte
                for bitno in range(bits_left+(8-bits_left)-1, -1, -1):

                    buf = valpart >> bitno & 1
                    row += bytes([curval << bits_left])[:buf]

                    if buf == 0: 
                        curval &= ~(~(0xFF) << bits_left)
                    elif buf == 1: 
                        curval |= 1<<(8-bits_left)

                    bits_left -= 1
                            
                prev = curval
    
            # append last value if it was incomplete
            if bits_left > 0 and curval: 
                row += bytes([prev << bits_left])
                
            imgdat += row

        # calculate indexed colors
        indexed_colors = [imgdat[k::3] for k in range(3)]

        # create local color table; make sure that alpha channel comes before RGB so that fancier 
        # editors won't screw with the transparency setting later
        paldata = [[indexed_colors[k][j:j+3] for j in range(k*len(indexed_colors[k]), (k+1)*(len(indexed_colors[k]))) 
                    for k in range(3)]]


        # construct local color tables
        for i in range(1, colormap_size):
            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]

        # write to output stream
        out.write(header + struct.pack('<HHBB', image_width, 
            image_height, ctsize | 0b10000000, 
            0b1 << 7 | 0b1 << 5 | 0b1 << 4 | 
            0b111 << 3 | 0b111) + palents)
    
    # get remaining lzw compressed info from the RNG
    datalen = int.from_bytes(rng.read(2), 'big')
    data = rng.read(datalen)
    out.write(header + data)
