#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each (2 bytes)
    w, h = struct.unpack('<hh', rng.read(4))

    # 5 bits for color table size (1 byte)
    c = ord(rng.read(1)) & 0b00011111

    # VERY HACKY FIX FOR JUMPING PAST COLOR TABLE SECTION
    if c > 7:
        return False
    
    # Value is -1 when no transparency used
    t = -1

    with open('color-table.raw', 'rb') as f:
        f.seek(c * 3)

        colors = list()
        while True:
            try:
                colors += [f.read(3)]
            except Exception as e:
                break
    

    # 1 bit for global color table flag and then by default it's set to on
    gct_flag = struct.unpack('<b?', rng.read(1))[0]

    # We have to fixup the pixel data section of the image in order to properly
    # handle the global color table that we generated earlier
    # For more information check out https://www.w3resource.com/python-exercises/basic/python-basic-1-exercise-85.php
    pixels = []
        
    # If there was a palette defined before this frame we need to offset our
    # reading point by its length so we can read the proper amount of raw data
    # Otherwise just start after the control block since we're not going to be
    # using that anyway
    if gct_flag == True:
        offset = len(colors) * 3 + 1
        pixels += map((lambda x : int.to_bytes(x, 1, "big")), range(offset, ((h*w) * 3) + offset)) 
    else:
        pixels += map((lambda x : int.to_bytes(x, 1, "big")), range(((h*w) * 3))) 

    # Every frame must end with an EOF block which is denoted by ';'
    pixels += b';'

    out.write(header)
    out.write(struct.pack("<HHB", w, h, c))

    # Doing this inefficiently, but whatever ¯\_(ツ)_/¯
    for i in range(len(colors)):
        out.write(colors[i])

    out.write(int.to_bytes(t, 1, "big"))

    # This should equal zero because we didn't implement interlacing or anything
    lzw_min_size = struct.unpack('<B', rng.read(1))[0]

    # LZW min code size has to be less than 12. No idea what I'm doing lol
    assert lzw_min_size < 12

    packed_fields = struct.pack("BBB", gct_flag, lzw_min_size, t)
    out.write(packed_fields)

    out.write(b'\x00\x00\x00')

    for val in pixels:
        out.write(val)

