#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def gen_gif(out: BinaryIO, rng):
    # Header
    out.write(b'GIF89a')

    # Logical Screen Descriptor (7 bytes)
    width = rng.randint(0x14, 0xFFF4 // 2 + 1) * 2
    height = rng.randint(0x14, 0xFFF4 // 2 + 1) * 2
    ncolors = rng.choice((2 ** i for i in range(1, 5)))
    background_color_index = rng.randrange(ncolors)
    aspect_ratio = b'\x00\x64'

    scrdesc = struct.pack('<HHBBB',
        width,
        height,
        ncolors,
        background_color_index,
        0b11000000 | len(aspect_ratio),
    )
    scrdesc += aspect_ratio

    out.write(scrdesc)

    # Global Color Table
    colorcount = min(ncolors, 256)
    colors = [rng.randrange(256) for _ in range(colorcount)]
    out.write(bytes([len(colors)]) + bytes(colors))

    # Image Descriptor (10 bytes)
    left = rng.randrange(width - 2)
    top = rng.randrange(height - 2)
    width = max(2, rng.randrange(min(left + 1, floor(width / 2)), width - left - 1))
    height = max(2, rng.randrange(min(top + 1, floor(height / 2)), height - top - 1))
    local_color_table = bool(rng.getrandbits(1))

    lctsize = None if not local_color_table else \
            rng.choice((2 ** i for i in range(1, 5)))
    interlace = False
    sort = True
    reserved = False
    lctbpp = None if not local_color_table or not lctsize else \
            8 * int(ceil(log2(lctsize)))
    lctsortflag = None if local_color_table and not sort else ord('=')

    imgdesc = struct.pack('<HHHHBB',
        left,
        top,
        width,
        height,
        0b11000000 | \
                ((local_color_table << 7) & 0b10000000) | \
                ((interlace     << 6) & 0b01000000) | \
                ((sort          << 5) & 0b00100000) | \
                ((reserved      << 3) & 0b00010000),
        lctsize or 0,
    )

    if lctsortflag is not None:
        imgdesc += bytes([lctsortflag])
    elif lctbpp is not None:
        imgdesc += lctbpp.to_bytes(1, 'little')

    out.write(imgdesc)

    # Local Color Table (optional)
    if local_color_table:
        colors = [rng.randrange(256) for _ in range(lctsize)]
        out.write(bytes([len(colors)]) + bytes(colors))

    # Graphics Control Extension Block
    hasdelay = bool(rng.getrandbits(1))
    disposalmethod = rng.randint(0, 2)
    userinput = False
    transparentcoloridx = rng.getrandbits(2)
    terminator = b'\x00'

    gce = struct.pack('<BBBBHBBBHB',
        b'\x21',           # Extension Introducer
        b'\xF9',           # Application Extension Label
        4,                # Application Extension Size
        b'\x04',           # Loop Count Sub-block Identifier
        rng.randint(0, 10),         # Loop count value
        ord('!') | \
                ((hasdelay    << 7) & 0b10000000) | \
                ((disposalmethod << 2) & 0b00001100) | \
                ((userinput   << 1) & 0b00000010) | \
                 (transparentcoloridx & 0b00000001),
        0,            # No transparancy color specification
        0,            # End of block marker
    )

    out.write(gce)

    # Plain Text Extension Block
    nxtpage = rng.randint(0x1E, 0xE9 + 1)

    pte = struct.pack('<BBBBHBBBHBB',
        b'\x21',           # Extension Introducer
        b'\x01',           # Plain text extension label
        16,               # Length of plain text sub-block
        b'\x0C',           # Plain text grid left position
        left,             # Plain text grid top position
        width,            # Plain text grid width
        height,           # Plain text grid height
        b'\x00',           # Cell grid left offset
        b'\x00',           # Cell grid top offset
        b'\x00',           # Cell width in pixels
        b'\x00',           # Cell height in pixels
        b'\x01',           # Number of characters in each cell
        nxtpage,          # Character cell data size
        b'\x00',           # Number of filler byte to make the cell data size even
    )

    out.write(pte)

    # Pixel Data Block
    lzwmincodesize = rng.randint(3, 12)
    lzwnonstandard = bool(rng.getrandbits(1))
    lzwnonsemi = bool(rng.getrandbits(1))
    lzwcrccheck = b'\x00' if lzwnonstandard else b'\x03'
    lzwsizeshift = rng.randint(2, 8)
    lzwmaxcodesize = ceil(log2(1 << lzwsizeshift))
    lzwhighbit = 0b10000000 >> lzwmincodesize
    mask = (lzwhighbit << (lzwmaxcodesize - 1)).to_bytes(1, 'big')
    eoi = b'\x00'
    
    flags = [ord('&'), lzwhighbit]
    flags += [None]*(lzwmincodesize - 1)

    oldbits = 0
    newbits = 0

    while True:
        for flag in flags:
            if flag is None:
                continue

            if rng.getrandbits(newbits):
                # Literal code

                val = rng.getrandbits(newbits)
                out.write(val.to_bytes(oldbits//8, 'big'))
                oldbits = newbits

            else:
                # Escaped code

                esccode = rng.getrandbits(newbits)
                assert 1 <= esccode < 2**lzwmincodesize
                
                out.write(((esccode & ~mask) | lzwhighbit).to_bytes(1, 'big'))
                out.write((esccode ^ lzwhighbit).to_bytes(1, 'big'))

        if newbits >= lzwmaxcodesize:
            break
        
        newbits += 1

    out.write(eoi)

