#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    print("Image dimensions:", (image_width, image_height))

    bgcoloridx = ord(rng.read(1)) & 0b11111
    bgcolor = struct.pack('B', bgcoloridx)

    bgcolorflag = 1<<7 if bool(ord(rng.read(1))) else 0
    ext_gct = 1<<7 if bool(ord(rng.read(1))) else 0
    lctsize = ord(rng.read(1)) & 0b1111111
    interlaced = 1 if bool(ord(rng.read(1))) else 0

    # Other flags can be ignored.

    gct = []

    # read in the global color table
    if ext_gct:
        ncolors = int.from_bytes(rng.read(2), 'little')
        gct = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ncolors)]]


    # skip past the rest of the field
    # (background index, pixel aspect ratio,
        # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):

            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]

    else:
        palents = b''

    # colormap is followed by a local color table
    # TODO: we could do something with this; but not now...
    rng.seek((3 << 8) | ctsize)

    datalen = int.from_bytes(rng.read(2), 'big')
    data = rng.read(datalen)
    out.write(header + data)
