#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(rng.read(2), 'little') & 0xffff

    # Image Header
    ihdr = [0] * 10
    ihdr[0] = 0x2c                   # image separator
    ihdr[1] = (int.from_bytes(rng.read(4), 'little')) & 0xff
    ihdr[2] = (int.from_bytes(rng.read(4), 'little')) & 0xff
    ihdr[3] = (int.from_bytes(rng.read(4), 'little')) & 0xff
    ihdr[4] = (int.from_bytes(rng.read(4), 'little')) & 0xff
    ihdr[5] = (int.from_bytes(rng.read(4), 'little')) & 0xf
    ihdr[6] = (int.from_bytes(rng.read(4), 'little')) & 0xf
    ihdr[7] = 0x00                   # no local color table
    ihdr[8] = struct.pack('>B', ((int.from_bytes(rng.read(1), 'little')) & 0xFF))[0]
    ihdr[9] = struct.pack('>B', ((int.from_bytes(rng.read(1), 'little')) & 0xFF))[0]

    out.write(header + bytes(ihdr))

    # Packed Field Section - Only LZW Minimum Code Size is set to 8
    pf = [0b00000000]
    pf[0] |= 0b011
    out.write(bytes(pf))

    # Application Extension
    app = [0] * 11
    app[0] = 0xff                    # application extension block identifier
    app[1] = len(b'NETSCAPE2.0')     # size of application data
    app[2] = ord('N')                # NETSCAPE char
    app[3] = ord('E')
    app[4] = ord('T')
    app[5] = ord('S')
    app[6] = ord('C')
    app[7] = ord('A')
    app[8] = ord('P')
    app[9] = ord('E')
    app[10] = ord('2')
    app[10] = ord('.')
    app[10] = ord('0')
    out.write(bytes(app))

    # Sub-Block containing loop count and terminator
    loop = []
    loop += [0b0] + [0]              # Loop Count sub-block
    loop += [0]                      # Terminator sub-block for loop count
    out.write(bytes(loop))

    # Plain Text Extension
    txt = [0] * 12
    txt[0] = 0xfe                    # text extension block identifier
    txt[1] = 2                       # size of plain text data
    txt[2] = 0x11                    # label flag
    txt[3] = 0xff                    # netscape char
    txt[4] = 0x00                    # terminating byte
    imgdesc = [[0]] * 12             # image descriptor
    imgdesc[0][0] = 0x2c             # image separator
    imgdesc[1][0] = (int.from_bytes(rng.read(4), 'little')) & 0xff
    imgdesc[2][0] = (int.from_bytes(rng.read(4), 'little')) & 0xff
    imgdesc[3][0] = (int.from_bytes(rng.read(4), 'little')) & 0xff
    imgdesc[4][0] = (int.from_bytes(rng.read(4), 'little')) & 0xff
    imgdesc[5][0] = (int.from_bytes(rng.read(4), 'little')) & 0xf
    imgdesc[6][0] = (int.from_bytes(rng.read(4), 'little')) & 0xf
    imgdesc[7][0] = 0                 # local color table present
    imgdesc[8][0] = (int.from_bytes(rng.read(1), 'little')) & 0XFF
    imgdesc[9][0] = (int.from_bytes(rng.read(1), 'little')) & 0XFF
    imgdesc[10][0] = 0                # no background color index specified
    imgdesc[11][0] = 0                # no aspect ratio specified
    out.write(bytes(txt))           # write plain text data
    out.write(bytes([imgdesc]))      # write image descriptor
    out.write(b'\x3b')               # literal block end marker

    # General Graphic Control Extension
    gce = [0] * 8
    gce[0] = 0xf9                    # graphics control extension block identifier
    gce[1] = 0x04                    # size of graphic control data
    gce[2] = (int.from_bytes(rng.read(1), 'little')[0]) | 0b11111000;       # packed field (delay time)
    gce[2] = 0b11                     # delay time value
    gce[3] = 0x01                    # transparent color index bit 0
    gce[4] = (int.from_bytes(rng.read(1), 'little')[0]) & 0b1;          # transparant color index bit 1
    gce[5] = 0                       # blocking execution of extensions
    gce[6] = 0x00                    # size of terminator for GCE
    gce[7] = 0x00                    # terminator for GCE
    out.write(bytes(gce))

    # Image Data
    imgdata = [0] * datalen
    for i in range(datalen):
        imgdata[i] = (int.from_bytes(rng.read(1), 'little')) & 0xff
    out.write(bytes(imgdata));

    # Trailer
    trailer = [0] * 1
    trailer[0] = 0x3b            # gif trailer
    out.write(bytes(trailer))

    # Data Block Terminator