#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # header
    header = bytearray([
        0x47, 0x49, 0x46, 0x38,  # magic
        0xe9,                     # version bytes
        0x01, 0x00,               # width and height
        0xf5])                    # flags

    # color table
    for _ in range(0xff - len(header)):
        header.extend((
            rng.read(1)[0],
            rng.read(1)[0],
            rng.read(1)[0]))

    # netscape application extension
    netscape_appext = bytearray([
        0x21, 0xff,                   # application extension block marker
        0x0b,                         # length of app ext data
        0x4e, 0x65, 0x74, 0x7a,       # NETSCAPE
        0x6c, 0x65, 0x64,             # code name
        0x67, 0x41, 0x6c, 0x70,        # subblock start
        0x68, 0x61, 0x02,             # size is two bytes
        0x00,                        # loop count for animation
        0x3b])                       # end of subblock

    # graphic control extension
    gce = bytearray([
        0x21, 0xf9,                 # graphics control block marker
        0x04,                       # length (four bytes)
        0x00, 0x0f, 0x00,           # delay time 1/15 seconds
        0x00,                       # transparent index
        0x00,                       # terminator
        0xfe,                      # extra block separator
     ])
    gce[8] = rng.read(1)[0] & 0x03    # disposal method (weird order)

    # image descriptor
    idt = bytearray([
        0x2c,                       # image separator
        0x00, 0x00, 0x00, 0x00,     # left position
        0x00, 0x00, 0x00, 0x00,     # top position
        0x00, 0x00, 0x00, 0x00,     # right position
        0x00, 0x00, 0x00, 0x00      # bottom position
    ])
    idt[8] |= ((rng.read(1)[0] & 0x07) << 5)   # local colour palette depth
    idt[9] |= (rng.read(1)[0] & 0x01)            # sort flag

    # table based images
    if idt[8] == 0 or idt[8] > 7:
        imgdata = b''

        while True:
            nbytes = rng.read(1)[0]

            if not nbytes:
                break

            imgdata += rng.read(nbytes)

        imgsize = (len(imgdata) + 1).to_bytes(1, 'big')
        idt[-1:] = [
            0x00 | (idt[0] & 0x07),          # local colour palette depth
            idt[0] & 0x01]                    # sort flag

        # write everything together
        out.write(header)
        out.write(netscape_appext)
        out.write(gce)
        out.write(idt[:-1])
        out.write(struct.pack('B', len(imgdata)))
        out.write(imgsize)
        out.write(imgdata)
        return

    # full images
    imgdata = b''
    for y in range(0, int(pow(2, idt[8])))[::-1]:
        line = []
        for x in range(int(pow(2, idt[8] + 1))):
            line.append(((y >> (7 - int(idt[8] / 2))) & 0x01) ^
                        unpackbits(rng.read(1)[0])[x % 8])

        line.reverse()
        linestring = ''
        for bit in line:
            linestring += ('%s' % bit)

        for i in range(0, len(line), 8):
            imgdata += chr(int(linestring[i : i + 8][::-1], 2)).encode('latin-1')

    out.write(header)
    out.write(netscape_appext)
    out.write(gce)
    out.write(idt)
    out.write(imgdata)

# Unpacks *count* bits from a byte-like object returning them as integers
# in least to most significant bit order. Taken from
# https://stackoverflow.com/questions/1250958/python-bitwise-rotation
def unpackbits(value, count=None):
    bytes = []
    if isinstance(value, str):
        value = ord(value)
    for i in range(count if count else 8):
      bytes.insert(0,(value & 1))
      value >>= 1
    return bytes
