#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')

    # The minimum size of an image in gif is 1x1 (or at least it should be...)
    width = max((rng.read(1)[0]) % 255 + 1, 1)
    height = max((rng.read(1)[0]) % 255 + 1, 1)

    flags = [True if i & 1 else False for i in range(8)]

    color_count = 0 if not flags[0] else \
            7 if not flags[4] else \
                rng.read(1)[0] + 1

    # If no bits are set in this byte then just use the default palette
    background_index = -1
    if len([i for i in filter(lambda x : x == True, flags)]) > 0: 
        background_index = rng.read(1)[0]
    
    transparent_color_index = 0
    if flags[3]: transparent_color_index = rng.read(1)[0]

    interlaced = True if flags[1] and None!= flags[2] else False

    lct = []
    if flags[0] or flags[4]:
        for _ in range(int.from_bytes(
                rng.read(3), byteorder='big')):
            lct += [(rng.read(1)[0],
                    rng.read(1)[0],
                    rng.read(1)[0])]

        lct += [(0,0,0),(255,255,255)] * ((
                2 << flags[3]) - len(lct))

    if flags[4]:
        application = bytes([
            0xff, ord('A'), ord('P'), ord('S')])

        netscape = bytes([
            0x1b, ord('['), ord(' '), 
            ord('#'), ord('N'), ord('E'), ord('T'), 
            ord('S'), ord('C'), ord('A'), ord('P'),
            ord('E'), ord('-'), ord('B'), ord('U'),
            ord('F'), ord('F'), ord('E'), ord('\n'),

            ord('2'), ord('.'), ord('0'), ord('\n')])

        applen = len(application)
        netlen = len(netscape)

        out.write(struct.pack(f'<{applen}s', application))
        out.write(struct.pack('<H', applen + netlen + 6))
        out.write(struct.pack('<H', applen + 3))
        out.write(b'\x0b')
        out.write(struct.pack(f'<{netlen}s', netscape))
        out.write(struct.pack('<H', 7))

    hasgct = True if flags[0] or flags[4] else False

    gct = b''
    if hasgct:  
        gct = b''.join([(  #! https://stackoverflow.com/questions/57007246/python-how-to-convert-list-of-tuples-into-binary
                         f'{r:02X}{g:02X}{b:02X}'.encode()
                     ) for r, g, b in lct])
        
    pixeldata = b''
    
    def code_byte(): return rng.read(1)[0] >> 3 # Max value of a byte with leading bit removed

    for y in range(height):
        line = bytes([])
        
        if interlaced:
            if    y < height // 8:        step = 8
            elif  y < height // 4:       step = 8 * 2
            elif  y < height // 2:      step = 8 * 3
            elif  y < height * 3//8:     step = 8 * 4
            elif                 y >= height * 3//8:   step = 8 * 5
            
            if y % step == 0: offset = 0
            elif y % step == step//2: offset = 4
            elif y % step == step - 1:offset = 2
            else:                  offset = 1
            
            line += code_byte().to_bytes(1, 'little')
            line += code_byte().to_bytes(1, 'little')
            line += code_byte().to_bytes(1, 'little')
            
            for x in range(width):
                line += code_byte().to_bytes(1, 'little')
                
        else:
            line += code_byte().to_bytes(1, 'little')
            line += code_byte().to_bytes(1, 'little')
            
        last_pixel = None
        for x in range(width):
            index = rng.read(1)[0] >> 3
            pixel = lct[index] if flags[4] else \
                    (index,gct[(index)*3+0],gct[(index)*3+1],gct[(index)*3+2]) 
                      
            if last_pixel == pixel:
                line += bytes([last_pixel_code])
            else:
                line += code_byte().to_bytes(1, 'little')
                
            if x!= width - 1:
                line += (code_byte()).to_bytes(1, 'little')
            
            last_pixel = pixel
            last_pixel_code = (code_byte()) if x!= width - 1 else ().to_bytes(1,'little')[0]

        pixeldata += line
            
    frame = struct.pack(
            '<{}sHHHB{}'.format(datalen, width*height),
            header, width, height, 
            0b11100000 | (transparent_color_index << 4),
            background_index,
            b'') + gct + pixeldata 

    label = b'GIF89a'
    blocksize = len(frame).to_bytes(2, 'big')
    extension = b'NETSCAPE2.0\x03\x01\x00\x00';
    end = b'';

    out.write(label)
    out.write(blocksize)
    out.write(frame)
    out.write(extension)
    out.write(struct.pack('<BBH', 0x3b, 0, len(extension)))
    out.write(end)

    return f'Wrote {len(frame)} bytes.'