#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # sort flag and interlace flag map to the same values here as
    # they do in the real world so it's not much of an exercise.)
    out.write(header
              + rng.read(7) # Logical Screen Descriptor
              )

    block = rng.read(1)[0]
    # the next field describes if there'll be a global color table or not
    has_global_color_table = bool(block & (1 << 7))
    # how many colors will be available from the color tables?
    # 2 ** (this value + 1); valid values are [1.. 7] where 1 means
    # only a single palette entry / a black + white pair
    gctsize = ((block >> 3) & 0b111) + 1

    # one byte for the block containing the LZW minimum code size
    codesizelzw = (rng.read(1)[0] & 0b00011111) + 1

    subblocks = []
    while True:

        # read the size of the current sub-block
        sub_block_size = int.from_bytes(rng.read(1), 'big')

        # if the lower bit of the size byte was set then there's more blocks
        if sub_block_size & (1 << 7):
            sub_block_size ^= (1 << 7)

        # otherwise we're done
        elif sub_block_size == 0: break

        subblocks.append(rng.read(sub_block_size).strip())


    # after these two blocks come a table of color indexes to be used
    # as a palette for the actual color data
    ctsize = int.from_bytes(rng.read(1), 'big')
    assert ctsize < 256,'something went wrong with the encoding of the GCT'

    # are there any extra "special" flags?
    flags = ord(rng.read(1))
    transparencyflag = bool(flags & 1)
    userinputflag = bool(flags & 2)
    delaytimeflag = bool(flags & 4)
    transparantindexflag = bool(flags & 8)

    if has_global_color_table:
        # calculate the amount of bytes consumed by all the GCT entries
        # (computed as 3 * 2**(gctsize+1)). Even though most modern gif
        # viewers can handle up to 256 colors per frame, some allow arbitrary
        # amounts of palette (up to 256 * 2**gctsize total).
        gctentries = 3 << (gctsize + 1)

        # consume those bytes of data
        rng.seek(-gctentries, 1)

        # read them again so we don't have to worry about having rand()'ed
        # past the end of our buffer. the seed should always be big enough!
        gctdata = list(map(lambda x: tuple(x),
                           struct.iter_unpack('BBB',
                                             rng.read(gctentries))))

        # old style gifs also had a minimal code size here before the GCT
        # which seems to have been replaced by just setting the maximum code
        # size as part of the LZW compression parameters. We won't bother
        # implementing anything related to it right now, since it doesn't seem
        # to really matter anyways. If you care about it you're probably doing
        # things wrong.

        # finally construct the palette data out of the raw bytes
        paldata = [[r, g, b]
                   for r, g, b in
                    [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):

            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]

    else:
        palents = b''

    # colormap is followed by a local color table
    # TODO: we could do something with this; but not now...
    rng.seek((3 << 8) | ctsize)

    datalen = int.from_bytes(rng.read(2), 'big')
    data = rng.read(datalen)
    out.write(header + data)
