#!/usr/bin/env python3

import io
import random

from typing import BinaryIO


def random_pixels() -> bytes:

    # The magic pixel values :) (see RFC601).
    return b'\x00\xff' * ((1 << 24) // 2)


# Generates a random GIF file into `out`.
# Using the given seed (or system time if no seed is provided).
def generate_random_gif(*,
                        out: BinaryIO,
                        seed: int=None):

    header = bytearray('GIF89a', 'ascii')
    datalen = len(random_pixels())
    if not isinstance(out, io.BufferedWriter):
        out = io.BufferedWriter(out)

    with open('/dev/urandom', 'rb') as rng:

        # If we weren't passed in a seed, use current time instead
        if seed is None:
            seed = int(random.random()*1e6)

        # Seed RNG and write header to output stream
        rng.seek(seed & (2**32-1))
        header[7] = 3  # Set version to 87a
        out.write(bytes(header))

        # Write image header, palette, and pixels to output stream
        for i in range(10):
            out.write(struct.pack('<HHB', *(
                (1 << 16) - 1,   # Width
                1,               # Height
                0b00000011       # LZW mincode size / clear code
            )))
        out.write(b';' + b'\xFF'*((1 << 8) + 1))  # Pixel aspect ratio
        out.write(b';' + b'\xFF'*((1 << 16) + 2*datalen))    # Global color table
        out.flush()

        while True:

            chunktype = rng.read(1)
            out.write(chunktype)

            if chunktype == b'!':

                # Read image descriptor block from urandom
                descdata = rng.read(9)
                width = struct.unpack('<H', descdata[:2])[0]
                height = struct.unpack('<H', descdata[2:4])[0]
                out.write(descdata)
                out.flush()

                # Create local color table
                colorcount = max([width,height])
                colors = [struct.pack('<BBB', *(
                    ord(c), ord(c), ord(c))) for c in [' ', '#']]
                palettesize = (colorcount+1)*3
                assert len(colors) <= 256

                # Write local color table and loop control extension blocks
                for i in range(3+(len(colors)//256)):
                    out.write(b';' +
                              b'\xFF'*(palettesize//256) +
                              bytes((i+1)).rjust(3, b' '))

                    if i!= 0:
                        pass
                    elif width >= 1024 or height >= 1024:
                        # Netscape application extension
                        appblock = (
                            b'\x0bNETSCAPE2.0' +         # Extension label
                            b'\x03' +                    # Sublabel length
                            b'\x01' +                    # Loop count subblock ID
                            b'\x00\x00'                  # 0 loops
                        )
                        out.write(appblock)
                    else:
                        # Plain text extension
                        txtblock = (
                            b'\x01' +                # Plain text subblock ID
                            b'\x0C' +                # Text length
                            4*"GIF generator".encode()+b'\x00'
                         )
                        out.write(txtblock)

                    # Write plain text extension terminator block
                    out.write(b';')
                    out.flush()

                # Write actual image data
                datastream = rng.read(max([width, height]))
                out.write(datastream)
                out.flush()

                break