#!/usr/bin/env python3

import base64
import io
from typing import Optional

def generate_random_gif() -> bytes:
    with open('/dev/urandom', 'rb') as rng:
        buf = io.BytesIO()
        generate_random_gif(rng, buf)
        return buf.getvalue()


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # gif header magic number
    out.write(b'GIF89a')

    # image size
    width = None
    height = None

    while True:

        c = next(rng).to_bytes(1, byteorder='big')
        
        if len(c):
            assert c[0] in [
                255,    # application extension
                219,    # global color table
                220,    # local color table
                221,    # graphics control block
                33,     # image separator
                ord(';'),# comment separator
            ]

            if c[0] == 255 and (width is not None or height is not None):
                break
            elif c[0] == 219:
                # we can't use this to set our dimensions because it's an arbitrary length
                pass
            elif c[0] == 220:
                # i think this means that the last palette was a global one instead of local, which means there's only 256 colors available in the palette
                pass
            elif c[0] == 221:
                # end graphic control block
                raise RuntimeError("can't handle graphic control")
            else:
                continue
            
            
        n = int.from_bytes(rng.read(1), byteorder='big', signed=False)
        data = rng.read(n)
                    
        out.write(data)

        if c[0] == 33:
            imgdesc = data[:9]
            
            lsb = lambda idx: imgdesc[idx] & 0x01
            
            flagsgct = bool((imgdesc[-1] >> 7) & 0x01)
            interlaced = bool((imgdesc[-1] >> 6) & 0x01)

            if flagsgct:
                offset = -1
                
                for i in range(len(imgdesc)-1)[::-1]:
                    if imgdesc[i] & 0x01:
                        offset = i+1
                        break

                gctsize = ((imgdesc[-1] & 0xf8) >> 3) + 1
                gct = imgdesc[offset:]

                numcolors = max([x for x in gct]) + 1
                
            w = int.from_bytes(data[5:6], byteorder='big', signed=False) 
            h = int.from_bytes(data[7:8], byteorder='big', signed=False)

            haslocalpalette = lsb(7)
            localpalette = []
            if haslocalpalette:
                localpalette = data[9-w*h:(9+numcolors)]
            
            hasinterlace = lsb(5)
            hassortflag   = lsb(4)
            hassizefields = lsb(3)
            hascolortable = lsb(2)
            hasglobalpalette = lsb(1)

            if flagsgct:
                numcolors += gctsize

            if (not lsb(0)) or (width!= w or height!= h):
                width = w
                height = h
            
    def chunks(lst: list, chunk_size : int = 8):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), chunk_size):
            yield lst[i:i + chunk_size]
    
    def getbyte():
        return int.from_bytes(next(rng).to_bytes(1,'big'), byteorder='big', signed=False)

    # frame count may be longer than the value encoded in this field
    framecount = 0
    while True:        
        fc = next(rng).to_bytes(1, byteorder='big')

        if len(fc):
            assert fc[0] >= 0 and fc[0] <= 9
            framecount |= fc[0] << (framecount.bit_length()-8)
        else:
            break
        
    out.seek(-1, 1)
    out.write(framecount.to_bytes(1, byteorder='big'))    
    print(f"set GIF framecount to {framecount}")
    