#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Returns a function that generates an unsigned integer of size `n`.
#
# The size is in bits; it can be either 8 or 16 (for example).
#
# This will loop over multiple bytes if necessary to fill up `n` bits.
def uint_generator(n):
    assert n == 8 or n == 16

    def f():
        r = []

        while len(r) * 8 < n:
            res = ord(next(gen))

            # Loop until we have enough space for this many bits
            if not len(r) and res >> (n - 8):
                continue

            r += [res]

        return sum([i << s for i, s in enumerate(reversed(range(len(r))))]) % (
            2 ** n)

    return f


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    head = b'GIF89a'
    gen = iter(bytearray(rng.read()))


    out.write(head)
    _append = lambda t: type(t)(out.write(t))[0]
    append = lambda v: _append(v.to_bytes(1, 'big'))

    width, height = map(uint_generator(16), range(2))()

    # Global color table exists iff the next bit is set.
    global_color_table_exists = bool(ord(next(gen)))
    append(global_color_table_exists)

    # Color resolution
    append((7 & ord(next(gen))) >> 4 | ((7 & ord(next(gen))) & 0xF) << 4)

    # Sort flag
    append(((bool(ord(next(gen))) ^ 1) << 1) |
           (((bool(ord(next(gen))) ^ 1) << 1) & 1))

    bits_per_pixel = {
        0b10: 1,
        0b11: 2,
        0b100: 4,
        0b101: 8
    }[ord(next(gen))]

    # Size of local color table
    local_color_table_size = \
        (bits_per_pixel if bits_per_pixel!= 8 else 3) * 3 * 256
    append(local_color_table_size)

    # Background color index
    append(None if local_color_table_size <= 2 else ord(next(gen)))

    # Pixel aspect ratio
    append(None if local_color_table_size >= 2**(bits_per_pixel+1) else
           ord(next(gen)))



    # Write fixed length section
    # Header
    append('NETSCAPE2.0')

    0x3B.to_bytes(1, 'big', out) # Extension introducer
    0x00.to_bytes(1, 'big', out) # Subblock length
    append(b'\x03\x01\x00\xFF') # The whole subblock as one string

    # Graphics control extension
    append('\xFF')          # Extension introducer
    append('')              # Subblock length
    append(b'\xf9')         # Block label
    append(None)            # Delay time
    append(None)            # Transparent color index
    append(False)           # Terminator

    # Image descriptor
    append('\x2C')                 # Image separator
    append(width.to_bytes(2, 'big'))             # Image left position
    append(height.to_bytes(2, 'big'))            # Image top position
    append(width.to_bytes(2, 'big'))             # Image width
    append(height.to_bytes(2, 'big'))            # Image height
    append(False)                    # Packed field
    append(bits_per_pixel)                   # LZW minimum code size
    append(None if True else '\x80')        # Plain RGB client color table
    append(None if False else '\x00')       # First entry in local color table
    append(None if False else '\xff')       # Last entry in local color table

    # Image data
    append(None if True else ''.join(chr(y) for y in height * [width]))









if __name__ == '__main__':
    import sys
    from os import urandom

    with open('/tmp/test.gif', mode='wb') as fd:
        generate_random_gif(urandom(), fd)