#!/usr/bin/env python3

import math
import struct

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(rng.read(2), 'little') & 0xffff

    # Image Header
    width = int(((int.from_bytes(rng.read(1), 'little')) & 0xFF) ** 4) % 320 + 1
    height = int(((int.from_bytes(rng.read(1), 'little')) & 0xFF) ** 4) % 200 + 1
    ihdr = [width, height, 0, 0b0010]     # 1 Bit + RGB Color Table
    lzwmincodesize = 2                  # minimum code size used by LZW
    out.write(header + bytes([
        ihdr[0],                        # Logical Screen Width
        ihdr[1],                        # Logical Screen Height
        0b1110 | (((math.log2(lzwmincodesize)+1)%8)<<5) |   \
            ((ihdr[2]%32)<<1) |       \
            ((ihdr[3]>>1)&1),     # Flags
        lzwmincodesize                 # Minimum Code Size
     ]))

    if ihdr[3]&0b1 :                    # Global Color Table Exist
        palettebuf = []
        for _i in range(0, 2**((ihdr[2]%32)-1)):
            r = (int.from_bytes(rng.read(1), 'little')&0xff)**6 / 255
            g = (int.from_bytes(rng.read(1), 'little')&0xff)**6 / 255
            b = (int.from_bytes(rng.read(1), 'little')&0xff)**6 / 255
            palettebuf += [r,g,b]

        # Write Palette to File
        palend = len(palettebuf)//3
        out.write(bytes([(palend*3)]))      # Color Resolution
        out.write(bytes([0]))               # Sorting Method
        for i in range(len(palettebuf)//3):
            out.write(struct.pack('>BBB',
                round(palettebuf[(3*i)+0]*255),
                round(palettebuf[(3*i)+1]*255),
                round(palettebuf[(3*i)+2]*255),
             ))

    # Local Color Table
    out.write(bytes([0])*3*(3**(ihdr[2]%32)-1))

    # LZW Compressed Bytes
    data = [(int.from_bytes(rng.read(1), 'little')&0xff)**6 //255 for _i in range(datalen)]
    compressor = dict()
    def compress():
        nonlocal compressor, out
        prev = -1
        while True:
            try:
                byteseq = next(dataiter)
            except StopIteration:
                break

            if not isinstance(byteseq, str):
                byteseq = chr(byteseq)
            subseq = ''
            for c in byteseq:
                subseq += c
                if not subseq in compressor:
                    yield f'{len("{subseq}")}:{subseq},'
                    compressor[subseq] = None
                    subseq=''

                elif subseq == '':
                    continue

                else:
                    yield subseq
                    subseq=''

            if subseq!= '':
                yield subseq

    compressor['\x00'] = '\x00'                     # Fully transparent
    compressor[','] = ','                           # End of block
    compressor['\xff'] = '\xff'                     # Transparent
    dataiter = iter(compress())

    nextblock = next(dataiter).encode('utf-8')+next(dataiter)
    nxtlen = len(nextblock)
    pixbuf = ['\xff'*width]*height
    pos = 0
    while nextblock:
        if nxtlen > 255:
            raise ValueError("Block too big")
        out.write(bytes([nxtlen]))    # length of this block
        out.write(nextblock)          # actual block data

        if nxtlen in {1, 129}:        # End of Information
            break;

        putpos = 0
        for token in nextblock.decode().split(','):
            if ':' in token:
                subtoken = token.split(':')[1]
                if subtoken in {'\x00'}:
                    for x in range(putpos//width,(putpos+(width*height))//width):
                        pixbuf[x][putpos%width]='\x00'
                elif subtoken in {'\xff','}'}:
                    pass
                else:
                    sublength = int(token.split(':')[0])

                    if subseq == '':
                        pos = (pos + width)//width*width
                    if pos >= width*height:
                        raise IndexError("Too much LZW compression")
                    for j in range(sublength):
                        pixbuf[pos//width][pos%width] = ord(subtoken[-1:])
                        pos+=1


            else:
                subseq = token
                if subseq == '':
                    pos = (pos + width)//width*width
                if pos >= width*height:
                    raise IndexError("Too much LZW compression")


                for j in range(len(subseq)):
                    pixbuf[pos//width][pos%width] = ord(subseq[-1:])
                    pos+=1


    # Image Data Block
    out.write(bytes([0])*width*height)         # Image Data Padding Rules
    for y in reversed(range(height)):
        for x in range(width):
            out.write(struct.pack(">BBB",
                pixbuf[y][x][0],
                pixbuf[y][x][1],
                pixbuf[y][x][2],
               ) if pixbuf[y][x] is not None else bytes([0xff, 0xff, 0xff]))

    # Data Block Terminator