#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # Generate global header.
    global_hdr = b''

    # Header version field is set to "87a".
    hdrver = chr(int('8', base=16)) + chr(int('7', base=16)) + \
             chr(ord('a'))
    global_hdr += hdrver.encode()

    # Field in which we will store the width and height of our
    # generated frame.
    width_height = [0x00] * 2
    
    # Determine if we're going to write more than one frame or not.
    numframes = ord(rng.read(1)) % 0xFF

    # Set global colors table flag.
    gctflag = (numframes & 0x80) >> 7 == 1;
    if gctflag:
        # We're writing multiple frames. Should have a global color 
        # table after this flag. Add a null at the end of it so that
        # there's always an even amount of bytes.
        numframes &= ~0x80
        numframes |= len(width_height)
        width_height.append(0x00)

    # Global palette size flag - should be zero for now!
    palsizflg = rng.read(1);

    # Field for storing how many bits per pixel each frame has.
    pxldepth = [0x00]

    # Fields for storing the LZW minimum code sizes. These may change 
    # every time we produce a new output!
    lzwsizes = [[0x00], [0x00]]

    # Get global color table length, if any.
    glbctlen = []

    # If there was a GCT, read its length from our random source.
    # The length is encoded as a 3-bit number with the highest bit being
    # a sentinel value indicating whether or not this color table exists.
    if gctflag:
        hi = ord(rng.read(1));

        while ((hi & 0x80)!= 0):
            glbctlen.append((hi & 0x7f))
            hi = ord(rng.read(1));

        glbctlen.append(hi & 0x7f)

    # Write out fields for each individual frame.
    for i in range(numframes):
       # Frame descriptor field.
       frmdescrpt = [];

       # Flags field contains type of graphic extension used by this 
       # frame. We'll use plain old graphics control extensions for now.
       flags = b'\xf9';

       # To calculate delay in ms between frames, we need to know the
       # duration of the previous frame. This means we can't just add
       # the 16-bit values inside `frmdscrpt`, because those wrap around
       # when they get too big. So what we do instead is read both the
       # numbers directly from the RNG and then reset when they wrap over.
       frmdualfrmtime = [0x00]*2
       frmlastfrmtime = [0x00]*2
       frmdualfrmtime[0] = ord(rng.read(1))
       frmlastfrmtime[0] = ord(rng.read(1))

       # Same story here; we save the entire 16-bit duration before
       # converting them to milliseconds again.
       frmdualfrmtime[1] = ord(rng.read(1))
       frmlastfrmtime[1] = ord(rng.read(1))

       # Check if either duration field wrapped over midway through.
       if (((frmdualfrmtime[0] < frmlastfrmtime[0]) and 
            (frmdualfrmtime[1] >= frmlastfrmtime[1])) or 
           ((frmdualfrmtime[0] >= frmlastfrmtime[0]) and
            (frmdualfrmtime[1] < frmlastfrmtime[1]))):
           # Reset all relative timing fields.
           frmdualfrmtime = [0x00]*2
           frmlastfrmtime = [0x00]*2

       # Add absolute timing information to frame descriptor field.
       frmdescrpt.extend([
                frmlastfrmtime[0], 
                frmlastfrmtime[1]])

       # Add current duration delta to frame descriptor field.
       frmdescrpt.extend([
                frmdualfrmtime[0], 
                frmdualfrmtime[1]])

       # Two reserved bytes at end of frame descriptor field.
       frmdescrpt.extend([0x00, 0x00]);

       # Write frame descriptor field.
       global_hdr += b''.join([
               int.to_bytes(
                    val, 
                    byteorder='big', 
                    signed=(val <= 127))[::-1] 
               for val in frmdescrpt][::-1])

       # Write out graphic control block size.
       global_hdr += b'\x04'

       # Write out graphic control block contents.
       global_hdr += flags + \
                     int.to_bytes(i,
                                  byteorder='little',
                                  signed=True)[::-1] + \
                     b'\x00'*2

       # Write out local color table flag. Should be false since we don't
       # have a local color table yet.
       localctflag = b'\x00'

       # Local color table length field, but only if needed...
       loccstlentmp = []
       if (localctflag == b'\xff'):
           # There was a GCT before us, so let's see if we need one now.
           if (glbctlen!= []):
               loccstlentmp.append(glbctlen.pop())

           else:
               # Nope, no GCT either. Let's make some up.
               loclcnt = ord(rng.read(1)) % 0xFC

               for _ in range(loccnt // len(width_height)):
                   loccstlentmp.extend(width_height)

              # Fill up all remaining slots with random data.
              loccstlentmp.extend([
                       ord(rng.read(1)) 
                       for _ in range(
                           loccstlentmp[-1] - len(loccstlentmp))
                      ])
              pass
           pass

       # Local color table length field itself.
       localctllen = len(loccstltmp).to_bytes(length=1,
                                              byteorder='big')

       # Write out local color table length.
       global_hdr += localctllen

       # Write out actual local color table.
       global_hdr += b''.join([
               int.to_bytes(
                    val, 
                    byteorder='big')[::-1]
               for val in loccsllentmp][::-1])
 
       # Image separator must be present before every frame except for first.
       imsep = b'\x2C'

       # Image block size. Should never exceed 0xFFFF.
       imgblksz = int.to_bytes(((pxldepth[0]+1)*9)//8*width_height[0]*
                               height_width[1],
                               length=2,
                               byteorder='big')

       # Image left position. Always set to zero.
       imglposn = b'\x00'*2

       # Minimum code size for LZW encoding. May vary between different
       # outputs.
       mincodesz = lzwsizes[(i+1) % len(lzwsizes)]

       # Write out image block contents.
       global_hdr += imsep + imgblksz + imglposn + mincodesz

       # Data subblock length field. Should be equal to PxlDepth + 1.
       datsubblkln = b'\x00' + bytes([pxldepth[0]+1])

      # Write data subblock length.
       global_hdr += datsubblkln

       # Color table identifier. Must be zero unless there actually is a
       # color table somewhere in the file already.
       clrtbleid = b'\x00'

       # Write data subblock contents.
       global_hdr += clrtbleid + \
                     b''.join([
                         int.to_bytes(
                             0,
                             spanned=1,
                             byteorder='big',
                             signed=False)
                          for x in range(imgblksz)])


    # Now comes the tricky part. We've got to determine where the global 
    # headers ends and the image data starts. And unfortunately, there isn't
    # any strict format standard for GIFs, so this is probably as accurate
    # as we'll ever get.

    # Get total length of global header, including trailing NULL byte.
    globhdrln = sum([len(field) 
                     for field in global_hdr]) + 1 

    # As long as the global header length doesn't start with a BOM, we can
    # assume we're dealing with the normal GIF format.
    # TODO: Check if this is faster?
    if globhdrln & 0xFF == 0xFF:
        # Need to pad global header until it does start with a BOM.
        padding_needed = (globhdrln | 0xFFFFFF00) ^ 0xFFFFFFFF
        global