#!/usr/bin/env python3

import sys
import argparse
import struct
try:
    from PIL import Image
except ImportError as e:
    print("You need Pillow installed, try:")
    if sys.platform == "linux":
        print("$ apt install libjpeg-dev zlib-dev")
    elif sys.platform == "darwin":
        print("$ brew install libjpeg zlib")
    else: raise e

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    # We need to make sure there is a valid palette so we can do the 
    # actual encoding into the LZW code
    if ctsize >= 2**int(args.minbits):

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):
            
            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]

        # write to output stream
        out.write(header + struct.pack('<HHBB', image_width, 
            image_height, ctsize | 0b10000000, 
            0b1 << 7 | 0b1 << 5 | 0b1 << 4 | 
            0b111 << 3 | 0b111) + palents)
        
        # now that we have encoded the color table, rewind to where 
        # we started reading from the RNG to actually encode the image
        rng.seek(2+len(palents))

        # `image_width` pixels per row, `image_height` rows total
        imdata = []
        for i in range(image_height):
            
            row_start = i*image_width*3
            row_end   = row_start + image_width*3

            imdata.append([tuple(rng.read(3)) for j in 
                range(row_start, row_end)])
    else:
        # recreate the first frame but with only black pixels
        imdata = [[(0, 0, 0)]*(image_width*3)]*(image_height)
    
    # encode every frame into its own block
    frames = list()
    while True:

        # encoder requires us to put all previous images into one image
        # so we'll create an array of images here.
        combined_im = None
        for f in range(args.frames):

            # starting bytes for this frame's header 
            frame_header = [
                struct.pack('<BB', 
                    image_width & 0xFF,
                    image_height & 0xFF),

                # the rest of the bytes in the frame header don't matter since
                # we're not writing them
                0b10 << 7 | 
                0b11 << 4 | 
                0b1111
            ]

            # read in LZW compression info
            lzllen = int.from_bytes(rng.read(1), 'little')
            lc = int.from_bytes(rng.read(1), 'little')
            lext_flag = bool(lext_mask & rng.read(1)[0])
            ldisplen = int((lc & 0b11110000) >> 4)
            ldistlen = int((lc & 0b1111))
            lext_mask  = rng.read(1)[0]
    
            # read in the LZW mincode sizes
            ext_minlens = [int.from_bytes(rng.read(1), 'little')]
            if lext_flag:
                ext_minlens.extend(\
                    [int.from_bytes(rng.read(1), 'little')
                        for _ in range(ldisplen)])
            rle_minlens = [int.from_bytes(rng.read(1), 'little'),
                int.from_bytes(rng.read(1), 'little')]

            # read in LZW codes themselves
            ext_codes  = []
            rle_codes  = []
            ext_rle    = False
            while len(ext_codes)+len(rle_codes)<lc:
                
                code_byte = rng.read(1)[0]

                # check if it's a clear code
                if code_byte==0x00:

                    # reset our current decoder state variables
                    ext_codes=list()
                    rle_codes=list()
                    ext_rle  =False
                    
                    continue
                
                # check if this code is long or short
                code_length = -(~code_byte)
                if ~code_length&0b11000000:

                    # this next part gets tricky because we might be looking at
                    # a displmented long code, or we might just be looking at
                    # something like a regular unextended code. This stanza
                    # figures out which case is happening by checking if the
                    # top two bits of the byte are set. If they are set then
                    # this is definitely a displacement code, otherwise it could
                    # still be an extended code
                    disp_bitfield = code_byte & 0b11000000
                    if ~disp_bitfield&0b11000000:

                        # this must be a displaced code! let's grab the 
                        # displacement bitcount and figure out how many more
                        # bytes we will need to read before decoding
                        disp_length = ~(~code_byte)&0b11000000
                        displens = [
                            int.from_bytes(
                                rng.read(1), 
                                'little'
                            ) for _ in range(ldisplen)
                        ]
                        
                        # decode the displacment code itself
                        disp_code = sum([(displens[j]*pow(2,(ldisplen-(j+1))) 
                            if j<ldisplen-1 else displens[-1])
                            for j in range(ldisplen)])
                        
                        # perform substitution on existing code
                        # NOTE: THIS IS NOT THE TRUELY WEIRD WAY OF DOING
                        # DISPLACED CODES BUT IT WORKS SO WHATEVER
                        # NOTE: also, I'm sure there's a simpler way of doing
                        # the following line... I just haven't worked through
                        # it yet :P
                        new_code = ext_codes[disp_code]+\
                            rle_codes[disp_code]
                        ext_codes.append(new_code)

                        continue
                        
                    else:
                        # nope, this was just a regular unextended code
                        # add it to our current code list
                        ext_codes.append(code_length)
                        continue
                else:
                    # handle special cases before ordinary codes
                    if code_length>0:
                        # this means we have an RLE code
                        rle_codes.append(code_length)
                        continue
                    else:
                        # this means we have an extended code
                        ext_rle =True
                        break
                    

            # ok, we've got our codes! Now what?
            if ext_rle:

                # read in the remainder of the bits needed to fully decode the 
                # codestream
                ext_stream = bytearray()
                while len(ext_stream) < pow(2,lext_mask)+1:
                    ext_stream.extend(rng.read(1))

                # use the bitstream to assemble the final codestream
                ext_stream = ''.join(['{:08b}'.format(b)[::-1]\
                                        for b in ext_stream])

                # construct the full codestream
                ext_stream = ext_stream[:lext_mask+1].encode()

                # build dictionary based on minimum lengths that were provided
                ext_dict = dict()
                for idx, l in enumerate(ext_minlens[:-1]):
                    ext_dict[(idx+1, False)] = ext_stream[:l][::-1]\
                                               .decode().strip('\0')
                for idx, l in enumerate(rle_minlens):
                    ext_dict[(len(ext_codes)+idx, True)]=\
                            '{}{}'.format(chr(rle_minlens[0]), chr(rle_minlens[1]))

                # decode the full codestream
                # NOTE: The usual way of implementing this algorithm would take up too much memory
                # So instead I made it work stepwise until it ran out of space in memory
                image_data = ''
                while len(ext_dict)>1:
                    #