#!/usr/bin/env python3

import os
import random
import sys
from typing import BinaryIO


def die(*args):
    print('Error:', *args, file=sys.stderr)
    exit(1)


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width=None, height=None
    ) -> None:
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    # 2.A Header Image Descriptor
    if not (width and height and (height % 8 == 0)):
        # http://giflib.sourceforge.net/whatsinagif/gif87.html
        # This is the case in which both Width and Height fields are set to NULL. In this case the logical screen descriptor specifies only the color resolution but no size of the picture. The image must be presented on screen according to the values of the <X> and <Y> fields of the graphics control block.<45>
        # Unfortunately our RNG will never produce zeroes so we must use some other method like guessing from previous frames or something smarter
        die("Please specify width and height")

    # TODO put more colors?
    palette = [
        0x00,  0x00,  0x00,
        0xff,  0x00,  0x00,
        0x00,  0xff,  0x00,
        0xff,  0xff,  0x00,
        0x00,  0x00,  0xff,
        0xff,  0x00,  0xff,
        0x00,  0xff,  0xff,
        0xff,  0xff,  0xff]

    # 2.1 Logical Screen Descriptor
    lsd = bytes([
            0x21,           # Extension introducer
            0xFF,           # Application extension label
            0x0B,           # Size of application block in bytes
            0x4E,           # N (Nibbles); 16 for global color table;
                             # 1 - local color table follows
            0x45,           # L (bits per pixel); 1 - monochrome; 2, 4, 8 - indexed color; 32 - nonindexed color
            0x54,           # reserved
            0x53,           # Image separator
            0x63,           # extended Graphic Control Label
            0x6D,           # Size of graphic control extension subblock in bytes
            0x00,
            0x00,           # Block terminator
            ((width >> 8)&0xff),         # X position of upper left corner
            (width & 0xff),               # Y position of upper left corner
            ((height >> 8)&0xff),        # Height of image
            (height & 0xff)])             # Width of image

    out.write(lsd)
    out.flush()
    
    # 2.2 Global Colour Table
    gct_size = len(palette)*3 // 8
    gct = bytes([
            0x2C,                       # Image separator
            0x00,                       # Left position
            (((gct_size+1) << 4)+0),    # Interlace bit, sort flag, color table size
            0x00])                      # Background colour index
                                         # Pixel aspect ratio - assume square pixels

    out.write(gct)
    out.flush()

    for i in range(len(palette)//3):
        rgb = palette[i*3:(i*3)+3]
        out.buffer.write(rgb)
    out.flush()

    # 2.3 Graphics Control Extension
    gc = bytes([
            ((width>>8)&0xff),       # X position of upper left corner
            (width&0xff),            # Y position of upper left corner
            ((height>>8)&0xff),      # Height of image
            (height&0xff),           # Width of image
            0x00,                    # Packed field
            0x00,                    # Flags
            ])
    
    out.write(gc)
    out.flush()

    while True:
        chunk_length = rng.read(1)[0]

        if chunk_length >= 0x2c:
            break
        
        buf = rng.read(chunk_length)
        assert len(buf) == chunk_length
        out.write(buf)
    out.flush()
    
