#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng`. The returned value is
# the size of the random data written to `out`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, max_size=None
    ) -> int:

    # If not specified, find the maximum possible byte for each field
    if max_size == None or max_size < (6+7) :
        max_size = min((6+7), ((1<<24)-max_size))
    elif max_size > (1<<24):
        raise ValueError('Maximum size cannot be larger than '
                         f'the GIF file\'s capacity ({1<<24})')

    # Generate the basic header
    palette = [0] * (3 << 16)
    out.write(b'GIF89a')
    headerlen = out.tell()

    while True:
        ctype = ord(rng.read(1))

        # break out when there's no more data
        if len(palette) <= 0 and len(data) <= 0:
            return out.tell() - headerlen

        # already filled the palette -- now we need data
        if len(palette) <= 0:
            continue

        # Add color table
        if ctype == 0x2c:
            # Write real colors from the last palette
            for i in range(len(last_colors)):
                out.write(struct.pack('<BBB', *(
                    last_colors[i][j] for j in reversed(range(3))
                        )))

            # Now fill with zeros up until 256 colors are left
            for _ in range(-len(last_colors), -len(palette)//3):
                out.write(b'\x00\x00\x00')

        # handle image descriptor block
        elif ctype & 0xf0 == 0x20:
            imlen = ord(rng.read(1))
            if imlen!= 0:
                out.write(imlen.to_bytes(1, byteorder='little'))
            else:
                out.write(b'\xff')

            out.write(b'\x00'*9)

        # handle local color map
        elif ctype == 0x21:
            size = ord(rng.read(1))
            if size == 0:
                out.write(b'\x00')
                out.write(b'\x00'*9)
                continue

            out.write(size.to_bytes(1, byteorder='little'))
            subctype = ord(rng.read(1))

            # Global color map
            if subctype == 0x01:
                for _ in range(size):
                    r = ord(rng.read(1))
                    g = ord(rng.read(1))
                    b = ord(rng.read(1))

                    palette += [r, g, b]

                print(f'{len(palette)} colors generated...')
                continue

            # Local palette
            if subctype == 0x03:
                # Compute the number of bytes necessary
                needed = size*3-len(palette) % 3

                # Read remaining extra bytes
                rawcmap = rng.read(needed)
                cmap = []

                # Split them evenly assuming 8 bit values
                for i in range(needed//3):
                    x = ord(rawcmap[(i*3+0)])*(1<<16)+(ord(rawcmap[(i*3+1)]))
                    y = ord(rawcmap[(i*3+2)])*(1<<8)
                    z = x | y
                    cmap.append(((z >> 16)&0xff,
                                 (z >>  8)&0xff,
                                  (z      )&0xff ))


                # Does it work? Let's try...
                assert sum([sum(clr) for clr in cmap]) < MAX_COLORMAP_SIZE, \
                       'More colors requested then supported by this decoder.'
                palette += [(clr[R], clr[G], clr[B])
                            for clr in cmap ]
                print(f'{len(palette)} colors generated...')

        # handle application extension
        elif ctype == 0x21:
            appsize = ord(rng.read(1))
            appsub = rng.read(appsize).decode('ascii')
            if appsub == 'NETSCAPE':
                # First data is what type of animation follows
                animtype = ord(rng.read(1))
                # Second one tells how long it will see the animation repeated
                numrepeats = ord(rng.read(1))

                out.write(animtype.to_bytes(1, byteorder='big'))
                out.write(numrepeats.to_bytes(1, byteorder='big'))
                out.write(b'\x00\x00')

