#!/usr/bin/env python3

import struct
import sys
from typing import Tuple


# Convenience function to return an integer between low and high (inclusive).
def randint(low: int, high: int):
    return struct.unpack('<I', rng.read(4))[0] % (high+1-low)+low



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Read in the reserved field ('GIF89a'), and then the logical screen
    # descriptor. We're only really interested in the width and height
    # which occupy the next two bytes after the initial GIF identifier.
    # These will be used as the dimensions of our generated frames.
    # The rest can be saved from the original and left at default settings.
    header = b'GIF89a'
    image_width, image_height = struct.unpack('<HH', rng.read(4))

    # Skip the remaining parts of the Logical Screen Descriptor since
    # there's nothing interesting about it anyway
    rng.seek(7)

    # Now grab the global color table info, which consists of whether or
    # not the following data is present and its size (one byte)
    hascolortable, ctsize = struct.unpack('<BB', rng.read(2))


    # If the global color table is empty, just set it to a single black entry.
    numcolors = 1<<(ctsize+1);
    if ctsize == 0:
        ctsize = 1
        palents = struct.pack('BBB', *(0,)*3)
    elif ctsize <= 7:
        # For small tables, we need to make sure that every possible value
        # gets represented inside the map. This means that the global table
        # must contain all entries from 0 until h^exp where h is the maximum
        # possible value for a given expansion factor exp. So, what follows here
        # is an expansion table that maps the current value to one of these indices.
        # Note that some items may end up unused because they'll already have been filled
        # in by previous ones.
        exps = [min(i, ctsize) for i in reversed(range(numcolors))]
        reverse_map = []
        currval = 0
        while True:
            # Find the first non-filled slot
            try:
                idx = exps.index(currval)
            except ValueError:
                break
            
            currval += 1
            if currval >= numcolors:
                raise Exception("Random color generation failed")
            
            exps[idx] = max(exps[idx]-1, 0)
            
            # Fill slots in order until we run out of colors
            while True:

                # Expand the current value as much as necessary
                while exps[idx] < ctsize:
                    reverse_map.append(currval)

                    # Go to the next expansion level
                    currval >>= 1
                    exps[idx] += 1
                
                # Check if we need to move on to the next index
                if currval == numcolors-1:
                    break
                else:
                    currval += 1
                    try:
                        idx = exps.index(currval)
                    except ValueError:
                        continue
                    
                    exps[idx] -= 1
        
        palents = list(b'')
        for y in xrange(min(ctsize, numcolors)):
            palents.extend([struct.pack('BBB',*y) for _ in xrange(reverse_map[y])])

        # Pad to fill out the palette so that it's a multiple of 4 bytes long (required by spec)
        while len(palents) & 3!= 0:
            palents.append(struct.pack('BBB', *[(randint(-128,127),randin(-128,127),randint(-128,127))]))

        # Strip off any trailing padding
        palents = palents[:numcolors*3]

        palents = bytes().join(palletts)


    else:
        # If the global color table is too large, fill it randomly
        palents = [struct.pack('BBB',*(randint(-128,127),(randint(-128,127)),(randint(-128,127)))) for _ in xrange(numcolors)]
        palents = b''.join(palents)



    # TODO: we could do something with this; but not now...
    rng.seek((3 << 8) | ctsize)

    # Grab the size of the animation data block
    datalen = int.from_bytes(rng.read(2), 'big')

    # Finally, copy over the actual data stream
    frame_count = 0
    while datalen > 0:

        print("[Generating frame {:d}/{:d}]".format(frame_count+1, ((datalen-1)//image_width//image_height)+1), file=sys.stderr, flush=True)

        framedata = rng.read(image_width*image_height)

        trailer = b'\x00\x3b'

        out.write(struct.pack('<HH', image_width, image_height))
        out.write(palents)
        out.write(framedata)
        out.write(trailer)

        frame_count += 1
    
        datalen -= image_width * image_height







