#!/usr/bin/env python3

import struct
import sys
import os
import random
import argparse

# Generates a random GIF file with given size `size`,
# and palette size `palette_size` to stdout
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--size', type=str, required=True, help='Size of the random data block (e.g., "1MB").')
    parser.add_argument('--transparency', action='store_true', default=False,
                        help='Add transparency extension.')
    parser.add_argument('--loopcount', type=int, default=sys.maxsize,
                        help='How many times should this gif play?')
    args = parser.parse_args()

    if not hasattr(__builtins__, 'bytes'):
        setattr(__builtins__, 'bytes', str)


    def convert_to_bytes(string):
        try:
            return float(string[:-2]) * {
                'KB': 1 << 10,
                'MB': 1 << 20,
                'GB': 1 << 30}.get(string[-2:], 1)
        except ValueError as e:
            raise Exception("Invalid argument") from e

    size = convert_to_bytes(args.size)

    print("\nGenerating {}...".format(args.size))
    out = bytearray()
    rnd = open("/dev/urandom", mode="rb")
    while len(out) < size:
        out += rnd.read((size - len(out)) % 1024)

        if args.transparency and (len(out) >= size / 2):
            Transparency = random.randint(0x10, 0x20)

            # Write transparent pixel to first two pixel locations
            out[0:Transparency] = [(random.randint(0xF0, 0xFF)), (random.randint(0xF0, 0xFF))]
            break

    header = b"GIF89a"
    datalen = len(out)

    # Image Header
    ihdr = [0] * 10
    ihdr[0] = 0x2C   # image separator

    width = random.randint(100, 800)
    height = random.randint(100, 600)
    ihdr[1], ihdr[2] = divmod(width, 256)      # Width
    ihdr[3], ihdr[4] = divmod(height, 256)     # Height
    ihdr[5] = 0b0111                      # Size of Local Color Table
    ihdr[6] = 0                          # No interlaced images used
    ihdr[7] = 0                          # No local color table
    ihdr[8] = 7                           # Sort by G at beginning of image data stream
    ihdr[9] = 0                           # Global color table is not included
    out += bytes(ihdr)

    # Graphics Control Extension
    gce = [0] * 8
    gce[0] = 0x21                       # Extension Introducer
    gce[1] = 0xF9                       # Graphic Control Label
    gce[2] = 0x04                       # Byte Length
    gce[3] = 0                         # Packed field: Reserved
    gce[4] = 0                         # Disposal Method: None
    gce[5] = 0                         # User Input Flag: False

    if args.transparency:
        gce[6] = 0xA0                     # Transparent Color Index
        gce[7] = 0                        # Delay Time: 2 seconds
    else:
        gce[6] = random.randint(
            0,
            255
        )                               # Transparent Color Index
        gce[7] = ord(os.urandom(1)[0])    # Delay Time: 2 seconds

    out += bytes(gce)

    # Application Extension
    appext = [0]*11
    appext[0] = 0x21                  # Extension introducer
    appext[1] = 0xFF                  # Application Extension label
    appext[2] = 11                    # Size of application block
    appext[3] = b'NETSCAPE2.0'       # Application Identifier
    appext[4] = 3                    # Subblock length
    appext[5] = 1                    # Application data sub-block index
    appext[6] = 0                    # Loop count
    appext[7] = 0                    # Block terminator
    out += bytes(appext)

    # Comment Extension
    comext = [
        0x21,                       # Extension Introducer
        0xFE,                       # Comment Extension Label
        0x0B,                       # Byte Length of comment
        0x47,                       # Homepage URL
        0x49,
        0x46,
        0x38,
        0x39,
        0x61,
        0x01,                       # Block terminator
        0x00                        # Padding
    ]
    out += bytes(comext)

    # Plain Text Extension
    txtext = [
        0x21,                       # Extension Introducer
        0x01,                       # Plain Text Extension
        0x0D,                       # Total byte length
        0x54,                       # Text Grid Left Position
        0x00,
        0x00,
        0x00,
        0x54,                       # Text Grid Top Position
        0x00,
        0x00,
        0x00,
        0x0A,                       # Text Grid Cell Width
        0x00,
        0x00,
        0x00,
        0x0C,                       # Text Grid Cell Height
        0x00,
        0x00,
        0x00,
        0x00,                       # Filler (Background Color Index)
        0x00,                       # Default text foreground color index
        0x0F,                       # Font Style
        0x41,                       # Font Family name length
        0x6E,                       # Font Family name
        0x67,
        0x65,
        0x6C,
        0x00,                       # Block Terminator
        0x00,                       # Filler (Block Size)
    ]
    out += bytes(txtext)

    # Logical Screen Descriptor
    lsd = [
        0x2C,                       # Image Separator
        0x00,                       # Image left position
        0x00,                       # Image top position
        0x10,                       # Width of screen
        0x0C,                       # Height of screen
        0x00,                       # Flags:
                                        #   itsColourTableSortedGCT: True
                                        #   Size of global colour table present: No
                                        #   Size of local colour table present: Yes
        0x00,                       # Background Colour Index
        0x00,                       # Pixel aspect ratio
        0x00,                       # BinData (Sub-block begin)
        0x0F,                       #   BinData Block Size
        0x41,                       #   BinData Block Type
        0x6E,                       #   Number of entries
        0x67,                       #     Entry 1
        0x65,                       #     Entry 2
        0x6C,                       #     Entry 3
        0x00,                       #   Block Terminator
        0x00,                       # End of BinData
    ]
    out += bytes(lsd)

    # Local Color Table
    colortable = [0] * 768           # Area of interest is only 256 entries
    for x in range(256):             # but we need 768 for the palette.
        colortable[x*3+0] = hex(x)[-1:] * 2
        colortable[x*3+1] = hex(x)[-1:] * 2
        colortable[x*3+2] = hex(x)[-1:] * 2
    out += bytes(colortable)

    # Image Data Block
    imdata = []
    codesize = 0x08                 # Initial Code Size for LZW compression
    tabclear = 1                    # The next clear code in use
    bitsleft = 0                    # Amount of bit space remaining in current byte
    byte = 0                        # The current byte under construction
    bufbyte = 0                     # The last read byte
    buflen = 0                      # How much data is contained in inputbuf[]
    bpos = 0                        # Current bit offset within curbyte
    codemask = (1 << codesize) - 1  # Mask used on codes
    mincodesize = codesize + 1      # Minimum codesize before resetting it back to
                                    # original value
    maxcodesize = codesize + 1      # Maximum codesize after which it can be reset
                                    # back to original value
    inputbuf = []                   # List containing uncompressed data
    outputbuf = []                  # List containing compressed data
    codetab = [-1] * 4096           # Actual lookup table
    prevcode = -1                   # Previously generated code
    curcode = 0                     # Currently generated code

    # Initialize
    inputbuf