#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Header
    out.write("GIF89a".encode())

    # Logical Screen Descriptor
    width = rng.randint(2 ** 10, 2 ** 12 - 1) // 2
    height = rng.randint(2 ** 10, 2 ** 12 - 1) // 2
    out.write((width).to_bytes(2, "little"))
    out.write((height).to_bytes(2, "little"))
    bgcolor = rng.randrange(1 << 8)
    out.write(struct.pack("<B", bgcolor))
    aspectratio = 0
    out.write(struct.pack("<B", aspectratio))

    # Global Color Table
    globalctlen = min(
        2**(rng.randrange(6)+1), 
        2**bpp)
    for i in range(globalctlen):
        r = rng.randrange(1<<8)
        g = rng.randrange(1<<8)
        b = rng.randrange(1<<8)
        out.write(r.to_bytes(1, "little"))
        out.write(g.to_bytes(1, "little"))
        out.write(b.to_bytes(1, "little"))

    # Graphic Control Extension
    out.write("!\xf9".encode())
    delaytime = rng.randrange(1 << 8)
    transindex = 0
    disposalmethod = rng.randrange(4)
    userinputreqd = False
    out.write(delaytime.to_bytes(2,"little"))
    out.write(transindex.to_bytes(1,"little"))
    out.write(disposalmethod.to_bytes(1,"little"))
    out.write(userinputreqd.to_bytes(1,"little"))
    transparentcolor = None 
    if transparentcolor is not None:
        out.write(transparentcolor.to_bytes(3,"little")+b"\x00")
    else: 
        out.write("\x00"*(3+1))
    # Image descriptor
    out.write(",".encode())
    leftpos = rng.randrange(-2**(16-1), 2**(16-1)-1)
    toppos = rng.randrange(-2**(16-1), 2**(16-1)-1)
    imagewidth = rng.randrange(2 ** 10, 2 ** 12 - 1) // 2
    imageheight = rng.randrange(2 ** 10, 2 ** 12 - 1) // 2
    packedbits = (leftpos & 0x0f).to_bytes(1, "little") + \
                 ((toppos >> 4) & 0x0f).to_bytes(1, "little") + \
                (((imagewidth // 2) - 1) & 0xf0 | ((imageheight // 2) - 1)).to_bytes(1, "little") + \
                 ((0 & 0b1111) << 4).to_bytes(1, "little") + \
                 8.to_bytes(1, "little")
    
    out.write(packedbits)


    # LZW Compression
    lzwmincodesize = 2
    numlzwcodes = 2**lzwmincodesize
    lzwwords = []
    maxlzwword = 255    
    
    while len(lzwwords) < numlzwcodes:
        newword  = (int.from_bytes(
                    rng.read(1), 
                    byteorder="big")).to_bytes(1,'big')
        oldbyte = newword[0]
        repeat = rng.randrange(maxlzwword)              

        if rng.randint(0,1) == 1 or \
           oldbyte > maxlzwword or \
           len(lzwwords) >= numlzwcodes or \
           repeat == 0 :
            lzwwords += [(oldbyte,)]
        
        elif rng.randint(0,repeat//numlzwcodes)==0: 
            lzwwords += [(oldbyte,) + tuple(newword) * repeat]
            
        else: 
            lzwwords += [(oldbyte,)] 
            
    codetable = {b'\xff':0}
    nextcode = 1
    usedcodes = set([0])

    def writewords():
        nonlocal codetable,nextcode,usedcodes
        sortedwords = list(sorted(lzwwords, key=lambda x: len(x)))
        for word in sortedwords:
            if bytes(word[:]) in codetable: continue

            if len(word) > 1: 
                codetable[bytes(word[:])] = nextcode
                nextcode += 1

            if len(word)>2: 
                subcode = bytes(word[:-1])
                sublen = len(subcode)

                if sublen <= lzwmincodesize:
                    raise ValueError("Invalid code length")

                if codetable[subcode]:
                    bitstream = bytearray()

                    bitsneeded = 8 - sublen - 1 
                    bitstream += ((codetable[subcode]) & ~(2**bitsneeded-1)).to_bytes(1, "little")
                    bitstream[-1] |= (2**bitsneeded-1)
                    bitstream += word[-1:] 

                    out.write(bitstream)
                
                else:
                    bitstream = bytearray()
                    bitstream += ((sublen - lzwmincodesize ) | 0b10000000).to_bytes(1, "little")
                    
                    bitsneeded = lzwmincodesize + 1
                    bitstream[-1] |= (word[-1] & ~((2**(bitsneeded+1)-1)>>1) ).to_bytes(1, "little")
                    bitstream += ((word[-1]>>1) & ~(2**(bitsneeded-1)-1)).to_bytes(1, "little")
                    bitstream += word[-1].to_bytes(1, "little")

                    out.write(bitstream)
            
            elif len(word)>1:  
                out.write(((word[0]|0b10000000).to_bytes(1, "little")))
        
            if nextcode in usedcodes: 
                raise OverflowError("Exhausted codes!")

            usedcodes.add(nextcode)
            
    maxwords = numlzwcodes - (2**(lzwmincodesize+1))-1 
    
    while len(lzwwords) > maxwords:
        writewords()
        
    assert nextcode < 2**lzwmincodesize 

    writewords()

    if len(lzwwords) < numlzwcodes: 
        diff = numlzwcodes - len(lzwwords)
        lzwwords += [b'\xff']*diff
        writewords()
        
    lzwwords = dict(zip(lzwwords,[None]*len(lzwwords)))
    lzwwords[(b'\x00',)] = b''
    lzwwords[(b'')]=b''
    lzwwords[tuple([])]=b''
    
    while True:
        newword = (int.from_bytes(
                    rng.read(1),
                    byteorder="big")).to_bytes(1,'big')[0]
        oldword = tuple(lzwwords[tuple(lzwwords.keys())[
                            rng.randrange(len(lzwwords))]])
        oldbyte = oldword[0]
        repeat = rng.randrange(maxlzwword)

        if rng.randint(0,1) == 1 or \
           oldbyte > maxlzwword or \
           len(lzwwords) >= numlzwcodes or \
           repeat == 0 :
            lzwwords[(oldbyte,)]

        elif rng.randint(0,repeat//numlzwcodes)==0: 
            lzwwords[(oldbyte,)*(repeat+1)] = (oldbyte,) + newword*repeat

        else: 
            lzwwords[(oldbyte,)]

        if len(lzwwords) > maxwords:  
            writewords()

        if len(lzwwords)>=len(set(lzwwords.values())): 
            writewords()

        if len(lzwwords) == numlzwcodes: 
            break 
        
    lzwwords = list(sorted(lzwwords.items(),key=lambda x: len(x[0])))

    lzwtable = {}
    for idx,(key,value) in enumerate(lzwwords):
        lzwtable[key] = idx

    def writelzw(source:BinaryIO)->None:
        nonlocal lzwtable
        bitstream = bytearray()
        readbytes = source.read(4)
        while readbytes!= b"":
            code = lzwtable[tuple(readbytes)]
            bitsneeded = lzwmincodesize  

            bitstream += ((code) & ~(2**bitsneeded-1)).to_bytes(1, "little")
            bitstream[-1] |= (2**bitsneeded-1)

            lastbytes = readbytes[-1:]