#!/usr/bin/env python3

import struct
from typing import BinaryIO


def _generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO
    ) -> None:
    header = b"GIF89a"
    datalen = int.from_bytes(rng.read(2), "little") & 0xffff

    # Write the Image Header Frame
    ihdr = bytearray([
        0x2C,                       # Image Separator
        *(int.to_bytes(
            rng.read(4), "little", signed=False)),   # Image Left Position
        0x00, 0x00,                  # Image Top Position
        0x00, 0x00,                  # Image Width
        (((int.from_bytes(
            rng.read(2), "big")) >> 4) << 4) | (
            1 if (((int.from_bytes(
                rng.read(2), "big")) >> 4) == 0) else (
                2 if (((int.from_bytes(
                    rng.read(2), "big")) >> 4) == 1) else 0)),    # Image Height
        *((int.to_bytes(((int.from_bytes(
            rng.read(2), "big")) >> 4) << 4), "big")),        # Local Color Table Size
        0x00])                       # No Interlace / Sort Flag

    out.write(header + ihdr)

    # LZW compressed data block
    for i in range((datalen - len(ihdr)) - 1):
        out.write(bytes([(int.from_bytes(rng.read(1), "little"))]))

    # Add on final frame terminator to end off the animation
    out.write(b";")

def make_color_table(rng: BinaryIO, size: int) -> bytes:
    """Generates a color table of `size`, which should be in the
    range 0 <= n < 256"""
    return bytes([0xFE, *[
        int((((rng.read(1)[0] >> 4) << 4) | 0xF) % 255),
        *[int((((rng.read(1)[0] >> 4) << 4) | 0xF) % 255)]*3]*(size-1)])

def generate_random_animated_gif(
        rng: BinaryIO, filename: str, colors: int=1<<2, frames: int=1<<11
    ) -> None:
    with open(filename, "wb") as f_out:

        # Write out global header information
        f_out.write(b"\x47\x49\x46\x38\x39\x61\x01\x2C\x00\x00\x00\x00\x00\x00" + \
            make_color_table(rng, colors))

        # Write out each frame
        for i in range(frames):

            # If this is not the first iteration write the frame seperator
            if i!= 0:
                f_out.write(b"\x2C\x00\x00\x00\x00\x00\x00")
            
            # Generate an animated GIF frame and write it out
            _generate_random_gif(rng, f_out)