#!/usr/bin/env python3

import struct
from typing import Iterable, Optional

# Generates a random GIF file into `out` using the
# random number generator `rng` (with a given seed). If
# seed is None, it will use /dev/urandom.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int, colorspace: int=0,
        palette: Optional[Iterable]=[
                [0x00, 0x00, 0x00], # black
                [0xff, 0xfe, 0xfd], # white
                [0xff, 0xfc, 0xf6], # pink
                ],
        numframes: int=12,
        delay: int=5
    ) -> None:

    if colorspace not in {0, 1}:
        raise ValueError('colorspace must be either 0 or 1!')

    # Header
    header = b'GIF89a'
    global_flags = ord(b'\x80') | (ord(b'\x00') << 1) \
                   | ((numframes & 7) << 2) | (((numframes >> 3) & 1) << 3)
    flags = b''

    # Logical screen descriptor and logical screen extension block
    lsd = (width & 0xFF).to_bytes(2, 'little', signed=False) + \
          (height & 0xFF).to_bytes(2, 'little', signed=False) + \
          b'\x08\x00' + bytes([global_flags]) + \
          (b'\x00\x00' if colorspace == 0 else b'\xff\xff')
    lseb = b'\xF9\x04' + len(palette).to_bytes(1, 'little') + \
           b'\x00\x00' + struct.pack('<H', delay) + \
           b'\x00' * 8
    out.write(header)
    out.write((len(lsd)).to_bytes(3, 'big'))
    out.write(lsd)

    if len(palette):
        out.write(lseb)

        # Global color table
        for rgb in palette:
            out.write(b'\x21\xF9\x04')
            out.write(rgb)

    blocks = [] # Frames are built as an array of string literals

    def write_graphics_control():
        out.write(blocks[-1][:3])
        out.write(flags)
        out.write(struct.pack('<B', delay))
        out.write(b'\x00')

    # Loop control extension block
    loopext = b'\x21\xFF\x0b' + b'NETSCAPE2.0' + b'\x03\x01\x00\x00'

    for frameidx in range(numframes - 1):
        blocks += [
            b'\x2c', # Image separator
            struct.pack('<HHHHHB',
                0, 0, width, height, 0,
                ord(b'\x80')), # LZW
            ]
        out.write(loopext)
        write_graphics_control()

    # Trailer
    trailer = b'\x3b'
    out.write(trailer)