#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (by default /dev/urandom).
# If you want to use this with stdin, pass open('/dev/stdin', 'rb') as the rng.
def generate_random_gif(
        rng :BinaryIO=open('/dev/urandom', 'rb'), out: BinaryIO=open('./test.gif','wb')
    ) -> None:

    width = input('width? ') if not rng else int.from_bytes(rng.read(4), byteorder='big')
    height = input('height? ') if not rng else int.from_bytes(rng.read(4), byteorder='big')
    packedBitsPerPixel = input('bits per pixel? ') if not rng else ord(rng.read(1))
    hasColorTable = bool(ord(input('has color table? y/n ').strip().lower()[0]) - ord('y'))
    print("{} x {}".format(width,height))
    print("packed bits per pixel: {0}".format(hex(packedBitsPerPixel)))
    print("has color table?: {0}".format(str(hasColorTable)))

    # Header
    out.write(b'GIF89a')
    
    # Logical screen descriptor
    out.write(struct.pack('<HHBBB', width, height, packedBitsPerPixel & 7 | ((bool(hasColorTable))*768), 0, 255, 0));

    backgroundIndex = ord(input('background index of transparent color (-1 for none)? ')) if not rng else ord(rng.read(1))
    aspectRatioByte = input('aspect ratio? ') if not rng else ord(rng.read(1))
    print("background index: %d" % backgroundIndex)
    print("aspect ration: %d" % aspectRatioByte)
    out.write(struct.pack("<BB", backgroundIndex, aspectRatioByte))

    # Color Table
    if hasColorTable:
        numColorsPresent = pow(2,(packedBitsPerPixel >> 4))
        globalColorTable = []

        for i in range(numColorsPresent):
            red = ord(input("{0}th color Red intensity?\t".format(i+1)).strip()) if not rng else ord(rng.read(1))
            green = ord(input("{0}th color Green intensity?\t".format(i+1)).strip()) if not rng else ord(rng.read(1))
            blue = ord(input("{0}th color Blue intensity?\t".format(i+1)).strip()) if not rng else ord(rng.read(1))

            print("\t{0}, {1}, {2}".format(red,green,blue))
            
            globalColorTable += [red,green,blue]
        
        out.write(bytearray(globalColorTable))

    # Data Block
    # packed field: local color table, interlace, sort order
    packedField = 2 * (int(not hasColorTable) & 1) + \
                  4 * (int(hasColorTable) & 1)    + \ 
                  8 * ord(input("interlace flag ([Y]es/[N]o)? ").strip().lower()[0] == "y")\
                  + 16 * ord(input("sort flag ([Y]es/[N]o)? ").strip().lower()[0] == "y")
    
    lzwMinCodeSize = int(input("minimum code size? [-1 means 256] ")) if not rng else ord(rng.read(1))
    print("lzw min code size: " + str(lzwMinCodeSize))
    out.write(struct.pack("<B", packedField))
    if lzwMinCodeSize > 0:
        out.write(struct.pack("<B", lzwMinCodeSize))

    dataChunk = bytes()
    while True:
        # Set up LZW minimum code size (either defined here or from a prior graphic control block)
        if lzwMinCodeSize < 0: 
            lzwMinCodeSize = ord(input("next lzw min code size? ").strip().lower()[0])
        elif lzwMinCodeSize < 2 and lzwMinCodeSize!= 1:
            raise Exception("Invalid lzw minimum code size.")
        
        bytesToRead = pow(2, lzwMinCodeSize) # Number of bytes to read at once
        dataChunk += rng.read(bytesToRead)
        while len(dataChunk) >= 2**lzwMinCodeSize:
            chunk = dataChunk[:pow(2, lzwMinCodeSize)]
            del dataChunk[:pow(2, lzwMinCodeSize)]
            out.write(chunk)
            yield chunk

