#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the signature bytes
    out.write(b"GIF89a")

    # Write the logical screen descriptor
    width = struct.unpack(">H", rng.read(2))[0] % 640
    height = struct.unpack(">H", rng.read(2))[0] % 480
    global_color_table = bool(struct.unpack("B", rng.read(1))[0]) ^ True
    background_color = struct.unpack("B", rng.read(1))[0]
    aspect_ratio = 0 << 4 | 0 # no defined ratio

    out.write(width.to_bytes(2, "big"))
    out.write(height.to_bytes(2, "big"))
    out.write((global_color_table << 7).to_bytes(1, "big"))
    out.write(background_color.to_bytes(1, "big"))
    out.write(aspect_ratio.to_bytes(1))

    # Write the local color table section
    size = rng.read(1)[0] & 0x07
    for _ in range(size ** 2):
        r, g, b = tuple([c & ((1<<5)-1) for c in rng.read(3)])
        out.write(((r >> 3) << 5 | (g >> 3) << 2 | b >> 6).to_bytes(1, "big"))

    # Write the image descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = struct.unpack(">H", rng.read(2))[0] % 640
    height = struct.unpack(">H", rng.read(2))[0] % 480
    has_local_color_table = False << 7
    interlace = False << 6
    sort = False << 5
    reserved = 0b0000 << 2
    lct_bit = False << 1
    lct_size = 1 << 1

    out.write(left)
    out.write(top)
    out.write(width.to_bytes(2, "big"))
    out.write(height.to_bytes(2, "big"))
    out.write(((has_local_color_table | interlace | sort |
                reserved | lct_bit | lct_size) << 8).to_bytes(1, "big"))

    # Write the local color table section
    if lct_size > 0:
        size = rng.read(1)[0] & 0x07
        for _ in range(size ** 2):
            r, g, b = tuple([c & ((1<<5)-1) for c in rng.read(3)])
            out.write(((r >> 3) << 5 | (g >> 3) << 2 | b >> 6).to_bytes(1, "big"))

    # Write the graphics control extension
    transparent_color_index = 0b0000 << 4 | 0b0000 # no transparency
    delay = 0b0000 << 2 | 0b00 # 2-frame delay
    trans_flag = False << 0
    user_input = False << 1
    disposal_method = 0b0000 << 2
    cbg_color_index = 0b0000 << 4 | 0b0000 # black background

    out.write((0x21).to_bytes(1, "big"))
    out.write(b"\xf9")
    out.write(transparent_color_index.to_bytes(1, "big"))
    out.write(delay.to_bytes(1, "big"))
    out.write((trans_flag << 7).to_bytes(1, "big"))
    out.write((user_input << 6).to_bytes(1, "big"))
    out.write(disposal_method.to_bytes(1, "big"))
    out.write(cbg_color_index.to_bytes(1, "big"))

    # Write the plain image data
    local_color_table = False << 7
    reserve = 0b00001 << 3 | 0b0000
    lct_size = 1 << 1

    out.write((local_color_table | reserve | lct_size).to_bytes(1, "big"))
    out.write(b"\x08")
    out.write(struct.pack(">%dB"%len(rng.read())))

    # Write the trailing padding bytes
    out.write(rng.read())
