#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        version: bytes=b'89a',
        width: int=100, height: int=100,
        hascolor: bool=False, numcolors: int=256, bgindex: int=-1,
        hasglobalpalette: bool=True, globalpalsize: int=7,
        localpalsize: int=7, loop: int=-1, blocksize: int=256
    ):

    header = (b'GIF').join([version[i:i+1] for i in range(4)])
    colorbits = 1 << ((numcolors <= 2) * 2 + \
                      (hascolor or hasglobalpalette))
    globalpalsize += (~globalpalsize & 1) # must be even
    localpalsize += (~localpalsize & 1)   # must be even
    
    headerlen = 10 + globalpalsize + localpalsize;

    # some of these haven't been tested very much yet...
    flags = (
            1 |  # image separator
            2 |  # use global palette
            hascolor    # true = local color table present
        ) << 5

    # this is a little bit over-engineered right now, but I'm hoping to add a
    # few more features before releasing it
    extflags = (
                hasglobalpalette      # set iff global palette available
                 | localpalsize        << 1 # size of local palette
                 | width               << 3 # horizontal canvas size
                 | height              << 5 # vertical canvas size
                 | colorbits           << 7 # bits per primary color
                 | bgindex             << 8 # background index
                 | globalpalsize       << 9 # length of global palette
                 | 0                    << 10 # pixel aspect ratio
                 | hascolor            << 11 # set iff local colors are used
                 | 0                    << 12 # user input flag
                 | loop!= -1          << 14 # set iff looping is desired
                 | False                << 15 # set iff interlace mode used
           )

    # write header
    out.write(header + b'\x00')
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    out.write(((blocksize - headerlen) // 8).to_bytes(1, 'little'))
    out.write((flags).to_bytes(1, 'little'))

    # global color table (if needed)
    if hasglobalpalette:
        def palentry():
            return (
                    (int.from_bytes(rng.read(1), 'little') >> 7) &
                    1,                   # gcte
                    (int.from_bytes(rng.read(1), 'little') >> 6) & 1,
                                                      # sort flag 
                    *map(lambda x: int(x / 3),
                        map(ord, rng.read(3))) # R, G, B
                )

        globalpal = list(map(palentry, range(globalpalsize)))

        for entry in globalpal:
            out.write(entry[-1:])

        # write local color table
        if hascolor:
            for _ in range(localpalsize):
                r = int(ord(rng.read(1)) / 3)
                g = int(ord(rng.read(1)) / 3)
                b = int(ord(rng.read(1)) / 3)

                out.write(r.to_bytes(1, 'little'))
                out.write(g.to_bytes(1, 'little'))
                out.write(b.to_bytes(1, 'little'))

            # write graphic control extension
            out.write(b'\x21\xf9\x04' +
                       (loop ^ 0xffff).to_bytes(2, 'little') +
                       b'\x00')


    # masking area
    out.write(b'\x21\xfe' +
               randrange(0x000000, 0xffffff + 1).to_bytes(3, 'little') + 
               b'\x00')

    # frame duration
    out.write(b'\x21\xff\x0b' +
               int(round(randrange(-30, 30) / 10)).to_bytes(1, 'little'))

    # repetition count (if applicable)
    




    # graphics data
    padding = [b'\x00'] * blocksize

    while True:
        byte = rng.read(blocksize)
        
        if not byte and not padding:
            break

        out.write(byte[:blocksize])