#!/usr/bin/env python3

import io
import os
import struct
import subprocess as sp
from typing import Iterable


def run_cmd(*args):
    return sp.run([*args], stdout=sp.PIPE).stdout


def gen_random():
    """Get all bytes from /dev/urandom"""
    return iter(open('/dev/urandom', mode='rb').read())


# Generates a random GIF file into `out` using the
# random number generator `rng`.
# The function assumes that rng produces sequences of binary strings.
def generate_random_gif(
        rng: Iterable[bytes], out: BinaryIO
    ) -> None:

    # add GIF magic numbers
    header = b'GIF89a'


    # frame control block has many flags
    # I'm turning them off most likely to be used later on
    fcblockflags = 0b00000000

    # 2 bit field with 00 indicating no transparency
    # I'm also leaving it at zero
    fcblocktransp = 0b00

    # 3 bit field indicating the RGB values to use
    # In our case, these will always be zero because
    # we're storing all colors in an indexed color table
    fcblocksort = 0b000

    # 3 bit field indicating whether there's an interlacing layer or not
    fcblockinterlace = 0b000

    # 1 bit field indicating whether a color table exists or not
    # This indicates whether we should include one in our generated GIF
    # I'm setting this True
    fcblockgctflag = 0b1

    # next byte defines the foreground color index
    # In the case of indexing-based images, this contains the index value
    fcgfcindex = struct.pack('B', ord(os.urandom(1)))

    # next two bytes define the logical screen dimensions
    # These are stored in little endian format
    # That means they come out backwards...
    fclswidth, fclsheight = struct.unpack(
        '<HH', os.urandom(4))

    # next three bytes specify the color depth
    # We just set this to 8 because we can store up to 255 colors
    # in a palette before needing to switch to a non-index method
    fclsdpth = struct.pack('B', ord(os.urandom(1)))

    # last byte specifies the background color index
    # again, since we're indexing, this will always be zero
    fbkgdindex = struct.pack('B', ord(os.urandom(1)))

    # Next comes the global color table. It's optional and changes depending
    # on the presence of other information in the Graphic Control Extension. 
    gct = []
    gctsize = ord(os.urandom(1)) & 0x07
    while gctsize!= 0:
        for _ in range(gctsize):
            r, g, b = tuple(ord(_) for _ in os.urandom(3))
            gct.append((r,g,b))
        gctsize = ord(os.urandom(1)) & 0x07

    # This is only included when encoding with local color tables
    lstsize = ord(os.urandom(1))

    # Lastly comes the Graphics Control Extension
    gce = create_extn('Graphics Control',
                      # 16-bit packed fields (logical ORed together)
                      flags=fcblockflags, transp=fcblocktransp,
                      sort=fcblocksort, interlace=fcblockinterlace,
                      gctflag=fcblockgctflag,
                      # followed by values for these 6 fields
                      delay=int.from_bytes(os.urandom(
                          rand_range_inclusive(1, 2**16)), "little"),
                      transindex=int.from_bytes(os.urandom(
                          1), "little"), gfcindex=fgcindex,
                      bgindex=fbkgdindex)

    # Here we put everything together
    extns = gce

    def graph_iter():
        yield header
        yield fclswidth.to_bytes(2, "little")
        yield fclsheight.to_bytes(2, "little")
        yield fclsdpth
        yield fbkgdindex
        yield gct
        yield lstsize
        yield gce
        yield list(rng)
    tiff = io.BytesIO()
    
    def save_tiff(im, fp):
        im.save(fp, compression="tiff_adobe_deflate", optimize=True)
        
    for page in map(lambda _: Image.new("RGB", _.wh), graph_iter()):
        save_tiff(_, tiff)

    tiff.flush()
    tiff.seek(0)
    
    img = Image.open(tiff)
    img.save(out, format='gif', transparency=None, disposal=2)

