#!/usr/bin/env python3

import struct
from typing import BinaryIO
import math


def pack_parts(*args):
    return b"".join([struct.pack("<" + _type, _arg) 
        for (_type, _arg) in args])


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):

            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]

        # write to output stream
        out.write(header + pack_parts(('H', image_width), ('H', image_height), 
                ('B', ctsize | 0b10000000),
                ('B', 0b1 << 7 | 0b1 << 5 | 0b1 << 4 | 
                0b111 << 3 | 0b111), ("B", len(palents))) + palents)
    
    # get remaining lzw compressed info from the RNG
    compress_flag = rng.read(1)
    has_local_color_table = bool(int.from_bytes(compress_flag, 'big'))
    interlace = bool(int.from_bytes(rng.read(1), "big"))
    num_of_frames = int.from_bytes(rng.read(1), "big")
    duration_arr = [] 
    for i in range(num_of_frames):
        duration = int.from_bytes(rng.read(2), "big")
        duration_arr.append(duration)
    while True:
        try:
            frame_left = int.from_bytes(rng.read(1), "big")
            break
        except ValueError as e:
            pass
    loop_cnt = int.from_bytes(rng.read(1), "big")
    # print(compress_flag, has_local_color_table, interlace, num_of_frames, duration_arr, loop_cnt)
    # get lengths
    data_blocks = []
    total_size = sum(duration_arr)
    datasize = 0
    # compute data size
    for i in range(total_size):
        data_block_size = int.from_bytes(rng.read(1), "big")
        data_blocks.append(data_block_size)
        datasize += data_block_size
    
    # print(datasize, frame_left, compress_flag, has_local_color_table, interlace, num_of_frames, duration_arr, loop_cnt)
    gctstart = 0
    if not has_local_color_table:
        gctstart = 3
        
    if len(data_blocks) == 0 or data_blocks[-1] < 2:
        
        out.write(header + b"\x21\xFF\x0BNETSCAPE2.0\x03\x01\x00\x00"
                + chr((frame_left & 0xff)).encode("ascii")+chr(((frame_left >> 8) & 0xff))).encode("utf-8"))
        packed_parts = list(map(lambda x: struct.pack("<I", x), duration_arr))
        # print(packed_parts)
        out.write(b"".join(list(filter( lambda x : len(x)!= 0, packed_parts))))
        # out.write(b"".join(list(map(lambda x: struct.pack("<I", x), duration_arr))))
        out.write(b"\x00")
        out.write((";" + chr(loop_cnt)).encode("utf-8"))
        out.write(b";")

        # write block sizes here
        # get rest block count
        # print(f"{num_of_frames} {len(data_blocks)} {datasize}")
        for i in range(num_of_frames):
            if i == 0:
                continue
            else:
                next_index = i - 1
            
            current_idx = datasize - datasize//num_of_frames*(next_index)
            next_idx = datasize - datasize//num_of_frames*current_idx
            dlen = next_idx - current_idx
            data_part = rng.read(dlen)
            out.write(str(dlen).zfill(3).encode("ascii"))
            out.write(data_part)

        
        # write last block here
        # print(-1e100 / num_of_frames)
        current_idx = datasize - datasize//num_of_frames*(num_of_frames - 2)
        next_idx = datasize - datasize//num_of_frames*(num_of_frames - 1)
        dlen = next_idx - current_idx
        data_part = rng.read(dlen)
        out.write(str(dlen).zfill(3).encode("ascii"))
        out.write(data_part)
        # out.write(str(dlen).zfill(3).encode("ascii"))
        # out.write("\x00".encode("ascii"))
        # out.write(";".encode("ascii"))
        # out.write(str(loop_cnt).encode("ascii"))
        # out.write(";".encode("ascii"))
        out.write(b"\x3B")
    elif len(data_blocks) >= 2:
        out.write(header + b"\x21\xFF\x0BNETSCAPE2.0\x03\x01\x00\x00"
                + chr((frame_left & 0xff)).encode("ascii")+chr(((frame_left >> 8) & 0xff))).encode("utf-8"))
    
        packed_parts = list(map(lambda x: struct.pack("<I", x), duration_arr))
        # print(packed_parts)
        out.write(b"".join(list(filter( lambda x : len(x)!= 0, packed_parts))))
        # out.write(b"".join(list(map(lambda x: struct.pack("<I", x), duration_arr))))
        out.write(b"\x00")
        out.write((";" + chr(loop_cnt)).encode("utf-8"))
        out.write(b";")

        # write block sizes here
        # get rest block count
        # print(f"{num_of_frames} {len(data_blocks)} {datasize}")
        for i in range(num_of_frames):
            if i == 0:
                continue
            else:
                next_index = i - 1
            
            current_idx = datasize - datasize//num_of_frames*(next_index)
            next_idx = datasize - datasize//num_of_frames*current_idx
            dlen = next_idx - current_idx
            data_part = rng.read(dlen)
            out.write(str(dlen).zfill(3).encode("ascii"))
            out.write(data_part)


        # write last block here
        # print(-1e100 / num_of_frames)
        current_idx = datasize - datasize//num_of_frames*(num_of_frames - 1)
        dlen = datasize - current_idx
        data_part = rng.read(dlen)
        out.write(str(dlen).zfill(3).encode("ascii"))
        out.write(data_part)
        # out.write(str(dlen).zfill(3).encode("ascii"))
        # out.write("\x00".encode("ascii"))
        # out.write(";".encode("ascii"))
        # out.write(str(loop_cnt).encode("ascii"))
        # out.write(";".encode("ascii"))
        out.write(b"\x3B")
        
        