#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Generate a random GIF."""


import io as StringIO
import struct
from typing import BinaryIO

from fvcore.common.file_io import PathManager

from.utils import open_like


def generate_random_gif(
        rng :BinaryIO, 
        path :str
    ):

    with open_like(rng, "rb") as fptr:

        # read the first three bytes
        version = fptr.read(3).decode("ascii")
        print(version)

        # get the logical screen descriptor
        width, height, flags = struct.unpack("<HHB", fptr.read(6))

        # calculate global color table size
        gctlen = ((flags >> 7) & 1) << ((flags & 7) + 1)

        # skip the background index field
        fptr.seek(1, 1)

        # skip extensions blocks field if any
        while True:
            blocksize = fptr.read(1)[0]

            # EOF reached without finding an image descriptor
            if not blocksize:
                raise RuntimeError("Could not find ImageDescriptor block.")

            blocktype = fptr.read(1)[0]
            
            # 255 indicates a 'graphic control extension'. 
            # Skip to next block.
            if blocktype == 0xff: continue

            elif blocktype!= 0x2c: 
                raise RuntimeError(f"Invalid block type {blocktype}. Expected 0x2c or 0xff.".format())

            else: break

        
        # read the local color table for each frame
        lctsize = [fptr.read(3)[-1] for _ in range((flags & 7) + 1)]
        
        # calculate the palette size based on the sizes calculated above
        palsize = sum([l*g for l,g in zip(lctsize, gctlen)])

        # calculate number of colors used from global color table
        numcolors = min([(2**i)*(1<<j) - 1 for i in range(1, 9) for j in range(7, 9)], key=(lambda x: abs(x - palsize)))
        
        # read the local color tables
        lctdata = {}
        if len(lctsize):
            for n, s in enumerate(lctsize):
                lctdata[n+1] = fptr.read(s*3)

     
        # read all frames
        frames = []
        while True:
            # read the frame's left position
            left = fptr.read(2)
            if len(left) < 2: 
                break

            # read the frame's top position
            top = fptr.read(2)
            if len(top) < 2: 
                break

            # read the frame's width and height
            w, h = struct.unpack("<HH", fptr.read(4))

            # read the packed fields
            flags = struct.unpack("B", fptr.read(1))[0]

            # calculate local color table length and create list
            flag_table = [(flags >> i) & 1 for i in reversed(range(7))]
            lctsize = [flag * (n+1) for n, flag in enumerate(flag_table[:3])]

            # check that we have enough bits to store our values
            for val, size in zip([w,h], [12,4]):
                assert val <= 2**(size-1)
                
            # unpack the field bit-by-bit
            interlace = bool(flags & 0b00000100)
            sort      = bool(flags & 0b00001000)
            reserved  = bool(flags & 0b00010000)
            lctused   = [val-1 for val, flag in zip(lctsize, flag_table[3:]) if flag]
            transparent = (flags>>4)&1
            userinput = (flags>>5)&1

            if 0:#print:
                print("%dx%d at %dx%d" % (w, h, ord(left), ord(top)), end="")

                if interlace:
                    print(", interlaced", end="")
                
                if lctused:
                    if len(lctused) > 1:
                        print(", using LCT[%d,%d]"%(lctused[-2], lctused[-1]), end="")
                        
                    else: 
                        print(", using LCT[%d]"%lctused[-1], end="")
                    
                    try:
                        lctdata[(lctused[-2], lctused[-1])][:]
                    except KeyError:
                        print("LCT not found!")
                    
                if sort: print(", sorted", end="")
                if reserved: print(", reserved", end="")
                print()


            # read the local color table for the current frame
            lctdat = None
            if len(lctsize): lctdat = [fptr.read(length) for length in lctsize]
        
            '''
            Read the optional graphic control extension if present
            The content of this block can be ignored I think
            '''
            if blocktype == 0x21:
                if ord(fptr.read(1)[0])!= 0xf9: 
                    raise Exception("Unknown block type {}".format(ord(fptr.read(1)[0])))
                
            
                tmp = fptr.read(1)
                while not tmp==b'\x00': tmp += fptr.read(1)
                for _ in range(4): fptr.read(1)

                del tmp
            
    
            '''
            For each text chunk there are two parts. First is 'text_block_header'.
            Then comes the actual data which may contain null characters.
            '''
            text_chunks = []
            while True:
                text_block_header = fptr.read(1)
                if ord(text_block_header[0]) == 0x01:

                    # add the first part to the array
                    tmp = ""
                    ch = fptr.read(1)[0]
                    while ch:
                        tmp += chr(ch)
                        ch = fptr.read(1)[0]
                    
                    text_chunks.append({"method": ord(text_block_header[0]), "text":tmp})
                    
                    # check whether this chunk is finished by checking the last byte
                    if fptr.read(1)==b'\x00':break
                else: 
                    text_chunks.append({"method": ord(text_block_header[0]), "text":""})
            
            # the second part of the text chunks follows after the first one ends
            # so we move the pointer by its length here
            if len(text_chunks)>1:
                fptr.seek(sum([len(t["text"]) for t in text_chunks[:-1]]), 1)
            
            '''
            Finally read the image data itself
            '''
            imginfo = {"width": w, 
                      "height": h,
                      "lct": lctdat}
            
            frames.append(imginfo)
            
        '''
        Since we're interested only in randomly generated images it doesn't really matter what
        gets written into these tags, but if you want to copy other samples make sure your gif writer
        supports them properly otherwise the tags will cause trouble when viewing the gif on some browsers
        '''
        appname = "<NAME>"
        comment = "https://www.facebook.com/avikde"
    
        '''
        Generate our gif
        '''
        with open(path,"wb") as fp:
            # Header
            fp.write('GIF89a'.encode('ascii'))

            # Logical Screen Descriptor
            fp.write(struct.pack('<H', width))         # Width
            fp.write(struct.pack('<H', height))        # Height
            fp.write('\x87\x00')                       # Flags: Global Color Table Flag set
            fp.write('\x00')                           # Background Color Index
            fp.write('\x00')                           # Pixel Aspect Ratio
            fp.write('\xe0\x00')                       # GCT Size: Externally Defined
            fp.write(struct.pack('<H',numcolors))       # Number of colors in global color table
            fp.write("\x00"*3*(2**(((flags&7)+1)-1)))  # Write empty color table   
            
            # Application Extension
            fp.write("\x21\xFF\x0B".encode('ascii'))     # application extension introducer
            fp.write(appname.upper().encode('ascii'))   # application identifier
            fp.write(b"\x03\x01")                      # application authentication code
            fp.write(comment.encode('ascii'))          # comment string
            fp.write("\x00")                           # terminating zero byte

            

            # Graphics Control Extension
            fp.write("\x21\xF9\x04".encode('ascii'))     # graphics control extension introducer
            fp.write("\x08")                           # label size
            fp.write(struct.pack("!BBHB", 0, 0, 0, 0))  # bitmap properties
            fp.write('\x00')                           # terminator
            
            # Plain Text Extension
            fp.write("\x21\x01\x0F".encode('ascii'))   # plain