#!/usr/bin/env python3

import struct
from typing import Iterable, Iterator

class ColorTableIterator(Iterable):

    def __init__(self, count=256):
        self._count = count

    def __iter__(self) -> Iterator[int]:
        return iter([(c, c, c) for c in range(self._count)])


# Generates a random GIF file into `out` using the
# random number generator `rng`
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width=200, height=200, ctsize=None
    ):

    header = b'GIF89a'

    # logical screen descriptor
    lsd = struct.pack('<HH', width & 0xffff, height & 0xffff)
    lsd += b'\xf0\x0f'

    out.write(header); out.write(lsd)


    if ctsize == None:
        ctsize = max(2, rng.randrange(256))

    # color table
    colortable = bytes()
    for entry in ColorTableIterator(ctsize):
        n = rng.randrange(256)
        colortable += struct.pack("BBB", *(entry))
    colormap = chr(ctsize).encode('ascii')
    colormap += struct.pack("<H", len(colortable)).decode('ascii').encode('ascii')
    out.write(colormap)
    out.write(colortable)


    # graphics control extension block
    gce = b'\x21\xf9\x04'
    gce += struct.pack(">H", ((rng.randrange(10<<8)|8)<<8|0))
    gce += b"\x00"
    out.write(gce)


    pb = rng.randrange(0xFFFF)
    pf = pb + rng.randrange(pb+2)

    raster = bytes()
    for row in range(height):
        raster += struct.pack(">H", 1)    # Raster operation to use
        raster += struct.pack(">H", pf & 0xFFF)
        raster += bytes([rng.choice([0b00, 0b01, 0b11])])     # LZW min code size (0-8 inclusive)


        line = bytes()
        st = pb&0xff
        while len(line) < (width//st)*st:      # fill current line with pixels
            line += struct.pack('>BBBB',
                rng.randrange(0x100),
                rng.randrange(0x100),
                rng.randrange(0x100),
                0x00)
        raster += line[:-(width%st)-st]       # don't include partial pixel at right edge
        raster += bytes([rng.choice([0b00, 0b01, 0b11]), 0b00])   # LZW end of stream (sort bit set)
        pass
    out.write(raster)
