#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file using as an input some bytes from /dev/urandom and write them to the given output stream
def generate_random_gif(instream: BinaryIO, outstream: BinaryIO):
    def _generate_size():
        return int.from_bytes(outstream.read(2), byteorder='little') & 0xFFFF

    # Write Header
    outstream.write(b'GIF89a')

    # Write Logical Screen Descriptor
    width = _generate_size() % 640 + 1
    height = (_generate_size() // 256 ** 3 - width) % 480 + 1
    packed_bits = [(width >> 8) & 0xFF,
                   (width & 0xFF),
                   (height >> 8) & 0xFF,
                   (height & 0xFF)]
    bits = [_generate_size(),
            *_generate_size().to_bytes(length=3, byteorder='big'),
            *_generate_size().to_bytes(length=3, byteorder='big')]

    packed_bits += list(((bits[i] >> 8) & 0xFF)
                        for i in range(len(packed_bits)))
    packed_bits += [(bits[-1] & 0xFF)]
    packed_bits += list((bits[i] & 0xFF)
                        for i in range(len(packed_bits)))
    packed_bits += [0, 0, 0, 0]

    outstream.write(bytes([*packed_bits]))
    print(f"width: {width}, height: {height}")

    # Write Global Color Table
    global_color_table_size = min(_generate_size(), 256)
    global_color_table = []
    if global_color_table_size!= 0:
        for index in range(global_color_table_size):
            rgb = [_generate_size() for i in range(3)]

            flags = [('R' if (rgb[0] & 0x80) else '_'),
                     ('G' if (rgb[1] & 0x80) else '_'),
                     ('B' if (rgb[2] & 0x80) else '_')]

            rgb[:] = [(rgb[index] & (~0x80)) << shift
                      for shift, flag in enumerate(flags) if flag == "_"]
            rgb[:] = [(value >> 1 & ~0x80) | value >> 7
                      for value in rgb[:]]

            rgb[:] = [min(max(value, 0), 255) for value in rgb[:-1]] + \
                [0]

            outstream.write(bytes(rgb))

        outstream.write(b'\x00\x00')

    while instream.peek():
        data_block = instream.read(1)[0]
        if data_block <= 0x7F:
            # Raw block
            outstream.write(data_block.to_bytes(length=1,
                                                 byteorder="big"))
        elif data_block >= 0xF9:
            break
        elif data_block == 0x2C or data_block == 0x21:
            # Extension block or Control extension block
            outstream.write(data_block.to_bytes(length=1,
                                                 byteorder="big"))
            size = _generate_size()
            outstream.write(size.to_bytes(length=2, byteorder="big"))
            for byte in range(size):
                value = instream.read(1)[0]

                outstream.write(value.to_bytes(
                    length=1, byteorder="big")
                )
        else:
            # Command extension type
            outstream.write(data_block.to_bytes(length=1,
                                                 byteorder="big"))
            size = _generate_size()
            outstream.write(size.to_bytes(length=2, byteorder="big"))
            for byte in range(size):
                value = _generate_size()

                outstream.write(value.to_bytes(
                    length=1, byteorder="big")
                )

    # Write Trailer
    outstream.write(b"\x3B")