#!/usr/bin/env python3

import sys
import os
import time
import random
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (`<os.urandom>` by default).
# If the argument `--color-table` is present on the command line
# then it will be used as an upper bound for choosing the size
# of the color table; otherwise, it is possible to have no color table.
def generate_random_gif(
        rng=None, out=sys.stdout, ctmax=128
    ) -> None:

    if rng == None:
        rng = os.urandom

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width = random.randrange(0, 65536)
    image_height = random.randrange(0, 65536)

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rl = lambda n: rng.read(n)[::-1]

    # size of the color table
    ctsize = max([0, min((int(rl(1)), ctmax))])


    palents = b''
    if ctsize > 0:

        # generate a list of RGB tuples with entries between 0 and 255
        colors = []
        while True:

            r = random.randint(0, 255)
            g = random.randint(0, 255)
            b = random.randint(0, 255)
            
            if not any([(r,g,b) == c[:3] for c in colors]):
                colors.append((r,g,b))

                if len(colors) == ctsize:
                    break

        # build up the packed palette data (packed inside a sub block)
        for r,g,b in colors:
            palents += struct.pack('BBB', b, g, r)

        # Write out the palette information
        out.write(header + struct.pack('<HHBB', image_width, 
            image_height, ctsize | 0b10000000,
            0b1 << 7 | 0b1 << 5 | 0b1 << 4 |  
            0b111 << 3 | 0b111) + palents)
        
    else:
        
        # No palette
        out.write(header + struct.pack('<HHBB', image_width, 
            image_height, 0, 
            0b1 << 7 | 0b1 << 5 | 0b1 << 4 |  
            0b111 << 3 | 0b111))

    # Build up the LZW encoded image data 
    # (the actual bytes we want to compress!)
    #firstbyte = out.tell() & 0xff
    #print("first byte:", firstbyte)

    lastchar = None
    data = b''
    while True:

        nextchar = rng.read(1)

        if nextchar!= None and len(nextchar):
            nextchar = ord(nextchar[0])
        else:
            return


        if nextchar in {255}:
            data += bytes((ctrlcodes[nextchar],))
            continue
                    
        if nextchar <= lastchar or lastchar == None:
            data += bytes((lastchar or 0,))
        elif lastchar <= ctrlcodes['clear']:
            data += bytes((ctrlcodes['clear'], lastchar,))
        else:
            data += bytes((ctrlcodes['clear'], 
                ctrlcodes['clear'], lastchar,),)
        
        if nextchar < lastchar or lastchar == None:
            pass
        elif lastchar <= ctrlcodes['clear'] and nextchar >= ctrlcodes['clear']:        
            data += bytes((nextchar,))
        elif lastchar <= ctrlcodes['clear']:
            data += bytes((ctrlcodes['clear'], nextchar,))
        else:
            data += bytes((ctrlcodes['clear'], 
                ctrlcodes['clear'], nextchar,),)

        lastchar = nextchar
        

        

    dlen = len(data)
    print("Data Length", dlen, "of type", type(dlen))
    print("Last char was", lastchar)
    print("Writing to offset", 
        out.tell(), "with seek at", out.seek(-1, io.SEEK_CUR))
    
    
    out.write(data)
    