import struct


# The maximum size of an uncompressed image in bytes (256x256).
MAX_IMAGE_SIZE = MAX_DIMENSION * MAX_DIMENSION * 4 // 100


def write_header(width: int, height: int):
    return b'' \
            + b'GIF87a' \
            + width.to_bytes(2, 'little') \
            + height.to_bytes(2, 'little')


def write_color_table():
    # TODO
    pass


def write_image_descriptor(left: int, top: int, width: int, height: int):
    return struct.pack('<BHHHHHH',
            0b10000000 | 0b11100,   # Local Color Table Flag and Size
            left, top,                # Left, Top positions
            width, height,            # Width, Height
            0b10000                   # LZW Compression
        )


def write_graphic_control_extension():
    # TODO
    pass


def write_application_extension(identifier: str):
    if identifier == 'NETSCAPE':
        subblock = b'\X03\X01\X00\X00\XFF\X00'
    else:
        raise ValueError("Application Extension not supported")

    return b'' \
            + b'\X21'       # Application Extension Label
            + b'\XFF'       # Application Block Length
            + identifier    # Identifier
            + subblock      # Subblock


def write_graphics_control_label():
    return '\XFC'


def write_comment_extension(comment: str):
    return '' \
            + '\XFE'           # Comment Extension Label
            + comment          # Comment String


def create_graphics_control_extension(duration: float):
    delay = round(duration * 100 / 10.0)
    packed = b'%c%c%c%c' % (
        0x04,              # Flags
        delay >> 16,       # Delay Hi
        delay >> 8,        # Delay Lo
        0,                 # Transparent Index
    )
    
    return b'' \
           + b'\X21'         # Graphic Control Extension Label
           + b'\XF9'         # Graphics Control Block Length
           + packed          # Packed Flags
           + b'\X00'         # Block Terminator


def copy_framedata(instream: BinaryIO, outstream: BinaryIO):
    while True:
        blocksize = instream.read(1)[0]

        if blocksize == 0x00 or blocksize == 0xff:
            break

        outstream.write(instream.read(blocksize))


def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    """Generate a GIF file produced by feeding `rng`'s bytes to
    `generate_random_gif`. This will generate images of up to
    20kB in size."""
    frames = []
    framedurations = [i for i in range(4)]     # Number of frames per second
    durationindex = 0                           # Current frame duration index
    colordepth = None                           # Pixel bit depth
    left = 0                                    # Position of current frame
    top = 0                                     # Position of current frame
    width = 0                                   # Width of current frame
    height = 0                                  # Height of current frame

    maxdimension = min(MAX_DIMENSION, 255)
    totalpixels = maxdimension ** 2
    pixels = b'\0'\ * totalpixels

    while len(frames) < FRAMECOUNT:
        if colordepth is None:
            framecount = (maxdimension - top) // 4

            if framecount > PIXELS:
                framecount = PIXELS
            
            if len(frames) >= FRAMECOUNT:
                framecount = FRAMECOUNT - len(frames)

            palette = create_palette()

            encodedframes += encode_frame(
                pixelsperline, framecount, palette[top:]
            )

            frames += [
                (pixels[top:], palette), 
            ]

            top += framecount * 4
        elif width <= 1 or height <= 1:
            # Done with this frame
            framedata = b''.join([encode_pixel(p) for p in pixels])
            graphicsctl = create_graphics_control_extension(FRAMEDURATIONS[framedurations[durationindex]])
            extensions = b''.join((
                graphicsctl,
                write_comment_extension('Made with Python using urandom'),
                write_application_extension('NETSCAPE'),
                b'\X00',             # End of File Marker
            ))

            header = write_header(width, height)
            descriptor = write_image_descriptor(left, top, width, height)
            graphicsextension = write_graphics_control_extension()

            outfile.write(header + descriptor + framedata + extensions)

            # Create new frame
            print('Frame generated:', infile.tell())
            framedurations.rotate(-1)
            durationindex = (durationindex - 1) % len(framedurations)
            colordepth = None
            left = 0
            top = 0
            width = 0
            height = 0
            pixels = b'\0'*totalpixels
        elif colordepth == 1:
            pixels[left//4*height+(top+3)//4] |= 1 << ((top&3)*4 + (left&3))
            left += 4
        elif colordepth == 2:
            pixels[left//4 * height + (top//2 + 1)/2] |= (1 << (((top&1)<<2)+left&3))
            left += 2
        elif colordepth == 4:
            pixels[left//2 * height + top//2] |= (1 << (((top&1)<<1)+left&1))
            left += 1
            
        width += COLORDEPTH_TO_PIXELSPERLINE[colordepth]
        
        if left >= WIDTH:
            left -= WIDTH
            height += 1
        
    copy_framedata(infile, outfile)