#!/usr/bin/env python3

import gzip
import struct
from typing import BinaryIO

# Generates a random GIF file with up to numframes frames
# each frame being of size maxsize (up to 65500 bytes).
# The generated gif will be written in outfile (gzip-compressed by default).
def generate_random_gif(
        numframes: int, maxsize: int, outfile: str='out', compressed=True
    ) -> None:

    assert maxsize <= 65534 # 65535 - 2 length bytes
    frametypes = [b';;'] * (numframes // len(frametypes))
    for i in range((numframes % len(frametypes))):
        frametypes[i] += struct.pack('B', i)

    print("Generating {} frames".format(len(frametypes)))
    if not compressed: outfile += '.gz'
    with open(outfile, 'wb') as f:
        write_random_gif(f, frametypes, maxsize)


# Write a list of frames to an opened binary output stream
def write_random_gif(
        outstream: BinaryIO, framelist: list, framemaxsize: int,
        compressionlevel=9
    ) -> None:

    header = b"GIF89a"
    globalcolortable = bytearray([0x00, 0x00, 0x00]) * 256
    gct = len(globalcolortable)
    lzwdictaddr = 0
    lzwwindowsize = 2**12 + 2**13

    # Concatenate all the frames
    frames = []
    for frametype in framelist:
        framedata = globalcolortable + frametype

        def copychunk(subdata):
            s = subdata[:framemaxsize]
            outstream.write(struct.pack('<HH', len(s), 0))
            return subdata[framemaxsize:]
        
        while True:
            chunk = framedata[:framemaxsize]
            framedata = framedata[framemaxsize:]

            if len(chunk) < framemaxsize or len(framedata) == 0:
                compressor = LZ77Compressor()
                c = compressor.compress(chunk)
                outstream.write(c)

                if len(chunk) < framemaxsize and len(framedata) > 0:
                    # More data is available but we need to flush the compressor first
                    continue
                
                break
            else:
                if len(framedata) >= framemaxsize:
                    framedata = copychunk(framedata)
        
        frames.append(compressor.dictionary)
        pass
    
    # Convert dictionary offsets from relative addresses to absolute ones
    dictsizesum = sum([len(d) for d in frames])
    dictoffsets = [lzwdictaddr]
    for frame in frames[:-1]:
        dictoffsets.append(dictoffsets[-1]+len(frame))
    
    # Write the image descriptor block
    outstream.write(b'\x2C\x00\x00\x00\x00\x00\x00\x00\xFF\x00\x00\x00')
    outstream.write(struct.pack("<HHHBBBHB",
        0, 0,
        dictoffsets[-1],   # Left
        0,                 # Top
        min(8,gct-1),      # Width
        0,                 # Height
        0,                 # Flags
        gct                # Color table size
    ))
    
    # Write color table (if necessary)
    if gct > 0:
        outstream.write(bytes(globalcolortable))
    
    # Write graphics control extension
    outstream.write(b'\x21\xF9\x04\x00\x0A\x00')
    
    # Write application extension
    outstream.write("\x21\xff\x0bNETSCAPE2.0\x03\x01")
    outstream.write(struct.pack(">H", dictoffsets[-1]))
    outstream.write(b"\x00")
    
    # Write plain subblocks
    for fo in dictoffsets:
        outstream.write(b'\x2c')
        outstream.write(struct.pack('<HHH',
            0, fo+1, fo
        ))
    
    # Write last subblock (end of animation)
    outstream.write(b'\x3B')


    if compressionlevel!= 0:
        # Compress the whole thing before returning
        outstream = gzip.open(outstream, "wb", compresslevel=compressionlevel)
        outstream.write(header)
        for i in range(len(frames)):
            outstream.write(struct.pack('<HH', 1, 0))
            outstream.write(globalcolortable)
            outstream.write(framelist[i])
        outstream.write(b'\x3B')




class LZWDictEntry():
    def __init__(self, code: int):
        self.code = code
        self.children = []
        self.parent = None
        self.value = ""


class LZWCompressor():
    def __init__(self):
        # Special codes
        self.clearcode = 256
        self.endofinfo = 257
        self.prevcode = 0

        self.dictionary = {None}
        self.currententry = LZWDictEntry(-1)

    def _find_node(self, parent: LZWDictEntry, nodeindex: int)->LZWDictEntry:
        cur = parent.children[nodeindex]
        if isinstance(cur, LZWDictEntry):
            return cur
        elif isinstance(cur, set):
            return self._find_child(parent, nodeindex)
        else:
            raise RuntimeError("Data structure error!")
        
    def _find_child(self, parent: LZWDictEntry, childname: bytes):
        for n in parent.children:
            if n.value == childname.decode('utf-8'):
                return n
        raise KeyError("Node {} does not have child {}".format(
            parent.value, childname.decode('utf-8')))
            
    def compress(self, data: bytes)->bytearray:
        ret = bytearray()
        prevdata = None
        nextdata = data[:2]

        data = data[2:]

        while True:
            try:
                entry = self._find_child(self.currententry, nextdata)
            except KeyError:
                # New character encountered
                newentry = LZWDictEntry(self.currententry.code+1)
                self.currententry.children.append({nextdata})
                self.currententry = newentry
                self.currententry.value = (
                        (self.currententry.parent.value + nextdata.decode('utf-8'))
                        if self.currententry.parent!= None
                            else nextdata.decode('utf-8')
                    )

                # Add to dictionary and reset pointers
                self.dictionary |= {newentry, }
                self.prevcode = newentry.code
                
                if len(ret) > 1 and ret[-2] == 1:
                    ret.pop()
                    if ret[-1] & 0x80: ret.pop()
                else:
                    ret.extend(struct.pack('<BH',
                        1, self.currententry.code
                    ))
            
            else: 
                # Character found
                if len(data) > 0: 
                    prevdata = nextdata
                    nextdata = data[:2]
                    data = data[2:]

                    # Reset pointers
                    self.prevcode = entry.code
                    
                    if len(ret) > 1 and ret[-2] == 1:
                        ret.pop()
                        if ret[-1] & 0x80: ret.pop()
                    else:
                        ret.extend(struct.pack('<BH',
                            1, entry.code
                        ))
                    
                else:
                    if len(ret) > 1 and ret[-2] == 1:
                        ret[-1] &= ~0x80
                    ret.extend(struct.pack('<BH',
                        0, entry.code    # End of information code
                    ))
                    break
        return ret

