#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header for GIF file. Magic string "GIF" followed by version
    # and size of logical screen descriptor. The logical screen
    # is always 768x768 pixels with black background color index zero.
    header = b'GIF89a\x00\x00\x00\x00,\x00\x00\x000\xf4'
    # Logical Screen Descriptor block: 0x1f = 31 means that the
    # following field contains *bits* describing this descriptor rather
    # than raw values. We use bits here because it's much easier to write.
    lsdlen = 0b11111 << 5 | 0b11111 << 2 | 0b1
    lsd = struct.pack('>HHBBB',
        0x2c00,      # Block size in bytes
        0x0007,      # Width
        0x0007,      # Height
        0b10000001,  # Global colour table flag set and size
        0b10111000,   # Background colour index, pixel aspect ratio
        )
    # Image Descriptor block - specifies canvas dimensions and
    # transparent background index value. If there were no transparency,
    # we'd simply omit this part.
    idlen = 0b10001110
    id = struct.pack('>HHBBHB',
        0x2c00,
        0x0007,
        0x0007,
        0b10000001,
        0x0000,
        0x0000,
        )
    # Code stream - LZW compression algorithm used, each code specifier
    # is made up from one or more bytes stored in LZW dictionary. Each
    # code has its own length which is represented as one byte.
    # This encoding results in poor IMG quality but okay compression rate. 
    cs = b'\x08\xff\xff\xff\x00\xf8\xff\xff\xff\xe6\xf0\xff\xff\xff\xce\xea\xff\xff\xff\x0e\x0d\xff\xff\xff\xa0\xb6\xff\xff\xff\xfe\xfd\xff\xff\xff\xc4\xdc\xff\xff\xffz_\xff\xff\xff\x00\xe0\xff\xff\xff\xff\x8c\xff\xff\xff\x00\xec\xff\xff\xff\x15\x00\xff\xff\xff\x14\x00\xff\xff\xff!\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x00\x00\xff\xff\xff\x08'
    # Padding block. Introduced between blocks A and B at offset 616. It
    # encodes padding values until the next multiple of four bytes.
    padlen = 624 - len(lsd) - len(id) - len(cs) - 1
    pad = bytearray()
    while len(pad) < padlen:
        pad += rng.read(padlen - len(pad))
    pad[-1] = 0x0

    # Write out all components of our GIF file.
    out.write(header)
    out.write(lsd)
    out.write(id)
    out.write(cs)
    out.write(bytes(pad))

# Generates a random JPEG file into `out` using the
# random number genrator `rng` (/dev/urandom).
def generate_random_jpeg(
        rng: BinaryIO, out: BinaryIO
    ):
    # 1024 bytes of binary random data, to be written
    # afterwards as JFIF-formatted image.
    jfif = rng.read(1024)

    # First two bytes are SOI marker (start of image).
    # Last five bytes are EOI marker (end of image).
    jfif = b'\xff\xd8' + jfif + \
           b'\xff\xd9' 

    # Write out data section of JFIF.
    out.write(jfif[2:-5])

