#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Frame header
    width = 100 + int.from_bytes(
            rng.read(3), byteorder='big', signed=False) % (1 << 8)
    height = 100 + int.from_bytes(
            rng.read(3), byteorder='big', signed=False) % (1 << 8)
    
    background_color = int.from_bytes(
            rng.read(3), byteorder='big', signed=True) & 0xffffff
    pixel_aspect_ratio = int.from_bytes(
            rng.read(1), byteorder='big', signed=False) % (1 << 8)

    # Color table
    color_table = []
    for _ in range(3*2**(pixel_aspect_ratio+1)):
        color_table.append(int.from_bytes(
                rng.read(3), byteorder='big', signed=True))
        
    num_of_color_table = 3**pixel_aspect_ratio
    palette_len = num_of_color_table << 3
    global_color_table = True
    interlace_flag = False
    sort_flag = False
    local_color_table_flag = True

    frame_header = bytearray([
            *(width.to_bytes()),
            *(height.to_bytes()),

            background_color >> 16,
            background_color >> 8,
            background_color,

            pixel_aspect_ratio
        ])
    frame_header[4] |= \
        1<<7 | \
        1<<5 | \
        (sort_flag << 3) |\
        (interlace_flag << 2) |\
        (local_color_table_flag << 1) |\
        global_color_table


    frame_header += bytes(palette_len - len(frame_header))
    frame_header.extend((c>>16 for c in color_table))
    frame_header.extend((c>>8 for c in color_table))
    frame_header.extend((c for c in color_table))

    # LZW compression
    min_code_size = 8

    code_size = min_code_size + 1
    clear_code = 1 << code_size
    end_code = 1 << code_size;
    code_mask = (1 << code_size) - 1
    next_code = None

    dictionary = {}
    dict_index = 2**code_size
    prev_char = ''
    output = bytearray()

    def add_to_dictionary(*chars):
        nonlocal dict_index

        for char in chars:
            key = f'{prev_char}{char}'
            value = (dict_index << 4) | len(key)
            dictionary[key] = value
            prev_char = char
            dict_index += 1

    add_to_dictionary(' ') # EOF
    add_to_dictionary(',', '.')
    add_to_dictionary('0', '1', 'I', '+', 'i', ';', '-', '|')

    input = iter(rng.read())
    while byte := next(input, None):
        if not byte: continue
        if not isinstance(byte, bytes): byte = ord(byte)

        key = f'{prev_char}{chr(byte)}'
        if result := dictionary.get(key):
            prev_char = chr(byte)
            continue
        elif result is None:
            key = f'{prev_char}'

        dict_index += 1
        value = (dict_index << 4) | len(key)
        dictionary[key] = value

        if result is None or dict_index == (clear_code - 1):
            output.append(((value >> 4) & code_mask).to_bytes(
                length=1, byteorder='big')[0])
            if result is None:
                output[-1] &= ~0xf0

            if dict_index!= (clear_code - 1):
                output.append(((value >> 8) & code_mask).to_bytes(
                    length=1, byteorder='big')[0])

                if dict_index > (end_code - 1):
                    output.append((value & code_mask).to_bytes(
                        length=1, byteorder='big')[0])
                    code_size += 1

                    if code_size == max(min_code_size + 1, 12):
                        pass

            if dict_index == (clear_code-1):
                add_to_dictionary(',') # COM
                add_to_dictionary('.') # EOS
        
        prev_char = chr(byte)

    if dict_index <= (end_code - 1):
        output.append(((value >> 4) & code_mask).to_bytes(
            length=1, byteorder='big')[0])

    output.append(0x0)
    output.append(0x3B) 

    # Write to final output
    out.write(header + frame_header + output)


