#!/usr/bin/env python3

import math
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # create the header
    image_width = next((x for x in range(
        128, 151)), None)

    image_height = next((x for x in range(
        128, 151)), None)

    if image_width is not None and image_height is not None:
        
        # we have a valid dimensions
        out.write(header + struct.pack('<HHBB', image_width, 
                image_height, 0b11111111 & 0b11111, 
                0b1 << 7 | 0b1 << 5 | 0b1 << 4 | 
                0b111 << 3 | 0b111))

        # now create a buffer with enough space allocated to hold all
        # relevant pixels at once
        buff = bytearray([next(rng) for _ in range(image_width*image_height)])

        # calculate the LZW-compressed form of our pixels
        lzwbuff, pos = compress(buff, True)

        # write it out to the output buffer
        out.write(lzwbuff)

        return


    else:

        print("There was an issue generating a random GIF.")
        return


# Calculates LZW compression based on https://rosettacode.org/wiki/LZW_compression#Python
# Almost directly copied over the code. The only difference is that rather than calculating
# the minimum value of position within the current buffer as they do in their example, I instead
# use a lookup table which I store in the `pos` parameter
def compress(buf, final=False):
    """Compress a string using the modified Lempel–Ziv–Welch algorithm."""

    counts = dict(((chr(n), n) for n in range(256)))
    w = ''    # Empty string
    result = []   # Output list
    wc = chr(0)     # Empty character
    enlargeIn = 4      # Initial size of data structure (power of two)
    dictSize = 2 ** 8 + 2
    position = 0       # Current position within data array == bit index
    mask = 4095        # Mask to remove upper bits

    def addToDictionary(c):
        """Add a character to the dictionary"""

        nonlocal enlargeIn, dictSize, position, wordCount

        if c in counts:           # If already in dictionary
            wc = c                  # Then keep this same
            w += wc                 # Add new letter to end of current word
        elif len(w) < dictSize:            # Otherwise add to dictionary
            position |= 1 << len(wordCount)
            counts[wc + c] = position
            wc = c
            w += c
            enlargeIn -= 1
        elif c!= EOF:             # Unless at end of file
            raise Exception("Bad PNG")

        if enlargeIn == 0:         # Time to grow data structure?
            enlargeIn = pow(2, len(wordCount))
            prev = counts[EOF]
            for i in range(dictSize):

                curr = counts.get(w+chr(i), prev)
                counts[w+chr(i)] = prev
                prev = curr
            
            dictSize = dictSize + enlargeIn
    
        wordCount += 1

    def copyToOutput():
        """Copy the output window to its final location"""

        nonlocal position, result, buf, ret

        r = max(position, 1)
        for j in range(1,r):
            val = buf[(j + position) % len(buf)]
            result.append(val)

        position = 0

    wordCount = 1          # Initialize a few variables
    ret = ''               # Initialize a few variables
    EOF = '\0\0\xff\xff'   # End-of-file marker

    while True:     
        c = EOF
        try:
            c = buf[position]
            inc = 1
        except IndexError:
            pass       # Outside buffer bounds

        addToDictionary(c)

        if (not final and position >= 128) or \
           (final and c == EOF):

            result.append(ord(w))     # Copy last char but force out
            while len(result) > 0:
                ret += struct.pack("<H", result.pop())
        
            ret += struct.pack("<B", wordCount-1)
    
            copyToOutput()
                
            if final:
                break
            w = ""   
            result = []  
            wordCount = 1
            enlargeIn = 4      # Initial size of data structure (power of two)
            dictSize = 2 ** 8 + 2
            position = 0       # Current position within data array == bit index

        position += inc

    return ret, pos