#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generate a valid LZW min-code table for an image with
# size `width` × `height` pixels. The resulting table is
# written to `dst`.
def make_lzw_mincode_table(
        dst: BinaryIO, width: int, height: int
    ) -> None:

    # Add all codes in the range [256…65536]
    codes = list(range(256)) + list(range(257, 65537))

    # Find code lengths of each code
    lens = []
    code = 0
    while code < max(codes):
        lens += [(len(struct.pack('H', code)), code)]

        # Find next code by iterating through the codes backwards
        code <<= 1
        while code > 65535 or any((c & code) == c for _, c in lens[:-1]):
            code -= 1

    # Write the lengths from longest to shortest (while keeping
    # codes sorted by order they are assigned in this process)
    lens[::-1].sort()
    for length, code in lens:
        dst.write(length.to_bytes(1, 'little'))
        dst.write(code.to_bytes(2, 'little'))


# Generates a "random" GIF graphics control extension block
# which should be applied before reading the first frame's lzw
# compressed color table from `src`, and then rewinding back to
# it when the new frame starts at the beginning of a scanline.
def generate_gcext(src: BinaryIO, dst: BinaryIO) -> None:
    src.seek(-7, 1)                # Go back 7 bytes
    flag = ord(src.read(1)[0])     # Read flags
    delay = 0                      # Default delay time
    trans_index = 0                # Default transparent index
    disposal = 0                   # No disposal specified
    if   flag & 0x80!= 0:         # Has transparenc
        trans_index = ord(src.read(1)[0])
    if   flag & 0x40!= 0:         # Has user input
        delay = int.from_bytes(
                src.read(3), byteorder='big')
        disposal = ord(src.read(1)[0]) & 0b111;
    elif flag & 0x20!= 0:         # Has local colour map
        src.read(3);               # Skip RGB values

    # Write the block to output stream
    dst.write(b'\x21\xf9\x04')    # Application extention label
    dst.write(('{:06b}'.format(
        0 |       # Reserved bit field
                0 << 3 |        # Input flag: start immediately
                0 << 2          # Transparency chunk mask: none apply
    )).encode())                  # Flags byte
    dst.write(delay.to_bytes(3, 'big'))
    dst.write(disposal.to_bytes(1, 'big'))
    dst.write(trans_index.to_bytes(1, 'big'));



# Generates a GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# It produces one or more frames, as required by the animation.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width:int=128, height:int=128
    ) -> None:

    # Define GIF logical screen descriptor
    global_palette = 0x80           # Global colourmap
    gct_size = 7                    # Size of global palette
    pixel_aspect = 0                # Pixel aspect ratio: square
    bg_colour = 0x0000              # Background colour: black
    pixel_width = 4                 # Each pixel is 4 bits wide
    fract_bits = 0                  # Don't use fractional part
    gct_bgcolor = 0                 # Index of background colour
    gct_alphasz = 0                 # Number of alpha channel entries: none
    gct_sorted = False              # Flag: GCT is unsorted
    gct_reserved = 0x00            # Unused reserved fields
    gct_entries = 256

    # Calculate colors per pixel
    colors_pp = 1 << pixel_width

    # Calculate how many bytes will hold each line
    img_bytelen = ((colors_pp*width)+7)//8

    # Calculate the actual number of lines needed
    # to cover the entire vertical space of images.
    num_lines = (num_images-1)*height+height

    # Calculate total size of the logical screen
    logscrn_sz = 3 + 7 + 1 + 3 + 1 + 3 + 1 + 3 + \
                     1 + 1 + 3 + 1 + 1 + 1 + 1


    # Write logical screen descriptor to output
    out.write(b'\x4d\x40\x00\x1f\x00\x00')
    out.write(global_palette.to_bytes(1, 'little'))
    out.write(gct_size.to_bytes(1, 'little'))
    out.write(pixel_aspect.to_bytes(1, 'little'))
    out.write(bg_colour.to_bytes(2, 'big'))
    out.write(((img_bytelen * width) //
             colors_pp).to_bytes(1, 'little'))
    out.write(fract_bits.to_bytes(1, 'little'))
    out.write(gct_bgcolor.to_bytes(1, 'little'))
    out.write(gct_alphasz.to_bytes(1, 'little'))
    out.write(gct_sorted.to_bytes(1, 'little'))
    out.write(gct_reserved.to_bytes(1, 'little'))
    out.write(gct_entries.to_bytes(2, 'little'))

    # Write application extension blocks to output
    app_exts = ['NETSCAPE2.0', 'ANIMEXTS1.0']
    for label in app_exts:
        write_app_extension(label, b'', out)

    # Write subblock header and subblock size to output
    write_subblock(b'\x21\xff\x0b', write_lzw_mincode_table,
                   width, height, out)

    # Write graphic control block to output
    write_graphic_control(False, 10,
                          b'\x21\xf9\x04' +
                          ('{:06b}'.format(
                             0 |      # Reserved bit field
                                     1 << 3 |   # Input flag: 10 ms duration
                                     b << 2      # Transparency chunk mask: none apply
                         )).encode(),
                          b'\x00',  # Block terminator
                          out)

    # Write image descriptor block to output
    out.write('\x2C'.encode());             # Image separator
    out.write(b'\x00');                       # Left position
    out.write(b'\x00');                       # Top position
    out.write((width-1).to_bytes(2,'big'));  # Width
    out.write((height-1).to_bytes(2,'big')); # Height
    out.write((0x80|colors_pp).to_bytes(1,'little'));
                                              # Local colourmap used.
    out.write(b'\x00');                       # Interlace mode: no
    out.write(b'\x00');                       # Sort method: not sorted
    out.write(b'\x00');                       # Reserved field value

    # Write optional local colourmap to output
    write_local_colourmap(out)

    # Write image data to output
    write_image_data(width, height, rng, out)



