import os
from functools import partial

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom by default)
def generate_random_gif(
        rng: Callable[[int], bytes] = os.urandom,
        out: BinaryIO = open('test.gif', 'wb'),
        width: int = 128, height: int = 128,
        color_depth: int = 7, loop_count: int = -1
    ):
    # Header
    out.write(b'GIF89a')

    # Screen descriptor
    out.write((width).to_bytes(2, 'big'))
    out.write((height).to_bytes(2, 'big'))

    bitfield = ((color_depth-1) << 4 | 0x0) & 0xff
    if loop_count == 0:
        bitfield |= 0x80
    elif loop_count > 0:
        bitfield |= 0x80 | ((loop_count & 0xFF) >> 8)
    out.write(bitfield.to_bytes(1, 'big'))
    out.write(((loop_count & 0xFF) ).to_bytes(1, 'big'))
    out.write(b'\x00\x00')

    # Global palette
    for i in range(0, 2**(color_depth+1)):
        r = g = b = 0
        if i!= 0:
            r = rng(1)[0] & 0xf8
            g = rng(1)[0] & 0xfc
            b = rng(1)[0] & 0xf8

        out.write(b'\x00')                                     # Reserved
        out.write(((r>>5)&0x7c)|0x1f)                         # Red
        out.write(((g>>2)&0xe0)|0x1f)                         # Green
        out.write(((b<<3)&0xf8)|0x1f)                         # Blue

    # Image Descriptor
    out.write(b'\x2C')                                         # Image Separator
    out.write((0).to_bytes(2, 'big'))                          # Left Position
    out.write((0).to_bytes(2, 'big'))                          # Top Position
    out.write((width).to_bytes(2, 'big'))                      # Width
    out.write((height).to_bytes(2, 'big'))                     # Height
    out.write((0x0).to_bytes(1, 'big'))                        # Local Color Table Flag
    out.write((0xf0 | 0x0).to_bytes(1, 'big'))                 # LZW minimum code size


    # LZW compression interlude
    def write_table() -> None:
        lzwsizelookup = [
            3, 4, 5, 6, 7, 8, 9, 10, 11, 13,
            15, 17, 19, 23, 27, 31, 35, 43, 51, 59,
            67, 83, 99, 115, 131, 163, 195, 227,
            258, 0, 0, 0, 0, 0, 0, 0
        ]
        
        # Write table header
        out.write(lzwsizelookup[color_depth])                   # Size of lookup table is fixed based on the global color depth
        out.write(b'\x00')                                     # First entry is always blank

        # Build table and send to output stream
        lzt = { }
        valuelist = []
        for i in range(0, 2**(color_depth+1)):
            valuelist.append(i)

        code = len(valuelist)-1   # Start with maximum possible code length (indicates end of compressed frame)
        maxcode = 2**code         # The maximum value that can be used as an index to the lookuptable 
        begincode = maxcode       # Value at beginning of next search window
        buffer = b''              # Buffer to hold values to compress
        for i in range(0, 2**(color_depth)):                  # Send first part of LZW dictionary
            lzt[(i)] = valuelist[i]
            buff = lzt[i].to_bytes(1, 'big') 
            outbuf = b'\x01' + buff                             # Marks new string
            out.write(outbuf[:len(outbuf)])

        while True:                                              # Begin compression
            # Get the next pixel from the input data and add it to our buffer
            pix = ord(rng(1))                                    # Pixel will now be a number rather than binary representation

            buffer += (((pix&0xFF)>>color_depth)).to_bytes(1, 'big')
            
            # See if we have any matching strings in the dictionary
            key = tuple(buffer)                                  # Key is a tuple containing all elements of buffer
            try:
                val = lzt[key]
                buffer = b''                                      # Reset buffer since we added a match

                out.write(val.to_bytes(1, 'big'))                 # Output match
            except KeyError:                                     # No matches found
                pass                                             # Just continue like nothing happened


            # If we reached the end of a search sequence...
            if begincode >= maxcode:                              #...then reset the search window and start over
                begincode = maxcode                               # Max code indicates end of compression

                if len(valuelist) < 4096:
                    print("Adding more entries")
                    # Extend the lzw dictionary 
                    valuelist.append(maxcode)                    # Add previously encountered code to list

                    for j in range(begincode, maxcode):          # Fill new entries up
                        lzt[(j)] = maxcode                       # Insert copy of old code

                        buff = lzt[j].to_bytes(1, 'big') 
                        outbuf = b'\x01' + buff                 # Marks new string
                        out.write(outbuf[:len(outbuf)])         # Append to output

                    # Increase next code length until there are enough values defined
                    code += 1
                    maxcode = pow(2, code)                       # Maximum value that can be used as an index to the lookuptable

                    if maxcode == 511:                           # If we just increased past 511, then reset everything
                        maxcode = 2**code
                        code -= 1
                               
                else:                                               # Just reset buffer and keep trying
                    print("Maximum lzw dictionary exceded.")     # This shouldn't happen
                    buffer = b''                                  # Reset buffer since we couldn't find anything more
                    out.write(b'\x00')                            # Single zero byte sentinel indicating no match
                    break                                       # Break out of main while loop


        # End of Table marker
        out.write(b'\x00')

    # Write table once before image data begins
    write_table()
    
    # Image Data
    blocksize = 1    # TODO: Determine block size
    currentblock = 0
    xpos = ypos = 0  
    while True:     
        if xpos >= width or ypos >= height:               # Are we done?
            break                                           # Go home

        # Calculate the offset where this block should go
        offset = (ypos * width + xpos)//8
        out.seek(offset, 0)                                # Set cursor position

        # Figure out how much padding needs to be written after this block
        remainingbits = (xpos % 8) ^ 7
        padbits = (remainingbits) // blocksize             # How many bits do we need to fill the last incomplete byte?

        # Generate the next block
        datablock = rng(1<<(blocksize+(padbits)))        # Read a full byte, plus whatever padding needed
        out.write(datablock)

        # Increment coordinates
        xpos += blocksize
        if xpos >= width:                                  # Did we reach horizontal boundary?
            xpos = 0                                        # Yes! Reset X coordinate back to left edge
            ypos += 1                                       # Increment Y coordinate

    # Block terminator
    out.write(b'\x00')

