#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng`.
# `mode`:
#    - 'binary': binary output (default)
#    - 'ascii': printable ASCII characters
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, mode='binary',
        width=None, height=None, bitdepth=None, nframes=1,
        bgcol=(0x00, 0x00, 0x00), colmap=None, xoff=None, yoff=None,
        delay=None):

    if mode not in ('binary','ascii'):
        raise ValueError('Invalid mode')

    # Header starts with "GIF" and identifier bytes
    h = [0x47, 0x49, 0x46]

    # Build the logical screen descriptor block
    w = None
    h = None
    bgcol = None
    colinfo = []
    gctsize = None
    lsdlen = len(h)+5
    if width is None or height is None:
        while True:
            # Read from stdin until we get an int value > 1
            try:
                d = read_int(sys.stdin)
                break;
            except ValueError as e:
                pass

            print("Please enter valid values.")

        W, H = d, d
        while True:
            if W % 2 == 0 and H % 2 == 0:
                break
            else:
                if W % 2!= 0:
                    W += 1
                elif H % 2!= 0:
                    H += 1

                print("Width ({}) or Height ({}) must be even integers.".format(W, H))
    
    if nframes < 1:
        raise ValueError('Number of frames must be at least one.')

    if bitdepth is not None and bitdepth <= 0 and bitdepth >= 9:
        raise ValueError('Bit depth should be between 1 and 8 bits')

    if nframes > 1:
        if delay is None:
            print("Delay cannot be None when generating multiple frames")
            os._exit(-1)
        if delay < 0:
            raise ValueError('Delay must be positive')

    # Parse command line arguments for optional LSD parameters
    if width is not None:
        width &= 0xffff
        h.extend([width & 0xff, width >> 8])
        lsdlen += 2
    if height is not None:
        height &= 0xffff
        h.extend([height&0xff, height>>8])
        lsdlen += 2
    if bgcolor is not None:
        bgcol |= 0xffffff
        h.extend([(bgcol >> i)&0xff for i in range(24, -1, -8)])
        lsdlen += 3
        
    block_terminator = False
    while True:
        if block_terminator:
            break

        # Get GCT size
        d = read_int(sys.stdin)
        if d < 1:
            raise ValueError('GCT size must be greater than zero')
        
        gctsize = d*3

        # Get color table
        coltable = [] 
        for _ in range(gctsize):
            v = ord(sys.stdin.read())
            if v < 0:
                raise ValueError('RGB values must be non-negative')
            
            if mode == 'ascii':
                c = chr((v+0x20)%0x7f)
            else:
                c = v%0xff

            coltable.append(c)
        lsdlen += gctsize
        
        codelens = []
        if bitdepth is not None:
            block_terminator = True
            
        # Calculate code sizes
        while True:
            # Read next int off stdin
            d = read_int(sys.stdin)
            if d < 0:
                raise ValueError('Code length must be non-negative')


            if d == 0:
                raise ValueError('Code length can\'t be zero')
            codelens.append(d)

            if sum(codelens) > 12:
                raise ValueError('Total code lengths must be less than 12')

            if min(codelens) < 2:
                continue
                
            if max(codelens) > 8:
                continue
                

            if mode == 'ascii':
                symbs = [''] * ((gctsize+1)/max(codelens[i:])*8)
            else:
                symbs = [(b'\x00\x00')[j//8] << j%8 for i in range(min(codelens))]

            clstr = ''.join(chr(cl) for cl in codelens)
            break
        

    # Add global color table to frame laters
    if gctsize is not None:
        # Add the global color table block
        lsdlen += 1
        h.append(lsdlen & 0xff)
        h.append(lsdlen >> 8)
        h.append(bitdepth)
        h.append(0b10000000)
        h.append(gctsize & 0xff)
        h.append(gctsize >> 8)
        h.extend(bytes(coltable))

        if gctsize == 14:
            h.append(0)
            h.append(0x00ff)
            h.append(0)
            h.append(0xff00)

    # Parse local image descriptors
    idblocks = []
    for idx in range(nframes):
        # Start with the image separator
        h.extend([0x2C,
                   # Image left offset
                   xmin & 0xff, (xmin >> 8) & 0xff, 
                   # Image top offset
                   ymin & 0xff, (ymin >> 8) & 0xff, 
                   # Width
                   width & 0xff, (width >> 8) & 0xff, 
                   # Height
                   height & 0xff, (height >> 8) & 0xff])

        # Flag block terminator
        if nframes == 1:
            # No block termination flag
            h.append(0x00)
        else:
            # Block termination flag set
            # Background index set to None
            # Transparency flags set to all false
            h.append(0x00)
            h.append(0x00)
            h.append(0x00)

        # Code size
        h.extend(codestr)

        # End image data
        h.append(0x00)

        # Set up ID block
        idblock = [idx ^ 0x80]
        if delay is not None:
            delay = delay * 10
            idblock.append(delay&0xff)
            idblock.append((delay>>8) & 0xff)
            idblock.append((delay>>16) & 0xff)
            idblock.append((delay>>24) & 0xff)

        idblocks.append(idblock)


    # Write everything to stdout
    out.write(struct.pack('!'+'B'*len(h), *h))

    # Now write the per-frame data
    for idx in range(nframes):
        # Write the ID block first
        out.write(bytes(idblock))

        # Generate pixels for this frame
        img = []
        for row in range(height):
            rle = []
            for col in range(row, width, step):
                bval = bytearray(rng.read(step))[::-1]
                cc = [coltable[i]*2**(codelens[i]-1)-1 for i in bval]
                ll = conv(cc, symbs)
                rle.append(ll)
                
        # Interleave RLE streams
        interleaved = zip(*rle)[::2]
        interleaved = list(interleaved)
                
        for row in interleaved:
            obuf = bytearray()
            last = row[0][0]
            cnt = 0
            for pair in row:
                if pair[0] == last:
                    cnt += 1
                else:
                    obuf.append(cnt | last)
                    last = pair[0]
                    cnt = 1                    
                    
            obuf.append(cnt|last)
            out.write(obuf)
            
if __name__=='__main__':
    import argparse

    parser = argparse.ArgumentParser(description='Generate random GIF images')
    parser.add_argument('-w', '--width', type=int, help="Image width", default=None)
    parser.add_argument('-H', '--height', type=int, help="Image height", default=None)
    parser.add_argument('-d', '--depth', type=int, choices=[2**k for k in range(1,9)], help="Color resolution / bit depth", default=None)
    parser.add_argument('-c', '--ncolors', type=int, help="Number of colors used", default=None)
    parser.add_argument('-n', '--nframes', type=int, help="Number of frames in animation", default=1)
    parser.add_argument('-C', '--colorfile', metavar='FILE', type=argparse.FileType(), help="File containing RGB palette data", default=None)
    parser.add_argument('--delay', metavar='INT', type=int, nargs='+', help="Frame