#!/usr/bin/env python3

import argparse
from math import ceil
import os
import sys
from typing import Iterable, Optional, List

from PIL import Image


class Rng():
    def __init__(self,
                 seedfile: str):
        """Create an RNG based on the contents of /dev/urandom."""
        self._f = open(seedfile, "rb")

    def randint(self, minv: int=None, maxv: int=None) -> int:
        """Return a random integer within the specified interval.

        If only one argument is given it will be treated as the maximum
        value to return and zero will be used for the minimum.

        Throws IndexError if you specify both arguments as None.
        """
        if minv == None or maxv == None:
            raise IndexError("You must supply two integers.")

        delta = maxv - minv
        if delta < 0:
            raise ValueError("Maximum ({}) less than minimum ({})"
                            .format(maxv, minv))

        res = int.from_bytes(self._f.read(ceil(delta.bit_length() / 8)),
                             'little', signed='false') & delta

        return minv + res

    def choice(self,
               seq: Iterable,
               weightfunc: Optional[callable]=None
              ) -> object:
        """Choose an element from a sequence at random weighted by the given
           function.

           This works by choosing a base random number between 0 and summed
           weights of all choices. The chosen item is then returned.

           Note that there's no way to reject an entry during picking until
           after its weight has been reduced. So your distribution might end up
           being skewed towards a few entries.

           If weightfunc is not supplied, the default is equal weighting. It'll
           be called as ``weightfunc(entry, iteration)`` where iteration starts
           counting at zero.

           Example::

               >>> rng = Rng('/dev/urandom')
               >>> fruit = ['apple', 'banana', 'orange']
               >>> rng.choice(fruit)
               'orange'

       """

        if weightfunc!= None:
            weights = list([weightfunc(e, i)
                for (i, e) in enumerate(seq)])
        else:
            weights = [1 for _ in seq]

        totalw = float(sum(weights))

        sel = self.randint(*[0, round(totalw)])

        wacc = 0
        for (idx, w) in enumerate(weights):

            wacc += w
            if sel <= wacc:

                return seq[idx]


def generate_image(rng: Rng) -> Image.Image:
    """Generate a new image in memory."""

    img = Image.new('L',
                    (rng.randint(minv=32, maxv=128),
                     rng.randint(minv=32, maxv=128)))

    img.putdata([(rng.randint()) for _ in range(img.width * img.height)])

    return img


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-o", "--output",
                        help="The output GIF image.", required=True)
    parser.add_argument("--seed",
                        help="/dev/urandom or equivalent source of entropy"
                       .format(os.path.getsize(__random__)),
                        default="/dev/urandom")
    args = parser.parse_args()


    rng = Rng(args.seed)

    images = []
    num_images = rng.randint(minv=5, maxv=10)
    print("Generating {} images".format(num_images))
    while len(images) < num_images:

        try:
            images.append(generate_image(rng))
        except Exception as ex:
            print(ex)
            continue


        # Try again!

    imgesizes = {i.size for i in images}
    largest = sorted(imgesizes)[-1]
    #print(largest)
    framedurations = [rng.randint(1, 10)
                      for _ in range(len(images))]
    #print(framedurations)
    assert sum(framedurations) >= len(images)

    speed = rng.choice([100, 200])

    fcount = 0
    totalsizetomake = sum([s[0] * s[1] for s in imgesizes])
    imgsizestotmake = set([])
    remainingsize = totalsizetomake
    lastimg = None
    try:
        while remainingsize > 0:

            if len(imgsizestotmake) > 0 and\
               (lastimg is None or
                   ((remainingsize == totalsizetomake) and
                       (lastimg.size not in imgsizestotmake))):


                targsize = rng.choice(list(imgsizestotmake),
                                      lambda s, i: abs(s - largest))

            elif len(imgsizestotmake) == 0:

                targsize = largest


            print("New Size:", targsize,
                  "Remanining:", remainingsize, "TotalSize:", totalsizetomake)
            fcount += 1
            baseimg = images[-1].resize(targsize, resample=Image.BOX).convert('P')

            if lastimg is None:

                lastimg = baseimg

            else:

                # Overlay onto last image and draw line down middle
                overlay = Image.new('P',
                                    (baseimg.width,
                                     lastimg.height + baseimg.height))
                overlay.paste(lastimg,
                              box=(0, 0,
                                   lastimg.width,
                                   lastimg.height))
                overlay.paste(baseimg,
                              box=(0,
                                   lastimg.height,
                                   baseimg.width,
                                   baseimg.height+lastimg.height))
                linesize = int(overlay.height * 0.1)
                drawr = overlay.crop(((0,
                                      0,
                                      overlay.width,
                                      linesize)))
                drawg = overlay.crop(((0,
                                      linesize,
                                      overlay.width,
                                      linesize*2)))
                drawb = overlay.crop(((0,
                                      linesize*2,
                                      overlay.width,
                                      linesize*3)))
                white = Image.new('P',
                                  (overlay.width,
                                   overlay.height)).paste(255)
                drawr.paste(white, mask=drawr)
                drawg.paste(white, mask=drawg)
                drawb.paste(white, mask=drawb)

                lastimg = overlay

            remainingsize -= targsize[0] * targsize[1]
            imgsizestotmake.remove(targsize)



        totalduration = sum(framedurations)

        # Write out frames
        print("Writing Images...")
        for idx, (i, d) in enumerate(zip(images, framedurations)):

            outfile = "{}-{:05d}.gif".format(args.output,
                                             idx)
            print("\t{}".format(outfile))
            i.save(outfile, transparency=255, duration=speed//d)

        # Also write out a merged animation
        mergedfname = "-merged.gif"
        print("Merging animations...", mergedfname)
        cmdstr = ("convert -delay {:d} {files}"
                  " {mergedfname}")

        files = ["{}-{:05d}.gif".format(args.output,
                                        idx) for idx in range(len(images))]
        cmdparams = {'files': " ".join(["\"" + f + "\"" for f in files]),
                    'mergedfname': mergedfname}

        print("\t{}".format(cmdparams['mergedfname']))

        rtncode = os.system(cmdstr.format(**cmdparams)) >> 8
        if rtncode!= 0:
            print("Error code {} when running '{}'"
                  "".format(rtncode,
                            cmdstr.format(**cmdparams)),
                  file=sys.stderr)



    finally:

        # Clean up any partial files
        for idx in range(len(images)):
            if not os.path.exists("{}-{:05d}.gif".format(args.output,
                                                          idx)):
                break
            os.unlink("{}-{:05d}.gif".format(args.output,
                                            idx))








