#!/usr/bin/env python3

import struct


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). Returns True on success
# or False otherwise
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bool:

    def _get_byte(): return rng.read(1)


    try:
        
        header = b'GIF89a'

        # width and height are 16 bits each
        image_width, image_height = \
            struct.unpack('<HH', _get_byte()*4)

        # skip fields that aren't particularly interesting
        # (the logical screen descriptor - color table size, background
        # color index, pixel aspect ratio, global color table flag,
        # color resolution, sort flag, size of global color table.)
        # we'll just assume these all start with zeroes
        [next(_get_byte() for x in range(7)) for x in range(7)]


        # size of the color table
        ctsize = int.from_bytes(_get_byte(), 'little')

        if ctsize == 0: return False

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):

            # write out the length byte first
            palval = paldata[i]
            palents += struct.pack('B', len(palval))
            palents += palval

        # write to output stream
        out.write(header + struct.pack('<HHBB', image_width, 
            image_height, ctsize | 0b10000000, 
            0b1 << 7 | 0b1 << 5 | 0b1 << 4 | 
            0b111 << 3 | 0b111) + palents)
        
        # get remaining lzw compressed info from the RNG
        datalen = int.from_bytes(_get_byte()*2,'big')
        data = rng.read(datalen-1)[:-1]
        out.write(data)
    except Exception as e:
        print('[!] Failed to generate valid Gif:\n{:s}'.format(e))
        return False

    return True