#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize == 0:

        out.write(header +
            struct.pack('<H', image_width) +
            struct.pack('<H', image_height) +
            b'\xF7\xFF' +
            struct.pack('<B', 0) +       # no local color table
            b'\xCE\x00\x00'             # color map follows immediately
        )


    elif ctsize <= 256:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):

            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]


        out.write(header +
            struct.pack('<H', image_width) +
            struct.pack('<H', image_height) +
            b'\xF7\xFF' +
            struct.pack('<B', 1<<7|int(log(ctsize)/log(2))) +   # local color table
                                                                # plus size
            palents                      +               # palette entries
            b'\xCE\x00\x00'             # color map follows immediately
        )

        # colormap is followed by a local color table
        # TODO: we could do something with this; but not now...
        rng.seek((3 << 8) | ctsize)

