#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    width = int.from_bytes(
            rng.read(2), byteorder='little') + 1
    height = int.from_bytes(
            rng.read(2), byteorder='little') + 1

    # Write the global color table
    numcolors = 2**(width > height)
    for _ in range(numcolors):
        rgb = rng.read(3)
        out.write(rgb + bytes([0]))
    
    for i in range(height):
        lzwmincodesize = min((i+1)**3*3//8, 12)

        # Write an LZW code size change command to set what we want here
        out.write(bytes([lzwmincodesize | 0xe0]) +
                  bytes([(lzwmincodesize & 0xf)]))
        
        for j in range(width // (2**lzwmincodesize)):
            code = 0

            while code < 256:
                bitstring = bin(code)[2:]

                # Add enough leading zeroes so that this is at least as long
                # as the lzwmincodesize
                bitstring = ('0'*(lzwmincodesize-len(bitstring))
                            + bitstring[::-1])
                
                # Take however many bytes we need from our bitstream
                bits = bitstring[:lzwmincodesize]
                restofbits = bitstring[lzwmincodesize:]

                # The current byte we're looking at
                pos = 0
                # The last cached pixel
                previouspixel = None

                try:
                    # Try reading until you get more than three
                    # consecutive identical pixels, or run out of data
                    while (pos < 3) or \
                          ((previouspixel == data[index+2])
                           and (previouspixel == data[index+1])
                           and (previouspixel == data[index])):

                        index += 1
                        if len(restofbits) == 0:
                            index += 1
                            restofbits = rng.read(1)[0].to_bytes(
                                    8, 'little')[::-1][lzwmincodesize:]
                        if len(bits) <= 0:
                            bits = restofbits[:lzwmincodesize]
                            restofbits = restofbits[lzwmincodesize:]

                    previouspixel = data[index]

                    if len(restofbits)!= 0:
                        index -= 1
                    
                    # Generate some new bytes based on how big these are
                    length = ((int(''.join('1' if x else '0'
                                for x in bits), base=2)+1)*3)//8

                    # Account for the fact they're one less byte than they could be
                    if length == 257:
                        length = 256
                    elif length == 513:
                        length = 512

                    # Read up to two bytes because there's no way to know ahead of time
                    # whether this will be encoded with a single byte or two
                    tmp = data[index+length//8:][:length%8//8+1]                    

                    code = int(str(bits+tmp[-1]), base=2)
                    out.write(bytes([code]))
                except IndexError:
                    raise Exception("ran off end")


# Generates a random JPEG file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_jpeg(
        rng: BinaryIO, out: BinaryIO,
        blocksize: int=256
    ):
    header = b'\xff\xe0\x00'
    padding = [b'\x00'] * blocksize
    out.write(header + len(data).to_bytes(3, 'big'))
    while True:
        data, padding = rng.read(blocksize - 4), \
                        rng.read(blocksize - 17)
        if not (padding and data): break
        out.write(data + padding)
