#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) of length `len`.
# Maximum size is 256MiB (262144*1024).
def generate_random_gif(
        len: int, rng: BinaryIO, out: BinaryIO
    ):
    if not isinstance(len, int):
        raise TypeError('length must be an integer')

    if len < 0 or len > 262144 * 1024:
        raise ValueError('maximum length is 256 MiB (262144*1024)')

    def make_block(data):
        return bytes([
            0x21,   # Extension introducer
            0xFF,   # Application extension
            block_size % 0x100,
            (block_size >> 8) % 0x100,
            (block_size >> 16) % 0x100,
            (block_size >> 24) % 0x100,
            ] + list(data))

    data = []
    BLOCK_SIZE = 256 * 1024    # Block size is fixed to 256 KiB
    while len > 0:
        n = min(BLOCK_SIZE, len)

        try:
            block = rng.read(n)
        except OSError as e:
            print(e, file=sys.stderr)
            exit(-1)

        if len == n:
            d = [
                0x2C,       # Image descriptor
                ((n >> 24) & 0xff),
                ((n >> 16) & 0xff),
                ((n >> 8) & 0xff),
                (n & 0xff),
                0,           # No local color table
                0,0,          # Left position
                frame_height-1,     # Top position
                frame_width-1,      # Width minus one - x coordinate
                frame_height-1,     # Height minus one - y coordinate
            ] + list(block)

            d += (b'\x00'*(BLOCK_SIZE+1))[:frame_width - n]
            data += make_block(bytes(d))
            break

        else:
            d = [
                0x2C,       # Image descriptor
                ((n >> 24) & 0xff),
                ((n >> 16) & 0xff),
                ((n >> 8) & 0xff),
                (n & 0xff),
                0,           # No local color table
                0,0,          # Left position
                frame_height-1,     # Top position
                frame_width-1,      # Width minus one - x coordinate
                frame_height-1,     # Height minus one - y coordinate
            ] + list(block)

            d += (b'\x00'*(BLOCK_SIZE+1))[:frame_width - n]
            data += make_block((bytes(d)))
            len -= n


    palette = [
        0x00, 0x00, 0x00, 0x00, 0x00, 0xCC, 0x33, 0x66,
        0x44, 0xBB, 0xAA, 0x99, 0x77, 0xEE, 0xDD, 0xCC,
        0xBB, 0xFF, 0xFF, 0xFF,
    ]
    data += make_block(palette)
    loop_count = 0
    LOOP_COUNT_DELTA = 0xf
    data += make_block(struct.pack('<B', loop_count & 0xff))

    global screen_width, screen_height
    screen_width = max(frame_width, 200)
    screen_height = max(frame_height, 200)

    logical_screen = [
        0x21,   # Extension introducer
        0xF9,   # Graphic control label
        4,       # Byte count
        0x01 << 2,        # Reserved
        0,
        0,
        0,
        0,
        (screen_width >> 8)%0x100,              # Screen width
        (screen_width & 0xff),(screen_height>>8)%0x100,         # Screen height
        (screen_height & 0xff),                  # Packed field:
                                                   # Global color table here?
        0x00,                                        # Background color index
        0x00,                                        # Pixel aspect ratio n/m
    ]
    data += make_block(bytes(logical_screen))
    out.write(('GIF89a'+''.join('%c'%i for i in data)).encode())

