#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import math
import struct
from math import ceil as ceiling
from pathlib import Path
from typing import BinaryIO


def main():

    # check if a filename is provided
    if len(sys.argv)!= 2:
        print('Error: Please provide an output file.')
        return

    # obtain seed from args
    try:
        seed = int(Path(os.path.expanduser(
            '.'+sys.argv[1]
        )).stem[-7:], base=16)
    except ValueError:
        print('Error: Wrong input file name format')
        return

    with open('/dev/urandom', 'rb') as rng:

        # set seed on rng
        rng.seek(seed % (rng.seek(0, 2)-4))

        # get width and height of the image
        w = 65536//int.from_bytes(rng.read(2), byteorder='big')
        h = 65536//int.from_bytes(rng.read(2), byteorder='big')

        # get number of colors in the palette and background color index
        nc = 2**(8*rng.read(1)+1)
        bc = rng.read(1)[0]

        maxbits = 2**b

        # generate main stream
        imdata = []
        for i in range((w*h)//maxbits):

            # get bits to write
            nbits = min(rng.read(1)[0]+1, maxbits)

            # get start position
            x = int.from_bytes(
                rng.read(math.ceil(nbits/8)), byteorder='little', signed=True
            ) % (w - nbits)
            y = int.from_bytes(
                rng.read(math.ceil(nbits/8)), byteorder='little', signed=True
            ) % (h - nbits)

            # write line
            imdata += [(x,y)]
            imdata += [rng.read(math.floor(nbits/8))] * ((nbits-1) // 8 + 1)


        # generate palette
        pal = []
        for cpos in range(nc):

            # generate r g b values
            r = rng.read(1)[0]<<2 | rng.read(1)[0]>>6
            g = rng.read(1)[0]<<2 | rng.read(1)[0]>>6
            b = rng.read(1)[0]<<2 | rng.read(1)[0]>>6

            # add this color to the palette
            pal += map(lambda v: v&0xff, (r,g,b))

        # merge everything into one file
        fdata = b''

        # add global headers
        fdata += b'\x47\x49\x46\x38\x39\x61'         # GIF89a
        fdata += b'\xf7\xa0'                         # Logical Screen Descriptor
        fdata += b'\x00\x00'                         # Width
        fdata += b'\x00\x00'                         # Height
        fdata += b'\x00'                             # Pixel aspect ratio
        fdata += bytes([bc])                        # Background Color Index (GIF Specific)
        fdata += b'\xff'                             # Transparent Color Index (GIF Specific)
        fdata += b'\xff\xff'                         # GCTF

        # convert pixel data into chunks
        chunksize = 255
        curchunk = 0
        while True:

            # get next sub block
            curchunk += 1
            pos = fdata.index(b'\x00'*chunksize,len(fdata)-(chunksize+curchunk*chunksize))+chunksize
            sblock = fdata[:pos-(chunksize+curchunk*chunksize)]

            # create a new frame
            f = b'\x2c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'

            # add size info
            f = f[:4] + \
                len(sblock).to_bytes(2,'big') + \
                f[6:]

            # add position info
            f = f[:8] + \
                (ceiling(pos / maxbits)*maxbits).to_bytes(2,'big') + \
                (ceiling(pos / maxbits)*maxbits).to_bytes(2,'big') + \
                f[10:]

            # add lzw compression info
            f = f[:17] + \
                b'\x08' + \
                f[18:]

            # add local palette info
            f = f[:21] + \
                ((ceiling(nc/256))*256).to_bytes(1,'big') + \
                ((nc-1)&0xff).to_bytes(1,'big') + \
                f[23:]

            # add main content
            f = f[:32] + sblock + f[32+(len(sblock)):]

            # do not append final block
            if len(imdata) == 0 or pos > (chunksize*(curchunk+1)):
                break


            # get next position
            x,y = imdata.pop()
            f = f[:36] + \
                x.to_bytes(2,byteorder='little',signed=True) + \
                y.to_bytes(2,byteorder='little',signed=True) + \
                f[38:]

            # pad remaining zeros
            f = f[:-2] + b'\x00\x00' + f[-2:]

            # add to final file
            fdata += f

        # add extension to store palette
        fdata += b'\x21\xf9\x04\x01\x00\x00'             # Graphics Control Extension Introducer
        fdata += b'\x00\x00'                             # Size of introducer
        fdata += b'\x00\x01'                             # Blocksize (always odd)
        fdata += b'\xff'                                 # No delay
        fdata += b'\x0b'                                 # Sub block type
        fdata += b'\x00\xe0'                             # Palette Size (64)
        fdata += b'\x00\x00\x00\x00\x00\x00\x00\x00'     # Reserved Field
        fdata += b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'   # Global Color Table
        fdata += b'\x00\x00'                             # Terminator