#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    width = 1 << (rng.read(1)[0] & 7)
    height = 1 << ((rng.read(1)[0] >> 4) & 7)

    screenwidth = [
        320, 640, 800, 1024, 1280, 1600, 
        2048, 2560, 3200, 4096][(rng.read(1)[0] & 15)]
    screenheight = [
        200, 400, 500, 800, 1000, 1200, 
        1600, 2000, 2500, 3200][(rng.read(1)[0] >> 4)]

    flags = bytes([((rng.read(1)[0] >> i) & 1) for i in range(8)])

    backgroundcolor = rng.read(1)[0]
    pixelaspectratio = rng.read(1)[0]

    palette = []
    for _ in range(2 ** (((flags[0] == 1) << 3) |
                       ((flags[1] == 1) << 2) |
                       ((flags[2] == 1) << 1))):

        rgb = [(rng.read(1)[0],
               rng.read(1)[0],
               rng.read(1)[0])]

        if flags[3]:
            alpha = rng.read(1)[0]

            rgba = [*rgb, alpha]

            palette += [[*rgba]]

        else:
            palette += [[*rgb]]


    globalcolortable = [rng.read(1)[0] for _ in range(3 * 2 ** (flags[-1]))]
    sortedcolortable = palette + globalcolortable
    
    colorresolution = {
        1 : "No colors",
        2 : "16 colors",
        3 : "256 colors"
    }[(rng.read(1)[0] >> 4) & 3]

    sortflag = ["Global table is sorted",
                "Color table is sorted"][rng.read(1)[0] & 1]

    interlacemode = {"None": False,
                     "Adam7": True}[["None",
                                       "Adam7",
                                      ][rng.read(1)[0] & 1]]

    data = []
    for rownum in range(height):
        lzwmincodesize = rng.read(1)[0] + 1


        subblocksizebits = min(lzwmincodesize+1,
                               max(((screenwidth + 15) // 16)))
        
        subblocksize = 2**subblocksizebits - 2
        firstcodeinsubblock = 2**(subblocksizebits-1)

        codebitlength = dict([(i, [])
                              for i in range(firstcodeinsubblock,
                                             firstcodeinsubblock+subblocksize)])
        
        chunkheader = []
        while len(chunkheader) < 1 or \
              sum(len(c) for c in chunkheader) > 255:
        
            chunkheader += [rng.read(1)[0].to_bytes(
                1, byteorder='big')]
            

        subblocks = dict()
        temp = []
        while len(temp) <= 255:
            zigzagdistance = rng.read(1)[0] 
            code = zigzagdistance // 16

            nonzerobits = zigzagdistance % 16
            codebitlength[code] = [*sorted({*reversed(codebitlength[code]),
                                            nonzerobits})]


            if zigzagdistance!= 0:
                temp += [zigzagdistance.to_bytes(1,
                                                 byteorder='big')]
            elif len(temp) >= firstcodeinsubblock and \
                 any(sum(map(lambda n: n > 0,
                             codebitlength[code+j]))
                        for j in range(-firstcodeinsubblock,
                                       1)):

                chunkheader += [struct.pack('>BB',
                                           *[*reversed(nonzerobits)],
                                           # Only ever one bit set here
                                           len(temp))]
                
                subblocks[len(temp)+1] = list(reversed(temp))
            
                temp = []
            
        codetable = dict()
        codesdict = dict()
        for code, codeinfo in enumerate(codebitlength):
            if len(codeinfo) == 0:
                continue

            codesdict[code] = (*zip(*{bitlength: index
                                     for bitlength, index
                                     in zip(*enumerate(
                                         reversed(codeinfo)))
                                     }.items()))

            codetable[code] = tuple(
                    (int.from_bytes(
                        chunkheader[-1][:codesdict[code]][index],
                        byteorder="big"),)
                    + (tuple(index << (n * 8)
                            for n, index
                            in zip(*codesdict[code])))
                        for index in range(2**codesdict[code][0][0]-1,
                                          -(1<<(8*(codesdict[code][0][0]+1))),
                                          -1))


        def get_next_code():
            return next(k
                       for k, v in codetable.items()
                           if len(v) > 0)

        
        currentval = 0
        currentcodesize = None
        currentcode = None
        while 1:
            if currentcodesize is None:
                currentcodesize = rng.read(1)[0] + 1
            
            if currentcode is None:
                currentcode = get_next_code()
                
            for _ in range(currentcodesize):
                yield from currentcode.to_bytes(1,
                                                byteorder='big')

                currentcode = next(codetable[currentcode].__iter__()
                                   )

                
    pass
    