#!/usr/bin/env python3

import io
import os
import platform
import random
import resource
import struct
from typing import BinaryIO


class RandomGifGenerator:

    def __init__(self):

        self._rng = io.BytesIO()

    @property
    def rng(self) -> BinaryIO:
        return self._rng

    def make_entry(
            self, entrysize: int, fileobj: BinaryIO
        ) -> bytes:
        if not isinstance(fileobj, (io.RawIOBase, io.BufferedIOBase)):
            raise TypeError('Not a valid binary IO object.')
        if not hasattr(os, 'urandom'):
            raise RuntimeError("The OS doesn't support urandom.")
        while len(self.rng.getbuffer()) < entrysize:
            line = os.urandom(min((
                max([256 ** i for i in range(7)]), entrysize - len(self.rng)
            )))
            self.rng.write(line)
        
        # If we're generating GIFs on Windows, add some padding to the end of
        # each block, since Windows seems to dislike it when there are extra
        # nulls at the ends of blocks.
        if platform.system() == "Windows":
            
            # The maximum value that can be stored inside an int is (2**31)-1, 
            # so we are safe from overflowing here.
            randint = lambda x, y: random.randint(x, min(y, 2**31-1))

            # Add padding between every 256 bytes and randomize its length.
            pos = [0]
            while True:
                newpos = randint(0, len(self.rng.getvalue()))

                # We don't want all the remaining space to be zeroes or else 
                # Windows Gifs will misbehave, so only append a nonzero byte
                # if the last two bytes aren't already zero.
                if len(self.rng.getvalue()[newpos:]) > 2 \
                        and sum(self.rng.getvalue()[newpos:][:2])!= 0:
                    self.rng.seek(-1, io.SEEK_CUR)
                    self.rng.write(struct.pack('<B', 
                        randint(range(0, 0xff))))
                    break
                
                pos.append(newpos)
            self.rng.seek(pos[-1], io.SEEK_SET)
        
        output = self.rng.read(entrysize)
        assert len(output) == entrysize
        return output
        
    def add_frame(
            self, width: int, height: int, colordepth: int
        ) -> None:
        """Adds a frame containing pixels of the given size."""
        raise NotImplementedError

    
