#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates an animated GIF with the given frames in their respective
# `frames`, which should be of type List[List[int]] where each inner list is a
# pixel row represented as tuple of R, G, B values. The first frame will have to
# be passed to this function as its first element. Every new frame must differ by
# at least one pixel color value from the previous one.
def generate_custom_animated_gif(
        frames: [[(int, int, int)]], out: BinaryIO
    ):
    # Get dimensions of every frame
    widths, heights = [], []
    for frame in frames:
        width, height = max((x for x, _, _ in frame)), len(frame)
        widths.append(width)
        heights.append(height)

    # Start writing everything to output stream
    out.write('GIF89a'.encode())

    # Write global headers to output stream
    out.write(struct.pack('<HHB', min(widths), max(heights), len(frames)))
    out.write("\0".encode() * 3)

    # Write local headers
    for width, height in zip(widths, heights):
        out.write(",".join([str(height), str(width), "\x87", '\0']).encode())

    # Loop through all frames and write them individually
    for i, frame in enumerate(frames):

        # Write separator bytes so that we can later easily patch the sizes when
        # they change per frame
        out.write('\x2c\0\0\0'.encode())

        # Write current frame offset as it could need to be patched
        out.write(','.join(["\0", "0", str(i), "\0"]).encode())

        # Write lzw minimum code size value
        out.write('\x08\0'.encode())

        # Write table and palette immediately after
        out.write('\x00'.encode())

        # Loop until we've written enough bits to fill up the whole image
        for y in range(heights[i]):

            # Check whether the next row exists
            try:
                next_row = frames[i+1][y]
            except IndexError:
                next_row = None

            # If yes, encode difference between rows to save space
            if next_row:

                # Loop through columns
                for x in range(min(widths)):

                    # Encode differences
                    r, g, b = map(lambda x, y: abs(x-y),
                                  frame[y][x], next_row[x])
                    out.write(chr(r << 4 | g))

            else:

                # Otherwise just copy the whole image over again
                for x in range(min(widths)):
                    out.write(chr(frame[y][x]))

