# -*- coding: utf-8 -*-
'''
Random animated images are great for stress testing your application's ability
to render and animate in real time. Animated gifs tend to be smaller than
animated mp4s or webms while still providing useful feedback for the user. This
tool provides an easy way to create a random animated gif from a set of
pregenerated frames.
'''
import argparse
from typing import List 

import numpy as np
import PIL.Image                    # type: ignore

# Returns a sequence of frame numbers [f1, f2,..., fn]
# that make up a cut with length n
def get_cut_frame_sequence(n):
    return list(range(0, n))

# Returns a sequence of frame numbers [f1, f2,..., fn]
# that make up a loop with length n over m iterations.
def get_looping_frame_sequence(m: int, n: int):
    if not len(str(n)) == len(str(m)):
        raise ValueError('Number of frames must have same digits as num loops.')
    seq = []
    
    for i in range(m):
        seq += list(range((i*n), ((i+1)*n)))
        
    return seq

def load_frames(filepaths: List[str],
                start_frame: int=0, end_frame: int=-1,
                skip_every: int=1) -> List[PIL.Image]:
    '''Loads all frames between start_frame and end_frame from each file
    specified by the filepath vector.
    If no filepath is specified, will just read off the first frame in 
    file. The default value for end_frame is -1, which means this function
    will read until there are no more frames left to read.'''
    frames = [PIL.Image.open(filepath)
              for filepath in filepaths]
    sliced_frames = [frame
                     for index, frame in enumerate(frames)
                     if (index >= start_frame
                         and index <= end_frame)]

    if skip_every > 1: 
        skipped_frames = [frame
                          for index, frame in enumerate(sliced_frames)
                          if index % skip_every == 0]

        return skipped_frames

    else:
        return sliced_frames

def merge_frames(frames: List[PIL.Image],
                 duration=None) -> PIL.Image.Image:
    '''Merges a series of frames together according to their durations.
    Assumes each frame in the list contains exactly one color channel only.
    Arguments:
        * frames :: A list of PIL Images representin the images you'd like
                      to merge together.
        * duration :: An optional argument representing the duration of 
                      each frame. Defaults to None. If provided it should
                      be a list containing integers denoting how many ms
                      long the corresponding frame is.
    '''
    width, height = frames[0].size
    result = PIL.Image.new("RGBA", (width, height))
    current = 0
    previous = None

    # Convert each frame into a numpy array
    # then append them together vertically.
    for index, frame in enumerate(frames):
        if duration!= None:
            current += duration[index]
        
        converted = np.asarray(
                        frame.convert('RGBA')).copy()

        if previous == None:
            previous = converted
        else:
            combined = np.vstack([previous, converted])
            
            # Reverse rows so that higher pixels appear on top
            reversed = combined[::-1]

            result.paste(
                    PIL.Image.fromarray(reversed), 
                    mask=converted[..., 3])

            del combined
            del reversed
            
        del converted

    return result
