#!/usr/bin/env python3

import struct
from typing import BinaryIO
from hashlib import md5

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)

    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')

    if ctsize > 0:

        # read in palette values (RGB triples)
        paldata = [struct.pack('BBB', *x) for x in
            [(y % 256, y // 256, y // 256 * 256)
                for y in range(ctsize)]]

        palents = bytes()
        for i in range(len(paldata)):

            # write out the length byte first
            palents += struct.pack('B', len(paldata[i]))
            palents += paldata[i]

    else:
        palents = b''

    # compute hash sum over the rest of the data
    digester = md5()
    while True:
        block = rng.read(16*1024)
        if len(block) == 0: break
        digester.update(block)
    GS = digester.digest()[:3]

    # finally the actual payload starts here
    out.write(b'\x21\xFF\x0B') # application extension
    out.write(b'\x4E\x45\x54\x53\x43\x41\x50\x45\x32\x2E\x30') # NETSCAPE2.0
    out.write(struct.pack('>H', len(GS))) # length field
    out.write(GS)               # identifier
    out.write(struct.pack('B', 3))   # version
    out.write(struct.pack('B', 1))   # flags
    out.write(struct.pack('B', 0))   # delay 0ms

    out.write(b'\x2C\x00\x00\x00\x00\x00,\xF9\x04')
    out.write(palents)
    out.write(struct.pack('H', image_width))
    out.write(struct.pack('H', image_height))

    out.write(b'\x00') # no LZW compression

    # write a bunch of colors from the RNG to create an initial frame
    data = []
    for _ in range(image_width * image_height):
        data.append(int.from_bytes(rng.read(1), "little"))

    # encode these as run-length encoded values
    chunked = {}
    chunk = list()
    prevval = None
    count = 0
    for val in data:

        if prevval!= val or count >= 255:
            k = str(prevval) + ':' + str(count)
            if k in chunked:
                chunked[k].append(chunk)
            else:
                chunked[k] = [chunk]
            chunk = []
            prevval = val
            count = 0

        chunk.append(val)
        count += 1

    k = str(prevval) + ':' + str(count)
    if k in chunked:
        chunked[k].append(chunk)
    else:
        chunked[k] = [chunk]

    c = sorted([len(x) for _, v in chunked.items() for x in v])[-1]
    d = b''
    for vals in chunked.values():
        for i in range(c):
            if i < len(vals):
                d += vals[i]
            else:
                d += b'\x00\x00\x00'

    out.write(d)
