from typing import BinaryIO


def from_uint24le(data):
    # Python integers are unlimited precision, so we have to do this manually
    num = sum((x << (i * 8)) for i, x in enumerate(data[::-1]))

    if num > pow(2, 23):
        return -(pow(2, 24) - num)
    else:
        return num


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    global_color_table_size = rng.read(1)[0] & 7
    width = from_uint24le(rng.read(3))
    height = from_uint24le(rng.read(3))
    flags = rng.read(1)[0] & 15
    background_index = rng.read(1)[0] & 255
    aspect_ratio = rng.read(1)
    background_color = bytes([
        rng.read(1)[0], 
        rng.read(1)[0], 
        rng.read(1)[0]    
    ])
    pixel_block_size = rng.read(1)[0] & 0b11111
    bg_color = None
    
    palette = []
    color_table_size = rng.read(1)[0] & 7

    if color_table_size == 0 and global_color_table_size!= 0:
        color_table_size = global_color_table_size
    
    for _ in range(0, color_table_size):
        red = rng.read(1)[0] & 255
        green = rng.read(1)[0] & 255
        blue = rng.read(1)[0] & 255

        palette += [red, green, blue]
        
        if red == background_color[0] \
           and green == background_color[1] \
           and blue == background_color[2]:
            bg_color = len(palette)-1


    out.write(header)
    out.write(global_color_table_size.to_bytes(1, 'big'))
    out.write(width.to_bytes(3, 'little'))
    out.write(height.to_bytes(3, 'little'))
    out.write(flags.to_bytes(1, 'big'))
    out.write(background_index.to_bytes(1, 'big'))
    out.write(aspect_ratio)
    out.write(background_color)
    out.write(pixel_block_size.to_bytes(1, 'big'))
    out.write(bg_color.to_bytes(1, 'big'))
    out.write(len(palette).to_bytes(1, 'big'))

    try:
        for p in palette:
            out.write(p.to_bytes(1, 'big')*3)

        while True:
            block_type = rng.read(1)[0]
            
            if block_type == ord('!'):
                break

            compressed_subblocks_amount = rng.read(1)[0]            
            subblocks = []

            for _ in range(compressed_subblocks_amount):
                byte = rng.read(1)[0]

                if not ((byte & 0b10000000) >> 7):
                    subblocks.append(
                        byte
                    )
                
                elif ((byte & 0b11111000) >> 3) == 0:
                    subblocks.append(
                        0b11111000 | 
                        rng.read(1)[0]
                    )
                
                elif ((byte & 0b11100000) >> 6) == 1:
                    subblocks.append(
                        rng.read(
                            (byte & 0b11111000) >> 3        
                        ).lstrip(b'\xff')
                    )
                
                else:
                    subblocks.append(
                        rng.read(
                            rng.read(
                                (byte & 0b11111000) >> 3
                            )[0]
                        ).lstrip(b'\xff')
                    )

            pallete_indices = []
            prev_index = 0

            for s in subblocks:
                if isinstance(s, int):
                    prev_index = s
                    pallete_indices.append(prev_index)

                else:
                    length = min(len(s), 255-len(pallete_indices))
                    
                    for i in range(length):
                        pallete_indices.append(prev_index+i+1)

                    if length < len(s):
                        pallete_indices += list(range(0, len(s)-length))
            
            
            lzw_minimum_codeword_size = max(
                2, 
                len(str(len(palette)))//2
            )

            out.write(((lzw_minimum_codeword_size-1)<<4)|0xf)
            out.write((len(palette)-1).to_bytes(1, 'big'))
            out.write(b'\0\0')
            out.write(b'\0'*12)

            for i in pallete_indices[:-1]:
                out.write(i.to_bytes(1, 'big'))


            codewords = {
                k : v for v,k in zip(
                    range(len(palette)+1, len(pallete_indices)+len(palette)+1), 
                    sorted(set(pallete_indices)).reverse()
                )
            }

            out.write(
                b''.join(
                    str(codewords[v]).encode('utf-8').zfill(lzw_minimum_codeword_size)
                    for v in pallete_indices
                )
            )
        
            out.write(b'\0')

    except Exception as e:
        pass