#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header length (13 bytes) plus 2 for width and height
    out.write(header + 2)

    # Width in pixels (1-4 bytes)
    width = rng.read(1)[0] % 64 + 1
    out.write(width)

    # Height in pixels (1-4 bytes)
    height = rng.read(1)[0] % 64 + 1
    out.write(height)

    # Color table flag and size (1-4 bytes). We will always use one color
    # table with its max index set to one (0x7f).
    out.write(b'\xf7\x00\x00\x00')

    # Size of global color table (1-4 bytes). We will also always use one color
    # table so we can skip this field entirely.
    out.write(b'\x00\x00\x00\x00')

    # Background color index (1-4 bytes). This is always zero since we only have
    # one color table entry.
    out.write(b'\x00\x00\x00\x00')

    # Pixel aspect ratio. This is usually 1 or 4/3 which equals to 0x5468.
    out.write(b'\x54\x68')

    # Data subblock (2 bytes)
    datalen = int.from_bytes(rng.read(3), byteorder='big')
    out.write(datalen.to_bytes(2, 'big', signed=False))

    # The actual image pixel values. For each row we repeat the previous pixel
    # value to make sure the encoding uses the LZW compression algorithm as much
    # as possible. In reality all rows should be different even though the
    # encoder may ignore that fact.
    prevpixel = None
    while rowlen:= rng.read(1)[0]:
        if rowlen > width:
            raise ValueError('Row length larger than image width!')

        # If there was no previous pixel then fill it up with zeros
        if not prevpixel:
            prevpixel = b'\x00'*rowlen
        
        # Write the actual row from our pseudorandom stream
        out.write(prevpixel*int((width - len(prevpixel))/len(prevpixel) + 1))

        # Read another row from the pseudorandom stream but store the last
        # element because it's the first next time around
        newpixel = rng.read(1)
        prevpixel = (newpixel + prevpixel[-1:])[1:]
    
    # End of image marker
    out.write(b'\x00')
    
