#!/usr/bin/env python3

import struct
from typing import BinaryIO


class ColorTable():
    """A color map for indexed images."""

    def __init__(self):
        self._colors = []


    @property
    def colors(self):
        return self._colors[:]


    def greyscale(self):
        """Generate Greyscale RGB color values.
        0 -> black
        255 -> white"""

        self._colors.clear()
        for i in range(256):
             self._colors.append(((i >>  5 & 7) << 5,
                                  (i >>  2 & 7) << 5,
                                  (i      & 3) << 6))


    def rgb(self):
        """Generate Red-Green-Blue RGB color values."""

        self._colors.clear()
        for i in range(256):
             self._colors.append((i      ,
                                 i >>  8,
                                ((i >> 6) & 3) << 8))



# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The
# color table can be generated or provided as a list
# of tuples of R, G B values. If a list isn't given, then it will
# construct one based on the cutoff value passed in.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, cmap=None, csizex=2, csizey=2,
        cut=255, bgcolor=(0,0,0), repeat=False
    ):

    header = b'GIF89a'

    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))

    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)
    
    if cmap == None:
        
        # Generate some kind of color map
        color_table = ColorTable()
        color_table.greyscale()
    
        # size of the color table
        ctsize = int.from_bytes(rng.read(1), 'little')
    
        if ctsize > 0:
    
            # read in palette values (RGB triples)
            d = [color_table.colors[(int(y / 256 ** j) % 256)]
                 for j in reversed(range(csizex))
                     for y in range(csizey*256**j)]
    
            palents = bytes()
            for i in range(len(d)):
    
                # write out the length byte first
                palents += struct.pack('B', len(pallets[i]))
                palents += d[i]
    
        else:
            palents = b''
    
        # colormap is followed by a local color table
        # TODO: we could do something with this; but not now...
        rng.seek((3 << 8) | ctsize)
    
        # get the size of the LZW encoded output
        datalen = int.from_bytes(rng.read(2), 'big')
    
    else:
        print("Using custom colour map!")
        pallets = cmap
        palents = bytes([len(pallets)])
        for c in pallets:
            palents += struct.pack('BBB', c[0], c[1], c[2])


    # get the actual data from urandom
    # assumes foreground colour is black?
    fgcol = 0 
    bgcol = 1<<7
    buf = bytes()
    while True:
        val = ord(rng.read(1))
        if val == 0xFF: break
        elif val <= cut: 
            if val == fgcol: 
                val |= 1
            elif val == bgcol: 
                pass
            else: 
                val |= 1|1<<7
            buf += struct.pack('B',val) 
        elif val >= 192: 
            linebuf = bytes()        
            nrept = (((val&7) + 3) << 8)|ord(rng.read(1))
            while len(linebuf) < nrept:
                nextbyte = ord(rng.read(1))
                offset = 0
                while offset < 8:
                    if nextbyte&(1<<(7-offset)):
                        linebuf += struct.pack('B',
                            ord(rng.read(1)) + 
                            (val >> 4) * 256 + 
                            (nextbyte>>offset)*256**(val&7)) 
                    offset+=1
            buf += linebuf
        else:
            raise RuntimeError(f"I don't know how to handle {hex(val)}")
        
    # compute the logical screen descriptor
    lsd = struct.pack('<BBHBBBB',
        image_width % 256, image_width//256,
        image_height % 256, image_height//256,
        8,    # packed field
        7,    # default colourmap
        0,    # sorted flag
        ctsize     )
    lsdbody = (bgcol << 4 | 0xF).to_bytes(1,'big')+palents
    
    if repeat:
        lsdbody += struct.pack("<BBB",0,image_width%256,image_width//256)
        lsdbody += struct.pack("<BBB",0,image_height%256,image_height//256)
    
    lsd += lsdbody + bytes([datalen//256,(datalen)%256])+b'\0\0'
    
    # add the various structures together and write them to disk
    out.write( header + lsd + buf)