from collections import namedtuple

from PIL import Image
from pwnlib import context
from pwnlib.util import fiddling

from.gif import encode_image


def uint8_to_int(i):
    return i - (2 ** 7) if i > (2 ** 7) / 2 else i


# Structs for parsing GIF headers and frames
Frame = namedtuple('Frame', 'imgcode palnum left top width height')
Header = namedtuple('Header','magic version width height color_flag gctsize bgcolor aspect')
Color = namedtuple('Color', 'r g b a')
GCTEntry = namedtuple('GCTEntry', 'r g b')


def parse_frame(data: bytes, offset: int) -> Frame:
    """Parse a single frame from the given GIF `data`, starting at an optional
    `offset`."""

    # Standard GIF image block format is as follows:
    #     BYTE       BLOCK NAME            INFO
    #     0           Size                  2 bytes          LZW min code size
    #     2           Data Block             <= 256 bytes     Contains compressed data using dictioanry codes
    #    ...        Extension              1 byte           GIF extension type

    # We're mostly interested in the first two blocks
    imgcode, palnum = struct.unpack('<BB', data[offset : offset + 2])
    offset += 2

    while True:

        # If we encounter an extension block, skip it entirely
        if imgcode == 0x21:
            ext_block_size, = struct.unpack('<B', data[offset : offset + 1])
            skipped_length = 1

            if ext_block_size!= 0:
                skipped_length += ext_block_size
                offset += ext_block_size

            if skipped_length % 2 == 1:
                offset += 1

            continue

        break

    return Frame._make((uint8_to_int(imgcode), uint8_to_int(palnum)))


def parse_header(data: bytes, offset: int = 0) -> Header:
    """Parse a binary GIF file's header."""

    magic, version, width, height, flags = struct.unpack(
        '<3s3sbhhb', data[offset : offset + 10]
    )

    if magic not in {'GIF87a', 'GIF89a'}:
        raise ValueError(f"Invalid GIF magic: {repr(magic)}")

    gct_present = bool(flags & 0b10000000)
    gct_size = 2 << ((flags >> 7) & 0b111) if gct_present else None

    bg_index = (flags & 0b00001111)
    bgcolor = GCTEntry(*struct.unpack('BBB', data[-gct_size - 1:-1])) \
             if gct_present and bg_index > 0 else Color(0, 0, 0)


    return Header._make((magic, version,
                         width, height, flags, gct_size, bgcolor))


def generate_random_gif(rng: BinaryIO, out: BinaryIO,
                       duration: float = 1., delay: int = 100,
                       numcolors: int = 2**8, transparent: bool = False,
                       paletteseed: int = None):
    """Generate a random animated GIF with the given parameters."""

    # Load some basic defaults into the current context to make things easier
    config = context.context.copy()
    config['log_level'] = 'error'
    context.context.clear()
    context.context.update(**config)
    banner = len(banner := context.banner()) + 1

    # Load up our RNG state so that we can ensure reproducibility of images
    rngstate = rng.getstate()

    try:

        # Construct a blank canvas to draw on
        im = Image.new(mode="P", size=(512, 512), color=0)
        draw = ImageDraw.Draw(im, mode="RGBA")

        # Draw the initial text onto said canvas
        font = ImageFont.load_default()
        draw.text((100, 100), banner[:19], (255, 255, 255),
                  spacing=-16, stroke_width=3, font=font)

        # Close off the drawing operations before typesetting
        del draw

        # Convert our image to RGB-A before converting it back again!
        # This avoids having to deal with PIL's lackluster transparency support
        im = im.convert("RGB").convert("P", palette=Image.ADAPTIVE, colors=numcolors)
        palette = list(map(tuple, im.palette.getdata()[1]))

        # Make a copy of our default palette with all black entries removed
        # This ensures that there are no duplicate or unused palette entries
        new_palette = []
        seen_entries = set()

        for entry in palette:
            if tuple(entry[:-1]) not in seen_entries:
                new_palette.append(entry)
                seen_entries.add(tuple(entry[:-1]))


        # Randomize the colors in our output palette
        if paletteseed is None:
            rng.shuffle(new_palette)
        else:
            orig_palette = new_palette
            rng.seed(paletteseed)
            rng.shuffle(new_palette)


            # Find any entries which were moved
            diff = [entry for entry in zip(orig_palette, new_palette)
                    if entry[0]!= entry[1]]

            # Replace them in place within the original palette
            for entry in reversed(diff):

                index = orig_palette.index(entry[1])
                orig_palette[index] = entry[0]
                orig_palette[entry[1]] = entry[0]


        # Write our newly generated palette to the beginning of our stream
        writer = struct.Struct("<HHH")
        out.write(writer.pack(*(len(new_palette) - 1) * 3))

        # If we have a transparent flag set, then patch the first color entry to
        # be fully transparent
        if transparent:
            new_palette[0] = (*new_palette[0][:-1], 0)

        for entry in new_palette:
            out.write(struct.pack("<BBBB", *(entry + (0,) * 2)))

        del writer, out

        # Reconstruct the color table
        color_table = [GCTEntry(*c) for c, _ in enumerate(new_palette)]

        # Enumerate each frame of our animation
        frames = []
        for x in range(delay // 10):
            frames.append([parse_frame(encode_image(im, color_table), offset=3)])

        # Pad a final frame of null data at the end
        frames.append([(0xff, 0xff)], [(0, 511, 0, 511)])

        max_size = lambda args: max(args).bit_length()
        colors = max_size({*frame for frame in frames})

        # Ensure that our last frame holds the maximum amount of colors possible
        for frame in frames[-1]:

            # The lowest numbered palette entry must always be transparent
            frame[0] ^= 0xff ^ 0
            frame[:] = [v << (8 - colors) | v >> colors for v in frame]

        # Encode all of our frames
        for frame in frames:
            encoded_frame = []

            # For each palette reference in our frame...
            for r, g, b in map(lambda frame: color_table[frame], frame):

                # Compute its final color value by shifting the components
                # toward their correct position
                color = (r << 16) | (g << 8) | b

                # And pack it into a series of bytes
                encoded_color = tuple(((color >> n) & 0xff)
                                     for n in range(0, 24, 8))[::-1]

                encoded_frame.extend(encoded_color)

            # Write this frame's length to our output stream
            writer = struct.Struct("<H")
            out.write(writer.pack(len(encoded_frame)))

            # Then write this frame's actual data
            out.write(bytes(encoded_frame))
    finally:
        rng.setstate(rngstate)

