import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Header
    version = b'\x47\x49\x46\x38\x39\x61'
    flags   = b'\x1f' # Color table defined, sorted, color resolution == 8 bit, no global palette
    width   = rng.read(3)
    height  = rng.read(3)
    out.write(version + flags + width + height)

    # LZW minimum code size
    lzw_mincodesize = rng.read(1)[0]
    out.write(lzw_mincodesize.to_bytes(1))

    # Global color table
    gctlen = rng.read(1)[0] & 0xf
    gct = []
    for _ in range(gctlen):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        gct += bytes([r, g, b])
    out.write(gctlen.to_bytes(1, 'little'))
    out.write(gct)

    # Graphics control extension
    has_ext = bool(rng.read(1)[0] % 2)
    ext_label = b'\x21\xf9\x04' if has_ext else b''
    ext_block = bytes()
    if has_ext:
        ext_block += b'\x00'    # Reserved
        repetitions = rng.read(3)[::-1]
        ext_block += bytes([repetitions[i] | ((repetitions[i+1] << 8) if i < 2 else 0)
                            for i in range(3)])

        ext_block += b'\x00'    # Transparent index
        ext_block += bytes([rng.read(1)[0], rng.read(1)[0]])
        out.write(ext_label + ext_block)


    # Image descriptor
    left      = rng.read(2)
    top       = rng.read(2)
    w         = rng.read(2)
    h         = rng.read(2)
    local_map = bytes([rng.read(1)[0]])

    out.write(left + top + w + h + local_map)

    # Table based image data
    idx = 0; block = 0
    while True:
        bit = min((8 - block) * 8, len(data)-idx)
        blockbyte = 0
        for i in reversed(range(bit)):
            blockbyte |= (data[idx+i//8] >> (7-i%8)) & 1 << (i%8)
        idx += bit // 8
        block += bit // 8
        if idx >= len(data): idx -= len(data); break
        out.write(struct.pack('B', blockbyte))

    # Padding
    out.write(b'\x00' * (8 - block))

