#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, nframes=None
    ):
    
    if nframes == None:
        nframes = struct.unpack('<H', rng.read(2))[0]
    
    header = b'GIF89a'
    
    # width and height are 16 bits each
    image_width, image_height = \
        struct.unpack('<HH', rng.read(4))
        
    
    # skip fields that aren't particularly interesting
    # (the logical screen descriptor - color table size, background
    # color index, pixel aspect ratio, global color table flag,
    # color resolution, sort flag, size of global color table.)
    rng.seek(7)
    
    # size of the color table
    ctsize = int.from_bytes(rng.read(1), 'little')
        
    if ctsize > 0:
        
        # allocate space for the palette
        paldata = []
        for i in range(ctsize):
            
            # store the rgb tuple as the index
            paldata.append(int(i / 256))
            
        # write out the palette
        for v in paldata: 
            out.write(struct.pack('BBB', *(v % 256, v // 256, v // 256*256)))
    
        # seek to end of file:
        #
        #   - current position
        #   - current seek position inside file
        #   - size of file minus current position
        pos = out.tell()
        currseekpos = out.seek(0, os.SEEK_CUR)
        rem = out.seek(0, os.SEEK_END) - pos
        out.seek(currseekpos)
        
        # set up application extension block which will reference it
        applens = len(paldata).to_bytes(1, 'little', signed=False) +\
                    chr(b'\x0b').encode("utf-8") + b'\x00\x0b\x00\xff'
                    
        out.write(applens)
        out.write(b'\x04NETSCAPE2.0')
        out.write(b'\x00\x03')
        out.write(b'\x01\x00\x00\x00')
        
        # write null bytes until there's enough padding so the next frame can fit
        while len(out.getbuffer()) < ((rem >> 1) - 6):
            out.write(b'\x00')
        
        # write out the loop count footer - zero means infinite looping
        out.write(b'\x00\x00\xff\xfe')
    
    else:
        palents = b''
    
    # colormap is followed by a local color table
    # TODO: we could do something with this; but not now...
    rng.seek((3 << 8) | ctsize)
    
    datalen = int.from_bytes(rng.read(2), 'big')
    data = rng.read(datalen)
    
    for _ in range(nframes):
        out.write(header + data)