#!/usr/bin/env python3

import struct
from typing import BinaryIO
import os


class RandomState:

    def __init__(self):
        self.rng = open('/dev/urandom', 'rb')

    # Return an integer in [0..n-1]
    def nextInt(self, n):
        return int.from_bytes(
                self.rng.read(int((n).bit_length()//8)), byteorder='big') % n


# Generate the color palette (256 colors)
def pallete():
    p = []
    for i in range(768 // 3):
        r = g = b = 0
        while not ((r!= 0 and g!= 0 and b!= 0) or
                   (r == 0xFF and g == 0xFF and b == 0xFF)):
            r = state.nextInt(256)
            g = state.nextInt(256)
            b = state.nextInt(256)
        p += [b, g, r]
    return bytes(p)


# Generates a random GIF file into `out` using the
# random number generator `/dev/urandom`
def generate_random_gif(
        out: BinaryIO, width=42, height=42, maxframes=181, bgcolor=(0xDE, 0xAD, 0xBE), transparence=True
    ):
    global state

    wbytes = width.to_bytes(2, "little")
    hbytes = height.to_bytes(2, "little")
    
    emptybyte = b'\x00'
    if bgcolor[3]:
        bgcolorstr = f"#{bgcolor[0]:X}{bgcolor[1]:X}{bgcolor[2]:X}"
        bgcolorcode = f"\x{bgcolor[3]:X}".encode("utf8")
    else:
        bgcolorstr = "... transparent..."
        bgcolorcode = emptybyte * 3

    pal = pallete()
    state = RandomState()

    out.write(b"GIF89a")
    out.write(wbytes)
    out.write(hbytes)
    out.write(b"\xF7\x00")        # No global color table
    out.write(emptybyte*3)       # Background color
    out.write(b"\x00")           # Pixel aspect ratio
    out.write(pal[:768])         # Color table

    # Write frames
    for frame in range(maxframes):

        # Frame duration
        out.write(state.nextInt(256).to_bytes(1, 'big'))
        out.write(state.nextInt(256).to_bytes(1, 'big'))

        # Image descriptor block
        out.write(b"\x2C")                       # Image separator
        xoffset = random.randint(0, width - 1)   # Position X
        yoffset = random.randint(0, height - 1)  # Position Y
        out.write(struct.pack("<HHH", xoffset, yoffset, 1))    # Offset & size

        lzwsize = state.nextInt(width * height // 2)     # LZW min code size
        out.write(((lzwsize + 1) | 0xE0).to_bytes(1, 'big'))   # LZW min code size

        # Write subblocks
        endofblock = False
        while not endofblock:

            # Block size
            numbytes = state.nextInt(128)            # Number of bytes to write
            if numbytes < 255:                      # Ending subblock?
                out.write(numbytes.to_bytes(1, 'big'))      # Yes!
                endofblock = True                     # Ok we are done
            elif numbytes > 255:                    # Next subblock is another one
                out.write(255.to_bytes(1, 'big'))          # Mark it as such
                out.write(numbytes.to_bytes(1, 'big'))      # Size of the following block
            else:                                   # Weird case but possible
                raise Exception('Unexpected error during image generation')

            # Block itself
            for _ in range(numbytes):                 # For all the bytes to write :
                codelen = state.nextInt(17)              # Compute a length of a new code

                # The first bit determines whether our code will be used
                # as a literal value or as a reference to something already written
                copyflag = bool(state.nextInt(1))
                
                # If it's a usefull value get its prefix from the previous codes
                if copyflag:
                    prevlen = state.nextInt(codelen) + 1
                    code = list(prevcodes[-prevlen:])
                else:
                    code = [state.nextInt(256)]             # Else: just getting a random byte

                # Add this new code to the list
                prevcodes.append(code[0])                  # Add it to the previous codes

                # Write the code (or the reference) on the output stream
                out.write(bs.pack('B', len(code)-1))                # Length of the encoded symbol
                out.write(bs.pack('%dB' % len(code), *code))        # The encoded symbol

            

    # Write final image
    out.write(b";")