#!/usr/bin/env python3

import struct
from sys import stderr
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b"GIF89a")

    # Logical screen descriptor
    width  = _generate_random_int(rng, (1 <<  4))
    height = _generate_random_int(rng, (1 <<  4))
    flags  = _generate_random_int(rng, (1 <<  1))
    bg     = _generate_random_int(rng, (1 <<  8))
    fg     = _generate_random_int(rng, (1 <<  8))
    
    gctsize = _generate_random_int(rng, (1 <<  3))
    gctbpp  = _generate_random_int(rng, (1 <<  3))
    gctsort = _generate_random_int(rng, (1 <<  1))
    gctlen  = _generate_random_int(rng, ((1 << 8) - 1))
    gct = bytes()
    for i in range(gctlen):
        r = _generate_random_int(rng, (1 <<  8))
        g = _generate_random_int(rng, (1 <<  8))
        b = _generate_random_int(rng, (1 <<  8))
        gct += struct.pack("BBB", r, g, b)

    out.write(struct.pack(">HHBBB",
            width, height,
            flags, bg, fg) + \
            struct.pack("<BHBHBH",
                gctsize, gctbpp, gctsort,
                gctlen, 0) 
    )

    # Global color table
    pass

    # Image descriptor
    left       = _generate_random_int(rng, (1 <<  4))
    top        = _generate_random_int(rng, (1 <<  4))
    width      = _generate_random_int(rng, (1 <<  4))
    height     = _generate_random_int(rng, (1 <<  4))
    lflags     = _generate_random_int(rng, (1 <<  1))
    interlace  = _generate_random_int(rng, (1 <<  1))
    sortflag   = _generate_random_int(rng, (1 <<  1))
    localcolor = _generate_random_int(rng, (1 <<  3))
    nlct       = _generate_random_int(rng, ((1 << 8) - 1))
    nbg        = _generate_random_int(rng, ((1 << 8) - 1))
    nfg        = _generate_random_int(rng, ((1 << 8) - 1))

    # Local color table
    lct = bytes()
    lbg = ''
    lfg = ''
    for i in range(nlct):
        r = _generate_random_int(rng, (1 <<  8))
        g = _generate_random_int(rng, (1 <<  8))
        b = _generate_random_int(rng, (1 <<  8))
        lct += struct.pack("BBB", r, g, b)

        if i == nbg:
            lbg = f"\nBG = {r} {g} {b}"
        
        elif i == nfg:
            lfg = f"\nFG = {r} {g} {b}"
    
    print(f"{lct}{lbg}{lfg}", end='', file=stderr)
    out.write(struct.pack(">HHBBBBBB",
            left, top, width, height, 
            lflags, interlace, sortflag, localcolor) + \
            struct.pack("<H", nlct) + \
            lct[localcolor*3:(localcolor+1)*3] 
    )

    # Graphics control extension
    flag       = _generate_random_int(rng, (1 << 1))
    delaytime  = _generate_random_int(rng, ((1 << 8) - 1))
    transparentindex = _generate_random_int(rng, ((1 << 8) - 1))
    disposal   = _generate_random_int(rng, (1 << 3))
    userinput  = _generate_random_int(rng, (1 << 1))

    out.write(struct.pack(">BBHHB",
            0xF9, 0x04,
            delaytime, transparentindex, disposal) + \
            struct.pack("<BB", 
            0, flag)
    )


    # Plain text extension
    index      = _generate_random_int(rng, ((1 << 8) - 1))
    sizeX      = _generate_random_int(rng, ((1 << 8) - 1))
    sizeY      = _generate_random_int(rng, ((1 << 8) - 1))
    posX       = _generate_random_int(rng, ((1 << 8) - 1))
    posY       = _generate_random_int(rng, ((1 << 8) - 1))
    cellwidth  = _generate_random_int(rng, ((1 << 8) - 1))
    cellheight = _generate_random_int(rng, ((1 << 8) - 1))
    foreR      = _generate_random_int(rng, (1 << 8))
    foreG      = _generate_random_int(rng, (1 << 8))
    foreB      = _generate_random_int(rng, (1 << 8))
    backR      = _generate_random_int(rng, (1 << 8))
    backG      = _generate_random_int(rng, (1 << 8))
    backB      = _generate_random_int(rng, (1 << 8))

    out.write(struct.pack(">BBHHHHBBBBBBB",
            0xFE, 0x01,
            index, sizeX, sizeY, posX, posY,
            cellwidth, cellheight,
            foreR, foreG, foreB, backR, backG, backB
        ))

    # Application extension
    appidlen = _generate_random_int(rng, ((1 << 8) - 1))
    authcode = ''.join([chr(_generate_random_int(rng, (1 << 8))) for n in range(8)])
    applen = _generate_random_int(rng, ((1 << 8) - 1))
    appname = ''.join([chr(_generate_random_int(rng, (1 << 8))) for n in range(applen)])

    out.write(struct.pack(f">BB{appidlen}sBB{applen}s",
            0xFF, 0xFF,
            authcode.encode(),
            0,
            appname.encode()))

    # Comment extension
    commentlen = _generate_random_int(rng, ((1 << 8) - 1))
    comment = ''.join([chr(_generate_random_int(rng, (1 << 8))) for n in range(commentlen)])

    out.write(struct.pack(f">BB{commentlen}s",
            0xFF, 0xFE,
            comment.encode())
    )

    # Trailer
    out.write(bytes([0x3B]))

def _generate_random_int(rng: BinaryIO, modulus: int) -> int:
    return struct.unpack('<I', rng.read(4))[0] % modulus

