#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the LICENSE file in
# the root directory of this source tree.

"""
Generate a random gif for testing purposes. The GIF has no animation, just an
image frame.
The amount of bytes used by the image can be controlled using the'size' parameter
(defaults to ~25kB).
A seed for the RNG can optionally be specified as well ('seed'). If none is given,
the current time will be used instead.
Author: <NAME> <<EMAIL>>
"""

import argparse
import datetime
import os
import sys
import zlib


def write_header(out):
    """Write the GIF file header."""

    # File header:
    #   - Version (always "GIF87a" or "GIF89a") => always "GIF89a" since we do not use extensions.
    #   - Logical Screen Descriptor block:
    #       - Width (2 bytes, LSB first)
    #           - 0x0e20: 320px
    #       - Height (2 bytes, LSB first)
    #           - 0x0e20: 320px
    #       - Global Color Table Flag set to 0.
    #       - Background color index (= transparent background) => set to default value (-1 == last entry).
    #       - Pixel aspect ratio (block terminator, always 0x00)
    out.write(b"\x47\x49\x46\x38\x39\x61\x0e\x20\x0e\x20\x00\xff\xff\x00\x00\x00")


def write_logical_screen(out, width, height):
    """Write the logical screen descriptor block."""

    # See https://www.w3.org/Graphics/GIF/spec-gif89a.txt section 10.
    # Logical Screen Descriptor block:
    #   - Width (2 bytes, LSB first)
    out.write((width % 0x10000).to_bytes(2, "little"))
    #     - 0x0e20: 320px
    out.write(((height % 0x10000) << 8 | ((width >> 8) & 0x00ff)).to_bytes(2, "little"))
    #   - Height (2 bytes, LSB first)
    #     - 0x0e20: 320px
    #   - Global Color Table Flag set to 0.
    #   - Background color index (= transparent background) => set to default value (-1 == last entry).
    out.write(b"\x00\xff\xff")
    #   - Pixel aspect ratio (block terminator, always 0x00)
    out.write(b"\x00")


def write_global_color_table(out, num_entries, size):
    """Write the global color table with ``num_entries`` entries of ``size``.
    All colors are initialized randomly."""

    # Each color takes up two bytes so there should be enough bytes left after writing all entries.
    assert len(out.peek(2*num_entries)) >= 2*num_entries

    # Initialize all colors to 0xAABBGGRR where AA is the alpha channel which is always 0x00, BB,
    # GG, RR are random values between [0x00, 0xFF]
    def create_random_color():
        return bytes([
            0x00,
            rng.randint(0x00, 0xFF),
            rng.randint(0x00, 0xFF),
            rng.randint(0x00, 0xFF)
        ])

    # Write the color table to the output stream.
    out.writelines(create_random_color()
                   for _ in range(num_entries))


def write_frame(out, width, height, compressed_data):
    """Write a single frame into the output stream describing a single-pixel transparent image."""
    assert width <= 0xFFFF
    assert height <= 0xFFFF

    # Frame header:
    #   - Image separator (0x2C)
    #   - Left position (2 bytes, LSB first)
    #       - 0x0000: Place image at top left corner.
    #   - Top position (2 bytes, LSB first)
    #       - Same as above.
    #   - Width of the image (2 bytes, LSB first)
    #       - Set to match the previously defined window width.
    #   - Height of the image (2 bytes, LSB first)
    #       - Set to match the previously defined window height.
    #   - Flags:
    #       - Local color table flag set to 0.
    #       - Interlace mode (bits 6..7) set to 00.
    #       - Sorting method (bit 5) set to 0.
    #       - Reserved bits (bits 2..4) set to 000.
    #       - Size of local color table (bits 1..3) set to 000.
    #       - Encoded fields size (1 byte, always 0x03)
    out.write(b"\x2C\x00\x00\x00"
              + ((width % 0x10000).to_bytes(2, "little")
                 + ((height % 0x10000) << 8 |
                    ((width >> 8) & 0x00ff))).to_bytes(2, "little")
              + b"\x00\x00\x00\x00\x80")
    #   - Frame data sub-block:
    #       - Extension introducer (0x21)
    #       - Graphics control extension label (0xF9)
    #       - Block size (always 0x04)
    #       - Packed field:
    #           - Disposal method (bits 2..4) set to 000.
    #               - 000: No disposal specified.
    #           - User input flag (bit 0) set to 0.
    #           - Transparent color flag set to 1.
    #           - Delay time (2 bytes, little endian, LSB first.)
    #               - 0x0000: Don't change the delay time on display.
    #       - Transparent index (1 byte, must be the same as the one in the main header's palette.)
    out.write(b"\x21\xf9\x04\x00\x01\x00\x00\xff\xff")
    #   A second control sub-block containing a plain text sub-block that tells the GIF decoder how long it
    #   should keep displaying the image before switching over to displaying the next frame. It is very
    #   important that this sub-block comes BEFORE any other graphics chunks!
    #       - Plain text extension label (0x01)
    #       - Sub-block size (2 bytes, big endian, LSB first)
    #          - 0x000B is the length of the content described below.
    out.write("\x21\x01\x0b".encode("ascii"))
    #       - Text command label (0x1A)
    #       - Block size (always 0x01)
    #       - Character to print after character 0x01 (0x14)
    out.write(b"\x01\x14")
    #   - Application extension:
    #       - Extension Introducer (0x21)
    #       - Application Extension Label (0xFF)
    #       - Block size (always 0x0B)
    #       - Application Identifier (ASCII string "NETSCAPE" padded to 8 characters)
    appext = b"\x21\xFF\x0B\x4E\x45\x54\x53\x43\x41\x50\x45"
    def pad_to_eight_chars(s):
        if len(s) > 8:
            raise ValueError('String "' + s + '" is too long.')
        else:
            return s.ljust(8, chr(0))
    out.write(appext[:8])
    #       - Application Authentication Code (ASCII string "\x03\xE8", ie the packed integer 0x03E8h)
    out.write(struct.pack('<H', 0x03E8))
    #   - End of extension block (0x00)
    out.write(b"\x00")
    #   - Graphic Control Extension block:
    #       - Extension introducer (0x21)
    #       - Graphics control extension label (0xF9)
    #       - Block size (always 0x04)
    #       - Packed field:
    #           - Reserved