#!/usr/bin/env python3

import struct
from typing import BinaryIO

def main():
    with open('/dev/urandom', 'rb') as rng,\
         open('generated-random.gif', 'wb') as out:
        generate_random_gif(rng, out)


# A list of block lengths to use in our lzw decoder.
lzw_lengths = [
    256,    # The first code size is always 257 entries long (for all intents and purposes.)
    10, 52, 17, 68, 34, 132, 66,  5, 137, 14,  50,
     25, 120, 24, 8, 49, 16, 86, 48, 27, 129,  3,
      67, 51, 99, 104, 23, 81,  6, 130,  77, 74,
    79, 26, 125, 38, 116, 100, 13, 106, 69, 87,
     35, 149, 64, 84, 101, 111, 30, 136, 133, 62,
     12, 29, 95, 15, 96, 126, 59, 19, 139, 92,
     36, 107, 117, 53, 114, 91, 21, 90, 54, 122,
     93, 108, 119, 109, 46, 70, 127, 112, 105, 28,
     143, 146, 144, 80, 138, 37, 58, 60, 72, 102,
    140, 75, 115, 147, 142, 148, 152, 151, 55,  57,
    154, 94, 22, 45, 88, 155, 83, 118, 141, 156, 131,  7, 135, 76, 124, 150, 134, 98,
     128, 44, 113, 103, 47, 78, 1, 97, 41, 89, 172,
    18, 20, 23, 160, 166, 167, 174, 169, 173, 211, 195,
    199, 203, 212, 198, 190, 184, 215, 157, 159, 218, 207,
    205, 219, 171, 216, 168, 170, 185, 189, 165, 164, 204,
    209, 162, 163, 194, 161, 192, 196, 210, 206, 213, 200,
    202, 208, 214, 191, 197, 186, 187, 181, 182, 179, 180,
    177, 178]


# Generate a LZW encoded gif by reading from `in`, writing to `out`.
def generate_gif_lzw(
        in_: BinaryIO, out: BinaryIO
    ) -> None:

    # Read in bytes until we have enough for a full header.
    while True:

        # Read another byte from the input stream.
        ch = next(in_)

        if ch == ord(','):
            break

        assert isinstance(ch, int), type(ch).__name__


    # Now that we know how many frames there are, write it back out.
    w, h = struct.unpack("<HH", in_.read(4))
    print("w x h:", w, "x", h)
    out.write(struct.pack('<6B', *b"GIF89a"))

    # Write out width and height
    out.write(struct.pack("<HH", w,h))

    # Write out color table
    out.write(bytes([0]))


    # Write out image descriptor
    local_color_table_flag = False
    interlace_flag = False
    sort_flag = False
    reserved = 0
    lct_size = 0
    lzw_min_code_size = max(max(lzw_lengths) - 1, 2)
    out.write(struct.pack("<BBBBBHB", *((local_color_table_flag << 7)|
                                          (interlace_flag << 6)|
                                          (sort_flag << 5)|
                                          reserved,
                                          0,
                                            0,
                                              w,
                                            h,
                                               lzw_min_code_size,
                                              lct_size)))

    def output_prefix(initial):
        out.write(bytes([initial]))

    def output_suffix(c):
        assert c >= 1
        out.write(bytes([(c >> 8)|((c & 255)<<8)]))

    curr_chunk = []
    used_codes = set()
    initial = 0
    prefix = ''

    # Read the rest of this frame's data. We do this by
    #  - splitting incoming chunks up into individual codes (with their length prefixes removed).
    #  - Decoding each code, and emitting its corresponding value.
    for chunk in iter(lambda: in_.read(256), b''):
        prev_chunk = curr_chunk
        curr_chunk = []

        for i in range(len(prev_chunk)):

            # Make sure we're given a big endian uint16.
            two_byte_code = prev_chunk[i] | ((prev_chunk[i+1])<<8)

            # Decode the code.
            codelength = len(curr_chunk)
            decoded = lzw_decode(two_byte_code, codelength=codelength+1)


            # Emit any new values.
            #print(f"decoded {decoded} ({two_byte_code})")
            
            if decoded!= '':
                for char in decoded:
                    output_suffix(char)

                used_codes |= set(decoded)
                
            else:
                
                # It was an empty string, so let's reset to the previous prefix
                #  which has been emitted.
                assert len(prefix) > 0
                output_suffix(int(prefix[-1]))

            # Grab the last character of the sequence. This will be used as 
            # part of the prefix on the next iteration.
            if decoded!= '':
                prefix += str(decoded[-1])

        # Flush remaining data onto the current chunk.
        curr_chunk += chunk


    # Set gct size based on how many unique colors were used.
    num_colors = len(used_codes)
    lct_size = min(num_colors, 256)
    
    # Reset the writer pointer to before the color table entry.
    pos = out.tell()
    out.seek(-1, 1)
    out.write(bytes([lct_size]))
    out.seek(pos, 0)

    # Write out the actual color table entries.
    for i in range(0, lct_size*3, 3):
        out.write(bytes(((i//3 % 256),)*(i // 3 < num_colors)))


# Given a single integer representing a variable length code, decode
# it according to our hard coded table.
def lzw_decode(
        n: int, *, 
        codelength: int = 0
    ) -> str:
    """Decode a single lzw code."""

    
    global lzw_lengths

    # If the codelength hasn't been specified explicitly, try to infer it from
    # the input number itself.
    if not codelength:
        
        # Use the highest bit position which isn't set.
        for i in reversed(range(n.bit_length())):
          if ~n & (1<<(i)):
              codelength = i + 1
              break
              
        # If none is found, then just default to the maximum allowed codelength.
        if not codelength:
            codelength = 