#!/usr/bin/python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out`. The output will contain an animation
# with three frames. Each frame is 64x64 pixels in size and contains one color
# chosen randomly from the set {black, grey, white}. The GCT for this animation
# has only two entries, so it's optimized to use just these two colors instead
# of all possible ones.  This method uses `/dev/urandom` as its source of
# entropy.
def generate_random_gif(out: BinaryIO):
    frame_data = []

    rng = open('/dev/urandom', mode="rb")
    
    black_color = struct.pack("HH", 0, 0)
    grey_color  = struct.pack("HHH", 0b111110000,
                               0b11111100,
                               0b11111111)
    white_color = struct.pack("HHH", 0b111111000,
                               0b111111100,
                               0b111111111)
    palette     = [black_color,
                   grey_color]
    gct         = len(palette) * 3
    lzwminsize  = 2
    lzwmaxsize  = 2**lzwminsize - 1
    mincodesize = 2
    numframes   = 3
    bgcolorindex = 0  # Colors from index zero in global color table
    width       = 64
    height      = 64

    header = b''

    header += b"GIF87a"            # Signature "GIF87a" or "GIF89a".
    header += b"\x00\x0A"          # Version string "87a" or "89a".
    header += struct.pack("<HBBH", 
                            width,
                            height,
                            0b1000,  # Global Color Table Present Flag.
                            bgColorIndex=bgcolorindex)
    header += struct.pack("<BBBHB", 
                            lzwminsize,
                            lzwmaxsize,
                            mincodesize,
                            0b10000000 | gct,  # Size of Global Color Table.
                            numframes)
    _append_data_block(header, out)


    for i in range(numFrames):
        # LZW minimum code size.
        _append_data_block([lzwminsize], out)

        # Graphic Control Extension.
        delay = rng.read(2)
        transparentFlag = False if rng.read(1)[0] == 0 else True 
        disposalMethod = rng.read(1)[0] & 0b111
        subblocks = [disposalMethod,
                     0,
                     *delay,
                     0]
        if transparentFlag:
            subblocks.extend((1,
                              *rng.read(2)))
        _append_graphic_control(subblocks, out)
        
        # Image Descriptor.
        xpos = rng.read(2)
        ypos = rng.read(2)
        w = b"\xFF\xFF" if rng.read(1)[0] == 0 else struct.pack("<B", rng.read(1)[0]+1)
        h = b"\xFF\xFF" if rng.read(1)[0] == 0 else struct.pack("<B", rng.read(1)[0]+1)
        localColorTableSize = rng.read(1)[0] & 0b11111
        interlaceFlag = bool(rng.read(1)[0])
        sortFlag = bool(rng.read(1)[0])
        reserved = rng.read(3)
        localColorTable = [(rng.read(3)+b"\xFF"*3)[i:i+3]
                           for i in range(localColorTableSize*3)]
        _append_image_descriptor(
                xpos, 
                ypos, 
                w, 
                h, 
                0b10000000 | (gct if localColorTable!= [] else 0),
                localColorTable,
                out)

        # Local Color Table.
        _append_data_block(localColorTable, out)

        # Actual image data.
        # Encode data with LZW compression algorithm.
        encoder = LzwEncoder()
        compressedDataBlocks = []
        textSubBlock = b""
        for datum in frame_data[i]:
            chunk = datum.to_bytes(length=1, byteorder="little")

            newTextSubBlock = encoder.addByteToBuffer(chunk)
            if not newTextSubBlock: continue
            
            compressedDataBlocks.append((textSubBlock,
                                         encoder._buildCodetable()))

            textSubBlock = newTextSubBlock
            
        # Terminate last block.
        compressedDataBlocks.append((textSubBlock,
                                     encoder._buildCodetable(),
                                     1))

        # Compress each block separately.
        for encodedDataBlock in compressedDataBlocks:
            _compress_data(*encodedDataBlock, out)
        
    # Append EOF indicator.
    _append_data_block([], out)
    
        
class LzwEncoder():
    def __init__(self):
        self.clear_dictionary()
        self._currentValue = None
        self._firstFreeCodeword = 256
        self._nextAvailableCodeword = self._firstFreeCodeword

    def addByteToBuffer(self, nextByte):
        """ Adds given byte to buffer and compresses it
            accordingly. If necessary adds more dictionary
            entries and resets current value."""
        
        # Add byte to input buffer and increment compressed data length counter.
        self._compressedDataLength += 1;
        self._buffer += nextByte

        # Check whether there was enough information added yet to make a decision.
        if self._currentValue is None or not self._buffer: return None

        # Get codeword of current bytes sequence.
        try: codeword = self._dictionary[self._buffer]
        except KeyError: 
            print("Error!")
            exit(-1)
        
        if codeword >= 0: 
            # Codeword found! We can write it now.
            self._codedOutput += codeword.to_bytes(length=1,
                                                   byteorder="little")
            # Reset buffer and reset current value.
            self._buffer = ""
            self._currentValue = None
            return None
        elif codeword < 0:
            # Not all bytes were consumed yet but no codeword
            # was found either. Hence append current value
            # to buffer again.
            self._currentValue <<= 8
            self._currentValue |= nextByte

            return self._buffer
        
    def decodeIntegerFromBitstream(self, bitstream):
        stream = iter(bitstream)
        integer = next(stream) << 8

        try:
            while True:
                integer |= next(stream)
        except StopIteration: pass

        return integer

    def clear_dictionary(self):
        self._dictionary = {}
        alphabet = list(range(0, 256))
    
        self._dictionary[""] = -1
        for ch in alphabet:
            self._dictionary[bytes([ch])] = ch
     
    def _buildCodetable(self):
        codewords = sorted([(k,v)
                        for k,v in self._dictionary.items() 
                        if v >= 0],
                        key=lambda t:t[-1])

        return [len(cw)-1 for cw,_ in codewords], codewords
            

if __name__ == "__main__":
    rng = open("/dev/urandom", mode="rb")
  
    out = open("./test.gif", "wb+") 
    generate_random_gif(out)
    out.close()
