#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
#
# The output file has a gritty hand-rolled GIF header which can be used to
# test various GIF parsers without needing any external tools.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):

    # Write the signature ("GIF89a") at offset 0x0
    out.seek(0);
    out.write(b"GIF89a");

    # We'll fill in these later on when writing out the extensions
    # and graphic control extension block
    imgdatapos = None

    # Image Descriptor Block
    # -----------------------

    # Write out position at offset 0x6
    out.seek(6);
    imgdatapos = out.tell();
    out.write(b'\x00\x00\x00\x00');

    # Now move to where our actual data will go
    out.seek(imgdatapos);

    # Note: the spec allows for a 16 bit delay field here, we're only going
    # to use one byte instead.

    # TODO: add support for different disposal methods etc...
    disptype = ord(rng.read(1));
    assert disptype == 0 or disptype == 2;

    # write the frame dimensions and color table indicator
    out.write((struct.pack("<H", 256)).to_bytes(2, 'little'));
    out.write((struct.pack("<H", 256)).to_bytes(2, 'little'));
    
    # full alpha channel transparency
    out.write(b"\x80");

    # sort flag set to false
    out.write(b"\x00");


    # Local Color Table
    # -----------------
    out.write(b"!" + b"!"*255);

    # Graphics Control Extension Block
    # --------------------------------
    out.seek(out.tell()+12);
    out.write(b"\x04");   # block type

    # leave these as zeroes for now
    out.write(b"\x00"*8);

    # LZW minimum code size
    out.write(b"\x02");

    # Image Data Block
    # ---------------- 

    # TODO: implement interlacing

    # TODO: figure out how big our buffer needs to be?
    dpos = out.tell();
    out.seek(dpos+1);  
    out.write(((256*256)*3)+3);      # writeword
    out.seek(dpos);                  # seek back to start of data

    for i in range(0, 256*256):
        
        val = ord(rng.read(1));

        rgbval = ((val & 0xF0)>>4,
                  (val & 0xF0)>>0,
                  0xFF);

        pixbuf = struct.pack('BBB', *(rgbval))
        out.write(pixbuf)


