#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # The spec says that the screen width and height should be 2 to the power
    # of 16 <= x <= 4096 (you're full of bytes); we shift it by two bits so that
    # it's possible for us to have a maximum value of 1024 but still have some
    # room for error
    width = ((int.from_bytes(
                    rng.read(2), byteorder='big', signed=False)) >>
            2) % (1 << 10) + 1
    height = ((int.from_bytes(
                     rng.read(2), byteorder='big', signed=False)) >>
              2) % (1 << 10) + 1
    
    bitdepth = min((int.from_bytes(
                        rng.read(1), byteorder='big', signed=False)),
                   (7)) + 1

    # We will set the gct flag iff there are colors in the palette; otherwise
    # it is just plain pixels; to do this easily, we check if the bitdepth > 8
    # (so: 2^bitdepth >= 256)
    global_color_table_flag = bool(((1 << bitdepth) - 1) < 256)

    color_resolution = max(min(int.from_bytes(
                                rng.read(1), byteorder='big', signed=False),
                               ((1 << bitdepth) - 1)).bit_length(),
                           1).bit_length() - 1

    sort_flag = False
    size_of_global_color_table = 0
    bg_color_index = 0
    pixelaspectratio = 0

    extraflags = [(sort_flag << 7) |
                  (size_of_global_color_table << 3) |
                  (bg_color_index)]
    extraflags += [pixelaspectratio] * 3
    extraflags = tuple([elem for elem in reversed(extraflags)])

    fc = frame_control_block(width,
                             height,
                             bitdepth,
                             global_color_table_flag,
                             color_resolution,
                             extraflags,)
                             
    out.write(header)
    out.write(fc.header())
    if global_color_table_flag:
        out.write(generate_random_palette().header())
        
    data = generate_random_image(fc).header()
    out.write(struct.pack('B', len(data)))
    out.write(data)
                

class frame_control_block():
    def __init__(self,
                 width: int,
                 height: int,
                 bitdepth: int,
                 global_color_table_flag: bool,
                 color_resolution: int,
                 extraflags: tuple):
        
        self.__width = width
        self.__height = height

        self.__bitdepth = bitdepth
        self.__global_color_table_flag = global_color_table_flag 
        self.__color_resolution = color_resolution
        self.__extraflags = extraflags

    def header(self):
        return struct.pack('<HHBBBHB',
                           self.__width,
                           self.__height,
                           # We convert from little-endian to big endian because
                           # that's how the image decoder expects them to be
                           ((self.__bitdepth - 1) << 7) |
                           (((not self.__global_color_table_flag) & 1) << 5) |
                           ((self.__color_resolution & 1) << 3) |
                           sum(map(lambda elem: 1<<elem[0], enumerate(reversed(self.__extraflags)))),
                           self.__bitdepth,
                          )


class palette():
    def __init__(self, entries: list):
        self.__entries = entries

    def header(self):
        result = b''
        for entry in self.__entries:
                result += struct.pack('BBBx',
                                      entry[0],
                                      entry[1],
                                      entry[2])
        return result
        
    
def generate_random_palette():
    palette = []

    while True:
        r = int.from_bytes(rng.read(1), byteorder='big', signed=True)
        g = int.from_bytes(rng.read(1), byteorder='big', signed=True)
        b = int.from_bytes(rng.read(1), byteorder='big', signed=True)

        if not any([(r == e[0] and 
                     g == e[1] and
                     b == e[2])
                    for e in palette]):
            palette.append((r,g,b))
            
        if len(palette) == 2**16:
            break
            
    return palette

def generate_random_image(framectrl):
    result = b''
    for _y in range(framectrl.height):
        for _x in range(framectrl.width):
            rgb = (int.from_bytes(
                        rng.read(1), byteorder='big', signed=True),
                   int.from_bytes(
                        rng.read(1), byteorder='big', signed=True),
                   int.from_bytes(
                        rng.read(1), byteorder='big', signed=True))

            matching_indices=[e[0] for e in framectrl.palette if all([rgb[i]==e[i] for i in range(len(rgb))])]
            result+=(matching_indices or ([0]))[-1].to_bytes(1, 'big', signed=True)

    return result