#!/usr/bin/env python3

import math
import os
import random as rand
import sys
import struct


# Generate a random gif file with the given parameters
# (all must be greater than zero except framestep)
def generate_random_gif(
        outpath : str,       # path to create gif at
        fps     : float,     # frames per second
        duration: float,     # seconds long
        overlapp: float,     # overlap between successive frames
        w       : int        # width of image
    ) -> None: 

    # check input validity
    assert fps >= 0.5
    assert duration > 0
    assert isinstance(overlapp, float) or overlapp == 0
    assert isinstance(w, int) and w > 0

    # how many frames should we have? 
    endframe = round(duration*(fps*overlapp)+1)*2

    # maximum amount of colors in our palette (max is 256)
    maxcolors = min([int((w+7)//8) - 1, 256])

    # determine palette sizes
    ctsizes = []
    while _get_colorspacesum(*ctsizes) < maxcolors**3:
        ctsizes.append(rand.randint(2, maxcolors-1))

    # compute the total number of colors we will actually need on average
    totcols = sum(_get_colorspacesum(cs) for cs in ctsizes)

    # now, figure out what's going on
    chansize = len(str(totcols))
    numframes = endframe//2

    # prepare to write the file
    with open(outpath, 'wb+') as f:
        
        f.write(b'GIF89a')

        # size of the global color table (will be updated later)
        gctsize = 0

        # write basic information about the animation 
        f.write(struct.pack("<HHBBBB", w, w, gctsize & 0xff,
            0b1 << 7| gctsize >> 8, 0, 0))

        # write out all of the color tables
        for cst in reversed(ctsizes):
            
            gctsize += cst
            r = list(range(cst))
            rand.shuffle(r)
            coltable = bytes().join(
                struct.pack("BBB", *(_rgb(n) for n in range(
                    cst*sum(ctsizes[:i]), cst*sum(ctsizes[:i+1])))
                ) for i in range(len(ctsizes)-1,-1,-1)
            )
            f.write(coltable)

        # write out local color map
        f.write(struct.pack("<B", 0))

        # write graphics control extension block
        f.write(b'\x21\xf9\x04')
        f.write(struct.pack("<BHBBBB", 0, 0, 0, 0, 0, 0))
        f.write(b';')

        # indicate transparent pixels by setting
        # the appropriate bit
        f.write(b'\x2C')
        f.write(struct.pack("<HHHBB", 0, 0, 0, 0, 0))

        # define the actual image area
        f.write(b'\x21\x00\x01')
        f.write(b'\x00')

        # LZW minimum code size, which is equal to the 
        # highest number among the color table sizes
        f.write(struct.pack("BB", ctsizes[-1], 0))

        # set up variables for tracking progress
        byteswritten = f.tell()
        totalbytes   = byteswritten + numframes*(2+math.ceil(w/8))
        percent      =.0
  
        print("Generating {}...".format(outpath))

        # iterate through every pair in the sequence
        for i in range(numframes):

            # clear the appropriate amount of space
            if i%2 == 0:

                f.seek(byteswritten)
                f.truncate()
                f.seek(-1, os.SEEK_END)
                byteswritten -= 1

                # initialize lzw buffer and its current position
                bufpos    = 1 
                cursize   = 1
                buf       = b"\x00"
                prevpixls = tuple(0 for _ in range(len(ctsizes)))
 
            else:
                
                # update the progress bar
                newpercent = i/(2*numframes)*100
                if newpercent!= percent:
                    print("\rProgress: [{:{}}] {:.0f}% ".format(
                        "="*round(newpercent/100*50)+" "*(50-round(
                            newpercent/100*50)), 
                        "#>-"*round(newpercent/100*50)), end="")
                    sys.stdout.flush()
                    percent=newpercent

                # find the next distingiushed pixel value
                pixvals = [_distinguishable_pixelval(prevpixls, 
                    ctsizes, rand.randint(0, (1<<cursize)-1))]
                cursize = min(cursize+1, 12)
                while any(((not _is_distinguishable(pixvals[0], pv, 
                    ctsizes)) or ((pv>>cursize)==pixvals[0]>>(cursize-1))) 
                    for pv in pixvals):
                    
                    pixvals.insert(
                        0, _distinguishable_pixelval(prevpixls, 
                            ctsizes, rand.randint(
                                0,(1<<cursize)+cursize-1)))

                # insert the three most prominent pixel values
                # into the lzw buffer and increment the buffer pos
                buf     = b"".join(
                    (_unshifted_bits(v, cursize+(bufpos>0))+buf)[-1:]
                    for v in pixvals
                )
                bufpos += len(pixvals)

                # rewind the start of the buffer back one position
                bufpos = max(1, bufpos-1)

            # update previous values
            prevpixls[:] = pixvals

            # get ready to insert this into the lzw buffer 
            f.write(buf[(bufpos-1)%bufpos:])
            f.write((b"\0"+buf[:bufpos])[bufpos%bufpos:])

            # update how much has been written so far
            byteswritten += len(buf[bufpos%bufpos:]) + bufpos

            # seek to the beginning of the GCT
            f.seek(3)
            f.write(struct.pack("<H", gctsize))

            # go back to the end of the file
            f.seek(totalbytes)
    
        
# Returns true iff 'a' and 'b' are distinguishable by 
# counting the number of different entries of both lists;
# returns false otherwise
def _is_distinguishable(a, b, lst:list) -> bool:
    
    return not any(x==y for x, y in zip(
        [lst[a%(d//2)][a//(d//2)] for d in lst], 
        [lst[b%(d//2)][b//(d//2)] for d in lst]
    ))
    

# Given two lists of color spaces (colorspace being defined by
# the summation of the product of their respective dimensions,
# e.g., (2 + 5)² = 2² + 2³ + 2⁴ + 2⁵ + 3² + 3³ + 5²
# is the first entry of a list of three color spaces),
# computes the value of an unshifted index whose corresponding
# rgb values differ in no more than k channels from those
# of another index, where k is the maximum depth of either
# color space
def _distinguishable_pixelval(
        lstA  : list,    # distinctive channel sequences
        lstB  : list,    # distinctive channel sequences
        val   : int,     # target color value
        k     : int = 5  # upper bound on allowed numer of channel diffs
    )-> int:

        # calculate largest power less than k such that
        # all entries in either list are less than it       
        maxpow = max(min(k, s**(2/len(lstB))) for s in lstA)
    
        # calculate smallest power larger than val
        minpow = min(s**(2/len(lstB)) for s in lstA if s <= maxpow)

        # if there exists none, then just pick a randomly
        # from the available powers and multiply them together
        if minpow == maxpow:
            return val*math.floor(rand.uniform(minpow, maxpow)**(
                1/len(lstB)))

        # take the floor of the mean of the extremes
        return int(math.floor(.5*(minpow + maxpow)*(
            1 + rand.triangular(-.5,.5))))

        
# Return the sum of the products of the dimensions of
# the given list of color spaces
def _get_colorspacesum(*args)->int:

    return sum(p for arg in args for p in sorted(arg, reverse=True)
        ) if args else 1
    
    
# Convert a single integer to RGB values in order (red, green, blue);
# assumes that the input is