#!/usr/bin/env python3

import os
import sys
import struct

from itertools import accumulate

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The returned list of bytes is the metadata for the gif.
def generate_random_gif(
        rng: BinaryIO, width=10, height=5
    ) -> bytearray:

    # Header
    out = bytearray()
    out += b"GIF89a"

    # Logical Screen Descriptor
    out += struct.pack("<H", width)   # Width
    out += struct.pack("<H", height)  # Height
    out += b"\x87\x00"               # Flags (Has colour map, sorted palette)
    bgcolor = 0                      # Background color index       (0 == black)
    aspectratio = 0                  # Pixel aspect ratio          (0 == square pixels)
    out += struct.pack("<BBB", bgcolor, aspectratio, 0)


    # Global Colour Table
    size = 6
    out += struct.pack("<BHH", size, size*3, 0)

    def make_rgb(i):
        r = i % 256; g = (i//2**8) % 256; b = (i//2**16) % 256
        return [r,g,b]

    rgbiter = iter([make_rgb(i+1) for i in range((size)**3)])
    rgbbytes = iter([c for c in accumulate(rgbiter, lambda x,y: [*x,*y])])

    out += bytes(next(rgbbytes))    # Black or white? We don't really care.

    # Image Descriptor
    lzwmincodesize = min(width,height).bit_length()  # LZW minimum code size
    flagbits = 0                                         # No interlacing
    out += struct.pack("<HHBB", 0, 0, lzwmincodesize, flagbits)

    # Plain Text Extension Block (not required by specification but allows tools to correctly display GIFs)
    text = "Created with GifRandom by <NAME>"      # The comment string
    extlen = len(text)+1                                 # Extension block length (including block type and length fields)
    out += struct.pack("<BBBBBB",
        0xfe,                                                    # Extension Introducer
        ord('t'), ord('e'), ord('X'), ord('t'),                # Extension Label
        extlen & 0xFF,                                           # Length Lo
        extlen >> 8                                              # Length Hi
    )
    out += text.encode("ascii")                             # Comment string

    # Application Extension Block (not required but useful otherwise)
    appdata = b"NETSCAPE2.0"                               # Application ID
    appdatalen = len(appdata)-1                            # Application Data Length
    expflag = 1                                           # Looping Allowed (1 == yes)
    out += struct.pack("<BBBBBB",
        0xff,                                                    # Extension Introducer
        ord('N'), ord('E'), ord('T'), ord('S'), ord('C'),     # Extension Label
        appdatalen & 0xFF,                                       # Length Lo
        appdatalen >> 8                                          # Length Hi
    )
    out += appdata[0:appdatalen]                           # Application ID

    # Graphics Control Extension Block (not required but useful otherwise)
    delaytime = b'\x00\xfa'                                # Delay time (in hundredths of seconds)
    transindex = 0                                        # Transparent Color Index             (0 == transparent)
    disposalmethod = 2                                    # Disposal method                     (2  == no dispose)
    userinputflag = 0                                     # User input expected                 (0  == no)
    transpflag = 0                                        # Transparency given in palette entry (0  == no)
    out += struct.pack("<BBBBBBBBBB",
        0x21,                                                   # Extension Introducer
        ord('g'), ord('c'), ord('F'), ord('9'), ord(' '),        # Extension Label
        4,                                                        # Block Size
        delaytime[0],                                            # Delay Time Lo
        delaytime[1],                                            # Delay Time Hi
        transindex & 0xFF,                                       # Transparent Color Index Lo
        transindex >> 8,                                         # Transparent Color Index Hi
        disposalmethod,                                          # Disposal Method
        userinputflag,                                           # User Input Flag
        transpflag                                               # Transparency Flag
    )

    data = []                                             # List of bytes making up image data

    # TODO: encode all different kinds of images here

    # Encode the first frame as an RGB image
    lzwtable = bytearray()
    lzwmincodesize = min(width,height).bit_length()  # LZW minimum code size

    dctable = {None:0}                                  # Dictionary table
    dctablepos = 1                                      # Position within dictionary table (we start at position 1 since we use None as placeholder)
    posstackpos = {}                                    # Stack pointer array - this keeps track of the current dictionary usage position so that we can later restore it after encoding each line without affecting any other line
    posstackpos[0] = 0                                  # Initialise stack pointer to zero on first dictionary
    posstackpos[width] = 0                              # Also initialise at end of last line

    bitstack = 0                                        # Bit stack
    valuemask = ((1<<(lzwmincodesize))-1)<<bitstack      # Mask used when clearing bits from bitstack
    data.append(0x80 | lzwmincodesize)                   # Clearing code at beginning of image

    # Encode horizontal lines
    x = y = 0
    while True:

        if x >= width:
            break

        if y >= height:

            # End of image
            break


        # Go through one line of image data
        imgrow = []
        while True:


            # Read next pixel
            if x > width-1 or y > height-1:
                imgrow.append(struct.pack('<B', 0))
            else:

                # Get pixel intensity
                imgrow.append(rng.read(1)[0])



            # Encode previous row
            linelen = len(imgrow)
            mincodesize = lzwmincodesize
            while True:

                # Find longest matching sequence
                maxmatchlen = 0
                bestmatchval = None
                for matchlen in range(maxmatchlen,2,-1):
                    if not imgrow[linelen-matchlen:(linelen)] in dctable:
                        continue

                    matchval = dctable[imgrow[linelen-matchlen:(linelen)]]
                    if matchlen > maxmatchlen and matchval>mincodesize:

                        # New longest match found
                        maxmatchlen = matchlen
                        bestmatchval = matchval

                if maxmatchlen <= 0:

                    # No more matches possible


                    # Write new literal character to the buffer
                    val = imgrow[-1]
                    if val in dctable and dctable[val]>mincodesize:
                        
                        # Value already present in table



                        # Subtract with minimum code size
                        val -= mincodesize

                        # Extract highest n bits from val and add them to bit stream
                        bitstack |= (val>>(lzwmincodesize-bitstack))&valuemask


                        # Check whether done inserting bits
                        if bitstack>=8:

                            # Output accumulated literal characters
                            accu = bitstack>>8
                            while accu > 0:
                                data.append(accu & 0xFF)
                                accu >>= 8
                            bitstack &= 0xFF


                        # Add literal to dictionary
                        dctable[imgrow[(linelen-1):]] = dctablepos

                        # Increment dictionary position
                        dctablepos += 1

                        # Reset lowest bit in bitstream
                        bitstack = 0

                        # Can stop because there's nowhere left to go back to
                        break

                else:

                    # Match found!



                    # Subtract with minimum code size
                    bestmatchval -= mincodesize

                    # Extract highest n bits from val and add them to bit stream
                    bitstack |= (bestmatchval>>(lzwmincodesize-bitstack))&valuemask

                    # Check whether done inserting bits
                    if bitstack>=8:

                        # Output accumulated literal characters
                        accu = bitstack>>8
                        while accu > 0:
                            data.append(accu & 0xFF)
                            accu >>= 8
                        bitstack &= 0xFF



                    # Possibly write code to the output buffer
                    if bestmatchval<=(dictlen:=dctablepos-1):

                        # Code is present in table



                        # Extract highest n bits from val and add them to bit stream
                        bitstack |= (bestmatchval>>(lzwmincodesize-bitstack))&valuemask

                        # Check whether done inserting bits
                        if bitstack>=8:

                            # Output accumulated codes
                            accu = bitstack>>8
                            while accu > 0:
                                data.append(accu & 0xFF)
                                accu >>= 8
                            bitstack &= 0xFF

                    elif dictlen<=255:

                        # Code greater than maximum allowed in single operation
                        # Therefore, store remainder in dictionary and encode remaining literals afterwards
                        remcode = (((bestmatchval-dictlen)>>8)&0xFF)|(((bestmatchval-dictlen)&0xFF)<<8)
                        remlength = maxmatchlen-1

                        # Write match to buffer
                        bestmatchval = dict