#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 27 16:04:59 2020

@author: jonas
"""
import numpy as np
from scipy.stats import norm, multivariate_normal
from scipy.optimize import minimize
import matplotlib.pyplot as plt

def get_data(n=100):
    """
    Generates n samples from a bimodal distribution with means mu1 and mu2
    
    Parameters
    ----------
    n : int (default = 100)
        Number of samples to generate
        
    Returns
    -------
    X : array of shape (n, )
        Samples generated from the bimodal distribution
    """
    # Set up parameters for generating data
    mu1 = 10
    mu2 = 20
    sigma1 = 1
    sigma2 = 2
    p1 = .5
    p2 = .5
    
    # Generate data
    X = np.random.choice([mu1, mu2], size=n, replace=True, p=[p1, p2])
    X += np.random.randn(n)*np.array([sigma1, sigma2]).reshape(-1, 1)[X]
    
    return X

def log_likelihood(theta, X):
    """
    Computes the log likelihood function for the bimodal distribution
    
    Parameters
    ----------
    theta : array of shape (2, )
        Array containing the mean values of the two distributions
    X : array of shape (n, )
        Sample data
        
    Returns
    -------
    ll : float
        Log likelihood value
    """
    # Unpack theta
    mu1, mu2 = theta
    
    # Get probabilities
    p1 = norm.pdf(X, loc=mu1)
    p2 = norm.pdf(X, loc=mu2)
    
    # Return log likelihood
    return np.sum(np.log(p1 + p2))

def mle(X):
    """
    Estimates the MLEs for the parameters of the bimodal distribution
    
    Parameters
    ----------
    X : array of shape (n, )
        Sample data
        
    Returns
    -------
    theta : array of shape (2, )
        Estimated parameter values
    """
    # Initialize theta
    theta = np.zeros((2, ))
    
    # Define objective function
    def obj(theta):
        return -log_likelihood(theta, X)
    
    # Minimize
    res = minimize(obj, theta)
    
    return res.x

def plot_bimodal():
    """
    Plots the bimodal distribution
    """
    # Set up parameters for generating data
    mu1 = 10
    mu2 = 20
    sigma1 = 1
    sigma2 = 2
    p1 = .5
    p2 = .5
    
    # Create grid of x values
    x = np.linspace(min(mu1, mu2)-5, max(mu1, mu2)+5, 100)
    
    # Plot pdfs
    plt.figure()
    plt.plot(x, norm.pdf(x, loc=mu1), label='$\\mathcal{N}(\\mu_1, \\sigma_1^2)$')
    plt.plot(x, norm.pdf(x, loc=mu2), label='$\\mathcal{N}(\\mu_2, \\sigma_2^2)$')
    plt.legend()
    plt.title('Bimodal Distribution PDF')
    plt.show()

def plot_mle(X):
    """
    Plots the estimated MLEs for the parameters of the bimodal distribution
    
    Parameters
    ----------
    X : array of shape (n, )
        Sample data
    """
    # Get estimates
    theta = mle(X)
    
    # Create grid of x values
    x = np.linspace(min(theta[0], theta[1])-5