#!/usr/bin/env python3

import struct
from typing import BinaryIO, List, Optional, Tuple

from .common import (
    BLOCK_SIZE,
    BlockType,
    FileHeader,
    Header,
    HeaderFlags,
    MagicNumber,
    SectionHeader,
)


def read_file(f: BinaryIO):
    magic = f.read(4)
    if len(magic) != 4 or magic != b"BMAP":
        raise ValueError("Invalid file header")

    version = struct.unpack("<I", f.read(4))[0]
    if version != 1:
        raise ValueError(f"Unsupported version {version}")

    flags = HeaderFlags(*struct.unpack("<II", f.read(8)))
    block_size = struct.unpack("<Q", f.read(8))[0]
    num_blocks = struct.unpack("<Q", f.read(8))[0]

    return FileHeader(flags, block_size, num_blocks)


def write_file(f: BinaryIO, hdr: FileHeader):
    f.write(MagicNumber)
    f.write(struct.pack("<I", 1))
    f.write(struct.pack("<II", *hdr.flags))
    f.write(struct.pack("<Q", hdr.block_size))
    f.write(struct.pack("<Q", hdr.num_blocks))


def read_header(f: BinaryIO):
    magic = f.read(4)
    if len(magic) != 4 or magic != b"BHDR":
        raise ValueError("Invalid header")

    version = struct.unpack("<I", f.read(4))[0]
    if version != 1:
        raise ValueError(f"Unsupported version {version}")

    flags = HeaderFlags(*struct.unpack("<II", f.read(8)))
    block_size = struct.unpack("<Q", f.read(8))[0]
    num_blocks = struct.unpack("<Q", f.read(8))[0]
    num_sections = struct.unpack("<I", f.read(4))[0]

    sections = []
    for _ in range(num_sections):
        section_type = BlockType(struct.unpack("<I", f.read(4))[0])
        offset = struct.unpack("<Q", f.read(8))[0]
        size = struct.unpack("<Q", f.read(8))[0]
        sections.append((section_type, offset, size))

    return Header(flags, block_size, num_blocks, sections)


def write_header(f: BinaryIO, hdr: Header):
    f.write(MagicNumber)
    f.write(struct.pack("<I", 1))
    f.write(struct.pack("<II", *hdr.flags))
    f.write(struct.pack("<Q", hdr.block_size))
    f.write(struct.pack("<Q", hdr.num_blocks))
    f.write(struct.pack("<I", len(hdr.sections)))
    for section in hdr.sections:
        f.write(struct.pack("<I", section[0].value))
        f.write(struct.pack("<QQ", section[1], section[2]))


def read_section_header(f: BinaryIO):
    magic = f.read(4)
    if len(magic) != 4 or magic != b"BSHT":
        raise ValueError("Invalid section header")

    version = struct.unpack("<I", f.read(4))[0]
    if version != 1:
        raise ValueError(f"Unsupported version {version}")

    flags = HeaderFlags(*struct.unpack("<II", f.read(8)))
    block_size = struct.unpack("<Q", f.read(8))[0]
    num_blocks = struct.unpack("<Q", f.read(8))[0]
    num_entries = struct.unpack("<Q", f.read(8))[0]

    entries = []
    for _ in range(num_entries):
        offset = struct.unpack("<Q", f.read(8))[