#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Write header
    out.write(b'GIF89a')

    # Write screen descriptor
    width = int.from_bytes(rng.read(2), 'big', signed=False) % 100 + 1
    height = int.from_bytes(rng.read(2), 'big', signed=False) % 100 + 1
    background_color_index = int.from_bytes(rng.read(1), 'big', signed=False) % 256
    pixel_aspect_ratio = int.from_bytes(rng.read(1), 'big', signed=False) % 16
    out.write(struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio))

    # Write global color table
    for _ in range(7):
        red = int.from_bytes(rng.read(1), 'big', signed=False) % 256
        green = int.from_bytes(rng.read(1), 'big', signed=False) % 256
        blue = int.from_bytes(rng.read(1), 'big', signed=False) % 256
        out.write(struct.pack('<BBB', red, green, blue))

    # Write application extension block
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01')
    loop_count = int.from_bytes(rng.read(2), 'big', signed=False) % (2 ** 16 - 1)
    out.write(struct.pack('<H', loop_count))

    # Write image descriptor
    left = int.from_bytes(rng.read(2), 'big', signed=False) % width
    top = int.from_bytes(rng.read(2), 'big', signed=False) % height
    width = int.from_bytes(rng.read(2), 'big', signed=False) % (width - left) + 1
    height = int.from_bytes(rng.read(2), 'big', signed=False) % (height - top) + 1
    local_color_table_flag = bool(int.from_bytes(rng.read(1), 'big', signed=False) & 1)
    interlace_flag = bool(int.from_bytes(rng.read(1), 'big', signed=False) & 1)
    sort_flag = bool(int.from_bytes(rng.read(1), 'big', signed=False) & 1)
    reserved = int.from_bytes(rng.read(1), 'big', signed=False) & 7
    out.write(struct.pack('<HHBBBB', left, top, width, height, local_color_table_flag, interlace_flag, sort_flag, reserved))

    # Write local color table
    if local_color_table_flag:
        for _ in range(7):
            red = int.from_bytes(rng.read(1), 'big', signed=False) % 256
            green = int.from_bytes(rng.read(1), 'big', signed=False) % 256
            blue = int.from_bytes(rng.read(1), 'big', signed=False) % 256
            out.write(struct.pack('<BBB', red, green, blue))

    # Write LZW-compressed image data
    clear_code = 0
    end_of_information_code = 0
    code_size = 9
    dictionary = {}
    current_code = None
    previous_code = None
    while True:
        # Read next byte from input
        byte = int.from_bytes(rng.read(1), 'big', signed=False) % 256

        # Check if we should write a clear code
        if len(dictionary) == 4096:
            out.write(struct.pack('<B', clear_code))
            code_size += 1
            if code_size > 12:
                raise ValueError("Code size too large")
            dictionary = {}
            current