#!/usr/bin/env python3

import struct
from typing import BinaryIO


def read_int(f: BinaryIO) -> int:
    return struct.unpack('<i', f.read(4))[0]


def write_int(f: BinaryIO, i: int):
    f.write(struct.pack('<i', i))


def read_float(f: BinaryIO) -> float:
    return struct.unpack('<f', f.read(4))[0]


def write_float(f: BinaryIO, fv: float):
    f.write(struct.pack('<f', fv))


def read_string(f: BinaryIO) -> str:
    length = read_int(f)
    if length == 0:
        return ''
    else:
        return f.read(length).decode()


def write_string(f: BinaryIO, s: str):
    write_int(f, len(s))
    f.write(s.encode())


def read_vec2(f: BinaryIO) -> tuple:
    x = read_float(f)
    y = read_float(f)
    return (x, y)


def write_vec2(f: BinaryIO, v: tuple):
    assert len(v) == 2
    write_float(f, v[0])
    write_float(f, v[1])


def read_vec3(f: BinaryIO) -> tuple:
    x = read_float(f)
    y = read_float(f)
    z = read_float(f)
    return (x, y, z)


def write_vec3(f: BinaryIO, v: tuple):
    assert len(v) == 3
    write_float(f, v[0])
    write_float(f, v[1])
    write_float(f, v[2])


def read_vec4(f: BinaryIO) -> tuple:
    x = read_float(f)
    y = read_float(f)
    z = read_float(f)
    w = read_float(f)
    return (x, y, z, w)


def write_vec4(f: BinaryIO, v: tuple):
    assert len(v) == 4
    write_float(f, v[0])
    write_float(f, v[1])
    write_float(f, v[2])
    write_float(f, v[3])


def read_quat(f: BinaryIO) -> tuple:
    x = read_float(f)
    y = read_float(f)
    z = read_float(f)
    w = read_float(f)
    return (x, y, z, w)


def write_quat(f: BinaryIO, q: tuple):
    assert len(q) == 4
    write_float(f, q[0])
    write_float(f, q[1])
    write_float(f, q[2])
    write_float(f, q[3])


def read_mat4(f: BinaryIO) -> list:
    m = []
    for _ in range(4):
        row = []
        for _ in range(4):
            row.append(read_float(f))
        m.append(row)
    return m


def write_mat4(f: BinaryIO, m: list):
    assert len(m) == 4 and all([len(r) == 4 for r in m])
    for row in m:
        for col in row:
            write_float(f, col)


def read_color(f: BinaryIO) -> tuple:
    r = read_float(f)
    g = read_float(f)
    b = read_float(f)
    a = read_float(f)
    return (r, g, b, a)


def write_color(f: BinaryIO, c: tuple):
    assert len(c) == 4
    write_float(f, c[0])
    write_float(f, c[1])
    write_float(f, c[2])
    write_float(f, c[3])


def read_bool(