#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # GIF header
    out.write(b'GIF89a')

    # Screen descriptor
    width = 16 + (rng.read(1)[0] & 0xF) * 16
    height = 16 + (rng.read(1)[0] & 0xF) * 16
    color_table_size = 2 ** ((rng.read(1)[0] >> 4) + 1)
    background_color_index = rng.read(1)[0] % color_table_size
    pixel_aspect_ratio = rng.read(1)[0] // 16
    out.write(struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio))

    # Global color table
    for _ in range(color_table_size):
        out.write(rng.read(3))

    # Image descriptor
    left = rng.read(2)
    top = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    local_color_table_flag = bool(rng.read(1)[0] & 0x80)
    interlaced = bool(rng.read(1)[0] & 0x40)
    sorted = bool(rng.read(1)[0] & 0x20)
    reserved = bool(rng.read(1)[0] & 0x10)
    color_table_size = 2 ** ((rng.read(1)[0] >> 4) + 1)
    out.write(left + top + width + height + bytes([local_color_table_flag | color_table_size]))

    # Local color table
    if local_color_table_flag:
        for _ in range(color_table_size):
            out.write(rng.read(3))

    # Graphic control extension
    block_size = 4
    extension_introducer = b'\x21'
    graphic_control_label = b'\xf9'
    disposal_method = rng.read(1)[0] & 0x7
    user_input_flag = bool(rng.read(1)[0] & 0x8)
    transparent_color_flag = bool(rng.read(1)[0] & 0x10)
    delay_time = rng.read(2)
    transparent_color_index = rng.read(1)[0]
    block_terminator = b'\x00'
    out.write(block_size + extension_introducer + graphic_control_label + bytes([disposal_method | user_input_flag | transparent_color_flag]) + delay_time + transparent_color_index + block_terminator)

    # Image data
    lzw_minimum_code_size = rng.read(1)[0]
    out.write(lzw_minimum_code_size)

    while True:
        code_size = rng.read(1)[0]
        if code_size == 0:
            break
        out.write(code_size)

        while True:
            code = rng.read(1)[0]
            if code == 0:
                break
            out.write(bytes([code]))

    # Trailer
    out.write(b';')
