#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Magic header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + rng.read(2)[0] % 640
    height = 1 + rng.read(2)[0] % 480
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 8
    out.write(struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio))

    # Global color table
    global_color_table_flag = 0x80 | (rng.read(1)[0] & 0x7F)
    global_color_table_size = 1 << ((global_color_table_flag >> 4) + 1)
    global_color_table = [
            rng.read(3),
            *[rng.read(3) for _ in range(global_color_table_size - 1)]
        ]
    out.write(bytes([global_color_table_flag]))
    out.write(b''.join(global_color_table))

    # Image block
    image_separator = b';'
    out.write(image_separator)

    # Local image descriptor
    left = 0
    top = 0
    width = width
    height = height
    local_color_table_flag = 0x80 | (rng.read(1)[0] & 0x7F)
    interlace_flag = 0
    sort_flag = 0
    local_color_table_size = 1 << ((local_color_table_flag >> 4) + 1)
    local_color_table = [
            rng.read(3),
            *[rng.read(3) for _ in range(local_color_table_size - 1)]
        ]
    out.write(struct.pack('<HHBBB', left, top, width, height, local_color_table_flag))
    out.write(b''.join(local_color_table))

    # Raster data
    lzw_minimum_code_size = 8
    clear_code = 0x100
    end_of_information_code = 0x101
    code_size = lzw_minimum_code_size + 1
    code_mask = (1 << code_size) - 1
    available_codes = 1 << lzw_minimum_code_size
    old_code = None
    stack = []
    while True:
        if len(stack) == 0:
            code = rng.read(1)[0] & code_mask
        else:
            code = stack[-1]
            if code < available_codes:
                stack.append(old_code)
                stack.append(code)
            elif code == clear_code:
                code_size = lzw_minimum_code_size + 1
                code_mask = (1 << code_size) - 1
                available_codes = 1 << lzw_minimum_code_size
                old_code = None
                continue
            elif code == end_of_information_code:
                break
            else:
                raise Exception("Invalid LZW code")
        if code == clear_code or available_codes > 4096:
            code_size += 1
            code_mask = (1 << code_size) - 1
            available_codes = 1 << lzw_minimum_code_size
        if old_code is not None:
            if code == clear_code:
                pass
            elif code == end_of_information_code:
                pass
            elif code >= available_codes:
                raise Exception("Invalid LZW code")
            else:
                stack.append(old_code)
                stack.append(code)
        old_code = code
        if len(stack) > 0:
            out.write(struct.pack('<H', available_codes))
            available_codes += 1
            if available_codes == 512:
                out.write(struct.pack('<H',