#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 25 10:46:49 2019

@author: jlee
"""

import numpy as np
from astropy.io import fits
from astropy.table import Table, vstack
from scipy.interpolate import interp1d


def get_wavelength(fitsfile):
    """
    Get wavelength array from a FITS file.
    
    Parameters
    ----------
    fitsfile : str
        Name of the FITS file.
        
    Returns
    -------
    wave : array
        Wavelength array in Angstroms.
    """
    
    hdu = fits.open(fitsfile)
    head = hdu[0].header
    crval = head['CRVAL1']
    cdelt = head['CDELT1']
    nwave = head['NAXIS1']
    wave = np.arange(nwave)*cdelt + crval
    
    return wave


def read_spectrum(fitsfile, ext=0, fluxunits='erg/s/cm^2/A', verbose=True):
    """
    Read spectrum from a FITS file.
    
    Parameters
    ----------
    fitsfile : str
        Name of the FITS file.
    ext : int
        Extension number to read. Default is 0.
    fluxunits : str
        Units for the flux. Default is 'erg/s/cm^2/A'.
    verbose : bool
        Print out information about the spectrum? Default is True.
        
    Returns
    -------
    spec : array
        Spectral data.
    wave : array
        Corresponding wavelength array in Angstroms.
    """
    
    # Read spectrum
    hdu = fits.open(fitsfile)[ext]
    head = hdu.header
    spec = hdu.data
    
    # Convert units if necessary
    if (fluxunits == 'erg/s/cm^2/A'):
        pass
    elif (fluxunits == 'Jy'):
        spec *= 1e-23
    else:
        raise ValueError('Unrecognized flux units!')
    
    # Get wavelength array
    try:
        wave = hdu.wave
    except AttributeError:
        wave = get_wavelength(fitsfile)
    
    # Print out info
    if verbose:
        print('Read spectrum from {}'.format(fitsfile))
        print('\tWavelength range: {:.2f} - {:.2f} A'.format(np.min(wave), np.max(wave)))
        print('\tFlux range: {:.2f} - {:.2f} erg/s/cm^2/A'.format(np.min(spec), np.max(spec)))
    
    return spec, wave


def read_spectra(fitsfiles, ext=0, fluxunits='erg/s/cm^2/A', verbose=True):
    """
    Read spectra from a list of FITS files.
    
    Parameters
    ----------
    fitsfiles : list
        List of names of the FITS files.
    ext : int
        Extension number to read. Default is 0.
    fluxunits : str
        Units for the flux. Default is 'erg/s/cm^2/A'.
    verbose : bool
        Print out information about the spectra? Default is True.
        
    Returns
    -------
    spec : array
        Spectral data.
    wave : array
        Corresponding wavelength array in Angstroms.
    """
    
    # Initialize arrays
    spec = []
    wave = []
    
    # Loop through all files
    for i, f in enumerate(fitsfiles):
        
        # Read spectrum
        s, w = read_spectrum(f, ext=ext, fluxunits=fluxunits, verbose=(i==0 and verbose))
        
        # Append to arrays
        spec.append(s)
        wave.append(w)
    
    return spec, wave


def write_spectrum(fitsfile, spec, wave