#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# returns its size in bytes.
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 256 + (rng.read(1)[0] % 4) * 256
    height = 256 + (rng.read(1)[0] % 4) * 256
    flags = b'\x87\x00' if rng.read(1)[0] & 1 else b'\x07\x00'
    bgColorIndex = rng.read(1)[0] % 256
    pixelAspectRatio = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBBB', width, height, flags[0], flags[1],
        bgColorIndex, pixelAspectRatio))

    # Global color table
    for i in range(256):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    leftPos = rng.read(1)[0] % 256
    topPos = rng.read(1)[0] % 256
    imageWidth = 256 + (rng.read(1)[0] % 4) * 256
    imageHeight = 256 + (rng.read(1)[0] % 4) * 256
    flags = b'\x87\x00' if rng.read(1)[0] & 1 else b'\x07\x00'
    localColorTableFlag = b'\x00' if rng.read(1)[0] & 1 else b'\x80'
    interlaceFlag = b'\x00' if rng.read(1)[0] & 1 else b'\x40'
    sortFlag = b'\x00' if rng.read(1)[0] & 1 else b'\x20'
    reserved = b'\x00'
    lctSize = 2 << ((rng.read(1)[0] % 3) + 1)
    backgroundIndex = rng.read(1)[0] % 256
    out.write(struct.pack('<HHHHBBBBB', leftPos, topPos, imageWidth,
        imageHeight, flags[0], flags[1], localColorTableFlag[0],
        interlaceFlag[0], sortFlag[0], reserved[0], lctSize, backgroundIndex))

    # Local color table
    for i in range(lctSize):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        out.write(struct.pack('BBB', r, g, b))

    # Image data
    numPixels = imageWidth * imageHeight
    while True:
        blockLength = min(numPixels, rng.read(1)[0] % 256)

        if blockLength == 0:
            break

        out.write(bytes([blockLength]))
        for i in range(blockLength):
            out.write(bytes([(rng.read(1)[0] % 256)]))

        numPixels -= blockLength

    return out.tell()

