#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    header = b'GIF89a'

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 10)
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 10)
    color_table_size = 7
    global_color_table = bytes([
        0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0xff, 0xff,
        0x00, 0xff, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00,
        0xff, 0xff, 0xff, 0xff])[:color_table_size * 3]
    background_color = 0
    pixel_aspect_ratio = 0
    logical_screen_descriptor = struct.pack('<HHBBB',
        width, height,
        color_table_size,
        background_color,
        pixel_aspect_ratio)

    # Global color table
    data = logical_screen_descriptor + global_color_table

    # Image descriptor
    left   = 0
    top    = 0
    width  = width - 1
    height = height - 1
    local_color_table_flag = False
    interlace_flag        = False
    sort_flag             = False
    reserved              = 0
    local_color_table     = None
    image_descriptor      = struct.pack('<HHHHBB',
        left, top, width, height,
        local_color_table_flag,
        interlace_flag,
        sort_flag,
        reserved)

    # Local color table
    if local_color_table_flag:
        local_color_table = bytes([
            0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0xff, 0xff,
            0x00, 0xff, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff, 0xff])[:color_table_size * 3]

        data += local_color_table

    # Table of contents
    num_of_blocks = 1
    block_size = len(data)
    block_offsets = [block_size]
    for i in range(num_of_blocks-1):
        block_offsets.append(int.from_bytes(rng.read(2),
            byteorder='little'))

    # Trailer
    trailer = b''
    for offset in reversed(block_offsets):
        trailer += offset.to_bytes(2, byteorder='little')
    trailer += b'\x00\x3b'

    # Write everything to output stream
    data += trailer

    out.write(header + data)
