#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    out.write(b'GIF89a')

    width = 1024 + int.from_bytes(rng.read(2), byteorder='little', signed=False) % (512 - 1024)
    height = 768 + int.from_bytes(rng.read(2), byteorder='little', signed=False) % (384 - 768)
    flags = b'\x87\x00' if bool(int.from_bytes(rng.read(1), byteorder='little', signed=False)) else b'\x00\x00'
    background_color_index = int.from_bytes(rng.read(1), byteorder='little', signed=False) & 0xF
    pixel_aspect_ratio = int.from_bytes(rng.read(1), byteorder='little', signed=False) & 0xFF
    out.write(struct.pack('<HHBBB', width, height, flags[0], flags[1], background_color_index, pixel_aspect_ratio))

    # Global Color Table
    color_table_size = min((width * height) // 256, 256)
    for i in range(color_table_size):
        r = int.from_bytes(rng.read(1), byteorder='little', signed=False)
        g = int.from_bytes(rng.read(1), byteorder='little', signed=False)
        b = int.from_bytes(rng.read(1), byteorder='little', signed=False)
        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Image Descriptor
    left = 0
    top = 0
    width = width
    height = height
    flags = b'\x87\x00' if bool(int.from_bytes(rng.read(1), byteorder='little', signed=False)) else b'\x00\x00'
    local_color_table_flag = True if bool(int.from_bytes(rng.read(1), byteorder='little', signed=False)) else False
    interlace_flag = True if bool(int.from_bytes(rng.read(1), byteorder='little', signed=False)) else False
    sort_flag = True if bool(int.from_bytes(rng.read(1), byteorder='little', signed=False)) else False
    reserved = int.from_bytes(rng.read(2), byteorder='little', signed=False)
    local_color_table_size = int.from_bytes(rng.read(1), byteorder='little', signed=False) & 0xFF
    out.write(struct.pack('<HHHHBB', left, top, width, height, flags[0], flags[1]))
    out.write(local_color_table_flag.to_bytes(1, 'little'))
    out.write(interlace_flag.to_bytes(1, 'little'))
    out.write(sort_flag.to_bytes(1, 'little'))
    out.write(reserved.to_bytes(2, 'little'))
    out.write(local_color_table_size.to_bytes(1, 'little'))

    # Local Color Table
    for i in range(local_color_table_size):
        r = int.from_bytes(rng.read(1), byteorder='little', signed=False)
        g = int.from_bytes(rng.read(1), byteorder='little', signed=False)
        b = int.from_bytes(rng.read(1), byteorder='little', signed=False)
        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # LZW Minimum Code Size
    lzw_minimum_code_size = int.from_bytes(rng.read(1), byteorder='little', signed=False) & 0xFF
    out.write(lzw_minimum_code_size.to_bytes(1, 'little'))

    # Data Sub-blocks
    data_subblock_size = 2 ** lzw_minimum_code_size
    while len(data_subblock := rng.read(data_subblock_size)):
        out.write(len(data_subblock).