#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 21 09:45:07 2019

@author: <NAME> (<EMAIL>)
"""


import numpy as np
from scipy.stats import norm, lognorm
from scipy.optimize import minimize_scalar
from sklearn.metrics import mean_squared_error



def get_lognormal_params(x):
    """
    Estimate the parameters of a lognormal distribution from data x
    
    Parameters
    ----------
    x : array
        Data to estimate the parameters from
        
    Returns
    -------
    mu : float
        Mean of the normal distribution
    sigma : float
        Standard deviation of the normal distribution
    """

    # Get the shape and scale parameters for the lognormal distribution
    shape, loc, scale = lognorm.fit(x)
    
    # Convert the shape parameter into a sigma value
    sigma = (shape - 1)/np.sqrt(shape * np.pi)
    
    return loc, sigma



def fit_lognormal_distribution(x, nbins=None):
    """
    Fit a lognormal distribution to data using maximum likelihood estimation
    
    Parameters
    ----------
    x : array
        Data to fit the distribution to
    nbins : int or None
        Number of bins to use in histogramming the data before fitting the 
        distribution. If None is given then no binning will be performed. The 
        default is None.
        
    Returns
    -------
    distr : scipy.stats._distn_infrastructure.rv_frozen
        Lognormal distribution with estimated parameters
    """
    
    if nbins is not None:
        
        # Bin the data
        hist, edges = np.histogram(x, bins=nbins)
        
        # Calculate the bin centers
        bin_centers = [(edges[i] + edges[i+1])/2 for i in range(len(hist))]
        
        # Calculate the bin widths
        bin_widths = [abs(edges[i] - edges[i+1]) for i in range(len(hist))]
        
        # Calculate the weights used when calculating the MLE
        weights = [hist[i]/bin_widths[i] for i in range(len(hist))]
        
    else:
        
        # No binning so set all weights equal
        bin_centers = x
        weights = np.ones((len(x),))
    
    
    # Get the initial guess at the parameters
    mu, sigma = get_lognormal_params(bin_centers)
    
    # Define the function that returns minus the loglikelihood
    def neg_loglike(mu, sigma):
        return -sum([weights[i]*lognorm.pdf(bin_centers[i], s=sigma, scale=np.exp(mu)) 
                     for i in range(len(bin_centers))])
    
    
    # Minimize the negative loglikelihood
    res = minimize_scalar(neg_loglike, args=(sigma,), method='bounded', bounds=[0., 10.],
                          options={'disp': False})
    
    # Extract the optimized values
    mu = res.x
    sigma = abs(res.fun)**(-1./2.)
    
    # Return the fitted distribution
    return lognorm(s=sigma, scale=np.exp(mu))



def get_normal_params(x):
    """
    Estimate the parameters of a normal distribution from data x
    
    Parameters
    ----------
    x : array
        Data to estimate the parameters from
        
    Returns
    -------
    mu : float
        Mean of the normal distribution
    sigma : float
        Standard deviation of the normal distribution
    """

    # Get the mean and standard deviation of the normal distribution
    mu, sigma = norm.fit(x)
    
    return mu, sigma



def fit_normal_distribution(x, nbins=None):
    """
    Fit a normal distribution to data using maximum likelihood estimation
    
    Parameters
    ----------
    x : array
        Data to fit the distribution to
    nbins : int or None
        Number of bins to use in histogramming the data before fitting the 
        distribution. If None is given then no binning will be performed. The 
        default is None.
        
    Returns
    -------
    distr : scipy.stats._distn_infrastructure.rv_frozen
        Normal distribution with estimated parameters
    """
    
    if nbins is not None:
        
        # Bin the data
        hist, edges = np.histogram(x, bins=nbins)
        
        # Calculate the bin centers
        bin_centers = [(edges[i] + edges[i+1])/2 for i in range(len(hist))]
        
        # Calculate the bin widths
        bin_widths = [abs(edges[i] - edges[i+1]) for i in range(len(hist))]
        
        # Calculate the weights used when calculating the MLE
        weights = [hist[i]/bin_widths[i] for i in range(len(hist))]
        
    else:
        
        # No binning so set