#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    header = b'GIF89a' + bytes([0x12, 0x34])

    # Logical screen descriptor
    width = rng.read(1)[0] % (65536 - 1) + 1
    height = rng.read(1)[0] % (65536 - 1) + 1
    flags = rng.read(1)[0] & 7
    bgColorIndex = rng.read(1)[0] % 256
    pixelAspectRatio = rng.read(1)[0] % 9

    lsd = bytearray()
    lsd += struct.pack('>H', width)
    lsd += struct.pack('>H', height)
    lsd += struct.pack('B', flags)
    lsd += struct.pack('B', bgColorIndex)
    lsd += struct.pack('B', pixelAspectRatio)

    # Global color table
    globalColorTableSize = rng.read(1)[0] % 2**3 + 1
    globalColorTable = bytearray()
    for i in range(globalColorTableSize * 3):
        globalColorTable.append(rng.read(1)[0] % 256)

    # Image descriptor
    leftPos = rng.read(1)[0] % (width - 1)
    topPos = rng.read(1)[0] % (height - 1)
    imageWidth = rng.read(1)[0] % (width - leftPos) + 1
    imageHeight = rng.read(1)[0] % (height - topPos) + 1
    flags = rng.read(1)[0] & 7
    localColorTableSize = rng.read(1)[0] % 2**3 + 1
    interlaceFlag = bool(flags & 0b10000000)
    sortFlag = bool(flags & 0b01000000)
    reserved = rng.read(1)[0] & 0xF
    localColorTable = bytearray()
    for i in range(localColorTableSize * 3):
        localColorTable.append(rng.read(1)[0] % 256)

    id = bytearray()
    id += struct.pack('B', leftPos)
    id += struct.pack('B', topPos)
    id += struct.pack('B', imageWidth)
    id += struct.pack('B', imageHeight)
    id += struct.pack('B', flags)
    id += struct.pack('B', localColorTableSize)

    # LZW minimum code size
    minCodeSize = rng.read(1)[0] % 2**3 + 1

    # LZW data
    # TODO: Figure out how to make this more efficient
    lzwData = bytearray()
    while True:
        if len(lzwData) == 0 or rng.read(1)[0] < 0x80:
            lzwData.append(rng.read(1)[0])
        else:
            length = rng.read(1)[0]
            if length > len(lzwData):
                break

            lzwData.append(lzwData[-length:])

        if len(lzwData) >= 256 and len(set(lzwData)) <= 256:
            break

    lzwData += b'\xFF\x00'

    # Trailer
    trailer = b';'

    out.write(header)
    out.write(lsd)
    out.write(globalColorTable)
    out.write(id)
    out.write(struct.pack('B', minCodeSize))
    out.write(lzwData)
    out.write(trailer)

