#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# returns its size in bytes.
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write("GIF89a".encode())

    # Logical screen descriptor
    width = 256 + (rng.read(1)[0] & 0x7f)
    height = 256 + (rng.read(1)[0] & 0x7f)
    flags = ((rng.read(1)[0] & 0x7f) << 8) | rng.read(1)[0]
    bgColorIndex = rng.read(1)[0]
    pixelAspectRatio = rng.read(1)[0]
    out.write(struct.pack("<HHBBB", width, height, flags, bgColorIndex, pixelAspectRatio))

    # Global color table
    globalColorTableSize = 2 ** ((flags >> 4) & 0b111)
    for i in range(globalColorTableSize):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(struct.pack("<BBB", r, g, b))

    # Image descriptor
    leftPos = 0
    topPos = 0
    imageWidth = width - leftPos
    imageHeight = height - topPos
    flags = 0
    localColorTableFlag = bool((rng.read(1)[0] & 0x80)!= 0)
    interlaceFlag = bool((rng.read(1)[0] & 0x40)!= 0)
    sortFlag = bool((rng.read(1)[0] & 0x20)!= 0)
    reserved = rng.read(1)[0]
    lctSize = 2 ** ((rng.read(1)[0] & 0x07) + 1) if localColorTableFlag else None
    lctBackgroundColorIndex = rng.read(1)[0] if localColorTableFlag else None
    out.write(struct.pack("<HHBBBBB", leftPos, topPos, imageWidth, imageHeight, flags, lctSize or 0, lctBackgroundColorIndex or 0))

    # Local color table
    if localColorTableFlag:
        for i in range(lctSize):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            out.write(struct.pack("<BBB", r, g, b))

    # LZW minimum code size
    minCodeSize = rng.read(1)[0] + 1
    out.write(minCodeSize.to_bytes(1, byteorder="little"))

    # Data blocks
    numDataBlocks = int(((imageWidth * imageHeight) / 256) + (((imageWidth * imageHeight) % 256) > 0))
    for i in range(numDataBlocks):
        blockSize = rng.read(1)[0] + 1
        out.write(blockSize.to_bytes(1, byteorder="little"))

        block = []
        while len(block) < blockSize:
            val = rng.read(1)[0]

            if val == 0:
                block += [val]
            elif val <= 127:
                block += list(range(1, val+1))
            else:
                block += list(reversed([val-1]))

        out.write(bytearray(block))

    return out.tell()

