#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header
    out.write(b'GIF89a')

    width = 1 + int.from_bytes(rng.read(1), 'big') % 256
    height = 1 + int.from_bytes(rng.read(1), 'big') % 256
    flags = b'\x07\x00' if (width == 1 or height == 1) else b'\x00\x00'
    out.write(struct.pack('<HH', width, height))
    out.write(flags)

    # Color table
    for _ in range(int.from_bytes(rng.read(1), 'big')):
        out.write(rng.read(3))

    # Image data
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x2C:
            # Image descriptor
            left = int.from_bytes(rng.read(1), 'big')
            top = int.from_bytes(rng.read(1), 'big')
            width = 1 + int.from_bytes(rng.read(1), 'big') % 256
            height = 1 + int.from_bytes(rng.read(1), 'big') % 256

            flags = [
                bool((int.from_bytes(rng.read(1), 'big') >> i) & 1)
                for i in reversed([4, 3, 2])
            ]

            lct = False
            interlace = False
            sort = False
            local_color_table = False

            if flags[0]:
                lct = bool((int.from_bytes(rng.read(1), 'big') >> 7) & 1)

                if not lct:
                    raise ValueError('LCT must be set when LCT flag is set')

            if flags[1]:
                interlace = bool((int.from_bytes(rng.read(1), 'big') >> 6) & 1)

            if flags[2]:
                sort = bool((int.from_bytes(rng.read(1), 'big') >> 5) & 1)

            if lct:
                local_color_table = bool((int.from_bytes(rng.read(1), 'big') >> 7) & 1)

            if local_color_table:
                for _ in range(int.from_bytes(rng.read(1), 'big')):
                    out.write(rng.read(3))

            lzw_min_size = int.from_bytes(rng.read(1), 'big')

            out.write(struct.pack('<HHBBBBB',
                                 left, top, width, height,
                                 *map(int, flags)))

            # Image data
            while True:
                block_size = int.from_bytes(rng.read(1), 'big')
                if block_size > 0:
                    out.write(rng.read(block_size))
                elif block_size == 0:
                    break
                else:
                    raise ValueError('Invalid block size')

        elif block_type == 0x21:
            # Extension block
            label = rng.read(1)[0]

            if label == 0xF9:
                # Graphic control extension
                out.write(rng.read(1))
                out.write(rng.read(1))
                out.write(rng.read(1))
                out.write(b'\x00')
            elif label == 0xFE:
                # Comment extension
                while True:
                    block_size = int.from_bytes(rng.read(1), 'big')
                    if block_size > 0:
                        out.write(rng.read(block_size))
                    elif block_size == 0:
                        break
                    else:
                        raise ValueError('Invalid block size')
            elif label == 0xFF:
                # Application extension
                app_id = rng.read(1)
                app_auth_code = rng.read(1)

                if app_id!= b'NETSCAPE':
                    raise ValueError('Unknown application ID')

                while True:
                    block_size = int.from_bytes(rng.read(1), 'big')

                    if block_size >= 3 and \
                            app_auth_code == b'\x03\x01' and \
                            rng.read(1) == b'\x00':

                        loop_count = int.from_bytes(rng.read(2), 'little')
                        out.write(app_id)
                        out.write(app_auth_code)
                        out.write(struct.pack('<H', block