#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng):
    # Header
    out.write(b'GIF89a')

    width = 1024
    height = 768
    color_table_size = 256
    background_color = 0x000000FF
    aspect_ratio = b'\x00\x00'

    out.write(struct.pack('<HHBBB',
        width, height,
        color_table_size,
        0xF7, 0x00))
    out.write(aspect_ratio)

    # Global Color Table
    for i in range(color_table_size):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)

        out.write(struct.pack('BBBB',
            r, g, b,
            0xFF))

    # Image Descriptor
    out.write(struct.pack('<BHHBBBBB',
        0x2C,
        width, height,
        0x00,
        0x00,
        0x00,
        background_color,
        0x00))

    # LZW Minimum Code Size (always 2 for this example)
    lzw_min_code_size = 2
    out.write(bytes([lzw_min_code_size]))

    # Pixel Data
    pixel_data = bytearray()

    # Generate some random data to compress
    for _ in range(width * height // 10):
        pixel_data += bytes([rng.randint(0, 255)])

    # Compress it with LZW
    compressed_pixel_data = compress_with_lzw(pixel_data, lzw_min_code_size)

    # Write the compressed data
    out.write(compressed_pixel_data)


# Compresses `data` using LZW with minimum code size of `min_code_size`.
# Returns the compressed data as a byte array.
def compress_with_lzw(data: bytes, min_code_size: int):
    # Initialize dictionary and clear code table
    dict_size = 2 ** min_code_size + 2
    dict = {}
    code_table = [None] * dict_size

    # Add all characters from input to the dictionary
    for c in data:
        if not c in dict:
            dict[c] = len(dict)

    # Start compression
    output = []
    prev_char = None
    cur_code_len = min_code_size + 1
    cur_code = -1
    max_code = 2 ** cur_code_len - 1

    for c in data:
        if c == prev_char or prev_char is None:
            cur_code += 1

            if cur_code > max_code:
                # Reset code length
                cur_code = 0

                while True:
                    cur_code_len += 1

                    if 2 ** cur_code_len <= dict_size:
                        break

                max_code = 2 ** cur_code_len - 1
        else:
            # Output previous character's code
            output.append(cur_code)

            # If we've reached maximum code length, reset it
            if cur_code_len >= 12:
                cur_code = 0
                cur_code_len = min_code_size + 1
                max_code = 2 ** cur_code_len - 1

            # Add new character to the dictionary
            if not c in dict:
                dict[c] = len(dict)

        prev_char = c

    # Output last character's code
    output.append(cur_code)

    # Output end-of-stream marker
    output.append(max_code + 1)

    # Convert codes to bytes
    return bytes([(output[i] >> 8) & 0xFF, output[i] & 0xFF])


if