#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' + bytes([0x12, 0x34])

    # Header
    out.write(header)

    # Logical screen descriptor
    width = int.from_bytes(rng.read(2), 'little') % (2**16 - 1) + 1
    height = int.from_bytes(rng.read(2), 'little') % (2**16 - 1) + 1
    flags = bytearray()
    flags += [int('b1', 2)]    # Global color table flag
    flags += [int('b1', 2)]    # Color resolution
    flags += [int('b1', 2)]    # Sort flag
    flags += [int('b3', 2)]    # Size of global color table
    out.write(struct.pack('<HHBBB', width, height, *flags))

    # Global color table
    for i in range(7):
        out.write(rng.read(3))

    # Application extension block
    out.write(b'\x21\xFF\x0BNETSCAPE2.0')
    out.write(b'\x03\x01\x00\x00')
    out.write(b'\x00')

    # Image descriptor
    left = 0
    top = 0
    width = int.from_bytes(rng.read(2), 'little') % (width - left) + left
    height = int.from_bytes(rng.read(2), 'little') % (height - top) + top
    flags = bytearray()
    flags += [int('b1', 2)]    # Local color table flag
    flags += [int('b3', 2)]    # Interlace flag
    flags += [int('b1', 2)]    # Sort flag
    flags += [int('b2', 2)]    # Reserved
    codesize = int.from_bytes(rng.read(1), 'little')
    out.write(struct.pack('<HHHHBB', left, top, width, height, *flags))

    # Local color table
    if flags[0] == 1:
        for i in range((1 << codesize)):
            out.write(rng.read(3))

    # LZW minimum code size
    lzwmincodesize = int.from_bytes(rng.read(1), 'little')
    out.write(lzwmincodesize.to_bytes(1, 'little'))

    # Data sub-block
    data = b''
    while True:
        chunktype = rng.read(1)[0]

        if chunktype == 0x00:   # End of image
            break

        elif chunktype == 0x01: # Plain text
            raise NotImplementedError("Plain text chunks are not supported")

        else:                   # LZW compressed sub-blocks
            length = int.from_bytes(rng.read(1), 'little')

            if length > 0:
                data += rng.read(length)

                if len(data) >= (1 << lzwmincodesize):
                    raise ValueError("LZW min code size is too small")

            else:               # Repeated pixels
                pixel = rng.read(1)[0]
                count = abs(length)
                data += pixel.to_bytes(1, 'little') * count

    # Write data sub-blocks to output
    while len(data)!= 0:
        length = min(len(data), 255)
        out.write(length.to_bytes(1, 'little'))
        out.write(data[:length])
        del data[:length]

    # Trailer
    out.write(b';')

