#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' + bytes([0x21]) + b'\xf9\x04'

    # 0-indexed color table size (1 << n)
    color_table_size = rng.read(1)[0] & 7

    # Global Color Table Flag
    global_color_table_flag = bool((rng.read(1)[0] >> 7) & 1)

    # Size of local color table
    if not global_color_table_flag:
        local_color_table_size = rng.read(1)[0] & 7

        # Local Color Table Flag
        local_color_table_flag = bool((rng.read(1)[0] >> 7) & 1)
    
    # Background Color Index
    background_color_index = rng.read(1)[0]

    # Pixel Aspect Ratio
    pixel_aspect_ratio = rng.read(1)[0]

    # Global Color Table
    global_color_table = [None for _ in range(pow(2, color_table_size))]
    for i in range(len(global_color_table)):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        global_color_table[i] = (r,g,b)

    # Image Descriptor
    image_descriptor = bytearray()
    image_descriptor += struct.pack('B', rng.read(1)[0])
    image_descriptor += struct.pack('<HHHBB', 
        rng.read(2), rng.read(2),
        0, 0, 
        0b10000 | rng.read(1)[0], 
    )

    # LZW Minimum Code Size
    lzw_minimum_code_size = rng.read(1)[0] & 0b1111

    # Block Terminator
    block_terminator = rng.read(1)[0]

    # Image Data
    image_data = []
    while True:
        data_block_length = rng.read(1)[0]
        if data_block_length == 0:
            break
        
        image_data += list(rng.read(data_block_length))

    # Write to output stream
    out.write(header)
    out.write(struct.pack('B', color_table_size))
    out.write(bytes([(int(global_color_table_flag) << 7) | int(local_color_table_flag)]))
    if not local_color_table_flag:
        out.write(struct.pack('B', local_color_table_size))
    out.write(struct.pack('B', background_color_index))
    out.write(struct.pack('B', pixel_aspect_ratio))
    for rgb in global_color_table:
        out.write(rgb)
    out.write(image_descriptor)
    out.write(struct.pack('B', lzw_minimum_code_size))
    out.write(block_terminator)
    for byte in image_data:
        out.write(byte)