#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    out.write(b'GIF89a')

    width = 1024
    height = 768
    bpp = 8
    color_table_size = (bpp == 8 and 256 or 1 << bpp) * 3
    global_color_table = True
    background_index = -1
    aspect_ratio = 0x40
    flags = 0x80 | (global_color_table and 0x80 or 0)
    palette_offset = 0
    screen_descriptor = struct.pack('<HHBBB',
            width, height, bpp, flags, background_index)
    out.write(screen_descriptor)

    if not global_color_table:
        raise NotImplementedError('local color tables are not supported yet')
    else:
        for i in range(color_table_size):
            r = rng.read(1)[0] % 256
            g = rng.read(1)[0] % 256
            b = rng.read(1)[0] % 256
            out.write(struct.pack('<BBB', r, g, b))

    data_block_offset = len(screen_descriptor) + \
                        color_table_size * 3 + \
                        1
    extension_block_offset = data_block_offset + 1
    image_descriptor = struct.pack('<BHHBB',
            0x2C, width, height, bpp, flags)
    out.write(image_descriptor)

    lzw_min_code_size = 2
    clear_code = 1 << lzw_min_code_size
    end_of_information_code = clear_code + 1
    next_code = clear_code + 2
    code_size = lzw_min_code_size + 1
    max_code_size = 12
    table = [i for i in range(clear_code)]
    output_buffer = []
    while True:
        if next_code < 2 ** code_size:
            table += [next_code]

        if next_code == end_of_information_code:
            break

        if next_code >= 2 ** max_code_size:
            raise Exception("Code size overflow")

        input_byte = rng.read(1)[0] & 0xFF
        if input_byte!= 0:
            output_buffer += [input_byte]
        elif len(output_buffer) > 0:
            output_buffer += [output_buffer[0]]

        if len(table) <= 2 ** code_size:
            code = table.index(output_buffer[0])
        else:
            code = 0

        for byte in output_buffer:
            out.write(bytes([byte]))

            if len(table) <= 2 ** code_size:
                new_entry = list(table[-1])
                new_entry[0] = code
                table += [new_entry]

                if len(table) > 2 ** max_code_size:
                    code_size += 1
                    max_code_size += 1

        del output_buffer[:]
        next_code += 1

    out.write(b'\x00\x3B')

