#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    header = b'GIF89a' + bytes([0x21]) + \
             b'\xf9\x04' + bytes([0x08, 0x00, 0x00])

    # Logical screen descriptor
    width = rng.read(1)[0] % (65536 - 100) + 100
    height = rng.read(1)[0] % (65536 - 100) + 100
    color_table_size = rng.read(1)[0] & 7
    global_color_table = rng.read(
        3 * ((1 << color_table_size) if color_table_size else 0))
    background_index = rng.read(1)[0] % len(global_color_table)
    pixel_aspect_ratio = rng.read(1)[0] & 7

    logical_screen_descriptor = struct.pack('>HHBBB',
                                            width, height,
                                            global_color_table[background_index],
                                            color_table_size,
                                            pixel_aspect_ratio)

    # Global color table
    palette = [b'\x00'] * (len(global_color_table) // 3)
    for i in range(len(palette)):
        palette[i] += global_color_table[(i*3)+1:(i*3)+4]

    # Image descriptor
    image_left = rng.read(1)[0] % (width - 100)
    image_top = rng.read(1)[0] % (height - 100)
    image_width = rng.read(1)[0] % (width - image_left - 100) + 100
    image_height = rng.read(1)[0] % (height - image_top - 100) + 100
    local_color_table_flag = bool(rng.read(1)[0] & 1)
    interlace_flag = bool(rng.read(1)[0] & 2)
    sort_flag = bool(rng.read(1)[0] & 4)
    reserved = rng.read(2)
    lzw_minimum_code_size = rng.read(1)[0] & 0x0f
    lzw_minimum_code_size += 1

    image_descriptor = struct.pack('>HHHHBB',
                                   image_left, image_top,
                                   image_width, image_height,
                                   local_color_table_flag,
                                   interlace_flag,
                                   sort_flag)

    # Local color table
    if local_color_table_flag:
        local_color_table_size = rng.read(1)[0] & 7
        local_color_table = rng.read(
            3 * ((1 << local_color_table_size) if local_color_table_size else 0))

        palette += [b'\x00'] * (len(local_color_table) // 3)
        for i in range(len(palette)-len(global_color_table)//3):
            palette[i+len(global_color_table)//3] += local_color_table[(i*3)+1:(i*3)+4]

        image_descriptor += struct.pack('B',
                                        local_color_table_size)
        image_descriptor += local_color_table

    # LZW minimum code size
    image_descriptor += struct.pack('B',
                                    lzw_minimum_code_size)

    # Data sub-block
    data_sub_block = bytearray()
    while True:
        block_type = rng.read(1)[0]
        if block_type == 0x2c or block_type == 0x21:
            break
        elif block_type!= 0x21:
            continue
        else:
            raise Exception("Unsupported block type")

    data_sub_block += struct.pack('B',
                                  block_type)

    while True:
        block_length = rng.read(1)[0]
        if not block_length:
            break
        data_sub_block += struct.pack('B',
                                      block_length)
        data_sub_block += rng.read(block_length)

    # Trailer
    trailer = b''

    # Write to output stream
    out.write(header)
    out.write(logical_screen_descriptor)
    out.write(bytes(palette))
    out.write(image_descriptor)
    out.write(data_sub_block)
    out.write(trailer)