#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1

    # Global color table
    global_color_table_size = int.from_bytes(rng.read(1), byteorder='little') & 0x7f
    if global_color_table_size == 0:
        global_color_table_size = 1 << 8
    elif global_color_table_size == 1:
        return

    global_color_table = [int.from_bytes(rng.read(3), byteorder='big') for _ in range(global_color_table_size)]

    # Image descriptor
    image_descriptor = bytes([
        0b10000000 | ((width >> 8) & 0b00001111),
        (width       ) & 0xff,
        (height      ) & 0xff,
        (height >>  8) & 0xff,
        0,
        0,
        0,
        0,
    ])

    # Local color table
    local_color_table_flag = bool(int.from_bytes(rng.read(1), byteorder='little'))
    if not local_color_table_flag:
        local_color_table = []
    else:
        local_color_table_size = int.from_bytes(rng.read(1), byteorder='little') & 0x7f
        if local_color_table_size == 0:
            local_color_table_size = 1 << 8

        local_color_table = [int.from_bytes(rng.read(3), byteorder='big') for _ in range(local_color_table_size)]

    # LZW minimum code size
    lzw_minimum_code_size = int.from_bytes(rng.read(1), byteorder='little') & 0x0f + 2

    # Image data
    image_data = bytes()
    while len(image_data) < width * height:
        block_type = int.from_bytes(rng.read(1), byteorder='little') & 0x03

        if block_type == 0:
            # Raw data sub-block
            length = int.from_bytes(rng.read(1), byteorder='little')

            image_data += rng.read(length)
        elif block_type == 1:
            # Clear code sub-block
            pass
        elif block_type == 2:
            # End of information sub-block
            break
        elif block_type == 3:
            # Code table sub-block
            length = int.from_bytes(rng.read(1), byteorder='little')
            codes = [(
                int.from_bytes(rng.read(1), byteorder='little'),
                int.from_bytes(rng.read(1), byteorder='little'),
            ) for _ in range(length)]

            # TODO: Actually use this code table somehow...
            print('Code table:', codes)

    # Output everything to `out`
    out.write(header)
    out.write(struct.pack('<H', width))
    out.write(struct.pack('<H', height))
    out.write(struct.pack('<B', global_color_table_size - 1))
    out.write(b'\x00' * 3)
    out.write(image_descriptor)

    if local_color_table_flag:
        out.write(bytes([local_color_table_size - 1]))
        out.write(b'\x00' * 3)

    out.write(bytes([lzw_minimum_code_size]))

    out.write(image_data)

