#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (e.g., /dev/urandom).
# The width and height of the image are given by `width` and `height`.
# The color depth is given by `colordepth`, which must be 1 or 2.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int, colordepth: int):

    # Check that the color depth is valid
    if not ((colordepth == 1) or (colordepth == 2)):
        raise ValueError('Invalid color depth.')

    # Write the header
    out.write(b'GIF89a')

    # Write the logical screen descriptor
    out.write(struct.pack('<HHBBB', width, height,
            0xf7 & (colordepth - 1), 0x00, 0x00))

    # Write the global color table
    for _ in range(int((2**(colordepth * 3)) / 8)):
        out.write(rng.read(1))

    # Write the graphics control extension
    out.write(b'\x21\xF9\x04\x05\x00\x00\x00\x00\x00\x00\x00')

    # Write the application extension
    out.write(b'\x21\xFF\x0BNETSCAPE2.0')
    out.write(b'\x03\x01\x00\x00')
    out.write(b'\x00\x00')

    # Write the image descriptor
    out.write(struct.pack('<BBHHHBB', 0x2c, 0x00, 0x00,
            width, height, 0x00, 0x00, 0x00))

    # Write the local color table
    for _ in range(int((2**(colordepth * 3)) / 8)):
        out.write(rng.read(1))

    # Write the image data
    while True:
        chunklen = rng.read(1)[0]

        if chunklen == 0x00:
            break

        out.write(chunklen.to_bytes(1, 'little'))
        out.write(rng.read(chunklen))