#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the magic bytes and version string
    out.write(b'GIF89a')

    # Generate a random screen descriptor
    width  = (1 << rng.read(1)[0]) % 500
    height = (1 << rng.read(1)[0]) % 500
    flags  = rng.read(1)[0] & 7
    bgcol  = rng.read(1)[0]
    aspect = rng.read(1)[0]

    # Write the screen descriptor to the output stream
    out.write(struct.pack('>HHBBB', width, height, flags, bgcol, aspect))

    # Generate a random global color table
    gctsize = ((flags >> 2) & 7) + 1
    gct = [None for _ in range(gctsize)]
    for i in range(gctsize):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        gct[i] = (r, g, b)

    # Write the global color table to the output stream
    for entry in gct:
        out.write(struct.pack('BBB', *entry))

    # Generate a random local color table
    lctsize = 2 ** (((flags >> 4) & 7) + 1)
    lct = [None for _ in range(lctsize)]
    for i in range(lctsize):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        lct[i] = (r, g, b)

    # Write the local color table to the output stream
    if len(lct) > 0:
        out.write(struct.pack('B', lctsize - 1))
        for entry in lct:
            out.write(struct.pack('BBB', *entry))

    # Generate a random image descriptor
    left   = rng.read(2)[0]
    top    = rng.read(2)[0]
    width  = (1 << rng.read(1)[0]) % 500
    height = (1 << rng.read(1)[0]) % 500
    flags  = rng.read(1)[0]
    lctidx = rng.read(1)[0]

    # Write the image descriptor to the output stream
    out.write(struct.pack('>HHHHBB', left, top, width, height, flags, lctidx))

    # Generate a random LZW code size
    mincodesize = rng.read(1)[0]
    maxcodesize = rng.read(1)[0]
    codesize = rng.read(1)[0]
    while not (mincodesize <= codesize <= maxcodesize):
        codesize = rng.read(1)[0]

    # Write the LZW code size to the output stream
    out.write(struct.pack('BB', mincodesize, maxcodesize))

    # Generate a random LZW dictionary
    dictlen = 2 ** codesize
    dict = {chr(i).encode(): chr((i + 1) % dictlen).encode() for i in range(dictlen)}

    # Write the LZW dictionary to the output stream
    for key, value in dict.items():
        out.write(key + value)

    # Generate a random LZW block
    blocklen = rng.read(1)[0]
    block = []
    while True:

        # Read a new character from the input stream
        char = rng.read(1)
        if not char or char == b'\x00': break

        # If the character is already in the dictionary...
        if char in dict:

            #... write it directly to the output stream
            out.write(char)

        else:

            # Otherwise, find the longest prefix of the character that's also in the dictionary
            match = ''
            for k in reversed(range(len(char))):
                candidate = char[:k+1].decode()
                if candidate in dict:
                    match = candidate
                    break

            # Write the matching prefix to the output stream
            out.write(match.encode())

            # Add the remaining suffix to the dictionary
            dict[char] = chr((ord(dict[match]) + 1) % dictlen).