#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = rng.randint(1024, 65536) // 16 * 16
    height = rng.randint(1024, 65536) // 16 * 16
    flags  = (width >> 7) & 1 | ((height >> 7) & 1) << 1
    global_color_table_flag = bool(rng.randint(0, 1))
    background_color_index   = rng.randint(0, 255)
    pixel_aspect_ratio       = 0
    out.write(struct.pack('>HHBBB',
                           width, height,
                           flags,
                           global_color_table_flag,
                           background_color_index,
                           pixel_aspect_ratio
                          ))

    # Global color table
    if global_color_table_flag:
        for i in range(0, 256):
            r = rng.randint(0, 255)
            g = rng.randint(0, 255)
            b = rng.randint(0, 255)
            out.write(struct.pack('BBB',
                                   r, g, b
                                  ))

    # Image descriptor
    left      = 0
    top       = 0
    image_width  = width
    image_height = height
    flags     = 0x80
    local_color_table_flag = False
    interlace_flag         = False
    sort_flag              = False
    reserved               = 0
    lct_size               = 0
    out.write(struct.pack('>HHHHBBBBB',
                           left, top,
                           image_width, image_height,
                           flags,
                           local_color_table_flag,
                           interlace_flag,
                           sort_flag,
                           reserved,
                           lct_size
                          ))

    # Local color table
    if local_color_table_flag:
        for i in range(0, 256):
            r = rng.randint(0, 255)
            g = rng.randint(0, 255)
            b = rng.randint(0, 255)
            out.write(struct.pack('BBB',
                                   r, g, b
                                  ))

    # Table based image data
    lzwmincodesize = 2
    out.write(struct.pack('B', lzwmincodesize))
    clear_code = 1 << lzwmincodesize
    end_of_information_code = clear_code + 1
    next_code = clear_code + 2
    code_mask = (1 << lzwmincodesize) - 1
    max_code_size = 12
    bit_count = 0
    buf = []
    while True:
        c = rng.read(1)[0]

        if c == 0x00 or len(buf) >= 255:
            break

        buf += [c]

        if bit_count < 8:
            bit_count += 8
            continue

        if not (next_code <= max_code_size and
                next_code > clear_code):

            raise Exception("Invalid LZW code")

        out.write(struct.pack('B', next_code - clear_code))

        if next_code == end_of_information_code:
            break

        if next_code == clear_code + 1:
            lzwmincodesize = 2
            out.write(struct.pack('B', lzwmincodesize))
            clear_code = 1 << lzwmincodesize
            end_of_information_code = clear_code + 1
            next_code = clear_code + 2
            code_mask = (1 << lzwmincodesize) - 1
            max_code_size = 12
            bit_count = 0
            continue

        old_code = buf[0]
        buf = buf[1:]

        if next_code == max_code_size:
            out.write(struct.pack('B', old_code))
            lzwmincodesize += 1
            max_code_size = 1 << lzwmincodesize
            code_mask = (1 << lzwmincodesize) - 1
            continue

        new_code = None