#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 << (int.from_bytes(rng.read(1), 'big') & 0x07)
    height = 1 << ((int.from_bytes(rng.read(1), 'big') >> 3) & 0x07)
    flags = int.from_bytes(rng.read(1), 'big')
    bgcol = int.from_bytes(rng.read(1), 'big')
    aspect = int.from_bytes(rng.read(1), 'big')
    out.write(struct.pack('>HHBBB', width, height, flags, bgcol, aspect))

    # Global color table
    gctsize = 2 ** ((flags & 0x07) + 1)
    for i in range(gctsize):
        r = int.from_bytes(rng.read(1), 'big')
        g = int.from_bytes(rng.read(1), 'big')
        b = int.from_bytes(rng.read(1), 'big')
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left = int.from_bytes(rng.read(2), 'big')
    top = int.from_bytes(rng.read(2), 'big')
    width = 1 << (int.from_bytes(rng.read(1), 'big') & 0x07)
    height = 1 << ((int.from_bytes(rng.read(1), 'big') >> 3) & 0x07)
    flags = int.from_bytes(rng.read(1), 'big')
    lctsize = 2 ** (((flags >> 4) & 0x07) + 1)
    interlace = bool((flags >> 6) & 0x01)
    sortflag = bool((flags >> 7) & 0x01)
    reserved = int.from_bytes(rng.read(1), 'big')
    lct = [None] * lctsize
    if lctsize > 0:
        for i in range(lctsize):
            r = int.from_bytes(rng.read(1), 'big')
            g = int.from_bytes(rng.read(1), 'big')
            b = int.from_bytes(rng.read(1), 'big')
            lct[i] = (r, g, b)
    else:
        lct = None
    out.write(struct.pack('>HHBBBBBB', left, top, width, height, flags,
                          lctsize, interlace, sortflag, reserved))

    # Local color table
    for i in range(lctsize):
        r = int.from_bytes(rng.read(1), 'big')
        g = int.from_bytes(rng.read(1), 'big')
        b = int.from_bytes(rng.read(1), 'big')
        out.write(struct.pack('BBB', r, g, b))

    # Image data
    while True:
        blocktype = int.from_bytes(rng.read(1), 'big')

        if blocktype == 0x2C:
            break

        elif blocktype == 0x21:
            pass

        elif blocktype == 0x3B:
            return

        else:
            raise Exception("Invalid block type")

    while True:
        blocklen = int.from_bytes(rng.read(1), 'big')
        if not blocklen:
            break
        out.write(rng.read(blocklen))