#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write('GIF89a'.encode())

    # Logical screen descriptor
    width  = rng.randint(1024, 65536)
    height = rng.randint(1024, 65536)
    flags  = (
            0b10000 <<  7 |    # Global color table flag
            0b00000 <<  4 |    # Color resolution
            0b00000 <<  3 |    # Sorting method
            0b00000 <<  2     # Size of global color table
        ).to_bytes(1, 'big')
    bgcolorindex = 0x00.to_bytes(1, 'big')
    pixelaspectratio = 0x00.to_bytes(1, 'big')
    out.write(struct.pack('<HHBBB', width, height, flags[0], bgcolorindex[0], pixelaspectratio[0]))

    # Global color table
    for _ in range(0, 2**int((flags[0] & 0b00000001)*3+1)):
        r = rng.randint(0, 255).to_bytes(1, 'big')
        g = rng.randint(0, 255).to_bytes(1, 'big')
        b = rng.randint(0, 255).to_bytes(1, 'big')
        out.write(r + g + b)

    # Image descriptor
    left      = 0x00.to_bytes(2, 'big')
    top       = 0x00.to_bytes(2, 'big')
    width     = width.to_bytes(2, 'big')
    height    = height.to_bytes(2, 'big')
    flags     = (
            0b10000000 << 7 | # Local color table flag
            0b00000000 << 6 | # Interlace flag
            0b00000000 << 5 | # Sorting flag
            0b00000000 << 4 | # Reserved
            0b00000000 << 3 | # Reserved
            0b00000000 << 2 | # Reserved
            0b00000000 << 1 | # Reserved
            0b00000000        # LZW minimum code size
        ).to_bytes(1, 'big')
    bgcolorindex = 0x00.to_bytes(1, 'big')
    pixelaspectratio = 0x00.to_bytes(1, 'big')
    out.write(left + top + width + height + flags + bgcolorindex + pixelaspectratio)

    # Table based image data
    lzwminsize = flags[-1:]
    while True:
        blocktype = rng.read(1)[0]

        if blocktype == 0x2c: # Image descriptor
            break

        elif blocktype == 0x21: # Extension
            extlabel = rng.read(1)[0]

            if extlabel == 0xf9:
                # Graphic control extension
                pass

            else:
                raise Exception("Unknown extension label")

        elif blocktype == 0x3b: # Trailer
            return

        else:
            raise Exception("Unknown block type")


    # Plain text extension
    extlabel = 0x01.to_bytes(1, 'big')
    authcode = bytes([rng.read(1)[0] for i in range(0, 3)])
    txtlen = rng.read(1)[0]
    txtdat = bytes([rng.read(1)[0] for i in range(0, txtlen)])
    out.write(extlabel + authcode + txtdat)

    # Application extension
    extlabel = 0xff.to_bytes(1, 'big')
    appid = bytes([rng.read(1)[0] for i in range(0, 11)])
    appauthcode = bytes([rng.read(1)[0] for i in range(0, 3)])
    appdat = bytes([rng.read(1)[0] for i in range(0, 3)])
    out.