#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom or /dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The first 6 bytes are always 'GIF87a', but we can't just write that to
    # the output because it's not valid ASCII. Instead, we'll use the fact
    # that Python will automatically convert any integer to its string
    # representation when concatenated with another string.
    out.write('GIF87a'.encode())

    # Next comes the width and height of the image as two unsigned shorts in
    # big endian order. We'll choose them at random from [10..250]x[10..250].
    w = struct.pack('>H', int.from_bytes(rng.read(2), byteorder='big'))
    h = struct.pack('>H', int.from_bytes(rng.read(2), byteorder='big'))
    out.write(w + h)

    # Then there is a single byte for the color table size - we'll set it to
    # zero since our images have no palette.
    ctsize = struct.pack('B', 0)
    out.write(ctsize)

    # Finally, we get to the actual pixel data. This consists of an infinite
    # sequence of subblocks, each containing either a graphic control block or
    # a frame. Each subblock starts with a byte indicating how many bytes follow
    # until the end of this subblock. If the value is less than 0xF9 then it's
    # the length of the entire subblock; if it's greater than 0xF9 then it's only
    # the last byte of the subblock which contains the remaining length.
    while True:

        # First, let's pick a random length between 4 and 128 inclusive.
        l = rng.read(1)[0] % 124 + 4

        # Now we need to decide whether to put the length in one byte or two.
        # To do so, we'll take the least significant bit of the length and add
        # it to itself shifted left by one position. If the result is equal to
        # 0xFF then we'll use two bytes instead of one.
        if ((l & 1) + (l >> 1)) == 0xff:
            out.write(struct.pack('B', 0xf9 | (l >> 8)))
            out.write(struct.pack('B', l & 0xff))
        else:
            out.write(struct.pack('B', l))

        # Now we're ready to actually write some pixels! Let's start with the
        # background color index. It's stored as a single byte.
        bgidx = rng.read(1)[0]
        out.write(struct.pack('B', bgidx))

        # Next up is the aspect ratio - again, a single byte where the least
        # significant bit indicates whether the aspect ratio should be treated
        # as a whole number rather than a fractional one.
        aspect = rng.read(1)[0]
        out.write(struct.pack('B', aspect))

        # After that come the global color table - we don't want one, so we'll
        # just skip over all the colors.
        for _ in range(2 ** (aspect & 7)):
            out.write(b'\x00\x00\x00')

        # Now comes the local color table - again, we don't care about it, so
        # we'll just skip over all the colors.
        for _ in range(2 ** ((aspect & 7) + 1)):
            out.write(b'\x00\x00\x00')

        # Now comes the LZW minimum code size - again, we won't compress the
        # data, so we'll just set it to zero.
        mincodesize = struct.pack('B', 0)
        out.write(mincodesize)

        # And now comes the actual pixel data. For simplicity, we'll just copy
        # the input stream verbatim.
        out.write(rng.read(l-1))

        # That was the last subblock, so we're done here.
        if l < 0xf9: break


if __name__ == '__main__':
    # Open the RNG source. If the user specified a filename on the command line
    # then open that file, otherwise try opening /dev/ur