#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (1 << rng.read(1)[0]) | ((1 << rng.read(1)[0]) >> 7)
    height = (1 << rng.read(1)[0]) | ((1 << rng.read(1)[0]) >> 7)
    flags  = rng.read(1)[0]
    bg     = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(struct.pack('>HHBBB', width, height, flags, bg, aspect))

    # Global color table
    global_color_table_size = rng.read(1)[0] & 0x07
    if global_color_table_size!= 0:
        global_color_table = [
                rng.read(1)[0] for _ in range(global_color_table_size**2)]
        out.write(bytes(global_color_table))

    # Image descriptor
    left   = (1 << rng.read(1)[0]) | ((1 << rng.read(1)[0]) >> 7)
    top    = (1 << rng.read(1)[0]) | ((1 << rng.read(1)[0]) >> 7)
    width  = (1 << rng.read(1)[0]) | ((1 << rng.read(1)[0]) >> 7)
    height = (1 << rng.read(1)[0]) | ((1 << rng.read(1)[0]) >> 7)
    flags  = rng.read(1)[0]
    local_color_table_size = rng.read(1)[0] & 0x07
    lct = []
    if local_color_table_size!= 0:
        lct = [rng.read(1)[0] for _ in range(local_color_table_size ** 2)]
    interlace = bool((flags & 0x40) > 0)
    sort      = bool((flags & 0x20) > 0)
    reserved  = rng.read(1)[0]
    lct_flag  = bool((flags & 0x80) > 0)
    out.write(struct.pack('>HHHHBB', left, top, width, height, flags,
                          local_color_table_size))
    if len(lct):
        out.write(bytes(lct))

    # Table based image data
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x2C:
            break

        elif block_type == 0x21:
            ext_label = rng.read(1)[0]

            if ext_label == 0xF9:
                size = rng.read(1)[0]

                while size >= 0:
                    subblock = rng.read(1)[0]

                    if subblock == 0x01:
                        delay = (1 << rng.read(1)[0]) | \
                                ((1 << rng.read(1)[0]) >> 7)

                        transparent_index = rng.read(1)[0]
                        terminator        = rng.read(1)[0]

                        assert terminator == 0, 'Invalid terminator'

                        out.write(struct.pack('>BHB', ext_label,
                                              size, delay))
                        out.write(struct.pack('>BBB', transparent_index,
                                              0, 0))

                    else:
                        raise ValueError('Unknown extension label')

                    size -= 1

            else:
                raise ValueError('Unknown extension label')

        else:
            raise ValueError('Unknown block type')


    # Plain text extension
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x21:
            ext_label = rng.read(1)[0]

            if ext_label == 0x01:
                size = rng.read(1)[0] - 1
                text = bytes([rng.read(1)[0] for _ in range(size)])
                terminator = rng.read(1)[0]

                assert terminator == 0, 'Invalid terminator'

                out.write(struct.pack('>BB', ext_label,