#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The width and height of the image are determined by
# `width` and `height`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int
    ):

    # Write the header
    out.write(b'GIF89a')

    # Write the logical screen descriptor
    out.write(struct.pack('>HHBBB', width, height, 0xF7, 0x00))

    # Write the global color table (black only)
    for i in range(256):
        out.write(b'\x00\x00\x00')

    # Write the first frame
    write_frame(rng, out, width, height)


# Writes one frame to `out`, which is assumed to be positioned at the start
# of the frame. This function assumes that the global color table has already
# been written.
def write_frame(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int
    ):

    # Write the local color table
    for i in range(256):
        out.write(b'\x00\x00\x00')

    # Write the graphics control extension
    out.write(b'\x21\xF9\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF')

    # Write the image descriptor
    out.write(struct.pack('>HHHHB', 0, 0, width, height, 0x00))

    # Write the LZW minimum code size
    out.write(b'\x80')

    # Write the compressed data
    lzw_compress(rng, out, width * height)

    # Write the end-of-image marker
    out.write(b'\x00')


# Compresses the input from `rng` into `out` using the LZW algorithm.
# The length of the compressed data must be known beforehand.
def lzw_compress(
        rng: BinaryIO, out: BinaryIO,
        len: int
    ):

    # Initialize the dictionary with all possible codes
    dict = {}
    for i in range(256):
        dict[i] = [i]

    # Read the initial code
    code = ord(rng.read(1)) << 8 | ord(rng.read(1))

    # Write it as a literal character
    out.write(code.to_bytes(2, 'little'))

    # Read the next two bytes
    while len > 0:

        # Get the current code
        cur = code >> 8

        # If this code exists in our dictionary...
        if cur in dict:

            #...then get its prefix
            pre = dict[cur][0]

            # If we have a new prefix...
            if pre!= cur:

                #...add it to the dictionary
                dict[len(dict)] = dict[cur] + [pre]

            # Set the code to the prefix's value
            code = pre << 8 | ord(rng.read(1))

            # Write the prefix as a literal character
            out.write((pre & 0xff).to_bytes(1, 'little'))

        else:

            # Add the new code to the dictionary
            dict[len(dict)] = [code]

            # Write the code as a literal character
            out.write(((code >> 8) & 0xff).to_bytes(1, 'little'))

        # Decrement the remaining length
        len -= 1