#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Generate the header and write it to the output stream
    header = b'GIF89a'
    out.write(header)

    # Generate the logical screen descriptor and write it to the output stream
    width = 1024
    height = 768
    color_table_size = 0
    global_color_table = False
    background_color = 0
    pixel_aspect_ratio = 0
    lsd = struct.pack('>HHBBB', width, height,
                      color_table_size, global_color_table,
                      background_color, pixel_aspect_ratio)
    out.write(lsd)

    # Generate the global color table (if any) and write it to the output stream
    if global_color_table == True:
        for i in range(pow(2, color_table_size)):
            r = rng.read(1)[0] % 256
            g = rng.read(1)[0] % 256
            b = rng.read(1)[0] % 256
            entry = struct.pack('BBB', r, g, b)
            out.write(entry)

    # Generate the image descriptor and write it to the output stream
    left = 0
    top = 0
    width = 1024
    height = 768
    local_color_table = False
    interlace = False
    sort = False
    reserved = 0
    lzw_min_code_size = 2
    id = struct.pack('>HHHHBB', left, top,
                     width, height, local_color_table,
                     interlace << 2 | sort << 1 | reserved)
    out.write(id)

    # Generate the local color table (if any) and write it to the output stream
    if local_color_table == True:
        for i in range(pow(2, color_table_size)):
            r = rng.read(1)[0] % 256
            g = rng.read(1)[0] % 256
            b = rng.read(1)[0] % 256
            entry = struct.pack('BBB', r, g, b)
            out.write(entry)

    # Generate the LZW minimum code size and write it to the output stream
    out.write(struct.pack('B', lzw_min_code_size))

    # Generate the compressed image data and write it to the output stream
    num_pixels = width * height
    bytes_per_pixel = 1
    clear_code = pow(2, lzw_min_code_size)
    end_of_info_code = clear_code + 1
    end_of_image_code = end_of_info_code + 1
    max_code_size = lzw_min_code_size + 1
    min_code_size = lzw_min_code_size + 1
    current_code_size = lzw_min_code_size + 1
    next_code = end_of_info_code + 1
    codes = []
    buffer = ''
    while len(codes) < num_pixels:

        # If we have enough bits in our buffer, read another byte from the input
        if len(buffer) >= current_code_size:

            # Read one byte at a time until we get enough bits
            newbyte = rng.read(1)[0]
            for i in range(8):

                # Add this bit to the buffer
                buffer += str((newbyte >> (7 - i)) & 1)

                # If we now have enough bits, process them
                if len(buffer) >= current_code_size:

                    # Get the code corresponding to these bits
                    code = int(buffer[:current_code_size], 2)
                    buffer = buffer[current_code_size:]

                    # If the code is not a valid prefix of some other code, add it to the list
                    if code > next_code or \
                            (len(codes) <= next_code and
                             code!= end_of_info_code):
                        codes.append(code)

                        # If that was the last code, we're done!
                        if code == end_of_image_code:
                            break

                        # If the code is equal to the maximum possible value for this code size,
                        # increase the code size by one
                        if