#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header and logical screen descriptor
    header = b'GIF89a'
    width  = (1 << rng.read(1)[0]) % 65536
    height = (1 << rng.read(1)[0]) % 65536
    flags  = bytes([
        0b00000000 | rng.read(1)[0] & 0b00000011,
        rng.read(1)[0],
        rng.read(1)[0],
        0b00000000 | rng.read(1)[0] & 0b00000011
    ])
    background = rng.read(1)[0]
    aspect     = rng.read(1)[0]
    out.write(header + struct.pack('>HHBBB', width, height, *flags))
    out.write(struct.pack('B', background))
    out.write(struct.pack('B', aspect))

    # Global color table
    global_color_table_size = rng.read(1)[0]
    if global_color_table_size > 0:
        out.write(struct.pack('<B', global_color_table_size))

        for _ in range(global_color_table_size):
            r   = rng.read(1)[0]
            g   = rng.read(1)[0]
            b   = rng.read(1)[0]
            out.write(struct.pack('BBB', r, g, b))

    # Image block
    image_left      = rng.read(2)
    image_top       = rng.read(2)
    image_width     = rng.read(2)
    image_height    = rng.read(2)
    packed          = rng.read(1)[0]
    local_color_map = bool((packed >> 7) & 1)
    interlace       = bool((packed >> 6) & 1)
    sort            = bool((packed >> 5) & 1)
    reserved        = bool((packed >> 4) & 1)
    lct_size        = ((packed & 0x0F) << 8) | rng.read(1)[0]
    lct             = []
    for i in range(lct_size):
        r   = rng.read(1)[0]
        g   = rng.read(1)[0]
        b   = rng.read(1)[0]
        lct += [r, g, b]
    out.write(image_left + image_top + image_width + image_height +
              struct.pack('B', packed))
    if local_color_map:
        out.write(struct.pack('<B', lct_size))
        for rgb in lct:
            out.write(struct.pack('B', rgb))

    # Table based image data
    while True:
        code = rng.read(1)[0]
        if code == 0x2C:
            break
        elif code == 0x21:
            subcode = rng.read(1)[0]
            if subcode == 0xF9:
                length = rng.read(1)[0]
                out.write(struct.pack('BB', code, subcode))
                out.write(length)

                # Application extension
                app_block_size = rng.read(1)[0]
                app_id         = rng.read(app_block_size)
                app_auth_code  = rng.read(1)[0]
                out.write(struct.pack('B', app_block_size))
                out.write(app_id)
                out.write(struct.pack('B', app_auth_code))

            else:
                raise ValueError("Unknown image control block")

        elif code == 0x3B:
            continue
        else:
            raise ValueError("Unknown image block")


    # Plain text extension
    plain_text_extension_label = rng.read(1)[0]
    plain_text_extension_block_size = rng.read(1)[0]
    plain_text_extension_text_grid_left = rng.read(2)
    plain_text_extension_text_grid_top = rng.read(2)
    plain_text_extension_text_