#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (e.g., /dev/urandom).
# The generated GIF will have a width of 1024 pixels and
# a height of 768 pixels.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    out.write(struct.pack('>HHBBB',
        1024, 768, 0x80, 0x00, 0x00))

    # Global color table
    for i in range(0, 256):
        r = rng.read(1)[0] & 0xf0 | rng.read(1)[0] >> 4
        g = rng.read(1)[0] & 0xf0 | rng.read(1)[0] >> 4
        b = rng.read(1)[0] & 0xf0 | rng.read(1)[0] >> 4
        out.write(struct.pack('BBBB', r, g, b, 0xff))

    # Image descriptor
    out.write(struct.pack('>HHHHB',
        0, 0, 1024, 768, 0x00))

    # Local color table
    for i in range(0, 256):
        r = rng.read(1)[0] & 0xf0 | rng.read(1)[0] >> 4
        g = rng.read(1)[0] & 0xf0 | rng.read(1)[0] >> 4
        b = rng.read(1)[0] & 0xf0 | rng.read(1)[0] >> 4
        out.write(struct.pack('BBBB', r, g, b, 0xff))

    # LZW minimum code size
    out.write(b'\x08')

    # Image data
    lzw_mincodesize = 8
    lzw_maxcodesize = 12
    lzw_clearcode = 1 << lzw_mincodesize
    lzw_endofinfo = lzw_clearcode + 1
    lzw_endofdata = lzw_endofinfo + 1
    lzw_nextcode = lzw_endofdata + 1
    lzw_prevcode = None
    lzw_curcode = None
    lzw_curbits = 0
    lzw_curbyte = 0
    lzw_curbit = 0
    lzw_buf = []
    while True:

        # Read a new bit if we need to
        if not lzw_curbit:

            # Read another byte from the input stream?
            if len(lzw_buf) == 0:
                buf = rng.read(1)

                # End of image?
                if len(buf) < 1:
                    break

                # Add it to our buffer
                lzw_buf += list(buf)

            # Get the next byte
            lzw_curbyte = lzw_buf[0]
            del lzw_buf[0]

            # Increment the current bit position
            lzw_curbit = 8

        # Extract the next bit
        lzw_curbits >>= 1
        if ord(lzw_curbyte) & (1 << (7 - lzw_curbit)):
            lzw_curbits |= 0x8000

        # Is this the last bit?
        if lzw_curbit == 1:

            # Decode the code
            if lzw_curcode == lzw_endofdata:
                pass
            elif lzw_curcode == lzw_endofinfo:
                pass
            elif lzw_curcode == lzw_clearcode:
                lzw_curcode = lzw_endofinfo
                lzw_curbits = 0
                lzw_curbyte = 0
                lzw_curbit = 0
            else:
                if lzw_prevcode is None or \
                        lzw_curcode >= lzw_nextcode:
                    lzw_buf += [chr(lzw_curbyte)]
                    continue
                lzw_buf += [chr(lzw_curbyte)]
                lzw_buf += [chr(lzw_curbyte)]
                lzw_buf += [chr(lzw_curbyte)]

            # Reset state variables
            lzw_prevcode = lzw_curcode
            lzw_curcode = None