#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (e.g., /dev/urandom).
# The generated GIF will have 10 frames and each frame
# will be of size 256x256 pixels with 4 bits per pixel.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Write the GIF header to the output stream
    def write_header():
        out.write(b'GIF89a')

    # Write the logical screen descriptor to the output stream
    def write_logical_screen_descriptor():
        width = 256
        height = 256

        packed = struct.pack('>HHBBB',
                width, height,
                0xf7 | 0x80, # Global color table flag set, sort flag clear
                0,            # Background color index
                0             # Pixel aspect ratio
        )
        out.write(packed)

    # Write the global color table to the output stream
    def write_global_color_table():
        palette = [
            0xff, 0xff, 0xff,    # White
            0x00, 0x00, 0x00     # Black
        ] * 256
        for i in range(len(palette)):
            out.write(struct.pack('B', palette[i]))

    # Write the image descriptor to the output stream
    def write_image_descriptor():
        left = 0
        top = 0
        width = 256
        height = 256

        flags = 0xc0 | 0x00 # Local color table flag set, interlace flag clear, sorted flag clear
        packed = struct.pack('>HHBBBBBB',
                left, top,
                width, height,
                flags,
                0,      # LZW minimum code size
                0       # No local color table
        )
        out.write(packed)

    # Write the local color table to the output stream
    def write_local_color_table():
        palette = [
            0xff, 0xff, 0xff,    # White
            0x00, 0x00, 0x00     # Black
        ] * 256
        for i in range(len(palette)):
            out.write(struct.pack('B', palette[i]))

    # Write the graphics control extension block to the output stream
    def write_graphics_control_extension():
        packed = struct.pack('>BBHBBBBB',
                0x21,   # Extension introducer
                0xf9,   # Graphics control label
                4,      # Block size
                0,      # Packed fields
                0,      # Delay time
                0,      # Transparent color index
                0       # Block terminator
        )
        out.write(packed)

    # Write the application extension block to the output stream
    def write_application_extension():
        packed = struct.pack('>BBHBBBBB',
                0x21,   # Extension introducer
                0xff,   # Application extension label
                11,     # Block size
                0,      # Application identifier
                0,      # Sub-block size
                0,      # Data sub-block
                0       # Block terminator
        )
        out.write(packed)

    # Write the comment extension block to the output stream
    def write_comment_extension():
        packed = struct.pack('>BBHB',
                0x21,   # Extension introducer
                0xfe,   # Comment extension label
                0,      # Block size
                0       # Block terminator
        )
        out.write(packed)

    # Write the plain text extension block to the output stream
    def write_plain_text_extension():
        packed = struct.pack('>BBHBBBBBBB',
                0x21,   # Extension introducer
                0x01,   # Plain text extension label
                12,     # Block size
                0,      # Text grid left position
                0,      # Text grid top position
                256,    # Text grid width
                256,    # Text grid height
                0,      # Cell width
                0,      # Cell height
                0       # Block terminator
        )
        out.write(packed)

    # Write the graphic control label to the output stream
    def write_graphic_control_label():
        packed = struct.pack('>BBHBBBBBBB',
                0x21,   # Extension introducer
                0xf9,   # Graphic control label
                4,      # Block size
                0,      # Packed fields
                0,      # Delay