#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (1 << rng.read(1)[0]) & 0xFF
    height = (1 << rng.read(1)[0]) & 0xFF
    flags  = (1 << rng.read(1)[0]) & 0xFF
    bg     =   rng.read(1)[0]
    aspect =   rng.read(1)[0]
    out.write(struct.pack('>HHBBB', width, height, flags, bg, aspect))

    # Global color table
    global_color_table_size = ((flags >> 7) & 1) + \
                              ((flags >> 5) & 2) * 3
    for _ in range(global_color_table_size):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left      =   0
    top       =   0
    image_width  = (1 << rng.read(1)[0]) & 0xFF
    image_height = (1 << rng.read(1)[0]) & 0xFF
    flags      = (1 << rng.read(1)[0]) & 0xFF
    local_color_table_flag = (flags >> 7) & 1
    interlace_flag         = (flags >> 6) & 1
    sort_flag              = (flags >> 5) & 1
    reserved               = (flags >> 4) & 1
    lct_size               = (1 << flags & 0xF)
    out.write(struct.pack('>HHHHBB',
                          left, top, image_width, image_height,
                          flags, lct_size))

    if local_color_table_flag == 1:
        for _ in range(lct_size):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            out.write(struct.pack('BBB', r, g, b))

    # LZW minimum code size
    min_code_size = rng.read(1)[0]
    out.write(min_code_size.to_bytes(1, 'little'))

    # Data sub-block
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x00:
            break

        elif block_type == 0x01:
            raise Exception("Not implemented")

        else:
            raise Exception("Invalid block type")

    # Trailer
    out.write(b'\x3B')