#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = rng.randint(1024, 65536)
    height = rng.randint(1024, 65536)
    flags  = (
            0x80 |    # Global color table flag
            7        # Color resolution
        ) << 4
    bgcol  = rng.randint(0, 255)
    aspect = 0
    out.write(struct.pack('<HHBBB', width, height, flags, bgcol, aspect))

    # Global color table
    for _ in range(7):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left   = 0
    top    = 0
    width  = rng.randint(width // 2, width)
    height = rng.randint(height // 2, height)
    flags  = (
            0x80 |     # Local color table flag
            0         # Interlace flag
        ) << 4
    lctsize = 7
    out.write(struct.pack('<HHHHB', left, top, width, height, flags))

    # Local color table
    for i in range(lctsize):
        if i == 0:
            r = g = b = 0
        else:
            r = rng.randint(r - 32, r + 32) & 255
            g = rng.randint(g - 32, g + 32) & 255
            b = rng.randint(b - 32, b + 32) & 255

        out.write(struct.pack('BBB', r, g, b))

    # LZW minimum code size
    mincodesize = rng.randint(2, 12)
    out.write(mincodesize.to_bytes(1, 'little'))

    # Data sub-block
    while True:
        blocktype = rng.read(1)[0]
        if blocktype == 0x00:
            break
        elif blocktype == 0x01:
            # Plain text extension
            extlen = rng.read(1)[0]
            label = rng.read(extlen)
            bkgdindex = rng.read(1)[0]
            foreindex = rng.read(1)[0]

            out.write(
                    bytes([
                        0x21, 0x01,
                        extlen,
                        *label,
                        0x00,
                        0x01,
                        bkgdindex,
                        foreindex]))
        elif blocktype == 0xfe:
            # Application extension
            appid = rng.read(8)
            appdata = rng.read(11)
            out.write(
                    bytes([
                        0x21, 0xff,
                        0x0b,
                        *appid,
                        *appdata]))
        elif blocktype == 0xf9:
            # Graphic control extension
            delaytime = rng.read(2)
            transindex = rng.read(1)[0]
            flags = rng.read(1)[0]
            out.write(
                    bytes([
                        0x21, 0xf9,
                        0x04,
                        *delaytime,
                        transindex,
                        flags]))
        elif blocktype == 0x02 or blocktype == 0x03:
            # Image data
            lzwminsize = rng.read(1)[0]
            imgdata = rng.read(1024)
            out.write(
                    bytes([
                        blocktype,
                        lzwminsize]) +
                    imgdata)

