#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the magic string and version
    out.write(b'GIF89a')

    # Generate the screen descriptor block
    width = 1024
    height = 768
    color_table_size = 0
    global_color_table = False
    background_color = 0x00
    aspect_ratio = 0x00
    flags = (
            0x80 | # Global Color Table Flag
            0x00 | # Sort Flag
            0x00   # Size of Local Color Table
        ).to_bytes(1, 'little', signed=False)
    screen_descriptor = bytes([
        width % 256,
        width // 256,
        height % 256,
        height // 256,
        color_table_size,
        *flags,
        background_color,
        aspect_ratio
    ])
    out.write(screen_descriptor)

    # Generate the local color table if needed
    if not global_color_table:
        for i in range(0, pow(2, color_table_size)):
            r = rng.read(1)[0] & 0xFF
            g = rng.read(1)[0] & 0xFF
            b = rng.read(1)[0] & 0xFF
            out.write(struct.pack('BBB', r, g, b))

    # Generate the image descriptor block
    left = 0
    top = 0
    width = 1024
    height = 768
    packed = (
            0x80 | # Local Color Table Flag
            0x00 | # Interlace Flag
            0x00 | # Sort Flag
            0x00   # Reserved
        ).to_bytes(1, 'little', signed=False)
    image_descriptor = bytes([
        left % 256,
        left // 256,
        top % 256,
        top // 256,
        width % 256,
        width // 256,
        height % 256,
        height // 256,
        *packed
    ])
    out.write(image_descriptor)

    # Generate the LZW minimum code size
    lzw_min_code_size = 2
    out.write(lzw_min_code_size.to_bytes(1, 'little'))

    # Generate the compressed image data
    def get_next_byte() -> int:
        return rng.read(1)[0] & 0xFF

    def write_byte(value: int):
        out.write(value.to_bytes(1, 'little'))

    def write_short(value: int):
        out.write((value % 256).to_bytes(1, 'little'))
        out.write((value // 256).to_bytes(1, 'little'))

    def write_long(value: int):
        out.write((value % 256).to_bytes(1, 'little'))
        out.write(((value >> 8) % 256).to_bytes(1, 'little'))
        out.write(((value >> 16) % 256).to_bytes(1, 'little'))
        out.write(((value >> 24) % 256).to_bytes(1, 'little'))

    def write_chunk(data: list[int]):
        length = len(data)
        write_byte(length)
        for value in data:
            write_byte(value)

    clear_code = 1 << lzw_min_code_size
    end_of_information_code = clear_code + 1
    next_code = end_of_information_code + 1
    codes = [None] * 4096
    output = []
    current_code_size = lzw_min_code_size + 1
    max_code_size = 12
    bitmask = (1 << current_code_size) - 1
    buffer = get_next_byte()
    while True:
        code = buffer & bitmask

        if code == end_of_information_code:
            break

        if code == clear_code:
            current_code_size = lzw_min_code_size + 1
            bitmask = (1 << current_code_size) - 