#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (1 << rng.read(1)[0]) % 256
    height = (1 << rng.read(1)[0]) % 256
    flags  = rng.read(1)[0] & 7
    bg     = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(struct.pack('>HHBBB', width, height, flags, bg, aspect))

    # Global color table
    global_colors = [None for _ in range(2**flags)]
    if flags == 7:
        for i in range(len(global_colors)):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            global_colors[i] = bytes([r,g,b])
        out.write(b''.join(global_colors))

    # Image descriptor
    left   = rng.read(2)
    top    = rng.read(2)
    width  = (1 << rng.read(1)[0]) % 256
    height = (1 << rng.read(1)[0]) % 256
    flags  = rng.read(1)[0] & 7
    local_colors = []
    if flags == 7:
        for i in range((1 << flags) - 1):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            local_colors.append(bytes([r,g,b]))
        out.write(left+top+width.to_bytes(1,'little')+height.to_bytes(1,'little'))
        out.write(flags.to_bytes(1,'little'))
        out.write(b''.join(local_colors))

    # LZW minimum code size
    mincodesize = rng.read(1)[0]
    out.write(mincodesize.to_bytes(1,'little'))

    # Data sub-block
    lzwdata = b''
    while True:
        blocktype = rng.read(1)[0]
        if blocktype == 0x00: break

        elif blocktype == 0x01:
            lzwdata += rng.read(1)[0].to_bytes(1,'little')

        else:
            length = rng.read(1)[0]
            lzwdata += rng.read(length).to_bytes(length,'little')

    out.write(lzwdata)