#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (e.g., /dev/urandom).
# The width and height of the generated image are given by
# `width` and `height`.  The color depth is determined from
# the size of `palette`, which must be either 256 or 1024 bytes.
def generate_random_gif(
        rng: BinaryIO, palette: bytes, width: int, height: int, out: BinaryIO
    ):

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    out.write(struct.pack('>HHBBB', width, height, 0xf7, 0x00, 0))

    # Global color table
    if len(palette) == 256 * 3:
        out.write(palette)
    elif len(palette) == 1024 * 3:
        out.write(b'\x00\x00\x00')
        out.write(palette[::3])
        out.write(palette[1::3])
        out.write(palette[2::3])
    else:
        raise ValueError("Palette length must be one of {256*3, 1024*3}")

    # Image descriptor
    out.write(struct.pack('<BHHBBBBBHB', 0x2c, 0, 0, 0, 0, 0, 0, 0, 0, 0))

    # Local color table
    out.write(b'\x00\x00\x00')

    # LZW minimum code size
    out.write(b'\x08')

    # Data sub-block
    for i in range(0, width * height):
        out.write(b'\x00')

        # Write up to 255 bytes at a time
        while True:
            n = min(int.from_bytes(rng.read(1), 'little'), 255 - 1)

            if n < 255:
                out.write(n.to_bytes(1, 'little'))
                break

            out.write((255).to_bytes(1, 'little'))
            out.write(rng.read(n))

