#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom or /dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The first 6 bytes are always "GIF87a" or "GIF89a".
    header = b'GIF89a'
    out.write(header)

    # Next comes the logical screen descriptor. It's size is 7 bytes and it
    # contains two fields: width and height of the image in pixels.
    width = int.from_bytes(
            rng.read(2), byteorder='little', signed=False) % 1000
    height = int.from_bytes(
            rng.read(2), byteorder='little', signed=False) % 1000
    out.write(struct.pack('<HHB', width, height, 0))

    # Then there's an optional global color table with up to 256 colors. We'll
    # just skip this part for now.
    numcolors = int.from_bytes(
            rng.read(1), byteorder='little', signed=False) & 0x07
    if numcolors!= 0:
        raise NotImplementedError('global color tables not supported yet')

    # Finally we have the actual image data. Each frame consists of a series of
    # blocks. A block starts with a single byte that tells us what kind of
    # block it is. There are four kinds of blocks:
    #   0x2C - Image Descriptor
    #   0x21 - Extension Block
    #   0x22 - Local Color Table
    #   0x3B - Trailer
    while True:

        # An image descriptor starts with a left position, top position,
        # width and height of the image in pixels.
        left = int.from_bytes(
                rng.read(2), byteorder='little', signed=False) % 1000
        top = int.from_bytes(
                rng.read(2), byteorder='little', signed=False) % 1000
        width = int.from_bytes(
                rng.read(2), byteorder='little', signed=False) % 1000
        height = int.from_bytes(
                rng.read(2), byteorder='little', signed=False) % 1000
        flags = int.from_bytes(
                rng.read(1), byteorder='little', signed=False)

        # If bit 7 is set then there's a local color table following the
        # image descriptor. Skip it for now.
        has_local_color_table = bool((flags >> 7) & 1)
        if has_local_color_table:
            numcolors = ((flags << 4) |
                         int.from_bytes(
                             rng.read(1), byteorder='little', signed=False)) & 0x0F
            if numcolors!= 0:
                raise NotImplementedError('local color tables not supported yet')

        # Bit 6 determines whether the image uses interlacing. We don't support
        # interlacing so ignore this flag too.
        use_interlacing = bool((flags >> 6) & 1)

        # Bit 5 determines whether the image is sorted. We don't support sorting
        # either so ignore this flag too.
        is_sorted = bool((flags >> 5) & 1)

        # Bit 4 determines whether the image has a local palette. We don't care
        # about palettes so ignore this flag too.
        has_local_palette = bool((flags >> 4) & 1)

        # Bit 3 determines whether the image has transparency. We don't care
        # about transparency so ignore this flag too.
        has_transparency = bool((flags >> 3) & 1)

        # Bit 2 determines whether the image has user input. We don't care about
        # user input so ignore this flag too.
        has_user_input = bool((flags >> 2) & 1)

        # Bit 1 determines whether the image has a delay time. We don't care
        # about delays so ignore this flag too.
        has_delay_time = bool((flags >> 1) & 1)

        # Bit 0 determines whether the image has transparent color index. We
        # don't care about this flag either so ignore it too.
        has_transparent_color_index = bool(flags & 1)

        # Write the image descriptor back