#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The first 6 bytes are always "GIF87a" or "GIF89a".
    # We'll use the latter for compatibility with more
    # browsers and tools.
    header = b'GIF89a'

    # Next comes the logical screen descriptor. This is where we specify
    # the width of our image (in pixels). Since this is a random number,
    # it will be between 1024 and 5120 pixels wide.
    w = rng.read(2)
    width = int.from_bytes(w, 'big') % 4096 + 1024

    # And the height of our image in pixels. Again, since this is a random
    # number, it will be between 1024 and 5120 pixels high.
    h = rng.read(2)
    height = int.from_bytes(h, 'big') % 4096 + 1024

    # Then there's a field that specifies the color table size. For now,
    # we'll just set this to zero, which means no color table.
    ctsize = b'\x00\x00'

    # Finally, we have two fields specifying the background color index
    # and pixel aspect ratio. Both of these values can be left at their
    # default value of 0.
    bgindex = b'\x00'
    aspect = b'\x00'

    # Now we're ready to write all of those values to the output stream.
    out.write(header + w + h + ctsize + bgindex + aspect)

    # Now let's get to the fun part! Let's generate some random colors!
    # First off, we need an array of RGB triples. Each triple consists of three
    # bytes representing the red, green, and blue components respectively.
    # These values must range from 0-255 inclusive.
    rgb = [struct.pack('BBB', *rng.read(3)) for _ in range(256)]

    # Now we need to create the actual color table by concatenating each
    # of the individual RGB triples together.
    colortable = b''.join(rgb)

    # Now we can write the length of the entire color table as well as
    # the contents itself to the output stream.
    out.write(len(colortable).to_bytes(2, 'big'))
    out.write(colortable)

    # Now let's move on to generating the image data. To do so, we need
    # to know how many frames our animated GIF should contain. It turns
    # out that this value has to be between 2 and 256 inclusive. So we'll
    # pick one randomly and then stick with it.
    numframes = rng.read(1)[0] % 256 + 2

    # Now we can start writing the image data. First up, we need to write
    # the global color table flag. If this value is nonzero, then the
    # following 3 bytes represent the index of the color palette to use.
    # Otherwise, they represent the transparent color index. In either case,
    # both of these values must be less than the total number of entries
    # in the color table.
    gctflag = rng.read(1)[0] % 2
    if gctflag == 0:
        gct = rng.read(1)[0] % len(rgb)
    else:
        gct = rng.read(1)[0] % len(rgb)
    out.write(gctflag.to_bytes(1, 'big') + gct.to_bytes(1, 'big'))

    # Next up, we need to write the local color table flag. If this value
    # is nonzero, then the next 3 bytes represent the index of the color
    # palette to use. Otherwise, they represent the transparent color index.
    # In either case, both of these values must be less than the total number
    # of entries in the color table.
    lctflag = rng.read(1)[0] % 2
    if lctflag == 0:
        lct = rng.read(1)[0] % len(rgb)
    else:
        lct = rng.read(1)[0] % len(rgb)
    out.write(lctflag.to_