#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (1 << rng.read(1)[0]) | ((1 << rng.read(1)[0]) >> 7)
    height = (1 << rng.read(1)[0]) | ((1 << rng.read(1)[0]) >> 7)
    flags  = rng.read(1)[0] & 0x4C
    bg     = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(struct.pack('>HHBBB', width, height, flags, bg, aspect))

    # Global color table
    for i in range(256):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left   = 0
    top    = 0
    width  = (1 << rng.read(1)[0]) | ((1 << rng.read(1)[0]) >> 7)
    height = (1 << rng.read(1)[0]) | ((1 << rng.read(1)[0]) >> 7)
    flags  = rng.read(1)[0] & 0xC0
    local  = rng.read(1)[0]
    interlaced = bool((flags & 0x40)!= 0)
    lctsize = 2 ** (local + 1) if local else 0
    out.write(struct.pack('>HHBBBBB', left, top, width, height, flags, local))

    # Local color table
    for i in range(lctsize):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(struct.pack('BBB', r, g, b))

    # Block size
    blocklen = rng.read(1)[0]
    while True:
        if blocklen == 0: break

        # Extension
        exttype = rng.read(1)[0]
        out.write(struct.pack('B', exttype))

        # Sub-block length
        subblocklen = rng.read(1)[0]
        out.write(struct.pack('B', subblocklen))

        # Data
        data = rng.read(subblocklen)
        out.write(data)

        # Block terminator
        blockterm = rng.read(1)[0]
        assert blockterm == 0

        blocklen = rng.read(1)[0]

    # Image data
    blocklen = rng.read(1)[0]
    while True:
        if blocklen == 0: break

        # Plain text extension
        exttype = rng.read(1)[0]
        out.write(struct.pack('B', exttype))

        # Sub-block length
        subblocklen = rng.read(1)[0]
        out.write(struct.pack('B', subblocklen))

        # Data
        data = rng.read(subblocklen)
        out.write(data)

        # Block terminator
        blockterm = rng.read(1)[0]
        assert blockterm == 0

        blocklen = rng.read(1)[0]