#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 << (int.from_bytes(rng.read(1), 'big') & 0x07)
    height = 1 << ((int.from_bytes(rng.read(1), 'big') >> 3) & 0x07)
    flags = int.from_bytes(rng.read(1), 'big')
    bgcolor = int.from_bytes(rng.read(1), 'big')
    aspectratio = int.from_bytes(rng.read(1), 'big')
    out.write(struct.pack('>HHBBB', width, height, flags, bgcolor, aspectratio))

    # Global color table
    ncolors = 2 ** ((flags >> 4) & 0x07)
    for i in range(ncolors):
        r = int.from_bytes(rng.read(1), 'big')
        g = int.from_bytes(rng.read(1), 'big')
        b = int.from_bytes(rng.read(1), 'big')
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left = int.from_bytes(rng.read(2), 'big')
    top = int.from_bytes(rng.read(2), 'big')
    width = 1 << (int.from_bytes(rng.read(1), 'big') & 0x07)
    height = 1 << ((int.from_bytes(rng.read(1), 'big') >> 3) & 0x07)
    flags = int.from_bytes(rng.read(1), 'big')
    lctsize = 2 ** ((flags >> 4) & 0x07)
    interlace = bool((flags >> 6) & 0x01)
    sortflag = bool((flags >> 7) & 0x01)
    out.write(struct.pack('>HHHHBB', left, top, width, height, flags, lctsize))

    # Local color table
    if lctsize!= 0:
        for i in range(lctsize):
            r = int.from_bytes(rng.read(1), 'big')
            g = int.from_bytes(rng.read(1), 'big')
            b = int.from_bytes(rng.read(1), 'big')
            out.write(struct.pack('BBB', r, g, b))

    # Table based image data
    while True:
        blocktype = int.from_bytes(rng.read(1), 'big')

        if blocktype == 0x2C:
            break

        elif blocktype == 0x21:
            pass

        else:
            raise Exception("Unknown block type")


    # Plain text extension
    while True:
        blocktype = int.from_bytes(rng.read(1), 'big')

        if blocktype == 0x01:
            pass

        elif blocktype == 0x00:
            break

        else:
            raise Exception("Unknown block type")

    # Application extension
    while True:
        blocktype = int.from_bytes(rng.read(1), 'big')

        if blocktype == 0xFF:
            pass

        elif blocktype == 0x00:
            break

        else:
            raise Exception("Unknown block type")

    # Comment extension
    while True:
        blocktype = int.from_bytes(rng.read(1), 'big')

        if blocktype == 0xFE:
            pass

        elif blocktype == 0x00:
            break

        else:
            raise Exception("Unknown block type")

    # Graphics control extension
    while True:
        blocktype = int.from_bytes(rng.read(1), 'big')

        if blocktype == 0xF9:
            pass

        elif blocktype == 0x00:
            break

        else:
            raise Exception("Unknown block type")

    # Plain text data
    while True:
        blocktype = int.from_bytes(rng.read(1), 'big')

        if blocktype == 0x01:
            pass

        elif blocktype == 0x00:
            break

        else:
            raise Exception("Unknown block type")

    # Image