#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom or /dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Generate the header and write it to `out`.
    header = b'GIF89a'
    out.write(header)

    # Generate the logical screen descriptor and write it to `out`.
    width  = int.from_bytes(
            rng.read(2), byteorder='little', signed=False) % 65536
    height = int.from_bytes(
            rng.read(2), byteorder='little', signed=False) % 65536
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lsd    = struct.pack('<HHBBB', width, height, *flags, *bg, *aspect)
    out.write(lsd)

    # Generate the global color table and write it to `out`.
    gctlen = int.from_bytes(rng.read(1), byteorder='little') & 0x7f
    if gctlen == 0:
        gctlen = 2 ** ((gctlen + 1) as int)
    gct = [rng.read(4) for _ in range(gctlen)]
    out.write(struct.pack('B', len(gct)))
    for entry in gct:
        out.write(entry)

    # Generate the image descriptor and write it to `out`.
    left   = int.from_bytes(rng.read(2), byteorder='little') % width
    top    = int.from_bytes(rng.read(2), byteorder='little') % height
    width  = int.from_bytes(rng.read(2), byteorder='little') % width
    height = int.from_bytes(rng.read(2), byteorder='little') % height
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    idesc  = struct.pack('<HHHHBB', left, top, width, height, *flags)
    out.write(idesc)

    # Generate the local color table and write it to `out`.
    lctlen = int.from_bytes(rng.read(1), byteorder='little') & 0x7f
    if lctlen == 0:
        lctlen = 2 ** ((lctlen + 1) as int)
    lct = [rng.read(4) for _ in range(lctlen)]
    out.write(struct.pack('B', len(lct)))
    for entry in lct:
        out.write(entry)

    # Generate the image data and write it to `out`.
    data = []
    while True:
        blocktype = rng.read(1)[0]

        if blocktype == 0x2c:
            break

        elif blocktype == 0x21:
            pass

        else:
            raise ValueError("Invalid block type")

        length = int.from_bytes(
                rng.read(1), byteorder='little', signed=False)
        data += list(rng.read(length))

    out.write(b'\x00'.join(data))

