#!/usr/bin/env python3

import io
import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom or /dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Generate the header and write it to the output stream.
    header = b'GIF89a'
    out.write(header)

    # Read the logical screen descriptor from the input stream.
    lsd = read_logical_screen_descriptor(rng)
    print('Logical Screen Descriptor:', lsd)
    out.write(lsd)

    # Read the global color table from the input stream.
    gct = read_global_color_table(rng, lsd['gct_size'])
    if len(gct):
        print('Global Color Table:')
        for i in range(0, 256, 16):
            print(' '.join(['{:04x}'.format(c) for c in gct[i : min(len(gct), i+16)]]))
        out.write(b''.join([struct.pack('<H', c) for c in gct]))

    # Read each image block from the input stream.
    while True:

        # Read an image block from the input stream.
        imgblock = read_image_block(rng)
        if not imgblock: break

        # Write the image block to the output stream.
        print('Image Block:', imgblock)
        out.write(imgblock)


# Reads a logical screen descriptor from the input stream.
def read_logical_screen_descriptor(
        rng: BinaryIO
    ) -> dict:

    return {
        'width':   read_uint16(rng),
        'height':  read_uint16(rng),
        'flags':   read_byte(rng),
        'bgcolor': read_byte(rng),
        'aspect':  read_byte(rng),
        'gct_size': read_byte(rng)}


# Reads a global color table from the input stream.
def read_global_color_table(
        rng: BinaryIO, size: int
    ) -> list:

    return [read_uint16(rng) for _ in range(size)]


# Reads an image block from the input stream.
def read_image_block(
        rng: BinaryIO
    ) -> bytes:

    # Read the image separator from the input stream.
    sep = read_byte(rng)
    if sep!= 0x2C: raise Exception()

    # Read the image descriptor from the input stream.
    idd = read_image_descriptor(rng)
    print('\tImage Descriptor:', idd)

    # Read the local color table from the input stream.
    lct = read_local_color_table(rng, idd['lct_size'])
    if len(lct):
        print('\tLocal Color Table:')
        for i in range(0, 256, 16):
            print('\t\t'+' '.join(['{:04x}'.format(c) for c in lct[i : min(len(lct), i+16)]]))

    # Read the image data from the input stream.
    data = read_image_data(rng, idd['lzwmin'], idd['lzwmax'], idd['lzwsize'])
    if len(data):
        print('\tImage Data:')
        for i in range(0, len(data), 16):
            print('\t\t'+' '.join(['{:02x}'.format(c) for c in data[i : min(len(data), i+16)]]))

    # Return the image block as a binary string.
    return sep + idd + b''.join([struct.pack('<H', c) for c in lct]) + data


# Reads an image descriptor from the input stream.
def read_image_descriptor(
        rng: BinaryIO
    ) -> dict:

    return {
        'left':     read_uint16(rng),
        'top':      read_uint16(rng),
        'width':    read_uint16(rng),
        'height':   read_uint16(rng),
        'flags':    read_byte(rng),
        'lct_size': read_byte(rng)}


# Reads a local color table from the input stream.
def read_local_color_table(
        rng: BinaryIO, size: int
    ) -> list:

    return [read_