#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (e.g., /dev/urandom).
# The generated GIF has 10 frames and each frame is
# 512 bytes long.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate the logical screen descriptor
    width = rng.randint(64, 1024)
    height = rng.randint(64, 1024)
    flags = 0x87
    bgcolorindex = 0
    pixelaspectratio = 0
    out.write(struct.pack('<HHBBB',
            width, height, flags, bgcolorindex, pixelaspectratio))

    # Generate global color table
    numcolors = min((width * height)//256, 256)
    for i in range(numcolors):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write(struct.pack('BBB', r, g, b))

    # Generate local color table
    numcolors = min((width * height)//256, 256)
    for i in range(numcolors):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write(struct.pack('BBB', r, g, b))

    # Generate image descriptors
    for _ in range(10):

        # Image separator
        out.write(b'\x2C')

        # Image left position
        xpos = rng.randint(0, width-1)
        out.write(struct.pack('<H', xpos))

        # Image top position
        ypos = rng.randint(0, height-1)
        out.write(struct.pack('<H', ypos))

        # Image width
        w = rng.randint(1, width - xpos)
        out.write(struct.pack('<H', w))

        # Image height
        h = rng.randint(1, height - ypos)
        out.write(struct.pack('<H', h))

        # Flags
        flags = 0x00
        out.write(struct.pack('B', flags))

        # LZW minimum code size
        lzwmincodesize = 2
        out.write(struct.pack('B', lzwmincodesize))

        # Image data
        datalen = w*h//2
        data = rng.read(datalen)
        out.write(data)

    # Terminator
    out.write(b';')

