#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical Screen Descriptor
    width  = (1 << rng.read(1)[0]) % 256
    height = (1 << rng.read(1)[0]) % 256
    flags  = rng.read(1)[0] & 7
    bg     = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(struct.pack('>HHBBB', width, height, flags, bg, aspect))

    # Global Color Table
    global_color_table_size = rng.read(1)[0] & 7
    if global_color_table_size!= 0:
        raise NotImplementedError()

    # Image Descriptor
    left   = rng.read(2)
    top    = rng.read(2)
    width  = (1 << rng.read(1)[0]) % 256
    height = (1 << rng.read(1)[0]) % 256
    flags  = rng.read(1)[0] & 7
    local_color_table_flag = bool((flags >> 7) & 1)
    interlace               = bool((flags >> 6) & 1)
    sort                    = bool((flags >> 5) & 1)
    reserved                = ((flags >> 3) & 3) == 3
    lct                     = None
    if local_color_table_flag:
        lct_size = (global_color_table_size + 1) * 3
        lct      = rng.read(lct_size)
    else:
        lct_size = 0
    lzw_mincode_size        = rng.read(1)[0] & 0x0F
    out.write(left + top + bytes([width, height]))
    out.write(bytes([flags | (local_color_table_flag << 7)]))
    if local_color_table_flag:
        out.write(bytes([lct_size]))
        out.write(lct)
    out.write(bytes([lzw_mincode_size]))

    # Local Color Table
    if not local_color_table_flag and lct is not None:
        out.write(lct)

    # Data
    while True:
        blocktype = rng.read(1)[0]

        if blocktype == 0x2C:
            # Image Descriptor
            left   = rng.read(2)
            top    = rng.read(2)
            width  = (1 << rng.read(1)[0]) % 256
            height = (1 << rng.read(1)[0]) % 256
            flags  = rng.read(1)[0] & 7
            local_color_table_flag = bool((flags >> 7) & 1)
            interlace               = bool((flags >> 6) & 1)
            sort                    = bool((flags >> 5) & 1)
            reserved                = ((flags >> 3) & 3) == 3
            lct                     = None
            if local_color_table_flag:
                lct_size = (global_color_table_size + 1) * 3
                lct      = rng.read(lct_size)
            else:
                lct_size = 0
            lzw_mincode_size        = rng.read(1)[0] & 0x0F

            out.write(blocktype)
            out.write(left + top + bytes([width, height]))
            out.write(bytes([flags | (local_color_table_flag << 7)]))
            if local_color_table_flag:
                out.write(bytes([lct_size]))
                out.write(lct)
            out.write(bytes([lzw_mincode_size]))

            # Local Color Table
            if not local_color_table_flag and lct is not None:
                out.write(lct)

            continue

        elif blocktype in [0x21, 0x22]:
            # Extension Block
            ext_label = rng.read(1)[0]
            out.write(blocktype)
            out.write(ext_label)

            if ext_label == 0xF9:
                # Graphics Control Extension
                out