#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom or /dev/urandom).
# If `use_urandom`, then use /dev/urandom instead of os.urandom.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, use_urandom=False
    ) -> None:

    # Read 10 bytes from the RNG to get the size of the image in bytes.
    if not use_urandom and hasattr(os, 'urandom'):
        # Use urandom() on Unix-like systems
        imgsize = int.from_bytes(
                os.urandom(4), byteorder='big', signed=False)
    else:
        # Otherwise, fall back to reading directly from /dev/urandom
        imgsize = int.from_bytes(
                rng.read(4), byteorder='big', signed=False)

    # Write the GIF header
    out.write(b'GIF89a')

    # Write the logical screen descriptor
    out.write(struct.pack('>HHBBB',
            256, 256, 7, 0, 0))

    # Write the global color table
    for i in range(0, 256):
        out.write(struct.pack('BBB', *rng.read(3)))

    # Write the image descriptor
    out.write(struct.pack('>HHHHB',
            0, 0, imgsize, 0, 0))

    # Write the image data
    while True:
        chunklen = min(imgsize, 2**12 - 1)

        # Write the length of this chunk
        out.write(chunklen.to_bytes(1, byteorder='big'))

        # Write the actual data
        out.write(rng.read(chunklen))

        imgsize -= chunklen

        if imgsize <= 0: break