#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The width and height of the image are given by `width` and `height`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int
    ):

    # Header (always 6 bytes long)
    out.write(b'GIF89a')

    # Logical screen descriptor (always 7 bytes long)
    # Width is stored as two bytes in big endian order
    out.write(struct.pack('>H', width))
    # Height is stored as two bytes in big endian order
    out.write(struct.pack('>H', height))
    # Flags for color resolution, sort flag, size of global color table
    # We use no color table so we set this to zero
    out.write(b'\x00\x00')
    # Background color index is always zero
    out.write(b'\x00')
    # Pixel aspect ratio - default value is 1:1
    out.write(b'\x00')

    # Image descriptor (always 10 bytes long)
    # Left position is always zero
    out.write(b'\x00\x00')
    # Top position is always zero
    out.write(b'\x00\x00')
    # Width is stored as two bytes in big endian order
    out.write(struct.pack('>H', width))
    # Height is stored as two bytes in big endian order
    out.write(struct.pack('>H', height))
    # Flags for local color table, interlace flag, sort flag, reserved bits
    # No color table used, not interlaced, not sorted, all reserved bits must be zero
    out.write(b'\x00\x00')

    # LZW minimum code size (always one byte long)
    # This is hardcoded to 2 because that's what most GIF viewers seem to do
    out.write(b'\x02')

    # Data sub-block
    # First byte indicates how many bytes follow
    # Then comes the actual data
    # Since we don't have any compression scheme other than LZW,
    # we just write the raw data here
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(data)