#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Magic bytes
    out.write(b'GIF89a')
    
    # Screen descriptor
    width = 1 + (int.from_bytes(rng.read(2), 'little') % 640)
    height = 1 + (int.from_bytes(rng.read(2), 'little') % 480)
    flags = int.from_bytes(rng.read(1), 'little') & 7
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('<HBBB', width, height, flags, background))
    out.write(aspect_ratio)
    
    # Global color table
    if flags & 1 == 1:
        colors = [rng.read(3) for _ in range(2 ** ((flags >> 4) + 1))]
        out.write(b''.join(colors))
        
    # Image data
    while True:
        block_type = rng.read(1)[0]
        if block_type == 0x3b:
            break
            
        length = int.from_bytes(rng.read(1), 'little')
        if block_type == 0x2c:
            left = int.from_bytes(rng.read(2), 'little')
            top = int.from_bytes(rng.read(2), 'little')
            width = int.from_bytes(rng.read(2), 'little')
            height = int.from_bytes(rng.read(2), 'little')
            flags = int.from_bytes(rng.read(1), 'little')
            local_color_table = []
            if flags & 1 == 1:
                local_color_table = [rng.read(3) for _ in range(2 ** ((flags >> 4) + 1))]
                
            lzw_minimum_code_size = rng.read(1)[0]
            lzw_data = b''
            while len(lzw_data) < length - 5:
                byte = rng.read(1)[0]
                if byte <= 0xf7:
                    lzw_data += byte.to_bytes(1, 'big')
                elif byte == 0xf8:
                    pass
                else:
                    clear_code = 1 << (byte & 0x07)
                    end_of_information = clear_code + 1
                    available_codes = clear_code + 2
                    old_code = None
                    code_stack = []
                    while True:
                        try:
                            code = int.from_bytes(rng.read(1), 'big')
                        except EOFError as e:
                            raise Exception("Unexpected EOF") from e
                        if code == end_of_information:
                            break
                        if code == clear_code:
                            code_stack = []
                            available_codes = clear_code + 2
                            old_code = None
                        elif code > available_codes or old_code is None:
                            code_stack.append(code)
                        else:
                            if code_stack:
                                code_stack.append(old_code)
                            while code_stack:
                                lzw_data += code_stack.pop().to_bytes(1, 'big')
                            code_stack.append(available_codes)
                            old_code = code
                        if available_codes >= 0xfff:
                            break
                        available_codes *= 2
                    lzw_data += b'\x00'
                    lzw_data += struct.pack('<HH', left, top)
                    lzw_data += struct.pack('<HH', width, height)
                    lzw_data += struct.pack('<B', flags)
                    lzw_data += struct.pack('<B', lzw_minimum_code_size)
                    lzw_data += b''.join(local_color_table)
                    lzw_data += b'\x