#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# returns its size in bytes.
def generate_random_gif(out: BinaryIO, rng):
    # Header
    out.write(b'GIF89a')

    width = 1 + rng.randint(0x7F)
    height = 1 + rng.randint(0x7F)
    flags = b'\x87\x00' if (width >= 64 or height >= 64) else b'\x07\x00'
    out.write(struct.pack('<HHB', width, height, ord(flags[0])))

    # Global Color Table
    for i in range(256):
        r = rng.randint(0xFF)
        g = rng.randint(0xFF)
        b = rng.randint(0xFF)
        out.write(bytes([r,g,b]))

    # Image Descriptor
    left = 0
    top = 0
    width -= 1
    height -= 1
    packed = ((left & 0xF8) << 8) | ((top & 0xFE) << 9) | \
             (((width - 1) & 0x07) << 12) | (((height - 1) & 0x07) << 15) | \
             0x8000
    out.write(struct.pack('<HHH', packed, 0, 0))

    # Local Color Table
    for i in range(256):
        r = rng.randint(0xFF)
        g = rng.randint(0xFF)
        b = rng.randint(0xFF)
        out.write(bytes([r,g,b]))

    # Image Data
    lzwmincodesize = 2
    clearcode = 1 << lzwmincodesize
    endofinfo = clearcode + 1
    nextcode = None
    codesize = lzwmincodesize + 1
    maxcodesize = 12
    bitbuffer = 0
    bitsinbitbuffer = 0
    bytestoread = 0
    bytecount = 0
    while True:
        if not nextcode:
            if bytecount == bytestoread:
                break

            byte = out.read(1)[0]
            bytecount += 1
            bitbuffer |= (byte & 0xFF) << bitsinbitbuffer
            bitsinbitbuffer += 8
            if bitsinbitbuffer < codesize: continue

        if nextcode == endofinfo: break

        if nextcode == clearcode:
            table = [i for i in range(clearcode)]
            table.extend((nextcode,))
            codesize = lzwmincodesize + 1
            maxcodesize = 12
            nextcode = None
            continue

        result = []
        if len(table) <= nextcode:
            result.append(table[-1])
            result.append(table[-1])
        else:
            result.extend(table[nextcode-len(table)])
            result.extend(table[nextcode-len(table)+1])

        if len(result) > 1 and result[-1] == result[-2]:
            result.pop()
            result.pop()
            table.extend(result)
            table.extend((nextcode,))
            codesize = min(maxcodesize, codesize+1)
            if codesize == maxcodesize and nextcode == (1<<codesize)-1:
                codesize += 1
            nextcode = None
            continue

        table.extend(result)
        nextcode = None

        if len(table) >= (1<<(codesize)):
            nextcode = len(table)
            if codesize < maxcodesize:
                codesize += 1

    return out.tell()

