#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (1 << rng.read(1)[0]) % 65536
    height = (1 << rng.read(1)[0]) % 65536
    flags  = rng.read(1)[0] & 7
    bgcol  = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(struct.pack('>HHBBB', width, height, flags, bgcol, aspect))

    # Global color table
    for i in range(2**flags):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left   = rng.read(2)
    top    = rng.read(2)
    width  = (1 << rng.read(1)[0]) % 65536
    height = (1 << rng.read(1)[0]) % 65536
    flags  = rng.read(1)[0] & 7
    lctflag = bool((flags >> 7) & 1)
    interlace = bool((flags >> 6) & 1)
    sortflag = bool((flags >> 5) & 1)
    reserved = rng.read(1)[0]
    lctsize = 2**(flags & 7) if lctflag else 0
    out.write(left + top + struct.pack('>HHBB', width, height, flags, lctsize))

    # Local color table
    for i in range(lctsize):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(struct.pack('BBB', r, g, b))

    # Table based image data
    while True:
        blocktype = rng.read(1)[0]

        if blocktype == 0x2C:
            # Image descriptor
            left   = rng.read(2)
            top    = rng.read(2)
            width  = (1 << rng.read(1)[0]) % 65536
            height = (1 << rng.read(1)[0]) % 65536
            flags  = rng.read(1)[0] & 7
            lctflag = bool((flags >> 7) & 1)
            interlace = bool((flags >> 6) & 1)
            sortflag = bool((flags >> 5) & 1)
            reserved = rng.read(1)[0]
            lctsize = 2**(flags & 7) if lctflag else 0
            out.write(
                left + top + struct.pack('>HHBB', width, height, flags, lctsize)
            )

            # Local color table
            for i in range(lctsize):
                r = rng.read(1)[0]
                g = rng.read(1)[0]
                b = rng.read(1)[0]
                out.write(struct.pack('BBB', r, g, b))

            # LZW minimum code size
            mincodesize = rng.read(1)[0]
            out.write(mincodesize.to_bytes(1, 'little'))

            # Data sub-blocks
            while True:
                databyte = rng.read(1)[0]

                if databyte!= 0x00:
                    out.write(databyte.to_bytes(1, 'little'))
                elif databyte == 0x00 and rng.read(1)[0] == 0x00:
                    break
                else:
                    raise Exception("Invalid GIF data")

        elif blocktype == 0x21:
            # Extension
            extlabel = rng.read(1)[0]
            out.write(extlabel.to_bytes(1, 'little'))

            # Application extension
            if extlabel == 0xFF:
                appid = rng.read(1)[0]
                appauthcode = rng.read(1)[0]
                out.write(appid.to_bytes(1, 'little'))
               