#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # The first 6 bytes are always "GIF87a" or "GIF89a".
    # We'll use "GIF89a", since it's more modern and has
    # better support for transparency.
    header = b'GIF89a'

    # Next comes the logical screen descriptor. It contains
    # some metadata about the image as well as information
    # on how to render it in a viewer.
    width = rng.randint(100, 250)
    height = rng.randint(100, 250)
    color_table_size = rng.choice([2**n for n in range(1, 4)])
    global_color_table = True if rng.randint(0, 1) else False
    background_index = rng.randint(0, color_table_size - 1)
    pixel_aspect_ratio = rng.randint(0, 2)
    flags = [
        0b1 << i for i in reversed(range(7))
    ]
    flags[0] |= bool(global_color_table) * 0b10000000
    flags[0] |= pixel_aspect_ratio * 0b01000000
    flags[3] |= background_index * 0b00000111
    screen_descriptor = struct.pack('>HHBBB',
            width, height, color_table_size,
            flags[0], flags[1], flags[2], flags[3])

    # If we have a global color table, then we need to write that
    # before any other chunks.
    if global_color_table:
        palette = []
        for _ in range(color_table_size):
            r = rng.randint(0, 255)
            g = rng.randint(0, 255)
            b = rng.randint(0, 255)
            palette += [r, g, b]

        global_color_table_chunk = struct.pack('>B',
                len(palette) // 3) + bytes(palette)
        out.write(global_color_table_chunk)

    # Now we can add all of our frames! Each frame is made up of
    # one or more graphics control blocks followed by an image
    # block.
    num_frames = rng.randint(1, 10)
    for _ in range(num_frames):
        # First, let's make sure this frame will be visible at least once.
        delay = rng.randint(30, 1000)
        transparent_color_flag = rng.randint(0, 1)
        disposal_method = rng.randint(0, 3)
        user_input_flag = rng.randint(0, 1)
        flags = [
            0b1 << i for i in reversed(range(8))
        ]
        flags[0] |= transparent_color_flag * 0b10000000
        flags[0] |= disposal_method * 0b01100000
        flags[0] |= user_input_flag * 0b00010000
        flags[3] |= delay % 256
        flags[4] |= delay // 256
        flags[4] |= disposal_method * 0b00001111
        graphics_control_block = struct.pack('>BBHB',
                0x21, 0xF9, 4, flags[0], flags[1], flags[2], flags[3], flags[4])

        # Now let's get the actual image data. This is just a bunch
        # of pixels with their colors encoded as indexes into the
        # global color table.
        num_pixels = width * height
        image_data = []
        for _ in range(num_pixels):
            index = rng.randint(0, color_table_size - 1)
            image_data.append(index)

        image_data_chunk = struct.pack('>B',
                len(image_data)) + bytes(image_data)
        out.write(graphics_control_block + image_data_chunk)

    # Finally, we need to close off the file with an end-of-file chunk.
    eof_