#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The size of the generated image is determined by
# the first 24 bits in `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Read the width and height from the first 24 bits
    w = int.from_bytes(rng.read(1), 'big') & 0x7F
    h = (int.from_bytes(rng.read(1), 'big') & 0x7F) << 8

    # Generate the palette
    palsize = ((w * h) // 64) + 1
    pal = [struct.pack('BBB', *(rng.read(3))) for _ in range(palsize)]

    # Write the header
    out.write(b'GIF89a')

    # Write the logical screen descriptor
    out.write(struct.pack('<HHBBBBB', w, h, 0xF0 | len(pal), 0, 0, 0))

    # Write the global color table
    out.writelines(pal)

    # Write the graphics control extension
    out.write(b'\x21\xF9\x04\x05\x00\x00\x00')

    # Write the application extension
    out.write(b'\x21\xFF\x0BNETSCAPE2.0')
    out.write(b'\x03\x01\x00\x00')
    out.write(b'\x00\x00')

    # Write the image descriptor
    out.write(struct.pack('<HHHHBB', 0, 0, w, h, 0, 0))

    # Write the local color table
    out.write(b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')

    # Write the image data
    lzwmincodesize = 2
    out.write(lzwmincodesize.to_bytes(1, 'little'))
    lzwmaxcodesize = 12
    out.write(lzwmaxcodesize.to_bytes(1, 'little'))

    # LZW encoding
    codesize = lzwmincodesize + 1
    clearcode = 1 << codesize
    endofinfo = clearcode + 1
    nextcode = clearcode + 2
    codes = []
    prev = None
    while True:
        if prev == None or nextcode >= (1 << codesize):
            if nextcode >= (1 << lzwmaxcodesize):
                raise Exception("Too many bytes")

            # Emit a CLEARCODE
            out.write(clearcode.to_bytes(1, 'little'))
            codesize += 1
            if codesize > lzwmaxcodesize:
                raise Exception("Code size too large")

            # Add new codes to the dictionary
            for i in range(nextcode - clearcode):
                codes.append((i+1, chr(i+1)))

        # Find the longest prefix that exists in the dictionary
        maxlen = min([len(c[1]) for c in codes] + [-1])
        match = ''
        for c in codes:
            if c[1][:maxlen] == prev[:maxlen]:
                match = c[1]

        # If there was no match then emit a literal instead
        if not match:
            out.write(prev[0].to_bytes(1, 'little'))
            codes.append((nextcode, prev))
            nextcode += 1
            continue

        # Emit a REPEATEDCODE
        out.write(((codes.index(match)+1) >> 8).to_bytes(1, 'little'))
        out.write(((codes.index(match)+1) % 256).to_bytes(1, 'little'))

        # Update the dictionary with the new sequence
        codes.append((nextcode, prev[:maxlen]))
        nextcode += 1

        # Advance the previous character
        prev = prev[maxlen:]

