#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    header = b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\xff\xff\xff\x21\xf9\x04\x01\x0a\x00\x01\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b'

    # Next comes the length of the data section.  It's stored as a 3-byte big-endian
    # integer, so we need to read three bytes from /dev/urandom and interpret those
    # as an unsigned 24-bit number.
    data_len = int.from_bytes(rng.read(3), byteorder='big') + 1

    # Now we know where the data section ends up, let's go ahead and write our header
    # with the correct size for the data section.
    header += (data_len).to_bytes(3, byteorder='big')

    # Finally, we can write all this stuff out to disk!
    out.write(header)

    # Now we're going to start writing pixels into the file.  Each pixel consists of
    # three bytes representing the red, green, and blue values respectively.  Since
    # each value is between [0, 255], we can use one byte per color channel.
    while True:

        # Read three bytes from /dev/urandom and interpret them as an RGB tuple.
        rgb = rng.read(3)

        if len(rgb)!= 3:
            break
        
        # Write these bytes to the output file.
        out.write(rgb)


if __name__ == '__main__':
    
    # Open /dev/urandom for reading randomness.
    with open('/dev/urandom', 'rb') as rng:
    
        # For each command line argument, create a corresponding gif file.
        for arg in sys.argv[1:]:

            # Create the output file.
            with open(arg, 'wb') as f:

                generate_random_gif(rng, f)