#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    height = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    color_table_size = 2 ** ((rng.read(1)[0] & 0b111) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    palette = [bytes([rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]]) for i in range(color_table_size)]
    out.write(struct.pack('<HHBBB', width, height, color_table_size, background_color_index, pixel_aspect_ratio))

    # Write the global color table
    for color in palette:
        out.write(color)

    # Write the application extension block
    app_extension_block_size = 11
    application_identifier = 'NETSCAPE2.0'
    loop_count = 0
    out.write(struct.pack('<H', app_extension_block_size))
    out.write(application_identifier.encode())
    out.write(struct.pack('<BB', 3, loop_count))

    # Write the image descriptor
    image_descriptor_size = 10
    left_position = rng.read(1)[0] & 0xFF | (rng.read(1)[0] & 0xFF) << 8
    top_position = rng.read(1)[0] & 0xFF | (rng.read(1)[0] & 0xFF) << 8
    width = rng.read(1)[0] & 0xFF | (rng.read(1)[0] & 0xFF) << 8
    height = rng.read(1)[0] & 0xFF | (rng.read(1)[0] & 0xFF) << 8
    local_color_table_flag = rng.read(1)[0] >> 7
    interlace_flag = rng.read(1)[0] >> 6
    sort_flag = rng.read(1)[0] >> 5
    reserved_field = rng.read(1)[0] >> 4
    palette = [bytes([rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]]) for i in range(color_table_size)]
    out.write(struct.pack('<BBHHBB', image_descriptor_size, local_color_table_flag, left_position, top_position, width, height))

    # Write the local color table
    for color in palette:
        out.write(color)

    # Write the LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0]
    out.write(lzw_minimum_code_size)

    # Write the compressed data
    while True:
        byte = rng.read(1)
        if not byte:
            break
        out.write(byte)
