#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0]))

    # Global color table
    for i in range(768):
        out.write(rng.read(3))

    while True:
        # Graphic control extension
        gce = bytes([0x21, 0xf9, 0x04])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        out.write(gce)

        # Image descriptor
        idesc = bytes([0x2c])
        idesc += generate_image_descriptor(rng, width, height)
        out.write(idesc)

        # Local color table
        if bool(idesc[9] & 0x80):
            for i in range((idesc[9] & 0x07) + 1):
                out.write(rng.read(3))

        # Pixel data
        for y in range(idesc[5]):
            for x in range(idesc[4]):
                out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

        # Trailer
        trailer = bytes([0x3b])
        out.write(trailer)

        # End of stream?
        if bool(rng.read(1)):
            break

# Generates a random image descriptor using the
# random number generator `rng` (/dev/urandom)
def generate_image_descriptor(
        rng: BinaryIO, width: int, height: int
    ) -> bytes:
    left   = int.from_bytes(rng.read(2), byteorder='little') % (width  // 2) + 1
    top    = int.from_bytes(rng.read(2), byteorder='little') % (height // 2) + 1
    width  = int.from_bytes(rng.read(2), byteorder='little') % (width - left) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (height - top) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    return struct.pack('<HHBBB', left, top, width, height, flags[0])
