#!/usr/bin/env python3

import struct
from typing import BinaryIO


def read_bytes(f: BinaryIO, count: int) -> bytes:
    return f.read(count)


def write_bytes(f: BinaryIO, data: bytes):
    f.write(data)


def read_int8(f: BinaryIO) -> int:
    return struct.unpack('b', read_bytes(f, 1))[0]


def write_int8(f: BinaryIO, value: int):
    assert -128 <= value < 128
    write_bytes(f, struct.pack('b', value))


def read_uint8(f: BinaryIO) -> int:
    return struct.unpack('B', read_bytes(f, 1))[0]


def write_uint8(f: BinaryIO, value: int):
    assert 0 <= value < 256
    write_bytes(f, struct.pack('B', value))


def read_int16(f: BinaryIO) -> int:
    return struct.unpack('h', read_bytes(f, 2))[0]


def write_int16(f: BinaryIO, value: int):
    assert -32768 <= value < 32768
    write_bytes(f, struct.pack('h', value))


def read_uint16(f: BinaryIO) -> int:
    return struct.unpack('H', read_bytes(f, 2))[0]


def write_uint16(f: BinaryIO, value: int):
    assert 0 <= value < 65536
    write_bytes(f, struct.pack('H', value))


def read_int32(f: BinaryIO) -> int:
    return struct.unpack('i', read_bytes(f, 4))[0]


def write_int32(f: BinaryIO, value: int):
    assert -2147483648 <= value < 2147483648
    write_bytes(f, struct.pack('i', value))


def read_uint32(f: BinaryIO) -> int:
    return struct.unpack('I', read_bytes(f, 4))[0]


def write_uint32(f: BinaryIO, value: int):
    assert 0 <= value < 4294967296
    write_bytes(f, struct.pack('I', value))


def read_float32(f: BinaryIO) -> float:
    return struct.unpack('f', read_bytes(f, 4))[0]


def write_float32(f: BinaryIO, value: float):
    write_bytes(f, struct.pack('f', value))


def read_string(f: BinaryIO) -> str:
    length = read_uint32(f)
    return read_bytes(f, length).decode()


def write_string(f: BinaryIO, value: str):
    write_uint32(f, len(value))
    write_bytes(f, value.encode())


def read_vector2(f: BinaryIO) -> tuple:
    x = read_float32(f)
    y = read_float32(f)
    return (x, y)


def write_vector2(f: BinaryIO, value: tuple):
    write_float32(f, value[0])
    write_float32(f, value[1])


def read_vector3(f: BinaryIO) -> tuple:
    x = read_float32(f)
    y = read_float32(f)
    z = read_float32(f)
    return (x, y, z)


def write_vector3(f: BinaryIO, value: tuple):
    write_float32(f, value[0])
    write_float32(f, value[1])
    write_float32(f, value[2])


def read_quaternion(f: BinaryIO) -> tuple:
    w = read_float32(f)
    x = read_float32(f)
