#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar 15 20:49:36 2021

@author: zhantao
"""
import os
from pathlib import Path
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import scipy.stats as stats
from sklearn.model_selection import train_test_split, cross_val_score
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import r2_score
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.metrics import confusion_matrix


def load_data(path):
    """Load data from csv file.
    
    Parameters
    ----------
    path : string
        The path of the csv file.
        
    Returns
    -------
    X : array
        Features.
    y : array
        Target values.
    """
    df = pd.read_csv(path)
    X = df[['age', 'sex', 'cp', 'trestbps', 'chol', 'fbs', 'restecg',
            'thalach', 'exang', 'oldpeak', 'slope', 'ca']].values
    y = df['target'].values
    return X, y


def split_train_test(X, y, test_size=0.2, random_state=None):
    """Split data into training set and testing set.
    
    Parameters
    ----------
    X : array
        Features.
    y : array
        Target values.
    test_size : float
        Proportion of the dataset to include in the test split.
    random_state : int or None
        Controls the shuffling applied to the data before applying the split.
        Pass an integer for reproducible output across multiple function calls.
        Default is None.
        
    Returns
    -------
    X_train : array
        Training features.
    X_test : array
        Testing features.
    y_train : array
        Training target values.
    y_test : array
        Testing target values.
    """
    X_train, X_test, y_train, y_test = \
        train_test_split(X, y, test_size=test_size, random_state=random_state)
    return X_train, X_test, y_train, y_test


def scale_features(X_train, X_test):
    """Scale features.
    
    Parameters
    ----------
    X_train : array
        Training features.
    X_test : array
        Testing features.
        
    Returns
    -------
    X_train_scaled : array
        Scaled training features.
    X_test_scaled : array
        Scaled testing features.
    scaler : object
        The fitted scaler.
    """
    # fit scaler
    scaler = StandardScaler()
    X_train_scaled = scaler.fit_transform(X_train)
    X_test_scaled = scaler.transform(X_test)
    return X_train_scaled, X_test_scaled, scaler


def plot_feature_distribution(X, y, feature_name, class_names):
    """Plot the distribution of a feature by classes.
    
    Parameters
    ----------
    X : array
        Features.
    y : array
        Target values.
    feature_name : string
        Name of the feature.
    class_names : list of strings
        Names of the classes.
    """
    fig, ax = plt.subplots(figsize=(10, 7))
    for i in range(len(class_names)):
        sns.distplot(X[:, i][y ==