#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a' # GIF magic string
    width = 160 # Width of the image
    height = 120 # Height of the image
    color_table_size = 7 # Size of the color table (2**n colors)
    background_color_index = 0 # Index of the background color in the color table
    pixel_aspect_ratio = 0 # Pixel aspect ratio (unused)

    # Write the GIF header
    out.write(header)

    # Write the Logical Screen Descriptor
    out.write(struct.pack('<HHBB', width, height, color_table_size, background_color_index))

    # Write the Global Color Table
    for i in range(2 ** color_table_size):
        out.write(bytes([i]))

    # Write the Graphic Control Extension
    gce = bytearray()
    gce += bytes([0x21]) # Block label
    gce += bytes([0xf9]) # Block size
    gce += bytes([0x04]) # Disposal method
    gce += bytes([0x00]) # User input flag
    gce += bytes([0xff]) # Transparent color index
    gce += bytes([0x00]) # Delay time (in hundredths of seconds)
    gce += bytes([0x00]) # Transparent color index
    gce += bytes([0x00]) # Block terminator
    out.write(gce)

    # Write the Image Descriptor
    idesc = bytearray()
    idesc += bytes([0x2c]) # Image separator
    idesc += struct.pack('<HH', 0, 0) # Image position
    idesc += struct.pack('<HH', width, height) # Image size
    idesc += bytes([0x00]) # Local color table flag
    idesc += bytes([0x00]) # Interlace flag
    idesc += bytes([0x00]) # Sort flag
    idesc += bytes([0x00]) # Size of local color table
    idesc += bytes([0x00]) # Reserved
    out.write(idesc)

    # Write the LZW-compressed image data
    out.write(b'\x00\x00') # Minimum code size
    while True:
        # Read a random byte from /dev/urandom
        c = rng.read(1)[0]

        if c == 0xFF:
            # EOF
            break

        elif c < 0x80:
            # Literal
            out.write(bytes([c]))

        else:
            # Code
            nbits = c & 0x0F
            if nbits > 0:
                # Read the next nbits bits from /dev/urandom and write them to the output
                out.write(rng.read(nbits // 8 + 1))

    # Write the trailer
    out.write(b';')
