#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0]))

    # Global color table
    for i in range(768):
        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

    # Image descriptor
    left   = int.from_bytes(rng.read(2), byteorder='little') % (width - 1) + 1
    top    = int.from_bytes(rng.read(2), byteorder='little') % (height - 1) + 1
    width  = int.from_bytes(rng.read(2), byteorder='little') % ((width - left) // 2) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % ((height - top) // 2) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lct    = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    interlaced = False if not bool(flags[0] & 0x40) else True
    sort       = False if not bool(flags[0] & 0x20) else True
    size       = 2 ** (((flags[0] >> 4) & 0b11) + 1)
    out.write(struct.pack('<HHBB', left, top, width, height))
    out.write(struct.pack('<BB', flags[0], lct[0]))

    # Local color table
    for i in range(size * size):
        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

    # Image data
    def get_pixel():
        return bytes([int.from_bytes(rng.read(1), byteorder='little')])

    def get_block(n):
        block = []
        while len(block) < n:
            block += [get_pixel()]
        return block[:n]

    def get_code(n):
        code = 0
        for _ in range(n):
            code = (code << 8) | ord(get_pixel())
        return code

    def get_subblocks():
        subblocks = []
        while True:
            subblock = get_pixel()
            if subblock == b'\0':
                break
            subblocks.append(subblock)
        return subblocks

    def decode_lzw(minimum_code_size):
        clear_code = 1 << minimum_code_size
        end_of_information_code = clear_code + 1
        available_codes = 1 << minimum_code_size + 2
        old_code = None
        code_size = minimum_code_size + 1
        dictionary = {}
        for i in range(clear_code, end_of_information_code):
            dictionary[i] = bytes([i])
        for i in range(available_codes, 1 << code_size):
            dictionary[i] = None
        while True:
            code = get_code(code_size)
            if code == end_of_information_code:
                break
            elif code == clear_code:
                code_size = minimum_code_size + 1
                available_codes = 1 << code_size + 2
                old_code = None
            elif old_code is None:
                yield dictionary[code]
                old_code = code
            elif code <= available_codes:
                entry = dictionary[old_code] + dictionary[code][0].to_bytes(1, 'big')
                yield entry
                dictionary[available_codes] = entry
                available_codes += 1
                if available_codes >= 1 << code_size:
                    code_size += 1
                old_code = code
            else:
                raise Exception("Invalid LZW code")

    def write_image_data():
        if interlaced:
            pass_offsets = [(0, 8), (4, 8), (2, 4), (1, 2)]
        else:
            pass_offsets = [(0, 1)]
        for y_offset, x_multiplier in pass_offsets:
            for y in range(y_offset, height, y_multiplier):
                for x in range(0, width, x_multiplier):
                    pixel = get_pixel()
                    out.write(pixel)

    def write_image_data_lzw():
        minimum_code_size = get_pixel()[0]
        subblocks = get_subblocks()
        for subblock in subblocks:
            for pixel in decode_lzw(minimum_code_size):
                out.write(pixel)

    if lct[0]:
        write_image_data()
    else:
        write_image_data_lzw()

    # Trailer
    out.write(b';')
