#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0]))

    # Global color table
    for i in range(768):
        out.write(rng.read(3))

    while True:
        # Graphic control extension
        gce = bytes([0x21, 0xf9, 0x04])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        out.write(gce)

        # Image descriptor
        idesc = bytes([0x2c])
        idesc += generate_image_descriptor(rng, width, height)
        out.write(idesc)

        # Local color table
        lct = bytes()
        if (flags[0] & 0b10000000) != 0:
            lct = bytes([0])
            for _ in range((1 << ((flags[0] >> 4) & 0b111)) * 3):
                lct += rng.read(1)
        out.write(lct)

        # Pixel data
        pixels = bytes()
        for y in range(height):
            for x in range(width):
                pixels += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        out.write(pixels)

        # Trailer
        trailer = bytes([0x3b])
        out.write(trailer)

def generate_image_descriptor(
        rng: BinaryIO, width: int, height: int
    ) -> bytes:
    x = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16)
    y = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16)
    w = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16)
    h = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16)
    flags = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    return struct.pack('<HHHHB', x, y, w, h, flags[0])
