#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    height = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    out.write(struct.pack('<HHBBB', width, height, 0xF0, 0x00, 0x00))

    # Write the global color table
    for i in range(768):
        out.write(bytes([(i >> ((~i + 1) % 8)) & 0xFF]))

    # Write the application extension block
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00\x00')

    # Write the image descriptor
    out.write(b'\x2c\x00\x00\x00\x00\x00\x00\x00\x00')

    # Write the local color table
    for i in range(768):
        out.write(bytes([(i >> ((~i + 1) % 8)) & 0xFF]))

    # Write the image data
    out.write(b'\x00\x00')

    # Write the LZW compressed image data
    lzw_min_code_sizes = [8, 4, 2, 0]
    lzw_clear_codes = [0x11, 0x11, 0x11, 0x11]
    lzw_end_of_information_codes = [0x11, 0x11, 0x11, 0x11]
    lzw_max_bits = [12, 12, 12, 12]

    # Initialize the LZW dictionary
    lzw_dictionary = {}
    lzw_current_bit_count = 0
    lzw_next_free_entry = 0
    lzw_output_buffer = []

    # Read the first code word
    lzw_code = read_lzw_code(rng, lzw_min_code_sizes[lzw_current_bit_count])

    while True:
        if lzw_code == lzw_clear_codes[lzw_current_bit_count]:
            # Clear the dictionary and reset the bit count
            lzw_dictionary = {}
            lzw_current_bit_count = 0
            lzw_next_free_entry = 0

            # Read the next code word
            lzw_code = read_lzw_code(rng, lzw_min_code_sizes[lzw_current_bit_count])

            continue

        if lzw_code == lzw_end_of_information_codes[lzw_current_bit_count]:
            break

        if lzw_code not in lzw_dictionary:
            # If this is the first time we have seen this code, add it to the dictionary
            lzw_dictionary[lzw_code] = [lzw_output_buffer[-1]]

            # Increment the bit count
            lzw_current_bit_count += 1

            # Check that we are still within our maximum allowed bits
            if lzw_current_bit_count > lzw_max_bits[lzw_current_bit_count]:
                raise Exception("LZW code size exceeded")

            # If we are at the maximum allowed bits, clear the dictionary
            if lzw_current_bit_count == lzw_max_bits[lzw_current_bit_count]:
                lzw_dictionary = {}
                lzw_current_bit_count = 0
                lzw_next_free_entry = 0

            # Add the new entry to the output buffer
            lzw_output_buffer.append(lzw_dictionary[lzw_code][0])

            # Read the next code word
            lzw_code = read_lzw_code(rng, lzw_min_code_sizes[lzw_current_bit_count])

            continue

        # Otherwise, we have already seen this code before
        lzw_output_buffer.extend(lzw_dictionary[lzw_code])

        # Update the dictionary with the new entry
        lzw_new_entry = lzw_output_buffer[:-1] + [lzw_output_buffer[-1]]
        lzw_dictionary[lzw_next_free_entry] = lzw_new_entry
        lzw_next_free_entry += 1

        # Read the next code word
        lzw_code = read_lzw_code(rng, lzw_min_code_sizes[lzw_current_bit_count])

    # Write the output buffer as LZW compressed data
    write_lzw_data(out, lzw_output_buffer)

    # Write the trailer
    out.write(b';')

# Reads an LZW code from the given RNG
def read_lzw_code(rng: BinaryIO, min_size: int) -> int:
    result = 0
    shift = 0

    while True:
        byte = rng.read(1)[0]

        result |= (byte & 0x7F) << shift
        shift += 7

        if byte & 0x80 != 0:
            return result

        if shift >= min_size:
            return result

# Writes LZW compressed data to the given stream
def write_lzw_data(out: BinaryIO, data: list):
    current_byte = 0
    current_