#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    height = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    out.write(struct.pack('<HHBBB', width, height, 0xF0, 0x00, 0x00))

    # Write the global color table
    for i in range(768):
        out.write(bytes([(i >> ((~i + 1) % 8)) & 0xFF]))

    # Write the image descriptor
    out.write(b'\x2C\x00\x00\x00\x00\x00\x00\x00')
    out.write(b'\x00\x00\x00\x00\x00\x00\x00\x00')

    # Write the local color table
    for i in range(768):
        out.write(bytes([(i >> ((~i + 1) % 8)) & 0xFF]))

    # Write the LZW minimum code size byte
    out.write(b'\x08')

    # Write the image data
    while True:

        # Generate a random length between 0 and 127 inclusive
        length = rng.read(1)[0] & 0x7F

        if length == 0x00:
            break

        elif length < 0x03:

            # Copy the next byte directly as many times as specified by the
            # length field
            out.write(bytes([length]) * length)

        else:

            # Read the first two bytes of the sequence
            seq = [
                rng.read(1)[0],
                rng.read(1)[0]]

            # If the second byte is less than or equal to the first one then
            # we need to add it back to the sequence so that the sequence can
            # be decoded properly
            if seq[1] <= seq[0]:

                # Add the second byte back to the sequence
                seq.append(seq[1])

                # Remove the last element from the sequence since it's already
                # been added back
                del seq[-2:]

            # Write the sequence
            out.write(bytes(seq))

