#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given PRNG.
def generate_gif(out: BinaryIO, rng):
    # Header
    header = b'GIF89a' + \
        struct.pack('<HH',
            rng.read(2)[0],       # width
            400                  # height
        )
    out.write(header)

    # Logical screen descriptor
    lsd = struct.pack('<BBBHBBBB',
        0xF7,                   # color resolution
        0,                      # sort order
        0,                      # global color table size
        0,                      # background color index
        0,                      # pixel aspect ratio
    )
    out.write(lsd)

    # Global color table
    gct = [struct.pack('BBB',
        rng.read(1)[0] % 256,   # red
        rng.read(1)[0] % 256,   # green
        rng.read(1)[0] % 256    # blue
    ) for _ in range(2**lsd[3])]
    gct += [(b'\x00\x00\x00')] * ((2**lsd[3]) - len(gct))
    out.write(bytes().join(gct))

    # Image descriptor
    idd = struct.pack('<BBHHHHHH',
        0x2C,                   # image separator
        0,                      # left position
        0,                      # top position
        400,                    # width
        400,                    # height
        0,                      # local color table size
        0                       # LZW minimum code size
    )
    out.write(idd)

    # Local color table
    lct = [struct.pack('BBB',
        rng.read(1)[0] % 256,   # red
        rng.read(1)[0] % 256,   # green
        rng.read(1)[0] % 256    # blue
    ) for _ in range(2**idd[6])]
    lct += [(b'\x00\x00\x00')] * ((2**idd[6]) - len(lct))
    out.write(bytes().join(lct))

    # Image data
    image_data = bytearray()
    while True:
        if rng.read(1)[0] > 128:
            break

        image_data += bytes([rng.read(1)[0]])

    out.write(image_data)

    # Graphics control extension
    packed = struct.pack('<BBHBHHB',
        0x21, 0xF9, 4,          # application extension block
        0,                     # no transparancy
        0,                     # delay time
        0,                     # transparent color index
        0,                     # terminator
    )
    out.write(packed)

    # Comment extension
    comment = bytes([rng.read(1)[0] % 256]) * (rng.read(1)[0] % 100)
    packed = struct.pack('<BB%dsB' % len(comment),
        0x21, 0xFE, len(comment), comment, 0x00
    )
    out.write(packed)

    # Plain text extension
    plain_text = bytes([rng.read(1)[0] % 256]) * (rng.read(1)[0] % 100)
    packed = struct.pack('<BB%dsB' % len(plain_text),
        0x21, 0x01, len(plain_text), plain_text, 0x00
    )
    out.write(packed)

    # Trailer
    out.write(b';')