#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0]))

    # Global color table
    for i in range(768):
        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

    # Image descriptor
    left   = int.from_bytes(rng.read(2), byteorder='little') % (width - 1) + 1
    top    = int.from_bytes(rng.read(2), byteorder='little') % (height - 1) + 1
    width  = int.from_bytes(rng.read(2), byteorder='little') % ((width - left) // 2) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % ((height - top) // 2) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lct    = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    interlaced = False if not bool(flags[0] & 0x40) else True
    sort       = False if not bool(flags[0] & 0x20) else True
    local_color_table_size = lct[0]
    background_index      = int.from_bytes(rng.read(1), byteorder='little')
    pixel_aspect_ratio    = int.from_bytes(rng.read(1), byteorder='little')
    out.write(struct.pack('<HHHHBB', left, top, width, height, flags[0], lct[0]))

    # Local color table
    for i in range((local_color_table_size ** 3)):
        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

    # Block size
    block_size = int.from_bytes(rng.read(1), byteorder='little')
    while block_size!= 0:
        # Extension
        label = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        ext_size = int.from_bytes(rng.read(1), byteorder='little')

        if label == b'\x21':
            sublabel = bytes([int.from_bytes(rng.read(1), byteorder='little')])

            if sublabel == b'\xf9':
                # Graphics control extension
                out.write(b'\x21\xf9')

                # Size of graphics control extension data
                out.write(ext_size.to_bytes(1, 'big'))

                # Packed fields
                packed = int.from_bytes(rng.read(1), byteorder='little')
                out.write(packed.to_bytes(1, 'big'))

                # Delay time
                delay_time = int.from_bytes(rng.read(2), byteorder='little')
                out.write(delay_time.to_bytes(2, 'big'))

                # Transparent color index
                trans_index = int.from_bytes(rng.read(1), byteorder='little')
                out.write(trans_index.to_bytes(1, 'big'))

                # Terminator
                out.write(b'\x00')
            elif sublabel == b'\xff':
                # Application extension
                out.write(b'\x21\xff')

                # Size of application extension data
                out.write(ext_size.to_bytes(1, 'big'))

                # Application identifier
                app_id = rng.read(8).rstrip(b'\x00').decode('ascii')
                out.write(app_id[:11].encode())

                # Authentication code
                auth_code = rng.read(3).rstrip(b'\x00').decode('ascii')
                out.write(auth_code[:3].encode())

                # Data sub-blocks
                block_size = int.from_bytes(rng.read(1), byteorder='little')
                while block_size!= 0:
                    out.write(block_size.to_bytes(1, 'big'))

                    data = rng.read(block_size)
                    out.write(data)

                    block_size = int.from_bytes(rng.read(1), byteorder='little')