#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bool:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = struct.pack('>H', 1024)
    height = struct.pack('>H',  768)
    flags  = struct.pack('B',   0x80 | 0x40 | 0x20)
    bg     = struct.pack('B',   0xff)
    aspect = struct.pack('B',   0x00)
    out.write(width + height + flags + bg + aspect)

    # Global color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f

        if ((r == 0) and (g == 0) and (b == 0)):
            r += 1
            g += 1
            b += 1

        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left      = struct.pack('<H',      0)
    top       = struct.pack('<H',      0)
    width     = struct.pack('<H', 1024)
    height    = struct.pack('<H',  768)
    flags     = struct.pack('B',       0x80)
    trans_idx = struct.pack('B',        0x00)
    out.write(left + top + width + height + flags + trans_idx)

    # LZW minimum code size
    min_code_size = struct.pack('B', 8)
    out.write(min_code_size)

    # Frame data
    lzw_data = []
    while len(lzw_data) < 1024 * 768 - 1:
        byte = ord(rng.read(1))

        if not byte:
            continue

        if byte >= 0x80:
            lzw_data.append((byte >> 8) & 0xFF)
            lzw_data.append(byte & 0xFF)
        else:
            lzw_data.append(byte)

    lzw_data.append(0x00)
    lzw_data.append(0x00)

    # LZW compression
    lzw_dict = {}
    lzw_prev = None
    lzw_cur  = None
    lzw_code = 256
    lzw_bits = 9
    lzw_buf  = 0
    lzw_pos  = 0
    lzw_end  = False

    def flush():
        nonlocal lzw_data, lzw_dict, lzw_prev, lzw_cur, \
                   lzw_code, lzw_bits, lzw_buf, lzw_pos, lzw_end

        if lzw_prev is not None:
            lzw_dict[lzw_code] = [lzw_prev, lzw_cur]

            lzw_data.append(lzw_code)
            lzw_data.append(lzw_prev)

        lzw_prev = lzw_cur
        lzw_cur  = None

        lzw_code += 1
        lzw_bits = 9
        lzw_buf  = 0
        lzw_pos  = 0
        lzw_end  = False

    while True:
        if lzw_end or (len(lzw_data) > 1024 * 768 - 1):
            break

        if lzw_cur is None:
            lzw_cur = ord(rng.read(1))

        if lzw_cur!= lzw_prev:
            if lzw_prev is not None:
                lzw_dict[lzw_code] = [lzw_prev, lzw_cur]

                lzw_data.append(lzw_code)
                lzw_data.append(lzw_prev)

            lzw_prev = lzw_cur
            lzw_cur  = None

            lzw_code += 1
            lzw_bits = 9
            lzw_buf  = 0
            lzw_pos  = 0
            lzw_end  = False
        elif lzw_bits == 0:
            lzw_data.append(lzw_code)
            lzw_data.append(lzw_prev)

            lzw_prev = None
            lzw_cur  = None

            lzw_code = 256
            lzw_bits = 9
            lzw_buf  = 0
            lzw_pos  = 0
            lzw_end  = False
        else:
            lzw_buf |= (lzw_cur << lzw_pos)
            lzw_pos += 1
            lzw_bits -= 1

            if lzw_pos == 8:
                lzw_data.append(lzw_code)
                lzw_data.append(lzw_buf)

                lzw_prev = None
                lzw_cur  = None

                lzw_code = 256
                lzw_bits = 9
                lzw_buf  = 0
                lzw_pos  = 0
                lzw_end  = False

    flush()

    # Write compressed data to output stream
    lzw_data = bytes(lzw_data)
    out.write(bytes([len(lzw_data)]))
    out.write(lzw_data)

    return True


if