#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):
    # Header
    out.write('GIF89a'.encode())

    # Logical screen descriptor
    width = 1 + (rng.read(1)[0] % 256)
    height = 1 + (rng.read(1)[0] % 256)
    flags = b'\x87' if rng.read(1)[0] & 1 else b'\x00'
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBBB', width, height, flags[0],
                          background_color_index, pixel_aspect_ratio))

    # Global color table
    global_color_table_size = min((width * height), 256) // 4
    for _ in range(global_color_table_size):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        out.write(struct.pack('<BBBB', r, g, b, 0xff))

    # Image descriptor
    left = 0
    top = 0
    width = 1 + (rng.read(1)[0] % 256)
    height = 1 + (rng.read(1)[0] % 256)
    flags = b'\x87' if rng.read(1)[0] & 1 else b'\x00'
    local_color_table_flag = bytes([flags[0] | ((rng.read(1)[0] & 1) << 7)])
    interlace_flag = bytes([(flags[0] >> 6) |
                            (((rng.read(1)[0] & 1) << 1) &
                             (flags[0] ^ 0b10000000))])
    sort_flag = bytes([(flags[0] >> 5) |
                       (((rng.read(1)[0] & 1) << 2) &
                        (flags[0] ^ 0b01000000))])
    reserved = b'\x00'
    local_color_table_size = min(((width * height)), 256) // 4
    local_color_table_sort_flag = \
            bytes([(local_color_table_flag[0] >> 4) |
                   (((rng.read(1)[0] & 1) << 4) &
                    local_color_table_flag[0] ^ 0b10000000)])
    local_color_table_interlace_flag = \
            bytes([(local_color_table_flag[0] >> 3) |
                   (((rng.read(1)[0] & 1) << 5) &
                    local_color_table_flag[0] ^ 0b01000000)])
    local_color_table_size = min(local_color_table_size,
                                 2 ** int(local_color_table_size / 2 -
                                         1).bit_length()
                                 )
    out.write(struct.pack('<HHBBBBB', left, top, width, height,
                          local_color_table_flag[0],
                          interlace_flag[0], sort_flag[0]))
    out.write(reserved)
    out.write(local_color_table_sort_flag)
    out.write(local_color_table_interlace_flag)
    out.write(bytes([local_color_table_size]))
    for i in range(local_color_table_size):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        out.write(struct.pack('<BBBB', r, g, b, 0xff))

    # LZW minimum code size
    lzw_minimum_code_size = max(rng.read(1)[0] % 256, 2)
    out.write(lzw_minimum_code_size.to_bytes(1, 'little'))

    # Image data
    image_data_size = 0
    while True:
        byte = rng.read(1)[0]

        if byte == 0x00 or byte == 0xff:
            out.write(byte.to_bytes(1,