#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0]))

    # Global color table
    for i in range(768):
        out.write(rng.read(3))

    while True:
        # Graphic control extension
        gce = bytes([0x21, 0xf9, 0x04])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        out.write(gce)

        # Image descriptor
        idesc = bytes([0x2c])
        idesc += generate_image_descriptor(rng, width, height)
        out.write(idesc)

        # Local color table
        if bool(idesc[9] & 0x80):
            for i in range((idesc[9] & 0x07) + 1):
                out.write(rng.read(3))

        # Image data
        lzwmincodesize = int.from_bytes(rng.read(1), byteorder='little')
        out.write(lzwmincodesize.to_bytes(1, 'big'))

        # LZW code stream
        while True:
            blocktype = int.from_bytes(rng.read(1), byteorder='little')

            if blocktype == 0x00: break

            elif blocktype == 0x01:
                numcodes = int.from_bytes(rng.read(1), byteorder='little')

                for i in range(numcodes):
                    length = int.from_bytes(rng.read(1), byteorder='little')

                    if length > 128:
                        length -= 128

                        while length >= 255:
                            out.write(b'\xff\xff')
                            length -= 255

                        out.write(((length).to_bytes(1, 'big')))

                    else:
                        out.write(length.to_bytes(1, 'big'))

                    out.write(rng.read(length))

            elif blocktype == 0xfe:
                out.write(rng.read(1))

            else: raise Exception('Invalid LZW block type.')


# Generate an image descriptor from the given parameters
def generate_image_descriptor(
        rng: BinaryIO, width: int, height: int
    ) -> bytes:
    left   = int.from_bytes(rng.read(2), byteorder='little') % (width - 1)
    top    = int.from_bytes(rng.read(2), byteorder='little') % (height - 1)
    width  = int.from_bytes(rng.read(2), byteorder='little') % ((width - left) + 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % ((height - top) + 1) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    return struct.pack('<HHBBBB', left, top, width, height, flags[0], 0)

