#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = rng.read(1)[0] % 40 + 10
    height = rng.read(1)[0] % 25 + 10
    flags = b'\x87\x00' if rng.read(1)[0] & 1 else b'\x07\x00'
    bg_color = bytes([rng.read(1)[0] for _ in range(3)])
    aspect_ratio = b'\x00'
    out.write(struct.pack('<HHBBB', width, height, *flags))
    out.write(bg_color)
    out.write(aspect_ratio)

    # Global color table
    num_colors = min((width+height)//2, 64)
    colors = [bytes([rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]])
              for i in range(num_colors)]
    while len(colors) < 2**int.from_bytes(flags[0:1], byteorder='big'):
        colors += [colors[-1]]
    out.write(b''.join(colors))

    # Image descriptor
    left = rng.read(1)[0] % (width - 10)
    top = rng.read(1)[0] % (height - 10)
    flags = b'\x87\x00' if rng.read(1)[0] & 1 else b'\x07\x00'
    delay = int.to_bytes(rng.read(1)[0] % 100, length=1, byteorder='little')
    transparent_index = rng.read(1)[0] % num_colors
    out.write(struct.pack('<HHB', left, top, *flags))
    out.write(delay)
    out.write(transparent_index.to_bytes(length=1, byteorder='little'))
    out.write(b'\x00')

    # LZW minimum code size
    lzw_min_code_size = rng.read(1)[0] % 12 + 1
    out.write(lzw_min_code_size.to_bytes(length=1, byteorder='little'))

    # Image data
    image_data = []
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x2C:
            break

        elif block_type == 0x21:
            subblock_type = rng.read(1)[0]

            if subblock_type == 0xF9:
                # Graphic control extension
                out.write(subblock_type.to_bytes(
                    length=1, byteorder='little'))

                # Block size
                out.write(b'\x04')

                # Flags and transparency index
                flags = b'\x04' if rng.read(1)[0] & 1 else b'\x00'
                out.write(flags)
                out.write(transparent_index.to_bytes(
                    length=1, byteorder='little'))

                # Delay time
                delay = int.to_bytes(rng.read(1)[0] % 100,
                                     length=1, byteorder='little')
                out.write(delay)

                # Transparent color index
                out.write(transparent_index.to_bytes(
                    length=1, byteorder='little'))

                continue

            elif subblock_type == 0xFF:
                # Application extension
                out.write(subblock_type.to_bytes(
                    length=1, byteorder='little'))

                # Block size
                out.write(b'\x0B')

                # Application identifier
                app_identifier = b'NETSCAPE2.0'
                out.write(app_identifier)

                # Application authentication code
                auth_code = b'\x03\x01'
                out.write(auth_code)

                # Sub-blocks
                while True:
                    subblock_type = rng.read(1)[0]

                    if subblock_type == 0x01:
                        # Loop count sub-block
                        out.write(subblock_type.to_bytes(
                            length=1, byteorder='little'))

                        # Block size
                        out.write(b'\x03')

                        # Loop count
                        loop_count = int.to_bytes(rng.read(1)[0] % 65536,
                                                  length=2, byteorder='little')
                        out.write(loop_count)

                        continue

                    elif subblock_type == 0x00:
                        # End of application extension sub-block
                        out.write(subblock_type.to_bytes(
                            length