#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0]))

    # Global color table
    for i in range(768):
        out.write(rng.read(3))

    while True:
        # Graphic control extension
        gce = bytes([0x21, 0xf9, 0x04])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        gce += bytes([int.from_bytes(rng.read(1), byteorder='little')])
        out.write(gce)

        # Image descriptor
        idesc = bytes([0x2c])
        idesc += generate_image_descriptor(rng, width, height)
        out.write(idesc)

        # LZW minimum code size
        lzwmincodesize = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        out.write(lzwmincodesize)

        # Data sub-blocks
        data = []
        while True:
            blocklen = int.from_bytes(rng.read(1), byteorder='little')

            if not blocklen:
                break

            data += [blocklen] * blocklen
            data += rng.read(blocklen)

        out.write(b''.join(data))


# Generates an image descriptor with the given dimensions and
# returns it as a list of bytes
def generate_image_descriptor(
        rng: BinaryIO, width: int, height: int
    ) -> bytes:
    xpos   = int.from_bytes(rng.read(2), byteorder='little') % (width - 1) + 1
    ypos   = int.from_bytes(rng.read(2), byteorder='little') % (height - 1) + 1
    width  = min(max(int.from_bytes(rng.read(2), byteorder='little'), 1), width)
    height = min(max(int.from_bytes(rng.read(2), byteorder='little'), 1), height)
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    return struct.pack('<HHBB', xpos, ypos, width, height) + flags
