#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bool:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
    height = 1 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
    flags = b'\x07\x00'
    bg_color = bytes([int.from_bytes(rng.read(1), byteorder='big', signed=False)]) * 3
    pixel_aspect_ratio = b'\x00'
    out.write(struct.pack('>HHBBB', width, height, flags[0], flags[1], bg_color[0], bg_color[1]))
    out.write(pixel_aspect_ratio)

    # Global color table
    for _ in range(0, 2**((flags[0] & 0x07) + 1)):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left = 0
    top = 0
    width = 1 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
    height = 1 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
    flags = b'\x00\x00'
    local_color_table_flag = b'\x00'
    interlace_flag = b'\x00'
    sort_flag = b'\x00'
    reserved = b'\x00\x00'
    lct_size = 2**(flags[0] & 0x07)
    out.write(struct.pack('>HHBBBBB', left, top, width, height, flags[0], flags[1], local_color_table_flag[0]))
    out.write(interlace_flag)
    out.write(sort_flag)
    out.write(reserved)
    if not ((local_color_table_flag[0] == 0 and lct_size!= 0) or
            (local_color_table_flag[0] == 1 and lct_size == 0)):
        print("Invalid LCT flag")
        return False
    for i in range(lct_size):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
        out.write(struct.pack('BBB', r, g, b))

    # Image data
    for y in range(height):
        for x in range(width):
            palette_index = int.from_bytes(rng.read(1), byteorder='big', signed=False) % len(bg_color)
            out.write(palette_index.to_bytes(1, 'big'))
    # Trailer
    out.write(b';')

    return True


