#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0]))

    # Global color table
    for i in range(768):
        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

    # Image descriptor
    left   = int.from_bytes(rng.read(2), byteorder='little') % (width - 1) + 1
    top    = int.from_bytes(rng.read(2), byteorder='little') % (height - 1) + 1
    width  = int.from_bytes(rng.read(2), byteorder='little') % ((width - left) // 2) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % ((height - top) // 2) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lct    = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    interlaced = False if not bool(flags[0] & 0x40) else True
    sort       = False if not bool(flags[0] & 0x20) else True
    local_color_table_size = lct[0]
    background_index      = int.from_bytes(rng.read(1), byteorder='little')
    pixel_aspect_ratio    = int.from_bytes(rng.read(1), byteorder='little')
    out.write(struct.pack('<HHHHBB', left, top, width, height, flags[0], lct[0]))

    # Local color table
    for i in range((local_color_table_size ** 3)):
        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

    # Block size
    block_size = int.from_bytes(rng.read(1), byteorder='little')
    while block_size!= 0:
        data = rng.read(block_size)

        # Graphics control extension
        if block_size == 4 and data[:3] == b'\x21\xF9\x04':
            out.write(data)

            # Plain text extension
            if data[3] == 0x01:
                out.write(b'\x0B')

                # Text sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

            elif data[3] == 0x02:
                out.write(b'\x0C')

                # Font selection sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

                # Text sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

            elif data[3] == 0x03:
                out.write(b'\x0D')

                # Font definition sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

                # Text sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

            elif data[3] == 0x04:
                out.write(b'\x0E')

                # Font selection sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

                # Text sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

            elif data[3] == 0x05:
                out.write(b'\x0F')

                # Font selection sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

                # Text sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

            elif data[3] == 0x06:
                out.write(b'\x10')

                # Font selection sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

                # Text sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

            elif data[3] == 0x07:
                out.write(b'\x11')

                # Font selection sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

                # Text sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

            elif data[3] == 0x08:
                out.write(b'\x12')

                # Font selection sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

                # Text sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

            elif data[3] == 0x09:
                out.write(b'\x13')

                # Font selection sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

                # Text sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\x00')

            elif data[3] == 0x0A:
                out.write(b'\x14')

                # Font selection sub-block
                out.write(b'\x01')
                out.write(b'\xFF\xFE')
                out.write(b'\x00\        block_size = int.from_bytes(rng.read(1), byteorder='little')