#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The
# width and height of each frame are determined by
# `width` and `height`.
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO,
        width: int,
        height: int):

    # The size of the entire file (including the header),
    # including the palette and all frames. This will be
    # written at the end once we know how big it actually
    # is.
    total_size = None

    def write_header():

        nonlocal total_size

        # The initial GIF header
        header = b'GIF89a'

        # Add some padding until we reach the next byte boundary
        while len(header) % 2!= 0:
            header += b'\x00'

        # Generate the rest of the header
        header += struct.pack('<H', total_size - 6)
        header += b'\x01'
        header += bytes([rng.read(1)[0]])
        header += b'\xFF\xFF\xFF'

        # Write the header
        out.write(header)


    def write_logical_screen_descriptor():

        # Write the logical screen descriptor
        out.write(struct.pack('<H', 0x200))
        out.write(b'\x01')
        out.write(bytes([rng.read(1)[0]]))
        out.write(b'\xFF\xFF\xFF')


    def write_global_color_table():

        # Write the global color table
        for i in range(0x200):
            out.write(bytes([rng.read(1)[0]]))


    def write_frame(left: int, top: int):

        # Write the first frame
        out.write(b'\x2C')
        out.write(struct.pack('<HHHHB',
            0x200, 0x200, left, top, 0x00))

        # Write the image data
        for i in range(0x4000):
            out.write(bytes([rng.read(1)[0]]))


    # Start writing the file
    write_header()

    # Calculate the total size of the file
    total_size = 0x200 + 7 + 0x200 + \
                 ((width * height) // 256) + \
                 1 + 0x4000

    # Write the logical screen descriptor
    write_logical_screen_descriptor()

    # Write the global color table
    write_global_color_table()

    # Write the first frame
    write_frame((total_size - 0x200) % 0x200,
                (total_size - 0x200) // 0x200)