#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    def write_byte(value: int) -> None:
        out.write(bytes([value]))

    def write_short(value: int) -> None:
        out.write(value.to_bytes(2, 'little'))

    def write_long(value: int) -> None:
        out.write(value.to_bytes(4, 'little'))


    # Header
    write_byte(ord('G'))
    write_byte(ord('I'))
    write_byte(ord('F'))
    write_byte(ord('8'))
    write_byte(ord('9'))
    write_byte(ord('a'))

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    write_short(width)
    write_short(height)
    write_byte(flags[0])
    write_byte(bg[0])
    write_byte(aspect[0])

    # Global color table
    for _ in range(768):
        write_byte(int.from_bytes(rng.read(1), byteorder='little'))

    # Image descriptor
    left   = int.from_bytes(rng.read(2), byteorder='little') % (width - 1) + 1
    top    = int.from_bytes(rng.read(2), byteorder='little') % (height - 1) + 1
    width  = int.from_bytes(rng.read(2), byteorder='little') % ((width - left) // 2) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % ((height - top) // 2) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lct    = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    interlaced = False if not bool(flags[0] & 0x40) else True
    sort       = False if not bool(flags[0] & 0x20) else True
    local_color_table_size = lct[0]
    background_index      = int.from_bytes(rng.read(1), byteorder='little')
    pixel_aspect_ratio    = int.from_bytes(rng.read(1), byteorder='little')
    write_short(left)
    write_short(top)
    write_short(width)
    write_short(height)
    write_byte(flags[0])
    write_byte(lct[0])

    # Local color table
    for _ in range((local_color_table_size ** 3)):
        write_byte(int.from_bytes(rng.read(1), byteorder='little'))

    # Block size
    block_size = int.from_bytes(rng.read(1), byteorder='little')
    while block_size!= 0:
        data = rng.read(block_size)

        # Graphics control extension
        if block_size == 4 and data[:3] == b'\x21\xF9\x04':
            write_byte(ord('!'))
            write_byte(ord(' '))
            write_byte(ord('g'))
            write_byte(ord('c'))
            write_byte(data[3:])

            # Plain text extension
            if data[-1:] == b'\x01':
                write_byte(ord(','))
                write_byte(ord(' '))
                write_byte(ord('P'))
                write_byte(ord('t'))
                write_byte(data[5:-1])

                # Text sub-block
                length = int.from_bytes(rng.read(1), byteorder='little')
                write_byte(length)
                write_byte(rng.read