#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    height = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    out.write(struct.pack('<HHBBB', width, height, 0xF0, 0x00, 0x00))

    # Write the global color table
    for i in range(768):
        out.write(bytes([(i >> ((~i + 1) % 8)) & 0xFF]))

    # Write the application extension block
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00\x00')

    # Write the image descriptor
    out.write(b'\x2c\x00\x00\x00\x00\x00\x00\x00\x00')

    # Write the local color table
    for i in range(768):
        out.write(bytes([(i >> ((~i + 1) % 8)) & 0xFF]))

    # Write the image data
    out.write(b'\x00\x00')

    # Write the LZW compressed image data
    while True:
        code = rng.read(1)[0]

        if code >= 0x00 and code <= 0x0f:
            # Raw byte
            out.write(bytes([code]))
        elif code >= 0x10 and code <= 0x1f:
            # Repeat previous byte
            count = 1 + (code - 0x10)
            out.write(bytes([out.peek()[0]] * count))
            out.seek(-count, 1)
        elif code >= 0x20 and code <= 0x2f:
            # Repeat next byte
            count = 1 + (code - 0x20)
            out.write(bytes([out.read(1)[0]] * count))
        elif code >= 0x30 and code <= 0x3f:
            # Repeat previous byte with offset
            count = 3 + (code - 0x30)
            out.write(bytes([out.peek((count // 3))[0]]))
            out.seek(-count, 1)
        elif code >= 0x40 and code <= 0x4f:
            # Repeat next byte with offset
            count = 3 + (code - 0x40)
            out.write(bytes([out.read(1)[0]]))
            out.write(bytes([out.read(1)[0]]))
            out.write(bytes([out.read(1)[0]] * (count - 3)))
        elif code == 0x50 or code == 0x51:
            # End of stream
            break
        elif code == 0x52:
            # End of frame
            pass
        elif code == 0x53:
            # Start of frame
            pass
        elif code == 0x54:
            # Define size
            pass
        elif code == 0x55:
            # Wait for user input
            pass
        elif code == 0x56:
            # Application specific
            pass
        elif code == 0x57:
            # Comment
            pass
        elif code == 0x58:
            # Graphic control
            pass
        elif code == 0x59:
            # Plain text
            pass
        elif code == 0x5A:
            # Application extension
            pass
        elif code == 0x5B:
            # File control
            pass
        elif code == 0x5C:
            # Trailer
            break
        elif code == 0x5D:
            # Extension introducer
            pass
        elif code == 0x5E:
            # Image descriptor
            pass
       