#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 20 15:49:56 2020

@author: jonasg
"""
import numpy as np
from scipy.stats import norm
from scipy.optimize import minimize, Bounds

def get_bounds(x):
    return Bounds([np.min(x), 0], [np.max(x), np.inf])

def loglikelihood(params, x, y):
    mu, sigma = params
    
    ll = np.sum(norm.logpdf(y, loc=mu, scale=sigma))
    
    return -ll

def fit_normal(x, y):
    bounds = get_bounds(x)
    init_guess = (np.mean(y), np.std(y))
    res = minimize(loglikelihood, init_guess, args=(x, y), method='L-BFGS-B', bounds=bounds)
    
    return res

def get_fit_data(x, y, nbins=100):
    # Fit normal distribution to data
    res = fit_normal(x, y)
    mu, sigma = res.x
    
    # Get fitted values and errors
    x_fit = np.linspace(np.min(x), np.max(x), nbins)
    y_fit = norm.pdf(x_fit, loc=mu, scale=sigma)
    y_err = np.sqrt(y_fit * (1 - y_fit) / len(x))
    
    return x_fit, y_fit, y_err

def plot_fit(ax, x, y, x_fit, y_fit, y_err, color='k'):
    ax.errorbar(x, y, fmt='o', ms=3, ls='none', markeredgewidth=0.5, elinewidth=0.5, capsize=2, ecolor=color, label='Data')
    ax.plot(x_fit, y_fit, '-', lw=1.5, color=color, label='Fitted curve')
    ax.fill_between(x_fit, y_fit + y_err, y_fit - y_err, alpha=0.2, color=color)
    ax.set_xlim(np.min(x), np.max(x))
    ax.set_ylim(bottom=0)
    ax.legend()

def plot_histogram(ax, x, y, bins=None, color='k'):
    if bins is None:
        bins = int((np.max(x) - np.min(x)) / 0.1)
        
    ax.hist(x, weights=y, bins=bins, histtype='stepfilled', color=color)
    ax.set_xlim(np.min(x), np.max(x))
    ax.set_ylim(bottom=0)

def plot_residuals(ax, x, y, y_fit, color='k'):
    residuals = y - y_fit
    ax.scatter(x, residuals, c=color, s=1)
    ax.set_xlim(np.min(x), np.max(x))
    ax.set_ylim(-0.1, 0.1)

def plot_qq(ax, x, y, color='k'):
    qq = norm.ppf(norm.cdf(x))
    ax.scatter(x, qq, c=color, s=1)
    ax.set_xlim(np.min(x), np.max(x))
    ax.set_ylim(np.min(x), np.max(x))

def plot_residual_distribution(ax, x, y, y_fit, color='k'):
    residuals = y - y_fit
    ax.hist(residuals, bins=int((np.max(residuals) - np.min(residuals)) / 0.1), histtype='stepfilled', color=color)
    ax.set_xlim(np.min(residuals), np.max(residuals))
    ax.set_ylim(bottom=0)