#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):
    # Header
    out.write("GIF89a".encode())
    
    # Logical Screen Descriptor
    width = 16 + (rng.read(1)[0] & 0xF) * 256 + rng.read(1)[0]
    height = 16 + (rng.read(1)[0] & 0xF) * 256 + rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = 0
    out.write(struct.pack("<HHBB", width, height, background_color_index, pixel_aspect_ratio))
    
    # Global Color Table
    color_table_size = 7 + rng.read(1)[0]
    for i in range(3 ** color_table_size):
        out.write(rng.read(3))
        
    # Application Extension
    block_type = 0xFF
    application_identifier = "RANDOM"
    application_authentication_code = "RANDOM"
    data_subblocks = []
    while True:
        subblock_length = rng.read(1)[0]
        if subblock_length == 0:
            break
        else:
            data_subblocks.append(rng.read(subblock_length))
            
    out.write((block_type).to_bytes(1, byteorder="little"))
    out.write(application_identifier.encode() + b"\x00")
    out.write(application_authentication_code.encode() + b"\x00")
    out.write(b"\x00")
    for subblock in data_subblocks:
        out.write(len(subblock).to_bytes(1, byteorder="little"))
        out.write(subblock)
    out.write(b"\x00")
    
    # Image Descriptor
    left = 0
    top = 0
    width = 16 + (rng.read(1)[0] & 0xF) * 256 + rng.read(1)[0]
    height = 16 + (rng.read(1)[0] & 0xF) * 256 + rng.read(1)[0]
    local_color_table_flag = 0
    interlace_flag = 0
    sort_flag = 0
    reserved_local_color_table_size = 0
    out.write(struct.pack("<HHHBBB", left, top, width, height, local_color_table_flag, interlace_flag, sort_flag, reserved_local_color_table_size))
    
    # Local Color Table
    for i in range(3 ** reserved_local_color_table_size):
        out.write(rng.read(3))
        
    # LZW Minimum Code Size
    minimum_code_size = 8 + rng.read(1)[0]
    out.write(minimum_code_size.to_bytes(1, byteorder="little"))
    
    # Image Data
    code_size = minimum_code_size + 1
    clear_code = 1 << minimum_code_size
    end_of_information_code = clear_code + 1
    table = [chr(i) for i in range(clear_code, end_of_information_code)]
    stack = []
    while True:
        try:
            code = int.from_bytes(rng.read(code_size // 8), byteorder="big")
        except ValueError:
            break
        if code == clear_code:
            code_size = minimum_code_size + 1
            table = [chr(i) for i in range(clear_code, end_of_information_code)]
            continue
        elif code == end_of_information_code:
            break
        elif len(stack) > 4096:
            raise Exception("Stack overflow")
        elif code < len(table):
            stack.append(table[code])
        else:
            if code >= len(table):
                raise Exception("Invalid code")
            stack.append(table[code])
            if len(