#!/usr/bin/env python3

import os
import sys
from pathlib import Path
import subprocess as sp
import shutil
import argparse
import re
import json
import tempfile
import time
import datetime
import logging
import multiprocessing
import threading
import queue
import signal
import traceback
import psutil
import socket
import requests
import urllib.request
import xml.etree.ElementTree as ET
import yaml

# Global variables
logger = None
log_queue = None
log_thread = None
log_lock = threading.Lock()

def get_logger(name):
    global logger, log_queue, log_thread, log_lock
    if not logger:
        # Create a new logger
        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)

        # Create the log queue and start the log thread
        log_queue = queue.Queue(-1)
        log_thread = threading.Thread(target=log_worker, args=(log_queue,))
        log_thread.daemon = True
        log_thread.start()

    return logger

def log_worker(q):
    while True:
        try:
            record = q.get()
            if record is None:
                break
            logger.handle(record)
        except Exception as e:
            print("Exception in log worker:", file=sys.stderr)
            traceback.print_exc()
            continue

def log_init():
    global logger, log_queue, log_thread, log_lock
    with log_lock:
        if logger:
            raise RuntimeError("Logging system already initialized")

        # Set up a handler to write messages to stderr
        consoleHandler = logging.StreamHandler()
        consoleHandler.setFormatter(logging.Formatter('%(message)s'))
        logger.addHandler(consoleHandler)

        # Set up a handler to write messages to a temporary file
        tmpdir = tempfile.mkdtemp()
        logFilePath = os.path.join(tmpdir, "test-runner.log")
        fileHandler = logging.FileHandler(logFilePath)
        fileHandler.setFormatter(logging.Formatter("%(asctime)s %(levelname)-8s %(message)s"))
        logger.addHandler(fileHandler)

        # Add a handler that will forward messages to the main process via the
        # pipe
        class QueueHandler(logging.Handler):
            def __init__(self, q):
                super().__init__()
                self.q = q

            def emit(self, record):
                self.q.put(record)

        queueHandler = QueueHandler(log_queue)
        logger.addHandler(queueHandler)

def log_shutdown():
    global logger, log_queue, log_thread, log_lock
    with log_lock:
        if not logger:
            return

        # Remove all handlers so we don't get duplicate output
        for h in list(logger.handlers):
            logger.removeHandler(h)

        # Send a sentinel value to tell the log thread to shut down
        log_queue.put(None)
        log_thread.join()

        # Close the log file
        for h in logger.handlers:
            if isinstance(h, logging.FileHandler):
                h.close()

        # Clear out the logger instance
        logger = None
        log_queue = None
        log_thread = None
