#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom or /dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    header = b'GIF89a'
    out.write(header)

    # Logical screen descriptor
    width = 2**7 * (struct.unpack('B', rng.read(1))[0] % 512 + 1)
    height = 2**7 * (struct.unpack('B', rng.read(1))[0] % 512 + 1)
    packed = struct.pack('<HHBBB', width, height, 0xf7, 0, 0)
    out.write(packed)

    # Global color table
    global_color_table_size = 2**3 * (struct.unpack('B', rng.read(1))[0] % 8 + 1)
    for _ in range(global_color_table_size):
        rgb = [struct.unpack('B', rng.read(1))[0] for i in range(3)]
        out.write(bytes(rgb))

    # Application extension block
    application_extension_block_label = bytes([0x21, 0xff])
    application_identifier = b'NETSCAPE2.0'
    application_auth_code = b'\<PASSWORD>'
    subblocks = []
    subblocks += [(1, b'\x0b')]
    subblocks += [(1, b'\x03\x01\x00')])
    subblocks += [(1, b'\x00')])
    for length, content in subblocks:
        packed = struct.pack('BB', length, ord(content[0]))
        out.write(packed)
        out.write(content[1:])

    # Graphics control extension block
    graphics_control_extension_block_label = bytes([0x21, 0xf9])
    packed = struct.pack('<BBBBHB', 0x04, 0x04, 0x00, 0x00, 0x00, 0x00)
    out.write(packed)

    # Plain text extension block
    plain_text_extension_block_label = bytes([0x21, 0x01])
    packed = struct.pack('<BBBB', 0x00, 0x00, 0x00, 0x00)
    out.write(packed)

    # Comment extension block
    comment_extension_block_label = bytes([0x21, 0xfe])
    comment = b'Created by <NAME>!'
    packed = struct.pack('B', len(comment))
    out.write(packed)
    out.write(comment)

    # Frames
    # Image data
    datalen = 2**10 * (struct.unpack('B', rng.read(1))[0] % 64 + 1)
    data = rng.read(datalen)
    out.write(data)

    # Trailer
    trailer = bytes([0x3b])
    out.write(trailer)


if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == '-':
        rng = sys.stdin.buffer
    else:
        try:
            rng = open('/dev/urandom', 'rb')
        except FileNotFoundError:
            rng = open(os.path.join('/', 'dev', 'urandom'), 'rb')
    with open('out.gif', 'wb') as f:
        generate_random_gif(rng, f)