#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom or /dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    header = b'GIF89a'
    out.write(header)

    # Logical screen descriptor
    width = 2**10 * (struct.unpack('B', rng.read(1))[0] % 7 + 1)
    height = 2**10 * (struct.unpack('B', rng.read(1))[0] % 5 + 1)
    packed = struct.pack('<HHBBB', width, height, 0xF7, 0x00, 0x00)
    out.write(packed)

    # Global color table
    num_colors = min((width+height)//4, 256)
    for i in range(num_colors):
        r = struct.unpack('B', rng.read(1))[0]
        g = struct.unpack('B', rng.read(1))[0]
        b = struct.unpack('B', rng.read(1))[0]
        out.write(bytes([r,g,b]))

    # Image descriptor
    packed = struct.pack('<HHHHBB', 0, 0, width, height, 0x00, 0x00)
    out.write(packed)

    # LZW minimum code size
    lzw_mincodesize = max(struct.unpack('B', rng.read(1))[0], 2)
    out.write(bytes([lzw_mincodesize]))

    # Pixel data
    pixeldata = bytes()
    while len(pixeldata) < width*height:

        # Get next byte from RNG and append to pixeldata
        nextbyte = struct.unpack('B', rng.read(1))[0]
        if nextbyte == 0xFF:
            continue
        elif nextbyte == 0xFE:
            pixeldata += b'\x00\x00'
        else:
            pixeldata += bytes([nextbyte])

    # Compress with LZW
    compressed = compress_with_lzw(pixeldata, lzw_mincodesize)
    out.write(compressed)

    # Trailer
    trailer = b';'
    out.write(trailer)



# Compresses `data` using LZW compression with minimum code size of
# `mincodesize`. Returns the compressed data as a bytestring.
def compress_with_lzw(
        data: bytes, mincodesize: int
    ) -> bytes:

    # Initialize dictionary
    dict_size = 2**(mincodesize+1)
    dict = [None]*dict_size
    for i in range(dict_size):
        dict[i] = bytes([i])

    # Initialize variables
    clear_code = 2**mincodesize
    end_of_info_code = clear_code + 1
    end_of_stream_code = end_of_info_code + 1
    prev_code = -1
    curr_code = -1
    bitbuffer = 0
    bitcount = 0
    output = []

    def flushbits():
        nonlocal bitbuffer
        nonlocal bitcount
        if bitcount > 0:
            output.append(bitbuffer & ((1<<bitcount)-1))
            bitbuffer >>= bitcount
            bitcount = 0

    # Process each byte in input data
    for c in data:

        # If we have reached the end of our current code, write it to the
        # output stream
        if curr_code >= dict_size-1:

            # Write clear code
            if curr_code == clear_code:
                output.append(clear_code)

                # Flush bits
                flushbits()

            # Write end of info code
            elif curr_code == end_of_info_code:
                output.append(end_of_info_code)

                # Flush bits
                flushbits()

            # Write previous code
            else:
                assert prev_code!= -1
                output.extend(dict[prev_code])

                # Flush bits
                flushbits()

            # Reset code
            prev_code = curr_code

        # Add new character to dictionary
        dict.append(bytes([c]))
        dict_size += 1

        # Update current code
        curr_code += 1

        # Check if we need more bits
        if curr_code >= (1<<(bitcount+1)):

            # Append to buffer
            bitbuffer |= (curr_code >> bitcount)
            bitcount += 1

            # Output buffer
            while bitcount >= 8:
                output.append(bitbuffer & 0xff)
                bitbuffer >>= 8
                bitcount -= 8


    # End of stream
    output.append(end_of_stream_code)

    return bytes(output)




if