#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng):
    # Header (13 bytes)
    out.write(b'GIF89a')

    # Logical screen descriptor (7 bytes)
    width = 256 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 100
    height = 256 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 100
    flags = b'\x87\x00' if bool(int.from_bytes(rng.read(1), byteorder='big', signed=False)) else b'\x00\x00'
    bg_color_index = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xf
    aspect_ratio = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
    out.write(struct.pack('>HHBBB', width, height, flags[0], flags[1], bg_color_index, aspect_ratio))

    # Global color table (3 * 256 = 768 bytes)
    for i in range(256):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor (10 bytes)
    left = 0
    top = 0
    width = 256 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 100
    height = 256 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 100
    flags = b'\x87\x00' if bool(int.from_bytes(rng.read(1), byteorder='big', signed=False)) else b'\x00\x00'
    local_color_table_flag = b'\x01' if bool(int.from_bytes(rng.read(1), byteorder='big', signed=False)) else b'\x00'
    interlace_flag = b'\x40' if bool(int.from_bytes(rng.read(1), byteorder='big', signed=False)) else b'\x00'
    sort_flag = b'\x20' if bool(int.from_bytes(rng.read(1), byteorder='big', signed=False)) else b'\x00'
    reserved = b'\x00'
    local_color_table_size = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x0f
    out.write(struct.pack('>HHBBBBB', left, top, width, height, flags[0], flags[1], local_color_table_flag[0]))

    # Local color table (3 * 2^(local_color_table_size+1) = 2^((local_color_table_size+1)*3) bytes)
    for i in range(1 << ((local_color_table_size + 1) * 3)):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        out.write(struct.pack('BBB', r, g, b))

    # Image data block (width * height bytes)
    for i in range(height):
        for j in range(width):
            out.write(struct.pack('B', int.from_bytes(rng.read(1), byteorder='big', signed=False)))

    # Trailer (1 byte)
    out.write(b';')

