#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big') + 1

    # Write the header and length of data to the output stream
    out.write(header)
    out.write(datalen.to_bytes(3, 'big'))

    # Generate the palette for this image
    palette = [0] * (768 // 4)
    for i in range(len(palette)):
        r = next(rng) & 255
        g = next(rng) & 255
        b = next(rng) & 255

        palette[i] = ((r << 24) |
                      (g << 16) |
                      (b <<  8))

    # Write the palette to the output stream
    for p in palette:
        out.write(p.to_bytes(4, 'little', signed=False))

    # Write the actual pixel data to the output stream
    while True:
        code = next(rng) & 255
        if code == 0x00: break;
        elif code < 0x21: continue;
        elif code <= 0x2C or \
             code == 0x3B or \
             code >= 0x3F:
            raise Exception('Invalid LZW code encountered!')

        # Read the bytes from the input stream that make up the prefix
        prefix = []
        for _ in range(code):
            prefix += [next(rng)]

        # Read the first byte of the suffix
        suffix = [next(rng)]

        # Read the rest of the suffix until we get back to the prefix
        while not all([prefix[i] == suffix[-1] for i in range(len(prefix))]):
            suffix += [next(rng)]

        # Write the encoded sequence to the output stream
        out.write((code - 0x21).to_bytes(1, 'little', signed=False))
        out.write(struct.pack('<H', len(suffix)))
        out.write(bytearray(suffix))

