#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
    out: BinaryIO,
    rng: BinaryIO):

    # Write header
    out.write(b'GIF89a')

    # Write logical screen descriptor
    width = 1 + int.from_bytes(rng.read(1), byteorder='big') % 256
    height = 1 + int.from_bytes(rng.read(1), byteorder='big') % 256
    flags = b'\x07\x00' if (width == height and width <= 64) else b'\x00\x00'
    out.write(struct.pack('<HHBBB',
        width, height,
        0xf0 | ((int.from_bytes(flags[0], byteorder='big') & 0x01) << 7)
            | ((int.from_bytes(flags[1], byteorder='big') & 0x01) << 6),
        0xff,
        0xff))

    # Write global color table
    for i in range(0, 2 ** int.from_bytes(flags[0], byteorder='big')):
        r = int.from_bytes(rng.read(1), byteorder='big') % 256
        g = int.from_bytes(rng.read(1), byteorder='big') % 256
        b = int.from_bytes(rng.read(1), byteorder='big') % 256
        out.write(struct.pack('BBB', r, g, b))

    # Write image descriptor
    left = 0
    top = 0
    width = 1 + int.from_bytes(rng.read(1), byteorder='big') % 256
    height = 1 + int.from_bytes(rng.read(1), byteorder='big') % 256
    flags = b'\x0c\x00'
    out.write(struct.pack('<HHHHB',
        left, top, width, height,
        int.from_bytes(flags[0], byteorder='big')))

    # Write local color table
    for i in range(0, 2 ** int.from_bytes(flags[1], byteorder='big')):
        r = int.from_bytes(rng.read(1), byteorder='big') % 256
        g = int.from_bytes(rng.read(1), byteorder='big') % 256
        b = int.from_bytes(rng.read(1), byteorder='big') % 256
        out.write(struct.pack('BBB', r, g, b))

    # Write graphics control extension
    label = b'\x21\xf9\x04'
    blocksize = len(label) + 1
    flags = b'\x04'
    delay = 1 + int.from_bytes(rng.read(1), byteorder='big') % 256
    transparent = int.from_bytes(rng.read(1), byteorder='big') % 256
    out.write(struct.pack('BBBBBB',
        label[0], label[1],
        blocksize,
        flags[0],
        delay,
        transparent))

    # Write plain text extension
    label = b'\x21\x01\x00\x00'
    blocksize = len(label) + 1
    text = bytes([int.from_bytes(rng.read(1), byteorder='big') % 256]) * \
           int.from_bytes(rng.read(1), byteorder='big') % 256
    out.write(struct.pack('BBBB',
        label[0], label[1],
        blocksize,
        len(text)))
    out.write(text)

    # Write application extension
    label = b'\x21\xff\x0bNETSCAPE2.0'
    blocksize = len(label) + 1
    appname = b'NS20'
    applength = len(appname) + 3
    subblocksize = applength + 1
    loopcount = int.from_bytes(rng.read(2), byteorder='big') % 65536
    out.write(struct.pack('BBBB',
        label[0], label[1],
        blocksize,
        subblocksize))
    out.write(applength.to_bytes(1, byteorder='big'))
    out.write(appname)
    out.write(subblocksize.to_bytes(1, byteorder='big'))
    out.write(loopcount.