#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO, Tuple


def read_uint8(f: BinaryIO) -> int:
    return ord(f.read(1))


def read_uint16(f: BinaryIO) -> int:
    return struct.unpack("<H", f.read(2))[0]


def read_uint32(f: BinaryIO) -> int:
    return struct.unpack("<I", f.read(4))[0]


def read_int32(f: BinaryIO) -> int:
    return struct.unpack("<i", f.read(4))[0]


def read_float(f: BinaryIO) -> float:
    return struct.unpack("<f", f.read(4))[0]


def read_string(f: BinaryIO) -> str:
    length = read_uint16(f)

    if length == 0xFFFF:
        # String is null-terminated
        string = ""

        while True:
            char = f.read(1).decode("ascii")

            if not char or char == "\x00":
                break

            string += char
    else:
        # String has a fixed size
        string = f.read(length).decode("utf-8")

    return string


def write_uint8(f: BinaryIO, value: int):
    f.write(struct.pack("<B", value))


def write_uint16(f: BinaryIO, value: int):
    f.write(struct.pack("<H", value))


def write_uint32(f: BinaryIO, value: int):
    f.write(struct.pack("<I", value))


def write_int32(f: BinaryIO, value: int):
    f.write(struct.pack("<i", value))


def write_float(f: BinaryIO, value: float):
    f.write(struct.pack("<f", value))


def write_string(f: BinaryIO, value: str):
    encoded = value.encode()

    if len(encoded) > 0xFFFF:
        raise Exception("String too long to be written as a UTF-8 string!")

    write_uint16(f, len(encoded))
    f.write(encoded)


class