#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write(b'GIF89a')

    width = 1 + int.from_bytes(rng.read(2), byteorder='big', signed=False) % 640
    height = 1 + int.from_bytes(rng.read(2), byteorder='big', signed=False) % 480
    flags = b'\x87\x00' if (width >= 256 or height >= 256) else b'\x07\x00'
    out.write(struct.pack('<HHB', width, height, ord(flags[0])))

    # Global color table
    colors = [
        bytes([int.from_bytes(rng.read(1), byteorder='big', signed=False)]) for _ in range(2 ** ((ord(flags[1]) & 0b00000110) >> 1))
    ]
    out.write(b''.join(colors))

    # Image descriptor
    out.write(b'\x2C')
    left = 0
    top = 0
    width = 1 + int.from_bytes(rng.read(2), byteorder='big', signed=False) % 640
    height = 1 + int.from_bytes(rng.read(2), byteorder='big', signed=False) % 480
    flags = b'\x87\x00' if (width >= 256 or height >= 256) else b'\x07\x00'
    out.write(struct.pack('<HHBBBBB', left, top, width, height, ord(flags[0]), ord(flags[1])))

    # Local color table
    if len(colors) < 2 ** ((ord(flags[1]) & 0b00000110) >> 1):
        colors += [
            bytes([int.from_bytes(rng.read(1), byteorder='big', signed=False)]) for _ in range(
                2 ** ((ord(flags[1]) & 0b00000110) >> 1) - len(colors)
            )
        ]
        out.write(b''.join(colors[-len(colors) :]))

    # Image data
    lzw_mincodesize = 2 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 12
    out.write(lzw_mincodesize.to_bytes(1, 'big'))

    codesize = 2 ** lzw_mincodesize
    clear_code = 1 << lzw_mincodesize
    end_of_information_code = clear_code + 1
    next_code = None
    codes = []
    buffer = ''
    while True:
        if not next_code:
            try:
                next_byte = rng.read(1)[0]
            except IndexError:
                break

            buffer += f'{next_byte:08b}'

        if len(codes) == codesize and next_code!= end_of_information_code:
            raise ValueError('Code size exceeded!')

        if len(buffer) > 8:
            bit = buffer[:8]
            del buffer[:8]
        elif len(buffer) <= 8:
            bit = buffer
            buffer = ''

        if next_code is None:
            if bit == '0':
                next_code = int(buffer, base=2)
                del buffer[:]
            elif bit == '1':
                next_code = int(buffer + rng.read(1).hex(), base=2)
                del buffer[:]
        else:
            next_code <<= 1
            if bit == '1':
                next_code |= 1

        if next_code == clear_code:
            codes = list(range(clear_code, clear_code + codesize))
            next_code = None
        elif next_code == end_of_information_code:
            break
        elif next_code < len(codes):
            pass
        else:
            codes.append(next_code)
            next_code = None

    codes = {c: i for i, c in enumerate(codes)}
    out.write(end_of_information_code.to_bytes(2, 'big'))
    out.write((clear_code - 1).to_bytes(2, 'big'))
    out.write(b'\x00\x00')

    def write_code(code):
        if code < clear_code:
            out.write(code.to_bytes(1, '