#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = 1024
    height = 768
    packed = struct.pack('>HHBBB', width, height, 0x80, 0, 0)
    out.write(packed)

    # Global color table
    for i in range(0, 256):
        r = next(rng) & 0xFF
        g = next(rng) & 0xFF
        b = next(rng) & 0xFF

        out.write(struct.pack('<BBB', r, g, b))

    # Image descriptor
    packed = struct.pack('>HHBBH', 0, 0, 0, 0, 0)
    out.write(packed)

    # LZW minimum code size
    lzwminsize = 2
    out.write(bytes([lzwminsize]))

    # Frame data
    framedata = bytearray()

    while len(framedata) < 1024 * 768 / 2:
        # Generate a new byte
        if len(framedata) % 2 == 0:
            framedata += bytes([next(rng)])
        else:
            framedata[-1] |= ((next(rng) >> 7) & 0x01) << 1

    # Compress it with LZW
    compressed = compress_with_lzw(framedata)

    # Write it to the output stream
    out.write(compressed)


# Compresses the given input buffer using LZW compression
def compress_with_lzw(inputbuf: bytearray) -> bytearray:
    # Initialize the dictionary
    dictsize   = 256
    dict       = [i for i in range(dictsize)]
    dictindex  = dictsize - 1

    # Output buffer
    output     = bytearray()

    # The first two codes are always written directly
    output    += bytes([dict[0]])
    output    += bytes([dict[1]])

    # The third code is always written as a literal
    output    += bytes([inputbuf[0]])

    # The fourth and further codes are encoded using LZW
    prevcode   = inputbuf[0] | (inputbuf[1] << 8)
    curcode    = None
    bitcount   = 16
    maxbit     = 12

    for i in range(2, len(inputbuf)):
        # Get the current code from the input buffer
        curcode = inputbuf[i] | (inputbuf[i + 1] << 8)

        # Check if we need more bits
        if bitcount >= maxbit:
            # Add the most significant bit of the previous code
            output += bytes([(prevcode >> (maxbit - 1))])

            # If there's still another bit available...
            if bitcount > maxbit:
                #...add that too
                output += bytes([(prevcode >> (maxbit - 2)) & 0x01])

                # Update the maximum amount of bits used
                maxbit += 1

        # Encode the code
        output += bytes([(curcode >> (maxbit - bitcount - 1)) & 0x01])

        # Update the bit counter
        bitcount += 1

        # If this was not the last code...
        if i!= len(inputbuf) - 1:
            #...check if the code exists already
            index = find_in_dictionary(dict, dictindex, prevcode, curcode)

            # If so, write its index instead
            if index!= -1:
                output += bytes([index])
            else:
                # Otherwise add both codes to the dictionary
                dictindex += 1
                dict.append((prevcode << 8) | curcode)
                dict.append(None)

                # And write their indices
                output += bytes([dictindex])
                output += bytes([dictindex - 1])

        # Update the previous code
        prevcode = curcode

    return output


# Searches the dictionary for an entry matching the given codes
def find_in_dictionary(
        dict: list, dictindex: int, prevcode: int, curcode: int
    ) -> int:
    # Try all possible entries starting at the end of the dictionary
    for i in reversed(range(dictindex + 1)):
        # If the entry matches our codes...
        if dict[i] == (prevcode << 8) | curcode:
            # Return its index
            return i

    # No match found
    return -1


if