#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (os.urandom(1)[0] % 64) + 128
    height = (os.urandom(1)[0] % 64) + 128
    flags  = bytes([
        0x80 | ((os.urandom(1)[0] & 7) << 5),
        (os.urandom(1)[0] & 0x01) << 7,
        (os.urandom(1)[0] & 0x01) << 6])
    bgcol  = bytes([os.urandom(1)[0], os.urandom(1)[0], os.urandom(1)[0]])
    pxsize = bytes([os.urandom(1)[0]])
    out.write(struct.pack('>HHBBB', width, height, flags[0], flags[1], flags[2]))
    out.write(bgcol)
    out.write(pxsize)

    # Global color table
    gctlen = (os.urandom(1)[0] % 256) + 1
    for i in range(gctlen):
        r = os.urandom(1)[0]
        g = os.urandom(1)[0]
        b = os.urandom(1)[0]
        out.write(bytes([r, g, b]))

    # Image descriptor
    left   = os.urandom(1)[0] % width
    top    = os.urandom(1)[0] % height
    width  = (os.urandom(1)[0] % min(width - left, 64)) + 1
    height = (os.urandom(1)[0] % min(height - top, 64)) + 1
    flags  = bytes([
        0x80 | ((os.urandom(1)[0] & 7) << 5),
        (os.urandom(1)[0] & 0x01) << 7,
        (os.urandom(1)[0] & 0x01) << 6])
    lctlen = (os.urandom(1)[0] % 256) + 1
    interl = bytes([os.urandom(1)[0]])
    out.write(struct.pack('>HHBBBBBB', left, top, width, height, flags[0], flags[1], flags[2], lctlen))
    if lctlen > 0:
        for i in range(lctlen):
            r = os.urandom(1)[0]
            g = os.urandom(1)[0]
            b = os.urandom(1)[0]
            out.write(bytes([r, g, b]))
    else:
        out.write(b'\x00\x00\x00')

    # Local image data
    lzwmincodesize = os.urandom(1)[0] % 129
    out.write(bytes([lzwmincodesize]))
    lzwmaxcodesize = max(lzwmincodesize + 1, 129)
    out.write(bytes([lzwmaxcodesize]))

    # LZW code stream
    while True:

        # Generate a new byte
        c = os.urandom(1)[0]

        # If it's not a clear code or end of information code...
        if c!= 0 and c!= 127:

            # Write it to the output
            out.write(bytes([c]))

        elif c == 0:

            # Write the clear code
            out.write(bytes([c]))

            # And set the minimum code size to be one more than the previous maximum
            lzwmincodesize += 1
            out.write(bytes([lzwmincodesize]))

        elif c == 127:

            # Write the end-of-information code
            out.write(bytes([c]))

            # We're done!
            break

    return


if