#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write("GIF89a".encode())

    width = 1 + int.from_bytes(rng.read(2), byteorder="little") % 50
    height = 1 + int.from_bytes(rng.read(2), byteorder="little") % 50

    flags = b"\x87\x00" if (width > 64 or height > 64) else b"\x00\x00"

    out.write(struct.pack("<HHBBB", width, height, *flags))

    palette = [
        bytes([int.from_bytes(rng.read(1), "little"),
                int.from_bytes(rng.read(1), "little"),
                int.from_bytes(rng.read(1), "little")]).ljust(3, b'\x00') for _ in range(2**((flags[0] & 0b111)+1))]

    out.write(b"".join(palette))

    out.write(b"\xff\x00")

    lzw_mincodesize = min(max(int.from_bytes(rng.read(1), "little"), 2), 12)

    out.write(lzw_mincodesize.to_bytes(1, "little"))

    lzw_table = []

    def add_to_lzw_table(*args):
        lzw_table += list(args)

        return len(lzw_table)-len(args)

    lzw_clear_code = add_to_lzw_table(0b11111110, 0b10)
    lzw_end_of_info_code = add_to_lzw_table(0b11111111, 0b10)
    lzw_end_of_image_code = add_to_lzw_table(0b11111111, 0b11)

    lzw_next_code = lzw_clear_code+1

    while True:
        lzw_current_byte = ord(rng.read(1))

        if lzw_current_byte == 0:
            break

        if lzw_current_byte < lzw_next_code and lzw_next_code <= 4095:
            lzw_table += [(lzw_table[-1][0], lzw_table[-1][1]+1)]

            lzw_next_code += 1

        if lzw_next_code >= 4096:
            out.write(lzw_table[lzw_clear_code])

            lzw_next_code = lzw_end_of_info_code+1

        current_code = -1

        for code in lzw_table:
            if code[0] == lzw_current_byte and code[1] <= lzw_next_code-lzw_clear_code:
                current_code = lzw_table.index(code)

                break

        if current_code!= -1:
            out.write(lzw_table[current_code])

            if lzw_next_code < 4096:
                lzw_table += [(lzw_table[current_code][0], lzw_table[current_code][1]+1)]

                lzw_next_code += 1
        elif current_code == -1:
            out.write(lzw_table[lzw_table.index((lzw_table[-1][0], lzw_table[-1][1]-1))])

            lzw_table[-1] = (lzw_table[-1][0], lzw_table[-1][1]+1)

    out.write(lzw_table[lzw_end_of_image_code])

