#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom or /dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    header = b'GIF89a'
    out.write(header)

    # Logical screen descriptor
    width = 2**7 * (struct.unpack('B', rng.read(1))[0] % 512 + 1)
    height = 2**7 * (struct.unpack('B', rng.read(1))[0] % 512 + 1)
    flags = struct.pack('<H', 0xf7de)
    bgcolorindex = struct.pack('<B', 0)
    pixelaspectratio = struct.pack('<B', 0)
    lsd = width + height + flags + bgcolorindex + pixelaspectratio
    out.write(lsd)

    # Global color table
    numcolors = min((width*height)//64, 256)
    gct = bytearray()
    for i in range(numcolors):
        r = struct.pack('<B', struct.unpack('B', rng.read(1))[0] % 256)
        g = struct.pack('<B', struct.unpack('B', rng.read(1))[0] % 256)
        b = struct.pack('<B', struct.unpack('B', rng.read(1))[0] % 256)
        gct += r + g + b
    out.write(gct)

    # Application extension block
    appext = b'\x21\xff\x0bNETSCAPE2.0'
    appext += b'\x03\x01'
    loops = struct.pack('<H', 0)
    appext += loops
    out.write(appext)

    # Image data
    datalen = 2**10 * (struct.unpack('B', rng.read(1))[0] % 64 + 1)
    data = rng.read(datalen)
    out.write(data)

    # Trailer
    trailer = bytes([0x3b])
    out.write(trailer)


if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == '-':
        rng = sys.stdin.buffer
    else:
        try:
            rng = open('/dev/urandom', 'rb')
        except FileNotFoundError:
            rng = open(os.path.join('/', 'dev', 'urandom'), 'rb')
    with open('out.gif', 'wb') as f:
        generate_random_gif(rng, f)