#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The width and height of each frame are between 10px and 256px.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Generate the header
    header = b'GIF89a'
    out.write(header)

    # Generate the logical screen descriptor
    w = rng.randint(10, 256)
    h = rng.randint(10, 256)
    flags = (
            0b1 << 7 | # Global color table flag
            0b1 << 4 | # Color resolution
            0b1 << 3 | # Sort flag
            0b1 << 2   # Size of global color table
        )
    gctsize = rng.randint(2, 256) - 1
    lsd = bytes([w, h]) + flags.to_bytes(1, 'little') \
          + gctsize.to_bytes(1, 'little')
    out.write(lsd)

    # Generate the global color table if necessary
    if gctsize > 0:
        for i in range(gctsize):
            r = rng.randint(0, 256)
            g = rng.randint(0, 256)
            b = rng.randint(0, 256)
            out.write(r.to_bytes(1, 'little'))
            out.write(g.to_bytes(1, 'little'))
            out.write(b.to_bytes(1, 'little'))

    # Generate the frames
    while True:

        # Generate the image descriptor
        x = rng.randint(0, w-1)
        y = rng.randint(0, h-1)
        w = rng.randint(10, min(x+10, w))
        h = rng.randint(10, min(y+10, h))
        flags = (
                0b1 << 7 | # Local color table flag
                0b1 << 6 | # Interlace flag
                0b1 << 5 | # Sort flag
                0b1 << 4   # Reserved
            )
        lctsize = rng.randint(2, 256) - 1
        id = bytes([x, y]) + bytes([w, h]) + flags.to_bytes(1, 'little') \
             + lctsize.to_bytes(1, 'little')
        out.write(id)

        # Generate the local color table if necessary
        if lctsize > 0:
            for i in range(lctsize):
                r = rng.randint(0, 256)
                g = rng.randint(0, 256)
                b = rng.randint(0, 256)
                out.write(r.to_bytes(1, 'little'))
                out.write(g.to_bytes(1, 'little'))
                out.write(b.to_bytes(1, 'little'))

        # Generate the LZW minimum code size
        lzwmincodesize = rng.randint(2, 256)
        out.write(lzwmincodesize.to_bytes(1, 'little'))

        # Generate the compressed data
        datalen = rng.randint(10, 256) * rng.randint(10, 256)
        data = rng.read(datalen)
        out.write(data)

        # Check if we're done
        if len(data) < 2: break