#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The first 6 bytes are always "GIF87a" or "GIF89a", depending on whether
    # we want to support old software (which doesn't understand newer features)
    # or new software (which does). We'll just use the latest version.
    out.write(b'GIF89a')

    # Next comes a 2-byte field that tells us how many bytes there will be in
    # the actual image data section. This is not known until after we've written
    # all the headers and such.
    datalen = rng.read(2)
    out.write(datalen)

    # Now comes the logical screen descriptor. It's pretty much fixed for now.
    # The width and height are each stored as an unsigned short, so they can
    # only go up to 65535 pixels wide or tall.
    out.write(struct.pack('>HHBBB',
        100, 100,  # Width and height
        8,         # Bits per pixel
        0,         # Background color index
        0          # Pixel aspect ratio
    ))

    # Then come the global color table. There are two options here: either this
    # is empty (meaning no palette was used), or it contains exactly 256 RGB
    # triplets. For now, let's assume the latter.
    out.write(b'\x00\x00\xff' * 256)

    # Finally comes the loop control block. This one says that the animation
    # should repeat forever.
    out.write(b'\x00\x00')

    # Now comes the image descriptor. This one says that the image has no local
    # color table, uses the entire global color table, starts at position x=y=0,
    # and is 100x100 pixels large.
    out.write(struct.pack('>HHHHBB',
        0, 0,      # Left and top coordinates
        100, 100,  # Width and height
        0,        # Local color table flag
        0,        # Interlace flag
        0,        # Sort flag
        0         # Reserved
    ))

    # And finally comes the image data! This is where the fun begins.
    # First off, we need to know what kind of image data we're going to write.
    # There are three possible values:
    #   0 - Plain text
    #   1 - Compressed LZW
    #   2 - Uncompressed LZW
    # Let's pick one of those randomly.
    imgtype = rng.read(1)[0] % 3

    if imgtype == 0:
        # Plain text images are encoded by simply writing the raw bytes from our
        # /dev/urandom source directly into the output stream.
        out.write(rng.read(int.from_bytes(datalen, byteorder='big')))
    elif imgtype == 1:
        # Compressed LZW images start with a single byte saying which code size
        # to use. We'll just choose something random between 2 and 12 bits.
        out.write(rng.read(1))

        # Now comes the compressed data itself. Each byte represents a sequence
        # of codes. To decode them, you read the next byte, find its value in
        # the dictionary, and then look up the corresponding string in the
        # dictionary again. If the string ends in a null character (\x00), stop;
        # otherwise, add the decoded string to your result buffer. Repeat until
        # you run out of input.
        while True:
            c = rng.read(1)[0]

            if c < 0x80:
                out.write(c.to_bytes(1, 'little'))

                if c == 0x00:
                    break
            else:
                s = [c & 0x7f]
                i = 0
                while True:
                    c = rng.read(1)[0]

                    if c >= 0x80:
                        s[i] |= ((c ^ 0x80) << i*7)
                        i += 1

                        if c <= 0xfe:
                            continue
                        else:
                            break
                    else:
                        s[i] |= (c << i*7)
                        i += 1
                        break

                out.write(s)
    elif imgtype == 2:
        # Un