#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (default is /dev/urandom).
# If `seed` is not None, it will be used to seed the RNG.
def generate_random_gif(
        rng: BinaryIO=None, out: BinaryIO=sys.stdout, seed: bytes=None):
    if rng is None:
        if seed is not None:
            rng = open('/dev/urandom', 'rb')
            rng.seek(0)
            rng.write(seed)

        else:
            rng = open('/dev/urandom', 'rb')

    # Read and write the GIF header
    header = b'GIF89a'
    out.write(header)

    # Read the logical screen descriptor
    width = int.from_bytes(rng.read(2), byteorder='little')
    height = int.from_bytes(rng.read(2), byteorder='little')
    flags = int.from_bytes(rng.read(1), byteorder='little')
    bgcolor = int.from_bytes(rng.read(1), byteorder='little')
    aspectratio = int.from_bytes(rng.read(1), byteorder='little')
    out.write(struct.pack('<HHBBB', width, height, flags, bgcolor, aspectratio))

    # Read the global color table
    gctsize = 2 ** ((flags & 7) + 1)
    for _ in range(gctsize):
        r = int.from_bytes(rng.read(1), byteorder='little')
        g = int.from_bytes(rng.read(1), byteorder='little')
        b = int.from_bytes(rng.read(1), byteorder='little')
        out.write(struct.pack('BBB', r, g, b))

    # Read the image descriptor
    left = int.from_bytes(rng.read(2), byteorder='little')
    top = int.from_bytes(rng.read(2), byteorder='little')
    width = int.from_bytes(rng.read(2), byteorder='little')
    height = int.from_bytes(rng.read(2), byteorder='little')
    flags = int.from_bytes(rng.read(1), byteorder='little')
    lctsize = 2 ** ((flags & 7) + 1)
    interlace = bool((flags >> 6) & 1)
    sortflag = bool((flags >> 7) & 1)
    out.write(struct.pack('<HHHHBB', left, top, width, height, flags, lctsize))

    # Read the local color table
    for _ in range(lctsize):
        r = int.from_bytes(rng.read(1), byteorder='little')
        g = int.from_bytes(rng.read(1), byteorder='little')
        b = int.from_bytes(rng.read(1), byteorder='little')
        out.write(struct.pack('BBB', r, g, b))

    # Read the LZW minimum code size
    mincodesize = int.from_bytes(rng.read(1), byteorder='little')
    out.write(mincodesize.to_bytes(1, byteorder='little'))

    # Read the compressed image data
    while True:
        blocktype = int.from_bytes(rng.read(1), byteorder='little')
        if blocktype == 0x00:
            break

        elif blocktype == 0x01:
            raise Exception("No support for GIF extension blocks")

        elif blocktype == 0xfe:
            raise Exception("No support for application extensions")

        elif blocktype == 0xff:
            raise Exception("No support for comments")

        elif blocktype >= 0x02 and blocktype <= 0x0b:
            length = int.from_bytes(rng.read(1), byteorder='little')

            if length!= 0:
                data = rng.read(length)

                # Write the uncompressed subblock
                out.write(struct.pack('B', blocktype))
                out.write(data)

        elif blocktype >= 0x0c and blocktype <= 0x0f:
            length = int.from_bytes(rng.read(1), byteorder='little')

            if length!= 0:
                data = rng.read(length)

                # Write the compressed subblock
                out.write(struct.pack('BB', blocktype,