#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF will have a width of `width`, and height of `height`.
# If either dimension is not specified (or if both are None), then it will be chosen randomly between [256, 4096].
# If `loopcount` is set to an integer greater than zero, then that many iterations of the animation will be performed.
# Otherwise, the loop count will be chosen randomly between [0, 256] inclusive.
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO,
        width=None,
        height=None,
        loopcount=None):

    def choose_random(lo, hi):
        return lo + rng.read(1)[0] % (hi - lo + 1)

    # Header
    out.write(b"GIF89a")

    # Logical screen descriptor
    w = choose_random(256, 4096) if width is None else width
    h = choose_random(256, 4096) if height is None else height
    flags = bytes([0])
    bgcolorindex = bytes([0])
    pixelaspectratio = bytes([0])
    out.write(struct.pack("<HHBBB", w, h, flags[0], bgcolorindex[0], pixelaspectratio[0]))

    # Global color table
    globalcolortablesize = choose_random(0, 7)
    globalcolortable = []
    for i in range(globalcolortablesize):
        r = rng.read(1)[0] & 0xff
        g = rng.read(1)[0] & 0xff
        b = rng.read(1)[0] & 0xff
        globalcolortable += [r, g, b]
    out.write(bytes(globalcolortable))

    # Image descriptor
    left = bytes([0])
    top = bytes([0])
    imagewidth = choose_random(w // 2, w * 2)
    imageheight = choose_random(h // 2, h * 2)
    flags = bytes([0b10000000 | ((choose_random(0, 3) << 2) | choose_random(0, 3))])
    localcolortablesize = choose_random(0, 7)
    interlacedflag = bytes([0])
    sortflag = bytes([0])
    reserved = bytes([0, 0])
    lct = bytes([])
    for i in range(localcolortablesize):
        r = rng.read(1)[0] & 0xff
        g = rng.read(1)[0] & 0xff
        b = rng.read(1)[0] & 0xff
        lct += bytes([r, g, b])
    out.write(left + top + imagewidth.to_bytes(2, "little"))
    out.write(imageheight.to_bytes(2, "little"))
    out.write(flags + localcolortablesize.to_bytes(1, "little"))
    out.write(interlacedflag + sortflag + reserved)
    out.write(lct)

    # Local color table
    lzwmincodesize = choose_random(2, 12)
    out.write(bytes([lzwmincodesize]))

    # LZW code size
    lzwsizes = {
        2:  3,
        3:  3,
        4:  3,
        5:  3,
        6:  3,
        7:  3,
        8:  3,
        9:  3,
        10: 3,
        11: 3,
        12: 3}
    lzwcodesize = lzwmincodesize + lzwsizes[lzwmincodesize]
    out.write(bytes([lzwcodesize]))

    # Data sub-block
    lzwdata = []
    while True:
        blocktype = rng.read(1)[0] & 0b1111
        if blocktype == 0:
            break

        elif blocktype == 1:
            # Plain text
            length = rng.read(1)[0]
            lzwdata += list(map(lambda x: x & 0xff, rng.read(length)))

        elif blocktype == 2:
            # Repeat previous character `length` times
            length = rng.read(1)[0]
            char = rng.read(