#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The first 6 bytes are always "GIF87a" or "GIF89a", depending on whether
    # we want to support an older version (which is not supported by most
    # browsers anymore) or the newer one.
    out.write(b'GIF89a')

    # Next comes the logical screen descriptor. It contains information about
    # how many colors there will be in the final image as well as its size. We
    # just use some arbitrary values here that should work for all images.
    width = 1024
    height = 512
    color_depth = 8
    global_color_table_size = 2 << color_depth - 1
    background_color = 0x00
    aspect_ratio = 0x00
    out.write(struct.pack('<HHBBB', width, height, global_color_table_size,
                          background_color, aspect_ratio))

    # Now comes the local color table. This is optional but it's recommended
    # because it makes things faster when displaying the image. Since we don't
    # have any special requirements, let's just put a single entry with index 0
    # which means transparent black.
    out.write(b'\x00\x00\x00')

    # Finally come the image descriptor. It tells us where the image starts,
    # what its dimensions are and if it uses a local color table or not.
    left = top = 0
    width -= 1
    height -= 1
    local_color_table_flag = False
    interlace_flag = False
    sort_flag = False
    reserved = 0
    lct_size = 0
    out.write(struct.pack('<HHHHBB', left, top, width, height,
                          local_color_table_flag, interlace_flag,
                          sort_flag, reserved))

    # If this flag was set, then we need to write the local color table size
    # next.
    if local_color_table_flag:
        out.write((lct_size).to_bytes(1, 'little'))

    # Now comes the actual image data. First off, we need to know how long our
    # data is going to be so we can write it at the beginning of the frame.
    # Let's assume each pixel takes up two bytes.
    datalen = width * height * 2
    out.write(datalen.to_bytes(2, 'little'))

    # Then we need to write the pixels themselves. For now, we'll just write
    # random numbers from /dev/urandom.
    out.write(rng.read(datalen))