#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # The first 6 bytes are always "GIF87a" or "GIF89a", depending on whether
    # we're generating an old-style (version 87a) or new-style (version 89a)
    # GIF file. We'll just use version 89a for now.
    out.write(b'GIF89a')
    # Next comes the logical screen descriptor. This is where we specify how many
    # colors there will be in our palette as well as the size of each pixel.
    # Since this is a simple example program, let's just make it so that every
    # pixel is one byte long.
    width = 1024
    height = 512
    depth = 8
    global_color_table_flag = True
    color_resolution = 7
    sort_flag = False
    global_color_table_size = pow(2, depth) - 1
    background_color_index = 0
    pixel_aspect_ratio = 0
    out.write(struct.pack('>HHBBB', width, height, depth,
                          global_color_table_flag, color_resolution))
    out.write(struct.pack('>?BB', sort_flag, global_color_table_size))
    out.write(background_color_index.to_bytes(1, 'little'))
    out.write(pixel_aspect_ratio.to_bytes(1, 'little'))
    # Now comes the local color table. If we have a global color table flag set,
    # then this section will contain exactly 2^(depth+1) entries. Otherwise, if
    # the global color table flag is not set, then this section contains only
    # 2^depth entries. In either case, these entries are all black.
    for i in range(pow(2, depth)):
        out.write((i << 3 | i >> 2 & 7).to_bytes(1, 'little'))
    # Finally, we come to the image data. First up is the graphic control
    # extension which tells us what to do with the next frame after this one.
    # For simplicity, we'll just say that we want to keep the current image
    # until we see another graphic control extension telling us otherwise.
    out.write(b'\x21\xF9\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF\x00')
    # Then comes the image descriptor. Here we tell the decoder about the location
    # of the image data relative to the beginning of the file and its dimensions.
    left = 0
    top = 0
    width = 1024
    height = 512
    packed = ((left & 0b11111111) << 2) \
             | (((top & 0b11111111) << 2) >> 8)
    out.write(packed.to_bytes(1, 'little'))
    out.write(((width & 0b11111111) << 2) \
             .to_bytes(1, 'little'))
    out.write((((height & 0b11111111) << 2) >> 8) \
             .to_bytes(1, 'little'))
    packed = (1 << 7) \
             | (0 << 6) \
             | (0 << 5) \
             | (0 << 4) \
             | (0 << 3) \
             | (0 << 2) \
             | (0 << 1) \
             | (0 << 0)
    out.write(packed.to_bytes(1, 'little'))
    # And finally comes the image data itself. It consists of the actual pixels
    # themselves, but since they're all zeroes, we can skip them entirely.
    # Instead, we'll just write some arbitrary data here.
    for _ in range(width * height):
        out.write(b'\x00')