#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom or /dev/urandom).
# The size of the generated image is determined by `width`, and
# its color depth by `colordepth`.
def generate_random_gif(
        width: int, height: int, colordepth: int,
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    out.write(struct.pack('<HHBBB', width, height, 0xf7, 0x00, colordepth))

    # Global color table
    for i in range(2**(1 << colordepth)):
        out.write(struct.pack('B', i))

    # Image descriptor
    out.write(struct.pack('<HHHHBB', 0, 0, width - 1, height - 1, 0, 0))

    # LZW minimum code size
    out.write(b'\x08')

    # Pixels
    lzwmincodesize = 8
    pixelcount = width * height
    while True:
        if pixelcount == 0: break

        # Get next byte from RNG
        c = ord(rng.read(1)) & 0xff

        # If it's not a clear code nor an end-of-stream marker...
        if c!= 0 and c!= 256:

            #...then write it to output stream as-is
            out.write(c.to_bytes(1, 'little'))

            # Decrease remaining pixels counter
            pixelcount -= 1

            continue

        # Otherwise, we have either a clear code or an end-of-stream marker
        # We need to find the longest prefix that exists in our dictionary
        # and write it to output stream instead.

        # Initialize current length with the value of the last written code
        clen = lzwmincodesize

        # Try to find the longest prefix that exists in our dictionary
        while True:
            # Read next bytes from input stream
            c1 = ord(rng.read(1)) & 0xff
            c2 = ord(rng.read(1)) & 0xff

            # Check if these two codes are already present in our dictionary
            key = ((c1 << 8) | c2) % (1 << lzwmincodesize)
            if key in dict:
                # If so, update current length with this entry's length
                clen += len(dict[key])

                # Write this entry to output stream
                out.write((len(dict[key])).to_bytes(1, 'little'))
                out.write(dict[key])

                # Decrease remaining pixels counter
                pixelcount -= len(dict[key])

                # And stop looking for longer prefixes
                break

            else:
                # Otherwise, add them to our dictionary
                dict[(c1 << 8) | c2] = [
                    c1.to_bytes(1, 'little'),
                    c2.to_bytes(1, 'little')]

                # Update current length with the new entry's length
                clen += 2

        # Add the new entry to our dictionary
        dict[clen] = [
            c1.to_bytes(1, 'little'),
            c2.to_bytes(1, 'little')]

        # If we reached maximum code size, increase it
        if clen >= (1 << lzwmincodesize):
            lzwmincodesize += 1

        # Finally, check if there are still some pixels left
        if pixelcount > 0:
            # If so, write a clear code to output stream
            out.write(b'\x00')

            # And decrease remaining pixels counter
            pixelcount -= 1

    # End of stream
    out.write(b'\x00\x3b')



if __name__ == '__main__':
    # Open /dev/urandom if available; otherwise use os.urandom()
    try:
        rng = open('/dev/urandom', mode='rb')
    except FileNotFoundError:
        print("Couldn't open /dev/urandom! Falling back on os.urandom().",
              file=sys.stderr)
        rng = os.urandom

    # Generate a random GIF file
    generate_random_gif(
        400, 400, 8,
        rng, sys.stdout.buffer)