#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The first 6 bytes are always "GIF87a" or "GIF89a", depending on whether
    # we want to use the old (version 87) or new (version 89) format. We'll
    # just write that one in here.
    out.write(b'GIF89a')

    # Next comes the logical screen descriptor. This is basically metadata about
    # how the image will be displayed. It's not actually used by the decoder, so
    # it doesn't matter what we put here.
    #
    #   - Width: 128 pixels
    #   - Height: 128 pixels
    #   - Global color table flag: False (no global color table)
    #   - Color resolution: 0 bits per pixel
    #   - Sorting method: No sorting
    #   - Size of global color table: 0 colors
    #   - Background color index: 0
    #   - Pixel aspect ratio: 0
    out.write(struct.pack('>HHBBB',
        128, 128, 0x80, 0, 0, 0))

    # Then comes the local color table. This contains all the colors that can be
    # used for drawing this frame. Since we're generating a completely random
    # image, there won't be any colors here yet.
    out.write(b'\x00\x00\x00')

    # Finally, we get to the actual image data. This is where things start getting
    # interesting!
    #
    # Each frame consists of two parts: A control block and an image data block.
    # The control block tells us information about the current frame, while the
    # image data block contains the actual bitmap.
    #
    # First up, let's write the control block.
    #
    #   - Image separator: Start of image
    #   - Left position: 0 pixels from left edge of screen
    #   - Top position: 0 pixels from top edge of screen
    #   - Width: 128 pixels wide
    #   - Height: 128 pixels high
    #   - Local color table flag: True (there is a local color table)
    #   - Interlace flag: False (no interlacing)
    #   - Sorting flag: False (not sorted)
    #   - Reserved flags: 0
    #   - LZW minimum code size: 2 (minimum code size for LZW compression)
    #   - Block size: 0 (block size for LZW compression)
    #   - Disposal method: 0 (no disposal specified)
    #   - User input flag: False (user cannot interact with the image)
    #   - Transparent color index: 0 (transparent color is not set)
    #   - Delay time: 0 milliseconds (delay between frames)
    #   - Transparency flag: False (transparency is not enabled)
    #   - Number of loops: 0 (infinite looping)
    out.write(struct.pack('>BHHHHBBBBBH',
        0x2C, 0, 0, 128, 128, 0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0))

    # Now we need to write the image data. To do this, we have to compress our
    # raw data using LZW compression.
    #
    # For more details on LZW compression, see https://en.wikipedia.org/wiki/LZW
    #
    # In short, LZW works by storing a dictionary of previously seen strings. If
    # you encounter a string that has already been stored before, instead of
    # storing it again, you simply store its reference ID.
    #
    # Let's take a look at an example. Suppose we've encountered these three
    # strings before:
    #
    #   1. "Hello"
    #   2. "World!"
    #   3. "Hello World!"
    #
    # If we now encounter the string "Hello World!", we don't have to store it
    # again; we can just reuse the entry for "Hello World!" which was created when
    # we saw "Hello".
    #
    # So if we were to encode "Hello World!" as follows:
    #
    #   1. "H" => 1