#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The size of the output is determined by the first 10 bytes read from `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Read the header and determine the length of the image data
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big') + 1

    # Write the header to the output
    out.write(header)

    # Generate the palette for this frame
    palette = [struct.pack('<BBB', *rng.read(3)) for _ in range(256)]

    # Write the global color table
    out.write(b'\x00\x00')
    out.write(b''.join(palette))

    # Write the logical screen descriptor
    out.write(struct.pack('<HHHHBB',
            1024, 768, 0xF0 | (datalen >> 10 & 0x0F),
            0xF0 | (datalen >> 2 & 0x0F), 0x00, 0x00))

    # Write the local color table
    out.write(b'\xFF\x00')
    out.write(b'\x00\xFF')
    out.write(b'\x00\x00')

    # Write the graphics control extension block
    out.write(b'\x21\xF9\x04\x00\x00\x00\x00\x00\x00\x00')

    # Write the application extension block
    out.write(b'\x21\xFF\x0BNETSCAPE2.0')
    out.write(b'\x03\x01\x00\x00')

    # Write the image descriptor
    out.write(struct.pack('<HHHHBB',
            0, 0, 1024, 768, 0x00, 0x00))

    # Write the image data in chunks of 256 bytes at a time
    while datalen > 0:
        chunklen = min(256, datalen)
        chunk = rng.read(chunklen)

        if len(chunk)!= chunklen:
            raise RuntimeError('Unexpected end of input')

        out.write(chunk)
        datalen -= chunklen