#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (os.urandom(1)[0] % 64) + 256
    height = (os.urandom(1)[0] % 64) + 256
    flags  = bytes([
        0x70 | ((os.urandom(1)[0] & 0x07) << 3),
        os.urandom(1)[0],
        0x80])
    bgcolorindex = os.urandom(1)[0]
    pixelaspectratio = os.urandom(1)[0]
    out.write(struct.pack('>HHBBB',
        width, height, flags[0], flags[1], flags[2]))
    out.write(bgcolorindex.to_bytes(1, 'little'))
    out.write(pixelaspectratio.to_bytes(1, 'little'))

    # Global color table
    globalcolortablesize = os.urandom(1)[0] & 0x07
    if globalcolortablesize > 0:
        for i in range(globalcolortablesize):
            r = os.urandom(1)[0]
            g = os.urandom(1)[0]
            b = os.urandom(1)[0]
            out.write(r.to_bytes(1, 'little'))
            out.write(g.to_bytes(1, 'little'))
            out.write(b.to_bytes(1, 'little'))

    # Image descriptor
    left   = os.urandom(2)
    top    = os.urandom(2)
    width  = (os.urandom(1)[0] % 64) + 256
    height = (os.urandom(1)[0] % 64) + 256
    flags  = bytes([
        0x70 | ((os.urandom(1)[0] & 0x07) << 3),
        os.urandom(1)[0]])
    bgcolorindex = os.urandom(1)[0]
    pixelaspectratio = os.urandom(1)[0]
    out.write(left + top)
    out.write(width.to_bytes(1, 'little'))
    out.write(height.to_bytes(1, 'little'))
    out.write(flags[0].to_bytes(1, 'little'))
    out.write(flags[1].to_bytes(1, 'little'))
    out.write(bgcolorindex.to_bytes(1, 'little'))
    out.write(pixelaspectratio.to_bytes(1, 'little'))

    # Local color table
    localcolortablesize = os.urandom(1)[0] & 0x07
    if localcolortablesize > 0:
        for i in range(localcolortablesize):
            r = os.urandom(1)[0]
            g = os.urandom(1)[0]
            b = os.urandom(1)[0]
            out.write(r.to_bytes(1, 'little'))
            out.write(g.to_bytes(1, 'little'))
            out.write(b.to_bytes(1, 'little'))

    # Image data
    lzwmincodesize = os.urandom(1)[0] & 0x0F
    out.write(lzwmincodesize.to_bytes(1, 'little'))
    imagedata = []
    while True:
        chunklen = os.urandom(1)[0]
        if not chunklen: break
        chunktype = os.urandom(1)[0]
        if chunktype == 0x2C: continue

        # LZW min code size
        elif chunktype == 0xF9:
            lzwmincodesize = os.urandom(1)[0] & 0x0F
            out.write(chunktype.to_bytes(1, 'little'))
            out.write(lzwmincodesize.to_bytes(1, 'little'))

        else:
            print("Unknown chunk type", hex(chunktype))
            exit(-1)

        imagedata += [chunklen, chunktype] * chunklen

    out.write(bytes(