#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Generate the header (always 6 bytes)
    header = b'GIF89a'
    out.write(header)

    # Generate the logical screen descriptor (always 7 bytes)
    width = int.from_bytes(
            rng.read(2), byteorder='little', signed=False) % 1000
    height = int.from_bytes(
            rng.read(2), byteorder='little', signed=False) % 1000
    flags = int.from_bytes(
            rng.read(1), byteorder='little', signed=False) & 0b11111110
    global_color_table_flag = flags >> 7
    color_resolution = ((flags << 1) & 0b11000000) | \
                       (((flags << 1) & 0b11000000) >> 4)
    sort_flag = ((flags << 5) & 0b10000000) >> 7
    size_of_global_color_table = ((flags << 5) & 0b00111000) | \
                                  (((flags << 5) & 0b00111000) >> 2)
    background_color_index = int.from_bytes(
            rng.read(1), byteorder='little', signed=False)
    pixel_aspect_ratio = int.from_bytes(
            rng.read(1), byteorder='little', signed=False)
    logical_screen_descriptor = struct.pack('<HHBBB',
                                            width,
                                            height,
                                            global_color_table_flag,
                                            color_resolution,
                                            sort_flag,
                                            size_of_global_color_table)
    logical_screen_descriptor += background_color_index.to_bytes(
            1, byteorder='little', signed=False)
    logical_screen_descriptor += pixel_aspect_ratio.to_bytes(
            1, byteorder='little', signed=False)
    out.write(logical_screen_descriptor)

    # Generate the global color table if necessary
    if global_color_table_flag == 1:
        for i in range(size_of_global_color_table):
            r = int.from_bytes(
                    rng.read(1), byteorder='little', signed=False)
            g = int.from_bytes(
                    rng.read(1), byteorder='little', signed=False)
            b = int.from_bytes(
                    rng.read(1), byteorder='little', signed=False)
            out.write((r,g,b))

    # Generate the image descriptor (always 10 bytes)
    left = int.from_bytes(
            rng.read(2), byteorder='little', signed=False) % width
    top = int.from_bytes(
            rng.read(2), byteorder='little', signed=False) % height
    width = int.from_bytes(
            rng.read(2), byteorder='little', signed=False) % width
    height = int.from_bytes(
            rng.read(2), byteorder='little', signed=False) % height
    flags = int.from_bytes(
            rng.read(1), byteorder='little', signed=False) & 0b11111110
    local_color_table_flag = flags >> 7
    interlace_flag = ((flags << 1) & 0b10000000) >> 7
    sort_flag = ((flags << 5) & 0b10000000) >> 7
    reserved = ((flags << 5) & 0b01110000) >> 4
    lzw_minimum_code_size = flags & 0b1111
    image_descriptor = struct.pack('<HHHHBB',
                                   left,
                                   top,
                                   width,
                                   height,
                                   local_color_table_flag,
                                   interlace_flag,
                                   sort_flag,
                                   reserved,
                                   lzw_minimum_code_size)
    out.write(image_descriptor)

    # Generate the local color table if necessary
    if local_color_table_flag == 1:
        for i in range(size_