#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The size of the generated image is determined by
# the first 24 bits of the random number stream.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Generate the header and write it to the output
    header = b'GIF89a'
    out.write(header)

    # Read in the width and height from the RNG
    width = int.from_bytes(
            rng.read(2), byteorder='little', signed=False)
    height = int.from_bytes(
            rng.read(2), byteorder='little', signed=False)

    # Write these dimensions to the output
    out.write(width.to_bytes(2, 'little'))
    out.write(height.to_bytes(2, 'little'))

    # Determine how many bytes are needed for each pixel
    color_depth = int.from_bytes(
            rng.read(1), byteorder='little', signed=False)
    if (color_depth == 0):
        raise ValueError('Color depth must be nonzero.')
    elif (color_depth > 8):
        raise NotImplementedError()

    # Write this value to the output
    out.write(color_depth.to_bytes(1, 'little'))

    # Write the rest of the header to the output
    out.write(b'\x87\x00\x00')

    # Calculate the length of the global color table
    gctlen = pow(2, min(int(color_depth), 7)) * 3
    # Write that value to the output
    out.write((gctlen - 1).to_bytes(1, 'little'))

    # Write the background color index to the output
    bgindex = int.from_bytes(
            rng.read(1), byteorder='little', signed=False)
    out.write(bgindex.to_bytes(1, 'little'))

    # Write the aspect ratio to the output
    out.write(b'\x00')

    # Write the palette entries to the output
    for i in range(pow(2, min(int(color_depth), 7))):
        r = int.from_bytes(
                rng.read(1), byteorder='little', signed=False)
        g = int.from_bytes(
                rng.read(1), byteorder='little', signed=False)
        b = int.from_bytes(
                rng.read(1), byteorder='little', signed=False)

        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Write the local color table to the output
    lctsize = int.from_bytes(
            rng.read(1), byteorder='little', signed=False)
    for i in range(lctsize):
        r = int.from_bytes(
                rng.read(1), byteorder='little', signed=False)
        g = int.from_bytes(
                rng.read(1), byteorder='little', signed=False)
        b = int.from_bytes(
                rng.read(1), byteorder='little', signed=False)

        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Write the image descriptor to the output
    out.write(width.to_bytes(2, 'little'))
    out.write(height.to_bytes(2, 'little'))
    out.write(b'\x00\x00')
    out.write(b'\x00')

    # Write the LZW minimum code size to the output
    out.write(min(int(color_depth), 8).to_bytes(1, 'little'))

    # Create an array of all possible codes
    codes = [i for i in range(256)]

    # Initialize the current code to the minimum code size
    curcode = min(int(color_depth), 8)

    # Initialize the next code to the maximum code size
    nxtcode = max(int(color_depth), 8)

    # Initialize the clear code to 2^(current code size)
    clrcd = pow(2, curcode)

    # Initialize the end