#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (1 << rng.read(1)[0]) % 640
    height = (1 << rng.read(1)[0]) % 480
    flags  = bytes([
        0x80 | ((rng.read(1)[0] & 0x70) >> 4),
        rng.read(1)[0] & 0x0f,
    ])
    bgcol  = rng.read(1)[0]
    pxsize = rng.read(1)[0] & 0x07
    colres = rng.read(1)[0] & 0x07
    sort   = rng.read(1)[0] & 0x03
    gctlen = rng.read(1)[0] if colres == 7 else 0
    out.write(struct.pack('>HHBBB', width, height, flags[0], flags[1], bgcol))

    # Global color table
    for _ in range(gctlen):
        out.write(rng.read(1))

    # Image descriptor
    left      = rng.read(2)
    top       = rng.read(2)
    width     = (1 << rng.read(1)[0]) % 640
    height    = (1 << rng.read(1)[0]) % 480
    flags     = bytes([
        0x80 | ((rng.read(1)[0] & 0x70) >> 4),
        rng.read(1)[0] & 0x0f,
    ])
    lctlen    = rng.read(1)[0] if flags[0] & 0x80!= 0 else 0
    interlace = bool((flags[0] & 0x40) >> 6)
    sorted    = bool((flags[0] & 0x20) >> 5)
    reserved  = rng.read(1)[0]
    lctbpp    = rng.read(1)[0] & 0x07
    lctsort   = rng.read(1)[0] & 0x03
    out.write(left + top + bytes([width, height]))
    out.write(flags[0] + flags[1])
    out.write(lctlen)

    # Local color table
    for _ in range(lctlen):
        out.write(rng.read(1))

    # Image data
    while True:
        code = rng.read(1)[0]
        if code == 0x2c:
            break
        elif code == 0x21:
            extcode = rng.read(1)[0]
            if extcode == 0xf9:
                out.write(bytes([extcode, 4, 4, 0, 0]))
            else:
                size = int.from_bytes(rng.read(2), byteorder='big')
                out.write(bytes([extcode, size & 0xff, size >> 8]))
                for _ in range(size - 2):
                    out.write(rng.read(1))
        else:
            print("Unknown block type:", hex(code))

    # Trailer
    out.write(b'\x3B')


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('./test.gif', 'wb') as f:
            generate_random_gif(rng, f)