#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 12 15:46:09 2020

@author: jonasg
"""
import numpy as np
from scipy.stats import norm, multivariate_normal
from scipy.optimize import minimize
import matplotlib.pyplot as plt

def log_likelihood(theta, x):
    """
    Computes the log likelihood of a normal distribution with mean and covariance given by theta.
    
    Parameters
    ----------
    theta : array_like
        Array containing the mean vector (first n elements) and the covariance matrix (last n*(n+1)/2 elements).
        
    x : array_like
        Data to compute the log likelihood for.
        
    Returns
    -------
    float
        Log likelihood of the data under the model.
    """
    # Extract parameters from theta
    mu = theta[:len(x)]
    sigma = np.zeros((len(x), len(x)))
    i = 0
    j = 0
    for k in range(len(x)):
        for l in range(k + 1):
            if k == l:
                sigma[k][l] = theta[i]
                i += 1
            else:
                sigma[k][l] = theta[j]
                sigma[l][k] = theta[j]
                j += 1
                
    # Compute log likelihood
    return np.sum([np.log(multivariate_normal.pdf(x=xi, mean=mu, cov=sigma)) for xi in x])

def log_prior(theta, bounds):
    """
    Computes the log prior of a normal distribution with mean zero and covariance given by theta.
    
    Parameters
    ----------
    theta : array_like
        Array containing the mean vector (first n elements) and the covariance matrix (last n*(n+1)/2 elements).
        
    bounds : array_like
        Bounds on the parameters.
        
    Returns
    -------
    float
        Log prior of the parameters.
    """
    # Extract parameters from theta
    mu = theta[:len(bounds)]
    sigma = np.zeros((len(bounds), len(bounds)))
    i = 0
    j = 0
    for k in range(len(bounds)):
        for l in range(k + 1):
            if k == l:
                sigma[k][l] = theta[i]
                i += 1
            else:
                sigma[k][l] = theta[j]
                sigma[l][k] = theta[j]
                j += 1
            
    # Check that parameters are within bounds
    if not all([mu[i] >= bounds[i][0] and mu[i] <= bounds[i][1] for i in range(len(bounds))]):
        return -np.inf
    elif not all([sigma[i][i] > 0 for i in range(len(bounds))]):
        return -np.inf
    elif not all([sigma[i][j] == 0 for i in range(len(bounds)) for j in range(i + 1, len(bounds)) if i != j]):
        return -np.inf
    else:
        return 0

def log_posterior(theta, x, bounds):
    """
    Computes the log posterior of a normal distribution with mean and covariance given by theta.
    
    Parameters
    ----------
    theta : array_like
        Array containing the mean vector (first n elements) and the covariance matrix (last n*(n+1)/2 elements).
        
    x : array_like
        Data to compute the log likelihood for.
        
    bounds : array_like
        Bounds on the parameters.
        
    Returns
    -------
    float
        Log posterior of the data under the model.
    """
    # Compute log likelihood
    llh = log_likelihood(theta, x)
    
    # Compute log prior
    lpr = log_prior(theta, bounds)
    
    # Return