#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 17 20:46:59 2019

@author: dberke

Script to generate a table of the average number of days between each pair of
planets in the Solar System.
"""

import argparse as ap
from collections import defaultdict
import csv
import datetime as dt
import itertools as it
import os
import re
import sys

import numpy as np
import pandas as pd

import varconlib as vcl


def main(args):
    """
    Run the program.

    Parameters
    ----------
    args : argparse namespace
        Command line arguments.

    Returns
    -------
    None.

    """

    # Get the data from the input file.
    planets = get_data()

    # Generate the table.
    table = generate_table(planets)

    # Write the table to the output file.
    with open(args.output, 'w') as f:
        writer = csv.writer(f)
        writer.writerows([['Planet', 'Average days']] + list(table))


def generate_table(planets):
    """
    Generate the table of average days between each planet pair.

    Parameters
    ----------
    planets : dict of strs to lists of floats
        The orbital elements for each planet.

    Returns
    -------
    table : list of lists of strings
        The table of average days between each planet pair.

    """

    # Make a list of all the pairs of planets.
    pairs = [list(p) for p in it.combinations(sorted(planets), 2)]

    # Calculate the average time difference between each pair of planets.
    table = []
    for i, (name1, name2) in enumerate(pairs):
        diffs = [(planets[name2][0] - planets[name1][0]).days for _ in range(3)]
        avg = sum(diffs)/len(diffs)
        table.append(['{}–{}'.format(name1, name2), '{:.2f}'.format(avg)])

    return table


def get_data():
    """
    Read the orbital elements from the input file and return them in a
    dictionary.

    Returns
    -------
    planets : dict of strs to lists of floats
        The orbital elements for each planet.

    """

    # Read the input file.
    filename = os.path.join('..', 'data', 'orbital_elements.csv')
    df = pd.read_csv(filename)

    # Extract the orbital elements for each planet.
    planets = {}
    for row in df.itertuples():
        name = row.Name
        if name not in planets:
            planets[name] = []
        planets[name].append(dt.datetime(*row.OrbitalElements))

    return planets

