#!/usr/bin/env python3

import os
import sys
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from collections import defaultdict
from datetime import date
from itertools import chain
from operator import itemgetter

from pysam import AlignmentFile

# This script is used to generate a table of the number of reads mapped to each
# chromosome. It takes as input the output from samtools view -c and outputs
# a table with one row per chromosome. The first column contains the name of the
# chromosome (e.g., chr1) and the second column contains the number of reads
# mapped to that chromosome.

def main(args):
    # Parse command line arguments
    parser = ArgumentParser(description=__doc__, formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('-i', '--input', dest='infile', required=True, help="Input file")
    parser.add_argument('-o', '--output', dest='outfile', required=True, help="Output file")
    args = parser.parse_args()
    
    # Read in counts for each chromosome
    counts = read_counts(args.infile)
    
    # Write out results
    write_results(args.outfile, counts)

def read_counts(filename):
    """Read in counts for each chromosome"""
    counts = {}
    with open(filename) as f:
        for line in f:
            fields = line.strip().split('\t')
            if len(fields) != 2:
                raise Exception("Expected two columns in %s but found %d" % (filename, len(fields)))
            chrom = fields[0]
            count = int(fields[1])
            
            # Add this count to the total for this chromosome
            try:
                counts[chrom] += count
            except KeyError:
                counts[chrom] = count
                
    return counts

def write_results(filename, counts):
    """Write out results"""
    with open(filename, 'w') as f:
        # Sort by chromosome name
        sorted_counts = sorted(counts.items(), key=itemgetter(0))
        
        # Print header
        print("\t".join(['Chromosome', 'Count']), file=f)
        
        # Print counts
        for chrom, count in sorted_counts:
            print("%s\t%d" % (chrom, count), file=f)
