#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a' + 7 * b'\0' # Header
    out.write(header)

    # Logical screen descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed = rng.read(1)[0]
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    lsd = struct.pack('<HHBBB', *struct.unpack('>HHLBBB', width+height+packed+background_color_index+pixel_aspect_ratio))
    out.write(lsd)

    # Global color table
    global_color_table_flag = rng.read(1)[0] & 0b10000000
    if global_color_table_flag == 0:
        raise Exception("No global color table")
    global_color_table_size = (global_color_table_flag >> 4) + 1
    global_color_table = [rng.read(3) for _ in range(global_color_table_size)]
    out.write(bytes().join(global_color_table))

    # Application extension
    block_type = b'\xff\xfe'
    application_id = b'NETSCAPE2.0'
    loop_count = b'\x00\x01'
    block_terminator = b';'
    app_ext = bytes([len(block_type)+len(application_id)+len(loop_count)+len(block_terminator)]) + \
              block_type + \
              application_id + \
              loop_count + \
              block_terminator
    out.write(app_ext)

    while True:
        # Image descriptor
        image_descriptor = b'\x2c' + rng.read(9)
        out.write(image_descriptor)

        # Local color table
        local_color_table_flag = rng.read(1)[0] & 0b10000000
        if local_color_table_flag == 0:
            raise Exception("No local color table")
        local_color_table_size = (local_color_table_flag >> 4) + 1
        local_color_table = [rng.read(3) for _ in range(local_color_table_size)]
        out.write(bytes().join(local_color_table))

        # Image data
        image_data = []
        while True:
            code_size = rng.read(1)[0]
            if code_size == 0:
                break
            image_data.append(code_size)
            image_data += rng.read(2**code_size)
        out.write(bytes(image_data))

        # Trailer
        trailer = b'\x3b'
        out.write(trailer)
