#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Header
    header = b'GIF89a' + \
             bytes([
                 rng.read(1)[0],  # width
                 rng.read(1)[0]   # height
            ])

    # Logical screen descriptor
    lsd = [
        0x2c,    # image separator
        0x00,    # left position
        0x00,    # top position
        0x00,    # width
        0x00,    # height
        0xf7,    # flags (global color table)
        0x00     # background color index
    ]

    # Global color table
    gct = []
    for i in range(0, 256):
        r = rng.read(1)[0] & 0b11111000 | rng.read(1)[0] >> 4
        g = rng.read(1)[0] & 0b11111100 | rng.read(1)[0] >> 2
        b = rng.read(1)[0] & 0b11111000 | rng.read(1)[0] >> 4

        gct += [r,g,b]

    lsd[4] |= len(gct) - 1

    # Image data
    idat = bytearray()

    # Frames
    frames = rng.read(1)[0] % 10 + 1
    for f in range(frames):
        # Frame duration
        fd = [
            rng.read(1)[0],  # frame delay low
            rng.read(1)[0]   # frame delay high
        ]

        # LZW minimum code size
        lzw = rng.read(1)[0] % 13 + 2

        # Image descriptor
        idd = [
            rng.read(1)[0],  # left position
            rng.read(1)[0],  # top position
            rng.read(1)[0],  # width
            rng.read(1)[0],  # height
            0x00             # flags
        ]

        # Local color table
        lct = []
        if rng.read(1)[0] > 0:
            for i in range(0, 2**lzw):
                r = rng.read(1)[0] & 0b11111000 | rng.read(1)[0] >> 4
                g = rng.read(1)[0] & 0b11111100 | rng.read(1)[0] >> 2
                b = rng.read(1)[0] & 0b11111000 | rng.read(1)[0] >> 4

                lct += [r,g,b]

            idd[4] |= len(lct) - 1

        # Encode pixels with LZW compression
        while True:
            c = rng.read(1)[0]
            if c == 0xff: break

            elif c < 0x80:
                idat += bytes([c])

            else:
                idat += bytes([(c >> 8), c & 0xff])

                if not idat[-1:] in idat[:-1]:
                    idat += bytes([idat[-1]])
                else:
                    idat += bytes([0x00])

                    while idat[-1]!= 0x00:
                        idat += bytes([idat[-2]])

        # Write to output
        out.write(header)
        out.write(bytes(lsd))
        out.write(struct.pack('B'*len(gct), *gct))

        for i in range(f+1):
            out.write(bytes(idd))
            if idd[4] & 0b10000000:
                out.write(struct.pack('B'*len(lct), *lct))
            out.write(fd)
            out.write(bytes([lzw]))
            out.write(idat)

