#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (1 << rng.read(1)[0]) % 640
    height = (1 << rng.read(1)[0]) % 480
    flags  = bytes([
        0x80 | ((rng.read(1)[0] & 0x70) >> 4),
        rng.read(1)[0] & 0x0f,
        0,
        0,
    ])
    bg     = bytes([rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]])
    aspect = bytes([0, 0, rng.read(1)[0], rng.read(1)[0]])
    out.write(struct.pack('<HHBBB', width, height, *flags))
    out.write(bg + aspect)

    # Global color table
    for _ in range(256):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(bytes([r, g, b]))

    # Image descriptor
    left   = 0
    top    = 0
    width  = (1 << rng.read(1)[0]) % 640
    height = (1 << rng.read(1)[0]) % 480
    flags  = bytes([
        0x80 | ((rng.read(1)[0] & 0x70) >> 4),
        rng.read(1)[0] & 0x0f,
    ])
    out.write(struct.pack('<HHHHB', left, top, width, height, *flags))

    # Local color table
    for _ in range((1 << flags[1]) if flags[1] else 2**flags[0]):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(bytes([r, g, b]))

    # Image data
    lzwmincodesize = rng.read(1)[0]
    out.write(bytes([lzwmincodesize]))

    # LZW code stream
    while True:

        # Read next byte and check if it's the end of the image
        c = rng.read(1)[0]
        if c == 0xff:
            break

        # Check if we need to add another bit to our current code
        if not hasattr(generate_random_gif, 'current'):

            # Initialize first code with its length
            generate_random_gif.current = [c]
            generate_random_gif.length  = len(generate_random_gif.current) - 1

        elif generate_random_gif.length < 12:

            # Add new bit to existing code
            generate_random_gif.current.append(c & (1 << generate_random_gif.length))
            generate_random_gif.length += 1

        else:

            # Write code to output
            out.write(bytes([
                generate_random_gif.current[0] |
                    (((generate_random_gif.current[-1] ^
                        generate_random_gif.current[-2]) << 7) & 0x80),
                *(generate_random_gif.current[i+1:]
                  for i in reversed(range(len(generate_random_gif.current)-1))),
            ]))

            # Reset state
            delattr(generate_random_gif, 'current')
            delattr(generate_random_gif, 'length')


    # End of image marker
    out.write(bytes([0x00]))



if