#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # 16-bit width and height
    w = rng.read(2)
    h = rng.read(2)

    # packed field of color resolution (bits per pixel)
    r = rng.read(1)[0] & 7
    g = rng.read(1)[0] & 7
    b = rng.read(1)[0] & 7
    
    # background color index
    bg = rng.read(1)[0]

    # aspect ratio
    aspect = rng.read(1)[0] & 7

    # logical screen descriptor
    lsd = struct.pack('<6sHHBBB', header, int.from_bytes(w, 'big'),
                      int.from_bytes(h, 'big'), r, g, b, bg)

    # global color table size
    ctsize = rng.read(1)[0] & 7

    # sort flag
    sorted = bool(rng.read(1)[0])

    # global color table
    if ctsize == 0:
        # no global color table
        pass
    elif ctsize == 1:
        # 256 colors
        for i in range(256):
            rgb = [rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]]
            lsd += bytes(rgb)
    else:
        raise Exception('Invalid global color table size')

    # graphics control extension
    hasgce = bool(rng.read(1)[0])
    if hasgce:
        # delay time
        dt = rng.read(2)

        # transparent color index
        tc = rng.read(1)[0]

        # block terminator
        bt = rng.read(1)[0]

        gce = struct.pack('<BHB4B', 0x21, 0xF9,
                           int.from_bytes(dt, 'big'),
                           0b10000000 | tc, 0) + bt.to_bytes(1, 'little')

    # image descriptor
    idesc = struct.pack('<BBH', 0x2C, 0,
                        0) + bytes([r, g, b])

    # local color table
    lct = []
    for i in range(int((1 << (r+1)) * (1 << (g+1)) * (1 << (b+1)))):
        rgb = [rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]]
        lct += bytes(rgb)

    # data subblock
    dsub = bytearray()
    while True:
        # data length
        dl = rng.read(1)[0]
        if dl == 0: break
        
        # data subblock
        dsub += rng.read(dl)

    # graphic control extension
    hasgce = bool(rng.read(1)[0])
    if hasgce:
        # delay time
        dt = rng.read(2)

        # transparent color index
        tc = rng.read(1)[0]

        # block terminator
        bt = rng.read(1)[0]

        gce = struct.pack('<BHB4B', 0x21, 0xF9,
                           int.from_bytes(dt, 'big'),
                           0b10000000 | tc, 0) + bt.to_bytes(1, 'little')

    # application extension
    hasae = bool(rng.read(1)[0])
    if hasae:
        # app id
        aid = rng.read(1)[0]

        # app auth code
        ac = rng.read(1)[0]

        # app data
        ad = rng.read(1)[0]

        ae = struct.pack('<BBBBB', 0xFF, 0x0B,
                          aid, ac, ad)

    # comment extension
    hasce = bool(rng.read(1)[0])
    if hasce:
        # comment
        cmt = rng.read(1)[0]

        ce = struct.pack('<BB', 0xFE, 0x01) + cmt

    # plain text extension
    haspte = bool(rng.read(1)[0])
    if haspte:
        # tx region left
        trl = rng.read(1)[0]

        # tx region top
        trt = rng.read(1)[0]

        # tx region right
        trr = rng.read(1)[0]

        # tx region bottom
        trb = rng.read(1)[0]

        # cell width
        cw = rng.read(1)[0]

        # cell height
        ch = rng.read(1)[0]

        # foreground color index
        fc = rng.read(1)[0]

        # background color index
        bc = rng.read(1)[0]

        pte = struct.pack('<BBBBBBB', 0xFF, 0x01,
                          trl, trt, trr, trb, cw, ch) \
             