#!/usr/bin/env python3

import struct
from typing import BinaryIO

def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
) -> None:
    """Generate a random GIF file.

    :param rng: A binary stream containing the RNG state to use for generation.
    :param out: The output file to write the generated data to.
    """
    # Header
    out.write(b'GIF89a')  # Magic number
    width = rng.read(1)[0] & 0x7f + 256 * (rng.read(1)[0] & 0xff)
    height = rng.read(1)[0] & 0x7f + 256 * (rng.read(1)[0] & 0xff)
    flags = rng.read(1)[0]
    bg_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('<HHBBB', width, height, flags, bg_color, aspect_ratio))

    # Logical screen descriptor
    color_table_size = rng.read(1)[0] & 0x7f
    global_color_table = [rng.read(1)[0] for _ in range(pow(2, color_table_size+1))]
    background_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('BBBBBB', color_table_size, *global_color_table,
                          background_index, pixel_aspect_ratio))

    # Image descriptor
    left = rng.read(1)[0] & 0x7f + 256 * (rng.read(1)[0] & 0xff)
    top = rng.read(1)[0] & 0x7f + 256 * (rng.read(1)[0] & 0xff)
    width = rng.read(1)[0] & 0x7f + 256 * (rng.read(1)[0] & 0xff)
    height = rng.read(1)[0] & 0x7f + 256 * (rng.read(1)[0] & 0xff)
    flags = rng.read(1)[0]
    local_color_table_flag = bool((flags >> 7) & 0x1)
    interlace_flag = bool((flags >> 6) & 0x1)
    sort_flag = bool((flags >> 5) & 0x1)
    reserved_bits = ((flags << 2) & 0xc0) | (((flags << 4) & 0xc0) >> 2)
    lct_size = pow(2, (flags & 0x7) + 1) if local_color_table_flag else 0
    lct = [rng.read(1)[0] for _ in range(lct_size)]
    lzw_min_code_size = rng.read(1)[0]
    out.write(struct.pack('<HHHHB', left, top, width, height, flags))
    if local_color_table_flag:
        out.write(bytes(lct))
    out.write(struct.pack('BB', lzw_min_code_size, reserved_bits))

