#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    out.write("GIF89a".encode())

    width = 1 + int.from_bytes(rng.read(1), "big") % 256
    height = 1 + int.from_bytes(rng.read(1), "big") % 256
    flags = b"\x80" if (int.from_bytes(rng.read(1), "big") & 1) == 1 else b"\x00"
    global_color_table_size = 1 << ((int.from_bytes(rng.read(1), "big") & 7) + 1)
    background_index = int.from_bytes(rng.read(1), "big") % \
                       global_color_table_size
    aspect_ratio = int.from_bytes(rng.read(1), "big") % 4

    out.write(struct.pack("<HHBBB",
                          width, height, flags[0],
                          global_color_table_size,
                          background_index))
    out.write(aspect_ratio.to_bytes(1, byteorder="little"))

    for i in range(global_color_table_size):
        r = int.from_bytes(rng.read(1), "big") % 256
        g = int.from_bytes(rng.read(1), "big") % 256
        b = int.from_bytes(rng.read(1), "big") % 256

        out.write(r.to_bytes(1, byteorder="little"))
        out.write(g.to_bytes(1, byteorder="little"))
        out.write(b.to_bytes(1, byteorder="little"))

    # Image descriptor
    left = 0
    top = 0
    width = 1 + int.from_bytes(rng.read(1), "big") % 256
    height = 1 + int.from_bytes(rng.read(1), "big") % 256
    packed = (left |
              (top << 8) |
              (((width - 1) & 0xF) << 12) |
              (((height - 1) & 0xF) << 16) |
              0x80)
    out.write(packed.to_bytes(2, byteorder="little"))

    local_color_table_flag = int.from_bytes(rng.read(1), "big") % 2
    interlace_flag = int.from_bytes(rng.read(1), "big") % 2
    sort_flag = int.from_bytes(rng.read(1), "big") % 2
    reserved_flags = int.from_bytes(rng.read(1), "big") % 2
    size_of_local_color_table = 1 << ((int.from_bytes(rng.read(1), "big") & 7) + 1)

    out.write((local_color_table_flag << 7) |
              (interlace_flag << 6) |
              (sort_flag << 5) |
              (reserved_flags << 3) |
              (size_of_local_color_table))

    for i in range(size_of_local_color_table):
        r = int.from_bytes(rng.read(1), "big") % 256
        g = int.from_bytes(rng.read(1), "big") % 256
        b = int.from_bytes(rng.read(1), "big") % 256

        out.write(r.to_bytes(1, byteorder="little"))
        out.write(g.to_bytes(1, byteorder="little"))
        out.write(b.to_bytes(1, byteorder="little"))

    lzw_min_code_size = int.from_bytes(rng.read(1), "big") % 256
    out.write(lzw_min_code_size.to_bytes(1, byteorder="little"))

    image_data = []
    while True:
        data_block_size = int.from_bytes(rng.read(1), "big") % 256
        if not data_block_size:
            break
        image_data += [data_block_size]
        image_data += list(rng.read(data_block_size))

    compressed_image_data = compress_image_data(image_data)
    out.write(compressed_image_data)