#!/usr/bin/env python3

import struct
from typing import BinaryIO

def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(rng.read(2), 'big')

    # Write the GIF header and screen descriptor
    out.write(header)
    out.write(struct.pack('>HHBBB', 100, 50, 0xF7, 0x00, 0))

    # Write the global color table (palette)
    for _ in range(0, 768):
        out.write(rng.read(1))

    # Write the image data
    while True:
        blocktype = rng.read(1)[0]

        if blocktype == 0x2C:
            # Image descriptor

            left = int.from_bytes(rng.read(2), 'little')
            top = int.from_bytes(rng.read(2), 'little')
            width = int.from_bytes(rng.read(2), 'little')
            height = int.from_bytes(rng.read(2), 'little')

            flags = rng.read(1)[0]
            lctsize = rng.read(1)[0]

            # Skip local color table
            if lctsize!= 0:
                rng.seek(lctsize * 3, 1)

            # Get the LZW minimum code size
            mincodesize = rng.read(1)[0] + 1

            # Read the image data
            imgdata = bytearray()
            while len(imgdata) < width * height:
                chunklen = rng.read(1)[0]

                if chunklen > 0:
                    imgdata += rng.read(chunklen)

                    if len(imgdata) >= width * height:
                        break

                    continue

                elif chunklen == 0:
                    imgdata += [0]

                else:
                    raise Exception("Invalid LZW length")

            # Write the image descriptor
            out.write(b'\x2c')
            out.write(left.to_bytes(2, 'little'))
            out.write(top.to_bytes(2, 'little'))
            out.write(width.to_bytes(2, 'little'))
            out.write(height.to_bytes(2, 'little'))
            out.write(flags.to_bytes(1, 'little'))
            out.write(mincodesize.to_bytes(1, 'little'))

            # Write the image data
            out.write(imgdata)

        elif blocktype == 0x21:
            # Extension block

            exttype = rng.read(1)[0]

            if exttype == 0xf9:
                # Graphic control extension

                # Skip over the rest of this block
                rng.seek(4, 1)

                # Read the next block type
                blocktype = rng.read(1)[0]

                if blocktype!= 0x01:
                    raise Exception("Invalid graphic control block")

                # Skip over the rest of this block
                rng.seek(3, 1)

            elif exttype == 0xff:
                # Application extension

                appid = rng.read(1)[0]
                appauthcode = rng.read(1)[0]

                # Skip over the rest of this block
                rng.seek(int.from_bytes(rng.read(2), 'big'), 1)

                # Read the next block type
                blocktype = rng.read(1)[0]

                if blocktype!= 0x01:
                    raise Exception("Invalid application extension block")

                # Skip over the rest of this block
                rng.seek(int.from_bytes(rng.read(2), 'big'), 1)

            else:
                raise Exception("Unknown extension block")

        elif blocktype == 0x3B:
            # Trailer
            return

        else:
            raise Exception("Unknown block type")

