#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), 'little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), 'little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), 'little')])
    bg     = bytes([int.from_bytes(rng.read(1), 'little')])
    aspect = bytes([int.from_bytes(rng.read(1), 'little')])
    out.write(struct.pack('<HHBBB', width, height, *flags))
    out.write(bg)
    out.write(aspect)

    # Global color table
    for _ in range(0, 256):
        r = int.from_bytes(rng.read(1), 'little') & 0xff
        g = int.from_bytes(rng.read(1), 'little') & 0xff
        b = int.from_bytes(rng.read(1), 'little') & 0xff
        out.write(bytes([r,g,b]))

    # Image descriptor
    left   = 0
    top    = 0
    width  = int.from_bytes(rng.read(2), 'little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), 'little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), 'little')])
    lct    = bool(flags[7])
    interlaced = False if not len(flags) else bool(flags[6])
    sort   = True if not len(flags) else bool(flags[5])
    local_color_table_size = int.from_bytes(rng.read(1), 'little')
    out.write(struct.pack('<HHHHB', left, top, width, height, *flags))

    # Local color table
    for _ in range(0, local_color_table_size):
        r = int.from_bytes(rng.read(1), 'little') & 0xff
        g = int.from_bytes(rng.read(1), 'little') & 0xff
        b = int.from_bytes(rng.read(1), 'little') & 0xff
        out.write(bytes([r,g,b]))

    # Table based image data
    while True:
        block_type = int.from_bytes(rng.read(1), 'little')

        if block_type == 0x2c:
            break

        elif block_type == 0x21:
            while True:
                sub_block_type = int.from_bytes(rng.read(1), 'little')

                if sub_block_type == 0xfe:
                    length = int.from_bytes(rng.read(1), 'little')

                    if length > 0:
                        out.write(rng.read(length))

                elif sub_block_type == 0x01:
                    pass

                else:
                    raise Exception('Unknown extension type {}'.format(sub_block_type))

        elif block_type == 0x3b:
            continue

        else:
            raise Exception('Unknown block type {}'.format(block_type))


    # Plain text data
    while True:
        block_type = int.from_bytes(rng.read(1), 'little')

        if block_type == 0x01 or block_type == 0xff:
            length = int.from_bytes(rng.read(1), 'little')

            if length > 0:
                out.write(rng.read(length))

        elif block_type == 0x00:
            break

        else:
            raise Exception('Unknown block type {}'.format(block_type))

    # Graphics control extension
    out.write(bytes([0x21, 0xf9, 4, 1, 0, 0]))

    # Application extension
    out.write(bytes([0x21, 0xff, 11, 78, 69, 84, 83, 67, 65, 80, 69, 58, 32, 49, 46, 48]))

    # Comment extension
    out.write(bytes([0x21, 0xfe, 0]))

    # Trailer
    out.write(b';')