#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (1 << rng.read(1)[0]) % 640
    height = (1 << rng.read(1)[0]) % 480
    flags  = bytes([
        0x80 | ((rng.read(1)[0] & 0x70) >> 4),
        rng.read(1)[0] & 0x0f,
    ])
    bgcol  = rng.read(1)[0]
    pxsize = rng.read(1)[0] & 0x07
    colres = rng.read(1)[0] & 0x07
    sort   = rng.read(1)[0] & 0x03
    out.write(bytes([
        width,
        height,
        flags[0],
        flags[1],
        bgcol,
        pxsize,
        colres,
        sort,
    ]))

    # Global color table
    gctlen = (1 << rng.read(1)[0]) & 0xff
    left   = b'' if gctlen == 0 else rng.read(gctlen)
    right  = b'' if gctlen == 0 else rng.read(2 ** (gctlen+1)-gctlen-1)
    out.write(left + right)

    # Graphics control extension
    exttype      = b'\xF9'
    blocksize    = b'\x04'
    loopcount    = rng.read(1)[0]
    terminator   = b'\x00'
    disposal     = rng.read(1)[0] & 0x03
    userinput    = bool((disposal & 0x1c) >> 2)
    transparent  = bool((disposal & 0x20) >> 5)
    out.write(exttype + blocksize + bytes([loopcount]))
    out.write(terminator + bytes([disposal]))
    out.write(b'\x00\x00\x00')

    # Plain text extension
    exttype       = b'\x01'
    blocksize     = b'\x0C'
    left          = rng.read(1)
    top           = rng.read(1)
    width         = rng.read(1)[0]
    height        = rng.read(1)[0]
    cellwidth     = rng.read(1)[0]
    cellheight    = rng.read(1)[0]
    fgcolorindex  = rng.read(1)[0]
    bgcolorindex  = rng.read(1)[0]
    interlacetype = rng.read(1)[0]
    terminator    = b'\x00'
    out.write(exttype + blocksize + left + top)
    out.write(bytes([width, height, cellwidth, cellheight]))
    out.write(bytes([fgcolorindex, bgcolorindex, interlacetype]))
    out.write(terminator)

    # Application extension
    exttype       = b'\xFF'
    blocksize     = b'\x0B'
    appid         = b'NETSCAPE2.0'
    authcode      = b'\x03\x01'
    applen        = b'\x03'
    junk          = b'\x00\x00'
    loopcount     = rng.read(1)[0]
    terminator    = b'\x00'
    out.write(exttype + blocksize + appid + authcode + applen)
    out.write(junk + bytes([loopcount]))
    out.write(terminator)

    # Comment extension
    exttype       = b'\xFE'
    blocksize     = b'\x00'
    comment       = b''
    while len(comment)!= 256:
        comment += rng.read(1)
    out.write(exttype + blocksize + comment[:256])

    # Graphic control label
    label = b'\x21'
    out.write(label)

    # Graphic control block size
    blocksize = b'\x04'
    out.write(blocksize)

    # Delay time
    delaytime = rng.read(2)
    out.write(delaytime)

    # Transparent color index
    transindex = rng.read(1)[0]
    out.write(transindex)

    # Block terminator
    term = b'\x00'
    out.write(term)

    # Image descriptor
    label = b'\x2C'
    out.write(label)

    # Left position
    left = rng.read(1)

    # Top position
    top = rng.read(1)

    # Width and height of image
    width  = (1 << rng.read(1)[0]) % 640
    height = (1 << rng.read(1)[0]) % 480

    # Flags
    flags = [0x80, 0x00]
    localcolortable = bool((rng.read(1)[0] & 0x80) >> 7)
    interlace       = bool((rng.read(1)[0] & 0x40) >> 6)
    sorted    = bool((flags[0] & 0x20) >> 5)
    reserved  = rng.read(1)[0]
    lctbpp    = rng.read(1)[0] & 0x07
    lctsort   = rng.read(1)[0] & 0x03
    out.write(left + top + bytes([width, height]))
    out.write(flags[0] + flags[1])
    out.write(lctlen)

    # Local color table
    for _ in range(lctlen):
        out.write(rng.read(1))

    # Image data
    for y in range(height):
        for x in range(width):
            out.write(rng.read(pxsize // 8))

    # Trailer
    out.write(b'\x3B')


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('./test.gif', 'wb') as f:
            generate_random_gif(rng, f)