#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (1 << rng.read(1)[0]) % 640
    height = (1 << rng.read(1)[0]) % 480
    flags  = bytes([
        0x80 | ((rng.read(1)[0] & 0x70) >> 4),
        rng.read(1)[0] & 0x0f,
    ])
    bgcol  = rng.read(1)[0]
    pxsize = rng.read(1)[0] & 0x07
    colres = rng.read(1)[0] & 0x07
    sort   = rng.read(1)[0] & 0x03
    gctlen = rng.read(1)[0] if colres == 7 else 0
    out.write(struct.pack('>HHBBB', width, height, flags[0], flags[1], bgcol))

    # Global color table
    for _ in range(gctlen):
        out.write(rng.read(1))

    # Image descriptor
    left      = rng.read(2)
    top       = rng.read(2)
    width     = (1 << rng.read(1)[0]) % 640
    height    = (1 << rng.read(1)[0]) % 480
    flags     = bytes([
        0x80 | ((rng.read(1)[0] & 0x70) >> 4),
        rng.read(1)[0] & 0x0f,
    ])
    lctlen    = rng.read(1)[0] if flags[0] & 0x80!= 0 else 0
    interlace = bool((flags[0] & 0x40) >> 6)
    sorted    = bool((flags[0] & 0x20) >> 5)
    reserved  = rng.read(1)[0]
    lct       = [rng.read(1) for _ in range(lctlen)]
    out.write(left + top + bytes([width, height]))
    out.write(bytes([flags[0], flags[1]]))
    out.write(bytes([lctlen]))

    # Local color table
    for c in lct:
        out.write(c)

    # Pixel data
    while True:

        # Block size
        blocksz = rng.read(1)[0]

        # End of image?
        if blocksz == 0x3B:
            break

        # LZW minimum code size
        mincode = rng.read(1)[0]

        # Data sub-blocks
        while blocksz > 0:

            # Read next byte from input stream
            b = rng.read(1)[0]
            blocksz -= 1

            # If it's not a clear code or end of information code...
            if b >= mincode and b <= 0xFF:

                #...if we're at the beginning of an encoded sequence
                if len(seq) == 0:

                    # Start with this byte
                    seq += bytes([b])

                elif len(seq) < 255:

                    # Add to current sequence
                    seq += bytes([b])

                else:

                    # Flush current sequence
                    out.write(bytes([len(seq)]))
                    out.write(seq)

                    # Start new sequence
                    seq = bytes([b])

            elif b == 0x00:

                # Flush current sequence
                out.write(bytes([len(seq)]))
                out.write(seq)

                # Reset sequence
                seq = bytes()

            elif b == 0xFF:

                # Flush current sequence
                out.write(bytes([len(seq)]))
                out.write(seq)

                # Reset sequence
                seq = bytes()

                # Next byte is literal
                b = rng.read(1)[0]
                blocksz -= 1

                # Flush literal
                out.write(bytes([b]))

            else:

                # Flush current sequence
                out.write(bytes([len(seq)]))
                out.write(seq)

                # Reset sequence
                seq = bytes()

                # Get length of copy sequence
                n = b + 2
                blocksz -= 1

                # Copy sequence
                for i in range(n):
                    out.write(seq[-i - 1:])


def main():

    try:
        seed = int(sys.argv[1])
    except IndexError:
        print("Usage: {} SEED".format(os.path.basename(__file__)))
        return 1

    with open('/dev/urandom', 'rb') as rng:
        rng.seek(seed)
        with open('test.gif', 'wb') as f:
            generate_random_gif(rng, f)


if