#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a\x00\x00\x00\x00\xff\xff\xff\x00,\x00\x00'

    # width and height of image in pixels
    w = int.from_bytes(rng.read(2), byteorder='big')
    h = int.from_bytes(rng.read(2), byteorder='big')

    # flags for global color table
    gctsize = rng.read(1)[0] & 7
    gctsortflag = bool(rng.read(1)[0])

    # palette size
    psize = 2 ** ((gctsize + 1) * 3)

    # local color table flag
    lctflag = bool(rng.read(1)[0])

    # interlace flag
    interlaceflag = bool(rng.read(1)[0])

    # sort flag
    sortflag = bool(rng.read(1)[0])

    # reserved bits
    rsvd = rng.read(2)

    # background color index
    bg = rng.read(1)[0]

    # pixel aspect ratio
    par = rng.read(1)[0] & 0xf

    # global color table
    if gctsize > 0:
        gct = [struct.unpack('BBB', rng.read(3))
                for _ in range(psize)]

        out.write(header + bytes(gct))

    # LZW minimum code size
    mincodesize = rng.read(1)[0] & 0xf

    # image descriptor block
    out.write(bytes([
        0x2c, # image separator
        w % 256, w // 256, # image width
        h % 256, h // 256, # image height
        0x00, # no local color table
        0x00 # image is not interlaced
    ]))

    # local color table
    if lctflag:
        lctsize = 2 ** ((mincodesize + 1) * 3)
        lct = [struct.unpack('BBB', rng.read(3))
               for _ in range(lctsize)]

        out.write(bytes(lct))

    # LZW compressed data
    # TODO: implement this

    b = rng.read(1)[0] & 7
    packed = bytes([r | (g << 3) | (b << 6)])

    # background color index
    bg = rng.read(1)[0]

    # aspect ratio
    ar = rng.read(1)[0] & 0xf

    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)
    out.write(header + data)
