#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (1 << rng.read(1)[0]) % 640
    height = (1 << rng.read(1)[0]) % 480
    flags  = bytes([
        0x80 | ((rng.read(1)[0] & 0x70) >> 4),
        rng.read(1)[0] & 0x0f,
        0, 0, 0
    ])
    bg     = bytes([rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]])
    aspect = bytes([0, 0, 0])
    out.write(width + b'\x00\x00' + height + flags + bg + aspect)

    # Global color table
    for _ in range(256):
        r   = rng.read(1)[0]
        g   = rng.read(1)[0]
        b   = rng.read(1)[0]
        out.write(bytes([r, g, b]))

    # Image descriptor
    left      = bytes([(rng.read(1)[0] & 0xf8) >> 3])
    top       = bytes([(rng.read(1)[0] & 0xfe) >> 1])
    width     = bytes([(rng.read(1)[0] & 0xfc) >> 2])
    height    = bytes([(rng.read(1)[0] & 0xfe) >> 1])
    flags     = bytes([0x80 | (((rng.read(1)[0] & 0x06) >> 1) ^ 0x01)])
    trans_idx = bytes([rng.read(1)[0] & 0xff])
    out.write(left + top + width + height + flags + trans_idx)

    # LZW minimum code size
    min_code_size = bytes([rng.read(1)[0] & 0xff])
    out.write(min_code_size)

    # Pixel data
    lzw_dict = [None] * 256
    lzw_dict[0] = ''
    lzw_dict[1] = '\x01'
    lzw_dict[2] = '\x01\x00'
    lzw_dict[3] = '\x01\x00\x00'
    lzw_dict[4] = '\x01\x00\x00\x00'
    lzw_dict[5] = '\x01\x00\x00\x00\x00'
    lzw_dict[6] = '\x01\x00\x00\x00\x00\x00'
    lzw_dict[7] = '\x01\x00\x00\x00\x00\x00\x00'
    lzw_dict[8] = '\x01\x00\x00\x00\x00\x00\x00\x00'
    lzw_dict[9] = '\x01\x00\x00\x00\x00\x00\x00\x00\x00'
    lzw_dict[10] = '\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    lzw_dict[11] = '\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    lzw_dict[12] = '\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    lzw_dict[13] = '\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    lzw_dict[14] = '\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    lzw_dict[15] = '\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    lzw