#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # 16-bit width and height
    w = rng.read(2)
    h = rng.read(2)

    # packed field of color resolution (bits per pixel)
    r = rng.read(1)[0] & 7
    g = rng.read(1)[0] & 7
    b = rng.read(1)[0] & 7
    packed = bytes([r | (g << 3) | (b << 6)])

    # background color index
    bg = rng.read(1)[0]

    # aspect ratio - fixed to 4:3
    ar = b'\x00'

    # logical screen descriptor
    lsd = header + w + h + packed + bg + ar

    # global color table size in bits
    ctsize = rng.read(1)[0] & 7

    # sort flag - set if sorted by decreasing importance
    sf = rng.read(1)[0] >> 7

    # global color table flag - set if there is one
    gf = rng.read(1)[0] >> 7

    # size of global color table
    ctsz = 3 * ((1 << ctsize) - 1)

    # global color table
    gct = None
    if gf == 1:
        gct = bytearray()

        for _ in range((1 << ctsize)):
            gct += rng.read(3)

    # image block
    img = bytearray()

    # LZW minimum code size
    mincodesize = rng.read(1)[0]

    # first frame control block
    fc = b'\x2c'

    # delay time in hundredths of a second
    dt = rng.read(2)

    # transparent color index
    tc = rng.read(1)[0]

    # image left position
    il = rng.read(2)

    # image top position
    it = rng.read(2)

    # image width
    iw = rng.read(2)

    # image height
    ih = rng.read(2)

    # local color table flag - set if there is one
    lf = rng.read(1)[0] >> 7

    # interlace flag - set if interlaced
    intf = rng.read(1)[0] >> 6

    # sort flag - set if sorted by decreasing importance
    srtf = rng.read(1)[0] >> 5

    # reserved flags
    resv = rng.read(1)[0] & 0x1f

    # size of local color table
    lctsz = 3 * ((1 << ctsize) - 1)

    # local color table
    lct = None
    if lf == 1:
        lct = bytearray()

        for _ in range((1 << ctsize)):
            lct += rng.read(3)

    # LZW data subblock
    lzwdata = bytearray()

    # LZW end of information code
    eoi = b'\x00\x00'

    # LZW clear code
    clrcd = b'\x01\x00'

    # LZW end of stream code
    eoscd = b'\x00\xff'

    # LZW codes
    codes = [clrcd, eoi, eoscd]

    # LZW minimum code length
    minlen = len(codes[mincodesize])

    # LZW maximum code length
    maxlen = 12

    # LZW current code length
    currlen = minlen

    # LZW next available code
    availcode = 2 ** currlen

    # LZW previous code
    prevcode = None

    # LZW dictionary
    d = {bytes([i]): i for i in range(availcode)}

    # LZW buffer
    buf = []

    # LZW buffer size
    bufsz = 256

    # LZW buffer pointer
    ptr = 0

    # LZW buffer mask
    msk = 0

    # LZW buffer bit counter
    cnt = 0

    # LZW buffer bit accumulator
    acc = 0

    # LZW buffer bit accumulator count
    acnt = 0

    # LZW buffer bit accumulator mask
    amsk = 0xffff

    while True:
        # get a random byte
        b = rng.read(1)[0]

        # check if we're at the end of the buffer
        if ptr >= bufsz:
            # write the buffer to the output
            out.write(struct.pack('B', ptr))

            # reset the buffer
            buf = []
            ptr = 0

        # check if this is an LZW code
        if currlen > minlen and \
                (b & 0x80 or
                    (prevcode!= None and
                        (currlen < maxlen and
                            (b & 0x7f ^ 0x7f) <=
                                (prevcode[-1] & 0x7f)))):
            # add the code to the buffer
            buf.append(b)

            # increment the buffer pointer
            ptr += 1

            # shift the accumulator right
            acc >>= 1

            # decrement the accumulator count
            acnt -= 1

            # check if the accumulator count has reached zero
            if acnt == 0:
                # reset the accumulator count
                acnt = 16

                # reset the accumulator