#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom or /dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the magic string and version number to the output stream.
    out.write(b'GIF89a')

    # Generate a 16-bit integer with the length of the image data.
    datalen = struct.pack('>H', int.from_bytes(
            rng.read(2), byteorder='big'))

    # Write the image data length to the output stream.
    out.write(datalen)

    # Generate an unsigned byte for each color channel in the palette.
    red   = rng.read(1)[0]
    green = rng.read(1)[0]
    blue  = rng.read(1)[0]

    # Write the palette colors to the output stream.
    out.write(struct.pack('BBB', red, green, blue))

    # Generate a single bit indicating whether the image is interlaced.
    interlace = rng.read(1)[0] & 1

    # Write the image descriptor to the output stream.
    out.write(struct.pack('<BBHHB',
            0x2C,              # Image separator
            0,                 # Left position
            0,                 # Top position
            100,               # Width
            100,               # Height
            0 | interlace      # LZW code size + interlace flag
        ))

    # Generate the image data.
    imgdata = b'\x00' * 100*100

    # Write the image data to the output stream.
    out.write(imgdata)

    # Write the trailer to the output stream.
    out.write(b';')

if __name__ == '__main__':
    if len(sys.argv)!= 2:
        print("Usage: {} <output>".format(sys.argv[0]), file=sys.stderr)
        exit(-1)

    try:
        with open('/dev/urandom', 'rb') as rng:
            with open(sys.argv[1], 'wb') as out:
                generate_random_gif(rng, out)
    except PermissionError:
        print("Permission denied when trying to access '/dev/urandom'",
              file=sys.stderr)
        exit(-1)