#!/usr/bin/env python3

import struct
from typing import BinaryIO, Tuple


def read_uint16(f: BinaryIO) -> int:
    return struct.unpack('H', f.read(2))[0]


def write_uint16(f: BinaryIO, x: int):
    f.write(struct.pack('H', x))


def read_int16(f: BinaryIO) -> int:
    return struct.unpack('h', f.read(2))[0]


def write_int16(f: BinaryIO, x: int):
    f.write(struct.pack('h', x))


def read_uint32(f: BinaryIO) -> int:
    return struct.unpack('I', f.read(4))[0]


def write_uint32(f: BinaryIO, x: int):
    f.write(struct.pack('I', x))


def read_int32(f: BinaryIO) -> int:
    return struct.unpack('i', f.read(4))[0]


def write_int32(f: BinaryIO, x: int):
    f.write(struct.pack('i', x))


def read_float(f: BinaryIO) -> float:
    return struct.unpack('f', f.read(4))[0]


def write_float(f: BinaryIO, x: float):
    f.write(struct.pack('f', x))


def read_double(f: BinaryIO) -> float:
    return struct.unpack('d', f.read(8))[0]


def write_double(f: BinaryIO, x: float):
    f.write(struct.pack('d', x))


def read_string(f: BinaryIO) -> str:
    length = read_uint16(f)

    if length == 0xFFFF:
        raise ValueError("String is null")

    s = f.read(length).decode()
    assert len(s) == length
    return s


def write_string(f: BinaryIO, s: str):
    encoded = s.encode()
    write_uint16(f, len(encoded))
    f.write(encoded)


def read_bytes(f: BinaryIO) -> bytes:
    length = read_uint16(f)

    if length == 0xFFFF:
        raise ValueError("Bytes are null")

    b = f.read(length)
    assert len(b) == length
    return b


def write_bytes(f: BinaryIO, b: bytes):
    write_uint16(f, len(b))
    f.write(b)


def read_bool(f: BinaryIO) -> bool:
    return read_uint16(f)!= 0


def write_bool(f: BinaryIO, b: bool):
    write_uint16(f, 1 if b else 0)


def read_array(f: BinaryIO, reader) -> list:
    count = read_uint16(f)

    if count == 0xFFFF:
        raise ValueError("Array is null")

    array = []
    for _ in range(count):
        array.append(reader(f))

    return array


def write_array(f: BinaryIO, writer, array):
    if not isinstance(array, (list, tuple)):
        raise TypeError("Expected a list or tuple")

    write_uint16(f, len(array))
    for item in array:
        writer(f, item)


def read_tuple(f: BinaryIO, *readers) -> tuple:
    values = [reader(f) for reader in readers]
    return tuple(values)


def write_tuple(f: BinaryIO, *writers, values: tuple):
    if len(values)!= len(writers):
        raise ValueError("Tuple size does not match writers")

    for value, writer in zip(values, writers):
        writer(f, value)


def read_pair(f: BinaryIO, r1, r2) -> Tuple[any, any]:
    v1 = r1(f)
    v2 = r2(f)
    return v1, v2


def write_pair(f: BinaryIO, w1, w2, pair: Tuple[any, any]):
    w1(f, pair[0])
    w2(f, pair[1])
