#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical Screen Descriptor
    width = 1 + (rng.read(1)[0] % 256)
    height = 1 + (rng.read(1)[0] % 256)
    flags = b'\x87\x00' if rng.read(1)[0] & 1 else b'\x07\x00'
    bg_color = bytes([rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]])
    pixel_aspect_ratio = b'\x00'
    out.write((width).to_bytes(2, byteorder='little'))
    out.write((height).to_bytes(2, byteorder='little'))
    out.write(flags)
    out.write(bg_color)
    out.write(pixel_aspect_ratio)

    # Global Color Table
    color_table_size = min(max(1 << ((rng.read(1)[0] >> 3) + 1), 2), 256)
    global_color_table = [None for _ in range(color_table_size)]
    for i in range(color_table_size):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        global_color_table[i] = bytes([r,g,b])
    for entry in global_color_table:
        out.write(entry)

    # Image Descriptor
    left = 0
    top = 0
    width = 1 + (rng.read(1)[0] % 256)
    height = 1 + (rng.read(1)[0] % 256)
    flags = b'\x87\x00' if rng.read(1)[0] & 1 else b'\x07\x00'
    local_color_table_flag = b'\x00'
    interlace_flag = b'\x00'
    sort_flag = b'\x00'
    reserved = b'\x00\x00'
    local_color_table_size = min(max(1 << ((rng.read(1)[0] >> 3) + 1), 2), 256)
    local_color_table = [None for _ in range(local_color_table_size)]
    for i in range(local_color_table_size):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        local_color_table[i] = bytes([r,g,b])
    for entry in local_color_table:
        out.write(entry)
    out.write(((left).to_bytes(2, byteorder='little')))
    out.write(((top).to_bytes(2, byteorder='little')))
    out.write(((width).to_bytes(2, byteorder='little')))
    out.write(((height).to_bytes(2, byteorder='little')))
    out.write(flags)
    out.write(local_color_table_flag)
    out.write(interlace_flag)
    out.write(sort_flag)
    out.write(reserved)

    # LZW Minimum Code Size
    lzw_min_code_size = max(rng.read(1)[0] % 256, 2)
    out.write(lzw_min_code_size.to_bytes(1, byteorder='little'))

    # Frame Data
    data_block = []
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x00 or len(data_block) >= 2**lzw_min_code_size - 1:
            # End of image
            break

        elif block_type == 0x01:
            # Extension Block
            extension_label = rng.read(1)[0]

            if extension_label == 0xF9:
                # Graphic Control Extension
                size = rng.read(1)[0]

                # Skip over all fields except delay time and transparent color index
                skip_fields = {
                    4, # Block Size
                    1, # Packed Fields
                    3, # Delay Time
                    1, # Transparent Color Index
                    }
                while size > 0:
                    field_byte = rng.read(1)[0]

                    for bit in reversed(range(8)):
                        if not skip_fields