#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The width and height of the image are specified by `width` and `height`.
# The color depth is 256 (i.e., each pixel is an unsigned char).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int
    ):

    # Header
    out.write('GIF89a'.encode())

    # Logical screen descriptor
    # Width x Height = 100x100 pixels
    # Color depth = 256 colors
    out.write((
        7 << 4 |  # Global color table flag set
        0 << 3 |  # Sort flag clear
        0b111 << 0  # Size of global color table = 2^7 = 128 entries
    ).to_bytes(1, 'little'))
    out.write((width >> 8 & 0xff).to_bytes(1, 'little'))
    out.write((width & 0xff).to_bytes(1, 'little'))
    out.write((height >> 8 & 0xff).to_bytes(1, 'little'))
    out.write((height & 0xff).to_bytes(1, 'little'))
    out.write((
        0b111 << 5 |   # Background color index = transparent
        0b111 << 2 |   # Pixel aspect ratio = 1:1
        0b11     # Color resolution = 256 colors
    ).to_bytes(1, 'little'))

    # Global color table
    for i in range(2**7):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        out.write((r << 3 | g >> 3 | 1 << 7).to_bytes(1, 'little'))
        out.write(((g & 0b111) << 5 | b << 3).to_bytes(1, 'little'))

    # Image descriptor
    # Left position = 0
    # Top position = 0
    # Width x Height = 100x100 pixels
    out.write((0).to_bytes(1, 'little'))
    out.write((0).to_bytes(1, 'little'))
    out.write((width >> 8 & 0xff).to_bytes(1, 'little'))
    out.write((width & 0xff).to_bytes(1, 'little'))
    out.write((height >> 8 & 0xff).to_bytes(1, 'little'))
    out.write((height & 0xff).to_bytes(1, 'little'))
    out.write((
        0b111 << 5 |    # Local color table flag set
        0b111 << 3 |    # Interlace flag clear
        0b111         # LZW minimum code size
    ).to_bytes(1, 'little'))

    # Data sub-block
    # Each pixel is encoded as one byte with value between 0 and 255
    # (since we have 256 colors in our palette)
    for y in range(height):
        for x in range(width):
            out.write(rng.read(1))