#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical Screen Descriptor (LSD)
    width  = int.from_bytes(
            rng.read(2), byteorder='big', signed=False
        ) % 65536
    height = int.from_bytes(
            rng.read(2), byteorder='big', signed=False
        ) % 65536
    packed = int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        )
    global_color_table   = bool((packed & 0x80) >> 7)
    color_resolution     = ((packed & 0x70) >> 4) + 1
    sort                 = bool((packed & 0x08) >> 3)
    size                 = ((packed & 0x07) >> 0) + 1
    background_color_idx = int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        ) % 256
    pixel_aspect_ratio   = int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        ) % 256
    lsd = [width, height]
    lsd += [(global_color_table << 7) |
           (color_resolution - 1 << 4) |
           (sort << 3) |
           (size - 1)]
    lsd += [background_color_idx, pixel_aspect_ratio]
    out.write(struct.pack('>HHBBBBBB', *lsd))

    # Global Color Table
    if global_color_table:
        for i in range(pow(2, size)):
            r = int.from_bytes(
                    rng.read(1), byteorder='big', signed=False
                ) % 256
            g = int.from_bytes(
                    rng.read(1), byteorder='big', signed=False
                ) % 256
            b = int.from_bytes(
                    rng.read(1), byteorder='big', signed=False
                ) % 256
            out.write(struct.pack('BBB', r, g, b))

    # Image Descriptor
    left      = int.from_bytes(
            rng.read(2), byteorder='big', signed=False
        ) % 65536
    top       = int.from_bytes(
            rng.read(2), byteorder='big', signed=False
        ) % 65536
    packed    = int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        )
    local_ct  = bool((packed & 0x80) >> 7)
    interlace = bool((packed & 0x40) >> 6)
    sorted    = bool((packed & 0x20) >> 5)
    reserved  = bool((packed & 0x10) >> 4)
    lzw       = bool((packed & 0x08) >> 3)
    size      = ((packed & 0x07) >> 0) + 1
    idd = [left, top]
    idd += [(local_ct << 7) |
           (interlace << 6) |
           (sorted << 5) |
           (reserved << 4) |
           (lzw << 3) |
           (size - 1)]
    out.write(struct.pack('>HHBBBB', *idd))

    # Local Color Table
    if local_ct:
        for i in range(pow(2, size)):
            r = int.from_bytes(
                    rng.read(1), byteorder='big', signed=False
                ) % 256
            g = int.from_bytes(
                    rng.read(1), byteorder='big', signed=False
                ) % 256
            b = int.from_bytes(
                    rng.read(1), byteorder='big', signed=False
                ) % 256
            out.write(struct.pack('BBB', r, g, b))

    # LZW Minimum Code Size
    mincodesize = int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        ) % 256
   