#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # Header
    header = b'GIF89a'
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), 'big') % 65536
    height = int.from_bytes(rng.read(2), 'big') % 65536
    flags  = bytes([int.from_bytes(rng.read(1), 'big')])
    bgcol  = bytes([int.from_bytes(rng.read(1), 'big')])
    pxsize = bytes([int.from_bytes(rng.read(1), 'big')])
    gctlen = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(struct.pack('>HHBBB', width, height, *flags))
    out.write(bgcol + pxsize + gctlen)

    # Global color table
    for _ in range(0, 2**(gctlen[0] & 7)):
        r = bytes([int.from_bytes(rng.read(1), 'big')])
        g = bytes([int.from_bytes(rng.read(1), 'big')])
        b = bytes([int.from_bytes(rng.read(1), 'big')])
        out.write(r+g+b)

    # Image descriptor
    left   = int.from_bytes(rng.read(2), 'big') % 65536
    top    = int.from_bytes(rng.read(2), 'big') % 65536
    width  = int.from_bytes(rng.read(2), 'big') % 65536
    height = int.from_bytes(rng.read(2), 'big') % 65536
    flags  = bytes([int.from_bytes(rng.read(1), 'big')])
    lctlen = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(struct.pack('>HHHHBB', left, top, width, height, *flags))
    out.write(lctlen)

    # Local color table
    if lctlen!= b'\x00':
        for _ in range(0, 2**(lctlen[0] & 7)):
            r = bytes([int.from_bytes(rng.read(1), 'big')])
            g = bytes([int.from_bytes(rng.read(1), 'big')])
            b = bytes([int.from_bytes(rng.read(1), 'big')])
            out.write(r+g+b)

    # Image data
    while True:
        blocktype = bytes([int.from_bytes(rng.read(1), 'big')])

        if blocktype == b'\x2C':
            # Image descriptor
            left   = int.from_bytes(rng.read(2), 'big') % 65536
            top    = int.from_bytes(rng.read(2), 'big') % 65536
            width  = int.from_bytes(rng.read(2), 'big') % 65536
            height = int.from_bytes(rng.read(2), 'big') % 65536
            flags  = bytes([int.from_bytes(rng.read(1), 'big')])
            lctlen = bytes([int.from_bytes(rng.read(1), 'big')])

            out.write(blocktype)
            out.write(struct.pack('>HHHHBB', left, top, width, height, *flags))
            out.write(lctlen)

            # Local color table
            if lctlen!= b'\x00':
                for _ in range(0, 2**(lctlen[0] & 7)):
                    r = bytes([int.from_bytes(rng.read(1), 'big')])
                    g = bytes([int.from_bytes(rng.read(1), 'big')])
                    b = bytes([int.from_bytes(rng.read(1), 'big')])
                    out.write(r+g+b)

            # Image data
            while len(data := rng.read(width*height)) > 