#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # Write the magic string and version to the output stream
    out.write(b'GIF89a')

    # Generate a random width between 256 and 4096 pixels
    width = (int.from_bytes(rng.read(1), 'big') % 3072) + 256

    # Generate a random height between 256 and 4096 pixels
    height = (int.from_bytes(rng.read(1), 'big') % 3072) + 256

    # Write the screen descriptor block to the output stream
    out.write(struct.pack('<HHBBB', width, height, 0xF0 | 0x07, 0, 0))

    # Write the global color table to the output stream
    palette = [i for i in range(256)]
    rng.shuffle(palette)
    out.write(bytes([(i << 4) | j for i,j in zip(palette[::2], palette[1::2])]))

    # Write the image descriptor block to the output stream
    out.write(struct.pack('<HHBHB', 0, 0, 0xF0 | 0x07, 0, 0))

    # Generate a random length of LZW compressed data
    lzwlen = int.from_bytes(rng.read(1), 'big') % 1024

    # Write the LZW compressed data to the output stream
    compressor = LzwCompressor()
    out.write(compressor.compress(rng.read(lzwlen)))

    # Write the end-of-file marker to the output stream
    out.write(b'\x3B')


class LzwCompressor():
    def __init__(self):
        self._table = {
            0xFF: b'\x00\x00',
            0xFE: b'\x00\x01',
            0xFD: b'\x00\x02',
            0xFC: b'\x00\x03',
            0xFB: b'\x00\x04',
            0xFA: b'\x00\x05',
            0xF9: b'\x00\x06',
            0xF8: b'\x00\x07',
            0xF7: b'\x00\x08',
            0xF6: b'\x00\t',
            0xF5: b'\x00\n',
            0xF4: b'\x00\x0B',
            0xF3: b'\x00\x0C',
            0xF2: b'\x00\r',
            0xF1: b'\x00\x0E',
            0xF0: b'\x00\x0F',
            0xEF: b'\x00\x10',
            0xEE: b'\x00\x11',
            0xED: b'\x00\x12',
            0xEC: b'\x00\x13',
            0xEB: b'\x00\x14',
            0xEA: b'\x00\x15',
            0xE9: b'\x00\x16',
            0xE8: b'\x00\x17',
            0xE7: b'\x00\x18',
            0xE6: b'\x00\x19',
            0xE5: b'\x00\x1A',
            0xE4: b'\x00\x1B',
            0xE3: b'\x00\x1C',
            0xE2: b'\x00\x1D',
            0xE1: b'\x00\x1E',
            0xE0: b'\x00\x1F',
            0xDF: b'\x00 ',
            0xDE: b'\x00!',
            0xDD: b'\x00"',
            0xDC: b'\x00#',
            0xDB: b'\x00$',
            0xDA: b'\x00%',
            0xD9: b'\x00&',
            0xD8: b'\x