#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Header
    out.write(b'GIF89a')

    # Logical Screen Descriptor
    width  = int.from_bytes(
            rng.read(2), byteorder='big', signed=False
        ) % (1 << 16)
    height = int.from_bytes(
            rng.read(2), byteorder='big', signed=False
        ) % (1 << 16)
    flags  = int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        ) & 0x7
    bgcol  = int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        )
    aspect = int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        )

    out.write(struct.pack('>HHBBB', width, height, flags, bgcol, aspect))

    # Global Color Table
    for _ in range((flags >> 1) + 1):
        out.write(int.to_bytes(
                int.from_bytes(
                    rng.read(1), byteorder='big', signed=False
                ), length=3, byteorder='big'))

    # Image Descriptor
    left   = int.from_bytes(
            rng.read(2), byteorder='big', signed=False
        ) % (1 << 16)
    top    = int.from_bytes(
            rng.read(2), byteorder='big', signed=False
        ) % (1 << 16)
    width  = int.from_bytes(
            rng.read(2), byteorder='big', signed=False
        ) % (1 << 16)
    height = int.from_bytes(
            rng.read(2), byteorder='big', signed=False
        ) % (1 << 16)
    flags  = int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        ) & 0x7
    lctsize = ((flags >> 1) + 1) * 3
    interlace = bool(flags & 1)
    sortflag = bool(flags & 2)
    reserved = int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        )
    lct = [None] * lctsize
    for i in range(lctsize):
        lct[i] = int.to_bytes(
                int.from_bytes(
                    rng.read(1), byteorder='big', signed=False
                ), length=3, byteorder='big')
    out.write(struct.pack('>HHHHBB', left, top, width, height, flags, lctsize))
    for i in range(lctsize):
        out.write(lct[i])

    # Local Color Table
    if lctsize > 0:
        for _ in range(lctsize // 3):
            out.write(int.to_bytes(
                    int.from_bytes(
                        rng.read(1), byteorder='big', signed=False
                    ), length=3, byteorder='big'))

    # Image Data
    while True:
        blocktype = int.from_bytes(
                rng.read(1), byteorder='big', signed=False
            )

        if blocktype == 0x2C:
            break
        elif blocktype!= 0x21:
            raise ValueError("Invalid block type")

        blocklen = int.from_bytes(
                rng.read(1), byteorder='big', signed=False
            )

        if blocklen < 1:
            continue

        if blocklen >= 2 and \
           rng.read(blocklen - 2).startswith(b'\xFF\x0B'):
            out.write(b'\xFF\x0B')
            continue

        out.write(rng.read(blocklen))

    # Trailer
    out.write(b';')