#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Generate a 16-bit integer for the image width and height.
    width = rng.read(2)
    height = rng.read(2)
    width = struct.unpack('>H', width)[0] % (2**14 - 1) + 1
    height = struct.unpack('>H', height)[0] % (2**14 - 1) + 1

    # Write the GIF header to the output file.
    out.write(b'GIF89a')

    # Write the logical screen descriptor to the output file.
    out.write(struct.pack('<HHBBB', width, height, 0xF7, 0x0B, 0))

    # Write the global color table to the output file.
    palette = [i * i & 0xFF for i in range(256)]
    out.write(bytearray(palette))

    # Write the local color table to the output file.
    out.write(b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')

    # Write the graphics control extension to the output file.
    out.write(b'\x21\xF9\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')

    # Write the application extension to the output file.
    out.write(b'\x21\xFF\x0BNETSCAPE2.0\x03\x01\x00\x00\x00')

    # Write the comment extension to the output file.
    out.write(b'\x21\xFE\x00')

    # Write the plain text extension to the output file.
    out.write(b'\x21\x01\x0FComment:\x00')

    # Write the end of file marker to the output file.
    out.write(b';')

    # Write the pixel data to the output file.
    for _ in range(height):
        row = []

        # Generate a random number of pixels per row.
        num_pixels = rng.read(1)
        num_pixels = ord(num_pixels[0]) % ((width // 2) or 1) + 1

        # Generate each pixel value.
        for _ in range(num_pixels):

            # Generate a random pixel index from the palette.
            pixel = rng.read(1)
            pixel = ord(pixel[0]) % len(palette)

            # Add the pixel to the row.
            row += [pixel]

            # If this is an even pixel...
            if not len(row) % 2:

                #...add its mirror to the row as well.
                row += [(len(palette) - 1) ^ pixel]

        # Write the row to the output file.
        out.write(bytes([len(row)]))
        out.write(bytes(row))


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('./test.gif', 'wb') as out:
            generate_random_gif(rng, out)