#!/usr/bin/env python3

import struct
from typing import BinaryIO

def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write('GIF89a'.encode())

    # Logical screen descriptor
    width  = rng.randint(1024, 65536)
    height = rng.randint(1024, 65536)
    flags  = (
            0b10000000 | # Global color table flag
            0b01111000 | # Color resolution
            7           # Sorting method
        ).to_bytes(1, 'little', signed=False)
    bgcolorindex = 0
    pixelaspectratio = 0
    out.write(struct.pack('<HHBBB', width, height, flags[0], bgcolorindex, pixelaspectratio))

    # Global color table
    for i in range(2**((flags[0] & 0x07)+1)):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        a = 0xff if i == 0 else rng.randint(0, 255)
        out.write(struct.pack('<BBBB', r, g, b, a))

    # Image descriptor
    left   = 0
    top    = 0
    width  = rng.randint(width // 2, width - 1)
    height = rng.randint(height // 2, height - 1)
    flags  = (
            0b10000000 | # Local color table flag
            0b00000000   # Interlace and sort flags
        ).to_bytes(1, 'little', signed=False)
    bgcolorindex = 0
    pixelaspectratio = 0
    out.write(struct.pack('<HHBBB', left, top, width, height, flags[0]))

    # Local color table
    for i in range(2**((flags[0] & 0x07)+1)):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        a = 0xff if i == 0 else rng.randint(0, 255)
        out.write(struct.pack('<BBBB', r, g, b, a))

    # LZW minimum code size
    lzwmincodesize = 2**(rng.randint(2, 12))
    out.write(lzwmincodesize.to_bytes(1, 'little'))

    # Data sub-block
    datablocklen = rng.randint(1, 2**lzwmincodesize//2)
    while datablocklen > 0:
        datablocklen -= 1

        # RLE block
        if rng.randrange(0, 2):

            # RLE block length
            rlentab = [
                len(data) for data in [
                    b'\x00\x00',
                    b'\x00\x01',
                    b'\x00\x02',
                    b'\x00\xfe',
                    b'\x00\xff',
                    b'\xff\x00',
                    b'\xff\x01',
                    b'\xff\x02',
                    b'\xff\xfe',
                    b'\xff\xff'
                ]
            ]
            rlelen = rng.choice(rlentab)

            # RLE block contents
            if rlelen < 3:
                out.write(data[:rlelen])
            elif rlelen == 3:
                out.write(data[-2:])
            else:
                out.write(data[::2])

        # Raw block
        else:
            rawlen = rng.randint(1, 2**lzwmincodesize//2)
            out.write(rng.read(rawlen))