#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header
    header = b'GIF89a'

    # Generate the logical screen descriptor
    width = 100 + int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        ) % 50
    height = 100 + int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        ) % 50
    flags = bytes([int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        )])
    background = bytes([int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        )])
    aspect = bytes([int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        )])
    lsd = struct.pack('>HHBBB', width, height, *flags, *background, *aspect)

    # Write it to the output stream and flush it
    out.write(header + lsd)
    out.flush()

    # Generate the global color table
    gctsize = 2 << bool(lsd[7] & 0b10000000)
    for i in range(gctsize):
        r = int.from_bytes(
                rng.read(1), byteorder='big', signed=False
            ) % 256
        g = int.from_bytes(
                rng.read(1), byteorder='big', signed=False
            ) % 256
        b = int.from_bytes(
                rng.read(1), byteorder='big', signed=False
            ) % 256
        out.write(struct.pack('BBB', r, g, b))
    out.flush()

    # Generate the image descriptor
    left = 0
    top = 0
    width = 100 + int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        ) % 50
    height = 100 + int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        ) % 50
    flags = bytes([int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        )])
    id = struct.pack('>HHHB', left, top, width, height, *flags)

    # Write it to the output stream and flush it
    out.write(id)
    out.flush()

    # Generate the local color table
    lctsize = 2 << bool(id[7] & 0b10000000)
    for i in range(lctsize):
        r = int.from_bytes(
                rng.read(1), byteorder='big', signed=False
            ) % 256
        g = int.from_bytes(
                rng.read(1), byteorder='big', signed=False
            ) % 256
        b = int.from_bytes(
                rng.read(1), byteorder='big', signed=False
            ) % 256
        out.write(struct.pack('BBB', r, g, b))
    out.flush()

    # Generate the graphics control extension
    label = b'\x21\xF9\x04'
    blocksize = len(label) + 1
    flags = bytes([int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        )])
    delay = int.from_bytes(
            rng.read(2), byteorder='big', signed=False
        ) % 65536
    transindex = int.from_bytes(
            rng.read(1), byteorder='big', signed=False
        ) % 256
    ext = struct.pack('>BBHBB', blocksize, *flags, delay, transindex)

    # Write it to the output stream and flush it
    out.write(ext)
    out.flush()

    # Generate the plain text extension
    label = b'\x21\x01\x00\xFF'
    blocksize = len(label) + 1
    txt = rng.read(blocksize - 1)
    ext = struct.pack('B', blocksize) + txt

    # Write it to the output