#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (stdin by default).
def generate_random_gif(
        rng: BinaryIO=sys.stdin.buffer, out: BinaryIO=sys.stdout.buffer
    ):
    header = b'GIF89a'

    # Read 10 bytes for the logical screen descriptor
    lsd = rng.read(10)
    width = int.from_bytes(lsd[6:8], 'little', signed=False)
    height = int.from_bytes(lsd[8:], 'little', signed=False)

    # Write the header and LSD to output
    out.write(header + lsd)

    # Calculate how many pixels there are in this image
    pixelcount = width * height

    # Read all of the palette entries from stdin
    palentries = []
    while len(palentries) < 256*3:
        entry = rng.read(1)

        if not entry or entry == b'\x3B': break
        else: palentries += list(entry)

    # Write the palette entries to stdout
    out.write(b''.join([bytes((i,)) for i in palentries]))

    # Write the global color table flag and size
    out.write(b'\x87\x00')

    # Write the background color index
    bgindex = rng.read(1)[0] % 256
    out.write(bgindex.to_bytes(1, 'little'))

    # Write the pixel aspect ratio
    out.write(b'\x00')

    # Write the local color table flag and size
    out.write(b'\x00\x00')

    # Write the rest of the image data
    while True:
        # Read up to 255 bytes at a time until we get an EOF
        chunk = rng.read(min(pixelcount, 255))

        # If we got an EOF, write the final block
        if not chunk:
            out.write(b'\x00')

            # Pad with zeroes until we reach the next multiple of 256
            padding = (-len(chunk)) % 256
            out.write(b'\x00'*padding)

            # Write the trailer
            out.write(b';')

            return

        # Otherwise, write the length of the current chunk
        out.write(len(chunk).to_bytes(1, 'little'))

        # Write the actual chunk
        out.write(chunk)

        # Decrement our counter
        pixelcount -= len(chunk)

