#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom or /dev/urandom).
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Read the magic bytes and version
    header = b'GIF89a'
    out.write(header)

    # Read the logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % 65536
    height = int.from_bytes(rng.read(2), byteorder='little') % 65536
    flags  = int.from_bytes(rng.read(1), byteorder='little') & 0b1111
    bg     = int.from_bytes(rng.read(1), byteorder='little')
    aspect = int.from_bytes(rng.read(1), byteorder='little')
    out.write(width.to_bytes(2, 'little'))
    out.write(height.to_bytes(2, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bg.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Read the global color table
    gctsize = int.from_bytes(rng.read(1), byteorder='little')
    if gctsize > 0:
        for i in range(gctsize):
            r = int.from_bytes(rng.read(1), byteorder='little')
            g = int.from_bytes(rng.read(1), byteorder='little')
            b = int.from_bytes(rng.read(1), byteorder='little')
            out.write(r.to_bytes(1, 'little'))
            out.write(g.to_bytes(1, 'little'))
            out.write(b.to_bytes(1, 'little'))

    # Read the image descriptor
    left   = int.from_bytes(rng.read(2), byteorder='little') % width
    top    = int.from_bytes(rng.read(2), byteorder='little') % height
    width  = int.from_bytes(rng.read(2), byteorder='little') % width
    height = int.from_bytes(rng.read(2), byteorder='little') % height
    flags  = int.from_bytes(rng.read(1), byteorder='little') & 0b1111
    lctsize= int.from_bytes(rng.read(1), byteorder='little')
    interlaced = bool((flags >> 7) & 1)
    sortflag   = bool((flags >> 6) & 1)
    reserved   = bool((flags >> 5) & 1)
    lctbpp     = ((flags << 4) | (reserved << 3)) & 0b1111
    lctsort    = ((flags << 3) | (sortflag << 2)) & 0b11
    lctsize    = (((flags << 2) | (interlaced << 1)) & 0b11) | lctsize
    out.write(left.to_bytes(2, 'little'))
    out.write(top.to_bytes(2, 'little'))
    out.write(width.to_bytes(2, 'little'))
    out.write(height.to_bytes(2, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(lctsize.to_bytes(1, 'little'))

    # Read the local color table
    if lctsize > 0:
        for i in range(lctsize):
            r = int.from_bytes(rng.read(1), byteorder='little')
            g = int.from_bytes(rng.read(1), byteorder='little')
            b = int.from_bytes(rng.read(1), byteorder='little')
            out.write(r.to_bytes(1, 'little'))
            out.write(g.to_bytes(1, 'little'))
            out.write(b.to_bytes(1, 'little'))

    # Read the LZW minimum code size
    mincodesize = int.from_bytes(rng.read(1), byteorder='little')
    out.write(min