#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = (1 << rng.read(1)[0]) % 65536
    height = (1 << rng.read(1)[0]) % 65536
    flags  = ((1 << rng.read(1)[0]) & 0x7c) | 0x80
    bgcol  = rng.read(1)[0]
    pxsize = rng.read(1)[0]
    gctlen = rng.read(1)[0] if flags & 0x80 else 0
    out.write(struct.pack('>HHBBB', width, height, flags, bgcol, pxsize))

    # Global color table
    for _ in range(gctlen):
        out.write(rng.read(3))

    # Image descriptor
    left   = (1 << rng.read(1)[0]) % width
    top    = (1 << rng.read(1)[0]) % height
    width2 = (1 << rng.read(1)[0]) % (width - left)
    height2= (1 << rng.read(1)[0]) % (height - top)
    flags  = ((1 << rng.read(1)[0]) & 0xe0) | 0x01
    lctlen = rng.read(1)[0] if flags & 0x80 else 0
    interl = rng.read(1)[0] if flags & 0x40 else 0
    out.write(struct.pack('>HHHHBB', left, top, width2, height2, flags, lctlen))

    # Local color table
    for _ in range(lctlen):
        out.write(rng.read(3))

    # Data sub-block
    while True:
        blocktype = rng.read(1)[0]

        if blocktype == 0x2C:
            break

        elif blocktype == 0x21:
            while True:
                exttype = rng.read(1)[0]

                if exttype == 0xf9:
                    break

                elif exttype == 0xff:
                    while True:
                        appid = rng.read(1)[0]

                        if appid == 0xfe:
                            break

                        length = rng.read(1)[0]
                        out.write(appid + length.to_bytes(1, 'big'))
                        out.write(rng.read(length))

                    continue

                raise Exception("Unknown extension type")

            continue

        elif blocktype == 0x3b:
            break

        raise Exception("Unknown block type")

    # Image data
    while True:
        blocktype = rng.read(1)[0]

        if blocktype == 0x21:
            while True:
                exttype = rng.read(1)[0]

                if exttype == 0xff:
                    while True:
                        appid = rng.read(1)[0]

                        if appid == 0x00 or appid == 0xff:
                            break

                        length = rng.read(1)[0]
                        out.write(appid + length.to_bytes(1, 'big'))
                        out.write(rng.read(length))

                    continue

                raise Exception("Unknown extension type")

            continue

        elif blocktype == 0x2c:
            break

        elif blocktype == 0x3b:
            break

        raise Exception("Unknown block type")

    # LZW minimum code size
    mincodesize = rng.read(1)[0]
    out.write(mincodesize.to_bytes(1, 'big'))

    # Pixels
    while True:
        blocktype = rng.read(1)[0]

        if blocktype == 0x00:
            break

        elif blocktype == 0x01:
            break

        elif blocktype >= 0x02 and blocktype <= 0xbf:
            pass

        elif blocktype == 0xc0:
            break

        elif blocktype == 0xc4:
            break

        elif blocktype == 0xcc:
            break

        elif blocktype == 0xcd:
            break

        raise Exception("Unknown block type")