#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The width and height of the image are determined by
# `width` and `height`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int
    ):
    # Header
    header = b'GIF89a'
    out.write(header)

    # Logical screen descriptor
    logical_screen_descriptor = (
            0x21 << 8 | 7
        ) << 4 | 0x00 | \
        0x00 | 0x00 | 0x00 | \
        ((width & 0xff) << 8) | (width >> 8) | \
        ((height & 0xff) << 8) | (height >> 8)
    out.write(struct.pack('>H', logical_screen_descriptor))

    # Global color table
    global_color_table = [0] * 64
    for i in range(len(global_color_table)):
        global_color_table[i] = rng.read(3)
    out.write(b''.join([ bytes([c]) for c in global_color_table ]))

    # Image descriptor
    image_descriptor = (
            0x2C << 8 | \
            0x00 | 0x00 | 0x00 | \
            ((width & 0xff) << 8) | (width >> 8) | \
            ((height & 0xff) << 8) | (height >> 8) | \
            0x00 | 0x00 | 0x00 | 0x00 | \
            0x00 | 0x00 | 0x00
        )
    out.write(struct.pack('>H', image_descriptor))

    # Local color table
    local_color_table = [0] * 64
    for i in range(len(local_color_table)):
        local_color_table[i] = rng.read(3)
    out.write(b''.join([ bytes([c]) for c in local_color_table ]))

    # LZW minimum code size
    lzw_minimum_code_size = 2
    out.write(bytes([lzw_minimum_code_size]))

    # Data sub-block
    data_sub_block = []
    while True:
        # Read a block from /dev/urandom
        block = rng.read(128)

        if len(block) == 0 or block[-1]!= 0:
            break

        data_sub_block += list(block[:-1])

    # Write the data sub-blocks to the output stream
    while len(data_sub_block):
        chunk = min(len(data_sub_block), 255)
        out.write(bytes([chunk]))
        out.write(bytes(data_sub_block[:chunk]))
        del data_sub_block[:chunk]

    # End of file
    out.write(b'\x3B')