#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (stdin by default).
def generate_random_gif(
        rng: BinaryIO=sys.stdin.buffer, out: BinaryIO=sys.stdout.buffer
    ):
    # Read the magic bytes and version string
    header = b'GIF89a'
    version = rng.read(6)

    # Write the header to stdout
    out.write(header + version)

    # Read the logical screen descriptor
    lsd = rng.read(7)
    width = int.from_bytes(lsd[0:2], 'little', signed=False)
    height = int.from_bytes(lsd[2:4], 'little', signed=False)
    flags = int.from_bytes(lsd[5:], 'little', signed=False)

    # Write the logical screen descriptor to stdout
    out.write(lsd)

    # Read the global color table if it exists
    gctsize = 1 << ((flags & 0b111) + 1)
    gct = None
    if gctsize > 0:
        gct = rng.read(gctsize * 3)

        # Write the global color table to stdout
        out.write(gct)

    # Read the image block
    while True:
        # Read the image descriptor
        idd = rng.read(9)
        left = int.from_bytes(idd[0:2], 'little', signed=False)
        top = int.from_bytes(idd[2:4], 'little', signed=False)
        width = int.from_bytes(idd[4:6], 'little', signed=False)
        height = int.from_bytes(idd[6:8], 'little', signed=False)
        flags = int.from_bytes(idd[8:], 'little', signed=False)

        # Write the image descriptor to stdout
        out.write(idd)

        # Read the local color table if it exists
        lctsize = 1 << ((flags >> 4) + 1)
        lct = None
        if lctsize > 0:
            lct = rng.read(lctsize * 3)

            # Write the local color table to stdout
            out.write(lct)

        # Read the LZW minimum code size
        mincodesize = rng.read(1)[0]

        # Write the LZW minimum code size to stdout
        out.write(mincodesize.to_bytes(1, 'little'))

        # Read the compressed data
        data = []
        while True:
            datum = rng.read(1)[0]
            if datum == 0x00:
                break
            elif datum <= 0x0f:
                for _ in range(datum):
                    data.append(rng.read(1))
            else:
                length = 1 << (datum - 0x10)
                data.append(rng.read(length))

        # Write the compressed data to stdout
        out.write(b'\x00'.join(data))