#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The first 6 bytes are always "GIF87a" or "GIF89a".
    # We want to make sure that we use the correct version of the format.
    header = b'GIF89a'
    out.write(header)

    # Next comes the logical screen descriptor. It contains information about
    # the image size and color palette used in this image.
    width = rng.randint(1024, 2048)
    height = rng.randint(512, 1024)
    flags = (
        0b00000001 | # Global Color Table Flag
        0b00000100   # Sorting Flag
    )
    bgcolorindex = 0
    pixelaspectratio = 0
    out.write(struct.pack('>HHBBB', width, height, flags, bgcolorindex,
                          pixelaspectratio))

    # Then follows the global color table. This is an optional field which can be
    # omitted if all colors from the images are already defined in the local color
    # tables of each frame. If it's present, then there must be at least as many
    # entries in the table as specified by the value of bits per pixel in the
    # Logical Screen Descriptor. Each entry consists of three bytes indicating the
    # red, green and blue components for one color.
    numcolors = 2 ** ((flags & 0b00000001) * 1 +
                      (flags & 0b00000010) * 2 +
                      (flags & 0b00000100) * 4 +
                      (flags & 0b00001000) * 8)
    for i in range(numcolors):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write((r,g,b).to_bytes(3,'little'))

    # Afterwards come the frames. A frame starts with its own local color table.
    # This means that every frame can have a different set of colors. However,
    # since we don't support transparency yet, we'll just reuse the same color
    # table for every frame.
    #
    # Since we're not doing anything fancy here, we'll just write the entire
    # image into one single frame.
    #
    # First comes the frame header.
    flags = 0b00000001 # Use local color table flag
    duration = 10      # Duration in milliseconds
    transindex = 0     # Transparent color index
    out.write(struct.pack('>BBH', flags, duration, transindex))

    # Now comes the local color table.
    numcolors = 2 ** ((flags & 0b00000001) * 1 +
                      (flags & 0b00000010) * 2 +
                      (flags & 0b00000100) * 4 +
                      (flags & 0b00001000) * 8)
    for i in range(numcolors):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write((r,g,b).to_bytes(3,'little'))

    # Finally comes the actual image data.
    #
    # For simplicity, we'll just put a single black pixel into the center of the
    # image.
    xpos = width // 2
    ypos = height // 2
    out.write(struct.pack('<HHB', xpos, ypos, 0x00))