#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (defaults to /dev/urandom).
# The generated image is 1024x768 pixels with 256 colors and
# has a color table of size 256.
def generate_random_gif(
        rng: BinaryIO=None, out: BinaryIO=sys.stdout.buffer
    ):

    # Default to reading from /dev/urandom if no RNG was provided
    if not rng:
        try:
            rng = open('/dev/urandom', 'rb')
        except FileNotFoundError as e:
            print('Could not find /dev/urandom!', file=sys.stderr)
            raise e
    
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    out.write(int.to_bytes(
            0x00, length=1, byteorder='big'))    # Screen width
    out.write(int.to_bytes(
            0x00, length=1, byteorder='big'))    # Screen height
    out.write(b'\xf0\x00')                      # Flags: global color table
    out.write(b'\x00\x00')                      # Background color index
    out.write(b'\x00\x00')                      # Pixel aspect ratio

    # Global color table
    for i in range(256):
        out.write(int.to_bytes(
                i, length=1, byteorder='big'))   # Color index
        out.write(int.to_bytes(
                0xff, length=1, byteorder='big')) # Red
        out.write(int.to_bytes(
                0xff, length=1, byteorder='big')) # Green
        out.write(int.to_bytes(
                0xff, length=1, byteorder='big')) # Blue

    # Image descriptor
    out.write(b'\x2c\x00')                      # Image separator
    out.write(int.to_bytes(
            0x00, length=1, byteorder='big'))    # Left position
    out.write(int.to_bytes(
            0x00, length=1, byteorder='big'))    # Top position
    out.write(int.to_bytes(
            0x00, length=1, byteorder='big'))    # Width
    out.write(int.to_bytes(
            0x00, length=1, byteorder='big'))    # Height
    out.write(b'\x08\x00')                      # Flags: local color table
    out.write(b'\x00\x00')                      # Interlace flag
    out.write(b'\x00\x00')                      # Sort flag
    out.write(b'\x00\x00')                      # Reserved

    # Local color table
    for i in range(256):
        out.write(int.to_bytes(
                i, length=1, byteorder='big'))   # Color index
        out.write(int.to_bytes(
                0xff, length=1, byteorder='big')) # Red
        out.write(int.to_bytes(
                0xff, length=1, byteorder='big')) # Green
        out.write(int.to_bytes(
                0xff, length=1, byteorder='big')) # Blue

    # Image data
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')        # Data length
    data = rng.read(datalen)                     # Actual data
    out.write(data)                              # Write it all down!