#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom).
# The size of the generated image is determined by
# `width`, `height`.
def generate_random_gif(
        rng: BinaryIO, width: int, height: int, out: BinaryIO
    ):

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    out.write(struct.pack('>HHBBB', 0, 0, 7, 0, 0))

    # Global color table
    for i in range(0, 1 << 7):
        out.write(struct.pack('BBBB', *rng.read(4)))

    # Image descriptor
    out.write(struct.pack('>HHHHB', 0, 0, width - 1, height - 1, 0))

    # Local color table
    for i in range(0, 1 << 7):
        out.write(struct.pack('BBBB', *rng.read(4)))

    # LZW minimum code size
    out.write(b'\x08')

    # Data sub-block
    lzw_mincodesize = 8
    lzw_clearcode = 1 << lzw_mincodesize
    lzw_endofinfo = lzw_clearcode + 1
    lzw_endofdata = lzw_endofinfo + 1
    lzw_firstcode = lzw_endofdata + 1

    lzw_dictionary = [i for i in range(lzw_firstcode)]
    lzw_dictionary[0] = None
    lzw_dictionary[1] = None
    lzw_dictionary[256] = None

    lzw_nextcode = len(lzw_dictionary)
    lzw_prevcode = None
    lzw_curcode = None

    def lzw_addcode():
        if lzw_nextcode == 2 ** lzw_mincodesize and \
                lzw_curcode!= lzw_endofinfo and \
                lzw_curcode!= lzw_endofdata:

            lzw_mincodesize += 1

        lzw_dictionary.append(None)
        return lzw_nextcode

    while True:
        # Generate a new code from the input stream
        lzw_curbyte = ord(rng.read(1))
        lzw_curcode = ((lzw_curcode or lzw_firstcode) << 8) | lzw_curbyte

        # If we have not seen this code before...
        if lzw_curcode >= lzw_nextcode:
            #... add it to our dictionary
            lzw_prevcode = lzw_curcode
            lzw_curcode = lzw_addcode()

        # If we are at the end of the input stream...
        if lzw_curbyte == 0xFF:
            break

        # Write the current code as a sequence of bytes
        curcode = lzw_curcode
        while curcode > lzw_clearcode:
            out.write(struct.pack('BB',
                    lzw_dictionary[curcode][0],
                    lzw_dictionary[curcode][1]))
            curcode = lzw_dictionary[curcode][2]

    # End of information block
    out.write(struct.pack('BB', 0x00, 0x00))
