#!/usr/bin/env python3

import struct
from typing import BinaryIO

def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical Screen Descriptor
    width  = (rng.read(1)[0] % 256) + 1
    height = (rng.read(1)[0] % 256) + 1
    flags  = bytes([
        0x80 | ((rng.read(1)[0] & 0x70) >> 4),
        rng.read(1)[0] & 0x0f])
    background = rng.read(1)[0]
    aspect     = rng.read(1)[0]
    out.write(struct.pack('>HHBBB', width, height, flags[0], flags[1], background))

    # Global Color Table
    for _ in range(0, (2 << (flags[0] & 0x07))):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        a = 0xff if (flags[0] & 0x80) else 0x00
        out.write(struct.pack('BBBB', r, g, b, a))

    # Image Descriptor
    left   = 0
    top    = 0
    width  = (rng.read(1)[0] % 256) + 1
    height = (rng.read(1)[0] % 256) + 1
    flags  = bytes([
        0x80 | ((rng.read(1)[0] & 0x70) >> 4),
        rng.read(1)[0] & 0x0f])
    local_color_table = bool((flags[0] & 0x80)!= 0)
    interlace         = bool((flags[1] & 0x40)!= 0)
    sort              = bool((flags[1] & 0x20)!= 0)
    reserved          = rng.read(1)[0]
    lctsize           = 2 << (flags[0] & 0x07)
    out.write(struct.pack('>HHHHBB', left, top, width, height, flags[0], flags[1]))

    # Local Color Table
    if local_color_table:
        for _ in range(0, lctsize):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            a = 0xff if (flags[0] & 0x80) else 0x00
            out.write(struct.pack('BBBB', r, g, b, a))

    # Image Data
    lzwmincodesize = rng.read(1)[0] + 1
    out.write(lzwmincodesize.to_bytes(1, 'little'))
    imagedata = []
    while True:
        blocktype = rng.read(1)[0]
        if blocktype == 0x00: break

        elif blocktype == 0x01:
            subblocklen = rng.read(1)[0]
            subblock = rng.read(subblocklen)
            imagedata += list(subblock)

        elif blocktype == 0xfe:
            subblocklen = rng.read(1)[0]
            subblock = rng.read(subblocklen)
            pass

        elif blocktype == 0xff:
            subblocklen = rng.read(1)[0]
            subblock = rng.read(subblocklen)
            pass

        else: raise Exception("Invalid GIF block type")

    # LZW Decompression
    codesize = lzwmincodesize
    prefix = [None] * 256
    suffix = [None] * 256
    table = {}
    for i in range(0, 256):
        table[(i,)] = bytes([i])
    for i in range(0, len(imagedata)):
        c = imagedata[i]
        if codesize < 256 and c not in table:
            codesize += 1
        if c in table:
            s = table[(c,)]
        else:
            s = prefix[-1] + table[(prefix[-1][-1:],)][0:1]
        imagedata[i] = s
        if len(s) >= 256: continue
        prefix.append(