#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        datalen = int.from_bytes(
            rng.read(2), 'big') % (0xffff + 1 - len(header))
        data = rng.read(datalen)
        if data == header:
            break
    else:
        raise Exception('Failed to find GIF header')

    width = int.from_bytes(rng.read(2), 'big')
    height = int.from_bytes(rng.read(2), 'big')
    flags = int.from_bytes(rng.read(1), 'big')
    background = rng.read(1)[0]
    aspect = rng.read(1)[0]

    out.write(b'\x47\x49\x46\x38\x39\x61')
    out.write(struct.pack('<H', width))
    out.write(struct.pack('<H', height))
    out.write(struct.pack('B', flags | 0x80))
    out.write(struct.pack('BB', background, aspect))

    while True:
        blocktype = rng.read(1)[0]
        if blocktype == 0x3b:
            out.write(b'\x3b')
            return
        elif blocktype != 0x2c:
            raise Exception('Unexpected block type {}'.format(blocktype))

        length = int.from_bytes(rng.read(2), 'big')
        out.write(struct.pack('<H', length))

        left = int.from_bytes(rng.read(2), 'big')
        top = int.from_bytes(rng.read(2), 'big')
        width = int.from_bytes(rng.read(2), 'big')
        height = int.from_bytes(rng.read(2), 'big')
        flags = int.from_bytes(rng.read(1), 'big')
        localpalette = bool(flags & 0x80)
        interlaced = bool(flags & 0x40)
        sorted = bool(flags & 0x20)
        colorres = flags & 0x07
        if colorres > 7 or colorres < 1:
            raise Exception('Invalid color resolution {}'.format(colorres))
        size = 3 * (2 ** colorres)
        if localpalette:
            palette = rng.read(size)
            out.write(struct.pack('B', size // 3 - 1))
            out.write(palette)
        else:
            out.write(b'\x00')

        lzwmincodesize = rng.read(1)[0]
        if lzwmincodesize > 12:
            raise Exception('Invalid LZW min code size {}'.format(lzwmincodesize))
        out.write(struct.pack('B', lzwmincodesize))

        clearcode = 1 << lzwmincodesize
        endofinfo = clearcode + 1
        nextcode = endofinfo + 1
        maxcode = 1 << (lzwmincodesize + 1) - 1
        firstcode = None
        oldcode = None
        stack = []
        while True:
            code = int.from_bytes(rng.read(2), 'big')
            if code == clearcode:
                firstcode = None
                oldcode = None
                continue
            if code == endofinfo:
                out.write(struct.pack('!H', 0))
                break

            if firstcode is None:
                firstcode = code
                oldcode = code
                stack.append(firstcode >> lzwmincodesize)
                continue

            incode = code
            if code >= nextcode:
                incode = oldcode

            if incode == clearcode:
                firstcode = None
                oldcode = None
                continue

            if incode == endofinfo:
                out.write(struct.pack('!H', 0))
                break

            stack.append((incode & ((1 << lzwmincodesize) - 1)))
            if len(stack) > maxcode + 2 - lzwmincodesize:
                raise Exception('Stack overflow')

            if firstcode is not None:
                prefix = (oldcode & ((1