#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = struct.pack('>H', 1024)
    height = struct.pack('>H',  768)
    flags  = struct.pack('B',   0x80 | 0x40 | 0x20)
    bg     = struct.pack('B',   0xff)
    aspect = struct.pack('B',   0x00)
    out.write(width + height + flags + bg + aspect)

    # Global color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f

        if ((r == 0) and (g == 0) and (b == 0)):
            r += 1
            g += 1
            b += 1

        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left      = struct.pack('>H', 0)
    top       = struct.pack('>H', 0)
    width     = struct.pack('>H', 1024)
    height    = struct.pack('>H',  768)
    flags     = struct.pack('B',   0x80)
    local_ct  = struct.pack('B',   0x00)
    interlace = struct.pack('B',   0x00)
    sort      = struct.pack('B',   0x00)
    size      = struct.pack('B',   0x00)
    out.write(left + top + width + height + flags + local_ct + interlace + sort + size)

    # Graphic control extension
    ext_code  = struct.pack('B',   0xf9)
    block_len = struct.pack('B',   0x04)
    flags     = struct.pack('B',   0x00)
    delay     = struct.pack('>H',  int(rng.read(1)))
    trans     = struct.pack('B',   0xff)
    out.write(ext_code + block_len + flags + delay + trans)

    # Plain text extension
    ext_code  = struct.pack('B',   0xfe)
    block_len = struct.pack('B',   0x0c)
    grid_left = struct.pack('>H',  int(rng.read(1)))
    grid_top  = struct.pack('>H',  int(rng.read(1)))
    grid_w    = struct.pack('>H',  int(rng.read(1)))
    grid_h    = struct.pack('>H',  int(rng.read(1)))
    char_w    = struct.pack('B',   0x00)
    char_h    = struct.pack('B',   0x00)
    fg        = struct.pack('B',   0xff)
    bg        = struct.pack('B',   0xff)
    out.write(ext_code + block_len + grid_left + grid_top + grid_w + grid_h + char_w + char_h + fg + bg)

    # Comment extension
    ext_code  = struct.pack('B',   0xfe)
    block_len = struct.pack('B',   0x01)
    comment   = struct.pack('B',   0x00)
    out.write(ext_code + block_len + comment)

    # Application extension
    ext_code  = struct.pack('B',   0xff)
    block_len = struct.pack('B',   0x0b)
    app_id    = struct.pack('11s', b'Python GIF')
    auth_code = struct.pack('3s',  b'RNG')
    version   = struct.pack('B',   0x00)
    out.write(ext_code + block_len + app_id + auth_code + version)

    # Image data
    for y in range(0, 768):
        for x in range(0, 1024):
            pixel = struct.pack('B', ord(rng.read(1)))
            out.write(pixel)

    # Trailer
    out.write(b'\x3b')

