#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = struct.pack('>H', 1024)
    height = struct.pack('>H',  768)
    flags  = struct.pack('B',   0x80 | 0x40 | 0x20)
    bg     = struct.pack('B',   0xff)
    aspect = struct.pack('B',   0x00)
    out.write(width + height + flags + bg + aspect)

    # Global color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f
        out.write(struct.pack('BBB', r, g, b))

    # Application extension
    ext_type = struct.pack('B', 0xFF)
    app_id   = 'RANDOM'.encode()
    auth_cod = 'RANDM'.encode()
    data     = rng.read(10)
    block_size = struct.pack('B', len(app_id)+len(auth_cod)+len(data))
    out.write(ext_type+block_size+app_id+auth_cod+data)

    # Image descriptor
    image_separator = struct.pack('B', 0x2C)
    left            = struct.pack('>H', 0)
    top             = struct.pack('>H', 0)
    width           = struct.pack('>H', 1024)
    height          = struct.pack('>H',  768)
    flags           = struct.pack('B',   0x80 | 0x08)
    local_color_table_size = struct.pack('B', 0x00)
    interlace_flag         = struct.pack('B', 0x00)
    sort_flag              = struct.pack('B', 0x00)
    out.write(image_separator + left + top + width + height + flags + local_color_table_size + interlace_flag + sort_flag)

    # Local color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f
        out.write(struct.pack('BBB', r, g, b))

    # LZW minimum code size
    lzw_minimum_code_size = struct.pack('B', 0x08)
    out.write(lzw_minimum_code_size)

    # Pixel data
    while True:
        pixel_count = int.from_bytes(rng.read(1), byteorder='big')
        if not pixel_count:
            break
        pixels = [ord(rng.read(1)) for _ in range(pixel_count)]
        out.write(bytes(pixels))

    # Trailer
    trailer = struct.pack('B', 0x3B)
    out.write(trailer)

