#!/usr/bin/env python3

import struct
from typing import List, Tuple

from . import (
    AES_BLOCKSIZE,
    AES_KEYSIZE,
    AES_ROUNDS,
    AES_SUBKEYS,
    AES_TABLE,
    AES_TABLE_INV,
)


def aes_keygen(key: bytes) -> List[List[int]]:
    """Generate subkeys from key.

    Args:
        key (bytes): The 16-byte key to generate the subkeys for.

    Returns:
        list: The subkeys generated from the given key.
    """
    assert len(key) == AES_KEYSIZE
    subkeys = [None] * AES_SUBKEYS
    subkeys[0] = [struct.unpack(">I", key[i : i + 4])[0] for i in range(0, AES_KEYSIZE, 4)]
    for r in range(1, AES_ROUNDS + 1):
        subkeys[r] = []
        for i in range(AES_BLOCKSIZE // 4):
            t = subkeys[r - 1][i]
            if r % Nk == 0:
                t = _aes_subword(t) ^ Rcon[r // Nk]
            subkeys[r].append(_aes_rotword(subkeys[r - 1][i + 4]) ^ t)
    return subkeys


def aes_encrypt(plaintext: bytes, key: bytes) -> bytes:
    """Encrypt plaintext using the given key.

    Args:
        plaintext (bytes): The plaintext to encrypt.
        key (bytes): The 16-byte key to use for encryption.

    Returns:
        bytes: The encrypted ciphertext.
    """
    assert len(plaintext) == AES_BLOCKSIZE
    assert len(key) == AES_KEYSIZE
    subkeys = aes_keygen(key)
    state = [struct.unpack(">I", plaintext[i : i + 4])[0] for i in range(0, AES_BLOCKSIZE, 4)]
    add_round_key(state, subkeys[0])
    for r in range(1, AES_ROUNDS):
        sub_bytes(state)
        shift_rows(state)
        mix_columns(state)
        add_round_key(state, subkeys[r])
    sub_bytes(state)
    shift_rows(state)
    add_round_key(state, subkeys[-1])
    return b"".join([struct.pack(">I", s) for s in state])


def aes_decrypt(ciphertext: bytes, key: bytes) -> bytes:
    """Decrypt ciphertext using the given key.

    Args:
        ciphertext (bytes): The ciphertext to decrypt.
        key (bytes): The 16-byte key to use for decryption.

    Returns:
        bytes: The decrypted plaintext.
    """
    assert len(ciphertext) == AES_BLOCKSIZE
    assert len(key) == AES_KEYSIZE
    subkeys = aes_keygen(key)
    state = [struct.unpack(">I", ciphertext[i : i + 4])[0] for i in range(0, AES_BLOCKSIZE, 4)]
    add_round_key(state, subkeys[-1])
    for r in reversed(range(1, AES_ROUNDS)):
        inv_shift_rows(state)
        inv_sub_bytes(state)
        add_round_key(state, subkeys[r - 1])
        inv_mix_columns(state)
    inv_shift_rows(state)
    inv_sub_bytes(state)
    add_round_key(state, subkeys[0])
    return b"".join([struct.pack(">I", s) for s in state])


#
# Helper functions
#


def _aes_subword(w: int) -> int:
    """Substitute word w with S-box values."""
    return sum((AES_TABLE[b] << (8 * i)) for