#!/usr/bin/env python3

import sys
import struct
from collections import namedtuple

# https://github.com/python-pillow/Pillow/blob/master/src/PIL/TiffImagePlugin.py
TIFF_HEADER = b"II\x2a\x00"  # little endian, magic number
TIFF_BIGENDIAN = b"MM\x00*"  # big endian, magic number

# TIFF File Directory (IFD) Entry
Entry = namedtuple("Entry", ["tag", "type", "count", "value"])

def read_entry(f):
    tag = struct.unpack("<H", f.read(2))[0]
    type = struct.unpack("<H", f.read(2))[0]
    count = struct.unpack("<I", f.read(4))[0]
    value = None
    if type == 1:  # BYTE
        value = struct.unpack("<B", f.read(1))[0]
    elif type == 2:  # ASCII
        value = f.read(count).decode()
    elif type == 3:  # SHORT
        value = struct.unpack("<H", f.read(2))[0]
    elif type == 4:  # LONG
        value = struct.unpack("<I", f.read(4))[0]
    elif type == 5:  # RATIONAL
        num, denom = struct.unpack("<II", f.read(8))
        value = (num / denom)
    else:
        raise Exception("Unknown type")
    return Entry(tag, type, count, value)

def write_entry(f, entry):
    f.write(struct.pack("<H", entry.tag))
    f.write(struct.pack("<H", entry.type))
    f.write(struct.pack("<I", entry.count))
    if entry.type == 1:  # BYTE
        f.write(struct.pack("<B", entry.value))
    elif entry.type == 2:  # ASCII
        f.write(entry.value.encode())
    elif entry.type == 3:  # SHORT
        f.write(struct.pack("<H", entry.value))
    elif entry.type == 4:  # LONG
        f.write(struct.pack("<I", entry.value))
    elif entry.type == 5:  # RATIONAL
        f.write(struct.pack("<II", entry.value[0], entry.value[1]))
    else:
        raise Exception("Unknown type")

def read_ifd(f):
    entries = []
    num_entries = struct.unpack("<H", f.read(2))[0]
    for i in range(num_entries):
        entries.append(read_entry(f))
    next_offset = struct.unpack("<I", f.read(4))[0]
    return entries, next_offset

def write_ifd(f, entries):
    f.write(struct.pack("<H", len(entries)))
    for entry in entries:
        write_entry(f, entry)
    f.write(b"\x00\x00\x00\x00")  # next IFD offset

def read_tiff(f):
    header = f.read(4)
    if header != TIFF_HEADER and header != TIFF_BIGENDIAN:
        raise Exception("Not a valid TIFF file")
    is_bigendian = (header == TIFF_BIGENDIAN)
    if is_bigendian:
        raise Exception("Big endian not supported")
    entries, next_offset = read_ifd(f)
    while next_offset > 0:
        f.seek(next_offset)
        entries_, next_offset = read_ifd(f)
        entries += entries_
    return entries

def write_tiff(f, entries):
    f.write(TIFF_HEADER)
    write_ifd(f, entries)

def main():
    if len(sys.argv) < 2:
        print("Usage: {} <input> [<output>]".format(sys.argv[0]), file=sys.stderr)
        exit(