#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_gif(out: BinaryIO, rng: BinaryIO):
    # Magic header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + (rng.read(2)[0] % 640)
    height = 1 + (rng.read(2)[0] % 480)
    flags = rng.read(1)[0] & 7
    background = rng.read(1)[0] & 0xff
    aspect = rng.read(1)[0] & 0xf
    out.write(struct.pack('<HBBB', width, height, flags, background))
    if aspect != 0:
        out.write(aspect.to_bytes(1, 'little'))

    # Global color table
    colors = [None] * 256
    for i in range(2 ** ((flags >> 4) + 1)):
        red = rng.read(1)[0] & 0xff
        green = rng.read(1)[0] & 0xff
        blue = rng.read(1)[0] & 0xff
        colors[i] = (red, green, blue)

    # Image descriptor
    left = rng.read(2)[0] & 0xffff
    top = rng.read(2)[0] & 0xffff
    width = rng.read(2)[0] & 0xffff
    height = rng.read(2)[0] & 0xffff
    flags = rng.read(1)[0] & 0xff
    local_color_table = None
    if flags & 0x80:
        local_color_table = []
        for i in range(2 ** ((flags >> 4) + 1)):
            red = rng.read(1)[0] & 0xff
            green = rng.read(1)[0] & 0xff
            blue = rng.read(1)[0] & 0xff
            local_color_table.append((red, green, blue))

    # Local color table
    if not local_color_table and global_color_table:
        local_color_table = global_color_table[:]

    # Raster data
    lzw_minimum_code_size = rng.read(1)[0] & 0xff
    clear_code = 1 << lzw_minimum_code_size
    end_of_information_code = clear_code + 1
    code_size = lzw_minimum_code_size + 1
    dictionary = {}
    for i in range(clear_code):
        dictionary[i] = bytes([i])
    old_code = -1
    while True:
        new_code = rng.read(1)[0] & 0xff
        if new_code == clear_code:
            code_size = lzw_minimum_code_size + 1
            dictionary = {}
            for i in range(clear_code):
                dictionary[i] = bytes([i])
            old_code = -1
            continue
        elif new_code == end_of_information_code:
            break
        if old_code < 0:
            output = dictionary[new_code]
            old_code = new_code
        else:
            if new_code not in dictionary:
                if len(dictionary) >= (1 << code_size):
                    code_size += 1
                dictionary[new_code] = dictionary[old_code] + \
                    dictionary[old_code][0].to_bytes(1, 'little')
            output = dictionary[new_code]
            old_code = new_code
        out.write(output)
