#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(out:BinaryIO, rng:callable):
    # Header
    out.write(b'GIF89a')
    
    # Logical Screen Descriptor
    width = 10 + rng(256)
    height = 10 + rng(256)
    color_table_size = 7 + rng(8)
    background_color_index = rng(2**color_table_size)
    pixel_aspect_ratio = 0
    out.write(struct.pack('<HHBB', width, height, color_table_size, background_color_index))
    out.write(pixel_aspect_ratio.to_bytes(1, 'little'))
    
    # Global Color Table
    for i in range(2 ** color_table_size):
        red = rng(256)
        green = rng(256)
        blue = rng(256)
        out.write(struct.pack('<BBB', red, green, blue))
        
    # Application Extension Block
    block_length = 11
    application_identifier = b'NETSCAPE2.0'
    application_authentication_code = b'A' * 3
    loop_count = 0xffff
    out.write(struct.pack('<H', block_length))
    out.write(application_identifier)
    out.write(application_authentication_code)
    out.write(struct.pack('<H', loop_count))
    
    # Image Descriptor
    left = rng(width - 10)
    top = rng(height - 10)
    width = 10 + rng(width - left)
    height = 10 + rng(height - top)
    local_color_table_flag = False
    interlace_flag = False
    sort_flag = True
    reserved_local_color_table_size = 0
    out.write(struct.pack('<HHBB', left, top, width, height))
    packed = local_color_table_flag << 7 | \
             interlace_flag << 6 | \
             sort_flag << 5 | \
             reserved_local_color_table_size
    out.write(packed.to_bytes(1, 'little'))
    
    # Local Color Table
    for i in range(2 ** reserved_local_color_table_size):
        red = rng(256)
        green = rng(256)
        blue = rng(256)
        out.write(struct.pack('<BBB', red, green, blue))
        
    # LZW Minimum Code Size
    minimum_code_size = 2 + rng(8)
    out.write(minimum_code_size.to_bytes(1, 'little'))
    
    # Pixels
    pixels = []
    for y in range(height):
        row = []
        for x in range(width):
            index = rng(2 ** color_table_size)
            row.append(index)
        pixels.append(row)
    lzw_encode(pixels, out, minimum_code_size)
    
    # Trailer
    out.write(b'\x00\x3b')

# Encodes an image as a GIF LZW stream
def lzw_encode(image:list, out:BinaryIO, minimum_code_size:int):
    clear_code = 1 << minimum_code_size
    end_of_information_code = clear_code + 1
    code_mask = (1 << minimum_code_size) - 1
    available_codes = 1 << minimum_code_size + 2
    old_code = None
    table = {}
    for y in range(len(image)):
        for x in range(len(image[y])):
            code = image[y][x]
            if old_code is not None:
                if (old_code, code) not in table:
                    table[(old_code, code)] = available_codes
                    available_codes += 1
                    if available_codes > 4096:
                        raise Exception("Too many codes")
                code =