#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = struct.pack('>H', 1024)
    height = struct.pack('>H',  768)
    flags  = struct.pack('B',   0x80 | 0x40 | 0x20)
    bg     = struct.pack('B',   0xff)
    aspect = struct.pack('B',   0x00)
    out.write(width + height + flags + bg + aspect)

    # Global color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f

        if ((r == 0) and (g == 0) and (b == 0)):
            r += 1
            g += 1
            b += 1

        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left      = struct.pack('>H', 0)
    top       = struct.pack('>H', 0)
    width     = struct.pack('>H', 1024)
    height    = struct.pack('>H',  768)
    flags     = struct.pack('B',   0x00 | 0x08 | 0x04)
    bg        = struct.pack('B',   0x00)
    trans     = struct.pack('B',   0x00)
    interlace = struct.pack('B',   0x00)
    sort      = struct.pack('B',   0x00)
    out.write(left + top + width + height + flags + bg + trans + interlace + sort)

    # Local color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f

        if ((r == 0) and (g == 0) and (b == 0)):
            r += 1
            g += 1
            b += 1

        out.write(struct.pack('BBB', r, g, b))

    # Graphic control extension
    ext_code = struct.pack('B', 0xf9)
    size     = struct.pack('B', 0x04)
    flags    = struct.pack('B', 0x00)
    delay    = struct.pack('<H', 0xffff)
    trans    = struct.pack('B', 0x00)
    block    = struct.pack('BBB', *[ord(c) for c in 'NET'])
    out.write(ext_code + size + flags + delay + trans + block)

    # Plain text extension
    ext_code = struct.pack('B', 0xfe)
    size     = struct.pack('B', 0x0c)
    grid     = struct.pack('B', 0x00)
    cell     = struct.pack('B', 0x00)
    fg       = struct.pack('B', 0xff)
    bg       = struct.pack('B', 0x00)
    x        = struct.pack('>H', 0)
    y        = struct.pack('>H', 0)
    width    = struct.pack('>H', 1024)
    height   = struct.pack('>H',  768)
    data     = struct.pack('B', 0x01)
    block    = struct.pack('BBBBHHB', *[ord(c) for c in 'NET'])
    out.write(ext_code + size + grid + cell + fg + bg + x + y + width + height + data + block)

    # Application extension
    ext_code = struct.pack('B', 0xff)
    size     = struct.pack('B', 0x0b)
    app      = struct.pack('BBBB', *[ord(c) for c in 'NET '])
    block    = struct.pack('BBBB', *[ord(c) for c in 'NET'])
    out.write(ext_code + size + app + block)

    # Comment extension
    ext_code = struct.pack('B', 0xfe)
    size     = struct.pack('B', 0x0d)
    grid     = struct.pack('B', 0x00)
    cell     = struct.pack('B', 0x00)
    fg       = struct.pack('B', 0xff)
    bg       = struct.pack('B', 0x00)
    x        = struct.pack('>H', 0)
    y        = struct.pack('>H', 0)
    width    = struct.pack('>H', 1024)
    height   = struct.pack('>H',  768)
    data     = struct.pack('B', 0x01)
    block    = struct.pack('BBBBHHB', *[ord(c) for c in 'NET'])
    out.write(ext_code + size + grid + cell + fg + bg + x + y + width + height + data + block)

    # Trailer
    out.write(b'\x3b')

