#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = struct.pack('>H', 1024)
    height = struct.pack('>H',  768)
    flags  = struct.pack('B',   0x80 | 0x40 | 0x20)
    bg     = struct.pack('B',   0xff)
    aspect = struct.pack('B',   0x00)
    out.write(width + height + flags + bg + aspect)

    # Global color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f

        if ((r == 0) and (g == 0) and (b == 0)):
            r += 1
            g += 1
            b += 1

        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left      = struct.pack('>H', 0)
    top       = struct.pack('>H', 0)
    width     = struct.pack('>H', 1024)
    height    = struct.pack('>H',  768)
    flags     = struct.pack('B',   0x80)
    local_map = struct.pack('B',   0x00)
    interlace = struct.pack('B',   0x00)
    sort      = struct.pack('B',   0x00)
    size      = struct.pack('B',   0x00)
    out.write(left + top + width + height + flags + local_map + interlace + sort + size)

    # LZW minimum code size
    out.write(struct.pack('B', 0x08))

    # Pixel data
    for i in range(0, 1024 * 768):
        out.write(struct.pack('B', ord(rng.read(1))))

    # Trailer
    out.write(b'\x3b')


# Checks whether the given file is a valid GIF file
def check_file(filename: str) -> bool:
    with open(filename, 'rb') as f:
        header = f.read(6)
        if header != b'GIF89a':
            print("Invalid GIF header")
            return False

    return True


