#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + rng.read(2)[0] % 640
    height = 1 + rng.read(2)[0] % 480
    flags = rng.read(1)[0] & 7
    background = rng.read(1)[0] % 256
    aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack('<HBBB', width, height, flags, background))
    out.write(aspect_ratio.to_bytes(1, 'little'))

    # Global color table
    if flags & 1 != 0:
        for _ in range(2 ** ((flags >> 4) + 1)):
            out.write(rng.read(3))

    # Application extension
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01')
    out.write(b'\x00\x00\x00')

    # Image descriptor
    left = rng.read(2)[0] % width
    top = rng.read(2)[0] % height
    width = 1 + rng.read(2)[0] % (width - left)
    height = 1 + rng.read(2)[0] % (height - top)
    flags = rng.read(1)[0] & 7
    local_color_table = []
    if flags & 1 == 0:
        local_color_table = [rng.read(3) for _ in range(2 ** ((flags >> 4) + 1))]
    out.write(struct.pack('<HHBB', left, top, width, height))
    out.write(flags.to_bytes(1, 'little'))
    if len(local_color_table) > 0:
        for color in local_color_table:
            out.write(color)

    # Graphic control extension
    delay = rng.read(2)[0] * 10
    transparent = rng.read(1)[0] % 256
    disposal = rng.read(1)[0] % 8
    out.write(b'\x21\xf9\x04\x08\x00\x00'.replace(b'\x00', delay.to_bytes(2, 'little')))
    out.write((disposal << 2).to_bytes(1, 'little'))
    out.write(transparent.to_bytes(1, 'little'))

    # Plain text extension
    block_size = rng.read(1)[0] % 256
    grid_left = rng.read(2)[0] % width
    grid_top = rng.read(2)[0] % height
    grid_width = rng.read(2)[0] % (width - grid_left)
    grid_height = rng.read(2)[0] % (height - grid_top)
    cell_width = rng.read(1)[0] % 256
    cell_height = rng.read(1)[0] % 256
    foreground = rng.read(1)[0] % 256
    background = rng.read(1)[0] % 256
    out.write(struct.pack('<BBHHBBBB', block_size, grid_left, grid_top, grid_width, grid_height, cell_width, cell_height, foreground, background))
    for _ in range(block_size // 12):
        out.write(rng.read(12))

    # Comment extension
    comment = b''
    while True:
        byte = rng.read(1)[0]
        if byte == 0:
            break
        comment += bytes([byte])
    out.write(b'\x21\xfe')
    out.write(len(comment).to_bytes(1, 'little'))
    out.write(comment)

    # Image data
    lzw_min