#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bool:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = struct.pack('>H', 1024)
    height = struct.pack('>H',  768)
    flags  = struct.pack('B',   0x80 | 0x40 | 0x20)
    bg     = struct.pack('B',   0xff)
    aspect = struct.pack('B',   0x00)
    out.write(width + height + flags + bg + aspect)

    # Global color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f

        if ((r == 0) and (g == 0) and (b == 0)):
            r += 1

        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left      = struct.pack('>H',      0)
    top       = struct.pack('>H',      0)
    width     = struct.pack('>H', 1024)
    height    = struct.pack('>H',  768)
    flags     = struct.pack('B',       0x00)
    trans_idx = struct.pack('B',       0x00)
    out.write(left + top + width + height + flags + trans_idx)

    # LZW minimum code size
    min_code_size = struct.pack('B', 8)
    out.write(min_code_size)

    # Image data
    lzw_data = bytearray()
    while len(lzw_data) < 1024 * 768 - 1:
        c = ord(rng.read(1))
        if not c: continue
        lzw_data.append(c)

    lzw_data[len(lzw_data)-1] |= 0xe0

    # LZW compression
    prev_code = None
    cur_code  = None
    max_code  = 2**8
    clear_code = 1 << 8
    end_code   = clear_code + 1
    next_code  = end_code + 1
    codes     = [None] * max_code
    lzw_dict  = {}

    def emit_code(code):
        nonlocal prev_code, cur_code, next_code, lzw_dict, lzw_data

        if code is None or code >= next_code: return

        if code == clear_code:
            codes[:] = [(i, chr(i)) for i in range(max_code)]

            lzw_dict = {chr(i): i for i in range(max_code)}
            lzw_dict[''] = clear_code
            lzw_dict['\x00'] = end_code

            lzw_data.append(clear_code)
            prev_code = None
            cur_code  = None
            next_code = end_code + 1
            return

        if code == end_code:
            lzw_data.append(end_code)
            prev_code = None
            cur_code  = None
            next_code = end_code + 1
            return

        if prev_code is not None:
            lzw_data.append(prev_code)

        if code in lzw_dict:
            lzw_data.append(code)
        else:
            pair = codes[code]
            lzw_data.extend([pair[0], pair[1]])

            new_pair = (next_code, pair[1] + pair[1][0])
            codes[next_code] = new_pair
            lzw_dict[new_pair[1]] = next_code

            next_code += 1

        prev_code = code
        cur_code  = None


    for c in lzw_data:
        if cur_code is None:
            cur_code = c
            continue

        emit_code((cur_code << 8) | c)
        cur_code = None

    emit_code(end_code)

    # Write compressed image to output stream
    out.write(bytes(lzw_data))

    # Trailer
    out.write(b'\x3b')

    return True



if