#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO, Optional


def main(argv: list[str] = []):
    if len(argv) < 2 or argv[1].startswith('-'):
        print('Usage:', argv[0], '<file>', file=sys.stderr)
        return 1

    with open(argv[1], 'rb') as f:
        try:
            data = read_data(f)
        except ValueError as e:
            print(e, file=sys.stderr)
            return 1

        if not data:
            print('<empty>', file=sys.stderr)
            return 1

        for i in range(len(data)):
            print('%d: %s' % (i + 1, format_byte(data[i])))


def read_data(f: BinaryIO) -> bytes:
    magic = f.read(4)
    if magic!= b'\x7FELF':
        raise ValueError('Not an ELF file.')

    endianess = f.read(1)
    if endianess == b'\x01':
        endianess = '<'
    elif endianess == b'\x02':
        endianess = '>'
    else:
        raise ValueError('Invalid endianness.')

    bits = int.from_bytes(f.read(1), byteorder='little', signed=False)
    if bits!= 64:
        raise ValueError('Only 64-bit files are supported.')

    header_type = int.from_bytes(f.read(1), byteorder='little', signed=False)
    if header_type!= 1:
        raise ValueError('Only executable files are supported.')

    machine = int.from_bytes(f.read(2), byteorder='little', signed=False)
    if machine!= 0xF3:
        raise ValueError('Only x86-64 is supported.')

    version = int.from_bytes(f.read(1), byteorder='little', signed=False)
    if version!= 1:
        raise ValueError('Only version 1 is supported.')

    entrypoint = int.from_bytes(f.read(8), byteorder=endianess, signed=False)

    phoff = int.from_bytes(f.read(8), byteorder=endianess, signed=False)
    shoff = int.from_bytes(f.read(8), byteorder=endianess, signed=False)
    flags = int.from_bytes(f.read(4), byteorder=endianess, signed=False)
    ehsize = int.from_bytes(f.read(2), byteorder=endianess, signed=False)
    phentsize = int.from_bytes(f.read(2), byteorder=endianess, signed=False)
    phnum = int.from_bytes(f.read(2), byteorder=endianess, signed=False)
    shentsize = int.from_bytes(f.read(2), byteorder=endianess, signed=False)
    shnum = int.from_bytes(f.read(2), byteorder=endianess, signed=False)
    shstrndx = int.from_bytes(f.read(2), byteorder=endianess, signed=False)

    if shoff > 0 and shnum > 0:
        # Read section headers.
        f.seek(shoff)
        sections = [None] * shnum
        for _ in range(shnum):
            name = read_string(f, shstrndx)

            shname = int.from_bytes(f.read(4), byteorder=endianess, signed=False)
            shtype = int.from_bytes(f.read(4), byteorder=endianess, signed=False)
            offset = int.from_bytes(f.read(8), byteorder=endianess, signed=False)
            size = int.from_bytes(f.read(8), byteorder=endianess, signed=False)
            addralign = int.from_bytes(
                f.read(8), byteorder=endianess, signed=False)
            entsize = int.from_bytes(
                f.read(8), byteorder=endianess, signed=False)

            sections[shname - 1] = {
                'name': name,
                'type': shtype,
                'offset': offset,
               'size': size,
                'addralign': addralign,
                'entsize': entsize,
            }

        # Find the.text section.
        textsection = None
        for s in sections:
            if s['name'] == '.text':
                textsection = s
                break

        if not textsection:
            raise ValueError('.text section missing from binary.')

        #