#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The size of the output is determined by the first 10 bytes read from `rng`.


def main(rng: BinaryIO, out: BinaryIO):
    # Read the header.
    header = rng.read(6)

    # Write the header to the output.
    out.write(header)

    # Read the logical screen descriptor and write it to the output.
    lsd_bytes = rng.read(7)
    width, height = struct.unpack("<HH", lsd_bytes[4:])
    out.write(lsd_bytes[:4] + struct.pack("<BBH", *struct.unpack(">BBH", lsd_bytes[4:])))

    # Generate the global color table.
    gct_size = (2 << ((lsd_bytes[5] & 7) + 1)) - 1
    gct_bytes = bytearray([0x00]) * (gct_size * 3)
    for i in range(gct_size):
        r = next(rng) % 256
        g = next(rng) % 256
        b = next(rng) % 256

        gct_bytes[i*3+0] = r
        gct_bytes[i*3+1] = g
        gct_bytes[i*3+2] = b

    # Write the global color table to the output.
    out.write(b"\x2c" + struct.pack("<BBBHB", *struct.unpack(">BBBHB", lsd_bytes[:5])) +
              struct.pack("B", gct_size) + gct_bytes)

    # Write the image descriptor to the output.
    out.write(b"\x2c\x00\x00\x00\x00\x00\x00")

    # Write the local color table to the output.
    out.write(b"\x2c\x00\x00\x00\x00\x00\x00")

    # Write the image data to the output.
    while True:
        block_type = rng.read(1)[0]
        if block_type == 0x21:
            # Extension blocks are ignored.
            continue
        elif block_type == 0x2C:
            # Image descriptor block.
            break
        else:
            raise ValueError(f"Invalid block type {block_type}")

    out.write(block_type)
    out.write(rng.read(8))

    # Write the actual image data to the output.
    while True:
        block_type = rng.read(1)[0]
        if block_type == 0x21:
            # Extension blocks are ignored.
            continue
        elif block_type == 0x2C:
            # Image descriptor block.
            break
        elif block_type == 0x3B:
            # Trailer block.
            return
        else:
            raise ValueError(f"Invalid block type {block_type}")

    out.write(block_type)
    out.write(rng.read())


if