#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bool:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = struct.pack('>H', 1024)
    height = struct.pack('>H',  768)
    flags  = struct.pack('B',   0x80 | 0x40 | 0x20)
    bg     = struct.pack('B',   0xff)
    aspect = struct.pack('B',   0x00)
    out.write(width + height + flags + bg + aspect)

    # Global color table
    for i in range(0, 256):
        r = int(rng.read(1)[0] % 256)
        g = int(rng.read(1)[0] % 256)
        b = int(rng.read(1)[0] % 256)

        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left      = struct.pack('>H',      0)
    top       = struct.pack('>H',      0)
    width     = struct.pack('>H', 1024)
    height    = struct.pack('>H',  768)
    flags     = struct.pack('B',       0x00 | 0x04 | 0x01)
    trans_idx = struct.pack('B',        0x00)
    out.write(left + top + width + height + flags + trans_idx)

    # LZW minimum code size
    lzw_min_code_size = struct.pack('B', 2)
    out.write(lzw_min_code_size)

    # Data sub-blocks
    data = []
    while len(data) < 1024 * 768 // 2:
        byte = rng.read(1)[0]

        if byte == 0x00 or byte == 0xff:
            data += [byte, 0x00]
        else:
            data += [byte]

    # Encode data with LZW compression
    lzw_table = {
        0x00 : 0x00,
        0xff : 0xff,
    }
    lzw_prev = None
    lzw_curr = None
    lzw_next = None
    lzw_code = 256
    lzw_bits = 9
    lzw_buf  = 0
    lzw_mask = 0x1ff

    def lzw_flush():
        nonlocal lzw_table, lzw_prev, lzw_curr, lzw_next, \
                   lzw_code, lzw_bits, lzw_buf, lzw_mask

        if not lzw_prev is None and not lzw_curr is None:

            if lzw_prev!= lzw_curr:
                lzw_table[lzw_code] = lzw_prev << 8 | lzw_curr

                lzw_code += 1
                lzw_bits += 9

                if lzw_bits >= 12:
                    lzw_buf |= lzw_curr << (lzw_bits - 12)

                    out.write(bytes([
                        ((lzw_buf >> 21) & 0x7f),
                        ((lzw_buf >> 14) & 0x7f),
                        ((lzw_buf >>  7) & 0x7f),
                        0x7f ^ (lzw_buf & 0x7f)]))

                    lzw_buf &= ~((1 << (lzw_bits - 12)) - 1)
                    lzw_bits -= 12

            elif lzw_next!= lzw_curr:
                lzw_table[lzw_code] = lzw_prev << 8 | lzw_curr

                lzw_code += 1
                lzw_bits += 9

                if lzw_bits >= 12:
                    lzw_buf |= lzw_curr << (lzw_bits - 12)

                    out.write(bytes([
                        ((lzw_buf >> 21) & 0x7f),
                        ((lzw_buf >> 14) & 0x7f),
                        ((lzw_buf >>  7) & 0x7f),
                        0x7f ^ (lzw_buf & 0x7f)]))

                    lzw_buf &= ~((1 << (lzw_bits - 12)) - 1)
                    lzw_bits -= 12

            lzw_prev = lzw_curr

        return True


    for byte in data:
        lzw_next = byte

        if lzw_next in lzw_table:
            lzw_curr = lzw_next

        else:
            lzw_flush()

            lzw_prev = lzw_next
            lzw_curr = lzw_next


        if lzw_code == 511:
            lzw_flush()

            lzw_prev = None
            lzw_curr = None
            lzw_next = None

            lzw_code = 256
            lzw