#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + int.from_bytes(rng.read(1), 'big') % 256
    height = 1 + int.from_bytes(rng.read(1), 'big') % 256
    flags = b'\x07\x00' if (width == 1 and height == 1) else b'\x00\x00'
    bgColorIndex = int.from_bytes(rng.read(1), 'big') & 0xf
    pixelAspectRatio = int.from_bytes(rng.read(1), 'big') & 0xf
    out.write(struct.pack('<HHBBB',
        width, height,
        7 << 4 | len(flags),
        bgColorIndex,
        pixelAspectRatio))

    # Global color table
    globalColorTableSize = min((int.from_bytes(rng.read(1), 'big') & 0x7) + 1, 2**len(flags))
    for i in range(globalColorTableSize):
        r = int.from_bytes(rng.read(1), 'big') & 0xff
        g = int.from_bytes(rng.read(1), 'big') & 0xff
        b = int.from_bytes(rng.read(1), 'big') & 0xff
        out.write(struct.pack('BBB',
            r, g, b))

    # Image descriptor
    leftPos = int.from_bytes(rng.read(1), 'big') & 0xff
    topPos = int.from_bytes(rng.read(1), 'big') & 0xff
    imageWidth = 1 + int.from_bytes(rng.read(1), 'big') % 256
    imageHeight = 1 + int.from_bytes(rng.read(1), 'big') % 256
    flags = b'\x80' if (imageWidth == 1 and imageHeight == 1) else b'\x00'
    localColorTableFlag = bytes([int.from_bytes(rng.read(1), 'big')])[0] >> 7
    interlaceFlag = bytes([int.from_bytes(rng.read(1), 'big')])[0] >> 6
    sortFlag = bytes([int.from_bytes(rng.read(1), 'big')])[0] >> 5
    reserved = bytes([int.from_bytes(rng.read(1), 'big')])[0] & 0x1f
    lctSize = min((int.from_bytes(rng.read(1), 'big') & 0x7) + 1, 2**localColorTableFlag)
    out.write(struct.pack('<HHBBBBBB',
        leftPos, topPos,
        imageWidth, imageHeight,
        flags[0],
        lctSize,
        localColorTableFlag << 7 | interlaceFlag << 6 | sortFlag << 5 | reserved))

    # Local color table
    for i in range(lctSize):
        r = int.from_bytes(rng.read(1), 'big') & 0xff
        g = int.from_bytes(rng.read(1), 'big') & 0xff
        b = int.from_bytes(rng.read(1), 'big') & 0xff
        out.write(struct.pack('BBB',
            r, g, b))

    # Table based image data
    while True:
        blockType = rng.read(1)[0]

        if blockType == 0x2c:
            # Image descriptor
            leftPos = int.from_bytes(rng.read(1), 'big') & 0xff
            topPos = int.from_bytes(rng.read(1), 'big') & 0xff
            imageWidth = 1 + int.from_bytes(rng.read(1), 'big') % 256
            imageHeight = 1 + int.from_bytes(rng.read(1), 'big') % 256
            flags = b'\x80' if (imageWidth == 1 and imageHeight == 1) else b'\x00'
            localColorTableFlag = bytes([int.from_bytes(rng.read(1), 'big')])[0] >> 7
            interlaceFlag = bytes([int.from_bytes(rng.read(1), 'big')])[0] >> 6
            sortFlag = bytes([int.from_bytes