#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + rng.randint(0x7F)
    height = 1 + rng.randint(0x7F)
    flags = (width & 0x80) | ((height & 0x80) >> 7)
    out.write(struct.pack('<HHBBB',
        width, height,
        0xF0 | rng.randint(0x07),
        0, flags))

    # Global color table
    for _ in range(256):
        r = rng.randint(0xFF)
        g = rng.randint(0xFF)
        b = rng.randint(0xFF)
        out.write(bytes([r, g, b]))

    # Image descriptor
    left = 0
    top = 0
    width = 1 + rng.randint(0x7F)
    height = 1 + rng.randint(0x7F)
    flags = (left & 0x80) | ((top & 0x80) >> 7) \
            | ((width & 0x80) >> 1) | ((height & 0x80) >> 8)
    out.write(struct.pack('<HHHHB',
        left, top, width, height, flags))

    # LZW minimum code size
    mincodesize = rng.randint(0x02)
    out.write(bytes([mincodesize]))

    # Frame data
    lzw_clear_code = 1 << mincodesize
    lzw_end_of_information_code = lzw_clear_code + 1
    lzw_next_code = lzw_end_of_information_code + 1
    lzw_max_code_size = 1 << mincodesize
    lzw_dictionary = [None] * (lzw_max_code_size + 2)
    lzw_dictionary[lzw_clear_code] = b''
    lzw_dictionary[lzw_end_of_information_code] = None
    lzw_code_size = mincodesize + 1
    lzw_buffer = bytes()
    while True:
        if len(lzw_buffer) < lzw_code_size:
            lzw_buffer += rng.randbytes(1)

        current_byte = ord(lzw_buffer[:1])
        lzw_buffer = lzw_buffer[1:]

        if current_byte == 0x00 or len(lzw_dictionary) >= 4096:
            out.write(bytes([current_byte]))

            if current_byte == 0x00:
                break

            continue

        if not lzw_dictionary[current_byte]:
            lzw_dictionary[current_byte] = bytes([current_byte])
        else:
            new_entry = lzw_dictionary[current_byte][-1:] + lzw_dictionary[current_byte][:1]
            lzw_dictionary.append(new_entry)

        out.write(lzw_dictionary[current_byte])

        if len(lzw_dictionary) - 1 == lzw_next_code:
            lzw_next_code <<= 1
            lzw_code_size += 1

        if lzw_next_code > lzw_max_code_size and lzw_code_size!= 12:
            lzw_max_code_size <<= 1
            lzw_code_size = lzw_code_size + 1
            lzw_next_code = lzw_end_of_information_code + 1
            lzw_dictionary = [None] * (lzw_max_code_size + 2)
            lzw_dictionary[lzw_clear_code] = b''
            lzw_dictionary[lzw_end_of_information_code] = None

