#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # 16-bit width and height
    w, h = struct.unpack('>HH', rng.read(4))

    # Global color table (0 means no global color table)
    gct = int.from_bytes(rng.read(1), 'big') & 127

    # Color resolution (bits per pixel)
    cr = int.from_bytes(rng.read(1), 'big') & 7

    # Sort flag (sort colors in palette by importance)
    sf = bool(int.from_bytes(rng.read(1), 'big'))

    # Size of local color table (0 means use maximum size allowed)
    lct_size = int.from_bytes(rng.read(1), 'big') & 127

    # Background color index
    bg = int.from_bytes(rng.read(1), 'big')

    # Pixel aspect ratio
    pa = int.from_bytes(rng.read(1), 'big') & 15

    # Write header
    out.write(header + bytes([w]) + bytes([h]))

    # Write logical screen descriptor
    out.write(struct.pack('>BHB', gct, cr, sf))
    out.write(bytes([lct_size]))
    out.write(bytes([bg]))
    out.write(bytes([pa]))

    # Write global color table if applicable
    if gct == 1:
        for i in range(0, 3 * pow(2, cr)):
            r, g, b = [
                int.from_bytes(rng.read(1), 'big') % 256
                for _ in range(0, 3)]

            out.write(bytes([r, g, b]))

    # Write image descriptor
    out.write(b'\x2c')
    out.write(bytes([0, 0, w, h]))

    # Write local color table
    for i in range(0, 3 * pow(2, cr)):
        r, g, b = [
            int.from_bytes(rng.read(1), 'big') % 256
            for _ in range(0, 3)]

        out.write(bytes([r, g, b]))

    # Write LZW minimum code size
    out.write(b'\x80')

    # Generate LZW data stream
    while True:
        # Read next byte from /dev/urandom
        b = int.from_bytes(rng.read(1), 'big')

        # If this is an end marker, break
        if b == 0xff:
            break

        # If this is not an end marker, write it to output
        else:
            out.write(bytes([b]))

