#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    palette = [struct.pack('BBB', *rng.read(3)) for _ in range(0x100)]
    width = int.from_bytes(rng.read(2), byteorder='little') % (64*512) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (64*512) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    background = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect_ratio = bytes([int.from_bytes(rng.read(1), byteorder='little')])

    out.write(header)
    out.write(b'\x00\x00')
    out.write(width.to_bytes(2, byteorder='little'))
    out.write(height.to_bytes(2, byteorder='little'))
    out.write(flags)
    out.write(background)
    out.write(aspect_ratio)

    # Write global color table
    for i in range(len(palette)):
        if len(palette[i]) == 3:
            out.write(palette[i] + b'\x00')
        else:
            assert len(palette[i]) == 4 and palette[i][-1] == 0
            out.write(palette[i])

    # Write frames
    frame_count = int.from_bytes(rng.read(1), byteorder='little') % 100 + 1
    for f in range(frame_count):

        left = int.from_bytes(rng.read(2), byteorder='little') % (64*512)
        top = int.from_bytes(rng.read(2), byteorder='little') % (64*512)
        width = int.from_bytes(rng.read(2), byteorder='little') % (64*512) + 1
        height = int.from_bytes(rng.read(2), byteorder='little') % (64*512) + 1
        flags = bytes([int.from_bytes(rng.read(1), byteorder='little')])

        lct = False
        interlace = False
        sort = False
        local_color_table = False
        reserved = False
        size = None
        bgnd_index = None
        pixel_aspect_ratio = None

        if flags & 0x80!= 0:
            lct = True
            size = int.from_bytes(rng.read(1), byteorder='little') % 0x100
        if flags & 0x40!= 0:
            interlace = True
        if flags & 0x20!= 0:
            sort = True
        if flags & 0x10!= 0:
            local_color_table = True
        if flags & 0x08!= 0:
            reserved = True
        if flags & 0x07 == 0:
            pass
        elif flags & 0x07 == 1:
            bgnd_index = int.from_bytes(rng.read(1), byteorder='little') % 0x100
        elif flags & 0x07 == 2:
            bgnd_index = int.from_bytes(rng.read(2), byteorder='little') % 0x10000
        elif flags & 0x07 == 3:
            bgnd_index = int.from_bytes(rng.read(3), byteorder='little') % 0x1000000
        elif flags & 0x07 == 4:
            bgnd_index = int.from_bytes(rng.read(4), byteorder='little') % 0x100000000
        else:
            raise Exception("Invalid flag")
        if flags & 0x04!= 0:
            pixel_aspect_ratio = int.from_bytes(rng.read(1), byteorder='little')

        if lct:
            for i in range(size):
                out.write(palette[i])

        out.write(left.to_bytes(2, byteorder='little'))
        out.write(top.to_bytes(2, byteorder='little'))
        out.write(width.to_bytes(2, byteorder='little'))
        out.write(height.to_bytes(2, byteorder='little'))
        out.write(flags)

        if lct:
            out