#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed = bytearray()
    packed[0] = (width[0] & 0b11111111) | ((height[0] & 0b11111111) << 8)
    packed[1] = (width[1]) | ((height[1]) << 8)
    packed[2] = rng.read(1)[0]
    packed[3] = 0b10000000
    packed[4] = rng.read(1)[0]
    out.write(packed)

    # Write the global color table
    palette = [None] * 256
    for i in range(0, 256):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        palette[i] = (r, g, b)

    out.write(bytes([(len(palette))]))
    for p in palette:
        out.write(struct.pack('BBB', p[0], p[1], p[2]))

    # Write the image descriptor
    x = rng.read(2)
    y = rng.read(2)
    w = rng.read(2)
    h = rng.read(2)

    packed = bytearray()
    packed[0] = (x[0] & 0b11111111) | ((y[0] & 0b11111111) << 8)
    packed[1] = (w[0] & 0b11111111) | ((h[0] & 0b11111111) << 8)
    packed[2] = (x[1]) | ((y[1]) << 8)
    packed[3] = (w[1]) | ((h[1]) << 8)
    packed[4] = 0b10000000
    packed[5] = rng.read(1)[0]
    out.write(packed)

    # Write the local color table
    if len(palette) > 256:
        raise Exception("Too many colors")

    out.write(bytes([(len(palette))]))
    for p in palette:
        out.write(struct.pack('BBB', p[0], p[1], p[2]))

    # Write the LZW code size and clear code
    out.write(b'\x08\x00')

    # Write the data stream
    while True:
        # Generate a random length between 2 and 127 inclusive
        lzw_length = rng.read(1)[0] % 127 + 2;

        # Read that amount of bytes from /dev/urandom
        buf = rng.read(lzw_length);

        # If we have reached the end of the image then write an EOI marker
        if not buf or buf == b'\xff':
            break

        # Otherwise write this as normal
        out.write(buf)

