#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(out: BinaryIO, rng: bytes):
    # Header
    out.write(b"GIF89a")

    # Logical screen descriptor
    width = 1 + int.from_bytes(rng[0:2], "big", signed=False) % 50
    height = 1 + int.from_bytes(rng[2:4], "big", signed=False) % 50
    packed = b""
    packed += ((width & 0x07) << 4).to_bytes(1, byteorder="little", signed=False)
    packed += (((height >> 3) & 0x07) << 4).to_bytes(
        1, byteorder="little", signed=False
    )
    packed += b"\x00\x00"
    packed += b"\x00"
    packed += b"\x00"
    out.write(packed)

    # Global color table
    palette = [
        (int.from_bytes(rng[i : i + 3], "big", signed=False) & 0xFFFFFF)
        for i in range(0, 3 * 256, 3)
    ]
    for p in palette:
        out.write((p & 0xFF).to_bytes(1, byteorder="little"))
        out.write(((p >> 8) & 0xFF).to_bytes(1, byteorder="little"))
        out.write(((p >> 16) & 0xFF).to_bytes(1, byteorder="little"))

    # Image descriptor
    packed = b""
    packed += ((0 & 0x07) << 4).to_bytes(1, byteorder="little", signed=False)
    packed += (((0 >> 3) & 0x07) << 4).to_bytes(
        1, byteorder="little", signed=False
    )
    packed += b"\x00\x00"
    packed += b"\x00"
    packed += b"\x00"
    out.write(packed)

    # Local color table
    out.write(b"\x00" * 3 * 256)

    # Graphic control extension
    packed = b""
    packed += b"\x21"
    packed += b"\xF9"
    packed += b"\x04"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    out.write(packed)

    # Application extension
    packed = b""
    packed += b"\x21"
    packed += b"\xFF"
    packed += b"\x0B"
    packed += b"NETSCAPE2.0"
    packed += b"\x03"
    packed += b"\x01"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    out.write(packed)

    # Comment extension
    comment = b"Comment"
    out.write(b"\x21")
    out.write(len(comment).to_bytes(1, byteorder="little"))
    out.write(comment)

    # Plain text extension
    plain_text = b"Plain Text Extension"
    out.write(b"\x21")
    out.write(len(plain_text).to_bytes(1, byteorder="little"))
    out.write(plain_text)

    # Graphics Control Extension block
    packed = b""
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    out.write(packed)

    # Plain text data
    packed = b""
    packed += b"\x01"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x00"
    packed += b"\x