#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = struct.pack('>H', 1024)
    height = struct.pack('>H',  768)
    flags  = struct.pack('B',   0x80 | 0x40 | 0x20)
    bg     = struct.pack('B',   0xff)
    aspect = struct.pack('B',   0x00)
    out.write(width + height + flags + bg + aspect)

    # Global color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f

        if ((r == 0) and (g == 0) and (b == 0)):
            r += 1
            g += 1
            b += 1

        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left      = struct.pack('>H', 0)
    top       = struct.pack('>H', 0)
    width     = struct.pack('>H', 1024)
    height    = struct.pack('>H',  768)
    flags     = struct.pack('B',   0x00 | 0x08 | 0x04)
    trans_idx = struct.pack('B',   0x00)
    out.write(left + top + width + height + flags + trans_idx)

    # LZW minimum code size
    min_code_size = struct.pack('B', 0xe0)
    out.write(min_code_size)

    # Encode image data
    lzw_table = [0] * 256
    lzw_next  = 1
    lzw_clear = False
    lzw_eoi   = False
    while not lzw_eoi:
        if lzw_clear:
            lzw_clear = False

            # Clear LZW table
            for i in range(0, 256):
                lzw_table[i] = lzw_next

                lzw_next += 1
                if lzw_next >= 256:
                    lzw_next = 256 + 1

            # Write clear code to output stream
            out.write(bytes([lzw_table[0]]))

        # Read next byte from input stream
        c = ord(rng.read(1))

        # Check for EOI marker
        if c == 0x3b:
            lzw_eoi = True
            continue

        # Check for CLEAR marker
        elif c == 0x2c:
            lzw_clear = True
            continue

        # Check for END of information code
        elif c == 0x00:
            # Write END OF INFORMATION code to output stream
            out.write(bytes([lzw_table[0]]))
            continue

        # Get previous sequence length
        prev_len = len(lzw_table) - 1

        # Find longest prefix match
        seq_len = 1
        while (seq_len <= prev_len) and \
              (lzw_table[(seq_len-1)]!= c):
            seq_len += 1

        # Output current character as literal value
        if seq_len > 1:
            out.write(bytes([lzw_table[(seq_len-1)]]))

        # Add new entry to LZW table
        lzw_table[prev_len] = lzw_next
        lzw_next += 1
        if lzw_next >= 256:
            lzw_next = 256 + 1

        # If we have reached the end of our LZW table...
        if prev_len+1 == len(lzw_table):
            # Increase LZW table size by one
            lzw_table.append(-1)


    # Write EOI marker to output stream
    eoi = bytes([0x00])
    out.write(eoi)


# Main function
if __name__ == '__main__':
    try:
        with open('/dev/urandom', 'rb') as f:
            generate_random_gif(f, sys.stdout.buffer)
    except Exception as ex:
        print("Error generating random GIF:", str(ex), file=sys.stderr)
    return True


