#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# returns its size in bytes.
def generate_random_gif(out: BinaryIO, rng):
    # Header (13 bytes).
    out.write(b'GIF89a')

    # Logical screen descriptor (7 bytes).
    width = 2**rng.randint(5, 10) - 1
    height = 2**rng.randint(5, 10) - 1
    flags = b'\x87\x00' if rng.choice([True, False]) else b'\x07\x00'
    bg_color_index = rng.randint(0, 255)
    pixel_aspect_ratio = b'\x00'
    out.write(struct.pack('<HHBBB',
        width, height, flags[0], flags[1],
        bg_color_index))
    out.write(pixel_aspect_ratio)

    # Global color table (3 * 256 + 1 bytes).
    for i in range(256):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write(struct.pack('BBB', r, g, b))
    out.write(b'\x00')

    # Image descriptor (10 bytes).
    left = 0
    top = 0
    width = 2**rng.randint(5, 10) - 1
    height = 2**rng.randint(5, 10) - 1
    flags = b'\x87\x00' if rng.choice([True, False]) else b'\x07\x00'
    local_color_table_flag = b'\x01'
    interlace_flag = b'\x00'
    sort_flag = b'\x00'
    reserved = b'\x00\x00'
    local_color_table_size = b'\xff'
    out.write(struct.pack('<HHHHBBBBB',
        left, top, width, height, flags[0], flags[1],
        local_color_table_flag[0], local_color_table_flag[1]))
    out.write(interlace_flag)
    out.write(sort_flag)
    out.write(reserved)
    out.write(local_color_table_size)

    # Local color table (3 * 256 + 1 bytes).
    for i in range(256):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write(struct.pack('BBB', r, g, b))
    out.write(b'\x00')

    # LZW minimum code size (1 byte).
    lzw_min_code_size = rng.randint(2, 12)
    out.write(bytes([lzw_min_code_size]))

    # Image data.
    image_data = []
    while len(image_data) < width * height:
        # Generate a new command.
        cmd = rng.randint(0, 4)

        # RLE-encoded pixels.
        if cmd == 0:
            count = min(width * height - len(image_data), rng.randint(3, 12))

            # Repeated pixel.
            if rng.choice([True, False]):
                value = rng.randint(0, 255)
                image_data += [value] * count

                # Escape sequence.
                if rng.choice([True, False]):
                    image_data += [value]

            # Raw pixels.
            else:
                image_data += list(rng.getrandbits(8).to_bytes(count, 'big'))

        # Plain text.
        elif cmd == 1:
            length = min(width * height - len(image_data), rng.randint(3, 12))
            image_data += list(rng.getrandbits(length * 8).to_bytes(length, 'big'))

        # LZW encoded pixels.
        elif cmd == 2:
            length = min(width * height - len(image_data), rng.randint(3, 12))
            image_data += list(rng.getrandbits(length * 8).to_bytes(length, 'big'))

        # EOF.
        elif cmd == 3:
            break

        # Invalid command.
        else:
            raise Exception()

    #