#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The size of the output is determined by the first 10 bytes read from `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Read the header and determine the length of the image data
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')

    # Write the header to the output
    out.write(header)

    # Copy the image data in chunks of 256 bytes at a time
    while datalen > 0:
        chunklen = min(datalen, 256)

        # Generate a random color palette for this chunk
        colors = [
                (
                    rng.read(1)[0],
                    rng.read(1)[0],
                    rng.read(1)[0]
                )
                for _ in range(768)]

        # Write the chunk's LZW-compressed data
        lzwdata = []
        prevcolor = -1
        while len(lzwdata) < chunklen:

            # If we're not yet at the end of the chunk...
            if len(lzwdata) < chunklen:

                # Get a new code from the input stream
                code = int.from_bytes(
                        rng.read(1), byteorder='big', signed=False)

                # If it's a clear code, reset the dictionary
                if code == 256:

                    # Add an empty string as the last entry
                    lzwdict[256] = ''

                    # Add all the colors to the dictionary
                    for i, color in enumerate(colors):

                        # Convert RGB values to a single integer
                        rgbint = ((color[0] << 16) |
                                  (color[1] <<  8) |
                                   color[2])

                        # Add the color to the dictionary with its index
                        lzwdict[i + 1] = chr(rgbint)

                    # Start a new block of compressed data
                    lzwblock = []

                    continue

                elif code >= 257:

                    # Look up the previous character in the dictionary
                    char = lzwdict[prevcode] + \
                           lzwdict[code][0]

                else:

                    # Just look up the character directly
                    char = lzwdict[code]

                # Append the character to our current block
                lzwblock += char

                # Update the previous character
                prevcode = code

            # We've reached the end of the chunk; compress the block
            else:

                # Compress the block
                lzwsymbols = ''.join([
                        str(lzwdict.index(c))
                        for c in lzwblock]).encode('ascii')

                # Write the compressed data to the output
                out.write(struct.pack('<H', len(lzwsymbols)))
                out.write(lzwsymbols)

                # Reset the block
                lzwblock = []

        # Decrement the remaining length of the image data
        datalen -= chunklen


