#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The size of the output is determined by the first 10 bytes read from `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Read the header and determine the length of the image data
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')

    # Write the header to the output
    out.write(header)

    # Copy the image data in chunks of 256 bytes at a time
    while datalen > 0:

        # Determine how many bytes we can copy this iteration
        chunklen = min(datalen, 256)

        # Generate some random pixels for each color channel
        red = rng.read(chunklen)
        green = rng.read(chunklen)
        blue = rng.read(chunklen)

        # Create an array of pixel tuples (R, G, B)
        pixels = [
                tuple([
                    r, g, b])
                for r, g, b in zip(red, green, blue)]

        # Convert the list of pixels to binary data
        pixeldata = b''
        for p in pixels:

            # Each pixel is represented as three bytes:
            # - one for the red value
            # - one for the green value
            # - one for the blue value
            pixeldata += struct.pack('BBB', *p)

        # Write the pixel data to the output stream
        out.write(pixeldata)

        # Update the remaining amount of data that needs to be copied
        datalen -= chunklen