#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The size of the output is determined by the first 10 bytes read from `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Read the header and determine the length of the image data
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')

    # Write the header to the output
    out.write(header)

    # Generate a palette with 256 colors
    pal = [struct.pack('BBB', *rng.read(3)) for _ in range(256)]

    # Write the palette to the output
    out.write(b''.join(pal))

    # Copy the image data in chunks of 256 bytes at a time
    while datalen > 0:
        chunklen = min(256, datalen)

        # Generate an LZW code table if this is the first iteration
        if 'codetable' not in locals():
            codetable = []

            # Add all single-byte codes (0x00 through 0xFF) to the table
            for i in range(256):
                codetable.append([i])

            # Add all two-byte codes (0x0000 through 0xFFFF) to the table
            for i in range(256):
                for j in range(256):
                    codetable.append([(j << 8) | i])

            # Add all three-byte codes (0x000000 through 0xFFFFFF) to the table
            for i in range(256):
                for j in range(256):
                    for k in range(256):
                        codetable.append([(k << 16) | (j << 8) | i])

        # Read one byte from the input
        c = ord(rng.read(1)) & 0xff

        # If it's a clear code, reset the code table
        if c == 0x00:
            codetable = []

            # Add all single-byte codes (0x00 through 0xFF) to the table
            for i in range(256):
                codetable.append([i])

            # Add all two-byte codes (0x0000 through 0xFFFF) to the table
            for i in range(256):
                for j in range(256):
                    codetable.append([(j << 8) | i])

            # Add all three-byte codes (0x000000 through 0xFFFFFF) to the table
            for i in range(256):
                for j in range(256):
                    for k in range(256):
                        codetable.append([(k << 16) | (j << 8) | i])

        # Otherwise, look up the next code in the table
        else:
            # Look up the last code in the table
            lastcode = codetable[-1]

            # If there are no more codes left in the current table, add another
            if len(lastcode) >= 4:
                raise Exception("Code overflow")

            # Append the new code to the end of the last code
            lastcode.append(c)

            # If the last code has reached its maximum length, add it back to the table
            if len(lastcode) == 4:
                codetable.append(lastcode)

                # If we've added too many codes, remove some
                if len(codetable) > 4096:
                    del codetable[0]

        # Find the index of the last code in the table that matches ours
        matchidx = -1
        for idx, entry in enumerate(codetable):
            if entry == lastcode:
                matchidx = idx
                break

        # Write the matching index as a variable-length integer
        varint = []
        while True:
            varint.insert(0, matchidx % 0x80)
            matchidx >>= 7
            if matchidx == 0:
                break
        out.write(bytes(varint + [0]))

        datalen -= chunklen