#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The size of the output is determined by the first 10 bytes read from `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Read the header and determine the length of the image data
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')

    # Write the header to the output
    out.write(header)

    # Generate the palette
    palsize = (datalen // 4 + 1) * 4
    palette = [struct.unpack('BBB', rng.read(3)) for _ in range(palsize)]

    # Write the palette to the output
    for color in palette:
        out.write(b'\x00\x00\x00')

    # Write the image data in chunks of 256 bytes at a time
    while datalen > 0:
        chunklen = min(256, datalen)

        # Write the chunk length as a LZW minimum code size
        out.write((chunklen - 1).to_bytes(1, 'little'))

        # Write the chunk itself
        chunk = rng.read(chunklen)
        out.write(chunk)

        # Pad the chunk with zeroes if necessary
        padding = ((chunklen + 1) % 2 == 0)
        if padding:
            out.write(b'\x00')

        # Decrement the remaining data length
        datalen -= chunklen