#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write(b'GIF89a')
    
    # Logical Screen Descriptor (LSD)
    width = 16 + rng.read(2)[0] % 15
    height = 16 + rng.read(2)[0] % 15
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio))
    
    # Global Color Table
    color_table_size = 7 + rng.read(1)[0] % 7
    color_resolution = 1 + rng.read(1)[0] % 7
    sort_flag = rng.read(1)[0] & 1
    global_color_table = []
    for i in range(3 ** color_table_size):
        red = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue = rng.read(1)[0] % 256
        global_color_table.append((red, green, blue))
    out.write(struct.pack('B', (color_table_size << 4) | (color_resolution << 2) | sort_flag))
    for red, green, blue in global_color_table:
        out.write(struct.pack('BBB', red, green, blue))
        
    # Application Extension Block
    application_identifier = b'Python'
    application_authentication_code = b'RANDOM'
    data_string = b''
    while len(data_string) < 255 - 11 - len(application_identifier) - len(application_authentication_code):
        data_string += rng.read(1)
    out.write(struct.pack('<H', 11 + len(application_identifier) + len(application_authentication_code)))
    out.write(b'\x21\xff')
    out.write(struct.pack('<H', len(application_identifier)))
    out.write(application_identifier)
    out.write(struct.pack('<H', len(application_authentication_code)))
    out.write(application_authentication_code)
    out.write(struct.pack('B', len(data_string)))
    out.write(data_string)
    
    # Image Descriptor
    left = rng.read(2)[0] % width
    top = rng.read(2)[0] % height
    width = rng.read(2)[0] % (width - left)
    height = rng.read(2)[0] % (height - top)
    local_color_table_flag = rng.read(1)[0] & 1
    interlaced_flag = rng.read(1)[0] & 1
    sorted_flag = rng.read(1)[0] & 1
    reserved_bits = rng.read(1)[0] & 7
    if local_color_table_flag == 0:
        local_color_table_size = 0
        local_color_table = None
    else:
        local_color_table_size = 7 + rng.read(1)[0] % 7
        local_color_table = []
        for i in range(3 ** local_color_table_size):
            red = rng.read(1)[0] % 256
            green = rng.read(1)[0] % 256
            blue = rng.read(1)[0] % 256
            local_color_table.append((red, green, blue))
    out.write(struct.pack('<HHBBBBBB', left, top, width, height, local_color_table_flag, interlaced_flag, sorted_flag, reserved_bits))
    
    # Local Color Table
    if local_color_table is not None:
        out.write(struct.pack('B', (local_color_table_size