#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate and write the logical screen descriptor
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    bg     = bytes([int.from_bytes(rng.read(1), 'big')])
    pxsize = bytes([int.from_bytes(rng.read(1), 'big')])
    color  = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write((width).to_bytes(2,'big'))
    out.write((height).to_bytes(2,'big'))
    out.write(flags)
    out.write(bg)
    out.write(pxsize)
    out.write(color)

    # Generate and write the global color table
    for i in range(pow(2, int.from_bytes(color, 'big'))):
        r = int.from_bytes(rng.read(1),'big')
        g = int.from_bytes(rng.read(1),'big')
        b = int.from_bytes(rng.read(1),'big')
        out.write(bytes([r,g,b]))

    # Generate and write each frame
    while True:

        # Generate the frame header
        left   = int.from_bytes(rng.read(1), 'big')
        top    = int.from_bytes(rng.read(1), 'big')
        width  = int.from_bytes(rng.read(1), 'big')
        height = int.from_bytes(rng.read(1), 'big')
        flags  = bytes([0x80 | int.from_bytes(rng.read(1), 'big')])
        delay  = int.from_bytes(rng.read(2), 'big')
        transp = bytes([int.from_bytes(rng.read(1), 'big')])
        out.write(left.to_bytes(2,'big'))
        out.write(top.to_bytes(2,'big'))
        out.write(width.to_bytes(2,'big'))
        out.write(height.to_bytes(2,'big'))
        out.write(flags)
        out.write(delay.to_bytes(2,'big'))
        out.write(transp)

        # Generate and write the local color table
        for i in range(pow(2, int.from_bytes(color, 'big'))):
            r = int.from_bytes(rng.read(1),'big')
            g = int.from_bytes(rng.read(1),'big')
            b = int.from_bytes(rng.read(1),'big')
            out.write(bytes([r,g,b]))

        # Read the image data
        blocklen = int.from_bytes(rng.read(1), 'big')
        if not blocklen: break
        blockdata = rng.read(blocklen)

        # Write the block length
        out.write(blocklen.to_bytes(1,'big'))

        # Encode the block with LZW
        encoded = []
        prev = ''
        for c in blockdata:

            # Get the next character to encode
            char = chr(c)

            # If we have no previous character or this is a new one...
            if len(prev)==0 or char!= prev[-1]:
                encoded += [prev]

                # Add the current character as the first entry
                encoded += [[char]]
            else:
                # Append it to the last entry
                encoded[-1].append(char)

            # Update the previous character
            prev = ''.join(encoded[-1])

        # Write the encoded characters
        for e in encoded:
            out.write(bytes([len(e)]))
            for c in e:
                out.write(bytes([ord(c)]))