#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    header = b'GIF89a'

    # 16-bit width and height
    w = rng.read(2)
    h = rng.read(2)

    # 8-bit background color index
    bgcolor = rng.read(1)

    # 8-bit pixel aspect ratio
    par = rng.read(1)

    # 7-bit global color table flag (0 or 1),
    # 3-bit color resolution,
    # 1-bit sort flag (0 or 1),
    # 3-bit size of global color table
    gct_flag = rng.read(1)[0] >> 7
    gct_size = ((gct_flag << 5) | (par[0] & 0b00011100)) + 1
    gct_sort = (par[0] >> 4) & 1
    gct_res = (par[0] >> 1) & 0b111

    # Global color table
    gct = []
    for _ in range(gct_size):
        gct.append(struct.unpack('BBB', rng.read(3)))

    # Graphic control extension
    gce = [
        0x21,                   # Extension Introducer
        0xf9,                   # Graphic Control Label
        0x04,                   # Block Size
        0x00,                   # Reserved
        0x00,                   # Disposal Method
        0xff,                   # User Input Flag
        0x00,                   # Transparent Color Index
        0x00,                   # Block Terminator
    ]

    # Image descriptor
    idesc = [
        0x2c,                   # Image Separator
        0x00,                   # Image Left Position
        0x00,                   # Image Top Position
        *w,                     # Image Width
        *h,                     # Image Height
        0x00,                   # Local Color Table Flag
        0x00,                   # Interlace Flag
        0x00,                   # Sort Flag
        0x00,                   # Reserved
        0x00,                   # LCT Size
    ]

    # Pixel data
    pixels = []
    for y in range(int.from_bytes(h, 'big')):
        for x in range(int.from_bytes(w, 'big')):
            pixels.extend([*rng.read(3)])

    # Output everything
    out.write(header)
    out.write(w)
    out.write(h)
    out.write(bgcolor)
    out.write(par)
    out.write(bytes([gct_flag << 7 | gct_res << 4 | gct_sort << 3 | gct_size - 1]))
    for rgb in gct:
        out.write(bytes(rgb))
    out.write(bytes(gce))
    out.write(bytes(idesc))
    out.write(bytes(pixels))
