#!/usr/bin/env python3

import struct
from typing import BinaryIO, List, Optional, Tuple

from . import (
    _decode_varint,
    _encode_varint,
    _read_bytes,
    _write_bytes,
)


def read_message(f: BinaryIO) -> bytes:
    length = _decode_varint(f)
    return _read_bytes(f, length)


def write_message(f: BinaryIO, message: bytes):
    _write_bytes(f, _encode_varint(len(message)))
    _write_bytes(f, message)


def read_messages(f: BinaryIO) -> List[bytes]:
    messages = []
    while True:
        try:
            message = read_message(f)
        except EOFError:
            break
        else:
            messages.append(message)
    return messages


def write_messages(f: BinaryIO, messages: List[bytes]):
    for message in messages:
        write_message(f, message)


def read_header(f: BinaryIO) -> Tuple[str, int]:
    magic = f.read(4).decode("ascii")
    if magic != "BMTP":
        raise ValueError(f"Invalid magic {magic}")
    version = struct.unpack("<I", f.read(4))[0]
    return magic, version


def write_header(f: BinaryIO, magic: str, version: int):
    assert len(magic) == 4
    f.write(magic.encode("ascii"))
    f.write(struct.pack("<I", version))


def read_request(f: BinaryIO) -> Tuple[Optional[str], Optional[str]]:
    command = None
    if not f.read(1):
        command = read_message(f).decode()
    service = None
    if not f.read(1):
        service = read_message(f).decode()
    return command, service


def write_request(f: BinaryIO, command: Optional[str], service: Optional[str]):
    if command is None:
        f.write(b"\x00")
    else:
        write_message(f, command.encode())
    if service is None:
        f.write(b"\x00")
    else:
        write_message(f, service.encode())


def read_response(f: BinaryIO) -> bool:
    return bool(f.read(1)[0])


def write_response(f: BinaryIO, success: bool):
    f.write(b"\x01" if success else b"\x00")


def read_error(f: BinaryIO) -> Exception:
    error_type = read_message(f).decode()
    error_msg = read_message(f).decode()
    return globals()[error_type](error_msg)


def write_error(f: BinaryIO, error: Exception):
    write_message(f, type(error).__name__.encode())
    write_message(f, str(error).encode())

