#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
	# Header
	out.write(b'GIF89a')

	# Logical screen descriptor
	width = 1 + int.from_bytes(rng.read(2), byteorder='big', signed=False) % 50
	height = 1 + int.from_bytes(rng.read(2), byteorder='big', signed=False) % 50
	packed = (
		0x87 # Global color table flag
		| ((width - 1) << 4)
		| ((height - 1) >> 4)
	).to_bytes(1, 'little')
	out.write(packed)
	out.write((height & 0xF).to_bytes(1, 'little'))
	out.write(b'\x00\x00') # Background color index
	out.write(b'\x00\x00') # Pixel aspect ratio

	# Color table
	for i in range(0, 2 ** packed[0] * 3):
		out.write(rng.read(1))

	# Image descriptor
	out.write(b'\x2C') # Image separator
	left = 0
	top = 0
	width = width if width > left else left
	height = height if height > top else top
	packed = (
		0x87 # Local color table flag
		| ((width - 1) << 4)
		| ((height - 1) >> 4)
	).to_bytes(1, 'little')
	out.write(packed)
	out.write((height & 0xF).to_bytes(1, 'little'))
	out.write(struct.pack('<HHH', left, top, 0)) # Image data position and size

	# LZW minimum code size
	out.write(b'\x08')

	# Image data
	lzw_mincodesize = 8
	lzw_clearcode = 1 << lzw_mincodesize
	lzw_endofinfo = lzw_clearcode + 1
	lzw_endofdata = lzw_endofinfo + 1
	lzw_nextcode = lzw_endofdata + 1
	lzw_maxcode = 1 << lzw_mincodesize
	lzw_bitmask = [1 << n for n in reversed(range(lzw_mincodesize))]
	lzw_table = {i : bytes([i]) for i in range(256)}
	lzw_prevbuf = b''
	lzw_curbuf = b''
	while True:
		if len(lzw_curbuf) == 0:
			lzw_curbuf += rng.read(1)

		if len(lzw_curbuf) < lzw_mincodesize or lzw_nextcode >= lzw_maxcode:
			lzw_prevbuf += lzw_curbuf[:1]
			lzw_curbuf = lzw_curbuf[1:]

			if len(lzw_curbuf) == 0:
				break

			continue

		if lzw_curbuf not in lzw_table:
			lzw_table[lzw_curbuf] = bytes([len(lzw_table)])
			lzw_prevbuf += lzw_table[lzw_prevbuf[-1:]]
			lzw_curbuf = lzw_curbuf[1:]

			if len(lzw_curbuf) == 0:
				break

			continue

		lzw_prevbuf += lzw_table[lzw_curbuf]
		lzw_curbuf = b''

		if lzw_prevbuf == lzw_table[lzw_clearcode]:
			lzw_prevbuf = b''
			lzw_curbuf = b''
			lzw_nextcode = lzw_endofdata + 1
			lzw_maxcode = 1 << lzw_mincodesize
			lzw_bitmask = [1 << n for n in reversed(range(lzw_mincodesize))]
			lzw_table = {i : bytes([i]) for i in range(256)}
			continue

		if lzw_prevbuf == lzw_table[lzw_endofinfo]:
			break

		if lzw_nextcode >= lzw_maxcode:
			lzw_prevbuf = lzw_prevbuf[:-1]
		