#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    header = b'GIF89a'

    # 16-bit width and height
    w = rng.read(2)
    h = rng.read(2)

    # packed field of color resolution (bits per pixel)
    r = rng.read(1)[0] & 7
    g = rng.read(1)[0] & 7
    b = rng.read(1)[0] & 7
    packed = bytes([r | (g << 3) | (b << 6)])

    # background color index
    bg = rng.read(1)[0]

    # aspect ratio
    ar = rng.read(1)[0] & 0xf

    # image descriptor
    out.write(header + w + h + packed +
             bytes([bg]) + bytes([ar]))

    # LZW minimum code size
    mincodesize = rng.read(1)[0] & 0x0f
    out.write(bytes([mincodesize]))

    # palette
    for i in range(256):
        out.write(struct.pack('BBB', *rng.read(3)))

    # graphics control extension
    out.write(b'\x21\xf9\x04')
    flags = rng.read(1)[0] & 0x0c
    if flags == 0:
        flags |= rng.read(1)[0] & 0x30
    delay = rng.read(2)
    transindex = rng.read(1)[0]
    out.write(flags + delay + bytes([transindex]))

    # application extension
    out.write(b'\x21\xff\x0bNETSCAPE2.0')
    appdata = b''
    while len(appdata) < 11:
        appdata += rng.read(1)
    out.write(appdata[:11])

    # comment extension
    out.write(b'\x21\xfe')
    comlen = rng.read(1)[0] % 256
    out.write(bytes([comlen]))
    out.write(rng.read(comlen))

    # plain text extension
    out.write(b'\x21\x01')
    txtlen = rng.read(1)[0] % 256
    out.write(bytes([txtlen]))
    out.write(rng.read(txtlen))

    # graphic data block
    lzwmincodesize = rng.read(1)[0] & 0x0f
    out.write(bytes([lzwmincodesize]))

    # LZW codes
    lzwcodes = []
    while True:
        c = rng.read(1)[0]

        # end of stream
        if c == 0x00:
            break

        # clear code
        elif c == 0xe0:
            lzwcodes = [None] * 256

            # end of information
            if rng.read(1)[0]!= 0x00:
                raise Exception("Invalid LZW stream")
            continue

        # end of information
        elif c == 0x01:
            break

        # literal code
        else:
            lzwcodes.append(bytes([c]))
            continue

        # length of prefix code
        plen = rng.read(1)[0]

        # get prefix code
        pfx = lzwcodes[plen - 1]

        # add suffix to prefix code
        sfx = rng.read(1)[0]
        lzwcodes.append(pfx + bytes([sfx]))

    # write compressed data
    for c in lzwcodes:
        if not c:
            out.write(b'\x00')
        else:
            out.write(c)

    # end of stream
    out.write(b'\x00')
