#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate and write the logical screen descriptor
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    bg     = bytes([int.from_bytes(rng.read(1), 'big')])
    pxsize = bytes([int.from_bytes(rng.read(1), 'big')])
    color  = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write((width).to_bytes(2,'big'))
    out.write((height).to_bytes(2,'big'))
    out.write(flags)
    out.write(bg)
    out.write(pxsize)
    out.write(color)

    # Generate and write the global color table
    for i in range(pow(2, int.from_bytes(color, 'big'))):
        r = int.from_bytes(rng.read(1),'big')
        g = int.from_bytes(rng.read(1),'big')
        b = int.from_bytes(rng.read(1),'big')
        out.write(r.to_bytes(1,'big'))
        out.write(g.to_bytes(1,'big'))
        out.write(b.to_bytes(1,'big'))

    # Generate and write the image descriptor
    left   = bytes([0x00])
    top    = bytes([0x00])
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x80 | ((int.from_bytes(rng.read(1), 'big') & 0x07)<<4)])
    trans  = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(left)
    out.write(top)
    out.write(width.to_bytes(2,'big'))
    out.write(height.to_bytes(2,'big'))
    out.write(flags)
    out.write(trans)

    # Generate and write the local color table
    if int.from_bytes(flags, 'big') & 0x80 == 0x80:
        for i in range(pow(2, int.from_bytes(pxsize, 'big'))):
            r = int.from_bytes(rng.read(1),'big')
            g = int.from_bytes(rng.read(1),'big')
            b = int.from_bytes(rng.read(1),'big')
            out.write(r.to_bytes(1,'big'))
            out.write(g.to_bytes(1,'big'))
            out.write(b.to_bytes(1,'big'))

    # Generate and write the image data
    lzwmincodesize = pow(2, int.from_bytes(pxsize, 'big'))
    lzwmaxcodesize = min(lzwmincodesize+1, 12)
    lzwsizemask = pow(2, lzwmaxcodesize)-1
    lzwclearcode = pow(2, lzwmaxcodesize)
    lzwendofinfo = lzwclearcode+1
    lzwendofdata = lzwclearcode+2
    lzwnextcode = lzwclearcode+3
    lzwprevcode = -1
    lzwcurcode = lzwclearcode+1
    lzwbitcount = lzwmincodesize+1
    lzwbuf = bytearray()
    lzwbufpos = 0
    lzwbuflen = 0
    while True:

        # If we have enough bits to read another code from the input stream...
        if lzwbitcount > 0:

            # Read one bit at a time until we get all of them
            while lzwbitcount <= 8:
                lzwbuf[lzwbufpos] |= int.from_bytes(
                    rng.read(1), 'big') << (8-lzwbitcount)
                lzwbitcount += 8
                lzwbufpos += 1

                # If we've filled up our buffer...
                if lzwbufpos >= len(lzwbuf):

                    # Flush it to the output stream
                    out.write(lzwbuf[:lzwbufpos])
                    lzwbuflen += lzwbufpos

                    # Reset the buffer position counter
                    lzwbufpos = 0

        # Get the next code from the input stream
        lzwcurcode = \
            (lzwcurcode << 1) | \
            (((lzwbuf[lzwbufpos] >> (8-(lzwbitcount))) & 0x01)!= 0)
        lzwbitcount -= 1

        # If this is the end of an encoded sequence...
        if lzwcurcode == lzwendofdata or lzwcurcode == lzwendofinfo:

            # If this was not the last code in the image...
            if lzwcurcode!= lzwendofinfo:

                # Add a zero-length block to mark the end of the image
                out.write(bytes([0]))

            break

        # If this is a clear code...
        elif lzwcurcode == lzwclearcode:

            # Reset the dictionary
            lzwdict = dict([(i, [i]) for i in range(lzwclearcode+2, 2**lzwmaxcodesize)])
            lzwdictlen = lzwclearcode+2
            lzwdictmax = 2**lzwmaxcodesize
            lzwdictsize = lzwclearcode+2
            lzwprevcode = -1
            lzwcurcode = lzwclearcode+1
            continue

        # If this is an end-of-information code...
        elif lzwcurcode == lzwendofinfo:

            # Flush any remaining buffered codes to the