#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# returns its size in bytes.
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # Header
    out.write("GIF89a".encode())

    width = rng.read(1)[0] % 256 + 1
    height = rng.read(1)[0] % 256 + 1
    flags = b"\x87" if (rng.read(1)[0] & 1) else b"\x00"
    background = rng.read(1)[0] % 256
    aspect_ratio = rng.read(1)[0] % 4 + 1
    palette = [
        (rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]) for _ in range(256)
    ]

    # Logical screen descriptor
    out.write(struct.pack("<HHBBB", width, height, flags[0], background, aspect_ratio))

    # Global color table
    out.write(b"".join([bytes((r, g, b)) for r, g, b in palette]))

    # Image descriptor
    out.write(struct.pack("<HHBBH", 0, 0, width, height, len(palette)))

    # Image data
    lzw_mincodesize = rng.read(1)[0] % 9 + 2
    out.write(lzw_mincodesize.to_bytes(1, "little"))

    def next_byte():
        return rng.read(1)[0] % 256

    codesize = lzw_mincodesize + 1
    clear_code = 1 << lzw_mincodesize
    end_of_information_code = clear_code + 1
    max_code_size = 1 << codesize
    bitmask = (1 << codesize) - 1
    current_code = None
    previous_code = None
    codes = {}
    output = []

    while True:
        # Read a byte from /dev/urandom
        c = next_byte()

        # If we're at the end of information code, stop generating more codes
        if current_code == end_of_information_code:
            break

        # If this is a new code, read another one to determine how many times it's repeated
        if not current_code or c!= previous_code:
            count = 1

            while True:
                d = next_byte()

                if d >= 128:
                    break

                count += 1

            # Add the new code to our dictionary
            current_code = ((previous_code << codesize) | c) & bitmask
            codes[current_code] = previous_code

            # If there are too many codes, reset the dictionary by clearing all codes
            if len(codes) > max_code_size:
                codes = {i: i for i in range(clear_code)}
                codesize = lzw_mincodesize + 1
                max_code_size = 1 << codesize
                bitmask = (1 << codesize) - 1

            # Write the code to the output buffer
            output.append(current_code)
        else:
            # This is an old code; look up what it means
            output.append(codes[current_code])

        # Advance the state machine
        previous_code = c

    # End of stream marker
    output.append(end_of_information_code)

    # LZW compression
    output = bytes(output)
    while len(output) > 0:
        chunk = min(len(output), 255)
        out.write(chunk.to_bytes(1, "little"))
        out.write(output[:chunk])
        del output[:chunk]

    # Trailer
    out.write(b";")

    return 13 + 7 + 7 + 3 * 256 + sum(map(lambda x: 3, palette)) + \
        1 + len(output)

