#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # 16-bit width and height
    w = rng.read(2)
    h = rng.read(2)

    # packed field of color resolution (bits per pixel)
    r = rng.read(1)[0] & 7
    g = rng.read(1)[0] & 7
    b = rng.read(1)[0] & 7
    packed = bytes([r | (g << 3) | (b << 6)])

    # background color index
    bg = rng.read(1)[0]

    # aspect ratio - fixed to 4:3
    ar = rng.read(1)[0] & 7

    data = bytearray()

    for i in range(int.from_bytes(h, 'little')):
        for j in range(int.from_bytes(w, 'little')):
            data += rng.read(1)

    out.write(header + w + h + packed + bg.to_bytes(1, 'little') + ar.to_bytes(1, 'little'))
    out.write(data)

# Reads an existing GIF file from `in`, modifies it by replacing all pixels with their
# grayscale equivalents, then writes the result back to `out`.
def convert_to_grayscale(
        in_: BinaryIO, out: BinaryIO
    ):
    header = in_.read(6)
    
    if not header == b'GIF89a':
        raise ValueError('Invalid header.')

    w = int.from_bytes(in_.read(2), 'little')
    h = int.from_bytes(in_.read(2), 'little')

    packed = in_.read(1)[0]
    bg = in_.read(1)[0]
    ar = in_.read(1)[0]

    data = bytearray()

    for i in range(h):
        for j in range(w):
            p = in_.read(1)[0]

            red   = (p >> 5) * 0x11
            green = ((p >> 2) & 7) * 0x11
            blue  = (p & 3) * 0x11
            
            data += bytes([(red + green + blue) // 3])

    out.write(header)
    out.write(w.to_bytes(2, 'little'))
    out.write(h.to_bytes(2, 'little'))
    out.write(packed.to_bytes(1, 'little'))
    out.write(bg.to_bytes(1, 'little'))
    out.write(ar.to_bytes(1, 'little'))
    out.write(data)