#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# returns its size in bytes.
def generate_gif(out: BinaryIO, rng):
    # Header (6 bytes).
    out.write(b'GIF89a')

    # Logical screen descriptor (7 bytes).
    width = 1 + rng.randint(0x20, 0x40) * 8
    height = 1 + rng.randint(0x20, 0x40) * 8
    flags = b'\x87\x00' if rng.randint(0, 1) else b'\x07\x00'
    bg_color = rng.randint(0, 0xFF)
    pixel_aspect_ratio = b'\x00'
    out.write(struct.pack('<HHBBB', width, height, flags[0], flags[1],
        bg_color, pixel_aspect_ratio[0]))

    # Global color table (3 bytes).
    num_colors = min((width * height), 256)
    for i in range(num_colors):
        r = g = b = rng.randint(0, 0xFF)
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor (10 bytes).
    left = top = 0
    width -= 1
    height -= 1
    flags = b'\x80\x00'
    local_color_table_flag = b'\x00'
    interlace_flag = b'\x00'
    sort_flag = b'\x00'
    reserved_flags = b'\x00'
    lct_size = b'\x00'
    out.write(struct.pack('<HHHHBBBBB', left, top, width, height, flags[0],
        flags[1], local_color_table_flag[0], interlace_flag[0], sort_flag[0],
        reserved_flags[0], lct_size[0]))

    # Local color table (3 bytes).
    for i in range(num_colors):
        r = g = b = rng.randint(0, 0xFF)
        out.write(struct.pack('BBB', r, g, b))

    # Image data block (variable length).
    while True:
        # Block size (1 byte).
        block_size = rng.randint(0xFB, 0xFF)

        # LZW minimum code size (1 byte).
        lzw_min_code_size = rng.randint(0x02, 0x08)

        # Subblocks (variable length).
        subblock_count = 0
        while True:
            # Terminator subblock (1 byte).
            if rng.randint(0, 1):
                break

            # Data subblock (variable length).
            if rng.randint(0, 1):
                subblock_count += 1

                # Data size (1 byte).
                data_size = rng.randint(0x02, 0xFE)

                # Data (data_size bytes).
                for _ in range(data_size - 1):
                    out.write(bytes([rng.randint(0x00, 0xFF)]))

                # Escape sequence (1 byte).
                out.write(bytes([0x00]))

                # Zero-length subblock (1 byte).
                if rng.randint(0, 1):
                    continue

            # Control subblock (variable length).
            elif rng.randint(0, 1):
                subblock_count += 1

                # Control size (1 byte).
                control_size = rng.randint(0x01, 0x03)

                # Control (control_size bytes).
                for _ in range(control_size):
                    out.write(bytes([rng.randint(0x00, 0xFF)]))

                # Zero-length subblock (1 byte).
                if rng.randint(0, 1):
                    continue

            # End of image subblock (1 byte).
            else:
                subblock_count += 1
                break

        # Check that we have generated at least one subblock.
        assert subblock_count > 0

        # Write block size (1 byte).
        out.write(bytes([block_size]))

        # Write LZW minimum code size (1 byte).
        out.write(bytes([lzw_min_code_size]))

        # Write subblocks (variable length).
        return

