#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
	"""Generate a random GIF file.

	Args:
		rng (BinaryIO): A binary stream to read from for the RNG seed.
		out (BinaryIO): The output file to write the GIF data into.
	"""

    # Header
    header = b'GIF89a'
    out.write(header)

    # Logical screen descriptor
    width = 1024
    height = 768
    color_table_size = 256
    global_color_table = True
    background_index = 0
    pixel_aspect_ratio = 0
    logical_screen_descriptor = struct.pack('<HHBBB',
            width, height, color_table_size,
            global_color_table << 7 | background_index,
            pixel_aspect_ratio)
    out.write(logical_screen_descriptor)

    # Global color table
    palette = [struct.pack('BBB', *tuple([i] * 3)) for i in range(256)]
    global_color_table_data = b''.join(palette)
    out.write(global_color_table_data)

    # Image descriptor
    image_left = 0
    image_top = 0
    image_width = width
    image_height = height
    local_color_table = False
    interlace = False
    sort = False
    reserved = 0
    lct_size = 0
    image_descriptor = struct.pack('<BBHHB',
            0x2C, 0,
            image_left, image_top,
            image_width, image_height,
            local_color_table << 7 | interlace << 6 | sort << 5 | reserved)
    out.write(image_descriptor)

    # Local color table
    local_color_table_data = b''
    if not local_color_table:
        local_color_table_data = b'\x00\x00\x00' * lct_size
    out.write(local_color_table_data)

    # LZW minimum code size
    min_code_size = 2
    out.write(bytes([min_code_size]))

    # Data blocks
    block_size = 256 - 2
    block = bytes()
    while len(block) < block_size:
        value = int.from_bytes(rng.read(1), 'little') % 256
        block += bytes([value])
    out.write(block)

    # Trailer
    trailer = b';'
    out.write(trailer)


if