#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# returns its size in bytes.
def generate_gif(out: BinaryIO, rng):
    # Header
    out.write(b"GIF89a")

    # Logical screen descriptor
    width = 256 + (rng.randint(0, 1024) % 256)
    height = 256 + (rng.randint(0, 1024) % 256)
    flags = b"\x87\x00" if rng.randint(0, 1) else b"\x07\x00"
    bgcolor = rng.randint(0, 255).to_bytes(1, "little") * 3
    aspect = b"\x00\x00"
    out.write(struct.pack("<HHBBB", width, height, flags[0], flags[1], bgcolor[0]))
    out.write(aspect)

    # Global color table
    for i in range(0, 256):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write(r.to_bytes(1, "little"))
        out.write(g.to_bytes(1, "little"))
        out.write(b.to_bytes(1, "little"))

    # Image descriptor
    left = 0
    top = 0
    width = 256 + (rng.randint(0, 1024) % 256)
    height = 256 + (rng.randint(0, 1024) % 256)
    flags = b"\x87\x00" if rng.randint(0, 1) else b"\x07\x00"
    bgcolor = rng.randint(0, 255).to_bytes(1, "little") * 3
    aspect = b"\x00\x00"
    out.write(struct.pack("<HHHHB", left, top, width, height, flags[0] | flags[1]))
    out.write(bgcolor)
    out.write(aspect)

    # Local color table
    for i in range(0, 256):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write(r.to_bytes(1, "little"))
        out.write(g.to_bytes(1, "little"))
        out.write(b.to_bytes(1, "little"))

    # Image data
    for y in range(height):
        lzwmincodesize = 2
        clearcode = 1 << lzwmincodesize
        endofinformation = clearcode + 1
        nextcode = None

        def emitbyte(b):
            nonlocal nextcode

            if nextcode is not None:
                if nextcode == clearcode:
                    codetable = [i for i in range(clearcode)]

                    for i in range(1, 1 << lzwmincodesize):
                        codetable.append((i << 8) | i)

                elif nextcode < len(codetable):
                    pass
                else:
                    assert nextcode == endofinformation
                    codetable.append(nextcode)

                code = codetable[nextcode]
                codelen = lzwmincodesize + 1

                while True:
                    if code >= len(codetable):
                        break

                    out.write(codetable[code].to_bytes(1, "little"))
                    del codetable[code]

                    if code <= maxcode:
                        break

                    codelen += 1
                    code >>= 1

                if codetable:
                    maxcode = max(codetable)
                else:
                    maxcode = -1

                nextcode = None

            if b!= 255:
                out.write(b.to_bytes(1, "little"))

        def emitcode(c):
            nonlocal nextcode

            if nextcode is None or nextcode == endofinformation:
                nextcode = c
            else:
                nextcode = ((nextcode << 1) & 0xff) | bool(c)

        prevpixel = None
        runlength = 0

        for x in range(width):
            pixel = rng.randint(0, 255)
            if prevpixel == pixel:
                runlength += 1
            else:
                emitcode(prevpixel is not None)
                if prevpixel is not None:
                    emitbyte(run