#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Header
    out.write(b'GIF89a')

    # Logical Screen Descriptor
    width = 1 + int.from_bytes(rng.read(1), 'big') % 256
    height = 1 + int.from_bytes(rng.read(1), 'big') % 256
    flags = b'\x07\x00' if (width <= 240 and height <= 240) else b'\x08\x00'
    background_color_index = int.from_bytes(rng.read(1), 'big') & 0xf
    pixel_aspect_ratio = b'\x00'
    out.write(struct.pack('<HHBBB', width, height, flags[0], flags[1], background_color_index))
    out.write(pixel_aspect_ratio)

    # Global Color Table
    global_color_table_size = 2 << ((int.from_bytes(rng.read(1), 'big') >> 3) & 0x3)
    for i in range(global_color_table_size):
        r = int.from_bytes(rng.read(1), 'big') & 0xff
        g = int.from_bytes(rng.read(1), 'big') & 0xff
        b = int.from_bytes(rng.read(1), 'big') & 0xff
        out.write(struct.pack('BBB', r, g, b))

    # Image Descriptor
    left = 0
    top = 0
    width = 1 + int.from_bytes(rng.read(1), 'big') % 256
    height = 1 + int.from_bytes(rng.read(1), 'big') % 256
    flags = b'\x0c\x00'
    local_color_table_flag = b'\x00'
    interlace_flag = b'\x00'
    sort_flag = b'\x00'
    reserved = b'\x00\x00'
    local_color_table_size = 2 << ((int.from_bytes(rng.read(1), 'big') >> 3) & 0x3)
    out.write(struct.pack('<HHHHBBBBB', left, top, width, height, flags[0], flags[1], local_color_table_flag[0], local_color_table_size, interlace_flag[0]))
    out.write(sort_flag)
    out.write(reserved)
    for i in range(local_color_table_size):
        r = int.from_bytes(rng.read(1), 'big') & 0xff
        g = int.from_bytes(rng.read(1), 'big') & 0xff
        b = int.from_bytes(rng.read(1), 'big') & 0xff
        out.write(struct.pack('BBB', r, g, b))

    # Image Data
    lzw_min_code_size = int.from_bytes(rng.read(1), 'big') & 0xff
    out.write(lzw_min_code_size.to_bytes(1, byteorder='little'))
    image_data_length = 1 + int.from_bytes(rng.read(1), 'big') % 256
    while True:
        block_type = int.from_bytes(rng.read(1), 'big') & 0xff
        if block_type == 0x00:
            break

        elif block_type == 0x01:
            subblock_length = int.from_bytes(rng.read(1), 'big') & 0xff
            out.write(subblock_length.to_bytes(1, byteorder='little'))

            subblock_data = bytes()
            while len(subblock_data) < subblock_length:
                subblock_data += rng.read(1)
            
            out.write(subblock_data[:subblock_length])
        
        elif block_type == 0xfe:
            subblock_length = int.from_bytes(rng.read(1), 'big') & 0xff
            out.write(subblock_length.to_bytes(1, byteorder='little'))

            subblock_data = bytes()
            while len(subblock_data) < subblock_length:
                subblock_data += rng.read(1)
            
            out.write(subblock_data[:subblock_length])
    
    out.write(image_data_length.