#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        width: int = 1024, height: int = 768,
        max_color_depth: int = 8,
        num_frames: int = 5):

    # Header
    header = b'GIF89a'
    out = bytearray()
    out += header

    # Logical screen descriptor
    out += (width).to_bytes(2, 'little')
    out += (height).to_bytes(2, 'little')
    out += bytes([max_color_depth])
    out += bytes([0])      # Background color index
    out += bytes([0])      # Pixel aspect ratio

    # Global color table
    for i in range(pow(2, max_color_depth)):
        out += bytes([i % 256, i // 256, i * 2 % 256])

    # Image descriptor
    out += bytes([0x2C])   # Image separator
    out += (0).to_bytes(2, 'little')     # Left position
    out += (0).to_bytes(2, 'little')     # Top position
    out += (width).to_bytes(2, 'little')
    out += (height).to_bytes(2, 'little')
    out += bytes([0])      # Local color table flag
    out += bytes([0b11111110])       # LZW minimum code size
    out += bytes([0])      # Block terminator

    # Data sub-block
    data = []
    for _ in range(num_frames):

        # Frame control block
        out += bytes([0x21])           # Extension introducer
        out += bytes([0xF9])           # Graphic control label
        out += bytes([0x04])           # Block size
        out += bytes([0b10000000 | 0b00000000])         # Flags: disposal method=none, user input=false
        out += bytes([0xFF])           # Delay time
        out += bytes([0xFF])
        out += bytes([0xFF])
        out += bytes([0xFF])
        out += bytes([0x00])           # Transparent color index
        out += bytes([0x00])           # Block terminator

        # Plain text extension
        out += bytes([0x21])           # Extension introducer
        out += bytes([0x01])           # Text label
        out += bytes([0x0B])           # Block size
        out += b'text\0'               # Text string
        out += bytes([0x00])           # Block terminator

        # Graphics control extension
        out += bytes([0x21])           # Extension introducer
        out += bytes([0xFF])           # Application extension label
        out += bytes([0x0B])           # Block size
        out += b'NETSCAPE2.0\0'        # Application identifier
        out += bytes([0x03])           # Sub-block size
        out += bytes([0b00000000])     # Loop count is zero
        out += bytes([0x00])           # Block terminator

        # Image data
        for y in range(height):
            for x in range(width):
                data.append(rng.read(1)[0] & ((1 << max_color_depth) - 1))

    # Write to output file
    datalen = len(data)
    while datalen > 255:
        out.extend(bytes([255]))
        out.extend(data[:255])
        del data[:255]
        datalen -= 255
    if datalen!= 0:
        out.write(header + datalen.to_bytes(3, 'big') + data)
