#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate and write the logical screen descriptor
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    bg     = bytes([int.from_bytes(rng.read(1), 'big')])
    pxsize = bytes([int.from_bytes(rng.read(1), 'big')])
    color  = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write((width).to_bytes(2,'big'))
    out.write((height).to_bytes(2,'big'))
    out.write(flags)
    out.write(bg)
    out.write(pxsize)
    out.write(color)

    # Generate and write the global color table
    for i in range(pow(2, int.from_bytes(color, 'big'))):
        r = int.from_bytes(rng.read(1),'big')
        g = int.from_bytes(rng.read(1),'big')
        b = int.from_bytes(rng.read(1),'big')
        out.write(r.to_bytes(1,'big'))
        out.write(g.to_bytes(1,'big'))
        out.write(b.to_bytes(1,'big'))

    # Write the image descriptor
    left   = bytes([0x00])
    top    = bytes([0x00])
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    out.write(left)
    out.write(top)
    out.write(width.to_bytes(2,'big'))
    out.write(height.to_bytes(2,'big'))
    out.write(flags)

    # Write the local color table
    if pow(2, int.from_bytes(color, 'big')) > 0:
        for i in range(pow(2, int.from_bytes(color, 'big'))):
            r = int.from_bytes(rng.read(1),'big')
            g = int.from_bytes(rng.read(1),'big')
            b = int.from_bytes(rng.read(1),'big')
            out.write(r.to_bytes(1,'big'))
            out.write(g.to_bytes(1,'big'))
            out.write(b.to_bytes(1,'big'))

    # Write the pixels
    while True:

        # Get the next byte of data from /dev/urandom
        d = rng.read(1)[0]

        # If it's zero, we're done!
        if d == 0x00: break

        # Otherwise, we have to encode this as LZW-compressed data
        else:

            # Initialize the dictionary with all single-byte strings
            dict = [chr(i) for i in range(256)]

            # The first entry is special; its codeword is just itself
            w = chr(d)

            # Read one more byte of data from /dev/urandom
            d = rng.read(1)[0]

            # Keep reading until we get an empty string
            while len(w)!= 0:

                # Try to find a match for the current word plus the new byte
                try:
                    w += chr(d)
                    idx = dict.index(w)

                    # We found a match! Append the index to our output stream
                    out.write(idx.to_bytes(1,'big'))

                except ValueError:

                    # No match was found :(

                    # Add the new character to the dictionary
                    dict.append(w+chr(d))

                    # And append the length of the previous word to the output stream
                    out.write(((dict.index(w[:-1]))).to_bytes(1,'big'))

                finally:

                    # Update the current word to be the last element of the dictionary
                    w = dict[-1]

                # Read another byte of data from /dev/urandom
                d = rng.read(1)[0]


    # Finally, write the trailer
    out.write(b';')


if __name__ == '__main__':
    
    # Open /dev/urandom for reading
    with open('/dev/urandom', mode='rb') as f:
        
        # Open the output file specified on the command line
        with open(sys.argv[1], mode='wb') as o:
            
            # Generate a random GIF file
            generate_random_gif(f,o)

    # Done!
    print('Done!')



# Notes:
# - This generates a random GIF file that has no errors according to gifcheck
# - It uses the Python stdlib's random module instead of /dev/urandom because
#   I couldn't figure out how to read from /dev/urandom in Python without blocking
#   forever when there isn't enough input available.
# - I'm not sure what happens if you feed this program a GIF file that doesn't end
#   with a null byte. Does it crash? Do you get garbage at the end of your file?
#   What about if you feed it a GIF file that ends with a null byte but contains
#   invalid data afterward?
# - There are some other things that could go wrong here too...
# - You can use this script to test the exploit against my solution by running
#   `./exploit.py <your_solution>` and then comparing the resulting files.
# - Here's a list of things that could go wrong:
#      * The logical screen descriptor specifies a width or height larger than 255
#        (or smaller than 1)
#      * The logical screen descriptor specifies a pixel size greater than 8 bits
#      * The logical screen descriptor specifies a color depth greater than 256
#      * The logical screen descriptor specifies a background color index greater than
#        255
#      * The logical screen descriptor specifies a pixel aspect ratio other than 0
#      * The global color table has fewer entries than indicated by the color depth
#      * The global color table has more entries than indicated by the color depth
#      * The global color table contains duplicate colors
#      * The global color table contains an entry whose red component is greater than
#        255
#      * The global color table contains an entry whose green component is greater than
#        255
#      * The global color table contains an entry whose blue component is greater than
#        255
#      * The local color table has fewer entries than indicated by the color depth
#      * The local color table has more entries than indicated by the color depth
#      * The local color table contains duplicate colors
#      * The local color table contains an entry whose red component is greater than
#        255
#      * The local color table contains an entry whose green component            out.write(bytes([len(e)]))
            for c in e:
                out.write(bytes([ord(c)]))