#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# writes it to `out`.
def generate_random(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write("GIF89a".encode())

    # Logical screen descriptor (7 bytes)
    width = 1 + int.from_bytes(rng.read(1), byteorder="big") % 50
    height = 1 + int.from_bytes(rng.read(1), byteorder="big") % 50
    flags = b"\x87" if bool(int.from_bytes(rng.read(1), byteorder="big")) else b"\x00"
    bg_color = int.from_bytes(rng.read(1), byteorder="big") & 0xf0 | 0x0f
    aspect_ratio = int.from_bytes(rng.read(1), byteorder="big") & 0xff
    out.write(struct.pack("<HHBBB", width, height, flags[0], bg_color, aspect_ratio))

    # Global color table (256 * 3 bytes)
    for i in range(256):
        r = int.from_bytes(rng.read(1), byteorder="big") & 0xff
        g = int.from_bytes(rng.read(1), byteorder="big") & 0xff
        b = int.from_bytes(rng.read(1), byteorder="big") & 0xff
        out.write(struct.pack("<BBB", r, g, b))

    # Image descriptor (10 bytes)
    left = 0
    top = 0
    width = 1 + int.from_bytes(rng.read(1), byteorder="big") % 50
    height = 1 + int.from_bytes(rng.read(1), byteorder="big") % 50
    flags = b"\xc4" if bool(int.from_bytes(rng.read(1), byteorder="big")) else b"\x00"
    delay = int.from_bytes(rng.read(2), byteorder="big") & 0xffff
    transparent_index = int.from_bytes(rng.read(1), byteorder="big") & 0xff
    out.write(struct.pack("<HHHHBB", left, top, width, height, flags[0], delay, transparent_index))

    # Local color table (256 * 3 bytes)
    for i in range(256):
        r = int.from_bytes(rng.read(1), byteorder="big") & 0xff
        g = int.from_bytes(rng.read(1), byteorder="big") & 0xff
        b = int.from_bytes(rng.read(1), byteorder="big") & 0xff
        out.write(struct.pack("<BBB", r, g, b))

    # Image data
    lzw_min_code_size = int.from_bytes(rng.read(1), byteorder="big") & 0xff
    out.write(lzw_min_code_size.to_bytes(1, "little"))

    # LZW code stream
    while True:
        value = int.from_bytes(rng.read(1), byteorder="big") & 0xff

        if value == 0x00:
            break

        elif value <= 0x7f:
            out.write(value.to_bytes(1, "little"))

        else:
            length = ((value >> 4) & 0x07) + 2

            if length == 10:
                value &= 0x0f

                if not value:
                    break

            out.write(((length - 2) << 4) | value).to_bytes(1, "little")

