#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and returns it.
def generate_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write("GIF89a".encode())

    # Logical screen descriptor
    width = 1 + rng.read(1)[0] % 256
    height = 1 + rng.read(1)[0] % 256
    flags = (
        ((rng.read(1)[0] & 1) << 7) |
        ((rng.read(1)[0] & 1) << 6) |
        ((rng.read(1)[0] & 1) << 5) |
        0b00000
    )
    bgColorIndex = rng.read(1)[0] % 256
    pixelAspectRatio = 0
    out.write(struct.pack("<HHBBB", width, height, flags, bgColorIndex, pixelAspectRatio))

    # Global color table
    for i in range(2**((flags >> 4) + 1)):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        out.write(struct.pack("<BBB", r, g, b))

    # Image descriptor
    leftPos = 0
    topPos = 0
    width = 1 + rng.read(1)[0] % 256
    height = 1 + rng.read(1)[0] % 256
    flags = (
        ((rng.read(1)[0] & 1) << 7) |
        ((rng.read(1)[0] & 1) << 6) |
        ((rng.read(1)[0] & 1) << 5) |
        0b00000
    )
    delayTime = 1 + rng.read(1)[0] % 256
    transparentColorIndex = rng.read(1)[0] % 256
    out.write(struct.pack("<HHHHBB", leftPos, topPos, width, height, flags, delayTime, transparentColorIndex))

    # Local color table
    if (flags & 0b10000)!= 0:
        for i in range(2**((flags >> 4) + 1)):
            r = rng.read(1)[0] % 256
            g = rng.read(1)[0] % 256
            b = rng.read(1)[0] % 256
            out.write(struct.pack("<BBB", r, g, b))

    # Image data
    lzwMinCodeSize = rng.read(1)[0] % 256
    out.write(bytes([lzwMinCodeSize]))

    codeLength = [0]*(lzwMinCodeSize+1)
    for i in range(lzwMinCodeSize+1):
        codeLength[i] = rng.read(1)[0] % 256
    maxCodeLength = lzwMinCodeSize + 1
    clearCode = 1 << lzwMinCodeSize
    endOfInformationCode = clearCode + 1
    nextCode = endOfInformationCode + 1
    codes = []
    for i in range(nextCode):
        codes.append([])
    currentCode = bytes()
    while True:
        bit = 0x80
        byte = rng.read(1)[0]
        while bit > 0:
            if len(currentCode) == 0 or (byte & bit) == 0:
                currentCode += bytes([byte])
                break

            currentCode += bytes([codes[-len(currentCode)][-1]])

        if len(currentCode) >= maxCodeLength:
            code = -1
            for i in range(maxCodeLength):
                code <<= 1
                if currentCode[i] == 0xff:
                    continue

                code |= currentCode[i]

            if code < nextCode:
                out.write(bytes([code]))
                if code == endOfInformationCode:
                    return

                if code == clearCode:
                    maxCodeLength = lzwMinCodeSize + 1
                    nextCode = endOfInformationCode + 1
                    codes = []
                    for i in range(nextCode):
                        codes.append([])

                    continue

                if len(codes) <= code:
                    codes.append([])

                codes[code].append(codes[-1][-1])
                if len(codes[code]) == 1 << codeLength[code]:
                    maxCodeLength = code + 1
                    nextCode = nextCode + 1

                del codes[-1]
               