#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    out.write("GIF89a".encode())

    # Logical screen descriptor
    width = 1 + int.from_bytes(rng.read(1), byteorder="big") % 256
    height = 1 + int.from_bytes(rng.read(1), byteorder="big") % 256
    flags = b"\x07" if (width == 1 or height == 1) else b"\x00"
    bgColorIndex = int.from_bytes(rng.read(1), byteorder="big") & 0xf
    pixelAspectRatio = int.from_bytes(rng.read(1), byteorder="big") & 0xff
    out.write(struct.pack("<HHBBB", width, height, flags[0], bgColorIndex,
                          pixelAspectRatio))

    # Global color table
    globalColorTableSize = min((int.from_bytes(rng.read(1), byteorder="big") &
                                0x07) + 1, 7) * 3
    for i in range(globalColorTableSize):
        r = int.from_bytes(rng.read(1), byteorder="big") & 0xff
        g = int.from_bytes(rng.read(1), byteorder="big") & 0xff
        b = int.from_bytes(rng.read(1), byteorder="big") & 0xff
        out.write(struct.pack("<BBB", r, g, b))
    for _ in range(4 - globalColorTableSize // 3):
        out.write(b'\x00\x00\x00')

    # Image descriptor
    leftPos = int.from_bytes(rng.read(1), byteorder="big")
    topPos = int.from_bytes(rng.read(1), byteorder="big")
    imageWidth = 1 + int.from_bytes(rng.read(1), byteorder="big") % 256
    imageHeight = 1 + int.from_bytes(rng.read(1), byteorder="big") % 256
    flags = b"\x0c" if ((imageWidth == 1 and imageHeight == 1)
                        or (leftPos == 0 and topPos == 0
                            and imageWidth == width and imageHeight == height)) \
            else b"\x00"
    localColorTableFlag = bytes([flags[0] |
                                  (int.from_bytes(rng.read(1),
                                                  byteorder="big") &
                                   0x01)])
    interlaceFlag = bytes([(flags[0] >> 2) |
                           (int.from_bytes(rng.read(1),
                                           byteorder="big") &
                            0x01 << 1)])
    sortFlag = bytes([(flags[0] >> 3) |
                      (int.from_bytes(rng.read(1),
                                      byteorder="big") &
                        0x01 << 2)])
    reserved = b"\x00"
    localColorTableSize = min((int.from_bytes(rng.read(1),
                                               byteorder="big") &
                               0x07) + 1, 7) * 3
    out.write(struct.pack("<HHBBBBB", leftPos, topPos, imageWidth, imageHeight,
                          localColorTableFlag[0], interlaceFlag[0],
                          sortFlag[0], reserved[0]))
    for i in range(localColorTableSize):
        r = int.from_bytes(rng.read(1), byteorder="big") & 0xff
        g = int.from_bytes(rng.read(1), byteorder="big") & 0xff
        b = int.from_bytes(rng.read(1), byteorder="big") & 0xff
        out.write(struct.pack("<BBB", r, g, b))
    for _ in range(4 - localColorTableSize // 3):
        out.write(b'\x00\x00\x00')

    # LZW minimum code size
    lzwMinimumCodeSize = int.from_bytes(rng.read(1), byteorder="big") & 0xff
    out.write(lzwMinimumCodeSize.to_bytes(1, "little"))

    # Image data
    def get_next_byte(): return int.from_bytes(rng.read(1), byteorder="big")
    def put_back_byte(b): rng.seek(-1, 1)
    def read_code():
        while True:
            c = get_next_byte()

            if c >= 