#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header
    header = b'GIF89a'

    # Logical screen descriptor
    width = 1 + (rng.read(1)[0] % 256)
    height = 1 + (rng.read(1)[0] % 256)
    flags = bytes([
        0b11 << 7 |     # Global color table flag
        0b1   << 4 |     # Color resolution
        0b1   << 3 |     # Sort flag
        0b1   << 2      # Size of global color table
    ])
    bgColorIndex = 0
    pixelAspectRatio = 0
    logicalScreenDescriptor = \
            header + \
            width.to_bytes(1, 'little') + \
            height.to_bytes(1, 'little') + \
            flags + \
            bgColorIndex.to_bytes(1, 'little') + \
            pixelAspectRatio.to_bytes(1, 'little')

    # Global color table
    globalColorTableSize = 2 ** ((flags[0] & 0b11) + 1)
    globalColorTable = bytearray()
    for i in range(globalColorTableSize * 3):
        globalColorTable += rng.read(1)

    # Image descriptor
    imageLeftPosition = 0
    imageTopPosition = 0
    imageWidth = width
    imageHeight = height
    flags = bytes([
        0b1    << 7 |     # Local color table flag
        0b10   << 5 |     # Interlace flag
        0b10   << 3 |     # Sort flag
        0b1    << 2 |     # Reserved
        0b100  << 0       # LZW minimum code size
    ])
    localColorTableSize = 2 ** (((flags[0] >> 4) & 0b11) + 1)
    localColorTable = bytearray()
    if localColorTableSize > 0:
        for i in range(localColorTableSize * 3):
            localColorTable += rng.read(1)
    lzwMinimumCodeSize = flags[-1]
    imageDescriptor = \
            imageLeftPosition.to_bytes(2, 'little') + \
            imageTopPosition.to_bytes(2, 'little') + \
            imageWidth.to_bytes(2, 'little') + \
            imageHeight.to_bytes(2, 'little') + \
            flags + \
            localColorTableSize.to_bytes(1, 'little')

    # Image data
    imageData = bytearray()
    while len(imageData) < imageWidth * imageHeight:
        chunkSize = min((rng.read(1)[0] % 128),
                        imageWidth * imageHeight - len(imageData))

        if rng.read(1)[0] % 2 == 0:
            # Raw data block
            imageData += rng.read(chunkSize)
        else:
            # LZW compressed data block

            # First byte is clear code
            prefix = 0x00
            suffix = rng.read(1)[0]
            imageData += bytes([suffix])

            # Generate codes from the input stream until we get to the end
            while True:
                # Read next bit
                bit = rng.read(1)[0]

                # If it's zero, read another byte and add its high bits as well
                if bit == 0:
                    imageData += bytes([prefix])

                    # Add the new byte as a prefix
                    prefix = rng.read(1)[0]
                    imageData += bytes([prefix])
                elif bit <= lzwMinimumCodeSize:
                    # We've reached an end-of-stream marker
                    break
                else:
                    # Otherwise, use the current prefix and the next byte as a suffix
                    imageData += bytes([prefix])
                    imageData += bytes([suffix])

                    # Update the prefix with the last two bytes
                    prefix = (
                            (prefix << lzwMinimumCodeSize) |
                            (suffix >> (8 - lzwMinimumCodeSize))) & 0xff

                    # Use the low bits of the suffix as the new suffix
                    suffix = rng.read(1)[0] & ((1 << lzwMinimumCodeSize) - 1)

    # Trailer
    trailer = b'\x3B'

    out.write(logicalScreenDescriptor)
    out.write(globalColorTable)
    out.write(imageDescriptor)
    out.write(imageData)
    out.write(trailer)

