#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    header = b'GIF89a' + bytes([0x12, 0x34]) + \
             bytes([0x56, 0x78]) + bytes([0x9A, 0xBC])
    out.write(header)

    # Logical screen descriptor
    width = int.from_bytes(rng.read(2), byteorder='little') % (2**12 - 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (2**12 - 1) + 1
    flags = [int.from_bytes(rng.read(1), byteorder='little')] * 3
    bg_color = int.from_bytes(rng.read(1), byteorder='little') & 0xF
    aspect_ratio = int.from_bytes(rng.read(1), byteorder='little') & 0xFF
    lsd = struct.pack('<HHBBB', width, height,
                      flags[0], flags[1], flags[2], bg_color)
    lsd += bytes([aspect_ratio])
    out.write(lsd)

    # Global color table
    global_ct_size = int.from_bytes(rng.read(1), byteorder='little') & 0x7F
    if global_ct_size!= 0:
        for _ in range((global_ct_size ** 2) // 3):
            rgb = []
            for i in range(3):
                rgb.append(int.from_bytes(rng.read(1),
                                          byteorder='little'))
            out.write(struct.pack('BBB',
                                  rgb[0] & 0xFF,
                                  rgb[1] & 0xFF,
                                  rgb[2] & 0xFF))

        for _ in range(((global_ct_size ** 2) // 3) % 3):
            out.write(b'\x00\x00\x00')

    # Loop control block
    loop_count = int.from_bytes(rng.read(1), byteorder='little')
    loop_flags = [int.from_bytes(rng.read(1), byteorder='little')] * 3
    loop_block = struct.pack('<BBBBB',
                             loop_count,
                             loop_flags[0],
                             loop_flags[1],
                             loop_flags[2],
                             0)
    out.write(loop_block)

    # Image descriptor
    left = int.from_bytes(rng.read(2), byteorder='little') % (width - 1)
    top = int.from_bytes(rng.read(2), byteorder='little') % (height - 1)
    image_width = int.from_bytes(rng.read(2),
                                 byteorder='little') % (width - left) + 1
    image_height = int.from_bytes(rng.read(2),
                                  byteorder='little') % (height - top) + 1
    flags = [int.from_bytes(rng.read(1), byteorder='little')] * 3
    local_ct_size = int.from_bytes(rng.read(1),
                                   byteorder='little') & 0x7F
    interlace = int.from_bytes(rng.read(1),
                               byteorder='little') & 0x1
    sort = int.from_bytes(rng.read(1),
                          byteorder='little') & 0x1
    img_descriptor = struct.pack('<HHHHB',
                                 left, top,
                                 image_width, image_height,
                                 flags[0])
    img_descriptor += bytes([flags[1]])
    img_descriptor += bytes([flags[2]])
    img_descriptor += bytes([local_ct_size])
    img_descriptor += bytes([interlace])
    img_descriptor += bytes([sort])
    out.write(img_descriptor)

    # Local color table
    if local_ct_size!= 0:
        for _ in range((local_ct_size ** 2) // 3):
            rgb = []
            for i in range(3):
                rgb.append(int.from_bytes(rng.read(1),
                                          byteorder='little'))
            out.write(struct.pack('BBB',
                                  rgb[0] & 0xFF,
                                  rgb[1] & 0xFF,
                                  rgb[2] & 0xFF))

        for _ in range(((local_ct_size ** 2) // 3) % 3):
            out.write(b'\x00\x00\x00')