#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate and write the logical screen descriptor
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    bg     = bytes([int.from_bytes(rng.read(1), 'big')])
    pxsize = bytes([int.from_bytes(rng.read(1), 'big')])
    color  = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(width + b',' + height + flags + bg + pxsize + color)

    # Generate and write global color table
    for i in range(int.from_bytes(color, 'big')):
        r = int.from_bytes(rng.read(1), 'big') & 0xff
        g = int.from_bytes(rng.read(1), 'big') & 0xff
        b = int.from_bytes(rng.read(1), 'big') & 0xff
        out.write(r.to_bytes(1, byteorder='big'))
        out.write(g.to_bytes(1, byteorder='big'))
        out.write(b.to_bytes(1, byteorder='big'))

    # Generate and write image data
    lzwmincodesize = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(lzwmincodesize)

    # Generate LZW code size table
    lzwsizes = [
        2**i for i in range(
            int.from_bytes(lzwmincodesize, 'big'),
            12+1
        )
    ]

    # Generate LZW code table
    codes = []
    for sz in lzwsizes:
        codes += list(range(sz))

    # Generate LZW dictionary
    dict = {}
    for c in codes:
        dict[c] = bytes([c])

    # Generate LZW stream
    while len(codes):

        # Get next LZW code from input
        buf = rng.read(1)
        if not buf or len(buf) == 0: break
        code = int.from_bytes(buf, 'big')

        # If this is a clear code, reset the dictionary
        if code == 0x00:

            # Reset dictionary to initial state
            dict = {
                k : v for k,v in zip(
                    codes,
                    [
                        bytes([k]) for k in codes
                    ]
                )
            }

            # Continue with next code
            continue

        # If this is an end of information code, stop
        elif code == 0x01: break

        # If this is a literal code, emit corresponding character
        elif code < len(dict):
            out.write(dict[code])

        # Otherwise, emit longest match
        else:

            # Find longest match
            maxlen = -1
            maxidx = -1
            for idx, d in enumerate(dict.values()):

                # Check if current entry matches beginning of code
                if d[:len(dict)-maxlen-1]!= dict[code][:len(d)]: continue

                # If it does, check if longer than previous best
                if len(d) > maxlen:

                    # Update maximum length and index
                    maxlen = len(d)
                    maxidx = idx

            # Emit longest match
            out.write(dict[maxidx][:-1])

            # Add new entry to dictionary
            dict[len(dict)] = \
                dict[code][:len(dict)-maxlen-1] + \
                dict[maxidx][-1:]


if