#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom) and
# returns its size in bytes.
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + (rng.read(1)[0] % 256)
    height = 1 + (rng.read(1)[0] % 256)
    flags = b'\x87\x00' if width <= 64 else b'\x00\x00'
    out.write(struct.pack('<HHBBB', width, height, 0xF0 | ((width >> 8) & 0x0F), 0, len(flags)))
    out.write(flags)

    # Global color table
    colors = [
        (rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]) for _ in range((rng.read(1)[0] % 256))
    ]
    out.write(bytes([len(colors)]))
    for r, g, b in colors:
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    out.write(b'\x2C\x00\x00\x00\x00\x00,\x00\x00\x00\x00\x01\x00\x18\x00\x00\x00\x00\xFF\xFF\xFF')

    # LZW minimum code size
    out.write(b'\x08')

    # Frame data
    lzw_mincodesize = rng.read(1)[0] % 256
    lzw_maxcodesize = min(lzw_mincodesize + 1, 12)
    lzw_clearcode = 1 << lzw_mincodesize
    lzw_endofinfo = lzw_clearcode + 1
    lzw_nextcode = lzw_endofinfo + 1
    lzw_prevcode = None
    lzw_curcode = None
    lzw_curbit = -1
    lzw_curbyte = 0
    lzw_curmask = 0x80
    lzw_table = {}
    lzw_stack = []
    lzw_buf = bytearray()
    lzw_bufpos = 0
    lzw_buflen = 0
    while True:
        # Generate new pixel
        x = rng.read(1)[0] % width
        y = rng.read(1)[0] % height

        # Get previous pixel
        prev_x = max(x-1, 0)
        prev_y = max(y-1, 0)
        prev_color = None
        try:
            prev_color = lzw_buf[lzw_bufpos - 1]
        except IndexError:
            pass

        # Write pixel to buffer
        lzw_buf += struct.pack('B', prev_color)
        lzw_bufpos += 1
        lzw_buflen += 1

        # Check if we need to flush the buffer
        if lzw_buflen >= 255 or (lzw_bufpos > 1 and lzw_buf[-2:] == b'\x00\x00'):
            # Flush buffer
            out.write(struct.pack('BB', lzw_buflen, prev_color))

            # Reset buffer
            lzw_buf = bytearray()
            lzw_bufpos = 0
            lzw_buflen = 0

        # Check if we have enough pixels to encode
        if lzw_buflen < 2: continue

        # Encode pixels as LZW codes
        while lzw_bufpos >= 2:
            # Get current pixel
            cur_color = lzw_buf[lzw_bufpos - 1]
            next_color = lzw_buf[lzw_bufpos - 2]

            # Check if this is a clear code
            if cur_color == 0 and next_color == 0:
                # Clear code
                lzw_curcode = lzw_clearcode

                # Add clear code to stack
                lzw_stack.append(lzw_curcode)

                # Update bit mask
                lzw_curbit = lzw_curbit + 1
                lzw_curmask = 0x80 << lzw_curbit

                # Move cursor back two positions
                lzw_bufpos -= 2

                # Continue with loop
                break

            # Check if this is an end of information code
            elif cur_color == 0