#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate some random bytes for the length of the image data
    # and write them to the output stream
    datalen = int.from_bytes(
            rng.read(3), byteorder='big', signed=False)
    out.write(datalen.to_bytes(
            3, byteorder='big', signed=False))

    # Write the palette color table (256 colors)
    for i in range(0, 768):
        out.write(rng.read(1))

    # Write the global color table flag
    out.write(b'\x87\x00')

    # Write the background color index
    bgidx = int.from_bytes(
            rng.read(1), byteorder='big', signed=False) % 256
    out.write(bgidx.to_bytes(
            1, byteorder='big', signed=False))

    # Write the pixel aspect ratio
    out.write(b'\x00')

    # Write the local color table flags
    for _ in range(0, 256):
        out.write(b'\x00')

    # Write the images
    for _ in range(0, 4):

        # Write the left position
        pos = int.from_bytes(
                rng.read(2), byteorder='big', signed=False) % 65536
        out.write(pos.to_bytes(
                2, byteorder='big', signed=False))

        # Write the top position
        pos = int.from_bytes(
                rng.read(2), byteorder='big', signed=False) % 65536
        out.write(pos.to_bytes(
                2, byteorder='big', signed=False))

        # Write the width
        size = int.from_bytes(
                rng.read(2), byteorder='big', signed=False) % 65536
        out.write(size.to_bytes(
                2, byteorder='big', signed=False))

        # Write the height
        size = int.from_bytes(
                rng.read(2), byteorder='big', signed=False) % 65536
        out.write(size.to_bytes(
                2, byteorder='big', signed=False))

        # Write the local color table flag
        out.write(b'\x00')

        # Write the interlace flag
        out.write(b'\x00')

        # Write the sort flag
        out.write(b'\x00')

        # Write the reserved field
        out.write(b'\x00')

        # Write the LZW minimum code size
        lzwmincodesize = int.from_bytes(
                rng.read(1), byteorder='big', signed=False) % 256
        out.write(lzwmincodesize.to_bytes(
                1, byteorder='big', signed=False))

        # Write the image data
        while True:

            # Read an unsigned byte from the input stream
            b = ord(rng.read(1))

            if b == 0x00 or b == 0xff:

                # Write it to the output stream
                out.write(b.to_bytes(
                        1, byteorder='big', signed=False))

                break

            elif b >= 0x01 and b <= 0xfe:

                # Write it to the output stream
                out.write(b.to_bytes(
                        1, byteorder='big', signed=False))

            else:

                # Read another byte from the input stream
                b2 = ord(rng.read(1))

                # Write the first byte as is
                out.write(b.to_bytes(
                        1, byteorder='big', signed=False))

                # Write the second byte with its high bit set
                out.write((b2 | 0x80).to_bytes(
                        1, byteorder='big', signed=False))

