#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = int.from_bytes(rng.read(2), byteorder='big') % (2**16 - 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='big') % (2**16 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    background = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    out.write(struct.pack('>HHBBB', width, height, *flags))
    out.write(background)
    out.write(aspect)

    # Global color table
    global_color_table_size = int.from_bytes(rng.read(1), byteorder='big') & 0b00011111
    global_color_table_flag = bool(global_color_table_size)
    global_color_table = []
    for i in range(pow(2, global_color_table_size)):
        r = int.from_bytes(rng.read(1), byteorder='big')
        g = int.from_bytes(rng.read(1), byteorder='big')
        b = int.from_bytes(rng.read(1), byteorder='big')
        global_color_table += [r,g,b]
    if global_color_table_flag:
        out.write(struct.pack('B', global_color_table_size))
        out.write(bytearray(global_color_table))

    # Image descriptor
    left = int.from_bytes(rng.read(2), byteorder='big') % (2**16 - 1)
    top = int.from_bytes(rng.read(2), byteorder='big') % (2**16 - 1)
    image_width = int.from_bytes(rng.read(2), byteorder='big') % (2**16 - 1) + 1
    image_height = int.from_bytes(rng.read(2), byteorder='big') % (2**16 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    local_color_table_size = int.from_bytes(rng.read(1), byteorder='big') & 0b00011111
    local_color_table_flag = bool(local_color_table_size)
    interlace = bool((int.from_bytes(flags, byteorder='big') >> 6) & 0b0000001)
    sort = bool((int.from_bytes(flags, byteorder='big') >> 5) & 0b0000001)
    reserved = int.from_bytes(rng.read(1), byteorder='big')
    lct = []
    for i in range(pow(2, local_color_table_size)):
        r = int.from_bytes(rng.read(1), byteorder='big')
        g = int.from_bytes(rng.read(1), byteorder='big')
        b = int.from_bytes(rng.read(1), byteorder='big')
        lct += [r,g,b]
    out.write(struct.pack('>HHBBBBB', left, top, image_width, image_height, *flags))
    if local_color_table_flag:
        out.write(struct.pack('B', local_color_table_size))
        out.write(bytearray(lct))

    # Image data
    data = rng.read(image_width*image_height)
    out.write(data)

    # Trailer
    out.write(b';')