#!/usr/bin/env python3

import os
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'\x47\x49\x46\x38\x39\x61')

    # Logical screen descriptor
    width  = struct.unpack('>H', rng.read(2))[0] % 500
    height = struct.unpack('>H', rng.read(2))[0] % 500
    color_depth = struct.unpack('B', rng.read(1))[0] & 7
    global_color_table_flag = bool(struct.unpack('B', rng.read(1))[0] & 0x80)
    background_color_index = struct.unpack('B', rng.read(1))[0] & 0xf
    pixel_aspect_ratio = struct.unpack('B', rng.read(1))[0] & 0x7

    out.write(width.to_bytes(2, 'big'))
    out.write(height.to_bytes(2, 'big'))
    out.write((color_depth | ((global_color_table_flag << 7) & 0x80)).to_bytes(1, 'big'))
    out.write(((background_color_index << 4) | pixel_aspect_ratio).to_bytes(1, 'big'))

    # Global color table
    if global_color_table_flag:
        for _ in range(pow(2, color_depth)):
            r = struct.unpack('B', rng.read(1))[0]
            g = struct.unpack('B', rng.read(1))[0]
            b = struct.unpack('B', rng.read(1))[0]

            out.write(r.to_bytes(1, 'big'))
            out.write(g.to_bytes(1, 'big'))
            out.write(b.to_bytes(1, 'big'))

    # Image descriptor
    left   = struct.unpack('>H', rng.read(2))[0] % width
    top    = struct.unpack('>H', rng.read(2))[0] % height
    width  = struct.unpack('>H', rng.read(2))[0] % (width - left)
    height = struct.unpack('>H', rng.read(2))[0] % (height - top)
    local_color_table_flag = bool(struct.unpack('B', rng.read(1))[0] & 0x80)
    interlace_flag = bool(struct.unpack('B', rng.read(1))[0] & 0x40)
    sort_flag = bool(struct.unpack('B', rng.read(1))[0] & 0x20)
    reserved = struct.unpack('B', rng.read(1))[0] & 0x1f

    out.write(left.to_bytes(2, 'big'))
    out.write(top.to_bytes(2, 'big'))
    out.write(width.to_bytes(2, 'big'))
    out.write(height.to_bytes(2, 'big'))
    out.write((local_color_table_flag << 7).to_bytes(1, 'big'))
    out.write((interlace_flag << 6).to_bytes(1, 'big'))
    out.write((sort_flag << 5).to_bytes(1, 'big'))
    out.write(reserved.to_bytes(1, 'big'))

    # Local color table
    if local_color_table_flag:
        for _ in range(pow(2, color_depth)):
            r = struct.unpack('B', rng.read(1))[0]
            g = struct.unpack('B', rng.read(1))[0]
            b = struct.unpack('B', rng.read(1))[0]

            out.write(r.to_bytes(1, 'big'))
            out.write(g.to_bytes(1, 'big'))
            out.write(b.to_bytes(1, 'big'))

    # LZW minimum code size
    lzw_minimum_code_size = struct.unpack('B', rng.read(1))[0] & 0x07

    out.write(lzw_minimum_code_size.