#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    header = b'GIF89a'

    # 16-bit width and height
    w = rng.read(2)
    h = rng.read(2)

    # packed field of color resolution (bits per pixel)
    r = rng.read(1)[0] & 7
    g = rng.read(1)[0] & 7
    b = rng.read(1)[0] & 7
    packed = bytes([r | (g << 3) | (b << 6)])

    # background color index
    bg = rng.read(1)[0]

    # aspect ratio
    ar = rng.read(1)[0] & 0xf

    # image descriptor
    out.write(header + w + h + packed +
             bytes([bg]) + bytes([ar]))

    # LZW minimum code size
    mincodesize = rng.read(1)[0]
    out.write(bytes([mincodesize]))

    # image data
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # convert to lzw encoded form
    compressed = compress(data)

    # write it all out
    out.write(compressed)


# Compresses an input string with lzw encoding
def compress(input: bytes) -> bytes:
    output = []
    dict_size = 256
    dictionary = [chr(i) for i in range(dict_size)]
    word = ''
    word_len = 0

    def add_to_dictionary(c):
        if len(dictionary) == dict_size:
            return False

        dictionary += c
        return True

    def get_index(c):
        try:
            return dictionary.index(c)
        except ValueError:
            return -1

    for char in input:
        wc = word + chr(char)
        if not add_to_dictionary(wc):
            idx = get_index(word[0])
            output.append(idx)

            while len(output) >= 128:
                out = bytes([(len(output) // 256) | 0x80])
                out += bytes([len(output) % 256])

                for o in output[:128]:
                    out += bytes([(o >> 8) & 0xff])
                    out += bytes([o & 0xff])

                output = output[128:]

                yield out

            dictionary = dictionary[-dict_size:]
            dict_size *= 4
            add_to_dictionary(wc)

        word = wc
        word_len += 1

    if word!= '':
        idx = get_index(word[0])
        output.append(idx)

        while len(output) >= 128:
            out = bytes([(len(output) // 256) | 0x80])
            out += bytes([len(output) % 256])

            for o in output[:128]:
                out += bytes([(o >> 8) & 0xff])
                out += bytes([o & 0xff])

            output = output[128:]

            yield out

    out = bytes([(len(output) // 256) | 0x80])
    out += bytes([len(output) % 256])

    for o in output:
        out += bytes([(o >> 8) & 0xff])
        out += bytes([o & 0xff])

    yield out