#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = rng.randint(100, 256)
    height = rng.randint(100, 256)
    flags  = (
            0x80 | # Global color table
            0x70   # Color resolution
        ) << 4
    bgcol  = rng.randint(0, 256)
    aspect = 0x00
    out.write(struct.pack('<HHBBB', width, height, flags, bgcol, aspect))

    # Global color table
    for _ in range(0, 2**flags & 0x07):
        r = rng.randint(0, 256)
        g = rng.randint(0, 256)
        b = rng.randint(0, 256)
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left    = 0
    top     = 0
    width   = rng.randint(width // 2, width * 2 - 1)
    height  = rng.randint(height // 2, height * 2 - 1)
    flags   = 0x00
    lctsize = 0x00
    out.write(struct.pack('<HHHHB', left, top, width, height, flags))

    # Local color table
    if lctsize!= 0x00:
        for _ in range(lctsize):
            r = rng.randint(0, 256)
            g = rng.randint(0, 256)
            b = rng.randint(0, 256)
            out.write(struct.pack('BBB', r, g, b))

    # Table-based image data
    lzwmincodesize = 2
    while True:
        codesize = rng.randint(lzwmincodesize, 12)

        # End of image marker
        if codesize == 12 and rng.randint(0, 1) == 0:
            break

        # LZW minimum clear code
        elif codesize == 12 and rng.randint(0, 1) == 0:
            out.write(b'\x00\x00')

        # LZW maximum clear code
        elif codesize == 12 and rng.randint(0, 1) == 0:
            out.write(b'\xFF\x00')

        # LZW end of information code
        elif codesize == 12 and rng.randint(0, 1) == 0:
            out.write(b'\x00\x00')

            # Block terminator
            out.write(b'\x00')
            continue

        # Other codes
        else:
            count = rng.randint(0, 256)
            out.write(count.to_bytes(1, 'little'))
            for i in range(count):
                out.write((i % 256).to_bytes(1, 'little'))