#!/usr/bin/env python3

import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    header = b'GIF89a'

    # 16-bit width and height
    w = rng.read(2)
    h = rng.read(2)

    # packed field of color resolution (bits per pixel)
    r = rng.read(1)[0] & 7
    g = rng.read(1)[0] & 7
    b = rng.read(1)[0] & 7
    packed = bytes([r | (g << 3) | (b << 6)])

    # background color index
    bg = rng.read(1)[0]

    # aspect ratio
    ar = rng.read(1)[0] & 0xf

    # image descriptor
    out.write(header + w + h + packed +
             bytes([bg]) + bytes([ar]))

    # LZW minimum code size
    mincodesize = rng.read(1)[0]
    out.write(bytes([mincodesize]))

    # image data
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # LZW compression
    lzw_compressed = []
    lzw_dict = {}
    lzw_dict[b'\x00'] = b'\x00'
    lzw_dict[b'\xff'] = b'\xff'

    def add_to_lzw_dict(key):
        if key not in lzw_dict:
            lzw_dict[key] = bytes([len(lzw_dict)])

    while len(data) > 0:
        c = data[:256]

        for i in range(0, len(c)):
            add_to_lzw_dict(c[i:i+1])

        data = data[256:]

    # The first character is always the empty string
    lzw_compressed += [b'\x00']

    while True:
        try:
            chunk = data[:256]

            for i in range(0, len(chunk)):
                lzw_compressed += [
                    lzw_dict[chunk[i:i+1]]
                ]

                add_to_lzw_dict(chunk[i:i+1])

            data = data[256:]
        except IndexError:
            break

    # Write the compressed data to the output stream
    for i in range(0, len(lzw_compressed)):
        out.write(lzw_compressed[i])