#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), 'big')])
    bgcolorindex = bytes([int.from_bytes(rng.read(1), 'big')])
    aspectratio = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(struct.pack('<HHBBB', width, height, *flags, *bgcolorindex, *aspectratio))

    # Global color table
    globalcolortablesize = int.from_bytes(rng.read(1), 'big') & 7
    globalcolortable = [rng.read(3) for _ in range(globalcolortablesize)]
    out.write(bytes([(globalcolortablesize << 5)]))
    for col in globalcolortable:
        out.write(col)

    # Image descriptor
    left = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    top = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    width = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), 'big')])
    bgcolorindex = bytes([int.from_bytes(rng.read(1), 'big')])
    pixelaspectratio = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(struct.pack('<HHHHBBBBB', left, top, width, height, *flags, *bgcolorindex, *pixelaspectratio))

    # Local color table
    localcolortablesize = int.from_bytes(rng.read(1), 'big') & 7
    localcolortable = [rng.read(3) for _ in range(localcolortablesize)]
    out.write(bytes([(localcolortablesize << 5)]))
    for col in localcolortable:
        out.write(col)

    # Image data
    lzwmincodesize = int.from_bytes(rng.read(1), 'big')
    lzwwords = []
    while True:
        wordlen = int.from_bytes(rng.read(1), 'big')
        if not wordlen:
            break
        elif wordlen == 255:
            continue

        lzwword = bytearray()
        for i in range(wordlen):
            lzwword += bytes([int.from_bytes(rng.read(1), 'big')])
        lzwwords.append(lzwword)

    lzwmaxcodesize = max((len(w).bit_length()) for w in lzwwords)
    lzwmaxcodesize = min(lzwmaxcodesize, 12)
    lzwmaxcodesize = max(lzwmaxcodesize, lzwmincodesize)
    out.write(bytes([lzwmincodesize]))

    codesize = lzwmincodesize + 1
    clearcode = 1 << codesize
    endofinformationcode = clearcode + 1
    nextcode = clearcode + 2
    codetable = {clearcode: b'\x00\x00', endofinformationcode: b'\xff\xff'}
    for lzwword in lzwwords:
        for c in lzwword:
            if len(codetable[nextcode]) < codesize:
                codetable[nextcode] += bytes([c])
            else:
                codetable[nextcode] = bytes([c])

            nextcode += 1
            if nextcode >= (1 << codesize):
                codesize += 1
                nextcode = clearcode + 2

                if codesize > lzwmaxcodesize:
                    raise Exception('