#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # The first 6 bytes are always "GIF87a" or "GIF89a", depending on whether
    # we're generating an old-style (version 87a) or new-style (version 89a)
    # GIF. We'll just use version 89a for now since it's more modern.
    header = b'GIF89a'
    # Next comes the logical screen descriptor. This is where we specify how
    # many pixels wide our image will be, as well as its height in pixels.
    # Since this is a random image, let's make it relatively small so that we
    # can see what it looks like when displayed.
    width = 10
    height = 10
    # Then there's a color table. In a normal GIF, this would contain all the
    # colors used by each pixel in the image. However, since we don't have any
    # transparency, we only need one entry here - the background color.
    bgcolor = 0x000000FF
    colortable = [bgcolor] * 256
    # Finally, we come to the graphic control extension. This tells us how long
    # the animation should last before looping back to the beginning, as well
    # as which color index from the palette should be treated as transparent.
    # Since we want the entire image to be visible at once, we'll set the delay
    # time to zero, and tell it not to loop.
    delaytime = 0
    transindex = 0
    gce = struct.pack('!BBHBHHB',
            0x21, 0xF9, 4,
            delaytime, transindex, 0b10000000, 0
        )
    # Now we get to the actual image data. First up is the logical screen
    # descriptor again, but with different values. These describe the size of
    # the image in blocks instead of pixels, and also include some flags about
    # the way the image was encoded. For now, we'll just leave these blank.
    lsd = struct.pack('!HHBBB',
            width, height, 0, 0, 0
        )
    # After that, we have the global color table. This contains the same entries
    # as the local color table, but applies to every single block in the image.
    # Since we don't have any transparency, we only need one entry here - the
    # background color.
    gct = struct.pack('!BBBB',
            *(colortable[i] & 0xFF for i in range(len(colortable)))
        )
    # Finally, we have the image descriptor. This describes the position of the
    # image within the logical screen, as well as its local color table. Again,
    # we only have one entry in the local color table, so we'll just duplicate
    # the value from above.
    id = struct.pack('!HHBBB',
            0, 0, 0, 0, 0
        )
    # Now we start writing everything to the output stream.
    out.write(header)
    out.write(lsd)
    out.write(gct)
    out.write(id)
    out.write(gce)
    # Now we write the actual image data. Each block consists of two parts - a
    # subblock containing the LZW codestream for the current row, followed by
    # another subblock containing the corresponding bitmap data for the current
    # row.
    #
    # The LZW codestream starts off with a clear code, then continues with a
    # series of codes representing individual characters in the image. When the
    # end of the line is reached, a flush code is sent to indicate that no more
    # codes follow.
    #
    # The bitmap data consists of a series of bitplanes, each of which begins
    # with a byte indicating the length of the following bitmap data. If the
    # length is zero, that means that the rest of the row is empty. Otherwise,
    # the next N bits represent the color index of the pixel at column N.
    #
    # Note that the order of the rows is reversed compared to the order they
    # appear in the image. This is because the GIF format requires that the
    # image data be written bottom-up rather than top-down.
    def encode_row(row):
        # Start with the clear code
        yield 0b11