#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # get some of our data into the file before writing the rest.
    header = b'GIF89a'
    width = 256
    height = 256
    color_table_size = 7
    global_color_table = [
            0x00, 0x00, 0x00,
            0xff, 0x00, 0x00,
            0x00, 0xff, 0x00,
            0xff, 0xff, 0x00,
            0x00, 0x00, 0xff,
            0xff, 0x00, 0xff,
            0x00, 0xff, 0xff]
    local_color_table = []
    for i in range(0, (1 << color_table_size)):
        local_color_table += [
                *global_color_table[i*3:(i+1)*3]]
    background_index = 0
    aspect_ratio = 4
    flags = 0b00000000
    gct_flag = True if len(local_color_table) > 0 else False
    lct_flag = True if len(local_color_table) > 0 else False
    sorted_gct_flag = False
    interlace_flag = False
    palette_size = 2 ** color_table_size
    transparent_index = 0
    pixel_aspect_ratio = 0
    screen_descriptor = bytes([
            width & 0xff,
            (width >> 8) & 0xff,
            height & 0xff,
            (height >> 8) & 0xff])
    screen_descriptor += bytes([(flags << 2) | color_table_size])
    out.write(header + screen_descriptor)

    # Write the GCT (Global Color Table).
    if gct_flag:
        out.write(struct.pack('<B', palette_size))
        for rgb in local_color_table:
            out.write(rgb.to_bytes(1, 'little'))

    # Write the LCT (Local Color Table).
    if lct_flag:
        out.write(struct.pack('<B', palette_size))
        for rgb in local_color_table:
            out.write(rgb.to_bytes(1, 'little'))

    # Write the image descriptor block.
    left = 0
    top = 0
    width -= 1
    height -= 1
    packed_fields = ((left & 0xf8) << 8) \
                    | ((top & 0xfe) << 1) \
                    | ((width & 0x07) << 5) \
                    | ((height & 0x07) << 2) \
                    | ((packed_fields & 0xc0) << 1)
    out.write(struct.pack('<BBHHHBB',
            packed_fields,
            background_index,
            transparent_index,
            left,
            top,
            width,
            height))

    # Write the LZW code size.
    out.write(b'\x08')

    # Write the pixels themselves.
    for y in range(0, height):
        for x in range(0, width):
            out.write((y % 2 ^ x % 3).to_bytes(1, 'little'))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)