#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, width: int=256, height: int=256
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # get some of our data into the file before writing the rest.
    header = b'GIF89a'
    datalen = (width * height // 8) + 7
    out.write(header + datalen.to_bytes(3, 'big'))

    # Write the logical screen descriptor.
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    out.write(b'\x87\x00')     # Global color table flag set
    out.write(b'\x00')         # Background color index
    out.write(b'\x00')         # Pixel aspect ratio

    # Write the global color table.
    for i in range(0, 256):
        r = rng.read(1)[0] & ~0x40
        g = rng.read(1)[0] & ~0x40
        b = rng.read(1)[0] & ~0xc0

        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Write the image descriptor.
    out.write(b'\x2c')             # Image separator
    out.write(((width >> 7) | 0x80).to_bytes(1, 'little'))   # Left position
    out.write(((height >> 7) | 0x80).to_bytes(1, 'little'))  # Top position
    out.write(((width - 1) & 0xff).to_bytes(1, 'little'))      # Width
    out.write(((height - 1) & 0xff).to_bytes(1, 'little'))     # Height
    out.write(b'\x00')             # Local color table flag not set
    out.write(b'\x00')             # Interlace flag not set
    out.write(b'\x00')             # Sort flag not set
    out.write(b'\x00')             # Reserved

    # Write the local color table.
    for i in range(0, 256):
        r = rng.read(1)[0] & ~0x40
        g = rng.read(1)[0] & ~0x40
        b = rng.read(1)[0] & ~0xc0

        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Finally, write the pixel data.
    for y in range(0, height):
        for x in range(0, width):

            if ((y % 8) == 0):
                byte = rng.read(1)[0]

            mask = 0x80 >> (x % 8)
            out.write(byte.to_bytes(1, 'little') if (byte & mask) else b'\x00')


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)