#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), 'big')])
    bgcolor = bytes([int.from_bytes(rng.read(1), 'big')])
    aspect = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(struct.pack('<HHBBB', width, height, *flags, bgcolor[0], aspect[0]))

    # Global color table
    global_colors = [
        bytes([int.from_bytes(rng.read(1), 'big'),
               int.from_bytes(rng.read(1), 'big'),
               int.from_bytes(rng.read(1), 'big')]) for _ in range(256)]
    out.write(b''.join(global_colors))

    # Image descriptor
    left = int.from_bytes(rng.read(2), 'big')
    top = int.from_bytes(rng.read(2), 'big')
    width = int.from_bytes(rng.read(2), 'big') % (width - left) + 1
    height = int.from_bytes(rng.read(2), 'big') % (height - top) + 1
    flags = bytes([int.from_bytes(rng.read(1), 'big')])
    lctsize = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(struct.pack('<HHHHB', left, top, width, height, *flags))
    if lctsize!= b'\x00':
        raise ValueError('LCT size must be zero!')

    # Local color table
    local_colors = [
        bytes([int.from_bytes(rng.read(1), 'big'),
               int.from_bytes(rng.read(1), 'big'),
               int.from_bytes(rng.read(1), 'big')]) for _ in range(256)]
    out.write(b''.join(local_colors))

    # Image data
    data = []
    while True:
        blocktype = rng.read(1)[0]

        if blocktype == 0x2c:
            # Image separator
            pass
        elif blocktype == 0x21:
            # Extension
            exttype = rng.read(1)[0]

            if exttype == 0xf9:
                # Graphic control extension
                flags = bytes([int.from_bytes(rng.read(1), 'big')])
                delay = int.from_bytes(rng.read(2), 'big')
                transparent = bytes([int.from_bytes(rng.read(1), 'big')])

                if len(transparent) > 0 or flags[0] & 1:
                    print("Warning: Transparent color is not supported!")

                if flags[0] & 4:
                    print("Warning: User input flag is set!")

                if flags[0] & 2:
                    print("Warning: Disposal method is not supported!")

                out.write(struct.pack('<BBHB', exttype, *flags, delay))
                out.write(transparent)
            else:
                # Unknown extension type!
                print("Warning: Unknown extension type")
        elif blocktype == 0x3b:
            break
        else:
            raise ValueError(f"Unknown block type {blocktype}")

    # Trailer
    out.write(b';')