#!/usr/bin/env python3

import argparse
import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # get some of our data into the file before writing the rest.
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    out.write(header + datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    for _ in range(0, datalen):
        out.write(rng.read(1))


def main() -> int:
    parser = argparse.ArgumentParser(description="Generate a random.gif")
    parser.add_argument("outfile", type=str, help="The output filename.")
    args = parser.parse_args()

    if not args.outfile or len(args.outfile) < 5:
        print("Please specify an output filename.", file=sys.stderr)
        return -1

    try:
        with open('/dev/urandom', mode='rb') as rng:
            with open(args.outfile, mode='wb') as f:
                generate_random_gif(rng, f)
    except OSError as e:
        print("Error opening {}: {}".format(args.outfile, str(e)),
              file=sys.stderr)
        return -2

    return 0


if __name__ == '__main__':
    exitcode = main()
    sys.exit(exitcode)