#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (e.g., /dev/urandom).
# The size of each frame is between 1 and max_frame_size bytes long.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO, max_frame_size=256
):
    # Header
    out.write('GIF89a'.encode())

    # Logical screen descriptor
    width = rng.randint(0x1, 0xFFFF)
    height = rng.randint(0x1, 0xFFFF)
    flags = 0b10000000 | rng.randint(0x0, 0xF)
    bgcolorindex = rng.randint(0x0, 0xFF)
    pixelaspectratio = 0x0
    out.write(struct.pack('<HHBBB', width, height, flags, bgcolorindex,
                          pixelaspectratio))

    # Global color table
    numcolors = min(max_frame_size // 4, 0xFF)
    for i in range(numcolors):
        r = rng.randint(0x0, 0xFF)
        g = rng.randint(0x0, 0xFF)
        b = rng.randint(0x0, 0xFF)
        out.write(struct.pack('<BBBB', r, g, b, 0x0))

    # Frames
    while True:
        # Image descriptor
        left = rng.randint(0x0, 0xFFFF)
        top = rng.randint(0x0, 0xFFFF)
        width = rng.randint(0x1, 0xFFFF)
        height = rng.randint(0x1, 0xFFFF)
        flags = 0b10000000 | rng.randint(0x0, 0xF)
        localcolortableflag = bool(flags & 0b10000000)
        interlaceflag = bool(flags & 0b01000000)
        sortflag = bool(flags & 0b00100000)
        reserved = rng.randint(0x0, 0x7)
        lctsize = rng.randint(0x0, 0xFF) if localcolortableflag else 0x0
        out.write(struct.pack('<HHHHBB', left, top, width, height, flags,
                              lctsize))

        # Local color table
        if localcolortableflag:
            for i in range(lctsize):
                r = rng.randint(0x0, 0xFF)
                g = rng.randint(0x0, 0xFF)
                b = rng.randint(0x0, 0xFF)
                out.write(struct.pack('<BBBB', r, g, b, 0x0))

        # LZW minimum code size
        lzwmincodesize = rng.randint(0x2, 0x8)
        out.write(struct.pack('<B', lzwmincodesize))

        # Data sub-blocks
        blocklen = rng.randint(0x1, max_frame_size - 1)
        while blocklen > 0:
            blockdata = rng.read(blocklen)
            out.write(struct.pack('<B', len(blockdata)))
            out.write(blockdata)

            blocklen -= len(blockdata) + 1

        # Trailer
        out.write(b'\x00')

        # End of image?
        if not blocklen: break

