#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (e.g., /dev/urandom).
# The generated image has dimensions `width` x `height`,
# and is composed of `num_colors` distinct colors.
def generate_random_gif(
        rng: BinaryIO, width: int, height: int, num_colors: int, out: BinaryIO
    ) -> None:

    # Write the GIF header to the output stream.
    def write_header() -> None:
        out.write('GIF89a'.encode())

        # Compute the global color table size.
        global_color_table_size = 2 ** ((num_colors - 1).bit_length())

        # Write the logical screen descriptor to the output stream.
        out.write((
            0x21, 0xF9, 0x04,    # Extension block introducer & label.
            0x08,                # Block length.
            0x00, 0x00,          # Background color index.
            0x00,                # Pixel aspect ratio.
            0x00 |              # Global color table flag.
                (global_color_table_size >> 7) & 0x1,
            0x00 | (global_color_table_size << 1) & 0xFF,

            # Write the global color table.
            *([0] * (3 * global_color_table_size))
        ))

        # Write the local color table.
        for i in range(num_colors):
            rgb = [int(c) for c in rng.read(3)]
            out.write(rgb)

        # Write the graphics control extension to the output stream.
        out.write((
            0x21, 0xF9, 0x04,    # Extension block introducer & label.
            0x0A,                # Block length.
            0x00,                # Flags.
            0x00, 0x00,          # Delay time.
            0x00,                # Transparent color index.
            0x00                 # Terminator.
        ))

        # Write the image descriptor to the output stream.
        out.write((
            0x2C,                # Image separator.
            0x00, 0x00,          # Left position.
            0x00, 0x00,          # Top position.
            width % 0x100,       # Width.
            width // 0x100,      # Height.
            0x00,                # Local color table flag.
            0x00                 # LZW minimum code size.
        ))

    # Write the palette-based image data to the output stream.
    def write_palette_data():
        # Initialize the current pixel with an invalid value so that we can
        # detect when it's been set to something valid by the end of this loop.
        cur_pixel = (-1, -1)

        # Loop through all pixels in the image.
        for y in range(height):
            for x in range(width):

                # Choose a new color from our list of available colors.
                color = next(available_colors)

                # Set the current pixel to the chosen color.
                cur_pixel = (x, y)

                # Write the pixel to the output stream.
                out.write(struct.pack('<BBB', *color))

        # Check if the last pixel was written successfully.
        assert cur_pixel == (width - 1, height - 1)


    # Generate the initial color map.
    available_colors = iter([(r, g, b) for r in range(num_colors)
                             for g in range(num_colors)
                             for b in range(num_colors)])

    # Write the GIF header to the output stream.
    write_header()

    # Write the palette-based image data to the output stream.
    write_palette_data()