#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    header = b'GIF89a'
    out.write(header)

    # Logical screen descriptor
    width = (int.from_bytes(rng.read(2), 'big') % 65536) + 1
    height = (int.from_bytes(rng.read(2), 'big') % 65536) + 1
    flags = bytes([0])
    bgcolorindex = bytes([0])
    pixelaspectratio = bytes([0])
    lsd = [width, height] + list(flags) + list(bgcolorindex) + list(pixelaspectratio)
    lsdlen = len(lsd)
    out.write(struct.pack('B' * lsdlen, *lsd))

    # Global color table
    global_colors = []
    for i in range(7):
        r = int.from_bytes(rng.read(1), 'big') & 0xff
        g = int.from_bytes(rng.read(1), 'big') & 0xff
        b = int.from_bytes(rng.read(1), 'big') & 0xff
        global_colors += [r, g, b]
    global_colors += [0] * ((2 << lsd[7]) - len(global_colors))
    out.write(bytes(global_colors))

    # Image descriptor
    left = bytes([0])
    top = bytes([0])
    width = bytes([(width >> 8) & 0xff]) + bytes([width & 0xff])
    height = bytes([(height >> 8) & 0xff]) + bytes([height & 0xff])
    flags = bytes([0])
    imd = list(left) + list(top) + list(width) + list(height) + list(flags)
    imdlen = len(imd)
    out.write(struct.pack('B' * imdlen, *imd))

    # Local color table
    local_colors = []
    for i in range(3):
        r = int.from_bytes(rng.read(1), 'big') & 0xff
        g = int.from_bytes(rng.read(1), 'big') & 0xff
        b = int.from_bytes(rng.read(1), 'big') & 0xff
        local_colors += [r, g, b]
    local_colors += [0] * ((2 << lsd[7]) - len(local_colors))
    out.write(bytes(local_colors))

    # Image data
    image_data = []
    while True:
        chunksize = int.from_bytes(rng.read(1), 'big')
        if not chunksize:
            break

        chunktype = int.from_bytes(rng.read(1), 'big')
        if chunktype == 0x2c or chunktype == 0x21:
            image_data += [chunksize]
            image_data += [chunktype]

            subchunks = []
            while True:
                subchunksize = int.from_bytes(rng.read(1), 'big')
                if not subchunksize:
                    break

                subchunktype = int.from_bytes(rng.read(1), 'big')
                subchunkdata = rng.read(subchunksize)
                subchunks += [subchunksize]
                subchunks += [subchunktype]
                subchunks += list(subchunkdata)

            image_data += subchunks
        else:
            raise Exception("Unknown chunk type")

    out.write(bytes(image_data))

    # Trailer
    trailer = b'\x3b'
    out.write(trailer)