#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF has no global color table and is 1024x768 pixels large.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1024
    height = 768
    packed = (
            0b10000000 |     # Global color table flag set
            0b00000000 |     # Sort flag clear
            0b00000000 |     # Size of global color table
            0b00000000       # Background color index
    ) << 5 | (
            0b00000000 |     # Pixel aspect ratio
            0b00000000 |     # Color resolution
            0b00000000 |     # Globla color table size minus one
            0b00000000       # Sort flag clear
    )
    out.write(struct.pack('<HHBBB', width, height, packed, 0))

    # Global color table
    for i in range(0, 256 * 3 // 4):
        out.write(rng.read(1))

    # Image descriptor
    packed = (
            0b00000000 |     # Local color table flag set
            0b00000000 |     # Interlace flag clear
            0b00000000 |     # Sort flag clear
            0b00000000 |     # Reserved
            0b00000000 |     # LZW minimum code size
            0b00000000       # Left position
    ) << 5 | (
            0b00000000 |     # Top position
            0b00000000 |     # Width - 1
            0b00000000 |     # Height - 1
            0b00000000       # Packed field
    )
    out.write(struct.pack('<HHBBBB', 0, 0, width - 1, height - 1, packed, 0))

    # Image data
    datalen = width * height
    while datalen > 0:
        chunklen = min(datalen, 255)
        out.write(chunklen.to_bytes(1, 'little'))
        out.write(rng.read(chunklen))
        datalen -= chunklen

    # Trailer
    out.write(b'\x3B')