#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # get some of our data into the file before writing the rest.
    header = b'GIF89a'
    width = 256
    height = 256
    color_table_size = 7
    global_color_table = [
            0x00, 0x00, 0x00,
            0xFF, 0x00, 0x00,
            0x00, 0xFF, 0x00,
            0xFF, 0xFF, 0x00,
            0x00, 0x00, 0xFF,
            0xFF, 0x00, 0xFF,
            0x00, 0xFF, 0xFF]
    background_index = 0
    pixel_aspect_ratio = 0
    flags = (b'\x80\x00' if color_table_size > 0 else b'') \
            + (b'\x40\x00' if global_color_table else b'') \
            + (b'\x20\x00' if background_index!= 0 else b'') \
            + (b'\x10\x00' if pixel_aspect_ratio!= 0 else b'')
    screen_descriptor = bytes([width & 0xff,
                               (width >> 8) & 0xff,
                               height & 0xff,
                               (height >> 8) & 0xff]) \
                       + ((flags).ljust(3, b'\x00')[::-1]) \
                       + (b'\x00' if color_table_size < 2**8 else
                          bytes([(color_table_size >> 8) & 0xff]))
    logical_screen_descriptor = screen_descriptor \
                                + bytes([background_index,
                                         pixel_aspect_ratio])
    out.write(header + len(logical_screen_descriptor).to_bytes(3, 'big'))
    out.write(logical_screen_descriptor)

    # If there's a global color table, write it now.
    if global_color_table:
        for entry in global_color_table:
            out.write(entry.to_bytes(1, 'little'))

    # Now we can write the actual data to the file.
    palette_size = min((1 << color_table_size), 2**8)
    for y in range(0, height):
        for x in range(0, width):
            index = rng.read(1)[0] % palette_size
            out.write(index.to_bytes(1, 'little'))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)