#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1024
    height = 768
    color_depth = 8
    global_color_table = False
    sort_flag = True
    size = (width * height).bit_length()
    background_color = 0
    aspect_ratio = 0
    out.write(struct.pack('<HHBBB',
                          width, height,
                          color_depth,
                          size,
                          global_color_table))
    if not global_color_table:
        raise NotImplementedError('Local color tables are not supported.')

    # Global color table
    for i in range(size**2 // 3):
        r = ord(os.urandom(1)) & ~0x80
        g = ord(os.urandom(1)) & ~0x80
        b = ord(os.urandom(1)) & ~0x80
        out.write(struct.pack('<BBB',
                              r, g, b))

    # Image descriptor
    left = top = 0
    width = 1024
    height = 768
    local_color_table = False
    interlace_flag = False
    sort_flag = True
    reserved = 0
    lct_size = 0
    out.write(struct.pack('<BBHHB',
                          0x2c,
                          left, top,
                          width, height,
                          local_color_table << 7 |
                          interlace_flag << 6 |
                          sort_flag << 5 |
                          reserved))
    if local_color_table:
        raise NotImplementedError('Local color tables are not supported.')

    # Local color table
    for i in range(lct_size ** 2 // 3):
        r = ord(os.urandom(1)) & ~0x80
        g = ord(os.urandom(1)) & ~0x80
        b = ord(os.urandom(1)) & ~0x80
        out.write(struct.pack('<BBB',
                              r, g, b))

    # Table based image data
    lzw_mincodesize = 2
    out.write(struct.pack('<B', lzw_mincodesize))
    clear_code = 1 << lzw_mincodesize
    end_of_information_code = clear_code + 1
    code_size = lzw_mincodesize + 1
    next_code = dict((i, i+1)
                     for i in range(clear_code - 1))
    codes = []
    while len(codes) < 256:
        code = ord(os.urandom(1))
        if code == end_of_information_code:
            break

        if code >= clear_code and code <= max(next_code.values()):
            codes += [code]
            continue

        if code == clear_code:
            next_code = dict((i, i+1)
                             for i in range(end_of_information_code-1))
            code_size = lzw_mincodesize + 1
            continue

        prefix = codes[code - clear_code - 1]
        suffix = ord(os.urandom(1))
        new_code = ((prefix << 8) |
                    suffix) % (1 << code_size)
        codes += [new_code]
        next_code[new_code] = len(codes)
        if code_size < 12 and \
           len(codes) > (1 << code_size):
            code_size += 1

    # LZW minimum code size
    out.write(struct.pack('<B', lzw_mincodesize))

    # Codes
    for c in sorted(set(codes)):
        out.write(struct.pack('<H', c))

    # End of information code
    out.write(struct.pack('<H', end_of_information_code))

