#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bool:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 16 + rng.read(1)[0] % 256
    height = 16 + rng.read(1)[0] % 256
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio))

    # Global color table
    for i in range(7):
        red = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue = rng.read(1)[0] % 256
        out.write(struct.pack('<BBB', red, green, blue))
    
    # Image descriptor
    left = rng.read(1)[0] % 256
    top = rng.read(1)[0] % 256
    width = 16 + rng.read(1)[0] % 256
    height = 16 + rng.read(1)[0] % 256
    local_color_table_flag = rng.read(1)[0] & 0x80 != 0
    interlace_flag = rng.read(1)[0] & 0x40 != 0
    sort_flag = rng.read(1)[0] & 0x20 != 0
    reserved_bits = rng.read(1)[0] & 0x1C != 0
    size_of_local_color_table = rng.read(1)[0] & 0x07
    out.write(struct.pack('<HHBB', left, top, width, height))
    out.write(struct.pack('<B', local_color_table_flag | interlace_flag | sort_flag | reserved_bits | size_of_local_color_table))

    # Local color table
    if local_color_table_flag:
        for i in range(2 ** (size_of_local_color_table + 1)):
            red = rng.read(1)[0] % 256
            green = rng.read(1)[0] % 256
            blue = rng.read(1)[0] % 256
            out.write(struct.pack('<BBB', red, green, blue))

    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0] % 256
    out.write(struct.pack('<B', lzw_minimum_code_size))

    # Pixel data
    while True:
        block_type = rng.read(1)[0]
        if block_type == 0x00:
            break
        
        elif block_type == 0x2c:
            # Image data
            image_data_length = int.from_bytes(rng.read(2), 'little')
            image_data = rng.read(image_data_length)
            out.write(struct.pack('<H', image_data_length))
            out.write(image_data)
            
        else:
            # Extension
            extension_function_code = rng.read(1)[0]
            extension_block_length = int.from_bytes(rng.read(2), 'little')
            extension_block = rng.read(extension_block_length)
            out.write(struct.pack('<BBH', block_type, extension_function_code, extension_block_length))
            out.write(extension_block)

    return True

