#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')
    
    # Logical Screen Descriptor
    width = 1 + rng.read(2)[0]
    height = 1 + rng.read(2)[0]
    packed = rng.read(1)[0]
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('<HBBB', width, height, packed, background))
    out.write(aspect_ratio)

    # Global Color Table
    for _ in range(2**((packed & 7)+1)):
        color = rng.read(3)
        out.write(color)

    # Image Data
    while True:
        block_type = rng.read(1)[0]
        
        if block_type == 0x2c:
            # Graphic Control Extension
            
            # Block size
            out.write(block_type)
            
            # Packed fields
            packed = rng.read(1)[0]
            out.write(packed)
            
            # Delay time
            delay = rng.read(2)
            out.write(delay)
            
            # Transparent color index
            transparent = rng.read(1)
            out.write(transparent)
            
            # Block terminator
            out.write(b'\x00')
        
        elif block_type == 0x21:
            # Application Extension
            
            # Block size
            out.write(block_type)
            
            # Application ID
            application_id = b'A' * 8
            out.write(application_id)
            
            # Authentication code
            authentication_code = b' ' * 3
            out.write(authentication_code)
            
            # Data sub-blocks
            data_sub_blocks = []
            while True:
                data_sub_block = rng.read(1)
                
                if not data_sub_block or data_sub_block[0] != 0xff:
                    break
                
                data_sub_blocks.append(data_sub_block)
            
            # Block terminator
            out.write(b'\x00')
            
            # Write data sub-blocks
            for data_sub_block in data_sub_blocks:
                out.write(data_sub_block)
        
        elif block_type == 0x3b:
            # Trailer
            out.write(block_type)
            return
        
        else:
            raise Exception("Invalid block type")

