#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    packed = rng.read(1)[0]
    color_table_size = 2 ** ((packed & 0b00000111) + 1)
    background_color = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('<HBBB', width, height, packed, background_color))

    # Global color table
    for _ in range(color_table_size):
        out.write(rng.read(3))

    # Application extension block
    application_extension_block_size = 11
    application_identifier = 'NETSCAPE2.0'
    application_authentication_code = b'\x03\x01'
    loop_count = 0
    out.write(struct.pack('<HH', application_extension_block_size, 0xFFFE))
    out.write(application_identifier.encode())
    out.write(application_authentication_code)
    out.write(struct.pack('<H', loop_count))

    # Image descriptor
    image_descriptor_size = 10
    left = 0
    top = 0
    width = width - 1
    height = height - 1
    local_color_table_flag = False
    interlace_flag = False
    sort_flag = False
    reserved = 0
    packed = (local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | reserved)
    out.write(struct.pack('<HHBB', image_descriptor_size, left, top, packed))

    # Local color table
    for _ in range(color_table_size):
        out.write(rng.read(3))

    # Raster data
    for y in range(height):
        lzw_minimum_code_size = 8
        out.write(struct.pack('B', lzw_minimum_code_size))
        lzw_data = []
        while True:
            code = int.from_bytes(rng.read(1), byteorder='big')
            if code == 0:
                break
            elif code <= len(lzw_data):
                prefix = lzw_data[-code:]
                suffix = [rng.read(1)]
                lzw_data += prefix + suffix
            else:
                lzw_data += [rng.read(1)]
