#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')
    
    # Logical Screen Descriptor
    width = 2**rng.read(1)[0] + 1
    height = 2**rng.read(1)[0] + 1
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 8
    out.write(struct.pack('<HHBB', width, height, background_color_index, pixel_aspect_ratio))

    # Global Color Table
    color_table_size = 2**rng.read(1)[0]
    if color_table_size > 0:
        for _ in range(3 * color_table_size):
            out.write(b'\xFF')

    # Image Descriptor
    left = rng.read(1)[0] % width
    top = rng.read(1)[0] % height
    width = rng.read(1)[0] % (width - left) + 1
    height = rng.read(1)[0] % (height - top) + 1
    local_color_table_flag = bool(rng.read(1)[0] & 1)
    interlace_flag = bool(rng.read(1)[0] & 2)
    sort_flag = bool(rng.read(1)[0] & 4)
    reserved_bits = rng.read(1)[0] >> 3
    out.write(struct.pack('<HHBBB', left, top, width, height, local_color_table_flag | interlace_flag << 1 | sort_flag << 2 | reserved_bits << 3))

    # Local Color Table
    if local_color_table_flag:
        color_table_size = 2**rng.read(1)[0]
        for _ in range(3 * color_table_size):
            out.write(b'\xFF')

    # LZW Minimum Code Size
    lzw_minimum_code_size = rng.read(1)[0] % 8
    out.write(lzw_minimum_code_size.to_bytes(1, 'big'))

    # Pixel Data
    while True:
        code_size = lzw_minimum_code_size + 1
        code_mask = (1 << code_size) - 1
        
        # Read first code
        code = rng.read(1)[0] & code_mask
        if code == 0:
            break
        out.write(code.to_bytes(1, 'big'))

        # Read subsequent codes
        while True:
            code = (code << 8) | rng.read(1)[0]
            
            # Check for clear code
            if code == ((1 << (code_size+1))-1):
                code_size = lzw_minimum_code_size + 1
                code_mask = (1 << code_size) - 1
                
                # Read next code
                code = rng.read(1)[0] & code_mask
                if code == 0:
                    break
                out.write(code.to_bytes(1, 'big'))
                continue
            
            # Check for end of stream
            elif code == ((1 << code_size)-1):
                break
            
            # Write code to output
            else:
                out.write((code & code_mask).to_bytes(1, 'big'))

