#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = rng.read(1)[0] % 40 + 10
    height = rng.read(1)[0] % 40 + 10
    packed = struct.pack('<HHBBB', width, height, 0xF7, 0x00, 0x00)
    out.write(packed)

    # Global color table
    for _ in range(0xFF):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        packed = struct.pack('BBB', r, g, b)
        out.write(packed)

    # Application extension
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01')
    out.write(struct.pack('>H', 0))

    # Loop forever
    out.write(b'\x00\x00')

    # Image block
    out.write(b'\x2c\x00\x00\x00\x00')

    # Local color table
    for _ in range(0xFF):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        packed = struct.pack('BBB', r, g, b)
        out.write(packed)

    # Image descriptor
    packed = struct.pack('<HHHH', 0, 0, width, height)
    out.write(packed)

    # LZW-compressed image data
    lzw_minimum_code_size = 8
    out.write(bytes([lzw_minimum_code_size]))
    clear_code = 1 << lzw_minimum_code_size
    end_of_information_code = clear_code + 1
    code_table = [chr(i) for i in range(clear_code, end_of_information_code)]
    code_table_index = end_of_information_code
    current_code_size = lzw_minimum_code_size + 1

    def write_bits(bits: int, bit_count: int):
        nonlocal out
        while bits >= 0x100:
            out.write(bytes([bits & 0xFF]))
            bits >>= 8
            bit_count -= 8
        if bit_count > 0:
            out.write(bytes([bits]))

    def flush():
        nonlocal out
        while code_table_index > end_of_information_code:
            code_table.append(code_table[code_table_index - 1])
            code_table_index += 1
        write_bits(end_of_information_code, current_code_size)
        write_bits(clear_code, current_code_size)
        out.write(b'\x00')

    def add_to_code_table(prefix: str, suffix: str):
        nonlocal code_table, code_table_index, current_code_size
        if code_table_index == 0xFFFF:
            flush()
            code_table = [chr(i) for i in range(clear_code, end_of_information_code)]
            code_table_index = end_of_information_code
            current_code_size = lzw_minimum_code_size + 1
        else:
            code_table.append(suffix)
            code_table_index += 1
        if code_table_index == (1 << current_code_size):
            current_code_size += 1
            write_bits(clear_code, current_code_size)

    def write_code(code: int):
        nonlocal out
        while code >= 0x100:
            out.write(bytes([code & 0xFF]))
            code >>= 8
        out.write(bytes([code]))

    def write_string(s: str):
        nonlocal code_table, code_table_index, current_code_size
        if s == '':
            return
        if len(s) == 1:
            add_to_code_table('', s)
            write_code(code_table_index - 1)
            return
        prefix, suffix = s[:-1], s[-1]
        try:
            index = code_table.index(prefix + suffix)
        except ValueError:
            add_to_code_table(prefix, suffix)
            write_code(code_table_index - 1)
        else:
            write_code(index)
        write_string(prefix)

    def get_pixel(x: int, y: int):
        return