#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Write the GIF header
    out.write(b'GIF89a')

    # Write the screen descriptor
    width = 1 + rng.read(2)[0] % 640
    height = 1 + rng.read(2)[0] % 480
    background_color_index = rng.read(1)[0] & 0xF
    pixel_aspect_ratio = rng.read(1)[0] & 0x7
    out.write(struct.pack('<HBB', width, height, background_color_index))

    # Write the global color table (all black)
    for i in range(1 << ((pixel_aspect_ratio + 1) * 3)):
        out.write(b'\x00\x00\x00')

    # Write the application extension block
    out.write(b'!\xFF\x0BNETSCAPE2.0\x03\x01')
    out.write(struct.pack('<H', 0))

    # Write the image descriptor
    out.write(b',')
    left = rng.read(2)[0] % width
    top = rng.read(2)[0] % height
    width = 1 + rng.read(2)[0] % (width - left)
    height = 1 + rng.read(2)[0] % (height - top)
    local_color_table_flag = rng.read(1)[0] & 0x80
    interlace_flag = rng.read(1)[0] & 0x40
    sort_flag = rng.read(1)[0] & 0x20
    reserved_bits = rng.read(1)[0] & 0x1C
    size_of_local_color_table = rng.read(1)[0] & 0x07
    out.write(struct.pack('<HHBB', left, top, width, height))
    out.write(struct.pack('B', local_color_table_flag | interlace_flag | sort_flag | reserved_bits | size_of_local_color_table))

    # Write the local color table (all black)
    for i in range(1 << (size_of_local_color_table + 1) * 3):
        out.write(b'\x00\x00\x00')

    # Write the LZW-compressed data stream
    clear_code = 1 << size_of_local_color_table
    end_of_information_code = clear_code + 1
    code_size = size_of_local_color_table + 1
    max_code_size = 12
    min_code_size = 2
    first_code = None
    old_code = None
    code_stack = []
    while True:
        # Read a new code from the RNG
        code = rng.read(code_size // 8)[0]
        if code >= end_of_information_code:
            break

        # If this is the first code of the image, write it to the output
        if first_code == None:
            first_code = code
            out.write(struct.pack('B', first_code))
            continue

        # If we have an old code, add the current code to the stack
        if old_code != None:
            code_stack.append(old_code)

        # If the code is the clear code, reset the state and write the clear code
        if code == clear_code:
            code_stack = []
            out.write(struct.pack('B', clear_code))
            continue

        # If the code is the end of information code, terminate the stream
        if code == end_of_information_code:
            break

        # If the code is too large, increase the code size
        if code > (1 << code_size) - 1:
            code_size += 1
            if code_size > max_code_size:
                raise Exception("Code size exceeded maximum")
            continue

        # If the code is not in the dictionary, add it to the dictionary
        if len(code_stack) <