#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = (rng.read(1)[0] % 50) + 1
    height = (rng.read(1)[0] % 50) + 1
    out.write(struct.pack('<HHBBB', width * 10, height * 10, 0xF0 | rng.read(1)[0], 0, 0))

    # Write the global color table
    for i in range(768):
        out.write(bytes([(i >> ((i // 3) * 8)) & 0xFF]))

    # Write the application extension block
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01')
    out.write(struct.pack('<I', 0))

    # Write the image descriptor
    out.write(b'\x2c\x00\x00\x00\x00')
    out.write(b'\x00')
    out.write(b'\x00')
    out.write(b'\x00')
    out.write(b'\x00')

    # Write the local color table
    for i in range(768):
        out.write(bytes([(i >> ((i // 3) * 8)) & 0xFF]))

    # Write the image data
    while True:
        lzw_min_code_size = rng.read(1)[0] & 0x0F
        if lzw_min_code_size == 0:
            break

        out.write(bytes([lzw_min_code_size]))

        code_size = lzw_min_code_size + 1
        clear_code = 1 << code_size
        end_of_information_code = clear_code + 1
        available_codes = 1 << code_size
        old_code = -1
        stack = []
        bits_in_stack = 0

        while True:
            # Read a new code
            code = 0
            code_bits_remaining = code_size + 1
            while code_bits_remaining > 0:
                if bits_in_stack < code_bits_remaining:
                    byte = rng.read(1)[0]
                    stack.append((byte >> (8 - bits_in_stack)) & 0xFF)
                    bits_in_stack += 8
                else:
                    stack.append((stack[-1] >> (code_bits_remaining - bits_in_stack)))
                    bits_in_stack -= code_bits_remaining

                code |= stack.pop() << (code_bits_remaining - 1)
                code_bits_remaining -= 1

            # Handle special codes
            if code == clear_code:
                code_size = lzw_min_code_size + 1
                available_codes = 1 << code_size
                old_code = -1
                continue
            elif code == end_of_information_code:
                break

            # Add the current code to the dictionary
            if old_code != -1 and available_codes < 4096:
                entry = [old_code]
                entry.extend(stack)
                stack = entry[:]

                if len(entry) <= code_size:
                    out.write(bytes([clear_code]))
                    code_size = lzw_min_code_size + 1
                    available_codes = 1 << code_size
                    old_code = -1
                    continue

                available_codes += 1

            # Output the current code
            if old_code != -1:
                out.write(bytes([old_code]))

            old_code = code

            # If we have filled up the buffer, write it out
            if len(stack) >= 255:
                out.write(bytes([len(stack)]))
                out.write(bytes(stack))
                stack = []

        # Flush any remaining data
        if len(stack) > 0:
            out.write(bytes([len(stack)]))
            out.write(bytes(stack))

    # Write the trailer
    out.write(b';')
