#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0]))

    # Global color table
    for i in range(768):
        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

    # Image descriptor
    left   = int.from_bytes(rng.read(2), byteorder='little') % (width - 1) + 1
    top    = int.from_bytes(rng.read(2), byteorder='little') % (height - 1) + 1
    width  = int.from_bytes(rng.read(2), byteorder='little') % ((width - left) // 2) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % ((height - top) // 2) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lct    = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    interlaced = False if not bool(flags[0] & 0x40) else True
    sort    = False if not bool(flags[0] & 0x20) else True
    local_color_table_size = lct[0]
    background_index      = int.from_bytes(rng.read(1), byteorder='little')
    pixel_aspect_ratio    = int.from_bytes(rng.read(1), byteorder='little')
    out.write(struct.pack('<HHHHBB', left, top, width, height, flags[0], lct[0]))

    # Local color table
    for i in range((local_color_table_size ** 3)):
        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

    # Block size
    block_size = int.from_bytes(rng.read(1), byteorder='little')
    while block_size!= 0:

        # Extension
        label = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        if label == b'\xff':
            # Graphic control extension
            block_size -= 5
            ext_label = bytes([int.from_bytes(rng.read(1), byteorder='little')])
            delay = int.from_bytes(rng.read(2), byteorder='little')
            transp = int.from_bytes(rng.read(1), byteorder='little')
            disposal = int.from_bytes(rng.read(1), byteorder='little')
            out.write(b'\x21\xF9\x04' + struct.pack('<HBB', delay, transp, disposal))

        elif label == b'\xf9':
            # Comment extension
            block_size -= 1
            ext_label = bytes([int.from_bytes(rng.read(1), byteorder='little')])
            comment = rng.read(block_size-1)
            out.write(b'\x21\xFE\x00' + comment)

        block_size = int.from_bytes(rng.read(1), byteorder='little')

    # Image data
    for y in range(height):
        for x in range(width):
            index = int.from_bytes(rng.read(1), byteorder='little')
            out.write(bytes([index]))

