#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    out.write(struct.pack('<H', 0x200))
    out.write(b'\x01')
    out.write(bytes([rng.read(1)[0]]))
    out.write(b'\xFF\xFF\xFF')

    # Write the global color table
    for _ in range(768):
        out.write(bytes([rng.read(1)[0]]))

    # Write the application extension block
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00\x00')

    # Write the image descriptor
    out.write(b'\x2c\x00\x00\x00\x00\x00\x00\x00\x00')

    # Write the local color table
    for _ in range(768):
        out.write(bytes([rng.read(1)[0]]))

    # Write the image data
    out.write(b'\x00\x00')

    # Write the LZW compressed image data
    lzw_min_code_sizes = [8, 4, 2, 0]
    lzw_max_code_size = max(lzw_min_code_sizes) + 1

    # The first byte of the image data must not be 0
    out.write(bytes([rng.read(1)[0]]).replace(b'\x00', b'\x01'))

    while True:

        # Generate a new code size if needed
        if len(lzw_min_code_sizes)!= 0 and \
                (rng.read(1)[0] % 5 == 0 or
                 len(lzw_min_code_sizes) == lzw_max_code_size - 1):

            lzw_min_code_sizes.pop()
            lzw_max_code_size -= 1

            if len(lzw_min_code_sizes) == 0:

                # If we have no more minimum code sizes left then stop
                break

            else:

                # Otherwise write another control character
                out.write(b'\x00')

        # Read the next byte from /dev/urandom
        c = bytes([rng.read(1)[0]])

        # Check if it's a control character
        if c[0] >= 0x80:

            # It is so write it directly as-is
            out.write(c)

        elif c[0] < 0x40:

            # It isn't so write an escape sequence
            out.write(b'\x00')
            out.write(c)

        else:

            # It is a literal value so write its length minus one times
            # the current code size followed by that many copies of itself
            out.write((c[0] & 0x3f).to_bytes(1, 'little'))
            out.write(c * ((c[0] & 0x3f) + 1))

    # Write the trailer
    out.write(b';')
