#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO):
    # Header
    header = b'GIF89a' + bytes([0x12, 0x34])

    # Logical screen descriptor
    width = rng.read(1)[0] % 65536
    height = rng.read(1)[0] % 65536
    flags = rng.read(1)[0] & 0b11111000 | 0b00000100
    bgColorIndex = rng.read(1)[0] % 256
    pixelAspectRatio = rng.read(1)[0] % 256
    logicalScreenDescriptor = \
            struct.pack('<HHBBB',
                width, height, flags, bgColorIndex, pixelAspectRatio)

    # Global color table
    globalColorTableSize = rng.read(1)[0] % 2 ** 2
    globalColorTable = bytearray()
    for i in range(globalColorTableSize * 3):
        globalColorTable += rng.read(1)

    # Image descriptor
    leftPos = rng.read(1)[0] % 65536
    topPos = rng.read(1)[0] % 65536
    imageWidth = rng.read(1)[0] % 65536
    imageHeight = rng.read(1)[0] % 65536
    flags = rng.read(1)[0] & 0b11111000 | 0b00000100
    localColorTableFlag = rng.read(1)[0] % 2
    interlaceFlag = rng.read(1)[0] % 2
    sortFlag = rng.read(1)[0] % 2
    reserved = rng.read(1)[0] % 2 ** 3
    localColorTableSize = rng.read(1)[0] % 2 ** 2
    localColorTable = bytearray()
    if localColorTableFlag == 1:
        for i in range(localColorTableSize * 3):
            localColorTable += rng.read(1)
    else:
        localColorTable = None
    imageDescriptor = \
            struct.pack('<HHHHBBBBB',
                leftPos, topPos, imageWidth, imageHeight,
                flags, localColorTableFlag, interlaceFlag,
                sortFlag, reserved, localColorTableSize)

    # Image data
    lzwMinimumCodeSize = rng.read(1)[0] % 2 ** 3
    while True:
        chunkSize = rng.read(1)[0] % 2 ** 12
        if chunkSize == 0: break

        chunkType = rng.read(1)[0] % 2 ** 3
        if chunkType!= 0: continue

        blockLength = rng.read(1)[0] % 2 ** 8
        blockSize = blockLength + 1
        compressedData = bytearray()
        while len(compressedData) < blockSize:
            compressedData += rng.read(blockSize - len(compressedData))

        uncompressedData = bytearray()
        while len(uncompressedData) < len(compressedData):
            code = rng.read(1)[0] % 2 ** lzwMinimumCodeSize

            if code == lzwMinimumCodeSize and len(uncompressedData) > 0:
                uncompressedData[-1:] = [uncompressedData[-1]]
            elif code == lzwMinimumCodeSize and len(uncompressedData) == 0:
                pass
            elif code == lzwMinimumCodeSize + 1 and len(uncompressedData) >= 2:
                uncompressedData = uncompressedData[:-2]
            elif code <= len(uncompressedData):
                uncompressedData += uncompressedData[code-1:-1]
            else:
                raise Exception('Invalid LZW code')

        out.write(struct.pack('<B', blockLength))
        out.write(bytes(uncompressedData))

    # Trailer
    trailer = b'\x3B'

    # Write everything to output
    out.write(header)
    out.write(logicalScreenDescriptor)
    out.write(globalColorTable)
    out.write(imageDescriptor)


if