#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(out: BinaryIO, rng: bytes):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + rng[0] % 256
    height = 1 + rng[1] % 256
    flags = b'\x87\x00' if rng[2] & 1 else b'\x07\x00'
    bg_color = rng[3] % 256
    pixel_aspect_ratio = b'\x00'
    out.write(struct.pack('<HHBBB', width, height, flags, bg_color, pixel_aspect_ratio))

    # Global color table
    num_colors = min((width * height), 256)
    for i in range(num_colors):
        r = g = b = rng[(i*4)+0] % 256
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left = top = 0
    width = 1 + rng[4] % 256
    height = 1 + rng[5] % 256
    flags = b'\x87\x00' if rng[6] & 1 else b'\x07\x00'
    local_color_table_flag = b'\x01' if rng[6] & 2 else b'\x00'
    interlace_flag = b'\x00'
    sort_flag = b'\x00'
    reserved = b'\x00\x00'
    lct_size = b'\x00'
    out.write(struct.pack('<HHHHBBBBB', left, top, width, height, flags, local_color_table_flag, interlace_flag, sort_flag, reserved, lct_size))

    # Local color table
    for i in range(num_colors):
        r = g = b = rng[(i*4)+0] % 256
        out.write(struct.pack('BBB', r, g, b))

    # LZW minimum code size
    min_code_size = rng[7] % 256
    out.write(min_code_size.to_bytes(1, byteorder='little'))

    # Data sub-blocks
    data_size = width * height // 2
    data = []
    while len(data) < data_size:
        block_type = rng[len(data)] % 2
        if block_type == 0:
            # Raw data block
            length = rng[(len(data)*2)+0] % 128
            data += [rng[(len(data)*2)+1+i] for i in range(length)]

        elif block_type == 1:
            # Repeat previous block
            count = rng[(len(data)*2)+0] % 3
            data += data[-count:]

        else:
            assert False

    # Write compressed data
    write_lzw_compressed_data(out, data, min_code_size)


# Writes LZW compressed data to `out`.
# The input is given as a list of integers and the minimum code size.
def write_lzw_compressed_data(out: BinaryIO, data: list, min_code_size: int):
    # Initialize dictionary with all possible codes
    dict_size = 2 ** min_code_size
    dict = {i : bytes([i]) for i in range(dict_size)}
    prev = None

    # Compress data
    for c in data:
        key = prev + bytes([c]) if prev!= None else bytes([c])
        if key not in dict:
            # Add new entry to dictionary
            dict[key] = key

            # Check if we need to increase the dictionary size
            if len(dict) >= 2**12 - 1 and min_code_size < 12:
                min_code_size += 1
                dict_size = 2 ** min_code_size

                # Rebuild dictionary
                dict = {}
                for k, v in dict.items():
                    dict[k] = v[:-(len(v)-min_code_size)]

        prev = dict[key]
        out.write(prev)

    # Flush remaining bits
    flush_bits(out, 0)


# Flushes any remaining bits from the bit buffer.
def flush_bits(out: BinaryIO, value: int):
    while True:
        if value <= 0xFF:
            break

        out.write(value.to_bytes(1, byteorder='little'))
        value >>= 8


if