#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([0b00000000])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])

    lsd = struct.pack('<HHBBB', width, height, *flags, *bg, *aspect)
    out.write(lsd)

    # Global color table
    global_color_table_size = int.from_bytes(rng.read(1), byteorder='little') & 7
    if global_color_table_size!= 0:
        raise Exception('Invalid global color table size')

    global_color_table = [
        bytes([
            int.from_bytes(rng.read(1), byteorder='little'),
            int.from_bytes(rng.read(1), byteorder='little'),
            int.from_bytes(rng.read(1), byteorder='little'),
            0x00
        ]) for _ in range(256)
    ]

    gct = bytes()
    for i in range(len(global_color_table)):
        gct += global_color_table[i]

    out.write(gct)

    # Image descriptor
    left   = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16)
    top    = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16)
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([
        0b00000000 | \
        ((int.from_bytes(rng.read(1), byteorder='little') & 1) << 7) | \
        ((int.from_bytes(rng.read(1), byteorder='little') & 1) << 6) | \
        ((int.from_bytes(rng.read(1), byteorder='little') & 1) << 5) | \
        ((int.from_bytes(rng.read(1), byteorder='little') & 1) << 4) | \
        ((int.from_bytes(rng.read(1), byteorder='little') & 1) << 3) | \
        ((int.from_bytes(rng.read(1), byteorder='little') & 1) << 2) | \
        ((int.from_bytes(rng.read(1), byteorder='little') & 1) << 1) | \
        (int.from_bytes(rng.read(1), byteorder='little') & 1)
    ])

    id = struct.pack('<HHHHB', left, top, width, height, *flags)
    out.write(id)

    # Local color table
    local_color_table_size = int.from_bytes(rng.read(1), byteorder='little') & 7
    if local_color_table_size == 0:
        local_color_table = []
    elif local_color_table_size == 1:
        local_color_table = [
            bytes([
                int.from_bytes(rng.read(1), byteorder='little'),
                int.from_bytes(rng.read(1), byteorder='little'),
                int.from_bytes(rng.read(1), byteorder='little'),
                0x00
            ]) for _ in range(1 << 1)
        ]
    else:
        raise Exception('Invalid local color table size')

    lct = bytes()
    for i in range(len(local_color_table)):
        lct += local_color_table[i]

    out.write(lct)

    # Block size
    block_size = int.from_bytes(rng.read(1), byteorder='little') & 7

    # LZW minimum code size
    min_code_size = int.from_bytes(rng.read(1), byteorder='little') & 7

    # Data sub-blocks
    data = bytes()
    while True:
        # Data sub-block
        length = int.from_bytes(rng.read(1), byteorder='little')

        if length > 0:
            data += rng.read(length)

            continue

        # End of data sub-block
        break

    # Codes
    codes = {}
    for i in