#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0]))

    # Global color table
    for i in range(768):
        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

    # Image descriptor
    left   = int.from_bytes(rng.read(2), byteorder='little') % (width - 1) + 1
    top    = int.from_bytes(rng.read(2), byteorder='little') % (height - 1) + 1
    width  = int.from_bytes(rng.read(2), byteorder='little') % ((width - left) // 2) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % ((height - top) // 2) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lct    = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    interlaced = False if not bool(flags[0] & 0x40) else True
    sort    = False if not bool(flags[0] & 0x20) else True
    local_color_table_size = lct[0]
    background_index      = int.from_bytes(rng.read(1), byteorder='little')
    pixel_aspect_ratio    = int.from_bytes(rng.read(1), byteorder='little')
    out.write(struct.pack('<HHHHBB', left, top, width, height, flags[0], lct[0]))

    # Local color table
    for i in range((local_color_table_size ** 3)):
        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

    # Block size
    block_size = int.from_bytes(rng.read(1), byteorder='little')

    # LZW minimum code size
    mincodesize = int.from_bytes(rng.read(1), byteorder='little')

    # Data sub-blocks
    while True:
        index = int.from_bytes(rng.read(1), byteorder='little')

        if index == 0xFF:
            break

        elif index <= mincodesize:
            out.write(bytes([index]))

        else:
            length = int.from_bytes(rng.read(1), byteorder='little')

            if length > 0:
                out.write(bytes([(length >> 8) | 0xC0]))
                out.write(bytes([length & 0xFF]))

                for i in range(length):
                    out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

            else:
                length = (-length) * 256 + int.from_bytes(rng.read(1), byteorder='little') + 1

                for i in range(length):
                    out.write(bytes([index]))

