#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical Screen Descriptor
    width  = rng.read(1)[0] % 256 + 1
    height = rng.read(1)[0] % 256 + 1
    flags  = bytes([
        0b00000000 | ((width >> 7) & 0x01),
        (width <<  1) & 0xFF,
        (height >> 7) & 0x01,
        (height << 1) & 0xFF,
        0b00000000
    ])
    bg_color_index    = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 4 + 1
    out.write(flags + bytes([bg_color_index]) + bytes([pixel_aspect_ratio]))

    # Global Color Table
    global_color_table_size = rng.read(1)[0] % 256 + 1
    global_color_table      = bytearray()
    for _ in range(global_color_table_size):
        global_color_table += bytes([
            rng.read(1)[0],
            rng.read(1)[0],
            rng.read(1)[0],
            0b00000000
        ])
    out.write(bytes([global_color_table_size * 3]) + global_color_table)

    # Image Descriptor
    image_left     = rng.read(1)[0] % 256
    image_top      = rng.read(1)[0] % 256
    image_width    = rng.read(1)[0] % 256 + 1
    image_height   = rng.read(1)[0] % 256 + 1
    flags          = bytes([
        0b00000000 | ((image_width >> 7) & 0x01),
        (image_width <<  1) & 0xFF,
        (image_height >> 7) & 0x01,
        (image_height << 1) & 0xFF,
        0b00000000
    ])
    local_color_map_flag = rng.read(1)[0] % 2
    interlace_flag       = rng.read(1)[0] % 2
    sort_flag            = rng.read(1)[0] % 2
    reserved             = rng.read(1)[0] % 2
    local_color_map_size = rng.read(1)[0] % 256 + 1
    out.write(bytes([image_left, image_top]) + flags + bytes([local_color_map_flag]) + bytes([interlace_flag]) + bytes([sort_flag]) + bytes([reserved]) + bytes([local_color_map_size]))

    if local_color_map_flag == 1:
        color_table = bytearray()
        for _ in range(local_color_map_size):
            color_table += bytes([
                rng.read(1)[0],
                rng.read(1)[0],
                rng.read(1)[0],
                0b00000000
            ])
        out.write(color_table)

    lzw_min_code_size = rng.read(1)[0] % 256 + 1
    out.write(bytes([lzw_min_code_size]))

    # LZW Data
    code_size = 2 ** lzw_min_code_size
    clear_code = 1 << lzw_min_code_size
    end_of_information_code = clear_code + 1
    next_code = None
    previous_code = None
    buffer = []
    while True:
        if not next_code or len(buffer) >= code_size - 1:
            next_code = rng.read(1)[0] % 256

            if next_code == end_of_information_code: break

        if next_code < clear_code:
            buffer.append(next_code)
            previous_code = next_code
            continue
        
        if next_code > previous_code:
            buffer.append(previous_code)
        else:
            buffer.append(previous_code)
            index = len(buffer) - 1
            while index >= 0 and buffer[index]!= clear_code:
                del buffer[index]
                index -= 1
            del buffer[-1]
            buffer.