#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = 1024 + int.from_bytes(rng.read(2), byteorder='little', signed=False) % 512
    height =  768 + int.from_bytes(rng.read(2), byteorder='little', signed=False) % 512
    flags  =   0b00000000 | \
              ((int.from_bytes(rng.read(1), byteorder='little', signed=False) & 0x01) << 7) | \
              (((int.from_bytes(rng.read(1), byteorder='little', signed=False) >> 1) & 0x07) << 3) | \
              (((int.from_bytes(rng.read(1), byteorder='little', signed=False) >> 4) & 0x01) << 1) | \
              (((int.from_bytes(rng.read(1), byteorder='little', signed=False) >> 5) & 0x01) << 0)
    bg_color_index     = int.from_bytes(rng.read(1), byteorder='little', signed=False)
    pixel_aspect_ratio = int.from_bytes(rng.read(1), byteorder='little', signed=False)
    out.write(struct.pack('<HHBBB', width, height, flags, bg_color_index, pixel_aspect_ratio))

    # Global color table
    global_color_table_size = min((width * height) // 256, 256)
    for i in range(global_color_table_size):
        r = int.from_bytes(rng.read(1), byteorder='little', signed=False)
        g = int.from_bytes(rng.read(1), byteorder='little', signed=False)
        b = int.from_bytes(rng.read(1), byteorder='little', signed=False)
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left      =      0
    top       =      0
    width     = width - 1
    height    = height - 1
    flags     =   0b00000000 | \
                 ((int.from_bytes(rng.read(1), byteorder='little', signed=False) & 0x01) << 7) | \
                 (((int.from_bytes(rng.read(1), byteorder='little', signed=False) >> 1) & 0x07) << 3) | \
                 (((int.from_bytes(rng.read(1), byteorder='little', signed=False) >> 4) & 0x01) << 1) | \
                 (((int.from_bytes(rng.read(1), byteorder='little', signed=False) >> 5) & 0x01) << 0)
    delay_time        = int.from_bytes(rng.read(2), byteorder='little', signed=False)
    transparent_color = int.from_bytes(rng.read(1), byteorder='little', signed=False)
    out.write(struct.pack('<HHHHBB', left, top, width, height, flags, delay_time, transparent_color))

    # Local color table
    local_color_table_flag = int.from_bytes(rng.read(1), byteorder='little', signed=False)
    if local_color_table_flag == 1:
        local_color_table_size = min(((width * height) // 256) + 1, 256)
        for i in range(local_color_table_size):
            r = int.from_bytes(rng.read(1), byteorder='little', signed=False)
            g = int.from_bytes(rng.read(1), byteorder='little', signed=False)
            b = int.from_bytes(rng.read(1), byteorder='little', signed=False)
            out.write(struct.pack('BBB', r, g, b))

    # LZW minimum code size
    lzw_minimum_code_size = max(min(int.from_bytes(rng.read(1), byteorder='little', signed=False), 12), 2)
    out.write(struct.pack('B', lzw_minimum_code_size))

    # Data sub-blocks
    data_subblock_size = 2 ** lzw_minimum_code_size
    while True:
        # Get next block