#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    out.write(struct.pack('<H', 0x200))
    out.write(b'\x01')
    out.write(bytes([rng.read(1)[0]]))
    out.write(b'\xFF\xFF\xFF')

    # Write the global color table
    for _ in range(768):
        out.write(bytes([rng.read(1)[0]]))

    # Write the application extension block
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00\x00')

    # Write the image descriptor
    out.write(b'\x2c\x00\x00\x00\x00\x00\x00\x00\x00')

    # Write the local color table
    for _ in range(768):
        out.write(bytes([rng.read(1)[0]]))

    # Write the image data
    out.write(b'\x00\x00')

    # Write the LZW compressed image data
    lzw_min_code_sizes = [8, 4, 2, 0]
    for lzw_min_code_size in lzw_min_code_sizes:

        # Initialize the dictionary with all possible values
        dict_size = (1 << lzw_min_code_size) + 256
        dict = bytearray()
        for i in range(dict_size):
            dict.append(i)

        # The first code is always the clear code
        dict[0] = 0xE0
        dict[1] = 0xE1
        dict[2] = 0xE2
        dict[3] = 0xE3
        dict[4] = 0xE4
        dict[5] = 0xE5
        dict[6] = 0xE6
        dict[7] = 0xE7
        dict[8] = 0xE8
        dict[9] = 0xE9
        dict[10] = 0xEA
        dict[11] = 0xEB
        dict[12] = 0xEC
        dict[13] = 0xED
        dict[14] = 0xEE
        dict[15] = 0xEF
        dict[16] = 0xF0
        dict[17] = 0xF1
        dict[18] = 0xF2
        dict[19] = 0xF3
        dict[20] = 0xF4
        dict[21] = 0xF5
        dict[22] = 0xF6
        dict[23] = 0xF7
        dict[24] = 0xF8
        dict[25] = 0xF9
        dict[26] = 0xFA
        dict[27] = 0xFB
        dict[28] = 0xFC
        dict[29] = 0xFD
        dict[30] = 0xFE
        dict[31] = 0xFF

        # The next two codes are always the end of stream and end of information codes
        dict[dict_size - 2] = 0xD9
        dict[dict_size - 1] = 0xDA

        # The third code is the first actual code
        current_code = 0

        while True:

            # Read one bit from the RNG
            bit = rng.read(1)[0] & 1

            if bit == 0:
                # If it's zero then we read another bit
                bit = rng.read(1)[0] & 1

                if bit == 0:
                    # If that was also zero then we write the current code twice
                    out.write(bytes([current_code]))
                    out.write(bytes([current_code]))

                    # And add it to the dictionary
                    dict.append(current_code)

                    # We need to reset the current code because it's no longer valid
                    current_code = 0

                else:
                    # Otherwise we write the current code once
                    out.write(bytes([current_code]))

                    # And add it to the dictionary
                    dict.append(current_code)

                    # We need to reset the current code because it's no longer valid
                    current_code = 0

                    # Then we read another bit
                    bit = rng.read(1)[0] & 1

                    if bit == 0:
                        # If that was also zero then we write the current code twice
                        out.write(bytes([current_code]))
                        out.write(bytes([current_code]))

                        # And add it to the dictionary
                        dict.append(current_code)

                        # We need to reset the current code because it's no longer valid
                        current_code = 0

                    else:
                        # Otherwise we write the current code once
                        out.write(bytes([current_code]))

                        # And add it to the dictionary
                        dict.append(current_code)

                        # We need to reset the current code because it's no longer valid
                        current_code = 0

            else:
                # Otherwise we read another bit
                bit = rng.read(1)[0] & 1

                if bit == 0:
                    # If it's zero then we write the current code once
                    out.write(bytes([current_code]))

                    # And add it to the dictionary
                    dict.append(current_code)

                    # We need to reset the current code because it's no longer valid
                    current_code = 0

                    # Then we read another bit
                    bit = rng.read(1)[0] & 1

                    if bit == 0:
                        # If that was also zero then we write the current code twice
                        out.write(bytes([current_code]))
                        out.write(bytes([current_code]))

                        # And add it to the dictionary
                        dict.append(current_code)

                        # We need to reset the current code because it's no longer valid
                        current_code