#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 256 + int.from_bytes(rng.read(1), 'big') % 40
    height = 256 + int.from_bytes(rng.read(1), 'big') % 40
    flags = b'\x87\x00' if bool(int.from_bytes(rng.read(1), 'big')) else b'\x07\x00'
    bgColorIndex = int.from_bytes(rng.read(1), 'big') & 0xf
    pixelAspectRatio = int.from_bytes(rng.read(1), 'big') & 0xff
    out.write(struct.pack('<HHBBB', width, height, flags[0], flags[1], bgColorIndex, pixelAspectRatio))

    # Global color table
    globalColorTableSize = min((width * height) // 256, 256)
    for i in range(globalColorTableSize):
        r = int.from_bytes(rng.read(1), 'big') & 0xff
        g = int.from_bytes(rng.read(1), 'big') & 0xff
        b = int.from_bytes(rng.read(1), 'big') & 0xff
        out.write(struct.pack('BBB', r, g, b))
    for _ in range(globalColorTableSize, 256):
        out.write(b'\x00\x00\x00')

    # Image descriptor
    left = 0
    top = 0
    flags = b'\x80\x00'
    localColorTableSize = 0
    interlaceFlag = False
    sortFlag = False
    reserved = b'\x00\x00'
    lct = None
    if not bool(int.from_bytes(rng.read(1), 'big')):
        localColorTableSize = min((width * height) // 256, 256)
        lct = bytearray()
        for i in range(localColorTableSize):
            r = int.from_bytes(rng.read(1), 'big') & 0xff
            g = int.from_bytes(rng.read(1), 'big') & 0xff
            b = int.from_bytes(rng.read(1), 'big') & 0xff
            lct += struct.pack('BBB', r, g, b)
        for _ in range(localColorTableSize, 256):
            lct += b'\x00\x00\x00'

        lct = bytes(lct)
    else:
        lct = b''
    delayTime = int.from_bytes(rng.read(2), 'big')
    transparentColorIndex = int.from_bytes(rng.read(1), 'big') & 0xf
    endOfFile = b'\x00'
    out.write(struct.pack('<HHBBBBB', left, top, width, height, flags[0], flags[1], localColorTableSize))
    out.write(struct.pack('B', interlaceFlag << 4 | sortFlag << 3 | reserved[0]))
    out.write(struct.pack('B', delayTime >> 8))
    out.write(struct.pack('B', delayTime & 0xff))
    out.write(struct.pack('B', transparentColorIndex))
    out.write(endOfFile)

    # LZW minimum code size
    out.write(b'\x08')

    # Pixel data
    if lct == b'':
        lzwMinimumCodeSize = 8
    elif len(lct) <= 16:
        lzwMinimumCodeSize = 3
    elif len(lct) <= 256:
        lzwMinimumCodeSize = 4
    else:
        lzwMinimumCodeSize = 8
    out.write(struct.pack('B', lzwMinimumCodeSize))

    # LZW dictionary
    dict = {}
    for i in range(256):
        dict[i] = chr(i).encode('ascii')
    dictSize = 256

    # LZW stream
    while True:
        char = rng.read(1)[0]
        if char!= 0x00 and char!= 0xff:
            if char >= dictSize:
                out.write(dict[dictSize - 1])

            if char < dictSize:
                out.write(dict[char])

                if dictSize < 4096:
                    dict[dictSize] = dict[char][:-1] + bytes([dict[