#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = (rng.read(1)[0] % 50) + 1
    height = (rng.read(1)[0] % 50) + 1
    out.write(struct.pack('<HHBBB', width * 10, height * 10, 0xF0 | rng.read(1)[0], 0, 0))

    # Write the global color table
    for i in range(768):
        out.write(bytes([(i >> ((i // 3) * 8)) & 0xFF]))

    # Write the application extension block
    out.write(b'\x21\xFF\x0BNETSCAPE2.0')
    out.write(b'\x03\x01')
    out.write(b'\x00\x00')

    # Write the image descriptor
    out.write(b'\x2C')
    out.write(struct.pack('<HHBBBBB', 0, 0, 0, 0xF0 | rng.read(1)[0], 0, 0))

    # LZW minimum code size
    lzw_min_code_size = rng.read(1)[0] + 1
    out.write(bytes([lzw_min_code_size]))

    # LZW initial dictionary
    dict_size = 2 ** lzw_min_code_size
    for i in range(dict_size):
        out.write(bytes([i]))

    # LZW data
    while True:
        code = rng.read(1)[0]

        if code == 0x00 or code == 0xFF:
            break;

        if code < dict_size:
            out.write(bytes([code]))

            # Add new entry to the dictionary
            out.write(bytes([rng.read(1)[0]]))
            dict_size += 1
        elif code > dict_size:
            raise Exception('Invalid LZW code!')
        else:
            prev = out.getvalue()[-2:]
            if len(prev)!= 2:
                continue

            index = int.from_bytes(prev, 'little') << 8 | code
            if index >= dict_size:
                raise Exception('Invalid LZW code!')

            out.seek(-2, 1)
            if index <= 0xFFFF:
                out.write(index.to_bytes(2, byteorder='little'))
            else:
                out.write((0x10000 + index).to_bytes(2, byteorder='little'))

            dict_size += 1

            if dict_size == 4096:
                # Reset the dictionary on overflow
                dict_size = 2 ** lzw_min_code_size

                # Increase the LZW min code size by one
                lzw_min_code_size += 1

                out.write(bytes([code]))
            else:
                out.write(bytes([code - lzw_min_code_size]))

    # Write the trailer
    out.write(b';')