#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bool:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = struct.pack('>H', 1024)
    height = struct.pack('>H',  768)
    flags  = struct.pack('B',   0x80 | 0x40 | 0x20)
    bg     = struct.pack('B',   0xff)
    aspect = struct.pack('B',   0x00)
    out.write(width + height + flags + bg + aspect)

    # Global color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f

        if ((r == 0) and (g == 0) and (b == 0)):
            r += 1
            g += 1
            b += 1

        out.write(struct.pack('BBB', r, g, b))

    # Image data
    for y in range(0, 768):
        lzw_mincodesize = ord(rng.read(1))
        out.write(bytes([lzw_mincodesize]))

        lzw_clearcode = bytes([1 << lzw_mincodesize])
        lzw_endcode   = bytes([(1 << lzw_mincodesize) + 1])

        lzw_nextcode = len(lzw_clearcode) + 2
        lzw_prevcode = None

        lzw_buffer = []

        while True:

            if not lzw_prevcode or \
               (len(lzw_buffer) >= 255):

                if len(lzw_buffer)!= 0:
                    out.write(bytes([len(lzw_buffer)]))
                else:
                    out.write(bytes([0]))

                lzw_buffer = [ord(rng.read(1))]

            if lzw_prevcode is None:
                code = lzw_buffer[0]
            elif lzw_prevcode == lzw_endcode:
                break
            else:
                code = lzw_buffer[-1] ^ (lzw_prevcode >> 8)

            if code < lzw_nextcode:
                out.write(bytes([code]))

                if code!= lzw_clearcode:

                    if len(lzw_buffer) > 1:
                        out.write(bytes(lzw_buffer[:-1]))

                lzw_buffer = []
                lzw_prevcode = None

            else:
                out.write(bytes([lzw_nextcode - 256]))

                if len(lzw_buffer) > 1:
                    out.write(bytes(lzw_buffer[:-1]))

                lzw_buffer = [(code ^ (lzw_nextcode - 256)),
                              ord(rng.read(1))]

            lzw_nextcode += 1

            if lzw_nextcode == (1 << lzw_mincodesize) + 2:
                out.write(lzw_clearcode)
                lzw_nextcode = len(lzw_clearcode) + 2


    # Trailer
    out.write(b';')

    return True


if __name__ == '__main__':

    try:
        with open('/dev/urandom', 'rb') as f:
            generate_random_gif(f, sys.stdout.buffer)
    except Exception as e:
        print("Error:", str(e), file=sys.stderr)
        return True


