#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bool:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = struct.pack('>H', 1024)
    height = struct.pack('>H',  768)
    flags  = struct.pack('B',   0x80 | 0x40 | 0x20)
    bg     = struct.pack('B',   0xff)
    aspect = struct.pack('B',   0x00)
    out.write(width + height + flags + bg + aspect)

    # Global color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f

        if ((r == 0) and (g == 0) and (b == 0)):
            r += 1
            g += 1
            b += 1

        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left      = struct.pack('>H',      0)
    top       = struct.pack('>H',      0)
    width     = struct.pack('>H', 1024)
    height    = struct.pack('>H',  768)
    flags     = struct.pack('B',       0x00)
    transparent_color = struct.pack('B', 0x00)
    out.write(left + top + width + height + flags + transparent_color)

    # LZW minimum code size
    min_code_size = struct.pack('B', 0xf)
    out.write(min_code_size)

    # Frame data
    lzw_table_size = 2 ** 8
    lzw_table = [None] * lzw_table_size
    lzw_table[0] = ''
    lzw_table[1] = '\x01'
    lzw_table_index = 2
    lzw_prev_char = None
    lzw_curr_char = None
    while True:
        lzw_curr_char = ord(rng.read(1)) & 0xff

        if not lzw_curr_char:
            break

        if (lzw_curr_char < len(lzw_table)):
            lzw_next_char = lzw_table[lzw_curr_char]

            if (len(lzw_next_char) > 1):
                lzw_prev_char = lzw_next_char[:-1]

                if (not lzw_prev_char):
                    continue

                out.write(bytes([ord(c) for c in lzw_prev_char]))
                lzw_next_char = lzw_next_char[-1:]

            lzw_table_index = lzw_curr_char
            lzw_prev_char = lzw_next_char
        else:
            lzw_next_char = lzw_prev_char + lzw_prev_char[0]

            if (len(lzw_next_char) > 1):
                lzw_prev_char = lzw_next_char[:-1]

                if (not lzw_prev_char):
                    continue

                out.write(bytes([ord(c) for c in lzw_prev_char]))
                lzw_next_char = lzw_next_char[-1:]

            lzw_table[lzw_table_index] = lzw_next_char
            lzw_table_index += 1

            if (lzw_table_index >= lzw_table_size):
                return False


    # Block terminator
    out.write(b'\x00')

    return True



if