#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = rng.read(1)[0] % 50 + 10
    height = rng.read(1)[0] % 50 + 10
    out.write(struct.pack('<HHBBB', width * 10, height * 10, 0xF0 | (rng.read(1)[0] & 0x0F), 0, 0))

    # Write the global color table
    for i in range(768):
        out.write(bytes([(i >> ((i // 3) * 8)) & 0xFF]))

    # Write the application extension block
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00\x00')

    # Write the image descriptor
    out.write(b'\x2c\x00\x00\x00\x00\x00\x00\x00\x00')

    # Write the local color table
    for i in range(768):
        out.write(bytes([(i >> ((i // 3) * 8)) & 0xFF]))

    # Write the image data
    out.write(b'\x00\x00')

    # Write the LZW compressed image data
    lzw_min_code_sizes = [8, 4, 2, 0]
    lzw_min_code_size = lzw_min_code_sizes[rng.read(1)[0] % len(lzw_min_code_sizes)]
    out.write(bytes([lzw_min_code_size]))

    # Generate the LZW code stream
    codes = []
    while True:
        code = rng.read(1)[0]
        if code == 0:
            break
        elif code < lzw_min_code_size or code >= lzw_min_code_size + 1 << lzw_min_code_size:
            raise Exception('Invalid LZW code')
        else:
            codes.append(code - lzw_min_code_size)

    # Write the LZW code stream
    prev_code = None
    while codes:
        code = codes.pop()
        if prev_code is not None:
            out.write(bytes([prev_code]))
        prev_code = code

    # Write the trailer
    out.write(b';')