#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The size of the output is determined by the first 10 bytes read from `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Read the header and determine the length of the image data
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')

    # Write the header to the output
    out.write(header)

    # Generate the palette (RGB triples)
    pal = [struct.pack('BBB', *rng.read(3)) for _ in range(256)]

    # Write the palette to the output
    out.write(b''.join(pal))

    # Write the global color table flag
    out.write(b'\x87\x00')

    # Write the background color index
    bgidx = rng.read(1)[0] % len(pal)
    out.write(bgidx.to_bytes(1, 'little'))

    # Write the pixel aspect ratio
    out.write(b'\x00')

    # Write the local color table flag
    out.write(b'\x00')

    # Write the graphics control extension block
    out.write(b'\x21\xF9\x04\x00\x00\x00\x00\x00\x00\x00')

    # Write the application extension block
    out.write(b'\xFF\x0BNETSCAPE2.0\x03\x01\x00\x00\x00')

    # Write the comment extension block
    out.write(b'\xFE\x01\x00')

    # Write the plain text extension block
    out.write(b'\x01\x0CComment:\x00')

    # Write the end of file block
    out.write(b'\x3B')

    # Copy the image data in chunks of 256 bytes at a time
    while datalen > 0:
        chunklen = min(256, datalen)
        out.write(rng.read(chunklen))
        datalen -= chunklen
