#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 29 14:07:56 2019

@author: jlee
"""

import numpy as np
from astropy.io import fits
from scipy.interpolate import interp1d

def read_fits(filename, ext=0):
    """Reads a FITS file and returns the data in an array.
    
    Parameters
    ----------
    filename : str
        Name of the FITS file to be opened.
        
    Returns
    -------
    data : array
        Data contained in the FITS file.
    """
    
    hdu = fits.open(filename)
    data = hdu[ext].data
    return data

def write_fits(filename, data, overwrite=True, clobber=False):
    """Writes a FITS file from an array.
    
    Parameters
    ----------
    filename : str
        Name of the FITS file to be created.
        
    data : array
        Data to be written into the FITS file.
        
    overwrite : bool (default: True)
        Whether or not to overwrite existing files.
        
    clobber : bool (default: False)
        Whether or not to overwrite existing files.
    """
    
    hdu = fits.PrimaryHDU()
    hdu.data = data
    hdu.writeto(filename, overwrite=overwrite, clobber=clobber)

def get_wavelengths(filename, ext=0):
    """Returns the wavelengths for each pixel in a FITS file.
    
    Parameters
    ----------
    filename : str
        Name of the FITS file to be opened.
        
    ext : int (default: 0)
        Extension number of the FITS file to be opened.
        
    Returns
    -------
    wave : array
        Wavelengths corresponding to each pixel in the FITS file.
    """
    
    hdu = fits.open(filename)[ext]
    crval1 = hdu.header['CRVAL1']
    cdelt1 = hdu.header['CDELT1']
    nx = hdu.header['NAXIS1']
    wave = np.arange(nx)*cdelt1 + crval1
    return wave

def get_flux(filename, ext=0):
    """Returns the flux for each pixel in a FITS file.
    
    Parameters
    ----------
    filename : str
        Name of the FITS file to be opened.
        
    ext : int (default: 0)
        Extension number of the FITS file to be opened.
        
    Returns
    -------
    flux : array
        Flux values corresponding to each pixel in the FITS file.
    """
    
    hdu = fits.open(filename)[ext]
    flux = hdu.data
    return flux

def get_flux_err(filename, ext=0):
    """Returns the uncertainty in the flux for each pixel in a FITS file.
    
    Parameters
    ----------
    filename : str
        Name of the FITS file to be opened.
        
    ext : int (default: 0)
        Extension number of the FITS file to be opened.
        
    Returns
    -------
    flux_err : array
        Uncertainty in the flux values corresponding to each pixel in the FITS file.
    """
    
    hdu = fits.open(filename)[ext]
    flux_err = hdu.header['FLUXERR']
    return flux_err

def get_mask(filename, ext=0):
    """Returns the mask for each pixel in a FITS file.
    
    Parameters
    ----------
    filename : str
        Name of the FITS file to be opened.
        
    ext : int (default: 0)
        Extension number of the FITS file to be opened.
        
    Returns
    -------
    mask : array
        Mask values corresponding to each pixel in the FITS file.
    """
    
    hdu