#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header (6 bytes)
    out.write(b'GIF89a')

    # Logical screen descriptor (7 bytes)
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0xff | 1
    flags = rng.read(1)[0]
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bgcolor.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Global color table (3 bytes)
    ncolors = min(max(1 << ((flags >> 4) & 7), 2), 256)
    for i in range(ncolors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Image descriptor (9 bytes)
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0xff | 1
    flags = rng.read(1)[0]
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(left.to_bytes(2, 'little'))
    out.write(top.to_bytes(2, 'little'))
    out.write(width.to_bytes(1, 'little'))
    out.write(height.to_bytes(1, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bgcolor.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Local color table (3 bytes * ncolors)
    if ncolors > 1:
        for i in range(ncolors):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            out.write(r.to_bytes(1, 'little'))
            out.write(g.to_bytes(1, 'little'))
            out.write(b.to_bytes(1, 'little'))

    # LZW minimum code size (1 byte)
    lzwmincodesize = rng.read(1)[0]
    out.write(lzwmincodesize.to_bytes(1, 'little'))

    # Data sub-block (variable length)
    while True:
        blocktype = rng.read(1)[0]

        if blocktype == 0x2c:
            break

        elif blocktype == 0x21:
            extensiontype = rng.read(1)[0]

            if extensiontype == 0xf9:
                # Graphic Control Extension (4 bytes + 1 byte)
                flags = rng.read(1)[0]
                delay = rng.read(2)[0]
                transpcolor = rng.read(1)[0]
                term = rng.read(1)[0]
                out.write(flags.to_bytes(1, 'little'))
                out.write(delay.to_bytes(2, 'little'))
                out.write(transpcolor.to_bytes(1, 'little'))
                out.write(term.to_bytes(1, 'little'))

            else:
                # Application Extension (11 bytes + variable length)
                appname = rng.read(8)
                appdata = rng.read(rng.read(1)[0])
                out.write(appname)
                out.write(len(appdata).to_bytes(1, 'little'))
                out.write(appdata)

        else:
            raise Exception('unexpected block type {}'.format(hex(blocktype)))

    # Data sub-block (variable length)
    while True:
        blocksize = rng.read(1)[0]
        data = rng.read(blocksize)
        out.write(blocksize.to_bytes(1, 'little'))
        out.write(data)

        if blocksize == 0:
            break
