#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # Generate a random GIF file.
    #
    # Args:
    #     rng (BinaryIO): Random source.
    #     out (BinaryIO): Output stream.
    #
    # Returns:
    #     None

    # Magic number
    out.write(b'GIF89a')

    # Screen descriptor
    width = 1 + int.from_bytes(rng.read(2), 'big') % 640
    height = 1 + int.from_bytes(rng.read(2), 'big') % 480
    flags = rng.read(1)[0] & 0x7F
    background = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(struct.pack('<HBBB', width, height, flags, background))

    # Global color table
    if flags & 0x80 != 0:
        colors = 1 << ((flags >> 4) & 0x07)
        for _ in range(colors):
            out.write(rng.read(3))

    # Image data
    while True:
        # Image header
        left = int.from_bytes(rng.read(2), 'big') % width
        top = int.from_bytes(rng.read(2), 'big') % height
        width = 1 + int.from_bytes(rng.read(2), 'big') % (width - left)
        height = 1 + int.from_bytes(rng.read(2), 'big') % (height - top)
        flags = rng.read(1)[0] & 0x87
        local_palette = False
        interlaced = False
        if flags & 0x80 != 0:
            local_palette = True
        if flags & 0x40 != 0:
            interlaced = True
        sort = rng.read(1)[0] & 0x08
        size = rng.read(1)[0] & 0x07
        out.write(struct.pack('<HHBBBB', left, top, width, height, flags, sort, size))

        # Local color table
        if local_palette:
            colors = 1 << size
            for _ in range(colors):
                out.write(rng.read(3))

        # LZW image data
        clear_code = 1 << size
        end_of_information = clear_code + 1
        available = clear_code + 2
        old_code = -1
        code_size = size + 1
        stack = []
        bits = 0
        mask = 1 << code_size
        while True:
            if len(stack) == 4096:
                break
            try:
                while True:
                    bits |= rng.read(1)[0] << (code_size - 8)
                    code_size += 8
            except IndexError:
                pass
            if bits >= mask:
                break

        while True:
            if len(stack) == 4096:
                break
            code = bits & mask
            bits >>= code_size
            code_size -= 1
            if code == clear_code:
                code_size = size + 1
                mask = 1 << code_size
                continue
            elif code == end_of_information:
                break
            elif code > available or old_code < 0:
                stack.append(suffix[clear_code])
                old_code = clear_code
                continue
            prefix = suffix[old_code]
            suffix[available] = suffix[old_code] + [prefix[-1]]
            stack.extend(suffix[old_code])
            if available == 4095:
                break
            available += 1
            if available == (1 << code_size) and code_size < 12:
                code_size += 1
                mask = 1 << code_size
            old_code = code

        # Write the output
        for i in range(len(stack)):
            if i % width == 0 and i != 0:
                out.write(b