#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header (6 bytes)
    out.write(b'GIF89a')

    # Logical screen descriptor (7 bytes)
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0x3f | 1
    flags = rng.read(1)[0]
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('<HHBBB', width, height, flags, background, aspect_ratio))

    # Global color table (variable size)
    if flags & 0x80 != 0:
        colors = 2 ** ((flags >> 4) + 1)
        for i in range(colors):
            red = rng.read(1)[0]
            green = rng.read(1)[0]
            blue = rng.read(1)[0]
            out.write(bytes([red, green, blue]))

    # Image block (variable size)
    while True:
        # Graphic control extension (optional, 4-8 bytes)
        if rng.read(1)[0] == 0xf9:
            length = rng.read(1)[0]
            assert length >= 4 and length <= 8
            disposal = rng.read(1)[0] & 0x7
            delay = rng.read(2)[0]
            transparent = rng.read(1)[0]
            terminator = rng.read(1)[0]
            assert terminator == 0

        # Image descriptor (always 9 bytes)
        image_left = rng.read(2)[0]
        image_top = rng.read(2)[0]
        image_width = rng.read(2)[0]
        image_height = rng.read(2)[0]
        flags = rng.read(1)[0]
        local_palette = None
        if flags & 0x80 != 0:
            local_palette = []
            colors = 2 ** ((flags >> 4) + 1)
            for i in range(colors):
                red = rng.read(1)[0]
                green = rng.read(1)[0]
                blue = rng.read(1)[0]
                local_palette.append((red, green, blue))

        lzw_minimum_code_size = rng.read(1)[0]

        # LZW data (variable size)
        while True:
            code = rng.read(1)[0]
            if code == 0:
                break

        # Trailer (always 1 byte)
        trailer = rng.read(1)[0]
        assert trailer == 0

        # End of image block
        if image_left == 0 and image_top == 0 \
           and image_width == 0 and image_height == 0:
            break
