#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    #
    # See https://en.wikipedia.org/wiki/GIF for more information about the
    # structure of a GIF file.

    # Write the header
    out.write(b'GIF89a')

    # Write the screen descriptor
    width = int.from_bytes(rng.read(2), byteorder='little') % 100 + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % 100 + 1
    flags = rng.read(1)[0] & 0x7F
    background = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0]
    out.write(struct.pack('<HBBB', width, height, flags, background))

    # Write the global color table
    if flags & 0x80 != 0:
        num_colors = 2 ** ((flags & 0x07) + 1)
        for _ in range(num_colors):
            out.write(rng.read(3))

    # Write the application extension
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01')
    out.write(struct.pack('<I', 0))

    # Now we need to figure out how many bytes are left in the file, which we
    # can do by subtracting the size of the header from the size of the file.
    # We'll do this by seeking back to the beginning of the file and getting
    # its current position.
    out.seek(0)
    size = out.tell()

    # Write the image descriptor
    out.write(b'\x2c')
    out.write(struct.pack('<HHBB', 0, 0, width, height))
    flags = rng.read(1)[0] & 0x7F
    local_color_table = (flags >> 7) & 1
    interlaced = (flags >> 6) & 1
    sorted = (flags >> 5) & 1
    reserved = (flags >> 4) & 1
    if local_color_table != 0:
        num_colors = 2 ** ((flags & 0x07) + 1)
        for _ in range(num_colors):
            out.write(rng.read(3))

    # Write the LZW-compressed image data
    lzw_minimum_code_size = rng.read(1)[0]
    out.write(b'\x00')
    while True:
        code = int.from_bytes(rng.read(2), byteorder='little')
        if code == 0:
            break
        out.write(struct.pack('<H', code))

    # Write the trailer
    out.write(b';')

if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)