#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate and write the logical screen descriptor
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    bg     = bytes([int.from_bytes(rng.read(1), 'big')])
    pxsize = bytes([int.from_bytes(rng.read(1), 'big')])
    color  = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write((width).to_bytes(2,'big'))
    out.write((height).to_bytes(2,'big'))
    out.write(flags)
    out.write(bg)
    out.write(pxsize)
    out.write(color)

    # Generate and write the global color table
    for i in range(pow(2, int.from_bytes(color, 'big'))):
        r = int.from_bytes(rng.read(1),'big')
        g = int.from_bytes(rng.read(1),'big')
        b = int.from_bytes(rng.read(1),'big')
        out.write(r.to_bytes(1,'big'))
        out.write(g.to_bytes(1,'big'))
        out.write(b.to_bytes(1,'big'))

    # Write the image descriptor
    left   = bytes([0x00])
    top    = bytes([0x00])
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    out.write(left)
    out.write(top)
    out.write(width.to_bytes(2,'big'))
    out.write(height.to_bytes(2,'big'))
    out.write(flags)

    # Write the local color table
    if pow(2, int.from_bytes(color, 'big')) > 0:
        for i in range(pow(2, int.from_bytes(color, 'big'))):
            r = int.from_bytes(rng.read(1),'big')
            g = int.from_bytes(rng.read(1),'big')
            b = int.from_bytes(rng.read(1),'big')
            out.write(r.to_bytes(1,'big'))
            out.write(g.to_bytes(1,'big'))
            out.write(b.to_bytes(1,'big'))

    # Generate and write the LZW-compressed data stream
    def lzw_encode(s: str) -> list[list[str]]:
        """LZW encode an input string s"""
        # Initialize the dictionary
        dict_size = 256
        dictionary = {chr(i): chr(i) for i in range(dict_size)}

        # Encode the string
        w = ''
        result = []
        for c in s:
            wc = w + c
            if wc in dictionary:
                w = wc
            else:
                result.append(dictionary[w])
                # Add wc to the dictionary.
                dictionary[wc] = chr(dict_size)
                dict_size += 1
                w = c

        # Output the code for w.
        if w:
            result.append(dictionary[w])
        return result

    def lzw_decode(l: list[str], dict_size: int) -> str:
        """Decode an LZW-encoded list of strings"""
        # Build the dictionary
        dictionary = {chr(i): [chr(i)] for i in range(dict_size)}
        for word in l[:dict_size]:
            dictionary[word].append(word)

        # Decode
        words = list()
        for word in l[dict_size:]:
            if word in dictionary:
                words.extend(dictionary[word])
            elif word == '\x00':
                break
            else:
                raise ValueError('Bad compressed string!')

        return ''.join(words)

    with open('/dev/urandom', 'rb') as f:
        # Read the entire file
        data = f.read().decode()

        # Encode the data
        encoded = []
        prev = ''
        for char in data:
            # If we have no previous character or this is the start of a new
            # encoding sequence or this character is not in the previous
            # encoding sequence, then add a new encoding sequence
            if len(encoded)==0 or \
                    len(encoded[-1])>=255 or \
                    char not in encoded[-1]:

                # Start a new encoding sequence
                if len(encoded)!= 0:
                    # Write the last entry
                    encoded += [prev]

                # Add the current character as the first entry
                encoded += [[char]]
            else:
                # Append it to the last entry
                encoded[-1].append(char)

            # Update the previous character
            prev = ''.join(encoded[-1])

        # Write the encoded characters
        for e in encoded:
            out.write(bytes([len(e)]))
            for c in e:
                out.write(bytes([ord(c)]))