#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # Write the magic header
    out.write('GIF89a'.encode())

    # Generate some random bytes for the logical screen descriptor
    # and write them to the output stream
    width = int.from_bytes(
            rng.read(2), byteorder='big', signed=False)
    height = int.from_bytes(
            rng.read(2), byteorder='big', signed=False)
    packed = int.from_bytes(
            rng.read(1), byteorder='big', signed=False)
    bgcolor = int.from_bytes(
            rng.read(1), byteorder='big', signed=False)
    aspectratio = int.from_bytes(
            rng.read(1), byteorder='big', signed=False)
    lsd = struct.pack(
            '<HHBBB', width, height, packed, bgcolor, aspectratio)
    out.write(lsd)

    # Generate some random bytes for the global color table
    # and write them to the output stream
    gctsize = 0x07 & (packed >> 4)
    if gctsize != 0:
        gct = rng.read(3 * (2 ** (gctsize + 1)))
        out.write(gct)

    # Generate some random bytes for the background color index
    # and write them to the output stream
    bgi = rng.read(1)
    out.write(bgi)

    # Generate some random bytes for the pixel aspect ratio
    # and write them to the output stream
    par = rng.read(1)
    out.write(par)

    # Generate some random bytes for the application extension
    # and write them to the output stream
    appextlen = int.from_bytes(
            rng.read(3), byteorder='big', signed=False)
    out.write(appextlen.to_bytes(3, 'big'))

    # Generate some random bytes for the application extension block
    # and write them to the output stream
    appext = rng.read(appextlen)
    out.write(appext)

    # Generate some random bytes for the graphic control label
    # and write them to the output stream
    gclabel = rng.read(1)
    out.write(gclabel)

    # Generate some random bytes for the graphic control block
    # and write them to the output stream
    gcblock = rng.read(6)
    out.write(gcblock)

    # Generate some random bytes for the image descriptor
    # and write them to the output stream
    ildescr = struct.pack(
            '<HHHBB', width, height, 0x00, 0x00, 0x00)
    out.write(ildescr)

    # Generate some random bytes for the local color table
    # and write them to the output stream
    lct = rng.read(3 * (2 ** (gctsize + 1)))
    out.write(lct)

    # Generate some random bytes for the interlaced flag
    # and write them to the output stream
    interlaceflag = rng.read(1)
    out.write(interlaceflag)

    # Generate some random bytes for the image data length
    # and write them to the output stream
    datalen = int.from_bytes(
            rng.read(3), byteorder='big', signed=False)
    out.write(datalen.to_bytes(3, 'big'))

    # Generate some random bytes for the image data itself
    # and write them to the output stream
    data = rng.read(datalen)
    out.write(data)