#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    header = b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\xff\xff\xff\x21\xf9\x04\x01\x0a\x00\x01\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b'
    out.write(header)

    # Write the data section.  It consists of a bunch of blocks.  Each block
    # starts with a byte indicating its type (either image or extension).
    # Then there are two bytes indicating the length of the block.  Finally,
    # there is the block itself.
    while True:
        # Generate a random block type.
        block_type = rng.read(1)[0] % 256

        # Generate a random block size.
        block_size = rng.read(2)
        if len(block_size) != 2:
            break
        block_size = struct.unpack('<H', block_size)[0]

        # Generate a random block.
        block = rng.read(block_size)
        if len(block) != block_size:
            break

        # Write the block to the output file.
        out.write(bytes([block_type]))
        out.write(struct.pack('<H', block_size))
        out.write(block)


if __name__ == '__main__':
    try:
        seed = int(sys.argv[1])
    except IndexError:
        print("Usage: {} <seed> [<output directory>]".format(os.path.basename(__file__)), file=sys.stderr)
        exit(1)

    try:
        output_dir = sys.argv[2]
    except IndexError:
        output_dir = ''

    with open('/dev/urandom', 'rb') as rng:
        rng.seek(seed)
        for i in range(10):
            with open(os.path.join(output_dir, "{:02d}.gif".format(i)), 'wb') as f:
                generate_random_gif(rng, f)