#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate and write the logical screen descriptor
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    bg     = bytes([int.from_bytes(rng.read(1), 'big')])
    pxsize = bytes([int.from_bytes(rng.read(1), 'big')])
    color  = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(width.to_bytes(1, 'big'))
    out.write(height.to_bytes(1, 'big'))
    out.write(flags)
    out.write(bg)
    out.write(pxsize)
    out.write(color)

    # Generate and write the global color table
    for _ in range(2**(pxsize+1)):
        out.write(bytes([int.from_bytes(rng.read(1), 'big')]))

    # Generate and write the image data
    while True:
        # Generate the image descriptor
        left   = int.from_bytes(rng.read(1), 'big')
        top    = int.from_bytes(rng.read(1), 'big')
        width  = int.from_bytes(rng.read(1), 'big')
        height = int.from_bytes(rng.read(1), 'big')
        flags  = bytes([int.from_bytes(rng.read(1), 'big') & 0xF8])
        lzwmin = bytes([int.from_bytes(rng.read(1), 'big')])
        out.write(left.to_bytes(1, 'big'))
        out.write(top.to_bytes(1, 'big'))
        out.write(width.to_bytes(1, 'big'))
        out.write(height.to_bytes(1, 'big'))
        out.write(flags)
        out.write(lzwmin)

        # Generate the LZW-compressed image data
        encoded = []
        prev = ''
        while True:
            # Read a byte from the RNG
            char = chr(int.from_bytes(rng.read(1), 'big'))

            # Check if we need to start a new encoding sequence
            if len(encoded) == 0 or \
               char not in encoded[-1]:

                # Start a new encoding sequence
                if len(encoded)!= 0:
                    # Write the last entry
                    encoded += [prev]

                # Add the current character as the first entry
                encoded += [[char]]
            else:
                # Append it to the last entry
                encoded[-1].append(char)

            # Update the previous character
            prev = ''.join(encoded[-1])

        # Write the encoded characters
        for e in encoded:
            out.write(bytes([len(e)]))
            for c in e:
                out.write(bytes([ord(c)]))