#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate and write the logical screen descriptor
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    bg     = bytes([int.from_bytes(rng.read(1), 'big')])
    pxsize = bytes([int.from_bytes(rng.read(1), 'big')])
    color  = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(width.to_bytes(2, 'big'))
    out.write(height.to_bytes(2, 'big'))
    out.write(flags)
    out.write(bg)
    out.write(pxsize)
    out.write(color)

    # Generate and write the global color table
    colors = []
    for i in range(2**(pxsize[0] + 1)):
        colors.append((
            int.from_bytes(rng.read(1), 'big'),
            int.from_bytes(rng.read(1), 'big'),
            int.from_bytes(rng.read(1), 'big')
        ))
    for c in colors:
        out.write(bytes([c[0]]))
        out.write(bytes([c[1]]))
        out.write(bytes([c[2]]))

    # Generate and write the image data
    while True:
        # Generate the image descriptor
        left   = (int.from_bytes(rng.read(1), 'big') % width)
        top    = (int.from_bytes(rng.read(1), 'big') % height)
        width  = (int.from_bytes(rng.read(1), 'big') % (width - left)) + 1
        height = (int.from_bytes(rng.read(1), 'big') % (height - top)) + 1
        flags  = bytes([0x80 | (int.from_bytes(rng.read(1), 'big') & 0x7f)])
        if not width or not height: break
        out.write(left.to_bytes(2, 'big'))
        out.write(top.to_bytes(2, 'big'))
        out.write(width.to_bytes(2, 'big'))
        out.write(height.to_bytes(2, 'big'))
        out.write(flags)

        # Generate the local color table
        lct_size = (int.from_bytes(rng.read(1), 'big') % 8) + 1
        lct_sort = bool(int.from_bytes(rng.read(1), 'big') & 1)
        lct      = []
        for i in range(2**lct_size):
            lct.append((
                int.from_bytes(rng.read(1), 'big'),
                int.from_bytes(rng.read(1), 'big'),
                int.from_bytes(rng.read(1), 'big')
            ))
        for c in lct:
            out.write(bytes([c[0]]))
            out.write(bytes([c[1]]))
            out.write(bytes([c[2]]))

        # Generate the LZW-encoded data
        clear = int.from_bytes(rng.read(1), 'big') & 0xff
        code_size = int.from_bytes(rng.read(1), 'big') & 0xff
        codes = {}
        stack = []
        for _ in range(clear + 2):
            codes[len(stack)] = ''.join(chr(i) for i in stack)
            stack.append(int.from_bytes(rng.read(1), 'big') & 0xff)
        next_code = len(codes)
        encoded = [codes[len(stack)]]
        while True:
            index = int.from_bytes(rng.read(1), 'big') & 0xff
            if index == clear:
                code_size += 1
                next_code = len(codes)
                continue
            elif index == 0:
                break
            elif index < next_code:
                encoded.append(codes[index])
                stack = list(stack[:code_size]) + [index]
                codes[next_code] = ''.join(chr(i) for i in stack)
                next_code += 1
            else:
                encoded.append(codes[index])
                stack = list(stack[:code_size]) + [stack[-1]]
                codes[next_code] = ''.join(chr(i) for i in stack)
                next_code        # Write the encoded characters
        for e in encoded:
            out.write(bytes([len(e)]))
            for c in e:
                out.write(bytes([ord(c)]))