#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # get some of our data into the file before writing the rest.
    header = b'GIF89a'
    header += bytes([
        0x21, 0xF9, 0x04, 0x05, 0x00, 0x00, 0x00, 0x00, 0x2C, 0x00, 0x00, 0x00,
        0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x02, 0x02, 0x60
    ])
    out.write(header)

    # Now we're going to write the image data.  First up are the palette colors.
    for _ in range(768):
        r = next(rng) & 0xFF
        g = next(rng) & 0xFF
        b = next(rng) & 0xFF

        out.write(struct.pack('<BBB', r, g, b))

    # Next comes the logical screen descriptor.
    width = (next(rng) % 100) + 1
    height = (next(rng) % 100) + 1
    flags = 0b10000000 | ((next(rng) % 8) << 2) | (((next(rng) % 3) + 1) << 5)
    bg_color = next(rng) & 0xFF
    pixel_aspect_ratio = 0
    out.write(struct.pack('<HHBBBHB', width, height, flags, bg_color,
                          pixel_aspect_ratio, 0, 0))

    # Finally come the loop control block.
    out.write(bytes([
        0x21, 0xFF, 0x0B, 0x4E, 0x45, 0x54, 0x53, 0x43, 0x41, 0x50, 0x45, 0x32,
        0x2E, 0x30, 0x03, 0x01, 0x00, 0x00, 0x00
    ]))

    # And now we start writing the actual image data.  This will be done by
    # encoding each row as a series of subblocks with lzw compression.
    # Each subblock starts with a byte indicating how many bytes follow.
    # Then follows the compressed data itself.
    # When there's no more data left to encode, we add another zero-length
    # subblock which signals the end of the image data.
    while True:
        length = min((next(rng) % 10), 127)
        if length == 0:
            break

        out.write(bytes([length]))
        for i in range(length):
            out.write(bytes([(next(rng) % 256)]))

    out.write(bytes([0]))


def main() -> int:
    try:
        seed = int(sys.argv[1])
    except IndexError:
        print('Usage: {} SEED'.format(os.path.basename(__file__)),
              file=sys.stderr)
        return 1

    with open('/dev/urandom', 'rb') as f:
        f.seek(seed)
        generate_random_gif(f, sys.stdout.buffer)

    return 0


if __name__ == '__main__':
    exit(main())