#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header (6 bytes)
    out.write(b'GIF89a')

    # Logical screen descriptor (7 bytes)
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0xff | 1
    flags = rng.read(1)[0]
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bgcolor.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Global color table (3 bytes)
    ncolors = min(max(1 << ((flags >> 4) & 7), 2), 256)
    for i in range(ncolors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Image descriptor (9 bytes)
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0xff | 1
    flags = rng.read(1)[0]
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(left.to_bytes(2, 'little'))
    out.write(top.to_bytes(2, 'little'))
    out.write(width.to_bytes(1, 'little'))
    out.write(height.to_bytes(1, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bgcolor.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Local color table (3 bytes * ncolors)
    if ncolors > 1:
        for i in range(ncolors):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            out.write(r.to_bytes(1, 'little'))
            out.write(g.to_bytes(1, 'little'))
            out.write(b.to_bytes(1, 'little'))

    # Image data blocks
    while True:
        # Data sub-block

        # Block size (1 byte)
        blocksize = rng.read(1)[0]

        # Block contents
        if blocksize!= 0:
            # Plain text data sub-block

            # Text length (1 byte)
            datalen = rng.read(1)[0]

            # Text data (datalen bytes)
            out.write(rng.read(datalen))

        else:
            # Image data sub-block

            # LZW minimum code size (1 byte)
            lzwmincodesize = rng.read(1)[0]

            # Sub-blocks of compressed data
            while True:
                # Compressed data sub-block

                # Block size (1 byte)
                blocksize = rng.read(1)[0]

                # Block contents
                if blocksize!= 0:
                    # Plain text data sub-block

                    # Text length (1 byte)
                    datalen = rng.read(1)[0]

                    # Text data (datalen bytes)
                    out.write(rng.read(datalen))

                else:
                    # End of compressed data sub-block

                    blocktype = rng.read(1)[0]

                if blocktype == 0x00:
                    break

                elif blocktype == 0x01:
                    raise Exception('Invalid GIF image')

                else:
                    raise Exception('Unknown GIF image block type')

        elif blocktype == 0x21:
            # Extension sub-block

            # Extension label (1 byte)
            extlabel = rng.read(1)[0]

            if extlabel == 0xf9:
                # Graphic control extension

                # Block size (1 byte)
                blocksize = rng.read(1)[0]
                assert blocksize == 4

                # Flags (1 byte)
                flags = rng.read(1)[0]

                # Delay time (2 bytes)