#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(out: BinaryIO, rng: BinaryIO):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + (rng.read(1)[0] % 256)
    height = 1 + (rng.read(1)[0] % 256)
    packed = bytearray([
        0x80 | ((width >> 7) & 0x01),
        (width >> 1) & 0xff,
        (height >> 1) & 0xff,
        0x80 | (((height * 7) // 8) & 0x07)])
    out.write(packed)

    # Global color table
    palette = [i for i in range(256)]
    rng.shuffle(palette)
    for c in palette:
        out.write(struct.pack('BBB', c, c, c))

    # Image descriptor
    packed = bytearray([
        0x2c,
        0x00,
        0x00,
        0x00,
        width - 1,
        (height - 1) & 0xff,
        0x00])
    out.write(packed)

    # LZW minimum code size
    out.write(b'\x02\x00')

    # Pixel data
    lzw_mincodesize = 2
    lzw_clearcode = 1 << lzw_mincodesize
    lzw_endofinfo = lzw_clearcode + 1
    lzw_endofdata = lzw_endofinfo + 1
    lzw_nextcode = lzw_endofdata + 1
    lzw_maxcode = (1 << lzw_mincodesize) + 2
    lzw_bitmask = (1 << lzw_mincodesize) - 1
    lzw_buf = b''
    lzw_buflen = 0
    lzw_bufpos = 0
    lzw_prevcode = None
    while True:
        if lzw_prevcode is not None and lzw_prevcode == lzw_endofdata:
            break

        if lzw_prevcode is None or lzw_prevcode >= len(palette):
            lzw_prevcode = ord(rng.read(1))
            lzw_buf += bytes((lzw_prevcode,))
            lzw_buflen += 1
        elif lzw_prevcode < lzw_endofdata:
            lzw_buf += bytes((palette[lzw_prevcode],))
            lzw_buflen += 1

        if lzw_buflen > 254:
            out.write(bytes((len(lzw_buf) - 1,)))
            out.write(lzw_buf)

            lzw_buf = b''
            lzw_buflen = 0

    out.write(bytes((0,)))

