#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header (6 bytes)
    out.write(b'GIF89a')

    # Logical screen descriptor (7 bytes)
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0xff | 1
    flags = rng.read(1)[0]
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bgcolor.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Global color table (3 bytes)
    ncolors = min(max(1 << ((flags >> 4) & 7), 2), 256)
    for i in range(ncolors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Image descriptor (9 bytes)
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0xff | 1
    flags = rng.read(1)[0]
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(left.to_bytes(2, 'little'))
    out.write(top.to_bytes(2, 'little'))
    out.write(width.to_bytes(1, 'little'))
    out.write(height.to_bytes(1, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bgcolor.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Local color table (3 bytes * ncolors)
    if (flags & 0x80)!= 0:
        for i in range(ncolors):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            out.write(r.to_bytes(1, 'little'))
            out.write(g.to_bytes(1, 'little'))
            out.write(b.to_bytes(1, 'little'))

    # LZW minimum code size (1 byte)
    lzwsize = rng.read(1)[0] + 1
    out.write(lzwsize.to_bytes(1, 'little'))

    # Data blocks (variable length)
    while True:
        blocktype = rng.read(1)[0]

        if blocktype == 0x2c:
            # Image data block

            # Left position (2 bytes)
            left = rng.read(2)[0]
            out.write(left.to_bytes(2, 'little'))

            # Top position (2 bytes)
            top = rng.read(2)[0]
            out.write(top.to_bytes(2, 'little'))

            # Width (1 byte)
            width = 1 << rng.read(1)[0] & 0x3f | 1
            out.write(width.to_bytes(1, 'little'))

            # Height (1 byte)
            height = 1 << rng.read(1)[0] & 0xff | 1
            out.write(height.to_bytes(1, 'little'))

            # Flags (1 byte)
            flags = rng.read(1)[0]
            out.write(flags.to_bytes(1, 'little'))

            # LZW minimum code size (1 byte)
            lzwsize = rng.read(1)[0] + 1
            out.write(lzwsize.to_bytes(1, 'little'))

            # Data blocks (variable length)
            while True:
                blocktype = rng.read(1)[0]

                if blocktype == 0x00:
                    break

                elif blocktype == 0x01:
                    # Extension block

                    # Block size (1 byte)
                    extsize = rng.read(1)[0]
                    out.write(extsize.to_bytes(1, 'little'))

                    # Extension type (1 byte)
                    exttype = rng.read(1)[0]
                    out.write(exttype.to_bytes(1, 'little'))

                    # Data blocks (variable length)
                    while True:
                        blocktype = rng.read(1)[0]

                        if blocktype == 0xf9:
                            # Graphics control extension

                            # Size of graphics control extension block (1 byte)
                            gcsize = rng.read(1)[0]
                            out.write(gcsize.to_bytes(1, 'little'))

                            # Packed fields (1 byte)
                            packedfields = rng.read(1)[0]
                            out.write(packedfields.to_bytes(1, 'little'))

                            # Delay time (2 bytes)
                            delaytime = rng.read(2)[0]
                            out.write(delaytime.to_bytes(2, 'little'))

                            # Transparent color index (1 byte)
                            transindex = rng.read(1)[0]
                            out.write(transindex.to_bytes(1, 'little'))

                            # Terminator (1 byte)
                            terminator = rng.read(1)[0]
                            assert terminator == 0

                            break

                        else:
                            raise ValueError('Unknown extension type {}'.format(
                                hex(blocktype)))

                else:
                    raise ValueError('Unknown block type {}'.format(hex(blocktype)))

        elif blocktype == 0x3b:
            # Trailer
            break

        else:
            raise ValueError('Unknown block type {}'.format(hex(blocktype)))

