#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate and write the logical screen descriptor
    width = 1 + int.from_bytes(rng.read(2), byteorder='big', signed=False) % 500
    height = 1 + int.from_bytes(rng.read(2), byteorder='big', signed=False) % 500
    flags = b'\x87\x00' if (width > height) else b'\x00\x00'
    background_color_index = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xf
    pixel_aspect_ratio = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
    out.write((width).to_bytes(2, 'big'))
    out.write((height).to_bytes(2, 'big'))
    out.write(flags)
    out.write(background_color_index.to_bytes(1, 'big'))
    out.write(pixel_aspect_ratio.to_bytes(1, 'big'))

    # Generate and write global color table
    num_colors = min(int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff, 256)
    for i in range(num_colors):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        out.write(r.to_bytes(1, 'big'))
        out.write(g.to_bytes(1, 'big'))
        out.write(b.to_bytes(1, 'big'))

    # Generate and write local color table
    num_colors = min(int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff, 256)
    for i in range(num_colors):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        out.write(r.to_bytes(1, 'big'))
        out.write(g.to_bytes(1, 'big'))
        out.write(b.to_bytes(1, 'big'))

    # Generate and write image data
    num_blocks = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
    for _ in range(num_blocks):

        # Generate and write block size
        size = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        out.write(size.to_bytes(1, 'big'))

        # Generate and write LZW minimum code size
        lzw_min_code_size = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        out.write(lzw_min_code_size.to_bytes(1, 'big'))

        # Generate and write subblock data
        subblock_data = bytes()
        while True:
            bits = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff

            # If this is the last byte of the subblock...
            if not ((bits >> 7) & 0x1):
                break
            
            # Otherwise, append this to our subblock data
            subblock_data += bits.to_bytes(1, 'big')
        
        # Write the subblock data
        out.write(subblock_data)
    
    # Write trailer
    out.write(b';')


if