#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    """Generate a random GIF image."""
    
    header = b'GIF89a'

    # The first 6 bytes of the GIF are fixed and contain information about
    # the size of the image (width x height). We'll use the same width and 
    # height values that were used when generating the original sample images.
    out.write(header + struct.pack('<HHB', 256, 256, 0x7A))

    # Next is the global color table. This contains all colors used by this
    # particular frame. Since we're using only one color, it's just black.
    out.write(b'\x00\x00\x00')

    # Then comes the "logical screen descriptor". It tells us how many frames
    # there will be in total, and what their dimensions will be. For now, we
    # have only one frame so these two fields don't matter much; they're set
    # to zero.
    out.write(struct.pack('<HHBBBBB', 0, 0, 0xF7, 0, 0, 0))

    # Finally, we come to the actual image data. Each frame consists of an
    # LZW compressed stream of pixel data. To make things more interesting,
    # we'll compress our own randomly generated pixel data instead!

    # First up, let's write the "image descriptor" which tells us some details
    # about the current frame. In this case, we'll use the same width and
    # height from before, but we'll also specify that this is the last frame
    # in the animation.
    out.write(struct.pack('<HHHHB', 0, 0, 256, 256, 0x7A))

    # Next up is the local color table. Again, since we're using only one color,
    # it's just black.
    out.write(b'\x00\x00\x00')

    # After that, we get to the actual image data itself. The format of this
    # data is described here: https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    # Basically, each block starts with a single byte containing either a code
    # or a literal value. If it's a code, then the next three bytes indicate
    # how long the following sequence is. If it's a literal value, then the
    # next byte indicates its length. The end of the image data is marked by
    # a special code indicating that no further codes follow.

    # So let's start off by writing the first code. A code is indicated by
    # setting the most significant bit to 1. The remaining seven bits indicate
    # the number of subsequent pixels that should be drawn in the same color.
    # Since we want to draw 256 pixels in the same color, we'll write a code
    # whose value is 0x81.
    out.write(b'\xFE')

    # Now we need to write the length of the following sequence. Since we want
    # to draw 256 pixels in the same color, we'll write 0x01 followed by 0xFF.
    # Note that while the GIF standard specifies that lengths must be encoded
    # in big endian order, Python's struct module doesn't support encoding
    # numbers larger than 4 bytes in big endian form. Therefore, we need to
    # manually convert the number into a byte array before passing it on to
    # the output buffer.
    out.write(int.to_bytes(0x01FF, 3, 'big'))

    # Now we can finally write the actual pixel data. Since we've chosen to
    # encode every pixel in the same color, we simply repeat the same color
    # 256 times.
    for _ in range(0, 256):
        out.write(b'\x00\x00\x00')

    # Now that we've written the actual pixel data, we can finish up the image
    # data section by writing a final code. Unlike the previous code, however,
    # this time we won't be drawing any pixels in the same color anymore. Instead,
    # we'll just draw a single pixel at position (0, 0) in the top left corner
    # of the screen.
    out.write(b'\x01')

    # As mentioned above, the length of the following sequence needs to be
    # encoded in big endian order. However, unlike the previous example where
    # we had to manually convert the number into a byte array, Python's struct
    # module does not provide a    # interpreting them as a little-endian unsigned integer.
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    out.write(header + datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    for _ in range(0, datalen):
        out.write(rng.read(1))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)