#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header (6 bytes)
    out.write(b'GIF89a')

    # Logical screen descriptor (7 bytes)
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0xff | 1
    flags = rng.read(1)[0]
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bgcolor.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Global color table (3 bytes)
    ncolors = min(max(1 << ((flags >> 4) & 7), 2), 256)
    for i in range(ncolors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Image descriptor (9 bytes)
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0xff | 1
    flags = rng.read(1)[0]
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(left.to_bytes(2, 'little'))
    out.write(top.to_bytes(2, 'little'))
    out.write(width.to_bytes(1, 'little'))
    out.write(height.to_bytes(1, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bgcolor.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Local color table (3 bytes * ncolors)
    if ncolors > 1:
        for i in range(ncolors):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            out.write(r.to_bytes(1, 'little'))
            out.write(g.to_bytes(1, 'little'))
            out.write(b.to_bytes(1, 'little'))

    # LZW minimum code size (1 byte)
    lzwmincodesize = rng.read(1)[0]
    out.write(lzwmincodesize.to_bytes(1, 'little'))

    # Data blocks (variable length)
    while True:
        # Data sub-block header (1 byte)
        datablockheader = rng.read(1)[0]

        # End of image data sub-block
        if datablockheader == 0x3b:
            break

        # Plain text data sub-block
        elif datablockheader == 0x01:
            # Sub-block size (2 bytes)
            datasubblocksize = rng.read(2)[0]

            # Text data (variable length)
            while len(out.getbuffer()) < datasubblocksize + 2:
                char = rng.read(1)[0]
                out.write(char.to_bytes(1, 'little'))

        # Graphics data sub-block
        else:
            # Sub-block size (2 bytes)
            datasubblocksize = rng.read(2)[0]

            # LZW minimum code size (1 byte)
            lzwmincodesize = rng.read(1)[0]

            # Data sub-blocks (variable length)
            while len(out.getbuffer()) < datasubblocksize + 2:
                # Codes (variable length)
                while True:
                    # Code (1 or more bytes)
                    code = []

                    # Read first byte from stream
                    c = rng.read(1)[0]
                    code += [c]

                    # Check if we need to read more bytes
                    if not (c & 0x80):
                        break

                    # Read remaining bytes from stream
                    while True:
                        c = rng.read(1)[0]
                        code += [c]

                        if not (c & 0x80):
                            break

                # Write codes as bytes
                for c in code:
                    out.write(c.to_bytes(1, 'little'))

                # Graphic control extension

                # Block size (1 byte)
                blocksize = rng.read(1)[0]
                assert blocksize == 4

                # Flags (1 byte)
                flags = rng.read(1)[0]

                # Delay time (2 bytes)