#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.

    # Header fields are as follows (in order):
    #   Signature ("GIF89a")
    #   Version ("\x00\x21")
    #   Logical Screen Descriptor block (7 bytes)
    #       Width (2 bytes)
    #       Height (2 bytes)
    #       Flags (1 byte)
    #           Global Color Table Flag (bit 7)
    #               If set to zero, there's no global color table; if set to one,
    #               there is a global color table with 256 entries of size 3 bytes
    #           Color Resolution (bits 4-6)
    #           7 = 256 colors
    #           Sort Flag (bit 3)
    #               If set to zero, the image should not be sorted
    #           Size of Global Color Table (bits 0-2)
    #               0 = No global color table
    #               1 = 2^1 byte per entry
    #               2 = 2^2 bytes per entry
    #               3 = 2^3 bytes per entry
    #               4 = 2^4 bytes per entry
    #               5 = 2^5 bytes per entry
    #               6 = 2^6 bytes per entry
    #               7 = 2^7 bytes per entry
    #       Background Color Index (1 byte)
    #       Pixel Aspect Ratio (1 byte)
    #   Global Color Table (optional)
    #       This field contains up to 256 RGB triplets for the global color table.
    #       It only exists if the Global Color Table flag was set earlier.

    signature = b"GIF89a"
    version = b"\x00\x21"
    width = rng.read(2)
    height = rng.read(2)
    flags = rng.read(1)
    bg_color_index = rng.read(1)
    pixel_aspect_ratio = rng.read(1)

    # Now we need to calculate the length of the data section by writing some
    # dummy data into a temporary file and see where it ends up.

    tmpfile = open("tmp", "wb+")
    tmpfile.write(signature + version +
                  width + height + flags +
                  bg_color_index + pixel_aspect_ratio)
    tmpfile.flush()
    tmpfile.seek(0, os.SEEK_END)
    data_length = tmpfile.tell() - len(signature) - len(version) - \
            len(width) - len(height) - len(flags) - \
            len(bg_color_index) - len(pixel_aspect_ratio)
    tmpfile.close()

    # Now let's go back and actually write all those fields.

    out.write(signature)
    out.write(version)
    out.write(width)
    out.write(height)
    out.write(flags)
    out.write(bg_color_index)
    out.write(pixel_aspect_ratio)

    # And now comes the fun part!  Let's generate the actual data section.
    # First off, we need to decide whether or not we're going to use a global
    # color table.

    if int.from_bytes(flags, 'big') & 0b10000000 == 0:
        # There's no global color table, so let's just write a bunch of black
        # pixels instead.

        num_pixels = int.from_bytes(width, 'big') * int.from_bytes(height, 'big')
        for i in range(num_pixels):
            out.write(b'\x00\x00\x00')
    else:
        # There's a global color table, so let's generate random colors from our
        # PRNG and write them to the output stream.

        num_colors = 2 ** ((int.from_bytes(flags, 'big') >> 4) & 0b111)
        for i in range(num_colors):
            r = rng.read(1)[0] % 256
            g = rng.read(1)[0] % 256
            b = rng.read(1)[0] % 256

            out.write(struct.pack('BBB', r, g, b))

        # Now that we've written the global color table, we can write the actual
        # pixels themselves.

        num_pixels = int.from_bytes(width, 'big') * int.from_bytes(height, 'big')
        for i in range(num_pixels):
            out.write(rng.read(1))

    # Finally, we need to pad the end of the file out to