#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate and write the logical screen descriptor
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    bg     = bytes([int.from_bytes(rng.read(1), 'big')])
    pxsize = bytes([int.from_bytes(rng.read(1), 'big')])
    color  = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write((width).to_bytes(2,'big'))
    out.write((height).to_bytes(2,'big'))
    out.write(flags)
    out.write(bg)
    out.write(pxsize)
    out.write(color)

    # Generate and write the global color table
    for i in range(pow(2, int.from_bytes(color, 'big'))):
        r = int.from_bytes(rng.read(1),'big')
        g = int.from_bytes(rng.read(1),'big')
        b = int.from_bytes(rng.read(1),'big')
        out.write(r.to_bytes(1,'big'))
        out.write(g.to_bytes(1,'big'))
        out.write(b.to_bytes(1,'big'))

    # Write the image descriptor
    left   = bytes([0x00])
    top    = bytes([0x00])
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    out.write(left)
    out.write(top)
    out.write(width.to_bytes(2,'big'))
    out.write(height.to_bytes(2,'big'))
    out.write(flags)

    # Write the local color table
    if pow(2, int.from_bytes(color, 'big')) > 0:
        for i in range(pow(2, int.from_bytes(color, 'big'))):
            r = int.from_bytes(rng.read(1),'big')
            g = int.from_bytes(rng.read(1),'big')
            b = int.from_bytes(rng.read(1),'big')
            out.write(r.to_bytes(1,'big'))
            out.write(g.to_bytes(1,'big'))
            out.write(b.to_bytes(1,'big'))

    # Write the LZW minimum code size
    lzwmincodesize = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(lzwmincodesize)

    # Generate and write the compressed image data
    while True:

        # Read a block of data from /dev/urandom
        blocklen = int.from_bytes(rng.read(1), 'big')
        if not blocklen: break
        blockdata = rng.read(blocklen)

        # If this is an uncompressed block...
        if blockdata[0] == 0x01:

            # Write it to the output file
            out.write(blockdata)

        else:

            # Otherwise, decompress it
            decompresseddata = bytearray()
            while len(blockdata) > 0:

                # Get the next control byte
                ctrlbyte = blockdata[0]
                del blockdata[0]

                # If we have a clear code...
                if ctrlbyte == 0xE0:

                    # Reset the dictionary
                    dict = {
                        x : [
                            x,
                            0
                        ] for x in range(256)
                    }

                    # Add the clear code to the output stream
                    decompresseddata += bytes([ctrlbyte])

                elif ctrlbyte == 0xEE:

                    # End of image marker; stop processing
                    break

                # If we have a literal run...
                elif ctrlbyte <= 0x7F:

                    # Copy that many bytes literally
                    decompresseddata += blockdata[:ctrlbyte+1]
                    del blockdata[:ctrlbyte+1]

                else:

                    # Otherwise, get the length and offset of the match
                    length = ((ctrlbyte & 0x07) + 3)
                    offset = (((ctrlbyte >> 3) & 0xFF) + 1)

                    # Look up the previous sequence in our dictionary
                    prevseq = decompresseddata[-offset:]

                    # Append the matched sequence to the output stream
                    decompresseddata += prevseq * length

            # Write the decompressed data to the output file
            out.write(decompresseddata)


if