#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate and write the logical screen descriptor
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    bg     = bytes([int.from_bytes(rng.read(1), 'big')])
    pxsize = bytes([int.from_bytes(rng.read(1), 'big')])
    color  = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write((width).to_bytes(2,'big'))
    out.write((height).to_bytes(2,'big'))
    out.write(flags)
    out.write(bg)
    out.write(pxsize)
    out.write(color)

    # Generate and write the global color table
    colors = int.from_bytes(rng.read(1), 'big') + 1
    palette = []
    for i in range(colors):
        r = int.from_bytes(rng.read(1), 'big')
        g = int.from_bytes(rng.read(1), 'big')
        b = int.from_bytes(rng.read(1), 'big')
        palette += [(r << 16) | (g << 8) | b]
    out.write((colors * 3).to_bytes(1,'big'))
    for p in palette:
        out.write(((p >> 16) & 0xff).to_bytes(1,'big'))
        out.write(((p >>  8) & 0xff).to_bytes(1,'big'))
        out.write(((p      ) & 0xff).to_bytes(1,'big'))

    # Generate and write the image data
    frames = int.from_bytes(rng.read(1), 'big') + 1
    for f in range(frames):
        lzwmincodesize = int.from_bytes(rng.read(1), 'big') + 1
        out.write(lzwmincodesize.to_bytes(1,'big'))

        # Encode the frame's textual content with LZW compression
        encoded = [['']]
        prev = ''
        while True:
            # Read one byte from /dev/urandom
            char = chr(int.from_bytes(rng.read(1),'big'))

            # Check if we should stop
            if char == '\n': break

            if len(encoded) == 0 or \
               char not in encoded[-1]:

                # Start a new encoding sequence
                if len(encoded)!= 0:
                    # Write the last entry
                    encoded += [prev]

                # Add the current character as the first entry
                encoded += [[char]]
            else:
                # Append it to the last entry
                encoded[-1].append(char)

            # Update the previous character
            prev = ''.join(encoded[-1])

        # Write the encoded characters
        for e in encoded:
            out.write(bytes([len(e)]))
            for c in e:
                out.write(bytes([ord(c)]))