#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate some random bytes for the length of the image data
    # and write them to the output stream
    datalen = int.from_bytes(
            rng.read(3), byteorder='big', signed=False)
    out.write(datalen.to_bytes(3, byteorder='big'))

    # Write the screen descriptor
    width  = (int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256) + 1
    height = (int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256) + 1
    packed = ((width - 1) << 4) | (height - 1)
    out.write(packed.to_bytes(1, byteorder='big'))
    out.write(b'\x00\x00')
    out.write((int.from_bytes(rng.read(1), byteorder='big', signed=False) & 7).to_bytes(1, byteorder='big'))
    out.write(b'\x00')

    # Write the global color table
    colors = [
        b'\xff\x00\x00',
        b'\x00\xff\x00',
        b'\x00\x00\xff',
        b'\xff\xff\xff']
    out.write(colors[int.from_bytes(rng.read(1), byteorder='big', signed=False) % len(colors)])
    out.write(colors[int.from_bytes(rng.read(1), byteorder='big', signed=False) % len(colors)])

    # Write the local color table
    num_colors = int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
    if num_colors > 0:
        out.write(num_colors.to_bytes(1, byteorder='big'))

        for _ in range(num_colors):
            out.write(colors[int.from_bytes(rng.read(1), byteorder='big', signed=False) % len(colors)])

    # Write the graphics control extension
    out.write(b'\x21\xf9\x04')
    out.write(struct.pack('>H', int.from_bytes(rng.read(2), byteorder='big', signed=False)))
    out.write(b'\x00')

    # Write the application extension
    out.write(b'\x21\xff\x0bNETSCAPE2.0')
    out.write(b'\x03\x01')
    out.write(b'\x00\x00')
    out.write(b'\x00')

    # Write the plain text extension
    out.write(b'\x21\x01\x0f')
    out.write(b'Comment:\x00')
    out.write(str(int.from_bytes(rng.read(1), byteorder='big', signed=False)).encode())
    out.write(b'\x00')

    # Write the image descriptor
    out.write(b'\x2c')
    out.write(((width-1) >> 8).to_bytes(1, byteorder='big'))
    out.write(((width-1) & 0xff).to_bytes(1, byteorder='big'))
    out.write(((height-1) >> 8).to_bytes(1, byteorder='big'))
    out.write(((height-1) & 0xff).to_bytes(1, byteorder='big'))
    out.write(b'\x00')
    out.write(b'\x00')
    out.write(b'\x00')
    out.write(b'\x00')

    # Write the LZW minimum code size
    out.write(b'\x80')

    # Write the actual image data
    while True:
        # Get a random byte from /dev/urandom
        c = rng.read(1)[0]

        # If it's not part of the LZW compression scheme, just write it directly
        if c < 0x80 or c >= 0xe0:
            out.write(c.to_bytes(1, byteorder='big'))

            continue

        # Otherwise, we need to look up the next byte
        # in the dictionary and write that instead
        out.write(lzwdict[(c ^ lzwdict[len(lzwdict)-1])].to_bytes(1, byteorder='big'))

        # Add this new character to our dictionary
        lzwdict.append(c)

        # Check whether we've reached the end of the dictionary
        if len(lzwdict) == 4096:
            break