#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header (6 bytes)
    out.write(b'GIF89a')

    # Logical screen descriptor (7 bytes)
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0xff | 1
    flags = rng.read(1)[0]
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bgcolor.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Global color table (3 bytes)
    ncolors = min(max(1 << ((flags >> 4) & 7), 2), 256)
    for i in range(ncolors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Image descriptor (9 bytes)
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0xff | 1
    flags = rng.read(1)[0]
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(left.to_bytes(2, 'little'))
    out.write(top.to_bytes(2, 'little'))
    out.write(width.to_bytes(1, 'little'))
    out.write(height.to_bytes(1, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bgcolor.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Local color table (3 bytes)
    ncolors = min(max(1 << ((flags >> 4) & 7), 2), 256)
    for i in range(ncolors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Image data block (LZW minimum code size byte + LZW compressed data)
    lzwmincodesize = rng.read(1)[0]
    out.write(lzwmincodesize.to_bytes(1, 'little'))
    lzwdata = [i % 256 for i in range(256)] * 256
    while len(lzwdata) > 0xffff:
        lzwdata += [i % 256 for i in range(len(lzwdata))] * 256
    out.write(struct.pack('<H', len(lzwdata)))
    out.write(bytes(lzwdata))

