#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate and write the logical screen descriptor
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    bg     = bytes([int.from_bytes(rng.read(1), 'big')])
    pxsize = bytes([int.from_bytes(rng.read(1), 'big')])
    color  = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write((width).to_bytes(2,'big'))
    out.write((height).to_bytes(2,'big'))
    out.write(flags)
    out.write(bg)
    out.write(pxsize)
    out.write(color)

    # Generate and write the global color table
    colors = int.from_bytes(rng.read(1), 'big') + 1
    for i in range(colors):
        r = int.from_bytes(rng.read(1),'big') & 0xff
        g = int.from_bytes(rng.read(1),'big') & 0xff
        b = int.from_bytes(rng.read(1),'big') & 0xff
        out.write(bytes([r]))
        out.write(bytes([g]))
        out.write(bytes([b]))

    # Generate and write the image descriptor
    left   = bytes([0])
    top    = bytes([0])
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    out.write(left)
    out.write(top)
    out.write((width).to_bytes(2,'big'))
    out.write((height).to_bytes(2,'big'))
    out.write(flags)

    # Generate and write the local color table
    colors = int.from_bytes(rng.read(1), 'big') + 1
    for i in range(colors):
        r = int.from_bytes(rng.read(1),'big') & 0xff
        g = int.from_bytes(rng.read(1),'big') & 0xff
        b = int.from_bytes(rng.read(1),'big') & 0xff
        out.write(bytes([r]))
        out.write(bytes([g]))
        out.write(bytes([b]))

    # Generate and write the graphics control extension
    label       = bytes([0xf9])
    size        = bytes([4])
    flags       = bytes([0x87])
    delay       = bytes([0])
    transparent = bytes([0])
    out.write(label)
    out.write(size)
    out.write(flags)
    out.write(delay)
    out.write(transparent)

    # Generate and write the application extension
    label      = bytes([0xff])
    size       = bytes([11])
    appname    = bytes('NETSCAPE2.0', 'ascii')
    applength  = bytes([3])
    authcode   = bytes([1, 0, 0])
    out.write(label)
    out.write(size)
    out.write(appname)
    out.write(applength)
    out.write(authcode)

    # Generate and write the comment extension
    label = bytes([0xfe])
    size  = bytes([11])
    text  = bytes('Comment!', 'ascii')
    out.write(label)
    out.write(size)
    out.write(text)

    # Generate and write the plain text extension
    label = bytes([0x01])
    size  = bytes([11])
    text  = bytes('Plain Text!', 'ascii')
    out.write(label)
    out.write(size)
    out.write(text)

    # Generate and write the graphic data
    num_blocks = ((width * height) // 256) + 1
    for i in range(num_blocks):

        # Generate the block data
        blockdata = bytearray()
        for j in range(256):
            blockdata += rng.read(1)

        # Encode the block with LZW
        encoded = []
        prev = ''
        for c in blockdata:

            # Get the next character to encode
            char = chr(int.from_bytes(c, 'big'))

            # Check if we need to start a new encoding sequence
            if len(encoded) == 0 or \
               char not in encoded[-1]:

                # Start a new encoding sequence
                if len(encoded)!= 0:
                    # Write the last entry
                    encoded += [prev]

                # Add the current character as the first entry
                encoded += [[char]]
            else:
                # Append it to the last entry
                encoded[-1].append(char)

            # Update the previous character
            prev = ''.join(encoded[-1])

        # Write the encoded characters
        for e in encoded:
            out.write(bytes([len(e)]))
            for c in e:
                out.write(bytes([ord(c)]))