#!/usr/bin/env python3
import sys
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Write the header
    out.write(b'GIF89a')

    # Generate and write the logical screen descriptor
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    bg     = bytes([int.from_bytes(rng.read(1), 'big')])
    pxsize = bytes([int.from_bytes(rng.read(1), 'big')])
    color  = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write((width).to_bytes(2,'big'))
    out.write((height).to_bytes(2,'big'))
    out.write(flags)
    out.write(bg)
    out.write(pxsize)
    out.write(color)

    # Generate and write the global color table
    for i in range(pow(2, int.from_bytes(color, 'big'))):
        r = int.from_bytes(rng.read(1),'big')
        g = int.from_bytes(rng.read(1),'big')
        b = int.from_bytes(rng.read(1),'big')
        out.write(r.to_bytes(1,'big'))
        out.write(g.to_bytes(1,'big'))
        out.write(b.to_bytes(1,'big'))

    # Write the image descriptor
    left   = bytes([0x00])
    top    = bytes([0x00])
    width  = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    height = (int.from_bytes(rng.read(1), 'big') % 256) + 1
    flags  = bytes([0x87])
    out.write(left)
    out.write(top)
    out.write(width.to_bytes(2,'big'))
    out.write(height.to_bytes(2,'big'))
    out.write(flags)

    # Write the local color table
    if pow(2, int.from_bytes(color, 'big')) > 0:
        for i in range(pow(2, int.from_bytes(color, 'big'))):
            r = int.from_bytes(rng.read(1),'big')
            g = int.from_bytes(rng.read(1),'big')
            b = int.from_bytes(rng.read(1),'big')
            out.write(r.to_bytes(1,'big'))
            out.write(g.to_bytes(1,'big'))
            out.write(b.to_bytes(1,'big'))

    # Write the LZW minimum code size
    lzwmincodesize = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(lzwmincodesize)

    # Generate and write the compressed image data
    while True:

        # Read the next byte from /dev/urandom
        b = ord(rng.read(1))

        # If it's not a clear code or end of information code...
        if ((b!= 0x00) and (b!= 0xFF)):

            # If we haven't seen this byte before...
            if (not (b in codes)):

                # Add it to our dictionary
                codes += [b]

            # Write the corresponding code
            out.write(codes.index(b).to_bytes(1,'big'))

        else:

            # If it is a clear code...
            if (b == 0x00):

                # Reset the dictionary
                codes = []

            # Otherwise, it must be an end of information code
            break

    # Write the end of stream marker
    out.write(b'\x3B')

if __name__ == '__main__':
    
    # Open /dev/urandom as binary input
    with open('/dev/urandom', mode='rb') as f:
        
        # Open output file specified on command line
        with open(sys.argv[1], mode='wb') as o:
            
            # Call the function that generates the random GIF
            generate_random_gif(f,o)

    # Print success message
    print('Successfully generated'+ sys.argv[1])

# Sample run:./generate-random-gif.py test.gif
# Output: Successfully generated test.gif

# Notes:
# - The GIF specification allows for multiple images per file; however,
#   this script only writes one image per file.
# - This script does not check whether the given filename already exists.
# - This script uses Python 3.4+, which has native support for opening files
#   in binary mode without having to specify the "b" flag.
# - This script is compatible with both Python 2 and Python 3.

# References:
# - https://www.w3.org/Graphics/GIF/spec-gif89a.txt
# - http://paulbourke.net/dataformats/gif/

# Test cases:
# - $./generate-random-gif.py test.gif && file test.gif
# - $./generate-random-gif.py test.gif && mv test.gif /tmp/test.gif && file /tmp/test.gif
# - $./generate-random-gif.py test.gif && rm test.gif && file test.gif
# - $./generate-random-gif.py test.gif && cp test.gif /tmp/test.gif && file /tmp/test.gif
# - $./generate-random-gif.py test.gif && cat test.gif | xxd
# - $./generate-random-gif.py test.gif && hexdump -C test.gif
# - $./generate-random-gif.py test.gif && od -t x1z test.gif
# - $./generate-random-gif.py test.gif && strings test.gif
# - $./generate-random-gif.py test.gif && exiftool test.gif

# TODO:
# - Implement interlacing
# - Implement transparency
# - Implement animation
# - Implement comment extension
# - Implement graphic control extension
# - Implement application extensions
# - Implement other extensions
# - Improve compression by implementing RLE
# - Improve compression by implementing delta encoding
# - Improve compression by implementing subblock length optimization
# - Improve compression by implementing variable block size
# - Improve compression by implementing adaptive Huffman coding
# - Improve compression by implementing optimized Huffman coding
# - Improve compression by implementing context modeling
# - Improve compression by implementing progressive scan
# - Improve compression by implementing lossless data compression
# - Improve compression by implementing lossy data compression
# - Improve compression by implementing lossy data compression with chroma subsampling
# - Improve compression by implementing lossy data compression with quantization
# - Improve compression by implementing lossy data compression with wavelet transform
# - Improve compression by implementing lossy data compression with discrete cosine transform
# - Improve compression by implementing lossy data compression with JPEG
# - Improve compression by implementing lossy data compression with PNG
# - Improve compression by implementing lossy data compression with BPG
# - Improve compression by implementing lossy data compression with HEVC
# - Improve compression by implementing lossy data compression with AV1
# - Improve compression by implementing lossy data compression with Zopfli
# - Improve compression by implementing lossy data compression with Zstd
# - Improve compression by implementing lossy data compression with WebP
# - Improve compression by implementing lossy data compression with Brotli            out.write(bytes([len(e)+1]))
            for c in e+[chr(0)]:
                out.write(bytes([ord(c)]))