#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header (6 bytes)
    out.write(b'GIF89a')

    # Logical screen descriptor (7 bytes)
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0xff | 1
    flags = rng.read(1)[0]
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bgcolor.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Global color table (3 bytes)
    ncolors = min(max(1 << ((flags >> 4) & 7), 2), 256)
    for i in range(ncolors):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]
        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Image descriptor (9 bytes)
    left = rng.read(2)[0]
    top = rng.read(2)[0]
    width = 1 << rng.read(1)[0] & 0x3f | 1
    height = 1 << rng.read(1)[0] & 0xff | 1
    flags = rng.read(1)[0]
    bgcolor = rng.read(1)[0]
    aspect = rng.read(1)[0]
    out.write(left.to_bytes(2, 'little'))
    out.write(top.to_bytes(2, 'little'))
    out.write(width.to_bytes(1, 'little'))
    out.write(height.to_bytes(1, 'little'))
    out.write(flags.to_bytes(1, 'little'))
    out.write(bgcolor.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Local color table (3 bytes * ncolors)
    if ncolors > 1:
        for i in range(ncolors):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            out.write(r.to_bytes(1, 'little'))
            out.write(g.to_bytes(1, 'little'))
            out.write(b.to_bytes(1, 'little'))

    # LZW minimum code size (1 byte)
    lzwmincodesize = rng.read(1)[0]
    out.write(lzwmincodesize.to_bytes(1, 'little'))

    # Data sub-block (variable length)
    while True:
        blocktype = rng.read(1)[0]

        if blocktype == 0x2c:
            # Image data sub-block

            # LZW minimum code size (1 byte)
            lzwmincodesize = rng.read(1)[0]
            out.write(lzwmincodesize.to_bytes(1, 'little'))

            # Data sub-block (variable length)
            while True:
                blocktype = rng.read(1)[0]

                if blocktype == 0x00:
                    break

                elif blocktype == 0x01:
                    raise Exception('Invalid GIF image data')

                else:
                    # Packed field
                    packedfield = rng.read(1)[0]

                    # Table size
                    tablesize = 2 ** (packedfield & 0xf) + 1

                    # Entropy encoding method
                    if packedfield & 0x10!= 0:
                        raise Exception('Invalid GIF image data')

                    # Sorting flag
                    if packedfield & 0x20!= 0:
                        raise Exception('Invalid GIF image data')

                    # Reserved bits
                    if packedfield & 0xc0!= 0:
                        raise Exception('Invalid GIF image data')

                    # Codes
                    codes = []
                    for _ in range(tablesize):
                        codes.append(struct.unpack('<H', rng.read(2))[0])

                    # Minimum code size
                    lzwmincodesize = rng.read(1)[0]

                    # End of stream marker
                    if lzwmincodesize == 0:
                        break

                    # Output code
                    outputcode = rng.read(1)[0]
                    while outputcode < len(codes):
                        out.write(codes[outputcode].to_bytes(2, 'little'))

                        # Get next code
                        if outputcode == 0xffff:
                            outputcode = 0
                        else:
                            outputcode += 1

                            # Find longest prefix match
                            maxlen = -1
                            for i in reversed(range(len(codes))):
                                if codes[i] <= outputcode and \
                                        len(codes) - i <= maxlen:
                                    maxlen = len(codes) - i
                                    outputcode = codes[i]

                            # Add new code to table
                            if len(codes) < tablesize:
                                codes.append(outputcode)
                                continue

                            # Remove old code from table
                            del codes[-1]

        elif blocktype == 0x21:
            # Extension sub-block

            # Extension label
            extlabel = rng.read(1)[0]

            # Application extension
            if extlabel == 0xff:
                appextlabel = rng.read(1)[0]
                appid = ''
                for _ in range(appextlabel):
                    appid += chr(rng.read(1)[0])
                appauthcode = ''
                for _ in range(3):
                    appauthcode += chr(rng.read(1)[0])
                out.write(appextlabel.to_bytes(1, 'little'))
                out.write(appid.encode())
                out.write(appauthcode.encode())

            # Comment extension
            elif extlabel == 0xfe:
                commentlength = 0
                while True:
                    commentbyte = rng.read(1)[0]
                    out.write(commentbyte.to_bytes(1, 'little'))
                    commentlength += 1
                    if commentbyte == 0:
                        break
                padding = (-commentlength) % 2
                for _ in range(padding):
                    out.write(b'\x00')

            # Graphic control extension
            elif extlabel == 0xf9:
                gcflags = rng.read(1)[0]
                gctimes = struct.unpack('<HH', rng.read(4))
                gcbgindex = rng.read(1)[0]
                gcratio = rng.read(1)[0]
                out.write(gcflags.to_bytes(1, 'little'))
                out.write(gctimes[0].to_bytes(2, 'little'))
                out.write(gctimes[1].to_bytes(2, 'little'))
                out.write(gcbgindex.to_bytes(1, 'little'))
                out.write(gcratio.to_bytes(1, 'little'))

            # Plain text extension
            elif extlabel == 0x01:
                ptflags = rng.read(1)[0]
                ptx = struct.unpack('<HH', rng.read(4))
                pty = struct.unpack('<HH', rng.read(4))
                ptgridsize = rng.read(1)[0]
                ptcharsize = rng.read(1)[0]
                ptfgindex = rng.read(1)[0]
                ptbgindex = rng.read(1)[0]
                out.write(ptflags.to_bytes(1, 'little'))
                out.write(ptx[0].to_bytes(2, 'little'))
                out.write(ptx[1].to_bytes(2, 'little'))
                out.write(pty[0].to_bytes(2, 'little