#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    width  = int.from_bytes(rng.read(2), 'little') % (1 << 10) + 1
    height = int.from_bytes(rng.read(2), 'little') % (1 << 10) + 1

    # Write header and screen descriptor
    out.write(header)
    out.write(struct.pack('<HHBBB', width, height, 0xF7, 0x0B, 0))

    # Write color table
    for i in range(4):
        out.write(b'\xFF\xFF\xFF')

    # Write image data
    while True:
        lzw_mincodesize = int.from_bytes(rng.read(1), 'little') & 0x0F

        if lzw_mincodesize == 0:
            break

        out.write(lzw_mincodesize.to_bytes(1, byteorder='little'))

        # Generate LZW code size table
        lzw_table = [i for i in range(256)]

        for i in range(256 - lzw_mincodesize):
            j = int.from_bytes(rng.read(1), 'little')

            k = lzw_table[i]
            del lzw_table[i]
            lzw_table += [k] * j

        # Write LZW code size table
        out.write(len(lzw_table).to_bytes(1, byteorder='little'))
        out.write(bytes([lzw_table[i] for i in range(len(lzw_table))]))

        # Generate LZW compressed data
        lzw_dict = {i : bytes([i]) for i in range(256)}
        lzw_prev = bytes()
        lzw_cur  = bytes()

        while len(lzw_cur)!= 0 or len(lzw_prev)!= 0:
            if len(lzw_cur) > 0:
                c = lzw_cur[:1]

                lzw_cur = lzw_cur[1:]
            else:
                c = lzw_prev[:1]

                lzw_prev = lzw_prev[1:]

            if c not in lzw_dict:
                continue

            d = lzw_dict[c]

            if len(d) >= 255:
                raise Exception('LZW dictionary overflowed!')

            lzw_dict[len(lzw_dict)] = lzw_dict[len(lzw_dict)-1] + d

            out.write(d)

            if len(lzw_dict) == 2 ** lzw_mincodesize:
                lzw_mincodesize += 1

                out.write((lzw_mincodesize-1).to_bytes(1, byteorder='little'))

                # Regenerate LZW code size table
                lzw_table = list(range(256))

                for i in range(256 - lzw_mincodesize):
                    j = int.from_bytes(rng.read(1), 'little')

                    k = lzw_table[i]
                    del lzw_table[i]
                    lzw_table += [k] * j

                # Write LZW code size table
                out.write(len(lzw_table).to_bytes(1, byteorder='little'))
                out.write(bytes([lzw_table[i] for i in range(len(lzw_table))]))

                # Regenerate LZW dictionary
                lzw_dict = {}

                for i in range(256):
                    lzw_dict[i] = bytes([i])

                for i in range(len(lzw_table)):
                    lzw_dict[len(lzw_dict)] = \
                        lzw_dict[len(lzw_dict)-1] + lzw_dict[lzw_table[i]]

        # Write end of stream marker
        out.write(b'\x00')

    # Write trailer
    out.write(b';')
    out.write(data)

