#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The width is determined by `width`, height by `height`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int
    ):
    # Header (always 6 bytes long)
    out.write(b'GIF89a')

    # Logical screen descriptor (always 7 bytes long)
    # Width in pixels
    out.write((width & 0xFF).to_bytes(1, 'little'))
    out.write(((width >> 8) & 0xFF).to_bytes(1, 'little'))
    # Height in pixels
    out.write((height & 0xFF).to_bytes(1, 'little'))
    out.write(((height >> 8) & 0xFF).to_bytes(1, 'little'))
    # Flags for color resolution / global color table
    # 1 - Global color table flag set
    # 7 - Color resolution bits set to 7
    out.write(b'\x87\x00')
    # Background color index
    out.write(b'\x00')
    # Pixel aspect ratio
    out.write(b'\x00')

    # Image descriptor block (always 10 bytes long)
    # Left position
    out.write(b'\x00')
    out.write(b'\x00')
    # Top position
    out.write(b'\x00')
    out.write(b'\x00')
    # Width in pixels
    out.write((width & 0xFF).to_bytes(1, 'little'))
    out.write(((width >> 8) & 0xFF).to_bytes(1, 'little'))
    # Height in pixels
    out.write((height & 0xFF).to_bytes(1, 'little'))
    out.write(((height >> 8) & 0xFF).to_bytes(1, 'little'))
    # Flags for local color table
    # 1 - Local color table flag set
    out.write(b'\x01')
    # LZW minimum code size
    out.write(b'\x08')

    # Global color table (always 3 * 256 = 768 bytes long)
    for i in range(256):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        out.write(r.to_bytes(1, 'little'))
        out.write(g.to_bytes(1, 'little'))
        out.write(b.to_bytes(1, 'little'))

    # Frame data
    # Each frame is made up of blocks, each block starts with a single byte
    # that determines how many subsequent bytes are part of this block.
    # If the first byte has the highest bit set, then it's followed by an
    # additional byte which specifies the amount of bytes that follow until
    # another block begins. This repeats until a block with the highest bit
    # not set is encountered.
    while True:
        # Read one byte from the input stream
        c = rng.read(1)[0]

        if c == 0x00:
            break;

        # Check whether we need to read more bytes or not
        if c >= 0x80:
            # We do!
            n = rng.read(1)[0]

            # Copy n bytes from the input stream to the output stream
            out.write(rng.read(n))
        else:
            # Nope, just copy the byte as-is
            out.write(c.to_bytes(1, 'little'))