#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The generated GIF will have a width between 10px and 500px,
# a height between 10px and 500px, and a color depth of either 4 or 8 bits per pixel.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Generate the header for the GIF file
    header = b'GIF89a'
    width = rng.randint(10, 500)
    height = rng.randint(10, 500)
    colordepth = rng.choice([4, 8])
    globalcolortablesize = (
        2 ** colordepth if colordepth <= 8 else None
    )
    backgroundindex = rng.randrange(globalcolortablesize) \
        if globalcolortablesize is not None else None
    aspectratio = rng.choice([0b0000, 0b0001, 0b0010, 0b0011,
                              0b0100, 0b0101, 0b0110, 0b0111,
                              0b1000, 0b1001, 0b1010, 0b1011,
                              0b1100, 0b1101, 0b1110, 0b1111])
    
    # Write the header to the output
    out.write(header)

    # Write the logical screen descriptor to the output
    out.write(struct.pack('>HHBBB',
                          width, height,
                          colordepth,
                          globalcolortablesize,
                          backgroundindex))
    out.write(aspectratio.to_bytes(1, 'big'))

    # Write the global color table to the output
    if globalcolortablesize is not None:
        for i in range(globalcolortablesize):
            r = rng.randrange(256)
            g = rng.randrange(256)
            b = rng.randrange(256)

            out.write((r,g,b))

    # Write the image descriptor to the output
    out.write(struct.pack('>HBHBB',
                          0x2c, # Image separator
                          left, top,
                          width, height,
                          0b0000 | localcolortablesize))

    # Write the local color table to the output
    if localcolortablesize is not None:
        for i in range(localcolortablesize):
            r = rng.randrange(256)
            g = rng.randrange(256)
            b = rng.randrange(256)

            out.write((r,g,b))

    # Write the image data to the output
    for row in range(height):
        lzwmincodesize = rng.randrange(2, 12)

        out.write(lzwmincodesize.to_bytes(1, 'big'))

        # TODO: Implement LZW compression