#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    header = b'GIF89a'
    out.write(header)

    # Logical screen descriptor
    width = int.from_bytes(rng.read(2), byteorder='big') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='big') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    bgcolorindex = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    pixelaspectratio = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    lsd = struct.pack('>HHBBB', width, height, *flags, *bgcolorindex,
                      *pixelaspectratio)
    out.write(lsd)

    # Global color table
    global_colors = [
        bytes([
            int.from_bytes(rng.read(1), byteorder='big'),
            int.from_bytes(rng.read(1), byteorder='big'),
            int.from_bytes(rng.read(1), byteorder='big'),
            0xff]) for _ in range(768)]
    gct = b''
    for i in range(len(global_colors)):
        if len(gct) < 256*4 and i == bgcolorindex[0]: continue
        gct += global_colors[i]
    while len(gct) < 256*4:
        gct += b'\x00\x00\x00\x00'
    out.write(gct)

    # Image descriptor
    left = int.from_bytes(rng.read(2), byteorder='big') % (width - 1)
    top = int.from_bytes(rng.read(2), byteorder='big') % (height - 1)
    width = int.from_bytes(rng.read(2), byteorder='big') % (width - left) + 1
    height = int.from_bytes(rng.read(2), byteorder='big') % (height - top) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    id = struct.pack('>HHHHB', left, top, width, height, *flags)
    out.write(id)

    # Local color table
    local_colors = []
    for _ in range(3):
        local_colors.append(bytes([
            int.from_bytes(rng.read(1), byteorder='big'),
            int.from_bytes(rng.read(1), byteorder='big'),
            int.from_bytes(rng.read(1), byteorder='big'),
            0xff]))
    lct = b''
    for c in local_colors:
        lct += c
    while len(lct) < 256*4:
        lct += b'\x00\x00\x00\x00'
    out.write(lct)

    # Table of image data
    data = rng.read((width+1)*height)
    out.write(data)

    # Trailer
    trailer = b'\x3b'
    out.write(trailer)

