#!/usr/bin/env python3
# Copyright lowRISC contributors.
# Licensed under the Apache License, Version 2.0, see LICENSE for details.
# SPDX-License-Identifier: Apache-2.0
r"""Generate a random animated GIF image.

This script generates a random animated GIF image using /dev/urandom. It writes
the resulting image to the specified output file.
"""
import argparse
import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # get some of our data into the file before writing the rest.
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    out.write(header + datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    for _ in range(0, datalen):
        out.write(rng.read(1))


def main() -> int:
    parser = argparse.ArgumentParser('generate_random_gif')
    parser.add_argument('--out', required=True)
    args = parser.parse_args()

    if not os.path.exists(args.out):
        try:
            with open(args.out, mode='xb'):
                pass
        except FileExistsError:
            print("{} already exists".format(args.out),
                  file=sys.stderr)
            return -1

    with open(args.out, mode='wb') as f:
        with open('/dev/urandom', mode='rb') as rng:
            generate_random_gif(rng, f)

    return 0


if __name__ == '__main__':
    exit(main())