#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), 'big')])
    background_color = bytes([int.from_bytes(rng.read(1), 'big')])
    pixel_aspect_ratio = bytes([int.from_bytes(rng.read(1), 'big')])
    out.write(struct.pack('>HHBBB', width, height, *flags))
    out.write(background_color)
    out.write(pixel_aspect_ratio)

    # Global color table
    global_color_table_size = int.from_bytes(rng.read(1), 'big') & 7
    if global_color_table_size!= 0:
        global_color_table = bytearray()
        for i in range(global_color_table_size):
            global_color_table += bytes([
                int.from_bytes(rng.read(1), 'big'),
                int.from_bytes(rng.read(1), 'big'),
                int.from_bytes(rng.read(1), 'big')
            ])
        out.write(global_color_table)

    # Image descriptor
    left = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    top = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    width = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), 'big') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), 'big')])
    local_color_table_flag = bool((flags[0] >> 7) & 1)
    interlace_flag = bool((flags[0] >> 6) & 1)
    sort_flag = bool((flags[0] >> 5) & 1)
    reserved = bytes([int.from_bytes(rng.read(1), 'big')])
    local_color_table_size = int.from_bytes(rng.read(1), 'big') & 7
    local_color_table = bytearray()
    for i in range(local_color_table_size):
        local_color_table += bytes([
            int.from_bytes(rng.read(1), 'big'),
            int.from_bytes(rng.read(1), 'big'),
            int.from_bytes(rng.read(1), 'big')
        ])
    out.write(struct.pack('>HHHHB', left, top, width, height, *flags))
    if local_color_table_flag:
        out.write(local_color_table)

    # Table based image data
    lzw_min_code_size = int.from_bytes(rng.read(1), 'big')
    lzw_max_code_size = min(lzw_min_code_size + 1, 12)
    lzw_clear_code = 1 << lzw_min_code_size
    lzw_end_of_information_code = lzw_clear_code + 1
    lzw_next_code = lzw_end_of_information_code + 1
    lzw_dictionary = {}
    lzw_buffer = []
    while True:
        code = int.from_bytes(rng.read(1), 'big')
        if code == lzw_end_of_information_code:
            break

        if code < lzw_next_code:
            lzw_buffer.append(code)

            if len(lzw_buffer) >= lzw_max_code_size:
                lzw_word = tuple(lzw_buffer[:lzw_max_code_size])

                if not lzw_