#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom). The
# width and height are specified in pixels as `width`
# and `height`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO,
        width: int, height: int
    ):
    # Create the header for this image
    header = (b'GIF89a'
              + width.to_bytes(2, byteorder='big')
              + height.to_bytes(2, byteorder='big'))
    
    # Generate the palette
    palsize = 768
    palette = [0] * palsize
    for i in range(palsize):
        palette[i] = rng.read(1)[0]

    # Write the header with the palette size
    out.write(header + bytes([palsize]))

    # Write the palette
    for p in palette:
        out.write(p.to_bytes(1, byteorder='big'))

    # Get the color table size from the first byte
    colortable = rng.read(1)[0]
    out.write(colortable.to_bytes(1, byteorder='big'))

    # Write the color table
    for _ in range(colortable):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        out.write((r << 4 | g >> 4).to_bytes(1, byteorder='big'))
        out.write(((g & 0xf) << 4 | b >> 4).to_bytes(1, byteorder='big'))

    # Get the LZW minimum code size from the next byte
    lzwmincodesize = rng.read(1)[0]
    out.write(lzwmincodesize.to_bytes(1, byteorder='big'))

    # Get the data block size from the next two bytes
    datablocklen = int.from_bytes(
            rng.read(2), byteorder='big', signed=False
        )
    out.write(datablocklen.to_bytes(2, byteorder='big'))

    # Write the compressed data
    while True:
        # Read one byte at a time until we get an EOI marker
        c = rng.read(1)[0]
        if c == 0x00: break
        
        # If it's not an EOI marker then write it directly
        if c!= 0x01:
            out.write(c.to_bytes(1, byteorder='big'))
            continue

        # Otherwise read another byte which is the count
        # of how many times to repeat that previous byte
        count = rng.read(1)[0]
        out.write(count.to_bytes(1, byteorder='big'))

        # And finally write the repeated byte itself
        out.write(c.to_bytes(1, byteorder='big'))