#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor (LSD)
    width = rng.read(1)[0] % 64 + 1
    height = rng.read(1)[0] % 64 + 1
    flags = bytes([
        0b10000000 | rng.read(1)[0],   # Global color table flag
        0b00000000 | rng.read(1)[0],   # Color resolution
        0b00000000 | rng.read(1)[0],   # Sorting method
        0b00000000 | rng.read(1)[0],   # Size of global color table
        0b00000000 | rng.read(1)[0],   # Background color index
        0b00000000 | rng.read(1)[0],   # Pixel aspect ratio
    ])
    gctsize = rng.read(1)[0] & 7
    bgindex = rng.read(1)[0]
    pixelaspectratio = rng.read(1)[0]
    lsd = [width, height, flags, gctsize, bgindex, pixelaspectratio]
    for i in range(len(lsd)):
        out.write(struct.pack('B', lsd[i]))

    # Global color table
    if gctsize > 0:
        for _ in range(gctsize):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            out.write(r)
            out.write(g)
            out.write(b)

    # Image descriptor
    left = rng.read(1)[0] % width
    top = rng.read(1)[0] % height
    flags = bytes([
        0b10000000 | rng.read(1)[0],   # Local color table flag
        0b00000000 | rng.read(1)[0],   # Interlace flag
        0b00000000 | rng.read(1)[0],   # Sorting flag
        0b00000000 | rng.read(1)[0],   # Reserved
        0b00000000 | rng.read(1)[0],   # Size of local color table
        0b00000000 | rng.read(1)[0],   # Reserved
    ])
    lctsize = rng.read(1)[0] & 7
    interleavemode = rng.read(1)[0]
    sortflag = rng.read(1)[0]
    reserved = rng.read(1)[0]
    id = [left, top, flags, lctsize, interleavemode, sortflag, reserved]
    for i in range(len(id)):
        out.write(struct.pack('B', id[i]))

    # Local color table
    if lctsize > 0:
        for _ in range(lctsize):
            r = rng.read(1)[0]
            g = rng.read(1)[0]
            b = rng.read(1)[0]
            out.write(r)
            out.write(g)
            out.write(b)

    # Table based image data
    lzwmincodesize = rng.read(1)[0]
    out.write(lzwmincodesize)
    while True:
        blocktype = rng.read(1)[0]
        if blocktype == 0x00:    # End of information code
            break

        elif blocktype == 0x01:  # Plain text extension
            pass

        elif blocktype == 0x02:  # Application extension
            pass

        elif blocktype == 0x03:  # Comment extension
            pass

        else:                    # Image data
            numcodes = rng.read(1)[0]
            if numcodes < 256:
                for _ in range(numcodes):
                    out.write(rng.read(1))

            else:
                for _ in range(numcodes - 256):
                    out.write(rng.read(1))