#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical Screen Descriptor
    width = int.from_bytes(rng.read(2), byteorder='big') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='big') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    background = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    aspect_ratio = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    out.write(struct.pack('>HHBBB', width, height, *flags))
    out.write(background)
    out.write(aspect_ratio)

    # Global Color Table
    num_colors = int.from_bytes(rng.read(1), byteorder='big') & 7
    if num_colors == 0:
        num_colors += 1
    color_table = [None] * num_colors
    for i in range(num_colors):
        r = int.from_bytes(rng.read(1), byteorder='big')
        g = int.from_bytes(rng.read(1), byteorder='big')
        b = int.from_bytes(rng.read(1), byteorder='big')
        color_table[i] = bytes((r,g,b))
    for c in color_table:
        out.write(c)

    # Image Descriptor
    left = int.from_bytes(rng.read(2), byteorder='big')
    top = int.from_bytes(rng.read(2), byteorder='big')
    width = int.from_bytes(rng.read(2), byteorder='big') % (2**10 - 1) + 1
    height = int.from_bytes(rng.read(2), byteorder='big') % (2**10 - 1) + 1
    flags = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    local_color_table = bool(int.from_bytes(rng.read(1), byteorder='big'))
    interlace = bool(int.from_bytes(rng.read(1), byteorder='big'))
    sort = bool(int.from_bytes(rng.read(1), byteorder='big'))
    reserved = bytes([int.from_bytes(rng.read(1), byteorder='big')])
    lct_size = int.from_bytes(rng.read(1), byteorder='big') & 7
    if lct_size == 0:
        lct_size += 1
    lct = []
    for _ in range(lct_size):
        r = int.from_bytes(rng.read(1), byteorder='big')
        g = int.from_bytes(rng.read(1), byteorder='big')
        b = int.from_bytes(rng.read(1), byteorder='big')
        lct.append(bytes((r,g,b)))
    for lc in lct:
        out.write(lc)
    out.write(struct.pack('>HHBBB', left, top, width, height, *flags))

    # Local Color Table
    while len(lct) < 2**(int(flags[-1]) + 1):
        r = int.from_bytes(rng.read(1), byteorder='big')
        g = int.from_bytes(rng.read(1), byteorder='big')
        b = int.from_bytes(rng.read(1), byteorder='big')
        lct.append(bytes((r,g,b)))
    for lc in lct:
        out.write(lc)

    # Image Data
    image_data = rng.read(width*height)
    out.write(image_data)

    # Trailer
    out.write(b';')

