#!/usr/bin/env python3
import sys
from typing import BinaryIO


def main():
    with open('/dev/urandom', 'rb') as rng,\
         open('out.gif', 'wb') as out:
        generate_random_gif(rng, out)


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1024
    height = 768
    color_depth = 8
    global_color_table = False
    sort_flag = True
    size = (width * height).bit_length() - 1
    packed = ((size & 0x07) << 4 |
              (color_depth & 0x07) << 1 |
              0 if global_color_table else 1 << 5 |
               0 if sort_flag else 1 << 3)
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    out.write(packed.to_bytes(1, 'little'))

    # Global color table
    for _ in range(2 ** color_depth):
        out.write(rng.read(3))

    # Image descriptor
    left = 0
    top = 0
    width -= 1
    height -= 1
    local_color_table = False
    interlace_flag = False
    sort_flag = True
    lct_size = 0
    size = (width * height).bit_length() - 1
    packed = ((size & 0x07) << 4 |
              (lct_size & 0x07) << 1 |
              0 if local_color_table else 1 << 5 |
               0 if interlace_flag else 1 << 6 |
                 0 if sort_flag else 1 << 3)
    out.write(((left & 0xFFF8) >> 3).to_bytes(2, 'little'))
    out.write(((top & 0xFFF8) >> 3).to_bytes(2, 'little'))
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    out.write(packed.to_bytes(1, 'little'))

    # Local color table
    for _ in range(2 ** lct_size):
        out.write(rng.read(3))

    # Data sub-block
    while True:
        block_type = rng.read(1)[0]

        if block_type == 0x2C:
            break

        elif block_type!= 0x21:
            raise ValueError("Invalid block type")

        block_size = int.from_bytes(
                rng.read(1), byteorder='big')
        if block_size < 2:
            raise ValueError("Invalid block size")

        label = rng.read(1)[0]
        if label!= 0xF9:
            raise ValueError("Invalid extension label")

        flags = rng.read(1)[0]
        delay = int.from_bytes(
                rng.read(2), byteorder='big')
        transparent_index = rng.read(1)[0]
        terminator = rng.read(1)[0]
        if terminator!= 0:
            raise ValueError("Invalid terminator")

        if flags & 1:
            continue

        while block_size > 0:
            chunk_size = min(
                    rng.read(1)[0], block_size)

            out.write(chunk_size.to_bytes(1, 'little'))
            out.write(rng.read(chunk_size))
            block_size -= chunk_size

    # Trailer
    out.write(b'\x3B')


if __name__ == '__main__':
    main()