#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    height = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    out.write(struct.pack('<HHBBB', width, height, 0xF0, 0x00, 0x00))

    # Write the global color table
    for i in range(768):
        out.write(bytes([(i >> ((~i + 1) % 8)) & 0xFF]))

    # Write the application extension block
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00\x00')

    # Write the image descriptor
    out.write(b'\x2c\x00\x00\x00\x00\x00\x00\x00\x00')

    # Write the local color table
    for i in range(768):
        out.write(bytes([(i >> ((~i + 1) % 8)) & 0xFF]))

    # Write the image data
    out.write(b'\x00\x00')

    # Write the LZW compressed image data
    lzw_min_code_sizes = [8, 4, 2, 0]
    code_size = 8
    clear_code = 0x11
    end_of_information_code = 0x100
    next_code = end_of_information_code + 1
    max_code = (1 << code_size) - 1
    first_code = True
    old_code = None
    for _ in range(width * height):
        if first_code:
            first_code = False
            code = rng.read(1)[0] & 0xFF
        else:
            code = rng.read(1)[0] & 0xFF

        if code == clear_code:
            code_size = lzw_min_code_sizes[next_code - end_of_information_code]
            max_code = (1 << code_size) - 1
            old_code = None
        elif code <= max_code:
            if old_code is not None:
                if code < next_code:
                    out.write(bytes([old_code]))
                    out.write(bytes([code]))
                else:
                    out.write(bytes([old_code]))
                    out.write(bytes([old_code]))
                    old_code = code
            else:
                old_code = code

            if next_code == max_code:
                code_size += 1
                max_code = (1 << code_size) - 1

            next_code += 1
            if next_code > max_code:
                break
            else:
                out.write(bytes([code]))

    # Write the trailer
    out.write(b';')
