#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = rng.read(1)[0] % 256 + 1
    height = rng.read(1)[0] % 256 + 1
    background_color = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBB', width, height, background_color, pixel_aspect_ratio))

    # Global color table
    global_color_table_size = 7 + rng.read(1)[0] % 8
    global_color_table_sort_flag = rng.read(1)[0] & 1
    global_color_table_colors = [None] * (2 ** global_color_table_size)
    for i in range(len(global_color_table_colors)):
        red = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue = rng.read(1)[0] % 256
        global_color_table_colors[i] = (red, green, blue)
    out.write(struct.pack('B', global_color_table_size | (global_color_table_sort_flag << 7)))
    for rgb in global_color_table_colors:
        out.write(struct.pack('BBB', *rgb))

    # Image block
    image_left = rng.read(1)[0] % 256
    image_top = rng.read(1)[0] % 256
    image_width = rng.read(1)[0] % 256 + 1
    image_height = rng.read(1)[0] % 256 + 1
    local_color_table_flag = rng.read(1)[0] >> 7
    interlace_flag = rng.read(1)[0] & 1
    sort_flag = rng.read(1)[0] & 1
    local_color_table_size = 7 + rng.read(1)[0] % 8
    out.write(struct.pack('<HHBB', image_left, image_top, image_width, image_height))
    out.write(struct.pack('BB', local_color_table_flag | (interlace_flag << 1) | (sort_flag << 2), local_color_table_size))
    if local_color_table_flag == 1:
        local_color_table_colors = [None] * (2 ** local_color_table_size)
        for i in range(len(local_color_table_colors)):
            red = rng.read(1)[0] % 256
            green = rng.read(1)[0] % 256
            blue = rng.read(1)[0] % 256
            local_color_table_colors[i] = (red, green, blue)
        for rgb in local_color_table_colors:
            out.write(struct.pack('BBB', *rgb))

    # Graphic control extension
    gce_block_size = 4
    gce_reserved = rng.read(1)[0]
    gce_disposal_method = rng.read(1)[0] % 8
    gce_user_input_flag = rng.read(1)[0] >> 7
    gce_transparent_color_flag = rng.read(1)[0] & 1
    delay_time = rng.read(2)
    transparent_color_index = rng.read(1)[0]
    block_terminator = b'\x00'
    out.write(struct.pack('BBBBB', gce_block_size, gce_reserved, gce_disposal_method | (gce_user_input_flag << 2) | (gce_transparent_color_flag << 3), *delay_time, transparent_color_index))
    out.write(block_terminator)

    # Application extension
    app_extension_block_size = 11
    application_identifier = 'A'*8
    authentication_code = 'B'*3
    out.write(struct.pack('BB', app_extension_block_size, len(application_identifier)))
    out.write(application_identifier.encode())
    out.write(authentication_code.encode())
    out.write(block_terminator)

    # Comment extension
    comment_extension_block_size = 1
    comment_data = 'C'*255
    out.write(struct.pack('BB', comment_extension_block_size, len(comment_data)))
    out.write(comment_data.encode())
    out.write(block_terminator)

    # Plain text extension
    plain_text_extension_block_size = 12
    plain_text_grid_left = rng.read(1)[0] % 256
    plain_text_grid_top = rng.read(1)[0] % 256
    plain_text_grid_width = rng.read(1)[0] % 256 + 1
    plain_text_grid_height = rng.read(1)[0] % 256 + 1
    character_cell_width = rng.read(1)[0] % 256
    character_cell_height = rng.read(1)[0] % 256
    text_fg_color_index = rng.read(1)[0] % 256
    text_bg_color_index = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBBBBB', plain_text_grid_left, plain_text_grid_top, plain_text_grid_width, plain_text_grid_height, character_cell_width, character_cell_height, text_fg_color_index, text_bg_color_index))
    out.write(block_terminator)

    # Image descriptor
    image_descriptor_block_size = 9
    image_lct_flag = rng.read(1)[0] >> 7
    image_interlace_flag = rng.read(1)[0] & 1
    image_sort_flag = rng.read(1)[0] & 1
    image_lct_size = 7 + rng.read(1)[0] % 8
    out.write(struct.pack('BB', image_descriptor_block_size, image_lct_flag | (image_interlace_flag << 1) | (image_sort_flag << 2), image_lct_size))
    if image_lct_flag == 1:
        image_lct_colors = [None] * (2 ** image_lct_size)
        for i in range(len(image_lct_colors)):
            red = rng.read(1)[0] % 256
            green = rng.read(1)[0] % 256
            blue = rng.read(1)[0] % 256
            image_lct_colors[i] = (red, green, blue)    out.write(struct.pack('B', lzw_minimum_code_size))

    # Pixel data
    for y in range(height):
        for x in range(width):
            index = rng.read(1)[0] % 256
            out.write(struct.pack('B', index))