#!/usr/bin/env python3

import sys
import struct
from collections import namedtuple

# https://docs.python.org/3/library/struct.html#format-characters

def read_int(f):
    return int.from_bytes(f.read(4), byteorder='little')

def read_float(f):
    return struct.unpack('<f', f.read(4))[0]

def read_string(f):
    length = read_int(f)
    return f.read(length).decode()

def write_int(f, i):
    f.write(i.to_bytes(4, 'little'))

def write_float(f, x):
    f.write(struct.pack('<f', x))

def write_string(f, s):
    if type(s) is str:
        s = s.encode()
    write_int(f, len(s))
    f.write(s)

def read_vec2(f):
    return [read_float(f) for _ in range(2)]

def write_vec2(f, v):
    assert len(v) == 2
    for x in v:
        write_float(f, x)

def read_vec3(f):
    return [read_float(f) for _ in range(3)]

def write_vec3(f, v):
    assert len(v) == 3
    for x in v:
        write_float(f, x)

def read_vec4(f):
    return [read_float(f) for _ in range(4)]

def write_vec4(f, v):
    assert len(v) == 4
    for x in v:
        write_float(f, x)

def read_mat4(f):
    return [[read_float(f) for _ in range(4)] for _ in range(4)]

def write_mat4(f, m):
    assert len(m) == 4 and all([len(r) == 4 for r in m])
    for row in m:
        for x in row:
            write_float(f, x)

def read_quat(f):
    return [read_float(f) for _ in range(4)]

def write_quat(f, q):
    assert len(q) == 4
    for x in q:
        write_float(f, x)

def read_color(f):
    return [read_float(f) for _ in range(4)]

def write_color(f, c):
    assert len(c) == 4
    for x in c:
        write_float(f, x)

def read_mesh(f):
    num_vertices = read_int(f)
    vertices = []
    for _ in range(num_vertices):
        vertices.append({
            'position': read_vec3(f),
            'normal': read_vec3(f),
            'uv': read_vec2(f),
            'tangent': read_vec4(f),
            'bone_weights': read_vec4(f),
            'bone_indices': read_vec4(f),
        })
    num_triangles = read_int(f)
    triangles = []
    for _ in range(num_triangles):
        triangles.append({
            'a': read_int(f),
            'b': read_int(f),
            'c': read_int(f),
        })
    return {
        'vertices': vertices,
        'triangles': triangles,
    }

def write_mesh(f, mesh):
    write_int(f, len(mesh['vertices']))
    for vertex in mesh['vertices']:
        write_vec3(f, vertex['position'])
        write_vec3(f, vertex['normal'])
        write_vec2(f, vertex['uv'])
        write_vec4(f, vertex['tangent'])
        write_vec4(f, vertex['bone_weights'])
        write_vec4(f, vertex['bone_indices'])
    write_int(f, len(mesh['triangles']))
    for triangle in mesh['triangles']: