#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul 15 20:47:49 2019

@author: jinyi
"""
import numpy as np
from scipy.stats import norm, multivariate_normal

def get_data(n):
    """
    Generate data from a bivariate normal distribution with mean [0, 0] and covariance matrix [[1, 0], [0, 1]]
    
    Parameters
    ----------
    n : int
        Number of samples to generate
        
    Returns
    -------
    x : array
        Samples generated from the bivariate normal distribution
    """
    # Generate random numbers from standard normal distribution
    r = np.random.randn(2, n)
    # Multiply by cholesky decomposition of covariance matrix
    return r @ np.linalg.cholesky([[1, 0], [0, 1]])

def log_likelihood(x, mu, sigma):
    """
    Compute the log likelihood of the data under a bivariate normal distribution with given parameters
    
    Parameters
    ----------
    x : array
        Data points
    mu : array
        Mean vector
    sigma : array
        Covariance matrix
        
    Returns
    -------
    ll : float
        Log likelihood of the data under the bivariate normal distribution
    """
    # Get dimension of the data
    d = len(mu)
    # Compute the determinant of the covariance matrix
    det = np.linalg.det(sigma)
    # Compute the inverse of the covariance matrix
    inv = np.linalg.inv(sigma)
    # Compute the exponent term in the exponential
    exp = -0.5 * (x - mu).T @ inv @ (x - mu)
    # Return the log likelihood
    return -d / 2 * np.log(2 * np.pi) - 0.5 * np.log(det) + exp

def mle(x):
    """
    Estimate the maximum likelihood estimator for the mean vector and covariance matrix of the bivariate normal distribution using the EM algorithm
    
    Parameters
    ----------
    x : array
        Data points
        
    Returns
    -------
    mu : array
        Maximum likelihood estimate of the mean vector
    sigma : array
        Maximum likelihood estimate of the covariance matrix
    """
    # Get number of samples
    N = len(x)
    # Initialize the mean vector and covariance matrix
    mu = np.zeros((2,))
    sigma = np.eye(2)
    # Initialize the old log likelihood
    old_ll = -np.inf
    # Iterate until convergence
    while True:
        # E step: compute the responsibilities
        resp = np.exp(log_likelihood(x, mu, sigma))
        # M step: update the mean vector and covariance matrix
        mu = np.sum(resp[:, None] * x, axis=0) / np.sum(resp)
        sigma = (x - mu).T @ (x - mu) / N
        # Check if converged
        new_ll = log_likelihood(x, mu, sigma)
        if abs(new_ll - old_ll) < 1e-6:
            break
        else:
            old_ll = new_ll
            
    return mu, sigma
