#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = rng.read(1)[0] % 40 + 10
    height = rng.read(1)[0] % 40 + 10
    color_table_size = 7
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256

    out.write(struct.pack('<HHBB', width, height, color_table_size, background_color_index))
    out.write(pixel_aspect_ratio)

    # Global color table
    for i in range(2 ** color_table_size):
        red   = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue  = rng.read(1)[0] % 256
        out.write(struct.pack('BBB', red, green, blue))

    # Image descriptor
    left = rng.read(1)[0] % width
    top = rng.read(1)[0] % height
    width = rng.read(1)[0] % (width - left) + 1
    height = rng.read(1)[0] % (height - top) + 1
    local_color_table_flag = rng.read(1)[0] & 0x80 != 0
    interlace_flag = rng.read(1)[0] & 0x40 != 0
    sort_flag = rng.read(1)[0] & 0x20 != 0
    reserved_local_color_table_size = rng.read(1)[0] & 0x07
    out.write(struct.pack('<HHBB', left, top, width, height))
    out.write(struct.pack('BB', local_color_table_flag | interlace_flag << 1 | sort_flag << 2 | reserved_local_color_table_size << 3, 0))

    # Local color table
    if local_color_table_flag:
        for i in range(2 ** reserved_local_color_table_size):
            red   = rng.read(1)[0] % 256
            green = rng.read(1)[0] % 256
            blue  = rng.read(1)[0] % 256
            out.write(struct.pack('BBB', red, green, blue))

    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0] % 12 + 1
    out.write(lzw_minimum_code_size)

    # Pixels
    clear_code = 1 << lzw_minimum_code_size
    end_of_information_code = clear_code + 1
    next_available_code = end_of_information_code + 1
    old_code = None
    while True:
        new_code = rng.read(1)[0] % next_available_code
        if new_code == clear_code:
            lzw_minimum_code_size += 1
            next_available_code = end_of_information_code + 1
            old_code = None
        elif new_code == end_of_information_code:
            break
        else:
            if old_code is not None:
                if new_code >= next_available_code:
                    out.write(struct.pack('B', old_code))
                    out.write(struct.pack('B', old_code))
                else:
                    out.write(struct.pack('B', old_code))
            old_code = new_code
            next_available_code += 1

    # Trailer
    out.write(b'\x00')
