#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
def generate_gif(out: BinaryIO, rng: BinaryIO):
    # Magic header.
    out.write(b'GIF89a')

    # Screen descriptor.
    width = 16 + (rng.read(1)[0] % 256)
    height = 16 + (rng.read(1)[0] % 256)
    background = rng.read(1)[0] & 0xFF
    aspect_ratio = rng.read(1)[0] & 0xFF
    out.write(struct.pack('<HHBB', width, height, background, aspect_ratio))

    # Global color table.
    for i in range(7):
        out.write(rng.read(3))

    # Image descriptor.
    left = rng.read(2)[0] & 0xFFFF
    top = rng.read(2)[0] & 0xFFFF
    width = rng.read(2)[0] & 0xFFFF
    height = rng.read(2)[0] & 0xFFFF
    flags = rng.read(1)[0] & 0xFF
    out.write(struct.pack('<HHHHB', left, top, width, height, flags))

    # Local color table.
    for i in range(3):
        out.write(rng.read(3))

    # Graphic control extension.
    block_size = 4
    packed = rng.read(1)[0] & 0xFF
    delay = rng.read(2)[0] & 0xFFFF
    transparent = rng.read(1)[0] & 0xFF
    out.write(struct.pack('<BBHB', block_size, packed, delay, transparent))

    # Plain text extension.
    block_size = 12
    grid_left = rng.read(2)[0] & 0xFFFF
    grid_top = rng.read(2)[0] & 0xFFFF
    grid_width = rng.read(2)[0] & 0xFFFF
    grid_height = rng.read(2)[0] & 0xFFFF
    cell_width = rng.read(1)[0] & 0xFF
    cell_height = rng.read(1)[0] & 0xFF
    fg = rng.read(1)[0] & 0xFF
    bg = rng.read(1)[0] & 0xFF
    out.write(struct.pack('<BBHHHHBB', block_size, grid_left, grid_top, grid_width, grid_height, cell_width, cell_height, fg, bg))

    # Comment extension.
    block_size = 11
    comment = b'Hello world!'
    out.write(struct.pack('<BB', block_size, len(comment)))
    out.write(comment)

    # Application extension.
    block_size = 11
    application = b'Python'
    out.write(struct.pack('<BB', block_size, len(application)))
    out.write(application)

    # Image data.
    lzw_minimum_code_size = rng.read(1)[0] & 0xFF
    out.write(lzw_minimum_code_size.to_bytes(1, 'big'))

    # LZW-compressed image data.
    while True:
        code = rng.read(1)[0] & 0xFF
        if code == 0:
            break
        out.write(code.to_bytes(1, 'big'))
