#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = rng.read(1)[0] % 256 + 1
    height = rng.read(1)[0] % 256 + 1
    background_color = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBB', width, height, background_color, pixel_aspect_ratio))

    # Global color table
    global_color_table_size = 7 + rng.read(1)[0] % 8
    global_color_table_sort_flag = rng.read(1)[0] & 1
    global_color_table_colors = [None] * (2 ** global_color_table_size)
    for i in range(len(global_color_table_colors)):
        red = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue = rng.read(1)[0] % 256
        global_color_table_colors[i] = (red, green, blue)
    out.write(struct.pack('B', global_color_table_size | (global_color_table_sort_flag << 7)))
    for rgb in global_color_table_colors:
        out.write(struct.pack('BBB', *rgb))

    # Image block
    image_left = rng.read(1)[0] % 256
    image_top = rng.read(1)[0] % 256
    image_width = rng.read(1)[0] % 256 + 1
    image_height = rng.read(1)[0] % 256 + 1
    local_color_table_flag = rng.read(1)[0] >> 7
    interlace_flag = rng.read(1)[0] & 1
    sort_flag = rng.read(1)[0] & 1
    local_color_table_size = 7 + rng.read(1)[0] % 8
    out.write(struct.pack('<HHBB', image_left, image_top, image_width, image_height))
    out.write(struct.pack('BB', local_color_table_flag | (interlace_flag << 1) | (sort_flag << 2), local_color_table_size))
    if local_color_table_flag == 1:
        local_color_table_colors = [None] * (2 ** local_color_table_size)
        for i in range(len(local_color_table_colors)):
            red = rng.read(1)[0] % 256
            green = rng.read(1)[0] % 256
            blue = rng.read(1)[0] % 256
            local_color_table_colors[i] = (red, green, blue)
        for rgb in local_color_table_colors:
            out.write(struct.pack('BBB', *rgb))

    # Graphic control extension
    gce_block_size = 4
    gce_reserved = 0
    gce_disposal_method = rng.read(1)[0] % 8
    gce_user_input_flag = rng.read(1)[0] >> 7
    gce_transparent_color_flag = rng.read(1)[0] & 1
    delay_time = rng.read(2)
    transparent_color_index = rng.read(1)[0] % 256
    block_terminator = 0
    out.write(struct.pack('BBBBB', gce_block_size, gce_reserved, gce_disposal_method | (gce_user_input_flag << 2) | (gce_transparent_color_flag << 3), *delay_time, transparent_color_index))
    out.write(struct.pack('B', block_terminator))

    # Plain text extension
    pte_block_size = 12
    pte_block_identifier = b'PTE'
    pte_block_length = 2
    pte_text_grid_left = rng.read(1)[0] % 256
    pte_text_grid_top = rng.read(1)[0] % 256
    pte_text_grid_width = rng.read(1)[0] % 256 + 1
    pte_text_grid_height = rng.read(1)[0] % 256 + 1
    pte_character_cell_width = rng.read(1)[0] % 256
    pte_character_cell_height = rng.read(1)[0] % 256
    pte_text_fg_color_index = rng.read(1)[0] % 256
    pte_text_bg_color_index = rng.read(1)[0] % 256
    pte_text_data = rng.read(pte_block_length - 1)
    out.write(struct.pack('BBBBBBBBBB', pte_block_size, *pte_block_identifier, pte_block_length, pte_text_grid_left, pte_text_grid_top, pte_text_grid_width, pte_text_grid_height, pte_character_cell_width, pte_character_cell_height, pte_text_fg_color_index, pte_text_bg_color_index))
    out.write(pte_text_data)

    # Application extension
    app_extension_block_size = 11
    app_identification_code = b'NETSCAPE2.0'
    app_authentication_code = b'APPLE'
    app_data_sub_block_size = 3
    loop_count = rng.read(2)
    out.write(struct.pack('BB', app_extension_block_size, *app_identification_code))
    out.write(struct.pack('BB', *app_authentication_code))
    out.write(struct.pack('BB', app_data_sub_block_size, 1))
    out.write(loop_count)

    # Comment extension
    comment_extension_block_size = 13
    comment_data = rng.read(comment_extension_block_size - 1)
    out.write(struct.pack('BB', comment_extension_block_size, 0xFE))
    out.write(comment_data)

    # Trailer
    out.write(b'\x3B')
