#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0]))

    # Global color table
    for i in range(7):
        r = int.from_bytes(rng.read(1), byteorder='little')
        g = int.from_bytes(rng.read(1), byteorder='little')
        b = int.from_bytes(rng.read(1), byteorder='little')
        out.write(struct.pack('<BBB', r, g, b))

    # Image data
    while True:
        # Graphic control extension
        if rng.read(1)[0] == 0x21:
            out.write(b'\xF9\x04\x08\x00\x00')

            # Disposal method
            disposal = bytes([int.from_bytes(rng.read(1), byteorder='little') & 0x07])
            out.write(disposal)

            # User input flag
            user_input = bytes([int.from_bytes(rng.read(1), byteorder='little') & 0x01])
            out.write(user_input)

            # Transparent color index
            transp = bytes([int.from_bytes(rng.read(1), byteorder='little')])
            out.write(transp)

            # Delay time
            delay = int.from_bytes(rng.read(2), byteorder='little')
            out.write(delay.to_bytes(2, 'little'))

            # Block terminator
            out.write(b'\x00')

        # Image descriptor
        elif rng.read(1)[0] == 0x2C:
            out.write(b'\x2C\x00\x00\x00\x00')

            # Left position
            left = int.from_bytes(rng.read(2), byteorder='little')
            out.write(left.to_bytes(2, 'little'))

            # Top position
            top = int.from_bytes(rng.read(2), byteorder='little')
            out.write(top.to_bytes(2, 'little'))

            # Width
            width = int.from_bytes(rng.read(2), byteorder='little')
            out.write(width.to_bytes(2, 'little'))

            # Height
            height = int.from_bytes(rng.read(2), byteorder='little')
            out.write(height.to_bytes(2, 'little'))

            # Local color table
            local_table_flag = bytes([int.from_bytes(rng.read(1), byteorder='little') & 0x80])
            out.write(local_table_flag)

            # Interlaced image
            interlace = bytes([int.from_bytes(rng.read(1), byteorder='little') & 0x40])
            out.write(interlace)

            # Sort flag
            sort = bytes([int.from_bytes(rng.read(1), byteorder='little') & 0x08])
            out.write(sort)

            # Size of local color table
            size = bytes([int.from_bytes(rng.read(1), byteorder='little') & 0x07])
            out.write(size)

            # Local color table
            for i in range(2 ** (size[0] + 1)):
                r = int.from_bytes(rng.read(1), byteorder='little')
                g = int.from_bytes(rng.read(1), byteorder='little')
                b = int.from_bytes(rng.read(1), byteorder='little')
                out.write(struct.pack('<BBB', r, g, b))

            # LZW minimum code size
            lzw_minimum_code_size = bytes([int.from_bytes(rng.read(1), byteorder='little')])
            out.write(lzw_minimum_code_size)

            # Image data
            while True:
                # Data sub-blocks
                length = int.from_bytes(rng.read(