#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = rng.read(1)[0] % 256 + 1
    height = rng.read(1)[0] % 256 + 1
    background_color = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBB', width, height, background_color, pixel_aspect_ratio))

    # Global color table
    global_color_table_size = 7 + rng.read(1)[0] % 8
    global_color_table_sort_flag = rng.read(1)[0] & 1
    global_color_table_colors = [None] * (2 ** global_color_table_size)
    for i in range(len(global_color_table_colors)):
        red = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue = rng.read(1)[0] % 256
        global_color_table_colors[i] = (red, green, blue)
    out.write(bytes([
        0xF8 | global_color_table_size,
        global_color_table_sort_flag << 7,
    ]))
    for color in global_color_table_colors:
        out.write(bytes(color))

    # Application extension
    application_extension_block_size = 11
    application_identifier = b'NETSCAPE2.0'
    application_authentication_code = b' '
    loop_count = 0
    out.write(struct.pack('<BHB',
        application_extension_block_size,
        len(application_identifier),
        len(application_authentication_code)))
    out.write(application_identifier)
    out.write(application_authentication_code)
    out.write(struct.pack('<H', loop_count))

    # Image descriptor
    image_descriptor_width = width
    image_descriptor_height = height
    image_descriptor_left = 0
    image_descriptor_top = 0
    local_color_table_flag = 0
    interlace_flag = 0
    sort_flag = 0
    local_color_table_size = 0
    out.write(struct.pack('<BBHHBB',
        image_descriptor_width,
        image_descriptor_height,
        image_descriptor_left,
        image_descriptor_top,
        local_color_table_flag << 7 | interlace_flag << 6 | sort_flag << 5 | local_color_table_size,
        0))

    # Local color table
    local_color_table_colors = []
    for i in range(len(local_color_table_colors)):
        red = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue = rng.read(1)[0] % 256
        local_color_table_colors.append((red, green, blue))
    for color in local_color_table_colors:
        out.write(bytes(color))

    # LZW minimum code size
    lzw_minimum_code_size = 8
    out.write(lzw_minimum_code_size.to_bytes(1, byteorder='big'))

    # LZW data
    lzw_data = []
    while True:
        if rng.read(1)[0] % 2 == 0:
            lzw_data.append(rng.read(1)[0])
        else:
            break
    out.write(bytes(lzw_data))

    # Trailer
    out.write(b'\x3B')
