#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 10 15:26:49 2019

@author: jlee
"""

import numpy as np
from scipy import interpolate, integrate
import matplotlib.pyplot as plt

def get_flux(wavelength, flux):
    """
    

    Parameters
    ----------
    wavelength : array
        Wavelength in Angstroms.
    flux : array
        Flux in erg/cm^2/s/Angstrom.

    Returns
    -------
    f_lambda : array
        Flux in erg/cm^2/s/Hz.
    lambda_axis : array
        Wavelength in Hz.

    """
    
    # Convert to Hz
    lambda_axis = (1e8 / wavelength) * 1e-12
    
    # Interpolate the flux onto a new axis
    f_lambda = interpolate.interp1d(lambda_axis, flux, kind='cubic')
    
    return f_lambda, lambda_axis

def get_spectrum(f_lambda, lambda_axis, n_points=10000):
    """
    

    Parameters
    ----------
    f_lambda : function
        Interpolated function of the spectrum.
    lambda_axis : array
        Wavelength in Hz.
    n_points : int, optional
        Number of points to use for the spectrum. The default is 10000.

    Returns
    -------
    spec : array
        Spectrum in erg/cm^2/s/Hz.
    lambda_spec : array
        Wavelength in Hz.

    """
    
    # Create a new axis for the spectrum
    lambda_spec = np.linspace(min(lambda_axis), max(lambda_axis), n_points)
    
    # Evaluate the spectrum on this new axis
    spec = f_lambda(lambda_spec)
    
    return spec, lambda_spec

def get_bolometric_luminosity(f_lambda, lambda_axis, n_points=10000):
    """
    

    Parameters
    ----------
    f_lambda : function
        Interpolated function of the spectrum.
    lambda_axis : array
        Wavelength in Hz.
    n_points : int, optional
        Number of points to use for the spectrum. The default is 10000.

    Returns
    -------
    bol_lum : float
        Bolometric luminosity in erg/s.

    """
    
    # Get the spectrum and its axis
    spec, lambda_spec = get_spectrum(f_lambda, lambda_axis, n_points=n_points)
    
    # Integrate over the whole spectrum
    bol_lum = integrate.trapz(spec, x=lambda_spec)
    
    return bol_lum

def get_bolometric_temperature(bol_lum, distance):
    """
    

    Parameters
    ----------
    bol_lum : float
        Bolometric luminosity in erg/s.
    distance : float
        Distance in pc.

    Returns
    -------
    bol_temp : float
        Bolometric temperature in K.

    """
    
    # Calculate the bolometric temperature
    bol_temp = bol_lum/(distance**2 * 4 * np.pi * 6.67e-8)
    
    return bol_temp

def get_bolometric_magnitude(bol_lum, distance):
    """
    

    Parameters
    ----------
    bol_lum : float
        Bolometric luminosity in erg/s.
    distance : float
        Distance in pc.

    Returns
    -------
    bol_mag : float
        Bolometric magnitude in Vega magnitudes.

    """
    
    # Calculate the bolometric magnitude
    bol_mag = -2.5 * np.log10((bol_lum / (4 * np.pi * distance**2)) * 3.826e26) + 48.6
    
