#!/usr/bin/env python3

import os
from pathlib import Path
import sys
from typing import List, Optional, Tuple

import click
import yaml

from . import __version__ as version
from . import config
from . import utils


@click.group()
def cli():
    pass


@cli.command(help="Show the current configuration")
def show_config():
    print(yaml.dump(config.get()))


@cli.command(help="Set a configuration value")
@click.argument("key", nargs=1)
@click.argument("value", nargs=-1)
def set_config(key: str, value: List[str]):
    if len(value) == 0:
        raise click.UsageError("No value provided.")
    elif len(value) > 1:
        value = " ".join(value)
    else:
        value = value[0]
    config.set(key, value)


@cli.command(help="Unset a configuration value")
@click.argument("key", nargs=1)
def unset_config(key: str):
    config.unset(key)


@cli.command(help="List all available templates")
def list_templates():
    for template in config.list_templates():
        print(template)


@cli.command(help="Create a new project from a template")
@click.option(
    "-t",
    "--template",
    help="The name of the template to use.",
    required=True,
    type=click.Choice(config.list_templates()),
)
@click.option(
    "-n",
    "--name",
    help="The name of the project to create.",
    required=True,
)
@click.option(
    "-d",
    "--directory",
    help=(
        "The directory where the project should be created. If not specified,"
        " the current working directory is used."
    ),
    default=os.getcwd(),
    type=click.Path(exists=False),
)
@click.option(
    "-f",
    "--force",
    help="Overwrite existing files without asking.",
    is_flag=True,
)
def init(template: str, name: str, directory: str, force: bool):
    # Create the directory if it doesn't exist yet.
    directory = Path(directory).resolve()
    if not directory.is_dir():
        directory.mkdir(parents=True)
    # Get the template and check whether it exists.
    template = config.get_template(template)
    if not template.exists():
        raise click.ClickException(f"Template {template} does not exist.")
    # Check whether the target directory already exists.
    target = directory / name
    if target.exists():
        if not force:
            overwrite = click.prompt(
                f"Directory {target} already exists. Overwrite? [y/N] ",
                default="N",
            )
            if overwrite.lower().startswith("n"):
                return
        shutil.rmtree(target)
    # Copy the template into the target directory.
    shutil.copytree(template, target)
    # Rename the project name in the source code.
    rename_project(target, name)
    # Run any post-init scripts.
    run_post_init(target)


def rename_project(path: Path, name: str):
    """Rename the project name in the source code."""
    for file in path.glob("**/*.py"):
        content = file.read_text()
        content = content.replace("{{ cookiecutter.project_slug }}", name)
        file.write_text(content)


def run_post_init(path: Path):
    """Run any post-init scripts."""
    script = path / "post-init.sh"
    if script.exists():
        utils.run([script], cwd=path)

