#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1 + rng.read(2)[0]
    height = 1 + rng.read(2)[0]
    global_color_table_flag = bool(rng.read(1)[0])
    color_resolution = rng.read(3)[0] & 7
    sort_flag = bool(rng.read(1)[0])
    global_color_table_size = 2 ** ((rng.read(3)[0] & 7) + 1)
    background_color_index = rng.read(1)[0]
    pixel_aspect_ratio = rng.read(1)[0]

    if global_color_table_flag:
        for _ in range(global_color_table_size * 3):
            out.write(rng.read(1))

    # Image descriptor
    image_left_position = rng.read(2)[0]
    image_top_position = rng.read(2)[0]
    image_width = width - image_left_position
    image_height = height - image_top_position
    local_color_table_flag = bool(rng.read(1)[0])
    interlace_flag = bool(rng.read(1)[0])
    sort_flag = bool(rng.read(1)[0])
    local_color_table_size = 2 ** ((rng.read(3)[0] & 7) + 1)
    lzw_minimum_code_size = rng.read(1)[0]

    if local_color_table_flag:
        for _ in range(local_color_table_size * 3):
            out.write(rng.read(1))

    # Raster data
    def get_lzw_code():
        code = 0
        while True:
            code <<= 1
            code |= bool(rng.read(1)[0])
            yield code

    lzw_codes = get_lzw_code()
    clear_code = 1 << lzw_minimum_code_size
    end_of_information_code = clear_code + 1

    code_stack = []
    current_code_size = lzw_minimum_code_size + 1

    try:
        while True:
            code = next(lzw_codes)
            if code == clear_code:
                current_code_size = lzw_minimum_code_size + 1
                code_stack = []
            elif code == end_of_information_code:
                break
            else:
                if len(code_stack) > 0 and code_stack[-1] == clear_code:
                    code_stack = [code]
                else:
                    code_stack.append(code)

                if len(code_stack) >= current_code_size:
                    prefix = code_stack[0]
                    suffix = code_stack[-current_code_size:]
                    if prefix not in suffix:
                        out.write(struct.pack('B', prefix))
                        code_stack = suffix[:-1]
                        current_code_size += 1
                        if current_code_size == 1 << 12:
                            out.write(struct.pack('BB', clear_code, 0))
                            current_code_size = lzw_minimum_code_size + 1
    except StopIteration:
        pass

    out.write(struct.pack('BB', clear_code, 0))
