#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bool:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = struct.pack('>H', 1024)
    height = struct.pack('>H',  768)
    flags  = struct.pack('B',   0x80 | 0x40 | 0x20)
    bg     = struct.pack('B',   0xff)
    aspect = struct.pack('B',   0x00)
    out.write(width + height + flags + bg + aspect)

    # Global color table
    for _ in range(0, 256):
        r = next(rng) & 0xff
        g = next(rng) & 0xff
        b = next(rng) & 0xff
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    left      = struct.pack('<H', 0)
    top       = struct.pack('<H', 0)
    width     = struct.pack('<H', 1024)
    height    = struct.pack('<H',  768)
    flags     = struct.pack('B',   0x80 | 0x00 | 0x00)
    trans_idx = struct.pack('B',   0x00)
    out.write(left + top + width + height + flags + trans_idx)

    # LZW minimum code size
    min_code_size = struct.pack('B', 8)
    out.write(min_code_size)

    # Encode image data
    lzw_table = [None] * 256
    lzw_next  = len(lzw_table) - 1
    lzw_clear = 1 << lzw_next
    lzw_end   = lzw_clear + 1
    lzw_prev  = None
    lzw_cur   = None

    while True:

        if lzw_cur is not None and lzw_cur == lzw_end:
            out.write(bytes([lzw_cur]))

            lzw_cur = None
            continue

        if lzw_cur is None or lzw_cur >= lzw_clear:
            lzw_cur = bytes([next(rng)])[0]
            out.write(lzw_cur)

            if lzw_cur < lzw_next:
                out.write(lzw_cur)

                lzw_table.append(lzw_table[lzw_prev][:-1] + lzw_cur)
                lzw_next += 1

            elif lzw_cur == lzw_end:
                out.write(lzw_cur)

            else:
                out.write(lzw_table[lzw_prev][-1])

        else:
            out.write(lzw_table[lzw_prev][-1])

        lzw_prev = lzw_cur
        lzw_cur  = None


if