#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = rng.read(1)[0] % 256 + 1
    height = rng.read(1)[0] % 256 + 1
    background_color = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBB', width, height, background_color, pixel_aspect_ratio))

    # Global color table
    global_color_table_size = 7 + rng.read(1)[0] % 8
    global_color_table_sort_flag = rng.read(1)[0] & 1
    global_color_table_colors = [None] * (2 ** global_color_table_size)
    for i in range(len(global_color_table_colors)):
        red = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue = rng.read(1)[0] % 256
        global_color_table_colors[i] = (red, green, blue)
    out.write(struct.pack('B', global_color_table_size | (global_color_table_sort_flag << 7)))
    for rgb in global_color_table_colors:
        out.write(struct.pack('BBB', *rgb))

    # Image block
    image_left = rng.read(1)[0] % 256
    image_top = rng.read(1)[0] % 256
    image_width = rng.read(1)[0] % 256 + 1
    image_height = rng.read(1)[0] % 256 + 1
    local_color_table_flag = rng.read(1)[0] >> 7
    interlace_flag = rng.read(1)[0] & 1
    sort_flag = rng.read(1)[0] & 1
    local_color_table_size = 7 + rng.read(1)[0] % 8
    out.write(struct.pack('<HHBB', image_left, image_top, image_width, image_height))
    out.write(struct.pack('BB', local_color_table_flag | (interlace_flag << 1) | (sort_flag << 2), local_color_table_size))
    if local_color_table_flag == 1:
        local_color_table_colors = [None] * (2 ** local_color_table_size)
        for i in range(len(local_color_table_colors)):
            red = rng.read(1)[0] % 256
            green = rng.read(1)[0] % 256
            blue = rng.read(1)[0] % 256
            local_color_table_colors[i] = (red, green, blue)
        for rgb in local_color_table_colors:
            out.write(struct.pack('BBB', *rgb))
    lzw_min_code_size = rng.read(1)[0] % 256
    out.write(bytes([lzw_min_code_size]))

    # LZW data
    code_length = 2 ** lzw_min_code_size
    clear_code = 1 << lzw_min_code_size
    end_of_information_code = clear_code + 1
    next_code = None
    dictionary = {}
    dictionary_keys = []
    def add_to_dictionary(*args):
        nonlocal next_code
        key = bytes(args)
        if not key in dictionary:
            dictionary[key] = len(dictionary_keys)
            dictionary_keys.append(key)
            next_code = len(dictionary_keys) - 1
        else:
            next_code = dictionary[key]
    add_to_dictionary()
    add_to_dictionary()
    output_buffer = bytearray()
    while True:
        bit = rng.read(1)[0] & 1
        output_buffer += b'\x00\xff'[bit]

        if next_code is None or next_code >= clear_code:
            input_byte = rng.read(1)[0]
            if input_byte == 0:
                break

            if next_code is None and input_byte > clear_code:
                raise Exception("Invalid LZW stream")

            if next_code is not None:
                output_buffer[-1] = dictionary_keys[next_code][-1:]
                add_to_dictionary(output_buffer[-1])

            if next_code < clear_code:
                output_buffer.extend(dictionary_keys[input_byte][:])
                add_to_dictionary(*(dictionary_keys[input_byte]))
            elif next_code == clear_code:
                pass
            else:
                output_buffer.extend(dictionary_keys[next_code - clear_code][:-1])
                add_to_dictionary(*(dictionary_keys[next_code - clear_code][:-1]))
        else:
            output_buffer.extend(dictionary_keys[next_code][:])
            add_to_dictionary(*(dictionary_keys[next_code]))

        if len(output_buffer) >= code_length:
            code_length <<= 1
            if code_length > 4096:
                raise Exception("LZW code length overflowed")

    # Trailer
    out.write(b';')


if