#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    width = rng.read(2)
    height = rng.read(2)
    # TODO: This should probably use the same algorithm as the one used by
    #       generate_random_png() instead of just copying it from there
    palette = [
        0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xF5, 0xF5, 0xF5, 0xFA, 0xFA, 0xFA,
        0xFD, 0xFD, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF]
    out.write(width + height + bytes(palette))

    # TODO: This could probably be improved with some more sophisticated logic
    #       than just repeating the same byte over and over again
    color_table_size = (len(palette) // 3) - 1
    for i in range(color_table_size):
        out.write(bytes([i]))

    # TODO: This should probably use the same algorithm as the one used by
    #       generate_random_png() instead of just copying it from there
    lzw_mincodesize = 2
    out.write(lzw_mincodesize.to_bytes(1, 'little'))

    # TODO: This should probably use the same algorithm as the one used by
    #       generate_random_png() instead of just copying it from there
    image_descriptor = b'\x2C\x00\x00\x00' \
                       + width \
                       + height \
                       + b'\x00\x00'
    out.write(image_descriptor)

    # TODO: This should probably use the same algorithm as the one used by
    #       generate_random_png() instead of just copying it from there
    local_color_table_flag = 1
    interlace_flag = 0
    sort_flag = 0
    reserved_bits = 0b000
    local_color_table_size = len(palette) // 3 - 1
    table_sorted_flag = int((sort_flag << 3) |
                            (reserved_bits << 2) |
                            (local_color_table_size << 4) |
                            (local_color_table_flag << 7))
    out.write(table_sorted_flag.to_bytes(1, 'little'))

    # TODO: This should probably use the same algorithm as the one used by
    #       generate_random_png() instead of just copying it from there
    code_size = lzw_mincodesize + 1
    clear_code = 1 << code_size
    end_of_information_code = clear_code + 1
    max_code_size = 12
    min_code_size = 2
    initial_code_size = min_code_size + 1
    initial_code = 1 << initial_code_size
    codes = []
    next_code = initial_code
    output_buffer = b''
    while True:
        if next_code == end_of_information_code or \
           next_code == clear_code:
            if next_code!= end_of_information_code:
                out.write(output_buffer[0].to_bytes(1, 'little'))

            if next_code == clear_code:
                codes = list(range(clear_code,
                                  end_of_information_code))

                # The first two codes are special; they're not actually LZW
                # codes, so we have to hard-code them here
                codes[0] = 0xFFFF
                codes[1] = 0xFFFE

                # We also need to reset our buffer since we've cleared all its
                # contents
                output_buffer = b''

            # Reset the next code size based on how many codes we already have
            # NOTE: This will never decrease the code size once it's set above
            new_code_size = min(max_code_size,
                                len(codes).bit_length())
            if new_code_size < code_size:
                out.write(new_code_size.to_bytes(1, 'little'))
                code_size = new_code_size

            # If this was the last code, then we can stop writing codes now
            if next_code == end_of_information_code:
                break

        # Get the next input byte
        input_byte = rng.read(1)
        if len(input_byte) == 0:
            raise Exception('ran out of input data')

        # Add it to