#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0]))

    # Global color table
    for i in range(768):
        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

    # Image descriptor
    left   = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16)
    top    = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16)
    width  = int.from_bytes(rng.read(2), byteorder='little') % ((1 << 16) - left) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % ((1 << 16) - top) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lct    = bool((flags[0] & 0b10000000) >> 7)
    interlaced = bool((flags[0] & 0b01000000) >> 6)
    sorted      = bool((flags[0] & 0b00100000) >> 5)
    reserved    = bool((flags[0] & 0b00010000) >> 4)
    lctSize     = 2 ** ((flags[0] & 0b00001111))
    out.write(struct.pack('<HHHHB', left, top, width, height, flags[0]))

    # Local color table
    if lct:
        for i in range(3 * lctSize**2):
            out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

    # Table based image data
    while True:
        blockType = int.from_bytes(rng.read(1), byteorder='little')

        if blockType == 0x2C:
            break

        elif blockType == 0x21:
            while True:
                extType = int.from_bytes(rng.read(1), byteorder='little')

                if extType == 0xFF:
                    break

                else:
                    length = int.from_bytes(rng.read(1), byteorder='little')

                    for _ in range(length):
                        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

            continue

        else:
            raise Exception('Invalid block type.')

    # Plain text extension
    out.write(bytes([0x21, 0xF9, 0x04, 0x05, 0x00, 0x00, 0x00, 0x00, 0x2C, 0x00, 0x00, 0x00, 0x00]))

    # Application extension
    out.write(bytes([
        0x21, 0xFF, 0x0B,

        0x4D, 0x45, 0x54, 0x41, 0x2D, 0x49, 0x4E, 0x46, 0x2F, 0x67, 0x69, 0x66,

        0x38, 0x39, 0x61,

        0x03, 0x01, 0x00, 0x00, 0x00
    ]))

    # Graphics control extension
    out.write(bytes([
        0x21, 0xF9, 0x04,

        0x08,

        0x00, 0x00,

        0x00, 0x00,

        0x00
    ]))

    # Image data
    while True:
        blockType = int.from_bytes(rng.read(1), byteorder='little')

        if blockType == 0x2C:
            break

        elif blockType == 0x21:
            while True:
                extType = int.from_bytes(rng.read(1), byteorder='little')

                if extType == 0xFF:
                    break

                else:
                    length = int.from_bytes(rng.read(1), byteorder='little')

                    for _ in range(length):
                        out.write(bytes([int.from_bytes(rng.read(1), byteorder='little')]))

            continue

        else:
            raise Exception('Invalid block type.')


    # Trailer
    out.write(b';')

