#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = rng.read(1)[0] % 256 + 1
    height = rng.read(1)[0] % 256 + 1
    background_color = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBB', width, height, background_color, pixel_aspect_ratio))

    # Global color table
    global_color_table_size = 7 + rng.read(1)[0] % 8
    global_color_table_sort_flag = rng.read(1)[0] & 1
    global_color_table_colors = [None] * (2 ** global_color_table_size)
    for i in range(len(global_color_table_colors)):
        red = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue = rng.read(1)[0] % 256
        global_color_table_colors[i] = bytes([red, green, blue])
    if global_color_table_sort_flag == 1:
        global_color_table_colors.sort()
    for color in global_color_table_colors:
        out.write(color)

    # Image descriptor
    left = rng.read(1)[0] % width
    top = rng.read(1)[0] % height
    image_width = rng.read(1)[0] % (width - left) + 1
    image_height = rng.read(1)[0] % (height - top) + 1
    packed = rng.read(1)[0]
    local_color_table_flag = (packed >> 7) & 1
    interlace_flag = (packed >> 6) & 1
    sort_flag = (packed >> 5) & 1
    reserved = (packed >> 4) & 1
    lct_size = 2 << ((packed & 0x0F) + 1)
    lct_entries = []
    for _ in range(lct_size):
        red = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue = rng.read(1)[0] % 256
        lct_entries.append(bytes([red, green, blue]))
    if sort_flag == 1:
        lct_entries.sort()
    for entry in lct_entries:
        out.write(entry)
    code_size = rng.read(1)[0] % 256 + 1
    out.write(code_size.to_bytes(1, byteorder='little'))

    # LZW minimum code size
    min_code_size = rng.read(1)[0] % 256 + 1
    out.write(min_code_size.to_bytes(1, byteorder='little'))

    # Image data
    lzw_dictionary = {}
    lzw_dictionary_size = 2 ** min_code_size
    lzw_dictionary_next_index = len(lzw_dictionary)
    lzw_buffer = b''
    lzw_buffer_index = 0
    lzw_clear_code = 1 << min_code_size
    lzw_end_of_information_code = lzw_clear_code + 1
    lzw_next_code = lzw_end_of_information_code + 1
    lzw_current_code = None
    while True:
        next_byte = rng.read(1)[0]

        # Clear buffer and dictionary
        if next_byte == 0x2C or next_byte == 0x21:
            lzw_buffer = b''
            lzw_buffer_index = 0
            lzw_dictionary = {
                lzw_clear_code : b'',
                lzw_end_of_information_code : b'\xFF\xFF'
            }
            lzw_dictionary_size = 2 ** min_code_size
            lzw_dictionary_next_index = len(lzw_dictionary)
            lzw_next_code = lzw_end_of_information_code + 1

            # End of information block
            if next_byte == 0x21:
                break
        
        # Add to buffer
        elif next_byte >= 0x30 and next_byte <= 0x3F:
            lzw_buffer += bytes([next_byte])
            lzw_buffer_index += 1
            
            # Generate new codes from buffer contents
            while lzw_buffer_index > 0:
                # Get longest prefix that exists in the dictionary
                longest_prefix = b''
                longest_prefix_length = 0
                for i in reversed(range(lzw_buffer_index)):
                    test_prefix = lzw_buffer[:i+1]
                    if test_prefix in lzw_dictionary:
                        longest_prefix = test_prefix
                        longest_prefix_length = i + 1

                # If no such prefix exists, add current character as-is
                if longest_prefix_length == 0:
                    lzw_dictionary[lzw_next_code] = bytes([lzw_buffer[0]])
                    lzw_next_code += 1

                    # Check if we need to increase our dictionary size
                    if lzw_next_code == 2 ** max(min_code_size + 1, 12):
                        min_code_size += 1
                    
                    # Reset buffer
                    lzw_buffer = b''
                    lzw_buffer_index = 0
                
               