#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = rng.read(1)[0] % 256 + 1
    height = rng.read(1)[0] % 256 + 1
    background_color = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBB', width, height, background_color, pixel_aspect_ratio))

    # Global color table
    global_color_table_size = 7 + rng.read(1)[0] % 8
    global_color_table_sort_flag = rng.read(1)[0] & 1
    global_color_table_colors = [None] * (2 ** global_color_table_size)
    for i in range(len(global_color_table_colors)):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        global_color_table_colors[i] = bytes([r,g,b])
    if global_color_table_sort_flag == 1:
        global_color_table_colors.sort()
    for c in global_color_table_colors:
        out.write(c)

    # Image descriptor
    image_left = rng.read(1)[0] % 256
    image_top = rng.read(1)[0] % 256
    image_width = rng.read(1)[0] % 256 + 1
    image_height = rng.read(1)[0] % 256 + 1
    local_color_table_flag = rng.read(1)[0] & 1
    interlace_flag = rng.read(1)[0] & 1
    sort_flag = rng.read(1)[0] & 1
    reserved = rng.read(1)[0]
    lct_size = 7 + rng.read(1)[0] % 8
    lct_colors = []
    for _ in range(lct_size):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        lct_colors.append(bytes([r,g,b]))
    if sort_flag == 1:
        lct_colors.sort()
    for c in lct_colors:
        out.write(c)
    out.write(struct.pack('<HHBBB', image_left, image_top, image_width, image_height, local_color_table_flag, interlace_flag, sort_flag, reserved))

    # LZW minimum code size
    min_code_size = rng.read(1)[0] % 256
    out.write(min_code_size.to_bytes(1, byteorder='big'))

    # Image data
    num_pixels = image_width * image_height
    pixels = set()
    while len(pixels) < num_pixels:
        pixels.add(rng.read(1)[0] % 256)
    lzw_data = bytes(sorted(list(pixels)))
    lzw_data += bytes([0x00, 0xff])
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_data = bytes([0x00]) + lzw_data
    lzw_