#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    width = rng.read(1)[0] % 40 + 1
    height = rng.read(1)[0] % 40 + 1
    flags = bytes([
        0x80 | rng.read(1)[0], # Global color table flag
        0x07 & rng.read(1)[0]]) # Color resolution and sort order
    bg = bytes([rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]])
    aspect = bytes([rng.read(1)[0], rng.read(1)[0]])
    out.write(width.to_bytes(1, 'little'))
    out.write(height.to_bytes(1, 'little'))
    out.write(flags)
    out.write(bg)
    out.write(aspect)

    # Global color table
    for _ in range(0xFF):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        packed = struct.pack('BBB', r, g, b)
        out.write(packed)

    # Image descriptor
    left = 0
    top = 0
    width -= 1
    height -= 1
    flags = bytes([
        0x80 | rng.read(1)[0], # Local color table flag
        0x07 & rng.read(1)[0]]) # Color resolution and sort order
    lct = bytes([rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]])
    out.write(left.to_bytes(2, 'little'))
    out.write(top.to_bytes(2, 'little'))
    out.write(width.to_bytes(2, 'little'))
    out.write(height.to_bytes(2, 'little'))
    out.write(flags)
    out.write(lct)

    # LZW minimum code size
    mincodesize = rng.read(1)[0] % 0xF + 2
    out.write(mincodesize.to_bytes(1, 'little'))

    # Data sub-block
    data = []
    while True:
        byte = rng.read(1)[0]

        if byte == 0x00:
            break
        
        elif byte <= 0x0F:
            data += [byte] * (1 + byte)

        else:
            count = ((byte - 0x10) << 8) | rng.read(1)[0]

            if count > len(data):
                raise Exception("Invalid LZW length")
            
            for i in reversed(range(count)):
                data[i] = data[i - count]

            data += list(reversed(data[-count:]))
    
    # Write compressed data
    def write_compressed():
        prev = None
        for d in data:
            if prev is not None:
                out.write((prev << 4).to_bytes(1, 'little'))
                out.write(((d >> 4) | 0x80).to_bytes(1, 'little'))
                out.write((((d & 0xF) << 4) | 0x80).to_bytes(1, 'little'))

                prev = (((d >> 4) | 0x80) << 4) | (((d & 0xF) << 4) | 0x80)

            else:
                out.write((d >> 4).to_bytes(1, 'little'))
                out.write(((d & 0xF) << 4).to_bytes(1, 'little'))
                
                prev = d

    write_compressed()

    # Trailer
    out.write(b'\x00\x3B')


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open(sys.argv[1], 'wb') as out:
            generate_random_gif(rng, out)
            out.write(struct.pack('B', index))