#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):

    # Write header
    out.write(b'GIF89a')

    # Write logical screen descriptor
    width = 1 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
    height = 1 + int.from_bytes(rng.read(1), byteorder='big', signed=False) % 256
    packed = b'\x07\x00\x00' + \
             width.to_bytes(2, byteorder='little', signed=False) + \
             height.to_bytes(2, byteorder='little', signed=False)
    out.write(packed)

    # Write global color table
    num_colors = min(width * height // 4, 256)
    for i in range(num_colors):
        r = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        g = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff
        b = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0xff

        packed = bytes([r,g,b])
        if len(packed) < 3:
            packed += b'\x00'*(3-len(packed))
        out.write(packed)

    # Write local color table
    if num_colors > 256:
        raise Exception("Too many colors")
    elif num_colors == 256:
        out.write(b'\x00'*3*256)

    # Write image data
    lzw_mincodesize = 2
    out.write(lzw_mincodesize.to_bytes(1, byteorder='little'))
    lzw_clearcode = 1 << lzw_mincodesize
    lzw_endofinfo = lzw_clearcode + 1
    lzw_endofdata = lzw_endofinfo + 1
    lzw_maxcodesize = 12
    lzw_bitmask = ((1<<lzw_maxcodesize)-1)<<lzw_mincodesize
    lzw_nextcode = lzw_endofinfo+1
    lzw_prevcode = None
    lzw_curcode = None
    lzw_curbits = 0
    lzw_curbyte = 0
    lzw_buf = []
    lzw_bufpos = 0
    lzw_bufsize = 2**lzw_mincodesize - 1

    def write_lzw():
        nonlocal lzw_prevcode, lzw_curcode, lzw_curbits, lzw_curbyte, lzw_buf, lzw_bufpos, lzw_bufsize

        while True:
            if not lzw_prevcode or lzw_curcode!= lzw_prevcode:
                break

            lzw_buf[lzw_bufpos] = lzw_prevcode >> 8
            lzw_bufpos += 1
            lzw_buf[lzw_bufpos] = lzw_prevcode & 0xff
            lzw_bufpos += 1

            lzw_curbits -= lzw_prevcode.bit_length()
            lzw_curbyte |= lzw_prevcode << lzw_curbits
            lzw_curcode = None

            if lzw_curbits <= 0:
                lzw_buf[lzw_bufpos] = lzw_curbyte
                lzw_bufpos += 1

                lzw_curbits += 8
                lzw_curbyte = 0

            if lzw_bufpos >= lzw_bufsize:
                out.write(struct.pack('B'*lzw_bufsize, *lzw_buf))
                lzw_bufpos = 0

    while True:
        cur_blocktype = int.from_bytes(rng.read(1), byteorder='big', signed=False) & 0x07
        if cur_blocktype == 0x00:
            # Uncompressed block
            size = int.from_bytes(rng.read(1), byteorder='big', signed=False)
            while size > 0:
                chunk = rng.read(min(size, 255))
                out.write(chunk)
                size -= len(chunk)
        else:
            # LZW compressed block
            if cur_blocktype == 0x01:
                # Clear code
                lzw_curcode = lzw_clearcode
                lzw_curbits = lzw_mincodesize