#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    width = rng.read(1)[0] % 40 + 1
    height = rng.read(1)[0] % 40 + 1
    flags = bytes([
        0x80 | rng.read(1)[0], # Global color table flag
        0x07 & rng.read(1)[0]]) # Color resolution and sort order
    bg = bytes([rng.read(1)[0], rng.read(1)[0], rng.read(1)[0]])
    aspect = bytes([rng.read(1)[0], rng.read(1)[0]])
    out.write(width.to_bytes(1, 'little'))
    out.write(height.to_bytes(1, 'little'))
    out.write(flags)
    out.write(bg)
    out.write(aspect)

    # Global color table
    for _ in range(0xFF):
        r = rng.read(1)[0] % 256
        g = rng.read(1)[0] % 256
        b = rng.read(1)[0] % 256
        packed = struct.pack('BBB', r, g, b)
        out.write(packed)

    # Image data
    lzw_mincodesize = rng.read(1)[0] % 0x10 + 1
    out.write(lzw_mincodesize.to_bytes(1, 'little'))
    out.write(b'\x00\x00') # LZW minimum code size
    out.write((width * height).to_bytes(2, 'little'))
    out.write(b'\x00') # Background index
    out.write(b'\x00') # Pixel aspect ratio

    # Frames
    for frame in range(0x10):
        # Descriptor
        left = rng.read(1)[0] % width
        top = rng.read(1)[0] % height
        width = rng.read(1)[0] % (width - left) + 1
        height = rng.read(1)[0] % (height - top) + 1
        flags = bytes([
            0xC0 | rng.read(1)[0], # Local color table flag
            0x07 & rng.read(1)[0]]) # Color resolution and sort order
        delay = rng.read(1)[0] % 0x100
        trans = rng.read(1)[0] % 0x100
        out.write(left.to_bytes(1, 'little'))
        out.write(top.to_bytes(1, 'little'))
        out.write(width.to_bytes(1, 'little'))
        out.write(height.to_bytes(1, 'little'))
        out.write(flags)
        out.write(delay.to_bytes(1, 'little'))
        out.write(trans.to_bytes(1, 'little'))

        # Local color table
        for _ in range(0xFF):
            r = rng.read(1)[0] % 256
            g = rng.read(1)[0] % 256
            b = rng.read(1)[0] % 256
            packed = struct.pack('BBB', r, g, b)
            out.write(packed)

        # LZW compressed image data
        lzw_clearcode = 1 << lzw_mincodesize
        lzw_endofinfo = lzw_clearcode + 1
        lzw_endofdata = lzw_endofinfo + 1
        lzw_nextcode = lzw_endofdata + 1
        lzw_nextbits = 9
        lzw_nextmask = (1 << lzw_nextbits) - 1
        lzw_curcode = None
        lzw_curbyte = None
        lzw_curbit = 0
        lzw_prevbuf = []
        lzw_prevlen = 0
        lzw_prevpos = 0
        lzw_prevcode = None
        lzw_prevbyte = None
        lzw_prevbit = 0
        lzw_prevcodesize = lzw_mincodesize + 1
        lzw_prevmaxcode = (1 << lzw_prevcodesize) + lzw_endofdata
        lzw_prevfirstcode = lzw_endofdata + 1
        lzw_prevlastcode = lzw_prevfirstcode + ((1 << lzw_prevcodesize) - 1)
        lzw_prevtable = [None] * lzw_prevmaxcode
        lzw_prevtable[lzw_clearcode] = ''
        lzw_prevtable[lzw_endofinfo] = None
        lzw_prevtable[lzw_endofdata] = None
        while True:
            if not lzw_curcode or lzw_curbit == 0:
                buf = bytearray()
                while len(buf) < 1:
                    buf += rng.read(1)
                lzw_curbyte = buf[0]
                del buf
                lzw_curbit = 8

            curbit = lzw_curbit - lzw_prevcodesize
            curmask = (1 << curbit) - 1
            curcode = (lzw_curbyte >> curbit) & curmask
            lzw_curbit -= lzw_prevcodesize
            if curcode >= lzw_prevmaxcode:
                raise Exception("Invalid LZW code")

            if curcode == lzw_endofdata:
                break

            if curcode == lzw_endofinfo:
                continue

            if not lzw_prevcode:
                prevcode = curcode
                continue

            if curcode == lzw_clearcode:
                lzw_prevcodesize = lzw_mincodesize + 1
                lzw_prevmaxcode = (