#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    height = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    out.write(struct.pack('<HHBBB', width, height, 0xF0, 0x00, 0x00))

    # Write the global color table
    for i in range(768):
        out.write(bytes([(i >> ((~i + 1) % 8)) & 0xFF]))

    # Write the image descriptor
    left = 0
    top = 0
    width = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    height = rng.read(1)[0] & 0x7F | (rng.read(1)[0] & 0x7F) << 8
    flags = 0x8C
    out.write(struct.pack('<HHHHB', left, top, width, height, flags))

    # Write the local color table
    for i in range(256):
        out.write(bytes([(i >> ((~i + 1) % 8)) & 0xFF]))

    # Write the image data
    for y in range(height):
        lzw_mincodesize = rng.read(1)[0]
        out.write(bytes([lzw_mincodesize]))

        codesize = 1 << lzw_mincodesize
        clear_code = 1 << lzw_mincodesize
        end_of_information_code = clear_code + 1
        next_code = None
        prev_code = None
        codes = []
        for x in range(width):
            if not next_code or len(codes) == codesize and next_code!= end_of_information_code:
                # We need more bits!
                while True:
                    byte = rng.read(1)[0]

                    for bit in reversed(range(8)):
                        code = (byte >> bit) & 1

                        if next_code:
                            code |= next_code << 1

                            if code < len(codes):
                                break
                        else:
                            assert code == 0

                        codes.append(code)

                    if next_code:
                        break

                next_code = codes[0]

            if next_code == clear_code:
                # Clear all of our state
                next_code = None
                prev_code = None
                codes = [end_of_information_code]
                continue

            if next_code == end_of_information_code:
                # End of stream
                break

            if prev_code is None:
                # Just write this code as-is
                out.write(bytes([next_code]))
            elif next_code >= len(codes):
                # This shouldn't happen because we're always reading from the same place
                raise Exception('Invalid LZW code!')
            else:
                # Emit a sequence of codes that represent the previous one
                seq = codes[:len(codes)-next_code+prev_code]
                out.write(bytes([seq[-1]]))
                del codes[:len(codes)-next_code+prev_code]

            prev_code = next_code
            next_code = None

        # Write the trailer
    out.write(b';')
