#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, flags[0], bg[0], aspect[0]))

    # Global color table
    global_color_table_size = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    if global_color_table_size!= b'\x00':
        raise Exception('Global color table size must be zero!')
    out.write(global_color_table_size)

    # Background color index
    background_color_index = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(background_color_index)

    # Pixel aspect ratio
    pixel_aspect_ratio = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(pixel_aspect_ratio)

    # Image descriptor
    image_left   = int.from_bytes(rng.read(2), byteorder='little') % (width - 1)
    image_top    = int.from_bytes(rng.read(2), byteorder='little') % (height - 1)
    image_width  = int.from_bytes(rng.read(2), byteorder='little') % ((width - image_left) + 1)
    image_height = int.from_bytes(rng.read(2), byteorder='little') % ((height - image_top) + 1)
    flags        = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    local_color_table_flag = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    interlace_flag         = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    sort_flag              = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    reserved               = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lct_size               = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBBBBB',
                         image_left, image_top, image_width, image_height,
                         flags[0], local_color_table_flag[0], interlace_flag[0], sort_flag[0],
                         reserved[0], lct_size[0]
                        ))

    # Local color table
    for i in range(lct_size):
        r = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        g = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        b = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        out.write(r + g + b)

    # Table based image data
    while True:
        block_type = bytes([int.from_bytes(rng.read(1), byteorder='little')])

        if block_type == b'\x2C':
            # Image descriptor
            left   = int.from_bytes(rng.read(2), byteorder='little') % (image_width - 1)
            top    = int.from_bytes(rng.read(2), byteorder='little') % (image_height - 1)
            width  = int.from_bytes(rng.read(2), byteorder='little') % ((image_width - left) + 1)
            height = int.from_bytes(rng.read(2), byteorder='little') % ((image_height - top) + 1)
            flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])

            # LZW minimum code size
            lzw_min_code_size = bytes([int.from_bytes(rng.read(1), byteorder='little')])
            out.write(block_type +
                      struct.pack('<HHBBBBB',
                                  left, top, width, height,
                                  flags[0], lzw_min_code_size[0]
                                 )
                     )

            # Data sub-blocks
            while True:
                block_size = bytes([int.from_bytes(rng.read(1), byteorder='little')])

                if block_size == b'\x00':
                    break

                out.write(block_size)
                out.write(rng.read(block_size))

            out.write(b'\x00')

        elif block_type == b'\x21':
            # Extension
            label = bytes([int.from_bytes(rng.read(1), byteorder='little')])

            if label == b'\xF9':
                # Graphics control extension
                length       = bytes([int.from