#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), 'little') % (1 << 10) + 1
    height = int.from_bytes(rng.read(2), 'little') % (1 << 10) + 1
    flags  = bytes([int.from_bytes(rng.read(1), 'little')])
    bg     = bytes([int.from_bytes(rng.read(1), 'little')])
    aspect = bytes([int.from_bytes(rng.read(1), 'little')])
    out.write(struct.pack('<HHBBB', width, height, *flags))
    out.write(bg)
    out.write(aspect)

    # Global color table
    colors = [b'\x00\x00\x00']
    for _ in range((width*height)//64):
        r = bytes([int.from_bytes(rng.read(1), 'little')])
        g = bytes([int.from_bytes(rng.read(1), 'little')])
        b = bytes([int.from_bytes(rng.read(1), 'little')])
        colors += [r,g,b]
    out.write(b''.join(colors))

    # Image block
    image_descriptor = struct.pack('<BBH', 0x2C, 0, 0)
    out.write(image_descriptor)

    # Local color table
    local_color_table = []
    if len(colors) > 1:
        local_color_table = [
            colors[i] for i in sorted(
                list(
                    set(
                        [
                            int.from_bytes(
                                rng.read(1), 
                                'little'
                            ) % len(colors) 
                            for _ in range(len(colors)-1)
                        ]
                    )
                ), 
                reverse=True
            )
        ]
    out.write(b''.join(local_color_table))

    # LZW minimum code size
    lzw_min_code_size = bytes([int.from_bytes(rng.read(1), 'little')])
    out.write(lzw_min_code_size)

    # Data sub-block
    data_sub_block = bytearray()
    while True:
        # Get next bit from RNG and add it to our data stream
        bit = bytes([int.from_bytes(rng.read(1), 'little')])
        data_sub_block += bit

        # If we have more than one byte of data, check if this is the end
        if len(data_sub_block) >= 255:
            # Check if last two bits are equal
            if data_sub_block[-2:] == b'\xff\xff': break

            # Otherwise, write current byte as is and continue
            else:
                out.write(bytes([len(data_sub_block)]))
                out.write(data_sub_block[:255])
                del data_sub_block[:255]
        
        # If we only have one byte of data, check if this is the end
        elif len(data_sub_block) == 1: 
            # Check if last bit is equal
            if data_sub_block[-1:] == b'\xff': break
            
            # Otherwise, write current byte as is and continue
            else:
                out.write(bytes([len(data_sub_block)]))
                out.write(data_sub_block[:1])
                del data_sub_block[:1]
    
    # Write final byte with length zero
    out.write(b'\x00')