#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    signature = b'GIF89a'
    version = b'\x00\x00'
    width = rng.read(1)[0] % 400 + 1
    height = rng.read(1)[0] % 400 + 1
    flags = b'\x87\x00'
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = b'\x00'
    out.write(signature)
    out.write(version)
    out.write(struct.pack('<HHB', width, height, flags[0]))
    out.write(background_color_index.to_bytes(1, byteorder='little'))
    out.write(pixel_aspect_ratio)

    # Logical screen descriptor
    global_color_table_flag = rng.read(1)[0] % 2
    interlace_flag = rng.read(1)[0] % 2
    sort_flag = rng.read(1)[0] % 2
    color_resolution = rng.read(1)[0] % 3
    global_color_table_size = 7 + rng.read(1)[0] % 8
    background_color_index = rng.read(1)[0] % 256
    pixel_aspect_ratio = b'\x00'
    out.write(struct.pack('BBBBBBB', global_color_table_flag | (interlace_flag << 1) | (sort_flag << 2), color_resolution, global_color_table_size, background_color_index, pixel_aspect_ratio[0], pixel_aspect_ratio[1]))

    # Global color table
    if global_color_table_flag == 1:
        global_color_table_colors = [None] * (2 ** global_color_table_size)
        for i in range(len(global_color_table_colors)):
            red = rng.read(1)[0] % 256
            green = rng.read(1)[0] % 256
            blue = rng.read(1)[0] % 256
            global_color_table_colors[i] = (red, green, blue)
        for rgb in global_color_table_colors:
            out.write(struct.pack('BBB', *rgb))
    
    # Image descriptor
    image_left = rng.read(1)[0] % 400
    image_top = rng.read(1)[0] % 400
    image_width = rng.read(1)[0] % 400 + 1
    image_height = rng.read(1)[0] % 400 + 1
    local_color_table_flag = rng.read(1)[0] % 2
    local_color_table_size = 7 + rng.read(1)[0] % 8
    out.write(struct.pack('<HHBB', image_left, image_top, image_width, image_height))
    out.write(struct.pack('BB', local_color_table_flag | (interlace_flag << 1) | (sort_flag << 2), local_color_table_size))
    if local_color_table_flag == 1:
        local_color_table_colors = [None] * (2 ** local_color_table_size)
        for i in range(len(local_color_table_colors)):
            red = rng.read(1)[0] % 256
            green = rng.read(1)[0] % 256
            blue = rng.read(1)[0] % 256
            local_color_table_colors[i] = (red, green, blue)
        for rgb in local_color_table_colors:
            out.write(struct.pack('BBB', *rgb))

    # LZW minimum code size
    lzw_minimum_code_size = rng.read(1)[0] % 128
    out.write(struct.pack('B', lzw_minimum_code_size))

    # Pixel data
    for y in range(height):
        for x in range(width):
            index = rng.read(1)[0] % 256
            out.write(struct.pack('B', index))