#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = rng.read(2)
    height = rng.read(2)
    packed = (width + height).to_bytes(2, byteorder='little', signed=False)
    out.write(packed)
    out.write(b'\xf7\xff')

    # Write the global color table
    palette = [None] * 256
    for i in range(256):
        r = rng.read(1)[0] & 0xf8
        g = rng.read(1)[0] & 0xfc
        b = rng.read(1)[0] & 0xf8

        palette[i] = ((r << 8) | (g << 3) | (b >> 3))

    for p in palette:
        out.write(struct.pack('>H', p))

    # Write the image data
    num_frames = rng.read(1)[0] % 4
    for frame in range(num_frames):
        # Write the local color table
        for i in range(256):
            r = rng.read(1)[0] & 0xf8
            g = rng.read(1)[0] & 0xfc
            b = rng.read(1)[0] & 0xf8

            palette[i] = ((r << 8) | (g << 3) | (b >> 3))

        for p in palette:
            out.write(struct.pack('>H', p))

        # Write the LZW minimum size
        minsize = rng.read(1)[0] % 9

        # Write the first code
        out.write(bytes([minsize + 1]))

        # Write the rest of the codes
        while True:
            code = rng.read(1)[0]
            if code == 0:
                break
            elif code <= minsize:
                out.write(bytes([code]))
            else:
                out.write(bytes([(code - minsize)]))

        # Write the end-of-information code
        while True:
            code = rng.read(1)[0]
            if code == 0:
                break
            else:
                out.write(bytes([code]))

    # Write the trailer
    out.write(b';')
