#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom) and
# returns its size in bytes.
def generate_gif(out: BinaryIO, rng):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = 1024 + int.from_bytes(rng(2), byteorder='little', signed=False) % 512
    height = 768 + int.from_bytes(rng(2), byteorder='little', signed=False) % 512
    flags = b'\x80\x00' if width > height else b'\x00\x00'
    bgColorIndex = int.from_bytes(rng(1), byteorder='little', signed=False) & 0xF
    pixelAspectRatio = b'\x00'
    out.write(struct.pack('<HHBBB', width, height, flags[0], flags[1], bgColorIndex))
    out.write(pixelAspectRatio)

    # Global color table
    globalColorTableSize = min((width * height // 64) + 1, 256)
    for i in range(globalColorTableSize):
        r = int.from_bytes(rng(1), byteorder='little', signed=False) & 0xFF
        g = int.from_bytes(rng(1), byteorder='little', signed=False) & 0xFF
        b = int.from_bytes(rng(1), byteorder='little', signed=False) & 0xFF
        out.write(struct.pack('BBB', r, g, b))

    # Image descriptor
    leftPos = 0
    topPos = 0
    imageWidth = width - leftPos
    imageHeight = height - topPos
    flags = b'\x80\x00'
    localColorTableFlag = b'\x00'
    interlaceFlag = b'\x00'
    sortFlag = b'\x00'
    reserved = b'\x00\x00'
    lctSize = b'\x00'
    out.write(struct.pack('<HHHHBBBBB', leftPos, topPos, imageWidth, imageHeight, flags[0], flags[1], localColorTableFlag[0], localColorTableFlag[1]))
    out.write(interlaceFlag)
    out.write(sortFlag)
    out.write(reserved)
    out.write(lctSize)

    # Local color table
    for i in range(localColorTableSize):
        r = int.from_bytes(rng(1), byteorder='little', signed=False) & 0xFF
        g = int.from_bytes(rng(1), byteorder='little', signed=False) & 0xFF
        b = int.from_bytes(rng(1), byteorder='little', signed=False) & 0xFF
        out.write(struct.pack('BBB', r, g, b))

    # LZW minimum code size
    lzwMinCodeSize = max(int.from_bytes(rng(1), byteorder='little', signed=False) & 0xFF, 2)
    out.write(lzwMinCodeSize.to_bytes(1, 'little'))

    # Pixel data
    pixels = []
    while len(pixels) < width * height:
        n = int.from_bytes(rng(1), byteorder='little', signed=False) & 0xFF

        if n == 0xFF:
            continue

        elif n <= 0x8F:
            pixels += [n]

        elif n >= 0xC0:
            count = ((n - 0xC0) << 8) | int.from_bytes(rng(1), byteorder='little', signed=False) & 0xFF
            pixels += list(reversed([pixels[-i-1] for i in range(count)]))

        else:
            raise Exception("Invalid LZW code")

    assert len(pixels) == width * height
    out.write(bytearray(pixels))

    return out.tell()


if