#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> bool:

    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width  = struct.pack('>H', 1024)
    height = struct.pack('>H',  768)
    flags  = struct.pack('B',   0x80 | 0x40 | 0x20)
    bg     = struct.pack('B',   0xff)
    aspect = struct.pack('B',   0x00)
    out.write(width + height + flags + bg + aspect)

    # Global color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f

        out.write(struct.pack('<BBB', r, g, b))

    # Image descriptor
    left      = struct.pack('>H',      0)
    top       = struct.pack('>H',      0)
    width     = struct.pack('>H', 1024)
    height    = struct.pack('>H',  768)
    flags     = struct.pack('B',        0x80)
    localsize = struct.pack('B',         0x00)
    out.write(left + top + width + height + flags + localsize)

    # Local color table
    for i in range(0, 256):
        r = ord(rng.read(1)) & 0x3f
        g = ord(rng.read(1)) & 0x3f
        b = ord(rng.read(1)) & 0x3f

        out.write(struct.pack('<BBB', r, g, b))

    # LZW minimum code size
    mincodesize = struct.pack('B', 0x08)
    out.write(mincodesize)


    # Data sub-block
    data = bytearray()
    while len(data)!= 1024 * 768 // 2:
        if rng.read(1) == b'\x00':
            continue

        data += rng.read(1)

    lzw_compressed = compress(data)
    out.write(lzw_compressed)

    return True



# Compresses an input string with the LZW algorithm and returns it as bytes
def compress(input: bytes) -> bytes:
    output = bytearray()

    dictionary = {bytes([i]): bytes([i]) for i in range(256)}

    word = bytes([ord(input[0])])
    for c in input[1:]:
        wc = word + bytes([ord(c)])

        if wc in dictionary:
            word = wc
        else:
            output += dictionary[word]

            if len(dictionary) < 4096:
                dictionary[wc] = dictionary[word] + bytes([len(dictionary)])

            word = bytes([ord(c)])

    output += dictionary[word]

    return output

