#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'

    # Header
    out.write(header)

    # Logical screen descriptor
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    bg     = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    aspect = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    out.write(struct.pack('<HHBBB', width, height, *flags))
    out.write(bg)
    out.write(aspect)

    # Global color table
    num_colors = min((width*height)//50, 256)
    for i in range(num_colors):
        r = int.from_bytes(rng.read(1), byteorder='little')
        g = int.from_bytes(rng.read(1), byteorder='little')
        b = int.from_bytes(rng.read(1), byteorder='little')
        out.write(b'\x00\x00\x00')

    # Image descriptor
    left   = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16)
    top    = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16)
    width  = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    height = int.from_bytes(rng.read(2), byteorder='little') % (1 << 16) + 1
    flags  = bytes([int.from_bytes(rng.read(1), byteorder='little')])
    lct    = bool(int.from_bytes(rng.read(1), byteorder='little'))
    interlaced = False if not flags[0] & 0x40 else True
    sort ='sort' if flags[0] & 0x20 else ''
    local_color_table = None
    if lct:
        num_colors = min((width*height)//50, 256)
        local_color_table = [None]*num_colors
        for i in range(num_colors):
            r = int.from_bytes(rng.read(1), byteorder='little')
            g = int.from_bytes(rng.read(1), byteorder='little')
            b = int.from_bytes(rng.read(1), byteorder='little')
            local_color_table[i] = b'\x00\x00\x00'
            out.write(local_color_table[i])
    out.write(struct.pack('<HHHHB', left, top, width, height, *flags))

    # Local color table
    if lct:
        for i in range(num_colors):
            r = int.from_bytes(rng.read(1), byteorder='little')
            g = int.from_bytes(rng.read(1), byteorder='little')
            b = int.from_bytes(rng.read(1), byteorder='little')
            local_color_table[i] = b'\x00\x00\x00'
            out.write(local_color_table[i])

    # Graphics control extension
    has_gce = bool(int.from_bytes(rng.read(1), byteorder='little'))
    if has_gce:
        block_size = int.from_bytes(rng.read(1), byteorder='little')
        assert block_size == 4

        packed = int.from_bytes(rng.read(1), byteorder='little')
        disposal_method = (packed >> 2) & 7
        user_input      = bool(packed & 2)
        transparent_idx = packed & 1
        delay           = int.from_bytes(rng.read(2), byteorder='little')
        transparancy    = bytes([int.from_bytes(rng.read(1), byteorder='little')])
        out.write(b'\x21\xF9\x04')
        out.write(struct.pack('BBHBB',
                             block_size,
                             0x01 | ((disposal_method & 7) << 2) |
                                (user_input       << 1) |
                                 (transparent_idx),
                             delay,
                             *transparancy
                            ))

    # Plain text extension
    has_pte = bool(int.from_bytes(rng.read(1), byteorder='little'))
    if has_pte:
        block_size = int.from_bytes(rng.read(1), byte