#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (os.urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    # Header
    out.write(b'GIF89a')

    # Logical screen descriptor
    width = rng.read(1)[0] % 256 + 1
    height = rng.read(1)[0] % 256 + 1
    background_color = rng.read(1)[0] % 256
    pixel_aspect_ratio = rng.read(1)[0] % 256
    out.write(struct.pack('<HHBB', width, height, background_color, pixel_aspect_ratio))

    # Global color table
    global_color_table_size = 7 + rng.read(1)[0] % 8
    global_color_table_sort_flag = rng.read(1)[0] & 1
    global_color_table_colors = [None] * (2 ** global_color_table_size)
    for i in range(len(global_color_table_colors)):
        red = rng.read(1)[0] % 256
        green = rng.read(1)[0] % 256
        blue = rng.read(1)[0] % 256
        global_color_table_colors[i] = (red, green, blue)
    out.write(struct.pack('B', global_color_table_size | (global_color_table_sort_flag << 7)))
    for rgb in global_color_table_colors:
        out.write(struct.pack('BBB', *rgb))

    # Image descriptor
    left = rng.read(1)[0] % 256
    top = rng.read(1)[0] % 256
    image_width = rng.read(1)[0] % 256 + 1
    image_height = rng.read(1)[0] % 256 + 1
    local_color_table_flag = rng.read(1)[0] & 1
    interlace_flag = rng.read(1)[0] & 1
    sort_flag = rng.read(1)[0] & 1
    reserved_flags = rng.read(1)[0] & 7
    lct_size = 2 ** ((reserved_flags >> 3) + 1) if local_color_table_flag else None
    lct_sort_flag = reserved_flags & 1
    out.write(struct.pack('<HHBBBBBB', left, top, image_width, image_height,
            local_color_table_flag, interlace_flag, sort_flag, reserved_flags))
    if local_color_table_flag:
        colors = [None] * lct_size
        for i in range(lct_size):
            red = rng.read(1)[0] % 256
            green = rng.read(1)[0] % 256
            blue = rng.read(1)[0] % 256
            colors[i] = (red, green, blue)
        out.write(struct.pack('B', lct_size | (lct_sort_flag << 7)))
        for rgb in colors:
            out.write(struct.pack('BBB', *rgb))

    # Local color table
    if not local_color_table_flag and lct_size is not None:
        colors = [None] * lct_size
        for i in range(lct_size):
            red = rng.read(1)[0] % 256
            green = rng.read(1)[0] % 256
            blue = rng.read(1)[0] % 256
            colors[i] = (red, green, blue)
        out.write(struct.pack('B', lct_size | (lct_sort_flag << 7)))
        for rgb in colors:
            out.write(struct.pack('BBB', *rgb))

    # Table based image data
    lzw_min_code_size = rng.read(1)[0] % 256
    out.write(struct.pack('B', lzw_min_code_size))

    # LZW code length table
    lzw_code_length_table = []
    for _ in range(4096):
        lzw_code_length_table += [(rng.read(1)[0] % 256)]
    out.write(bytes(lzw_code_length_table))

    # Pixels
    pixels = bytearray()
    while len(pixels) < image_width * image_height:
        current_byte = rng.read(1)[0]

        # Check if we need to add more bytes from the stream
        if current_byte == 0x00 or current_byte == 0xff:
            continue
        
        # Add the new byte to our buffer
        pixels += bytes([current_byte])

        # If this was the last byte of the block, decode it
        if current_byte >= lzw_min_code_size:
            # Get the next byte which tells us how many times to repeat the previous byte
            num_repeats = rng.read(1)[0]

            # Repeat the previous byte as many times as specified by the next byte
            for _ in range(num_repeats):
                pixels[-1] = current_byte
    
    # Write the compressed image data
    out.write(bytes(pixels))


if