#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The size of the output is determined by the first 10 bytes read from `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Read the header and determine the length of the image data
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')

    # Write the header to the output
    out.write(header)

    # Write the rest of the data to the output
    while datalen > 0:
        chunklen = min(datalen, 256)

        # Generate some random pixels for this chunk
        pixels = [
                (
                    rng.read(1)[0],
                    rng.read(1)[0],
                    rng.read(1)[0]
                )
                for _ in range(chunklen)]

        # Encode these pixels as an LZW-compressed subblock
        compressed = lzw_compress(pixels)

        # Write the compressed subblock to the output
        out.write(struct.pack('B', len(compressed)))
        out.write(compressed)

        datalen -= chunklen


# Compresses a list of RGB triples with the LZW algorithm.
# Returns the compressed data as a byte string.
def lzw_compress(data):
    dictsize = 256
    dictionary = {i : chr(i) for i in range(dictsize)}

    # Add all characters to the dictionary
    for char in ''.join([chr(c[0]) + chr(c[1]) + chr(c[2]) for c in data]):
        if not char in dictionary:
            dictionary[dictsize] = char
            dictsize += 1

    # Initialize the compression state
    prevcode = ''
    code = ''
    result = []

    # Compress each pixel triple
    for rgb in data:
        code = chr(rgb[0]) + chr(rgb[1]) + chr(rgb[2])

        # If we've seen this sequence before, just use its index
        if code == prevcode + prevcode[0]:
            result.append((dictionary[prevcode].encode(), False))
            prevcode = code

        else:

            # Find the longest prefix that's already been added to the dictionary
            maxprefix = ''
            for i in reversed(range(1, len(code))):

                prefix = code[:i]
                if prefix in dictionary:

                    # If it's longer than any previous match, save it
                    if len(maxprefix) < len(prefix):
                        maxprefix = prefix

            # Save the longest matching prefix
            result.append((dictionary[maxprefix].encode(), True))

            # Add the new character to the dictionary
            dictionary[dictsize] = code[-1:]
            dictsize += 1

            # Update the previous code
            prevcode = code

    return b''.join([r[0] for r in result])