#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = rng.read(1)[0] & 0xF + 1
    height = (rng.read(1)[0] & 0xF) << 4 | (rng.read(1)[0] & 0xF) + 1
    flags = rng.read(1)[0] & 0x80 | 0x70 | ((rng.read(1)[0] & 0x01) << 5)
    bg_color_index = rng.read(1)[0]
    aspect_ratio = rng.read(1)[0] & 0xC0 | (rng.read(1)[0] & 0x3F) >> 2
    out.write(struct.pack('<HHBBB', width, height, flags, bg_color_index, aspect_ratio))

    # Write the global color table
    for i in range(768):
        out.write(bytes([rng.read(1)[0]] * int((i % 3 == 2))))

    # Write the application extension block
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00\x00')

    # Write the image descriptor
    left = rng.read(1)[0] & 0xF | (rng.read(1)[0] & 0xF) << 4
    top = rng.read(1)[0] & 0xF | (rng.read(1)[0] & 0xF) << 4
    flags = rng.read(1)[0] & 0xE0 | (rng.read(1)[0] & 0x10) << 2 | (rng.read(1)[0] & 0x01) << 7
    delay = rng.read(1)[0] | (rng.read(1)[0] & 0x0F) << 8
    transparent_color_index = rng.read(1)[0]
    out.write(struct.pack('<HHBBBBBB', left, top, width, height, flags, delay, transparent_color_index, 0))

    # Write the local color table
    for i in range(768):
        out.write(bytes([rng.read(1)[0]] * int((i % 3 == 2))))

    # Write the image data
    out.write(b'\x00\x00')

    # Write the LZW compressed image data
    lzw_min_code_sizes = [8, 4, 2, 0]
    for lzw_min_code_size in lzw_min_code_sizes:
        out.write(bytes([lzw_min_code_size]))

        while True:
            code = rng.read(1)[0]
            if code == 0:
                break
            elif code < lzw_min_code_size:
                out.write(bytes([code]))
            elif code == lzw_min_code_size:
                out.write(b'\x00\x00')
            else:
                raise Exception('Invalid LZW min code size')

    # Write the trailer
    out.write(b';')