#!/usr/bin/env python3

import sys
import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom).
# The size of the output is determined by the first 10 bytes read from `rng`.
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # Read the header and determine the length of the image data
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')

    # Write the header to the output
    out.write(header)

    # Write the rest of the data to the output
    while datalen > 0:
        chunklen = min(datalen, 256)

        # Generate some random pixels for this chunk
        pixels = [
                (
                    rng.read(1)[0],
                    rng.read(1)[0],
                    rng.read(1)[0]
                )
                for i in range(chunklen)
            ]

        # Write them as an LZW-compressed block
        lzw_compress(pixels, out)

        datalen -= chunklen


# Compresses a list of RGB tuples using the LZW algorithm.
# This function writes the compressed data directly to `out`, without returning it.
def lzw_compress(
        pixels: [(int, int, int)], out: BinaryIO
    ) -> None:

    # Initialize the dictionary with all possible codes
    dictsize = 256
    codesize = 9
    dict = {i: chr(i).encode('ascii') for i in range(dictsize)}

    # Start with the clear code
    out.write(b'\x80\x00')

    # Initialize the buffer with the first pixel's color
    buf = pixels[0]

    # Iterate over every pixel after the first one
    for p in pixels[1:]:
        # Find the longest prefix that matches the current buffer
        match = ''
        for c in reversed(range(codesize)):
            if len(match) == 0 or \
               dict[c].decode() == match:

                match += dict[c].decode()

            else:
                break

        # If we found no matching prefix, add a new entry to the dictionary
        if len(match) < 1:
            dictsize += 1
            dict[dictsize - 1] = match + dict[buf[0]]
            codesize += 1
            match = match + dict[buf[0]].decode()

        # Add the next character to the end of our match string
        match += dict[p[0]].decode()

        # Output the encoded value
        out.write((len(match) - 1).to_bytes(1, 'big'))
        out.write(dict[ord(match[-1])].ljust(2, '\x00'.encode()))

        # Update the buffer
        buf = p