#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO):

    header = b'GIF89a'
    # 0x100 is arbitrary but should be enough to cover most cases
    for _ in range(0x100):
        header += bytes([rng.read(1)[0]])

    # Write the header
    out.write(header)

    # Write the logical screen descriptor
    width = rng.read(1)[0] & 0xF + 1
    height = (rng.read(1)[0] & 0xF) << 4 | rng.read(1)[0] & 0xF + 1
    packed = ((width - 1) & 0xF) << 4 | (height - 1) & 0xF
    out.write(struct.pack('<H', 0x200))
    out.write(packed.to_bytes(1, 'little'))
    out.write(b'\x08')
    out.write((rng.read(1)[0] & 0x0F).to_bytes(1, 'little'))

    # Write the global color table
    for i in range(768):
        out.write(((i >> 5) & 0x07).to_bytes(1, 'little'))
        out.write(((i >> 2) & 0x07).to_bytes(1, 'little'))
        out.write(((i >> 0) & 0x03).to_bytes(1, 'little'))

    # Write the application extension block
    out.write(b'\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00\x00')

    # Write the image descriptor
    out.write(b'\x2c\x00\x00\x00\x00\x00\x00\x00\x00')

    # Write the local color table
    for i in range(768):
        out.write(((i >> 5) & 0x07).to_bytes(1, 'little'))
        out.write(((i >> 2) & 0x07).to_bytes(1, 'little'))
        out.write(((i >> 0) & 0x03).to_bytes(1, 'little'))

    # Write the image data
    out.write(b'\x00\x00')

    # Write the LZW compressed image data
    lzw_min_code_sizes = [8, 4, 2, 0]
    for lzw_min_code_size in lzw_min_code_sizes:
        out.write(bytes([lzw_min_code_size]))

        while True:
            code = rng.read(1)[0]
            if code == 0:
                break
            elif code <= lzw_min_code_size:
                out.write(bytes([code]))
            else:
                out.write(bytes([(code >> 8) & 0xFF]))
                out.write(bytes([code & 0xFF]))

    # Write the trailer
    out.write(b';')
