#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    out.write(b"GIF89a")
    out.write(struct.pack("<H", 16))
    out.write(struct.pack("<HHBB", 0, 0, 0xF7, 0))
    out.write(b"\x2C\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01;")

    # Now we're ready to start writing the data section.  We'll use a simple
    # LZW compression scheme to compress the data.  This is not the most
    # efficient way to do this, but it should work well enough for our purposes.
    #
    # First, we need to initialize the dictionary with all of the possible
    # byte values.
    dict = [chr(i) for i in range(256)]
    dict += [""]

    # Next, we'll read some random bytes from the RNG and write them as
    # compressed data.
    while True:
        # Read a new byte from the RNG.
        b = ord(rng.read(1))

        # If the byte isn't already in the dictionary, add it.
        if b >= len(dict):
            dict.append(dict[b] + dict[b])

        # Write the current dictionary index (which is one more than the
        # previous byte).
        out.write(struct.pack("B", b+1))

        # If we've reached the end of the dictionary, stop.
        if b == 255:
            break

    # Finally, terminate the stream by writing a single zero byte.
    out.write(struct.pack("B", 0))

