#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    header = b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\xff\xff\xff\x21\xf9\x04\x01\x0a\x00\x01\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b'
    out.write(header)

    # Next, we need to figure out how many bytes of random data we should write
    # into the file.  This is done by reading three bytes from /dev/urandom and
    # interpreting them as a little-endian unsigned integer.
    datalen = int.from_bytes(rng.read(3), 'little')

    # Now we write the data section itself.  It consists of a bunch of "nets"
    # which are just blocks of pixels.  Each net starts with a byte indicating
    # the number of rows in the net (up to 255).  Then each row is a variable
    # length block of pixels.  Each pixel is represented as a single byte,
    # where the high four bits are the red component and the low four bits are
    # the blue component.  So for example, if you wanted to represent a red
    # pixel, you would write \xF0.
    while datalen > 0:
        # First, pick a random number of rows between 1 and 255 inclusive.
        numrows = rng.read(1)[0] % 256 + 1

        # Write this out as a one-byte integer.
        out.write(struct.pack('B', numrows))

        # Now write out the actual pixel data.  For each row, we pick a random
        # number of pixels between 1 and 255 inclusive.  Then we write that
        # many bytes of random data.
        for _ in range(numrows):
            numpixels = rng.read(1)[0] % 256 + 1
            out.write(rng.read(numpixels))

        # Finally, write a zero byte to indicate the end of this net.
        out.write(b'\x00')

        # Update our running count of how much data we have left to write.
        datalen -= 1 + numrows * (1 + numpixels)

