#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # We need to write a GIF header first.
    out.write(struct.pack("!6s", b"GIF89a"))

    # Then we need to write the screen descriptor.
    width = int.from_bytes(rng.read(2), byteorder="little") % 256 + 1
    height = int.from_bytes(rng.read(2), byteorder="little") % 256 + 1
    background = rng.read(1)[0] & 0xFF
    color_table_size = (rng.read(1)[0] & 0xF) + 1
    flags = rng.read(1)[0] & 0x7F
    out.write(struct.pack("!4B", width, height, background, flags))

    # The global color table is next.
    if color_table_size > 0:
        for _ in range(0, 3 * color_table_size):
            out.write(rng.read(1))

    # Next comes the image descriptor.
    left = int.from_bytes(rng.read(2), byteorder="little") % 256
    top = int.from_bytes(rng.read(2), byteorder="little") % 256
    width = int.from_bytes(rng.read(2), byteorder="little") % 256 + 1
    height = int.from_bytes(rng.read(2), byteorder="little") % 256 + 1
    flags = rng.read(1)[0] & 0x87
    out.write(struct.pack("!8B", left, top, width, height, flags, 0, 0, 0))

    # Finally, we have the local color table.
    if flags & 0x80 != 0:
        for _ in range(0, 3 * (2 ** ((flags & 0x07) + 1))):
            out.write(rng.read(1))

    # Now we need to calculate how many bytes of random data we'll be writing.
    # This is done by subtracting the size of the GIF header and the image
    # descriptor from the total size of the file.
    datalen = os.fstat(out.fileno()).st_size - 10 - 8
    # Now we can write the actual data to the file.
    for _ in range(0, datalen):
        out.write(rng.read(1))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)