#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    header = b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\xff\xff\xff\x21\xf9\x04\x01\x0a\x00\x01\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b'
    out.write(header)

    # Now we're going to write the data section.  It consists of a bunch of
    # blocks, each of which starts with a byte indicating its type.  There are
    # two types of blocks: extension blocks (which start with `!`) and image
    # blocks (which start with `,`).  Extension blocks have a subtype, but
    # image blocks don't.  Image blocks also contain a bunch of metadata, like
    # the width and height of the image.
    while True:
        block_type = rng.read(1)
        if not block_type:
            break
        if block_type != b'!':
            continue
        subtype = rng.read(1)
        if subtype == b'NETSCAPE2.0':
            # This is the application extension block that tells the browser
            # that this is an animated GIF.  It contains a loop count of zero,
            # which means "loop forever".
            out.write(subtype + b'\x03\x01\x00\x00\x00')
        else:
            out.write(rng.read(1))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)