#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    # Generate a random GIF.
    #
    # The format is described in the following document:
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt
    #
    # This function generates a valid GIF with a single frame.
    # It does not generate a valid GCE or PTE (Graphic Control Extension and
    # Plain Text Extension) block.
    #
    # The generated GIF has the following properties:
    # - 1x1 pixel size
    # - 256 color palette
    # - No transparency
    # - No user input flag
    # - No disposal method
    # - No delay time
    # - No comment
    # - No application extension
    # - No loop count
    #
    # Parameters:
    #   rng: A file object that provides random bytes.
    #   out: A file object to write the output to.
    #
    # Returns: Nothing.

    # Magic number "GIF"
    out.write(b'GIF')

    # Version "89a"
    out.write(b'89a')

    # Logical screen descriptor
    out.write(struct.pack('<H', 1))  # Width
    out.write(struct.pack('<H', 1))  # Height
    out.write(b'\x00\x00')           # Background color index
    out.write(b'\x00')               # Aspect ratio

    # Global color table
    for _ in range(7):
        out.write(rng.read(1))       # Red
        out.write(rng.read(1))       # Green
        out.write(rng.read(1))       # Blue

    # Image descriptor
    out.write(b'\x2c')               # Separator
    out.write(struct.pack('<H', 0))  # Left position
    out.write(struct.pack('<H', 0))  # Top position
    out.write(struct.pack('<H', 1))  # Width
    out.write(struct.pack('<H', 1))  # Height
    out.write(b'\x00')               # Local color table flag
    out.write(b'\x00')               # Interlace flag
    out.write(b'\x00')               # Sort flag
    out.write(b'\x00')               # Size of local color table

    # LZW minimum code size
    out.write(b'\x08')

    # Image data
    out.write(b'\x00\x00\xff\x00')

    # Trailer
    out.write(b';')

