#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 17 20:56:49 2019

@author: jinyi
"""
import numpy as np
from scipy.stats import norm, multivariate_normal
from scipy.optimize import minimize

def log_likelihood(theta, X):
    """
    Log likelihood function for theta and X
    
    Parameters
    ----------
    theta : array of shape (n_features,)
        The parameters to estimate by MLE
        
    X : array of shape (n_samples, n_features)
        The data set
        
    Returns
    -------
    float
        The log likelihood value
    """
    mu = theta[:-1]
    sigma = theta[-1]
    n_samples = X.shape[0]
    return -(n_samples / 2) * np.log(2 * np.pi) \
            - (n_samples / 2) * np.log(sigma) \
            - (1 / (2 * sigma)) * np.sum((X - mu) ** 2)

def mle(X):
    """
    Estimate the maximum likelihood estimator of the parameters of a Gaussian
    distribution given a dataset X
    
    Parameters
    ----------
    X : array of shape (n_samples, n_features)
        The data set
        
    Returns
    -------
    array of shape (n_features,)
        The maximum likelihood estimator of the parameters of a Gaussian
        distribution
    """
    def f(theta):
        return -log_likelihood(theta, X)
    initial_guess = np.zeros(X.shape[1])
    res = minimize(f, initial_guess, method='L-BFGS-B')
    return res.x

def pdf(theta, x):
    """
    Probability density function of a Gaussian distribution with mean mu and
    standard deviation sigma evaluated at x
    
    Parameters
    ----------
    theta : array of shape (n_features + 1,)
        The parameters of the Gaussian distribution
        
    x : array of shape (n_features,)
        The point where the probability density function is evaluated
        
    Returns
    -------
    float
        The probability density function evaluated at x
    """
    mu = theta[:-1]
    sigma = theta[-1]
    return np.exp(-np.sum((mu - x)**2) / (2 * sigma**2)) / (np.sqrt(2 * np.pi * sigma**2))

def log_pdf(theta, x):
    """
    Logarithm of the probability density function of a Gaussian distribution
    with mean mu and standard deviation sigma evaluated at x
    
    Parameters
    ----------
    theta : array of shape (n_features + 1,)
        The parameters of the Gaussian distribution
        
    x : array of shape (n_features,)
        The point where the probability density function is evaluated
        
    Returns
    -------
    float
        The logarithm of the probability density function evaluated at x
    """
    mu = theta[:-1]
    sigma = theta[-1]
    return -np.sum((mu - x)**2) / (2 * sigma**2) - np.log(np.sqrt(2 * np.pi * sigma**2))

def sample(theta, size=1):
    """
    Sample from a Gaussian distribution with mean mu and standard deviation
    sigma
    
    Parameters
    ----------
    theta : array of shape (n_features + 1,)
        The parameters of the Gaussian distribution
        
    size : int, optional
        Number of samples to draw (default is 1)
        
    Returns
    -------
    array of shape (size, n_features)
        Samples drawn from the Gaussian distribution
    """
    mu = theta[:-1]
    sigma = theta[-1]
    return np.random.randn(*mu.shape) * sigma + mu

def rvs(theta, size=1):
    """
    Draw random samples from a Gaussian distribution with