#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    out.write(b"GIF89a")

    # The next two bytes are the width and height of the image.  We're going to
    # make them both 1024 pixels wide and tall.
    out.write(struct.pack("<HH", 1024, 1024))

    # Next byte is the "global color table flag".  If this is set, then there
    # will be a global color table after the header.  Since we don't have one,
    # we'll clear this bit.
    out.write(b"\x00")

    # The next byte is the background color index.  We'll use black (index 0).
    out.write(b"\x00")

    # The next byte is the pixel aspect ratio.  We'll leave this at its default
    # value of 0, which means that each pixel is square.
    out.write(b"\x00")

    # Now we need to write the global color table.  This is a list of RGB
    # triples, where each triple is represented as three bytes.  There are 256
    # colors in the table, so we'll write 768 bytes total.
    for _ in range(768):
        out.write(os.urandom(3))

    # Now we need to write the application extension block.  This is a special
    # kind of block that allows us to store arbitrary data in the file.  It's
    # used here to store the seed for the PRNG.
    out.write(b"\x21\xff\x0bNETSCAPE2.0\x03\x01\x00\x00\x00")

    # Now we need to write the graphic control extension block.  This tells the
    # decoder what to do with the upcoming frame.  In our case, we want to
    # indicate that the frame should not be displayed until all previous frames
    # have been displayed.
    out.write(b"\x21\xf9\x04\x01\x00\x00\x00")

    # Now we need to write the image descriptor block.  This contains the
    # coordinates of the top left corner of the image and the size of the image.
    # We'll put the image in the upper left corner of the canvas and make it
    # fill the entire canvas.
    out.write(b"\x2c\x00\x00\x00\x00\x00\x00\x00\x00")

    # Now we need to write the local color table.  This is the same as the
    # global color table above, except that it only applies to this particular
    # image.
    for _ in range(768):
        out.write(os.urandom(3))

    # Finally, we need to write the actual image data.  Each pixel is stored as
    # a single byte, indicating which color from the color table to use.  So
    # we'll read 1024 * 1024 = 1048576 bytes from /dev/urandom.
    for _ in range(1048576):
        out.write(rng.read(1))

    # And now we need to write the trailer.  This indicates the end of the GIF
    # file.
    out.write(b"\x3b")

