#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
) -> None:
    """Generate a random GIF image."""

    # First 6 bytes are fixed.
    out.write(b'GIF89a')

    # The next 7 bytes are the screen descriptor.
    width = int.from_bytes(rng.read(2), byteorder='little') + 1
    height = int.from_bytes(rng.read(2), byteorder='little') + 1
    packed = int.from_bytes(rng.read(1), byteorder='little')
    background = (packed & 0x07) << 4 | ((packed >> 4) & 0x07)
    aspect = (packed >> 4) & 0x0f
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    out.write(background.to_bytes(1, 'little'))
    out.write(aspect.to_bytes(1, 'little'))

    # Then comes the global color table.
    numcolors = 1 << ((packed >> 5) & 0x07)
    if numcolors > 0:
        for _ in range(0, numcolors * 3):
            out.write(rng.read(1))

    # Next is the graphic control extension.
    out.write(b'\x21\xf9\x04\x01\x00\x00\xff\x00')

    # And finally the image descriptor.
    left = int.from_bytes(rng.read(2), byteorder='little') + 1
    top = int.from_bytes(rng.read(2), byteorder='little') + 1
    width = int.from_bytes(rng.read(2), byteorder='little') + 1
    height = int.from_bytes(rng.read(2), byteorder='little') + 1
    packed = int.from_bytes(rng.read(1), byteorder='little')
    localct = bool(packed & 0x80)
    interlaced = bool(packed & 0x40)
    sortflag = bool(packed & 0x20)
    reserved = bool(packed & 0x10)
    size = (packed & 0x07) + 1
    out.write(b'\x2c')
    out.write((left).to_bytes(2, 'little'))
    out.write((top).to_bytes(2, 'little'))
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    out.write(localct.to_bytes(1, 'little'))
    out.write(interlaced.to_bytes(1, 'little'))
    out.write(sortflag.to_bytes(1, 'little'))
    out.write(reserved.to_bytes(1, 'little'))
    out.write(size.to_bytes(1, 'little'))

    # Local color table follows.
    if localct:
        for _ in range(0, 3 * (2 ** size)):
            out.write(rng.read(1))

    # Image data starts here.
    datalen = width * height
    out.write(datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    for _ in range(0, datalen):
        out.write(rng.read(1))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)