#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

def main(argv: list[str]) -> int:
    if len(argv) < 2:
        print("Usage: {} <file>".format(os.path.basename(__file__)), file=sys.stderr)
        return 1
    
    with open(argv[1], "rb") as f:
        data = f.read()
        
    # Read the header
    magic, version, flags, count = struct.unpack("<4sIHH", data[:10])
    assert magic == b"BMP ", "Not a BMP file!"
    assert version in (19778, 19779), "Unknown BMP version!"
    assert not flags & 0x000C, "Compressed BMP files are unsupported!"
    assert not flags & 0x00F0, "Unsupported color depth!"
    assert not flags & 0xFF00, "Reserved bits must be zero!"
    
    # Read the DIB header
    dib_header_size = struct.unpack("<I", data[10:14])[0]
    assert dib_header_size >= 40, "Invalid DIB header size!"
    width, height, planes, bitcount = struct.unpack("<iihh", data[14:26])
    compression, image_size, x_pixels_per_meter, y_pixels_per_meter, colors_used, important_colors = struct.unpack("<IIiiII", data[26:40])
    assert compression == 0, "Compressed BMP files are unsupported!"
    assert bitcount in (1, 4, 8, 24, 32), "Unsupported color depth!"
    
    # Read the palette
    palette_offset = 40 + dib_header_size
    palette_length = 4 * ((1 << bitcount) - 1)
    palette = []
    for i in range(palette_length):
        r, g, b, _ = struct.unpack("<BBBB", data[palette_offset+i*4:palette_offset+(i+1)*4])
        palette.append((r, g, b))
    
    # Read the pixels
    pixel_data_offset = 54 + dib_header_size
    row_size = (width * bitcount + 7) // 8
    row_padding = (row_size % 4) or 4
    rows = []
    for y in range(height-1, -1, -1):
        row = []
        for x in range(width):
            if bitcount == 1:
                byte = data[pixel_data_offset + y//8*row_size + x//8]
                bit = 7 - x%8
                
                if byte & (1<<bit):
                    row.append(1)
                else:
                    row.append(0)
            
            elif bitcount == 4:
                byte = data[pixel_data_offset + y//2*row_size + x//2]
                if y%2 and x%2:
                    bit = 0
                elif y%2:
                    bit = 4
                elif x%2:
                    bit = 4
                else:
                    bit = 0
                
                value = (byte >> bit) & 0xF
                row.append(value)
            
            elif bitcount == 8:
                row.append(data[pixel_data_offset + y*row_size + x])
            
            elif bitcount == 24:
                row.extend([
                    data[pixel_data_offset + y*row_size + 3*x],
                    data[pixel_data_offset + y*row_size + 3*x+1],
                    data[pixel_data_offset + y*row_size + 3*x+2],
                ])
            
            elif bitcount == 32:
                row.extend([
                    data[pixel_data_offset + y*row_size + 4*x],
                    data[pixel_data_offset + y*row_size + 4*x+1],
                    data[pixel