#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # a variable number of chunks. We'll just write the header first, and then
    # we can write the rest of the file later.
    out.write(b"GIF89a")  # Magic bytes
    out.write(struct.pack("<H", 10))  # Width
    out.write(struct.pack("<H", 10))  # Height
    out.write(b"\xF4\x00")  # Background color (white)
    out.write(b"\x00")  # Aspect ratio (unused)

    # Now we need to write the image data. This will be a series of "subblocks".
    # Each subblock starts with a byte indicating its length, followed by that
    # many bytes of actual data. Since we don't know how long the data will be,
    # we'll have to write it all at once after we're done generating it.
    #
    # First, though, we need to write the initial subblock, which contains the
    # LZW dictionary size.
    out.write(b"\x2C")  # Image block indicator
    out.write(b"\x00")  # Left coordinate
    out.write(b"\x00")  # Top coordinate
    out.write(b"\x0A")  # Width
    out.write(b"\x0A")  # Height
    out.write(b"\x00")  # No local color table
    out.write(b"\x00")  # Reserved bits
    out.write(b"\x07")  # Size of LZW dictionary

    # Now we can start writing the image data itself. To do this, we'll use an
    # LZW encoder from the standard library.
    from zlib import compressobj
    lzw = compressobj()

    # We'll also keep track of how much data we've written so far.
    total_written = 0

    while True:
        # Read a random byte from stdin.
        b = rng.read(1)
        if not b:
            break
        b = int.from_bytes(b, "big")

        # Encode it using the LZW encoder.
        encoded = lzw.compress(b.to_bytes(1, "big"))

        # If there was any output, write it to stdout.
        if len(encoded):
            out.write(len(encoded).to_bytes(1, "big"))
            out.write(encoded)
            total_written += len(encoded) + 1

    # Write the final chunk.
    out.write(b"\x00")

    # Now we can go back and fill in the size of the image data.
    out.seek(10)
    out.write(total_written.to_bytes(3, "little"))

    # Finally, write the trailer.
    out.write(b";")

