#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an arbitrary number of frames. We'll just write the header here and then
    # we can write each frame as we go along.
    #
    # The header is 13 bytes long and looks like this:
    #
    #   +0: "GIF89a" (6 bytes)
    #   +6: Logical Screen Width (2 bytes, little endian)
    #   +8: Logical Screen Height (2 bytes, little endian)
    #  +10: Packed Fields
    #       +0: Global Color Table Flag (1 bit)
    #           If set to 1, there will be a global color table following the
    #           header. Otherwise, no global color table follows.
    #       +1: Color Resolution (3 bits)
    #           This field indicates how many bits are used to represent each
    #           primary color. It should always be 7 for 256-color images.
    #       +4: Sort Flag (1 bit)
    #           If set to 0, the global color table is sorted by frequency, with
    #           the most frequent color first. If set to 1, the colors are sorted
    #           alphabetically.
    #       +5: Size of Global Color Table (3 bits)
    #           This field indicates how many colors are in the global color
    #           table. It should be 7 for 256-color images.
    #  +11: Background Color Index (1 byte)
    #  +12: Pixel Aspect Ratio (1 byte)
    #       For 256-color images, this should always be 0.
    #
    # After the header comes the global color table. Each entry in the table is
    # 3 bytes long and represents one of the 256 colors in the image. The order
    # of the entries in the table doesn't matter; the indices in the image data
    # refer to the position of each color in the table.
    #
    # Finally, after the global color table comes the image data. Each frame is
    # preceded by a single-byte length field indicating how many bytes follow.
    # Then comes the actual image data.

    # Write the header.
    out.write(b"GIF89a")
    out.write(struct.pack("<H", 256))
    out.write(struct.pack("<H", 256))
    out.write(b"\x87\x00\x00")
    out.write(b"\xff")
    out.write(b"\x00")

    # Write the global color table.
    for i in range(256):
        out.write(struct.pack("BBB", i, i, i))

    # Write the image data.
    while True:
        # Read a random byte from /dev/urandom.
        b = rng.read(1)
        if len(b) == 0:
            break
        b = b[0]

        # Write a single-byte length field followed by the byte itself.
        out.write(struct.pack("B", 1))
        out.write(bytes([b]))

