#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can use
    # `struct` to generate the rest of the file.
    out.write(b"GIF89a")

    # The logical screen descriptor contains information about the size of the
    # image as well as some flags for whether or not there are global color
    # tables.  We'll just pick random values here.
    width = rng.read(2)
    height = rng.read(2)
    flags = rng.read(1)
    background = rng.read(1)[0] % 256
    aspect = rng.read(1)[0] % 256

    # Write the logical screen descriptor to the output stream.
    out.write(width + height + flags + bytes([background]) + bytes([aspect]))

    # Now we need to write the global color table.  This is optional, but if
    # present it must be preceded by a byte indicating how many colors are in
    # the table.  We'll just write a random number of colors.
    num_colors = rng.read(1)[0] % 256
    out.write(bytes([num_colors]))

    # Each color entry consists of three bytes, one each for red, green, and
    # blue.  We'll just write random numbers for those.
    for _ in range(num_colors):
        out.write(rng.read(3))

    # After the color table comes a single byte indicating how long the
    # application extension is.  We'll just write a random value.
    app_ext_len = rng.read(1)[0] % 256
    out.write(bytes([app_ext_len]))

    # Next comes the application extension itself.  It starts with "NETSCAPE2.0",
    # followed by a single byte indicating the length of the loop count (always
    # zero).  We'll just write those two things.
    out.write(b"NETSCAPE2.0\x03\x01")

    # Finally, we need to write the actual animation frames.  Each frame begins
    # with a graphic control extension which indicates the duration of the
    # current frame.  We'll just write a random duration.
    gce_duration = rng.read(2)
    gce_flags = rng.read(1)
    out.write(gce_duration + gce_flags)

    # Then comes the image descriptor, which is pretty much the same as the
    # logical screen descriptor from earlier.
    x = rng.read(2)
    y = rng.read(2)
    width = rng.read(2)
    height = rng.read(2)
    flags = rng.read(1)
    lct_flag = rng.read(1)[0] & 0x80
    interlace_flag = rng.read(1)[0] & 0x40
    sort_flag = rng.read(1)[0] & 0x08
    local_color_table_length = rng.read(1)[0] & 0x07
    out.write(x + y + width + height + flags + bytes([lct_flag | interlace_flag | sort_flag | local_color_table_length]))

    # If there is a local color table, we need to write it now.  Otherwise,
    # we're done.
    if local_color_table_length != 0:
        num_colors = 2 ** (local_color_table_length + 1)
        out.write(bytes([num_colors]))
        for _ in range(num_colors):
            out.write(rng.read(3))

    # Finally, we need to write the image data.  There are a few different ways
    # this could be encoded, but we'll just write raw RGB data.  First, we need
    # to write the LZW minimum code size.
    min_code_size = rng.read(1)[0] % 256
    out.write(bytes([min_code_size]))

    # Then we need to write the compressed data.  To do this, we'll read a
    # random amount of data from the input stream and write it directly to