#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    out.write(b"GIF89a")

    # Write a dummy logical screen descriptor.  This is where you would put
    # things like the width and height of the image, but for now we'll just
    # use some arbitrary values.
    out.write(struct.pack("<H", 10))
    out.write(struct.pack("<H", 20))
    out.write(b"\xF7\x00")

    # Now we need to write the global color table.  It must have exactly one
    # entry for each possible pixel value (i.e., 256 entries).  Each entry is
    # three bytes, representing red, green, and blue components respectively.
    # For this example, we'll just make all pixels white.
    for _ in range(256):
        out.write(b"\xFF\xFF\xFF")

    # Now we need to write the graphic control extension.  This is basically
    # metadata about the next frame, such as whether or not it should be
    # displayed with transparency.  In our case, we don't care about any of
    # those details, so we'll just set everything to zero.
    out.write(b"\x21\xF9\x04\x00\x00\x00\x00")

    # Finally, we need to write the actual image data.  This is where the
    # real magic happens!  First, we're going to write a single byte that
    # tells us how many times to repeat the previous pixel.  Since we want
    # every pixel to be different, we'll just write 0 here.
    out.write(b"\x00")

    # Next, we need to write the LZW code size.  This is the number of bits
    # per pixel in the image data.  We'll just use 8 bits per pixel since
    # that's what most images use.
    out.write(b"\x08")

    # Now we need to write the image data itself.  This is where the real
    # compression comes into play.  Instead of writing raw pixel values, we
    # write codes that represent groups of pixels.  These codes are mapped to
    # pixel values using a lookup table called the "prefix tree".  To start
    # off, we'll just write a bunch of random codes.
    while True:
        try:
            out.write(rng.read(1))
        except OSError:
            break

