#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    """Generate a GIF image containing a single frame that contains a single
    pixel whose color is chosen randomly using the given binary stream."""

    # First, we need to write the magic number "GIF89a" to the output file.
    out.write(b'GIF89a')

    # Then we need to write the screen descriptor block.  The first field is an
    # unsigned short (2 bytes).  We'll just use zeroes here because they don't
    # really matter.
    out.write(struct.pack('<HHHHHBBBBB',
            0x0000,   # Screen width
            0x0000,   # Screen height
            0xf000,   # Flags (global color table present; no sort flag set)
            0x0000,   # Background color index
            0x0000,   # Pixel aspect ratio
            0x00,     # Global color table size (logarithmic lookup table)
            0x00      # Background color index
        ))

    # Finally, we need to write the global color table.  Since this example only
    # uses one color, it's easy enough to hardcode its value.
    out.write(b'\xff\x00\x00\x00')

    # Now comes the image descriptor block.  Again, all fields are unsigned
    # integers except for the flags field which is a bitmask.
    out.write(struct.pack('<HHHHHHBHB',
            0x0000,       # Left position
            0x0000,       # Top position
            0x0001,       # Image width
            0x0001,       # Image height
            0x0000,       # Flags (no local color table; interlace not used)
            0x00,         # Local color table size (logarithmic lookup table)
            0x00          # LZW minimum code size
        ))

    # And now comes the image data itself.  It consists of blocks of up to
     # 255 bytes each.  Each block starts with a byte indicating how long the
    # following data is.  If it's less than 255, then there will be another byte
    # after the length byte indicating whether or not the end of image has been
    # reached.  Otherwise, if it's equal to 255, then the next byte indicates
    # how many more bytes follow before the end of image.
    #
    # In our case, since we're generating a single frame consisting of a single
    # pixel, we know exactly what the data will look like.  So let's go ahead
    # and write it!
    out.write(b'\x07\x00\x00\x00\x00\x01\x00\x01\x00')

    # Next, we need to figure out how many bytes of random data we should write
    # into the file.  This is done by reading three bytes from /dev/urandom and
    # interpreting them as a little-endian unsigned integer.
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    out.write(datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    for _ in range(0, datalen):
        out.write(rng.read(1))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)