#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    header = b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\xff\xff\xff\x21\xf9\x04\x01\x0a\x00\x01\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b'
    out.write(header)

    # Next comes the logical screen descriptor.  It's always exactly 7 bytes.
    # For now, let's just hardcode this one.
    out.write(struct.pack('<HHBBB',
            500, 500, 8, 7, 0, 0))

    # Finally, we need to write the length of the image data section.  Since
    # there are no variable-length integers in Python, we have to convert our
    # desired length (which is arbitrary but must fit within 2^16 - 1) to big-
    # endian before writing it to disk.  Note that since we're using a binary
    # stream here, we don't need to worry about padding or endianness; those
    # details are handled automatically.
    #
    # In order to make things interesting, we'll also read some random bytes
    # into the file.  This is done by reading three bytes from /dev/urandom and
    # interpreting them as a little-endian unsigned integer.
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    out.write(datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    for _ in range(0, datalen):
        out.write(rng.read(1))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)