#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def main(argv):
    if len(argv)!= 2:
        print("Usage: {} <file>".format(os.path.basename(__file__)))
        return -1

    with open(argv[1], "rb") as f:
        magic = f.read(4).decode()

        if magic == "Yaz0":
            decompress_yaz0(f, argv[1])
        elif magic in ("Yaz", "Yaz1"):
            decompress_yaz1(f, argv[1])
        else:
            print("{} is not a Yaz file".format(argv[1]))
            return -1

    return 0


def decompress_yaz0(f: BinaryIO, filename: str):
    """Decompresses yaz0 files."""
    # Read header
    header = f.read(8)
    (magic, uncompressed_size,) = struct.unpack("<4sI", header)
    assert magic == b"Yaz0"

    # Decompress data
    output = bytearray(uncompressed_size)
    offset = 0
    while True:
        control = f.read(1)[0]
        for i in range(8):
            bit = bool((control >> i) & 1)

            if bit:
                copy_offset = f.read(1)[0] | ((f.read(1)[0]) << 8)

                length = min(0x1000 + 1 - copy_offset, uncompressed_size - offset)
                f.seek(-2, os.SEEK_CUR)

                output[offset : offset + length] = output[copy_offset : copy_offset + length]
                offset += length
            else:
                input_byte = f.read(1)[0]
                output[offset] = input_byte
                offset += 1

            if offset >= uncompressed_size:
                break

        if offset >= uncompressed_size:
            break

    # Write to disk
    with open(filename[:-4], "wb") as out:
        out.write(output)


def decompress_yaz1(f: BinaryIO, filename: str):
    """Decompresses yaz1 files."""
    # Read header
    header = f.read(8)
    (magic, uncompressed_size,) = struct.unpack("<4sI", header)
    assert magic in (b"Yaz", b"Yaz1")

    # Decompress data
    output = bytearray(uncompressed_size)
    offset = 0
    while True:
        control = f.read(1)[0]
        for i in range(8):
            bit = bool((control >> i) & 1)

            if bit:
                copy_length = f.read(1)[0]
                copy_offset = f.read(1)[0] | ((f.read(1)[0]) << 8)

                length = min(0x1000 + 1 - copy_offset, uncompressed_size - offset)
                f.seek(-2, os.SEEK_CUR)

                output[offset : offset + copy_length] = output[
                    copy_offset : copy_offset + copy_length
                ]
                offset += copy_length
            else:
                input_byte = f.read(1)[0]
                output[offset] = input_byte
                offset += 1

            if offset >= uncompressed_size:
                break

        if offset >= uncompressed_size:
            break

    # Write to disk
    with open(filename[:-4], "wb") as out:
        out.write(output)


if