#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    """Generate a GIF image using the given random number generator."""

    # First, we need to construct the header for our GIF image.  We'll start off
    # with the magic string "GIF89a".
    header = b'GIF89a'

    # The next two bytes are the logical screen width and height (in that order).
    # Since this is an arbitrary value, let's just use something small.
    w = h = 256
    header += w.to_bytes(2, 'little') + h.to_bytes(2, 'little')

    # Then comes the global color table flag.  If it's set, then there will be a
    # global color table following immediately after these values.  For now, we're
    # not going to bother setting one up, so we'll just put a zero here.
    header += b'\x00'

    # Finally, we have the background color index.  This is an index into the
    # global color table.  Since we don't have any colors yet, we'll just pick
    # some arbitrary value.
    bgcol = 0
    header += bgcol.to_bytes(1, 'little')

    # Now we've constructed the header, we can write it to the output stream.
    out.write(header)

    # Next, we need to figure out how many bytes of random data we should write
    # into the file.  This is done by reading three bytes from /dev/urandom and
    # interpreting them as a little-endian unsigned integer.
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    out.write(datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    for _ in range(0, datalen):
        out.write(rng.read(1))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)