#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def main(argv):
    if len(argv)!= 2:
        print("Usage: {} <file>".format(os.path.basename(__file__)), file=sys.stderr)
        return -1

    with open(argv[1], "rb") as f:
        data = bytearray(f.read())

        # Check for valid header
        if not (data[:4] == b"Yaz0" and
                ((data[8] & 0x7F) in [0, 1]) and
                0 <= int.from_bytes(data[4:8], byteorder="little")):
            raise ValueError("Invalid Yaz0 file!")

        # Decompress the data
        decompressed_size = int.from_bytes(data[4:8], byteorder="little")
        compressed_size = len(data) - 16
        decompressed_data = yaz0_decompress(data[16:], compressed_size, decompressed_size)

        # Write to output file
        out_filename = argv[1][:-4] + ".out"
        with open(out_filename, "wb") as of:
            of.write(decompressed_data)

        print("{} bytes written to {}".format(len(decompressed_data), out_filename))


def yaz0_decompress(data: bytes, compressed_size: int, decompressed_size: int) -> bytes:
    """Decompresses a Yaz0-compressed buffer."""
    assert isinstance(data, bytes)
    assert isinstance(compressed_size, int)
    assert isinstance(decompressed_size, int)

    uncompressed_data = bytearray()
    offset = 0
    while True:
        control_byte = data[offset]
        offset += 1

        for i in range(8):
            bit = bool((control_byte >> i) & 1)

            if bit:
                copy_length = min(4, decompressed_size - len(uncompressed_data))

                # Read the length of the copy from the next two bytes
                copy_offset = int.from_bytes(
                    data[offset : offset+2], byteorder="little", signed=False
                )
                offset += 2

                # Copy the data
                uncompressed_data += data[copy_offset : copy_offset+copy_length]
            else:
                literal_byte = data[offset]
                offset += 1
                uncompressed_data.append(literal_byte)

            if len(uncompressed_data) >= decompressed_size:
                break

        if len(uncompressed_data) >= decompressed_size:
            break

    assert len(uncompressed_data) == decompressed_size
    return uncompressed_data


if