#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    header = b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\xff\xff\xff\x21\xf9\x04\x01\x0a\x00\x01\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b'
    out.write(header)

    # Next, we need to figure out how many bytes of random data we should write
    # into the file.  This is done by reading three bytes from /dev/urandom and
    # treating them as a little endian unsigned integer.
    num_bytes = int.from_bytes(rng.read(3), byteorder='little')

    # Now, we can actually write some data!
    for _ in range(num_bytes):
        out.write(os.urandom(1))

    # Finally, we need to pad the end of the file with null bytes until the
    # length of the entire file is a multiple of four.  Since this padding may
    # also contain our randomly generated data, we need to keep track of where
    # the actual image data ends.
    padding_length = (4 - ((out.tell() + 1) % 4)) % 4
    padding = b'\x00' * padding_length
    out.write(padding)

    # And now comes the tricky part...we need to update the size field in the
    # header based on what we've written above.  To do this, we need to seek back
    # to the beginning of the header and overwrite its last two bytes with the
    # correct value.
    out.seek(6)
    out.write(struct.pack('<H', out.tell()))

    return


if