#!/usr/bin/env python3

import os, sys
from pathlib import Path

# Add the parent directory to the path so we can import our modules.
sys.path.append(str(Path(__file__).parent))

import argparse
import json
import logging
import re
import subprocess
import tempfile


def main():
    parser = argparse.ArgumentParser()

    # Required arguments.
    parser.add_argument('--input', required=True)
    parser.add_argument('--output')
    parser.add_argument('--config-file', required=True)

    args = parser.parse_args()

    with open(args.config_file) as f:
        config = json.load(f)

    if not args.output:
        output = os.path.splitext(os.path.basename(args.input))[0] + '.mp4'
    else:
        output = args.output

    tmpdir = tempfile.TemporaryDirectory().name
    os.makedirs(tmpdir, exist_ok=True)

    input_filename = os.path.join(tmpdir, 'input.mp4')
    output_filename = os.path.join(tmpdir, 'output.mp4')

    ffmpeg_command = [
        'ffmpeg', '-y', '-i', args.input,
        '-c:v', 'libx265', '-preset','slow', '-crf', str(config['crf']),
        '-vf','scale=-1:'+str(config['height']),
        '-pix_fmt', 'yuv420p',
        '-movflags', '+faststart',
        '-an',  # No audio
        '-sn',  # No subtitles
        output_filename
    ]

    logging.info('Running command: %s',''.join(ffmpeg_command))
    result = subprocess.run(ffmpeg_command).returncode
    if result!= 0:
        raise Exception("FFmpeg failed")

    # Copy metadata from original file.
    ffprobe_command = ['ffprobe', '-loglevel', 'quiet', '-print_format', 'json',
                       '-show_streams', '-select_streams', 'a:0', input_filename]
    logging.info('Running command: %s',''.join(ffprobe_command))
    process = subprocess.Popen(ffprobe_command, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    out, _ = process.communicate()
    data = json.loads(out)
    for stream in data["streams"]:
        if "tags" in stream and "title" in stream["tags"] and len(stream["tags"]["title"]) > 0:
            title = stream["tags"]["title"].strip()

            # Remove any non-alphanumeric characters.
            title = re.sub("[^A-Za-z0-9]", "", title)

            # Replace spaces with underscores.
            title = title.replace(" ", "_")

            new_filename = "{} - {}.mp4".format(
                os.path.splitext(os.path.basename(output_filename))[0],
                title
            )
            os.rename(output_filename, new_filename)
            break

    print(new_filename)


if