#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # the image data.  We'll just write the header first.

    # Header (13 bytes)
    # ------------------
    # Signature ("GIF89a")
    signature = b"GIF89a"
    assert len(signature) == 6
    out.write(signature)

    # Logical screen descriptor (7 bytes)
    # -----------------------------------
    # Screen width (2 bytes)
    width = rng.read(2)
    out.write(width)

    # Screen height (2 bytes)
    height = rng.read(2)
    out.write(height)

    # Packed field (1 byte)
    packed_field = rng.read(1)[0]
    out.write(packed_field)

    # Background color index (1 byte)
    background_color_index = rng.read(1)[0]
    out.write(background_color_index)

    # Pixel aspect ratio (1 byte)
    pixel_aspect_ratio = rng.read(1)[0]
    out.write(pixel_aspect_ratio)

    # Global color table (3 * 256 = 768 bytes)
    for _ in range(0, 256):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        out.write(r)
        out.write(g)
        out.write(b)


    # Image descriptor (10 bytes)
    # ----------------------------
    # Image left position (2 bytes)
    image_left_position = rng.read(2)
    out.write(image_left_position)

    # Image top position (2 bytes)
    image_top_position = rng.read(2)
    out.write(image_top_position)

    # Image width (2 bytes)
    image_width = rng.read(2)
    out.write(image_width)

    # Image height (2 bytes)
    image_height = rng.read(2)
    out.write(image_height)

    # Packed field (1 byte)
    packed_field = rng.read(1)[0]
    out.write(packed_field)

    # Local color table (3 * 256 = 768 bytes)
    for _ in range(0, 256):
        r = rng.read(1)[0]
        g = rng.read(1)[0]
        b = rng.read(1)[0]

        out.write(r)
        out.write(g)
        out.write(b)

    # LZW minimum code size (1 byte)
    lzw_minimum_code_size = rng.read(1)[0]
    out.write(lzw_minimum_code_size)

    # Image data block (variable length)
    # ----------------------------------
    # Each byte of the image data is encoded using a variable number of bits
    # from the LZW compression algorithm.  This means we need to know how many
    # bits are used per byte before decoding each byte.  To do this, we use an
    # LZW decoder which keeps track of the last N codes seen so far.  For each
    # new code encountered, if its prefix matches one of those previously seen
    # codes, we can encode it with fewer bits than if it were a new code.  If
    # there's no match, we must emit the entire code as a literal value.
    #
    # In order to keep track of all possible prefixes, we need to store them in
    # a tree structure.  There are two kinds of nodes: internal nodes represent
    # prefixes of longer codes; leaf nodes represent complete codes.  Leaf node
    # values are stored at the end of the file after all other nodes have been
    # written.
    #
    # Here's what the tree might look like when encoding the string "Hello":
    #
    #    Internal Node:
    #      Prefix: ""
    #      Child Nodes: [
    #        Internal Node:
    #          Prefix: "H"
    #          Child Nodes: [
    #            Internal Node:
    #              Prefix: "He"
    #              Child Nodes: [
    #                Internal Node:
    #                  Prefix: "Hel"
    #                  Child Nodes: [
    #                    Internal Node:
    #                      Prefix: "Hell"
    #                      Child Nodes: [
    #                        Leaf Node:
    #                          Value: "Hello"
    #                      ]
    #                  ]
    #                ]
    #              ]
    #            ]
    #          ]
    #        ]
    #      ]
    #
    # As you can see, the longest code is only three characters long, but the
    # tree contains more than 4K nodes!  Since we