#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # First off, let's construct our GIF header.  We'll use version "89a", which
    # is the most common version used today.  The logical screen descriptor
    # (logical screen width, height, color depth, etc.) will be set to 64x64 at
     # 256 colors.  Finally, we'll specify that there are no global color table
    # entries or background color index.
    header = b'GIF89a\x00\x00\x00\x00\xff\xff\xff!\xf9\x04\x01\x00\x00\x00\x00,\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;'
    out.write(header)

    # Next, we need to figure out how many bytes of random data we should write
    # into the file.  This is done by reading three bytes from /dev/urandom and
    # interpreting them as a little-endian unsigned integer.
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    out.write(datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    for _ in range(0, datalen):
        out.write(rng.read(1))


def generate_random_png(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # First off, let's construct our PNG header.  We'll use version "1.0", which
    # is the most common version used today.  The IHDR chunk will be set to 64x64
    # at 256 colors.  Finally, we'll specify that there are no global color table
    # entries or background color index.
    header = b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00`\x00\x00\x00`\x08\x06\x00\x00\x00\xe2\x98w8\x00\x00\x00\x04sBIT\x08\x08\x08\x08|\x08d\x88\x00\x00\x00\tpHYs\x00\x00\x0f\xaf\x00\x00\x0f\xaf\x01\x7b\x8c\xd3\xea\x00\x00\x00\x19tEXtSoftware\x00www.inkscape.org\x9b\xee<\x1a\x00\x00\x00\x0eIDAT(\xcf\xc5\x90\xcd\x0e\x80 \x0cC\x97\xa6\x13\x00\x00\x00\x00IEND\xaeB`\x82'
    out.write(header)

    # Next, we need to figure out how many bytes of random data we should write
    # into the file.  This is done by reading three bytes from /dev/urandom and
    # interpreting them as a little-endian unsigned integer.
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    out.write(datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    for _ in range(0, datalen):
        out.write(rng.read(1))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)