#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    header = b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\xff\xff\xff\x21\xf9\x04\x01\x0a\x00\x01\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b'
    out.write(header)

    # Next, we need to figure out how many bytes of random data we should write
    # into the file.  This is done by reading three bytes from /dev/urandom and
    # interpreting them as a little-endian unsigned integer.
    datalen = int.from_bytes(
        rng.read(3), byteorder='little', signed=False
    )

    # Now we know how much data there's going to be, let's actually write some
    # data!
    for i in range(datalen):

        # Each byte is written as two nybbles (half-bytes).  We use the lower
        # nybble here because it's easier to read.
        out.write(
            struct.pack('B', rng.read(1)[0] & 0xF)
        )

    # Finally, we have to pad the end of the file with enough null bytes to make
    # up the difference between the size of the original header and the actual
    # length of our data.
    padding = len(header) - out.tell()
    if padding > 0:
        out.write(b'\x00'*padding)


if __name__ == '__main__':

    # Open /dev/urandom for reading.
    try:
        rng = open('/dev/urandom', 'rb')
    except FileNotFoundError:
        print("Couldn't find /dev/urandom!")
        sys.exit(-1)

    # Open the output file for writing.
    try:
        outfile = sys.argv[1]
    except IndexError:
        print("No output filename specified.")
        sys.exit(-1)

    try:
        with open(outfile, 'wb') as f:
            generate_random_gif(rng, f)