#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO,
        out: BinaryIO = None,
        width: int = 640, height: int = 480,
        color_depth: int = 256,
        bg_color: int = 0xFFFFFF,
        loop_count: int = 0,
        duration: float = 0.1,
        frame_delays: list[int] = [10],
        palette: bytes = b'\xFF\xFF\xFF\x00\x00\x00' * 256,
        transparency: bool = False,
        disposal_method: int = 0,
        user_input: bool = True,
        comment: str = '',
        ) -> None:
    """Generate a random GIF image."""
    
    if not isinstance(width, int) or (not 1 <= width <= 65535):
        raise ValueError("Width must be an integer between 1 and 65535.")
        
    if not isinstance(height, int) or (not 1 <= height <= 65535):
        raise ValueError("Height must be an integer between 1 and 65535.")
        
    if not isinstance(color_depth, int) or (not 1 <= color_depth <= 256):
        raise ValueError("Color depth must be an integer between 1 and 256.")
        
    if not isinstance(bg_color, int) or (not 0 <= bg_color < pow(2, color_depth)):
        raise ValueError("Background color must be an integer between "
                         "0 and {}.".format(pow(2, color_depth)))
        
    if not isinstance(loop_count, int) or (not -1 <= loop_count <= 65535):
        raise ValueError("Loop count must be an integer between -1 and 65535.")
        
    if not isinstance(duration, (float, int)) or (not 0 <= duration <= 65535):
        raise ValueError("Duration must be a number between 0 and 65535.")
        
    if len(frame_delays)!= 0 and max(frame_delays) > 65535:
        raise ValueError("Maximum delay value is 65535.")
        
    if len(palette)!= color_depth * 3:
        raise ValueError("Palette length must match color depth.")
        
    if not isinstance(transparency, bool):
        raise TypeError("Transparency must be boolean.")
        
    if not isinstance(disposal_method, int) or \
       (not 0 <= disposal_method <= 7):
        raise ValueError("Disposal method must be an integer between 0 and 7.")
        
    if not isinstance(user_input, bool):
        raise TypeError("User input flag must be boolean.")
        
    if not isinstance(comment, str):
        raise TypeError("Comment must be string.")
        
    # We need to know how many frames there will be before we start writing anything.
    numframes = sum([d!= 0 for d in frame_delays])
    
    # The header contains some metadata about the image.
    header = bytearray()
    header += b"GIF89a"
    header += width.to_bytes(2, 'little')
    header += height.to_bytes(2, 'little')
    header += ((color_depth & 0x0F) << 4 |
               ((color_depth >> 4) & 0x0F)).to_bytes(1, 'little')
    header += b"\x80"
    header += b"\x00"
    header += b"\x00"
    header += b"\x00"
    header += numframes.to_bytes(1, 'little')
    header += b"\x00"
    header += b"".join([(b"\x00", b"\xF9")[i % 2]
                        for i in range(numframes)])
    header += b"".join([struct.pack('<H', d) for d in frame_delays])
    header += b"\x00"
    header += b"".join([(b"\x00", b"\xFE")[i % 2]
                        for i in range(numframes)])
    header += b"".join([struct.pack('<B', dm) for dm in range(8)])
    header += b"\x00"
    header += b"\x00"
    header += b"\x00"
    header += b"".join([(b"\x00", b"\xFC")[i % 2]
                        for i in range(numframes)])
    header += b"".join([struct.pack('<B', disposal_method)
                        for _ in    out.write(header + datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    for _ in range(0, datalen):
        out.write(rng.read(1))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)