#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
	"""Generate a random GIF file.
	
	The generated GIF will be 1024x768 pixels and contain one frame with the
	given RNG as its palette."""
	
	# Header
	out.write(b'GIF89a') # Signature
	out.write(struct.pack('<HH', 1024, 768)) # Screen size (width x height)
	out.write(b'\x87\x00\x00') # Flags: Global Color Table Present | Sort Flag
	out.write(b'\x00') # Background color index
	out.write(b'\x00') # Pixel aspect ratio
	
	# Logical screen descriptor
	out.write(b'\x00') # GCT min value
	out.write(b'\x3B') # GCT max value
	out.write(b'\x00') # GCT sort flag
	out.write(b'\x00\x00') # GCT resolution
	out.write(b'\x00\x00') # GCT size
	out.write(b'\x00') # GCT packed field
	
	# Global color table
	for i in range(256):
		r = next(rng) % 256
		g = next(rng) % 256
		b = next(rng) % 256
		
		out.write(bytes([r, g, b]))
	
	# Image descriptor
	out.write(b'\x2C') # Image separator
	out.write(struct.pack('<HHHHBB',
		0, 0, 1024, 768, 0, 0
	)) # Left-top corner coordinates + width & height of image
	out.write(b'\x00') # Local color table flag
	out.write(b'\x00') # Interlace flag
	out.write(b'\x00') # Sort flag
	out.write(b'\x00') # Reserved for future use
	out.write(b'\x00') # LZW minimum code size
	
	# Image data
	data = bytearray()
	
	for y in reversed(range(768)):
		for x in range(1024):
			data += bytes([next(rng)])
	
	lzw_mincodesize = 12
	lzw_maxcodesize = 12
	
	while len(data) > 2**lzw_mincodesize - 1:
		lzw_maxcodesize += 1
	
	lzw_clearcode = 1 << lzw_mincodesize
	lzw_endofinfo = lzw_clearcode + 1
	lzw_endofblock = lzw_endofinfo + 1
	lzw_firstfreecode = lzw_endofblock + 1
	lzw_numextra = lzw_firstfreecode + 1
	lzw_maxcode = lzw_firstfreecode + 2**lzw_maxcodesize
	
	lzw_dict = {i: [i] for i in range(lzw_numextra)}
	lzw_dict[lzw_endofblock].append(None)
	lzw_dict[lzw_endofinfo].append(None)
	
	lzw_prev = None
	lzw_cur = []
	
	def emit():
		nonlocal lzw_prev, lzw_cur
		
		if not lzw_cur or lzw_cur == [lzw_endofblock]:
			return
		
		if lzw_cur[-1] is None:
			raise ValueError('Invalid LZW stream!')
		
		if lzw_prev is None:
			lzw_prev = lzw_cur[0]
			del lzw_cur[0]
			
			emit()
			
		else:
			lzw_dict[len(lzw_dict)].extend(
				lzw_dict[lzw_prev][:-1] + [lzw_cur[0]]
			)
			
			lzw_prev = lzw_cur[0]
			del lzw_cur[0]
		
	def flush():
		nonlocal lzw_prev, lzw_cur
		
		if not lzw_cur:
			return
		
		if lzw_cur!= [lzw_endofblock]:
			raise ValueError('Invalid LZW stream!')
		
		lzw_prev = None