#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    header = b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\xff\xff\xff\x21\xf9\x04\x01\x0a\x00\x01\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b'

    # Next comes the length of the data section.  It's stored as a 3-byte big-endian
    # integer, so we need to read three bytes from /dev/urandom and interpret those
    # into the file.  This is done by reading three bytes from /dev/urandom and
    # interpreting them as a little-endian unsigned integer.
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    out.write(header + datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.  Each frame is made up of two parts:
    #    - A control block (which tells the decoder what to do with this frame)
    #    - An image descriptor (which tells the decoder where to draw the next frame)
    for _ in range(datalen):

        # Control blocks are always four bytes long.  They start with a single byte which
        # indicates whether or not there should be a delay before displaying the next frame.
        # If the delay flag is set, then the next byte specifies the number of hundredths
        # of a second to wait before showing the next frame.  Otherwise, if the delay flag
        # isn't set, then the next byte is interpreted as a zero and ignored.
        ctrlblock = rng.read(1)[0]
        if ctrlblock & 0x80 == 0x80:
            delay = rng.read(1)[0] * 10
        else:
            delay = 0

        # After the control block comes the image descriptor.  This consists of five bytes.
        # The first byte is the left position of the image.  The second byte is the top
        # position of the image.  The third byte is the width of the image.  The fourth
        # byte is the height of the image.  Finally, the fifth byte is the color table size.
        # For now, we're going to ignore all of these values except for the color table size.
        imgdesc = rng.read(5)

        # Finally, after the image descriptor comes the actual image data.  This is encoded
        # using LZW compression.  To decode the image, we need to know its dimensions, but
        # since we don't have any way to get at that information until we've decoded the
        # image itself, we'll just store the entire image in memory and then dump it when
        # we've finished decoding.
        imgdata = []

        while True:

            # First, we need to read one byte from the input stream.  This byte contains
            # either a literal value or a command.
            cmd = rng.read(1)[0]

            # Literal values are simply copied directly to the output buffer.  So if the
            # current command byte is less than 0x80, then we copy the current byte to
            # the output buffer and continue processing.
            if cmd < 0x80:
                imgdata.append(cmd)
                continue

            # Commands are more complicated.  They consist of a prefix byte followed by
            # a variable number of bytes.  The prefix byte determines how many bytes follow
            # the prefix byte.  There are several different types of commands, each with their
            # own rules about how they work.  For our purposes here, we only care about the
            # RLE command, which repeats a previously seen sequence of bytes a certain number
            # of times.  The RLE command starts with a 0x00 byte followed by another byte.
            # That other byte is called the "index".  The index refers to a previous sequence
            # of bytes that was already written to the output buffer.  The RLE command also
            # contains a final byte which specifies how many times to repeat the indexed bytes.
            elif cmd == 0x00:

                # Read the index and count bytes from the input stream.
                idx = rng.read(1)[0]
                cnt = rng.read(1)[0]

                # Copy the indexed bytes over to the output buffer.
                imgdata += imgdata[idx : idx+cnt]

            # If none of the above conditions were met, then something went wrong.  Print
            # an error message and exit.
            else:
                print('error: invalid command', hex(cmd))
                return

        # Once we've finished decoding the image, we can write it to the output file.
        # First, we need to convert the list of integers into a string of bytes.
        imgstr = ''.join([chr(i) for i in imgdata])

        # Next, we need to compress the image using LZW compression.  Since Python doesn't
        # include a built-in LZW implementation, we'll use a third-party library instead.
        try:
            import lzw
        except ImportError:
            print("error: couldn't find lzw module")
            return

        compressed = lzw.compress(imgstr)

        # Finally, we can write the compressed image to the output file.
        out.write(compressed)


if