#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    header = b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\xff\xff\xff\x21\xf9\x04\x01\x0a\x00\x01\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b'
    out.write(header)

    # Next, we need to figure out how many bytes of random data we should write
    # into the file.  This is done by reading three bytes from /dev/urandom and
    # interpreting them as a little endian unsigned integer.
    num_bytes = int.from_bytes(rng.read(3), byteorder=sys.byteorder, signed=False)

    # Now we know how much data there is, let's write it all at once!
    out.write(os.urandom(num_bytes))

    # Finally, we need to pad the output file up to the end of the last byte of
    # the image data.  To do this, we read one more byte from /dev/urandom and
    # interpret it as a single byte value.  If it's zero, we're good; if not,
    # we'll keep reading until we get a zero or run out of input.
    while True:
        padding_byte = rng.read(1)

        if len(padding_byte)!= 1:
            raise RuntimeError("ran out of padding bytes")

        if padding_byte == b'\x00':
            break

        out.write(padding_byte)


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)