#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: 'random.Random',
        out: BinaryIO):
    """Generate a random GIF file."""

    # Header
    out.write(b'GIF89a')  # magic number (6 bytes)
    width = rng.randint(100, 250)
    height = rng.randint(100, 250)
    flags = b'\x87\x00' if rng.choice([True, False]) else b'\x00\x00'
    bg_color = rng.randint(0, 255)
    aspect_ratio = rng.randint(0, 255)
    out.write(struct.pack('<HHBBB', width, height, *flags))
    out.write(bytes((bg_color, aspect_ratio)))

    # Global color table
    for _ in range(rng.randint(2, 256)):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write(bytes((r, g, b)))

    # Image descriptor
    left = rng.randint(0, width - 1)
    top = rng.randint(0, height - 1)
    flags = b'\x87\x00' if rng.choice([True, False]) else b'\x00\x00'
    local_color_table_flag = \
        1 << 7 if rng.choice([True, False]) else 0
    interlace_flag = 1 << 6 if rng.choice([True, False]) else 0
    sort_flag = 1 << 5 if rng.choice([True, False]) else 0
    reserved_bits = rng.randint(0, 7)
    lct_size = rng.randint(0, 256) | local_color_table_flag
    out.write(struct.pack('<HHBBBBB',
                          left, top, width, height, *flags))
    out.write(bytes((lct_size, interlace_flag, sort_flag, reserved_bits)))

    # Local color table
    for _ in range(lct_size):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        out.write(bytes((r, g, b)))

    # Pixel data
    num_pixels = width * height
    while True:
        block_size = min(num_pixels, rng.randint(2, 254)) + 1
        out.write(bytes((block_size,)))

        for _ in range(block_size):
            palette_index = rng.randint(0, lct_size - 1)
            out.write(bytes((palette_index,)))

            num_extra_bytes = rng.randint(0, 2)
            for _ in range(num_extra_bytes):
                out.write(bytes((rng.randint(0, 255),)))

        num_pixels -= block_size
        if not num_pixels: break

    # Trailer
    out.write(b';')


if __name__ == '__main__':
    import argparse
    import random

    parser = argparse.ArgumentParser()
    parser.add_argument('out', nargs='?', type=argparse.FileType('wb'),
                        default=sys.stdout.buffer)
    args = parser.parse_args()

    with args.out as f:
        rng = random.Random(os.urandom(8))
                generate_random_gif(rng, f)