#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # start writing frames into the file.

    # Header magic number (4 bytes)
    out.write(b'GIF89a')

    # Screen descriptor block (7 bytes)
    width = rng.randint(1024, 5120)
    height = rng.randint(1024, 5120)
    flags = b'\x80\x00' if rng.choice([True, False]) else b'\x00\x00'
    bg_color_index = rng.randint(0, 255)
    pixel_aspect_ratio = b'\xff'
    screen_descriptor = struct.pack('<HHBBB', width, height, flags[0],
                                    flags[1], bg_color_index,
                                    pixel_aspect_ratio)
    out.write(screen_descriptor)

    # Global color table (3 * 256 bytes)
    global_color_table = bytearray()
    for _ in range(256):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        global_color_table += struct.pack('<BBB', r, g, b)
    out.write(global_color_table)

    # Loop control extension (8 bytes)
    loop_control_extension = b'\x21\xFF\x0B\x4E\x45\x54\x53\x43\x41\x50\x45' \
                             b'\x32\x2E\x30\x03\x01\x00\x00\x00'
    out.write(loop_control_extension)

    # Image descriptor block (10 bytes)
    left = rng.randint(0, width - 1)
    top = rng.randint(0, height - 1)
    image_width = rng.randint(left + 1, min(width, left + 128))
    image_height = rng.randint(top + 1, min(height, top + 128))
    flags = b'\x80\x00' if rng.choice([True, False]) else b'\x00\x00'
    local_color_table_flag = b'\x00'
    interlace_flag = b'\x00'
    sort_flag = b'\x00'
    reserved = b'\x00\x00'
    image_descriptor = struct.pack('<HHHHBBBBB', left, top, image_width,
                                   image_height, flags[0], flags[1],
                                   local_color_table_flag, interlace_flag,
                                   sort_flag, reserved[0], reserved[1])
    out.write(image_descriptor)

    # Local color table (3 * 256 bytes)
    local_color_table = bytearray()
    for _ in range(256):
        r = rng.randint(0, 255)
        g = rng.randint(0, 255)
        b = rng.randint(0, 255)
        local_color_table += struct.pack('<BBB', r, g, b)
    out.write(local_color_table)

    # Frame data (variable length)
    lzw_min_code_size = rng.randint(2, 12)
    out.write(struct.pack('<B', lzw_min_code_size))
    frame_data = bytearray()
    while len(frame_data) < 1024:
        num_bytes = rng.randint(1, 128)
        frame_data += rng.read(num_bytes)
    out.write(frame_data)

    # Trailer (3 bytes)
    trailer = b'\x3B\x00\x00'
    out.write(trailer)


if __name__ == '__main__':
    with open('/dev/urandom', 'rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)