#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def read_u8(f: BinaryIO) -> int:
    return struct.unpack('B', f.read(1))[0]


def read_u16(f: BinaryIO) -> int:
    return struct.unpack('<H', f.read(2))[0]


def read_u32(f: BinaryIO) -> int:
    return struct.unpack('<I', f.read(4))[0]


def read_u64(f: BinaryIO) -> int:
    return struct.unpack('<Q', f.read(8))[0]


def write_u8(f: BinaryIO, v: int):
    f.write(struct.pack('B', v))


def write_u16(f: BinaryIO, v: int):
    f.write(struct.pack('<H', v))


def write_u32(f: BinaryIO, v: int):
    f.write(struct.pack('<I', v))


def write_u64(f: BinaryIO, v: int):
    f.write(struct.pack('<Q', v))


def align(f: BinaryIO, alignment=4):
    pos = f.tell()
    if (pos % alignment)!= 0:
        padding = alignment - (pos % alignment)
        f.seek(padding, os.SEEK_CUR)


def main():
    with open(sys.argv[1], 'rb') as f:

        # Read header
        magic = f.read(4).decode('ascii')
        assert magic == 'Yaz0'
        flags = read_u8(f)
        decomp_size = read_u32(f)
        comp_size = read_u32(f)

        print(f"flags={flags:#x}")
        print(f"decomp_size={decomp_size}")
        print(f"comp_size={comp_size}")

        # Read block info table
        num_blocks = read_u16(f)
        block_info_table = []
        for i in range(num_blocks):
            block_info_table += [
                {
                    "offset": read_u32(f),
                    "decomp_size": read_u32(f),
                    "comp_size": read_u32(f),
                }
            ]

        # Read compressed data
        f.seek(0x10 + num_blocks * 0xC, os.SEEK_SET)
        comp_data = bytearray(f.read())

    # Decompress
    out = bytearray(decomp_size)
    src = memoryview(comp_data)
    dst = memoryview(out)
    offset = 0
    for bi in block_info_table:
        size = min(bi['decomp_size'], len(dst) - offset)
        while True:
            ctrl = src[bi['offset']]

            # Literal run
            if ctrl < 0x80:
                count = ctrl >> 5
                offs = ((ctrl & 0x1F) << 8) | src[bi['offset']+1]

                if count == 7:
                    count += src[bi['offset']+2]
                    bi['offset'] += 1

                bi['offset'] += 2

                for _ in range(count):
                    dst[offset] = src[offs]
                    offset += 1
                    offs += 1

                    if offset >= size:
                        break

                continue

            # Backward reference
            if ctrl < 0xE0:
                count = (ctrl & 0xF) + 3
                offs = (((ctrl >> 4) & 1) << 8) | src[bi['offset']+1]

                bi['offset'] += 2

                for _ in range(count):
                    dst[offset] = dst[offset-offs-1]
                    offset += 1

                    if offset >= size:
                        break

                continue

            # End of stream marker
            if ctrl == 0xFD:
                break

            raise Exception("Invalid control code")

    assert offset == len(dst)

    # Write output file
    with open(sys.argv[2], 'wb') as f:
        f.write(out[:comp_size])


if