#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:
    """Generate a random GIF image."""
    
    # First off, let's write the header.
    out.write(b'GIF89a')  # Magic number and version string.
    width = int.from_bytes(rng.read(2), byteorder='little')
    height = int.from_bytes(rng.read(2), byteorder='little')
    out.write((width).to_bytes(2, 'little'))
    out.write((height).to_bytes(2, 'little'))
    flags = b'\x87\x00'   # Global color table flag set + sort flag unset.
    out.write(flags)
    bgcolorindex = rng.read(1)[0] & 0x0F
    out.write(bgcolorindex.to_bytes(1, 'little'))
    pixelaspectratio = b'\x00'
    out.write(pixelaspectratio)

    # Next up is the global colour table. We'll just use the default palette.
    for i in range(0, 256 * 3):
        out.write(rng.read(1))

    # Finally, we need to write the image descriptor.
    left = int.from_bytes(rng.read(2), byteorder='little')
    top = int.from_bytes(rng.read(2), byteorder='little')
    width = int.from_bytes(rng.read(2), byteorder='little')
    height = int.from_bytes(rng.read(2), byteorder='little')
    flags = b'\x80\x00'     # LZW compression flag set.
    out.write(left.to_bytes(2, 'little'))
    out.write(top.to_bytes(2, 'little'))
    out.write(width.to_bytes(2, 'little'))
    out.write(height.to_bytes(2, 'little'))
    out.write(flags)

    # And now comes the fun part! Let's write the compressed data.
    # The first two bytes are the minimum code size (which should be at least
    # 2 bits long). We'll choose 4 bits here.
    out.write(b'\x04')

    # Then comes the actual compressed data.
    datalen = int.from_bytes(
                rng.read(3), byteorder='big')
    out.write(datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    for _ in range(0, datalen):
        out.write(rng.read(1))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)