#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    header = b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\xff\xff\xff\x21\xf9\x04\x01\x0a\x00\x01\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b'
    out.write(header)

    # Next, we need to figure out how many bytes of random data we should write
    # into the file.  This is done by reading three bytes from /dev/urandom and
    # interpreting them as a little-endian unsigned integer.
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    out.write(datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    for i in range(0, datalen):

        # Every other frame, we'll add a single color palette entry at the end of
        # the image.  This makes it easier to spot the boundaries between frames.
        if (i % 50) == 0:

            # First, we need to determine what color this new palette entry should
            # be.  We do this by reading four bytes from /dev/urandom and
            # interpreting them as a big-endian RGB value.
            r = int.from_bytes(
                    rng.read(1), byteorder='big')
            g = int.from_bytes(
                    rng.read(1), byteorder='big')
            b = int.from_bytes(
                    rng.read(1), byteorder='big')
            rgb = [r,g,b]

            # Then we need to write our new palette entry to the output stream.
            out.write(struct.pack('<BBB', *rgb))

        # Finally, we can write one more byte of random data to the output stream.
        out.write(rng.read(1))


if __name__ == '__main__':
    with open('/dev/urandom', mode='rb') as rng:
        with open('test.gif', mode='wb') as f:
            generate_random_gif(rng, f)