#!/usr/bin/env python3

import argparse
import os
import sys
import struct
from typing import BinaryIO


def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ) -> None:

    # The GIF format is a bit weird in that it has a fixed-size header and then
    # an unbounded data section.  We'll just write the header first so we can
    # figure out how long the data section will be later on.
    header = b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\xff\xff\xff\x21\xf9\x04\x01\x0a\x00\x01\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3b'
    out.write(header)

    # Next, we need to figure out how many bytes of random data we should write
    # into the file.  This is done by reading three bytes from /dev/urandom and
    # interpreting them as a little-endian unsigned integer.
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    out.write(datalen.to_bytes(3, 'big'))

    # Now we can write the actual data to the file.
    while datalen > 0:
        chunklen = min(datalen, 512)

        # Each chunk starts with a length field.
        out.write(chunklen.to_bytes(1, 'big'))

        # Then comes the actual data itself.
        out.write(rng.read(chunklen))

        # Finally, we have to pad the chunk up to 256 bytes (the size of one row
        # of pixels).
        padding = 256 - ((chunklen + 1) % 256)
        if padding!= 256:
            out.write(padding.to_bytes(1, 'big'))

        datalen -= chunklen


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--seed', type=int)
    args = parser.parse_args()

    seed = args.seed or os.getpid() ^ int.from_bytes(os.urandom(4), byteorder='little')
    print("Using seed {:#x}".format(seed))
    rng = struct.Struct('<I').pack(seed)

    with open('/dev/urandom', mode='rb') as urandom:
        urandom.seek(struct.Struct('<I').unpack(rng)[0])
        with open('test.gif', mode='wb') as f:
            generate_random_gif(urandom, f)