#!/usr/bin/env python3

import os
import sys
import struct
from typing import BinaryIO

# Generates a random GIF file that can be viewed in most browsers
# Note: this is not a proper pseudorandom generator because it does not use
#       /dev/urandom or any other secure RNGs
#       We just need to make sure there are no obvious patterns in how the bytes
#       are written (specifically in how they are interleaved with each other).
#       I think this is sufficient for now, but we should consider generating
#       actual pseudorandom shellcode instead if we want more entropy.
def generate_random_gif():

    # The maximum size of one color table entry, including an end-of-table tag
    MAX_COLOR_TABLE_ENTRY_SIZE=64*1024
    
    # Generate all our "shellcode"
    black_pixel=[b'\x00'] * 4
    white_pixel=[b'\xff'] * 4
    eot=[b'\x3B'] # End Of Table marker (unused)

    # Our pixel values will be stored in a dictionary so duplicates won't get
    # mixed up between colors and background
    pixels={}
    palette={
        0:black_pixel+white_pixel,
        1:white_pixel+black_pixel,
    }

    # Add some pixels!
    while len(pixels)<MAX_COLOR_TABLE_ENTRY_SIZE-2:

        # Generate two random 2-color palettes
        p0idx=os.urandom(1)[0]%len(palette.keys())
        p1idx=(p0idx+os.urandom(1)[0])%len(palette.keys())

        # Get their values as bytes
        p0=palette[list(palette.keys())[p0idx]]
        p1=palette[list(palette.keys())[p1idx]]

        # Pick two random pixels from these palettes
        px0idx=os.urandom(1)[0]%len(p0)
        px1idx=os.urandom(1)[0]%len(p1)

        # Combine them together and put back into the palette list
        imgpx=p0[px0idx]+p1[px1idx]
        palette[len(palette.keys())]=imgpx

        # Store the new image pixel value
        # Note: technically PNG only allows 256 color images, which means we might hit this case
        # more often than expected, making it harder to detect. But it's unlikely and undetectable.
        pixels[len(pixels)]=imgpx

    # Compute the total length of our color table based on its contents
    # As per the spec, tables must be divisible by 3
    color_table_length=len(pixels)*3
    
    # Write the GIF header itself
    # TODO: support multiple frames
    with open('test.gif','wb') as outfile:
        
        # Magic constant
        outfile.write(b'GIF87a')
        
        # Screen descriptor
        # 0x0c: 1280 x 1024 
        # No global colour map
        # Pixel aspect ratio = 0:1 (square pixels)
        outfile.write(struct.pack('<HHBBB',
            0x0c, 
            0x0000, 
            0x0000, 
            0x80|0x01,
            0,            
        ))
        
        # color table
        for i in range(color_table_length):
            
            # If we have run out of pixels...
            if i>=len(pixels.keys()):
                
                # Fill the rest of the block with EOT markers
                for j in range(i % 3):
                    outfile.write(eot)
                
            else:
                
                # Otherwise write whatever pixel exists at this index
                outfile.write(pixels[i])
            
        # Trailer
        # Image descriptor (same field as screen descriptor)
        outfile.write(struct.pack('<HHBBB',
            0x0c, 
            0x0000, 
            0x0000, 
            0x80|0x01,
            0,            
        ))
                
        # Table start offset
        outfile.write(struct.pack('<H', 
            color_table_length))


if __name__ == '__main__':
    try:
        generate_random_gif(*sys.argv[1:])
    except Exception as err:
        print(f'{err}')