#!/usr/bin/env python3

import os
import re
import sys
import struct
import zlib
from typing import Optional

try:
    from tqdm import trange
except ModuleNotFoundError:
    def trange(*args, **kwargs): return args[0]


# Reads a GIF file from f and returns its content as bytes
def read_gif_file(f: BinaryIO) -> bytes:
    return f.read()


# Reads a single GIF frame from f and decodes its contents appropriately
def load_frame(f: BinaryIO, index: int, global_table: bool) -> bytes:
    assert f.read(5) == b'GIF87a' or \
           f.read(5) == b'GIF89a' # If one of those fails, you have no business reading this...

    blocktype = next(f)
    assert blocktype == b'!' # Block type should be screen descriptor

    width = f.read(2)[::-1]
    height = f.read(2)[::-1]

    # TODO: Could make use of local color tables later but they're not necessary for the exploit
    bg = f.read(1)
    aspect = f.read(1)
    flags = f.read(1)
    backgroundcolor = f.read(1)
    pixelaspectratio = f.read(1)
    # Skip any extra flags, like transparency or interlacing

    if global_table:
        palette = [
            f.read(3 + 1)
            for i in range(int(bg)+1)
        ]
        palette += [(b'\x00',)*(3+1)]*(2**(ord(flags)>>5)-len(palette))
    else:
        assert ord(bg)==2**ord(flags)>>5, 'Invalid color table size'
        palette = []
        for i in range(2**ord(bg)):
            r = f.read(1)
            g = f.read(1)
            b = f.read(1)
            f.read(1)
            palette.append((r,g,b))

    transparency = None
    delaytime = None
    framedata = None
    while True:
        blocktype = next(f)

        if len(blocktype)!= 1: break
        blocktype = blocktype[0]

        if blocktype == 0x21: # Extension introducer

            exttype = next(f)
            if exttype == b'L':
                subblock = next(f)

                if subblock == b'F':
                    framedata = next(f)

                    crc = f.read(1)
                    framedata = framedata[:-1] # Drop the CRC byte

                elif subblock == b'D':
                    duration = f.read(2)
                    assert duration == b'\x00\x00'

                    transparancy = f.read(1)
                    tsortflag = next(f)
                    backgroundindex = f.read(1)

                    crc = f.read(1)

            elif exttype == b'z':
                subblock = next(f)

                if subblock == b'F':
                    framedata = next(f)

                    delimiter = next(f)
                    assert delimiter == b';'

                    try:
                        compressedsize = int(next(f))
                    except ValueError:
                        continue

                    data = b''
                    while len(data) < compressedsize:
                        data += f.read(compressedsize - len(data))
                    framedata = framedata[:-1] # Drop the CRC byte

                    crc = f.read(1)

                    framedata = zlib.decompress(framedata)
                    framedata = framedata[:width*height]
                else:
                    raise RuntimeError("Unsupported subblock")



        elif blocktype == 0x2C:
            assert framedata is None, "Duplicate bitmap"
            xpos = f.read(2)[::-1]
            ypos = f.read(2)[::-1]
            width = f.read(2)[::-1]
            height = f.read(2)[::-1]

            lctsize = next(f)
            lctflag = next(f)


            if lctflag:
                raise NotImplementedError("Local color tables not supported")

            lzwminsize = next(f)


            subimg = f.read(width*height)

            crc = f.read(1)


        elif blocktype == 0x3B:
            break

        else:
            raise RuntimeError("Unknown block type")


    return framedata


# Returns a list containing the contents of each individual layer in the given framedata
def get_layers(framedata: bytes) -> list:
    layers = []

    # TODO: implement interlacing detection
    top, bottom = 0, len(framedata) // 2

    # Add the leftmost plane to the stack
    layers.append([framedata[top:bottom].tolist()])

    offset = 1
    while offset<len(framedata):
        # Get the current layer
        layer = layers[-1][offset % len(layers[-1])]

        # Copy the current layer into a new empty list
        layers.append([])
        for rowi in range(len(layer) // 2):
            newrow = []
            for coli in range(len(layer[0])):
                # Swap the current cell with its mirror position across the middle of the image
                tmp = layer[coli][rowi]
                layer[coli][rowi] = layer[(len(layer)-col-1)][len(layer)//2+rowi]
                layer[(len(layer)-col-1)][len(layer)//2+rowi] = tmp


                # Get the bits for the colors of our mirrored cells
                newrow.append(''.join(str(i) for i in map(int, str(tmp))))
            layers[-1].append(newrow)

        # Move down to the next layer
        offset += 1


    return layers



# Writes the provided framedata back to disk as a valid GIF file named `name`.ext
def write_gif_file(name: str, framedata: bytes, extension=""):
    path = name + "." + extension
    with open(path, 'wb') as f:
        f.write(b'GIF89a')
        # Write out some dummy data for descriptor length, etc.
        f.write(struct.pack(">H", 1))
        f.write(b'\0'*10)
        f.write(b'!,\0\0\0\0\xff\0\0\0\0\0\0\0') # Static example
        f.write(b'!\xf9\0\0\0\0\0\0\0\0\0\0\0\0') # Static example
        f.write(b'!;\0\0\0\0\x0b\0\0\0\0\0\0\0') # Static example

        f.write(b';')

        f.write(framedata)

        f.write(b'\xFF')


# Generates a random GIF file into `out` using the
# given RNG as source of randomness. Included here
# mostly to demonstrate that the vulnerability can
# be triggered by a crafted framedata field in an
# image descriptor
def generate_random_gif(rng: BinaryIO, out: BinaryIO, global_table:bool = False):
    header = b'GIF89a'
    datalen = int.from_bytes(rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header and all bytes preceding the first descriptor
    out.write(header + data[:10])

    # Now we're going to write 256 descriptors at once
    # Using this loop means we avoid making multiple calls to read()
    for _ in trange(trange(257), desc='Generating frames...', unit='frames'):
        # The label says 10 but there are actually two more
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

        # TODO: Should experiment with variations on these values too
        #       see https://github.com/decompme/decompilation-guide/issues/47
        out.write(b'\x2C\x00\x00\x00\x00\x00\x00\xFB\x00\x00\x00')


    last_byte = out.tell() - 1
    # At the moment this code doesn't support appending arbitrary
    # bytes onto the output so we'll have to manually compute where
    # our data will start relative to the rest of the file
    print(f'The generated gif is {last_byte} bytes long.')

    # For now we'll hardcode the number of loops to use, since we don't
    # yet know how much space remains between the final descriptor and EOF
    numloops = 0xC0

    paddinglen = (numloops * 3).to_bytes(size=2, byteorder='little')

    # Write the leftover bytes before the last descriptor
    out.seek(-1, 2)
    out