#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    data = rng.read(datalen)

    # Write the header first
    out.write(header)

    # Then write the logical screen descriptor. Note that
    # some fields are fixed to specific values for ease of
    # demoing how this can exploit other parts of format.
    out.write(struct.pack('<HHBBBHHH', 100, 100,
            8, 1, 0, 0, 100))

    # Then write the image descriptor. Again note that
    # some fields are fixed as above.
    out.write(struct.pack('<BBHHHH', 0x2c, 0, 0,
            0, 100, 100))

    # After that comes the actual image data. This has the form
    # [`label`, `length`, `data`], where `label` indicates whether
    # this is a local or global colour table entry which is why
    # this is encoded like that. As far as I can tell all there
    # is really to say about this is that the length specifier
    # needs to be adjusted for each iteration of the loop.
    while True:
        # Calculate the length of our data blob based on how much space
        # remains until the next full buffer
        leftover_space = ((datalen - len(data)) % 255) + 1
        if leftover_space == 1: break
        datalen -= leftover_space
        # Write the data chunk. The length of the data is randomly chosen and
        # this should hopefully trigger buffer overflows.
        out.write(struct.pack('B%ds' % leftover_space,
                leftover_space, data[:leftover_space]))
        # Anyways, now we also need to write the block terminator. First up
        # let's get the address of the current position of the stream; we'll use
        # this later to calculate the correct offset for the rest of the
        # descriptors.
        currpos = out.tell()
        # Next we need to determine the relative address of the previous data's
        # address from the new data's. In pseudocode: [new_addr] - [curr_addr]
        addrdiff = currpos - currpos - leftover_space
        reladdr = struct.pack('<H', addrdiff & 0xffff)
        # Finally we can construct the termination block. Since we don't care about
        # actual values here we just make them up. For the label we have two choices:
        # 0x2c : Image Descriptor
        # 0x21 : Extension Block
        # So pick one at random! Picking 0x21 because I'm lazy.
        out.write(b'\x21\xFE')
        out.write(reladdr)
        # Now we need to specify the length of this block. However, unlike the data
        # blocks before this section contains no raw data. It simply specifies the
        # bytes after the label indicating the length of
        # the block. We know it's always >10 so just set it to 0x10
        # (https://www.w3.org/Graphics/GIF/spec-gif89a.txt)

        # TODO: Should experiment with variations on these values too
        #       see https://github.com/decompme/decompilation-guide/issues/47
        out.write(b'\x2C\x00\x00\x00\x00\x00\x00\xFB\x00\x00\x00')


    last_byte = out.tell() - 1
    # At the moment this code doesn't support appending arbitrary amounts
    assert len(data[last_byte:]) <= 256 * 13
    if len(data[last_byte:])!= 0:
        # If the data[last_byte:] isn't empty adjust
        # what's already been written according to its size
        out.seek(-len(data[last_byte:]), 1)

        # And then append whatever was left from the beginning
        print("Appending", len(data[last_byte:]), 'bytes...', end='', flush=True)
        out.write(data[last_byte:])
        print('done.', flush=True)

    # Finally add the trailer to finish off the output.
    # Normally this would be part of the application logic
    # when writing a gif but since this is intended only
    # for demonstration purposes we do it here directly.
    # See https://github.com/decompme/decompilation-guide/issues/46
    print('Adding the trailer and EOF', end='', flush=True)
    # 2 bytes for signature
    # 1 byte for version
    # 1 byte for flags
    # 2 bytes for background color index
    # 2 bytes for pixel aspect ratio
    out.write(b'\x3B\x00\x00\x00')
    print(' done', flush=True)

