#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# given RNG as source of randomness. This is mainly here to show
# that the vulnerability can be triggered by a crafted
# framedata field in an image descriptor
def generate_random_gif(rng: BinaryIO, out: BinaryIO):
    # The header has to follow the format described here:
    # https://www.w3.org/Graphics/GIF/spec-gif89a.txt

    # First a 6-character fixed string "GIF87a", which stands for
    # a GIF version 87a, followed by three bytes that indicate the size
    # of the entire file in bytes, most significant byte first.
    # Here we don't care about the exact value because we never need
    # to append anything else to the file later
    header = b'GIF87a' + b'\xFF\xFF\xFF'

    # Next up is a single byte determining whether the global color table
    # should be used (`0x7A`). We'll choose false (`0x41`) here
    block1 = b'\x41'

    # Then comes another 3 bytes that indicates the logical screen width and height
    # in pixels, again with the most significant byte first. Again, we won't mind
    # about the actual value here
    block2 = b'\x01\x00\x01\x00'

    # Following the logical screen specifier follows one or more blocks containing
    # various bits of metadata such as background color index, pixel aspect ratio etc.
    # In practice they aren't needed for exploitation, so I decided not to touch them
    palette = [i & 0xFF for i in range(256)]
    color_table = b''
    for i in range(0x100):
        color_table += bytes([palette[i]])*3

    # Lastly come some variable length fields describing the global color table
    gctable_spec = b'\x00'   # We only want one entry in the GCT, so let's specify 0x00
    gctable_entry = b'\xFF'*6  # That entry should be made entirely opaque

    # Putting everything together gives us the full header
    header += block1 + block2 + color_table + gctable_spec + gctable_entry
    datalen = len(header)

    # Now we're going to write some frames
    for _ in range(257):
        # Since we know what the length of each frame is going to be going in
        # advance, we can calculate its offset from the beginning of the file easily
        frameno = out.tell()-datalen

        # Each frame consists of several parts; a control block, a local color table
        # (if applicable), and then finally the frame data itself
        descriptor_spec = b'\x21\xF9\x04'    # Indicates a graphic control extension
        control_block = b'\x80\x00\x00'     # No disposal effect, no user input required
        lctable_spec = b'\xFE'               # A local color table is present
        lctable = b'\xFF'*6                  # And it must be completely opaque
        data_len = b'\x00\x00'              # How many bytes of frame data is next? (we'll figure out later)
        data = b'\x00'*0                     # Can't be empty, but we'll overwrite it later
    
        # Assemble the complete frame
        descriptor = descriptor_spec + control_block + lctable_spec + lctable + data_len + data
        
        # Generate a bunch of random entropy
        data_len_int = rng.randint(10, 100)
        data = rng.getrandbits(8*data_len_int).to_bytes(length=data_len_int, byteorder='little')
        
        # Finally overwrite the frame data length and frame data fields
        descriptor = descriptor[:3] + data_len + descriptor[-2:] + data
        
        # Actually write the completed frame to disk
        out.write(descriptor)
    
    # After writing all frames and their associated data, we still need to add a few more things:
    # -   A Graphics Control Extension that tells the browser how many times to
    #     repeat those frames when displaying them
    # -   A frame whose contents is sufficient to waste a lot of memory (this makes
    #     it harder for the exploit search tool to find any results)
    # Let's do both right away

    numloops = rnd.randint(0, 2**8)
    
    # Remember that we already wrote part of the header to disk earlier. That's why we
    # have to go back to the beginning of the file to change the number of loops
    out.seek(0)
    paddinglen = (numloops * 3).to_bytes(size=2, byteorder='little')
    out.write(paddinglen)
    out.write(struct.pack('<B', numloops))


