#!/usr/bin/env python3

import struct
from typing import BinaryIO

# Generates a random GIF file into `out` using the
# random number generator `rng` (/dev/urandom)
def generate_random_gif(
        rng: BinaryIO, out: BinaryIO
    ):
    header = b'GIF89a'
    datalen = int.from_bytes(
            rng.read(3), byteorder='big')
    width = 1 << rng.read(1)[0]

    global color_table
    if not color_table:
        num_colors = rng.read(1)[0]+1
        for _ in range((num_colors+1)*3):
            color_table += rng.read(1)

    imgdata = []
    while len(imgdata) < datalen - 4:
        packetlen = min([datalen-4-len(imgdata), 255])

        cmdb = rng.read(1)[0] & 7
        ext = 0x21 <= cmdb and cmdb <= 0x2f or \
              cmdb == 0x3b or \
              cmdb >= 0x80
        
        if ext:
            extcode = ord(rng.read(1))

            if extcode == 1:
                newent = 0
                while newent==0 or newent in used_colors:
                    newent = rng.read(1)[0]

                addcolor = [newent]*3
                for i in range(3):
                    addcolor[i] ^= (color_table[newent*3+i]^255)&255

                color_table += bytes(addcolor)
                used_colors += {newent}

                imgdata += [cmdb, extcode, packetlen//3+1] + addcolor
            else:
                imgdata += [cmdb, extcode, packetlen] + \
                           rng.read(packetlen)
        elif cmdb == 0x2c:
            imgdata += [cmdb,
                         packetlen,
                         rng.read(1)[0],
                         rng.read(1)[0]] + \
                        rng.read(packetlen)
        elif cmdb == 0x21:
            appcode = ord(rng.read(1))
            
            if appcode == 0xf9:
                apppacket = list(rng.read(6))
                
                apppacket[0] |= 0xe0
                apppacket[5] &= ~1 

                imgdata += [cmdb, appcode,
                            bytes(apppacket)] + \
                           rng.read(packetlen - 6)
            else:
                imgdata += [cmdb, appcode,
                            rng.read(packetlen)]
        else:
            imgdata += [cmdb, packetlen] + rng.read(packetlen)
    
    out.write(header + \
              struct.pack('>H', width)+ \
              bytes(used_colors) + \
              255 * (width-len(used_colors)))

    lastbytepos = out.tell()
    out.seek(lastbytepos-4)
    out.write(struct.pack('<I', lastbytepos-12))

    out.write(bytes(imgdata))
    out.write(b'\xff\x00')

# Overwrites every other frame with our own custom
# content as a proof-of-concept.
#
# Note that this is designed to be called from within
# an already running Python interpreter, because we're
# going to be modifying global variables in that context.
def overwrite_frame_data():
    global fds, savedfds  

    states = ['normal','stuck']
    gifnames = ['default', 'overwritten', 
                'og_start', 'og_stop']
    startframes = [None, None, 1, 5]
    stopframes = [None, None, 3, 5]
    frames_to_overwrite = []

    for state in range(len(states)):
        for nameidx in range(len(gifnames)-1):
            frames_to_overwrite += [[state,
                                     nameidx,
                                     1,
                                     True]]

    for sidx,nameidx,stepsize,reverse in frames_to_overwrite:
        # Set up files.
        curfile = open('../gifs/'+gifnames[nameidx+1]+'.gif',
                       mode='rb+')
        newfd = {'file':curfile,'startframe':None,'stopframe':None}
        savedfds[sidx][nameidx] = newfd
        curfile.seek(-1,2)
        newfd['stopframe'] = curfile.tell()-1

        if reverse:
            stepdir=-1
        else:
            stepdir=1

        startfrm = startframes[nameidx]
        for x in range(int(((savedfds[sidx][nameidx]['stopframe']-
                             savedfds[sidx][nameidx]['startframe'])/
                             max([1,stepsize]))/stepdir)):
            frm = startfrm+(x*stepdir*stepsize)
            curfile.seek(frm)
            gifhdr = curfile.read(6)

            curfile.seek(frm)
            if ((framelen := struct.unpack('<H',
                                          curfile.read(2))[0])+
                1 > curfile.seek(0,2)-curfile.tell()):
                break
            curfile.seek(curfile.tell()+1)
            datatocpy = curfile.read(max([1,framelen-1]))
            curfile.seek(frm+12)

            # Prepare for writing by overwriting the current frame with its
            # header.
            curfile.write(b'\x21\xff\x0b\x4e\x45\x54\x53\x43\x41\x50\x45\x32\x2e\x30')

            # Seek to after end of frame, then write our own contents.
            curfile.seek(12+framelen)
            curfile.write(bytearray(gifhdr[:-1]
                                    + b'\x21\xff\x0b\x4e\x45\x54\x53\x43\x41\x50\x45\x32\x2e\x30'))

            # Go back to end of overwritten frame, copy original data forward.
            curfile.seek(12+len(datatocpy))
            curfile.write(datatocpy)

        curfile.close()

        # Calculate next frame's offset relative to beginning
        # of file
        newfd['startframe'] = newfd['stopframe'] + \
                              len(newfd['file'].readline()) + \
                              6  
        print('Wrote'+ str(newfd['startframe']))


    return

# Read in entirety of given file handle, store it in memory,
# then close the handle. Useful if you want to pass a
# raw file descriptor to another process via multiprocessing.Pool.map().
def read_in_chunks(fileobj, chunksize=1024):
    """Generator to read a file piece by piece."""
    try:
        while True:
            chunk = fileobj.read(chunksize)
            if not chunk:
                break
            yield chunk
    finally:
        fileobj.close()

# Generate a new JFIF APP0 segment containing the encrypted version of
# the provided `plaintext`.
def gen_jpeg_segment(
        plaintext: bytes,
        rng: BinaryIO,
        jfif: bool = False
    ) -> bytes:
    # First two bytes are "JFIF" signature.
    segsig = b'\xFF\xE0'

    # Next eight bytes are null-terminated application identifier string.
    segid = b''
    if jfif:
        segid = b'JFIF\x00'
    else:
        segid = b'Adobe\x00'

    # Next byte is zero meaning no thumbnail present.
    segtnb = '\x00'

    # Next three bytes indicate the major version of application note.
    segmajver = b'\x01\x02\x03'

    # Next three bytes indicate minor version.
    segminver = b'\x00\x04\x05'

    # Next three bytes are DPI units.
    segunits = b'\x00\x01\x00'

    # Next six bytes are X density.
    segxdens = b'\x00\x01\x02\x03\x04\x05'

    # Next six bytes are Y density.
    segydens = b'\x00\x01\x02\x03\x04\x05'

    # Next six bytes represent horizontal thumbnail size (?) 
    segthumx = b'\x00\x01\x02\x03\x04\x05'

    # Next six bytes represent vertical thumbnial size (?).
    segthumy = b'\x00\x01\x02\x03\x04\x05'

    # Next byte represents thumbnail compression method.
    segcompr = b'\x01'

    # Next four bytes indicate length of compressed thumbnail.
    segcomprlen = b'\x00\x00\x00\x00'

    # Next four bytes contain encryption key.
    segkey = rng.read(4)

    # Make sure we've got enough plain text to fill the resultant JPEG.
    paddedtxtlen = (((len(plaintext)//16)+
                     (-len(plaintext)%16))+
                     1)*16
    padding = paddedtxt    # Last five bytes are EOI marker (end of image).
    jfif = b'\xff\xd8' + jfif + \
           b'\xff\xd9' 

    # Write out data section of JFIF.
    out.write(jfif[2:-5])

